"""Data downloaded from NCBI Gene data converted into Python data."""

# Copyright (C) 2014-2018, DV Klopfenstein. All rights reserved

downloaded = "2016_02_22" # 21955 items

import collections as cx

NtData = cx.namedtuple('NtData', 'tax_id Org_name GeneID CurrentID Status Symbol Aliases description other_designations map_location chromosome genomic_nucleotide_accession_version start_position_on_the_genomic_accession end_position_on_the_genomic_accession orientation exon_count OMIM no_hdr0')
GeneID2nt = {
  21846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21846, CurrentID=0, Status='live', Symbol='Tie1', Aliases='D430008P04Rik, TIE, tie-1', description='tyrosine kinase with immunoglobulin-like and EGF-like domains 1', other_designations='tyrosine kinase receptor 1', map_location='4 54.67 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118471191, end_position_on_the_genomic_accession=118489849, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  21855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21855, CurrentID=0, Status='live', Symbol='Timm17b', Aliases='17kDa, Sfc3, mTim17b', description='translocase of inner mitochondrial membrane 17b', other_designations='translocator of inner mitochondrial membrane 17 kDa, b|translocator of inner mitochondrial membrane 17b|translocator of inner mitochondrial membrane b', map_location='X 3.56 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7899242, end_position_on_the_genomic_accession=7907652, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  546144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546144, CurrentID=0, Status='live', Symbol='Wdr72', Aliases='AW558070, D230040E23', description='WD repeat domain 72', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=74110331, end_position_on_the_genomic_accession=74283203, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  21869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21869, CurrentID=0, Status='live', Symbol='Nkx2-1', Aliases='AV026640, Nkx2.1, T/EBP, Titf1, Ttf-1', description='NK2 homeobox 1', other_designations='thyroid nuclear factor 1|thyroid transcription factor 1|thyroid-specific enhancer-binding protein|tinman', map_location='12 C1-C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=56531935, end_position_on_the_genomic_accession=56536908, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21873, CurrentID=0, Status='live', Symbol='Tjp2', Aliases='ZO-2', description='tight junction protein 2', other_designations='Tight junction protein ZO-2 (Zonula occludens 2 protein) (Zona occludens 2 protein) (Tight junction protein 2)|zona occludens protein 2|zonula occludens protein 2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24094502, end_position_on_the_genomic_accession=24225128, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  21899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21899, CurrentID=0, Status='live', Symbol='Tlr6', Aliases='', description='toll-like receptor 6', other_designations='', map_location='5 33.54 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=64952030, end_position_on_the_genomic_accession=64971173, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  21925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21925, CurrentID=0, Status='live', Symbol='Tnnc2', Aliases='Tncs', description='troponin C2, fast', other_designations='STNC|troponin C, fast skeletal', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164777161, end_position_on_the_genomic_accession=164779734, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  677296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=677296, CurrentID=0, Status='live', Symbol='Fcrl6', Aliases='ENSMUSG00000070504, FcRH6, mIFGP6, moFcRH6', description='Fc receptor-like 6', other_designations='cell surface glycoprotein gp42|fcR-like protein 6', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172596640, end_position_on_the_genomic_accession=172618322, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  394252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=394252, CurrentID=0, Status='live', Symbol='Serpinb3d', Aliases='Serpinb3l1', description='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 3D', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 3D|serine proteinase inhibitor B3-like 1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107078163, end_position_on_the_genomic_accession=107084125, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  394430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=394430, CurrentID=0, Status='live', Symbol='Ugt1a10', Aliases='A13, Ugt1a13', description='UDP glycosyltransferase 1 family, polypeptide A10', other_designations='UDP glycosyltransferase 1 family polypeptide A13', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88055411, end_position_on_the_genomic_accession=88220002, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  394432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=394432, CurrentID=0, Status='live', Symbol='Ugt1a7c', Aliases="A10', A13, Ugt1a10", description='UDP glucuronosyltransferase 1 family, polypeptide A7C', other_designations='UDP glucuronosyltransferase 1 family, polypeptide A10|UDP glycosyltransferase 1 family, polypeptide A10|UDP-glucuronosyltransferase 1A7C|UDPGT 1-7C|UGT1*7C|UGT1-07C|UGT1.7C', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88095001, end_position_on_the_genomic_accession=88220002, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  394433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=394433, CurrentID=0, Status='live', Symbol='Ugt1a5', Aliases='', description='UDP glucuronosyltransferase 1 family, polypeptide A5', other_designations='UDP glycosyltransferase 1 family, polypeptide A5', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88166012, end_position_on_the_genomic_accession=88220002, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  394434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=394434, CurrentID=0, Status='live', Symbol='Ugt1a9', Aliases='UDPGT, UDPGT 1-9, UGT1-09, UGT1.9, UGTP4, Ugt1a12, mUGTBr/P', description='UDP glucuronosyltransferase 1 family, polypeptide A9', other_designations='UDP glycosyltransferase 1 family polypeptide A12|UDP-glucuronosyltransferase 1-7|UDP-glucuronosyltransferase 1A9', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88070779, end_position_on_the_genomic_accession=88220002, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  394435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=394435, CurrentID=0, Status='live', Symbol='Ugt1a6b', Aliases="A9', Ugt1a9", description='UDP glucuronosyltransferase 1 family, polypeptide A6B', other_designations='UDP glycosyltransferase 1 family polypeptide A9', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88103257, end_position_on_the_genomic_accession=88218998, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  394436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=394436, CurrentID=0, Status='live', Symbol='Ugt1a1', Aliases='Gnt1, UDPGT 1-1, UGT1A01, Udpgt-1a, UgtBr1', description='UDP glucuronosyltransferase 1 family, polypeptide A1', other_designations='UDP-glucuronosyltransferase 1 family, member 1|UDP-glucuronosyltransferase 1 family, polypeptide A1|UDP-glucuronosyltransferase 1A1|UDPGT|UGT1*1|UGT1-01|UGT1.1', map_location='1 44.55 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88211959, end_position_on_the_genomic_accession=88220002, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  263406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=263406, CurrentID=0, Status='live', Symbol='Plekhg3', Aliases='BC030417', description='pleckstrin homology domain containing, family G (with RhoGef domain) member 3', other_designations='PH domain-containing family G member 3', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76533247, end_position_on_the_genomic_accession=76579042, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  263764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=263764, CurrentID=0, Status='live', Symbol='Creg2', Aliases='A830098L22Rik', description='cellular repressor of E1A-stimulated genes 2', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39618406, end_position_on_the_genomic_accession=39651510, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100271704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100271704, CurrentID=0, Status='live', Symbol='Gm17660', Aliases='Scpppq1', description='predicted gene, 17660', other_designations='secretory calcium-binding phosphoprotein proline-glutamine rich 1', map_location='5 50.54 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104070060, end_position_on_the_genomic_accession=104077608, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  263803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=263803, CurrentID=0, Status='live', Symbol='Pkn3', Aliases='AW209115, BC034126', description='protein kinase N3', other_designations='protein kinase PKN-beta|protein kinase PKNbeta|protein-kinase C-related kinase 3', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30077544, end_position_on_the_genomic_accession=30091019, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  100009600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100009600, CurrentID=0, Status='live', Symbol='Zglp1', Aliases='Glp1', description='zinc finger, GATA-like protein 1', other_designations='GATA-like 1|GATA-like protein 1|GLP-1', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21062393, end_position_on_the_genomic_accession=21067093, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  263876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=263876, CurrentID=0, Status='live', Symbol='Spata2', Aliases='AI504642, mKIAA0757', description='spermatogenesis associated 2', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167481136, end_position_on_the_genomic_accession=167492874, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  264064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=264064, CurrentID=0, Status='live', Symbol='Cdk8', Aliases='', description='cyclin-dependent kinase 8', other_designations='cell division protein kinase 8|mediator complex subunit CDK8|mediator of RNA polymerase II transcription subunit CDK8', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146230736, end_position_on_the_genomic_accession=146302874, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  264134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=264134, CurrentID=0, Status='live', Symbol='Ttc26', Aliases='9330141E21Rik, 9430097H08Rik', description='tetratricopeptide repeat domain 26', other_designations='TPR repeat protein 26|protein hop-sterile|tetratricopeptide repeat protein 26', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38381474, end_position_on_the_genomic_accession=38427647, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100534273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100534273, CurrentID=0, Status='live', Symbol='Fer1l5', Aliases='4930533C12Rik', description='fer-1-like 5 (C. elegans)', other_designations='fer-1-like 5 protein', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36372199, end_position_on_the_genomic_accession=36422110, orientation='plus', exon_count=54, OMIM='', no_hdr0=''),
  109569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109569, CurrentID=0, Status='live', Symbol='Rn7s8', Aliases='Rn7s-8', description='7S RNA 8', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109593, CurrentID=0, Status='live', Symbol='Lmo3', Aliases='AI854781, BB106490, Rbtn-3, Rbtn3', description='LIM domain only 3', other_designations='LIM only 3|LMO-3|neuronal-specific transcription factor DAT1', map_location='6 69.02 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=138362918, end_position_on_the_genomic_accession=138581968, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  240672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240672, CurrentID=0, Status='live', Symbol='Dusp5', Aliases='Gm337', description='dual specificity phosphatase 5', other_designations='', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53529318, end_position_on_the_genomic_accession=53541322, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  240675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240675, CurrentID=0, Status='live', Symbol='Vwa2', Aliases='Amaco', description='von Willebrand factor A domain containing 2', other_designations='VWA-like domains, matrilins and collagens|a domain-containing protein similar to matrilin and collagen', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56874416, end_position_on_the_genomic_accession=56912078, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  22258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22258, CurrentID=0, Status='live', Symbol='Usp4', Aliases='F730026I20Rik, Unp, mKIAA4155', description='ubiquitin specific peptidase 4 (proto-oncogene)', other_designations='deubiquitinating enzyme 4|ubiquitin specific protease 4 (proto-oncogene)|ubiquitin thioesterase 4|ubiquitin thiolesterase 4|ubiquitin-specific-processing protease 4|ubiquitous nuclear protein', map_location='9 59.25 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108347820, end_position_on_the_genomic_accession=108392529, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  328108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328108, CurrentID=0, Status='live', Symbol='Fam179b', Aliases='A430041B07Rik, BC062107, C130034K06, mKIAA0423', description='family with sequence similarity 179, member B', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=64965742, end_position_on_the_genomic_accession=65022573, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  264895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=264895, CurrentID=0, Status='live', Symbol='Acsf2', Aliases='', description='acyl-CoA synthetase family member 2', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94557102, end_position_on_the_genomic_accession=94601891, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  102632135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102632135, CurrentID=0, Status='live', Symbol='Gm30289', Aliases='', description='predicted gene, 30289', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40980026, end_position_on_the_genomic_accession=41040207, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  22327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22327, CurrentID=0, Status='live', Symbol='Vbp1', Aliases='VBP-1', description='von Hippel-Lindau binding protein 1', other_designations='VHL-binding protein 1|von Hippel-Lindau-binding protein 1', map_location='X 38.25 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75514278, end_position_on_the_genomic_accession=75534946, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  396184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=396184, CurrentID=0, Status='live', Symbol='Flrt1', Aliases='AW742165, D630040I23Rik', description='fibronectin leucine rich transmembrane protein 1', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7087464, end_position_on_the_genomic_accession=7166030, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  546643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546643, CurrentID=0, Status='live', Symbol='I830127L07Rik', Aliases='', description='RIKEN cDNA I830127L07 gene', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75131292, end_position_on_the_genomic_accession=75135194, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  240817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240817, CurrentID=0, Status='live', Symbol='Teddm2', Aliases='5830403L16Rik, AJ554213, Me9, e92', description='transmembrane epididymal family member 2', other_designations='epididymal protein Me9', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153849542, end_position_on_the_genomic_accession=153851195, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  328234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328234, CurrentID=0, Status='live', Symbol='Rnf182', Aliases='C630023L15Rik', description='ring finger protein 182', other_designations='', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43615158, end_position_on_the_genomic_accession=43670946, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  328250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328250, CurrentID=0, Status='live', Symbol='Nutm2', Aliases='Fam22, Gm806', description='NUT family member 2', other_designations='NUT family member Gm806|protein FAM22', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=50467307, end_position_on_the_genomic_accession=50475355, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  240869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240869, CurrentID=0, Status='live', Symbol='Zbtb37', Aliases='D430004I08Rik', description='zinc finger and BTB domain containing 37', other_designations='', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161008792, end_position_on_the_genomic_accession=161034864, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  109801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109801, CurrentID=0, Status='live', Symbol='Glo1', Aliases='0610009E22Rik, 1110008E19Rik, 2510049H23Rik, AW550643, GLY1, Glo-1, Glo-1r, Glo-1s-r, Glo1-s, Qglo, Glo1', description='glyoxalase 1', other_designations='S-D-lactoylglutathione methylglyoxal lyase|aldoketomutase|glx I|glyoxalase 1 complex|glyoxalase 1 regulatory|glyoxalase 1 structural|glyoxalase I|ketone-aldehyde mutase|methylglyoxalase', map_location='17 15.69 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=30592861, end_position_on_the_genomic_accession=30612659, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  328258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328258, CurrentID=0, Status='live', Symbol='Slc25a48', Aliases='AU042651, E230025K15', description='solute carrier family 25, member 48', other_designations='putative mitochondrial carrier protein FLJ44862 homolog', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56437963, end_position_on_the_genomic_accession=56472367, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  240880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240880, CurrentID=0, Status='live', Symbol='Scyl3', Aliases='1200016D23Rik, 6030457O16, AW214499, Pace1', description='SCY1-like 3 (S. cerevisiae)', other_designations='SCY1-like protein 3|ezrin-binding partner PACE-1|ezrin-binding protein PACE-1', map_location='1 H2.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=163927958, end_position_on_the_genomic_accession=163955126, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  634104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=634104, CurrentID=0, Status='live', Symbol='Olfr287', Aliases='MOR286-1', description='olfactory receptor 287', other_designations='GA_x6K02T2NBG7-5548201-5549154|olfactory receptor MOR286-1', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98204566, end_position_on_the_genomic_accession=98221056, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  240892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240892, CurrentID=0, Status='live', Symbol='Dusp27', Aliases='C130085G02Rik, Gm209', description='dual specificity phosphatase 27 (putative)', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166098148, end_position_on_the_genomic_accession=166127898, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22441, CurrentID=0, Status='live', Symbol='Xlr', Aliases='', description='X-linked lymphocyte-regulated', other_designations='X-linked gene family of B-cell surface antigens|X-linked lymphocyte-regulated complex', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53777118, end_position_on_the_genomic_accession=53797716, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  240921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240921, CurrentID=0, Status='live', Symbol='Gm4955', Aliases='EG240921, Ifi206', description='predicted gene 4955', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173468413, end_position_on_the_genomic_accession=173486872, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  328329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328329, CurrentID=0, Status='live', Symbol='Mast4', Aliases='4930420O11Rik, AI642422, mKIAA0303', description='microtubule associated serine/threonine kinase family member 4', other_designations='mast4', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=102732488, end_position_on_the_genomic_accession=103335840, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  66203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66203, CurrentID=0, Status='live', Symbol='Lce1m', Aliases='1110059L13Rik, AI131654, Lce5a, Sprrl10', description='late cornified envelope 1M', other_designations='late cornified envelope 5A|small proline rich-like 10', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93017807, end_position_on_the_genomic_accession=93019061, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102633458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102633458, CurrentID=0, Status='live', Symbol='Gm11100', Aliases='ENSMUSG00000079357', description='predicted gene 11100', other_designations='', map_location='14 7.08 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=13932894, end_position_on_the_genomic_accession=13949153, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  328370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328370, CurrentID=0, Status='live', Symbol='Rft1', Aliases='D930025H04', description='RFT1 homolog (S. cerevisiae)', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30654341, end_position_on_the_genomic_accession=30691317, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  328381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328381, CurrentID=0, Status='live', Symbol='Sh2d4b', Aliases='A430109M18Rik, D030001E08', description='SH2 domain containing 4B', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=40811521, end_position_on_the_genomic_accession=40902337, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  328401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328401, CurrentID=0, Status='live', Symbol='Rnase9', Aliases='ESRL, Rak1', description='ribonuclease, RNase A family, 9 (non-active)', other_designations='epididymal RNase|ribonuclease A family, member 9|ribonuclease A k1|ribonuclease-like protein 9', map_location='14 B-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51038459, end_position_on_the_genomic_accession=51041905, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  109963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109963, CurrentID=0, Status='live', Symbol='Fre2', Aliases='', description='Friend erythroleukemia 2', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  328417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328417, CurrentID=0, Status='live', Symbol='Parp4', Aliases='Adprtl1, C030027K23Rik, E230037B21Rik, Gm743, PARPL, PH5P, VAULT3, VPARP, p193', description='poly (ADP-ribose) polymerase family, member 4', other_designations='ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase)-like 1', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56575619, end_position_on_the_genomic_accession=56659798, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  266614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266614, CurrentID=0, Status='live', Symbol='Ly6g5b', Aliases='', description='lymphocyte antigen 6 complex, locus G5B', other_designations='Ly6g5b splicing isoform 611|Ly6g5b splicing isoform 707', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35113944, end_position_on_the_genomic_accession=35115400, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  266615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266615, CurrentID=0, Status='live', Symbol='Svet1', Aliases='', description='subventricular expressed transcript 1', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  266620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266620, CurrentID=0, Status='live', Symbol='Defb36', Aliases='BD-36, Defb123, mBD-36', description='defensin beta 36', other_designations='defensin beta 123', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152604327, end_position_on_the_genomic_accession=152612729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  266632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266632, CurrentID=0, Status='live', Symbol='Irak4', Aliases='8430405M07Rik, 9330209D03Rik, IRAK-4, NY-REN-64', description='interleukin-1 receptor-associated kinase 4', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=94543611, end_position_on_the_genomic_accession=94568317, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  266645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266645, CurrentID=0, Status='live', Symbol='Acmsd', Aliases='', description='amino carboxymuconate semialdehyde decarboxylase', other_designations='picolinate carboxylase', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=127729413, end_position_on_the_genomic_accession=127767564, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  790956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=790956, CurrentID=0, Status='live', Symbol='Rs5-8s1', Aliases='', description='5.8S ribosomal RNA', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  266690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266690, CurrentID=0, Status='live', Symbol='Cyb5r4', Aliases='2810034J18Rik, B5+B5R, C79736, Cb5cb5r, Ncb5or, b5/b5r, b5b5r', description='cytochrome b5 reductase 4', other_designations='N-terminal cytochrome b5 and cytochrome b5 oxidoreductase domain-containing protein|NADPH cytochrome B5 oxidoreductase|cb5/cb5R|cytochrome b-type NAD(P)H oxidoreductase|flavohemoprotein b5/b5R', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=87022008, end_position_on_the_genomic_accession=87077774, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  266692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266692, CurrentID=0, Status='live', Symbol='Cpne1', Aliases='1810028N16Rik, mKIAA4108', description='copine I', other_designations='', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156071841, end_position_on_the_genomic_accession=156111965, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  66296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66296, CurrentID=0, Status='live', Symbol='Haus2', Aliases='1700101G24Rik, 2410007P03Rik, AU020217, Cep27', description='HAUS augmin-like complex, subunit 2', other_designations='centrosomal protein 27|centrosomal protein of 27 kDa', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120609382, end_position_on_the_genomic_accession=120621560, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22608, CurrentID=0, Status='live', Symbol='Ybx1', Aliases='1700102N10Rik, C79409, EF1A, MSY1, Nsep1, YB-1, dbpB, mYB-1a', description='Y box protein 1', other_designations='CBF-A|CCAAT-binding transcription factor I subunit A|DNA binding protein B|DNA-binding protein B|EFI-A|Y-box binding protein 1|Y-box transcription factor|Y-box-binding protein 1|enhancer factor I subunit A|nuclease sensitive element binding protein 1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119277327, end_position_on_the_genomic_accession=119294659, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  109990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109990, CurrentID=0, Status='live', Symbol='Saac', Aliases='', description='sialic acid O-acetylation', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  266744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266744, CurrentID=0, Status='live', Symbol='Lgsn', Aliases='Gluld1, Lgs, lengsin', description='lengsin, lens protein with glutamine synthetase domain', other_designations='glutamate-ammonia ligase (glutamine synthase) domain containing 1|glutamate-ammonia ligase (glutamine synthetase) domain containing 1|glutamate-ammonia ligase domain-containing protein 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=31176435, end_position_on_the_genomic_accession=31204725, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  266781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266781, CurrentID=0, Status='live', Symbol='Snx17', Aliases='5830447M19Rik, AI790646, D5Ertd260e, b2b1625.1Clo, mKIAA0064', description='sorting nexin 17', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31193287, end_position_on_the_genomic_accession=31198900, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  791070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791070, CurrentID=0, Status='live', Symbol='Mir744', Aliases='Mirn744, mir-744, mmu-mir-744', description='microRNA 744', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=65734733, end_position_on_the_genomic_accession=65734832, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  791071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791071, CurrentID=0, Status='live', Symbol='Mir758', Aliases='Mirn758, mir-758, mmu-mir-758', description='microRNA 758', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109712810, end_position_on_the_genomic_accession=109712890, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  791073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791073, CurrentID=0, Status='live', Symbol='Mir762', Aliases='Mirn762, mir-762, mmu-mir-762', description='microRNA 762', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127708487, end_position_on_the_genomic_accession=127708562, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  791074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791074, CurrentID=0, Status='live', Symbol='Mir802', Aliases='Mirn802, mir-802, mmu-mir-802', description='microRNA 802', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93369720, end_position_on_the_genomic_accession=93369816, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  791081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791081, CurrentID=0, Status='live', Symbol='Mir764', Aliases='Mirn764, mir-764, mmu-mir-764', description='microRNA 764', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=147002259, end_position_on_the_genomic_accession=147002366, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  266815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=266815, CurrentID=0, Status='live', Symbol='Mill1', Aliases='5530400I18Rik', description='MHC I like leukocyte 1', other_designations='MHC I-like leukocyte 1', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18245232, end_position_on_the_genomic_accession=18266092, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  791260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791260, CurrentID=0, Status='live', Symbol='Tomt', Aliases='Comt2, F930017I19Rik', description='transmembrane O-methyltransferase', other_designations='catechol O-methyltransferase 2|catechol-O-methyltransferase 2', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101899803, end_position_on_the_genomic_accession=101906359, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  267019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=267019, CurrentID=0, Status='live', Symbol='Rps15a', Aliases='A630031B11Rik', description='ribosomal protein S15A', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118104374, end_position_on_the_genomic_accession=118116147, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  791364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791364, CurrentID=0, Status='live', Symbol='Hottip', Aliases='ENSMUSG00000055408, Gm15053, OTTMUSG00000018874', description='Hoxa distal transcript antisense RNA', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52262775, end_position_on_the_genomic_accession=52267603, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  328505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328505, CurrentID=0, Status='live', Symbol='Skint7', Aliases='C130057D23Rik', description='selection and upkeep of intraepithelial T cells 7', other_designations='novel immunoglobulin V-set and CD80-like C2-set immunoglobulin domain containing protein|skint 7|skint-7', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=111972924, end_position_on_the_genomic_accession=111988260, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22687, CurrentID=0, Status='live', Symbol='Zpr1', Aliases='AI303781, Zfp259, Znf259', description='ZPR1 zinc finger', other_designations='zinc finger protein 259', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46273064, end_position_on_the_genomic_accession=46282643, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  22695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22695, CurrentID=0, Status='live', Symbol='Zfp36', Aliases='Gos24, Nup475, TIS11D, TISII, Tis11, Ttp, Zfp-36', description='zinc finger protein 36', other_designations='TPA-induced sequence 11|growth factor-inducible nuclear protein NUP475|tristetraproline', map_location='7 16.72 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28376784, end_position_on_the_genomic_accession=28379228, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22701, CurrentID=0, Status='live', Symbol='Zfp41', Aliases='CTfin92, Zfp-41', description='zinc finger protein 41', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75616142, end_position_on_the_genomic_accession=75627446, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  328563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328563, CurrentID=0, Status='live', Symbol='Apol11b', Aliases='A330102K04Rik', description='apolipoprotein L 11b', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77633951, end_position_on_the_genomic_accession=77643286, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  241196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241196, CurrentID=0, Status='live', Symbol='Serpinb13', Aliases='5430417G24, HUR7, HURPIN, PI13, headpin', description='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 13', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 13', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=106980984, end_position_on_the_genomic_accession=107001195, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100144395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100144395, CurrentID=0, Status='live', Symbol='Nct1', Aliases='', description='non-coding in testis 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100144396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100144396, CurrentID=0, Status='live', Symbol='Nct2', Aliases='', description='non-coding in testis 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241201, CurrentID=0, Status='live', Symbol='Cdh7', Aliases='9330156F07Rik, CDH7L1', description='cadherin 7, type 2', other_designations='', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=109982327, end_position_on_the_genomic_accession=110141917, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  110158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110158, CurrentID=0, Status='live', Symbol='Si-r', Aliases='Suc-1', description='sucrase isomaltase, regulatory', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  231506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231506, CurrentID=0, Status='live', Symbol='Lin54', Aliases='4832414L08, AI461788, mKIAA2037', description='lin-54 homolog (C. elegans)', other_designations='', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100442033, end_position_on_the_genomic_accession=100500649, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  399101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399101, CurrentID=0, Status='live', Symbol='Snhg3', Aliases='Rnu17d, U17HG', description='small nucleolar RNA host gene 3', other_designations='RNA, U17d small nucleolar|small nucleolar RNA host gene (non-protein coding) 3', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132351933, end_position_on_the_genomic_accession=132353686, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102635290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102635290, CurrentID=0, Status='live', Symbol='Ttc39aos1', Aliases='Gm12750, OTTMUSG00000008290, lincRNA-EPS', description='Ttc39a opposite strand RNA 1', other_designations='lincRNA erythroid prosurvival', map_location='4 51.29 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109402273, end_position_on_the_genomic_accession=109406267, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102635415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102635415, CurrentID=0, Status='live', Symbol='Halr1', Aliases='Gm15055, Haunt, Linchoxa1, OTTMUSG00000018881, linc-Evx1, linc-Hoxa1, linc1547', description='Hoxa adjacent long noncoding RNA 1', other_designations='long intergenic non-coding Hoxa1 regulatory RNA', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52102949, end_position_on_the_genomic_accession=52113684, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102635496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102635496, CurrentID=0, Status='live', Symbol='Malrd1', Aliases='', description='MAM and LDL receptor class A domain containing 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=15526479, end_position_on_the_genomic_accession=16255557, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  268281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268281, CurrentID=0, Status='live', Symbol='Shprh', Aliases='2610103K11Rik, AA450458, AU024614, BC006883, D230017O13Rik, E130018M05', description='SNF2 histone linker PHD RING helicase', other_designations='SNF2, histone-linker, PHD and RING finger domain-containing helicase', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=11149321, end_position_on_the_genomic_accession=11217595, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  268288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268288, CurrentID=0, Status='live', Symbol='Samd3', Aliases='Gm623', description='sterile alpha motif domain containing 3', other_designations='SAM domain-containing protein 3', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=26229618, end_position_on_the_genomic_accession=26272916, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  268291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268291, CurrentID=0, Status='live', Symbol='Rnf217', Aliases='AU016819, Ibrdc1', description='ring finger protein 217', other_designations='IBR domain containing 1|IBR domain-containing protein 1', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=31501887, end_position_on_the_genomic_accession=31609725, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  268294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268294, CurrentID=0, Status='live', Symbol='Zbtb24', Aliases='BC055367, Bif1, Bsg1, ZNF450', description='zinc finger and BTB domain containing 24', other_designations='BMP-induced factor 1|bone morphogenetic protein-induced factor 1|brain specific protein 1|zinc finger protein 450', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41450356, end_position_on_the_genomic_accession=41465582, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  268297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268297, CurrentID=0, Status='live', Symbol='Scml4', Aliases='9330161D17, mFLJ00197', description='sex comb on midleg-like 4 (Drosophila)', other_designations='', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=42860317, end_position_on_the_genomic_accession=42960782, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  268301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268301, CurrentID=0, Status='live', Symbol='Sowahc', Aliases='4921515A04Rik, 9430099J10Rik, Ankrd57, C820004L04Rik', description='sosondowah ankyrin repeat domain family member C', other_designations='ankyrin repeat domain 57|ankyrin repeat domain-containing protein 57|protein sosondowah homolog C', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59221922, end_position_on_the_genomic_accession=59226434, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  241327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241327, CurrentID=0, Status='live', Symbol='Olfml2a', Aliases='4932431K08Rik, mFLJ00237', description='olfactomedin-like 2A', other_designations='photomedin-1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38931980, end_position_on_the_genomic_accession=38960585, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  268345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268345, CurrentID=0, Status='live', Symbol='Kcnc2', Aliases='AW047325, B230117I07, KShIIIA, Kv3.2', description='potassium voltage gated channel, Shaw-related subfamily, member 2', other_designations='shaw-like potassium channel|voltage-gated potassium channel Kv3.2', map_location='10 60.3 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=112270459, end_position_on_the_genomic_accession=112466304, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  268354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268354, CurrentID=0, Status='live', Symbol='Fam19a2', Aliases='6330575M02, FAM19A2 TAFA-2, TAFA2, Tafa-2', description='family with sequence similarity 19, member A2', other_designations='chemokine-like protein TAFA-2', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=123264076, end_position_on_the_genomic_accession=123741204, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  268373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268373, CurrentID=0, Status='live', Symbol='Ppia', Aliases='2700098C05, Cphn, CyP-18, CypA', description='peptidylprolyl isomerase A', other_designations='PPIase A|SP18|cyclophilin A|cyclosporin A-binding protein|rotamase A', map_location='11 3.97 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6415870, end_position_on_the_genomic_accession=6419810, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  268379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268379, CurrentID=0, Status='live', Symbol='Abca13', Aliases='9830132L24, A930002G16Rik, AI956815', description='ATP-binding cassette, sub-family A (ABC1), member 13', other_designations='ATP-binding cassette transporter', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=9191354, end_position_on_the_genomic_accession=9684871, orientation='plus', exon_count=63, OMIM='', no_hdr0=''),
  268390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268390, CurrentID=0, Status='live', Symbol='Ahsa2', Aliases='', description='AHA1, activator of heat shock protein ATPase 2', other_designations='', map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23487882, end_position_on_the_genomic_accession=23498039, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  268395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268395, CurrentID=0, Status='live', Symbol='Mpg', Aliases='9830006D05, AI326268, APNG, Aag, Mid1', description='N-methylpurine-DNA glycosylase', other_designations='3-alkyladenine DNA glycosylase|3-methyladenine DNA glycosidase|3-methyladenine DNA glycosylase|ADPG|alkylpurine-DNA-N-glycosylase', map_location='11 18.83 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32226425, end_position_on_the_genomic_accession=32232702, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  268396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268396, CurrentID=0, Status='live', Symbol='Sh3pxd2b', Aliases='G431001E03Rik, TSK4, fad49', description='SH3 and PX domains 2B', other_designations='factor for adipocyte differentiation 49|tyrosine kinase substrate with four SH3 domains', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32347773, end_position_on_the_genomic_accession=32428183, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  268417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268417, CurrentID=0, Status='live', Symbol='Zkscan17', Aliases='Nizp1, Zfp496, Znf496', description='zinc finger with KRAB and SCAN domains 17', other_designations='NSD1-interacting zinc finger protein 1|zinc finger protein with KRAB and SCAN domains 17', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59485520, end_position_on_the_genomic_accession=59506640, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  268420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268420, CurrentID=0, Status='live', Symbol='Alkbh5', Aliases='AW050020, AW215868, Abh5, E130207K11, Ofoxd', description='alkB homolog 5, RNA demethylase', other_designations='alkB, alkylation repair homolog 5|alkylated DNA repair protein alkB homolog 5|alpha-ketoglutarate-dependent dioxygenase alkB homolog 5|oxoglutarate and iron-dependent oxygenase domain containing|probable alpha-ketoglutarate-dependent dioxygenase ABH5', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60537683, end_position_on_the_genomic_accession=60558512, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  399510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399510, CurrentID=0, Status='live', Symbol='Map4k5', Aliases='4432415E19Rik, GCKR, KHS, MAPKKKK5', description='mitogen-activated protein kinase kinase kinase kinase 5', other_designations='MAPK/ERK kinase kinase kinase 5|MEK kinase kinase 5|MEKKK 5|map4k5', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69803757, end_position_on_the_genomic_accession=69893163, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  268445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268445, CurrentID=0, Status='live', Symbol='Ankrd13b', Aliases='AW124583, B930093C12Rik', description='ankyrin repeat domain 13b', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77470485, end_position_on_the_genomic_accession=77489678, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  268448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268448, CurrentID=0, Status='live', Symbol='Phf12', Aliases='2410142K10Rik, F630045O13, PF1, mKIAA1523', description='PHD finger protein 12', other_designations='PHD factor 1|PHD zinc finger transcription factor|phf12', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77982754, end_position_on_the_genomic_accession=78030539, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  268449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268449, CurrentID=0, Status='live', Symbol='Rpl23a', Aliases='BC029892, MDA20', description='ribosomal protein L23A', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78180935, end_position_on_the_genomic_accession=78183584, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  268451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268451, CurrentID=0, Status='live', Symbol='Rab11fip4', Aliases='A730072L08Rik, RAB11-FIP4, mKIAA1821', description='RAB11 family interacting protein 4 (class II)', other_designations='FIP4-Rab11|mRab11-FIP4A', map_location='11 46.88 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79591021, end_position_on_the_genomic_accession=79698023, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  268465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268465, CurrentID=0, Status='live', Symbol='Eme1', Aliases='6820428D13', description='essential meiotic structure-specific endonuclease 1', other_designations='essential meiotic endonuclease 1 homolog 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94644998, end_position_on_the_genomic_accession=94653785, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  268469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268469, CurrentID=0, Status='live', Symbol='Zfp652', Aliases='9530033F24Rik, ZNF652', description='zinc finger protein 652', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95749067, end_position_on_the_genomic_accession=95764715, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  268470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268470, CurrentID=0, Status='live', Symbol='Ube2z', Aliases='AW049199, C030047H17Rik, D11Moh35', description='ubiquitin-conjugating enzyme E2Z', other_designations='E2 ubiquitin-conjugating enzyme Z|uba6-specific E2 conjugating enzyme 1|ubiquitin carrier protein Z|ubiquitin-conjugating enzyme E2Z (putative)|ubiquitin-protein ligase Z|use1', map_location='11 59.28 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96047431, end_position_on_the_genomic_accession=96065364, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  399548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399548, CurrentID=0, Status='live', Symbol='Scn4b', Aliases='Gm1471', description='sodium channel, type IV, beta', other_designations='sodium channel accessory subunit|sodium channel beta 4 subunit|sodium channel, type IV, beta polypeptide|voltage gated sodium channel beta 4 subunit', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45139042, end_position_on_the_genomic_accession=45154061, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  399549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399549, CurrentID=0, Status='live', Symbol='H2-M10.6', Aliases='', description='histocompatibility 2, M region locus 10.6', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36812175, end_position_on_the_genomic_accession=36815566, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  268480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268480, CurrentID=0, Status='live', Symbol='Rapgefl1', Aliases='B230105J10', description='Rap guanine nucleotide exchange factor (GEF)-like 1', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98836785, end_position_on_the_genomic_accession=98853006, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  268481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268481, CurrentID=0, Status='live', Symbol='Krt222', Aliases='6330509G02Rik', description='keratin 222', other_designations='keratin-222', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99232761, end_position_on_the_genomic_accession=99244281, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  268482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268482, CurrentID=0, Status='live', Symbol='Krt12', Aliases='AI835216, K12, Krt-12, Krt1-12', description='keratin 12', other_designations='CK-12|Keratin, type I cytoskeletal 12 (Cytokeratin 12)|cytokeratin-12|keratin complex 1, acidic, gene 12|keratin-12', map_location='11 62.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99415664, end_position_on_the_genomic_accession=99422593, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  399558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399558, CurrentID=0, Status='live', Symbol='Flrt2', Aliases='', description='fibronectin leucine rich transmembrane protein 2', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=95692038, end_position_on_the_genomic_accession=95785215, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  268490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268490, CurrentID=0, Status='live', Symbol='Lsm12', Aliases='1110032E16Rik, 1110059P07, 2600001B17Rik', description='LSM12 homolog', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102163489, end_position_on_the_genomic_accession=102186270, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  268491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268491, CurrentID=0, Status='live', Symbol='Meioc', Aliases='Gm1564, Gm663', description='meiosis specific with coiled-coil domain', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102664966, end_position_on_the_genomic_accession=102682240, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  399566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399566, CurrentID=0, Status='live', Symbol='Btbd6', Aliases='', description='BTB (POZ) domain containing 6', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112974998, end_position_on_the_genomic_accession=112979493, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  399568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399568, CurrentID=0, Status='live', Symbol='BC052040', Aliases='', description='cDNA sequence BC052040', other_designations='', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=115581618, end_position_on_the_genomic_accession=115778768, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  268510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268510, CurrentID=0, Status='live', Symbol='Mgat5b', Aliases='C330018B01, GnT-IX, mGnTVB', description='mannoside acetylglucosaminyltransferase 5, isoenzyme B', other_designations='GNT-Vb|N-acetylglucosaminyl-transferase Vb|N-acetylglucosaminyltranferase VB|N-acetylglucosaminyltransferase IX|UDP-N-acetylglucosamine: alpha1,6-D-mannoside beta1,6-N-acetylglucosaminyltransferase b|alpha-mannoside beta-1,6-N-acetylglucosaminyltransferase B|beta1,6-N-acetylglucosaminyltransferase IX (GnT-IX)|glcNAc-T Vb|mannoside acetylglucosaminyltransferase 5B', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116918856, end_position_on_the_genomic_accession=116986944, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  268512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268512, CurrentID=0, Status='live', Symbol='Slc26a11', Aliases='F630021I08Rik', description='solute carrier family 26, member 11', other_designations='anion exchanger Slc26a11|novel sulfate transporter protein (Slc26a11)|slc26a11', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119355532, end_position_on_the_genomic_accession=119381076, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  268515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268515, CurrentID=0, Status='live', Symbol='Bahcc1', Aliases='B930044J06, BC060615', description='BAH domain and coiled-coil containing 1', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120232904, end_position_on_the_genomic_accession=120292297, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  399591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399591, CurrentID=0, Status='live', Symbol='Tmsb15l', Aliases='4930488E11Rik, Tb15r, Tmsb15b1-Tmsb15b2, Tmsb15r', description='thymosin beta 15b like', other_designations='Tmsb15b1-Tmsb15b2 readthrough transcript', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136954988, end_position_on_the_genomic_accession=136976874, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  399599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399599, CurrentID=0, Status='live', Symbol='Ccdc87', Aliases='4931419P11Rik', description='coiled-coil domain containing 87', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4839366, end_position_on_the_genomic_accession=4842528, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  399603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399603, CurrentID=0, Status='live', Symbol='Fam84b', Aliases='BF642811, D330050I23Rik', description='family with sequence similarity 84, member B', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=60818996, end_position_on_the_genomic_accession=60825080, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  268534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268534, CurrentID=0, Status='live', Symbol='Sntg2', Aliases='2210008K22Rik, 9530013L23, BB121248', description='syntrophin, gamma 2', other_designations='G2SYN|SYN5|syntrophin-5', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=30174480, end_position_on_the_genomic_accession=30373392, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  268564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268564, CurrentID=0, Status='live', Symbol='Zbtb1', Aliases='9430077A12, AI256369, Bsg2, C430003J21Rik', description='zinc finger and BTB domain containing 1', other_designations='brain specific protein 2', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76370192, end_position_on_the_genomic_accession=76396955, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  268566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268566, CurrentID=0, Status='live', Symbol='Gphn', Aliases='5730552E08Rik, AI662856, BC027112, C230040D23, GPH, GPHRYN, geph', description='gephyrin', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=78224153, end_position_on_the_genomic_accession=78684772, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  268567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268567, CurrentID=0, Status='live', Symbol='Tmem229b', Aliases='6330442E10Rik, 9830138K24, AU019070', description='transmembrane protein 229B', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=78961795, end_position_on_the_genomic_accession=79007659, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  268591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268591, CurrentID=0, Status='live', Symbol='Serpina5', Aliases='4933415L04, PAI-3, Pci', description='serine (or cysteine) peptidase inhibitor, clade A, member 5', other_designations='PAI3|alpha-1 antiproteinase|antitrypsin|plasminogen activator inhibitor 3|plasminogen activator inhibitor-3|protein C inhibitor|serine (or cysteine) proteinase inhibitor, clade A, member 5|serpin A5', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104099313, end_position_on_the_genomic_accession=104106138, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  268595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268595, CurrentID=0, Status='live', Symbol='D430019H16Rik', Aliases='AI852661', description='RIKEN cDNA D430019H16 gene', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105453856, end_position_on_the_genomic_accession=105493095, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  399673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399673, CurrentID=0, Status='live', Symbol='Tdpoz2', Aliases='', description='TD and POZ domain containing 2', other_designations='TD and POZ domain containing gene 2|tdpoz2', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93651542, end_position_on_the_genomic_accession=93652686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  399674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399674, CurrentID=0, Status='live', Symbol='Tdpoz3', Aliases='', description='TD and POZ domain containing 3', other_designations='TD and POZ domain containing gene 3|tdpoz3', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93826020, end_position_on_the_genomic_accession=93827117, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  399675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399675, CurrentID=0, Status='live', Symbol='Tdpoz4', Aliases='', description='TD and POZ domain containing 4', other_designations='tdpoz4', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93796398, end_position_on_the_genomic_accession=93797510, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  399676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=399676, CurrentID=0, Status='live', Symbol='Tdpoz5', Aliases='', description='TD and POZ domain containing 5', other_designations='TD and POZ domain containing gene 5|tdpoz5', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93960292, end_position_on_the_genomic_accession=93961314, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  268656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268656, CurrentID=0, Status='live', Symbol='Sptlc1', Aliases='AW552086, C77762, E030036H05, Lcb1', description='serine palmitoyltransferase, long chain base subunit 1', other_designations='LCB 1|SPT 1|SPT1|long chain base biosynthesis protein 1|serine-palmitoyl-CoA transferase 1', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=53332748, end_position_on_the_genomic_accession=53377417, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  268663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268663, CurrentID=0, Status='live', Symbol='Cdhr2', Aliases='Gm624, PC-LKC, Pcdh24', description='cadherin-related family member 2', other_designations='protocadherin 24|protocadherin LKC', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54701463, end_position_on_the_genomic_accession=54736662, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  268670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268670, CurrentID=0, Status='live', Symbol='Zfp759', Aliases='BC028265, Rslcan-8, Rslcan8', description='zinc finger protein 759', other_designations='regulator of sex-limitation candidate 8', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67125128, end_position_on_the_genomic_accession=67142404, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  268686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268686, CurrentID=0, Status='live', Symbol='S100z', Aliases='Gm625', description='S100 calcium binding protein, zeta', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95477266, end_position_on_the_genomic_accession=95479379, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  268697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268697, CurrentID=0, Status='live', Symbol='Ccnb1', Aliases='Ccnb1-rs1-rs13, CycB1, Cycb-4, Cycb-5, Cycb1-rs1, Ccnb1', description='cyclin B1', other_designations='G2/mitotic-specific cyclin B1|cyclin B1, related sequence 1|cyclin B1, related sequence 13', map_location='13 53.23 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100778732, end_position_on_the_genomic_accession=100786543, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110319, CurrentID=0, Status='live', Symbol='Tcrb-V8', Aliases='', description='T cell receptor beta, variable 8', other_designations='T-cell receptor beta, variable 8', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  268706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268706, CurrentID=0, Status='live', Symbol='Slc38a9', Aliases='4833412L08Rik, 6720411P22Rik, 9130023D20Rik, 9430067K09Rik, A730092C09', description='solute carrier family 38, member 9', other_designations='putative sodium-coupled neutral amino acid transporter 9|solute carrier family 38 member 9', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112660691, end_position_on_the_genomic_accession=112738752, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  268709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268709, CurrentID=0, Status='live', Symbol='Fam107a', Aliases='DRR1', description='family with sequence similarity 107, member A', other_designations='', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8296270, end_position_on_the_genomic_accession=8318018, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  268721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268721, CurrentID=0, Status='live', Symbol='Zswim8', Aliases='2310021P13Rik, 4832404P21Rik, Kiaa0913, mKIAA0913', description='zinc finger SWIM-type containing 8', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20707319, end_position_on_the_genomic_accession=20723619, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  268729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268729, CurrentID=0, Status='live', Symbol='Gm626', Aliases='ENSMUSG00000071536, Frmpd2, Gm1582', description='predicted gene 626', other_designations='FERM and PDZ domain containing 2', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33471696, end_position_on_the_genomic_accession=33575269, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  66634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66634, CurrentID=0, Status='live', Symbol='Mcm8', Aliases='5730432L01Rik', description='minichromosome maintenance 8 homologous recombination repair factor', other_designations='DNA replication licensing factor MCM8|minichromosome maintenance 8|minichromosome maintenance deficient 8', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132816141, end_position_on_the_genomic_accession=132844197, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  268739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268739, CurrentID=0, Status='live', Symbol='Arhgef40', Aliases='E130112L23Rik, Gm669, Solo', description='Rho guanine nucleotide exchange factor (GEF) 40', other_designations='protein SOLO', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51984681, end_position_on_the_genomic_accession=52006247, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  268741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268741, CurrentID=0, Status='live', Symbol='Tox4', Aliases='5730589K01Rik, A630040M18, AA410149, LCP1', description='TOX high mobility group box family member 4', other_designations='epidermal Langerhans cell protein LCP1', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52279146, end_position_on_the_genomic_accession=52295509, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  268747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268747, CurrentID=0, Status='live', Symbol='Lrrc16b', Aliases='mFLJ00240', description='leucine rich repeat containing 16B', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55490656, end_position_on_the_genomic_accession=55508272, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  268749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268749, CurrentID=0, Status='live', Symbol='Rnf31', Aliases='AL033293, BC031509, Flj10111, HOIP, Paul, mFLJ00217', description='ring finger protein 31', other_designations='HOIL-1-interacting protein|putative Ariadne-like ubiquitin ligase', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55591713, end_position_on_the_genomic_accession=55603671, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  268752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268752, CurrentID=0, Status='live', Symbol='Wdfy2', Aliases='5830485M08, B130024L21Rik, ProF, ZFYVE22', description='WD repeat and FYVE domain containing 2', other_designations='WD40- and FYVE domain-containing protein 2|propeller-FYVE protein', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62837690, end_position_on_the_genomic_accession=62961509, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  268755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268755, CurrentID=0, Status='live', Symbol='Mir124a-1hg', Aliases='A930011O12Rik, Rncr3, mir-124-1, mir-3078', description='Mir124-1 host gene (non-protein coding)', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64588115, end_position_on_the_genomic_accession=64593961, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  268756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268756, CurrentID=0, Status='live', Symbol='Gulo', Aliases='5730581M22, AU018375, BC028822, GLO, LGO, sfx, unh, unhip', description='gulonolactone (L-) oxidase', other_designations='L-gulono-gamma-lactone oxidase', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65986786, end_position_on_the_genomic_accession=66009254, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  268759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268759, CurrentID=0, Status='live', Symbol='9930012K11Rik', Aliases='', description='RIKEN cDNA 9930012K11 gene', other_designations='', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70154405, end_position_on_the_genomic_accession=70159502, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  268780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268780, CurrentID=0, Status='live', Symbol='Egflam', Aliases='5930412K08, AU040377', description='EGF-like, fibronectin type III and laminin G domains', other_designations='EGF-like, fibronectin type-III and laminin G-like domain-containing protein|nectican', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=7206120, end_position_on_the_genomic_accession=7398746, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  268782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268782, CurrentID=0, Status='live', Symbol='Agxt2', Aliases='AI303810, AI663818', description='alanine-glyoxylate aminotransferase 2', other_designations='(R)-3-amino-2-methylpropionate--pyruvate transaminase|AGT 2|D-AIBAT|beta-ALAAT II|beta-alanine-pyruvate aminotransferase', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10358516, end_position_on_the_genomic_accession=10410159, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  268783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268783, CurrentID=0, Status='live', Symbol='Mtmr12', Aliases='3Pap, 4932703C11, C730015A02Rik, Pip3ap, mKIAA1682', description='myotubularin related protein 12', other_designations='phosphatidylinositol-3 phosphate 3-phosphatase adaptor subunit|phosphatidylinositol-3-phosphatase associated protein', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=12205094, end_position_on_the_genomic_accession=12272240, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  328789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328789, CurrentID=0, Status='live', Symbol='Lhfpl5', Aliases='9330179O15Rik, Tmhs, hscy', description='lipoma HMGIC fusion partner-like 5', other_designations='lipoma HMGIC fusion partner-like 5 protein|tetraspan transmembrane protein, hair cell stereocilia', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28575371, end_position_on_the_genomic_accession=28583593, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  268807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268807, CurrentID=0, Status='live', Symbol='Klhl38', Aliases='8230402K04Rik', description='kelch-like 38', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58314573, end_position_on_the_genomic_accession=58324169, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  268822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268822, CurrentID=0, Status='live', Symbol='Adck5', Aliases='A230108P17', description='aarF domain containing kinase 5', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76576359, end_position_on_the_genomic_accession=76595812, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  268857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268857, CurrentID=0, Status='live', Symbol='Nlrc3', Aliases='CLR16.2, D230007K08Rik, mFLJ00348', description='NLR family, CARD domain containing 3', other_designations='Caterpiller 16.2|caterpillar 16.2', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3943807, end_position_on_the_genomic_accession=3979032, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  268859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268859, CurrentID=0, Status='live', Symbol='Rbfox1', Aliases='A2bp, A2bp1, Hrnbp1, fox-1', description='RNA binding protein, fox-1 homolog (C. elegans) 1', other_designations='ataxin 2 binding protein 1|ataxin-2-binding protein 1|fox-1 homolog A|hexaribonucleotide binding protein 1', map_location='16 3.34 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5884793, end_position_on_the_genomic_accession=7412480, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  268860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268860, CurrentID=0, Status='live', Symbol='Abat', Aliases='9630038C02Rik, AI255750, ENSMUSG00000051226, Gabaat, Gabat, Gm9851, I54, Laibat, X61497', description='4-aminobutyrate aminotransferase', other_designations='(S)-3-amino-2-methylpropionate transaminase|4-aminobutyrate transaminase|GABA aminotransferase|GABA transaminase|GABA transferase|beta-alanine oxoglutarate aminotransferase|gamma-amino-N-butyrate transaminase', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8513429, end_position_on_the_genomic_accession=8621568, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  268878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268878, CurrentID=0, Status='live', Symbol='Atp13a5', Aliases='C630015F21Rik', description='ATPase type 13A5', other_designations='P5-ATPase', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=29231851, end_position_on_the_genomic_accession=29378764, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  268880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268880, CurrentID=0, Status='live', Symbol='Xxylt1', Aliases='AI480653', description='xyloside xylosyltransferase 1', other_designations='UDP-xylose:alpha-xyloside alpha-1,3-xylosyltransferase', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30928348, end_position_on_the_genomic_accession=31081551, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  268882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268882, CurrentID=0, Status='live', Symbol='Fbxo45', Aliases='2610017J04Rik, AI463119, BC026799', description='F-box protein 45', other_designations='F-box only protein 45|fbxo45|mFbxo45', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32230112, end_position_on_the_genomic_accession=32247025, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  268885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268885, CurrentID=0, Status='live', Symbol='Stfa2l1', Aliases='', description='stefin A2 like 1', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36156811, end_position_on_the_genomic_accession=36161948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  268890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268890, CurrentID=0, Status='live', Symbol='Lsamp', Aliases='5430428I19, AW046674, B130007O04Rik, D930023J12Rik, Lam, Lamp', description='limbic system-associated membrane protein', other_designations='', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=41532788, end_position_on_the_genomic_accession=42181679, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102636131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102636131, CurrentID=0, Status='live', Symbol='Gm13067', Aliases='OTTMUSG00000010272', description='predicted gene 13067', other_designations='', map_location='4 80.81 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150029405, end_position_on_the_genomic_accession=150058717, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  268902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268902, CurrentID=0, Status='live', Symbol='Robo2', Aliases='2600013A04Rik, 9430089E08Rik, BB097918, D230004I22Rik, mKIAA1568', description='roundabout homolog 2 (Drosophila)', other_designations='', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=73892061, end_position_on_the_genomic_accession=75447298, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  268903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268903, CurrentID=0, Status='live', Symbol='Nrip1', Aliases='6030458L20Rik, 8430438I05Rik, 9630050P12, AA959574, AW456757, RIP140', description='nuclear receptor interacting protein 1', other_designations='nrip1|nuclear factor RIP140|receptor-interacting protein 140', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=76287403, end_position_on_the_genomic_accession=76374195, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  268930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268930, CurrentID=0, Status='live', Symbol='Pkmyt1', Aliases='6230424P17, AW209059, Myt1', description='protein kinase, membrane associated tyrosine/threonine 1', other_designations='MYT1 kinase|membrane-associated tyrosine-and threonine-specific cdc2-inhibitory kinase', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23726336, end_position_on_the_genomic_accession=23736735, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  268932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268932, CurrentID=0, Status='live', Symbol='Caskin1', Aliases='3300002N10Rik, C130061I24, C630036E02Rik, mKIAA1306', description='CASK interacting protein 1', other_designations='CASK-interacting protein 1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24487852, end_position_on_the_genomic_accession=24508909, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  268933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268933, CurrentID=0, Status='live', Symbol='Wdr24', Aliases='BC037651', description='WD repeat domain 24', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25823627, end_position_on_the_genomic_accession=25828730, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  268934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268934, CurrentID=0, Status='live', Symbol='Grm4', Aliases='Gprc1d, mGluR4', description='glutamate receptor, metabotropic 4', other_designations='G protein-coupled receptor GRM4|G protein-coupled receptor, family C, group 1, member D', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27422387, end_position_on_the_genomic_accession=27513679, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  268935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268935, CurrentID=0, Status='live', Symbol='Scube3', Aliases='CEGF3, D030038I21Rik', description='signal peptide, CUB domain, EGF-like 3', other_designations='scube3|signal peptide, CUB and EGF-like domain containing protein 3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28142100, end_position_on_the_genomic_accession=28174851, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  268936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268936, CurrentID=0, Status='live', Symbol='Brpf3', Aliases='AI413466, AW493400, mKIAA1286', description='bromodomain and PHD finger containing, 3', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28800557, end_position_on_the_genomic_accession=28838789, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  268949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268949, CurrentID=0, Status='live', Symbol='Dpcr1', Aliases='Gm630', description='diffuse panbronchiolitis critical region 1 (human)', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35635755, end_position_on_the_genomic_accession=35643695, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  268958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268958, CurrentID=0, Status='live', Symbol='Capn11', Aliases='', description='calpain 11', other_designations='CANP 11|calcium-activated neutral proteinase 11|calcium-dependent thiol protease|capn11', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45630191, end_position_on_the_genomic_accession=45660090, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  268970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268970, CurrentID=0, Status='live', Symbol='Arhgap28', Aliases='AU044757, AW550892, E130310N06', description='Rho GTPase activating protein 28', other_designations='rho-type GTPase-activating protein 28', map_location='17 E1.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=67842706, end_position_on_the_genomic_accession=68004108, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  268973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268973, CurrentID=0, Status='live', Symbol='Nlrc4', Aliases='9530011P19Rik, CLAN, CLAN1, CLANA, CLANB, CLANC, CLAND, Card12, IPAF', description='NLR family, CARD domain containing 4', other_designations='caspase recruitment domain family, member 12|caspase recruitment domain-containing protein 12|ice protease-activating factor', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74425716, end_position_on_the_genomic_accession=74459143, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  268977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268977, CurrentID=0, Status='live', Symbol='Ltbp1', Aliases='9430031G15Rik, 9830146M04, Ltbp-1, Tgfb, b2b1000Clo', description='latent transforming growth factor beta binding protein 1', other_designations='TGF-beta1-BP-1|latent TGF beta binding protein|latent transforming growth factor beta binding protein 1L|transforming growth factor beta-1-binding protein 1', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=75005529, end_position_on_the_genomic_accession=75392967, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  268980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268980, CurrentID=0, Status='live', Symbol='Strn', Aliases='AU022939, D130055P19', description='striatin, calmodulin binding protein', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78649908, end_position_on_the_genomic_accession=78737191, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  268996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268996, CurrentID=0, Status='live', Symbol='Ss18', Aliases='D130059H17, Ssxt, Syt', description='synovial sarcoma translocation, Chromosome 18', other_designations='synovial sarcoma associated SS18-beta|synovial sarcoma associated SS18-delta|synovial sarcoma associated SS18-gamma|synovial sarcoma, translocated to X chromosome|synovial sarcoma-associated Ss18-alpha', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=14625629, end_position_on_the_genomic_accession=14683002, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  269003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269003, CurrentID=0, Status='live', Symbol='Sap130', Aliases='2610304F09Rik, 6720406D06', description='Sin3A associated protein', other_designations='130 kDa Sin3-associated polypeptide|Sin3A associated protein, 130kDa|sin3-associated polypeptide p130', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31634327, end_position_on_the_genomic_accession=31723068, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  269016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269016, CurrentID=0, Status='live', Symbol='Sh3rf2', Aliases='2310046K19, 9130023G24Rik, AI452083, Ppp1r39, RNF158', description='SH3 domain containing ring finger 2', other_designations='RING finger protein 158|SH3 domain-containing RING finger protein 2|protein phosphatase 1 regulatory subunit 39', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42051743, end_position_on_the_genomic_accession=42158695, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  269019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269019, CurrentID=0, Status='live', Symbol='Stk32a', Aliases='A930015B13Rik, YANK1', description='serine/threonine kinase 32A', other_designations='stk32a', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=43207697, end_position_on_the_genomic_accession=43317481, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  269023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269023, CurrentID=0, Status='live', Symbol='Zfp608', Aliases='4932417D18Rik, D430007A19Rik, Znf608', description='zinc finger protein 608', other_designations='', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=54888045, end_position_on_the_genomic_accession=54992598, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  102636264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102636264, CurrentID=0, Status='live', Symbol='Gm33378', Aliases='', description='predicted gene, 33378', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82011695, end_position_on_the_genomic_accession=82016869, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  269033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269033, CurrentID=0, Status='live', Symbol='4930503L19Rik', Aliases='9930107F24, BB148262, Lars, Las2', description='RIKEN cDNA 4930503L19 gene', other_designations='', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=70452210, end_position_on_the_genomic_accession=70472504, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  269037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269037, CurrentID=0, Status='live', Symbol='Ctif', Aliases='Gm672', description='CBP80/20-dependent translation initiation factor', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=75431205, end_position_on_the_genomic_accession=75697696, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66686, CurrentID=0, Status='live', Symbol='Dcbld1', Aliases='4631413K11Rik', description='discoidin, CUB and LCCL domain containing 1', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52233584, end_position_on_the_genomic_accession=52321378, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  269053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269053, CurrentID=0, Status='live', Symbol='Gpr152', Aliases='A930009H15Rik, Gm673', description='G protein-coupled receptor 152', other_designations='putative G-protein coupled receptor', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4139799, end_position_on_the_genomic_accession=4145741, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  269060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269060, CurrentID=0, Status='live', Symbol='Dagla', Aliases='Nsddr', description='diacylglycerol lipase, alpha', other_designations='DGL-alpha|neural stem cell-derived dendrite regulator', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10245265, end_position_on_the_genomic_accession=10304962, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  269061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269061, CurrentID=0, Status='live', Symbol='Cpsf7', Aliases='5730453I16Rik, AL022757, C330017N18Rik', description='cleavage and polyadenylation specific factor 7', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10525244, end_position_on_the_genomic_accession=10547735, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  66689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66689, CurrentID=0, Status='live', Symbol='Klhl28', Aliases='2810440N09Rik, 4122402F11Rik, 4931401E10Rik, Btbd5', description='kelch-like 28', other_designations='BTB (POZ) domain containing 5|BTB/POZ domain-containing protein 5', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=64942198, end_position_on_the_genomic_accession=64965605, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  110380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110380, CurrentID=0, Status='live', Symbol='Shroom2', Aliases='4832440C16, Apxl, C630003H05Rik, Shrm2', description='shroom family member 2', other_designations='', map_location='X 68.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152609509, end_position_on_the_genomic_accession=152770251, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  269109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269109, CurrentID=0, Status='live', Symbol='Dpp10', Aliases='6430601K09Rik, DPP X, Dprp3', description='dipeptidylpeptidase 10', other_designations='dipeptidyl peptidase X', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=123332138, end_position_on_the_genomic_accession=124845960, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  269113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269113, CurrentID=0, Status='live', Symbol='Nup54', Aliases='3110079L04Rik, 54kDa', description='nucleoporin 54', other_designations='54 kDa nucleoporin|nucleoporin 54kDa|nucleoporin Nup54', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92415540, end_position_on_the_genomic_accession=92435265, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  269116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269116, CurrentID=0, Status='live', Symbol='Nfasc', Aliases='AA387016, D430023G06Rik, NF, mKIAA0756', description='neurofascin', other_designations='', map_location='1 57.42 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132564690, end_position_on_the_genomic_accession=132741834, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  269120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269120, CurrentID=0, Status='live', Symbol='Optc', Aliases='', description='opticin', other_designations='oculoglycan', map_location='1 58.02 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133897193, end_position_on_the_genomic_accession=133909714, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  269132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269132, CurrentID=0, Status='live', Symbol='Colgalt2', Aliases='AI427652, D730024P12, Glt25d2', description='collagen beta(1-O)galactosyltransferase 2', other_designations='glycosyltransferase 25 domain containing 2|glycosyltransferase 25 family member 2|hydroxylysine galactosyltransferase 2', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152399867, end_position_on_the_genomic_accession=152510695, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  328845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328845, CurrentID=0, Status='live', Symbol='Acsbg2', Aliases='Bgr', description='acyl-CoA synthetase bubblegum family member 2', other_designations='bubblegum-related protein', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56843103, end_position_on_the_genomic_accession=56874504, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  269152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269152, CurrentID=0, Status='live', Symbol='Kif26b', Aliases='4832420M10, BC056349, D230039L06Rik', description='kinesin family member 26B', other_designations='N-11 kinesin|kinesin family protein 26b', map_location='1 83.16 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=178525741, end_position_on_the_genomic_accession=178939200, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  269180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269180, CurrentID=0, Status='live', Symbol='Inpp4a', Aliases='107kDa, 9630012D15, D130048C09Rik, R74740', description='inositol polyphosphate-4-phosphatase, type I', other_designations='inositol polyphosphate 4-phosphatase type I', map_location='1 15.46 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37299838, end_position_on_the_genomic_accession=37410740, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  269181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269181, CurrentID=0, Status='live', Symbol='Mgat4a', Aliases='9530018I07Rik, GnT-IVa, glcNAc-T-IVa', description='mannoside acetylglucosaminyltransferase 4, isoenzyme A', other_designations='N-acetylglucosaminyltransferase IVa|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase IVa|UDP-N-acetylglucosamine: alpha-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IVa', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37439340, end_position_on_the_genomic_accession=37541047, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  269198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269198, CurrentID=0, Status='live', Symbol='Nbeal1', Aliases='2310076G13Rik, A530050O19Rik, A530083I02Rik, AI662228, ALS2CR17', description='neurobeachin like 1', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60178629, end_position_on_the_genomic_accession=60334705, orientation='plus', exon_count=62, OMIM='', no_hdr0=''),
  269209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269209, CurrentID=0, Status='live', Symbol='Stk36', Aliases='1700112N14Rik, B930045J24, FU, Fused, mKIAA1278', description='serine/threonine kinase 36', other_designations='fused homolog|serine/threonine kinase 36 (fused homolog, Drosophila)|stk36', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74601450, end_position_on_the_genomic_accession=74636894, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  269224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269224, CurrentID=0, Status='live', Symbol='Pask', Aliases='mKIAA0135', description='PAS domain containing serine/threonine kinase', other_designations='PAS-kinase|paskin', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93309437, end_position_on_the_genomic_accession=93343350, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  269233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269233, CurrentID=0, Status='live', Symbol='Fam171a1', Aliases='9630050M13Rik, AA409679', description='family with sequence similarity 171, member A1', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3114202, end_position_on_the_genomic_accession=3227809, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  269252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269252, CurrentID=0, Status='live', Symbol='Gtf3c4', Aliases='5330400C03, AI426938, AU014771, AU017413, KAT12, TFIIIC90', description='general transcription factor IIIC, polypeptide 4', other_designations='TF3C-delta|TFIIIC 90 kDa subunit|general transcription factor IIIC 4|transcription factor IIIC 90 kDa subunit|transcription factor IIIC subunit delta', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28822299, end_position_on_the_genomic_accession=28840360, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  269254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269254, CurrentID=0, Status='live', Symbol='Setx', Aliases='A130090N03, A930037J23Rik, AOA2, AW060766, Als4, SCAR1, Sen1, mKIAA0625', description='senataxin', other_designations='SEN1 homolog|amyotrophic lateral sclerosis 4 homolog|amyotrophic lateral sclerosis 4 protein homolog', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29122588, end_position_on_the_genomic_accession=29182471, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  269261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269261, CurrentID=0, Status='live', Symbol='Rpl12', Aliases='E430018F03', description='ribosomal protein L12', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32961712, end_position_on_the_genomic_accession=32964045, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  269275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269275, CurrentID=0, Status='live', Symbol='Acvr1c', Aliases='ACVRLK7, ALK7, Alk-7, C230097P10', description='activin A receptor, type IC', other_designations='ACTR-IC|activin receptor type IC|activin receptor-like kinase 7|acvr1c', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=58267453, end_position_on_the_genomic_accession=58357896, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  241489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241489, CurrentID=0, Status='live', Symbol='Pde11a', Aliases='6330414F14Rik, A630086N24Rik, Gm350, PDE11A1', description='phosphodiesterase 11A', other_designations='cAMP and cGMP phosphodiesterase 11A', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=75987026, end_position_on_the_genomic_accession=76339145, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  269295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269295, CurrentID=0, Status='live', Symbol='Rtn4rl2', Aliases='Ngr2, Ngrh1, Ngrl3', description='reticulon 4 receptor-like 2', other_designations='nogo receptor-like 3|nogo-66 receptor homolog 1|nogo-66 receptor-related protein 2|rtn4rl2', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84871309, end_position_on_the_genomic_accession=84888507, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  269328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269328, CurrentID=0, Status='live', Symbol='Muc15', Aliases='4732460E09, D730046L02Rik, MUC-15', description='mucin 15', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=110721494, end_position_on_the_genomic_accession=110739537, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  269336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269336, CurrentID=0, Status='live', Symbol='Ccdc32', Aliases='Gm631', description='coiled-coil domain containing 32', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119017779, end_position_on_the_genomic_accession=119029406, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  269338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269338, CurrentID=0, Status='live', Symbol='Vps39', Aliases='A230065P22Rik, AW743070, Vam6, Vam6P, mVam6', description='vacuolar protein sorting 39 (yeast)', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120316461, end_position_on_the_genomic_accession=120353144, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  269344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269344, CurrentID=0, Status='live', Symbol='Ell3', Aliases='A930015D22Rik', description='elongation factor RNA polymerase II-like 3', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121439025, end_position_on_the_genomic_accession=121444278, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  269346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269346, CurrentID=0, Status='live', Symbol='Slc28a2', Aliases='2010208B10Rik, B430217P18, BB152493, Cnt2', description='solute carrier family 28 (sodium-coupled nucleoside transporter), member 2', other_designations='CNT 2|Na(+)/nucleoside cotransporter 2|SPNT|concentrative nucleoside transporter 2|purine-selective Na+ nucleoside cotransporter|sodium-coupled nucleoside transporter 2|sodium/purine nucleoside cotransporter|solute carrier family 28 member 2', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122426355, end_position_on_the_genomic_accession=122461152, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  269356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269356, CurrentID=0, Status='live', Symbol='Slc4a11', Aliases='AI503023, BTR1, NaBC1', description='solute carrier family 4, sodium bicarbonate transporter-like, member 11', other_designations='Sodium borate cotransporter 1|Solute carrier family 4 member 11|bicarbonate transporter-related protein 1', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130684107, end_position_on_the_genomic_accession=130697519, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  269378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269378, CurrentID=0, Status='live', Symbol='Ahcy', Aliases='AA987153, AL024110, CuBP, SAHH', description='S-adenosylhomocysteine hydrolase', other_designations='S-adenosyl-L-homocysteine hydrolase|adoHcyase|liver copper-binding protein', map_location='2 76.94 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155059310, end_position_on_the_genomic_accession=155074497, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  269389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269389, CurrentID=0, Status='live', Symbol='Tox2', Aliases='AI851523, AV026525, BI987407, Gcx1, RxHMG1', description='TOX high mobility group box family member 2', other_designations='granulosa cell HMG-box protein 1', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163202643, end_position_on_the_genomic_accession=163323102, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  269397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269397, CurrentID=0, Status='live', Symbol='Ss18l1', Aliases='A230053O16Rik, CREST', description='synovial sarcoma translocation gene on chromosome 18-like 1', other_designations='SS18-like protein 1', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180042483, end_position_on_the_genomic_accession=180070201, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  269400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269400, CurrentID=0, Status='live', Symbol='Rtel1', Aliases='AI451565, AW540478, Rtel', description='regulator of telomere elongation helicase 1', other_designations='DEAH helicase|regulator of telomere length', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181288028, end_position_on_the_genomic_accession=181356616, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  269401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269401, CurrentID=0, Status='live', Symbol='Zfp512b', Aliases='Gm632, Znf512b', description='zinc finger protein 512B', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181582103, end_position_on_the_genomic_accession=181592798, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  269423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269423, CurrentID=0, Status='live', Symbol='Abhd18', Aliases='2310068E01Rik, 3110057O12Rik, AI645591', description='abhydrolase domain containing 18', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40894249, end_position_on_the_genomic_accession=40937738, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  269424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269424, CurrentID=0, Status='live', Symbol='Jade1', Aliases='AU041499, D530048A03Rik, Phf17, mKIAA1807', description='jade family PHD finger 1', other_designations='Jade1 PHD finger protein|PHD finger protein 17|jade family PHD finger protein 1', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=41555734, end_position_on_the_genomic_accession=41616864, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  269437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269437, CurrentID=0, Status='live', Symbol='Plch1', Aliases='BC042549, PLC-eta-1, PLCeta1, Plcl3', description='phospholipase C, eta 1', other_designations='phosphoinositide phospholipase C-eta-1|phospholipase C-like protein 3', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=63696232, end_position_on_the_genomic_accession=63899677, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  269470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269470, CurrentID=0, Status='live', Symbol='Wdr3', Aliases='AW546279, D030020G18Rik', description='WD repeat domain 3', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100138180, end_position_on_the_genomic_accession=100162403, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  269473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269473, CurrentID=0, Status='live', Symbol='Lrig2', Aliases='4632419I10Rik, BB096938, LIG-2', description='leucine-rich repeats and immunoglobulin-like domains 2', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104453980, end_position_on_the_genomic_accession=104511938, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  269513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269513, CurrentID=0, Status='live', Symbol='Nkain3', Aliases='C530048M05, E130310K16Rik', description='Na+/K+ transporting ATPase interacting 3', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 3', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=20118336, end_position_on_the_genomic_accession=20780310, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  269514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269514, CurrentID=0, Status='live', Symbol='Fbxl4', Aliases='AI836810, FBL4, FBL5', description='F-box and leucine-rich repeat protein 4', other_designations='', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=22357543, end_position_on_the_genomic_accession=22434091, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  269523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269523, CurrentID=0, Status='live', Symbol='Vcp', Aliases='3110001E05, CDC48, p97, p97/VCP', description='valosin containing protein', other_designations='15S Mg(2+)-ATPase p97 subunit|AAA ATPase p97|Inv protein|TER ATPase|homolog of yeast cdc48|valosin-containing protein', map_location='4 22.95 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42979964, end_position_on_the_genomic_accession=43000507, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  269529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269529, CurrentID=0, Status='live', Symbol='Fbxo10', Aliases='FBX10, Gm634', description='F-box protein 10', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45034248, end_position_on_the_genomic_accession=45084604, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  269536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269536, CurrentID=0, Status='live', Symbol='Tex10', Aliases='2610206N19Rik, 2810462N03Rik, BC006867', description='testis expressed gene 10', other_designations='clone 18330', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48430858, end_position_on_the_genomic_accession=48473464, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  269582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269582, CurrentID=0, Status='live', Symbol='Clspn', Aliases='B130025E01, C85083, E130314M08Rik', description='claspin', other_designations='claspin homolog', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126556951, end_position_on_the_genomic_accession=126594533, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  269585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269585, CurrentID=0, Status='live', Symbol='Zscan20', Aliases='C130001F22, Zfp31, Zkscan20', description='zinc finger and SCAN domains 20', other_designations='zinc finger protein 31|zinc finger with KRAB and SCAN domains 20', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128578670, end_position_on_the_genomic_accession=128610172, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  269587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269587, CurrentID=0, Status='live', Symbol='Epb41', Aliases='4.1R, AI415518, D4Ertd442e, Elp-1, Elp1, Epb4.1, mKIAA4056', description='erythrocyte membrane protein band 4.1', other_designations='P4.1|Protein 4.1R|band 4.1|erythrocyte protein band 4.1', map_location='4 64.54 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=131923413, end_position_on_the_genomic_accession=132075579, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  269589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269589, CurrentID=0, Status='live', Symbol='Sytl1', Aliases='9030417L18, PSGL-1, Slp1', description='synaptotagmin-like 1', other_designations='exophilin-7', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133253090, end_position_on_the_genomic_accession=133263087, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  269593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269593, CurrentID=0, Status='live', Symbol='Luzp1', Aliases='2700072H04Rik, AI266952, Luzp, mFLJ00226', description='leucine zipper protein 1', other_designations='leucine zipper motif-containing protein', map_location='4 68.8 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136469749, end_position_on_the_genomic_accession=136554725, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  269604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269604, CurrentID=0, Status='live', Symbol='Gpr157', Aliases='F730108M23Rik', description='G protein-coupled receptor 157', other_designations='novel 7 transmembrane receptor (Secretin family) domain containing protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150087503, end_position_on_the_genomic_accession=150105927, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  269608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269608, CurrentID=0, Status='live', Symbol='Plekhg5', Aliases='BC023181, Syx', description='pleckstrin homology domain containing, family G (with RhoGef domain) member 5', other_designations='PH domain-containing family G member 5|RhoA-specific guanine exchange factor SYX1|RhoA-specific guanine exchange factor SYX2|synectin-binding RhoA exchange factor', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152072433, end_position_on_the_genomic_accession=152115404, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  269610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269610, CurrentID=0, Status='live', Symbol='Chd5', Aliases='4930532L22Rik, AW060752, B230399N07Rik, CHD-5', description='chromodomain helicase DNA binding protein 5', other_designations='ATP-dependent helicase CHD5', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152338651, end_position_on_the_genomic_accession=152390194, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  269614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269614, CurrentID=0, Status='live', Symbol='Pank4', Aliases='D030031I12Rik, R75150', description='pantothenate kinase 4', other_designations='pantothenic acid kinase 4', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154963948, end_position_on_the_genomic_accession=154980938, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  269615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269615, CurrentID=0, Status='live', Symbol='Plch2', Aliases='A930027K05Rik, PLCeta2, Plc-eta2, Plcl4', description='phospholipase C, eta 2', other_designations='phosphoinositide phospholipase C-eta-2|phosphoinositide phospholipase C-like 4|phospholipase C-like protein 4', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154983115, end_position_on_the_genomic_accession=155064091, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  269623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269623, CurrentID=0, Status='live', Symbol='Rbm48', Aliases='AW548102, C030048B08Rik', description='RNA binding motif protein 48', other_designations='UPF0712 protein C7orf64 homolog', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3583978, end_position_on_the_genomic_accession=3596547, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  269629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269629, CurrentID=0, Status='live', Symbol='Lhfpl3', Aliases='A930031L14Rik', description='lipoma HMGIC fusion partner-like 3', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=22745636, end_position_on_the_genomic_accession=23275597, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  269630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269630, CurrentID=0, Status='live', Symbol='5031425E22Rik', Aliases='', description='RIKEN cDNA 5031425E22 gene', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23431808, end_position_on_the_genomic_accession=23434353, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  269633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269633, CurrentID=0, Status='live', Symbol='Wdr86', Aliases='2810046M22Rik, BC059069', description='WD repeat domain 86', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24704229, end_position_on_the_genomic_accession=24730698, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  269637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269637, CurrentID=0, Status='live', Symbol='Cnpy1', Aliases='1500012D20Rik, 9630008K15Rik, AI853839', description='canopy FGF signaling regulator 1', other_designations='canopy 1 homolog', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=28200826, end_position_on_the_genomic_accession=28237873, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  269639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269639, CurrentID=0, Status='live', Symbol='Zfp512', Aliases='2500002M11Rik, AI450173, D230008H22, Znf512', description='zinc finger protein 512', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31452416, end_position_on_the_genomic_accession=31481754, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  269642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269642, CurrentID=0, Status='live', Symbol='Nat8l', Aliases='1110038O08Rik, Shati', description='N-acetyltransferase 8-like', other_designations='N-acetyltransferase 8-like protein|NAA synthetase|protein Shati', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33995984, end_position_on_the_genomic_accession=34005916, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  269643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269643, CurrentID=0, Status='live', Symbol='Ppp2r2c', Aliases='6330548O06Rik, IMYPNO, IMYPNO1, PR52', description='protein phosphatase 2, regulatory subunit B, gamma', other_designations='PP2A subunit B isoform B55-gamma|PP2A subunit B isoform PR55-gamma|PP2A subunit B isoform R2-gamma|PP2A subunit B isoform gamma|protein phosphatase 2 (formerly 2A), regulatory subunit B (PR 52), gamma isoform', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36866843, end_position_on_the_genomic_accession=36955078, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  269682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269682, CurrentID=0, Status='live', Symbol='Golga3', Aliases='5330413L04, 5430416E01Rik, AI449376, AW490576, G1-499-14, Mea-2, Mea2, repro27', description='golgi autoantigen, golgin subfamily a, 3', other_designations='Mea2/Golga3|golgin-160|male-enhanced antigen 2', map_location='5 53.36 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110176659, end_position_on_the_genomic_accession=110226470, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  269693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269693, CurrentID=0, Status='live', Symbol='Ccdc60', Aliases='AU018954, C130098D09', description='coiled-coil domain containing 60', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=116125581, end_position_on_the_genomic_accession=116288998, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  269695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269695, CurrentID=0, Status='live', Symbol='Rnft2', Aliases='AW049082, B830028P19Rik, Tmem118', description='ring finger protein, transmembrane 2', other_designations='transmembrane protein 118', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=118190734, end_position_on_the_genomic_accession=118245111, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  269700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269700, CurrentID=0, Status='live', Symbol='Gm15800', Aliases='A130010L24, AU042671, BC019639, EG545802, OTTMUSG00000026283, mKIAA0614', description='predicted gene 15800', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121219339, end_position_on_the_genomic_accession=121368577, orientation='plus', exon_count=78, OMIM='', no_hdr0=''),
  269701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269701, CurrentID=0, Status='live', Symbol='Wdr66', Aliases='4930415N18Rik, 4933428F06Rik, 5031404N07', description='WD repeat domain 66', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123254560, end_position_on_the_genomic_accession=123327484, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  269702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269702, CurrentID=0, Status='live', Symbol='Mphosph9', Aliases='4930548D04Rik, 9630025B04Rik, AW413446, AW547060, B930097C17Rik, C87456, MPP-9, MPP9', description='M-phase phosphoprotein 9', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124250959, end_position_on_the_genomic_accession=124330759, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  269704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269704, CurrentID=0, Status='live', Symbol='Zfp664', Aliases='AW060232, D930038J03Rik, ZFOC1, Znf664', description='zinc finger protein 664', other_designations='zinc finger protein ZFOC1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124862705, end_position_on_the_genomic_accession=124888630, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  269713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269713, CurrentID=0, Status='live', Symbol='Clip2', Aliases='B230327O20, CLIP-115, Clip1, Cyln2, WSCR4, mKIAA0291, wbscr4', description='CAP-GLY domain containing linker protein 2', other_designations='cytoplasmic linker protein 1, 115 kDa|cytoplasmic linker protein 2', map_location='5 74.63 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134489383, end_position_on_the_genomic_accession=134552468, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  269717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269717, CurrentID=0, Status='live', Symbol='Orai2', Aliases='A730041O15Rik, Tmem142b', description='ORAI calcium release-activated calcium modulator 2', other_designations='calcium release-activated calcium channel 2|transmembrane protein 142B', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136147459, end_position_on_the_genomic_accession=136170656, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66805, CurrentID=0, Status='live', Symbol='Tspan1', Aliases='9030418M05Rik', description='tetraspanin 1', other_designations='tetraspan 1|tspan-1', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116161881, end_position_on_the_genomic_accession=116167621, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  269774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269774, CurrentID=0, Status='live', Symbol='Aak1', Aliases='5530400K14Rik, 9630042K20, AU067724, AU067726, BC028270, C79663, D6Ertd245e, R75501, mKIAA1048', description='AP2 associated kinase 1', other_designations='adaptor-associated kinase 1', map_location='6 37.75 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86849482, end_position_on_the_genomic_accession=87003227, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  269784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269784, CurrentID=0, Status='live', Symbol='Cntn4', Aliases='9630050B05, Axcam, BIG-2', description='contactin 4', other_designations='axonal-associated cell adhesion molecule|brain-derived immunoglobulin superfamily molecule|brain-derived immunoglobulin superfamily protein 2', map_location='6 E2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=105677651, end_position_on_the_genomic_accession=106700141, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  269788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269788, CurrentID=0, Status='live', Symbol='Lhfpl4', Aliases='1190004M23Rik, AI604880, B230384L07, mKIAA4027', description='lipoma HMGIC fusion partner-like protein 4', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113168087, end_position_on_the_genomic_accession=113195384, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  269799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269799, CurrentID=0, Status='live', Symbol='Clec4a1', Aliases='BC049354, Dcir4', description='C-type lectin domain family 4, member a1', other_designations='dendritic cell inhibitory receptor 4', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122921848, end_position_on_the_genomic_accession=122934619, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  269800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269800, CurrentID=0, Status='live', Symbol='Zfp384', Aliases='BB163993, C130073D16Rik, Ciz, NP, Nmp4', description='zinc finger protein 384', other_designations='nuclear matrix protein 4', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125009110, end_position_on_the_genomic_accession=125037870, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  269823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269823, CurrentID=0, Status='live', Symbol='Pon3', Aliases='2810004E20, AI786302', description='paraoxonase 3', other_designations='', map_location='6 1.99 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=5220852, end_position_on_the_genomic_accession=5256250, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  269831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269831, CurrentID=0, Status='live', Symbol='Tspan12', Aliases='9030619E17, AI426782, AI663988, AW111457, Tm4sf12', description='tetraspanin 12', other_designations='transmembrane 4 superfamily member 12|tspan-12', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=21771391, end_position_on_the_genomic_accession=21853349, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  269846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269846, CurrentID=0, Status='live', Symbol='Trbv31', Aliases='Gm16809, Tcrb-V14', description='T cell receptor beta, variable 31', other_designations='T-cell receptor beta, variable 14|T-cell receptor beta, variable 31', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41557693, end_position_on_the_genomic_accession=41558371, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  269854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269854, CurrentID=0, Status='live', Symbol='Nat14', Aliases='BC047219', description='N-acetyltransferase 14', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4918888, end_position_on_the_genomic_accession=4925006, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  269855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269855, CurrentID=0, Status='live', Symbol='Ssc5d', Aliases='A430110N23Rik, S5D-SRCRB', description='scavenger receptor cysteine rich family, 5 domains', other_designations='group B scavenger receptor cysteine-rich domain-containing|scavenger receptor cysteine rich domain containing (5 domains)|scavenger receptor cysteine-rich domain-containing protein LOC284297 homolog', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4925130, end_position_on_the_genomic_accession=4944828, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  269862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269862, CurrentID=0, Status='live', Symbol='Olfr1349', Aliases='MOR232-8', description='olfactory receptor 1349', other_designations='GA_x6K02T2QGBW-3245761-3244808|olfactory receptor MOR232-8', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6514199, end_position_on_the_genomic_accession=6520355, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  269870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269870, CurrentID=0, Status='live', Symbol='Zfp446', Aliases='A630035I11Rik, A930031C22, AA120513, AV306632', description='zinc finger protein 446', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12976397, end_position_on_the_genomic_accession=12985717, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  328968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328968, CurrentID=0, Status='live', Symbol='Mir143hg', Aliases='Carmen, E330013P06, Gm16908', description='Mir143 and Mir145 host gene (non-protein coding)', other_designations='Mir143 and Mir145 host gene (non-coding RNA)|cardiac mesoderm enhancer-associated noncoding RNA', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61645878, end_position_on_the_genomic_accession=61665538, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  269878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269878, CurrentID=0, Status='live', Symbol='Megf8', Aliases='AW049492, Egfl4, b2b1702.2Clo, b2b1702Clo, b2b288Clo, m687Ddg, mKIAA0817', description='multiple EGF-like-domains 8', other_designations='EGF-like protein 4|EGF-like-domain, multiple 4|epidermal growth factor-like protein 4|multiple EGF-like domains protein 8', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25316742, end_position_on_the_genomic_accession=25365917, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  269881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269881, CurrentID=0, Status='live', Symbol='Map3k10', Aliases='BC028668, BC046514, MST, Mlk2', description='mitogen-activated protein kinase kinase kinase 10', other_designations='MKN28 derived nonreceptor_type|MKN28 kinase|mixed lineage kinase 2|serine/threonine kinase', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27656375, end_position_on_the_genomic_accession=27674607, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  328971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328971, CurrentID=0, Status='live', Symbol='Spink10', Aliases='A230091H23Rik', description='serine peptidase inhibitor, Kazal type 10', other_designations='serine protease inhibitor Kazal type 10', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=62548911, end_position_on_the_genomic_accession=62662933, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  269902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269902, CurrentID=0, Status='live', Symbol='Vmn2r57', Aliases='D430014M15, EG269902', description='vomeronasal 2, receptor 57', other_designations='vomeronasal receptor Vmn2r57', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=41399732, end_position_on_the_genomic_accession=41448641, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  110521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110521, CurrentID=0, Status='live', Symbol='Hivep1', Aliases='Cryabp1', description='human immunodeficiency virus type I enhancer binding protein 1', other_designations='HIV type I enhancer binding protein 1|alpha A-CRYBP1|alpha A-crystallin-binding protein 1|alpha A-crystallin-binding protein I|alphaA-CRYBP1|crystallin, alpha binding protein 1|human immunodeficiency virus type I enhancer-binding protein 1|transcription factor alphaA-CRYBP1', map_location='13 20.72 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=42051969, end_position_on_the_genomic_accession=42185030, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  269941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269941, CurrentID=0, Status='live', Symbol='Chsy1', Aliases='mKIAA0990', description='chondroitin sulfate synthase 1', other_designations='N-acetylgalactosaminyl-proteoglycan 3-beta-glucuronosyltransferase 1|N-acetylgalactosaminyltransferase 1|carbohydrate (chondroitin) synthase 1|chSy-1|chondroitin glucuronyltransferase 1|chondroitin synthase 1|glucuronosyl-N-acetylgalactosaminyl-proteoglycan 4-beta-N-acetylgalactosaminyltransferase 1', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66109515, end_position_on_the_genomic_accession=66173798, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  269951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269951, CurrentID=0, Status='live', Symbol='Idh2', Aliases='E430004F23, IDPm, Idh-2', description='isocitrate dehydrogenase 2 (NADP+), mitochondrial', other_designations='ICD-M|IDH|IDP|NADP(+)-specific ICDH|NADP+-specific isocitrate dehydrogenase|oxalosuccinate decarboxylase', map_location='7 45.43 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80094846, end_position_on_the_genomic_accession=80115387, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  269952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269952, CurrentID=0, Status='live', Symbol='Gdpgp1', Aliases='D330012F22Rik', description='GDP-D-glucose phosphorylase 1', other_designations='GDP-D-glucose phosphorylase C15orf58 homolog', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80232893, end_position_on_the_genomic_accession=80241420, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  269954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269954, CurrentID=0, Status='live', Symbol='Ttll13', Aliases='1700111A04Rik, 4921538O11, BC024578', description='tubulin tyrosine ligase-like family, member 13', other_designations='ttll13|tubulin--tyrosine ligase-like protein 13', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80246376, end_position_on_the_genomic_accession=80260821, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  269955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269955, CurrentID=0, Status='live', Symbol='Rccd1', Aliases='5830436H09, E430018M08Rik', description='RCC1 domain containing 1', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80294598, end_position_on_the_genomic_accession=80324696, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  269959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269959, CurrentID=0, Status='live', Symbol='Adamtsl3', Aliases='9230119C12Rik, C130057K09, mKIAA1233', description='ADAMTS-like 3', other_designations='a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif like, 3|punctin-2', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=82335446, end_position_on_the_genomic_accession=82614450, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  269966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269966, CurrentID=0, Status='live', Symbol='Nup98', Aliases='4732457F17, AI849286, Nup96', description='nucleoporin 98', other_designations='', map_location='7 54.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102119400, end_position_on_the_genomic_accession=102210166, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  269980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269980, CurrentID=0, Status='live', Symbol='Gm5053', Aliases='EG269980', description='predicted gene 5053', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108527727, end_position_on_the_genomic_accession=108529070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  269994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269994, CurrentID=0, Status='live', Symbol='Gsg1l', Aliases='AI852434, C230098I05Rik, G630023A01Rik, Gm637, g1l', description='GSG1-like', other_designations='GSG1-like protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125878412, end_position_on_the_genomic_accession=126082772, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  269997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269997, CurrentID=0, Status='live', Symbol='Zfp747', Aliases='6430604K15Rik', description='zinc finger protein 747', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127372537, end_position_on_the_genomic_accession=127376050, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  269999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=269999, CurrentID=0, Status='live', Symbol='Orai3', Aliases='9930124N15, BC061259, Tmem142c', description='ORAI calcium release-activated calcium modulator 3', other_designations='calcium release-activated calcium channel 3|transmembrane protein 142C', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127769589, end_position_on_the_genomic_accession=127775150, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  270004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270004, CurrentID=0, Status='live', Symbol='Foxi2', Aliases='B130055A05Rik', description='forkhead box I2', other_designations='foxi2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=135410367, end_position_on_the_genomic_accession=135413615, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  270028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270028, CurrentID=0, Status='live', Symbol='Fam155a', Aliases='6430500D16, AW121567, Tmem28', description='family with sequence similarity 155, member A', other_designations='transmembrane protein 28', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=9206010, end_position_on_the_genomic_accession=9773363, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  270035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270035, CurrentID=0, Status='live', Symbol='Letm2', Aliases='6030453H13, D030041N04Rik', description='leucine zipper-EF-hand containing transmembrane protein 2', other_designations='LETM1 and EF-hand domain-containing protein 2', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25578481, end_position_on_the_genomic_accession=25597899, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  270049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270049, CurrentID=0, Status='live', Symbol='Galntl6', Aliases='1700021K10Rik, 4930431L04Rik, A830023L05', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 6', other_designations='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 20', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=57772489, end_position_on_the_genomic_accession=58911627, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  270058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270058, CurrentID=0, Status='live', Symbol='Map1s', Aliases='6430517J16Rik, Bpy2ip1, Map8, Mtap1s, VCY2IP1', description='microtubule-associated protein 1S', other_designations='BPY2 interacting protein 1|BPY2-interacting protein 1|MAP-1S|microtubule-associated protein 8', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70905957, end_position_on_the_genomic_accession=70930112, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  270066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270066, CurrentID=0, Status='live', Symbol='Slc35e1', Aliases='6030458H05, AA408278', description='solute carrier family 35, member E1', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72477995, end_position_on_the_genomic_accession=72492614, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  270076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270076, CurrentID=0, Status='live', Symbol='Gcdh', Aliases='9030411L18, AI266902, D17825, GCD', description='glutaryl-Coenzyme A dehydrogenase', other_designations='', map_location='8 41.28 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84886387, end_position_on_the_genomic_accession=84893921, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66859, CurrentID=0, Status='live', Symbol='Slc16a9', Aliases='1200003C15Rik, 4930425B13Rik, AI552636', description='solute carrier family 16 (monocarboxylic acid transporters), member 9', other_designations='MCT 9|solute carrier family 16 member 9', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=70245276, end_position_on_the_genomic_accession=70285951, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  270084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270084, CurrentID=0, Status='live', Symbol='Lpcat2', Aliases='A330042H22, Aytl1, Aytl1a, lpafat1', description='lysophosphatidylcholine acyltransferase 2', other_designations='1-AGP acyltransferase 11|1-AGPAT 11|1-acylglycerol-3-phosphate O-acyltransferase 11|1-acylglycerophosphocholine O-acyltransferase|1-alkylglycerophosphocholine O-acetyltransferase|LPC acyltransferase 2|LPCAT-2|acetyl-CoA:lyso-PAF acetyltransferase|acetyl-CoA:lyso-platelet-activating factor acetyltransferase|acyl-CoA:lysophosphatidylcholine acyltransferase member 1|acyltransferase like 1|acyltransferase like 1A|acyltransferase-like 1|lyso-PAF acetyltransferase|lysoPAFAT|lysoPAFAT/LPCAT2|lysoPC acyltransferase 2', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=92855330, end_position_on_the_genomic_accession=92919279, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  270086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270086, CurrentID=0, Status='live', Symbol='Ogfod1', Aliases='4930415J21Rik, AA387199, AA939912, AW061076, mKIAA1612', description='2-oxoglutarate and iron-dependent oxygenase domain containing 1', other_designations='2-oxoglutarate and iron-dependent oxygenase domain-containing protein 1', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94037141, end_position_on_the_genomic_accession=94067923, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  270091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270091, CurrentID=0, Status='live', Symbol='Lrrc36', Aliases='BC036564', description='leucine rich repeat containing 36', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105411389, end_position_on_the_genomic_accession=105464086, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  270096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270096, CurrentID=0, Status='live', Symbol='Mon1b', Aliases='5031407H10, 5033413H12Rik', description='MON1 homolog b (yeast)', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=113635586, end_position_on_the_genomic_accession=113645192, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  270097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270097, CurrentID=0, Status='live', Symbol='Vat1l', Aliases='9430073I07, AI427515, mKIAA1576', description='vesicle amine transport protein 1 homolog-like (T. californica)', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=114205582, end_position_on_the_genomic_accession=114374074, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  270106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270106, CurrentID=0, Status='live', Symbol='Rpl13', Aliases='A52, L13', description='ribosomal protein L13', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123102350, end_position_on_the_genomic_accession=123105244, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  270109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270109, CurrentID=0, Status='live', Symbol='Pcnxl2', Aliases='A830048P05, AU067631, E330039K12Rik, mKIAA0435', description='pecanex-like 2 (Drosophila)', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125751504, end_position_on_the_genomic_accession=125898317, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  270110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270110, CurrentID=0, Status='live', Symbol='Irf2bp2', Aliases='E130305N23Rik, IRF-2BP2', description='interferon regulatory factor 2 binding protein 2', other_designations='IRF-2-binding protein 2', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=126588296, end_position_on_the_genomic_accession=126593436, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  270118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270118, CurrentID=0, Status='live', Symbol='Maml2', Aliases='5930431H10, BC032967', description='mastermind like 2 (Drosophila)', other_designations='', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=13619989, end_position_on_the_genomic_accession=13709533, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  270120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270120, CurrentID=0, Status='live', Symbol='Fat3', Aliases='9430076A06Rik, D430038H04Rik, Gm1132, Gm510, MFAT3F', description='FAT atypical cadherin 3', other_designations='FAT tumor suppressor homolog 3|protocadherin Fat3', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15910189, end_position_on_the_genomic_accession=16502879, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  270151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270151, CurrentID=0, Status='live', Symbol='Nlrx1', Aliases='BC034204, NOD9', description='NLR family member X1', other_designations='nucleotide-binding domain and leucine-rich repeat containing', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44252713, end_position_on_the_genomic_accession=44268599, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  270152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270152, CurrentID=0, Status='live', Symbol='Amica1', Aliases='AMICA, Crea7, Gm638', description='adhesion molecule, interacts with CXADR antigen 1', other_designations='adhesion molecule AMICA|dendritic cell-specific protein CREA7|dendritic-cell specific protein Crea7|mCrea7', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45079162, end_position_on_the_genomic_accession=45108531, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  329015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329015, CurrentID=0, Status='live', Symbol='Atg2a', Aliases='1810013C15Rik, A830054M12, BC023754, mKIAA0404', description='autophagy related 2A', other_designations='ATG2 autophagy related 2 homolog A', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6241666, end_position_on_the_genomic_accession=6262304, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  270156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270156, CurrentID=0, Status='live', Symbol='AU019823', Aliases='Gm639', description='expressed sequence AU019823', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50605240, end_position_on_the_genomic_accession=50617589, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  270160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270160, CurrentID=0, Status='live', Symbol='Rab39', Aliases='C230094F14Rika, Rab39', description='RAB39, member RAS oncogene family', other_designations='rab-39', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53684110, end_position_on_the_genomic_accession=53706232, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  270162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270162, CurrentID=0, Status='live', Symbol='Elmod1', Aliases='4831417L10, AI428506, AI834991', description='ELMO/CED-12 domain containing 1', other_designations='ELMO domain containing 1', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53911457, end_position_on_the_genomic_accession=53975310, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  270163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270163, CurrentID=0, Status='live', Symbol='Myo9a', Aliases='4732465J09Rik, C130068I12Rik, C230003M11', description='myosin IXa', other_designations='myosin-IXa|unconventional myosin-9a', map_location='9 32.13 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59750856, end_position_on_the_genomic_accession=59928866, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  270166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270166, CurrentID=0, Status='live', Symbol='Clpx', Aliases='AU014732, E330029I21', description='caseinolytic mitochondrial matrix peptidase chaperone subunit', other_designations='caseinolytic peptidase X|caseinolytic protease X', map_location='9 35.22 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65294250, end_position_on_the_genomic_accession=65330658, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  110564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110564, CurrentID=0, Status='live', Symbol='Lth1', Aliases='Lth-1, Lvth-1', description='liver 30-40 thousand Mol. Wt. protein 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  270190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270190, CurrentID=0, Status='live', Symbol='Ephb1', Aliases='9330129L11, AW488255, C130099E04Rik, Cek6, ENSMUSG00000074119, Elk, Elkh, Hek6, Net', description='Eph receptor B1', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=101922128, end_position_on_the_genomic_accession=102354693, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  270192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270192, CurrentID=0, Status='live', Symbol='Rab6b', Aliases='9630015C03, AI844641, C330006L04Rik, D9Bwg0185e', description='RAB6B, member RAS oncogene family', other_designations='', map_location='9 54.88 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103112074, end_position_on_the_genomic_accession=103185276, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  270198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270198, CurrentID=0, Status='live', Symbol='Pfkfb4', Aliases='C230090D14', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 4', other_designations='6PF-2-K/Fru-2,6-P2ase 4|6PF-2-K/Fru-2,6-P2ase testis-type isozyme|PFK/FBPase 4', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108988033, end_position_on_the_genomic_accession=109032236, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  270201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270201, CurrentID=0, Status='live', Symbol='Klhl18', Aliases='A930041K15, AW545966', description='kelch-like 18', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110425926, end_position_on_the_genomic_accession=110476700, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  270210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270210, CurrentID=0, Status='live', Symbol='Zfp651', Aliases='4732420M22Rik, BC030045, R74626', description='zinc finger protein 651', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121758638, end_position_on_the_genomic_accession=121771742, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  270328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270328, CurrentID=0, Status='live', Symbol='Gsdmc3', Aliases='9930109F21Rik', description='gasdermin C3', other_designations='gsdmc3', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=63857716, end_position_on_the_genomic_accession=63878583, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66921, CurrentID=0, Status='live', Symbol='Prpf38b', Aliases='1110021E09Rik, AU018955', description='PRP38 pre-mRNA processing factor 38 (yeast) domain containing B', other_designations='', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108902807, end_position_on_the_genomic_accession=108911752, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  241688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241688, CurrentID=0, Status='live', Symbol='Dzank1', Aliases='2810039F03Rik, 6330439K17Rik, 6430408C01, AI851115, Ankrd64', description='double zinc ribbon and ankyrin repeat domains 1', other_designations='ankyrin repeat-containing protein C20orf12 homolog|likely ortholog of H. sapiens chromosome 20 open reading frame 12 (C20orf12)', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144470555, end_position_on_the_genomic_accession=144527425, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  329093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329093, CurrentID=0, Status='live', Symbol='Cpa6', Aliases='9030616D13Rik', description='carboxypeptidase A6', other_designations='metallocarboxypeptidase', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=10295791, end_position_on_the_genomic_accession=10720058, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  270624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270624, CurrentID=0, Status='live', Symbol='Spin4', Aliases='9630042H07Rik', description='spindlin family, member 4', other_designations='novel Spin/Ssty repeat containing protein', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=95022507, end_position_on_the_genomic_accession=95026682, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  270627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270627, CurrentID=0, Status='live', Symbol='Taf1', Aliases='AU015687, B430306D02Rik, Ccg-1, Ccg1, KAT4, N-TAF1, TAFII250, Taf2a, p250', description='TATA-box binding protein associated factor 1', other_designations='TAF1 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF1 RNA polymerase II, TATA box binding protein (TBP)-associated factor, neuron specific isoform|TBP-associated factor 250 kDa|cell cycle gene 1 protein|cell cycle, G1 phase defect|transcription initiation factor TFIID 250 kDa subunit', map_location='X 44.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101527574, end_position_on_the_genomic_accession=101601789, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  270669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270669, CurrentID=0, Status='live', Symbol='Mbtps2', Aliases='9630032G22Rik, AI662535', description='membrane-bound transcription factor peptidase, site 2', other_designations='S2P endopeptidase|endopeptidase S2P|membrane-bound transcription factor protease, site 2|site-2 protease', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=157547822, end_position_on_the_genomic_accession=157598715, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  270672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270672, CurrentID=0, Status='live', Symbol='Map3k15', Aliases='BC031147, MEKK15', description='mitogen-activated protein kinase kinase kinase 15', other_designations='MAPK/ERK kinase kinase 15|MEK kinase 15|MEKK 15|map3k15', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=159988433, end_position_on_the_genomic_accession=160123351, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  270685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270685, CurrentID=0, Status='live', Symbol='Mthfd1l', Aliases='2410004L15Rik, AI647056, Fthfsdc1', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 1-like', other_designations='formyltetrahydrofolate synthetase domain containing 1', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=3970441, end_position_on_the_genomic_accession=4167081, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  270711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270711, CurrentID=0, Status='live', Symbol='Fam26d', Aliases='4732454E20Rik, AI586164, Gm643', description='family with sequence similarity 26, member D', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34038784, end_position_on_the_genomic_accession=34044313, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  241727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241727, CurrentID=0, Status='live', Symbol='Snph', Aliases='6430515A01, AW045671, AW556958, AW743098, mKIAA0374', description='syntaphilin', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151590549, end_position_on_the_genomic_accession=151633030, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  241732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241732, CurrentID=0, Status='live', Symbol='Tspyl3', Aliases='AW212607, Gm36', description='TSPY-like 3', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153222370, end_position_on_the_genomic_accession=153225441, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  270757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270757, CurrentID=0, Status='live', Symbol='Bpifc', Aliases='4732454E24, Bpil2', description='BPI fold containing family C', other_designations='bactericidal/permeability-increasing protein-like 2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85959340, end_position_on_the_genomic_accession=86015898, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  270802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270802, CurrentID=0, Status='live', Symbol='BC048403', Aliases='E030027L10', description='cDNA sequence BC048403', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121739603, end_position_on_the_genomic_accession=121757613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  270893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270893, CurrentID=0, Status='live', Symbol='Tmem132e', Aliases='Gm644', description='transmembrane protein 132E', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82387446, end_position_on_the_genomic_accession=82447618, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  270906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=270906, CurrentID=0, Status='live', Symbol='Prr11', Aliases='B930067F20Rik', description='proline rich 11', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87089156, end_position_on_the_genomic_accession=87108714, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67008, CurrentID=0, Status='live', Symbol='Yae1d1', Aliases='1600012F09Rik', description='Yae1 domain containing 1', other_designations='', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=17986640, end_position_on_the_genomic_accession=17993351, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  329154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329154, CurrentID=0, Status='live', Symbol='Ankrd44', Aliases='4930444A19Rik, A130096K20, E130014H08Rik, PP6-ARS-B', description='ankyrin repeat domain 44', other_designations='ankyrin repeat domain-containing protein 44|serine/threonine-protein phosphatase 6 regulatory subunit ARS-B', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=54645340, end_position_on_the_genomic_accession=54926733, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  271005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271005, CurrentID=0, Status='live', Symbol='Klhdc1', Aliases='', description='kelch domain containing 1', other_designations='', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69226917, end_position_on_the_genomic_accession=69284646, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  329160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329160, CurrentID=0, Status='live', Symbol='9130024F11Rik', Aliases='B530002L08, EG329160', description='RIKEN cDNA 9130024F11 gene', other_designations='', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=56971469, end_position_on_the_genomic_accession=56975196, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  271036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271036, CurrentID=0, Status='live', Symbol='Catsperb', Aliases='4931410B03, 4932415G16Rik', description='catsper channel auxiliary subunit beta', other_designations='CatSper channel subunit beta|catSper-beta|cation channel, sperm-associated, beta', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=101404645, end_position_on_the_genomic_accession=101626009, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  271047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271047, CurrentID=0, Status='live', Symbol='Serpina3b', Aliases='6A1, A030003A19Rik', description='serine (or cysteine) peptidase inhibitor, clade A, member 3B', other_designations='alpha-1 antiproteinase|antitrypsin|serine (or cysteine) proteinase inhibitor, clade A, member 3B|serpin A3B', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104127531, end_position_on_the_genomic_accession=104139546, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  271127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271127, CurrentID=0, Status='live', Symbol='Adamts16', Aliases='', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 16', other_designations='ADAM-TS 16|ADAM-TS16|ADAMTS-16|a disintegrin-like and metallopetidase (reprolysin type) with thrombospondin type 1 motif, 16|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 16 preproprotein', map_location='13 35.97 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=70727798, end_position_on_the_genomic_accession=70841848, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  271144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271144, CurrentID=0, Status='live', Symbol='Ankdd1b', Aliases='9330128J19Rik', description='ankyrin repeat and death domain containing 1B', other_designations='ankyrin repeat and death domain-containing protein ENSP00000345065 homolog', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=96415030, end_position_on_the_genomic_accession=96471354, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  271209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271209, CurrentID=0, Status='live', Symbol='Rp1l1', Aliases='BC031365, BE983540, Dcdc4, Rp1hl1', description='retinitis pigmentosa 1 homolog (human)-like 1', other_designations='retinitis pigmentosa 1-like protein 1', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63992431, end_position_on_the_genomic_accession=64033506, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  271221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271221, CurrentID=0, Status='live', Symbol='5031414D18Rik', Aliases='Gm744, Gm911, Kiaa0226l', description='RIKEN cDNA 5031414D18 gene', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75015878, end_position_on_the_genomic_accession=75052537, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  271278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271278, CurrentID=0, Status='live', Symbol='BC024139', Aliases='6230424I18Rik', description='cDNA sequence BC024139', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76119517, end_position_on_the_genomic_accession=76126556, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  271305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271305, CurrentID=0, Status='live', Symbol='Phf21b', Aliases='A730032D07Rik', description='PHD finger protein 21B', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84785376, end_position_on_the_genomic_accession=84856132, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67063, CurrentID=0, Status='live', Symbol='Tmem246', Aliases='2810432L12Rik, 9330170P15Rik, AI835809', description='transmembrane protein 246', other_designations='transmembrane protein C9orf125 homolog', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49584506, end_position_on_the_genomic_accession=49597870, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  271375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271375, CurrentID=0, Status='live', Symbol='Cd200r2', Aliases='AY230198, Cd200r1l', description='Cd200 receptor 2', other_designations='CD200 cell surface glycoprotein receptor|CD200 cell surface glycoprotein receptor-like 2|CD200 cell surface glycoprotein receptor-like c|CD200 receptor-like 2|CD200RLc|cell surface glycoprotein OX2 receptor 2', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44867097, end_position_on_the_genomic_accession=44915840, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  271377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271377, CurrentID=0, Status='live', Symbol='Zbtb11', Aliases='9230110G02Rik, ZNF-U69274', description='zinc finger and BTB domain containing 11', other_designations='', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=55972697, end_position_on_the_genomic_accession=56008913, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  664608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664608, CurrentID=0, Status='live', Symbol='Rhox4g', Aliases='Rhox4.6, Rhox4.7, Rhox4f', description='reproductive homeobox 4G', other_designations='reproductive homeobox 4F', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37646501, end_position_on_the_genomic_accession=37651327, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  664609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664609, CurrentID=0, Status='live', Symbol='Rhox4a', Aliases='Rhox4, Rhox4.1', description='reproductive homeobox 4A', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37265318, end_position_on_the_genomic_accession=37270136, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  664610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664610, CurrentID=0, Status='live', Symbol='Rhox4d', Aliases='Rhox4.4, Rhox4c', description='reproductive homeobox 4D', other_designations='', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37514389, end_position_on_the_genomic_accession=37519176, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  271424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271424, CurrentID=0, Status='live', Symbol='Ip6k3', Aliases='D830007E07Rik, Ihpk3', description='inositol hexaphosphate kinase 3', other_designations='insP6 kinase 3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27143971, end_position_on_the_genomic_accession=27167764, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  241846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241846, CurrentID=0, Status='live', Symbol='Lsm14b', Aliases='9430078B04, BC040823', description='LSM family member 14B', other_designations='LSM14 homolog B|RNA-associated protein 55B|mRAP55B', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180024902, end_position_on_the_genomic_accession=180035466, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  271457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271457, CurrentID=0, Status='live', Symbol='Rab5a', Aliases='2410015H04Rik, AI663973, AU021172, nnyRab5a', description='RAB5A, member RAS oncogene family', other_designations='', map_location='17 27.82 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=53479234, end_position_on_the_genomic_accession=53507678, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  140474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140474, CurrentID=0, Status='live', Symbol='Muc4', Aliases='4933405I11Rik, Asgp', description='mucin 4', other_designations='ascites sialoglycoprotein|mucin 4, ASGP|pancreatic adenocarcinoma mucin|testis mucin', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32735886, end_position_on_the_genomic_accession=32782391, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  140475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140475, CurrentID=0, Status='live', Symbol='Bsnd', Aliases='', description='Bartter syndrome, infantile, with sensorineural deafness (Barttin)', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106483456, end_position_on_the_genomic_accession=106492243, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  140476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140476, CurrentID=0, Status='live', Symbol='Strc', Aliases='DFNB16', description='stereocilin', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121363248, end_position_on_the_genomic_accession=121381370, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  140477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140477, CurrentID=0, Status='live', Symbol='Dmbx1', Aliases='Atx, Cdmx, Mbx, Otx3', description='diencephalon/mesencephalon homeobox 1', other_designations='diencephalon/mesencephalon-expressed brain homeobox gene 1 protein|dmbx1|homeobox gene Atx|orthodenticle homolog 3|paired-like homeobox protein DMBX1|paired-type homeobox Atx', map_location='4 53.07 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115915119, end_position_on_the_genomic_accession=115939926, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  140481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140481, CurrentID=0, Status='live', Symbol='Man2a2', Aliases='1700052O22Rik, 4931438M07Rik, AI480988, MX, Man IIx', description='mannosidase 2, alpha 2', other_designations='alpha mannosidase IIx|alpha-mannosidase IIx|mannosidase alpha class 2A member 2|mannosidase, alpha, class 2A, member 2|mannosyl-oligosaccharide 1,3-1,6-alpha-mannosidase', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80349097, end_position_on_the_genomic_accession=80371486, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  140482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140482, CurrentID=0, Status='live', Symbol='Zfp358', Aliases='Zfend', description='zinc finger protein 358', other_designations='', map_location='8 1.92 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3493075, end_position_on_the_genomic_accession=3497208, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  140483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140483, CurrentID=0, Status='live', Symbol='Hnmt', Aliases='1500031F01Rik, AI788969', description='histamine N-methyltransferase', other_designations='HMT', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24002911, end_position_on_the_genomic_accession=24049757, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  140484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140484, CurrentID=0, Status='live', Symbol='Pofut1', Aliases='O-FucT-1, mKIAA0180', description='protein O-fucosyltransferase 1', other_designations='peptide-O-fucosyltransferase 1', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153241532, end_position_on_the_genomic_accession=153271029, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  140486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140486, CurrentID=0, Status='live', Symbol='Igf2bp1', Aliases='AL024068, AW549074, CRD-BP, Crdbp, D030026A21Rik, D11Moh40e, D11Moh45, IMP-1, Neilsen, ZBP-1, Zbp1, mir-3063', description='insulin-like growth factor 2 mRNA binding protein 1', other_designations='IGF-II mRNA-binding protein 1|IGF2 mRNA-binding protein 1|VICKZ family member 1|coding region determinant binding protein|coding region determinant-binding protein|insulin-like growth factor 2 mRNA binding protein 1 dN CRDBP-2|insulin-like growth factor 2 mRNA binding protein 1 deltaN CRDBP|insulin-like growth factor 2, binding protein 1|zipcode binding protein 1|zipcode-binding protein 1', map_location='11 59.08 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95957163, end_position_on_the_genomic_accession=96005944, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  140488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140488, CurrentID=0, Status='live', Symbol='Igf2bp3', Aliases='2610101N11Rik, AA522010, AL022933, AU045931, IMP-3, IMP3, Koc13, Neilsen, mimp3', description='insulin-like growth factor 2 mRNA binding protein 3', other_designations='IGF-II mRNA-binding protein 3|IGF2 mRNA-binding protein 3|VICKZ family member 3|insulin-like growth factor 2, binding protein 3|mIMP-3', map_location='6 23.82 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49085218, end_position_on_the_genomic_accession=49215009, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  140489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140489, CurrentID=0, Status='live', Symbol='Bhlhe23', Aliases='A930001L02Rik, BETA4, Bhlhb4', description='basic helix-loop-helix family, member e23', other_designations='basic helix-loop-helix domain containing, class B4|beta-cell E-box transactivating factor 4|class B basic helix-loop-helix protein 4', map_location='2 103.34 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180774381, end_position_on_the_genomic_accession=180776900, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  140491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140491, CurrentID=0, Status='live', Symbol='Ppp1r3a', Aliases='GM, RG1, RGL', description='protein phosphatase 1, regulatory (inhibitor) subunit 3A', other_designations='protein phosphatase 1 glycogen-associated regulatory subunit|protein phosphatase type-1 glycogen targeting subunit', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=14713822, end_position_on_the_genomic_accession=14755274, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  140492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140492, CurrentID=0, Status='live', Symbol='Kcnn2', Aliases='KCa2.2, SK2, SKCA2, bc, fri', description='potassium intermediate/small conductance calcium-activated channel, subfamily N, member 2', other_designations='bouncy|small conductance calcium-activated potassium channel subfamily N member 2', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=45268764, end_position_on_the_genomic_accession=45685887, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  140493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140493, CurrentID=0, Status='live', Symbol='Kcnn3', Aliases='KCa2.3, SK3, SKCA3', description='potassium intermediate/small conductance calcium-activated channel, subfamily N, member 3', other_designations='SKCa 3|small conductance calcium-activated potassium channel 3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89520164, end_position_on_the_genomic_accession=89672494, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  140494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140494, CurrentID=0, Status='live', Symbol='Atp6v0a4', Aliases='Atp6n1b, a4', description='ATPase, H+ transporting, lysosomal V0 subunit A4', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) noncatalytic accessory protein 1B|V-ATPase 116 kDa|V-ATPase alpha 4|V-type proton ATPase 116 kDa subunit a|vacuolar proton translocating ATPase 100 kDa a4 subunit|vacuolar proton translocating ATPase 116 kDa subunit a kidney isoform', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38048483, end_position_on_the_genomic_accession=38124586, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  664783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664783, CurrentID=0, Status='live', Symbol='Dux', Aliases='AW8220734, Duxbl, EG664783, Dux', description='double homeobox', other_designations='homeobox protein', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.5', start_position_on_the_genomic_accession=57693399, end_position_on_the_genomic_accession=57695423, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  140497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140497, CurrentID=0, Status='live', Symbol='AF251705', Aliases='CLM4, Cd300c, Cd300d, DIgR1, Igsf7, Lmir2, MAIR-II', description='cDNA sequence AF251705', other_designations='CLM|Cd300C antigen|Cd300D antigen|MAIR-2|dendritic cell-derived Ig-like receptor 1|dendritic cell-derived immunoglobulin receptor 1|immunoglobulin superfamily member 7|immunoglobulin superfamily, member 7|leukocyte mono-Ig-like receptor 2|myeloid-associated immunoglobulin-like receptor 2|myeloid-associated immunoglobulin-like receptor II', map_location='11 80.57 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114996721, end_position_on_the_genomic_accession=115004357, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  140498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140498, CurrentID=0, Status='live', Symbol='Rxfp2', Aliases='Gpr106, Great, Lgr8', description='relaxin/insulin-like family peptide receptor 2', other_designations='G protein coupled receptor affecting testicular descent|G protein-coupled receptor 106|leucine-rich repeat-containing G protein-coupled receptor 8', map_location='5 89.23 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=150017781, end_position_on_the_genomic_accession=150082550, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  140499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140499, CurrentID=0, Status='live', Symbol='Ube2j2', Aliases='1200007B18Rik, 2400008G19Rik, 5730472G04Rik, AL022923, NCUBE-2, Ubc6, Ubc6p', description='ubiquitin-conjugating enzyme E2J 2', other_designations='Ubc6p homolog|non-canonical ubiquitin-conjugating enzyme 2|ubiquitin conjugating enzyme 6|ubiquitin-conjugating enzyme E2, J2 homolog', map_location='4 87.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155943813, end_position_on_the_genomic_accession=155959604, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  140500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140500, CurrentID=0, Status='live', Symbol='Acap3', Aliases='Centb5, Kiaa1716-hp, mKIAA1716', description='ArfGAP with coiled-coil, ankyrin repeat and PH domains 3', other_designations='centaurin, beta 5', map_location='4 87.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155891875, end_position_on_the_genomic_accession=155907251, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  664799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664799, CurrentID=0, Status='live', Symbol='Ctcfl', Aliases='Boris, OTTMUSG00000016680', description='CCCTC-binding factor (zinc finger protein)-like', other_designations='BORIS|Brother of the regulator of imprinted sites|CTCF paralog|CTCF-like protein|likely orthologue of H. sapiens CCCTC-binding factor-like protein (CTCFL)', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173093609, end_position_on_the_genomic_accession=173119525, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  664805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664805, CurrentID=0, Status='live', Symbol='Skor2', Aliases='Corl2, EG664805, Fussel18, Gm7348', description='SKI family transcriptional corepressor 2', other_designations='LBX1 corepressor 1-like protein|functional Smad suppressor element on chromosome 18|fussel-18 homolog|ladybird homeobox corepressor 1-like protein|transcriptional corepressor Corl2', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=76846734, end_position_on_the_genomic_accession=76901032, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  664829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664829, CurrentID=0, Status='live', Symbol='Slx', Aliases='EG664958, Xmr', description='Sycp3 like X-linked', other_designations='', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=26522657, end_position_on_the_genomic_accession=26545565, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  241877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241877, CurrentID=0, Status='live', Symbol='Slc10a5', Aliases='Gm405, mP5', description='solute carrier family 10 (sodium/bile acid cotransporter family), member 5', other_designations='Na(+)/bile acid cotransporter 5|solute carrier family 10 member 5', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10331734, end_position_on_the_genomic_accession=10335656, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  140546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140546, CurrentID=0, Status='live', Symbol='Eri3', Aliases='PINT1, Prnpip1', description='exoribonuclease 3', other_designations='prion interactor 1|prion protein interacting protein 1|prion protein-interacting protein', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117550293, end_position_on_the_genomic_accession=117674297, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  329260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329260, CurrentID=0, Status='live', Symbol='Dennd1b', Aliases='4632404N19Rik, 4930467M19Rik, 6820401H01Rik, F730008N07Rik', description='DENN/MADD domain containing 1B', other_designations='ASL1/9AS1-1 fusion|connecdenn 2', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=138963446, end_position_on_the_genomic_accession=139176043, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  140557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140557, CurrentID=0, Status='live', Symbol='Smc1b', Aliases='SMC-1B, SMC1beta, Smc1l2', description='structural maintenance of chromosomes 1B', other_designations='SMC protein 1B|SMC-1-beta|structural maintenace of chromosomes 1B', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85064689, end_position_on_the_genomic_accession=85131957, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  140559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140559, CurrentID=0, Status='live', Symbol='Igsf8', Aliases='ESTM34, EWI-2, KCT-4, PGRL', description='immunoglobulin superfamily, member 8', other_designations='CD81 partner 3|PG regulatory-like protein|glu-Trp-Ile EWI motif-containing protein 2|keratinocyte-associated transmembrane protein 4|prostaglandin regulatory-like protein', map_location='1 79.59 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172312407, end_position_on_the_genomic_accession=172319841, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  271639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271639, CurrentID=0, Status='live', Symbol='Adcy10', Aliases='4930431D04Rik, 4931412F17, Sacy, sAC', description='adenylate cyclase 10', other_designations='germ cell soluble adenylyl cyclase|testicular soluble adenylyl cyclase', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165483869, end_position_on_the_genomic_accession=165576778, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  140570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140570, CurrentID=0, Status='live', Symbol='Plxnb2', Aliases='Debt, plexin-B2', description='plexin B2', other_designations='differentially expressed in brain tumors', map_location='15 44.68 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89155546, end_position_on_the_genomic_accession=89181140, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  140571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140571, CurrentID=0, Status='live', Symbol='Plxnb3', Aliases='AI451018, Plxn6', description='plexin B3', other_designations='plexin 6|plexin-6', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73756559, end_position_on_the_genomic_accession=73772510, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  664862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664862, CurrentID=0, Status='live', Symbol='Gpr137b-ps', Aliases='2310041G17Rik, ENSMUSG00000075118', description='G protein-coupled receptor 137B, pseudogene', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12614065, end_position_on_the_genomic_accession=12650395, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  140577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140577, CurrentID=0, Status='live', Symbol='Ankrd6', Aliases='', description='ankyrin repeat domain 6', other_designations='diversin', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=32804027, end_position_on_the_genomic_accession=32950884, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  140579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140579, CurrentID=0, Status='live', Symbol='Elmo2', Aliases='1190002F24Rik, CED-12', description='engulfment and cell motility 2', other_designations='CED-12 homolog A', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165288031, end_position_on_the_genomic_accession=165326479, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  140580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140580, CurrentID=0, Status='live', Symbol='Elmo1', Aliases='6330578D22Rik, C230095H21Rik, CED-12', description='engulfment and cell motility 1', other_designations='engulfment and cell motility 1, ced-12 homolog', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=20090240, end_position_on_the_genomic_accession=20608353, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  402734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=402734, CurrentID=0, Status='live', Symbol='C730036E19Rik', Aliases='lncLSTR', description='RIKEN cDNA C730036E19 gene', other_designations='liver-specific triglyceride regulator', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151138034, end_position_on_the_genomic_accession=151144095, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  664883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664883, CurrentID=0, Status='live', Symbol='Nova1', Aliases='9430099M15Rik, G630039L02, Nova-1', description='neuro-oncological ventral antigen 1', other_designations='ventral neuron-specific protein 1', map_location='12 21.46 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=46694517, end_position_on_the_genomic_accession=46819194, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  271697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271697, CurrentID=0, Status='live', Symbol='Cdk15', Aliases='Als2cr7, Pftk2', description='cyclin-dependent kinase 15', other_designations='PFTAIRE protein kinase 2|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 7|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 7 protein homolog|cell division protein kinase 15|serine/threonine-protein kinase ALS2CR7|serine/threonine-protein kinase PFTAIRE-2', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59256889, end_position_on_the_genomic_accession=59352612, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  140629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140629, CurrentID=0, Status='live', Symbol='Ubox5', Aliases='1500010O06Rik, C330018L13Rik, Rnf37, UIP5, Ubce7ip5, Ube7ip5', description='U box domain containing 5', other_designations='U-box domain-containing protein 5|UbcM4-interacting protein 5|ubiquitin conjugating enzyme 7 interacting protein 5|ubiquitin-conjugating enzyme 7-interacting protein 5', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130589996, end_position_on_the_genomic_accession=130630038, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  140630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140630, CurrentID=0, Status='live', Symbol='Ube4a', Aliases='4732444G18Rik, 9930123J21Rik, UFD2b', description='ubiquitination factor E4A', other_designations='ubiquitination factor E4A, UFD2 homolog', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44923127, end_position_on_the_genomic_accession=44965600, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  329274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329274, CurrentID=0, Status='live', Symbol='Fam163a', Aliases='A230106N23Rik', description='family with sequence similarity 163, member A', other_designations='cebelin', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156075956, end_position_on_the_genomic_accession=156087888, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  271711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271711, CurrentID=0, Status='live', Symbol='Tmem169', Aliases='A830020B06Rik', description='transmembrane protein 169', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72284362, end_position_on_the_genomic_accession=72303104, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67134, CurrentID=0, Status='live', Symbol='Nop56', Aliases='2310044F10Rik, Nol5a', description='NOP56 ribonucleoprotein', other_designations='56kDa with KKE/D repeat|NOP56 ribonucleoprotein homolog|nucleolar protein 5A', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130274367, end_position_on_the_genomic_accession=130279313, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  664968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664968, CurrentID=0, Status='live', Symbol='Tmem238', Aliases='2210411K11Rik', description='transmembrane protein 238', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4784785, end_position_on_the_genomic_accession=4789560, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  140703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140703, CurrentID=0, Status='live', Symbol='Emid1', Aliases='AW122071, CO-5, Emu1', description='EMI domain containing 1', other_designations='Emu1|emilin and multimerin domain-containing protein 1', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5106265, end_position_on_the_genomic_accession=5152298, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  664994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=664994, CurrentID=0, Status='live', Symbol='Isoc2a', Aliases='Isoc2', description='isochorismatase domain containing 2a', other_designations='isoc2a|isochorismatase domain-containing protein 2A, mitochondrial', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4877053, end_position_on_the_genomic_accession=4895717, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  140709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140709, CurrentID=0, Status='live', Symbol='Col26a1', Aliases='9430032K24Rik, BC002218, Col26a, Emid2, Emu2', description='collagen, type XXVI, alpha 1', other_designations='Collagen XXVI|EMI domain containing 2|EMI domain-containing protein 2|Emu2|alpha-1 type XXVI collagen|emilin and multimerin domain-containing protein 2|procollagen, type XXVI, alpha 1', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136741759, end_position_on_the_genomic_accession=136883376, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  665001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665001, CurrentID=0, Status='live', Symbol='Gm14391', Aliases='OTTMUSG00000016543', description='predicted gene 14391', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=175185165, end_position_on_the_genomic_accession=175206924, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  271786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271786, CurrentID=0, Status='live', Symbol='Galnt13', Aliases='A230002A12, A230020F20, BB182356', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 13', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 13|galNAc-T13|polypeptide GalNAc transferase 13|pp-GaNTase 13|pp-GalNAc-T13|protein-UDP acetylgalactosaminyltransferase 13', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=54436387, end_position_on_the_genomic_accession=55118309, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  140721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140721, CurrentID=0, Status='live', Symbol='Caskin2', Aliases='1600028L06Rik, mKIAA1139', description='CASK-interacting protein 2', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115799183, end_position_on_the_genomic_accession=115813592, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  140723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140723, CurrentID=0, Status='live', Symbol='Cacng5', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 5', other_designations='TARP gamma-5|neuronal voltage-gated calcium channel gamma-5 subunit|transmembrane AMPAR regulatory protein gamma-5|voltage dependent calcium channel gamma 5 subunit', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107874604, end_position_on_the_genomic_accession=107915055, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  140740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140740, CurrentID=0, Status='live', Symbol='Sec63', Aliases='5730478J10Rik, AI649014, AW319215', description='SEC63-like (S. cerevisiae)', other_designations='', map_location='10 22.89 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=42761496, end_position_on_the_genomic_accession=42832514, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  271813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271813, CurrentID=0, Status='live', Symbol='Agbl2', Aliases='4930524K04, A430081C19Rik, CCP2', description='ATP/GTP binding protein-like 2', other_designations='ATP/GTP-binding protein-like 2|carboxypeptidase 2, cytosolic', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90782587, end_position_on_the_genomic_accession=90834437, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  140742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140742, CurrentID=0, Status='live', Symbol='Sesn1', Aliases='1110002G11Rik, AU044290, Pa26, Sest1', description='sestrin 1', other_designations='p53 regulated Pa26 nuclear protein|p53-regulated protein PA26', map_location='10 22.77 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41810574, end_position_on_the_genomic_accession=41908436, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  140743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140743, CurrentID=0, Status='live', Symbol='Rem2', Aliases='AW411893', description='rad and gem related GTP binding protein 2', other_designations='rad and Gem-like GTP-binding protein 2', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54476100, end_position_on_the_genomic_accession=54480434, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  665033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665033, CurrentID=0, Status='live', Symbol='Col6a5', Aliases='Col29a1, EG665033, Gm7455', description='collagen, type VI, alpha 5', other_designations='collagen alpha-1(XXIX) chain', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105856070, end_position_on_the_genomic_accession=105960643, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  140765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140765, CurrentID=0, Status='live', Symbol='Tmprss3', Aliases='', description='transmembrane protease, serine 3', other_designations='transmembrane proteinase serine 3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31179265, end_position_on_the_genomic_accession=31209743, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  271842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271842, CurrentID=0, Status='live', Symbol='Rpusd2', Aliases='4921503C21Rik, 9630001E10, BB231107', description='RNA pseudouridylate synthase domain containing 2', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119034790, end_position_on_the_genomic_accession=119042198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  271844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271844, CurrentID=0, Status='live', Symbol='Pla2g4f', Aliases='4732472I07Rik, Gm647, Pla2zeta', description='phospholipase A2, group IVF', other_designations='cPLA2-zeta', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120299952, end_position_on_the_genomic_accession=120314320, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  271849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271849, CurrentID=0, Status='live', Symbol='Shc4', Aliases='6230417E10Rik, 9930029B02Rik, Gm685, RaLP', description='SHC (Src homology 2 domain containing) family, member 4', other_designations='SH2 domain protein C4|SHC-transforming protein D|mShcD|novel phosphotyrosine interaction domain (PTB/PID) and SH2 domains containing protein|rai-like protein RaLP|src homology 2 domain-containing-transforming protein C4', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125627447, end_position_on_the_genomic_accession=125724160, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  140780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140780, CurrentID=0, Status='live', Symbol='Bmp2k', Aliases='4933417M22Rik, AA673486, AV128808, BIKE', description='BMP2 inducible kinase', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=96997689, end_position_on_the_genomic_accession=97091048, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  140781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140781, CurrentID=0, Status='live', Symbol='Myh7', Aliases='B-MHC, MyHC-I, Myhc-b, Myhcb, beta-MHC', description='myosin, heavy polypeptide 7, cardiac muscle, beta', other_designations='MYH-beta/slow|cardiac myosin heavy chain beta isoform|myHC-beta|myHC-slow|myosin heavy chain 7|myosin heavy chain slow isoform|myosin heavy chain, cardiac muscle beta isoform|myosin heavy chain, cardiac muscle, fetal', map_location='14 28.01 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54970688, end_position_on_the_genomic_accession=54994549, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  140792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140792, CurrentID=0, Status='live', Symbol='Colec12', Aliases='CL-P1, SRCL, Scara4', description='collectin sub-family member 12', other_designations='collectin placenta protein 1|scavenger receptor with C-type lectin', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=9707648, end_position_on_the_genomic_accession=9877995, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  140795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140795, CurrentID=0, Status='live', Symbol='P2ry14', Aliases='A330108O13Rik, Gpr105, P2Y14', description='purinergic receptor P2Y, G-protein coupled, 14', other_designations='G-protein coupled receptor 105|UDP-glucose receptor', map_location='3 28.96 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=59114623, end_position_on_the_genomic_accession=59153492, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  140806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140806, CurrentID=0, Status='live', Symbol='Il25', Aliases='IL-17e, Il17e', description='interleukin 25', other_designations='interleukin 17E', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54929401, end_position_on_the_genomic_accession=54935841, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  140810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140810, CurrentID=0, Status='live', Symbol='Ttbk2', Aliases='2610507N02Rik, AI326283, B930008N24Rik, TTK, Ttbk, Ttbk1, mKIAA0847', description='tau tubulin kinase 2', other_designations='protein bartleby', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120732816, end_position_on_the_genomic_accession=120850584, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  271887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271887, CurrentID=0, Status='live', Symbol='Rbbp8nl', Aliases='4732416F18', description='RBBP8 N-terminal like', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180275246, end_position_on_the_genomic_accession=180289900, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  665113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665113, CurrentID=0, Status='live', Symbol='Tnik', Aliases='1500031A17Rik, 4831440I19Rik, AI451411, C530008O15Rik, C630040K21Rik', description='TRAF2 and NCK interacting kinase', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=28263214, end_position_on_the_genomic_accession=28670585, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  665119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665119, CurrentID=0, Status='live', Symbol='Sec14l5', Aliases='', description='SEC14-like lipid binding 5', other_designations='SEC14-like 5', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5146901, end_position_on_the_genomic_accession=5187271, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  140858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140858, CurrentID=0, Status='live', Symbol='Wdr5', Aliases='2410008O07Rik, AA408785, AA960360, Big, Big-3', description='WD repeat domain 5', other_designations='BMP2-induced 3-kb gene protein|WD repeat-containing protein BIG-3', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27515147, end_position_on_the_genomic_accession=27536535, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  140859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140859, CurrentID=0, Status='live', Symbol='Nek8', Aliases='4632401F23Rik, b2b1449Clo, jck', description='NIMA (never in mitosis gene a)-related expressed kinase 8', other_designations='NIMA-related expressed kinase 8|juvenile cystic kidney|never in mitosis A-related kinase 8|nimA-related protein kinase 8', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78166106, end_position_on_the_genomic_accession=78176695, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  665155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665155, CurrentID=0, Status='live', Symbol='Srp54b', Aliases='', description='signal recognition particle 54B', other_designations='SRP54', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55155104, end_position_on_the_genomic_accession=55189213, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  140887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140887, CurrentID=0, Status='live', Symbol='Lnx2', Aliases='9630046H24, AW209022', description='ligand of numb-protein X 2', other_designations='numb-binding protein 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147016655, end_position_on_the_genomic_accession=147076572, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  271970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271970, CurrentID=0, Status='live', Symbol='Arsj', Aliases='9330196J05Rik, D830047F08', description='arylsulfatase J', other_designations='ASJ', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=126363250, end_position_on_the_genomic_accession=126440379, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140904, CurrentID=0, Status='live', Symbol='Caln1', Aliases='9630012C17Rik, Cabp8, MNCb-0849', description='calneuron 1', other_designations='calneuron I|calneuron-1|calneuron-I', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130369356, end_position_on_the_genomic_accession=130846325, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  271981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271981, CurrentID=0, Status='live', Symbol='Tbck', Aliases='1700120J03Rik, 9430001M19, A630047E20Rik, C030007I09Rikl, Tbck', description='TBC1 domain containing kinase', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=132683994, end_position_on_the_genomic_accession=132841688, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  140917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140917, CurrentID=0, Status='live', Symbol='Dclre1b', Aliases='AI452214, Apollo, SNMIB, mSNM1B', description='DNA cross-link repair 1B', other_designations='DNA cross-link repair 1B protein|DNA cross-link repair 1B, PSO2 homolog|SNM1 homolog B', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103800605, end_position_on_the_genomic_accession=103809387, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  140918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140918, CurrentID=0, Status='live', Symbol='Slc7a12', Aliases='Asc-2, XAT1', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 12', other_designations='Asc-2,asc-type amino acid transporter 2|asc-type amino acid transporter 2', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14480691, end_position_on_the_genomic_accession=14505819, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  140919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140919, CurrentID=0, Status='live', Symbol='Slc17a6', Aliases='2900073D12Rik, DNPI, VGLUT2', description='solute carrier family 17 (sodium-dependent inorganic phosphate cotransporter), member 6', other_designations='differentiation-associated BNPI|differentiation-associated Na(+)-dependent inorganic phosphate cotransporter|solute carrier family 17 (vesicular glutamate transporter), member 6|solute carrier family 17 member 6', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=51621830, end_position_on_the_genomic_accession=51671126, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  665211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665211, CurrentID=0, Status='live', Symbol='Gm14326', Aliases='OTTMUSG00000016327', description='predicted gene 14326', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=177935993, end_position_on_the_genomic_accession=177958136, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  329324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329324, CurrentID=0, Status='live', Symbol='Syt14', Aliases='B230320I09Rik, sytXIV', description='synaptotagmin XIV', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=192891233, end_position_on_the_genomic_accession=193035862, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  403088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403088, CurrentID=0, Status='live', Symbol='Tcaf3', Aliases='Eapa2, Fam115e', description='TRPM8 channel-associated factor 3', other_designations='experimental autoimmune prostatitis antigen 2|family with sequence similarity 115, member E|protein FAM115E', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42587213, end_position_on_the_genomic_accession=42597372, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  272027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272027, CurrentID=0, Status='live', Symbol='Tstd2', Aliases='3010020C06', description='thiosulfate sulfurtransferase (rhodanese)-like domain containing 2', other_designations='likely ortholog of H. sapiens chromosome 9 open reading frame 97 (C9orf97)', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46114746, end_position_on_the_genomic_accession=46138581, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  272031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272031, CurrentID=0, Status='live', Symbol='Plppr1', Aliases='E130309F12Rik, Lppr1, PRG-3, mKIAA4247', description='phospholipid phosphatase related 1', other_designations='plasticity-related gene 3 protein|plasticity-related protein 3', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49059462, end_position_on_the_genomic_accession=49340261, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  665270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665270, CurrentID=0, Status='live', Symbol='Plb1', Aliases='4632413E21Rik, 4930433E17Rik, 4930539A06Rik, BC033606', description='phospholipase B1', other_designations='PLB/LIP|phospholipase B|phospholipase B/lipase', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=32232691, end_position_on_the_genomic_accession=32364356, orientation='plus', exon_count=61, OMIM='', no_hdr0=''),
  403171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403171, CurrentID=0, Status='live', Symbol='Banf2', Aliases='4930517K23Rik, Baf-L, Baf-like, Gm115', description='barrier to autointegration factor 2', other_designations='barrier-to-autointegration factor 2|likely ortholog of H. sapiens Chromosome 20 open reading frame 179 (C20orf179)', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144033102, end_position_on_the_genomic_accession=144073979, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  403174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403174, CurrentID=0, Status='live', Symbol='Msantd1', Aliases='A930005I04Rik, Gm144', description='Myb/SANT-like DNA-binding domain containing 1', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34907775, end_position_on_the_genomic_accession=34923853, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  403175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403175, CurrentID=0, Status='live', Symbol='Tigd4', Aliases='C130063O11Rik, Gm418', description='tigger transposable element derived 4', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84593574, end_position_on_the_genomic_accession=84597032, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  403178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403178, CurrentID=0, Status='live', Symbol='Plcxd1', Aliases='A330045H12Rik, Gm576', description='phosphatidylinositol-specific phospholipase C, X domain containing 1', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110099225, end_position_on_the_genomic_accession=110105953, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  403180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403180, CurrentID=0, Status='live', Symbol='Ccdc121', Aliases='6530421E24Rik, Gm755', description='coiled-coil domain containing 121', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181509633, end_position_on_the_genomic_accession=181511451, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  403183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403183, CurrentID=0, Status='live', Symbol='Mettl21e', Aliases='4832428D23Rik, Gm991', description='methyltransferase like 21E', other_designations='methyltransferase-like protein 21C pseudogene 1 homolog|methyltransferase-like protein 21E pseudogene homolog|methyltransferase-like protein LOC121952 homolog', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44204070, end_position_on_the_genomic_accession=44218931, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  403187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403187, CurrentID=0, Status='live', Symbol='Opa3', Aliases='D630048P19Rik, Gm1425, Gm472', description='optic atrophy 3', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19228389, end_position_on_the_genomic_accession=19246817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  403200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403200, CurrentID=0, Status='live', Symbol='4930504O13Rik', Aliases='Gm524', description='RIKEN cDNA 4930504O13 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58446143, end_position_on_the_genomic_accession=58452966, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  403205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403205, CurrentID=0, Status='live', Symbol='Agr3', Aliases='BCMP11, E030025L21Rik, Gm888', description='anterior gradient 3', other_designations='anterior gradient homolog 3|anterior gradient protein 3 homolog', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=35925621, end_position_on_the_genomic_accession=35949737, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  272158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272158, CurrentID=0, Status='live', Symbol='Poln', Aliases='POL4P', description='DNA polymerase N', other_designations='DNA-directed DNA polymerase nu', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34007179, end_position_on_the_genomic_accession=34169526, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  665378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665378, CurrentID=0, Status='live', Symbol='Gm7609', Aliases='EG665378', description='predicted pseudogene 7609', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85199592, end_position_on_the_genomic_accession=85215395, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  665433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665433, CurrentID=0, Status='live', Symbol='Hist1h2ao', Aliases='', description='histone cluster 1, H2ao', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21810367, end_position_on_the_genomic_accession=21810918, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  329360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329360, CurrentID=0, Status='live', Symbol='Rnf224', Aliases='Gm757', description='ring finger protein 224', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25234476, end_position_on_the_genomic_accession=25236787, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  329366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329366, CurrentID=0, Status='live', Symbol='Ccdc187', Aliases='4932418E24Rik', description='coiled-coil domain containing 187', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26243449, end_position_on_the_genomic_accession=26300476, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  329375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329375, CurrentID=0, Status='live', Symbol='Cfap77', Aliases='1700101E01Rik, Gm823', description='cilia and flagella associated protein 77', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28927749, end_position_on_the_genomic_accession=29055274, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  272322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272322, CurrentID=0, Status='live', Symbol='Arntl2', Aliases='4632430A05Rik, BMAL2, CLIF, MOP9, bHLHe6', description='aryl hydrocarbon receptor nuclear translocator-like 2', other_designations='brain and muscle ARNT-like 2|brain-muscle-ARNT-like protein 2', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146796015, end_position_on_the_genomic_accession=146833529, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  403395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=403395, CurrentID=0, Status='live', Symbol='Clec3a', Aliases='1110019O10Rik, Clecsf1, Gm796', description='C-type lectin domain family 3, member a', other_designations='C-type calcium dependent, carbohydrate-recognition domain lectin, superfamily member 1, cartilage-derived', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=114418034, end_position_on_the_genomic_accession=114428133, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102639566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102639566, CurrentID=0, Status='live', Symbol='Ptgs2os2', Aliases='Gm26687, Linc-cox2, Lincrna-cox2', description='prostaglandin-endoperoxide synthase 2, opposite strand 2', other_designations='long non-coding RNA Cox2', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=150159043, end_position_on_the_genomic_accession=150164948, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  272347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272347, CurrentID=0, Status='live', Symbol='Zfp398', Aliases='5730513I23Rik', description='zinc finger protein 398', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47835661, end_position_on_the_genomic_accession=47873537, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  272359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272359, CurrentID=0, Status='live', Symbol='Irf2bp1', Aliases='6330414O09Rik, C530033F24', description='interferon regulatory factor 2 binding protein 1', other_designations='IRF-2-binding protein 1|IRF-2BP1|probable E3 ubiquitin-protein ligase IRF2BP1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19004065, end_position_on_the_genomic_accession=19006763, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102639598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102639598, CurrentID=0, Status='live', Symbol='Gm14296', Aliases='', description='predicted gene 14296', other_designations='novel KRAB box and zinc finger, C2H2 type domain containing protein|zinc finger protein 431-like', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=176902774, end_position_on_the_genomic_accession=176927930, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  665596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665596, CurrentID=0, Status='live', Symbol='Hist1h2bq', Aliases='Gm13646, OTTMUSG00000013203', description='histone cluster 1, H2bq', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21806412, end_position_on_the_genomic_accession=21810199, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  272381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272381, CurrentID=0, Status='live', Symbol='Lrrc4b', Aliases='A830007M12, Lrig4', description='leucine rich repeat containing 4B', other_designations='NGL-3|leucine-rich repeats and immunoglobulin-like domains 4|netrin-G3 ligand', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44442487, end_position_on_the_genomic_accession=44463351, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  272382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272382, CurrentID=0, Status='live', Symbol='Spib', Aliases='Spi-B', description='Spi-B transcription factor (Spi-1/PU.1 related)', other_designations='Ets transcription factor Spi-B|Spi-1/PU.1 related', map_location='7 28.83 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44525995, end_position_on_the_genomic_accession=44532086, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  272396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272396, CurrentID=0, Status='live', Symbol='Tarsl2', Aliases='A530046H20Rik', description='threonyl-tRNA synthetase-like 2', other_designations='probable threonyl-tRNA synthetase 2, cytoplasmic|thrRS|threonine--tRNA ligase|threonyl-tRNA synthetase|threonyl-tRNA synthetase-like protein 2', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=65644898, end_position_on_the_genomic_accession=65692093, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  665622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665622, CurrentID=0, Status='live', Symbol='Hist1h2br', Aliases='Gm11277, OTTMUSG00000000449', description='histone cluster 1 H2br', other_designations='H2b histone family member', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21833743, end_position_on_the_genomic_accession=21837530, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  272411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272411, CurrentID=0, Status='live', Symbol='B3gnt6', Aliases='BC039789, Core-3, Core3', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 6 (core 3 synthase)', other_designations='BGnT-6|beta-1,3-Gn-T6|beta-1,3-N-acetylglucosaminyltransferase 6|beta3Gn-T6|core 3 synthase', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98192415, end_position_on_the_genomic_accession=98199475, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  272428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272428, CurrentID=0, Status='live', Symbol='Acsm5', Aliases='C730019D22, C730027J19Rik', description='acyl-CoA synthetase medium-chain family member 5', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119526131, end_position_on_the_genomic_accession=119543360, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  272465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272465, CurrentID=0, Status='live', Symbol='Tmem255b', Aliases='Fam70b, Gm687', description='transmembrane protein 255B', other_designations='family with sequence similarity 70, member B|protein FAM70B', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13435165, end_position_on_the_genomic_accession=13483667, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  665700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665700, CurrentID=0, Status='live', Symbol='Hmcn2', Aliases='D030068H01', description='hemicentin 2', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31313991, end_position_on_the_genomic_accession=31460738, orientation='plus', exon_count=99, OMIM='', no_hdr0=''),
  272538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272538, CurrentID=0, Status='live', Symbol='Tango6', Aliases='AW413431, E330010G16, Tmco7', description='transport and golgi organization 6', other_designations='transmembrane and coiled-coil domain-containing protein 7|transmembrane and coiled-coil domains 7|transport and Golgi organization 6 homolog', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106683068, end_position_on_the_genomic_accession=106851439, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  272551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272551, CurrentID=0, Status='live', Symbol='Gins2', Aliases='2210013I18Rik, 4833427B12Rik, AI323585, Pfs2', description='GINS complex subunit 2 (Psf2 homolog)', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120581266, end_position_on_the_genomic_accession=120589075, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  665775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665775, CurrentID=0, Status='live', Symbol='Bod1l', Aliases='A230054D04Rik, AI853319, FAM44A, mKIAA1327', description='biorientation of chromosomes in cell division 1-like', other_designations='family with sequence similarity 44, member A', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=41787538, end_position_on_the_genomic_accession=41844365, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  67272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67272, CurrentID=0, Status='live', Symbol='Cmtm5', Aliases='1500005P16Rik, 2900052H21Rik, Cklfsf5', description='CKLF-like MARVEL transmembrane domain containing 5', other_designations='chemokine-like factor super family 5|chemokine-like factor super family five variant 4|chemokine-like factor superfamily 5|chemokine-like factor superfamily member 5', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54936470, end_position_on_the_genomic_accession=54939277, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  665780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665780, CurrentID=0, Status='live', Symbol='Zscan4b', Aliases='EG665780', description='zinc finger and SCAN domain containing 4B', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=10900740, end_position_on_the_genomic_accession=10905050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  635253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=635253, CurrentID=0, Status='live', Symbol='Usp51', Aliases='AV136873', description='ubiquitin specific protease 51', other_designations='ubiquitin specific peptidase 51', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153006381, end_position_on_the_genomic_accession=153009913, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  272589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272589, CurrentID=0, Status='live', Symbol='Tbcel', Aliases='D330014K23, E130107N23Rik, Lrrc35', description='tubulin folding cofactor E-like', other_designations='leucine rich repeat containing 35|leucine-rich repeat-containing protein 35', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=42412316, end_position_on_the_genomic_accession=42472275, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  665848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665848, CurrentID=0, Status='live', Symbol='Zscan4e', Aliases='EG665848', description='zinc finger and SCAN domain containing 4E', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11306369, end_position_on_the_genomic_accession=11310682, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  272636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272636, CurrentID=0, Status='live', Symbol='Esyt3', Aliases='D930024E11, D9Ertd280e, Fam62c, mKIAA4186', description='extended synaptotagmin-like protein 3', other_designations='E-Syt3|family with sequence similarity 62 (C2 domain containing), member C|family with sequence similarity 62, member C', map_location='9 51.54 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99308739, end_position_on_the_genomic_accession=99360405, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  272643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272643, CurrentID=0, Status='live', Symbol='Prss43', Aliases='Gm661, Tessp3', description='protease, serine 43', other_designations='protease, serine, 43|testis serine protease 3', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110826690, end_position_on_the_genomic_accession=110831507, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  665902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665902, CurrentID=0, Status='live', Symbol='Zscan4f', Aliases='C330039H09Rik', description='zinc finger and SCAN domain containing 4F', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11397915, end_position_on_the_genomic_accession=11402318, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  665943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665943, CurrentID=0, Status='live', Symbol='E330014E10Rik', Aliases='', description='RIKEN cDNA E330014E10 gene', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=95775870, end_position_on_the_genomic_accession=95804230, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100861668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100861668, CurrentID=0, Status='live', Symbol='Gm21119', Aliases='', description='predicted gene, 21119', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=20614405, end_position_on_the_genomic_accession=20657119, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  666002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666002, CurrentID=0, Status='live', Symbol='Vmn2r114', Aliases='EG666002, V2Rp3', description='vomeronasal 2, receptor 114', other_designations='vomeronasal 2 receptor p3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23290934, end_position_on_the_genomic_accession=23312313, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  272790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=272790, CurrentID=0, Status='live', Symbol='Magee2', Aliases='9630059J11Rik, BB131182, Mage-e2', description='melanoma antigen, family E, 2', other_designations='', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=104854952, end_position_on_the_genomic_accession=104857267, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  666048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666048, CurrentID=0, Status='live', Symbol='Trabd2b', Aliases='Gm12824, Hkat, OTTMUSG00000008561', description='TraB domain containing 2B', other_designations='TRAB domain-containing protein 2B|UPF0632 protein A|heart, kidney and adipose-enriched transmembrane protein', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=114406194, end_position_on_the_genomic_accession=114615098, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  666060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666060, CurrentID=0, Status='live', Symbol='Frmpd1', Aliases='BC031840, mKIAA0967', description='FERM and PDZ domain containing 1', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45184417, end_position_on_the_genomic_accession=45285936, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  329470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329470, CurrentID=0, Status='live', Symbol='Accs', Aliases='2610203E10Rik, Phacs', description='1-aminocyclopropane-1-carboxylate synthase (non-functional)', other_designations='1-aminocyclopropane-1-carboxylate synthase homolog|ACC synthase-like protein 1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93833467, end_position_on_the_genomic_accession=93849995, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  666145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666145, CurrentID=0, Status='live', Symbol='1700016G14Rik', Aliases='', description='RIKEN cDNA 1700016G14 gene', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24733875, end_position_on_the_genomic_accession=24746592, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  666173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666173, CurrentID=0, Status='live', Symbol='Vps13b', Aliases='1810042B05Rik, 2310042E16Rik, 4732488H20, C330002D13Rik, C87206, Coh1, D230005K13, mKIAA0532, mKIAA5032', description='vacuolar protein sorting 13B (yeast)', other_designations='Cohen syndrome homolog 1|cohen syndrome protein 1 homolog', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=35369882, end_position_on_the_genomic_accession=35931229, orientation='plus', exon_count=66, OMIM='', no_hdr0=''),
  666185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666185, CurrentID=0, Status='live', Symbol='Gm7969', Aliases='', description='predicted gene 7969', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67949751, end_position_on_the_genomic_accession=67954507, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  666279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666279, CurrentID=0, Status='live', Symbol='Dspp', Aliases='Dmp2, Dmp3, Dpp, Dsp', description='dentin sialophosphoprotein', other_designations='DMP-3|dentin matrix protein 3', map_location='5 50.59 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104170712, end_position_on_the_genomic_accession=104180127, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  329502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329502, CurrentID=0, Status='live', Symbol='Pla2g4e', Aliases='2310026J01Rik, C230096D22, Pla2epsilon', description='phospholipase A2, group IVE', other_designations='cPLA2-epsilon|phospholipase A2, group 4E', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120166412, end_position_on_the_genomic_accession=120245625, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  666317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666317, CurrentID=0, Status='live', Symbol='Prl2c1', Aliases='Ghd22', description='Prolactin family 2, subfamily c, member 1', other_designations='growth hormone d22', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27849327, end_position_on_the_genomic_accession=27857791, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  404194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404194, CurrentID=0, Status='live', Symbol='Gfral', Aliases='AY457637, Gral', description='GDNF family receptor alpha like', other_designations='GDNF receptor alpha-like', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=76164102, end_position_on_the_genomic_accession=76213657, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  666339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666339, CurrentID=0, Status='live', Symbol='Muc3', Aliases='', description='mucin 3, intestinal', other_designations='CAAA01224633.1', map_location='5 76.22 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137134924, end_position_on_the_genomic_accession=137166048, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  666348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666348, CurrentID=0, Status='live', Symbol='Apol7e', Aliases='ENSMUSG00000071716', description='apolipoprotein L 7e', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77695719, end_position_on_the_genomic_accession=77719288, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  404212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404212, CurrentID=0, Status='live', Symbol='Pisrt1', Aliases='', description='polled intersex syndrome regulated transcript 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  404238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404238, CurrentID=0, Status='live', Symbol='Mrgprb3', Aliases='MrgB3', description='MAS-related GPR, member B3', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48642863, end_position_on_the_genomic_accession=48643801, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404239, CurrentID=0, Status='live', Symbol='Mrgprb5', Aliases='Mgrg11, MrgB5', description='MAS-related GPR, member B5', other_designations='Mas-related G protein-coupled receptor g11', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48168017, end_position_on_the_genomic_accession=48168985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404240, CurrentID=0, Status='live', Symbol='Mrgprb8', Aliases='MrgB8', description='MAS-related GPR, member B8', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48388526, end_position_on_the_genomic_accession=48389648, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404242, CurrentID=0, Status='live', Symbol='Mrgprx1', Aliases='MrgC11, Mrgprc11', description='MAS-related GPR, member X1', other_designations='MAS-related GPR, member C11|mas-related G-protein coupled receptor member C11|mrgprx1|sensory neuron-specific G-protein coupled receptor 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48020971, end_position_on_the_genomic_accession=48027597, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  404284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404284, CurrentID=0, Status='live', Symbol='Vmn1r59', Aliases='V1rd10', description='vomeronasal 1 receptor 59', other_designations='vomernasal 1 receptor Vmn1r59|vomeronasal 1 receptor, D10', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5453827, end_position_on_the_genomic_accession=5454759, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404285, CurrentID=0, Status='live', Symbol='Vmn1r64', Aliases='V1rd11', description='vomeronasal 1 receptor 64', other_designations='vomernasal 1 receptor Vmn1r64|vomeronasal 1 receptor, D11', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5883580, end_position_on_the_genomic_accession=5884542, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404286, CurrentID=0, Status='live', Symbol='Vmn1r179', Aliases='V1rd17', description='vomeronasal 1 receptor 179', other_designations='vomernasal 1 receptor Vmn1r179|vomeronasal 1 receptor, D17', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23928386, end_position_on_the_genomic_accession=23929357, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404287, CurrentID=0, Status='live', Symbol='V1rd19', Aliases='Vmn1r182', description='vomeronasal 1 receptor, D19', other_designations='vomernasal 1 receptor Vmn1r182', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24003111, end_position_on_the_genomic_accession=24004028, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404289, CurrentID=0, Status='live', Symbol='Vmn1r181', Aliases='V1rd20', description='vomeronasal 1 receptor 181', other_designations='vomernasal 1 receptor Vmn1r181', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23983961, end_position_on_the_genomic_accession=23985048, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404290, CurrentID=0, Status='live', Symbol='Vmn1r93', Aliases='V1rd21', description='vomeronasal 1 receptor 93', other_designations='vomernasal 1 receptor Vmn1r93|vomeronasal 1 receptor, D21', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=20138123, end_position_on_the_genomic_accession=20139088, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404291, CurrentID=0, Status='live', Symbol='Vmn1r174', Aliases='V1rd22', description='vomeronasal 1 receptor 174', other_designations='vomernasal 1 receptor Vmn1r174', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23753911, end_position_on_the_genomic_accession=23754852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404308, CurrentID=0, Status='live', Symbol='Olfr118', Aliases='MOR263-13', description='olfactory receptor 118', other_designations='GA_x5J8B7TVMCU-697-290|GA_x5J8B7TYMFE-883-533|GA_x6K02T2PSCP-2131124-2132089', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37672025, end_position_on_the_genomic_accession=37672990, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404316, CurrentID=0, Status='live', Symbol='Olfr403', Aliases='IA7, MOR125-5_p, Olfr43', description='olfactory receptor 403', other_designations='GA_x6K02T2P1NL-4348188-4349129|olfactory receptor 43', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74186249, end_position_on_the_genomic_accession=74196446, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  404317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404317, CurrentID=0, Status='live', Symbol='Olfr592', Aliases='MOR0-3P, MOR32-13, Olfr1525-ps1', description='olfactory receptor 592', other_designations='GA_x6K02T2PBJ9-5902266-5903204|olfactory receptor 1525, pseudogene 1|olfactory receptor MOR0-3P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103186603, end_position_on_the_genomic_accession=103187541, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404318, CurrentID=0, Status='live', Symbol='Olfr681', Aliases='MOR4014', description='olfactory receptor 681', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105121459, end_position_on_the_genomic_accession=105122406, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404319, CurrentID=0, Status='live', Symbol='Olfr750', Aliases='GA_x5J8B7W5WBF-6267395-6266441, GA_x6K02T2PMLR-6808276-6807281, MOR103-18', description='olfactory receptor 750', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51070311, end_position_on_the_genomic_accession=51071442, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404322, CurrentID=0, Status='live', Symbol='Olfr924', Aliases='MOR171-27P, MOR171-47, Olfr1520-ps1', description='olfactory receptor 924', other_designations='GA_x6K02T2PVTD-32543982-32544908|olfactory receptor 1520, pseudogene 1|olfactory receptor MOR171-27P', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38848116, end_position_on_the_genomic_accession=38849042, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  666468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666468, CurrentID=0, Status='live', Symbol='Atg4a', Aliases='AI627006, AV169859, Apg4al, Autl2, Atg4a', description='autophagy related 4A, cysteine peptidase', other_designations='APG4 (ATG4) autophagy-related homolog A|AUT-like 2 cysteine endopeptidase|AUT-like 2, cysteine endopeptidase|autophagin 2|autophagin-2|autophagy 4 homolog A|autophagy-related 4A|autophagy-related cysteine endopeptidase 2|autophagy-related protein 4 homolog A|cysteine protease APG4A', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=141158699, end_position_on_the_genomic_accession=141164270, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  404325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404325, CurrentID=0, Status='live', Symbol='Olfr1057', Aliases='GA_x5J8B7TRQ4E-1174-939, GA_x5J8B7TY464-2-187, MOR185-11', description='olfactory receptor 1057', other_designations='GA_x6K02T2Q125-47844843-47843896|olfactory receptor GA_x5J8B7TRQ4E-1174-939|olfactory receptor GA_x5J8B7TY464-2-187', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86374463, end_position_on_the_genomic_accession=86375410, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404330, CurrentID=0, Status='live', Symbol='Olfr1198', Aliases='GA_x5J8B7W62NC-67655-66851, GA_x6K02T2Q125-50221692-50220766, MOR225-13', description='olfactory receptor 1198', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88745960, end_position_on_the_genomic_accession=88746886, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404331, CurrentID=0, Status='live', Symbol='Olfr1252', Aliases='GA_x5J8B7W62NC-992588-991951', description='olfactory receptor 1252', other_designations='GA_x6K02T2Q125-51162884-51161940|MOR231-22_p|olfactory receptor GA_x5J8B7W62NC-992588-991951', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89721165, end_position_on_the_genomic_accession=89722109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404335, CurrentID=0, Status='live', Symbol='Olfr1535', Aliases='MOR256-36, MOR256-63, Olfr1365', description='olfactory receptor 1535', other_designations='GA_x6K02T2QHY8-11688984-11689964|olfactory receptor 1365|olfactory receptor MOR256-36', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21555040, end_position_on_the_genomic_accession=21556020, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  404337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404337, CurrentID=0, Status='live', Symbol='Olfr1383', Aliases='MOR256-56', description='olfactory receptor 1383', other_designations='GA_x6K02T2QP88-5912627-5911692|odorant receptor|olfactory receptor GA_x5J8B7W6B6J-2556603-2556836', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49523725, end_position_on_the_genomic_accession=49524660, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  404346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404346, CurrentID=0, Status='live', Symbol='Olfr322', Aliases='GA_x5J8B7U1E65-1-838, MOR256-64, OTTMUSG00000005755, Olfr179, Olfr179-ps1, Olfr216, Olfr216-ps1-ps1, Olfr322', description='olfactory receptor 322', other_designations='GA_x6K02SYUNE5-2-727|GA_x6K02SYUPNB-557-3|GA_x6K02T2NKPP-754816-754257|olfactory receptor 216|olfactory receptor GA_x5J8B7U1E65-1-838', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58665561, end_position_on_the_genomic_accession=58666544, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  666528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666528, CurrentID=0, Status='live', Symbol='Zfp541', Aliases='EG666528, Znf541', description='zinc finger protein 541', other_designations='spermatogenic cell HDAC-interacting protein 1', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16061775, end_position_on_the_genomic_accession=16096340, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  666532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666532, CurrentID=0, Status='live', Symbol='Gm13139', Aliases='OTTMUSG00000010657', description='predicted gene 13139', other_designations='novel KRAB box domain containing protein', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147131931, end_position_on_the_genomic_accession=147181762, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  30924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30924, CurrentID=0, Status='live', Symbol='Angptl3', Aliases='hypl', description='angiopoietin-like 3', other_designations='angiopoietin-like protein 3', map_location='4 45.6 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=99030955, end_position_on_the_genomic_accession=99038188, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  30926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30926, CurrentID=0, Status='live', Symbol='Glrx3', Aliases='PICOT, Txnl2', description='glutaredoxin 3', other_designations='PKC interacting cousin of thioredoxin|PKC-interacting cousin of thioredoxin|PKC-theta-interacting protein|PKCq-interacting protein|thioredoxin-like 2|thioredoxin-like protein 2', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=137437643, end_position_on_the_genomic_accession=137468594, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  404473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404473, CurrentID=0, Status='live', Symbol='Olfr1082', Aliases='GA_x5J8B7W4CQV-18360-18784, MOR192-4_p-ps1, Olfr1082', description='olfactory receptor 1082', other_designations='GA_x6K02T2Q125-48079993-48079157|olfactory receptor 1082, pseudogene 1|olfactory receptor GA_x5J8B7W4CQV-18360-18784', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86593885, end_position_on_the_genomic_accession=86598809, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  11287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11287, CurrentID=0, Status='live', Symbol='Pzp', Aliases='A1m, A2m, AI893533, MAM', description='pregnancy zone protein', other_designations='alpha 1 macroglobulin|alpha-2-M|alpha-2-macroglobulin', map_location='6 F1-G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128483567, end_position_on_the_genomic_accession=128526720, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  11298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11298, CurrentID=0, Status='live', Symbol='Aanat', Aliases='AA-NAT, Nat-2, Nat4, Snat', description='arylalkylamine N-acetyltransferase', other_designations='aralkylamine N-acetyltransferase|serotonin acetylase', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116591721, end_position_on_the_genomic_accession=116597680, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11302, CurrentID=0, Status='live', Symbol='Aatk', Aliases='AATYK, aatyk1, mKIAA0641', description='apoptosis-associated tyrosine kinase', other_designations='apoptosis-associated tyrosine kinase 1B|brain apoptosis-associated tyrosine kinase|lemur tyrosine kinase 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120007311, end_position_on_the_genomic_accession=120047182, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11303, CurrentID=0, Status='live', Symbol='Abca1', Aliases='ABC-1, Abc1', description='ATP-binding cassette, sub-family A (ABC1), member 1', other_designations='ATP-binding cassette 1|ATP-binding cassette transporter 1', map_location='4 A5-B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=53030789, end_position_on_the_genomic_accession=53159988, orientation='minus', exon_count=52, OMIM='', no_hdr0=''),
  11304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11304, CurrentID=0, Status='live', Symbol='Abca4', Aliases='AW050280, Abc10, Abcr, D430003I15Rik, RmP', description='ATP-binding cassette, sub-family A (ABC1), member 4', other_designations='ATP-binding cassette 10|ATP-binding cassette sub-family A member 4|RIM ABC transporter|Rim protein', map_location='3 52.94 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122043893, end_position_on_the_genomic_accession=122180070, orientation='plus', exon_count=54, OMIM='', no_hdr0=''),
  11305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11305, CurrentID=0, Status='live', Symbol='Abca2', Aliases='AI413825, Abc2, D2H0S1474E, mKIAA1062', description='ATP-binding cassette, sub-family A (ABC1), member 2', other_designations='ATP-binding cassette 2|ATP-binding cassette transporter 2', map_location='2 A2-B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25428604, end_position_on_the_genomic_accession=25448544, orientation='plus', exon_count=50, OMIM='', no_hdr0=''),
  11306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11306, CurrentID=0, Status='live', Symbol='Abcb7', Aliases='AA517758, AU019072, Abc7', description='ATP-binding cassette, sub-family B (MDR/TAP), member 7', other_designations='ABC transporter 7 protein|ATP-binding cassette 7|ATP-binding cassette transporter 7', map_location='X C-D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=104280565, end_position_on_the_genomic_accession=104413846, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  11307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11307, CurrentID=0, Status='live', Symbol='Abcg1', Aliases='AW413978, Abc8, White', description='ATP-binding cassette, sub-family G (WHITE), member 1', other_designations='ATP-binding cassette 8|ATP-binding cassette transporter 8|ATP-binding cassette, subfamily G, member 1|white protein homolog', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31057694, end_position_on_the_genomic_accession=31117984, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  11308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11308, CurrentID=0, Status='live', Symbol='Abi1', Aliases='E3B1, NAP1, Ssh3bp1, abi-1', description='abl-interactor 1', other_designations='abelson interactor 1|ablphilin-1|eps8 SH3 domain-binding protein|eps8 binding protein|eps8-binding protein|spectrin SH3 domain binding protein 1|spectrin SH3 domain-binding protein 1', map_location='2 15.18 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=22939989, end_position_on_the_genomic_accession=23040241, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  404545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404545, CurrentID=0, Status='live', Symbol='Ano7', Aliases='Ngep, Ngep-L, Pcanap5, Tmem16g', description='anoctamin 7', other_designations='prostate cancer associated protein 5|transmembrane protein 16g', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93373796, end_position_on_the_genomic_accession=93404304, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  404549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404549, CurrentID=0, Status='live', Symbol='Ifna14', Aliases='OTTMUSG00000007662', description='interferon alpha 14', other_designations='', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88571229, end_position_on_the_genomic_accession=88571798, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  11350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11350, CurrentID=0, Status='live', Symbol='Abl1', Aliases='AI325092, Abl, E430008G22Rik, c-Abl', description='c-abl oncogene 1, non-receptor tyrosine kinase', other_designations='Abelson murine leukemia oncogene|Abelson tyrosine-protein kinase 1|abelson murine leukemia viral oncogene homolog 1|c-abl oncogene 1, receptor tyrosine kinase|c-abl protein, type II|c-abl protein, type III|p150|proto-oncogene c-Abl|proto-oncogene tyrosine-protein kinase ABL1|v-abl Abelson murine leukemia oncogene 1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31688252, end_position_on_the_genomic_accession=31807093, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  11352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11352, CurrentID=0, Status='live', Symbol='Abl2', Aliases='AA536808, Abll, Arg', description='v-abl Abelson murine leukemia viral oncogene 2 (arg, Abelson-related gene)', other_designations='Abelson-related gene protein|abelson murine leukemia viral oncogene homolog 2|abl2|tyrosine kinase ARG|tyrosine-protein kinase ABL2|tyrosine-protein kinase ARG|v-abl Abelson murine leukemia viral oncogene homolog 2 (arg, Abelson-related gene)', map_location='1 67.71 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156558171, end_position_on_the_genomic_accession=156649619, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11354, CurrentID=0, Status='live', Symbol='Scgb1b27', Aliases='Abp, Abpa, Abpa27, Sal-1, Tcp', description='secretoglobin, family 1B, member 27', other_designations='androgen binding protein alpha|androgen-binding protein|major allergen 1-like; lacrimal gland protein|salivary androgen-binding alpha subunit|salivary androgen-binding protein alpha subunit', map_location='7 19.38 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34021567, end_position_on_the_genomic_accession=34022881, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  11358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11358, CurrentID=0, Status='live', Symbol='ac', Aliases='', description='absent corpus callosum', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11363, CurrentID=0, Status='live', Symbol='Acadl', Aliases='AA960361, AU018452, C79855, LCAD', description='acyl-Coenzyme A dehydrogenase, long-chain', other_designations='acetyl-Coenzyme A dehydrogenase, long-chain', map_location='1 33.64 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=66830839, end_position_on_the_genomic_accession=66863309, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  11364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11364, CurrentID=0, Status='live', Symbol='Acadm', Aliases='AU018656, MCAD', description='acyl-Coenzyme A dehydrogenase, medium chain', other_designations='', map_location='3 78.77 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=153922353, end_position_on_the_genomic_accession=153944643, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  11370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11370, CurrentID=0, Status='live', Symbol='Acadvl', Aliases='vlcad', description='acyl-Coenzyme A dehydrogenase, very long chain', other_designations='MVLCAD', map_location='11 42.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70010183, end_position_on_the_genomic_accession=70015428, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  666731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666731, CurrentID=0, Status='live', Symbol='Trim43c', Aliases='EG666731, Trim43', description='tripartite motif-containing 43C', other_designations='trim43c|tripartite motif-containing 43', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88839023, end_position_on_the_genomic_accession=88848198, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  666747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666747, CurrentID=0, Status='live', Symbol='Trim43b', Aliases='EG666747, Gm8269', description='tripartite motif-containing 43B', other_designations='trim43b', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89084624, end_position_on_the_genomic_accession=89092835, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67437, CurrentID=0, Status='live', Symbol='Ssr3', Aliases='0610038P07Rik, AL022999, AU022074, AW553833, TRAPG', description='signal sequence receptor, gamma', other_designations='SSR-gamma|TRAP-gamma|signal sequence receptor subunit gamma|translocon-associated protein gamma', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=65379657, end_position_on_the_genomic_accession=65392553, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11409, CurrentID=0, Status='live', Symbol='Acads', Aliases='AI196007, Bcd-1, Bcd1, Hdlq8, SCAD', description='acyl-Coenzyme A dehydrogenase, short chain', other_designations='butyryl-CoA dehydrogenase', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115110299, end_position_on_the_genomic_accession=115119346, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11416, CurrentID=0, Status='live', Symbol='Slc33a1', Aliases='AI315656, AI788741, Acatn, D630022N01Rik', description='solute carrier family 33 (acetyl-CoA transporter), member 1', other_designations='AT-1|acetyl-CoA transporter 1|solute carrier family 33 member 1', map_location='3 E1-E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=63942323, end_position_on_the_genomic_accession=63964733, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  404634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404634, CurrentID=0, Status='live', Symbol='H2afy2', Aliases='macroH2A2', description='H2A histone family, member Y2', other_designations='histone macroH2A2|mH2A2', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61738647, end_position_on_the_genomic_accession=61784104, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11419, CurrentID=0, Status='live', Symbol='Asic1', Aliases='AI843610, ASIC, ASIC1a, Accn2, B530003N02Rik, BNaC2', description='acid-sensing (proton-gated) ion channel 1', other_designations='acid sensing ion channel|amiloride-sensitive cation channel 2, neuronal|brain sodium channel 2|degenerin 2', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99670266, end_position_on_the_genomic_accession=99701130, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  11421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11421, CurrentID=0, Status='live', Symbol='Ace', Aliases='AW208573, CD143', description='angiotensin I converting enzyme (peptidyl-dipeptidase A) 1', other_designations='angiotensin converting enzyme|dipeptidyl carboxypeptidase I|dipeptidyl peptidase|kininase II', map_location='11 68.84 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105967945, end_position_on_the_genomic_accession=105989964, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  11423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11423, CurrentID=0, Status='live', Symbol='Ache', Aliases='', description='acetylcholinesterase', other_designations='', map_location='5 76.32 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137285849, end_position_on_the_genomic_accession=137294466, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11425, CurrentID=0, Status='live', Symbol='Apoc4', Aliases='Acl, apo-CIV, apoC-IV', description='apolipoprotein C-IV', other_designations='apolipoprotein C2-linked|apolipoprotein C4', map_location='7 9.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19678084, end_position_on_the_genomic_accession=19681460, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11426, CurrentID=0, Status='live', Symbol='Macf1', Aliases='ABP620, Acf7, Aclp7, MACF, R74989, mKIAA0465', description='microtubule-actin crosslinking factor 1', other_designations='actin cross-linking family 7|actin-crosslinking protein 7|trabeculin alpha', map_location='4 57.42 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123349633, end_position_on_the_genomic_accession=123687315, orientation='minus', exon_count=112, OMIM='', no_hdr0=''),
  11428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11428, CurrentID=0, Status='live', Symbol='Aco1', Aliases='Aco-1, Irebp, Irp1', description='aconitase 1', other_designations='IRE-BP 1|citrate hydro-lyase|cytoplasmic aconitase|iron regulatory protein 1|iron-responsive element-binding protein 1', map_location='4 20.24 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40143265, end_position_on_the_genomic_accession=40199009, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  11429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11429, CurrentID=0, Status='live', Symbol='Aco2', Aliases='Aco-2, Aco3, D10Wsu183e', description='aconitase 2, mitochondrial', other_designations='citrate hydro-lyase', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81872463, end_position_on_the_genomic_accession=81915137, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  11430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11430, CurrentID=0, Status='live', Symbol='Acox1', Aliases='AOX, Acox, D130055E20Rik, Paox', description='acyl-Coenzyme A oxidase 1, palmitoyl', other_designations='palmitoyl-CoA oxidase|peroxisomal fatty acyl-coenzyme A oxidase|peroxisomal straight-chain Acyl-CoA oxidase', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116171883, end_position_on_the_genomic_accession=116199045, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11431, CurrentID=0, Status='live', Symbol='Acp1', Aliases='4632432E04Rik, AI427468, Acp-1, LMW-PTP', description='acid phosphatase 1, soluble', other_designations='LMW-PTPase|low molecular weight cytosolic acid phosphatase', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=30893326, end_position_on_the_genomic_accession=30911621, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11432, CurrentID=0, Status='live', Symbol='Acp2', Aliases='Acp-2, LAP', description='acid phosphatase 2, lysosomal', other_designations='', map_location='2 50.54 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91202812, end_position_on_the_genomic_accession=91214098, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11433, CurrentID=0, Status='live', Symbol='Acp5', Aliases='TRACP, TRAP', description='acid phosphatase 5, tartrate resistant', other_designations='TR-AP|acid phosphatase type V|tartrate-resistant acid ATPase|trATPase|type 5 acid phosphatase', map_location='9 8.38 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22126727, end_position_on_the_genomic_accession=22135746, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11434, CurrentID=0, Status='live', Symbol='Acr', Aliases='', description='acrosin prepropeptide', other_designations='preproacrosin', map_location='15 E-F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89567562, end_position_on_the_genomic_accession=89578904, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11435, CurrentID=0, Status='live', Symbol='Chrna1', Aliases='AI385656, AI608266, Achr-1, Acra', description='cholinergic receptor, nicotinic, alpha polypeptide 1 (muscle)', other_designations='', map_location='2 43.76 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73563281, end_position_on_the_genomic_accession=73580338, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11438, CurrentID=0, Status='live', Symbol='Chrna4', Aliases='Acra-4, Acra4, EBN1, ENFL1', description='cholinergic receptor, nicotinic, alpha polypeptide 4', other_designations='a4 nicotinic receptor|acetylcholine receptor alpha 4 neural|acetylcholine receptor, neuronal, nicotinic, alpha-4 subunit|alpha4 nAChR|alpha4-nAChR', map_location='2 103.54 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181022311, end_position_on_the_genomic_accession=181043580, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11440, CurrentID=0, Status='live', Symbol='Chrna6', Aliases='Acra6, Nica6', description='cholinergic receptor, nicotinic, alpha polypeptide 6', other_designations='acetylcholine receptor alpha 6 neural|alpha6 nAChR|nicotinic acetylcholine receptor subunit alpha 6', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27403212, end_position_on_the_genomic_accession=27413944, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11441, CurrentID=0, Status='live', Symbol='Chrna7', Aliases='Acra7, alpha7, nAchR', description='cholinergic receptor, nicotinic, alpha polypeptide 7', other_designations='acetylcholine receptor alpha 7 neural|alpha7 nicotinic receptor|alpha7-nAChR', map_location='7 34.47 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=63098692, end_position_on_the_genomic_accession=63212526, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11443, CurrentID=0, Status='live', Symbol='Chrnb1', Aliases='Achr-2, Acrb', description='cholinergic receptor, nicotinic, beta polypeptide 1 (muscle)', other_designations='AChR beta|acetylcholine receptor beta', map_location='11 42.87 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69784036, end_position_on_the_genomic_accession=69795974, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  11444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11444, CurrentID=0, Status='live', Symbol='Chrnb2', Aliases='Acrb-2, Acrb2, C030030P04Rik, [b]2-nAchR', description='cholinergic receptor, nicotinic, beta polypeptide 2 (neuronal)', other_designations='acetylcholine receptor beta 2 neural', map_location='3 39.19 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89753448, end_position_on_the_genomic_accession=89764632, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  666806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666806, CurrentID=0, Status='live', Symbol='Gm8300', Aliases='EG666806', description='predicted gene 8300', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87514316, end_position_on_the_genomic_accession=87518265, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  11447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11447, CurrentID=0, Status='live', Symbol='Chrnd', Aliases='Achr-4, Acrd, L10076', description='cholinergic receptor, nicotinic, delta polypeptide', other_designations='AChR delta subunit', map_location='1 44.07 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87190597, end_position_on_the_genomic_accession=87200070, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11448, CurrentID=0, Status='live', Symbol='Chrne', Aliases='AChrepsilon, Acre, nAChRE', description='cholinergic receptor, nicotinic, epsilon polypeptide', other_designations='AchR epsilon subunit|nicotinic acetylcholine receptor epsilon subunit', map_location='11 43.14 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70614883, end_position_on_the_genomic_accession=70619835, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  11449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11449, CurrentID=0, Status='live', Symbol='Chrng', Aliases='Achr-3, Acrg', description='cholinergic receptor, nicotinic, gamma polypeptide', other_designations='chrng variant N1|chrng variant N2|nicotinic acetylcholine receptor gamma subunit', map_location='1 44.07 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87205811, end_position_on_the_genomic_accession=87212410, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11450, CurrentID=0, Status='live', Symbol='Adipoq', Aliases='30kDa, APN, Acdc, Acrp30, Ad, GBP28, adipo, apM1', description='adiponectin, C1Q and collagen domain containing', other_designations='30 kDa adipocyte complement-related protein|adipocyte complement related protein|adipocyte complement-related 30 kDa protein|adipocyte, C1Q and collagen domain containing|adipocyte, C1q and collagen domain-containing protein|adipocyte-specific protein AdipoQ', map_location='16 13.96 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23146536, end_position_on_the_genomic_accession=23157968, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  11451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11451, CurrentID=0, Status='live', Symbol='Acrv1', Aliases='Msa63, SP-10', description='acrosomal vesicle protein 1', other_designations='MSA-63', map_location='9 20.67 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=36693220, end_position_on_the_genomic_accession=36698837, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11459, CurrentID=0, Status='live', Symbol='Acta1', Aliases='AA959943, Acta-2, Acts, Actsk-1', description='actin, alpha 1, skeletal muscle', other_designations='alpha-actin-1', map_location='8 72.26 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123891758, end_position_on_the_genomic_accession=123894775, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11461, CurrentID=0, Status='live', Symbol='Actb', Aliases='Actx, E430023M04Rik, beta-actin', description='actin, beta', other_designations='A-X actin-like protein|actin, beta, cytoplasmic|melanoma X-actin', map_location='5 81.8 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142903116, end_position_on_the_genomic_accession=142906724, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11464, CurrentID=0, Status='live', Symbol='Actc1', Aliases='Actc-1', description='actin, alpha, cardiac muscle 1', other_designations='actin alpha cardiac|alpha-cardiac actin|alphac-actin', map_location='2 57.55 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=114047289, end_position_on_the_genomic_accession=114052811, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11465, CurrentID=0, Status='live', Symbol='Actg1', Aliases='AL023024, Actg, Actl, E51', description='actin, gamma, cytoplasmic 1', other_designations='beta-actin FE-3|gamma-actin', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120345689, end_position_on_the_genomic_accession=120348491, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11468, CurrentID=0, Status='live', Symbol='Actg2', Aliases='ACTA3, Act-4, Act4, SMGA', description='actin, gamma 2, smooth muscle, enteric', other_designations='alpha-actin-3|gamma-2-actin|smooth muscle gamma-actin', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83512909, end_position_on_the_genomic_accession=83536251, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11470, CurrentID=0, Status='live', Symbol='Actl7a', Aliases='Tact2', description='actin-like 7a', other_designations='T-actin-2|actin-like Tact2|actin-like-7-alpha|t-actin 2|testis-specific actin-2', map_location='4 31.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56743422, end_position_on_the_genomic_accession=56744925, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  11471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11471, CurrentID=0, Status='live', Symbol='Actl7b', Aliases='Tact1', description='actin-like 7b', other_designations='T-actin-1|actin-like-7-beta|t-actin 1|testis-specific actin-1', map_location='4 27.0 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56740005, end_position_on_the_genomic_accession=56741425, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  11472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11472, CurrentID=0, Status='live', Symbol='Actn2', Aliases='1110008F24Rik', description='actinin alpha 2', other_designations='F-actin cross-linking protein|alpha-actinin skeletal muscle', map_location='13 4.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12269426, end_position_on_the_genomic_accession=12340732, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  11474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11474, CurrentID=0, Status='live', Symbol='Actn3', Aliases='', description='actinin alpha 3', other_designations='F-actin cross-linking protein|alpha-actinin skeletal muscle', map_location='19 4.12 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4861216, end_position_on_the_genomic_accession=4877909, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  11475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11475, CurrentID=0, Status='live', Symbol='Acta2', Aliases='0610041G09Rik, Actvs, SMalphaA, a-SMA, alphaSMA', description='actin, alpha 2, smooth muscle, aorta', other_designations='actin, alpha, vascular smooth muscle|alpha-actin-2', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34240336, end_position_on_the_genomic_accession=34255585, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11477, CurrentID=0, Status='live', Symbol='Acvr1', Aliases='ALK2, ActR-I, ActRIA, Acvr, Acvrlk2, Alk-2, Alk8, D330013D15Rik, SKR1, Tsk7L', description='activin A receptor, type 1', other_designations='TGF-B superfamily receptor type I|TSK-7L|TSR-I|activin receptor type I|serine/threonine-protein kinase receptor R1', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=58446438, end_position_on_the_genomic_accession=58566828, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11479, CurrentID=0, Status='live', Symbol='Acvr1b', Aliases='6820432J04, ActR-IB, ActRIB, Acvrlk4, Alk4, SKR2', description='activin A receptor, type 1B', other_designations='ALK-4|activin receptor type IB|activin receptor-like kinase 4|serine/threonine-protein kinase receptor R2', map_location='15 56.48 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101174125, end_position_on_the_genomic_accession=101212601, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11480, CurrentID=0, Status='live', Symbol='Acvr2a', Aliases='ActrIIa, Acvr2, TactrII', description='activin receptor IIA', other_designations='ACTR-IIA|activin receptor type IIA', map_location='2 28.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=48814109, end_position_on_the_genomic_accession=48903264, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11481, CurrentID=0, Status='live', Symbol='Acvr2b', Aliases='4930516B21Rik, ActRIIB', description='activin receptor IIB', other_designations='activin receptor type IIB', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119401583, end_position_on_the_genomic_accession=119442134, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11482, CurrentID=0, Status='live', Symbol='Acvrl1', Aliases='AI115505, AI427544, Acvrlk1, Alk1', description='activin A receptor, type II-like 1', other_designations='SKR3|TGF-B superfamily receptor type I|TSR-I|activin receptor-like kinase 1|activin receptor-like kinase-1', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101128522, end_position_on_the_genomic_accession=101145336, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11484, CurrentID=0, Status='live', Symbol='Aspa', Aliases='Acy-2, Acy2, nur7', description='aspartoacylase', other_designations='aminoacylase-2|aspartoacylase (aminoacylase) 2', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73304988, end_position_on_the_genomic_accession=73326868, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11486, CurrentID=0, Status='live', Symbol='Ada', Aliases='', description='adenosine deaminase', other_designations='adenosine aminohydrolase', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163726571, end_position_on_the_genomic_accession=163750239, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  11487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11487, CurrentID=0, Status='live', Symbol='Adam10', Aliases='1700031C13Rik, MADM, kuz, kuzbanian', description='a disintegrin and metallopeptidase domain 10', other_designations='ADAM 10|a disintegrin and metalloprotease domain (ADAM) 10|a disintegrin and metalloprotease domain 10|kuzbanian protein homolog|mammalian disintegrin-metalloprotease', map_location='9 39.53 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70679001, end_position_on_the_genomic_accession=70780229, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  11488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11488, CurrentID=0, Status='live', Symbol='Adam11', Aliases='AW060611, Mdc', description='a disintegrin and metallopeptidase domain 11', other_designations='a disintegrin and metalloprotease domain (ADAM) 11|a disintegrin and metalloprotease domain 11|metalloproteinase-like, disintegrin-like, and cysteine-rich protein', map_location='11 66.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102761439, end_position_on_the_genomic_accession=102780262, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  11489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11489, CurrentID=0, Status='live', Symbol='Adam12', Aliases='Mltna, mKIAA4001', description='a disintegrin and metallopeptidase domain 12 (meltrin alpha)', other_designations='ADAM 12|a disintegrin and metalloprotease domain 12|a disintegrin and metalloproteinase domain 12 (meltrin alpha)|meltrin alpha|meltrin-alpha', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133883196, end_position_on_the_genomic_accession=134232149, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  11490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11490, CurrentID=0, Status='live', Symbol='Adam15', Aliases='MDC15, metargidin', description='a disintegrin and metallopeptidase domain 15 (metargidin)', other_designations='AD56|ADAM 15|MDC-15|a disintegrin and metalloprotease domain 15 (metargidin)|a disintegrin and metalloproteinase domain 15 (metargidin)|disintegrin and metalloprotease 15|metalloprotease RGD disintegrin protein|metalloproteinase-like, disintegrin-like, and cysteine-rich protein 15', map_location='3 39.07 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89339640, end_position_on_the_genomic_accession=89350010, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  11491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11491, CurrentID=0, Status='live', Symbol='Adam17', Aliases='CD156b, Tace', description='a disintegrin and metallopeptidase domain 17', other_designations='TNF-alpha convertase|TNF-alpha converting enzyme|a disintegrin and metalloproteinase domain 17', map_location='12 8.3 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=21323509, end_position_on_the_genomic_accession=21373632, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  11492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11492, CurrentID=0, Status='live', Symbol='Adam19', Aliases='AL024287, MADDAM, MTLNB, Mltnb', description='a disintegrin and metallopeptidase domain 19 (meltrin beta)', other_designations='MDC beta|Mel beta|a disintegrin and metalloproteinase domain 19 (meltrin beta)|adamalysin 19', map_location='11 A5-B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46054502, end_position_on_the_genomic_accession=46147347, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  404710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404710, CurrentID=0, Status='live', Symbol='Iqgap3', Aliases='AI593484, D030034H08', description='IQ motif containing GTPase activating protein 3', other_designations='', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88082051, end_position_on_the_genomic_accession=88121048, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  11495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11495, CurrentID=0, Status='live', Symbol='Adam2', Aliases='AI323749, Ftnb, Ph30-beta', description='a disintegrin and metallopeptidase domain 2', other_designations='ADAM 2|PH-30|PH30|a disintegrin and metalloprotease domain 2 (fertilin beta)|fertilin beta|fertilin subunit beta', map_location='14 34.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66027329, end_position_on_the_genomic_accession=66077856, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  11496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11496, CurrentID=0, Status='live', Symbol='Adam22', Aliases='2900022I03Rik, AI854032, MDC2', description='a disintegrin and metallopeptidase domain 22', other_designations='a disintegrin and metalloprotease domain (ADAM) 22', map_location='5 3.39 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8072348, end_position_on_the_genomic_accession=8368268, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  11497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11497, CurrentID=0, Status='live', Symbol='Adam3', Aliases='Cyrn1, Taz83, tMDC', description='a disintegrin and metallopeptidase domain 3 (cyritestin)', other_designations='', map_location='8 13.09 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24677232, end_position_on_the_genomic_accession=24726044, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  11498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11498, CurrentID=0, Status='live', Symbol='Adam4', Aliases='AV274654, tMDCV', description='a disintegrin and metallopeptidase domain 4', other_designations='a disintegrin and metalloprotease domain (ADAM) 4|a disintegrin and metalloprotease domain 4', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81419549, end_position_on_the_genomic_accession=81421878, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  11499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11499, CurrentID=0, Status='live', Symbol='Adam5', Aliases='tMDC II, tMDCII', description='a disintegrin and metallopeptidase domain 5', other_designations='a disintegrin and metalloprotease domain (ADAM) 5|a disintegrin and metalloprotease domain 5|transmembrane metalloproteinase-like, disintegrin-like, and cysteine-rich protein II', map_location='8 13.26 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24727093, end_position_on_the_genomic_accession=24824369, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  11500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11500, CurrentID=0, Status='live', Symbol='Adam7', Aliases='ADAM 7, EAP1, EAPI', description='a disintegrin and metallopeptidase domain 7', other_designations='a disintegrin and metalloprotease domain (ADAM) 7|epididymal apical protein I', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=68497336, end_position_on_the_genomic_accession=68533689, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  11501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11501, CurrentID=0, Status='live', Symbol='Adam8', Aliases='ADAM 8, CD156, CD156a, E430039A18Rik, MS2', description='a disintegrin and metallopeptidase domain 8', other_designations='a disintegrin and metalloprotease domain 8|cell surface antigen MS2|macrophage cysteine-rich glycoprotein', map_location='7 F3-F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139978932, end_position_on_the_genomic_accession=139993121, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  11502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11502, CurrentID=0, Status='live', Symbol='Adam9', Aliases='AU020942, MDC9, Mltng, mKIAA0021', description='a disintegrin and metallopeptidase domain 9 (meltrin gamma)', other_designations='a disintegrin and metalloproteinase domain 9 (meltrin gamma)|metalloprotease/disintegrin/cysteine-rich protein 9|myeloma cell metalloproteinase', map_location='8 13.7 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24949611, end_position_on_the_genomic_accession=25016933, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  11504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11504, CurrentID=0, Status='live', Symbol='Adamts1', Aliases='ADAM-TS1, ADAMTS, ADAMTS-1, C3-C5, METH-1, METH1', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 1', other_designations='ADAM-TS 1', map_location='16 C3-C5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=85793827, end_position_on_the_genomic_accession=85803115, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11512, CurrentID=0, Status='live', Symbol='Adcy6', Aliases='AC6, mKIAA0422', description='adenylate cyclase 6', other_designations='ATP pyrophosphate-lyase 6|adenylate cyclase type VI|adenylyl cyclase 6|ca(2+)-inhibitable adenylyl cyclase', map_location='15 F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98589978, end_position_on_the_genomic_accession=98610140, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  11513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11513, CurrentID=0, Status='live', Symbol='Adcy7', Aliases='AA407758', description='adenylate cyclase 7', other_designations='ATP pyrophosphate-lyase 7|adenylate cyclase type VII|adenylyl cyclase 7|adenylyl cyclase type VII', map_location='8 43.06 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88272367, end_position_on_the_genomic_accession=88329965, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  11514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11514, CurrentID=0, Status='live', Symbol='Adcy8', Aliases='AC8, AW060868', description='adenylate cyclase 8', other_designations='ATP pyrophosphate-lyase 8|adenylate cyclase type VIII|adenylyl cyclase 8|ca(2+)/calmodulin-activated adenylyl cyclase', map_location='15 29.03 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=64699035, end_position_on_the_genomic_accession=64922296, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  11515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11515, CurrentID=0, Status='live', Symbol='Adcy9', Aliases='AC9, ACtp10, AW125421, D16Wsu65e, mKIAA0520', description='adenylate cyclase 9', other_designations='ATP pyrophosphate-lyase 9|adenylate cyclase type IX|adenylyl cyclase 9|adenylyl cyclase type 10', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4284886, end_position_on_the_genomic_accession=4437236, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  11516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11516, CurrentID=0, Status='live', Symbol='Adcyap1', Aliases='PACAP', description='adenylate cyclase activating polypeptide 1', other_designations='', map_location='17 E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=93198904, end_position_on_the_genomic_accession=93206541, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11517, CurrentID=0, Status='live', Symbol='Adcyap1r1', Aliases='2900024I10Rik, AI846590, PAC1, PAC1R, PACAP1-R', description='adenylate cyclase activating polypeptide 1 receptor 1', other_designations='PACAP type I receptor', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55451501, end_position_on_the_genomic_accession=55501455, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  11518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11518, CurrentID=0, Status='live', Symbol='Add1', Aliases='AI256389', description='adducin 1 (alpha)', other_designations='erythrocyte adducin subunit alpha', map_location='5 17.9 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34573669, end_position_on_the_genomic_accession=34632305, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  11519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11519, CurrentID=0, Status='live', Symbol='Add2', Aliases='2900072M03Rik, add97', description='adducin 2 (beta)', other_designations='erythrocyte adducin subunit beta|snoRNA MBII-396', map_location='6 37.55 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86028681, end_position_on_the_genomic_accession=86124409, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  11520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11520, CurrentID=0, Status='live', Symbol='Plin2', Aliases='AA407157, ADPH, Adfp, Adrp', description='perilipin 2', other_designations='adipophilin|adipose differentiation related protein|adipose differentiation-related protein', map_location='4 40.69 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86626910, end_position_on_the_genomic_accession=86670059, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11522, CurrentID=0, Status='live', Symbol='Adh1', Aliases='ADH-AA, AI194826, Adh-1, Adh-1-t, Adh-1e, Adh-1t, Adh-3e-e, Adh1-t, Adh1tl, Adh3-e, Adh1', description='alcohol dehydrogenase 1 (class I)', other_designations='ADH-A2|alcohol dehydrogenase 1 temporal, liver|alcohol dehydrogenase 1, complex|alcohol dehydrogenase 1, electrophoretic|alcohol dehydrogenase 3, electrophoretic|alcohol dehydrogenase A subunit|class I alcohol dehydrogenase', map_location='3 64.16 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138277645, end_position_on_the_genomic_accession=138290691, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  404739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404739, CurrentID=0, Status='live', Symbol='Iglj2', Aliases='Igl-J2', description='immunoglobulin lambda joining 2', other_designations='immunoglobulin lambda chain, joining region 2', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19200198, end_position_on_the_genomic_accession=19200235, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  11529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11529, CurrentID=0, Status='live', Symbol='Adh7', Aliases='AI325182, Adh-3, Adh-3e, Adh-3t, Adh3, Adh3-e, Adh3-t, Adh4, Adt-1', description='alcohol dehydrogenase 7 (class IV), mu or sigma polypeptide', other_designations='ADH-C2|IV ADH|alcohol dehydrogenase 3 complex|alcohol dehydrogenase 3 temporal|alcohol dehydrogenase 4|alcohol dehydrogenase class IV mu/sigma chain|gastric alcohol dehydrogenase|retinol dehydrogenase', map_location='3 64.13 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138198053, end_position_on_the_genomic_accession=138232042, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11532, CurrentID=0, Status='live', Symbol='Adh5', Aliases='Adh-5, Adh3, GSNOR', description='alcohol dehydrogenase 5 (class III), chi polypeptide', other_designations='ADH-B2|FALDH|FDH|GSH-FDH|S-(hydroxymethyl)glutathione dehydrogenase|S-nitrosoglutathione reductase|alcohol dehydrogenase 2|alcohol dehydrogenase B2|alcohol dehydrogenase class-III|glutathione-dependent formaldehyde dehydrogenase', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138437200, end_position_on_the_genomic_accession=138455499, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11534, CurrentID=0, Status='live', Symbol='Adk', Aliases='2310026J05Rik, 5033405D03Rik, AI255373, AI987814, Ak', description='adenosine kinase', other_designations="adenosine 5'-phosphotransferase", map_location='14 A2-B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21052573, end_position_on_the_genomic_accession=21448569, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  11535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11535, CurrentID=0, Status='live', Symbol='Adm', Aliases='AM', description='adrenomedullin', other_designations='', map_location='7 57.7 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110627669, end_position_on_the_genomic_accession=110629819, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11536, CurrentID=0, Status='live', Symbol='Gpr182', Aliases='AM-R, Admr, G10-D, Gpcr17, Gpcr22, MB10, NOW', description='G protein-coupled receptor 182', other_designations='G10D|adrenomedullin receptor', map_location='10 75.07 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127749602, end_position_on_the_genomic_accession=127751798, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  11537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11537, CurrentID=0, Status='live', Symbol='Cfd', Aliases='Adn, DF', description='complement factor D (adipsin)', other_designations='28 kDa adipocyte protein|C3 convertase activator|D component (adipsin) of complement|adipsin|properdin factor D', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79890853, end_position_on_the_genomic_accession=79892660, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11538, CurrentID=0, Status='live', Symbol='Adnp', Aliases='AA589558, mKIAA0784', description='activity-dependent neuroprotective protein', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168180965, end_position_on_the_genomic_accession=168207062, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11539, CurrentID=0, Status='live', Symbol='Adora1', Aliases='A1-AR, A1AR, A1R, AA1R, AI848715, BB176431, Ri', description='adenosine A1 receptor', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134199215, end_position_on_the_genomic_accession=134235457, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11540, CurrentID=0, Status='live', Symbol='Adora2a', Aliases='A2AAR, A2aR, AA2AR', description='adenosine A2a receptor', other_designations='', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75316860, end_position_on_the_genomic_accession=75334792, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11541, CurrentID=0, Status='live', Symbol='Adora2b', Aliases='A2BAR, A2BR, A2b, AA2BR, AI480866, AI605384', description='adenosine A2b receptor', other_designations='A2b, Rs', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62248984, end_position_on_the_genomic_accession=62266452, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  11542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11542, CurrentID=0, Status='live', Symbol='Adora3', Aliases='A3AR, A3R, AA3R, ARA3, Gpcr2', description='adenosine A3 receptor', other_designations='G-protein coupled receptor 2', map_location='3 46.45 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105904421, end_position_on_the_genomic_accession=105908928, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102640920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102640920, CurrentID=0, Status='live', Symbol='Gm36869', Aliases='', description='predicted gene, 36869', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138341770, end_position_on_the_genomic_accession=138359898, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11545, CurrentID=0, Status='live', Symbol='Parp1', Aliases='5830444G22Rik, AI893648, ARTD1, Adprp, Adprt1, C80510, PARP, PPOL, parp-1, sPARP-1', description='poly (ADP-ribose) polymerase family, member 1', other_designations='ADP-ribosyltransferase (NAD+, poly (ADP-ribose) polymerase) 1|ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase) 1|ADP-ribosyltransferase diphtheria toxin-like 1|ADPRT 1|NAD(+) ADP-ribosyltransferase 1|msPARP|poly(ADP-ribose) polymerase|poly[ADP-ribose] synthase 1|poly[ADP-ribose] synthetase 1', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180568975, end_position_on_the_genomic_accession=180601254, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  11546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11546, CurrentID=0, Status='live', Symbol='Parp2', Aliases='ARTD2, Adprt2, Adprtl2, Aspartl2, C78626, PARP-2', description='poly (ADP-ribose) polymerase family, member 2', other_designations='ADP-ribosyltransferase (NAD+, poly(ADP-ribose) polymerase)-like 2|ADP-ribosyltransferase (NAD+; poly (ADP-ribose) polymerase) 2|ADP-ribosyltransferase (NAD+; poly(ADP-ribose) polymerase)-like 2|ADP-ribosyltransferase diphtheria toxin-like 2|ADPRT-2|NAD(+) ADP-ribosyltransferase 2|mPARP-2|pADPRT-2|poly[ADP-ribose] synthase 2|poly[ADP-ribose] synthetase 2', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50807898, end_position_on_the_genomic_accession=50821301, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  11548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11548, CurrentID=0, Status='live', Symbol='Adra1b', Aliases='[a]1b', description='adrenergic receptor, alpha 1b', other_designations='alpha-1B adrenoceptor|alpha-1B adrenoreceptor|alpha1B-adrenergic receptor', map_location='11 25.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43774605, end_position_on_the_genomic_accession=43901251, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11549, CurrentID=0, Status='live', Symbol='Adra1a', Aliases='Adra1c', description='adrenergic receptor, alpha 1a', other_designations='alpha-1A adrenoceptor|alpha-1A adrenoreceptor|alpha-1C adrenergic receptor', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66635017, end_position_on_the_genomic_accession=66771240, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11550, CurrentID=0, Status='live', Symbol='Adra1d', Aliases='Adra-1, Adra1, Adra1a, Gpcr8, Spr8, [a]1d, alpha1D-AR', description='adrenergic receptor, alpha 1d', other_designations='alpha-1A adrenergic receptor|alpha-1D adrenoceptor|alpha-1D adrenoreceptor', map_location='2 63.5 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131545357, end_position_on_the_genomic_accession=131562285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  11551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11551, CurrentID=0, Status='live', Symbol='Adra2a', Aliases='AW122659, Adra-2, Adra-2a, alpha(2A)AR, alpha2-C10, alpha2A, alpha2A-AR', description='adrenergic receptor, alpha 2a', other_designations='alpha-2A adrenoceptor|alpha-2A adrenoreceptor|alpha-2AAR|alpha2A-adrenergic receptor|subtype alpha2-C10', map_location='19 49.04 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=54045182, end_position_on_the_genomic_accession=54048982, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  11552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11552, CurrentID=0, Status='live', Symbol='Adra2b', Aliases='Adra-2b, [a]2B, alpha2-C2, alpha2B', description='adrenergic receptor, alpha 2b', other_designations='alpha-2B adrenoceptor|alpha-2B adrenoreceptor|alpha-2BAR|subtype alpha2-C2', map_location='2 61.95 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127363286, end_position_on_the_genomic_accession=127367221, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  11553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11553, CurrentID=0, Status='live', Symbol='Adra2c', Aliases='Adra-2c, [a]2C, alpha2-C4, alpha2C', description='adrenergic receptor, alpha 2c', other_designations='alpha-2 adrenergic receptor subtype C4|alpha-2C adrenoceptor|alpha-2C adrenoreceptor|alpha-2CAR|subtype alpha2-C4', map_location='5 18.09 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35278566, end_position_on_the_genomic_accession=35281763, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  11554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11554, CurrentID=0, Status='live', Symbol='Adrb1', Aliases='Adrb-1, beta-AR', description='adrenergic receptor, beta 1', other_designations='beta 1-AR|beta-1 adrenoceptor|beta-1 adrenoreceptor|cardiac beta adrenergic receptor', map_location='19 51.96 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56722372, end_position_on_the_genomic_accession=56724862, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  11555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11555, CurrentID=0, Status='live', Symbol='Adrb2', Aliases='Adrb-2, Badm, Gpcr7', description='adrenergic receptor, beta 2', other_designations='beta 2-AR|beta-2 adrenoceptor|beta-2 adrenoreceptor', map_location='18 35.1 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=62177713, end_position_on_the_genomic_accession=62179981, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  11556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11556, CurrentID=0, Status='live', Symbol='Adrb3', Aliases='Adrb-3', description='adrenergic receptor, beta 3', other_designations='Beta-3 AR|Beta-3 adrenoceptor|beta 3-AR|beta-3 adrenoreceptor|beta3-adrenergic receptor', map_location='8 A2-A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27225776, end_position_on_the_genomic_accession=27230805, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11564, CurrentID=0, Status='live', Symbol='Adsl', Aliases='Adl, Asl', description='adenylosuccinate lyase', other_designations='ASase|adenylosuccinase|adenylosuccinate lyase 1', map_location='15 37.95 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80948490, end_position_on_the_genomic_accession=80970947, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11565, CurrentID=0, Status='live', Symbol='Adssl1', Aliases='AI528595, Adss, Adss1', description='adenylosuccinate synthetase like 1', other_designations='AMPSase 1|IMP--aspartate ligase 1|M-type adenylosuccinate synthetase|adSS 1|adenylosuccinate synthetase 1, muscle|adenylosuccinate synthetase, basic isozyme|adenylosuccinate synthetase, muscle isozyme', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112620047, end_position_on_the_genomic_accession=112641355, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11566, CurrentID=0, Status='live', Symbol='Adss', Aliases='AI314886, AS2, Adss', description='adenylosuccinate synthetase, non muscle', other_designations='AMPSase 2|IMP--aspartate ligase 2|L-type adenylosuccinate synthetase|adSS 2|adenylosuccinate synthetase 2, non muscle|adenylosuccinate synthetase, acidic isozyme|adenylosuccinate synthetase, liver isozyme|adenylosuccinate synthetase, non-muscle isozyme', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=177763178, end_position_on_the_genomic_accession=177796694, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  11567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11567, CurrentID=0, Status='live', Symbol='Avil', Aliases='Advil, DOC6', description='advillin', other_designations='actin-binding protein DOC6|p92', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127000075, end_position_on_the_genomic_accession=127020994, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  11568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11568, CurrentID=0, Status='live', Symbol='Aebp1', Aliases='ACLP', description='AE binding protein 1', other_designations='aortic carboxypeptidase-like protein', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5861853, end_position_on_the_genomic_accession=5872225, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  11569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11569, CurrentID=0, Status='live', Symbol='Aebp2', Aliases='AU023766, B230313N05Rik', description='AE binding protein 2', other_designations='AE-binding protein 2|adipocyte enhancer-binding protein 2', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=140622454, end_position_on_the_genomic_accession=140677634, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11571, CurrentID=0, Status='live', Symbol='Crisp1', Aliases='AW324342, Aeg1, CRISP-1', description='cysteine-rich secretory protein 1', other_designations='SCP 1|acidic epididymal glycoprotein 1|sperm-coating glycoprotein 1', map_location='17 19.47 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40293758, end_position_on_the_genomic_accession=40319246, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11572, CurrentID=0, Status='live', Symbol='Crisp3', Aliases='Aeg2, CRISP-3, CRS3, SGP28', description='cysteine-rich secretory protein 3', other_designations='SCP 2|acidic epididymal glycoprotein 2|sperm-coating glycoprotein 2', map_location='17 19.45 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40221777, end_position_on_the_genomic_accession=40242258, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11576, CurrentID=0, Status='live', Symbol='Afp', Aliases='', description='alpha fetoprotein', other_designations='alpha-1-fetoprotein|alpha-fetoglobulin|alpha-foetoprotein', map_location='5 44.77 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90490714, end_position_on_the_genomic_accession=90508907, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  666938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666938, CurrentID=0, Status='live', Symbol='Bend4', Aliases='D330027G24Rik, EG666938', description='BEN domain containing 4', other_designations='bend4', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=67392147, end_position_on_the_genomic_accession=67428250, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11592, CurrentID=0, Status='live', Symbol='ag', Aliases='', description='agitans', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11593, CurrentID=0, Status='live', Symbol='Aga', Aliases='AW060726', description='aspartylglucosaminidase', other_designations='N4-(N-acetyl-beta-glucosaminyl)-L-asparagine amidase|glycosylasparaginase', map_location='8 B3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=53511687, end_position_on_the_genomic_accession=53523422, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11595, CurrentID=0, Status='live', Symbol='Acan', Aliases='Agc, Agc1, Cspg1, b2b183Clo, cmd', description='aggrecan', other_designations='CSPCP|aggrecan 1|cartilage-specific proteoglycan core protein', map_location='7 44.88 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79053202, end_position_on_the_genomic_accession=79115099, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  11596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11596, CurrentID=0, Status='live', Symbol='Ager', Aliases='RAGE', description='advanced glycosylation end product-specific receptor', other_designations='advanced glycation end-products receptor|receptor for advanced glycation end-products variant 20|receptor for advanced glycosylation end products', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34597410, end_position_on_the_genomic_accession=34600937, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11600, CurrentID=0, Status='live', Symbol='Angpt1', Aliases='1110046O21Rik, Ang-1, Ang1', description='angiopoietin 1', other_designations='', map_location='15 16.69 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=42424667, end_position_on_the_genomic_accession=42676977, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11601, CurrentID=0, Status='live', Symbol='Angpt2', Aliases='Agpt2, Ang-2, Ang2', description='angiopoietin 2', other_designations='', map_location='8 10.3 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=18690263, end_position_on_the_genomic_accession=18741562, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11602, CurrentID=0, Status='live', Symbol='Angpt4', Aliases='Agpt4, Ang3', description='angiopoietin 4', other_designations='ANG-3|ANG-4|angiopoietin 3|angiopoietin-3', map_location='2 74.83 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151911332, end_position_on_the_genomic_accession=151944654, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11603, CurrentID=0, Status='live', Symbol='Agrn', Aliases='Agrin, nmf380', description='agrin', other_designations='', map_location='4 88.55 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156165290, end_position_on_the_genomic_accession=156197519, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  11604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11604, CurrentID=0, Status='live', Symbol='Agrp', Aliases='Agrt, Art', description='agouti related neuropeptide', other_designations='agouti related protein|orexigenic neuropeptide', map_location='8 D1-D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105566695, end_position_on_the_genomic_accession=105579845, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11605, CurrentID=0, Status='live', Symbol='Gla', Aliases='Ags', description='galactosidase, alpha', other_designations='alpha-D-galactosidase A|alpha-D-galactoside galactohydrolase|melibiase', map_location='X E-F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134588169, end_position_on_the_genomic_accession=134601005, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11606, CurrentID=0, Status='live', Symbol='Agt', Aliases='AI265500, AngI, AngII, Aogen, Serpina8', description='angiotensinogen (serpin peptidase inhibitor, clade A, member 8)', other_designations='angiotensin I|angiotensin ll|serpin A8', map_location='8 72.81 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124556587, end_position_on_the_genomic_accession=124569707, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11607, CurrentID=0, Status='live', Symbol='Agtr1a', Aliases='1810074K20Rik, AG2S, AI551199, AT1, AT1a, AT2R1, AT2R1A, Agtr-1a, Agtr1, Angtr-1a', description='angiotensin II receptor, type 1a', other_designations='angiotensin II type-1A receptor|angiotensin receptor 1|angiotensin receptor 1a', map_location='13 13.19 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=30336356, end_position_on_the_genomic_accession=30382867, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11608, CurrentID=0, Status='live', Symbol='Agtr1b', Aliases='AT1B, AT2R1B, Agtr-1b, Angtr-1b', description='angiotensin II receptor, type 1b', other_designations='AT3|angiotensin II type-1B receptor', map_location='3 6.37 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=20314473, end_position_on_the_genomic_accession=20367177, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  11609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11609, CurrentID=0, Status='live', Symbol='Agtr2', Aliases='AI316812, AW107640', description='angiotensin II receptor, type 2', other_designations='AT2 receptor|angiotensin II type-2 receptor', map_location='X 16.71 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=21484549, end_position_on_the_genomic_accession=21488833, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  142682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=142682, CurrentID=0, Status='live', Symbol='Zcchc14', Aliases='AA792890, BDG29', description='zinc finger, CCHC domain containing 14', other_designations='BDG-29 protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121598697, end_position_on_the_genomic_accession=121652761, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  11611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11611, CurrentID=0, Status='live', Symbol='Agxt', Aliases='AGT, Agt11, SPT, Agxt', description='alanine-glyoxylate aminotransferase', other_designations='alanine--glyoxylate aminotransferase|alanine-glyoxylate aminotransferase 1|serine-pyruvate aminotransferase', map_location='1 47.0 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93135240, end_position_on_the_genomic_accession=93145421, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11614, CurrentID=0, Status='live', Symbol='Nr0b1', Aliases='AHX, Ahc, Ahch, DAX-1, Dax1', description='nuclear receptor subfamily 0, group B, member 1', other_designations='adrenal hypoplasia congenita-like protein|adrenal hypoplasia, congenital homolog|nuclear receptor DAX-1', map_location='X 39.67 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=86191775, end_position_on_the_genomic_accession=86195944, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  142687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=142687, CurrentID=0, Status='live', Symbol='Asb14', Aliases='Asb-14', description='ankyrin repeat and SOCS box-containing 14', other_designations='ankyrin repeat and SOCS box-containing protein 14|ankyrin repeat domain-containing SOCS box protein 14', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26894591, end_position_on_the_genomic_accession=26916100, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  142688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=142688, CurrentID=0, Status='live', Symbol='Asb13', Aliases='2210015B19Rik, 6430573K02Rik, Asb-13, C85285', description='ankyrin repeat and SOCS box-containing 13', other_designations='ankyrin repeat and SOCS box-containing protein 13|ankyrin repeat domain-containing SOCS box protein 13', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3634012, end_position_on_the_genomic_accession=3651779, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11622, CurrentID=0, Status='live', Symbol='Ahr', Aliases='Ah, Ahhe, In, bHLHe76, Ahr', description='aryl-hydrocarbon receptor', other_designations='ah receptor|dioxin receptor', map_location='12 15.78 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=35497979, end_position_on_the_genomic_accession=35534989, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  11624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11624, CurrentID=0, Status='live', Symbol='Ahrr', Aliases='mKIAA1234', description='aryl-hydrocarbon receptor repressor', other_designations='ahR repressor', map_location='13 C2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74211117, end_position_on_the_genomic_accession=74292330, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11625, CurrentID=0, Status='live', Symbol='Ahsg', Aliases='', description='alpha-2-HS-glycoprotein', other_designations='countertrypin|fetuin-A', map_location='16 13.74 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22892015, end_position_on_the_genomic_accession=22899451, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11628, CurrentID=0, Status='live', Symbol='Aicda', Aliases='Aid, Arp2', description='activation-induced cytidine deaminase', other_designations='activation induced deaminase|cytidine aminohydrolase', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122553809, end_position_on_the_genomic_accession=122564180, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11629, CurrentID=0, Status='live', Symbol='Aif1', Aliases='AI607846, AIF-1, D17H6S50E, G1, Iba1', description='allograft inflammatory factor 1', other_designations='ionized calcium binding adapter molecule 1|ionized calcium-binding adapter molecule 1|testis specific', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35170991, end_position_on_the_genomic_accession=35176087, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11630, CurrentID=0, Status='live', Symbol='Aim1', Aliases='AI462491, AI463325', description='absent in melanoma 1', other_designations='', map_location='10 23.14 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43950307, end_position_on_the_genomic_accession=44148924, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  11632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11632, CurrentID=0, Status='live', Symbol='Aip', Aliases='AA408703, AW476050, Ara9, D19Bwg1412e, Fkbp16, Xap2', description='aryl-hydrocarbon receptor-interacting protein', other_designations='', map_location='19 3.82 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4113756, end_position_on_the_genomic_accession=4127236, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11634, CurrentID=0, Status='live', Symbol='Aire', Aliases='', description='autoimmune regulator (autoimmune polyendocrinopathy candidiasis ectodermal dystrophy)', other_designations='APECED protein homolog|autoimmune polyendocrinopathy candidiasis ectodermal dystrophy protein homolog|autoimmune regulator AIRE1a', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78030022, end_position_on_the_genomic_accession=78043610, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11636, CurrentID=0, Status='live', Symbol='Ak1', Aliases='Ak-1, B430205N08Rik', description='adenylate kinase 1', other_designations='cytosolic adenylate kinase', map_location='2 22.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32621758, end_position_on_the_genomic_accession=32635058, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11637, CurrentID=0, Status='live', Symbol='Ak2', Aliases='Ak-2, D4Ertd220e', description='adenylate kinase 2', other_designations='AK 2|ATP-AMP transphosphorylase 2|ATP:AMP phosphotransferase|adenylate monophosphate kinase', map_location='4 62.62 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128993217, end_position_on_the_genomic_accession=129011529, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11639, CurrentID=0, Status='live', Symbol='Ak4', Aliases='AK 4, Ak-3, Ak-4, Ak3, Ak3l1, D4Ertd274e', description='adenylate kinase 4', other_designations='AK 4|ATP-AMP transphosphorylase|GTP:AMP phosphotransferase AK4, mitochondrial|adenylate kinase 3 alpha-like 1|adenylate kinase 3-like 1|adenylate kinase isoenzyme 4, mitochondrial', map_location='4 46.84 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101419289, end_position_on_the_genomic_accession=101467771, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11640, CurrentID=0, Status='live', Symbol='Akap1', Aliases='AKAP121, AKAP84, Akap, C76494, C81186, DAKAP1, S-AKAP84', description='A kinase (PRKA) anchor protein 1', other_designations='D-AKAP-1|PRKA1|S-AKAP|a kinase anchor protein 1, mitochondrial|dual specificity A-kinase-anchoring protein 1|protein kinase A-anchoring protein 1|spermatid A-kinase anchor protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88830792, end_position_on_the_genomic_accession=88864586, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11641, CurrentID=0, Status='live', Symbol='Akap2', Aliases='AA959716, AI649048, Akap-2, Akap-kl, B230340M18Rik, Prka2', description='A kinase (PRKA) anchor protein 2', other_designations='AKAP expressed in kidney and lung|protein kinase A-anchoring protein 2', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=57845248, end_position_on_the_genomic_accession=57896984, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11642, CurrentID=0, Status='live', Symbol='Akap3', Aliases='Prka3', description='A kinase (PRKA) anchor protein 3', other_designations='A-kinase anchor protein 110 kDa|AKAP 110|AKAP-3|Protein kinase A-anchoring protein 3', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126853098, end_position_on_the_genomic_accession=126874308, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11643, CurrentID=0, Status='live', Symbol='Akap4', Aliases='AKAP-4, AKAP82, Fsc1, PRKA4, mAKAP82, p82', description='A kinase (PRKA) anchor protein 4', other_designations='A-kinase anchor protein 82 kDa|major sperm fibrous sheath protein|protein kinase A-anchoring protein 4', map_location='X 3.17 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7067515, end_position_on_the_genomic_accession=7078609, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11647, CurrentID=0, Status='live', Symbol='Alpl', Aliases='APTNAP, Akp-2, Akp2, TNAP, TNSALP', description='alkaline phosphatase, liver/bone/kidney', other_designations='alkaline phosphatase 2, liver', map_location='4 70.02 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137741731, end_position_on_the_genomic_accession=137796384, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  11648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11648, CurrentID=0, Status='live', Symbol='Akp3', Aliases='Akp-3, IAP, dIAP', description='alkaline phosphatase 3, intestine, not Mn requiring', other_designations='intestinal alkaline phosphatase', map_location='1 44.06 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87125008, end_position_on_the_genomic_accession=87127912, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11650, CurrentID=0, Status='live', Symbol='Alppl2', Aliases='Akp5, C77216, D1Ertd816e, EAP', description='alkaline phosphatase, placental-like 2', other_designations='alkaline phosphatase 5|embryonic alkaline phosphatase|embryonic-type alkaline phosphatase', map_location='1 44.05 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87086690, end_position_on_the_genomic_accession=87096250, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11651, CurrentID=0, Status='live', Symbol='Akt1', Aliases='Akt, PKB, PKB/Akt, PKBalpha, Rac', description='thymoma viral proto-oncogene 1', other_designations='AKT1 kinase|Akt1m protein|PKB alpha|RAC-PK-alpha|protein kinase B alpha|protein kinase B-alpha|proto-oncogene c-AKT|related to A and C kinases', map_location='12 61.2 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112653821, end_position_on_the_genomic_accession=112674802, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  11652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11652, CurrentID=0, Status='live', Symbol='Akt2', Aliases='2410016A19Rik, AW554154, PKB, PKBbeta', description='thymoma viral proto-oncogene 2', other_designations='PKB beta|RAC-PK-beta|RAC-beta serine/threonine protein kinase|protein kinase Akt-2|protein kinase B beta|protein kinase B, beta', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27591557, end_position_on_the_genomic_accession=27639453, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  11655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11655, CurrentID=0, Status='live', Symbol='Alas1', Aliases='ALAS, ALAS-N, Alas-1, Alas-h', description='aminolevulinic acid synthase 1', other_designations='5-aminolevulinic acid synthase 1|aminolevulinate synthase H|delta-ALA synthase 1|delta-aminolevulinate synthase 1|succinyl-CoA: glycine C-succinyl transferase', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106233455, end_position_on_the_genomic_accession=106247954, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  11656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11656, CurrentID=0, Status='live', Symbol='Alas2', Aliases='ALAS, ALAS-E, ALASE, Alas-2', description='aminolevulinic acid synthase 2, erythroid', other_designations='5-aminolevulinic acid synthase 2|delta-ALA synthase 2|delta-aminolevulinate synthase 2|erythroid-specific ALAS', map_location='X 68.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=150547417, end_position_on_the_genomic_accession=150570622, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11657, CurrentID=0, Status='live', Symbol='Alb', Aliases='Alb-11, BCL002, Alb', description='albumin', other_designations='albumin 1', map_location='5 44.7 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90460889, end_position_on_the_genomic_accession=90476603, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  11658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11658, CurrentID=0, Status='live', Symbol='Alcam', Aliases='AI853494, BEN, CD166, DM-GRASP, MuSC, SC1', description='activated leukocyte cell adhesion molecule', other_designations='', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=52248996, end_position_on_the_genomic_accession=52453255, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  11666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11666, CurrentID=0, Status='live', Symbol='Abcd1', Aliases='ALDP, Ald, Aldgh', description='ATP-binding cassette, sub-family D (ALD), member 1', other_designations='ATP-binding cassette, sub-family D, member 1|X-linked adrenoleukodystrophy (ALD) gene homolog|adrenoleukodystrophy protein', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73716597, end_position_on_the_genomic_accession=73738287, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11668, CurrentID=0, Status='live', Symbol='Aldh1a1', Aliases='Ahd-2, Ahd2, Aldh1, Aldh1a2, E1, Raldh1', description='aldehyde dehydrogenase family 1, subfamily A1', other_designations='ALDH-E1|ALHDII|RALDH 1|alcohol dehydrogenase family 1, subfamily A1|alcohol dehydrogenase family 1, subfamily A2|aldehyde dehydrogenase 1, liver cytosolic (class 1)|aldehyde dehydrogenase family 1 member A1|aldehyde dehydrogenase, cytosolic', map_location='19 13.91 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=20601982, end_position_on_the_genomic_accession=20643462, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11669, CurrentID=0, Status='live', Symbol='Aldh2', Aliases='AHD-M1, ALDH-E2, ALDHI, Ahd-5, Ahd5', description='aldehyde dehydrogenase 2, mitochondrial', other_designations='ALDH class 2', map_location='5 F-G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121567687, end_position_on_the_genomic_accession=121593824, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  11670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11670, CurrentID=0, Status='live', Symbol='Aldh3a1', Aliases='Ahd-4, Ahd4, Aldh, Aldh3', description='aldehyde dehydrogenase family 3, subfamily A1', other_designations='alcohol dehydrogenase family 3, subfamily A1|aldehyde dehydrogenase 3, stomach cytosolic (class 3)|aldehyde dehydrogenase 4|aldehyde dehydrogenase family 3 member A1|dioxin-inducible aldehyde dehydrogenase 3', map_location='11 37.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61207537, end_position_on_the_genomic_accession=61218418, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  11671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11671, CurrentID=0, Status='live', Symbol='Aldh3a2', Aliases='AI194803, Ahd-3, Ahd-3r, Ahd3, Ahd3-r, Aldh4, Aldh4-r, FALDH', description='aldehyde dehydrogenase family 3, subfamily A2', other_designations='alcohol dehydrogenase family 3, subfamily A2|aldehyde dehydrogenase 3|aldehyde dehydrogenase 4, liver microsomal (class 3) regulator|aldehyde dehydrogenase family 3 member A2', map_location='11 37.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61224486, end_position_on_the_genomic_accession=61267465, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  667034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667034, CurrentID=0, Status='live', Symbol='Pnp2', Aliases='', description='purine-nucleoside phosphorylase 2', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50956141, end_position_on_the_genomic_accession=50964749, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11676, CurrentID=0, Status='live', Symbol='Aldoc', Aliases='AI847350, AU040929, Aldo3, Scrg2', description='aldolase C, fructose-bisphosphate', other_designations='aldolase 3, C isoform|brain-type aldolase|scrapie responsive gene 2|scrapie-responsive protein 2|zebrin II', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78323073, end_position_on_the_genomic_accession=78327781, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11677, CurrentID=0, Status='live', Symbol='Akr1b3', Aliases='ALR2, AR, Ahr-1, Ahr1, Akr1b1, Aldor1, Aldr1', description='aldo-keto reductase family 1, member B3 (aldose reductase)', other_designations='aldehyde reductase 1|aldo-keto reductase family 1, member B1 (aldose reductase)', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34303930, end_position_on_the_genomic_accession=34317489, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11682, CurrentID=0, Status='live', Symbol='Alk', Aliases='CD246, Tcrz', description='anaplastic lymphoma kinase', other_designations='', map_location='17 43.77 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71868988, end_position_on_the_genomic_accession=72605410, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  11684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11684, CurrentID=0, Status='live', Symbol='Alox12', Aliases='9930022G08Rikp, P-12LO, Alox12', description='arachidonate 12-lipoxygenase', other_designations='12S-LOX|12S-lipoxygenase|lipoxin synthase 12-LO|platelet-type lipoxygenase 12', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70241455, end_position_on_the_genomic_accession=70255386, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11685, CurrentID=0, Status='live', Symbol='Alox12e', Aliases='8-LOX, AW259591, Alox12-ps1, Alox12-ps2, Aloxe, e-LOX1', description='arachidonate lipoxygenase, epidermal', other_designations='12-LOX|arachidonate 12-lipoxygenase, epidermal', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70315503, end_position_on_the_genomic_accession=70322727, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11686, CurrentID=0, Status='live', Symbol='Alox12b', Aliases='12R-LOX, Aloxe2, e-LOX2', description='arachidonate 12-lipoxygenase, 12R type', other_designations='12R-lipoxygenase|epidermis-type lipoxygenase 12|epidermis-type lipoxygenase 2', map_location='11 42.38 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69157072, end_position_on_the_genomic_accession=69169791, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  11687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11687, CurrentID=0, Status='live', Symbol='Alox15', Aliases='12-LO, 12/15-LO, 15-LOX, Alox12l, L-12LO', description='arachidonate 15-lipoxygenase', other_designations='12-LOX|12/15-lipoxygenase|arachidonate 12-lipoxygenase, leukocyte-type|arachidonate omega-6 lipoxygenase', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70344147, end_position_on_the_genomic_accession=70375918, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11688, CurrentID=0, Status='live', Symbol='Alox8', Aliases='15-LOX-2, 15-LOX-B, 8-LOX, 8S-LOX, Alox15b', description='arachidonate 8-lipoxygenase', other_designations='15-lipoxygenase 2|8S-lipoxygenase|arachidonate 15-lipoxygenase B|arachidonate 15-lipoxygenase type II|arachidonate 15-lipoxygenase, second type', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69183885, end_position_on_the_genomic_accession=69197843, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11689, CurrentID=0, Status='live', Symbol='Alox5', Aliases='5-LO, 5-LOX, 5LO, 5LX, AI850497, F730011J02', description='arachidonate 5-lipoxygenase', other_designations='5-lipoxygenase', map_location='6 53.79 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116410069, end_position_on_the_genomic_accession=116461210, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11690, CurrentID=0, Status='live', Symbol='Alox5ap', Aliases='Flap', description='arachidonate 5-lipoxygenase activating protein', other_designations='MK-886-binding protein', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149265058, end_position_on_the_genomic_accession=149288047, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11692, CurrentID=0, Status='live', Symbol='Gfer', Aliases='Alr, ERV1', description='growth factor, erv1 (S. cerevisiae)-like (augmenter of liver regeneration)', other_designations='augmenter of liver regeneration', map_location='17 12.48 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24693189, end_position_on_the_genomic_accession=24696227, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11694, CurrentID=0, Status='live', Symbol='Alx3', Aliases='', description='aristaless-like homeobox 3', other_designations='proline-rich transcription factor Alx3', map_location='3 46.83 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107595031, end_position_on_the_genomic_accession=107605875, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11695, CurrentID=0, Status='live', Symbol='Alx4', Aliases='lst', description='aristaless-like homeobox 4', other_designations="ALX-4|Aristaless-like 4|Strong's luxoid|aristaless 4", map_location='2 51.62 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93642434, end_position_on_the_genomic_accession=93681341, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11697, CurrentID=0, Status='live', Symbol='am', Aliases='', description='amputated', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11698, CurrentID=0, Status='live', Symbol='Ambn', Aliases='', description='ameloblastin', other_designations='', map_location='5 43.63 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88455991, end_position_on_the_genomic_accession=88468531, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11699, CurrentID=0, Status='live', Symbol='Ambp', Aliases='AI194774, ASPI, HI-30, Intin4, Itil, UTI', description='alpha 1 microglobulin/bikunin', other_designations='Urinary Trypsin Inhibitor|ulinastatin', map_location='4 C1-C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63143279, end_position_on_the_genomic_accession=63154142, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11702, CurrentID=0, Status='live', Symbol='Amd1', Aliases='AdoMetDC, Amd-1, SAMDC, SAMDC 1, adoMetDC1', description='S-adenosylmethionine decarboxylase 1', other_designations='adoMetDC 1', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40287458, end_position_on_the_genomic_accession=40302188, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11704, CurrentID=0, Status='live', Symbol='Amelx', Aliases='ALGN, AMGL, AMGX, Amel, Amg, LRAP, Rgsc888', description='amelogenin, X-linked', other_designations='amelogenin X chromosome|enamel matrix protein|leucine-rich amelogenin peptide', map_location='X 78.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=169175044, end_position_on_the_genomic_accession=169187287, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11705, CurrentID=0, Status='live', Symbol='Amh', Aliases='MIS', description='anti-Mullerian hormone', other_designations='Mullerian inhibiting substance|anti-Muellerian hormone|muellerian-inhibiting substance', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80805021, end_position_on_the_genomic_accession=80807679, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11717, CurrentID=0, Status='live', Symbol='Ampd3', Aliases='', description='adenosine monophosphate deaminase 3', other_designations='AMP deaminase H-type|heart-type AMPD', map_location='7 E2-E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110767837, end_position_on_the_genomic_accession=110812409, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  11720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11720, CurrentID=0, Status='live', Symbol='Mat1a', Aliases='AI046368, AdoMet, Ams, MAT, MATA1, SAMS, SAMS1', description='methionine adenosyltransferase I, alpha', other_designations='adoMet synthetase 1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=41105012, end_position_on_the_genomic_accession=41124428, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11722, CurrentID=0, Status='live', Symbol='Amy1', Aliases='Amy-1a, C030014B17Rik, Amy1', description='amylase 1, salivary', other_designations='1,4-alpha-D-glucan glucanohydrolase 1|alpha amylase 1|salivary and hepatic alpha-amylase', map_location='3 49.35 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113555952, end_position_on_the_genomic_accession=113577750, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  11727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11727, CurrentID=0, Status='live', Symbol='Ang', Aliases='AI3855861, Rnase5, Rnase5a, Ang', description='angiogenin, ribonuclease, RNase A family, 5', other_designations='RNase 5|angiogenin, ribonuclease A family, member 1|angiogenin-1|ribonuclease 5', map_location='14 B-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51091077, end_position_on_the_genomic_accession=51102009, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  11731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11731, CurrentID=0, Status='live', Symbol='Ang2', Aliases='Angrp, Raa3, Rnase5b', description='angiogenin, ribonuclease A family, member 2', other_designations='angiogenin related protein|angiogenin-related protein|ribonuclease A a3', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51195486, end_position_on_the_genomic_accession=51195923, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  11732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11732, CurrentID=0, Status='live', Symbol='Ank', Aliases='Ankh, D15Ertd221e, ank, mKIAA1581', description='progressive ankylosis', other_designations='fn54 protein', map_location='15 10.23 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=27466677, end_position_on_the_genomic_accession=27594909, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11733, CurrentID=0, Status='live', Symbol='Ank1', Aliases='Ank-1, nb, pale', description='ankyrin 1, erythroid', other_designations='erythrocyte ankyrin|normoblastic anemia', map_location='8 11.42 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22974836, end_position_on_the_genomic_accession=23150502, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  11735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11735, CurrentID=0, Status='live', Symbol='Ank3', Aliases='2900054D09Rik, AI314020, Ank-3, AnkG, Ankyrin-3, Ankyrin-G', description='ankyrin 3, epithelial', other_designations='brain-specific ankyrin-G', map_location='10 36.1 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=69533580, end_position_on_the_genomic_accession=70027436, orientation='plus', exon_count=51, OMIM='', no_hdr0=''),
  11736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11736, CurrentID=0, Status='live', Symbol='Ankfy1', Aliases='Ankhzn, ZFYVE14, mKIAA1255', description='ankyrin repeat and FYVE domain containing 1', other_designations='Rabankyrin-5|ankyrin repeat and FYVE domain-containing protein 1|ankyrin repeat hooked to a zinc finger motif|ankyrin repeat hooked to zinc finger motif|ankyrin repeats hooked to a zinc finger motif|rank-5', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72690002, end_position_on_the_genomic_accession=72772146, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  11737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11737, CurrentID=0, Status='live', Symbol='Anp32a', Aliases='Anp32, I1PP2A, LANP, PHAP1, W91701, pp32', description='acidic (leucine-rich) nuclear phosphoprotein 32 family, member A', other_designations='acidic nuclear phosphoprotein 32|acidic nuclear phosphoprotein pp32|leucine-rich acidic nuclear protein|potent heat-stable protein phosphatase 2A inhibitor I1PP2A|vacidic (leucine-rich) nuclear phosphoprotein 32 family, member A', map_location='9 33.68 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62341343, end_position_on_the_genomic_accession=62378804, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11739, CurrentID=0, Status='live', Symbol='Slc25a4', Aliases='AU019225, Ant1', description='solute carrier family 25 (mitochondrial carrier, adenine nucleotide translocator), member 4', other_designations='ADP,ATP carrier protein 1|ADP,ATP carrier protein, heart/skeletal muscle|ANT 1|adenine nucleotide translocase-1|adenine nucleotide translocator 1, skeletal muscle|mANC1|solute carrier family 25 member 4', map_location='8 26.22 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46207339, end_position_on_the_genomic_accession=46211009, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11740, CurrentID=0, Status='live', Symbol='Slc25a5', Aliases='Ant2', description='solute carrier family 25 (mitochondrial carrier, adenine nucleotide translocator), member 5', other_designations='ADP, ATP carrier protein 2|ADP,ATP carrier protein 2|adenine nucleotide translocase|adenine nucleotide translocator 2, fibroblast', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36795652, end_position_on_the_genomic_accession=36798806, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11743, CurrentID=0, Status='live', Symbol='anx', Aliases='', description='anorexia', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11744, CurrentID=0, Status='live', Symbol='Anxa11', Aliases='A830099O17Rik, Anx11', description='annexin A11', other_designations='Annexin A11 (Annexin XI) (Calcyclin-associated annexin 50) (CAP-50)|CAP-50|annexin XI|annexin-11|calcyclin-associated annexin 50', map_location='14 15.06 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25842151, end_position_on_the_genomic_accession=25886804, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  11745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11745, CurrentID=0, Status='live', Symbol='Anxa3', Aliases='Anx3', description='annexin A3', other_designations='35-alpha calcimedin|PAP-III|annexin III|annexin-3|lipocortin III|placental anticoagulant protein III', map_location='5 47.29 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=96793385, end_position_on_the_genomic_accession=96845968, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11746, CurrentID=0, Status='live', Symbol='Anxa4', Aliases='AI265406, AIV, AW106930, Anx4, Xanx-4', description='annexin A4', other_designations='annexin IV|annexin-4', map_location='6 37.75 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86736840, end_position_on_the_genomic_accession=86793593, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  11747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11747, CurrentID=0, Status='live', Symbol='Anxa5', Aliases='Anx5, R74653', description='annexin A5', other_designations='CBP-I|PAP-I|PP4|VAC-alpha|anchorin CII|annexin V|annexin-5|calphobindin I|endonexin II|lipocortin V|placental anticoagulant protein 4|placental anticoagulant protein I|thromboplastin inhibitor|vascular anticoagulant-alpha', map_location='3 17.46 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36448923, end_position_on_the_genomic_accession=36475887, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  11749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11749, CurrentID=0, Status='live', Symbol='Anxa6', Aliases='AW107198, Anx6, AnxVI, Cabm, Camb', description='annexin A6', other_designations='67 kDa calelectrin|CPB-II|annexin VI, p68|annexin-6|calphobindin-II|chromobindin-20|lipocortin VI|p68|p70', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54978962, end_position_on_the_genomic_accession=55033471, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  11750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11750, CurrentID=0, Status='live', Symbol='Anxa7', Aliases='AI265384, AI316497, Anx7, synexin', description='annexin A7', other_designations='annexin VII|annexin-7', map_location='14 11.53 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20455261, end_position_on_the_genomic_accession=20480133, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11752, CurrentID=0, Status='live', Symbol='Anxa8', Aliases='AI466995, Anx8', description='annexin A8', other_designations='annexin 8|annexin VIII', map_location='14 20.8 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34085979, end_position_on_the_genomic_accession=34100571, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11754, CurrentID=0, Status='live', Symbol='Aoc3', Aliases='SSAO, VAP1', description='amine oxidase, copper containing 3', other_designations='VAP-1|copper amine oxidase|semicarbazide-sensitive amine oxidase|vascular adhesion protein 1', map_location='11 B2-B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101330606, end_position_on_the_genomic_accession=101339430, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11757, CurrentID=0, Status='live', Symbol='Prdx3', Aliases='AW822249, Aop1, D0Tohi1, Ef2l, Mer5, Prx3, SP22, TDXM', description='peroxiredoxin 3', other_designations='AOP-1|Prx III|anti-oxidant protein 1|antioxidant protein|mitochondrial Trx dependent peroxide reductase|mitochondrial thioredoxin dependent peroxide reductase|perioredoxin-3|peroxiredoxin III', map_location='19 56.52 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60864066, end_position_on_the_genomic_accession=60874538, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11758, CurrentID=0, Status='live', Symbol='Prdx6', Aliases='1-Cys Prx, 1-cysPrx, 9430088D19Rik, AA690119, Aop2, Brp-12, CP-3, GPx, Ltw-4, Ltw4, Lvtw-4, NSGPx, ORF06, Prdx5, aiPLA2', description='peroxiredoxin 6', other_designations='1-Cys peroxiredoxin|acidic calcium-independent phospholipase A2|anti-oxidant protein 2, related sequence 3|antioxidant protein 2|non-selenium glutathione peroxidase|peroxiredoxin 5, related sequence 3|peroxiredoxin 6, related sequence 3', map_location='1 69.75 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161240112, end_position_on_the_genomic_accession=161251210, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11761, CurrentID=0, Status='live', Symbol='Aox1', Aliases='AI196512, AI255253, Ao, Aox-1, Aox-2, Aox2, Moro, Ro', description='aldehyde oxidase 1', other_designations='aldehyde oxidase 2|azaheterocycle hydroxylase 1|retinal oxidase', map_location='1 C1-C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58029809, end_position_on_the_genomic_accession=58106410, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  11764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11764, CurrentID=0, Status='live', Symbol='Ap1b1', Aliases='Adtb1, b2b1660Clo', description='adaptor protein complex AP-1, beta 1 subunit', other_designations='Mutant line 1660|adapter-related protein complex 1 subunit beta-1|adaptor protein complex AP-1 subunit beta-1|adaptor-related protein complex 1 subunit beta-1|adaptor-related protein complex AP-1, beta 1 subunit|beta-1-adaptin|beta-adaptin 1|beta-prime adaptin|beta1-adaptin|clathrin assembly protein complex 1 beta large chain|golgi adaptor HA1/AP1 adaptin beta subunit', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4947521, end_position_on_the_genomic_accession=5042794, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  11765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11765, CurrentID=0, Status='live', Symbol='Ap1g1', Aliases='AA409002, AU041323, AW551707, Adtg, D8Ertd374e', description='adaptor protein complex AP-1, gamma 1 subunit', other_designations='adapter-related protein complex 1 subunit gamma-1|adaptor protein complex AP-1 subunit gamma-1|adaptor-related protein complex 1 subunit gamma-1|clathrin assembly protein complex 1 gamma-1 large chain|gamma-adaptin|gamma1-adaptin|golgi adaptor HA1/AP1 adaptin subunit gamma-1', map_location='8 57.26 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109778583, end_position_on_the_genomic_accession=109864210, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  11766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11766, CurrentID=0, Status='live', Symbol='Ap1g2', Aliases='Adtg2, G2ad', description='adaptor protein complex AP-1, gamma 2 subunit', other_designations='gamma 2-adaptin', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55098835, end_position_on_the_genomic_accession=55106593, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  11767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11767, CurrentID=0, Status='live', Symbol='Ap1m1', Aliases='AA408894, AP47, Adtm1A, Cltnm, [m]1A, mu1A', description='adaptor-related protein complex AP-1, mu subunit 1', other_designations='AP-mu chain family member mu1A|adapter-related protein complex 1 subunit mu-1|adaptor protein complex AP-1 mu-1 subunit|adaptor protein complex AP-1 subunit mu-1|adaptor-related protein complex 1 mu-1 subunit|adaptor-related protein complex 1 subunit mu-1|clathrin assembly protein complex 1 medium chain 1|clathrin assembly protein complex 1 mu-1 medium chain 1|clathrin coat assembly protein AP47|clathrin coat-associated protein AP47|golgi adaptor HA1/AP1 adaptin mu-1 subunit|mu-adaptin 1|mu1A-adaptin', map_location='8 34.92 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72240050, end_position_on_the_genomic_accession=72257385, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  11768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11768, CurrentID=0, Status='live', Symbol='Ap1m2', Aliases='D9Ertd818e, [m]1B, mu1B', description='adaptor protein complex AP-1, mu 2 subunit', other_designations='AP-mu chain family member mu1B|adaptor protein complex AP-1 mu-2 subunit|adaptor protein complex AP-1 subunit mu-2|adaptor-related protein complex 1 mu-2 subunit|adaptor-related protein complex 1 subunit mu-2|adaptor-related protein complex AP-1, mu 2 subunit|clathrin assembly protein complex 1 medium chain 2|clathrin assembly protein complex 1 mu-2 medium chain 2|golgi adaptor HA1/AP1 adaptin mu-2 subunit|mu-adaptin 2|mu1B-adaptin', map_location='9 7.76 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21294613, end_position_on_the_genomic_accession=21312349, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  11769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11769, CurrentID=0, Status='live', Symbol='Ap1s1', Aliases='AP19', description='adaptor protein complex AP-1, sigma 1', other_designations='HA1 19 kDa subunit|adapter-related protein complex 1 sigma-1A subunit|adapter-related protein complex 1 subunit sigma-1A|adaptor protein complex AP-1 sigma-1A subunit|adaptor protein complex AP-1 subunit sigma-1A|adaptor-related protein complex 1 subunit sigma-1A|adaptor-related protein complex AP-1, sigma 1|clathrin assembly protein complex 1 sigma-1A small chain|clathrin coat assembly protein AP19|golgi adaptor HA1/AP1 adaptin sigma-1A subunit|sigma 1a subunit of AP-1 clathrin|sigma-adaptin 1A|sigma1A-adaptin', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137034993, end_position_on_the_genomic_accession=137046060, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11770, CurrentID=0, Status='live', Symbol='Fabp4', Aliases='422/aP2, ALBP/Ap2, Ap2, Lbpl', description='fatty acid binding protein 4, adipocyte', other_designations='3T3-L1 lipid-binding protein|A-FABP|AFABP|ALBP|P15|P2 adipocyte protein|adipocyte lipid-binding protein|adipocyte protein aP2|adipocyte-type fatty acid-binding protein|fatty acid-binding protein 4|myelin P2 protein homolog|protein 422', map_location='3 2.56 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10204342, end_position_on_the_genomic_accession=10208597, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11771, CurrentID=0, Status='live', Symbol='Ap2a1', Aliases='Adtaa', description='adaptor-related protein complex 2, alpha 1 subunit', other_designations='100 kDa coated vesicle protein A|adapter-related protein complex 2 alpha-1 subunit|adapter-related protein complex 2 subunit alpha-1|adaptor protein complex AP-2 subunit alpha-1|adaptor protein complex AP-2, alpha 1 subunit|adaptor-related protein complex 2 subunit alpha-1|adaptor-related protein complex AP-2, alpha 1 subunit|alpha-adaptin A|alpha1-adaptin|clathrin assembly protein complex 2 alpha-A large chain|plasma membrane adaptor HA2/AP2 adaptin alpha A subunit', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44900373, end_position_on_the_genomic_accession=44929490, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  11772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11772, CurrentID=0, Status='live', Symbol='Ap2a2', Aliases='2410074K14Rik, AF006990, AW146353, Adtab, C78001, L25', description='adaptor-related protein complex 2, alpha 2 subunit', other_designations='100 kDa coated vesicle protein C|adapter-related protein complex 2 alpha-2 subunit|adapter-related protein complex 2 subunit alpha-2|adaptor protein complex AP-2 subunit alpha-2|adaptor protein complex AP-2, alpha 2 subunit|adaptor-related protein complex 2 subunit alpha-2|adaptor-related protein complex AP-2, alpha 2 subunit|alpha-C adaptin|alpha-adaptin C|alpha2-adaptin|clathrin assembly protein complex 2 alpha-C large chain|plasma membrane adaptor HA2/AP2 adaptin alpha C subunit', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141562180, end_position_on_the_genomic_accession=141633011, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  11773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11773, CurrentID=0, Status='live', Symbol='Ap2m1', Aliases='', description='adaptor-related protein complex 2, mu 1 subunit', other_designations='AP-2 mu chain|adapter-related protein complex 2 mu subunit|adaptor protein complex AP-2 subunit mu|adaptor protein complex AP-2, mu1|adaptor-related protein complex 2 subunit mu|adaptor-related protein complex AP-2, mu1|clathrin assembly protein complex 2 medium chain|clathrin assembly protein complex 2 mu medium chain|clathrin coat assembly protein AP50|clathrin coat-associated protein AP50|clathrin-associated AP-2|mu2-adaptin|plasma membrane adaptor AP-2 50 kDa protein', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20535480, end_position_on_the_genomic_accession=20544909, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11774, CurrentID=0, Status='live', Symbol='Ap3b1', Aliases='AP-3, AU015684, C78395, Hps2, beta3A, pe, pearl, rim2', description='adaptor-related protein complex 3, beta 1 subunit', other_designations='adapter-related protein complex 3 subunit beta-1|adaptor protein complex AP-3 beta-1 subunit|adaptor protein complex AP-3 subunit beta-1|adaptor-related protein complex 3 subunit beta-1|adaptor-related protein complex AP-3, beta 1 subunit|beta-3A-adaptin|beta3A-adaptin|clathrin assembly protein complex 3 beta-1 large chain|recombination induced mutation 2', map_location='13 49.22 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=94358952, end_position_on_the_genomic_accession=94566316, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  11775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11775, CurrentID=0, Status='live', Symbol='Ap3b2', Aliases='AI549966, AU042881, Naptb, [b]-NAP, beta-NAP, beta3B', description='adaptor-related protein complex 3, beta 2 subunit', other_designations='AP-3 complex beta-2 subunit|adapter-related protein complex 3 subunit beta-2|adaptor protein complex AP-3 beta-2 subunit|adaptor protein complex AP-3 subunit beta-2|adaptor-related protein complex 3 subunit beta-2|adaptor-related protein complex AP-3, beta 2 subunit|beta-3B-adaptin|beta3B-adaptin|clathrin assembly protein complex 3 beta-2 large chain', map_location='7 45.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81460399, end_position_on_the_genomic_accession=81494038, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  11776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11776, CurrentID=0, Status='live', Symbol='Ap3d1', Aliases='AA407035, Ap3d, Bolvr, mBLVR1, mh, mocha', description='adaptor-related protein complex 3, delta 1 subunit', other_designations='AP-3 complex subunit delta|adapter-related protein complex 3 subunit delta-1|adaptor-related protein complex 3 subunit delta-1|adaptor-related protein complex 3, delta subunit|adaptor-related protein complex AP-3, delta subunit|delta-adaptin', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80706956, end_position_on_the_genomic_accession=80742282, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  11777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11777, CurrentID=0, Status='live', Symbol='Ap3s1', Aliases='', description='adaptor-related protein complex 3, sigma 1 subunit', other_designations='AP-3 complex sigma-3A subunit|adapter-related protein complex 3 sigma-1 subunit|adaptor-related protein complex AP-3, sigma 1 subunit|sigma-3A-adaptin|sigma-adaptin 3a', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46741876, end_position_on_the_genomic_accession=46790826, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11778, CurrentID=0, Status='live', Symbol='Ap3s2', Aliases='[s]3B', description='adaptor-related protein complex 3, sigma 2 subunit', other_designations='AP-3 complex sigma-3B subunit|AP-3 complex subunit sigma-3B|adapter-related protein complex 3 sigma-2 subunit|adapter-related protein complex 3 subunit sigma-2|adaptor-related protein complex 3 subunit sigma-2|adaptor-related protein complex AP-3, sigma 2 subunit|sigma 3B|sigma-3B-adaptin|sigma-adaptin 3b|sigma3B-adaptin', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79875325, end_position_on_the_genomic_accession=79920640, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11781, CurrentID=0, Status='live', Symbol='Ap4m1', Aliases='4930443L05Rik, Ap4m4', description='adaptor-related protein complex AP-4, mu 1', other_designations='AP-4 adapter complex mu subunit|AP-4 adaptor complex mu subunit|adapter-related protein complex 4 mu-1 subunit|adapter-related protein complex 4 subunit mu-1|adaptor-related protein complex 4 subunit mu-1|adaptor-related protein complex AP-4 mu4 subunit|mu subunit of AP-4|mu-ARP2|mu-adaptin-related protein 2|mu4|mu4-adaptin', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138172021, end_position_on_the_genomic_accession=138178687, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  11782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11782, CurrentID=0, Status='live', Symbol='Ap4s1', Aliases='AI314282', description='adaptor-related protein complex AP-4, sigma 1', other_designations='AP-4 adapter complex subunit sigma-1|AP-4 adaptor complex subunit sigma-1|adapter-related protein complex 4 subunit sigma-1|adaptor-related protein complex 4 subunit sigma-1|sigma-1 subunit of AP-4|sigma-4-adaptin|sigma4-adaptin', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51689888, end_position_on_the_genomic_accession=51738939, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11783, CurrentID=0, Status='live', Symbol='Apaf1', Aliases='6230400I06Rik, Apaf-1l, fog, mKIAA0413, Apaf1', description='apoptotic peptidase activating factor 1', other_designations='apoptotic protease activating factor 1|forebrain overgrowth', map_location='10 C3-D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=90989311, end_position_on_the_genomic_accession=91082743, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  11784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11784, CurrentID=0, Status='live', Symbol='Apba2', Aliases='X11-like, X11L, XllL, mXllL', description='amyloid beta (A4) precursor protein-binding, family A, member 2', other_designations='Mint 2|adapter protein X11beta|mint-2|neuron-specific X11L protein|neuronal Munc18-1-interacting protein 2', map_location='7 34.65 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64501706, end_position_on_the_genomic_accession=64753878, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  11785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11785, CurrentID=0, Status='live', Symbol='Apbb1', Aliases='Fe65, Rir', description='amyloid beta (A4) precursor protein-binding, family B, member 1', other_designations='', map_location='7 55.9 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105558465, end_position_on_the_genomic_accession=105581653, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  11787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11787, CurrentID=0, Status='live', Symbol='Apbb2', Aliases='2310007D03Rik, FE65L1, Rirl1, TR2L, Zfra', description='amyloid beta (A4) precursor protein-binding, family B, member 2', other_designations='zinc finger-like protein', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=66298725, end_position_on_the_genomic_accession=66618817, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  11789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11789, CurrentID=0, Status='live', Symbol='Apc', Aliases='AI047805, AU020952, AW124434, CC1, Min, mAPC', description='adenomatosis polyposis coli', other_designations='multiple intestinal neoplasia', map_location='18 18.53 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34220307, end_position_on_the_genomic_accession=34322190, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  11790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11790, CurrentID=0, Status='live', Symbol='Speg', Aliases='AW125581, Apeg1, BPEG, D1Bwg1450e, SPEGalpha, SPEGbeta, mKIAA1297', description='SPEG complex locus', other_designations='APEG-1|aortic preferentially expressed gene 1|aortic preferentially expressed protein 1', map_location='1 38.88 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75375297, end_position_on_the_genomic_accession=75432320, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  11792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11792, CurrentID=0, Status='live', Symbol='Apex1', Aliases='APE, Apex, HAP1, Ref-1', description='apurinic/apyrimidinic endonuclease 1', other_designations='AP endonuclease 1|APEN|APEX nuclease|apurinic-apyrimidinic endonuclease 1|redox factor-1', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50924944, end_position_on_the_genomic_accession=50927188, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11793, CurrentID=0, Status='live', Symbol='Atg5', Aliases='2010107M05Rik, 3110067M24Rik, AW319544, Apg5ll, C88337, Paddy, Atg5', description='autophagy related 5', other_designations='autophagy-related 5', map_location='10 23.24 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=44268358, end_position_on_the_genomic_accession=44364299, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11796, CurrentID=0, Status='live', Symbol='Birc3', Aliases='AW107670, Api1, Api2, Birc2, HIAP2, IAP1, IAP2, MIAP1, MIAP2, MIHB, MIHC, RNF49, cIAP-1, cIAP-2, cIAP1, cIAP2', description='baculoviral IAP repeat-containing 3', other_designations='IAP-1|apoptosis inhibitor 1|baculoviral IAP repeat-containing 2|inhibitor of apoptosis protein 1|mIAP-1', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7848699, end_position_on_the_genomic_accession=7873213, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  11797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11797, CurrentID=0, Status='live', Symbol='Birc2', Aliases='AW146227, Api1, Api2, Birc3, C330006D17Rik, HIAP1, HIAP2, IAP1, IAP2, MIAP1, MIAP2, MIHB, MIHC, RNF48, cIAP1, cIAP2, mcIAP1', description='baculoviral IAP repeat-containing 2', other_designations='IAP-2|apoptosis inhibitor 2|baculoviral IAP repeat-containing 3|inhibitor of apoptosis protein 2|mIAP-2', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7818226, end_position_on_the_genomic_accession=7837133, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11798, CurrentID=0, Status='live', Symbol='Xiap', Aliases='1110015C02Rik, Aipa, Api3, Birc4, IAP3, ILP-1, MIHA', description='X-linked inhibitor of apoptosis', other_designations='IAP homolog A|X-linked IAP|baculoviral IAP repeat-containing protein 4|inhibitor of apoptosis protein 3', map_location='X A3-A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=42060402, end_position_on_the_genomic_accession=42109664, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11799, CurrentID=0, Status='live', Symbol='Birc5', Aliases='AAC-11, Api4, TIAP, survivin40', description='baculoviral IAP repeat-containing 5', other_designations='apoptosis inhibitor 4|apoptosis inhibitor survivin|survivin', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117849237, end_position_on_the_genomic_accession=117855743, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  11800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11800, CurrentID=0, Status='live', Symbol='Api5', Aliases='AAC-11, AI196452, API-5', description='apoptosis inhibitor 5', other_designations='apoptosis inhibitory protein 5', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=94411727, end_position_on_the_genomic_accession=94438181, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11801, CurrentID=0, Status='live', Symbol='Cd5l', Aliases='1/6, AAC-11, AI047839, Api6, CT2, Pdp, Sp-alpha', description='CD5 antigen-like', other_designations='AIM/Spalpha|CT-2|Pdp 1/6|apoptosis inhibitor 6|apoptosis inhibitor expressed by macrophages', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87357881, end_position_on_the_genomic_accession=87371074, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11803, CurrentID=0, Status='live', Symbol='Aplp1', Aliases='', description='amyloid beta (A4) precursor-like protein 1', other_designations='APLP|APLP-1', map_location='7 18.32 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30434980, end_position_on_the_genomic_accession=30445582, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  11804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11804, CurrentID=0, Status='live', Symbol='Aplp2', Aliases='AI790698, APLP-2', description='amyloid beta (A4) precursor-like protein 2', other_designations='CDEBP|CDEI box-binding protein', map_location='9 A2-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=31149557, end_position_on_the_genomic_accession=31211843, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  11806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11806, CurrentID=0, Status='live', Symbol='Apoa1', Aliases='Alp-1, Apoa-1, Brp-14, Ltw-1, Lvtw-1, Sep-1, Sep-2, Sep2, apo-AI, apoA-I', description='apolipoprotein A-I', other_designations='apolipoprotein A1', map_location='9 A2-A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46228630, end_position_on_the_genomic_accession=46230469, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11807, CurrentID=0, Status='live', Symbol='Apoa2', Aliases='Alp-2, Apo-AII, ApoA-II, ApoAII, Apoa-2, Hdl-1', description='apolipoprotein A-II', other_designations='apolipoprotein A2', map_location='1 79.22 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171225011, end_position_on_the_genomic_accession=171226379, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11808, CurrentID=0, Status='live', Symbol='Apoa4', Aliases='Apoa-4', description='apolipoprotein A-IV', other_designations='apo-AIV|apoA-IV|apolipoprotein A4', map_location='9 25.36 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46240844, end_position_on_the_genomic_accession=46243458, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  11810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11810, CurrentID=0, Status='live', Symbol='Apobec1', Aliases='Cdar1', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide 1', other_designations='apolipoprotein B editing complex 1|mRNA(cytosine(6666)) deaminase 1', map_location='6 57.68 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122577792, end_position_on_the_genomic_accession=122603041, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11811, CurrentID=0, Status='live', Symbol='Apobec2', Aliases='Arp1', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide 2', other_designations='apolipoprotein B editing complex 2|mRNA(cytosine(6666)) deaminase 2|probable C->U-editing enzyme APOBEC-2', map_location='17 24.02 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48419231, end_position_on_the_genomic_accession=48432728, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11812, CurrentID=0, Status='live', Symbol='Apoc1', Aliases='Apo-CIB, ApoC-IB, apo-CI, apoC-I', description='apolipoprotein C-I', other_designations='apolipoprotein C1', map_location='7 9.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19689480, end_position_on_the_genomic_accession=19692659, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11813, CurrentID=0, Status='live', Symbol='Apoc2', Aliases='apo-CII, apoC-II', description='apolipoprotein C-II', other_designations='apolipoprotein C2', map_location='7 9.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19671579, end_position_on_the_genomic_accession=19677941, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11814, CurrentID=0, Status='live', Symbol='Apoc3', Aliases='apo-CIII, apoC-III', description='apolipoprotein C-III', other_designations='apolipoprotein C3', map_location='9 25.36 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46233050, end_position_on_the_genomic_accession=46235636, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11815, CurrentID=0, Status='live', Symbol='Apod', Aliases='', description='apolipoprotein D', other_designations='', map_location='16 21.41 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31296192, end_position_on_the_genomic_accession=31314808, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11816, CurrentID=0, Status='live', Symbol='Apoe', Aliases='AI255918, Apo-E', description='apolipoprotein E', other_designations='', map_location='7 9.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19696244, end_position_on_the_genomic_accession=19699166, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11818, CurrentID=0, Status='live', Symbol='Apoh', Aliases='B2GPI, beta-2-GPI, beta2-GPI', description='apolipoprotein H', other_designations='APC inhibitor|activated protein C-binding protein|apo-H|beta(2)GPI|beta-2-glycoprotein I', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=108395297, end_position_on_the_genomic_accession=108414396, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11819, CurrentID=0, Status='live', Symbol='Nr2f2', Aliases='2700033K02Rik, 9430015G03Rik, ARP-1, Aporp1, COUP-TF2, COUP-TFII, COUPTFB, EAR3, SVP40, Tcfcoup2', description='nuclear receptor subfamily 2, group F, member 2', other_designations='COUP transcription factor II|COUP-TF II|apolipoprotein AI regulatory protein 1|apolipoprotein regulatory protein 1', map_location='7 38.66 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=70351946, end_position_on_the_genomic_accession=70366746, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11820, CurrentID=0, Status='live', Symbol='App', Aliases='Abeta, Abpp, Adap, Ag, Cvap, E030013M08Rik, betaApp', description='amyloid beta (A4) precursor protein', other_designations='alzheimer disease amyloid A4 protein homolog|amyloidogenic glycoprotein|appican|protease nexin II', map_location='16 C3-qter', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=84954436, end_position_on_the_genomic_accession=85173952, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  11821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11821, CurrentID=0, Status='live', Symbol='Aprt', Aliases='C85684', description='adenine phosphoribosyl transferase', other_designations='', map_location='8 71.91 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122574637, end_position_on_the_genomic_accession=122576907, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11826, CurrentID=0, Status='live', Symbol='Aqp1', Aliases='CHIP28', description='aquaporin 1', other_designations='AQP-1|DER2|aquaporin-CHIP|delayed early response protein 2|growth factor-induced delayed early response protein|water channel protein for red blood cells and kidney proximal tubule', map_location='6 27.38 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55336299, end_position_on_the_genomic_accession=55348555, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11827, CurrentID=0, Status='live', Symbol='Aqp2', Aliases='AQP-CD, WCH-CD, cph, jpk', description='aquaporin 2', other_designations='ADH water channel|AQP-2|aquaporin-CD|collecting duct water channel protein|water channel protein for renal collecting duct', map_location='15 56.13 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99579056, end_position_on_the_genomic_accession=99584545, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11828, CurrentID=0, Status='live', Symbol='Aqp3', Aliases='AQP-2', description='aquaporin 3', other_designations='AQP-3|aquaglyceroporin-3', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41092724, end_position_on_the_genomic_accession=41098183, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11829, CurrentID=0, Status='live', Symbol='Aqp4', Aliases='WCH4', description='aquaporin 4', other_designations='mercurial-insensitive water channel', map_location='18 8.74 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=15389394, end_position_on_the_genomic_accession=15410977, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11830, CurrentID=0, Status='live', Symbol='Aqp5', Aliases='', description='aquaporin 5', other_designations='AQP-5', map_location='15 56.13 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99591028, end_position_on_the_genomic_accession=99594829, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  11831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11831, CurrentID=0, Status='live', Symbol='Aqp6', Aliases='AI790337', description='aquaporin 6', other_designations='AQP-6', map_location='15 56.13 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99599053, end_position_on_the_genomic_accession=99605477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11832, CurrentID=0, Status='live', Symbol='Aqp7', Aliases='AQP7L, AQPap', description='aquaporin 7', other_designations='AQP-7|aquaglyceroporin-7', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41033074, end_position_on_the_genomic_accession=41048215, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11833, CurrentID=0, Status='live', Symbol='Aqp8', Aliases='AI255744, AQP-8', description='aquaporin 8', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=123462294, end_position_on_the_genomic_accession=123468003, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11834, CurrentID=0, Status='live', Symbol='Aqr', Aliases='AW495846, mKIAA0560', description='aquarius', other_designations='', map_location='2 E5-F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=114101161, end_position_on_the_genomic_accession=114175339, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  11835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11835, CurrentID=0, Status='live', Symbol='Ar', Aliases='AW320017, Tfm', description='androgen receptor', other_designations='dihydrotestosterone receptor|nuclear receptor subfamily 3 group C member 4|testicular feminization', map_location='X 42.82 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=98149750, end_position_on_the_genomic_accession=98317147, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11836, CurrentID=0, Status='live', Symbol='Araf', Aliases='1200013E08Rik, A-Raf, AW4954441, Araf', description='Araf proto-oncogene, serine/threonine kinase', other_designations='A-Raf proto-oncogene serine/threonine-protein kinase|proto-oncogene A-Raf|raf-related oncogene|v-raf murine sarcoma 3611 viral oncogene homolog 1', map_location='X A2-A3.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20797787, end_position_on_the_genomic_accession=20860521, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  11837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11837, CurrentID=0, Status='live', Symbol='Rplp0', Aliases='36B4, Arbp, L10E', description='ribosomal protein, large, P0', other_designations='60S ribosomal protein L10E|acidic ribosomal phosphoprotein P0', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115559467, end_position_on_the_genomic_accession=115563729, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11838, CurrentID=0, Status='live', Symbol='Arc', Aliases='Arc3.1, C86064, arg3.1, mArc', description='activity regulated cytoskeletal-associated protein', other_designations='ARC/ARG3.1|activity-regulated gene 3.1 protein homolog|arc', map_location='15 34.25 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74669081, end_position_on_the_genomic_accession=74672570, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11839, CurrentID=0, Status='live', Symbol='Areg', Aliases='AR, Mcub, Sdgf', description='amphiregulin', other_designations='schwannoma-derived growth factor', map_location='5 44.78 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=91139615, end_position_on_the_genomic_accession=91148432, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11840, CurrentID=0, Status='live', Symbol='Arf1', Aliases='', description='ADP-ribosylation factor 1', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59211411, end_position_on_the_genomic_accession=59228267, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11841, CurrentID=0, Status='live', Symbol='Arf2', Aliases='', description='ADP-ribosylation factor 2', other_designations='', map_location='11 67.64 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103966718, end_position_on_the_genomic_accession=103985337, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11842, CurrentID=0, Status='live', Symbol='Arf3', Aliases='5430400P17Rik, AI854770', description='ADP-ribosylation factor 3', other_designations='', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98737624, end_position_on_the_genomic_accession=98763118, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11843, CurrentID=0, Status='live', Symbol='Arf4', Aliases='AA407803', description='ADP-ribosylation factor 4', other_designations='', map_location='14 16.09 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26638197, end_position_on_the_genomic_accession=26657258, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11844, CurrentID=0, Status='live', Symbol='Arf5', Aliases='', description='ADP-ribosylation factor 5', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=28423640, end_position_on_the_genomic_accession=28426499, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11845, CurrentID=0, Status='live', Symbol='Arf6', Aliases='AI788669, AW496366', description='ADP-ribosylation factor 6', other_designations='', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69372150, end_position_on_the_genomic_accession=69375980, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  11846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11846, CurrentID=0, Status='live', Symbol='Arg1', Aliases='AI, AI256583, Arg-1, PGIF', description='arginase, liver', other_designations='arginase 1, liver|arginase I|liver-type arginase|type I arginase', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24915207, end_position_on_the_genomic_accession=24927470, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11847, CurrentID=0, Status='live', Symbol='Arg2', Aliases='AII, AU022422', description='arginase type II', other_designations='kidney-type arginase|non-hepatic arginase|type II arginase', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79130788, end_position_on_the_genomic_accession=79156301, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11848, CurrentID=0, Status='live', Symbol='Rhoa', Aliases='Arha, Arha1, Arha2', description='ras homolog gene family, member A', other_designations='aplysia ras-related homolog A', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108306205, end_position_on_the_genomic_accession=108337939, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  11852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11852, CurrentID=0, Status='live', Symbol='Rhob', Aliases='AA017882, Arh6, Arhb', description='ras homolog gene family, member B', other_designations='aplysia ras-related homolog B|ras homolog B|ras homolog gene family, member AB', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8497758, end_position_on_the_genomic_accession=8499985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  11853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11853, CurrentID=0, Status='live', Symbol='Rhoc', Aliases='AI324259, Arh9, Arhc', description='ras homolog gene family, member C', other_designations='SIG-61|aplysia ras-related homolog 9 (RhoC)|ras homolog 9 (RhoC)|silica-induced gene 61 protein', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104789034, end_position_on_the_genomic_accession=104794459, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11854, CurrentID=0, Status='live', Symbol='Rhod', Aliases='AI326383, Arhd, Rho, RhoHP1, RhoM', description='ras homolog gene family, member D', other_designations='aplysia ras-related homolog D|ras homolog D (RhoD)', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4425457, end_position_on_the_genomic_accession=4439432, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11855, CurrentID=0, Status='live', Symbol='Arhgap5', Aliases='p190-B, p190B', description='Rho GTPase activating protein 5', other_designations='rho-type GTPase-activating protein 5', map_location='12 22.16 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=52503771, end_position_on_the_genomic_accession=52571978, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11856, CurrentID=0, Status='live', Symbol='Arhgap6', Aliases='AI504284, RhoGAPX-1', description='Rho GTPase activating protein 6', other_designations='rho-type GTPase-activating protein 6|rho-type GTPase-activating protein RhoGAPX-1', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=168795094, end_position_on_the_genomic_accession=169304440, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  11857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11857, CurrentID=0, Status='live', Symbol='Arhgdib', Aliases='D4, Gdid4, Ly-GDI', description='Rho, GDP dissociation inhibitor (GDI) beta', other_designations='rho GDI 2|rho-GDI beta', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136923655, end_position_on_the_genomic_accession=136941899, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11858, CurrentID=0, Status='live', Symbol='Rnd2', Aliases='Arhn, Rohn', description='Rho family GTPase 2', other_designations='aplysia ras-related homolog N (RhoN)|ras homolog N (RhoN)|ras homolog gene family, member N|rho-related GTP-binding protein Rho7|rho7', map_location='11 C-D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101468338, end_position_on_the_genomic_accession=101471306, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11859, CurrentID=0, Status='live', Symbol='Phox2a', Aliases='Arix, Phox2, Pmx2, Pmx2a', description='paired-like homeobox 2a', other_designations='PHOX2A homeodomain protein|aristaless homeobox gene homolog|aristaless homeobox protein homolog|paired mesoderm homeobox 2a', map_location='7 54.66 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101818313, end_position_on_the_genomic_accession=101822726, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  11861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11861, CurrentID=0, Status='live', Symbol='Arl4a', Aliases='AI467555, Arl4', description='ADP-ribosylation factor-like 4A', other_designations='ADP-ribosylation factor-like 4|ADP-ribosylation-like 4', map_location='12 18.06 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=40033291, end_position_on_the_genomic_accession=40037987, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11863, CurrentID=0, Status='live', Symbol='Arnt', Aliases='D3Ertd557e, Drnt, ESTM42, Hif1b, W08714, bHLHe2, mKIAA4051', description='aryl hydrocarbon receptor nuclear translocator', other_designations='HIF-1-beta|HIF1-beta|dioxin receptor, nuclear translocator|hypoxia-inducible factor 1-beta', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95434368, end_position_on_the_genomic_accession=95497240, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  11864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11864, CurrentID=0, Status='live', Symbol='Arnt2', Aliases='Hif-2b, bHLHe1, mKIAA0307', description='aryl hydrocarbon receptor nuclear translocator 2', other_designations='ARNT protein 2', map_location='7 48.35 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=84246274, end_position_on_the_genomic_accession=84410171, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  11865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11865, CurrentID=0, Status='live', Symbol='Arntl', Aliases="Arnt3, BMAL1b, Bmal1, MOP3, bHLHe5, bmal1b'", description='aryl hydrocarbon receptor nuclear translocator-like', other_designations='brain and muscle ARNT-like 1', map_location='7 F2-F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=113178473, end_position_on_the_genomic_accession=113314126, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  11867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11867, CurrentID=0, Status='live', Symbol='Arpc1b', Aliases='41kDa, AA408064, AA408534, AA571392, AF007010, AW208418, L72, SOP2Hs, p41-ARC', description='actin related protein 2/3 complex, subunit 1B', other_designations='arp2/3 complex 41 kDa subunit', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145114256, end_position_on_the_genomic_accession=145128186, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11870, CurrentID=0, Status='live', Symbol='Art1', Aliases='ADPRT, ARTC1, Yac-1', description='ADP-ribosyltransferase 1', other_designations='ADP-ribosyltransferase C2 and C3 toxin-like 1|mono(ADP-ribosyl)transferase 1', map_location='7 54.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102101706, end_position_on_the_genomic_accession=102111758, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11871, CurrentID=0, Status='live', Symbol='Art2a-ps', Aliases='ARTC2, Art2a, Ly92a, Rt-6, Rt6, Rt6-1', description='ADP-ribosyltransferase 2a, pseudogene', other_designations='ADP-ribosyltransferase 2a pseudogene|ADP-ribosyltransferase C2 and C3 toxin-like 2|T-cell NAD(P)(+)--arginine ADP-ribosyltransferase 1|T-cell differentiation marker Rt6 homolog 1|T-cell mono(ADP-ribosyl)transferase 1|mono(ADP-ribosyl)transferase 2A', map_location='7 54.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101552453, end_position_on_the_genomic_accession=101560865, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11872, CurrentID=0, Status='live', Symbol='Art2b', Aliases='ART2.2, ARTC2, Art, Ly92b, Rt-6, Rt6, Rt6-2', description='ADP-ribosyltransferase 2b', other_designations='ADP-ribosyltransferase C2 and C3 toxin-like 2|T-cell NAD(P)(+)--arginine ADP-ribosyltransferase 2|T-cell differentiation marker Rt6 homolog 2|T-cell mono(ADP-ribosyl)transferase 2|mono(ADP-ribosyl)transferase 2B', map_location='7 54.61 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101575729, end_position_on_the_genomic_accession=101585583, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11875, CurrentID=0, Status='live', Symbol='Art5', Aliases='ARTC5, Yac-2', description='ADP-ribosyltransferase 5', other_designations='ADP-ribosyltransferase C2 and C3 toxin-like 5|NAD(P)(+)--arginine ADP-ribosyltransferase 5|mono(ADP-ribosyl)transferase 5', map_location='7 54.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102096879, end_position_on_the_genomic_accession=102100306, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11876, CurrentID=0, Status='live', Symbol='Artn', Aliases='neublastin', description='artemin', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117926160, end_position_on_the_genomic_accession=117930791, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11877, CurrentID=0, Status='live', Symbol='Arvcf', Aliases='', description='armadillo repeat gene deleted in velo-cardio-facial syndrome', other_designations='', map_location='16 11.37 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18348182, end_position_on_the_genomic_accession=18407076, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  11878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11878, CurrentID=0, Status='live', Symbol='Arx', Aliases='', description='aristaless related homeobox', other_designations='aristaless-related homeobox', map_location='X 41.05 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=93286450, end_position_on_the_genomic_accession=93298357, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  11881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11881, CurrentID=0, Status='live', Symbol='Arsb', Aliases='1110007C02Rik, AI480648, As-1, As-1r, As-1s, As-1t, As1, As1-r, As1-s, As1-t, Asr-1, Ast-1', description='arylsulfatase B', other_designations='ASB|G4S|N-acetylgalactosamine-4-sulfatase|arylsulfatase B complex|arylsulfatase B regulation|arylsulfatase B structural|arylsulfatase B temporal regulation', map_location='13 47.88 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93771023, end_position_on_the_genomic_accession=93943016, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11883, CurrentID=0, Status='live', Symbol='Arsa', Aliases='AS-A, ASA, AW212749, As-2, As2, TISP73', description='arylsulfatase A', other_designations='cerebroside-sulfatase', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89472476, end_position_on_the_genomic_accession=89477424, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11885, CurrentID=0, Status='live', Symbol='asa', Aliases='', description='anti-sarcolemmal autoantibodies', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11886, CurrentID=0, Status='live', Symbol='Asah1', Aliases='2310081N20Rik, AC, Asah', description='N-acylsphingosine amidohydrolase 1', other_designations='ACDase|acid CDase|acylsphingosine deacylase', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=41340643, end_position_on_the_genomic_accession=41374697, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11889, CurrentID=0, Status='live', Symbol='Asgr1', Aliases='ASGPR1, Asgr, Asgr-1, HL-1', description='asialoglycoprotein receptor 1', other_designations='ASGP-R 1|ASGPR 1|hepatic lectin 1', map_location='11 42.98 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70054124, end_position_on_the_genomic_accession=70057895, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11890, CurrentID=0, Status='live', Symbol='Asgr2', Aliases='ASGPR2, Asgr, Asgr-2, HL-2', description='asialoglycoprotein receptor 2', other_designations='hepatic lectin 2', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70092307, end_position_on_the_genomic_accession=70106187, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11891, CurrentID=0, Status='live', Symbol='Rab27a', Aliases='2210402C08Rik, 2410003M20Rik, 4933437C11Rik, ash', description='RAB27A, member RAS oncogene family', other_designations='ashen', map_location='9 40.08 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=73044810, end_position_on_the_genomic_accession=73097623, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11893, CurrentID=0, Status='live', Symbol='asp1', Aliases='Ias, asp-1', description='audiogenic seizure prone 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11894, CurrentID=0, Status='live', Symbol='asp2', Aliases='asp, asp-2', description='audiogenic seizure prone 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  667256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667256, CurrentID=0, Status='live', Symbol='4921511M17Rik', Aliases='', description='RIKEN cDNA 4921511M17 gene', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154022780, end_position_on_the_genomic_accession=154033830, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11898, CurrentID=0, Status='live', Symbol='Ass1', Aliases='AA408052, ASS, Ass-1, fold', description='argininosuccinate synthetase 1', other_designations='arginosuccinate synthetase 1|citrulline--aspartate ligase', map_location='2 21.81 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31470270, end_position_on_the_genomic_accession=31520670, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  11899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11899, CurrentID=0, Status='live', Symbol='Astn1', Aliases='Astn, GC14, mKIAA0289', description='astrotactin 1', other_designations='neuronal migration protein GC14', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=158362102, end_position_on_the_genomic_accession=158691786, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  11904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11904, CurrentID=0, Status='live', Symbol='at', Aliases='', description='atrichosis', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  11905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11905, CurrentID=0, Status='live', Symbol='Serpinc1', Aliases='AI114908, ATIII, At-3, At3', description='serine (or cysteine) peptidase inhibitor, clade C (antithrombin), member 1', other_designations='anti-thrombin 3|antithrombin|serine (or cysteine) proteinase inhibitor, clade C (antithrombin), member 1|serpin C1|serpin peptidase inhibitor, clade C, member 1', map_location='1 69.75 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160978606, end_position_on_the_genomic_accession=161003015, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11906, CurrentID=0, Status='live', Symbol='Zfhx3', Aliases='A230102L03Rik, Atbf1, Sci, WBP9, mKIAA4228', description='zinc finger homeobox 3', other_designations='AT motif binding factor 1|AT motif-binding factor 1|AT-binding transcription factor 1|ZFH-3|alpha-fetoprotein enhancer-binding protein|zinc finger homeodomain protein 3', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=108704034, end_position_on_the_genomic_accession=108961636, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11907, CurrentID=0, Status='live', Symbol='Ate1', Aliases='AI225793, AW547406', description='arginyltransferase 1', other_designations='R-transferase 1|arginine-tRNA--protein transferase 1|arginine-tRNA-protein transferase 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=130391493, end_position_on_the_genomic_accession=130520371, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  11908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11908, CurrentID=0, Status='live', Symbol='Atf1', Aliases='', description='activating transcription factor 1', other_designations='TCR-ATF1|cAMP-dependent transcription factor ATF-1', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100227859, end_position_on_the_genomic_accession=100261248, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  11909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11909, CurrentID=0, Status='live', Symbol='Atf2', Aliases='Atf-2, CRE-BP, Creb2, D130078H02Rik, D18875, Tg(Gzma-Klra1)7Wum, mXBP', description='activating transcription factor 2', other_designations='cAMP response element-binding protein CRE-BP1|cAMP-dependent transcription factor ATF-2', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73816509, end_position_on_the_genomic_accession=73892639, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  11910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11910, CurrentID=0, Status='live', Symbol='Atf3', Aliases='LRG-21', description='activating transcription factor 3', other_designations='cAMP-dependent transcription factor ATF-3|transcription factor LRG-21', map_location='1 96.28 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191170296, end_position_on_the_genomic_accession=191183333, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11911, CurrentID=0, Status='live', Symbol='Atf4', Aliases='Atf-4, C/ATF, CREB2, TAXREB67', description='activating transcription factor 4', other_designations='c/EBP-related ATF|cAMP-dependent transcription factor ATF-4|tax-responsive enhancer element-binding protein 67 homolog|taxREB67 homolog', map_location='15 37.85 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80255184, end_position_on_the_genomic_accession=80257545, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  667277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667277, CurrentID=0, Status='live', Symbol='C1rb', Aliases='Gm8551, mC1rB', description='complement component 1, r subcomponent B', other_designations='complement component 1 subcomponent r-B|complement component 1, r-B subcomponent|complement component C1rb', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124570430, end_position_on_the_genomic_accession=124581044, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  11920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11920, CurrentID=0, Status='live', Symbol='Atm', Aliases='AI256621, C030026E19Rik', description='ataxia telangiectasia mutated', other_designations='A-T mutated homolog|ataxia telangiectasia gene mutated in human beings|ataxia telangiectasia mutated homolog', map_location='9 C-D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53437122, end_position_on_the_genomic_accession=53536816, orientation='minus', exon_count=64, OMIM='', no_hdr0=''),
  11921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11921, CurrentID=0, Status='live', Symbol='Atoh1', Aliases='Hath1, MATH-1, Math1, bHLHa14', description='atonal bHLH transcription factor 1', other_designations='atonal homolog 1|helix-loop-helix protein mATH-1', map_location='6 30.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=64729146, end_position_on_the_genomic_accession=64731243, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  11922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11922, CurrentID=0, Status='live', Symbol='Neurod6', Aliases='Atoh2, Math-2, Math2, Nex, Nex1m, bHLHa2', description='neurogenic differentiation 6', other_designations='atonal homolog 2|helix-loop-helix protein mATH-2|protein NEX-1|protein atonal homolog 2', map_location='6 27.53 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55677818, end_position_on_the_genomic_accession=55681263, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  11923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11923, CurrentID=0, Status='live', Symbol='Neurod4', Aliases='AI846749, ATH-3, Atoh3, MATH-3, Math3, bHLHa4', description='neurogenic differentiation 4', other_designations='atonal homolog 3|helix-loop-helix protein mATH-3|protein atonal homolog 3', map_location='10 77.81 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130268152, end_position_on_the_genomic_accession=130280240, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11924, CurrentID=0, Status='live', Symbol='Neurog2', Aliases='Atoh4, Math4A, bHLHa8, ngn-2, ngn2', description='neurogenin 2', other_designations='atonal homolog 4|helix-loop-helix protein mATH-4A|protein atonal homolog 4', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127633145, end_position_on_the_genomic_accession=127635631, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  11925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11925, CurrentID=0, Status='live', Symbol='Neurog3', Aliases='Atoh5, Math4B, bHLHa7, ngn3', description='neurogenin 3', other_designations='NGN-3|Neurogenin 3 (Atonal protein homolog 5) (Helix-loop-helix protein mATH-4B) (MATH4B)|atonal homolog 5|helix-loop-helix protein mATH-4B|protein atonal homolog 5', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62133090, end_position_on_the_genomic_accession=62134763, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  11927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11927, CurrentID=0, Status='live', Symbol='Atox1', Aliases='AI256639, Atx1', description='antioxidant 1 copper chaperone', other_designations='ATX1 (antioxidant protein 1) homolog 1|metal transport protein ATX1', map_location='11 33.07 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55446643, end_position_on_the_genomic_accession=55461138, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11928, CurrentID=0, Status='live', Symbol='Atp1a1', Aliases='Atpa-1, BC010319', description='ATPase, Na+/K+ transporting, alpha 1 polypeptide', other_designations='Na(+)/K(+) ATPase alpha-1 subunit|sodium pump subunit alpha-1', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=101576219, end_position_on_the_genomic_accession=101604707, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  11931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11931, CurrentID=0, Status='live', Symbol='Atp1b1', Aliases='Atp4b, Atpb, Atpb-1, NKbeta1', description='ATPase, Na+/K+ transporting, beta 1 polypeptide', other_designations='sodium/potassium ATPase beta subunit|sodium/potassium-dependent ATPase subunit beta-1', map_location='1 71.75 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164437099, end_position_on_the_genomic_accession=164458355, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11932, CurrentID=0, Status='live', Symbol='Atp1b2', Aliases='Amog, Atpb-2', description='ATPase, Na+/K+ transporting, beta 2 polypeptide', other_designations='adhesion molecule in glia|glial cell adhesion molecule|snoRNA MBI-85|sodium/potassium-dependent ATPase subunit beta-2', map_location='11 42.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69599750, end_position_on_the_genomic_accession=69605960, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11933, CurrentID=0, Status='live', Symbol='Atp1b3', Aliases='AA409958, AI664000, AW212096', description='ATPase, Na+/K+ transporting, beta 3 polypeptide', other_designations='ATPB-3|ATPase, Na+/K+ beta 3 polypeptide|Na K-ATPase beta-3 subunit|sodium/potassium-dependent ATPase subunit beta-3', map_location='9 50.31 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96332662, end_position_on_the_genomic_accession=96364905, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11936, CurrentID=0, Status='live', Symbol='Fxyd2', Aliases='Atp1g1', description='FXYD domain-containing ion transport regulator 2', other_designations='ATPase, Na+/K+ transporting, gamma 1 polypeptide|Na(+)/K(+) ATPase subunit gamma|sodium pump gamma chain|sodium/potassium-transporting ATPase gamma chain', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45399709, end_position_on_the_genomic_accession=45410278, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  11937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11937, CurrentID=0, Status='live', Symbol='Atp2a1', Aliases='SERCA1', description='ATPase, Ca++ transporting, cardiac muscle, fast twitch 1', other_designations='SR Ca(2+)-ATPase 1|calcium pump 1|calcium-transporting ATPase sarcoplasmic reticulum type, fast twitch skeletal muscle isoform|endoplasmic reticulum class 1/2 Ca(2+) ATPase', map_location='7 69.04 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126445860, end_position_on_the_genomic_accession=126463103, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  11938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11938, CurrentID=0, Status='live', Symbol='Atp2a2', Aliases='9530097L16Rik, D5Wsu150e, SERCA2, SERCA2B, Serca2a, mKIAA4195', description='ATPase, Ca++ transporting, cardiac muscle, slow twitch 2', other_designations='SERCA2a isoform|SR Ca(2+)-ATPase 2|calcium pump 2|calcium-transporting ATPase sarcoplasmic reticulum type, slow twitch skeletal muscle isoform|endoplasmic reticulum class 1/2 Ca(2+) ATPase|sarco/endoplasmic reticulum Ca2+-ATPase 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122453513, end_position_on_the_genomic_accession=122502225, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  11941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11941, CurrentID=0, Status='live', Symbol='Atp2b2', Aliases='D6Abb2e, Gena300, PMCA2, Tmy, dfw, jog, wms, wri', description='ATPase, Ca++ transporting, plasma membrane 2', other_designations='deaf waddler|plasma membrane calcium ATPase|plasma membrane calcium pump|wriggle', map_location='6 52.85 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113743842, end_position_on_the_genomic_accession=114042613, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  11944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11944, CurrentID=0, Status='live', Symbol='Atp4a', Aliases='', description='ATPase, H+/K+ exchanging, gastric, alpha polypeptide', other_designations='ATPase, H+/K+ exchanging, alpha polypeptide|ATPase, H+/K+ transporting, alpha polypeptide|H+/K+-ATPase alpha|H+K+-transporting alpha 1|gastric H(+)/K(+) ATPase subunit alpha|proton pump', map_location='7 19.2 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30712209, end_position_on_the_genomic_accession=30725534, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  11945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11945, CurrentID=0, Status='live', Symbol='Atp4b', Aliases='AV080843', description='ATPase, H+/K+ exchanging, beta polypeptide', other_designations='ATPase, H+/K+ transporting, beta polypeptide, gastric specific|H+,K+-ATPase|H+/K+-ATPase beta|H,K-ATPase-Beta|gastric H(+)/K(+) ATPase subunit beta|proton pump beta chain', map_location='8 5.89 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13386209, end_position_on_the_genomic_accession=13396778, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  11946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11946, CurrentID=0, Status='live', Symbol='Atp5a1', Aliases='AI035633, AL022851, AL023067, Atpm, D18Ertd206e, Mom2', description='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit 1', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, alpha subunit|modifier of Min 2', map_location='18 52.38 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77773768, end_position_on_the_genomic_accession=77782869, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11947, CurrentID=0, Status='live', Symbol='Atp5b', Aliases='', description='ATP synthase, H+ transporting mitochondrial F1 complex, beta subunit', other_designations='ATP synthase, H+ transporting mitochondrial F1 complex, alpha subunit|mitochondrial ATP synthase, H+ transporting F1 complex beta subunit', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128083307, end_position_on_the_genomic_accession=128090388, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11949, CurrentID=0, Status='live', Symbol='Atp5c1', Aliases='1700094F02Rik', description='ATP synthase, H+ transporting, mitochondrial F1 complex, gamma polypeptide 1', other_designations='ATP synthase gamma-subunit gene encoding a mitochondrial protein|F-ATPase gamma subunit|F1 gamma', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10056030, end_position_on_the_genomic_accession=10080510, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11950, CurrentID=0, Status='live', Symbol='Atp5f1', Aliases='C76477', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit B1', other_designations='ATP synthase subunit b, mitochondrial|ATPase subunit b', map_location='3 46.46 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105942698, end_position_on_the_genomic_accession=105960248, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11951, CurrentID=0, Status='live', Symbol='Atp5g1', Aliases='', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C1 (subunit 9)', other_designations='ATP synthase lipid-binding protein, mitochondrial|ATP synthase proteolipid P1|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9)|ATPase protein 9|ATPase subunit c', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96072793, end_position_on_the_genomic_accession=96075694, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  11957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11957, CurrentID=0, Status='live', Symbol='Atp5j', Aliases='CF6', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F', other_designations='ATP synthase coupling factor VI|ATPase subunit F6', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=84827866, end_position_on_the_genomic_accession=84835648, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  11958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11958, CurrentID=0, Status='live', Symbol='Atp5k', Aliases='2610008D24Rik, Atp5i, Lfm1', description='ATP synthase, H+ transporting, mitochondrial F1F0 complex, subunit E', other_designations='ATPase subunit e', map_location='5 53.11 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108433253, end_position_on_the_genomic_accession=108434378, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11964, CurrentID=0, Status='live', Symbol='Atp6v1a', Aliases='AI647066, Atp6a1, Atp6a21, VA68, VPP2, Atp6v1a', description='ATPase, H+ transporting, lysosomal V1 subunit A', other_designations='70-kDa subunit|ATPase, H+ transporting, V1 subunit A|ATPase, H+ transporting, V1 subunit A1|ATPase, H+ transporting, lysosomal (vacuolar proton pump), alpha 70 kDa|ATPase, H+ transporting, lysosomal 70kD, V1 subunit A|V-ATPase 69 kDa subunit|V-ATPase subunit A|lysosomal 70kDa|vacuolar proton pump subunit alpha', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44085402, end_position_on_the_genomic_accession=44139700, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  11966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11966, CurrentID=0, Status='live', Symbol='Atp6v1b2', Aliases='AI194269, AI790362, Atp6b2, HO57, R74844', description='ATPase, H+ transporting, lysosomal V1 subunit B2', other_designations='ATPase, H+ transporting, V1 subunit B|ATPase, H+ transporting, lysosomal (vacuolar proton pump), beta 56/58 kDa|ATPase, H+ transporting, lysosomal 56/58kD, V1 subunit B|V-ATPase subunit B 2|endomembrane proton pump 58 kDa subunit|lysosomal 56/58kDa|vacuolar proton pump subunit B 2', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69088736, end_position_on_the_genomic_accession=69113718, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  11972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11972, CurrentID=0, Status='live', Symbol='Atp6v0d1', Aliases='AI267038, Ac39, Atp6d, P39, VATX, Vma6', description='ATPase, H+ transporting, lysosomal V0 subunit D1', other_designations='ATPase, H+ transporting, V0 subunit D|ATPase, H+ transporting, lysosomal (vacuolar proton pump), 42 kDa|ATPase, H+ transporting, lysosomal 38kDa, V0 subunit D|V-ATPase 40 kDa accessory protein|V-ATPase AC39 subunit|V-ATPase subunit d 1|lysosomal 38kDa|physophilin|vacuolar proton pump subunit d 1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105524470, end_position_on_the_genomic_accession=105566040, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  11973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11973, CurrentID=0, Status='live', Symbol='Atp6v1e1', Aliases='2410029D23Rik, Atp6e, Atp6e2, Atp6v1e, D6Ertd385e, E2, P31, Vma4', description='ATPase, H+ transporting, lysosomal V1 subunit E1', other_designations='ATPase, H+ transporting lysosomal (vacuolar proton pump), 32 kDa|ATPase, H+ transporting, V1 subunit E|ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E|H(+)-ATPase E-like protein|H+ ATPase subunit E|V-ATPase 31 kDa subunit|V-ATPase subunit E 1|VATPase, H+ transporting, lysosomal V1 subunit E1|lysosomal 31kDa|vacuolar proton pump subunit E 1', map_location='6 57.01 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120795244, end_position_on_the_genomic_accession=120822685, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11974, CurrentID=0, Status='live', Symbol='Atp6v0e', Aliases='Atp6k1, M9.2, Atp6v0e', description='ATPase, H+ transporting, lysosomal V0 subunit E', other_designations='ATPase, H+ transporting lysosomal (vacuolar proton pump), 9.2 kDa|ATPase, H+ transporting, V0 subunit e|ATPase, H+ transporting, lysosomal 9kD V0 subunit E|V-ATPase subunit e 1|lysosomal 9kDa|vacuolar proton pump subunit e 1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26676369, end_position_on_the_genomic_accession=26699647, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11975, CurrentID=0, Status='live', Symbol='Atp6v0a1', Aliases='AA959968, ATP6a1, Atp6n1, Atp6n1a, Atpv0a1, Vpp-1, Vpp1', description='ATPase, H+ transporting, lysosomal V0 subunit A1', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) noncatalytic accessory protein 1 (110/160 kDa)|ATPase, H+ transporting, lysosomal (vacuolar proton pump) noncatalytic accessory protein 1A (110/160 kDa)|ATPase, H+ transporting, lysosomal V0 subunit a|V-ATPase 116 kDa|V-ATPase a1|V-type proton ATPase 116 kDa subunit a|clathrin-coated vesicle/synaptic vesicle proton pump 116 kDa subunit|vacuolar adenosine triphosphatase subunit Ac116|vacuolar proton pump subunit 1|vacuolar proton translocating ATPase 116 kDa subunit a', map_location='11 64.04 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101009447, end_position_on_the_genomic_accession=101063719, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  11977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11977, CurrentID=0, Status='live', Symbol='Atp7a', Aliases='MNK', description='ATPase, Cu++ transporting, alpha polypeptide', other_designations='Menkes protein|copper pump 1|menkes disease-associated protein homolog', map_location='X 47.36 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106027224, end_position_on_the_genomic_accession=106128160, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  11979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11979, CurrentID=0, Status='live', Symbol='Atp7b', Aliases='Atp7a, WND, tx', description='ATPase, Cu++ transporting, beta polypeptide', other_designations='Wilson protein|copper pump 2|toxic milk|wilson disease-associated protein homolog', map_location='8 10.78 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21992783, end_position_on_the_genomic_accession=22060074, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  11980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11980, CurrentID=0, Status='live', Symbol='Atp8a1', Aliases='AI481521, AI853962, APLT, AW743152, AW822227, Atp3a2, B230107D19Rik, ClassI', description='ATPase, aminophospholipid transporter (APLT), class I, type 8A, member 1', other_designations='ATPase 8A1, aminophospholipid transporter (APLT), class I|ATPase 8A1, p type|ATPase class I type 8A member 1|P4-ATPase flippase complex alpha subunit ATP8A1|chromaffin granule ATPase II|probable phospholipid-transporting ATPase IA', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=67618139, end_position_on_the_genomic_accession=67847487, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  11981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11981, CurrentID=0, Status='live', Symbol='Atp9a', Aliases='IIa', description='ATPase, class II, type 9A', other_designations='ATPase 9A, class II|ATPase 9A, p type|ATPase, class 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168634438, end_position_on_the_genomic_accession=168742136, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  11982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11982, CurrentID=0, Status='live', Symbol='Atp10a', Aliases='Atp10c, pfatp', description='ATPase, class V, type 10A', other_designations='ATPase 10A, class V|ATPase, class 5|P-locus fat-associated ATPase|P4-ATPase flippase complex alpha subunit ATP10A', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=58656143, end_position_on_the_genomic_accession=58830035, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  11983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11983, CurrentID=0, Status='live', Symbol='Atpif1', Aliases='Atpi, IF(1), If1', description='ATPase inhibitory factor 1', other_designations='inhibitor of F(1)F(o)-ATPase', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132530555, end_position_on_the_genomic_accession=132535409, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  11984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11984, CurrentID=0, Status='live', Symbol='Atp6v0c', Aliases='Atp6c, Atp6c2, Atp6l, Atpl, Atpl-rs1, PL16, VATL, Vma3', description='ATPase, H+ transporting, lysosomal V0 subunit C', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) 16 kDa|ATPase, H+ transporting, lysosomal 16kD, V0 subunit C|ATPase, H+ transporting, lysosomal 2|H(+)-ATPase (mvp)|V-ATPase 16 kDa proteolipid subunit|vacuolar proton pump 16 kDa proteolipid subunit', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24163866, end_position_on_the_genomic_accession=24169429, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  11987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11987, CurrentID=0, Status='live', Symbol='Slc7a1', Aliases='4831426K01Rik, AI447493, Atrc-1, Atrc1, Cat1, Rec-1, Rev-1, mCAT-1', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 1', other_designations='CAT-1|ERR|cationic amino acid transporter 1|ecotropic retroviral leukemia receptor|ecotropic retrovirus receptor|solute carrier family 7 member 1|system Y+ basic amino acid transporter', map_location='5 88.48 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=148327410, end_position_on_the_genomic_accession=148399904, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11988, CurrentID=0, Status='live', Symbol='Slc7a2', Aliases='20.5, AI158848, Atrc2, CAT-2, Cat2, Tea', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 2', other_designations='T-cell early activation protein|amino acid transporter, cationic 2 (low affinity)|cationic amino acid transporter 2A|cationic amino acid transporter 2B|low affinity cationic amino acid transporter 2|solute carrier family 7 member 2', map_location='8 23.89 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40862367, end_position_on_the_genomic_accession=40922070, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  11989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11989, CurrentID=0, Status='live', Symbol='Slc7a3', Aliases='Atrc3, CAT-3, CAT3, SLC7A1, SLC7A2', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 3', other_designations='amino acid transporter, cationic 3|cationic amino acid transporter y+|solute carrier family 7 member 3', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101079204, end_position_on_the_genomic_accession=101090452, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  11990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11990, CurrentID=0, Status='live', Symbol='Atrn', Aliases='AW558010, Mgca, mKIAA0548, mahogany, mg', description='attractin', other_designations='protein mahogany', map_location='2 63.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130906447, end_position_on_the_genomic_accession=131030333, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  11991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11991, CurrentID=0, Status='live', Symbol='Hnrnpd', Aliases='Auf1, Hnrpd', description='heterogeneous nuclear ribonucleoprotein D', other_designations='AU-rich element RNA-binding protein 1|hnRNP D0', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=99955935, end_position_on_the_genomic_accession=99979015, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  11992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11992, CurrentID=0, Status='live', Symbol='Auh', Aliases='C77140, W91705', description='AU RNA binding protein/enoyl-coenzyme A hydratase', other_designations='AU-binding enoyl-CoA hydratase|AU-specific RNA-binding enoyl-CoA hydratase|muAUH', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=52835102, end_position_on_the_genomic_accession=52929706, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  11993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11993, CurrentID=0, Status='live', Symbol='Aup1', Aliases='AA589454', description='ancient ubiquitous protein 1', other_designations='', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83054521, end_position_on_the_genomic_accession=83057682, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  11994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11994, CurrentID=0, Status='live', Symbol='Pcdh15', Aliases='BB078305, ENSMUSG00000046980, Gm9815, Ush1f, av, nmf19, roda', description='protocadherin 15', other_designations='Ames waltzer|protocadherin 15 CD2|protocadherin 15 CD3 isoform', map_location='10 37.43 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=73100028, end_position_on_the_genomic_accession=74649737, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  11997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11997, CurrentID=0, Status='live', Symbol='Akr1b7', Aliases='AR, Avdp, MVDP', description='aldo-keto reductase family 1, member B7', other_designations='aldehyde reductase|androgen regulated vas deferens protein|vas deferens androgen-dependent protein', map_location='6 14.95 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34412362, end_position_on_the_genomic_accession=34423137, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  11998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11998, CurrentID=0, Status='live', Symbol='Avp', Aliases='Vp, Vsp', description='arginine vasopressin', other_designations='AVP-NPII|pre-pro-arginine-vasopressin-neurophysin II|prepro-AVP-NP II', map_location='2 63.24 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130580677, end_position_on_the_genomic_accession=130582541, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12000, CurrentID=0, Status='live', Symbol='Avpr2', Aliases='ADHR, DI1, DIR, ND1, V2R, VPV2R', description='arginine vasopressin receptor 2', other_designations='antidiuretic hormone receptor|nephrogenic diabetes insipidus|renal-type arginine vasopressin receptor', map_location='X 37.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73890591, end_position_on_the_genomic_accession=73894428, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12005, CurrentID=0, Status='live', Symbol='Axin1', Aliases='AI316800, Axin, Fu, Kb, Ki, fused, kinky, knobbly', description='axin 1', other_designations='axis inhibition protein 1|protein Fused', map_location='17 13.07 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26130972, end_position_on_the_genomic_accession=26195811, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12006, CurrentID=0, Status='live', Symbol='Axin2', Aliases='Axi1, Axil, Conductin', description='axin2', other_designations='axin-like protein|axis inhibition protein 2', map_location='11 71.83 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=108917828, end_position_on_the_genomic_accession=108950783, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12007, CurrentID=0, Status='live', Symbol='Azgp1', Aliases='Zag', description='alpha-2-glycoprotein 1, zinc', other_designations='major histocompatibility complex class I-like protein|zn-alpha-2-GP', map_location='5 76.92 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137981521, end_position_on_the_genomic_accession=137990233, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12009, CurrentID=0, Status='live', Symbol='Cep131', Aliases='AZ1, Azi, Azi1', description='centrosomal protein 131', other_designations='5-azacytidine induced gene 1|5-azacytidine-induced protein 1|pre-acrosome localization protein 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120064429, end_position_on_the_genomic_accession=120086859, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  12010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12010, CurrentID=0, Status='live', Symbol='B2m', Aliases='Ly-m11, beta2-m, beta2m', description='beta-2 microglobulin', other_designations='', map_location='2 F1-F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122147687, end_position_on_the_genomic_accession=122153082, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12012, CurrentID=0, Status='live', Symbol='Baat', Aliases='AI118337, AI158864, BAT', description='bile acid-Coenzyme A: amino acid N-acyltransferase', other_designations='BACAT|bile acid-Coenzyme A dehydrogenase: amino acid n-acyltransferase|glycine N-choloyltransferase|long-chain fatty-acyl-CoA hydrolase|taurine N-acyltransferase', map_location='4 26.51 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49489416, end_position_on_the_genomic_accession=49507912, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  667373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667373, CurrentID=0, Status='live', Symbol='Ifit1bl1', Aliases='Gm14446, Ifit1c, OTTMUSG00000016644', description='interferon induced protein with tetratricpeptide repeats 1B like 1', other_designations='', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34592888, end_position_on_the_genomic_accession=34663000, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12014, CurrentID=0, Status='live', Symbol='Bach2', Aliases='E030004N02Rik', description='BTB and CNC homology, basic leucine zipper transcription factor 2', other_designations='BTB and CNC homolog 2|BTB and CNC homology 2|Transcription regulator protein BACH2 (BTB and CNC homolog 2)', map_location='4 A4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=32238574, end_position_on_the_genomic_accession=32586108, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12015, CurrentID=0, Status='live', Symbol='Bad', Aliases='AI325008, Bbc2', description='BCL2-associated agonist of cell death', other_designations='bcl-2-binding component 6|bcl-xL/Bcl-2-associated death promoter|bcl2 antagonist of cell death', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6941861, end_position_on_the_genomic_accession=6951905, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12017, CurrentID=0, Status='live', Symbol='Bag1', Aliases='BAG-1, Rap46', description='BCL2-associated athanogene 1', other_designations='Bcl-2-binding protein|bcl-2-associated athanogene 1', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40936398, end_position_on_the_genomic_accession=40948294, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12018, CurrentID=0, Status='live', Symbol='Bak1', Aliases='Bak, N-BAK1, N-Bak', description='BCL2-antagonist/killer 1', other_designations='Bcl2 homologous antagonist/killer|apoptosis regulator BAK', map_location='17 B', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27019812, end_position_on_the_genomic_accession=27028839, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12020, CurrentID=0, Status='live', Symbol='Nkx3-2', Aliases='Bapx1, NKX3.2, Nkx-3.2', description='NK3 homeobox 2', other_designations='bagpipe homeobox gene 1 homolog|bagpipe homeobox protein homolog 1|homeobox protein NK-3 homolog B|homeodomain protein Nkx-3.2', map_location='5 22.58 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=41761483, end_position_on_the_genomic_accession=41764220, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12021, CurrentID=0, Status='live', Symbol='Bard1', Aliases='ENSMUSG00000060893, ENSMUSG00000073653', description='BRCA1 associated RING domain 1', other_designations='BARD-1', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=71027531, end_position_on_the_genomic_accession=71103233, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12022, CurrentID=0, Status='live', Symbol='Barx1', Aliases='', description='BarH-like homeobox 1', other_designations='', map_location='13 24.9 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48663036, end_position_on_the_genomic_accession=48666507, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12023, CurrentID=0, Status='live', Symbol='Barx2', Aliases='2310006E12Rikb, Barx2', description='BarH-like homeobox 2', other_designations='homeodomain transcription factor BARX2', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=31846044, end_position_on_the_genomic_accession=31913285, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12028, CurrentID=0, Status='live', Symbol='Bax', Aliases='', description='BCL2-associated X protein', other_designations='', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45461695, end_position_on_the_genomic_accession=45466903, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12029, CurrentID=0, Status='live', Symbol='Bcl6b', Aliases='Bazf', description='B cell CLL/lymphoma 6, member B', other_designations='B-cell CLL/lymphoma 6, member B|BcL6-associated zinc finger protein', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70224127, end_position_on_the_genomic_accession=70229798, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12032, CurrentID=0, Status='live', Symbol='Bcan', Aliases='Cspg7', description='brevican', other_designations='', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87987531, end_position_on_the_genomic_accession=88000893, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12033, CurrentID=0, Status='live', Symbol='Bcap29', Aliases='AW208404, Bap29', description='B cell receptor associated protein 29', other_designations='BCR-associated protein 29', map_location='12 13.63 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31595354, end_position_on_the_genomic_accession=31634658, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12034, CurrentID=0, Status='live', Symbol='Phb2', Aliases='AU044498, BAP, Bap37, Bcap37, REA', description='prohibitin 2', other_designations='B-cell receptor-associated protein 37|B-cell receptor-associated protein BAP37|phb2|repressor of estrogen receptor activity', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124712289, end_position_on_the_genomic_accession=124716945, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12035, CurrentID=0, Status='live', Symbol='Bcat1', Aliases='BCATc, Eca39', description='branched chain aminotransferase 1, cytosolic', other_designations='BCAT(c)', map_location='6 77.27 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=144993835, end_position_on_the_genomic_accession=145076157, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12036, CurrentID=0, Status='live', Symbol='Bcat2', Aliases='Bcat-2, Eca40', description='branched chain aminotransferase 2, mitochondrial', other_designations='BCAT(m)', map_location='7 29.36 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45569855, end_position_on_the_genomic_accession=45589710, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12038, CurrentID=0, Status='live', Symbol='Bche', Aliases='C730038G20Rik', description='butyrylcholinesterase', other_designations='acylcholine acylhydrolase|butyrylcholine esterase|choline esterase II|pseudocholinesterase', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=73635805, end_position_on_the_genomic_accession=73708437, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12039, CurrentID=0, Status='live', Symbol='Bckdha', Aliases='', description='branched chain ketoacid dehydrogenase E1, alpha polypeptide', other_designations='BCKAD E1[a]|BCKDE1A|BCKDH E1-alpha|branched-chain alpha-keto acid dehydrogenase E1 component alpha chain', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25629852, end_position_on_the_genomic_accession=25658784, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12040, CurrentID=0, Status='live', Symbol='Bckdhb', Aliases='BCKDE1B, BCKDH E1-beta', description='branched chain ketoacid dehydrogenase E1, beta polypeptide', other_designations='branched-chain alpha-keto acid dehydrogenase E1 component beta chain', map_location='9 E2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=83925111, end_position_on_the_genomic_accession=84124240, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12041, CurrentID=0, Status='live', Symbol='Bckdk', Aliases='AI327402', description='branched chain ketoacid dehydrogenase kinase', other_designations='BCKD-kinase|BCKDHKIN|branched-chain alpha-ketoacid dehydrogenase kinase', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127904073, end_position_on_the_genomic_accession=127909664, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12042, CurrentID=0, Status='live', Symbol='Bcl10', Aliases='AI132454, BCL-10, C81403, CARMEN, CIPER, CLAP, ME10, cE10', description='B cell leukemia/lymphoma 10', other_designations='B-cell CLL/lymphoma 10|B-cell leukemia/lymphoma 10|CARD-containing molecule enhancing NF-kappa-B|CARD-like apoptotic protein|CED-3/ICH-1 prodomain homologous E10-like regulator|cellular homolog of vCARMEN|cellular-E10|mammalian CARD-containing adapter molecule E10', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145924262, end_position_on_the_genomic_accession=145934366, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12043, CurrentID=0, Status='live', Symbol='Bcl2', Aliases='AW986256, Bcl-2, C430015F12Rik, D630044D05Rik, D830018M01Rik', description='B cell leukemia/lymphoma 2', other_designations='B-cell leukemia/lymphoma 2', map_location='1 49.76 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=106538178, end_position_on_the_genomic_accession=106714290, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12044, CurrentID=0, Status='live', Symbol='Bcl2a1a', Aliases='A1, BB218357, Bcl2a1, Bfl-1, Hbpa1', description='B cell leukemia/lymphoma 2 related protein A1a', other_designations='A1-A|B-cell leukemia/lymphoma 2 related protein A1a|hemopoietic-specific early response protein|protein BFL-1', map_location='9 47.24 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88956920, end_position_on_the_genomic_accession=88962416, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12045, CurrentID=0, Status='live', Symbol='Bcl2a1b', Aliases='A1-b', description='B cell leukemia/lymphoma 2 related protein A1b', other_designations='anti-apoptotic protein A1-b', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89199273, end_position_on_the_genomic_accession=89207838, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12047, CurrentID=0, Status='live', Symbol='Bcl2a1d', Aliases='A1-d', description='B cell leukemia/lymphoma 2 related protein A1d', other_designations='B-cell leukemia/lymphoma 2 related protein A1d', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88723288, end_position_on_the_genomic_accession=88731850, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12048, CurrentID=0, Status='live', Symbol='Bcl2l1', Aliases='Bcl(X)L, Bcl-XL, Bcl2l, BclX, bcl-x, bcl2-L-1', description='BCL2-like 1', other_designations='B-cell leukemia/lymphoma x|anti-apoptosis regulatory protein|apoptosis regulator Bcl-X', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152758732, end_position_on_the_genomic_accession=152832581, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12049, CurrentID=0, Status='live', Symbol='Bcl2l10', Aliases='AA420380, AU023065, Boo, C85687, Diva', description='Bcl2-like 10', other_designations='anti-apoptotic protein Boo|apoptosis regulator Bcl-B|bcl-2 homolog Diva|bcl2-L-10', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75347758, end_position_on_the_genomic_accession=75351640, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12050, CurrentID=0, Status='live', Symbol='Bcl2l2', Aliases='AW048834, Bcl-w, Gtrgal2, Gtrosa41, bclw, c98', description='BCL2-like 2', other_designations='apoptosis regulator Bcl-W|bcl2-L-2', map_location='14 27.98 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54883366, end_position_on_the_genomic_accession=54889095, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12051, CurrentID=0, Status='live', Symbol='Bcl3', Aliases='AI528691, Bcl-3', description='B cell leukemia/lymphoma 3', other_designations='B-cell CLL/lymphoma 3|B-cell leukemia/lymphoma 3', map_location='7 9.95 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19808462, end_position_on_the_genomic_accession=19822811, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12053, CurrentID=0, Status='live', Symbol='Bcl6', Aliases='Bcl5', description='B cell leukemia/lymphoma 6', other_designations='B-cell leukemia/lymphoma 5|B-cell leukemia/lymphoma 6', map_location='16 15.26 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23965052, end_position_on_the_genomic_accession=23988612, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12054, CurrentID=0, Status='live', Symbol='Bcl7b', Aliases='', description='B cell CLL/lymphoma 7B', other_designations='B-cell CLL/lymphoma 7B', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135168372, end_position_on_the_genomic_accession=135181852, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12055, CurrentID=0, Status='live', Symbol='Bcl7c', Aliases='C230096E12Rik', description='B cell CLL/lymphoma 7C', other_designations='B-cell CLL/lymphoma 7C|B-cell chronic lymphocytic leukemia/lymphoma 7C protein', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127661454, end_position_on_the_genomic_accession=127708780, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12057, CurrentID=0, Status='live', Symbol='Opn1sw', Aliases='AW551857, Bcp', description='opsin 1 (cone pigments), short-wave-sensitive (color blindness, tritan)', other_designations='BOP|S opsin|SWS opsin|UV cone pigment|blue cone opsin|blue cone photoreceptor pigment|blue cone pigment|blue opsin|blue visual pigment protein|blue-sensitive opsin|blue/UV opsin|short wavelength sensitive opsin|short wavelength-sensitive cone opsin', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29376671, end_position_on_the_genomic_accession=29388729, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12061, CurrentID=0, Status='live', Symbol='Bdkrb1', Aliases='B1BKR, B1R, BKR1, BRADYB1, Bdkrb', description='bradykinin receptor, beta 1', other_designations='BK-1 receptor|kinin B1', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105604091, end_position_on_the_genomic_accession=105605428, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12062, CurrentID=0, Status='live', Symbol='Bdkrb2', Aliases='B(2), B2, B2R, BK-2, BK2, BK2R, BKR2, BRB2', description='bradykinin receptor, beta 2', other_designations='BK-2 receptor|kinin B2', map_location='12 55.76 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105563172, end_position_on_the_genomic_accession=105595244, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12064, CurrentID=0, Status='live', Symbol='Bdnf', Aliases='', description='brain derived neurotrophic factor', other_designations='', map_location='2 56.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=109674700, end_position_on_the_genomic_accession=109727043, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12067, CurrentID=0, Status='live', Symbol='Ube2n-ps1', Aliases='1500026J17Rik, Benh-ps1', description='ubiquitin-conjugating enzyme E2N, pseudogene 1', other_designations='bendless homolog, pseudogene 1', map_location='5 40.68 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76231885, end_position_on_the_genomic_accession=76233097, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  12068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12068, CurrentID=0, Status='live', Symbol='Bet1', Aliases='AW555236, Bet-1', description='Bet1 golgi vesicular membrane trafficking protein', other_designations='blocked early in transport 1 homolog|golgi vesicular membrane trafficking protein p18|golgi vesicular membrane-trafficking protein p18|mBET1', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=4076904, end_position_on_the_genomic_accession=4086995, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12069, CurrentID=0, Status='live', Symbol='Bex2', Aliases='AL024066, Bex1', description='brain expressed X-linked 2', other_designations='brain-expressed X-linked protein 2 homolog', map_location='X 57.37 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136066565, end_position_on_the_genomic_accession=136068236, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12070, CurrentID=0, Status='live', Symbol='Ngfrap1', Aliases='AL033356, Bex3, DXWsu67e, Gcap27, Nade', description='nerve growth factor receptor (TNFRSF16) associated protein 1', other_designations='brain expressed X-linked 3|brain-expressed X-linked protein 3 homolog|granule cell antiserum positive 27|nerve growth factor receptor-associated protein 1|p75NTR-associated Cell Death Executor', map_location='X 57.64 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136270253, end_position_on_the_genomic_accession=136271978, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12075, CurrentID=0, Status='live', Symbol='Bfsp1', Aliases='CP95', description='beaded filament structural protein 1, in lens-CP94', other_designations='beaded filament structural protein in lens-CP94|lens fiber cell beaded-filament structural protein CP 95', map_location='2 G', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=143826528, end_position_on_the_genomic_accession=143915274, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67549, CurrentID=0, Status='live', Symbol='Gpr89', Aliases='4933412D19Rika, SH120, Gpr89', description='G protein-coupled receptor 89', other_designations='protein GPR89', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96871066, end_position_on_the_genomic_accession=96905343, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12091, CurrentID=0, Status='live', Symbol='Glb1', Aliases='AW125515, Bge, Bgl, Bgl-e, Bgl-s, Bgl-t, Bgs, Bgt, C130097A14Rik', description='galactosidase, beta 1', other_designations='acid beta-galactosidase|lactase', map_location='9 64.4 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114401078, end_position_on_the_genomic_accession=114474379, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  12095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12095, CurrentID=0, Status='live', Symbol='Bglap3', Aliases='AI461847, Bglap-rs1, OC-X, ORG, mOC-X', description='bone gamma-carboxyglutamate protein 3', other_designations='bone gamma-carboxyglutamate protein, related sequence 1|gamma-carboxyglutamic acid-containing protein 3|nephrocalcin', map_location='3 38.81 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88368615, end_position_on_the_genomic_accession=88372749, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12096, CurrentID=0, Status='live', Symbol='Bglap', Aliases='BGP1, OC, OG1, mOC-A, Bglap', description='bone gamma carboxyglutamate protein', other_designations='bone Gla protein|bone gamma carboxyglutamate protein 1|gamma-carboxyglutamic acid-containing protein', map_location='3 38.82 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88383495, end_position_on_the_genomic_accession=88384466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12097, CurrentID=0, Status='live', Symbol='Bglap2', Aliases='BGP2, Bglap1, Bgp, Og2, mOC-B', description='bone gamma-carboxyglutamate protein 2', other_designations='bone Gla protein 2|gamma-carboxyglutamic acid-containing protein 2|osteocalcin', map_location='3 38.81 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88377736, end_position_on_the_genomic_accession=88378699, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12111, CurrentID=0, Status='live', Symbol='Bgn', Aliases='BG, DSPG1, PG-S1, PGI, SLRR1A', description='biglycan', other_designations='bone/cartilage proteoglycan I', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73483612, end_position_on_the_genomic_accession=73495936, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12116, CurrentID=0, Status='live', Symbol='Bhmt', Aliases='', description='betaine-homocysteine methyltransferase', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93616891, end_position_on_the_genomic_accession=93637758, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12121, CurrentID=0, Status='live', Symbol='Bicd1', Aliases='B830009D06Rik', description='bicaudal D homolog 1 (Drosophila)', other_designations='bic-D 1', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=149408819, end_position_on_the_genomic_accession=149563329, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12122, CurrentID=0, Status='live', Symbol='Bid', Aliases='2700049M22Rik, AI875481, AU022477', description='BH3 interacting domain death agonist', other_designations='p22 BID', map_location='6 57.02 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120893119, end_position_on_the_genomic_accession=120917142, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12123, CurrentID=0, Status='live', Symbol='Hrk', Aliases='AI838259, Bid3, DP5, harakiri', description='harakiri, BCL2 interacting protein (contains only BH3 domain)', other_designations='BH3 interacting (with BCL2 family) domain, apoptosis agonist|BH3-interacting domain-containing protein 3|neuronal death protein DP5', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=118169764, end_position_on_the_genomic_accession=118189478, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12124, CurrentID=0, Status='live', Symbol='Bik', Aliases='Biklk, Blk, Nbk', description='BCL2-interacting killer', other_designations='apoptosis inducer NBK|bik-like killer protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83526792, end_position_on_the_genomic_accession=83544635, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12125, CurrentID=0, Status='live', Symbol='Bcl2l11', Aliases='1500006F24Rik, Bim, Bod, bcl2-L-11', description='BCL2-like 11 (apoptosis facilitator)', other_designations='Bcl2 interacting mediator of cell death|bcl2-interacting mediator of cell death', map_location='2 F3-G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128126038, end_position_on_the_genomic_accession=128162547, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12131, CurrentID=0, Status='live', Symbol='Bkm1', Aliases='Bkm-1', description='banded krait minor satellite DNA 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12140, CurrentID=0, Status='live', Symbol='Fabp7', Aliases='B-FABP, BFABP, Blbp, MRG', description='fatty acid binding protein 7, brain', other_designations='brain lipid binding protein|brain lipid-binding protein|brain-type fatty acid-binding protein|fatty acid-binding protein 7|mammary-derived growth inhibitor-related', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=57784923, end_position_on_the_genomic_accession=57788450, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12142, CurrentID=0, Status='live', Symbol='Prdm1', Aliases='Blimp-1, Blimp1, PRDI-BF1, ZNFPR1A1, b2b1765Clo', description='PR domain containing 1, with ZNF domain', other_designations='B lymphocyte induced maturation protein|B lymphocyte-induced maturation protein 1|PR domain containing 1 with ZNF domain|PR domain-containing protein 1|beta-interferon gene positive regulatory domain I-binding factor', map_location='10 23.24 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=44437174, end_position_on_the_genomic_accession=44528560, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12143, CurrentID=0, Status='live', Symbol='Blk', Aliases='', description='B lymphoid kinase', other_designations='b lymphocyte kinase|p55-Blk', map_location='14 33.25 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63372831, end_position_on_the_genomic_accession=63417187, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12144, CurrentID=0, Status='live', Symbol='Blm', Aliases='', description='Bloom syndrome, RecQ helicase-like', other_designations='Bloom syndrome homolog|mBLM|recQ helicase homolog', map_location='7 45.65 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80454993, end_position_on_the_genomic_accession=80535119, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  12145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12145, CurrentID=0, Status='live', Symbol='Cxcr5', Aliases='Blr1, CXC-R5, CXCR-5, Gpcr6, MDR15', description='chemokine (C-X-C motif) receptor 5', other_designations='burkitt lymphoma receptor 1 homolog|chemochine (C-X-C motif) receptor 5', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44511534, end_position_on_the_genomic_accession=44561874, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12151, CurrentID=0, Status='live', Symbol='Bmi1', Aliases='AW546694, Bmi-1, Pcgf4', description='Bmi1 polycomb ring finger oncogene', other_designations='B lymphoma Mo-MLV insertion region 1|polycomb group RING finger protein 4|polycomb group ring finger 4', map_location='2 12.9 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=18677018, end_position_on_the_genomic_accession=18686629, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12153, CurrentID=0, Status='live', Symbol='Bmp1', Aliases='Pcp, Tld', description='bone morphogenetic protein 1', other_designations='mammalian tolloid protein|procollagen C-proteinase', map_location='14 36.32 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70474555, end_position_on_the_genomic_accession=70520712, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  12154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12154, CurrentID=0, Status='live', Symbol='Bmp10', Aliases='', description='bone morphogenetic protein 10', other_designations='BMP-10', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87429002, end_position_on_the_genomic_accession=87434512, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12155, CurrentID=0, Status='live', Symbol='Bmp15', Aliases='AU015375, AU018861, AU021453, Bmp-15, C86824, C87336, GDF-9B', description='bone morphogenetic protein 15', other_designations='growth/differentiation factor 9B', map_location='X 2.81 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=6314106, end_position_on_the_genomic_accession=6320723, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12156, CurrentID=0, Status='live', Symbol='Bmp2', Aliases='AI467020a, Bmp2', description='bone morphogenetic protein 2', other_designations='bone morphogenetic protein 2A', map_location='2 65.21 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=133552159, end_position_on_the_genomic_accession=133562896, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12159, CurrentID=0, Status='live', Symbol='Bmp4', Aliases='Bmp-4, Bmp2b, Bmp2b-1, Bmp2b1', description='bone morphogenetic protein 4', other_designations='bone morphogenetic protein 2B', map_location='14 23.95 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=46383520, end_position_on_the_genomic_accession=46390599, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12160, CurrentID=0, Status='live', Symbol='Bmp5', Aliases='AU023399, se', description='bone morphogenetic protein 5', other_designations='', map_location='9 42.34 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75775365, end_position_on_the_genomic_accession=75899017, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12161, CurrentID=0, Status='live', Symbol='Bmp6', Aliases='D13Wsu115e, Vgr1', description='bone morphogenetic protein 6', other_designations='VG-1-related protein', map_location='13 18.15 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38345716, end_position_on_the_genomic_accession=38499728, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12162, CurrentID=0, Status='live', Symbol='Bmp7', Aliases='OP1', description='bone morphogenetic protein 7', other_designations='osteogenic protein 1', map_location='2 95.54 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172868012, end_position_on_the_genomic_accession=172940321, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12163, CurrentID=0, Status='live', Symbol='Bmp8a', Aliases='Bmp7r1, OP-2, OP2', description='bone morphogenetic protein 8a', other_designations='osteogenic protein 2', map_location='4 57.42 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123312645, end_position_on_the_genomic_accession=123343252, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12164, CurrentID=0, Status='live', Symbol='Bmp8b', Aliases='Op3', description='bone morphogenetic protein 8b', other_designations='', map_location='4 57.38 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123104307, end_position_on_the_genomic_accession=123126094, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12165, CurrentID=0, Status='live', Symbol='Gdf2', Aliases='Bmp9', description='growth differentiation factor 2', other_designations='BMP-9|GDF-2|bone morphogenetic protein 9', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33941039, end_position_on_the_genomic_accession=33947198, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12166, CurrentID=0, Status='live', Symbol='Bmpr1a', Aliases='1110037I22Rik, ALK3, AU045487, BMPR-IA, Bmpr', description='bone morphogenetic protein receptor, type 1A', other_designations='ALK-3|BMP type-1A receptor|BMP-2/BMP-4 receptor|BMPR-1A|SKR5|activin receptor-like kinase 3|serine/threonine-protein kinase receptor R5', map_location='14 20.81 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34411061, end_position_on_the_genomic_accession=34503532, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12167, CurrentID=0, Status='live', Symbol='Bmpr1b', Aliases='AI385617, ALK-6, AV355320, Acvrlk6, Alk6, BMPR-1B, BMPR-IB, CFK-43a, SKR6', description='bone morphogenetic protein receptor, type 1B', other_designations='BMP type-1B receptor|activin receptor-like kinase 6|serine/threonine-protein kinase receptor R6', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=141837136, end_position_on_the_genomic_accession=142169590, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  12168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12168, CurrentID=0, Status='live', Symbol='Bmpr2', Aliases='2610024H22Rik, AL117858, AW546137, BB189135, BMP-2, BMPR-2, BMPR-II, BMPRII, BRK-3, Gm20272', description='bone morphogenetic protein receptor, type II (serine/threonine kinase)', other_designations='BMP type II receptor|BMP type-2 receptor|bone morphogenic protein receptor, type II (serine/threonine kinase)', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59764270, end_position_on_the_genomic_accession=59878081, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12169, CurrentID=0, Status='live', Symbol='Bmx', Aliases='Etk, Etk/Bmx, Tyro8', description='BMX non-receptor tyrosine kinase', other_designations='TYRO8 protein tyrosine kinase 8|bone marrow tyrosine kinase gene in chromosome X protein homolog', map_location='X F', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164192842, end_position_on_the_genomic_accession=164258193, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  12173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12173, CurrentID=0, Status='live', Symbol='Bnc1', Aliases='AI047752, AW546376, Bnc', description='basonuclin 1', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81966657, end_position_on_the_genomic_accession=81992299, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12175, CurrentID=0, Status='live', Symbol='Bnip2', Aliases='5730523P12Rik, BNIP2beta, Nip21', description='BCL2/adenovirus E1B interacting protein 2', other_designations='BCL2/adenovirus E1B 19 kDa-interacting protein 2 beta|BCL2/adenovirus E1B 19kDa-interacting protein 1, NIP2|BCL2/adenovirus E1B interacting protein 1, NIP2', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=69989425, end_position_on_the_genomic_accession=70011659, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12176, CurrentID=0, Status='live', Symbol='Bnip3', Aliases='Nip3', description='BCL2/adenovirus E1B interacting protein 3', other_designations='BCL2/adenovirus E1B 19 kDa-interacting protein 1, NIP3|BCL2/adenovirus E1B 19kDa-interacting protein 1, NIP3|BCL2/adenovirus E1B interacting protein 1, NIP3', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=138890836, end_position_on_the_genomic_accession=138909506, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12177, CurrentID=0, Status='live', Symbol='Bnip3l', Aliases='C86132, D14Ertd719e, Nip3L, Nix', description='BCL2/adenovirus E1B interacting protein 3-like', other_designations='BCL2/adenovirus E1B 19 kDa-interacting protein 3-like|BCL2/adenovirus E1B 19kDa-interacting protein 3-like|NIP3-like protein X', map_location='14 34.6 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66985238, end_position_on_the_genomic_accession=67008894, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12180, CurrentID=0, Status='live', Symbol='Smyd1', Aliases='4632404M21Rik, Bop, C78565, Zmynd18', description='SET and MYND domain containing 1', other_designations='CD8b-opposite|CD8beta opposite strand|SET and MYND domain-containing protein 1|m-BOP|zinc finger protein BOP', map_location='6 32.14 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71213940, end_position_on_the_genomic_accession=71262281, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12181, CurrentID=0, Status='live', Symbol='Bop1', Aliases='AU020183, AW146150, D18861, Erb1p, Kiaa0124, mKIAA0124', description='block of proliferation 1', other_designations='block of proliferation 1 protein', map_location='15 35.91 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76452996, end_position_on_the_genomic_accession=76477269, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  12182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12182, CurrentID=0, Status='live', Symbol='Bst1', Aliases='114/A10, A530073F09, BP-3, Bp3, Bsta1, CD157, Ly65', description='bone marrow stromal cell antigen 1', other_designations='ADP-ribosyl cyclase 2|BP-3 alloantigen|BST-1|antigen BP3|bone marrow stromal antigen 1|cADPr hydrolase 2|cyclic ADP-ribose hydrolase 2|leukocyte antigen 65|ly-65', map_location='5 23.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43818880, end_position_on_the_genomic_accession=43843468, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12183, CurrentID=0, Status='live', Symbol='Bpgm', Aliases='AI323730, AL022789, C86192', description='2,3-bisphosphoglycerate mutase', other_designations='2,3-bisphosphoglycerate mutase, erythrocyte|2,3-bisphosphoglycerate synthase|BPG-dependent PGAM', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34476254, end_position_on_the_genomic_accession=34505610, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12189, CurrentID=0, Status='live', Symbol='Brca1', Aliases='', description='breast cancer 1, early onset', other_designations='breast cancer 1|breast cancer associated 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101488754, end_position_on_the_genomic_accession=101551956, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  12190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12190, CurrentID=0, Status='live', Symbol='Brca2', Aliases='Fancd1, RAB163', description='breast cancer 2, early onset', other_designations='breast cancer 2|fanconi anemia group D1 protein homolog', map_location='5 89.52 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=150522021, end_position_on_the_genomic_accession=150570147, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  12192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12192, CurrentID=0, Status='live', Symbol='Zfp36l1', Aliases='AW742437, AW743212, Berg36, Brf1, D530020L18Rik, ERF1, TIS11b, cMG1', description='zinc finger protein 36, C3H type-like 1', other_designations='butyrate response factor 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80107760, end_position_on_the_genomic_accession=80113013, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12193, CurrentID=0, Status='live', Symbol='Zfp36l2', Aliases='Brf2, ERF2, Tis11d', description='zinc finger protein 36, C3H type-like 2', other_designations='ZFP36-like 2|butyrate response factor 2', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84183924, end_position_on_the_genomic_accession=84187947, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12209, CurrentID=0, Status='live', Symbol='Brs3', Aliases='BRS-3', description='bombesin-like receptor 3', other_designations='', map_location='X A7.1-A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=57043074, end_position_on_the_genomic_accession=57048758, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12211, CurrentID=0, Status='live', Symbol='Birc6', Aliases='A430032G04Rik, A430040A19Rik, AA501170, Bruce, D630005A10Rik, mKIAA1289', description='baculoviral IAP repeat-containing 6', other_designations='BIR repeat-containing ubiquitin-conjugating enzyme|apollon|baculovirus inhibitor of apoptosis repeat containing ubiquitin-conjugating enzyme|baculovirus inhibitor of apoptosis repeat containing ubiquitin-conjugating enzyme; accccbaculoviral IAP repeat-containing 6|ubiquitin-conjugating BIR domain enzyme apollon', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74528216, end_position_on_the_genomic_accession=74703356, orientation='plus', exon_count=75, OMIM='', no_hdr0=''),
  12212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12212, CurrentID=0, Status='live', Symbol='Chic1', Aliases='Brx', description='cysteine-rich hydrophobic domain 1', other_designations='brain X-linked protein|cystein-rich hydrophobic domain 1|cysteine-rich hydrophobic domain 1 protein', map_location='X 42.3 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103356328, end_position_on_the_genomic_accession=103409096, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12215, CurrentID=0, Status='live', Symbol='Bsg', Aliases='AI115436, AI325119, CD147, EMMPRIN, HT-7', description='basigin', other_designations='5A11/Basigin|HT7 antigen|basic immunoglobulin superfamily|gp 42|membrane glycoprotein gp42|neurothelin', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79704358, end_position_on_the_genomic_accession=79711979, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12217, CurrentID=0, Status='live', Symbol='Bsn', Aliases='', description='bassoon', other_designations='presynaptic cytomatrix protein', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108096022, end_position_on_the_genomic_accession=108191938, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12223, CurrentID=0, Status='live', Symbol='Btc', Aliases='Bcn', description='betacellulin, epidermal growth factor family member', other_designations='probetacellulin', map_location='5 44.78 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=91357261, end_position_on_the_genomic_accession=91402994, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12224, CurrentID=0, Status='live', Symbol='Klf5', Aliases='4930520J07Rik, Bteb2, CKLF, IKLF', description='Kruppel-like factor 5', other_designations='BTE-binding protein 2|basic transcription element binding protein 2|basic transcription element-binding protein 2|intestinal-enriched krueppel-like factor|transcription factor BTEB2', map_location='14 E2.2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=99298691, end_position_on_the_genomic_accession=99313412, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12226, CurrentID=0, Status='live', Symbol='Btg1', Aliases='AI426953, AW546738', description='B cell translocation gene 1, anti-proliferative', other_designations='B-cell translocation gene 1 protein|B-cell translocation gene 1, anti-proliferative', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=96617001, end_position_on_the_genomic_accession=96622813, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12227, CurrentID=0, Status='live', Symbol='Btg2', Aliases='AA959598, APRO1, Pc3, TIS21', description='B cell translocation gene 2, anti-proliferative', other_designations='B-cell translocation gene 2, anti-proliferative|BTG family member 2|NGF-inducible protein TIS21', map_location='1 58.1 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134074865, end_position_on_the_genomic_accession=134079155, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12228, CurrentID=0, Status='live', Symbol='Btg3', Aliases='ANA, tob5', description='B cell translocation gene 3', other_designations='BTG family member 3|abundant in neuroepithelium area protein', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=78359860, end_position_on_the_genomic_accession=78377181, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12229, CurrentID=0, Status='live', Symbol='Btk', Aliases='AI528679, xid', description='Bruton agammaglobulinemia tyrosine kinase', other_designations="ATK|B-cell progenitor kinase|BPK|Bruton's tyrosine kinase|Tyrosine-protein kinase BTK (Brutons tyrosine kinase) (Agammaglobulinaemia tyrosine kinase) (ATK) (B cell progenitor kinase) (BPK) (Kinase EMB)|X-linked immune deficiency|agammaglobulinaemia tyrosine kinase|agammaglobulinemia tyrosine kinase|bruton tyrosine kinase|kinase EMB", map_location='X 56.18 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134542334, end_position_on_the_genomic_accession=134583623, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  12231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12231, CurrentID=0, Status='live', Symbol='Btn1a1', Aliases='Btn', description='butyrophilin, subfamily 1, member A1', other_designations='BT', map_location='13 9.72 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23456984, end_position_on_the_genomic_accession=23465914, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12234, CurrentID=0, Status='live', Symbol='Btrc', Aliases='Beta-Trcp1, E3RS-IkappaB, E3RSIkappaB, FWD1, Fbw1a, HOS, Slimb, b-TrCP, beta-TrCP', description='beta-transducin repeat containing protein', other_designations='E3RSIkappaB|HOS|SCF b-TRCP|beta-TrCP|btrc', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45363714, end_position_on_the_genomic_accession=45533343, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  12235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12235, CurrentID=0, Status='live', Symbol='Bub1', Aliases='AL022991a, C80208, D2Xrf87, Bub1', description='BUB1, mitotic checkpoint serine/threonine kinase', other_designations='budding uninhibited by benzimidazoles 1 homolog|mBUB1', map_location='2 62.1 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127800200, end_position_on_the_genomic_accession=127831876, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  12236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12236, CurrentID=0, Status='live', Symbol='Bub1b', Aliases='AU045529, BUBR1', description='BUB1B, mitotic checkpoint serine/threonine kinase', other_designations='MAD3/BUB1-related protein kinase|budding uninhibited by benzimidazoles 1 beta|budding uninhibited by benzimidazoles 1 homolog, beta|mitotic checkpoint kinase MAD3L', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118598203, end_position_on_the_genomic_accession=118641592, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  12237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12237, CurrentID=0, Status='live', Symbol='Bub3', Aliases='AU019800, AU021329, AU043350, AW146323, C78067', description='BUB3 mitotic checkpoint protein', other_designations='WD repeat type I transmembrane protein A72.5|budding uninhibited by benzimidazoles 3 homolog', map_location='7 74.36 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131560370, end_position_on_the_genomic_accession=131571898, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12238, CurrentID=0, Status='live', Symbol='Commd3', Aliases='AW550818, Bup, D2Ertd542e', description='COMM domain containing 3', other_designations='Bmi1 upstream|bmi-1 upstream gene protein', map_location='2 12.9 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=18672462, end_position_on_the_genomic_accession=18676216, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  667612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667612, CurrentID=0, Status='live', Symbol='Ncr3-ps', Aliases='1C7, EG667612, Gm8726, Ly117, Ncr3', description='natural cytotoxicity triggering receptor 3, pseudogene', other_designations='lymphocyte antigen 117', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35180536, end_position_on_the_genomic_accession=35184809, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  111269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111269, CurrentID=0, Status='live', Symbol='Hprt-ps1', Aliases='D17Rp10, RP10', description='hypoxanthine guanine phosphoribosyl transferase pseudogene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12256, CurrentID=0, Status='live', Symbol='Bxv1', Aliases='Bxv-1, Xmv-43, Xmv43', description='B10 xenotropic virus 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12257, CurrentID=0, Status='live', Symbol='Tspo', Aliases='Bzrp, IBP, PBR1, Tspo', description='translocator protein', other_designations='PKBS|benzodiazepine receptor, peripheral|isoquinoline-binding protein|mitochondrial benzodiazepine receptor|peripheral-type benzodiazepine receptor', map_location='15 39.4 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83563573, end_position_on_the_genomic_accession=83574203, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12258, CurrentID=0, Status='live', Symbol='Serping1', Aliases='C1INH, C1nh', description='serine (or cysteine) peptidase inhibitor, clade G, member 1', other_designations='C1 Inh|C1 esterase inhibitor|C1 inhibitor|C1-inhibiting factor|complement component 1 inhibitor|serine (or cysteine) proteinase inhibitor, clade G (C1 inhibitor), member 1|serine (or cysteine) proteinase inhibitor, clade G, member 1|serpin G1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84765360, end_position_on_the_genomic_accession=84775445, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12259, CurrentID=0, Status='live', Symbol='C1qa', Aliases='AI255395, C1q', description='complement component 1, q subcomponent, alpha polypeptide', other_designations='', map_location='4 69.05 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136895916, end_position_on_the_genomic_accession=136898844, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12260, CurrentID=0, Status='live', Symbol='C1qb', Aliases='', description='complement component 1, q subcomponent, beta polypeptide', other_designations='', map_location='4 69.05 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136880145, end_position_on_the_genomic_accession=136886177, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12261, CurrentID=0, Status='live', Symbol='C1qbp', Aliases='AA407365, AA986492, D11Wsu182e, HABP1, P32, gC1qBP', description='complement component 1, q subcomponent binding protein', other_designations='GC1q-R protein|glycoprotein gC1qBP', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70977846, end_position_on_the_genomic_accession=70983026, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12262, CurrentID=0, Status='live', Symbol='C1qc', Aliases='AI385742, C1qg, Ciqc', description='complement component 1, q subcomponent, C chain', other_designations='complement component 1, q subcomponent, c polypeptide|complement component 1, q subcomponent, gamma polypeptide', map_location='4 69.05 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136889802, end_position_on_the_genomic_accession=136892914, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12263, CurrentID=0, Status='live', Symbol='C2', Aliases='', description='complement component 2 (within H-2S)', other_designations='MHC complement component C2|classical-complement pathway C3/C5 convertase|complement factor C2', map_location='17 18.41 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34862602, end_position_on_the_genomic_accession=34882100, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  12265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12265, CurrentID=0, Status='live', Symbol='Ciita', Aliases='C2ta, EG669998, Gm9475, Mhc2ta', description='class II transactivator', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10480015, end_position_on_the_genomic_accession=10528606, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  12266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12266, CurrentID=0, Status='live', Symbol='C3', Aliases='AI255234, ASP, HSE-MSF, Plp', description='complement component 3', other_designations='acylation stimulating protein|complement component 3d|complement factor 3', map_location='17 E1-E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57203968, end_position_on_the_genomic_accession=57228136, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  12267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12267, CurrentID=0, Status='live', Symbol='C3ar1', Aliases='AZ3B, C3AR, HNFAG09', description='complement component 3a receptor 1', other_designations='C3a-R|anaphylatoxin C3a receptor', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122847140, end_position_on_the_genomic_accession=122856157, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12268, CurrentID=0, Status='live', Symbol='C4b', Aliases='C4, Ss', description='complement component 4B (Chido blood group)', other_designations='complement component 4 (within H-2S)|complement component 4B (Childo blood group)', map_location='17 18.29 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34728046, end_position_on_the_genomic_accession=34743901, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  12269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12269, CurrentID=0, Status='live', Symbol='C4bp', Aliases='AI195242a, C4bp', description='complement component 4 binding protein', other_designations='', map_location='1 56.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130635921, end_position_on_the_genomic_accession=130661618, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12270, CurrentID=0, Status='live', Symbol='C4bp-ps1', Aliases='C4bpb', description='complement component 4 binding protein, pseudogene 1', other_designations='', map_location='1 56.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130670202, end_position_on_the_genomic_accession=130681571, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12273, CurrentID=0, Status='live', Symbol='C5ar1', Aliases='C5aR, C5r1, Cd88, D7Msu1', description='complement component 5a receptor 1', other_designations='C5a anaphylatoxin receptor|C5a ligand|C5a-R|complement component 5, receptor 1', map_location='7 8.77 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16246743, end_position_on_the_genomic_accession=16259540, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12274, CurrentID=0, Status='live', Symbol='C6', Aliases='AW111623', description='complement component 6', other_designations='', map_location='15 1.97 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=4727068, end_position_on_the_genomic_accession=4823967, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  12279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12279, CurrentID=0, Status='live', Symbol='C9', Aliases='', description='complement component 9', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=6445333, end_position_on_the_genomic_accession=6498476, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12281, CurrentID=0, Status='live', Symbol='cab', Aliases='', description='cardiac abnormality', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12282, CurrentID=0, Status='live', Symbol='Hyou1', Aliases='AI415631, CBP-140, Cab140, Grp170, Orp150', description='hypoxia up-regulated 1', other_designations='140 kDa Ca(2+)-binding protein|150-kDa oxygen regulated protein|170 kDa glucose regulated protein GRP170|GRP-170|calcium binding protein 140|calcium binding protein, 140 kDa', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44379490, end_position_on_the_genomic_accession=44392369, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  12283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12283, CurrentID=0, Status='live', Symbol='Cab39', Aliases='AA408805, AA960512, C78372, MO25, MO25alpha', description='calcium binding protein 39', other_designations='calcium binding protein, 39 kDa', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85793410, end_position_on_the_genomic_accession=85851577, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12285, CurrentID=0, Status='live', Symbol='cac', Aliases='cat', description='recessive cataract', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12286, CurrentID=0, Status='live', Symbol='Cacna1a', Aliases='APCA, BI, Caca1a, Cacnl1a4, Cav2.1, Ccha1a, EA2, FHM, HPCA, MHP, MHP1, SCA6, alpha1A, la, nmf352, rkr, smrl, tg', description='calcium channel, voltage-dependent, P/Q type, alpha 1A subunit', other_designations='brain calcium channel I|calcium channel, L type, alpha-1 polypeptide|leaner|rocker|tottering|voltage-gated calcium channel subunit alpha Cav2.1', map_location='8 40.95 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84338633, end_position_on_the_genomic_accession=84640249, orientation='plus', exon_count=55, OMIM='', no_hdr0=''),
  12287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12287, CurrentID=0, Status='live', Symbol='Cacna1b', Aliases='AW050276, AW060892, AW822256, Cav2.2, Cchn1a, alpha(1B)', description='calcium channel, voltage-dependent, N type, alpha 1B subunit', other_designations='BIII|brain calcium channel III|calcium channel, L type, alpha-1 polypeptide|calcium channel, voltage-dependent, L type, alpha 1B subunit|voltage-gated calcium channel subunit alpha Cav2.2', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24603860, end_position_on_the_genomic_accession=24763195, orientation='minus', exon_count=52, OMIM='', no_hdr0=''),
  12288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12288, CurrentID=0, Status='live', Symbol='Cacna1c', Aliases='Cav1.2, Cchl1a1, D930026N18Rik, MBC, MELC-CC', description='calcium channel, voltage-dependent, L type, alpha 1C subunit', other_designations='brain class C|calcium channel voltage-dependent alpha1c subunit|calcium channel, L type, alpha-1 polypeptide, isoform 1, cardiac muscle|neuronal voltage-gated calcium channel alpha 1C subunit|skeletal muscle-specific calcium channel|voltage-gated calcium channel subunit alpha Cav1.2', map_location='6 55.86 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118587240, end_position_on_the_genomic_accession=119198921, orientation='minus', exon_count=53, OMIM='', no_hdr0=''),
  12289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12289, CurrentID=0, Status='live', Symbol='Cacna1d', Aliases='8430418G19Rik, C79217, Cach3, Cacn4, Cacnl1a2, Cav1.3, Cchl1a, Cchl1a2, D-LTCC', description='calcium channel, voltage-dependent, L type, alpha 1D subunit', other_designations='Cav1.3alpha1|alpha 1 E|calcium channel, L type, alpha-1 polypeptide|voltage-gated calcium channel subunit alpha Cav1.3', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30039941, end_position_on_the_genomic_accession=30492363, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  12290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12290, CurrentID=0, Status='live', Symbol='Cacna1e', Aliases='A430040I15, BII, Cach6, Cacnl1a6, Cav2.3, Cchra1, alpha1E', description='calcium channel, voltage-dependent, R type, alpha 1E subunit', other_designations='brain calcium channel II|calcium channel, L type, alpha-1 polypeptide|calcium channel, voltage-dependent, L type, alpha 1E subunit|voltage-gated calcium channel subunit alpha Cav2.3', map_location='1 66.14 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=154390806, end_position_on_the_genomic_accession=154884782, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  12291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12291, CurrentID=0, Status='live', Symbol='Cacna1g', Aliases='Cav3.1d, [a]1G, a1G, alpha-1G, mKIAA1123', description='calcium channel, voltage-dependent, T type, alpha 1G subunit', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94408391, end_position_on_the_genomic_accession=94474390, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  12292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12292, CurrentID=0, Status='live', Symbol='Cacna1s', Aliases='AW493108, Cav1.1, Cchl1a3, fmd, mdg, sj', description='calcium channel, voltage-dependent, L type, alpha 1S subunit', other_designations='DHPR alpha1s|calcium channel, L type, alpha-1 polypeptide, isoform 3, skeletal muscle|dihydropyridine receptor alpha 1S|muscle dysgenesis|voltage-gated calcium channel subunit alpha Cav1.1', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136052805, end_position_on_the_genomic_accession=136119822, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  12293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12293, CurrentID=0, Status='live', Symbol='Cacna2d1', Aliases='Ca(v)alpha2delta1, Cacna2, Cchl2a', description='calcium channel, voltage-dependent, alpha2/delta subunit 1', other_designations='dihydropyridine-sensitive calcium channel alpha 2|voltage-gated calcium channel subunit alpha-2/delta-1', map_location='5 6.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=15934691, end_position_on_the_genomic_accession=16374511, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  12294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12294, CurrentID=0, Status='live', Symbol='Cacna2d3', Aliases='Cacnad3, alapha2delta3', description='calcium channel, voltage-dependent, alpha2/delta subunit 3', other_designations='calcium channel, voltage dependent, alpha2/delta subunit 3|voltage-gated calcium channel subunit alpha-2/delta-3', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=28904943, end_position_on_the_genomic_accession=29721864, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  12295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12295, CurrentID=0, Status='live', Symbol='Cacnb1', Aliases='CAB1, Cchb1, Cchlb1', description='calcium channel, voltage-dependent, beta 1 subunit', other_designations='calcium channel voltage-dependent subunit beta 1|voltage-dependent calcium channel beta1 subunit', map_location='11 61.5 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98001508, end_position_on_the_genomic_accession=98022975, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  12296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12296, CurrentID=0, Status='live', Symbol='Cacnb2', Aliases='AW060387, CAB2, Cavbeta2, Cchb2', description='calcium channel, voltage-dependent, beta 2 subunit', other_designations='voltage-gated calcium channel beta2 subunit transcript variant X13', map_location='2 10.7 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14604049, end_position_on_the_genomic_accession=14988811, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  12297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12297, CurrentID=0, Status='live', Symbol='Cacnb3', Aliases='Beta3, CAB3, Ca(v)beta3, Cchb3', description='calcium channel, voltage-dependent, beta 3 subunit', other_designations='calcium channel voltage-dependent subunit beta 3', map_location='15 54.64 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98626873, end_position_on_the_genomic_accession=98644536, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  12298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12298, CurrentID=0, Status='live', Symbol='Cacnb4', Aliases='3110038O15Rik, Cchb4, lh', description='calcium channel, voltage-dependent, beta 4 subunit', other_designations='CAB4|calcium channel voltage-dependent subunit beta 4', map_location='2 29.98 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=52428320, end_position_on_the_genomic_accession=52676609, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12299, CurrentID=0, Status='live', Symbol='Cacng1', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 1', other_designations='dihydropyridine receptor gamma 1|dihydropyridine-sensitive L-type, skeletal muscle calcium channel subunit gamma', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107703218, end_position_on_the_genomic_accession=107716476, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12300, CurrentID=0, Status='live', Symbol='Cacng2', Aliases='AW060990, B230105C07Rik, B930041E13Rik, stargazer, stargazin, stg, wag, waggler', description='calcium channel, voltage-dependent, gamma subunit 2', other_designations='TARP gamma 2|TARP gamma-2|neuronal voltage-gated calcium channel gamma-2 subunit|transmembrane AMPAR regulatory protein gamma-2', map_location='15 36.92 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77993623, end_position_on_the_genomic_accession=78119280, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12301, CurrentID=0, Status='live', Symbol='Cacybp', Aliases='SIP', description='calcyclin binding protein', other_designations='siah-interacting protein', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160202367, end_position_on_the_genomic_accession=160212892, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  667663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667663, CurrentID=0, Status='live', Symbol='Myo3a', Aliases='9030416P08Rik', description='myosin IIIA', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=22542413, end_position_on_the_genomic_accession=22618252, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  12304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12304, CurrentID=0, Status='live', Symbol='Pdia4', Aliases='AI987846, Cai, ERp-72, Erp72', description='protein disulfide isomerase associated 4', other_designations='ER protein 72|calcium binding protein, intestinal|endoplasmic reticulum protein 72|endoplasmic reticulum resident protein 72', map_location='6 22.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47796141, end_position_on_the_genomic_accession=47813512, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12305, CurrentID=0, Status='live', Symbol='Ddr1', Aliases='6030432F18, AI323681, CD167a, Cak, Nep, PTK3A', description='discoidin domain receptor family, member 1', other_designations='CD167 antigen-like family member A|cell adhesion kinase|discoidin receptor tyrosine kinase|epithelial discoidin domain receptor 1|protein-tyrosine kinase MPK-6|tyrosine kinase DDR|tyrosine-protein kinase CAK', map_location='17 18.7 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35681567, end_position_on_the_genomic_accession=35704629, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  12306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12306, CurrentID=0, Status='live', Symbol='Anxa2', Aliases='AW215814, Cal1h', description='annexin A2', other_designations='33-kDa calcimedin|33-kDa lymphocyte Ca<2+>- binding protein|33kDa lymphocyte Ca<2+>- binding protein|36-kDa calelectrin|PAP-IV|annexin II|annexin-2|calpactin I heavy chain|calpactin-1 heavy chain|chromobindin 8|chromobindin-8|lipocortin II|p36|placental anticoagulant protein IV|protein I', map_location='9 38.58 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=69453683, end_position_on_the_genomic_accession=69491795, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12307, CurrentID=0, Status='live', Symbol='Calb1', Aliases='Brain-2, CB, Calb, Calb-1', description='calbindin 1', other_designations='CalbindinD28K|D-28K|PCD-29|calbindin 1, 28kDa|calbindin D28|calbindin-28K|calbindin-D (28k)|spot 35 protein|vitamin D-dependent calcium-binding protein, avian-type', map_location='4 6.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=15881264, end_position_on_the_genomic_accession=15906709, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12308, CurrentID=0, Status='live', Symbol='Calb2', Aliases='CR', description='calbindin 2', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110142538, end_position_on_the_genomic_accession=110168206, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12309, CurrentID=0, Status='live', Symbol='S100g', Aliases='AU042539, CABP1, CaBP-D9K, Cabp9k, Calb3', description='S100 calcium binding protein G', other_designations="CABP|S100 calcium-binding protein G|calbindin 3, (vitamin D-dependent calcium binding protein)|calbindin D9k|calbindin-D9K major form|cytidine 5'-triphosphate synthase 2|vitamin D-dependent calcium-binding protein, intestinal", map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=162961992, end_position_on_the_genomic_accession=162964599, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12310, CurrentID=0, Status='live', Symbol='Calca', Aliases='CA, CGRP-1, CGRP1, Calc, Calc1, Cgrp, Ct, Ctn', description='calcitonin/calcitonin-related polypeptide, alpha', other_designations='alpha-type CGRP|calcitonin gene-related peptide I', map_location='7 59.99 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114625985, end_position_on_the_genomic_accession=114636460, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12311, CurrentID=0, Status='live', Symbol='Calcr', Aliases='Clr', description='calcitonin receptor', other_designations='CT-R', map_location='6 1.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3685720, end_position_on_the_genomic_accession=3764732, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12313, CurrentID=0, Status='live', Symbol='Calm1', Aliases='AI256814, AI327027, AI461935, AL024000, CaM, Calm', description='calmodulin 1', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100199541, end_position_on_the_genomic_accession=100209806, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12314, CurrentID=0, Status='live', Symbol='Calm2', Aliases='1500001E21Rik, AL024017', description='calmodulin 2', other_designations='caM', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87433401, end_position_on_the_genomic_accession=87446935, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12315, CurrentID=0, Status='live', Symbol='Calm3', Aliases='CaMA, R75142', description='calmodulin 3', other_designations='caM', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16915379, end_position_on_the_genomic_accession=16924032, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12316, CurrentID=0, Status='live', Symbol='Aspm', Aliases='Calmbp1, D330028K02Rik, MCPH5, Sha1', description='asp (abnormal spindle)-like, microcephaly associated (Drosophila)', other_designations='calmodulin binding protein 1|calmodulin-binding protein 1|calmodulin-binding protein Sha1|spindle and hydroxyurea checkpoint abnormal protein', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=139454773, end_position_on_the_genomic_accession=139494091, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  12317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12317, CurrentID=0, Status='live', Symbol='Calr', Aliases='CRTegulin, Calr', description='calreticulin', other_designations='CRP55|ERp60|HACBP|endoplasmic reticulum resident protein 60', map_location='8 41.21 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84842088, end_position_on_the_genomic_accession=84846931, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12319, CurrentID=0, Status='live', Symbol='Car8', Aliases='AW546993, Ca8, Cals, Cals1, Carp, wdl', description='carbonic anhydrase 8', other_designations='CA-RP VIII|CA-VIII|carbonic anhydrase VIII|carbonic anhydrase-like sequence 1', map_location='4 3.53 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=8141493, end_position_on_the_genomic_accession=8239041, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12321, CurrentID=0, Status='live', Symbol='Calu', Aliases='9530075H20Rik', description='calumenin', other_designations='crocalbin', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29348106, end_position_on_the_genomic_accession=29376675, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12322, CurrentID=0, Status='live', Symbol='Camk2a', Aliases='CaMKII, R74975, mKIAA0968', description='calcium/calmodulin-dependent protein kinase II alpha', other_designations='CaMK II|alpha-CaMKII|caM kinase II subunit alpha|caMK-II subunit alpha', map_location='18 34.41 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60925580, end_position_on_the_genomic_accession=60988152, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  667683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667683, CurrentID=0, Status='live', Symbol='Igkv4-59', Aliases='EG667683, Gm8760, kk4', description='immunoglobulin kappa variable 4-59', other_designations='immunoglobulin kappa chain variable 4-59|monoclonal antibody kappa light chain', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69438218, end_position_on_the_genomic_accession=69438741, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  12325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12325, CurrentID=0, Status='live', Symbol='Camk2g', Aliases='Camkg', description='calcium/calmodulin-dependent protein kinase II gamma', other_designations='Ca2+/calmodulin-dependent protein kinase II|CaMK II|caMK-II subunit gamma', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20734873, end_position_on_the_genomic_accession=20794130, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  12326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12326, CurrentID=0, Status='live', Symbol='Camk4', Aliases='A430110E23Rik, AI666733, CaMKIV, CaMKIV/Gr, D18Bwg0362e', description='calcium/calmodulin-dependent protein kinase IV', other_designations='CAM kinase-GR|Ca2+/calmodulin-dependent protein kinase type IV/Gr|caMK IV', map_location='18 18.15 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32934957, end_position_on_the_genomic_accession=33195767, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  12328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12328, CurrentID=0, Status='live', Symbol='Caml', Aliases='AI385748g, Caml', description='calcium modulating ligand', other_designations='', map_location='13 30.06 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55623005, end_position_on_the_genomic_accession=55632416, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12329, CurrentID=0, Status='live', Symbol='can', Aliases='', description='cartilage anomaly', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12330, CurrentID=0, Status='live', Symbol='Canx', Aliases='1110069N15Rik, AI988026, Cnx, D11Ertd153e', description='calnexin', other_designations='', map_location='11 30.46 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50293957, end_position_on_the_genomic_accession=50325673, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12331, CurrentID=0, Status='live', Symbol='Cap1', Aliases='', description='CAP, adenylate cyclase-associated protein 1 (yeast)', other_designations='CAP 1|adenylyl cyclase-associated CAP protein, yeast homolog 1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=122859048, end_position_on_the_genomic_accession=122886071, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12332, CurrentID=0, Status='live', Symbol='Capg', Aliases='gCap39, mbh1', description='capping protein (actin filament), gelsolin-like', other_designations='actin regulatory protein CAP-G|actin-capping protein GCAP39|myc basic motif homolog 1', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72544391, end_position_on_the_genomic_accession=72562983, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12333, CurrentID=0, Status='live', Symbol='Capn1', Aliases='Capa-1, Capa1, mu-calpin', description='calpain 1', other_designations='CANP 1|calcium-activated neutral proteinase 1|calpain mu-type|calpain-1 large subunit|micromolar-calpain|mu-calpain|muCANP', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5988545, end_position_on_the_genomic_accession=6015825, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  12334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12334, CurrentID=0, Status='live', Symbol='Capn2', Aliases='AI326419, Capa-2, Capa2, m-calpain, m-calpin', description='calpain 2', other_designations='80 kDa M-calpain subunit|CALP80|CANP 2|calcium-activated neutral proteinase 2|calpain M-type|calpain-2 large subunit|cysteine proteinase|m-calpain 80 kDa large subunit|m-calpain large subunit|millimolar-calpain', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182467259, end_position_on_the_genomic_accession=182517483, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  12335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12335, CurrentID=0, Status='live', Symbol='Capn3', Aliases='AI323605, Capa-3, Capa3, Lp82, p94', description='calpain 3', other_designations='CANP 3|calcium-activated neutral proteinase 3|calpain L3|calpain p94|muscle-specific calcium-activated neutral protease 3|skeletal muscle specific calpain p94', map_location='2 60.31 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120463593, end_position_on_the_genomic_accession=120504919, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  12336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12336, CurrentID=0, Status='live', Symbol='Capns1', Aliases='Capa-4, Capa4, Capn4, D7Ertd146e', description='calpain, small subunit 1', other_designations='CANP small subunit|CDPS|CSS1|calcium-activated neutral proteinase small subunit|calcium-dependent protease small subunit 1|calpain 4|calpain regulatory subunit', map_location='7 17.31 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30186942, end_position_on_the_genomic_accession=30198794, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12337, CurrentID=0, Status='live', Symbol='Capn5', Aliases='nCL-3', description='calpain 5', other_designations='new calpain 3', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98121559, end_position_on_the_genomic_accession=98179298, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12338, CurrentID=0, Status='live', Symbol='Capn6', Aliases='', description='calpain 6', other_designations='', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=143802236, end_position_on_the_genomic_accession=143827412, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12339, CurrentID=0, Status='live', Symbol='Capn7', Aliases='AU022319, PalBH', description='calpain 7', other_designations='palB homolog', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31336620, end_position_on_the_genomic_accession=31371991, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  12340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12340, CurrentID=0, Status='live', Symbol='Capza1', Aliases='CAPZ, CAZ1, Cappa1', description='capping protein (actin filament) muscle Z-line, alpha 1', other_designations='F-actin capping protein alpha-1 subunit|capZ alpha-1|capping protein alpha 1', map_location='3 45.88 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104822785, end_position_on_the_genomic_accession=104864505, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12343, CurrentID=0, Status='live', Symbol='Capza2', Aliases='1110053K06Rik, AW208808, Cappa2', description='capping protein (actin filament) muscle Z-line, alpha 2', other_designations='capZ alpha-2|capping protein alpha 2', map_location='6 7.96 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=17637098, end_position_on_the_genomic_accession=17666534, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12344, CurrentID=0, Status='live', Symbol='Capza3', Aliases='510-4, Cappa3, Gsg3, Tex8, repro32', description='capping protein (actin filament) muscle Z-line, alpha 3', other_designations='capZ alpha-3|capping protein alpha 3|cp alpha3|germ cell-specific protein 3', map_location='6 69.77 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=140041525, end_position_on_the_genomic_accession=140042789, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12345, CurrentID=0, Status='live', Symbol='Capzb', Aliases='1700120C01Rik, AI325129, CPB1, CPB2, CPbeat2, CPbeta1, CPbeta2, Cappb1', description='capping protein (actin filament) muscle Z-line, beta', other_designations='capZ beta|capping protein beta 1', map_location='4 70.59 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139192899, end_position_on_the_genomic_accession=139291820, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12346, CurrentID=0, Status='live', Symbol='Car1', Aliases='AW555628, Ca1, Car-1', description='carbonic anhydrase 1', other_designations='CA I|CA-I|carbonate dehydratase I|carbonic anhydrase I', map_location='3 3.18 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14766214, end_position_on_the_genomic_accession=14808365, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12348, CurrentID=0, Status='live', Symbol='Car11', Aliases='Ca11', description='carbonic anhydrase 11', other_designations='CA-RP II|CA-RP XI|CA-XI|CARP XI|CARP-2|carbonic anhydrase-related protein 2', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45699848, end_position_on_the_genomic_accession=45704671, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12349, CurrentID=0, Status='live', Symbol='Car2', Aliases='AI131712, CAII, Ca2, Car-2, Ltw-5, Lvtw-5', description='carbonic anhydrase 2', other_designations='CA II|CA-II|carbonate dehydratase II|carbonic anhydrase II', map_location='3 3.23 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14886278, end_position_on_the_genomic_accession=14900770, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12350, CurrentID=0, Status='live', Symbol='Car3', Aliases='BB219044, Ca3, Car-3', description='carbonic anhydrase 3', other_designations='CA-III|carbonate dehydratase III|carbonic anhydrase III', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14863490, end_position_on_the_genomic_accession=14872536, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12351, CurrentID=0, Status='live', Symbol='Car4', Aliases='AW456718, Ca4', description='carbonic anhydrase 4', other_designations='CA IV|CA-IV|carbonate dehydratase IV|carbonic anhydrase IV', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84957754, end_position_on_the_genomic_accession=84966054, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12352, CurrentID=0, Status='live', Symbol='Car5a', Aliases='CAV, Ca5a, Car5', description='carbonic anhydrase 5a, mitochondrial', other_designations='CA Y|CA-VA|carbonate dehydratase VA|carbonic anhydrase 5, mitochondrial|carbonic anhydrase 5b, mitochondrial|carbonic anhydrase VA', map_location='8 70.81 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121916135, end_position_on_the_genomic_accession=121944912, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12353, CurrentID=0, Status='live', Symbol='Car6', Aliases='Ca6, DOC1', description='carbonic anhydrase 6', other_designations='CA-VI|carbonate dehydratase VI|carbonic anhydrase VI|salivary carbonic anhydrase|secreted carbonic anhydrase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150187015, end_position_on_the_genomic_accession=150201135, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12354, CurrentID=0, Status='live', Symbol='Car7', Aliases='AV343731, Ca7', description='carbonic anhydrase 7', other_designations='CA-VII|carbonate dehydratase VII|carbonic anhydrase VII', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104534700, end_position_on_the_genomic_accession=104550347, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12355, CurrentID=0, Status='live', Symbol='Nr1i3', Aliases='AA209988, AI551208, CAR, CAR-beta, Care2, ESTM32, MB67', description='nuclear receptor subfamily 1, group I, member 3', other_designations='constitutive androstane receptor|constitutively transactivates RAREs|orphan nuclear receptor MB67', map_location='1 79.21 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171213935, end_position_on_the_genomic_accession=171218845, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12359, CurrentID=0, Status='live', Symbol='Cat', Aliases='2210418N07, Cas-1, Cas1, Cs-1', description='catalase', other_designations='catalase 1', map_location='2 54.43 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103453904, end_position_on_the_genomic_accession=103485153, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12361, CurrentID=0, Status='live', Symbol='Cask', Aliases='DXPri1, DXRib1, LIN-2, Pals3, mLin-2', description='calcium/calmodulin-dependent serine protein kinase (MAGUK family)', other_designations='', map_location='X 8.43 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=13517080, end_position_on_the_genomic_accession=13846784, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  12362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12362, CurrentID=0, Status='live', Symbol='Casp1', Aliases='ICE, Il1bc', description='caspase 1', other_designations='CASP-1|IL-1 beta-converting enzyme|IL-1B converting enzyme|IL-1BC|interleukin 1 beta-converting enzyme|interleukin-1 beta convertase|interleukin-1 beta-converting enzyme|p45', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=5298517, end_position_on_the_genomic_accession=5307281, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12363, CurrentID=0, Status='live', Symbol='Casp4', Aliases='CASP-11, CASP-4, Casp11, Caspl, ich-3', description='caspase 4, apoptosis-related cysteine peptidase', other_designations='caspase 11, apoptosis-related cysteine peptidase|caspase 11, apoptosis-related cysteine protease|caspase 4, apoptosis-related cysteine protease|caspase-11 short form splicing|protease ICH-3', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=5308825, end_position_on_the_genomic_accession=5336791, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12364, CurrentID=0, Status='live', Symbol='Casp12', Aliases='', description='caspase 12', other_designations='CASP-12', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=5345442, end_position_on_the_genomic_accession=5373034, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12365, CurrentID=0, Status='live', Symbol='Casp14', Aliases='MICE, mini-ICE', description='caspase 14', other_designations='CASP-14', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78711995, end_position_on_the_genomic_accession=78718347, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12366, CurrentID=0, Status='live', Symbol='Casp2', Aliases='CASP-2, ICH-1, NEDD-2, Nedd2', description='caspase 2', other_designations='neural precursor cell expressed developmentally down-regulated protein 2|protease ICH-1', map_location='6 20.54 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42264791, end_position_on_the_genomic_accession=42282508, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12367, CurrentID=0, Status='live', Symbol='Casp3', Aliases='A830040C14Rik, AC-3, CASP-3, CC3, CPP-32, CPP32, Lice, SCA-1, Yama, mldy', description='caspase 3', other_designations='SREBP cleavage activity 1|apopain|caspase 3, apoptosis related cysteine protease|cysteine protease CPP32|protein Yama', map_location='8 26.39 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46617291, end_position_on_the_genomic_accession=46639698, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12368, CurrentID=0, Status='live', Symbol='Casp6', Aliases='CASP-6, Mch2', description='caspase 6', other_designations='apoptotic protease Mch-2', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129901425, end_position_on_the_genomic_accession=129914102, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12369, CurrentID=0, Status='live', Symbol='Casp7', Aliases='AI314680, CMH-1, ICE-IAP3, Mch3, caspase-7, mCASP-7', description='caspase 7', other_designations='CASP-7|apoptotic protease Mch-3|cysteine protease LICE2', map_location='19 51.84 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56396091, end_position_on_the_genomic_accession=56442347, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12370, CurrentID=0, Status='live', Symbol='Casp8', Aliases='CASP-8, FLICE, MACH, Mch5', description='caspase 8', other_designations='Fas-linked ICE-like protease', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58795229, end_position_on_the_genomic_accession=58847503, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12371, CurrentID=0, Status='live', Symbol='Casp9', Aliases='AI115399, APAF-3, AW493809, CASP-9, Caspase-9, ICE-LAP6, Mch6', description='caspase 9', other_designations='ICE-like apoptotic protease 6|apoptotic protease Mch-6|apoptotic protease-activating factor 3', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141793612, end_position_on_the_genomic_accession=141815978, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12372, CurrentID=0, Status='live', Symbol='Casq1', Aliases='CSQ, CSQ-1, CSQ1, sCSQ', description='calsequestrin 1', other_designations='calmitine|calsequestrin, skeletal muscle isoform', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172209894, end_position_on_the_genomic_accession=172219895, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12373, CurrentID=0, Status='live', Symbol='Casq2', Aliases='AA033488, AW146219, Csq2, ESTM52, cCSQ', description='calsequestrin 2', other_designations='calsequestrin, cardiac muscle isoform|cardiac calsequestrin', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102086420, end_position_on_the_genomic_accession=102146517, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12374, CurrentID=0, Status='live', Symbol='Casr', Aliases='CaR, Gprc2a', description='calcium-sensing receptor', other_designations='G protein coupled receptor, family C, group 2, member A|G protein-coupled receptor, family C, group 2, member A|PCaR1|cation sensing receptor|parathyroid cell calcium-sensing receptor', map_location='16 25.57 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36490585, end_position_on_the_genomic_accession=36562141, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12380, CurrentID=0, Status='live', Symbol='Cast', Aliases='', description='calpastatin', other_designations='calpain inhibitor|calpastatin type III', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74692366, end_position_on_the_genomic_accession=74808874, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  667742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667742, CurrentID=0, Status='live', Symbol='Piezo2', Aliases='5930434P17, 9030411M15Rik, 9430028L06Rik, Fam38b, Fam38b2', description='piezo-type mechanosensitive ion channel component 2', other_designations='family with sequence similarity 38, member B|family with sequence similarity 38, member B2|protein PIEZO2', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=63010213, end_position_on_the_genomic_accession=63387307, orientation='minus', exon_count=55, OMIM='', no_hdr0=''),
  12385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12385, CurrentID=0, Status='live', Symbol='Ctnna1', Aliases='2010010M04Rik, AA517462, AI988031, Catna1', description='catenin (cadherin associated protein), alpha 1', other_designations='102 kDa cadherin-associated protein|CAP102|alpha E catenin|alpha E-catenin|alpha(E)-catenin|cadherin associated protein|catenin alpha 1', map_location='18 18.89 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35118912, end_position_on_the_genomic_accession=35254775, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  12386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12386, CurrentID=0, Status='live', Symbol='Ctnna2', Aliases='AI481747, Catna, Catna2, cdf, chp', description='catenin (cadherin associated protein), alpha 2', other_designations='alpha N-catenin|alpha(N)-catenin|cadherin associated protein|catenin alpha 2|cerebellar deficient folia', map_location='6 B3-D', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=76881637, end_position_on_the_genomic_accession=77979694, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  12387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12387, CurrentID=0, Status='live', Symbol='Ctnnb1', Aliases='Bfc, Catnb, Mesc', description='catenin (cadherin associated protein), beta 1', other_designations='beta-catenin', map_location='9 72.19 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120933400, end_position_on_the_genomic_accession=120960507, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12388, CurrentID=0, Status='live', Symbol='Ctnnd1', Aliases='AA409437, AU019353, Catns, Ctnnd, P120, mKIAA0384', description='catenin (cadherin associated protein), delta 1', other_designations='CAS|cadherin-associated Src substrate|catenin src|p120 catenin|p120(cas)|p120(ctn)|p120-catenin', map_location='2 49.45 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84600781, end_position_on_the_genomic_accession=84658972, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  12389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12389, CurrentID=0, Status='live', Symbol='Cav1', Aliases='Cav, Cav-1', description='caveolin 1, caveolae protein', other_designations='caveolin, caveolae protein 1|caveolin, caveolae protein, 22 kDa', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=17306335, end_position_on_the_genomic_accession=17341328, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12390, CurrentID=0, Status='live', Symbol='Cav2', Aliases='AI447843', description='caveolin 2', other_designations='', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=17281185, end_position_on_the_genomic_accession=17289130, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12391, CurrentID=0, Status='live', Symbol='Cav3', Aliases='AI385751, Cav-3, M-cav', description='caveolin 3', other_designations='M-caveolin', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=112459505, end_position_on_the_genomic_accession=112472872, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12393, CurrentID=0, Status='live', Symbol='Runx2', Aliases='AML3, Cbf, Cbfa-1, Cbfa1, LS3, Osf2, PEBP2aA, Pebp2a1, Pebpa2a', description='runt related transcription factor 2', other_designations='AKV core binding factor|CBF-alpha-1|PEA2-alpha A|PEBP2 alpha A|SL3-3 enhancer factor 1 alpha A subunit|acute myeloid leukemia 3 protein|core binding factor alpha 1|osteoblast-specific transcription factor 2|polyomavirus enhancer-binding protein 2 alpha A subunit|runt domain, alpha subunit 1', map_location='17 21.33 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=44495986, end_position_on_the_genomic_accession=44814831, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12394, CurrentID=0, Status='live', Symbol='Runx1', Aliases='AML1, CBF-alpha-2, Cbfa2, Pebp2a2, Pebpa2b', description='runt related transcription factor 1', other_designations='SL3-3 enhancer factor 1 alpha B subunit|SL3/AKV core-binding factor alpha B subunit|acute myeloid leukemia 1 protein|core binding factor alpha 2|core-binding factor subunit alpha-2|oncogene AML-1|polyomavirus enhancer-binding protein 2 alpha B subunit|runt domain, alpha subunit 2', map_location='16 53.7 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92601466, end_position_on_the_genomic_accession=92826074, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12395, CurrentID=0, Status='live', Symbol='Runx1t1', Aliases='Cbfa2t1h, ETO, MTG8', description='runt-related transcription factor 1; translocated to, 1 (cyclin D-related)', other_designations='CBFA2T1 identified gene homolog|acute myelogenous leukemia 1 translocation 1 protein', map_location='4 A', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=13743182, end_position_on_the_genomic_accession=13895056, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12396, CurrentID=0, Status='live', Symbol='Cbfa2t2', Aliases='A430091M07, C330013D05Rikh, MTGR1, Cbfa2t2', description='core-binding factor, runt domain, alpha subunit 2, translocated to, 2 (human)', other_designations='CBFA2T2 identified gene homolog|MTG8-like protein|MTG8-related protein 1|core-binding factor runt domain alpha subunit 2 translocated to 2 homolog|core-binding factor, runt domain, alpha subunit 2, translocated to, 2 homolog', map_location='2 C3.2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154436463, end_position_on_the_genomic_accession=154539356, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12398, CurrentID=0, Status='live', Symbol='Cbfa2t3', Aliases='A630044F12Rik, AI465270, AW229127h, ETO-2, Eto2, MTGR2, Cbfa2t3', description='core-binding factor, runt domain, alpha subunit 2, translocated to, 3 (human)', other_designations='ETO/MTG8-related protein ETO-2|MTG8-related protein 2|eight twenty one protein 2|myeloid translocation gene-related protein 2', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122625136, end_position_on_the_genomic_accession=122699639, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12399, CurrentID=0, Status='live', Symbol='Runx3', Aliases='AML2, Cbfa3, Pebp2a3', description='runt related transcription factor 3', other_designations='CBF-alpha-3|PEA2-alpha C|PEBP2-alpha C|SL3-3 enhancer factor 1 alpha C subunit|SL3/AKV core-binding factor alpha C subunit|acute myeloid leukemia 2 protein|core binding factor alpha 3|core-binding factor subunit alpha-3|oncogene AML-2|polyomavirus enhancer-binding protein 2 alpha C subunit|runt domain, alpha subunit 3|transcription factor AML2/CBFA3', map_location='4 67.19 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135120645, end_position_on_the_genomic_accession=135177990, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12400, CurrentID=0, Status='live', Symbol='Cbfb', Aliases='AI893578, PEA2, PEBP2b, Pebp2, Pebpb2', description='core binding factor beta', other_designations='CBF-beta|PEA2-beta|PEBP2-beta|SL3-3 enhancer factor 1 subunit beta|SL3/AKV core-binding factor beta subunit|polyomavirus enhancer-binding protein 2 beta subunit', map_location='8 53.04 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105170674, end_position_on_the_genomic_accession=105217989, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12401, CurrentID=0, Status='live', Symbol='Serpina6', Aliases='Cbg', description='serine (or cysteine) peptidase inhibitor, clade A, member 6', other_designations='corticosteroid binding globulin|serine (or cysteine) proteinase inhibitor, clade A, member 6|serpin A6|transcortin', map_location='12 52.98 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103646632, end_position_on_the_genomic_accession=103657071, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12402, CurrentID=0, Status='live', Symbol='Cbl', Aliases='4732447J05Rik-2, c-Cbl, cbl, Cbl', description='Casitas B-lineage lymphoma', other_designations='casitas B-lineage lymphoma proto-oncogene|proto-oncogene c-CBL|signal transduction protein CBL', map_location='9 24.72 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44145040, end_position_on_the_genomic_accession=44234046, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  12404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12404, CurrentID=0, Status='live', Symbol='Cbln1', Aliases='AI323299', description='cerebellin 1 precursor protein', other_designations='brain protein D3|precerebellin', map_location='8 42.16 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=87468853, end_position_on_the_genomic_accession=87472592, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12405, CurrentID=0, Status='live', Symbol='Cbln2', Aliases='6330593N19Rik, A730004O05', description='cerebellin 2 precursor protein', other_designations='', map_location='18 58.63 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=86711053, end_position_on_the_genomic_accession=86718285, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12406, CurrentID=0, Status='live', Symbol='Serpinh1', Aliases='BERF-1, Cbp1, Cbp2, Hsp47, J6, Serpinh2, gp46', description='serine (or cysteine) peptidase inhibitor, clade H, member 1', other_designations='47 kDa heat shock protein|beta enolase repressor factor 1|collagen binding protein 1|colligin|serine (or cysteine) proteinase inhibitor, clade H (heat shock protein 47), member 1|serine (or cysteine) proteinase inhibitor, clade H (heat shock protein 47), member 2|serine protease inhibitor J6', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99345375, end_position_on_the_genomic_accession=99353239, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12408, CurrentID=0, Status='live', Symbol='Cbr1', Aliases='AW261796, CR, Cbr', description='carbonyl reductase 1', other_designations='15-hydroxyprostaglandin dehydrogenase|Carbonyl reductase [NADPH] 1 (NADPH-dependent carbonyl reductase 1)|NADPH-dependent carbonyl reductase 1|prostaglandin 9-ketoreductase|prostaglandin-E(2) 9-reductase', map_location='16 54.53 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93607837, end_position_on_the_genomic_accession=93610349, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12409, CurrentID=0, Status='live', Symbol='Cbr2', Aliases='MLCR', description='carbonyl reductase 2', other_designations='AP27|LCR|NADPH-dependent carbonyl reductase 2|adipocyte protein P27|lung carbonyl reductase', map_location='11 84.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120729485, end_position_on_the_genomic_accession=120732025, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12411, CurrentID=0, Status='live', Symbol='Cbs', Aliases='AI047524, AI303044, HIP4', description='cystathionine beta-synthase', other_designations='beta-thionase|serine sulfhydrase', map_location='17 A-C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31612623, end_position_on_the_genomic_accession=31637233, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  667772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667772, CurrentID=0, Status='live', Symbol='Myh15', Aliases='EG667772', description='myosin, heavy chain 15', other_designations='', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=49057486, end_position_on_the_genomic_accession=49199104, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  12416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12416, CurrentID=0, Status='live', Symbol='Cbx2', Aliases='M33, MOD2, pc', description='chromobox 2', other_designations='M33 polycomb-like protein|chromobox homolog 2|modifier 3 protein|polycomb', map_location='11 83.33 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119023019, end_position_on_the_genomic_accession=119031274, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12417, CurrentID=0, Status='live', Symbol='Cbx3', Aliases='HP1g, M32', description='chromobox 3', other_designations='HP1 gamma|chromobox homolog 3|heterochromatin protein 1 gamma|heterochromatin protein 1 homolog gamma|modifier 2 protein', map_location='6 B-C', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=51470616, end_position_on_the_genomic_accession=51483704, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12418, CurrentID=0, Status='live', Symbol='Cbx4', Aliases='MPc2, PC2', description='chromobox 4', other_designations='chromobox homolog 4|chromobox protein homolog 4|polycomb 2 homolog', map_location='11 83.34 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119077571, end_position_on_the_genomic_accession=119086237, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12419, CurrentID=0, Status='live', Symbol='Cbx5', Aliases='2610029O15Rik, C75991, HP1, Hp1a, Hp1alpha', description='chromobox 5', other_designations='HP1 alpha|chromobox homolog 5|heterochromatin protein 1 alpha|heterochromatin protein 1 homolog alpha|mHP1(alpha)', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103191546, end_position_on_the_genomic_accession=103239904, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12420, CurrentID=0, Status='live', Symbol='cby', Aliases='', description='chubby', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12421, CurrentID=0, Status='live', Symbol='Rb1cc1', Aliases='2900055E04Rik, 5930404L04Rik, Cc1, FIP200, LaXp180', description='RB1-inducible coiled-coil 1', other_designations='200 kDa FAK family kinase-interacting protein|FAK family kinase-interacting protein of 200 kDa|coiled coil forming protein 1|coiled-coil-forming protein 1', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=6214642, end_position_on_the_genomic_accession=6276372, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  12424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12424, CurrentID=0, Status='live', Symbol='Cck', Aliases='', description='cholecystokinin', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121489824, end_position_on_the_genomic_accession=121495694, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12425, CurrentID=0, Status='live', Symbol='Cckar', Aliases='AW106902', description='cholecystokinin A receptor', other_designations='CCK-A receptor|CCK-AR|CCK1-R|cholecystokinin-1 receptor', map_location='5 29.52 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=53698485, end_position_on_the_genomic_accession=53707704, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12426, CurrentID=0, Status='live', Symbol='Cckbr', Aliases='CCK-BR, CCK2-R, CCK2R, CCKR-2', description='cholecystokinin B receptor', other_designations='CCK-B receptor|CCK-B/gastrin receptor|CCK2/gastrin receptor|cholecystokinin-2 receptor', map_location='7 55.86 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105425558, end_position_on_the_genomic_accession=105436339, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12427, CurrentID=0, Status='live', Symbol='Ccna1', Aliases='', description='cyclin A1', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=55045469, end_position_on_the_genomic_accession=55055330, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12428, CurrentID=0, Status='live', Symbol='Ccna2', Aliases='AA408589, Ccn-1, Ccn1, Ccna, CycA2, Cyca', description='cyclin A2', other_designations='Cyclin A2 (Cyclin A)|cyclin-A', map_location='3 17.67 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36564865, end_position_on_the_genomic_accession=36571996, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12442, CurrentID=0, Status='live', Symbol='Ccnb2', Aliases='CycB2', description='cyclin B2', other_designations='G2/mitotic-specific cyclin B2', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70407689, end_position_on_the_genomic_accession=70421554, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  667803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667803, CurrentID=0, Status='live', Symbol='C920025E04Rik', Aliases='H2-T11', description='RIKEN cDNA C920025E04 gene', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36109030, end_position_on_the_genomic_accession=36111729, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12444, CurrentID=0, Status='live', Symbol='Ccnd2', Aliases='2600016F06Rik, AI256817, BF642806, C86853, Vin-1, Vin1, cD2', description='cyclin D2', other_designations='', map_location='6 61.92 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127125162, end_position_on_the_genomic_accession=127152188, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12445, CurrentID=0, Status='live', Symbol='Ccnd3', Aliases='9230106B05Rik, AA682053, AL024085, AW146355, C78795', description='cyclin D3', other_designations='', map_location='17 23.37 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47504738, end_position_on_the_genomic_accession=47599689, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12447, CurrentID=0, Status='live', Symbol='Ccne1', Aliases='AW538188, CycE1', description='cyclin E1', other_designations='', map_location='7 25.35 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=38097984, end_position_on_the_genomic_accession=38107490, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12448, CurrentID=0, Status='live', Symbol='Ccne2', Aliases='', description='cyclin E2', other_designations='', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11189214, end_position_on_the_genomic_accession=11204779, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12449, CurrentID=0, Status='live', Symbol='Ccnf', Aliases='CycF, Fbxo1', description='cyclin F', other_designations='G2/mitotic-specific cyclin-F', map_location='17 12.32 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24223230, end_position_on_the_genomic_accession=24251409, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12450, CurrentID=0, Status='live', Symbol='Ccng1', Aliases='AI314029', description='cyclin G1', other_designations='cyclin-G', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=40748552, end_position_on_the_genomic_accession=40755315, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12452, CurrentID=0, Status='live', Symbol='Ccng2', Aliases='', description='cyclin G2', other_designations='', map_location='5 E3.3-F1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=93267573, end_position_on_the_genomic_accession=93276231, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12453, CurrentID=0, Status='live', Symbol='Ccni', Aliases='', description='cyclin I', other_designations='', map_location='5 E3.3-F1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=93181933, end_position_on_the_genomic_accession=93206495, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12454, CurrentID=0, Status='live', Symbol='Ccnk', Aliases='AW123198, AW413594, CPR4, CycK', description='cyclin K', other_designations='', map_location='12 59.23 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108179446, end_position_on_the_genomic_accession=108203362, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12455, CurrentID=0, Status='live', Symbol='Ccnt1', Aliases='2810478G24Rik, AI115585, CycT1', description='cyclin T1', other_designations='cyclin-T', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98543211, end_position_on_the_genomic_accession=98568497, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12457, CurrentID=0, Status='live', Symbol='Noct', Aliases='AU043840, Ccr4, Ccrn4l', description='nocturnin', other_designations='CCR4 carbon catabolite repression 4-like|CCR4 protein homolog|carbon catabolite repression 4 homolog|carbon catabolite repression 4-like protein|circadian deadenylase NOC', map_location='3 B-D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51224447, end_position_on_the_genomic_accession=51251654, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12458, CurrentID=0, Status='live', Symbol='Ccr6', Aliases='CC-CKR-6, CCR-6, Cmkbr6, KY411', description='chemokine (C-C motif) receptor 6', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8236043, end_position_on_the_genomic_accession=8257141, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12460, CurrentID=0, Status='live', Symbol='Ccs', Aliases='Ccsd', description='copper chaperone for superoxide dismutase', other_designations='superoxide dismutase copper chaperone', map_location='19 4.11 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4825366, end_position_on_the_genomic_accession=4839327, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12461, CurrentID=0, Status='live', Symbol='Cct2', Aliases='Cctb', description='chaperonin containing Tcp1, subunit 2 (beta)', other_designations='CCT-beta|TCP-1-beta|chaperonin subunit 2 (beta)', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117050997, end_position_on_the_genomic_accession=117063814, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12462, CurrentID=0, Status='live', Symbol='Cct3', Aliases='AL024092, Cctg, Tcp1-rs3, TriC-P5', description='chaperonin containing Tcp1, subunit 3 (gamma)', other_designations='CCT-gamma|TCP-1-gamma|chaperonin subunit 3 (gamma)|matricin', map_location='3 F', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88297135, end_position_on_the_genomic_accession=88321766, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  667823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667823, CurrentID=0, Status='live', Symbol='Trim5', Aliases='EG667823, Gm8833', description='tripartite motif-containing 5', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104263386, end_position_on_the_genomic_accession=104288421, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12464, CurrentID=0, Status='live', Symbol='Cct4', Aliases='2610204B21Rik, A45, C78323, Cctd', description='chaperonin containing Tcp1, subunit 4 (delta)', other_designations='CCT-delta|T complex protein 1, delta|TCP-1 delta|TCP-1-delta|chaperonin subunit 4 (delta)', map_location='11 14.25 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=22990593, end_position_on_the_genomic_accession=23003336, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12465, CurrentID=0, Status='live', Symbol='Cct5', Aliases='Ccte, TCPE, mKIAA0098', description='chaperonin containing Tcp1, subunit 5 (epsilon)', other_designations='CCT-epsilon|TCP-1-epsilon|chaperonin subunit 5 (epsilon)', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=31590813, end_position_on_the_genomic_accession=31601804, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12466, CurrentID=0, Status='live', Symbol='Cct6a', Aliases='CCT-zeta, CCT-zeta-1, Cct6, Cctz-1, TCP-1-zeta', description='chaperonin containing Tcp1, subunit 6a (zeta)', other_designations='chaperonin containing TCP-1|chaperonin subunit 6a (zeta)', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129844962, end_position_on_the_genomic_accession=129846443, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12467, CurrentID=0, Status='live', Symbol='Cct6b', Aliases='CCT-zeta-2, CCTzeta-2, Cctz-2, TCP-1-zeta-2', description='chaperonin containing Tcp1, subunit 6b (zeta)', other_designations='chaperonin subunit 6b (zeta)', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82719248, end_position_on_the_genomic_accession=82764321, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  12468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12468, CurrentID=0, Status='live', Symbol='Cct7', Aliases='AA408524, AL022769, Ccth, Cctz', description='chaperonin containing Tcp1, subunit 7 (eta)', other_designations='CCT-eta|TCP-1-eta|chaperonin containing t-complex subunit eta|chaperonin subunit 7 (eta)', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85451505, end_position_on_the_genomic_accession=85468477, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12469, CurrentID=0, Status='live', Symbol='Cct8', Aliases='AI132397, Cctq, Tcpq', description='chaperonin containing Tcp1, subunit 8 (theta)', other_designations='CCT-theta|TCP-1-theta|chaperonin subunit 8 (theta)', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87483325, end_position_on_the_genomic_accession=87495869, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12475, CurrentID=0, Status='live', Symbol='Cd14', Aliases='', description='CD14 antigen', other_designations='myeloid cell-specific leucine-rich glycoprotein', map_location='18 19.46 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36725064, end_position_on_the_genomic_accession=36726815, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12476, CurrentID=0, Status='live', Symbol='Cd151', Aliases='PETA-3, SFA-1, Tspan24', description='CD151 antigen', other_designations='GP27|membrane glycoprotein SFA-1|platelet-endothelial tetraspan antigen 3', map_location='7 86.82 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141467362, end_position_on_the_genomic_accession=141471481, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12477, CurrentID=0, Status='live', Symbol='Ctla4', Aliases='Cd152, Ctla-4, Ly-56', description='cytotoxic T-lymphocyte-associated protein 4', other_designations='CD152 antigen|cytotoxic T-lymphocyte-associated antigen 4', map_location='1 C', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60909025, end_position_on_the_genomic_accession=60915832, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  667838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667838, CurrentID=0, Status='live', Symbol='Naip3', Aliases='Birc1c, EG667838, Mitf, Naip-rs5, pORF1, pORF2', description='NLR family, apoptosis inhibitory protein 3', other_designations='baculoviral IAP repeat-containing 1c|neuronal apoptosis inhibitory protein 3|neuronal apoptosis inhibitory protein related sequence 5', map_location='13 53.13 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100340005, end_position_on_the_genomic_accession=100389242, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  12479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12479, CurrentID=0, Status='live', Symbol='Cd1d1', Aliases='AI747460, CD1.1, Cd1a, Cd1d, Ly-38', description='CD1d1 antigen', other_designations='T-cell surface glycoprotein CD1d1', map_location='3 37.91 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=86995834, end_position_on_the_genomic_accession=86999340, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12480, CurrentID=0, Status='live', Symbol='Cd1d2', Aliases='CD1.2, Cd1b, Ly-38', description='CD1d2 antigen', other_designations='T-cell surface glycoprotein CD1d2', map_location='3 37.88 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=86986586, end_position_on_the_genomic_accession=86989533, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12481, CurrentID=0, Status='live', Symbol='Cd2', Aliases='LFA-2, Ly-37, Ly37', description='CD2 antigen', other_designations='LFA-3 receptor|T-cell surface antigen T11/Leu-5|lymphocyte antigen 37', map_location='3 44.17 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=101275908, end_position_on_the_genomic_accession=101287939, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12482, CurrentID=0, Status='live', Symbol='Ms4a1', Aliases='AA960661, Cd20, Ly-44, Ms4a2', description='membrane-spanning 4-domains, subfamily A, member 1', other_designations='B-cell differentiation antigen Ly-44|CD20 antigen|lymphocyte antigen 44|membrane-spanning 4-domains, subfamily A, member 2', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11250603, end_position_on_the_genomic_accession=11266151, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12483, CurrentID=0, Status='live', Symbol='Cd22', Aliases='A530093D23, Lyb-8, Lyb8', description='CD22 antigen', other_designations='B-lymphocyte cell adhesion molecule|BL-CAM|T-cell surface antigen Leu-14|sialic acid-binding Ig-like lectin 2|siglec-2', map_location='7 19.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30865404, end_position_on_the_genomic_accession=30880342, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12484, CurrentID=0, Status='live', Symbol='Cd24a', Aliases='Cd24, HSA, Ly-52, nectadrin', description='CD24a antigen', other_designations='M1/69-J11D heat stable antigen|R13-Ag|X62 heat stable antigen|cluster of differentiation 24|heat stable antigen|lymphocyte antigen 52', map_location='10 23.01 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43579169, end_position_on_the_genomic_accession=43584265, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12487, CurrentID=0, Status='live', Symbol='Cd28', Aliases='', description='CD28 antigen', other_designations='', map_location='1 C1-C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60746388, end_position_on_the_genomic_accession=60773359, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12488, CurrentID=0, Status='live', Symbol='Cd2ap', Aliases='AL024079, C78928, METS-1, Mets1', description='CD2-associated protein', other_designations='SH3 domain-containing adapter protein|SH3 domain-containing adaptor protein METS-1|mesenchyme to epithelium transition protein, SH3 domains|mesenchyme-to-epithelium transition protein with SH3 domains 1', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=42749098, end_position_on_the_genomic_accession=42876712, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  12489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12489, CurrentID=0, Status='live', Symbol='Cd33', Aliases='Siglec-3, gp67', description='CD33 antigen', other_designations='sialic acid-binding Ig-like lectin 3', map_location='7 28.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43527399, end_position_on_the_genomic_accession=43544428, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12490, CurrentID=0, Status='live', Symbol='Cd34', Aliases='AU040960', description='CD34 antigen', other_designations='cluster designation 34', map_location='1 98.38 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=194938821, end_position_on_the_genomic_accession=194961292, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12491, CurrentID=0, Status='live', Symbol='Cd36', Aliases='FAT, GPIV, Scarb3', description='CD36 antigen', other_designations='GPIIIB|PAS IV|PAS-4|fatty acid translocase|glycoprotein IIIb|platelet glycoprotein IV', map_location='5 8.11 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=17781690, end_position_on_the_genomic_accession=17888959, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  12492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12492, CurrentID=0, Status='live', Symbol='Scarb2', Aliases='9330185J12Rik, Cd36l2, LGP85, LIMP II, LIMP-2, MLGP85', description='scavenger receptor class B, member 2', other_designations='85 kDa lysosomal membrane sialoglycoprotein|CD36 antigen (collagen type I receptor, thrombospondin receptor)-like 2|lysosome membrane protein II', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92443873, end_position_on_the_genomic_accession=92505608, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12493, CurrentID=0, Status='live', Symbol='Cd37', Aliases='Tspan26', description='CD37 antigen', other_designations='', map_location='7 29.2 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45233632, end_position_on_the_genomic_accession=45239235, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12494, CurrentID=0, Status='live', Symbol='Cd38', Aliases='ADPRC 1-rs1, I-19, Cd38', description='CD38 antigen', other_designations="2'-phospho-ADP-ribosyl cyclase|2'-phospho-ADP-ribosyl cyclase/2'-phospho-cyclic-ADP-ribose transferase|2'-phospho-cyclic-ADP-ribose transferase|ADP-ribosyl cyclase 1|NIM-R5 antigen|cADPr hydrolase 1|cyclic ADP-ribose hydrolase 1", map_location='5 23.85 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43868827, end_position_on_the_genomic_accession=43912374, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12495, CurrentID=0, Status='live', Symbol='Entpd1', Aliases='2610206B08Rik, AA408691, Cd39, NTPDase-1', description='ectonucleoside triphosphate diphosphohydrolase 1', other_designations='CD39 antigen|NTPDase 1|ecto-ATP diphosphohydrolase 1|ecto-ATPDase 1|ecto-ATPase 1|ectoapyrase|lymphoid cell activation antigen', map_location='19 34.25 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40612365, end_position_on_the_genomic_accession=40741602, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12496, CurrentID=0, Status='live', Symbol='Entpd2', Aliases='Cd39l1, NTPDase2', description='ectonucleoside triphosphate diphosphohydrolase 2', other_designations='CD39 antigen-like 1|NTPDase 2|ecto-ATP diphosphohydrolase 2|ecto-ATPDase 2|ecto-ATPase 2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25395874, end_position_on_the_genomic_accession=25401323, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12497, CurrentID=0, Status='live', Symbol='Entpd6', Aliases='2700026H11Rik, Cd39l2, NTPDase-6', description='ectonucleoside triphosphate diphosphohydrolase 6', other_designations='CD39 antigen-like 2', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150749038, end_position_on_the_genomic_accession=150771675, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12499, CurrentID=0, Status='live', Symbol='Entpd5', Aliases='AI196558, AI987697, Cd39l4, ER-UDPase, NTPDase-5, NTPDase5, Pcph, mNTPase', description='ectonucleoside triphosphate diphosphohydrolase 5', other_designations='CD39 antigen-like 4|GDPase ENTPD5|NTPDase 5|UDPase ENTPD5|guanosine-diphosphatase ENTPD5|nucleoside diphosphatase|uridine-diphosphatase ENTPD5', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84373875, end_position_on_the_genomic_accession=84409029, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12500, CurrentID=0, Status='live', Symbol='Cd3d', Aliases='T3d', description='CD3 antigen, delta polypeptide', other_designations='T-cell receptor T3 delta chain', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44981769, end_position_on_the_genomic_accession=44987052, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12501, CurrentID=0, Status='live', Symbol='Cd3e', Aliases='AI504783, CD3, CD3epsilon, T3e', description='CD3 antigen, epsilon polypeptide', other_designations='T-cell surface antigen T3/Leu-4 epsilon chain', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44998735, end_position_on_the_genomic_accession=45009662, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12502, CurrentID=0, Status='live', Symbol='Cd3g', Aliases='Ctg-3, Ctg3, T3g', description='CD3 antigen, gamma polypeptide', other_designations='T-cell receptor T3 gamma chain', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44969572, end_position_on_the_genomic_accession=44980431, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12503, CurrentID=0, Status='live', Symbol='Cd247', Aliases='4930549J05Rik, A430104F18Rik, AW552088, Cd3, Cd3-eta, Cd3-zeta, Cd3h, Cd3z, Cd3zeta, T3z, Tcrk, Tcrz', description='CD247 antigen', other_designations='CD3 antigen, zeta polypeptide|CD3-zeta/eta|T-cell receptor T3 eta chain|T-cell surface glycoprotein CD3 zeta chain', map_location='1 G-H', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165780357, end_position_on_the_genomic_accession=165871153, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12504, CurrentID=0, Status='live', Symbol='Cd4', Aliases='L3T4, Ly-4', description='CD4 antigen', other_designations='T-cell differentiation antigen L3T4|T-cell surface antigen T4/Leu-3|T-cell surface glycoprotein CD4 precursor (T-cell surface antigen T4/Leu-3) (T-cell differentiation antigen L3T4)', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124864693, end_position_on_the_genomic_accession=124888248, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  667865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667865, CurrentID=0, Status='live', Symbol='Igkv6-17', Aliases='Igk-V19-17', description='immunoglobulin kappa variable 6-17', other_designations='Ig kappa chain V-V region MPC11|Ig kappa chain V19-17|V kappa 19-17|immunoglobulin kappa chain variable 19 (V19)-17', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70371469, end_position_on_the_genomic_accession=70371993, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  12506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12506, CurrentID=0, Status='live', Symbol='Cd48', Aliases='AI449234, AW610730, BCM1, BLAST, BLAST-1, BLAST1, Bcm-1, MEM-102, SLAMF2, Sgp-60', description='CD48 antigen', other_designations='BCM1 surface antigen|HM48-1|MRC OX-45 surface antigen|SLAM family member 2|signaling lymphocytic activation molecule 2', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171681421, end_position_on_the_genomic_accession=171705257, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12507, CurrentID=0, Status='live', Symbol='Cd5', Aliases='Ly-1, Ly-12, Ly-A, Lyt-1', description='CD5 antigen', other_designations='lymphocyte antigen 1', map_location='19 7.16 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10718107, end_position_on_the_genomic_accession=10739026, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12508, CurrentID=0, Status='live', Symbol='Cd53', Aliases='AI323659, Ox-44, Tspan25', description='CD53 antigen', other_designations='cell surface glycoprotein CD53', map_location='3 46.53 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106758861, end_position_on_the_genomic_accession=106790149, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12509, CurrentID=0, Status='live', Symbol='Cd59a', Aliases='AA987121, Cd59, protectin', description='CD59a antigen', other_designations='CD59 antigen|MAC-IP|MAC-inhibitory protein|MACIF|membrane attack complex inhibition factor', map_location='2 54.53 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104095801, end_position_on_the_genomic_accession=104115413, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12511, CurrentID=0, Status='live', Symbol='Cd6', Aliases='', description='CD6 antigen', other_designations='', map_location='19 7.16 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10789339, end_position_on_the_genomic_accession=10830059, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12512, CurrentID=0, Status='live', Symbol='Cd63', Aliases='C75951, ME491, Tspan30', description='CD63 antigen', other_designations='melanoma 1 antigen', map_location='10 77.19 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128908919, end_position_on_the_genomic_accession=128912822, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12514, CurrentID=0, Status='live', Symbol='Cd68', Aliases='Lamp4, Scard1, gp110', description='CD68 antigen', other_designations='', map_location='11 42.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69664213, end_position_on_the_genomic_accession=69666170, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12515, CurrentID=0, Status='live', Symbol='Cd69', Aliases='5830438K24Rik, AI452015, AIM, VEA', description='CD69 antigen', other_designations='Very Early Activation Antigen', map_location='6 63.21 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129267324, end_position_on_the_genomic_accession=129275369, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12516, CurrentID=0, Status='live', Symbol='Cd7', Aliases='', description='CD7 antigen', other_designations='', map_location='11 84.85 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121036747, end_position_on_the_genomic_accession=121039478, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12517, CurrentID=0, Status='live', Symbol='Cd72', Aliases='CD72c, Ly-19, Ly-32, Ly-m19, Lyb-2', description='CD72 antigen', other_designations='lymphocyte antigen 32', map_location='4 23.04 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43447724, end_position_on_the_genomic_accession=43455468, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12518, CurrentID=0, Status='live', Symbol='Cd79a', Aliases='Ig-alpha, Iga, Igalpha, Ly-54, Ly54, mb-1', description='CD79A antigen (immunoglobulin-associated alpha)', other_designations='Ig alpha|MB-1 membrane glycoprotein|immunoglobulin-associated alpha|membrane-bound immunoglobulin-associated protein|surface IgM-associated protein', map_location='7 13.49 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24897511, end_position_on_the_genomic_accession=24902197, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12519, CurrentID=0, Status='live', Symbol='Cd80', Aliases='B71, Cd28l, Ly-53, Ly53, MIC17, TSA1', description='CD80 antigen', other_designations='B7 protein|activation B7-1 antigen', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38455568, end_position_on_the_genomic_accession=38498546, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12520, CurrentID=0, Status='live', Symbol='Cd81', Aliases='Tapa-1, Tapa1, Tspan28', description='CD81 antigen', other_designations='26 kDa cell surface protein TAPA-1|target of the antiproliferative antibody 1', map_location='7 88.1 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143052750, end_position_on_the_genomic_accession=143067930, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12521, CurrentID=0, Status='live', Symbol='Cd82', Aliases='AA682076, AL023070, C33, IA4, Kai1, Tspan27', description='CD82 antigen', other_designations='C33 antigen|inducible membrane protein R2|kangai 1 (suppression of tumorigenicity 6, prostate)|metastasis suppressor Kangai-1 homolog', map_location='2 51.62 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93419102, end_position_on_the_genomic_accession=93462950, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  667882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667882, CurrentID=0, Status='live', Symbol='Usp-ps', Aliases='Usp17, Usp17-3', description='ubiquitin specific peptidase, pseudogene (USP17 homolog)', other_designations='ubiquitin specific peptidase 17 (homolog)|ubiquitin specific protease 17', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104647249, end_position_on_the_genomic_accession=104653154, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  12523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12523, CurrentID=0, Status='live', Symbol='Cd84', Aliases='A130013D22Rik, CDw84, SLAMF5', description='CD84 antigen', other_designations='leukocyte differentiation antigen CD84|signaling lymphocytic activation molecule 5', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171839697, end_position_on_the_genomic_accession=171890718, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12524, CurrentID=0, Status='live', Symbol='Cd86', Aliases='B7, B7-2, B7.2, B70, CLS1, Cd28l2, ETC-1, Ly-58, Ly58, MB7, MB7-2, TS/A-2', description='CD86 antigen', other_designations='activation B7-2 antigen|early T cell costimulatory molecule-1|early T-cell costimulatory molecule 1', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36568956, end_position_on_the_genomic_accession=36666077, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12525, CurrentID=0, Status='live', Symbol='Cd8a', Aliases='BB154331, Ly-2, Ly-35, Ly-B, Lyt-2', description='CD8 antigen, alpha chain', other_designations='Lyt-2.1 lymphocyte differentiation antigen (AA at 100)|T-cell surface glycoprotein Lyt-2', map_location='6 C', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71352772, end_position_on_the_genomic_accession=71379171, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12526, CurrentID=0, Status='live', Symbol='Cd8b1', Aliases='Cd8b, Ly-3, Ly-C, Lyt-3', description='CD8 antigen, beta chain 1', other_designations='T-cell membrane glycoprotein Ly-3|T-cell surface glycoprotein Lyt-3|lymphocyte antigen 3', map_location='6 32.14 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71322812, end_position_on_the_genomic_accession=71337451, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12527, CurrentID=0, Status='live', Symbol='Cd9', Aliases='Tspan29', description='CD9 antigen', other_designations='', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125460266, end_position_on_the_genomic_accession=125494755, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12530, CurrentID=0, Status='live', Symbol='Cdc25a', Aliases='D9Ertd393e', description='cell division cycle 25A', other_designations='dual specificity phosphatase Cdc25A', map_location='9 59.81 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109875337, end_position_on_the_genomic_accession=109893895, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12531, CurrentID=0, Status='live', Symbol='Cdc25b', Aliases='AI604853', description='cell division cycle 25B', other_designations='dual specificity phosphatase Cdc25B', map_location='2 63.29 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131186436, end_position_on_the_genomic_accession=131198511, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  12532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12532, CurrentID=0, Status='live', Symbol='Cdc25c', Aliases='Cdc25', description='cell division cycle 25C', other_designations='cell cycle phosphatase CDC25C|cell division cycle control protein 25C|dual specificity phosphatase Cdc25C', map_location='18 18.71 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34732993, end_position_on_the_genomic_accession=34752130, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12534, CurrentID=0, Status='live', Symbol='Cdk1', Aliases='Cdc2, Cdc2a, p34<CDC2>', description='cyclin-dependent kinase 1', other_designations='cell cycle p34 CDC2 kinase protein|cell division control protein 2 homolog|cell division cycle 2 homolog A|cell division cycle control protein 2a|cell division protein kinase 1|p34 protein kinase', map_location='10 36.07 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=69336635, end_position_on_the_genomic_accession=69352912, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12537, CurrentID=0, Status='live', Symbol='Cdk11b', Aliases='AA989746, CDK11-p110, CDK11-p46, CDK11-p58, Cdc11b, Cdc2l1, Cdc2l2, Cdk11, p58', description='cyclin-dependent kinase 11B', other_designations='PITSLRE proteins|PITSLRE serine/threonine-protein kinase CDC2L1|cell division cycle 2 homolog-like 1|cell division cycle 2 homolog-like 2|cell division cycle 2-like 1|cell division cycle 2-like 2|cell division cycle 2-like protein kinase 1|cell division protein kinase 11|cyclin-dependent kinase 11|galactosyltransferase-associated protein kinase p58/GTA', map_location='4 86.73 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155624846, end_position_on_the_genomic_accession=155649935, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  12539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12539, CurrentID=0, Status='live', Symbol='Cdc37', Aliases='p50, p50Cdc37', description='cell division cycle 37', other_designations='cell division cycle 37 homolog|hsp90 chaperone protein kinase-targeting subunit', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21138541, end_position_on_the_genomic_accession=21149977, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12540, CurrentID=0, Status='live', Symbol='Cdc42', Aliases='AI747189, AU018915', description='cell division cycle 42', other_designations='G25K GTP-binding protein|cell division cycle 42 homolog', map_location='4 69.83 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137319696, end_position_on_the_genomic_accession=137357759, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12544, CurrentID=0, Status='live', Symbol='Cdc45', Aliases='Cdc45l', description='cell division cycle 45', other_designations='CDC45-related protein|PORC-PI-1|cell division cycle 45 homolog', map_location='16 11.63 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18780447, end_position_on_the_genomic_accession=18811972, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12545, CurrentID=0, Status='live', Symbol='Cdc7', Aliases='AI597260l1, muCdc7, Cdc7', description='cell division cycle 7 (S. cerevisiae)', other_designations='CDC7-related kinase|cell division cycle 7-like 1', map_location='5 E', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=106964315, end_position_on_the_genomic_accession=106984440, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12549, CurrentID=0, Status='live', Symbol='Arhgap31', Aliases='5830477L08Rik, AU041750, Cdgap, D330026I07Rik', description='Rho GTPase activating protein 31', other_designations='CDC42 GTPase-activating protein', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38598343, end_position_on_the_genomic_accession=38713035, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12550, CurrentID=0, Status='live', Symbol='Cdh1', Aliases='AA960649, ARC-1, E-cad, Ecad, L-CAM, UVO, Um', description='cadherin 1', other_designations='E-cadherin|epithelial cadherin|uvomorulin', map_location='8 D', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106603368, end_position_on_the_genomic_accession=106670246, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  12552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12552, CurrentID=0, Status='live', Symbol='Cdh11', Aliases='Cad11, OSF-4', description='cadherin 11', other_designations='OB-cadherin|osteoblast cadherin|osteoblast-cadherin', map_location='8 50.44 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=102632095, end_position_on_the_genomic_accession=102785480, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12554, CurrentID=0, Status='live', Symbol='Cdh13', Aliases='4932416G01Rik, Cdht, Tcad', description='cadherin 13', other_designations='H-cadherin|T-cadherin|heart cadherin', map_location='8 65.97 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=118032249, end_position_on_the_genomic_accession=119324929, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  12555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12555, CurrentID=0, Status='live', Symbol='Cdh15', Aliases='AI323380, Cdh14, Mcad', description='cadherin 15', other_designations='M-cadherin|cadherin-14|muscle cadherin', map_location='8 71.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122848374, end_position_on_the_genomic_accession=122867397, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12556, CurrentID=0, Status='live', Symbol='Cdh16', Aliases='', description='cadherin 16', other_designations='KSP-cadherin|kidney-specific cadherin', map_location='8 53.04 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104601915, end_position_on_the_genomic_accession=104624396, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  12557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12557, CurrentID=0, Status='live', Symbol='Cdh17', Aliases='HPT-1, HPT-1/LI', description='cadherin 17', other_designations='BILL-cadherin|LI-cadherin|P130|liver-intestine cadherin', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11758157, end_position_on_the_genomic_accession=11817905, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  12558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12558, CurrentID=0, Status='live', Symbol='Cdh2', Aliases='CDHN, N-CAD, Ncad', description='cadherin 2', other_designations='N-cadherin|neural cadherin', map_location='18 10.1 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=16588877, end_position_on_the_genomic_accession=16809049, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12560, CurrentID=0, Status='live', Symbol='Cdh3', Aliases='AI385538, Cadp, Cdhp, P-cadherin, Pcad', description='cadherin 3', other_designations='RPE-specific cadherin|placental cadherin', map_location='8 53.16 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106510852, end_position_on_the_genomic_accession=106556911, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  12561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12561, CurrentID=0, Status='live', Symbol='Cdh4', Aliases='AW120700, R-CAD, R-Cadh, Rcad', description='cadherin 4', other_designations='R-cadherin|retina cadherin|retinal cadherin', map_location='2 101.86 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=179442478, end_position_on_the_genomic_accession=179899377, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  12562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12562, CurrentID=0, Status='live', Symbol='Cdh5', Aliases='7B4, AA408225, Cd144, VE-Cad, VECD, VEcad, Vec', description='cadherin 5', other_designations='7B4/cadherin-5|VE-cadherin|vascular endothelial cadherin', map_location='8 53.04 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104101613, end_position_on_the_genomic_accession=104144504, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12563, CurrentID=0, Status='live', Symbol='Cdh6', Aliases='cad6', description='cadherin 6', other_designations='K-cadherin|kidney cadherin', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=13028699, end_position_on_the_genomic_accession=13173689, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  667924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667924, CurrentID=0, Status='live', Symbol='Igkv3-10', Aliases='IgK21-10, Igk-V21-10', description='immunoglobulin kappa variable 3-10', other_designations='Ig kappa chain V-III region MOPC 63|V-kappa-21B|immunoglobulin kappa chain variable 21 (V21)-10|immunoglobulin kappa chain variable 3-10', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70572633, end_position_on_the_genomic_accession=70573230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  12565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12565, CurrentID=0, Status='live', Symbol='Cdh9', Aliases='', description='cadherin 9', other_designations='T1-cadherin', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=16728549, end_position_on_the_genomic_accession=16857097, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12566, CurrentID=0, Status='live', Symbol='Cdk2', Aliases='A630093N05Rik', description='cyclin-dependent kinase 2', other_designations='cell division protein kinase 2', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128697938, end_position_on_the_genomic_accession=128705051, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12567, CurrentID=0, Status='live', Symbol='Cdk4', Aliases='Crk3', description='cyclin-dependent kinase 4', other_designations='PSK-J3|cell division protein kinase 4|cyclin dependent kinase 4|p34<PSK-J3>/cdk4|serine/threonine kinase', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127063537, end_position_on_the_genomic_accession=127067288, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12568, CurrentID=0, Status='live', Symbol='Cdk5', Aliases='AW048668, Crk6', description='cyclin-dependent kinase 5', other_designations='CR6 protein kinase|TPKII catalytic subunit|cell division protein kinase 5|serine/threonine-protein kinase PSSALRE|tau protein kinase II catalytic subunit', map_location='5 11.73 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24407597, end_position_on_the_genomic_accession=24423591, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12569, CurrentID=0, Status='live', Symbol='Cdk5r1', Aliases='Cdk5r, D11Bwg0379e, p25, p35', description='cyclin-dependent kinase 5, regulatory subunit 1 (p35)', other_designations='CDK5 activator 1|TPKII regulatory subunit|cyclin-dependent kinase 5 regulatory subunit 1|cyclin-dependent kinase 5, regulatory subunit (p35) 1', map_location='11 47.94 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80477046, end_position_on_the_genomic_accession=80481179, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12570, CurrentID=0, Status='live', Symbol='Cdk5r2', Aliases='B230310J22Rik, p39', description='cyclin-dependent kinase 5, regulatory subunit 2 (p39)', other_designations='CDK5 activator 2|cyclin-dependent kinase 5 regulatory subunit 2|p39I', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74855029, end_position_on_the_genomic_accession=74857732, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12571, CurrentID=0, Status='live', Symbol='Cdk6', Aliases='AI504062, Crk2', description='cyclin-dependent kinase 6', other_designations='CR2 protein kinase|cell division protein kinase 6|serine/threonine-protein kinase PLSTIRE', map_location='5 2.04 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3343891, end_position_on_the_genomic_accession=3531008, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12572, CurrentID=0, Status='live', Symbol='Cdk7', Aliases='AI323415, AI528512, C230069N13, Cdkn7, Crk4', description='cyclin-dependent kinase 7', other_designations='39 kDa protein kinase|CAK|CDK-activating kinase|CR4 protein kinase|CRK4 PK (CDC2-related-kinase-4 protein kinase)|P39 Mo15|TFIIH basal transcription factor complex kinase subunit|cell division protein kinase 7|cyclin-dependent kinase 7 (homolog of Xenopus MO15 cdk-activating kinase)|protein-tyrosine kinase MPK-7', map_location='13 53.23 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100697024, end_position_on_the_genomic_accession=100730942, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12575, CurrentID=0, Status='live', Symbol='Cdkn1a', Aliases='CAP20, CDKI, CIP1, Cdkn1, P21, SDI1, Waf1, mda6, p21Cip1, p21WAF', description='cyclin-dependent kinase inhibitor 1A (P21)', other_designations='CDK-interacting protein 1|melanoma differentiation-associated protein', map_location='17 15.12 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29090986, end_position_on_the_genomic_accession=29100722, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12576, CurrentID=0, Status='live', Symbol='Cdkn1b', Aliases='AA408329, AI843786, Kip1, p27, p27Kip1', description='cyclin-dependent kinase inhibitor 1B', other_designations='cyclin-dependent kinase inhibitor p27', map_location='6 65.77 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134920401, end_position_on_the_genomic_accession=134925525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12577, CurrentID=0, Status='live', Symbol='Cdkn1c', Aliases='AL024410, CDKI, Kip2, p57(kip2), p57Kip2', description='cyclin-dependent kinase inhibitor 1C (P57)', other_designations='cyclin-dependent kinase inhibitor p57', map_location='7 88.22 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143458339, end_position_on_the_genomic_accession=143461072, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12578, CurrentID=0, Status='live', Symbol='Cdkn2a', Aliases='ARF-INK4a, Arf, INK4a-ARF, Ink4a/Arf, MTS1, Pctr1, p16, p16(INK4a), p16INK4a, p19<ARF>, p19ARF', description='cyclin-dependent kinase inhibitor 2A', other_designations='CDK4I|alternative reading frame|cdkn2a|cyclin-dependent kinase 4 inhibitor A|cyclin-dependent kinase inhibitor 2A (p16, inhibits CDK4)|cyclin-dependent kinase inhibitor 2A, isoforms 1/2|cyclin-dependent kinase inhibitor protein|mitochondrial smARF|p16-INK4|p16-INK4a', map_location='4 C3-C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=89274473, end_position_on_the_genomic_accession=89294619, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12579, CurrentID=0, Status='live', Symbol='Cdkn2b', Aliases='AV083695, INK4b, MTS2, p15, p15(INK4b), p15INK4b', description='cyclin-dependent kinase inhibitor 2B (p15, inhibits CDK4)', other_designations='cyclin-dependent kinase inhibitor p15|cyclin-dependent kinase inhibitor p15INK4b|cyclin-dependent kinase inhibitor protein|p14-INK4b|p15-INK4b', map_location='4 C3-C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=89306289, end_position_on_the_genomic_accession=89311032, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  667940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667940, CurrentID=0, Status='live', Symbol='Igkv3-5', Aliases='Gm16774, IgK21-5, Igk-V21-5', description='immunoglobulin kappa chain variable 3-5', other_designations='V-kappa-21C|immunoglobulin kappa chain variable 21 (V21)-5', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70663298, end_position_on_the_genomic_accession=70663895, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  12581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12581, CurrentID=0, Status='live', Symbol='Cdkn2d', Aliases='INK4d, p19, p19INK4d', description='cyclin-dependent kinase inhibitor 2D (p19, inhibits CDK4)', other_designations='p19-INK4d', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21288459, end_position_on_the_genomic_accession=21291402, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12582, CurrentID=0, Status='live', Symbol='cdm', Aliases='', description='cadmium resistance', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12583, CurrentID=0, Status='live', Symbol='Cdo1', Aliases='1300002L19Rik, Cdo, D18Ucla3', description='cysteine dioxygenase 1, cytosolic', other_designations='CDO-I|cysteine dioxygenase type I|cytosolic cysteine dioxygenase 1', map_location='18 24.75 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46713205, end_position_on_the_genomic_accession=46728542, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12585, CurrentID=0, Status='live', Symbol='Cdr2', Aliases='AA617262', description='cerebellar degeneration-related 2', other_designations='', map_location='7 65.04 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120957036, end_position_on_the_genomic_accession=120982312, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12587, CurrentID=0, Status='live', Symbol='Mia', Aliases='CD-RAP, Cdrap1, Mia', description='melanoma inhibitory activity', other_designations='cartilage derived retinoic acid sensitive protein|cartilage-derived retinoic acid-sensitive protein|melanoma inhibitory activity 1|melanoma inhibitory activity protein|melanoma-inhibitory-activity', map_location='7 15.84 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27179741, end_position_on_the_genomic_accession=27181149, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12589, CurrentID=0, Status='live', Symbol='Ift81', Aliases='AW060663, Cdv-1, Cdv-1r, Cdv1', description='intraflagellar transport 81', other_designations='carnitine deficiency-associated gene expressed in ventricle 1|carnitine deficiency-associated protein expressed in ventricle 1|intraflagellar transport 81 homolog', map_location='5 62.44 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122540265, end_position_on_the_genomic_accession=122614546, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12590, CurrentID=0, Status='live', Symbol='Cdx1', Aliases='Cdx, Cdx-1', description='caudal type homeobox 1', other_designations='caudal type homeo box 1|caudal type homeobox transcription factor 1|caudal-type homeobox protein 1', map_location='18 34.41 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61018862, end_position_on_the_genomic_accession=61036199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12591, CurrentID=0, Status='live', Symbol='Cdx2', Aliases='Cdx-2', description='caudal type homeobox 2', other_designations='caudal type homeo box 2|caudal-type homeobox protein 2', map_location='5 G2-G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147300900, end_position_on_the_genomic_accession=147307249, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12592, CurrentID=0, Status='live', Symbol='Cdx4', Aliases='Cdx-3, Cdx-4, Cdx3', description='caudal type homeobox 4', other_designations='caudal type homeo box 4|caudal-type homeobox protein 4', map_location='X 46.06 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103321398, end_position_on_the_genomic_accession=103330224, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12593, CurrentID=0, Status='live', Symbol='Cdyl', Aliases='AI325931', description='chromodomain protein, Y chromosome-like', other_designations='CDY-like', map_location='13 14.39 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=35659811, end_position_on_the_genomic_accession=35874064, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12606, CurrentID=0, Status='live', Symbol='Cebpa', Aliases='C/ebpalpha, CBF-A, Cebp', description='CCAAT/enhancer binding protein (C/EBP), alpha', other_designations='C/EBP alpha', map_location='7 21.02 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35119293, end_position_on_the_genomic_accession=35121928, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12607, CurrentID=0, Status='live', Symbol='Cebpz', Aliases='AI848081, CBF2, Cbf, Cebpa-rs1', description='CCAAT/enhancer binding protein zeta', other_designations='CCAAT-binding factor|CCAAT-box-binding transcription factor|CCAAT/enhancer binding protein alpha (C/EBP) related sequence 1', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78919003, end_position_on_the_genomic_accession=78937070, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  12608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12608, CurrentID=0, Status='live', Symbol='Cebpb', Aliases='C/EBPbeta, CRP2, IL-6DBP, LAP, LIP, NF-IL6, NF-M, Nfil6', description='CCAAT/enhancer binding protein (C/EBP), beta', other_designations='AGP/EBP|C/EBP BETA|interleukin-6-dependent-binding protein|liver-enriched transcriptional activator|nuclear protein Il6', map_location='2 87.58 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167688915, end_position_on_the_genomic_accession=167690432, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12609, CurrentID=0, Status='live', Symbol='Cebpd', Aliases='c/EBPdelta', description='CCAAT/enhancer binding protein (C/EBP), delta', other_designations='C/EBP delta|c/EBP-related protein 3', map_location='16 10.09 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=15887286, end_position_on_the_genomic_accession=15889545, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12611, CurrentID=0, Status='live', Symbol='Cebpg', Aliases='C/EBP[g], C/EBPgamma, GPE1-BP, Gpe1bp, Ig/EBP', description='CCAAT/enhancer binding protein (C/EBP), gamma', other_designations='G-CSF promoter element 1 binding protein|GPE1-BP (C/EBP gamma)|GPE1-binding protein|IG/EBP-1|c/EBP gamma|granulocyte colony-stimulating factor promoter element 1-binding protein|immunoglobulin enhancer-binding protein 1', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35046422, end_position_on_the_genomic_accession=35056590, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12613, CurrentID=0, Status='live', Symbol='Cel', Aliases='1810036E18Rik, BAL', description='carboxyl ester lipase', other_designations='BSSL|bile salt-stimulated lipase|cholesterol esterase|pancreatic lysophospholipase|sterol esterase', map_location='2 19.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28555819, end_position_on_the_genomic_accession=28563362, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12614, CurrentID=0, Status='live', Symbol='Celsr1', Aliases='Adgrc1, Crsh, Scy, crash', description='cadherin, EGF LAG seven-pass G-type receptor 1', other_designations='cadherin, EGF LAG seven-pass G-type receptor 1 (flamingo homolog, Drosophila)|seven-pass transmembrane receptor', map_location='15 40.42 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85898758, end_position_on_the_genomic_accession=86034217, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  12615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12615, CurrentID=0, Status='live', Symbol='Cenpa', Aliases='Cenp-A', description='centromere protein A', other_designations='centromere autoantigen A|centrosomin A', map_location='5 16.76 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30666876, end_position_on_the_genomic_accession=30674837, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12616, CurrentID=0, Status='live', Symbol='Cenpb', Aliases='CENP-B', description='centromere protein B', other_designations='centromere autoantigen B', map_location='2 63.29 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131177289, end_position_on_the_genomic_accession=131180012, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12617, CurrentID=0, Status='live', Symbol='Cenpc1', Aliases='CENP-C, Cenpc', description='centromere protein C1', other_designations='CENP-C 1|CENTROMERE PROTEIN C (CENP-C) (CENTROMERE AUTOANTIGEN C)|centromere autoantigen C1', map_location='5 E2-E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86012018, end_position_on_the_genomic_accession=86066019, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12622, CurrentID=0, Status='live', Symbol='Cer1', Aliases='Cerl, Cerl1, Cerr1, cer-1', description='cerberus 1, DAN family BMP antagonist', other_designations='cer-l|cerberus 1 homolog|cerberus homolog|cerberus-like protein|cerberus-related protein', map_location='4 39.4 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=82879909, end_position_on_the_genomic_accession=82892647, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12623, CurrentID=0, Status='live', Symbol='Ces1g', Aliases='Ces-1, Ces1, Ses-1', description='carboxylesterase 1G', other_designations='ES-x|acyl coenzyme A:cholesterol acyltransferase|acyl-coenzyme A:cholesterol acyltransferase|carboxylesterase 1', map_location='8 45.29 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93302369, end_position_on_the_genomic_accession=93337209, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12626, CurrentID=0, Status='live', Symbol='Cetn3', Aliases='MmCEN3', description='centrin 3', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=81783247, end_position_on_the_genomic_accession=81797157, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12627, CurrentID=0, Status='live', Symbol='Cfc1', Aliases='AV265756, b2b970Clo, cryptic', description='cripto, FRL-1, cryptic family 1', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34535648, end_position_on_the_genomic_accession=34544311, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12628, CurrentID=0, Status='live', Symbol='Cfh', Aliases='Mud-1, NOM, Sas-1, Sas1', description='complement component factor h', other_designations='complement factor H related protein 3A4/5G4|protein beta-1-H', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=140085855, end_position_on_the_genomic_accession=140183411, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  12630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12630, CurrentID=0, Status='live', Symbol='Cfi', Aliases='', description='complement component factor i', other_designations='C3B/C4B inactivator', map_location='3 59.21 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129835889, end_position_on_the_genomic_accession=129875329, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12631, CurrentID=0, Status='live', Symbol='Cfl1', Aliases='AA959946, Cof', description='cofilin 1, non-muscle', other_designations='cofilin, non-muscle isoform|n-cofilin', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5490455, end_position_on_the_genomic_accession=5494031, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  667992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667992, CurrentID=0, Status='live', Symbol='Tas2r103', Aliases='EG667992, T2R3, TRB2, Tas2r10, Tas2r3, mGR03, mt2r63', description='taste receptor, type 2, member 103', other_designations='T2R10|T2R103|taste receptor family B member 2|taste receptor type 2 member 10', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=133036163, end_position_on_the_genomic_accession=133037101, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12633, CurrentID=0, Status='live', Symbol='Cflar', Aliases='2310024N18Rik, A430105C05Rik, CLARP, Cash, Casper, ENSMUSG00000072980, FLAME, FLAME-1, Flip, Gm9845, I-FLICE, MRIT, c-Flip', description='CASP8 and FADD-like apoptosis regulator', other_designations='FADD-like antiapoptotic molecule 1|MACH-related inducer of toxicity|caspase homolog|caspase-eight-related protein|caspase-like apoptosis regulatory protein|cellular FLICE-like inhibitory protein|inhibitor of FLICE|usurpin', map_location='1 29.16 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58711310, end_position_on_the_genomic_accession=58759209, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12638, CurrentID=0, Status='live', Symbol='Cftr', Aliases='AW495489, Abcc7', description='cystic fibrosis transmembrane conductance regulator', other_designations='ATP-binding cassette sub-family C member 7|ATP-binding cassette transporter sub-family C member 7|ATP-binding cassette, subfamily c, member 7|cAMP-dependent chloride channel|channel conductance-controlling ATPase|cystic fibrosis transmembrane conductance regulator homolog|cystic fibrosis transmembrane conductance regulator homolog; ATP-binding cassette, subfamily c, member 7', map_location='6 8.1 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=18170262, end_position_on_the_genomic_accession=18322774, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  12640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12640, CurrentID=0, Status='live', Symbol='Cga', Aliases='CG-alpha, FSHA, GPHA1, GPHalpha, HCG, LHA, Tsha, aGSU, alpha-GSU, alphaGSU, alphaSU', description='glycoprotein hormones, alpha subunit', other_designations='FSH-alpha|LSH-alpha|TSH-alpha|anterior pituitary glycoprotein hormones common subunit alpha|follicle-stimulating hormone alpha chain|follitropin alpha chain|luteinizing hormone alpha chain|lutropin alpha chain|thyroid-stimulating hormone alpha chain|thyrotropin alpha chain', map_location='4 16.86 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34893779, end_position_on_the_genomic_accession=34907374, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12642, CurrentID=0, Status='live', Symbol='Ch25h', Aliases='AI462618, m25OH', description='cholesterol 25-hydroxylase', other_designations='cholesterol 25-monooxygenase', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34473783, end_position_on_the_genomic_accession=34475135, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12643, CurrentID=0, Status='live', Symbol='Chad', Aliases='SLRR4A', description='chondroadherin', other_designations='cartilage leucine-rich protein', map_location='11 58.9 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94565074, end_position_on_the_genomic_accession=94569127, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  111334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111334, CurrentID=0, Status='live', Symbol='Cd1', Aliases='', description='CD1 antigen complex', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12647, CurrentID=0, Status='live', Symbol='Chat', Aliases='B230380D24Rik', description='choline acetyltransferase', other_designations='CHOACTase|choline acetylase', map_location='14 19.4 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32408203, end_position_on_the_genomic_accession=32465909, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  12648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12648, CurrentID=0, Status='live', Symbol='Chd1', Aliases='4930525N21Rik, AI851787, AW555109', description='chromodomain helicase DNA binding protein 1', other_designations='ATP-dependent helicase CHD1|CHD-1', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15704967, end_position_on_the_genomic_accession=15772612, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  12649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12649, CurrentID=0, Status='live', Symbol='Chek1', Aliases='C85740, Chk1, rad27', description='checkpoint kinase 1', other_designations='CHK1 checkpoint homolog|checkpoint kinase 1 homolog|checkpoint kinase-1|rad27 homolog', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=36708482, end_position_on_the_genomic_accession=36726658, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12651, CurrentID=0, Status='live', Symbol='Chkb', Aliases='Chetk, Chkl, Ck, Ck/Ek, Ck/Ek-beta, Ckb, Ckekb, Ek', description='choline kinase beta', other_designations='choline/ethanolamine kinase-b|choline/ethanolaminekinase|ethanolamine kinase', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89426349, end_position_on_the_genomic_accession=89429927, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12652, CurrentID=0, Status='live', Symbol='Chga', Aliases='ChrA, cgA', description='chromogranin A', other_designations='', map_location='12 51.66 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102554969, end_position_on_the_genomic_accession=102565027, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12653, CurrentID=0, Status='live', Symbol='Chgb', Aliases='Scg-1', description='chromogranin B', other_designations='cgB|chromogranin-B|secretogranin I|sgI', map_location='2 F-G', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132781278, end_position_on_the_genomic_accession=132795072, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12654, CurrentID=0, Status='live', Symbol='Chil1', Aliases='AW208766, Brp39, Chi3l1, Gp39', description='chitinase-like 1', other_designations='CGP-39|GP-39|breast regression protein 39|cartilage glycoprotein 39|chitinase 3-like 1', map_location='1 58.15 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134182164, end_position_on_the_genomic_accession=134190031, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12655, CurrentID=0, Status='live', Symbol='Chil3', Aliases='AI505981, Chi3l3, ECF-L, Ym1', description='chitinase-like 3', other_designations='beta-N-acetylhexosaminidase Ym1|eosinophil chemotactic cytokine|eosinophil chemotactic factor-L|secreted protein Ym1', map_location='3 46.49 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106147554, end_position_on_the_genomic_accession=106167564, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  242408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242408, CurrentID=0, Status='live', Symbol='Fam221b', Aliases='4930412F15Rik', description='family with sequence similarity 221, member B', other_designations='uncharacterized protein C9orf128 homolog', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43659622, end_position_on_the_genomic_accession=43668859, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12659, CurrentID=0, Status='live', Symbol='Ovgp1', Aliases='120kDa, AU016433, AU019448, Chit5, MOGP, OGP, muc9', description='oviductal glycoprotein 1', other_designations='chitinase 5|estrogen-dependent oviduct protein|mucin 9|oviduct glycoprotein OGP|oviductal glycoprotein 1, 120kD|oviductin', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105973802, end_position_on_the_genomic_accession=105987423, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12660, CurrentID=0, Status='live', Symbol='Chka', Aliases='CK, CK/EK-alpha, Chetk-alpha, Chk, ChoK, EK, EtnK-alpha', description='choline kinase alpha', other_designations='choline/ethanolamine kinase alpha|ethanolamine kinase', map_location='19 3.57 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3851585, end_position_on_the_genomic_accession=3894369, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12661, CurrentID=0, Status='live', Symbol='Chl1', Aliases='A530023M13Rik, AI465420, CALL, LICAM2', description='cell adhesion molecule L1-like', other_designations='cell adhesion molecule with homology to L1CAM|chl1-like protein|close homolog of L1', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=103510582, end_position_on_the_genomic_accession=103733038, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  12662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12662, CurrentID=0, Status='live', Symbol='Chm', Aliases='Rep-1', description='choroidermia (RAB escort protein 1)', other_designations='choroideraemia protein homolog|choroideremia protein homolog|rab escort protein 1', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=113040590, end_position_on_the_genomic_accession=113185539, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  12663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12663, CurrentID=0, Status='live', Symbol='Chml', Aliases='AI450814, E030003F13Rik, REP-2, Rep2', description='choroideremia-like', other_designations='choroideraemia-like protein|choroideremia-like protein|rab escort protein 2|rab escort protein-2', map_location='1 H5-H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175682237, end_position_on_the_genomic_accession=175688353, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12667, CurrentID=0, Status='live', Symbol='Chrd', Aliases='Chd', description='chordin', other_designations='', map_location='16 12.51 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20731417, end_position_on_the_genomic_accession=20742384, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  12669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12669, CurrentID=0, Status='live', Symbol='Chrm1', Aliases='Chrm-1, M1, M1R', description='cholinergic receptor, muscarinic 1, CNS', other_designations='AChR M1|m1 muscarinic acetylcholine receptor protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8664005, end_position_on_the_genomic_accession=8683606, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12671, CurrentID=0, Status='live', Symbol='Chrm3', Aliases='Chrm-3, M3, M3R', description='cholinergic receptor, muscarinic 3, cardiac', other_designations='AChR M3|mm3 mAChR|muscarinic acetylcholine receptor 3', map_location='13 3.72 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=9875488, end_position_on_the_genomic_accession=10360803, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12672, CurrentID=0, Status='live', Symbol='Chrm4', Aliases='Chrm-4, M4', description='cholinergic receptor, muscarinic 4', other_designations='AChR M4|mm4 mAChR|muscarinic acetylcholine receptor 4', map_location='2 50.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91919619, end_position_on_the_genomic_accession=91929835, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12675, CurrentID=0, Status='live', Symbol='Chuk', Aliases='AI2566581, Fbx24, Fbxo24, IKBKA, IKK1, Ikka, NFKBIKA, Chuk', description='conserved helix-loop-helix ubiquitous kinase', other_designations='I kappa B kinase 1|I-kappa-B kinase 1|I-kappa-B kinase alpha|IKK alpha|IKK-A|IkappaB kinase alpha|nuclear factor NF-kappa-B inhibitor kinase alpha', map_location='19 36.71 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44073334, end_position_on_the_genomic_accession=44107506, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  12677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12677, CurrentID=0, Status='live', Symbol='Vsx2', Aliases='Chx10, Hox-10, Hox10, or', description='visual system homeobox 2', other_designations='C. elegans ceh-10 homeo domain containing homolog|ceh-10 homeodomain-containing homolog|homeobox protein CHX10|ocular retardation', map_location='12 39.28 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84569828, end_position_on_the_genomic_accession=84595457, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12683, CurrentID=0, Status='live', Symbol='Cidea', Aliases='AW212747', description='cell death-inducing DNA fragmentation factor, alpha subunit-like effector A', other_designations='cell death-inducing DFFA-like effector A', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67343564, end_position_on_the_genomic_accession=67367794, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12684, CurrentID=0, Status='live', Symbol='Cideb', Aliases='1110030C18Rik, AI790179, CIDE-B', description='cell death-inducing DNA fragmentation factor, alpha subunit-like effector B', other_designations='DFFA-like B|cell death-inducing DFFA-like effector B', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55754052, end_position_on_the_genomic_accession=55758424, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12686, CurrentID=0, Status='live', Symbol='Elovl3', Aliases='CIN-2, Cig30', description='elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 3', other_designations='3-keto acyl-CoA synthase Elovl3|ELOVL FA elongase 3|ELOVL fatty acid elongase 3|cold inducible glycoprotein 30|cold-inducible glycoprotein of 30 kDa|elongation of very long chain fatty acids-like 3|very long chain 3-ketoacyl-CoA synthase 3|very long chain 3-oxoacyl-CoA synthase 3|very-long-chain 3-oxoacyl-CoA synthase 3', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46131899, end_position_on_the_genomic_accession=46135694, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  111341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111341, CurrentID=0, Status='live', Symbol='Ph', Aliases='', description='patch deletion region', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12695, CurrentID=0, Status='live', Symbol='Inadl', Aliases='Cipp, Patj', description='InaD-like (Drosophila)', other_designations='PALS1-associated tight junction protein|channel-interacting PDZ domain protein|channel-interacting PDZ domain-containing protein|inactivation no after-potential D-like protein|novel ankyrin-repeat protein|protein associated to tight junctions', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=98395534, end_position_on_the_genomic_accession=98719603, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  12696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12696, CurrentID=0, Status='live', Symbol='Cirbp', Aliases='Cirp, R74941', description='cold inducible RNA binding protein', other_designations='A18 hnRNP|cold inducible RNA-binding protein|glycine-rich RNA-binding protein CIRP', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80165990, end_position_on_the_genomic_accession=80172691, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12700, CurrentID=0, Status='live', Symbol='Cish', Aliases='CIS-1, CIS1, Cis, F17, F23, SOCS', description='cytokine inducible SH2-containing protein', other_designations='cytokine-inducible SH2 protein|suppressor of cytokine signaling', map_location='9 57.99 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107296685, end_position_on_the_genomic_accession=107301962, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12702, CurrentID=0, Status='live', Symbol='Socs3', Aliases='Cis3, Cish3, EF-10, Ef10, SSI-3, Ssi3', description='suppressor of cytokine signaling 3', other_designations='CIS-3|E2a-Pbx1 target gene in fibroblasts 10|STAT-induced STAT inhibitor 3|cytokine inducible SH2-containing protein 3', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117966079, end_position_on_the_genomic_accession=117970341, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12703, CurrentID=0, Status='live', Symbol='Socs1', Aliases='Cish1, Cish7, JAB, SOCS-1, SSI-1', description='suppressor of cytokine signaling 1', other_designations='JAK2-binding protein|STAT-induced STAT inhibitor 1|cytokine inducible SH2-containing protein 1|cytokine inducible SH2-containing protein 7', map_location='16 5.81 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10783808, end_position_on_the_genomic_accession=10785536, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12704, CurrentID=0, Status='live', Symbol='Cit', Aliases='C030025P15Rik, CRIK, CRIK-SK-k, Cit', description='citron', other_designations='citron kinase|citron-N|rho-interacting, serine/threonine-protein kinase 21', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115845224, end_position_on_the_genomic_accession=116008586, orientation='plus', exon_count=54, OMIM='', no_hdr0=''),
  12705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12705, CurrentID=0, Status='live', Symbol='Cited1', Aliases='AI316840, AU019144, Msg1', description='Cbp/p300-interacting transactivator with Glu/Asp-rich carboxy-terminal domain 1', other_designations='melanocyte-specific protein 1', map_location='X 45.25 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102247378, end_position_on_the_genomic_accession=102252181, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12709, CurrentID=0, Status='live', Symbol='Ckb', Aliases='B-CK, Bck, Ck-3, Ck3b, Ckb', description='creatine kinase, brain', other_designations='creatine kinase B chain', map_location='12 61.09 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111669355, end_position_on_the_genomic_accession=111672338, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12715, CurrentID=0, Status='live', Symbol='Ckm', Aliases='Ckmm, M-CK, MCK', description='creatine kinase, muscle', other_designations='creatine kinase M chain', map_location='7 9.67 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19411094, end_position_on_the_genomic_accession=19421583, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12716, CurrentID=0, Status='live', Symbol='Ckmt1', Aliases='Mt-CK, ScCKmit, UbCKmit, mi-CK', description='creatine kinase, mitochondrial 1, ubiquitous', other_designations='U-MtCK|acidic-type mitochondrial creatine kinase|mia-CK|ubiquitous mitochondrial creatine kinase', map_location='2 F1-F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121357751, end_position_on_the_genomic_accession=121363737, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12717, CurrentID=0, Status='live', Symbol='ckr', Aliases='', description='chakragati', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12721, CurrentID=0, Status='live', Symbol='Coro1a', Aliases='Clabp, Lmb3, TACO, p57', description='coronin, actin binding protein 1A', other_designations='clipin-A|coronin 1|coronin-like actin binding protein|coronin-like protein A|coronin-like protein p57|tryptophan aspartate-containing coat protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126699774, end_position_on_the_genomic_accession=126704816, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12722, CurrentID=0, Status='live', Symbol='Clca3a1', Aliases='Cacc, Clca1', description='chloride channel accessory 3A1', other_designations='chloride channel calcium activated 1', map_location='3 H2-H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144729697, end_position_on_the_genomic_accession=144760972, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12723, CurrentID=0, Status='live', Symbol='Clcn1', Aliases='Clc-1, Clc1, SMCC1, adr, mto, myotonia, nmf355', description='chloride channel, voltage-sensitive 1', other_designations='arrested development of righting response|chloride channel 1|chloride channel protein, skeletal muscle', map_location='6 20.57 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42282593, end_position_on_the_genomic_accession=42315759, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  12724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12724, CurrentID=0, Status='live', Symbol='Clcn2', Aliases='AL118368, ClC-2, Clc2, nmf240', description='chloride channel, voltage-sensitive 2', other_designations='chloride channel 2', map_location='16 12.5 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20702966, end_position_on_the_genomic_accession=20717741, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  12725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12725, CurrentID=0, Status='live', Symbol='Clcn3', Aliases='Clc3', description='chloride channel, voltage-sensitive 3', other_designations='chloride channel 3|chloride channel protein 3|chloride transporter ClC-3|clC-3', map_location='8 30.9 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=60910389, end_position_on_the_genomic_accession=60983311, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12727, CurrentID=0, Status='live', Symbol='Clcn4', Aliases='Clc4-2-2, Clcn4', description='chloride channel, voltage-sensitive 4', other_designations='chloride channel 4-2|chloride channel protein 4|chloride transporter ClC-4|clC-4|voltage-gated chloride channel ClC-4A', map_location='7 4.23 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7282309, end_position_on_the_genomic_accession=7300851, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12728, CurrentID=0, Status='live', Symbol='Clcn5', Aliases='5430408K11Rik, ClC-5, Clc5, D930009B12Rik, DXImx42e, Sfc13, T25545', description='chloride channel 5', other_designations='chloride channel protein 5|chloride transporter ClC-5', map_location='X 3.21 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7153810, end_position_on_the_genomic_accession=7319359, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  12729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12729, CurrentID=0, Status='live', Symbol='Clns1a', Aliases='2610036D06Rik, 2610100O04Rik, Clci, Clcni, ICLN', description='chloride channel, nucleotide-sensitive, 1A', other_designations='chloride channel current inducer|chloride channel regulator|chloride channel, nucleotide sensitive 1A|chloride conductance regulatory protein ICln|chloride ion current inducer protein|i(Cln)', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97696657, end_position_on_the_genomic_accession=97720793, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12733, CurrentID=0, Status='live', Symbol='Clcnka', Aliases='C75963, CLC-K1, Clcnk1', description='chloride channel, voltage-sensitive Ka', other_designations='chloride channel K1|chloride channel Ka|putative basolateral cTAL chloride channel ClC-Ka', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141384611, end_position_on_the_genomic_accession=141398706, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12737, CurrentID=0, Status='live', Symbol='Cldn1', Aliases='AI596271', description='claudin 1', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=26356645, end_position_on_the_genomic_accession=26371839, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12738, CurrentID=0, Status='live', Symbol='Cldn2', Aliases='AL022813', description='claudin 2', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139800808, end_position_on_the_genomic_accession=139811386, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12739, CurrentID=0, Status='live', Symbol='Cldn3', Aliases='AI182374, Cpetr2, mRVP1', description='claudin 3', other_designations='CPE-R 2|CPE-receptor 2|clostridium perfringens enterotoxin receptor 2', map_location='5 74.93 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134986214, end_position_on_the_genomic_accession=134987476, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12740, CurrentID=0, Status='live', Symbol='Cldn4', Aliases='Cep-r, Cpetr, Cpetr1', description='claudin 4', other_designations='CPE-R|CPE-receptor|clostridium perfringens enterotoxin receptor', map_location='5 74.9 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134945123, end_position_on_the_genomic_accession=134946934, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  668101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668101, CurrentID=0, Status='live', Symbol='Sirpb1b', Aliases='', description='signal-regulatory protein beta 1B', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=15495754, end_position_on_the_genomic_accession=15575065, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12744, CurrentID=0, Status='live', Symbol='clf2', Aliases='', description='cleft lip 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12745, CurrentID=0, Status='live', Symbol='Clgn', Aliases='4930459O04Rik, AI528775, Cln', description='calmegin', other_designations='A2/6|MEG 1 antigen|calnexin-T', map_location='8 39.97 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83389826, end_position_on_the_genomic_accession=83428552, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  12747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12747, CurrentID=0, Status='live', Symbol='Clk1', Aliases='STY', description='CDC-like kinase 1', other_designations='protein kinase STY', map_location='1 29.09 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58411988, end_position_on_the_genomic_accession=58424088, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12748, CurrentID=0, Status='live', Symbol='Clk2', Aliases='AU041688', description='CDC-like kinase 2', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89164805, end_position_on_the_genomic_accession=89177087, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  668110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668110, CurrentID=0, Status='live', Symbol='Syce1l', Aliases='4930481F22Rik, MMRP, PSESS, mmrp2', description='synaptonemal complex central element protein 1 like', other_designations='meiosis-related protein|predicted second spermatocyte specific protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=113643213, end_position_on_the_genomic_accession=113655533, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12751, CurrentID=0, Status='live', Symbol='Tpp1', Aliases='Cln2, LPIC, TPP-1, TPP-I', description='tripeptidyl peptidase I', other_designations='ceroid-lipofuscinosis, neuronal 2|lysosomal pepstatin-insensitive protease|tripeptidyl aminopeptidase|tripeptidyl-peptidase I', map_location='7 55.97 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105744847, end_position_on_the_genomic_accession=105752282, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12752, CurrentID=0, Status='live', Symbol='Cln3', Aliases='AI323623', description='ceroid lipofuscinosis, neuronal 3, juvenile (Batten, Spielmeyer-Vogt disease)', other_designations='', map_location='7 69.16 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126571400, end_position_on_the_genomic_accession=126585817, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12753, CurrentID=0, Status='live', Symbol='Clock', Aliases='5330400M04Rik, KAT13D', description='circadian locomotor output cycles kaput', other_designations='', map_location='5 40.63 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76209868, end_position_on_the_genomic_accession=76305007, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  12757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12757, CurrentID=0, Status='live', Symbol='Clta', Aliases='AV026556, Lca', description='clathrin, light polypeptide (Lca)', other_designations='clathryn light chain (LCA2)', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44003816, end_position_on_the_genomic_accession=44032848, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12759, CurrentID=0, Status='live', Symbol='Clu', Aliases='AI893575, ApoJ, Cli, D14Ucla3, SP-40, Sgp-2, Sgp2, Sugp-2', description='clusterin', other_designations='Apolipoprotein J|clustrin|complement lysis inhibitor|sulfated glycoprotein 2|testosterone repressed prostate message|testosterone repressed prostate message-2', map_location='14 34.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65968483, end_position_on_the_genomic_accession=65981547, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12763, CurrentID=0, Status='live', Symbol='Cmah', Aliases='', description='cytidine monophospho-N-acetylneuraminic acid hydroxylase', other_designations='CMP-N-acetylneuraminate monooxygenase|CMP-N-acetylneuraminic acid hydroxylase|CMP-Neu5Ac hydroxylase|CMP-NeuAc hydroxylase', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24326980, end_position_on_the_genomic_accession=24477289, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  12764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12764, CurrentID=0, Status='live', Symbol='Cmas', Aliases='AW208911, CMPNeu5Ac, D6Bwg0250e', description='cytidine monophospho-N-acetylneuraminic acid synthetase', other_designations='CMP-N-acetylneuraminic acid synthase|CMP-N-acetylneuraminic acid synthetase|CMP-Neu5Ac synthase|CMP-NeuNAc synthase|CMP-NeuNAc synthetase|CMP-sialic acid synthetase', map_location='6 74.66 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142756686, end_position_on_the_genomic_accession=142775714, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12765, CurrentID=0, Status='live', Symbol='Cxcr2', Aliases='CD128, CDw128, Cmkar2, Gpcr16, IL-8Rh, IL-8rb, IL8RA, Il8rb, mIL-8RH', description='chemokine (C-X-C motif) receptor 2', other_designations='CXC-R2|CXCR-2|GRO/MGSA receptor|IL-8 receptor alpha chain|IL-8R B|chemokine (C-X-C) receptor 2|high affinity interleukin-8 receptor B|interleukin 8 receptor, beta|interleukin-8 receptor type B', map_location='1 38.41 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74153994, end_position_on_the_genomic_accession=74161246, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12766, CurrentID=0, Status='live', Symbol='Cxcr3', Aliases='Cd183, Cmkar3', description='chemokine (C-X-C motif) receptor 3', other_designations='IP-10 receptor|interferon-inducible protein 10 receptor', map_location='X 44.58 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101731536, end_position_on_the_genomic_accession=101734147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12767, CurrentID=0, Status='live', Symbol='Cxcr4', Aliases='CD184, Cmkar4, LESTR, PB-CKR, PBSF/SDF-1, Sdf1r, b2b220Clo', description='chemokine (C-X-C motif) receptor 4', other_designations='CXC-R4|CXCR-4|SDF-1 receptor|chemokine (C-X-C) receptor 4|chemokine receptor 4|fusin|leukocyte-derived seven transmembrane domain receptor|leukocyte-expressed seven-transmembrane-domain|pre-B-cell-derived chemokine receptor|stromal cell-derived factor 1 receptor', map_location='1 56.43 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=128588199, end_position_on_the_genomic_accession=128592315, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12768, CurrentID=0, Status='live', Symbol='Ccr1', Aliases='Cmkbr1, Mip-1a-R', description='chemokine (C-C motif) receptor 1', other_designations='C-C CKR-1|CC-CKR-1|CCR-1|MIP-1 alpha R|MIP-1 alphaR|MIP-1alpha-R|RANTES-R|chemokine (C-C) receptor 1|macrophage inflammatory protein 1-alpha receptor', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123962126, end_position_on_the_genomic_accession=123968692, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12769, CurrentID=0, Status='live', Symbol='Ccr9', Aliases='A130091K22Rik, Cmkbr10, GPR-9-6', description='chemokine (C-C motif) receptor 9', other_designations='C-C CKR-9|CC-CKR-9|CCR-9|chemokine C-C receptor 10', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123767211, end_position_on_the_genomic_accession=123783457, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12770, CurrentID=0, Status='live', Symbol='Ccr1l1', Aliases='Cmkbr1l1', description='chemokine (C-C motif) receptor 1-like 1', other_designations='MIP-1 alphaRL1|chemokine (C-C) receptor 1-like 1|macrophage inflammatory protein 1 alpha receptor-like 1|macrophage inflammatory protein-1 alpha receptor-like 1', map_location='9 75.05 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123976949, end_position_on_the_genomic_accession=123978408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12771, CurrentID=0, Status='live', Symbol='Ccr3', Aliases='CC-CKR3, CKR3, Cmkbr1l2, Cmkbr3', description='chemokine (C-C motif) receptor 3', other_designations='C-C CKR-3|CC chemokine receptor 3|CC-CKR-3|CCR-3|MIP-1 alpha RL2|MIP-1 alphaRL2|chemokine (C-C) receptor 1,-like 2|chemokine (C-C) receptor 3|macrophage inflammatory protein 1-alpha receptor-like 2', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=124013629, end_position_on_the_genomic_accession=124031689, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12772, CurrentID=0, Status='live', Symbol='Ccr2', Aliases='Cc-ckr-2a, Ccr2b, Ckr2, Ckr2a, Ckr2b, Cmkbr2, mJe-r, Ccr2', description='chemokine (C-C motif) receptor 2', other_designations='C-C CHEMOKINE RECEPTOR TYPE 2 (C-C CKR-2) (CC-CKR-2) (CCR-2) (CCR2) (JE/FIC RECEPTOR) (MCP-1 RECEPTOR)|C-C CKR-2|CCR-2|JE/FIC receptor|MCP-1 receptor|MIP-1 alphaR|chemoattractant protein-1 receptor|chemokine (C-C) receptor 2', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=124101918, end_position_on_the_genomic_accession=124109140, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12773, CurrentID=0, Status='live', Symbol='Ccr4', Aliases='C-C CKR-4, CHEMR1, Cmkbr4, LESTR, Sdf1r', description='chemokine (C-C motif) receptor 4', other_designations='chemokine (C-C) receptor 4|fusin|leukocyte-expressed seven-transmembrane-domain receptor', map_location='9 64.49 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114490316, end_position_on_the_genomic_accession=114505473, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12774, CurrentID=0, Status='live', Symbol='Ccr5', Aliases='AM4-7, CD195, Cmkbr5', description='chemokine (C-C motif) receptor 5', other_designations='C-C CKR-5|CC-CKR-5|CCR-5|MIP-1 alpha receptor|chemokine (C-C) receptor 5', map_location='9 75.05 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=124121543, end_position_on_the_genomic_accession=124127183, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12775, CurrentID=0, Status='live', Symbol='Ccr7', Aliases='CC-CKR-7, CCR-7, CD197, Cdw197, Cmkbr7, EBI1, Ebi1h', description='chemokine (C-C motif) receptor 7', other_designations='EBV-induced G-protein coupled receptor 1|MIP-3 beta receptor|chemokine (C-C) receptor 7|epstein-Barr virus-induced G-protein coupled receptor 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99144199, end_position_on_the_genomic_accession=99155077, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12776, CurrentID=0, Status='live', Symbol='Ccr8', Aliases='C-C, C-C CKR-8, CC-CKR-8, CCR-8, CKR-8, Cmkbr8, mCCR8', description='chemokine (C-C motif) receptor 8', other_designations='C-C motif chemokine receptor 8', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120092133, end_position_on_the_genomic_accession=120094906, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12777, CurrentID=0, Status='live', Symbol='Ccr10', Aliases='Cmkbr9, Gpr2', description='chemokine (C-C motif) receptor 10', other_designations='C-C CKR-10|CC-CKR-10|CCR-10|G protein-coupled receptor 2|G-protein coupled receptor 2|chemokine C-C receptor 9', map_location='11 64.34 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101172922, end_position_on_the_genomic_accession=101175695, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12778, CurrentID=0, Status='live', Symbol='Ackr3', Aliases='AW541270, CXC-R7, CXCR-7, Cmkor1, Cxcr7, RDC-1, Rdc1', description='atypical chemokine receptor 3', other_designations='C-X-C chemokine receptor type 7|G-protein coupled receptor RDC1 homolog|chemokine (C-X-C motif) receptor 7|chemokine orphan receptor 1', map_location='1 45.28 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90203980, end_position_on_the_genomic_accession=90215724, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12780, CurrentID=0, Status='live', Symbol='Abcc2', Aliases='AI173996, Abc30, Cmoat, Mrp2, cMRP', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 2', other_designations='ATP-binding cassette sub-family C member 2|multidrug resistance protein 2', map_location='19 36.67 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43782308, end_position_on_the_genomic_accession=43838332, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  12785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12785, CurrentID=0, Status='live', Symbol='Cnbp', Aliases='AA4087101, Znf9, Cnbp', description='cellular nucleic acid binding protein', other_designations='cellular nucleic acid binding protein 1|zinc finger protein 9', map_location='6 D1-D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87842615, end_position_on_the_genomic_accession=87851106, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12788, CurrentID=0, Status='live', Symbol='Cnga1', Aliases='Cncg', description='cyclic nucleotide gated channel alpha 1', other_designations='CNG channel alpha-1|CNG-1|CNG1|cyclic nucleotide gated channel, cGMP gated|cyclic nucleotide-gated cation channel 1|cyclic nucleotide-gated channel alpha-1|cyclic nucleotide-gated channel, photoreceptor|rod photoreceptor cGMP-gated channel subunit alpha', map_location='5 38.44 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72603696, end_position_on_the_genomic_accession=72642752, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12789, CurrentID=0, Status='live', Symbol='Cnga2', Aliases='CNG2, Cnca, Cncg4, OCNC1', description='cyclic nucleotide gated channel alpha 2', other_designations='CNG channel alpha-2|cyclic nucleotide gated channel 4|cyclic nucleotide-gated cation channel 2', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71991875, end_position_on_the_genomic_accession=72010218, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12790, CurrentID=0, Status='live', Symbol='Cnga3', Aliases='CNG3', description='cyclic nucleotide gated channel alpha 3', other_designations='CNG channel alpha-3|CNG-3|cone photoreceptor cGMP-gated channel subunit alpha|cyclic nucleotide-gated channel alpha-3', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37218278, end_position_on_the_genomic_accession=37263384, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  12793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12793, CurrentID=0, Status='live', Symbol='Cnih1', Aliases='0610007J15, CNIH-1, Cnih', description='cornichon family AMPA receptor auxiliary protein 1', other_designations='cornichon homolog 1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=46775567, end_position_on_the_genomic_accession=46788888, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12794, CurrentID=0, Status='live', Symbol='Cnih2', Aliases='CNIH-2, Cnil', description='cornichon family AMPA receptor auxiliary protein 2', other_designations='cornichon homolog 2|cornichon-like protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5092868, end_position_on_the_genomic_accession=5098521, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12795, CurrentID=0, Status='live', Symbol='Plk3', Aliases='Cnk, Fnk, PLK-3, PRK', description='polo-like kinase 3', other_designations='FGF-inducible kinase|cytokine-inducible serine/threonine-protein kinase', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117128655, end_position_on_the_genomic_accession=117133952, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12796, CurrentID=0, Status='live', Symbol='Camp', Aliases='CAP18, Cnlp, Cramp, FALL39, MCLP', description='cathelicidin antimicrobial peptide', other_designations='CLP|cathelin-like protein|cathelin-related antimicrobial peptide', map_location='9 59.8 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109847377, end_position_on_the_genomic_accession=109849456, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12797, CurrentID=0, Status='live', Symbol='Cnn1', Aliases='CN, CnnI', description='calponin 1', other_designations='basic calponin|calponin H1, smooth muscle', map_location='9 A2-A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22099221, end_position_on_the_genomic_accession=22109221, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12798, CurrentID=0, Status='live', Symbol='Cnn2', Aliases='AA408047, AI324678, Calpo2', description='calponin 2', other_designations='calponin H2, smooth muscle|h2-calponin|neutral calponin', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79988600, end_position_on_the_genomic_accession=79995400, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  12799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12799, CurrentID=0, Status='live', Symbol='Cnp', Aliases='CNPase-1, Cnp1, Cnp', description="2',3'-cyclic nucleotide 3' phosphodiesterase", other_designations='cyclic nucleotide phosphodiesterase 1', map_location='11 63.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100574909, end_position_on_the_genomic_accession=100591521, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12801, CurrentID=0, Status='live', Symbol='Cnr1', Aliases='CB-R, CB1, CB1R', description='cannabinoid receptor 1 (brain)', other_designations='brain-type cannabinoid receptor|striatal cannabinoid receptor type 1 protein', map_location='4 16.28 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=33923844, end_position_on_the_genomic_accession=33948831, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12802, CurrentID=0, Status='live', Symbol='Cnr2', Aliases='CB-2, CB2, CB2-R', description='cannabinoid receptor 2 (macrophage)', other_designations='cannabinoid receptor 2 (spleen)', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135895404, end_position_on_the_genomic_accession=135920215, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12803, CurrentID=0, Status='live', Symbol='Cntf', Aliases='AI429687', description='ciliary neurotrophic factor', other_designations='', map_location='19 8.73 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12763528, end_position_on_the_genomic_accession=12765632, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12804, CurrentID=0, Status='live', Symbol='Cntfr', Aliases='Cntfralpha', description='ciliary neurotrophic factor receptor', other_designations='CNTF receptor subunit alpha', map_location='4 21.81 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41657498, end_position_on_the_genomic_accession=41697091, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  12805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12805, CurrentID=0, Status='live', Symbol='Cntn1', Aliases='AW495098, CNTN, F3cam, usl', description='contactin 1', other_designations='neural cell surface protein F3', map_location='15 F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=92051142, end_position_on_the_genomic_accession=92341967, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  668166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668166, CurrentID=0, Status='live', Symbol='Zxdb', Aliases='Zxda', description='zinc finger, X-linked, duplicated B', other_designations='zinc finger X-linked protein ZXDA/ZXDB', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94724569, end_position_on_the_genomic_accession=94730191, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  12807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12807, CurrentID=0, Status='live', Symbol='Hps3', Aliases='coa', description='Hermansky-Pudlak syndrome 3', other_designations='Hermansky-Pudlak syndrome 3 homolog|cocoa protein', map_location='3 6.12 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19995945, end_position_on_the_genomic_accession=20035376, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12808, CurrentID=0, Status='live', Symbol='Cobl', Aliases='', description='cordon-bleu WH2 repeat', other_designations='', map_location='11 7.3 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=12236676, end_position_on_the_genomic_accession=12464997, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  12810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12810, CurrentID=0, Status='live', Symbol='Coch', Aliases='AW122937-5B2, D12H14S564E, Coch', description='cochlin', other_designations='coagulation factor C homolog (Limulus polyphemus)', map_location='12 22.11 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51593341, end_position_on_the_genomic_accession=51605773, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12812, CurrentID=0, Status='live', Symbol='Coil', Aliases='C79982, Cln80, p80, p80-coilin', description='coilin', other_designations='nuclear phosphoprotein of Cajal body', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88973935, end_position_on_the_genomic_accession=88991613, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  668173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668173, CurrentID=0, Status='live', Symbol='Pex10', Aliases='AV128229, Gm142', description='peroxisomal biogenesis factor 10', other_designations='peroxin-10|peroxisome assembly protein 10', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155067030, end_position_on_the_genomic_accession=155072406, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12814, CurrentID=0, Status='live', Symbol='Col11a1', Aliases='C530001D20Rik, cho', description='collagen, type XI, alpha 1', other_designations='a1(XI) collagen|pro-alpha1(XI) collagen|procollagen, type XI, alpha 1', map_location='3 49.35 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=114030479, end_position_on_the_genomic_accession=114220327, orientation='plus', exon_count=69, OMIM='', no_hdr0=''),
  12815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12815, CurrentID=0, Status='live', Symbol='Col11a2', Aliases='', description='collagen, type XI, alpha 2', other_designations='procollagen, type XI, alpha 2', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34038486, end_position_on_the_genomic_accession=34066684, orientation='plus', exon_count=68, OMIM='', no_hdr0=''),
  12816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12816, CurrentID=0, Status='live', Symbol='Col12a1', Aliases='', description='collagen, type XII, alpha 1', other_designations='procollagen, type XII, alpha 1', map_location='9 43.82 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=79598987, end_position_on_the_genomic_accession=79720841, orientation='minus', exon_count=66, OMIM='', no_hdr0=''),
  12817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12817, CurrentID=0, Status='live', Symbol='Col13a1', Aliases='', description='collagen, type XIII, alpha 1', other_designations='procollagen, type XIII, alpha 1|type XIII collagen', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61838235, end_position_on_the_genomic_accession=61979209, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  12818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12818, CurrentID=0, Status='live', Symbol='Col14a1', Aliases='5730412L22Rik, AW108078', description='collagen, type XIV, alpha 1', other_designations='col14a1|procollagen, type XIV, alpha 1', map_location='15 D', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=55307706, end_position_on_the_genomic_accession=55520803, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  12819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12819, CurrentID=0, Status='live', Symbol='Col15a1', Aliases='', description='collagen, type XV, alpha 1', other_designations='procollagen type XV|procollagen, type XV|type XV collagen', map_location='4 B1-B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=47207633, end_position_on_the_genomic_accession=47313167, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  12821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12821, CurrentID=0, Status='live', Symbol='Col17a1', Aliases='BP180, Bpag, Bpag2', description='collagen, type XVII, alpha 1', other_designations='180 kDa bullous pemphigoid antigen 2|bullous pemphigoid antigen 2|procollagen, type XVII, alpha 1', map_location='19 40.07 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47646341, end_position_on_the_genomic_accession=47692042, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  12822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12822, CurrentID=0, Status='live', Symbol='Col18a1', Aliases='', description='collagen, type XVIII, alpha 1', other_designations='endostatin|procollagen, type XVIII, alpha 1', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77052178, end_position_on_the_genomic_accession=77166530, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  12823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12823, CurrentID=0, Status='live', Symbol='Col19a1', Aliases='', description='collagen, type XIX, alpha 1', other_designations='col19a1|collagen alpha-1(Y) chain|procollagen, type XIX, alpha 1', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=24257683, end_position_on_the_genomic_accession=24588036, orientation='minus', exon_count=52, OMIM='', no_hdr0=''),
  12824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12824, CurrentID=0, Status='live', Symbol='Col2a1', Aliases='Col2, Col2a, Col2a-1, Del1, Dmm, Lpk, Rgsc856', description='collagen, type II, alpha 1', other_designations='alpha-1 type II collagen|chondrocalcin|procollagen, type II, alpha 1', map_location='15 53.97 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97975602, end_position_on_the_genomic_accession=98004724, orientation='minus', exon_count=55, OMIM='', no_hdr0=''),
  12825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12825, CurrentID=0, Status='live', Symbol='Col3a1', Aliases='AW550625, Col3a-1, Tsk-2, Tsk2', description='collagen, type III, alpha 1', other_designations='procollagen, type III, alpha 1', map_location='1 23.67 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=45311538, end_position_on_the_genomic_accession=45349706, orientation='plus', exon_count=51, OMIM='', no_hdr0=''),
  12826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12826, CurrentID=0, Status='live', Symbol='Col4a1', Aliases='Bru, Col4a-1, Raw, Svc', description='collagen, type IV, alpha 1', other_designations='alpha1(IV) collagen|procollagen, type IV, alpha 1', map_location='8 5.53 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11198423, end_position_on_the_genomic_accession=11312826, orientation='minus', exon_count=53, OMIM='', no_hdr0=''),
  12827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12827, CurrentID=0, Status='live', Symbol='Col4a2', Aliases='Col4a-2', description='collagen, type IV, alpha 2', other_designations='canstatin|procollagen, type IV, alpha 2', map_location='8 5.62 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11312829, end_position_on_the_genomic_accession=11449287, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  12828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12828, CurrentID=0, Status='live', Symbol='Col4a3', Aliases='[a]3(IV), alpha3(IV)', description='collagen, type IV, alpha 3', other_designations='col4a3|collagen type IV alpha3 chain|procollagen, type IV, alpha 3|tumstatin', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=82586921, end_position_on_the_genomic_accession=82722059, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  12829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12829, CurrentID=0, Status='live', Symbol='Col4a4', Aliases='E130010M05Rik, [a]4(IV)', description='collagen, type IV, alpha 4', other_designations='col4a4|procollagen, type IV, alpha 4', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=82448419, end_position_on_the_genomic_accession=82586849, orientation='minus', exon_count=50, OMIM='', no_hdr0=''),
  12830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12830, CurrentID=0, Status='live', Symbol='Col4a5', Aliases='', description='collagen, type IV, alpha 5', other_designations='collagen type IV alpha5 chain|procollagen type IV alpha 5|procollagen, type IV, alpha 5', map_location='X 62.16 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=141475369, end_position_on_the_genomic_accession=141689236, orientation='plus', exon_count=53, OMIM='', no_hdr0=''),
  12831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12831, CurrentID=0, Status='live', Symbol='Col5a1', Aliases='AI413331', description='collagen, type V, alpha 1', other_designations='alpha1(V) collagen|pro-alpha1(V) collagen|procollagen type V alpha 1|procollagen, type V, alpha 1', map_location='2 A2-B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27886395, end_position_on_the_genomic_accession=28039510, orientation='plus', exon_count=67, OMIM='', no_hdr0=''),
  12832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12832, CurrentID=0, Status='live', Symbol='Col5a2', Aliases='1110014L14Rik', description='collagen, type V, alpha 2', other_designations='procollagen, type V, alpha 2', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=45374331, end_position_on_the_genomic_accession=45503282, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  12833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12833, CurrentID=0, Status='live', Symbol='Col6a1', Aliases='AI747156, Col6a-1', description='collagen, type VI, alpha 1', other_designations='procollagen, type VI, alpha 1', map_location='10 39.71 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76708792, end_position_on_the_genomic_accession=76726044, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  12834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12834, CurrentID=0, Status='live', Symbol='Col6a2', Aliases='Col6a-2', description='collagen, type VI, alpha 2', other_designations='procollagen, type VI, alpha 2', map_location='10 39.32 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76595756, end_position_on_the_genomic_accession=76623625, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  12835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12835, CurrentID=0, Status='live', Symbol='Col6a3', Aliases='AI507288, Col6a-3', description='collagen, type VI, alpha 3', other_designations='collagen alpha 3 chain type VI|collagen alpha3(VI)|procollagen, type VI, alpha 3|type VI collagen alpha 3 subunit', map_location='1 45.53 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90766860, end_position_on_the_genomic_accession=90843971, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  12836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12836, CurrentID=0, Status='live', Symbol='Col7a1', Aliases='AW209154', description='collagen, type VII, alpha 1', other_designations='LC collagen|col7a1|long-chain collagen|procollagen, type VII, alpha 1', map_location='9 59.63 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108953575, end_position_on_the_genomic_accession=108984875, orientation='plus', exon_count=119, OMIM='', no_hdr0=''),
  12837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12837, CurrentID=0, Status='live', Symbol='Col8a1', Aliases='Col8a-1', description='collagen, type VIII, alpha 1', other_designations='procollagen type VIII alpha 1|procollagen, type VIII, alpha 1', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57624256, end_position_on_the_genomic_accession=57754737, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12839, CurrentID=0, Status='live', Symbol='Col9a1', Aliases='Col9a-1', description='collagen, type IX, alpha 1', other_designations='alpha 1 type IX collagen|procollagen, type IX, alpha 1', map_location='1 9.95 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=24177590, end_position_on_the_genomic_accession=24252738, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  12840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12840, CurrentID=0, Status='live', Symbol='Col9a2', Aliases='AI427499, Col9a-2', description='collagen, type IX, alpha 2', other_designations='procollagen, type IX, alpha 2', map_location='4 56.78 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=121039566, end_position_on_the_genomic_accession=121055325, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  12841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12841, CurrentID=0, Status='live', Symbol='Col9a3', Aliases='AV006866', description='collagen, type IX, alpha 3', other_designations='procollagen, type IX, alpha 3', map_location='2 103.14 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180597795, end_position_on_the_genomic_accession=180622185, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  12842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12842, CurrentID=0, Status='live', Symbol='Col1a1', Aliases='Col1a-1, Cola-1, Cola1, Mov-13, Mov13', description='collagen, type I, alpha 1', other_designations='alpha-1 type 1 collagen|alpha-1 type I collagen|procollagen, type I, alpha 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94936270, end_position_on_the_genomic_accession=94951867, orientation='plus', exon_count=51, OMIM='', no_hdr0=''),
  12843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12843, CurrentID=0, Status='live', Symbol='Col1a2', Aliases='AA960264, AI325291, Col1a-2, Cola-2, Cola2, oim', description='collagen, type I, alpha 2', other_designations='alpha-2 type I collagen|collagen COL1A2|procollagen, type I, alpha 2', map_location='6 1.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=4505697, end_position_on_the_genomic_accession=4541543, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  12845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12845, CurrentID=0, Status='live', Symbol='Comp', Aliases='TSP5', description='cartilage oligomeric matrix protein', other_designations='thrombospondin-5', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70373548, end_position_on_the_genomic_accession=70382066, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  12846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12846, CurrentID=0, Status='live', Symbol='Comt', Aliases='Comt1, D16Wsu103e, D330014B15Rik', description='catechol-O-methyltransferase', other_designations='catechol-O-methyltransferase 1', map_location='16 11.4 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18406882, end_position_on_the_genomic_accession=18426716, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12847, CurrentID=0, Status='live', Symbol='Copa', Aliases='AU040324, xenin', description='coatomer protein complex subunit alpha', other_designations='alpha-COP|alpha-coat protein', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172082529, end_position_on_the_genomic_accession=172122332, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  12848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12848, CurrentID=0, Status='live', Symbol='Cops2', Aliases='AI315723, C85265, Csn2, Sgn2, TRIP-15, Trip15', description='COP9 signalosome subunit 2', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 2|JAB1-containing signalosome subunit 2|TR-interacting protein 15|alien homolog|alien-like|thyroid receptor-interacting protein 15', map_location='2 61.76 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125830302, end_position_on_the_genomic_accession=125859082, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  12850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12850, CurrentID=0, Status='live', Symbol='Coq7', Aliases='clk-1', description='demethyl-Q 7', other_designations='DMQ hydroxylase|coenzyme Q biosynthesis protein 7 homolog|timing protein clk-1 homolog|ubiquinone biosynthesis monooxygenase COQ7|ubiquinone biosynthesis protein COQ7 homolog', map_location='7 63.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118509659, end_position_on_the_genomic_accession=118533368, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  668212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668212, CurrentID=0, Status='live', Symbol='Efr3b', Aliases='AI852640, C030014M07Rik, mKIAA0953', description='EFR3 homolog B', other_designations='', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3962554, end_position_on_the_genomic_accession=4039094, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  12854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12854, CurrentID=0, Status='live', Symbol='Cort', Aliases='CST, PCST', description='cortistatin', other_designations='', map_location='4 78.89 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149125191, end_position_on_the_genomic_accession=149126741, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12856, CurrentID=0, Status='live', Symbol='Cox17', Aliases='AI037035', description='cytochrome c oxidase assembly protein 17', other_designations='cytochrome c oxidase subunit XVII assembly protein homolog|cytochrome c oxidase, subunit XVII assembly protein homolog', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38346984, end_position_on_the_genomic_accession=38352763, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12857, CurrentID=0, Status='live', Symbol='Cox4i1', Aliases='AL024441, COX, COX IV-1, COXIV, Cox4, Cox4a, IV-1', description='cytochrome c oxidase subunit IV isoform 1', other_designations='cytochrome c oxidase polypeptide IV|cytochrome c oxidase, subunit IVa', map_location='8 69.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120668225, end_position_on_the_genomic_accession=120674209, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  668218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668218, CurrentID=0, Status='live', Symbol='Bin2', Aliases='', description='bridging integrator 2', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100641082, end_position_on_the_genomic_accession=100669576, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12859, CurrentID=0, Status='live', Symbol='Cox5b', Aliases='', description='cytochrome c oxidase subunit Vb', other_designations='cytochrome c oxidase polypeptide Vb', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36691487, end_position_on_the_genomic_accession=36693388, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12861, CurrentID=0, Status='live', Symbol='Cox6a1', Aliases='VIaL', description='cytochrome c oxidase subunit VIa polypeptide 1', other_designations='cytochrome c oxidase polypeptide VIa-liver|subunit VIaL (liver-type)', map_location='5 56.06 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115345654, end_position_on_the_genomic_accession=115348955, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12862, CurrentID=0, Status='live', Symbol='Cox6a2', Aliases='CoxVIaH, VIaH', description='cytochrome c oxidase subunit VIa polypeptide 2', other_designations='cytochrome c oxidase polypeptide VIa-heart|subunit VIaH (heart-type)', map_location='7 70.04 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128205435, end_position_on_the_genomic_accession=128206366, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12864, CurrentID=0, Status='live', Symbol='Cox6c', Aliases='', description='cytochrome c oxidase subunit VIc', other_designations='cytochrome c oxidase polypeptide VIc', map_location='15 B3.3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=35931976, end_position_on_the_genomic_accession=35938246, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12865, CurrentID=0, Status='live', Symbol='Cox7a1', Aliases='COX7A, COX7AH, COX7AM', description='cytochrome c oxidase subunit VIIa 1', other_designations='cytochrome c oxidase subunit VIIa-H|cytochrome c oxidase subunit VIIa-M|cytochrome c oxidase subunit VIIa-heart|cytochrome c oxidase subunit VIIa-muscle|cytochrome c oxidase, subunit VIIa 1', map_location='7 17.31 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30184171, end_position_on_the_genomic_accession=30186030, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12866, CurrentID=0, Status='live', Symbol='Cox7a2', Aliases='COX7AL, Cox7a3, CoxVIIa-L', description='cytochrome c oxidase subunit VIIa 2', other_designations='cytochrome c oxidase subunit VIIa 3|cytochrome c oxidase subunit VIIa-L|cytochrome c oxidase subunit VIIa-liver/heart', map_location='9 43.82 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=79755241, end_position_on_the_genomic_accession=79759853, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12867, CurrentID=0, Status='live', Symbol='Cox7c', Aliases='COXVIIc1, Cox7c', description='cytochrome c oxidase subunit VIIc', other_designations='cytochrome c oxidase polypeptide VIIc', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=86044798, end_position_on_the_genomic_accession=86046795, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12868, CurrentID=0, Status='live', Symbol='Cox8a', Aliases='COX8L', description='cytochrome c oxidase subunit VIIIa', other_designations='COX VIII-L|cytochrome c oxidase polypeptide VIII-liver|cytochrome c oxidase subunit 8-2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7215158, end_position_on_the_genomic_accession=7217616, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12869, CurrentID=0, Status='live', Symbol='Cox8b', Aliases='Cox8h, CoxVIII-H', description='cytochrome c oxidase subunit VIIIb', other_designations='COX VIII-H|cytochrome c oxidase polypeptide VIII-heart|cytochrome c oxidase subunit 8-1|cytochrome c oxidase subunit 8H|cytochrome c oxidase subunit VIII-H', map_location='7 86.11 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140898942, end_position_on_the_genomic_accession=140900446, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12870, CurrentID=0, Status='live', Symbol='Cp', Aliases='D3Ertd555e', description='ceruloplasmin', other_designations='ferroxidase', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19956987, end_position_on_the_genomic_accession=20009750, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  12873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12873, CurrentID=0, Status='live', Symbol='Cpa3', Aliases='MC-CPA', description='carboxypeptidase A3, mast cell', other_designations='', map_location='3 6.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=20215616, end_position_on_the_genomic_accession=20242209, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12874, CurrentID=0, Status='live', Symbol='Cpd', Aliases='AA960140, AW322530, D830034L15Rik', description='carboxypeptidase D', other_designations='gp180|metallocarboxypeptidase D', map_location='11 46.05 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76777208, end_position_on_the_genomic_accession=76847008, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  12876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12876, CurrentID=0, Status='live', Symbol='Cpe', Aliases='CPH, Cph-1, Cph1, R74677, fat', description='carboxypeptidase E', other_designations='carboxypeptidase H|enkephalin convertase|prohormone-processing carboxypeptidase', map_location='8 32.3 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=64592542, end_position_on_the_genomic_accession=64693040, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12877, CurrentID=0, Status='live', Symbol='Cpeb1', Aliases='AU024112, Cpeb, mCPEB', description='cytoplasmic polyadenylation element binding protein 1', other_designations='CPE-BP1|CPE-binding protein 1|mCPEB-1', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81347026, end_position_on_the_genomic_accession=81455457, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12879, CurrentID=0, Status='live', Symbol='Cys1', Aliases='2900006B19Rik, AV218859, Ccap, ck, cpk', description='cystin 1', other_designations='cilia-associated protein|congenital polycystic kidneys|cystin cilia associated protein', map_location='12 8.49 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24665833, end_position_on_the_genomic_accession=24681795, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  329828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329828, CurrentID=0, Status='live', Symbol='AI464131', Aliases='AA408153, AI604836, Gm762, Kiaa1161, NET37', description='expressed sequence AI464131', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41491847, end_position_on_the_genomic_accession=41503136, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12890, CurrentID=0, Status='live', Symbol='Cplx2', Aliases='921-L, AI413745, AW492120', description='complexin 2', other_designations='CPX II|complexin II|synaphin-1', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54371349, end_position_on_the_genomic_accession=54383917, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12891, CurrentID=0, Status='live', Symbol='Cpne6', Aliases='AU067659, BB076446', description='copine VI', other_designations='N-copine|copine 6|neuronal copine|neuronal-copine', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55510063, end_position_on_the_genomic_accession=55517431, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  12892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12892, CurrentID=0, Status='live', Symbol='Cpox', Aliases='CPX, Cpo, HCP, cac, nct', description='coproporphyrinogen oxidase', other_designations='COX|Coproporphyrinogen III oxidase, mitochondrial precursor (Coproporphyrinogenase) (Coprogen oxidase) (COX)|clone 560|coprogen oxidase|coproporphyrinogen-III oxidase, mitochondrial|coproporphyrinogenase', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58670208, end_position_on_the_genomic_accession=58680391, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  668253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668253, CurrentID=0, Status='live', Symbol='Dleu2', Aliases='1810047A16Rik, 4932411A06Rik, 7730401J12Rik, AI197277, AI788824, Alt1, Gm9069, Leu2', description='deleted in lymphocytic leukemia, 2', other_designations='', map_location='14 33.16 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61602836, end_position_on_the_genomic_accession=61682373, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12894, CurrentID=0, Status='live', Symbol='Cpt1a', Aliases='C730027G07, CPTI, Cpt1', description='carnitine palmitoyltransferase 1a, liver', other_designations='CPT I|CPT1-L|CPTI-L|L-CPT I|carnitine O-palmitoyltransferase I, liver isoform|carnitine palmitoyltransferase 1, liver', map_location='19 3.08 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3322326, end_position_on_the_genomic_accession=3385735, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  12895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12895, CurrentID=0, Status='live', Symbol='Cpt1b', Aliases='Cpt1, Cpt1-m, Cpti, Cpti-m, M-cpti', description='carnitine palmitoyltransferase 1b, muscle', other_designations='muscle-type carnitine palmitoyltransferase I', map_location='15 44.89 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89416405, end_position_on_the_genomic_accession=89425862, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  12896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12896, CurrentID=0, Status='live', Symbol='Cpt2', Aliases='AI323697, CPTII', description='carnitine palmitoyltransferase 2', other_designations='CPT II|carnitine palmitoyltransferase II', map_location='4 50.18 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107903982, end_position_on_the_genomic_accession=107923589, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  668257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668257, CurrentID=0, Status='live', Symbol='Dgat2l6', Aliases='9930021F22Rik', description='diacylglycerol O-acyltransferase 2-like 6', other_designations='', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100516094, end_position_on_the_genomic_accession=100546108, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  12902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12902, CurrentID=0, Status='live', Symbol='Cr2', Aliases='C3DR, CD21, CD35, Cr-1, Cr-2, Cr1', description='complement receptor 2', other_designations='complement C3d receptor|complement receptor type 1|expressed in B lymphocytes', map_location='1 98.44 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=195136811, end_position_on_the_genomic_accession=195177190, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  12903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12903, CurrentID=0, Status='live', Symbol='Crabp1', Aliases='AI326249, CRABP-I, Crabp-1, CrabpI, Rbp-5', description='cellular retinoic acid binding protein I', other_designations='cellular retinoic acid-binding protein I', map_location='9 29.76 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54764748, end_position_on_the_genomic_accession=54773110, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12904, CurrentID=0, Status='live', Symbol='Crabp2', Aliases='AI893628, Crabp-2, CrabpII', description='cellular retinoic acid binding protein II', other_designations='CRABP-II|cellular retinoic acid-binding protein II', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87948693, end_position_on_the_genomic_accession=87953372, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12905, CurrentID=0, Status='live', Symbol='Cradd', Aliases='RAIDD', description='CASP2 and RIPK1 domain containing adaptor with death domain', other_designations='CASP2 and RIPK1 domain-containing adaptor with death domain|RIP-associated protein with a death domain|caspase and RIP adapter with death domain', map_location='10 49.26 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=95174739, end_position_on_the_genomic_accession=95324345, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67687, CurrentID=0, Status='live', Symbol='1700011L22Rik', Aliases='', description='RIKEN cDNA 1700011L22 gene', other_designations='', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79210423, end_position_on_the_genomic_accession=79248897, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12908, CurrentID=0, Status='live', Symbol='Crat', Aliases='AW107812, CARAT', description='carnitine acetyltransferase', other_designations='CAT|carnitine acetylase', map_location='2 21.69 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30400476, end_position_on_the_genomic_accession=30415748, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12909, CurrentID=0, Status='live', Symbol='Crcp', Aliases='AL022669', description='calcitonin gene-related peptide-receptor component protein', other_designations='CGRP-RCP|CGRP-receptor component protein|CGRPRCP|RNA polymerase III subunit C9', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130029306, end_position_on_the_genomic_accession=130060787, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12912, CurrentID=0, Status='live', Symbol='Creb1', Aliases='2310001E10Rik, 3526402H21Rik, AV083133, Creb, Creb-1', description='cAMP responsive element binding protein 1', other_designations='cAMP-responsive element-binding protein 1', map_location='1 32.74 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=64532804, end_position_on_the_genomic_accession=64604548, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12913, CurrentID=0, Status='live', Symbol='Creb3', Aliases='AU044960, AW538053, C80076, LZIP, LZIP-1, LZIP-2, Luman', description='cAMP responsive element binding protein 3', other_designations='CREB-3|cAMP-responsive element-binding protein 3|transcription factor LZIP', map_location='4 23.05 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43562634, end_position_on_the_genomic_accession=43567060, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12914, CurrentID=0, Status='live', Symbol='Crebbp', Aliases='AW558298, CBP, CBP/p300, KAT3A, p300/CBP', description='CREB binding protein', other_designations='', map_location='16 2.4 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4081334, end_position_on_the_genomic_accession=4213454, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  12915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12915, CurrentID=0, Status='live', Symbol='Atf6b', Aliases='AA617266, ATF6beta, Creb-rp, Crebl1', description='activating transcription factor 6 beta', other_designations='ATF6-beta|cAMP response element binding protein-related protein|cAMP response element-binding protein-related protein|cAMP responsive element binding protein-like 1|cAMP-dependent transcription factor ATF-6 beta|cAMP-responsive element-binding protein-like 1', map_location='17 18.22 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34647146, end_position_on_the_genomic_accession=34655074, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  12916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12916, CurrentID=0, Status='live', Symbol='Crem', Aliases='ICER, ICERI', description='cAMP responsive element modulator', other_designations='cAMP responsive element moderator|inducible cAMP early repressor', map_location='18 A', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=3266354, end_position_on_the_genomic_accession=3340355, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  12917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12917, CurrentID=0, Status='live', Symbol='crf', Aliases='', description='cream fur', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12918, CurrentID=0, Status='live', Symbol='Crh', Aliases='CRF, Gm1347', description='corticotropin releasing hormone', other_designations='corticotropin releasing factor|corticotropin-releasing factor|corticotropin-releasing hormone', map_location='3 5.75 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19693401, end_position_on_the_genomic_accession=19695396, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  12919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12919, CurrentID=0, Status='live', Symbol='Crhbp', Aliases='CRH-BP', description='corticotropin releasing hormone binding protein', other_designations='CRF-BP|CRF-binding protein|corticotropin-releasing hormone-binding protein', map_location='13 49.5 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95431376, end_position_on_the_genomic_accession=95444831, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12921, CurrentID=0, Status='live', Symbol='Crhr1', Aliases='CRFR1, Crhr', description='corticotropin releasing hormone receptor 1', other_designations='CRF 1 receptor|CRF-R-1|CRF-R1|CRF-R1alpha|CRFR-1|CRH receptor subtype-1 isoform-3|CRH-R-1|CRH-R1|corticotropin-releasing hormone receptor 1', map_location='11 67.77 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104131998, end_position_on_the_genomic_accession=104175518, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  12922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12922, CurrentID=0, Status='live', Symbol='Crhr2', Aliases='CRF-R2, CRFR2alpha, CRFR2beta, Crfr2', description='corticotropin releasing hormone receptor 2', other_designations='CRF 2 receptor|CRF-R-2|CRF-RB|CRFR-2|CRH-R-2|CRH-R2|corticotropin releasing factor receptor type 2 beta|corticotropin-releasing factor 2|corticotropin-releasing hormone receptor 2', map_location='6 27.33 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55090048, end_position_on_the_genomic_accession=55133016, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12924, CurrentID=0, Status='live', Symbol='cri', Aliases='', description='cribriform degeneration', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12925, CurrentID=0, Status='live', Symbol='Crip1', Aliases='CRHP, CRP1, Crip', description='cysteine-rich protein 1 (intestinal)', other_designations='CRP-1|cysteine rich intestinal protein|cysteine-rich intestinal protein', map_location='12 61.59 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113152012, end_position_on_the_genomic_accession=113153879, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12927, CurrentID=0, Status='live', Symbol='Bcar1', Aliases='AI385681, Cas, Crkas', description='breast cancer anti-estrogen resistance 1', other_designations='CRK-associated substrate|p130 Cas|p130cas|v-crk-associated tyrosine kinase substrate', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111710474, end_position_on_the_genomic_accession=111743849, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12928, CurrentID=0, Status='live', Symbol='Crk', Aliases='Crk-I-II, Crk-III, Crk3, CrkIII, Crko, c-Crk, p38, Crk', description='v-crk avian sarcoma virus CT10 oncogene homolog', other_designations='CT-10 related kinase 3|avian sarcoma virus CT10 (v-crk) oncogene homolog|proto-oncogene c-crk|v-crk sarcoma virus CT10 oncogene homolog|v-crk sarcoma virus CT10 oncogene-like protein', map_location='11 A-D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75679248, end_position_on_the_genomic_accession=75708428, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12929, CurrentID=0, Status='live', Symbol='Crkl', Aliases='1110025F07Rik, AA589403, AI325100, Crkol', description='v-crk avian sarcoma virus CT10 oncogene homolog-like', other_designations='v-crk sarcoma virus CT10 oncogene homolog (avian)-like', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17451985, end_position_on_the_genomic_accession=17487440, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12931, CurrentID=0, Status='live', Symbol='Crlf1', Aliases='CLF-1, CRLM-3, CRLM3, NR6, NR6.1', description='cytokine receptor-like factor 1', other_designations='cytokine receptor like molecule 3|cytokine receptor-like molecule 3|cytokine-like factor 1|novel cytokine receptor 6', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70493107, end_position_on_the_genomic_accession=70504081, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12933, CurrentID=0, Status='live', Symbol='Crmp1', Aliases='CRMP-1, DRP-1, Dpysl1, ULIP-3, Ulip3', description='collapsin response mediator protein 1', other_designations='unc-33-like phosphoprotein 3', map_location='5 19.96 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37242058, end_position_on_the_genomic_accession=37292163, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12934, CurrentID=0, Status='live', Symbol='Dpysl2', Aliases='AI851130, Crmp2, DRP2, Musunc33, TOAD-64, Ulip2', description='dihydropyrimidinase-like 2', other_designations='DRP-2|ULIP-2|collapsin response mediator protein 2|unc-33-like phosphoprotein 2', map_location='14 34.6 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66802864, end_position_on_the_genomic_accession=66910919, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12936, CurrentID=0, Status='live', Symbol='Pcdha4', Aliases='Cnr1, Crnr1, R75250', description='protocadherin alpha 4', other_designations='PCDH-alpha-4|cadherin-related neuronal receptor 1', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36952689, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12937, CurrentID=0, Status='live', Symbol='Pcdha6', Aliases='CNR, Cnr2, Crnr2', description='protocadherin alpha 6', other_designations='cadherin-related neuronal receptor 2', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36967756, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12939, CurrentID=0, Status='live', Symbol='Pcdha7', Aliases='Cnr4, Crnr4', description='protocadherin alpha 7', other_designations='cadherin-related neuronal receptor 4', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36973924, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12941, CurrentID=0, Status='live', Symbol='Pcdha5', Aliases='Cnr6, Crnr6', description='protocadherin alpha 5', other_designations='cadherin-related neuronal receptor 6', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36960440, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12942, CurrentID=0, Status='live', Symbol='Pcdha11', Aliases='A830022B16Rik, Cnr7, Crnr7', description='protocadherin alpha 11', other_designations='cadherin-related neuronal receptor 7', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37010858, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12943, CurrentID=0, Status='live', Symbol='Pcdha10', Aliases='Cnr3, Cnr8, Crnr3, Crnr8, Pcdha14', description='protocadherin alpha 10', other_designations='PCDH-alpha-10|cadherin-related neuronal receptor 3|protocadherin alpha 14', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37005320, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12944, CurrentID=0, Status='live', Symbol='Crp', Aliases='AI255847', description='C-reactive protein, pentraxin-related', other_designations='C-reactive protein, petaxin related', map_location='1 80.13 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172698056, end_position_on_the_genomic_accession=172699966, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  12945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12945, CurrentID=0, Status='live', Symbol='Dmbt1', Aliases='CRP, CRP-[a], CRP-[b], Crpd, DBMT1, gp300, p80', description='deleted in malignant brain tumors 1', other_designations='apactin|crp-ductin|deleted in malignant brain tumors 1 pseudogene|ebnerin|glycoprotein 300|hensin|mucin-like glycoprotein|muclin|vomeroglandin', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131032054, end_position_on_the_genomic_accession=131121628, orientation='plus', exon_count=51, OMIM='', no_hdr0=''),
  12946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12946, CurrentID=0, Status='live', Symbol='Cr1l', Aliases='Crry, Mcp, mCRY', description='complement component (3b/4b) receptor 1-like', other_designations='complement receptor related protein|complement regulatory protein Crry|protein p65', map_location='1 98.43 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=195098747, end_position_on_the_genomic_accession=195131570, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  12947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12947, CurrentID=0, Status='live', Symbol='Crry-ps', Aliases='Crrx, Mcp-ps, mCRX', description='complement receptor related protein, pseudogene', other_designations='', map_location='8 35.31 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=73612902, end_position_on_the_genomic_accession=73614459, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  12948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12948, CurrentID=0, Status='live', Symbol='crsp', Aliases='', description='cryptorchidism with white spotting, deletion region', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12950, CurrentID=0, Status='live', Symbol='Hapln1', Aliases='BB099155, CLP, Crtl1, Crtl1l, LP, LP-1', description='hyaluronan and proteoglycan link protein 1', other_designations='cartilage link protein 1|cartilage linking protein 1|cartilage-link protein|cartilage-linking protein 1|link protein|proteoglycan link protein', map_location='13 45.5 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=89540529, end_position_on_the_genomic_accession=89611832, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12951, CurrentID=0, Status='live', Symbol='Crx', Aliases='Crx1', description='cone-rod homeobox', other_designations='cone-rod homeobox containing', map_location='7 8.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15865947, end_position_on_the_genomic_accession=15879955, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12952, CurrentID=0, Status='live', Symbol='Cry1', Aliases='AU020726, AU021000, Phll1', description='cryptochrome 1 (photolyase-like)', other_designations='', map_location='10 C', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85131700, end_position_on_the_genomic_accession=85185054, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  12953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12953, CurrentID=0, Status='live', Symbol='Cry2', Aliases='AV006279, D130054K12Rik', description='cryptochrome 2 (photolyase-like)', other_designations='', map_location='2 E', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92403646, end_position_on_the_genomic_accession=92434069, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12954, CurrentID=0, Status='live', Symbol='Cryaa', Aliases='Acry-1, Crya-1, Crya1, DAcry-1, lop18', description='crystallin, alpha A', other_designations='alpha-crystallin|crystallin, alpha 1|lens opacity 18', map_location='17 A3-B', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31674348, end_position_on_the_genomic_accession=31681730, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12955, CurrentID=0, Status='live', Symbol='Cryab', Aliases='Crya-2, Crya2, HspB5, P23', description='crystallin, alpha B', other_designations='alpha B-crystallin|alpha(B)-crystallin|crystallin, alpha 2|heat shock protein CryAB|heat shock protein HspB5', map_location='9 27.75 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50751325, end_position_on_the_genomic_accession=50756636, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12957, CurrentID=0, Status='live', Symbol='Cryba1', Aliases='BA3/A1, Cryb', description='crystallin, beta A1', other_designations='beta-A1-crystallin|beta-A3/A1-crystallin', map_location='11 B-C1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77718732, end_position_on_the_genomic_accession=77725781, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12958, CurrentID=0, Status='live', Symbol='Cryba2', Aliases='E130107M19Rik', description='crystallin, beta A2', other_designations='beta-A2 crystallin|betaA2-crystallin|lens structural protein', map_location='1 38.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74889934, end_position_on_the_genomic_accession=74898843, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12959, CurrentID=0, Status='live', Symbol='Cryba4', Aliases='', description='crystallin, beta A4', other_designations='beta-A4 crystallin', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112246492, end_position_on_the_genomic_accession=112263555, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  406176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=406176, CurrentID=0, Status='live', Symbol='Olfr151', Aliases='M71, MOR171-2', description='olfactory receptor 151', other_designations='GA_x6K02T2PVTD-31408136-31407207|M71 odorant receptor|odorant receptor M71|olfactory receptor 171-2|olfactory receptor 7H', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37729977, end_position_on_the_genomic_accession=37731061, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12961, CurrentID=0, Status='live', Symbol='Crybb2', Aliases='Cryb-2', description='crystallin, beta B2', other_designations='beta-crystallin Bp', map_location='5 55.38 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113058258, end_position_on_the_genomic_accession=113070111, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12962, CurrentID=0, Status='live', Symbol='Crybb3', Aliases='AI852419', description='crystallin, beta B3', other_designations='beta-B3 crystallin', map_location='5 55.38 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113075839, end_position_on_the_genomic_accession=113082233, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  12964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12964, CurrentID=0, Status='live', Symbol='Cryga', Aliases='Cryg-4, DGcry-4', description='crystallin, gamma A', other_designations='gamma-A-crystallin|gamma-crystallin 4', map_location='1 32.84 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65100389, end_position_on_the_genomic_accession=65103363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12965, CurrentID=0, Status='live', Symbol='Crygb', Aliases='Cryg-3, DGcry-3, Nop', description='crystallin, gamma B', other_designations='gamma-B-crystallin|gamma-crystallin 3|gammaB-crystallin|nuclear opacity', map_location='1 32.82 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65080222, end_position_on_the_genomic_accession=65082290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12966, CurrentID=0, Status='live', Symbol='Crygc', Aliases='Cryg-5', description='crystallin, gamma C', other_designations='gamma-C-crystallin|gammaC-crystallin', map_location='1 32.81 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65071525, end_position_on_the_genomic_accession=65073532, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12967, CurrentID=0, Status='live', Symbol='Crygd', Aliases='Aey4, Cryg-1, DGcry-1, Lop12', description='crystallin, gamma D', other_designations='gamma-D-crystallin|gamma-crystallin 1|lens opacity 12', map_location='1 32.8 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65061838, end_position_on_the_genomic_accession=65063440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  12968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12968, CurrentID=0, Status='live', Symbol='Cryge', Aliases='Cryg-2, Cryg-6, DGcry-2, Elo', description='crystallin, gamma E', other_designations='eye lens obsolescence|gamma-E-crystallin', map_location='1 32.8 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65048555, end_position_on_the_genomic_accession=65057116, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  12969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12969, CurrentID=0, Status='live', Symbol='Crygf', Aliases='3110001K11Rik, Cryg-2, DGcry-2, Len-2', description='crystallin, gamma F', other_designations='gamma-F-crystallin', map_location='1 33.31 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65926522, end_position_on_the_genomic_accession=65928304, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  406186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=406186, CurrentID=0, Status='live', Symbol='Olfr142', Aliases='K20, MOR227-2', description='olfactory receptor 142', other_designations='GA_x6K02T2Q125-51607674-51606757|olfactory receptor 227-2|olfactory receptor 4C', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90252069, end_position_on_the_genomic_accession=90252986, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12971, CurrentID=0, Status='live', Symbol='Crym', Aliases='', description='crystallin, mu', other_designations='NADP-regulated thyroid-hormone-binding protein|mu-crystallin homolog|thiomorpholine-carboxylate dehydrogenase', map_location='7 64.47 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120186384, end_position_on_the_genomic_accession=120201988, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12972, CurrentID=0, Status='live', Symbol='Cryz', Aliases='Sez9', description='crystallin, zeta', other_designations='NADPH:quinone reductase|quinone reductase|zeta-crystallin', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=154596711, end_position_on_the_genomic_accession=154623182, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12973, CurrentID=0, Status='live', Symbol='crz', Aliases='', description='control of early proliferation of Trypanosoma cruzi', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  12974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12974, CurrentID=0, Status='live', Symbol='Cs', Aliases='2610511A05Rik, 9030605P22Rik, Ahl4, BB234005, Cis', description='citrate synthase', other_designations='citrate (Si)-synthase', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128337832, end_position_on_the_genomic_accession=128362479, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12977, CurrentID=0, Status='live', Symbol='Csf1', Aliases='C87615, Csfm, MCSF, op', description='colony stimulating factor 1 (macrophage)', other_designations='osteopetrosis', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107741048, end_position_on_the_genomic_accession=107760469, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  12978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12978, CurrentID=0, Status='live', Symbol='Csf1r', Aliases='AI323359, CD115, CSF-1R, Csfmr, Fim-2, Fms, M-CSF-R, M-CSFR', description='colony stimulating factor 1 receptor', other_designations='CSF-1 receptor|CSF-1-R|c-fms|proto-oncogene c-Fms|proto-oncogene fms', map_location='18 D', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61096808, end_position_on_the_genomic_accession=61131139, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  12981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12981, CurrentID=0, Status='live', Symbol='Csf2', Aliases='Csfgm, GMCSF, Gm-CSf, MGI-IGM', description='colony stimulating factor 2 (granulocyte-macrophage)', other_designations='CSF|colony-stimulating factor|granulocyte-macrophage colony stimulating factor 2|put. GM-CSF', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54247270, end_position_on_the_genomic_accession=54249899, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12982, CurrentID=0, Status='live', Symbol='Csf2ra', Aliases='CD116, Csfgmra, GM-CSF-Ra, GM-CSFR', description='colony stimulating factor 2 receptor, alpha, low-affinity (granulocyte-macrophage)', other_designations='GM-CSF-R-alpha|GM-CSFRalpha|GMCSFR-alpha|GMR-alpha|colony stimulating factor 2 receptor alpha chain', map_location='19 56.89 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=61224401, end_position_on_the_genomic_accession=61228465, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  12983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12983, CurrentID=0, Status='live', Symbol='Csf2rb', Aliases='AI848964, AIC2B, Bc, CDw1311, Csfgmrb, Il3r, Il3rb, Il3rb1, Il5rb, Csf2rb', description='colony stimulating factor 2 receptor, beta, low-affinity (granulocyte-macrophage)', other_designations='CD131 antigen|GM-CSF/IL-3/IL-5 receptor common beta subunit|GM-CSF/IL-3/IL-5 receptor common beta-chain|beta c|colony stimulating factor 2 receptor, beta 1, low-affinity (granulocyte-macrophage)|common beta chain', map_location='15 37.36 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78325797, end_position_on_the_genomic_accession=78351001, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  12984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12984, CurrentID=0, Status='live', Symbol='Csf2rb2', Aliases='AIC2A, BetaIl3, Bil3, Csfgmrb, Il3r, Il3rb, Il3rb2', description='colony stimulating factor 2 receptor, beta 2, low-affinity (granulocyte-macrophage)', other_designations='IL-3 receptor class 2 subunit beta|IL-3R class 2 subunit beta|colony-stimulating factor 2 receptor subunit beta-2|interleukin-3 receptor class II beta chain', map_location='15 37.22 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78282507, end_position_on_the_genomic_accession=78305721, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  12985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12985, CurrentID=0, Status='live', Symbol='Csf3', Aliases='Csfg, G-CSF, MGI-IG', description='colony stimulating factor 3 (granulocyte)', other_designations='', map_location='11 62.45 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98701095, end_position_on_the_genomic_accession=98704836, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12986, CurrentID=0, Status='live', Symbol='Csf3r', Aliases='Cd114, Csfgr, G-CSF-R, G-CSFR', description='colony stimulating factor 3 receptor (granulocyte)', other_designations='G-CSF receptor', map_location='4 59.77 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126023441, end_position_on_the_genomic_accession=126044975, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  12988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12988, CurrentID=0, Status='live', Symbol='Csk', Aliases='AW212630, p50CSK', description='c-src tyrosine kinase', other_designations='C-Src kinase|protein-tyrosine kinase MPK-2', map_location='9 31.18 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57626646, end_position_on_the_genomic_accession=57653626, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  12990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12990, CurrentID=0, Status='live', Symbol='Csn1s1', Aliases='AA959832, AW212882, Csna', description='casein alpha s1', other_designations='alpha-casein', map_location='5 43.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87666208, end_position_on_the_genomic_accession=87682578, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  12991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12991, CurrentID=0, Status='live', Symbol='Csn2', Aliases='Csnb', description='casein beta', other_designations='', map_location='5 43.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87692619, end_position_on_the_genomic_accession=87699429, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12992, CurrentID=0, Status='live', Symbol='Csn1s2b', Aliases='AW987150, Csnd, Csne', description='casein alpha s2-like B', other_designations='casein delta|epsilon-casein', map_location='5 43.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87808121, end_position_on_the_genomic_accession=87824421, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  12993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12993, CurrentID=0, Status='live', Symbol='Csn1s2a', Aliases='Csng, PP22', description='casein alpha s2-like A', other_designations='gamma casein|milk protein', map_location='5 43.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87774567, end_position_on_the_genomic_accession=87788797, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  12994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12994, CurrentID=0, Status='live', Symbol='Csn3', Aliases='AW208918, CSN10, Csnk', description='casein kappa', other_designations='', map_location='5 43.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87925575, end_position_on_the_genomic_accession=87932545, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  12995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12995, CurrentID=0, Status='live', Symbol='Csnk2a1', Aliases='Csnk2a1-rs4', description='casein kinase 2, alpha 1 polypeptide', other_designations='CK II alpha|casein kinase II, alpha 1 polypeptide|casein kinase II, alpha 1 related sequence 4', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152226731, end_position_on_the_genomic_accession=152281856, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  668357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668357, CurrentID=0, Status='live', Symbol='Dmrtc1c2', Aliases='', description='DMRT-like family C1c2', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102847636, end_position_on_the_genomic_accession=102857207, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  668359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668359, CurrentID=0, Status='live', Symbol='Gm9125', Aliases='EG668359', description='predicted gene 9125', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94047981, end_position_on_the_genomic_accession=94054885, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13000, CurrentID=0, Status='live', Symbol='Csnk2a2', Aliases='1110035J23Rik, C77789, CK2', description='casein kinase 2, alpha prime polypeptide', other_designations="CK II alpha'|casein kinase II, alpha 2, polypeptide", map_location='8 47.12 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95446096, end_position_on_the_genomic_accession=95488904, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  406217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=406217, CurrentID=0, Status='live', Symbol='Bex4', Aliases='', description='brain expressed gene 4', other_designations='BEX1-like protein 1|brain expressed X-linked 4|brain expressed, X-linked 4|brain-expressed X-linked protein 4 homolog', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136139045, end_position_on_the_genomic_accession=136140439, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  406218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=406218, CurrentID=0, Status='live', Symbol='Panx2', Aliases='', description='pannexin 2', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89059726, end_position_on_the_genomic_accession=89073571, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  406219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=406219, CurrentID=0, Status='live', Symbol='Krt83', Aliases='Krt2-25, Krt87', description='keratin 83', other_designations='K83|keratin complex 2, basic, gene 25|keratin-83|krt83|type II hair keratin Hb3|type-II keratin Kb23|type-II keratin Kb25', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101431490, end_position_on_the_genomic_accession=101438804, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  406220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=406220, CurrentID=0, Status='live', Symbol='Krt77', Aliases='4732484G22Rik, K1emb, Krt239, mK1b', description='keratin 77', other_designations='CK-1B|K77|cytokeratin-1B|embryonic type II keratin 1|embryonic type II keratin-1|keratin 1b|keratin complex 2, basic, gene 39|keratin-77|type II keratin Kb39|type-II intermediate filament protein|type-II keratin Kb39', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101858731, end_position_on_the_genomic_accession=101869706, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  406221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=406221, CurrentID=0, Status='live', Symbol='Krt40', Aliases='CK-40, K40', description='keratin 40', other_designations='cytokeratin-40|keratin-40', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99537485, end_position_on_the_genomic_accession=99543203, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13006, CurrentID=0, Status='live', Symbol='Smc3', Aliases='Bamacan, Cspg6, HCAP, Mmip1, SMC-3, SmcD', description='structural maintenance of chromosomes 3', other_designations='SMC protein 3|basement membrane-associated chondroitin proteoglycan|chondroitin sulfate proteoglycan 6|chromosome segregation protein SmcD|mad member-interacting protein 1|structural maintenace of chromosomes 3', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53600396, end_position_on_the_genomic_accession=53645831, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  13007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13007, CurrentID=0, Status='live', Symbol='Csrp1', Aliases='AA408841, AA959891, AW545626, CRP1, Csrp', description='cysteine and glycine-rich protein 1', other_designations='CRP|cysteine rich protein 1|cysteine-rich protein 1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135729197, end_position_on_the_genomic_accession=135752229, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13008, CurrentID=0, Status='live', Symbol='Csrp2', Aliases='AW551867, Crp2, SmLim', description='cysteine and glycine-rich protein 2', other_designations='DLP-1|cysteine-rich protein 2|double LIM protein 1', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=110920176, end_position_on_the_genomic_accession=110939514, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13009, CurrentID=0, Status='live', Symbol='Csrp3', Aliases='CRP3, MLP, MMLP', description='cysteine and glycine-rich protein 3', other_designations='LIM domain protein, cardiac|cardiac LIM protein|cysteine-rich protein 3|muscle LIM protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48830398, end_position_on_the_genomic_accession=48848051, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13010, CurrentID=0, Status='live', Symbol='Cst3', Aliases='CysC', description='cystatin C', other_designations='cystatin 3|cystatin-3', map_location='2 73.6 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148871732, end_position_on_the_genomic_accession=148875468, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13011, CurrentID=0, Status='live', Symbol='Cst7', Aliases='Cmap', description='cystatin F (leukocystatin)', other_designations='cystatin 7|cystatin-7|cystatin-like metastasis-associated protein', map_location='2 G1-G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150570415, end_position_on_the_genomic_accession=150578944, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13012, CurrentID=0, Status='live', Symbol='Cst8', Aliases='Cres, Cst-rs1', description='cystatin 8 (cystatin-related epididymal spermatogenic)', other_designations='cystatin-related epididymal spermatogenic protein|cystatin-related epididymal-specific protein', map_location='2 73.56 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148798839, end_position_on_the_genomic_accession=148805596, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13013, CurrentID=0, Status='live', Symbol='Cst9', Aliases='M12, cresp, testatin', description='cystatin 9', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148835147, end_position_on_the_genomic_accession=148838737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13014, CurrentID=0, Status='live', Symbol='Cstb', Aliases='AA960480, Stfb', description='cystatin B', other_designations='stefin-B', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78425670, end_position_on_the_genomic_accession=78427622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13016, CurrentID=0, Status='live', Symbol='Ctbp1', Aliases='BARS, CtBP3/BARS, D4S115h, D5H4S115, D5H4S115E', description='C-terminal binding protein 1', other_designations='', map_location='5 17.52 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33247723, end_position_on_the_genomic_accession=33275004, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13017, CurrentID=0, Status='live', Symbol='Ctbp2', Aliases='AA407280, D7Ertd45e, Gtrgeo6, Ribeye', description='C-terminal binding protein 2', other_designations='gene trap ROSA b-geo 6', map_location='7 76.32 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132986417, end_position_on_the_genomic_accession=133123877, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  13018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13018, CurrentID=0, Status='live', Symbol='Ctcf', Aliases='AW108038', description='CCCTC-binding factor', other_designations='11-zinc finger protein|CTCFL paralog', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105636538, end_position_on_the_genomic_accession=105682922, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  13019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13019, CurrentID=0, Status='live', Symbol='Ctf1', Aliases='CT-1', description='cardiotrophin 1', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127712704, end_position_on_the_genomic_accession=127718188, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  668381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668381, CurrentID=0, Status='live', Symbol='Scgb1b12', Aliases='Abpa12, EG668381, Gm9140', description='secretoglobin, family 1B, member 12', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=32334188, end_position_on_the_genomic_accession=32335501, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13024, CurrentID=0, Status='live', Symbol='Ctla2a', Aliases='Ctla-2a', description='cytotoxic T lymphocyte-associated protein 2 alpha', other_designations='cytotoxic T lymphocyte-associated protein 2-alpha|cytotoxic T-lymphocyte-associated protein 2-alpha', map_location='13 32.56 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60934155, end_position_on_the_genomic_accession=60936625, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13025, CurrentID=0, Status='live', Symbol='Ctla2b', Aliases='AW545568, Ctla-2b', description='cytotoxic T lymphocyte-associated protein 2 beta', other_designations='cytotoxic T lymphocyte-associated protein 2-beta|cytotoxic T-lymphocyte-associated protein 2-beta', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60895351, end_position_on_the_genomic_accession=60897932, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13026, CurrentID=0, Status='live', Symbol='Pcyt1a', Aliases='CTalpha, Cctalpha, Ctpct, Cttalpha', description='phosphate cytidylyltransferase 1, choline, alpha isoform', other_designations='CCT A|CCT-alpha|CT A|CTP phosphocholine cytidylyl transferase|CTP:phosphocholine cytidylyltransferase A|CTP:phosphocholine cytidylyltransferase alpha|phosphorylcholine transferase A', map_location='16 22.69 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32430921, end_position_on_the_genomic_accession=32475065, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13030, CurrentID=0, Status='live', Symbol='Ctsb', Aliases='CB', description='cathepsin B', other_designations='cathepsin B1', map_location='14 33.24 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63122462, end_position_on_the_genomic_accession=63145923, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13032, CurrentID=0, Status='live', Symbol='Ctsc', Aliases='AI047818, CatC, DPP1, DPPI', description='cathepsin C', other_designations='cathepsin J|dipeptidyl peptidase I|dipeptidyl transferase', map_location='7 D3-E1.1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=88278023, end_position_on_the_genomic_accession=88310875, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13033, CurrentID=0, Status='live', Symbol='Ctsd', Aliases='CD, CatD', description='cathepsin D', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142375916, end_position_on_the_genomic_accession=142387870, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13034, CurrentID=0, Status='live', Symbol='Ctse', Aliases='A430072O03Rik, C920004C08Rik, CE, CatE', description='cathepsin E', other_designations='', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131638314, end_position_on_the_genomic_accession=131675507, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13035, CurrentID=0, Status='live', Symbol='Ctsg', Aliases='CatG, VSP', description='cathepsin G', other_designations='vimentin-specific protease', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56099881, end_position_on_the_genomic_accession=56102574, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13036, CurrentID=0, Status='live', Symbol='Ctsh', Aliases='AL022844', description='cathepsin H', other_designations='Cat H|cathepsin B3|cathepsin BA', map_location='9 47.4 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=90054267, end_position_on_the_genomic_accession=90076095, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13038, CurrentID=0, Status='live', Symbol='Ctsk', Aliases='AI323530, MMS10-Q, Ms10q, catK', description='cathepsin K', other_designations='Cat K', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95499210, end_position_on_the_genomic_accession=95509387, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13039, CurrentID=0, Status='live', Symbol='Ctsl', Aliases='1190035F06Rik, CatL1, MEP, fs, nkt, Ctsl', description='cathepsin L', other_designations='Cat L|major excreted protein|p39 cysteine proteinase', map_location='13 33.26 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64363214, end_position_on_the_genomic_accession=64370883, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13040, CurrentID=0, Status='live', Symbol='Ctss', Aliases='Cats', description='cathepsin S', other_designations='Cat S', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95526786, end_position_on_the_genomic_accession=95556405, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13041, CurrentID=0, Status='live', Symbol='Ctsw', Aliases='lymphopain', description='cathepsin W', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5465240, end_position_on_the_genomic_accession=5470387, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13043, CurrentID=0, Status='live', Symbol='Cttn', Aliases='1110020L01Rik, Ems1', description='cortactin', other_designations='mammary tumor and squamous cell carcinoma associated (p80/85 src substrate)', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144435723, end_position_on_the_genomic_accession=144471004, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  13046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13046, CurrentID=0, Status='live', Symbol='Celf1', Aliases='1600010O03Rik, AA407467, Brunol2, CUG-BP, CUG-BP1, CUGBP, Cugbp1, D2Wsu101e, HNAB50, NAB50', description='CUGBP, Elav-like family member 1', other_designations='50 kDa nuclear polyadenylated RNA-binding protein|CELF-1|CUG triplet repeat RNA-binding protein 1|CUG triplet repeat, RNA binding protein 1|CUG-BP- and ETR-3-like factor 1|EDEN-BP homolog|RNA-binding protein BRUNOL-2|brain protein F41|bruno-like protein 2|deadenylation factor CUG-BP|deadenylation factor EDEN-BP|embryo deadenylation element-binding protein homolog', map_location='2 50.44 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90940387, end_position_on_the_genomic_accession=91019497, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  13047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13047, CurrentID=0, Status='live', Symbol='Cux1', Aliases='CDP, Cutl1, Cux, Cux-1', description='cut-like homeobox 1', other_designations='CCAAT displacement protein|Cux/CDP homeoprotein|cut-like 1|homeobox protein cut-like 1|homeobox protein cux-1', map_location='5 75.96 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136248135, end_position_on_the_genomic_accession=136567490, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  13048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13048, CurrentID=0, Status='live', Symbol='Cux2', Aliases='1700051K22Rik, Cutl2, Cux-2, ENSMUSG00000072641', description='cut-like homeobox 2', other_designations='homeobox protein Cux-2', map_location='5 62.02 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121857962, end_position_on_the_genomic_accession=122050562, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  13049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13049, CurrentID=0, Status='live', Symbol='cw', Aliases='thd', description='curly whiskers', other_designations='tail hair depletion', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13051, CurrentID=0, Status='live', Symbol='Cx3cr1', Aliases='', description='chemokine (C-X3-C motif) receptor 1', other_designations='C-X3-C CKR-1|chemokine (C-X3-C) receptor 1|fractalkine receptor', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120048683, end_position_on_the_genomic_accession=120068296, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13052, CurrentID=0, Status='live', Symbol='Cxadr', Aliases='2610206D03Rik, AU016810, AW553441, CAR, MCAR, MCVADR', description='coxsackie virus and adenovirus receptor', other_designations='', map_location='16 C', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=78301460, end_position_on_the_genomic_accession=78359785, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13056, CurrentID=0, Status='live', Symbol='Cyb561', Aliases='', description='cytochrome b-561', other_designations='', map_location='11 68.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105933702, end_position_on_the_genomic_accession=105945114, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13057, CurrentID=0, Status='live', Symbol='Cyba', Aliases='b558, nmf333, p22-phox, p22phox', description='cytochrome b-245, alpha polypeptide', other_designations='cytochrome b(558) alpha chain|cytochrome b558 subunit alpha|cytochrome beta-558|neutrophil cytochrome b 22 kDa polypeptide|p22 phagocyte B-cytochrome|p22 phox|superoxide-generating NADPH oxidase light chain subunit', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122424771, end_position_on_the_genomic_accession=122432965, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13058, CurrentID=0, Status='live', Symbol='Cybb', Aliases='C88302, CGD91-phox, Cgd, Cyd, Nox2, gp91-1, gp91phox', description='cytochrome b-245, beta polypeptide', other_designations='NADPH oxidase 2|cytochrome b558 subunit beta|heme-binding membrane glycoprotein gp91phox|neutrophil cytochrome b 91 kDa polypeptide|p22 phagocyte B-cytochrome', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9435254, end_position_on_the_genomic_accession=9487766, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  668419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668419, CurrentID=0, Status='live', Symbol='Scgb2b33', Aliases='Abpbg33, EG668419, Gm9161', description='secretoglobin, family 2B, member 33', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33111669, end_position_on_the_genomic_accession=33113458, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13063, CurrentID=0, Status='live', Symbol='Cycs', Aliases='', description='cytochrome c, somatic', other_designations='', map_location='6 24.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=50562563, end_position_on_the_genomic_accession=50566474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13067, CurrentID=0, Status='live', Symbol='Cyct', Aliases='T-Cc', description='cytochrome c, testis', other_designations='', map_location='2 44.94 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76353942, end_position_on_the_genomic_accession=76360453, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13070, CurrentID=0, Status='live', Symbol='Cyp11a1', Aliases='Cyp11a, Cypxia1, D9Ertd411e, P450scc, Scc, cscc', description='cytochrome P450, family 11, subfamily a, polypeptide 1', other_designations='cholesterol desmolase|cholesterol side chain cleavage|cytochrome P450 11A1|cytochrome P450 side-chain cleavage enzyme|cytochrome P450(scc)|cytochrome P450, 11a, cholesterol side-chain cleavage', map_location='9 31.63 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58015017, end_position_on_the_genomic_accession=58027023, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13072, CurrentID=0, Status='live', Symbol='Cyp11b2', Aliases='Cpn2, Cyp11b, Cyp11b-2', description='cytochrome P450, family 11, subfamily b, polypeptide 2', other_designations='CYPXIB2|aldosterone synthase|cytochrome P450C11|steroid 11-beta-hydroxylase|steroid-11-beta-hydroxylase', map_location='15 34.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74851010, end_position_on_the_genomic_accession=74856318, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  668433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668433, CurrentID=0, Status='live', Symbol='4930544D05Rik', Aliases='', description='RIKEN cDNA 4930544D05 gene', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70615834, end_position_on_the_genomic_accession=70616892, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13074, CurrentID=0, Status='live', Symbol='Cyp17a1', Aliases='Cyp17, p450c17', description='cytochrome P450, family 17, subfamily a, polypeptide 1', other_designations='17-alpha-hydroxyprogesterone aldolase|CYPXVII|cytochrome P450 17A1|cytochrome P450, 17|cytochrome P450, 17a1|cytochrome P450-C17|cytochrome P450c17|steroid 17-alpha hydroxylase', map_location='19 38.97 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46667165, end_position_on_the_genomic_accession=46673000, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13075, CurrentID=0, Status='live', Symbol='Cyp19a1', Aliases='Ar, ArKO, Cyp19, Int-5, Int5, p450arom', description='cytochrome P450, family 19, subfamily a, polypeptide 1', other_designations='CYPXIX|cytochrome P-450AROM|cytochrome P450 19A1|cytochrome P450, 19, aromatase|estrogen synthase', map_location='9 29.49 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54165862, end_position_on_the_genomic_accession=54229532, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  13076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13076, CurrentID=0, Status='live', Symbol='Cyp1a1', Aliases='AHH, AHRR, CP11, Cyp1a2, P450-1', description='cytochrome P450, family 1, subfamily a, polypeptide 1', other_designations='CYPIA1|aromatic compound inducible|cytochrome P450 family 1 subfamily a polypeptide 2|cytochrome P450 subfamily I, polypeptide 1|cytochrome P450, 1a1, aromatic compound inducible|cytochrome P450-P1', map_location='9 31.34 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57687928, end_position_on_the_genomic_accession=57703824, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13077, CurrentID=0, Status='live', Symbol='Cyp1a2', Aliases='CP12, Cyp1a1, P450-3', description='cytochrome P450, family 1, subfamily a, polypeptide 2', other_designations='CYPIA2|aromatic compound inducible|cholesterol 25-hydroxylase|cytochrome P450 family 1 subfamily a polypeptide 1|cytochrome P450, 1a2, aromatic compound inducible|cytochrome P450-P2|cytochrome P450-P3', map_location='9 31.3 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57676937, end_position_on_the_genomic_accession=57683655, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13078, CurrentID=0, Status='live', Symbol='Cyp1b1', Aliases='CP1B, P4501b1', description='cytochrome P450, family 1, subfamily b, polypeptide 1', other_designations='CYPIB1|benz[a]anthracene inducible|cytochrome P450, 1b1, benz[a]anthracene inducible|cytochrome P450CMEF|cytochrome P450EF', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=79696282, end_position_on_the_genomic_accession=79715065, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13079, CurrentID=0, Status='live', Symbol='Cyp21a1', Aliases='21-OH, 21OH, 21OHA, 21OHB, CYP21OH-A, Cyp21, Cyp21-ps1, Cyp21B, Cyp21a2-ps, Cyp21a2ps, Oh21-1, Oh21-2', description='cytochrome P450, family 21, subfamily a, polypeptide 1', other_designations='21-OHase|21-hydroxylase|cytochrome P-450|cytochrome P-450c21|cytochrome P450 21|cytochrome P450 XXI|cytochrome P450 hydroxylase A|cytochrome P450, 21, pseudogene|cytochrome P450, 21, steroid 21 hydroxylase|cytochrome P450, family 21, subfamily a, polypeptide 2, pseudogene|cytochrome P450-C21|steroid 21 hydroxylase', map_location='17 18.36 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34801348, end_position_on_the_genomic_accession=34804426, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13081, CurrentID=0, Status='live', Symbol='Cyp24a1', Aliases='24-OHase, CP24, Cyp24', description='cytochrome P450, family 24, subfamily a, polypeptide 1', other_designations='25-hydroxyvitamin D-24-hydroxylase|cytochrome P450 24A1|cytochrome P450, 24|cytochrome P450, 24a1|cytochrome P450-CC24|vitamin D(3) 24-hydroxylase', map_location='2 91.91 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=170482965, end_position_on_the_genomic_accession=170497138, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13082, CurrentID=0, Status='live', Symbol='Cyp26a1', Aliases='Cyp26, P450RA, P450RAI, RAH', description='cytochrome P450, family 26, subfamily a, polypeptide 1', other_designations='cytochrome P450, 26, retinoic acid A1|cytochrome P450RAI|retinoic acid 4-hydroxylase|retinoic acid hydrolase|retinoic acid-metabolizing cytochrome', map_location='19 C2-C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37697808, end_position_on_the_genomic_accession=37701536, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  13085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13085, CurrentID=0, Status='live', Symbol='Cyp2a12', Aliases='', description='cytochrome P450, family 2, subfamily a, polypeptide 12', other_designations='CYPIIA12|cytochrome P450, 2a12|steroid hormones 7-alpha-hydroxylase|testosterone 7-alpha-hydroxylase', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27029090, end_position_on_the_genomic_accession=27036818, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13086, CurrentID=0, Status='live', Symbol='Cyp2a4', Aliases='AI893559, Cyp15a1, D7Ucla4', description='cytochrome P450, family 2, subfamily a, polypeptide 4', other_designations='CYPIIA4|cytochrome P450, 2a4|cytochrome P450-15-alpha|cytochrome P450-IIA3.1|testosterone 15-alpha-hydroxylase|testosterone 15alpha-hydroxylase', map_location='7 14.39 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26307192, end_position_on_the_genomic_accession=26315088, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13087, CurrentID=0, Status='live', Symbol='Cyp2a5', Aliases='CYPIIA5, Coh, Cyp15a2', description='cytochrome P450, family 2, subfamily a, polypeptide 5', other_designations='coumarin 7-hydroxylase|cytochrome P450, 2a5|cytochrome P450-15-COH|cytochrome P450-IIA3.2', map_location='7 15.54 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26835339, end_position_on_the_genomic_accession=26843264, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13088, CurrentID=0, Status='live', Symbol='Cyp2b10', Aliases='Cyp2b, Cyp2b20, p16', description='cytochrome P450, family 2, subfamily b, polypeptide 10', other_designations='CYPIIB10|CYPIIB20|P24|cytochrome P450 2B20|cytochrome P450 clone PF3/46|cytochrome P450, 2b10, phenobarbitol inducible, type b|cytochrome P450, 2b20|cytochrome P450, family 2, subfamily b, polypeptide 20|cytochrome P450-16-alpha|phenobarbitol inducible, type b|testosterone 16-alpha hydroxylase', map_location='7 14.11 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25897658, end_position_on_the_genomic_accession=25926624, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13089, CurrentID=0, Status='live', Symbol='Cyp2b13', Aliases='', description='cytochrome P450, family 2, subfamily b, polypeptide 13', other_designations='cytochrome P450, 2b13, phenobarbitol inducible, type c|phenobarbital inducible, type c', map_location='7 14.11 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26061344, end_position_on_the_genomic_accession=26096197, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13090, CurrentID=0, Status='live', Symbol='Cyp2b19', Aliases='CYPIIB19', description='cytochrome P450, family 2, subfamily b, polypeptide 19', other_designations='cytochrome P450, 2b19', map_location='7 15.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26757187, end_position_on_the_genomic_accession=26772630, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13094, CurrentID=0, Status='live', Symbol='Cyp2b9', Aliases='Cyp2b', description='cytochrome P450, family 2, subfamily b, polypeptide 9', other_designations='16alphaoh-a|CYPIIB9|cytochrome P450 clone PF26|cytochrome P450, 2b9, phenobarbitol inducible, type a|cytochrome P450-16-alpha|phenobarbitol inducible, type a|testosterone 16-alpha hydroxylase', map_location='7 14.11 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26173339, end_position_on_the_genomic_accession=26210661, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  13095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13095, CurrentID=0, Status='live', Symbol='Cyp2c29', Aliases='AHOH, AHOHase, AI159681, AI196010, Ah-2, Ahh-1, Cyp2c, P450-2C', description='cytochrome P450, family 2, subfamily c, polypeptide 29', other_designations='CYPIIC29|aldehyde oxygenase|cytochrome P-450 MUT-2|cytochrome P450, 2c29', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39287044, end_position_on_the_genomic_accession=39330713, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13096, CurrentID=0, Status='live', Symbol='Cyp2c37', Aliases='', description='cytochrome P450, family 2. subfamily c, polypeptide 37', other_designations='CYPIIC37|cytochrome P450, 2c37', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39992424, end_position_on_the_genomic_accession=40012243, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13097, CurrentID=0, Status='live', Symbol='Cyp2c38', Aliases='', description='cytochrome P450, family 2, subfamily c, polypeptide 38', other_designations='CYPIIC38|cytochrome P450, 2c38', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39389556, end_position_on_the_genomic_accession=39463089, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13098, CurrentID=0, Status='live', Symbol='Cyp2c39', Aliases='CYPIIC39', description='cytochrome P450, family 2, subfamily c, polypeptide 39', other_designations='cytochrome P450, 2c39', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39510822, end_position_on_the_genomic_accession=39568529, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13099, CurrentID=0, Status='live', Symbol='Cyp2c40', Aliases='AI265721, AI662255', description='cytochrome P450, family 2, subfamily c, polypeptide 40', other_designations='CYPIIC40|cytochrome P450, 2c40', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39767073, end_position_on_the_genomic_accession=39812814, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13101, CurrentID=0, Status='live', Symbol='Cyp2d10', Aliases='AI303445, Cyp2d, P450-2D', description='cytochrome P450, family 2, subfamily d, polypeptide 10', other_designations='CYPIID10|cytochrome P450, 2d10|cytochrome P450-16-alpha|cytochrome P450CB|testosterone 16-alpha hydroxylase', map_location='15 38.58 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82402845, end_position_on_the_genomic_accession=82407208, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13105, CurrentID=0, Status='live', Symbol='Cyp2d9', Aliases='Cyp2d, P450-2D', description='cytochrome P450, family 2, subfamily d, polypeptide 9', other_designations='CYPIID9|cytochrome P450, 2d9|cytochrome P450-16-alpha|cytochrome P450CA|testosterone 16-alpha hydroxylase|testosterone 16alpha-hydroxylase', map_location='15 38.6 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82452377, end_position_on_the_genomic_accession=82456827, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13106, CurrentID=0, Status='live', Symbol='Cyp2e1', Aliases='Cyp2e', description='cytochrome P450, family 2, subfamily e, polypeptide 1', other_designations='4-nitrophenol 2-hydroxylase|CYPIIE1|cytochrome P450, 2e1, ethanol inducible|cytochrome P450-ALC|cytochrome P450-J', map_location='7 85.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140763832, end_position_on_the_genomic_accession=140774981, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13107, CurrentID=0, Status='live', Symbol='Cyp2f2', Aliases='Cyp2f', description='cytochrome P450, family 2, subfamily f, polypeptide 2', other_designations='CYPIIF2|cytochrome P450, 2f2|cytochrome P450-NAH-2|naphthalene dehydrogenase|naphthalene hydroxylase', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27119955, end_position_on_the_genomic_accession=27133660, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13108, CurrentID=0, Status='live', Symbol='Cyp2g1', Aliases='', description='cytochrome P450, family 2, subfamily g, polypeptide 1', other_designations='cytochrome P450, 2g1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26808927, end_position_on_the_genomic_accession=26821197, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13109, CurrentID=0, Status='live', Symbol='Cyp2j5', Aliases='', description='cytochrome P450, family 2, subfamily j, polypeptide 5', other_designations='CYPIIJ5|arachidonic acid epoxygenase|cytochrome P450, 2j5', map_location='4 44.71 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=96627432, end_position_on_the_genomic_accession=96664170, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13110, CurrentID=0, Status='live', Symbol='Cyp2j6', Aliases='', description='cytochrome P450, family 2, subfamily j, polypeptide 6', other_designations='CYPIIJ6|arachidonic acid epoxygenase|cytochrome P450, 2j6', map_location='4 44.62 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=96516138, end_position_on_the_genomic_accession=96560835, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13112, CurrentID=0, Status='live', Symbol='Cyp3a11', Aliases='AI256190, Cyp3a, IIIAm1, Pcn', description='cytochrome P450, family 3, subfamily a, polypeptide 11', other_designations='CYPIIIA11|cytochrome P-450IIIAM1|cytochrome P-450UT|cytochrome P450, steroid inducible 3a11|steroid inducible', map_location='5 85.17 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145854607, end_position_on_the_genomic_accession=145879854, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13113, CurrentID=0, Status='live', Symbol='Cyp3a13', Aliases='IIIAm2', description='cytochrome P450, family 3, subfamily a, polypeptide 13', other_designations='CYPIIIA13|cytochrome P450, steroid inducible 3a13', map_location='5 76.89 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137892933, end_position_on_the_genomic_accession=137921619, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13114, CurrentID=0, Status='live', Symbol='Cyp3a16', Aliases='', description='cytochrome P450, family 3, subfamily a, polypeptide 16', other_designations='CYPIIIA16|cytochrome P450, 3a16', map_location='5 85.17 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145436309, end_position_on_the_genomic_accession=145469737, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13115, CurrentID=0, Status='live', Symbol='Cyp27b1', Aliases='Cp2b, Cyp1, Cyp27b, Cyp40, P450c1, Pddr, Vdd1, Vddr, VddrI, Vdr', description='cytochrome P450, family 27, subfamily b, polypeptide 1', other_designations='1alpha(OH)ase|25(OH)D 1alpha-hydroxylase|25-OHD-1 alpha-hydroxylase|25-hydroxyvitamin D(3) 1-alpha-hydroxylase|25-hydroxyvitamin D-1 alpha hydroxylase, mitochondrial precursor (25-OHD-1 alpha-hydroxylase) (25-hydroxyvitamin D3 1-alpha-hydroxylase) (VD3 1A hydroxylase) (P450C1 alpha) (P450VD1-alpha)|25-hydroxyvitamin D3 1alpha-hydroxylase|P450VD1alpha|VD3 1A hydroxylase|calcidiol 1-monooxygenase|cytochrome P450 subfamily XXVIIB polypeptide 1|cytochrome P450, 27b1|cytochrome P450, 40 (25-hydroxyvitamin D3 1 alpha-hydroxylase)|cytochrome P450C1 alpha|cytochrome P450VD1-alpha|cytochrome p450 27B1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127048246, end_position_on_the_genomic_accession=127053422, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13116, CurrentID=0, Status='live', Symbol='Cyp46a1', Aliases='Cyp46', description='cytochrome P450, family 46, subfamily a, polypeptide 1', other_designations='CH24H|cholestrol 24-hydroxylase|cytochrome P450 46A1|cytochrome P450, 46 (cholesterol 24-hydroxylase)|cytochrome P450, 46a1 (cholesterol 24-hydroxylase)', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108334261, end_position_on_the_genomic_accession=108362248, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13117, CurrentID=0, Status='live', Symbol='Cyp4a10', Aliases='AI647584, Cyp4a, D4Rp1, RP1', description='cytochrome P450, family 4, subfamily a, polypeptide 10', other_designations='CYPIVA10|cyp4a10|cytochrome P450, 4a10|cytochrome P450-LA-omega 1|cytochrome P452|lauric acid omega-hydroxylase|long-chain fatty acid omega-monooxygenase', map_location='4 53.05 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115518287, end_position_on_the_genomic_accession=115533649, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13118, CurrentID=0, Status='live', Symbol='Cyp4a12b', Aliases='BC060945, Cyp4a12', description='cytochrome P450, family 4, subfamily a, polypeptide 12B', other_designations='cytochrome P450, family 4, subfamily a, polypeptide 12', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115411624, end_position_on_the_genomic_accession=115439034, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  13119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13119, CurrentID=0, Status='live', Symbol='Cyp4a14', Aliases='AI314743', description='cytochrome P450, family 4, subfamily a, polypeptide 14', other_designations='CYPIVA14|cyp4a14|cytochrome P450, 4a14|cytochrome P450-LA-omega 3|lauric acid omega-hydroxylase 3|long-chain fatty acid omega-monooxygenase', map_location='4 53.04 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115486200, end_position_on_the_genomic_accession=115496159, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13120, CurrentID=0, Status='live', Symbol='Cyp4b1', Aliases='', description='cytochrome P450, family 4, subfamily b, polypeptide 1', other_designations='CYPIVB1|cytochrome P450, subfamily IV B, polypeptide 1', map_location='4 53.06 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115624728, end_position_on_the_genomic_accession=115647705, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13121, CurrentID=0, Status='live', Symbol='Cyp51', Aliases='AI426508, CYPLIa1, Ldm, P450-14DM, P450LI, Cyp51', description='cytochrome P450, family 51', other_designations='cytochrome P450 51A1|cytochrome P450, 51|cytochrome P450-14DM|cytochrome P45014DM|cytochrome P450LI|lanosterol 14-alpha-demethylase|sterol 14-alpha demethylase', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=4080674, end_position_on_the_genomic_accession=4104697, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13122, CurrentID=0, Status='live', Symbol='Cyp7a1', Aliases='', description='cytochrome P450, family 7, subfamily a, polypeptide 1', other_designations='CYPVII|cholesterol 7 alpha hydroxylase|cholesterol 7-alpha-hydroxylase|cytochrome P450 7A1|cytochrome P450, 7a1', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=6264113, end_position_on_the_genomic_accession=6275631, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13123, CurrentID=0, Status='live', Symbol='Cyp7b1', Aliases='AW261589, D3Ertd552e, hct-1', description='cytochrome P450, family 7, subfamily b, polypeptide 1', other_designations='cytochrome P450 7B1|cytochrome P450, 7b1|hippocampal transcript 1 protein|oxysterol 7-alpha-hydroxylase', map_location='3 4.98 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=18071950, end_position_on_the_genomic_accession=18243893, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13124, CurrentID=0, Status='live', Symbol='Cyp8b1', Aliases='', description='cytochrome P450, family 8, subfamily b, polypeptide 1', other_designations='7-alpha-hydroxy-4-cholesten-3-one 12-alpha-hydroxylase|CYPVIIIB1|cytochrome P450 8B1|cytochrome P450, 8b1, sterol 12 alpha-hydrolase|sterol 12-alpha-hydrolase|sterol 12-alpha-hydroxylase', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121914356, end_position_on_the_genomic_accession=121916305, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  13131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13131, CurrentID=0, Status='live', Symbol='Dab1', Aliases='AI956902, C630028C02Rik, scm, scr, scrambler, yot', description='disabled 1', other_designations='yotari', map_location='4 47.79 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=103619546, end_position_on_the_genomic_accession=104744844, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  13132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13132, CurrentID=0, Status='live', Symbol='Dab2', Aliases='5730435J12Rik, AA960054, AI957090, D15Wsu122e, D630005B22Rik, Doc-2, Doc2, p96', description='disabled 2, mitogen-responsive phosphoprotein', other_designations='adaptor molecule disabled-2|differentially expressed in ovarian carcinoma 2|mitogen-responsive phosphoprotein', map_location='15 2.15 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=6299788, end_position_on_the_genomic_accession=6440715, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  13134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13134, CurrentID=0, Status='live', Symbol='Dach1', Aliases='Dac, Dach', description='dachshund 1 (Drosophila)', other_designations='', map_location='14 E3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=97786846, end_position_on_the_genomic_accession=98169765, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13135, CurrentID=0, Status='live', Symbol='Dad1', Aliases='AI323713', description='defender against cell death 1', other_designations='DAD-1|oligosaccharyl transferase subunit DAD1', map_location='14 27.7 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54235485, end_position_on_the_genomic_accession=54253929, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13136, CurrentID=0, Status='live', Symbol='Cd55', Aliases='Daf, Daf-GPI, Daf1, GPI-DAF', description='CD55 molecule, decay accelerating factor for complement', other_designations='CD55 antigen|Cromer blood group|GPI anchor addition signal|complement-glycosylphosphatidylinositol|decay accelerating factor 1', map_location='1 56.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130437196, end_position_on_the_genomic_accession=130463012, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13137, CurrentID=0, Status='live', Symbol='Cd55b', Aliases='AI323748, Daf, Daf-TM, Daf2, TM-DAF', description='CD55 molecule, decay accelerating factor for complement B', other_designations='complement-transmembrane|decay accelerating factor 2', map_location='1 56.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130388528, end_position_on_the_genomic_accession=130423037, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  13138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13138, CurrentID=0, Status='live', Symbol='Dag1', Aliases='D9Wsu13e, DG, Dp427, Dp71', description='dystroglycan 1', other_designations='dystrophin associated glycoprotein 1', map_location='9 59.08 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108204861, end_position_on_the_genomic_accession=108263958, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13139, CurrentID=0, Status='live', Symbol='Dgka', Aliases='80kDa, AW146112, Dagk1', description='diacylglycerol kinase, alpha', other_designations='80 kDa diacylglycerol kinase|DAG kinase alpha|DGK-alpha|diacylglycerol kinase alpha Flt2-Pr2|diglyceride kinase alpha|truncated diacylglycerol kinase alpha Flt1-Pr1|truncated diacylglycerol kinase alpha Flt1-Pr2', map_location='10 77.14 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128720134, end_position_on_the_genomic_accession=128748420, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  668501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668501, CurrentID=0, Status='live', Symbol='Zfp507', Aliases='1810022O10Rik, A230056M16Rik, AL024009, AU021817, BB105092, Gm480, Znf507, mKIAA1084', description='zinc finger protein 507', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35772343, end_position_on_the_genomic_accession=35802989, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13142, CurrentID=0, Status='live', Symbol='Dao', Aliases='AI987963, DAAO, DAMOX-1, Dao1, Dao', description='D-amino acid oxidase', other_designations='D-amino acid oxidase 1', map_location='5 55.93 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113999636, end_position_on_the_genomic_accession=114025675, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13143, CurrentID=0, Status='live', Symbol='Dapk2', Aliases='', description='death-associated protein kinase 2', other_designations='DAP kinase 2|DAP-kinase-related protein 1 beta isoform|DRP-1|death-associated kinase 2', map_location='9 35.75 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66158226, end_position_on_the_genomic_accession=66272242, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  13144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13144, CurrentID=0, Status='live', Symbol='Dapk3', Aliases='ZIPK, dlk', description='death-associated protein kinase 3', other_designations='DAP kinase 3|DAP-like kinase|MYPT1 kinase|ZIP kinase|ZIP-kinase|death-associated kinase 3', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81182987, end_position_on_the_genomic_accession=81193197, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67728, CurrentID=0, Status='live', Symbol='Dph2', Aliases='9130020C19Rik, AI467389l2, Dph2', description='DPH2 homolog', other_designations='DPH2-like 2|diphthamide biosynthesis protein 2 homolog-like 2|diptheria toxin resistance protein required for diphthamide biosynthesis (Saccharomyces)-like 1|diptheria toxin resistance protein required for diphthamide biosynthesis (Saccharomyces)-like 2', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117888643, end_position_on_the_genomic_accession=117892048, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13162, CurrentID=0, Status='live', Symbol='Slc6a3', Aliases='DAT, Dat1', description='solute carrier family 6 (neurotransmitter transporter, dopamine), member 3', other_designations='DA transporter|dopamine transporter 1|solute carrier family 6 member 3', map_location='13 40.1 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73535528, end_position_on_the_genomic_accession=73578672, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13163, CurrentID=0, Status='live', Symbol='Daxx', Aliases='', description='Fas death domain-associated protein', other_designations='death-domain associated protein', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33909381, end_position_on_the_genomic_accession=33915590, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13164, CurrentID=0, Status='live', Symbol='Dazl', Aliases='Daz-like, Dazh1, Dazla, Tpx-2, Tpx2, Dazl', description='deleted in azoospermia-like', other_designations='DAZ-like autosomal|deleted in azoospermia-like 1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=50279393, end_position_on_the_genomic_accession=50293620, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13166, CurrentID=0, Status='live', Symbol='Dbh', Aliases='', description='dopamine beta hydroxylase', other_designations='dopamine beta-monooxygenase', map_location='2 19.29 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27165507, end_position_on_the_genomic_accession=27183204, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  13167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13167, CurrentID=0, Status='live', Symbol='Dbi', Aliases='ACBD1, Acbp, EP, endozepine', description='diazepam binding inhibitor', other_designations='acyl-CoA binding protein|diazepam binding inhibitor, splice form 1b|diazepam-binding inhibitor', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120113280, end_position_on_the_genomic_accession=120121096, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13168, CurrentID=0, Status='live', Symbol='Dbil5', Aliases='ELP', description='diazepam binding inhibitor-like 5', other_designations='endozepine-like peptide', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76217613, end_position_on_the_genomic_accession=76218665, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13169, CurrentID=0, Status='live', Symbol='Dbnl', Aliases='Abp1, SH3P7, mAbp1', description='drebrin-like', other_designations='SH3 domain-containing protein 7|actin-binding protein 1|dbnl', map_location='11 3.87 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5788483, end_position_on_the_genomic_accession=5800980, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13170, CurrentID=0, Status='live', Symbol='Dbp', Aliases='', description='D site albumin promoter binding protein', other_designations='albumin D box-binding protein|albumin D-element-binding protein', map_location='7 29.45 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45705095, end_position_on_the_genomic_accession=45710203, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13171, CurrentID=0, Status='live', Symbol='Dbt', Aliases='D3Wsu60e', description='dihydrolipoamide branched chain transacylase E2', other_designations='BCKAD E2|BCKAD-E2|BCKADE2|branched-chain alpha-keto acid dehydrogenase complex component E2|branched-chain alpha-ketoacid dehydrogenase, E2 subunit|dihydrolipoamide acetyltransferase component of branched-chain alpha-keto acid dehydrogenase complex|dihydrolipoyl transacylase|dihydrolipoyllysine-residue (2-methylpropanoyl)transferase|part of the BCKAD complex', map_location='3 50.37 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116513079, end_position_on_the_genomic_accession=116549981, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  13172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13172, CurrentID=0, Status='live', Symbol='Dbx1', Aliases='AI426026, Dbx', description='developing brain homeobox 1', other_designations='Mmox C|developing brain homeobox protein 1', map_location='7 31.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=49631499, end_position_on_the_genomic_accession=49636835, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13175, CurrentID=0, Status='live', Symbol='Dclk1', Aliases='1700113D08Rik, 2810480F11Rik, AI836758, Click-I, Cpg16, Dcamkl1, Dcl, Dclk, mKIAA0369', description='doublecortin-like kinase 1', other_designations='CLICK-I beta|doublecortin and calcium/calmodulin-dependent protein kinase-like 1|doublecortin-like and CAM kinase-like 1', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=55242369, end_position_on_the_genomic_accession=55539068, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  13176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13176, CurrentID=0, Status='live', Symbol='Dcc', Aliases='C030036D22Rik, Igdcc1', description='deleted in colorectal carcinoma', other_designations='tumor suppressor protein DCC', map_location='18 45.24 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=71253613, end_position_on_the_genomic_accession=72351231, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  13177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13177, CurrentID=0, Status='live', Symbol='Eci1', Aliases='Dci, eci', description='enoyl-Coenzyme A delta isomerase 1', other_designations='3,2 trans-enoyl-Coenyme A isomerase|3,2-trans-enoyl-CoA isomerase, mitochondrial|D3,D2-enoyl-CoA isomerase|delta(3),Delta(2)-enoyl-CoA isomerase|dodecenoyl-CoA isomerase|dodecenoyl-Coenzyme A delta isomerase (3,2 trans-enoyl-Coenyme A isomerase)', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24426683, end_position_on_the_genomic_accession=24439316, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  13178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13178, CurrentID=0, Status='live', Symbol='Dck', Aliases='', description='deoxycytidine kinase', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88765013, end_position_on_the_genomic_accession=88783277, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  13179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13179, CurrentID=0, Status='live', Symbol='Dcn', Aliases='DC, DSPG2, PG40, PGII, PGS2, SLRR1B', description='decorin', other_designations='bone proteoglycan II', map_location='10 50.27 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=97479500, end_position_on_the_genomic_accession=97518162, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13180, CurrentID=0, Status='live', Symbol='Pcbd1', Aliases='Dcoh, Pcbd, Pcd, Phs', description='pterin 4 alpha carbinolamine dehydratase/dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1) 1', other_designations='4-alpha-hydroxy-tetrahydropterin dehydratase|6-pyruvoyl-tetrahydropterin synthase/dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1)|dimerization cofactor of HNF1|phenylalanine hydroxylase-stimulating protein|pterin carbinolamine dehydratase', map_location='10 32.14 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61089331, end_position_on_the_genomic_accession=61094329, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  668542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668542, CurrentID=0, Status='live', Symbol='Ighv1-62', Aliases='IgVH', description='immunoglobulin heavy variable V1-62', other_designations='anti-RNA antibody 564 immunoglobulin G heavy chain|anti-RNA antibody 564 immunoglobulin M heavy chain|immunoglobulin G2a gamma chain|immunoglobulin gamma-chain|low affinity anti-nucleosome binding monoclonal antibody|moderate affinity anti-nucleosome binding monoclonal antibody', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=115371986, end_position_on_the_genomic_accession=115372278, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  67733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67733, CurrentID=0, Status='live', Symbol='Itgb3bp', Aliases='4930471O16Rik, AU022583', description='integrin beta 3 binding protein (beta3-endonexin)', other_designations='CENP-R|beta3-endonexin|nuclear receptor-interacting factor 3', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=99765402, end_position_on_the_genomic_accession=99830689, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13184, CurrentID=0, Status='live', Symbol='Dcpp1', Aliases='BC005655, C86550, Dcpp, Dcpp-1, p20', description='demilune cell and parotid protein 1', other_designations='common salivary protein 1', map_location='17 12.11 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23880876, end_position_on_the_genomic_accession=23882853, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13185, CurrentID=0, Status='live', Symbol='Dscr3', Aliases='AW538125, Dcra', description='Down syndrome critical region gene 3', other_designations='Down syndrome critical region gene A|Down syndrome critical region protein A homolog', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94497724, end_position_on_the_genomic_accession=94526629, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13190, CurrentID=0, Status='live', Symbol='Dct', Aliases='DT, TRP-2, TRP2, Tyrp-2, Tyrp2, slaty, slt', description='dopachrome tautomerase', other_designations='L-dopachrome Delta-isomerase|SLATY locus protein|tyrosinase-related protein 2|tyrosinase-related protein-2', map_location='14 61.6 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118012790, end_position_on_the_genomic_accession=118052323, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13191, CurrentID=0, Status='live', Symbol='Dctn1', Aliases='AL022633, Glued, p150, p150<glued>', description='dynactin 1', other_designations='150 kDa dynein-associated polypeptide|DAP-150|DP-150|dynactin 1, retrograde axonal transport|p150-glued', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83165915, end_position_on_the_genomic_accession=83200118, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  13193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13193, CurrentID=0, Status='live', Symbol='Dcx', Aliases='Dbct', description='doublecortin', other_designations='doublecortex|doublin|lis-X|lissencephalin-X|lissencephaly, X-linked (doublecortin)', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=143855842, end_position_on_the_genomic_accession=143933270, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13194, CurrentID=0, Status='live', Symbol='Ddb1', Aliases='127kDa, AA408517, p127-Ddb1', description='damage specific DNA binding protein 1', other_designations='DDB p127 subunit|DNA repair protein|UV-damaged DNA-binding factor|damage-specific DNA-binding protein 1|damage-specific DNA-binding protein, DNA repair|damaged-DNA recognition protein 1', map_location='19 centromere', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10605568, end_position_on_the_genomic_accession=10629822, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  13195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13195, CurrentID=0, Status='live', Symbol='Ddc', Aliases='Aadc', description='dopa decarboxylase', other_designations='', map_location='11 A1-A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=11814101, end_position_on_the_genomic_accession=11898144, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  13196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13196, CurrentID=0, Status='live', Symbol='Asap1', Aliases='AV239055, DEF-1, Ddef1, PAP, mKIAA1249, s19', description='ArfGAP with SH3 domain, ankyrin repeat and PH domain1', other_designations='130 kDa phosphatidylinositol 4,5-biphosphate-dependent ARF1 GTPase-activating protein|130 kDa phosphatidylinositol 4,5-bisphosphate-dependent ARF1 GTPase-activating protein|ADP-ribosylation factor-directed GTPase-activating protein 1|ARF GTPase-activating protein 1|PIP2-dependent ARF1 GAP|development and differentiation-enhancing factor 1', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=64086840, end_position_on_the_genomic_accession=64382919, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  13197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13197, CurrentID=0, Status='live', Symbol='Gadd45a', Aliases='AA545191, Ddit1, GADD45', description='growth arrest and DNA-damage-inducible 45 alpha', other_designations='DDIT-1|DNA damage-inducible transcript 1 protein|DNA-damage inducible transcript 1|DNA-damage-inducible transcript 1|Gadd45alpha|growth arrest and DNA-damage-inducible protein GADD45 alpha|growth arrest and DNA-damage-inducible, alpha', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67035096, end_position_on_the_genomic_accession=67037407, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13198, CurrentID=0, Status='live', Symbol='Ddit3', Aliases='CHOP-10, CHOP10, chop, gadd153', description='DNA-damage inducible transcript 3', other_designations='C/EBP homoologous protein 10|C/EBP zeta|CCAAT/enhancer-binding protein homologous protein|DDIT-3|c/EBP-homologous protein 10|growth arrest and DNA-damage-inducible protein GADD153', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127290793, end_position_on_the_genomic_accession=127296288, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13199, CurrentID=0, Status='live', Symbol='Ddn', Aliases='Gm748', description='dendrin', other_designations='', map_location='15 54.69 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98803782, end_position_on_the_genomic_accession=98807925, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13200, CurrentID=0, Status='live', Symbol='Ddost', Aliases='', description='dolichyl-di-phosphooligosaccharide-protein glycotransferase', other_designations='DDOST 48 kDa subunit|oligosaccharyl transferase 48 kDa subunit', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138304738, end_position_on_the_genomic_accession=138312611, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  13202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13202, CurrentID=0, Status='live', Symbol='Ddt', Aliases='C78655', description='D-dopachrome tautomerase', other_designations='', map_location='10 38.58 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75771233, end_position_on_the_genomic_accession=75773374, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13204, CurrentID=0, Status='live', Symbol='Dhx15', Aliases='DBP1, DEAH9, Ddx15, HRH2, mDEAH9', description='DEAH (Asp-Glu-Ala-His) box polypeptide 15', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 15 (RNA helicase A)|DEAH box protein 15|putative pre-mRNA-splicing factor ATP-dependent RNA helicase DHX15', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52150203, end_position_on_the_genomic_accession=52190546, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13205, CurrentID=0, Status='live', Symbol='Ddx3x', Aliases='D1Pas1-rs2, Ddx3, Fin14', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 3, X-linked', other_designations='D-E-A-D (aspartate-glutamate-alanine-aspartate) box polypeptide 3|D1Pas1-related sequence 2|DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 3|DEAD box RNA helicase DEAD3|DEAD box protein 3, X-chromosomal|DEAD-box protein 3 (DEAD-box RNA helicase DEAD3) (mDEAD3) (Embryonic RNA helicase) (D1PAS1 related sequence 2)|embryonic RNA helicase|fibroblast growth factor inducible 14|mDEAD3', map_location='X 8.17 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=13280496, end_position_on_the_genomic_accession=13293983, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  13206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13206, CurrentID=0, Status='live', Symbol='Ddx4', Aliases='AV206478, Mvh, VASA', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 4', other_designations="D-E-A-D (aspartate-glutamate-alanine-aspartate) box polypeptide 4|DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 4|DEAD box polypeptide 4|DEAD box protein 4|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 4|mvh / m'vasa|vasa homolog", map_location='13 63.87 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112598333, end_position_on_the_genomic_accession=112652629, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  13207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13207, CurrentID=0, Status='live', Symbol='Ddx5', Aliases='2600009A06Rik, G17P1, HUMP68, Hlr1, p68', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 5', other_designations='D-E-A-D (aspartate-glutamate-alanine-aspartate) box polypeptide 5|DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 5|DEAD box RNA helicase DEAD1|DEAD box protein 5|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 5|RNA helicase p68|mDEAD1|p68 RNA helicase', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106780355, end_position_on_the_genomic_accession=106789268, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  13209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13209, CurrentID=0, Status='live', Symbol='Ddx6', Aliases='1110001P04Rik, E230023J21Rik, HLR2, mRCK/P54, p54, rck', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 6', other_designations='ATP-dependent RNA helicase p54|D-E-A-D (aspartate-glutamate-alanine-aspartate) box polypeptide 6|DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 6|DEAD box protein 6|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 6|oncogene RCK homolog', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44602716, end_position_on_the_genomic_accession=44640731, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  13211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13211, CurrentID=0, Status='live', Symbol='Dhx9', Aliases='AI326842, Ddx9, HEL-5, NDHII, RHA, mHEL-5', description='DEAH (Asp-Glu-Ala-His) box polypeptide 9', other_designations='D-E-A-D (aspartate-glutamate-alanine-aspartate) box polypeptide 9|DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 9|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 9|DEAH box protein 9|RNA helicase|leukophysin|nuclear DNA helicase II', map_location='1 65.37 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153455756, end_position_on_the_genomic_accession=153487686, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  13214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13214, CurrentID=0, Status='live', Symbol='Defb1', Aliases='AW260221, BD-1', description='defensin beta 1', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21776555, end_position_on_the_genomic_accession=21795185, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13215, CurrentID=0, Status='live', Symbol='Defb2', Aliases='BD-2', description='defensin beta 2', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21839926, end_position_on_the_genomic_accession=21843482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13216, CurrentID=0, Status='live', Symbol='Defa1', Aliases='Defcr, Defcr1', description='defensin, alpha 1', other_designations='cryptdin-1|defensin related cryptdin peptide 1|defensin-1|defensin-related cryptdin peptide', map_location='8', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13218, CurrentID=0, Status='live', Symbol='Defa-rs1', Aliases='2010300L12Rik, 2010319H24Rik, CRS1C, CRS1C-2, CRS1C-4, CRS1C-5, Defcr-rs1', description='defensin, alpha, related sequence 1', other_designations='cryptdin-related protein 1C|defensin related sequence cryptdin peptide (paneth cells)|defensin-related cryptdin-related sequence 1|testis defensin', map_location='8 10.35 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21325888, end_position_on_the_genomic_accession=21327020, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13222, CurrentID=0, Status='live', Symbol='Defa-rs2', Aliases='CRS4C, CRS4C-1, CRS4C1d, CRS4C1e, CRS4C1f, CRS4C1g, CRS4C1h, CRS4C1i, CRS4C1j, Defcr-rs2, Defcr-rs3', description='defensin, alpha, related sequence 2', other_designations='cryptdin related sequence peptide|cryptdin-related protein 4C-1|defensin related cryptdin, related sequence 2', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13226, CurrentID=0, Status='live', Symbol='Defa-rs7', Aliases='CRS4C-2, CRS4C2, CRS4C2b, CRS4C2c, CRS4C2d, CRS4C3a, CRS4C3a2, CRS4C3b, CRS4C3c, CRS4C3d, CRS4C3e, Defcr-rs7', description='defensin, alpha, related sequence 7', other_designations='cryptdin related sequence peptide|cryptdin-related protein 4C-2|defensin related cryptdin, related sequence 7|defensin-related cryptdin-related sequence 7', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21191573, end_position_on_the_genomic_accession=21192550, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13236, CurrentID=0, Status='live', Symbol='Defa25', Aliases='Defcr-2, Defcr2, Defcr25', description='defensin, alpha, 25', other_designations='defensin related cryptdin 2|defensin-related cryptdin-25', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21084442, end_position_on_the_genomic_accession=21085285, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13237, CurrentID=0, Status='live', Symbol='Defa3', Aliases='Defa27, Defcr-3, Defcr3', description='defensin, alpha, 3', other_designations='defensin, alpha, 27|defensin-related cryptdin-3', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21287409, end_position_on_the_genomic_accession=21288377, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13238, CurrentID=0, Status='live', Symbol='Defa4', Aliases='Crp4, Defa28, Defcr4', description='defensin, alpha, 4', other_designations='alpha-defensin, 28|defensin related cryptdin 4', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13239, CurrentID=0, Status='live', Symbol='Defa5', Aliases='Defa29, Defcr29, Defcr5', description='defensin, alpha, 5', other_designations='defensin, alpha, 29|defensin-related cryptdin-5', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21297394, end_position_on_the_genomic_accession=21298375, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13240, CurrentID=0, Status='live', Symbol='Defa6', Aliases='Defcr6', description='defensin, alpha, 6', other_designations='alpha-defensin, 6|cryptdin-2|defensin related cryptdin 6|defensin-related cryptdin-6/12', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13244, CurrentID=0, Status='live', Symbol='Degs1', Aliases='AA536663, Degs, Des1, Mdes', description='delta(4)-desaturase, sphingolipid 1', other_designations='degenerative spermatocyte homolog 1|dihydroceramide desaturase', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182275770, end_position_on_the_genomic_accession=182282759, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13259, CurrentID=0, Status='live', Symbol='Del(10)1H', Aliases='Del1H', description='deletion, Chr 10, Harwell 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13268, CurrentID=0, Status='live', Symbol='Del(13)36H', Aliases='Del36H, Svea', description='deletion, Chr 13, Harwell 36', other_designations='Del(13)Svea36H|small with variable eye anomalies', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  668661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668661, CurrentID=0, Status='live', Symbol='2410002F23Rik', Aliases='', description='RIKEN cDNA 2410002F23 gene', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44246720, end_position_on_the_genomic_accession=44252319, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13321, CurrentID=0, Status='live', Symbol='Del(7)Tyr<c-3H>-p', Aliases='D7Ssb3Hp, Del(7)Tyrc-3H-p, Tyr<c-3H>', description='deletion, Chr 7, Tyr c-3H, proximal', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242519, CurrentID=0, Status='live', Symbol='Ifna12', Aliases='', description='interferon alpha 12', other_designations='IFN alpha 12|IFN-alpha-12', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88602580, end_position_on_the_genomic_accession=88603376, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  13332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13332, CurrentID=0, Status='live', Symbol='Del(Y)1Ct-d', Aliases='Del(Y)1H', description='deletion, Chr Y, 1 Cattanach, distal breakpoint', other_designations='deletion, Chr Y, Yd1, Rlll, Cattanach, distal breakpoint', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13333, CurrentID=0, Status='live', Symbol='Del(Y)1Ct-p', Aliases='Del(Y)1H', description='deletion, Chr Y, 1 Cattanach, proximal breakpoint', other_designations='deletion, Chr Y, Yd1, Rlll, Cattanach, proximal breakpoint', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13334, CurrentID=0, Status='live', Symbol='Del(Y)B10.BR-Y<del>/Ms', Aliases='Del(Y)B10.BR-Y<del>/Ma', description='deletion, Chr Y, B10.BR-Y<del>, Mishima', other_designations='Del(B10.Br-Y<del>)', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242521, CurrentID=0, Status='live', Symbol='Klhl9', Aliases='8030469P05, C530050O22Rik, ENSMUSG00000070923, mKIAA1354', description='kelch-like 9', other_designations='', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88718292, end_position_on_the_genomic_accession=88722508, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  13340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13340, CurrentID=0, Status='live', Symbol='Slc29a2', Aliases='Der12, Ent2, Hnp36', description='solute carrier family 29 (nucleoside transporters), member 2', other_designations='36 kDa hydrophobic nucleolar protein|36 kDa nucleolar protein HNP36|delayed early response gene 12|delayed-early response protein 12|equilibrative NBMPR-insensitive nucleoside transporter|equilibrative nitrobenzylmercaptopurine riboside-insensitive nucleoside transporter|nucleoside transporter, ei-type|solute carrier family 29 member 2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5023407, end_position_on_the_genomic_accession=5031972, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13345, CurrentID=0, Status='live', Symbol='Twist2', Aliases='Dermo1, bHLHa39', description='twist basic helix-loop-helix transcription factor 2', other_designations='dermis-expressed protein 1|twist homolog 2', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91801477, end_position_on_the_genomic_accession=91848027, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13346, CurrentID=0, Status='live', Symbol='Des', Aliases='', description='desmin', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75360292, end_position_on_the_genomic_accession=75368579, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13347, CurrentID=0, Status='live', Symbol='Dffa', Aliases='A330085O09Rik, DFF35, Dff45, ICAD, ICAD-L, ICAD-S', description='DNA fragmentation factor, alpha subunit', other_designations='DNA fragmentation factor 45 kDa subunit|inhibitor of CAD', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149104142, end_position_on_the_genomic_accession=149120653, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13349, CurrentID=0, Status='live', Symbol='Ackr1', Aliases='AA162249, CCBP1, CD234, Darc, Dfy, ESTM35, FY, GPD', description='atypical chemokine receptor 1 (Duffy blood group)', other_designations='Duffy antigen receptor for chemokines|Duffy blood group, chemokine receptor|duffy antigen/chemokine receptor|glycoprotein D', map_location='1 80.33 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173331886, end_position_on_the_genomic_accession=173333503, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13350, CurrentID=0, Status='live', Symbol='Dgat1', Aliases='ARAT, C75990, D15Ertd23e, Dgat', description='diacylglycerol O-acyltransferase 1', other_designations='acyl-CoA retinol O-fatty-acyltransferase|diacylglycerol acyltransferase 1|diglyceride acyltransferase|retinol O-fatty-acyltransferase', map_location='15 35.99 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76502014, end_position_on_the_genomic_accession=76512020, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  13353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13353, CurrentID=0, Status='live', Symbol='Dgcr6', Aliases='', description='DiGeorge syndrome critical region gene 6', other_designations='DiGeorge syndrome chromosome region 6|diGeorge syndrome critical region 6 homolog', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18052860, end_position_on_the_genomic_accession=18071632, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13356, CurrentID=0, Status='live', Symbol='Dgcr2', Aliases='9930034O06Rik, DGS-C, Dgsc, Idd, Lan, Sez12, mKIAA0163', description='DiGeorge syndrome critical region gene 2', other_designations='DiGeorge syndrome gene c|DiGeorge syndrome protein C|seizure-related membrane-bound adhesion protein', map_location='16 A-B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17840355, end_position_on_the_genomic_accession=17894098, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13358, CurrentID=0, Status='live', Symbol='Slc25a1', Aliases='1300019P08Rik, 2610100G11Rik, AI194714, Ctp, Dgsj, Slc20a3', description='solute carrier family 25 (mitochondrial carrier, citrate transporter), member 1', other_designations='DiGeorge syndrome gene j|citrate transport protein|solute carrier family 25 member 1|tricarboxylate carrier protein', map_location='16 11.11 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17925211, end_position_on_the_genomic_accession=17928219, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13360, CurrentID=0, Status='live', Symbol='Dhcr7', Aliases='AA409147', description='7-dehydrocholesterol reductase', other_designations='7-DHC reductase|sterol Delta(7)-reductase', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143823167, end_position_on_the_genomic_accession=143848410, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  13361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13361, CurrentID=0, Status='live', Symbol='Dhfr', Aliases='8430436I03Rik, AA607882, AI662710, AW555094', description='dihydrofolate reductase', other_designations='', map_location='13 47.64 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92354783, end_position_on_the_genomic_accession=92389053, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13363, CurrentID=0, Status='live', Symbol='Dhh', Aliases='C78960', description='desert hedgehog', other_designations='HHG-3', map_location='15 55.05 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98881630, end_position_on_the_genomic_accession=98898572, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  668725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668725, CurrentID=0, Status='live', Symbol='Mrgpra9', Aliases='EG668725, MrgA9', description='MAS-related GPR, member A9', other_designations='G-protein coupled adenine receptor', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=47234919, end_position_on_the_genomic_accession=47252848, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13367, CurrentID=0, Status='live', Symbol='Diaph1', Aliases='D18Wsu154e, Dia1, Diap1, Drf1, p140mDia', description='diaphanous related formin 1', other_designations='diaphanous homolog 1|diaphanous-related formin-1', map_location='18 19.71 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37843601, end_position_on_the_genomic_accession=37935617, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  13368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13368, CurrentID=0, Status='live', Symbol='Dffb', Aliases='40kDa, 5730477D02Rik, CAD, CPAN, DFF40, Didff', description='DNA fragmentation factor, beta subunit', other_designations='DFF-40|DNA fragmentation factor 40 kDa subunit|DNA fragmentation factor, 40 kD, beta subunit|DNase inhibited by DNA fragmentation factor|caspase-activated DNase|caspase-activated deoxyribonuclease', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=153964447, end_position_on_the_genomic_accession=153975121, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13370, CurrentID=0, Status='live', Symbol='Dio1', Aliases='5DI, D1, ITDI1, TXDI1', description='deiodinase, iodothyronine, type I', other_designations="DIOI|type 1 DI|type-I 5'-deiodinase", map_location='4 50.18 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107291465, end_position_on_the_genomic_accession=107307143, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13371, CurrentID=0, Status='live', Symbol='Dio2', Aliases='5DII, AI324267, DIOII', description='deiodinase, iodothyronine, type II', other_designations="iodothyronine type II deiodinase|type 2 DI|type-II 5' deiodinase|type-II 5'-deiodinase", map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=90724552, end_position_on_the_genomic_accession=90738438, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13380, CurrentID=0, Status='live', Symbol='Dkk1', Aliases='mdkk-1', description='dickkopf WNT signaling pathway inhibitor 1', other_designations='dickkopf homolog 1|dickkopf-1|dickkopf-like protein 1|dkk-1', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=30545876, end_position_on_the_genomic_accession=30549496, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13382, CurrentID=0, Status='live', Symbol='Dld', Aliases='', description='dihydrolipoamide dehydrogenase', other_designations='branched chain alpha-keto acid dehydrogenase complex subunit E3', map_location='12 13.43 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31331562, end_position_on_the_genomic_accession=31351471, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13383, CurrentID=0, Status='live', Symbol='Dlg1', Aliases='B130052P05Rik, Dlgh1, E-dlg/SAP97, SAP-97, SAP97, mKIAA4187', description='discs, large homolog 1 (Drosophila)', other_designations='embryo-dlg/synapse-associated protein 97', map_location='16 22.4 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31663443, end_position_on_the_genomic_accession=31873356, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  13384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13384, CurrentID=0, Status='live', Symbol='Mpp3', Aliases='6430514B01, Dlgh3', description='membrane protein, palmitoylated 3 (MAGUK p55 subfamily member 3)', other_designations='Mpp3 membrane protein, palmitoylated 3|discs large homolog 3', map_location='11 65.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101999652, end_position_on_the_genomic_accession=102027048, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  13385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13385, CurrentID=0, Status='live', Symbol='Dlg4', Aliases='Dlgh4, PSD-95, PSD95, SAP90, SAP90A', description='discs, large homolog 4 (Drosophila)', other_designations='PSD-95 alpha 2b|PSD-95 beta|SAP-90|discs large homolog 4|postsynaptic density protein 95|synapse-associated protein 90', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70018455, end_position_on_the_genomic_accession=70045532, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  13386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13386, CurrentID=0, Status='live', Symbol='Dlk1', Aliases='AW742678, DLK-1, DlkI, FA1, Ly107, Peg9, SCP1, ZOG, pG2, pref-1', description='delta-like 1 homolog (Drosophila)', other_designations='adipocyte differentiation inhibitor protein|fetal antigen 1|preadipocyte factor 1|stromal cell derived protein 1', map_location='12 E-F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109452823, end_position_on_the_genomic_accession=109463336, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  13388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13388, CurrentID=0, Status='live', Symbol='Dll1', Aliases='Delta1', description='delta-like 1 (Drosophila)', other_designations='delta like-1|drosophila Delta homolog 1', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15367354, end_position_on_the_genomic_accession=15376113, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13389, CurrentID=0, Status='live', Symbol='Dll3', Aliases='pu, pudgy', description='delta-like 3 (Drosophila)', other_designations='M-Delta-3|delta3|drosophila Delta homolog 3', map_location='7 16.67 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28293555, end_position_on_the_genomic_accession=28301785, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13390, CurrentID=0, Status='live', Symbol='Dlx1', Aliases='Dlx, Dlx-1', description='distal-less homeobox 1', other_designations='DII B', map_location='2 42.61 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71528041, end_position_on_the_genomic_accession=71533981, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13392, CurrentID=0, Status='live', Symbol='Dlx2', Aliases='AW121999, Dlx-2, Tes-1', description='distal-less homeobox 2', other_designations='DII A|homeobox protein TES-1', map_location='2 42.65 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71543408, end_position_on_the_genomic_accession=71546754, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13393, CurrentID=0, Status='live', Symbol='Dlx3', Aliases='AV237891, Dlx-3', description='distal-less homeobox 3', other_designations='DII C', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95120117, end_position_on_the_genomic_accession=95125293, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13394, CurrentID=0, Status='live', Symbol='Dlx4', Aliases='Dlx-4, Dlx7', description='distal-less homeobox 4', other_designations='DII D|distal-less homeo box 7|homeobox protein DLX-7', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95140447, end_position_on_the_genomic_accession=95146499, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13395, CurrentID=0, Status='live', Symbol='Dlx5', Aliases='AI385752', description='distal-less homeobox 5', other_designations='', map_location='6 2.83 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=6877801, end_position_on_the_genomic_accession=6882068, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13396, CurrentID=0, Status='live', Symbol='Dlx6', Aliases='', description='distal-less homeobox 6', other_designations='', map_location='6 2.83 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=6863334, end_position_on_the_genomic_accession=6867970, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  668758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668758, CurrentID=0, Status='live', Symbol='Fbxw28', Aliases='Gm9337', description='F-box and WD-40 domain protein 28', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109322886, end_position_on_the_genomic_accession=109339659, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  13400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13400, CurrentID=0, Status='live', Symbol='Dmpk', Aliases='DM, DMK, Dm15, MDPK, MT-PK', description='dystrophia myotonica-protein kinase', other_designations='DM-kinase|dystrophia myotonica kinase, B15|myotonic dystrophy protein kinase', map_location='7 9.46 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19083834, end_position_on_the_genomic_accession=19093821, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13401, CurrentID=0, Status='live', Symbol='Dmwd', Aliases='DMR-N9, Dm9', description='dystrophia myotonica-containing WD repeat motif', other_designations='dystrophia myotonica linked gene, N9|dystrophia myotonica-containing WD repeat motif protein', map_location='7 9.46 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19076089, end_position_on_the_genomic_accession=19082776, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13404, CurrentID=0, Status='live', Symbol='Dmc1', Aliases='Dmc1h, Lim15, Mei11, sgdp', description='DNA meiotic recombinase 1', other_designations='DMC1 dosage suppressor of mck1 homolog, meiosis-specific homologous recombination|disrupted meiotic cDNA 1 homolog', map_location='15 37.79 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79561497, end_position_on_the_genomic_accession=79605154, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13405, CurrentID=0, Status='live', Symbol='Dmd', Aliases='DXSmh7, DXSmh9, Dp427, Dp71, dys, mdx, pke', description='dystrophin, muscular dystrophy', other_designations='X-linked muscular dystrophy', map_location='X C', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=82310713, end_position_on_the_genomic_accession=85205050, orientation='plus', exon_count=85, OMIM='', no_hdr0=''),
  13406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13406, CurrentID=0, Status='live', Symbol='Dmp1', Aliases='AV020965, Dmp, PP', description='dentin matrix protein 1', other_designations='AG1|DMP-1|serine rich acidic phosphoprotein', map_location='5 50.61 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104197558, end_position_on_the_genomic_accession=104214109, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  13409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13409, CurrentID=0, Status='live', Symbol='Tmc1', Aliases='4933416G09Rik, Beethoven, Bth, CWEA1, dn', description='transmembrane channel-like gene family 1', other_designations='deafness protein|transmembrane cochlear-expressed protein 1|transmembrane, cochlear expressed 1', map_location='19 13.98 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=20783458, end_position_on_the_genomic_accession=21037152, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  13411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13411, CurrentID=0, Status='live', Symbol='Dnah11', Aliases='Dnahc11, b2b1203Clo, b2b1279Clo, b2b1289Clo, b2b1727Clo, b2b598Clo, iv, lrd', description='dynein, axonemal, heavy chain 11', other_designations='dynein, axon, heavy chain 11|situs inversus viscerum', map_location='12 63.25 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=117877982, end_position_on_the_genomic_accession=118199043, orientation='minus', exon_count=87, OMIM='', no_hdr0=''),
  13417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13417, CurrentID=0, Status='live', Symbol='Dnah8', Aliases='ATPase, Dnahc8, Hst6.7b, P1-Loop', description='dynein, axonemal, heavy chain 8', other_designations='axonemal beta dynein heavy chain 8|axonemal dynein heavy chain 8 Dnahc8|ciliary dynein heavy chain 8|dynein, axon, heavy chain 8', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=30626936, end_position_on_the_genomic_accession=30875762, orientation='plus', exon_count=95, OMIM='', no_hdr0=''),
  13418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13418, CurrentID=0, Status='live', Symbol='Dnajc1', Aliases='4733401K02Rik, AA960110, D230036H06Rik, Dnajl1, ERdj1, ERj1p, MTJ1', description='DnaJ heat shock protein family (Hsp40) member C1', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 1|DnaJ-like protein 1|dnaJ protein homolog MTJ1', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=18206333, end_position_on_the_genomic_accession=18392851, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13419, CurrentID=0, Status='live', Symbol='Dnase1', Aliases='AI788650, DNaseI, Dnl1', description='deoxyribonuclease I', other_designations='', map_location='16 2.37 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4036786, end_position_on_the_genomic_accession=4040024, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13421, CurrentID=0, Status='live', Symbol='Dnase1l3', Aliases='DNasegamma, Dhp2, Lsd', description='deoxyribonuclease 1-like 3', other_designations='DNase I homolog protein DHP2|DNase I-like 3|DNase gamma|LS-DNase|deoxyribonuclease I-like 3|liver and spleen DNase|macrophage specific DNase I', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=7965187, end_position_on_the_genomic_accession=7994653, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13423, CurrentID=0, Status='live', Symbol='Dnase2a', Aliases='Dnase2', description='deoxyribonuclease II alpha', other_designations='DNase II alpha|acid DNase|lysosomal DNase II', map_location='8 41.31 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84908624, end_position_on_the_genomic_accession=84911461, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13424, CurrentID=0, Status='live', Symbol='Dync1h1', Aliases='9930018I23Rik, AI894280, DHC1, DHC1a, DNCL, Dnchc1, Dnec1, Dnecl, Loa, MAP1C, P22, Swl, mKIAA0325', description='dynein cytoplasmic 1 heavy chain 1', other_designations='cytoplasmic dynein heavy chain 1|dynein heavy chain, cytosolic|dynein heavy chain, retrograde transport|dynein, cytoplasmic, heavy chain 1|legs at odd angle|sprawling', map_location='12 60.7 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110601395, end_position_on_the_genomic_accession=110666944, orientation='plus', exon_count=77, OMIM='', no_hdr0=''),
  13426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13426, CurrentID=0, Status='live', Symbol='Dync1i1', Aliases='DH IC-1, DHIC-1, DIC, Dncic1, IC74', description='dynein cytoplasmic 1 intermediate chain 1', other_designations='Dynein intermediate chain 1, cytosolic (DH IC-1) (Cytoplasmic dynein intermediate chain 1)|cytoplasmic dynein intermediate chain 1 isoform 1.6|dynein, cytoplasmic, intermediate chain 1', map_location='6 2.26 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=5725633, end_position_on_the_genomic_accession=6028039, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  13427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13427, CurrentID=0, Status='live', Symbol='Dync1i2', Aliases='3110079H08Rik, AW554389, Dncic2', description='dynein cytoplasmic 1 intermediate chain 2', other_designations='DH IC-2|cytoplasmic dynein intermediate chain 2 isoform 2.1|cytoplasmic dynein intermediate chain 2 isoform 2.2|cytoplasmic dynein intermediate chain 2 isoform 2.5|dynein intermediate chain 2, cytosolic|dynein, cytoplasmic, intermediate chain 2', map_location='2 42.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71211706, end_position_on_the_genomic_accession=71263303, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  13429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13429, CurrentID=0, Status='live', Symbol='Dnm1', Aliases='AI838169, Dnm, Ftfl, mKIAA4093', description='dynamin 1', other_designations='', map_location='2 22.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32308471, end_position_on_the_genomic_accession=32353339, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  13430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13430, CurrentID=0, Status='live', Symbol='Dnm2', Aliases='Dyn2, Udnm, b2b2159Clo', description='dynamin 2', other_designations='', map_location='9 7.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21424908, end_position_on_the_genomic_accession=21507759, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  13433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13433, CurrentID=0, Status='live', Symbol='Dnmt1', Aliases='Cxxc9, Dnmto, MCMT, MTase, Met-1, Met1, MommeD2, m.MmuI, Dnmt1', description='DNA methyltransferase (cytosine-5) 1', other_designations='DNA MTase MmuI|DNA methyltransferase MmuI', map_location='9 7.66 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20907205, end_position_on_the_genomic_accession=20959888, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  13434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13434, CurrentID=0, Status='live', Symbol='Trdmt1', Aliases='Dnmt2, Rnmt2', description='tRNA aspartic acid methyltransferase 1', other_designations='DNA (cytosine-5)-methyltransferase-like protein 2|DNA MTase homolog MmuIIP|DNA methyltransferase 2|DNA methyltransferase homolog MmuIIP|m.MmuIIP|met-2|tRNA (cytosine-5-)-methyltransferase', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=13509014, end_position_on_the_genomic_accession=13544971, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  67775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67775, CurrentID=0, Status='live', Symbol='Rtp4', Aliases='5830458K16Rik, Ifrg28', description='receptor transporter protein 4', other_designations='28 kDa interferon-responsive protein|28kD interferon alpha responsive protein|receptor transporting protein 4', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23601379, end_position_on_the_genomic_accession=23614222, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13436, CurrentID=0, Status='live', Symbol='Dnmt3b', Aliases='MmuIIIB', description='DNA methyltransferase 3B', other_designations='DNA MTase MmuIIIB|DNA methyltransferase MmuIIIB', map_location='2 A2-A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153649165, end_position_on_the_genomic_accession=153687730, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  13437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13437, CurrentID=0, Status='live', Symbol='Dnpep', Aliases='AA407814', description='aspartyl aminopeptidase', other_designations='', map_location='1 C3-C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75308565, end_position_on_the_genomic_accession=75317654, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  13445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13445, CurrentID=0, Status='live', Symbol='Cdk2ap1', Aliases='Apc10, Cdkap1, DORC1, Doc1, ST19, doc-1, p12', description='CDK2 (cyclin-dependent kinase 2)-associated protein 1', other_designations='CDK2-associated protein 1|deleted in oral cancer 1|putative oral cancer suppressor', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124345439, end_position_on_the_genomic_accession=124354628, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13446, CurrentID=0, Status='live', Symbol='Doc2a', Aliases='', description='double C2, alpha', other_designations='doc2-alpha', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126847349, end_position_on_the_genomic_accession=126852705, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  13447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13447, CurrentID=0, Status='live', Symbol='Doc2b', Aliases='', description='double C2, beta', other_designations='doc2-beta|double C2 protein beta', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75769090, end_position_on_the_genomic_accession=75796057, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13448, CurrentID=0, Status='live', Symbol='Dok1', Aliases='AW557123, p62DOK', description='docking protein 1', other_designations='downstream of tyrosine kinase 1|p62(dok)', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83030932, end_position_on_the_genomic_accession=83033471, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13449, CurrentID=0, Status='live', Symbol='Dok2', Aliases='DokR, Frip', description='docking protein 2', other_designations='IL-Four Receptor Interacting Protein|IL-four receptor-interacting protein|dok-R|dok-related protein|downstream of tyrosine kinase 2|p56 dok-2|p56(dok-2)', map_location='14 D2-D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70764528, end_position_on_the_genomic_accession=70778494, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13476, CurrentID=0, Status='live', Symbol='Reep5', Aliases='AI324241, AU022809, AW495741, DP1/TB2, Dp1, R74856, TB2/DP1', description='receptor accessory protein 5', other_designations='GP106|deleted in polyposis 1|polyposis locus protein 1 homolog|protein TB2 homolog|receptor expression enhancing protein 5', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34344885, end_position_on_the_genomic_accession=34373415, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13478, CurrentID=0, Status='live', Symbol='Dpagt1', Aliases='AU021132, Dpagt2, GPT, Gnpta', description='dolichyl-phosphate (UDP-N-acetylglucosamine) acetylglucosaminephosphotransferase 1 (GlcNAc-1-P transferase)', other_designations='G1PT|N-acetylglucosamine-1-phosphate transferase|UDP-N-acetylglucosamine lysosomal enzyme N-acetylglucosamine phosphotransferase|dolichyl-phosphate alpha-N-acetylglucosaminephosphotransferase 2|glcNAc-1-P transferase', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44326017, end_position_on_the_genomic_accession=44333606, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13479, CurrentID=0, Status='live', Symbol='Dpep1', Aliases='AI327012, MBD', description='dipeptidase 1 (renal)', other_designations='MBD-1|membrane bound dipeptidase type II|membrane-bound dipeptidase 1|microsomal dipeptidase|renal dipeptidase', map_location='8 72.08 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123186235, end_position_on_the_genomic_accession=123201812, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  13480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13480, CurrentID=0, Status='live', Symbol='Dpm1', Aliases='AI118379, AI194292', description='dolichol-phosphate (beta-D) mannosyltransferase 1', other_designations='DPM synthase subunit 1|MPD synthase|dolichol-phosphate mannose synthase subunit 1|dolichyl-phosphate beta-D-mannosyltransferase subunit 1|mannose-P-dolichol synthase subunit 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168209048, end_position_on_the_genomic_accession=168230379, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13481, CurrentID=0, Status='live', Symbol='Dpm2', Aliases='AW557993, R75484', description='dolichol-phosphate (beta-D) mannosyltransferase 2', other_designations='DPM synthase subunit 2|dolichol-phosphate mannose synthase subunit 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32570858, end_position_on_the_genomic_accession=32573571, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13482, CurrentID=0, Status='live', Symbol='Dpp4', Aliases='Cd26, Dpp-4, THAM', description='dipeptidylpeptidase 4', other_designations='DPP IV|T-cell activation antigen CD26|dipeptidyl peptidase IV|thymocyte-activating molecule', map_location='2 C2-D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=62330073, end_position_on_the_genomic_accession=62412317, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  13483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13483, CurrentID=0, Status='live', Symbol='Dpp6', Aliases='B930011P16Rik, D5Buc3, D5Buc4, D5Buc5, DPP VI, Dpp-6, Gm1377, In(5)6H-p, Peplb, Rw', description='dipeptidylpeptidase 6', other_designations='DPPX|dipeptidyl aminopeptidase-related protein|dipeptidyl peptidase IV-like protein|dipeptidyl peptidase VI|dipeptidylpeptidase VI|inversion, Chr 5, Harwell 6, proximal', map_location='5 12.92 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=26817357, end_position_on_the_genomic_accession=27727500, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  13486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13486, CurrentID=0, Status='live', Symbol='Dr1', Aliases='1700121L09Rikl, NC2, NC2beta, Dr1', description='down-regulator of transcription 1', other_designations='NC2-beta|TATA-binding protein-associated phosphoprotein|negative cofactor 2-beta', map_location='5 52.82 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108268897, end_position_on_the_genomic_accession=108280521, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13487, CurrentID=0, Status='live', Symbol='Slc26a3', Aliases='9030623B18Rik, 9130013M11Rik, AV376035, Dra', description='solute carrier family 26, member 3', other_designations='down-regulated in adenoma', map_location='12 13.45 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31390876, end_position_on_the_genomic_accession=31473922, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  13488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13488, CurrentID=0, Status='live', Symbol='Drd1', Aliases='C030036C15Rik, Drd-1a, Gpcr15, Drd1', description='dopamine receptor D1', other_designations='D1 receptor|D1A dopamine receptor|G protein-coupled receptor DRD1|dopamine D1 receptor|dopamine receptor D1A', map_location='13 28.4 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54051183, end_position_on_the_genomic_accession=54055658, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13489, CurrentID=0, Status='live', Symbol='Drd2', Aliases='D2R, Drd-2', description='dopamine receptor D2', other_designations='D2 receptor|dopamine D2 receptor|dopamine receptor 2', map_location='9 26.72 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=49340428, end_position_on_the_genomic_accession=49408174, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13490, CurrentID=0, Status='live', Symbol='Drd3', Aliases='D3R', description='dopamine receptor D3', other_designations='dopamine D3 receptor|dopamine receptor 3', map_location='16 28.44 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=43754015, end_position_on_the_genomic_accession=43823846, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13491, CurrentID=0, Status='live', Symbol='Drd4', Aliases='AW125663, D4R, Drd-4', description='dopamine receptor D4', other_designations='d(2C) dopamine receptor|dopamine D4 receptor|dopamine receptor 4', map_location='7 86.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141291976, end_position_on_the_genomic_accession=141296238, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13492, CurrentID=0, Status='live', Symbol='Drd5', Aliases='D5R, Drd-5, Drd1b, Gpcr1', description='dopamine receptor D5', other_designations='d(5) dopamine receptor|dopamine D5 receptor|dopamine receptor 5', map_location='5 20.4 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38319509, end_position_on_the_genomic_accession=38322310, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  13494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13494, CurrentID=0, Status='live', Symbol='Drg1', Aliases='AA408859, AI132520, Nedd3', description='developmentally regulated GTP binding protein 1', other_designations='DRG-1|GTP-binding protein|NEDD-3|neural precursor cell expressed developmentally down-regulated protein 3', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3249922, end_position_on_the_genomic_accession=3266386, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13495, CurrentID=0, Status='live', Symbol='Drg2', Aliases='AI255295, AI461605', description='developmentally regulated GTP binding protein 2', other_designations='DRG-2|developmentally regulated GTP-binding protein 2', map_location='11 37.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60454617, end_position_on_the_genomic_accession=60468724, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13496, CurrentID=0, Status='live', Symbol='Arid3a', Aliases='Bright, Dri1, Dril1', description='AT rich interactive domain 3A (BRIGHT-like)', other_designations='ARID domain-containing protein 3A|AT rich interactive domain 3A (Bright like)|B-cell regulator of IgH transcription|dead ringer homolog 1|dead ringer-like protein 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79927043, end_position_on_the_genomic_accession=79955018, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13497, CurrentID=0, Status='live', Symbol='Drp2', Aliases='AW495265, DRP-2', description='dystrophin related protein 2', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134404628, end_position_on_the_genomic_accession=134456574, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  13498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13498, CurrentID=0, Status='live', Symbol='Atn1', Aliases='Atr1, Drpla, atrophin-1', description='atrophin 1', other_designations='dentatorubral pallidoluysian atrophy|dentatorubral-pallidoluysian atrophy protein homolog', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124742544, end_position_on_the_genomic_accession=124756487, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13499, CurrentID=0, Status='live', Symbol='Drr1', Aliases='Drr1ab, Drr1c, Drr1', description='developmentally regulated repeat element-containing transcript 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13505, CurrentID=0, Status='live', Symbol='Dsc1', Aliases='1110020A10Rik, AI507491', description='desmocollin 1', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20083471, end_position_on_the_genomic_accession=20114773, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  13506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13506, CurrentID=0, Status='live', Symbol='Dsc2', Aliases='AW228162', description='desmocollin 2', other_designations='epithelial type 2 desmocollin', map_location='18 11.14 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20030633, end_position_on_the_genomic_accession=20059552, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  13507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13507, CurrentID=0, Status='live', Symbol='Dsc3', Aliases='5430426I24Rik', description='desmocollin 3', other_designations='', map_location='18 11.11 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=19960930, end_position_on_the_genomic_accession=20002413, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  13508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13508, CurrentID=0, Status='live', Symbol='Dscam', Aliases='4932410A21Rik', description='Down syndrome cell adhesion molecule', other_designations='', map_location='16 C', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96587452, end_position_on_the_genomic_accession=97170735, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  13510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13510, CurrentID=0, Status='live', Symbol='Dsg1a', Aliases='DG1, DGI, Dsg1, dsg1-alpha', description='desmoglein 1 alpha', other_designations='desmosomal glycoprotein I', map_location='18 11.29 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20310873, end_position_on_the_genomic_accession=20343353, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13511, CurrentID=0, Status='live', Symbol='Dsg2', Aliases='AA408168, D18Ertd293e', description='desmoglein 2', other_designations='', map_location='18 11.42 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20558116, end_position_on_the_genomic_accession=20604526, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13512, CurrentID=0, Status='live', Symbol='Dsg3', Aliases='bal', description='desmoglein 3', other_designations='130 kDa pemphigus vulgaris antigen homolog', map_location='18 11.39 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20510304, end_position_on_the_genomic_accession=20541310, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  13516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13516, CurrentID=0, Status='live', Symbol='Epyc', Aliases='Dspg3, PG-Lb, SLRR3B', description='epiphycan', other_designations='dermatan sulfate proteoglycan 3|dermatan sulphate proteoglycan 3|proteoglycan-Lb|small chondroitin/dermatan sulfate proteoglycan', map_location='10 C2-C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=97644068, end_position_on_the_genomic_accession=97681908, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13518, CurrentID=0, Status='live', Symbol='Dst', Aliases='2310001O04Rik, A830042E19Rik, AW554249, BP230, BPAG1-n, Bpag, Bpag1, Macf2, ah, athetoid, dt, mKIAA0728, nmf203, nmf339', description='dystonin', other_designations='BPA|bullous pemphigoid antigen 1|bullous pemphigoid antigen 1, isoforms 1/2/3/4|bullous pemphigoid antigen 1, isoforms 6/7|dystonia musculorum protein|hemidesmosomal plaque protein|microtubule actin cross-linking factor 2', map_location='1 12.91 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=33908208, end_position_on_the_genomic_accession=34308662, orientation='plus', exon_count=111, OMIM='', no_hdr0=''),
  668880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668880, CurrentID=0, Status='live', Symbol='Stard9', Aliases='4831403C07Rik, E230025N21Rik, Kif16a, mKIAA1300', description='START domain containing 9', other_designations='N-3 kinesin|START domain-containing protein 9|StAR-related lipid transfer protein 9|kinesin 16A|kinesin family member 16A|kinesin-like protein Kif16a', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120629091, end_position_on_the_genomic_accession=120716086, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  13521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13521, CurrentID=0, Status='live', Symbol='Slc26a2', Aliases='Dtd, ST-OB', description='solute carrier family 26 (sulfate transporter), member 2', other_designations='diastrophic dysplasia protein homolog|solute carrier family 26 member 2', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61196854, end_position_on_the_genomic_accession=61211596, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13522, CurrentID=0, Status='live', Symbol='Adam28', Aliases='C130072N01Rik, D430033C21Rik, Dtgn1, MDC-L, MDCL, TECADAM, eMDCII', description='a disintegrin and metallopeptidase domain 28', other_designations='ADAM 28|a disintegrin and metalloprotease domain 28|a disintegrin and metalloproteinase domain 28|disintegrin 1|metalloproteinase-disintegrin domain containing protein TECADAM|thymic epithelial cell-ADAM', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=68604994, end_position_on_the_genomic_accession=68655887, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  13524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13524, CurrentID=0, Status='live', Symbol='Adam18', Aliases='Adam27, Dtgn3, Tmdc3, tMDCIII', description='a disintegrin and metallopeptidase domain 18', other_designations='ADAM 18|ADAM 27|disintegrin 3|disintegrin and metalloproteinase domain-containing protein 27|tMDC III|transmembrane metalloproteinase-like, disintegrin-like, and cysteine-rich protein III', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24602246, end_position_on_the_genomic_accession=24674770, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  67790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67790, CurrentID=0, Status='live', Symbol='Rab39b', Aliases='6330580M05Rik', description='RAB39B, member RAS oncogene family', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75572045, end_position_on_the_genomic_accession=75578231, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13526, CurrentID=0, Status='live', Symbol='Adam24', Aliases='Dtgn5', description='a disintegrin and metallopeptidase domain 24 (testase 1)', other_designations='ADAM 24|a disintegrin and metalloproteinase domain 24 (testase 1)|disintegrin 5|testase 1|testase-1', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40675094, end_position_on_the_genomic_accession=40682199, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13527, CurrentID=0, Status='live', Symbol='Dtna', Aliases='2210407P21Rik, DTN-A, Dtn, Gm19389, a-DB-1, adbn', description='dystrobrevin alpha', other_designations='alpha-dystrobrevin', map_location='18 12.08 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=23310020, end_position_on_the_genomic_accession=23659719, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  13528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13528, CurrentID=0, Status='live', Symbol='Dtnb', Aliases='', description='dystrobrevin, beta', other_designations='DTN-B|beta-dystrobrevin|mDTN-B', map_location='12 1.88 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3572391, end_position_on_the_genomic_accession=3782611, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  13529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13529, CurrentID=0, Status='live', Symbol='Prl8a2', Aliases='D/tPRP, DPRP, Dtprp, Ghd11, mdPRP', description='prolactin family 8, subfamily a, member 2', other_designations='decidual/trophoblast prolactin-related protein|growth hormone d11', map_location='13 12.41 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27345673, end_position_on_the_genomic_accession=27354216, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13531, CurrentID=0, Status='live', Symbol='Usp17la', Aliases='Dub1, USP17-A', description='ubiquitin specific peptidase 17-like A', other_designations='deubiquitinating enzyme 1|ubiquitin carboxyl-terminal hydrolase DUB-1|ubiquitin thioesterase DUB-1|ubiquitin thiolesterase DUB-1|ubiquitin-specific-processing protease DUB-1', map_location='7 55.69 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104857016, end_position_on_the_genomic_accession=104862667, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13532, CurrentID=0, Status='live', Symbol='Usp17lc', Aliases='Dub-2, Dub2, Dub2b, Usp17l5', description='ubiquitin specific peptidase 17-like C', other_designations='ubiquitin specific peptidase 17-like 5', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103416696, end_position_on_the_genomic_accession=103419174, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  668894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668894, CurrentID=0, Status='live', Symbol='Gm14025', Aliases='OTTMUSG00000015417', description='predicted gene 14025', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129024141, end_position_on_the_genomic_accession=129065905, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  13537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13537, CurrentID=0, Status='live', Symbol='Dusp2', Aliases='PAC1', description='dual specificity phosphatase 2', other_designations='dual specificity protein phosphatase PAC-1', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127336159, end_position_on_the_genomic_accession=127338377, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13542, CurrentID=0, Status='live', Symbol='Dvl1', Aliases='Dvl, mKIAA4029', description='dishevelled segment polarity protein 1', other_designations='DSH homolog 1|dishevelled, dsh homolog 1|dishevelled-1', map_location='4 87.61 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155847317, end_position_on_the_genomic_accession=155859303, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  13543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13543, CurrentID=0, Status='live', Symbol='Dvl2', Aliases='', description='dishevelled segment polarity protein 2', other_designations='DSH homolog 2|dishevelled 2, dsh homolog|dishevelled-2', map_location='11 42.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70000626, end_position_on_the_genomic_accession=70010109, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13544, CurrentID=0, Status='live', Symbol='Dvl3', Aliases='', description='dishevelled segment polarity protein 3', other_designations='DSH homolog 3|dishevelled 3, dsh homolog|dishevelled-3', map_location='16 12.46 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20516938, end_position_on_the_genomic_accession=20532548, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  13548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13548, CurrentID=0, Status='live', Symbol='Dyrk1a', Aliases='2310043O08Rik, D16Ertd272e, D16Ertd493e, Dyrk, ENSMUSG00000074897, Gm10783, Mnbh, Mp86, mmb', description='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 1a', other_designations='dual specificity YAK1-related kinase|protein kinase minibrain homolog', map_location='16 55.3 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94569951, end_position_on_the_genomic_accession=94695519, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13549, CurrentID=0, Status='live', Symbol='Dyrk1b', Aliases='Mirk', description='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 1b', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28179483, end_position_on_the_genomic_accession=28187793, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13555, CurrentID=0, Status='live', Symbol='E2f1', Aliases='E2F-1, mKIAA4009', description='E2F transcription factor 1', other_designations='', map_location='2 76.79 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154559400, end_position_on_the_genomic_accession=154569892, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  668917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668917, CurrentID=0, Status='live', Symbol='Zfp133-ps', Aliases='OTTMUSG00000003825, Zfp133', description='zinc finger protein 133, pseudogene', other_designations='', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144459280, end_position_on_the_genomic_accession=144467794, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13559, CurrentID=0, Status='live', Symbol='E2f5', Aliases='AU024671, E2F-5', description='E2F transcription factor 5', other_designations='', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14578602, end_position_on_the_genomic_accession=14606309, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13560, CurrentID=0, Status='live', Symbol='E4f1', Aliases='p120E4F, phi-AP3', description='E4F transcription factor 1', other_designations='putative E3 ubiquitin-protein ligase E4F1|transcription factor E4F|transcription factor phi AP3', map_location='17 12.39 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24443778, end_position_on_the_genomic_accession=24455590, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67797, CurrentID=0, Status='live', Symbol='Snrnp48', Aliases='1110050F08Rik, 6530403A03Rik, C6orf151', description='small nuclear ribonucleoprotein 48 (U11/U12)', other_designations='U11/U12 snRNP 48 kDa protein', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38204911, end_position_on_the_genomic_accession=38227663, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  668929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668929, CurrentID=0, Status='live', Symbol='Rad21l', Aliases='Gm14160, MeiRAD21L1, Rad21l', description='RAD21-like (S. pombe)', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151644366, end_position_on_the_genomic_accession=151670622, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  668940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668940, CurrentID=0, Status='live', Symbol='Myh7b', Aliases='Myh14', description='myosin, heavy chain 7B, cardiac muscle, beta', other_designations='myosin heavy chain 7B, cardiac muscle beta isoform', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155611243, end_position_on_the_genomic_accession=155634307, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  13586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13586, CurrentID=0, Status='live', Symbol='Ear1', Aliases='EAR-1, ECP 1, Raf1', description='eosinophil-associated, ribonuclease A family, member 1', other_designations='RNase 3-1|eosinophil secondary granule ribonuclease 1|eosinophil-associated ribonuclease 1|ribonuclease 3-1|ribonuclease A f1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=43818765, end_position_on_the_genomic_accession=43819639, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13587, CurrentID=0, Status='live', Symbol='Ear2', Aliases='EAR-13, EAR-2, ECP, ECP 2, Raf3, Rnase2', description='eosinophil-associated, ribonuclease A family, member 2', other_designations='RNase 3-2|eosinophil secondary granule ribonuclease 2|eosinophil-associated ribonuclease 13|eosinophil-associated ribonuclease 2|eosinophil-derived neurotoxin, liver|ribonuclease 3-2|ribonuclease A f3', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=44102654, end_position_on_the_genomic_accession=44103531, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13589, CurrentID=0, Status='live', Symbol='Mapre1', Aliases='5530600P05Rik, AI462499, AI504412, AW260097, BIM1p, D2Ertd459e, Eb1', description='microtubule-associated protein, RP/EB family, member 1', other_designations='APC-binding protein EB1|adenomatosis polyposis coli binding protein Eb1|end-binding protein 1', map_location='2 75.95 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153741287, end_position_on_the_genomic_accession=153773317, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  13590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13590, CurrentID=0, Status='live', Symbol='Lefty1', Aliases='AI450052, Ebaf, Leftb, Stra3, Tgfb4, lefty, lefty-1', description='left right determination factor 1', other_designations='TGF-beta-4|left-right determination, factor B|protein lefty-1|stimulated by retinoic acid gene 3 protein|transforming growth factor beta-4', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180935039, end_position_on_the_genomic_accession=180938401, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13591, CurrentID=0, Status='live', Symbol='Ebf1', Aliases='Ebf, O/E-1, OE-1, Olf-1, Olf1', description='early B cell factor 1', other_designations='early B-cell factor', map_location='11 26.45 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=44618100, end_position_on_the_genomic_accession=45008096, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  13592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13592, CurrentID=0, Status='live', Symbol='Ebf2', Aliases='D14Ggc1e, EBF-2, Mmot1, O/E-3, OE-3', description='early B cell factor 2', other_designations='MET-mesencephalon-olfactory TF1|MET-mesencephalon-olfactory transcription factor 1|Olf-1/EBF-like 3|metencephalon-mesencephalon-olfactory transcription factor 1', map_location='14 D-E1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=67233292, end_position_on_the_genomic_accession=67430952, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  13593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13593, CurrentID=0, Status='live', Symbol='Ebf3', Aliases='3110018A08Rik, O/E-2, mKIAA4201', description='early B cell factor 3', other_designations='EBF-3|OE-2|Olf-1/EBF-like 2|early B-cell factor 3', map_location='7 F3-F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=137193671, end_position_on_the_genomic_accession=137322810, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  13595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13595, CurrentID=0, Status='live', Symbol='Ebp', Aliases='AI255399, Pabp, SI, Td, mSI', description='phenylalkylamine Ca2+ antagonist (emopamil) binding protein', other_designations='D8-D7 sterol isomerase|cholestenol Delta-isomerase|delta(8)-Delta(7) sterol isomerase|tattered', map_location='X 3.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8185331, end_position_on_the_genomic_accession=8193512, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13599, CurrentID=0, Status='live', Symbol='Ecel1', Aliases='DINE, xce', description='endothelin converting enzyme-like 1', other_designations='damage-induced neuronal endopeptidase', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87142874, end_position_on_the_genomic_accession=87157085, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  13600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13600, CurrentID=0, Status='live', Symbol='Eclb', Aliases='ecl', description='epistatic circling B C57L/J', other_designations='', map_location='4 69.05 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13601, CurrentID=0, Status='live', Symbol='Ecm1', Aliases='AI663821, p85', description='extracellular matrix protein 1', other_designations='secretory component p85', map_location='3 40.91 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95734147, end_position_on_the_genomic_accession=95739819, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  13602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13602, CurrentID=0, Status='live', Symbol='Sparcl1', Aliases='Ecm2, Sc1, hevin, mast9', description='SPARC-like 1', other_designations='SPARC-like 1 (mast9, hevin)|extracellular matrix protein 2|matrix glycoprotein Sc1', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104079108, end_position_on_the_genomic_accession=104114088, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13603, CurrentID=0, Status='live', Symbol='Opn3', Aliases='ERO, Ecpn', description='opsin 3', other_designations='encephalopsin|opsin (encephalopsin)|panopsin', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175662421, end_position_on_the_genomic_accession=175692902, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13605, CurrentID=0, Status='live', Symbol='Ect2', Aliases='AI528536, mKIAA4037', description='ect2 oncogene', other_designations='epithelial cell-transforming sequence 2 oncogene', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=27097222, end_position_on_the_genomic_accession=27153878, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  13607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13607, CurrentID=0, Status='live', Symbol='Eda', Aliases='EDA1, Ed1-A1, Eda-A2, HED, Ta, Tnlg7c, XLHED, tabby, Eda', description='ectodysplasin-A', other_designations='EDA protein homolog|tumor necrosis factor ligand 7c', map_location='X 43.59 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=99975606, end_position_on_the_genomic_accession=100400760, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13608, CurrentID=0, Status='live', Symbol='Edar', Aliases='ED1R, ED3, ED5, EDA-A1R, EDA3, dl', description='ectodysplasin-A receptor', other_designations='anhidrotic ectodysplasin receptor 1|downless|ectodermal dysplasia receptor', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=58600780, end_position_on_the_genomic_accession=58675696, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13609, CurrentID=0, Status='live', Symbol='S1pr1', Aliases='AI849002, Edg1, Lpb1, S1p, S1p1', description='sphingosine-1-phosphate receptor 1', other_designations='S1P receptor 1|S1P receptor Edg-1|endothelial differentiation G-protein coupled receptor 1|endothelial differentiation sphingolipid G-protein-coupled receptor 1|lysophospholipid receptor B1|sphingosine 1-phosphate receptor Edg-1', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=115710433, end_position_on_the_genomic_accession=115715055, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13610, CurrentID=0, Status='live', Symbol='S1pr3', Aliases='AI132464, Edg3, Lpb3, S1p3', description='sphingosine-1-phosphate receptor 3', other_designations='S1P receptor Edg-3|S1p receptor 3|endothelial differentiation G-protein coupled receptor 3|endothelial differentiation sphingolipid G-protein-coupled receptor 3|lysophospholipid receptor B3|sphingosine 1-phosphate receptor Edg-3', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51408618, end_position_on_the_genomic_accession=51422797, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13611, CurrentID=0, Status='live', Symbol='S1pr4', Aliases='Edg6, Lpc1, S1p4', description='sphingosine-1-phosphate receptor 4', other_designations='S1P receptor Edg-6|S1p receptor 4|endothelial differentiation G-protein coupled receptor 6|endothelial differentiation G-protein-coupled receptor 6|lysophospholipid receptor C1|sphingosine 1-phosphate receptor Edg-6', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81497745, end_position_on_the_genomic_accession=81500137, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  13612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13612, CurrentID=0, Status='live', Symbol='Edil3', Aliases='Del-1, Del1', description='EGF-like repeats and discoidin I-like domains 3', other_designations='developmental endothelial locus-1|developmentally-regulated endothelial cell locus 1 protein|integrin-binding protein DEL1', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=88821472, end_position_on_the_genomic_accession=89323225, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  13614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13614, CurrentID=0, Status='live', Symbol='Edn1', Aliases='ET-1, PPET1, preproET', description='endothelin 1', other_designations='preproendothelin-1', map_location='13 20.82 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=42301270, end_position_on_the_genomic_accession=42307989, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13615, CurrentID=0, Status='live', Symbol='Edn2', Aliases='ET-2, PPET2, VIC', description='endothelin 2', other_designations='preproendothelin-2|vasoactive intestinal contractor peptide', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120161424, end_position_on_the_genomic_accession=120167360, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13616, CurrentID=0, Status='live', Symbol='Edn3', Aliases='ET-3, PPET3, ls, tmgc48', description='endothelin 3', other_designations='preproendothelin-3', map_location='2 98.1 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174760758, end_position_on_the_genomic_accession=174784042, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13617, CurrentID=0, Status='live', Symbol='Ednra', Aliases='ET-AR, ETa, Gpcr10', description='endothelin receptor type A', other_designations='ET-A|G-protein coupled receptor 10|endothelin A receptor', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=77663029, end_position_on_the_genomic_accession=77724452, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13618, CurrentID=0, Status='live', Symbol='Ednrb', Aliases='ET-B, ET-BR, ETR-b, ETb, Sox10m1', description='endothelin receptor type B', other_designations='endothelin receptor non-selective type|piebald', map_location='14 53.05 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=103814615, end_position_on_the_genomic_accession=103844476, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13619, CurrentID=0, Status='live', Symbol='Phc1', Aliases='AW557034, Edr, Edr1, Mph1, Rae-28, rae28', description='polyhomeotic-like 1 (Drosophila)', other_designations='early development regulator1 (homolog of polyhomeotic 1)|early development regulatory protein 1', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122317731, end_position_on_the_genomic_accession=122340227, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  13623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13623, CurrentID=0, Status='live', Symbol='Edv', Aliases='', description='endogenous sequence related to the Duplan murine retrovirus', other_designations='murine leukemia virus gag, pol, and env pseudogene', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13626, CurrentID=0, Status='live', Symbol='Eed', Aliases='ENSMUSG00000039373, l(7)5Rn, l7Rn5, lusk', description='embryonic ectoderm development', other_designations='lethal, Chr 7, Rinchik 5', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=89954654, end_position_on_the_genomic_accession=89981419, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13627, CurrentID=0, Status='live', Symbol='Eef1a1', Aliases='', description='eukaryotic translation elongation factor 1 alpha 1', other_designations='EF-1-alpha-1|EF-Tu|eEF1A-1|elongation factor 1 A-1|elongation factor Tu|eukaryotic elongation factor 1 A-1', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78478453, end_position_on_the_genomic_accession=78481724, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13628, CurrentID=0, Status='live', Symbol='Eef1a2', Aliases='Eef1a, S1, wasted, wst', description='eukaryotic translation elongation factor 1 alpha 2', other_designations='EF-1-alpha-2|eEF1A-2|eukaryotic elongation factor 1 A-2|statin-S1', map_location='2 103.6 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181147692, end_position_on_the_genomic_accession=181157015, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13629, CurrentID=0, Status='live', Symbol='Eef2', Aliases='Ef-2', description='eukaryotic translation elongation factor 2', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81176631, end_position_on_the_genomic_accession=81182509, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  13631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13631, CurrentID=0, Status='live', Symbol='Eef2k', Aliases='C86191, eEF-2K', description='eukaryotic elongation factor-2 kinase', other_designations='calcium/calmodulin-dependent eukaryotic elongation factor 2 kinase|eEF-2 kinase|eukaroytic elongation factor 2 kinase', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120842831, end_position_on_the_genomic_accession=120907219, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  13636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13636, CurrentID=0, Status='live', Symbol='Efna1', Aliases='AI325262, B61, Efl1, Epl1, Eplg1, Lerk1', description='ephrin A1', other_designations='EPH-related receptor tyrosine kinase ligand 1|immediate early response protein B61', map_location='3 39.04 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89271730, end_position_on_the_genomic_accession=89280951, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13637, CurrentID=0, Status='live', Symbol='Efna2', Aliases='CEK7L, Elf1, Epl6, Eplg6, Lerk6', description='ephrin A2', other_designations='CEK7-ligand|EPH related receptor tyrosine kinase ligand 6|EPH-related receptor tyrosine kinase ligand 6|ephrin A6', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80179482, end_position_on_the_genomic_accession=80190010, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13638, CurrentID=0, Status='live', Symbol='Efna3', Aliases='AW494418, EFL-2, Ehk1-L, Epl3, LERK-3', description='ephrin A3', other_designations='EHK1 ligand|EPH-related receptor tyrosine kinase ligand 3', map_location='3 39.06 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89313899, end_position_on_the_genomic_accession=89322960, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100862177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100862177, CurrentID=0, Status='live', Symbol='Gm21541', Aliases='', description='predicted gene, 21541', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41903613, end_position_on_the_genomic_accession=41904743, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13640, CurrentID=0, Status='live', Symbol='Efna5', Aliases='AL-1, AV158822, EFL-5, Ephrin-A5, Epl7, LERK-7, RAGS', description='ephrin A5', other_designations='EPH-related receptor tyrosine kinase ligand 7', map_location='17 32.57 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=62602957, end_position_on_the_genomic_accession=62881317, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13641, CurrentID=0, Status='live', Symbol='Efnb1', Aliases='Cek5-L, EFL-3, Elk-L, Epl2, Eplg2, LERK-2, Lerk2, Stra1', description='ephrin B1', other_designations='CEK5 receptor ligand|Cek ligand|ELK ligand|EPH-related receptor tyrosine kinase ligand 2|stimulated by retinoic acid gene 1 protein', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=99136130, end_position_on_the_genomic_accession=99148991, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13642, CurrentID=0, Status='live', Symbol='Efnb2', Aliases='ELF-2, Epl5, Eplg5, Htk-L, LERK-5, Lerk5, NLERK-1', description='ephrin B2', other_designations='EPH-related receptor tyrosine kinase ligand 5|HTK ligand', map_location='8 3.42 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=8617235, end_position_on_the_genomic_accession=8660773, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13643, CurrentID=0, Status='live', Symbol='Efnb3', Aliases='EFL-6, ELF-3, Elk-L3, Epl8, LERK-8, NLERK-2', description='ephrin B3', other_designations='', map_location='11 42.8 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69554092, end_position_on_the_genomic_accession=69560595, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13644, CurrentID=0, Status='live', Symbol='Efs', Aliases='', description='embryonal Fyn-associated substrate', other_designations='SRC-interacting protein|signal-integrating protein', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54916543, end_position_on_the_genomic_accession=54926788, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13645, CurrentID=0, Status='live', Symbol='Egf', Aliases='AI790464', description='epidermal growth factor', other_designations='Pro-epidermal growth factor precursor (EGF)', map_location='3 58.5 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129677569, end_position_on_the_genomic_accession=129755322, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  13646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13646, CurrentID=0, Status='live', Symbol='Klk1b22', Aliases='Egfbp-1, Egfbp1, Klk22, mGk-22', description='kallikrein 1-related peptidase b22', other_designations='EGF-BP A|beta-NGF-endopeptidase|epidermal growth factor binding protein type A|epidermal growth factor-binding protein type A|glandular kallikrein K22|kallikrein 22|nerve growth factor beta chain endopeptidase|tissue kallikrein 22', map_location='7 28.61 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44112683, end_position_on_the_genomic_accession=44116876, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13647, CurrentID=0, Status='live', Symbol='Egfbp2', Aliases='Egfbp-2, Klk1b26, mGk-13', description='epidermal growth factor binding protein type B', other_designations='EGF-BP B|PRECE-1|glandular kallikrein K13|kallikrein 1b26|prorenin-converting enzyme 1|tissue kallikrein 13', map_location='7', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13648, CurrentID=0, Status='live', Symbol='Klk1b9', Aliases='Egfbp-3, Egfbp3, Klk9, mGk-9', description='kallikrein 1-related peptidase b9', other_designations='EGF-BP C|epidermal growth factor binding protein type C|epidermal growth factor-binding protein type C|glandular kallikrein K9|kallikrein 9|tissue kallikrein-9', map_location='7 28.45 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43976061, end_position_on_the_genomic_accession=43980376, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13649, CurrentID=0, Status='live', Symbol='Egfr', Aliases='9030024J15Rik, AI552599, Erbb, Errb1, Errp, Wa5, wa-2, wa2', description='epidermal growth factor receptor', other_designations='avian erythroblastic leukemia viral (v-erb-b) oncogene homolog|waved 2', map_location='11 A1-A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=16752203, end_position_on_the_genomic_accession=16913907, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  13650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13650, CurrentID=0, Status='live', Symbol='Rhbdf1', Aliases='C16ORF8, Dist, Dist1, Egfr-rs, mKIAA4242', description='rhomboid family 1 (Drosophila)', other_designations='epidermal growth factor receptor related sequence|epidermal growth factor receptor-related protein|iRhom1|rhomboid family member 1', map_location='11 18.83 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32209584, end_position_on_the_genomic_accession=32222331, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  13653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13653, CurrentID=0, Status='live', Symbol='Egr1', Aliases='A530045N19Rik, ETR103, Egr-1, Krox-1, Krox-24, Krox24, NGF1-A, NGFI-A, NGFIA, TIS8, Zenk, Zfp-6, Zif268, egr', description='early growth response 1', other_designations='Krox-24 nuclear protein|nerve growth factor-induced protein A|transcription factor Zif268|zinc finger protein Krox-24', map_location='18 C/D', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34861207, end_position_on_the_genomic_accession=34864956, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13654, CurrentID=0, Status='live', Symbol='Egr2', Aliases='Egr-2, Krox-20, Krox20, NGF1-B, Zfp-25, Zfp-6', description='early growth response 2', other_designations='early growth response protein 2|zinc finger protein Krox-20', map_location='10 B5', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=67534456, end_position_on_the_genomic_accession=67542188, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13655, CurrentID=0, Status='live', Symbol='Egr3', Aliases='Pilot', description='early growth response 3', other_designations='EGR-3', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70077445, end_position_on_the_genomic_accession=70082613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13656, CurrentID=0, Status='live', Symbol='Egr4', Aliases='NGF1-C, NGFI-C, NGFIC, pAT133', description='early growth response 4', other_designations='EGR-4', map_location='6 37.48 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85511122, end_position_on_the_genomic_accession=85513542, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  13660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13660, CurrentID=0, Status='live', Symbol='Ehd1', Aliases='AA409636, Past1, RME-1', description='EH-domain containing 1', other_designations='PAST homolog 1|mPAST1', map_location='19 4.4 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6276896, end_position_on_the_genomic_accession=6300096, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13661, CurrentID=0, Status='live', Symbol='Ehf', Aliases='9030625L19Rik, AU019492', description='ets homologous factor', other_designations='ETS domain-containing transcription factor|ehf', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103263431, end_position_on_the_genomic_accession=103303275, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  13663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13663, CurrentID=0, Status='live', Symbol='Ei24', Aliases='AA536736, AI115355, PIG8', description='etoposide induced 2.4 mRNA', other_designations='p53-induced gene 8 protein', map_location='9 20.68 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=36779153, end_position_on_the_genomic_accession=36797393, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13664, CurrentID=0, Status='live', Symbol='Eif1a', Aliases='C76390, Ef1a, Eftu, Eif4c, eIF-1A, eIF-4C', description='eukaryotic translation initiation factor 1A', other_designations='eukaryotic translation initiation factor 4C', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46597658, end_position_on_the_genomic_accession=46610225, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13665, CurrentID=0, Status='live', Symbol='Eif2s1', Aliases='0910001O23Rik, 2410026C18Rik, 35kDa, Eif2a, eIF2alpha', description='eukaryotic translation initiation factor 2, subunit 1 alpha', other_designations='eIF-2-alpha|eIF-2A|eIF-2alpha|eukaryotic translation initiation factor 2 subunit alpha|eukaryotic translation initiation factor 2A', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=78862072, end_position_on_the_genomic_accession=78887010, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13666, CurrentID=0, Status='live', Symbol='Eif2ak3', Aliases='Pek, Perk', description='eukaryotic translation initiation factor 2 alpha kinase 3', other_designations='PRKR-like endoplasmic reticulum kinase|pancreatic eIF2-alpha kinase', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70844484, end_position_on_the_genomic_accession=70905240, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  13667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13667, CurrentID=0, Status='live', Symbol='Eif2b4', Aliases='Eif2b', description='eukaryotic translation initiation factor 2B, subunit 4 delta', other_designations='eIF-2B GDP-GTP exchange factor subunit delta', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31187558, end_position_on_the_genomic_accession=31193139, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13669, CurrentID=0, Status='live', Symbol='Eif3a', Aliases='A830012B05Rik, Csma, Eif3, Eif3s10, mKIAA0139', description='eukaryotic translation initiation factor 3, subunit A', other_designations='EUKARYOTIC TRANSLATION INITIATION FACTOR 3 SUBUNIT 10 (EIF-3 THETA) (EIF3 P167) (EIF3 P180) (EIF3 P185) (P162 PROTEIN) (CENTROSOMIN)|centrosomin|eIF-3-theta|eIF3 p167|eIF3 p180|eIF3 p185|eIF3a|eukaryotic translation initiation factor 3, subunit 10 (theta)|p162', map_location='19 D', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60761116, end_position_on_the_genomic_accession=60790693, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  13680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13680, CurrentID=0, Status='live', Symbol='Ddx19a', Aliases='DBP5, Ddx19, Eif4a-rs1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 19a', other_designations='DEAD box RNA helicase DEAD5|DEAD box protein 19A|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 19|Ddx19-like protein|eukaryotic translation initiation factor 4A related sequence 1|eukaryotic translation initiation factor 4A-related sequence 1|mDEAD5', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110974556, end_position_on_the_genomic_accession=110997823, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13681, CurrentID=0, Status='live', Symbol='Eif4a1', Aliases='BM-010, Ddx2a, Eif4', description='eukaryotic translation initiation factor 4A1', other_designations='ATP-dependent RNA helicase eIF4A-1|eIF-4A-I|eIF4A-I|initiation factor eIF-4A long form', map_location='11 42.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69666936, end_position_on_the_genomic_accession=69672423, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  13682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13682, CurrentID=0, Status='live', Symbol='Eif4a2', Aliases='4833432N07Rik, BM-010, Ddx2b, Eif4, eIF-4A-II, eIF4A-II', description='eukaryotic translation initiation factor 4A2', other_designations='ATP-dependent RNA helicase eIF4A-2', map_location='16 13.93 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23107468, end_position_on_the_genomic_accession=23114132, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  13684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13684, CurrentID=0, Status='live', Symbol='Eif4e', Aliases='EG668879-ps, If4e, eIF-4E, Eif4e', description='eukaryotic translation initiation factor 4E', other_designations='eIF-4F 25 kDa subunit|mRNA cap-binding protein', map_location='3 64.3 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138526191, end_position_on_the_genomic_accession=138557599, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13685, CurrentID=0, Status='live', Symbol='Eif4ebp1', Aliases='4e-bp1, AA959816, PHAS-I', description='eukaryotic translation initiation factor 4E binding protein 1', other_designations='eIF4E-binding protein 1|phosphorylated heat- and acid-stable protein regulated by insulin 1', map_location='8 A4-B1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27260327, end_position_on_the_genomic_accession=27275656, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13688, CurrentID=0, Status='live', Symbol='Eif4ebp2', Aliases='2810011I19Rik, 4E-BP2, AA792569, BC010348, PHAS-II', description='eukaryotic translation initiation factor 4E binding protein 2', other_designations='eIF4E-binding protein 2|phosphorylated heat- and acid-stable protein regulated by insulin 2', map_location='10 B4-B5', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61432497, end_position_on_the_genomic_accession=61452669, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  13690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13690, CurrentID=0, Status='live', Symbol='Eif4g2', Aliases='DAP-5, E130105L11Rik, Nat1, Natm1, p97', description='eukaryotic translation initiation factor 4, gamma 2', other_designations='eIF-4-gamma 2|eIF-4G 2|eIF4G 2|eif4g2|novel APOBEC-1 target 1|translation repressor Nat1', map_location='7 58.0 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=111067985, end_position_on_the_genomic_accession=111083030, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  13706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13706, CurrentID=0, Status='live', Symbol='Cela2a', Aliases='Ela-2, Ela2, Ela2a', description='chymotrypsin-like elastase family, member 2A', other_designations='elastase 2A|elastase-2|elastase-2A|neutrophilic elastase', map_location='4 74.75 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141814963, end_position_on_the_genomic_accession=141826003, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13708, CurrentID=0, Status='live', Symbol='elap', Aliases='lap', description='eye lens aplasia', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13709, CurrentID=0, Status='live', Symbol='Elf1', Aliases='Elf-1, Sts1, mElf-1, p70', description='E74-like factor 1', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79481194, end_position_on_the_genomic_accession=79582491, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  13710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13710, CurrentID=0, Status='live', Symbol='Elf3', Aliases='ESE-1, ESX, jen', description='E74-like factor 3', other_designations='elf3|epithelial-restricted with serine box|epithelium-restricted Ets protein ESX|epithelium-specific Ets transcription factor 1', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135253574, end_position_on_the_genomic_accession=135258565, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13711, CurrentID=0, Status='live', Symbol='Elf5', Aliases='ESE-2, ESE-5, ESE-5.', description='E74-like factor 5', other_designations='Epithelium specific Ets-2', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103411693, end_position_on_the_genomic_accession=103450989, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13712, CurrentID=0, Status='live', Symbol='Elk1', Aliases='Elk-1', description='ELK1, member of ETS oncogene family', other_designations='', map_location='X A1-A3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20933395, end_position_on_the_genomic_accession=20950608, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13713, CurrentID=0, Status='live', Symbol='Elk3', Aliases='D430049E23Rik, Erp, Etrp, Net, Sap-2', description='ELK3, member of ETS oncogene family', other_designations='ETS-domain protein ELK-3|ETS-related protein ERP|ETS-related protein NET', map_location='10 C-D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93247415, end_position_on_the_genomic_accession=93311167, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13714, CurrentID=0, Status='live', Symbol='Elk4', Aliases='2310011G17Rik, A130026I01Rik, BB162516, SAP-1, Sap1', description='ELK4, member of ETS oncogene family', other_designations='SRF accessory protein 1|serum response factor accessory protein 1', map_location='1 E3-G', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132007234, end_position_on_the_genomic_accession=132032612, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13716, CurrentID=0, Status='live', Symbol='Ell', Aliases='Ell1, Men', description='elongation factor RNA polymerase II', other_designations='RNA POLYMERASE II ELONGATION FACTOR ELL (ELEVEN-NINETEEN LYSINE-RICH LEUKEMIA PROTEIN)|eleven-nineteen lysine-rich leukemia protein', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70539675, end_position_on_the_genomic_accession=70592858, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  13717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13717, CurrentID=0, Status='live', Symbol='Eln', Aliases='AI385707, AI480567, E030024M20Rik', description='elastin', other_designations='tropoelastin', map_location='5 74.76 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134702596, end_position_on_the_genomic_accession=134747367, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  13722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13722, CurrentID=0, Status='live', Symbol='Aimp1', Aliases='9830137A06Rik, AIMP1/p43, EMAPII, Emap2, Scye1, p43', description='aminoacyl tRNA synthetase complex-interacting multifunctional protein 1', other_designations='ARS inducible multifunctional protein 1|endothelial monocyte activating polypeptide 2|multisynthase complex auxiliary component p43|multisynthetase complex auxiliary component p43|small inducible cytokine subfamily E, member 1', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=132660498, end_position_on_the_genomic_accession=132683879, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13723, CurrentID=0, Status='live', Symbol='Emb', Aliases='AL022799, Gp70', description='embigin', other_designations='glycoprotein 70|teratocarcinoma glycoprotein Gp-70', map_location='13 D2.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=117220573, end_position_on_the_genomic_accession=117274415, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13726, CurrentID=0, Status='live', Symbol='Emd', Aliases='AW550900, Sta', description='emerin', other_designations='', map_location='X 37.92 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74254673, end_position_on_the_genomic_accession=74261563, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  13728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13728, CurrentID=0, Status='live', Symbol='Mark2', Aliases='EMK-1, Emk, Par-1, Par-1b', description='MAP/microtubule affinity regulating kinase 2', other_designations='ELKL motif kinase 1|MAP/microtubule affinity-regulating kinase 2|PAR1 homolog b', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7275396, end_position_on_the_genomic_accession=7341867, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  13730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13730, CurrentID=0, Status='live', Symbol='Emp1', Aliases='I-8-09, TMP', description='epithelial membrane protein 1', other_designations='EMP-1|tumor-associated membrane protein', map_location='6 66.25 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135362550, end_position_on_the_genomic_accession=135383173, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13731, CurrentID=0, Status='live', Symbol='Emp2', Aliases='XMP', description='epithelial membrane protein 2', other_designations='EMP-2', map_location='16 5.54 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10281749, end_position_on_the_genomic_accession=10313968, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13732, CurrentID=0, Status='live', Symbol='Emp3', Aliases='H-4, H4, HNMP-1, MI-35, Ymp', description='epithelial membrane protein 3', other_designations='EMP-3|hematopoietic neural membrane protein 1', map_location='7 29.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45918023, end_position_on_the_genomic_accession=45921426, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13733, CurrentID=0, Status='live', Symbol='Adgre1', Aliases='DD7A5-7, EGF-TM7, Emr1, F4/80, Gpf480, Ly71, TM7LN3', description='adhesion G protein-coupled receptor E1', other_designations='EGF-like module containing, mucin-like, hormone receptor-like sequence 1|EGF-like module receptor 1|EGF-like module-containing mucin-like hormone receptor-like 1|EMR1 hormone receptor|cell surface glycoprotein F4/80|lymphocyte antigen 71', map_location='17 29.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57358651, end_position_on_the_genomic_accession=57483529, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  669149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=669149, CurrentID=0, Status='live', Symbol='Vmn2r88', Aliases='V2r13, V2r3', description='vomeronasal 2, receptor 88', other_designations='vomeronasal 2, receptor, 13|vomeronasal 2, receptor, 3|vomeronasal organ family 2, receptor, 13|vomeronasal organ family 2, receptor, 3|vomeronasal receptor Vmn2r88', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51411001, end_position_on_the_genomic_accession=51419527, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13796, CurrentID=0, Status='live', Symbol='Emx1', Aliases='', description='empty spiracles homeobox 1', other_designations='empty spiracles homolog 1|empty spiracles-like protein 1', map_location='6 37.21 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85187931, end_position_on_the_genomic_accession=85204463, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13797, CurrentID=0, Status='live', Symbol='Emx2', Aliases='Pdo', description='empty spiracles homeobox 2', other_designations='empty spiracles homolog 2|empty spiracles-like protein 2', map_location='19 56.28 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=59458116, end_position_on_the_genomic_accession=59465357, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13798, CurrentID=0, Status='live', Symbol='En1', Aliases='En-1, Mo-en.1', description='engrailed 1', other_designations='engrailed-1|homeobox protein en-1|mo-En-1', map_location='1 C2-qter', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120602391, end_position_on_the_genomic_accession=120607991, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13799, CurrentID=0, Status='live', Symbol='En2', Aliases='BB131122, En-2', description='engrailed 2', other_designations='homeobox protein en-2|mo-En-2', map_location='5 13.94 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=28165696, end_position_on_the_genomic_accession=28172166, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13800, CurrentID=0, Status='live', Symbol='Enah', Aliases='Mena, NDPP-1, Ndpp1, WBP8', description='enabled homolog (Drosophila)', other_designations='NPC derived proline rich protein 1', map_location='1 84.93 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181896384, end_position_on_the_genomic_accession=182019980, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  13801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13801, CurrentID=0, Status='live', Symbol='Enam', Aliases='abte', description='enamelin', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88487975, end_position_on_the_genomic_accession=88506049, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  13803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13803, CurrentID=0, Status='live', Symbol='Enc1', Aliases='Nrpb, PIG10', description='ectodermal-neural cortex 1', other_designations='ENC-1', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=97241105, end_position_on_the_genomic_accession=97253040, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  13804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13804, CurrentID=0, Status='live', Symbol='Endog', Aliases='', description='endonuclease G', other_designations='endo G', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30171401, end_position_on_the_genomic_accession=30174069, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  13805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13805, CurrentID=0, Status='live', Symbol='Eng', Aliases='AI528660, AI662476, CD105, Endo, S-endoglin', description='endoglin', other_designations='cell surface MJ7/18 antigen|transmembrane glycoprotein', map_location='2 22.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32646532, end_position_on_the_genomic_accession=32682669, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13806, CurrentID=0, Status='live', Symbol='Eno1', Aliases='0610008I15, AL022784, Eno-1, MBP-1', description='enolase 1, alpha non-neuron', other_designations='2-phospho-D-glycerate hydro-lyase|2-phospho-D-glycerate hydrolase|NNE|c-Myc promoter binding protein|non-neural enolase', map_location='4 81.24 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150236726, end_position_on_the_genomic_accession=150248873, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13807, CurrentID=0, Status='live', Symbol='Eno2', Aliases='AI837106, D6Ertd375e, Eno-2, NSE', description='enolase 2, gamma neuronal', other_designations='2-phospho-D-glycerate hydro-lyase|neural enolase|neuron-specific enolase', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124760053, end_position_on_the_genomic_accession=124769673, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13808, CurrentID=0, Status='live', Symbol='Eno3', Aliases='Eno-3, MSE', description='enolase 3, beta muscle', other_designations='2-phospho-D-glycerate hydro-lyase|muscle-specific enolase|skeletal muscle enolase', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70654763, end_position_on_the_genomic_accession=70662513, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  13809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13809, CurrentID=0, Status='live', Symbol='Enpep', Aliases='6030431M22Rik, APA, Bp-1/6C3, Ly-51, Ly51', description='glutamyl aminopeptidase', other_designations='AP-A|BP-1/6C3 antigen|EAP|aminopeptidase A|aminopeptidase-A', map_location='3 57.92 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129269177, end_position_on_the_genomic_accession=129332749, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  13812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13812, CurrentID=0, Status='live', Symbol='eob', Aliases='', description='eyelids open at birth', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13813, CurrentID=0, Status='live', Symbol='Eomes', Aliases='C77258, TBR-2, Tbr2', description='eomesodermin', other_designations='T-box brain protein 2|T-brain-2', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=118478180, end_position_on_the_genomic_accession=118486132, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  13819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13819, CurrentID=0, Status='live', Symbol='Epas1', Aliases='HIF-2alpha, HIF2A, HLF, HRF, MOP2, bHLHe73', description='endothelial PAS domain protein 1', other_designations='EPAS-1|HIF-1-alpha-like factor|HIF-2-alpha|HIF-related factor|HIF1 alpha-like factor|HIF1alpha-like factor|HIF2-alpha|HLF (HIF1alpha-like factor)|Hif like protein|hypoxia inducible transcription factor 2alpha|hypoxia-inducible factor 2-alpha|mHLF', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=86753864, end_position_on_the_genomic_accession=86833410, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  13821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13821, CurrentID=0, Status='live', Symbol='Epb41l1', Aliases='4.1N, Epb4.1l1, NBL1, mKIAA0338', description='erythrocyte membrane protein band 4.1 like 1', other_designations='erythrocyte protein band 4.1 like 1|erythrocyte protein band 4.1-like 1|neuronal protein 4.1', map_location='2 77.39 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156420487, end_position_on_the_genomic_accession=156543214, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  13822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13822, CurrentID=0, Status='live', Symbol='Epb41l2', Aliases='4.1G, AW555191, D10Ertd398e, Epb4.1l2, NBL2', description='erythrocyte membrane protein band 4.1 like 2', other_designations='erythrocyte membrane protein band 4.1-like 2 retinal|erythrocyte protein band 4.1 like 2|erythrocyte protein band 4.1-like 2|erythrocyte protein band 4.1l2 long|erythrocyte protein band 4.1l2 middle|generally expressed protein 4.1', map_location='10 12.26 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=25359798, end_position_on_the_genomic_accession=25523519, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  13823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13823, CurrentID=0, Status='live', Symbol='Epb41l3', Aliases='4.1B, DAL1P, Dal1, Epb4.1l3, NBL3', description='erythrocyte membrane protein band 4.1 like 3', other_designations='130kDa Protein 4.1B MEF cell isoform|60kDa 4.1B MEF cell isoform|DAL-1|differentially expressed in adenocarcinoma of the lung protein 1|erythrocyte protein band 4.1 like 3|erythrocyte protein band 4.1-like 3|mDAL-1', map_location='17 40.15 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=69075670, end_position_on_the_genomic_accession=69289991, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  13824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13824, CurrentID=0, Status='live', Symbol='Epb41l4a', Aliases='Epb4.1l4, Epb4.1l4a, NBL4', description='erythrocyte membrane protein band 4.1 like 4a', other_designations='erythrocyte protein band 4.1 like 4a|erythrocyte protein band 4.1-like 4|erythrocyte protein band 4.1-like 4a', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=33796327, end_position_on_the_genomic_accession=34007678, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  13828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13828, CurrentID=0, Status='live', Symbol='Epb42', Aliases='Epb4.2', description='erythrocyte membrane protein band 4.2', other_designations='P4.2|erythrocyte protein 4.2|erythrocyte protein band 4.2', map_location='2 60.37 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121018215, end_position_on_the_genomic_accession=121036755, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13829, CurrentID=0, Status='live', Symbol='Dmtn', Aliases='AI325486, Epb4.9, Epb49, dematin', description='dematin actin binding protein', other_designations='dematin actin-binding protein|erythrocyte membrane protein band 4.9|erythrocyte protein band 4.9', map_location='14 36.32 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70602184, end_position_on_the_genomic_accession=70636048, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  13830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13830, CurrentID=0, Status='live', Symbol='Stom', Aliases='Epb7.2', description='stomatin', other_designations='erythrocyte protein band 7.2|erythrocyte protein band 7.2; protein 7.2b|protein 7.2b', map_location='2 B-C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35313990, end_position_on_the_genomic_accession=35337009, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13831, CurrentID=0, Status='live', Symbol='Epc1', Aliases='2400007E14Rik, 5730566F07Rik, A930032N02Rik, mir-1893', description='enhancer of polycomb homolog 1 (Drosophila)', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=6435951, end_position_on_the_genomic_accession=6516108, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  13835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13835, CurrentID=0, Status='live', Symbol='Epha1', Aliases='5730453L17Rik, AL033318, Eph, Esk', description='Eph receptor A1', other_designations='embryonic stem cell kinase|mEpha1|tyrosine-protein kinase receptor ESK', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42358487, end_position_on_the_genomic_accession=42373310, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  13836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13836, CurrentID=0, Status='live', Symbol='Epha2', Aliases='AW545284, Eck, Myk2, Sek-2, Sek2', description='Eph receptor A2', other_designations='epithelial cell kinase|tyrosine-protein kinase receptor ECK|tyrosine-protein kinase receptor MPK-5|tyrosine-protein kinase receptor SEK-2', map_location='4 D-E', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141301221, end_position_on_the_genomic_accession=141329384, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  13837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13837, CurrentID=0, Status='live', Symbol='Epha3', Aliases='AW492086, Cek4, ETK1, End3, Hek, Hek4, Mek4, Tyro4', description='Eph receptor A3', other_designations='EK4|EPH-like kinase 4|tyrosine-protein kinase TYRO4|tyrosine-protein kinase receptor ETK1', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=63545218, end_position_on_the_genomic_accession=63864157, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  13838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13838, CurrentID=0, Status='live', Symbol='Epha4', Aliases='2900005C20Rik, AI385584, Cek8, Hek8, Sek, Sek1, Tyro1, rb', description='Eph receptor A4', other_designations='tyrosine-protein kinase receptor MPK-3|tyrosine-protein kinase receptor SEK-1', map_location='1 C1-C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=77367182, end_position_on_the_genomic_accession=77515103, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  13839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13839, CurrentID=0, Status='live', Symbol='Epha5', Aliases='AI854630, AW125296, Cek7, Ehk1, Els1, Hek7, Rek7, bsk', description='Eph receptor A5', other_designations='EPH homology kinase 1|brain-specific kinase|eck-like sequence 1', map_location='5 43.0 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=84054761, end_position_on_the_genomic_accession=84417818, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  13840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13840, CurrentID=0, Status='live', Symbol='Epha6', Aliases='Ehk2, Hek12, m-ehk2', description='Eph receptor A6', other_designations='EHK-2|EPH homology kinase 2', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59653061, end_position_on_the_genomic_accession=60606484, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  13841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13841, CurrentID=0, Status='live', Symbol='Epha7', Aliases='Cek11, Ebk, Ehk3, Hek11, Mdk1', description='Eph receptor A7', other_designations='EPH homology kinase 3|developmental kinase 1|embryonic brain kinase', map_location='4 12.42 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=28813044, end_position_on_the_genomic_accession=28967511, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  13842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13842, CurrentID=0, Status='live', Symbol='Epha8', Aliases='AW047546, Eek, Hek3, mKIAA1459', description='Eph receptor A8', other_designations='EPH- and ELK-related kinase|Ephrin type-A receptor 8 precursor (Tyrosine-protein kinase receptor EEK) (EPH-and ELK-related kinase)|tyrosine-protein kinase receptor EEK', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136929419, end_position_on_the_genomic_accession=136956816, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  13844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13844, CurrentID=0, Status='live', Symbol='Ephb2', Aliases='Cek5, Drt, ETECK, Erk, Hek5, Nuk, Prkm5, Qek5, Sek3, Tyro5', description='Eph receptor B2', other_designations='neural kinase|nuk receptor tyrosine kinase|tyrosine-protein kinase receptor EPH-3|tyrosine-protein kinase receptor SEK-3', map_location='4 D-E', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136647536, end_position_on_the_genomic_accession=136836295, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  13845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13845, CurrentID=0, Status='live', Symbol='Ephb3', Aliases='AW456895, Cek10, Etk2, HEK2, MDK5, Sek4, Tyro6', description='Eph receptor B3', other_designations='developmental kinase 5|mDK-5|tyrosine-protein kinase receptor SEK-4', map_location='16 B1-B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21204795, end_position_on_the_genomic_accession=21223309, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  13846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13846, CurrentID=0, Status='live', Symbol='Ephb4', Aliases='AI042935, Htk, MDK2, Myk1, Tyro11', description='Eph receptor B4', other_designations='developmental kinase 2|hepatoma transmembrane kinase|mDK-2|tyrosine kinase MYK-1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137350109, end_position_on_the_genomic_accession=137374522, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  13848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13848, CurrentID=0, Status='live', Symbol='Ephb6', Aliases='Cekl, Mep', description='Eph receptor B6', other_designations='chicken eph/elk receptor-like protein|ephrin type A receptor|tyrosine-protein kinase-defective receptor EPH-6', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41605040, end_position_on_the_genomic_accession=41620507, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  13849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13849, CurrentID=0, Status='live', Symbol='Ephx1', Aliases='EH, Eph-1, Eph1, Epxh1, mEH', description='epoxide hydrolase 1, microsomal', other_designations='epoxide hydratase|microsomal epoxide hydrolase', map_location='1 84.48 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180989556, end_position_on_the_genomic_accession=181017495, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  13850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13850, CurrentID=0, Status='live', Symbol='Ephx2', Aliases='CEH, Eph2, SEH, sEP', description='epoxide hydrolase 2, cytoplasmic', other_designations='cytosolic epoxide hydrolase|epoxide hydratase|epoxide hydrolase 2C|soluble epoxide hydrolase', map_location='14 D', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66084372, end_position_on_the_genomic_accession=66124522, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  13851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13851, CurrentID=0, Status='live', Symbol='epi', Aliases='', description='exocrine pancreatic insufficiency', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  13852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13852, CurrentID=0, Status='live', Symbol='Stx2', Aliases='AW538950, Epim, G1-536-1, Syn-2, repro34', description='syntaxin 2', other_designations='epimorphin', map_location='5 67.99 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=128984557, end_position_on_the_genomic_accession=129008600, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  13853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13853, CurrentID=0, Status='live', Symbol='Epm2a', Aliases='TG-B, Tg(TcraK,TcrbK)TG-BFlv', description='epilepsy, progressive myoclonic epilepsy, type 2 gene alpha', other_designations='LAFPTPase|glucan phosphatase|lafora PTPase', map_location='10 A', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=11342690, end_position_on_the_genomic_accession=11457477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13854, CurrentID=0, Status='live', Symbol='Epn1', Aliases='AA269831, Ibp1', description='epsin 1', other_designations='EPS-15-interacting protein 1|intersectin-EH-binding protein 1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5080203, end_position_on_the_genomic_accession=5098179, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13855, CurrentID=0, Status='live', Symbol='Epn2', Aliases='9530051D10Rik, AA536924, Ibp2', description='epsin 2', other_designations='EPS-15-interacting protein 2|intersectin-EH-binding protein 2', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61517249, end_position_on_the_genomic_accession=61579881, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  13856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13856, CurrentID=0, Status='live', Symbol='Epo', Aliases='', description='erythropoietin', other_designations='', map_location='5 G', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137483020, end_position_on_the_genomic_accession=137485816, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13857, CurrentID=0, Status='live', Symbol='Epor', Aliases='', description='erythropoietin receptor', other_designations='EPO-R', map_location='9 7.93 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21958897, end_position_on_the_genomic_accession=21963576, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13858, CurrentID=0, Status='live', Symbol='Eps15', Aliases='2410112D09Rik', description='epidermal growth factor receptor pathway substrate 15', other_designations='epidermal growth factor pathway substrate 15|protein AF-1p', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109280256, end_position_on_the_genomic_accession=109387817, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  13859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13859, CurrentID=0, Status='live', Symbol='Eps15l1', Aliases='9830147J04Rik, AI593686, Eps15-rs, Eps15R', description='epidermal growth factor receptor pathway substrate 15-like 1', other_designations='epidermal growth factor receptor pathway substrate 15-related sequence|eps15-related protein', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72340990, end_position_on_the_genomic_accession=72421565, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  13860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13860, CurrentID=0, Status='live', Symbol='Eps8', Aliases='AW261790', description='epidermal growth factor receptor pathway substrate 8', other_designations='', map_location='6 66.78 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=137477245, end_position_on_the_genomic_accession=137649397, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  13861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13861, CurrentID=0, Status='live', Symbol='Epx', Aliases='EPO', description='eosinophil peroxidase', other_designations='', map_location='11 52.25 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87863997, end_position_on_the_genomic_accession=87875536, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13863, CurrentID=0, Status='live', Symbol='Lcn5', Aliases='E-RABP, Erabp, MEP10', description='lipocalin 5', other_designations='epididymal retinoic acid binding protein|epididymal secretory protein 10|lcn5', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25657932, end_position_on_the_genomic_accession=25662230, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  13864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13864, CurrentID=0, Status='live', Symbol='Nr2f6', Aliases='AV090102, COUP-TF3, EAR2, Erbal2', description='nuclear receptor subfamily 2, group F, member 6', other_designations='COUP transcription factor 3|EAR-2|V-erbA-related protein 2|avian erythroblastic leukemia viral (v-erb-a) oncogene homolog-like 2', map_location='8 34.43 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71374118, end_position_on_the_genomic_accession=71381955, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  13865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13865, CurrentID=0, Status='live', Symbol='Nr2f1', Aliases='COUP-TF1, COUP-TFI, COUPTFA, EAR3, Erbal3, SVP44, Tcfcoup1', description='nuclear receptor subfamily 2, group F, member 1', other_designations='COUP transcription factor I|COUP-TF I|EAR-3|V-erbA-related protein 3|avian erythroblastic leukemia viral (v-erb-a) oncogene homolog-like 3', map_location='13 41.38 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=78188971, end_position_on_the_genomic_accession=78199172, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  13866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13866, CurrentID=0, Status='live', Symbol='Erbb2', Aliases='Erbb-2, HER-2, HER2, Neu, c-erbB2, c-neu, l11Jus8, mKIAA3023', description='erb-b2 receptor tyrosine kinase 2', other_designations='Neu oncogene|avian erythroblastosis oncogene B 2|c-erbB-2|p185erbB2|proto-oncogene NEU|proto-oncogene c-ErbB-2|v-erb-b2 erythroblastic leukemia viral oncogene homolog 2, neuro/glioblastoma derived oncogene homolog', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98412484, end_position_on_the_genomic_accession=98437716, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  13867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13867, CurrentID=0, Status='live', Symbol='Erbb3', Aliases='C76256, Erbb-3r, Her3, Erbb3', description='erb-b2 receptor tyrosine kinase 3', other_designations='avian erythroblastosis oncogene B 3 receptor|c-erbB-3|glial growth factor receptor|proto-oncogene-like protein c-ErbB-3|receptor tyrosine kinase|v-erb-b2 erythroblastic leukemia viral oncogene homolog 3', map_location='10 77.1 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128569368, end_position_on_the_genomic_accession=128589501, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  13869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13869, CurrentID=0, Status='live', Symbol='Erbb4', Aliases='Her4, c-erbB-4', description='erb-b2 receptor tyrosine kinase 4', other_designations='avian erythroblastosis oncogene B 4|proto-oncogene-like protein c-ErbB-4|v-erb-a erythroblastic leukemia viral oncogene homolog 4', map_location='1 33.8 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=68031904, end_position_on_the_genomic_accession=69108794, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  13870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13870, CurrentID=0, Status='live', Symbol='Ercc1', Aliases='Ercc-1', description='excision repair cross-complementing rodent repair deficiency, complementation group 1', other_designations='', map_location='7 9.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19344422, end_position_on_the_genomic_accession=19356524, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  13871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13871, CurrentID=0, Status='live', Symbol='Ercc2', Aliases='AA407812, AU020867, AW240756, CXPD, Ercc-2, Mhdarco15, XPD', description='excision repair cross-complementing rodent repair deficiency, complementation group 2', other_designations='DNA excision repair protein ERCC-2|DNA repair protein complementing XP-D cells|TFIIH basal transcription factor complex helicase subunit|xeroderma pigmentosum group D-complementing protein', map_location='7 9.62 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19382039, end_position_on_the_genomic_accession=19395692, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  13872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13872, CurrentID=0, Status='live', Symbol='Ercc3', Aliases='BTF2 p89, Ercc-3, XPB', description='excision repair cross-complementing rodent repair deficiency, complementation group 3', other_designations='BTF2-p89|DNA excision repair protein ERCC-3|DNA repair protein complementing XP-B cells|TFIIH 89 kDa subunit|basic transcription factor 2 89 kDa subunit|excision repair 3|xeroderma pigmentosum group B-complementing protein', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32240305, end_position_on_the_genomic_accession=32270151, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  13874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13874, CurrentID=0, Status='live', Symbol='Ereg', Aliases='EPR', description='epiregulin', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=91074617, end_position_on_the_genomic_accession=91093649, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  13875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13875, CurrentID=0, Status='live', Symbol='Erf', Aliases='', description='Ets2 repressor factor', other_designations='Est2 repressor factor', map_location='7 13.73 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25242560, end_position_on_the_genomic_accession=25250758, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13876, CurrentID=0, Status='live', Symbol='Erg', Aliases='D030036I24Rik', description='avian erythroblastosis virus E-26 (v-ets) oncogene related', other_designations='', map_location='16 56.04 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=95359169, end_position_on_the_genomic_accession=95586593, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  13877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13877, CurrentID=0, Status='live', Symbol='Erh', Aliases='Mer, Prei1', description='enhancer of rudimentary homolog (Drosophila)', other_designations='preimplantation protein 1', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80634023, end_position_on_the_genomic_accession=80643861, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13884, CurrentID=0, Status='live', Symbol='Ces1c', Aliases='Ces-N, Ee-1, Ee1, Es-4, Es-N, Es1, Es4, EsN, PESN', description='carboxylesterase 1C', other_designations='PES-N|esterase 1|liver carboxylesterase N|lung surfactant convertase', map_location='8 44.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93099015, end_position_on_the_genomic_accession=93131283, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13885, CurrentID=0, Status='live', Symbol='Esd', Aliases='Es-10, Es10, FGH, sid478', description='esterase D/formylglutathione hydrolase', other_designations='Sid 478|esterase 10|esterase D', map_location='14 D2-E2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=74732297, end_position_on_the_genomic_accession=74750765, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13897, CurrentID=0, Status='live', Symbol='Ces1e', Aliases='Eg, Es-22, Es22, egasyn', description='carboxylesterase 1E', other_designations='esterase 22|esterase-22|liver carboxylesterase 22', map_location='8 45.11 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93201218, end_position_on_the_genomic_accession=93229632, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13909, CurrentID=0, Status='live', Symbol='Ces3b', Aliases='Ces31L, Ces3L, EG13909, ES31L, Es31, Gm4738', description='carboxylesterase 3B', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105083730, end_position_on_the_genomic_accession=105093929, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  13924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13924, CurrentID=0, Status='live', Symbol='Ptprv', Aliases='Esp, OST, OST-PTP, mOST-PTP', description='protein tyrosine phosphatase, receptor type, V', other_designations='ES cell phosphatase|R-PTP-V|embryonic stem cell phosphatase|embryonic stem cell protein-tyrosine phosphatase|tyrosine phosphatase', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135108498, end_position_on_the_genomic_accession=135132575, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  13929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13929, CurrentID=0, Status='live', Symbol='Amz2', Aliases='AA408420, ESTM12, X83328', description='archaelysin family metallopeptidase 2', other_designations='archeobacterial metalloproteinase-like protein 2', map_location='11 71.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109425946, end_position_on_the_genomic_accession=109438148, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  13972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13972, CurrentID=0, Status='live', Symbol='Gnb1l', Aliases='ESTM55, Gm16314, Me49f07, OTTMUSG00000033458, Wdr14, Wdvcf', description='guanine nucleotide binding protein (G protein), beta polypeptide 1-like', other_designations='WD repeat-containing protein 14|WD40 repeat-containing protein deleted in VCFS|g protein subunit beta-like protein 1', map_location='16 11.46 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18498021, end_position_on_the_genomic_accession=18566680, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  330010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330010, CurrentID=0, Status='live', Symbol='Ttll10', Aliases='4833412E22Rik, 4930595O22Rik, AU022799, BC042775, Ttll5', description='tubulin tyrosine ligase-like family, member 10', other_designations='tubulin tyrosine ligase-like family, member 5|tubulin--tyrosine ligase-like protein 10', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156023748, end_position_on_the_genomic_accession=156084440, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  13982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13982, CurrentID=0, Status='live', Symbol='Esr1', Aliases='ER, ER-alpha, ERa, ERalpha, ESR, Estr, Estra, Nr3a1', description='estrogen receptor 1 (alpha)', other_designations='estradiol receptor|nuclear receptor subfamily 3 group A member 1', map_location='10 2.03 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=4611989, end_position_on_the_genomic_accession=5005633, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  13983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13983, CurrentID=0, Status='live', Symbol='Esr2', Aliases='ER[b], ERbeta, Estrb', description='estrogen receptor 2 (beta)', other_designations='ER beta|ER-beta|nuclear receptor subfamily 3 group A member 2|oestrogen receptor beta', map_location='12 D1-D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76120419, end_position_on_the_genomic_accession=76177259, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  13984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13984, CurrentID=0, Status='live', Symbol='Esx1', Aliases='Spx1', description='extraembryonic, spermatogenesis, homeobox 1', other_designations='extraembryonic spermatogenesis homeobox 1', map_location='X 59.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=137115397, end_position_on_the_genomic_accession=137122131, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  13990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13990, CurrentID=0, Status='live', Symbol='Smarcad1', Aliases='AV081750, AW226546, D6Pas1, Etl1, mKIAA1122', description='SWI/SNF-related, matrix-associated actin-dependent regulator of chromatin, subfamily a, containing DEAD/H box 1', other_designations='ATP-dependent helicase SMARCAD1|enhancer trap locus 1|enhancer trap locus homolog 1|etl-1', map_location='6 30.11 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=65042588, end_position_on_the_genomic_accession=65116061, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  13992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13992, CurrentID=0, Status='live', Symbol='Khdrbs3', Aliases='Etle, SLM-2, Salp, Slm2, T-STAR', description='KH domain containing, RNA binding, signal transduction associated 3', other_designations='RNA-binding protein Etoile|etoile|sam68-like mammalian protein 2', map_location='15 30.36 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=68928405, end_position_on_the_genomic_accession=69101163, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  13998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13998, CurrentID=0, Status='live', Symbol='Fgd6', Aliases='AA123052, Etohd4, ZFYVE24', description='FYVE, RhoGEF and PH domain containing 6', other_designations='ethanol decreased 4', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=94036001, end_position_on_the_genomic_accession=94145339, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  13999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13999, CurrentID=0, Status='live', Symbol='Gm14288', Aliases='OTTMUSG00000016199', description='predicted gene 14288', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=175792312, end_position_on_the_genomic_accession=175801038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14000, CurrentID=0, Status='live', Symbol='Drosha', Aliases='1110013A17Rik, AI874853, Etohi2, Rn3, Rnasen', description='drosha, ribonuclease type III', other_designations='RNase III|ethanol induced 2|protein Drosha|ribonuclease III, nuclear', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=12824804, end_position_on_the_genomic_accession=12935291, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  14004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14004, CurrentID=0, Status='live', Symbol='Chchd2', Aliases='AL033347, Etohi6', description='coiled-coil-helix-coiled-coil-helix domain containing 2', other_designations='coiled-coil-helix-coiled-coil-helix domain-containing protein 2, mitochondrial|ethanol induced 6', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129881161, end_position_on_the_genomic_accession=129887470, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14007, CurrentID=0, Status='live', Symbol='Celf2', Aliases='B230218O03, B230345P09Rik, C88023, CELF-2, CUG-BP2, Cugbp2, D230046B21Rik, Etr-3, Napor, Napor-2, mETR-3', description='CUGBP, Elav-like family member 2', other_designations='CUG triplet repeat RNA-binding protein 2|CUG triplet repeat, RNA binding protein 2|CUG-BP- and ETR-3-like factor 2|RNA-binding protein BRUNOL-3|bruno-like protein 3|elav-type RNA-binding protein 3|mNapor|neuroblastoma apoptosis-related RNA-binding protein', map_location='2 A2-A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=6539694, end_position_on_the_genomic_accession=7396227, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  14008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14008, CurrentID=0, Status='live', Symbol='Etv2', Aliases='Etsrp71', description='ets variant 2', other_designations='ets related protein 71|ets variant gene 2|ets-related protein 71', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30633616, end_position_on_the_genomic_accession=30636509, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14009, CurrentID=0, Status='live', Symbol='Etv1', Aliases='ER81, Etsrp81', description='ets variant 1', other_designations='ets related protein 81|ets variant gene 1|ets-related protein 81', map_location='12 17.99 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=38779656, end_position_on_the_genomic_accession=38868215, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14011, CurrentID=0, Status='live', Symbol='Etv6', Aliases='AW123102, AW557856, Tel', description='ets variant 6', other_designations='ETS translocation variant 6|ETS-related protein Tel1|ets variant gene 6 (TEL oncogene)|translocation-ets-leukemia', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134034583, end_position_on_the_genomic_accession=134270158, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14012, CurrentID=0, Status='live', Symbol='Mpzl2', Aliases='Eva, Eva1', description='myelin protein zero-like 2', other_designations='epithelial V-like antigen 1', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45040291, end_position_on_the_genomic_accession=45054043, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14013, CurrentID=0, Status='live', Symbol='Mecom', Aliases='D630039M04Rik, Evi-1, Evi1, Jbo, Mds, Mds1, Mds1-Evi1, Prdm3, Znfpr1b1', description='MDS1 and EVI1 complex locus', other_designations='MDS1 and EVI1 complex locus protein MDS1|ecotropic viral integration site 1|ecotropic virus integration site 1 protein|myelodysplasia syndrome 1 homolog|myelodysplasia syndrome 1 protein homolog', map_location='3 12.66 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=29951296, end_position_on_the_genomic_accession=30509783, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  14017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14017, CurrentID=0, Status='live', Symbol='Evi2a', Aliases='AW491894, Evi-2, Evi2', description='ecotropic viral integration site 2a', other_designations='EVI-2A|ecotropic viral integration site 2A protein', map_location='11 46.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79526560, end_position_on_the_genomic_accession=79530609, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14020, CurrentID=0, Status='live', Symbol='Evi5', Aliases='NB4S', description='ecotropic viral integration site 5', other_designations='EVI-5', map_location='5 52.23 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107744795, end_position_on_the_genomic_accession=107875107, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  14025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14025, CurrentID=0, Status='live', Symbol='Bcl11a', Aliases='2810047E18Rik, BCL-11A, Ctip1, D930021L15Rik, Evi9, Evi9a, Evi9b, Evi9c, mKIAA1809', description='B cell CLL/lymphoma 11A (zinc finger protein)', other_designations='B-cell CLL/lymphoma 11A (zinc finger protein)|COUP-TF interacting protein 1|COUP-TF-interacting protein 1|EVI-9|ecotropic viral integration site 9 protein|myeloid leukemia', map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=24076564, end_position_on_the_genomic_accession=24173558, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14026, CurrentID=0, Status='live', Symbol='Evl', Aliases='AI528774', description='Ena-vasodilator stimulated phosphoprotein', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108554720, end_position_on_the_genomic_accession=108688516, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  407243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407243, CurrentID=0, Status='live', Symbol='Tmem189', Aliases='AI840826, Kua', description='transmembrane protein 189', other_designations='tmem189|ubiquitin-conjugating enzyme variant Kua', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167643225, end_position_on_the_genomic_accession=167661544, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14028, CurrentID=0, Status='live', Symbol='Evx1', Aliases='Evx-1', description='even-skipped homeobox 1', other_designations='even skipped homeotic gene 1 homolog', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52313498, end_position_on_the_genomic_accession=52318389, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14029, CurrentID=0, Status='live', Symbol='Evx2', Aliases='Evx-2', description='even-skipped homeobox 2', other_designations='even skipped homeotic gene 2 homolog', map_location='2 44.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74655524, end_position_on_the_genomic_accession=74661518, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14030, CurrentID=0, Status='live', Symbol='Ewsr1', Aliases='Ews, Ewsh', description='Ewing sarcoma breakpoint region 1', other_designations='Ewing sarcoma homolog', map_location='11 A1-A3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5069687, end_position_on_the_genomic_accession=5099099, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  14033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14033, CurrentID=0, Status='live', Symbol='exed', Aliases='', description='extraembryonic ectoderm development', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14034, CurrentID=0, Status='live', Symbol='exf', Aliases='ex', description='exfoliative', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14038, CurrentID=0, Status='live', Symbol='Wfdc18', Aliases='Expi, Kal1, WDNM1', description='WAP four-disulfide core domain 18', other_designations='Kallmann syndrome 1 sequence|extracellular peptidase inhibitor|extracellular proteinase inhibitor', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83709004, end_position_on_the_genomic_accession=83711360, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14042, CurrentID=0, Status='live', Symbol='Ext1', Aliases='AA409028', description='exostoses (multiple) 1', other_designations='glucuronosyl-N-acetylglucosaminyl-proteoglycan/N-acetylglucosaminyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|multiple exostoses protein 1 homolog', map_location='15 20.0 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=53068261, end_position_on_the_genomic_accession=53346183, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14043, CurrentID=0, Status='live', Symbol='Ext2', Aliases='AI893565', description='exostoses (multiple) 2', other_designations='glucuronosyl-N-acetylglucosaminyl-proteoglycan/N-acetylglucosaminyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|multiple exostoses protein 2 homolog', map_location='2 51.62 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93695631, end_position_on_the_genomic_accession=93822588, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  14048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14048, CurrentID=0, Status='live', Symbol='Eya1', Aliases='bor', description='EYA transcriptional coactivator and phosphatase 1', other_designations='eyes absent 1 homolog', map_location='1 4.31 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=14168953, end_position_on_the_genomic_accession=14310265, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14049, CurrentID=0, Status='live', Symbol='Eya2', Aliases='Eab1', description='EYA transcriptional coactivator and phosphatase 2', other_designations='eyes absent 2 homolog', map_location='2 85.83 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165595028, end_position_on_the_genomic_accession=165771727, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14050, CurrentID=0, Status='live', Symbol='Eya3', Aliases='AI844637', description='EYA transcriptional coactivator and phosphatase 3', other_designations='eyes absent 3 homolog', map_location='4 65.68 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132639016, end_position_on_the_genomic_accession=132724765, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  14051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14051, CurrentID=0, Status='live', Symbol='Eya4', Aliases='B130023L16Rik', description='EYA transcriptional coactivator and phosphatase 4', other_designations='eyes absent 4 homolog', map_location='10 10.44 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23102962, end_position_on_the_genomic_accession=23350856, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  14055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14055, CurrentID=0, Status='live', Symbol='Ezh1', Aliases='ENX-2', description='enhancer of zeste 1 polycomb repressive complex 2 subunit', other_designations='enhancer of zeste homolog 1', map_location='11 C-D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101191115, end_position_on_the_genomic_accession=101226463, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  14056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14056, CurrentID=0, Status='live', Symbol='Ezh2', Aliases='Enx-1, Enx1h, KMT6, mKIAA4065', description='enhancer of zeste 2 polycomb repressive complex 2 subunit', other_designations='enhancer of zeste homolog 2', map_location='6 B', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47530274, end_position_on_the_genomic_accession=47595340, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  14057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14057, CurrentID=0, Status='live', Symbol='Sfxn1', Aliases='2810002O05Rik, A930015P12Rik, f', description='sideroflexin 1', other_designations='flexed tail|flexed-tail', map_location='13 28.4 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54071824, end_position_on_the_genomic_accession=54108346, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14058, CurrentID=0, Status='live', Symbol='F10', Aliases='Cf10, fX', description='coagulation factor X', other_designations='stuart factor', map_location='8 5.73 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13037308, end_position_on_the_genomic_accession=13056676, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14060, CurrentID=0, Status='live', Symbol='F13b', Aliases='Cf-13b, Cf13b', description='coagulation factor XIII, beta subunit', other_designations='protein-glutamine gamma-glutamyltransferase B chain|transglutaminase B chain', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=139501551, end_position_on_the_genomic_accession=139523756, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14061, CurrentID=0, Status='live', Symbol='F2', Aliases='Cf-2, Cf2, FII', description='coagulation factor II', other_designations='thrombin', map_location='2 50.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91613986, end_position_on_the_genomic_accession=91636452, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14062, CurrentID=0, Status='live', Symbol='F2r', Aliases='AI482343, Cf2r, Par1, ThrR', description='coagulation factor II (thrombin) receptor', other_designations='PAR-1|thrombin receptor', map_location='13 50.21 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95601789, end_position_on_the_genomic_accession=95618433, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14063, CurrentID=0, Status='live', Symbol='F2rl1', Aliases='Gpcr11, PAR-2, Par2', description='coagulation factor II (thrombin) receptor-like 1', other_designations='G-protein coupled receptor 11|Protease-activated receptor-2|coagulation factor II receptor-like 1|proteinase-activated receptor-2|thrombin receptor-like 1', map_location='13 49.53 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95511718, end_position_on_the_genomic_accession=95525240, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14064, CurrentID=0, Status='live', Symbol='F2rl2', Aliases='F730031A08, PAR3', description='coagulation factor II (thrombin) receptor-like 2', other_designations='PAR-3|Proteinase activated receptor 3 precursor (PAR-3) (Thrombin receptor-like 2) (Coagulation factor II receptor-like 2)|coagulation factor II receptor-like 2|protease-activated receptor 3|thrombin receptor-like 2', map_location='13 50.38 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95696920, end_position_on_the_genomic_accession=95702768, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14065, CurrentID=0, Status='live', Symbol='F2rl3', Aliases='PAR4', description='coagulation factor II (thrombin) receptor-like 3', other_designations='PAR-4|coagulation factor II receptor-like 3|thrombin receptor-like 3', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72761186, end_position_on_the_genomic_accession=72763898, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14066, CurrentID=0, Status='live', Symbol='F3', Aliases='AA409063, CD142, Cf-3, Cf3, TF', description='coagulation factor III', other_designations='', map_location='3 52.94 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121723537, end_position_on_the_genomic_accession=121735052, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14067, CurrentID=0, Status='live', Symbol='F5', Aliases='AI173222, Cf-5, Cf5', description='coagulation factor V', other_designations='activated protein C cofactor', map_location='1 71.46 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164151859, end_position_on_the_genomic_accession=164220274, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  14068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14068, CurrentID=0, Status='live', Symbol='F7', Aliases='AI132620, Cf7, FVII', description='coagulation factor VII', other_designations='serum prothrombin conversion accelerator', map_location='8 5.73 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13026034, end_position_on_the_genomic_accession=13035809, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14069, CurrentID=0, Status='live', Symbol='F8', Aliases='Cf-8, Cf8, FVIII', description='coagulation factor VIII', other_designations='Factor VIII|procoagulant component', map_location='X A7-B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75172715, end_position_on_the_genomic_accession=75383525, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  14070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14070, CurrentID=0, Status='live', Symbol='F8a', Aliases='AI852759, DXUcsf11, F8a', description='factor 8-associated gene A', other_designations='Factor VIII associated protein|cpG island protein', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73228306, end_position_on_the_genomic_accession=73230795, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14071, CurrentID=0, Status='live', Symbol='F9', Aliases='AW111646, Cf-9, Cf9', description='coagulation factor IX', other_designations='christmas factor', map_location='X A6-A7', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=59999464, end_position_on_the_genomic_accession=60030758, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14073, CurrentID=0, Status='live', Symbol='Faah', Aliases='AW412498', description='fatty acid amide hydrolase', other_designations='amide hydrolase|anandamide amidohydrolase 1|fatty acid amide hydrolase protein|oleamide hydrolase 1', map_location='4 53.08 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115996656, end_position_on_the_genomic_accession=116017902, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14077, CurrentID=0, Status='live', Symbol='Fabp3', Aliases='Fabph-1, Fabph-4, Fabph1, Fabph4, H-FABP, Mdgi', description='fatty acid binding protein 3, muscle and heart', other_designations='fatty acid binding protein heart 1|fatty acid binding protein heart 4|fatty acid-binding protein 3|heart-type fatty acid-binding protein|mammary-derived growth inhibitor', map_location='4 63.43 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130308778, end_position_on_the_genomic_accession=130315463, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14079, CurrentID=0, Status='live', Symbol='Fabp2', Aliases='Fabpi, I-FABP', description='fatty acid binding protein 2, intestinal', other_designations='intestinal fatty acid binding protein 2|intestinal-type fatty acid-binding protein', map_location='3 53.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122895072, end_position_on_the_genomic_accession=122899506, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14080, CurrentID=0, Status='live', Symbol='Fabp1', Aliases='Fabpl, L-FABP', description='fatty acid binding protein 1, liver', other_designations='14 kDa selenium-binding protein|fatty acid binding protein liver|fatty acid-binding protein 1|liver-type fatty acid-binding protein', map_location='6 32.14 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71199888, end_position_on_the_genomic_accession=71205023, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14081, CurrentID=0, Status='live', Symbol='Acsl1', Aliases='Acas, Acas1, Acs, FACS, Facl2, LACS 1, LACS1', description='acyl-CoA synthetase long-chain family member 1', other_designations='acetate-CoA ligase|acetyl-CoA synthetase|acetyl-Coenzyme A synthetase 1 (ADP forming)|fatty acid Coenzyme A ligase, long chain 2|long-chain acyl-CoA synthetase 1', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46471037, end_position_on_the_genomic_accession=46536051, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  14082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14082, CurrentID=0, Status='live', Symbol='Fadd', Aliases='Mort1/FADD', description='Fas (TNFRSF6)-associated via death domain', other_designations='FAS-associating death domain-containing protein|Fas-associating protein with death domain|mediator of receptor induced toxicity|protein FADD', map_location='7 88.85 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144578323, end_position_on_the_genomic_accession=144582436, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14083, CurrentID=0, Status='live', Symbol='Ptk2', Aliases='FAK, FRNK, Fadk, mKIAA4203, p125FAK', description='PTK2 protein tyrosine kinase 2', other_designations='FADK 1|focal adhesion kinase-related nonkinase|focal ashension kinase 1|pp125FAK|protein-tyrosine kinase 2', map_location='15 33.94 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73205102, end_position_on_the_genomic_accession=73423275, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  14084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14084, CurrentID=0, Status='live', Symbol='Faf1', Aliases='AA408698, Dffrx, Fam', description='Fas-associated factor 1', other_designations='', map_location='4 51.33 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109676627, end_position_on_the_genomic_accession=109963960, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14085, CurrentID=0, Status='live', Symbol='Fah', Aliases='', description='fumarylacetoacetate hydrolase', other_designations='FAA|beta-diketonase', map_location='7 48.36 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=84585159, end_position_on_the_genomic_accession=84605942, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14086, CurrentID=0, Status='live', Symbol='Fscn1', Aliases='AI663989, Fan1, fascin-1', description='fascin actin-bundling protein 1', other_designations='fascin 1|singed-like protein', map_location='5 81.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142960355, end_position_on_the_genomic_accession=142973189, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14087, CurrentID=0, Status='live', Symbol='Fanca', Aliases='AW208693, FACA', description='Fanconi anemia, complementation group A', other_designations='', map_location='8 72.1 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123268244, end_position_on_the_genomic_accession=123318651, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  14088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14088, CurrentID=0, Status='live', Symbol='Fancc', Aliases='Facc, mir-3074-1', description='Fanconi anemia, complementation group C', other_designations='', map_location='13 32.8 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=63301134, end_position_on_the_genomic_accession=63497290, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  14089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14089, CurrentID=0, Status='live', Symbol='Fap', Aliases='SIMP', description='fibroblast activation protein', other_designations='FAPalpha|dipeptidyl peptidase FAP|fibroblast activation protein alpha|gelatine degradation protease FAP|integral membrane serine protease|post-proline cleaving enzyme|seprase|serine integral membrane protease|surface-expressed protease', map_location='2 35.85 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=62500936, end_position_on_the_genomic_accession=62575508, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  14102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14102, CurrentID=0, Status='live', Symbol='Fas', Aliases='AI196731, APO1, APT1, CD95, TNFR6, Tnfrsf6, lpr', description='Fas (TNF receptor superfamily member 6)', other_designations='FASLG receptor|Fas (TNF receptor superfamily member)|Fas antigen|apo-1 antigen|apoptosis-mediating surface antigen FAS|lymphoproliferation', map_location='19 29.48 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34290659, end_position_on_the_genomic_accession=34327775, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14103, CurrentID=0, Status='live', Symbol='Fasl', Aliases='APT1LG1, CD178, CD95-L, CD95L, Fas-Lg, Tnfsf6, Tnlg1a, gld, Fasl', description='Fas ligand (TNF superfamily, member 6)', other_designations='CD95 ligand|Fas antigen ligand|generalized lymphoproliferative disease|tumor necrosis factor ligand 1a', map_location='1 69.95 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161779664, end_position_on_the_genomic_accession=161795538, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14104, CurrentID=0, Status='live', Symbol='Fasn', Aliases='A630082H08Rik, FAS', description='fatty acid synthase', other_designations='', map_location='11 84.56 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120805958, end_position_on_the_genomic_accession=120824547, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  14105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14105, CurrentID=0, Status='live', Symbol='Srsf10', Aliases='FUSIP2, Fusip1, NSSR1, NSSR2, Nssr, SRrp40, Sfrs13a, Srsf13a, TASR, TASR1, TASR2', description='serine/arginine-rich splicing factor 10', other_designations='FUS-interacting serine-arginine-rich protein 1|Fus-associated protein with serine-arginine repeats|TLS-associated SR protein|TLS-associated protein with SR repeats|TLS-associated protein with Ser-Arg repeats|TLS-associated serine-arginine protein|neural-salient serine/arginine-rich protein|neural-specific SR protein|serine/arginine-rich splicing factor 13A', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135855728, end_position_on_the_genomic_accession=135869908, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14106, CurrentID=0, Status='live', Symbol='Foxh1', Aliases='Fast1, Fast2, fast-1, fast-2', description='forkhead box H1', other_designations='FAST1 transcription factor|forkhead activin signal transducer 1|forkhead activin signal transducer 2', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76667829, end_position_on_the_genomic_accession=76670317, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14107, CurrentID=0, Status='live', Symbol='Fat1', Aliases='2310038E12Rik, AU023433, Fath, mFat1', description='FAT atypical cadherin 1', other_designations='FAT tumor suppressor homolog 1|fat 1 cadherin', map_location='8 24.81 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=44934725, end_position_on_the_genomic_accession=45052257, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  14109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14109, CurrentID=0, Status='live', Symbol='Fau', Aliases='MNSFbeta', description='Finkel-Biskis-Reilly murine sarcoma virus (FBR-MuSV) ubiquitously expressed (fox derived)', other_designations='MNSF-beta|monoclonal non-specific suppressor factor beta|monoclonal nonspecific suppressor factor beta|monoclonal nonspecific suppressor factor betamonoclonal nonspecific suppressor factor beta|ubiquitin-like protein FUBI', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6057888, end_position_on_the_genomic_accession=6059524, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14113, CurrentID=0, Status='live', Symbol='Fbl', Aliases='AL022665, FIB, FLRN, RNU3IP1', description='fibrillarin', other_designations='histone-glutamine methyltransferase|nucleolar protein 1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28169748, end_position_on_the_genomic_accession=28179269, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14114, CurrentID=0, Status='live', Symbol='Fbln1', Aliases='', description='fibulin 1', other_designations='BM-90|BM-90/fibulin extracellular matrix glyroprotein|FIBL-1|basement-membrane protein 90', map_location='15 E-F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85205944, end_position_on_the_genomic_accession=85286294, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14115, CurrentID=0, Status='live', Symbol='Fbln2', Aliases='5730577E14Rik, FIBL-2', description='fibulin 2', other_designations='', map_location='6 40.42 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91212764, end_position_on_the_genomic_accession=91272540, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14118, CurrentID=0, Status='live', Symbol='Fbn1', Aliases='AI536462, B430209H23, Fib-1, Tsk', description='fibrillin 1', other_designations='tight skin', map_location='2 61.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125300594, end_position_on_the_genomic_accession=125506485, orientation='minus', exon_count=69, OMIM='', no_hdr0=''),
  14119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14119, CurrentID=0, Status='live', Symbol='Fbn2', Aliases='BC063774, Fib-2, mKIAA4226, sne, sy', description='fibrillin 2', other_designations='syndatyly ems', map_location='18 D-E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=58008623, end_position_on_the_genomic_accession=58209926, orientation='minus', exon_count=65, OMIM='', no_hdr0=''),
  14120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14120, CurrentID=0, Status='live', Symbol='Fbp2', Aliases='Fbp-1, Fbp1, Rae-30', description='fructose bisphosphatase 2', other_designations='D-fructose-1,6-bisphosphate 1-phosphohydrolase 2|FBPase 2|FBPase muscle|muscle FBPase', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=62836885, end_position_on_the_genomic_accession=62858370, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14121, CurrentID=0, Status='live', Symbol='Fbp1', Aliases='Fbp-2, Fbp2, Fbp3', description='fructose bisphosphatase 1', other_designations='D-fructose-1,6-bisphosphate 1-phosphohydrolase 1|D-fructose-1,6-bisphosphate 1-phosphohydrolase 3|FBPase brain isoform|FBPase liver|fructose-1,6-bisphosphatase isozyme 3|liver FBPase', map_location='13 B3-C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=62864753, end_position_on_the_genomic_accession=62888296, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14123, CurrentID=0, Status='live', Symbol='Fbrs', Aliases='Fbs, Fbs1', description='fibrosin', other_designations='fibrosin 1|fibrosin-1|probable fibrosin-1 long transcript protein', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127477246, end_position_on_the_genomic_accession=127491513, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14125, CurrentID=0, Status='live', Symbol='Fcer1a', Aliases='FcERI, Fce1a, Fcr-5, fcepsilonri', description='Fc receptor, IgE, high affinity I, alpha polypeptide', other_designations='fc-epsilon RI-alpha|igE Fc receptor subunit alpha', map_location='1 80.33 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173221269, end_position_on_the_genomic_accession=173227232, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14126, CurrentID=0, Status='live', Symbol='Ms4a2', Aliases='FcRB, Fce1b, Fcer1b, Fcrbeta, Ms4a1, fcERI', description='membrane-spanning 4-domains, subfamily A, member 2', other_designations='Fc receptor, IgE, high affinity I, beta polypeptide|fc epsilon receptor I beta-chain|membrane-spanning 4-domains, subfamily A, member 1', map_location='19 8.46 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11615517, end_position_on_the_genomic_accession=11624034, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14127, CurrentID=0, Status='live', Symbol='Fcer1g', Aliases='AI573376, CD23, FcR-gamma, FcR[g], FcRgamma, Fce1g, FcepsilonRI, Ly-50', description='Fc receptor, IgE, high affinity I, gamma polypeptide', other_designations='Fc receptor gamma-chain|fc-epsilon RI-gamma|fceRI gamma|igE Fc receptor subunit gamma', map_location='1 79.23 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171229572, end_position_on_the_genomic_accession=171234349, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14128, CurrentID=0, Status='live', Symbol='Fcer2a', Aliases='CD23, Fce2, Fcer2, Ly-42', description='Fc receptor, IgE, low affinity II, alpha polypeptide', other_designations='FC epsilon RII|fc-epsilon-RII|low-affinity IgE receptor|lymphocyte IgE receptor', map_location='8 A1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3680726, end_position_on_the_genomic_accession=3694174, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14129, CurrentID=0, Status='live', Symbol='Fcgr1', Aliases='AI323638, AV092959, CD64, FcgammaRI, IGGHAFC', description='Fc receptor, IgG, high affinity I', other_designations='IgG high affinity Fc receptor|fc-gamma RI|fcRI|igG Fc receptor I', map_location='3 41.72 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96282909, end_position_on_the_genomic_accession=96293969, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14130, CurrentID=0, Status='live', Symbol='Fcgr2b', Aliases='AI528646, CD32, F630109E10Rik, Fc[g]RII, FcgRII, Fcgr2, Fcgr2a, Fcr-2, Fcr-3, Ly-17, Ly-m20, LyM-1, Lym-1, fcRII', description='Fc receptor, IgG, low affinity IIb', other_designations='Fc gamma RIIB|Fc-gamma-RII|FcgammaRIIB|fc gamma receptor IIB|fc-gamma RII|fc-gamma-RIIB|igG Fc receptor II beta|low affinity immunoglobulin gamma Fc region receptor II|lymphocyte antigen 17', map_location='1 78.02 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170960558, end_position_on_the_genomic_accession=170976071, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14131, CurrentID=0, Status='live', Symbol='Fcgr3', Aliases='CD16', description='Fc receptor, IgG, low affinity III', other_designations='Fc gamma receptor III|Fcg receptor III|FcgammaRIII|fc-gamma RIII|fcRIII|igG Fc receptor III', map_location='1 78.8 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171051169, end_position_on_the_genomic_accession=171065933, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14132, CurrentID=0, Status='live', Symbol='Fcgrt', Aliases='FcRn', description='Fc receptor, IgG, alpha chain transporter', other_designations='igG Fc fragment receptor transporter alpha chain|neonatal Fc receptor', map_location='7 29.12 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45092992, end_position_on_the_genomic_accession=45103846, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14133, CurrentID=0, Status='live', Symbol='Fcna', Aliases='Fcn1', description='ficolin A', other_designations='M-ficolin|collagen/fibrinogen domain-containing protein 1|ficolin-A|ficolin-alpha', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25624666, end_position_on_the_genomic_accession=25635443, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14134, CurrentID=0, Status='live', Symbol='Fcnb', Aliases='Fcn2', description='ficolin B', other_designations='L-ficolin|collagen/fibrinogen domain-containing protein 2|ficolin-B|ficolin-beta', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28076479, end_position_on_the_genomic_accession=28084913, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14137, CurrentID=0, Status='live', Symbol='Fdft1', Aliases='SQS, SS', description='farnesyl diphosphate farnesyl transferase 1', other_designations='FPP:FPP farnesyltransferase|farnesyl-diphosphate farnesyltransferase', map_location='14 33.24 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63145150, end_position_on_the_genomic_accession=63179655, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14148, CurrentID=0, Status='live', Symbol='Fdx1', Aliases='', description='ferredoxin 1', other_designations='adrenal ferredoxin|ferredoxin-1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51943023, end_position_on_the_genomic_accession=51963602, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14149, CurrentID=0, Status='live', Symbol='Fdxr', Aliases='', description='ferredoxin reductase', other_designations='AR|adrenodoxin reductase|ferredoxin--NADP(+) reductase', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115268024, end_position_on_the_genomic_accession=115277095, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14150, CurrentID=0, Status='live', Symbol='fe', Aliases='', description='faded', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14151, CurrentID=0, Status='live', Symbol='Fech', Aliases='AI894116, Fcl, fch', description='ferrochelatase', other_designations='heme synthase|heme synthetase|protoheme ferro-lyase', map_location='18 37.43 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=64456542, end_position_on_the_genomic_accession=64489095, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14154, CurrentID=0, Status='live', Symbol='Fem1a', Aliases='AW611390a, Fem1a', description='feminization 1 homolog a (C. elegans)', other_designations='FEM1-alpha-A|FEM1a-A', map_location='17 29.24 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56256793, end_position_on_the_genomic_accession=56263608, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14155, CurrentID=0, Status='live', Symbol='Fem1b', Aliases='mKIAA0396', description='feminization 1 homolog b (C. elegans)', other_designations='F1A-alpha|FEM1-beta|fem-1-like death receptor-binding protein alpha|fem-1-like in apoptotic pathway protein alpha|mt-Fem', map_location='9 33.87 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62791824, end_position_on_the_genomic_accession=62811648, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14156, CurrentID=0, Status='live', Symbol='Fen1', Aliases='AW538437, FEN-1', description='flap structure specific endonuclease 1', other_designations='FEN-1|fen1|flap structure-specific endonuclease 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10199132, end_position_on_the_genomic_accession=10204187, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14158, CurrentID=0, Status='live', Symbol='Fer', Aliases='AV082135, C330004K01Rikt, Fert2, Fer', description='fer (fms/fps related) protein kinase', other_designations='fer (fms/fps related) protein kinase, testis specific 2|p94-Fer|proto-oncogene c-Fer|proto-oncogene tyrosine-protein kinase Fer', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=63829950, end_position_on_the_genomic_accession=64146519, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  14159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14159, CurrentID=0, Status='live', Symbol='Fes', Aliases='AI586313, BB137047, FPS, c-fes', description='feline sarcoma oncogene', other_designations='proto-oncogene c-Fes|proto-oncogene tyrosine-protein kinase Fes/Fps|tyrosine kinase Fps/Fes', map_location='7 45.65 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80377758, end_position_on_the_genomic_accession=80387954, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14160, CurrentID=0, Status='live', Symbol='Lgr5', Aliases='FEX, Gpr49', description='leucine rich repeat containing G protein coupled receptor 5', other_designations='G protein-coupled receptor 49|G-protein coupled receptor 49|follicle expressed hormone|orphan G-protein coupled receptor FEX', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115450314, end_position_on_the_genomic_accession=115587780, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  14161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14161, CurrentID=0, Status='live', Symbol='Fga', Aliases='ENSMUSG00000059807, Fib', description='fibrinogen alpha chain', other_designations='fibrinogen, A alpha polypeptide', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=83026153, end_position_on_the_genomic_accession=83033627, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14163, CurrentID=0, Status='live', Symbol='Fgd1', Aliases='ZFYVE3', description='FYVE, RhoGEF and PH domain containing 1', other_designations='faciogenital dysplasia 1 protein homolog|rho/Rac GEF|rho/Rac guanine nucleotide exchange factor FGD1|zinc finger FYVE domain-containing protein 3', map_location='X 68.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151046118, end_position_on_the_genomic_accession=151089686, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14164, CurrentID=0, Status='live', Symbol='Fgf1', Aliases='Dffrx, Fam, Fgf-1, Fgfa', description='fibroblast growth factor 1', other_designations='HBGF-1|aFGF|acidic fibroblast growth factor|fibroblast growth factor 1 (acidic)|heparin-binding growth factor 1', map_location='18 20.74 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38838672, end_position_on_the_genomic_accession=38929470, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14165, CurrentID=0, Status='live', Symbol='Fgf10', Aliases='AEY17, BB213776, Fgf-10, Gsfaey17', description='fibroblast growth factor 10', other_designations='keratinocyte growth factor 2', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=118714699, end_position_on_the_genomic_accession=118792573, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14166, CurrentID=0, Status='live', Symbol='Fgf11', Aliases='Fgf-11, Fhf-3, Fhf3', description='fibroblast growth factor 11', other_designations='fibroblast growth factor homologous factor 3', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69796068, end_position_on_the_genomic_accession=69802945, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14167, CurrentID=0, Status='live', Symbol='Fgf12', Aliases='AV114868, B230343J05Rik, FGF-12, FHF-1, Fhf1', description='fibroblast growth factor 12', other_designations='fibroblast growth factor homologous factor 1|myocyte-activating factor', map_location='16 B1-B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=28157783, end_position_on_the_genomic_accession=28753243, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14168, CurrentID=0, Status='live', Symbol='Fgf13', Aliases='Fhf2', description='fibroblast growth factor 13', other_designations='fibroblast growth factor homologous factor 2', map_location='X 33.31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=59062139, end_position_on_the_genomic_accession=59585572, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14169, CurrentID=0, Status='live', Symbol='Fgf14', Aliases='FHF-4, Fhf4, mFHF-4(1B)', description='fibroblast growth factor 14', other_designations='FGF-14|fibroblast growth factor homologous factor 4', map_location='14 66.18 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=123978291, end_position_on_the_genomic_accession=124677666, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14170, CurrentID=0, Status='live', Symbol='Fgf15', Aliases='FGF19', description='fibroblast growth factor 15', other_designations='FGF-15', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144896532, end_position_on_the_genomic_accession=144900953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14171, CurrentID=0, Status='live', Symbol='Fgf17', Aliases='', description='fibroblast growth factor 17', other_designations='FGF-17', map_location='14 36.4 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70636203, end_position_on_the_genomic_accession=70643014, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14172, CurrentID=0, Status='live', Symbol='Fgf18', Aliases='D130055P09Rik, FGF-18', description='fibroblast growth factor 18', other_designations='zFGF5', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=33116970, end_position_on_the_genomic_accession=33147400, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14173, CurrentID=0, Status='live', Symbol='Fgf2', Aliases='Fgf-2, Fgfb, bFGF', description='fibroblast growth factor 2', other_designations='HBGF-2|basic fibroblast growth factor|heparin-binding growth factor 2', map_location='3 A2-B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37348469, end_position_on_the_genomic_accession=37405538, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14174, CurrentID=0, Status='live', Symbol='Fgf3', Aliases='Fgf-3, Int-2, Int-P', description='fibroblast growth factor 3', other_designations='HBGF-3|INT-2 proto-oncogene protein|heparin-binding growth factor 3|proto-oncogene Int-2', map_location='7 88.85 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144837521, end_position_on_the_genomic_accession=144844447, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14175, CurrentID=0, Status='live', Symbol='Fgf4', Aliases='Fgf-4, Fgfk, Hst1, Hstf-1, KS3, hst, hst-1, kFGF', description='fibroblast growth factor 4', other_designations='HBGF-4|K-fibroblast growth factor|heparin-binding growth factor 4', map_location='7 88.88 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144861386, end_position_on_the_genomic_accession=144865243, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14176, CurrentID=0, Status='live', Symbol='Fgf5', Aliases='Fgf-5, HBGF-5, angora, go', description='fibroblast growth factor 5', other_designations='heparin-binding growth factor 5', map_location='5 E1-F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=98254184, end_position_on_the_genomic_accession=98277033, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14177, CurrentID=0, Status='live', Symbol='Fgf6', Aliases='Fgf-6, HSTF-2', description='fibroblast growth factor 6', other_designations='HBGF-6|HST-2|heparin secretory-transforming protein 2|heparin-binding growth factor 6', map_location='6 F3-G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127015542, end_position_on_the_genomic_accession=127028182, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14178, CurrentID=0, Status='live', Symbol='Fgf7', Aliases='Kgf', description='fibroblast growth factor 7', other_designations='FGF-7|HBGF-7|Keratinocyte growth factor|heparin-binding growth factor 7', map_location='2 F-G', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126001108, end_position_on_the_genomic_accession=126091185, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14179, CurrentID=0, Status='live', Symbol='Fgf8', Aliases='Aigf, Fgf-8', description='fibroblast growth factor 8', other_designations='HBGF-8|androgen-induced growth factor|heparin-binding growth factor 8', map_location='19 C3-D', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45736798, end_position_on_the_genomic_accession=45742935, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14180, CurrentID=0, Status='live', Symbol='Fgf9', Aliases='Eks', description='fibroblast growth factor 9', other_designations='FGF-9|GAF|HBGF-9|elbow knee synostosis|glia activating factor|glia-activating factor', map_location='14 D', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=58071037, end_position_on_the_genomic_accession=58112720, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14181, CurrentID=0, Status='live', Symbol='Fgfbp1', Aliases='FGF-BP, FGF-BP1, FGFBP-1', description='fibroblast growth factor binding protein 1', other_designations='FGF-binding protein 1', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43978858, end_position_on_the_genomic_accession=43981827, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14182, CurrentID=0, Status='live', Symbol='Fgfr1', Aliases='AW208770, Eask, FLG, Fgfr-1, Flt-2, Hspy, MFR, bFGF-R-1, c-fgr', description='fibroblast growth factor receptor 1', other_designations='basic fibroblast growth factor receptor 1|proto-oncogene c-Fgr', map_location='8 14.12 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25513623, end_position_on_the_genomic_accession=25575718, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  14183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14183, CurrentID=0, Status='live', Symbol='Fgfr2', Aliases='AU043015, AW556123, Bek, Fgfr-2, Fgfr-7, Fgfr7, KGFR, KGFRTr, svs', description='fibroblast growth factor receptor 2', other_designations='keratinocyte growth factor receptor', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=130162451, end_position_on_the_genomic_accession=130266944, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  14184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14184, CurrentID=0, Status='live', Symbol='Fgfr3', Aliases='CD333, FR3, Fgfr-3, Flg-2, HBGFR, Mfr3, sam3', description='fibroblast growth factor receptor 3', other_designations='heparin-binding growth factor receptor', map_location='5 B', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33721724, end_position_on_the_genomic_accession=33737068, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  14186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14186, CurrentID=0, Status='live', Symbol='Fgfr4', Aliases='Fgfr-4', description='fibroblast growth factor receptor 4', other_designations='CTLA-2-beta protein|fibroblast growth factor receptor 4 16 minus form|protein-tyrosine kinase receptor MPK-11', map_location='13 29.8 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55152818, end_position_on_the_genomic_accession=55168759, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14187, CurrentID=0, Status='live', Symbol='Akr1b8', Aliases='FR-1, Fgfrp, Fgrp', description='aldo-keto reductase family 1, member B8', other_designations='AR|aldehyde reductase|fibroblast growth factor regulated protein|fibroblast growth factor-regulated protein|protein FR-1', map_location='6 14.91 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34354164, end_position_on_the_genomic_accession=34368454, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14190, CurrentID=0, Status='live', Symbol='Fgl2', Aliases='AI385601, musfiblp', description='fibrinogen-like protein 2', other_designations='associated with MHV-3 induced prothrombinase activity|cytotoxic T-lymphocyte-specific protein|prothrombinase', map_location='5 9.83 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21372673, end_position_on_the_genomic_accession=21378386, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14191, CurrentID=0, Status='live', Symbol='Fgr', Aliases='', description='FGR proto-oncogene, Src family tyrosine kinase', other_designations='Gardner-Rasheed feline sarcoma viral (Fgr) oncogene homolog|p55-Fgr|proto-oncogene c-Fgr|proto-oncogene tyrosine-protein kinase Fgr', map_location='4 66.11 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132974095, end_position_on_the_genomic_accession=133001893, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  14194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14194, CurrentID=0, Status='live', Symbol='Fh1', Aliases='Fh, Fh-1', description='fumarate hydratase 1', other_designations='EF-3|fumarase', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175601378, end_position_on_the_genomic_accession=175625635, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14198, CurrentID=0, Status='live', Symbol='Fhit', Aliases='AW045638, Fra14A2', description='fragile histidine triad gene', other_designations="AP3A hydrolase|AP3Aase|diadenosine 5',5'''-P1,P3-triphosphate hydrolase|dinucleosidetriphosphatase|fragile histidine triad protein", map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=9550092, end_position_on_the_genomic_accession=11162054, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14199, CurrentID=0, Status='live', Symbol='Fhl1', Aliases='FHL-1, KyoT, RAM14-1, SLIM, SLIM-1', description='four and a half LIM domains 1', other_designations='RBP-associated molecule 14-1|skeletal muscle LIM-protein 1', map_location='X A6-A7.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56731720, end_position_on_the_genomic_accession=56793346, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  14200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14200, CurrentID=0, Status='live', Symbol='Fhl2', Aliases='FHL-2, SLIM-3, SLIM3', description='four and a half LIM domains 2', other_designations='skeletal muscle LIM-protein 3', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=43123071, end_position_on_the_genomic_accession=43198299, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14201, CurrentID=0, Status='live', Symbol='Fhl3', Aliases='SLIM2', description='four and a half LIM domains 3', other_designations='skeletal muscle LIM-protein 2', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124695963, end_position_on_the_genomic_accession=124708611, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14202, CurrentID=0, Status='live', Symbol='Fhl4', Aliases='', description='four and a half LIM domains 4', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85097019, end_position_on_the_genomic_accession=85102495, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14204, CurrentID=0, Status='live', Symbol='Il4i1', Aliases='Fig1, Fig1-ps, LAAO, LAO', description='interleukin 4 induced 1', other_designations='IL4-induced protein 1|L-amino acid oxidase|interleukin-4-induced protein 1|interleukin-four induced gene 1|interleukin-four induced gene-1 protein|interleukin-four induced protein-1|protein Fig-1', map_location='7 28.93 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44836288, end_position_on_the_genomic_accession=44840809, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14205, CurrentID=0, Status='live', Symbol='Figf', Aliases='AI325264, VEGF-D, Vegfd', description='c-fos induced growth factor', other_designations='', map_location='X 76.35 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164372998, end_position_on_the_genomic_accession=164402650, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14208, CurrentID=0, Status='live', Symbol='Ppm1g', Aliases='AI385652, Fin13', description='protein phosphatase 1G (formerly 2C), magnesium-dependent, gamma isoform', other_designations='PP2C-gamma|fibroblast growth factor inducible 13|fibroblast growth factor-inducible protein 13|protein phosphatase 1C|protein phosphatase 2C isoform gamma|protein phosphatase magnesium-dependent 1 gamma', map_location='5 17.27 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31202668, end_position_on_the_genomic_accession=31220545, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14211, CurrentID=0, Status='live', Symbol='Smc2', Aliases='5730502P04Rik, AI255214, AW545314, CAP-E, CAPE, Fin16, SMC-2l1, Smc2', description='structural maintenance of chromosomes 2', other_designations='FGF-inducible protein 16|SMC protein 2|SMC2 structural maintenance of chromosomes 2-like 1|XCAP-E homolog|chromosome-associated protein E|fibroblast growth factor inducible 16', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=52439221, end_position_on_the_genomic_accession=52488365, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  14218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14218, CurrentID=0, Status='live', Symbol='Sh3pxd2a', Aliases='2310014D11Rik, AA589508, AI256723, AI413738, C230050L11, EG329070, Fish, Gm5098, Sh3md1, TSK5, Tks5', description='SH3 and PX domains 2A', other_designations='SH3 multiple domains 1|SH3 multiple domains protein 1|five SH3 domain-containing protein|five SH3 domains|tyrosine kinase substrate with five SH3 domains', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47260174, end_position_on_the_genomic_accession=47464611, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  14219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14219, CurrentID=0, Status='live', Symbol='Ctgf', Aliases='Ccn2, Fisp12, Hcs24, fisp-12', description='connective tissue growth factor', other_designations='CCN family member 2|Connective tissue growth factor precursor (CTGF) (FISP-12 protein) (Hypertrophic chondrocyte-specific protein 24)|fibroblast inducible secreated protein|fibroblast inducible secreted protein|hypertrophic chondrocyte-specific gene product 24|hypertrophic chondrocyte-specific protein 24', map_location='10 A3-B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24595442, end_position_on_the_genomic_accession=24598683, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14221, CurrentID=0, Status='live', Symbol='Fjx1', Aliases='', description='four jointed box 1 (Drosophila)', other_designations='four-jointed protein homolog|four-jointed protein x1', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=102449366, end_position_on_the_genomic_accession=102451792, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14225, CurrentID=0, Status='live', Symbol='Fkbp1a', Aliases='FKBP12, Fkbp, Fkbp1', description='FK506 binding protein 1a', other_designations='FK506 binding protein 12|PPIase FKBP1A|calstabin-1|immunophilin FKBP12', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151542483, end_position_on_the_genomic_accession=151561691, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14226, CurrentID=0, Status='live', Symbol='Fkbp1b', Aliases='12.6kDa, AW494148, FKBP12.6, calstabin2', description='FK506 binding protein 1b', other_designations='12.6 kDa FK506-binding protein|12.6 kDa FKBP|FK506-binding protein 12.6|FK506-binding protein 1B|FKBP-12.6|FKBP-1B|PPIase FKBP1B|immunophilin FKBP12.6|rotamase', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4833174, end_position_on_the_genomic_accession=4842463, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14227, CurrentID=0, Status='live', Symbol='Fkbp2', Aliases='13kDa, FKBP-2, mFKBP13, mFKBP2', description='FK506 binding protein 2', other_designations='13 kDa FK506-binding protein|13 kDa FKBP|FK506-binding protein 2|FKBP-13|PPIase FKBP2|immunophilin FKBP13|rotamase', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6977739, end_position_on_the_genomic_accession=6980501, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14228, CurrentID=0, Status='live', Symbol='Fkbp4', Aliases='59kDa, AL022792, AW208983, FKBP-4, FKBP-52, FKBP52, FKPB52, p59', description='FK506 binding protein 4', other_designations='52 kDa FK506-binding protein|52 kDa FKBP|59 kDa immunophilin|FK506-binding protein 4|FKBP59|HBI|HSP-binding immunophilin|PPIase FKBP4|immunophilin FKBP52|rotamase', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128430107, end_position_on_the_genomic_accession=128438631, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14229, CurrentID=0, Status='live', Symbol='Fkbp5', Aliases='D17Ertd592e, Dit1, FKBP-5, FKBP51', description='FK506 binding protein 5', other_designations='51 kDa FK506-binding protein|51 kDa FKBP|PPIase FKBP5|rotamase', map_location='17 14.66 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28398753, end_position_on_the_genomic_accession=28486149, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14230, CurrentID=0, Status='live', Symbol='Fkbp10', Aliases='AI325255, FKBP-10, FKBP-65, FKBP65, Fkbp-rs1, Fkbp1-rs, Fkbp6, Fkbprp', description='FK506 binding protein 10', other_designations='65 kDa FK506-binding protein|65 kDa FKBP|FK506 binding protein 6 (65 kDa)|FK506-binding protein 10|PPIase FKBP10|immunophilin FKBP65|rotamase', map_location='11 63.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100415694, end_position_on_the_genomic_accession=100424840, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14231, CurrentID=0, Status='live', Symbol='Fkbp7', Aliases='23kDa, FKBP-7, FKBP23', description='FK506 binding protein 7', other_designations='23 kDa FK506-binding protein|23 kDa FKBP|FK506-binding protein 7|FKBP-23|PPIase FKBP7|rotamase', map_location='2 45.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76663034, end_position_on_the_genomic_accession=76673139, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14232, CurrentID=0, Status='live', Symbol='Fkbp8', Aliases='38kDa, FKBP-38, FKBP-8, Fkbp38', description='FK506 binding protein 8', other_designations='38 kDa FK506-binding protein|38 kDa FKBP|FK506-binding protein 38|FK506-binding protein 8|FKBPR38|PPIase FKBP8|mFKBP38|rotamase', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70527743, end_position_on_the_genomic_accession=70535328, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14233, CurrentID=0, Status='live', Symbol='Foxi1', Aliases='5830401E05Rik, FREAC6, Fkh10, HFH-3, Hfh3', description='forkhead box I1', other_designations='HNF-3/forkhead homolog 3|forkhead homolog 10', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=34204341, end_position_on_the_genomic_accession=34208089, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14234, CurrentID=0, Status='live', Symbol='Foxc2', Aliases='Fkh14, Hfhbf3, MFH-1, Mfh1', description='forkhead box C2', other_designations='BF-3|brain factor 3|forkhead homolog 14|forkhead-related protein FKHL14|mesenchyme fork head protein 1|transcription factor FKH-14', map_location='8 70.33 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121116171, end_position_on_the_genomic_accession=121118895, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14235, CurrentID=0, Status='live', Symbol='Foxm1', Aliases='AA408308, AW554517, BB238854, Fkh16b, HFH-11B, MPHOSPH2, Mpm2, WIN, Foxm1', description='forkhead box M1', other_designations='forkhead homolog 16|trident|winged-helix transcription factor Trident', map_location='6 62.98 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128362994, end_position_on_the_genomic_accession=128375886, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14236, CurrentID=0, Status='live', Symbol='Foxn2', Aliases='3230402J05Rik, 6030465J18Rik, Fkh19, Foxn1, HTLF', description='forkhead box N2', other_designations='forkhead box N1|forkhead homolog 19', map_location='17 E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=88440712, end_position_on_the_genomic_accession=88490533, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14237, CurrentID=0, Status='live', Symbol='Foxd4', Aliases='FREAC5, Fkh2', description='forkhead box D4', other_designations='FREAC-5|forkhead homolog 2|forkhead-related protein FKHL9|forkhead-related transcription factor 5|transcription factor FKH-2', map_location='19 19.86 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24898965, end_position_on_the_genomic_accession=24901309, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14238, CurrentID=0, Status='live', Symbol='Foxf2', Aliases='FREAC2, Fkh20, LUN', description='forkhead box F2', other_designations='forkhead homolog 20', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=31625816, end_position_on_the_genomic_accession=31631406, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14239, CurrentID=0, Status='live', Symbol='Foxs1', Aliases='FREAC10, Fkh3, Fkhl18', description='forkhead box S1', other_designations='FREAC-10|forkhead-like 18 protein|forkhead-related transcription factor 10|transcription factor FKH-3', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152931898, end_position_on_the_genomic_accession=152933208, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14240, CurrentID=0, Status='live', Symbol='Foxb2', Aliases='Fkh4', description='forkhead box B2', other_designations='forkhead homolog 4|transcription factor FKH-4', map_location='19 12.26 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=16872316, end_position_on_the_genomic_accession=16873830, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14241, CurrentID=0, Status='live', Symbol='Foxl1', Aliases='FREAC7, Fkh6, fkh-6', description='forkhead box L1', other_designations='forkhead homolog 6|forkhead-related protein FKHL11|transcription factor FKH-6', map_location='8 70.34 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121127685, end_position_on_the_genomic_accession=121130644, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14245, CurrentID=0, Status='live', Symbol='Lpin1', Aliases='4631420P06, Kiaa0188, Lipin1, fld, mKIAA0188', description='lipin 1', other_designations='fatty liver dystrophy protein|lipin-1', map_location='12 7.9 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=16535669, end_position_on_the_genomic_accession=16646968, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  14246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14246, CurrentID=0, Status='live', Symbol='Flg', Aliases='AW107830, ft', description='filaggrin', other_designations='fillagrin|keratin intermediate filament-associated protein|profilaggrin', map_location='3 40.16 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93273546, end_position_on_the_genomic_accession=93293689, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14247, CurrentID=0, Status='live', Symbol='Fli1', Aliases='EWSR2, Fli-1, SIC-1, Sic1', description='Friend leukemia integration 1', other_designations='retroviral integration site protein Fli-1', map_location='9 17.74 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=32422204, end_position_on_the_genomic_accession=32543427, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14248, CurrentID=0, Status='live', Symbol='Flii', Aliases='3632430F08Rikh, Flii', description='flightless I actin binding protein', other_designations='cytoskeletal actin-modulating protein|flightless I homolog', map_location='11 B', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60714143, end_position_on_the_genomic_accession=60727263, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  14251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14251, CurrentID=0, Status='live', Symbol='Flot1', Aliases='reggie-2', description='flotillin 1', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35823320, end_position_on_the_genomic_accession=35832787, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  14252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14252, CurrentID=0, Status='live', Symbol='Flot2', Aliases='AI573412, Esa, reggie-2', description='flotillin 2', other_designations='epidermal surface antigen|membrane component chromosome 17 surface marker 1 homolog', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78037868, end_position_on_the_genomic_accession=78060434, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  14254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14254, CurrentID=0, Status='live', Symbol='Flt1', Aliases='AI323757, Flt-1, VEGFR-1, VEGFR1, sFlt1', description='FMS-like tyrosine kinase 1', other_designations='embryonic receptor kinase 2|tyrosine-protein kinase receptor FLT|vascular endothelial growth factor receptor-1|vascular permeability factor receptor', map_location='5 G', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147562196, end_position_on_the_genomic_accession=147726444, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  14255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14255, CurrentID=0, Status='live', Symbol='Flt3', Aliases='B230315G04, CD135, Flk-2, Flk2, Flt-3, Ly72, wmfl', description='FMS-like tyrosine kinase 3', other_designations='FL cytokine receptor|fetal liver kinase 2|tyrosine-protein kinase FLT3|tyrosine-protein kinase receptor flk-2', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147330741, end_position_on_the_genomic_accession=147400489, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  14256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14256, CurrentID=0, Status='live', Symbol='Flt3l', Aliases='Flt3lg, Ly72L', description='FMS-like tyrosine kinase 3 ligand', other_designations='flt3 ligand', map_location='7 B2-C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45131184, end_position_on_the_genomic_accession=45136490, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14257, CurrentID=0, Status='live', Symbol='Flt4', Aliases='AI323512, Chy, Flt-4, VEGFR-3, VEGFR3', description='FMS-like tyrosine kinase 4', other_designations='chylous ascites|receptor protein tyrosine kinase|tyrosine-protein kinase receptor FLT4|vascular endothelial growth factor receptor-3', map_location='11 A5-B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49609679, end_position_on_the_genomic_accession=49652739, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  14259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14259, CurrentID=0, Status='live', Symbol='fm', Aliases='', description='foam cell reticulosis', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14260, CurrentID=0, Status='live', Symbol='Fmn1', Aliases='Fmn, ld', description='formin 1', other_designations='limb deformity protein', map_location='2 C1-qter', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=113327699, end_position_on_the_genomic_accession=113716768, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  14261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14261, CurrentID=0, Status='live', Symbol='Fmo1', Aliases='', description='flavin containing monooxygenase 1', other_designations='FMO 1|dimethylaniline oxidase 1|hepatic flavin-containing monooxygenase 1', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162829561, end_position_on_the_genomic_accession=162866548, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14262, CurrentID=0, Status='live', Symbol='Fmo3', Aliases='AW111792', description='flavin containing monooxygenase 3', other_designations='dimethylaniline oxidase 3|hepatic flavin-containing monooxygenase 3|trimethylamine monooxygenase', map_location='1 70.34 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162953800, end_position_on_the_genomic_accession=162984491, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14263, CurrentID=0, Status='live', Symbol='Fmo5', Aliases='5033418D19Rik, AI195026', description='flavin containing monooxygenase 5', other_designations='FMO 5|dimethylaniline oxidase 5|hepatic flavin-containing monooxygenase 5', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97610185, end_position_on_the_genomic_accession=97655287, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14264, CurrentID=0, Status='live', Symbol='Fmod', Aliases='AI131919, AU041740, SLRR2E', description='fibromodulin', other_designations='FM|KSPG fibromodulin|collagen-binding 59 kDa protein|keratan sulfate proteoglycan fibromodulin', map_location='1 58.09 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134037515, end_position_on_the_genomic_accession=134048277, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14265, CurrentID=0, Status='live', Symbol='Fmr1', Aliases='FMRP, Fmr-1', description='fragile X mental retardation syndrome 1', other_designations='', map_location='X 34.83 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=68678541, end_position_on_the_genomic_accession=68717963, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14266, CurrentID=0, Status='live', Symbol='Aff2', Aliases='FMR2P, Fmr2, Ox19, Oxh', description='AF4/FMR2 family, member 2', other_designations='Ox19 homolog|fragile X mental retardation 2 homolog|fragile X mental retardation protein 2 homolog|protein FMR-2', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=69360331, end_position_on_the_genomic_accession=69872054, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  14268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14268, CurrentID=0, Status='live', Symbol='Fn1', Aliases='E330027I09, Fn, Fn-1', description='fibronectin 1', other_designations='', map_location='1 36.05 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=71585473, end_position_on_the_genomic_accession=71653234, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  14269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14269, CurrentID=0, Status='live', Symbol='Fnbp1', Aliases='1110057E06Rik, 2210010H06Rik, FBP1, Fbp17', description='formin binding protein 1', other_designations='formin binding protein 17', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31026206, end_position_on_the_genomic_accession=31142027, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14270, CurrentID=0, Status='live', Symbol='Srgap2', Aliases='9930124L22Rik, AI448945, FBP2, Fnbp2, srGAP3', description='SLIT-ROBO Rho GTPase activating protein 2', other_designations='FBP-27|formin binding protein 2|formin-binding protein 2|formin-binding protein 27', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131285251, end_position_on_the_genomic_accession=131527369, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  14272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14272, CurrentID=0, Status='live', Symbol='Fnta', Aliases='FTA', description='farnesyltransferase, CAAX box, alpha', other_designations='CAAX farnesyltransferase subunit alpha|FTase-alpha|GGTase-I-alpha|ras proteins prenyltransferase subunit alpha|type I protein geranyl-geranyltransferase subunit alpha', map_location='8 14.27 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25998722, end_position_on_the_genomic_accession=26015601, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14275, CurrentID=0, Status='live', Symbol='Folr1', Aliases='FBP1, Folbp-1, Folbp1', description='folate receptor 1 (adult)', other_designations='FR-alpha|folate binding protein 1|folate-binding protein 1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101858331, end_position_on_the_genomic_accession=101870788, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14276, CurrentID=0, Status='live', Symbol='Folr2', Aliases='FBP2, FR-P3, FR-beta, Folbp-2, Folbp2', description='folate receptor 2 (fetal)', other_designations='folate binding protein 2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101839988, end_position_on_the_genomic_accession=101857190, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14281, CurrentID=0, Status='live', Symbol='Fos', Aliases='D12Rfj1, c-fos, cFos', description='FBJ osteosarcoma oncogene', other_designations='cellular oncogene fos|proto-oncogene protein c-fos', map_location='12 39.7 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85473901, end_position_on_the_genomic_accession=85477269, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14282, CurrentID=0, Status='live', Symbol='Fosb', Aliases='', description='FBJ osteosarcoma oncogene B', other_designations='osteosarcoma oncoprotein B FosB', map_location='7 A2-B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19302696, end_position_on_the_genomic_accession=19310226, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14283, CurrentID=0, Status='live', Symbol='Fosl1', Aliases='AW538199, Fra1, fra-1', description='fos-like antigen 1', other_designations='', map_location='19 4.33 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5447698, end_position_on_the_genomic_accession=5455938, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14284, CurrentID=0, Status='live', Symbol='Fosl2', Aliases='Fra-2', description='fos-like antigen 2', other_designations='', map_location='5 17.33 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=32136472, end_position_on_the_genomic_accession=32157840, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14287, CurrentID=0, Status='live', Symbol='Fpgs', Aliases='AA408187', description='folylpolyglutamyl synthetase', other_designations='folylpoly-gamma-glutamate synthetase|tetrahydrofolate synthase|tetrahydrofolylpolyglutamate synthase', map_location='2 22.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32682609, end_position_on_the_genomic_accession=32704148, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14289, CurrentID=0, Status='live', Symbol='Fpr2', Aliases='E330010I07Rik, Fpr-rs2', description='formyl peptide receptor 2', other_designations='N-formylpeptide receptor-like 2|formyl peptide receptor, related sequence 2|formylpeptide receptor-related sequence 2|lipoxin A4 receptor-like protein', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17887824, end_position_on_the_genomic_accession=17893952, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14290, CurrentID=0, Status='live', Symbol='Fpr-rs3', Aliases='', description='formyl peptide receptor, related sequence 3', other_designations='N-formylpeptide receptor-like 3', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20623846, end_position_on_the_genomic_accession=20624877, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14291, CurrentID=0, Status='live', Symbol='Fpr-rs4', Aliases='', description='formyl peptide receptor, related sequence 4', other_designations='N-formylpeptide receptor-like 4', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=18021733, end_position_on_the_genomic_accession=18022704, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14293, CurrentID=0, Status='live', Symbol='Fpr1', Aliases='FPR, LXA4R, fMLF-R', description='formyl peptide receptor 1', other_designations='N-formyl peptide receptor|N-formylpeptide chemoattractant receptor|fMLP receptor|lipoxin A4 receptor', map_location='17 10.63 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17876471, end_position_on_the_genomic_accession=17883939, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14294, CurrentID=0, Status='live', Symbol='Fpr3', Aliases='ALX, Fpr-rs1, Fpr-s1, Fprl1, LXA4-R, Lxa4r', description='formyl peptide receptor 3', other_designations='FMLP-R-I|FMLP-related receptor I|LXA4 receptor|N-formyl peptide receptor 2|N-formyl peptide receptor 3|N-formylpeptide receptor-like 1|formyl peptide receptor-like 1|lipoxin A4 receptor', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17970458, end_position_on_the_genomic_accession=17971677, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14296, CurrentID=0, Status='live', Symbol='Frat1', Aliases='AW060382', description='frequently rearranged in advanced T cell lymphomas', other_designations='frequently rearranged in advanced T-cell lymphomas 1', map_location='19 35.3 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41829970, end_position_on_the_genomic_accession=41832583, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14297, CurrentID=0, Status='live', Symbol='Fxn', Aliases='FA, FARR, Frda, X25', description='frataxin', other_designations='Friedreich ataxia', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24261453, end_position_on_the_genomic_accession=24280586, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14299, CurrentID=0, Status='live', Symbol='Ncs1', Aliases='9430075O15Rik, A730032G13Rik, AI836659, Freq, Mfreq, NCS-1', description='neuronal calcium sensor 1', other_designations='frequenin homolog|neuronal calcium sensor-1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31245828, end_position_on_the_genomic_accession=31295989, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14300, CurrentID=0, Status='live', Symbol='Frg1', Aliases='', description='FSHD region gene 1', other_designations='FSHD region gene 1 protein', map_location='8 23.89 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=41397454, end_position_on_the_genomic_accession=41417118, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14302, CurrentID=0, Status='live', Symbol='Frk', Aliases='BSK, BSK/IYK, C85044, GTK, RAK', description='fyn-related kinase', other_designations='B-cell src-homology tyrosine kinase|beta-cell Src-homology tyrosine kinase|intestine tyrosine kinase', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34483400, end_position_on_the_genomic_accession=34611534, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67920, CurrentID=0, Status='live', Symbol='Mak16', Aliases='2600016B03Rik, AI314911, Rbm13', description='MAK16 homolog', other_designations='RNA binding motif protein 13', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=31159468, end_position_on_the_genomic_accession=31168724, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14308, CurrentID=0, Status='live', Symbol='Fshb', Aliases='FSH, FSH-B, FSH-betaeta, Fshb', description='follicle stimulating hormone beta', other_designations='follicle-stimulating hormone beta subunit|follitropin beta chain', map_location='2 56.02 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=107056140, end_position_on_the_genomic_accession=107059651, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14309, CurrentID=0, Status='live', Symbol='Fshr', Aliases='', description='follicle stimulating hormone receptor', other_designations='follitropin receptor', map_location='17 E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=88984952, end_position_on_the_genomic_accession=89200675, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14311, CurrentID=0, Status='live', Symbol='Cidec', Aliases='CIDE-3, Fsp27', description='cell death-inducing DFFA-like effector c', other_designations='CIDE-3alpha|cell death-inducing DFFA-like effector protein C|fat specific gene 27|fat-specific protein FSP27', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113424634, end_position_on_the_genomic_accession=113435760, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14312, CurrentID=0, Status='live', Symbol='Brd2', Aliases='AW228947, D17H6S113E, Frg-1, Fsrg-1, Fsrg1, Nat, Ring3, Rnf3, mKIAA4005', description='bromodomain containing 2', other_designations='female sterile homeotic-related gene 1', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34112019, end_position_on_the_genomic_accession=34122607, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  14313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14313, CurrentID=0, Status='live', Symbol='Fst', Aliases='AL033346, FS', description='follistatin', other_designations='activin-binding protein', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=114452262, end_position_on_the_genomic_accession=114458951, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14314, CurrentID=0, Status='live', Symbol='Fstl1', Aliases='AI316791, AW107808, Fstl, TSC-36', description='follistatin-like 1', other_designations='TGF-beta-inducible protein TSC-36|follistatin-like protein 1', map_location='16 26.48 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37777055, end_position_on_the_genomic_accession=37836516, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14317, CurrentID=0, Status='live', Symbol='Ftcd', Aliases='', description='formiminotransferase cyclodeaminase', other_designations='formiminotransferase-cyclodeaminase|glutamate formiminotransferase', map_location='10 39.25 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76575648, end_position_on_the_genomic_accession=76590340, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  14319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14319, CurrentID=0, Status='live', Symbol='Fth1', Aliases='FHC, Fth, HFt, MFH', description='ferritin heavy polypeptide 1', other_designations='H-ferritin|ferritin H subunit|ferritin heavy chain 1', map_location='19 A-C', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=9980600, end_position_on_the_genomic_accession=9985111, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14325, CurrentID=0, Status='live', Symbol='Ftl1', Aliases='Ftl, Ftl-1, L-ferritin', description='ferritin light polypeptide 1', other_designations='ferritin L subunit 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45457944, end_position_on_the_genomic_accession=45459886, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14337, CurrentID=0, Status='live', Symbol='Ftl2-ps', Aliases='Ftl, Ftl-2, Ftl2', description='ferritin light polypeptide 2, pseudogene', other_designations='ferritin L subunit 2|ferritin light chain 2|ferritin light chain 2, pseudogene|ferritin subunit LG', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=125504080, end_position_on_the_genomic_accession=125505018, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  14339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14339, CurrentID=0, Status='live', Symbol='Aktip', Aliases='AL023020, Fif, Ft, Ft1, Fts', description='thymoma viral proto-oncogene 1 interacting protein', other_designations='fused toes protein', map_location='8 D', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=91109871, end_position_on_the_genomic_accession=91135494, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14343, CurrentID=0, Status='live', Symbol='Fut1', Aliases='', description='fucosyltransferase 1', other_designations='GDP-L-fucose:beta-D-galactoside 2-alpha-L-fucosyltransferase 1|H transferase|alpha(1,2)FT 1|beta-galactoside alpha-1,2-fucosyltransferase', map_location='7 29.39 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45617266, end_position_on_the_genomic_accession=45621059, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14344, CurrentID=0, Status='live', Symbol='Fut2', Aliases='', description='fucosyltransferase 2', other_designations='GDP-L-fucose:beta-D-galactoside 2-alpha-L-fucosyltransferase 2|alpha(1,2)FT 2|secretory blood group protein 2', map_location='7 29.41 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45648591, end_position_on_the_genomic_accession=45666394, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14345, CurrentID=0, Status='live', Symbol='Fut4', Aliases='AI451562, CD15, FAL, FucT-IV, LeX, SSEA-1, Ssea1', description='fucosyltransferase 4', other_designations='3-fucosyl-N-acetyl-lactosamine epitope|Lewis x|alpha (1,3) fucosyltransferase, myeloid specific|alpha 1,3-fucosyltransferase|fuc-TIV|fucosyltransferase IV|galactoside 3-L-fucosyltransferase|stage-specific embryonic antigen 1', map_location='9 4.3 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14748459, end_position_on_the_genomic_accession=14752122, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14347, CurrentID=0, Status='live', Symbol='Fut7', Aliases='AI853193, FTVII, Fuc-TVII, FucT-VII', description='fucosyltransferase 7', other_designations='fucosyltransferase VII|galactoside 3-L-fucosyltransferase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25422351, end_position_on_the_genomic_accession=25426374, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14348, CurrentID=0, Status='live', Symbol='Fut9', Aliases='AI746471, AU067636, mFUT9, mFuc-TIX', description='fucosyltransferase 9', other_designations='fuc-TIX|fucT-IX|fucosyltransferase IX|galactoside 3-L-fucosyltransferase', map_location='4 10.59 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=25609333, end_position_on_the_genomic_accession=25800003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14349, CurrentID=0, Status='live', Symbol='Fv1', Aliases='Fv-1, Rv-1, Rv1', description='Friend virus susceptibility 1', other_designations='Fv1 gammaretrovirus restriction factor|Fv1 restriction factor|Rauscher leukemia virus susceptibility 1|retrovirus restriction protein', map_location='4 78.56 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147868979, end_position_on_the_genomic_accession=147870358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14352, CurrentID=0, Status='live', Symbol='Fv4', Aliases='Akvr-1, Fv-4', description='Friend virus susceptibility 4', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111619, CurrentID=0, Status='live', Symbol='Tcra-V10', Aliases='', description='T cell receptor alpha, variable 10', other_designations='T-cell receptor alpha, variable 10', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14356, CurrentID=0, Status='live', Symbol='Timm10b', Aliases='Fxc1, Tim10b, Tim9b', description='translocase of inner mitochondrial membrane 10B', other_designations='Fracture Callus 1|fractured callus 1|fractured callus expressed transcript 1|mitochondrial import inner membrane translocase subunit Tim9 B|small zinc finger-like protein', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105640250, end_position_on_the_genomic_accession=105641849, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14357, CurrentID=0, Status='live', Symbol='Dtx1', Aliases='Fxit1, mKIAA4160', description='deltex 1, E3 ubiquitin ligase', other_designations='FXI-T1|deltex 1 homolog|deltex1|fractionated x-irradiation induced transcript 1|mDTX1|protein deltex-1', map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120680203, end_position_on_the_genomic_accession=120711669, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14359, CurrentID=0, Status='live', Symbol='Fxr1', Aliases='1110050J02Rik, 9530073J07Rik, AI851072h, Fxr1p, Fxr1', description='fragile X mental retardation gene 1, autosomal homolog', other_designations='fragile X mental retardation gene, autosomal homolog', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=34020058, end_position_on_the_genomic_accession=34069980, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14360, CurrentID=0, Status='live', Symbol='Fyn', Aliases='AI448320, AW552119', description='Fyn proto-oncogene', other_designations='Src Kinase p59|p59-Fyn|proto-oncogene c-Fyn|proto-oncogene tyrosine-protein kinase Fyn', map_location='10 20.51 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39369770, end_position_on_the_genomic_accession=39565381, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14362, CurrentID=0, Status='live', Symbol='Fzd1', Aliases='AW227548, FZ-1, Fz1', description='frizzled class receptor 1', other_designations='frizzled homolog 1|mFz1', map_location='5 2.61 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=4753839, end_position_on_the_genomic_accession=4758216, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14365, CurrentID=0, Status='live', Symbol='Fzd3', Aliases='AU020229, D930050A07Rik, Fz3', description='frizzled class receptor 3', other_designations='frizzled homolog 3|fz-3|mFz3', map_location='14 34.09 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65192441, end_position_on_the_genomic_accession=65262463, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14366, CurrentID=0, Status='live', Symbol='Fzd4', Aliases='Fz4', description='frizzled class receptor 4', other_designations='frizzled homolog 4|fz-4|mFz4', map_location='7 49.32 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=89404366, end_position_on_the_genomic_accession=89413133, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14367, CurrentID=0, Status='live', Symbol='Fzd5', Aliases='5330434N09Rik, AI427138, Fz-5, Fz5, mFz5', description='frizzled class receptor 5', other_designations='frizzled homolog 5', map_location='1 32.74 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=64730558, end_position_on_the_genomic_accession=64737750, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14368, CurrentID=0, Status='live', Symbol='Fzd6', Aliases='Fz6', description='frizzled class receptor 6', other_designations='frizzled 6|frizzled homolog 6|fz-6|mFz6', map_location='15 15.22 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39005817, end_position_on_the_genomic_accession=39038194, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14369, CurrentID=0, Status='live', Symbol='Fzd7', Aliases='Fz7', description='frizzled class receptor 7', other_designations='frizzled homolog 7|fz-7|mFz7', map_location='1 30.08 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59482147, end_position_on_the_genomic_accession=59486955, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14370, CurrentID=0, Status='live', Symbol='Fzd8', Aliases='Fz8', description='frizzled class receptor 8', other_designations='frizzled homolog 8|fz-8', map_location='18 4.91 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=9212856, end_position_on_the_genomic_accession=9216201, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14371, CurrentID=0, Status='live', Symbol='Fzd9', Aliases='mfz9', description='frizzled class receptor 9', other_designations='frizzled homolog 9|fz-9|mFz3', map_location='5 75.08 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135248938, end_position_on_the_genomic_accession=135251047, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14373, CurrentID=0, Status='live', Symbol='G0s2', Aliases='AI255151, AV006465', description='G0/G1 switch gene 2', other_designations='G0S2-like protein|putative lymphocyte G0/G1 switch protein 2', map_location='1 97.69 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193272160, end_position_on_the_genomic_accession=193273188, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14375, CurrentID=0, Status='live', Symbol='Xrcc6', Aliases='70kDa, G22p1, Ku70', description='X-ray repair complementing defective repair in Chinese hamster cells 6', other_designations="5'-dRP/AP lyase Ku70|5'-deoxyribose-5-phosphate lyase Ku70|ATP-dependent DNA helicase 2 subunit 1|ATP-dependent DNA helicase II 70 kDa subunit|CTC box-binding factor 75 kDa subunit|CTC75|CTCBF|DNA repair protein XRCC6|Ku p70|ku autoantigen protein p70 homolog|thyroid autoantigen 70 kDa", map_location='15 E-F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81987826, end_position_on_the_genomic_accession=82040085, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14376, CurrentID=0, Status='live', Symbol='Ganab', Aliases='AU042638, G2an, GluII, mKIAA0088', description='alpha glucosidase 2 alpha neutral subunit', other_designations='alpha-glucosidase 2|carbohydrate processing enzyme of the endoplasmic reticulum|glucosidase II alpha-subunit|glucosidase II subunit alpha', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8898057, end_position_on_the_genomic_accession=8916742, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  14377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14377, CurrentID=0, Status='live', Symbol='G6pc', Aliases='AW107337, G6Pase, G6pt, Glc-6-Pase', description='glucose-6-phosphatase, catalytic', other_designations='Glc-6-Pase-alpha', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101367730, end_position_on_the_genomic_accession=101377903, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14378, CurrentID=0, Status='live', Symbol='G6pc2', Aliases='G6pc-rs, IGRP', description='glucose-6-phosphatase, catalytic, 2', other_designations='G-6-Pase 2|glucose-6-phosphatase, catalytic, related sequence|islet-specific glucose-6-phosphatase catalytic subunit-related protein', map_location='2 39.66 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69211073, end_position_on_the_genomic_accession=69227993, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14380, CurrentID=0, Status='live', Symbol='G6pd2', Aliases='G6pdx-ps1, Gpd-2, Gpd2', description='glucose-6-phosphate dehydrogenase 2', other_designations='G6PD|glucose-6-phosphate dehydrogenase X-linked, pseudogene 1', map_location='5 31.99 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=61808843, end_position_on_the_genomic_accession=61810477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14381, CurrentID=0, Status='live', Symbol='G6pdx', Aliases='G28A, G6pd, Gpdx', description='glucose-6-phosphate dehydrogenase X-linked', other_designations='', map_location='X A2-A3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74409486, end_position_on_the_genomic_accession=74429161, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14385, CurrentID=0, Status='live', Symbol='Slc37a4', Aliases='G6PT, G6pt1, GSD-1b, mG6PT', description='solute carrier family 37 (glucose-6-phosphate transporter), member 4', other_designations='glucose 6-phosphate transporter|glucose-6-phosphatase, transport protein 1|solute carrier family 37 (glycerol-6-phosphate transporter), member 4|transporter', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44396857, end_position_on_the_genomic_accession=44402968, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14387, CurrentID=0, Status='live', Symbol='Gaa', Aliases='E430018M07Rik', description='glucosidase, alpha, acid', other_designations='acid maltase', map_location='11 D-E', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119267967, end_position_on_the_genomic_accession=119285698, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  14388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14388, CurrentID=0, Status='live', Symbol='Gab1', Aliases='AA408973, AW107238', description='growth factor receptor bound protein 2-associated protein 1', other_designations='GRB2-associated binder 1', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=80764431, end_position_on_the_genomic_accession=80880755, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  14389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14389, CurrentID=0, Status='live', Symbol='Gab2', Aliases='AI463667, D130058I17Rik, p97', description='growth factor receptor bound protein 2-associated protein 2', other_designations='GRB2-associated binder 2|Grb2 associated binder 2|PH domain-containing adaptor molecule p97', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97081751, end_position_on_the_genomic_accession=97308951, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14390, CurrentID=0, Status='live', Symbol='Gabpa', Aliases='GABPalpha', description='GA repeat binding protein, alpha', other_designations='GA repeat binding protein alpha|GA-binding protein alpha-subunit|GABP subunit alpha', map_location='16 46.92 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=84834925, end_position_on_the_genomic_accession=84863783, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14391, CurrentID=0, Status='live', Symbol='Gabpb1', Aliases='BABPB2, E4TF1, E4TF1-47, E4TF1-53, E4Tf1B, GABPB, GABPB-1, GABPB-2, GABPB1-1, GABPB1-2, NRF2B1, NRF2B2', description='GA repeat binding protein, beta 1', other_designations='GA binding protein subunit beta1|GABP subunit beta-1|GABP subunit beta-2|Gabp beta1', map_location='2 61.76 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126628907, end_position_on_the_genomic_accession=126676677, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14394, CurrentID=0, Status='live', Symbol='Gabra1', Aliases='Gabra-1', description='gamma-aminobutyric acid (GABA) A receptor, subunit alpha 1', other_designations='GABA(A) receptor subunit alpha-1|GABAA alpha 1|GABAAR alpha1|gamma-aminobutyric acid (GABA-A) receptor, subunit alpha 1', map_location='11 24.97 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=42130939, end_position_on_the_genomic_accession=42182951, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14395, CurrentID=0, Status='live', Symbol='Gabra2', Aliases='C630048P16Rik, Gabra-2', description='gamma-aminobutyric acid (GABA) A receptor, subunit alpha 2', other_designations='GABA(A) receptor subunit alpha-2|gamma-aminobutyric acid (GABA-A) receptor, subunit alpha 2', map_location='5 37.59 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=70961037, end_position_on_the_genomic_accession=71095849, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14396, CurrentID=0, Status='live', Symbol='Gabra3', Aliases='Gabra-3', description='gamma-aminobutyric acid (GABA) A receptor, subunit alpha 3', other_designations='GABA(A) receptor subunit alpha-3', map_location='X 37.24 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72432676, end_position_on_the_genomic_accession=72656678, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14397, CurrentID=0, Status='live', Symbol='Gabra4', Aliases='Gabra-4', description='gamma-aminobutyric acid (GABA) A receptor, subunit alpha 4', other_designations='GABA(A) receptor subunit alpha-4|gamma-aminobutyric acid (GABA-A) receptor, subunit alpha 4', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=71569729, end_position_on_the_genomic_accession=71658308, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14399, CurrentID=0, Status='live', Symbol='Gabra6', Aliases='GABA-ARalpha6, Gabra-6, alpha6', description='gamma-aminobutyric acid (GABA) A receptor, subunit alpha 6', other_designations='GABA(A) receptor subunit alpha-6|GABAA receptor alpha 6 subunit|gamma-aminobutyric acid (GABA-A) receptor, subunit alpha 6|subunit alpha 6 gamma-aminobutyric acid (GABA-A) receptor', map_location='11 25.03 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=42306437, end_position_on_the_genomic_accession=42321127, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14400, CurrentID=0, Status='live', Symbol='Gabrb1', Aliases='AW061132, B230208N19Rik, Gabrb-1', description='gamma-aminobutyric acid (GABA) A receptor, subunit beta 1', other_designations='GABA(A) receptor subunit beta-1|gamma-aminobutyric acid (GABA-A) receptor, subunit beta 1', map_location='5 38.18 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=71658056, end_position_on_the_genomic_accession=72149572, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14401, CurrentID=0, Status='live', Symbol='Gabrb2', Aliases='AI834970, C030002O17Rik, C030021G16Rik, Gabrab2, Gabrb-2', description='gamma-aminobutyric acid (GABA) A receptor, subunit beta 2', other_designations='GABA(A) receptor subunit beta-2|gamma-aminobutyric acid (GABA-A) receptor, subunit beta 2|gamma-aminobutyric acid A receptor, subunit beta 2|gamma-aminobutyric acid receptor beta-2 subunit', map_location='11 25.08 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=42419757, end_position_on_the_genomic_accession=42632591, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14402, CurrentID=0, Status='live', Symbol='Gabrb3', Aliases='A230092K12Rik, AW049585, Cp1, Gabrb-3, beta3', description='gamma-aminobutyric acid (GABA) A receptor, subunit beta 3', other_designations='GABA(A) receptor subunit beta-3|cleft palate 1|gamma-aminobutyric acid (GABA-A) receptor, subunit beta 3', map_location='7 33.53 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=57590518, end_position_on_the_genomic_accession=57828802, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14403, CurrentID=0, Status='live', Symbol='Gabrd', Aliases='AI853201', description='gamma-aminobutyric acid (GABA) A receptor, subunit delta', other_designations='GABA(A) receptor subunit delta|gamma-aminobutyric acid (GABA-A) receptor, subunit delta', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155384979, end_position_on_the_genomic_accession=155398129, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14404, CurrentID=0, Status='live', Symbol='Gabre', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, subunit epsilon', other_designations='gamma-aminobutyric acid (GABA-A) receptor, subunit epsilon', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72256926, end_position_on_the_genomic_accession=72274822, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14405, CurrentID=0, Status='live', Symbol='Gabrg1', Aliases='GabaA, GabaA/BZ', description='gamma-aminobutyric acid (GABA) A receptor, subunit gamma 1', other_designations='GABA(A) receptor subunit gamma-1|gamma-aminobutyric acid (GABA-A) receptor, subunit gamma 1', map_location='5 37.43 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=70751047, end_position_on_the_genomic_accession=70843508, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14406, CurrentID=0, Status='live', Symbol='Gabrg2', Aliases='GABAA-R, Gabrg-2, gamma2', description='gamma-aminobutyric acid (GABA) A receptor, subunit gamma 2', other_designations='', map_location='11 24.8 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=41910190, end_position_on_the_genomic_accession=42000840, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14407, CurrentID=0, Status='live', Symbol='Gabrg3', Aliases='B230362M20Rik, Gabrg-3', description='gamma-aminobutyric acid (GABA) A receptor, subunit gamma 3', other_designations='GABA(A) receptor subunit gamma-3|gamma-aminobutyric acid (GABA-A) receptor, subunit gamma 3', map_location='7 33.48 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=56724242, end_position_on_the_genomic_accession=57386871, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14408, CurrentID=0, Status='live', Symbol='Gabrr1', Aliases='GABA-C', description='gamma-aminobutyric acid (GABA) C receptor, subunit rho 1', other_designations='GABA(A) receptor subunit rho-1|GABA(C) receptor|gamma-aminobutyric acid (GABA-A) receptor, subunit rho 1|gamma-aminobutyric acid (GABA-C) receptor, subunit rho 1', map_location='4 14.68 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=33120424, end_position_on_the_genomic_accession=33163588, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14409, CurrentID=0, Status='live', Symbol='Gabrr2', Aliases='', description='gamma-aminobutyric acid (GABA) C receptor, subunit rho 2', other_designations='GABA(A) receptor subunit rho-2|GABA(C) receptor|gamma-aminobutyric acid (GABA-A) receptor, subunit rho 2|gamma-aminobutyric acid (GABA-C) receptor, subunit rho 2|gamma-aminobutyric acid receptor rho2 subunit', map_location='4 14.64 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=33057712, end_position_on_the_genomic_accession=33095865, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14411, CurrentID=0, Status='live', Symbol='Slc6a12', Aliases='BGT1, GAT2, Gabt2', description='solute carrier family 6 (neurotransmitter transporter, betaine/GABA), member 12', other_designations='GAT-2|Na(+)/Cl(-) betaine/GABA transporter|gamma-aminobutyric acid (GABA-A) transporter 2|sodium- and chloride-dependent GABA transporter 2|solute carrier family 6 member 12', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121342599, end_position_on_the_genomic_accession=121365773, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  14412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14412, CurrentID=0, Status='live', Symbol='Slc6a13', Aliases='GAT3, Gabt3, Gat2', description='solute carrier family 6 (neurotransmitter transporter, GABA), member 13', other_designations='GAT-2|GAT-3|gamma-aminobutyric acid (GABA-A) transporter 3|sodium- and chloride-dependent GABA transporter 3|solute carrier family 6 member 13', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121300296, end_position_on_the_genomic_accession=121337724, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14415, CurrentID=0, Status='live', Symbol='Gad1', Aliases='EP10, Gad-1', description='glutamate decarboxylase 1', other_designations='67 kDa glutamic acid decarboxylase|glutamate decarboxylase 67 kDa isoform|glutamic acid decarboxylase 1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=70561763, end_position_on_the_genomic_accession=70602014, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  14417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14417, CurrentID=0, Status='live', Symbol='Gad2', Aliases='6330404F12Rik, GAD(65), GAD65, Gad-2', description='glutamic acid decarboxylase 2', other_designations='65 kDa glutamic acid decarboxylase|GAD-65|glutamate decarboxylase 65 kDa isoform', map_location='2 15.15 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=22621423, end_position_on_the_genomic_accession=22693877, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14419, CurrentID=0, Status='live', Symbol='Gal', Aliases='Galn', description='galanin', other_designations='', map_location='19 3.16 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3409917, end_position_on_the_genomic_accession=3414457, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14420, CurrentID=0, Status='live', Symbol='Galc', Aliases='2310068B06Rik, A930008M05Rik, AW212969, AW413532, Gacy, twi, twitcher', description='galactosylceramidase', other_designations='GALCERase|galactocerebroside beta-galactosidase|galactosylceramide beta-galactosidase', map_location='12 49.83 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98202300, end_position_on_the_genomic_accession=98259461, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  14421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14421, CurrentID=0, Status='live', Symbol='B4galnt1', Aliases='4933429D13Rik, Gal-NAc-T, GalNAc-T, GalNAcT, Galgt1, Ggm-2, Ggm2', description='beta-1,4-N-acetyl-galactosaminyl transferase 1', other_designations='(N-acetylneuraminyl)-galactosylglucosylceramide|GM2/GD2 synthase|UDP-N-acetyl-alpha-D-galactosamine:(N-acetylneuraminyl)-galactosylglucosylceramide-beta-1, 4-N-acetylgalactosaminyltransferase|beta1,4GalNAC-T', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127165156, end_position_on_the_genomic_accession=127172340, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14422, CurrentID=0, Status='live', Symbol='B4galnt2', Aliases='AI593864, Dlb-1, Dlb1, Galg2, Galgt2, Ggm3, Igf2bp1', description='beta-1,4-N-acetyl-galactosaminyl transferase 2', other_designations='CT-Gal NAc transferase|N-acetylgalactosaminyltransferase|UDP-N-acetyl-alpha-D-galactosamine:(N-acetylneuraminyl)-galactosyl-N-acetylglucosaminylpolypeptide-beta-1, 4-N-acetylgalactosaminyltransferase|dolichos lectin binding 1', map_location='11 59.02 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95863559, end_position_on_the_genomic_accession=95914920, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14423, CurrentID=0, Status='live', Symbol='Galnt1', Aliases='', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 1', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 1|galNAc-T1|polypeptide GalNAc transferase 1|pp-GaNTase 1|ppGaNTase-T1|protein-UDP acetylgalactosaminyltransferase 1', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24205344, end_position_on_the_genomic_accession=24286816, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  14425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14425, CurrentID=0, Status='live', Symbol='Galnt3', Aliases='', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 3', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 3|galNAc-T3|polypeptide GalNAc transferase 3|pp-GaNTase 3|ppGaNTase-T3|protein-UDP acetylgalactosaminyltransferase 3', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=66082766, end_position_on_the_genomic_accession=66124793, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14426, CurrentID=0, Status='live', Symbol='Galnt4', Aliases='AV011803', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 4', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 4|galNAc-T4|polypeptide GalNAc transferase 4|pp-GaNTase 4|ppGaNTase-T4|protein-UDP acetylgalactosaminyltransferase 4', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=99108135, end_position_on_the_genomic_accession=99113247, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14427, CurrentID=0, Status='live', Symbol='Galr1', Aliases='Galnr1', description='galanin receptor 1', other_designations='GAL1-R|GALR-1', map_location='18 E3-E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=82392496, end_position_on_the_genomic_accession=82406777, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14428, CurrentID=0, Status='live', Symbol='Galr2', Aliases='mGalR', description='galanin receptor 2', other_designations='GAL2-R|GALR-2', map_location='11 81.1 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116280232, end_position_on_the_genomic_accession=116287498, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14429, CurrentID=0, Status='live', Symbol='Galr3', Aliases='Galnr3', description='galanin receptor 3', other_designations='GAL3-R|GALR-3', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79041885, end_position_on_the_genomic_accession=79043558, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14430, CurrentID=0, Status='live', Symbol='Galt', Aliases='AW553376', description='galactose-1-phosphate uridyl transferase', other_designations='UDP-glucose--hexose-1-phosphate uridylyltransferase|gal-1-P uridylyltransferase', map_location='4 22.07 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41755228, end_position_on_the_genomic_accession=41759224, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14431, CurrentID=0, Status='live', Symbol='Gamt', Aliases='AA571402, Spintz1', description='guanidinoacetate methyltransferase', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80258151, end_position_on_the_genomic_accession=80261024, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14432, CurrentID=0, Status='live', Symbol='Gap43', Aliases='B-50, Basp2, GAP-43', description='growth associated protein 43', other_designations='axonal membrane protein GAP-43|brain abundant, membrane attached signal protein 2|calmodulin-binding protein P-57|growth accentuating protein 43|growth-associated protein 43', map_location='16 28.37 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=42248552, end_position_on_the_genomic_accession=42340651, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14433, CurrentID=0, Status='live', Symbol='Gapdh', Aliases='Gapd', description='glyceraldehyde-3-phosphate dehydrogenase', other_designations='peptidyl-cysteine S-nitrosylase GAPDH', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125161338, end_position_on_the_genomic_accession=125166511, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14447, CurrentID=0, Status='live', Symbol='Gapdhs', Aliases='Gapd-s, Gapds', description='glyceraldehyde-3-phosphate dehydrogenase, spermatogenic', other_designations='GAPDH-2|spermatogenic cell-specific glyceraldehyde 3-phosphate dehydrogenase 2|spermatogenic glyceraldehyde-3-phosphate dehydrogenase', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30729777, end_position_on_the_genomic_accession=30744791, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14450, CurrentID=0, Status='live', Symbol='Gart', Aliases='Gaps, Prgs', description='phosphoribosylglycinamide formyltransferase', other_designations='glycinamide ribonucleotide synthetase', map_location='16 C3-C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91621395, end_position_on_the_genomic_accession=91647481, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  14451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14451, CurrentID=0, Status='live', Symbol='Gas1', Aliases='AW554192, Gas-1', description='growth arrest specific 1', other_designations='', map_location='13 B3-C2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60174405, end_position_on_the_genomic_accession=60177535, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14453, CurrentID=0, Status='live', Symbol='Gas2', Aliases='Gas-2', description='growth arrest specific 2', other_designations='', map_location='7 32.87 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=51862360, end_position_on_the_genomic_accession=51995017, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14455, CurrentID=0, Status='live', Symbol='Gas5', Aliases='Gas-5, Mir5117', description='growth arrest specific 5', other_designations='', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161035166, end_position_on_the_genomic_accession=161038537, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14456, CurrentID=0, Status='live', Symbol='Gas6', Aliases='Gas-6', description='growth arrest specific 6', other_designations='AXL receptor tyrosine kinase ligand|GAS 6', map_location='8 6.06 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13465374, end_position_on_the_genomic_accession=13494535, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14457, CurrentID=0, Status='live', Symbol='Gas7', Aliases='AW124766, B230343A10Rik-cb, Gas7', description='growth arrest specific 7', other_designations='growth arrest-specific 7 protein', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67455024, end_position_on_the_genomic_accession=67688992, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  14459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14459, CurrentID=0, Status='live', Symbol='Gast', Aliases='GAS', description='gastrin', other_designations='', map_location='11 63.46 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100334412, end_position_on_the_genomic_accession=100336996, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14460, CurrentID=0, Status='live', Symbol='Gata1', Aliases='Gata-1, Gf-1, eryf1', description='GATA binding protein 1', other_designations='GATA-binding factor 1|GATA-binding protein 1|NF-E1 DNA-binding protein', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7959260, end_position_on_the_genomic_accession=7976677, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14461, CurrentID=0, Status='live', Symbol='Gata2', Aliases='Gata-2', description='GATA binding protein 2', other_designations='GATA-binding protein 2', map_location='6 39.2 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88192578, end_position_on_the_genomic_accession=88207032, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14462, CurrentID=0, Status='live', Symbol='Gata3', Aliases='Gata-3, jal', description='GATA binding protein 3', other_designations='GATA-binding factor 3|GATA-binding protein 3', map_location='2 6.69 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=9857078, end_position_on_the_genomic_accession=9887950, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14463, CurrentID=0, Status='live', Symbol='Gata4', Aliases='Gata-4', description='GATA binding protein 4', other_designations='GATA-binding factor 4|GATA-binding protein 4', map_location='14 33.24 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63198915, end_position_on_the_genomic_accession=63271714, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14464, CurrentID=0, Status='live', Symbol='Gata5', Aliases='', description='GATA binding protein 5', other_designations='GATA-binding factor 5|GATA-binding protein 5|Transcription factor GATA-5 (GATA binding factor-5)', map_location='2 102.85 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180325088, end_position_on_the_genomic_accession=180335512, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14465, CurrentID=0, Status='live', Symbol='Gata6', Aliases='AA410133, GATA-6', description='GATA binding protein 6', other_designations='GATA-binding factor 6|GATA-binding protein 6', map_location='18 5.63 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=11052475, end_position_on_the_genomic_accession=11085636, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14466, CurrentID=0, Status='live', Symbol='Gba', Aliases='GBA1, GC, GCase, GLUC, betaGC', description='glucosidase, beta, acid', other_designations='D-glucosyl-N-acylsphingosine glucohydrolase|acid beta glucosidase|beta-glucocerebrosidase|glucocerebrosidase', map_location='3 39.01 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89202914, end_position_on_the_genomic_accession=89208966, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14467, CurrentID=0, Status='live', Symbol='Gbas', Aliases='AV006093, Nipsnap2', description='glioblastoma amplified sequence', other_designations='glioblastoma-amplified sequence', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129725044, end_position_on_the_genomic_accession=129758327, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14468, CurrentID=0, Status='live', Symbol='Gbp2b', Aliases='Gbp-1, Gbp1, Mag-1, Mpa-1, Mpa1', description='guanylate binding protein 2b', other_designations='GTP-binding protein 1|guanylate nucleotide binding protein 1|interferon-gamma-inducible protein MAG-1|interferon-induced guanylate-binding protein 1', map_location='3 66.69 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142594816, end_position_on_the_genomic_accession=142619179, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14469, CurrentID=0, Status='live', Symbol='Gbp2', Aliases='', description='guanylate binding protein 2', other_designations='GBP-2|GTP-binding protein 2|guanine nucleotide-binding protein 2|guanylate nucleotide binding protein 2|interferon-induced guanylate-binding protein 2|mGBP-2|mGBP2', map_location='3 66.69 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142620663, end_position_on_the_genomic_accession=142638008, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14470, CurrentID=0, Status='live', Symbol='Rabac1', Aliases='2310040I06Rik, Gbpap1, PRA1, prenylin', description='Rab acceptor 1 (prenylated)', other_designations='GTP binding protein associated protein 1|PRA1 family protein 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24969750, end_position_on_the_genomic_accession=24972728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14472, CurrentID=0, Status='live', Symbol='Gbx2', Aliases='D130058E05Rik, Gbx-2, MMoxA, Stra7', description='gastrulation brain homeobox 2', other_designations='gastrulation and brain-specific homeobox protein 2|stimulated by retinoic acid gene 7 protein', map_location='1 C5-E1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=89927962, end_position_on_the_genomic_accession=89931176, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14473, CurrentID=0, Status='live', Symbol='Gc', Aliases='DBP, VDB', description='group specific component', other_designations='gc-globulin', map_location='5 44.32 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=89417511, end_position_on_the_genomic_accession=89457898, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14479, CurrentID=0, Status='live', Symbol='Usp15', Aliases='4921514G19Rik, AI327321, E430033I05Rik, Gcap18', description='ubiquitin specific peptidase 15', other_designations='deubiquitinating enzyme 15|granule cell antiserum positive 18|ubiquitin thioesterase 15|ubiquitin thiolesterase 15|ubiquitin-specific-processing protease 15', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=123105006, end_position_on_the_genomic_accession=123197019, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  14489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14489, CurrentID=0, Status='live', Symbol='Mtpn', Aliases='5033418D15Rik, Gcdp, V1', description='myotrophin', other_designations='granule cell differentiation protein|protein V-1', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=35508824, end_position_on_the_genomic_accession=35539888, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  330096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330096, CurrentID=0, Status='live', Symbol='Shisa3', Aliases='D830007B15Rik', description='shisa family member 3', other_designations='shisa homolog 3', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=67607883, end_position_on_the_genomic_accession=67613987, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14525, CurrentID=0, Status='live', Symbol='Gcsam', Aliases='Gcet, Gcet2, M17, M17-L', description='germinal center associated, signaling and motility', other_designations='germinal center B-cell-expressed transcript 2 protein|germinal center expressed transcript 2', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45610440, end_position_on_the_genomic_accession=45622867, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14526, CurrentID=0, Status='live', Symbol='Gcg', Aliases='GLP-1, Glu, PPG', description='glucagon', other_designations='glucagon-like peptide I', map_location='2 35.85 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=62474530, end_position_on_the_genomic_accession=62483655, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14527, CurrentID=0, Status='live', Symbol='Gcgr', Aliases='GR', description='glucagon receptor', other_designations='GL-R', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120529190, end_position_on_the_genomic_accession=120538987, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14528, CurrentID=0, Status='live', Symbol='Gch1', Aliases='GTP-CH, GTPCH, Gch', description='GTP cyclohydrolase 1', other_designations='GTP cyclohydrolase I|GTP-CH-I', map_location='14 C2-3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47153895, end_position_on_the_genomic_accession=47189402, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14531, CurrentID=0, Status='live', Symbol='Gcm1', Aliases='GCMa, Gcm-rs2-rs1, glide, Gcm1', description='glial cells missing homolog 1 (Drosophila)', other_designations='GCM motif protein 1|Gcm 1|Gcm a|glial cell deficient|mGCM1|mGCMa', map_location='9 E1-E3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78051958, end_position_on_the_genomic_accession=78065624, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14533, CurrentID=0, Status='live', Symbol='Bloc1s1', Aliases='AI839753, BLOS1, Gcn5l1', description='biogenesis of lysosomal organelles complex-1, subunit 1', other_designations='BLOC-1 subunit 1|GCN5 general control of amino acid synthesis-like 1|GCN5-like protein 1|biogenesis of lysosome-related organelles complex-1, subunit 1|general control of amino acid synthesis, yeast homolog-like 1|general control of amino acid synthesis-like 1', map_location='10 77.19 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128919220, end_position_on_the_genomic_accession=128923547, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14534, CurrentID=0, Status='live', Symbol='Kat2a', Aliases='1110051E14Rik, AW212720, Gcn5, Gcn5l2, mmGCN5', description='K(lysine) acetyltransferase 2A', other_designations='GCN5 general control of amino acid synthesis-like 2|PCAF-B/GCN5|general control of amino acid synthesis 5-like 2|general control of amino acid synthesis protein 5-like 2|general control of amino acid synthesis, yeast homolog-like 2|general control of amino acid synthesis-like 2|histone acetyltransferase GCN5|lysine acetyltransferase 2A', map_location='11 63.53 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100704746, end_position_on_the_genomic_accession=100712467, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  111649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111649, CurrentID=0, Status='live', Symbol='Igl-1', Aliases='', description='immunoglobulin lambda chain, cluster 1', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14536, CurrentID=0, Status='live', Symbol='Nr6a1', Aliases='1700113M01Rik, Gcnf, NCNF, RTR', description='nuclear receptor subfamily 6, group A, member 1', other_designations='germ cell nuclear factor|mGCNF|retinoid receptor-related testis-specific receptor', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38723369, end_position_on_the_genomic_accession=38927688, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14537, CurrentID=0, Status='live', Symbol='Gcnt1', Aliases='5630400D21Rik, B130048E03, C2 GlcNAcT, C2GNT, IGnT', description='glucosaminyl (N-acetyl) transferase 1, core 2', other_designations='beta-1,3-galactosyl-O-glycosyl-glycoprotein beta-1,6-N-acetylglucosaminyltransferase (Core 2 branching enzyme) (Core2-GlcNAc-transferase) (C2GNT)|core 2-branching enzyme|core2-GlcNAc-transferase', map_location='19 12.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=17326141, end_position_on_the_genomic_accession=17372844, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14538, CurrentID=0, Status='live', Symbol='Gcnt2', Aliases='5330430K10Rik, IGnT, IGnTA, IGnTB, IGnTC', description='glucosaminyl (N-acetyl) transferase 2, I-branching enzyme', other_designations='I-branching enzyme|N-acetylglucosaminyltransferase|beta-1,6-N-acetylglucosaminyltransferase|glucosaminyltransferase, I-branching enzyme|large I antigen-forming beta-1,6-N-acetylglucosaminyltransferase', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=40859768, end_position_on_the_genomic_accession=40960892, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14539, CurrentID=0, Status='live', Symbol='Opn1mw', Aliases='Gcp', description='opsin 1 (cone pigments), medium-wave-sensitive (color blindness, deutan)', other_designations='MWS opsin|green LWS photopigment|green cone photoreceptor pigment|green long wavelength sensitive cone opsin|green-sensitive opsin|medium wavelength-sensitive cone opsin|midwavelength sensitive opsin', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74127466, end_position_on_the_genomic_accession=74150756, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14544, CurrentID=0, Status='live', Symbol='Gda', Aliases='AU015411, AW047581', description='guanine deaminase', other_designations='GAH|guanase|guanine aminase|guanine aminohydrolase', map_location='19 14.32 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=21391307, end_position_on_the_genomic_accession=21472661, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14545, CurrentID=0, Status='live', Symbol='Gdap1', Aliases='', description='ganglioside-induced differentiation-associated-protein 1', other_designations='', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=17145373, end_position_on_the_genomic_accession=17164271, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14547, CurrentID=0, Status='live', Symbol='Gdap2', Aliases='C77050, D3Ertd801e', description='ganglioside-induced differentiation-associated-protein 2', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100162463, end_position_on_the_genomic_accession=100206989, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  14548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14548, CurrentID=0, Status='live', Symbol='Mrps33', Aliases='AI841153, CGI-139, Gdap3, MRP-S33, PTD003', description='mitochondrial ribosomal protein S33', other_designations='S33mt|ganglioside-induced differentiation-associated-protein 3', map_location='6 B4', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39801804, end_position_on_the_genomic_accession=39810936, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14555, CurrentID=0, Status='live', Symbol='Gpd1', Aliases='AI747587, Gdc-1, Gdc1, mKIAA4010', description='glycerol-3-phosphate dehydrogenase 1 (soluble)', other_designations='GPD-C|GPDH-C|glycerol phosphate dehydrogenase 1, cytoplasmic adult|glycerol-3-phosphate dehydrogenase [NAD+], cytoplasmic|glycerolphosphate dehydrogenase 1, cytoplasmic adult', map_location='15 56.13 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99717593, end_position_on_the_genomic_accession=99725008, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14559, CurrentID=0, Status='live', Symbol='Gdf1', Aliases='AI385651, Gdf-1', description='growth differentiation factor 1', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70315775, end_position_on_the_genomic_accession=70331592, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14560, CurrentID=0, Status='live', Symbol='Gdf10', Aliases='Bmp3b', description='growth differentiation factor 10', other_designations='BMP-3b|GDF-10|bone morphogenetic protein 3b', map_location='14 20.8 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33923587, end_position_on_the_genomic_accession=33935285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14561, CurrentID=0, Status='live', Symbol='Gdf11', Aliases='BMP-11, Bmp11', description='growth differentiation factor 11', other_designations='GDF-11|bone morphogenetic protein 11', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128884546, end_position_on_the_genomic_accession=128891718, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14562, CurrentID=0, Status='live', Symbol='Gdf3', Aliases='C78318, Gdf-3, Vgr-2, Vgr2, ecat9', description='growth differentiation factor 3', other_designations='VG-1-related protein 2', map_location='6 57.7 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122605403, end_position_on_the_genomic_accession=122610071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14563, CurrentID=0, Status='live', Symbol='Gdf5', Aliases='BMP-14, Cdmp-1, bp, brp', description='growth differentiation factor 5', other_designations='GDF-5|bone morphogenetic protein 14|brachypodism|cartilage-derived morphogenetic protein-1', map_location='2 77.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155941023, end_position_on_the_genomic_accession=155945364, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  242726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242726, CurrentID=0, Status='live', Symbol='Padi6', Aliases='Pad6, Padi5, ePAD', description='peptidyl arginine deiminase, type VI', other_designations='arginine deiminase-like protein|egg and embryo abundant PAD|peptidyl arginine deiminase, egg and embryo abundant|peptidylarginine deiminase VI|protein-arginine deiminase type VI', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140727355, end_position_on_the_genomic_accession=140742643, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  14566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14566, CurrentID=0, Status='live', Symbol='Gdf9', Aliases='Gdf-9', description='growth differentiation factor 9', other_designations='', map_location='11 31.94 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53430878, end_position_on_the_genomic_accession=53437903, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14567, CurrentID=0, Status='live', Symbol='Gdi1', Aliases='GDIA, GDIalpha', description='guanosine diphosphate (GDP) dissociation inhibitor 1', other_designations='GDI-1|Rab GDIalpha|guanosine diphosphate dissociation inhibitor 1|rab GDI alpha', map_location='X B-C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74305012, end_position_on_the_genomic_accession=74311867, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14569, CurrentID=0, Status='live', Symbol='Gdi2', Aliases='GDI-B, GDIB, Gdi3', description='guanosine diphosphate (GDP) dissociation inhibitor 2', other_designations='GDI beta|GDI-2|GDI-3|RAB GDP dissociation inhibitor beta-2|guanosine diphosphate (GDP) dissociation inhibitor 3|guanosine diphosphate dissociation inhibitor 2|rab GDI beta', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3538075, end_position_on_the_genomic_accession=3566261, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14570, CurrentID=0, Status='live', Symbol='Arhgdig', Aliases='Gdi5, RIP2, Rho-GDI-3, Rho-GDI2', description='Rho GDP dissociation inhibitor (GDI) gamma', other_designations='guanosine diphosphate (GDP) dissociation inhibitor 5, Rho associated|rho GDI 3|rho-GDI gamma', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26199183, end_position_on_the_genomic_accession=26201350, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14571, CurrentID=0, Status='live', Symbol='Gpd2', Aliases='AA408484, AI448216, AU021455, AW494132, GPDH, Gdm1, Gpd-m, Gpdh-m, TISP38', description='glycerol phosphate dehydrogenase 2, mitochondrial', other_designations='FAD-linked glycerol-3-phosphate dehydrogenase|glycerol phosphate dehydrogenase 1, mitochondrial|glycerol-3-phosphate dehydrogenase 2, mitochondrial', map_location='2 31.66 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=57237678, end_position_on_the_genomic_accession=57370719, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  407788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407788, CurrentID=0, Status='live', Symbol='BC051142', Aliases='NG8, TSBP, Tesb, Tesbp', description='cDNA sequence BC051142', other_designations='testis specific basic protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34398779, end_position_on_the_genomic_accession=34460734, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  14573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14573, CurrentID=0, Status='live', Symbol='Gdnf', Aliases='AI385739', description='glial cell line derived neurotrophic factor', other_designations='ATF|astrocyte-derived trophic factor|mGDNF|neurotrophic factor', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=7810048, end_position_on_the_genomic_accession=7837580, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  407790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407790, CurrentID=0, Status='live', Symbol='Ndufa4l2', Aliases='BC064011', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4-like 2', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127509304, end_position_on_the_genomic_accession=127517154, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14579, CurrentID=0, Status='live', Symbol='Gem', Aliases='AV020497', description='GTP binding protein (gene overexpressed in skeletal muscle)', other_designations='GTP-binding mitogen-induced T-cell protein|RAS-like protein KIR', map_location='4 5.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11704447, end_position_on_the_genomic_accession=11714993, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14580, CurrentID=0, Status='live', Symbol='Gfap', Aliases='AI836096', description='glial fibrillary acidic protein', other_designations='intermediate filament protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102887336, end_position_on_the_genomic_accession=102897212, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14581, CurrentID=0, Status='live', Symbol='Gfi1', Aliases='AW495828, Gfi-1, Pal-1, Pal1', description='growth factor independent 1', other_designations='growth factor independent protein 1', map_location='5 52.23 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107716655, end_position_on_the_genomic_accession=107726031, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14582, CurrentID=0, Status='live', Symbol='Gfi1b', Aliases='Gfi-1B', description='growth factor independent 1B', other_designations='growth factor independent protein 1B', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28609450, end_position_on_the_genomic_accession=28621982, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14583, CurrentID=0, Status='live', Symbol='Gfpt1', Aliases='2810423A18Rik, AI324119, AI449986, GFA, GFAT, GFAT1, GFAT1m, Gfpt', description='glutamine fructose-6-phosphate transaminase 1', other_designations='D-fructose-6-phosphate amidotransferase 1|GFAT 1|glucosamine--fructose-6-phosphate aminotransferase [isomerizing] 1|glutamine:fructose 6 phosphate amidotransferase 1|glutamine:fructose-6-phosphate amidotransferase 1|hexosephosphate aminotransferase 1', map_location='6 37.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87042846, end_position_on_the_genomic_accession=87092207, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14584, CurrentID=0, Status='live', Symbol='Gfpt2', Aliases='AI480523, GFAT2', description='glutamine fructose-6-phosphate transaminase 2', other_designations='D-fructose-6-phosphate amidotransferase 2|GFAT 2|glucosamine--fructose-6-phosphate aminotransferase [isomerizing] 2|glutamine:fructose 6 phosphate amidotransferase 2|glutamine:fructose-6-phosphate amidotransferase 2|hexosephosphate aminotransferase 2', map_location='11 29.9 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49794155, end_position_on_the_genomic_accession=49838620, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14585, CurrentID=0, Status='live', Symbol='Gfra1', Aliases='AU042498', description='glial cell line derived neurotrophic factor family receptor alpha 1', other_designations='GDNF receptor alpha-1|GDNFR-alpha|GDNFR-alpha-1|GFR alpha-1|GFR-alpha-1|TGF-beta-related neurotrophic factor receptor 1', map_location='19 D2-D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58235581, end_position_on_the_genomic_accession=58455946, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  14586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14586, CurrentID=0, Status='live', Symbol='Gfra2', Aliases='', description='glial cell line derived neurotrophic factor family receptor alpha 2', other_designations='GDNF receptor alpha-2|GDNF receptor beta|GDNFR-alpha-2|GDNFR-beta|GFR alpha 2|GFR alpha-2|NRTNR-alpha|NTNR-alpha|TGF-beta-related neurotrophic factor receptor 2|neurturin receptor alpha', map_location='14 D3-E1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70890107, end_position_on_the_genomic_accession=70979840, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14587, CurrentID=0, Status='live', Symbol='Gfra3', Aliases='GFRalpha3, Y15110', description='glial cell line derived neurotrophic factor family receptor alpha 3', other_designations='GDNFR-alpha-3|GFR alpha-3', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34689903, end_position_on_the_genomic_accession=34720387, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14588, CurrentID=0, Status='live', Symbol='Gfra4', Aliases='AW060733, G630015H18Rik', description='glial cell line derived neurotrophic factor family receptor alpha 4', other_designations='GDNF receptor alpha-4|GDNFR-alpha-4|GFR alpha-4|GFR-alpha-4|persephin receptor', map_location='2 63.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131039632, end_position_on_the_genomic_accession=131043088, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14590, CurrentID=0, Status='live', Symbol='Ggh', Aliases='gamma-GH', description='gamma-glutamyl hydrolase', other_designations='FGPH|GH|conjugase|folylpolyglutamate hydrolase|gamma-Glu-x carboxypeptidase', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=20041987, end_position_on_the_genomic_accession=20066111, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14593, CurrentID=0, Status='live', Symbol='Ggps1', Aliases='1810026C22Rik, 9530089B04Rik, AI843169, C79210', description='geranylgeranyl diphosphate synthase 1', other_designations='(2E,6E)-farnesyl diphosphate synthase|GGPP synthase|GGPP synthetase|GGPPSase|dimethylallyltranstransferase|farnesyl diphosphate synthase|farnesyltranstransferase|geranylgeranyl pyrophosphate synthetase|geranyltranstransferase', map_location='13 5.29 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=14049793, end_position_on_the_genomic_accession=14063734, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14594, CurrentID=0, Status='live', Symbol='Ggta1', Aliases='AW108479, GALT, Gal, Ggta, Ggta-1, alpha Gal, alpha3GalT', description='glycoprotein galactosyltransferase alpha 1, 3', other_designations='UDP-galactose:beta-D-galactosyl-1,4-N-acetyl-D-glucosaminide alpha-1,3-galactosyltransferase|glycoprotein alpha galactosyl transferase 1', map_location='2 23.6 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35400179, end_position_on_the_genomic_accession=35464381, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14595, CurrentID=0, Status='live', Symbol='B4galt1', Aliases='B-1,4-GalT, B-1,4-GalT1, GalT, Ggtb, Ggtb2, b4Gal-T1, beta-1,4-GalT, beta-1,4-GalT1, beta4Gal-T1', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 1', other_designations='UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 1|b1,4-Galactosyltransferase I|beta-1,4-GalTase 1|galactosyltransferase 2 beta 1, 4|glycoprotein galactosyltransferase beta 1, 4', map_location='4 20.46 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40804582, end_position_on_the_genomic_accession=40854537, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14598, CurrentID=0, Status='live', Symbol='Ggt1', Aliases='CD224, GGT, GGT 1, GGT-1, Ggtp, dwg', description='gamma-glutamyltransferase 1', other_designations='glutathione hydrolase 1|leukotriene-C4 hydrolase', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75564114, end_position_on_the_genomic_accession=75586193, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14599, CurrentID=0, Status='live', Symbol='Gh', Aliases='Gh1b1, Gh', description='growth hormone', other_designations='growth hormone b1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106300261, end_position_on_the_genomic_accession=106303704, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14600, CurrentID=0, Status='live', Symbol='Ghr', Aliases='GHBP, GHR/BP', description='growth hormone receptor', other_designations='GH receptor|Growth hormone receptor precursor (GH receptor) (GH binding protein) (GHBP) (Serum binding protein)|growth hormone receptor/binding protein|somatotropin receptor', map_location='15 1.84 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=3317755, end_position_on_the_genomic_accession=3583527, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14601, CurrentID=0, Status='live', Symbol='Ghrh', Aliases='Ghrf', description='growth hormone releasing hormone', other_designations='GRF|growth hormone-releasing factor|growth hormone-releasing hormone', map_location='2 78.25 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157329496, end_position_on_the_genomic_accession=157346736, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14602, CurrentID=0, Status='live', Symbol='Ghrhr', Aliases='Ghrfr, lit, little', description='growth hormone releasing hormone receptor', other_designations='GHRH receptor|GRF receptor|GRFR|growth hormone-releasing factor receptor', map_location='6 27.38 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55376295, end_position_on_the_genomic_accession=55388530, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14603, CurrentID=0, Status='live', Symbol='Gif', Aliases='AV073125', description='gastric intrinsic factor', other_designations='IF|INF|intrinsic factor', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11747543, end_position_on_the_genomic_accession=11763447, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14605, CurrentID=0, Status='live', Symbol='Tsc22d3', Aliases='DIP, Dsip1, Gilz, TSC-22R, Tilz3', description='TSC22 domain family, member 3', other_designations='TSC22-related inducible leucine zipper 3|delta sleep inducing peptide, immunoreactor|glucocorticoid-induced leucine zipper protein|long glucocorticoid-induced leucine zipper protein', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140539528, end_position_on_the_genomic_accession=140600522, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14606, CurrentID=0, Status='live', Symbol='Gpin1', Aliases='8430411H10Rik, Gin-1, Gin1', description='gross passage A viral integration region 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14607, CurrentID=0, Status='live', Symbol='Gip', Aliases='', description='gastric inhibitory polypeptide', other_designations='glucose-dependent insulinotropic polypeptide', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96024534, end_position_on_the_genomic_accession=96031296, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14608, CurrentID=0, Status='live', Symbol='Gpr83', Aliases='AW045253, Gir, Gpr72, RP105, RP39, RP82', description='G protein-coupled receptor 83', other_designations='glucocorticoid induced receptor|glucocorticoid-induced receptor', map_location='9 4.4 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14859803, end_position_on_the_genomic_accession=14870789, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14609, CurrentID=0, Status='live', Symbol='Gja1', Aliases='AU042049, AW546267, Cnx43, Cx43, Cx43alpha1, Gja-1, Npm1, connexin43', description='gap junction protein, alpha 1', other_designations='alpha 1 connexin|connexin 43|connexin-43|gap junction 43 kDa heart protein|gap junction membrane channel protein alpha 1', map_location='10 28.64 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=56377201, end_position_on_the_genomic_accession=56390423, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14610, CurrentID=0, Status='live', Symbol='Gja10', Aliases='Cx-57, Cx59, cx57', description='gap junction protein, alpha 10', other_designations='connexin-57|gap junction membrane channel protein alpha 10', map_location='4 14.25 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=32600865, end_position_on_the_genomic_accession=32602382, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14611, CurrentID=0, Status='live', Symbol='Gja3', Aliases='Cnx46, Cx43, Cx46, Gja-3', description='gap junction protein, alpha 3', other_designations='alpha 3 connexin|connexin-46|gap junction membrane channel protein alpha 3', map_location='14 29.82 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57034460, end_position_on_the_genomic_accession=57058110, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14612, CurrentID=0, Status='live', Symbol='Gja4', Aliases='AU020209, AW558810, Cnx37, Cx37, Gja-4', description='gap junction protein, alpha 4', other_designations='connexin-37|gap junction membrane channel protein alpha 4', map_location='4 61.47 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127311420, end_position_on_the_genomic_accession=127314039, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14613, CurrentID=0, Status='live', Symbol='Gja5', Aliases='5730555N10Rik, Cnx40, Cx40, Gja-5', description='gap junction protein, alpha 5', other_designations='connexin-40|gap junction membrane channel protein alpha 5', map_location='3 42.04 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97032402, end_position_on_the_genomic_accession=97053634, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14615, CurrentID=0, Status='live', Symbol='Gjc1', Aliases='C130009G16Rik, Cnx45, Cx45, Gja-7, Gja7', description='gap junction protein, gamma 1', other_designations='connexin 45|connexin-45|gap junction alpha-7 protein|gap junction membrane channel protein alpha 7', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102799472, end_position_on_the_genomic_accession=102819686, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14616, CurrentID=0, Status='live', Symbol='Gja8', Aliases='Aey5, Cnx50, Cx50, Lop10', description='gap junction protein, alpha 8', other_designations='alpha 8 connexin|connexin-50|gap junction membrane channel protein alpha 8|lens fiber protein MP70|lens opacity 10', map_location='3 42.03 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96913566, end_position_on_the_genomic_accession=96926051, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14617, CurrentID=0, Status='live', Symbol='Gjd2', Aliases='Gja9, connexin36, cx36', description='gap junction protein, delta 2', other_designations='connexin-36|gap junction alpha-9 protein|gap junction membrane channel protein alpha 9|gap junction protein, alpha 9', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=114009601, end_position_on_the_genomic_accession=114013619, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14618, CurrentID=0, Status='live', Symbol='Gjb1', Aliases='Cnx32, Cx32', description='gap junction protein, beta 1', other_designations='connexin 32|connexin-32|connexin-43|gap junction membrane channel protein beta 1', map_location='X D-F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101376376, end_position_on_the_genomic_accession=101385629, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14619, CurrentID=0, Status='live', Symbol='Gjb2', Aliases='AI325222, Cnx26, Cx26, Gjb-2', description='gap junction protein, beta 2', other_designations='connexin 26|connexin-26|gap junction membrane channel protein beta 2', map_location='14 D1-E1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57098602, end_position_on_the_genomic_accession=57104702, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14620, CurrentID=0, Status='live', Symbol='Gjb3', Aliases='Cnx31, Cx31, D4Wsu144e, Gjb-3', description='gap junction protein, beta 3', other_designations='connexin 31|connexin-31|gap junction membrane channel protein beta 3', map_location='4 61.48 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127325235, end_position_on_the_genomic_accession=127330836, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14621, CurrentID=0, Status='live', Symbol='Gjb4', Aliases='Cnx30.3, Cx30.3, Gjb-4', description='gap junction protein, beta 4', other_designations='connexin-30.3|gap junction membrane channel protein beta 4', map_location='4 D1-D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127331418, end_position_on_the_genomic_accession=127354141, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14622, CurrentID=0, Status='live', Symbol='Gjb5', Aliases='Cnx31.1, Cx31.1, Gjb-5', description='gap junction protein, beta 5', other_designations='connexin 31.1|connexin-31.1|gap junction membrane channel protein beta 5', map_location='4 61.51 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127354809, end_position_on_the_genomic_accession=127358164, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14623, CurrentID=0, Status='live', Symbol='Gjb6', Aliases='AA958971, Cx30, D14Bwg0506e', description='gap junction protein, beta 6', other_designations='connexin-30|gap junction membrane channel protein beta 6', map_location='14 30.1 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57123301, end_position_on_the_genomic_accession=57133611, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14625, CurrentID=0, Status='live', Symbol='Gykl1', Aliases='AI449806, Gk-rs1', description='glycerol kinase-like 1', other_designations='glucokinase activity, related sequence 1', map_location='18 28.25 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=52693679, end_position_on_the_genomic_accession=52695593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14626, CurrentID=0, Status='live', Symbol='Gk2', Aliases='GKTA, Gk-rs2', description='glycerol kinase 2', other_designations='ATP:glycerol 3-phosphotransferase 2|GK 2|glucokinase activity, related sequence 2|glycerokinase 2|glycerol kinase, testis specific 2', map_location='5 47.71 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=97455161, end_position_on_the_genomic_accession=97457008, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14628, CurrentID=0, Status='live', Symbol='Ostm1', Aliases='1200002H13Rik, HSPC019, gl', description='osteopetrosis associated transmembrane protein 1', other_designations='chloride channel 7 beta subunit|grey lethal osteopetrosis', map_location='10 22.89 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=42678916, end_position_on_the_genomic_accession=42702462, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14629, CurrentID=0, Status='live', Symbol='Gclc', Aliases='D9Wsu168e, GLCL-H, Ggcs-hs, Glclc', description='glutamate-cysteine ligase, catalytic subunit', other_designations='GCS heavy chain|gamma GCS-HS|gamma-ECS|gamma-glutamylcysteine synthetase heavy subunit', map_location='9 D-E', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=77754516, end_position_on_the_genomic_accession=77794489, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14630, CurrentID=0, Status='live', Symbol='Gclm', Aliases='AI649393, Gcmc, Glclr', description='glutamate-cysteine ligase, modifier subunit', other_designations='GCS light chain|gamma-ECS regulatory subunit|gamma-glutamylcysteine synthetase regulatory subunit|glutamate--cysteine ligase modifier subunit', map_location='3 H1-3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122245557, end_position_on_the_genomic_accession=122267215, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14632, CurrentID=0, Status='live', Symbol='Gli1', Aliases='AV235269, Zfp-5, Zfp5', description='GLI-Kruppel family member GLI1', other_designations='glioma-associated oncogene homolog|zinc finger protein 5', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127329881, end_position_on_the_genomic_accession=127341579, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14633, CurrentID=0, Status='live', Symbol='Gli2', Aliases='', description='GLI-Kruppel family member GLI2', other_designations='Tax helper protein', map_location='1 52.17 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=118834061, end_position_on_the_genomic_accession=119053619, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  14634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14634, CurrentID=0, Status='live', Symbol='Gli3', Aliases='AI854843, AU023367, Bph, GLI3-190, GLI3FL, Pdn, Xt, add', description='GLI-Kruppel family member GLI3', other_designations='GLI3 form of 190 kDa|GLI3 full length protein|anterior digit pattern deformity|brachyphalangy|extra toes|polydactyly Nagoya|zinc finger protein GLI3', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=15463235, end_position_on_the_genomic_accession=15730026, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14635, CurrentID=0, Status='live', Symbol='Galk1', Aliases='AA409894, GALK, Gk, Glk, Glk1', description='galactokinase 1', other_designations='galactose kinase', map_location='11 80.91 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116008357, end_position_on_the_genomic_accession=116012719, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14636, CurrentID=0, Status='live', Symbol='Gln3-1', Aliases='', description='GLN retrotransposon family 3-1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14645, CurrentID=0, Status='live', Symbol='Glul', Aliases='GS, Glns', description='glutamate-ammonia ligase (glutamine synthetase)', other_designations='glutamate decarboxylase|glutamate-ammonia ligase (glutamine synthase)', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153899929, end_position_on_the_genomic_accession=153909723, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14646, CurrentID=0, Status='live', Symbol='Glns-ps1', Aliases='', description='glutamine synthetase pseudogene 1', other_designations='glutamate-ammonia ligase (glutamine synthetase) pseudogene', map_location='11 12.26 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=19720222, end_position_on_the_genomic_accession=19721576, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  14651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14651, CurrentID=0, Status='live', Symbol='Hagh', Aliases='BC019817, Glo-2, Glo2, Rsp29', description='hydroxyacyl glutathione hydrolase', other_designations='glx II|glyoxalase II|glyoxylase 2|round spermatid protein RSP29', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24840079, end_position_on_the_genomic_accession=24864450, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14652, CurrentID=0, Status='live', Symbol='Glp1r', Aliases='GLP-1R, GLP1Rc', description='glucagon-like peptide 1 receptor', other_designations='GLP-1 receptor|GLP-1-R', map_location='17 15.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=30901867, end_position_on_the_genomic_accession=30936510, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67978, CurrentID=0, Status='live', Symbol='Tctn2', Aliases='4432405B04Rik, Tect2', description='tectonic family member 2', other_designations='tectonic 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124598749, end_position_on_the_genomic_accession=124627738, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14654, CurrentID=0, Status='live', Symbol='Glra1', Aliases='nmf11, oscillator, ot, spasmodic, spd', description='glycine receptor, alpha 1 subunit', other_designations='glycine receptor 48 kDa subunit|glycine receptor strychnine-binding subunit', map_location='11 33.12 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55514239, end_position_on_the_genomic_accession=55608198, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14657, CurrentID=0, Status='live', Symbol='Glra4', Aliases='', description='glycine receptor, alpha 4 subunit', other_designations='', map_location='X 59.1 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136756628, end_position_on_the_genomic_accession=136792158, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14658, CurrentID=0, Status='live', Symbol='Glrb', Aliases='AI853901, Glyrb, spa, spastic', description='glycine receptor, beta subunit', other_designations='glycine receptor 58 kDa subunit|inhibitory glycine receptor', map_location='3 E3-F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=80843598, end_position_on_the_genomic_accession=80913660, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14660, CurrentID=0, Status='live', Symbol='Gls', Aliases='6330442B14, AI314027, B230365M23Rik', description='glutaminase', other_designations='', map_location='1 26.86 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52163449, end_position_on_the_genomic_accession=52233232, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  14661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14661, CurrentID=0, Status='live', Symbol='Glud1', Aliases='AI118167, Gdh-X, Glud, Gludl', description='glutamate dehydrogenase 1', other_designations='GDH 1', map_location='14 20.8 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34310727, end_position_on_the_genomic_accession=34345033, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14663, CurrentID=0, Status='live', Symbol='Glycam1', Aliases='MC26, Sgp50, glyCAM-1', description='glycosylation dependent cell adhesion molecule 1', other_designations='GlyCAM 1|Sele ligand|endothelial ligand FOR L-selectin|sulfated 50 kDa glycoprotein', map_location='15 59.03 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103562759, end_position_on_the_genomic_accession=103565081, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14664, CurrentID=0, Status='live', Symbol='Slc6a9', Aliases='Glyt-1, Glyt1', description='solute carrier family 6 (neurotransmitter transporter, glycine), member 9', other_designations='glycine transporter 1|solute carrier family 6 member 9', map_location='4 53.62 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117835258, end_position_on_the_genomic_accession=117869305, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14667, CurrentID=0, Status='live', Symbol='Gm2a', Aliases='AA408702, AW215435, GM2-AP, SAP-3', description='GM2 ganglioside activator protein', other_designations='cerebroside sulfate activator protein|shingolipid activator protein 3|sphingolipid activator protein 3', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55097985, end_position_on_the_genomic_accession=55113029, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14670, CurrentID=0, Status='live', Symbol='Gnl1', Aliases='Gna-rs1, Gnal1', description='guanine nucleotide binding protein-like 1', other_designations='GTP-binding protein MMR1|guanine nucleotide binding protein, related sequence 1', map_location='17 C-D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35979955, end_position_on_the_genomic_accession=35989462, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14672, CurrentID=0, Status='live', Symbol='Gna11', Aliases='Dsk7, E430025L19', description='guanine nucleotide binding protein, alpha 11', other_designations='G-protein subunit alpha-11|g alpha-11', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81528732, end_position_on_the_genomic_accession=81545046, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14673, CurrentID=0, Status='live', Symbol='Gna12', Aliases='AI414047, AI504261, Galpha12', description='guanine nucleotide binding protein, alpha 12', other_designations='G-protein subunit alpha-12|g alpha-12', map_location='5 79.3 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140759944, end_position_on_the_genomic_accession=140830431, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14674, CurrentID=0, Status='live', Symbol='Gna13', Aliases='AU024132, AU043124, Galpha13', description='guanine nucleotide binding protein, alpha 13', other_designations='G-protein subunit alpha-13|g alpha-13', map_location='11 71.88 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109362783, end_position_on_the_genomic_accession=109401369, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14675, CurrentID=0, Status='live', Symbol='Gna14', Aliases='AU023208', description='guanine nucleotide binding protein, alpha 14', other_designations='G-protein subunit alpha-14|g alpha-14', map_location='19 11.29 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=16435667, end_position_on_the_genomic_accession=16610818, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14676, CurrentID=0, Status='live', Symbol='Gna15', Aliases='G[a]15, Galpha15', description='guanine nucleotide binding protein, alpha 15', other_designations='G-protein subunit alpha-15|g alpha-15', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81502313, end_position_on_the_genomic_accession=81524472, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14677, CurrentID=0, Status='live', Symbol='Gnai1', Aliases='AU046200, Gialpha1, Gnai-1', description='guanine nucleotide binding protein (G protein), alpha inhibiting 1', other_designations='adenylate cyclase-inhibiting G alpha protein|guanine nucleotide binding protein, alpha inhibiting 1', map_location='5 8.16 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=18265135, end_position_on_the_genomic_accession=18360413, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14678, CurrentID=0, Status='live', Symbol='Gnai2', Aliases='C76432, Galphai2, Gia, Gnai-2', description='guanine nucleotide binding protein (G protein), alpha inhibiting 2', other_designations='adenylate cyclase-inhibiting G alpha protein|guanine nucleotide binding protein, alpha inhibiting 2', map_location='9 58.43 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107614125, end_position_on_the_genomic_accession=107635391, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14679, CurrentID=0, Status='live', Symbol='Gnai3', Aliases='AI158965, AW537698, Galphai3, Gnai-3', description='guanine nucleotide binding protein (G protein), alpha inhibiting 3', other_designations='g(i) alpha-3|guanine nucleotide binding protein, alpha inhibiting 3', map_location='3 46.83 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108107275, end_position_on_the_genomic_accession=108149956, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14680, CurrentID=0, Status='live', Symbol='Gnal', Aliases='2610011C15Rik, 9630020G10Rik, Galphaolf, Gna10, Golf', description='guanine nucleotide binding protein, alpha stimulating, olfactory type', other_designations='G alpha 10|adenylate cyclase-stimulating G alpha protein, olfactory type|guanine nucleotide binding protein, alpha 10', map_location='18 39.85 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67088298, end_position_on_the_genomic_accession=67226792, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14681, CurrentID=0, Status='live', Symbol='Gnao1', Aliases='AW050213, Galphao, Gnao, alphaO', description='guanine nucleotide binding protein, alpha O', other_designations='GTP-binding protein alpha-o subunit|Go alpha', map_location='8 45.94 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93810561, end_position_on_the_genomic_accession=93969388, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14682, CurrentID=0, Status='live', Symbol='Gnaq', Aliases='1110005L02Rik, 6230401I02Rik, AA408290, AW060788, Dsk1, Dsk10, Galphaq, Gq, GqI', description='guanine nucleotide binding protein, alpha q polypeptide', other_designations='G alpha q|guanine nucleotide-binding protein alpha-q', map_location='19 11.01 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=16132831, end_position_on_the_genomic_accession=16387453, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14683, CurrentID=0, Status='live', Symbol='Gnas', Aliases='5530400H20Rik, A930027G11Rik, C130027O20Rik, GPSA, GSP, Galphas1, Gnasxl, Gs-alpha, Gsa, Nesp, Nesp55, Nespl, Oed-Sml, Oedsml, P1, P2, P3, PHP1A, PHP1B, POH, SCG6, Gnas', description='GNAS (guanine nucleotide binding protein, alpha stimulating) complex locus', other_designations='Gs alpha subunit|adenylate cyclase-stimulating G alpha protein|alternative gene product encoded by XL-exon|extra large alphas protein|guanine nucleotide-binding protein G(s) subunit alpha isoforms XLas|neuroendocrine secretory protein 55|secretogranin VI|stimulatory G-protein alpha subunit', map_location='2 E1-H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174283444, end_position_on_the_genomic_accession=174346744, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  14685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14685, CurrentID=0, Status='live', Symbol='Gnat1', Aliases='Gnat-1, Ird1, Ird2, Tralpha, irdc, irdr, transducin', description='guanine nucleotide binding protein, alpha transducing 1', other_designations='transducin alpha-1 chain', map_location='9 58.86 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107674474, end_position_on_the_genomic_accession=107679592, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14686, CurrentID=0, Status='live', Symbol='Gnat2', Aliases='AW490837, Gnat-2, Gt-2, Tcalpha', description='guanine nucleotide binding protein, alpha transducing 2', other_designations='transducin alpha-2 chain', map_location='3 46.83 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108086804, end_position_on_the_genomic_accession=108101432, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14687, CurrentID=0, Status='live', Symbol='Gnaz', Aliases='AI847979, Gz', description='guanine nucleotide binding protein, alpha z subunit', other_designations='g(x) alpha chain|gz-alpha', map_location='10 38.48 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=74967090, end_position_on_the_genomic_accession=75016907, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14688, CurrentID=0, Status='live', Symbol='Gnb1', Aliases='AA409223, C77571, Gnb-1', description='guanine nucleotide binding protein (G protein), beta 1', other_designations='guanine nucleotide binding protein, beta 1|transducin beta chain 1', map_location='4 86.17 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155491361, end_position_on_the_genomic_accession=155559269, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14693, CurrentID=0, Status='live', Symbol='Gnb2', Aliases='Gnb-2', description='guanine nucleotide binding protein (G protein), beta 2', other_designations='g protein subunit beta-2|guanine nucleotide binding protein, beta 2|guanine nucleotide-binding protein, beta-2 subunit|transducin beta chain 2', map_location='5 76.54 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137528127, end_position_on_the_genomic_accession=137533229, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14694, CurrentID=0, Status='live', Symbol='Rack1', Aliases='AL033335, GB-like, Gnb2-rs1, Gnb2l1, p205', description='receptor for activated C kinase 1', other_designations='12-3|activated protein kinase C receptor|guanine nucleotide binding protein (G protein), beta polypeptide 2 like 1 sequence 1|guanine nucleotide binding protein (G protein), beta polypeptide 2-like 1|guanine nucleotide binding protein related|guanine nucleotide binding protein, beta 2, related sequence 1|guanine nucleotide binding protein, beta-2, related sequence 1|receptor for activated C kinase|receptor of activated protein kinase C 1', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48800360, end_position_on_the_genomic_accession=48806241, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14695, CurrentID=0, Status='live', Symbol='Gnb3', Aliases='Gnb-3', description='guanine nucleotide binding protein (G protein), beta 3', other_designations='guanine nucleotide binding protein, beta 3|transducin beta chain 3', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124834240, end_position_on_the_genomic_accession=124840275, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14696, CurrentID=0, Status='live', Symbol='Gnb4', Aliases='6720453A21Rik, G(beta)4', description='guanine nucleotide binding protein (G protein), beta 4', other_designations='guanine nucleotide binding protein, beta 4|transducin beta chain 4', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32580332, end_position_on_the_genomic_accession=32616535, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  14697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14697, CurrentID=0, Status='live', Symbol='Gnb5', Aliases='GBS, Gbeta5, flr', description='guanine nucleotide binding protein (G protein), beta 5', other_designations='flail|flailer|guanine nucleotide binding protein, beta 5|hug|transducin beta chain 5', map_location='9 42.3 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75306288, end_position_on_the_genomic_accession=75345923, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  14699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14699, CurrentID=0, Status='live', Symbol='Gngt1', Aliases='G(y)1, Gng1', description='guanine nucleotide binding protein (G protein), gamma transducing activity polypeptide 1', other_designations='guanine nucleotide binding protein (G protein), gamma 1 subunit|transducin gamma chain', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3989072, end_position_on_the_genomic_accession=3997436, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14700, CurrentID=0, Status='live', Symbol='Gng10', Aliases='AV006524', description='guanine nucleotide binding protein (G protein), gamma 10', other_designations='', map_location='4 32.36 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59035156, end_position_on_the_genomic_accession=59041899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14701, CurrentID=0, Status='live', Symbol='Gng12', Aliases='2010305F15Rik, AA536815, AI115529, AI314170, AI842738', description='guanine nucleotide binding protein (G protein), gamma 12', other_designations='', map_location='6 30.68 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66896397, end_position_on_the_genomic_accession=67021361, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14702, CurrentID=0, Status='live', Symbol='Gng2', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 2', other_designations='g gamma-I', map_location='14 11.17 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=19872558, end_position_on_the_genomic_accession=19977462, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14704, CurrentID=0, Status='live', Symbol='Gng3', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 3', other_designations='guanine nucleotide binding protein (G protein), gamma 3 subunit', map_location='19 5.76 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8836929, end_position_on_the_genomic_accession=8839246, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14705, CurrentID=0, Status='live', Symbol='Bscl2', Aliases='2900097C17Rik, AI046355, Gng3lg', description='Berardinelli-Seip congenital lipodystrophy 2 (seipin)', other_designations='Berardinelli-Seip congenital lipodystrophy 2 homolog (seipin)|Bernardinelli-Seip congenital lipodystrophy 2 homolog|G protein gamma 3 linked|bernardinelli-Seip congenital lipodystrophy type 2 protein homolog|bscl2', map_location='19 5.76 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8837467, end_position_on_the_genomic_accession=8848683, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  14706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14706, CurrentID=0, Status='live', Symbol='Gng4', Aliases='', description='guanine nucleotide binding protein (G protein), gamma 4', other_designations='', map_location='13 5.29 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=13784059, end_position_on_the_genomic_accession=13827895, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14707, CurrentID=0, Status='live', Symbol='Gng5', Aliases='G(y)5', description='guanine nucleotide binding protein (G protein), gamma 5', other_designations='guanine nucleotide binding protein (G protein), gamma 5 subunit', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146499836, end_position_on_the_genomic_accession=146505543, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14708, CurrentID=0, Status='live', Symbol='Gng7', Aliases='AI840417', description='guanine nucleotide binding protein (G protein), gamma 7', other_designations='guanine nucleotide binding protein (G protein), gamma 7 subunit', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80948624, end_position_on_the_genomic_accession=81014961, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14709, CurrentID=0, Status='live', Symbol='Gng8', Aliases='G(y)8', description='guanine nucleotide binding protein (G protein), gamma 8', other_designations='gamma-9|guanine nucleotide binding protein (G protein), gamma 8 subunit', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16891737, end_position_on_the_genomic_accession=16895451, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14710, CurrentID=0, Status='live', Symbol='Gngt2', Aliases='AV096488', description='guanine nucleotide binding protein (G protein), gamma transducing activity polypeptide 2', other_designations='G-gamma-8|g gamma-C|guanine nucleotide binding protein (G protein), gamma transducing polypeptide 2', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95842219, end_position_on_the_genomic_accession=95845734, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14711, CurrentID=0, Status='live', Symbol='Gnmt', Aliases='', description='glycine N-methyltransferase', other_designations='glycine N methyl transferase', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46725664, end_position_on_the_genomic_accession=46729196, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14712, CurrentID=0, Status='live', Symbol='Gnpat', Aliases='AU019525, D1Ertd819e, DHAPAT', description='glyceronephosphate O-acyltransferase', other_designations='DAP-AT|DHAP-AT|acyl-CoA:dihydroxyacetonephosphateacyltransferase|glycerone-phosphate O-acyltransferase', map_location='8 72.81 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124863033, end_position_on_the_genomic_accession=124890057, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14714, CurrentID=0, Status='live', Symbol='Gnrh1', Aliases='Gnrh, Gnrh2, LHRH, Lhrh1, Lnrh, hpg', description='gonadotropin releasing hormone 1', other_designations='gonadotropin releasing hormone 2|luteinizing hormone-releasing hormone I|progonadoliberin I', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=67745229, end_position_on_the_genomic_accession=67749446, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14715, CurrentID=0, Status='live', Symbol='Gnrhr', Aliases='', description='gonadotropin releasing hormone receptor', other_designations='gnRH receptor', map_location='5 43.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86181994, end_position_on_the_genomic_accession=86197896, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14718, CurrentID=0, Status='live', Symbol='Got1', Aliases='AI789014, Got-1, cAspAT, cCAT', description='glutamic-oxaloacetic transaminase 1, soluble', other_designations='cysteine aminotransferase, cytoplasmic|cysteine transaminase, cytoplasmic|cytosolic aspartate aminotransferase|glutamate oxaloacetate transaminase 1, soluble|transaminase A', map_location='19 36.67 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43499752, end_position_on_the_genomic_accession=43524605, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14719, CurrentID=0, Status='live', Symbol='Got2', Aliases='AL022787, FABP-pm, Got-2, mAspAT', description='glutamatic-oxaloacetic transaminase 2, mitochondrial', other_designations='FABP-1|FABPpm|fatty acid-binding protein|glutamate oxaloacetate transaminase 2, mitochondrial|kynurenine aminotransferase 4|kynurenine aminotransferase IV|kynurenine--oxoglutarate transaminase 4|kynurenine--oxoglutarate transaminase IV|mitochondrial aspartate aminotransferase|plasma membrane fatty acid binding protein|plasma membrane-associated fatty acid-binding protein|transaminase A', map_location='8 47.79 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95864137, end_position_on_the_genomic_accession=95888365, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14720, CurrentID=0, Status='live', Symbol='Got2-ps1', Aliases='', description='glutamatic-oxaloacetic transaminase 2, mitochondrial, pseudogene 1', other_designations='glutamate oxaloacetate transaminase 2, mitochondrial, pseudogene 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138363272, end_position_on_the_genomic_accession=138365608, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  14723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14723, CurrentID=0, Status='live', Symbol='Gp1ba', Aliases='GPIba, GPIbalpha', description='glycoprotein 1b, alpha polypeptide', other_designations='GP-Ib alpha|glycoprotein Ibalpha', map_location='11 43.2 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70636651, end_position_on_the_genomic_accession=70644030, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14724, CurrentID=0, Status='live', Symbol='Gp1bb', Aliases='', description='glycoprotein Ib, beta polypeptide', other_designations='GP-Ib beta|GPIb-beta|GPIbB', map_location='16 11.53 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18620319, end_position_on_the_genomic_accession=18622403, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14725, CurrentID=0, Status='live', Symbol='Lrp2', Aliases='AI315343, AW536255, D230004K18Rik, Gp330, Megalin, b2b1625.2Clo', description='low density lipoprotein receptor-related protein 2', other_designations='LRP-2|glycoprotein 330', map_location='2 40.74 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69424335, end_position_on_the_genomic_accession=69586067, orientation='minus', exon_count=79, OMIM='', no_hdr0=''),
  14726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14726, CurrentID=0, Status='live', Symbol='Pdpn', Aliases='Gp38, OTS-8, RANDAM-2, T1-alpha, T1a, T1alpha', description='podoplanin', other_designations='PA2.26 antigen|aggrus|glycoprotein 38|transmembrane glycoprotein E11', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143267409, end_position_on_the_genomic_accession=143299564, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14727, CurrentID=0, Status='live', Symbol='Lilr4b', Aliases='Gp49a, gp49', description='leukocyte immunoglobulin-like receptor, subfamily B, member 4B', other_designations='glycoprotein 49 A', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=51480612, end_position_on_the_genomic_accession=51486329, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14728, CurrentID=0, Status='live', Symbol='Lilrb4a', Aliases='CD85K, Gp49b, HM18, ILT3, LIR-5, Lilrb4, gp49', description='leukocyte immunoglobulin-like receptor, subfamily B, member 4A', other_designations='glycoprotein 49 B|leukocyte immunoglobulin-like receptor, subfamily B, member 4|mast cell surface glycoprotein Gp49B', map_location='10 25.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=51490883, end_position_on_the_genomic_accession=51496613, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14729, CurrentID=0, Status='live', Symbol='Gp5', Aliases='GPV', description='glycoprotein 5 (platelet)', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30307685, end_position_on_the_genomic_accession=30310781, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14731, CurrentID=0, Status='live', Symbol='Gpaa1', Aliases='C80044, mGAA1', description='GPI anchor attachment protein 1', other_designations='GAA1 protein homolog', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76331294, end_position_on_the_genomic_accession=76334900, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14732, CurrentID=0, Status='live', Symbol='Gpam', Aliases='GPAT, GPAT1', description='glycerol-3-phosphate acyltransferase, mitochondrial', other_designations='GPAT-1|P90', map_location='19 50.81 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=55067635, end_position_on_the_genomic_accession=55127233, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  14733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14733, CurrentID=0, Status='live', Symbol='Gpc1', Aliases='AI462976', description='glypican 1', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92831686, end_position_on_the_genomic_accession=92860211, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14734, CurrentID=0, Status='live', Symbol='Gpc3', Aliases='OCI-5', description='glypican 3', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52272426, end_position_on_the_genomic_accession=52614245, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14735, CurrentID=0, Status='live', Symbol='Gpc4', Aliases='9530073D23Rik, AF311610, AI385680, AI661372, K-glypican', description='glypican 4', other_designations='', map_location='X 28.33 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52050885, end_position_on_the_genomic_accession=52164923, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14738, CurrentID=0, Status='live', Symbol='Gpr12', Aliases='Gpcr01, Gpcr12, Gpcr20, Gpcr21', description='G-protein coupled receptor 12', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146582150, end_position_on_the_genomic_accession=146586487, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14739, CurrentID=0, Status='live', Symbol='S1pr2', Aliases='1100001A16Rik, Edg5, Gpcr13, H218, LPb2, S1P2', description='sphingosine-1-phosphate receptor 2', other_designations='G-protein coupled receptor 13|S1P receptor 2|S1P receptor Edg-5|endothelial differentiation G-protein coupled receptor 5|endothelial differentiation, sphingolipid G-protein-coupled receptor, 5|lysophospholipid receptor B2|sphingosine 1-phosphate receptor Edg-5|sphingosine-1-phosphate receptor S1P(2)', map_location='9 7.68 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20965952, end_position_on_the_genomic_accession=20976793, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14744, CurrentID=0, Status='live', Symbol='Gpr65', Aliases='Dig1, Gpcr25, TDAG8', description='G-protein coupled receptor 65', other_designations='G-protein coupled receptor 25|T-cell death-associated gene 8 protein', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98268635, end_position_on_the_genomic_accession=98276722, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14745, CurrentID=0, Status='live', Symbol='Lpar1', Aliases='AI326300, Edg2, Gpcr26, Kdt2, lpA1, vzg-1', description='lysophosphatidic acid receptor 1', other_designations='G-protein coupled receptor 26|LPA receptor 1|endothelial differentiation lysophosphatidic acid G-protein-coupled receptor 2|lysophosphatidic acid receptor Edg-2|rec1.3', map_location='4 32.2 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58435252, end_position_on_the_genomic_accession=58554027, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  276891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276891, CurrentID=0, Status='live', Symbol='Timd4', Aliases='B430010N18Rik, TIM-4, Tim4', description='T cell immunoglobulin and mucin domain containing 4', other_designations='T-cell immunoglobulin and mucin domain containing 4|T-cell immunoglobulin mucin receptor 4|T-cell membrane protein 4|TIMD-4|smuckler|spleen, mucin-containing, knockout of lymphotoxin protein', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46810786, end_position_on_the_genomic_accession=46844333, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14748, CurrentID=0, Status='live', Symbol='Gpr3', Aliases='Gpcr20, Gpcr21, Gpcr3', description='G-protein coupled receptor 3', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133209340, end_position_on_the_genomic_accession=133212536, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14751, CurrentID=0, Status='live', Symbol='Gpi1', Aliases='Amf, Gpi, Gpi-1, Gpi-1r, Gpi-1s, Gpi-1t-r, Gpi1-s, Gpi1-t, Gpi1s, MF, NK, NK/GPI, Nlk, Org, Pgi, Phi, Gpi1', description='glucose phosphate isomerase 1', other_designations='autocrine motility factor|glucose phosphate isomerase 1 complex|glucose phosphate isomerase 1, regulatory|glucose phosphate isomerase 1, structural|glucose phosphate isomerase 1, temporal|maturation factor|neuroleukin|phosphoglucose isomerase|phosphohexose isomerase', map_location='7 19.46 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34201327, end_position_on_the_genomic_accession=34230336, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  14755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14755, CurrentID=0, Status='live', Symbol='Pigq', Aliases='Gpi1, Gpi1h, Gpi1p, Gpih', description='phosphatidylinositol glycan anchor biosynthesis, class Q', other_designations='N-acetylglucosamyl transferase component GPI1|PIG-Q|glycosylphosphatidylinositol 1 homolog|phosphatidylinositol-glycan biosynthesis class Q protein', map_location='17 B', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25926420, end_position_on_the_genomic_accession=25942877, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14756, CurrentID=0, Status='live', Symbol='Gpld1', Aliases='6330541J12Rik, AW546131', description='glycosylphosphatidylinositol specific phospholipase D1', other_designations='GPI-PLD|GPI-specific phospholipase D|PI-G PLD|glycoprotein phospholipase D|glycosyl-phosphatidylinositol-specific phospholipase D', map_location='13 10.79 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24943152, end_position_on_the_genomic_accession=24991936, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  14758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14758, CurrentID=0, Status='live', Symbol='Gpm6b', Aliases='AI593561, Gpm6, M6B', description='glycoprotein m6b', other_designations='', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=166238943, end_position_on_the_genomic_accession=166389033, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  14760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14760, CurrentID=0, Status='live', Symbol='Gpr19', Aliases='', description='G protein-coupled receptor 19', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134869092, end_position_on_the_genomic_accession=134901063, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14761, CurrentID=0, Status='live', Symbol='Gpr27', Aliases='', description='G protein-coupled receptor 27', other_designations='super conserved receptor expressed in brain 1', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=99692679, end_position_on_the_genomic_accession=99693818, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14762, CurrentID=0, Status='live', Symbol='Gpr33', Aliases='', description='G protein-coupled receptor 33', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=52023003, end_position_on_the_genomic_accession=52028063, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14763, CurrentID=0, Status='live', Symbol='Gpr37', Aliases='AI848630, Pael-R', description='G protein-coupled receptor 37', other_designations='G-protein coupled receptor 37|parkin-associated endothelin B-like receptor|probable G-protein coupled receptor 37', map_location='6 11.65 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=25668523, end_position_on_the_genomic_accession=25689980, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14764, CurrentID=0, Status='live', Symbol='Ptgdr2', Aliases='Crth2, Gpr44, Grp45', description='prostaglandin D2 receptor 2', other_designations='G protein-coupled receptor 44|G-protein coupled receptor 44|PGD2 receptor|chemoattractant receptor-homologous molecule expressed on TH2 cells|prostaglandin D2 receptor CRTH2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10927861, end_position_on_the_genomic_accession=10942511, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14765, CurrentID=0, Status='live', Symbol='Gpr50', Aliases='H9', description='G-protein-coupled receptor 50', other_designations='', map_location='X 36.98 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71663767, end_position_on_the_genomic_accession=71669257, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14766, CurrentID=0, Status='live', Symbol='Adgrg1', Aliases='Cyt28, Gpr56, TM7LN4, TM7XN1', description='adhesion G protein-coupled receptor G1', other_designations='G protein-coupled receptor 56|secretin receptor|serpentine receptor cyt28', map_location='8 47.12 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94974756, end_position_on_the_genomic_accession=95014208, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  14767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14767, CurrentID=0, Status='live', Symbol='Nmur1', Aliases='FM-3, Gpr66, NMU-R1, NMU1R, NmU-R', description='neuromedin U receptor 1', other_designations='G protein-coupled receptor 66', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86386325, end_position_on_the_genomic_accession=86388141, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14768, CurrentID=0, Status='live', Symbol='Lancl1', Aliases='AW124738, Gpr69, Gpr69a, p40', description='LanC (bacterial lantibiotic synthetase component C)-like 1', other_designations='40 kDa erythrocyte membrane protein|G protein-coupled receptor 69|G protein-coupled receptor 69A', map_location='1 C2-C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=67000517, end_position_on_the_genomic_accession=67038872, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14772, CurrentID=0, Status='live', Symbol='Grk4', Aliases='A830025H08Rik, GRK, Gprk2l, Gprk4', description='G protein-coupled receptor kinase 4', other_designations='G protein-coupled receptor kinase 2, groucho gene related|g protein-coupled receptor kinase GRK4', map_location='5 17.91 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34660287, end_position_on_the_genomic_accession=34755305, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14773, CurrentID=0, Status='live', Symbol='Grk5', Aliases='Gprk5', description='G protein-coupled receptor kinase 5', other_designations='g protein-coupled receptor kinase GRK5', map_location='19 56.52 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60889593, end_position_on_the_genomic_accession=61095968, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14775, CurrentID=0, Status='live', Symbol='Gpx1', Aliases='AI195024, AL033363, CGPx, GPx-1, GSHPx-1, Gpx', description='glutathione peroxidase 1', other_designations='cellular glutathione peroxidase|selenium-dependent glutathione peroxidase 1', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108339080, end_position_on_the_genomic_accession=108340342, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14776, CurrentID=0, Status='live', Symbol='Gpx2', Aliases='GI-GPx, GPx-GI, GSHPx-2, GSHPx-GI', description='glutathione peroxidase 2', other_designations='GPx-2|glutathione peroxidase-gastrointestinal|intestinal GPx', map_location='12 33.73 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76792335, end_position_on_the_genomic_accession=76795554, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  14778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14778, CurrentID=0, Status='live', Symbol='Gpx3', Aliases='AA960521, EGPx, GPx, GSHPx-3, GSHPx-P', description='glutathione peroxidase 3', other_designations='GPx-3|GPx-P|extracellular GPx|plasma GPx|plasma glutathione peroxidase', map_location='11 B3-B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54902854, end_position_on_the_genomic_accession=54910382, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  14780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14780, CurrentID=0, Status='live', Symbol='Gpx5', Aliases='AV379049, Arep', description='glutathione peroxidase 5', other_designations='EGLP|GPx-5|GSHPx-5|arMEP24|epididymis-specific glutathione peroxidase-like protein|lysosomal trafficking regulator, long splice form|major androgen-regulated protein', map_location='13 A2-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21286429, end_position_on_the_genomic_accession=21292729, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14781, CurrentID=0, Status='live', Symbol='gr', Aliases='', description='grizzled', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14782, CurrentID=0, Status='live', Symbol='Gsr', Aliases='AI325518, D8Ertd238e, Gr-1, Gr1', description='glutathione reductase', other_designations='GR|GRase|glutathione reductase 1', map_location='8 20.69 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33653238, end_position_on_the_genomic_accession=33698162, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14783, CurrentID=0, Status='live', Symbol='Grb10', Aliases='5730571D09Rik, AI325020, Meg1, mKIAA0207', description='growth factor receptor bound protein 10', other_designations='GRB10 adapter protein|maternally expressed gene 1 protein', map_location='11 7.15 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=11930499, end_position_on_the_genomic_accession=12037422, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14784, CurrentID=0, Status='live', Symbol='Grb2', Aliases='AA408164, Ash', description='growth factor receptor bound protein 2', other_designations='SH2/SH3 adapter GRB2|adapter protein GRB2', map_location='11 80.91 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115644130, end_position_on_the_genomic_accession=115708597, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14786, CurrentID=0, Status='live', Symbol='Grb7', Aliases='mKIAA4028', description='growth factor receptor bound protein 7', other_designations='GRB7 adapter protein|Growth factor receptor-bound protein 7 (GRB7 adapter protein) (Epidermal growth factor receptor GRB-7)|epidermal growth factor receptor GRB-7', map_location='11 61.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98446834, end_position_on_the_genomic_accession=98455373, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14787, CurrentID=0, Status='live', Symbol='Rhpn1', Aliases='BB023497, Grbp, Rhophilin, mKIAA1929', description='rhophilin, Rho GTPase binding protein 1', other_designations='GTP-Rho-binding protein 1|GTP-rho binding protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75704285, end_position_on_the_genomic_accession=75714419, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  14788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14788, CurrentID=0, Status='live', Symbol='Gpr162', Aliases='A-2, BC031437, Grca', description='G protein-coupled receptor 162', other_designations='gene rich cluster, A', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124858444, end_position_on_the_genomic_accession=124864219, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14789, CurrentID=0, Status='live', Symbol='P3h3', Aliases='BC016431, Grcb, Leprel2', description='prolyl 3-hydroxylase 3', other_designations='gene rich cluster, B|leprecan-like 2|leprecan-like protein 2|protein B', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124841095, end_position_on_the_genomic_accession=124857746, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14790, CurrentID=0, Status='live', Symbol='Grcc10', Aliases='2310033H05Rik, C10', description='gene rich cluster, C10 gene', other_designations='B-cell receptor-associated protein 37', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124739183, end_position_on_the_genomic_accession=124741079, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14791, CurrentID=0, Status='live', Symbol='Emg1', Aliases='C2f, Grcc2f', description='EMG1 N1-specific pseudouridine methyltransferase', other_designations='18S rRNA (pseudouridine(1248)-N1)-methyltransferase|18S rRNA (pseudouridine-N1-)-methyltransferase NEP1|18S rRNA Psi1248 methyltransferase|EMG1 nucleolar protein homolog|gene rich cluster, C2f|nucleolar protein EMG1 homolog|probable ribosome biogenesis protein NEP1|protein C2f|ribosome biogenesis protein NEP1', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124704370, end_position_on_the_genomic_accession=124712178, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14792, CurrentID=0, Status='live', Symbol='Lpcat3', Aliases='C3f, Grcc3f, Lpcat, Lpeat, Lplat5, Lpsat, Mboat5, Moact5, Oact5, PTG', description='lysophosphatidylcholine acyltransferase 3', other_designations='1-acylglycerophosphoserine O-acyltransferase|LPC acyltransferase 3|O-acyltransferase domain-containing protein 5|lyso-PC acyltransferase 3|lyso-PE acyltransferase|lyso-PS acyltransferase|lysophosphatidylethanolamine acyltransferase|lysophosphatidylserine acyltransferase', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124663104, end_position_on_the_genomic_accession=124704716, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14793, CurrentID=0, Status='live', Symbol='Cdca3', Aliases='2410005A12Rik, C8, Grcc8, Tome-1', description='cell division cycle associated 3', other_designations='gene rich cluster, C8|gene-rich cluster protein C8|trigger of mitotic entry 1|trigger of mitotic entry protein 1', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124829595, end_position_on_the_genomic_accession=124833701, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14794, CurrentID=0, Status='live', Symbol='Spsb2', Aliases='AI461677, C9, Grcc9, SSB2', description='splA/ryanodine receptor domain and SOCS box containing 2', other_designations='gene-rich cluster protein C9', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124808094, end_position_on_the_genomic_accession=124810883, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14797, CurrentID=0, Status='live', Symbol='Aes', Aliases='AL024115, Esp1, Grg, Grg-5, Grg5, Tle5', description='amino-terminal enhancer of split', other_designations='groucho-related protein 5|related to Drosophila groucho', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81559444, end_position_on_the_genomic_accession=81566371, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14799, CurrentID=0, Status='live', Symbol='Gria1', Aliases='2900051M01Rik, Glr-1, Glr1, GluA1, GluR-A, GluRA, Glur-1, Glur1, HIPA1, gluR-K1', description='glutamate receptor, ionotropic, AMPA1 (alpha 1)', other_designations='AMPA-selective glutamate receptor 1|glutamate receptor A|glutamate receptor ionotropic, AMPA 1', map_location='11 34.51 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=57011571, end_position_on_the_genomic_accession=57330244, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  14800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14800, CurrentID=0, Status='live', Symbol='Gria2', Aliases='GluA2, GluR-B, Glur-2, Glur2', description='glutamate receptor, ionotropic, AMPA2 (alpha 2)', other_designations='AMPA selective glutamate receptor|AMPA-selective glutamate receptor 2|gluR-K2|glutamate receptor B|glutamate receptor ionotropic, AMPA 2', map_location='3 35.5 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=80684935, end_position_on_the_genomic_accession=80803204, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  14802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14802, CurrentID=0, Status='live', Symbol='Gria4', Aliases='GluA4, GluR-D, Glur-4, Glur4, Gluralpha4, spkw1', description='glutamate receptor, ionotropic, AMPA4 (alpha 4)', other_designations='AMPA-selective glutamate receptor 4|glutamate receptor channel alpha 4|glutamate receptor ionotropic, AMPA 4|spike wave 1', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=4417893, end_position_on_the_genomic_accession=4796247, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  14803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14803, CurrentID=0, Status='live', Symbol='Grid1', Aliases='GluD1, GluRdelta1', description='glutamate receptor, ionotropic, delta 1', other_designations='gluR delta-1 subunit|glutamate receptor delta-1 subunit', map_location='14 20.84 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34819996, end_position_on_the_genomic_accession=35583387, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14804, CurrentID=0, Status='live', Symbol='Grid2', Aliases='B230104L07Rik, GluD2, GluRdelta2, Lc, Lc<J>, MMS10-AC, Ms10ac, cpr, ho, nmf408, tpr', description='glutamate receptor, ionotropic, delta 2', other_designations='creeper|gluR delta-2 subunit|glutamate receptor delta-2 subunit|hotfoot|lurcher|minisatellite 10ac detected by probe MMS10', map_location='6 29.77 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=63256832, end_position_on_the_genomic_accession=64704322, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14805, CurrentID=0, Status='live', Symbol='Grik1', Aliases='A830007B11Rik, D16Ium24, D16Ium24e, GluK1, GluK5, Glur-5, Glur5, Glurbeta1', description='glutamate receptor, ionotropic, kainate 1', other_designations='glutamate receptor subunit 5', map_location='16 50.23 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87895897, end_position_on_the_genomic_accession=88290602, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  14806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14806, CurrentID=0, Status='live', Symbol='Grik2', Aliases='AW124492, GluK2, Glur-6, Glur6, Glurbeta2', description='glutamate receptor, ionotropic, kainate 2 (beta 2)', other_designations='gluR beta-2|glutamate receptor 6|glutamate receptor beta-2|glutamate receptor, ionotropic kainate 2', map_location='10 24.87 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=49094842, end_position_on_the_genomic_accession=49789250, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  14807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14807, CurrentID=0, Status='live', Symbol='Grik3', Aliases='9630027E11, GluK3, GluR7-3, Glur-7, Glur7', description='glutamate receptor, ionotropic, kainate 3', other_designations='glutamate receptor 7', map_location='4 58.91 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=125490709, end_position_on_the_genomic_accession=125714173, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  276952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276952, CurrentID=0, Status='live', Symbol='Rasl10b', Aliases='B230331P10Rik, Gm729, VTS58635', description='RAS-like, family 10, member B', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83409142, end_position_on_the_genomic_accession=83421039, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14809, CurrentID=0, Status='live', Symbol='Grik5', Aliases='GluK5, GluRgamma2, KA2', description='glutamate receptor, ionotropic, kainate 5 (gamma 2)', other_designations='gluR gamma-2|glutamate receptor KA-2|glutamate receptor gamma-2|glutamate receptor, ionotropic kainate 5', map_location='7 13.73 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25009849, end_position_on_the_genomic_accession=25072499, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  14810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14810, CurrentID=0, Status='live', Symbol='Grin1', Aliases='GluN1, GluRdelta1, GluRzeta1, M100174, NMD-R1, NMDAR1, NR1, Nmdar, Rgsc174', description='glutamate receptor, ionotropic, NMDA1 (zeta 1)', other_designations='N-methyl-D-aspartate receptor subunit NR1|glutamate [NMDA] receptor subunit zeta-1', map_location='2 17.14 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25291177, end_position_on_the_genomic_accession=25319198, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  14811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14811, CurrentID=0, Status='live', Symbol='Grin2a', Aliases='GluN2A, NMDAR2A, NR2A', description='glutamate receptor, ionotropic, NMDA2A (epsilon 1)', other_designations='GluRepsilon1|N-methyl D-aspartate receptor subtype 2A|glutamate [NMDA] receptor subunit epsilon-1', map_location='16 5.28 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=9573737, end_position_on_the_genomic_accession=9996500, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14812, CurrentID=0, Status='live', Symbol='Grin2b', Aliases='AW490526, GluN2B, NR2B, Nmdar2b', description='glutamate receptor, ionotropic, NMDA2B (epsilon 2)', other_designations='GluRepsilon2|N-methyl D-aspartate receptor 2B|N-methyl D-aspartate receptor subtype 2B|glutamate [NMDA] receptor subunit epsilon-2', map_location='6 66.38 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135715272, end_position_on_the_genomic_accession=136173819, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14813, CurrentID=0, Status='live', Symbol='Grin2c', Aliases='GluN2C, NMDAR2C, NR2C', description='glutamate receptor, ionotropic, NMDA2C (epsilon 3)', other_designations='GluRepsilon3|N-methyl D-aspartate receptor subtype 2C|glutamate [NMDA] receptor subunit epsilon-3', map_location='11 80.8 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115249169, end_position_on_the_genomic_accession=115267297, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14814, CurrentID=0, Status='live', Symbol='Grin2d', Aliases='GluN2D, NMDAR2D, NR2D', description='glutamate receptor, ionotropic, NMDA2D (epsilon 4)', other_designations='GluRepsilon4|N-methyl D-aspartate receptor subtype 2D|glur epsilon 4|glutamate [NMDA] receptor subunit epsilon-4|ionotropic NMDA glutamate receptor epsilon 4 subunit type 2', map_location='7 29.54 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45831883, end_position_on_the_genomic_accession=45872689, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14815, CurrentID=0, Status='live', Symbol='Nr3c1', Aliases='GR, Grl-1, Grl1', description='nuclear receptor subfamily 3, group C, member 1', other_designations='glucocorticoid receptor 1', map_location='18 21.09 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=39410545, end_position_on_the_genomic_accession=39519416, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14816, CurrentID=0, Status='live', Symbol='Grm1', Aliases='4930455H15Rik, ENSMUSG00000075319, Gm10828, Gprc1a, mGluR1, nmf373, rcw, wobl', description='glutamate receptor, metabotropic 1', other_designations='G protein coupled receptor, family C, group 1, member A|G protein-coupled receptor, family C, group 1, member A', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=10686059, end_position_on_the_genomic_accession=11082377, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14823, CurrentID=0, Status='live', Symbol='Grm8', Aliases='A230002O04, Gprc1h, mGluR8', description='glutamate receptor, metabotropic 8', other_designations='G protein coupled receptor, family C, group 1, member H', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=27275121, end_position_on_the_genomic_accession=28135837, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14824, CurrentID=0, Status='live', Symbol='Grn', Aliases='Pgrn, epithelin', description='granulin', other_designations='PC cell-derived growth factor|PCDGF|PEPI|acrogranulin|proepithelin|progranulin', map_location='11 66.29 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102430322, end_position_on_the_genomic_accession=102436809, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14825, CurrentID=0, Status='live', Symbol='Cxcl1', Aliases='Fsp, Gro1, KC, Mgsa, N51, Scyb1, gro', description='chemokine (C-X-C motif) ligand 1', other_designations='C-X-C motif chemokine 1|GRO1 oncogene|KC/GR)-alpha|KC/GRO-alpha|alpha-chemokine|platelet-derived growth factor-inducible protein KC|secretory protein N51', map_location='5 E-F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90891245, end_position_on_the_genomic_accession=90893121, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14827, CurrentID=0, Status='live', Symbol='Pdia3', Aliases='58kDa, ERp57, ERp60, ERp61, Erp, Grp58, PDI, PDI-Q2, PI-PLC, PLC[a], Plca', description='protein disulfide isomerase associated 3', other_designations='58 kDa glucose-regulated protein|58 kDa microsomal protein|ER protein 57|ER protein 60|disulfide isomerase ER-60|endoplasmic reticulum protein|endoplasmic reticulum resident protein 57|endoplasmic reticulum resident protein 60|glucose regulated protein, 58 kDa|p58|phospholipase C, alpha', map_location='2 60.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121413902, end_position_on_the_genomic_accession=121438687, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14828, CurrentID=0, Status='live', Symbol='Hspa5', Aliases='AL022860, AU019543, Bip, D2Wsu141e, D2Wsu17e, Grp78, Hsce70, SEZ-7, Sez7, baffled, mBiP', description='heat shock protein 5', other_designations='GRP-78|XAP-1 antigen|glucose regulated protein, 78 kDa|heat shock 70 kDa protein 5|heat shock 70kD protein 5 (glucose-regulated protein, 78kD)|immunoglobulin heavy chain-binding protein', map_location='2 22.94 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34772090, end_position_on_the_genomic_accession=34776529, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14829, CurrentID=0, Status='live', Symbol='Grpr', Aliases='GRP-R', description='gastrin releasing peptide receptor', other_designations='GRP-preferring bombesin receptor', map_location='X 75.63 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=163513904, end_position_on_the_genomic_accession=163549647, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14836, CurrentID=0, Status='live', Symbol='Gsc', Aliases='', description='goosecoid homeobox', other_designations='', map_location='12 54.32 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104471209, end_position_on_the_genomic_accession=104473236, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14840, CurrentID=0, Status='live', Symbol='Gsg1', Aliases='', description='germ cell-specific gene 1', other_designations='germ cell-associated protein 1', map_location='6 66.07 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135237329, end_position_on_the_genomic_accession=135254336, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14841, CurrentID=0, Status='live', Symbol='Gsg2', Aliases='Haspin', description='germ cell-specific gene 2', other_designations='germ cell-specific gene 2 protein|haploid germ cell-specific nuclear protein kinase', map_location='11 45.25 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73135484, end_position_on_the_genomic_accession=73138294, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14842, CurrentID=0, Status='live', Symbol='Gsx1', Aliases='Gsh-1, Gsh1', description='GS homeobox 1', other_designations='genomic screened homeo box 1|homeobox protein GSH-1', map_location='5 86.79 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147188696, end_position_on_the_genomic_accession=147190947, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  14843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14843, CurrentID=0, Status='live', Symbol='Gsx2', Aliases='Gsh-2, Gsh2', description='GS homeobox 2', other_designations='genetic-screened homeobox 2|genomic screened homeo box 2|homeobox protein GSH-2', map_location='5 39.55 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=75075601, end_position_on_the_genomic_accession=75077893, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  408062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408062, CurrentID=0, Status='live', Symbol='Zfp873', Aliases='', description='zinc finger protein 873', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82048086, end_position_on_the_genomic_accession=82077727, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  408065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408065, CurrentID=0, Status='live', Symbol='Zfp456', Aliases='BC065418, Rslcan-13', description='zinc finger protein 456', other_designations='', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67363584, end_position_on_the_genomic_accession=67375763, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  408067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408067, CurrentID=0, Status='live', Symbol='Zfp874b', Aliases='9630025I21Rik, Rslcan16', description='zinc finger protein 874b', other_designations='regulator of sex-limitation candidate 16', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67471513, end_position_on_the_genomic_accession=67486236, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14852, CurrentID=0, Status='live', Symbol='Gspt1', Aliases='AI314175, AI326449, AV307676, C79774, G1st, Gst-1', description='G1 to S phase transition 1', other_designations='G1 to S phase transition protein 1 homolog|G1 to phase transition 1|G1-to-S transition GTP binding protein|eRF3a|eukaryotic peptide chain release factor subunit 3a|eukaryotic polypeptide chain release factor 3|guanine nucleotide regulatory protein', map_location='16 6.38 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11216240, end_position_on_the_genomic_accession=11254325, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  14853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14853, CurrentID=0, Status='live', Symbol='Gspt2', Aliases='', description='G1 to S phase transition 2', other_designations='G1 to S phase transition protein 2 homolog|eRF3b|eukaryotic peptide chain release factor subunit 3b', map_location='X B1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94636069, end_position_on_the_genomic_accession=94638562, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14854, CurrentID=0, Status='live', Symbol='Gss', Aliases='AI314904, GS-A/GS-B, GSH-S', description='glutathione synthetase', other_designations='GSH synthetase|glutathione synthase', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155563181, end_position_on_the_genomic_accession=155592810, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  14857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14857, CurrentID=0, Status='live', Symbol='Gsta1', Aliases='Gst2-1, OTTMUSG00000031890', description='glutathione S-transferase, alpha 1 (Ya)', other_designations='GST class-alpha member 1|glutathione S-transferase Ya|glutathione S-transferase Ya1', map_location='9 43.65 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78230669, end_position_on_the_genomic_accession=78242683, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  14858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14858, CurrentID=0, Status='live', Symbol='Gsta2', Aliases='Gst2-2, Gstc-2, Gstc2', description='glutathione S-transferase, alpha 2 (Yc2)', other_designations='GST class-alpha member 2|glutathione S-transferase GT41A', map_location='9 43.65 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78331018, end_position_on_the_genomic_accession=78347192, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  14859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14859, CurrentID=0, Status='live', Symbol='Gsta3', Aliases='Gst2-3', description='glutathione S-transferase, alpha 3', other_designations='GST class-alpha member 3|glutathione S-transferase Ya3|glutathione S-transferase Yc', map_location='1 6.5 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=21240585, end_position_on_the_genomic_accession=21265575, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14860, CurrentID=0, Status='live', Symbol='Gsta4', Aliases='GST5.7, mGsta4', description='glutathione S-transferase, alpha 4', other_designations='GST A4-4|GST class-alpha member 4|GSTA4-4|glutathione S-transferase 5.7', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78191966, end_position_on_the_genomic_accession=78209349, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  170711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170711, CurrentID=0, Status='live', Symbol='Otud7a', Aliases='AJ430384, Otud7', description='OTU domain containing 7A', other_designations='Cezanne 2 protein|zinc finger protein Cezanne 2', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=63444745, end_position_on_the_genomic_accession=63759905, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14863, CurrentID=0, Status='live', Symbol='Gstm2', Aliases='Gstb-2, Gstb2', description='glutathione S-transferase, mu 2', other_designations='GST 5-5|GST class-mu 2|glutathione S-transferase pmGT2|glutathione-S-transferase, mu 2', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107981702, end_position_on_the_genomic_accession=107986436, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14864, CurrentID=0, Status='live', Symbol='Gstm3', Aliases='Fsc2, mGSTM5', description='glutathione S-transferase, mu 3', other_designations='GST class-mu 3|glutathione S-transferase GT9.3', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107963696, end_position_on_the_genomic_accession=107969174, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  14865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14865, CurrentID=0, Status='live', Symbol='Gstm4', Aliases='1110004G14Rik, GSTM7-7, Gstb-4, Gstb4', description='glutathione S-transferase, mu 4', other_designations='glutathione-S-transferase, mu 4', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108040408, end_position_on_the_genomic_accession=108045080, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14866, CurrentID=0, Status='live', Symbol='Gstm5', Aliases='', description='glutathione S-transferase, mu 5', other_designations='GST class-mu 5|fibrous sheath component 2|fsc2|glutathione-S-transferase, mu 5', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107895854, end_position_on_the_genomic_accession=107898685, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14867, CurrentID=0, Status='live', Symbol='Gstm6', Aliases='', description='glutathione S-transferase, mu 6', other_designations='GST class-mu 6|glutathione-S-transferase class M5|glutathione-S-transferase, mu 6', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107938847, end_position_on_the_genomic_accession=107953491, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14869, CurrentID=0, Status='live', Symbol='Gstp2', Aliases='GSTpiA, Gst-3, Gst3', description='glutathione S-transferase, pi 2', other_designations='GST YF-YF|GST class-pi|GST-piA|Gst p-2|gst P2', map_location='19 3.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4040288, end_position_on_the_genomic_accession=4042221, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14870, CurrentID=0, Status='live', Symbol='Gstp1', Aliases='GstpiB', description='glutathione S-transferase, pi 1', other_designations='GST YF-YF|GST class-pi|GST-piB|Gst p-1|gst P1|preadipocyte growth factor', map_location='19 3.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4035411, end_position_on_the_genomic_accession=4037912, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14871, CurrentID=0, Status='live', Symbol='Gstt1', Aliases='AI255817-1, Gstt1', description='glutathione S-transferase, theta 1', other_designations='GST class-theta-1', map_location='10 B5-C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75783813, end_position_on_the_genomic_accession=75798584, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14872, CurrentID=0, Status='live', Symbol='Gstt2', Aliases='AI266894, Yrs, mGSTT2', description='glutathione S-transferase, theta 2', other_designations='GST class-theta-2', map_location='10 B5-C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75831842, end_position_on_the_genomic_accession=75837419, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14873, CurrentID=0, Status='live', Symbol='Gsto1', Aliases='AA407097, AI194287, AU018802, Gsto-1, Gstx, Spg-r, p28', description='glutathione S-transferase omega 1', other_designations='Glutathione transferase omega 1 (GSTO 1-1) (p28)|MMA(V) reductase|S-(Phenacyl)glutathione reductase|glutathione S-transferase omega 1-1|glutathione-S-transferase like|glutathione-dependent dehydroascorbate reductase|monomethylarsonic acid reductase', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47854989, end_position_on_the_genomic_accession=47864790, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14874, CurrentID=0, Status='live', Symbol='Gstz1', Aliases='MAAI', description='glutathione transferase zeta 1 (maleylacetoacetate isomerase)', other_designations='GSTZ1-1|glutathione S-transferase zeta 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87147165, end_position_on_the_genomic_accession=87164723, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14875, CurrentID=0, Status='live', Symbol='gt', Aliases='', description='gray tremor', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14884, CurrentID=0, Status='live', Symbol='Gtf2h1', Aliases='62kDa, AW743425, AW822074, BTF2 p62, C77871, p62', description='general transcription factor II H, polypeptide 1', other_designations='BTF2-p62|TFIIH basal transcription factor complex p62 subunit|basic transcription factor 2 62 kDa subunit|general transcription factor IIH, polypeptide 1 (62kD subunit)', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46795812, end_position_on_the_genomic_accession=46823800, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  14885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14885, CurrentID=0, Status='live', Symbol='Gtf2h4', Aliases='AW545633, BTF2 p52, p52', description='general transcription factor II H, polypeptide 4', other_designations='TFIIH basal transcription factor complex p52 subunit|basic transcription factor 2 52 kDa subunit', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35667728, end_position_on_the_genomic_accession=35673750, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  14886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14886, CurrentID=0, Status='live', Symbol='Gtf2i', Aliases='6030441I21Rik, BAP-135, Diws1t, GtfII-I, Spin, TFII-I, WBSCR6', description='general transcription factor II I', other_designations='BTK-associated protein 135|Bruton tyrosine kinase-associated protein 135|SRF-Phox1-interacting protein|Williams-Beuren syndrome chromosomal region 6 protein', map_location='5 74.48 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134237832, end_position_on_the_genomic_accession=134314760, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  14894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14894, CurrentID=0, Status='live', Symbol='Cfap20', Aliases='2600014O15Rik, AL024127, Gtl3, T10-2A2', description='cilia and flagella associated protein 20', other_designations='UPF0468 protein C16orf80 homolog|gene trap locus 3 protein', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95420249, end_position_on_the_genomic_accession=95434869, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14897, CurrentID=0, Status='live', Symbol='Trip12', Aliases='1110036I07Rik, 6720416K24Rik, AA410158, Gtl6, TRIP-12', description='thyroid hormone receptor interactor 12', other_designations='TR-interacting protein 12|gene trap locus 6|thyroid receptor interactor TRIP12|thyroid receptor-interacting protein 12', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=84721189, end_position_on_the_genomic_accession=84850008, orientation='minus', exon_count=50, OMIM='', no_hdr0=''),
  14904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14904, CurrentID=0, Status='live', Symbol='Gtpbp1', Aliases='AL022987, AW045683, C85212, GP-1, Gtpbp', description='GTP binding protein 1', other_designations='G-protein 1|GP1', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79690896, end_position_on_the_genomic_accession=79721479, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14910, CurrentID=0, Status='live', Symbol='Gt(ROSA)26Sor', Aliases='AV258896, Gtrgeo26, Gtrosa26, R26, ROSA26', description='gene trap ROSA 26, Philippe Soriano', other_designations='beta geo', map_location='6 52.73 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113067428, end_position_on_the_genomic_accession=113077244, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14911, CurrentID=0, Status='live', Symbol='Thumpd3', Aliases='AW556087, Gt(ROSA)26asSor, Gtrosa26as', description='THUMP domain containing 3', other_designations='gene trap ROSA 26 antisense, Philippe Soriano|gtROSA26asSor', map_location='6 52.73 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113046252, end_position_on_the_genomic_accession=113068278, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  14912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14912, CurrentID=0, Status='live', Symbol='Nkx6-2', Aliases='Gtx, Nkx6.2', description='NK6 homeobox 2', other_designations='Drosophila NK transcription factor related, gene family 6, locus 2|NK6 transcription factor related, locus 2|glial and testis specific homeobox', map_location='7 84.57 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139579376, end_position_on_the_genomic_accession=139582797, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  14913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14913, CurrentID=0, Status='live', Symbol='Guca1a', Aliases='GC-A, Gcap1, Guca1, mGCAP1', description='guanylate cyclase activator 1a (retina)', other_designations='GCAP 1', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47394558, end_position_on_the_genomic_accession=47413514, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14915, CurrentID=0, Status='live', Symbol='Guca2a', Aliases='Guca2', description='guanylate cyclase activator 2a (guanylin)', other_designations='guanylate cyclase activator 2 (guanylin 2, intestinal, heatstable)', map_location='4 55.47 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119637732, end_position_on_the_genomic_accession=119639465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  14916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14916, CurrentID=0, Status='live', Symbol='Guca2b', Aliases='AV066530, Gcap2, Ugn', description='guanylate cyclase activator 2b (retina)', other_designations='granule cell antiserum positive 2|uroguanylin', map_location='4 55.48 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119656601, end_position_on_the_genomic_accession=119658945, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14917, CurrentID=0, Status='live', Symbol='Gucy2c', Aliases='AI893437, GC-C, Gcc', description='guanylate cyclase 2c', other_designations='STA receptor|guanylyl cyclase C|heat stable enterotoxin receptor|intestinal guanylate cyclase', map_location='6 66.67 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136697284, end_position_on_the_genomic_accession=136781742, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  14918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14918, CurrentID=0, Status='live', Symbol='Gucy2d', Aliases='', description='guanylate cyclase 2d', other_designations='guanylyl cyclase 2d|guanylyl cyclase D', map_location='7 53.86 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98440416, end_position_on_the_genomic_accession=98477572, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  14919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14919, CurrentID=0, Status='live', Symbol='Gucy2e', Aliases='GC-E, GC1, ROS-GC1', description='guanylate cyclase 2e', other_designations='guanylyl cyclase 2e', map_location='11 42.51 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69218117, end_position_on_the_genomic_accession=69237358, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  14923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14923, CurrentID=0, Status='live', Symbol='Guk1', Aliases='AA409738, AL033299, AL033300, AV026611, Gmk, Guk-1', description='guanylate kinase 1', other_designations='GMP kinase', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59183855, end_position_on_the_genomic_accession=59192207, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  14924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14924, CurrentID=0, Status='live', Symbol='Magi1', Aliases='AIP3, BAP1, Baiap1, Gukmi1, MAGI1c, Magi-1, TNRC19, WWP3, mKIAA4129', description='membrane associated guanylate kinase, WW and PDZ domain containing 1', other_designations='BAI1-associated protein 1|membrane-associated guanylate kinase inverted 1', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=93675453, end_position_on_the_genomic_accession=94283920, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  14933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14933, CurrentID=0, Status='live', Symbol='Gk', Aliases='D930012N15Rik, Gyk', description='glycerol kinase', other_designations='ATP:glycerol 3-phosphotransferase|glycerokinase', map_location='X C-D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=85701937, end_position_on_the_genomic_accession=85776819, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  14934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14934, CurrentID=0, Status='live', Symbol='Gypa', Aliases='AI853584, CD235a, GPA', description='glycophorin A', other_designations='', map_location='8 38.27 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=80493981, end_position_on_the_genomic_accession=80510785, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  14936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14936, CurrentID=0, Status='live', Symbol='Gys1', Aliases='Gys3, MGS', description='glycogen synthase 1, muscle', other_designations='glycogen synthase 3, brain', map_location='7 29.31 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45434839, end_position_on_the_genomic_accession=45456617, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  14938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14938, CurrentID=0, Status='live', Symbol='Gzma', Aliases='AW494114, Ctla-3, Ctla3, Hf, SE1, TSP-1, TSP1', description='granzyme A', other_designations='BLT esterase|H factor|Hanukah factor|autocrine thymic lymphoma granzyme-like serine protease|fragmentin-1|serine esterase 1|t cell-specific serine protease 1', map_location='13 D', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113093827, end_position_on_the_genomic_accession=113100981, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14939, CurrentID=0, Status='live', Symbol='Gzmb', Aliases='AI553453, CCP-1/C11, CCP1, Ctla-1, Ctla1, GZB', description='granzyme B', other_designations='cytotoxic cell protease 1|fragmentin-2', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56258856, end_position_on_the_genomic_accession=56262260, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14940, CurrentID=0, Status='live', Symbol='Gzmc', Aliases='AI323531, B10, CCP2, Ctla-5, Ctla5', description='granzyme C', other_designations='cytotoxic cell protease 2', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56231397, end_position_on_the_genomic_accession=56241639, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14941, CurrentID=0, Status='live', Symbol='Gzmd', Aliases='CCP5, Ctla-5, Ctla5', description='granzyme D', other_designations='', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56129556, end_position_on_the_genomic_accession=56132603, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14942, CurrentID=0, Status='live', Symbol='Gzme', Aliases='CCP3, Ctla-6, Ctla6, MCSP-2', description='granzyme E', other_designations='CTL serine protease 2|D12|MCSP2|cytotoxic cell protease 3|cytotoxic serine protease 2', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56117619, end_position_on_the_genomic_accession=56120625, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14943, CurrentID=0, Status='live', Symbol='Gzmf', Aliases='CCP4, Ctla-7, Ctla7, MCSP-3', description='granzyme F', other_designations='AKA granzyme G|C134|CTL serine protease 3|MCSP3|cytotoxic cell protease 4|cytotoxic serine protease 3|granzyme G', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56205263, end_position_on_the_genomic_accession=56211407, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14944, CurrentID=0, Status='live', Symbol='Gzmg', Aliases='Ctla-7, Ctla7', description='granzyme G', other_designations='AKA granzyme G|CTL serine protease 1|CTL serine protease 3|MCSP-1', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56156581, end_position_on_the_genomic_accession=56159579, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14945, CurrentID=0, Status='live', Symbol='Gzmk', Aliases='', description='granzyme K', other_designations='', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113171874, end_position_on_the_genomic_accession=113180897, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68027, CurrentID=0, Status='live', Symbol='Tmem178', Aliases='2810417M05Rika, Tmem178', description='transmembrane protein 178', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80944632, end_position_on_the_genomic_accession=81001816, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  14950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14950, CurrentID=0, Status='live', Symbol='H13', Aliases='1200006O09Rik, 4930443L17Rik, 5031424B04Rik, AV020344, H-13, Hm13, PSL3, Spp', description='histocompatibility 13', other_designations='presenilin-like aspartyl protease|presenilin-like protein 3|signal peptide peptidase beta', map_location='2 75.41 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152669437, end_position_on_the_genomic_accession=152708668, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  14955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14955, CurrentID=0, Status='live', Symbol='H19', Aliases='AI747191', description='H19, imprinted maternally expressed transcript', other_designations='H19 fetal liver mRNA', map_location='7 87.97 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142575532, end_position_on_the_genomic_accession=142578146, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14957, CurrentID=0, Status='live', Symbol='Hist1h1d', Aliases='H1.3, H1D, H1f3, H1s-4', description='histone cluster 1, H1d', other_designations='H1 VAR.4|H1 histone family, member 3|histone 1, H1d|histone 1, family 3', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23555032, end_position_on_the_genomic_accession=23555807, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14958, CurrentID=0, Status='live', Symbol='H1f0', Aliases='D130017D06Rik, H1(0), H1fv', description='H1 histone family, member 0', other_designations="H1-0 histone|histone H1'|histone H1(0)|myD196", map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79028212, end_position_on_the_genomic_accession=79030504, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  14960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14960, CurrentID=0, Status='live', Symbol='H2-Aa', Aliases='Aalpha, H-2Aa, H2Aa, I-Aalpha, IAalpha, Ia-1, Ia1', description='histocompatibility 2, class II antigen A, alpha', other_designations='A alpha|H-2 class II histocompatibility antigen, A-B alpha chain|H-2 class II histocompatibility antigen, A-F alpha chain|H-2 class II histocompatibility antigen, A-K alpha chain|H-2 class II histocompatibility antigen, A-Q alpha chain|H-2 class II histocompatibility antigen, A-R alpha chain|H-2 class II histocompatibility antigen, A-S alpha chain|H-2 class II histocompatibility antigen, A-U alpha chain|MHC class II antigen A alpha isoform-004|MHC class II antigen alpha chain|MHC-II I-A(d) alpha|major histocompatibility protein class II alpha chain', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34282751, end_position_on_the_genomic_accession=34287771, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14961, CurrentID=0, Status='live', Symbol='H2-Ab1', Aliases='AI845868, Abeta, H-2Ab, H2-Ab, I-Abeta, IAb, Ia-2, Ia2, Rmcs1', description='histocompatibility 2, class II antigen A, beta 1', other_designations='H-2 class II histocompatibility antigen, A beta chain|MHC class II H2-IA-beta-psi|MHC class II antigen A beta|major histocompatibility complex class II beta chain|response to metastatic cancers 1', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34263227, end_position_on_the_genomic_accession=34269418, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14962, CurrentID=0, Status='live', Symbol='Cfb', Aliases='AI195813, AI255840, B, Bf, C2, Fb, H2-Bf', description='complement factor B', other_designations='C3/C5 convertase|Factor B|alternative-complement pathway C3/C5 convertase|complement component 2 (within H-2S)|histocompatibility 2, complement component factor B|properdin', map_location='17 18.41 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34856374, end_position_on_the_genomic_accession=34862514, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  14963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14963, CurrentID=0, Status='live', Symbol='H2-Bl', Aliases='H2-B1-like, H2-Bl', description='histocompatibility 2, blastocyst', other_designations='H2 blastocyst|MHC class I antigen splice variant Bl.1|blastocyst MHC', map_location='17 18.89 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36080189, end_position_on_the_genomic_accession=36084249, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14964, CurrentID=0, Status='live', Symbol='H2-D1', Aliases='H-2D, H2-D', description='histocompatibility 2, D region locus 1', other_designations='H-2D cell surface glycoprotein|H2-DC1-beta transplantation antigen protein|MHC H2-D-q alpha-chain|MHC class I H-2Dp|MHC class I H2 antigen|MHC class I antigen|Q5k protein', map_location='17 18.61 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35263094, end_position_on_the_genomic_accession=35267497, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14967, CurrentID=0, Status='live', Symbol='H2-D4', Aliases='H-2D4', description='histocompatibility 2, D region locus 4', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14969, CurrentID=0, Status='live', Symbol='H2-Eb1', Aliases='Eb, H-2Eb, H2Eb, Ia-4, Ia4', description='histocompatibility 2, class II antigen E beta', other_designations='I-E beta MHC class II|MHC H2-IE-beta cell surface glycoprotein|MHC class II IE-b|MHC class II antigen beta chain|cell surface glycoprotein|major histocompatibility complex H-2E beta chain', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34305867, end_position_on_the_genomic_accession=34316674, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14972, CurrentID=0, Status='live', Symbol='H2-K1', Aliases='H-2K, H-2K(d), H2-K, K-f', description='histocompatibility 2, K1, K region', other_designations='H-2 class I histocompatibility antigen, K-B alpha chain|H-2 class I histocompatibility antigen, K-D alpha chain|H-2 class I histocompatibility antigen, K-K alpha chain|H-2 class I histocompatibility antigen, K-Q alpha chain|H-2K(B)|H-2K(K)|H-2K(Q)|MHC H2-K transplantation antigen|MHC class I H2-K-b-alpha-2 cell surface glycoprotein|MHC class I antigen|MHC class I heavy chain H2-K|class I major histocompatibility antigen H-2Kd', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33996012, end_position_on_the_genomic_accession=34000342, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  408190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408190, CurrentID=0, Status='live', Symbol='Wfdc13', Aliases='Wfdc13l1', description='WAP four-disulfide core domain 13', other_designations='WAP four-disulfide core domain 13-like 1', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164685107, end_position_on_the_genomic_accession=164687706, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  408191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408191, CurrentID=0, Status='live', Symbol='Gm5415', Aliases='EG408191, Senp14', description='predicted gene 5415', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=32543546, end_position_on_the_genomic_accession=32547293, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  14976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14976, CurrentID=0, Status='live', Symbol='Pfdn6', Aliases='H-2Ke2, H2-Ke2, Ke-2', description='prefoldin subunit 6', other_designations='H2-K region expressed gene 2|protein Ke2', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33938909, end_position_on_the_genomic_accession=33940343, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14977, CurrentID=0, Status='live', Symbol='Slc39a7', Aliases='AA408174, AI117660, AL024048, H-2Ke4, H2-Ke4, Ke-4, Ke4, Ring5, Zip7', description='solute carrier family 39 (zinc transporter), member 7', other_designations='H2-K region expressed gene 4|histidine-rich membrane protein Ke4|solute carrier family 39 member 7|zrt-, Irt-like protein 7', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34028266, end_position_on_the_genomic_accession=34031690, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  14979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14979, CurrentID=0, Status='live', Symbol='H2-Ke6', Aliases='D17H6S112E, H-2Ke6, Hsd17b8, Ke-6, Ke6, Ring2', description='H2-K region expressed gene 6', other_designations='17-beta-HSD 8|17-beta-hydroxysteroid dehydrogenase 8|3-oxoacyl-[acyl-carrier-protein] reductase|estrogen 17-oxidoreductase|testosterone 17-beta-dehydrogenase 8', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34026033, end_position_on_the_genomic_accession=34028055, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  408196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408196, CurrentID=0, Status='live', Symbol='Gm5416', Aliases='EG408196, Stfna1l1', description='predicted gene 5416', other_designations='stefin-A1-like 1', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36210403, end_position_on_the_genomic_accession=36217788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68033, CurrentID=0, Status='live', Symbol='Cox19', Aliases='1810074D02Rik, 2810437L13Rik', description='cytochrome c oxidase assembly protein 19', other_designations='COX19 cytochrome c oxidase assembly homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139337822, end_position_on_the_genomic_accession=139345166, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  14985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14985, CurrentID=0, Status='live', Symbol='H2-M10.1', Aliases='9.5H', description='histocompatibility 2, M region locus 10.1', other_designations='', map_location='17 19.01 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36322860, end_position_on_the_genomic_accession=36326150, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14990, CurrentID=0, Status='live', Symbol='H2-M2', Aliases='H-2M2, Thy19.4', description='histocompatibility 2, M region locus 2', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37480851, end_position_on_the_genomic_accession=37483624, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14991, CurrentID=0, Status='live', Symbol='H2-M3', Aliases='H-2M3, Hmt, M3a, MuprM3, MuprM301, MuprM302, R4B2', description='histocompatibility 2, M region locus 3', other_designations='MHC class I-b antigen M3', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37270234, end_position_on_the_genomic_accession=37274485, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  14992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14992, CurrentID=0, Status='live', Symbol='H2-M4-ps', Aliases='H-2M4, H2-M4', description='histocompatibility 2, M region locus 4, pseudogene', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36980633, end_position_on_the_genomic_accession=36980914, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  14994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14994, CurrentID=0, Status='live', Symbol='H2-M6-ps', Aliases='H-2M6, H2-M6', description='histocompatibility 2, M region locus 6, pseudogene', other_designations='histocompatibility 2, D region locus 1 pseudogene', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36969207, end_position_on_the_genomic_accession=36971396, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  14995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14995, CurrentID=0, Status='live', Symbol='H2-M7-ps', Aliases='H-2M7, H2-M7', description='histocompatibility 2, M region locus 7, pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  14997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14997, CurrentID=0, Status='live', Symbol='H2-M9', Aliases='M9', description='histocompatibility 2, M region locus 9', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36639351, end_position_on_the_genomic_accession=36642661, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  14998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14998, CurrentID=0, Status='live', Symbol='H2-DMa', Aliases='H-2Ma, H2-Ma', description='histocompatibility 2, class II, locus DMa', other_designations='H2-M alpha|MHC class II antigen DMa|histocompatibility 2, class II, locus Ma', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34122718, end_position_on_the_genomic_accession=34139101, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  14999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14999, CurrentID=0, Status='live', Symbol='H2-DMb1', Aliases='AI385589, H-2Mb1, H2-Mb1', description='histocompatibility 2, class II, locus Mb1', other_designations='H2-M beta 1 chain|H2-M beta1|MHC class II antigen Mb1', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34153191, end_position_on_the_genomic_accession=34160229, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15000, CurrentID=0, Status='live', Symbol='H2-DMb2', Aliases='H-2Mb2, H2-Mb2', description='histocompatibility 2, class II, locus Mb2', other_designations='H2-M beta2|MHC class II antigen Mb2', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34145416, end_position_on_the_genomic_accession=34151095, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15001, CurrentID=0, Status='live', Symbol='H2-Oa', Aliases='H-2Oa', description='histocompatibility 2, O region alpha locus', other_designations='MHC class II antigen O region alpha locus', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34092379, end_position_on_the_genomic_accession=34095234, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15002, CurrentID=0, Status='live', Symbol='H2-Ob', Aliases='A-beta-2, A-beta2, H-2Ab, H-2I, H-2Ob, H2-Ab, H2-Ab2, H2-IAb2, Ob', description='histocompatibility 2, O region beta locus', other_designations='H2-IA-beta-2 antigen|MHC class II antigen O region beta locus isoform-003', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34238895, end_position_on_the_genomic_accession=34245908, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15004, CurrentID=0, Status='live', Symbol='H2-Pb', Aliases='H-2Ab, H-2I, H-2Pb, H2-Ab, H2-Ab3, Ia7, Pb', description='histocompatibility 2, P region beta locus', other_designations='A beta3', map_location='17 18.51 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15006, CurrentID=0, Status='live', Symbol='H2-Q1', Aliases='H-2Q1, Q1, Q1b, Q1d, Q1k, Qa-1, Qa1, Qed-1', description='histocompatibility 2, Q region locus 1', other_designations='', map_location='17 18.63 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35320516, end_position_on_the_genomic_accession=35327699, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  15007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15007, CurrentID=0, Status='live', Symbol='H2-Q10', Aliases='H-2Q10, Q10, Qa10', description='histocompatibility 2, Q region locus 10', other_designations='MHC class I transplantation antigen', map_location='17 18.68 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35470067, end_position_on_the_genomic_accession=35474563, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  277154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277154, CurrentID=0, Status='live', Symbol='Nynrin', Aliases='AA409373, C80530, mKIAA1305', description='NYN domain and retroviral integrase containing', other_designations='NYN domain and retroviral integrase catalytic domain-containing protein|pol-like protein', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55854115, end_position_on_the_genomic_accession=55874736, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  15013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15013, CurrentID=0, Status='live', Symbol='H2-Q2', Aliases='BE136769, ENSMUSG00000079507, Gm11132, H-2Q2, gs14-2', description='histocompatibility 2, Q region locus 2', other_designations='MHC classIb like protein GS14-2|class Ib major histocompatibility complex', map_location='17 18.64 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35342120, end_position_on_the_genomic_accession=35346764, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  15015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15015, CurrentID=0, Status='live', Symbol='H2-Q4', Aliases='H-2Q4, H2-Gs10, Qa-4, Qa4, Qat-4, Qb-1, Qb1', description='histocompatibility 2, Q region locus 4', other_designations='MHC class I like protein GS10', map_location='17 18.65 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35379555, end_position_on_the_genomic_accession=35384674, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  15016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15016, CurrentID=0, Status='live', Symbol='H2-Q5', Aliases='H-2Q5, Qa-5, Qa5, Qat-5', description='histocompatibility 2, Q region locus 5', other_designations='MHC class I antigen splice variant a1-a3-TM', map_location='17 18.65 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35394099, end_position_on_the_genomic_accession=35395630, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15018, CurrentID=0, Status='live', Symbol='H2-Q7', Aliases='H-2Q7, Ped, Q9, Qa-2, Qa-7, Qa7', description='histocompatibility 2, Q region locus 7', other_designations='QA-2 antigen|class Ib MHC antigen Qa-2', map_location='17 18.67 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35439147, end_position_on_the_genomic_accession=35443773, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15019, CurrentID=0, Status='live', Symbol='H2-Q8', Aliases='H-2Q8, MMS10-T, Ms10t, Qa-2, Qa-8, Qa8', description='histocompatibility 2, Q region locus 8', other_designations='minisatellite 10t detected by probe MMS10', map_location='17', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15023, CurrentID=0, Status='live', Symbol='H2-T1', Aliases='H-2T1', description='histocompatibility 2, T region locus 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15024, CurrentID=0, Status='live', Symbol='H2-T10', Aliases='H-2T10', description='histocompatibility 2, T region locus 10', other_designations='', map_location='17 18.93 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36115871, end_position_on_the_genomic_accession=36121444, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  15027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15027, CurrentID=0, Status='live', Symbol='H2-T13', Aliases='H-2T13', description='histocompatibility 2, T region locus 13', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15033, CurrentID=0, Status='live', Symbol='H2-T18', Aliases='H-2T18, MumuTL, TLAg, Tla', description='histocompatibility 2, T region locus 18', other_designations='MHC class I antigen|TL Ag', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242805, CurrentID=0, Status='live', Symbol='Ankrd65', Aliases='E230028L10Rik, Gm439, Gm45', description='ankyrin repeat domain 65', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155790358, end_position_on_the_genomic_accession=155793328, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15040, CurrentID=0, Status='live', Symbol='H2-T23', Aliases='37b, 37c, H-2T23, H2-Qa1, Qa-1, Qa-1(b), Qa1, Qed-1, T18c, T18c(37), T23b, T23d', description='histocompatibility 2, T region locus 23', other_designations='MHC class I antigen Qa1|MHC classIb Qa1', map_location='17 18.86 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36029765, end_position_on_the_genomic_accession=36032850, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15042, CurrentID=0, Status='live', Symbol='H2-T24', Aliases='H-2T24', description='histocompatibility 2, T region locus 24', other_designations='', map_location='17 18.84 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36003992, end_position_on_the_genomic_accession=36020842, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15043, CurrentID=0, Status='live', Symbol='H2-T3', Aliases='H-2T3, H2-Tw3, MumuTL, TL', description='histocompatibility 2, T region locus 3', other_designations='MHC c5/g1L protein|MHC class I antigen|MHC thymus leukemia antigen|MHC-TLA-T13a|TL antigen|thymus leukemia antigen', map_location='17 18.99 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36185571, end_position_on_the_genomic_accession=36190151, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  330188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330188, CurrentID=0, Status='live', Symbol='Ccdc63', Aliases='4921511C16Rik', description='coiled-coil domain containing 63', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122108052, end_position_on_the_genomic_accession=122138080, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  15051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15051, CurrentID=0, Status='live', Symbol='H2-T9', Aliases='H-2T9', description='histocompatibility 2, T region locus 9', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36109227, end_position_on_the_genomic_accession=36111670, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  330189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330189, CurrentID=0, Status='live', Symbol='Tmem120b', Aliases='', description='transmembrane protein 120B', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123076265, end_position_on_the_genomic_accession=123117445, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  15061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15061, CurrentID=0, Status='live', Symbol='Ifi44l', Aliases='H-28, H28, H28-1, NS1178', description='interferon-induced protein 44 like', other_designations='H28c minor histocompatibility antigen 1|histocompatibility 28', map_location='3 76.94 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=151758737, end_position_on_the_genomic_accession=151762891, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15064, CurrentID=0, Status='live', Symbol='Mr1', Aliases='H2ls', description='major histocompatibility complex, class I-related', other_designations='MCH class I|MHC class I-related gene protein|MHC class-I related-gene protein|class I histocompatibility antigen-like protein|histocompatibility-2 complex class 1-like sequence', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155127872, end_position_on_the_genomic_accession=155146813, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68048, CurrentID=0, Status='live', Symbol='Aen', Aliases='2700083B06Rik, Isg20l1', description='apoptosis enhancing nuclease', other_designations='interferon stimulated exonuclease gene 20-like 1|interferon-stimulated 20 kDa exonuclease-like 1', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=78895927, end_position_on_the_genomic_accession=78908833, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  15077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15077, CurrentID=0, Status='live', Symbol='Hist2h3c1', Aliases='BE691662, H3-615, H3f2, Hist2h3ca1', description='histone cluster 2, H3c1', other_designations='H3 histone, family 2|histone 2, H3c1|histone 2, H3ca1', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96246685, end_position_on_the_genomic_accession=96247348, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15078, CurrentID=0, Status='live', Symbol='H3f3a', Aliases='H3.3A', description='H3 histone, family 3A', other_designations='', map_location='1 D2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180802568, end_position_on_the_genomic_accession=180813603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15081, CurrentID=0, Status='live', Symbol='H3f3b', Aliases='9430068D06Rik, H3.3B', description='H3 histone, family 3B', other_designations='H3L-like histone', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116021961, end_position_on_the_genomic_accession=116025450, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101202683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101202683, CurrentID=0, Status='live', Symbol='Dreh', Aliases='', description='down-regulated in hepatocellular carcinoma', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=64767111, end_position_on_the_genomic_accession=64767931, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15101, CurrentID=0, Status='live', Symbol='H60a', Aliases='H60', description='histocompatibility 60a', other_designations='histocompatibility 60', map_location='10', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  277250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277250, CurrentID=0, Status='live', Symbol='Kdm3b', Aliases='5830462I21Rik, JHDM2B, Jmjd1b, mKIAA1082', description='KDM3B lysine (K)-specific demethylase 3B', other_designations='jmjC domain-containing histone demethylation protein 2B|jumonji domain containing 1B|jumonji domain-containing protein 1B', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34776243, end_position_on_the_genomic_accession=34839370, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  15107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15107, CurrentID=0, Status='live', Symbol='Hadh', Aliases='AA409008, AU019341, AW742602, HCDHsc, Schad, Hadh', description='hydroxyacyl-Coenzyme A dehydrogenase', other_designations='L-3-hydroxyacyl-Coenzyme A dehydrogenase, short chain|hydroxylacyl-Coenzyme A dehydrogenase-dehydrogenase|medium and short chain L-3-hydroxyacyl-coenzyme A dehydrogenase|short chain 3-hydroxyacyl-CoA dehydrogenase|short-chain 3-hydroxyacyl-CoA dehydrogenase', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=131233419, end_position_on_the_genomic_accession=131272101, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15108, CurrentID=0, Status='live', Symbol='Hsd17b10', Aliases='17bHSD10, Ads9, ERAB, Hadh2', description='hydroxysteroid (17-beta) dehydrogenase 10', other_designations='17-beta-HSD 10|17-beta-hydroxysteroid dehydrogenase 10|3-hydroxy-2-methylbutyryl-CoA dehydrogenase|3-hydroxyacyl-CoA dehydrogenase type II|endoplasmic reticulum-associated amyloid beta-peptide-binding protein|hydroxyacyl-Coenzyme A dehydrogenase type II|hydroxyacyl-Coenzyme A dehydrogenase, type II|mitochondrial RNase P protein 2|mitochondrial ribonuclease P protein 2|type II HADH', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152001896, end_position_on_the_genomic_accession=152004442, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15109, CurrentID=0, Status='live', Symbol='Hal', Aliases='Hsd, his, histidase', description='histidine ammonia lyase', other_designations='', map_location='10 C2-D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93488768, end_position_on_the_genomic_accession=93516743, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  15110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15110, CurrentID=0, Status='live', Symbol='Hand1', Aliases='Ehand1, Hxt, Th1, Thing1, bHLHa27, eHAND', description='heart and neural crest derivatives expressed transcript 1', other_designations='extraembryonic tissues, heart, autonomic nervous system and neural crest derivatives-expressed protein 1|helix-loop-helix transcription factor expressed in extraembryonic mesoderm and trophoblast|thing-1', map_location='11 35.27 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=57828713, end_position_on_the_genomic_accession=57832147, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15111, CurrentID=0, Status='live', Symbol='Hand2', Aliases='AI225906, AI661148, Ehand2, Hed, Th2, Thing2, bHLHa26, dHAND', description='heart and neural crest derivatives expressed transcript 2', other_designations='deciduum, heart, autonomic nervous system and neural crest derivatives-expressed protein 2|helix-loop-helix transcription factor expressed in embryo and deciduum|thing-2', map_location='8 29.8 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=57320983, end_position_on_the_genomic_accession=57324517, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15112, CurrentID=0, Status='live', Symbol='Hao1', Aliases='GOX, Gox1, Hao-1', description='hydroxyacid oxidase 1, liver', other_designations='HAOX1|glycolate oxidase 1', map_location='2 65.65 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=134497361, end_position_on_the_genomic_accession=134554352, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15114, CurrentID=0, Status='live', Symbol='Hap1', Aliases='HAP-1', description='huntingtin-associated protein 1', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100347327, end_position_on_the_genomic_accession=100356353, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  15115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15115, CurrentID=0, Status='live', Symbol='Hars', Aliases='MMHRS', description='histidyl-tRNA synthetase', other_designations='hisRS|histidine tRNA ligase|histidyl-tRNA synthetase, cytoplasmic', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36766528, end_position_on_the_genomic_accession=36783205, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  15116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15116, CurrentID=0, Status='live', Symbol='Has1', Aliases='HAS', description='hyaluronan synthase1', other_designations='HA synthase 1|hyaluronate synthase 1|hyaluronic acid synthase 1', map_location='17 10.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17840587, end_position_on_the_genomic_accession=17858011, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15117, CurrentID=0, Status='live', Symbol='Has2', Aliases='', description='hyaluronan synthase 2', other_designations='HA synthase 2|hyaluronan synthase homolog|hyaluronate synthase 2|hyaluronic acid synthase 2', map_location='15 23.31 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=56665627, end_position_on_the_genomic_accession=56694546, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15118, CurrentID=0, Status='live', Symbol='Has3', Aliases='', description='hyaluronan synthase 3', other_designations='HA synthase 3|hyaluronate synthase 3|hyaluronic acid synthase 3', map_location='8 53.28 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106870242, end_position_on_the_genomic_accession=106882902, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15121, CurrentID=0, Status='live', Symbol='Hba', Aliases='', description='hemoglobin alpha chain complex', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15122, CurrentID=0, Status='live', Symbol='Hba-a1', Aliases='Hba, Hba1, Hbat1', description='hemoglobin alpha, adult chain 1', other_designations='alpha 1 globin|alpha globin 1|alpha-globin|haemaglobin alpha 1|hemoglobin alpha chain', map_location='11 18.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32283672, end_position_on_the_genomic_accession=32284486, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15124, CurrentID=0, Status='live', Symbol='Hba-ps3', Aliases='Hba-3ps', description='hemoglobin alpha, pseudogene 3', other_designations='', map_location='15 22.75 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=56171880, end_position_on_the_genomic_accession=56172421, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  15126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15126, CurrentID=0, Status='live', Symbol='Hba-x', Aliases='AI450015, Hbz', description='hemoglobin X, alpha-like embryonic chain in Hba complex', other_designations='alpha-like embryonic globin chain x|hemaglobin zeta|hemoglobin zeta chain|zeta globin|zeta-globin', map_location='11 18.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32276600, end_position_on_the_genomic_accession=32278116, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15127, CurrentID=0, Status='live', Symbol='Hbb', Aliases='', description='hemoglobin beta chain complex', other_designations='beta-globin', map_location='7', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15128, CurrentID=0, Status='live', Symbol='Hbb-ar', Aliases='HS2, HS3, HS4, HS5, LCR', description='hemoglobin, activating region', other_designations='locus control region', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15129, CurrentID=0, Status='live', Symbol='Hbb-b1', Aliases='AA409645, Hbb1, Hbbt1, Hbbt2, MommeD7, beta1', description='hemoglobin, beta adult major chain', other_designations='beta maj|beta major globin', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15130, CurrentID=0, Status='live', Symbol='Hbb-b2', Aliases='AI036344, Hbb2, Hbbt2, beta2', description='hemoglobin, beta adult minor chain', other_designations='beta min|beta minor globin', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15131, CurrentID=0, Status='live', Symbol='Hbb-bh0', Aliases='Gm16324, OTTMUSG00000033600, beta-HO', description='hemoglobin, beta, pseudogene bh0', other_designations='beta-globin HO', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15132, CurrentID=0, Status='live', Symbol='Hbb-bh1', Aliases='betaH1', description='hemoglobin Z, beta-like embryonic chain', other_designations='beta Hl globin|beta-H1-globin|hemoglobin beta-H1 chain|minor beta-globin|protein Z', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103841638, end_position_on_the_genomic_accession=103843162, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15134, CurrentID=0, Status='live', Symbol='Hbb-bh3', Aliases='Gm15119, OTTMUSG00000019070', description='hemoglobin beta, pseudogene bh3', other_designations='', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103832628, end_position_on_the_genomic_accession=103833732, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  15135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15135, CurrentID=0, Status='live', Symbol='Hbb-y', Aliases='Ey, Hby', description='hemoglobin Y, beta-like embryonic chain', other_designations='epsilon Y globin|epsilon-Y2-globin|hemoglobin epsilon-Y2 chain', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103851754, end_position_on_the_genomic_accession=103853207, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15139, CurrentID=0, Status='live', Symbol='Hc', Aliases='C5, C5a, He', description='hemolytic complement', other_designations='', map_location='2 cen-C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34983331, end_position_on_the_genomic_accession=35061441, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  15140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15140, CurrentID=0, Status='live', Symbol='Hc1', Aliases='Cen1', description='heterochromatin, Chr 1', other_designations='satellite head (k<t)|satellite head (k<t>)', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15141, CurrentID=0, Status='live', Symbol='Hc10', Aliases='Cen10', description='heterochromatin, Chr 10', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15142, CurrentID=0, Status='live', Symbol='Hc11', Aliases='Cen11', description='heterochromatin, Chr 11', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15143, CurrentID=0, Status='live', Symbol='Hc12', Aliases='Cen12', description='heterochromatin, Chr 12', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15144, CurrentID=0, Status='live', Symbol='Hc13', Aliases='Cen13', description='heterochromatin, Chr 13', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15145, CurrentID=0, Status='live', Symbol='Hc14', Aliases='Cen14', description='heterochromatin, Chr 14', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15146, CurrentID=0, Status='live', Symbol='Hc15', Aliases='Cen15', description='heterochromatin, Chr 15', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15147, CurrentID=0, Status='live', Symbol='Hc16', Aliases='Cen16', description='heterochromatin, Chr 16', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15148, CurrentID=0, Status='live', Symbol='Hc17', Aliases='Cen17', description='heterochromatin, Chr 17', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15149, CurrentID=0, Status='live', Symbol='Hc18', Aliases='Cen18', description='heterochromatin, Chr 18', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15150, CurrentID=0, Status='live', Symbol='Hc19', Aliases='Cen19', description='heterochromatin, Chr 19', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15151, CurrentID=0, Status='live', Symbol='Hc2', Aliases='Cen2', description='heterochromatin, Chr 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15152, CurrentID=0, Status='live', Symbol='Hc3', Aliases='Cen3', description='heterochromatin, Chr 3', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15153, CurrentID=0, Status='live', Symbol='Hc4', Aliases='Cen4', description='heterochromatin, Chr 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15154, CurrentID=0, Status='live', Symbol='Hc5', Aliases='Cen5', description='heterochromatin, Chr 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15155, CurrentID=0, Status='live', Symbol='Hc6', Aliases='Cen6', description='heterochromatin, Chr 6', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15156, CurrentID=0, Status='live', Symbol='Hc7', Aliases='Cen7', description='heterochromatin, Chr 7', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15157, CurrentID=0, Status='live', Symbol='Hc8', Aliases='Cen8', description='heterochromatin, Chr 8', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15158, CurrentID=0, Status='live', Symbol='Hc9', Aliases='Cen9', description='heterochromatin, Chr 9', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15159, CurrentID=0, Status='live', Symbol='Hccs', Aliases='', description='holocytochrome c synthetase', other_designations='CCHL|holocytochrome c-type synthase', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=169311530, end_position_on_the_genomic_accession=169320343, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15160, CurrentID=0, Status='live', Symbol='Serpind1', Aliases='AA985900, AI303446, HCII, Hcf2', description='serine (or cysteine) peptidase inhibitor, clade D, member 1', other_designations='HC-II|heparin cofactor II|protease inhibitor leuserpin-2|serine (or cysteine) proteinase inhibitor, clade D (heparin cofactor), member 1|serpin D1', map_location='16 10.74 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17331339, end_position_on_the_genomic_accession=17343574, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15161, CurrentID=0, Status='live', Symbol='Hcfc1', Aliases='AW060991, HCF, HCF1', description='host cell factor C1', other_designations='VP16 accessory protein', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73942792, end_position_on_the_genomic_accession=73967135, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  15162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15162, CurrentID=0, Status='live', Symbol='Hck', Aliases='AI849071, Bmk', description='hemopoietic cell kinase', other_designations='B cell/myeloid kinase|B-cell/myeloid kinase|hematopoietic cell kinase|p56-HCK/p59-HCK', map_location='2 75.41 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153108468, end_position_on_the_genomic_accession=153151441, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  15163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15163, CurrentID=0, Status='live', Symbol='Hcls1', Aliases='AW213261, HS1', description='hematopoietic cell specific Lyn substrate 1', other_designations='hematopoietic cell-specific LYN substrate 1|lckBP1', map_location='16 B', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36934983, end_position_on_the_genomic_accession=36963214, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15164, CurrentID=0, Status='live', Symbol='Hcm', Aliases='', description='heterochromatin, multiple chromosomes', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15165, CurrentID=0, Status='live', Symbol='Hcn1', Aliases='Bcng1, C630013B14Rik, HAC2', description='hyperpolarization-activated, cyclic nucleotide-gated K+ 1', other_designations='BCNG-1|HAC-2|brain cyclic nucleotide-gated channel 1|hyperpolarization-activated cation channel 2', map_location='13 66.34 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=117602320, end_position_on_the_genomic_accession=117987414, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  15166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15166, CurrentID=0, Status='live', Symbol='Hcn2', Aliases='BCNG2, HAC1, trls', description='hyperpolarization-activated, cyclic nucleotide-gated K+ 2', other_designations='BCNG-2|HAC-1|brain cyclic nucleotide-gated channel 2|hyperpolarization-activated cation channel 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79716634, end_position_on_the_genomic_accession=79736108, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  15168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15168, CurrentID=0, Status='live', Symbol='Hcn3', Aliases='BCNG-4, Bcng4, Hac3, Hcn4', description='hyperpolarization-activated, cyclic nucleotide-gated K+ 3', other_designations='HAC-3|hyperpolarization-activated cation channel 3|hyperpolarization-activated, cyclic nucleotide-gated K+ 4', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89146775, end_position_on_the_genomic_accession=89160228, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15170, CurrentID=0, Status='live', Symbol='Ptpn6', Aliases='70Z-SHP, Hcph, PTPTY-42, Ptp1C, SH-PTP1, SHP-1, hcp, me', description='protein tyrosine phosphatase, non-receptor type 6', other_designations='hematopoietic cell protein-tyrosine phosphatase|motheaten|protein-tyrosine phosphatase 1C', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124720707, end_position_on_the_genomic_accession=124738709, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  15171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15171, CurrentID=0, Status='live', Symbol='Hcrt', Aliases='PPOX', description='hypocretin', other_designations='orexin A|orexin B|prepro-orexin', map_location='11 63.6 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100761693, end_position_on_the_genomic_accession=100762931, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15181, CurrentID=0, Status='live', Symbol='Gm10093', Aliases='EG15181, ENSMUSG00000061062, Gm1824, HD1, Hdac1, Hdac1-ps, RPD3', description='predicted pseudogene 10093', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78491565, end_position_on_the_genomic_accession=78493541, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  15182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15182, CurrentID=0, Status='live', Symbol='Hdac2', Aliases='D10Wsu179e, YAF1, Yy1bp, mRPD3', description='histone deacetylase 2', other_designations='HD2|YY1 transcription factor-binding protein', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=36974544, end_position_on_the_genomic_accession=37001889, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  15183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15183, CurrentID=0, Status='live', Symbol='Hdac3', Aliases='AW537363', description='histone deacetylase 3', other_designations='HD3|HDAC3 splicing HDAC3alpha|HDAC3 splicing HDAC3beta|HDAC3 splicing HDAC3delta|HDAC3 splicing HDAC3gamma', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37936971, end_position_on_the_genomic_accession=37955000, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  15184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15184, CurrentID=0, Status='live', Symbol='Hdac5', Aliases='AI426555, Hdac4, mHDA1, mKIAA0600', description='histone deacetylase 5', other_designations='HD5|histone deacetylase 4|histone deacetylase mHDA1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102195747, end_position_on_the_genomic_accession=102230186, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  15185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15185, CurrentID=0, Status='live', Symbol='Hdac6', Aliases='Hd6, Hdac5, Sfc6, mHDA2', description='histone deacetylase 6', other_designations='histone deacetylase 5|histone deacetylase mHDA2|scurfy candidate 6', map_location='X 3.58 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7930120, end_position_on_the_genomic_accession=7947889, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  15186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15186, CurrentID=0, Status='live', Symbol='Hdc', Aliases='AW108189-a, Hdc-c, Hdc-e, Hdc-s, Hdc', description='histidine decarboxylase', other_designations='L-histidine decarboxylase', map_location='2 61.76 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126593659, end_position_on_the_genomic_accession=126624022, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  15191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15191, CurrentID=0, Status='live', Symbol='Hdgf', Aliases='AI118077, D3Ertd299e', description='hepatoma-derived growth factor', other_designations='', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87906321, end_position_on_the_genomic_accession=87916132, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  15192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15192, CurrentID=0, Status='live', Symbol='Hdgfl1', Aliases='4921520N01Rik, HRP-1, Hdgfrp1, Pwwp1', description='hepatoma derived growth factor-like 1', other_designations='PWWP domain containing 1|hepatoma-derived growth factor, related protein 1', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=26768172, end_position_on_the_genomic_accession=26770165, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15193, CurrentID=0, Status='live', Symbol='Hdgfrp2', Aliases='HRP-2', description='hepatoma-derived growth factor, related protein 2', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56079604, end_position_on_the_genomic_accession=56100601, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  15194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15194, CurrentID=0, Status='live', Symbol='Htt', Aliases='AI256365, C430023I11Rik, Hd, Hdh, IT15', description='huntingtin', other_designations='HD protein homolog|Huntington disease gene homolog|huntington disease protein homolog', map_location='5 17.92 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34761736, end_position_on_the_genomic_accession=34912534, orientation='plus', exon_count=67, OMIM='', no_hdr0=''),
  670558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=670558, CurrentID=0, Status='live', Symbol='H60c', Aliases='4632413I24Rik, Gm9491', description='histocompatibility 60c', other_designations='histocompatibility antigen 60c', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=3256202, end_position_on_the_genomic_accession=3268092, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15199, CurrentID=0, Status='live', Symbol='Hebp1', Aliases='Hebp, p22HBP', description='heme binding protein 1', other_designations='p22 HBP', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135137519, end_position_on_the_genomic_accession=135168215, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15200, CurrentID=0, Status='live', Symbol='Hbegf', Aliases='AW047313, Dtr, Dts, Hegfl', description='heparin-binding EGF-like growth factor', other_designations='diphtheria toxin receptor|heparin binding epidermal growth factor-like growth factor', map_location='18 19.46 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36504927, end_position_on_the_genomic_accession=36515805, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15201, CurrentID=0, Status='live', Symbol='Hells', Aliases='AI323785, E130115I21Rik, LSH, Lysh, PASG, YFK8', description='helicase, lymphoid specific', other_designations='hells|proliferation-associated SNF2-like protein', map_location='19 C3-D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38930990, end_position_on_the_genomic_accession=38970470, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  15202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15202, CurrentID=0, Status='live', Symbol='Gml2', Aliases='1700057K19Rik, HemT, Hemt1', description='glycosylphosphatidylinositol anchored molecule like 2', other_designations='hematopoietic cell-specific transcript', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74818880, end_position_on_the_genomic_accession=74846164, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15203, CurrentID=0, Status='live', Symbol='Heph', Aliases='C130006F04Rik, Cpl, mKIAA0698, sla', description='hephaestin', other_designations='haphaestin|sex linked anemia', map_location='X 42.69 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=96455436, end_position_on_the_genomic_accession=96574485, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  15204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15204, CurrentID=0, Status='live', Symbol='Herc2', Aliases='D15F32S1h, D7H15F32S1, D7H15F37S1, jdf2, mKIAA0393, rjs', description='HECT and RLD domain containing E3 ubiquitin protein ligase 2', other_designations='HECT domain and RCC1-like domain-containing protein 2|hect (homologous to the E6-AP (UBE3A) carboxyl terminus) domain and RCC1 (CHC1)-like domain (RLD) 2|herc2|mutant HERC2 isoform|probable E3 ubiquitin-protein ligase HERC2', map_location='7 33.42 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=56050155, end_position_on_the_genomic_accession=56231800, orientation='plus', exon_count=94, OMIM='', no_hdr0=''),
  15205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15205, CurrentID=0, Status='live', Symbol='Hes1', Aliases='Hry, bHLHb39', description='hairy and enhancer of split 1 (Drosophila)', other_designations='', map_location='16 21.09 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30065357, end_position_on_the_genomic_accession=30067796, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15206, CurrentID=0, Status='live', Symbol='Hes2', Aliases='HES-2', description='hairy and enhancer of split 2 (Drosophila)', other_designations='', map_location='4 82.92 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152158867, end_position_on_the_genomic_accession=152162469, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15207, CurrentID=0, Status='live', Symbol='Hes3', Aliases='Hes-3, bHLHb43', description='hairy and enhancer of split 3 (Drosophila)', other_designations='', map_location='4 83.01 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152285250, end_position_on_the_genomic_accession=152291903, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15208, CurrentID=0, Status='live', Symbol='Hes5', Aliases='bHLHb38', description='hairy and enhancer of split 5 (Drosophila)', other_designations='', map_location='4 86.01 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154958680, end_position_on_the_genomic_accession=154962372, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  277353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277353, CurrentID=0, Status='live', Symbol='Tcfl5', Aliases='AV260129, Figlb', description='transcription factor-like 5 (basic helix-loop-helix)', other_designations='FIG beta|factor in the germline beta', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180621956, end_position_on_the_genomic_accession=180642691, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15211, CurrentID=0, Status='live', Symbol='Hexa', Aliases='Hex-1', description='hexosaminidase A', other_designations='N-acetyl-beta-glucosaminidase subunit alpha|beta-N-acetylhexosaminidase subunit alpha|hexosaminidase subunit A', map_location='9 32.02 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59539545, end_position_on_the_genomic_accession=59565109, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15212, CurrentID=0, Status='live', Symbol='Hexb', Aliases='', description='hexosaminidase B', other_designations='N-acetyl-beta-glucosaminidase subunit beta|beta-N-acetylhexosaminidase subunit beta|hexosaminidase subunit B', map_location='13 50.66 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=97176331, end_position_on_the_genomic_accession=97198357, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  15213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15213, CurrentID=0, Status='live', Symbol='Hey1', Aliases='AI316788, AI414254, CHF2, HRT1, Herp2, Hesr1, bHLHb31, hesr-1', description='hairy/enhancer-of-split related with YRPW motif 1', other_designations='HRT-1|Hairy/E(spl)-related with YRPW motif 1|hairy and enhancer of split-related protein 1|hairy-related transcription factor 1|mHRT1', map_location='3 2.15 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=8663359, end_position_on_the_genomic_accession=8667038, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15214, CurrentID=0, Status='live', Symbol='Hey2', Aliases='CHF1, Herp1, Hrt2, bHLHb32, hesr2', description='hairy/enhancer-of-split related with YRPW motif 2', other_designations='HES-related repressor protein 2|HESR-2|HRT-2|Hairy-E(spl)-related with YRPW motif 2|hairy and enhancer of split related 2|hairy and enhancer of split-related protein 2|hairy-related transcription factor 2|mHRT2|protein gridlock homolog', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=30832359, end_position_on_the_genomic_accession=30842783, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  277360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277360, CurrentID=0, Status='live', Symbol='Prex1', Aliases='G630042G04, P-REX1, Setd6', description='phosphatidylinositol-3,4,5-trisphosphate-dependent Rac exchange factor 1', other_designations='SET domain containing 6|ptdIns(3,4,5)-dependent Rac exchanger 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=166566345, end_position_on_the_genomic_accession=166713832, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  15218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15218, CurrentID=0, Status='live', Symbol='Foxn1', Aliases='D11Bhm185e, Fkh19, HFH-11, Hfh11, Whn, nu, nude', description='forkhead box N1', other_designations='HNF-3/forkhead homolog 11|hepatocyte nuclear factor 3 forkhead homolog 11|winged-helix transcription factor nude', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78357577, end_position_on_the_genomic_accession=78386922, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15220, CurrentID=0, Status='live', Symbol='Foxq1', Aliases='HFH-1, Hfh1, Hfh1l, sa', description='forkhead box Q1', other_designations='HFH-1l|HNF-3/forkhead homolog 1 like|HNF-3/forkhead-like protein 1|hepatocyte nuclear factor 3 forkhead homolog 1|satin', map_location='13 13.46 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=31558170, end_position_on_the_genomic_accession=31560976, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15221, CurrentID=0, Status='live', Symbol='Foxd3', Aliases='CWH3, Genesis, Hfh2', description='forkhead box D3', other_designations='HFH-2|HNF-3/forkhead homolog 2|HNF3/FH transcription factor genesis|hepatocyte nuclear factor 3 forkhead homolog 2', map_location='4 45.71 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=99656299, end_position_on_the_genomic_accession=99658671, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15223, CurrentID=0, Status='live', Symbol='Foxj1', Aliases='FKHL-13, HFH-4, Hfh4', description='forkhead box J1', other_designations='HNF-3/forkhead homolog 4|forkhead-related protein FKHL13|hepatocyte nuclear factor 3 forkhead homolog 4', map_location='11 81.16 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116330704, end_position_on_the_genomic_accession=116335516, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15227, CurrentID=0, Status='live', Symbol='Foxf1', Aliases='AI450827, FREAC1a, Freac-1, HFH-8, Hfh8, Foxf1', description='forkhead box F1', other_designations='HNF-3/forkhead homolog 8|forkhead box F1a|forkhead-related protein FKHL5|forkhead-related transcription factor 1|hepatocyte nuclear factor 3 forkhead homolog 8', map_location='8 70.31 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121084386, end_position_on_the_genomic_accession=121088154, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15228, CurrentID=0, Status='live', Symbol='Foxg1', Aliases='2900064B05Rik, BF-1, Bf1, Hfh9, Hfhbf1', description='forkhead box G1', other_designations='HNF-3/forkhead homolog, brain factor 1|forkhead-related protein FKHL1', map_location='12 21.78 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=49382883, end_position_on_the_genomic_accession=49386867, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15229, CurrentID=0, Status='live', Symbol='Foxd1', Aliases='AI385632, BF-2, FREAC4, Hfh10, Hfhbf2', description='forkhead box D1', other_designations='FORKHEAD BOX PROTEIN D1 (FORKHEAD-RELATED PROTEIN FKHL8) (FORKHEAD-RELATED TRANSCRIPTION FACTOR 4) (FREAC-4)|FREAC-4|HFH-BF-2|HNF-3/forkhead homolog, brain factor 2|brain factor 2|forkhead-related protein FKHL8|forkhead-related transcription factor 4', map_location='13 51.96 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98354245, end_position_on_the_genomic_accession=98356705, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15233, CurrentID=0, Status='live', Symbol='Hgd', Aliases='Hgo, aku', description='homogentisate 1, 2-dioxygenase', other_designations='alkaptonuria|homogentisate oxygenase|homogentisic acid oxidase|homogentisicase', map_location='16 26.4 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37580153, end_position_on_the_genomic_accession=37632026, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15234, CurrentID=0, Status='live', Symbol='Hgf', Aliases='C230052L06Rik, HGF/SF, NK1, NK2, SF, SF/HGF', description='hepatocyte growth factor', other_designations='hepatopoietin-A|scatter factor', map_location='5 7.07 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=16553335, end_position_on_the_genomic_accession=16622408, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  15235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15235, CurrentID=0, Status='live', Symbol='Mst1', Aliases='D3F15S2h, D9H3F15S2, DNF15S2h, Hgfl', description='macrophage stimulating 1 (hepatocyte growth factor-like)', other_designations='MSP|macrophage stimulatory protein', map_location='9 59.07 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108080380, end_position_on_the_genomic_accession=108085027, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  15239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15239, CurrentID=0, Status='live', Symbol='Hgs', Aliases='Hgr, Hrs, ZFYVE8, tn', description='HGF-regulated tyrosine kinase substrate', other_designations='teetering', map_location='11 84.16 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120467635, end_position_on_the_genomic_accession=120483984, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  15242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15242, CurrentID=0, Status='live', Symbol='Hhex', Aliases='Hex, Hex1-rs2, Prh, Prhx, Hhex', description='hematopoietically expressed homeobox', other_designations='hematopoietically expressed homeobox, related sequence 2|homeobox protein HEX|homeobox protein PRH|mHex', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37434841, end_position_on_the_genomic_accession=37440731, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  15245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15245, CurrentID=0, Status='live', Symbol='Hhip', Aliases='Hhip1', description='Hedgehog-interacting protein', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79965850, end_position_on_the_genomic_accession=80058070, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  15247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15247, CurrentID=0, Status='live', Symbol='Mfsd14a', Aliases='Hiat1', description='major facilitator superfamily domain containing 14A', other_designations='hippocampus abundant gene transcript 1|tetracycline transporter-like protein', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116631164, end_position_on_the_genomic_accession=116662692, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  15248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15248, CurrentID=0, Status='live', Symbol='Hic1', Aliases='AA408311, HIC-1', description='hypermethylated in cancer 1', other_designations='', map_location='11 45.76 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75164565, end_position_on_the_genomic_accession=75170255, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15251, CurrentID=0, Status='live', Symbol='Hif1a', Aliases='AA959795, HIF-1-alpha, HIF1-alpha, HIF1alpha, MOP1, bHLHe78', description='hypoxia inducible factor 1, alpha subunit', other_designations='ARNT-interacting protein', map_location='12 31.99 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73901329, end_position_on_the_genomic_accession=73947530, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  277396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277396, CurrentID=0, Status='live', Symbol='Klhl23', Aliases='C130068N17Rik', description='kelch-like 23', other_designations='', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69821920, end_position_on_the_genomic_accession=69836651, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15254, CurrentID=0, Status='live', Symbol='Hint1', Aliases='AA673479, Hint, Ipk1, PKCI-1, PRKCNH1', description='histidine triad nucleotide binding protein 1', other_designations="PKC inhibitor/ interacting protein|adenosine 5'-monophosphoramidase|protein kinase C inhibitor 1|protein kinase C-interacting protein 1", map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54866438, end_position_on_the_genomic_accession=54870496, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15257, CurrentID=0, Status='live', Symbol='Hipk1', Aliases='1110062K04Rik, Myak', description='homeodomain interacting protein kinase 1', other_designations='nuclear body-associated kinase 2|protein kinase Myak', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103739815, end_position_on_the_genomic_accession=103791894, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  15258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15258, CurrentID=0, Status='live', Symbol='Hipk2', Aliases='1110014O20Rik, B230339E18Rik, Stank', description='homeodomain interacting protein kinase 2', other_designations='nuclear body-associated kinase 1|sialophorin tail-associated nuclear serine/threonine-protein kinase', map_location='6 B', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38687512, end_position_on_the_genomic_accession=38876190, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  15259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15259, CurrentID=0, Status='live', Symbol='Hipk3', Aliases='DYRK6, FIST3, PKY, mir-1902', description='homeodomain interacting protein kinase 3', other_designations='ANPK|FIST|HCMV-interacting protein|androgen receptor-interacting nuclear protein kinase|fas-interacting serine/threonine-protein kinase|nbak3|nuclear body-associated kinase 3', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104426481, end_position_on_the_genomic_accession=104494489, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  15260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15260, CurrentID=0, Status='live', Symbol='Hira', Aliases='AA138857, D16Ertd95e, Gm15797, N28177, OTTMUSG00000026264, Tuple1', description='histone cell cycle regulator', other_designations='TUP1-like enhancer of split gene 1|TUP1-like enhancer of split protein 1|histone cell cycle regulation defective homolog A', map_location='16 A-B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18876750, end_position_on_the_genomic_accession=18970309, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  15267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15267, CurrentID=0, Status='live', Symbol='Hist2h2aa1', Aliases='AI448581, H2a-614, H2a-615, Hist2', description='histone cluster 2, H2aa1', other_designations='histone gene complex 2', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96245497, end_position_on_the_genomic_accession=96246052, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15268, CurrentID=0, Status='live', Symbol='Hist3', Aliases='', description='histone cluster 3', other_designations='histone 3 protein', map_location='11', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  277414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277414, CurrentID=0, Status='live', Symbol='Trp53i11', Aliases='Tp53i11', description='transformation related protein 53 inducible protein 11', other_designations='Trp53 inducible protein 11 (Trp53i11)|p53-induced gene 11 protein|tumor protein p53 inducible protein 11', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93187534, end_position_on_the_genomic_accession=93201757, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  15273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15273, CurrentID=0, Status='live', Symbol='Hivep2', Aliases='MIBP1, Schnurri-2, Shn-2', description='human immunodeficiency virus type I enhancer binding protein 2', other_designations='MIBP-1|human immunodeficiency virus type I enhancer-binding protein 2 homolog|myc intron-binding protein 1', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13965779, end_position_on_the_genomic_accession=14151378, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15275, CurrentID=0, Status='live', Symbol='Hk1', Aliases='BB404130, Hk-1-s, dea, mHk1-s, Hk1', description='hexokinase 1', other_designations='HK I|downeast anemia|hexokinase type I|hexokinase, tumor isozyme', map_location='10 32.37 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62268855, end_position_on_the_genomic_accession=62379920, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  15277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15277, CurrentID=0, Status='live', Symbol='Hk2', Aliases='AI642394, HKII', description='hexokinase 2', other_designations='HK II|hexokinase type II', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=82725025, end_position_on_the_genomic_accession=82774454, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  15278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15278, CurrentID=0, Status='live', Symbol='Tfb2m', Aliases='Hkp1', description='transcription factor B2, mitochondrial', other_designations="S-adenosylmethionine-6-N', N'-adenosyl(rRNA) dimethyltransferase 2|house-keeping protein 1|mTFB2M|mitochondrial 12S rRNA dimethylase 2|mitochondrial transcription factor B2|mtTFB2", map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=179528050, end_position_on_the_genomic_accession=179546305, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  15280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15280, CurrentID=0, Status='live', Symbol='hl', Aliases='', description='hair loss', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15284, CurrentID=0, Status='live', Symbol='Hlx', Aliases='Hlx1', description='H2.0-like homeobox', other_designations='H2.0-like homeo box 1|homeobox protein HLX1', map_location='1 88.97 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=184727143, end_position_on_the_genomic_accession=184732493, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15285, CurrentID=0, Status='live', Symbol='Mnx1', Aliases='HB9, Hlxb9, MNR2', description='motor neuron and pancreas homeobox 1', other_designations='homeobox gene HB9|homeobox protein HB9', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=29473823, end_position_on_the_genomic_accession=29478470, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  277432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277432, CurrentID=0, Status='live', Symbol='Vstm2l', Aliases='Gm691', description='V-set and transmembrane domain containing 2-like', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157914653, end_position_on_the_genomic_accession=157944719, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  15289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15289, CurrentID=0, Status='live', Symbol='Hmgb1', Aliases='HMG-1, Hmg1, SBP-1, p30', description='high mobility group box 1', other_designations='amphoterin|high mobility group protein 1', map_location='5 89.18 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149047222, end_position_on_the_genomic_accession=149053037, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15312, CurrentID=0, Status='live', Symbol='Hmgn1', Aliases='HMG-14, Hmg14', description='high mobility group nucleosomal binding domain 1', other_designations='high mobility group nucleosome-binding domain-containing protein 1|high mobility group protein 14|high-mobility group nucleosome-binding domain-containing protein 1', map_location='16 56.83 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96121588, end_position_on_the_genomic_accession=96127725, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  277463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277463, CurrentID=0, Status='live', Symbol='Gpr107', Aliases='9930033D15Rik, AI790205, C530034M11, mKIAA1624', description='G protein-coupled receptor 107', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31151997, end_position_on_the_genomic_accession=31216567, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  277468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277468, CurrentID=0, Status='live', Symbol='Slc39a12', Aliases='AW046938, Gm731', description='solute carrier family 39 (zinc transporter), member 12', other_designations='ZIP-12|solute carrier family 39 member 12|zrt- and Irt-like protein 12', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14388316, end_position_on_the_genomic_accession=14494978, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  15331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15331, CurrentID=0, Status='live', Symbol='Hmgn2', Aliases='HMG-17, Hmg17', description='high mobility group nucleosomal binding domain 2', other_designations='high mobility group nucleosome-binding domain-containing protein 2|high mobility group protein 17|high-mobility group nucleosome-binding domain-containing protein 2', map_location='4 66.47 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133964739, end_position_on_the_genomic_accession=133967991, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15352, CurrentID=0, Status='live', Symbol='Gm4739', Aliases='EG15352', description='predicted gene 4739', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=26107749, end_position_on_the_genomic_accession=26108737, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  15353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15353, CurrentID=0, Status='live', Symbol='Hmg20b', Aliases='AW610687, BRAF35, Hmgx2, Hmgxb2, Smarce1r', description='high mobility group 20B', other_designations='BRCA2-associated factor 35|HMG box-containing protein 20B|SMARCE1-related protein|high mobility group 20 B|high mobility group protein 20 B|structural DNA-binding protein BRAF35', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81346046, end_position_on_the_genomic_accession=81350511, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  15354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15354, CurrentID=0, Status='live', Symbol='Hmgb3', Aliases='Hmg2a, Hmg4', description='high mobility group box 3', other_designations='high mobility group protein 2a|high mobility group protein 4|high mobilty group protein 2A|high-mobility group protein 4', map_location='X A1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71555917, end_position_on_the_genomic_accession=71560673, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15356, CurrentID=0, Status='live', Symbol='Hmgcl', Aliases='AW476067, HL', description='3-hydroxy-3-methylglutaryl-Coenzyme A lyase', other_designations='3-hydroxy-3-methylglutarate-CoA lyase|HMG-CoA lyase', map_location='4 68.14 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135946448, end_position_on_the_genomic_accession=135962617, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  15357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15357, CurrentID=0, Status='live', Symbol='Hmgcr', Aliases='HMG-CoAR, Red', description='3-hydroxy-3-methylglutaryl-Coenzyme A reductase', other_designations='3-hydroxy-3-methylglutaryl-CoA reductase|HMG-CoA reductase', map_location='13 50.65 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=96648962, end_position_on_the_genomic_accession=96670936, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  15360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15360, CurrentID=0, Status='live', Symbol='Hmgcs2', Aliases='1300002P16, mHS', description='3-hydroxy-3-methylglutaryl-Coenzyme A synthase 2', other_designations='3-hydroxy-3-methylglutaryl coenzyme A synthase|HMG-CoA synthase', map_location='3 42.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98280431, end_position_on_the_genomic_accession=98310738, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  15361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15361, CurrentID=0, Status='live', Symbol='Hmga1', Aliases='AL023995a, Hmga1b, Hmgi, Hmgiy, Hmgy, Hmga1', description='high mobility group AT-hook 1', other_designations='high mobility group AT-hook protein 1|high mobility group protein A1|high mobility group protein I', map_location='17 14.5 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27556574, end_position_on_the_genomic_accession=27563672, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15364, CurrentID=0, Status='live', Symbol='Hmga2', Aliases='9430083A20Rik, HMGI-C, Hmgic, pg, pygmy', description='high mobility group AT-hook 2', other_designations='high mobility group AT-hook protein 2|high mobility group protein I', map_location='10 67.94 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=120361275, end_position_on_the_genomic_accession=120476935, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15366, CurrentID=0, Status='live', Symbol='Hmmr', Aliases='AA386826, CD168, Rhamm', description='hyaluronan mediated motility receptor (RHAMM)', other_designations='intracellular hyaluronic acid-binding protein|receptor for hyaluronan-mediated motility', map_location='11 24.4 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=40701388, end_position_on_the_genomic_accession=40733464, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  15368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15368, CurrentID=0, Status='live', Symbol='Hmox1', Aliases='D8Wsu38e, HO-1, HO1, Hemox, Hmox, Hsp32', description='heme oxygenase 1', other_designations='P32 protein|heme oxygenase (decycling) 1|hemoxygenase', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=75093618, end_position_on_the_genomic_accession=75100593, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15369, CurrentID=0, Status='live', Symbol='Hmox2', Aliases='HO-2, HO2', description='heme oxygenase 2', other_designations='heme oxygenase (decycling) 2', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4726361, end_position_on_the_genomic_accession=4766249, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  15370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15370, CurrentID=0, Status='live', Symbol='Nr4a1', Aliases='GFRP1, Gfrp, Hbr-1, Hbr1, Hmr, N10, NGFI-B, NGFIB, NP10, TIS1, TR3, nur77', description='nuclear receptor subfamily 4, group A, member 1', other_designations='hormone receptor|nuclear hormone receptor NUR/77|nuclear protein N10|orphan nuclear receptor HMR', map_location='15 F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101254213, end_position_on_the_genomic_accession=101274795, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  15371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15371, CurrentID=0, Status='live', Symbol='Hmx1', Aliases='Nkx5-3', description='H6 homeobox 1', other_designations='H6 homeo box 1|homeobox protein H6', map_location='5 18.14 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35388495, end_position_on_the_genomic_accession=35400455, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15372, CurrentID=0, Status='live', Symbol='Hmx2', Aliases='Nkx-5.2, Nkx5-2', description='H6 homeobox 2', other_designations='H6 homeo box 2|NK-related homeobox|homeobox protein Nkx-5.2', map_location='7 74.34 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131548773, end_position_on_the_genomic_accession=131558867, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15373, CurrentID=0, Status='live', Symbol='Hmx3', Aliases='Nkx-5.1, Nkx5-1, Nkx5.1', description='H6 homeobox 3', other_designations='H6 homeo box 3|homeobox protein H6 family member 3|homeobox protein Nkx-5.1', map_location='7 74.32 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131541138, end_position_on_the_genomic_accession=131546027, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15374, CurrentID=0, Status='live', Symbol='Hn1', Aliases='', description='hematological and neurological expressed sequence 1', other_designations='', map_location='11 80.84 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115497353, end_position_on_the_genomic_accession=115514410, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15375, CurrentID=0, Status='live', Symbol='Foxa1', Aliases='Hnf-3a, Hnf3a, Tcf-3a, Tcf3a', description='forkhead box A1', other_designations='HNF-3-alpha|fork head domain|forkhead box protein A1|hepatocyte nuclear factor 3 alpha (winged helix transcription factor)', map_location='12 24.7 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=57540628, end_position_on_the_genomic_accession=57546911, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15376, CurrentID=0, Status='live', Symbol='Foxa2', Aliases='HNF3-beta, HNF3beta, Hnf-3b, Hnf3b, Tcf-3b, Tcf3b', description='forkhead box A2', other_designations='HNF-3-beta|forkhead box protein A2|hepatocyte nuclear factor 3 beta (winged helix transcription factor)', map_location='2 73.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148042877, end_position_on_the_genomic_accession=148046969, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15377, CurrentID=0, Status='live', Symbol='Foxa3', Aliases='Hnf-3g, Hnf3g, Tcf-3g, Tcf3g', description='forkhead box A3', other_designations='HNF-3-gamma|forkhead box protein A3|hepatocyte nuclear factor 3 gamma (winged helix transcription factor)', map_location='7 9.46 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19013283, end_position_on_the_genomic_accession=19023539, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15378, CurrentID=0, Status='live', Symbol='Hnf4a', Aliases='HNF-4, Hnf4lpha, MODY1, Nr2a1, TCF-14, Tcf14, Hnf4a', description='hepatic nuclear factor 4, alpha', other_designations='HNF4 alpha|Nuclear receptor 2A1|nuclear receptor subfamily 2 group A member 1|transcription factor 14|transcription factor HNF-4', map_location='2 84.32 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163506809, end_position_on_the_genomic_accession=163572907, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  15379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15379, CurrentID=0, Status='live', Symbol='Onecut1', Aliases='D9Ertd423e, HNF-6, Hfh12, Hnf6, Hnf6b, OC-1, Oc1', description='one cut domain, family member 1', other_designations='HNF-3/forkhead homolog 12|hepatic nuclear factor 6|one cut homeobox 1', map_location='9 41.93 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=74861921, end_position_on_the_genomic_accession=74889648, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15381, CurrentID=0, Status='live', Symbol='Hnrnpc', Aliases='AL022939, D14Wsu171e, Hnrpc, Hnrpc1, Hnrpc2, hnRNPC1, hnRNPC2, hnrnp-C', description='heterogeneous nuclear ribonucleoprotein C', other_designations='heterogeneous nuclear ribonucleoprotein C1|heterogeneous nuclear ribonucleoprotein C2|heterogeneous ribonucleoprotein C1/C2|hnRNP C1 / hnRNP C2|hnRNP C1/C2|hnRNP C2|snoRNA MBI-122', map_location='14 26.79 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52073380, end_position_on_the_genomic_accession=52104028, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  15382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15382, CurrentID=0, Status='live', Symbol='Hnrnpa1', Aliases='HDP-1, Hdp, Hnrpa1, hnRNP A1, hnrnp-A1', description='heterogeneous nuclear ribonucleoprotein A1', other_designations='helix-destabilizing protein|hnRNP core protein A1|single-strand-binding protein|topoisomerase-inhibitor suppressed', map_location='15 58.58 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103240397, end_position_on_the_genomic_accession=103246698, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  15384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15384, CurrentID=0, Status='live', Symbol='Hnrnpab', Aliases='3010025C11Rik, CBF-A, Cgbfa, Hnrpab', description='heterogeneous nuclear ribonucleoprotein A/B', other_designations='CArG-binding factor-A|S1 protein D2|hnRNP A/B', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51600100, end_position_on_the_genomic_accession=51606881, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15387, CurrentID=0, Status='live', Symbol='Hnrnpk', Aliases='Hnrpk, KBBP, NOVA', description='heterogeneous nuclear ribonucleoprotein K', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58391132, end_position_on_the_genomic_accession=58403343, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  15388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15388, CurrentID=0, Status='live', Symbol='Hnrnpl', Aliases='C79783, D830027H13Rik, Hnrpl', description='heterogeneous nuclear ribonucleoprotein L', other_designations='hnRNP L', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28808546, end_position_on_the_genomic_accession=28822268, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  15392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15392, CurrentID=0, Status='live', Symbol='hop', Aliases='hpy', description='hop-sterile', other_designations='hydrocephalic-polydactyl', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242864, CurrentID=0, Status='live', Symbol='Napepld', Aliases='A530089G06, Mbldc1, NAPE-PLD', description='N-acyl phosphatidylethanolamine phospholipase D', other_designations='Metallo-beta-lactamase domain containing 1|NAPE-hydrolyzing phospholipase D', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21662901, end_position_on_the_genomic_accession=21702153, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15394, CurrentID=0, Status='live', Symbol='Hoxa1', Aliases='ERA1, Hox-1.6', description='homeobox A1', other_designations='early retinoic acid 1|homeobox protein Hox-1.6|homeoboxless protein ERA-1-399|homeotic protein ERA-1-993', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52155367, end_position_on_the_genomic_accession=52160410, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15395, CurrentID=0, Status='live', Symbol='Hoxa10', Aliases='Hox-1.8, Hoxa-10', description='homeobox A10', other_designations='homeo box A10|homeobox protein Hox-1.8', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52231197, end_position_on_the_genomic_accession=52240854, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15396, CurrentID=0, Status='live', Symbol='Hoxa11', Aliases='Hox-1.9, Hoxa-11', description='homeobox A11', other_designations='homeo box A11|homeobox protein Hox-1.9', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52242105, end_position_on_the_genomic_accession=52245810, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15397, CurrentID=0, Status='live', Symbol='Hoxa11os', Aliases='Hoxa11as, Hoxa11s', description='homeobox A11, opposite strand', other_designations='HOXA11 antisense RNA (non-protein coding)|HOXA11 opposite strand RNA (non-protein coding)|Hoxa11 opposite strand transcript (non-protein coding)|homeo box A11, opposite strand transcript', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52245243, end_position_on_the_genomic_accession=52249769, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15398, CurrentID=0, Status='live', Symbol='Hoxa13', Aliases='Hd, Hox-1.10', description='homeobox A13', other_designations='homeo box A13|homeobox protein Hox-1.10|hypodactyly', map_location='6 25.41 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52258853, end_position_on_the_genomic_accession=52260880, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15399, CurrentID=0, Status='live', Symbol='Hoxa2', Aliases='AI324701, HOX1.11, Hox-1.11', description='homeobox A2', other_designations='homeo box A2|homeobox protein Hox-1.11', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52162417, end_position_on_the_genomic_accession=52164831, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15400, CurrentID=0, Status='live', Symbol='Hoxa3', Aliases='Hox-1.5, Mo-10', description='homeobox A3', other_designations='homeo box A3|homeobox protein Hox-1.5|homeobox protein MO-10', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52169062, end_position_on_the_genomic_accession=52213067, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15401, CurrentID=0, Status='live', Symbol='Hoxa4', Aliases='AV206827, Hox-1.4', description='homeobox A4', other_designations='homeo box A4|homeobox protein Hox-1.4|homeobox protein MH-3', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52189687, end_position_on_the_genomic_accession=52191703, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15402, CurrentID=0, Status='live', Symbol='Hoxa5', Aliases='Hox-1.3', description='homeobox A5', other_designations='homeo box A5|homeobox protein Hox-1.3|homeobox protein M2', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52201754, end_position_on_the_genomic_accession=52204587, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15403, CurrentID=0, Status='live', Symbol='Hoxa6', Aliases='Hox-1.2', description='homeobox A6', other_designations='homeo box A6|homeobox protein Hox-1.2|homeobox protein M5-4', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52206365, end_position_on_the_genomic_accession=52208624, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15404, CurrentID=0, Status='live', Symbol='Hoxa7', Aliases='AV118143, Hox-1.1, M6', description='homeobox A7', other_designations='HOMEOBOX PROTEIN HOX-A7 (HOX-1.1) (M6-12) (M6)', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52214491, end_position_on_the_genomic_accession=52221852, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15405, CurrentID=0, Status='live', Symbol='Hoxa9', Aliases='D6a9, Hox-1.7', description='homeobox A9', other_designations='homeo box A9|homeobox protein Hox-1.7', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52223097, end_position_on_the_genomic_accession=52227370, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15406, CurrentID=0, Status='live', Symbol='Hoxb', Aliases='Hox-2', description='homeobox B cluster', other_designations='homeo box B cluster', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15407, CurrentID=0, Status='live', Symbol='Hoxb1', Aliases='Hox-2.9', description='homeobox B1', other_designations='homeo box B1|homeobox protein Hox-2.9', map_location='11 59.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96365746, end_position_on_the_genomic_accession=96368254, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15408, CurrentID=0, Status='live', Symbol='Hoxb13', Aliases='', description='homeobox B13', other_designations='homeo box B13', map_location='11 59.63 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96194361, end_position_on_the_genomic_accession=96196599, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15410, CurrentID=0, Status='live', Symbol='Hoxb3', Aliases='Hox-2.7', description='homeobox B3', other_designations='homeo box B3|homeobox protein Hox-2.7|homeobox protein MH-23', map_location='11 59.84 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96307079, end_position_on_the_genomic_accession=96347930, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  15412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15412, CurrentID=0, Status='live', Symbol='Hoxb4', Aliases='Hox-2.6', description='homeobox B4', other_designations='homeo box B4|homeobox protein Hox-2.6', map_location='11 59.83 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96318267, end_position_on_the_genomic_accession=96321638, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15413, CurrentID=0, Status='live', Symbol='Hoxb5', Aliases='AI385717, Hox-2.1', description='homeobox B5', other_designations='homeo box B5|homeobox protein H24.1|homeobox protein Hox-2.1|homeobox protein Mu-1', map_location='11 59.82 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96303512, end_position_on_the_genomic_accession=96306121, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15414, CurrentID=0, Status='live', Symbol='Hoxb6', Aliases='Hox-2.2', description='homeobox B6', other_designations='homeo box B6|homeobox protein Hox-2.2|homeobox protein MH-22A', map_location='11 59.82 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96292669, end_position_on_the_genomic_accession=96301569, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15415, CurrentID=0, Status='live', Symbol='Hoxb7', Aliases='AI325018, Hox-2.3', description='homeobox B7', other_designations='homeo box B7|homeobox protein Hox-2.3|homeobox protein MH-22B|homeobox protein MuB1', map_location='11 59.82 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96286646, end_position_on_the_genomic_accession=96290163, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15416, CurrentID=0, Status='live', Symbol='Hoxb8', Aliases='Hox-2.4', description='homeobox B8', other_designations='homeo box B8|homeobox protein Hox-2.4', map_location='11 59.82 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96281905, end_position_on_the_genomic_accession=96285325, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15417, CurrentID=0, Status='live', Symbol='Hoxb9', Aliases='Hox-2.5', description='homeobox B9', other_designations='homeo box B9|homeobox protein Hox-2.5', map_location='11 59.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96250938, end_position_on_the_genomic_accession=96276595, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  277562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277562, CurrentID=0, Status='live', Symbol='Olfr1286', Aliases='GA_x5J8B7TU5VU-1563-1372, GA_x6K02T2Q125-72472405-72471488, MOR248-21', description='olfactory receptor 1286', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111420032, end_position_on_the_genomic_accession=111420949, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15421, CurrentID=0, Status='live', Symbol='Hoxc12', Aliases='Hox-3.8', description='homeobox C12', other_designations='Homeobox protein Hox-C12 (Hox-3F)|homeo box C12|homeobox protein Hox-3.8', map_location='15 58.01 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102936723, end_position_on_the_genomic_accession=102940855, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15422, CurrentID=0, Status='live', Symbol='Hoxc13', Aliases='', description='homeobox C13', other_designations='homeo box C13', map_location='15 57.99 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102921116, end_position_on_the_genomic_accession=102929135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15423, CurrentID=0, Status='live', Symbol='Hoxc4', Aliases='Hox-3.5', description='homeobox C4', other_designations='homeo box C4|homeobox protein Hox-3.5', map_location='15 58.08 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103018099, end_position_on_the_genomic_accession=103036856, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  15424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15424, CurrentID=0, Status='live', Symbol='Hoxc5', Aliases='AI181837, Hox-3.4, Hox-6.2', description='homeobox C5', other_designations='homeo box C5|homeobox protein Hox-3.4|homeobox protein Hox-6.2', map_location='15 58.07 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103014008, end_position_on_the_genomic_accession=103017429, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15425, CurrentID=0, Status='live', Symbol='Hoxc6', Aliases='Hox-3.3, Hox-6.1', description='homeobox C6', other_designations='homeo box C6|homeobox protein Hox-3.3|homeobox protein Hox-6.1', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102998226, end_position_on_the_genomic_accession=103011882, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15426, CurrentID=0, Status='live', Symbol='Hoxc8', Aliases='D130011F21Rik, Hox-3.1', description='homeobox C8', other_designations='homeo box C8|homeobox protein Hox-3.1|homeobox protein M31', map_location='15 58.05 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102990539, end_position_on_the_genomic_accession=102994254, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15427, CurrentID=0, Status='live', Symbol='Hoxc9', Aliases='Hox-3.2', description='homeobox C9', other_designations='homeo box C9|homeobox protein Hox-3.2', map_location='15 58.04 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102977032, end_position_on_the_genomic_accession=102984444, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15429, CurrentID=0, Status='live', Symbol='Hoxd1', Aliases='Hox-4.9', description='homeobox D1', other_designations='homeo box D1|homeobox protein Hox-4.9', map_location='2 44.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74762980, end_position_on_the_genomic_accession=74765142, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15430, CurrentID=0, Status='live', Symbol='Hoxd10', Aliases='AI385591, AI874987, Hox-4.5, Hox-5.3', description='homeobox D10', other_designations='homeo box D10|homeobox protein Hox-4.5|homeobox protein Hox-5.3', map_location='2 44.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74684275, end_position_on_the_genomic_accession=74695106, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15431, CurrentID=0, Status='live', Symbol='Hoxd11', Aliases='E230017H14Rik, Hox-4.6, Hox-5.4, Hox-5.5', description='homeobox D11', other_designations='homeo box D11|homeobox protein Hox-4.6|homeobox protein Hox-5.5', map_location='2 44.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74679558, end_position_on_the_genomic_accession=74687016, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15432, CurrentID=0, Status='live', Symbol='Hoxd12', Aliases='Hox-4.7, Hox-5.6', description='homeobox D12', other_designations='homeo box D12|homeobox protein Hox-4.7|homeobox protein Hox-5.6', map_location='2 44.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74675013, end_position_on_the_genomic_accession=74677705, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15433, CurrentID=0, Status='live', Symbol='Hoxd13', Aliases='Hox-4.8, spdh', description='homeobox D13', other_designations='homeo box D13|homeobox protein Hox-4.8', map_location='2 44.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74668310, end_position_on_the_genomic_accession=74671599, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15434, CurrentID=0, Status='live', Symbol='Hoxd3', Aliases='Hox-4.1, Hox-5.5', description='homeobox D3', other_designations='homeo box D3|homeo box D4|homeobox protein Hox-4.1|homeobox protein MH-19', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74711993, end_position_on_the_genomic_accession=74748271, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15436, CurrentID=0, Status='live', Symbol='Hoxd4', Aliases='6030436D05Rik, Hox-4.2, Hox-5.1', description='homeobox D4', other_designations='homeo box D4|homeobox protein Hox-4.2|homeobox protein Hox-5.1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74721978, end_position_on_the_genomic_accession=74729160, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15437, CurrentID=0, Status='live', Symbol='Hoxd8', Aliases='4921540P06Rik, AI047735, Hox-4.3, Hox-5.4', description='homeobox D8', other_designations='homeo box D8|homeobox protein Hox-4.3|homeobox protein Hox-5.4', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74704615, end_position_on_the_genomic_accession=74707933, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15438, CurrentID=0, Status='live', Symbol='Hoxd9', Aliases='Hox-4.4, Hox-5.2', description='homeobox D9', other_designations='homeo box D9|homeobox protein Hox-4.4|homeobox protein Hox-5.2', map_location='2 44.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74697763, end_position_on_the_genomic_accession=74700208, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15439, CurrentID=0, Status='live', Symbol='Hp', Aliases='HP-1, preHP2', description='haptoglobin', other_designations='zonulin', map_location='8 57.11 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109575128, end_position_on_the_genomic_accession=109579172, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15441, CurrentID=0, Status='live', Symbol='Hp1bp3', Aliases='Hp1bp74', description='heterochromatin protein 1, binding protein 3', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138216582, end_position_on_the_genomic_accession=138244683, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  15442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15442, CurrentID=0, Status='live', Symbol='Hpse', Aliases='HSE1, Hpa, Hpr1', description='heparanase', other_designations='endo-glucoronidase|endoglycosidase heparanase', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100679486, end_position_on_the_genomic_accession=100719683, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15443, CurrentID=0, Status='live', Symbol='hpc', Aliases='', description='hyperspiny Purkinje cell', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15444, CurrentID=0, Status='live', Symbol='Hpca', Aliases='', description='hippocalcin', other_designations='', map_location='4 D2-D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129111570, end_position_on_the_genomic_accession=129122070, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15445, CurrentID=0, Status='live', Symbol='Hpd', Aliases='Fla, Flp, Hppd, Laf', description='4-hydroxyphenylpyruvic acid dioxygenase', other_designations='4-hydroxyphenylpyruvic acid oxidase|4HPPD|F liver antigen|HPPDase|f Alloantigen|f protein', map_location='5 62.82 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123171807, end_position_on_the_genomic_accession=123182686, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  15446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15446, CurrentID=0, Status='live', Symbol='Hpgd', Aliases='15-PGDH, AV026552', description='hydroxyprostaglandin dehydrogenase 15 (NAD)', other_designations='15-hydroxyprostaglandin dehydrogenase|PGDH|prostaglandin dehydrogenase 1', map_location='8 B3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=56294552, end_position_on_the_genomic_accession=56321046, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15447, CurrentID=0, Status='live', Symbol='hph1', Aliases='hph-1', description='hyperphenylalaninemia 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15448, CurrentID=0, Status='live', Symbol='hph2', Aliases='hph-2', description='hyperphenylalaninemia 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15450, CurrentID=0, Status='live', Symbol='Lipc', Aliases='AI256194, HL, Hpl', description='lipase, hepatic', other_designations='hepatic lipase|lipase member C', map_location='9 39.57 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70798128, end_position_on_the_genomic_accession=70952221, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  15451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15451, CurrentID=0, Status='live', Symbol='Hpn', Aliases='Hlb320', description='hepsin', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31098725, end_position_on_the_genomic_accession=31115326, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  15452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15452, CurrentID=0, Status='live', Symbol='Hprt', Aliases='C81579, HPGRT1, Hprt', description='hypoxanthine guanine phosphoribosyl transferase', other_designations='HGPRT|HGPRTase|HPRT B|hypoxanthine guanine phosphoribosyl transferase 1', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52988078, end_position_on_the_genomic_accession=53021660, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  432602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432602, CurrentID=0, Status='live', Symbol='Krtap31-2', Aliases='OTTMUSG00000002177', description='keratin associated protein 31-2', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99936291, end_position_on_the_genomic_accession=99937225, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15458, CurrentID=0, Status='live', Symbol='Hpx', Aliases='Hpxn, hx', description='hemopexin', other_designations='', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105591611, end_position_on_the_genomic_accession=105600116, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  15460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15460, CurrentID=0, Status='live', Symbol='Hr', Aliases='ALUNC, AU, N, ba, bldy, hr, rh, rh-bmh, rhino', description='hairless', other_designations='protein hairless', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70554056, end_position_on_the_genomic_accession=70573548, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  15461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15461, CurrentID=0, Status='live', Symbol='Hras', Aliases='H-ras, Ha-ras, Harvey-ras-1, Hras1, Kras2, c-H-ras, c-Ha-ras, c-rasHa, ras, Hras', description='Harvey rat sarcoma virus oncogene', other_designations='H-Ras-1|H-ras 1 protein|Harvey rat sarcoma virus oncogene 1|c-Ha-ras p21 protein|c-Ha-ras transgene|p21ras|transforming protein P21', map_location='7 86.48 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141189934, end_position_on_the_genomic_accession=141194004, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15463, CurrentID=0, Status='live', Symbol='Agfg1', Aliases='AU045498, C130049H11Rik, C85612, D730048C23Rik, Hrb, RAB, Rip', description='ArfGAP with FG repeats 1', other_designations='HIV-1 Rev binding protein|HIV-1 Rev-binding protein homolog|arf-GAP domain and FG repeats-containing protein 1|nucleoporin-like protein RIP', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=82839460, end_position_on_the_genomic_accession=82901001, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15464, CurrentID=0, Status='live', Symbol='Hrc', Aliases='', description='histidine rich calcium binding protein', other_designations='', map_location='7 29.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45335269, end_position_on_the_genomic_accession=45338973, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15465, CurrentID=0, Status='live', Symbol='Hrh1', Aliases='Bphs, H1R, HH1R, Hir', description='histamine receptor H1', other_designations='Bordetella pertussis induced histamine sensitization', map_location='6 53.05 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=114397925, end_position_on_the_genomic_accession=114483296, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15466, CurrentID=0, Status='live', Symbol='Hrh2', Aliases='', description='histamine receptor H2', other_designations='H2R|HH2R|gastric receptor I', map_location='13 28.4 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54191970, end_position_on_the_genomic_accession=54236182, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15467, CurrentID=0, Status='live', Symbol='Eif2ak1', Aliases='HCR, Hri', description='eukaryotic translation initiation factor 2 alpha kinase 1', other_designations='heme-controlled repressor|heme-regulated eukaryotic initiation factor eIF-2-alpha kinase|heme-regulated inhibitor|hemin-sensitive initiation factor 2-alpha kinase', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143871862, end_position_on_the_genomic_accession=143902717, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15468, CurrentID=0, Status='live', Symbol='Prmt2', Aliases='AI504737, Hrmt1l1', description='protein arginine N-methyltransferase 2', other_designations='heterogeneous nuclear ribonucleoprotein methyltransferase-like 1|histone-arginine N-methyltransferase PRMT2', map_location='10 38.74 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76207226, end_position_on_the_genomic_accession=76237865, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15469, CurrentID=0, Status='live', Symbol='Prmt1', Aliases='6720434D09Rik, AW214366, Hrmt1l2, Mrmt1', description='protein arginine N-methyltransferase 1', other_designations='arginine N-methyltransferase 1|heterogeneous nuclear ribonucleoproteins methyltransferase-like 2|histone-arginine N-methyltransferase PRMT1', map_location='7 29.07 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44976755, end_position_on_the_genomic_accession=44986420, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15473, CurrentID=0, Status='live', Symbol='Hrsp12', Aliases='HR12, HRP12', description='heat-responsive protein 12', other_designations='', map_location='15 B3.3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34484022, end_position_on_the_genomic_accession=34495246, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15475, CurrentID=0, Status='live', Symbol='hs', Aliases='', description='head spot', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15476, CurrentID=0, Status='live', Symbol='Hs3st1', Aliases='3-Ost, D5Wsu110e, Hsg3ost', description='heparan sulfate (glucosamine) 3-O-sulfotransferase 1', other_designations='heparan sulfate 3-O-sulfotransferase 1|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 1', map_location='5 21.14 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=39613935, end_position_on_the_genomic_accession=39644631, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  15478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15478, CurrentID=0, Status='live', Symbol='Hs3st3a1', Aliases='3Ost3a, Hs3st3a', description='heparan sulfate (glucosamine) 3-O-sulfotransferase 3A1', other_designations='heparan sulfate 3-O-sulfotransferase 3A1|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 3A1', map_location='11 40.24 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=64435332, end_position_on_the_genomic_accession=64522835, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15481, CurrentID=0, Status='live', Symbol='Hspa8', Aliases='2410008N15Rik, Hsc70, Hsc71, Hsc73, Hsp73, Hspa10', description='heat shock protein 8', other_designations='heat shock 70 kDa protein 8|heat shock 70kD protein 8|heat shock cognate hsc73|heat shock protein cognate 70', map_location='9 21.55 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40801273, end_position_on_the_genomic_accession=40805199, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  15482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15482, CurrentID=0, Status='live', Symbol='Hspa1l', Aliases='Hsc70t, Msh5', description='heat shock protein 1-like', other_designations='heat shock 70 kDa protein 1L|heat shock 70 kDa-like protein 1|heat shock 70kDa protein 1-like|heat shock protein cognate 70, testis|heat shock-related protein hsc70t|inducible heat shock protein 70|spermatid-specific heat shock protein 70', map_location='17 18.51 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34972703, end_position_on_the_genomic_accession=34979228, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15483, CurrentID=0, Status='live', Symbol='Hsd11b1', Aliases='', description='hydroxysteroid 11-beta dehydrogenase 1', other_designations='11-DH|11-beta-HSD1|11-beta-hydroxysteroid dehydrogenase 1|11beta-HSD-1|11beta-HSD1A|11beta-hydroxysteroid dehydrogenase type 1|Corticosteroid 11-beta-dehydrogenase, isozyme 1 (11-DH) (11-beta-hydroxysteroid dehydrogenase 1) (11-beta-HSD1) (11beta-HSD1A)', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193221640, end_position_on_the_genomic_accession=193264049, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15484, CurrentID=0, Status='live', Symbol='Hsd11b2', Aliases='11HSD2', description='hydroxysteroid 11-beta dehydrogenase 2', other_designations='11(beta)-HSD2|11-DH2|11-beta-HSD2|11-beta-hydroxysteroid dehydrogenase type 2|NAD-dependent 11-beta-hydroxysteroid dehydrogenase', map_location='8 53.04 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105518746, end_position_on_the_genomic_accession=105523988, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15485, CurrentID=0, Status='live', Symbol='Hsd17b1', Aliases='17HSDB1, 17beta-HSD, E2DH, Hsd17ba', description='hydroxysteroid (17-beta) dehydrogenase 1', other_designations='17-beta-HSD 1|17-beta-hydroxysteroid dehydrogenase type 1', map_location='11 64.17 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101077968, end_position_on_the_genomic_accession=101081188, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15486, CurrentID=0, Status='live', Symbol='Hsd17b2', Aliases='AI194836, AI194967, AI255511', description='hydroxysteroid (17-beta) dehydrogenase 2', other_designations='17 HSD type 2|17-beta-HSD 2|17-beta-hydroxysteroid dehydrogenase 2|17-beta-hydroxysteroid dehydrogenase type 2|testosterone 17-beta-dehydrogenase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=117700982, end_position_on_the_genomic_accession=117759030, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15487, CurrentID=0, Status='live', Symbol='Hsd17b3', Aliases='', description='hydroxysteroid (17-beta) dehydrogenase 3', other_designations='17(beta)HSD type 3|17-beta-HSD 3|17-beta-hydroxysteroid dehydrogenase type 3|testicular 17-beta-hydroxysteroid dehydrogenase', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64058274, end_position_on_the_genomic_accession=64089258, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15488, CurrentID=0, Status='live', Symbol='Hsd17b4', Aliases='17-beta-HSD, 17[b]-HSD, DBP, MFE-2, MFP2, MPF-2, Mfp-2, perMFE-2', description='hydroxysteroid (17-beta) dehydrogenase 4', other_designations='17-beta-HSD 4|17-beta-hydroxysteroid dehydrogenase 4|D-bifunctional protein|multifunctional protein 2', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=50128201, end_position_on_the_genomic_accession=50196270, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  15490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15490, CurrentID=0, Status='live', Symbol='Hsd17b7', Aliases='AI266814, ERG27', description='hydroxysteroid (17-beta) dehydrogenase 7', other_designations='17-beta-HSD 7|17-beta-hydroxysteroid dehydrogenase 7|estradiol 17-beta-dehydrogenase 7', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=169949535, end_position_on_the_genomic_accession=169969238, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  15492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15492, CurrentID=0, Status='live', Symbol='Hsd3b1', Aliases='3-beta-HSD I, D3Ertd383e', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 1', other_designations='', map_location='3 42.89 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98842356, end_position_on_the_genomic_accession=98859794, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  15493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15493, CurrentID=0, Status='live', Symbol='Hsd3b2', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 2', other_designations='3 beta-hydroxysteroid dehydrogenase 2|3 beta-hydroxysteroid dehydrogenase/Delta 5-->4-isomerase type II|3-beta-HSD II|hydroxysteroid dehydrogenase-2, delta<5>-3-beta', map_location='3 42.85 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98711107, end_position_on_the_genomic_accession=98724594, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15494, CurrentID=0, Status='live', Symbol='Hsd3b3', Aliases='9030618K22Rik, AI790201', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 3', other_designations='3 beta-hydroxysteroid dehydrogenase 3|3 beta-hydroxysteroid dehydrogenase/Delta 5-->4-isomerase type III|3-beta-HSD III|3-beta-hydroxy-5-ene steroid dehydrogenase|delta-5-3-ketosteroid isomerase|hydroxysteroid dehydrogenase-2, delta<5>-3-beta|progesterone reductase|steroid delta-isomerase', map_location='3 42.86 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98741476, end_position_on_the_genomic_accession=98763128, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15495, CurrentID=0, Status='live', Symbol='Hsd3b4', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 4', other_designations='3 beta-hydroxysteroid dehydrogenase 4|3 beta-hydroxysteroid dehydrogenase type IV|3-beta-HSD IV|3-beta-hydroxy-5-ene steroid dehydrogenase|3-beta-hydroxysteroid dehydrogenase/delta-5-delta-4 isomerase|NADPH-dependent 3-beta-hydroxy-Delta(5)-steroid dehydrogenase|hydroxysteroid dehydrogenase-4, delta<5>-3-beta|progesterone reductase', map_location='3 42.79 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98499071, end_position_on_the_genomic_accession=98510567, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15496, CurrentID=0, Status='live', Symbol='Hsd3b5', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 5', other_designations='3 beta-hydroxysteroid dehydrogenase type V|3(beta)HSDV|3-beta-HSD V|3-beta-hydroxy-5-ene steroid dehydrogenase|NADPH-dependent 3-beta-hydroxy-Delta(5)-steroid dehydrogenase|hydroxysteroid dehydrogenase-5, delta<5>-3-beta|progesterone reductase', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98618636, end_position_on_the_genomic_accession=98645534, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15497, CurrentID=0, Status='live', Symbol='Hsd3b6', Aliases='D19210', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 6', other_designations='3 beta-hydroxysteroid dehydrogenase/Delta 5-->4-isomerase type VI|3-beta-HSD VI|3beta-HSD VI|hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 6 pseudogene|hydroxysteroid dehydrogenase-6, delta<5>-3-beta', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98805504, end_position_on_the_genomic_accession=98814443, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15499, CurrentID=0, Status='live', Symbol='Hsf1', Aliases='AA960185alpha, Hsf1beta, Hsf1', description='heat shock factor 1', other_designations='HSF 1|HSTF 1|Hsf1gammaalpha|Hsf1gammabeta|heat shock transcription factor 1 alpha isoform|heat shock transcription factor 1 beta isoform|heat shock transcription factor 1 gammaalpha isoform|heat shock transcription factor 1 gammabeta isoform', map_location='15 35.95 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76477395, end_position_on_the_genomic_accession=76500978, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15500, CurrentID=0, Status='live', Symbol='Hsf2', Aliases='AI661205', description='heat shock factor 2', other_designations='HSF 2|HSTF 2|heat shock transcription factor 2', map_location='10 B3-B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=57486187, end_position_on_the_genomic_accession=57513143, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  15501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15501, CurrentID=0, Status='live', Symbol='hsh', Aliases='', description='hind shaker', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15502, CurrentID=0, Status='live', Symbol='Dnaja1', Aliases='HSJ-2, Hsj2, Nedd7', description='DnaJ heat shock protein family (Hsp40) member A1', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 1|dnaJ protein homolog 2|heat shock 40 kDa protein 4|heat shock protein J2|heat shock protein, Dnaj-like 2|neural precursor cell expressed, developmentally down-regulated gene 7', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40722468, end_position_on_the_genomic_accession=40737153, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  15504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15504, CurrentID=0, Status='live', Symbol='Dnajb3', Aliases='Hsj3, MSJ-1, Msj1', description='DnaJ heat shock protein family (Hsp40) member B3', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 3|HSJ-3|dnaJ protein homolog 3|heat shock protein J3|heat shock protein, DNAJ-like 3', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88204732, end_position_on_the_genomic_accession=88205748, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15505, CurrentID=0, Status='live', Symbol='Hsph1', Aliases='105kDa, AI790491, Hsp105, Hsp110, hsp-E7I, hsp110/105', description='heat shock 105kDa/110kDa protein 1', other_designations='42 degrees C-HSP|HSP105 42 C-HSP|heat shock 110 kDa protein|heat shock protein 110|heat shock protein, 105 kDa|heat shock-related 100 kDa protein E7I', map_location='5 89.18 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149616843, end_position_on_the_genomic_accession=149637155, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  15507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15507, CurrentID=0, Status='live', Symbol='Hspb1', Aliases='27kDa, Hsp25', description='heat shock protein 1', other_designations='HSP 25|HSP 27|growth-related 25 kDa protein|heat shock 25 kDa protein|heat shock 27 kDa protein|heat shock 27kDa protein 1|heat shock protein, 25 kDa|p25|truncated hsp25', map_location='5 75.51 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135887919, end_position_on_the_genomic_accession=135889563, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15508, CurrentID=0, Status='live', Symbol='Hsp25-ps1', Aliases='25kDa, ENSMUSG00000007852', description='heat shock protein 25, pseudogene 1', other_designations='heat shock protein, 25 kDa, pseudogene 1|heat shock protein, pseudogene 1', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=45078370, end_position_on_the_genomic_accession=45079184, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  15510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15510, CurrentID=0, Status='live', Symbol='Hspd1', Aliases='60kDa, Hsp60', description='heat shock protein 1 (chaperonin)', other_designations='60 kDa chaperonin|CPN60|HSP-60|HSP-65|chaperonin 60|heat shock 60kDa protein 1 (chaperonin)|heat shock protein 60|heat shock protein, 60 kDa|mitochondrial matrix protein P1', map_location='1 C1.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55077834, end_position_on_the_genomic_accession=55088112, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  15511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15511, CurrentID=0, Status='live', Symbol='Hspa1b', Aliases='Hsp70, Hsp70-1, Hsp70.1, hsp68', description='heat shock protein 1B', other_designations='68 kDa heat shock protein|heat shock 70 kDa protein 1|heat shock 70kDa protein 1B|heat shock protein 70.1|heat shock protein, 70 kDa 1', map_location='17 18.5 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34956429, end_position_on_the_genomic_accession=34959238, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15512, CurrentID=0, Status='live', Symbol='Hspa2', Aliases='70kDa, HSP70.2, HSP70A2, Hsp70-2', description='heat shock protein 2', other_designations='heat shock 70kDa protein 2|heat shock protein 70.2|heat shock protein, 70 kDa 2', map_location='12 33.73 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76403844, end_position_on_the_genomic_accession=76406938, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15516, CurrentID=0, Status='live', Symbol='Hsp90ab1', Aliases='90kDa, AL022974, C81438, Hsp84, Hsp84-1, Hsp90, Hspcb', description='heat shock protein 90 alpha (cytosolic), class B member 1', other_designations='HSP 84|TSTA|heat shock 84 kDa|heat shock 90kDa protein 1, beta|heat shock protein 1, beta|heat shock protein 90kDa alpha (cytosolic), class B member 1|heat shock protein, 84 kDa 1|retinal degeneration slow protein|tumor-specific transplantation 84 kDa antigen', map_location='17 22.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45567778, end_position_on_the_genomic_accession=45573261, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  15519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15519, CurrentID=0, Status='live', Symbol='Hsp90aa1', Aliases='86kDa, 89kDa, AL024080, AL024147, Hsp86-1, Hsp89, Hsp90, Hspca, hsp4', description='heat shock protein 90, alpha (cytosolic), class A member 1', other_designations='HSP 86|HSP86|TSTA|heat shock 86 kDa|heat shock protein 1, alpha|heat shock protein 90kDa alpha (cytosolic), class A member 1|heat shock protein, 1|heat shock protein, 86 kDa 1|heat shock protein, 89 kDa|tumor-specific transplantation 86 kDa antigen', map_location='12 60.75 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110691036, end_position_on_the_genomic_accession=110696395, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  15525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15525, CurrentID=0, Status='live', Symbol='Hspa4', Aliases='70kDa, AI317151, APG-2, Hsp110, Hsp70RY, mKIAA4025', description='heat shock protein 4', other_designations='heat shock 70-related protein APG-2|heat shock protein, 110 kDa', map_location='11 31.91 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53259814, end_position_on_the_genomic_accession=53300479, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  15526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15526, CurrentID=0, Status='live', Symbol='Hspa9', Aliases='74kDa, Csa, Grp75, Hsc74, Hsp74, Hsp74aa, Mortalin, Mot-2, Mot2, Mthsp70, Pbp74, Hspa9', description='heat shock protein 9', other_designations='75 kDa glucose-regulated protein|C3H-specific antigen|GRP-75|heat shock 70 kDa protein 9|heat shock protein 9A|heat shock protein cognate 74|heat shock protein, 74 kDa, A|heat shock protein, A|p66 MOT|peptide-binding protein 74', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34937414, end_position_on_the_genomic_accession=34954351, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  15527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15527, CurrentID=0, Status='live', Symbol='Hspa9-ps1', Aliases='70kDa, Hsc74-ps1, Hsp74-ps1, Hsp74b, Hspa9b', description='heat shock protein 9, pseudogene 1', other_designations='heat shock 70kD protein 9B (mortalin-2)|heat shock protein cognate 74, pseudogene 1|heat shock protein, 74 kDa, pseudogene 1', map_location='X 33.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=60577758, end_position_on_the_genomic_accession=60579557, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  15528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15528, CurrentID=0, Status='live', Symbol='Hspe1', Aliases='10kDa, Hsp10, mt-cpn10', description='heat shock protein 1 (chaperonin 10)', other_designations='10 kDa chaperonin|CPN10|chaperonin 10|heat shock 10 kDa protein 1 (chaperonin 10)|mitochondrial chaperonin 10', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55088148, end_position_on_the_genomic_accession=55091317, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  15529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15529, CurrentID=0, Status='live', Symbol='Sdc2', Aliases='4833414L08Rik, AA960457, Hspg1, Synd2, syndecan-2', description='syndecan 2', other_designations='HSPG|fibroglycan|heparan sulfate proteoglycan 1, cell surface-associated (fibroglycan)|heparan sulfate proteoglycan core protein', map_location='15 13.74 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=32920723, end_position_on_the_genomic_accession=33034721, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15530, CurrentID=0, Status='live', Symbol='Hspg2', Aliases='AI852380, Pcn, Plc, per', description='perlecan (heparan sulfate proteoglycan 2)', other_designations='HSPG', map_location='4 69.93 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137468764, end_position_on_the_genomic_accession=137570630, orientation='plus', exon_count=98, OMIM='', no_hdr0=''),
  15531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15531, CurrentID=0, Status='live', Symbol='Ndst1', Aliases='1200015G06Rik, Hsst, NDST-1, b2b2230Clo', description='N-deacetylase/N-sulfotransferase (heparan glucosaminyl) 1', other_designations='HSNST 1|N-HSST 1|N-deacetylase/N-sulfotransferase-1|N-heparan sulfate sulfotransferase 1|[Heparan sulfate]-glucosamine N-sulfotransferase 1|glucosaminyl N-deacetylase/N-sulfotransferase 1|heparan sulfate-N-deacetylase/N-sulfotransferase', map_location='18 D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60684494, end_position_on_the_genomic_accession=60748892, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  15547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15547, CurrentID=0, Status='live', Symbol='Trmt2a', Aliases='Htf9c', description='TRM2 tRNA methyltransferase 2A', other_designations='TRM2 tRNA methyltransferase 2 homolog A|hpaII tiny fragments locus 9c protein', map_location='16 11.31 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18248830, end_position_on_the_genomic_accession=18254772, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  15550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15550, CurrentID=0, Status='live', Symbol='Htr1a', Aliases='Gpcr18', description='5-hydroxytryptamine (serotonin) receptor 1A', other_designations='5-HT-1A|5-HT1A receptor|G-protein coupled receptor 18|serotonin receptor 1A', map_location='13 56.92 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=105443693, end_position_on_the_genomic_accession=105448133, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15551, CurrentID=0, Status='live', Symbol='Htr1b', Aliases='', description='5-hydroxytryptamine (serotonin) receptor 1B', other_designations='5-HT-1B|5-HT1B receptor|5-HT<1B> receptor|serotonin receptor 1B', map_location='9 44.61 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=81631392, end_position_on_the_genomic_accession=81632552, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15552, CurrentID=0, Status='live', Symbol='Htr1d', Aliases='5-HT-1D, 5-HT1D, AI853647, Gpcr14b, Htr1d', description='5-hydroxytryptamine (serotonin) receptor 1D', other_designations='5-hydroxytryptamine (serotonin) receptor 1D beta|G-protein coupled receptor 14|serotonin receptor 1D', map_location='4 68.74 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136423522, end_position_on_the_genomic_accession=136444405, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15553, CurrentID=0, Status='live', Symbol='Htr1da', Aliases='', description='5-hydroxytryptamine (serotonin) receptor 1D alpha', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100416706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100416706, CurrentID=0, Status='live', Symbol='Zfp729b', Aliases='AA987127', description='zinc finger protein 729b', other_designations='Novel C2H2 type zinc finger protein (AA987161)', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67589439, end_position_on_the_genomic_accession=67609707, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15557, CurrentID=0, Status='live', Symbol='Htr1f', Aliases='Htr1eb', description='5-hydroxytryptamine (serotonin) receptor 1F', other_designations='5-HT-1E-beta|5-HT-1F|5-HT1F|5-hydroxytryptamine (serotonin) receptor 1E beta|serotonin receptor 1F', map_location='16 37.1 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=64922423, end_position_on_the_genomic_accession=65105784, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15558, CurrentID=0, Status='live', Symbol='Htr2a', Aliases='E030013E04, Htr-2, Htr2', description='5-hydroxytryptamine (serotonin) receptor 2A', other_designations='5-HT-2|5-HT-2A|5-HT2A receptor|5-hydroxytryptamine 2 receptor|serotonin receptor 2A', map_location='14 39.37 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=74640840, end_position_on_the_genomic_accession=74706859, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15559, CurrentID=0, Status='live', Symbol='Htr2b', Aliases='5-HT2B, AJ012488, AV377389', description='5-hydroxytryptamine (serotonin) receptor 2B', other_designations='5-HT-2B|5-HT-2F|NP75 protein|serotonin receptor 2B', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86097546, end_position_on_the_genomic_accession=86113662, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15560, CurrentID=0, Status='live', Symbol='Htr2c', Aliases='5-HT2C, 5-HT2cR, 5-HTR2C, 5HT1c, Htr1c, SR1', description='5-hydroxytryptamine (serotonin) receptor 2C', other_designations='5-HT-1C|5-HT-2C|5-HT1C|5-HT2C receptor|5-hydroxytryptamine receptor 1C|serotonin receptor 2C', map_location='X D-F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=146962513, end_position_on_the_genomic_accession=147197277, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15561, CurrentID=0, Status='live', Symbol='Htr3a', Aliases='5-HT3', description='5-hydroxytryptamine (serotonin) receptor 3A', other_designations='5-HT-3|5-HT3-A|5-HT3A|5-HT3R|5-hydroxytryptamine receptor 3|5HT3 serotonin receptor|serotonin receptor 3A|serotonin-gated ion channel receptor', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48899213, end_position_on_the_genomic_accession=48911151, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  15562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15562, CurrentID=0, Status='live', Symbol='Htr4', Aliases='5-HT-4, 5-HT4, 5-HT<4L>, 5HTR4', description='5 hydroxytryptamine (serotonin) receptor 4', other_designations='serotonin receptor 4', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=62322160, end_position_on_the_genomic_accession=62498026, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  15563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15563, CurrentID=0, Status='live', Symbol='Htr5a', Aliases='Htr5', description='5-hydroxytryptamine (serotonin) receptor 5A', other_designations='5-HT-5|5-HT-5A|5-HT5A|serotonin receptor 5A', map_location='5 A3-B', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=27841702, end_position_on_the_genomic_accession=27855086, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15564, CurrentID=0, Status='live', Symbol='Htr5b', Aliases='5-Ht5b', description='5-hydroxytryptamine (serotonin) receptor 5B', other_designations='5-HT-5B|protein-tyrosine-phosphatase (EC 3.1.3.48) Ly-5 precursor (B-cell variant)|serotonin receptor 5B', map_location='1 E4-G', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=121509758, end_position_on_the_genomic_accession=121529312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15565, CurrentID=0, Status='live', Symbol='Htr6', Aliases='5-HT6', description='5-hydroxytryptamine (serotonin) receptor 6', other_designations='5-HT-6|serotonin receptor 6', map_location='4 70.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139059705, end_position_on_the_genomic_accession=139075519, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15566, CurrentID=0, Status='live', Symbol='Htr7', Aliases='5-HT7', description='5-hydroxytryptamine (serotonin) receptor 7', other_designations='5-HT-7|5-HT-X|serotonin receptor 7', map_location='19 30.3 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=35958415, end_position_on_the_genomic_accession=36058286, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15567, CurrentID=0, Status='live', Symbol='Slc6a4', Aliases='5-HTT, AI323329, Htt, Sert', description='solute carrier family 6 (neurotransmitter transporter, serotonin), member 4', other_designations='5-hydroxytryptamine (serotonin) transporter|5HT transporter|5HTT|Sodium-dependent serotonin transporter (5HT transporter) (5HTT)|solute carrier family 6 member 4', map_location='11 46.18 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76998466, end_position_on_the_genomic_accession=77032345, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  15568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15568, CurrentID=0, Status='live', Symbol='Elavl1', Aliases='2410055N02Rik, HUR, Hua, W91709', description='ELAV (embryonic lethal, abnormal vision)-like 1 (Hu antigen R)', other_designations='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 1 (Hu antigen R)|HU-antigen A|elav-like generic protein|hu-antigen R|melG', map_location='8 2.0 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4284781, end_position_on_the_genomic_accession=4325100, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15569, CurrentID=0, Status='live', Symbol='Elavl2', Aliases='Hub, mel-N1', description='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 2 (Hu antigen B)', other_designations='ELAV-like neuronal protein 1|HU-antigen B|nervous system-specific RNA-binding protein Mel-N1', map_location='4 42.36 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=91250766, end_position_on_the_genomic_accession=91400064, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15570, CurrentID=0, Status='live', Symbol='hubb', Aliases='hub, jaundice', description='hyper-unconjugated bilirubinemia', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  15571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15571, CurrentID=0, Status='live', Symbol='Elavl3', Aliases='2600009P04Rik, Huc, PLE21, mHuC', description='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 3 (Hu antigen C)', other_designations='HU-antigen C', map_location='9 8.07 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22015005, end_position_on_the_genomic_accession=22052023, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  15572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15572, CurrentID=0, Status='live', Symbol='Elavl4', Aliases='Elav, Hud, PNEM', description='ELAV (embryonic lethal, abnormal vision, Drosophila)-like 4 (Hu antigen D)', other_designations='HU-antigen D|RNA-binding protein HUD3|paraneoplastic encephalomyelitis antigen HuD', map_location='4 51.42 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=110203737, end_position_on_the_genomic_accession=110351946, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  15574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15574, CurrentID=0, Status='live', Symbol='Hus1', Aliases='', description='HUS1 checkpoint clamp component', other_designations='Hus1 homolog', map_location='11 5.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=8993137, end_position_on_the_genomic_accession=9011534, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68133, CurrentID=0, Status='live', Symbol='Gcsh', Aliases='1100001L02Rik, 5730591C18Rik', description='glycine cleavage system protein H (aminomethyl carrier)', other_designations='lipoic acid-containing protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116981828, end_position_on_the_genomic_accession=116993449, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  15586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15586, CurrentID=0, Status='live', Symbol='Hyal1', Aliases='Hya1, Hyal-1', description='hyaluronoglucosaminidase 1', other_designations='hyaluronidase 1|hyaluronoglucosaminidase-1', map_location='9 F1-F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107576952, end_position_on_the_genomic_accession=107580137, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  15587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15587, CurrentID=0, Status='live', Symbol='Hyal2', Aliases='AI256841, AU020858, AW555733', description='hyaluronoglucosaminidase 2', other_designations='hyal-2|hyaluronidase 2|hyaluronoglucosaminidase-2', map_location='9 58.12 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107567888, end_position_on_the_genomic_accession=107572778, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  277744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277744, CurrentID=0, Status='live', Symbol='Gm694', Aliases='', description='predicted gene 694', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141432672, end_position_on_the_genomic_accession=141436105, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  277753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277753, CurrentID=0, Status='live', Symbol='Cyp4a12a', Aliases='BC025936, Cyp4a12', description='cytochrome P450, family 4, subfamily a, polypeptide 12a', other_designations='CYPIVA12|cytochrome P450, family 4, subfamily a, polypeptide 12', map_location='4 52.91 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115299046, end_position_on_the_genomic_accession=115332815, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  277773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277773, CurrentID=0, Status='live', Symbol='Fam205c', Aliases='4931430D02, FAM205CP', description='family with sequence similarity 205, member C', other_designations='protein FAM205CP homolog|transmembrane protein ENSP00000340100 homolog', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42868003, end_position_on_the_genomic_accession=42874838, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  330286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330286, CurrentID=0, Status='live', Symbol='D630045J12Rik', Aliases='Kiaa1549, mKIAA1549', description='RIKEN cDNA D630045J12 gene', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38123174, end_position_on_the_genomic_accession=38254009, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  277854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277854, CurrentID=0, Status='live', Symbol='Depdc5', Aliases='AV016528', description='DEP domain containing 5', other_designations='', map_location='5 17.35 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=32863647, end_position_on_the_genomic_accession=32994233, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  100416868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100416868, CurrentID=0, Status='live', Symbol='Pgk1-ps3', Aliases='Pgk-1ps3', description='phosphoglycerate kinase 1, pseudogene 3', other_designations='phosphoglycerate kinase 1 pseudogene', map_location='3 5.16 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=18564493, end_position_on_the_genomic_accession=18566138, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  68157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68157, CurrentID=0, Status='live', Symbol='6720475J19Rik', Aliases='', description='RIKEN cDNA 6720475J19 gene', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.5', start_position_on_the_genomic_accession=9910228, end_position_on_the_genomic_accession=9911661, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  277898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277898, CurrentID=0, Status='live', Symbol='Slc15a5', Aliases='9830102E05Rik', description='solute carrier family 15, member 5', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=137983590, end_position_on_the_genomic_accession=138079916, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  277939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277939, CurrentID=0, Status='live', Symbol='C2cd3', Aliases='AU020772', description='C2 calcium-dependent domain containing 3', other_designations='protein hearty', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100372233, end_position_on_the_genomic_accession=100470159, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  100416952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100416952, CurrentID=0, Status='live', Symbol='Hmgb1-ps6', Aliases='Gm15695, Hmg-1RI-A, Hmg-F, Hmg1-rs7, Hmgb1-rs7, OTTMUSG00000025338', description='high mobility group box 1, pseudogene 6', other_designations='high mobility group box 1, related sequence 7|high mobility group box 2 pseudogene', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31991253, end_position_on_the_genomic_accession=31991866, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  277973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277973, CurrentID=0, Status='live', Symbol='Slc9a5', Aliases='Gm696', description='solute carrier family 9 (sodium/hydrogen exchanger), member 5', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105348104, end_position_on_the_genomic_accession=105369881, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  277978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277978, CurrentID=0, Status='live', Symbol='Exoc3l', Aliases='C730015A04Rik1, Exoc3l', description='exocyst complex component 3-like', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105289923, end_position_on_the_genomic_accession=105296962, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  671232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=671232, CurrentID=0, Status='live', Symbol='Topaz1', Aliases='Gm9524', description='testis and ovary specific PAZ domain containing 1', other_designations='testis and ovary-specific PAZ domain gene 1|topaz 1', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122747217, end_position_on_the_genomic_accession=122802074, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  15874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15874, CurrentID=0, Status='live', Symbol='Iapp', Aliases='DAP', description='islet amyloid polypeptide', other_designations='amylin|diabetes-associated peptide', map_location='6 73.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142298425, end_position_on_the_genomic_accession=142303820, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15891, CurrentID=0, Status='live', Symbol='Ibsp', Aliases='BSP, BSP II, BSPII, Bsp2', description='integrin binding sialoprotein', other_designations='bone sialoprotein II|cell-binding sialoprotein|integrin-binding sialoprotein', map_location='5 50.68 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104299287, end_position_on_the_genomic_accession=104311472, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15893, CurrentID=0, Status='live', Symbol='Ica1', Aliases='69kDa, ICA69', description='islet cell autoantigen 1', other_designations='69 kDa islet cell autoantigen|ICAp69|ica1|islet cell autoantigen 1, 69 kDa|islet cell autoantigen p69|p69', map_location='6 A1-A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=8630527, end_position_on_the_genomic_accession=8778537, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  15894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15894, CurrentID=0, Status='live', Symbol='Icam1', Aliases='CD54, Icam-1, Ly-47, MALA-2', description='intercellular adhesion molecule 1', other_designations='myD10', map_location='9 7.69 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21015960, end_position_on_the_genomic_accession=21028797, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15896, CurrentID=0, Status='live', Symbol='Icam2', Aliases='CD102, Icam-2, Ly-60', description='intercellular adhesion molecule 2', other_designations='lymphocyte function-associated AG-1 counter-receptor', map_location='11 69.09 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106377656, end_position_on_the_genomic_accession=106388122, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15898, CurrentID=0, Status='live', Symbol='Icam5', Aliases='CD50, ICAM-3, Icam3, TLN, Tlcn', description='intercellular adhesion molecule 5, telencephalin', other_designations='ICAM-5|intercellular adhesion molecule 3|telencephalin', map_location='9 7.7 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21032038, end_position_on_the_genomic_accession=21039036, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  15900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15900, CurrentID=0, Status='live', Symbol='Irf8', Aliases='AI893568, ICSBP, IRF-8, Icsbp1, Myls', description='interferon regulatory factor 8', other_designations='interferon consensus sequence binding protein 1|interferon consensus sequence-binding protein', map_location='8 70.05 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120736358, end_position_on_the_genomic_accession=120756694, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  15901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15901, CurrentID=0, Status='live', Symbol='Id1', Aliases='AI323524, D2Wsu140e, Idb1, bHLHb24', description='inhibitor of DNA binding 1', other_designations='inhibitor of differentiation 1', map_location='2 75.41 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152736237, end_position_on_the_genomic_accession=152737410, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15902, CurrentID=0, Status='live', Symbol='Id2', Aliases='AI255428, C78922, Idb2, bHLHb26', description='inhibitor of DNA binding 2', other_designations='inhibitor of differentiation 2', map_location='12 B', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=25093799, end_position_on_the_genomic_accession=25096092, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  15903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15903, CurrentID=0, Status='live', Symbol='Id3', Aliases='Hlh462, Idb3, bHLHb25', description='inhibitor of DNA binding 3', other_designations='ID-like protein inhibitor HLH 462|inhibitor of differentiation 3', map_location='4 68.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136143822, end_position_on_the_genomic_accession=136145392, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15904, CurrentID=0, Status='live', Symbol='Id4', Aliases='Idb4, bHLHb27', description='inhibitor of DNA binding 4', other_designations='inhibitor of differentiation 4', map_location='13 B', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48261427, end_position_on_the_genomic_accession=48264036, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  15925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15925, CurrentID=0, Status='live', Symbol='Ide', Aliases='1300012G03Rik, 4833415K22Rik, AA675336, AI507533', description='insulin degrading enzyme', other_designations='insulin protease|insulinase|insulysin', map_location='19 32.24 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37268741, end_position_on_the_genomic_accession=37334544, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  15926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15926, CurrentID=0, Status='live', Symbol='Idh1', Aliases='AI314845, AI788952, E030024J03Rik, Id-1, Idh-1, Idpc', description='isocitrate dehydrogenase 1 (NADP+), soluble', other_designations='IDH|IDP|NADP(+)-specific ICDH|cytosolic NADP-isocitrate dehydrogenase|oxalosuccinate decarboxylase', map_location='1 32.91 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65158616, end_position_on_the_genomic_accession=65186479, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  15929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15929, CurrentID=0, Status='live', Symbol='Idh3g', Aliases='', description='isocitrate dehydrogenase 3 (NAD+), gamma', other_designations='NAD(+)-specific ICDH subunit gamma|isocitrate dehydrogenase [NAD] subunit gamma, mitochondrial|isocitric dehydrogenase subunit gamma', map_location='X A7.3-B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73778963, end_position_on_the_genomic_accession=73786953, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  15930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15930, CurrentID=0, Status='live', Symbol='Ido1', Aliases='Ido, Indo', description='indoleamine 2,3-dioxygenase 1', other_designations='IDO-1|indole 2,3-dioxygenase|indoleamine-pyrrole 2,3 dioxygenase|indoleamine-pyrrole 2,3-dioxygenase', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24584133, end_position_on_the_genomic_accession=24596952, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  15931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15931, CurrentID=0, Status='live', Symbol='Ids', Aliases='AW214631', description='iduronate 2-sulfatase', other_designations='alpha-L-iduronate sulfate sulfatase|iduronate sulfatase|iduronate-2-sulfatase', map_location='X 36.01 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=70343069, end_position_on_the_genomic_accession=70365114, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  15932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15932, CurrentID=0, Status='live', Symbol='Idua', Aliases='6030426D08', description='iduronidase, alpha-L-', other_designations='', map_location='5 53.24 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108660331, end_position_on_the_genomic_accession=108685446, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  15936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15936, CurrentID=0, Status='live', Symbol='Ier2', Aliases='AI317238, Ch1, Pip92', description='immediate early response 2', other_designations='CHX1|T-lymphocyte-activated protein|cycloheximide-induced gene protein|growth factor-inducible immediate early protein|proline-rich-induced protein 92', map_location='8 41.02 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84661331, end_position_on_the_genomic_accession=84662852, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15937, CurrentID=0, Status='live', Symbol='Ier3', Aliases='AI663993, IEX-1, cI-3, gly96', description='immediate early response 3', other_designations='cAMP inducible gene 3|immediate early protein GLY96|immediate early response 3 protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35821713, end_position_on_the_genomic_accession=35822911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15939, CurrentID=0, Status='live', Symbol='Ier5', Aliases='', description='immediate early response 5', other_designations='', map_location='1 66.45 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155096367, end_position_on_the_genomic_accession=155099636, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15944, CurrentID=0, Status='live', Symbol='Irgm1', Aliases='Ifggd3, Ifi1, Iigp3, Iipg3, Irgm, LRG-47', description='immunity-related GTPase family M member 1', other_designations='LPS-stimulated RAW 264.7 macrophage protein 47|immunity related GTPase m1|immunity-related GTPase family, M|interferon inducible protein 1|interferon-gamma-inducible GTPase Ifggd3 protein|interferon-inducible GTPase 3|interferon-inducible protein 1', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48865245, end_position_on_the_genomic_accession=48871415, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15945, CurrentID=0, Status='live', Symbol='Cxcl10', Aliases='C7, CRG-2, INP10, IP-10, IP10, Ifi10, Scyb10, gIP-10, mob-1', description='chemokine (C-X-C motif) ligand 10', other_designations='10 kDa interferon gamma-induced protein|gamma-IP10|interferon activated gene 10|interferon-gamma induced protein CRG-2|interferon-gamma-induced protein CRG-2|small inducible cytokine B subfamily (Cys-X-Cys), member 10|small-inducible cytokine B10', map_location='5 46.57 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92346638, end_position_on_the_genomic_accession=92348889, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  15950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15950, CurrentID=0, Status='live', Symbol='Ifi203', Aliases='', description='interferon activated gene 203', other_designations='interferon-inducible protein p203', map_location='1 80.76 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173920400, end_position_on_the_genomic_accession=173942847, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15951, CurrentID=0, Status='live', Symbol='Ifi204', Aliases='Ifi16, p204', description='interferon activated gene 204', other_designations='ifi-204|interferon, gamma-inducible gene 204|interferon, gamma-inducible protein 16|interferon-inducible protein p204', map_location='1 80.63 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173747293, end_position_on_the_genomic_accession=173767068, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  278097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278097, CurrentID=0, Status='live', Symbol='Armcx6', Aliases='AW060994', description='armadillo repeat containing, X-linked 6', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134748447, end_position_on_the_genomic_accession=134756342, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  15957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15957, CurrentID=0, Status='live', Symbol='Ifit1', Aliases='ISG56, Ifi56, P56', description='interferon-induced protein with tetratricopeptide repeats 1', other_designations='GARG-16|IFI-56K|IFIT-1|glucocorticoid-attenuated response gene 16 protein|interferon-induced 56 kDa protein', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34640889, end_position_on_the_genomic_accession=34650009, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15958, CurrentID=0, Status='live', Symbol='Ifit2', Aliases='AV302338, Ifi54, P54', description='interferon-induced protein with tetratricopeptide repeats 2', other_designations='GARG-39|IFI-54K|IFIT-2|glucocorticoid-attenuated response gene 39 protein|interferon-induced 54 kDa protein', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34550549, end_position_on_the_genomic_accession=34576534, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  15959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15959, CurrentID=0, Status='live', Symbol='Ifit3', Aliases='Ifi49, P49', description='interferon-induced protein with tetratricopeptide repeats 3', other_designations='GARG-49|IFIT-3|IRG2|glucocorticoid-attenuated response gene 49 protein', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34583142, end_position_on_the_genomic_accession=34588982, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15962, CurrentID=0, Status='live', Symbol='Ifna1', Aliases='Ifa1', description='interferon alpha 1', other_designations='IFN-alpha-1|interferon alpha family gene 1', map_location='4 42.04 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88850087, end_position_on_the_genomic_accession=88850656, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15964, CurrentID=0, Status='live', Symbol='Ifna11', Aliases='IFN-[a]11', description='interferon alpha 11', other_designations='IFN-alpha-11|interferon alpha family, gene 11|limitin', map_location='4 42.03 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88819919, end_position_on_the_genomic_accession=88820565, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15965, CurrentID=0, Status='live', Symbol='Ifna2', Aliases='Ifa2', description='interferon alpha 2', other_designations='IFN-alpha-2|interferon alpha family, gene 2', map_location='4 42.02 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88683207, end_position_on_the_genomic_accession=88683779, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15967, CurrentID=0, Status='live', Symbol='Ifna4', Aliases='Ifa4', description='interferon alpha 4', other_designations='IFN-alpha-4|MuIFN-alpha 4|interferon alpha family, gene 4', map_location='4 42.03 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88841816, end_position_on_the_genomic_accession=88842459, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15968, CurrentID=0, Status='live', Symbol='Ifna5', Aliases='Ifa5', description='interferon alpha 5', other_designations='IFN-alpha-5|interferon alpha family, gene 5', map_location='4 42.03 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88835525, end_position_on_the_genomic_accession=88836094, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15969, CurrentID=0, Status='live', Symbol='Ifna6', Aliases='IFNa8/6, Ifa6, Ifa8, Ifna8', description='interferon alpha 6', other_designations='IFN-alpha-6|IFN-alpha-8|interferon alpha 8/6|interferon alpha family, gene 6|interferon alpha family, gene 8|interferon alpha-8', map_location='4 42.03 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88827416, end_position_on_the_genomic_accession=88827985, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15970, CurrentID=0, Status='live', Symbol='Ifna7', Aliases='Ifa7', description='interferon alpha 7', other_designations='IFN-alpha-7|interferon alpha family, gene 7', map_location='4 42.03 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88816225, end_position_on_the_genomic_accession=88816800, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15972, CurrentID=0, Status='live', Symbol='Ifna9', Aliases='Ifa9', description='interferon alpha 9', other_designations='IFN-alpha-9|interferon alpha family, gene 9', map_location='4 42.01 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88591813, end_position_on_the_genomic_accession=88592385, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15974, CurrentID=0, Status='live', Symbol='Ifnab', Aliases='', description='interferon alpha B', other_designations='', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88690599, end_position_on_the_genomic_accession=88691268, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15975, CurrentID=0, Status='live', Symbol='Ifnar1', Aliases='CD118, Ifar, Ifnar, Ifrc, Infar', description='interferon (alpha and beta) receptor 1', other_designations='IFN-R-1|IFN-alpha/beta receptor 1|IFN-alpha/betaR|INF-a receptor|interferon-alpha/beta receptor 1|type I interferon receptor 1', map_location='16 52.98 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91485087, end_position_on_the_genomic_accession=91510523, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  15976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15976, CurrentID=0, Status='live', Symbol='Ifnar2', Aliases='AI747302, Ifnar-2', description='interferon (alpha and beta) receptor 2', other_designations="IFN-R-2|IFN-alpha/beta receptor 2|interferon receptor 2a'|interferon receptor 2c|interferon-alpha/beta receptor 2|type I interferon alpha/beta receptor 2a|type I interferon receptor 2", map_location='16 52.82 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91372215, end_position_on_the_genomic_accession=91405589, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  15977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15977, CurrentID=0, Status='live', Symbol='Ifnb1', Aliases='IFN-beta, IFNB, Ifb', description='interferon beta 1, fibroblast', other_designations='', map_location='4 41.91 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88522025, end_position_on_the_genomic_accession=88522794, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  15978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15978, CurrentID=0, Status='live', Symbol='Ifng', Aliases='IFN-g, Ifg', description='interferon gamma', other_designations='IFN-gamma|gamma interferon', map_location='10 66.75 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=118441047, end_position_on_the_genomic_accession=118445892, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  15979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15979, CurrentID=0, Status='live', Symbol='Ifngr1', Aliases='CD119, IFN-gammaR, Ifgr, Ifngr, Nktar', description='interferon gamma receptor 1', other_designations='IFN-gamma receptor 1|IFN-gamma-R1|INF-g receptor|interferon-gamma receptor 1', map_location='10 8.49 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19591958, end_position_on_the_genomic_accession=19610225, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15980, CurrentID=0, Status='live', Symbol='Ifngr2', Aliases='Ifgr2, Ifgt', description='interferon gamma receptor 2', other_designations='', map_location='16 53.07 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91547094, end_position_on_the_genomic_accession=91564007, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  15982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15982, CurrentID=0, Status='live', Symbol='Ifrd1', Aliases='Ifnl, PC4, Tis7', description='interferon-related developmental regulator 1', other_designations='TPA-induced sequence 7|nerve growth factor-inducible protein PC4', map_location='12 18.06 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=40203129, end_position_on_the_genomic_accession=40223189, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15983, CurrentID=0, Status='live', Symbol='Ifrd2', Aliases='1810034A24Rik, AA470234, AI838527, SKMc15', description='interferon-related developmental regulator 2', other_designations='', map_location='9 58.24 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107587570, end_position_on_the_genomic_accession=107593038, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  15985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15985, CurrentID=0, Status='live', Symbol='Cd79b', Aliases='B29, Ig-beta, Igb, Igbeta', description='CD79B antigen', other_designations='B-cell-specific glycoprotein B29|immunoglobulin-associated B29 protein|immunoglobulin-associated beta', map_location='11 68.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106311341, end_position_on_the_genomic_accession=106314529, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  57271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57271, CurrentID=0, Status='live', Symbol='Olfr1509', Aliases='MOR244-3, Mor83, Or83', description='olfactory receptor 1509', other_designations='GA_x6K02T2RJGY-534312-533386|gene for odorant receptor MOR83|olfactory receptor MOR244-3', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52450393, end_position_on_the_genomic_accession=52451402, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  16000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16000, CurrentID=0, Status='live', Symbol='Igf1', Aliases='C730016P09Rik, Igf-1, Igf-I', description='insulin-like growth factor 1', other_designations='somatomedin', map_location='10 43.7 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=87858270, end_position_on_the_genomic_accession=87937047, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  16001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16001, CurrentID=0, Status='live', Symbol='Igf1r', Aliases='A330103N21Rik, CD221, D930020L01, IGF-1R, hyft', description='insulin-like growth factor I receptor', other_designations='IGF-I receptor|line 186', map_location='7 37.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=67952257, end_position_on_the_genomic_accession=68233668, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  16002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16002, CurrentID=0, Status='live', Symbol='Igf2', Aliases='AL033362, Igf-2, Igf-II, M6pr, Mpr, Peg2', description='insulin-like growth factor 2', other_designations='multiplication-stimulating polypeptide', map_location='7 87.99 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142650768, end_position_on_the_genomic_accession=142662270, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16004, CurrentID=0, Status='live', Symbol='Igf2r', Aliases='AI661837, CD222, CI-MPR, M6P/IGF2R, Mpr300', description='insulin-like growth factor 2 receptor', other_designations='300 kDa mannose 6-phosphate receptor|CI Man-6-P receptor|IGF-II receptor|IGF-II/CI-MPR|M6P/IGF2 receptor|M6PR|MPR 300|cation-independent MPR|cation-independent mannose 6-phosphate receptor|insulin-like growth factor II receptor|mannose-6-phosphate receptor, cation independent', map_location='17 A-C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12682406, end_position_on_the_genomic_accession=12769706, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  16005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16005, CurrentID=0, Status='live', Symbol='Igfals', Aliases='ALS, Albs, mKIAA4111', description='insulin-like growth factor binding protein, acid labile subunit', other_designations='', map_location='17 A2-A3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24877653, end_position_on_the_genomic_accession=24882011, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16006, CurrentID=0, Status='live', Symbol='Igfbp1', Aliases='', description='insulin-like growth factor binding protein 1', other_designations='IBP-1|IGF-binding protein 1|INSULIN-LIKE GROWTH FACTOR BINDING PROTEIN 1 PRECURSOR (IGFBP-1) (IBP-1) (IGF-BINDING PROTEIN 1)', map_location='11 4.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=7197787, end_position_on_the_genomic_accession=7202546, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16007, CurrentID=0, Status='live', Symbol='Cyr61', Aliases='AI325051, CCN1, Igfbp10', description='cysteine rich protein 61', other_designations='3CH61|CCN family member 1|IBP-10|IGF-binding protein 10|IGFBP-10|cysteine-rich angiogenic inducer 61|insulin-like growth factor binding protein 10|insulin-like growth factor-binding protein 10|intermediate early', map_location='3 70.18 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145646971, end_position_on_the_genomic_accession=145649985, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16008, CurrentID=0, Status='live', Symbol='Igfbp2', Aliases='AI255832, IBP-2, Igfbp-2, mIGFBP-2', description='insulin-like growth factor binding protein 2', other_designations='IGF-binding protein 2', map_location='1 36.94 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72824480, end_position_on_the_genomic_accession=72852474, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16009, CurrentID=0, Status='live', Symbol='Igfbp3', Aliases='AI649005, IGFBP-3, IGgfbp3', description='insulin-like growth factor binding protein 3', other_designations='IBP-3|IGF-binding protein 3', map_location='11 4.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=7206086, end_position_on_the_genomic_accession=7213931, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16010, CurrentID=0, Status='live', Symbol='Igfbp4', Aliases='AI875747, Deb2, IGFBP-4', description='insulin-like growth factor binding protein 4', other_designations='Band 3A|IBP-4|IGF-binding protein 4|differentially expressed in B16F10 2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99041260, end_position_on_the_genomic_accession=99052646, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16011, CurrentID=0, Status='live', Symbol='Igfbp5', Aliases='AI256729, AW208790, IGFBP-5, IGFBP-5P', description='insulin-like growth factor binding protein 5', other_designations='IBP-5|IGF-binding protein 5', map_location='1 36.94 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72858065, end_position_on_the_genomic_accession=72874865, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16012, CurrentID=0, Status='live', Symbol='Igfbp6', Aliases='IGFBP-6', description='insulin-like growth factor binding protein 6', other_designations='IBP-6|IGF-binding protein 6|INSULIN-LIKE GROWTH FACTOR BINDING PROTEIN 6 PRECURSOR (IGFBP-6) (IBP-6) (IGF-BINDING PROTEIN 6)', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102144186, end_position_on_the_genomic_accession=102149512, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16016, CurrentID=0, Status='live', Symbol='Ighg2b', Aliases='IgG2b, Igh-3, gamma2b', description='immunoglobulin heavy constant gamma 2B', other_designations='immunoglobulin heavy chain 3 (serum IgG2b)', map_location='12 62.1 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113304314, end_position_on_the_genomic_accession=113307933, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  16017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16017, CurrentID=0, Status='live', Symbol='Ighg1', Aliases='IgG1, Igh-4, VH7183', description='immunoglobulin heavy constant gamma 1 (G1m marker)', other_designations='immunoglobulin heavy chain 4 (serum IgG1)|immunoglobulin heavy chain V region', map_location='12 62.1 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113326544, end_position_on_the_genomic_accession=113330523, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  16019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16019, CurrentID=0, Status='live', Symbol='Ighm', Aliases='AI326478, Igh-6, Igh-M, Igh6, Igm, muH', description='immunoglobulin heavy constant mu', other_designations='Ig mu|immunoglobulin heavy chain 6 (heavy chain of IgM)|mu-immunoglobulin', map_location='12 F1-2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113418826, end_position_on_the_genomic_accession=113422730, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  16020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16020, CurrentID=0, Status='live', Symbol='Igh-7', Aliases='', description='immunoglobulin heavy chain 7 (heavy chain of IgE)', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16021, CurrentID=0, Status='live', Symbol='Igh-8', Aliases='', description='immunoglobulin heavy chain 8 (heavy chain of IgG3)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16031, CurrentID=0, Status='live', Symbol='Igh-Ia', Aliases='ENSMUSG00000078149, Igh2', description='immunoglobulin heavy chain Ia', other_designations='Ig germline alpha-chain gene C-switch region', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16033, CurrentID=0, Status='live', Symbol='Igh-J', Aliases='Jh', description='immunoglobulin heavy chain, joining region', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  278180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278180, CurrentID=0, Status='live', Symbol='Vsig4', Aliases='A530061A11, BC025105, CRIg, Z39IG', description='V-set and immunoglobulin domain containing 4', other_designations='complement receptor of the immunoglobulin superfamily', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=96247201, end_position_on_the_genomic_accession=96293518, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16043, CurrentID=0, Status='live', Symbol='Igh-Sa1', Aliases='', description='immunoglobulin heavy chain Sa1', other_designations='alpha switch region', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16049, CurrentID=0, Status='live', Symbol='Igh-V', Aliases='B1H12, B4H2, Gal13, IGHV2B, M86', description='immunoglobulin heavy chain variable region', other_designations='pot. H26-1 variable region', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16050, CurrentID=0, Status='live', Symbol='Igh-V10', Aliases='', description='immunoglobulin heavy chain (V10 family)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16051, CurrentID=0, Status='live', Symbol='Igh-V11', Aliases='', description='immunoglobulin heavy chain (V11 family)', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16057, CurrentID=0, Status='live', Symbol='Igh-V3609N', Aliases='', description='immunoglobulin heavy chain (V3609N non-productive)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='NT_096355.1', start_position_on_the_genomic_accession=1483, end_position_on_the_genomic_accession=17432, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  16058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16058, CurrentID=0, Status='live', Symbol='Igh-V3660', Aliases='', description='immunoglobulin heavy chain (V3660 family)', other_designations='immunoglobulin mu-chain', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16059, CurrentID=0, Status='live', Symbol='Igh-V7183', Aliases='B9-scFv, IgG, IgH, IgVH1(VSG), VH7183, VI24H', description='immunoglobulin heavy chain (V7183 family)', other_designations='Ig heavy chain V-region|IgG heavy chain|IgG1 H chain|anti-DNA immunoglobulin heavy chain IgG|anti-glycophorin A type M Ig heavy chain|antibody heavy chain variable region|immunoglobulin gamma-chain|immunoglobulin heavy chain V region|immunoglobulin heavy chain variable region 37.1|immunoglobulin heavy chain variable region 7183.14|monoclonal antibody 3G-10 H chain|polyreactive autoantibody, immunoglobulin IgM heavy chain|trypanosome variant surface glycoprotein specific immunoglobulin VH1', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16061, CurrentID=0, Status='live', Symbol='Igh-VJ558', Aliases='AI893585', description='immunoglobulin heavy chain (J558 family)', other_designations='Igh-A (1g2)|Vh186.2/Jh2|immunoglobulin heavy chain variable region', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16062, CurrentID=0, Status='live', Symbol='Igh-VJ606', Aliases='', description='immunoglobulin heavy chain (J606 family)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16065, CurrentID=0, Status='live', Symbol='Igh-VS107', Aliases='IgG, IgH, VhNZA6', description='immunoglobulin heavy chain (S107 family)', other_designations='IgMk heavy chain variable region|VH coding region|anti-poly(dC) monoclonal antibody heavy chain|immunoglobulin variable region', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16068, CurrentID=0, Status='live', Symbol='Il18bp', Aliases='IL-18BP, Igifbp, MC54L', description='interleukin 18 binding protein', other_designations='interferon gamma inducing factor binding protein|interferon gamma-inducing factor-binding protein', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102015077, end_position_on_the_genomic_accession=102019041, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16069, CurrentID=0, Status='live', Symbol='Jchain', Aliases='9530090F24Rik, AI323815, Igj, Jch', description='immunoglobulin joining chain', other_designations='J chain|immunoglobulin joining gene (not J region)', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88519809, end_position_on_the_genomic_accession=88527897, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16071, CurrentID=0, Status='live', Symbol='Igkc', Aliases='Igk-C', description='immunoglobulin kappa constant', other_designations='If kappa light chain|immunoglobulin kappa chain, constant region|immunoglobulin variable region|kappa light chain C_region', map_location='6 C', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70726435, end_position_on_the_genomic_accession=70726754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  16073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16073, CurrentID=0, Status='live', Symbol='Igk-J', Aliases='', description='immunoglobulin kappa chain, joining region', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16080, CurrentID=0, Status='live', Symbol='Igk-V', Aliases='', description='immunoglobulin kappa chain complex variable region', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.5', start_position_on_the_genomic_accession=43029303, end_position_on_the_genomic_accession=43032096, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  16081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16081, CurrentID=0, Status='live', Symbol='Igk-V1', Aliases='', description='immunoglobulin kappa chain variable 1 (V1)', other_designations='IgG1/kappa antibody', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16087, CurrentID=0, Status='live', Symbol='Igk-V12/13', Aliases='IGKV12-e, Igk-V12, Igk-V13', description='immunoglobulin kappa chain variable 12/13', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  278240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278240, CurrentID=0, Status='live', Symbol='Spin2c', Aliases='BC068162, SPIN-2C, Spin2', description='spindlin family, member 2C', other_designations='SPIN-2|spindlin family, member 2|spindlin-2|spindlin-like protein 2|spindlin-like protein 2C', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153832002, end_position_on_the_genomic_accession=153834243, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16097, CurrentID=0, Status='live', Symbol='Igk-V20', Aliases='', description='immunoglobulin kappa chain variable 20 (V20 family)', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16098, CurrentID=0, Status='live', Symbol='Igkv1-117', Aliases='Cr1, Gm16797, IgK, Igk-V21, V(k)45.1', description='immunoglobulin kappa variable 1-117', other_designations='Ig kappa chain|anti-phenyloxazolone mAb L chain|anti-poly(dC) monoclonal antibody kappa light chain|immunoglobulin kappa chain variable 1-117|immunoglobulin kappa chain variable 21 (V21)|polyreactive autoantibody, immunoglobulin light chain kappa', map_location='6 31.07 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68121091, end_position_on_the_genomic_accession=68121825, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  16113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16113, CurrentID=0, Status='live', Symbol='Igk-V24', Aliases='IGKV2-f', description='immunoglobulin kappa chain variable 24 (V24)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16114, CurrentID=0, Status='live', Symbol='Igk-V28', Aliases='AV061724, Igk-Ef1, Igk-Trp, Igk-VSer', description='immunoglobulin kappa chain variable 28 (V28)', other_designations='Ig kappa chain V-II region 7S34.1|immunoglobulin kappa light chain|immunoglobulin light chain variable region', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16118, CurrentID=0, Status='live', Symbol='Igk-V34', Aliases='', description='immunoglobulin kappa chain variable 34 (V34)', other_designations='immunoglobulin light chain variable region', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.3', start_position_on_the_genomic_accession=68924428, end_position_on_the_genomic_accession=68924918, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16121, CurrentID=0, Status='live', Symbol='Igk-V4', Aliases='1700129E19Rik', description='immunoglobulin kappa chain variable 4 (V4)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16122, CurrentID=0, Status='live', Symbol='Igk-V5', Aliases='B1H12, B4H2, Gal13, M86', description='immunoglobulin kappa chain variable 5 (V5 family)', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16129, CurrentID=0, Status='live', Symbol='Igk-V9', Aliases='', description='immunoglobulin kappa chain variable 9 (V9)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  278279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278279, CurrentID=0, Status='live', Symbol='Tmtc2', Aliases='8430438D04Rik, D330034A10Rik', description='transmembrane and tetratricopeptide repeat containing 2', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=105187663, end_position_on_the_genomic_accession=105574482, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16136, CurrentID=0, Status='live', Symbol='Igll1', Aliases='BB139905, Igl-5, Igll, Lambda5', description='immunoglobulin lambda-like polypeptide 1', other_designations='CD179 antigen-like family member B|ig lambda-5|immunoglobulin lambda chain 5|lambda 5', map_location='16 10.45 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16860671, end_position_on_the_genomic_accession=16863985, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16142, CurrentID=0, Status='live', Symbol='Iglv1', Aliases='1810027O01Rik, Igl-V1', description='immunoglobulin lambda variable 1', other_designations='A6 anti-[4-hydroxy-3-nitrophenyl(phenolate form)] acetyl mAb V-L region|immunoglobulin lambda chain, variable 1|immunoglobulin lambda-chain|imunoglobulin light chain variable region|light chain variable region', map_location='16 11.82 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19085017, end_position_on_the_genomic_accession=19085460, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  16145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16145, CurrentID=0, Status='live', Symbol='Igtp', Aliases='AW558444, Irgm3', description='interferon gamma induced GTPase', other_designations='', map_location='11 36.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58199556, end_position_on_the_genomic_accession=58207592, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16147, CurrentID=0, Status='live', Symbol='Ihh', Aliases='HHG-2', description='Indian hedgehog', other_designations='', map_location='1 38.55 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74945318, end_position_on_the_genomic_accession=74951651, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16149, CurrentID=0, Status='live', Symbol='Cd74', Aliases='CLIP, DHLAG, HLADG, Ia-GAMMA, Ii', description='CD74 antigen (invariant polypeptide of major histocompatibility complex, class II antigen-associated)', other_designations='HLA-DR-GAMMA|Ia-associated invariant chain|MHC class II-associated invariant chain|class II-associated invariant chain peptide|dinucleotide microsatellite|histocompatibility: class II antigens, gamma chain of|ia antigen-associated invariant chain|invariant polypeptide of major histocompatibility complex, class II antigen-associated', map_location='18 34.41 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60803849, end_position_on_the_genomic_accession=60812652, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16150, CurrentID=0, Status='live', Symbol='Ikbkb', Aliases='AI132552, IKK-2, IKK-beta, IKK2, IKK[b], IKKbeta', description='inhibitor of kappaB kinase beta', other_designations='I-kappa-B kinase 2|I-kappa-B-kinase beta|IKK-B|NFKBIKB|inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase beta|nuclear factor NF-kappa-B inhibitor kinase beta', map_location='8 A3-A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22659205, end_position_on_the_genomic_accession=22706585, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  16151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16151, CurrentID=0, Status='live', Symbol='Ikbkg', Aliases='1110037D23Rik, AI848108, AI851264, AW124339, IKK[g], NEMO', description='inhibitor of kappaB kinase gamma', other_designations='I-kappa-B kinase subunit gamma|IKK-gamma|IKKAP1|IKKG|NF-kappa-B essential modifier|ikB kinase subunit gamma|ikB kinase-associated protein 1|inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase gamma|inhibitor of nuclear factor kappa-B kinase subunit gamma|mFIP-3', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74393291, end_position_on_the_genomic_accession=74456792, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16153, CurrentID=0, Status='live', Symbol='Il10', Aliases='CSIF, Il-10', description='interleukin 10', other_designations='cytokine synthesis inhibitory factor', map_location='1 56.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131019845, end_position_on_the_genomic_accession=131024970, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16154, CurrentID=0, Status='live', Symbol='Il10ra', Aliases='AW553859, CDw210, CDw210a, Il10r, mIL-10R', description='interleukin 10 receptor, alpha', other_designations='IL-10 receptor subunit alpha|IL-10R subunit 1|IL-10R subunit alpha|IL-10R1|IL-10RA|interleukin-10 receptor subunit 1', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45253837, end_position_on_the_genomic_accession=45269251, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16155, CurrentID=0, Status='live', Symbol='Il10rb', Aliases='6620401D04Rik, AI528744, CRF2-4, Crfb4, D16H21S58, D21S58h, IL-10R2, Il10r2', description='interleukin 10 receptor, beta', other_designations='IL-10 receptor subunit beta|IL-10R subunit 2|IL-10R subunit beta|IL-10RB|cytokine receptor class-II member 4|cytokine receptor family 2 member 4|cytokine receptor family II, member 4|interleukin 10 receptor 2|interleukin-10 receptor subunit 2', map_location='16 52.87 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91406235, end_position_on_the_genomic_accession=91425834, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16156, CurrentID=0, Status='live', Symbol='Il11', Aliases='IL-11', description='interleukin 11', other_designations='', map_location='7 2.76 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4772378, end_position_on_the_genomic_accession=4782857, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16157, CurrentID=0, Status='live', Symbol='Il11ra1', Aliases='AI314697, GP130, Il-11ra, Il11ra, Il11ra2, NR1', description='interleukin 11 receptor, alpha chain 1', other_designations='IL-11 receptor subunit alpha-1|IL-11R subunit alpha-1|IL-11R-alpha-1|IL-11RA1|Il-11ra-alpha|NR-1|enhancer trap locus homolog 2|etl-2|interleukin 11 receptor, alpha chain 2|locus 2|novel cytokine receptor 1', map_location='4 21.94 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41760443, end_position_on_the_genomic_accession=41769473, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  16158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16158, CurrentID=0, Status='live', Symbol='Il11ra2', Aliases='ENSMUSG00000066201, Etl-2, Etl2, Il11ra-ps1', description='interleukin 11 receptor, alpha chain 2', other_designations='IL-11 receptor subunit alpha-2|IL-11 receptor subunit beta|IL-11R subunit alpha-2|IL-11R subunit beta|IL-11R-alpha-2|IL-11R-beta|IL-11RA2|IL-11RB|interleukin-11 receptor subunit beta', map_location='4 22.73 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42655275, end_position_on_the_genomic_accession=42665763, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  16159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16159, CurrentID=0, Status='live', Symbol='Il12a', Aliases='IL-12p35, Il-12a, Ll12a, p35', description='interleukin 12a', other_designations='CLMF p35|IL-12 subunit p35|cytotoxic lymphocyte maturation factor 35 kDa subunit|interleukin 12a p35 subunit', map_location='3 31.92 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=68690644, end_position_on_the_genomic_accession=68698550, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  278304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278304, CurrentID=0, Status='live', Symbol='Zfp385c', Aliases='A930006D11Rik, Znf385c', description='zinc finger protein 385C', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100626219, end_position_on_the_genomic_accession=100692659, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16161, CurrentID=0, Status='live', Symbol='Il12rb1', Aliases='CD212, IL-12R[b]', description='interleukin 12 receptor, beta 1', other_designations='IL-12 receptor beta component|IL-12 receptor subunit beta-1|IL-12R subunit beta-1|IL-12R-beta-1|Interleukin-12 receptor beta-1 chain precursor (IL-12R-beta1) (Interleukin-12 receptor beta) (IL-12 receptor beta component)', map_location='8 34.21 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70808360, end_position_on_the_genomic_accession=70821424, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  16162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16162, CurrentID=0, Status='live', Symbol='Il12rb2', Aliases='A930027I18Rik, IL-12RB2, Ifnm', description='interleukin 12 receptor, beta 2', other_designations='IL-12 receptor subunit beta-2|IL-12R subunit beta-2|interferon response to microorganisms', map_location='6 30.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67285294, end_position_on_the_genomic_accession=67376294, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  16163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16163, CurrentID=0, Status='live', Symbol='Il13', Aliases='Il-13', description='interleukin 13', other_designations='T-cell activation protein P600', map_location='11 31.98 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53631323, end_position_on_the_genomic_accession=53634702, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16164, CurrentID=0, Status='live', Symbol='Il13ra1', Aliases='AI882074, CD213a1, IL-13R-alpha-1, IL-13r[a], Il13ra, NR4', description='interleukin 13 receptor, alpha 1', other_designations='IL-13 receptor subunit alpha-1|IL-13R subunit alpha-1|interleukin-13-binding protein|novel cytokine receptor 4', map_location='X 20.49 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36112139, end_position_on_the_genomic_accession=36171262, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  16165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16165, CurrentID=0, Status='live', Symbol='Il13ra2', Aliases='CD213a2, IL-13R-alpha-2', description='interleukin 13 receptor, alpha 2', other_designations='', map_location='X 68.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=147383478, end_position_on_the_genomic_accession=147429192, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  16168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16168, CurrentID=0, Status='live', Symbol='Il15', Aliases='AI503618, IL-15', description='interleukin 15', other_designations='', map_location='8 39.33 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=82331624, end_position_on_the_genomic_accession=82403227, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16169, CurrentID=0, Status='live', Symbol='Il15ra', Aliases='AA690181, IL-15RA', description='interleukin 15 receptor, alpha chain', other_designations='IL-15 receptor subunit alpha|IL-15R-alpha', map_location='2 8.97 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=11705169, end_position_on_the_genomic_accession=11733985, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  16170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16170, CurrentID=0, Status='live', Symbol='Il16', Aliases='mKIAA4048', description='interleukin 16', other_designations='', map_location='7 D2-D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=83643063, end_position_on_the_genomic_accession=83746271, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  16171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16171, CurrentID=0, Status='live', Symbol='Il17a', Aliases='Ctla-8, Ctla8, IL-17, IL-17A, Il17', description='interleukin 17A', other_designations='cytotoxic T-lymphocyte-associated antigen 8', map_location='1 A', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20730905, end_position_on_the_genomic_accession=20734496, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16172, CurrentID=0, Status='live', Symbol='Il17ra', Aliases='AW538159, Cdw217, Il17r, VDw217', description='interleukin 17 receptor A', other_designations='IL-17 receptor A|IL-17RA', map_location='6 56.95 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120463181, end_position_on_the_genomic_accession=120483731, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  16173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16173, CurrentID=0, Status='live', Symbol='Il18', Aliases='Igif, Il-18', description='interleukin 18', other_designations='IFN-gamma-inducing factor|IL-1 gamma|interferon gamma-inducing factor|interferon-gamma-inducing factor|interleukin-1 gamma', map_location='9 27.75 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50554705, end_position_on_the_genomic_accession=50581841, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  16174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16174, CurrentID=0, Status='live', Symbol='Il18rap', Aliases='AcPL, IL-18R-beta, IL-18RAcP, IL-18Rbeta, IL-1RAcPL', description='interleukin 18 receptor accessory protein', other_designations='AcPL accessory protein-like)|CD218 antigen-like family member B|IL-18 receptor accessory protein|accessory protein-like|interleukin-18 receptor beta', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40515362, end_position_on_the_genomic_accession=40551707, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  671535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=671535, CurrentID=0, Status='live', Symbol='Parp10', Aliases='', description='poly (ADP-ribose) polymerase family, member 10', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76232995, end_position_on_the_genomic_accession=76243440, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16176, CurrentID=0, Status='live', Symbol='Il1b', Aliases='IL-1beta, Il-1b', description='interleukin 1 beta', other_designations='IL-1 beta', map_location='2 F', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129363843, end_position_on_the_genomic_accession=129371173, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16177, CurrentID=0, Status='live', Symbol='Il1r1', Aliases='CD121a, CD121b, IL-1R1, IL-iR, Il1r-1', description='interleukin 1 receptor, type I', other_designations='CD121 antigen-like family member A|IL-1 receptor alpha chain|IL-1R-1|IL-1R-alpha|IL-1RT-1|IL-1RT1|interleukin-1 receptor alpha|interleukin-1 receptor type I|p80', map_location='1 18.81 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40225014, end_position_on_the_genomic_accession=40316201, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  16178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16178, CurrentID=0, Status='live', Symbol='Il1r2', Aliases='CD121b, Il1r-2', description='interleukin 1 receptor, type II', other_designations='CD121 antigen-like family member B|IL-1 receptor beta chain|IL-1 type II receptor|IL-1R-2|IL-1R-beta|IL-1RT-2|IL-1RT2|interleukin-1 receptor beta|interleukin-1 receptor type II', map_location='1 18.65 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40081750, end_position_on_the_genomic_accession=40125230, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  16179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16179, CurrentID=0, Status='live', Symbol='Irak1', Aliases='AA408924, IRAK, IRAK-1, IRAK1-S, IRAK1b, Il1rak, Plpk, mPLK', description='interleukin-1 receptor-associated kinase 1', other_designations='interleukin-1 receptor-associated kinase-1|interleukin-1 receptor-associated kinase-1b|pelle-like protein kinase', map_location='X 37.61 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74013914, end_position_on_the_genomic_accession=74023921, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  16180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16180, CurrentID=0, Status='live', Symbol='Il1rap', Aliases='6430709H04Rik, AI255955, AV239853, IL-1RAcP', description='interleukin 1 receptor accessory protein', other_designations='IL-1 receptor accessory protein', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=26581688, end_position_on_the_genomic_accession=26730128, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  16181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16181, CurrentID=0, Status='live', Symbol='Il1rn', Aliases='F630041P17Rik, IL-1ra', description='interleukin 1 receptor antagonist', other_designations='IL-1RN|IL1 inhibitor|IRAP', map_location='2 16.36 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24336860, end_position_on_the_genomic_accession=24351491, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16182, CurrentID=0, Status='live', Symbol='Il18r1', Aliases='Il18ralpha, Il1rrp', description='interleukin 18 receptor 1', other_designations='CD218 antigen-like family member A|IL-18R-1|IL-18R1|IL-1Rrp|IL1 receptor-related protein|IL1R-rp|interleukin 1 receptor related protein', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40465552, end_position_on_the_genomic_accession=40500854, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  16183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16183, CurrentID=0, Status='live', Symbol='Il2', Aliases='Il-2', description='interleukin 2', other_designations='T-cell growth factor|TCGF', map_location='3 B-C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37120713, end_position_on_the_genomic_accession=37125954, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16184, CurrentID=0, Status='live', Symbol='Il2ra', Aliases='CD25, Il2r, Ly-43', description='interleukin 2 receptor, alpha chain', other_designations='IL-2 receptor subunit alpha|IL-2-RA|IL-2R alpha chain|IL-2R subunit alpha|IL2-RA|p55 chain', map_location='2 A2-A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=11642792, end_position_on_the_genomic_accession=11693194, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16185, CurrentID=0, Status='live', Symbol='Il2rb', Aliases='CD122, IL-15Rbeta, IL15Rbeta, Il-2/15Rbeta, Il-2Rbeta, p70', description='interleukin 2 receptor, beta chain', other_designations='IL-15 receptor beta chain|IL-2 receptor beta chain|IL-2 receptor subunit beta|IL-2/15 receptor-beta|IL-2R subunit beta|IL-2RB|high affinity IL-2 receptor subunit beta|p70-75', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78479256, end_position_on_the_genomic_accession=78511627, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  16186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16186, CurrentID=0, Status='live', Symbol='Il2rg', Aliases='CD132, [g]c, gamma(c), gc, p64', description='interleukin 2 receptor, gamma chain', other_designations='IL-2 receptor subunit gamma|IL-2R subunit gamma|common gamma chain|gamma C receptor|interleukin-2 receptor subunit gamma', map_location='X 43.9 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101264387, end_position_on_the_genomic_accession=101268206, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16187, CurrentID=0, Status='live', Symbol='Il3', Aliases='BPA, Csfmu, HCGF, Il-3, MCGF, PSF', description='interleukin 3', other_designations='P-cell-stimulating factor|hematopoietic growth factor|mast cell growth factor|multipotential colony-stimulating factor', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54265085, end_position_on_the_genomic_accession=54267279, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16188, CurrentID=0, Status='live', Symbol='Il3ra', Aliases='CD123, CDw123, SUT-1', description='interleukin 3 receptor, alpha chain', other_designations='IL-3 receptor alpha SP2 isoform|IL-3 receptor alpha chain|IL-3 receptor subunit alpha|IL-3R subunit alpha|IL-3R-alpha|IL-3RA|Il-3 alpha subunit|interleukin-3 receptor class II alpha chain', map_location='14 7.08 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=14346402, end_position_on_the_genomic_accession=14355493, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  16189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16189, CurrentID=0, Status='live', Symbol='Il4', Aliases='BSF-1, Il-4', description='interleukin 4', other_designations='B-cell IgG differentiation factor|B-cell growth factor 1|B-cell stimulatory factor 1|IGG1 induction factor|lymphocyte stimulatory factor 1', map_location='11 31.97 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53612460, end_position_on_the_genomic_accession=53618665, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16190, CurrentID=0, Status='live', Symbol='Il4ra', Aliases='CD124, Il4r', description='interleukin 4 receptor, alpha', other_designations='IL-4 receptor alpha chain|IL-4 receptor subunit alpha|IL-4R subunit alpha|IL-4R-alpha|IL-4RA', map_location='7 68.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125552282, end_position_on_the_genomic_accession=125579474, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  16191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16191, CurrentID=0, Status='live', Symbol='Il5', Aliases='Il-5', description='interleukin 5', other_designations='B-cell growth factor II|BCGF-II|T-cell replacing factor|TRF|cytotoxic T-lymphocyte inducer|eosinophil differentiation factor', map_location='11 A5/B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53719364, end_position_on_the_genomic_accession=53725623, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16192, CurrentID=0, Status='live', Symbol='Il5ra', Aliases='CD125, CDw125, Il5r', description='interleukin 5 receptor, alpha', other_designations='IL-5 receptor alpha chain|IL-5 receptor subunit alpha|IL-5R subunit alpha|IL-5R-alpha|IL-5RA', map_location='6 49.19 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=106710375, end_position_on_the_genomic_accession=106749037, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16193, CurrentID=0, Status='live', Symbol='Il6', Aliases='Il-6', description='interleukin 6', other_designations='B-cell hybridoma growth factor|interleukin HP-1', map_location='5 15.7 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30013161, end_position_on_the_genomic_accession=30019968, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16194, CurrentID=0, Status='live', Symbol='Il6ra', Aliases='CD126, IL-6R, IL-6R-alpha, IL-6RA, Il6r', description='interleukin 6 receptor, alpha', other_designations='IL-6 receptor alpha chain|IL-6 receptor subunit alpha|IL-6R 1|IL-6R subunit alpha', map_location='3 39.19 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89868829, end_position_on_the_genomic_accession=89913492, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16195, CurrentID=0, Status='live', Symbol='Il6st', Aliases='5133400A03Rik, AA389424, BB405851, CD130, D13Ertd699e, gp130', description='interleukin 6 signal transducer', other_designations='IL-6 receptor subunit beta|IL-6R subunit beta|IL-6R-beta|IL-6RB|interleukin-6 signal transducer|membrane glycoprotein 130|oncostatin-M receptor subunit alpha', map_location='13 63.73 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112464070, end_position_on_the_genomic_accession=112506860, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  16196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16196, CurrentID=0, Status='live', Symbol='Il7', Aliases='A630026I06Rik, Il-7, hlb368', description='interleukin 7', other_designations='', map_location='3 2.02 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=7572028, end_position_on_the_genomic_accession=7613427, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16197, CurrentID=0, Status='live', Symbol='Il7r', Aliases='CD127, IL-7Ralpha', description='interleukin 7 receptor', other_designations='IL-7 receptor alpha chain|IL-7 receptor subunit alpha|IL-7R subunit alpha|IL-7R-alpha|IL-7RA|interleukin 7 receptor alpha chain', map_location='15 4.16 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9506159, end_position_on_the_genomic_accession=9529876, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16198, CurrentID=0, Status='live', Symbol='Il9', Aliases='Il-9, P40', description='interleukin 9', other_designations='T-cell growth factor P40|cytokine P40', map_location='13 30.06 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56479277, end_position_on_the_genomic_accession=56482246, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16199, CurrentID=0, Status='live', Symbol='Il9r', Aliases='', description='interleukin 9 receptor', other_designations='IL-9 receptor|IL-9R', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32187544, end_position_on_the_genomic_accession=32205340, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16201, CurrentID=0, Status='live', Symbol='Ilf3', Aliases='MBII-26, MPHOSPH4, NF90, NFAR', description='interleukin enhancer binding factor 3', other_designations='M phase phosphoprotein 4|nuclear factor 90|nuclear factor associated with dsRNA', map_location='9 A4-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21367871, end_position_on_the_genomic_accession=21405363, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  16202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16202, CurrentID=0, Status='live', Symbol='Ilk', Aliases='AA511515, ESTM24', description='integrin linked kinase', other_designations='integrin binding protein kinase', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105736590, end_position_on_the_genomic_accession=105742925, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  671564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=671564, CurrentID=0, Status='live', Symbol='Rnf212', Aliases='EG330157, F830003B07', description='ring finger protein 212', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108728104, end_position_on_the_genomic_accession=108775109, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  16205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16205, CurrentID=0, Status='live', Symbol='Gimap1', Aliases='IAP38, Imap38, imap', description='GTPase, IMAP family member 1', other_designations='immune associated nucleotide family member|immune-associated protein 38|immunity-associated protein 1|immunity-associated protein, 38 kDa', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48739047, end_position_on_the_genomic_accession=48743795, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16206, CurrentID=0, Status='live', Symbol='Lrig1', Aliases='D6Bwg0781e, Img, LIG-1', description='leucine-rich repeats and immunoglobulin-like domains 1', other_designations='integral membrane glycoprotein', map_location='6 43.15 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=94604529, end_position_on_the_genomic_accession=94700253, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  16210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16210, CurrentID=0, Status='live', Symbol='Impact', Aliases='E430016J11Rik', description='impact, RWD domain protein', other_designations='imprinted and ancient gene protein', map_location='18 A2-B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12962943, end_position_on_the_genomic_accession=12992950, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  16211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16211, CurrentID=0, Status='live', Symbol='Kpnb1', Aliases='AA409963, IPOB, Impnb', description='karyopherin (importin) beta 1', other_designations='PTAC97|SCG|importin beta|karyopherin subunit beta-1|nuclear factor p97|pore targeting complex 97 kDa subunit', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97159710, end_position_on_the_genomic_accession=97187892, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  16261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16261, CurrentID=0, Status='live', Symbol='In(17)4t', Aliases='In4', description='inversion, Chr 17', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16314, CurrentID=0, Status='live', Symbol='In(X)1H', Aliases='In1H', description='inversion, Chr X, Harwell 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16319, CurrentID=0, Status='live', Symbol='Incenp', Aliases='2700067E22Rik, AU019509, C130081E20, C77457', description='inner centromere protein', other_designations='', map_location='19 6.1 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=9872297, end_position_on_the_genomic_accession=9899546, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  16322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16322, CurrentID=0, Status='live', Symbol='Inha', Aliases='AW555078', description='inhibin alpha', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75507077, end_position_on_the_genomic_accession=75510366, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16323, CurrentID=0, Status='live', Symbol='Inhba', Aliases='', description='inhibin beta-A', other_designations='activin beta-A chain', map_location='13 5.85 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=16007745, end_position_on_the_genomic_accession=16027723, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16324, CurrentID=0, Status='live', Symbol='Inhbb', Aliases='', description='inhibin beta-B', other_designations='activin beta-B chain', map_location='1 52.29 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=119415465, end_position_on_the_genomic_accession=119422248, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16325, CurrentID=0, Status='live', Symbol='Inhbc', Aliases='', description='inhibin beta-C', other_designations='activin beta-C chain', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127356325, end_position_on_the_genomic_accession=127370544, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16326, CurrentID=0, Status='live', Symbol='Inhbe', Aliases='', description='inhibin beta-E', other_designations='activin beta-E chain|activin betaE', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127349402, end_position_on_the_genomic_accession=127351772, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16328, CurrentID=0, Status='live', Symbol='Cep250', Aliases='AW490617, B230210E21Rik, Cep2, Inmp', description='centrosomal protein 250', other_designations='250 kDa centrosomal protein|C-Nap1|centrosomal Nek2-associated protein 1|intranuclear matrix protein', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155956272, end_position_on_the_genomic_accession=155998900, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  16329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16329, CurrentID=0, Status='live', Symbol='Inpp1', Aliases='2300002C06Rik, AV137389', description='inositol polyphosphate-1-phosphatase', other_designations='IPP|IPPase', map_location='1 26.99 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52785425, end_position_on_the_genomic_accession=52817688, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16330, CurrentID=0, Status='live', Symbol='Inpp5b', Aliases='5PTase, AW260155, INPP5P', description='inositol polyphosphate-5-phosphatase B', other_designations='inositol polyphosphate-5-phosphatase, 75 kDa|phosphoinositide 5-phosphatase', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124741829, end_position_on_the_genomic_accession=124801511, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  16331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16331, CurrentID=0, Status='live', Symbol='Inpp5d', Aliases='SHIP, SHIP-1, SHIP1, SIP-145, p150Ship, s-SHIP', description='inositol polyphosphate-5-phosphatase D', other_designations="SH2 domain-containing inositol 5'-phosphatase 1|SH2 domain-containing inositol phosphatase 1|SH2-containing inositol phosphatase SHIP|Src homology 2 domain-containing inositol-5-phosphatase|inositol polyphosphate-5-phosphatase, 145 kDa", map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87620312, end_position_on_the_genomic_accession=87720510, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  16332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16332, CurrentID=0, Status='live', Symbol='Inppl1', Aliases='51C, SHIP2', description='inositol polyphosphate phosphatase-like 1', other_designations="INPPL-1|SH2 domain-containing inositol 5'-phosphatase 2|SH2 domain-containing inositol phosphatase 2|SH2 domain-containing inositol-5'-phosphatase 2|SHIP-2|ablSH3-binding protein|inositol polyphosphate phosphatase-like protein 1|phosphatidylinositol-3,4,5-trisphosphate 5-phosphatase 2", map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101822632, end_position_on_the_genomic_accession=101838226, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  16333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16333, CurrentID=0, Status='live', Symbol='Ins1', Aliases='Ins-1, Ins2-rs1', description='insulin I', other_designations='', map_location='19 46.76 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=52264323, end_position_on_the_genomic_accession=52265476, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16334, CurrentID=0, Status='live', Symbol='Ins2', Aliases='AA986540, Ins-2, InsII, Mody, Mody4', description='insulin II', other_designations='proinsulin', map_location='7 88.0 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142678656, end_position_on_the_genomic_accession=142679708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16336, CurrentID=0, Status='live', Symbol='Insl3', Aliases='Rlf, Rlnl', description='insulin-like 3', other_designations='Ley I-L|ley-I-L|leydig insulin-like peptide|relaxin like factor|relaxin-like factor', map_location='8 34.43 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71689252, end_position_on_the_genomic_accession=71690577, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16337, CurrentID=0, Status='live', Symbol='Insr', Aliases='4932439J01Rik, CD220, D630014A15Rik, IR, IR-A, IR-B', description='insulin receptor', other_designations='', map_location='8 1.82 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3150922, end_position_on_the_genomic_accession=3279649, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  16341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16341, CurrentID=0, Status='live', Symbol='Eif3e', Aliases='48kDa, Eif3s6, Int6, eIF3-p46, eIF3-p48', description='eukaryotic translation initiation factor 3, subunit E', other_designations='MMTV integration site 6|eIF3e|eukaryotic translation initiation factor 3, subunit 6 48kDa|mammary tumor integration site 6|mammary tumor-associated protein INT-6|viral integration site protein INT-6', map_location='15 16.73 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=43250040, end_position_on_the_genomic_accession=43282736, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16348, CurrentID=0, Status='live', Symbol='Invs', Aliases='Nphp2, inv', description='inversin', other_designations='inversion of embryo turning protein|nephrocystin 2', map_location='4 B', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48279760, end_position_on_the_genomic_accession=48431953, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  16351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16351, CurrentID=0, Status='live', Symbol='Ipp', Aliases='D4Jhu8, Mipp', description='IAP promoted placental gene', other_designations='intracisternal A particle-promoted polypeptide|murine IAP-promoted placenta-expressed protein', map_location='4 53.2 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116507511, end_position_on_the_genomic_accession=116542561, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  16353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16353, CurrentID=0, Status='live', Symbol='Ipw', Aliases='', description='imprinted gene in the Prader-Willi syndrome region', other_designations='', map_location='7 34.0 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=59619158, end_position_on_the_genomic_accession=59678878, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  330406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330406, CurrentID=0, Status='live', Symbol='B4galnt3', Aliases='AB114826, B4GalNAcT3, C330047A21', description='beta-1,4-N-acetyl-galactosaminyl transferase 3', other_designations='N-acetyl-beta-glucosaminyl-glycoprotein 4-beta-N-acetylgalactosaminyltransferase 2|NGalNAc-T2|beta 1,4-N-acetylgalactosaminyltransferase-transferase 3|beta 1,4-N-acetylgalactosaminyltransferase-transferase-III|beta-1,4-N-acetylgalactosaminyltransferase III|beta4GalNAc-T3|beta4GalNAcT3', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120200753, end_position_on_the_genomic_accession=120295005, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  16362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16362, CurrentID=0, Status='live', Symbol='Irf1', Aliases='AU020929, Irf-1', description='interferon regulatory factor 1', other_designations='', map_location='11 32.0 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53770014, end_position_on_the_genomic_accession=53778374, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  16363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16363, CurrentID=0, Status='live', Symbol='Irf2', Aliases='9830146E22Rik, AI646973, Irf-2', description='interferon regulatory factor 2', other_designations='', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46739745, end_position_on_the_genomic_accession=46847458, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  16364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16364, CurrentID=0, Status='live', Symbol='Irf4', Aliases='AI385587, IRF-4, LSIRF, Spip', description='interferon regulatory factor 4', other_designations='NF-EM5|PU.1 interaction partner|Sfpi1/PU.1 interaction partner|lymphocyte-specific interferon regulatory factor|transcriptional activator PIP', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=30749208, end_position_on_the_genomic_accession=30766982, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16365, CurrentID=0, Status='live', Symbol='Irg1', Aliases='AI323667, CAD', description='immunoresponsive gene 1', other_designations='aconitate decarboxylase|cis-aconitic acid decarboxylase|immune-responsive gene 1 protein', map_location='14 51.67 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=103047012, end_position_on_the_genomic_accession=103056573, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16367, CurrentID=0, Status='live', Symbol='Irs1', Aliases='G972R, IRS-1', description='insulin receptor substrate 1', other_designations='', map_location='1 42.0 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=82233105, end_position_on_the_genomic_accession=82291439, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16369, CurrentID=0, Status='live', Symbol='Irs3', Aliases='AA408304', description='insulin receptor substrate 3', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137643030, end_position_on_the_genomic_accession=137645748, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16370, CurrentID=0, Status='live', Symbol='Irs4', Aliases='IRS-4', description='insulin receptor substrate 4', other_designations='phosphoprotein of 160 kDa|pp160', map_location='X 62.43 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=141710994, end_position_on_the_genomic_accession=141725217, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16371, CurrentID=0, Status='live', Symbol='Irx1', Aliases='', description='Iroquois related homeobox 1 (Drosophila)', other_designations='homeodomain protein IRXA1|iroquois homeobox protein 1', map_location='13 38.43 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=71958232, end_position_on_the_genomic_accession=71963723, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16372, CurrentID=0, Status='live', Symbol='Irx2', Aliases='', description='Iroquois related homeobox 2 (Drosophila)', other_designations='homeodomain protein IRXA2|iroquois homeobox protein 2', map_location='13 39.23 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=72628071, end_position_on_the_genomic_accession=72634196, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16373, CurrentID=0, Status='live', Symbol='Irx3', Aliases='AI894186', description='Iroquois related homeobox 3', other_designations='homeodomain protein IRXB1|iroquois homeobox protein 3', map_location='8 44.55 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=91798511, end_position_on_the_genomic_accession=91801654, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68265, CurrentID=0, Status='live', Symbol='Iqcf3', Aliases='1700012F17Rik, AV040440', description='IQ motif containing F3', other_designations='testis-specific IQ motif containing protein 2|testis-specific IQ motif-containing protein 2', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106504666, end_position_on_the_genomic_accession=106561631, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  16376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16376, CurrentID=0, Status='live', Symbol='Is(17;In2)1Gso', Aliases='Is1Gso', description='insertion, Chr 17 piece into inverted Chr 2, Generoso 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16385, CurrentID=0, Status='live', Symbol='Is(HSR;1)1Lub', Aliases='Is1Lub', description='insertion, homogeneous staining region, Chr 1, Lubeck 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16388, CurrentID=0, Status='live', Symbol='Is(In7;X)1Ct', Aliases='Is1Ct', description='insertion, inverted Chr 7 piece into Chr X, Cattanach 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16389, CurrentID=0, Status='live', Symbol='Is(In7;X)1Ct', Aliases='Is1Ct', description='insertion, inverted Chr 7 piece into Chr X, Cattanach 1', other_designations='', map_location='7 27.0 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16391, CurrentID=0, Status='live', Symbol='Irf9', Aliases='Irf-9, Isgf3g, p48', description='interferon regulatory factor 9', other_designations='IFN-alpha-responsive transcription factor subunit|ISGF-3 gamma|ISGF3 p48 subunit|interferon dependent positive acting transcription factor 3 gamma|interferon-stimulated gene factor 3 gamma|transcriptional regulator ISGF3 subunit gamma', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55603985, end_position_on_the_genomic_accession=55610030, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  16392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16392, CurrentID=0, Status='live', Symbol='Isl1', Aliases='', description='ISL1 transcription factor, LIM/homeodomain', other_designations='Islet 1|Lim-homeodomain protein Islet1|insulin related protein|islet-1', map_location='13 D2.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=116298257, end_position_on_the_genomic_accession=116309721, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16396, CurrentID=0, Status='live', Symbol='Itch', Aliases='6720481N21Rik, 8030492O04Rik, A130065M08, AIP4, C230047C07Rik', description='itchy, E3 ubiquitin protein ligase', other_designations='', map_location='2 76.94 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155133379, end_position_on_the_genomic_accession=155226855, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  16398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16398, CurrentID=0, Status='live', Symbol='Itga2', Aliases='CD49B, DX5', description='integrin alpha 2', other_designations='CD49 antigen-like family member B|GPIa|VLA-2 receptor, alpha 2 subunit|VLA-2 subunit alpha|collagen receptor|platelet membrane glycoprotein Ia', map_location='13 64.61 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=114835912, end_position_on_the_genomic_accession=114932041, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  16399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16399, CurrentID=0, Status='live', Symbol='Itga2b', Aliases='AI172977, CD41, CD41B, GpIIb, alphaIIb', description='integrin alpha 2b', other_designations='GP IIb|GPalpha IIb|alpha IIb|platelet glycoprotein IIb of IIb/II Ia complex|platelet membrane glycoprotein IIb', map_location='11 66.29 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102453297, end_position_on_the_genomic_accession=102469883, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  16400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16400, CurrentID=0, Status='live', Symbol='Itga3', Aliases='AA407068, CD49C, GAPB3', description='integrin alpha 3', other_designations='CD49 antigen-like family member C|VLA-3 receptor, alpha 3 subunit|galactoprotein B3', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95044475, end_position_on_the_genomic_accession=95076757, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  16401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16401, CurrentID=0, Status='live', Symbol='Itga4', Aliases='CD49DB, Itga4', description='integrin alpha 4', other_designations='CD49 antigen-like family member D|LPAM subunit alpha|VLA-4 receptor, alpha 4 subunit|VLA-4 subunit alpha|integrin alpha-IV|integrin alpha4B|lymphocyte Peyer patch adhesion molecules subunit alpha', map_location='2 47.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=79255426, end_position_on_the_genomic_accession=79333114, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  16402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16402, CurrentID=0, Status='live', Symbol='Itga5', Aliases='Cd49e, Fnra, VLA5', description='integrin alpha 5 (fibronectin receptor alpha)', other_designations='CD49 antigen-like family member E|fibronectin receptor alpha polypeptide|fibronectin receptor subunit alpha|integrin alpha-F', map_location='15 58.9 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103344286, end_position_on_the_genomic_accession=103366748, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  16403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16403, CurrentID=0, Status='live', Symbol='Itga6', Aliases='5033401O05Rik, AI115430, Cd49f, VLA-6', description='integrin alpha 6', other_designations='CD49 antigen-like family member F', map_location='2 42.79 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71786885, end_position_on_the_genomic_accession=71858427, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  16404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16404, CurrentID=0, Status='live', Symbol='Itga7', Aliases='[a]7, alpha7', description='integrin alpha 7', other_designations='', map_location='10 77.2 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128933775, end_position_on_the_genomic_accession=128958286, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  16407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16407, CurrentID=0, Status='live', Symbol='Itgae', Aliases='A530055J10, CD103, aM290, alpha-E1, alpha-M290', description='integrin alpha E, epithelial-associated', other_designations='integrin alpha M290', map_location='11 45.22 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73090595, end_position_on_the_genomic_accession=73147449, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  16408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16408, CurrentID=0, Status='live', Symbol='Itgal', Aliases='(p180), Cd11a, LFA-1, LFA-1A, Ly-15, Ly-21', description='integrin alpha L', other_designations='CD11 antigen-like family member A|alpha L integrin|leukocyte adhesion glycoprotein LFA-1 alpha chain|leukocyte function-associated molecule 1 alpha chain|lymphocyte antigen 15|lymphocyte function associated antigen 1, alpha polypeptide', map_location='7 69.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127296230, end_position_on_the_genomic_accession=127335138, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  16409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16409, CurrentID=0, Status='live', Symbol='Itgam', Aliases='CD11b/CD18, CR3, CR3A, Cd11b, F730045J24Rik, Ly-40, MAC1, Mac-1, Mac-1a', description='integrin alpha M', other_designations='CD11 antigen-like family member B|CD11B (p170)|CR-3 alpha chain|Mac-1 alpha|cell surface glycoprotein MAC-1 alpha subunit|cell surface glycoprotein MAC-1 subunit alpha|complement component receptor 3 alpha-a|complement receptor type 3|leukocyte adhesion receptor MO1|macrophage antigen alpha', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128062640, end_position_on_the_genomic_accession=128128160, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  16410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16410, CurrentID=0, Status='live', Symbol='Itgav', Aliases='1110004F14Rik, 2610028E01Rik, CD51, D430040G12Rik', description='integrin alpha V', other_designations='alpha v integrin|vitronectin receptor alpha polypeptide (VNRA)|vitronectin receptor subunit alpha', map_location='2 49.33 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=83723819, end_position_on_the_genomic_accession=83806917, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  16411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16411, CurrentID=0, Status='live', Symbol='Itgax', Aliases='AI449405, Cd11c, Cr4, N418', description='integrin alpha X', other_designations='CD11 antigen-like family member C|CD11C (p150) alpha polypeptide|complement receptor 4|integrin aX|leukocyte adhesion glycoprotein p150,95 alpha chain|leukocyte adhesion receptor p150,95', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128129529, end_position_on_the_genomic_accession=128150657, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  16412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16412, CurrentID=0, Status='live', Symbol='Itgb1', Aliases='4633401G24Rik, AA409975, AA960159, CD29, ENSMUSG00000051907, Fnrb, Gm9863, gpIIa', description='integrin beta 1 (fibronectin receptor beta)', other_designations='VLA-4 subunit beta|fibronectin receptor subunit beta', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=128685284, end_position_on_the_genomic_accession=128733579, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16413, CurrentID=0, Status='live', Symbol='Itgb1bp1', Aliases='AI449260, AU019480, Icap1', description='integrin beta 1 binding protein 1', other_designations='bodenin|integrin cytoplasmic domain-associated protein 1', map_location='12 A1.2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=21267246, end_position_on_the_genomic_accession=21286279, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16414, CurrentID=0, Status='live', Symbol='Itgb2', Aliases='2E6, AI528527, Cd18, LAD, LCAMB, Lfa1, MF17', description='integrin beta 2', other_designations='Mac-1 beta|cell surface adhesion glycoproteins LFA-1/CR3/p150,95 subunit beta|complement receptor C3 subunit beta|lymphocyte function associated antigen 1|macrophage antigen-1 beta', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77530328, end_position_on_the_genomic_accession=77565681, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16415, CurrentID=0, Status='live', Symbol='Itgb2l', Aliases='5033406G21Rik, pactolus', description='integrin beta 2-like', other_designations='itgb2l|protein pactolus', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96422295, end_position_on_the_genomic_accession=96444365, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  16416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16416, CurrentID=0, Status='live', Symbol='Itgb3', Aliases='CD61, GP3A, INGRB3', description='integrin beta 3', other_designations='GPIIIa|platelet gpIIIa|platelet membrane glycoprotein IIIa', map_location='11 67.84 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104608000, end_position_on_the_genomic_accession=104670490, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  16418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16418, CurrentID=0, Status='live', Symbol='Eif6', Aliases='1110004P11, AA408895, CAB, Itgb4bp, eIF-6, imc-415, p27(BBP), p27BBP', description='eukaryotic translation initiation factor 6', other_designations='B4 integrin interactor|eIF-6|integrin beta 4 binding protein', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155819837, end_position_on_the_genomic_accession=155826925, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16419, CurrentID=0, Status='live', Symbol='Itgb5', Aliases='AA475909, AI874634, ESTM23, [b]-5, [b]5, [b]5A, [b]5B, beta-5, beta5', description='integrin beta 5', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33829665, end_position_on_the_genomic_accession=33949338, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  16420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16420, CurrentID=0, Status='live', Symbol='Itgb6', Aliases='2210409C20Rik, 4831415H04Rik', description='integrin beta 6', other_designations='', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=60598292, end_position_on_the_genomic_accession=60722603, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  16421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16421, CurrentID=0, Status='live', Symbol='Itgb7', Aliases='Ly69', description='integrin beta 7', other_designations='M290 IEL antigen|integrin beta-P', map_location='15 57.39 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102215994, end_position_on_the_genomic_accession=102231958, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  16423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16423, CurrentID=0, Status='live', Symbol='Cd47', Aliases='9130415E20Rik, AA407862, AI848868, AW108519, B430305P08Rik, IAP, Itgp', description='CD47 antigen (Rh-related antigen, integrin-associated signal transducer)', other_designations='Rh-related antigen|integrin-associated protein', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=49855252, end_position_on_the_genomic_accession=49912424, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  16424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16424, CurrentID=0, Status='live', Symbol='Itih1', Aliases='ITI-HC1, Intin1, Itih-1', description='inter-alpha trypsin inhibitor, heavy chain 1', other_designations='inter-alpha (globulin) inhibitor, H1 polypeptide', map_location='14 A2-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30929180, end_position_on_the_genomic_accession=30943321, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  16425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16425, CurrentID=0, Status='live', Symbol='Itih2', Aliases='AI747202, Intin2, Itih-2', description='inter-alpha trypsin inhibitor, heavy chain 2', other_designations='ITI heavy chain H2|ITI-HC2|inter-alpha-inhibitor heavy chain 2', map_location='2 6.89 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10094591, end_position_on_the_genomic_accession=10130683, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  16426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16426, CurrentID=0, Status='live', Symbol='Itih3', Aliases='AW108094, ITI-HC3, Intin3, Itih-3', description='inter-alpha trypsin inhibitor, heavy chain 3', other_designations='ITI heavy chain H3|inter-alpha-inhibitor H3 chain|inter-alpha-inhibitor heavy chain 3', map_location='14 A2-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30908572, end_position_on_the_genomic_accession=30923755, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  16427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16427, CurrentID=0, Status='live', Symbol='Itih4', Aliases='ITI-HC4, Itih-4, PK-120', description='inter alpha-trypsin inhibitor, heavy chain 4', other_designations='ITI heavy chain H4|inter-alpha-inhibitor heavy chain 4', map_location='14 19.09 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30886476, end_position_on_the_genomic_accession=30901986, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  16428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16428, CurrentID=0, Status='live', Symbol='Itk', Aliases='Emt, Tcsk, Tsk', description='IL2 inducible T cell kinase', other_designations='T-cell-specific kinase|interleukin-2-inducible T cell kinase|kinase TLK', map_location='11 27.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46325148, end_position_on_the_genomic_accession=46389515, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  16429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16429, CurrentID=0, Status='live', Symbol='Itln1', Aliases='IntL, Itln, Itln2, Itln3, Itln5, Itlna, Lfr', description='intelectin 1 (galactofuranose binding)', other_designations='calcium-dependent galactosyl-specific lectin|galactofuranose-binding lectin|intelectin 2|intelectin 5|intelectin a|intestinal lactoferrin receptor|truncated intelectin 3', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171518123, end_position_on_the_genomic_accession=171535325, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16430, CurrentID=0, Status='live', Symbol='Stt3a', Aliases='AA408947, BB081708, Itm1', description='STT3, subunit of the oligosaccharyltransferase complex, homolog A (S. cerevisiae)', other_designations='B5|STT3-A|integral membrane protein 1|intergral membrane protein 1|oligosaccharyl transferase subunit STT3A', map_location='9 20.67 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=36729345, end_position_on_the_genomic_accession=36767674, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  16431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16431, CurrentID=0, Status='live', Symbol='Itm2a', Aliases='AW108051, Bricd2a, E25, Itm2', description='integral membrane protein 2A', other_designations='', map_location='X A2-A3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107397195, end_position_on_the_genomic_accession=107403360, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16432, CurrentID=0, Status='live', Symbol='Itm2b', Aliases='AI256040, Bri, Bri2, Bricd2b, D14Sel6, E25BMM, imBRI2', description='integral membrane protein 2B', other_designations='Integral membrane protein 2B (E25B protein)|immature BRI2|transmembrane protein BRI', map_location='14 38.88 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73362231, end_position_on_the_genomic_accession=73385271, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16433, CurrentID=0, Status='live', Symbol='Cuzd1', Aliases='Erg-1, Itmap1, Uo-44, Usg, Utczp', description='CUB and zona pellucida-like domains 1', other_designations='CUB and ZP domain-containing protein 1|cuzd1|integral membrane-associated protein 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131308554, end_position_on_the_genomic_accession=131322292, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16434, CurrentID=0, Status='live', Symbol='Itpa', Aliases='2010016I08Rik, AU020102, Itp', description='inosine triphosphatase (nucleoside triphosphate pyrophosphatase)', other_designations='ITPase|NTPase|non-canonical purine NTP pyrophosphatase|non-standard purine NTP pyrophosphatase|nucleoside triphosphate pyrophosphatase|nucleoside-triphosphate diphosphatase|nucleoside-triphosphate pyrophosphatase', map_location='2 63.24 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130667841, end_position_on_the_genomic_accession=130681614, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16438, CurrentID=0, Status='live', Symbol='Itpr1', Aliases='D6Pas2, ENSMUSG00000072853, Gm10429, IP3R1, InsP3R, Ip3r, Itpr-1, P400, Pcp-1, Pcp1, opt, wblo', description='inositol 1,4,5-trisphosphate receptor 1', other_designations='IP3 receptor|IP3R 1|InsP3R type I|InsP>3<R type I|Purkinje cell protein 1|inositol 1,4,5-triphosphate receptor 1|inositol 1,4,5-trisphosphate-binding protein P400|insP3R1|opisthotonus|protein PCD-6|type 1 InsP3 receptor|type 1 inositol 1,4,5-trisphosphate receptor', map_location='6 49.74 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=108213083, end_position_on_the_genomic_accession=108551116, orientation='plus', exon_count=65, OMIM='', no_hdr0=''),
  16439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16439, CurrentID=0, Status='live', Symbol='Itpr2', Aliases='AI649341, InsP3R-2, InsP3R-5, Ip3r2, Itpr5, insP3R2', description='inositol 1,4,5-triphosphate receptor 2', other_designations='IP3 receptor|IP3R 2|inositol 1,4,5-triphosphate receptor 5|inositol 1,4,5-trisphosphate type V receptor|inositol trisphosphate receptor type 2|type 2 InsP3 receptor|type 2 inositol 1,4,5-trisphosphate receptor', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146108299, end_position_on_the_genomic_accession=146502223, orientation='minus', exon_count=58, OMIM='', no_hdr0=''),
  16440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16440, CurrentID=0, Status='live', Symbol='Itpr3', Aliases='Ip3r3, Itpr-3', description='inositol 1,4,5-triphosphate receptor 3', other_designations='IP3 receptor|IP3R 3|insP3R3|type 3 InsP3 receptor|type 3 inositol 1,4,5-trisphosphate receptor', map_location='17 19.0 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27057260, end_position_on_the_genomic_accession=27122223, orientation='plus', exon_count=58, OMIM='', no_hdr0=''),
  16443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16443, CurrentID=0, Status='live', Symbol='Itsn1', Aliases='AA517634, AA545208, AI316805, AI839402, AI848451, Ehsh1, Ese1, Itsn, Sh3p17', description='intersectin 1 (SH3 domain protein 1A)', other_designations='EH and SH3 domains protein 1|Eh domain, SH3 domain regulator of endocytosis 1|intersectin (SH3 domain protein 1A)|intersectin short form 2 variant 2|intersectin-L', map_location='16 C3.3-C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91729354, end_position_on_the_genomic_accession=91920591, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  16447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16447, CurrentID=0, Status='live', Symbol='Ivl', Aliases='1110019C06Rik', description='involucrin', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92570899, end_position_on_the_genomic_accession=92573790, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16449, CurrentID=0, Status='live', Symbol='Jag1', Aliases='ABE2, Gsfabe2, Htu, Ozz, Ser-1', description='jagged 1', other_designations='Serrate-1', map_location='2 67.73 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=137081451, end_position_on_the_genomic_accession=137116520, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  16450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16450, CurrentID=0, Status='live', Symbol='Jag2', Aliases='D12Ggc2e, Serh, mJagged2-1, sm', description='jagged 2', other_designations='syndactylism', map_location='12 61.37 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112907803, end_position_on_the_genomic_accession=112929764, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  16451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16451, CurrentID=0, Status='live', Symbol='Jak1', Aliases='AA960307, BAP004, C130039L05Rik', description='Janus kinase 1', other_designations='JAK-1', map_location='4 46.19 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101151974, end_position_on_the_genomic_accession=101265297, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  16452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16452, CurrentID=0, Status='live', Symbol='Jak2', Aliases='Fd17', description='Janus kinase 2', other_designations='JAK-2', map_location='19 23.73 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29251776, end_position_on_the_genomic_accession=29313080, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  16453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16453, CurrentID=0, Status='live', Symbol='Jak3', Aliases='fae, wil', description='Janus kinase 3', other_designations='', map_location='8 34.43 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71676383, end_position_on_the_genomic_accession=71690577, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  16456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16456, CurrentID=0, Status='live', Symbol='F11r', Aliases='9130004G24, AA638916, ESTM33, JAM, JAM-1, JAM-A, Jcam, Jcam1, Ly106', description='F11 receptor', other_designations='BV11 antigen|F11 receptor/F11R|junction cell adhesion molecule A|junction cell adhesion molecule1|junctional adhesion molecule 1', map_location='1 79.43 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171437561, end_position_on_the_genomic_accession=171464593, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  16467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16467, CurrentID=0, Status='live', Symbol='Atcay', Aliases='3322401A10Rik, BB077577, BNIP-H, CLAC, hes, ji', description='ataxia, cerebellar, Cayman type', other_designations='ataxia, cerebellar, Cayman type homolog|hesitant|jittery', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81204513, end_position_on_the_genomic_accession=81230779, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16468, CurrentID=0, Status='live', Symbol='Jarid2', Aliases='Jmj, jumonji', description='jumonji, AT rich interactive domain 2', other_designations='jumonji/ARID domain-containing protein 2', map_location='13 21.66 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=44729181, end_position_on_the_genomic_accession=44921643, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  16469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16469, CurrentID=0, Status='live', Symbol='Jrk', Aliases='', description='jerky', other_designations='', map_location='15 34.26 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74702412, end_position_on_the_genomic_accession=74709322, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16470, CurrentID=0, Status='live', Symbol='Ush1g', Aliases='Sans, js', description='Usher syndrome 1G', other_designations='Usher syndrome 1G homolog|jackson shaker protein|scaffold protein containing ankyrin repeats and SAM domain|scaffold protein, amkyrin repeats and SAM domain containing|scaffold protein, ankyrin repeats and SAM domain containing', map_location='11 80.84 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115315192, end_position_on_the_genomic_accession=115322142, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16475, CurrentID=0, Status='live', Symbol='Ajuba', Aliases='Jub', description='ajuba LIM protein', other_designations='protein ajuba', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54567469, end_position_on_the_genomic_accession=54577661, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16476, CurrentID=0, Status='live', Symbol='Jun', Aliases='AP-1c, c-jun, Jun', description='jun proto-oncogene', other_designations='AH119|AP1|Jun oncogene|V-jun avian sarcoma virus 17 oncogene homolog|activator protein 1|immediate early|jun A|proto-oncogene c-jun', map_location='4 C5-C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=95049036, end_position_on_the_genomic_accession=95052222, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  16477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16477, CurrentID=0, Status='live', Symbol='Junb', Aliases='', description='jun B proto-oncogene', other_designations='Jun-B oncogene|myD21', map_location='8 C2-D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84976909, end_position_on_the_genomic_accession=84978748, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  16478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16478, CurrentID=0, Status='live', Symbol='Jund', Aliases='Jund1', description='jun D proto-oncogene', other_designations='Jun proto-oncogene related gene d|Jun proto-oncogene related gene d1', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70697739, end_position_on_the_genomic_accession=70700616, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  16480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16480, CurrentID=0, Status='live', Symbol='Jup', Aliases='Ctnng, D930025P04Rik, PG', description='junction plakoglobin', other_designations='desmoplakin III|desmoplakin-3|gamma-catenin', map_location='11 63.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100368856, end_position_on_the_genomic_accession=100397790, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  16483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16483, CurrentID=0, Status='live', Symbol='Kap', Aliases='AW146435, FKSG22', description='kidney androgen regulated protein', other_designations='ARP', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=133849855, end_position_on_the_genomic_accession=133853667, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16485, CurrentID=0, Status='live', Symbol='Kcna1', Aliases='AI840627, Kca1-1, Kv1.1, MBK1, Mk-1, Shak, mceph', description='potassium voltage-gated channel, shaker-related subfamily, member 1', other_designations='MKI|brain potassium channel protein-1|megencephaly|potassium voltage gated channel, shaker related subfamily, member 1|voltage-gated potassium channel subunit Kv1.1', map_location='6 F1-F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126636463, end_position_on_the_genomic_accession=126645801, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16490, CurrentID=0, Status='live', Symbol='Kcna2', Aliases='Akr6a4, ENSMUSG00000074335, Gm10672, Kca1-2, Kv1.2, Mk-2', description='potassium voltage-gated channel, shaker-related subfamily, member 2', other_designations='MK2|voltage-gated potassium channel subunit Kv1.2', map_location='3 46.61 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107090852, end_position_on_the_genomic_accession=107115005, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16491, CurrentID=0, Status='live', Symbol='Kcna3', Aliases='Kca1-3, Kv1.3, Mk-3', description='potassium voltage-gated channel, shaker-related subfamily, member 3', other_designations='MK3|Voltage-gated potassium channel protein Kv1.3 (RGK5) (RCK3) (KV3)|potassium voltage gated channel, shaker related subfamily, member 3|voltage-gated potassium channel subunit Kv1.3', map_location='3 46.59 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107036162, end_position_on_the_genomic_accession=107038129, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  16492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16492, CurrentID=0, Status='live', Symbol='Kcna4', Aliases='Kv1.4', description='potassium voltage-gated channel, shaker-related subfamily, member 4', other_designations='potassium voltage gated channel, shaker related subfamily, member 4|voltage-gated potassium channel subunit Kv1.4', map_location='2 E', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=107290589, end_position_on_the_genomic_accession=107303587, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16493, CurrentID=0, Status='live', Symbol='Kcna5', Aliases='Kv1.5', description='potassium voltage-gated channel, shaker-related subfamily, member 5', other_designations='KV1-5|voltage-gated potassium channel subunit Kv1.5', map_location='6 61.35 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126532551, end_position_on_the_genomic_accession=126535555, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  16494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16494, CurrentID=0, Status='live', Symbol='Kcna6', Aliases='Kv1.6, MK1.6', description='potassium voltage-gated channel, shaker-related, subfamily, member 6', other_designations='potassium voltage gated channel, shaker related, subfamily, member 6|voltage-gated potassium channel subunit Kv1.6', map_location='6 61.72 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126708328, end_position_on_the_genomic_accession=126743108, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16495, CurrentID=0, Status='live', Symbol='Kcna7', Aliases='Kv1.7', description='potassium voltage-gated channel, shaker-related subfamily, member 7', other_designations='potassium voltage gated channel, shaker related subfamily, member 7|potassium voltage-gated channel, shaker-related subfamily, member 7; potassium voltage gated channel, shaker related subfamily, member 7|voltage-gated potassium channel subunit Kv1.7', map_location='7 29.3 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45405960, end_position_on_the_genomic_accession=45411382, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16497, CurrentID=0, Status='live', Symbol='Kcnab1', Aliases='Akr8a8, Kvbeta1.1, mKv(beta)1', description='potassium voltage-gated channel, shaker-related subfamily, beta member 1', other_designations='K(+) channel subunit beta-1|kv-beta-1|potassium voltage gated channel, shaker related subfamily, beta member 1', map_location='3 30.15 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=64949049, end_position_on_the_genomic_accession=65378225, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  16498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16498, CurrentID=0, Status='live', Symbol='Kcnab2', Aliases='F5, I2rf5, Kcnb3, kv-beta-2', description='potassium voltage-gated channel, shaker-related subfamily, beta member 2', other_designations='K(+) channel subunit beta-2|neuroimmune protein F5|potassium voltage gated channel, shaker related subfamily, beta member 2', map_location='4 83.08 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152390740, end_position_on_the_genomic_accession=152477549, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  16499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16499, CurrentID=0, Status='live', Symbol='Kcnab3', Aliases='C330022D06Rik, Kcnab4, mKv(beta)4', description='potassium voltage-gated channel, shaker-related subfamily, beta member 3', other_designations='K(+) channel subunit beta-3|kv-beta-3', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69326258, end_position_on_the_genomic_accession=69333041, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  16500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16500, CurrentID=0, Status='live', Symbol='Kcnb1', Aliases='Kcr1-1, Kv2.1, Shab', description='potassium voltage gated channel, Shab-related subfamily, member 1', other_designations='voltage-gated potassium channel subunit Kv2.1', map_location='2 87.22 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167095969, end_position_on_the_genomic_accession=167188818, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16502, CurrentID=0, Status='live', Symbol='Kcnc1', Aliases='C230009H10Rik, KShIIIB, KV4, Kcr2-1, Kv3.1, NGK2, Shaw', description='potassium voltage gated channel, Shaw-related subfamily, member 1', other_designations='voltage-gated potassium channel subunit Kv3.1|voltage-gated potassium channel subunit Kv4', map_location='7 30.1 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46396457, end_position_on_the_genomic_accession=46438706, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16504, CurrentID=0, Status='live', Symbol='Kcnc3', Aliases='KShIIID, Kcr2-3, Kv3.3', description='potassium voltage gated channel, Shaw-related subfamily, member 3', other_designations='voltage-gated potassium channel subunit Kv3.3', map_location='7 28.85 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44590661, end_position_on_the_genomic_accession=44604754, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16506, CurrentID=0, Status='live', Symbol='Kcnd1', Aliases='1110037K09Rik, Kca2-1, Kv4.1, Shal, Shal1, mShal1', description='potassium voltage-gated channel, Shal-related family, member 1', other_designations='voltage-gated potassium channel subunit Kv4.1', map_location='X 3.53 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7823843, end_position_on_the_genomic_accession=7836503, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16508, CurrentID=0, Status='live', Symbol='Kcnd2', Aliases='AI839615, AW555701, Kv4.2, R75121, mKIAA1044', description='potassium voltage-gated channel, Shal-related family, member 2', other_designations='potassium channel Kv4.2|voltage-gated potassium channel subunit Kv4.2', map_location='6 8.49 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=21216109, end_position_on_the_genomic_accession=21729805, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16509, CurrentID=0, Status='live', Symbol='Kcne1', Aliases='Isk, MinK, nmf190', description='potassium voltage-gated channel, Isk-related subfamily, member 1', other_designations='IKs producing slow voltage-gated potassium channel subunit beta Mink|delayed rectifier potassium channel subunit IsK|mISK|minimal potassium channel', map_location='16 53.57 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92345995, end_position_on_the_genomic_accession=92359468, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16510, CurrentID=0, Status='live', Symbol='Kcnh1', Aliases='EAG1, Kv10.1, M-eag', description='potassium voltage-gated channel, subfamily H (eag-related), member 1', other_designations='EAG channel 1|ether a go-go|ether-a-go-go potassium channel 1|voltage-gated potassium channel subunit Kv10.1', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=192190362, end_position_on_the_genomic_accession=192510159, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  16511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16511, CurrentID=0, Status='live', Symbol='Kcnh2', Aliases='AI326795, ERG1, LQT, Lqt2, M-erg, Merg1, merg1a, merg1b', description='potassium voltage-gated channel, subfamily H (eag-related), member 2', other_designations='eag-related protein 1|ether-a-go-go-related gene potassium channel 1|ether-a-go-go-related protein 1|voltage-gated potassium channel subunit Kv11.1', map_location='5 10.94 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24319589, end_position_on_the_genomic_accession=24351604, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  16512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16512, CurrentID=0, Status='live', Symbol='Kcnh3', Aliases='AU019351, C030044P22Rik, Elk2, Kv12.2, Melk2', description='potassium voltage-gated channel, subfamily H (eag-related), member 3', other_designations='ELK channel 2|ether a go-go like|ether-a-go-go-like potassium channel 2|voltage-gated potassium channel subunit Kv12.2', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99224358, end_position_on_the_genomic_accession=99242826, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16513, CurrentID=0, Status='live', Symbol='Kcnj10', Aliases='BIR10, BIRK-1, Kir1.2, Kir4.1', description='potassium inwardly-rectifying channel, subfamily J, member 10', other_designations='inward rectifier K(+) channel Kir4.1|potassium channel, inwardly rectifying subfamily J member 10', map_location='1 79.69 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172341210, end_position_on_the_genomic_accession=172374085, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16514, CurrentID=0, Status='live', Symbol='Kcnj11', Aliases='Kir6.2, mBIR', description='potassium inwardly rectifying channel, subfamily J, member 11', other_designations='inward rectifier K(+) channel Kir6.2', map_location='7 29.66 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46097123, end_position_on_the_genomic_accession=46100872, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16515, CurrentID=0, Status='live', Symbol='Kcnj12', Aliases='IRK-2, IRK2, Kir2.2, MB-IRK2', description='potassium inwardly-rectifying channel, subfamily J, member 12', other_designations='inward rectifier K(+) channel Kir2.2', map_location='11 37.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61022564, end_position_on_the_genomic_accession=61073269, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16516, CurrentID=0, Status='live', Symbol='Kcnj15', Aliases='4930414N08Rik, AI182284, AI267127, IRKK, Kir4.2', description='potassium inwardly-rectifying channel, subfamily J, member 15', other_designations='inward rectifier K+ channel Kir4.2|potassium channel, inwardly rectifying subfamily J member 15', map_location='16 55.86 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=95257558, end_position_on_the_genomic_accession=95300260, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  16517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16517, CurrentID=0, Status='live', Symbol='Kcnj16', Aliases='6430410F18Rik, AI132396, Kir5.1', description='potassium inwardly-rectifying channel, subfamily J, member 16', other_designations='inward rectifier K(+) channel Kir5.1|potassium channel, inwardly rectifying subfamily J member 16', map_location='11 75.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=110968033, end_position_on_the_genomic_accession=111027968, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16518, CurrentID=0, Status='live', Symbol='Kcnj2', Aliases='IRK1, Kcnf1, Kir2.1', description='potassium inwardly-rectifying channel, subfamily J, member 2', other_designations='IRK-1|inward rectifier K(+) channel Kir2.1|potassium channel, inwardly rectifying subfamily J member 2', map_location='11 75.23 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=111066164, end_position_on_the_genomic_accession=111076825, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16519, CurrentID=0, Status='live', Symbol='Kcnj3', Aliases='GIRK-1, GIRK1, Kcnf3, Kir3.1', description='potassium inwardly-rectifying channel, subfamily J, member 3', other_designations='inward rectifier K(+) channel Kir3.1|potassium channel, inwardly rectifying subfamily J member 3', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=55426974, end_position_on_the_genomic_accession=55598145, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16520, CurrentID=0, Status='live', Symbol='Kcnj4', Aliases='IRK3, Kcnf2, MB-IRK3', description='potassium inwardly-rectifying channel, subfamily J, member 4', other_designations='IRK-3|Kir 2.3|inward rectifier K(+) channel Kir2.3|potassium channel, inwardly rectifying subfamily J member 4', map_location='15 37.74 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79483714, end_position_on_the_genomic_accession=79511883, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16521, CurrentID=0, Status='live', Symbol='Kcnj5', Aliases='GIRK4, Kir3.4', description='potassium inwardly-rectifying channel, subfamily J, member 5', other_designations='CIR|GIRK-4|KATP-1|cardiac inward rectifier|heart KATP channel|inward rectifier K(+) channel Kir3.4|potassium channel, inwardly rectifying subfamily J member 5', map_location='9 17.65 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=32314706, end_position_on_the_genomic_accession=32344389, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16522, CurrentID=0, Status='live', Symbol='Kcnj6', Aliases='BIR1, GIRK2, KATP2, KCNJ7, Kir3.2, weaver, wv', description='potassium inwardly-rectifying channel, subfamily J, member 6', other_designations='G-protein-coupled inward rectifier K+ channel 2-B|GIRK-2|inward rectifier K(+) channel Kir3.2|potassium channel, inwardly rectifying subfamily J member 6|potassium inwardly-rectifying channel J6', map_location='16 55.44 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94748683, end_position_on_the_genomic_accession=94997696, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16523, CurrentID=0, Status='live', Symbol='Kcnj8', Aliases='AI448900, Kir6.1, gnite, slmbr, sltr', description='potassium inwardly-rectifying channel, subfamily J, member 8', other_designations='inward rectifier K(+) channel Kir6.1|inwardly rectifier K(+) channel Kir6.1|potassium channel, inwardly rectifying subfamily J member 8|uKATP-1', map_location='6 74.31 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142564837, end_position_on_the_genomic_accession=142571647, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16524, CurrentID=0, Status='live', Symbol='Kcnj9', Aliases='1700085N21Rik, Girk3, Kir3.3, mbGIRK3', description='potassium inwardly-rectifying channel, subfamily J, member 9', other_designations='GIRK-3|inward rectifier K(+) channel Kir3.3|inwardly rectifier K(+) channel Kir3.3|potassium channel, inwardly rectifying subfamily J member 9', map_location='1 79.66 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172320501, end_position_on_the_genomic_accession=172329331, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16525, CurrentID=0, Status='live', Symbol='Kcnk1', Aliases='AI788889, TWIK-1', description='potassium channel, subfamily K, member 1', other_designations='inward rectifying potassium channel protein TWIK-1', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125995102, end_position_on_the_genomic_accession=126030685, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16526, CurrentID=0, Status='live', Symbol='Kcnk2', Aliases='A430027H14Rik, AI848635, TREK-1', description='potassium channel, subfamily K, member 2', other_designations='TREK-1 K(+) channel subunit|outward rectifying potassium channel protein TREK-1|two pore potassium channel TPKC1', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=189207930, end_position_on_the_genomic_accession=189402705, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16527, CurrentID=0, Status='live', Symbol='Kcnk3', Aliases='TASK, Task-1, cTBAK-1', description='potassium channel, subfamily K, member 3', other_designations='TWIK-related acid-sensitive K(+) channel 1|acid-sensitive potassium channel protein TASK-1|cardiac two pore background K(+) channel|two pore K(+) channel KT3.1|two pore potassium channel KT3.1', map_location='5 B', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30588170, end_position_on_the_genomic_accession=30625270, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16528, CurrentID=0, Status='live', Symbol='Kcnk4', Aliases='MLZ-622, TRAAK, TRAAKt, Tex40', description='potassium channel, subfamily K, member 4', other_designations='TWIK-related arachidonic acid-stimulated potassium channel protein|protein expressed in male leptotene and zygotene spermatocytes 622', map_location='19 5.08 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6925662, end_position_on_the_genomic_accession=6935372, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16529, CurrentID=0, Status='live', Symbol='Kcnk5', Aliases='KCNK5b, TASK-2', description='potassium channel, subfamily K, member 5', other_designations='potassium channel, subfamily, member 5 (KCNK5)|potassium inwardly-rectifying channel, subfamily K, member 5', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20140057, end_position_on_the_genomic_accession=20181782, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16530, CurrentID=0, Status='live', Symbol='Kcnk7', Aliases='2310014G03Rik, Kcnk6, Kcnk8, Knot, Mlk3', description='potassium channel, subfamily K, member 7', other_designations='double-pore K(+) channel 3|neuromuscular two p domain potassium channel|putative potassium channel DP3', map_location='19 B2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5704476, end_position_on_the_genomic_accession=5707405, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16531, CurrentID=0, Status='live', Symbol='Kcnma1', Aliases='5730414M22Rik, BKCa, MaxiK, Slo, Slo1, mSlo, mSlo1', description='potassium large conductance calcium-activated channel, subfamily M, alpha member 1', other_designations='BK channel alpha subunit|BKCA alpha|KCa1.1|calcium-activated potassium channel, subfamily M subunit alpha-1|k(VCA)alpha|maxi K channel|slo homolog|slo-alpha|slowpoke homolog', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=23292291, end_position_on_the_genomic_accession=24004951, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  16532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16532, CurrentID=0, Status='live', Symbol='Kcnu1', Aliases='Kcnma3, Slo3, mSlo3', description='potassium channel, subfamily U, member 1', other_designations='KSper|calcium-activated potassium channel subunit alpha-3|calcium-activated potassium channel, subfamily M subunit alpha-3|pH-sensitive maxi potassium channel|pore-forming subunit of the sperm-specific alkalization activated K(+) current|potassium large conductance pH-sensitive channel, subfamily M, alpha member 3|slowpoke homolog 3', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25842654, end_position_on_the_genomic_accession=25937943, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  16533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16533, CurrentID=0, Status='live', Symbol='Kcnmb1', Aliases='BKbeta1', description='potassium large conductance calcium-activated channel, subfamily M, beta member 1', other_designations='BK channel beta subunit|BK channel subunit beta-1|BKbeta|calcium-activated potassium channel subunit beta|calcium-activated potassium channel, subfamily M subunit beta-1|charybdotoxin receptor subunit beta-1|k(VCA)beta-1|maxi K channel subunit beta-1|slo-beta|slo-beta-1', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=33962951, end_position_on_the_genomic_accession=33973641, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16534, CurrentID=0, Status='live', Symbol='Kcnn4', Aliases='IK1, IKCA1, KCA4, KCa3.1, SK4, SKCas, mIKCa1', description='potassium intermediate/small conductance calcium-activated channel, subfamily N, member 4', other_designations='SKCa 4|SKCa4', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24369833, end_position_on_the_genomic_accession=24385212, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  278679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278679, CurrentID=0, Status='live', Symbol='Apol7b', Aliases='', description='apolipoprotein L 7b', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77422209, end_position_on_the_genomic_accession=77447460, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16536, CurrentID=0, Status='live', Symbol='Kcnq2', Aliases='HNSPC, KQT2, Nmf134', description='potassium voltage-gated channel, subfamily Q, member 2', other_designations='KQT-like 2|potassium channel subunit alpha KvLQT2|voltage-gated potassium channel subunit Kv7.2', map_location='2 H3-4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181075579, end_position_on_the_genomic_accession=181135289, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  16538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16538, CurrentID=0, Status='live', Symbol='Kcns1', Aliases='Kv9.1', description='K+ voltage-gated channel, subfamily S, 1', other_designations='delayed-rectifier K(+) channel alpha subunit 1|voltage-gated potassium channel subunit Kv9.1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164163619, end_position_on_the_genomic_accession=164171113, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16539, CurrentID=0, Status='live', Symbol='Kcns2', Aliases='E130006J24Rik, Kv9.2', description='K+ voltage-gated channel, subfamily S, 2', other_designations='delayed-rectifier K(+) channel alpha subunit 2|voltage-gated potassium channel subunit Kv9.2', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34837355, end_position_on_the_genomic_accession=34843407, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16541, CurrentID=0, Status='live', Symbol='Napsa', Aliases='KAP, Kdap, NAP1, SNAPA, pronapsin', description='napsin A aspartic peptidase', other_designations='KDAP-1|kidney-derived aspartic protease-like protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44572445, end_position_on_the_genomic_accession=44586846, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16542, CurrentID=0, Status='live', Symbol='Kdr', Aliases='6130401C07, Flk-1, Flk1, Krd-1, Ly73, VEGFR-2, VEGFR2, sVEGFR-2', description='kinase insert domain protein receptor', other_designations='VEGF receptor-2|fetal liver kinase 1|kinase NYK|protein-tyrosine kinase receptor flk-1|soluble vascular endothelial growth factor receptor 2|vascular endothelial growth factor receptor- 2|vascular endothelial growth factor receptor-2|vascular endothelial growth factor receptor-3', map_location='5 40.23 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=75932827, end_position_on_the_genomic_accession=75979074, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  16543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16543, CurrentID=0, Status='live', Symbol='Mdfic', Aliases='Kdt1, Mdfid', description='MyoD family inhibitor domain containing', other_designations='I-mfa domain-containing protein|kidney cell line-derived transcript 1|mdfic', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=15720584, end_position_on_the_genomic_accession=15802169, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16545, CurrentID=0, Status='live', Symbol='Kera', Aliases='CNA2, SLRR2B', description='keratocan', other_designations='KTN|keratan sulfate proteoglycan keratocan', map_location='10 50.34 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=97606959, end_position_on_the_genomic_accession=97613693, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16548, CurrentID=0, Status='live', Symbol='Khk', Aliases='', description='ketohexokinase', other_designations='hepatic fructokinase', map_location='5 16.9 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30921436, end_position_on_the_genomic_accession=30931248, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16549, CurrentID=0, Status='live', Symbol='Khsrp', Aliases='6330409F21Rik, Fbp2, Fubp2, Ksrp', description='KH-type splicing regulatory protein', other_designations='FUSE-binding protein 2|KH type-splicing regulatory protein', map_location='17 29.63 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57021049, end_position_on_the_genomic_accession=57031507, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  16551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16551, CurrentID=0, Status='live', Symbol='Kif11', Aliases='Eg5, Kif8, Kifl1, Knsl1', description='kinesin family member 11', other_designations='kinesin 11|kinesin-like 1|kinesin-related motor protein Eg5', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37376403, end_position_on_the_genomic_accession=37421859, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  16552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16552, CurrentID=0, Status='live', Symbol='Kif12', Aliases='', description='kinesin family member 12', other_designations='N-9 kinesin|kinesin 12|kinesin superfamily protein 12', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63165629, end_position_on_the_genomic_accession=63172142, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  16553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16553, CurrentID=0, Status='live', Symbol='Kif13a', Aliases='4930505I07Rik, mKIAA4109', description='kinesin family member 13A', other_designations='N-3 kinesin|kinesin 13A', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=46749087, end_position_on_the_genomic_accession=46931300, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  16554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16554, CurrentID=0, Status='live', Symbol='Kif13b', Aliases='5330429L19Rik, 6030414C01, AI429803, C130021D12Rik, GAKIN', description='kinesin family member 13B', other_designations='N-3 kinesin|kinesin 13B', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64652490, end_position_on_the_genomic_accession=64809618, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  16558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16558, CurrentID=0, Status='live', Symbol='Kif16b', Aliases='8430434E15Rik, AA623607, AI789011, C80253, C80902, mKIAA1590', description='kinesin family member 16B', other_designations='N-3 kinesin|kinesin 16B', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=142618334, end_position_on_the_genomic_accession=142901595, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  16559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16559, CurrentID=0, Status='live', Symbol='Kif17', Aliases='5930435E01Rik, AW492270b, mKIAA1405, Kif17', description='kinesin family member 17', other_designations='N-4 kinesin|kinesin 17', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138262251, end_position_on_the_genomic_accession=138301973, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  16560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16560, CurrentID=0, Status='live', Symbol='Kif1a', Aliases='ATSV, C630002N23Rik, Gm1626, Kns1', description='kinesin family member 1A', other_designations='N-3 kinesin|axonal transporter of synaptic vesicles|kinesin heavy chain member 1A', map_location='1 E1-E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93015455, end_position_on_the_genomic_accession=93101890, orientation='minus', exon_count=51, OMIM='', no_hdr0=''),
  16561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16561, CurrentID=0, Status='live', Symbol='Kif1b', Aliases='A530096N05Rik, AI448212, AI506502, D4Mil1e, KIF1Bp130, KIF1Bp204', description='kinesin family member 1B', other_designations='N-3 kinesin|kinesin heavy chain member 1B', map_location='4 E', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149176319, end_position_on_the_genomic_accession=149307733, orientation='minus', exon_count=51, OMIM='', no_hdr0=''),
  16562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16562, CurrentID=0, Status='live', Symbol='Kif1c', Aliases='B430105J22Rik, D11Bwg1349e, Ltxs1', description='kinesin family member 1C', other_designations='N-3 kinsin|kinesin 1C|kinesin superfamily protein 1C|lethal factor toxin susceptibility 1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70700548, end_position_on_the_genomic_accession=70731970, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  16563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16563, CurrentID=0, Status='live', Symbol='Kif2a', Aliases='C530030B14Rik, Kif2, Kns2, M-kinesin', description='kinesin family member 2A', other_designations='kinesin heavy chain member 2|kinesin-2', map_location='13 F', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=106960585, end_position_on_the_genomic_accession=107022228, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  16564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16564, CurrentID=0, Status='live', Symbol='Kif21a', Aliases='', description='kinesin family member 21A', other_designations='N-5 kinesin', map_location='15 45.86 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=90933275, end_position_on_the_genomic_accession=91049951, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  16565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16565, CurrentID=0, Status='live', Symbol='Kif21b', Aliases='2610511N21Rik, mKIAA0449', description='kinesin family member 21B', other_designations='N-5 kinesin|kinesin-like protein KIF6', map_location='1 59.61 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136131374, end_position_on_the_genomic_accession=136178014, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  16568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16568, CurrentID=0, Status='live', Symbol='Kif3a', Aliases='Kif3, Kifl, Kns3', description='kinesin family member 3A', other_designations='N-4 kinesin|kinesin-II subunit|microtubule plus end-directed kinesin motor 3A', map_location='11 31.97 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53567195, end_position_on_the_genomic_accession=53604248, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  16569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16569, CurrentID=0, Status='live', Symbol='Kif3b', Aliases='AI854312, AW549267, mKIAA0359', description='kinesin family member 3B', other_designations='microtubule plus end-directed kinesin motor 3B', map_location='2 75.41 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153291409, end_position_on_the_genomic_accession=153333390, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  16570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16570, CurrentID=0, Status='live', Symbol='Kif3c', Aliases='mKIAA4058', description='kinesin family member 3C', other_designations='N-4 kinesin', map_location='12 1.77 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3365116, end_position_on_the_genomic_accession=3406494, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16571, CurrentID=0, Status='live', Symbol='Kif4', Aliases='AI323435, D330050K22Rik, Kns4', description='kinesin family member 4', other_designations='N-5 kinesin|chromokinesin|kinesin heavy chain member 4', map_location='X 43.72 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100626065, end_position_on_the_genomic_accession=100727271, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  16572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16572, CurrentID=0, Status='live', Symbol='Kif5a', Aliases='D10Bwg0738e, Khc, Kif5, Kns, mKIAA4086', description='kinesin family member 5A', other_designations='NKHC|kinesin heavy chain neuron-specific 1|neuronal kinesin heavy chain', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127225695, end_position_on_the_genomic_accession=127263363, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  16573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16573, CurrentID=0, Status='live', Symbol='Kif5b', Aliases='AL022807, Khc, Khcs, Kns1, Ukhc', description='kinesin family member 5B', other_designations='conventional kinesin heavy chain|kinesin heavy chain member 5B, ubiquitous|ubiquitous kinesin heavy chain', map_location='18 4.46 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=6201005, end_position_on_the_genomic_accession=6241524, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  16574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16574, CurrentID=0, Status='live', Symbol='Kif5c', Aliases='KINN, Khc, NKHC, NKHC2', description='kinesin family member 5C', other_designations='kinesin heavy chain member 5C|kinesin heavy chain neuron-specific 2', map_location='2 28.68 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=49619314, end_position_on_the_genomic_accession=49774778, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  16576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16576, CurrentID=0, Status='live', Symbol='Kif7', Aliases='', description='kinesin family member 7', other_designations='kinesin 7', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79698098, end_position_on_the_genomic_accession=79715772, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  16578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16578, CurrentID=0, Status='live', Symbol='Kif9', Aliases='', description='kinesin family member 9', other_designations='kinesin 9', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110476786, end_position_on_the_genomic_accession=110525174, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  16579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16579, CurrentID=0, Status='live', Symbol='Kifap3', Aliases='KAP-3, KAP3, SMAP', description='kinesin-associated protein 3', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=163762139, end_position_on_the_genomic_accession=163917104, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  16580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16580, CurrentID=0, Status='live', Symbol='Kifc5b', Aliases='Hset, KNSL2, Kifc5a', description='kinesin family member C5B', other_designations='kinesin family c-terminal 5A|kinesin family c-terminal 5B|kinesin family member C5A|kinesin-related protein KIFC5A', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26917091, end_position_on_the_genomic_accession=26932579, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  16581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16581, CurrentID=0, Status='live', Symbol='Kifc2', Aliases='', description='kinesin family member C2', other_designations='', map_location='15 D-E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76659837, end_position_on_the_genomic_accession=76668196, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  16582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16582, CurrentID=0, Status='live', Symbol='Kifc3', Aliases='AI325457, BB123200', description='kinesin family member C3', other_designations='', map_location='8 47.12 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95099825, end_position_on_the_genomic_accession=95202857, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  16588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16588, CurrentID=0, Status='live', Symbol='Kin', Aliases='Kin17', description='Kin17 DNA and RNA binding protein', other_designations='KIN, antigenic determinant of recA protein homolog|antigenic determinant of rec-A protein|binding to curved DNA', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10080598, end_position_on_the_genomic_accession=10092701, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16589, CurrentID=0, Status='live', Symbol='Uhmk1', Aliases='4732477C12Rik, 4930500M09Rik, AA673513, AI449218, AU021979, KIS, Kist, P-CIP2', description='U2AF homology motif (UHM) kinase 1', other_designations='PAM COOH-terminal interactor protein 2|U2AF homology motif kinase 1|kinase interacting with leukemia-associated gene (stathmin)|kinase interacting with stathmin', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170199256, end_position_on_the_genomic_accession=170215898, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16590, CurrentID=0, Status='live', Symbol='Kit', Aliases='Bs, CD117, Fdc, Gsfsco1, Gsfsco5, Gsfsow3, SCO1, SCO5, SOW3, Ssm, Tr-kit, W, c-KIT', description='kit oncogene', other_designations='Dominant white spotting|SCFR|Steel Factor Receptor|belly-spot|c-kit proto-oncogene protein|dominant spotting|proto-oncogene c-Kit|proto-oncogene tyrosine-protein kinase Kit|spotted sterile male|tyrosine-protein kinase Kit', map_location='5 39.55 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=75574987, end_position_on_the_genomic_accession=75656722, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  16591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16591, CurrentID=0, Status='live', Symbol='Kl', Aliases='alpha-kl', description='klotho', other_designations='secreted form of Klotho protein', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=150952607, end_position_on_the_genomic_accession=150993817, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16592, CurrentID=0, Status='live', Symbol='Fabp5', Aliases='E-FABP, Fabpe, Klbp, PA-FABP, mal1', description='fatty acid binding protein 5, epidermal', other_designations='epidermal-type fatty acid-binding protein|epithelial fatty acid-binding protein|keratinocyte lipid-binding protein|psoriasis-associated fatty acid-binding protein homolog', map_location='3 A1-A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10012585, end_position_on_the_genomic_accession=10016610, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16593, CurrentID=0, Status='live', Symbol='Klc1', Aliases='AI874768, Kns2', description='kinesin light chain 1', other_designations='KLC 1|kinesin 2 60/70kDa', map_location='12 61.13 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111758849, end_position_on_the_genomic_accession=111807844, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  16594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16594, CurrentID=0, Status='live', Symbol='Klc2', Aliases='8030455F02Rik, AW212649, KLC 2, KLC-2', description='kinesin light chain 2', other_designations='', map_location='19 4.25 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5107746, end_position_on_the_genomic_accession=5118570, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  16596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16596, CurrentID=0, Status='live', Symbol='Klf1', Aliases='Eklf, Nan', description='Kruppel-like factor 1 (erythroid)', other_designations='erythroid krueppel-like transcription factor|erythroid transcription factor|neonatal anemia', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84901928, end_position_on_the_genomic_accession=84905295, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16597, CurrentID=0, Status='live', Symbol='Klf12', Aliases='2700063E05Rik, AI225908, AI315654, AP-2rep, B130052C06Rik, D530033K05Rik', description='Kruppel-like factor 12', other_designations='transcriptional repressor AP-2rep', map_location='14 50.9 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=99866181, end_position_on_the_genomic_accession=100284714, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16598, CurrentID=0, Status='live', Symbol='Klf2', Aliases='Lklf', description='Kruppel-like factor 2 (lung)', other_designations='lung krueppel-like factor', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72319062, end_position_on_the_genomic_accession=72321654, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16599, CurrentID=0, Status='live', Symbol='Klf3', Aliases='9930027G08Rik, AL024007, Bklf, Tef-2', description='Kruppel-like factor 3 (basic)', other_designations='CACCC-box-binding protein BKLF|basic krueppel-like factor', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=64803523, end_position_on_the_genomic_accession=64832901, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16600, CurrentID=0, Status='live', Symbol='Klf4', Aliases='EZF, Gklf, Zie', description='Kruppel-like factor 4 (gut)', other_designations='Kruppel-like factor 4 (Gut enriched kruppel-like factor) (Epithelial zinc-finger protein EZF)|epithelial zinc finger protein EZF|gut-enriched krueppel-like factor', map_location='4 29.76 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=55527137, end_position_on_the_genomic_accession=55532475, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16601, CurrentID=0, Status='live', Symbol='Klf9', Aliases='2310051E17Rik, AA589643, BTEB-1, Bteb1, ENSMUSG00000055440, Gm9971', description='Kruppel-like factor 9', other_designations='BTE-binding protein 1|GC-box-binding protein 1|basic transcription element binding protein 1|basic transcription element-binding protein 1|transcription factor BTEB1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=23141226, end_position_on_the_genomic_accession=23166911, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16602, CurrentID=0, Status='live', Symbol='Klk1b2-ps', Aliases='Klk-ps1, mGK-2', description='kallikrein 1-related peptidase b2, pseudogene', other_designations='kallikrein 1-related petidase b26 pseudogene|kallikrein, pseudogene 1', map_location='7 28.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43958822, end_position_on_the_genomic_accession=43962996, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  16604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16604, CurrentID=0, Status='live', Symbol='Klk1b7-ps', Aliases='1700127D06Rik, Klk-ps2, mGK-7', description='kallikrein 1-related peptidase b7, pseudogene', other_designations='kallikrein, pseudogene 2', map_location='7 28.42 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43945012, end_position_on_the_genomic_accession=43945927, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16612, CurrentID=0, Status='live', Symbol='Klk1', Aliases='0610007D04Rik, Kalb6, Klk6, mGk-6, Klk1', description='kallikrein 1', other_designations='KAL-B|glandular kallikrein K1|kallikrein 6|kallikrein renal/pancreas/salivary|renal kallikrein|tissue kallikrein-6|true tissue kallikrein', map_location='7 28.74 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44225437, end_position_on_the_genomic_accession=44229617, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16613, CurrentID=0, Status='live', Symbol='Klk1b11', Aliases='Klk11, mGK-11', description='kallikrein 1-related peptidase b11', other_designations='glandular kallikrein K11|kallikrein 11|tissue kallikrein-11', map_location='7 28.48 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43995879, end_position_on_the_genomic_accession=43999875, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16615, CurrentID=0, Status='live', Symbol='Klk1b16', Aliases='Klk16, mGk-16', description='kallikrein 1-related peptidase b16', other_designations='gamma-renin, submandibular gland|glandular kallikrein K16|kallikrein 16', map_location='7 28.64 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44136767, end_position_on_the_genomic_accession=44141604, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16616, CurrentID=0, Status='live', Symbol='Klk1b21', Aliases='Klk21, mGk-21', description='kallikrein 1-related peptidase b21', other_designations='glandular kallikrein K21|kallikrein 21|tissue kallikrein 21', map_location='7 28.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44102292, end_position_on_the_genomic_accession=44106583, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16617, CurrentID=0, Status='live', Symbol='Klk1b24', Aliases='Klk24, mGk-24', description='kallikrein 1-related peptidase b24', other_designations='glandular kallikrein K24|kallikrein 24|tissue kallikrein 24', map_location='7 28.7 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44188230, end_position_on_the_genomic_accession=44192455, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16618, CurrentID=0, Status='live', Symbol='Klk1b26', Aliases='Egfbp2, Klk26, PRECE-2, mGK-26, mGK26', description='kallikrein 1-related petidase b26', other_designations='EGF-BP type B|glandular kallikrein K26|kallikrein 1-related petidase b25|kallikrein 26|prorenin-converting enzyme 2|tissue kallikrein 26', map_location='7 28.5 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44007953, end_position_on_the_genomic_accession=44016969, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16619, CurrentID=0, Status='live', Symbol='Klk1b27', Aliases='Gk27, Klk21l, Klk27, mGK-27', description='kallikrein 1-related peptidase b27', other_designations='glandular kallikrein 27|glandular kallikrein K27|kallikrein 27|mKlk27|tissue kallikrein 27', map_location='7 28.54 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44052290, end_position_on_the_genomic_accession=44056712, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16621, CurrentID=0, Status='live', Symbol='Klkb1', Aliases='APS, Kal-3, Kal3, Klk3, PSA', description='kallikrein B, plasma 1', other_designations='antigen, prostate specific|fletcher factor|kallikrein 3, plasma|kininogenin|plasma prekallikrein', map_location='8 B1-B3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45269442, end_position_on_the_genomic_accession=45294835, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  16622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16622, CurrentID=0, Status='live', Symbol='Klk1b5', Aliases='Klk5, mGK-5', description='kallikrein 1-related peptidase b5', other_designations='glandular kallikrein K5|kallikrein 5|tissue kallikrein-5', map_location='7 28.73 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44216475, end_position_on_the_genomic_accession=44220703, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16623, CurrentID=0, Status='live', Symbol='Klk1b1', Aliases='Klk1, TK, mGK-1, mK1', description='kallikrein 1-related peptidase b1', other_designations='glandular kallikrein K1|tissue kallikrein|tissue kallikrein-1', map_location='7 28.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43966768, end_position_on_the_genomic_accession=43971316, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16624, CurrentID=0, Status='live', Symbol='Klk1b8', Aliases='Klk8, Nrpn, TADG14, mGK-8', description='kallikrein 1-related peptidase b8', other_designations='glandular kallikrein K8|kallikrein 8|tissue kallikrein-8', map_location='7 28.43 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43950672, end_position_on_the_genomic_accession=43954938, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16625, CurrentID=0, Status='live', Symbol='Serpina3c', Aliases='1A1, Kalbp, Klkbp, spi2', description='serine (or cysteine) peptidase inhibitor, clade A, member 3C', other_designations='KBP|alpha-1 antiproteinase|kallikrein binding protein|kallikrein-binding protein|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 3C|serine (or cysteine) proteinase inhibitor, clade A, member 3C|serpin A3C', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104146907, end_position_on_the_genomic_accession=104153933, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16627, CurrentID=0, Status='live', Symbol='Klra1', Aliases='A1, CH29-493D4.3, Klra22, Ly49a, Ly49o<129>, Ly49v', description='killer cell lectin-like receptor, subfamily A, member 1', other_designations='class I MHC receptor KLRA22|ly-49a|lymphocyte antigen 49a|natural killer cell receptor Ly49A|novel killer cell lectin-like receptor subfamily A protein|t lymphocyte antigen A1', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130363918, end_position_on_the_genomic_accession=130386874, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16628, CurrentID=0, Status='live', Symbol='Klra10', Aliases='Ly49J, Ly49i2', description='killer cell lectin-like receptor subfamily A, member 10', other_designations='', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130266495, end_position_on_the_genomic_accession=130284998, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16630, CurrentID=0, Status='live', Symbol='Klra12', Aliases='Ly49L, Ly49l1, Ly49l2, Ly49l3, Ly49l4, Ly49p/d', description='killer cell lectin-like receptor subfamily A, member 12', other_designations='novel killer cell lectin-like receptor, subfamily A protein', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16631, CurrentID=0, Status='live', Symbol='Klra13-ps', Aliases='Klra13, Ly49M', description='killer cell lectin-like receptor subfamily A, member 13, pseudogene', other_designations='', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130291161, end_position_on_the_genomic_accession=130306432, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16633, CurrentID=0, Status='live', Symbol='Klra2', Aliases='Klra30, Ly49b', description='killer cell lectin-like receptor, subfamily A, member 2', other_designations='NK cell inhibitory receptor Klra2|T-cell surface glycoprotein Ly-49B|T-cell surface glycoprotein Ly49b|killer cell lectin-like receptor subfamily A member 30|ly-49b|lymphocyte antigen 49b|membrane protein Ly49b', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131198976, end_position_on_the_genomic_accession=131247723, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16634, CurrentID=0, Status='live', Symbol='Klra3', Aliases='5E6, Ly49c, NK-2.1, Nk-2, Nk2, Nk2.1', description='killer cell lectin-like receptor, subfamily A, member 3', other_designations='T-cell surface glycoprotein Ly-49C|lymphocyte antigen 49c|natural killer cell-associated antigen 2', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130323289, end_position_on_the_genomic_accession=130337626, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16635, CurrentID=0, Status='live', Symbol='Klra4', Aliases='Chok, Klra32, Klra33, Ly-49d, Ly49d, ly49r<129>', description='killer cell lectin-like receptor, subfamily A, member 4', other_designations='T-cell surface glycoprotein Ly-49D|killer cell lectin-like receptor subfamily A member 32|killer cell lectin-like receptor subfamily A member 33|lymphocyte antigen 49d', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130043730, end_position_on_the_genomic_accession=130067271, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16636, CurrentID=0, Status='live', Symbol='Klra5', Aliases='Ly49e', description='killer cell lectin-like receptor, subfamily A, member 5', other_designations='Ly49E receptor|T-cell surface glycoprotein Ly-49E|class I MHC receptor KLRA5|ly-49e|lymphocyte antigen 49e', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129897752, end_position_on_the_genomic_accession=129913224, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16637, CurrentID=0, Status='live', Symbol='Klra6', Aliases='Ly49f, ly-49f', description='killer cell lectin-like receptor, subfamily A, member 6', other_designations='T-cell surface glycoprotein Ly-49F|lymphocyte antigen 49f', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130013033, end_position_on_the_genomic_accession=130026960, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16638, CurrentID=0, Status='live', Symbol='Klra7', Aliases='LGL-1, Ly-49G.1, Ly-49G.2, Ly-49G.3, Ly49G.4, Ly49g, Ly49g1, Ly49g2', description='killer cell lectin-like receptor, subfamily A, member 7', other_designations='T-cell surface glycoprotein Ly-49G|ly-49g|lymphocyte antigen 49g', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130218605, end_position_on_the_genomic_accession=130233322, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16639, CurrentID=0, Status='live', Symbol='Klra8', Aliases='Cmv-1, Cmv1, Ly49h, Ly49u<129>', description='killer cell lectin-like receptor, subfamily A, member 8', other_designations='T-cell surface glycoprotein Ly-49H|cytomegalovirus resistance 1|ly-49h|lymphocyte antigen 49h', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130115226, end_position_on_the_genomic_accession=130129898, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16640, CurrentID=0, Status='live', Symbol='Klra9', Aliases='LY49I1, Ly49I', description='killer cell lectin-like receptor subfamily A, member 9', other_designations='Ly49I receptor', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130178626, end_position_on_the_genomic_accession=130194374, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16641, CurrentID=0, Status='live', Symbol='Klrc1', Aliases='CD159a, NKG2A, NKG2B', description='killer cell lectin-like receptor subfamily C, member 1', other_designations='Cluster of Differentiation 159a', map_location='6 G1-2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129666015, end_position_on_the_genomic_accession=129682852, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16642, CurrentID=0, Status='live', Symbol='Klrc2', Aliases='NKG2C', description='killer cell lectin-like receptor subfamily C, member 2', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129649602, end_position_on_the_genomic_accession=129660737, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16643, CurrentID=0, Status='live', Symbol='Klrd1', Aliases='CD94', description='killer cell lectin-like receptor, subfamily D, member 1', other_designations='', map_location='6 63.42 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129588092, end_position_on_the_genomic_accession=129598775, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16644, CurrentID=0, Status='live', Symbol='Kng1', Aliases='Kng', description='kininogen 1', other_designations='H-kininigen|L-kininogen', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23058220, end_position_on_the_genomic_accession=23082078, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  16646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16646, CurrentID=0, Status='live', Symbol='Kpna1', Aliases='AW494490, IPOA5, NPI1, Rch2, mSRP1', description='karyopherin (importin) alpha 1', other_designations='NPI-1|RAG cohort protein 2|SRP1-beta|importin alpha-S1|importin subunit alpha-1|karyopherin subunit alpha-1|m-importin-alpha-S1|nucleoprotein interactor 1', map_location='16 B4-B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35983363, end_position_on_the_genomic_accession=36037136, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16647, CurrentID=0, Status='live', Symbol='Kpna2', Aliases='2410044B12Rik, IPOA1, PTAC58, Rch1', description='karyopherin (importin) alpha 2', other_designations='RAG cohort protein 1|SRP1-alpha|importin alpha P1|importin subunit alpha-2|karyopherin subunit alpha-2|m-importin-alpha-P1|pendulin|pore targeting complex 58 kDa subunit', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106988629, end_position_on_the_genomic_accession=106999525, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16648, CurrentID=0, Status='live', Symbol='Kpna3', Aliases='IPOA4', description='karyopherin (importin) alpha 3', other_designations='importin alpha Q2|importin subunit alpha-3|karyopherin subunit alpha-3|qip2', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61365186, end_position_on_the_genomic_accession=61439947, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  16649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16649, CurrentID=0, Status='live', Symbol='Kpna4', Aliases='1110058D08Rik, IPOA3', description='karyopherin (importin) alpha 4', other_designations='importin alpha Q1|importin subunit alpha-4|karyopherin subunit alpha-4|qip1', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69072221, end_position_on_the_genomic_accession=69127092, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  16650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16650, CurrentID=0, Status='live', Symbol='Kpna6', Aliases='IPOA7, Kpna5, NPI-2', description='karyopherin (importin) alpha 6', other_designations='importin alpha-S2|karyopherin (importin) alpha 5|karyopherin subunit alpha-6', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129643979, end_position_on_the_genomic_accession=129672767, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  16651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16651, CurrentID=0, Status='live', Symbol='Sspn', Aliases='Krag', description='sarcospan', other_designations='K-ras oncogene-associated protein|Kras oncogene-associated|kirsten-Ras-associated protein', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145931150, end_position_on_the_genomic_accession=145965225, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16653, CurrentID=0, Status='live', Symbol='Kras', Aliases='AI929937, K-ras, Ki-ras-2, Kras2, p21B, ras, Kras', description='Kirsten rat sarcoma viral oncogene homolog', other_designations='K-Ras 2|Kirsten rat sarcoma oncogene 2, expressed|c-K-ras|c-Ki-ras|p21 protein|v-Ki-ras2 Kirsten rat sarcoma viral oncogene homolog', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145216699, end_position_on_the_genomic_accession=145250300, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16656, CurrentID=0, Status='live', Symbol='Hivep3', Aliases='2900056N03Rik, A130075N07, AI848000, E030045D18Rik, KBP-1, KBP1, Krc, Rc, Schnurri-3, Shn3, Zas3', description='human immunodeficiency virus type I enhancer binding protein 3', other_designations='KB-binding and recognition component|KB-binding and regognition component|human immunodeficiency virus type I enhancer-binding protein 3 homolog|kappa B and Rss recognition component|kappa-B and V(D)J recombination signal sequences-binding protein|kappa-binding protein 1|recombinant component|zinc finger protein ZAS3', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119733785, end_position_on_the_genomic_accession=120138045, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16658, CurrentID=0, Status='live', Symbol='Mafb', Aliases='Kreisler, Krml, Krml1, kr', description='v-maf musculoaponeurotic fibrosarcoma oncogene family, protein B (avian)', other_designations='Kreisler (maf-related) leucine zipper protein|V-maf musculoaponeurotic fibrosarcoma oncogene homolog B|maf-B|segmentation protein Kr|transcription factor Maf-1', map_location='2 80.92 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=160363677, end_position_on_the_genomic_accession=160367065, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  16659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16659, CurrentID=0, Status='live', Symbol='Krt31c', Aliases='Krt-1, Krt1', description='keratin 31 complex', other_designations='keratin gene complex 1, acidic', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16660, CurrentID=0, Status='live', Symbol='Krt31', Aliases='Ha1, Kha1, Krt1-1, MKHA-1', description='keratin 31', other_designations='HKA-1|K31|hair keratin, type I Ha1|keratin complex 1, acidic, gene 1|keratin complex 1, gene 1|keratin-31', map_location='11 63.39 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100046646, end_position_on_the_genomic_accession=100050551, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16661, CurrentID=0, Status='live', Symbol='Krt10', Aliases='D130054E02Rik, K10, K1C1, Krt-1.10, Krt1-10', description='keratin 10', other_designations='56 kDa cytokeratin|CK-10|cytokeratin 10|cytokeratin-10|epidermal keratin 10|intermediate filament protein|keratin complex 1, acidic, gene 10|keratin, type I cytoskeletal 59 kDa|keratin-10|suprabasal cytokeratin 10', map_location='11 62.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99385254, end_position_on_the_genomic_accession=99389699, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16663, CurrentID=0, Status='live', Symbol='Krt13', Aliases='CK13, K13, Krt-1.13, Krt1-13', description='keratin 13', other_designations='47 kDa cytokeratin|CK-13|cytokeratin 13|cytokeratin-13|keratin complex 1, acidic, gene 13|keratin-13', map_location='11 63.41 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100117596, end_position_on_the_genomic_accession=100121495, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16664, CurrentID=0, Status='live', Symbol='Krt14', Aliases='AI626930, CK-14, K14, Krt-1.14, Krt1-14', description='keratin 14', other_designations='cytokeratin 14|cytokeratin-14|keratin complex 1, acidic, gene 14|keratin-14', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100203162, end_position_on_the_genomic_accession=100207510, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16665, CurrentID=0, Status='live', Symbol='Krt15', Aliases='AI528832, K15, Krt1-15', description='keratin 15', other_designations='CK-15|cytokeratin-15|keratin complex 1, acidic, gene 15|keratin-15', map_location='11 63.41 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100131759, end_position_on_the_genomic_accession=100135949, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16666, CurrentID=0, Status='live', Symbol='Krt16', Aliases='AI324768, CK-16, K16, Krt1-16', description='keratin 16', other_designations='cytokeratin-16|keratin complex 1, acidic, gene 16|keratin intermediate filament 16a|keratin intermediate filament 16b|keratin-16', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100246091, end_position_on_the_genomic_accession=100248962, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16667, CurrentID=0, Status='live', Symbol='Krt17', Aliases='K17, Krt1-17', description='keratin 17', other_designations='CK-17|cytokeratin-17|keratin complex 1, acidic, gene 17|keratin-17', map_location='11 63.44 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100256217, end_position_on_the_genomic_accession=100260989, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16668, CurrentID=0, Status='live', Symbol='Krt18', Aliases='CK18, K18, Krt1-18', description='keratin 18', other_designations='cytokeratin endo B|cytokeratin-18|keratin D|keratin complex 1, acidic, gene 18', map_location='15 57.22 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102028216, end_position_on_the_genomic_accession=102032026, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16669, CurrentID=0, Status='live', Symbol='Krt19', Aliases='AI663979, CK-19, EndoC, K19, Krt-1.19, Krt1-19', description='keratin 19', other_designations='cytokeratin 19|keratin complex 1, acidic, gene 19|keratin-19', map_location='11 63.42 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100140810, end_position_on_the_genomic_accession=100145919, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16670, CurrentID=0, Status='live', Symbol='Krt32', Aliases='Ha3, Hka2, K32, Krt1-2, Krtha2', description='keratin 32', other_designations='hair keratin, type I Ha2|keratin complex 1, acidic, gene 2', map_location='11 63.4 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100080848, end_position_on_the_genomic_accession=100088226, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16671, CurrentID=0, Status='live', Symbol='Krt33b', Aliases='Ha3, Ha4, Krt1-3, mHa3', description='keratin 33B', other_designations='K33B|hair keratin, type I Ha3|keratin complex 1, acidic, gene 3|keratin complex 1, gene 3|keratin, type I cuticular Ha3|keratin-33B|ortholog of keratin, hair, acidic, 3B KRTHA3B', map_location='11 63.37 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100023634, end_position_on_the_genomic_accession=100029868, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16672, CurrentID=0, Status='live', Symbol='Krt34', Aliases='4733401E01Rik, Ha2, K34, Krt1-4', description='keratin 34', other_designations='hair keratin, type I Ha4|keratin complex 1, acidic, gene 4|keratin-34', map_location='11 63.38 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100037351, end_position_on_the_genomic_accession=100041554, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16673, CurrentID=0, Status='live', Symbol='Krt36', Aliases='HRa-1, K36, Krt1-22, Krt1-5, MHRa-1', description='keratin 36', other_designations='MHRa-1|keratin complex 1, acidic, gene 5|keratin-36|keratin-5|krt36|type I keratin 48 kDa (MHRa-1)', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100102012, end_position_on_the_genomic_accession=100114634, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16675, CurrentID=0, Status='live', Symbol='Krt27', Aliases='CK-27, Krt1-c29, mIRSa3.1, stpm', description='keratin 27', other_designations='cytokeratin-27|keratin complex-1, acidic, gene C29|keratin complex-1, gene C29|keratin-27|type I inner root sheath-specific keratin-K25irs3', map_location='11 62.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99345566, end_position_on_the_genomic_accession=99351094, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16678, CurrentID=0, Status='live', Symbol='Krt1', Aliases='Krt-2.1, Krt2-1, Krt86', description='keratin 1', other_designations='67 kDa cytokeratin|CK-1|K1|cytokeratin-1|keratin 86|keratin complex 2, basic, gene 1|keratin-1|type-II keratin Kb1', map_location='15 57.06 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101845426, end_position_on_the_genomic_accession=101850786, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16679, CurrentID=0, Status='live', Symbol='Krt86', Aliases='5430421N21Rik, AI325212, Khb4, Krt2-10, Krt2-11, MHb4', description='keratin 86', other_designations='K86|keratin complex 2, basic, gene 10|keratin-86|type II hair keratin Hb6|type-II keratin Kb26', map_location='15 56.9 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101473478, end_position_on_the_genomic_accession=101479983, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16680, CurrentID=0, Status='live', Symbol='Krt84', Aliases='AV235125, HRb-1, Krt2-16, Krt2-3', description='keratin 84', other_designations='65 kDa type II keratin|K84|keratin complex 2, basic, gene 16|keratin complex 2, gene 16|keratin-84|type II hair keratin Hb4|type-II keratin Kb24', map_location='15 56.9 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101516663, end_position_on_the_genomic_accession=101532959, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16681, CurrentID=0, Status='live', Symbol='Krt2', Aliases='BB005427-17, Krt2-2, Krt2e, Krt2', description='keratin 2', other_designations='CK-2e|K2e|cytokeratin-2e|epithelial keratin-2e|keratin 2 epidermis|keratin complex 2, basic, gene 17|keratin complex 2, gene 17|keratin-2 epidermis|keratin-2e|krt2|type-II keratin Kb2', map_location='15 57.03 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101810689, end_position_on_the_genomic_accession=101818169, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  16682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16682, CurrentID=0, Status='live', Symbol='Krt4', Aliases='AW108092, K4, Krt-2.4, Krt2-4', description='keratin 4', other_designations='57kDa keratin|57kd keratin|CK-4|cytokeratin-4|cytoskeletal 57 kDa keratin|keratin complex 2, basic, gene 4|keratin-4|type-II keratin Kb4', map_location='15 57.13 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101918535, end_position_on_the_genomic_accession=101924735, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16687, CurrentID=0, Status='live', Symbol='Krt6a', Aliases='2310016L08Rik, Krt2-6a, Krt2-6c, MK6a, mK6[a]', description='keratin 6A', other_designations='60-kDa keratin|CK-6A|K6A|cytokeratin-6A|keratin complex 2, basic, gene 6a|keratin complex 2, basic, gene 6c|keratin complex 2, gene 6a|keratin-6-alpha|keratin-6A|mK6-alpha', map_location='15 56.97 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101689928, end_position_on_the_genomic_accession=101694305, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16688, CurrentID=0, Status='live', Symbol='Krt6b', Aliases='BB144589, Krt2-6b, mK6[b]', description='keratin 6B', other_designations='CK-6B|K6B|cytokeratin-6B|keratin complex 2, basic, gene 6b|keratin-6-beta|keratin-6B|mK6-beta', map_location='15 56.97 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101676023, end_position_on_the_genomic_accession=101680289, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16691, CurrentID=0, Status='live', Symbol='Krt8', Aliases='AA960620, AL022697, AU019895, Card2, EndoA, K8, Krt-2.8, Krt2-8', description='keratin 8', other_designations='CK-8|cytokeratin 8|cytokeratin endo A|cytokeratin-8|keratin complex 2, basic, gene 8|keratin-8|type-II keratin Kb8', map_location='15 57.2 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101996711, end_position_on_the_genomic_accession=102004342, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16693, CurrentID=0, Status='live', Symbol='Krtap11-1', Aliases='AA162070, AA589467, AV018433, Hacl1, KAP11.1, PRED29', description='keratin associated protein 11-1', other_designations='keratin-associated protein 11-1|keratin-associated protein 11.1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=89570176, end_position_on_the_genomic_accession=89571183, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  16694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16694, CurrentID=0, Status='live', Symbol='Krtap12-1', Aliases='D10Jhu14e', description='keratin associated protein 12-1', other_designations='krtap12-1', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77720586, end_position_on_the_genomic_accession=77721256, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  16699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16699, CurrentID=0, Status='live', Symbol='Krtap13', Aliases='4C32', description='keratin associated protein 13', other_designations='keratin-associated protein 13', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88750744, end_position_on_the_genomic_accession=88751628, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  16700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16700, CurrentID=0, Status='live', Symbol='Krtap6-1', Aliases='HGTpII.1, HGTpII.2', description='keratin associated protein 6-1', other_designations='keratin-associated protein 6.1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=89031699, end_position_on_the_genomic_accession=89032292, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  16703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16703, CurrentID=0, Status='live', Symbol='Krtap8-1', Aliases='AI326176', description='keratin associated protein 8-1', other_designations='HGT keratin|HGTp type I F|HGTpI F|high-glycine/tyrosine protein keratin|keratin-associated protein 8.1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=89487374, end_position_on_the_genomic_accession=89487952, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  16705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16705, CurrentID=0, Status='live', Symbol='Krtap9-1', Aliases='', description='keratin associated protein 9-1', other_designations='keratin-associated protein 9-1|keratin-associated protein 9.1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99873389, end_position_on_the_genomic_accession=99874000, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  16706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16706, CurrentID=0, Status='live', Symbol='Ksr1', Aliases='AW492498, B-KSR1, D11Bhm183e, D11Bhm184e, Ksr, RSU2, hb', description='kinase suppressor of ras 1', other_designations='mKSR1|protein Hb', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79013440, end_position_on_the_genomic_accession=79146554, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  16709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16709, CurrentID=0, Status='live', Symbol='Ktn1', Aliases='', description='kinectin 1', other_designations='ktn1', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47648453, end_position_on_the_genomic_accession=47736564, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  16716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16716, CurrentID=0, Status='live', Symbol='Ky', Aliases='CD1, D9Mgc44e, ky', description='kyphoscoliosis peptidase', other_designations='', map_location='9 54.54 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=102505755, end_position_on_the_genomic_accession=102546252, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  16728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16728, CurrentID=0, Status='live', Symbol='L1cam', Aliases='CD171, L1, N-CAM-L1, NCAM-L1', description='L1 cell adhesion molecule', other_designations='', map_location='X A6-B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73853780, end_position_on_the_genomic_accession=73880841, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  16736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16736, CurrentID=0, Status='live', Symbol='L1Md-Tf30', Aliases='ORF1', description='L1 repeat, Tf subfamily, member 30', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16763, CurrentID=0, Status='live', Symbol='Lad1', Aliases='', description='ladinin', other_designations='LADA|lad-1|linear IgA disease antigen homolog', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135818598, end_position_on_the_genomic_accession=135833341, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  16764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16764, CurrentID=0, Status='live', Symbol='Aff3', Aliases='3222402O04Rik, A730046J16, LAF-4, Laf4', description='AF4/FMR2 family, member 3', other_designations='lymphoid nuclear protein 4|lymphoid nuclear protein related to AF4|protein LAF-4', map_location='1 16.58 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=38175991, end_position_on_the_genomic_accession=38627837, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  16765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16765, CurrentID=0, Status='live', Symbol='Stmn1', Aliases='19k, Lag, Lap18, Op18, P18, P19, Pig, Pp17, Pp18, Pp19, Pr22, Smn', description='stathmin 1', other_designations='leukemia associated phosphoprotein p18|leukemia-associated gene protein|leukemia-associated phosphoprotein p18|metablastin|oncoprotein 18|phosphoprotein p19|prosolin', map_location='4 66.76 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134468320, end_position_on_the_genomic_accession=134473843, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16768, CurrentID=0, Status='live', Symbol='Lag3', Aliases='CD223, LAG-3, Ly66', description='lymphocyte-activation gene 3', other_designations='', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124904359, end_position_on_the_genomic_accession=124912435, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16769, CurrentID=0, Status='live', Symbol='Dsg4', Aliases='CDHF13, lah', description='desmoglein 4', other_designations='', map_location='18 11.35 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20436175, end_position_on_the_genomic_accession=20471821, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16770, CurrentID=0, Status='live', Symbol='Lalba', Aliases='AW208827, Lyzg', description='lactalbumin, alpha', other_designations='lactose synthase B protein|lysozyme g', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98480400, end_position_on_the_genomic_accession=98482934, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16772, CurrentID=0, Status='live', Symbol='Lama1', Aliases='AA408497, Lama', description='laminin, alpha 1', other_designations='S-LAM alpha|S-laminin subunit alpha|laminin A chain|laminin-1 subunit alpha|laminin-3 subunit alpha', map_location='17 38.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=67697133, end_position_on_the_genomic_accession=67822645, orientation='plus', exon_count=64, OMIM='', no_hdr0=''),
  16773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16773, CurrentID=0, Status='live', Symbol='Lama2', Aliases='5830440B04, dy, mKIAA4087, mer, merosin', description='laminin, alpha 2', other_designations='dystrophia muscularis|laminin M chain|laminin-12 subunit alpha|laminin-2 subunit alpha|laminin-4 subunit alpha|merosin heavy chain', map_location='10 A4-B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=26981285, end_position_on_the_genomic_accession=27617051, orientation='minus', exon_count=66, OMIM='', no_hdr0=''),
  16774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16774, CurrentID=0, Status='live', Symbol='Lama3', Aliases='Lama3B, [a]3B', description='laminin, alpha 3', other_designations='epiligrin subunit alpha|kalinin subunit alpha|laminin-5 subunit alpha|laminin-6 subunit alpha|laminin-7 subunit alpha|nicein subunit alpha|nicein, 150kDa', map_location='18 A', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12333723, end_position_on_the_genomic_accession=12583013, orientation='plus', exon_count=78, OMIM='', no_hdr0=''),
  16775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16775, CurrentID=0, Status='live', Symbol='Lama4', Aliases='', description='laminin, alpha 4', other_designations='laminin [a]4|laminin-14 subunit alpha|laminin-8 subunit alpha|laminin-9 subunit alpha', map_location='10 20.02 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=38965515, end_position_on_the_genomic_accession=39110188, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  16776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16776, CurrentID=0, Status='live', Symbol='Lama5', Aliases='AA408760, AA408762, AI853660, [a]5, laminin-511, mKIAA0533', description='laminin, alpha 5', other_designations='laminin-10 subunit alpha|laminin-11 subunit alpha|laminin-15 subunit alpha', map_location='2 102.7 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180176373, end_position_on_the_genomic_accession=180225905, orientation='minus', exon_count=79, OMIM='', no_hdr0=''),
  16777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16777, CurrentID=0, Status='live', Symbol='Lamb1', Aliases='C77966, C80098, C81607, D130003D08Rik, Lamb-1-1, Lamb1', description='laminin B1', other_designations='laminin B1 chain|laminin B1 subunit 1|laminin-1 subunit beta|laminin-10 subunit beta|laminin-12 subunit beta|laminin-2 subunit beta|laminin-6 subunit beta|laminin-8 subunit beta', map_location='12 13.39 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31265148, end_position_on_the_genomic_accession=31329644, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  16779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16779, CurrentID=0, Status='live', Symbol='Lamb2', Aliases='AW211941, Lamb-2, Lams, npht', description='laminin, beta 2', other_designations='Laminin S|S-LAM beta|S-laminin subunit beta|laminin-11 subunit beta|laminin-14 subunit beta|laminin-15 subunit beta|laminin-3 subunit beta|laminin-4 subunit beta|laminin-7 subunit beta|laminin-9 subunit beta', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108479857, end_position_on_the_genomic_accession=108490530, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  16780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16780, CurrentID=0, Status='live', Symbol='Lamb3', Aliases='', description='laminin, beta 3', other_designations='epiligrin subunit bata|kalinin B1 chain|kalinin subunit beta|laminin-5 subunit beta|nicein subunit beta|nicein, 125kD|nicein, 125kDa', map_location='1 H2-H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193294353, end_position_on_the_genomic_accession=193343878, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  16782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16782, CurrentID=0, Status='live', Symbol='Lamc2', Aliases='AA589349', description='laminin, gamma 2', other_designations='epiligrin subunit gamma|kalinin subunit gamma|kalinin/nicein/epiligrin 100 kDa subunit|laminin B2t chain|laminin gamma2 chain|laminin-5 subunit gamma|nicein subunit gamma|nicein, 100kD|nicein, 100kDa', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153122756, end_position_on_the_genomic_accession=153186447, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  16783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16783, CurrentID=0, Status='live', Symbol='Lamp1', Aliases='CD107a, LGP-120, LGP-A, Lamp-1, P2B', description='lysosomal-associated membrane protein 1', other_designations='120 kDa lysosomal membrane glycoprotein|CD107 antigen-like family member A|LYSOSOME-ASSOCIATED MEMBRANE GLYCOPROTEIN 1 PRECURSOR (LAMP-1) (LGP-A) (LGP-120) (CD107A) (P2B)|Lamp I|lysosomal membrane glycoprotein 1|lysosomal membrane glycoprotein A|lysosome-associated membrane protein 1', map_location='8 5.73 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13159134, end_position_on_the_genomic_accession=13175338, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16784, CurrentID=0, Status='live', Symbol='Lamp2', Aliases='CD107b, LGP-B, Lamp II, Lamp-2, Lamp-2a, Lamp-2b, Lamp-2c, Mac3', description='lysosomal-associated membrane protein 2', other_designations='CD107 antigen-like family member B|lysosomal membrane glycoprotein 2|lysosomal membrane glycoprotein type B', map_location='X 22.67 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38401357, end_position_on_the_genomic_accession=38456460, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16785, CurrentID=0, Status='live', Symbol='Rpsa', Aliases='67kDa, 67lr, AL022858, Lamr, Lamr1, Lamrl1, MLR, P40, P40-3, P40-8', description='ribosomal protein SA', other_designations='37 kDa laminin receptor|37 kDa oncofetal antigen|37/67 kDa laminin receptor|37LRP|67 kDa laminin receptor|67LR|LBP/p40|LRP/LR|OFA/iLRP|P40-3, functional|P40-8, functional|lamR|laminin receptor 1 (67kD, ribosomal protein SA)|laminin-binding protein precursor p40', map_location='9 71.41 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120127766, end_position_on_the_genomic_accession=120132369, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16790, CurrentID=0, Status='live', Symbol='Anpep', Aliases='AP-M, AP-N, Apn, Cd13, P150', description='alanyl (membrane) aminopeptidase', other_designations='alanyl aminopeptidase|aminopeptidase M|aminopeptidase N/CD13|membrane protein p161|microsomal aminopeptidase', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79821803, end_position_on_the_genomic_accession=79848221, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  16792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16792, CurrentID=0, Status='live', Symbol='Laptm5', Aliases='E3', description='lysosomal-associated protein transmembrane 5', other_designations='lysosomal-associated multitransmembrane protein 5|retinoic acid-inducible E3 protein', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130913334, end_position_on_the_genomic_accession=130936148, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16795, CurrentID=0, Status='live', Symbol='Large', Aliases='BPFD#36, Gyltl1a, Mbp-1, Mbp1, enr, fg, froggy, myd', description='like-glycosyltransferase', other_designations='acetylglucosaminyltransferase-like 1A|myodystrophy', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72814599, end_position_on_the_genomic_accession=73353577, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  16796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16796, CurrentID=0, Status='live', Symbol='Lasp1', Aliases='AA408629, Def-4, SH3P6, Tg(Col1a1-lacZ)1Ngma', description='LIM and SH3 protein 1', other_designations='LASP-1|MLN 50|Tg(11Col1a1)NG|metastatic lymph node gene 50 protein', map_location='11 C-D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97799005, end_position_on_the_genomic_accession=97838765, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16797, CurrentID=0, Status='live', Symbol='Lat', Aliases='p36-38, pp36', description='linker for activation of T cells', other_designations='36 kDa phospho-tyrosine adapter protein|linker protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126363827, end_position_on_the_genomic_accession=126369589, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  16798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16798, CurrentID=0, Status='live', Symbol='Lats1', Aliases='AW208599', description='large tumor suppressor', other_designations='WARTS protein kinase|large tumor suppressor homolog 1|large tumor supressor, homolog 1|lats1', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7681209, end_position_on_the_genomic_accession=7716461, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16800, CurrentID=0, Status='live', Symbol='Arhgef2', Aliases='AA408978, GEF, GEF-H1, GEFH1, LFP40, Lbcl1, Lfc, P40, mKIAA0651', description='rho/rac guanine nucleotide exchange factor (GEF) 2', other_designations="LBC'S first cousin|guanine nucleotide exchange factor H1|lymphoid blast crisis-like 1|oncogene LFC|rhobin", map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88593114, end_position_on_the_genomic_accession=88649532, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  16801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16801, CurrentID=0, Status='live', Symbol='Arhgef1', Aliases='Lbcl2, Lsc', description='Rho guanine nucleotide exchange factor (GEF) 1', other_designations="lbc's second cousin|lymphoid blast crisis like 2|lymphoid blast crisis-like 2", map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24902917, end_position_on_the_genomic_accession=24926594, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  16803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16803, CurrentID=0, Status='live', Symbol='Lbp', Aliases='Bpifd2, Ly88', description='lipopolysaccharide binding protein', other_designations='', map_location='2 78.72 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158306493, end_position_on_the_genomic_accession=158332852, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  16814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16814, CurrentID=0, Status='live', Symbol='Lbx1', Aliases='Lbx1h', description='ladybird homeobox homolog 1 (Drosophila)', other_designations='lady bird-like homeobox 1 homolog|ladybird homeobox protein homolog 1', map_location='19 D', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45232684, end_position_on_the_genomic_accession=45237435, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16815, CurrentID=0, Status='live', Symbol='Lbx2', Aliases='Lbx2h', description='ladybird homeobox homolog 2 (Drosophila)', other_designations='lady bird-like homeobox 2 homolog|ladybird homeobox protein homolog 2', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83086365, end_position_on_the_genomic_accession=83088241, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16816, CurrentID=0, Status='live', Symbol='Lcat', Aliases='AI046659, D8Wsu61e', description='lecithin cholesterol acyltransferase', other_designations='lecithin-cholesterol acyltransferase Lcat|phospholipid-cholesterol acyltransferase', map_location='8 53.06 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105939551, end_position_on_the_genomic_accession=105943402, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16818, CurrentID=0, Status='live', Symbol='Lck', Aliases='Hck-3, Lsk, Lskt, p56<lck>, p56Lck', description='lymphocyte protein tyrosine kinase', other_designations='leukocyte C-terminal Src kinase|lymphocyte cell-specific protein-tyrosine kinase', map_location='4 63.26 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129548344, end_position_on_the_genomic_accession=129573641, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  68339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68339, CurrentID=0, Status='live', Symbol='Ccdc88c', Aliases='0610010D24Rik, AV000935, AW324073, Daple, mKIAA1509', description='coiled-coil domain containing 88C', other_designations='DVL-binding protein DAPLE|coiled-coil domain-containing protein 88C|dvl-associating protein with a high frequency of leucine residues', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100911521, end_position_on_the_genomic_accession=101028983, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  16820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16820, CurrentID=0, Status='live', Symbol='Lcn3', Aliases='Vnsp1', description='lipocalin 3', other_designations='VNSP I|lipocalin-3|vomeronasal secretory protein I', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25765569, end_position_on_the_genomic_accession=25768099, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16821, CurrentID=0, Status='live', Symbol='Lcn4', Aliases='A630045M08Rik, Vnsp2', description='lipocalin 4', other_designations='VNSP II|lipocalin-4|vomeronasal secretory protein II', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26667674, end_position_on_the_genomic_accession=26671282, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16822, CurrentID=0, Status='live', Symbol='Lcp2', Aliases='AI323664, BB161688, SLP-76, SLP76, twm', description='lymphocyte cytosolic protein 2', other_designations='SH2 domain-containing leukocyte protein of 76 kDa|SLP-76 tyrosine phosphoprotein', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=34047201, end_position_on_the_genomic_accession=34092280, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  16825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16825, CurrentID=0, Status='live', Symbol='Ldb1', Aliases='CLIM2, NLI', description='LIM domain binding 1', other_designations='CLIM-2|LDB-1|LIM domain-binding factor CLIM2|carboxyl-terminal LIM domain-binding protein 2|mLdb1|nuclear LIM interactor', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46032589, end_position_on_the_genomic_accession=46045516, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  16826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16826, CurrentID=0, Status='live', Symbol='Ldb2', Aliases='AI035351, AW146358, CLIM1, CLP-36, Ldb3', description='LIM domain binding 2', other_designations='LIM domain-binding factor CLIM1|carboxyl-terminal LIM domain-binding protein 1', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=44472131, end_position_on_the_genomic_accession=44799760, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16828, CurrentID=0, Status='live', Symbol='Ldha', Aliases='Ldh1, Ldhm, l7R2', description='lactate dehydrogenase A', other_designations='LDH muscle subunit|LDH-M|M-LDH|lactate dehydrogenase 1, A chain|lactate dehydrogenase A4|lactate dehydrogenase-A', map_location='7 30.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46845804, end_position_on_the_genomic_accession=46855627, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16832, CurrentID=0, Status='live', Symbol='Ldhb', Aliases='AI790582, H-Ldh, LDH-B, LDH-H, Ldh-2, Ldh2', description='lactate dehydrogenase B', other_designations='LDH heart subunit|lactate dehydrogenase 2, B chain', map_location='6 74.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142490249, end_position_on_the_genomic_accession=142507957, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16833, CurrentID=0, Status='live', Symbol='Ldhc', Aliases='LDH-C4, Ldh-3, Ldh-x, Ldh34, Ldhc', description='lactate dehydrogenase C', other_designations='LDH testis subunit|LDH-C|lactate dehydrogenase 3, C chain, sperm specific|lactate dehydrogenase-X', map_location='7 30.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46861263, end_position_on_the_genomic_accession=46878142, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  16834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16834, CurrentID=0, Status='live', Symbol='Cog1', Aliases='6030445I15, Ldlb, mKIAA1381', description='component of oligomeric golgi complex 1', other_designations='COG complex subunit 1|low density lipoprotein B|low density lipoprotein receptor defect B-complementing protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113649178, end_position_on_the_genomic_accession=113669331, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  16835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16835, CurrentID=0, Status='live', Symbol='Ldlr', Aliases='Hlb301', description='low density lipoprotein receptor', other_designations='LDL receptor', map_location='9 7.87 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21723576, end_position_on_the_genomic_accession=21749919, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  16840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16840, CurrentID=0, Status='live', Symbol='Lect1', Aliases='Bricd3, ChM-I', description='leukocyte cell derived chemotaxin 1', other_designations='Chondromodulin 1|chondromodulin-1|chondromodulin-I', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79637682, end_position_on_the_genomic_accession=79662199, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16841, CurrentID=0, Status='live', Symbol='Lect2', Aliases='', description='leukocyte cell-derived chemotaxin 2', other_designations='LECT-2|chM-II|chondromodulin II|chondromodulin-II', map_location='13 30.08 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56528959, end_position_on_the_genomic_accession=56548561, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16842, CurrentID=0, Status='live', Symbol='Lef1', Aliases='3000002B05, AI451430, Lef-1', description='lymphoid enhancer binding factor 1', other_designations='', map_location='3 60.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=131110263, end_position_on_the_genomic_accession=131224357, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  16846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16846, CurrentID=0, Status='live', Symbol='Lep', Aliases='ob, obese', description='leptin', other_designations='obesity factor', map_location='6 12.3 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29060195, end_position_on_the_genomic_accession=29073877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16847, CurrentID=0, Status='live', Symbol='Lepr', Aliases='LEPROTb, Modb1, OB-RGRP, Obr, db, diabetes, obese-like, obl, Lepr', description='leptin receptor', other_designations='B219|LEP-R|OB receptor|leptin receptor gene-related protein', map_location='4 46.96 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101717407, end_position_on_the_genomic_accession=101815354, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  16848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16848, CurrentID=0, Status='live', Symbol='Lfng', Aliases='AW061165', description='LFNG O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase', other_designations='O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase|lunatic fringe', map_location='5 79.15 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140607341, end_position_on_the_genomic_accession=140615545, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16851, CurrentID=0, Status='live', Symbol='lg', Aliases='', description='lid gap', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16852, CurrentID=0, Status='live', Symbol='Lgals1', Aliases='AA410090, Gal-1, Galbp, L-14.5, L14, Lect14, galectin-1', description='lectin, galactose binding, soluble 1', other_designations='14 kDa lectin|S-Lac lectin 1|beta-galactoside binding protein|beta-galactoside-binding lectin L-14-I|galaptin|lactose-binding lectin 1|lectin galactoside-binding soluble 1', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78926725, end_position_on_the_genomic_accession=78930465, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16854, CurrentID=0, Status='live', Symbol='Lgals3', Aliases='GBP, L-34, Mac-2, gal3', description='lectin, galactose binding, soluble 3', other_designations='35 kDa lectin|CBP 35|L-34 galactoside-binding lectin|carbohydrate-binding protein 35|gal-3|galactose-specific lectin 3|igE-binding protein|laminin-binding protein|lectin L-29|mac-2 antigen', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47373860, end_position_on_the_genomic_accession=47386167, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16855, CurrentID=0, Status='live', Symbol='Lgals4', Aliases='gal-4', description='lectin, galactose binding, soluble 4', other_designations='lactose-binding lectin 4', map_location='7 16.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28833609, end_position_on_the_genomic_accession=28841710, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  16857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16857, CurrentID=0, Status='live', Symbol='Lgals6', Aliases='galectin-6', description='lectin, galactose binding, soluble 6', other_designations='gal-6', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.4', start_position_on_the_genomic_accession=28542890, end_position_on_the_genomic_accession=28550393, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16858, CurrentID=0, Status='live', Symbol='Lgals7', Aliases='Galectin-7, Pig1', description='lectin, galactose binding, soluble 7', other_designations='gal-7', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28863853, end_position_on_the_genomic_accession=28866290, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16859, CurrentID=0, Status='live', Symbol='Lgals9', Aliases='AA407335, AI194909, AI265545, LGALS35, Lgals5, gal-9, galectin-9', description='lectin, galactose binding, soluble 9', other_designations='lectin, galactose binding, soluble 5', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78962979, end_position_on_the_genomic_accession=78984924, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  330490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330490, CurrentID=0, Status='live', Symbol='Nlrp9c', Aliases='AA636537, BB062542, Nalp-zeta, Nalp9c', description='NLR family, pyrin domain containing 9C', other_designations='NACHT, LRR and PYD containing protein 9c', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26364702, end_position_on_the_genomic_accession=26403682, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16862, CurrentID=0, Status='live', Symbol='lgl', Aliases='', description='legless', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  16865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16865, CurrentID=0, Status='live', Symbol='Eif2d', Aliases='D1Ertd5e, Lgtn', description='eukaryotic translation initiation factor 2D', other_designations='ligatin', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131153170, end_position_on_the_genomic_accession=131187659, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  16866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16866, CurrentID=0, Status='live', Symbol='Lhb', Aliases='LH, LH-B, LSH-B, LSH-beta', description='luteinizing hormone beta', other_designations='leutropin|luteinizing hormone subunit beta|lutropin beta chain', map_location='7 29.3 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45418166, end_position_on_the_genomic_accession=45421855, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16867, CurrentID=0, Status='live', Symbol='Lhcgr', Aliases='Gpcr19-rs1, LH-R, Lhr', description='luteinizing hormone/choriogonadotropin receptor', other_designations='LH/CG-R|LSH-R|luteinizing hormone receptor', map_location='17 58.35 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=88716488, end_position_on_the_genomic_accession=88792059, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16869, CurrentID=0, Status='live', Symbol='Lhx1', Aliases='Lim1', description='LIM homeobox protein 1', other_designations='LIM homeo box protein 1|homeobox protein Lim-1', map_location='11 51.31 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84519379, end_position_on_the_genomic_accession=84525534, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  16870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16870, CurrentID=0, Status='live', Symbol='Lhx2', Aliases='LH2A, Lh-2, Lim2, ap, apterous', description='LIM homeobox protein 2', other_designations='LIM homeo box protein 2|homeobox protein LH-2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38339281, end_position_on_the_genomic_accession=38369737, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  16871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16871, CurrentID=0, Status='live', Symbol='Lhx3', Aliases='Lim3, P-LIM, mLIM3, mLim-3', description='LIM homeobox protein 3', other_designations='LIM homeo box protein 3|homeobox protein LIM-3|homeobox protein P-LIM', map_location='2 A2-C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26200212, end_position_on_the_genomic_accession=26208249, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16872, CurrentID=0, Status='live', Symbol='Lhx4', Aliases='A330062J17Rik, Gsh-4, Gsh4', description='LIM homeobox protein 4', other_designations='LIM homeo box protein 4', map_location='1 67.47 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155698031, end_position_on_the_genomic_accession=155751726, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16873, CurrentID=0, Status='live', Symbol='Lhx5', Aliases='Lim2', description='LIM homeobox protein 5', other_designations='LIM homeo box protein 5', map_location='5 60.62 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120431886, end_position_on_the_genomic_accession=120441457, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16874, CurrentID=0, Status='live', Symbol='Lhx6', Aliases='Lhx6.1', description='LIM homeobox protein 6', other_designations='LIM-homeodomain protein 6.1|LIM/homeobox protein Lhx6.1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36081950, end_position_on_the_genomic_accession=36105997, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  16875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16875, CurrentID=0, Status='live', Symbol='Lhx8', Aliases='L3, Lhx7', description='LIM homeobox protein 8', other_designations='LIM homeo box protein 8|LIM homeobox protein 7|LIM/homeobox protein Lhx7', map_location='3 H3-H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=154306288, end_position_on_the_genomic_accession=154330560, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16876, CurrentID=0, Status='live', Symbol='Lhx9', Aliases='3110009O07Rik, BB104635, LH2B', description='LIM homeobox protein 9', other_designations='', map_location='1 G-H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=138825186, end_position_on_the_genomic_accession=138848649, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16878, CurrentID=0, Status='live', Symbol='Lif', Aliases='', description='leukemia inhibitory factor', other_designations='d factor|differentiation-stimulating factor|myeloid leukaemia inhibitory factor', map_location='11 A1-A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4257568, end_position_on_the_genomic_accession=4272514, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16880, CurrentID=0, Status='live', Symbol='Lifr', Aliases='A230075M04Rik, AW061234, LIF', description='leukemia inhibitory factor receptor', other_designations='D-factor/LIF receptor|LIF receptor|LIF-R|leukemia inhibitory factor receptor alpha chain|soluble differentiation-stimulating factor receptor', map_location='15 3.46 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=7090611, end_position_on_the_genomic_accession=7197489, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  16881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16881, CurrentID=0, Status='live', Symbol='Lig1', Aliases='AL033288, LigI', description='ligase I, DNA, ATP-dependent', other_designations='DNA ligase I|polydeoxyribonucleotide synthase [ATP] 1', map_location='7 7.82 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13277313, end_position_on_the_genomic_accession=13311427, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  16882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16882, CurrentID=0, Status='live', Symbol='Lig3', Aliases='D11Wsu78e', description='ligase III, DNA, ATP-dependent', other_designations='DNA ligase (ATP) 3|DNA ligase III|polydeoxyribonucleotide synthase [ATP] 3', map_location='11 50.23 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82781109, end_position_on_the_genomic_accession=82804274, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  279028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=279028, CurrentID=0, Status='live', Symbol='Adamts13', Aliases='ADAM-TS13, ADAMTS-13, Gm710, vWF-CP', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 13', other_designations='a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 13|vWF-CP mRNA for von Willebrand factor-cleaving|vWF-cleaving protease|von Willebrand factor-cleaving protease', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26973416, end_position_on_the_genomic_accession=27009628, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  16885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16885, CurrentID=0, Status='live', Symbol='Limk1', Aliases='KIZ-1, LIMK-1, Limk', description='LIM-domain containing, protein kinase', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134656039, end_position_on_the_genomic_accession=134688655, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  16886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16886, CurrentID=0, Status='live', Symbol='Limk2', Aliases='Limk-2, whe', description='LIM motif-containing protein kinase 2', other_designations='LIM kinase 2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3343297, end_position_on_the_genomic_accession=3409238, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  16889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16889, CurrentID=0, Status='live', Symbol='Lipa', Aliases='AA960673, Lal, Lip-1, Lip1', description='lysosomal acid lipase A', other_designations='acid cholesteryl ester hydrolase|cholesteryl esterase|lipase A|lysosomal acid lipase 1|sterol esterase', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34492316, end_position_on_the_genomic_accession=34527474, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16890, CurrentID=0, Status='live', Symbol='Lipe', Aliases='4933403G17Rik, HSL', description='lipase, hormone sensitive', other_designations='', map_location='7 13.78 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25379527, end_position_on_the_genomic_accession=25398704, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  16891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16891, CurrentID=0, Status='live', Symbol='Lipg', Aliases='3110013K01Rik, EL, lipase, mEDL', description='lipase, endothelial', other_designations='EDL|endothelial cell-derived lipase|lipose, endothelial', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74939322, end_position_on_the_genomic_accession=74961263, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  16897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16897, CurrentID=0, Status='live', Symbol='Llgl1', Aliases='AI325176, Lgl1, Llglh, Mgl1', description='lethal giant larvae homolog 1 (Drosophila)', other_designations='LLGL|mgl-1|mlgl', map_location='11 37.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60699670, end_position_on_the_genomic_accession=60714188, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  16898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16898, CurrentID=0, Status='live', Symbol='Rps2', Aliases='Llrep3, S4', description='ribosomal protein S2', other_designations='40S ribosomal protein S4|protein LLRep3|repeat family 3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24718121, end_position_on_the_genomic_accession=24721927, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16904, CurrentID=0, Status='live', Symbol='Gzmm', Aliases='Lmet1, MMET-1', description='granzyme M (lymphocyte met-ase 1)', other_designations='lymphoctye Met-ase 1', map_location='10 C', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79689020, end_position_on_the_genomic_accession=79695261, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16905, CurrentID=0, Status='live', Symbol='Lmna', Aliases='Dhe', description='lamin A', other_designations='lamin C|lamin-A/C', map_location='3 38.84 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88481148, end_position_on_the_genomic_accession=88509932, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  16906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16906, CurrentID=0, Status='live', Symbol='Lmnb1', Aliases='', description='lamin B1', other_designations='', map_location='18 30.84 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=56707813, end_position_on_the_genomic_accession=56753424, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  16907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16907, CurrentID=0, Status='live', Symbol='Lmnb2', Aliases='', description='lamin B2', other_designations='lamin B3', map_location='10 C', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80901363, end_position_on_the_genomic_accession=80918245, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16909, CurrentID=0, Status='live', Symbol='Lmo2', Aliases='Rbtn-2, Rbtn2, Rhom-2, Ttg2', description='LIM domain only 2', other_designations='LIM domain only protein 2|LIM only 2|LMO-2|T-cell translocation protein 2|cysteine-rich protein TTG-2', map_location='2 54.46 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103957995, end_position_on_the_genomic_accession=103981878, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  16911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16911, CurrentID=0, Status='live', Symbol='Lmo4', Aliases='A730077C12Rik, Crp3, Etohi4', description='LIM domain only 4', other_designations='LIM domain only protein 4|LMO-4|breast tumor autoantigen|ethanol induced 4', map_location='3 68.61 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144188530, end_position_on_the_genomic_accession=144205255, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  16912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16912, CurrentID=0, Status='live', Symbol='Psmb9', Aliases='Lmp-2, Lmp2', description='proteasome (prosome, macropain) subunit, beta type 9 (large multifunctional peptidase 2)', other_designations='LMP-2d|large multifunctional protease 2|low molecular mass protein 2|macropain chain 7|multicatalytic endopeptidase complex chain 7|proteasome chain 7|proteasome subunit beta type 9|proteasome subunit beta-1i|proteosome (prosome, macropain) subunit, beta type 9 (large multifunctional peptidase 2)|proteosome beta 9 subunit|really interesting new gene 12 protein', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34182030, end_position_on_the_genomic_accession=34187330, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16913, CurrentID=0, Status='live', Symbol='Psmb8', Aliases='Lmp-7, Lmp7', description='proteasome (prosome, macropain) subunit, beta type 8 (large multifunctional peptidase 7)', other_designations='MC13 (LMP7)|large multifunctional protease 7|low molecular mass protein 7|macropain subunit C13|multicatalytic endopeptidase complex subunit C13|proteasome component C13|proteasome subunit MC13|proteasome subunit beta type 8|proteasome subunit beta-5i|proteosome (prosome, macropain) subunit, beta type 8 (large multifunctional peptidase 7)|proteosome beta 8 subunit', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34198195, end_position_on_the_genomic_accession=34201454, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  16917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16917, CurrentID=0, Status='live', Symbol='Lmx1b', Aliases='LMX1.1, LMX1.2', description='LIM homeobox transcription factor 1 beta', other_designations='LIM homeo box transcription factor 1B|LIM/homeobox protein 1.2|LIM/homeobox protein LMX1B|LMX-1.2', map_location='2 22.48 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33560965, end_position_on_the_genomic_accession=33640992, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  16918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16918, CurrentID=0, Status='live', Symbol='Mycl', Aliases='AW536278, L-myc, Lmyc-1, Lmyc11, bHLHe38, Mycl', description='v-myc myelocytomatosis viral oncogene homolog, lung carcinoma derived (avian)', other_designations='L-myc-1 proto-oncogene protein|lung carcinoma myc related oncogene 1|protein L-Myc-1|v-myc myelocytomatosis viral oncogene homolog 1, lung carcinoma derived', map_location='4 57.36 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=122995613, end_position_on_the_genomic_accession=123002490, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16922, CurrentID=0, Status='live', Symbol='Phyh', Aliases='AI256161, AI265699, Lnap1, PAHX', description='phytanoyl-CoA hydroxylase', other_designations='lupus nephritis-associated peptide 1|phytanic acid oxidase|phytanoyl-CoA alpha-hydroxylase', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=4918996, end_position_on_the_genomic_accession=4938743, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  16923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16923, CurrentID=0, Status='live', Symbol='Sh2b3', Aliases='AI429800, Lnk', description='SH2B adaptor protein 3', other_designations='linker of T-cell receptor pathways|lymphocyte-specific adapter protein Lnk|signal transduction protein Lnk', map_location='5 61.99 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121815481, end_position_on_the_genomic_accession=121837655, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  672284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=672284, CurrentID=0, Status='live', Symbol='Nkx1-1', Aliases='Nkx-1.1, Sax2', description='NK1 transcription factor related, locus 1 (Drosophila)', other_designations='spinal cord axial homeobox gene 2', map_location='5 17.65 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33430734, end_position_on_the_genomic_accession=33434090, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16939, CurrentID=0, Status='live', Symbol='Lor', Aliases='AI036317, S77319', description='loricrin', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92080271, end_position_on_the_genomic_accession=92083142, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  16948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16948, CurrentID=0, Status='live', Symbol='Lox', Aliases='AI893619, TSC-160, rrg', description='lysyl oxidase', other_designations='ras excision protein|ras recision', map_location='18 28.22 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=52516060, end_position_on_the_genomic_accession=52529867, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16949, CurrentID=0, Status='live', Symbol='Loxl1', Aliases='Loxl', description='lysyl oxidase-like 1', other_designations='lysyl oxidase 2|lysyl oxidase-like protein 1', map_location='9 31.65 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58287723, end_position_on_the_genomic_accession=58313212, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  16950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16950, CurrentID=0, Status='live', Symbol='Loxl3', Aliases='Lor2, Loxl2', description='lysyl oxidase-like 3', other_designations='lysyl oxidase-like 2|lysyl oxidase-like protein 3|lysyl oxidase-related protein 2', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83034224, end_position_on_the_genomic_accession=83052564, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  16952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16952, CurrentID=0, Status='live', Symbol='Anxa1', Aliases='Anx-1, Anx-A1, C430014K04Rik, Lpc-1, Lpc1', description='annexin A1', other_designations='annexin I|annexin-1|calpactin II|calpactin-2|chromobindin-9|lipocortin 1|lipocortin I|p35|phospholipase A2 inhibitory protein', map_location='19 13.83 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=20373434, end_position_on_the_genomic_accession=20390671, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  16956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16956, CurrentID=0, Status='live', Symbol='Lpl', Aliases='', description='lipoprotein lipase', other_designations='O 1-4-5', map_location='8 33.88 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=68880555, end_position_on_the_genomic_accession=68906932, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  16963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16963, CurrentID=0, Status='live', Symbol='Xcl1', Aliases='AI661682, ATAC, LTN, Lptn, SCM-1, SCM-1a, Scyc1', description='chemokine (C motif) ligand 1', other_designations='c motif chemokine 1|cytokine SCM-1|lymphotaxin|small inducible cytokine subfamily C, member 1 (lymphotactin)|small-inducible cytokine C1', map_location='1 72.26 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164931648, end_position_on_the_genomic_accession=164935510, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16969, CurrentID=0, Status='live', Symbol='Zbtb7a', Aliases='9030619K07Rik, 9130006G12Rik, AI452336, FBI-1, Lrf, Pokemon, Zbtb7', description='zinc finger and BTB domain containing 7a', other_designations='POK erythroid myeloid ontogenic factor|POZ and Krueppel erythroid myeloid ontogenic factor|leukemia/lymphoma related factor|leukemia/lymphoma-related factor', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81134585, end_position_on_the_genomic_accession=81152222, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  16970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16970, CurrentID=0, Status='live', Symbol='Lrmp', Aliases='D6Int3, D6Int4, D6Int5, D6Int7, D6Int8, Jaw1', description='lymphoid-restricted membrane protein', other_designations='', map_location='6 77.34 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145111760, end_position_on_the_genomic_accession=145174928, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  16971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16971, CurrentID=0, Status='live', Symbol='Lrp1', Aliases='A2mr, AI316852, CD91, Lrp, b2b1554Clo', description='low density lipoprotein receptor-related protein 1', other_designations='LRP-1|alpha 2-macroglobulin receptor|alpha-2-macroglobulin receptor|lipoprotein receptor-related protein|lrp1', map_location='10 B2-D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127538157, end_position_on_the_genomic_accession=127621148, orientation='minus', exon_count=89, OMIM='', no_hdr0=''),
  16973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16973, CurrentID=0, Status='live', Symbol='Lrp5', Aliases='BMND1, HBM, LR3, LRP7, OPPG, mKIAA4142', description='low density lipoprotein receptor-related protein 5', other_designations='LRP-5|LRP-7|low-density lipoprotein receptor-related protein 7', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3584825, end_position_on_the_genomic_accession=3686564, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  16974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16974, CurrentID=0, Status='live', Symbol='Lrp6', Aliases='C030016K15Rik, Cd, Gw, ska26, ska<m26Jus>, skax26', description='low density lipoprotein receptor-related protein 6', other_designations='LRP-6|crooked|low density lipoprotein receptor-related protein 6; low density lipoprotein-related protein 6|low density lipoprotein-related protein 6', map_location='6 65.37 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134446476, end_position_on_the_genomic_accession=134566960, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  16975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16975, CurrentID=0, Status='live', Symbol='Lrp8', Aliases='4932703M08Rik, AA921429, AI848122, ApoER2, Lr8b', description='low density lipoprotein receptor-related protein 8, apolipoprotein e receptor', other_designations='apolipoprotein E receptor 2|low density lipoprotein receptor relative, 8 ligand binding repeats, brain', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107802259, end_position_on_the_genomic_accession=107876840, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  16976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16976, CurrentID=0, Status='live', Symbol='Lrpap1', Aliases='AA617339, AI790446, AU042172, C77774, HBP44, RAP', description='low density lipoprotein receptor-related protein associated protein 1', other_designations='Alpha-2-macroglobulin receptor-associated protein precursor (Alpha-2-MRAP) (Low density lipoprotein receptor-related protein-associated protein 1) (RAP) (Heparin binding protein-44) (HBP-44)|HBP-44|alpha-2-MRAP|heparin-binding protein 44|low density lipoprotein receptor related protein, associated protein 1|low density lipoprotein receptor-related protein-associated protein 1', map_location='5 18.01 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35091506, end_position_on_the_genomic_accession=35105697, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16977, CurrentID=0, Status='live', Symbol='Lrrc23', Aliases='4921537K05Rik, B7, Lrpb7', description='leucine rich repeat containing 23', other_designations='leucine-rich protein B7', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124769869, end_position_on_the_genomic_accession=124780341, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  16978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16978, CurrentID=0, Status='live', Symbol='Lrrfip1', Aliases='AU024550, Fliiap1', description='leucine rich repeat (in FLII) interacting protein 1', other_designations='FLAP (FLI LRR associated protein)|FLI-LRR-associated protein 1|H186 FLAP|LRR FLII-interacting protein 1|flap-1', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90998458, end_position_on_the_genomic_accession=91128944, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  16979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16979, CurrentID=0, Status='live', Symbol='Lrrn1', Aliases='2810047E21Rik, NLRR-1', description='leucine rich repeat protein 1, neuronal', other_designations='', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=107529726, end_position_on_the_genomic_accession=107570228, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16980, CurrentID=0, Status='live', Symbol='Lrrn2', Aliases='5730406J09Rik, NLRR-2', description='leucine rich repeat protein 2, neuronal', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132880355, end_position_on_the_genomic_accession=132940005, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16981, CurrentID=0, Status='live', Symbol='Lrrn3', Aliases='NLRR-3', description='leucine rich repeat protein 3, neuronal', other_designations='', map_location='12 B3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=41451668, end_position_on_the_genomic_accession=41486057, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  16985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16985, CurrentID=0, Status='live', Symbol='Lsp1', Aliases='Lsp-1, WP34, p50, pp52', description='lymphocyte specific 1', other_designations='52 kDa phosphoprotein|S37 protein|leukocyte specific protein 1|leukocyte-specific protein 1|lymphocyte-specific antigen WP34', map_location='7 87.93 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142460812, end_position_on_the_genomic_accession=142494868, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  16987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16987, CurrentID=0, Status='live', Symbol='Lss', Aliases='2810025N20Rik, D10Ertd116e, Osc', description='lanosterol synthase', other_designations='2,3-epoxysqualene--lanosterol cyclase|oxidosqualene--lanosterol cyclase', map_location='10 39.1 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76531565, end_position_on_the_genomic_accession=76557139, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  16988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16988, CurrentID=0, Status='live', Symbol='Lst1', Aliases='B144', description='leukocyte specific transcript 1', other_designations='', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35185095, end_position_on_the_genomic_accession=35188440, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16991, CurrentID=0, Status='live', Symbol='Lt1', Aliases='', description='lurcher transcript 1', other_designations='regulation of nuclear pre-mRNA domain containing 1A pseudogene', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=64896953, end_position_on_the_genomic_accession=64899441, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  16992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16992, CurrentID=0, Status='live', Symbol='Lta', Aliases='LT, LT-[a], LT-alpha, LT[a], LTalpha, Ltx, TNF-beta, TNFSF1, Tnfb, Tnfsf1b, Tnlg1e, hlb382', description='lymphotoxin A', other_designations='TNF beta|lymphotoxin alpha|tumor necrosis factor ligand 1e|tumor necrosis factor ligand superfamily member 1', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35203165, end_position_on_the_genomic_accession=35205625, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16993, CurrentID=0, Status='live', Symbol='Lta4h', Aliases='', description='leukotriene A4 hydrolase', other_designations='LTA-4 hydrolase|LTA4 hydrodase', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93453396, end_position_on_the_genomic_accession=93484898, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  16994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16994, CurrentID=0, Status='live', Symbol='Ltb', Aliases='AI662801, LTbeta, Tnfc, Tnfsf3, Tnlg1c, p33', description='lymphotoxin B', other_designations='LT-beta|TNF-C|lymphotoxin beta|tumor necrosis factor C|tumor necrosis factor ligand 1c|tumor necrosis factor ligand superfamily member 3', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35194507, end_position_on_the_genomic_accession=35196305, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  16995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16995, CurrentID=0, Status='live', Symbol='Ltb4r1', Aliases='BLT1, BLTR, Ltb4r, mBLTR', description='leukotriene B4 receptor 1', other_designations='LTB4 receptor|LTB4-R 1|LTB4-R1', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55765050, end_position_on_the_genomic_accession=55768494, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  16997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16997, CurrentID=0, Status='live', Symbol='Ltbp2', Aliases='AW208642', description='latent transforming growth factor beta binding protein 2', other_designations='LTBP-2', map_location='12 D', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84783212, end_position_on_the_genomic_accession=84877266, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  16998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16998, CurrentID=0, Status='live', Symbol='Ltbp3', Aliases='Ltbp2, mFLJ00070', description='latent transforming growth factor beta binding protein 3', other_designations='LTBP-3', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5740904, end_position_on_the_genomic_accession=5758532, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  17000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17000, CurrentID=0, Status='live', Symbol='Ltbr', Aliases='AI256028, CD18, LTbetaR, Ltar, TNF-R-III, TNFCR, TNFR-RP, TNFR2-RP, TNFRrp, Tnfbr, Tnfrsf3', description='lymphotoxin B receptor', other_designations='LT beta-R|LT-beta receptor|TNF receptor-related protein|lymphotoxin-beta receptor|tumor necrosis factor receptor superfamily, member 3', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125306571, end_position_on_the_genomic_accession=125313870, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17001, CurrentID=0, Status='live', Symbol='Ltc4s', Aliases='', description='leukotriene C4 synthase', other_designations='LTC4 synthase|leukotriene-C(4) synthase', map_location='11 B1.1-B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50235003, end_position_on_the_genomic_accession=50239175, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17002, CurrentID=0, Status='live', Symbol='Ltf', Aliases='Csp82, Lf, MMS10R, Ms10r', description='lactotransferrin', other_designations='N lobe|lactoferrin|minisatellite 10r detected by probe MMS10', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111019292, end_position_on_the_genomic_accession=111042766, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  17005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17005, CurrentID=0, Status='live', Symbol='Ltk', Aliases='', description='leukocyte tyrosine kinase', other_designations='leukocyte receptor tyrosine kinase', map_location='2 59.97 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119751326, end_position_on_the_genomic_accession=119760431, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  541307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=541307, CurrentID=0, Status='live', Symbol='Ccl26', Aliases='Ccl26l', description='chemokine (C-C motif) ligand 26', other_designations='eotaxin-3', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135560448, end_position_on_the_genomic_accession=135563569, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  17022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17022, CurrentID=0, Status='live', Symbol='Lum', Aliases='Ldc, SLRR2D', description='lumican', other_designations='KSPG lumican|keratan sulfate proteoglycan lumican', map_location='10 50.32 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=97565501, end_position_on_the_genomic_accession=97572703, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17025, CurrentID=0, Status='live', Symbol='Alad', Aliases='ALADH, Lv', description='aminolevulinate, delta-, dehydratase', other_designations='porphobilinogen synthase', map_location='4 33.17 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62505984, end_position_on_the_genomic_accession=62520063, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  17034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17034, CurrentID=0, Status='live', Symbol='Lx', Aliases='', description='luxate', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17035, CurrentID=0, Status='live', Symbol='Lxn', Aliases='', description='latexin', other_designations='ECI|TCI|endogenous carboxypeptidase inhibitor|tissue carboxypeptidase inhibitor', map_location='3 30.98 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=67457999, end_position_on_the_genomic_accession=67463907, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17055, CurrentID=0, Status='live', Symbol='Klra', Aliases='Ly-49, Ly49', description='killer cell lectin-like receptor, subfamily A', other_designations='lymphocyte antigen 49 complex', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17057, CurrentID=0, Status='live', Symbol='Klrb1a', Aliases='Ly55a, NKR-P1.7, NKR-P1A, NKRP12, Nkrp1-a, Nkrp1a, ly-55A', description='killer cell lectin-like receptor subfamily B member 1A', other_designations='CD161 antigen-like family member A|NKR-P1 2|lymphocyte antigen 55 complex, locus A|lymphocyte antigen 55A|natural killer cell receptor protein NKR-P1A|natural killer cell receptor-P1|natural killer cell surface protein P1-2', map_location='6 63.06 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128609227, end_position_on_the_genomic_accession=128622985, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17059, CurrentID=0, Status='live', Symbol='Klrb1c', Aliases='AI462337, CD161, Ly-59, Ly55c, Ly59, NK-RP1, NK1.1, NKR-P1.9, NKRP1, NKRP140, Nk-1, Nk-1.2, Nk1, Nk1.2, Nkrp1c, ly-55c', description='killer cell lectin-like receptor subfamily B member 1C', other_designations='CD161 antigen-like family member C|NKR-P1 40|killer cell lectin-like receptor subfamily A member 1C|lymphocyte antigen 55 complex, locus C|lymphocyte antigen 55c|lymphocyte antigen 59|natural killer cell receptor protein NKR-P1C|natural killer cell receptor-P1|natural killer cell surface protein P1-40|natural killer cell-associated antigen 1', map_location='6 63.09 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128778485, end_position_on_the_genomic_accession=128788641, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17060, CurrentID=0, Status='live', Symbol='Blnk', Aliases='BASH, Bca, Ly-57, Ly57, Lyw-57, SLP-65', description='B cell linker', other_designations='B-cell adapter containing a SH2 domain protein|B-cell adapter containing a Src homology 2 domain protein|cytoplasmic adapter protein|lymphocyte antigen 57|src homology 2 domain-containing leukocyte protein of 65 kDa', map_location='19 34.26 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40928927, end_position_on_the_genomic_accession=40994535, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17062, CurrentID=0, Status='live', Symbol='Ly6', Aliases='Ala-1, DAG, H9/25, Ly-27, Ly-6, Ly27', description='lymphocyte antigen 6 complex', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17063, CurrentID=0, Status='live', Symbol='Muc13', Aliases='114/A10, 14/A10, AI159736, Lrrp, Ly64, NJ-1', description='mucin 13, epithelial transmembrane', other_designations='MUC-13|cell surface antigen 114/A10|lymphocyte antigen 64', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33794037, end_position_on_the_genomic_accession=33819929, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  17064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17064, CurrentID=0, Status='live', Symbol='Cd93', Aliases='6030404G09Rik, AA145088, AA4.1, AW555904, C1qr1, C1qrp, Ly68', description='CD93 antigen', other_designations='C1q/MBL/SPA receptor|C1qR(p)|cell surface antigen AA4|complement component 1 q subcomponent receptor 1|complement component 1, q subcomponent, receptor 1|ly-68|lymphocyte antigen 68', map_location='2 73.48 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148436651, end_position_on_the_genomic_accession=148443535, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17067, CurrentID=0, Status='live', Symbol='Ly6c1', Aliases='AA682074, AA959465, Ly-6C, Ly-6C1, Ly6c', description='lymphocyte antigen 6 complex, locus C1', other_designations='Lymphocyte antigen Ly-6C|lymphocyte antigen 6 complex, locus C', map_location='15 34.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75044018, end_position_on_the_genomic_accession=75048853, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17068, CurrentID=0, Status='live', Symbol='Ly6d', Aliases='Ly-61, Ly61, Thb', description='lymphocyte antigen 6 complex, locus D', other_designations='ly-6D|thymocyte B-cell antigen', map_location='15 34.27 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74762056, end_position_on_the_genomic_accession=74763567, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  17069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17069, CurrentID=0, Status='live', Symbol='Ly6e', Aliases='Ly67, RIG-E, Sca-2, TSA-1, Tsa1', description='lymphocyte antigen 6 complex, locus E', other_designations='ly-6E|stem cell antigen 2|thymic shared antigen 1', map_location='15 34.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74955051, end_position_on_the_genomic_accession=74959905, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17071, CurrentID=0, Status='live', Symbol='Ly6f', Aliases='', description='lymphocyte antigen 6 complex, locus F', other_designations='ly-6F|ly-6F.1', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75268372, end_position_on_the_genomic_accession=75272236, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17075, CurrentID=0, Status='live', Symbol='Epcam', Aliases='CD326, EGP, EGP-2, Egp314, Ep-CAM, EpCAM1, GA733-2, Ly74, TROP1, Tacsd1, Tacstd1, gp40', description='epithelial cell adhesion molecule', other_designations='Trop-1 protein|epithelial glycoprotein 314|lymphocyte antigen 74|mEGP314|panepithelial glycoprotein 314|protein 289A|tumor-associated calcium signal transducer 1', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87635979, end_position_on_the_genomic_accession=87651127, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17076, CurrentID=0, Status='live', Symbol='Ly75', Aliases='CD205, DEC-205, DEC205', description='lymphocyte antigen 75', other_designations='ly-75', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=60293760, end_position_on_the_genomic_accession=60383231, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  17079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17079, CurrentID=0, Status='live', Symbol='Cd180', Aliases='F630107B15, Ly78, RP105', description='CD180 antigen', other_designations='ly-78|lymphocyte antigen 78|radioprotective 105 kDa protein', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=102693558, end_position_on_the_genomic_accession=102706631, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17082, CurrentID=0, Status='live', Symbol='Il1rl1', Aliases='DER4, Fit-1, Ly84, ST2L, St2, St2-rs1, T1, T1/ST2', description='interleukin 1 receptor-like 1', other_designations='lymphocyte antigen 84', map_location='1 19.19 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40404864, end_position_on_the_genomic_accession=40465418, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  17083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17083, CurrentID=0, Status='live', Symbol='Tmed1', Aliases='Il1rl1l, Ly84l, St2l', description='transmembrane emp24 domain containing 1', other_designations='interleukin 1 receptor-like 1 ligand|interleukin-1 receptor-like 1 ligand|lymphocyte antigen 84 ligand|p24 family protein gamma-1|p24gamma1|putative T1/ST2 receptor-binding protein', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21507380, end_position_on_the_genomic_accession=21510187, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17084, CurrentID=0, Status='live', Symbol='Ly86', Aliases='MD-1, MD1, ly86_tv2', description='lymphocyte antigen 86', other_designations='ly-86|protein MD-1', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=37345345, end_position_on_the_genomic_accession=37419036, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17085, CurrentID=0, Status='live', Symbol='Ly9', Aliases='AI893573, CD229, Lgp100, Ly-9, SLAMF3, T100', description='lymphocyte antigen 9', other_designations='SLAM family member 3|cell surface molecule Ly-9|signaling lymphocytic activation molecule 3', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171588613, end_position_on_the_genomic_accession=171607410, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17086, CurrentID=0, Status='live', Symbol='Ncr1', Aliases='Ly94, NKp46', description='natural cytotoxicity triggering receptor 1', other_designations='MAR1 (mouse activating receptor 1)|NK-p46|activating receptor 1|lymphocyte antigen 94|mAR-1|mNKp46|natural killer cell p46-related protein', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4337678, end_position_on_the_genomic_accession=4345164, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17087, CurrentID=0, Status='live', Symbol='Ly96', Aliases='ESOP-1, MD-2, MD2', description='lymphocyte antigen 96', other_designations='ly-96|myeloid differentiation factor-2|myeloid differentiation protein-2|protein MD-2', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16687851, end_position_on_the_genomic_accession=16709608, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  17089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17089, CurrentID=0, Status='live', Symbol='Lyar', Aliases='MLZ-264', description='Ly1 antibody reactive clone', other_designations='protein expressed in male leptotene and zygotene spermatocytes 264', map_location='5 20.29 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38220482, end_position_on_the_genomic_accession=38234306, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  17095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17095, CurrentID=0, Status='live', Symbol='Lyl1', Aliases='Lyl-1, bHLHa18', description='lymphoblastomic leukemia 1', other_designations='lymphoblastic leukemia-derived sequence 1', map_location='8 41.02 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84701422, end_position_on_the_genomic_accession=84704716, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17096, CurrentID=0, Status='live', Symbol='Lyn', Aliases='AA407514, Hck-2, p53Lyn, p56Lyn', description='LYN proto-oncogene, Src family tyrosine kinase', other_designations='V-yes-1 Yamaguchi sarcoma viral related oncogene homolog|Yamaguchi sarcoma viral (v-yes-1) oncogene homolog', map_location='4 2.05 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3678121, end_position_on_the_genomic_accession=3791612, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  17101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17101, CurrentID=0, Status='live', Symbol='Lyst', Aliases='D13Sfk13, beige, bg', description='lysosomal trafficking regulator', other_designations='CHS1 homolog', map_location='13 5.28 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=13590339, end_position_on_the_genomic_accession=13778803, orientation='plus', exon_count=55, OMIM='', no_hdr0=''),
  17105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17105, CurrentID=0, Status='live', Symbol='Lyz2', Aliases='AI326280, Lys, Lysm, Lyzf2, Lyzs, Lzm, Lzm-s1, Lzp', description='lysozyme 2', other_designations='1,4-beta-N-acetylmuramidase C|lysozyme C type M|lysozyme f2', map_location='10 65.34 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117277541, end_position_on_the_genomic_accession=117282272, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17110, CurrentID=0, Status='live', Symbol='Lyz1', Aliases='Lyz, Lyzf3, Lzp-s', description='lysozyme 1', other_designations='1,4-beta-N-acetylmuramidase C|P lysozyme structural|lysozyme C type P|lysozyme f3|renal amyloidosis', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117287795, end_position_on_the_genomic_accession=117292868, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17112, CurrentID=0, Status='live', Symbol='Tm4sf1', Aliases='L6, M3s1', description='transmembrane 4 superfamily member 1', other_designations='12A8 target antigen|L6 antigen|membrane component chromosome 3 surface marker 1 homolog|membrane component, surface marker 1|tumor-associated antigen L6', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=57287064, end_position_on_the_genomic_accession=57301983, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17113, CurrentID=0, Status='live', Symbol='M6pr', Aliases='CD-MPR, Mpr46', description='mannose-6-phosphate receptor, cation dependent', other_designations='46 kDa mannose 6-phosphate receptor|CD Man-6-P receptor|MPR 46', map_location='6 57.52 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122309010, end_position_on_the_genomic_accession=122317677, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17116, CurrentID=0, Status='live', Symbol='Mab21l1', Aliases='AW047968', description='mab-21-like 1 (C. elegans)', other_designations='', map_location='3 E3-F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=55782510, end_position_on_the_genomic_accession=55785287, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17117, CurrentID=0, Status='live', Symbol='Amacr', Aliases='Macr1', description='alpha-methylacyl-CoA racemase', other_designations='2-arylpropionyl-CoA epimerase|2-methylacyl-CoA racemase|alpha-methylacyl-Coenzyme A racemase', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10981756, end_position_on_the_genomic_accession=10996624, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17118, CurrentID=0, Status='live', Symbol='Marcks', Aliases='Macs, PKCSL', description='myristoylated alanine rich protein kinase C substrate', other_designations='80K protein', map_location='10 19.49 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=37133243, end_position_on_the_genomic_accession=37138926, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17119, CurrentID=0, Status='live', Symbol='Mxd1', Aliases='AW122478, Mad, Mad1', description='MAX dimerization protein 1', other_designations='max dimerizer 1', map_location='6 37.75 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86647044, end_position_on_the_genomic_accession=86671442, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17120, CurrentID=0, Status='live', Symbol='Mad1l1', Aliases='AI173502, AW550425, Mad1', description='MAD1 mitotic arrest deficient 1-like 1', other_designations='MAD1-like protein 1|mitotic arrest deficient 1-like 1|mitotic arrest deficient 1-like protein 1', map_location='5 78.82 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140008689, end_position_on_the_genomic_accession=140321591, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  17121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17121, CurrentID=0, Status='live', Symbol='Mxd3', Aliases='4631412E13Rik, Mad3, bHLHc13', description='Max dimerization protein 3', other_designations='max dimerizer 3|max-associated protein 3', map_location='13 29.8 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55325168, end_position_on_the_genomic_accession=55329770, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17122, CurrentID=0, Status='live', Symbol='Mxd4', Aliases='2810410A03Rik, Mad4, bHLHc12', description='Max dimerization protein 4', other_designations='max dimerizer 4|max-associated protein 4|max-interacting transcriptional repressor MAD4', map_location='5 17.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34173887, end_position_on_the_genomic_accession=34187750, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17123, CurrentID=0, Status='live', Symbol='Madcam1', Aliases='AV211525', description='mucosal vascular addressin cell adhesion molecule 1', other_designations='MAdCAM-1|mMAdCAM-1', map_location='10 C1-C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79664395, end_position_on_the_genomic_accession=79668541, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17125, CurrentID=0, Status='live', Symbol='Smad1', Aliases='AI528653, Madh1, Madr1, Mlp1, MusMLP', description='SMAD family member 1', other_designations='MAD homolog 1|Smad 1|dwarfin-A|dwf-A|mMad1|mad-related protein 1|mothers against DPP homolog 1|mothers-against-DPP-related 1', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79338395, end_position_on_the_genomic_accession=79399468, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  17126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17126, CurrentID=0, Status='live', Symbol='Smad2', Aliases='7120426M23Rik, Madh2, Madr2, Smad-2, mMad2', description='SMAD family member 2', other_designations='MAD homolog 2|Smad 2|mad-related protein 2|mothers against DPP homolog 2', map_location='18 51.42 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=76241251, end_position_on_the_genomic_accession=76311748, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  17127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17127, CurrentID=0, Status='live', Symbol='Smad3', Aliases='AU022421, Madh3', description='SMAD family member 3', other_designations='MAD homolog 3|Smad 3|mMad3|mad3|mothers against DPP homolog 3', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=63646766, end_position_on_the_genomic_accession=63757994, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  17128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17128, CurrentID=0, Status='live', Symbol='Smad4', Aliases='AW743858, D18Wsu70e, DPC4, Madh4', description='SMAD family member 4', other_designations='MAD homolog 4|Smad 4|deletion target in pancreatic carcinoma 4 homolog|mothers against DPP homolog 4', map_location='18 49.51 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=73638998, end_position_on_the_genomic_accession=73703741, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  17129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17129, CurrentID=0, Status='live', Symbol='Smad5', Aliases='1110051M15Rik, AI451355, Dwf-C, Madh5, MusMLP', description='SMAD family member 5', other_designations='MAD homolog 5|Smad 5|dwarfin-C|mSmad5|mothers against DPP homolog 5', map_location='13 30.12 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56702988, end_position_on_the_genomic_accession=56742379, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17130, CurrentID=0, Status='live', Symbol='Smad6', Aliases='Madh6, b2b390Clo', description='SMAD family member 6', other_designations='MAD homolog 6|Smad 6|mad homolog 7|mothers against DPP homolog 6', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=63953076, end_position_on_the_genomic_accession=64022085, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17131, CurrentID=0, Status='live', Symbol='Smad7', Aliases='Madh7', description='SMAD family member 7', other_designations='MAD homolog 7|MAD homolog 8|SMAD 7|madh8|mothers against DPP homolog 7|mothers against DPP homolog 8|mothers against decapentaplegic homolog 8', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=75367365, end_position_on_the_genomic_accession=75395935, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17132, CurrentID=0, Status='live', Symbol='Maf', Aliases='2810401A20Rik, A230108G15Rik, AW047063, c-maf', description='avian musculoaponeurotic fibrosarcoma oncogene homolog', other_designations='V-maf musculoaponeurotic fibrosarcoma oncogene homolog|avian musculoaponeurotic fibrosarcoma (v-maf) AS42 oncogene homolog|c-Maf long form|proto-oncogene c-maf', map_location='8 62.61 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=115703253, end_position_on_the_genomic_accession=115706894, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17133, CurrentID=0, Status='live', Symbol='Maff', Aliases='', description='v-maf musculoaponeurotic fibrosarcoma oncogene family, protein F (avian)', other_designations='V-maf musculoaponeurotic fibrosarcoma oncogene homolog F', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79346626, end_position_on_the_genomic_accession=79359078, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17134, CurrentID=0, Status='live', Symbol='Mafg', Aliases='AA545192, C630022N07Rik', description='v-maf musculoaponeurotic fibrosarcoma oncogene family, protein G (avian)', other_designations='V-maf musculoaponeurotic fibrosarcoma oncogene homolog G|avian musculoaponeurotic fibrosarcoma virus (v-maf) AS42 oncogene, protein G', map_location='11 84.35 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120628348, end_position_on_the_genomic_accession=120633547, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  17135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17135, CurrentID=0, Status='live', Symbol='Mafk', Aliases='AW061068, NF-E2, Nfe2u', description='v-maf musculoaponeurotic fibrosarcoma oncogene family, protein K (avian)', other_designations='erythroid transcription factor NF-E2 p18 subunit|nuclear factor, erythroid derived 2, ubiquitous|p18 NF-E2', map_location='5 78.81 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139791536, end_position_on_the_genomic_accession=139802652, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17136, CurrentID=0, Status='live', Symbol='Mag', Aliases='Gma, siglec-4a', description='myelin-associated glycoprotein', other_designations='', map_location='7 19.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30899175, end_position_on_the_genomic_accession=30914922, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  17137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17137, CurrentID=0, Status='live', Symbol='Magea1', Aliases='Mage-a1', description='melanoma antigen, family A, 1', other_designations='', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155088519, end_position_on_the_genomic_accession=155089790, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17138, CurrentID=0, Status='live', Symbol='Magea2', Aliases='Mage-a2', description='melanoma antigen, family A, 2', other_designations='', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155027201, end_position_on_the_genomic_accession=155033313, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17139, CurrentID=0, Status='live', Symbol='Magea3', Aliases='Mage-a3', description='melanoma antigen, family A, 3', other_designations='', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154948180, end_position_on_the_genomic_accession=154959031, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17140, CurrentID=0, Status='live', Symbol='Magea4', Aliases='Mage-a4', description='melanoma antigen, family A, 4', other_designations='', map_location='X 37.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72221015, end_position_on_the_genomic_accession=72222964, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17141, CurrentID=0, Status='live', Symbol='Magea5', Aliases='Mage-a5', description='melanoma antigen, family A, 5', other_designations='', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155053061, end_position_on_the_genomic_accession=155063145, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17142, CurrentID=0, Status='live', Symbol='Magea6', Aliases='Mage-a6', description='melanoma antigen, family A, 6', other_designations='', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154924012, end_position_on_the_genomic_accession=154926115, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17144, CurrentID=0, Status='live', Symbol='Magea8', Aliases='Mage-a8', description='melanoma antigen, family A, 8', other_designations='', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154985775, end_position_on_the_genomic_accession=154995850, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17145, CurrentID=0, Status='live', Symbol='Mageb1', Aliases='MAGE-Xp, Mage-b1, Mage-rs1, Magel1, Smage1, dam1', description='melanoma antigen, family B, 1', other_designations='', map_location='X 40.59 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=92015191, end_position_on_the_genomic_accession=92016333, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17146, CurrentID=0, Status='live', Symbol='Mageb2', Aliases='Mage-b2, Mage-rs2, Smage2', description='melanoma antigen, family B, 2', other_designations='', map_location='X 40.35 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=91331905, end_position_on_the_genomic_accession=91332897, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17147, CurrentID=0, Status='live', Symbol='Mageb3', Aliases='Mage-b3, Mage-ps1, Mage-rs3, Smage3', description='melanoma antigen, family B, 3', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121953771, end_position_on_the_genomic_accession=121956092, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17149, CurrentID=0, Status='live', Symbol='Magoh', Aliases='Mago-m, Mos2', description='mago homolog, exon junction complex core component', other_designations='mago-nashi homolog, proliferation-associated', map_location='4 50.18 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107879755, end_position_on_the_genomic_accession=107887424, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17150, CurrentID=0, Status='live', Symbol='Mfap2', Aliases='AI893631, Magp, Magp1', description='microfibrillar-associated protein 2', other_designations='MAGP-1|MFAP-2|microfibril-associated glycoprotein 1', map_location='4 D3-E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141010424, end_position_on_the_genomic_accession=141015984, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  672511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=672511, CurrentID=0, Status='live', Symbol='Rnf213', Aliases='6030403J01, D11Ertd759e, mKIAA1554', description='ring finger protein 213', other_designations='', map_location='11 83.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119389454, end_position_on_the_genomic_accession=119487423, orientation='plus', exon_count=69, OMIM='', no_hdr0=''),
  17152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17152, CurrentID=0, Status='live', Symbol='Mak', Aliases='A930010O05Rik', description='male germ cell-associated kinase', other_designations='protein kinase RCK', map_location='13 20.25 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41025007, end_position_on_the_genomic_accession=41079706, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17153, CurrentID=0, Status='live', Symbol='Mal', Aliases='Mpv17, Vip17', description='myelin and lymphocyte protein, T cell differentiation protein', other_designations='T-lymphocyte maturation-associated protein|myelin and lymphocyte protein, T-cell differentiation protein', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127633226, end_position_on_the_genomic_accession=127656695, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17155, CurrentID=0, Status='live', Symbol='Man1a', Aliases='Man1a1, PCR1', description='mannosidase 1, alpha', other_designations='alpha-1,2-mannosidase IA|man(9)-alpha-mannosidase|man9-mannosidase|mannosidase alpha class 1A member 1|mannosyl-oligosaccharide alpha-1,2-mannosidase|processing alpha-1,2-mannosidase IA', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=53904785, end_position_on_the_genomic_accession=54076586, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  17156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17156, CurrentID=0, Status='live', Symbol='Man1a2', Aliases='AI428775, AI528764, AI854422, Man1b, PCR2', description='mannosidase, alpha, class 1A, member 2', other_designations='alpha-1,2-mannosidase IB|mannosidase 1, beta|processing alpha-1,2-mannosidase IB', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100562203, end_position_on_the_genomic_accession=100685473, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  17158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17158, CurrentID=0, Status='live', Symbol='Man2a1', Aliases='Mana-2, Mana2, Map-2', description='mannosidase 2, alpha 1', other_designations='AMan II|golgi alpha-mannosidase II|man II|mannosidase alpha class 2A member 1|mannosyl-oligosaccharide 1,3-1,6-alpha-mannosidase', map_location='17 E1.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=64601649, end_position_on_the_genomic_accession=64755110, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  17159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17159, CurrentID=0, Status='live', Symbol='Man2b1', Aliases='AW107687, LAMAN', description='mannosidase 2, alpha B1', other_designations='lysosomal acid alpha-mannosidase|mannosidase alpha class 2B member 1|mannosidase alpha-B', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85082729, end_position_on_the_genomic_accession=85098739, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  17160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17160, CurrentID=0, Status='live', Symbol='Man2b2', Aliases='mKIAA0935', description='mannosidase 2, alpha B2', other_designations='135 kD alpha-D-mannosidase|135 kDa alpha-D-mannosidase|mannosidase alpha class 2B member 2', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36806813, end_position_on_the_genomic_accession=36830649, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  17161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17161, CurrentID=0, Status='live', Symbol='Maoa', Aliases='1110061B18Rik, AA407771', description='monoamine oxidase A', other_designations='MAO-A|monoamine oxidase type A', map_location='X 11.78 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=16619698, end_position_on_the_genomic_accession=16687812, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17164, CurrentID=0, Status='live', Symbol='Mapkapk2', Aliases='AA960234, MAPKAP-K2, MK-2, MK2, Rps6kc1', description='MAP kinase-activated protein kinase 2', other_designations='MAPK-activated protein kinase 2|MAPKAP kinase 2', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131053700, end_position_on_the_genomic_accession=131097543, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  17165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17165, CurrentID=0, Status='live', Symbol='Mapkapk5', Aliases='MK5, PRAK', description='MAP kinase-activated protein kinase 5', other_designations='MAPK-activated protein kinase 5|MAPKAP kinase 5|MAPKAPK-5', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121525051, end_position_on_the_genomic_accession=121545892, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  17167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17167, CurrentID=0, Status='live', Symbol='Marco', Aliases='AI323439, Ly112, Scara2', description='macrophage receptor with collagenous structure', other_designations='', map_location='1 E4-F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120474538, end_position_on_the_genomic_accession=120505084, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17168, CurrentID=0, Status='live', Symbol='Nprl3', Aliases='Aag, CGTHBA, HS-26, HS-40, Mare, Phg, Prox1, m(alpha)RE', description='nitrogen permease regulator-like 3', other_designations='-14|alpha globin regulatory element containing|alpha-globin regulatory element-containing gene protein|proximal locus to the hemoglobin alpha chain complex', map_location='11 18.83 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32231963, end_position_on_the_genomic_accession=32267707, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17169, CurrentID=0, Status='live', Symbol='Mark3', Aliases='1600015G02Rik, A430080F22Rik, C-TAK1, ETK-1, ETK1, Emk2, MPK10', description='MAP/microtubule affinity regulating kinase 3', other_designations='ELKL motif kinase 2|EMK-2|EMK2/MARK3 serine-threonine protein kinase|MPK-10', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111574510, end_position_on_the_genomic_accession=111656227, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  17171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17171, CurrentID=0, Status='live', Symbol='Mas1', Aliases='Mas-1, Mgra', description='MAS1 oncogene', other_designations='MAS proto-oncogene|Mas-related G protein-coupled receptor a|angiotensin-(1-7) receptor', map_location='17 8.69 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12838841, end_position_on_the_genomic_accession=12868145, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  17172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17172, CurrentID=0, Status='live', Symbol='Ascl1', Aliases='AI225900, ASH1, Mash1, bHLHa46', description='achaete-scute family bHLH transcription factor 1', other_designations='ASH-1|achaete-scute complex homolog 1|achaete-scute complex homolog-like 1|achaete-scute complex-like protein 1|mASH-1|mammalian achaete scute homolog 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=87491041, end_position_on_the_genomic_accession=87493660, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17173, CurrentID=0, Status='live', Symbol='Ascl2', Aliases='2410083I15Rik, Mash2, bHLHa45', description='achaete-scute family bHLH transcription factor 2', other_designations='ASH-2|achaete-scute complex homolog 2|achaete-scute complex homolog-like 2|long transient receptor potential-related channel 5|mASH-2|mammalian achaete scute homolog 2', map_location='7 88.08 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142966819, end_position_on_the_genomic_accession=143009052, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17174, CurrentID=0, Status='live', Symbol='Masp1', Aliases='AW048060, CCPII, Crarf/3, Masp1', description='mannan-binding lectin serine peptidase 1', other_designations='MASP-1|complement factor MASP-3|complement-activating component of Ra-reactive factor|mannose-binding lectin-associated serine protease 1|mannose-binding protein-associated serine protease|ra-reactive factor serine protease p100|raRF|serine protease 5', map_location='16 B2-B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23451785, end_position_on_the_genomic_accession=23521073, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  541463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=541463, CurrentID=0, Status='live', Symbol='Tex24', Aliases='1700108N07Rik, TESF-1', description='testis expressed gene 24', other_designations='testis-specific factor 1|testis-specific protein', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27344323, end_position_on_the_genomic_accession=27349188, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17178, CurrentID=0, Status='live', Symbol='Fxyd3', Aliases='Mat8', description='FXYD domain-containing ion transport regulator 3', other_designations='chloride conductance inducer protein Mat-8|mammary tumor 8 kDa protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31070535, end_position_on_the_genomic_accession=31076697, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  17179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17179, CurrentID=0, Status='live', Symbol='Matk', Aliases='CHK, HYL, Ntk, p56ntk', description='megakaryocyte-associated tyrosine kinase', other_designations='Csk homologous kinase|protein kinase NTK|tyrosine protein kinase|tyrosine-protein kinase CTK', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81252795, end_position_on_the_genomic_accession=81262982, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  17180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17180, CurrentID=0, Status='live', Symbol='Matn1', Aliases='CMP, Crtm, Mat1, matrilin-1', description='matrilin 1, cartilage matrix protein', other_designations='matrilin 1, cartilage matrix protein 1', map_location='4 64.09 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130944385, end_position_on_the_genomic_accession=130955476, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17181, CurrentID=0, Status='live', Symbol='Matn2', Aliases='Crtm2, matrilin-2', description='matrilin 2', other_designations='', map_location='15 B3.3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34306658, end_position_on_the_genomic_accession=34436242, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  17182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17182, CurrentID=0, Status='live', Symbol='Matn3', Aliases='AV009181', description='matrilin 3', other_designations='', map_location='12 3.96 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8947929, end_position_on_the_genomic_accession=8972028, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17183, CurrentID=0, Status='live', Symbol='Matn4', Aliases='matrilin-4', description='matrilin 4', other_designations='MAT-4', map_location='2 85.16 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164389393, end_position_on_the_genomic_accession=164405160, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  17184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17184, CurrentID=0, Status='live', Symbol='Matr3', Aliases='1110061A14Rik, 2810017I02Rik, AI841759, AW555618, D030046F20Rik, mKIAA0723', description='matrin 3', other_designations='', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35562158, end_position_on_the_genomic_accession=35592045, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17187, CurrentID=0, Status='live', Symbol='Max', Aliases='AA960152, AI875693, bHLHd4, bHLHd5, bHLHd6, bHLHd7, bHLHd8', description='Max protein', other_designations='myc-associated factor X|myc-binding novel HLH/LZ protein|protein myn', map_location='12 D1-D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76937269, end_position_on_the_genomic_accession=76962335, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17188, CurrentID=0, Status='live', Symbol='Maz', Aliases='PUR1, Pur-1, SAF-1, SAF-2', description='MYC-associated zinc finger protein (purine-binding transcription factor)', other_designations='MAZI|purine-binding transcription factor', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127022123, end_position_on_the_genomic_accession=127026553, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17189, CurrentID=0, Status='live', Symbol='Mb', Aliases='AI325109', description='myoglobin', other_designations='', map_location='15 36.36 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77015487, end_position_on_the_genomic_accession=77050668, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17190, CurrentID=0, Status='live', Symbol='Mbd1', Aliases='Cxxc3, PCM1', description='methyl-CpG binding domain protein 1', other_designations='methyl-CpG-binding protein MBD1', map_location='18 50.7 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74267615, end_position_on_the_genomic_accession=74282734, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  17191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17191, CurrentID=0, Status='live', Symbol='Mbd2', Aliases='MBD2a', description='methyl-CpG binding domain protein 2', other_designations='methyl-CpG-binding protein MBD2|testis-specific methyl-CpG binding protein 2', map_location='18 44.54 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=70568270, end_position_on_the_genomic_accession=70626131, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17192, CurrentID=0, Status='live', Symbol='Mbd3', Aliases='AI181826, AU019209', description='methyl-CpG binding domain protein 3', other_designations='methyl-CpG-binding protein MBD3', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80392539, end_position_on_the_genomic_accession=80400702, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17193, CurrentID=0, Status='live', Symbol='Mbd4', Aliases='Med1', description='methyl-CpG binding domain protein 4', other_designations='methyl-CpG-binding protein MBD4|mismatch-specific DNA N-glycosylase', map_location='6 53.72 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115840697, end_position_on_the_genomic_accession=115853438, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  17194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17194, CurrentID=0, Status='live', Symbol='Mbl1', Aliases='MBL-A, MBP-A, S-MBP', description='mannose-binding lectin (protein A) 1', other_designations='mannan-binding protein|mannose binding lectin (A)|mannose binding lectin, serum (A)|ra-reactive factor polysaccharide-binding component p28B|raRF p28B', map_location='14 22.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=41151456, end_position_on_the_genomic_accession=41159596, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  17195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17195, CurrentID=0, Status='live', Symbol='Mbl2', Aliases='L-MBP, MBL, MBL-C, MBP-C', description='mannose-binding lectin (protein C) 2', other_designations='RA-reactive factor P28A subunit|RARF/P28A|mannan-binding protein|mannose binding lectin (C)|mannose binding lectin, liver (C)', map_location='19 25.14 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=30232906, end_position_on_the_genomic_accession=30239686, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  17196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17196, CurrentID=0, Status='live', Symbol='Mbp', Aliases='C76307, Hmbpr, R75289, golli-mbp, mld, shi', description='myelin basic protein', other_designations='myelin A1 protein|myelin deficient|shiverer', map_location='18 E2-E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=82475123, end_position_on_the_genomic_accession=82585637, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17199, CurrentID=0, Status='live', Symbol='Mc1r', Aliases='Mshra, Tob, e', description='melanocortin 1 receptor', other_designations='MC1-R|MSH-R|extension recessive yellow|melanocortin receptor 1|melanocortin-1 receptor|tobacco darkening', map_location='8 72.1 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123407082, end_position_on_the_genomic_accession=123410744, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17200, CurrentID=0, Status='live', Symbol='Mc2r', Aliases='ACTH-R, ACTHR, MC2-R', description='melanocortin 2 receptor', other_designations='ACTH receptor|adrenocorticotropin receptor|melanocortin receptor 2', map_location='18 41.89 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=68406899, end_position_on_the_genomic_accession=68429320, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17201, CurrentID=0, Status='live', Symbol='Mc3r', Aliases='MC3-R', description='melanocortin 3 receptor', other_designations='', map_location='2 94.8 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172248492, end_position_on_the_genomic_accession=172251114, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17202, CurrentID=0, Status='live', Symbol='Mc4r', Aliases='Mc4-r', description='melanocortin 4 receptor', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=66857705, end_position_on_the_genomic_accession=66860487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17203, CurrentID=0, Status='live', Symbol='Mc5r', Aliases='', description='melanocortin 5 receptor', other_designations='MC5-R', map_location='18 41.46 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=68337260, end_position_on_the_genomic_accession=68342446, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17207, CurrentID=0, Status='live', Symbol='Mcf2l', Aliases='C130040G20Rik, Dbs, Ost, mKIAA0362', description='mcf.2 transforming sequence-like', other_designations="DBL's big sister|MCF2-transforming sequence-like protein|Ost gamma|exchange factor for RhoA and Cdc42|guanine nucleotide exchange factor OSTIII", map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=12873811, end_position_on_the_genomic_accession=13020509, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  17210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17210, CurrentID=0, Status='live', Symbol='Mcl1', Aliases='AW556805, Mcl-1', description='myeloid cell leukemia sequence 1', other_designations='bcl-2-related protein EAT/mcl1', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95658368, end_position_on_the_genomic_accession=95663179, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17215, CurrentID=0, Status='live', Symbol='Mcm3', Aliases='AL033361, C80350, Mcmd, P1, p1.m', description='minichromosome maintenance complex component 3', other_designations='DNA polymerase alpha holoenzyme-associated protein P1|P1-MCM3|mini chromosome maintenance deficient|minichromosome maintenance deficient 3', map_location='1 A3-A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20803014, end_position_on_the_genomic_accession=20820213, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17216, CurrentID=0, Status='live', Symbol='Mcm2', Aliases='AA959861, AW476101, BM28, CDCL1, Mcmd2, mKIAA0030', description='minichromosome maintenance complex component 2', other_designations='mini chromosome maintenance deficient 2|minichromosome maintenance deficient 2 mitotin|minichromosome maintenance protein 2 homolog|nuclear protein BM28', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88883474, end_position_on_the_genomic_accession=88898780, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17217, CurrentID=0, Status='live', Symbol='Mcm4', Aliases='19G, AI325074, AU045576, Cdc21, Mcmd4, mKIAA4003, mcdc21', description='minichromosome maintenance complex component 4', other_designations='CDC21 homolog|P1-CDC21|mini chromosome maintenance deficient 4 homolog|minichromosome maintenance deficient 4 homolog', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=15623897, end_position_on_the_genomic_accession=15637400, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17218, CurrentID=0, Status='live', Symbol='Mcm5', Aliases='AA617332, AI324988, AL033333, Cdc46, Mcmd5, P1-CDC46', description='minichromosome maintenance complex component 5', other_designations='CDC46 homolog|mini chromosome maintenance deficient 5|minichromosome maintenance deficient 5, cell division cycle 46', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=75109509, end_position_on_the_genomic_accession=75128439, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  17219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17219, CurrentID=0, Status='live', Symbol='Mcm6', Aliases='ASP-l1, D1Wsu22e, Mcmd6', description='minichromosome maintenance complex component 6', other_designations='mini chromosome maintenance deficient 6|minichromosome maintenance deficient 6 (MIS5 homolog, S. pombe)', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=128331591, end_position_on_the_genomic_accession=128359656, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17220, CurrentID=0, Status='live', Symbol='Mcm7', Aliases='AI747533, Mcmd7, mCDC47', description='minichromosome maintenance complex component 7', other_designations='CDC47 homolog|mini chromosome maintenance deficient 7|minichromosome maintenance deficient 7', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138164589, end_position_on_the_genomic_accession=138171862, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17221, CurrentID=0, Status='live', Symbol='Cd46', Aliases='Mcp', description='CD46 antigen, complement regulatory protein', other_designations='', map_location='1 98.41 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=195038968, end_position_on_the_genomic_accession=195098784, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17222, CurrentID=0, Status='live', Symbol='Anapc1', Aliases='2610021O03Rik, AI047775, AI853536, AW547281, Apc1, Mcpr, tsg24', description='anaphase promoting complex subunit 1', other_designations='cyclosome subunit 1|meiotic check point regulator|mitotic checkpoint regulator|testis-specific gene 24 protein', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128610083, end_position_on_the_genomic_accession=128687410, orientation='minus', exon_count=50, OMIM='', no_hdr0=''),
  17224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17224, CurrentID=0, Status='live', Symbol='Mcpt1', Aliases='AV080368, Mcp-1', description='mast cell protease 1', other_designations='MMCP-1', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56017964, end_position_on_the_genomic_accession=56020391, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17225, CurrentID=0, Status='live', Symbol='Mcpt2', Aliases='MMCP-2, Mcp-2', description='mast cell protease 2', other_designations='', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56021584, end_position_on_the_genomic_accession=56044635, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  17227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17227, CurrentID=0, Status='live', Symbol='Mcpt4', Aliases='MMCP-4, MMCP-4A, MMCP-4B, Mcp-4, Mcp4', description='mast cell protease 4', other_designations='MSMCP|myonase|serosal mast cell protease', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56059744, end_position_on_the_genomic_accession=56062310, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17228, CurrentID=0, Status='live', Symbol='Cma1', Aliases='MMCP-5, Mcp-5, Mcp5, Mcpt5', description='chymase 1, mast cell', other_designations='alpha-chymase|mast cell chymase 1|mast cell protease 5|mast cell protease I', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55941451, end_position_on_the_genomic_accession=55944661, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17229, CurrentID=0, Status='live', Symbol='Tpsb2', Aliases='AV011504, MMCP-6, Mcp-6, Mcp6, Mcpt6', description='tryptase beta 2', other_designations='mast cell protease 6|tryptase-2', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25366303, end_position_on_the_genomic_accession=25368096, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  17231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17231, CurrentID=0, Status='live', Symbol='Mcpt8', Aliases='BB139338, MMCP-8', description='mast cell protease 8', other_designations='', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56082163, end_position_on_the_genomic_accession=56085276, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17232, CurrentID=0, Status='live', Symbol='Mcpt9', Aliases='MMCP-1, MMCP-9, MMCP9', description='mast cell protease 9', other_designations='mast cell-specific serine protease', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56026758, end_position_on_the_genomic_accession=56030495, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17235, CurrentID=0, Status='live', Symbol='Smcp', Aliases='Mcs, Mcsp', description='sperm mitochondria-associated cysteine-rich protein', other_designations='mitochondrial capsule selenoprotein', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92583866, end_position_on_the_genomic_accession=92589024, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17237, CurrentID=0, Status='live', Symbol='Mgrn1', Aliases='2610042J20Rik, mKIAA0544, md, nc', description='mahogunin, ring finger 1', other_designations='mahoganoid|mahogunin RING finger protein 1|probable E3 ubiquitin-protein ligase MGRN1', map_location='16 2.48 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4886100, end_position_on_the_genomic_accession=4938296, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  17238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17238, CurrentID=0, Status='live', Symbol='Mdac', Aliases='', description='modifier of Dac', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17240, CurrentID=0, Status='live', Symbol='Mdfi', Aliases='I-mf, I-mfa', description='MyoD family inhibitor', other_designations='myogenic repressor I-mf', map_location='17 23.99 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47815328, end_position_on_the_genomic_accession=47834691, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17242, CurrentID=0, Status='live', Symbol='Mdk', Aliases='MK, Mek', description='midkine', other_designations='retinoic acid-induced differentiation factor|retinoic acid-responsive protein', map_location='2 50.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91929805, end_position_on_the_genomic_accession=91932297, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17245, CurrentID=0, Status='live', Symbol='Mdm1', Aliases='Arrd2, Mdm-1', description='transformed mouse 3T3 cell double minute 1', other_designations='double minute 1|mdm4 transformed 3T3 cell double minute 1 protein', map_location='10 C1-C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=118141537, end_position_on_the_genomic_accession=118168999, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  17246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17246, CurrentID=0, Status='live', Symbol='Mdm2', Aliases='1700007J15Rik, AA415488, Mdm-2', description='transformed mouse 3T3 cell double minute 2', other_designations='double minute 2 protein|oncoprotein Mdm2|p53-binding protein Mdm2', map_location='10 C1-C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117688875, end_position_on_the_genomic_accession=117710770, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100418400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100418400, CurrentID=0, Status='live', Symbol='Pafah1b1-ps2', Aliases='Lis4, Mdsh4, Pafaha-ps2', description='platelet-activating factor acetylhydrolase, isoform 1b, beta1 subunit, pseudogene 2', other_designations='platelet-activating factor acetylhydrolase, isoform 1b, subunit 1 pseudogene', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7061859, end_position_on_the_genomic_accession=7063452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  17250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17250, CurrentID=0, Status='live', Symbol='Abcc1', Aliases='Abcc1ab, MRP, Mdrap, Mrp1, Abcc1', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 1', other_designations='ATP-binding cassette, sub-family C (CFTR/MRP), member 1a|ATP-binding cassette, sub-family C (CFTR/MRP), member 1b|LTC4 transporter|leukotriene C(4) transporter|multiple drug resistance-associated protein', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14361521, end_position_on_the_genomic_accession=14474879, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  17252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17252, CurrentID=0, Status='live', Symbol='Rdh11', Aliases='2610319N22Rik, AI428145, AU045252, Arsdr1, C85936, CGI-82, HCBP12, M42C60, Mdt1, Psdr1, SCALD, UBE-1c1, Ube-1c', description='retinol dehydrogenase 11', other_designations='all-trans and 9-cis|androgen-regulated short-chain dehydrogenase/reductase 1|cell line MC/9.IL4 derived transcript 1|cell line MC/9.IL4-derived protein 1|prostate short-chain dehydrogenase/reductase 1|ralR1|retinal reductase 1|short-chain aldehyde dehydrogenase|short-chain dehydrogenase/reductase 1|ubiquitously expressing', map_location='12 35.51 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79174335, end_position_on_the_genomic_accession=79191819, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  17254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17254, CurrentID=0, Status='live', Symbol='Slc3a2', Aliases='4F2, 4F2HC, AI314110, Cd98, Ly-10, Ly-m10, Ly10, Mdu1, Mgp-2hc, NACAE', description='solute carrier family 3 (activators of dibasic and neutral amino acid transport), member 2', other_designations='CD98 antigen|CD98 heavy chain|antigen identified by monoclonal antibodies 4F2|solute carrier family 3 member 2|type II transmembrane protein', map_location='19 5.44 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8706882, end_position_on_the_genomic_accession=8723369, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  17255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17255, CurrentID=0, Status='live', Symbol='mea', Aliases='', description='meander tail', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17256, CurrentID=0, Status='live', Symbol='Mea1', Aliases='Mea-1', description='male enhanced antigen 1', other_designations='male-enhanced antigen-1', map_location='17 22.9 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46679138, end_position_on_the_genomic_accession=46683126, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17257, CurrentID=0, Status='live', Symbol='Mecp2', Aliases='1500041B07Rik, D630021H01Rik, Mbd5, WBP10', description='methyl CpG binding protein 2', other_designations='meCp-2 protein', map_location='X 37.63 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74026592, end_position_on_the_genomic_accession=74085690, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17258, CurrentID=0, Status='live', Symbol='Mef2a', Aliases='A430079H05Rik', description='myocyte enhancer factor 2A', other_designations='', map_location='7 36.72 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=67231163, end_position_on_the_genomic_accession=67373501, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  17259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17259, CurrentID=0, Status='live', Symbol='Mef2b', Aliases='AI451606', description='myocyte enhancer factor 2B', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70152778, end_position_on_the_genomic_accession=70167488, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17260, CurrentID=0, Status='live', Symbol='Mef2c', Aliases='5430401D19Rik, 9930028G15Rik, AV011172, Mef2', description='myocyte enhancer factor 2C', other_designations='', map_location='13 43.68 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=83504034, end_position_on_the_genomic_accession=83667079, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17261, CurrentID=0, Status='live', Symbol='Mef2d', Aliases='C80750', description='myocyte enhancer factor 2D', other_designations='', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88141966, end_position_on_the_genomic_accession=88172084, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17263, CurrentID=0, Status='live', Symbol='Meg3', Aliases='2900016C05Rik, 3110050O07Rik, 6330408G06Rik, AI425946, AW108224, D12Bwg1266e, Gtl2, R74756, R75394', description='maternally expressed 3', other_designations='GTL2, imprinted maternally expressed untranslated mRNA|gene trap locus 2|maternally expressed gene 3', map_location='12 60.25 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109540996, end_position_on_the_genomic_accession=109571729, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  17268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17268, CurrentID=0, Status='live', Symbol='Meis1', Aliases='C530044H18Rik, Evi8', description='Meis homeobox 1', other_designations='myeloid ecotropic viral integration site 1', map_location='11 11.11 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=18880428, end_position_on_the_genomic_accession=19020482, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  17274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17274, CurrentID=0, Status='live', Symbol='Rab8a', Aliases='AA409338, Mel', description='RAB8A, member RAS oncogene family', other_designations='cell line NK14 derived transforming oncogene|oncogene c-mel', map_location='8 34.84 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72161200, end_position_on_the_genomic_accession=72181366, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17276, CurrentID=0, Status='live', Symbol='Mela', Aliases='80kDa, Ag, env, gag, gag-pol, pol', description='melanoma antigen', other_designations='Gag-Pol polyprotein|endogenous virus encoded protein|envelope glycoprotein|envelope protein|melanoma antigen, 80 kDa|reverse transcriptase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.4', start_position_on_the_genomic_accession=126311968, end_position_on_the_genomic_accession=126314734, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17279, CurrentID=0, Status='live', Symbol='Melk', Aliases='AI327312, MPK38, mKIAA0175', description='maternal embryonic leucine zipper kinase', other_designations='protein kinase PK38|tyrosine-protein kinase MELK', map_location='4 23.46 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44300876, end_position_on_the_genomic_accession=44364675, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  17281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17281, CurrentID=0, Status='live', Symbol='Fyco1', Aliases='2810409M01Rik, Mem2, RUFY3, ZFYVE7', description='FYVE and coiled-coil domain containing 1', other_designations='maternal embryonic message 2', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123789500, end_position_on_the_genomic_accession=123851899, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  17283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17283, CurrentID=0, Status='live', Symbol='Men1', Aliases='AW045611', description='multiple endocrine neoplasia 1', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6334979, end_position_on_the_genomic_accession=6340894, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17285, CurrentID=0, Status='live', Symbol='Meox1', Aliases='AI385561, D330041M02Rik, Mox-1, Mox1', description='mesenchyme homeobox 1', other_designations='', map_location='11 65.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101877510, end_position_on_the_genomic_accession=101894354, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  17286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17286, CurrentID=0, Status='live', Symbol='Meox2', Aliases='AI528662, Gax, Mox-2, Mox2', description='mesenchyme homeobox 2', other_designations='', map_location='12 16.84 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=37108546, end_position_on_the_genomic_accession=37179528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17287, CurrentID=0, Status='live', Symbol='Mep1a', Aliases='AI098089, AW107200, Mep-1, Mep-1a, Mep1', description='meprin 1 alpha', other_designations='endopeptidase-2|meprin A alpha-subunit|meprin alpha', map_location='17 C1-D1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43474324, end_position_on_the_genomic_accession=43504144, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  17288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17288, CurrentID=0, Status='live', Symbol='Mep1b', Aliases='Mep-1b', description='meprin 1 beta', other_designations='endopeptidase-2|meprin B|meprin beta', map_location='18 11.81 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=21072344, end_position_on_the_genomic_accession=21100199, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17289, CurrentID=0, Status='live', Symbol='Mertk', Aliases='Eyk, Mer, Nyk, nmf12', description='c-mer proto-oncogene tyrosine kinase', other_designations='Tyro 12|proto-oncogene c-Mer|proto-oncogene tyrosine-protein kinase Mer|receptor tyrosine kinase MerTK', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128698997, end_position_on_the_genomic_accession=128803162, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  17292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17292, CurrentID=0, Status='live', Symbol='Mesp1', Aliases='bHLHc5', description='mesoderm posterior 1', other_designations='', map_location='7 45.16 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79792241, end_position_on_the_genomic_accession=79793590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17293, CurrentID=0, Status='live', Symbol='Mesp2', Aliases='bHLHc6', description='mesoderm posterior 2', other_designations='', map_location='7 45.18 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79810727, end_position_on_the_genomic_accession=79813431, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  17294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17294, CurrentID=0, Status='live', Symbol='Mest', Aliases='AA408879, AI256745, Peg1', description='mesoderm specific transcript', other_designations='paternally-expressed gene 1 protein', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30733506, end_position_on_the_genomic_accession=30748466, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  17295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17295, CurrentID=0, Status='live', Symbol='Met', Aliases='AI838057, HGF, HGFR, Par4, c-Met', description='met proto-oncogene', other_designations='HGF receptor|HGF/SF receptor|SF receptor|met proto-oncogene tyrosine kinase|proto-oncogene c-Met|scatter factor receptor|tyrosine-protein kinase Met', map_location='6 7.83 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=17463349, end_position_on_the_genomic_accession=17573980, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  17299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17299, CurrentID=0, Status='live', Symbol='Mettl1', Aliases='2810012D02Rik', description='methyltransferase like 1', other_designations='methyltransferase-like 1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127041643, end_position_on_the_genomic_accession=127046365, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17300, CurrentID=0, Status='live', Symbol='Foxc1', Aliases='FREAC3, Fkh1, Mf1, Mf4, ch, fkh-1, frkhda', description='forkhead box C1', other_designations='FREAC-3|MF-1|congenital hydrocephalus|forkhead-related protein FKHL7|forkhead-related transcription factor 3|mesoderm/mesenchyme forkhead 1|transcription factor FKH-1', map_location='13 13.52 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=31806646, end_position_on_the_genomic_accession=31810635, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17301, CurrentID=0, Status='live', Symbol='Foxd2', Aliases='AI426778, Mf2', description='forkhead box D2', other_designations='MF-2|mesoderm/mesenchyme forkhead 2', map_location='4 52.73 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=114906280, end_position_on_the_genomic_accession=114908898, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17304, CurrentID=0, Status='live', Symbol='Mfge8', Aliases='AA408458, AI325141, MFG-E8, MP47, Mfgm, P47, SED1, lactadherin', description='milk fat globule-EGF factor 8 protein', other_designations='EGF/factor VIII|milk fat globule glycoprotein|sperm surface protein SP47', map_location='7 44.9 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79133768, end_position_on_the_genomic_accession=79149060, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17305, CurrentID=0, Status='live', Symbol='Mfng', Aliases='AW546563', description='MFNG O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase', other_designations='O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase|manic fringe homolog', map_location='15 37.69 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78755882, end_position_on_the_genomic_accession=78773516, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17306, CurrentID=0, Status='live', Symbol='Sypl2', Aliases='AI552439, Mg29', description='synaptophysin-like 2', other_designations='mitsugumin 29|mitsugumin-29', map_location='3 F3-H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108210410, end_position_on_the_genomic_accession=108226669, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17308, CurrentID=0, Status='live', Symbol='Mgat1', Aliases='Mgat-1', description='mannoside acetylglucosaminyltransferase 1', other_designations='GNT-I|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase I|glcNAc-T I', map_location='11 29.27 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49244191, end_position_on_the_genomic_accession=49263030, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17309, CurrentID=0, Status='live', Symbol='Mgat3', Aliases='1110038J12Rik, AI848272, GnT-III', description='mannoside acetylglucosaminyltransferase 3', other_designations='N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase III|glcNAc-T III', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80173650, end_position_on_the_genomic_accession=80215519, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17311, CurrentID=0, Status='live', Symbol='Kitl', Aliases='Clo, Con, Gbg, Mgf, SCF, SF, SLF, Sl, blz, contrasted, Kitl', description='kit ligand', other_designations='C-kit ligand|Steel factor|cloud gray|grizzle-belly|hematopoietic growth factor KL|mast cell growth factor|stem cell factor', map_location='10 51.4 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=100015809, end_position_on_the_genomic_accession=100100416, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17312, CurrentID=0, Status='live', Symbol='Clec10a', Aliases='CD301a, M-ASGP-BP-1, Mgl, Mgl1', description='C-type lectin domain family 10, member A', other_designations='macrophage asialoglycoprotein-binding protein 1|macrophage galactose N-acetyl-galactosamine specific lectin 1', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70166623, end_position_on_the_genomic_accession=70170836, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17313, CurrentID=0, Status='live', Symbol='Mgp', Aliases='Mglap', description='matrix Gla protein', other_designations='matrix gamma-carboxyglutamate (gla) protein', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136872435, end_position_on_the_genomic_accession=136875805, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17314, CurrentID=0, Status='live', Symbol='Mgmt', Aliases='AGT, AI267024, Agat', description='O-6-methylguanine-DNA methyltransferase', other_designations='6-O-methylguanine-DNA methyltransferase|O-6-alkylguanine-DNA alkyltransferase|O-6-methylguanine-DNA-alkyltransferase', map_location='7 82.07 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=136894468, end_position_on_the_genomic_accession=137128193, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17318, CurrentID=0, Status='live', Symbol='Mid1', Aliases='61B3-R, DXHXS1141, Fxy, Trim18', description='midline 1', other_designations='RING finger protein Midline-1|midin|tripartite motif-containing protein 18', map_location='X 79.19 cM', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=90718657, end_position_on_the_genomic_accession=90765072, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  17319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17319, CurrentID=0, Status='live', Symbol='Mif', Aliases='GIF, Glif', description='macrophage migration inhibitory factor', other_designations='DER6|L-dopachrome isomerase|L-dopachrome tautomerase|delayed early response protein 6|glycosylation-inhibiting factor|phenylpyruvate tautomerase', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75859353, end_position_on_the_genomic_accession=75860250, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  541610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=541610, CurrentID=0, Status='live', Symbol='Trcg1', Aliases='', description='taste receptor cell gene 1', other_designations='taste receptor protein 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57236460, end_position_on_the_genomic_accession=57252175, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  17329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17329, CurrentID=0, Status='live', Symbol='Cxcl9', Aliases='BB139920, CMK, Mig, MuMIG, Scyb9, crg-10', description='chemokine (C-X-C motif) ligand 9', other_designations='M119|gamma interferon-induced monokine|gamma-interferon-induced monokine|monokine induced by gamma interferon|monokine induced by interferon-gamma|protein m119|small inducible cytokine B subfamily (Cys-X-Cys), member 9|small-inducible cytokine B9', map_location='5 46.51 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92321331, end_position_on_the_genomic_accession=92328079, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17330, CurrentID=0, Status='live', Symbol='Minpp1', Aliases='AA408516', description='multiple inositol polyphosphate histidine phosphatase 1', other_designations='2,3-BPG phosphatase|2,3-bisphosphoglycerate 3-phosphatase|inositol (1,3,4,5)-tetrakisphosphate 3-phosphatase|ins(1,3,4,5)P(4) 3-phosphatase', map_location='19 27.25 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=32485769, end_position_on_the_genomic_accession=32515370, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17339, CurrentID=0, Status='live', Symbol='Mip', Aliases='Aqp0, Cat, Cts, Hfi, Lop, MIP26, MP26, Svl, shrivelled', description='major intrinsic protein of lens fiber', other_designations='aquaporin-0|cataract and small eye|dominant cataract|hydropic fibers|lens opacity|major intrinsic protein of eye lens fiber', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128225838, end_position_on_the_genomic_accession=128231811, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17341, CurrentID=0, Status='live', Symbol='Bhlha15', Aliases='1810009C13Rik, Bhlhb8, MIST-1, Mist1', description='basic helix-loop-helix family, member a15', other_designations='basic helix-loop-helix domain containing, class B, 8|muscle, intestine and stomach expression 1', map_location='5 G2-G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144190221, end_position_on_the_genomic_accession=144194441, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  17342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17342, CurrentID=0, Status='live', Symbol='Mitf', Aliases='BCC2, Bhlhe32, Gsfbcc2, Vitiligo, Wh, bw, mi, vit', description='microphthalmia-associated transcription factor', other_designations='black eyed white|microphtalmia-associated transcription factor', map_location='6 45.05 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=97807003, end_position_on_the_genomic_accession=98021360, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  17344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17344, CurrentID=0, Status='live', Symbol='Pias2', Aliases='6330408K17Rik, AI462206, ARIP3, AU018068, Dib, Miz1, PIASxalpha, PIASxb, PIASxbeta, SIZ2', description='protein inhibitor of activated STAT 2', other_designations='DAB2-interacting protein|DIP|Msx-interacting-zinc finger|PIASxalpha6|androgen receptor-interacting protein 3|msx-interacting zinc finger protein|protein inhibitor of activated STAT x', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77065208, end_position_on_the_genomic_accession=77155710, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  17345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17345, CurrentID=0, Status='live', Symbol='Mki67', Aliases='D630048A14Rik, Ki-67, Ki67', description='antigen identified by monoclonal antibody Ki 67', other_designations='', map_location='7 F3-F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=135689782, end_position_on_the_genomic_accession=135716408, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  17346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17346, CurrentID=0, Status='live', Symbol='Mknk1', Aliases='2410048M24Rik, Mnk1', description='MAP kinase-interacting serine/threonine kinase 1', other_designations='MAPK signal-integrating kinase 1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115838984, end_position_on_the_genomic_accession=115879256, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  17347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17347, CurrentID=0, Status='live', Symbol='Mknk2', Aliases='2010016G11Rik, Gprk7, Mnk2', description='MAP kinase-interacting serine/threonine kinase 2', other_designations='G protein-coupled receptor kinase 7|MAP kinase signal-integrating kinase 2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80665318, end_position_on_the_genomic_accession=80676293, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  17349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17349, CurrentID=0, Status='live', Symbol='Mlf1', Aliases='HLS7', description='myeloid leukemia factor 1', other_designations='hematopoietic lineage switch 7|myelodysplasia-myeloid leukemia factor 1', map_location='3 30.91 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=67374097, end_position_on_the_genomic_accession=67400000, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17350, CurrentID=0, Status='live', Symbol='Mlh1', Aliases='1110035C23Rik, AI317206, AI325952, AI561766', description='mutL homolog 1 (E. coli)', other_designations='colon cancer, nonpolyposis type 2|mutL protein homolog 1', map_location='9 60.92 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111228228, end_position_on_the_genomic_accession=111271786, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  17354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17354, CurrentID=0, Status='live', Symbol='Mllt10', Aliases='Af10, B130021D15Rik, D630001B22Rik, mKIAA4140', description='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 10', other_designations='myeloid/lymphoid or mixed lineage-leukemia translocation to 10 homolog', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=18055175, end_position_on_the_genomic_accession=18212390, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  17355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17355, CurrentID=0, Status='live', Symbol='Aff1', Aliases='9630032B01Rik, AW319193, Af4, Mllt2h, Rob', description='AF4/FMR2 family, member 1', other_designations='homolog of human MLLT2 unidentified|myeloid/lymphoid or mixed-lineage leukemia translocated to 2 homolog|protein AF-4|proto-oncogene AF4|robotic', map_location='5 50.45 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=103692334, end_position_on_the_genomic_accession=103855322, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  17356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17356, CurrentID=0, Status='live', Symbol='Mllt4', Aliases='5033403D15Rik, AF6, Af-6, Afadin, Gm314, I-afadin, S-afadin', description='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 4', other_designations='myeloid/lymphoid or mixed lineage-leukemia translocation to 4 homolog|protein Af-6', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13760541, end_position_on_the_genomic_accession=13905798, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  17357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17357, CurrentID=0, Status='live', Symbol='Marcksl1', Aliases='AL022768, AW215397, AW536807, D4Bc1, F52, MacMARCKS, Macs2, Macs3, Mlp, Mrp', description='MARCKS-like 1', other_designations='MARCKS-like protein 1|brain protein F52|mac-MARCKS|macrophage myristoylated alanine-rich C kinase substrate', map_location='4 63.26 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129513581, end_position_on_the_genomic_accession=129515981, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  17364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17364, CurrentID=0, Status='live', Symbol='Trpm1', Aliases='4732499L03Rik, AI606771, Ltrpc1, Mlsn1', description='transient receptor potential cation channel, subfamily M, member 1', other_designations='long transient receptor potential channel 1|melastatin-1', map_location='7 34.61 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64153835, end_position_on_the_genomic_accession=64269772, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  17380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17380, CurrentID=0, Status='live', Symbol='Mme', Aliases='6030454K05Rik, C85356, CALLA, CD10, NEP, SFE', description='membrane metallo endopeptidase', other_designations='atriopeptidase|common acute lymphoblastic leukemia antigen|enkephalinase|neutral endopeptidase 24.11|skin fibroblast elastase', map_location='3 29.97 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=63295226, end_position_on_the_genomic_accession=63383713, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  17381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17381, CurrentID=0, Status='live', Symbol='Mmp12', Aliases='AV378681, MME, Mmel', description='matrix metallopeptidase 12', other_designations='macrophage elastase|matrix metalloproteinase 12', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7344397, end_position_on_the_genomic_accession=7360461, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  17384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17384, CurrentID=0, Status='live', Symbol='Mmp10', Aliases='AV377895, MMP-10, SL-2', description='matrix metallopeptidase 10', other_designations='matrix metalloproteinase 10|transin-2', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7502342, end_position_on_the_genomic_accession=7510242, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17385, CurrentID=0, Status='live', Symbol='Mmp11', Aliases='MMP-11, SL-3, ST3, Stmy3', description='matrix metallopeptidase 11', other_designations='matrix metalloproteinase 11|stromelysin 3', map_location='10 38.62 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75923221, end_position_on_the_genomic_accession=75932520, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  17386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17386, CurrentID=0, Status='live', Symbol='Mmp13', Aliases='Clg, MMP-13, Mmp1', description='matrix metallopeptidase 13', other_designations='collagenase-1|interstitial collagenase|matrix metalloproteinase 13', map_location='9 A1-A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7272514, end_position_on_the_genomic_accession=7283333, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17387, CurrentID=0, Status='live', Symbol='Mmp14', Aliases='AI325305, MMP-X1, MT-MMP-1, MT1-MMP, sabe', description='matrix metallopeptidase 14 (membrane-inserted)', other_designations='Membrane type 1-MMP|matrix metalloproteinase 14 (membrane-inserted)|membrane-type-1 matrix metalloproteinase|type 1 matrix metalloprotease 14', map_location='14 27.79 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54431604, end_position_on_the_genomic_accession=54441258, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17388, CurrentID=0, Status='live', Symbol='Mmp15', Aliases='AI503551, MT-MMP 2, MT2-MMP', description='matrix metallopeptidase 15', other_designations='Membrane type 2-MMP|membrane-type-2 matrix metalloproteinase', map_location='8 47.12 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95352337, end_position_on_the_genomic_accession=95374293, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17389, CurrentID=0, Status='live', Symbol='Mmp16', Aliases='MT-MMP 3, MT3-MMP, Mt3mmp', description='matrix metallopeptidase 16', other_designations='Membrane type 3-MMP|membrane-type matrix metalloproteinase 3', map_location='4 7.08 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=17853482, end_position_on_the_genomic_accession=18118734, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17390, CurrentID=0, Status='live', Symbol='Mmp2', Aliases='Clg4a, GelA, MMP-2', description='matrix metallopeptidase 2', other_designations='72 kDa gelatinase|72kD gelatinase|72kD type IV collagenase|72kDa gelatinase|72kDa type IV collagenase|gelatinase A|matrix metalloproteinase 2', map_location='8 44.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=92827292, end_position_on_the_genomic_accession=92853420, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  17391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17391, CurrentID=0, Status='live', Symbol='Mmp24', Aliases='AU040325, MMP-21, MT5-MMP, MT5MMP, MTMMP5', description='matrix metallopeptidase 24', other_designations='matrix metalloproteinase 24|matrix metalloproteinase-21|membrane type 5-MMP|membrane-type matrix metalloproteinase 5', map_location='2 77.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155775344, end_position_on_the_genomic_accession=155818366, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17392, CurrentID=0, Status='live', Symbol='Mmp3', Aliases='EMS-2, MMP-3, SL-1, SLN-1, SLN1, STR-1, Stmy1, Str1', description='matrix metallopeptidase 3', other_designations='matrix metalloproteinase 3|progelatinase|transin-1', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7445822, end_position_on_the_genomic_accession=7455975, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17393, CurrentID=0, Status='live', Symbol='Mmp7', Aliases='MAT', description='matrix metallopeptidase 7', other_designations='matrilysin, uterine|matrin|matrix metalloproteinase 7|pump-1 protease|uterine metalloproteinase', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7692111, end_position_on_the_genomic_accession=7699416, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  17394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17394, CurrentID=0, Status='live', Symbol='Mmp8', Aliases='BB138268', description='matrix metallopeptidase 8', other_designations='collagenase-2|matrix metalloproteinase-8', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7558429, end_position_on_the_genomic_accession=7568486, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17395, CurrentID=0, Status='live', Symbol='Mmp9', Aliases='AW743869, B/MMP9, Clg4b, Gel B, MMP-9, pro-MMP-9', description='matrix metallopeptidase 9', other_designations='92 kDa gelatinase|92 kDa type IV collagenase|92kD gelatinase|92kD type IV collagenase|collagenase type IVB|gelatinase B', map_location='2 H1-H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164940744, end_position_on_the_genomic_accession=164955850, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  279561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=279561, CurrentID=0, Status='live', Symbol='Wnk3', Aliases='Prkwnk3-ps, Wnk3', description='WNK lysine deficient protein kinase 3', other_designations='WNK lysine deficient protein kinase 3, pseudogene|protein kinase lysine-deficient 3|protein kinase with no lysine 3', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151168777, end_position_on_the_genomic_accession=151320194, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  17420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17420, CurrentID=0, Status='live', Symbol='Mnat1', Aliases='E130115E11Rik, MAT1, P36', description='menage a trois 1', other_designations='CDK7/cyclin H assembly factor|CDK7/cyclin-H assembly factor|RING finger protein MAT1|p35', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73112184, end_position_on_the_genomic_accession=73273988, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17423, CurrentID=0, Status='live', Symbol='Ndst2', Aliases='Mndns, NDST-2', description='N-deacetylase/N-sulfotransferase (heparan glucosaminyl) 2', other_designations='N-HSST 2|N-heparan sulfate sulfotransferase 2|[Heparan sulfate]-glucosamine N-sulfotransferase|glucosaminyl N-deacetylase/N-sulfotransferase 2|glucosaminyl N-deacetylase/N-sulphotransferase 2|heparan N-sulfotransferase 2|heparan glucosaminyl N-deacetylase/N-sulfotransferase 2|heparan sulfate N-deacetylase/N-sulfotransferase 2|mastocytoma N-deacetylase/N-sulfotransferase', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20723730, end_position_on_the_genomic_accession=20734621, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  17425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17425, CurrentID=0, Status='live', Symbol='Foxk1', Aliases='A630048H08Rik, AI463295, ENSMUSG00000075577, Gm10868, Mnf', description='forkhead box K1', other_designations='myocyte nuclear factor', map_location='5 81.53 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142401497, end_position_on_the_genomic_accession=142462015, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17427, CurrentID=0, Status='live', Symbol='Mns1', Aliases='AW546487', description='meiosis-specific nuclear structural protein 1', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72438518, end_position_on_the_genomic_accession=72458576, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  279572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=279572, CurrentID=0, Status='live', Symbol='Tlr13', Aliases='AI666735, Gm713', description='toll-like receptor 13', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106143275, end_position_on_the_genomic_accession=106160493, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17433, CurrentID=0, Status='live', Symbol='Mobp', Aliases='AI835053, AI839848, MOBP155, MOBP170, MOBP69, MOBP73, MOBP81, R74645', description='myelin-associated oligodendrocytic basic protein', other_designations='', map_location='9 71.41 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120149690, end_position_on_the_genomic_accession=120181329, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17434, CurrentID=0, Status='live', Symbol='Mocs2', Aliases='AI415403', description='molybdenum cofactor synthesis 2', other_designations='MOCS2B|molybdenum cofactor synthesis protein 2 large subunit|molybdenum cofactor synthesis protein 2B', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=114818237, end_position_on_the_genomic_accession=114829420, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17436, CurrentID=0, Status='live', Symbol='Me1', Aliases='D9Ertd267e, Mdh-1, Mod-1, Mod1', description='malic enzyme 1, NADP(+)-dependent, cytosolic', other_designations='NADP-ME|malic enzyme, supernatant', map_location='9 46.58 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=86581363, end_position_on_the_genomic_accession=86695914, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17441, CurrentID=0, Status='live', Symbol='Mog', Aliases='B230317G11Rik', description='myelin oligodendrocyte glycoprotein', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37010740, end_position_on_the_genomic_accession=37023398, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  17444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17444, CurrentID=0, Status='live', Symbol='Grap2', Aliases='GRAP-2, GRB2L, GRBLG, GRID, Gads, GrbX, Grf40, GrpL, Mona, P38', description='GRB2-related adaptor protein 2', other_designations='GRB-2-like protein|GRB-2-related monocytic adapter protein|Grb2-related adaptor downstream of Sch|adapter protein GRID|growth factor receptor-binding protein|hematopoietic cell-associated adaptor protein GrpL|monocytic adapter|monocytic adaptor', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80572422, end_position_on_the_genomic_accession=80653047, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17448, CurrentID=0, Status='live', Symbol='Mdh2', Aliases='MDH, Mdh-2, Mor-1, Mor1', description='malate dehydrogenase 2, NAD (mitochondrial)', other_designations='mitochondrial malate dehydrogenase 2, NAD', map_location='5 75.4 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135778649, end_position_on_the_genomic_accession=135790386, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17449, CurrentID=0, Status='live', Symbol='Mdh1', Aliases='B230377B03Rik, D17921, MDH-s, MDHA, Mor-2, Mor2', description='malate dehydrogenase 1, NAD (soluble)', other_designations='cytosolic malate dehydrogenase|malate dehydrogenase, soluble', map_location='11 13.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=21556692, end_position_on_the_genomic_accession=21571934, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  17450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17450, CurrentID=0, Status='live', Symbol='Morc1', Aliases='Morc', description='microrchidia 1', other_designations='morc1|protein microrchidia', map_location='16 30.75 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48431237, end_position_on_the_genomic_accession=48630905, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  17451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17451, CurrentID=0, Status='live', Symbol='Mos', Aliases='c-mos', description='Moloney sarcoma oncogene', other_designations='c-mos proto-oncogene|oocyte maturation factor mos|proto-oncogene c-Mos', map_location='4 A1-A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3870658, end_position_on_the_genomic_accession=3872105, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17454, CurrentID=0, Status='live', Symbol='Mov10', Aliases='C77703, Mov-10', description='Moloney leukemia virus 10', other_designations='moloney leukemia virus 10 protein', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104794828, end_position_on_the_genomic_accession=104818599, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  17463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17463, CurrentID=0, Status='live', Symbol='Psmd7', Aliases='AW107203, Mov-34, Mov34', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 7', other_designations='26S proteasome regulatory subunit S12|26S proteasome regulatory subunit rpn8|proteasome subunit p40', map_location='8 54.38 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107580380, end_position_on_the_genomic_accession=107588482, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17470, CurrentID=0, Status='live', Symbol='Cd200', Aliases='Mox2, OX2', description='CD200 antigen', other_designations='MRC OX-2 antigen|antigen identified by monoclonal antibody MRC OX-2', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45382111, end_position_on_the_genomic_accession=45409058, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17472, CurrentID=0, Status='live', Symbol='Gbp4', Aliases='AW228052, Mag-2, Mpa-2, Mpa2, mKIAA4245', description='guanylate binding protein 4', other_designations='guanylate nucleotide binding protein 4|guanylate-binding protein 4 homolog', map_location='5 50.68 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105115767, end_position_on_the_genomic_accession=105139630, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  17474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17474, CurrentID=0, Status='live', Symbol='Clec4d', Aliases='Clecsf8, Mpcl, mcl', description='C-type lectin domain family 4, member d', other_designations='C-type (calcium dependent, carbohydrate recognition domain) lectin, superfamily member 8|C-type lectin superfamily member 8|C-type lectin, superfamily member 8|macrophage C-type lectin|macrophage-restricted C-type lectin', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=123262107, end_position_on_the_genomic_accession=123275268, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  17475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17475, CurrentID=0, Status='live', Symbol='Mpdz', Aliases='AI225843, B930003D11Rik, MUPP1', description='multiple PDZ domain protein', other_designations='multi-PDZ domain protein 1', map_location='4 38.0 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=81278499, end_position_on_the_genomic_accession=81442918, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  17476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17476, CurrentID=0, Status='live', Symbol='Mpeg1', Aliases='MPS1, Mpg-1', description='macrophage expressed gene 1', other_designations='macrophage gene 1 protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12460709, end_position_on_the_genomic_accession=12465285, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17480, CurrentID=0, Status='live', Symbol='Mpl', Aliases='CD110, TPO-R, c-mpl, hlb219', description='myeloproliferative leukemia virus oncogene', other_designations='myeloproliferative leukemia protein|proto-oncogene c-Mpl', map_location='4 D', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118442412, end_position_on_the_genomic_accession=118457513, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  279653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=279653, CurrentID=0, Status='live', Symbol='Pcdh19', Aliases='B530002L05Rik, Gm717, mKIAA1313', description='protocadherin 19', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=133582860, end_position_on_the_genomic_accession=133688993, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17523, CurrentID=0, Status='live', Symbol='Mpo', Aliases='mKIAA4033', description='myeloperoxidase', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87793784, end_position_on_the_genomic_accession=87804412, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  17524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17524, CurrentID=0, Status='live', Symbol='Mpp1', Aliases='55kDa, C130070C03Rik, p55', description='membrane protein, palmitoylated', other_designations='erythrocyte protein p55|membrane protein, palmitoylated 1', map_location='X 38.15 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75109733, end_position_on_the_genomic_accession=75130949, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  17527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17527, CurrentID=0, Status='live', Symbol='Mpv17', Aliases='Tg.Mpv17', description='MpV17 mitochondrial inner membrane protein', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31140663, end_position_on_the_genomic_accession=31154532, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  17528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17528, CurrentID=0, Status='live', Symbol='Mpz', Aliases='Mpp, P-zero, P0', description='myelin protein zero', other_designations='myelin peripheral protein', map_location='1 79.05 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171150713, end_position_on_the_genomic_accession=171161125, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17532, CurrentID=0, Status='live', Symbol='Mras', Aliases='2900078C09Rik, AI326250', description='muscle and microspikes RAS', other_designations='X-Ras|ras-related protein R-Ras3', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99385420, end_position_on_the_genomic_accession=99436961, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  723818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723818, CurrentID=0, Status='live', Symbol='Mir135b', Aliases='Mirn135b, mir-135b', description='microRNA 135b', other_designations='mir 135b|mmu-mir-135b', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132198088, end_position_on_the_genomic_accession=132198184, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17534, CurrentID=0, Status='live', Symbol='Mrc2', Aliases='Endo180, mKIAA0709, uPARAP', description='mannose receptor, C type 2', other_designations='lectin lambda|macrophage mannose receptor 2|novel lectin', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105292576, end_position_on_the_genomic_accession=105351146, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  17535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17535, CurrentID=0, Status='live', Symbol='Mre11a', Aliases='Mre11, Mre11b', description='meiotic recombination 11 homolog A (S. cerevisiae)', other_designations='MRE11 homolog 1|MRE11 homolog A|meiotic recombination 11 homolog 1|mmMRE11A', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14784670, end_position_on_the_genomic_accession=14837125, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  17536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17536, CurrentID=0, Status='live', Symbol='Meis2', Aliases='A430109D20Rik, Mrg1, Stra10', description='Meis homeobox 2', other_designations='Cbp/p300-interacting transactivator with Glu/Asp-rich carboxy-terminal domain 2|Meis2 homeobox|meis1-related protein 1|myeloid ecotropic viral integration site-related gene 1', map_location='2 58.28 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=115861264, end_position_on_the_genomic_accession=116067555, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  17537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17537, CurrentID=0, Status='live', Symbol='Meis3', Aliases='AI573393, Mrg2', description='Meis homeobox 3', other_designations='meis1-related protein 2|myeloid ecotropic viral integration site-related gene 2', map_location='7 8.76 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16174849, end_position_on_the_genomic_accession=16186510, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17540, CurrentID=0, Status='live', Symbol='Mrvi1', Aliases='BB115629, Irag, Ris1', description='MRV integration site 1', other_designations='JAW1-related protein MRVI1|inositol 1,4,5-triphosphate receptor-associated cGMP kinase substrate|inositol 1,4,5-trisphosphate receptor-associated cGMP kinase substrate|murine retrovirus integration site 1 protein|retroviral integration site 1', map_location='7 57.9 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110868257, end_position_on_the_genomic_accession=110982908, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  68460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68460, CurrentID=0, Status='live', Symbol='Dhrs7c', Aliases='1110001P11Rik, AI120487, SRP-35', description='dehydrogenase/reductase (SDR family) member 7C', other_designations='short-chain dehydrogenase/reductase family 32C member 2', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67798271, end_position_on_the_genomic_accession=67816002, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  279706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=279706, CurrentID=0, Status='live', Symbol='Nup62cl', Aliases='Gm719', description='nucleoporin 62 C-terminal like', other_designations='novel Nsp1-like C-terminal region containing protein', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140006802, end_position_on_the_genomic_accession=140062707, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  723823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723823, CurrentID=0, Status='live', Symbol='Mir219a-1', Aliases='Mir219-1, Mirn219-1, mir-219a-1, mmu-mir-219a-1', description='microRNA 219a-1', other_designations='microRNA 219-1|mir 219-1|mmu-mir-219-1', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34024983, end_position_on_the_genomic_accession=34025092, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723825, CurrentID=0, Status='live', Symbol='Mir103-2', Aliases='Mirn103-2, mir-103-2', description='microRNA 103-2', other_designations='mir 103-2|mmu-mir-103-2', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131288052, end_position_on_the_genomic_accession=131288137, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68468, CurrentID=0, Status='live', Symbol='Ly6g6c', Aliases='1110003M04Rik, AU016360, G6c, NG24', description='lymphocyte antigen 6 complex, locus G6C', other_designations='lymphocyte antigen 6 complex G6C', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35067325, end_position_on_the_genomic_accession=35070048, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  723832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723832, CurrentID=0, Status='live', Symbol='Mir298', Aliases='Mirn298, mir-298, mmu-mir-298', description='microRNA 298', other_designations='mir 298', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174267504, end_position_on_the_genomic_accession=174267585, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  279766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=279766, CurrentID=0, Status='live', Symbol='Rhbdd3', Aliases='Ptag', description='rhomboid domain containing 3', other_designations='pituitary-derived proapoptotic', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5095631, end_position_on_the_genomic_accession=5106100, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  723835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723835, CurrentID=0, Status='live', Symbol='Mir302c', Aliases='Mirn302c, mir-302c', description='microRNA 302c', other_designations='mir 302c|mmu-mir-302c', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127545363, end_position_on_the_genomic_accession=127545430, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68477, CurrentID=0, Status='live', Symbol='Rmnd5a', Aliases='1110007A06Rik, AJ237586, AW125795, Gid2, MTA.D02.090', description='required for meiotic nuclear division 5 homolog A (S. cerevisiae)', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71388634, end_position_on_the_genomic_accession=71440637, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17653, CurrentID=0, Status='live', Symbol='Ms6hm', Aliases='PC-1', description='minisatellite 6 hypermutable', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  723839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723839, CurrentID=0, Status='live', Symbol='Mir323', Aliases='Mirn323, mir-323, mmu-mir-323', description='microRNA 323', other_designations='mir 323', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109712508, end_position_on_the_genomic_accession=109712593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723840, CurrentID=0, Status='live', Symbol='Mir326', Aliases='Mirn326, mir-326, mmu-mir-326', description='microRNA 326', other_designations='mir 326', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99552269, end_position_on_the_genomic_accession=99552363, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723841, CurrentID=0, Status='live', Symbol='Mir328', Aliases='Mirn328, miR-328, mmu-mir-328', description='microRNA 328', other_designations='mir 328', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105308364, end_position_on_the_genomic_accession=105308460, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17681, CurrentID=0, Status='live', Symbol='Msc', Aliases='MyoR, bHLHa22', description='musculin', other_designations='myogenic repressor', map_location='1 4.5 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=14753344, end_position_on_the_genomic_accession=14756012, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  723843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723843, CurrentID=0, Status='live', Symbol='Mir337', Aliases='Mirn337, mir-337, mmu-mir-337', description='microRNA 337', other_designations='mir 337', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109585789, end_position_on_the_genomic_accession=109585885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17684, CurrentID=0, Status='live', Symbol='Cited2', Aliases='AI835299, ER154-like, Mrg1, Msg2, p35srj', description='Cbp/p300-interacting transactivator, with Glu/Asp-rich carboxy-terminal domain, 2', other_designations='MSG-related protein 1|melanocyte specific gene 2', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=17723205, end_position_on_the_genomic_accession=17725674, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  17685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17685, CurrentID=0, Status='live', Symbol='Msh2', Aliases='AI788990', description='mutS homolog 2 (E. coli)', other_designations='mutS protein homolog 2', map_location='17 57.87 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87672557, end_position_on_the_genomic_accession=87723713, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  17686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17686, CurrentID=0, Status='live', Symbol='Msh3', Aliases='D13Em1, Rep-3, Rep3', description='mutS homolog 3 (E. coli)', other_designations='REP-1|protein repair-1|protein repair-3', map_location='13 47.63 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92211879, end_position_on_the_genomic_accession=92355003, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  17687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17687, CurrentID=0, Status='live', Symbol='Msh5', Aliases='Mut5', description='mutS homolog 5 (E. coli)', other_designations='DNA mismatch repair protein Msh5', map_location='17 18.57 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35028605, end_position_on_the_genomic_accession=35048501, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  17688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17688, CurrentID=0, Status='live', Symbol='Msh6', Aliases='AU044881, AW550279, GTBP, Gtmbp', description='mutS homolog 6 (E. coli)', other_designations='g/T mismatch-binding protein|mutS-alpha 160 kDa subunit|p160', map_location='17 57.87 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87975050, end_position_on_the_genomic_accession=87990892, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68484, CurrentID=0, Status='live', Symbol='Krtap6-5', Aliases='1110002D05Rik, Krtap16-8', description='keratin associated protein 6-5', other_designations='keratin associated protein 16-8|keratin-associated protein 16-8|keratin-associated protein 16.8', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=89047289, end_position_on_the_genomic_accession=89047899, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17690, CurrentID=0, Status='live', Symbol='Msi1', Aliases='Msi1h, Musahi1, m-Msi-1', description='musashi RNA-binding protein 1', other_designations='Musashi homolog 1(Drosophila)|musashi-1', map_location='5 E3-F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115359083, end_position_on_the_genomic_accession=115455698, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  17691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17691, CurrentID=0, Status='live', Symbol='Sik1', Aliases='Msk, Sik, Snf1lk', description='salt inducible kinase 1', other_designations='HRT-20|Probable serine/threonine protein kinase SNF1LK (HRT-20) (Myocardial SNF1-like kinase)|SIK-1|SNF1-like kinase|myocardial SNF1-like kinase|salt-inducible kinase 1|salt-inducible protein kinase 1|serine/threonine-protein kinase SNF1-like kinase 1|serine/threonine-protein kinase SNF1LK', map_location='17 17.25 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31844250, end_position_on_the_genomic_accession=31855792, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  17692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17692, CurrentID=0, Status='live', Symbol='Msl3', Aliases='AU0189311, Msl3l1, Msl3', description='male-specific lethal 3 homolog (Drosophila)', other_designations='male-specific lethal-3 homolog 1|male-specific lethal-3 protein-like 1', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=168651267, end_position_on_the_genomic_accession=168674388, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  17695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17695, CurrentID=0, Status='live', Symbol='Msmb', Aliases='PIP, PSP94, beta-MSP', description='beta-microseminoprotein', other_designations='PSP-94|beta-inhibin|prostate secreted seminal plasma protein|prostate secretory protein of 94 amino acids|prostatic inhibin protein', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32142023, end_position_on_the_genomic_accession=32158327, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17698, CurrentID=0, Status='live', Symbol='Msn', Aliases='C78546', description='moesin', other_designations='membrane-organizing extension spike protein', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=96096045, end_position_on_the_genomic_accession=96168553, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  17700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17700, CurrentID=0, Status='live', Symbol='Mstn', Aliases='Cmpt, Gdf8', description='myostatin', other_designations='', map_location='1 26.99 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53061663, end_position_on_the_genomic_accession=53068079, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17701, CurrentID=0, Status='live', Symbol='Msx1', Aliases='AA675338, AI324650, Hox-7, Hox7, Hox7.1, msh', description='msh homeobox 1', other_designations='homeo box, msh-like 1|homeobox protein Hox-7|homeobox, msh-like 1|hox-7.1|msh homeobox 1-like protein|muscle-segment homeobox', map_location='5 20.21 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37820491, end_position_on_the_genomic_accession=37824585, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17702, CurrentID=0, Status='live', Symbol='Msx2', Aliases='BB122635, Hox-8, Hox8, Hox8.1', description='msh homeobox 2', other_designations='homeo box, msh-like 2|homeobox protein Hox-8-1|homeobox, msh-like 2', map_location='13 27.84 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=53466881, end_position_on_the_genomic_accession=53472780, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17703, CurrentID=0, Status='live', Symbol='Msx3', Aliases='AI323377', description='msh homeobox 3', other_designations='homeo box, msh-like 3|homeobox, msh-like 3', map_location='7 85.04 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140046156, end_position_on_the_genomic_accession=140049970, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  17705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17705, CurrentID=0, Status='live', Symbol='mt-Atp6', Aliases='ATP6', description='ATP synthase 6, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=7927, end_position_on_the_genomic_accession=8607, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17706, CurrentID=0, Status='live', Symbol='mt-Atp8', Aliases='ATP8', description='ATP synthase 8, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=7766, end_position_on_the_genomic_accession=7969, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  68487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68487, CurrentID=0, Status='live', Symbol='Tmem140', Aliases='1110007F12Rik', description='transmembrane protein 140', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34863106, end_position_on_the_genomic_accession=34874946, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17708, CurrentID=0, Status='live', Symbol='mt-Co1', Aliases='CoxI, COX1', description='cytochrome c oxidase I, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=5328, end_position_on_the_genomic_accession=6872, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17709, CurrentID=0, Status='live', Symbol='mt-Co2', Aliases='COX2', description='cytochrome c oxidase II, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=7013, end_position_on_the_genomic_accession=7696, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17710, CurrentID=0, Status='live', Symbol='mt-Co3', Aliases='COX3', description='cytochrome c oxidase III, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=8607, end_position_on_the_genomic_accession=9390, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17711, CurrentID=0, Status='live', Symbol='mt-Cytb', Aliases='CYTB', description='cytochrome b, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=14145, end_position_on_the_genomic_accession=15288, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17713, CurrentID=0, Status='live', Symbol='Grpel1', Aliases='AA408748, mt-GrpE#1, mt-Grpel1', description='GrpE-like 1, mitochondrial', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36465185, end_position_on_the_genomic_accession=36474077, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17714, CurrentID=0, Status='live', Symbol='Grpel2', Aliases='mt-GrpE#2, mt-Grpel2', description='GrpE-like 2, mitochondrial', other_designations='', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61712424, end_position_on_the_genomic_accession=61726361, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17716, CurrentID=0, Status='live', Symbol='mt-Nd1', Aliases='ND1', description='NADH dehydrogenase 1, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=2751, end_position_on_the_genomic_accession=3707, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17717, CurrentID=0, Status='live', Symbol='mt-Nd2', Aliases='ND2', description='NADH dehydrogenase 2, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=3914, end_position_on_the_genomic_accession=4951, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17718, CurrentID=0, Status='live', Symbol='mt-Nd3', Aliases='ND3', description='NADH dehydrogenase 3, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=9459, end_position_on_the_genomic_accession=9806, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17719, CurrentID=0, Status='live', Symbol='mt-Nd4', Aliases='ND4', description='NADH dehydrogenase 4, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=10167, end_position_on_the_genomic_accession=11544, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17720, CurrentID=0, Status='live', Symbol='mt-Nd4l', Aliases='ND4L', description='NADH dehydrogenase 4L, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=9877, end_position_on_the_genomic_accession=10173, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17721, CurrentID=0, Status='live', Symbol='mt-Nd5', Aliases='ND5', description='NADH dehydrogenase 5, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=11742, end_position_on_the_genomic_accession=13565, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  17722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17722, CurrentID=0, Status='live', Symbol='mt-Nd6', Aliases='ND6', description='NADH dehydrogenase 6, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=13552, end_position_on_the_genomic_accession=14070, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  17725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17725, CurrentID=0, Status='live', Symbol='mt-Rnr2', Aliases='Rnr2', description='16S rRNA, mitochondrial', other_designations='', map_location='', chromosome='MT', genomic_nucleotide_accession_version='NC_005089.1', start_position_on_the_genomic_accession=1094, end_position_on_the_genomic_accession=2675, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  673094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=673094, CurrentID=0, Status='live', Symbol='Cd99', Aliases='1110061M03Rik, 2410026K10Rik, D4, Pilr-l, pilr-1', description='CD99 antigen', other_designations='paired immunoglobin-like type 2 receptor-ligand', map_location='4', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  68493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68493, CurrentID=0, Status='live', Symbol='Ndufaf4', Aliases='1110007M04Rik, 3000003G13Rik, AW214064', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 4', other_designations='hormone-regulated proliferation-associated protein of 20 kDa homolog', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=24898083, end_position_on_the_genomic_accession=24905001, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17748, CurrentID=0, Status='live', Symbol='Mt1', Aliases='MT-I, Mt-1', description='metallothionein 1', other_designations='metallothionein-I', map_location='8 46.34 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94179088, end_position_on_the_genomic_accession=94180327, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17749, CurrentID=0, Status='live', Symbol='Polr2k', Aliases='ABC10-alpha, MafY, Mt1a, RPABC4, RPB10alpha, RPB12, RPB7.0', description='polymerase (RNA) II (DNA directed) polypeptide K', other_designations='metallothionein-I gene transcription activator', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36174010, end_position_on_the_genomic_accession=36177012, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17750, CurrentID=0, Status='live', Symbol='Mt2', Aliases='AA409533, MT-II, Mt-2', description='metallothionein 2', other_designations='metallothionein-II', map_location='8 46.32 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94172618, end_position_on_the_genomic_accession=94173567, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17751, CurrentID=0, Status='live', Symbol='Mt3', Aliases='Mt-3', description='metallothionein 3', other_designations='GIF|MT-III|growth inhibitory factor|metallothionein-III', map_location='8 46.29 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94152607, end_position_on_the_genomic_accession=94154148, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17752, CurrentID=0, Status='live', Symbol='Mt4', Aliases='MT-IV', description='metallothionein 4', other_designations='MT-4|metallothionein-IV', map_location='8 46.26 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94137204, end_position_on_the_genomic_accession=94139031, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17754, CurrentID=0, Status='live', Symbol='Map1a', Aliases='6330416M19Rik, AI853608, Mtap-1, Mtap1, Mtap1a, moth1', description='microtubule-associated protein 1 A', other_designations='MAP-1A|modifier of tubby hearing 1|short MAP1A|short microtubule-associated protein 1A', map_location='2 60.37 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121289600, end_position_on_the_genomic_accession=121310832, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17755, CurrentID=0, Status='live', Symbol='Map1b', Aliases='A230055D22, AI843217, LC1, MAP5, Mtap-5, Mtap1b, Mtap5', description='microtubule-associated protein 1B', other_designations='MAP-1B|MAP1(X)|MAP1.2|Microtubule-associated protein 1B (MAP 1B) (MAP1.2) (MAP1(X))|microtubule-associated protein 5', map_location='13 52.9 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=99421464, end_position_on_the_genomic_accession=99516602, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  17756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17756, CurrentID=0, Status='live', Symbol='Map2', Aliases='G1-397-34, MAP-2, Mtap-2, Mtap2, repro4', description='microtubule-associated protein 2', other_designations='', map_location='1 33.49 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=66175246, end_position_on_the_genomic_accession=66442583, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  17758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17758, CurrentID=0, Status='live', Symbol='Map4', Aliases='AA407148, MAP-4, Mtap-4, Mtap4', description='microtubule-associated protein 4', other_designations='MAP 4', map_location='9 59.83 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109931774, end_position_on_the_genomic_accession=110083954, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  17760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17760, CurrentID=0, Status='live', Symbol='Map6', Aliases='2810411E12Rik, Map-6, Mtap6, STOP', description='microtubule-associated protein 6', other_designations='stable tubule-only polypeptide', map_location='7 E2-F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99267447, end_position_on_the_genomic_accession=99337137, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17761, CurrentID=0, Status='live', Symbol='Map7', Aliases='E-MAP-115, MAP-7, Mtap7, R75000, mshi, mste, ste', description='microtubule-associated protein 7', other_designations='epithelial microtubule-associated protein of 115 kDa', map_location='10 9.75 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=20148920, end_position_on_the_genomic_accession=20281590, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  17762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17762, CurrentID=0, Status='live', Symbol='Mapt', Aliases='AI413597, AW045860, Mtapt, Tau', description='microtubule-associated protein tau', other_designations='PHF-tau|neurofibrillary tangle protein|paired helical filament-tau', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104231353, end_position_on_the_genomic_accession=104332090, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17763, CurrentID=0, Status='live', Symbol='Mtcp1', Aliases='P13MTCP1', description='mature T cell proliferation 1', other_designations='MTCP-1 type B1|mature T-cell proliferation-1 type B1', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75404846, end_position_on_the_genomic_accession=75416584, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17764, CurrentID=0, Status='live', Symbol='Mtf1', Aliases='MTF-1, Thyls', description='metal response element binding transcription factor 1', other_designations='MRE-binding transcription factor|metalloregulatory transcription factor|transcription factor MTF-1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124802549, end_position_on_the_genomic_accession=124849800, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  17765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17765, CurrentID=0, Status='live', Symbol='Mtf2', Aliases='9230112N11Rik, AA537621, C76717, M96, Pcl2', description='metal response element binding transcription factor 2', other_designations='mPCl2|metal regulatory transcription factor 2|metal response element DNA-binding protein M96|metal-response element DNA-binding protein M96|polycomb-like protein 2|ziRF1|zinc-regulated factor 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108065669, end_position_on_the_genomic_accession=108109219, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  17766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17766, CurrentID=0, Status='live', Symbol='Nudt1', Aliases='Mth1', description='nudix (nucleoside diphosphate linked moiety X)-type motif 1', other_designations="2-hydroxy-dATP diphosphatase|8-oxo-7,8-dihydro-2'-dGTPase|8-oxo-dGTPase|mutT human homolog (8-oxo-dGTPase)|nucleoside diphosphate-linked moiety X motif 1|nudix motif 1", map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140321627, end_position_on_the_genomic_accession=140338137, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68497, CurrentID=0, Status='live', Symbol='Arel1', Aliases='1110018G07Rik, AI649076, AW610792, Kiaa0317, mKIAA0317', description='apoptosis resistant E3 ubiquitin protein ligase 1', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84918140, end_position_on_the_genomic_accession=84970886, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  17768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17768, CurrentID=0, Status='live', Symbol='Mthfd2', Aliases='AW558851, NMDMC', description='methylenetetrahydrofolate dehydrogenase (NAD+ dependent), methenyltetrahydrofolate cyclohydrolase', other_designations='', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83305704, end_position_on_the_genomic_accession=83325947, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17769, CurrentID=0, Status='live', Symbol='Mthfr', Aliases='AI323986', description='5,10-methylenetetrahydrofolate reductase', other_designations='', map_location='4 78.67 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148038018, end_position_on_the_genomic_accession=148059565, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17771, CurrentID=0, Status='live', Symbol='Mtl5', Aliases='tesmin', description='metallothionein-like 5, testis-specific (tesmin)', other_designations='testis-specific metallothionein-like protein', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3388857, end_position_on_the_genomic_accession=3407823, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  17772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17772, CurrentID=0, Status='live', Symbol='Mtm1', Aliases='AF073996, Mtm, mKIAA4176', description='X-linked myotubular myopathy gene 1', other_designations='phosphatidylinositol-3,5-bisphosphate 3-phosphatase|phosphatidylinositol-3-phosphate phosphatase', map_location='X 36.55 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71210767, end_position_on_the_genomic_accession=71315413, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  17773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17773, CurrentID=0, Status='live', Symbol='Mtnr1a', Aliases='MR, MelR', description='melatonin receptor 1A', other_designations='Mel1a receptor|mel-1A-R', map_location='8 24.95 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45069210, end_position_on_the_genomic_accession=45089233, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17775, CurrentID=0, Status='live', Symbol='Laptm4a', Aliases='AA286466, LAPTM4, MTP, Mtrp, mKIAA0108', description='lysosomal-associated protein transmembrane 4A', other_designations='golgi 4-transmembrane spanning transporter|golgi 4-transmembrane-spanning transporter|membrane transporter protein|transporter protein', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8921307, end_position_on_the_genomic_accession=8938742, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17776, CurrentID=0, Status='live', Symbol='Mast2', Aliases='MAST205, Mtssk', description='microtubule associated serine/threonine kinase 2', other_designations='mast2|microtubule associated testis specific serine/threonine protein kinase', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116306760, end_position_on_the_genomic_accession=116464254, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  17777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17777, CurrentID=0, Status='live', Symbol='Mttp', Aliases='1810043K16Rik, MTP', description='microsomal triglyceride transfer protein', other_designations='microsomal triglyceride transfer protein B|microsomal triglyceride transfer protein, large subunit', map_location='3 64.06 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138089853, end_position_on_the_genomic_accession=138143388, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  68499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68499, CurrentID=0, Status='live', Symbol='Mrpl53', Aliases='1110007K17Rik', description='mitochondrial ribosomal protein L53', other_designations='L53mt|MRP-L53', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83109108, end_position_on_the_genomic_accession=83109932, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68501, CurrentID=0, Status='live', Symbol='Nsmce2', Aliases='1110014D18Rik, AI661537', description='non-SMC element 2 homolog (MMS21, S. cerevisiae)', other_designations='MMS21 homolog|non-structural maintenance of chromosomes element 2 homolog', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=59374198, end_position_on_the_genomic_accession=59601689, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17807, CurrentID=0, Status='live', Symbol='Mtv43', Aliases='Mtv-43', description='mammary tumor virus locus 43', other_designations='Mtv-43 superantigen|Mtv-7 sag|endogenous superantigen', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  68505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68505, CurrentID=0, Status='live', Symbol='Vps51', Aliases='1110014N23Rik, 3110057M17Rik, AI837850, Ffr', description='vacuolar protein sorting 51 homolog (S. cerevisiae)', other_designations='protein fat-free homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6067842, end_position_on_the_genomic_accession=6077187, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17826, CurrentID=0, Status='live', Symbol='Fam89b', Aliases='1110021A21Rik, MMTV, Mtvr2', description='family with sequence similarity 89, member B', other_designations='C184M protein|MMTV receptor|mammary tumor virus receptor 2', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5728087, end_position_on_the_genomic_accession=5729666, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  723867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723867, CurrentID=0, Status='live', Symbol='Mir434', Aliases='Mirn434, mir-434, mmu-mir-434', description='microRNA 434', other_designations='mir 434', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109594506, end_position_on_the_genomic_accession=109594599, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17828, CurrentID=0, Status='live', Symbol='Bloc1s5', Aliases='1810074A19Rik, Muted, mu', description='biogenesis of lysosomal organelles complex-1, subunit 5, muted', other_designations='BLOC-1 subunit 5|biogenesis of organelles complex-1, subunit 5, muted|protein Muted homolog', map_location='13 18.35 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38602706, end_position_on_the_genomic_accession=38635884, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17829, CurrentID=0, Status='live', Symbol='Muc1', Aliases='CD227, EMA, Muc-1', description='mucin 1, transmembrane', other_designations='episialin', map_location='3 39.02 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89229056, end_position_on_the_genomic_accession=89233381, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17830, CurrentID=0, Status='live', Symbol='Prol1', Aliases='Muc10', description='proline rich, lacrimal 1', other_designations='mucin 10, submandibular gland salivary mucin', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88317312, end_position_on_the_genomic_accession=88328817, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17831, CurrentID=0, Status='live', Symbol='Muc2', Aliases='2010015E03Rik, MCM, wnn', description='mucin 2', other_designations='MCM|colonic mucin|muc2|secreted gel-forming mucin', map_location='7 87.1 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141744612, end_position_on_the_genomic_accession=141754694, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  17833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17833, CurrentID=0, Status='live', Symbol='Muc5ac', Aliases='2210005L13Rik, MGM', description='mucin 5, subtypes A and C, tracheobronchial/gastric', other_designations='gastric mucin', map_location='7 87.23 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141788958, end_position_on_the_genomic_accession=141819217, orientation='plus', exon_count=50, OMIM='', no_hdr0=''),
  17835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17835, CurrentID=0, Status='live', Symbol='Mug-ps1', Aliases='Mug3', description='murinoglobulin, pseudogene 1', other_designations='murinoglobulin 3', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122176053, end_position_on_the_genomic_accession=122248566, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  17836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17836, CurrentID=0, Status='live', Symbol='Mug1', Aliases='', description='murinoglobulin 1', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121838497, end_position_on_the_genomic_accession=121889057, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  17837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17837, CurrentID=0, Status='live', Symbol='Mug2', Aliases='Cpamd8', description='murinoglobulin 2', other_designations='C3 and PZP-like, alpha-2-macroglobulin domain containing 8', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122006719, end_position_on_the_genomic_accession=122085968, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  68509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68509, CurrentID=0, Status='live', Symbol='Ptx4', Aliases='1110018H23Rik', description='pentraxin 4', other_designations='neuronal pentraxin-like protein C16orf38 homolog', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25120697, end_position_on_the_genomic_accession=25125286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17840, CurrentID=0, Status='live', Symbol='Mup1', Aliases='Ltn-1, Lvtn-1, Mup-1, Mup-a0, Mup7, Up-1, Mup1', description='major urinary protein 1', other_designations='MUP 1|major urinary protein 10|major urinary protein 7', map_location='4 32.85 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60498012, end_position_on_the_genomic_accession=60501960, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17841, CurrentID=0, Status='live', Symbol='Mup2', Aliases='AA589603, Mup-2, Mup18, Mup4', description='major urinary protein 2', other_designations='2610016E04Rik|major urinary protein 4', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60135913, end_position_on_the_genomic_accession=60139872, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17842, CurrentID=0, Status='live', Symbol='Mup3', Aliases='MUP15, MUPIII, Mup-3, Mup25', description='major urinary protein 3', other_designations='MUP 3|alpha-2u-globulin III|major urinary protein 25|non-group 1/group 2 MUP15', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62083476, end_position_on_the_genomic_accession=62087370, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17843, CurrentID=0, Status='live', Symbol='Mup4', Aliases='Mup-4, Mup1', description='major urinary protein 4', other_designations='MUP 4|major urinary protein 1', map_location='4 32.82 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59956806, end_position_on_the_genomic_accession=59960665, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17844, CurrentID=0, Status='live', Symbol='Mup5', Aliases='Mup18', description='major urinary protein 5', other_designations='MUP 5|Mup V|major urinary protein 18', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61831319, end_position_on_the_genomic_accession=61835180, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68510, CurrentID=0, Status='live', Symbol='Ints1', Aliases='1110015K06Rik', description='integrator complex subunit 1', other_designations='int1', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139751282, end_position_on_the_genomic_accession=139775701, orientation='minus', exon_count=47, OMIM='', no_hdr0=''),
  17846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17846, CurrentID=0, Status='live', Symbol='Commd1', Aliases='AI256843, Murr1, U2/Mu', description='COMM domain containing 1', other_designations='U2af1-rs1 region 1|copper metabolism domain containing 1|protein Murr1', map_location='11 14.22 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=22899728, end_position_on_the_genomic_accession=22982843, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17847, CurrentID=0, Status='live', Symbol='Usp34', Aliases='A530081C03Rik, Murr2', description='ubiquitin specific peptidase 34', other_designations='U2af1-rs1 region 2|deubiquitinating enzyme 34|ubiquitin specific protease 34|ubiquitin thioesterase 34|ubiquitin thiolesterase 34|ubiquitin-specific-processing protease 34', map_location='11 14.29 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23306846, end_position_on_the_genomic_accession=23490560, orientation='plus', exon_count=80, OMIM='', no_hdr0=''),
  17849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17849, CurrentID=0, Status='live', Symbol='Murvy', Aliases='', description='murine repeated virus, Y-linked', other_designations='murine repeated virus, Y linked', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17850, CurrentID=0, Status='live', Symbol='Mut', Aliases='D230010K02Rik, Mcm', description='methylmalonyl-Coenzyme A mutase', other_designations='methylmalonyl-CoA isomerase', map_location='17 C-D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40934685, end_position_on_the_genomic_accession=40961989, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  17851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17851, CurrentID=0, Status='live', Symbol='mut1', Aliases='mut-1', description='mutator 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17855, CurrentID=0, Status='live', Symbol='Mvk', Aliases='2310010A05Rik, AI256848, AI414037, MK', description='mevalonate kinase', other_designations='', map_location='5 55.99 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114444257, end_position_on_the_genomic_accession=114460591, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  17857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17857, CurrentID=0, Status='live', Symbol='Mx1', Aliases='AI893580, Mx, Mx-1', description='MX dynamin-like GTPase 1', other_designations='myxoma resistance protein 1|myxovirus (influenza virus) resistance 1|myxovirus (influenza) resistance 1 polypeptide', map_location='16 57.46 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97447035, end_position_on_the_genomic_accession=97462906, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  17858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17858, CurrentID=0, Status='live', Symbol='Mx2', Aliases='AI528743, Mx-2', description='MX dynamin-like GTPase 2', other_designations='myxovirus (influenza virus) resistance 2|myxovirus resistance protein 2', map_location='16 57.51 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97536081, end_position_on_the_genomic_accession=97560901, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  17859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17859, CurrentID=0, Status='live', Symbol='Mxi1', Aliases='ENSMUSG00000067085, Gm10197, Mad2, bHLHc11', description='MAX interactor 1, dimerization protein', other_designations='Max interacting protein 1|max interactor 1', map_location='19 D', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53310371, end_position_on_the_genomic_accession=53375810, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  17863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17863, CurrentID=0, Status='live', Symbol='Myb', Aliases='AI550390, M16449, c-myb', description='myeloblastosis oncogene', other_designations='gag-myb protein|myb proto-oncogene protein|myeloblastosis proto-oncogene product|proto-oncogene c-Myb|tumor-specific myb protein', map_location='10 9.75 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=21124930, end_position_on_the_genomic_accession=21160984, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17864, CurrentID=0, Status='live', Symbol='Mybl1', Aliases='A-myb, G1-419-6, repro9', description='myeloblastosis oncogene-like 1', other_designations='myb-like protein 1', map_location='1 2.08 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9667407, end_position_on_the_genomic_accession=9700439, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17865, CurrentID=0, Status='live', Symbol='Mybl2', Aliases='B-Myb, Bmyb', description='myeloblastosis oncogene-like 2', other_designations='myb-like protein 2', map_location='2 84.0 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163054333, end_position_on_the_genomic_accession=163084688, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  17868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17868, CurrentID=0, Status='live', Symbol='Mybpc3', Aliases='', description='myosin binding protein C, cardiac', other_designations='C-protein, cardiac muscle isoform|cardiac C-protein|cardiac MyBP-C', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91118114, end_position_on_the_genomic_accession=91136518, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  17869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17869, CurrentID=0, Status='live', Symbol='Myc', Aliases='AU0167572, Niard, Nird, bHLHe39, Myc', description='myelocytomatosis oncogene', other_designations='c-myc proto-oncogene|transcription factor p64', map_location='15 D2-D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=61985341, end_position_on_the_genomic_accession=61990361, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17870, CurrentID=0, Status='live', Symbol='Mycs', Aliases='', description='myc-like oncogene, s-myc protein', other_designations='', map_location='X 2.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=5466904, end_position_on_the_genomic_accession=5469265, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  112205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112205, CurrentID=0, Status='live', Symbol='Ity2', Aliases='', description='immunity to S. typhimurium 2', other_designations='', map_location='11 105.28 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17872, CurrentID=0, Status='live', Symbol='Ppp1r15a', Aliases='9630030H21, Gadd34, Myd116', description='protein phosphatase 1, regulatory (inhibitor) subunit 15A', other_designations='growth arrest and DNA damage-inducible protein GADD34|growth arrest and DNA-damage-inducible 34|myeloid differentiation primary response gene 116|myeloid differentiation primary response protein MyD116', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45522917, end_position_on_the_genomic_accession=45526268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  17873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17873, CurrentID=0, Status='live', Symbol='Gadd45b', Aliases='AI323528, Myd118', description='growth arrest and DNA-damage-inducible 45 beta', other_designations='growth arrest and DNA-damage-inducible protein GADD45 beta|myeloid differentiation primary response gene 118|myeloid differentiation primary response protein MyD118|negative growth regulatory protein MyD118', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80930091, end_position_on_the_genomic_accession=80932204, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  17874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17874, CurrentID=0, Status='live', Symbol='Myd88', Aliases='', description='myeloid differentiation primary response gene 88', other_designations='', map_location='9 71.33 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119335988, end_position_on_the_genomic_accession=119340040, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17876, CurrentID=0, Status='live', Symbol='Myef2', Aliases='9430071B01, mKIAA1341', description='myelin basic protein expression factor 2, repressor', other_designations='MEF-2|myEF-2', map_location='2 61.16 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125082209, end_position_on_the_genomic_accession=125123661, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  17877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17877, CurrentID=0, Status='live', Symbol='Myf5', Aliases='B130010J22Rik, Myf-5, bHLHc2', description='myogenic factor 5', other_designations='myogenic regulatory factor MYF-5', map_location='10 55.95 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=107482908, end_position_on_the_genomic_accession=107486927, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17878, CurrentID=0, Status='live', Symbol='Myf6', Aliases='MRF4, bHLHc4, herculin', description='myogenic factor 6', other_designations='muscle-specific regulatory factor 4|myf-6', map_location='10 55.96 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=107492848, end_position_on_the_genomic_accession=107502515, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17879, CurrentID=0, Status='live', Symbol='Myh1', Aliases='A530084A17Rik, IId, IId/x, MHC-2X/D, MHC2X/D, MYHC-IIX, MdMs, MyHC-IId/x, MyHC-IIx/d, Myhs-f, Myhs-f2, Myhsf2', description='myosin, heavy polypeptide 1, skeletal muscle, adult', other_designations='myHC-2x|myosin heavy chain 1|myosin heavy chain 2X|myosin heavy chain, skeletal muscle, adult 1', map_location='11 40.59 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67200114, end_position_on_the_genomic_accession=67224575, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  17880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17880, CurrentID=0, Status='live', Symbol='Myh11', Aliases='AV071570, SM1, SM2, smMHC', description='myosin, heavy polypeptide 11, smooth muscle', other_designations='myosin heavy chain 11, smooth muscle|myosin heavy chain, smooth muscle isoform', map_location='16 9.71 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14194527, end_position_on_the_genomic_accession=14291408, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  723876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723876, CurrentID=0, Status='live', Symbol='Mir486', Aliases='Mirn486, mir-486a, mmu-mir-486, mmu-mir-486a', description='microRNA 486', other_designations='mir 486', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=23142555, end_position_on_the_genomic_accession=23142682, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17882, CurrentID=0, Status='live', Symbol='Myh2', Aliases='MHC2A, MyHC-IIaa, Myhs-f, Myhs-f1, Myhsf1, Myh2', description='myosin, heavy polypeptide 2, skeletal muscle, adult', other_designations='myosin heavy chain 2A|skeletal muscle myosin heavy chain II-a', map_location='11 40.59 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67171027, end_position_on_the_genomic_accession=67197517, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  17883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17883, CurrentID=0, Status='live', Symbol='Myh3', Aliases='MyHC-emb, Myhs-e, Myhse', description='myosin, heavy polypeptide 3, skeletal muscle, embryonic', other_designations='embryonic skeletal myosin heavy chain isoform|myosin heavy chain 3', map_location='11 40.59 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67078273, end_position_on_the_genomic_accession=67102291, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  17884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17884, CurrentID=0, Status='live', Symbol='Myh4', Aliases='AI506973, MHC2B, MM, MYH-2B, Minimsc, Minmus, MyHC-IIb, Myhsf', description='myosin, heavy polypeptide 4, skeletal muscle', other_designations='myHC-2b|myosin heavy chain 2b|myosin heavy chain 4', map_location='11 40.59 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67237812, end_position_on_the_genomic_accession=67260447, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  17885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17885, CurrentID=0, Status='live', Symbol='Myh8', Aliases='4832426G23Rik, AI327267, MHCp, MyHC-pn, Myhs-p, Myhsp', description='myosin, heavy polypeptide 8, skeletal muscle, perinatal', other_designations='Beta-MHC antisense|myHC-perinatal|myosin heavy chain 8|myosin heavy chain, skeletal muscle, perinatal', map_location='11 40.59 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67277124, end_position_on_the_genomic_accession=67308634, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  17886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17886, CurrentID=0, Status='live', Symbol='Myh9', Aliases='Fltn, Myhn-1, Myhn1, NMHCIIA, NMMHC-A, NMMHC-IIA, TU72.6', description='myosin, heavy polypeptide 9, non-muscle', other_designations='NMHC II-A|NMMHC II-a|cellular myosin heavy chain, type A|flectin|myosin IIA|myosin heavy chain IX|non-muscle myosin heavy chain IIa', map_location='15 36.81 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77760585, end_position_on_the_genomic_accession=77842175, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  17888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17888, CurrentID=0, Status='live', Symbol='Myh6', Aliases='A830009F23Rik, AA517445, Myhc-a, Myhca, alpha-MHC', description='myosin, heavy polypeptide 6, cardiac muscle, alpha', other_designations='Myosin heavy chain, cardiac muscle alpha isoform (MyHC-alpha)|alpha cardiac MHC|alpha myosin|cardiac myosin heavy chain alpha isoform|cardiomyopathy, hypertrophic 1|myHC-alpha|myosin heavy chain 6|myosin heavy chain, cardiac muscle, adult', map_location='14 28.01 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54941921, end_position_on_the_genomic_accession=54968014, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  723878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723878, CurrentID=0, Status='live', Symbol='Mir494', Aliases='Mirn494, mir-494, mmu-mir-494', description='microRNA 494', other_designations='mir 494', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109715318, end_position_on_the_genomic_accession=109715402, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17895, CurrentID=0, Status='live', Symbol='Myk103', Aliases='Myk-103', description='sequence flanking transgenic insert', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17896, CurrentID=0, Status='live', Symbol='Myl4', Aliases='ALC1, AMLC, ELC, ELC1a, GT1, MLC1a, Myla', description='myosin, light polypeptide 4', other_designations='MLC1EMB|myosin light chain 1, atrial/fetal isoform|myosin light chain, alkali, cardiac atria|myosin, light polypeptide 4, alkali; atrial, embryonic', map_location='11 E', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104550530, end_position_on_the_genomic_accession=104587219, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  17897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17897, CurrentID=0, Status='live', Symbol='Myl3', Aliases='MLC1s, MLC1v, Mylc, VLC1', description='myosin, light polypeptide 3', other_designations='MLC1SB|myosin alkali light chain 1, ventricular/slow skeletal muscle isoform|myosin light chain 1, slow-twitch muscle B/ventricular isoform|myosin light chain, alkali, cardiac ventricles|myosin, light polypeptide 3, alkali; ventricular, skeletal, slow|slow skeletal|ventricular|ventricular/slow twitch myosin alkali light chain', map_location='9 60.69 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110763681, end_position_on_the_genomic_accession=110769794, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  17898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17898, CurrentID=0, Status='live', Symbol='Myl7', Aliases='MLC-2alpha, MLC2a, MYL2A, Mylc2a, RLC-A', description='myosin, light polypeptide 7, regulatory', other_designations='MLC-2a|myosin light chain 2a|myosin light chain, regulatory A|myosin regulatory light chain 7', map_location='11 3.88 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5896637, end_position_on_the_genomic_accession=5898782, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68519, CurrentID=0, Status='live', Symbol='Eml1', Aliases='1110008N23Rik, A930030P13Rik, AA171013, AI847476, AI853955, ELP79, EMAP, EMAPL, heco', description='echinoderm microtubule associated protein like 1', other_designations='EMAP-1|echinoderm microtubule associated protein-like protein 1', map_location='12 59.46 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108371153, end_position_on_the_genomic_accession=108539564, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  17901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17901, CurrentID=0, Status='live', Symbol='Myl1', Aliases='AI325107, MLC1f, MLC3f, Mylf', description='myosin, light polypeptide 1', other_designations='A1 catalytic|A2 catalytic|MLC1/MLC3|MLC1F/MLC3F|myosin light chain A1/A2|myosin light chain alkali 1/2|myosin light chain, alkali, fast skeletal muscle|myosin, light polypeptide 1, alkali; atrial, embryonic|myosin, light polypeptide 4, alkali; atrial, embryonic', map_location='1 33.71 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=66924295, end_position_on_the_genomic_accession=66945473, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17904, CurrentID=0, Status='live', Symbol='Myl6', Aliases='ESMLC, LC17, LC17-GI, MLC-3, MLC1SM, Myln', description='myosin, light polypeptide 6, alkali, smooth muscle and non-muscle', other_designations='17 kDa myosin light chain|myosin light chain A3|myosin light chain alkali 3|myosin light chain, alkali, nonmuscle|smooth muscle and nonmuscle myosin light chain alkali 6', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128490861, end_position_on_the_genomic_accession=128493896, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  17906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17906, CurrentID=0, Status='live', Symbol='Myl2', Aliases='MLC-2, MLC-2s/v, MLC-2v, Mlc2v, Mylpc', description='myosin, light polypeptide 2, regulatory, cardiac, slow', other_designations='myosin light chain 2, slow skeletal/ventricular muscle isoform|myosin light chain 2v|myosin light chain, phosphorylatable, cardiac ventricles|ventricular myosin regulatory light chain', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122100904, end_position_on_the_genomic_accession=122113396, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  17907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17907, CurrentID=0, Status='live', Symbol='Mylpf', Aliases='2410014J02Rik, MLC-2, Mlc2', description='myosin light chain, phosphorylatable, fast skeletal muscle', other_designations='MLC2F|fast skeletal myosin light chain 2|myosin light chain 2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127208848, end_position_on_the_genomic_accession=127214288, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17909, CurrentID=0, Status='live', Symbol='Myo10', Aliases='AW048724, D15Ertd600e, mKIAA0799', description='myosin X', other_designations='myosin-X|unconventional myosin-10|unconventionnal myosin-X', map_location='15 9.36 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=25622550, end_position_on_the_genomic_accession=25813673, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  17910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17910, CurrentID=0, Status='live', Symbol='Myo15', Aliases='Myo15a, sh-2, sh2', description='myosin XV', other_designations='myosin XVA|myosin-XV|shaker 2|unconventional myosin-15', map_location='11 37.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60469339, end_position_on_the_genomic_accession=60528369, orientation='plus', exon_count=67, OMIM='', no_hdr0=''),
  723881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723881, CurrentID=0, Status='live', Symbol='Mir543', Aliases='Mirn543, mir-543, mmu-mir-543', description='microRNA 543', other_designations='mir 543', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109717258, end_position_on_the_genomic_accession=109717333, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17912, CurrentID=0, Status='live', Symbol='Myo1b', Aliases='AA406997, MIH-L, MMIa', description='myosin IB', other_designations='MMI-alpha|Myosin IA (Myosin I alpha) (MMI-alpha) (MMIa) (MIH-L)|molecular motor|myosin I alpha|myosin-Ib', map_location='1 26.58 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=51749758, end_position_on_the_genomic_accession=51915978, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  17913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17913, CurrentID=0, Status='live', Symbol='Myo1c', Aliases='C80397, MMIb, MYO1E, NMI, mm1beta, myr2', description='myosin IC', other_designations='MMI-beta|myosin I beta|myosin-Ibeta|myosin-Ic|nuclear myosin I beta', map_location='11 C-E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75650493, end_position_on_the_genomic_accession=75674643, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  17916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17916, CurrentID=0, Status='live', Symbol='Myo1f', Aliases='C330006B10Rik', description='myosin IF', other_designations='myosin-1F-like protein|myosin-If', map_location='17 B-C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33555707, end_position_on_the_genomic_accession=33607764, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  17918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17918, CurrentID=0, Status='live', Symbol='Myo5a', Aliases='9630007J19Rik, AI413174, AI661011, Dbv, MVa, Myo5, MyoVA, Sev-1, d, d-120J, flail, flr', description='myosin VA', other_designations='dilute lethal-20J protein|dilute myosin heavy chain, non-muscle|flailer|myosin-Va', map_location='9 42.26 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75069298, end_position_on_the_genomic_accession=75223688, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  17919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17919, CurrentID=0, Status='live', Symbol='Myo5b', Aliases='AI661750, mKIAA1119', description='myosin VB', other_designations='myosin-Vb', map_location='18 50.7 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74442378, end_position_on_the_genomic_accession=74772698, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  17920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17920, CurrentID=0, Status='live', Symbol='Myo6', Aliases='BC029719, Tlc, sv', description='myosin VI', other_designations="Snell's waltzer|myosin-VI|tailchaser|unconventional myosin-6", map_location='9 43.98 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=80164976, end_position_on_the_genomic_accession=80311729, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  17921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17921, CurrentID=0, Status='live', Symbol='Myo7a', Aliases='Hdb, Myo7, USH1B, nmf371, polka, sh-1, sh1', description='myosin VIIA', other_designations='motor protein|myosin-VIIa|shaker 1', map_location='7 53.57 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98051054, end_position_on_the_genomic_accession=98119522, orientation='minus', exon_count=53, OMIM='', no_hdr0=''),
  17922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17922, CurrentID=0, Status='live', Symbol='Myo7b', Aliases='', description='myosin VIIB', other_designations='myosin-VIIb', map_location='18 17.89 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31959234, end_position_on_the_genomic_accession=32036931, orientation='minus', exon_count=47, OMIM='', no_hdr0=''),
  68523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68523, CurrentID=0, Status='live', Symbol='Fam96b', Aliases='1110019N10Rik', description='family with sequence similarity 96, member B', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104639839, end_position_on_the_genomic_accession=104641728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17925, CurrentID=0, Status='live', Symbol='Myo9b', Aliases='', description='myosin IXb', other_designations='myosin-IXb|unconventional myosin-9b', map_location='8 34.43 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71272619, end_position_on_the_genomic_accession=71360713, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  17926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17926, CurrentID=0, Status='live', Symbol='Myoc', Aliases='AI957332, GLC1A, TIGR', description='myocilin', other_designations='trabecular meshwork-induced glucocorticoid response protein', map_location='1 70.29 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162639150, end_position_on_the_genomic_accession=162649694, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17927, CurrentID=0, Status='live', Symbol='Myod1', Aliases='AI503393, MYF3, MyoD, Myod-1, bHLHc1', description='myogenic differentiation 1', other_designations='', map_location='7 30.03 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46376474, end_position_on_the_genomic_accession=46379092, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17928, CurrentID=0, Status='live', Symbol='Myog', Aliases='MYF4, bHLHc3, myo', description='myogenin', other_designations='MYOD1-related protein', map_location='1 58.18 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134290004, end_position_on_the_genomic_accession=134292548, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17929, CurrentID=0, Status='live', Symbol='Myom1', Aliases='D430047A17Rik', description='myomesin 1', other_designations='myomesin family member 1|skelemin', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71019002, end_position_on_the_genomic_accession=71126856, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  17930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17930, CurrentID=0, Status='live', Symbol='Myom2', Aliases='AW146149', description='myomesin 2', other_designations='skeletal muscle 165kD protein|skeletal muscle 165kDa protein', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=15057653, end_position_on_the_genomic_accession=15133420, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  17931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17931, CurrentID=0, Status='live', Symbol='Ppp1r12a', Aliases='1200015F06Rik, 5730577I22Rik, AA792106, AV099298, D10Ertd625e, Mypt1', description='protein phosphatase 1, regulatory (inhibitor) subunit 12A', other_designations='myosin phosphatase target subunit 1|myosin phosphatase, target subunit 1|myosin phosphatase-targeting subunit 1', map_location='10 C', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=108162065, end_position_on_the_genomic_accession=108279646, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  17932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17932, CurrentID=0, Status='live', Symbol='Myt1', Aliases='Nzf2, Nzf2a, Nzf2b, Nztf2', description='myelin transcription factor 1', other_designations='NZF-2|neural zinc finger factor 2|neural zinc finger transcription factor 2', map_location='2 103.77 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181763332, end_position_on_the_genomic_accession=181827781, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  17933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17933, CurrentID=0, Status='live', Symbol='Myt1l', Aliases='2900046C06Rik, 2900093J19Rik, C630034G21Rik, Nztf1, Pmng1, Png-1, mKIAA1106', description='myelin transcription factor 1-like', other_designations='NZF-1|myT1-L|neural zinc finger factor 1|neural zinc finger protein NZF-1|neural zinc finger transcription factor 1|postmeiotic neural gene 1|postmitotic neural gene 1 protein|zinc finger protein Png-1', map_location='12 11.86 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=29528311, end_position_on_the_genomic_accession=29923216, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  68525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68525, CurrentID=0, Status='live', Symbol='Evc2', Aliases='1110017L09Rik, Lbn, limbin', description='Ellis van Creveld syndrome 2', other_designations='Ellis van Creveld syndrome 2 homolog', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37338478, end_position_on_the_genomic_accession=37425055, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  17936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17936, CurrentID=0, Status='live', Symbol='Nab1', Aliases='', description='Ngfi-A binding protein 1', other_designations='EGR-1-binding protein 1', map_location='1 26.99 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52455849, end_position_on_the_genomic_accession=52500448, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  17937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17937, CurrentID=0, Status='live', Symbol='Nab2', Aliases='AI451907', description='Ngfi-A binding protein 2', other_designations='EGR-1-binding protein 2|NGFI-A binding protein 2 (EGR-1 binding protein 2)', map_location='10 74.6 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127660918, end_position_on_the_genomic_accession=127666703, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  17938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17938, CurrentID=0, Status='live', Symbol='Naca', Aliases='AL022831, AL024382, Gm1878, mKIAA0363, skNAC', description='nascent polypeptide-associated complex alpha polypeptide', other_designations='alpha-NAC, muscle-specific form|alpha-NAC/1.9.2|nascent polypeptide-associated complex alpha subunit|nascent polypeptide-associated complex subunit alpha, muscle-specific form', map_location='10 D2-D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128035346, end_position_on_the_genomic_accession=128048637, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  17939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17939, CurrentID=0, Status='live', Symbol='Naga', Aliases='', description='N-acetyl galactosaminidase, alpha', other_designations='alpha-galactosidase B', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82329532, end_position_on_the_genomic_accession=82338840, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17940, CurrentID=0, Status='live', Symbol='Naip1', Aliases='AV364616, Birc1a, D13Lsd1, Naip, Naip-rs1', description='NLR family, apoptosis inhibitory protein 1', other_designations='baculoviral IAP repeat-containing 1a|neuronal apoptosis inhibitory protein 1', map_location='13 D1-D3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100407764, end_position_on_the_genomic_accession=100452864, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  723886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723886, CurrentID=0, Status='live', Symbol='Mir96', Aliases='Mirn96, mir-96, mmu-mir-96', description='microRNA 96', other_designations='mir 96', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30169446, end_position_on_the_genomic_accession=30169551, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  330671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330671, CurrentID=0, Status='live', Symbol='B4galnt4', Aliases='BC038881, Gm1095, beta4GalNAcT4', description='beta-1,4-N-acetyl-galactosaminyl transferase 4', other_designations='NGalNAc-T1|beta 1,4-N-acetylgalactosaminyltransferase-transferase 4|beta-1,4-N-acetylgalactosaminyltransferase IV|beta1,4-N-acetylgalactosaminyltransferase IV|beta4GalNAc-T4', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141061109, end_position_on_the_genomic_accession=141072119, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  17948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17948, CurrentID=0, Status='live', Symbol='Naip2', Aliases='Birc1b, Naip-rs6', description='NLR family, apoptosis inhibitory protein 2', other_designations='neuronal apoptosis inhibitory protein, related sequence 6', map_location='13 53.01 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100144063, end_position_on_the_genomic_accession=100202122, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17950, CurrentID=0, Status='live', Symbol='Naip4', Aliases='Birc1d, Naip-rs2', description='NLR family, apoptosis inhibitory protein 4', other_designations='neuronal apoptosis inhibitory protein 4|neuronal apoptosis inhibitory protein, related sequence 2', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  17951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17951, CurrentID=0, Status='live', Symbol='Naip5', Aliases='Birc1e, Lgn1, Naip-rs3', description='NLR family, apoptosis inhibitory protein 5', other_designations='Legionella, susceptiblility 1|baculoviral IAP repeat-containing 1e|neuronal apoptosis inhibitory protein 5|neuronal apoptosis inhibitory protein, related sequence 3', map_location='13 53.05 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100211739, end_position_on_the_genomic_accession=100254771, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  17952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17952, CurrentID=0, Status='live', Symbol='Naip6', Aliases='Birc1f, Naip-rs4, Naip-rs4A', description='NLR family, apoptosis inhibitory protein 6', other_designations='neuronal apoptosis inhibitory protein 6|neuronal apoptosis inhibitory protein, related sequence 4', map_location='13 53.09 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100281121, end_position_on_the_genomic_accession=100316616, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  17954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17954, CurrentID=0, Status='live', Symbol='Nap1l2', Aliases='Bpx', description='nucleosome assembly protein 1-like 2', other_designations='brain-specific protein, X-linked', map_location='X 45.97 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103184059, end_position_on_the_genomic_accession=103186664, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  17955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17955, CurrentID=0, Status='live', Symbol='Nap1l4', Aliases='2810410H14Rik, AI316776, D7Wsu30e, Nap2', description='nucleosome assembly protein 1-like 4', other_designations='nucleosome assembly protein 1 like 4', map_location='7 88.24 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143513579, end_position_on_the_genomic_accession=143549196, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  17957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17957, CurrentID=0, Status='live', Symbol='Napb', Aliases='Brp14, E161, I47, SNARE, b-SNAP', description='N-ethylmaleimide sensitive fusion protein attachment protein beta', other_designations='N-ethylmaleimide-sensitive factor attachment protein beta|SNAP-beta|brain protein 14|brain protein I47', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148693985, end_position_on_the_genomic_accession=148732420, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  17960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17960, CurrentID=0, Status='live', Symbol='Nat1', Aliases='Nat-1', description='N-acetyl transferase 1', other_designations='N(alpha)-acetyltransferase 15, NatA auxiliary subunit|N-acetyl transferase 1, liver, blood|N-acetyltransferase type 1|arylamide acetylase 1', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=67485170, end_position_on_the_genomic_accession=67492582, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17961, CurrentID=0, Status='live', Symbol='Nat2', Aliases='AV377607', description='N-acetyltransferase 2 (arylamine N-acetyltransferase)', other_designations='N-acetyl transferase 2|N-acetyltransferase type 2|NAT-2|arylamide acetylase 2', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=67494875, end_position_on_the_genomic_accession=67502644, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17962, CurrentID=0, Status='live', Symbol='Nat3', Aliases='', description='N-acetyltransferase 3', other_designations='N-acetyltransferase type 3|NAT-3|arylamide acetylase 3', map_location='8 33.39 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=67523854, end_position_on_the_genomic_accession=67548627, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  17965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17965, CurrentID=0, Status='live', Symbol='Nbl1', Aliases='D4H1S1733E, DAN, Dana, NO3', description='neuroblastoma, suppression of tumorigenicity 1', other_designations='N03|zinc finger protein DAN', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139082291, end_position_on_the_genomic_accession=139093013, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17966, CurrentID=0, Status='live', Symbol='Nbr1', Aliases='mKIAA0049', description='neighbor of Brca1 gene 1', other_designations='membrane component chromosome 17 surface marker 2 homolog|neighbor of BRCA1 gene 1 protein|next to the Brca1', map_location='11 65.36 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101552107, end_position_on_the_genomic_accession=101581951, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  17967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17967, CurrentID=0, Status='live', Symbol='Ncam1', Aliases='CD56, E-NCAM, NCAM-1, Ncam', description='neural cell adhesion molecule 1', other_designations='N-CAM-1', map_location='9 26.83 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=49502134, end_position_on_the_genomic_accession=49799427, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  17968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17968, CurrentID=0, Status='live', Symbol='Ncam2', Aliases='Ncam-2, Ocam, RNCAM', description='neural cell adhesion molecule 2', other_designations='N-CAM-2|R4B12 antigen|neural cell adhesion molecule RB-8|olfactory axon cell adhesion molecule', map_location='16 45.89 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=81200697, end_position_on_the_genomic_accession=81624290, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  17969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17969, CurrentID=0, Status='live', Symbol='Ncf1', Aliases='NCF-47K, NOXO2, Ncf-1, p47-phox, p47<phox>, p47phox', description='neutrophil cytosolic factor 1', other_designations='47 kDa neutrophil oxidase factor|NADPH oxidase subunit (47 kD)|NADPH oxidase subunit (47kDa)|neutrophil NADPH oxidase factor 1', map_location='5 74.47 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134220053, end_position_on_the_genomic_accession=134229625, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  17970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17970, CurrentID=0, Status='live', Symbol='Ncf2', Aliases='NOXA2, Ncf-2, p67phox', description='neutrophil cytosolic factor 2', other_designations='67 kDa neutrophil oxidase factor|NADPH oxidase activator 2|NADPH oxidase subunit (67kDa)|neutrophil NADPH oxidase factor 2', map_location='1 65.05 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152800158, end_position_on_the_genomic_accession=152837169, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  17972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17972, CurrentID=0, Status='live', Symbol='Ncf4', Aliases='AI451400, p40phox', description='neutrophil cytosolic factor 4', other_designations='NCF-4|neutrophil NADPH oxidase factor 4|p40-phox', map_location='15 37.1 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78244164, end_position_on_the_genomic_accession=78262582, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  17973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17973, CurrentID=0, Status='live', Symbol='Nck1', Aliases='6330586M15Rik, D230010O13Rik, Nck, Nck-1', description='non-catalytic region of tyrosine kinase adaptor protein 1', other_designations='NCK adaptor protein 1', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=100494981, end_position_on_the_genomic_accession=100546053, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  17974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17974, CurrentID=0, Status='live', Symbol='Nck2', Aliases='4833426I10Rik, Grb4, NCKbeta, Nck-2', description='non-catalytic region of tyrosine kinase adaptor protein 2', other_designations='NCK adaptor protein 2|SH2/SH3 adaptor protein NCK-beta|growth factor receptor-bound protein 4', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=43445749, end_position_on_the_genomic_accession=43570518, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  17975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17975, CurrentID=0, Status='live', Symbol='Ncl', Aliases='B530004O11Rik, C23, D0Nds28, D1Nds28, Nucl', description='nucleolin', other_designations='', map_location='1 43.94 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86344719, end_position_on_the_genomic_accession=86359455, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  17977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17977, CurrentID=0, Status='live', Symbol='Ncoa1', Aliases='KAT13A, SRC-1, SRC1, bHLHe74', description='nuclear receptor coactivator 1', other_designations='NCoA-1|SRC-a/NCoA-1|mNRC-1|nuclear receptor coactivator protein 1|steroid receptor coactivator 1|steroid receptor coactivator-1', map_location='12 A2-A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4247362, end_position_on_the_genomic_accession=4477210, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  17978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17978, CurrentID=0, Status='live', Symbol='Ncoa2', Aliases='D1Ertd433e, GRIP-1, Grip1, KAT13C, NCoA-2, SRC-2, TIF2, bHLHe75', description='nuclear receptor coactivator 2', other_designations='glucocorticoid receptor interacting protein 1|steroid receptor coactivator 2|transcriptional intermediary factor 2', map_location='1 4.12 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=13139105, end_position_on_the_genomic_accession=13374691, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  17979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17979, CurrentID=0, Status='live', Symbol='Ncoa3', Aliases='2010305B15Rik, AW321064, Actr, Aib1, KAT13B, Rac3, Src3, Tram-1, Tram1, bHLHe42, p/Cip, pCip', description='nuclear receptor coactivator 3', other_designations='AIB-1|CBP-interacting protein|NCoA-3|RAC-3|SRC-3|amplified in breast cancer-1 protein homolog|receptor-associated coactivator 3|steroid receptor coactivator protein 3|thyroid hormone receptor activator molecule 1', map_location='2 H2-H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165992637, end_position_on_the_genomic_accession=166073242, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  68533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68533, CurrentID=0, Status='live', Symbol='Mphosph6', Aliases='1110001M01Rik, AA536809, C86426', description='M phase phosphoprotein 6', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=117791642, end_position_on_the_genomic_accession=117801933, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17984, CurrentID=0, Status='live', Symbol='Ndn', Aliases='AI528698, Peg6', description='necdin', other_designations='', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=62348277, end_position_on_the_genomic_accession=62349928, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17986, CurrentID=0, Status='live', Symbol='Ndp', Aliases='NDh, Ndp', description='Norrie disease (pseudoglioma) (human)', other_designations='norrie disease protein homolog', map_location='X 12.07 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=16885521, end_position_on_the_genomic_accession=16911764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  17988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17988, CurrentID=0, Status='live', Symbol='Ndrg1', Aliases='CAP43, CMT4D, DRG1, HMSNL, NMSL, Ndr1, Ndrl, PROXY1, RTP, TDD5', description='N-myc downstream regulated gene 1', other_designations='N-myc downstream regulated 1|N-myc downstream-regulated gene 1 protein|protein Ndr1', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=66929318, end_position_on_the_genomic_accession=66969641, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  17991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17991, CurrentID=0, Status='live', Symbol='Ndufa2', Aliases='AV000592, B8, C1-B8, CI-B8', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 2', other_designations='NADH-ubiquinone oxidoreductase B8 subunit|complex I-B8', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36742332, end_position_on_the_genomic_accession=36744587, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  17992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17992, CurrentID=0, Status='live', Symbol='Ndufa4', Aliases='MLRQ', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4', other_designations='CI-MLRQ|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 4 (9 kDa, MLRQ)|NADH dehydrogenase [ubiquinone] 1 alpha subcomplex subunit 4|NADH-ubiquinone oxidoreductase MLRQ subunit|complex I-MLRQ', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=11900373, end_position_on_the_genomic_accession=11907446, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17993, CurrentID=0, Status='live', Symbol='Ndufs4', Aliases='6720411N02Rik, C1-18k', description='NADH dehydrogenase (ubiquinone) Fe-S protein 4', other_designations='CI-18 kDa|CI-AQDQ|NADH-ubiquinone oxidoreductase 18 kDa subunit|complex I-18 kDa|complex I-AQDQ', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=114287795, end_position_on_the_genomic_accession=114388094, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  17995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17995, CurrentID=0, Status='live', Symbol='Ndufv1', Aliases='CI-51kD', description='NADH dehydrogenase (ubiquinone) flavoprotein 1', other_designations='NADH dehydrogenase flavoprotein 1; 24 kDa (FP)|NADH-ubiquinone oxidoreductase 51 kDa subunit|complex I-51kD', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4007499, end_position_on_the_genomic_accession=4012755, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17996, CurrentID=0, Status='live', Symbol='Neb', Aliases='AI595938', description='nebulin', other_designations='', map_location='2 29.98 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=52136640, end_position_on_the_genomic_accession=52338798, orientation='minus', exon_count=166, OMIM='', no_hdr0=''),
  17997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17997, CurrentID=0, Status='live', Symbol='Nedd1', Aliases='', description='neural precursor cell expressed, developmentally down-regulated gene 1', other_designations='NEDD-1|neural precursor cell expressed developmentally down-regulated protein 1', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=92684745, end_position_on_the_genomic_accession=92723682, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  17999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17999, CurrentID=0, Status='live', Symbol='Nedd4', Aliases='AA959633, AL023035, AU019897, E430025J12Rik-1, Nedd4a, mKIAA0093, Nedd4', description='neural precursor cell expressed, developmentally down-regulated 4', other_designations='NEDD-4|neural precursor cell expressed developmentally down-regulated protein 4|neural precursor cell expressed, developmentally down-regulated gene 4|neural precursor cell expressed, developmentally down-regulated gene 4a|neural precursor cell expressed, developmentally down-regulted gene 4', map_location='9 40.08 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72662347, end_position_on_the_genomic_accession=72749851, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  18000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18000, CurrentID=0, Status='live', Symbol='Sept2', Aliases='AW208991, Nedd-5, Nedd5, mKIAA0158', description='septin 2', other_designations='neural precursor cell expressed, developmentally down-regulated gene 5', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93478967, end_position_on_the_genomic_accession=93509733, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  18002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18002, CurrentID=0, Status='live', Symbol='Nedd8', Aliases='Rub1', description='neural precursor cell expressed, developmentally down-regulated gene 8', other_designations='NEDD-8|neddylin|neural precursor cell expressed developmentally down-regulated protein 8|ubiquitin-like protein Nedd8', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55662267, end_position_on_the_genomic_accession=55671906, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18003, CurrentID=0, Status='live', Symbol='Nedd9', Aliases='Cas-L, CasL, HEF1, MEF1, Nedd-9, p105', description='neural precursor cell expressed, developmentally down-regulated gene 9', other_designations='CRK-associated substrate-related protein|Crk-associated substrate lymphocyte-specific protein', map_location='13 A3.3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41309916, end_position_on_the_genomic_accession=41487360, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18004, CurrentID=0, Status='live', Symbol='Nek1', Aliases='D8Ertd790e, kat', description='NIMA (never in mitosis gene a)-related expressed kinase 1', other_designations='NIMA-related kinase 1|NY-REN-55 antigen|Nek1 serine/threonine- and tyrosine-specific protein kinase|kidney, anemia and testis|never in mitosis A-related kinase 1|nimA-related protein kinase 1', map_location='8 30.91 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=60993193, end_position_on_the_genomic_accession=61131349, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  18005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18005, CurrentID=0, Status='live', Symbol='Nek2', Aliases='AA617254, C77054', description='NIMA (never in mitosis gene a)-related expressed kinase 2', other_designations='NIMA - related expressed kinase 2|NIMA-related expressed kinase 2|NIMA-related kinase 2|Nek2 kinase|never in mitosis A-related kinase 2|nimA-related protein kinase 2', map_location='1 96.94 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191820875, end_position_on_the_genomic_accession=191833049, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  18007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18007, CurrentID=0, Status='live', Symbol='Neo1', Aliases='2610028H22Rik, AI327052, D930014N22Rik, Igdcc2', description='neogenin', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58874679, end_position_on_the_genomic_accession=59036957, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  18008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18008, CurrentID=0, Status='live', Symbol='Nes', Aliases='AA166324, C78523, ESTM46', description='nestin', other_designations='intermediate filament protein', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87971093, end_position_on_the_genomic_accession=87980451, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18010, CurrentID=0, Status='live', Symbol='Neu1', Aliases='AA407268, AA407316, Aglp, Apl, Bat-7, Bat7, G9, Map-2, Neu, Neu-1', description='neuraminidase 1', other_designations='G9 sialidase|HLA-B-associated transcript 7|N-acetyl-alpha-neuraminidase 1|lysosomal sialidase|sialidase 1', map_location='17 18.48 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34931253, end_position_on_the_genomic_accession=34937297, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18011, CurrentID=0, Status='live', Symbol='Neurl1a', Aliases='2410129E16Rik, AI450910, AI481072, Neu1, Neur1, Neurl, Neurl1, Nlz, Rnf67', description='neuralized E3 ubiquitin protein ligase 1A', other_designations='m-neu1|m-neuralized 1|neuralized homolog 1A|neuralized-like homolog|neuralized-like protein 1|neuralized-like protein 1A|neuralized1', map_location='19 39.39 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47178820, end_position_on_the_genomic_accession=47259441, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18012, CurrentID=0, Status='live', Symbol='Neurod1', Aliases='BETA2, BHF-1, Neurod, bHLHa3', description='neurogenic differentiation 1', other_designations='beta-cell E-box transcriptional activator 2', map_location='2 47.58 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=79452641, end_position_on_the_genomic_accession=79456636, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18013, CurrentID=0, Status='live', Symbol='Neurod2', Aliases='Ndrf, bHLHa1', description='neurogenic differentiation 2', other_designations='neuroD-related factor', map_location='11 61.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98325417, end_position_on_the_genomic_accession=98329645, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18014, CurrentID=0, Status='live', Symbol='Neurog1', Aliases='AKA, Math4C, Neurod3, bHLHa6, ngn1', description='neurogenin 1', other_designations='NGN-1|helix-loop-helix protein mATH-4C|neurogenic basic-helix-loop-helix protein|neurogenic differentiation 3|neurogenic differentiation factor 3', map_location='13 30.06 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56250492, end_position_on_the_genomic_accession=56252163, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18015, CurrentID=0, Status='live', Symbol='Nf1', Aliases='AW494271, Dsk9, E030030H24Rik, Mhdadsk9, Nf-1', description='neurofibromatosis 1', other_designations='neurofibromatosis-related protein NF-1', map_location='11 B4-5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79335988, end_position_on_the_genomic_accession=79581609, orientation='plus', exon_count=59, OMIM='', no_hdr0=''),
  18016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18016, CurrentID=0, Status='live', Symbol='Nf2', Aliases='merlin', description='neurofibromatosis 2', other_designations='moesin-ezrin-radixin-like protein|neurofibromin-2|schwannomin', map_location='11 3.0 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4757650, end_position_on_the_genomic_accession=4849544, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  18018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18018, CurrentID=0, Status='live', Symbol='Nfatc1', Aliases='2210017P03Rik, AI449492, AV076380, NF-ATc, NFAT2, NFATc, Nfatcb', description='nuclear factor of activated T cells, cytoplasmic, calcineurin dependent 1', other_designations='NF-ATc1|NFAT transcription complex cytosolic component|nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 1|transcription factor NF-ATc', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80606205, end_position_on_the_genomic_accession=80713071, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  723899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723899, CurrentID=0, Status='live', Symbol='Mir365-1', Aliases='Mirn365-1, mir-365-1', description='microRNA 365-1', other_designations='mir 365-1|mmu-mir-365-1', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13453840, end_position_on_the_genomic_accession=13453926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18020, CurrentID=0, Status='live', Symbol='Nfatc2ip', Aliases='D7Ertd304e, NIP45', description='nuclear factor of activated T cells, cytoplasmic, calcineurin dependent 2 interacting protein', other_designations='45 kDa NF-AT-interacting protein|45 kDa NFAT-interacting protein|nuclear factor of activated T-cells, cytoplasmic 2 interacting protein|nuclear factor of activated T-cells, cytoplasmic 2-interacting protein|nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 2 interacting protein', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126382853, end_position_on_the_genomic_accession=126397252, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18021, CurrentID=0, Status='live', Symbol='Nfatc3', Aliases='C80703, D8Ertd281e, NFAT4, NFATx', description='nuclear factor of activated T cells, cytoplasmic, calcineurin dependent 3', other_designations='NF-AT4|NF-ATc3|T-cell transcription factor NFAT4|nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 3', map_location='8 D', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106058825, end_position_on_the_genomic_accession=106130537, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18022, CurrentID=0, Status='live', Symbol='Nfe2', Aliases='NF-E2, NF-E2/P45, p45, p45NFE2, p45nf-e2', description='nuclear factor, erythroid derived 2', other_designations='leucine zipper protein NF-E2|nuclear factor, erythroid derived 2, 45 kDa|nuclear factor, erythroid-derived 2 45 kDa subunit|p45 NF-E2', map_location='15 58.62 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103248212, end_position_on_the_genomic_accession=103258403, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18023, CurrentID=0, Status='live', Symbol='Nfe2l1', Aliases='AA408798, AW212678, LCR-F1, Lcrf1, NRF1, TCF-11, TCF11', description='nuclear factor, erythroid derived 2,-like 1', other_designations='NF-E2-related factor 1|NFE2-related factor 1', map_location='11 D-E', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96817413, end_position_on_the_genomic_accession=96829968, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18024, CurrentID=0, Status='live', Symbol='Nfe2l2', Aliases='Nrf2', description='nuclear factor, erythroid derived 2, like 2', other_designations='NF-E2-related factor 2|NFE2-related factor 2', map_location='2 44.75 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=75675519, end_position_on_the_genomic_accession=75704641, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18025, CurrentID=0, Status='live', Symbol='Nfe2l3', Aliases='Nrf3', description='nuclear factor, erythroid derived 2, like 3', other_designations='NF-E2-related factor 3|NFE2-related factor 3|nuclear factor erythroid 2-like 3', map_location='6 24.84 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=51432670, end_position_on_the_genomic_accession=51458768, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18027, CurrentID=0, Status='live', Symbol='Nfia', Aliases='1110047K16Rik, 9430022M17Rik, CTF, NF1-A, NF1A', description='nuclear factor I/A', other_designations='CCAAT-box-binding transcription factor|TGGCA-binding protein|nuclear factor 1/A', map_location='4 C4-C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=97581717, end_position_on_the_genomic_accession=98118876, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18028, CurrentID=0, Status='live', Symbol='Nfib', Aliases='6720429L07Rik, CTF, E030026I10Rik, NF-I/B, NF1-B, NFI-B', description='nuclear factor I/B', other_designations='CCAAT-box-binding transcription factor|TGGCA-binding protein|nuclear factor 1/B', map_location='4 C4-C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=82290173, end_position_on_the_genomic_accession=82505779, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18029, CurrentID=0, Status='live', Symbol='Nfic', Aliases='1110019L22Rik, 1500041O16Rik, AA589446, AI746521, NF1-C', description='nuclear factor I/C', other_designations='CCAAT-box-binding transcription factor|CTF|NF-I/C|NFI-C|TGGCA-binding protein|nuclear factor 1-C2|nuclear factor 1/C', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81396191, end_position_on_the_genomic_accession=81431167, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18030, CurrentID=0, Status='live', Symbol='Nfil3', Aliases='AV225605, E4BP4', description='nuclear factor, interleukin 3, regulated', other_designations='E4 promoter binding-protein 4|E4 promoter-binding protein 4|embryo implantation-related NFIL3/E4BP4-like transcription factor', map_location='13 27.68 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=52967209, end_position_on_the_genomic_accession=52982790, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18032, CurrentID=0, Status='live', Symbol='Nfix', Aliases='CTF, NF-I/X, NF1-X, NFI-X', description='nuclear factor I/X', other_designations='CCAAT-box-binding transcription factor|TGGCA-binding protein|nuclear factor 1/X', map_location='8 C1-C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84704712, end_position_on_the_genomic_accession=84800338, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18033, CurrentID=0, Status='live', Symbol='Nfkb1', Aliases='NF-KB1, NF-kappaB, NF-kappaB1, p105, p50, p50/p105', description='nuclear factor of kappa light polypeptide gene enhancer in B cells 1, p105', other_designations='DNA-binding factor KBF1|EBP-1|NF kappaB1|NF-kappa-B1 p84/NF-kappa-B1 p98|NF-kappaB p50|nuclear factor kappaB p50|nuclear factor of kappa light chain gene enhancer in B-cells 1, p105|nuclear factor of kappa light polypeptide gene enhancer in B-cells 1, p105|nuclear factor-kappaB p50|p50 subunit of NF kappaB|p50 subunit of NF-kappaB', map_location='3 62.82 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135584655, end_position_on_the_genomic_accession=135691547, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  18034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18034, CurrentID=0, Status='live', Symbol='Nfkb2', Aliases='NF-kappaB2, lyt, p49, p49/p100, p50B, p52', description='nuclear factor of kappa light polypeptide gene enhancer in B cells 2, p49/p100', other_designations='DNA-binding factor KBF2|NF kappaB2|nuclear factor of kappa light polypeptide gene enhancer in B-cells 2, p49/p100', map_location='19 C3-D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46303612, end_position_on_the_genomic_accession=46312400, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  18035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18035, CurrentID=0, Status='live', Symbol='Nfkbia', Aliases='AI462015, Nfkbi', description='nuclear factor of kappa light polypeptide gene enhancer in B cells inhibitor, alpha', other_designations='I(Kappa)B(alpha)|I-kappa-B-alpha|I-kappaBalpha|ikB-alpha|ikappaBalpha|nuclear factor of kappa light chain gene enhancer in B-cells inhibitor, alpha|nuclear factor of kappa light polyp gene enhancer in B-cell 1|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, alpha', map_location='12 C1-C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55489409, end_position_on_the_genomic_accession=55492647, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18036, CurrentID=0, Status='live', Symbol='Nfkbib', Aliases='I(Kappa)B(beta), I-kappa-B-beta, IKB-beta, IKappaBbeta, IkB, IkBb, NF-kappa-BIB, ikB-B', description='nuclear factor of kappa light polypeptide gene enhancer in B cells inhibitor, beta', other_designations='nuclear factor of kappa light chain gene enhancer in B-cells inhibitor, beta|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, beta', map_location='7 A3-B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28758251, end_position_on_the_genomic_accession=28766639, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  18037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18037, CurrentID=0, Status='live', Symbol='Nfkbie', Aliases='I-kappa-B-epsilon, IKBE, IkB-E, NF-kappa-BIE, ikB-epsilon', description='nuclear factor of kappa light polypeptide gene enhancer in B cells inhibitor, epsilon', other_designations='ikappaBepsilon|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, epsilon', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45555667, end_position_on_the_genomic_accession=45563169, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18038, CurrentID=0, Status='live', Symbol='Nfkbil1', Aliases='Def-7, IKBL', description='nuclear factor of kappa light polypeptide gene enhancer in B cells inhibitor like 1', other_designations='I-kappa-B-like protein|ikappaBL|inhibitor of kappa B-like protein|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor-like 1', map_location='17 18.6 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35220175, end_position_on_the_genomic_accession=35236452, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18039, CurrentID=0, Status='live', Symbol='Nefl', Aliases='AI847934, CMT2E, NF-L, NF68, Nfl', description='neurofilament, light polypeptide', other_designations='68 kDa neurofilament protein|neurofilament protein L|neurofilament triplet L protein', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=68083884, end_position_on_the_genomic_accession=68087745, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18040, CurrentID=0, Status='live', Symbol='Nefm', Aliases='NF-M, NF160, NF165, Nef3, Nfm', description='neurofilament, medium polypeptide', other_designations='160 kDa neurofilament protein|neurofilament 3, medium|neurofilament protein M|neurofilament triplet M protein|neurofilament-M', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=68119545, end_position_on_the_genomic_accession=68125004, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18041, CurrentID=0, Status='live', Symbol='Nfs1', Aliases='AA987187, m-Nfs1, m-Nfsl', description='nitrogen fixation gene 1 (S. cerevisiae)', other_designations='nifS-like (sic)|nitrogen fixation gene, yeast homolog 1', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156123637, end_position_on_the_genomic_accession=156144186, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18044, CurrentID=0, Status='live', Symbol='Nfya', Aliases='AA407810, Cbf-b, SEZ-10, SEZ10', description='nuclear transcription factor-Y alpha', other_designations='CAAT box DNA-binding protein subunit A|CAAT-box DNA-binding protein subunit A|NF-YA|nuclear transcription factor Y subunit A', map_location='17 C-D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48386885, end_position_on_the_genomic_accession=48409901, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18045, CurrentID=0, Status='live', Symbol='Nfyb', Aliases='AA985999, Cbf-A', description='nuclear transcription factor-Y beta', other_designations='CAAT box DNA-binding protein subunit B|CAAT-box DNA-binding protein subunit B|NF-YB|nuclear transcription factor Y subunit B', map_location='10 C3-D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82748700, end_position_on_the_genomic_accession=82764148, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18046, CurrentID=0, Status='live', Symbol='Nfyc', Aliases='', description='nuclear transcription factor-Y gamma', other_designations='CAAT box DNA-binding protein subunit C|nuclear transcription factor Y subunit C', map_location='4 56.52 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120757435, end_position_on_the_genomic_accession=120831579, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18048, CurrentID=0, Status='live', Symbol='Klk1b4', Aliases='Ngfa, mGk-4', description='kallikrein 1-related pepidase b4', other_designations='7S nerve growth factor alpha chain|7S nerve growth factor alpha subunit|alpha-NGF|nerve growth factor, alpha', map_location='7 28.72 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44207435, end_position_on_the_genomic_accession=44211754, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  723904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723904, CurrentID=0, Status='live', Symbol='Mir219a-2', Aliases='Mir219-2, Mirn219-2, mir-219a-2, mmu-mir-219a-2', description='microRNA 219a-2', other_designations='microRNA 219-2|mir 219-2|mmu-mir-219-2', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29845631, end_position_on_the_genomic_accession=29845727, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18050, CurrentID=0, Status='live', Symbol='Klk1b3', Aliases='Ngfg, gamma-NGF, mGk-3', description='kallikrein 1-related peptidase b3', other_designations='7S nerve growth factor gamma chain|7S nerve growth factor gamma subunit|glandular kallikrein K3|tissue kallikrein-3', map_location='7 28.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44198191, end_position_on_the_genomic_accession=44202351, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18053, CurrentID=0, Status='live', Symbol='Ngfr', Aliases='LNGFR, Tnfrsf16, p75, p75NGFR, p75NTR', description='nerve growth factor receptor (TNFR superfamily, member 16)', other_designations='NGF receptor|low affinity neurotrophin receptor p75NTR|low-affinity nerve growth factor receptor|p75 neurotrophin receptor', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95568818, end_position_on_the_genomic_accession=95587698, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18054, CurrentID=0, Status='live', Symbol='Ngp', Aliases='bectenecin', description='neutrophilic granule protein', other_designations='clone B6|cystatin-like protein|myeloid bactenecin protein|myeloid granule protein|myeloid secondary granule protein', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110419808, end_position_on_the_genomic_accession=110423012, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68545, CurrentID=0, Status='live', Symbol='Ecscr', Aliases='1110006O17Rik, ARIA', description='endothelial cell surface expressed chemotaxis and apoptosis regulator', other_designations='apoptosis regulator through modulating IAP expression|endothelial cell-specific molecule 2', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35713086, end_position_on_the_genomic_accession=35722725, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18071, CurrentID=0, Status='live', Symbol='Nhlh1', Aliases='Hen1, Nscl, Tal2, bHLHa35', description='nescient helix loop helix 1', other_designations='HELIX-LOOP-HELIX PROTEIN 1 (HEN1) (NSCL)|HEN-1|NSCL-1', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172052292, end_position_on_the_genomic_accession=172057596, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18072, CurrentID=0, Status='live', Symbol='Nhlh2', Aliases='6230401I09Rik, Hen2, NSCL2, Nscl-2, bHLHa34', description='nescient helix loop helix 2', other_designations='HEN-2', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102010110, end_position_on_the_genomic_accession=102015492, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18073, CurrentID=0, Status='live', Symbol='Nid1', Aliases='A630025O17, Nid, entactin, entactin-1, nidogen-1', description='nidogen 1', other_designations='NID-1|Nidogen precursor (Entactin)|entactin 1', map_location='13 5.26 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=13437602, end_position_on_the_genomic_accession=13512275, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  18074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18074, CurrentID=0, Status='live', Symbol='Nid2', Aliases='AW547149, Ly111, NID-2, entactin-2, nidogen-2', description='nidogen 2', other_designations='entactin 2', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=19751257, end_position_on_the_genomic_accession=19811787, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  723909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723909, CurrentID=0, Status='live', Symbol='Mir342', Aliases='Mirn342, mmu-mir-342', description='microRNA 342', other_designations='mir 342', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108658620, end_position_on_the_genomic_accession=108658718, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18080, CurrentID=0, Status='live', Symbol='Nin', Aliases='3110068G20Rik, AI385615, AU024711, mKIAA1565', description='ninein', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70011435, end_position_on_the_genomic_accession=70113830, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  18081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18081, CurrentID=0, Status='live', Symbol='Ninj1', Aliases='AU024536', description='ninjurin 1', other_designations='nerve injury-induced protein 1', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49187547, end_position_on_the_genomic_accession=49196251, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18082, CurrentID=0, Status='live', Symbol='Nipsnap1', Aliases='', description='4-nitrophenylphosphatase domain and non-neuronal SNAP25-like protein homolog 1 (C. elegans)', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4873965, end_position_on_the_genomic_accession=4894202, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68550, CurrentID=0, Status='live', Symbol='Tefm', Aliases='1110002N22Rik', description='transcription elongation factor, mitochondrial', other_designations='', map_location='11 47.53 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80136674, end_position_on_the_genomic_accession=80142720, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18087, CurrentID=0, Status='live', Symbol='Nktr', Aliases='5330401F18Rik, D9Wsu172e', description='natural killer tumor recognition sequence', other_designations='NK-TR protein|cyclophilin-related protein|natural killer tumor recognition protein|natural-killer cells cyclophilin-related protein', map_location='9 72.57 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121718181, end_position_on_the_genomic_accession=121756842, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  18088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18088, CurrentID=0, Status='live', Symbol='Nkx2-2', Aliases='Nkx-2.2, Nkx2.2, tinman', description='NK2 homeobox 2', other_designations='Drosophila NK2 transcription factor related, locus 2|NK2 transcription factor related, locus 2|glycoprotein GP330, renal|homeobox protein NK-2 homolog B|homeobox protein Nkx2-2', map_location='2 H', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=147177546, end_position_on_the_genomic_accession=147195783, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18089, CurrentID=0, Status='live', Symbol='Nkx2-3', Aliases='Nkx-2.3, Nkx2.3, tinman', description='NK2 homeobox 3', other_designations='Drosophila NK2 transcription factor related, locus 3|NK2 transcription factor related, locus 3|homeobox protein NK-2 homolog 3|homeobox protein NK-2 homolog C|nkx2-C', map_location='19 36.67 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43612325, end_position_on_the_genomic_accession=43615892, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  723911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723911, CurrentID=0, Status='live', Symbol='Mir367', Aliases='Mirn367, mir-367, mmu-mir-367', description='microRNA 367', other_designations='mir 367', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127545733, end_position_on_the_genomic_accession=127545807, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18092, CurrentID=0, Status='live', Symbol='Nkx2-6', Aliases='AV044928, Nkx-2.6, Nkx2.6, Tix, tinman', description='NK2 homeobox 6', other_designations='Drosophila NK2 transcription factor related, locus 6|NK2 transcription factor related, locus 6|homeobox protein NK-2 homolog F', map_location='14 36.01 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69171802, end_position_on_the_genomic_accession=69175540, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18094, CurrentID=0, Status='live', Symbol='Nkx2-9', Aliases='Nkx-2.9, Nkx2-8, Nkx2.9, tinman', description='NK2 homeobox 9', other_designations='Drosophila NK2 transcription factor related, locus 9|NK2 transcription factor related, locus 9|homeobox protein NK-2 homolog H|homeobox protein Nkx-2.9', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=56611389, end_position_on_the_genomic_accession=56614676, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18095, CurrentID=0, Status='live', Symbol='Nkx3-1', Aliases='Bax, NKX3.1, NKX3A, Nkx-3.1, bagpipe', description='NK-3 transcription factor, locus 1 (Drosophila)', other_designations='Drosophila NK-3 transcription factor, locus 1|homeobox protein NK-3 homolog A', map_location='14 36.02 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69190692, end_position_on_the_genomic_accession=69194658, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18096, CurrentID=0, Status='live', Symbol='Nkx6-1', Aliases='NKX6A, Nkx6.1', description='NK6 homeobox 1', other_designations='Drosophila NK transcription factor related, gene family 6, locus 1|NK6 transcription factor related, locus 1|homeobox protein NK-6 homolog A|homeodomain transcription factor Nkx6.1', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=101659184, end_position_on_the_genomic_accession=101664711, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68552, CurrentID=0, Status='live', Symbol='Smim14', Aliases='1110003E01Rik, 1700127H04Rik, 5430439C17Rik, MAd4', description='small integral membrane protein 14', other_designations='', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65446844, end_position_on_the_genomic_accession=65537179, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18099, CurrentID=0, Status='live', Symbol='Nlk', Aliases='AI194375', description='nemo like kinase', other_designations='nemo-like kinase|nlk|serine/threonine protein kinase NLK', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78567168, end_position_on_the_genomic_accession=78697595, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18100, CurrentID=0, Status='live', Symbol='Mrpl40', Aliases='Nlvcf', description='mitochondrial ribosomal protein L40', other_designations='L40mt|MRP-L40|nuclear localization signal protein absent in velo-cardio-facial patients|nuclear localization signal-containing protein deleted in velocardiofacial syndrome homolog', map_location='16 11.69 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18872018, end_position_on_the_genomic_accession=18876637, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18101, CurrentID=0, Status='live', Symbol='Nmbr', Aliases='BB182387', description='neuromedin B receptor', other_designations='NMB-R|neuromedin-B-preferring bombesin receptor', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=14753378, end_position_on_the_genomic_accession=14771509, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18102, CurrentID=0, Status='live', Symbol='Nme1', Aliases='AL024257, NDPK-A, NM23-M1, NM23A', description='NME/NM23 nucleoside diphosphate kinase 1', other_designations='NDK A|NDP kinase A|expressed in non-metastatic cells 1 protein|expressed in non-metastatic cells 1, protein|metastasis inhibition factor NM23|non-metastatic cells 1, protein (NM23A) expressed in|nucleoside-diphosphate kinase 1|nucleotide diphosphate kinase|tumor metastatic process-associated protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=93958925, end_position_on_the_genomic_accession=93968521, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  723913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723913, CurrentID=0, Status='live', Symbol='Mir412', Aliases='Mirn412, mir-412, mmu-mir-412', description='microRNA 412', other_designations='mir 412', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109743289, end_position_on_the_genomic_accession=109743368, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18104, CurrentID=0, Status='live', Symbol='Nqo1', Aliases='AV001255, Dia4, Dtd, Nmo-1, Nmo1, Nmor1, Ox-1, Ox1, Qr1', description='NAD(P)H dehydrogenase, quinone 1', other_designations='DT-diaphorase|NAD(P)H dehydrogenase (quinone)|NAD(P)H menadione oxidoreductase 1, dioxin inducible|NAD(P)H:quinone oxidoreductase 1|azoreductase|diaphorase 4 (NADH/NADPH)|menadione reductase|phylloquinone reductase|quinone reductase 1', map_location='8 54.08 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107388225, end_position_on_the_genomic_accession=107403205, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18105, CurrentID=0, Status='live', Symbol='Nqo2', Aliases='NMO2, Nmor2, Ox2', description='NAD(P)H dehydrogenase, quinone 2', other_designations='NAD(P)H menadione oxidoreductase 2, dioxin inducible|NRH dehydrogenase [quinone] 2|NRH: quinone oxidoreductase|NRH:quinone oxidoreductase 2|QR2|quinone reductase 2|ribosyldihydronicotinamide dehydrogenase', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33964655, end_position_on_the_genomic_accession=33988465, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18106, CurrentID=0, Status='live', Symbol='Cd244', Aliases='2B4, C9.1, F730046O15Rik, Ly90, NAIL, NKR2B4, Nmrk, SLAMF4', description='CD244 natural killer cell receptor 2B4', other_designations='CD244 long isoform|CD244 short isoform|NK cell receptor 2B4|NK cell type I receptor protein 2B4|SLAM family member 4|non MHC restricted killing associated|non-MHC restricted killing associated|signaling lymphocytic activation molecule 4', map_location='1 79.52 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171558743, end_position_on_the_genomic_accession=171585316, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18107, CurrentID=0, Status='live', Symbol='Nmt1', Aliases='AW536594', description='N-myristoyltransferase 1', other_designations='NMT 1|myristoyl-CoA:protein N-myristoyltransferase 1|peptide N-myristoyltransferase 1|type I N-myristoyltransferase', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103028344, end_position_on_the_genomic_accession=103068912, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18108, CurrentID=0, Status='live', Symbol='Nmt2', Aliases='A930001K02Rik, AI605445, AU044698, hNMT-2', description='N-myristoyltransferase 2', other_designations='myristoyl-CoA:protein N-myristoyltransferase 2|peptide N-myristoyltransferase 2|type II N-myristoyltransferase', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3284212, end_position_on_the_genomic_accession=3328877, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  18109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18109, CurrentID=0, Status='live', Symbol='Mycn', Aliases='N-myc, Nmyc, Nmyc-1, Nmyc1, bHLHe37, c-nmyc', description='v-myc myelocytomatosis viral related oncogene, neuroblastoma derived (avian)', other_designations='neuroblastoma myc-related oncogene 1', map_location='12 A3-B', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=12936093, end_position_on_the_genomic_accession=12941836, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18111, CurrentID=0, Status='live', Symbol='Nnat', Aliases='5730414I02Rik, AW107673, Peg5', description='neuronatin', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157560078, end_position_on_the_genomic_accession=157562525, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18113, CurrentID=0, Status='live', Symbol='Nnmt', Aliases='', description='nicotinamide N-methyltransferase', other_designations='', map_location='9 26.45 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48538001, end_position_on_the_genomic_accession=48607635, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18114, CurrentID=0, Status='live', Symbol='Rrp1', Aliases='AL033305, AL033306, AU018604, NNP-1, Nnp1', description='ribosomal RNA processing 1 homolog (S. cerevisiae)', other_designations='RRP1-like protein|novel nuclear protein 1|nucleolar protein Nop52|protein NNP-1', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78400362, end_position_on_the_genomic_accession=78413043, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18115, CurrentID=0, Status='live', Symbol='Nnt', Aliases='4930423F13Rik, AI323702, BB168308', description='nicotinamide nucleotide transhydrogenase', other_designations='pyridine nucleotide transhydrogenase', map_location='13 D2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119335527, end_position_on_the_genomic_accession=119409138, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  18117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18117, CurrentID=0, Status='live', Symbol='Emc8', Aliases='Cox4nb, Fam158b, Noc4', description='ER membrane protein complex subunit 8', other_designations='COX4 neighbor|neighbor of Cox4', map_location='8 69.98 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120653914, end_position_on_the_genomic_accession=120668112, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18119, CurrentID=0, Status='live', Symbol='Nodal', Aliases='Tg.413d', description='nodal', other_designations='early embryo mesoderm formation|transgenic lethal 413.d', map_location='10 32.21 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61417972, end_position_on_the_genomic_accession=61425337, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18120, CurrentID=0, Status='live', Symbol='Mrpl49', Aliases='4833419D18Rik, Nof1', description='mitochondrial ribosomal protein L49', other_designations='L49mt|MRP-L49|neighbor of fau 1', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6053630, end_position_on_the_genomic_accession=6057751, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  723916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723916, CurrentID=0, Status='live', Symbol='Mir484', Aliases='Mirn484, mir-484, mmu-mir-484', description='microRNA 484', other_designations='mir 484', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14159626, end_position_on_the_genomic_accession=14159692, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18124, CurrentID=0, Status='live', Symbol='Nr4a3', Aliases='AI573420, CHN, CSMF, MINOR, NOR-1, Nor1, TEC', description='nuclear receptor subfamily 4, group A, member 3', other_designations='neural orphan receptor 1|orphan nuclear receptor TEC|translocated in extraskeletal chondrosarcoma', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48044637, end_position_on_the_genomic_accession=48086447, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18125, CurrentID=0, Status='live', Symbol='Nos1', Aliases='2310005C01Rik, N-NOS, NC-NOS, NO, NOS, NOS-I, Nos-1, bNOS, nNOS', description='nitric oxide synthase 1, neuronal', other_designations='NOS type I|constitutive NOS|neuronal NOS|neuronal nitric oxide synthase|peptidyl-cysteine S-nitrosylase NOS1', map_location='5 57.29 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117780997, end_position_on_the_genomic_accession=117958840, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  18126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18126, CurrentID=0, Status='live', Symbol='Nos2', Aliases='MAC-NOS, NOS-II, Nos-2a, i-NOS, iNOS, Nos2', description='nitric oxide synthase 2, inducible', other_designations='NOS type II|inducible NO synthase|inducible NOS|inducible nitric oxide synthase|macrophage NOS|nitric oxide synthase 2, inducible, macrophage|peptidyl-cysteine S-nitrosylase NOS2', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78920859, end_position_on_the_genomic_accession=78960232, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  18127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18127, CurrentID=0, Status='live', Symbol='Nos3', Aliases='2310065A03Rik, Nos-3, eNOS, ecNOS', description='nitric oxide synthase 3, endothelial cell', other_designations='EC-NOS|NOS type III|NOSIII|cNOS|constitutive NOS|endothelial NOS|endothelial nitric oxide synthase', map_location='5 11.32 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24364819, end_position_on_the_genomic_accession=24384474, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  18128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18128, CurrentID=0, Status='live', Symbol='Notch1', Aliases='9930111A19Rik, Mis6, N1, Tan1, lin-12', description='notch 1', other_designations='Motch A|Notch gene homolog 1|mT14|major type A protein|p300|transmembrane receptor Notch1', map_location='2 18.91 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26457902, end_position_on_the_genomic_accession=26503822, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  18129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18129, CurrentID=0, Status='live', Symbol='Notch2', Aliases='AI853703, N2', description='notch 2', other_designations='Motch B|Notch gene homolog 2', map_location='3 42.42 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98013538, end_position_on_the_genomic_accession=98150367, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  18130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18130, CurrentID=0, Status='live', Symbol='Ints6', Aliases='2900075H24Rik, AI480962, DICE1, Ddx26, HDB, Notch2l', description='integrator complex subunit 6', other_designations='DBI-1|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 26|int6', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62676325, end_position_on_the_genomic_accession=62761370, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  18131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18131, CurrentID=0, Status='live', Symbol='Notch3', Aliases='AW229011, N3, hpbk', description='notch 3', other_designations='Notch gene homolog 3', map_location='17 17.37 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32120893, end_position_on_the_genomic_accession=32166852, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  18132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18132, CurrentID=0, Status='live', Symbol='Notch4', Aliases='Int-3, Int3, N4', description='notch 4', other_designations='Notch gene homolog 4', map_location='17 18.15 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34564282, end_position_on_the_genomic_accession=34588543, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  18133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18133, CurrentID=0, Status='live', Symbol='Nov', Aliases='C130088N23Rik, CCN3', description='nephroblastoma overexpressed gene', other_designations='CCN family member 3|nephroblastoma overexpressed gene protein homolog|nephroblastoma-overexpressed gene protein homolog|novH', map_location='15 21.49 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=54745613, end_position_on_the_genomic_accession=54753767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18139, CurrentID=0, Status='live', Symbol='Zfp638', Aliases='AI323587, Np220, Zfml, Znf638, mNP220', description='zinc finger protein 638', other_designations='nuclear protein 220|zinc finger matrin-like protein|zinc finger, matrin-like', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83867106, end_position_on_the_genomic_accession=83987119, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  18140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18140, CurrentID=0, Status='live', Symbol='Uhrf1', Aliases='AL022808, ICBP90, Np95, RNF106', description='ubiquitin-like, containing PHD and RING finger domains, 1', other_designations='mUhrf1|nuclear protein 95|nuclear zinc finger protein Np95|ubiquitin-like PHD and RING finger domain-containing protein 1|ubiquitin-like-containing PHD and RING finger domains protein 1', map_location='17 D-E1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56303337, end_position_on_the_genomic_accession=56323486, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18141, CurrentID=0, Status='live', Symbol='Nup50', Aliases='1700030K07Rik, AI413123, Npap60', description='nucleoporin 50', other_designations='50 kDa nucleoporin|nuclear pore associated protein 60 kDa|nucleoporin Nup50|nucleoprotein 50', map_location='15 40.25 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84923428, end_position_on_the_genomic_accession=84942963, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  18142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18142, CurrentID=0, Status='live', Symbol='Npas1', Aliases='MOP5, bHLHe11', description='neuronal PAS domain protein 1', other_designations='neuronal PAS1', map_location='7 9.14 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16455718, end_position_on_the_genomic_accession=16476823, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18143, CurrentID=0, Status='live', Symbol='Npas2', Aliases='MOP4, bHLHe9', description='neuronal PAS domain protein 2', other_designations='neuronal PAS2', map_location='1 17.98 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39193956, end_position_on_the_genomic_accession=39363240, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  18145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18145, CurrentID=0, Status='live', Symbol='Npc1', Aliases='A430089E03Rik, C85354, D18Ertd139e, D18Ertd723e, lcsd, nmf164, spm', description='Niemann-Pick type C1', other_designations='Niemann Pick type C1|sphingomyelinosis', map_location='18 6.15 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12189694, end_position_on_the_genomic_accession=12236386, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  18146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18146, CurrentID=0, Status='live', Symbol='Npdc1', Aliases='AI314472, NPDC-1', description='neural proliferation, differentiation and control 1', other_designations='neural proliferation differentiation control-1 protein|neural proliferation, differentiation and control gene 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25403050, end_position_on_the_genomic_accession=25409494, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18148, CurrentID=0, Status='live', Symbol='Npm1', Aliases='B23, NO38, Npm', description='nucleophosmin 1', other_designations='nucleolar phosphoprotein B23|nucleolar protein NO38|numatrin', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=33152498, end_position_on_the_genomic_accession=33163206, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18150, CurrentID=0, Status='live', Symbol='Npm3', Aliases='Nub1', description='nucleoplasmin 3', other_designations='', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45747734, end_position_on_the_genomic_accession=45749563, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18155, CurrentID=0, Status='live', Symbol='Pnoc', Aliases='N/OFQ, Npnc1, OFQ/N', description='prepronociceptin', other_designations='neuropeptide nociceptin 1|orphanin FQ|proorphanin', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65400673, end_position_on_the_genomic_accession=65425581, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18158, CurrentID=0, Status='live', Symbol='Nppb', Aliases='AA408272, BNF, BNP', description='natriuretic peptide type B', other_designations='brain natriuretic peptide|gamma-brain natriuretic peptide|natriuretic peptide precursor type B|type-B natriuretic factor', map_location='4 78.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147985778, end_position_on_the_genomic_accession=147987205, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18159, CurrentID=0, Status='live', Symbol='Nppc', Aliases='CNP, lbab', description='natriuretic peptide type C', other_designations='natriuretic peptide precursor C|natriuretic peptide precursor type C', map_location='1 43.98 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86666293, end_position_on_the_genomic_accession=86670573, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18160, CurrentID=0, Status='live', Symbol='Npr1', Aliases='AI893888, GC-A, NPR-A, NPRA, Pndr', description='natriuretic peptide receptor 1', other_designations='ANP-A|ANPR-A|Nppa receptor|atrial natriuretic peptide receptor type A|guanylate cyclase A|guanylyl cyclase-A', map_location='3 39.23 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90450592, end_position_on_the_genomic_accession=90465928, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  18162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18162, CurrentID=0, Status='live', Symbol='Npr3', Aliases='ANP-C, ANPR-C, EF-2, NPR-C, lgj, stri', description='natriuretic peptide receptor 3', other_designations='Nppc receptor|atrial natriuretic peptide clearance receptor|atrial natriuretic peptide receptor type C|longjohn|strigosus', map_location='15 5.83 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=11839896, end_position_on_the_genomic_accession=11907371, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18163, CurrentID=0, Status='live', Symbol='Ctnnd2', Aliases='Catnd2, Nprap, neurojugin', description='catenin (cadherin associated protein), delta 2', other_designations='catenin (cadherin-associated protein), delta 2 (neural plakophilin-related arm-repeat protein)|catenin delta 2|neural plakophilin-related arm-repeat protein|neurojungin', map_location='15 B2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=30172593, end_position_on_the_genomic_accession=31029344, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  18164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18164, CurrentID=0, Status='live', Symbol='Nptx1', Aliases='D11Bwg1004e, Np1', description='neuronal pentraxin 1', other_designations='NP-I|neuronal pentraxin I', map_location='11 E1-E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119538719, end_position_on_the_genomic_accession=119547820, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18166, CurrentID=0, Status='live', Symbol='Npy1r', Aliases='Npyr, Y1-R', description='neuropeptide Y receptor Y1', other_designations='NPY-1 receptor|NPY1-R|neuropeptide Y receptor subtype', map_location='8 B3-C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=66697192, end_position_on_the_genomic_accession=66706798, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18167, CurrentID=0, Status='live', Symbol='Npy2r', Aliases='NPY2-R', description='neuropeptide Y receptor Y2', other_designations='NPY-Y2 receptor|Y2 receptor', map_location='3 36.49 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=82538383, end_position_on_the_genomic_accession=82548989, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18168, CurrentID=0, Status='live', Symbol='Npy5r', Aliases='NPY5-R, NPYY5-R, Y5R', description='neuropeptide Y receptor Y5', other_designations='NPY-Y5 receptor|Y5 receptor', map_location='8 B3-C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=66679965, end_position_on_the_genomic_accession=66688581, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18169, CurrentID=0, Status='live', Symbol='Npy6r', Aliases='', description='neuropeptide Y receptor Y6', other_designations='NPY6-R|PP2|pancreatic polypeptide receptor 2', map_location='18 23.74 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44270127, end_position_on_the_genomic_accession=44277700, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18170, CurrentID=0, Status='live', Symbol='nr', Aliases='', description='nervous', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18171, CurrentID=0, Status='live', Symbol='Nr1i2', Aliases='PXR, PXR.1, PXR.2, PXR1, SXR, mPXR', description='nuclear receptor subfamily 1, group I, member 2', other_designations='Pregnane X receptor|orphan nuclear receptor PXR', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38248321, end_position_on_the_genomic_accession=38294851, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18173, CurrentID=0, Status='live', Symbol='Slc11a1', Aliases='Bcg, Ity, Lsh, Nramp, Nramp1, ity', description='solute carrier family 11 (proton-coupled divalent metal ion transporters), member 1', other_designations='NRAMP 1|host resistance locus Bcg/Ity/Lsh|solute carrier family 11 member 1', map_location='1 38.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74375014, end_position_on_the_genomic_accession=74386057, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  18174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18174, CurrentID=0, Status='live', Symbol='Slc11a2', Aliases='DCT1, DMT1, Nramp2, mk, van', description='solute carrier family 11 (proton-coupled divalent metal ion transporters), member 2', other_designations='DMT-1|NRAMP 2|divalent cation transporter 1|divalent metal transporter 1|microcytic anemia, viable anaemia|solute carrier family 11 member 2', map_location='15 56.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100387898, end_position_on_the_genomic_accession=100424209, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  18175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18175, CurrentID=0, Status='live', Symbol='Nrap', Aliases='', description='nebulin-related anchoring protein', other_designations='N-RAP|nrap', map_location='19 51.8 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56319954, end_position_on_the_genomic_accession=56390061, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  18176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18176, CurrentID=0, Status='live', Symbol='Nras', Aliases='AV095280, N-ras', description='neuroblastoma ras oncogene', other_designations='transforming protein N-Ras', map_location='3 45.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103058285, end_position_on_the_genomic_accession=103067914, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  243328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243328, CurrentID=0, Status='live', Symbol='Slc29a4', Aliases='ENT4, mPMAT', description='solute carrier family 29 (nucleoside transporters), member 4', other_designations='solute carrier family 29 member 4', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142692476, end_position_on_the_genomic_accession=142722490, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  723926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723926, CurrentID=0, Status='live', Symbol='Mir25', Aliases='Mirn25, mir-25, mmu-mir-25', description='microRNA 25', other_designations='mir 25', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138165321, end_position_on_the_genomic_accession=138165404, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18183, CurrentID=0, Status='live', Symbol='Nrg3', Aliases='ska', description='neuregulin 3', other_designations='pro-NRG3|scaramanga', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=38365102, end_position_on_the_genomic_accession=39473113, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18185, CurrentID=0, Status='live', Symbol='Nrl', Aliases='D14H14S46E', description='neural retina leucine zipper gene', other_designations='', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55518978, end_position_on_the_genomic_accession=55524981, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18186, CurrentID=0, Status='live', Symbol='Nrp1', Aliases='C530029I03, NP-1, NPN-1, Npn1, Nrp', description='neuropilin 1', other_designations='A5 protein|Neuropilin-1 precursor (A5 protein)|transmembrane receptor', map_location='8 E', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=128358285, end_position_on_the_genomic_accession=128505476, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  18187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18187, CurrentID=0, Status='live', Symbol='Nrp2', Aliases='1110048P06Rik, Np-2, Np2, Npn-2, Npn2', description='neuropilin 2', other_designations='vascular endothelial cell growth factor 165 receptor 2', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=62703317, end_position_on_the_genomic_accession=62818692, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18188, CurrentID=0, Status='live', Symbol='Nrtn', Aliases='NTN', description='neurturin', other_designations='', map_location='17 29.49 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56751323, end_position_on_the_genomic_accession=56757530, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18189, CurrentID=0, Status='live', Symbol='Nrxn1', Aliases='1700062G21Rik, 9330127H16Rik, A230068P09Rik, mKIAA0578', description='neurexin I', other_designations='alpha-latrotoxin receptor (calcium-dependent)|neurexin 1 NRXN1+7a|neurexin 1 NRXN1-7', map_location='17 E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=90033644, end_position_on_the_genomic_accession=91093100, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  18190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18190, CurrentID=0, Status='live', Symbol='Nrxn2', Aliases='6430591O13Rik, mKIAA0921', description='neurexin II', other_designations='neurexin II alpha|neurexin II beta', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6418738, end_position_on_the_genomic_accession=6533217, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  18191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18191, CurrentID=0, Status='live', Symbol='Nrxn3', Aliases='', description='neurexin III', other_designations='neurexin-3-alpha', map_location='12 42.94 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=88722865, end_position_on_the_genomic_accession=90334935, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  18193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18193, CurrentID=0, Status='live', Symbol='Nsd1', Aliases='AI528500, KMT3B', description='nuclear receptor-binding SET-domain protein 1', other_designations='H3-K36-HMTase|H4-K20-HMTase|NR-binding SET domain-containing protein|nsd1|nuclear receptor-binding SET domain-containing protein 1', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55209782, end_position_on_the_genomic_accession=55318325, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  18194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18194, CurrentID=0, Status='live', Symbol='Nsdhl', Aliases='AI747449, Bpa, H105E3, Str, XAP104', description='NAD(P) dependent steroid dehydrogenase-like', other_designations='NAD(P)-dependent steroid dehydrogenase-like protein|bare patches|striated', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72918521, end_position_on_the_genomic_accession=72958528, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  18195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18195, CurrentID=0, Status='live', Symbol='Nsf', Aliases='AI316878, AU020090, AU067812, SKD2', description='N-ethylmaleimide sensitive fusion protein', other_designations='N-ethylmaleimide sensitive factor|N-ethylmaleimide-sensitive fusion protein|NEM-sensitive fusion protein|suppressor of K(+) transport growth defect 2|vesicular-fusion protein NSF', map_location='11 67.54 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103821782, end_position_on_the_genomic_accession=103954056, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  18196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18196, CurrentID=0, Status='live', Symbol='Nsg1', Aliases='Neep21, m234, p21', description='neuron specific gene family member 1', other_designations='brain neuron cytoplasmic protein 1', map_location='5 20.21 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38137192, end_position_on_the_genomic_accession=38159467, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18197, CurrentID=0, Status='live', Symbol='Nsg2', Aliases='', description='neuron specific gene family member 2', other_designations='protein 8.5|protein p19', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32000386, end_position_on_the_genomic_accession=32059211, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  18198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18198, CurrentID=0, Status='live', Symbol='Musk', Aliases='Mdk4, Mlk, Nsk1, Nsk2, Nsk3', description='muscle, skeletal, receptor tyrosine kinase', other_designations='muscle localized kinase 2|muscle, skeletal receptor tyrosine protein kinase|muscle-specific kinase receptor|muscle-specific protein kinase secretory isoform|muscle-specific tyrosine protein kinase receptor|muscle-specific tyrosine-protein kinase receptor', map_location='4 31.87 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58285640, end_position_on_the_genomic_accession=58377848, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18201, CurrentID=0, Status='live', Symbol='Nsmaf', Aliases='AA959567, C630007J05, Fan', description='neutral sphingomyelinase (N-SMase) activation associated factor', other_designations='factor associated with N-SMase activation|factor associated with neutral sphingomyelinase activation', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=6396207, end_position_on_the_genomic_accession=6454668, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  18203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18203, CurrentID=0, Status='live', Symbol='Ntan1', Aliases='', description='N-terminal Asn amidase', other_designations='N-terminal asparagine amidohydrolase|PNAA|PNAD|asparagine-specific N-terminal amidase|protein N-terminal Asn amidase|protein NH2-terminal asparagine amidohydrolase|protein NH2-terminal asparagine deamidase|protein NTN-amidase', map_location='16 9.64 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13818834, end_position_on_the_genomic_accession=13835450, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18205, CurrentID=0, Status='live', Symbol='Ntf3', Aliases='AI316846, AI835689, Nt3, Ntf-3', description='neurotrophin 3', other_designations='HDNF|NGF-2|nerve growth factor 2|neurotrophic factor', map_location='6 60.45 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126101412, end_position_on_the_genomic_accession=126166791, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18207, CurrentID=0, Status='live', Symbol='Nthl1', Aliases='Nth1, Octs3', description='nth (endonuclease III)-like 1 (E.coli)', other_designations='DNA glycoslyase/AP lyase|bifunctional DNA N-glycoslyase/DNA-(apurinic or apyrimidinic site) lyase|thymine glycol DNA glycosylase/AP lyase', map_location='17 A3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24632682, end_position_on_the_genomic_accession=24638838, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18208, CurrentID=0, Status='live', Symbol='Ntn1', Aliases='AI561871, BC019633, Netrin-1', description='netrin 1', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68209364, end_position_on_the_genomic_accession=68386826, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  18209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18209, CurrentID=0, Status='live', Symbol='Ntn3', Aliases='Ntn2l', description='netrin 3', other_designations='netrin-2-like protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24203831, end_position_on_the_genomic_accession=24209387, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18211, CurrentID=0, Status='live', Symbol='Ntrk1', Aliases='C80751, Tkr, TrkA, trk', description='neurotrophic tyrosine kinase, receptor, type 1', other_designations='', map_location='3 38.62 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87778244, end_position_on_the_genomic_accession=87795242, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  18212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18212, CurrentID=0, Status='live', Symbol='Ntrk2', Aliases='GP145-TrkB/GP95-TrkB, Tkrb, trk-B, trkB', description='neurotrophic tyrosine kinase, receptor, type 2', other_designations='neurotrophic tyrosine receptor kinase type 2|trkB tyrosine kinase', map_location='13 31.2 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58806569, end_position_on_the_genomic_accession=59133970, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  18213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18213, CurrentID=0, Status='live', Symbol='Ntrk3', Aliases='AW125844_tv3, TrkC, Ntrk3', description='neurotrophic tyrosine kinase, receptor, type 3', other_designations='GP145-TrkC|neurotrophin-3|trk-C|trkC tyrosine kinase', map_location='7 44.01 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=78179986, end_position_on_the_genomic_accession=78578852, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  18214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18214, CurrentID=0, Status='live', Symbol='Ddr2', Aliases='AW495251, Ntrkr3, tyro10', description='discoidin domain receptor family, member 2', other_designations='CD167 antigen-like family member B|discoidin domain receptor 2|neurotrophic tyrosine kinase, receptor-related 3|receptor protein-tyrosine kinase TKT|tyrosine-protein kinase TYRO10', map_location='1 H1-H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=169972307, end_position_on_the_genomic_accession=170110836, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  18216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18216, CurrentID=0, Status='live', Symbol='Ntsr1', Aliases='NT-1R, NTR-1, NTR1', description='neurotensin receptor 1', other_designations='NT-R-1', map_location='2 H1-H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180499976, end_position_on_the_genomic_accession=180544979, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68572, CurrentID=0, Status='live', Symbol='Ict1', Aliases='1110001A02Rik, 1110002E03Rik, MRP-L58', description='immature colon carcinoma transcript 1', other_designations='39S ribosomal protein L58, mitochondrial|immature colon carcinoma transcript 1 protein homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115403725, end_position_on_the_genomic_accession=115410919, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18218, CurrentID=0, Status='live', Symbol='Dusp8', Aliases='5530400B01Rik, AI593498, Nttp1', description='dual specificity phosphatase 8', other_designations='neuronal tyrosine threonine phosphatase 1|neuronal tyrosine/threonine phosphatase 1', map_location='7 87.59 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142079487, end_position_on_the_genomic_accession=142095880, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18220, CurrentID=0, Status='live', Symbol='Nucb1', Aliases='B230337F23Rik, C77483, Calnuc, MTEST82, Nucb', description='nucleobindin 1', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45492674, end_position_on_the_genomic_accession=45510408, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  18221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18221, CurrentID=0, Status='live', Symbol='Nudc', Aliases='SIG-92, Silg92', description='nuclear distribution gene C homolog (Aspergillus)', other_designations='nuclear distribution protein C homolog|silica-induced gene 92 protein', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133532542, end_position_on_the_genomic_accession=133546027, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18222, CurrentID=0, Status='live', Symbol='Numb', Aliases='Nb', description='numb gene homolog (Drosophila)', other_designations='', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83794034, end_position_on_the_genomic_accession=83921934, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  18223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18223, CurrentID=0, Status='live', Symbol='Numbl', Aliases='nbl', description='numb-like', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27258438, end_position_on_the_genomic_accession=27282150, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18226, CurrentID=0, Status='live', Symbol='Nup62', Aliases='D7Ertd649e, Nupc1, p62', description='nucleoporin 62', other_designations='62 kDa nucleoporin|nuclear pore complex 1|nucleoporin Nup62', map_location='7 28.93 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44816420, end_position_on_the_genomic_accession=44830807, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18227, CurrentID=0, Status='live', Symbol='Nr4a2', Aliases='HZF-3, NOT, Nurr1, RNR-1, TINOR, TINUR', description='nuclear receptor subfamily 4, group A, member 2', other_designations='NUR-related factor 1|nur related protein 1|orphan nuclear receptor NURR1', map_location='2 31.66 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=57106830, end_position_on_the_genomic_accession=57124003, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18229, CurrentID=0, Status='live', Symbol='nv', Aliases='', description='Nijmegen waltzer', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18230, CurrentID=0, Status='live', Symbol='Nxn', Aliases='l11Jus13', description='nucleoredoxin', other_designations='protein Red-1', map_location='11 45.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76257226, end_position_on_the_genomic_accession=76399141, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18231, CurrentID=0, Status='live', Symbol='Nxph1', Aliases='C130005L03Rik', description='neurexophilin 1', other_designations='', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=8948614, end_position_on_the_genomic_accession=9259863, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18232, CurrentID=0, Status='live', Symbol='Nxph2', Aliases='', description='neurexophilin 2', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=23320899, end_position_on_the_genomic_accession=23401986, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18241, CurrentID=0, Status='live', Symbol='Gpr143', Aliases='Oa1', description='G protein-coupled receptor 143', other_designations='MOA1|homolog of human ocular albinism 1 (Nettleship-Falls)|ocular albinism type 1 protein homolog', map_location='X F2-F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152781921, end_position_on_the_genomic_accession=152808651, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18242, CurrentID=0, Status='live', Symbol='Oat', Aliases='AI194874', description='ornithine aminotransferase', other_designations='ornithine--oxo-acid aminotransferase', map_location='7 76.3 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132557475, end_position_on_the_genomic_accession=132576398, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18245, CurrentID=0, Status='live', Symbol='Oaz1', Aliases='AZ-1, AZ1, Antizyme, ODC-Az, Oaz', description='ornithine decarboxylase antizyme 1', other_designations='antizyme 1', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80826656, end_position_on_the_genomic_accession=80829290, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18247, CurrentID=0, Status='live', Symbol='Oaz2', Aliases='AZ-2, AZ2-ps, Sez15, Oaz2', description='ornithine decarboxylase antizyme 2', other_designations='ODC-Az 2|antizyme 2|seizure-related protein 15', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65676548, end_position_on_the_genomic_accession=65690300, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18249, CurrentID=0, Status='live', Symbol='Obp1a', Aliases='OBP-1a, OBP5', description='odorant binding protein IA', other_designations='odorant binding protein 5|odorant-binding protein Ia', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=78085505, end_position_on_the_genomic_accession=78091374, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  18256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18256, CurrentID=0, Status='live', Symbol='Oc90', Aliases='Ocn-95, PLA2L, Pla2ll, oc95', description='otoconin 90', other_designations='otoconin-95|phospholipase A2-like like', map_location='15 29.16 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=65876050, end_position_on_the_genomic_accession=65912411, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18260, CurrentID=0, Status='live', Symbol='Ocln', Aliases='AI503564, Ocl', description='occludin', other_designations='', map_location='13 53.23 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100496504, end_position_on_the_genomic_accession=100553689, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18261, CurrentID=0, Status='live', Symbol='Ocm', Aliases='', description='oncomodulin', other_designations='OM|parvalbumin beta', map_location='5 G1-3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144019807, end_position_on_the_genomic_accession=144050640, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18263, CurrentID=0, Status='live', Symbol='Odc1', Aliases='ODC', description='ornithine decarboxylase, structural 1', other_designations='', map_location='12 8.11 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=17544873, end_position_on_the_genomic_accession=17551502, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  280408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280408, CurrentID=0, Status='live', Symbol='Rilp', Aliases='AI835553, Gm857', description='Rab interacting lysosomal protein', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75510094, end_position_on_the_genomic_accession=75513168, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  280411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280411, CurrentID=0, Status='live', Symbol='Lix1l', Aliases='D130027M04Rik', description='Lix1-like', other_designations='', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96601133, end_position_on_the_genomic_accession=96625352, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  723941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723941, CurrentID=0, Status='live', Symbol='Mir541', Aliases='Mirn541, mir-541, mmu-mir-541', description='microRNA 541', other_designations='mir 541', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109742409, end_position_on_the_genomic_accession=109742498, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18285, CurrentID=0, Status='live', Symbol='Odf1', Aliases='', description='outer dense fiber of sperm tails 1', other_designations='outer dense fiber protein-1', map_location='15 B2-C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=38219203, end_position_on_the_genomic_accession=38226735, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18286, CurrentID=0, Status='live', Symbol='Odf2', Aliases='AI848335, MMTEST29', description='outer dense fiber of sperm tails 2', other_designations='84 kDa outer dense fiber protein|cenexin|outer dense fiber of sperm tails protein 2', map_location='2 20.79 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29889720, end_position_on_the_genomic_accession=29931746, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  18290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18290, CurrentID=0, Status='live', Symbol='Ofa', Aliases='59kDa', description='oncofetal antigen', other_designations='oncofetal antigen, 59 kDa', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18291, CurrentID=0, Status='live', Symbol='Nobox', Aliases='OG2, Og2x', description='NOBOX oogenesis homeobox', other_designations='OG2 homeobox|homeodomain-containing protein OG-2|newborn ovary homeobox protein|oocyte-specific homeobox protein', map_location='6 21.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43303674, end_position_on_the_genomic_accession=43317378, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18292, CurrentID=0, Status='live', Symbol='Sebox', Aliases='OG9, Og9x', description='SEBOX homeobox', other_designations='OG9 homeobox|homeobox OG-9|skin-, embryo-, brain- and oocyte-specific homeobox', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78503513, end_position_on_the_genomic_accession=78505081, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18293, CurrentID=0, Status='live', Symbol='Ogdh', Aliases='2210403E04Rik, 2210412K19Rik, AA409584, d1401, mKIAA4192', description='oxoglutarate (alpha-ketoglutarate) dehydrogenase (lipoamide)', other_designations='2-oxoglutarate dehydrogenase complex component E1|OGDC-E1|alpha-ketoglutarate dehydrogenase|oxoglutarate dehydrogenase (lipoamide)', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6291593, end_position_on_the_genomic_accession=6359094, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  18294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18294, CurrentID=0, Status='live', Symbol='Ogg1', Aliases='Mmh', description='8-oxoguanine DNA-glycosylase 1', other_designations='', map_location='6 E-F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113326976, end_position_on_the_genomic_accession=113334188, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18295, CurrentID=0, Status='live', Symbol='Ogn', Aliases='3110079A16Rik, OG, OIF, SLRR3A, mimecan, mimican', description='osteoglycin', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49608071, end_position_on_the_genomic_accession=49624500, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18300, CurrentID=0, Status='live', Symbol='Oit1', Aliases='2310076N21Rik, AV067083, EF-7, Fam3d', description='oncoprotein induced transcript 1', other_designations='oncoprotein-induced protein 1|protein EF-7', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8348937, end_position_on_the_genomic_accession=8378763, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  18301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18301, CurrentID=0, Status='live', Symbol='Fxyd5', Aliases='EF-8, Oit2, RIC', description='FXYD domain-containing ion transport regulator 5', other_designations='dysadherin|ion channel homolog RIC|oncoprotein induced transcript 2|oncoprotein-induced protein 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31032723, end_position_on_the_genomic_accession=31042331, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18302, CurrentID=0, Status='live', Symbol='Oit3', Aliases='EF-9, LZP', description='oncoprotein induced transcript 3', other_designations='liver-specific zona pellucida domain-containing protein', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59422960, end_position_on_the_genomic_accession=59441779, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18307, CurrentID=0, Status='live', Symbol='Olfr10', Aliases='GA_x6K02T2QP88-6117098-6116163, L45, MOR256-55', description='olfactory receptor 10', other_designations='odorant receptor L45', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49317548, end_position_on_the_genomic_accession=49318483, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18310, CurrentID=0, Status='live', Symbol='Olfr13', Aliases='K7, MOR261-6', description='olfactory receptor 13', other_designations='GA_x6K02T2P3E9-4384160-4383228|odorant receptor K7|olfactory receptor 261-6|olfactory receptor MOR261-6', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43173988, end_position_on_the_genomic_accession=43174920, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18312, CurrentID=0, Status='live', Symbol='Olfr15', Aliases='MOR256-17, OR3', description='olfactory receptor 15', other_designations='GA_x54KRFPKG5P-348087-349025|odorant receptor OR3|olfactory receptor 256-17|olfactory receptor MOR256-17', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3838975, end_position_on_the_genomic_accession=3839913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723948, CurrentID=0, Status='live', Symbol='Mir302b', Aliases='Mirn302b, mir-302b', description='microRNA 302b', other_designations='mir 302b|mmu-mir-302b', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127545228, end_position_on_the_genomic_accession=127545301, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18314, CurrentID=0, Status='live', Symbol='Olfr17', Aliases='MOR263-5, P2', description='olfactory receptor 17', other_designations='GA_x6K02T2PBJ9-9470146-9471093|P2 olfactory receptor|olfactory receptor MOR263-5', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107097467, end_position_on_the_genomic_accession=107098414, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18315, CurrentID=0, Status='live', Symbol='Olfr18', Aliases='MOR145-1, MTPCR34', description='olfactory receptor 18', other_designations='GA_x6K02T2PVTD-14054886-14053957|olfactory receptor 145-1|olfactory receptor TPCR34', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20313892, end_position_on_the_genomic_accession=20336094, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18316, CurrentID=0, Status='live', Symbol='Olfr19', Aliases='M12, MOR140-1, MTPCR15', description='olfactory receptor 19', other_designations='GA_x54KRFPKG5P-13123979-13123050|odorant receptor M12|olfactory receptor 140-1|olfactory receptor MOR140-1', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16673050, end_position_on_the_genomic_accession=16673979, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18317, CurrentID=0, Status='live', Symbol='Olfr2', Aliases='I7, MOR103-15, Olfr41', description='olfactory receptor 2', other_designations='GA_x6K02T2PBJ9-9381439-9380456|I7 olfactory receptor|odorant receptor B2|olfactory recepter I7|olfactory receptor 41', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107000875, end_position_on_the_genomic_accession=107005558, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18321, CurrentID=0, Status='live', Symbol='Olfr23', Aliases='MOR135-27, MTPCR50', description='olfactory receptor 23', other_designations='GA_x6K02T2P1NL-4097159-4098136|olfactory receptor MOR135-27', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73930360, end_position_on_the_genomic_accession=73941225, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18322, CurrentID=0, Status='live', Symbol='Olfr24', Aliases='MOR132-1, MTPCR51', description='olfactory receptor 24', other_designations='GA_x6K02T2PVTD-12498544-12497603|olfactory receptor 132-1|olfactory receptor MOR132-1|olfactory receptor TPCR51', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18754692, end_position_on_the_genomic_accession=18755633, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18324, CurrentID=0, Status='live', Symbol='Olfr26', Aliases='MOR171-9, MTPCR09', description='olfactory receptor 26', other_designations='GA_x6K02T2PVTD-32550930-32551856|olfactory receptor MOR171-9', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38855060, end_position_on_the_genomic_accession=38855990, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18328, CurrentID=0, Status='live', Symbol='Olfr3', Aliases='MOR136-14, Y71', description='olfactory receptor 3', other_designations='GA_x6K02T2NLDC-33507606-33506665|odorant receptor K10|olfactory receptor 136-14|olfactory receptor MOR136-14', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36812149, end_position_on_the_genomic_accession=36813090, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18329, CurrentID=0, Status='live', Symbol='Olfr30', Aliases='MOR2811, MTPCR07', description='olfactory receptor 30', other_designations='', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58454915, end_position_on_the_genomic_accession=58455980, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18330, CurrentID=0, Status='live', Symbol='Olfr31', Aliases='MOR274-1, MTPCR53', description='olfactory receptor 31', other_designations='GA_x6K02T2PLTE-6714644-6715597|olfactory receptor MOR274-1', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=14328113, end_position_on_the_genomic_accession=14329066, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18331, CurrentID=0, Status='live', Symbol='Olfr32', Aliases='MOR227-7P, MOR227-9_p, MTPCR05', description='olfactory receptor 32', other_designations='GA_x6K02T2Q125-51573576-51572650|olfactory receptor MOR227-7P', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90138081, end_position_on_the_genomic_accession=90268639, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18332, CurrentID=0, Status='live', Symbol='Olfr33', Aliases='MOR11-2, MTPCR33, ORL181', description='olfactory receptor 33', other_designations='GA_x6K02T2PBJ9-5431102-5430146|odorant receptor', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102713455, end_position_on_the_genomic_accession=102714411, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68592, CurrentID=0, Status='live', Symbol='Syf2', Aliases='1110018L13Rik, Cbpin, D4Bwg1551e, Gcipip, Ntc31, mp29, p29', description='SYF2 homolog, RNA splicing factor (S. cerevisiae)', other_designations='CCNDBP1 interactor|CCNDBP1-interactor|GCIP-interacting protein p29 (Gcipip)', map_location='4 67.15 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134930980, end_position_on_the_genomic_accession=134937537, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18341, CurrentID=0, Status='live', Symbol='Olfr263', Aliases='GA_x5J8B7TTCWC-858-415, IA3, MOR256-31, MOR256-37P, MOR256-61-ps1, Olfr42, Olfr263', description='olfactory receptor 263', other_designations='GA_x6K02T2N5E5-9379-8514|GA_x6K02T2QHY8-12114828-12113875|olfactory receptor 263, pseudogene 1|olfactory receptor 42|olfactory receptor GA_x5J8B7TTCWC-858-415|olfactory receptor MOR256-31|olfactory receptor MOR256-37P', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21132777, end_position_on_the_genomic_accession=21133730, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18343, CurrentID=0, Status='live', Symbol='Olfr44', Aliases='IB3, MOR171-5', description='olfactory receptor 44', other_designations='GA_x6K02T2PVTD-33181773-33180838|odorant receptor K11|olfactory receptor MOR171-5', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39484087, end_position_on_the_genomic_accession=39493988, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18344, CurrentID=0, Status='live', Symbol='Olfr45', Aliases='IB6, MOR253-2', description='olfactory receptor 45', other_designations='GA_x6K02T2PBJ9-42837030-42837962|olfactory receptor MOR253-2', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140690907, end_position_on_the_genomic_accession=140691842, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18345, CurrentID=0, Status='live', Symbol='Olfr46', Aliases='IB7, ID12, IF5, MOR253-8', description='olfactory receptor 46', other_designations='GA_x6K02T2PBJ9-42759973-42760905|olfactory receptor MOR253-8', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140601340, end_position_on_the_genomic_accession=140611124, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18346, CurrentID=0, Status='live', Symbol='Olfr47', Aliases='IB12, MOR261-9', description='olfactory receptor 47', other_designations='GA_x6K02T2P3E9-4322325-4321360|olfactory receptor MOR261-9', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43235610, end_position_on_the_genomic_accession=43236575, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18347, CurrentID=0, Status='live', Symbol='Olfr48', Aliases='C3, IC3, MOR232-5', description='olfactory receptor 48', other_designations='GA_x6K02T2Q125-51285881-51284976|olfactory receptor MOR232-5', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89844066, end_position_on_the_genomic_accession=89844971, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18348, CurrentID=0, Status='live', Symbol='Olfr49', Aliases='IC6, MOR1181', description='olfactory receptor 49', other_designations='olfactory receptor 118-1|olfactory receptor I-C6|olfr49', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54281896, end_position_on_the_genomic_accession=54282925, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18349, CurrentID=0, Status='live', Symbol='Olfr5', Aliases='MOR103-8, ORL120', description='olfactory receptor 5', other_designations='GA_x6K02T2QGBW-3210997-3210059|odorant receptor M41|olfactory receptor 103-8', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6480216, end_position_on_the_genomic_accession=6486909, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18350, CurrentID=0, Status='live', Symbol='Olfr50', Aliases='ID3, MOR136-6', description='olfactory receptor 50', other_designations='GA_x6K02T2NLDC-33487752-33488690|olfactory receptor 136-6|olfactory receptor I-D3|olfactory receptor MOR136-6|olfr50', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36793238, end_position_on_the_genomic_accession=36794176, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18351, CurrentID=0, Status='live', Symbol='Olfr51', Aliases='ID7, MOR129-3', description='olfactory receptor 51', other_designations='GA_x6K02T2QP88-4431129-4430206|olfactory receptor MOR129-3', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51006974, end_position_on_the_genomic_accession=51007897, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18352, CurrentID=0, Status='live', Symbol='Olfr52', Aliases='IE6, MOR185-6, ORL446', description='olfactory receptor 52', other_designations='GA_x6K02T2Q125-47650922-47649963', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86181150, end_position_on_the_genomic_accession=86182109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18354, CurrentID=0, Status='live', Symbol='Olfr54', Aliases='F3, MOR126-1', description='olfactory receptor 54', other_designations='GA_x6K02T2QP88-4411197-4410256|olfactory receptor MOR126-1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51027004, end_position_on_the_genomic_accession=51027945, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18356, CurrentID=0, Status='live', Symbol='Olfr56', Aliases='IF7, MOR276-1', description='olfactory receptor 56', other_designations='GA_x6K02T2QP88-6300500-6299553|olfactory receptor 276-1|olfactory receptor F7|olfactory receptor G3|olfactory receptor I-G7|olfactory receptor MOR276-1|olfr56', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49050733, end_position_on_the_genomic_accession=49135387, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18357, CurrentID=0, Status='live', Symbol='Olfr57', Aliases='IF12, MOR139-3', description='olfactory receptor 57', other_designations='GA_x6K02T2QGN0-2777431-2776472|olfactory receptor MOR139-3|olfactory receptor OR6-8', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79034723, end_position_on_the_genomic_accession=79035802, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18358, CurrentID=0, Status='live', Symbol='Olfr58', Aliases='IG6, MOR146-3, MOR146-7P', description='olfactory receptor 58', other_designations='GA_x6K02T2PVTD-13523015-13523944|olfactory receptor MOR146-7P', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19780283, end_position_on_the_genomic_accession=19784064, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18359, CurrentID=0, Status='live', Symbol='Olfr59', Aliases='IH3, MOR133-3P', description='olfactory receptor 59', other_designations='GA_x6K02T2P1NL-4434429-4435400|olfactory receptor MOR133-3P', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74283752, end_position_on_the_genomic_accession=74289646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18361, CurrentID=0, Status='live', Symbol='Olfr60', Aliases='IH6, MOR253-6', description='olfactory receptor 60', other_designations='GA_x6K02T2PBJ9-42496183-42495251|olfactory receptor MOR253-6', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140345052, end_position_on_the_genomic_accession=140345987, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18362, CurrentID=0, Status='live', Symbol='Olfr61', Aliases='IH7, MOR253-1', description='olfactory receptor 61', other_designations='GA_x6K02T2PBJ9-42783826-42784758|olfactory receptor MOR253-1', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140637703, end_position_on_the_genomic_accession=140638638, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18365, CurrentID=0, Status='live', Symbol='Olfr65', Aliases="5'[b]3, 5'beta3, MOR1-1, MOR1-2, ORL670, Olfr64", description='olfactory receptor 65', other_designations='GA_x6K02T2PBJ9-6634906-6633983', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103906342, end_position_on_the_genomic_accession=103907447, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18366, CurrentID=0, Status='live', Symbol='Olfr64', Aliases="5'[b]2, 5'beta2, MOR1-1, MOR1-2, ORL533", description='olfactory receptor 64', other_designations='GA_x6K02T2PBJ9-6648196-6649143', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103892810, end_position_on_the_genomic_accession=103894471, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18367, CurrentID=0, Status='live', Symbol='Olfr66', Aliases="5'[b]1, 5'beta1, MOR1-3, ORL532", description='olfactory receptor 66', other_designations='GA_x6K02T2PBJ9-6620959-6620024', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103881306, end_position_on_the_genomic_accession=103882241, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18368, CurrentID=0, Status='live', Symbol='Olfr67', Aliases="3'[b]1, 3'beta1, MOR31-1", description='olfactory receptor 67', other_designations='GA_x6K02T2PBJ9-6515150-6514170|olfactory receptor MOR31-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103787212, end_position_on_the_genomic_accession=103791830, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18369, CurrentID=0, Status='live', Symbol='Olfr68', Aliases="3'[b]2, 3'beta2, MOR22-3", description='olfactory receptor 68', other_designations='GA_x6K02T2PBJ9-6504436-6503489|olfactory receptor MOR22-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103777396, end_position_on_the_genomic_accession=103778343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18370, CurrentID=0, Status='live', Symbol='Olfr69', Aliases="3'[b]3, 3'beta3, MOR22-2", description='olfactory receptor 69', other_designations='GA_x6K02T2PBJ9-6494485-6493535|olfactory receptor MOR22-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103767277, end_position_on_the_genomic_accession=103771594, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18371, CurrentID=0, Status='live', Symbol='Olfr7', Aliases='M71, MOR171-2', description='olfactory receptor 7 complex', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18372, CurrentID=0, Status='live', Symbol='Olfr8', Aliases='GA_x5J8B7TNKDU-1250-1574, MOR139-5P', description='olfactory receptor 8', other_designations='GA_x6K02T2QGN0-2857086-2856154|odorant receptor M64|olfactory receptor GA_x5J8B7TNKDU-1250-1574', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78955207, end_position_on_the_genomic_accession=78956139, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18373, CurrentID=0, Status='live', Symbol='Olfr9', Aliases='MOR269-3', description='olfactory receptor 9', other_designations='GA_x6K02T2PULF-10675610-10676548|odorant receptor M25|olfactory receptor 269-3|olfactory receptor MOR269-3', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128989914, end_position_on_the_genomic_accession=128990852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18375, CurrentID=0, Status='live', Symbol='olt', Aliases='', description='oligotriche', other_designations='Del(9)olt1Pas|Del(9Ctdspl-Slc22a14)1Pas', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18377, CurrentID=0, Status='live', Symbol='Omg', Aliases='', description='oligodendrocyte myelin glycoprotein', other_designations='', map_location='11 46.79 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79500982, end_position_on_the_genomic_accession=79504082, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18378, CurrentID=0, Status='live', Symbol='Omp', Aliases='', description='olfactory marker protein', other_designations='', map_location='7 53.59 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98143359, end_position_on_the_genomic_accession=98145447, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18379, CurrentID=0, Status='live', Symbol='Omt2a', Aliases='OM2a', description='oocyte maturation, alpha', other_designations='', map_location='9 43.65 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78311972, end_position_on_the_genomic_accession=78314797, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  112290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112290, CurrentID=0, Status='live', Symbol='Lorp1', Aliases='', description='loss of righting due to propofol 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18383, CurrentID=0, Status='live', Symbol='Tnfrsf11b', Aliases='OCIF, Opg, TR1', description='tumor necrosis factor receptor superfamily, member 11b (osteoprotegerin)', other_designations='osteoclastogenesis inhibitory factor|osteoprotegerin', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=54250619, end_position_on_the_genomic_accession=54278484, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18386, CurrentID=0, Status='live', Symbol='Oprd1', Aliases='DOR, DOR-1, Nbor, mDOR', description='opioid receptor, delta 1', other_designations='D-OR-1|K56|MSL-2|delta 1 opioid receptor', map_location='4 D1-D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132110726, end_position_on_the_genomic_accession=132144486, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18387, CurrentID=0, Status='live', Symbol='Oprk1', Aliases='K-OR-1, KOR, KOR-1, MSL-1, Oprk2, R21', description='opioid receptor, kappa 1', other_designations='opioid receptor, kappa 2', map_location='1 A2-A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=5588450, end_position_on_the_genomic_accession=5606134, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18389, CurrentID=0, Status='live', Symbol='Oprl1', Aliases='KOR-3, KOR3, LC132, MOR-C, ORGC, ORL1, Oprl, XOR1, morc', description='opioid receptor-like 1', other_designations='K3 opiate receptor|N/OFQ receptor|kappa-type 3 opioid receptor|nociceptin/ orphaninFQ receptor', map_location='2 H2-4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181714806, end_position_on_the_genomic_accession=181720985, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18390, CurrentID=0, Status='live', Symbol='Oprm1', Aliases='M-OR-1, MOP-R, MOR-1, MOR-1O, Oprm, mor, muOR', description='opioid receptor, mu 1', other_designations='MOP receptor|mu opioid receptor MOR-1P|mu opioid receptor splice variant mMOR-1Z', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=6758593, end_position_on_the_genomic_accession=7038209, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  18391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18391, CurrentID=0, Status='live', Symbol='Sigmar1', Aliases='Oprs1, sigma1R', description='sigma non-opioid intracellular receptor 1', other_designations='opioid receptor, sigma 1|sigma 1 receptor variant SR-1D|sigma 1-type opioid receptor|sigma1-receptor', map_location='4 A5-B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41738493, end_position_on_the_genomic_accession=41741359, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18392, CurrentID=0, Status='live', Symbol='Orc1', Aliases='AA545195, MmORC1l, Orc1', description='origin recognition complex, subunit 1', other_designations='', map_location='4 D', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108575499, end_position_on_the_genomic_accession=108616048, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18393, CurrentID=0, Status='live', Symbol='Orc2', Aliases='AU041563l, Orc2', description='origin recognition complex, subunit 2', other_designations='', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58462771, end_position_on_the_genomic_accession=58505109, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  18399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18399, CurrentID=0, Status='live', Symbol='Slc22a6', Aliases='NKT, Oat1, Orctl1, mOat1', description='solute carrier family 22 (organic anion transporter), member 6', other_designations='kidney-specific transport protein|mNKT|mROAT1|novel kidney transcript|organic anion transporter 1|organic cationic transporter-like 1|renal organic anion transporter 1', map_location='19 5.44 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8617996, end_position_on_the_genomic_accession=8628299, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18400, CurrentID=0, Status='live', Symbol='Slc22a18', Aliases='AW260131, BWR1A, BWSCR1A, HET, ITM, Impt1, Orctl2, Slc22a1l, TSSC5, p45-BWR1A', description='solute carrier family 22 (organic cation transporter), member 18', other_designations='ORCTL-2|imprinted multi-membrane spanning polyspecific transporter-related protein 1|organic cationic transporter-like 2|tumor suppressing subtransferable candidate 5', map_location='7 88.22 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143473726, end_position_on_the_genomic_accession=143499337, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  68603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68603, CurrentID=0, Status='live', Symbol='Pmvk', Aliases='1110011E12Rik, 2900002L22Rik, AV001327, PMK, PMKA, PMKASE', description='phosphomevalonate kinase', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89454546, end_position_on_the_genomic_accession=89469014, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18405, CurrentID=0, Status='live', Symbol='Orm1', Aliases='Agp-1, Agp-2, Orm-1', description='orosomucoid 1', other_designations='AGP 1|OMD 1|orosomucoid-1', map_location='4 33.96 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63344566, end_position_on_the_genomic_accession=63348162, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18406, CurrentID=0, Status='live', Symbol='Orm2', Aliases='Agp1, Orm-2', description='orosomucoid 2', other_designations='AGP 2|OMD 2|bM340H1.5 (orosomucoid 2)|orosomucoid-2', map_location='4 33.96 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63362449, end_position_on_the_genomic_accession=63365877, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18407, CurrentID=0, Status='live', Symbol='Orm3', Aliases='Agp-3, Orm-3', description='orosomucoid 3', other_designations='AGP 3|OMD 3|orosomucoid-3', map_location='4 33.96 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63355279, end_position_on_the_genomic_accession=63359511, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18408, CurrentID=0, Status='live', Symbol='Slc25a15', Aliases='D630044L02Rik, Ornt1', description='solute carrier family 25 (mitochondrial carrier ornithine transporter), member 15', other_designations='ornithine transporter, mitochondrial 1|solute carrier family 25 member 15', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22375550, end_position_on_the_genomic_accession=22398621, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18412, CurrentID=0, Status='live', Symbol='Sqstm1', Aliases='A170, OSF-6, Osi, STAP, STONE14, p62', description='sequestosome 1', other_designations='oxidative stress induced|ubiquitin-binding protein p62', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50200152, end_position_on_the_genomic_accession=50210820, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18413, CurrentID=0, Status='live', Symbol='Osm', Aliases='OncoM', description='oncostatin M', other_designations='', map_location='11 2.94 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4236785, end_position_on_the_genomic_accession=4241026, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18414, CurrentID=0, Status='live', Symbol='Osmr', Aliases='OSMRB', description='oncostatin M receptor', other_designations='IL-31 receptor subunit beta|IL-31R subunit beta|IL-31R-beta|IL-31RB|interleukin-31 receptor subunit beta|oncostatin receptor|oncostatin-M specific receptor subunit beta', map_location='15 3.3 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=6813568, end_position_on_the_genomic_accession=6875246, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  18415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18415, CurrentID=0, Status='live', Symbol='Hspa4l', Aliases='94kDa, AI461691, APG-1, Osp94', description='heat shock protein 4 like', other_designations='heat shock 70-related protein APG-1|heat shock 70kDa protein 4 like|osmotic stress protein 94', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40744355, end_position_on_the_genomic_accession=40796103, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  18416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18416, CurrentID=0, Status='live', Symbol='Otc', Aliases='AI265390, Sf, spf', description='ornithine transcarbamylase', other_designations='OTCase', map_location='X A1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=10252297, end_position_on_the_genomic_accession=10321024, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18417, CurrentID=0, Status='live', Symbol='Cldn11', Aliases='Claudin-11, Claudin11, Osp, Otm', description='claudin 11', other_designations='oligodendrocyte transmembrane protein|oligodendrocyte-specific protein', map_location='3 15.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=31149920, end_position_on_the_genomic_accession=31164326, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18419, CurrentID=0, Status='live', Symbol='Otog', Aliases='4932420K09, Otgn, twt', description='otogelin', other_designations='twister', map_location='7 B4-C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46240987, end_position_on_the_genomic_accession=46311434, orientation='plus', exon_count=56, OMIM='', no_hdr0=''),
  18420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18420, CurrentID=0, Status='live', Symbol='Otp', Aliases='', description='orthopedia homolog (Drosophila)', other_designations='', map_location='13 49.22 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=94875627, end_position_on_the_genomic_accession=94885130, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68606, CurrentID=0, Status='live', Symbol='Ppm1f', Aliases='1110021B16Rik, 4933427B07Rik, CaMKPase, Popx2, mKIAA0015', description='protein phosphatase 1F (PP2C domain containing)', other_designations='ca(2+)/calmodulin-dependent protein kinase phosphatase|caM-kinase phosphatase|calcium/calmodulin-dependent protein kinase phosphatase|partner of PIX2', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16896469, end_position_on_the_genomic_accession=16927375, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18422, CurrentID=0, Status='live', Symbol='Ott', Aliases='AV272316', description='ovary testis transcribed', other_designations='', map_location='X 68.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=149097952, end_position_on_the_genomic_accession=149138152, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18423, CurrentID=0, Status='live', Symbol='Otx1', Aliases='A730044F23Rik, jv', description='orthodenticle homolog 1', other_designations='', map_location='11 14.1 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=21994764, end_position_on_the_genomic_accession=22001651, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18424, CurrentID=0, Status='live', Symbol='Otx2', Aliases='E130306E05Rik', description='orthodenticle homolog 2', other_designations='', map_location='14 25.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=48657452, end_position_on_the_genomic_accession=48667644, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18425, CurrentID=0, Status='live', Symbol='Oua1', Aliases='Oua-1', description='ouabain resistance 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18426, CurrentID=0, Status='live', Symbol='Ovol1', Aliases='BB147136, Ovo1, movo1', description='OVO homolog-like 1 (Drosophila)', other_designations='mOvo1a', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5549137, end_position_on_the_genomic_accession=5560575, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68607, CurrentID=0, Status='live', Symbol='Serhl', Aliases='0610008B10Rik, 1110019M09Rik, AU019786', description='serine hydrolase-like', other_designations='SHL', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83100191, end_position_on_the_genomic_accession=83116675, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18429, CurrentID=0, Status='live', Symbol='Oxt', Aliases='OT, Oxy', description='oxytocin', other_designations='OT-NPI|neurophysin 1', map_location='2 63.24 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130574967, end_position_on_the_genomic_accession=130577054, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18430, CurrentID=0, Status='live', Symbol='Oxtr', Aliases='OTR', description='oxytocin receptor', other_designations='OT-R', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=112473684, end_position_on_the_genomic_accession=112491308, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18431, CurrentID=0, Status='live', Symbol='Oca2', Aliases='D7H15S12, D7Icr28RN, D7Nic1, p', description='oculocutaneous albinism II', other_designations='melanocyte-specific transporter protein|pink-eyed dilution protein', map_location='7 33.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=56239593, end_position_on_the_genomic_accession=56536517, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  18432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18432, CurrentID=0, Status='live', Symbol='Mybbp1a', Aliases='AL024407, AU019902, P160, p160MBP, p67MBP', description='MYB binding protein (P160) 1a', other_designations='myb-binding protein of 160 kDa|nuclear protein P160', map_location='11 44.29 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72441378, end_position_on_the_genomic_accession=72451550, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  18436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18436, CurrentID=0, Status='live', Symbol='P2rx1', Aliases='AI323649, BB122383, P2x, Pdcd3, RP-2', description='purinergic receptor P2X, ligand-gated ion channel, 1', other_designations='ATP receptor|P2X1 receptor|programmed cell death 3', map_location='11 45.09 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72999056, end_position_on_the_genomic_accession=73015200, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18438, CurrentID=0, Status='live', Symbol='P2rx4', Aliases='AI504491, AW555605, D5Ertd444e, P2X4', description='purinergic receptor P2X, ligand-gated ion channel 4', other_designations='ionotropic purinergic receptor|purinergic receptor P2x4 subunit variant e', map_location='5 62.53 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122707504, end_position_on_the_genomic_accession=122729739, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18439, CurrentID=0, Status='live', Symbol='P2rx7', Aliases='AI467586, P2X(7), P2X7R', description='purinergic receptor P2X, ligand-gated ion channel, 7', other_designations='ATP receptor|P2X7 purinoceptor|P2X7 receptor|P2Z receptor', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122643911, end_position_on_the_genomic_accession=122691432, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  18440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18440, CurrentID=0, Status='live', Symbol='P2rx6', Aliases='P2rxl1, P2x6, P2xm', description='purinergic receptor P2X, ligand-gated ion channel, 6', other_designations='ATP receptor|purinergic receptor P2X-like 1, orphan receptor', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17561855, end_position_on_the_genomic_accession=17572012, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18441, CurrentID=0, Status='live', Symbol='P2ry1', Aliases='P2Y1', description='purinergic receptor P2Y, G-protein coupled 1', other_designations='ATP receptor|P2Y1 receptor', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=61002785, end_position_on_the_genomic_accession=61008982, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18442, CurrentID=0, Status='live', Symbol='P2ry2', Aliases='P2U1, P2Y2', description='purinergic receptor P2Y, G-protein coupled 2', other_designations='ATP receptor|P2U purinoceptor 1', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100996568, end_position_on_the_genomic_accession=101012053, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18451, CurrentID=0, Status='live', Symbol='P4ha1', Aliases='AL022634, P4ha', description='procollagen-proline, 2-oxoglutarate 4-dioxygenase (proline 4-hydroxylase), alpha 1 polypeptide', other_designations='4-PH alpha-1|procollagen-proline,2-oxoglutarate-4-dioxygenase subunit alpha-1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59323246, end_position_on_the_genomic_accession=59373304, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  18452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18452, CurrentID=0, Status='live', Symbol='P4ha2', Aliases='AA407196, C76437, P4hl', description='procollagen-proline, 2-oxoglutarate 4-dioxygenase (proline 4-hydroxylase), alpha II polypeptide', other_designations='4-PH alpha-2|Prolyl 4-hydroxylase alpha IIa subunit|procollagen-proline,2-oxoglutarate-4-dioxygenase subunit alpha-2', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54100924, end_position_on_the_genomic_accession=54131668, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  18453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18453, CurrentID=0, Status='live', Symbol='P4hb', Aliases='ERp59, PDI, Pdia1, Thbp', description='prolyl 4-hydroxylase, beta polypeptide', other_designations='ER protein 59|cellular thyroid hormone-binding protein|endoplasmic reticulum resident protein 59|p55|prolyl 4-hydroxylase subunit beta|protein disulfide isomerase', map_location='11 D-E', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120560304, end_position_on_the_genomic_accession=120572936, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18457, CurrentID=0, Status='live', Symbol='Bloc1s6', Aliases='BLOC-1, Pldn, Stx13bp1, pa', description='biogenesis of lysosomal organelles complex-1, subunit 6, pallidin', other_designations='BLOC-1 subunit 6|biogenesis of organelles complex-1, subunit 6, pallidin|pallid protein|pallidin|syntaxin 13 binding protein 1|syntaxin 13-interacting protein pallid', map_location='2 60.66 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122738505, end_position_on_the_genomic_accession=122749487, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18458, CurrentID=0, Status='live', Symbol='Pabpc1', Aliases='PABP, Pabp1, PabpI, Pabpl1, ePAB', description='poly(A) binding protein, cytoplasmic 1', other_designations='PABP-1|poly A binding protein 1|poly(A)-binding protein 1', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36595658, end_position_on_the_genomic_accession=36609581, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18459, CurrentID=0, Status='live', Symbol='Pabpc2', Aliases='Pabp, Pabp+, Pabp2', description='poly(A) binding protein, cytoplasmic 2', other_designations='poly A binding protein 2', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=39773497, end_position_on_the_genomic_accession=39776082, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  112304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112304, CurrentID=0, Status='live', Symbol='Pas5a', Aliases='Par1', description='pulmonary adenoma susceptibility 5a', other_designations='', map_location='11 62.4 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  723974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723974, CurrentID=0, Status='live', Symbol='Mir500', Aliases='Mirn500, mir-500, mmu-mir-500', description='microRNA 500', other_designations='mir 500', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7237683, end_position_on_the_genomic_accession=7237774, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18472, CurrentID=0, Status='live', Symbol='Pafah1b1', Aliases='LIS-1, Lis1, MMS10-U, Mdsh, Ms10u, Pafaha', description='platelet-activating factor acetylhydrolase, isoform 1b, subunit 1', other_designations='LIS-1|PAF acetylhydrolase 45 kDa subunit|PAF-AH 45 kDa subunit|PAF-AH alpha|PAFAH alpha|lissencephaly-1 protein|platelet-activating factor acetylhydrolase 1b, regulatory subunit 1|platelet-activating factor acetylhydrolase, isoform 1b, beta1 subunit', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74673949, end_position_on_the_genomic_accession=74724384, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18473, CurrentID=0, Status='live', Symbol='Pafah1b1-ps1', Aliases='Lis3, Mdsh3, Pafaha-ps1', description='platelet-activating factor acetylhydrolase, isoform 1b, beta1 subunit, pseudogene 1', other_designations='', map_location='3 41.82 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96407377, end_position_on_the_genomic_accession=96409464, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  18475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18475, CurrentID=0, Status='live', Symbol='Pafah1b2', Aliases='AI747451, AU021353, Pafahb, mus[b]', description='platelet-activating factor acetylhydrolase, isoform 1b, subunit 2', other_designations='PAF acetylhydrolase 30 kDa subunit|PAF-AH 30 kDa subunit|PAF-AH subunit beta|PAFAH subunit beta|platelet-activating factor acetylhydrolase, isoform 1b, alpha2 subunit', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45965311, end_position_on_the_genomic_accession=46012685, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18476, CurrentID=0, Status='live', Symbol='Pafah1b3', Aliases='Pafahg, mus[g]', description='platelet-activating factor acetylhydrolase, isoform 1b, subunit 3', other_designations='PAF acetylhydrolase 29 kDa subunit|PAF-AH 29 kDa subunit|PAF-AH gamma|PAF-AH subunit gamma|PAFAH subunit gamma|platelet-activating factor acetylhydrolase, isoform 1b, alpha1 subunit', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25295049, end_position_on_the_genomic_accession=25297955, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  280621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280621, CurrentID=0, Status='live', Symbol='BC089491', Aliases='Selv', description='cDNA sequence BC089491', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28284652, end_position_on_the_genomic_accession=28291134, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18478, CurrentID=0, Status='live', Symbol='Pah', Aliases='AW106920', description='phenylalanine hydroxylase', other_designations='phe-4-monooxygenase', map_location='10 C2-D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=87521795, end_position_on_the_genomic_accession=87584137, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18479, CurrentID=0, Status='live', Symbol='Pak1', Aliases='AW045634, PAK-1, Paka', description='p21 protein (Cdc42/Rac)-activated kinase 1', other_designations='CDC42/RAC effector kinase PAK-A|P21 CDKN1A-activated kinase 1|activated protein kinase alpha|alpha-PAK|p21-activated kinase 1|p65-PAK', map_location='7 53.57 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97788047, end_position_on_the_genomic_accession=97912381, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18481, CurrentID=0, Status='live', Symbol='Pak3', Aliases='PAK-3, Pak65alpha, Pak65beta, Stk4', description='p21 protein (Cdc42/Rac)-activated kinase 3', other_designations='CDC42/RAC effector kinase PAK-B|beta-PAK|p21-activated kinase 3', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=143518368, end_position_on_the_genomic_accession=143797796, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  18483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18483, CurrentID=0, Status='live', Symbol='Palm', Aliases='', description='paralemmin', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79793351, end_position_on_the_genomic_accession=79820896, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18484, CurrentID=0, Status='live', Symbol='Pam', Aliases='PHM', description='peptidylglycine alpha-amidating monooxygenase', other_designations='PEPTIDYL-GLYCINE ALPHA-AMIDATING MONOOXYGENASE PRECURSOR (PAM)', map_location='1 47.76 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=97821091, end_position_on_the_genomic_accession=98095632, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  18488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18488, CurrentID=0, Status='live', Symbol='Cntn3', Aliases='Pang', description='contactin 3', other_designations='BIG-1|brain-derived immunoglobulin superfamily protein 1|plasmacytoma-associated neuronal glycoprotein', map_location='6 47.52 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=102163304, end_position_on_the_genomic_accession=102564620, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  18489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18489, CurrentID=0, Status='live', Symbol='Reg3b', Aliases='HIP, PAP1, Pap, REG-III', description='regenerating islet-derived 3 beta', other_designations='REG-3-beta|RegIII (beta)|pancreatitis-associated protein 1|reg III-beta|regenerating islet-derived protein III-beta', map_location='6 C', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=78370885, end_position_on_the_genomic_accession=78373466, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  280635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280635, CurrentID=0, Status='live', Symbol='Emilin3', Aliases='1110013O17Rik, EMILIN-3, EMILIN-5, EMILIN-T, Emilin5', description='elastin microfibril interfacer 3', other_designations='elastin microfibril interface located protein 3|elastin microfibril interface located protein 5|elastin microfibril interface-located protein 3|elastin microfibril interfacer 5', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=160906438, end_position_on_the_genomic_accession=160912339, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  280645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280645, CurrentID=0, Status='live', Symbol='B3gat2', Aliases='GlcAT-S, Glcats', description='beta-1,3-glucuronyltransferase 2 (glucuronosyltransferase S)', other_designations='UDP-glucuronosyltransferase S|UDP-glucuronosyltransferase-S|glcAT-D|glucuronosyltransferase S|glucuronosyltransferase-S', map_location='1 A4-B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=23761926, end_position_on_the_genomic_accession=23847865, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18503, CurrentID=0, Status='live', Symbol='Pax1', Aliases='Pax-1, hbs, hunchback, un, undulated, wt', description='paired box 1', other_designations='paired box gene 1|wavy tail', map_location='2 72.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=147361925, end_position_on_the_genomic_accession=147375049, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18504, CurrentID=0, Status='live', Symbol='Pax2', Aliases='Opdc, Pax-2', description='paired box 2', other_designations='optic disc coloboma|paired box gene 2', map_location='19 38.09 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44745695, end_position_on_the_genomic_accession=44837871, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18505, CurrentID=0, Status='live', Symbol='Pax3', Aliases='Pax-3, Sp, Splchl2, splotch', description='paired box 3', other_designations='PAX3/FKHR fusion|paired box gene 3', map_location='1 39.79 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=78101267, end_position_on_the_genomic_accession=78197547, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18506, CurrentID=0, Status='live', Symbol='Pax4', Aliases='Pax-4', description='paired box 4', other_designations='paired box gene 4', map_location='6 11.99 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=28442334, end_position_on_the_genomic_accession=28449340, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18507, CurrentID=0, Status='live', Symbol='Pax5', Aliases='BSAP, EBB-1, KLP, Pax-5', description='paired box 5', other_designations='B-cell-specific transcription factor|Pax5 variant 1|Pax5 variant 2|paired box gene 5', map_location='4 23.55 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44526269, end_position_on_the_genomic_accession=44710481, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18508, CurrentID=0, Status='live', Symbol='Pax6', Aliases='1500038E17Rik, AEY11, Dey, Gsfaey11, Pax-6, Sey', description='paired box 6', other_designations="Dickie's small eye|oculorhombin|paired box gene 6", map_location='2 55.31 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=105667518, end_position_on_the_genomic_accession=105697364, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  18509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18509, CurrentID=0, Status='live', Symbol='Pax7', Aliases='Pax-7', description='paired box 7', other_designations='paired box gene 7|paired box transcription factor PAX7', map_location='4 70.83 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139737059, end_position_on_the_genomic_accession=139833026, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18510, CurrentID=0, Status='live', Symbol='Pax8', Aliases='Pax-8', description='paired box 8', other_designations='paired box gene 8', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24420551, end_position_on_the_genomic_accession=24480092, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18511, CurrentID=0, Status='live', Symbol='Pax9', Aliases='Pax-9', description='paired box 9', other_designations='paired box gene 9', map_location='12 24.53 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=56690898, end_position_on_the_genomic_accession=56712824, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18514, CurrentID=0, Status='live', Symbol='Pbx1', Aliases='2310056B04Rik, D230003C07Rik, Pbx-1', description='pre B cell leukemia homeobox 1', other_designations='homeobox protein PBX1', map_location='1 75.95 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=168119364, end_position_on_the_genomic_accession=168432264, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18515, CurrentID=0, Status='live', Symbol='Pbx2', Aliases='AU043397, G17', description='pre B cell leukemia homeobox 2', other_designations='homeobox protein PBX2|pre B-cell leukemia transcription factor 2', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34592471, end_position_on_the_genomic_accession=34597341, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18516, CurrentID=0, Status='live', Symbol='Pbx3', Aliases='', description='pre B cell leukemia homeobox 3', other_designations='homeobox protein PBX3', map_location='2 22.84 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34171457, end_position_on_the_genomic_accession=34372045, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  280662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280662, CurrentID=0, Status='live', Symbol='Afm', Aliases='Alf, alpha-Alb', description='afamin', other_designations='alpha albumin|alpha-albumin', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90518949, end_position_on_the_genomic_accession=90553544, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  18519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18519, CurrentID=0, Status='live', Symbol='Kat2b', Aliases='A930006P13Rik, AI461839, AW536563, Pcaf, p/CAF', description='K(lysine) acetyltransferase 2B', other_designations='histone acetylase PCAF|histone acetyltransferase PCAF|lysine acetyltransferase 2B|p300/CBP-associated factor', map_location='17 27.86 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=53566853, end_position_on_the_genomic_accession=53672721, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  18521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18521, CurrentID=0, Status='live', Symbol='Pcbp2', Aliases='AW412548, Hnrpx, alphaCP-2', description='poly(rC) binding protein 2', other_designations='CBP|CTBP|alpha-CP2|heterogeneous nuclear ribonucleoprotein X|hnRNP X|putative heterogeneous nuclear ribonucleoprotein X', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102470222, end_position_on_the_genomic_accession=102500059, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  280667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280667, CurrentID=0, Status='live', Symbol='Adam1b', Aliases='Adam1, Ftna', description='a disintegrin and metallopeptidase domain 1b', other_designations='ADAM 1b|PH-30 alpha|a disintegrin and metalloproteinase domain 1b|fertilin alpha|fertilin subunit alpha-b', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121500097, end_position_on_the_genomic_accession=121502980, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  280668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280668, CurrentID=0, Status='live', Symbol='Adam1a', Aliases='Adam1, Ftna', description='a disintegrin and metallopeptidase domain 1a', other_designations='ADAM 1a|PH-30 alpha|a disintegrin and metalloproteinase domain 1a (fertilin alpha)|fertilin alpha|fertilin subunit alpha-a', map_location='5 61.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121518604, end_position_on_the_genomic_accession=121521695, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18526, CurrentID=0, Status='live', Symbol='Pcdh10', Aliases='6430521D13Rik, 6430703F07Rik, OL-pc, mKIAA1400', description='protocadherin 10', other_designations='OL-protocadherin', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=45378398, end_position_on_the_genomic_accession=45434579, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18530, CurrentID=0, Status='live', Symbol='Pcdh8', Aliases='1700080P15Rik, Papc', description='protocadherin 8', other_designations='arcadlin|paraxial protocadherin', map_location='14 42.76 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79766772, end_position_on_the_genomic_accession=79771312, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18534, CurrentID=0, Status='live', Symbol='Pck1', Aliases='AI265463, PEPCK, Pck-1', description='phosphoenolpyruvate carboxykinase 1, cytosolic', other_designations='PEPCK-C|phosphoenolpyruvate carboxykinase, cytosolic|phosphoenolpyruvate carboxylase', map_location='2 C1-qter', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173153073, end_position_on_the_genomic_accession=173159254, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18536, CurrentID=0, Status='live', Symbol='Pcm1', Aliases='2600002H09Rik, 9430077F19Rik, C030044G17Rik', description='pericentriolar material 1', other_designations='PCM-1|mPCM-1|pericentriolar material gene 1', map_location='8 23.89 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=41239641, end_position_on_the_genomic_accession=41334088, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  18537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18537, CurrentID=0, Status='live', Symbol='Pcmt1', Aliases='C79501, PIMT', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase 1', other_designations='L-isoaspartyl protein carboxyl methyltransferase|protein L-isoaspartyl/D-aspartyl methyltransferase|protein carboxyl methyltransferase|protein-beta-aspartate methyltransferase', map_location='10 2.26 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7629380, end_position_on_the_genomic_accession=7680973, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18538, CurrentID=0, Status='live', Symbol='Pcna', Aliases='', description='proliferating cell nuclear antigen', other_designations='cyclin', map_location='2 64.15 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132249286, end_position_on_the_genomic_accession=132253180, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18540, CurrentID=0, Status='live', Symbol='Pcna-ps2', Aliases='', description='proliferating cell nuclear antigen pseudogene 2', other_designations='', map_location='19 6.08 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=9283238, end_position_on_the_genomic_accession=9284495, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  18541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18541, CurrentID=0, Status='live', Symbol='Pcnt', Aliases='AW476095, C86676, KEN2, kendrin, m239Asp, m275Asp, Pcnt', description='pericentrin (kendrin)', other_designations='pericentrin 2|pericentrin-250|pericentrin-360', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76351254, end_position_on_the_genomic_accession=76444122, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  18542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18542, CurrentID=0, Status='live', Symbol='Pcolce', Aliases='Astt-2, Astt2, P14, PCPE-1', description='procollagen C-endopeptidase enhancer protein', other_designations='astroglial cell transcript 2|procollagen C-proteinase enhancer 1|procollagen C-proteinase enhancer protein|procollagen COOH-terminal proteinase enhancer 1|type 1 procollagen C-proteinase enhancer protein|type I procollagen COOH-terminal proteinase enhancer', map_location='5 76.59 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137605105, end_position_on_the_genomic_accession=137611899, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18545, CurrentID=0, Status='live', Symbol='Pcp2', Aliases='L7, Pcp-2', description='Purkinje cell protein 2 (L7)', other_designations='Purkinje cell-specific protein L7|protein PCD-5|retinal Purkinje cell protein 2', map_location='8 1.92 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3623369, end_position_on_the_genomic_accession=3625545, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18546, CurrentID=0, Status='live', Symbol='Pcp4', Aliases='P16Rimb19, Pcp-4, Pep19', description='Purkinje cell protein 4', other_designations='brain-specific antigen PCP-4|brain-specific polypeptide PEP-19', map_location='16 56.97 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96467606, end_position_on_the_genomic_accession=96525789, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18548, CurrentID=0, Status='live', Symbol='Pcsk1', Aliases='Nec-1, Nec1, PC1, PC3, Phpp-1, SPC3', description='proprotein convertase subtilisin/kexin type 1', other_designations='NEC 1|furin homolog|prohormone convertase 1|prohormone convertase 1/3|propeptide-processing protease|proprotein convertase 1', map_location='13 C2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=75089415, end_position_on_the_genomic_accession=75134861, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  18549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18549, CurrentID=0, Status='live', Symbol='Pcsk2', Aliases='6330411F23Rik, AI839700, Nec-2, Nec2, PC2, Phpp-2, SPC2', description='proprotein convertase subtilisin/kexin type 2', other_designations='KEX2-like endoprotease 2|NEC 2|prohormone convertase 2|proprotein convertase 2', map_location='2 70.89 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=143546133, end_position_on_the_genomic_accession=143816284, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18550, CurrentID=0, Status='live', Symbol='Furin', Aliases='9130404I01Rik, Fur, PACE, Pcsk3, SPC1', description='furin (paired basic amino acid cleaving enzyme)', other_designations='dibasic-processing enzyme|paired basic amino acid residue cleaving enzyme|prohormone convertase 3|proprotein convertase subtilisin/kexin type 3|subtilisin pro-protein processing enzyme', map_location='7 D1-E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80389194, end_position_on_the_genomic_accession=80405445, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  18551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18551, CurrentID=0, Status='live', Symbol='Pcsk4', Aliases='AI647044, L21221, NEC 3, NEC3, PC4, SPC5', description='proprotein convertase subtilisin/kexin type 4', other_designations='KEX2-like endoprotease 3|neuroendocrine convertase 3|prohormone convertase 3', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80321281, end_position_on_the_genomic_accession=80329909, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18552, CurrentID=0, Status='live', Symbol='Pcsk5', Aliases='PC5, PC6, SPC6, b2b1549Clo, b2b585Clo', description='proprotein convertase subtilisin/kexin type 5', other_designations='proprotein convertase 5|proprotein convertase 6|proprotein convertase PC5|subtilisin-like proprotein convertase 6|subtilisin/kexin-like protease PC5', map_location='19 12.86 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=17432832, end_position_on_the_genomic_accession=17837632, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  18553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18553, CurrentID=0, Status='live', Symbol='Pcsk6', Aliases='C86343, Pace4, SPC4', description='proprotein convertase subtilisin/kexin type 6', other_designations='paired basic amino acid cleaving system 4', map_location='7 35.36 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=65862136, end_position_on_the_genomic_accession=66050386, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  18554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18554, CurrentID=0, Status='live', Symbol='Pcsk7', Aliases='AA959856, PC7, SPC7', description='proprotein convertase subtilisin/kexin type 7', other_designations='prohormone convertase 7|prohormone convertase PC7|proprotein convertase 7|proprotein convertase PC7|subtilisin-like proprotein convertase 7|subtilisin/kexin-like protease PC7', map_location='9 25.32 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45906569, end_position_on_the_genomic_accession=45929726, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  18555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18555, CurrentID=0, Status='live', Symbol='Cdk16', Aliases='Crk5, Pctaire1, Pctk1', description='cyclin-dependent kinase 16', other_designations='PCTAIRE-motif protein kinase 1|cell division protein kinase 16|serine/threonine-protein kinase PCTAIRE-1', map_location='X 16.18 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20687959, end_position_on_the_genomic_accession=20699880, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  18557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18557, CurrentID=0, Status='live', Symbol='Cdk18', Aliases='AA682070, Ctk18, Pctk3', description='cyclin-dependent kinase 18', other_designations='PCTAIRE-motif protein kinase 3|cell division protein kinase 18|serine/threonine-protein kinase PCTAIRE-3', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132113547, end_position_on_the_genomic_accession=132139957, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18559, CurrentID=0, Status='live', Symbol='Pctp', Aliases='PC-TP, StarD2', description='phosphatidylcholine transfer protein', other_designations='START domain-containing protein 2|stAR-related lipid transfer protein 2', map_location='11 54.7 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=89983417, end_position_on_the_genomic_accession=90002914, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18563, CurrentID=0, Status='live', Symbol='Pcx', Aliases='Pc, Pcb', description='pyruvate carboxylase', other_designations='pyruvate decarboxylase|pyruvic carboxylase', map_location='19 4.07 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4510472, end_position_on_the_genomic_accession=4621752, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  18566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18566, CurrentID=0, Status='live', Symbol='Pdcd1', Aliases='Ly101, PD-1, Pdc1', description='programmed cell death 1', other_designations='mPD-1|programmed cell death 1 protein|protein PD-1', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=94038305, end_position_on_the_genomic_accession=94052553, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18567, CurrentID=0, Status='live', Symbol='Pdcd2', Aliases='AI528543, Pcd2, RP-8', description='programmed cell death 2', other_designations='zinc finger protein Rp-8', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15521575, end_position_on_the_genomic_accession=15527301, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18569, CurrentID=0, Status='live', Symbol='Pdcd4', Aliases='D19Ucla1, Ma3, Tis', description='programmed cell death 4', other_designations='protein MA-3|topoisomerase-inhibitor suppressed protein', map_location='19 48.73 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53892231, end_position_on_the_genomic_accession=53929861, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  18570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18570, CurrentID=0, Status='live', Symbol='Pdcd6', Aliases='AV299538, Alg2, MA-3, PS2, alg-2', description='programmed cell death 6', other_designations='ALG-257|PMP41|probable calcium-binding protein ALG-2', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74303121, end_position_on_the_genomic_accession=74317575, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18571, CurrentID=0, Status='live', Symbol='Pdcd6ip', Aliases='AI480591, AW544830, Aip1, Alix, C76364, Eig2, mKIAA1375', description='programmed cell death 6 interacting protein', other_designations='ALG-2-interacting protein 1|ALG-2-interacting protein X|Alg2-interacting protein 1|Alg2-interacting protein X|E2f1-inducible protein', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=113651744, end_position_on_the_genomic_accession=113708260, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  18572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18572, CurrentID=0, Status='live', Symbol='Pdcd11', Aliases='1110021I22Rik, ALG-4, Pdcd7, mKIAA0185', description='programmed cell death 11', other_designations='apoptosis-linked gene 4 protein|programmed cell death protein 11|programmed cell death protein 7', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47090749, end_position_on_the_genomic_accession=47131146, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  18573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18573, CurrentID=0, Status='live', Symbol='Pde1a', Aliases='AI987702, AW125737', description='phosphodiesterase 1A, calmodulin-dependent', other_designations='61 kDa Cam-PDE|cam-PDE 1A', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=79834453, end_position_on_the_genomic_accession=80196518, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  18574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18574, CurrentID=0, Status='live', Symbol='Pde1b', Aliases='Pde1b1', description='phosphodiesterase 1B, Ca2+-calmodulin dependent', other_designations='63 kDa Cam-PDE|cam-PDE 1B|phosphodiesterase 1B, Ca2+-calmodulin dependent, 63 kDa|phosphodiesterase 1B1, Ca2+-calmodulin dependent, 63 kDa', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103503034, end_position_on_the_genomic_accession=103530056, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18575, CurrentID=0, Status='live', Symbol='Pde1c', Aliases='', description='phosphodiesterase 1C', other_designations='cam-PDE 1C', map_location='6 27.65 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56069798, end_position_on_the_genomic_accession=56652516, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  18576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18576, CurrentID=0, Status='live', Symbol='Pde3b', Aliases='9830102A01Rik, AI847709', description='phosphodiesterase 3B, cGMP-inhibited', other_designations='CGI-PDE B|CGIPDE1|cyclic GMP-inhibited phosphodiesterase B', map_location='7 59.46 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114414700, end_position_on_the_genomic_accession=114537938, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  18577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18577, CurrentID=0, Status='live', Symbol='Pde4a', Aliases='D9Ertd60e, Dpde2', description='phosphodiesterase 4A, cAMP specific', other_designations='cAMP-specific phosphodiesterase PDE4A|cyclic AMP specific phosphodiesterase PDE4A5A|dunce', map_location='9 7.73 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21165683, end_position_on_the_genomic_accession=21213248, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  18578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18578, CurrentID=0, Status='live', Symbol='Pde4b', Aliases='Dpde4, R74983, dunce', description='phosphodiesterase 4B, cAMP specific', other_designations='', map_location='4 46.99 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=102087537, end_position_on_the_genomic_accession=102607262, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  18582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18582, CurrentID=0, Status='live', Symbol='Pde6d', Aliases='AI841218, PrBP/delta', description='phosphodiesterase 6D, cGMP-specific, rod, delta', other_designations='GMP-PDE delta', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86543013, end_position_on_the_genomic_accession=86582793, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18583, CurrentID=0, Status='live', Symbol='Pde7a', Aliases='AU015378, AW047537', description='phosphodiesterase 7A', other_designations='', map_location='3 5.36 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19223108, end_position_on_the_genomic_accession=19311374, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18584, CurrentID=0, Status='live', Symbol='Pde8a', Aliases='AI551852, Pde8', description='phosphodiesterase 8A', other_designations='mmPDE8', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81213804, end_position_on_the_genomic_accession=81333622, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  18585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18585, CurrentID=0, Status='live', Symbol='Pde9a', Aliases='PDE9A1', description='phosphodiesterase 9A', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31386212, end_position_on_the_genomic_accession=31476309, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  18587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18587, CurrentID=0, Status='live', Symbol='Pde6b', Aliases='Pdeb, r, rd, rd-1, rd1, rd10', description='phosphodiesterase 6B, cGMP, rod receptor, beta polypeptide', other_designations='GMP-PDE beta|cGMP-phosphodiesterase beta-subunit|phosphodiesterase, cGMP, rod receptor, beta polypeptide|retinal degeneration 1', map_location='5 53.07 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108388373, end_position_on_the_genomic_accession=108431743, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  18588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18588, CurrentID=0, Status='live', Symbol='Pde6g', Aliases='Pdeg, p3AT', description='phosphodiesterase 6G, cGMP-specific, rod, gamma', other_designations='GMP-PDE gamma|cGMP phosphodiesterase gamma|rod cyclic-GMP phosphodiesterase 6 gamma splice variant p3', map_location='11 84.14 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120447607, end_position_on_the_genomic_accession=120453500, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18590, CurrentID=0, Status='live', Symbol='Pdgfa', Aliases='', description='platelet derived growth factor, alpha', other_designations='PDGF subunit A|PDGF-1|platelet-derived growth factor A chain|platelet-derived growth factor alpha polypeptide', map_location='5 77.65 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138976381, end_position_on_the_genomic_accession=138999916, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18591, CurrentID=0, Status='live', Symbol='Pdgfb', Aliases='PDGF-B, Sis', description='platelet derived growth factor, B polypeptide', other_designations='PDGF subunit B|PDGF-2|c-sis|platelet-derived growth factor B chain|platelet-derived growth factor beta polypeptide|proto-oncogene c-Sis', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79995865, end_position_on_the_genomic_accession=80014808, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18595, CurrentID=0, Status='live', Symbol='Pdgfra', Aliases='AI115593, CD140a, Pdgfr-2', description='platelet derived growth factor receptor, alpha polypeptide', other_designations='Alpha platelet-derived growth factor receptor precursor (PDGF-R-alpha)|CD140 antigen-like family member A|PDGF alpha chain|PDGF-R-alpha|PDGFR-alpha|alpha-type platelet-derived growth factor receptor|platelet-derived growth factor alpha receptor', map_location='5 39.55 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=75151401, end_position_on_the_genomic_accession=75198211, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  18596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18596, CurrentID=0, Status='live', Symbol='Pdgfrb', Aliases='AI528809, CD140b, PDGFR-1, Pdgfr', description='platelet derived growth factor receptor, beta polypeptide', other_designations='CD140 antigen-like family member B|PDGF beta chain|PDGF-R-beta|PDGFR-beta|beta platelet-derived growth factor receptor|beta-type platelet-derived growth factor receptor|platelet-derived growth factor receptor 1|platelet-derived growth factor receptor beta variant 1', map_location='18 34.41 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61045150, end_position_on_the_genomic_accession=61085068, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  18597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18597, CurrentID=0, Status='live', Symbol='Pdha1', Aliases='Pdha-1', description='pyruvate dehydrogenase E1 alpha 1', other_designations='PDHE1-A type I|pyruvate dehydrogenase E1alpha subunit', map_location='X F3-F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=160122219, end_position_on_the_genomic_accession=160138336, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18598, CurrentID=0, Status='live', Symbol='Pdha2', Aliases='Pdhal', description='pyruvate dehydrogenase E1 alpha 2', other_designations='PDHE1-A type II|PYRUVATE DEHYDROGENASE E1 COMPONENT ALPHA SUBUNIT, TESTIS-SPECIFIC FORM, MITOCHONDRIAL PRECURSOR (PDHE1-A TYPE II)', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=141210004, end_position_on_the_genomic_accession=141212355, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18599, CurrentID=0, Status='live', Symbol='Padi1', Aliases='AV236283, Pdi1', description='peptidyl arginine deiminase, type I', other_designations='PAD type I|PROTEIN-ARGININE DEIMINASE TYPE I (PEPTIDYLARGININE DEIMINASE I)|Pad type 1|peptidylarginine deiminase I', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140812981, end_position_on_the_genomic_accession=140845778, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18600, CurrentID=0, Status='live', Symbol='Padi2', Aliases='Pdi, Pdi2, mKIAA0994', description='peptidyl arginine deiminase, type II', other_designations='PAD type II|peptidyl arginine deiminase, type II; PAD type II|peptidylarginine deiminase II|protein-arginine deiminase type II', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140906334, end_position_on_the_genomic_accession=140952586, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  18601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18601, CurrentID=0, Status='live', Symbol='Padi3', Aliases='AI507004, Pad3, Pdi3, mKIAA4171', description='peptidyl arginine deiminase, type III', other_designations='PAD type III|PROTEIN-ARGININE DEIMINASE TYPE III (PEPTIDYLARGININE DEIMINASE III)|peptidylarginine deiminase III', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140785369, end_position_on_the_genomic_accession=140810648, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18602, CurrentID=0, Status='live', Symbol='Padi4', Aliases='Pad4, Pdi4', description='peptidyl arginine deiminase, type IV', other_designations='PAD type IV|peptidylarginine deiminase IV|protein-arginine deiminase type IV', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140745508, end_position_on_the_genomic_accession=140774239, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18604, CurrentID=0, Status='live', Symbol='Pdk2', Aliases='', description='pyruvate dehydrogenase kinase, isoenzyme 2', other_designations='PDH kinase 2|[Pyruvate dehydrogenase (acetyl-transferring)] kinase isozyme 2, mitochondrial|[Pyruvate dehydrogenase [lipoamide]] kinase isozyme 2, mitochondrial|pyruvate dehydrogenase 2|pyruvate dehydrogenase kinase, isozyme 2', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95026258, end_position_on_the_genomic_accession=95041371, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18605, CurrentID=0, Status='live', Symbol='Enpp1', Aliases='4833416E15Rik, AI428932, C76301, CD203c, E-NPP 1, E-NPP1, Ly-41, M6S1, NPP1, Npps, PC-1, Pca, Pca-1, Pdnp1, ttw, twy', description='ectonucleotide pyrophosphatase/phosphodiesterase 1', other_designations='lymphocyte antigen 41|phosphodiesterase I/nucleotide pyrophosphatase 1|plasma-cell membrane glycoprotein PC-1|tiptoe walking', map_location='10 12.26 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24637914, end_position_on_the_genomic_accession=24712154, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  18606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18606, CurrentID=0, Status='live', Symbol='Enpp2', Aliases='ATX, E-NPP 2, Npps2, PD-Ialpha, Pdnp2, lysoPLD', description='ectonucleotide pyrophosphatase/phosphodiesterase 2', other_designations='autotaxin|extracellular lysophospholipase D|phosphodiesterase I/nucleotide pyrophosphatase 2', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=54838679, end_position_on_the_genomic_accession=54952887, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  18607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18607, CurrentID=0, Status='live', Symbol='Pdpk1', Aliases='Pdk1', description='3-phosphoinositide dependent protein kinase 1', other_designations='3-phosphoinositide dependent protein kinase-1|Pkb kinase|mPDK1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24073680, end_position_on_the_genomic_accession=24150922, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18609, CurrentID=0, Status='live', Symbol='Pdx1', Aliases='IDX-1, IPF-1, Ipf1, Mody4, STF-1, pdx-1', description='pancreatic and duodenal homeobox 1', other_designations='insulin promoter factor 1, homeodomain transcription factor|islet/duodenum homeobox 1|pancreatic and duodenal homeobox gene 1|somatostatin-transactivating factor 1', map_location='5 86.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147270131, end_position_on_the_genomic_accession=147275149, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18610, CurrentID=0, Status='live', Symbol='Pdyn', Aliases='Dyn', description='prodynorphin', other_designations='alpha-neoendorphin|beta-neoendorphin|beta-neoendorphin-dynorphin|big dynorphin|dynorphin A(1-13)|dynorphin A(1-17)|dynorphin A(1-8)|leu-enkephalin|leumorphin|preprodynorphin|rimorphin', map_location='2 63.19 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129686549, end_position_on_the_genomic_accession=129699938, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18611, CurrentID=0, Status='live', Symbol='Pea15a', Aliases='Mat1, PEA-15, Pea15, Pkcs15', description='phosphoprotein enriched in astrocytes 15A', other_designations='15 kDa phosphoprotein enriched in astrocytes|mammary transforming gene 1|phosphoprotein enriched in astrocytes 15', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172196728, end_position_on_the_genomic_accession=172206781, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18612, CurrentID=0, Status='live', Symbol='Etv4', Aliases='AW414408, Pea-3, Pea3', description='ets variant 4', other_designations='ETS variant protein 4|POLYOMAVIRUS ENHANCER ACTIVATOR 3 (PEA3 PROTEIN) (ETS TRANSLOCATION VARIANT 4)|ets variant gene 4 (E1A enhancer binding protein, E1AF)', map_location='11 65.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101769742, end_position_on_the_genomic_accession=101785310, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18613, CurrentID=0, Status='live', Symbol='Pecam1', Aliases='C85791, Cd31, PECAM-1, Pecam', description='platelet/endothelial cell adhesion molecule 1', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106654217, end_position_on_the_genomic_accession=106750623, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  18616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18616, CurrentID=0, Status='live', Symbol='Peg3', Aliases='AL022617, ASF-1, End4, Gcap4, Pw1, Zfp102, mKIAA0287', description='paternally expressed 3', other_designations='granule cell antiserum positive 4|paternally expressed gene 3', map_location='7 A2-B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6705960, end_position_on_the_genomic_accession=6730419, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18617, CurrentID=0, Status='live', Symbol='Rhox5', Aliases='AA409564, Pem', description='reproductive homeobox 5', other_designations='homeobox protein Pem|placenta and embryonic expression protein|placentae and embryos oncofetal|reproductive homeobox on X chromosome, 5|reproductive homeobox on chromosome X 5', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37754608, end_position_on_the_genomic_accession=37808878, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18618, CurrentID=0, Status='live', Symbol='Pemt', Aliases='PEAMT, PEMT2, Pempt, Pempt2', description='phosphatidylethanolamine N-methyltransferase', other_designations='', map_location='11 37.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59970614, end_position_on_the_genomic_accession=60046512, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18619, CurrentID=0, Status='live', Symbol='Penk', Aliases='AI326464, ENK, PPA1, Penk', description='preproenkephalin', other_designations='preproenkephalin 1|spermatogenic-specific proenkephalin', map_location='4 2.31 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=4133530, end_position_on_the_genomic_accession=4138965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18624, CurrentID=0, Status='live', Symbol='Pepd', Aliases='Pep-4, Pep4, dal', description='peptidase D', other_designations='X-pro dipeptidase|imidodipeptidase|peptidase 4|prolidase|proline dipeptidase', map_location='7 20.7 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34912407, end_position_on_the_genomic_accession=35044708, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  18626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18626, CurrentID=0, Status='live', Symbol='Per1', Aliases='Per, m-rigui, mPer1', description='period circadian clock 1', other_designations='circadian clock protein PERIOD 1|circadian pacemaker protein Rigui|period homolog 1', map_location='11 B', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69095140, end_position_on_the_genomic_accession=69109965, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  18627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18627, CurrentID=0, Status='live', Symbol='Per2', Aliases='mKIAA0347, mPer2', description='period circadian clock 2', other_designations='circadian clock protein PERIOD 2|period 2|period homolog 2', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91415979, end_position_on_the_genomic_accession=91459393, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  18628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18628, CurrentID=0, Status='live', Symbol='Per3', Aliases='2810049O06Rik, mPer3', description='period circadian clock 3', other_designations='circadian clock protein PERIOD 3|period homolog 3', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=151003655, end_position_on_the_genomic_accession=151044675, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  18630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18630, CurrentID=0, Status='live', Symbol='Pet2', Aliases='PC231, PC326, Pex3', description='plasmacytoma expressed transcript 2', other_designations='', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=89403848, end_position_on_the_genomic_accession=89409689, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18631, CurrentID=0, Status='live', Symbol='Pex11a', Aliases='PEX11alpha', description='peroxisomal biogenesis factor 11 alpha', other_designations='PEX11-alpha|peroxin-11A|peroxisomal biogenesis factor 11a|protein PEX11 homolog alpha', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79735957, end_position_on_the_genomic_accession=79743126, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18632, CurrentID=0, Status='live', Symbol='Pex11b', Aliases='PEX11beta, Pex11pbeta', description='peroxisomal biogenesis factor 11 beta', other_designations='PEX11-beta|peroxin-11B|peroxisomal biogenesis factor 11b|protein PEX11 homolog beta', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96635357, end_position_on_the_genomic_accession=96645381, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18633, CurrentID=0, Status='live', Symbol='Pex16', Aliases='', description='peroxisomal biogenesis factor 16', other_designations='peroxin-16|peroxisomal membrane protein PEX16', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92374676, end_position_on_the_genomic_accession=92381220, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18634, CurrentID=0, Status='live', Symbol='Pex7', Aliases='MmPEX7', description='peroxisomal biogenesis factor 7', other_designations='PTS2 receptor|peroxin-7|peroxisomal targeting signal 2 receptor|peroxisome biogenesis factor 7', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19860090, end_position_on_the_genomic_accession=19907674, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18635, CurrentID=0, Status='live', Symbol='pf', Aliases='pd', description='pupoid fetus', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18636, CurrentID=0, Status='live', Symbol='Cfp', Aliases='BCFG, Pfc', description='complement factor properdin', other_designations='complement factor P|properdin factor, complement', map_location='X A3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20925535, end_position_on_the_genomic_accession=20931525, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18637, CurrentID=0, Status='live', Symbol='Pfdn2', Aliases='ESTM27, W48336', description='prefoldin 2', other_designations='', map_location='1 79.35 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171345699, end_position_on_the_genomic_accession=171358170, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18639, CurrentID=0, Status='live', Symbol='Pfkfb1', Aliases='PFK/FBPase 1', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 1', other_designations='6PF-2-K/Fru-2,6-P2ase liver isozyme|PFK-2/FBPase-2 gene A', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=150588257, end_position_on_the_genomic_accession=150643878, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  18640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18640, CurrentID=0, Status='live', Symbol='Pfkfb2', Aliases='4930568D07Rik', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 2', other_designations='6PF-2-K/Fru-2,6-P2ase 2|6PF-2-K/Fru-2,6-P2ase heart-type isozyme|PFK-2/FBPase-2 gene B|PFK/FBPase 2', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130689043, end_position_on_the_genomic_accession=130729253, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  18641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18641, CurrentID=0, Status='live', Symbol='Pfkl', Aliases='AA407869, ATP-PFK, PFK-L', description='phosphofructokinase, liver, B-type', other_designations='6-phosphofructokinase type B|6-phosphofructokinase, liver type|ATP-PFK|PFK-B|phosphofructo-1-kinase isozyme B|phosphofructokinase 1|phosphohexokinase', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77986947, end_position_on_the_genomic_accession=78009796, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  18642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18642, CurrentID=0, Status='live', Symbol='Pfkm', Aliases='AI131669, ATP-PFK, PFK-A, PFK-M, Pfk-4, Pfk4, Pfka, Pfkx', description='phosphofructokinase, muscle', other_designations='6-phosphofructokinase type A|6-phosphofructokinase, muscle type|ATP-PFK|phosphofructo-1-kinase isozyme A|phosphofructokinase 1|phosphofructokinase 4|phosphofructokinase M variant 3/4|phosphofructokinase, polypeptide X|phosphofructokinase-1 A isozyme|phosphofructokinase-M (connecting peptide)|phosphohexokinase', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98092589, end_position_on_the_genomic_accession=98132447, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  18643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18643, CurrentID=0, Status='live', Symbol='Pfn1', Aliases='Pfn', description='profilin 1', other_designations='actin binding protein|profilin I', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70651847, end_position_on_the_genomic_accession=70654650, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18645, CurrentID=0, Status='live', Symbol='Pfn2', Aliases='Pfn', description='profilin 2', other_designations='profilin II', map_location='3 28.15 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=57841895, end_position_on_the_genomic_accession=57847757, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18646, CurrentID=0, Status='live', Symbol='Prf1', Aliases='Pfn, Pfp, Prf-1', description='perforin 1 (pore forming protein)', other_designations='P1|cytolysin|lymphocyte pore-forming protein|pore forming protein', map_location='10 32.18 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61297680, end_position_on_the_genomic_accession=61304263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18647, CurrentID=0, Status='live', Symbol='Cdk14', Aliases='Pftk1, mKIAA0834', description='cyclin-dependent kinase 14', other_designations='cell division protein kinase 14|serine/threonine-protein kinase PFTAIRE-1', map_location='5 2.61 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=4803384, end_position_on_the_genomic_accession=5380251, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  18648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18648, CurrentID=0, Status='live', Symbol='Pgam1', Aliases='2310050F24Rik, Pgam-1', description='phosphoglycerate mutase 1', other_designations='BPG-dependent PGAM 1|PGAM-B|phosphoglycerate mutase isozyme B', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41911871, end_position_on_the_genomic_accession=41918665, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  330788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330788, CurrentID=0, Status='live', Symbol='Zfp866', Aliases='9830167H18Rik, AI875089, C85843, D330038O06Rik', description='zinc finger protein 866', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69761324, end_position_on_the_genomic_accession=69774911, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  243407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243407, CurrentID=0, Status='live', Symbol='C130060K24Rik', Aliases='', description='RIKEN cDNA C130060K24 gene', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=65381294, end_position_on_the_genomic_accession=65458150, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18654, CurrentID=0, Status='live', Symbol='Pgf', Aliases='AI854365, PIGF, Plgf', description='placental growth factor', other_designations='', map_location='12 D', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85166637, end_position_on_the_genomic_accession=85177785, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18655, CurrentID=0, Status='live', Symbol='Pgk1', Aliases='Pgk-1', description='phosphoglycerate kinase 1', other_designations='', map_location='X C-D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106187100, end_position_on_the_genomic_accession=106203699, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18656, CurrentID=0, Status='live', Symbol='Pgk1-ps1', Aliases='Pgk-1ps1', description='phosphoglycerate kinase 1, pseudogene 1', other_designations='', map_location='16 46.57 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=81909003, end_position_on_the_genomic_accession=81909715, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  68646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68646, CurrentID=0, Status='live', Symbol='Nadk2', Aliases='MNADK, Nadkd1', description='NAD kinase 2, mitochondrial', other_designations='NAD kinase domain-containing protein 1, mitochondrial', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9071260, end_position_on_the_genomic_accession=9110496, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18663, CurrentID=0, Status='live', Symbol='Pgk2', Aliases='Pgk-2, Tcp-2', description='phosphoglycerate kinase 2', other_designations='phosphoglycerate kinase, testis specific', map_location='17 19.44 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40207018, end_position_on_the_genomic_accession=40208609, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18667, CurrentID=0, Status='live', Symbol='Pgr', Aliases='9930019P03Rik, BB114106, ENSMUSG00000074510, NR3C3, PR, PR-A, PR-B', description='progesterone receptor', other_designations='nuclear receptor subfamily 3 group C member 3', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=8894736, end_position_on_the_genomic_accession=8968611, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18669, CurrentID=0, Status='live', Symbol='Abcb1b', Aliases='Abcb1, Mdr1, Mdr1b, Pgy-1, Pgy1, mdr', description='ATP-binding cassette, sub-family B (MDR/TAP), member 1B', other_designations='ATP-binding cassette sub-family B member 1|P glycoprotein 1|multidrug resistance protein 1', map_location='5 3.43 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8798147, end_position_on_the_genomic_accession=8866315, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  18670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18670, CurrentID=0, Status='live', Symbol='Abcb4', Aliases='Mdr2, Pgy-2, Pgy2, mdr-2', description='ATP-binding cassette, sub-family B (MDR/TAP), member 4', other_designations='ATP-binding cassette sub-family B member 4|P glycoprotein 2|P-glycoprotein 3|multidrug resistance protein 2|multidrug resistance protein 3', map_location='5 3.43 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8893598, end_position_on_the_genomic_accession=8959226, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  18671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18671, CurrentID=0, Status='live', Symbol='Abcb1a', Aliases='Abcb4, Evi32, Mdr1a, Mdr3, P-gp, Pgp, Pgy-3, Pgy3, mdr-3', description='ATP-binding cassette, sub-family B (MDR/TAP), member 1A', other_designations='ATP-binding cassette sub-family B member 1A|ATP-binding cassette, sub-family B (MDR/TAP), member 4|ATP-binding cassette, subfamily B, member 1A|P glycoprotein 3|P-glycoprotein 3|ecotropic viral integration site 32|multi-drug resistance 3|multidrug resistance protein 3|multiple drug resistant 1a', map_location='5 3.43 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8567091, end_position_on_the_genomic_accession=8748575, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  18673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18673, CurrentID=0, Status='live', Symbol='Phb', Aliases='Bap32', description='prohibitin', other_designations='B-cell receptor-associated protein 32', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95666957, end_position_on_the_genomic_accession=95680773, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18674, CurrentID=0, Status='live', Symbol='Slc25a3', Aliases='5730556H19Rik, PTP, Phc', description='solute carrier family 25 (mitochondrial carrier, phosphate carrier), member 3', other_designations='phosphate carrier, mitochondrial|phosphate transport protein|solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 3|solute carrier family 25 member 3', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=91116578, end_position_on_the_genomic_accession=91124052, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18675, CurrentID=0, Status='live', Symbol='Phex', Aliases='Gy, HPDR, HPDR1, Hyp, PEX', description='phosphate regulating endopeptidase homolog, X-linked', other_designations='Metalloendopeptidase homolog PEX (Phosphate regulating neutral endopeptidase) (X-linked hypophosphatemia protein) (HYP) (Vitamin D-resistant hypophosphatemic rickets protein)|X-linked hypophosphatemia protein|X-linked phosphate regulating endopeptidase|phosphate regulating gene with homologies to endopeptidases on the X chromosome (hypophosphatemia, vitamin D resistant rickets)|phosphate regulating neutral endopeptidase|vitamin D-resistant hypophosphatemic rickets protein', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=157162075, end_position_on_the_genomic_accession=157416551, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  18676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18676, CurrentID=0, Status='live', Symbol='Phf2', Aliases='GRC5', description='PHD finger protein 2', other_designations='', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48801750, end_position_on_the_genomic_accession=48870885, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  18679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18679, CurrentID=0, Status='live', Symbol='Phka1', Aliases='5330411D17, 9830108K24Rik, Phka', description='phosphorylase kinase alpha 1', other_designations='phosphorylase kinase alpha M subunit', map_location='X 45.47 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102513975, end_position_on_the_genomic_accession=102644278, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  18682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18682, CurrentID=0, Status='live', Symbol='Phkg1', Aliases='Phkg', description='phosphorylase kinase gamma 1', other_designations='Phosphorylase B kinase gamma catalytic chain, skeletal muscle isoform (Phosphorylase kinase gamma subunit 1)|phosphorylase kinase subunit gamma-1|serine/threonine-protein kinase PHKG1', map_location='5 68.26 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129863435, end_position_on_the_genomic_accession=129879083, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18685, CurrentID=0, Status='live', Symbol='Phtf1', Aliases='AU041898, AW049785, Phft, Phtf', description='putative homeodomain transcription factor 1', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103968110, end_position_on_the_genomic_accession=104007490, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  18693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18693, CurrentID=0, Status='live', Symbol='Pick1', Aliases='Prkcabp', description='protein interacting with C kinase 1', other_designations='protein kinase C-alpha-binding protein|protein that interacts with C kinase 1', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79229166, end_position_on_the_genomic_accession=79249466, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18700, CurrentID=0, Status='live', Symbol='Piga', Aliases='AI194334, Pig-a', description='phosphatidylinositol glycan anchor biosynthesis, class A', other_designations='glcNAc-PI synthesis protein|phosphatidylinositol glycan, class A|phosphatidylinositol-glycan biosynthesis class A protein|phosphatidylinositolglycan class A', map_location='X F3-F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164419680, end_position_on_the_genomic_accession=164433916, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18701, CurrentID=0, Status='live', Symbol='Pigf', Aliases='', description='phosphatidylinositol glycan anchor biosynthesis, class F', other_designations='PIG-F|phosphatidylinositol glycan, class F', map_location='17 E4-E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=86997258, end_position_on_the_genomic_accession=87025471, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  18703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18703, CurrentID=0, Status='live', Symbol='Pigr', Aliases='', description='polymeric immunoglobulin receptor', other_designations='phosphatidylinositol glycan, class R|poly-Ig receptor', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130826684, end_position_on_the_genomic_accession=130852249, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18704, CurrentID=0, Status='live', Symbol='Pik3c2a', Aliases='Cpk-m, PI3KC2', description='phosphatidylinositol 3-kinase, C2 domain containing, alpha polypeptide', other_designations='PI3K-C2-alpha|p170|phosphatidylinositol-4-phosphate 3-kinase C2 domain-containing subunit alpha|phosphoinositide 3-kinase-C2-alpha|ptdIns-3-kinase C2 subunit alpha', map_location='7 61.62 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=116337265, end_position_on_the_genomic_accession=116444557, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  18705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18705, CurrentID=0, Status='live', Symbol='Pik3c2g', Aliases='C80387, PI3K-C2-gamma', description='phosphatidylinositol 3-kinase, C2 domain containing, gamma polypeptide', other_designations='ptdIns-3-kinase C2 subunit gamma', map_location='6 69.7 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=139715859, end_position_on_the_genomic_accession=139969284, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  18706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18706, CurrentID=0, Status='live', Symbol='Pik3ca', Aliases='6330412C24Rik, caPI3K, p110, p110alpha', description='phosphatidylinositol 3-kinase, catalytic, alpha polypeptide', other_designations='PI3-kinase subunit alpha|PI3K-alpha|PI3Kalpha|phosphatidylinositol 4,5-bisphosphate 3-kinase 110 kDa catalytic subunit alpha|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit alpha|phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit alpha isoform|phosphoinositide-3-kinase catalytic alpha polypeptide|ptdIns-3-kinase subunit alpha|ptdIns-3-kinase subunit p110-alpha|serine/threonine protein kinase PIK3CA', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32397055, end_position_on_the_genomic_accession=32468486, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  18707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18707, CurrentID=0, Status='live', Symbol='Pik3cd', Aliases='2410099E07Rik, 2610208K16Rik, AW545373, p110delta', description='phosphatidylinositol 3-kinase catalytic delta polypeptide', other_designations='PI3-kinase subunit delta|PI3K-delta|PI3Kdelta|catalytic phosphatidylinositol 3-kinase delta|p100 delta|phosphatidylinositol 3-kinase catalytic subunit delta|phosphatidylinositol 3-kinase p110 delta|phosphatidylinositol 4,5-bisphosphate 3-kinase 110 kDa catalytic subunit delta|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit delta|phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit delta isoform|phosphoinositide 3 kinase P110delta|ptdIns-3-kinase subunit delta|ptdIns-3-kinase subunit p110-delta', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149649168, end_position_on_the_genomic_accession=149702566, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  18708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18708, CurrentID=0, Status='live', Symbol='Pik3r1', Aliases='PI3K, p50alpha, p55alpha, p85alpha', description='phosphatidylinositol 3-kinase, regulatory subunit, polypeptide 1 (p85 alpha)', other_designations='PI3-kinase subunit p85-alpha|PI3K regulatory subunit alpha|phosphatidylinositol 3-kinase 85 kDa regulatory subunit alpha', map_location='13 53.92 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=101680761, end_position_on_the_genomic_accession=101768217, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  18709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18709, CurrentID=0, Status='live', Symbol='Pik3r2', Aliases='p85beta', description='phosphatidylinositol 3-kinase, regulatory subunit, polypeptide 2 (p85 beta)', other_designations='PI3-kinase regulatory subunit beta|PI3-kinase subunit p85-beta|PI3K regulatory subunit beta|phosphatidylinositol 3-kinase 85 kDa regulatory subunit beta|ptdIns-3-kinase regulatory subunit beta|ptdIns-3-kinase regulatory subunit p85-beta', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70768178, end_position_on_the_genomic_accession=70776728, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  18710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18710, CurrentID=0, Status='live', Symbol='Pik3r3', Aliases='AA414954, p55pik', description='phosphatidylinositol 3 kinase, regulatory subunit, polypeptide 3 (p55)', other_designations='PI3-kinase regulatory subunit gamma|PI3-kinase subunit p55-gamma|PI3K regulatory subunit gamma|phosphatidylinositol 3-kinase 55 kDa regulatory subunit gamma|ptdIns-3-kinase regulatory subunit gamma|ptdIns-3-kinase regulatory subunit p55-gamma', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116221528, end_position_on_the_genomic_accession=116303056, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18711, CurrentID=0, Status='live', Symbol='Pikfyve', Aliases='5230400C17Rik, Pip5k, Pip5k3, Pipk5k3, PipkIII, p235', description='phosphoinositide kinase, FYVE finger containing', other_designations='1-phosphatidylinositol-3-phosphate 5-kinase|FYVE finger-containing phosphoinositide kinase|PIPkin-III|PtdIns(4)P-5-kinase|phosphatidylinositol 3-phosphate 5-kinase type III|phosphatidylinositol-3-phosphate 5-kinase type III|phosphatidylinositol-3-phosphate/phosphatidylinositol 5-kinase, type III|phosphatidylinositol-4-phosphate 5-kinase, type III|type III PIP kinase', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65186639, end_position_on_the_genomic_accession=65278696, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  18712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18712, CurrentID=0, Status='live', Symbol='Pim1', Aliases='Pim-1', description='proviral integration site 1', other_designations='pim-1 oncogene|proto-oncogene serine/threonine-protein kinase pim-1', map_location='17 15.38 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29491045, end_position_on_the_genomic_accession=29495465, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18715, CurrentID=0, Status='live', Symbol='Pim2', Aliases='DXCch3, Pim-2', description='proviral integration site 2', other_designations='', map_location='X 3.55 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7878306, end_position_on_the_genomic_accession=7883432, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18716, CurrentID=0, Status='live', Symbol='Pip', Aliases='GCDFP-15, GP17, mSMGP', description='prolactin induced protein', other_designations='14 kDa submandibular gland protein|SMGP|gross cystic disease fluid protein 15|prolactin-induced protein', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41847549, end_position_on_the_genomic_accession=41852062, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18717, CurrentID=0, Status='live', Symbol='Pip5k1c', Aliases='AI115456, AI835305, Pip5kIgamma', description='phosphatidylinositol-4-phosphate 5-kinase, type 1 gamma', other_designations='PIP5K1-gamma|phosphatidylinositol 4-phosphate 5-kinase type I gamma|phosphatidylinositol-4-phosphate 5-kinase type I gamma|phosphatidylinositol-4-phosphate 5-kinase type-1 gamma|ptdIns(4)P-5-kinase 1 gamma', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81292962, end_position_on_the_genomic_accession=81319974, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  18718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18718, CurrentID=0, Status='live', Symbol='Pip4k2a', Aliases='AW742916, Pip5k2a', description='phosphatidylinositol-5-phosphate 4-kinase, type II, alpha', other_designations='1-phosphatidylinositol 5-phosphate 4-kinase 2-alpha|1-phosphatidylinositol-5-phosphate 4-kinase 2-alpha|PI(5)P 4-kinase type II alpha|PIP4KII-alpha|diphosphoinositide kinase 2-alpha|phosphatidylinositol 5-phosphate 4-kinase type II alpha|phosphatidylinositol-4-phosphate 5-kinase, type II, alpha|phosphatidylinositol-5-phosphate 4-kinase type-2 alpha|ptdIns(5)P-4-kinase isoform 2-alpha', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=18842255, end_position_on_the_genomic_accession=18998527, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18719, CurrentID=0, Status='live', Symbol='Pip5k1b', Aliases='PI4P5K-I[b], PIP5K1-beta, PIP5KIbeta, Pipk5b, STM7', description='phosphatidylinositol-4-phosphate 5-kinase, type 1 beta', other_designations='phosphatidylinositol 4-phosphate 5-kinase type I beta|phosphatidylinositol-4-phosphate 5-kinase type I beta|phosphatidylinositol-4-phosphate 5-kinase type-1 beta|ptdIns(4)P-5-kinase 1 beta', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24294794, end_position_on_the_genomic_accession=24555827, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18720, CurrentID=0, Status='live', Symbol='Pip5k1a', Aliases='PI4P5K-I[a], PIP5K1-alpha, PIP5KIalpha, Pipk5a', description='phosphatidylinositol-4-phosphate 5-kinase, type 1 alpha', other_designations='68 kDa type I phosphatidylinositol 4-phosphate 5-kinase|68 kDa type I phosphatidylinositol-4-phosphate 5-kinase|phosphatidylinositol 4-phosphate 5-kinase type I alpha|phosphatidylinositol-4-phosphate 5-kinase type I alpha|phosphatidylinositol-4-phosphate 5-kinase type-1 alpha|ptdIns(4)P-5-kinase 1 alpha', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95058530, end_position_on_the_genomic_accession=95106930, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18722, CurrentID=0, Status='live', Symbol='Pira1', Aliases='6M21, Ly89, PIR-A1, Pir', description='paired-Ig-like receptor A1', other_designations='paired immunoglobulin-like receptor', map_location='7 2.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3731633, end_position_on_the_genomic_accession=3739870, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18725, CurrentID=0, Status='live', Symbol='Pira2', Aliases='6M23', description='paired-Ig-like receptor A2', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3836810, end_position_on_the_genomic_accession=3845099, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18726, CurrentID=0, Status='live', Symbol='Lilra6', Aliases='7M1, Pira3', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 6', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3908277, end_position_on_the_genomic_accession=3915586, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18729, CurrentID=0, Status='live', Symbol='Pira6', Aliases='12M1, Pira10, p91B', description='paired-Ig-like receptor A6', other_designations='killer cell inhibitory receptor-like protein p91B|paired-Ig-like receptor A10', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4274124, end_position_on_the_genomic_accession=4284827, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18733, CurrentID=0, Status='live', Symbol='Pirb', Aliases='Gp91, LIR-3, Lilrb3, PIR-B', description='paired Ig-like receptor B', other_designations='cell-surface glycoprotein p91|leukocyte immunoglobulin-like receptor 3|leukocyte immunoglobulin-like receptor, subfamily B (with TM and ITIM domains), member 3|paired immunoglobulin-like receptor B', map_location='7 2.14 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3712505, end_position_on_the_genomic_accession=3720456, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18736, CurrentID=0, Status='live', Symbol='Pou1f1', Aliases='GHF-1, Hmp1, Pit-1, Pit1, Pit1-rs1, dw, dwarf', description='POU domain, class 1, transcription factor 1', other_designations='POU domain, class 1, transcription factor 1 (Pit1, growth hormone factor 1)|growth hormone factor 1|pituitary specific transcription factor 1, related sequence 1|pituitary-specific positive transcription factor 1 variant w', map_location='16 37.25 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=65520529, end_position_on_the_genomic_accession=65534146, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18738, CurrentID=0, Status='live', Symbol='Pitpna', Aliases='Pitpn, vb, vibrator', description='phosphatidylinositol transfer protein, alpha', other_designations='PI-TP-alpha|Pitp alpha|ptdIns transfer protein alpha|ptdInsTP alpha', map_location='11 45.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75588108, end_position_on_the_genomic_accession=75628778, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18739, CurrentID=0, Status='live', Symbol='Pitpnm1', Aliases='DRES9, Mpt-1, Nir-2, PITPnm 1, Pitpnm, R75447, Rd9, RdgB, RdgB1', description='phosphatidylinositol transfer protein, membrane-associated 1', other_designations='drosophila retinal degeneration B homolog 1|pyk2 N-terminal domain-interacting receptor 2|retinal degeneration 9', map_location='19 3.81 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4099998, end_position_on_the_genomic_accession=4113965, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  18740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18740, CurrentID=0, Status='live', Symbol='Pitx1', Aliases='Bft, P-OTX, Potx, Ptx1', description='paired-like homeodomain transcription factor 1', other_designations='hindlimb-expressed homeobox protein backfoot|homeobox protein P-OTX|homeobox protein PITX1|pituitary OTX-related factor', map_location='13 30.06 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55825046, end_position_on_the_genomic_accession=55836194, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18741, CurrentID=0, Status='live', Symbol='Pitx2', Aliases='9430085M16Rik, Brx1, Brx1b, Munc30, Otlx2, Ptx2, Rieg', description='paired-like homeodomain transcription factor 2', other_designations='ALL1-responsive protein ARP1|BRX1 homeoprotein|orthodenticle-like homeobox 2|paired-like homeodomain transcription factor Munc 30|solurshin', map_location='3 57.84 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129199878, end_position_on_the_genomic_accession=129219594, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  18742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18742, CurrentID=0, Status='live', Symbol='Pitx3', Aliases='Ptx3, ak', description='paired-like homeodomain transcription factor 3', other_designations='Pituitary homeobox 3 (Homeobox protein PITX3)|aphakia|homeobox protein PITX3', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46135294, end_position_on_the_genomic_accession=46148325, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18744, CurrentID=0, Status='live', Symbol='Pja1', Aliases='praja-1', description='praja ring finger 1, E3 ubiquitin protein ligase', other_designations='praja1, RING-H2 motif containing', map_location='X 43.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=99465734, end_position_on_the_genomic_accession=99471268, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18745, CurrentID=0, Status='live', Symbol='pk', Aliases='', description='plucked', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18746, CurrentID=0, Status='live', Symbol='Pkm', Aliases='AA414905, AL024370, AL024424, Pk-2, Pk-3, Pk32, Pkm', description='pyruvate kinase, muscle', other_designations='pyruvate kinase 3|pyruvate kinase isozymes M1/M2|pyruvate kinase muscle isozyme', map_location='9 32.03 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59656368, end_position_on_the_genomic_accession=59679375, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  18747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18747, CurrentID=0, Status='live', Symbol='Prkaca', Aliases='PKCD, Pkaca', description='protein kinase, cAMP dependent, catalytic, alpha', other_designations='PKA C-alpha|sperm cAMP-dependent protein kinase catalytic subunit', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83972978, end_position_on_the_genomic_accession=83996444, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18749, CurrentID=0, Status='live', Symbol='Prkacb', Aliases='CbPKA, Pkacb', description='protein kinase, cAMP dependent, catalytic, beta', other_designations='Cbeta2 PKA|PKA C-beta|cAMP-dependent protein kinase C beta', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146729579, end_position_on_the_genomic_accession=146812946, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  18750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18750, CurrentID=0, Status='live', Symbol='Prkca', Aliases='AI875142, Pkca', description='protein kinase C, alpha', other_designations='PKC-A|PKC-alpha|protein kinase c-alpha', map_location='11 70.8 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107933387, end_position_on_the_genomic_accession=108343888, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  18751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18751, CurrentID=0, Status='live', Symbol='Prkcb', Aliases='PKC-B, PKC-Beta, Pkcb1, Prkcb2, Prkcb', description='protein kinase C, beta', other_designations='protein kinase C beta-II|protein kinase C, beta 1', map_location='7 65.75 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122289078, end_position_on_the_genomic_accession=122634402, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18752, CurrentID=0, Status='live', Symbol='Prkcg', Aliases='PKCgamma, Pkcc, Prkcc', description='protein kinase C, gamma', other_designations='PKC-gamma', map_location='7 1.93 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3303532, end_position_on_the_genomic_accession=3331099, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18753, CurrentID=0, Status='live', Symbol='Prkcd', Aliases='AI385711, D14Ertd420e, PKC[d], PKCdelta, Pkcd', description='protein kinase C, delta', other_designations='nPKC-delta|protein kinase C delta variant IX|protein kinase C, delta IV|protein kinase C, delta V|protein kinase C[d]|tyrosine-protein kinase PRKCD', map_location='14 18.82 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30595353, end_position_on_the_genomic_accession=30626414, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  18754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18754, CurrentID=0, Status='live', Symbol='Prkce', Aliases='5830406C15Rik, PKC[e], PKCepsilon, Pkce, R75156', description='protein kinase C, epsilon', other_designations='nPKC-epsilon', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=86166971, end_position_on_the_genomic_accession=86657919, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  18755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18755, CurrentID=0, Status='live', Symbol='Prkch', Aliases='Pkch', description='protein kinase C, eta', other_designations='PKC-L|nPKC-eta', map_location='12 C3-D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73585041, end_position_on_the_genomic_accession=73778185, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  18759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18759, CurrentID=0, Status='live', Symbol='Prkci', Aliases='2310021H13Rik, AI427505, PKClambda, Pkci, Pkcl, Prkcl, aPKClambda, mKIAA4165', description='protein kinase C, iota', other_designations='aPKC-lambda/iota|atypical protein kinase C-lambda/iota|nPKC-iota|protein kinase C, lambda', map_location='3 14.65 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30995742, end_position_on_the_genomic_accession=31052741, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18760, CurrentID=0, Status='live', Symbol='Prkd1', Aliases='PKD, PKD1, Pkcm, Prkcm', description='protein kinase D1', other_designations='nPKC-D1|nPKC-mu|protein kinase C mu type|protein kinase C, mu|protein kinase D', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=50341226, end_position_on_the_genomic_accession=50649248, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  18761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18761, CurrentID=0, Status='live', Symbol='Prkcq', Aliases='A130035A12Rik, AW494342, PKC-0, PKC-theta, PKCtheta, Pkcq', description='protein kinase C, theta', other_designations='PKC theta|nPKC-theta', map_location='2 8.42 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=11172382, end_position_on_the_genomic_accession=11301226, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  18762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18762, CurrentID=0, Status='live', Symbol='Prkcz', Aliases='AI098070, C80388, Pkcz, R74924, aPKCzeta, nPKC-zeta, zetaPKC', description='protein kinase C, zeta', other_designations='atypical protein kinase C|protein kinase C zeta subspecies', map_location='4 86.17 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155260118, end_position_on_the_genomic_accession=155361428, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  18763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18763, CurrentID=0, Status='live', Symbol='Pkd1', Aliases='PC1, mFLJ00285', description='polycystic kidney disease 1 homolog', other_designations='autosomal dominant polycystic kidney disease 1 protein homolog|polycystic kidney disease 1 homolog; polycystin-1|polycystin 1', map_location='17 12.4 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24549950, end_position_on_the_genomic_accession=24596514, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  18764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18764, CurrentID=0, Status='live', Symbol='Pkd2', Aliases='C030034P18Rik, PC2, TRPP2', description='polycystic kidney disease 2', other_designations='polycystic kidney disease 2 protein homolog|polycystin 2|transient receptor potential cation channel subfamily P member 2', map_location='5 50.68 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104457224, end_position_on_the_genomic_accession=104505819, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  18766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18766, CurrentID=0, Status='live', Symbol='Pkdrej', Aliases='', description='polycystic kidney disease (polycystin) and REJ (sperm receptor for egg jelly homolog, sea urchin)', other_designations='PKD and REJ homolog|receptor for egg jelly-like protein', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85814676, end_position_on_the_genomic_accession=85821733, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18767, CurrentID=0, Status='live', Symbol='Pkia', Aliases='AI415001, PKIalpha', description='protein kinase inhibitor, alpha', other_designations='PKI-alpha|cAMP-dependent protein kinase inhibitor, muscle/brain isoform', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=7366604, end_position_on_the_genomic_accession=7445366, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18768, CurrentID=0, Status='live', Symbol='Pkib', Aliases='Pkibeta, Prkacn2', description='protein kinase inhibitor beta, cAMP dependent, testis specific', other_designations='PKI-beta|cAMP-dependent protein kinase inhibitor, testis isoform|protein kinase, cAMP-dependent, catalytic, inhibitor 2', map_location='10 29.22 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=57631981, end_position_on_the_genomic_accession=57741112, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18769, CurrentID=0, Status='live', Symbol='Pkig', Aliases='PKIgamma', description='protein kinase inhibitor, gamma', other_designations='', map_location='2 84.4 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163658386, end_position_on_the_genomic_accession=163726158, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18770, CurrentID=0, Status='live', Symbol='Pklr', Aliases='Pk-1, Pk1, R-PK', description='pyruvate kinase liver and red blood cell', other_designations='L-PK|pyruvate kinase isozymes L/R|pyruvate kinase isozymes R/L', map_location='3 39.01 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89136122, end_position_on_the_genomic_accession=89146600, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18771, CurrentID=0, Status='live', Symbol='Pknox1', Aliases='D17Wsu76e, PREP1', description='Pbx/knotted 1 homeobox', other_designations='PBX/knotted homeobox 1', map_location='17 B-C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31564721, end_position_on_the_genomic_accession=31607693, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18772, CurrentID=0, Status='live', Symbol='Pkp1', Aliases='', description='plakophilin 1', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135871395, end_position_on_the_genomic_accession=135919361, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  18775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18775, CurrentID=0, Status='live', Symbol='Prl3d1', Aliases='AI325057, Csh1, Ghd4, PL-I, PL-I-alpha, PL-Ia, Pl-1, Pl1, mPL-I', description='prolactin family 3, subfamily d, member 1', other_designations='chorionic somatomammotropin hormone 1|growth hormone d4|placental lactogen 1|placental lactogen I|placental lactogen-I alpha|prolactin-like 2', map_location='13 12.27 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27094189, end_position_on_the_genomic_accession=27100260, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18776, CurrentID=0, Status='live', Symbol='Prl3b1', Aliases='Csh2, Ghd8, PL, Pl-2, Pl2, mPL-II, mplII', description='prolactin family 3, subfamily b, member 1', other_designations='PL-II|chorionic somatomammotropin hormone 2|growth hormone d8|placental lactogen 2|placental lactogen II', map_location='13 12.35 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27242430, end_position_on_the_genomic_accession=27249740, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18777, CurrentID=0, Status='live', Symbol='Lypla1', Aliases='Pla1a', description='lysophospholipase 1', other_designations='APT-1|LPL-I|lysoPLA I|lysophopholipase 1|lysophospholipase I|phospholipase 1a', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=4807563, end_position_on_the_genomic_accession=4846737, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18778, CurrentID=0, Status='live', Symbol='Pla2g1b', Aliases='Pla2a, sPLA2IB', description='phospholipase A2, group IB, pancreas', other_designations='PLA2-Ib|group IB phospholipase A2|phosphatidylcholine 2-acylhydrolase 1B', map_location='5 F1-G1.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115466241, end_position_on_the_genomic_accession=115474717, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18779, CurrentID=0, Status='live', Symbol='Pla2r1', Aliases='Pla2-r, Pla2g1br, Pla2r', description='phospholipase A2 receptor 1', other_designations='180 kDa secretory phospholipase A2 receptor|M-type receptor|PLA2-I receptor|phospholipase A2, group IB, pancreas, receptor', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=60417543, end_position_on_the_genomic_accession=60553350, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  18780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18780, CurrentID=0, Status='live', Symbol='Pla2g2a', Aliases='EF, Mom1, Pla2, sPLA2, sPla2-IIA', description='phospholipase A2, group IIA (platelets, synovial fluid)', other_designations='GIIC sPLA2|enhancing factor|group IIA phospholipase A2|modifier of Min1|non-pancreatic secreted type II phospholipase A2|phosphatidylcholine 2-acylhydrolase 2A|secretory group II phospholipase A2', map_location='4 70.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138831857, end_position_on_the_genomic_accession=138835189, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18781, CurrentID=0, Status='live', Symbol='Pla2g2c', Aliases='', description='phospholipase A2, group IIC', other_designations='14 kDa phospholipase A2|GIIC sPLA2|PLA2-8|phosphatidylcholine 2-acylhydrolase 2C|sPLA2-IIC', map_location='4 70.54 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138725325, end_position_on_the_genomic_accession=138744575, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18782, CurrentID=0, Status='live', Symbol='Pla2g2d', Aliases='Splash, mGIID', description='phospholipase A2, group IID', other_designations='GIID sPLA2|PLA2IID|phosphatidylcholine 2-acylhydrolase 2D|sPLA2-IID|secretory-type PLA, stroma-associated homolog', map_location='4 70.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138775735, end_position_on_the_genomic_accession=138782143, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18783, CurrentID=0, Status='live', Symbol='Pla2g4a', Aliases='Pla2g4, cPLA2, cPLA2-alpha, cPLA2alpha', description='phospholipase A2, group IVA (cytosolic, calcium-dependent)', other_designations='Cytosolic phospholipase A2 (CPLA2)|Type IV PLA2|cytosolic PLA2|phospholipase A2 group IVA|phospholipase A2, group 4', map_location='1 63.51 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=149829618, end_position_on_the_genomic_accession=149961284, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  18784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18784, CurrentID=0, Status='live', Symbol='Pla2g5', Aliases='PLA2, sPLA2', description='phospholipase A2, group V', other_designations='PLA2-10|group V phospholipase A2|phosphatidylcholine 2-acylhydrolase 5', map_location='4 70.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138799247, end_position_on_the_genomic_accession=138863469, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  18786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18786, CurrentID=0, Status='live', Symbol='Plaa', Aliases='2410007N06, AI536418, AU018445, AW208417, D4Ertd618e, PLA2P, PLAP, Ufd3', description='phospholipase A2, activating protein', other_designations='', map_location='4 43.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=94565139, end_position_on_the_genomic_accession=94603247, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  18787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18787, CurrentID=0, Status='live', Symbol='Serpine1', Aliases='PAI-1, PAI1, Planh1', description='serine (or cysteine) peptidase inhibitor, clade E, member 1', other_designations='PAI|Plasminogen activator inhibitor-1 precursor (PAI-1) (Endothelial plasminogen activator inhibitor) (PAI)|endothelial plasminogen activator inhibitor|nexin, plasminogen activator inhibitor type 1|plasminogen activator inhibitor, type I|serine (or cysteine) proteinase inhibitor, clade E, member 1|serpin E1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137061504, end_position_on_the_genomic_accession=137072272, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18788, CurrentID=0, Status='live', Symbol='Serpinb2', Aliases='PAI-2, Planh2, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 2', other_designations='plasminogen activator inhibitor, type II|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 2|serine (or cysteine) proteinase inhibitor, clade B, member 2|serpin B2', map_location='1 50.34 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107500892, end_position_on_the_genomic_accession=107525600, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18789, CurrentID=0, Status='live', Symbol='Papola', Aliases='Pap, PapIII, Plap', description='poly (A) polymerase alpha', other_designations='PAP-alpha|polynucleotide adenylyltransferase', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105784702, end_position_on_the_genomic_accession=105838949, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  18791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18791, CurrentID=0, Status='live', Symbol='Plat', Aliases='AU020998, AW212668, D8Ertd2e, tPA', description='plasminogen activator, tissue', other_designations='t-plasminogen activator', map_location='8 11.42 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22757744, end_position_on_the_genomic_accession=22782848, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  18792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18792, CurrentID=0, Status='live', Symbol='Plau', Aliases='u-PA, uPA', description='plasminogen activator, urokinase', other_designations='U-plasminogen activator', map_location='14 11.53 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20836662, end_position_on_the_genomic_accession=20843388, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18793, CurrentID=0, Status='live', Symbol='Plaur', Aliases='Cd87, u-PAR, uPAR', description='plasminogen activator, urokinase receptor', other_designations='urokinase plasminogen activator receptor', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24458741, end_position_on_the_genomic_accession=24475873, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18795, CurrentID=0, Status='live', Symbol='Plcb1', Aliases='3110043I21Rik, AI132408, Plcb, mKIAA0581', description='phospholipase C, beta 1', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase beta-1|PLC-154|PLC-beta-1|phosphoinositide phospholipase C|phosphoinositide phospholipase C-beta-1|phospholipase C-beta-1|phospholipase C-beta-1a', map_location='2 65.66 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=134786164, end_position_on_the_genomic_accession=135475258, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  18796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18796, CurrentID=0, Status='live', Symbol='Plcb2', Aliases='AI550384, B230205M18Rik, B230399N12', description='phospholipase C, beta 2', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase beta-2|PLC-beta-2|phosphoinositide phospholipase C-beta-2|phospholipase C-beta-2', map_location='2 59.43 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118707517, end_position_on_the_genomic_accession=118728659, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  18797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18797, CurrentID=0, Status='live', Symbol='Plcb3', Aliases='mKIAA4098', description='phospholipase C, beta 3', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase beta-3|PLC-beta-3|phosphoinositide phospholipase C-beta-3|phospholipase C-beta-3', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6953706, end_position_on_the_genomic_accession=6974336, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  18798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18798, CurrentID=0, Status='live', Symbol='Plcb4', Aliases='A930039J07Rik, AI854601, C230058B11Rik', description='phospholipase C, beta 4', other_designations='', map_location='2 66.64 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=135658437, end_position_on_the_genomic_accession=136014602, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  18799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18799, CurrentID=0, Status='live', Symbol='Plcd1', Aliases='AW212592, C79986', description='phospholipase C, delta 1', other_designations='PLC-delta 1|phosphoinositide phospholipase C-delta-1|phospholipase C delta 1 Lf', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119071527, end_position_on_the_genomic_accession=119093502, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  18802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18802, CurrentID=0, Status='live', Symbol='Plcd4', Aliases='4921507K24Rik', description='phospholipase C, delta 4', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase delta-4|PLC-delta-4|phosphoinositide phospholipase C-delta-4|phospholipase C-delta-4', map_location='1 38.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74541972, end_position_on_the_genomic_accession=74565977, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  18803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18803, CurrentID=0, Status='live', Symbol='Plcg1', Aliases='AI894140, Cded, Plc-1, Plc-gamma1, Plcg-1', description='phospholipase C, gamma 1', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase gamma 1|1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase gamma-1|PLC-gamma-1|cell differentiation and embryonic development|phosphoinositide phospholipase C-gamma-1|phospholipase C-gamma-1', map_location='2 80.97 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=160731310, end_position_on_the_genomic_accession=160775760, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  18805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18805, CurrentID=0, Status='live', Symbol='Pld1', Aliases='AA536939, C85393a, Pld1b, Pld1', description='phospholipase D1', other_designations='PLD 1|choline phosphatase 1|mPLD1|phosphatidylcholine-hydrolyzing phospholipase D1', map_location='3 11.15 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=27938680, end_position_on_the_genomic_accession=28133362, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  18806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18806, CurrentID=0, Status='live', Symbol='Pld2', Aliases='PLD1C', description='phospholipase D2', other_designations='choline phosphatase 2|phosphatidylcholine-hydrolyzing phospholipase D2', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70540063, end_position_on_the_genomic_accession=70558110, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  18807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18807, CurrentID=0, Status='live', Symbol='Pld3', Aliases='Sam-9', description='phospholipase D family, member 3', other_designations='PLD 3|choline phosphatase 3|phosphatidylcholine-hydrolyzing phospholipase D3|schwannoma-associated protein 9', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27531645, end_position_on_the_genomic_accession=27553213, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18809, CurrentID=0, Status='live', Symbol='ple', Aliases='Tg.ple, Tgple', description='perinatal lethality', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18810, CurrentID=0, Status='live', Symbol='Plec', Aliases='AA591047, AU042537, EBS1, PCN, PLTN1, Plec', description='plectin', other_designations='plectin 1|plectin-1|plectin-6', map_location='15 35.48 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76170974, end_position_on_the_genomic_accession=76231378, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  18811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18811, CurrentID=0, Status='live', Symbol='Prl2c2', Aliases='MRP-1, PLF-1, Plf, Plf1', description='prolactin family 2, subfamily c, member 2', other_designations='mitogen-regulated protein 1|proliferin|proliferin-1', map_location='13 5.18 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12996120, end_position_on_the_genomic_accession=13005330, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  18812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18812, CurrentID=0, Status='live', Symbol='Prl2c3', Aliases='Ghd1, MRP-2, MRP-3, Mrpplf3, PLF-2, PLF-3, Plf2, Prl2c4, mrp/plf3', description='prolactin family 2, subfamily c, member 3', other_designations='growth hormone d1|mitogen-regulated protein 2|mitogen-regulated protein 3|prolactin-2C4|proliferin 2|proliferin-2|proliferin-3', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12790816, end_position_on_the_genomic_accession=12800093, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18813, CurrentID=0, Status='live', Symbol='Pa2g4', Aliases='38kDa, AA672939, Ebp1, Plfap', description='proliferation-associated 2G4', other_designations='ErbB3-binding protein 1|IRES-specific cellular trans-acting factor 45 kDa|ITAF45|mpp1|proliferation-associated 2G4, 38kD|proliferation-associated protein 1|protein p38-2G4', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128557766, end_position_on_the_genomic_accession=128565934, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18814, CurrentID=0, Status='live', Symbol='Prl7d1', Aliases='Ghd19, PLF-RP, PRP, Plfr', description='prolactin family 7, subfamily d, member 1', other_designations='growth hormone d19|proliferin related protein|proliferin-related protein', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27708998, end_position_on_the_genomic_accession=27716748, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18815, CurrentID=0, Status='live', Symbol='Plg', Aliases='AI649309, Pg', description='plasminogen', other_designations='angiostatin|plasmin heavy chain A|plasmin light chain B', map_location='17 8.5 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12378609, end_position_on_the_genomic_accession=12419384, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  18816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18816, CurrentID=0, Status='live', Symbol='Serpinf2', Aliases='AI747498, Pli, Serpimf2', description='serine (or cysteine) peptidase inhibitor, clade F, member 2', other_designations='alpha 2 antiplasmin|alpha-2-AP|alpha-2-PI|alpha-2-plasmin inhibitor|pigment epithelium derived factor|plasmin inhibitor alpha 2|serine (or cysteine) proteinase inhibitor, clade F, member 2|serpin F2', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75431736, end_position_on_the_genomic_accession=75439501, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18817, CurrentID=0, Status='live', Symbol='Plk1', Aliases='Plk, STPK13', description='polo-like kinase 1', other_designations='PLK-1|serine/threonine-protein kinase 13', map_location='7 65.52 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122159437, end_position_on_the_genomic_accession=122169875, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  18818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18818, CurrentID=0, Status='live', Symbol='Plk-ps1', Aliases='', description='polo-like kinase, pseudogene 1', other_designations='polo-like kinase 1 pseudogene', map_location='5 18.25 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35554135, end_position_on_the_genomic_accession=35556252, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  18821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18821, CurrentID=0, Status='live', Symbol='Pln', Aliases='Plb', description='phospholamban', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=53337686, end_position_on_the_genomic_accession=53345999, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18822, CurrentID=0, Status='live', Symbol='Plod1', Aliases='2410042F05Rik, AI854890, AV007929, Lh1, Plod', description='procollagen-lysine, 2-oxoglutarate 5-dioxygenase 1', other_designations='', map_location='4 78.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147909753, end_position_on_the_genomic_accession=147936776, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  18823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18823, CurrentID=0, Status='live', Symbol='Plp1', Aliases='DM20, Plp, jimpy, jp, msd, rsh', description='proteolipid protein (myelin) 1', other_designations='lipophilin|myelin synthesis deficiency|rump shaker', map_location='X F1-F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136822746, end_position_on_the_genomic_accession=136838582, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  18824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18824, CurrentID=0, Status='live', Symbol='Plp2', Aliases='mIMA4', description='proteolipid protein 2', other_designations='', map_location='X A2-A3.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7668116, end_position_on_the_genomic_accession=7671278, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  18826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18826, CurrentID=0, Status='live', Symbol='Lcp1', Aliases='AW536232, D14Ertd310e, LCP-1, Pls2, pp65', description='lymphocyte cytosolic protein 1', other_designations='65 kDa macrophage protein|L-fimbrin|L-plastin (Lymphocyte cytosolic protein 1) (LCP-1) (65 kDa macrophage protein) (PP65)|plastin 2, L', map_location='14 39.63 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75131105, end_position_on_the_genomic_accession=75230842, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  18828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18828, CurrentID=0, Status='live', Symbol='Plscr2', Aliases='', description='phospholipid scramblase 2', other_designations='PL scramblase 2|ca(2+)-dependent phospholipid scramblase 2', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=92275602, end_position_on_the_genomic_accession=92297752, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  18829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18829, CurrentID=0, Status='live', Symbol='Ccl21a', Aliases='6CKBAC2, 6Ckine, ALP, AW987545, CKb9, SCYA21a, SLC, Scya21, Scya21b, Tca4, plt', description='chemokine (C-C motif) ligand 21A (serine)', other_designations='CC chemokine 6Ckine-ser|CCL21-Ser|beta chemokine exodus-2|beta-chemokine exodus-2|exodus-2|small inducible cytokine A21a|small-inducible cytokine A21a|thymus-derived chemotactic agent 4', map_location='4 22.81 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42772860, end_position_on_the_genomic_accession=42773991, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18830, CurrentID=0, Status='live', Symbol='Pltp', Aliases='Bpife, OD107', description='phospholipid transfer protein', other_designations='lipid transfer protein II', map_location='2 85.27 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164839518, end_position_on_the_genomic_accession=164857708, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  18843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18843, CurrentID=0, Status='live', Symbol='Bpifa1', Aliases='LUNX, NASG, Plunc, SPLUNC1, SPURT', description='BPI fold containing family A, member 1', other_designations='palate lung and nasal epithelium clone protein|palate, lung, and nasal epithelium associated|palate, lung, and nasal epithelium clone|palate, lung, and nasal epithelium expressed transcript|protein Plunc', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154142880, end_position_on_the_genomic_accession=154149217, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18844, CurrentID=0, Status='live', Symbol='Plxna1', Aliases='2600013D04Rik, NOV, PlexA1, Plxn1, mKIAA4053', description='plexin A1', other_designations='plex 1|plexin 1|plexin-1', map_location='6 E2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89316313, end_position_on_the_genomic_accession=89362613, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  18845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18845, CurrentID=0, Status='live', Symbol='Plxna2', Aliases='2810428A13Rik, AA589422, AW457381, OCT, PlexA2, Plxn2, mKIAA0463', description='plexin A2', other_designations='plex 2|plexin 2|plexin-2', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=194619829, end_position_on_the_genomic_accession=194816869, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  18846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18846, CurrentID=0, Status='live', Symbol='Plxna3', Aliases='PlexA3, Plxa3, Plxn3, Plxn4, SEX, mKIAA4078', description='plexin A3', other_designations='plexin 3', map_location='X B-C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74329066, end_position_on_the_genomic_accession=74344689, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  18849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18849, CurrentID=0, Status='live', Symbol='pma', Aliases='', description='peroneal muscular atrophy', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18854, CurrentID=0, Status='live', Symbol='Pml', Aliases='1200009E24Rik, AI661194, Trim19', description='promyelocytic leukemia', other_designations='probable transcription factor PML', map_location='9 31.63 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58217179, end_position_on_the_genomic_accession=58249796, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18857, CurrentID=0, Status='live', Symbol='Pmp2', Aliases='P2', description='peripheral myelin protein 2', other_designations='', map_location='3 2.54 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10179851, end_position_on_the_genomic_accession=10183885, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18858, CurrentID=0, Status='live', Symbol='Pmp22', Aliases='22kDa, Gas-3, HNPP, Tr, trembler', description='peripheral myelin protein 22', other_designations='PMP-22|growth arrest-specific protein 3|peripheral myelin protein, 22 kDa', map_location='11 38.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=63128982, end_position_on_the_genomic_accession=63159547, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  18861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18861, CurrentID=0, Status='live', Symbol='Pms2', Aliases='AW555130, Pmsl2', description='postmeiotic segregation increased 2 (S. cerevisiae)', other_designations='DNA mismatch repair protein PMS2|PMS1 protein homolog 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143909974, end_position_on_the_genomic_accession=143985719, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  83555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83555, CurrentID=0, Status='live', Symbol='Tex13', Aliases='4933403J23Rikb, Tex13', description='testis expressed gene 13', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140808307, end_position_on_the_genomic_accession=140813451, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  18933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18933, CurrentID=0, Status='live', Symbol='Prrx1', Aliases='A230024N07Rik, AA755424, AI385634, AI843499, K-2, Pmx, Pmx1, Prx1, mHox', description='paired related homeobox 1', other_designations='homeobox protein K-2|homeobox protein mHox|paired-related homeobox protein 1', map_location='1 70.53 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=163245119, end_position_on_the_genomic_accession=163313705, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  18935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18935, CurrentID=0, Status='live', Symbol='Phox2b', Aliases='Dilp1, NBPhox, Pmx2b', description='paired-like homeobox 2b', other_designations='GENA 269|PHOX2B homeodomain protein|dilated pupils 1|neuroblastoma Phox|paired mesoderm homeobox 2b', map_location='5 35.95 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=67094397, end_position_on_the_genomic_accession=67099126, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18938, CurrentID=0, Status='live', Symbol='Ppp1r14b', Aliases='AOM172, PHI-1, PLCB3N, Png', description='protein phosphatase 1, regulatory (inhibitor) subunit 14B', other_designations='phosphatase holoenzyme inhibitor 1|phospholipase C-beta-3 neighbouring gene protein|phospholipase c neighboring', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6975048, end_position_on_the_genomic_accession=6977324, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  18946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18946, CurrentID=0, Status='live', Symbol='Pnliprp1', Aliases='Plrp1', description='pancreatic lipase related protein 1', other_designations='PL-RP1|pancreatic lipase-related protein 1', map_location='19 54.49 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58728887, end_position_on_the_genomic_accession=58744169, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18947, CurrentID=0, Status='live', Symbol='Pnliprp2', Aliases='PLRP2', description='pancreatic lipase-related protein 2', other_designations='PL-RP2|cytotoxic T-lymphocyte lipase|galactolipase', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58759723, end_position_on_the_genomic_accession=58777534, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  18948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18948, CurrentID=0, Status='live', Symbol='Pnmt', Aliases='Pent', description='phenylethanolamine-N-methyltransferase', other_designations='PNMTase|noradrenaline N-methyltransferase', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98386632, end_position_on_the_genomic_accession=98388097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  18949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18949, CurrentID=0, Status='live', Symbol='Pnn', Aliases='AU045199, D12Ertd512e', description='pinin', other_designations='', map_location='12 26.0 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=59066919, end_position_on_the_genomic_accession=59074017, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  18950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18950, CurrentID=0, Status='live', Symbol='Pnp', Aliases='AL024301, AU015798, Np, Np-1, Np-21, Pnp', description='purine-nucleoside phosphorylase', other_designations='inosine phosphorylase|inosine-guanosine phosphorylase|purine-nucleoside phosphorylase 1', map_location='14 B-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50944303, end_position_on_the_genomic_accession=50953412, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18951, CurrentID=0, Status='live', Symbol='Sept5', Aliases='Cdcrel-1, Cdcrel1, Pnutl1', description='septin 5', other_designations='cell division control related protein 1|cell division control-related protein 1|peanut-like 1 homolog|peanut-like protein 1', map_location='16 11.53 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18621811, end_position_on_the_genomic_accession=18629938, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18952, CurrentID=0, Status='live', Symbol='Sept4', Aliases='ARTS, Bh5, OTTMUSG00000001265, Pnutl2', description='septin 4', other_designations='M-Septin|brain protein H5|cell division control-related protein 2b|peanut-like 2 homolog|peanut-like protein 2|septin H5', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87566727, end_position_on_the_genomic_accession=87590539, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18968, CurrentID=0, Status='live', Symbol='Pola1', Aliases='AW321876, Pola', description='polymerase (DNA directed), alpha 1', other_designations='DNA polymerase alpha 1, 180 kDa|DNA polymerase alpha catalytic subunit p180|G7a/Bat6Hom|polymerase, alpha 1', map_location='X C-D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=93304766, end_position_on_the_genomic_accession=93632202, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  18969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18969, CurrentID=0, Status='live', Symbol='Pola2', Aliases='AI573378', description='polymerase (DNA directed), alpha 2', other_designations='DNA polymerase alpha 2, 68 kDa|DNA polymerase alpha 70 kDa subunit', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5940543, end_position_on_the_genomic_accession=5964212, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  18970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18970, CurrentID=0, Status='live', Symbol='Polb', Aliases='A430088C08Rik', description='polymerase (DNA directed), beta', other_designations='Pol beta', map_location='8 11.42 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22628119, end_position_on_the_genomic_accession=22653468, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  18971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18971, CurrentID=0, Status='live', Symbol='Pold1', Aliases='125kDa', description='polymerase (DNA directed), delta 1, catalytic subunit', other_designations='DNA polymerase delta 1, catalytic domain', map_location='7 28.83 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44532744, end_position_on_the_genomic_accession=44548815, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  18972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18972, CurrentID=0, Status='live', Symbol='Pold2', Aliases='50kDa, p50, po1D2', description='polymerase (DNA directed), delta 2, regulatory subunit', other_designations='DNA polymerase delta subunit p50', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5872180, end_position_on_the_genomic_accession=5878256, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  18973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18973, CurrentID=0, Status='live', Symbol='Pole', Aliases='', description='polymerase (DNA directed), epsilon', other_designations='DNA polymerase II subunit A|DNA-directed DNA polymerase epsilon|pol-epsilon', map_location='5 E3-E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110286319, end_position_on_the_genomic_accession=110337453, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  18974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18974, CurrentID=0, Status='live', Symbol='Pole2', Aliases='', description='polymerase (DNA directed), epsilon 2 (p59 subunit)', other_designations='DNA polymerase II subunit 2|DNA polymerase epsilon small subunit|DNA polymerase epsilon subunit B|DNA polymerase epsilon, subunit 2', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69201773, end_position_on_the_genomic_accession=69228242, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  18975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18975, CurrentID=0, Status='live', Symbol='Polg', Aliases='AA409516A, Polg', description='polymerase (DNA directed), gamma', other_designations='DNA polymerase gamma subunit 1 (Mitochondrial DNA polymerase catalytic subunit) (PolG-alpha)|Pol gamma|mitochondrial DNA polymerase catalytic subunit|mitochondrial DNA polymerase gamma|mitochondrial DNA polymerase-gamma|mitochondrial polymerase gamma|polG-alpha|polymerase, gamma', map_location='7 E', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79449382, end_position_on_the_genomic_accession=79466305, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  18976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18976, CurrentID=0, Status='live', Symbol='Pomc', Aliases='ACTH, BE, Beta-LPH, Clip, Gamma-LPH, Npp-1, Pomc1, alpha-MSH, alphaMSH, beta-MSH, gamma-MSH, Pomc', description='pro-opiomelanocortin-alpha', other_designations='adrenal corticotropic hormone|adrenocorticotropic hormone|alpha-melanocyte stimulating hormone|beta-endorphin|corticotropin-like intermediary peptide|corticotropin-lipotropin|melanotropin gamma|met-enkephalin', map_location='12 1.99 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3954945, end_position_on_the_genomic_accession=3960643, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18977, CurrentID=0, Status='live', Symbol='Pomc-ps1', Aliases='Pomc-2, Pomc2', description='pro-opiomelanocortin, pseudogene 1', other_designations='', map_location='19 10.39 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=15362335, end_position_on_the_genomic_accession=15362870, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  18979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18979, CurrentID=0, Status='live', Symbol='Pon1', Aliases='Pon', description='paraoxonase 1', other_designations='A-esterase 1|PON 1|aromatic esterase 1|serum aryldialkylphosphatase 1', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=5168090, end_position_on_the_genomic_accession=5193987, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18983, CurrentID=0, Status='live', Symbol='Cnot7', Aliases='AU022737, CAF-1, Caf1, Pop2', description='CCR4-NOT transcription complex, subunit 7', other_designations='CCR4-associated factor 1|POP2 homolog|catabolite repressor protein (CCR4)-associative factor 1', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40492538, end_position_on_the_genomic_accession=40515847, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  18984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18984, CurrentID=0, Status='live', Symbol='Por', Aliases='4933424M13Rik, CPR, CYPOR, P450R', description='P450 (cytochrome) oxidoreductase', other_designations='NADH cytochrome P450 oxydoreductase', map_location='5 75.34 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135670040, end_position_on_the_genomic_accession=135735326, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  18985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18985, CurrentID=0, Status='live', Symbol='Pou2af1', Aliases='BOB.1, Bob-1, Bob1, OBF-1, OBF.1, OCA-B, OCAB', description='POU domain, class 2, associating factor 1', other_designations='B-cell-specific coactivator OBF-1|OCT-binding factor 1', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51213690, end_position_on_the_genomic_accession=51240079, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  18986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18986, CurrentID=0, Status='live', Symbol='Pou2f1', Aliases='2810482H01Rik, NF-A1, Oct-1, Oct1, Otf-1, Otf1', description='POU domain, class 2, transcription factor 1', other_designations='octamer-binding protein 1|octamer-binding transcription factor 1|ubiquitous transcription factor', map_location='1 73.21 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165865150, end_position_on_the_genomic_accession=166002672, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  18987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18987, CurrentID=0, Status='live', Symbol='Pou2f2', Aliases='Oct-2, Oct2a, Oct2b, Otf-2, Otf2', description='POU domain, class 2, transcription factor 2', other_designations='lymphoid-restricted immunoglobulin octamer-binding protein NF-A2|octamer-binding protein 2|octamer-binding transcription factor 2', map_location='7 13.73 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25091114, end_position_on_the_genomic_accession=25179992, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  18988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18988, CurrentID=0, Status='live', Symbol='Pou2f3', Aliases='Epoc-1, Oct-11a, Oct11, Otf-11, Otf11, Skin, Skin-1a, Skn-1a, Skn-li', description='POU domain, class 2, transcription factor 3', other_designations='oct-11|octamer-binding protein 11|octamer-binding transcription factor 11', map_location='9 24.21 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=43123927, end_position_on_the_genomic_accession=43205755, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  18991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18991, CurrentID=0, Status='live', Symbol='Pou3f1', Aliases='Oct-6, Oct6, Otf-6, Otf6, Scip, Test1, Tst-1, Tst1', description='POU domain, class 3, transcription factor 1', other_designations='POU domain transcription factor SCIP|octamer binding factor oct6|octamer-binding protein 6|octamer-binding transcription factor 6', map_location='4 57.86 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124657646, end_position_on_the_genomic_accession=124660655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18992, CurrentID=0, Status='live', Symbol='Pou3f2', Aliases='9430075J19Rik, A230098E07Rik, Brn-2, Brn2, OTF-7, Otf7, oct-7', description='POU domain, class 3, transcription factor 2', other_designations='brain-2 class III POU-domain protein|brain-specific homeobox/POU domain protein 2|nervous system-specific octamer-binding transcription factor N-Oct-3|octamer-binding protein 7|octamer-binding transcription factor 7', map_location='4 9.73 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=22482095, end_position_on_the_genomic_accession=22488366, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  18993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18993, CurrentID=0, Status='live', Symbol='Pou3f3', Aliases='Brn-1, Brn1, Otf8, Skin1', description='POU domain, class 3, transcription factor 3', other_designations='OTF-8|brain-1|brain-specific homeobox/POU domain protein 1|oct-8|octamer-binding protein 8|octamer-binding transcription factor 8', map_location='1 21.21 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=42697146, end_position_on_the_genomic_accession=42700210, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18994, CurrentID=0, Status='live', Symbol='Pou3f4', Aliases='Brn-4, Brn4, OTF-9, Otf9, Slf, oct-9', description='POU domain, class 3, transcription factor 4', other_designations='brain-4 class III POU-domain protein|brain-specific homeobox/POU domain protein 4|octamer-binding protein 9|octamer-binding transcription factor 9|sex linked fidget', map_location='X 48.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=110814465, end_position_on_the_genomic_accession=110815550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  18996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18996, CurrentID=0, Status='live', Symbol='Pou4f1', Aliases='Brn-3, Brn-3.0, Brn3, Brn3.0, Brn3a, E130119J07Rik', description='POU domain, class 4, transcription factor 1', other_designations='brain-3A|brain-specific homeobox/POU domain protein 3A|brn-3A', map_location='14 E1-E3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=104461675, end_position_on_the_genomic_accession=104468022, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18997, CurrentID=0, Status='live', Symbol='Pou4f2', Aliases='Brn-3.2, Brn-3b, Brn3b, Pou4f-rs1, mBrn3-3R', description='POU domain, class 4, transcription factor 2', other_designations='POU domain, class 4, transcription factor, related sequence 1|brain-3B|brain-specific homeobox/POU domain protein 3B', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=78433009, end_position_on_the_genomic_accession=78436652, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  18998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18998, CurrentID=0, Status='live', Symbol='Pou4f3', Aliases='Brn3.1, Brn3c, ddl, dreidel', description='POU domain, class 4, transcription factor 3', other_designations='brain-3C|brain-specific homeobox POU domain protein 3C|brain-specific homeobox/POU domain protein 3C|brn-3C', map_location='18 B3-E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42394597, end_position_on_the_genomic_accession=42396093, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18999, CurrentID=0, Status='live', Symbol='Pou5f1', Aliases='NF-A3, Oct-3, Oct-3/4, Oct-4, Oct3, Oct3/4, Oct4, Otf-3, Otf-4, Otf3, Otf3-rs7, Otf3g, Otf4', description='POU domain, class 5, transcription factor 1', other_designations='octamer-binding protein 3|octamer-binding protein 4|octamer-binding transcription factor 3', map_location='17 18.69 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35506032, end_position_on_the_genomic_accession=35510777, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19009, CurrentID=0, Status='live', Symbol='Pou6f1', Aliases='2310038G18Rik, C79593, Emb, cns-1', description='POU domain, class 6, transcription factor 1', other_designations='octamer-binding transcription factor EMB|transcription regulatory protein MCP-1', map_location='15 56.35 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100575318, end_position_on_the_genomic_accession=100601690, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19011, CurrentID=0, Status='live', Symbol='Endou', Aliases='Pp11r, Tcl-30', description='endonuclease, polyU-specific', other_designations='PP11-related protein|T-cell-specific protein 30|placental protein 11 related|placental protein 11-related protein|protein endoU|uridylate-specific endoribonuclease', map_location='15 53.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97711015, end_position_on_the_genomic_accession=97731405, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19012, CurrentID=0, Status='live', Symbol='Plpp1', Aliases='Hic53, Hpic53, LPP-1, LPP1, PAP-2a, PAP2-alpha, PAP2a, Ppap2a, mPAP', description='phospholipid phosphatase 1', other_designations='35 kDa PAP|hydrogen peroxide inducible protein 53|lipid phosphate phosphohydrolase 1|phosphatidate phosphohydrolase type 2a|phosphatidic acid phosphatase 2a|phosphatidic acid phosphatase type 2A', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112800727, end_position_on_the_genomic_accession=112867894, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19013, CurrentID=0, Status='live', Symbol='Ppara', Aliases='4933429D07Rik, AW742785, Nr1c1, PPAR-alpha, PPARalpha, Ppar', description='peroxisome proliferator activated receptor alpha', other_designations='nuclear receptor subfamily 1 group C member 1', map_location='15 40.42 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85728192, end_position_on_the_genomic_accession=85806851, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  19014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19014, CurrentID=0, Status='live', Symbol='Med1', Aliases='AI480703, CRSP210, DRIP205, PBP, Pparbp, TRAP220, l11Jus15', description='mediator complex subunit 1', other_designations='PPAR-binding protein|Ppar binding protein|TR-interacting protein 2|TRIP-2|peroxisome proliferator activated receptor binding protein|peroxisome proliferator-activated receptor-binding protein|thyroid hormone receptor-associated protein complex 220 kDa component|thyroid receptor-interacting protein 2', map_location='11 61.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98152154, end_position_on_the_genomic_accession=98193314, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  19015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19015, CurrentID=0, Status='live', Symbol='Ppard', Aliases='NUC-1, NUC1, Nr1c2, PPAR-beta, PPAR-delta, PPAR[b], PPARdelta, Pparb, Pparb/d', description='peroxisome proliferator activator receptor delta', other_designations='PPARdelta/beta|Peroxisome proliferator-activated receptor beta|nuclear hormone receptor 1|nuclear receptor subfamily 1 group C member 2', map_location='17 14.64 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28232754, end_position_on_the_genomic_accession=28301469, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19016, CurrentID=0, Status='live', Symbol='Pparg', Aliases='Nr1c3, PPAR-gamma, PPAR-gamma2, PPARgamma, PPARgamma2', description='peroxisome proliferator activated receptor gamma', other_designations='nuclear receptor subfamily 1 group C member 3|peroxisome proliferator activated receptor gamma 2|peroxisome proliferator activated receptor gamma 4', map_location='6 E3-F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115360879, end_position_on_the_genomic_accession=115490404, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  19017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19017, CurrentID=0, Status='live', Symbol='Ppargc1a', Aliases='A830037N07Rik, ENSMUSG00000079510, Gm11133, PGC-1, PPARGC-1-alpha, Pgc-1alpha, Pgc1, Pgco1, Ppargc1', description='peroxisome proliferative activated receptor, gamma, coactivator 1 alpha', other_designations='PPAR-gamma coactivator 1-alpha', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=51454249, end_position_on_the_genomic_accession=52115958, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19018, CurrentID=0, Status='live', Symbol='Scand1', Aliases='2310003H23Rik, Leap1, PGC-2, Ppargc2', description='SCAN domain-containing 1', other_designations='leucine-rich-domain inter-acting protein 1|peroxisome proliferative activated receptor, gamma, coactivator 2', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156311846, end_position_on_the_genomic_accession=156312704, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19023, CurrentID=0, Status='live', Symbol='Ppef2', Aliases='', description='protein phosphatase, EF hand calcium-binding domain 2', other_designations='PPEF-2', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92226691, end_position_on_the_genomic_accession=92256900, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19024, CurrentID=0, Status='live', Symbol='Ppfibp2', Aliases='Cclp1', description='PTPRF interacting protein, binding protein 2 (liprin beta 2)', other_designations='PTPRF-interacting protein-binding protein 2|coiled-coil-like protein 1|liprin beta 2|protein tyrosine phosphatase receptor type f polypeptide-interacting protein-binding protein 2|protein tyrosine phosphatase, receptor-type, F interacting protein, binding protein 2', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107595051, end_position_on_the_genomic_accession=107748583, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  19025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19025, CurrentID=0, Status='live', Symbol='Ctsa', Aliases='AU019505, PPCA, Ppgb', description='cathepsin A', other_designations='carboxypeptidase C|carboxypeptidase L|protective protein cathepsin A|protective protein for beta-galactosidase', map_location='2 85.27 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164832872, end_position_on_the_genomic_accession=164841032, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  19027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19027, CurrentID=0, Status='live', Symbol='Sypl', Aliases='AI314763, AI604763, D12Ertd446e, PanI, Pphn1, Sypl', description='synaptophysin-like protein', other_designations='Pan I|pantophysin', map_location='12 B2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=32953873, end_position_on_the_genomic_accession=32979502, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19035, CurrentID=0, Status='live', Symbol='Ppib', Aliases='AA408962, AA553318, AI844835, Cphn-2, Cphn2, CyP-20b', description='peptidylprolyl isomerase B', other_designations='CYP-S1|PPIase B|S-cyclophilin|SCYLP|cyclophilin B|rotamase B', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66060169, end_position_on_the_genomic_accession=66066629, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19038, CurrentID=0, Status='live', Symbol='Ppic', Aliases='CyP-20c', description='peptidylprolyl isomerase C', other_designations='PPIase C|cyclophilin C|rotamase C', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=53406341, end_position_on_the_genomic_accession=53418007, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19039, CurrentID=0, Status='live', Symbol='Lgals3bp', Aliases='90K, CyCAP, MAC-2BP, Ppicap, Tango10b', description='lectin, galactoside-binding, soluble, 3 binding protein', other_designations='cyp-C-associated protein|lectin galactoside-binding soluble 3-binding protein|mama|peptidylprolyl isomerase C-associated protein', map_location='11 E', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118392752, end_position_on_the_genomic_accession=118401931, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19041, CurrentID=0, Status='live', Symbol='Ppl', Aliases='AW553870', description='periplakin', other_designations='195kDa protein', map_location='16 A-B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5086291, end_position_on_the_genomic_accession=5132481, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  19042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19042, CurrentID=0, Status='live', Symbol='Ppm1a', Aliases='2310003C21Rik, 2900017D14Rik, AI427932, AU017636, MMPa-2, MPPa-1', description='protein phosphatase 1A, magnesium dependent, alpha isoform', other_designations='Mpp alpha|PP2C-alpha|protein phosphatase 2C isoform alpha|protein phosphatase IA', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72761211, end_position_on_the_genomic_accession=72799822, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19043, CurrentID=0, Status='live', Symbol='Ppm1b', Aliases='PP2CB', description='protein phosphatase 1B, magnesium dependent, beta isoform', other_designations='PP2C-beta|protein phosphatase 2C isoform beta', map_location='17 E4-E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84956741, end_position_on_the_genomic_accession=85023992, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19044, CurrentID=0, Status='live', Symbol='Ppox', Aliases='Ppo', description='protoporphyrinogen oxidase', other_designations='protoporhyrinogen oxidase', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171276991, end_position_on_the_genomic_accession=171281203, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19045, CurrentID=0, Status='live', Symbol='Ppp1ca', Aliases='Ppp1c, dism2', description='protein phosphatase 1, catalytic subunit, alpha isoform', other_designations='PP-1A', map_location='19 E3-F2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4192174, end_position_on_the_genomic_accession=4195419, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19046, CurrentID=0, Status='live', Symbol='Ppp1cb', Aliases='1200010B19', description='protein phosphatase 1, catalytic subunit, beta isoform', other_designations='PP-1B', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=32458848, end_position_on_the_genomic_accession=32493712, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19047, CurrentID=0, Status='live', Symbol='Ppp1cc', Aliases='PP-1G, PP1, dis2m1', description='protein phosphatase 1, catalytic subunit, gamma isoform', other_designations='PP1C gamma 1|PP1C gamma 2|protein phosphatase 1C catalytic subunit', map_location='5 F-G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122158255, end_position_on_the_genomic_accession=122175273, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19049, CurrentID=0, Status='live', Symbol='Ppp1r1b', Aliases='AU040756, DARPP-32, Darpp32', description='protein phosphatase 1, regulatory (inhibitor) subunit 1B', other_designations='dopamine- and cAMP-regulated neuronal phosphoprotein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98348738, end_position_on_the_genomic_accession=98357795, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19051, CurrentID=0, Status='live', Symbol='Ppp1r17', Aliases='Gsbs, Ppp1r2', description='protein phosphatase 1, regulatory subunit 17', other_designations='G substrate|G-substrate|protein phosphatase 1, regulatory (inhibitor) 2', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56017515, end_position_on_the_genomic_accession=56032688, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19052, CurrentID=0, Status='live', Symbol='Ppp2ca', Aliases='PP2A, R75353', description='protein phosphatase 2 (formerly 2A), catalytic subunit, alpha isoform', other_designations='PP2A-alpha|protein phosphatase 2a, catalytic subunit, alpha isoform', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52098824, end_position_on_the_genomic_accession=52122749, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19053, CurrentID=0, Status='live', Symbol='Ppp2cb', Aliases='AI115466, D8Ertd766e, PP2Ac', description='protein phosphatase 2 (formerly 2A), catalytic subunit, beta isoform', other_designations='PP2A-beta|protein phosphatase 2a, catalytic subunit, beta isoform', map_location='8 20.63 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33599621, end_position_on_the_genomic_accession=33619804, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19054, CurrentID=0, Status='live', Symbol='Ppp2r3d', Aliases="PP2A B''-PR59, PP2A PR59, PR59, Ppp2r3, Ppp2r3a", description="protein phosphatase 2 (formerly 2A), regulatory subunit B'', delta", other_designations="protein phosphatase 2 (formerly 2A), regulatory subunit B'', alpha|protein phosphatase 2, regulatory subunit (59 kDa)|protein phosphatase 2, regulatory subunit B, alpha|protein phosphatase 2A, 59 kDa regulatory subunit B|serine/threonine-protein phosphatase 2A regulatory subunit B'' subunit alpha", map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=124474346, end_position_on_the_genomic_accession=124476862, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19055, CurrentID=0, Status='live', Symbol='Ppp3ca', Aliases='2900074D19Rik, CN, Caln, Calna, CnA', description='protein phosphatase 3, catalytic subunit, alpha isoform', other_designations='CAM-PRP catalytic subunit|PP2B alpha 1|PP2BA alpha|calcineurin A alpha|calmodulin-dependent calcineurin A subunit alpha isoform', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=136670066, end_position_on_the_genomic_accession=136937727, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  19056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19056, CurrentID=0, Status='live', Symbol='Ppp3cb', Aliases='1110063J16Rik, Calnb, CnAbeta, Cnab', description='protein phosphatase 3, catalytic subunit, beta isoform', other_designations='CAM-PRP catalytic subunit|PP2BA beta|calmodulin-dependent calcineurin A subunit beta isoform', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20499314, end_position_on_the_genomic_accession=20546573, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  19057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19057, CurrentID=0, Status='live', Symbol='Ppp3cc', Aliases='Calnc, PP2BA gamma', description='protein phosphatase 3, catalytic subunit, gamma isoform', other_designations='CAM-PRP catalytic subunit|calcineurin, testis-specific catalytic subunit|calmodulin-dependent calcineurin A subunit gamma isoform', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70217864, end_position_on_the_genomic_accession=70289508, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  19058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19058, CurrentID=0, Status='live', Symbol='Ppp3r1', Aliases='CaNB1, Cnb1, MCIP1', description='protein phosphatase 3, regulatory subunit B, alpha isoform (calcineurin B, type I)', other_designations='PP2B beta 1|protein phosphatase 2B regulatory subunit 1|protein phosphatase 3 regulatory subunit B alpha', map_location='11 A3.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=17159298, end_position_on_the_genomic_accession=17200380, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19059, CurrentID=0, Status='live', Symbol='Ppp3r2', Aliases='CaNB2, CnB2', description='protein phosphatase 3, regulatory subunit B, alpha isoform (calcineurin B, type II)', other_designations='PP2B beta 2|calcineurin B subunit|protein phospatase 3, regulatory subunit B, alpha isoform (calcineurin B, type II)|protein phosphatase 2B regulatory subunit 2|protein phosphatase 3 regulatory subunit B beta isoform', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49678747, end_position_on_the_genomic_accession=49681983, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19060, CurrentID=0, Status='live', Symbol='Ppp5c', Aliases='AU020526, PP5', description='protein phosphatase 5, catalytic subunit', other_designations='ANP receptor|PPT|protein phosphatase T', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17004640, end_position_on_the_genomic_accession=17027914, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  19062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19062, CurrentID=0, Status='live', Symbol='Inpp5k', Aliases='C62, Pps, SKIP', description='inositol polyphosphate 5-phosphatase K', other_designations='PI-5-phosphatase related|putative PI-5-phosphatase|skeletal muscle and kidney enriched inositol polyphosphate phosphatase|skeletal muscle and kidney-enriched inositol phosphatase', map_location='11 45.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75630507, end_position_on_the_genomic_accession=75648865, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19063, CurrentID=0, Status='live', Symbol='Ppt1', Aliases='9530043G02Rik, AA960502, C77813, CLN1, D4Ertd184e, INCL, PPT', description='palmitoyl-protein thioesterase 1', other_designations='PPT-1|palmitoyl-protein hydrolase 1', map_location='4 D1-D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=122836075, end_position_on_the_genomic_accession=122859175, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19064, CurrentID=0, Status='live', Symbol='Ppy', Aliases='AV061413, IGSPP, PP', description='pancreatic polypeptide', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102099931, end_position_on_the_genomic_accession=102101337, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19065, CurrentID=0, Status='live', Symbol='Npy4r', Aliases='NYYR-D, Ppyr1, Y4', description='neuropeptide Y receptor Y4', other_designations='NPY4-R|NPYR-D|PP1|Y4 receptor|pancreatic polypeptide receptor 1', map_location='14 20.8 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34145646, end_position_on_the_genomic_accession=34152419, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19069, CurrentID=0, Status='live', Symbol='Nup88', Aliases='Nup84, Prei2', description='nucleoporin 88', other_designations='88 kDa nucleoporin|preimplantation protein 2', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70943058, end_position_on_the_genomic_accession=70969973, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  19070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19070, CurrentID=0, Status='live', Symbol='Mob4', Aliases='2610109B12Rik, Mobkl3, Phocn, Prei3', description='MOB family member 4, phocein', other_designations='MOB1, Mps One Binder kinase activator-like 3|class II mMOB1|mob1 homolog 3|mob3|mps one binder kinase activator-like 3|preimplantation protein 3', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55130936, end_position_on_the_genomic_accession=55154899, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19072, CurrentID=0, Status='live', Symbol='Prep', Aliases='AI047692, AI450383, D10Wsu136e, PEP, Pop', description='prolyl endopeptidase', other_designations='PE|post-proline cleaving enzyme|prolyl oligopeptidase', map_location='10 B2-B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=45067214, end_position_on_the_genomic_accession=45159981, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  19073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19073, CurrentID=0, Status='live', Symbol='Srgn', Aliases='Prg, Prg1, Sgc', description='serglycin', other_designations='gp600|mastocytoma proteoglycan core protein|proteoglycan 1, secretory granule|proteoglycan, secretory granule|secretory granule proteoglycan core protein', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62494427, end_position_on_the_genomic_accession=62527446, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19074, CurrentID=0, Status='live', Symbol='Prg2', Aliases='MBP, mMBP, mMBP-1', description='proteoglycan 2, bone marrow', other_designations='BMPG|major basic protein 1', map_location='2 49.45 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84980461, end_position_on_the_genomic_accession=84983632, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19075, CurrentID=0, Status='live', Symbol='Prim1', Aliases='AI324982', description='DNA primase, p49 subunit', other_designations='DNA primase 49 kDa subunit|p49', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128015215, end_position_on_the_genomic_accession=128030030, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19076, CurrentID=0, Status='live', Symbol='Prim2', Aliases='AI323589', description='DNA primase, p58 subunit', other_designations='DNA primase 58 kDa subunit', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=33453808, end_position_on_the_genomic_accession=33669794, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  19079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19079, CurrentID=0, Status='live', Symbol='Prkab1', Aliases='1300015D22Rik, AU021155, E430008F22', description='protein kinase, AMP-activated, beta 1 non-catalytic subunit', other_designations='AMPK subunit beta-1|AMPKb', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=116013586, end_position_on_the_genomic_accession=116024503, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19082, CurrentID=0, Status='live', Symbol='Prkag1', Aliases='AA571379, BB036179, Prkaac', description='protein kinase, AMP-activated, gamma 1 non-catalytic subunit', other_designations='AMPK gamma1|AMPK subunit gamma-1|AMPKg', map_location='15 54.73 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98812797, end_position_on_the_genomic_accession=98831508, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19084, CurrentID=0, Status='live', Symbol='Prkar1a', Aliases='1300018C22Rik, RIalpha, Tse-1, Tse1', description='protein kinase, cAMP dependent regulatory, type I, alpha', other_designations='protein kinase, cAMP dependent regulatory, type 1, alpha', map_location='11 72.33 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109649395, end_position_on_the_genomic_accession=109669659, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19085, CurrentID=0, Status='live', Symbol='Prkar1b', Aliases='AI385716, RIbeta', description='protein kinase, cAMP dependent regulatory, type I beta', other_designations='cAMP dependent regulatory type I beta subunit transcript 1|cAMP dependent regulatory type I beta subunit transcript 2|cAMP dependent regulatory type I beta subunit transcript 3|cAMP dependent regulatory type I beta subunit transcript 4', map_location='5 77.74 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139017304, end_position_on_the_genomic_accession=139130415, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  19087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19087, CurrentID=0, Status='live', Symbol='Prkar2a', Aliases='1110061A24Rik, AI317181, AI836829, RII(alpha)', description='protein kinase, cAMP dependent regulatory, type II alpha', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108692143, end_position_on_the_genomic_accession=108749512, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19088, CurrentID=0, Status='live', Symbol='Prkar2b', Aliases='AI451071, AW061005, PKARIIbeta, Pkarb2, RII(beta)', description='protein kinase, cAMP dependent regulatory, type II beta', other_designations='', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31958450, end_position_on_the_genomic_accession=32061343, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19089, CurrentID=0, Status='live', Symbol='Prkcsh', Aliases='80K-H, PKCSH', description='protein kinase C substrate 80K-H', other_designations='alpha glucosidase II, beta-subunit|carbohydrate processing enzyme of the endoplasmic reticulum|glucosidase II beta-subunit|hepatocystin|protein kinase C substrate 60.1 kDa protein heavy chain', map_location='9 8.04 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22002988, end_position_on_the_genomic_accession=22014245, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  19090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19090, CurrentID=0, Status='live', Symbol='Prkdc', Aliases='AI326420, AU019811, DNA-PKcs, DNAPDcs, DNAPK, DNPK1, DOXNPH, HYRC1, XRCC7, dxnph, p460, scid, slip', description='protein kinase, DNA activated, catalytic polypeptide', other_designations='DNA-PK catalytic subunit|doxorubicin nephropathy|severe combined immunodeficiency', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=15637444, end_position_on_the_genomic_accession=15842252, orientation='plus', exon_count=87, OMIM='', no_hdr0=''),
  19091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19091, CurrentID=0, Status='live', Symbol='Prkg1', Aliases='AW125416, CGKI, Gm19690b, Prkgr1b, Prkg1', description='protein kinase, cGMP-dependent, type I', other_designations='cGK 1|cGK1|cGMP-dependent protein kinase I', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=30564487, end_position_on_the_genomic_accession=31765304, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  19092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19092, CurrentID=0, Status='live', Symbol='Prkg2', Aliases='AW212535, CGKII, Prkgr2, cGK-II', description='protein kinase, cGMP-dependent, type II', other_designations='cGK 2|cGK2|cGMP-dependent protein kinase II', map_location='5 48.35 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=98929773, end_position_on_the_genomic_accession=99038718, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  19094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19094, CurrentID=0, Status='live', Symbol='Mapk11', Aliases='P38b, Prkm11, Sapk2, Sapk2b, p38-2, p38beta, p38beta2', description='mitogen-activated protein kinase 11', other_designations='MAP kinase 11|MAP kinase p38 beta|MAPK 11|mitogen activated protein kinase 11|mitogen-activated protein kinase p38 beta|protein kinase, mitogen activated kinase, 11, p38beta', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89142481, end_position_on_the_genomic_accession=89149623, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19099, CurrentID=0, Status='live', Symbol='Mapk8ip1', Aliases='IB1, JIP-1, Jip1, Prkm8ip, Skip, mjip-2a', description='mitogen-activated protein kinase 8 interacting protein 1', other_designations='IB-1|JNK MAP kinase scaffold protein 1|JNK-interacting protein 1|islet-brain-1|mitogen activated protein kinase 8 interacting protein 1|mitogen-activated protein kinase 8-interacting protein 1|protein kinase, mitogen-activated 8 interacting protein', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92383676, end_position_on_the_genomic_accession=92401263, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19106, CurrentID=0, Status='live', Symbol='Eif2ak2', Aliases='2310047A08Rik, 4732414G15Rik, AI467567, AI747578, Pkr, Prkr, Tik', description='eukaryotic translation initiation factor 2-alpha kinase 2', other_designations='IFN- type I-induced and dsRNA-activated kinase|IFN-induced and double-stranded RNA-activated kinase|P1/eIF-2A protein kinase|T-cell viral integration site|dsRNA-activated kinase|eIF-2 alpha|eIF-2A protein kinase 2|eukaryotic translation initiation factor 2 alpha kinase 2|interferon-inducible RNA-dependent protein kinase|p68 kinase|protein kinase R|protein kinase RNA-activated|protein kinase, interferon inducible double stranded RNA dependent|protein kinase, interferon-inducible double stranded RNA dependent|serine/threonine-protein kinase TIK|tyrosine-protein kinase EIF2AK2', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78852550, end_position_on_the_genomic_accession=78882572, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  19108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19108, CurrentID=0, Status='live', Symbol='Prkx', Aliases='PKX1, Pkare', description='protein kinase, X-linked', other_designations='PKA-related protein kinase|protein kinase X|protein kinase X-linked|putative serine/threonine kinase|serine/threonine-protein kinase PRKX', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=77760972, end_position_on_the_genomic_accession=77796288, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19109, CurrentID=0, Status='live', Symbol='Prl', Aliases='AV290867, Gha11a1, Prl', description='prolactin', other_designations='growth hormone a1', map_location='13 12.25 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27057570, end_position_on_the_genomic_accession=27065204, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19110, CurrentID=0, Status='live', Symbol='Prl4a1', Aliases='AA409771, Ghd23, PLP-A, Prlpa', description='prolactin family 4, subfamily a, member 1', other_designations='PRL-like protein A|growth hormone d23|placental prolactin-like protein A|prolactin-like protein A', map_location='13 12.66 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=28016223, end_position_on_the_genomic_accession=28023546, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19111, CurrentID=0, Status='live', Symbol='Prl6a1', Aliases='Ghd10, PLP-B, Prlpb', description='prolactin family 6, subfamily a, member 1', other_designations='PRL-like protein B|growth hormone d10|placental prolactin-like protein B|prolactin-like protein B', map_location='13 12.39 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27312627, end_position_on_the_genomic_accession=27319252, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19112, CurrentID=0, Status='live', Symbol='Prl8a6', Aliases='Ghd13, PLP-C1, PLP-Ca, PLP-calpha, Prlpc, Prlpc1', description='prolactin family 8, subfamily a, member 6', other_designations='PRL-like protein C1|growth hormone d13|placental prolactin-like protein C1|prolactin-like protein C 1|prolactin-like protein C-alpha', map_location='13 12.45 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27432681, end_position_on_the_genomic_accession=27438688, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19113, CurrentID=0, Status='live', Symbol='Prl7a1', Aliases='AA407021, AA409971, AA410116, PLP-E, PLP-G, Prlpe, Prlpg', description='prolactin family 7, subfamily a, member 1', other_designations='PRL-like protein E|PRL-like protein G|placental prolactin-like protein E|placental prolactin-like protein G|prolactin-like protein E|prolactin-like protein G', map_location='13 12.56 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27633364, end_position_on_the_genomic_accession=27642493, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19114, CurrentID=0, Status='live', Symbol='Prl7a2', Aliases='Ghd18, PLP-F, Prlpf', description='prolactin family 7, subfamily a, member 2', other_designations='PRL-like protein F|growth hormone d18|placental prolactin-like protein F|prolactin-like protein F', map_location='13 12.58 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27658584, end_position_on_the_genomic_accession=27668052, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19116, CurrentID=0, Status='live', Symbol='Prlr', Aliases='AI987712, Pr-1, Pr-3-rs1, Prlr', description='prolactin receptor', other_designations='PRL-R|prolactin receptor related sequence 1', map_location='15 5.23 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10177238, end_position_on_the_genomic_accession=10349190, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  19118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19118, CurrentID=0, Status='live', Symbol='Prm1', Aliases='Prm-1', description='protamine 1', other_designations='cysteine-rich protamine|sperm protamine-P1', map_location='16 cen-C3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10796332, end_position_on_the_genomic_accession=10796823, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19119, CurrentID=0, Status='live', Symbol='Prm2', Aliases='AI528784, Prm-2', description='protamine 2', other_designations='sperm histone P2|sperm protamine P2|sperm protamine-P2', map_location='16 5.84 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10791381, end_position_on_the_genomic_accession=10792097, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19120, CurrentID=0, Status='live', Symbol='Prm3', Aliases='BC061127, PX, Pxg', description='protamine 3', other_designations='sperm protamine P3', map_location='16 5.84 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10790508, end_position_on_the_genomic_accession=10790914, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19122, CurrentID=0, Status='live', Symbol='Prnp', Aliases='AA960666, AI325101, CD230, PrP, PrP<C>, PrPC, PrPSc, Prn-i, Prn-p, Sinc, prP27-30, prP33-35C', description='prion protein', other_designations='', map_location='2 64.07 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131909928, end_position_on_the_genomic_accession=131938436, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19123, CurrentID=0, Status='live', Symbol='Proc', Aliases='PC', description='protein C', other_designations='anticoagulant protein C|autoprothrombin IIA|blood coagulation factor XIV|inactivator of coagulation factors Va, VIII', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32123126, end_position_on_the_genomic_accession=32139570, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19124, CurrentID=0, Status='live', Symbol='Procr', Aliases='AI325044, Ccca, Ccd41, Epcr', description='protein C receptor, endothelial', other_designations='APC receptor|CD antigen CD201|activated protein C receptor|centrocyclin|centrosomal protein CCD41|endothelial cell protein C receptor', map_location='2 H1-3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155751125, end_position_on_the_genomic_accession=155755478, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19125, CurrentID=0, Status='live', Symbol='Prodh', Aliases='Pro-1, Pro1, Ym24d07', description='proline dehydrogenase', other_designations='proline dehydrogenase, mitochondrial|proline oxidase', map_location='16 11.19 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18071726, end_position_on_the_genomic_accession=18089283, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  19126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19126, CurrentID=0, Status='live', Symbol='Prom1', Aliases='4932416E19Rik, AC133, CD133, Prom, Prom-1, Proml1', description='prominin 1', other_designations='antigen AC133 homolog|prominin-like 1|prominin-like protein 1', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43993621, end_position_on_the_genomic_accession=44102252, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  19127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19127, CurrentID=0, Status='live', Symbol='Prop1', Aliases='Prop-1, df', description='paired like homeodomain factor 1', other_designations='Ames dwarf|pituitary specific homeodomain factor|pituitary-specific homeodomain factor|prophet of Pit-1', map_location='11 30.95 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50950806, end_position_on_the_genomic_accession=50953632, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19128, CurrentID=0, Status='live', Symbol='Pros1', Aliases='AW214361', description='protein S (alpha)', other_designations='', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=62854334, end_position_on_the_genomic_accession=62929346, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  19130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19130, CurrentID=0, Status='live', Symbol='Prox1', Aliases='A230003G05Rik', description='prospero homeobox 1', other_designations='PROX-1|homeobox prospero-like protein PROX1|prospero-related homeobox 1', map_location='1 95.85 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=190118035, end_position_on_the_genomic_accession=190170825, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19131, CurrentID=0, Status='live', Symbol='Prh1', Aliases='MP2, Prp', description='proline rich protein HaeIII subfamily 1', other_designations='proline-rich protein MP-2|proline-rich salivary protein', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132569842, end_position_on_the_genomic_accession=132572401, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19132, CurrentID=0, Status='live', Symbol='Prph', Aliases='Prph1', description='peripherin', other_designations='peripherin 1', map_location='15 E-F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99053575, end_position_on_the_genomic_accession=99058978, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19133, CurrentID=0, Status='live', Symbol='Prph2', Aliases='AOFMD, AVMD, Nmf193, PRPH, RP7, Rd-2, Rd2, Rds, Tspan22, rds', description='peripherin 2', other_designations='retinal degeneration 2|retinal degeneration slow protein|retinal degeneration, slow (retinitis pigmentosa 7)', map_location='17 22.91 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46910485, end_position_on_the_genomic_accession=46924926, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19134, CurrentID=0, Status='live', Symbol='Prpf4b', Aliases='2610037H07, Prp4, Prp4k, Prpk, cbp143', description='pre-mRNA processing factor 4B', other_designations='PRP4 pre-mRNA processing factor 4 homolog B|PRP4 pre-mRNA-processing factor 4 homolog|pre-mRNA protein kinase', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34875296, end_position_on_the_genomic_accession=34906064, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  19139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19139, CurrentID=0, Status='live', Symbol='Prps1', Aliases='2310010D17Rik, C76571, C76678, PRS-I, Prps-1', description='phosphoribosyl pyrophosphate synthetase 1', other_designations='phosphoribosyl pyrophosphate synthase I|phosphoribosyl pyrophosphate synthetase I', map_location='X F1-F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140456603, end_position_on_the_genomic_accession=140476140, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19141, CurrentID=0, Status='live', Symbol='Lgmn', Aliases='AEP, AI746452, AU022324, Prsc1', description='legumain', other_designations='asparaginyl endopeptidase|preprolegumain|protease, cysteine 1|protease, cysteine, 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102394098, end_position_on_the_genomic_accession=102439697, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  19142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19142, CurrentID=0, Status='live', Symbol='Prss12', Aliases='Bssp-3, motopsin', description='protease, serine 12 neurotrypsin (motopsin)', other_designations='brain-specific serine protease 3|protease, serine, 12 neurotrypsin (motopsin)|serine protease 12', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=123446913, end_position_on_the_genomic_accession=123506602, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19143, CurrentID=0, Status='live', Symbol='St14', Aliases='Epithin, MT-SP1, Prss14, Tmprss14, mCAP3, matriptase', description='suppression of tumorigenicity 14 (colon carcinoma)', other_designations='matriptase/MT-SP1|protease, serine, 14 (epithin)|serine protease 14', map_location='9 16.61 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=31088590, end_position_on_the_genomic_accession=31131874, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  19144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19144, CurrentID=0, Status='live', Symbol='Klk6', Aliases='AI849898, BSP, Bssp, Klk29, MSP, Prss18, Prss9, neurosin', description='kallikrein related-peptidase 6', other_designations='brain serine protease|kallikrein 29|kallikrein 6|myelencephalon-specific protease|myofibril-bound protease|protease M|protease, serine, 18|protease, serine, 9', map_location='7 B4-B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43824544, end_position_on_the_genomic_accession=43832027, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19146, CurrentID=0, Status='live', Symbol='Tmprss15', Aliases='Entk, Prss7', description='transmembrane protease, serine 15', other_designations='enterokinase|protease, serine, 7 (enterokinase)', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=78953008, end_position_on_the_genomic_accession=79091097, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  19152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19152, CurrentID=0, Status='live', Symbol='Prtn3', Aliases='PR3, mPR3', description='proteinase 3', other_designations='PR-3', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79879667, end_position_on_the_genomic_accession=79883174, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19153, CurrentID=0, Status='live', Symbol='Prx', Aliases='L-Periaxin', description='periaxin', other_designations='', map_location='7 15.91 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27499305, end_position_on_the_genomic_accession=27520041, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19155, CurrentID=0, Status='live', Symbol='Npepps', Aliases='AAP-S, MP100, Psa, R74825, goku', description='aminopeptidase puromycin sensitive', other_designations='cytosol alanyl aminopeptidase', map_location='11 60.95 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97205842, end_position_on_the_genomic_accession=97280638, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  19156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19156, CurrentID=0, Status='live', Symbol='Psap', Aliases='AI037048, SGP-1', description='prosaposin', other_designations='snoRNA MBII-198|sulfated glycoprotein 1', map_location='10 30.02 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60277628, end_position_on_the_genomic_accession=60302600, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  19157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19157, CurrentID=0, Status='live', Symbol='Cyth1', Aliases='CLM1, CTH-1, CYTIP, Pscd1, mKIAA4240', description='cytohesin 1', other_designations='PH, SEC7 and coiled-coil domain-containing protein 1|SEC7 homolog A|mSec7-1|pleckstrin homology, Sec7 and coiled-coil domains 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118164166, end_position_on_the_genomic_accession=118248592, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  19158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19158, CurrentID=0, Status='live', Symbol='Cyth2', Aliases='ARNO, CLM2, Pscd2', description='cytohesin 2', other_designations='ARF nucleotide-binding site opener|PH, SEC7 and coiled-coil domain-containing protein 2|SEC7 homolog B|mSec7-2|pleckstrin homology, Sec7 and coiled-coil domains 2', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45806637, end_position_on_the_genomic_accession=45814316, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  19159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19159, CurrentID=0, Status='live', Symbol='Cyth3', Aliases='AI648983, ARNO3, CLM3, Grp1, Pscd3', description='cytohesin 3', other_designations='ARF nucleotide-binding site opener 3|PH, SEC7 and coiled-coil domain-containing protein 3|SEC7 homolog C|general receptor of phosphoinositides 1|mSec7-3|pleckstrin homology, Sec7 and coiled-coil domains 3', map_location='5 82.36 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143622447, end_position_on_the_genomic_accession=143710250, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  19164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19164, CurrentID=0, Status='live', Symbol='Psen1', Aliases='Ad3h, PS-1, PS1, S182', description='presenilin 1', other_designations='', map_location='12 38.84 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83688146, end_position_on_the_genomic_accession=83738446, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19165, CurrentID=0, Status='live', Symbol='Psen2', Aliases='ALG-3, Ad4h, Alg3, PS-2, PS2, Psnl2, STM2', description='presenilin 2', other_designations='', map_location='1 84.19 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180227004, end_position_on_the_genomic_accession=180263476, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19166, CurrentID=0, Status='live', Symbol='Psma2', Aliases='Lmpc3', description='proteasome (prosome, macropain) subunit, alpha type 2', other_designations='macropain subunit C3|multicatalytic endopeptidase complex subunit C3|proteasome component C3', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=14613242, end_position_on_the_genomic_accession=14625673, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19167, CurrentID=0, Status='live', Symbol='Psma3', Aliases='Lmpc8', description='proteasome (prosome, macropain) subunit, alpha type 3', other_designations='macropain subunit C8|multicatalytic endopeptidase complex subunit C8|proteasome component C8|proteasome subunit K', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70974623, end_position_on_the_genomic_accession=70994937, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19170, CurrentID=0, Status='live', Symbol='Psmb1', Aliases='AA409053, C81484, Lmpc5', description='proteasome (prosome, macropain) subunit, beta type 1', other_designations='component C5 of proteasome|macropain subunit C5|multicatalytic endopeptidase complex subunit C5|proteasome beta-subunit C5|proteasome component C5|proteasome gamma chain|proteasome subunit C5', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15475721, end_position_on_the_genomic_accession=15498276, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19171, CurrentID=0, Status='live', Symbol='Psmb10', Aliases='Mecl-1, Mecl1', description='proteasome (prosome, macropain) subunit, beta type 10', other_designations='low molecular mass protein 10|macropain subunit MECl-1|multicatalytic endopeptidase complex subunit MECl-1|prosome Mecl1|proteasome (prosomome, macropain) subunit, beta type 10|proteasome MECl-1|proteasome subunit MECL1|proteasome subunit beta-2i', map_location='8 53.06 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105935728, end_position_on_the_genomic_accession=105938454, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19172, CurrentID=0, Status='live', Symbol='Psmb4', Aliases='Pros-27', description='proteasome (prosome, macropain) subunit, beta type 4', other_designations='low molecular mass protein 3|macropain beta chain|multicatalytic endopeptidase complex beta chain|proteasome beta chain|proteasome chain 3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94884324, end_position_on_the_genomic_accession=94886958, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19173, CurrentID=0, Status='live', Symbol='Psmb5', Aliases='', description='proteasome (prosome, macropain) subunit, beta type 5', other_designations='macropain epsilon chain|multicatalytic endopeptidase complex epsilon chain|proteasome beta type subunit 5|proteasome chain 6|proteasome epsilon chain|proteasome subunit X', map_location='14 C2-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54614120, end_position_on_the_genomic_accession=54617995, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19175, CurrentID=0, Status='live', Symbol='Psmb6', Aliases='Lmp19, Mpnd', description='proteasome (prosome, macropain) subunit, beta type 6', other_designations='low molecular mass protein 19|macropain delta chain|multicatalytic endopeptidase complex delta chain|proteasome delta chain|proteasome subunit Y|subunit 2', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70525357, end_position_on_the_genomic_accession=70527858, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19177, CurrentID=0, Status='live', Symbol='Psmb7', Aliases='AU020723, MC14', description='proteasome (prosome, macropain) subunit, beta type 7', other_designations='macropain chain Z|multicatalytic endopeptidase complex chain Z|proteasome subunit Z', map_location='2 24.42 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38588046, end_position_on_the_genomic_accession=38643952, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19179, CurrentID=0, Status='live', Symbol='Psmc1', Aliases='AI325227, P26s4, Rpt2/S4, S4, rpt2', description='protease (prosome, macropain) 26S subunit, ATPase 1', other_designations='26S proteasome AAA-ATPase subunit RPT2|proteasome 26S subunit ATPase 1', map_location='12 50.43 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100110202, end_position_on_the_genomic_accession=100123364, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19181, CurrentID=0, Status='live', Symbol='Psmc2', Aliases='', description='proteasome (prosome, macropain) 26S subunit, ATPase 2', other_designations='26S proteasome AAA-ATPase subunit RPT1|proteasome 26S subunit ATPase 2', map_location='5 9.97 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21785283, end_position_on_the_genomic_accession=21803784, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19182, CurrentID=0, Status='live', Symbol='Psmc3', Aliases='TBP-1', description='proteasome (prosome, macropain) 26S subunit, ATPase 3', other_designations='26S proteasome AAA-ATPase subunit RPT5|TAT-binding protein 1|Tat binding protein 1|proteasome 26S subunit ATPase 3', map_location='2 50.44 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91054016, end_position_on_the_genomic_accession=91059438, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19183, CurrentID=0, Status='live', Symbol='Psmc3ip', Aliases='C79099, GT198, HOP2, Tbpip', description='proteasome (prosome, macropain) 26S subunit, ATPase 3, interacting protein', other_designations='PSMC3-interacting protein|TBP-1-interacting protein|proteasome 26S ATPase subunit 3-interacting protein|tat-binding protein 1-interacting protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101092141, end_position_on_the_genomic_accession=101095435, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19184, CurrentID=0, Status='live', Symbol='Psmc5', Aliases='mSUG1', description='protease (prosome, macropain) 26S subunit, ATPase 5', other_designations='26S proteasome AAA-ATPase subunit RPT6|p45/SUG|proteasome 26S subunit ATPase 5|proteasome subunit p45', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106256185, end_position_on_the_genomic_accession=106263112, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19185, CurrentID=0, Status='live', Symbol='Psmd4', Aliases='Af1, Mcb1, angiocidin', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 4', other_designations='26S proteasome regulatory subunit RPN10|26S proteasome regulatory subunit S5A|RPN10 homolog|anti secretory factor 1|multiubiquitin chain-binding protein|multiubiquitin-chain-binding protein', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95032688, end_position_on_the_genomic_accession=95042614, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19186, CurrentID=0, Status='live', Symbol='Psme1', Aliases='AW413925, PA28a', description='proteasome (prosome, macropain) activator subunit 1 (PA28 alpha)', other_designations='11S regulator complex subunit alpha|PA28 alpha subunit|PA28alpha|REG-alpha|activator of multicatalytic protease subunit 1|protease (prosome, macropain) 28 subunit, alpha|proteasome (prosome, macropain) 28 subunit, alpha|proteasome activator 28 subunit alpha', map_location='14 C2-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55578494, end_position_on_the_genomic_accession=55581527, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19188, CurrentID=0, Status='live', Symbol='Psme2', Aliases='AA589371, AI788882, PA28b', description='proteasome (prosome, macropain) activator subunit 2 (PA28 beta)', other_designations='11S regulator complex subunit beta|PA28beta|REG-beta|activator of multicatalytic protease subunit 2|protease (prosome, macropain) 28 subunit, beta|proteasome (prosome, macropain) 28 subunit, beta|proteasome activator 28 subunit beta', map_location='14 C2-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55587440, end_position_on_the_genomic_accession=55591146, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  19192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19192, CurrentID=0, Status='live', Symbol='Psme3', Aliases='AA410043, AU020960, Ki, PA28gamma, REGgamma, pa28g', description='proteaseome (prosome, macropain) activator subunit 3 (PA28 gamma, Ki)', other_designations='11S regulator complex subunit gamma|REG-gamma|activator of multicatalytic protease subunit 3|ki nuclear autoantigen|proteaseome (prosome, macropain) 28 subunit, 3|proteasome activator 28 subunit gamma', map_location='11 64.67 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101316251, end_position_on_the_genomic_accession=101323531, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19193, CurrentID=0, Status='live', Symbol='Pipox', Aliases='LPIPOX, Pso', description='pipecolic acid oxidase', other_designations='L-pipecolate oxidase|L-pipecolic acid oxidase', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77880614, end_position_on_the_genomic_accession=77893872, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19194, CurrentID=0, Status='live', Symbol='Bpifa2', Aliases='Bpifa2e, Psp', description='BPI fold containing family A, member 2', other_designations='BPI fold containing family A, member 2E|parotid secretory protein', map_location='2 76.15 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154008276, end_position_on_the_genomic_accession=154016073, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19197, CurrentID=0, Status='live', Symbol='Pspn', Aliases='PSP', description='persephin', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56999457, end_position_on_the_genomic_accession=57000018, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19200, CurrentID=0, Status='live', Symbol='Pstpip1', Aliases='CD2BP1, def-2', description='proline-serine-threonine phosphatase-interacting protein 1', other_designations='PEST phosphatase-interacting protein 1', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56078445, end_position_on_the_genomic_accession=56128890, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  19201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19201, CurrentID=0, Status='live', Symbol='Pstpip2', Aliases='MAYP, cmo', description='proline-serine-threonine phosphatase-interacting protein 2', other_designations='PEST phosphatase-interacting protein 2|macrophage actin-associated tyrosine-phosphorylated protein|pp37', map_location='18 52.38 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77794550, end_position_on_the_genomic_accession=77882879, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  19202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19202, CurrentID=0, Status='live', Symbol='Rhox6', Aliases='AV213605, AW548842, Psx1', description='reproductive homeobox 6', other_designations='placenta specific homeobox 1', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37827055, end_position_on_the_genomic_accession=37829857, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19204, CurrentID=0, Status='live', Symbol='Ptafr', Aliases='PAFR', description='platelet-activating factor receptor', other_designations='PAF receptor', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132564067, end_position_on_the_genomic_accession=132580866, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19205, CurrentID=0, Status='live', Symbol='Ptbp1', Aliases='AA407203, AL033359, HNRPI, PTB-1, PTB2, PTB3, PTB4, Ptb, pPTB', description='polypyrimidine tract binding protein 1', other_designations='heterogeneous nuclear ribonucleoprotein I|hnRNP I', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79854432, end_position_on_the_genomic_accession=79864435, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  19206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19206, CurrentID=0, Status='live', Symbol='Ptch1', Aliases='A230106A15Rik, Ptc, Ptc1, Ptch, mes', description='patched homolog 1', other_designations='Patched 1|Ptch1 protein +12b|mesenchymal dysplasia', map_location='13 32.8 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=63508328, end_position_on_the_genomic_accession=63573616, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  19207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19207, CurrentID=0, Status='live', Symbol='Ptch2', Aliases='ptc2', description='patched homolog 2', other_designations='', map_location='4 C-D', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117095785, end_position_on_the_genomic_accession=117115383, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  19208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19208, CurrentID=0, Status='live', Symbol='Ptcra', Aliases='gp33, pT-alpha, pT-alpha-TCR, pT[a], pTa, pTalpha', description='pre T cell antigen receptor alpha', other_designations='', map_location='17 D-E1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46755663, end_position_on_the_genomic_accession=46763712, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19210, CurrentID=0, Status='live', Symbol='Ptdss1', Aliases='AU044268, AW539008, PSS-1, mKIAA0024', description='phosphatidylserine synthase 1', other_designations='PtdSer Synthase-1|ptdSer synthase 1|serine-exchange enzyme I', map_location='13 B-C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=66932830, end_position_on_the_genomic_accession=66998401, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19211, CurrentID=0, Status='live', Symbol='Pten', Aliases='2310035O07Rik, A130070J02Rik, AI463227, B430203M17Rik, MMAC1, TEP1', description='phosphatase and tensin homolog', other_designations='mitochondrial PTENalpha|mutated in multiple advanced cancers 1|phosphatidylinositol-3,4,5-trisphosphate 3-phosphatase and dual-specificity protein phosphatase PTEN', map_location='19 28.14 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=32757577, end_position_on_the_genomic_accession=32826160, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19212, CurrentID=0, Status='live', Symbol='Pter', Aliases='AI790318, Mpr56-1', description='phosphotriesterase related', other_designations='parathion hydrolase (phosphotriesterase)-related protein', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=12924034, end_position_on_the_genomic_accession=13003453, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19213, CurrentID=0, Status='live', Symbol='Ptf1a', Aliases='PTF1-p48, PTF1p48, bHLHa29', description='pancreas specific transcription factor, 1a', other_designations='bHLH transcription factor p48|p48 DNA-binding subunit of transcription factor PTF1|pancreas transcription factor1 p48 subunit|pancreas-specific transcription factor 1a', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=19445663, end_position_on_the_genomic_accession=19447501, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19214, CurrentID=0, Status='live', Symbol='Ptgdr', Aliases='DP, PGD', description='prostaglandin D receptor', other_designations='PGD receptor|PGD2 receptor|prostanoid DP receptor', map_location='14 22.59 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=44851235, end_position_on_the_genomic_accession=44859444, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19215, CurrentID=0, Status='live', Symbol='Ptgds', Aliases='21kDa, L-PGDS, PGD2, PGDS, PGDS2, Ptgs3', description='prostaglandin D2 synthase (brain)', other_designations='PGD2 synthase|glutathione-independent PGD synthase|glutathione-independent PGD synthetase|lipocalin-type prostaglandin-D synthase|prostaglandin D2 synthase (21 kDa, brain)|prostaglandin-D2 synthase', map_location='2 17.28 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25466709, end_position_on_the_genomic_accession=25470109, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19216, CurrentID=0, Status='live', Symbol='Ptger1', Aliases='42kDa, EP1, Ptgerep1', description='prostaglandin E receptor 1 (subtype EP1)', other_designations='PGE receptor EP1 subtype|PGE2 receptor EP1 subtype|prostanoid EP1 receptor', map_location='8 40.22 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83666530, end_position_on_the_genomic_accession=83670103, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19217, CurrentID=0, Status='live', Symbol='Ptger2', Aliases='EP2, Ptgerep2', description='prostaglandin E receptor 2 (subtype EP2)', other_designations='EP2 receptor|PGE receptor EP2 subtype|PGE2 receptor EP2 subtype|prostaglandin E receptor 2 (EP2 subtype)|prostaglandin E receptor EP2 subtype|prostaglandin E receptor subtype EP2|prostanoid EP2 receptor', map_location='14 22.68 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=44988111, end_position_on_the_genomic_accession=45003820, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19218, CurrentID=0, Status='live', Symbol='Ptger3', Aliases='EP3, Pgerep3, Ptgerep3', description='prostaglandin E receptor 3 (subtype EP3)', other_designations='PGE receptor EP3 subtype|PGE2 receptor EP3 subtype|prostaglandin E receptor EP3 subtype|prostanoid EP3 receptor', map_location='3 81.32 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=157566816, end_position_on_the_genomic_accession=157644758, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19219, CurrentID=0, Status='live', Symbol='Ptger4', Aliases='EP4, Ptgerep4', description='prostaglandin E receptor 4 (subtype EP4)', other_designations='PGE2 receptor EP4 subtype|prostanoid EP4 receptor', map_location='15 1.99 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=5233399, end_position_on_the_genomic_accession=5244187, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19220, CurrentID=0, Status='live', Symbol='Ptgfr', Aliases='AI957154, PGF, fp', description='prostaglandin F receptor', other_designations='PGF receptor|PGF2-alpha receptor|prostaglandin F2alpha receptor|prostanoid FP receptor', map_location='3 76.96 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=151798610, end_position_on_the_genomic_accession=151837528, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19221, CurrentID=0, Status='live', Symbol='Ptgfrn', Aliases='4833445A08Rik, AU042434, CD9P-1, FPRP, Trim45', description='prostaglandin F2 receptor negative regulator', other_designations='prostaglandin F2-alpha receptor regulatory protein|prostaglandin F2-alpha receptor-associated protein', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=101040230, end_position_on_the_genomic_accession=101111105, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  19222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19222, CurrentID=0, Status='live', Symbol='Ptgir', Aliases='IP, PGI2', description='prostaglandin I receptor (IP)', other_designations='PGI receptor|PGI2 receptor|prostaglandin I2 receptor|prostanoid IP receptor', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16905419, end_position_on_the_genomic_accession=16910905, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19223, CurrentID=0, Status='live', Symbol='Ptgis', Aliases='Cyp8, Cyp8a1, Pgis', description='prostaglandin I2 (prostacyclin) synthase', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167203194, end_position_on_the_genomic_accession=167252127, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19224, CurrentID=0, Status='live', Symbol='Ptgs1', Aliases='COX1, Cox-1, Cox-3, PGHS-1, PHS 1, Pghs1', description='prostaglandin-endoperoxide synthase 1', other_designations='PGH synthase 1|cyclooxygenase-1|prostaglandin H2 synthase 1', map_location='2 24.19 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36216743, end_position_on_the_genomic_accession=36252275, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  19225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19225, CurrentID=0, Status='live', Symbol='Ptgs2', Aliases='COX2, Cox-2, PES-2, PGHS-2, PHS II, PHS-2, Pghs2, TIS10, gripghs', description='prostaglandin-endoperoxide synthase 2', other_designations='PGH synthase 2|cyclooxygenase-2|glucocorticoid-regulated inflammatory cyclooxygenase|macrophage activation-associated marker protein P71/73|prostaglandin H2 synthase 2', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=150100124, end_position_on_the_genomic_accession=150108017, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19226, CurrentID=0, Status='live', Symbol='Pth', Aliases='Pthp', description='parathyroid hormone', other_designations='', map_location='7 F', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=113385576, end_position_on_the_genomic_accession=113388573, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19227, CurrentID=0, Status='live', Symbol='Pthlh', Aliases='PLP, PTH-like, Pthrp', description='parathyroid hormone-like peptide', other_designations='PTH-related peptide|parathyroid hormone-like hormone|parathyroid hormone-like protein|parathyroid hormone-related peptide', map_location='6 F-G', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=147252110, end_position_on_the_genomic_accession=147264085, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19228, CurrentID=0, Status='live', Symbol='Pth1r', Aliases='PPR, Pthr, Pthr1', description='parathyroid hormone 1 receptor', other_designations='PTH-related peptide receptor|PTH/PTHr receptor|PTH/PTHrP receptor|PTH/PTHrP type I receptor|PTH1 receptor|parathyroid hormone receptor 1', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110722085, end_position_on_the_genomic_accession=110747145, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  19229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19229, CurrentID=0, Status='live', Symbol='Ptk2b', Aliases='CADTK, CAKB, CAKbeta, E430023O05Rik, FADK2, FAK2, PYK2, Raftk', description='PTK2 protein tyrosine kinase 2 beta', other_designations='CAK-beta|FADK 2|calcium-dependent tyrosine kinase|calcium-regulated non-receptor proline-rich tyrosine kinase|cell adhesion kinase beta|cellular adhesion kinase beta|focal adhesion kinase 2|proline-rich tyrosine kinase 2|protein tyrosine kinase 2 beta|related adhesion focal tyrosine kinase', map_location='14 34.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66153257, end_position_on_the_genomic_accession=66281066, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  19230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19230, CurrentID=0, Status='live', Symbol='Twf1', Aliases='A6, Ptk9, twinfilin', description='twinfilin, actin-binding protein, homolog 1 (Drosophila)', other_designations='PTK9 protein tyrosine kinase 9|actin monomer-binding protein|protein A6|protein tyrosine kinase 9', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=94577948, end_position_on_the_genomic_accession=94589884, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19231, CurrentID=0, Status='live', Symbol='Ptma', Aliases='Thym', description='prothymosin alpha', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86526736, end_position_on_the_genomic_accession=86530698, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19240, CurrentID=0, Status='live', Symbol='Tmsb10', Aliases='Ptmb10, Tb10', description='thymosin, beta 10', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72957347, end_position_on_the_genomic_accession=72958748, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19241, CurrentID=0, Status='live', Symbol='Tmsb4x', Aliases='Ptmb4, Tb4', description='thymosin, beta 4, X chromosome', other_designations='prothymosin beta 4|t beta 4', map_location='X 78.28 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=167207093, end_position_on_the_genomic_accession=167213479, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19242, CurrentID=0, Status='live', Symbol='Ptn', Aliases='HARP, HB-GAM, HBBN, HBGF-8, HBNF, OSF, Osf-1, Osf1', description='pleiotrophin', other_designations='HBBM|heparin affin regulatory peptide|heparin-binding brain mitogen|heparin-binding growth factor 8|heparin-binding growth-associated molecule|heparin-binding neurite promoting factor|heparin-binding neurotorphic factor|heparin-binding neutrophic factor|osteoblast-specific factor 1|osteoblastic cell factor', map_location='6 15.48 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=36714909, end_position_on_the_genomic_accession=36811361, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19243, CurrentID=0, Status='live', Symbol='Ptp4a1', Aliases='AA415290, AU019864, C130021B01, Prl-1', description='protein tyrosine phosphatase 4a1', other_designations='protein-tyrosine phosphatase 4a1|protein-tyrosine phosphatase of regenerating liver 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=30940303, end_position_on_the_genomic_accession=30949765, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19244, CurrentID=0, Status='live', Symbol='Ptp4a2', Aliases='Prl-2', description='protein tyrosine phosphatase 4a2', other_designations='protein-tyrosine phosphatase 4a2|protein-tyrosine phosphatase of regenerating liver 2', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129820479, end_position_on_the_genomic_accession=129850003, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19245, CurrentID=0, Status='live', Symbol='Ptp4a3', Aliases='AV088979, Prl-3, pPtp4a3', description='protein tyrosine phosphatase 4a3', other_designations='protein-tyrosine phosphatase 4a3|protein-tyrosine phosphatase of regenerating liver 3', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73723145, end_position_on_the_genomic_accession=73758766, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19246, CurrentID=0, Status='live', Symbol='Ptpn1', Aliases='PTP-1B, PTP-HA2, PTP1B', description='protein tyrosine phosphatase, non-receptor type 1', other_designations='protein-protein-tyrosine phosphatase HA2|protein-tyrosine phosphatase 1B|protein-tyrosine phosphatase HA2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167932303, end_position_on_the_genomic_accession=167979385, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19247, CurrentID=0, Status='live', Symbol='Ptpn11', Aliases='2700084A17Rik, AW536184, PTP1D, PTP2C, SAP-2, SH-PTP2, SH-PTP3, SHP-2, Shp2, Syp', description='protein tyrosine phosphatase, non-receptor type 11', other_designations='SH2 domain-containing protein tyrosine phosphatase-2|protein-tyrosine phosphatase SYP', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121130533, end_position_on_the_genomic_accession=121191397, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  19248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19248, CurrentID=0, Status='live', Symbol='Ptpn12', Aliases='P19-PTP, PTP-P19, PTP-PEST, PTPG1', description='protein tyrosine phosphatase, non-receptor type 12', other_designations='MPTP-PEST|protein-tyrosine phosphatase P19', map_location='5 A3-B', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=20986645, end_position_on_the_genomic_accession=21055797, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  19249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19249, CurrentID=0, Status='live', Symbol='Ptpn13', Aliases='AI324989, PTP-BL, PTPL1, Ptpri, RIP', description='protein tyrosine phosphatase, non-receptor type 13', other_designations='PTP36|protein tyrosine phosphatase DPZPTP|protein tyrosine phosphatase PTP-BL|protein-tyrosine phosphatase RIP|tyrosine phosphatase', map_location='5 E-F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=103425187, end_position_on_the_genomic_accession=103598361, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  19250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19250, CurrentID=0, Status='live', Symbol='Ptpn14', Aliases='C130080N23Rik, OTTMUSG00000022087, PTP36', description='protein tyrosine phosphatase, non-receptor type 14', other_designations='protein-tyrosine phosphatase PTP36', map_location='1 95.03 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=189728216, end_position_on_the_genomic_accession=189876695, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  19252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19252, CurrentID=0, Status='live', Symbol='Dusp1', Aliases='3CH134, MKP1, Ptpn16, U19515, erp, mkp-1', description='dual specificity phosphatase 1', other_designations='MAP kinase phosphatase 1|mitogen-activated protein kinase phosphatase 1|mitogen-activated protein kinase phosphatase-1|protein tyrosine phosphatase, non-receptor type 16|protein-tyrosine phosphatase 3CH134|protein-tyrosine phosphatase ERP', map_location='17 A2-C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26505591, end_position_on_the_genomic_accession=26508472, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19253, CurrentID=0, Status='live', Symbol='Ptpn18', Aliases='FLP1, HSCF, PTP-HSCF, PTP-K1, Ptpk1', description='protein tyrosine phosphatase, non-receptor type 18', other_designations='FLP-1|fetal liver phosphatase 1', map_location='1 13.25 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34459746, end_position_on_the_genomic_accession=34473784, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  19255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19255, CurrentID=0, Status='live', Symbol='Ptpn2', Aliases='AI325124, Ptpt, TC-PTP', description='protein tyrosine phosphatase, non-receptor type 2', other_designations='MPTP|T cell protein-tyrosine phosphatase|protein-tyrosine phosphatase PTP-2', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67665501, end_position_on_the_genomic_accession=67724748, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19256, CurrentID=0, Status='live', Symbol='Ptpn20', Aliases='typ', description='protein tyrosine phosphatase, non-receptor type 20', other_designations='testis-specific tyrosine phosphatase', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33589201, end_position_on_the_genomic_accession=33640754, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  19258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19258, CurrentID=0, Status='live', Symbol='Ptpn4', Aliases='PTPMEG, Ptn4, TEP, TEP/mPTPMEG, hPTP-MEG', description='protein tyrosine phosphatase, non-receptor type 4', other_designations='testis-enriched protein tyrosine phosphatase', map_location='1 52.39 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=119652467, end_position_on_the_genomic_accession=119837542, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  19259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19259, CurrentID=0, Status='live', Symbol='Ptpn5', Aliases='Step', description='protein tyrosine phosphatase, non-receptor type 5', other_designations='neural-specific protein-tyrosine phosphatase|striatum-enriched protein-tyrosine phosphatase', map_location='7 B3-B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=47077795, end_position_on_the_genomic_accession=47133695, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  19260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19260, CurrentID=0, Status='live', Symbol='Ptpn22', Aliases='70zpep, PEP, Ptpn8', description='protein tyrosine phosphatase, non-receptor type 22 (lymphoid)', other_designations='PEST domain-enriched tyrosine phosphatase|hematopoietic cell protein-tyrosine phosphatase 70Z-PEP|protein tyrosine phosphatase, non-receptor type 8', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103856804, end_position_on_the_genomic_accession=103912252, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  19261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19261, CurrentID=0, Status='live', Symbol='Sirpa', Aliases='AI835480, Bit, CD172a, P84, Ptpns1, SHP-1, SHPS-1, SIRP', description='signal-regulatory protein alpha', other_designations='CD172 antigen-like family member A|SHP substrate 1|brain Ig-like molecule with tyrosine-based activation motifs|brain immunological-like with tyrosine-based motifs|inhibitory receptor SHPS-1|mSIRP-alpha1|myD-1 antigen|protein tyrosine phosphatase, non-receptor type substrate 1|signal-regulatory protein alpha-1|sirp-alpha-1', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129592606, end_position_on_the_genomic_accession=129632228, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19262, CurrentID=0, Status='live', Symbol='Ptpra', Aliases='Ptpa, Ptpalpha, Rptpalpha, Rptra, Rptralpha', description='protein tyrosine phosphatase, receptor type, A', other_designations='LCA-related phosphatase|PTPTY-28|R-PTP-alpha|protein-tyrosine phosphatase alpha', map_location='2 G', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130450252, end_position_on_the_genomic_accession=130554300, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  19263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19263, CurrentID=0, Status='live', Symbol='Ptprb', Aliases='3230402H02Rik, C130094E24, Ptpz, Rptpb, VE-PTP, Veptp', description='protein tyrosine phosphatase, receptor type, B', other_designations='R-PTP-beta|protein-tyrosine phosphatase beta|vascular endothelial protein tyrosine phosphatase', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=116275516, end_position_on_the_genomic_accession=116389538, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  19264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19264, CurrentID=0, Status='live', Symbol='Ptprc', Aliases='B220, CD45R, Cd45, L-CA, Ly-5, Lyt-4, T200, loc', description='protein tyrosine phosphatase, receptor type, C', other_designations='leukocyte common antigen|lymphocyte antigen 5|lymphocyte common antigen', map_location='1 60.73 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=138062859, end_position_on_the_genomic_accession=138175755, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  19265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19265, CurrentID=0, Status='live', Symbol='Ptprcap', Aliases='AI605264, CD-45-AP, LPAP, LSM-1', description='protein tyrosine phosphatase, receptor type, C polypeptide-associated protein', other_designations='CD45-AP|CD45-associated protein|PTPRC-associated protein', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4154646, end_position_on_the_genomic_accession=4156710, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19266, CurrentID=0, Status='live', Symbol='Ptprd', Aliases='1110002J03Rik, 3000002J10Rik, B230219D21Rik, R-PTP-delta', description='protein tyrosine phosphatase, receptor type, D', other_designations='', map_location='4 36.94 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=75941237, end_position_on_the_genomic_accession=78211895, orientation='minus', exon_count=52, OMIM='', no_hdr0=''),
  19267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19267, CurrentID=0, Status='live', Symbol='Ptpre', Aliases='PTPe, PTPepsilon, RPTPepsilon', description='protein tyrosine phosphatase, receptor type, E', other_designations='R-PTP-epsilon|protein-tyrosine phosphatase epsilon', map_location='7 81.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=135537645, end_position_on_the_genomic_accession=135686294, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  19268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19268, CurrentID=0, Status='live', Symbol='Ptprf', Aliases='AA591035, LAR, LARS, RPTP-LAR', description='protein tyrosine phosphatase, receptor type, F', other_designations='leukocyte antigen-related (LAR) PTP receptor|leukocyte common antigen related', map_location='4 C6-D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118208213, end_position_on_the_genomic_accession=118291400, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  19270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19270, CurrentID=0, Status='live', Symbol='Ptprg', Aliases='5430405N12Rik, AW046354, AW549872, RPTPgamma', description='protein tyrosine phosphatase, receptor type, G', other_designations='Protein-tyrosine phosphatase gamma precursor (R-PTP-gamma)|R-PTP-gamma', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=11553553, end_position_on_the_genomic_accession=12242039, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  19271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19271, CurrentID=0, Status='live', Symbol='Ptprj', Aliases='AI450271, BET, Byp, CD148, DEP-1, PTPbeta2, Ptpb2, RPTPJ, Scc-1, Scc1', description='protein tyrosine phosphatase, receptor type, J', other_designations='HPTP beta-like tyrosine phosphatase|R-PTP-J|R-PTP-eta|colon tumor susceptibility 1|protein-tyrosine phosphatase eta|protein-tyrosine phosphatase receptor type J|susceptibility to colon cancer 1', map_location='2 E1-2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90429756, end_position_on_the_genomic_accession=90580677, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  19272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19272, CurrentID=0, Status='live', Symbol='Ptprk', Aliases='AI853699, PTPk, RPTPkappa', description='protein tyrosine phosphatase, receptor type, K', other_designations='R-PTP-kappa|protein-tyrosine phosphatase kappa', map_location='10 15.06 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=28074331, end_position_on_the_genomic_accession=28597397, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  19273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19273, CurrentID=0, Status='live', Symbol='Ptpru', Aliases='Ftp-1, PCP-2, PTP, PTP-lambda, PTPlambda, Pcp2, Ptpf, Ptprl, R-PTP-U, R-PTP-psi, RPTPlambda', description='protein tyrosine phosphatase, receptor type, U', other_designations='Pancreatic carcinoma phosphatase 2|Protein-tyrosine phosphatase-lamda|Receptor-type protein-tyrosine phosphatase psi|protein tyrosine phosphatase, receptor type, L', map_location='4 64.5 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=131768457, end_position_on_the_genomic_accession=131838316, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  19274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19274, CurrentID=0, Status='live', Symbol='Ptprm', Aliases='RPTPmu, mKIAA4044', description='protein tyrosine phosphatase, receptor type, M', other_designations='R-PTP-mu|protein tyrosine phosphatase, receptor-type, M|protein-tyrosine phosphatase mu', map_location='17 37.88 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=66666848, end_position_on_the_genomic_accession=67354541, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  19275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19275, CurrentID=0, Status='live', Symbol='Ptprn', Aliases='IA-2, mIA-A, mKIAA4064', description='protein tyrosine phosphatase, receptor type, N', other_designations='PTP IA-2|R-PTP-N|protein tyrosine phosphatase, receptor-type, N', map_location='1 38.64 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75247035, end_position_on_the_genomic_accession=75264497, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  19276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19276, CurrentID=0, Status='live', Symbol='Ptprn2', Aliases='4930425H11Rik, IA2beta, PTP-NP, Phol, mKIAA0387, phogrin', description='protein tyrosine phosphatase, receptor type, N polypeptide 2', other_designations='IA-2 beta|PTP IA-2beta|R-PTP-N2|protein tyrosine phosphatase-NP', map_location='12 F', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=116485692, end_position_on_the_genomic_accession=117278167, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  19277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19277, CurrentID=0, Status='live', Symbol='Ptpro', Aliases='D28, GLEPP1, PTPBK, PTPROt, PTPU2, PTPoc, PTPphi, Ptpn15, R-PTP-O', description='protein tyrosine phosphatase, receptor type, O', other_designations='PTPase U2|glomerular epithelial protein 1|protein tyrosine phosphatase U2|protein tyrosine phosphatase, non-receptor type 15|receptor protein tyrosine phosphatase RO', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=137252299, end_position_on_the_genomic_accession=137464634, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  19279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19279, CurrentID=0, Status='live', Symbol='Ptprr', Aliases='Gmcp1, PTP-SL, PTPBR7, RPTPRR, mPTP213', description='protein tyrosine phosphatase, receptor type, R', other_designations='R-PTP-R|STEP-like PTPase|phosphotyrosine phosphatase 13|protein-tyrosine-phosphatase SL|protein-tyrosine-phosphatase-SL', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=116018200, end_position_on_the_genomic_accession=116274932, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  19280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19280, CurrentID=0, Status='live', Symbol='Ptprs', Aliases='AL022616, PTP, PTP-NU3, PTPNU-3, PTPsigma, Ptpt9, R-PTP-S, RPTPsigma', description='protein tyrosine phosphatase, receptor type, S', other_designations='R-PTP-sigma|Receptor-type tyrosine-protein phosphatase sigma|tissue-specific protein tyrosine phosphatase', map_location='17 29.32 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56412426, end_position_on_the_genomic_accession=56476480, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  19281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19281, CurrentID=0, Status='live', Symbol='Ptprt', Aliases='R-PTP-T, RPTP-rfo, RPTP-rho, RPTPmam4, RPTPrho, mKIAA0283, mRPTPrho', description='protein tyrosine phosphatase, receptor type, T', other_designations='receptor protein tyrosine phosphatase-rho|receptor-type tyrosine-protein phosphatase rho', map_location='2 81.91 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=161521988, end_position_on_the_genomic_accession=162661147, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  19283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19283, CurrentID=0, Status='live', Symbol='Ptprz1', Aliases='AI850339, DSD-1-PG, PTPbeta, PTPzeta, Ptprz, Ptpz, R-PTP-zeta, RPTPz, Rptpbeta', description='protein tyrosine phosphatase, receptor type Z, polypeptide 1', other_designations='3F8 chondroitin sulfate proteoglycan|3H1 keratan sulfate proteoglycan|phosphacan|protein tyrosine phosphate, receptor type, Z', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=22875458, end_position_on_the_genomic_accession=23052916, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  19284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19284, CurrentID=0, Status='live', Symbol='ptr', Aliases='', description='pulmonary tumor resistance', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19285, CurrentID=0, Status='live', Symbol='Ptrf', Aliases='2310075E07Rik, AW546441, Cav-p60, Cavin', description='polymerase I and transcript release factor', other_designations='cavin-1|ptrf', map_location='11 63.95 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100956736, end_position_on_the_genomic_accession=100971027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19286, CurrentID=0, Status='live', Symbol='Pts', Aliases='PTPS', description='6-pyruvoyl-tetrahydropterin synthase', other_designations='PTP synthase', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50521617, end_position_on_the_genomic_accession=50528641, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19288, CurrentID=0, Status='live', Symbol='Ptx3', Aliases='AI607804, TSG-14', description='pentraxin related gene', other_designations='pentaxin related|pentaxin-related protein PTX3|pentraxin 3|tumor necrosis factor-inducible gene 14 protein', map_location='3 30.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=66219887, end_position_on_the_genomic_accession=66225806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19289, CurrentID=0, Status='live', Symbol='Igdcc3', Aliases='2810401C09Rik, AI851425, Punc, WI-14920', description='immunoglobulin superfamily, DCC subclass, member 3', other_designations='amplimer WI-14920, mouse homolog|putative neuronal cell adhesion molecule', map_location='9 D-E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65141121, end_position_on_the_genomic_accession=65187031, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  19290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19290, CurrentID=0, Status='live', Symbol='Pura', Aliases='CAGER-1, Pur-alpha, ssCRE-BP', description='purine rich element binding protein A', other_designations='Pur alpha|purine-rich single-stranded DNA-binding protein alpha', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36280551, end_position_on_the_genomic_accession=36291721, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19291, CurrentID=0, Status='live', Symbol='Purb', Aliases='2310015K15Rik, AA114818, Cager-2, D11Bwg0414e', description='purine rich element binding protein B', other_designations='PurB beta|purine-rich element-binding protein B|purine-rich single-stranded DNA-binding protein beta|vascular actin single-stranded DNA-binding factor 2 p44 component', map_location='11 4.16 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6467599, end_position_on_the_genomic_accession=6476076, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19293, CurrentID=0, Status='live', Symbol='Pvalb', Aliases='PV, Parv, Pva', description='parvalbumin', other_designations='', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78191117, end_position_on_the_genomic_accession=78206351, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19294, CurrentID=0, Status='live', Symbol='Pvrl2', Aliases='AI325026, AI987993, Cd112, MPH, Pvr, Pvs, nectin-2', description='poliovirus receptor-related 2', other_designations='herpes virus entry mediator B|herpesvirus entry mediator B|hveB|mHveB|murine herpes virus entry protein B|murine herpesvirus entry protein B|poliovirus receptor homolog|poliovirus receptor-related protein 2|poliovirus sensitivity', map_location='7 9.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19716644, end_position_on_the_genomic_accession=19749573, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19296, CurrentID=0, Status='live', Symbol='Pvt1', Aliases='Ayu21-84Imeg, Gt(pU21)84Imeg, Mis-1, Mlvi-1, Pvt-1', description='plasmacytoma variant translocation 1', other_designations='', map_location='15 D2-D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=62037987, end_position_on_the_genomic_accession=62250975, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19298, CurrentID=0, Status='live', Symbol='Pex19', Aliases='Pxf', description='peroxisomal biogenesis factor 19', other_designations='peroxin-19|peroxisomal farnesylated protein|peroxisome biogenesis factor 19', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172126744, end_position_on_the_genomic_accession=172136497, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19299, CurrentID=0, Status='live', Symbol='Abcd3', Aliases='AI313901, AU018866, AW146054, PMP68, PMP70, Pxmp1', description='ATP-binding cassette, sub-family D (ALD), member 3', other_designations='68 kDa peroxisomal membrane protein|70 kDa peroxisomal membrane protein|peroxisomal membrane protein 1|peroxisomal membrane protein, 70 kDa', map_location='3 G-H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121758907, end_position_on_the_genomic_accession=121815215, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  19300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19300, CurrentID=0, Status='live', Symbol='Abcd4', Aliases='P69r, P70R, Pxmp1l', description='ATP-binding cassette, sub-family D (ALD), member 4', other_designations='PMP69|PMP70-related protein|PXMP1-L|peroxisomal membrane protein 69|peroxisomal membrane protein, 70 kDa-related', map_location='12 39.3 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84602531, end_position_on_the_genomic_accession=84617479, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  19301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19301, CurrentID=0, Status='live', Symbol='Pxmp2', Aliases='22kDa, PMP22', description='peroxisomal membrane protein 2', other_designations='22 kDa peroxisomal membrane protein|peroxisomal membrane protein 2, 22 kDa', map_location='5 53.44 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110274286, end_position_on_the_genomic_accession=110286168, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19302, CurrentID=0, Status='live', Symbol='Pex2', Aliases='D3Ertd138e, PAF-1, PMP35, Pxmp3', description='peroxisomal biogenesis factor 2', other_designations='Zellweger syndrome homolog|peroxin-2|peroxisomal membrane protein 3, 35 kDa|peroxisome assembly factor 1', map_location='3 1.96 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=5560188, end_position_on_the_genomic_accession=5576877, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19303, CurrentID=0, Status='live', Symbol='Pxn', Aliases='AW108311, AW123232, Pax', description='paxillin', other_designations='', map_location='5 56.1 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115506674, end_position_on_the_genomic_accession=115555987, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  19305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19305, CurrentID=0, Status='live', Symbol='Pex5', Aliases='AW212715, ESTM1, PTS1R, Pxr1, X83306', description='peroxisomal biogenesis factor 5', other_designations='PTS1 receptor|PTS1-BP|PXR1P|peroxin 5|peroxisomal C-terminal targeting signal import receptor|peroxisomal targeting signal 1 receptor|peroxisome biogenesis factor 5|peroxisome receptor 1', map_location='6 59.15 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124396816, end_position_on_the_genomic_accession=124416037, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19309, CurrentID=0, Status='live', Symbol='Pygm', Aliases='AI115133, PG', description='muscle glycogen phosphorylase', other_designations='myophosphorylase', map_location='19 4.53 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6384429, end_position_on_the_genomic_accession=6398459, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  19317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19317, CurrentID=0, Status='live', Symbol='Qk', Aliases='1110003F05Rik, 1500005P18I, l(17)-1Wis, l17Wis1, Qk', description='quaking', other_designations='QKI-5 protein|lethal, Chr 17, U Wisconsin 1', map_location='17 7.75 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=10202856, end_position_on_the_genomic_accession=10319361, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19324, CurrentID=0, Status='live', Symbol='Rab1a', Aliases='Gtbp, Rab-1, Rab1, Ypt1, mKIAA3012', description='RAB1A, member RAS oncogene family', other_designations='RAB1, member RAS oncogene family|YPT1-related protein|ras-related YPT1 protein', map_location='11 12.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=20201602, end_position_on_the_genomic_accession=20226856, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19325, CurrentID=0, Status='live', Symbol='Rab10', Aliases='AW107754', description='RAB10, member RAS oncogene family', other_designations='', map_location='12 1.71 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3247428, end_position_on_the_genomic_accession=3309969, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19326, CurrentID=0, Status='live', Symbol='Rab11b', Aliases='A730055L17Rik', description='RAB11B, member RAS oncogene family', other_designations='RAS oncogene family protein', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33742484, end_position_on_the_genomic_accession=33760486, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19328, CurrentID=0, Status='live', Symbol='Rab12', Aliases='2900054P15Rik, AW555448, C77700', description='RAB12, member RAS oncogene family', other_designations='rab-13', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=66494512, end_position_on_the_genomic_accession=66519934, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19329, CurrentID=0, Status='live', Symbol='Rab17', Aliases='AW413472', description='RAB17, member RAS oncogene family', other_designations='', map_location='1 45.84 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90958133, end_position_on_the_genomic_accession=90969674, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19330, CurrentID=0, Status='live', Symbol='Rab18', Aliases='AA959686', description='RAB18, member RAS oncogene family', other_designations='GTP-binding protein Rab18', map_location='18 4.53 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=6765014, end_position_on_the_genomic_accession=6791606, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19331, CurrentID=0, Status='live', Symbol='Rab19', Aliases='', description='RAB19, member RAS oncogene family', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39381428, end_position_on_the_genomic_accession=39390666, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19332, CurrentID=0, Status='live', Symbol='Rab20', Aliases='AA536966, D8Ertd350e', description='RAB20, member RAS oncogene family', other_designations='', map_location='8 5.73 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11453513, end_position_on_the_genomic_accession=11478499, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19334, CurrentID=0, Status='live', Symbol='Rab22a', Aliases='3732413A17Rik, AI662177, AW319644, AW550514, E130120E14Rik, Rab22', description='RAB22A, member RAS oncogene family', other_designations='RAB22, member RAS oncogene family|rab-14|rab-22', map_location='2 97.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173659845, end_position_on_the_genomic_accession=173702182, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19335, CurrentID=0, Status='live', Symbol='Rab23', Aliases='AW545388, opb, opb2', description='RAB23, member RAS oncogene family', other_designations='open brain 2|protein open brain|rab-15', map_location='1 12.8 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=33719896, end_position_on_the_genomic_accession=33742564, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19336, CurrentID=0, Status='live', Symbol='Rab24', Aliases='6530406O07Rik', description='RAB24, member RAS oncogene family', other_designations='rab-16', map_location='13 29.8 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55319223, end_position_on_the_genomic_accession=55321980, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19337, CurrentID=0, Status='live', Symbol='Rab33a', Aliases='Rabs10', description='RAB33A, member RAS oncogene family', other_designations='RAB33A, member of RAS oncogene family|small GTP-binding protein S10', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48519285, end_position_on_the_genomic_accession=48530240, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19338, CurrentID=0, Status='live', Symbol='Rab33b', Aliases='', description='RAB33B, member RAS oncogene family', other_designations='RAB33B, member of RAS oncogene family', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51483966, end_position_on_the_genomic_accession=51496228, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19339, CurrentID=0, Status='live', Symbol='Rab3a', Aliases='', description='RAB3A, member RAS oncogene family', other_designations='', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70754666, end_position_on_the_genomic_accession=70758686, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19340, CurrentID=0, Status='live', Symbol='Rab3d', Aliases='C130057E11Rik', description='RAB3D, member RAS oncogene family', other_designations='', map_location='9 A2-A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21907491, end_position_on_the_genomic_accession=21918197, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19341, CurrentID=0, Status='live', Symbol='Rab4a', Aliases='AI848268, Rab4', description='RAB4A, member RAS oncogene family', other_designations='', map_location='8 72.19 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123805847, end_position_on_the_genomic_accession=123835291, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19342, CurrentID=0, Status='live', Symbol='Rab4b', Aliases='1500031G17Rik', description='RAB4B, member RAS oncogene family', other_designations='', map_location='7 15.84 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27168424, end_position_on_the_genomic_accession=27178901, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19344, CurrentID=0, Status='live', Symbol='Rab5b', Aliases='C030027M18Rik', description='RAB5B, member RAS oncogene family', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128677172, end_position_on_the_genomic_accession=128696306, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19345, CurrentID=0, Status='live', Symbol='Rab5c', Aliases='AI326010, Rabl', description='RAB5C, member RAS oncogene family', other_designations='', map_location='11 63.54 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100715009, end_position_on_the_genomic_accession=100738210, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19346, CurrentID=0, Status='live', Symbol='Rab6a', Aliases='2610028L11Rik, AA419671, Rab6', description='RAB6A, member RAS oncogene family', other_designations='RAB6, member RAS oncogene family|rab-6', map_location='7 C4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100607586, end_position_on_the_genomic_accession=100641268, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19347, CurrentID=0, Status='live', Symbol='Dennd5a', Aliases='1500012B19Rik, AJ245569, AW496491, ORF37, Rab6ip1, mKIAA1091', description='DENN/MADD domain containing 5A', other_designations='Rab6 interacting protein 1|rab6-interacting protein 1', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109893780, end_position_on_the_genomic_accession=109960458, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  19348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19348, CurrentID=0, Status='live', Symbol='Kif20a', Aliases='AA415432, Rab6kifl', description='kinesin family member 20A', other_designations='Rabkinesin-6|kinesin-like protein 174|rab6-interacting kinesin-like protein', map_location='18 18.69 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34624624, end_position_on_the_genomic_accession=34633277, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  19349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19349, CurrentID=0, Status='live', Symbol='Rab7', Aliases='Rab7a', description='RAB7, member RAS oncogene family', other_designations='', map_location='6 39.13 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87999106, end_position_on_the_genomic_accession=88045270, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19352, CurrentID=0, Status='live', Symbol='Rabggtb', Aliases='', description='Rab geranylgeranyl transferase, b subunit', other_designations='GGTase-II-beta|RAB geranylgeranyltransferase, beta subunit|geranylgeranyl transferase type II subunit beta|rab GG transferase beta|rab GGTase beta|rab geranyl-geranyltransferase subunit beta|type II protein geranyl-geranyltransferase subunit beta', map_location='3 78.76 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=153907289, end_position_on_the_genomic_accession=153912966, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19353, CurrentID=0, Status='live', Symbol='Rac1', Aliases='AL023026, D5Ertd559e', description='RAS-related C3 botulinum substrate 1', other_designations='p21-Rac1', map_location='5 82.22 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143505481, end_position_on_the_genomic_accession=143528116, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19354, CurrentID=0, Status='live', Symbol='Rac2', Aliases='AI323801, AI452260', description='RAS-related C3 botulinum substrate 2', other_designations='EN-7 protein|p21-Rac2|protein EN-7', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78559169, end_position_on_the_genomic_accession=78572783, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19355, CurrentID=0, Status='live', Symbol='Rad1', Aliases='mRAD1', description='RAD1 checkpoint DNA exonuclease', other_designations='DNA repair exonuclease rad1 homolog|RAD1 homolog|rad1-like DNA damage checkpoint protein', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10486018, end_position_on_the_genomic_accession=10499063, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19356, CurrentID=0, Status='live', Symbol='Rad17', Aliases='MmRad24', description='RAD17 checkpoint clamp loader component', other_designations='RAD17 homolog', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100617164, end_position_on_the_genomic_accession=100651061, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19357, CurrentID=0, Status='live', Symbol='Rad21', Aliases='SCC1, mKIAA0078', description='RAD21 cohesin complex component', other_designations='PW29|RAD21 homolog|SCC1 homolog|pokeweed agglutinin-binding protein 29', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=51962604, end_position_on_the_genomic_accession=51993538, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  19358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19358, CurrentID=0, Status='live', Symbol='Rad23a', Aliases='2310040P19Rik, AL024030, HR23A, mHR23A', description='RAD23a homolog (S. cerevisiae)', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84834652, end_position_on_the_genomic_accession=84840750, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19359, CurrentID=0, Status='live', Symbol='Rad23b', Aliases='0610007D13Rik, AV001138, HR23B, mHR23B, p58', description='RAD23b homolog (S. cerevisiae)', other_designations='XP-C repair-complementing complex 58 kDa protein', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=55350042, end_position_on_the_genomic_accession=55392237, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19360, CurrentID=0, Status='live', Symbol='Rad50', Aliases='Mrelll, Rad50', description='RAD50 homolog (S. cerevisiae)', other_designations='mRad50', map_location='11 A5-B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53649519, end_position_on_the_genomic_accession=53707319, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  19361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19361, CurrentID=0, Status='live', Symbol='Rad51', Aliases='AV304093a, Reca, Rad51', description='RAD51 homolog', other_designations='RAD51 homolog A', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119112817, end_position_on_the_genomic_accession=119136070, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19362, CurrentID=0, Status='live', Symbol='Rad51ap1', Aliases='2510006L10Rik, RAB22', description='RAD51 associated protein 1', other_designations='RAD51-interacting protein|rad51ap1', map_location='6 61.92 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126923415, end_position_on_the_genomic_accession=126939621, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19363, CurrentID=0, Status='live', Symbol='Rad51b', Aliases='AI553500, R51H2, Rad51l1, mREC2', description='RAD51 homolog B', other_designations='RAD51-like 1|RAD51-like protein 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79297282, end_position_on_the_genomic_accession=79852895, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  19364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19364, CurrentID=0, Status='live', Symbol='Rad51d', Aliases='R51H3, Rad51l3, TRAD, Trad-d5', description='RAD51 homolog D', other_designations='RAD51-like 3|TRAD/RAD51L3', map_location='11 50.3 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82871960, end_position_on_the_genomic_accession=82890751, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19365, CurrentID=0, Status='live', Symbol='Rad52', Aliases='Rad52yh', description='RAD52 homolog (S. cerevisiae)', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119902207, end_position_on_the_genomic_accession=119922823, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19366, CurrentID=0, Status='live', Symbol='Rad54l', Aliases='RAD54', description='RAD54 like (S. cerevisiae)', other_designations='RAD54 homolog|mHR54|mRAD54', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116096727, end_position_on_the_genomic_accession=116123690, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  19367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19367, CurrentID=0, Status='live', Symbol='Rad9a', Aliases='Rad9', description='RAD9 homolog A', other_designations='DNA repair exonuclease rad9 homolog A|Rad9-like protein|mRAD9', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4195196, end_position_on_the_genomic_accession=4201639, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19368, CurrentID=0, Status='live', Symbol='Raet1a', Aliases='RAE-1alpha, Rae1alpha, Raet1', description='retinoic acid early transcript 1, alpha', other_designations='RAE-1-alpha', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  68764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68764, CurrentID=0, Status='live', Symbol='Cdhr3', Aliases='1110049B09Rik, BB172694, Cdh28', description='cadherin-related family member 3', other_designations='cadherin-like protein 28', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=33033796, end_position_on_the_genomic_accession=33093037, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  19370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19370, CurrentID=0, Status='live', Symbol='Raet1c', Aliases='RAE-1gamma', description='retinoic acid early transcript gamma', other_designations='RAE-1-gamma', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.4', start_position_on_the_genomic_accession=21863318, end_position_on_the_genomic_accession=21873333, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19373, CurrentID=0, Status='live', Symbol='Rag1', Aliases='Rag-1', description='recombination activating gene 1', other_designations='recombinase activating gene 1|recombination activating protein 1|recombination-activating gene 1', map_location='2 53.88 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=101638252, end_position_on_the_genomic_accession=101649532, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19374, CurrentID=0, Status='live', Symbol='Rag2', Aliases='Rag-2', description='recombination activating gene 2', other_designations='recombinase activating gene 2|recombination activating protein 2', map_location='2 53.87 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=101624748, end_position_on_the_genomic_accession=101632528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19376, CurrentID=0, Status='live', Symbol='Rab34', Aliases='AI326479, Narr, Rah, Rah1', description='RAB34, member RAS oncogene family', other_designations='RAB34, member of RAS oncogene family|rab-39|ras-related homolog|ras-related protein Rab-39|ras-related protein Rah', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78187879, end_position_on_the_genomic_accession=78192193, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19377, CurrentID=0, Status='live', Symbol='Rai1', Aliases='Gt1', description='retinoic acid induced 1', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60104907, end_position_on_the_genomic_accession=60199197, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19378, CurrentID=0, Status='live', Symbol='Aldh1a2', Aliases='AV116159, Aldh1a7, Raldh1, Raldh2', description='aldehyde dehydrogenase family 1, subfamily A2', other_designations='RALDH 2|RALDH(II)|alcohol dehydrogenase family 1, subfamily A2|alcohol dehydrogenase family 1, subfamily A7|aldehyde dehydrogenase family 1 member A2|retinaldehyde dehydrogenase 2|retinaldehyde-specific dehydrogenase type 2', map_location='9 39.85 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=71215789, end_position_on_the_genomic_accession=71296243, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  19383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19383, CurrentID=0, Status='live', Symbol='Raly', Aliases='AI663842, Merc', description='hnRNP-associated with lethal yellow', other_designations='hnRNP associated with lethal yellow protein|maternally-expressed hnRNP C-related protein', map_location='2 76.83 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154791097, end_position_on_the_genomic_accession=154871010, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19384, CurrentID=0, Status='live', Symbol='Ran', Aliases='', description='RAN, member RAS oncogene family', other_designations='GTPase Ran|ras-like protein TC4|ras-related nuclear protein', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129020156, end_position_on_the_genomic_accession=129024321, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19385, CurrentID=0, Status='live', Symbol='Ranbp1', Aliases='Htf9a', description='RAN binding protein 1', other_designations='Htf9-a/RanBP1|hpaII tiny fragments locus 9a protein|ran-binding protein 1', map_location='16 11.3 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18239979, end_position_on_the_genomic_accession=18248694, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19386, CurrentID=0, Status='live', Symbol='Ranbp2', Aliases='A430087B05Rik, AI256741, NUP358', description='RAN binding protein 2', other_designations='ran-binding protein 2', map_location='10 29.34 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=58446852, end_position_on_the_genomic_accession=58494155, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  19387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19387, CurrentID=0, Status='live', Symbol='Rangap1', Aliases='C79654, Fug1, mKIAA1835', description='RAN GTPase activating protein 1', other_designations='', map_location='15 38.21 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81704248, end_position_on_the_genomic_accession=81729919, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19395, CurrentID=0, Status='live', Symbol='Rasgrp2', Aliases='CDC25L, Caldaggef1', description='RAS, guanyl releasing protein 2', other_designations='CalDAG-GEFI|F25B3.3 kinase like protein|F25B3.3 kinase-like protein|RAP 1A protein-specific guanine nucleotide exchange factor 1|RAS, guanyl releasing protein 2; RAP 1A protein-specific guanine nucleotide exchange factor 1; CalDAG-GEFI|calcium and DAG-regulated guanine nucleotide exchange factor I', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6399284, end_position_on_the_genomic_accession=6415216, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  19400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19400, CurrentID=0, Status='live', Symbol='Rapsn', Aliases='43kDa, Nraps, Raps, rapsyn', description='receptor-associated protein of the synapse', other_designations='43 kDa postsynaptic protein|43kDa acetylcholine receptor-associated protein|acetylcholine receptor-associated 43 kDa protein|receptor-associated protein of the synapse, 43 kDa', map_location='2 50.44 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91035466, end_position_on_the_genomic_accession=91045729, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19401, CurrentID=0, Status='live', Symbol='Rara', Aliases='Nr1b1, RAR, RARalpha1', description='retinoic acid receptor, alpha', other_designations='RAR alpha 1|RAR-alpha|nuclear receptor subfamily 1 group B member 1', map_location='11 D-E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98927809, end_position_on_the_genomic_accession=98974942, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  68770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68770, CurrentID=0, Status='live', Symbol='Phtf2', Aliases='1110054G21Rik, 9530062N20Rik, AI447096, C77923', description='putative homeodomain transcription factor 2', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=20758664, end_position_on_the_genomic_accession=20882124, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19411, CurrentID=0, Status='live', Symbol='Rarg', Aliases='Nr1b3, RAR-gamma, RARD, RARgamma2', description='retinoic acid receptor, gamma', other_designations='RAR gamma 2|nuclear receptor subfamily 1 group B member 3|retinoic acid receptor gamma 4', map_location='15 E-F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102234938, end_position_on_the_genomic_accession=102257522, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  19412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19412, CurrentID=0, Status='live', Symbol='ras', Aliases='asr', description='resistance to audiogenic seizures', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19414, CurrentID=0, Status='live', Symbol='Rasa3', Aliases='AI326412, C86362, E130011G04, GAPIII, scat', description='RAS p21 protein activator 3', other_designations='GAP1(IP4BP)|GAPIII activator 3|R-Ras gap|R-ras RTPase activating protein-3|Ras GTPase-activating protein III|ins P4-binding protein', map_location='8 6.29 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13567218, end_position_on_the_genomic_accession=13677587, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  19415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19415, CurrentID=0, Status='live', Symbol='Rasal1', Aliases='MRASAL', description='RAS protein activator like 1 (GAP1 like)', other_designations='ras GTPase-activating-like protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120648040, end_position_on_the_genomic_accession=120679610, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  19416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19416, CurrentID=0, Status='live', Symbol='Rasd1', Aliases='Dexras1', description='RAS, dexamethasone-induced 1', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59963181, end_position_on_the_genomic_accession=59965080, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19417, CurrentID=0, Status='live', Symbol='Rasgrf1', Aliases='AI844718, CDC25, CDC25Mm, Gnrp, Grf1, Grfbeta, P190-A, Ras-GRF1, p190, p190RhoGEF', description='RAS protein-specific guanine nucleotide-releasing factor 1', other_designations='Ras guanine release factor beta|ras-specific nucleotide exchange factor CDC25', map_location='9 47.31 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89909472, end_position_on_the_genomic_accession=90026979, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  19418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19418, CurrentID=0, Status='live', Symbol='Rasgrf2', Aliases='6330417G04Rik, AW048350, Grf2, Ras-GRF2', description='RAS protein-specific guanine nucleotide-releasing factor 2', other_designations='ras guanine nucleotide exchange factor 2', map_location='13 C3-D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92022789, end_position_on_the_genomic_accession=92131494, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  19419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19419, CurrentID=0, Status='live', Symbol='Rasgrp1', Aliases='Rasgrp', description='RAS guanyl releasing protein 1', other_designations='calDAG-GEFII|calcium and DAG-regulated guanine nucleotide exchange factor II', map_location='2 59.19 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=117279993, end_position_on_the_genomic_accession=117342877, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  19428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19428, CurrentID=0, Status='live', Symbol='Rasl2-9', Aliases='M2, Ran, Ran/M2-ps, Rasl2-9', description='RAS-like, family 2, locus 9', other_designations='RAS-like, family 2, locus 9, pseudogene', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5124942, end_position_on_the_genomic_accession=5125950, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19434, CurrentID=0, Status='live', Symbol='Rax', Aliases='7530406A22Rik, E130303K03Rik, Rx, ey-1, ey1', description='retina and anterior neural fold homeobox', other_designations='Rx/rax homeoprotein|eyeless 1|retina and anterior neural fold homeobox protein', map_location='18 39.03 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65934639, end_position_on_the_genomic_accession=65939089, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  330921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330921, CurrentID=0, Status='live', Symbol='Pate2', Aliases='Gm846, Pate-M, mANLP1', description='prostate and testis expressed 2', other_designations='PATE-like protein M', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35574309, end_position_on_the_genomic_accession=35698689, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19448, CurrentID=0, Status='live', Symbol='Rb(1.3)1Bnr', Aliases='Rb1Bnr', description='Robertsonian translocation, Chr 1 and 3, Universitat Bonn/Rhein 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19477, CurrentID=0, Status='live', Symbol='Rb(11.16)2H', Aliases='Rb2H', description='Robertsonian translocation, Chr 11 and 16, Harwell 2', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19478, CurrentID=0, Status='live', Symbol='Rb(11.16)2H', Aliases='Rb2H', description='Robertsonian translocation, Chr 11 and 16, Harwell 2', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19499, CurrentID=0, Status='live', Symbol='Rb(16.17)7Bnr', Aliases='Rb7Bnr', description='Robertsonian translocation, Chr 16 and 17, Universitat Bonn/Rhein 7', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19500, CurrentID=0, Status='live', Symbol='Rb(16.17)7Bnr', Aliases='Rb7Bnr', description='Robertsonian translocation, Chr 16 and 17, Universitat Bonn/Rhein 7', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19501, CurrentID=0, Status='live', Symbol='Rb(16.17)8Lub', Aliases='Rb8Lub', description='Robertsonian translocation, Chr 16 and 17, Lubeck 8', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19502, CurrentID=0, Status='live', Symbol='Rb(16.17)8Lub', Aliases='Rb8Lub', description='Robertsonian translocation, Chr 16 and 17, Lubeck 8', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  674895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=674895, CurrentID=0, Status='live', Symbol='Nek10', Aliases='Gm282', description='NIMA (never in mitosis gene a)- related kinase 10', other_designations='nek10|never in mitosis A-related kinase 10|nimA-related protein kinase 10', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=14803580, end_position_on_the_genomic_accession=15012059, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  330938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330938, CurrentID=0, Status='live', Symbol='Dixdc1', Aliases='4930563F16Rik, BC048182, Ccd1, Ccd1Aalpha1L, Ccd1Abeta1L, Ccd1BalphaL, Ccd1BbetaL, Ccd1CL, mKIAA1735', description='DIX domain containing 1', other_designations='DIX domain-containing protein 1|coiled-coil protein DIX1|coiled-coil-DIX1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50662752, end_position_on_the_genomic_accession=50737737, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  19557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19557, CurrentID=0, Status='live', Symbol='Rb(4.6)2Bnr', Aliases='Rb2Bnr', description='Robertsonian translocation, Chr 4 and 6, Universitat Bonn/Rhein 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  330941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330941, CurrentID=0, Status='live', Symbol='AI593442', Aliases='E130102D23', description='expressed sequence AI593442', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=52673032, end_position_on_the_genomic_accession=52680064, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19579, CurrentID=0, Status='live', Symbol='Rb(5.19)1Wh', Aliases='Rb1Wh, T1Wh', description='Robertsonian translocation, Chr 5 and 19, White 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19580, CurrentID=0, Status='live', Symbol='Rb(5.19)1Wh', Aliases='Rb1Wh, T1Wh', description='Robertsonian translocation, Chr 5 and 19, White 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19595, CurrentID=0, Status='live', Symbol='Rb(6.16)24Lub', Aliases='Rb24Lub', description='Robertsonian translocation, Chr 6 and 16, Lubeck 24', other_designations='Rb(6.16)24Lubx', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19596, CurrentID=0, Status='live', Symbol='Rb(6.16)24Lub', Aliases='Rb24Lub', description='Robertsonian translocation, Chr 6 and 16, Lubeck 24', other_designations='Rb(6.16)24Lubx', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19608, CurrentID=0, Status='live', Symbol='Rb(7.18)9Lub', Aliases='Rb9Lub', description='Robertsonian translocation, Chr 7 and 18, Lubeck 9', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19613, CurrentID=0, Status='live', Symbol='Rb(8.12)5Bnr', Aliases='Rb5Bnr', description='Robertsonian translocation, Chr 8 and 12, Universitat Bonn/Rhein 5', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19614, CurrentID=0, Status='live', Symbol='Rb(8.12)5Bnr', Aliases='Rb5Bnr', description='Robertsonian translocation, Chr 8 and 12, Universitat Bonn/Rhein 5', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19645, CurrentID=0, Status='live', Symbol='Rb1', Aliases='Rb, Rb-1, pRb', description='retinoblastoma 1', other_designations='pp105', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73194453, end_position_on_the_genomic_accession=73325887, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  19646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19646, CurrentID=0, Status='live', Symbol='Rbbp4', Aliases='RBAP48, mRbAp48', description='retinoblastoma binding protein 4', other_designations='CAF-1 p48 subunit|CAF-1 subunit C|CAF-I 48 kDa subunit|CAF-I p48|RBBP-4|chromatin assembly factor 1 subunit C|chromatin assembly factor I p48 subunit|nucleosome-remodeling factor subunit RBAP48|retinoblastoma-binding protein 4|retinoblastoma-binding protein p48', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129307100, end_position_on_the_genomic_accession=129335370, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  19647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19647, CurrentID=0, Status='live', Symbol='Rbbp6', Aliases='4933422O15Rik, AI316869, BB233631, C030034J04Rik, C77662, P2P-R, PACT', description='retinoblastoma binding protein 6', other_designations='p53-associated cellular protein of testis|proliferation potential-related protein|protein P2P-R|retinoblastoma-binding protein 6', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122970546, end_position_on_the_genomic_accession=123002572, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  19649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19649, CurrentID=0, Status='live', Symbol='Robo3', Aliases='Rbig1, Rig-1, Rig1a, Robo3b, Robo3', description='roundabout homolog 3 (Drosophila)', other_designations='retinoblastoma inhibiting gene 1|retinoblastoma-inhibiting gene 1 protein', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37416045, end_position_on_the_genomic_accession=37433408, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  19650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19650, CurrentID=0, Status='live', Symbol='Rbl1', Aliases='AW547426, PRB1, p107', description='retinoblastoma-like 1 (p107)', other_designations='107 kDa retinoblastoma-associated protein', map_location='2 78.05 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157145893, end_position_on_the_genomic_accession=157204534, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  19651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19651, CurrentID=0, Status='live', Symbol='Rbl2', Aliases='PRB2, RBR-2, Rb2, p130', description='retinoblastoma-like 2', other_designations='130 kDa retinoblastoma-associated protein|retinoblastoma-related protein 2', map_location='8 44.25 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=91070057, end_position_on_the_genomic_accession=91123844, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  19652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19652, CurrentID=0, Status='live', Symbol='Rbm3', Aliases='2600016C11Rik', description='RNA binding motif protein 3', other_designations='RNA-binding motif protein 3|putative RNA-binding protein 3', map_location='X 3.68 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8138975, end_position_on_the_genomic_accession=8145880, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19653, CurrentID=0, Status='live', Symbol='Rbm4', Aliases='4921506I22Rik, Lark1, Mlarka, lark, Rbm4', description='RNA binding motif protein 4', other_designations='RNA-binding motif protein 4a|lark homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4784293, end_position_on_the_genomic_accession=4794009, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19654, CurrentID=0, Status='live', Symbol='Rbm6', Aliases='4930506F14Rik, Def-3, NY-LU-12, g16, mKIAA4015', description='RNA binding motif protein 6', other_designations='', map_location='9 F1-F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107773559, end_position_on_the_genomic_accession=107872819, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  19655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19655, CurrentID=0, Status='live', Symbol='Rbmx', Aliases='hnRNP G', description='RNA binding motif protein, X chromosome', other_designations='heterogeneous nuclear ribonucleoprotein G', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=57383348, end_position_on_the_genomic_accession=57393042, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19656, CurrentID=0, Status='live', Symbol='Rbmxl1', Aliases='Hnrpg, Rbmx, Rbmxrt', description='RNA binding motif protein, X-linked like-1', other_designations='RNA binding motif protein, X chromosome retrogene|RNA binding motif protein, X linked-like-1|RNA-binding motif protein, X chromosome|heterogeneous nuclear ribonucleoprotein G (hnRNP G)|heterogeneous nuclear ribonucleoprotein G-like 1|hnRNP G', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=78505269, end_position_on_the_genomic_accession=78508928, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19657, CurrentID=0, Status='live', Symbol='Rbmy', Aliases='RBM, Rbm1, Rbm1-rs11a1, Rbmy1a1-rs1, Rbmy1b, Rbmy', description='RNA binding motif protein, Y chromosome', other_designations='RNA binding motif protein, Y chromosome, family 1, member A1, related sequence 1|RNA-binding motif protein 1|RNA-binding motif protein, Y chromosome, family 1 member B|y chromosome RNA recognition motif 1', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=2830680, end_position_on_the_genomic_accession=2841858, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19659, CurrentID=0, Status='live', Symbol='Rbp1', Aliases='CRBPI, Crbp, Rbp-1', description='retinol binding protein 1, cellular', other_designations='CRBP-I|cellular retinol-binding protein I|mCRBPI', map_location='9 51.36 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98422961, end_position_on_the_genomic_accession=98446551, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19660, CurrentID=0, Status='live', Symbol='Rbp2', Aliases='Crbp-2, CrbpII, Rbp-2', description='retinol binding protein 2, cellular', other_designations='CRBP-II|cellular retinol-binding protein II', map_location='9 51.37 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98490537, end_position_on_the_genomic_accession=98509771, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19661, CurrentID=0, Status='live', Symbol='Rbp3', Aliases='Irbp, Rbp-3', description='retinol binding protein 3, interstitial', other_designations='interphotoreceptor retinoid-binding protein|interstitial retinol-binding protein', map_location='14 20.8 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33954022, end_position_on_the_genomic_accession=33964216, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19662, CurrentID=0, Status='live', Symbol='Rbp4', Aliases='Rbp-4', description='retinol binding protein 4, plasma', other_designations='PRBP|RBP|plasma retinol-binding protein|retinol binding protein 4, cellular', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38116620, end_position_on_the_genomic_accession=38125321, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19663, CurrentID=0, Status='live', Symbol='Rbpms', Aliases='2010300K22Rik, 2700019M19Rik, AU017537, RBP-MS, hermes', description='RNA binding protein gene with multiple splicing', other_designations='HEart, RRM Expressed Sequence|RNA-binding protein gene with multiple splicing|RNA-binding protein with multiple splicing (RBP-MS)', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33782644, end_position_on_the_genomic_accession=33929895, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  19664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19664, CurrentID=0, Status='live', Symbol='Rbpj', Aliases='AI843960, CBF1, Igkjrb, Igkrsbp, RBP-J, RBP-J kappa, RBP-Jkappa, RBPjk, Rbpsuh', description='recombination signal binding protein for immunoglobulin kappa J region', other_designations='J-kappa recombination signal binding protein', map_location='5 C1-C3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=53466117, end_position_on_the_genomic_accession=53657445, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  19668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19668, CurrentID=0, Status='live', Symbol='Rbpjl', Aliases='RBP-L, Rbpsuhl', description='recombination signal binding protein for immunoglobulin kappa J region-like', other_designations='RBP-J kappa-like|transcription factor RBP-L', map_location='2 85.16 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164403056, end_position_on_the_genomic_accession=164415448, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19671, CurrentID=0, Status='live', Symbol='Rce1', Aliases='D19Ertd283e, D19Ertd98e', description='RCE1 homolog, prenyl protein peptidase (S. cerevisiae)', other_designations='FACE-2|Ras and a-factor-converting enzyme 1 homolog|farnesylated proteins-converting enzyme 2|prenyl protein-specific endoprotease 2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4622551, end_position_on_the_genomic_accession=4625617, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19672, CurrentID=0, Status='live', Symbol='Rcn1', Aliases='Rcn', description='reticulocalbin 1', other_designations='', map_location='2 55.18 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=105385948, end_position_on_the_genomic_accession=105399319, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19674, CurrentID=0, Status='live', Symbol='Rcvrn', Aliases='CAR, S-modulin', description='recoverin', other_designations='23 kDa photoreceptor cell-specific protein|cancer associated retinopathy protein|cancer-associated retinopathy protein|guanylate cyclase activator', map_location='11 41.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67695326, end_position_on_the_genomic_accession=67703355, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  330959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330959, CurrentID=0, Status='live', Symbol='Snapc5', Aliases='2010103A03Rik', description='small nuclear RNA activating complex, polypeptide 5', other_designations='SNAPc 19 kDa subunit|SNAPc subunit 5|small nuclear RNA activating complex, polypeptide 5, 19kDa|small nuclear RNA-activating complex polypeptide 5|snRNA-activating protein complex 19 kDa subunit', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64179297, end_position_on_the_genomic_accession=64182688, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19679, CurrentID=0, Status='live', Symbol='Pitpnm2', Aliases='NIR3, RDGBA2, Rdgb2, mKIAA1457', description='phosphatidylinositol transfer protein, membrane-associated 2', other_designations='drosophila retinal degeneration B homolog 2|pyk2 N-terminal domain-interacting receptor 3|retinal degeneration B2 homolog', map_location='5 63.39 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124118689, end_position_on_the_genomic_accession=124249760, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  19682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19682, CurrentID=0, Status='live', Symbol='Rdh5', Aliases='9-cis, AI987873, RDH4, cRDH', description='retinol dehydrogenase 5', other_designations='11-cis RDH|11-cis RoDH|9-cis retinol dehydrogenase|9-cis-retinol dehydrogenase|cis-retinol dehydrogenase|rdh5|retinol dehydrogenase type 5', map_location='10 77.19 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128913588, end_position_on_the_genomic_accession=128919297, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19683, CurrentID=0, Status='live', Symbol='Rdh16', Aliases='CRAD, CRAD1, Rdh6', description='retinol dehydrogenase 16', other_designations='cis-retinol/androgen dehydrogenase 1|retinol dehydrogenase 6|retinol dehydrogenase type 6', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127801090, end_position_on_the_genomic_accession=127815839, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19684, CurrentID=0, Status='live', Symbol='Rdx', Aliases='AA516625', description='radixin', other_designations='ESP10', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=52047150, end_position_on_the_genomic_accession=52088738, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  19687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19687, CurrentID=0, Status='live', Symbol='Rfc1', Aliases='140kDa, Alp145, RFC140, Recc1', description='replication factor C (activator 1) 1', other_designations='A1 140 kDa subunit|A1-P145|ISRE-binding protein|RF-C 140 kDa subunit|activator 1 140 kDa subunit|activator 1 large subunit|activator 1 subunit 1|differentiation-specific element-binding protein|replication factor C 1|replication factor C 140 kDa subunit|replication factor C large subunit|replication factor C, 140 kDa', map_location='5 C', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65261852, end_position_on_the_genomic_accession=65336476, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  19691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19691, CurrentID=0, Status='live', Symbol='Recql', Aliases='RecQ1', description='RecQ protein-like', other_designations='DNA-dependent ATPase Q1|recQ protein-like 1', map_location='6 73.91 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142350342, end_position_on_the_genomic_accession=142387116, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19692, CurrentID=0, Status='live', Symbol='Reg1', Aliases='', description='regenerating islet-derived 1', other_designations='PSP|PTP|REG 1|islet of Langerhans regenerating protein 1|pancreatic stone protein 1|pancreatic thread protein 1|regenerating islet-derived, mouse homolog 1|regenerating protein 1', map_location='6 C-D', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=78420227, end_position_on_the_genomic_accession=78428668, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  330962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330962, CurrentID=0, Status='live', Symbol='Slc51b', Aliases='Ostb', description='solute carrier family 51, beta subunit', other_designations='OST-beta|Ost beta|organic solute transporter beta|solute carrier family 51 subunit beta', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65412753, end_position_on_the_genomic_accession=65422773, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19694, CurrentID=0, Status='live', Symbol='Reg3a', Aliases='AV063448', description='regenerating islet-derived 3 alpha', other_designations='REG-3-alpha|RegIII (alpha)|generating islet-derived, mouse homolog 3 alpha|islet of Langerhans regenerating protein 3|lithostathine 3|pancreatitis-associated protein 2|reg III-alpha|regenerating islet-derived protein III-alpha', map_location='6 34.76 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=78380709, end_position_on_the_genomic_accession=78383839, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19695, CurrentID=0, Status='live', Symbol='Reg3g', Aliases='AI449515', description='regenerating islet-derived 3 gamma', other_designations='REG-3-gamma|RegIII (gamma)|generating islet-derived, mouse homolog 3 gamma|pancreatitis-associated protein 3|reg III-gamma|regenerating islet-derived protein III-gamma', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=78466268, end_position_on_the_genomic_accession=78468874, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19696, CurrentID=0, Status='live', Symbol='Rel', Aliases='c-Rel', description='reticuloendotheliosis oncogene', other_designations='C-Rel proto-oncogene protein', map_location='11 14.36 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23741729, end_position_on_the_genomic_accession=23770970, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19697, CurrentID=0, Status='live', Symbol='Rela', Aliases='p65', description='v-rel reticuloendotheliosis viral oncogene homolog A (avian)', other_designations='avian reticuloendotheliosis viral (v-rel) oncogene homolog A|nuclear factor NF-kappa-B p65 subunit|nuclear factor of kappa light polypeptide gene enhancer in B-cells 3|p65 NF kappaB|p65 NF-kappa B|p65 NFkB', map_location='19 B1-3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5637440, end_position_on_the_genomic_accession=5648130, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19698, CurrentID=0, Status='live', Symbol='Relb', Aliases='shep', description='avian reticuloendotheliosis viral (v-rel) oncogene related B', other_designations='', map_location='7 9.93 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19606217, end_position_on_the_genomic_accession=19629438, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  19699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19699, CurrentID=0, Status='live', Symbol='Reln', Aliases='reeler, rl', description='reelin', other_designations='extracellular matrix serine protease', map_location='5 A3-B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21884454, end_position_on_the_genomic_accession=22344705, orientation='minus', exon_count=66, OMIM='', no_hdr0=''),
  19700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19700, CurrentID=0, Status='live', Symbol='Rem1', Aliases='E030011C07Rik, Ras, Rem', description='rad and gem related GTP binding protein 1', other_designations='RAS-like GTP binding protein Rem|rad and Gem-like GTP-binding protein 1', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152627008, end_position_on_the_genomic_accession=152635191, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19701, CurrentID=0, Status='live', Symbol='Ren1', Aliases='D19352, Ren, Ren-1, Ren-Ac, Ren1d, Rn-1, Rnr, Ren1', description='renin 1 structural', other_designations='angiotensinogenase|aspartyl-protease|kidney renin|renin b', map_location='1 57.91 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133350674, end_position_on_the_genomic_accession=133360320, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19702, CurrentID=0, Status='live', Symbol='Ren2', Aliases='Ren, Ren-2, Ren-B, Rn-2, Rnr', description='renin 2 tandem duplication of Ren1', other_designations='angiotensinogenase|submandibular gland renin', map_location='1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19703, CurrentID=0, Status='live', Symbol='Renbp', Aliases='Age, Rnbp', description='renin binding protein', other_designations='GlcNAc 2-epimerase|N-acetyl-D-glucosamine 2-epimerase|renin-binding protein', map_location='X 37.49 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73922121, end_position_on_the_genomic_accession=73930850, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  19704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19704, CurrentID=0, Status='live', Symbol='Upf1', Aliases='B430202H16Rik, NORF1, PNORF-1, Rent1, Upflp', description='UPF1 regulator of nonsense transcripts homolog (yeast)', other_designations='ATP-dependent helicase RENT1|Regulator of nonsense transcripts 1 (Nonsense mRNA reducing factor 1) (NORF1) (Up-frameshift suppressor 1 homolog)|mUpf1|nonsense mRNA reducing factor 1|up-frameshift suppressor 1 homolog', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70331522, end_position_on_the_genomic_accession=70353999, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  19707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19707, CurrentID=0, Status='live', Symbol='Reps1', Aliases='BB161292', description='RalBP1 associated Eps domain containing protein', other_designations='RalBP1 associated Eps domain containing 1|ralBP1-interacting protein 1', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18055561, end_position_on_the_genomic_accession=18125155, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  19708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19708, CurrentID=0, Status='live', Symbol='Dpf2', Aliases='2210010M07Rik, BAF45D, Req, ubi-d4', description='D4, zinc and double PHD fingers family 2', other_designations='BRG1-associated factor 45D|apoptosis response zinc finger protein|neuro-d4/requiem|protein requiem', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5896516, end_position_on_the_genomic_accession=5912871, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  19711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19711, CurrentID=0, Status='live', Symbol='Resp18', Aliases='AI851012', description='regulated endocrine-specific protein 18', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75272197, end_position_on_the_genomic_accession=75278455, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19712, CurrentID=0, Status='live', Symbol='Rest', Aliases='2610008J04Rik, AA407358, NRSF, REST4', description='RE1-silencing transcription factor', other_designations='neural-restrictive silencer factor', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77265494, end_position_on_the_genomic_accession=77286432, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19713, CurrentID=0, Status='live', Symbol='Ret', Aliases='PTC, RET51, RET9, c-Ret', description='ret proto-oncogene', other_designations='proto-oncogene c-Ret', map_location='6 E3-F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118151748, end_position_on_the_genomic_accession=118197744, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  19714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19714, CurrentID=0, Status='live', Symbol='Rev3l', Aliases='Rev, Rev3, Sez4', description='REV3-like, catalytic subunit of DNA polymerase zeta RAD54 like (S. cerevisiae)', other_designations='seizure-related protein 4', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39732099, end_position_on_the_genomic_accession=39875210, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  19716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19716, CurrentID=0, Status='live', Symbol='Bex1', Aliases='AA407839, AU018692, Bex4, Rex3', description='brain expressed gene 1', other_designations='REX-3|brain-expressed X-linked protein 1 homolog|reduced expression 3|reduced expression protein 3', map_location='X 57.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136213972, end_position_on_the_genomic_accession=136215513, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19718, CurrentID=0, Status='live', Symbol='Rfc2', Aliases='2610008M13Rik, 40kDa, AI326953, Recc2', description='replication factor C (activator 1) 2', other_designations='A1 40 kDa subunit|RF-C 40 kDa subunit|RFC40|activator 1 40 kDa subunit|activator 1 subunit 2|replication factor C 40 kDa subunit|replication factor C, 40 kDa', map_location='5 74.68 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134582690, end_position_on_the_genomic_accession=134598328, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19719, CurrentID=0, Status='live', Symbol='Rfng', Aliases='', description='RFNG O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase', other_designations='O-fucosylpeptide 3-beta-N-acetylglucosaminyltransferase|radical fringe gene homolog', map_location='11 84.53 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120780745, end_position_on_the_genomic_accession=120784217, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19720, CurrentID=0, Status='live', Symbol='Trim27', Aliases='AW538890, Gm19403, Rfp', description='tripartite motif-containing 27', other_designations='ret finger protein|tripartite motif protein 27|tripartite motif-containing protein 27', map_location='13 7.67 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21179440, end_position_on_the_genomic_accession=21194723, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19724, CurrentID=0, Status='live', Symbol='Rfx1', Aliases='AI047719, AI385641', description='regulatory factor X, 1 (influences HLA class II expression)', other_designations='regulatory factor X 1|transcription factor RFX1|vregulatory factor X, 1 (influences HLA class II expression)', map_location='8 40.28 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84066675, end_position_on_the_genomic_accession=84096992, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  19725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19725, CurrentID=0, Status='live', Symbol='Rfx2', Aliases='5430432H19Rik', description='regulatory factor X, 2 (influences HLA class II expression)', other_designations='regulatory factor X 2', map_location='17 29.5 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56775897, end_position_on_the_genomic_accession=56831054, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  19726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19726, CurrentID=0, Status='live', Symbol='Rfx3', Aliases='C230093O12Rik, MRFX3', description='regulatory factor X, 3 (influences HLA class II expression)', other_designations='regulatory factor X 3', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=27761718, end_position_on_the_genomic_accession=28011698, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  19727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19727, CurrentID=0, Status='live', Symbol='Rfxank', Aliases='Tvl1', description='regulatory factor X-associated ankyrin-containing protein', other_designations='RFX-B|ankyrin repeat-containing adapter protein Tvl-1|regulatory factor X subunit B', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70130806, end_position_on_the_genomic_accession=70139197, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19729, CurrentID=0, Status='live', Symbol='Slc50a1', Aliases='MmSWEET1, Rag1ap1, Rga', description='solute carrier family 50 (sugar transporter), member 1', other_designations='RAG1-activating protein 1|probable sugar transporter RAG1AP1|recombination activating gene 1 activating protein 1|recombination activating gene 1 gene activation|solute carrier family 50 member 1', map_location='3 39.04 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89268246, end_position_on_the_genomic_accession=89270570, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19730, CurrentID=0, Status='live', Symbol='Ralgds', Aliases='Gnds, Rgds, mKIAA1308', description='ral guanine nucleotide dissociation stimulator', other_designations='ral guanine nucleotide exchange factor|ralGEF', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28513089, end_position_on_the_genomic_accession=28553082, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  19731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19731, CurrentID=0, Status='live', Symbol='Rgl1', Aliases='Rgl, mKIAA0959', description='ral guanine nucleotide dissociation stimulator,-like 1', other_designations='ralGDS-like 1', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152516768, end_position_on_the_genomic_accession=152766351, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  19732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19732, CurrentID=0, Status='live', Symbol='Rgl2', Aliases='KE1.5, Rab2l, Rgt2, Rlf', description='ral guanine nucleotide dissociation stimulator-like 2', other_designations='RalGDS-like factor|ral guanine nucleotide dissociation stimulator,-like 2|ralGDS-like 2|ras-associated protein RAB2L', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33929435, end_position_on_the_genomic_accession=33937687, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  19733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19733, CurrentID=0, Status='live', Symbol='Rgn', Aliases='AI265316, GNL, SMP30', description='regucalcin', other_designations='RC|SMP-30|gluconolactonase|senescence marker protein 30', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20549767, end_position_on_the_genomic_accession=20562089, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19734, CurrentID=0, Status='live', Symbol='Rgs16', Aliases='Rgs14, Rgsr', description='regulator of G-protein signaling 16', other_designations='A28-RGS14P|RGS-R|retinal-specific RGS|retinally abundant regulator of G-protein signaling', map_location='1 65.43 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153740353, end_position_on_the_genomic_accession=153745468, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19735, CurrentID=0, Status='live', Symbol='Rgs2', Aliases='GOS8', description='regulator of G-protein signaling 2', other_designations='', map_location='1 62.56 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=143999338, end_position_on_the_genomic_accession=144004149, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19736, CurrentID=0, Status='live', Symbol='Rgs4', Aliases='AA004315, AA597169, ESTM48, ESTM50', description='regulator of G-protein signaling 4', other_designations='', map_location='1 76.84 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=169741477, end_position_on_the_genomic_accession=169747642, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19737, CurrentID=0, Status='live', Symbol='Rgs5', Aliases='1110070A02Rik', description='regulator of G-protein signaling 5', other_designations='', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=169655501, end_position_on_the_genomic_accession=169693526, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19739, CurrentID=0, Status='live', Symbol='Rgs9', Aliases='RGS9-1-2, Rgs9', description='regulator of G-protein signaling 9', other_designations='ROS GAP', map_location='11 71.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109225348, end_position_on_the_genomic_accession=109298181, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  19742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19742, CurrentID=0, Status='live', Symbol='rh', Aliases='', description='rachiterata', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19743, CurrentID=0, Status='live', Symbol='Rhag', Aliases='CD241, Rh50, Rh50A', description='Rhesus blood group-associated A glycoprotein', other_designations='50 kD glycoprotein|erythrocyte membrane glycoprotein Rh50|rh family type A glycoprotein|rh type A glycoprotein|rhesus blood group family type A glycoprotein', map_location='17 19.54 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40811099, end_position_on_the_genomic_accession=40840754, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19744, CurrentID=0, Status='live', Symbol='Rheb', Aliases='Rheb1', description='Ras homolog enriched in brain', other_designations='RAS-homolog enriched in brain', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24802823, end_position_on_the_genomic_accession=24842361, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19746, CurrentID=0, Status='live', Symbol='Rhd', Aliases='Rh, Rhced, Rhl1', description='Rh blood group, D antigen', other_designations='Rh blood group protein|Rh-like polypeptide|Rhesus blood group CE and D|erythrocyte membrane glycoprotein Rh30', map_location='4 67.13 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134864536, end_position_on_the_genomic_accession=134896172, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19752, CurrentID=0, Status='live', Symbol='Rnase1', Aliases='AI574248, Rib-1, Rib1', description='ribonuclease, RNase A family, 1 (pancreatic)', other_designations='RNase 1|RNase A|ribonuclease 1, pancreatic', map_location='14 B-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51145002, end_position_on_the_genomic_accession=51146767, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19762, CurrentID=0, Status='live', Symbol='Rit2', Aliases='RIBA, Rin, Roc2', description='Ras-like without CAAX 2', other_designations='GTP-binding protein Roc2|RAS-like protein expressed in neuron|ras-like protein expressed in neurons|ras-like without CAAX protein 2', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=30974314, end_position_on_the_genomic_accession=31317128, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19763, CurrentID=0, Status='live', Symbol='Ring1', Aliases='Ring1A', description='ring finger protein 1', other_designations='polycomb complex protein RING1|transcription repressor Ring1A', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34020792, end_position_on_the_genomic_accession=34024680, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19765, CurrentID=0, Status='live', Symbol='Ralbp1', Aliases='RLIP76, Rip1', description='ralA binding protein 1', other_designations='DNP-SG ATPase|Ral-interacting protein 1|dinitrophenyl S-glutathione ATPase', map_location='17 35.26 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65848415, end_position_on_the_genomic_accession=65885755, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  19766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19766, CurrentID=0, Status='live', Symbol='Ripk1', Aliases='D330015H01Rik, RIP, Rinp, Rip1', description='receptor (TNFRSF)-interacting serine-threonine kinase 1', other_designations='RIP-1|cell death protein RIP|receptor-interacting protein 1|serine/threonine-protein kinase RIP', map_location='13 14.01 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34000844, end_position_on_the_genomic_accession=34035170, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  19769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19769, CurrentID=0, Status='live', Symbol='Rit1', Aliases='RIBB, ROC1, Rit', description='Ras-like without CAAX 1', other_designations='RAS-like protein expressed in many tissues|ras-like without CAAX protein 1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88716840, end_position_on_the_genomic_accession=88731048, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19771, CurrentID=0, Status='live', Symbol='Rlbp1', Aliases='3110056M11Rik, CRALBP', description='retinaldehyde binding protein 1', other_designations='cellular retinaldehyde-binding protein', map_location='7 45.0 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79374865, end_position_on_the_genomic_accession=79387043, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  19773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19773, CurrentID=0, Status='live', Symbol='Rln1', Aliases='Rln, rlx', description='relaxin 1', other_designations='', map_location='19 23.83 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29331755, end_position_on_the_genomic_accession=29334670, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19775, CurrentID=0, Status='live', Symbol='Xpr1', Aliases='Rmc-1, Rmc1, Sxv, Syg1, XR', description='xenotropic and polytropic retrovirus receptor 1', other_designations='receptor for MCF virus 1|suppressor of yeast G<a> deletion|susceptibility to xenotropic virus|xenotropic and polytropic murine leukemia virus receptor|xenotropic and polytropic retrovirus receptor 1 homolog', map_location='1 66.49 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155275657, end_position_on_the_genomic_accession=155417444, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  19777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19777, CurrentID=0, Status='live', Symbol='Uri1', Aliases='C80913, NNX3, Rmp', description='URI1, prefoldin-like chaperone', other_designations='RNA polymerase II subunit 5-mediating protein|RPB5-mediating protein|protein phosphatase 1 regulatory subunit 19', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=37959992, end_position_on_the_genomic_accession=38023575, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  19782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19782, CurrentID=0, Status='live', Symbol='Rmrp', Aliases='1110032O22Rikr, Rmrp', description='RNA component of mitochondrial RNAase P', other_designations='RNA component of mitochondrial RNAase P, 1|RNase MRP', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43492785, end_position_on_the_genomic_accession=43493059, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19791, CurrentID=0, Status='live', Symbol='Rn18s', Aliases='', description='18S ribosomal RNA', other_designations='18S RNA', map_location='', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3180891, end_position_on_the_genomic_accession=3201670, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  19799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19799, CurrentID=0, Status='live', Symbol='Rn4.5s', Aliases='', description='4.5S RNA', other_designations='', map_location='6 22.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47654921, end_position_on_the_genomic_accession=47655094, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19804, CurrentID=0, Status='live', Symbol='Rn5s', Aliases='', description='5S RNA', other_designations='', map_location='8 E', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123598156, end_position_on_the_genomic_accession=123598276, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19817, CurrentID=0, Status='live', Symbol='Rn7sk', Aliases='', description='RNA, 7SK, nuclear', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78175302, end_position_on_the_genomic_accession=78175633, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19819, CurrentID=0, Status='live', Symbol='Rnaseh1', Aliases='', description='ribonuclease H1', other_designations='RNase H1', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28649602, end_position_on_the_genomic_accession=28659591, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19820, CurrentID=0, Status='live', Symbol='Rlim', Aliases='AL022832, AW743871, Ha1r1, Rnf12, Rlim', description='ring finger protein, LIM domain interacting', other_designations='Hoxa1 regulated|LIM domain-interacting RING finger protein|R-LIM|RING finger LIM domain-binding protein|ring finger protein 12', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103957164, end_position_on_the_genomic_accession=103981297, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19821, CurrentID=0, Status='live', Symbol='Rnf2', Aliases='AI326319, AI450156, AU019207, Ring1B, dinG', description='ring finger protein 2', other_designations='RING finger protein 1B', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151469406, end_position_on_the_genomic_accession=151500923, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  19822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19822, CurrentID=0, Status='live', Symbol='Rnf4', Aliases='AU018689, Gtrgeo8, SNURF', description='ring finger protein 4', other_designations='E3 ubiquitin ligase RNF4', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34336250, end_position_on_the_genomic_accession=34353445, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19823, CurrentID=0, Status='live', Symbol='Rnf7', Aliases='Rbx2, SAG', description='ring finger protein 7', other_designations='sensitive to apoptosis gene protein', map_location='9 D2.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96470957, end_position_on_the_genomic_accession=96478701, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19824, CurrentID=0, Status='live', Symbol='Trim10', Aliases='AI324236, BB139825, Herf1, Rnf9', description='tripartite motif-containing 10', other_designations='hematopoietic RING finger 1|ring finger protein 9|tripartite motif protein 10', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36869574, end_position_on_the_genomic_accession=36877833, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19826, CurrentID=0, Status='live', Symbol='Rnps1', Aliases='', description='ribonucleic acid binding protein S1', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24414675, end_position_on_the_genomic_accession=24425898, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  19842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19842, CurrentID=0, Status='live', Symbol='Rnu1a1', Aliases='Rnu1a-1', description='U1a1 small nuclear RNA', other_designations='', map_location='11 D-E', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87422867, end_position_on_the_genomic_accession=87423030, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19844, CurrentID=0, Status='live', Symbol='Rnu1b1', Aliases='Rnu1b-1', description='U1b1 small nuclear RNA', other_designations='', map_location='3 41.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96359953, end_position_on_the_genomic_accession=96360117, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19845, CurrentID=0, Status='live', Symbol='Rnu1b2', Aliases='pU1.2', description='U1b2 small nuclear RNA', other_designations='', map_location='3 41.79 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96374021, end_position_on_the_genomic_accession=96374185, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19848, CurrentID=0, Status='live', Symbol='Rnu2-10', Aliases='Rnu2, nmf291', description='U2 small nuclear RNA 10', other_designations='', map_location='11 65.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101658238, end_position_on_the_genomic_accession=101658424, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19850, CurrentID=0, Status='live', Symbol='Rnu3a', Aliases='', description='U3A small nuclear RNA', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40258576, end_position_on_the_genomic_accession=40258679, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19859, CurrentID=0, Status='live', Symbol='Rnu3b2', Aliases='Rnu3b-2', description='U3B small nuclear RNA 2', other_designations='', map_location='11 52.11 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87471368, end_position_on_the_genomic_accession=87471582, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19862, CurrentID=0, Status='live', Symbol='Rnu6', Aliases='', description='U6 small nuclear RNA', other_designations='', map_location='17 8.88 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13432033, end_position_on_the_genomic_accession=13432139, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19866, CurrentID=0, Status='live', Symbol='Rnu7', Aliases='', description='U7 small nuclear RNA', other_designations='', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124741225, end_position_on_the_genomic_accession=124741286, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19868, CurrentID=0, Status='live', Symbol='Rnu7-ps2', Aliases='', description='U7 small nuclear RNA, pseudogene 2', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=95565542, end_position_on_the_genomic_accession=95565604, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  19872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19872, CurrentID=0, Status='live', Symbol='Rny1', Aliases='', description='RNA, Y1 small cytoplasmic, Ro-associated', other_designations='', map_location='6 22.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47788070, end_position_on_the_genomic_accession=47788180, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  19874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19874, CurrentID=0, Status='live', Symbol='Rny3', Aliases='', description='RNA, Y3 small cytoplasmic (associated with Ro protein)', other_designations='', map_location='6 22.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47781624, end_position_on_the_genomic_accession=47781724, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  19876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19876, CurrentID=0, Status='live', Symbol='Robo1', Aliases='AW494633, AW742721, DUTT1, Gm310', description='roundabout homolog 1 (Drosophila)', other_designations='', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=72027553, end_position_on_the_genomic_accession=73047491, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  19877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19877, CurrentID=0, Status='live', Symbol='Rock1', Aliases='1110055K06Rik, Rock-I', description='Rho-associated coiled-coil containing protein kinase 1', other_designations='Rho-associated coiled-coil forming kinase 1|p160 ROCK-1|p160ROCK|rho-associated, coiled-coil-containing protein kinase 1|rho-associated, coiled-coil-containing protein kinase I', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10064401, end_position_on_the_genomic_accession=10181792, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  19878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19878, CurrentID=0, Status='live', Symbol='Rock2', Aliases='B230113H15Rik, ROKalpha, Rho-kinase, Rock-IIm, mKIAA0619, Rock2', description='Rho-associated coiled-coil containing protein kinase 2', other_designations='Rho-associated coiled-coil forming kinage 2|p164 ROCK-2|rho-associated, coiled-coil-containing protein kinase 2|rho-associated, coiled-coil-containing protein kinase II', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=16894978, end_position_on_the_genomic_accession=16988274, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  19879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19879, CurrentID=0, Status='live', Symbol='Slc22a8', Aliases='Oat3, Roct', description='solute carrier family 22 (organic anion transporter), member 8', other_designations='organic anion transporter 3|reduced in osteosclerosis transporter|solute carrier family 22, member 8', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8591254, end_position_on_the_genomic_accession=8611835, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  19881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19881, CurrentID=0, Status='live', Symbol='Rom1', Aliases='M101156, Rgsc1156, Rom-1, Rosp1, Tspan23', description='rod outer segment membrane protein 1', other_designations='', map_location='19 6.03 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8927382, end_position_on_the_genomic_accession=8929356, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  19882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19882, CurrentID=0, Status='live', Symbol='Mst1r', Aliases='CD136, CDw136, Fv-2, Fv2, PTK8, Ron, STK', description='macrophage stimulating 1 receptor (c-met-related tyrosine kinase)', other_designations='MSP receptor|friend virus susceptibility 2|p185-Ron|receptor protein tyrosine kinase, c-met-related|stem cell-derived tyrosine kinase', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107906860, end_position_on_the_genomic_accession=107920383, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  19883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19883, CurrentID=0, Status='live', Symbol='Rora', Aliases='9530021D13Rik, Nr1f1, ROR1, ROR2, ROR3, nmf267, sg, staggerer, tmgc26', description='RAR-related orphan receptor alpha', other_designations='RAR-related orphan receptor A|nuclear receptor RZR-alpha|nuclear receptor subfamily 1 group F member 1|retinoid-related orphan receptor-alpha', map_location='9 37.45 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=68653402, end_position_on_the_genomic_accession=69388246, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  19885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19885, CurrentID=0, Status='live', Symbol='Rorc', Aliases='Nr1f3, RORgamma, TOR, Thor', description='RAR-related orphan receptor gamma', other_designations='RAR-related orphan receptor C|nuclear receptor RZR-gamma|nuclear receptor subfamily 1 group F member 3|retinoid-related orphan receptor-gamma|thymus orphan receptor', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94372715, end_position_on_the_genomic_accession=94398276, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  19886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19886, CurrentID=0, Status='live', Symbol='Ros1', Aliases='Ros-1, c-ros', description='Ros1 proto-oncogene', other_designations='c-Ros receptor tyrosine kinase|c-ros-1|proto-oncogene c-Ros|proto-oncogene c-Ros-1|receptor tyrosine kinase c-ros oncogene 1', map_location='10 26.18 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52044865, end_position_on_the_genomic_accession=52195397, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  19888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19888, CurrentID=0, Status='live', Symbol='Rp1', Aliases='Dcdc3, Orp1h, mG145, Rp1', description='retinitis pigmentosa 1 (human)', other_designations='retinitis pigmentosa 1 homolog|retinitis pigmentosa RP1 protein homolog', map_location='1 1.65 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=4290846, end_position_on_the_genomic_accession=4409266, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  19889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19889, CurrentID=0, Status='live', Symbol='Rp2', Aliases='AI662636h, Rp2', description='retinitis pigmentosa 2 homolog (human)', other_designations='', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20364481, end_position_on_the_genomic_accession=20405653, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  19891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19891, CurrentID=0, Status='live', Symbol='Rpa2', Aliases='AA409079, AI325195, AU020965, RPA34, Rf-A2', description='replication protein A2', other_designations='30-kDa protein|RF-A protein 2|RP-A p32|RP-A p34|replication factor A protein 2|replication protein A 34 kDa subunit', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132768360, end_position_on_the_genomic_accession=132778746, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  19892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19892, CurrentID=0, Status='live', Symbol='Rpe65', Aliases='65kDa, A930029L06Rik, LCA2, Mord1, RP20, rd12', description='retinal pigment epithelium 65', other_designations='all-trans-retinyl-palmitate hydrolase|modifier of retinal degeneration 1|retinal pigment epithelium, 65 kDa|retinal pigment epithelium-specific 65 kDa protein|retinol isomerase', map_location='3 82.52 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=159599075, end_position_on_the_genomic_accession=159625307, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  19893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19893, CurrentID=0, Status='live', Symbol='Rpgr', Aliases='Rp3h', description='retinitis pigmentosa GTPase regulator', other_designations='mRpgr', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=10158216, end_position_on_the_genomic_accession=10216795, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  19894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19894, CurrentID=0, Status='live', Symbol='Rph3a', Aliases='2900002P20Rik, AU022689, AW108370', description='rabphilin 3A', other_designations='Doc2 family|exophilin-1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120940499, end_position_on_the_genomic_accession=121010102, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  19895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19895, CurrentID=0, Status='live', Symbol='Rpia', Aliases='RPI', description='ribose 5-phosphate isomerase A', other_designations='phosphoriboisomerase', map_location='6 32.06 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70765720, end_position_on_the_genomic_accession=70792175, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  19896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19896, CurrentID=0, Status='live', Symbol='Rpl10a', Aliases='CsA-19, Nedd6', description='ribosomal protein L10A', other_designations='NEDD-6|neural precursor cell expressed developmentally down-regulated protein 6', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28328471, end_position_on_the_genomic_accession=28331033, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19899, CurrentID=0, Status='live', Symbol='Rpl18', Aliases='L18a, Rpl18', description='ribosomal protein L18', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45718071, end_position_on_the_genomic_accession=45720835, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  330998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330998, CurrentID=0, Status='live', Symbol='Ankrd34c', Aliases='B230218L05Rik, BB276295, Gm771', description='ankyrin repeat domain 34C', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89723933, end_position_on_the_genomic_accession=89738877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  19921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19921, CurrentID=0, Status='live', Symbol='Rpl19', Aliases='', description='ribosomal protein L19', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98023080, end_position_on_the_genomic_accession=98030493, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19933, CurrentID=0, Status='live', Symbol='Rpl21', Aliases='8430440E03Rik, L21', description='ribosomal protein L21', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146832890, end_position_on_the_genomic_accession=146837032, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19934, CurrentID=0, Status='live', Symbol='Rpl22', Aliases='2700038K18Rik', description='ribosomal protein L22', other_designations='heparin-binding protein HBp15', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152324436, end_position_on_the_genomic_accession=152334082, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  19935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19935, CurrentID=0, Status='live', Symbol='Mrpl23', Aliases='L23mrp, Rpl23, Rpl23l', description='mitochondrial ribosomal protein L23', other_designations='L23 mitochondrial-related protein|L23mt|MRP-L23', map_location='7 87.95 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142533030, end_position_on_the_genomic_accession=142540747, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  19941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19941, CurrentID=0, Status='live', Symbol='Rpl26', Aliases='SIG-20', description='ribosomal protein L26', other_designations='silica-induced gene 20 protein', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68901566, end_position_on_the_genomic_accession=68904537, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19942, CurrentID=0, Status='live', Symbol='Rpl27', Aliases='', description='ribosomal protein L27', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101442245, end_position_on_the_genomic_accession=101445596, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19943, CurrentID=0, Status='live', Symbol='Rpl28', Aliases='D7Wsu21e', description='ribosomal protein L28', other_designations='', map_location='7 2.77 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4792888, end_position_on_the_genomic_accession=4794562, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19944, CurrentID=0, Status='live', Symbol='Rpl29', Aliases='Rpl43', description='ribosomal protein L29', other_designations='', map_location='9 F2-F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106429432, end_position_on_the_genomic_accession=106431568, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19946, CurrentID=0, Status='live', Symbol='Rpl30', Aliases='', description='ribosomal protein L30', other_designations='', map_location='15 14.33 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34440505, end_position_on_the_genomic_accession=34443276, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  19951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19951, CurrentID=0, Status='live', Symbol='Rpl32', Aliases='AU020185, rpL32-3A', description='ribosomal protein L32', other_designations='snoRNA MBI-141', map_location='6 53.72 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115805514, end_position_on_the_genomic_accession=115808743, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  19952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19952, CurrentID=0, Status='live', Symbol='Rpl32-ps', Aliases='', description='ribosomal protein L32, pseudogene', other_designations='', map_location='3 32.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69716941, end_position_on_the_genomic_accession=69717444, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  19981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19981, CurrentID=0, Status='live', Symbol='Rpl37a', Aliases='', description='ribosomal protein L37a', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72711260, end_position_on_the_genomic_accession=72713813, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  19982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19982, CurrentID=0, Status='live', Symbol='Rpl36a', Aliases='L44L, Rpl44', description='ribosomal protein L36A', other_designations='60S ribosomal protein L44|ribosomal protein L44', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134585654, end_position_on_the_genomic_accession=134588062, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  19988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19988, CurrentID=0, Status='live', Symbol='Rpl6', Aliases='Taxreb107', description='ribosomal protein L6', other_designations='TAX-responsive enhancer element-binding protein 107', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121204501, end_position_on_the_genomic_accession=121209241, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  19989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19989, CurrentID=0, Status='live', Symbol='Rpl7', Aliases='Rpl7a, Surf-3', description='ribosomal protein L7', other_designations='', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16101295, end_position_on_the_genomic_accession=16104433, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20005, CurrentID=0, Status='live', Symbol='Rpl9', Aliases='', description='ribosomal protein L9', other_designations='', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65388364, end_position_on_the_genomic_accession=65391431, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20014, CurrentID=0, Status='live', Symbol='Rpn2', Aliases='1300012C06Rik, AV261018, Rpn-2', description='ribophorin II', other_designations='RPN-II|dolichyl-diphosphooligosaccharide--protein glycosyltransferase 63 kDa subunit|ribophorin-2', map_location='2 78.2 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157279011, end_position_on_the_genomic_accession=157326318, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  20016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20016, CurrentID=0, Status='live', Symbol='Polr1c', Aliases='40kDa, AA409007, AA959927, AL024089, Polr1e, RPA40, Rpo1-1', description='polymerase (RNA) I polypeptide C', other_designations='AC40|DNA-directed RNA polymerase I subunit C|DNA-directed RNA polymerases I and III 40 kDa polypeptide|RNA polymerase 1-1 (40 kDa subunit)|RNA polymerases I and III subunit AC1|RPC40', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46243920, end_position_on_the_genomic_accession=46248045, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20017, CurrentID=0, Status='live', Symbol='Polr1b', Aliases='128kDa, D630020H17Rik, RPA116, RPA135, RPA2, Rpo1-2', description='polymerase (RNA) I polypeptide B', other_designations='DNA-directed RNA polymerase I 135 kDa polypeptide|DNA-directed RNA polymerase I 135kDa polypeptide|RNA polymerase 1-2, 128 kDa subunit|RNA polymerase I subunit 2', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129100970, end_position_on_the_genomic_accession=129128060, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20018, CurrentID=0, Status='live', Symbol='Polr1d', Aliases='1110003G10Rik, AC19, RPA16, Rpo1-3', description='polymerase (RNA) I polypeptide D', other_designations='DNA-directed RNA polymerase I 16 kDa polypeptide|RNA polymerase 1-3|RNA polymerases I and III subunit AC2', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147077346, end_position_on_the_genomic_accession=147111361, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20019, CurrentID=0, Status='live', Symbol='Polr1a', Aliases='194kDa, 3010014K16Rik, RPA194, Rpo1-4, mRPA1', description='polymerase (RNA) I polypeptide A', other_designations='DNA-directed RNA polymerase I largest subunit|DNA-directed RNA polymerase I subunit A|RNA polymerase 1-4 (194 kDa subunit)|RNA polymerase I 194 kDa subunit|RNA polymerase I subunit A1', map_location='6 32.21 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71909053, end_position_on_the_genomic_accession=71979435, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  20020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20020, CurrentID=0, Status='live', Symbol='Polr2a', Aliases='220kDa, Rpb1, Rpo2-1', description='polymerase (RNA) II (DNA directed) polypeptide A', other_designations='DNA-directed RNA polymerase II subunit A|DNA-directed RNA polymerase III largest subunit|RNA polymerase II 1|RNA polymerase II subunit B1', map_location='11 B1-C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69733997, end_position_on_the_genomic_accession=69758633, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  20021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20021, CurrentID=0, Status='live', Symbol='Polr2c', Aliases='33kDa, RPB3, Rpo2-3, mRBP31', description='polymerase (RNA) II (DNA directed) polypeptide C', other_designations='DNA-directed RNA polymerase II 33 kDa polypeptide|DNA-directed RNA polymerase II subunit C|RNA polymerase II 3|RNA polymerase II subunit 3|RNA polymerase II subunit B3|RPB31|RPB33', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94857450, end_position_on_the_genomic_accession=94864242, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20022, CurrentID=0, Status='live', Symbol='Polr2j', Aliases='14.5kDa, Polr2i, Rpb11a, Rpo2-4', description='polymerase (RNA) II (DNA directed) polypeptide J', other_designations='DNA-directed RNA polymerase II 13.3 kDa polypeptide|DNA-directed RNA polymerase II subunit J|RNA polymerase II 13.3 kDa subunit|RNA polymerase II 4 (14 kDa subunit)|RNA polymerase II subunit B11|RNA polymerase II subunit RPB14|RPB14|polymerase (RNA) II (DNA directed) polypeptide I', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136116691, end_position_on_the_genomic_accession=136122947, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20024, CurrentID=0, Status='live', Symbol='Sub1', Aliases='AI842364, P15, P9, Pc4, Rpo2tc1', description='SUB1 homolog (S. cerevisiae)', other_designations='RNA polymerase II transcriptional coactivator|p14|positive cofactor 4|single-stranded DNA-binding protein p9', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=11981338, end_position_on_the_genomic_accession=11996031, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20028, CurrentID=0, Status='live', Symbol='Pdc', Aliases='Rpr-1, Rpr1', description='phosducin', other_designations='33 kDa phototransducing protein|PHD|rod photoreceptor 1', map_location='1 63.84 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=150319417, end_position_on_the_genomic_accession=150333906, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20042, CurrentID=0, Status='live', Symbol='Rps12', Aliases='', description='ribosomal protein S12', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23785183, end_position_on_the_genomic_accession=23787209, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20044, CurrentID=0, Status='live', Symbol='Rps14', Aliases='2600014J02Rik, AL023078', description='ribosomal protein S14', other_designations='', map_location='18 34.4 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60774596, end_position_on_the_genomic_accession=60778546, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20054, CurrentID=0, Status='live', Symbol='Rps15', Aliases='rig', description='ribosomal protein S15', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80292431, end_position_on_the_genomic_accession=80294114, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20055, CurrentID=0, Status='live', Symbol='Rps16', Aliases='AA420385, AI317031', description='ribosomal protein S16', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28350689, end_position_on_the_genomic_accession=28352698, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20068, CurrentID=0, Status='live', Symbol='Rps17', Aliases='', description='ribosomal protein S17', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81342733, end_position_on_the_genomic_accession=81345234, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20084, CurrentID=0, Status='live', Symbol='Rps18', Aliases='H-2Ke3, H2-Ke3, Ke-3, ke3', description='ribosomal protein S18', other_designations='', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33951999, end_position_on_the_genomic_accession=33955641, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20085, CurrentID=0, Status='live', Symbol='Rps19', Aliases='Dsk3', description='ribosomal protein S19', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24884376, end_position_on_the_genomic_accession=24889806, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20088, CurrentID=0, Status='live', Symbol='Rps24', Aliases='', description='ribosomal protein S24', other_designations='MRP S24', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=24490633, end_position_on_the_genomic_accession=24496146, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20090, CurrentID=0, Status='live', Symbol='Rps29', Aliases='', description='ribosomal protein S29', other_designations='S29 ribosomal protein', map_location='12 H3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69157722, end_position_on_the_genomic_accession=69159186, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20091, CurrentID=0, Status='live', Symbol='Rps3a1', Aliases='Rps3a', description='ribosomal protein S3A1', other_designations='ribosomal protein S3A', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=86137940, end_position_on_the_genomic_accession=86142668, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20102, CurrentID=0, Status='live', Symbol='Rps4x', Aliases='Rps4, Rps4-1', description='ribosomal protein S4, X-linked', other_designations='', map_location='X 45.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102184941, end_position_on_the_genomic_accession=102189306, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20103, CurrentID=0, Status='live', Symbol='Rps5', Aliases='AA617411', description='ribosomal protein S5', other_designations='S5 ribosomal protein', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12922311, end_position_on_the_genomic_accession=12926686, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20104, CurrentID=0, Status='live', Symbol='Rps6', Aliases='S6R', description='ribosomal protein S6', other_designations='phosphoprotein NP33', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86854099, end_position_on_the_genomic_accession=86857367, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20111, CurrentID=0, Status='live', Symbol='Rps6ka1', Aliases='Mapkapk-1a, Rsk, Rsk-1, Rsk1, S6K-alpha-1, p90-Rsk1, p90Rsk1, p90S6K, p90rsk', description='ribosomal protein S6 kinase polypeptide 1', other_designations='90 kDa ribosomal protein S6 kinase 1|MAP kinase-activated protein kinase 1a|MAPK-activated protein kinase 1a|MAPKAP kinase 1a|ribosomal protein S6 kinase, 90kD, polypeptide 1; Ribosomal protein S6 kinase, 90kD, 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133847290, end_position_on_the_genomic_accession=133887797, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  20112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20112, CurrentID=0, Status='live', Symbol='Rps6ka2', Aliases='90kDa, D17Wsu134e, Rps6ka-rs1, Rsk3, p90rsk, pp90rsk', description='ribosomal protein S6 kinase, polypeptide 2', other_designations='90 kDa ribosomal protein S6 kinase 2|MAP kinase-activated protein kinase 1c|MAPK-activated protein kinase 1c|MAPKAP kinase 1c|MAPKAPK-1c|RSK-3|S6K-alpha 2|S6K-alpha-2|p90-RSK 2|p90RSK2|pp90RSK3|protein-tyrosine kinase Mpk-9|ribosomal S6 kinase 3|ribosomal protein S6 kinase alpha 2|ribosomal protein S6 kinase, 90kD, polypeptide 2', map_location='17 F4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=7088975, end_position_on_the_genomic_accession=7303315, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  20115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20115, CurrentID=0, Status='live', Symbol='Rps7', Aliases='MtuA, S7, Rps7', description='ribosomal protein S7', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28630847, end_position_on_the_genomic_accession=28635953, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20116, CurrentID=0, Status='live', Symbol='Rps8', Aliases='', description='ribosomal protein S8', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117153836, end_position_on_the_genomic_accession=117156132, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20128, CurrentID=0, Status='live', Symbol='Trim30a', Aliases='Rpt-1, Rpt1, Trim30', description='tripartite motif-containing 30A', other_designations='down regulatory protein of interleukin-2 receptor|regulatory protein, T lymphocyte 1|tripartite motif protein 30|tripartite motif-containing 30|tripartite motif-containing protein 30', map_location='7 55.69 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104409025, end_position_on_the_genomic_accession=104465193, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20129, CurrentID=0, Status='live', Symbol='Rptn', Aliases='', description='repetin', other_designations='', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93393699, end_position_on_the_genomic_accession=93399442, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20130, CurrentID=0, Status='live', Symbol='Rras', Aliases='AI573426', description='Harvey rat sarcoma oncogene, subgroup R', other_designations='p23', map_location='7 29.08 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45018007, end_position_on_the_genomic_accession=45021647, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20132, CurrentID=0, Status='live', Symbol='Rrh', Aliases='', description='retinal pigment epithelium derived rhodopsin homolog', other_designations='peropsin', map_location='3 59.09 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129808575, end_position_on_the_genomic_accession=129822505, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20133, CurrentID=0, Status='live', Symbol='Rrm1', Aliases='RnrM1', description='ribonucleotide reductase M1', other_designations='ribonucleoside-diphosphate reductase subunit M1|ribonucleotide reductase large subunit', map_location='7 54.72 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102441695, end_position_on_the_genomic_accession=102469771, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  20135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20135, CurrentID=0, Status='live', Symbol='Rrm2', Aliases='AA407299, R2', description='ribonucleotide reductase M2', other_designations='ribonucleoside-diphosphate reductase M2 chain|ribonucleotide reductase small chain|ribonucleotide reductase small subunit', map_location='12 8.5 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24708254, end_position_on_the_genomic_accession=24714146, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20146, CurrentID=0, Status='live', Symbol='rs', Aliases='gc, grc', description='recessive spotting', other_designations='grey coat', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20147, CurrentID=0, Status='live', Symbol='Rs1', Aliases='Rs1h, Xlrs1, tmgc1', description='retinoschisis (X-linked, juvenile) 1 (human)', other_designations='X-linked juvenile retinoschisis protein homolog|retinoschisis 1 homolog', map_location='X 73.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=160768013, end_position_on_the_genomic_accession=160799663, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20148, CurrentID=0, Status='live', Symbol='Dhrs3', Aliases='Rsdr1, retSDR1', description='dehydrogenase/reductase (SDR family) member 3', other_designations='retinal short-chain dehydrogenase/reductase 1', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144892609, end_position_on_the_genomic_accession=144927645, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20150, CurrentID=0, Status='live', Symbol='rsm1', Aliases='rsm-1', description='defective resistance to Schistosoma mansoni', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20163, CurrentID=0, Status='live', Symbol='Rsu1', Aliases='RsuI, rsp-1', description='Ras suppressor protein 1', other_designations='rsu-1', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=13076967, end_position_on_the_genomic_accession=13271415, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20164, CurrentID=0, Status='live', Symbol='Opn1lw', Aliases='Rsvp', description='opsin 1 (cone pigments), long-wave-sensitive', other_designations='red sensitive visual pigment', map_location='X', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20166, CurrentID=0, Status='live', Symbol='Rtkn', Aliases='', description='rhotekin', other_designations='rtkn', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83134439, end_position_on_the_genomic_accession=83152579, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  20167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20167, CurrentID=0, Status='live', Symbol='Rtn2', Aliases='MMS10-P, Ms10p, NSPLI, Nspl1', description='reticulon 2 (Z-band associated protein)', other_designations='NSP-like protein 1|NSP-like protein I|neuroendocrine-specific protein-like 1|neuroendocrine-specific protein-like I', map_location='7 9.54 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19282624, end_position_on_the_genomic_accession=19296164, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20168, CurrentID=0, Status='live', Symbol='Rtn3', Aliases='RTN3-A1', description='reticulon 3', other_designations='Nsp-like II', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7425895, end_position_on_the_genomic_accession=7483291, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20170, CurrentID=0, Status='live', Symbol='Hps6', Aliases='5330434M19Rik, BLOC2, ru', description='Hermansky-Pudlak syndrome 6', other_designations='ruby eye|ruby-eye protein', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46003478, end_position_on_the_genomic_accession=46006173, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  20174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20174, CurrentID=0, Status='live', Symbol='Ruvbl2', Aliases='mp47, p47', description='RuvB-like protein 2', other_designations='reptin', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45421896, end_position_on_the_genomic_accession=45435170, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  20181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20181, CurrentID=0, Status='live', Symbol='Rxra', Aliases='9530071D11Rik, Nr2b1, RXRalpha1', description='retinoid X receptor alpha', other_designations='RXR alpha 1|nuclear receptor subfamily 2 group B member 1', map_location='2 19.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27676324, end_position_on_the_genomic_accession=27763319, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  20182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20182, CurrentID=0, Status='live', Symbol='Rxrb', Aliases='AL023085, H-2RIIBP, Nr2b2, RCoR-1, Rub', description='retinoid X receptor beta', other_designations='MHC class I regulatory element-binding protein H-2RIIBP|nuclear receptor subfamily 2 group B member 2|putative retinoid X receptor beta', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34031812, end_position_on_the_genomic_accession=34038403, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20183, CurrentID=0, Status='live', Symbol='Rxrg', Aliases='Nr2b3', description='retinoid X receptor gamma', other_designations='nuclear receptor subfamily 2 group B member 3', map_location='1 74.99 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167598362, end_position_on_the_genomic_accession=167639623, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20184, CurrentID=0, Status='live', Symbol='Uimc1', Aliases='9430016E08Rik, D330018D10Rik, D630032M02Rik, RIP110, Rxrip110', description='ubiquitin interaction motif containing 1', other_designations='receptor-associated protein 80|retinoid X receptor interacting protein 110|ubiquitin interaction motif-containing protein 1', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55027880, end_position_on_the_genomic_accession=55100309, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  20185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20185, CurrentID=0, Status='live', Symbol='Ncor1', Aliases='5730405M06Rik, A230020K14Rik, N-CoR, RIP13, Rxrip13, mKIAA1047', description='nuclear receptor co-repressor 1', other_designations='N-CoR1|retinoid X receptor interacting protein 13|retinoid X receptor-interacting protein 13', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62315643, end_position_on_the_genomic_accession=62457332, orientation='minus', exon_count=49, OMIM='', no_hdr0=''),
  20186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20186, CurrentID=0, Status='live', Symbol='Nr1h4', Aliases='AI957360, Fxr, HRR1, RIP14, Rxrip14', description='nuclear receptor subfamily 1, group H, member 4', other_designations='RXR-interacting protein 14|farnesoid X activated receptor|farnesoid X-activated receptor|farnesol receptor HRR-1|retinoid X receptor-interacting protein 14', map_location='10 44.98 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=89454234, end_position_on_the_genomic_accession=89533645, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  20187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20187, CurrentID=0, Status='live', Symbol='Ryk', Aliases='AW536699, ERK-3, Vik', description='receptor-like tyrosine kinase', other_designations='NYK-R|growth factor receptor|kinase VIK|met-related kinase', map_location='9 54.72 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=102834920, end_position_on_the_genomic_accession=102908307, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20190, CurrentID=0, Status='live', Symbol='Ryr1', Aliases='AI528790, RYR-1, Ryr, skrr', description='ryanodine receptor 1, skeletal muscle', other_designations='calcium release channel|skeletal muscle calcium release channel|skeletal muscle ryanodine receptor|skeletal muscle-type ryanodine receptor|type 1 ryanodine receptor', map_location='7 A2-B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29003340, end_position_on_the_genomic_accession=29125179, orientation='minus', exon_count=110, OMIM='', no_hdr0=''),
  20191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20191, CurrentID=0, Status='live', Symbol='Ryr2', Aliases='9330127I20Rik, RYR-2', description='ryanodine receptor 2, cardiac', other_designations='calcium release channel|cardiac muscle ryanodine receptor|cardiac muscle ryanodine receptor-calcium release channel|ryanodine receptor type 2|type 2 ryanodine receptor', map_location='13 A1-A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=11553099, end_position_on_the_genomic_accession=12106945, orientation='minus', exon_count=108, OMIM='', no_hdr0=''),
  20192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20192, CurrentID=0, Status='live', Symbol='Ryr3', Aliases='AI851294, C230090H21, RYR-3', description='ryanodine receptor 3', other_designations='brain ryanodine receptor-calcium release channel|brain-type ryanodine receptor|calcium release channel|type 3 ryanodine receptor', map_location='2 E5-F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112631380, end_position_on_the_genomic_accession=113030331, orientation='minus', exon_count=104, OMIM='', no_hdr0=''),
  20193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20193, CurrentID=0, Status='live', Symbol='S100a1', Aliases='AI266795, S100, S100a', description='S100 calcium binding protein A1', other_designations='S-100 protein alpha chain|S-100 protein subunit alpha|S100 calcium-binding protein A1', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90511034, end_position_on_the_genomic_accession=90514330, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20194, CurrentID=0, Status='live', Symbol='S100a10', Aliases='42C, AA409961, AL024248, CAL12, CLP11, Cal1l, p10, p11', description='S100 calcium binding protein A10 (calpactin)', other_designations='S100 calcium binding protein A10 (calgizzarin)|S100 calcium-binding protein A10|calcium binding protein A11 (calgizzarin)|calpactin I light chain|calpactin-1 light chain|cellular ligand of annexin II', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93555117, end_position_on_the_genomic_accession=93564645, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20195, CurrentID=0, Status='live', Symbol='S100a11', Aliases='EMAPI, Emap1, S100a14, S100c, cal', description='S100 calcium binding protein A11', other_designations='EMAP|S100 calcium binding protein A11 (calgizzarin)|S100 calcium binding protein A11 (calizzarin)|S100 calcium binding protein A14|S100 calcium-binding protein A11|calgizzarin|calizzarin|endothelial monocyte activating polypeptide 1|endothelial monocyte-activating polypeptide|protein S100-C', map_location='3 E-F', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93520496, end_position_on_the_genomic_accession=93526288, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20196, CurrentID=0, Status='live', Symbol='S100a13', Aliases='', description='S100 calcium binding protein A13', other_designations='S100 calcium-binding protein A13', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90514435, end_position_on_the_genomic_accession=90524581, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20197, CurrentID=0, Status='live', Symbol='S100a3', Aliases='S100E', description='S100 calcium binding protein A3', other_designations='S100 calcium-binding protein A3', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90560356, end_position_on_the_genomic_accession=90602706, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20198, CurrentID=0, Status='live', Symbol='S100a4', Aliases='18A2, 42a, Capl, FSp1, Mts1, PeL98, metastasin, pk9a', description='S100 calcium binding protein A4', other_designations='S100 calcium-binding protein A4|calvasculin|metastatic cell protein|placental calcium-binding protein|protein 18A2|protein Mts1', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90603770, end_position_on_the_genomic_accession=90606045, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20199, CurrentID=0, Status='live', Symbol='S100a5', Aliases='S100D9', description='S100 calcium binding protein A5', other_designations='S100 calcium-binding protein A5', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90606479, end_position_on_the_genomic_accession=90611784, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20200, CurrentID=0, Status='live', Symbol='S100a6', Aliases='2A9, 5B10, CALCYCLIN, Cacy, PRA', description='S100 calcium binding protein A6 (calcyclin)', other_designations='S100 calcium-binding protein A6|calcium binding protein A6 (calcyclin)|prolactin receptor-associated protein', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90612894, end_position_on_the_genomic_accession=90614414, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20201, CurrentID=0, Status='live', Symbol='S100a8', Aliases='60B8Ag, AI323541, B8Ag, CFAg, CP-10, Caga, MRP8, p8', description='S100 calcium binding protein A8 (calgranulin A)', other_designations='MRP-8|S100 calcium-binding protein A8|calgranulin-A|chemotactic S100 protein|chemotactic cytokine CP-10|leukocyte L1 complex light chain|migration inhibitory factor-related protein 8|pro-inflammatory S100 cytokine', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90669071, end_position_on_the_genomic_accession=90670034, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20202, CurrentID=0, Status='live', Symbol='S100a9', Aliases='60B8Ag, AW546964, BEE22, Cagb, GAGB, L1Ag, MRP14, p14', description='S100 calcium binding protein A9 (calgranulin B)', other_designations='S100 calcium-binding protein A9|leukocyte L1 complex heavy chain|migration inhibitory factor-related protein 14', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90692630, end_position_on_the_genomic_accession=90695721, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20203, CurrentID=0, Status='live', Symbol='S100b', Aliases='AI850290, Bpb', description='S100 protein, beta polypeptide, neural', other_designations='S-100 protein beta chain|S-100 protein subunit beta|S100 calcium-binding protein B', map_location='10 38.76 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76253836, end_position_on_the_genomic_accession=76261319, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20204, CurrentID=0, Status='live', Symbol='Prrx2', Aliases='Prx2, S8', description='paired related homeobox 2', other_designations='PRX-2|homeo box of paired rule|homeobox protein S8|surface antigen, homeo box of paired rule', map_location='2 21.74 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30845064, end_position_on_the_genomic_accession=30881251, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20207, CurrentID=0, Status='live', Symbol='Saa-ps', Aliases='ENSMUSG00000061933', description='serum amyloid A, pseudogene', other_designations='Saa pseudogene', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46734688, end_position_on_the_genomic_accession=46735583, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  20208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20208, CurrentID=0, Status='live', Symbol='Saa1', Aliases='Saa-1, Saa2', description='serum amyloid A 1', other_designations='serum amyloid A 2', map_location='7 30.54 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46740497, end_position_on_the_genomic_accession=46743070, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20209, CurrentID=0, Status='live', Symbol='Saa2', Aliases='AW111173, Saa-2, Saa1', description='serum amyloid A 2', other_designations='serum amyloid A 1', map_location='7 30.56 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46751777, end_position_on_the_genomic_accession=46754314, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20210, CurrentID=0, Status='live', Symbol='Saa3', Aliases='AV098916, Saa-3, l7R3', description='serum amyloid A 3', other_designations='', map_location='7 30.51 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46711998, end_position_on_the_genomic_accession=46715676, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20211, CurrentID=0, Status='live', Symbol='Saa4', Aliases='Saa-4, Saa-5', description='serum amyloid A 4', other_designations='amyloid A-5 protein', map_location='7 30.53 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46727998, end_position_on_the_genomic_accession=46732543, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20215, CurrentID=0, Status='live', Symbol='Sag', Aliases='A930001K18Rik, Arr1, Irbp, arrestin', description='S-antigen, retina and pineal gland (arrestin)', other_designations='48 kDa protein|S-AG|arrestin 1|retinal S-antigen|rod arrestin|rod photoreceptor arrestin|visual arrestin 1', map_location='1 44.44 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87803680, end_position_on_the_genomic_accession=87845158, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  20216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20216, CurrentID=0, Status='live', Symbol='Acsm3', Aliases='Sa, Sah', description='acyl-CoA synthetase medium-chain family member 3', other_designations='SA hypertension-associated homolog|butyrate--CoA ligase 3|butyryl coenzyme A synthetase 3|butyryl-coenzyme A synthetase 3|middle-chain acyl-CoA synthetase 3|protein SA homolog', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119760883, end_position_on_the_genomic_accession=119784896, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20218, CurrentID=0, Status='live', Symbol='Khdrbs1', Aliases='Sam68, p62, p68', description='KH domain containing, RNA binding, signal transduction associated 1', other_designations='GAP-associated tyrosine phosphoprotein p62|khdrbs1|p21 Ras GTPase-activating protein-associated p62|src associated in mitosis, 68 kDa|src-associated in mitosis 68 kDa protein', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129703164, end_position_on_the_genomic_accession=129742768, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  20219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20219, CurrentID=0, Status='live', Symbol='Apcs', Aliases='Sap', description='serum amyloid P-component', other_designations='APCS amyloid P-component serum protein', map_location='1 80.33 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172893961, end_position_on_the_genomic_accession=172895054, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20220, CurrentID=0, Status='live', Symbol='Sap18', Aliases='C530046K05Rik, D11Ertd539e, EMegR4, Sinbp1', description='Sin3-associated polypeptide 18', other_designations='18 kDa Sin3-associated polypeptide|sin3-associated polypeptide p18', map_location='14 30.51 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57798189, end_position_on_the_genomic_accession=57804980, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20222, CurrentID=0, Status='live', Symbol='Sf3a2', Aliases='66kDa, AW047242, PRP11, SFA66, Sap62', description='splicing factor 3a, subunit 2', other_designations='SAP 62|SF3a66|spliceosomal protein gene 62|spliceosome-associated protein 62|splicing factor 3a, subunit 2, 66kD', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80798664, end_position_on_the_genomic_accession=80804924, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20224, CurrentID=0, Status='live', Symbol='Sar1a', Aliases='1600019H17Rik, Sara, Sara1', description='SAR1 gene homolog A (S. cerevisiae)', other_designations='SAR1a gene homolog 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61680304, end_position_on_the_genomic_accession=61693297, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20226, CurrentID=0, Status='live', Symbol='Sars', Aliases='Sars1, Strs, serRS', description='seryl-aminoacyl-tRNA synthetase', other_designations='seryl-aminoacyl-tRNA synthetase 1|seryl-tRNA synthetase, cytoplasmic|seryl-tRNA(Ser/Sec) synthetase', map_location='3 47.08 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108424864, end_position_on_the_genomic_accession=108445259, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20227, CurrentID=0, Status='live', Symbol='Sart1', Aliases='U5-110K', description='squamous cell carcinoma antigen recognized by T cells 1', other_designations='SART-1|hypoxia-associated factor|mSART-1|squamous cell carcinoma antigen recognized by T-cells 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5377523, end_position_on_the_genomic_accession=5389420, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  20229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20229, CurrentID=0, Status='live', Symbol='Sat1', Aliases='AA617398, SSAT, Sat', description='spermidine/spermine N1-acetyl transferase 1', other_designations='SSAT-1|polyamine N-acetyltransferase 1|putrescine acetyltransferase|spermidine/spermine N(1)-acetyltransferase 1', map_location='X F3-F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155213126, end_position_on_the_genomic_accession=155216449, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20230, CurrentID=0, Status='live', Symbol='Satb1', Aliases='2610306G12Rik, AW413156', description='special AT-rich sequence binding protein 1', other_designations='special AT-rich sequence-binding protein 1', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=51736187, end_position_on_the_genomic_accession=51833497, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20231, CurrentID=0, Status='live', Symbol='Nkx1-2', Aliases='Nkx-1.1, Nkx-1.2, Sax1', description='NK1 transcription factor related, locus 2 (Drosophila)', other_designations='homeobox protein Nkx1.2E427|homeobox protein SAX-1|spinal cord axial homeobox gene 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132596239, end_position_on_the_genomic_accession=132599637, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20234, CurrentID=0, Status='live', Symbol='Sbp', Aliases='AV082137, AV082259, p25', description='spermine binding protein', other_designations='major prostatic secretory glycoprotein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23941996, end_position_on_the_genomic_accession=23945607, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20237, CurrentID=0, Status='live', Symbol='sc', Aliases='', description='screwtail', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20238, CurrentID=0, Status='live', Symbol='Atxn1', Aliases='2900016G23Rik, Atx1, C85907, ENSMUSG00000074917, Gm10786, Sca1', description='ataxin 1', other_designations='spinocerebellar ataxia 1 homolog', map_location='13 21.98 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=45549755, end_position_on_the_genomic_accession=45965036, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  20239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20239, CurrentID=0, Status='live', Symbol='Atxn2', Aliases='9630045M23Rik, ATX2, AW544490, Sca2', description='ataxin 2', other_designations='spinocerebellar ataxia 2 (olivopontocerebellar ataxia 2, autosomal dominant, ataxin 2)|spinocerebellar ataxia 2 homolog|spinocerebellar ataxia type 2 protein homolog', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121711609, end_position_on_the_genomic_accession=121814950, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20248, CurrentID=0, Status='live', Symbol='Serpinb3a', Aliases='1110001H02Rik, 1110013A16Rik, Scca2, Serpinb4, Sqn5', description='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 3A', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 3A|squamous cell carcinoma antigen 2', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107045587, end_position_on_the_genomic_accession=107052303, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20249, CurrentID=0, Status='live', Symbol='Scd1', Aliases='AA589638, AI265570, Scd, Scd-1, ab', description='stearoyl-Coenzyme A desaturase 1', other_designations='asebia|delta(9)-desaturase 1|delta-9 desaturase 1|fatty acid desaturase 1|stearoyl-CoA desaturase 1', map_location='19 37.98 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44394450, end_position_on_the_genomic_accession=44407709, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20250, CurrentID=0, Status='live', Symbol='Scd2', Aliases='Mir5114, Scd-2, mir-5114, swty', description='stearoyl-Coenzyme A desaturase 2', other_designations='Acyl-CoA desaturase 2 (Stearoyl-CoA desaturase 2) (Fatty acid desaturase 2) (Delta(9)-desaturase 2)|delta(9)-desaturase 2|delta-9 desaturase 2|fatty acid desaturase 2|stearoyl-CoA desaturase 2', map_location='19 37.98 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44293676, end_position_on_the_genomic_accession=44306864, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20254, CurrentID=0, Status='live', Symbol='Scg2', Aliases='Chgc, SgII', description='secretogranin II', other_designations='chromogranin-C', map_location='1 40.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=79434669, end_position_on_the_genomic_accession=79440121, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20255, CurrentID=0, Status='live', Symbol='Scg3', Aliases='1B1075, AI385542, Chgd, SgIII', description='secretogranin III', other_designations='', map_location='9 42.32 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75643366, end_position_on_the_genomic_accession=75684082, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20256, CurrentID=0, Status='live', Symbol='Clec11a', Aliases='AW457320, Clecsf3, Scgf', description='C-type lectin domain family 11, member a', other_designations='lymphocyte secreted C-type lectin|stem cell growth factor', map_location='7 B3-B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44303766, end_position_on_the_genomic_accession=44306959, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20257, CurrentID=0, Status='live', Symbol='Stmn2', Aliases='AI159727, SCG10, Scgn10, Stmb2', description='stathmin-like 2', other_designations='superior cervical ganglion-10 protein|superiorcervical ganglia, neural specific 10', map_location='3 2.15 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=8509527, end_position_on_the_genomic_accession=8561604, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20259, CurrentID=0, Status='live', Symbol='Scin', Aliases='AW545522, adseverin', description='scinderin', other_designations='gelsolin-like protein', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=40059769, end_position_on_the_genomic_accession=40134228, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  20262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20262, CurrentID=0, Status='live', Symbol='Stmn3', Aliases='9330161A03Rik, AI604839, R75323, Sclip', description='stathmin-like 3', other_designations='SCG10-like protein|SCG10-related protein HiAT3|Scgn10 like-protein|hippocampus abundant transcript 3', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181306459, end_position_on_the_genomic_accession=181314500, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20264, CurrentID=0, Status='live', Symbol='Scn10a', Aliases='Nav1.8, PN3, SNS', description='sodium channel, voltage-gated, type X, alpha', other_designations='peripheral nerve sodium channel 3|sensory neuron sodium channel|sodium channel protein type X subunit alpha|voltage gated sodium channel|voltage-gated sodium channel Nav1.8 variant a|voltage-gated sodium channel Nav1.8 variant b|voltage-gated sodium channel subunit alpha Nav1.8', map_location='9 71.33 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119608456, end_position_on_the_genomic_accession=119719032, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  20265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20265, CurrentID=0, Status='live', Symbol='Scn1a', Aliases='B230332M13, Nav1.1', description='sodium channel, voltage-gated, type I, alpha', other_designations='', map_location='2 39.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=66270781, end_position_on_the_genomic_accession=66440847, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  20266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20266, CurrentID=0, Status='live', Symbol='Scn1b', Aliases='', description='sodium channel, voltage-gated, type I, beta', other_designations='sodium channel, voltage-gated, type I, beta polypeptide', map_location='7 19.3 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31116524, end_position_on_the_genomic_accession=31126945, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20269, CurrentID=0, Status='live', Symbol='Scn3a', Aliases='Gm1000, Nav1.3', description='sodium channel, voltage-gated, type III, alpha', other_designations='sodium channel, voltage-gated, type III, alpha polypeptide|voltage-gated sodium channel Nav1.3 variant a|voltage-gated sodium channel Nav1.3 variant b', map_location='2 38.55 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=65457118, end_position_on_the_genomic_accession=65567527, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  20271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20271, CurrentID=0, Status='live', Symbol='Scn5a', Aliases='Nav1.5, Nav1.5c, SkM1, SkM2, mH1', description='sodium channel, voltage-gated, type V, alpha', other_designations='sodium channel protein cardiac muscle subunit alpha|sodium channel protein type V subunit alpha|sodium channel voltage-gated type V alpha polypeptide|sodium channel, voltage-gated, type V, alpha polypeptide|voltage-gated sodium channel Nav1.5 variant a|voltage-gated sodium channel Nav1.5 variant b|voltage-gated sodium channel subunit alpha Nav1.5', map_location='9 71.33 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119483404, end_position_on_the_genomic_accession=119579030, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  20272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20272, CurrentID=0, Status='live', Symbol='Scn7a', Aliases='1110034K09Rik, AI642000, NaG, Nav2, Nav2.3, Nax, Scn6a', description='sodium channel, voltage-gated, type VII, alpha', other_designations='sodium channel, voltage-gated, type VI, alpha polypeptide', map_location='2 39.21 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=66673424, end_position_on_the_genomic_accession=66785091, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  20273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20273, CurrentID=0, Status='live', Symbol='Scn8a', Aliases='AI853486, C630029C19Rik, NaCh6, Nav1.6, dmu, med, mnd-2, mnd2, nmf2, nmf335, nmf58, nur14, seal', description='sodium channel, voltage-gated, type VIII, alpha', other_designations='ataxia 3|sodium channel protein type VIII subunit alpha|voltage-gated sodium channel Nav1.6 variant a|voltage-gated sodium channel Nav1.6 variant b|voltage-gated sodium channel alpha subunit|voltage-gated sodium channel subunit alpha Nav1.6', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100868549, end_position_on_the_genomic_accession=101045938, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  20274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20274, CurrentID=0, Status='live', Symbol='Scn9a', Aliases='Nav1.7, PN1, mKIAA4197', description='sodium channel, voltage-gated, type IX, alpha', other_designations='peripheral sodium channel 1|sodium channel 25|sodium channel, voltage-gated, type IX, alpha polypeptide|voltage-gated sodium channel Nav1.7 variant a|voltage-gated sodium channel Nav1.7 variant b|voltage-gated sodium channel alpha subunit Nav1.7', map_location='2 39.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=66480080, end_position_on_the_genomic_accession=66635345, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  20276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20276, CurrentID=0, Status='live', Symbol='Scnn1a', Aliases='ENaC, SCNEA, Scnn1, mENaC', description='sodium channel, nonvoltage-gated 1 alpha', other_designations='ENaC alpha|alpha-ENaC|alpha-NaCH|amiloride-sensitive epithelial sodium channel|epithelial Na(+) channel subunit alpha|epithelial sodium channel alpha subunit|nonvoltage-gated sodium channel 1 subunit alpha|sodium channel, nonvoltage-gated, type I, alpha polypeptide', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125321340, end_position_on_the_genomic_accession=125344943, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20277, CurrentID=0, Status='live', Symbol='Scnn1b', Aliases='', description='sodium channel, nonvoltage-gated 1 beta', other_designations='ENaC beta|SCNEB|beta-ENaC|beta-NaCH|epithelial Na(+) channel subunit beta|nonvoltage-gated sodium channel 1 subunit beta', map_location='7 65.07 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121865038, end_position_on_the_genomic_accession=121918728, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  20278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20278, CurrentID=0, Status='live', Symbol='Scnn1g', Aliases='SCNEG', description='sodium channel, nonvoltage-gated 1 gamma', other_designations='ENaC gamma|epithelial Na(+) channel subunit gamma|gamma-ENaC|gamma-NaCH|nonvoltage-gated sodium channel 1 subunit gamma', map_location='7 65.07 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121734420, end_position_on_the_genomic_accession=121768477, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  20280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20280, CurrentID=0, Status='live', Symbol='Scp2', Aliases='AA409774, AA409893, C76618, C79031, NSL-TP, SCP-2, SCP-X, SCPX, ns-LTP', description='sterol carrier protein 2, liver', other_designations='SCP-chi|nonspecific lipid transfer protein|propanoyl-CoA C-acyltransferase|sterol carrier protein X', map_location='4 50.2 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108043830, end_position_on_the_genomic_accession=108144971, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  20284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20284, CurrentID=0, Status='live', Symbol='Scrg1', Aliases='AW124307', description='scrapie responsive gene 1', other_designations='scRG-1|scrapie responsive protein one|scrapie-responsive gene 1 protein', map_location='8 29.8 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=57455923, end_position_on_the_genomic_accession=57477585, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20286, CurrentID=0, Status='live', Symbol='Zc3h7b', Aliases='Gm627, Scrg3, Zc2h7b, mKIAA1031', description='zinc finger CCCH type containing 7B', other_designations='scrapie responsive gene 3|zinc finger CCCH-type containing 7B', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81744848, end_position_on_the_genomic_accession=81796269, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  20287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20287, CurrentID=0, Status='live', Symbol='Sct', Aliases='', description='secretin', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141278329, end_position_on_the_genomic_accession=141279133, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20288, CurrentID=0, Status='live', Symbol='Msr1', Aliases='MRS-A, MSR, MSR-A, SR-AI, SR-AII, Scara1, Scvr', description='macrophage scavenger receptor 1', other_designations='SR-A|macrophage acetylated LDL receptor I and II|scavenger receptor type A', map_location='8 23.89 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=39543232, end_position_on_the_genomic_accession=39642751, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  20289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20289, CurrentID=0, Status='live', Symbol='Scx', Aliases='BB114693, Bhlha41, Scl', description='scleraxis', other_designations='', map_location='15 35.92 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76456033, end_position_on_the_genomic_accession=76459468, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20290, CurrentID=0, Status='live', Symbol='Ccl1', Aliases='BF534335, I-309, Scya1, Tca-3', description='chemokine (C-C motif) ligand 1', other_designations='CCR8 ligand|P500|SIS-epsilon|T-cell activation protein 3|TCA3|small inducible cytokine A1|small-inducible cytokine A1', map_location='11 49.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82176666, end_position_on_the_genomic_accession=82179884, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20292, CurrentID=0, Status='live', Symbol='Ccl11', Aliases='Scya11, eotaxin', description='chemokine (C-C motif) ligand 11', other_designations='C-C motif chemokine 11|eosinophil chemotactic protein|small chemokine (C-C motif) ligand 11|small inducible cytokine A11|small-inducible cytokine A11', map_location='11 49.84 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82057832, end_position_on_the_genomic_accession=82062955, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20293, CurrentID=0, Status='live', Symbol='Ccl12', Aliases='MCP-5, Scya12', description='chemokine (C-C motif) ligand 12', other_designations='MCP-1-related chemokine|monocyte chemoattractant protein 5|monocyte chemotactic protein 5|small inducible cytokine A12|small-inducible cytokine A12', map_location='11 49.9 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82101845, end_position_on_the_genomic_accession=82103399, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20295, CurrentID=0, Status='live', Symbol='Ccl17', Aliases='Abcd-2, Scya17, Scya17l, Tarc', description='chemokine (C-C motif) ligand 17', other_designations='small inducible cytokine subfamily A17', map_location='8 46.85 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94807404, end_position_on_the_genomic_accession=94812038, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20296, CurrentID=0, Status='live', Symbol='Ccl2', Aliases='AI323594, HC11, JE, MCAF, MCP-1, MCP1, SMC-CF, Scya2, Sigje', description='chemokine (C-C motif) ligand 2', other_designations='monocyte chemoattractant protein 1|monocyte chemoattractant protein-1|monocyte chemotactic protein 1|platelet-derived growth factor-inducible protein JE|small inducible cytokine A2|small-inducible cytokine A2', map_location='11 C-E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82035577, end_position_on_the_genomic_accession=82037452, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20297, CurrentID=0, Status='live', Symbol='Ccl20', Aliases='CKb4, LARC, MIP-3A, MIP-3[a], MIP3A, ST38, Scya20, exodus-1', description='chemokine (C-C motif) ligand 20', other_designations='CC chemokine LARC|CC chemokine ST38|MIP-3-alpha|beta chemokine exodus-1|beta-chemokine exodus-1|liver and activation-regulated chemokine|macrophage inflammatory protein 3 alpha|small inducible cytokine subfamily A20|small-inducible cytokine A20', map_location='1 42.69 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=83116744, end_position_on_the_genomic_accession=83119167, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20299, CurrentID=0, Status='live', Symbol='Ccl22', Aliases='ABCD-1, DCBCK, MDC, Scya22', description='chemokine (C-C motif) ligand 22', other_designations='CC chemokine ABCD-1|SMALL INDUCIBLE CYTOKINE A22 PRECURSOR (CC CHEMOKINE ABCD-1) (ACTIVATED B AND DENDRITIC CELL-DERIVED)|activated B and dendritic cell-derived|dendritic cell and B cell derived chemokine|small inducible cytokine subfamily A, member 22|small inducible cytokine subfamily A22|small-inducible cytokine A22', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94745684, end_position_on_the_genomic_accession=94751388, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20300, CurrentID=0, Status='live', Symbol='Ccl25', Aliases='A130072A22Rik, AI852536, CKb15, Scya25, TECK', description='chemokine (C-C motif) ligand 25', other_designations='small inducible chemokine 25|small inducible cytokine A25|thymus-expressed chemokine', map_location='8 2.01 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4325159, end_position_on_the_genomic_accession=4360020, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20301, CurrentID=0, Status='live', Symbol='Ccl27a', Aliases='ALP, AW558992, CTACK, CTAK, Ccl27, ESkine, ILC, PESKY, Scya27, Scya27a', description='chemokine (C-C motif) ligand 27A', other_designations='CC chemokine ILC|IL-11 R-alpha-locus chemokine|chemokine (C-C motif) ligand 27|cutaneous T-cell-attracting chemokine|skinkine|skinskine|small inducible cytokine A27|small inducible cytokine A27a|small-inducible cytokine A27', map_location='4 22.08 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41769467, end_position_on_the_genomic_accession=41774210, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20302, CurrentID=0, Status='live', Symbol='Ccl3', Aliases='AI323804, G0S19-1, LD78alpha, MIP-1alpha, MIP1-(a), MIP1-alpha, Mip1a, Scya3', description='chemokine (C-C motif) ligand 3', other_designations='L2G25B|MIP-1 alpha|MIP-1-alpha|MIP1 (a)|SIS-alpha|TY-5|heparin-binding chemotaxis protein|macrophage inflammatory protein 1-alpha|macrophage inflammatory protein-1alpha|small inducible cytokine A3|small-inducible cytokine A3', map_location='11 51.04 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83647843, end_position_on_the_genomic_accession=83649378, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20303, CurrentID=0, Status='live', Symbol='Ccl4', Aliases='AT744.1, Act-2, MIP-1B, Mip1b, Scya4', description='chemokine (C-C motif) ligand 4', other_designations='ACT2|MIP-1 beta|MIP-1-beta|SIS-gamma|immune activation protein 2|macrophage inflammatory protein 1-beta|small inducible cytokine A4|small-inducible cytokine A4', map_location='11 51.09 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83648624, end_position_on_the_genomic_accession=83664683, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20304, CurrentID=0, Status='live', Symbol='Ccl5', Aliases='MuRantes, RANTES, SISd, Scya5, TCP228', description='chemokine (C-C motif) ligand 5', other_designations='SIS-delta|T-cell-specific protein RANTES|small inducible cytokine A5|small-inducible cytokine A5', map_location='11 50.66 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83525778, end_position_on_the_genomic_accession=83530518, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20305, CurrentID=0, Status='live', Symbol='Ccl6', Aliases='MRP-1, Scya6, c10', description='chemokine (C-C motif) ligand 6', other_designations='CC chemokine C10|small inducible cytokine A6|small-inducible cytokine A6', map_location='11 50.85 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83587886, end_position_on_the_genomic_accession=83593087, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20306, CurrentID=0, Status='live', Symbol='Ccl7', Aliases='MCP-3, Scya7, fic, marc, mcp3', description='chemokine (C-C motif) ligand 7', other_designations='RANTES/sis homolog|intercrine/chemokine MARC|monocyte chemoattractant protein 3|monocyte chemotactic protein 3|small inducible cytokine A7|small-inducible cytokine A7', map_location='11 49.83 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82045712, end_position_on_the_genomic_accession=82047525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20307, CurrentID=0, Status='live', Symbol='Ccl8', Aliases='1810063B20Rik, AB023418, HC14, MCP-2, Mcp2, Scya8', description='chemokine (C-C motif) ligand 8', other_designations='monocyte chemoattractant protein 2|monocyte chemoattractant protein-2|monocyte chemotactic protein 2|small inducible cytokine A8|small-inducible cytokine A8', map_location='11 49.91 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82115185, end_position_on_the_genomic_accession=82116799, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20308, CurrentID=0, Status='live', Symbol='Ccl9', Aliases='CCF18, MRP-2, Scya10, Scya9', description='chemokine (C-C motif) ligand 9', other_designations='MIP-1 gamma|macrophage inflammatory protein 1-gamma|macrophage inflammatory protein-related protein 2|small inducible cytokine A10|small inducible cytokine A9', map_location='11 50.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83572917, end_position_on_the_genomic_accession=83578636, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20309, CurrentID=0, Status='live', Symbol='Cxcl15', Aliases='Il8, Scyb15, lungkine, weche', description='chemokine (C-X-C motif) ligand 15', other_designations='interleukin 8|small inducible cytokine subfamily B, member 15|small-inducible cytokine B15', map_location='5 44.78 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90794534, end_position_on_the_genomic_accession=90803067, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20310, CurrentID=0, Status='live', Symbol='Cxcl2', Aliases='CINC-2a, GROb, Gro2, MIP-2, MIP-2a, Mgsa-b, Mip2, Scyb, Scyb2', description='chemokine (C-X-C motif) ligand 2', other_designations='macrophage inflammatory protein 2|small inducible cytokine subfamily, member 2', map_location='5 44.78 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90903899, end_position_on_the_genomic_accession=90905938, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20311, CurrentID=0, Status='live', Symbol='Cxcl5', Aliases='AMCF-II, ENA-78, GCP-2, LIX, Scyb5, Scyb6', description='chemokine (C-X-C motif) ligand 5', other_designations='cytokine LIX|granulocyte chemotactic protein-2|small inducible cytokine B subfamily, member 5|small inducible cytokine subfamily B, member 15|small-inducible cytokine B5', map_location='5 44.78 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90759298, end_position_on_the_genomic_accession=90761625, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20312, CurrentID=0, Status='live', Symbol='Cx3cl1', Aliases='AB030188, ABCD-3, AI848747, CX3C, Cxc3, D8Bwg0439e, Scyd1', description='chemokine (C-X3-C motif) ligand 1', other_designations='C-X3-C motif chemokine 1|CX3C membrane-anchored chemokine|neurotactin|small inducible cytokine subfamily D, 1|small-inducible cytokine D1', map_location='8 46.79 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94772180, end_position_on_the_genomic_accession=94782427, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20315, CurrentID=0, Status='live', Symbol='Cxcl12', Aliases='Pbsf, Scyb12, Sdf1, Tlsf, Tpar1', description='chemokine (C-X-C motif) ligand 12', other_designations='12-O-tetradecanoylphorbol 13-acetate repressed protein 1|pre-B-cell growth-stimulating factor|thymic lymphoma cell-stimulating factor', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=117168535, end_position_on_the_genomic_accession=117181368, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20316, CurrentID=0, Status='live', Symbol='Sdf2', Aliases='AI853825', description='stromal cell derived factor 2', other_designations='SDF-2', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78245746, end_position_on_the_genomic_accession=78255496, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20317, CurrentID=0, Status='live', Symbol='Serpinf1', Aliases='AI195227, EPC-1, Pedf, Pedfl, Sdf3', description='serine (or cysteine) peptidase inhibitor, clade F, member 1', other_designations='SDF-3|alpha-2 antiplasmin|caspin|pigment epithelium derived factor|serine (or cysteine) proteinase inhibitor, clade F (alpha-2 antiplasmin, pigment epithelium derived factor), member 1|serine (or cysteine) proteinase inhibitor, clade F), member 1|serine (or cysteine) proteinase inhibitor, clade F, member 1|serpin F1|serpin-F1|stromal cell derived factor 3|stromal cell-derived factor 3', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75410029, end_position_on_the_genomic_accession=75422623, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20318, CurrentID=0, Status='live', Symbol='Sdf4', Aliases='Cab45', description='stromal cell derived factor 4', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155992843, end_position_on_the_genomic_accession=156013610, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20319, CurrentID=0, Status='live', Symbol='Sfrp2', Aliases='AI851596, Sdf5', description='secreted frizzled-related protein 2', other_designations='SARP-1|sFRP-2|secreted apoptosis-related protein 1|secreted frizzled-related sequence protein 2|secreted frizzled-related sequence protein 5|stromal cell derived factor 5', map_location='3 37.37 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=83766148, end_position_on_the_genomic_accession=83774314, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20320, CurrentID=0, Status='live', Symbol='Nptn', Aliases='AW554172, SDR-1, Sdfr1', description='neuroplastin', other_designations='stromal cell derived factor receptor 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58582240, end_position_on_the_genomic_accession=58652879, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20321, CurrentID=0, Status='live', Symbol='Frrs1', Aliases='AI131732, Sdfr2', description='ferric-chelate reductase 1', other_designations='SDR-2|stromal cell derived factor receptor 2|stromal cell-derived receptor 2', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116859567, end_position_on_the_genomic_accession=116904901, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  20322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20322, CurrentID=0, Status='live', Symbol='Sord', Aliases='Sdh-1, Sdh1, Sodh-1', description='sorbitol dehydrogenase', other_designations='L-iditol 2-dehydrogenase|sorbitol dehydrogenase 1', map_location='2 60.59 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122234839, end_position_on_the_genomic_accession=122265337, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20324, CurrentID=0, Status='live', Symbol='Sdpr', Aliases='', description='serum deprivation response', other_designations='cavin 2|cavin-2|phosphatidylserine-binding protein|sdpr', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=51289126, end_position_on_the_genomic_accession=51302960, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20331, CurrentID=0, Status='live', Symbol='seb', Aliases='', description='seborrheic dermatitis', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20333, CurrentID=0, Status='live', Symbol='Sec22b', Aliases='4930564D15Rik, AA517334, AI480645, C81333, ERS-24, Sec22l1', description='SEC22 homolog B, vesicle trafficking protein', other_designations='ER-Golgi SNARE of 24 kDa|ERS24|SEC22 vesicle trafficking protein homolog B|SEC22 vesicle trafficking protein-like 1|SEC22 vesicle-trafficking protein homolog B|SEC22 vesicle-trafficking protein-like 1|mSec22b', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97901227, end_position_on_the_genomic_accession=97922318, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20334, CurrentID=0, Status='live', Symbol='Sec23a', Aliases='Msec23, Sec23r', description='SEC23 homolog A, COPII coat complex component', other_designations='SEC23-related protein A|SEC23A', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=58958383, end_position_on_the_genomic_accession=59012026, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  20335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20335, CurrentID=0, Status='live', Symbol='Sec61g', Aliases='', description='SEC61, gamma subunit', other_designations='', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=16501638, end_position_on_the_genomic_accession=16508484, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20336, CurrentID=0, Status='live', Symbol='Exoc4', Aliases='C78892, Sec8, Sec8l1', description='exocyst complex component 4', other_designations='SEC8-like 1|exocyst complex component Sec8', map_location='6 14.59 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=33249090, end_position_on_the_genomic_accession=33972943, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  20338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20338, CurrentID=0, Status='live', Symbol='Sel1l', Aliases='AW493766, Sel1h, mKIAA4137', description='sel-1 suppressor of lin-12-like (C. elegans)', other_designations='Sel1 (suppressor of lin-12) 1 homolog|sel-1L|suppressor of lin-12-like protein 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=91806043, end_position_on_the_genomic_accession=91849157, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  20339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20339, CurrentID=0, Status='live', Symbol='Sele', Aliases='CD62E, E-selectin, ELAM-1, Elam, LECAM2', description='selectin, endothelial cell', other_designations='CD62 antigen-like family member E|endothelial leukocyte adhesion molecule 1|leukocyte-endothelial cell adhesion molecule 2', map_location='1 71.35 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164039604, end_position_on_the_genomic_accession=164058487, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20340, CurrentID=0, Status='live', Symbol='Glg1', Aliases='AI593353, AW537898, CFR, CFR-1, ESL-1, MG-160, MG160, Selel', description='golgi apparatus protein 1', other_designations='E-selectin ligand 1|golgi sialoglycoprotein MG-160|selectin, endothelial cell, ligand', map_location='8 57.98 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111157548, end_position_on_the_genomic_accession=111259202, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  20341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20341, CurrentID=0, Status='live', Symbol='Selenbp1', Aliases='Lp56, Lpsb, SBP56', description='selenium binding protein 1', other_designations='56 kDa selenium-binding protein|SP56', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94933083, end_position_on_the_genomic_accession=94944758, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20342, CurrentID=0, Status='live', Symbol='Selenbp2', Aliases='AP56, Lpsb2', description='selenium binding protein 2', other_designations='56 kDa acetaminophen-binding protein|acetaminophen-binding protein', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94693573, end_position_on_the_genomic_accession=94704406, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20343, CurrentID=0, Status='live', Symbol='Sell', Aliases='AI528707, CD62L, L-selectin, LECAM-1, Lnhr, Ly-22, Ly-m22, Lyam-1, Lyam1', description='selectin, lymphocyte', other_designations='CD62 antigen-like family member L|LAM-1|LECAM1|leukocyte adhesion molecule 1|leukocyte-endothelial cell adhesion molecule 1|lymph node homing receptor|lymphocyte antigen 22|lymphocyte surface MEL-14 antigen', map_location='1 71.37 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164061813, end_position_on_the_genomic_accession=164080785, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20344, CurrentID=0, Status='live', Symbol='Selp', Aliases='CD62P, GMP-140, Grmp, LECAM3, PADGEM', description='selectin, platelet', other_designations='CD62 antigen-like family member P|granule membrane protein 140|leukocyte-endothelial cell adhesion molecule 3|platelet activation dependent granule-external membrane protein', map_location='1 71.42 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164115264, end_position_on_the_genomic_accession=164150026, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  20345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20345, CurrentID=0, Status='live', Symbol='Selplg', Aliases='CD162, Psgl-1, Psgl1, Selp1, Selpl', description='selectin, platelet (p-selectin) ligand', other_designations='P-selectin glycoprotein ligand-1|selectin P ligand', map_location='5 55.59 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113817798, end_position_on_the_genomic_accession=113832638, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20346, CurrentID=0, Status='live', Symbol='Sema3a', Aliases='Hsema-I, SEMA1, SemD, Semad, coll-1', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3A', other_designations='collapsin-1|sema D|sema III|semaphorin III|semaphorin-D', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=13125371, end_position_on_the_genomic_accession=13606234, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  20347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20347, CurrentID=0, Status='live', Symbol='Sema3b', Aliases='SemA, Semaa, sema5, semaV', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3B', other_designations='sema A|semaphorin-A', map_location='9 58.31 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107597674, end_position_on_the_genomic_accession=107609303, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  20348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20348, CurrentID=0, Status='live', Symbol='Sema3c', Aliases='1110036B02Rik, SemE, Semae', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3C', other_designations='sema E|semaphorin E|semaphorin-E', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=17574816, end_position_on_the_genomic_accession=17730268, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  20349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20349, CurrentID=0, Status='live', Symbol='Sema3e', Aliases='6430702L12, AA408817, Semah, mKIAA0331', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3E', other_designations='sema H|semaphorin-H', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=14025276, end_position_on_the_genomic_accession=14256689, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  20350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20350, CurrentID=0, Status='live', Symbol='Sema3f', Aliases='Sema4, Semak', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3F', other_designations='semaphorin IV', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107681499, end_position_on_the_genomic_accession=107710475, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  20351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20351, CurrentID=0, Status='live', Symbol='Sema4a', Aliases='AI132332, SemB, Semab', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4A', other_designations='sema B|semaphorin 4A|semaphorin-B', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88435959, end_position_on_the_genomic_accession=88461182, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  20352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20352, CurrentID=0, Status='live', Symbol='Sema4b', Aliases='SemC, Semac, mKIAA1745', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4B', other_designations='sema C|semaphorin-C', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80186841, end_position_on_the_genomic_accession=80226527, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20353, CurrentID=0, Status='live', Symbol='Sema4c', Aliases='AI426163, M-Sema F, Semacl1, Semaf, Semai, sema I', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4C', other_designations='semaphorin I|semaphorin-C-like 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36548640, end_position_on_the_genomic_accession=36558381, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  20354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20354, CurrentID=0, Status='live', Symbol='Sema4d', Aliases='CD100, Semacl2, Semaj, Semcl2, coll-4', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4D', other_designations='M-sema G|sema J|semaphorin H|semaphorin-C-like 2|semaphorin-J', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51685522, end_position_on_the_genomic_accession=51793747, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  20355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20355, CurrentID=0, Status='live', Symbol='Sema4f', Aliases='Sema W', description='sema domain, immunoglobulin domain (Ig), TM domain, and short cytoplasmic domain', other_designations='semaphorin-W', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=82911884, end_position_on_the_genomic_accession=82939761, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20356, CurrentID=0, Status='live', Symbol='Sema5a', Aliases='5930434A13, 9130201M22Rik, AI464145, Semaf, semF', description='sema domain, seven thrombospondin repeats (type 1 and type 1-like), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 5A', other_designations='M-Sema D|sema F|semaphorin F|semaphorin-F', map_location='15 13.02 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=32244742, end_position_on_the_genomic_accession=32696341, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  20357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20357, CurrentID=0, Status='live', Symbol='Sema5b', Aliases='AI893641, SemG, Semag, mKIAA1445', description='sema domain, seven thrombospondin repeats (type 1 and type 1-like), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 5B', other_designations='sema G|semaphorin 5B|semaphorin-G', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35541154, end_position_on_the_genomic_accession=35664261, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  20358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20358, CurrentID=0, Status='live', Symbol='Sema6a', Aliases='A730020P05Rik, AI851735, Sema6A-1, Semaq, VIa', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6A', other_designations='Mutant line 997|sema Q|sema VIA|semaphorin Q|semaphorin VIA|semaphorin-6A-1', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=47245254, end_position_on_the_genomic_accession=47371487, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  20359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20359, CurrentID=0, Status='live', Symbol='Sema6b', Aliases='Sema, Seman, VIb, semaZ', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6B', other_designations='sema N|sema VIB|semaphorin VIB|semaphorin-N', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56123085, end_position_on_the_genomic_accession=56140343, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  20360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20360, CurrentID=0, Status='live', Symbol='Sema6c', Aliases='Semay, mKIAA1869', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6C', other_designations='Sema Y|m-Sema Y|semaphorin-Y', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95160420, end_position_on_the_genomic_accession=95174050, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  20361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20361, CurrentID=0, Status='live', Symbol='Sema7a', Aliases='2900057C09Rik, CDw108, H-Sema-L, M-Sema-L, Semal', description='sema domain, immunoglobulin domain (Ig), and GPI membrane anchor, (semaphorin) 7A', other_designations='H-Sema K1|Semaphorin K1|sema K1|sema L|semaphorin-K1|semaphorin-L', map_location='9 A3.3-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57940135, end_position_on_the_genomic_accession=57962865, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20362, CurrentID=0, Status='live', Symbol='Sept8', Aliases='AW046166, Sepl, mKIAA0202', description='septin 8', other_designations='', map_location='11 31.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53519736, end_position_on_the_genomic_accession=53549600, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20363, CurrentID=0, Status='live', Symbol='Sepp1', Aliases='AU018766, D15Ucla1, Se-P, selp', description='selenoprotein P, plasma, 1', other_designations='plasma selenoprotein P|seP', map_location='15 1.84 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=3270767, end_position_on_the_genomic_accession=3280508, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20364, CurrentID=0, Status='live', Symbol='Sepw1', Aliases='selW', description='selenoprotein W, muscle 1', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15917208, end_position_on_the_genomic_accession=15922371, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20365, CurrentID=0, Status='live', Symbol='Serf1', Aliases='4F5, Msmac1, m4F5', description='small EDRK-rich factor 1', other_designations='modifier of spinal muscular atrophy candidate 1|protein 4F5', map_location='13 52.98 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100108019, end_position_on_the_genomic_accession=100114233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20370, CurrentID=0, Status='live', Symbol='Sez6', Aliases='BSRP-C, D11Bhm177e, sez-6', description='seizure related gene 6', other_designations='brain-specific receptor-like protein C', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77930839, end_position_on_the_genomic_accession=77979052, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  20371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20371, CurrentID=0, Status='live', Symbol='Foxp3', Aliases='JM2, scurfin, sf', description='forkhead box P3', other_designations='scurfy', map_location='X 3.41 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7579676, end_position_on_the_genomic_accession=7595243, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20375, CurrentID=0, Status='live', Symbol='Spi1', Aliases='Dis-1, Dis1, PU.1, Sfpi-1, Sfpi1, Spi-1, Tcfpu1, Tfpu.1', description='spleen focus forming virus (SFFV) proviral integration oncogene', other_designations='31 kDa-transforming protein|SFFV proviral integration 1 protein|hematopoietic transcription factor PU.1', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91082381, end_position_on_the_genomic_accession=91115757, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20377, CurrentID=0, Status='live', Symbol='Sfrp1', Aliases='2210415K03Rik, AW011917, AW107218, AW742929, sFRP-1', description='secreted frizzled-related protein 1', other_designations='secreted frizzled-related sequence protein 1', map_location='8 11.48 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=23411382, end_position_on_the_genomic_accession=23449632, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20378, CurrentID=0, Status='live', Symbol='Frzb', Aliases='Frp, Sfrp3, frezzled, fritz, frzb-1', description='frizzled-related protein', other_designations='frizzled-related protein 1|sFRP-3|secreted frizzled-related sequence protein 3', map_location='2 48.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=80411970, end_position_on_the_genomic_accession=80447396, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20379, CurrentID=0, Status='live', Symbol='Sfrp4', Aliases='', description='secreted frizzled-related protein 4', other_designations='FRP-4|frizzled-related protein 4|sFRP-4', map_location='13 6.99 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19623175, end_position_on_the_genomic_accession=19632825, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20382, CurrentID=0, Status='live', Symbol='Srsf2', Aliases='D11Wsu175e, MRF-1, Pr264, SC35, Sfrs10, Sfrs2', description='serine/arginine-rich splicing factor 2', other_designations='myelin regulatory factor 1|putative myelin regulatory factor 1|splicing component, 35 kDa|splicing factor Sc35|splicing factor, arginine/serine-rich 2 (SC-35)', map_location='11 81.49 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116849897, end_position_on_the_genomic_accession=116853104, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20383, CurrentID=0, Status='live', Symbol='Srsf3', Aliases='AL024116, Sfrs3, X16', description='serine/arginine-rich splicing factor 3', other_designations='pre-mRNA-splicing factor SRP20|splicing factor, arginine/serine-rich 3 (SRp20)', map_location='17 15.03 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29032660, end_position_on_the_genomic_accession=29043372, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20384, CurrentID=0, Status='live', Symbol='Srsf5', Aliases='Sfrs5', description='serine/arginine-rich splicing factor 5', other_designations='delayed-early protein HRS|pre-mRNA-splicing factor SRP40|splicing factor, arginine/serine-rich 5 (SRp40, HRS)', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80945504, end_position_on_the_genomic_accession=80950507, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20387, CurrentID=0, Status='live', Symbol='Sftpa1', Aliases='SP-A, Sftp-1, Sftp1', description='surfactant associated protein A1', other_designations='PSAP|PSP-A|surfactant pulmonary associated protein A1', map_location='14 22.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=41131787, end_position_on_the_genomic_accession=41136373, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20388, CurrentID=0, Status='live', Symbol='Sftpb', Aliases='AI562151, SF-B, SP-B, Sftp-3, Sftp3', description='surfactant associated protein B', other_designations='pulmonary surfactant-associated proteolipid SPL(Phe)', map_location='6 32.27 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72304610, end_position_on_the_genomic_accession=72314373, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20389, CurrentID=0, Status='live', Symbol='Sftpc', Aliases='Bricd6, SP-C, SP5, SPC, Sftp-2, Sftp2, pro-SpC', description='surfactant associated protein C', other_designations='pulmonary surfactant-associated proteolipid SPL(Val)', map_location='14 36.32 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70520941, end_position_on_the_genomic_accession=70524081, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  544678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544678, CurrentID=0, Status='live', Symbol='Cfap74', Aliases='2010015L04Rik, A530045M11', description='cilia and flagella associated protein 74', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155398208, end_position_on_the_genomic_accession=155466823, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  20391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20391, CurrentID=0, Status='live', Symbol='Sgca', Aliases='50DAG, Asg', description='sarcoglycan, alpha (dystrophin-associated glycoprotein)', other_designations='50 kDa dystrophin-associated glycoprotein|adhalin|alpha-SG', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94962777, end_position_on_the_genomic_accession=94976327, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  20392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20392, CurrentID=0, Status='live', Symbol='Sgce', Aliases='e-SG', description='sarcoglycan, epsilon', other_designations='epsilon-SG', map_location='6 1.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=4674350, end_position_on_the_genomic_accession=4747204, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20393, CurrentID=0, Status='live', Symbol='Sgk1', Aliases='Sgk', description='serum/glucocorticoid regulated kinase 1', other_designations='serum and glucocorticoid-dependent protein kinase|serum/glucocorticoid-regulated kinase 1', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=21882184, end_position_on_the_genomic_accession=21999903, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  20394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20394, CurrentID=0, Status='live', Symbol='Scg5', Aliases='7B2, AI325031, Sgne-1, Sgne1', description='secretogranin V', other_designations='secretogranin-5|secretory granule endocrine protein I|secretory granule neuroendocrine protein 1, 7B2 protein', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=113776313, end_position_on_the_genomic_accession=113829091, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20397, CurrentID=0, Status='live', Symbol='Sgpl1', Aliases='AI428538, D10Xrf456, S1PL, Spl', description='sphingosine phosphate lyase 1', other_designations='SP-lyase 1|SPL 1|sphingosine-1-phosphate aldolase', map_location='10 32.14 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61098642, end_position_on_the_genomic_accession=61147686, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20399, CurrentID=0, Status='live', Symbol='Sh2b1', Aliases='AI425885, C530001K22Rik, Irip, Psm, SH2-B, SH2-Bb, Sh2bpsm1, mKIAA1299', description='SH2B adaptor protein 1', other_designations='SH2 domain-containing protein 1B|SH2-B PH domain containing signaling mediator 1|pro-rich, PH and SH2 domain-containing signaling mediator', map_location='7 69.06 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126466993, end_position_on_the_genomic_accession=126475684, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  20400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20400, CurrentID=0, Status='live', Symbol='Sh2d1a', Aliases='Gm686, SAP', description='SH2 domain containing 1A', other_designations='Duncan disease homolog|SH2 domain protein 1A|SLAM-associated protein isoform SAP-2|T-cell signal transduction molecule SAP|signaling lymphocytic activation molecule-associated protein', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=42502608, end_position_on_the_genomic_accession=42522075, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20401, CurrentID=0, Status='live', Symbol='Sh3bp1', Aliases='3BP-1', description='SH3-domain binding protein 1', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78899780, end_position_on_the_genomic_accession=78912052, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  20402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20402, CurrentID=0, Status='live', Symbol='Zfp106', Aliases='Cd-1, D2Dcr28, H3a, Sh3bp3, Znf106, sirm', description='zinc finger protein 106', other_designations='H3a minor histocompatibility antigen|SH3-domain binding protein 3|zfp-106|zinc finger protein 474', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120506824, end_position_on_the_genomic_accession=120563858, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  20403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20403, CurrentID=0, Status='live', Symbol='Itsn2', Aliases='AI327390, Ese2, Sh3d1B, Sh3p18, mKIAA1256', description='intersectin 2', other_designations='EH and SH3 domains protein 2|EH domain and SH3 domain regulator of endocytosis 2|Eh domain, SH3 domain regulator of endocytosis 2|SH3 domain protein 1B|SH3 domain-containing protein 1B', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4592638, end_position_on_the_genomic_accession=4713952, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  20404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20404, CurrentID=0, Status='live', Symbol='Sh3gl2', Aliases='9530001L19Rik, AI120490, AW555077, B930049H17Rik, EEN-B1, SH3PA, Sh3d2a', description='SH3-domain GRB2-like 2', other_designations='SH3 domain protein 2A|SH3 domain-containing GRB2-like protein 2|SH3p4|endophilin A1|endophilin I|endophilin-1', map_location='4 40.23 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=85205456, end_position_on_the_genomic_accession=85389380, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20405, CurrentID=0, Status='live', Symbol='Sh3gl1', Aliases='C77078, EEN, SH3P8, Sh3d2b', description='SH3-domain GRB2-like 1', other_designations='SH3 domain protein 2B|SH3 domain-containing GRB2-like protein 1|endophilin II|endophilin-2', map_location='17 29.11 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56016750, end_position_on_the_genomic_accession=56036638, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  20406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20406, CurrentID=0, Status='live', Symbol='Sh3d2c-ps1', Aliases='SH3P13, Sh3d2c1, Sh3gl3', description='SH3 domain protein 2C, pseudogene 1', other_designations='SH3 domain protein 2 C1', map_location='2 15.68 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=23680182, end_position_on_the_genomic_accession=23681590, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  20408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20408, CurrentID=0, Status='live', Symbol='Sh3gl3', Aliases='EEN-B2, SH3P13, Sh3d2c, Sh3d2c2', description='SH3-domain GRB2-like 3', other_designations='SH3 domain protein 2 C2|SH3 domain protein 2C|SH3 domain-containing GRB2-like protein 3|endophilin III|endophilin-3', map_location='7 D2-3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=82259908, end_position_on_the_genomic_accession=82307420, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20409, CurrentID=0, Status='live', Symbol='Ostf1', Aliases='C78236, SH3P2, Sh3d3', description='osteoclast stimulating factor 1', other_designations='SH3 domain protein 3', map_location='19 13.17 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=18580364, end_position_on_the_genomic_accession=18631813, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20410, CurrentID=0, Status='live', Symbol='Sorbs3', Aliases='SCAM-1, SH3P3, Sh3d4, vinexin-g', description='sorbin and SH3 domain containing 3', other_designations='SH3 domain protein 4|SH3 domain-containing protein SH3P3|SH3-containing adapter molecule 1|vinexin alpha|vinexin beta', map_location='14 36.27 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70180468, end_position_on_the_genomic_accession=70212457, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  20411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20411, CurrentID=0, Status='live', Symbol='Sorbs1', Aliases='2310065E01Rik, 9530001P15Rik, CAP, SH3P12, Sh3d5, mKIAA1296', description='sorbin and SH3 domain containing 1', other_designations='SH3 domain protein 5|c-Cbl-associated protein|ponsin|sorbs1', map_location='19 34.25 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40292040, end_position_on_the_genomic_accession=40513876, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  20415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20415, CurrentID=0, Status='live', Symbol='Shbg', Aliases='ABP', description='sex hormone binding globulin', other_designations='SBP|sex steroid-binding protein|testis-specific androgen-binding protein', map_location='11 42.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69610281, end_position_on_the_genomic_accession=69618025, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20416, CurrentID=0, Status='live', Symbol='Shc1', Aliases='Shc, ShcA, p66, p66shc', description='src homology 2 domain-containing transforming protein C1', other_designations='SH2 domain protein C1|SHC-transforming protein A|Src homology 2 domain-containing transforming protein 1 isoform p66|src homology 2 domain-containing-transforming protein C1', map_location='3 39.11 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89418481, end_position_on_the_genomic_accession=89430029, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20418, CurrentID=0, Status='live', Symbol='Shc3', Aliases='N-Shc, Rai, ShcC', description='src homology 2 domain-containing transforming protein C3', other_designations='SH2 domain protein C3|SHC-transforming protein C|neuronal Shc|src homology 2 domain-containing-transforming protein C3', map_location='13 26.12 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51431041, end_position_on_the_genomic_accession=51567084, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20419, CurrentID=0, Status='live', Symbol='Shcbp1', Aliases='mPAL', description='Shc SH2-domain binding protein 1', other_designations='SHC-binding protein|protein expressed in activated lymphocytes', map_location='8 A1.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4735980, end_position_on_the_genomic_accession=4779534, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  20420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20420, CurrentID=0, Status='live', Symbol='Shd', Aliases='AI413439', description='src homology 2 domain-containing transforming protein D', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55968392, end_position_on_the_genomic_accession=55976625, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  544710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544710, CurrentID=0, Status='live', Symbol='Krtap10-10', Aliases='EG544710', description='keratin associated protein 10-10', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77835947, end_position_on_the_genomic_accession=77837251, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  20423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20423, CurrentID=0, Status='live', Symbol='Shh', Aliases='9530036O11Rik, Dsh, Hhg1, Hx, Hxl3, M100081', description='sonic hedgehog', other_designations='HHG-1|hemimelic extra toes|short digits', map_location='5 14.39 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=28456840, end_position_on_the_genomic_accession=28467101, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20425, CurrentID=0, Status='live', Symbol='Shmt1', Aliases='AI324848, AI385541, C81125, Shmt, mshmt, mshmt1, mshmt2', description='serine hydroxymethyltransferase 1 (soluble)', other_designations='glycine hydroxymethyltransferase|serine methylase', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60788892, end_position_on_the_genomic_accession=60811265, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  20428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20428, CurrentID=0, Status='live', Symbol='sho', Aliases='', description='shorthead', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20429, CurrentID=0, Status='live', Symbol='Shox2', Aliases='6330543G17Rik, OG12, Og12x, Prx3, SHOT', description='short stature homeobox 2', other_designations='OG-12|homeobox protein Og12X|paired family homeodomain protein Prx3', map_location='3 E3-F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=66971723, end_position_on_the_genomic_accession=66981771, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20430, CurrentID=0, Status='live', Symbol='Cyfip1', Aliases='E030028J09Rik, P140SRA-1, P140sra1, Shyc, Sra-1, Sra1, l(7)1Rl, l71Rl, l7Rl1, mKIAA0068, pl-1', description='cytoplasmic FMR1 interacting protein 1', other_designations='cyfip1|specifically Rac1-associated protein 1', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=55842071, end_position_on_the_genomic_accession=55932633, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  20431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20431, CurrentID=0, Status='live', Symbol='Pmel', Aliases='D10H12S53E, D12S53Eh17, Si, Silv, gp100, gp87, Pmel', description='premelanosome protein', other_designations='melanocyte protein pmel 17|silver locus protein', map_location='10 77.13 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128705168, end_position_on_the_genomic_accession=128720238, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20437, CurrentID=0, Status='live', Symbol='Siah1a', Aliases='AA982064, AI853500, Sinh1a', description='seven in absentia 1A', other_designations='mSiah-1a|seven in absentia homolog 1a|siah-1a', map_location='8 42.1 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=86723920, end_position_on_the_genomic_accession=86746006, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20438, CurrentID=0, Status='live', Symbol='Siah1b', Aliases='AA960570, Sinh1b, siah-1b', description='seven in absentia 1B', other_designations='drosophila seven-in-absentia gene product homolog|seven in absentia homolog 1b', map_location='X 76.11 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164070703, end_position_on_the_genomic_accession=164076664, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20439, CurrentID=0, Status='live', Symbol='Siah2', Aliases='AA415433, Sinh2', description='seven in absentia 2', other_designations='mSiah2|seven in absentia homolog 2|siah-2', map_location='3 28.68 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=58674949, end_position_on_the_genomic_accession=58692388, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20440, CurrentID=0, Status='live', Symbol='St6gal1', Aliases='AW742324, Siat1, St6Gal-I, St6gal, St6galI', description='beta galactoside alpha 2,6 sialyltransferase 1', other_designations='CMP-N-acetylneuraminate-beta-galactosamide-alpha-2,6-sialyltransferase 1|ST6Gal I|alpha 2,6-ST 1', map_location='16 14.03 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23224728, end_position_on_the_genomic_accession=23360350, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20441, CurrentID=0, Status='live', Symbol='St3gal3', Aliases='ST3GalIII, ST3N, Siat3, Siat6', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 3', other_designations='N-acetyllactosaminide alpha-2,3-sialyltransferase|alpha 2,3-ST 3|beta-galactoside alpha-2,3-sialyltransferase 3|gal beta-1,3(4) GlcNAc alpha-2,3 sialyltransferase|sialyltransferase 3|sialyltransferase 6 (N-acetyllacosaminide alpha 2,3-sialyltransferase)', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117932152, end_position_on_the_genomic_accession=118134946, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  20442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20442, CurrentID=0, Status='live', Symbol='St3gal1', Aliases='5330418N22Rik, AI467004, ST3GalI, Siat4, Siat4a, St3gal-1', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 1', other_designations='CMP-N-acetylneuraminate: [beta-galactosidase alpha-2,3] sialytransferase|Gal beta1,3 GalNAc alpha2,3 sialyltransferase|SIAT4-A|ST3GAL-I sialyltransferase|ST3Gal I|ST3GalA.1|ST3O|alpha 2,3-ST 1|beta-galactoside alpha-2,3-sialyltransferase 1|gal-NAc6S|gal-beta-1,3-GalNAc-alpha-2,3-sialyltransferase|sialyltransferase 4A (beta-galactosidase alpha-2,3-sialytransferase)|sialyltransferase 4A (beta-galactoside alpha-2,3-sialytransferase)', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=67102875, end_position_on_the_genomic_accession=67176910, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  20443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20443, CurrentID=0, Status='live', Symbol='St3gal4', Aliases='Siat4c', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 4', other_designations='SIAT4-C|ST3Gal IV|ST3GalIV|alpha 2,3-ST 4|alpha 2,3-sialyltransferase IV|beta-galactoside alpha-2,3-sialyltransferase 4|gal-beta-1,4-GalNAc-alpha-2,3-sialyltransferase|sialyltransferase 4C (beta-galactosidase alpha-2,3-sialytransferase)|sialyltransferase 4C (beta-galactoside alpha-2,3-sialytransferase)', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35046576, end_position_on_the_genomic_accession=35117445, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  20444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20444, CurrentID=0, Status='live', Symbol='St3gal2', Aliases='AI429591, AW822065, ST3GalII, Siat5', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 2', other_designations='SIAT4-B|ST3Gal II|ST3GalA.2|alpha 2,3-ST 2|beta-galactoside alpha-2,3-sialyltransferase 2|gal-NAc6S|gal-beta-1,3-GalNAc-alpha-2,3-sialyltransferase|sialyltransferase 4B|sialyltransferase 5', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110919833, end_position_on_the_genomic_accession=110972497, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20445, CurrentID=0, Status='live', Symbol='St6galnac1', Aliases='Siat7a', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 1', other_designations='SIAT7-A|ST6GalNAc I|ST6GalNAcI|sialyltransferase 7 ((alpha-N-acetylneuraminyl 2,3-beta-galactosyl-1,3)-N-acetyl galactosaminde alpha-2,6-sialyltransferase) A|sialyltransferase 7A', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116765025, end_position_on_the_genomic_accession=116775507, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20446, CurrentID=0, Status='live', Symbol='St6galnac2', Aliases='II, ST6GalNAc, Siat7, Siat7b', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 2', other_designations='Gal-beta-1,3-GalNAc alpha-2,6-sialyltransferase|SIAT7-B|ST6GalNAc II|ST6GalNAcII|alpha-N-acetylgalactosaminide alpha-2,6-sialyltransferase II|galNAc alpha-2,6-sialyltransferase II|sialyltransferase 7 ((alpha-N-acetylneuraminyl 2,3-beta-galactosyl-1,3)-N-acetyl galactosaminde alpha-2,6-sialyltransferase) B|sialyltransferase 7B', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116676703, end_position_on_the_genomic_accession=116695231, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20447, CurrentID=0, Status='live', Symbol='St6galnac3', Aliases='Siat7c', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 3', other_designations='SIAT7-C|ST6GalNAc III|ST6GalNAcIII|STY|galNAc alpha-2,6-sialyltransferase III|sialyltransferase 7 ((alpha-N-acetylneuraminyl 2,3-betagalactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) C|sialyltransferase 7C', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=153202509, end_position_on_the_genomic_accession=153725133, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20448, CurrentID=0, Status='live', Symbol='St6galnac4', Aliases='SIAT7-D, ST6GalNAcIV, Siat7d', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 4', other_designations='GalNAc alpha-2,6-sialyltransferase|ST6 neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 4|ST6GalNAc IV|Sia-alpha-2-3-Gal-beta-1-3-GalNAc GalNAc-alpha-2,6-sialyltransferase|neuAc-alpha-2,3-Gal-beta-1,3-GalNAc-alpha-2,6-sialyltransferase|sialyltransferase 7 ((alpha-N-acetylneuraminyl 2,3-betagalactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) D|sialyltransferase 7D', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32587078, end_position_on_the_genomic_accession=32599696, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20449, CurrentID=0, Status='live', Symbol='St8sia1', Aliases='9330109E03Rik, GD3S, Sia-T, Siat8, Siat8a', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 1', other_designations='GD3 synthase|SIAT8-A|ST8Sia I|ST8SiaI|alpha-2, 8-sialyltransferase|alpha-2,8-sialyltransferase 8A|ganglioside GD3 synthase|ganglioside GT3 synthase|sialyltransferase 8 A|sialyltransferase 8A|sialyltransferase St8Sia I|sialytransferase St8Sia I', map_location='6 74.78 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142821541, end_position_on_the_genomic_accession=142989530, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20450, CurrentID=0, Status='live', Symbol='St8sia2', Aliases='AI323367, ST8SiaII, STX, Siat8b', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 2', other_designations='SIAT8-B|polysialic acid synthase|sialyltransferase 8 (alpha-2, 8-sialyltransferase) B|sialyltransferase 8B|sialyltransferase St8Sia II|sialyltransferase X|sialytransferase St8Sia II', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=73939119, end_position_on_the_genomic_accession=74013727, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20451, CurrentID=0, Status='live', Symbol='St8sia3', Aliases='AI847333, ST8SiaIII, Siat8c', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 3', other_designations='SIAT8-C|alpha-2,8-sialyltransferase 8C|alpha-2,8-sialyltransferase III|sialyltransferase 8 (alpha-2, 8-sialyltransferase) C|sialyltransferase 8C|sialyltransferase St8Sia III|sialytransferase St8Sia III', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=64254359, end_position_on_the_genomic_accession=64276144, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20452, CurrentID=0, Status='live', Symbol='St8sia4', Aliases='PST, PST-1, SIAT8-D, ST8SiaIV, Siat8d', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 4', other_designations='alpha-2,8-sialyltransferase 8D|polysialyltransferase-1|sialyltransferase 8 (alpha-2, 8-sialyltransferase) D|sialyltransferase 8D|sialyltransferase St8Sia IV|sialytransferase St8Sia IV', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=95587682, end_position_on_the_genomic_accession=95667649, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20454, CurrentID=0, Status='live', Symbol='St3gal5', Aliases='3S-T, Siat9, [a]2', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 5', other_designations='CMP-NeuAc:lactosylceramide alpha-2,3-sialyltransferase|GM3 synthase|GM3-specific sialytransferase|ST3GalV|ganglioside GM3 synthase|mST3Gal V|sialyltransferase 9 (CMP-NeuAc:lactosylceramide alpha-2,3-sialyltransferase)', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72097554, end_position_on_the_genomic_accession=72154570, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20459, CurrentID=0, Status='live', Symbol='Ptk6', Aliases='BRK, Sik, Tksk, tks', description='PTK6 protein tyrosine kinase 6', other_designations='protein tyrosine kinase 6|src-related intestinal kinase|tyrosine kinase identified from skin', map_location='2 103.62 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181195124, end_position_on_the_genomic_accession=181203810, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20460, CurrentID=0, Status='live', Symbol='Stil', Aliases='Sil', description='Scl/Tal1 interrupting locus', other_designations='Tal1 interrupting locus', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115000107, end_position_on_the_genomic_accession=115043203, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  20462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20462, CurrentID=0, Status='live', Symbol='Tra2b', Aliases='5730405G21Rik, D16Ertd266e, SIG-41, Sfrs10, Silg41, TRA2beta', description='transformer 2 beta homolog (Drosophila)', other_designations='TRA-2 beta|TRA2-beta|silica-induced gene 41 protein|splicing factor, arginine/serine-rich 10 (transformer 2 homolog, Drosophila)|transformer-2 protein homolog B', map_location='16 13.18 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22245741, end_position_on_the_genomic_accession=22266000, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20463, CurrentID=0, Status='live', Symbol='Cox7a2l', Aliases='COX7AR, COX7RP, EB1, SIG-81, SIG81, Silg81', description='cytochrome c oxidase subunit VIIa polypeptide 2-like', other_designations='cytochrome c oxidase subunit VIIa-related protein|silica-induced gene 81 protein|supercomplex assembly factor I', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=83501917, end_position_on_the_genomic_accession=83514333, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20464, CurrentID=0, Status='live', Symbol='Sim1', Aliases='bHLHe14, mSIM1', description='single-minded homolog 1 (Drosophila)', other_designations='single-minded 1', map_location='10 24.87 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=50894673, end_position_on_the_genomic_accession=50989156, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  20465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20465, CurrentID=0, Status='live', Symbol='Sim2', Aliases='bHLHe15', description='single-minded homolog 2 (Drosophila)', other_designations='SIM transcription factor|mSIM|single-minded 2', map_location='16 C3.3-C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94085198, end_position_on_the_genomic_accession=94127253, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  20466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20466, CurrentID=0, Status='live', Symbol='Sin3a', Aliases='AW553200, Sin3, mKIAA4126, mSin3A', description='transcriptional regulator, SIN3A (yeast)', other_designations='histone deacetylase complex subunit Sin3a|transcriptional corepressor Sin3a|transcriptional regulator, SIN3 yeast homolog A', map_location='9 30.89 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57071991, end_position_on_the_genomic_accession=57128368, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  20467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20467, CurrentID=0, Status='live', Symbol='Sin3b', Aliases='2810430C10Rik', description='transcriptional regulator, SIN3B (yeast)', other_designations='SIN3 homolog B, transcription regulator|histone deacetylase complex subunit Sin3b|transcriptional corepressor Sin3b|transcriptional regulator, SIN3 yeast homolog B', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72723270, end_position_on_the_genomic_accession=72758203, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  20469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20469, CurrentID=0, Status='live', Symbol='Sipa1', Aliases='Spa1', description='signal-induced proliferation associated gene 1', other_designations='GTPase-activating protein Spa-1|sipa-1', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5651185, end_position_on_the_genomic_accession=5663947, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  20471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20471, CurrentID=0, Status='live', Symbol='Six1', Aliases='BB138287', description='sine oculis-related homeobox 1', other_designations='sine oculis homeobox homolog 1|sine oculis-related homeobox 1 homolog', map_location='12 30.34 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73041827, end_position_on_the_genomic_accession=73046712, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20472, CurrentID=0, Status='live', Symbol='Six2', Aliases='', description='sine oculis-related homeobox 2', other_designations='sine oculis homeobox homolog 2|sine oculis-related homeobox 2 homolog', map_location='17 55.72 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=85662735, end_position_on_the_genomic_accession=85688254, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20473, CurrentID=0, Status='live', Symbol='Six3', Aliases='E130112M24Rika, Six3alpha, Six3b, Six3beta, Six3', description='sine oculis-related homeobox 3', other_designations='sine oculis homeobox homolog 3|sine oculis-related homeobox 3 homolog', map_location='17 55.42 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=85620834, end_position_on_the_genomic_accession=85626191, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20474, CurrentID=0, Status='live', Symbol='Six4', Aliases='AI047561, AREC3, TrexBF', description='sine oculis-related homeobox 4', other_designations='sine oculis homeobox homolog 4|sine oculis-related homeobox 4 homolog|skeletal muscle-specific ARE-binding protein AREC3', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73099609, end_position_on_the_genomic_accession=73114774, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  544763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544763, CurrentID=0, Status='live', Symbol='Hbq1b', Aliases='F830116E18Rik, Hbqt1', description='hemoglobin, theta 1B', other_designations="Hbq1 5'-theta", map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32286702, end_position_on_the_genomic_accession=32296033, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20476, CurrentID=0, Status='live', Symbol='Six6', Aliases='Optx2, Six9', description='sine oculis-related homeobox 6', other_designations='optic homeobox 2|sine oculis homeobox homolog 6|sine oculis-related homeobox 6 homolog', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72939945, end_position_on_the_genomic_accession=72944899, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20479, CurrentID=0, Status='live', Symbol='Vps4b', Aliases='8030489C12Rik, Skd1', description='vacuolar protein sorting 4b (yeast)', other_designations='suppressor of K(+) transport growth defect 1|suppressor of K+ transport defect 1', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=106770788, end_position_on_the_genomic_accession=106796725, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  20480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20480, CurrentID=0, Status='live', Symbol='Clpb', Aliases='AL118244, Skd3', description='ClpB caseinolytic peptidase B', other_designations='ClpB caseinolytic peptidase B homolog|suppressor of K+ transport defect 3|suppressor of potassium transport defect 3', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101663595, end_position_on_the_genomic_accession=101790168, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  20481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20481, CurrentID=0, Status='live', Symbol='Ski', Aliases='2310012I02Rik, 2610001A11Rik, AA062172, AA589460, BC004088', description='ski sarcoma viral oncogene homolog (avian)', other_designations='C-ski|Sloan-Kettering viral oncogene homolog|proto-oncogene c-Ski', map_location='4 86.17 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155154075, end_position_on_the_genomic_accession=155222593, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20482, CurrentID=0, Status='live', Symbol='Skil', Aliases='Skir, SnoN, sno', description='SKI-like', other_designations='ski-related oncogene|ski/sno related', map_location='3 15.12 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=31095058, end_position_on_the_genomic_accession=31122923, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20491, CurrentID=0, Status='live', Symbol='Sla', Aliases='Slapp-1, Sla', description='src-like adaptor', other_designations='src-like-adapter protein 1', map_location='15 29.3 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=66780819, end_position_on_the_genomic_accession=66831829, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20492, CurrentID=0, Status='live', Symbol='Slbp', Aliases='', description='stem-loop binding protein', other_designations='histone stem-loop-binding protein', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33640055, end_position_on_the_genomic_accession=33652574, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20493, CurrentID=0, Status='live', Symbol='Slc10a1', Aliases='Ntcp', description='solute carrier family 10 (sodium/bile acid cotransporter family), member 1', other_designations='Na(+)/bile acid cotransporter|Na(+)/taurocholate transport protein|bile acid cotransporting polypeptide|sodium bile acid cotransporting polypeptide|sodium-taurocholate cotransporting polypeptide|sodium/taurocholate cotransporting polypeptide', map_location='12 37.21 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80953185, end_position_on_the_genomic_accession=80968702, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20494, CurrentID=0, Status='live', Symbol='Slc10a2', Aliases='ASBT, ISBT', description='solute carrier family 10, member 2', other_designations='IBAT|Na(+)-dependent ileal bile acid transporter|apical sodium-dependent bile acid transporter|ileal Na(+)/bile acid cotransporter|ileal sodium-dependent bile acid transporter|sodium/taurocholate cotransporting polypeptide, ileal', map_location='8 2.16 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=5085623, end_position_on_the_genomic_accession=5105232, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20495, CurrentID=0, Status='live', Symbol='Slc12a1', Aliases='AI788571, D630042G03Rik, Nkcc2, mBSC1, urehr3', description='solute carrier family 12, member 1', other_designations='BSC1|bumetanide-sensitive cotransporter type 1|bumetanide-sensitive sodium-(potassium)-chloride cotransporter 2|kidney-specific Na-K-Cl symporter', map_location='2 61.23 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125152600, end_position_on_the_genomic_accession=125230002, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  20496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20496, CurrentID=0, Status='live', Symbol='Slc12a2', Aliases='9330166H04Rik, Nkcc1, mBSC2, sy-ns', description='solute carrier family 12, member 2', other_designations='basolateral Na-K-Cl symporter|bumetanide-sensitive sodium-(potassium)-chloride cotransporter 1|sodium/potassium/chloride cotransporters|sodium/potassium/chloride transporters', map_location='18 32.15 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=57878513, end_position_on_the_genomic_accession=57946821, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  20497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20497, CurrentID=0, Status='live', Symbol='Slc12a3', Aliases='AI035291, NCC, NCCT, TSC', description='solute carrier family 12, member 3', other_designations='Na-Cl symporter|thiazide-sensitive sodium-chloride cotransporter', map_location='8 46.46 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94329112, end_position_on_the_genomic_accession=94366222, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20498, CurrentID=0, Status='live', Symbol='Slc12a4', Aliases='AW546649, KCC1, RBCKCC1', description='solute carrier family 12, member 4', other_designations='K-Cl Co-transporter-1|K-Cl cotransporter KCC1|electroneutral potassium-chloride cotransporter 1|erythroid K-Cl cotransporter 1|mKCC1', map_location='8 53.06 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105943590, end_position_on_the_genomic_accession=105966149, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  20499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20499, CurrentID=0, Status='live', Symbol='Slc12a7', Aliases='AA408796, D13Ertd261e, Kcc4', description='solute carrier family 12, member 7', other_designations='K-Cl cotransporter 4|electroneutral potassium-chloride cotransporter 4', map_location='13 40.15 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73733053, end_position_on_the_genomic_accession=73816755, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  20500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20500, CurrentID=0, Status='live', Symbol='Slc13a2', Aliases='Nadc1, mNaDC-1', description='solute carrier family 13 (sodium-dependent dicarboxylate transporter), member 2', other_designations='Na(+)/dicarboxylate cotransporter 1|naDC-1|renal sodium/dicarboxylate cotransporter|sodium/dicarboxylate co-transporter', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78397276, end_position_on_the_genomic_accession=78422280, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  20501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20501, CurrentID=0, Status='live', Symbol='Slc16a1', Aliases='AL022710, Mct1', description='solute carrier family 16 (monocarboxylic acid transporters), member 1', other_designations='solute carrier family 16 member 1', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104638664, end_position_on_the_genomic_accession=104658462, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20502, CurrentID=0, Status='live', Symbol='Slc16a2', Aliases='AW105741, Mct8, Xpct', description='solute carrier family 16 (monocarboxylic acid transporters), member 2', other_designations='MCT 8|Monocarboxylate transporter 8 (MCT 8) (X-linked PEST-containing transporter)|X-linked PEST-containing transporter|solute carrier family 16 member 2', map_location='X 46.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103697414, end_position_on_the_genomic_accession=103821988, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  544791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544791, CurrentID=0, Status='live', Symbol='Myh13', Aliases='MyHC-eo', description='myosin, heavy polypeptide 13, skeletal muscle', other_designations='EO Myosin|extraocular myosin', map_location='11 40.85 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67327103, end_position_on_the_genomic_accession=67371502, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  20504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20504, CurrentID=0, Status='live', Symbol='Slc17a1', Aliases='NAPI-1, Napi1, Npt1', description='solute carrier family 17 (sodium phosphate), member 1', other_designations='Na(+)/PI cotransporter 1|renal Na(+)-dependent phosphate cotransporter 1|renal sodium-phosphate transport protein 1|sodium-dependent phosphate/anion cotransporter|sodium/phosphate type I cotransporter|solute carrier family 17 (sodium/hydrogen exchanger), member 1|solute carrier family 17 vesicular glutamate transporter), member 1', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23867723, end_position_on_the_genomic_accession=23895730, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  20505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20505, CurrentID=0, Status='live', Symbol='Slc34a1', Aliases='NaPi-IIa, Npt2, Npt2a, Slc17a2', description='solute carrier family 34 (sodium phosphate), member 1', other_designations='Na(+)-dependent phosphate cotransporter 2A|Na(+)/Pi cotransporter 2A|Na/Pi cotransporter|naPi-2a|naPi-7|renal Na+/Pi transporter|sodium-phosphate transport protein 2A|sodium/phosphate cotransporter 2A|solute carrier family 17 (sodium/hydrogen exchanger), member 2|solute carrier family 34 member 1|type IIa Na+/Pi-cotransporter', map_location='13 B', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55399622, end_position_on_the_genomic_accession=55414695, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20508, CurrentID=0, Status='live', Symbol='Slc18a3', Aliases='VAChT, VAT', description='solute carrier family 18 (vesicular monoamine), member 3', other_designations='solute carrier family 18 member 3', map_location='14 19.4 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32462437, end_position_on_the_genomic_accession=32464850, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  20509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20509, CurrentID=0, Status='live', Symbol='Slc19a1', Aliases='AI323572, RFC, RFC-1, RFC1', description='solute carrier family 19 (folate transporter), member 1', other_designations='IFC-1|intestinal folate carrier 1|intestinal folate carrier protein|reduced folate carrier 1|solute carrier family 19 (sodium/hydrogen exchanger), member 1|solute carrier family 19 member 1', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77032269, end_position_on_the_genomic_accession=77050432, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20510, CurrentID=0, Status='live', Symbol='Slc1a1', Aliases='D130048G10Rik, EAAC1, EAAC2, EAAT3, MEAAC1', description='solute carrier family 1 (neuronal/epithelial high affinity glutamate transporter, system Xag), member 1', other_designations='excitatory amino acid carrier 2|excitatory amino-acid carrier 1|glutamate transporter mEAAC2|sodium-dependent glutamate/aspartate transporter 3|solute carrier family 1, member 1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=28835166, end_position_on_the_genomic_accession=28913960, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20511, CurrentID=0, Status='live', Symbol='Slc1a2', Aliases='1700091C19Rik, 2900019G14Rik, AI159670, Eaat2, GLT-1, GLT1, MGLT1', description='solute carrier family 1 (glial high affinity glutamate transporter), member 2', other_designations='EAAT2 variant mEAAT2/5UT5|glial high affinity glutamate transporter|glutamate transporter 1c|sodium-dependent glutamate/aspartate transporter 2|solute carrier family 1 member 2|solute carrier family 1, member 2', map_location='2 54.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=102658594, end_position_on_the_genomic_accession=102790784, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20512, CurrentID=0, Status='live', Symbol='Slc1a3', Aliases='AI504299, B430115D02Rik, Eaat1, GLAST, GLAST-1, GLU-T, GluT-1, Gmt1, MGluT1', description='solute carrier family 1 (glial high affinity glutamate transporter), member 3', other_designations='glial high affinity glutamate transporter|glutamate/aspartate transporter|high-affinity neuronal glutamate transporter|sodium-dependent glutamate/aspartate transporter 1|solute carrier family 1 member 3|solute carrier family 1, member 3', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=8634124, end_position_on_the_genomic_accession=8710807, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20513, CurrentID=0, Status='live', Symbol='Slc1a6', Aliases='EAAT4', description='solute carrier family 1 (high affinity aspartate/glutamate transporter), member 6', other_designations='high-affinity neuronal glutamate transporter|sodium-dependent glutamate/aspartate transporter|solute carrier family 1 member 6|solute carrier family 1, member 6', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78780496, end_position_on_the_genomic_accession=78814825, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20514, CurrentID=0, Status='live', Symbol='Slc1a5', Aliases='AAAT, ASCT2, ATBO, M7V1, M7VS1, R16, RDRC, Slc1a7', description='solute carrier family 1 (neutral amino acid transporter), member 5', other_designations='ASC-like Na(+)-dependent neutral amino acid transporter ASCT2|ATB(0)|insulin-activated amino acid transporter|sodium-dependent neutral amino acid transporter type 2|solute carrier family 1 member 5|solute carrier family 1, member 7', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16781345, end_position_on_the_genomic_accession=16798274, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20515, CurrentID=0, Status='live', Symbol='Slc20a1', Aliases='AI607883, Glvr-1, Glvr1', description='solute carrier family 20, member 1', other_designations='gibbon ape leukemia virus receptor 1|leukemia virus receptor 1 homolog|phosphate transporter 1|piT-1', map_location='2 62.79 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129198773, end_position_on_the_genomic_accession=129211615, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20516, CurrentID=0, Status='live', Symbol='Slc20a2', Aliases='MolPit2, Pit-2, Pit2, Ram-1, Ram1', description='solute carrier family 20, member 2', other_designations='phosphate transporter 2|type III sodium-dependent phosphate transporter', map_location='8 11.42 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22476700, end_position_on_the_genomic_accession=22569616, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20517, CurrentID=0, Status='live', Symbol='Slc22a1', Aliases='Lx1, Oct1, Orct, Orct1', description='solute carrier family 22 (organic cation transporter), member 1', other_designations='mOCT1|organic cation transporter 1|solute carrier family 22, member 1', map_location='17 8.63 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12648874, end_position_on_the_genomic_accession=12675838, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  544806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544806, CurrentID=0, Status='live', Symbol='Tmem92', Aliases='AU016430-ps, Tmem92', description='transmembrane protein 92', other_designations='transmembrane protein 92, pseudogene', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94777217, end_position_on_the_genomic_accession=94782703, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  282663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=282663, CurrentID=0, Status='live', Symbol='Serpinb1b', Aliases='6330533H24Rik, EIB, Serpin1b1, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 1b', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 1b|serine protease inhibitor EIB|serpin B1b', map_location='13 13.81 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33078504, end_position_on_the_genomic_accession=33094380, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20520, CurrentID=0, Status='live', Symbol='Slc22a5', Aliases='Lstpl, Octn2, jvs', description='solute carrier family 22 (organic cation transporter), member 5', other_designations='high-affinity sodium-dependent carnitine cotransporter|juvenile visceral steatosis|organic cation/carnitine transporter 2', map_location='11 32.02 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53864542, end_position_on_the_genomic_accession=53891799, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20521, CurrentID=0, Status='live', Symbol='Slc22a12', Aliases='AI987855, OAT4L, Rst, Slc22al2, URAT1', description='solute carrier family 22 (organic anion/cation transporter), member 12', other_designations='renal-specific transporter|solute carrier family 22 (organic cation transporter)-like 2|urate anion exchanger 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6535854, end_position_on_the_genomic_accession=6543070, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20522, CurrentID=0, Status='live', Symbol='Slc23a1', Aliases='D18Ucla2, SVCT1, Slc23a2, YSPL3', description='solute carrier family 23 (nucleobase transporters), member 1', other_designations='Na(+)/L-ascorbic acid transporter 1|sodium-dependent vitamin C transporter 1|solute carrier family 23 (nucleobase transporters), member 2|yolk sac permease-like molecule 3', map_location='18 19.17 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35604041, end_position_on_the_genomic_accession=35630340, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  20523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20523, CurrentID=0, Status='live', Symbol='Slc25a14', Aliases='BMCP1, UCP5, UCP5L, UCP5S', description='solute carrier family 25 (mitochondrial carrier, brain), member 14', other_designations='mitochondrial uncoupling protein 5', map_location='X 25.69 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48623418, end_position_on_the_genomic_accession=48662298, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20524, CurrentID=0, Status='live', Symbol='Slc25a17', Aliases='34kDa, 47kDa, PMP34, Pmp47', description='solute carrier family 25 (mitochondrial carrier, peroxisomal membrane protein), member 17', other_designations='34 kDa peroxisomal membrane protein|peroxisomal integral membrane protein 47|peroxisomal membrane protein, 34 kDa|solute carrier family 25 member 17|solute carrier family 25, member 17', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81318921, end_position_on_the_genomic_accession=81360765, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20525, CurrentID=0, Status='live', Symbol='Slc2a1', Aliases='Glut-1, Glut1', description='solute carrier family 2 (facilitated glucose transporter), member 1', other_designations='GT1|glucose transporter type 1, erythrocyte/brain|solute carrier family 2, member 1', map_location='4 55.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119108745, end_position_on_the_genomic_accession=119137330, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20526, CurrentID=0, Status='live', Symbol='Slc2a2', Aliases='AI266973, Glut-2, Glut2', description='solute carrier family 2 (facilitated glucose transporter), member 2', other_designations='glucose transporter type 2, liver|liver-type glucose transporter', map_location='3 12.19 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=28697903, end_position_on_the_genomic_accession=28728361, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20527, CurrentID=0, Status='live', Symbol='Slc2a3', Aliases='AA408729, AL023014, AL024341, AU040424, C78366, Glut-3, Glut3', description='solute carrier family 2 (facilitated glucose transporter), member 3', other_designations='glucose transporter type 3, brain', map_location='6 57.82 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122727809, end_position_on_the_genomic_accession=122802270, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20528, CurrentID=0, Status='live', Symbol='Slc2a4', Aliases='Glut-4, Glut4', description='solute carrier family 2 (facilitated glucose transporter), member 4', other_designations='GT2|glucose transporter type 4, insulin-responsive', map_location='11 C-E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69942286, end_position_on_the_genomic_accession=69948221, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20529, CurrentID=0, Status='live', Symbol='Slc31a1', Aliases='4930445G01Rik, AI787263, AU016967, Ctr1', description='solute carrier family 31, member 1', other_designations='copper transporter 1|solute carrier family 13 (sodium/sulphate symporters), member 1', map_location='4 C1-C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62360582, end_position_on_the_genomic_accession=62391769, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20530, CurrentID=0, Status='live', Symbol='Slc31a2', Aliases='AI604396, CTR2', description='solute carrier family 31, member 2', other_designations='copper transporter 2|solute carrier family 31 (copper transporters), member 2', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62280393, end_position_on_the_genomic_accession=62298412, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20531, CurrentID=0, Status='live', Symbol='Slc34a2', Aliases='AA536683, D5Ertd227e, NaPi-2b, Npt2b', description='solute carrier family 34 (sodium phosphate), member 2', other_designations='Na(+)-dependent phosphate cotransporter 2B|Na(+)/Pi cotransporter 2B|sodium-phosphate transport protein 2B|sodium/phosphate cotransporter 2B|solute carrier family 34 member 2|type IIb Na/Picotransporter', map_location='5 28.92 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=53049203, end_position_on_the_genomic_accession=53071664, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  20532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20532, CurrentID=0, Status='live', Symbol='Slc3a1', Aliases='D2H, NBAT, NTAA, RBAT', description='solute carrier family 3, member 1', other_designations='b(0,+)-type amino acid transport protein|solute carrier family 3 member 1', map_location='17 55.17 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=85028347, end_position_on_the_genomic_accession=85064241, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20533, CurrentID=0, Status='live', Symbol='Slc4a1', Aliases='Ae1, CD233, Empb3, l11Jus51', description='solute carrier family 4 (anion exchanger), member 1', other_designations='AE 1|MEB3|anion exchange protein 1|anion exchanger 1|erythrocyte membrane protein band 3|solute carrier family 4 member 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102348820, end_position_on_the_genomic_accession=102365281, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  20534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20534, CurrentID=0, Status='live', Symbol='Slc4a1ap', Aliases='kanadaptin', description='solute carrier family 4 (anion exchanger), member 1, adaptor protein', other_designations='solute carrier family 4 (anion exchanger), member 1, adapter protein', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31526850, end_position_on_the_genomic_accession=31554038, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20535, CurrentID=0, Status='live', Symbol='Slc4a2', Aliases='Ae2, B3RP', description='solute carrier family 4 (anion exchanger), member 2', other_designations='AE 2|anion exchanger 2 type b1|anion exchanger 2 type b2|anion exchanger 2 type c1|anion exchanger 2 type c2|band 3-related protein|non-erythroid band 3-like protein|solute carrier family 4 member 2', map_location='5 11.74 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24423761, end_position_on_the_genomic_accession=24440947, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  20536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20536, CurrentID=0, Status='live', Symbol='Slc4a3', Aliases='A930038D23Rik, Ae3', description='solute carrier family 4 (anion exchanger), member 3', other_designations='AE 3|anion exchanger 3 cardiac isoform|neuronal band 3-like protein|solute carrier family 4 member 3', map_location='1 39.16 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75546041, end_position_on_the_genomic_accession=75559431, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20537, CurrentID=0, Status='live', Symbol='Slc5a1', Aliases='Sglt1', description='solute carrier family 5 (sodium/glucose cotransporter), member 1', other_designations='Na(+)/glucose cotransporter 1|high affinity sodium-glucose cotransporter|sodium glucose cotransporter 1|sodium-glucose cotransporter 1|solute carrier family 5 member 1|solute carrier family 5, member 1|solute carrier family 5, member alpha 1', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33104219, end_position_on_the_genomic_accession=33162699, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20538, CurrentID=0, Status='live', Symbol='Slc6a2', Aliases='NE-T, NET, Slc6a5', description='solute carrier family 6 (neurotransmitter transporter, noradrenalin), member 2', other_designations='NE transporter|norepinephrine transporter|solute carrier family 6 member 2', map_location='8 44.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=92960436, end_position_on_the_genomic_accession=93001667, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20539, CurrentID=0, Status='live', Symbol='Slc7a5', Aliases='4F2LC, D0H16S474E, LAT1, TA1', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 5', other_designations='4F2 LC|4F2 light chain|L-type amino acid transporter 1|solute carrier family 7 member 5', map_location='8 70.8 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121881146, end_position_on_the_genomic_accession=121907686, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20540, CurrentID=0, Status='live', Symbol='Slc7a7', Aliases='AI790233, my+lat1', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 7', other_designations='solute carrier family 7 member 7|y(+)L-type amino acid transporter 1|y+LAT-1|y+LAT1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54369442, end_position_on_the_genomic_accession=54417702, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  20541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20541, CurrentID=0, Status='live', Symbol='Slc8a1', Aliases='AI852629, AV344025, D930008O12Rik, Ncx1', description='solute carrier family 8 (sodium/calcium exchanger), member 1', other_designations='Na(+)/Ca(2+)-exchange protein 1|Na/Ca exchanger|sodium/calcium exchanger isoform NaCa7|solute carrier family 8 member 1', map_location='17 51.29 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=81373105, end_position_on_the_genomic_accession=81738387, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  20544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20544, CurrentID=0, Status='live', Symbol='Slc9a1', Aliases='AW554487, Apnh, Mir5122, Nhe1, mir-5122, swe', description='solute carrier family 9 (sodium/hydrogen exchanger), member 1', other_designations='NHE-1|Na(+)/H(+) exchanger 1|Na+/H+ antiporter|Na+/H+, amiloride sensitive|slow-wave epilepsy|solute carrier family 9 member 1|solute carrier family 9, member 1', map_location='4 D3-E', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133369772, end_position_on_the_genomic_accession=133423702, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  20555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20555, CurrentID=0, Status='live', Symbol='Slfn1', Aliases='AV316259', description='schlafen 1', other_designations='', map_location='11 50.3 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83116845, end_position_on_the_genomic_accession=83122659, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20556, CurrentID=0, Status='live', Symbol='Slfn2', Aliases='Shlf2', description='schlafen 2', other_designations='', map_location='11 50.3 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83065112, end_position_on_the_genomic_accession=83070678, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20557, CurrentID=0, Status='live', Symbol='Slfn3', Aliases='', description='schlafen 3', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83191301, end_position_on_the_genomic_accession=83218427, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20558, CurrentID=0, Status='live', Symbol='Slfn4', Aliases='AI450778', description='schlafen 4', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83175171, end_position_on_the_genomic_accession=83190216, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20562, CurrentID=0, Status='live', Symbol='Slit1', Aliases='Slil1, mKIAA0813', description='slit homolog 1 (Drosophila)', other_designations='slit-1', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41597824, end_position_on_the_genomic_accession=41747275, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  20563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20563, CurrentID=0, Status='live', Symbol='Slit2', Aliases='Drad-1, E030015M03Rik, E130320P19Rik, Slil3, mKIAA4141, slit-2', description='slit homolog 2 (Drosophila)', other_designations='downregulated during adipocyte differentiation-1|neurogenic extracellular slit protein', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=47983108, end_position_on_the_genomic_accession=48306282, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  20564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20564, CurrentID=0, Status='live', Symbol='Slit3', Aliases='Slil2, Slit1, b2b2362.1Clo', description='slit homolog 3 (Drosophila)', other_designations='slit-3', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=35121456, end_position_on_the_genomic_accession=35708507, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  20568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20568, CurrentID=0, Status='live', Symbol='Slpi', Aliases='', description='secretory leukocyte peptidase inhibitor', other_designations='ALP|secretory leukocyte protease inhibitor', map_location='2 H', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164354070, end_position_on_the_genomic_accession=164356507, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20583, CurrentID=0, Status='live', Symbol='Snai2', Aliases='Slug, Slugh, Snail2', description='snail family zinc finger 2', other_designations='neural crest transcription factor Slug|protein snail homolog 2|slug, chicken homolog|snail homolog 2', map_location='16 10.07 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14705859, end_position_on_the_genomic_accession=14709385, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20585, CurrentID=0, Status='live', Symbol='Hltf', Aliases='AF010600, BC057116, P113, Smarca3, Snf2l3', description='helicase-like transcription factor', other_designations='SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily A member 3|TNF-response element-binding protein|sucrose nonfermenting protein 2-like 3', map_location='3 6.15 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=20057811, end_position_on_the_genomic_accession=20118491, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20586, CurrentID=0, Status='live', Symbol='Smarca4', Aliases='Brg1, HP1-BP72, SNF2beta, SW1/SNF, b2b508.1Clo, b2b692Clo', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 4', other_designations='ATP-dependent helicase SMARCA4|BAF190A|BRG1-associated factor 190A|SNF2-beta|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily A member 4|protein brahma homolog 1', map_location='9 7.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21616106, end_position_on_the_genomic_accession=21704230, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  20587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20587, CurrentID=0, Status='live', Symbol='Smarcb1', Aliases='AU020204, Baf47, Ini1, SNF5/INI1, Snf5', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily b, member 1', other_designations='BRG1-associated factor 47|SNF5 homolog|integrase interactor 1 protein|mSNF5', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75896769, end_position_on_the_genomic_accession=75921614, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20588, CurrentID=0, Status='live', Symbol='Smarcc1', Aliases='AI115498, BAF155, Rsc8, SRG3, msp3', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily c, member 1', other_designations='BRG1-associated factor 155|SWI/SNF complex 155 kDa subunit|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily C member 1|SWI3-related protein', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110132024, end_position_on_the_genomic_accession=110240178, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  20589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20589, CurrentID=0, Status='live', Symbol='Ighmbp2', Aliases='AEP, Catf1, RIPE3b1, Smbp-2, Smbp2, Smubp2, nmd, sma', description='immunoglobulin mu binding protein 2', other_designations='ATP-dependent helicase IGHMBP2|antifreeze enhancer-binding protein|cardiac transcription factor 1|immunoglobulin S mu binding protein 2|immunoglobulin mu-binding protein 2|neuromuscular degeneration|p110 subunit', map_location='19 3.03 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3259076, end_position_on_the_genomic_accession=3283010, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20590, CurrentID=0, Status='live', Symbol='smc', Aliases='', description='spondylo-metaphyseal chondrodysplasia', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20591, CurrentID=0, Status='live', Symbol='Kdm5c', Aliases='D930009K15Rik, Jarid1c, Smcx, mKIAA0234', description='lysine (K)-specific demethylase 5C', other_designations='histone demethylase JARID1C|jumonji, AT rich interactive domain 1C (Rbp2 like)|jumonji/ARID domain-containing protein 1C|selected mouse cDNA on the X', map_location='X F2-F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152233196, end_position_on_the_genomic_accession=152279102, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20592, CurrentID=0, Status='live', Symbol='Kdm5d', Aliases='HY, Jarid1d, Smcy', description='lysine (K)-specific demethylase 5D', other_designations='H-Y|histocompatibility Y antigen|histone demethylase JARID1D|jumonji, AT rich interactive domain 1D (Rbp2 like)|jumonji/ARID domain-containing protein 1D|protein SmcY|selected mouse cDNA on the Y', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=897660, end_position_on_the_genomic_accession=943813, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  544881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544881, CurrentID=0, Status='live', Symbol='BB287469', Aliases='', description='expressed sequence BB287469', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87815519, end_position_on_the_genomic_accession=87820676, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20595, CurrentID=0, Status='live', Symbol='Smn1', Aliases='AI849087, Gemin1, Smn', description='survival motor neuron 1', other_designations='survival of motor neuron protein', map_location='13 D1/D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100123205, end_position_on_the_genomic_accession=100137699, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20597, CurrentID=0, Status='live', Symbol='Smpd1', Aliases='A-SMase, ASM, Zn-SMase, aSMase', description='sphingomyelin phosphodiesterase 1, acid lysosomal', other_designations='acid sphingomyelinase', map_location='7 55.9 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105554360, end_position_on_the_genomic_accession=105558389, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20598, CurrentID=0, Status='live', Symbol='Smpd2', Aliases='AW108287, nSMase, nSMase1', description='sphingomyelin phosphodiesterase 2, neutral', other_designations='N-SMase|lyso-PAF-PLC|lyso-platelet-activating factor-phospholipase C|neutral sphingomyelinase 1', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41487172, end_position_on_the_genomic_accession=41490340, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20599, CurrentID=0, Status='live', Symbol='Smr3a', Aliases='AA589601, MSG1, MSG3, Smr1, Smr3', description='submaxillary gland androgen regulated protein 3A', other_designations='salivary protein MSG1|submaxillary gland androgen regulated protein 1|submaxillary gland androgen regulated protein 3|submaxillary gland androgen-regulated protein 1', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87969504, end_position_on_the_genomic_accession=88008534, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20600, CurrentID=0, Status='live', Symbol='Smr2', Aliases='MSG2, Vcs2', description='submaxillary gland androgen regulated protein 2', other_designations='salivary protein MSG2', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88086556, end_position_on_the_genomic_accession=88109053, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20602, CurrentID=0, Status='live', Symbol='Ncor2', Aliases='N-CoR, SMRT, SMRTe', description='nuclear receptor co-repressor 2', other_designations='N-CoR2|T3 receptor-associating factor|TRAC|silencing mediator of retinoic acid and thyroid hormone receptor|silencing mediator of retinoid acid and thyroid hormone receptor|thyroid-, retinoic-acid-receptor-associated corepressor', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=125017153, end_position_on_the_genomic_accession=125179219, orientation='minus', exon_count=52, OMIM='', no_hdr0=''),
  20603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20603, CurrentID=0, Status='live', Symbol='Sms', Aliases='AI427066, Gy, SPMSY, SpmST, gyro', description='spermine synthase', other_designations='spermidine aminopropyltransferase', map_location='X 72.48 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=157443852, end_position_on_the_genomic_accession=157492046, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20604, CurrentID=0, Status='live', Symbol='Sst', Aliases='SOM, SRIF, SS, Smst', description='somatostatin', other_designations='preprosomatostatin', map_location='16 cen-C3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23889581, end_position_on_the_genomic_accession=23890844, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20605, CurrentID=0, Status='live', Symbol='Sstr1', Aliases='SRIF-2, SS-1-R, SS1-R, SS1R, Smstr-1, Smstr1, sst1', description='somatostatin receptor 1', other_designations='', map_location='12 25.61 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=58211387, end_position_on_the_genomic_accession=58216036, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20606, CurrentID=0, Status='live', Symbol='Sstr2', Aliases='SRIF-1, SS2R, SSTR-2, Smstr-2, Smstr2, sst2', description='somatostatin receptor 2', other_designations='', map_location='11 79.05 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113619481, end_position_on_the_genomic_accession=113625895, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20607, CurrentID=0, Status='live', Symbol='Sstr3', Aliases='Smstr-3, Smstr3, sst3', description='somatostatin receptor 3', other_designations='SS-3-R|SS3-R|SS3R|SSR-28', map_location='15 37.55 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78537014, end_position_on_the_genomic_accession=78549369, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  544896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544896, CurrentID=0, Status='live', Symbol='Ighv5-9', Aliases='Gm16949, VH7183', description='immunoglobulin heavy variable 5-9', other_designations='Ig heavy chain V region|VH7183-DSP2-JH3-CH1|immunoglobulin heavy variable V5-9', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113661771, end_position_on_the_genomic_accession=113662228, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  20609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20609, CurrentID=0, Status='live', Symbol='Sstr5', Aliases='Smstr5, sst5', description='somatostatin receptor 5', other_designations='SS-5-R|SS5-R|SS5R', map_location='17 12.62 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25489875, end_position_on_the_genomic_accession=25514304, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20610, CurrentID=0, Status='live', Symbol='Sumo3', Aliases='2810014B19Rik, D10Ertd345e, SMT3A, SUMO-3, Smt3h1', description='small ubiquitin-like modifier 3', other_designations='SMT3 (supressor of mif two, 3) homolog 1|SMT3 suppressor of mif two 3 homolog 3|ubiquitin-like protein SMT3A', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77606097, end_position_on_the_genomic_accession=77618331, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20611, CurrentID=0, Status='live', Symbol='Ssty1', Aliases='Smy, Ssty', description='spermiogenesis specific transcript on the Y 1', other_designations='', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=18089167, end_position_on_the_genomic_accession=18091551, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20612, CurrentID=0, Status='live', Symbol='Siglec1', Aliases='Cd169, Siglec-1, Sn', description='sialic acid binding Ig-like lectin 1, sialoadhesin', other_designations='SER|sheep erythrocyte receptor|sialic acid-binding Ig-like lectin 1', map_location='2 F-H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131069220, end_position_on_the_genomic_accession=131086787, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  20613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20613, CurrentID=0, Status='live', Symbol='Snai1', Aliases='Sna, Sna1, Snail, Snail1', description='snail family zinc finger 1', other_designations='snail homolog 1', map_location='2 87.33 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167538227, end_position_on_the_genomic_accession=167542814, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20614, CurrentID=0, Status='live', Symbol='Snap25', Aliases='Bdr, GENA70, SNAP-25, SUP, sp', description='synaptosomal-associated protein 25', other_designations='blind drunk|super protein|synaptosomal-associated 25 kDa protein|synaptosomal-associated protein, 25 kDa', map_location='2 67.56 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=136713438, end_position_on_the_genomic_accession=136782428, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20615, CurrentID=0, Status='live', Symbol='Snapin', Aliases='25kDa, AA407989, AV026596, Bloc1s7, Snap25bp, Snapap', description='SNAP-associated protein', other_designations='BLOC-1 subunit 7|biogenesis of lysosome-related organelles complex 1 subunit 7|synaptosomal-associated protein 25 binding protein|synaptosomal-associated protein 25-binding protein|synaptosomal-associated protein, 25 kDa, binding protein', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90488026, end_position_on_the_genomic_accession=90491013, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20616, CurrentID=0, Status='live', Symbol='Snap91', Aliases='91kDa, AP180, F1-20, mKIAA0656', description='synaptosomal-associated protein 91', other_designations='91 kDa synaptosomal-associated protein|clathrin coat-associated protein AP180|phosphoprotein F1-20|synaptosomal-associated protein, 91 kDa', map_location='9 E3.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=86765923, end_position_on_the_genomic_accession=86880708, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  20617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20617, CurrentID=0, Status='live', Symbol='Snca', Aliases='NACP, alphaSYN', description='synuclein, alpha', other_designations='non-A beta component of AD amyloid|non-A4 component of amyloid', map_location='6 29.15 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=60731573, end_position_on_the_genomic_accession=60829855, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20618, CurrentID=0, Status='live', Symbol='Sncg', Aliases='C79089, persyn', description='synuclein, gamma', other_designations='', map_location='14 B-C', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34370274, end_position_on_the_genomic_accession=34374669, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20619, CurrentID=0, Status='live', Symbol='Snap23', Aliases='23kDa, AA408749, SNAP-23, Sndt, Syndet', description='synaptosomal-associated protein 23', other_designations='synaptosomal-associated protein, 23kD|vesicle-membrane fusion protein SNAP-23', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120567651, end_position_on_the_genomic_accession=120601255, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20620, CurrentID=0, Status='live', Symbol='Plk2', Aliases='Snk', description='polo-like kinase 2', other_designations='PLK-2|serine/threonine-protein kinase SNK|serum-inducible kinase', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=110395044, end_position_on_the_genomic_accession=110400844, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20621, CurrentID=0, Status='live', Symbol='Snn', Aliases='2810407J07Rik, AI848521, AW547286', description='stannin', other_designations='', map_location='16 A2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11066298, end_position_on_the_genomic_accession=11074985, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20623, CurrentID=0, Status='live', Symbol='Snrk', Aliases='2010012F07Rik, AI448042, AW547029, E030034B15, R74830, mKIAA0096', description='SNF related kinase', other_designations='SNF1-related kinase|snrk', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122117266, end_position_on_the_genomic_accession=122169707, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20624, CurrentID=0, Status='live', Symbol='Eftud2', Aliases='116kDa, Snrp116, U5-116kD', description='elongation factor Tu GTP binding domain containing 2', other_designations='U5 small nuclear ribonucleoprotein 116 kDa|U5 snRNP-specific protein, 116 kDa|U5-116 kDa|elongation factor Tu GTP-binding domain-containing protein 2', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102838472, end_position_on_the_genomic_accession=102880996, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  20630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20630, CurrentID=0, Status='live', Symbol='Snrpc', Aliases='Snrp1c, U1-C, U1C', description='U1 small nuclear ribonucleoprotein C', other_designations='U1 small nuclear ribonucleoprotein 1C|U1 snRNP C|U1 snRNP protein C|U1-C|U1C', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27840087, end_position_on_the_genomic_accession=27851968, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20637, CurrentID=0, Status='live', Symbol='Snrnp70', Aliases='2700022N21Rik, 3200002N22Rik, AI325098, R74807, Rnu1p70, Rnulp70, Snrp70, Srnp70, U1-70, U1-70K', description='small nuclear ribonucleoprotein 70 (U1)', other_designations='U1 small nuclear ribonucleoprotein 70 kDa polypeptide A|U1 snRNP 70 kDa', map_location='7 29.28 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45376454, end_position_on_the_genomic_accession=45395742, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  20638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20638, CurrentID=0, Status='live', Symbol='Snrpb', Aliases='AL024368, AU018828, SM-B, SM11, SMB, SNRNP-B', description='small nuclear ribonucleoprotein B', other_designations='sm protein B', map_location='2 63.19 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130171636, end_position_on_the_genomic_accession=130179376, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20639, CurrentID=0, Status='live', Symbol='Snrpb2', Aliases='2810052G09Rik', description='U2 small nuclear ribonucleoprotein B', other_designations="U2 snRNP B''", map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=143063069, end_position_on_the_genomic_accession=143072052, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20641, CurrentID=0, Status='live', Symbol='Snrpd1', Aliases='AA407109, AL023031, SMD1', description='small nuclear ribonucleoprotein D1', other_designations='sm-D autoantigen|sm-D1|snRNP core protein D1', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10617796, end_position_on_the_genomic_accession=10628230, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20643, CurrentID=0, Status='live', Symbol='Snrpe', Aliases='AL022645, C76690, SME', description='small nuclear ribonucleoprotein E', other_designations='sm protein E|sm-E|snRNP-E', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133603871, end_position_on_the_genomic_accession=133610280, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20646, CurrentID=0, Status='live', Symbol='Snrpn', Aliases='2410045I01Rik, HCERN3, Peg4, Pwcr1, SMN', description='small nuclear ribonucleoprotein N', other_designations='sm protein D|sm protein N|sm-D|sm-N|snRNP-N|tissue-specific-splicing protein', map_location='7 34.04 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=59982502, end_position_on_the_genomic_accession=60140219, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  20648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20648, CurrentID=0, Status='live', Symbol='Snta1', Aliases='AW228934, Snt1', description='syntrophin, acidic 1', other_designations='59 kDa dystrophin-associated protein A1 acidic component 1|alpha1-syntrophin|syntrophin-1', map_location='2 76.52 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154376314, end_position_on_the_genomic_accession=154408084, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20649, CurrentID=0, Status='live', Symbol='Sntb1', Aliases='', description='syntrophin, basic 1', other_designations='59 kDa dystrophin-associated protein A1 basic component 1|59-1 DAP|DAPA1B|beta1-Syntrophin|syntrophin-2', map_location='15 22.14 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=55638505, end_position_on_the_genomic_accession=55907305, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20650, CurrentID=0, Status='live', Symbol='Sntb2', Aliases='Snt2', description='syntrophin, basic 2', other_designations='59 kDa dystrophin-associated protein A1 basic component 2|SNT3|SNTL|syntrophin-3', map_location='8 53.38 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106935750, end_position_on_the_genomic_accession=107014192, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20652, CurrentID=0, Status='live', Symbol='Soat1', Aliases='8430426K15Rik, ACAT-1, AW550831, Acact, ald, hid', description='sterol O-acyltransferase 1', other_designations='acyl coenzyme A:cholesterol acyltransferase 1|acyl-coenzyme A:cholesterol acyltransferase 1|adrenocortical lipid depletion|cholesterol acyltransferase 1', map_location='1 67.71 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156428108, end_position_on_the_genomic_accession=156474328, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  20654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20654, CurrentID=0, Status='live', Symbol='soc', Aliases='', description='soft coat', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20655, CurrentID=0, Status='live', Symbol='Sod1', Aliases='B430204E11Rik, Cu/Zn-SOD, CuZnSOD, Ipo-1, Ipo1, SODC, Sod-1', description='superoxide dismutase 1, soluble', other_designations='Cu(2+)-Zn2+ superoxide dismutase', map_location='16 B5-C3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90220742, end_position_on_the_genomic_accession=90226324, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20656, CurrentID=0, Status='live', Symbol='Sod2', Aliases='MnSOD, Sod-2', description='superoxide dismutase 2, mitochondrial', other_designations='manganese SOD|manganese superoxide dismutase', map_location='17 8.75 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13007839, end_position_on_the_genomic_accession=13018119, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20657, CurrentID=0, Status='live', Symbol='Sod3', Aliases='AI314465, EC-SOD', description='superoxide dismutase 3, extracellular', other_designations='extracellular superoxide dismutase', map_location='5 27.92 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52363804, end_position_on_the_genomic_accession=52369738, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20658, CurrentID=0, Status='live', Symbol='Son', Aliases='2900011L12Rik, AA409051, AU067731, C81487, mKIAA1019, nrebp', description='Son DNA binding protein', other_designations='NRE-binding protein|Son cell proliferation protein|negative regulatory element-binding protein', map_location='16 53.22 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91647824, end_position_on_the_genomic_accession=91679195, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  20660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20660, CurrentID=0, Status='live', Symbol='Sorl1', Aliases='2900010L19Rik, AI596264, AW261561, LR11, SorLA, gp250, mSorLA', description='sortilin-related receptor, LDLR class A repeats-containing', other_designations='LDLR relative with 11 ligand-binding repeats|low-density lipoprotein receptor relative with 11 ligand-binding repeats|sorLA-1|sorting protein-related receptor containing LDLR class A repeats', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=41968489, end_position_on_the_genomic_accession=42124289, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  20661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20661, CurrentID=0, Status='live', Symbol='Sort1', Aliases='2900053A11Rik, AI852375, Ntr3, Ntsr3', description='sortilin 1', other_designations='mNTR3|neurotensin receptor 3', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108283026, end_position_on_the_genomic_accession=108361519, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  20662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20662, CurrentID=0, Status='live', Symbol='Sos1', Aliases='4430401P03Rik, 9630010N06, AI449023', description='son of sevenless homolog 1 (Drosophila)', other_designations='guanine nucleotide releasing factor|ras guanine exchange factor|son-of-sevenless 1', map_location='17 50.67 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80393752, end_position_on_the_genomic_accession=80480453, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  20663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20663, CurrentID=0, Status='live', Symbol='Sos2', Aliases='SOS-2, mSOS-2', description='son of sevenless homolog 2 (Drosophila)', other_designations='', map_location='12 C3-D', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69583761, end_position_on_the_genomic_accession=69681852, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  20664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20664, CurrentID=0, Status='live', Symbol='Sox1', Aliases='BB176347, Sox-1', description='SRY (sex determining region Y)-box 1', other_designations='SRY-box containing gene 1', map_location='8 A1-A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=12395519, end_position_on_the_genomic_accession=12399555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  20665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20665, CurrentID=0, Status='live', Symbol='Sox10', Aliases='Dom, Sox21', description='SRY (sex determining region Y)-box 10', other_designations='SRY-box containing gene 10|dominant megacolon|transcription factor SOX-M', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79154909, end_position_on_the_genomic_accession=79165235, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20666, CurrentID=0, Status='live', Symbol='Sox11', Aliases='1110038H03Rik, 6230403H02Rik, AI836553, end1', description='SRY (sex determining region Y)-box 11', other_designations='SRY-box containing gene 11', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=27334268, end_position_on_the_genomic_accession=27342718, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  20667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20667, CurrentID=0, Status='live', Symbol='Sox12', Aliases='2010205A02Rik', description='SRY (sex determining region Y)-box 12', other_designations='SRY-box containing gene 12', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152393611, end_position_on_the_genomic_accession=152398046, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  20668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20668, CurrentID=0, Status='live', Symbol='Sox13', Aliases='AA407916, AW259412, AW540933, Sox-13', description='SRY (sex determining region Y)-box 13', other_designations='SRY-box containing gene 13|mSox13', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133382300, end_position_on_the_genomic_accession=133424212, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  20669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20669, CurrentID=0, Status='live', Symbol='Sox14', Aliases='', description='SRY (sex determining region Y)-box 14', other_designations='SRY-box containing gene 14', map_location='9 52.11 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99874106, end_position_on_the_genomic_accession=99876170, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  20670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20670, CurrentID=0, Status='live', Symbol='Sox15', Aliases='', description='SRY (sex determining region Y)-box 15', other_designations='SRY-box containing gene 15', map_location='11 42.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69655037, end_position_on_the_genomic_accession=69656727, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20671, CurrentID=0, Status='live', Symbol='Sox17', Aliases='', description='SRY (sex determining region Y)-box 17', other_designations='SRY-box containing gene 17', map_location='1 1.65 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=4490928, end_position_on_the_genomic_accession=4497388, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20672, CurrentID=0, Status='live', Symbol='Sox18', Aliases='AI385749, Ra, Ragl', description='SRY (sex determining region Y)-box 18', other_designations='SRY-box containing gene 18|Sry-related HMG-box gene 18|ragged', map_location='2 103.71 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181669837, end_position_on_the_genomic_accession=181671640, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20674, CurrentID=0, Status='live', Symbol='Sox2', Aliases='Sox-2, lcc, ysb', description='SRY (sex determining region Y)-box 2', other_designations='SRY-box containing gene 2', map_location='3 A2-B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=34650005, end_position_on_the_genomic_accession=34652461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  20675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20675, CurrentID=0, Status='live', Symbol='Sox3', Aliases='Sox-3', description='SRY (sex determining region Y)-box 3', other_designations='SRY-box containing gene 3', map_location='X A7.3-B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=60891366, end_position_on_the_genomic_accession=60893430, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  20677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20677, CurrentID=0, Status='live', Symbol='Sox4', Aliases='AA682046, Sox-4', description='SRY (sex determining region Y)-box 4', other_designations='SRY-box containing gene 4', map_location='13 A3-A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=28950716, end_position_on_the_genomic_accession=28953682, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  20678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20678, CurrentID=0, Status='live', Symbol='Sox5', Aliases='A730017D01Rik, AI528773', description='SRY (sex determining region Y)-box 5', other_designations='SRY-box containing gene 5|Sox5-BLM isoform', map_location='6 76.14 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=143828425, end_position_on_the_genomic_accession=144781989, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  20679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20679, CurrentID=0, Status='live', Symbol='Sox6', Aliases='AI987981, SOX-LZ', description='SRY (sex determining region Y)-box 6', other_designations='SRY-box containing gene 6', map_location='7 61.29 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=115470872, end_position_on_the_genomic_accession=116040484, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  20680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20680, CurrentID=0, Status='live', Symbol='Sox7', Aliases='', description='SRY (sex determining region Y)-box 7', other_designations='SRY-box containing gene 7|mSOX7', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63943706, end_position_on_the_genomic_accession=63950732, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20681, CurrentID=0, Status='live', Symbol='Sox8', Aliases='', description='SRY (sex determining region Y)-box 8', other_designations='SRY-box containing gene 8', map_location='17 12.69 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25565893, end_position_on_the_genomic_accession=25570686, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20682, CurrentID=0, Status='live', Symbol='Sox9', Aliases='2010306G03Rik, AV220920, mKIAA4243', description='SRY (sex determining region Y)-box 9', other_designations='SRY-box containing gene 9', map_location='11 77.27 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=112782210, end_position_on_the_genomic_accession=112787757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  544971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544971, CurrentID=0, Status='live', Symbol='Bdp1', Aliases='AI662220, AW049390, AW456745, B130055N23Rik, G630013P12Rik, TAF3B1, TFC5, TFIIIB150, TFIIIB90, Tfnr, mKIAA1241', description='B double prime 1, subunit of RNA polymerase III transcription initiation factor IIIB', other_designations='transcription factor IIIB 150|transcription factor-like nuclear regulator', map_location='13 52.92 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100017994, end_position_on_the_genomic_accession=100104073, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  20684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20684, CurrentID=0, Status='live', Symbol='Sp100', Aliases='A430075G10Rik', description='nuclear antigen Sp100', other_designations='nuclear dot-associated Sp100 protein|speckled 100 kDa', map_location='1 43.6 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85649840, end_position_on_the_genomic_accession=85709998, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  20685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20685, CurrentID=0, Status='live', Symbol='Sp100-rs1', Aliases='D1Lub1', description='nuclear antigen Sp100, related sequence 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20686, CurrentID=0, Status='live', Symbol='Spa17', Aliases='Sp17', description='sperm autoantigenic protein 17', other_designations='band 34', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37603294, end_position_on_the_genomic_accession=37613833, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20687, CurrentID=0, Status='live', Symbol='Sp3', Aliases='D130027J01Rik', description='trans-acting transcription factor 3', other_designations='', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=72936430, end_position_on_the_genomic_accession=72980446, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20688, CurrentID=0, Status='live', Symbol='Sp4', Aliases='5730497N03Rik, HF-1b, HF1-b', description='trans-acting transcription factor 4', other_designations='', map_location='12 63.48 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=118231686, end_position_on_the_genomic_accession=118301440, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20689, CurrentID=0, Status='live', Symbol='Sall3', Aliases='B130022O04Rik, Msal, Msal-1, Sal, Salt, Spalt', description='sal-like 3 (Drosophila)', other_designations='spalt-like protein 3', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80966359, end_position_on_the_genomic_accession=80986578, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20690, CurrentID=0, Status='live', Symbol='Spam1', Aliases='AV039194, Ph-20', description='sperm adhesion molecule 1', other_designations='hyal-PH20|hyaluronoglucosaminidase PH-20|sperm surface protein PH-20', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24791188, end_position_on_the_genomic_accession=24801048, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20692, CurrentID=0, Status='live', Symbol='Sparc', Aliases='BM-40, ON', description='secreted acidic cysteine rich glycoprotein', other_designations='basement-membrane protein 40|osteonectin|secreted protein acidic and rich in cysteine', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55394159, end_position_on_the_genomic_accession=55420080, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20698, CurrentID=0, Status='live', Symbol='Sphk1', Aliases='1110006G24Rik, Sk1, Spk1', description='sphingosine kinase 1', other_designations='SK 1|SPK 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116529817, end_position_on_the_genomic_accession=116536675, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20700, CurrentID=0, Status='live', Symbol='Serpina1a', Aliases='Aat-2, Aat2, Dom1, PI1, Spi1-1, Spi1-3', description='serine (or cysteine) peptidase inhibitor, clade A, member 1A', other_designations='AAT|alpha-1 protease inhibitor 1|alpha-1-antiproteinase|alpha-1-antitrypsin 2|serine (or cysteine) proteinase inhibitor, clade A, member 1a|serine protease inhibitor 1-1|serine protease inhibitor A1a|serpin A1a', map_location='12 52.98 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103853295, end_position_on_the_genomic_accession=103863619, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20701, CurrentID=0, Status='live', Symbol='Serpina1b', Aliases='D12Ucla2, Dom2, PI2, Spi1-2', description='serine (or cysteine) preptidase inhibitor, clade A, member 1B', other_designations='AAT|alpha-1 protease inhibitor 2|alpha-1-antiproteinase|serine (or cysteine) proteinase inhibitor, clade A, member 1b|serine protease inhibitor 1-2|serine protease inhibitor A1b|serpin A1b', map_location='12 52.98 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103728156, end_position_on_the_genomic_accession=103738189, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20702, CurrentID=0, Status='live', Symbol='Serpina1c', Aliases='Pi3, Pi6, Skalp, Spi1-3, Spi1-6, Wap3', description='serine (or cysteine) peptidase inhibitor, clade A, member 1C', other_designations='alpha-1 protease inhibitor 3|alpha-1 protease inhibitor 6|alpha-1-antitrypsin 1-6|apha-1 protease inhibitor 3|pre-elafin|serine (or cysteine) proteinase inhibitor, clade A, member 1c|serine protease inhibitor 1-3|serine protease inhibitor 1-6|serine protease inhibitor A1c|serpin A1c|trappin-2', map_location='12 52.98 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103894926, end_position_on_the_genomic_accession=103904950, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20703, CurrentID=0, Status='live', Symbol='Serpina1d', Aliases='AI118301, Dom4, PI4, Spi1-4', description='serine (or cysteine) peptidase inhibitor, clade A, member 1D', other_designations='alpha-1 protease inhibitor 4|serine (or cysteine) proteinase inhibitor, clade A, member 1d|serine protease inhibitor 1-4|serine protease inhibitor A1d|serpin A1d', map_location='12 52.98 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103763587, end_position_on_the_genomic_accession=103773633, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20704, CurrentID=0, Status='live', Symbol='Serpina1e', Aliases='Dom5, PI5, Spi1-5', description='serine (or cysteine) peptidase inhibitor, clade A, member 1E', other_designations='alpha-1 protease inhibitor 5|serine (or cysteine) proteinase inhibitor, clade A, member 1e|serine protease inhibitor 1-5|serine protease inhibitor A1e|serpin A1e', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103946931, end_position_on_the_genomic_accession=103956897, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20706, CurrentID=0, Status='live', Symbol='Serpinb9b', Aliases='1600019A21Rik, R86, SPI-CI, Spi10, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 9b', other_designations='serine protease inhibitor 10|serine proteinase inhibitor R86', map_location='13 13.79 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33027414, end_position_on_the_genomic_accession=33040559, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20707, CurrentID=0, Status='live', Symbol='Serpinb9c', Aliases='3830421J05Rik, Nk9, Spi11', description='serine (or cysteine) peptidase inhibitor, clade B, member 9c', other_designations='ovalbumin|serine protease inhibitor 11', map_location='13 13.82 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33149275, end_position_on_the_genomic_accession=33159754, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20708, CurrentID=0, Status='live', Symbol='Serpinb6b', Aliases='NK13, Spi12, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 6b', other_designations='serine protease inhibitor 12', map_location='13 13.78 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=32956809, end_position_on_the_genomic_accession=32979067, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20709, CurrentID=0, Status='live', Symbol='Serpinb9f', Aliases='BC052216, NK21, NK21L1, Spi13, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 9f', other_designations='serine protease inhibitor 13', map_location='13 13.86 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33324077, end_position_on_the_genomic_accession=33335370, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20710, CurrentID=0, Status='live', Symbol='Serpinb9e', Aliases='AW540195, NK26, Spi14, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 9e', other_designations='serine protease inhibitor 14', map_location='13 13.84 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33249610, end_position_on_the_genomic_accession=33260846, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20713, CurrentID=0, Status='live', Symbol='Serpini1', Aliases='AI837402, Ns, PI-12, PI12, Spi17', description='serine (or cysteine) peptidase inhibitor, clade I, member 1', other_designations='peptidase inhibitor 12|protease inhibitor 17|serine protease inhibitor 17|serpin I1', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=75557503, end_position_on_the_genomic_accession=75642523, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20714, CurrentID=0, Status='live', Symbol='Serpina3k', Aliases='1300001I07Rik, D12Rp54, MMCM2, MMSpi2, RP54, Spi-2, Spi2, contrapsin', description='serine (or cysteine) peptidase inhibitor, clade A, member 3K', other_designations='alpha-1 antiproteinase|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 3K|serine (or cysteine) proteinase inhibitor, clade A, member 3K|serine protease inhibitor 2|serpin A3K', map_location='12 53.99 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104338486, end_position_on_the_genomic_accession=104345739, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20715, CurrentID=0, Status='live', Symbol='Serpina3g', Aliases='2A2, AI119734, Spi2-1, Spi2/eb.1, Spi2A, spi2', description='serine (or cysteine) peptidase inhibitor, clade A, member 3G', other_designations='alpha-1 antiproteinase|serine (or cysteine) proteinase inhibitor, clade A, member 3G|serine protease inhibitor 2-1|serine protease inhibitor 2A|serpin 2A|serpin A3G', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104236242, end_position_on_the_genomic_accession=104241934, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20716, CurrentID=0, Status='live', Symbol='Serpina3n', Aliases='Spi2-2, Spi2.2, Spi2/eb.4', description='serine (or cysteine) peptidase inhibitor, clade A, member 3N', other_designations='alpha-1 antiproteinase|antitrypsin|serine (or cysteine) proteinase inhibitor, clade A, member 3N|serine protease inhibitor 2-2|serpin A3N', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104406675, end_position_on_the_genomic_accession=104414334, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20717, CurrentID=0, Status='live', Symbol='Serpina3m', Aliases='3e46, AI195004, MMCM7, MMSPi2.4, Spi-2l, Spi-2rs1, Spi2-rs1, Spi2.4, spi2', description='serine (or cysteine) peptidase inhibitor, clade A, member 3M', other_designations='alpha-1 antiproteinase|antitrypsin|serine (or cysteine) proteinase inhibitor, clade A, member 3M|serine protease inhibitor-2 related sequence 1|serpin A3M', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104387157, end_position_on_the_genomic_accession=104394259, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20719, CurrentID=0, Status='live', Symbol='Serpinb6a', Aliases='4930482L21Rik, D330015H01Rik, PI-6, Serpinb6, Spi3, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 6a', other_designations='placental thrombin inhibitor|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 6|serine (or cysteine) proteinase inhibitor, clade B, member 6a|serine protease inhibitor 3', map_location='13 14.0 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33917918, end_position_on_the_genomic_accession=34002794, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  20720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20720, CurrentID=0, Status='live', Symbol='Serpine2', Aliases='B230326M24Rik, PAI-1, PI-7, PI7, PN-1, Spi4', description='serine (or cysteine) peptidase inhibitor, clade E, member 2', other_designations='GDN|peptidase inhibitor 7|protease nexin 1|protease nexin I|serine (or cysteine) proteinase inhibitor, clade E (nexin, plasminogen activator inhibitor type 1), member 2|serine (or cysteine) proteinase inhibitor, clade E, member 2|serine protease inhibitor 4|serine protease-inhibitor 4|serpin E2', map_location='1 40.97 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=79794319, end_position_on_the_genomic_accession=79861175, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  20723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20723, CurrentID=0, Status='live', Symbol='Serpinb9', Aliases='BB283241, CAP-3, CAP3, PI-9, PI9, Spi6, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 9', other_designations='serine protease inhibitor 6', map_location='13 13.79 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=32997061, end_position_on_the_genomic_accession=33017955, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20724, CurrentID=0, Status='live', Symbol='Serpinb5', Aliases='1110036M19Rik, AI462524, AI646751, Maspin, PI-5, Spi7, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 5', other_designations='peptidase inhibitor 5|protease inhibitor 5|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 5|serine (or cysteine) proteinase inhibitor, clade B, member 5|serine protease inhibitor 7', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=106861117, end_position_on_the_genomic_accession=106883350, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20725, CurrentID=0, Status='live', Symbol='Serpinb8', Aliases='CAP-2, CAP2, NK10, Spi8, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 8', other_designations='serine (or cysteine) peptdiase inhibitor, clade B, member 8|serine (or cysteine) proteinase inhibitor, clade B, member 8|serine protease inhibitor 8', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107589925, end_position_on_the_genomic_accession=107610472, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20726, CurrentID=0, Status='live', Symbol='Serpinb9d', Aliases='AT2, Spi9, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 9d', other_designations='serine protease inhibitor 9', map_location='13 13.83 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33192959, end_position_on_the_genomic_accession=33203980, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20728, CurrentID=0, Status='live', Symbol='Spic', Aliases='AU019198, C76795, Prf, Spi-C', description='Spi-C transcription factor (Spi-1/PU.1 related)', other_designations='pu.1-related factor', map_location='10 C', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88675269, end_position_on_the_genomic_accession=88683023, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20729, CurrentID=0, Status='live', Symbol='Spin1', Aliases='Spin', description='spindlin 1', other_designations='30000 Mr metaphase complex|SSEC P|spindlin1', map_location='13 26.04 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51099834, end_position_on_the_genomic_accession=51152562, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20730, CurrentID=0, Status='live', Symbol='Spink1', Aliases='Spink3, p12', description='serine peptidase inhibitor, Kazal type 1', other_designations='prostatic secretory glycoprotein|serine peptidase inhibitor, Kazal type 3|serine protease inhibitor Kazal-type 3|serine protease inhibitor, Kazal type 3', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=43728069, end_position_on_the_genomic_accession=43737237, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20731, CurrentID=0, Status='live', Symbol='Spink4', Aliases='MPGC60', description='serine peptidase inhibitor, Kazal type 4', other_designations='peptide PEC-60 homolog|serine protease inhibitor, Kazal type 4', map_location='4 A', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40920056, end_position_on_the_genomic_accession=40931395, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20732, CurrentID=0, Status='live', Symbol='Spint1', Aliases='HAI-1', description='serine protease inhibitor, Kunitz type 1', other_designations='hepatocyte growth factor activator inhibitor type 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119237360, end_position_on_the_genomic_accession=119249535, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20733, CurrentID=0, Status='live', Symbol='Spint2', Aliases='AL024025, C76321, HAI-2', description='serine protease inhibitor, Kunitz type 2', other_designations='hepatocyte growth factor activator inhibitor type 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29256330, end_position_on_the_genomic_accession=29282064, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20737, CurrentID=0, Status='live', Symbol='Spn', Aliases='A630014B01Rik, Cd43, Galgp, Ly-48, Ly48', description='sialophorin', other_designations='3E8 antigen|B-cell differentiation antigen LP-3|leukocyte sialoglycoprotein|lymphocyte antigen 48', map_location='7 F3-F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127132232, end_position_on_the_genomic_accession=127144616, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20739, CurrentID=0, Status='live', Symbol='Spta1', Aliases='AF093576, AI451697, Spna-1, Spna1, ha, ihj, nmf4, sph', description='spectrin alpha, erythrocytic 1', other_designations='alpha-spectrin 1, erythroid|erythroid alpha-spectrin|hemolytic anemia|neuroscience mutagenesis facility, 4|spectrin alpha 1|spectrin alpha chain, erythrocyte|spherocytosis', map_location='1 80.97 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174172776, end_position_on_the_genomic_accession=174248449, orientation='plus', exon_count=53, OMIM='', no_hdr0=''),
  20740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20740, CurrentID=0, Status='live', Symbol='Sptan1', Aliases='2610027H02Rik, Spna-2, Spna2', description='spectrin alpha, non-erythrocytic 1', other_designations='alpha-II spectrin|alpha-spectrin 2, brain|fodrin alpha chain|spectrin alpha chain, brain|spectrin, non-erythroid alpha chain', map_location='2 20.93 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29965553, end_position_on_the_genomic_accession=30031451, orientation='plus', exon_count=59, OMIM='', no_hdr0=''),
  20741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20741, CurrentID=0, Status='live', Symbol='Sptb', Aliases='AI842465, D330027P03Rik, Gm1301, Spnb-1, Spnb1, ja, jaundiced, mKIAA4219', description='spectrin beta, erythrocytic', other_designations='beta-I spectrin|beta-spectrin 1|brain erythroid spectrin (235E)|spectrin R|spectrin beta 1|spectrin beta chain, erythrocyte', map_location='12 33.73 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76580488, end_position_on_the_genomic_accession=76710547, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  20742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20742, CurrentID=0, Status='live', Symbol='Sptbn1', Aliases='9930031C03Rik, AL033301, SPTB2, Spnb-2, Spnb2, elf1, elf3, mKIAA4049', description='spectrin beta, non-erythrocytic 1', other_designations='beta fodrin|beta-II spectrin|beta-spectrin 2, non-erythrocytic|brain spectrin|embryonic liver fodrin|fodrin beta chain|non-erythrocytic|spectrin G|spectrin beta 2|spectrin beta chain, brain 1|spectrin, non-erythroid beta chain 1', map_location='11 17.44 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30099395, end_position_on_the_genomic_accession=30268204, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  20743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20743, CurrentID=0, Status='live', Symbol='Sptbn2', Aliases='Spnb3, mKIAA0302', description='spectrin beta, non-erythrocytic 2', other_designations='beta-spectrin 3|spectrin beta 3', map_location='19 4.1 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4711223, end_position_on_the_genomic_accession=4752360, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  20744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20744, CurrentID=0, Status='live', Symbol='Strbp', Aliases='6430510M02Rik, AI852513, C230082I21Rik, C86322, Spnr', description='spermatid perinuclear RNA binding protein', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37569868, end_position_on_the_genomic_accession=37703969, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  20745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20745, CurrentID=0, Status='live', Symbol='Spock1', Aliases='Spock, Ticn1, testican', description='sparc/osteonectin, cwcv and kazal-like domains proteoglycan 1', other_designations='testican 1', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=57421195, end_position_on_the_genomic_accession=57908332, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  20747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20747, CurrentID=0, Status='live', Symbol='Spop', Aliases='AI315626, Pcif1, TEF2', description='speckle-type POZ protein', other_designations='HIB homolog 1|PDX-1 C-terminal-interacting factor 1', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95414064, end_position_on_the_genomic_accession=95493410, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20750, CurrentID=0, Status='live', Symbol='Spp1', Aliases='2AR, Apl-1, BNSP, BSPI, Bsp, ETA-1, Eta, OP, Opn, Opnl, Ric, Spp-1', description='secreted phosphoprotein 1', other_designations='44 kDa bone phosphoprotein|bone sialoprotein 1|calcium oxalate crystal growth inhibitor protein|early T-lymphocyte activation 1 protein|minopontin|osteopontin-like protein', map_location='5 50.68 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104435111, end_position_on_the_genomic_accession=104441053, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20751, CurrentID=0, Status='live', Symbol='Spr', Aliases='AA409688, ENSMUSG00000071345, Gm10328', description='sepiapterin reductase', other_designations='', map_location='6 37.15 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85133680, end_position_on_the_genomic_accession=85137764, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20753, CurrentID=0, Status='live', Symbol='Sprr1a', Aliases='AI528815, SPR1a, mSPRR1A', description='small proline-rich protein 1A', other_designations='SPR1 A', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92483954, end_position_on_the_genomic_accession=92485881, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20754, CurrentID=0, Status='live', Symbol='Sprr1b', Aliases='', description='small proline-rich protein 1B', other_designations='SPR1 B|SPR1B', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92436809, end_position_on_the_genomic_accession=92438779, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20755, CurrentID=0, Status='live', Symbol='Sprr2a1', Aliases='Sprr2a, Sprr2a3', description='small proline-rich protein 2A1', other_designations='sprr2a3', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92215835, end_position_on_the_genomic_accession=92222491, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20756, CurrentID=0, Status='live', Symbol='Sprr2b', Aliases='', description='small proline-rich protein 2B', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92316705, end_position_on_the_genomic_accession=92318085, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20758, CurrentID=0, Status='live', Symbol='Sprr2d', Aliases='', description='small proline-rich protein 2D', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92339140, end_position_on_the_genomic_accession=92340873, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20759, CurrentID=0, Status='live', Symbol='Sprr2e', Aliases='', description='small proline-rich protein 2E', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92352143, end_position_on_the_genomic_accession=92353449, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20760, CurrentID=0, Status='live', Symbol='Sprr2f', Aliases='', description='small proline-rich protein 2F', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92365187, end_position_on_the_genomic_accession=92366442, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20761, CurrentID=0, Status='live', Symbol='Sprr2g', Aliases='', description='small proline-rich protein 2G', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92373915, end_position_on_the_genomic_accession=92375229, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20762, CurrentID=0, Status='live', Symbol='Sprr2h', Aliases='AI661967', description='small proline-rich protein 2H', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92385685, end_position_on_the_genomic_accession=92387324, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20763, CurrentID=0, Status='live', Symbol='Sprr2i', Aliases='', description='small proline-rich protein 2I', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92407991, end_position_on_the_genomic_accession=92409271, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20764, CurrentID=0, Status='live', Symbol='Sprr2j-ps', Aliases='Sprr2j', description='small proline-rich protein 2J, pseudogene', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92418087, end_position_on_the_genomic_accession=92419396, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20765, CurrentID=0, Status='live', Symbol='Sprr2k', Aliases='', description='small proline-rich protein 2K', other_designations='', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92432582, end_position_on_the_genomic_accession=92433927, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  20766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20766, CurrentID=0, Status='live', Symbol='Sprr3', Aliases='SPR3', description='small proline-rich protein 3', other_designations='cornifin beta', map_location='3 40.14 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92456502, end_position_on_the_genomic_accession=92458720, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  20767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20767, CurrentID=0, Status='live', Symbol='sps', Aliases='dd', description='spontaneous seizure', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20768, CurrentID=0, Status='live', Symbol='Sephs2', Aliases='Sps2, Ysg3', description='selenophosphate synthetase 2', other_designations='clone 1000|selenium donor protein 2|selenophosphate synthase 2|the SelD gene product', map_location='7 69.43 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127271879, end_position_on_the_genomic_accession=127274059, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  20770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20770, CurrentID=0, Status='live', Symbol='Mucl2', Aliases='Spt-1, Spt1', description='mucin-like 2', other_designations='16.5 kDa submandibular gland glycoprotein|salivary protein 1', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103895855, end_position_on_the_genomic_accession=103899312, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20771, CurrentID=0, Status='live', Symbol='Mucl1', Aliases='Spt-2, Spt2', description='mucin-like 1', other_designations='salivary protein 2|spot 2 protein', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103751923, end_position_on_the_genomic_accession=103755498, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20773, CurrentID=0, Status='live', Symbol='Sptlc2', Aliases='LCB2, LCB2a', description='serine palmitoyltransferase, long chain base subunit 2', other_designations='LCB 2|SPT 2|long chain base biosynthesis protein 2|long chain base biosynthesis protein 2a|serine-palmitoyl-CoA transferase 2', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87307889, end_position_on_the_genomic_accession=87388230, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20775, CurrentID=0, Status='live', Symbol='Sqle', Aliases='AI323792', description='squalene epoxidase', other_designations='SE', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=59315092, end_position_on_the_genomic_accession=59331193, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20776, CurrentID=0, Status='live', Symbol='Tmie', Aliases='5131400L21Rik, Mm.87012, sr', description='transmembrane inner ear', other_designations='spinner', map_location='9 60.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110865707, end_position_on_the_genomic_accession=110880083, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20778, CurrentID=0, Status='live', Symbol='Scarb1', Aliases='AI120173, CD36, Cd36l1, Chohd1, Cla-1, Cla1, D5Ertd460e, Hdlq1, Hlb398, SR-B1, SR-BI, SRBI, Srb1, mSR-BI', description='scavenger receptor class B, member 1', other_designations='HDL QTL 1|scavenger receptor class B type I|scavenger receptor class B1', map_location='5 G1.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=125277087, end_position_on_the_genomic_accession=125341111, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20779, CurrentID=0, Status='live', Symbol='Src', Aliases='AW259666, pp60c-src', description='Rous sarcoma oncogene', other_designations='p60-Src|proto-oncogene c-Src', map_location='2 78.35 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157423901, end_position_on_the_genomic_accession=157471838, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  20787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20787, CurrentID=0, Status='live', Symbol='Srebf1', Aliases='ADD1, SREBP-1a, SREBP1, SREBP1c, bHLHd1', description='sterol regulatory element binding transcription factor 1', other_designations='adipocyte determination- and differentiation-dependent factor 1', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60199089, end_position_on_the_genomic_accession=60222686, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  20788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20788, CurrentID=0, Status='live', Symbol='Srebf2', Aliases='AI608257, SREBP-2, SREBP2, SREBP2gc, bHLHd2, lop13, nuc', description='sterol regulatory element binding factor 2', other_designations='srebf2|sterol regulatory element-binding transcription factor 2', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82147269, end_position_on_the_genomic_accession=82205379, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  545085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545085, CurrentID=0, Status='live', Symbol='Wdr70', Aliases='4833422F06Rik', description='WD repeat domain 70', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=7873055, end_position_on_the_genomic_accession=8099209, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  20807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20807, CurrentID=0, Status='live', Symbol='Srf', Aliases='AW049942, AW240594', description='serum response factor', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46546838, end_position_on_the_genomic_accession=46556870, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20810, CurrentID=0, Status='live', Symbol='Srm', Aliases='AA407669, SpdST, SpdSy', description='spermidine synthase', other_designations='putrescine aminopropyltransferase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148591513, end_position_on_the_genomic_accession=148594619, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20811, CurrentID=0, Status='live', Symbol='Srms', Aliases='A230069J08Rik, srm', description='src-related kinase lacking C-terminal regulatory tyrosine and N-terminal myristylation sites', other_designations='PTK70', map_location='2 103.62 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181205562, end_position_on_the_genomic_accession=181218783, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20812, CurrentID=0, Status='live', Symbol='srn', Aliases='', description='siren', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20813, CurrentID=0, Status='live', Symbol='Srp14', Aliases='14kDa, AW536328', description='signal recognition particle 14', other_designations='(homologous Alu RNA binding protein)|homologous Alu RNA-binding protein', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118475843, end_position_on_the_genomic_accession=118479696, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20815, CurrentID=0, Status='live', Symbol='Srpk1', Aliases='AU017960', description='serine/arginine-rich protein specific kinase 1', other_designations='SFRS protein kinase 1|SR protein-specific kinase 1|SR-protein-specific kinase 1|serine/arginine-rich protein-specific kinase 1|serine/threonine-protein kinase SRPK1|srpk1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28589590, end_position_on_the_genomic_accession=28622636, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  20817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20817, CurrentID=0, Status='live', Symbol='Srpk2', Aliases='AW226533, AW492537, AW547358, Wbp6, mSRPK2', description='serine/arginine-rich protein specific kinase 2', other_designations='SFRS protein kinase 2|SR-protein-specific kinase 2|WW domain binding protein 6|serine/arginine-rich protein-specific kinase 2|serine/threonine-protein kinase SRPK2|srpk2', map_location='5 10.36 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23503356, end_position_on_the_genomic_accession=23684641, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  20818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20818, CurrentID=0, Status='live', Symbol='Srprb', Aliases='AA409543', description='signal recognition particle receptor, B subunit', other_designations='SR-beta', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103188033, end_position_on_the_genomic_accession=103202065, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20819, CurrentID=0, Status='live', Symbol='Srst', Aliases='', description='simple repeat sequence-containing transcript', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20821, CurrentID=0, Status='live', Symbol='Trim21', Aliases='Ro52, Ssa1', description='tripartite motif-containing 21', other_designations='52 kDa Ro protein|52 kDa ribonucleoprotein autoantigen Ro/SS-A|52kD Ro/SSA autoantigen|SS-A|Sjogren syndrome antigen A1|ro(SS-A)|sjoegren syndrome type A antigen|tripartite motif protein 21|tripartite motif-containing protein 21', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102557920, end_position_on_the_genomic_accession=102565482, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20822, CurrentID=0, Status='live', Symbol='Trove2', Aliases='1810007I17Rik, A530054J02Rik, AI646302, SS-A/Ro, Ssa, Ssa2', description='TROVE domain family, member 2', other_designations='60 kDa Ro protein|60 kDa ribonucleoprotein Ro|Sjogren syndrome antigen A2|roRNP', map_location='1 62.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=143750790, end_position_on_the_genomic_accession=143777051, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20823, CurrentID=0, Status='live', Symbol='Ssb', Aliases='SS-B', description='Sjogren syndrome antigen B', other_designations='La protein|Lupus LA protein homolog (LA ribonucleoprotein) (LA autoantigen homolog)|autoantigen La|la autoantigen homolog|la ribonucleoprotein', map_location='2 40.95 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69861562, end_position_on_the_genomic_accession=69871846, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20826, CurrentID=0, Status='live', Symbol='Nhp2l1', Aliases='FA-1, Fta1, Snu13, Ssfa1', description='NHP2 non-histone chromosome protein 2-like 1 (S. cerevisiae)', other_designations='U4/U6.U5 small nuclear ribonucleoprotein SNU13|U4/U6.U5 tri-snRNP 15.5 kDa protein|fertilization antigen 1|fertilization antigen-1|high mobility group-like nuclear protein 2 homolog 1|sperm specific antigen 1|sperm-specific antigen 1', map_location='15 38.38 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82041345, end_position_on_the_genomic_accession=82047598, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20829, CurrentID=0, Status='live', Symbol='Ssm1', Aliases='Ssm-1', description='strain specific modifier of transgene methylation 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20832, CurrentID=0, Status='live', Symbol='Ssr4', Aliases='AW743063, C78062, Trap', description='signal sequence receptor, delta', other_designations='SSR-delta|TRAP-delta|signal sequence receptor subunit delta', map_location='X 37.41 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73787028, end_position_on_the_genomic_accession=73790828, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20833, CurrentID=0, Status='live', Symbol='Ssrp1', Aliases='C81323, Hmg1-rs1, Hmgi-rs3, Hmgox, T160', description='structure specific recognition protein 1', other_designations='facilitates chromatin transcription complex subunit SSRP1|recombination signal sequence recognition protein 1', map_location='2 49.45 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85037120, end_position_on_the_genomic_accession=85047114, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  20834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20834, CurrentID=0, Status='live', Symbol='Znrf4', Aliases='1700094M01Rik, Ssrzf1, spzn', description='zinc and ring finger 4', other_designations='sperizin|spermatid specific RING zinc finger 1', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56511248, end_position_on_the_genomic_accession=56512483, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  545123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545123, CurrentID=0, Status='live', Symbol='Cyp2d11', Aliases='Cyp2d, P450-2D', description='cytochrome P450, family 2, subfamily d, polypeptide 11', other_designations='CYPIID11|cytochrome P450, 2d11|cytochrome P450-16-alpha|cytochrome P450CC|testosterone 16-alpha hydroxylase', map_location='15 38.58 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82389154, end_position_on_the_genomic_accession=82394022, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20840, CurrentID=0, Status='live', Symbol='Stac', Aliases='', description='src homology three (SH3) and cysteine rich domain', other_designations='SRC homology 3 and cysteine-rich domain-containing protein', map_location='9 61.81 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111561434, end_position_on_the_genomic_accession=111690588, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  20841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20841, CurrentID=0, Status='live', Symbol='Zfp143', Aliases='AA959806, AU015869, D7Ertd805e, KRAB14, SBF, Staf, Zfp79, Zfp80-rs1, Znf143, pHZ-1', description='zinc finger protein 143', other_designations='mStaf|selenocysteine tRNA gene transcription activating factor|selenocysteine tRNA gene transcription-activating factor|zfp-143|zinc finger protein 80, related sequence 1', map_location='7 E3-F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110061670, end_position_on_the_genomic_accession=110095396, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  20842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20842, CurrentID=0, Status='live', Symbol='Stag1', Aliases='AU045003, SA-1, Scc3', description='stromal antigen 1', other_designations='SCC3 homolog 1', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=100597709, end_position_on_the_genomic_accession=100958544, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  20843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20843, CurrentID=0, Status='live', Symbol='Stag2', Aliases='9230105L23Rik, B230112I07Rik, SA-2, SAP2', description='stromal antigen 2', other_designations='SCC3 homolog 2|nuclear protein SA2', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=42149322, end_position_on_the_genomic_accession=42277186, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  20844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20844, CurrentID=0, Status='live', Symbol='Stam', Aliases='STAM1', description='signal transducing adaptor molecule (SH3 domain and ITAM motif) 1', other_designations='STAM-1', map_location='2 A2-B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14073646, end_position_on_the_genomic_accession=14149210, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20845, CurrentID=0, Status='live', Symbol='Star', Aliases='AV363654, D8Ertd419e', description='steroidogenic acute regulatory protein', other_designations='START domain-containing protein 1|luteinizing hormone-induced protein|stARD1', map_location='8 14.17 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25808513, end_position_on_the_genomic_accession=25815982, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20846, CurrentID=0, Status='live', Symbol='Stat1', Aliases='2010005J02Rik, AA408197', description='signal transducer and activator of transcription 1', other_designations='', map_location='1 26.81 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52119438, end_position_on_the_genomic_accession=52161865, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20847, CurrentID=0, Status='live', Symbol='Stat2', Aliases='1600010G07Rik, AW496480', description='signal transducer and activator of transcription 2', other_designations='', map_location='10 76.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128270559, end_position_on_the_genomic_accession=128292852, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  20848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20848, CurrentID=0, Status='live', Symbol='Stat3', Aliases='1110034C02Rik, AW109958, Aprf', description='signal transducer and activator of transcription 3', other_designations='acute phase response factor|acute-phase response factor', map_location='11 63.82 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100886810, end_position_on_the_genomic_accession=100939570, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  20849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20849, CurrentID=0, Status='live', Symbol='Stat4', Aliases='', description='signal transducer and activator of transcription 4', other_designations='', map_location='1 26.67 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=51987106, end_position_on_the_genomic_accession=52107189, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20850, CurrentID=0, Status='live', Symbol='Stat5a', Aliases='AA959963, STAT5', description='signal transducer and activator of transcription 5A', other_designations='mammary gland factor STAT5A', map_location='11 63.77 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100859180, end_position_on_the_genomic_accession=100885169, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  20851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20851, CurrentID=0, Status='live', Symbol='Stat5b', Aliases='', description='signal transducer and activator of transcription 5B', other_designations='mammary gland factor STAT5B', map_location='11 63.63 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100780731, end_position_on_the_genomic_accession=100859004, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  545140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545140, CurrentID=0, Status='live', Symbol='Olfr288', Aliases='C85333, GA_x5J8B7W76D0-7609197-7608973, MOR286-3P', description='olfactory receptor 288', other_designations='GA_x6K02T2NBG7-5568919-5569857|olfactory receptor GA_x5J8B7W76D0-7609197-7608973|olfactory receptor MOR286-3P', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98186013, end_position_on_the_genomic_accession=98195542, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  414069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=414069, CurrentID=0, Status='live', Symbol='BC024978', Aliases='', description='cDNA sequence BC024978', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27195781, end_position_on_the_genomic_accession=27205386, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20855, CurrentID=0, Status='live', Symbol='Stc1', Aliases='Stc', description='stanniocalcin 1', other_designations='STC-1', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69029289, end_position_on_the_genomic_accession=69041401, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20856, CurrentID=0, Status='live', Symbol='Stc2', Aliases='AW125853l, mustc2, Stc2', description='stanniocalcin 2', other_designations='STC-2', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=31359441, end_position_on_the_genomic_accession=31370061, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20859, CurrentID=0, Status='live', Symbol='Sult2a1', Aliases='ST2A1, Std, Sth1, mSTa1', description='sulfotransferase family 2A, dehydroepiandrosterone (DHEA)-preferring, member 1', other_designations='ST|hydroxysteroid sulfotransferase|sulfotransferase 2A1|sulfotransferase, DHEA preferring|sulfotransferase, hydroxysteroid preferring 1', map_location='7 8.02 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13796246, end_position_on_the_genomic_accession=13837410, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20860, CurrentID=0, Status='live', Symbol='Sult1e1', Aliases='EST, ST1E1, Ste', description='sulfotransferase family 1E, member 1', other_designations='sulfotransferase 1E1|sulfotransferase, estrogen preferring|sulfotransferase, estrogen-preferring', map_location='5 43.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87575968, end_position_on_the_genomic_accession=87591611, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20861, CurrentID=0, Status='live', Symbol='Stfa1', Aliases='Stefin-3, Stf-3, Stf1, Stf3, Stfa3', description='stefin A1', other_designations='', map_location='16 25.45 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36277148, end_position_on_the_genomic_accession=36285371, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20862, CurrentID=0, Status='live', Symbol='Stfa2', Aliases='Stf2', description='stefin A2', other_designations='', map_location='16 25.52 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36403944, end_position_on_the_genomic_accession=36420916, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20863, CurrentID=0, Status='live', Symbol='Stfa3', Aliases='Stf3', description='stefin A3', other_designations='', map_location='16 25.55 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36450537, end_position_on_the_genomic_accession=36455392, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20866, CurrentID=0, Status='live', Symbol='Stim1', Aliases='SIM', description='stromal interaction molecule 1', other_designations='', map_location='7 54.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102267271, end_position_on_the_genomic_accession=102437319, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20867, CurrentID=0, Status='live', Symbol='Stip1', Aliases='Hop, Sti1, p60', description='stress-induced phosphoprotein 1', other_designations='Hsp70/Hsp90 organizing protein|IEF SSP 3521|hsc70/Hsp90-organizing protein|mSTI1|stress-inducible protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7020696, end_position_on_the_genomic_accession=7040026, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  20868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20868, CurrentID=0, Status='live', Symbol='Stk10', Aliases='Gek1, Lok, mKIAA4026', description='serine/threonine kinase 10', other_designations='germ cell kinase 1|lymphocyte-oriented kinase', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32533266, end_position_on_the_genomic_accession=32624595, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  20869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20869, CurrentID=0, Status='live', Symbol='Stk11', Aliases='AA408040, Lkb1, Par-4, R75140', description='serine/threonine kinase 11', other_designations='liver kinase B1 homolog|serine/threonine-protein kinase 11|serine/threonine-protein kinase LKB1|stk11', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80115693, end_position_on_the_genomic_accession=80130685, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  20871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20871, CurrentID=0, Status='live', Symbol='Aurkc', Aliases='AIE1, AIK3, ARK-3, IAK3, Stk13', description='aurora kinase C', other_designations='aurora 3|aurora B|aurora-related kinase 3|aurora/IPL1-related kinase 3|aurora/Ipl1/Eg2 protein 1|serine/threonine kinase 13 (aurora/IPL-like)|serine/threonine-protein kinase 13|serine/threonine-protein kinase aurora-C', map_location='7 A2-A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6989892, end_position_on_the_genomic_accession=7003364, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20872, CurrentID=0, Status='live', Symbol='Stk16', Aliases='EDPK, Krct, MPSK, PKL12, TSF-1', description='serine/threonine kinase 16', other_designations='TGF-beta-stimulated factor 1|embryo-derived protein kinase|kinase related to cerevisiae and thaliana|myristoylated and palmitoylated serine/threonine-protein kinase|tyrosine-protein kinase STK16', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75210829, end_position_on_the_genomic_accession=75215606, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20873, CurrentID=0, Status='live', Symbol='Plk4', Aliases='1700028H20, AI385771, Sak, Stk18', description='polo-like kinase 4', other_designations='PLK-4|serine/threonine kinase 18|serine/threonine-protein kinase 18|serine/threonine-protein kinase Sak', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40799951, end_position_on_the_genomic_accession=40816883, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  20874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20874, CurrentID=0, Status='live', Symbol='Slk', Aliases='9A2, AV021402, AW411554, Etk4, SMAK, Stk2, mKIAA0204, mSLK', description='STE20-like kinase', other_designations='STE20-related kinase SMAK|STE20-related serine/threonine-protein kinase|serine/threonine-protein kinase 2', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47580019, end_position_on_the_genomic_accession=47645246, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  20877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20877, CurrentID=0, Status='live', Symbol='Aurkb', Aliases='AIM-1, AIRK2, AL022959, Aik2, Aim1, Ark2, AurB, IPL1, STK-1, Stk12, Stk5', description='aurora kinase B', other_designations='ARK-2|aurora 1|aurora B|aurora- and IPL1-like midbody-associated protein 1|aurora-related kinase 2|aurora/IPL1-related kinase 2|serine/threonine kinase 12|serine/threonine kinase 5|serine/threonine-protein kinase 12|serine/threonine-protein kinase 5|serine/threonine-protein kinase aurora-B', map_location='11 42.32 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69045625, end_position_on_the_genomic_accession=69051664, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20878, CurrentID=0, Status='live', Symbol='Aurka', Aliases='AIRK1, ARK-1, AU019385, AW539821, Ark1, Aurora-A, Ayk1, IAK, IAK1, Stk6', description='aurora kinase A', other_designations='aurora 2|aurora A|aurora family kinase 1|aurora-related kinase 1|aurora/IPL1-related kinase 1|ipl1- and aurora-related kinase 1|serine/threonine kinase 6|serine/threonine-protein kinase 6|serine/threonine-protein kinase Ayk1|serine/threonine-protein kinase aurora-A', map_location='2 94.84 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172356190, end_position_on_the_genomic_accession=172370706, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  20882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20882, CurrentID=0, Status='live', Symbol='stm', Aliases='', description='stumpy', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20887, CurrentID=0, Status='live', Symbol='Sult1a1', Aliases='AI266890, PST, ST1A1, Stp, Stp1', description='sulfotransferase family 1A, phenol-preferring, member 1', other_designations='ST1A4|aryl sulfotransferase|arylsulfotransferase ST1A4|mSTp1|phenol sulfotransferase|phenol/aryl sulfotransferase|sulfokinase|sulfotransferase, phenol preferring 1', map_location='7 69.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126672865, end_position_on_the_genomic_accession=126676483, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20888, CurrentID=0, Status='live', Symbol='Sult1c1', Aliases='(PST)G, P-SULT, ST1C1, Stp2, Sult1a2, mOLFST', description='sulfotransferase family, cytosolic, 1C, member 1', other_designations='phenol sulfotransferase|sulfotransferase family 1A, phenol-preferring, member 2|sult1c1', map_location='17 28.07 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=53961615, end_position_on_the_genomic_accession=53990631, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20890, CurrentID=0, Status='live', Symbol='Wnt8a', Aliases='Stra11, Wnt-8A, Wnt-8D, Wnt8d', description='wingless-type MMTV integration site family, member 8A', other_designations='protein Wnt-8d|stimulated by retinoic acid gene 11 protein|wingless-related MMTV integration site 8A|wingless-related MMTV integration site 8D|wnt-8', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34542058, end_position_on_the_genomic_accession=34548960, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20892, CurrentID=0, Status='live', Symbol='Stra13', Aliases='', description='stimulated by retinoic acid 13', other_designations='CENP-X|FANCM-interacting histone fold protein 2|Fanconi anemia-associated polypeptide of 10 kDa|immediate-early-response protein D9|retinoic acid-inducible gene D9 protein|stimulated by retinoic acid gene 13 protein homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120710941, end_position_on_the_genomic_accession=120713903, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  20893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20893, CurrentID=0, Status='live', Symbol='Bhlhe40', Aliases='Bhlhb2, C130042M06Rik, CR8, Clast5, Dec1, Sharp2, Stra13, Stra14', description='basic helix-loop-helix family, member e40', other_designations='E47 interaction protein 1|basic helix-loop-helix domain containing, class B2|class B basic helix-loop-helix protein 2|cytokine response gene 8|eip1 (E47 interaction protein 1)|stimulated by retinoic acid gene 13 protein', map_location='6 E-F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=108660629, end_position_on_the_genomic_accession=108666925, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  20897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20897, CurrentID=0, Status='live', Symbol='Stra6', Aliases='AI891933', description='stimulated by retinoic acid gene 6', other_designations='retinoic acid-responsive protein', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58063793, end_position_on_the_genomic_accession=58154187, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  20899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20899, CurrentID=0, Status='live', Symbol='Stra8', Aliases='', description='stimulated by retinoic acid gene 8', other_designations='stra8', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34920163, end_position_on_the_genomic_accession=34939344, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  20901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20901, CurrentID=0, Status='live', Symbol='Strap', Aliases='AW557906, C78091, C79202, Unrip', description='serine/threonine kinase receptor associated protein', other_designations='UNR-interacting protein', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=137735082, end_position_on_the_genomic_accession=137751930, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  545192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545192, CurrentID=0, Status='live', Symbol='Baiap3', Aliases='Bap3, Gm937', description='BAI1-associated protein 3', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25242659, end_position_on_the_genomic_accession=25256364, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  20905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20905, CurrentID=0, Status='live', Symbol='Sts', Aliases='ArsC', description='steroid sulfatase', other_designations='ASC|arylsulfatase C|steryl-sulfate sulfohydrolase', map_location='XY 75.0 cM', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20906, CurrentID=0, Status='live', Symbol='stu', Aliases='', description='stumbler', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20907, CurrentID=0, Status='live', Symbol='Stx1a', Aliases='HPC-1', description='syntaxin 1A (brain)', other_designations='neuron-specific antigen HPC-1|syntaxin 1 a', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135023487, end_position_on_the_genomic_accession=135051104, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20908, CurrentID=0, Status='live', Symbol='Stx3', Aliases='Syn-3', description='syntaxin 3', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11775118, end_position_on_the_genomic_accession=11820422, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20909, CurrentID=0, Status='live', Symbol='Stx4a', Aliases='Stx4, Syn-4, Syn4', description='syntaxin 4A (placental)', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127841769, end_position_on_the_genomic_accession=127849686, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20910, CurrentID=0, Status='live', Symbol='Stxbp1', Aliases='AI317162, AI326233, MMS10-G, Ms10g, Munc-18a, Munc18-1, N-sec1, Rb-sec1, Sxtbp1, Unc18-1, Unc18h, nsec1', description='syntaxin binding protein 1', other_designations='UNC-18 homolog|minisatellite 10g detected by probe MMS10|protein unc-18 homolog 1|protein unc-18 homolog A|unc-18-1|unc-18A', map_location='2 22.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32787607, end_position_on_the_genomic_accession=32847237, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  20911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20911, CurrentID=0, Status='live', Symbol='Stxbp2', Aliases='C79054, Munc-18-2, Munc-18b, Munc18b, Sxtbp2, Sxtp2, Unc18-2, Unc18b, muSec1', description='syntaxin binding protein 2', other_designations='munc18-2|protein unc-18 homolog 2|protein unc-18 homolog B|syntaxin binding protein Munc18-2|unc-18 homolog 2|unc-18B', map_location='8 A1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3630960, end_position_on_the_genomic_accession=3643707, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  20912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20912, CurrentID=0, Status='live', Symbol='Stxbp3', Aliases='Munc-18ca, Sxtbp3, Unc18-3, Stxbp3', description='syntaxin binding protein 3', other_designations='MUNC-18-3|mammalian homolog of Unc-18c|protein unc-18 homolog 3|protein unc-18 homolog C|syntaxin binding protein 3A|unc-18 homolog 3|unc-18C', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108793180, end_position_on_the_genomic_accession=108840502, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  20913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20913, CurrentID=0, Status='live', Symbol='Stxbp4', Aliases='6030470M02Rik, Synip', description='syntaxin binding protein 4', other_designations='STX4-interacting protein|syntaxin 4-interacting protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=90476492, end_position_on_the_genomic_accession=90638123, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  57435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57435, CurrentID=0, Status='live', Symbol='Plin4', Aliases='S3-12, mKIAA1881', description='perilipin 4', other_designations='adipocyte protein S3-12|plasma membrane associated protein, S3-12', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56100587, end_position_on_the_genomic_accession=56109947, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20915, CurrentID=0, Status='live', Symbol='su', Aliases='', description='surdescens', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20916, CurrentID=0, Status='live', Symbol='Sucla2', Aliases='4930547K18Rik', description='succinate-Coenzyme A ligase, ADP-forming, beta subunit', other_designations='ATP-specific succinyl-CoA synthetase subunit beta|SCS-betaA|succinyl-CoA synthetase beta-A chain', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73552786, end_position_on_the_genomic_accession=73596144, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20917, CurrentID=0, Status='live', Symbol='Suclg2', Aliases='AF171077, AW556404, D6Wsu120e, SCS-betaG', description='succinate-Coenzyme A ligase, GDP-forming, beta subunit', other_designations='GTP-specific succinyl-CoA synthetase beta subunit|succinyl-CoA synthetase beta-G chain', map_location='6 44.63 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=95473009, end_position_on_the_genomic_accession=95718846, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  20918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20918, CurrentID=0, Status='live', Symbol='Eif1', Aliases='Sui1-rs1', description='eukaryotic translation initiation factor 1', other_designations='protein translation factor SUI1 homolog|suppressor of initiator codon mutations, related sequence 1|suppressor of initiator codon mutations-Yeast homolog related sequence 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100319994, end_position_on_the_genomic_accession=100322096, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20922, CurrentID=0, Status='live', Symbol='Supt4a', Aliases='AL022777, Supt4h, Supt4h1, Supt4h1a', description='suppressor of Ty 4A', other_designations='DRB sensitivity-inducing factor small subunit 1|DSIF small subunit 1|suppressor of Ty 4 homolog 1|transcription elongation factor SPT4 1', map_location='11 52.2 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87737507, end_position_on_the_genomic_accession=87743628, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20924, CurrentID=0, Status='live', Symbol='Supt5', Aliases='AL033283, AL033321, AU019549, Spt5h, Supt5', description='suppressor of Ty 5', other_designations='DRB sensitivity-inducing factor large subunit|DSIF large subunit|suppressor of Ty 5 homolog', map_location='7 16.68 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28314891, end_position_on_the_genomic_accession=28338804, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  20926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20926, CurrentID=0, Status='live', Symbol='Supt6', Aliases='5131400N11Rik, AI132449, SPT6h, mKIAA0162, Supt6', description='suppressor of Ty 6', other_designations='suppressor of Ty 6 homolog', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78206749, end_position_on_the_genomic_accession=78245703, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  20927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20927, CurrentID=0, Status='live', Symbol='Abcc8', Aliases='D930031B21Rik, SUR1, Sur', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 8', other_designations='sulfonylurea receptor', map_location='7 29.66 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46104523, end_position_on_the_genomic_accession=46180039, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  20928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20928, CurrentID=0, Status='live', Symbol='Abcc9', Aliases='AI414027, AI449286, SUR2A, SUR2B, Sur2', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 9', other_designations='ATP-binding cassette transporter sub-family C member 9|sulfonylurea receptor 2|sulfonylurea-binding protein 2', map_location='6 74.35 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142587862, end_position_on_the_genomic_accession=142702760, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  20930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20930, CurrentID=0, Status='live', Symbol='Surf1', Aliases='0610010F23Rik, Surf-1', description='surfeit gene 1', other_designations='', map_location='2 19.1 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26913378, end_position_on_the_genomic_accession=26916530, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  20931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20931, CurrentID=0, Status='live', Symbol='Surf2', Aliases='Surf-2', description='surfeit gene 2', other_designations='', map_location='2 19.1 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26916372, end_position_on_the_genomic_accession=26920183, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20932, CurrentID=0, Status='live', Symbol='Surf4', Aliases='AL033340, AL033373, Surf-4', description='surfeit gene 4', other_designations='', map_location='2 19.1 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26920040, end_position_on_the_genomic_accession=26933776, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20933, CurrentID=0, Status='live', Symbol='Med22', Aliases='AW212655, AW558812, Surf-5, Surf5', description='mediator complex subunit 22', other_designations='surfeit gene 5|surfeit locus protein 5', map_location='2 19.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26905262, end_position_on_the_genomic_accession=26910672, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20935, CurrentID=0, Status='live', Symbol='Surf6', Aliases='D2Wsu129e, Surf-6', description='surfeit gene 6', other_designations='', map_location='2 19.08 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26890772, end_position_on_the_genomic_accession=26902813, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20937, CurrentID=0, Status='live', Symbol='Suv39h1', Aliases='AI852103, AL022883, DXHXS7466e, H3-K9-HMTase 1, KMT1A, mIS6', description='suppressor of variegation 3-9 homolog 1 (Drosophila)', other_designations='Suv39h histone methyltransferase|histone H3-K9 methyltransferase 1|position-effect variegation 3-9 homolog|su(var)3-9 homolog 1', map_location='X A1-A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8061171, end_position_on_the_genomic_accession=8074760, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  20939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20939, CurrentID=0, Status='live', Symbol='Sva', Aliases='', description='seminal vesicle antigen', other_designations='', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42038394, end_position_on_the_genomic_accession=42042851, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20941, CurrentID=0, Status='live', Symbol='Svs4', Aliases='SVS-IV, Svp-2, Svp-4, Svp2, Svp4', description='seminal vesicle secretory protein 4', other_designations='SVS IV|seminal vesicle protein 2|seminal vesicle protein 4|seminal vesicle secretory protein IV', map_location='2 85.01 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164275956, end_position_on_the_genomic_accession=164278307, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  20944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20944, CurrentID=0, Status='live', Symbol='Svs5', Aliases='BB115186, Semg1, Svp-1, Svp1, Svp5', description='seminal vesicle secretory protein 5', other_designations='SVS V|SVS protein F|semenogelin I|seminal vesicle protein 1|seminal vesicle protein F|seminal vesicle secretion 5|seminal vesicle secretory protein V', map_location='2 85.16 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164332765, end_position_on_the_genomic_accession=164334394, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  20945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20945, CurrentID=0, Status='live', Symbol='Svs6', Aliases='MSVSP99, Svp6', description='seminal vesicle secretory protein 6', other_designations='SVS VI|SVSP99|seminal vesicle protein 6|seminal vesicle secretion 6|seminal vesicle secretory protein VI', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164316751, end_position_on_the_genomic_accession=164318450, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20947, CurrentID=0, Status='live', Symbol='Swap70', Aliases='70kDa, AV235546', description='SWA-70 protein', other_designations='SWAP complex protein, 70 kDa|SWAP-70', map_location='7 57.7 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110221703, end_position_on_the_genomic_accession=110283506, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20954, CurrentID=0, Status='live', Symbol='sy', Aliases='', description='shaker-with-syndactylism deletion region', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20955, CurrentID=0, Status='live', Symbol='Vamp7', Aliases='Sybl1, TI-VAMP, VAMP-7', description='vesicle-associated membrane protein 7', other_designations='synaptobrevin like 1|synaptobrevin-like protein 1|tetanus neurotoxin-insensitive vesicle-associated membrane protein', map_location='X', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  20957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20957, CurrentID=0, Status='live', Symbol='Sycp1', Aliases='SCP1', description='synaptonemal complex protein 1', other_designations='SCP-1', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102818499, end_position_on_the_genomic_accession=102936312, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  20958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20958, CurrentID=0, Status='live', Symbol='Sycp1-ps1', Aliases='Sycp1-rs', description='synaptonemal complex protein 1, pseudogene 1', other_designations='', map_location='7 9.31 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18786302, end_position_on_the_genomic_accession=18789414, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  20962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20962, CurrentID=0, Status='live', Symbol='Sycp3', Aliases='Cor1, Scp3', description='synaptonemal complex protein 3', other_designations='SCP-3', map_location='10 C', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88459587, end_position_on_the_genomic_accession=88473236, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  20963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20963, CurrentID=0, Status='live', Symbol='Syk', Aliases='Sykb', description='spleen tyrosine kinase', other_designations='', map_location='13 B-C2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=52582329, end_position_on_the_genomic_accession=52648792, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  20964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20964, CurrentID=0, Status='live', Symbol='Syn1', Aliases='Syn-1-S, Syn1', description='synapsin I', other_designations='truncated synapsin 1-S', map_location='X A1-A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20860511, end_position_on_the_genomic_accession=20920918, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  20965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20965, CurrentID=0, Status='live', Symbol='Syn2', Aliases='2900074L19Rik, AI836018, AI841723', description='synapsin II', other_designations='', map_location='6 F', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115134902, end_position_on_the_genomic_accession=115282626, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  20969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20969, CurrentID=0, Status='live', Symbol='Sdc1', Aliases='AA408134, AA409076, CD138, Sstn, Synd, Synd1, syn-1', description='syndecan 1', other_designations='synstatin', map_location='12 3.94 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8771321, end_position_on_the_genomic_accession=8793716, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  20970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20970, CurrentID=0, Status='live', Symbol='Sdc3', Aliases='Synd3, mKIAA0468, syn-3', description='syndecan 3', other_designations='', map_location='4 63.63 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130792537, end_position_on_the_genomic_accession=130826318, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20971, CurrentID=0, Status='live', Symbol='Sdc4', Aliases='AA959608, AW108331, S4, Synd4, ryudocan, syndecan-4', description='syndecan 4', other_designations='ryudocan core protein', map_location='2 85.16 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164424247, end_position_on_the_genomic_accession=164443188, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  20972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20972, CurrentID=0, Status='live', Symbol='Syngr1', Aliases='R75468b, p29, Syngr1', description='synaptogyrin 1', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80091334, end_position_on_the_genomic_accession=80119501, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  20973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20973, CurrentID=0, Status='live', Symbol='Syngr2', Aliases='Clast2, cellugyrin', description='synaptogyrin 2', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117809667, end_position_on_the_genomic_accession=117814286, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20974, CurrentID=0, Status='live', Symbol='Syngr3', Aliases='', description='synaptogyrin 3', other_designations='Sgyr III|synaptogryin III', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24685092, end_position_on_the_genomic_accession=24689949, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20975, CurrentID=0, Status='live', Symbol='Synj2', Aliases='AI481647, SJ2, mKIAA0348', description='synaptojanin 2', other_designations='synaptic inositol-1,4,5-trisphosphate 5-phosphatase 2', map_location='17 3.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=5941280, end_position_on_the_genomic_accession=6045553, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  20977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20977, CurrentID=0, Status='live', Symbol='Syp', Aliases='A230093K24Rik, AI848995, Syn, p38', description='synaptophysin', other_designations='BM89 antigen|Syp I|major synaptic vesicle protein p38', map_location='X A-D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7638580, end_position_on_the_genomic_accession=7653256, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  20979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20979, CurrentID=0, Status='live', Symbol='Syt1', Aliases='AW124717, G630098F17Rik, SytI', description='synaptotagmin I', other_designations='p65|synaptotagmin 1', map_location='10 56.52 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=108497650, end_position_on_the_genomic_accession=109010975, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  20980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20980, CurrentID=0, Status='live', Symbol='Syt2', Aliases='IP4BP, R74640, mKIAA4194', description='synaptotagmin II', other_designations='IP4-binding protein|inositol polyphosphate-binding protein|synaptotagmin 2|sytII', map_location='1 58.24 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134646564, end_position_on_the_genomic_accession=134753149, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20981, CurrentID=0, Status='live', Symbol='Syt3', Aliases='AI385753, sytIII', description='synaptotagmin III', other_designations='', map_location='7 28.83 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44384099, end_position_on_the_genomic_accession=44400030, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  20983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20983, CurrentID=0, Status='live', Symbol='Syt4', Aliases='SytIV', description='synaptotagmin IV', other_designations='synaptotagmin 4', map_location='18 17.75 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31437808, end_position_on_the_genomic_accession=31447415, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  545276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545276, CurrentID=0, Status='live', Symbol='Gal3st3', Aliases='', description='galactose-3-O-sulfotransferase 3', other_designations="beta-galactose-3-O-sulfotransferase 3|gal-beta-1, 3-GalNAc 3'-sulfotransferase 3|gal3ST-3", map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5298331, end_position_on_the_genomic_accession=5308739, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  545279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545279, CurrentID=0, Status='live', Symbol='Ms4a15', Aliases='E530011F12Rik, Gm1286', description='membrane-spanning 4-domains, subfamily A, member 15', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10978303, end_position_on_the_genomic_accession=10993306, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  20997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20997, CurrentID=0, Status='live', Symbol='T', Aliases='Bra, D17Mit170, Low, Lr1, Tl2, Tl3, cou, me75, T', description='brachyury', other_designations='brachyury-like 2|brachyury-like 3|low ratio|protein T', map_location='17 4.92 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8422178, end_position_on_the_genomic_accession=8442496, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20998, CurrentID=0, Status='live', Symbol='t', Aliases='btm', description='t-complex', other_designations='brachyury interacting tail length modifier', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  545288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545288, CurrentID=0, Status='live', Symbol='Cyp2c67', Aliases='C730004C24Rik', description='cytochrome P450, family 2, subfamily c, polypeptide 67', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39608844, end_position_on_the_genomic_accession=39649083, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  545291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545291, CurrentID=0, Status='live', Symbol='Hpse2', Aliases='Gm968, Hpa2', description='heparanase 2', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42786541, end_position_on_the_genomic_accession=43388569, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  545366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545366, CurrentID=0, Status='live', Symbol='Cfhr2', Aliases='Cfhrb_4/2', description='complement factor H-related 2', other_designations='', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=139810292, end_position_on_the_genomic_accession=139858699, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  545370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545370, CurrentID=0, Status='live', Symbol='Hmcn1', Aliases='EG545370, Gm201', description='hemicentin 1', other_designations='', map_location='1 G1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=150562500, end_position_on_the_genomic_accession=150993687, orientation='minus', exon_count=108, OMIM='', no_hdr0=''),
  545378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545378, CurrentID=0, Status='live', Symbol='Sh2d1b2', Aliases='EAT-2B, Eat2b, Ert, Sh2d1c', description='SH2 domain containing 1B2', other_designations='EAT-2-related transducer|SH2 domain protein 1B2', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170232871, end_position_on_the_genomic_accession=170251587, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69051, CurrentID=0, Status='live', Symbol='Pycr2', Aliases='1810018M05Rik, P5cr2', description='pyrroline-5-carboxylate reductase family, member 2', other_designations='P5C reductase 2|P5CR 2', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180904274, end_position_on_the_genomic_accession=180908088, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  545384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545384, CurrentID=0, Status='live', Symbol='BC094916', Aliases='Ifi214, p214', description='cDNA sequence BC094916', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173521311, end_position_on_the_genomic_accession=173535957, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21098, CurrentID=0, Status='live', Symbol='T(16;17)43H', Aliases='T43H', description='reciprocal translocation, Chr 16 and 17, Harwell 43', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21099, CurrentID=0, Status='live', Symbol='T(16;17)43H', Aliases='T43H', description='reciprocal translocation, Chr 16 and 17, Harwell 43', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  545389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545389, CurrentID=0, Status='live', Symbol='Cep170', Aliases='4933426L22Rik, A330004A13Rik, AI195353', description='centrosomal protein 170', other_designations='KARP-1-binding protein|centrosomal protein 170kDa', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=176733653, end_position_on_the_genomic_accession=176814173, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  545417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545417, CurrentID=0, Status='live', Symbol='Olfr367-ps', Aliases='EG545417, MOR128-1, ORL1283, OTTMUSG00000012147, Olfr367', description='olfactory receptor 367, pseudogene', other_designations='GA_x6K02T2NLDC-33954853-33955794|olfactory receptor MOR128-1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37266951, end_position_on_the_genomic_accession=37271387, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  545428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545428, CurrentID=0, Status='live', Symbol='Ccdc141', Aliases='2610301F02Rik, CAMDI, ENSMUSG00000075261', description='coiled-coil domain containing 141', other_designations='coiled-coil protein associated with myosin II and DISC1', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=77009902, end_position_on_the_genomic_accession=77176062, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  21148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21148, CurrentID=0, Status='live', Symbol='T(2;8)26H', Aliases='T(5;18)26H, T26H', description='reciprocal translocation, Chr 2 and 8, Harwell 26', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21149, CurrentID=0, Status='live', Symbol='T(2;8)26H', Aliases='T(5;18)26H, T26H', description='reciprocal translocation, Chr 2 and 8, Harwell 26', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21151, CurrentID=0, Status='live', Symbol='T(2;8)2Wa', Aliases='T2Wa', description='reciprocal translocation, Chr 2 and 8, Wageningen 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  545471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545471, CurrentID=0, Status='live', Symbol='Zfp345', Aliases='OTTMUSG00000015743', description='zinc finger protein 345', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150470517, end_position_on_the_genomic_accession=150485077, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  545474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545474, CurrentID=0, Status='live', Symbol='Scrt2', Aliases='AW060659', description='scratch homolog 2, zinc finger protein (Drosophila)', other_designations='ortholog of human scratch homolog 2, zinc finger protein SCRT2|scratch homolog 2 zinc finger protein', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152081529, end_position_on_the_genomic_accession=152095802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  545475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545475, CurrentID=0, Status='live', Symbol='Defb28', Aliases='Gm1635, OTTMUSG00000015852', description='defensin beta 28', other_designations='beta-defensin 28', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152518255, end_position_on_the_genomic_accession=152521447, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  545477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545477, CurrentID=0, Status='live', Symbol='Bpifa6', Aliases='EG545477, Gm5840, Splunc6', description='BPI fold containing family A, member 6', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153974945, end_position_on_the_genomic_accession=154000495, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  545486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545486, CurrentID=0, Status='live', Symbol='Tubb1', Aliases='2810484G07Rik, M(beta)1', description='tubulin, beta 1 class VI', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174450595, end_position_on_the_genomic_accession=174458380, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  545527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545527, CurrentID=0, Status='live', Symbol='Erich6', Aliases='4932431H17Rik, Fam194a', description='glutamate rich 6', other_designations='family with sequence similarity 194, member A|protein FAM194A', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=58616300, end_position_on_the_genomic_accession=58665413, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  21258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21258, CurrentID=0, Status='live', Symbol='T(7;15)9H', Aliases='T9H', description='reciprocal translocation, Chr 7 and 15, Harwell 9', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21259, CurrentID=0, Status='live', Symbol='T(7;15)9H', Aliases='T9H', description='reciprocal translocation, Chr 7 and 15, Harwell 9', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  545548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545548, CurrentID=0, Status='live', Symbol='Lce3a', Aliases='EG545548', description='late cornified envelope 3A', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92925287, end_position_on_the_genomic_accession=92926230, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  21266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21266, CurrentID=0, Status='live', Symbol='T(7;18)50H', Aliases='T50H', description='reciprocal translocation, Chr 7 and 18, Harwell 50', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21267, CurrentID=0, Status='live', Symbol='T(7;18)50H', Aliases='T50H', description='reciprocal translocation, Chr 7 and 18, Harwell 50', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  545562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545562, CurrentID=0, Status='live', Symbol='Amy2b', Aliases='Amy-X, Amy2-2, OTTMUSG00000022462, mAmy2-1', description='amylase 2b', other_designations='amylase 2-2, pancreatic|amylase 2b, pancreatic', map_location='3 49.35 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113147656, end_position_on_the_genomic_accession=113156727, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  21297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21297, CurrentID=0, Status='live', Symbol='T(X;16)16H', Aliases='T16H', description='reciprocal translocation, Chr X and 16, Harwell 16', other_designations="Searle's translocation", map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21298, CurrentID=0, Status='live', Symbol='T(X;16)16H', Aliases='T16H', description='reciprocal translocation, Chr X and 16, Harwell 16', other_designations="Searle's translocation", map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21331, CurrentID=0, Status='live', Symbol='T2', Aliases='brachyury2', description='brachyury 2', other_designations='brachyury the second', map_location='17 4.92 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8372015, end_position_on_the_genomic_accession=8422728, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21333, CurrentID=0, Status='live', Symbol='Tac1', Aliases='4930528L02Rik, NK-1, NK1, Nkna, PPT-A, PPTA, SP', description='tachykinin 1', other_designations='neurokinin 1|neurokinin A|preprotachykinin A|substance P', map_location='6 3.31 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=7554776, end_position_on_the_genomic_accession=7562978, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21334, CurrentID=0, Status='live', Symbol='Tac2', Aliases='PPT-B, Tac3', description='tachykinin 2', other_designations='neurokinin 2|neurokinin A|neurokinin B-like protein Zneurok1|neurokinin alpha|neuromedin L|neuropeptide K|preprotachykinin-B|substance K', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127724478, end_position_on_the_genomic_accession=127731768, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21335, CurrentID=0, Status='live', Symbol='Tacc3', Aliases='Aint, C86661, Eric1', description='transforming, acidic coiled-coil containing protein 3', other_designations='ARNT-interacting protein', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33658147, end_position_on_the_genomic_accession=33672203, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  21336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21336, CurrentID=0, Status='live', Symbol='Tacr1', Aliases='Nk1r, Spr, Tac1r', description='tachykinin receptor 1', other_designations='Nk1 receptor|neurokinin receptor 1|substance P receptor', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=82402005, end_position_on_the_genomic_accession=82560104, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21337, CurrentID=0, Status='live', Symbol='Tacr2', Aliases='Nk2r, Skr, Tac2r', description='tachykinin receptor 2', other_designations='7 transmembrane receptor|NK-2 receptor|NK-2R|neurokinin A receptor|substance K receptor|tachykinin NK-2 receptor', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62252438, end_position_on_the_genomic_accession=62265990, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21338, CurrentID=0, Status='live', Symbol='Tacr3', Aliases='Nk3r, Nkr, Tac3r', description='tachykinin receptor 3', other_designations='NK-3 receptor|Nk3 receptor|neurokinin 3 receptor|neurokinin B receptor|neuromedin K receptor', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=134829007, end_position_on_the_genomic_accession=134934581, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21339, CurrentID=0, Status='live', Symbol='Taf1a', Aliases='TAFI48, mTAFI48', description='TATA-box binding protein associated factor, RNA polymerase I, A', other_designations='RNA polymerase I-specific TBP-associated factor 48 kDa|TATA box binding protein (Tbp)-associated factor, RNA polymerase I, A|TATA box-binding protein-associated factor 1A|TBP-associated factor 1A|transcription initiation factor SL1/TIF-IB subunit A', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=183388885, end_position_on_the_genomic_accession=183410201, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21340, CurrentID=0, Status='live', Symbol='Taf1b', Aliases='4930408G01Rik, A230108M10Rik, TAFI68, Tafi86, p63', description='TATA-box binding protein associated factor, RNA polymerase I, B', other_designations='RNA polymerase I-specific TBP-associated factor 68 kDa|TATA box binding protein (Tbp)-associated factor, RNA polymerase I, B|TATA box-binding protein-associated factor 1B|TBP-associated factor 1B|TBP-associated factor RNA polymerase I 68|transcription factor TAFI68|transcription initiation factor SL1/TIF-IB subunit B', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24498581, end_position_on_the_genomic_accession=24558571, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  21341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21341, CurrentID=0, Status='live', Symbol='Taf1c', Aliases='Tafi95, mTAFI95', description='TATA-box binding protein associated factor, RNA polymerase I, C', other_designations='RNA polymerase I-specific TBP-associated factor 95 kDa|TATA box binding protein (Tbp)-associated factor, RNA polymerase I, C|TATA box binding protein (Tbp)-associated factor, RNA polymerase I, C; transcription factor TAFI95|TATA box-binding protein-associated factor 1C|TBP-associated factor 1C|TBP-associated factor RNA polymerase I 95 kDa|transcription factor TAFI95|transcription initiation factor SL1/TIF-IB subunit C', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119597974, end_position_on_the_genomic_accession=119605240, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  21343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21343, CurrentID=0, Status='live', Symbol='Taf6', Aliases='80kDa, AW549759, TAF(II)80, TAFII70, Taf2e, p80', description='TATA-box binding protein associated factor 6', other_designations='TAF(II)70|TAF6 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF6 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 80 kDa|TAFII-70|TAFII-80|TAFII80|TATA box binding protein (Tbp)-associated factor, RNA polymerase II, E|transcription initiation factor TFIID 70 kDa subunit|transcription initiation factor TFIID 80 kDa subunit', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138178617, end_position_on_the_genomic_accession=138187446, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  21345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21345, CurrentID=0, Status='live', Symbol='Tagln', Aliases='Sm22, Sm22a, Ws310', description='transgelin', other_designations='SM22-alpha|Sm22alpha|actin-associated protein p27|smooth muscle protein 22-alpha', map_location='9 25.36 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45929628, end_position_on_the_genomic_accession=45936058, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21346, CurrentID=0, Status='live', Symbol='Tagln2', Aliases='2700094C18Rik, SM22beta, Sm22B, Sm22a', description='transgelin 2', other_designations='SM22-beta', map_location='1 79.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172500246, end_position_on_the_genomic_accession=172507375, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21349, CurrentID=0, Status='live', Symbol='Tal1', Aliases='SCL/tal-1, Scl, bHLHa17, tal-1', description='T cell acute lymphocytic leukemia 1', other_designations='T-cell acute lymphocytic leukemia 1|stem cell protein', map_location='4 52.73 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115056426, end_position_on_the_genomic_accession=115071755, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21350, CurrentID=0, Status='live', Symbol='Tal2', Aliases='bHLHa19', description='T cell acute lymphocytic leukemia 2', other_designations='T-cell acute lymphocytic leukemia 2|TAL-2', map_location='4 28.75 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=53779705, end_position_on_the_genomic_accession=53786885, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  21351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21351, CurrentID=0, Status='live', Symbol='Taldo1', Aliases='', description='transaldolase 1', other_designations='', map_location='7 F3-F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141392160, end_position_on_the_genomic_accession=141402976, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21353, CurrentID=0, Status='live', Symbol='Tank', Aliases='C86182, E430026L09Rik, I-TRAF', description='TRAF family member-associated Nf-kappa B activator', other_designations='TRAF-interacting protein', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=61578561, end_position_on_the_genomic_accession=61654171, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  21354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21354, CurrentID=0, Status='live', Symbol='Tap1', Aliases='ABC17, APT1, Abcb2, Ham-1, Ham1, MTP1, PSF1, RING4, TAP, Tap-1, Y3', description='transporter 1, ATP-binding cassette, sub-family B (MDR/TAP)', other_designations='ABC transporter, MHC, 1|ATP dependent transport protein family member|ATP-binding cassette sub-family B member 2|ATP-binding cassette transporter, major histocompatibility complex, 1|ATP-binding cassette, sub-family B (MDR/TAP), member 2|histocompatibility antigen modifier 1|peptide supply factor 1|peptide transporter PSF1|peptide transporter TAP1|transporter 1 ATP-binding cassette sub-family B|transporter 1, ABC (ATP binding cassette)|transporter associated with antigen processing 1|transporter, ABC, MHC, 1|transporter, ATP-binding cassette, major histocompatibility complex, 1', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34187556, end_position_on_the_genomic_accession=34197225, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21355, CurrentID=0, Status='live', Symbol='Tap2', Aliases='ABC18, AI462429, APT2, Abcb3, Ham-2, Ham2, MTP2, PSF2, RING11, Tap-2, Y1, jas', description='transporter 2, ATP-binding cassette, sub-family B (MDR/TAP)', other_designations='ATP dependent transport protein family member|ATP-binding cassette sub-family B member 3|ATP-binding cassette, sub-family B (MDR/TAP), member 3|histocompatibility antigen modifier 2|transporter 2 ATP-binding cassette sub-family B|transporter 2, ABC (ATP binding cassette)', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34204085, end_position_on_the_genomic_accession=34216321, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  21356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21356, CurrentID=0, Status='live', Symbol='Tapbp', Aliases='D17Wsu91e, TPN', description='TAP binding protein', other_designations='TAP-associated protein|TAP-binding protein|TPSN', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33919478, end_position_on_the_genomic_accession=33929290, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21357, CurrentID=0, Status='live', Symbol='Tarbp2', Aliases='Prbp, TRBP', description='TAR (HIV) RNA binding protein 2', other_designations='PRM-1 RNA-binding protein|TAR RNA-binding protein 2|protamine-1 RNA-binding protein', map_location='15 F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102518192, end_position_on_the_genomic_accession=102523676, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  545648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545648, CurrentID=0, Status='live', Symbol='Gm13272', Aliases='OTTMUSG00000011265', description='predicted gene 13272', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88779850, end_position_on_the_genomic_accession=88780660, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  21366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21366, CurrentID=0, Status='live', Symbol='Slc6a6', Aliases='AA589629, C80501, Taut', description='solute carrier family 6 (neurotransmitter transporter, taurine), member 6', other_designations='solute carrier family 6 member 6|taurine/beta-alanine transporter', map_location='6 40.73 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91684044, end_position_on_the_genomic_accession=91759063, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  545655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545655, CurrentID=0, Status='live', Symbol='Gm13287', Aliases='OTTMUSG00000011285', description='predicted gene 13287', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88803254, end_position_on_the_genomic_accession=88803802, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  21371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21371, CurrentID=0, Status='live', Symbol='Tbca', Aliases='Tbca13', description='tubulin cofactor A', other_designations='CFA|TCP1-chaperonin cofactor A|co-chaperonin associated with a- & b-tubulin|cofactor A|tubulin-folding cofactor A', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=94788943, end_position_on_the_genomic_accession=94842899, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  21372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21372, CurrentID=0, Status='live', Symbol='Tbl1x', Aliases='5330429M20, Tbl1', description='transducin (beta)-like 1 X-linked', other_designations='transducin beta-like protein 1X', map_location='X 38.31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=77510547, end_position_on_the_genomic_accession=77662985, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  21374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21374, CurrentID=0, Status='live', Symbol='Tbp', Aliases='GTF2D1, Gtf2d, SCA17, TFIID', description='TATA box binding protein', other_designations='TATA sequence-binding protein|TATA-binding factor|TATA-binding protein|TATA-box factor|TFIID core protein|transcription initiation factor TFIID TBP subunit', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15499888, end_position_on_the_genomic_accession=15517427, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21375, CurrentID=0, Status='live', Symbol='Tbr1', Aliases='', description='T-box brain gene 1', other_designations='T-brain-1|TBR-1|TES-56', map_location='2 35.56 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=61802734, end_position_on_the_genomic_accession=61814114, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  21376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21376, CurrentID=0, Status='live', Symbol='Tbrg1', Aliases='AA408552, AA409675, Niam, Tb-5, mFLJ00213', description='transforming growth factor beta regulated gene 1', other_designations='nuclear interactor of ARF and Mdm2', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37649182, end_position_on_the_genomic_accession=37657312, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  21379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21379, CurrentID=0, Status='live', Symbol='Tbrg4', Aliases='2310042P22Rik, AA120735, AA408001, AI527316, Cpr2, R74877, Tb-12, mKIAA0948', description='transforming growth factor beta regulated gene 4', other_designations='cell cycle progression 2 protein|transforming growth factor beta regulator 4', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6615598, end_position_on_the_genomic_accession=6626084, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  21380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21380, CurrentID=0, Status='live', Symbol='Tbx1', Aliases='', description='T-box 1', other_designations='T-box protein 1|testis-specific T-box protein', map_location='16 11.51 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18581704, end_position_on_the_genomic_accession=18590671, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  21384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21384, CurrentID=0, Status='live', Symbol='Tbx15', Aliases='Tbx14, Tbx8, de', description='T-box 15', other_designations='T-box 14|T-box protein 14|T-box protein 15|T-box transcription factor TBX14|droopy ear|mmTBx8', map_location='3 43.03 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=99240347, end_position_on_the_genomic_accession=99354260, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21385, CurrentID=0, Status='live', Symbol='Tbx2', Aliases='', description='T-box 2', other_designations='T-box protein 2', map_location='11 51.34 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=85832615, end_position_on_the_genomic_accession=85841948, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  21386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21386, CurrentID=0, Status='live', Symbol='Tbx3', Aliases='D5Ertd189e', description='T-box 3', other_designations='T-box protein 3', map_location='5 60.34 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=119670160, end_position_on_the_genomic_accession=119684724, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21387, CurrentID=0, Status='live', Symbol='Tbx4', Aliases='3930401C23', description='T-box 4', other_designations='T-box protein 4', map_location='11 51.42 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=85886343, end_position_on_the_genomic_accession=85916099, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21388, CurrentID=0, Status='live', Symbol='Tbx5', Aliases='', description='T-box 5', other_designations='T-box protein 5', map_location='5 60.42 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=119796515, end_position_on_the_genomic_accession=119885220, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  21389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21389, CurrentID=0, Status='live', Symbol='Tbx6', Aliases='rv', description='T-box 6', other_designations='T-box protein 6|rib-vertebrae', map_location='7 69.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126781483, end_position_on_the_genomic_accession=126785548, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21390, CurrentID=0, Status='live', Symbol='Tbxa2r', Aliases='TP, TXA2-R', description='thromboxane A2 receptor', other_designations='TXA2 receptor|prostanoid TP receptor|thromboxane prostanoid', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81328259, end_position_on_the_genomic_accession=81335184, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21391, CurrentID=0, Status='live', Symbol='Tbxas1', Aliases='CYP5, CYP5A1, THAS, TS, TXAS, TXS', description='thromboxane A synthase 1, platelet', other_designations='TXA synthase|cytochrome P450 5A1', map_location='6 F1-pter', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38918467, end_position_on_the_genomic_accession=39084590, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  21393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21393, CurrentID=0, Status='live', Symbol='Tcap', Aliases='AI503981, AJ223855', description='titin-cap', other_designations='titin cap protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98383811, end_position_on_the_genomic_accession=98384953, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  21399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21399, CurrentID=0, Status='live', Symbol='Tcea1', Aliases='S-II', description='transcription elongation factor A (SII) 1', other_designations='transcription elongation factor S-II protein 1|transcription elongation factor TFIIS.o', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=4857694, end_position_on_the_genomic_accession=4897909, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21400, CurrentID=0, Status='live', Symbol='Tcea2', Aliases='AI326274, S-II-T1, SII-T1, Tceat', description='transcription elongation factor A (SII), 2', other_designations='testis-specific S-II|transcription elongation factor S-II protein 2|transcription elongation factor TFIIS.l', map_location='2 103.72 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181680281, end_position_on_the_genomic_accession=181688069, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21401, CurrentID=0, Status='live', Symbol='Tcea3', Aliases='S-II, SII-K1', description='transcription elongation factor A (SII), 3', other_designations='tissue-specific transcription factor S-II|transcription elongation factor S-II protein 3|transcription elongation factor TFIIS.h', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136247936, end_position_on_the_genomic_accession=136274903, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21402, CurrentID=0, Status='live', Symbol='Skp1a', Aliases='15kDa, 2610043E24Rik, 2610206H23Rik, EMC19, OCP-II, OCP2, SKP1, Tceb1l, p19A, p19Skp1', description='S-phase kinase-associated protein 1A', other_designations='cyclin A/CDK2-associated protein p19|cyclin-A/CDK2-associated protein p19', map_location='11 31.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52230522, end_position_on_the_genomic_accession=52246858, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  21405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21405, CurrentID=0, Status='live', Symbol='Hnf1a', Aliases='AI323641, HNF1, HNF1-alpha, HNF1[a], Hnf-1lpha, LFB1, Tcf-1, Tcf1, Hnf1a', description='HNF1 homeobox A', other_designations='HNF-1-alpha|HNF-1A|hepatocyte nuclear factor 1|liver-specific transcription factor LF-B1|transcription factor 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114948980, end_position_on_the_genomic_accession=114971062, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  21406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21406, CurrentID=0, Status='live', Symbol='Tcf12', Aliases='A130037E08Rik, ALF1, HEB, HEBAlt, HTF-4, HTF4, ME1, REB, bHLHb20', description='transcription factor 12', other_designations='DNA-binding protein HTF4|E-box-binding protein|TCF-12|class A helix-loop-helix transcription factor ME1|transcription factor HTF-4', map_location='9 39.85 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=71844252, end_position_on_the_genomic_accession=72111819, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  21407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21407, CurrentID=0, Status='live', Symbol='Tcf15', Aliases='Meso1, bHLH-EC2', description='transcription factor 15', other_designations='TCF-15|paraxis|protein bHLH-EC2', map_location='2 74.83 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152143046, end_position_on_the_genomic_accession=152149097, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21408, CurrentID=0, Status='live', Symbol='Zfp354a', Aliases='AW488485, Tcf17, Znf354a, kid1', description='zinc finger protein 354A', other_designations='TCF-17|kidney, ischemia, and developmentally-regulated protein 1|renal transcription factor Kid-1|transcription factor 17', map_location='11 30.97 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51059244, end_position_on_the_genomic_accession=51072799, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21410, CurrentID=0, Status='live', Symbol='Hnf1b', Aliases='HNF-1-beta, HNF-1B, HNF-1Betaeta, LFB3, Tcf-2, Tcf2, vHNF1, Hnf1b', description='HNF1 homeobox B', other_designations='hepatocyte nuclear factor-1 beta|homeoprotein LFB3|transcription factor 2', map_location='11 51.23 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83850209, end_position_on_the_genomic_accession=83905917, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21411, CurrentID=0, Status='live', Symbol='Tcf20', Aliases='2810438H08Rik, SPBP, mKIAA0292', description='transcription factor 20', other_designations='SPRE-binding protein|TCF-20|nuclear factor SPBP|stromelysin 1 PDGF responsive element binding protein|stromelysin 1 PDGF-responsive element-binding protein|stromelysin PDGF responsive element binding protein transcription factor|stromelysin-1 PDGF-responsive element-binding protein', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82808447, end_position_on_the_genomic_accession=82987867, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  21412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21412, CurrentID=0, Status='live', Symbol='Tcf21', Aliases='Pod-1, Pod1, bHLHa23, capsulin, epc, epicardin', description='transcription factor 21', other_designations='TCF-21|podocyte-expressed 1', map_location='10 10.33 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=22817263, end_position_on_the_genomic_accession=22820128, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  21413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21413, CurrentID=0, Status='live', Symbol='Tcf4', Aliases='5730422P05Rik, ASP-I2, E2-2, E2.2, ITF-2, ITF-2b, ITF2, ME2, MITF-2A, MITF-2B, SEF-2, SEF2, SEF2-1, TFE, Tcf-4, bHLHb19', description='transcription factor 4', other_designations='MITF-2|SL3-3 enhancer factor 2|Transcription factor 4 (Immunoglobulin transcription factor 2) (ITF-2) (MITF-2) (SL3-3 enhancer factor 2) (SEF-2) (Class A helix-loop-helix transcription factor ME2)|class A helix-loop-helix transcription factor ME2|immunoglobulin transcription factor 2|immunoglobulin transcription factor-2', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=69344490, end_position_on_the_genomic_accession=69687967, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  21414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21414, CurrentID=0, Status='live', Symbol='Tcf7', Aliases='AI465550, TCF-1, Tcf1', description='transcription factor 7, T cell specific', other_designations='T cell factor-1|T-cell factor 1|T-cell-specific transcription factor 1|transcription factor 7, T-cell specific', map_location='11 31.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52252283, end_position_on_the_genomic_accession=52283378, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  21415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21415, CurrentID=0, Status='live', Symbol='Tcf7l1', Aliases='Tcf-3, Tcf3, bHLHb21', description='transcription factor 7 like 1 (T cell specific, HMG box)', other_designations='HMG box transcription factor 3|T-cell factor 3|transcription factor 3|transcription factor 7-like 1 (T-cell specific, HMG box)|transcription factor 7-like 2 (T-cell specific, HMG box)', map_location='6 32.27 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72626371, end_position_on_the_genomic_accession=72789045, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  21416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21416, CurrentID=0, Status='live', Symbol='Tcf7l2', Aliases='TCF4B, TCF4E, Tcf-4, Tcf4, mTcf-4B, mTcf-4E', description='transcription factor 7 like 2, T cell specific, HMG box', other_designations='HMG box transcription factor 4|HMG box transcription factor TCF7L2|T-cell factor 4|T-cell-specific transcription factor 4|mTCF-4|transcription factor 7-like 2, T-cell specific, HMG-box', map_location='19 51.59 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=55741810, end_position_on_the_genomic_accession=55933661, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  21417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21417, CurrentID=0, Status='live', Symbol='Zeb1', Aliases='3110032K11Rik, AREB6, BZP, MEB1, Nil2, TCF-8, Tcf18, Tcf8, Tw, ZEB, Zfhep, Zfhx1a, Zfx1a, Zfx1ha, [delta]EF1', description='zinc finger E-box binding homeobox 1', other_designations='transcription factor 8|twirler|zinc finger homeobox protein 1a', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=5591860, end_position_on_the_genomic_accession=5775468, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  21418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21418, CurrentID=0, Status='live', Symbol='Tfap2a', Aliases='AP-2, AP2alpha, Ap-2 (a), Ap2, Ap2tf, Tcfap2a', description='transcription factor AP-2, alpha', other_designations='activating enhancer-binding protein 2 alpha|activator protein 2', map_location='13 A5-B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=40713802, end_position_on_the_genomic_accession=40738238, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  21419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21419, CurrentID=0, Status='live', Symbol='Tfap2b', Aliases='AI606113, AP-2(beta), E130018K07Rik, Tcfap2b', description='transcription factor AP-2 beta', other_designations='AP2-beta|activating enhancer-binding protein 2 beta', map_location='1 A2-A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=19208914, end_position_on_the_genomic_accession=19238845, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21420, CurrentID=0, Status='live', Symbol='Tfap2c', Aliases='AA409384, AP2gamma, Ap-2.2, Stra2, Tcfap2c', description='transcription factor AP-2, gamma', other_designations='AP2-gamma|activating enhancer-binding protein 2 gamma|stimulated by retinoic acid gene 2', map_location='2 H3-H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172545678, end_position_on_the_genomic_accession=172558622, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21422, CurrentID=0, Status='live', Symbol='Tfcp2', Aliases='CP-2, CP2, D230015P20Rik, LBP-1c, LBP-1d, LBP1, LSF, Tcfcp2, UBP-1', description='transcription factor CP2', other_designations='', map_location='15 56.33 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100502748, end_position_on_the_genomic_accession=100552014, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  21423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21423, CurrentID=0, Status='live', Symbol='Tcf3', Aliases='A1, AA408400, ALF2, AW209082, E12, E12/E47, E2A, E47, ME2, Pan1, Pan2, TCF-3, Tcfe2a, VDIR, bHLHb21', description='transcription factor 3', other_designations='immunoglobulin enhancer-binding factor E12/E47|transcription factor A1|transcription factor E2a', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80409165, end_position_on_the_genomic_accession=80433653, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  21425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21425, CurrentID=0, Status='live', Symbol='Tfeb', Aliases='Tcfeb, bHLHe35', description='transcription factor EB', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47737037, end_position_on_the_genomic_accession=47792416, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  21426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21426, CurrentID=0, Status='live', Symbol='Tfec', Aliases='BB107417, Tcfec, bHLHe34', description='transcription factor EC', other_designations='TFE-C|mTFEC', map_location='6 E4-G', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=16830837, end_position_on_the_genomic_accession=16898462, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  21427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21427, CurrentID=0, Status='live', Symbol='Vps72', Aliases='Tcfl1, YL-1', description='vacuolar protein sorting 72 (yeast)', other_designations='protein YL-1|transcription factor-like 1', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95111042, end_position_on_the_genomic_accession=95123051, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  21428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21428, CurrentID=0, Status='live', Symbol='Mlx', Aliases='Tcfl4, bHLHd13', description='MAX-like protein X', other_designations='BigMax alpha|MAX-like bHLHZIP protein|protein BigMax|transcription factor like 4|transcription factor-like 4|transcription factor-like protein 4', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101086709, end_position_on_the_genomic_accession=101092456, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21429, CurrentID=0, Status='live', Symbol='Ubtf', Aliases='A930005G04Rik, NOR-90, Tcfubf, UBF, UBF-1, UBF1', description='upstream binding transcription factor, RNA polymerase I', other_designations='transcription factor UBF|upstream-binding factor 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102304563, end_position_on_the_genomic_accession=102319757, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  21430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21430, CurrentID=0, Status='live', Symbol='tch', Aliases='', description='tich', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21432, CurrentID=0, Status='live', Symbol='Tcl1', Aliases='Tcl1a', description='T cell lymphoma breakpoint 1', other_designations='T-cell lymphoma breakpoint 1|TCL-1|oncogene TCL-1|oncogene TCL1|protein p14 TCL1', map_location='12 55.28 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105216753, end_position_on_the_genomic_accession=105222737, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  545725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545725, CurrentID=0, Status='live', Symbol='Mterf1a', Aliases='4931431L11Rik, 9230106K09Rik, Mterf, Mterf1, mTERF1b', description='mitochondrial transcription termination factor 1a', other_designations='mitochondrial transcription termination factor 1|mitochondrial transcription termination factor 1b|mterf1b|transcription termination factor, mitochondrial', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3863845, end_position_on_the_genomic_accession=3894203, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  545732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545732, CurrentID=0, Status='live', Symbol='4933402N22Rik', Aliases='', description='RIKEN cDNA 4933402N22 gene', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=11918043, end_position_on_the_genomic_accession=11922804, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21452, CurrentID=0, Status='live', Symbol='Tcn2', Aliases='AW208754, Tcn-2', description='transcobalamin 2', other_designations='TC II|TC-2|TCII|transcobalamin II', map_location='11 2.76 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3917077, end_position_on_the_genomic_accession=3932078, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  21453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21453, CurrentID=0, Status='live', Symbol='Tcof1', Aliases='AA408847, AW209012, treacle', description='Treacher Collins Franceschetti syndrome 1, homolog', other_designations='treacher Collins syndrome protein homolog', map_location='18 34.41 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60813736, end_position_on_the_genomic_accession=60848983, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  21454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21454, CurrentID=0, Status='live', Symbol='Tcp1', Aliases='AI528772, CCT, Cct1, Ccta, TRic, Tcp-1, Tp63, c-cpn, p63', description='t-complex protein 1', other_designations='CCT-alpha|T-complex protein 1 subunit alpha A|T-complex protein 1 subunit alpha B|TCP-1-A|TCP-1-B|TCP-1-alpha|t-complex polypeptide 1|tailless complex polypeptide 1A|tailless complex polypeptide 1B', map_location='17 A3-B', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12916329, end_position_on_the_genomic_accession=12925067, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21458, CurrentID=0, Status='live', Symbol='Tcp10', Aliases='Tcp-10', description='t-complex protein 10', other_designations='t-complex protein-10 complex', map_location='17 A2-B', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21461, CurrentID=0, Status='live', Symbol='Tcp10a', Aliases='D17Leh66A, D17Leh66aa, Gm10326, T66A-a, Tcp-10a', description='t-complex protein 10a', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=7324656, end_position_on_the_genomic_accession=7345860, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21462, CurrentID=0, Status='live', Symbol='Tcp10b', Aliases='D17Leh66B, D17Leh66ba, T66B-a, Tcp-10b, Tcp-10bt', description='t-complex protein 10b', other_designations='', map_location='17 7.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13061109, end_position_on_the_genomic_accession=13082226, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21463, CurrentID=0, Status='live', Symbol='Tcp11', Aliases='D17Ken1, Tcp-11', description='t-complex protein 11', other_designations='testis-specific protein PBS13', map_location='17 A3-B', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28066747, end_position_on_the_genomic_accession=28080772, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  21473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21473, CurrentID=0, Status='live', Symbol='Tcra', Aliases='Tcralpha', description='T cell receptor alpha chain', other_designations='T-cell receptor alpha chain|Tcr alpha', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52427967, end_position_on_the_genomic_accession=54224198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  545762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545762, CurrentID=0, Status='live', Symbol='Gm16367', Aliases='', description='predicted gene 16367', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21475, CurrentID=0, Status='live', Symbol='Tcra-J', Aliases='', description='T cell receptor alpha, joining region', other_designations='T-cell receptor alpha, joining region', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21507, CurrentID=0, Status='live', Symbol='Trav5-4', Aliases='Tcra-V13.1', description='T cell receptor alpha variable 5-4', other_designations='T cell receptor alpha chain AV13S3|T-cell receptor alpha chain variable region 5-4|T-cell receptor alpha variable region family 13 subfamily 1', map_location='14 27.48 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53704007, end_position_on_the_genomic_accession=53704514, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  545798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545798, CurrentID=0, Status='live', Symbol='Tmem233', Aliases='DSPB2, Gm13843, OTTMUSG00000014722', description='transmembrane protein 233', other_designations='dispanin subfamily B member 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=116040532, end_position_on_the_genomic_accession=116083246, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  545812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545812, CurrentID=0, Status='live', Symbol='Pilrb2', Aliases='EG545812', description='paired immunoglobin-like type 2 receptor beta 2', other_designations='activating receptor PILR-beta-2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137865821, end_position_on_the_genomic_accession=137871914, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  545814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545814, CurrentID=0, Status='live', Symbol='Smok3a', Aliases='Smok2b1', description='sperm motility kinase 3A', other_designations='sperm motility kinase 2B1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138021276, end_position_on_the_genomic_accession=138034665, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  414758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=414758, CurrentID=0, Status='live', Symbol='Zfp950', Aliases='5830428H23Rik, AU016073, BC055757, Gm34518, Zfp826, mszf84', description='zinc finger protein 950', other_designations='zinc finger protein 826|zinc finger prtoein 950', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=61051147, end_position_on_the_genomic_accession=61313507, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  545847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545847, CurrentID=0, Status='live', Symbol='Igkv14-111', Aliases='Gm16939', description='immunoglobulin kappa variable 14-111', other_designations='Ig kappa V-region|Ig kappa chain V-V region L6|Ig kappa chain V-V region T1|immunoglobulin kappa chain variable 14-111', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68256404, end_position_on_the_genomic_accession=68256869, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  545851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545851, CurrentID=0, Status='live', Symbol='Igkv12-44', Aliases='Gm16851', description='immunoglobulin kappa variable 12-44', other_designations='immunoglobulin kappa chain variable 12-44|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69814631, end_position_on_the_genomic_accession=69815100, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  21577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21577, CurrentID=0, Status='live', Symbol='Tcrb', Aliases='TCRbeta, Tib', description='T cell receptor beta chain', other_designations='T-cell receptor beta chain', map_location='6 A-C', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40891296, end_position_on_the_genomic_accession=41558371, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  21580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21580, CurrentID=0, Status='live', Symbol='Tcrb-J', Aliases='5830405F06Rik', description='T cell receptor beta, joining region', other_designations='T-cell receptor beta, joining region', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  414801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=414801, CurrentID=0, Status='live', Symbol='Itprip', Aliases='4833424O12Rik, DANGER, mKIAA1754', description='inositol 1,4,5-triphosphate receptor interacting protein', other_designations='inositol 1,4,5-triphosphate receptor-interacting protein', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47894593, end_position_on_the_genomic_accession=47919519, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  545874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545874, CurrentID=0, Status='live', Symbol='Vmn2r25', Aliases='EG545874', description='vomeronasal 2, receptor 25', other_designations='', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=123822814, end_position_on_the_genomic_accession=123853190, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  545893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545893, CurrentID=0, Status='live', Symbol='Mansc4', Aliases='A430079E08, EG545893, Gm5887', description='MANSC domain containing 4', other_designations='MANSC domain-containing protein ENSP00000370673 homolog', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=147075062, end_position_on_the_genomic_accession=147087032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  545901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545901, CurrentID=0, Status='live', Symbol='Atp6v0c-ps1', Aliases='Atp6c, Atp6c-rs1, Atp6c1, Atp6c2, Atp6l-ps1', description='ATPase, H+ transporting, lysosomal V0 subunit C, pseudogene 1', other_designations='ATPase, H+ transporting, V0 subunit c, pseudogene 1|ATPase, H+ transporting, lysosomal (vacuolar proton pump), pseudogene 1|ATPase, H+ transporting, lysosomal 16kD, V0 subunit c, pseudogene 1|ATPase, H+ transporting, lysosomal, related sequence 1|MGPLlambda1-2|lysosomal 16kDa', map_location='7 2.55 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4412323, end_position_on_the_genomic_accession=4413440, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  545902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545902, CurrentID=0, Status='live', Symbol='Ptprh', Aliases='sap-1', description='protein tyrosine phosphatase, receptor type, H', other_designations='receptor type protein tyrosine phosphatase sap-1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4548614, end_position_on_the_genomic_accession=4604041, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  21616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21616, CurrentID=0, Status='live', Symbol='Tcrd-V1', Aliases='', description='T cell receptor delta, variable 1', other_designations='T-cell receptor delta, variable 1', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  545913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545913, CurrentID=0, Status='live', Symbol='Zscan4d', Aliases='EG545913', description='zinc finger and SCAN domain containing 4D', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11160883, end_position_on_the_genomic_accession=11167460, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  21632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21632, CurrentID=0, Status='live', Symbol='Tcrg-V1', Aliases='Gm16602, TRGV1', description='T cell receptor gamma, variable 1', other_designations='T-cell receptor gamma, variable 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19340016, end_position_on_the_genomic_accession=19340454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  21636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21636, CurrentID=0, Status='live', Symbol='Trgv2', Aliases='Tcrg-V2, Vgamma2', description='T cell receptor gamma variable 2', other_designations='T-cell receptor gamma, variable 2', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19336573, end_position_on_the_genomic_accession=19337026, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  21637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21637, CurrentID=0, Status='live', Symbol='Tcrg-V3', Aliases='TCRGV5S3, TRGV3, vgamma3', description='T cell receptor gamma, variable 3', other_designations='T-cell receptor gamma, variable 3|TCR V gamma 1.3', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19242845, end_position_on_the_genomic_accession=19243301, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  21638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21638, CurrentID=0, Status='live', Symbol='Tcrg-V4', Aliases='AI528624, TRGV4, Vg4', description='T cell receptor gamma, variable 4', other_designations='T-cell receptor gamma, variable 4', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19184981, end_position_on_the_genomic_accession=19185506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  21639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21639, CurrentID=0, Status='live', Symbol='Tcrg-V5', Aliases='TRGV5', description='T cell receptor gamma, variable 5', other_designations='T-cell receptor gamma, variable 5', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19192278, end_position_on_the_genomic_accession=19192725, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  21640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21640, CurrentID=0, Status='live', Symbol='Tcrg-V6', Aliases='TRGV6', description='T cell receptor gamma, variable 6', other_designations='T-cell receptor gamma, variable 6', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19190368, end_position_on_the_genomic_accession=19190929, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  545929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545929, CurrentID=0, Status='live', Symbol='Gm5891', Aliases='EG545929, ENSMUSG00000074313', description='predicted gene 5891', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=21909434, end_position_on_the_genomic_accession=21921291, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  21642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21642, CurrentID=0, Status='live', Symbol='tct1', Aliases='tct, tct-1', description='t complex tail interaction 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21645, CurrentID=0, Status='live', Symbol='Tcte1', Aliases='D17Sil1, Tcte-1', description='t-complex-associated testis expressed 1', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45523434, end_position_on_the_genomic_accession=45542679, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21646, CurrentID=0, Status='live', Symbol='Tcte2', Aliases='AI429268, Tcte-2', description='t-complex-associated testis expressed 2', other_designations='', map_location='17 8.9 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13716436, end_position_on_the_genomic_accession=13761394, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  21647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21647, CurrentID=0, Status='live', Symbol='Tcte3', Aliases='D17Leh117c, LC2, Tcte-3, Tctex-2, Tctex-4, Tctex2, Tctex4', description='t-complex-associated testis expressed 3', other_designations='T-complex testis-specific protein 2|T-complex testis-specific protein 3|T-complex-associated testis-expressed protein 3|T-complex-associated testis-expressed protein 4|t-complex testis-expressed 4', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15027154, end_position_on_the_genomic_accession=15041559, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  21648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21648, CurrentID=0, Status='live', Symbol='Dynlt1b', Aliases='AGS2, Dynlt1, Tctex-1, Tctex1', description='dynein light chain Tctex-type 1B', other_designations='T-complex testis-specific protein 1|activator of G-protein signaling 2|dynein light chain Tctex-type 1D|t-complex testis expressed 1', map_location='17 3.97 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6430112, end_position_on_the_genomic_accession=6436296, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  545938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545938, CurrentID=0, Status='live', Symbol='Zfp607', Aliases='4732475C15Rik', description='zinc finger proten 607', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27859827, end_position_on_the_genomic_accession=27909855, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21652, CurrentID=0, Status='live', Symbol='Phf1', Aliases='AW557215, D17Ertd455e, Pcl1, Phf2, Tctex3', description='PHD finger protein 1', other_designations='polycomb-like protein 1|t-complex testis-expressed 3', map_location='17 13.6 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26933127, end_position_on_the_genomic_accession=26937908, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  414872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=414872, CurrentID=0, Status='live', Symbol='Zyg11b', Aliases='1110046I03Rik, 2810482G21Rik, D4Mgi23, Gm431, mKIAA1730', description='zyg-ll family member B, cell cycle regulator', other_designations='zyg-ll homolog B', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108227755, end_position_on_the_genomic_accession=108301141, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  545947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545947, CurrentID=0, Status='live', Symbol='Scgb2b19', Aliases='Abpbg19, Gm5894', description='secretoglobin, family 2B, member 19', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33278366, end_position_on_the_genomic_accession=33280341, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  545948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545948, CurrentID=0, Status='live', Symbol='Scgb1b20', Aliases='Abpa20, EG545948, Gm5895', description='secretoglobin, family 1B, member 20', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33373234, end_position_on_the_genomic_accession=33374559, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21661, CurrentID=0, Status='live', Symbol='tda1', Aliases='Tda-1', description='testis-determining autosomal 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21664, CurrentID=0, Status='live', Symbol='Phlda1', Aliases='DT1P1B11, TDAG51, Tdag', description='pleckstrin homology-like domain, family A, member 1', other_designations='PQ-rich protein|PQR protein|T-cell death associated|T-cell death-associated gene 51 protein|proline- and glutamine-rich protein', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=111506286, end_position_on_the_genomic_accession=111508649, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  21665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21665, CurrentID=0, Status='live', Symbol='Tdg', Aliases='E130317C12Rik, JZA-3, Jza1', description='thymine DNA glycosylase', other_designations='C-JUN leucine zipper interactive protein JZA-3|thymine-DNA glycosylase', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82629247, end_position_on_the_genomic_accession=82650494, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21667, CurrentID=0, Status='live', Symbol='Tdgf1', Aliases='CR1, cripto', description='teratocarcinoma-derived growth factor 1', other_designations='TERATOCARCINOMA-DERIVED GROWTH FACTOR PRECURSOR (EPIDERMAL GROWTH FACTOR-LIKE CRIPTO PROTEIN) (CRIPTO GROWTH FACTOR)|cripto growth factor|epidermal growth factor-like Cripto protein', map_location='9 60.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110939608, end_position_on_the_genomic_accession=110946158, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  21672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21672, CurrentID=0, Status='live', Symbol='Prdx2', Aliases='AL022839, Band-8, NkefB, PRP, PrxII, TDX1, TPx, TPx-B, TR, TSA, Tdpx1, Torin', description='peroxiredoxin 2', other_designations='Calpromotin|Prx II-1|Trx dependent peroxide reductase 1|protector protein|thiol-specific antioxidant protein|thioredoxin peroxidase 1|thioredoxin reductase|thioredoxin-dependent peroxide reductase 1', map_location='8 41.4 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84969648, end_position_on_the_genomic_accession=84974317, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  21673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21673, CurrentID=0, Status='live', Symbol='Dntt', Aliases='Tdt', description='deoxynucleotidyltransferase, terminal', other_designations='terminal addition enzyme|terminal deoxynucleotidyl transferase', map_location='19 34.32 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41029275, end_position_on_the_genomic_accession=41059525, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21674, CurrentID=0, Status='live', Symbol='Sry', Aliases='Tdf, Tdy', description='sex determining region of Chr Y', other_designations='testis determining-Y|testis-determining factor', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=2662471, end_position_on_the_genomic_accession=2663658, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  21676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21676, CurrentID=0, Status='live', Symbol='Tead1', Aliases='2610024B07Rik, B230114H05Rik, Gtrgeo5, TEAD-1, TEF-1, Tcf13, mTEF-1', description='TEA domain family member 1', other_designations='NTEF-1|TCF-13|transcription factor 13', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=112679320, end_position_on_the_genomic_accession=112906805, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  21677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21677, CurrentID=0, Status='live', Symbol='Tead2', Aliases='ETF, Etdf, TEAD-2, TEF-4, TEF4', description='TEA domain family member 2', other_designations='ETEF-1|embryonic TEA domain-containing factor', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45215753, end_position_on_the_genomic_accession=45233636, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21678, CurrentID=0, Status='live', Symbol='Tead3', Aliases='DTEF-1, ETFR-1, TEAD-3, TEF-5, Tcf13r2', description='TEA domain family member 3', other_designations='ETF-related factor 1', map_location='17 14.66 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28331671, end_position_on_the_genomic_accession=28350805, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  21679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21679, CurrentID=0, Status='live', Symbol='Tead4', Aliases='ETFR-2, Etfr2, FR-19, Rtef1, TEAD-4, TEF-3, Tcf13r1, Tef3, Tefr, Tefr1, Tefr1a', description='TEA domain family member 4', other_designations='ETF-related factor 2|ETF-related factor-2|FGF-regulated 19|RTEF-1|TEF-1-related factor 1|TEF-1-related factor FR-19', map_location='6 62.92 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128227143, end_position_on_the_genomic_accession=128300813, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  21681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21681, CurrentID=0, Status='live', Symbol='Alyref', Aliases='ALY, REF1, Refbp1, Thoc4', description='Aly/REF export factor', other_designations='REF1-I|RNA and export factor binding protein 1|RNA and export factor-binding protein 1|THO complex 4|Tcra enhancer-binding factor interacting protein 1|ally of AML-1 and LEF-1|tho4|transcriptional coactivator Aly/REF', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120594516, end_position_on_the_genomic_accession=120598436, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  21682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21682, CurrentID=0, Status='live', Symbol='Tec', Aliases='', description='tec protein tyrosine kinase', other_designations='cytoplasmic tyrosine kinase, Dscr28C related|protein tyrosine kinase TecI', map_location='5 38.44 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72755718, end_position_on_the_genomic_accession=72868740, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  21683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21683, CurrentID=0, Status='live', Symbol='Tecta', Aliases='Tctna', description='tectorin alpha', other_designations='[a]-tectorin', map_location='9 23.63 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=42329622, end_position_on_the_genomic_accession=42403179, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  21684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21684, CurrentID=0, Status='live', Symbol='Tectb', Aliases='Tctnb', description='tectorin beta', other_designations='[b]-tectorin', map_location='19 51.13 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=55180733, end_position_on_the_genomic_accession=55196313, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21685, CurrentID=0, Status='live', Symbol='Tef', Aliases='2310028D20Rik', description='thyrotroph embryonic factor', other_designations='thyrotrophic embryonic factor', map_location='15 38.25 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81802673, end_position_on_the_genomic_accession=81826863, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21687, CurrentID=0, Status='live', Symbol='Tek', Aliases='AA517024, Cd202b, Hyk, STK1, Tie-2, Tie2', description='endothelial-specific receptor tyrosine kinase', other_designations='endothelial tyrosine kinase|p140 TEK|tunica interna endothelial cell kinase|tyrosine kinase with Ig and EGF homology domains-2|tyrosine-protein kinase receptor TEK|tyrosine-protein kinase receptor TIE-2', map_location='4 43.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=94739289, end_position_on_the_genomic_accession=94874976, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  21689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21689, CurrentID=0, Status='live', Symbol='Tekt1', Aliases='MT14', description='tektin 1', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72344717, end_position_on_the_genomic_accession=72362442, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  546024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546024, CurrentID=0, Status='live', Symbol='Crxos', Aliases='AA606869, AY5908911, Egam1, Crxos', description='cone-rod homeobox, opposite strand', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15882617, end_position_on_the_genomic_accession=15904030, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21743, CurrentID=0, Status='live', Symbol='Inmt', Aliases='Temt', description='indolethylamine N-methyltransferase', other_designations='amine N-methyltransferase|aromatic alkylamine N-methyltransferase|arylamine N-methyltransferase|indolamine N-methyltransferase|nicotine N-methyltransferase|thioether S-methyltransferase', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55170627, end_position_on_the_genomic_accession=55174990, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  21744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21744, CurrentID=0, Status='live', Symbol='Adad1', Aliases='Tenr', description='adenosine deaminase domain containing 1 (testis specific)', other_designations='testis nuclear RNA binding protein|testis nuclear RNA-binding protein', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37063093, end_position_on_the_genomic_accession=37121930, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  21745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21745, CurrentID=0, Status='live', Symbol='Tep1', Aliases='Tp1', description='telomerase associated protein 1', other_designations='p240|p80 homolog|p80 telomerase homolog|telomerase protein 1|telomerase-associated protein 1', map_location='14 C2-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50824061, end_position_on_the_genomic_accession=50870554, orientation='minus', exon_count=57, OMIM='', no_hdr0=''),
  21748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21748, CurrentID=0, Status='live', Symbol='Terc', Aliases='mTER, mTR', description='telomerase RNA component', other_designations='', map_location='3 F/G', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96414437, end_position_on_the_genomic_accession=96414833, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  21749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21749, CurrentID=0, Status='live', Symbol='Terf1', Aliases='Pin2, Trbf1, Trf1', description='telomeric repeat binding factor 1', other_designations='TTAGGG repeat-binding factor 1', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=15804982, end_position_on_the_genomic_accession=15844052, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21750, CurrentID=0, Status='live', Symbol='Terf2', Aliases='TRF2', description='telomeric repeat binding factor 2', other_designations='TTAGGG repeat-binding factor 2|telomeric DNA-binding protein', map_location='8 53.59 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107069400, end_position_on_the_genomic_accession=107096642, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  21752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21752, CurrentID=0, Status='live', Symbol='Tert', Aliases='EST2, TCS1, TP2, TR, TRT', description='telomerase reverse transcriptase', other_designations='telomerase catalytic subunit|telomerase reverse transcriptase beta', map_location='13 40.12 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73626797, end_position_on_the_genomic_accession=73649843, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  21753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21753, CurrentID=0, Status='live', Symbol='Tes', Aliases='D6Ertd352e, TESS1, Tes2, testin2, Tes', description='testis derived transcript', other_designations='TES1/TES2|testin 2', map_location='6 7.67 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=17065149, end_position_on_the_genomic_accession=17105825, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  21754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21754, CurrentID=0, Status='live', Symbol='Tesk1', Aliases='AI326901', description='testis specific protein kinase 1', other_designations='testicular protein kinase 1', map_location='4 A5-C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43442277, end_position_on_the_genomic_accession=43448075, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21755, CurrentID=0, Status='live', Symbol='Prss39', Aliases='Tesp1', description='protease, serine 39', other_designations='inactive testicular serine protease 1|protease, serine, 39|testicular serine protease 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34498430, end_position_on_the_genomic_accession=34503062, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21756, CurrentID=0, Status='live', Symbol='Prss40', Aliases='Tesp2', description='protease, serine 40', other_designations='protease, serine, 40|testicular serine protease 2', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34536515, end_position_on_the_genomic_accession=34560993, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  21761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21761, CurrentID=0, Status='live', Symbol='Morf4l1', Aliases='MORFRG15, MRG15, TEG-189, Tex189, mKIAA4002', description='mortality factor 4 like 1', other_designations='MORF-related gene 15 protein|testis expressed gene 189|testis-expressed gene 189 protein|transcription factor-like protein MRG15', map_location='9 D1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=90091669, end_position_on_the_genomic_accession=90114820, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  21762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21762, CurrentID=0, Status='live', Symbol='Psmd2', Aliases='9430095H01Rik, AA407121, TEG-190, Tex190', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 2', other_designations='26S proteasome regulatory subunit RPN1|26S proteasome regulatory subunit S2|26S proteasome subunit p97|testis expressed gene 190', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20651652, end_position_on_the_genomic_accession=20663414, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  21763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21763, CurrentID=0, Status='live', Symbol='Tex2', Aliases='4930568E07Rik, AI553404, Def-5, Taz4', description='testis expressed gene 2', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106502139, end_position_on_the_genomic_accession=106613418, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  21766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21766, CurrentID=0, Status='live', Symbol='Tex261', Aliases='3110001O07Rik, AA409339, AI480706, AL033351, TEG-261', description='testis expressed gene 261', other_designations='testis-expressed protein 261|testis-expressed sequence 261 protein', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83768881, end_position_on_the_genomic_accession=83775814, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  21767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21767, CurrentID=0, Status='live', Symbol='Tex264', Aliases='TEG-264', description='testis expressed gene 264', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106658746, end_position_on_the_genomic_accession=106685952, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  21769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21769, CurrentID=0, Status='live', Symbol='Zfand3', Aliases='AW539211, TEG-27, Tex27', description='zinc finger, AN1-type domain 3', other_designations='testis-expressed sequence 27', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=30005039, end_position_on_the_genomic_accession=30210020, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  21770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21770, CurrentID=0, Status='live', Symbol='Ppp2r5d', Aliases="B'delta, TEG-271, Tex271", description="protein phosphatase 2, regulatory subunit B', delta", other_designations='protein phosphatase 2, regulatory subunit B (B56), delta isoform|testis expressed gene 271', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46682991, end_position_on_the_genomic_accession=46705020, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  21771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21771, CurrentID=0, Status='live', Symbol='Cirh1a', Aliases='Cirhin, Naic, Teg-292, Tex292', description='cirrhosis, autosomal recessive 1A (human)', other_designations='testis-expressed gene 292 protein', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106893636, end_position_on_the_genomic_accession=106923094, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  21778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21778, CurrentID=0, Status='live', Symbol='Tex9', Aliases='Gm19407, tsec-1', description='testis expressed gene 9', other_designations='testis-specifically expressed protein 1', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72458055, end_position_on_the_genomic_accession=72510944, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  21779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21779, CurrentID=0, Status='live', Symbol='tf', Aliases='', description='tufted', other_designations='', map_location='17 13.3 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21780, CurrentID=0, Status='live', Symbol='Tfam', Aliases='AI661103, Hmgts, mtTFA, tsHMG', description='transcription factor A, mitochondrial', other_designations='TS-HMG|testis-specific HMG-box protein m-tsHMG|testis-specific high mobility group protein', map_location='10 36.83 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=71225468, end_position_on_the_genomic_accession=71238957, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  21781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21781, CurrentID=0, Status='live', Symbol='Tfdp1', Aliases='Dp1, Drtf1', description='transcription factor Dp 1', other_designations='DRTF1-polypeptide 1|E2F dimerization partner 1|cell cycle regulatory transcription factor DP1', map_location='8 A1-A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13338751, end_position_on_the_genomic_accession=13378448, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  546071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546071, CurrentID=0, Status='live', Symbol='Mast3', Aliases='BC024265, mKIAA0561', description='microtubule associated serine/threonine kinase 3', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70778117, end_position_on_the_genomic_accession=70805068, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  21784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21784, CurrentID=0, Status='live', Symbol='Tff1', Aliases='Bcei, PS2', description='trefoil factor 1', other_designations='protein pS2', map_location='17 15.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31161395, end_position_on_the_genomic_accession=31165065, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  21785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21785, CurrentID=0, Status='live', Symbol='Tff2', Aliases='SP, mSP', description='trefoil factor 2 (spasmolytic protein 1)', other_designations='spasmolytic polypeptide', map_location='17 15.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31141062, end_position_on_the_genomic_accession=31144282, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  21786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21786, CurrentID=0, Status='live', Symbol='Tff3', Aliases='ITF, mITF', description='trefoil factor 3, intestinal', other_designations='intestinal trefoil factor', map_location='17 15.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31125306, end_position_on_the_genomic_accession=31129611, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  21787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21787, CurrentID=0, Status='live', Symbol='Tfg', Aliases='AI173908', description='Trk-fused gene', other_designations='', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56690329, end_position_on_the_genomic_accession=56717530, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  21788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21788, CurrentID=0, Status='live', Symbol='Tfpi', Aliases='A630013F22Rik, AW552122, EPI, LACI', description='tissue factor pathway inhibitor', other_designations='extrinsic pathway inhibitor|lipoprotein-associated coagulation inhibitor', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84432855, end_position_on_the_genomic_accession=84476829, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  21789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21789, CurrentID=0, Status='live', Symbol='Tfpi2', Aliases='AV000670, PP5/TFPI-2', description='tissue factor pathway inhibitor 2', other_designations='TFPI-2', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3962589, end_position_on_the_genomic_accession=3988975, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  677156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=677156, CurrentID=0, Status='live', Symbol='Cyp4f37', Aliases='Gm9705', description='cytochrome P450, family 4, subfamily f, polypeptide 37', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32621319, end_position_on_the_genomic_accession=32636184, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  21802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21802, CurrentID=0, Status='live', Symbol='Tgfa', Aliases='wa-1, wa1', description='transforming growth factor alpha', other_designations='transforming growth factor-alpha', map_location='6 37.62 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86195251, end_position_on_the_genomic_accession=86275449, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  21803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21803, CurrentID=0, Status='live', Symbol='Tgfb1', Aliases='TGF-beta1, TGFbeta1, Tgfb, Tgfb-1', description='transforming growth factor, beta 1', other_designations='', map_location='7 13.98 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25687002, end_position_on_the_genomic_accession=25704996, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  21804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21804, CurrentID=0, Status='live', Symbol='Tgfb1i1', Aliases='ARA55, Hic5, TSC-5, hic-5', description='transforming growth factor beta 1 induced transcript 1', other_designations='TGF beta-stimulated clone 5|androgen receptor-associated protein of 55 kDa|hydrogen peroxide-inducible clone 5 protein', map_location='7 F1-F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128246799, end_position_on_the_genomic_accession=128255699, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21807, CurrentID=0, Status='live', Symbol='Tsc22d1', Aliases='AA589566, AW105905, Egr5, TSC-22, Tgfb1i4, Tsc, Tsc22', description='TSC22 domain family, member 1', other_designations='Tgfb-stimulated clone 22 homolog|Tsc22-related inducible leucine zipper 1b|regulatory protein TSC-22|regulatory protein Tsc22|transforming growth factor beta 1 induced transcript 4', map_location='14 40.44 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=76414961, end_position_on_the_genomic_accession=76507766, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21808, CurrentID=0, Status='live', Symbol='Tgfb2', Aliases='BB105277, Tgf-beta2, Tgfb-2', description='transforming growth factor, beta 2', other_designations='TGF-beta-2', map_location='1 89.95 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=186622787, end_position_on_the_genomic_accession=186709033, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  21809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21809, CurrentID=0, Status='live', Symbol='Tgfb3', Aliases='TGF-beta-3, Tgfb-3', description='transforming growth factor, beta 3', other_designations='', map_location='12 40.09 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86056743, end_position_on_the_genomic_accession=86079041, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  21810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21810, CurrentID=0, Status='live', Symbol='Tgfbi', Aliases='68kDa, AI181842, AI747162, Beta-ig, big-h3', description='transforming growth factor, beta induced', other_designations='beta ig-h3|transforming growth factor, beta induced, 68 kDa', map_location='13 30.09 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56609603, end_position_on_the_genomic_accession=56639339, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  21812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21812, CurrentID=0, Status='live', Symbol='Tgfbr1', Aliases='ALK5, AU017191, Alk-5, ESK2, TGFR-1, TbetaR-I, TbetaRI', description='transforming growth factor, beta receptor I', other_designations='TGF-beta receptor type I|transforming growth factor-beta receptor type I', map_location='4 26.02 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=47353113, end_position_on_the_genomic_accession=47414926, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21813, CurrentID=0, Status='live', Symbol='Tgfbr2', Aliases='1110020H15Rik, AU042018, DNIIR, RIIDN, TBR-II, TbetaR-II, TbetaRII', description='transforming growth factor, beta receptor II', other_designations='TGF-beta receptor II|TGF-beta receptor type II|TGF-beta type II receptor|TGFR-2|transforming growth factor-beta receptor type II', map_location='9 68.39 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=116087695, end_position_on_the_genomic_accession=116175363, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  21814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21814, CurrentID=0, Status='live', Symbol='Tgfbr3', Aliases='1110036H20Rik, AU015626, AW215636, TBRIII', description='transforming growth factor, beta receptor III', other_designations='TGF-beta receptor type 3|TGF-beta receptor type III|TGFR-3|betaglycan', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107106570, end_position_on_the_genomic_accession=107289595, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  21815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21815, CurrentID=0, Status='live', Symbol='Tgif1', Aliases='AA959811, AI462167, Tgif', description='TGFB-induced factor homeobox 1', other_designations="5'-TG-3'-interacting factor 1|TALE family homeobox|TG interacting factor 1", map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=70844205, end_position_on_the_genomic_accession=70853532, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  21816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21816, CurrentID=0, Status='live', Symbol='Tgm1', Aliases='2310004J08Rik, Tgase1', description='transglutaminase 1, K polypeptide', other_designations='K polypeptide|TG K|TG(K)|TGK|TGase 1|TGase K|TGase-1|epidermal TGase|protein-glutamine-gamma-glutamyltransferase|transglutaminase K|transglutaminase type 1|transglutaminase-1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55700009, end_position_on_the_genomic_accession=55713921, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  21817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21817, CurrentID=0, Status='live', Symbol='Tgm2', Aliases='G[a]h, TG2, TGase2, tTG, tTGas', description='transglutaminase 2, C polypeptide', other_designations='C polypeptide|TG C|TG(C)|TGC|TGase C|TGase-2|protein-glutamine-gamma-glutamyltransferase|tissue transglutaminase|transglutaminase C|transglutaminase-2', map_location='2 78.72 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158116405, end_position_on_the_genomic_accession=158146392, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  21818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21818, CurrentID=0, Status='live', Symbol='Tgm3', Aliases='AI893889, TG E, TG(E), TGE, TGase E, TGase-3', description='transglutaminase 3, E polypeptide', other_designations='E polypeptide|protein-glutamine-gamma-glutamyltransferase|transglutaminase E|transglutaminase-3', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129987794, end_position_on_the_genomic_accession=130050399, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  21819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21819, CurrentID=0, Status='live', Symbol='Tg', Aliases='Tgn, cog', description='thyroglobulin', other_designations='congenital goiter', map_location='15 D3-E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=66670756, end_position_on_the_genomic_accession=66850721, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  21821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21821, CurrentID=0, Status='live', Symbol='Ift88', Aliases='AW552028, Tg737, Tg737Rpw, TgN737Rpw, Ttc10, flexo, fxo, orpk, polaris', description='intraflagellar transport 88', other_designations='TPR repeat protein 10|intraflagellar transport 88 homolog|recessive polycystic kidney disease protein Tg737|tetratricopeptide repeat domain 10|tetratricopeptide repeat protein 10|tgN(Imorpk)737Rpw|transgene insert site 737, insertional mutation, polycystic kidney disease', map_location='14 30.1 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57423657, end_position_on_the_genomic_accession=57517936, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  21822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21822, CurrentID=0, Status='live', Symbol='Tgtp1', Aliases='Gtp2, Irgb6, Mg21, Tgtp', description='T cell specific GTPase 1', other_designations='T-cell specific GTPase 1', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48985329, end_position_on_the_genomic_accession=48992246, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  21823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21823, CurrentID=0, Status='live', Symbol='Th', Aliases='', description='tyrosine hydroxylase', other_designations='tyrosine 3-hydroxylase', map_location='7 88.06 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142892776, end_position_on_the_genomic_accession=142899966, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  21824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21824, CurrentID=0, Status='live', Symbol='Thbd', Aliases='AI385582, CD141, TM', description='thrombomodulin', other_designations='fetomodulin|snoRNA MBII-339', map_location='2 73.45 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148404471, end_position_on_the_genomic_accession=148408188, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  21825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21825, CurrentID=0, Status='live', Symbol='Thbs1', Aliases='TSP-1, TSP1, Thbs-1, tbsp1', description='thrombospondin 1', other_designations='', map_location='2 F1-F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118111890, end_position_on_the_genomic_accession=118127133, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  21826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21826, CurrentID=0, Status='live', Symbol='Thbs2', Aliases='TSP2, Thbs-2', description='thrombospondin 2', other_designations='', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14665500, end_position_on_the_genomic_accession=14694262, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  21827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21827, CurrentID=0, Status='live', Symbol='Thbs3', Aliases='TSP3, Thbs-3', description='thrombospondin 3', other_designations='', map_location='3 E3-F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89215142, end_position_on_the_genomic_accession=89226837, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  21828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21828, CurrentID=0, Status='live', Symbol='Thbs4', Aliases='TSP4', description='thrombospondin 4', other_designations='', map_location='13 47.75 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92751586, end_position_on_the_genomic_accession=92794818, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  21829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21829, CurrentID=0, Status='live', Symbol='thd', Aliases='th, ub, unbalanced', description='tilted head', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21830, CurrentID=0, Status='live', Symbol='Theg', Aliases='', description='testicular haploid expressed gene', other_designations='testicular haploid expressed gene product', map_location='10 B5-C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79576477, end_position_on_the_genomic_accession=79587136, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  21831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21831, CurrentID=0, Status='live', Symbol='thf', Aliases='', description='thin fur', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21832, CurrentID=0, Status='live', Symbol='Thpo', Aliases='Mgdf, Ml, Mpllg, Tpo', description='thrombopoietin', other_designations='C-mpl ligand|megakaryocyte colony-stimulating factor|megakaryocyte growth and development factor|myeloproliferative leukemia virus oncogene ligand', map_location='16 12.51 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20724454, end_position_on_the_genomic_accession=20734511, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  21833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21833, CurrentID=0, Status='live', Symbol='Thra', Aliases='6430529J03Rik, AW259572, Erba, Nr1a1, Rvr, T3R[a], T3Ralpha1, Thra2, c-erbA-1, c-erbA-alpha, c-erbAalpha, Thra', description='thyroid hormone receptor alpha', other_designations='TR alpha 1|TR alpha 2|nuclear receptor subfamily 1 group A member 1', map_location='11 62.58 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98741873, end_position_on_the_genomic_accession=98765111, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21834, CurrentID=0, Status='live', Symbol='Thrb', Aliases='Nr1a2, T3R[b], T3Rbeta1, Thrb2, c-erbAbeta, Thrb', description='thyroid hormone receptor beta', other_designations='TR beta|c-erbA-2|c-erbA-beta|nuclear receptor subfamily 1 group A member 2', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=17659218, end_position_on_the_genomic_accession=18040097, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  21835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21835, CurrentID=0, Status='live', Symbol='Thrsp', Aliases='S14, SPOT14', description='thyroid hormone responsive', other_designations='spot 14 protein|thyroid hormone responsive SPOT14 homolog', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97412957, end_position_on_the_genomic_accession=97417510, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  21837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21837, CurrentID=0, Status='live', Symbol='Tht', Aliases='TgN370Rpw', description='thick tail', other_designations='transgene insert site 370, insertional mutation TgN (Imusd) 370 Woychik', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21838, CurrentID=0, Status='live', Symbol='Thy1', Aliases='CD90, T25, Thy-1, Thy-1.2.1, Thy1.2, Thy1', description='thymus cell antigen 1, theta', other_designations='thy-1 antigen', map_location='9 24.52 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44043384, end_position_on_the_genomic_accession=44048579, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  21841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21841, CurrentID=0, Status='live', Symbol='Tia1', Aliases='2310050N03Rik, AI256674, TIA-1, mTIA-1', description='cytotoxic granule-associated RNA binding protein 1', other_designations='RNA-binding protein TIA-1|T-cell-restricted intracellular antigen-1|cytotoxic granule-associated RNA-binding protein 1', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86404219, end_position_on_the_genomic_accession=86433405, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  21843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21843, CurrentID=0, Status='live', Symbol='Tial1', Aliases='5330433G13Rik, AL033329, TIAR, mTIAR', description='Tia1 cytotoxic granule-associated RNA binding protein-like 1', other_designations='TIA-1-related protein|Tial1 cytotoxic granule-associated RNA binding protein-like 1|Tial1 cytotoxic granule-associated RNA-binding protein-like 1', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128439772, end_position_on_the_genomic_accession=128461668, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  21844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21844, CurrentID=0, Status='live', Symbol='Tiam1', Aliases='AI847750, D16Ium10, D16Ium10e', description='T cell lymphoma invasion and metastasis 1', other_designations='T-cell lymphoma invasion and metastasis 1|TIAM-1', map_location='16 51.5 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=89787111, end_position_on_the_genomic_accession=90145483, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  546134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546134, CurrentID=0, Status='live', Symbol='Gramd2', Aliases='BC064463', description='GRAM domain containing 2', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59679462, end_position_on_the_genomic_accession=59718871, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  21847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21847, CurrentID=0, Status='live', Symbol='Klf10', Aliases='AI115143, EGR[a], Egral, Gdnfif, TIEG-1, Tieg, Tieg1, mGIF', description='Kruppel-like factor 10', other_designations='GDNF-inducible factor|TGFB inducible early growth response 1|TGFB-inducible early growth response protein 1|transcription factor GIF|transforming growth factor-beta-inducible early growth response protein 1', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=38291464, end_position_on_the_genomic_accession=38300711, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  21848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21848, CurrentID=0, Status='live', Symbol='Trim24', Aliases='A130082H20Rik, AI447469, D430004I05Rik, TIF1, TIF1-alpha, TIF1alpha, Tif1a', description='tripartite motif-containing 24', other_designations='E3 ubiquitin-protein ligase Trim24|transcriptional intermediary factor 1, alpha|tripartite motif-containing protein 24', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=37869880, end_position_on_the_genomic_accession=37968445, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  21849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21849, CurrentID=0, Status='live', Symbol='Trim28', Aliases='AA408787, KAP-1, KRIP-1, MommeD9, Tif1b, Tif1beta', description='tripartite motif-containing 28', other_designations='E3 SUMO-protein ligase TRIM28|KRAB-A-interacting protein|TIF1-beta|transcriptional intermediary factor 1, beta|tripartite motif protein 28|tripartite motif-containing protein 28', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13024152, end_position_on_the_genomic_accession=13031032, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  21853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21853, CurrentID=0, Status='live', Symbol='Timeless', Aliases='C77407, Debt69, tim', description='timeless circadian clock 1', other_designations='mTim|timeless homolog', map_location='10 76.49 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128232063, end_position_on_the_genomic_accession=128252941, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  21854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21854, CurrentID=0, Status='live', Symbol='Timm17a', Aliases='17kDa, mTim17a', description='translocase of inner mitochondrial membrane 17a', other_designations='inner membrane preprotein translocase Tim17a|translocator of inner mitochondrial membrane 17 kDa, a|translocator of inner mitochondrial membrane 17a|translocator of inner mitochondrial membrane a', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135301535, end_position_on_the_genomic_accession=135313737, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  546143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546143, CurrentID=0, Status='live', Symbol='Ccpg1os', Aliases='EG546143, Gm5918', description='cell cycle progression 1, opposite strand', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72979717, end_position_on_the_genomic_accession=72985376, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21856, CurrentID=0, Status='live', Symbol='Timm44', Aliases='0710005E20Rik, D8Ertd118e, Mimt44, Tim44', description='translocase of inner mitochondrial membrane 44', other_designations='translocator of inner mitochondrial membrane 44|translocator of inner mitochondrial membrane 44; DNA segment, Chr 8, ERATO Doi 118, expressed', map_location='8 1.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4259731, end_position_on_the_genomic_accession=4275905, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  21857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21857, CurrentID=0, Status='live', Symbol='Timp1', Aliases='Clgi, EPA, TIMP-1, TPA-S1, Timp', description='tissue inhibitor of metalloproteinase 1', other_designations='TPA-induced protein|collagenase inhibitor 16C8 fibroblast|erythroid-potentiating activity|tissue inhibitor of metalloproteinase-1|tissue inhibitor of metalloproteinases 1', map_location='X 16.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20870166, end_position_on_the_genomic_accession=20874737, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  21858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21858, CurrentID=0, Status='live', Symbol='Timp2', Aliases='D11Bwg1104e, Timp-2', description='tissue inhibitor of metalloproteinase 2', other_designations='tissue inhibitor of metalloproteinases 2', map_location='11 83.09 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118301061, end_position_on_the_genomic_accession=118355411, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21859, CurrentID=0, Status='live', Symbol='Timp3', Aliases='Timp-3', description='tissue inhibitor of metalloproteinase 3', other_designations='tissue inhibitor of metalloproteinases 3', map_location='10 C1-D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=86300412, end_position_on_the_genomic_accession=86349505, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21863, CurrentID=0, Status='live', Symbol='tint', Aliases='', description='tail interaction with T complex', other_designations='interaction with T complex', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  546157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546157, CurrentID=0, Status='live', Symbol='7420426K07Rik', Aliases='', description='RIKEN cDNA 7420426K07 gene', other_designations='', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98903119, end_position_on_the_genomic_accession=98904622, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  21871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21871, CurrentID=0, Status='live', Symbol='Atp6v0a2', Aliases='8430408C20Rik, AI385560, ATP6a2, AW489264, Atp6n1d, Atp6n2, C76904, ISF, J6B7, SHIF, Stv1, TJ6M, TJ6s, Tj6, V-ATPase 116 kDa, V-ATPase a2', description='ATPase, H+ transporting, lysosomal V0 subunit A2', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump) non-catalytic accessory protein 2 (38kD)|ATPase, H+ transporting, lysosomal V0 subunit a|T-cell expressing clone j6|V-type proton ATPase 116 kDa subunit a|immune suppressor factor J6B7|vacuolar proton translocating ATPase 116 kDa subunit a', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124712202, end_position_on_the_genomic_accession=124724455, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  21872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21872, CurrentID=0, Status='live', Symbol='Tjp1', Aliases='ZO1', description='tight junction protein 1', other_designations='zona occludens protein 1|zonula occludens protein 1', map_location='7 35.02 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=65296165, end_position_on_the_genomic_accession=65527806, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  546161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546161, CurrentID=0, Status='live', Symbol='Fbxw18', Aliases='C85627', description='F-box and WD-40 domain protein 18', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109676732, end_position_on_the_genomic_accession=109702700, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  21874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21874, CurrentID=0, Status='live', Symbol='tk', Aliases='', description='tail-kinks', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21877, CurrentID=0, Status='live', Symbol='Tk1', Aliases='D530002A18Rik, Tk-1a, Tk1b, Tk1', description='thymidine kinase 1', other_designations='', map_location='11 E1-E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117815509, end_position_on_the_genomic_accession=117826014, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  21881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21881, CurrentID=0, Status='live', Symbol='Tkt', Aliases='p68', description='transketolase', other_designations='TK', map_location='14 B1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30549131, end_position_on_the_genomic_accession=30574724, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  21884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21884, CurrentID=0, Status='live', Symbol='Fabp9', Aliases='1700007P10Rik, Perf, Perf15, Tlbp', description='fatty acid binding protein 9, testis', other_designations='15 kDa perforatorial protein|PERF 15|T-FABP|testis lipid binding protein|testis lipid-binding protein|testis-type fatty acid-binding protein', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10193623, end_position_on_the_genomic_accession=10197283, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  21885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21885, CurrentID=0, Status='live', Symbol='Tle1', Aliases='C230057C06Rik, Estm14, Grg-1, Grg1, Tle4l', description='transducin-like enhancer of split 1, homolog of Drosophila E(spl)', other_designations='groucho-related protein 1|transducin-like enhancer of split 4 (homolog of Drosophila E(spl)) - like', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=72109945, end_position_on_the_genomic_accession=72200893, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  21886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21886, CurrentID=0, Status='live', Symbol='Tle2', Aliases='Grg2, mKIAA4188', description='transducin-like enhancer of split 2, homolog of Drosophila E(spl)', other_designations='groucho-related protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81574739, end_position_on_the_genomic_accession=81592713, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  21887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21887, CurrentID=0, Status='live', Symbol='Tle3', Aliases='2610103N05Rik, ESG, Grg3a, Grg3b, mKIAA1547', description='transducin-like enhancer of split 3, homolog of Drosophila E(spl)', other_designations='grg-3', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=61372326, end_position_on_the_genomic_accession=61418497, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  21888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21888, CurrentID=0, Status='live', Symbol='Tle4', Aliases='5730411M05Rik, AA792082, Bce-1, Bce1, ESTM13, ESTM14, Grg4, X83332, X83333, grg-4', description='transducin-like enhancer of split 4, homolog of Drosophila E(spl)', other_designations='B lymphocyte gene 1|groucho-related protein 4', map_location='19 9.11 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=14448072, end_position_on_the_genomic_accession=14598183, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  21892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21892, CurrentID=0, Status='live', Symbol='Tll1', Aliases='Tll, Tll-1', description='tolloid-like', other_designations='mTll', map_location='8 32.19 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=64014150, end_position_on_the_genomic_accession=64205993, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  21893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21893, CurrentID=0, Status='live', Symbol='Tlm', Aliases='', description='T lymphoma oncogene', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.4', start_position_on_the_genomic_accession=107551128, end_position_on_the_genomic_accession=107553683, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21894, CurrentID=0, Status='live', Symbol='Tln1', Aliases='Tln', description='talin 1', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43531513, end_position_on_the_genomic_accession=43562583, orientation='minus', exon_count=60, OMIM='', no_hdr0=''),
  21897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21897, CurrentID=0, Status='live', Symbol='Tlr1', Aliases='', description='toll-like receptor 1', other_designations='TIL|toll/interleukin-1 receptor-like protein', map_location='5 33.53 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=64924680, end_position_on_the_genomic_accession=64933589, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21898, CurrentID=0, Status='live', Symbol='Tlr4', Aliases='Lps, Ly87, Ran/M1, Rasl2-8', description='toll-like receptor 4', other_designations='Toll like receptor 4|lipopolysaccharide response', map_location='4 34.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=66827811, end_position_on_the_genomic_accession=66842797, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  415115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=415115, CurrentID=0, Status='live', Symbol='Neurl2', Aliases='Neur2, Ozz, Ozz-E3', description='neuralized E3 ubiquitin protein ligase 2', other_designations='neuralized-like 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164830730, end_position_on_the_genomic_accession=164833596, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  21906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21906, CurrentID=0, Status='live', Symbol='Otop1', Aliases='A530025J20Rik, Otp1, tlt', description='otopetrin 1', other_designations='tilted', map_location='5 20.35 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38277404, end_position_on_the_genomic_accession=38304217, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  21907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21907, CurrentID=0, Status='live', Symbol='Nr2e1', Aliases='Mtl1, Mtll, TLL, Tlx, XTLL, fierce, frc, tailless', description='nuclear receptor subfamily 2, group E, member 1', other_designations='nuclear receptor TLX|protein tailless homolog|tailless homolog', map_location='10 22.89 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=42561963, end_position_on_the_genomic_accession=42583588, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  21908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21908, CurrentID=0, Status='live', Symbol='Tlx1', Aliases='Hox-11, Hox11', description='T cell leukemia, homeobox 1', other_designations='T-cell leukemia, homeobox 1|homeobox TLX-1|homeobox protein Hox-11', map_location='19 38.45 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45150079, end_position_on_the_genomic_accession=45156943, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21909, CurrentID=0, Status='live', Symbol='Tlx2', Aliases='Enx, Hox11L.1, Hox11l1, NCX, Ncx1, Tlx1l1, Tlx1l2', description='T cell leukemia, homeobox 2', other_designations='PMUR10F|T-cell leukemia, homeobox 2|enteric neuron homeobox protein|homeo box 11-like 1|homeobox TLX-2|homeobox protein Hox-11L1', map_location='6 C3-D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83068325, end_position_on_the_genomic_accession=83070225, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  21911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21911, CurrentID=0, Status='live', Symbol='tm', Aliases='', description='tremulous', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21912, CurrentID=0, Status='live', Symbol='Tspan7', Aliases='1200014P11Rik, A15, AI323365, Cd231, Mxs1, PE31, R74651, TALLA, Tm4sf2', description='tetraspanin 7', other_designations='TALLA homolog|cell surface glycoprotein A15|transmembrane 4 superfamily member 2|tspan-7', map_location='X A1.3-A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=10485116, end_position_on_the_genomic_accession=10596604, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21915, CurrentID=0, Status='live', Symbol='Dtymk', Aliases='AU044245, CDC8, D10Ertd203e, TMPK, TYPK, Tmk, mtmk', description='deoxythymidylate kinase', other_designations='dTMP kinase', map_location='1 47.34 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93792576, end_position_on_the_genomic_accession=93801934, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21916, CurrentID=0, Status='live', Symbol='Tmod1', Aliases='E-Tmod, Tmod', description='tropomodulin 1', other_designations='erythrocyte tropomodulin', map_location='4 24.49 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46038940, end_position_on_the_genomic_accession=46116032, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  21917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21917, CurrentID=0, Status='live', Symbol='Tmpo', Aliases='5630400D24Rik, AI195756, AI606875, AW214352, AW547477, LAP2, TP', description='thymopoietin', other_designations='', map_location='10 45.66 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=91147571, end_position_on_the_genomic_accession=91181719, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  21922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21922, CurrentID=0, Status='live', Symbol='Clec3b', Aliases='Tna', description='C-type lectin domain family 3, member b', other_designations='TN|plasminogen kringle 4-binding protein|tetranectin (plasminogen binding protein)|tetranectin (plasminogen-binding protein)', map_location='9 F1-F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123150946, end_position_on_the_genomic_accession=123157432, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21923, CurrentID=0, Status='live', Symbol='Tnc', Aliases='AI528729, C130033P17Rik, Hxb, TN, TN-C, Ten, cytotactin, tenascin-C', description='tenascin C', other_designations='hexabrachion|tnc', map_location='4 34.06 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63959785, end_position_on_the_genomic_accession=64047015, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  21924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21924, CurrentID=0, Status='live', Symbol='Tnnc1', Aliases='AI874626, TnC, cTnC, cTnI, tncc', description='troponin C, cardiac/slow skeletal', other_designations='TN-C', map_location='14 19.09 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31208312, end_position_on_the_genomic_accession=31211711, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  546213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546213, CurrentID=0, Status='live', Symbol='Igkv4-53', Aliases='Gm1420, Gm1526, IgVkkh4, kh4', description='immunoglobulin kappa variable 4-53', other_designations='immunoglobulin kappa chain variable 4-53|immunoglobulin kappa-chain VK-1', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69648824, end_position_on_the_genomic_accession=69649354, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  21926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21926, CurrentID=0, Status='live', Symbol='Tnf', Aliases='DIF, TNF-a, TNF-alpha, TNFSF2, TNFalphaa, Tnfsf1a, Tnlg1f, Tnf', description='tumor necrosis factor', other_designations='cachectin|tumor necrosis factor ligand 1f|tumor necrosis factor ligand superfamily member 2|tumor necrosis factor-alpha', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35199367, end_position_on_the_genomic_accession=35202007, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  21927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21927, CurrentID=0, Status='live', Symbol='Tnfaip1', Aliases='Edp-1, Edp1, Tnfip1', description='tumor necrosis factor, alpha-induced protein 1 (endothelial)', other_designations='BTB/POZ domain-containing protein TNFAIP1|tumor necrosis factor induced protein 1', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78522850, end_position_on_the_genomic_accession=78536260, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  21928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21928, CurrentID=0, Status='live', Symbol='Tnfaip2', Aliases='B94, Exoc3l3, M-sec, Tnfip2, tnfb94', description='tumor necrosis factor, alpha-induced protein 2', other_designations='TNF alpha-induced protein 2|primary response gene B94 protein|tumor necrosis factor induced protein 2', map_location='12 60.99 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111442446, end_position_on_the_genomic_accession=111455018, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  21929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21929, CurrentID=0, Status='live', Symbol='Tnfaip3', Aliases='A20, Tnfip3', description='tumor necrosis factor, alpha-induced protein 3', other_designations='TNF alpha-induced protein 3|putative DNA-binding protein A20|tumor necrosis factor induced protein 3|zinc finger protein A20', map_location='10 8.08 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19000910, end_position_on_the_genomic_accession=19015410, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  21930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21930, CurrentID=0, Status='live', Symbol='Tnfaip6', Aliases='TSG-6, Tnfip6, Tsg6', description='tumor necrosis factor alpha induced protein 6', other_designations='TNF alpha-induced protein 6|TNF-stimulated gene 6 protein|tumor necrosis factor alpha-induced protein 6|tumor necrosis factor induced protein 6|tumor necrosis factor, alpha-induced protein 6', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=52038113, end_position_on_the_genomic_accession=52056681, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  21933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21933, CurrentID=0, Status='live', Symbol='Tnfrsf10b', Aliases='DR5, KILLER, Ly98, MK, TRAILR2, TRICK2A, TRICK2B, TRICKB', description='tumor necrosis factor receptor superfamily, member 10b', other_designations='KILLER/DR5 TRAIL death-inducing receptor|death receptor 5', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69767472, end_position_on_the_genomic_accession=69784411, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21934, CurrentID=0, Status='live', Symbol='Tnfrsf11a', Aliases='Ly109, ODFR, OFE, Rank, TRANCE-R, mRANK', description='tumor necrosis factor receptor superfamily, member 11a, NFKB activator', other_designations='osteoclast differentiation factor receptor|receptor activator of NF-KB|receptor activator of NF-kappaB', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=105780723, end_position_on_the_genomic_accession=105847981, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21935, CurrentID=0, Status='live', Symbol='Tnfrsf17', Aliases='BCM, BCMA, Tnfrsf13, Tnfrsf13a', description='tumor necrosis factor receptor superfamily, member 17', other_designations='B-cell maturation protein', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11313174, end_position_on_the_genomic_accession=11320074, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21936, CurrentID=0, Status='live', Symbol='Tnfrsf18', Aliases='AITR, Gitr', description='tumor necrosis factor receptor superfamily, member 18', other_designations='glucocorticoid-induced TNFR-related protein', map_location='4 E', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156026342, end_position_on_the_genomic_accession=156028891, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21937, CurrentID=0, Status='live', Symbol='Tnfrsf1a', Aliases='CD120a, FPF, TNF-R, TNF-R-I, TNF-R1, TNF-R55, TNF-alphaR1, TNFAR, TNFR60, TNFRI, TNFRp55, TNFalpha-R1, Tnfr-2, Tnfr1, p55, p55-R', description='tumor necrosis factor receptor superfamily, member 1a', other_designations='TNF receptor alpha chain|TNF-RI|TNF-alpha-R1|TNFR-I|p60|tumor necrosis factor receptor 1|tumor necrosis factor receptor type I', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125349723, end_position_on_the_genomic_accession=125362483, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21938, CurrentID=0, Status='live', Symbol='Tnfrsf1b', Aliases='CD120b, TNF-R-II, TNF-R2, TNF-R75, TNF-alphaR2, TNFBR, TNFR80, TNFRII, TNFalpha-R2, Tnfr-1, Tnfr2, p75', description='tumor necrosis factor receptor superfamily, member 1b', other_designations='TNF receptor beta chain|TNF-RII|TNFR-II|p75 TNFR|p80 TNF-alpha receptor|tumor necrosis factor receptor 2|tumor necrosis factor receptor type II', map_location='4 78.17 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=145212368, end_position_on_the_genomic_accession=145246870, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  21939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21939, CurrentID=0, Status='live', Symbol='Cd40', Aliases='AI326936, Bp50, GP39, HIGM1, IGM, IMD3, T-BAM, TRAP, Tnfrsf5, p50', description='CD40 antigen', other_designations='B-cell surface antigen CD40|CD40L receptor|T-cell differentiation antigen|tumor necrosis factor receptor superfamily, member 5', map_location='2 85.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165055604, end_position_on_the_genomic_accession=165071654, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21940, CurrentID=0, Status='live', Symbol='Cd27', Aliases='S152, Tnfrsf7, Tp55', description='CD27 antigen', other_designations='CD antigen 27|CD27L receptor|T-cell activation antigen CD27|tumor necrosis factor receptor superfamily member 7|tumor necrosis factor receptor superfamily, member 7', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125232622, end_position_on_the_genomic_accession=125237080, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  21941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21941, CurrentID=0, Status='live', Symbol='Tnfrsf8', Aliases='Cd30, D1S166E, Ki, Ki-1', description='tumor necrosis factor receptor superfamily, member 8', other_designations='CD30L receptor|lymphocyte activation antigen CD30', map_location='4 78.17 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=145268976, end_position_on_the_genomic_accession=145315147, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  21942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21942, CurrentID=0, Status='live', Symbol='Tnfrsf9', Aliases='4-1BB, A930040I11Rik, AA408498, AI325004, CDw137, Cd137, ILA, Ly63', description='tumor necrosis factor receptor superfamily, member 9', other_designations='4-1BB ligand receptor|CD137 antigen|T-cell antigen 4-1BB|secreted CD137 antigen', map_location='4 81.52 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150920155, end_position_on_the_genomic_accession=150946104, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  21943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21943, CurrentID=0, Status='live', Symbol='Tnfsf11', Aliases='Ly109l, ODF, OPG, OPGL, RANKL, Trance', description='tumor necrosis factor (ligand) superfamily, member 11', other_designations='OPG ligand|TNF-related activation-induced cytokine|osteoclast differentiation factor|osteoprotegerin ligand|receptor activator of NF-kappaB ligand|receptor activator of nuclear factor kappa B ligand|receptor activator of nuclear factor kappa-B ligand|tumor necrosis factor-related activation-induced cytokine', map_location='14 41.26 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=78277446, end_position_on_the_genomic_accession=78308043, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21944, CurrentID=0, Status='live', Symbol='Tnfsf12', Aliases='Apo3l, Dr3l, Dr3lg, Tweak', description='tumor necrosis factor (ligand) superfamily, member 12', other_designations='TNF-related weak inducer of apoptosis', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69686240, end_position_on_the_genomic_accession=69696098, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  21945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21945, CurrentID=0, Status='live', Symbol='Dedd', Aliases='CASP8IP1, DEFT, Dedpro1, FLDED1, KE05', description='death effector domain-containing', other_designations='tumor necrosis factor (ligand) superfamily, member 19', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171328529, end_position_on_the_genomic_accession=171342501, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  21946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21946, CurrentID=0, Status='live', Symbol='Pglyrp1', Aliases='PGRP, PGRP-S, Pglyrp, Tag7, Tasg7, Tnfsf3l', description='peptidoglycan recognition protein 1', other_designations='cytokine tag7', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18884690, end_position_on_the_genomic_accession=18890438, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21947, CurrentID=0, Status='live', Symbol='Cd40lg', Aliases='CD154, CD40-L, Cd40l, HIGM1, IGM, IMD3, Ly-62, Ly62, T-BAM, TRAP, Tnfsf5, Tnlg8b, gp39', description='CD40 ligand', other_designations='T-cell antigen Gp39|TNF-related activation protein|tumor necrosis factor ligand 8b|tumor necrosis factor ligand superfamily member 5', map_location='X 31.21 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=57212143, end_position_on_the_genomic_accession=57224042, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  21948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21948, CurrentID=0, Status='live', Symbol='Cd70', Aliases='CD27LG, Cd27l, Tnfsf7, Tnlg8a', description='CD70 antigen', other_designations='CD27 ligand|CD27-L|Ki-24 antigen|tumor necrosis factor (ligand) superfamily, member 7|tumor necrosis factor ligand 8a|tumor necrosis factor ligand superfamily member 7', map_location='17 C-D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57145997, end_position_on_the_genomic_accession=57149777, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  21949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21949, CurrentID=0, Status='live', Symbol='Tnfsf8', Aliases='CD153, CD30LG, Cd30l, Tnlg3a', description='tumor necrosis factor (ligand) superfamily, member 8', other_designations='CD30 ligand|CD30-L|tumor necrosis factor ligand 3a', map_location='4 34.06 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63832820, end_position_on_the_genomic_accession=63861681, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21950, CurrentID=0, Status='live', Symbol='Tnfsf9', Aliases='4-1BB-L, 4-1BBL, AI848817, Cd137l, Ly63l', description='tumor necrosis factor (ligand) superfamily, member 9', other_designations='4-1BB ligand', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57105277, end_position_on_the_genomic_accession=57107757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21951, CurrentID=0, Status='live', Symbol='Tnks', Aliases='4930554K12Rik, AI662855, ARTD5, C86528, D130072O21Rik, TANK1, mTNKS1', description='tankyrase, TRF1-interacting ankyrin-related ADP-ribose polymerase', other_designations='ADP-ribosyltransferase diphtheria toxin-like 5|TRF1-interacting ankyrin-related ADP-ribose polymerase 1|tankyrase 1|tankyrase I', map_location='8 21.16 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=34829179, end_position_on_the_genomic_accession=34965690, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  21952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21952, CurrentID=0, Status='live', Symbol='Tnni1', Aliases='2700018B22Rik, AI747285, ssTnI', description='troponin I, skeletal, slow 1', other_designations='troponin I, slow-twitch isoform', map_location='1 59.28 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135783035, end_position_on_the_genomic_accession=135810989, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  21953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21953, CurrentID=0, Status='live', Symbol='Tnni2', Aliases='', description='troponin I, skeletal, fast 2', other_designations='troponin I, fast-twitch isoform', map_location='7 87.93 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142440494, end_position_on_the_genomic_accession=142444410, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  21954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21954, CurrentID=0, Status='live', Symbol='Tnni3', Aliases='Tn1, cTnI', description='troponin I, cardiac 3', other_designations='cardiac troponin I', map_location='7 2.61 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4518308, end_position_on_the_genomic_accession=4522443, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  21955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21955, CurrentID=0, Status='live', Symbol='Tnnt1', Aliases='AW146156, Tnt, sTnT, ssTnT', description='troponin T1, skeletal, slow', other_designations='skeletal muscle slow-twitch TnT', map_location='7 2.6 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4504570, end_position_on_the_genomic_accession=4515999, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  21956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21956, CurrentID=0, Status='live', Symbol='Tnnt2', Aliases='Tnt, cTnT', description='troponin T2, cardiac', other_designations='cardiac TnT|cardiac muscle troponin T|tnTc', map_location='1 59.32 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135836316, end_position_on_the_genomic_accession=135852268, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  21957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21957, CurrentID=0, Status='live', Symbol='Tnnt3', Aliases='fTnT', description='troponin T3, skeletal, fast', other_designations='fast skeletal muscle troponin T|fast skeletal muscle troponin T3|skeletal muscle fast-twitch TnT|tnTf', map_location='7 87.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142498777, end_position_on_the_genomic_accession=142516213, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  21958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21958, CurrentID=0, Status='live', Symbol='Tnp1', Aliases='Stp-1, TP1, Tp-1', description='transition protein 1', other_designations='', map_location='1 36.94 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=73015074, end_position_on_the_genomic_accession=73015899, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  21959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21959, CurrentID=0, Status='live', Symbol='Tnp2', Aliases='TP2, Tp-2', description='transition protein 2', other_designations='transition protein 2 (alt.)', map_location='16 B1-B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10787935, end_position_on_the_genomic_accession=10788655, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  21960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21960, CurrentID=0, Status='live', Symbol='Tnr', Aliases='J1-tenascin, TN-R, janusin, restrictin', description='tenascin R', other_designations='neural recognition molecule J1-160/180', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=159523741, end_position_on_the_genomic_accession=159931729, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  21961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21961, CurrentID=0, Status='live', Symbol='Tns1', Aliases='1110018I21Rik, 1200014E20Rik, AI648117, E030018G17Rik, E030037J05Rik, Tns', description='tensin 1', other_designations='', map_location='1 38.17 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=73910231, end_position_on_the_genomic_accession=74124535, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  21968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21968, CurrentID=0, Status='live', Symbol='Tom1', Aliases='', description='target of myb1 homolog (chicken)', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=75033686, end_position_on_the_genomic_accession=75070121, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  21969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21969, CurrentID=0, Status='live', Symbol='Top1', Aliases='AI467334, D130064I21Rik, Top-1', description='topoisomerase (DNA) I', other_designations='DNA topoisomerase I', map_location='2 H2-3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=160645897, end_position_on_the_genomic_accession=160722764, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  21973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21973, CurrentID=0, Status='live', Symbol='Top2a', Aliases='Top-2', description='topoisomerase (DNA) II alpha', other_designations='DNA Topoisomerase II alpha|DNA topoisomerase II, alpha isozyme', map_location='11 62.91 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98992947, end_position_on_the_genomic_accession=99024206, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  21974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21974, CurrentID=0, Status='live', Symbol='Top2b', Aliases='D230016L12Rik, Top-2', description='topoisomerase (DNA) II beta', other_designations='DNA topoisomerase II, beta isozyme', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=16365206, end_position_on_the_genomic_accession=16430787, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  21975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21975, CurrentID=0, Status='live', Symbol='Top3a', Aliases='', description='topoisomerase (DNA) III alpha', other_designations='DNA topoisomerase III alpha|Top IIIa|topoisomerase (DNA) 3 alpha|topoisomerase 3 alpha', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60739473, end_position_on_the_genomic_accession=60777365, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  21976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21976, CurrentID=0, Status='live', Symbol='Top3b', Aliases='', description='topoisomerase (DNA) III beta', other_designations='DNA topoisomerase III beta-1|Topo III beta', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16870694, end_position_on_the_genomic_accession=16892986, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  21979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21979, CurrentID=0, Status='live', Symbol='Tp(Y)1Ct', Aliases='Sxr, Tp1Ct', description='transposition, Chr Y, Cattanach 1', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  21981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21981, CurrentID=0, Status='live', Symbol='Ppp1r13b', Aliases='AI449786, ASPP1, AW545810, Tp53bp2, Trp53bp2, p85', description='protein phosphatase 1, regulatory (inhibitor) subunit 13B', other_designations='protein phosphatase 1 regulatory subunit 13B|transformation related protein 53 binding protein 2|tumor protein p53 binding protein, 2|tumor protein p53-binding protein, 2', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111828458, end_position_on_the_genomic_accession=111908142, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  21982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21982, CurrentID=0, Status='live', Symbol='Tmem165', Aliases='AV026557, Tpardl, Tparl, pFT27', description='transmembrane protein 165', other_designations='TPA regulated locus|TPA-regulated locus protein|transmembrane protein 164|transmembrane protein PFT27|transmembrane protein TPARL', map_location='5 40.56 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76183880, end_position_on_the_genomic_accession=76209244, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  21983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21983, CurrentID=0, Status='live', Symbol='Tpbg', Aliases='5T4, AW495680, WAIF1', description='trophoblast glycoprotein', other_designations='5T4 oncofetal antigen|5T4 oncofetal trophoblast glycoprotein|5T4 oncotrophoblast glycoprotein|wnt-activated inhibitory factor 1', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=85842380, end_position_on_the_genomic_accession=85847055, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  21984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21984, CurrentID=0, Status='live', Symbol='Tpbpa', Aliases='AV033951, Tb-1, Tb1, Tpbp, b2b1247Clo', description='trophoblast specific protein alpha', other_designations='Mutant line 1247|clone 4311', map_location='13 32.57 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60938692, end_position_on_the_genomic_accession=60941935, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  21985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21985, CurrentID=0, Status='live', Symbol='Tpd52', Aliases='AI043038, mD52', description='tumor protein D52', other_designations='', map_location='3 A1-A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=8928626, end_position_on_the_genomic_accession=9004769, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  21987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21987, CurrentID=0, Status='live', Symbol='Tpd52l1', Aliases='D53', description='tumor protein D52-like 1', other_designations='mD53', map_location='10 A4-B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=31332375, end_position_on_the_genomic_accession=31446144, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  21990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21990, CurrentID=0, Status='live', Symbol='Tph1', Aliases='Tph', description='tryptophan hydroxylase 1', other_designations='tryptophan 5-monooxygenase 1', map_location='7 30.43 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46644641, end_position_on_the_genomic_accession=46672537, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  21991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21991, CurrentID=0, Status='live', Symbol='Tpi1', Aliases='AI255506, Tpi, Tpi-1', description='triosephosphate isomerase 1', other_designations='TIM|triose-phosphate isomerase', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124810592, end_position_on_the_genomic_accession=124814296, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22003, CurrentID=0, Status='live', Symbol='Tpm1', Aliases='AA986836, AI854628, TM2, Tm3, Tmpa, Tpm-1, alpha-TM', description='tropomyosin 1, alpha', other_designations='alpha tropomyosin|alpha-tropomyosin|tropomyosin-1', map_location='9 36.27 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=67022590, end_position_on_the_genomic_accession=67049401, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  22004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22004, CurrentID=0, Status='live', Symbol='Tpm2', Aliases='Tpm-2, Trop-2', description='tropomyosin 2, beta', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43513726, end_position_on_the_genomic_accession=43523583, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  546294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546294, CurrentID=0, Status='live', Symbol='Rhox3g', Aliases='OTTMUSG00000017157, Rhox3.7, Rhox3f-ps, Rhox3g', description='reproductive homeobox 3G', other_designations='reproductive homeobox 3F|reproductive homeobox 3G, pseudogene', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37623431, end_position_on_the_genomic_accession=37628139, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22017, CurrentID=0, Status='live', Symbol='Tpmt', Aliases='AW106912', description='thiopurine methyltransferase', other_designations='', map_location='13 24.5 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=47021987, end_position_on_the_genomic_accession=47043338, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22018, CurrentID=0, Status='live', Symbol='Tpo', Aliases='', description='thyroid peroxidase', other_designations='', map_location='12 C', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=30054661, end_position_on_the_genomic_accession=30132984, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  22019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22019, CurrentID=0, Status='live', Symbol='Tpp2', Aliases='TPP-2, TppII', description='tripeptidyl peptidase II', other_designations='TPP-II|tripeptidyl aminopeptidase|tripeptidyl-peptidase II', map_location='1 23.5 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=43933991, end_position_on_the_genomic_accession=44003000, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  22021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22021, CurrentID=0, Status='live', Symbol='Tpst1', Aliases='R75054, Tango13a', description='protein-tyrosine sulfotransferase 1', other_designations='TPST-1|tyrosylprotein sulfotransferase 1', map_location='5 F-G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130073326, end_position_on_the_genomic_accession=130135734, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22022, CurrentID=0, Status='live', Symbol='Tpst2', Aliases='AI448750, D5Ucla3, Tango13b, grm, grt', description='protein-tyrosine sulfotransferase 2', other_designations='TPST-2|tyrosylprotein sulfotransferase 2|tyrosylprotein sulfotransferase-2', map_location='5 54.65 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112276682, end_position_on_the_genomic_accession=112315356, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22024, CurrentID=0, Status='live', Symbol='Crisp2', Aliases='CRISP-2, GAPDL5, Tpx-1, Tpx1', description='cysteine-rich secretory protein 2', other_designations='testis specific gene 1|testis-specific protein TPX-1', map_location='17 19.54 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40764734, end_position_on_the_genomic_accession=40794146, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22025, CurrentID=0, Status='live', Symbol='Nr2c1', Aliases='4831444H07Rik, 80.3, Eenr, TR2, Tr2-11', description='nuclear receptor subfamily 2, group C, member 1', other_designations='80-3 cNDA|early embryonic nuclear receptor|mTR2|nr2c1|nuclear receptor subfamily 2, group H, member 1|orphan nuclear receptor TR2|orphan receptor, TR2-11|testicular receptor 2', map_location='10 48.81 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=94147517, end_position_on_the_genomic_accession=94197216, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  22026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22026, CurrentID=0, Status='live', Symbol='Nr2c2', Aliases='TAK1, Tr4, mKIAA4145', description='nuclear receptor subfamily 2, group C, member 2', other_designations='nuclear receptor subfamily 2, group H, member 2|orphan nuclear receptor TAK1|orphan nuclear receptor TR4|orphan receptor, TR4|testicular receptor 4', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=92090368, end_position_on_the_genomic_accession=92173058, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  22027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22027, CurrentID=0, Status='live', Symbol='Hsp90b1', Aliases='ERp99, GRP94, TA-3, Targ2, Tra-1, Tra1, endoplasmin, gp96', description='heat shock protein 90, beta (Grp94), member 1', other_designations='94 kDa glucose-regulated protein|endoplasmic reticulum resident protein 99|heat shock protein 90 kDa beta member 1|polymorphic tumor rejection antigen 1|transforming growth factor alpha regulated gene 2|tumor rejection antigen (gp96) 1|tumor rejection antigen gp96', map_location='10 43.05 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=86690841, end_position_on_the_genomic_accession=86705444, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  22029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22029, CurrentID=0, Status='live', Symbol='Traf1', Aliases='4732496E14Rik', description='TNF receptor-associated factor 1', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34943256, end_position_on_the_genomic_accession=34961772, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22030, CurrentID=0, Status='live', Symbol='Traf2', Aliases='AI325259', description='TNF receptor-associated factor 2', other_designations='E3 ubiquitin-protein ligase TRAF2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25517982, end_position_on_the_genomic_accession=25546940, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22031, CurrentID=0, Status='live', Symbol='Traf3', Aliases='AI528849, CAP-1, CD40bp, CRAF1, LAP1, T-BAM, amn', description='TNF receptor-associated factor 3', other_designations='CD40 receptor-associated factor 1|TRAFAMN', map_location='12 F', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111166370, end_position_on_the_genomic_accession=111267155, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  22032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22032, CurrentID=0, Status='live', Symbol='Traf4', Aliases='A530032M13Rik, CART1, msp2', description='TNF receptor associated factor 4', other_designations='cysteine-rich motif associated to RING and Traf domains protein 1', map_location='11 B5-C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78158423, end_position_on_the_genomic_accession=78165569, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22033, CurrentID=0, Status='live', Symbol='Traf5', Aliases='', description='TNF receptor-associated factor 5', other_designations='', map_location='1 97.11 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=192052688, end_position_on_the_genomic_accession=192092599, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22034, CurrentID=0, Status='live', Symbol='Traf6', Aliases='2310003F17Rik, AI851288, C630032O20Rik', description='TNF receptor-associated factor 6', other_designations='E3 ubiquitin-protein ligase TRAF6', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=101678420, end_position_on_the_genomic_accession=101701668, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22035, CurrentID=0, Status='live', Symbol='Tnfsf10', Aliases='A330042I21Rik, AI448571, APO-2L, Ly81, TL2, Tnlg6a, Trail', description='tumor necrosis factor (ligand) superfamily, member 10', other_designations='TNF-related apoptosis inducing ligand|TNF-related apoptosis-inducing ligand|TRAIL/APO2L|tumor necrosis factor ligand 6a', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=27317077, end_position_on_the_genomic_accession=27339665, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22036, CurrentID=0, Status='live', Symbol='Traip', Aliases='Trip', description='TRAF-interacting protein', other_designations='', map_location='9 59.07 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107950136, end_position_on_the_genomic_accession=107972270, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  22037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22037, CurrentID=0, Status='live', Symbol='Trap1a', Aliases='Tra-P1A', description='tumor rejection antigen P1A', other_designations='', map_location='X 61.35 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139333683, end_position_on_the_genomic_accession=139338165, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22038, CurrentID=0, Status='live', Symbol='Plscr1', Aliases='MmTRA1a, MmTRA1b, Nor1, Tra1, Tra1a, Tra1b, Tras1, Tras2', description='phospholipid scramblase 1', other_designations='PL scramblase 1|ca(2+)-dependent phospholipid scramblase 1|transplantability-associated protein 1', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=92250194, end_position_on_the_genomic_accession=92272561, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22040, CurrentID=0, Status='live', Symbol='Trex1', Aliases='AU041952', description='three prime repair exonuclease 1', other_designations="3'-5' exonuclease TREX1|DNase III|three prime exonuclease 1|trophoblast expressed 1", map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109057932, end_position_on_the_genomic_accession=109059723, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22041, CurrentID=0, Status='live', Symbol='Trf', Aliases='AI266983, Cd176, HP, Tf, Tfn, hpx', description='transferrin', other_designations='beta-1 metal-binding globulin|hypotransferrinemia with hemochromatosis|siderophilin', map_location='9 F1-F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103208876, end_position_on_the_genomic_accession=103230286, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  22042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22042, CurrentID=0, Status='live', Symbol='Tfrc', Aliases='2610028K12Rik, CD71, E430033M20Rik, Mtvr1, TFR, TFR1, TR, Trfr, p90', description='transferrin receptor', other_designations='mammary tumor virus receptor 1', map_location='16 23.06 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32608896, end_position_on_the_genomic_accession=32632794, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  22044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22044, CurrentID=0, Status='live', Symbol='Trh', Aliases='Pro-TRH, Trf', description='thyrotropin releasing hormone', other_designations='TSH-releasing factor|pro-thyrotropin-releasing hormone|prothyroliberin|protirelin|thyrotropin-releasing factor', map_location='6 41.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=92242061, end_position_on_the_genomic_accession=92244650, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22045, CurrentID=0, Status='live', Symbol='Trhr', Aliases='TRH-R1', description='thyrotropin releasing hormone receptor', other_designations='TRH-R|thyroliberin receptor', map_location='15 16.82 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=44195924, end_position_on_the_genomic_accession=44235912, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  546336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546336, CurrentID=0, Status='live', Symbol='Prrg1', Aliases='2010007L08Rik', description='proline rich Gla (G-carboxyglutamic acid) 1', other_designations='ortholog of proline rich Gla (G-carboxyglutamic acid) 1 PRRG1', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=78449610, end_position_on_the_genomic_accession=78583913, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22051, CurrentID=0, Status='live', Symbol='Trip6', Aliases='', description='thyroid hormone receptor interactor 6', other_designations='TR-interacting protein 6|TRIP-6|ZRP-1|zyxin-related protein 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137309899, end_position_on_the_genomic_accession=137314487, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22053, CurrentID=0, Status='live', Symbol='n-Th2', Aliases='Trnah, Trnah1', description='nuclear encoded tRNA histidine 2 (anticodon GTG)', other_designations='histidine transfer RNA 1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96435972, end_position_on_the_genomic_accession=96436043, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22057, CurrentID=0, Status='live', Symbol='Tob1', Aliases='Tob, Trob', description='transducer of ErbB-2.1', other_designations='transducer of erbB2', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94211454, end_position_on_the_genomic_accession=94215495, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  22059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22059, CurrentID=0, Status='live', Symbol='Trp53', Aliases='Tp53, bbl, bfy, bhy, p44, p53', description='transformation related protein 53', other_designations='mutant p53|p53 cellular tumor antigen|tumor suppressor p53|tumor supressor p53', map_location='11 B2-C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69580359, end_position_on_the_genomic_accession=69591873, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  22060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22060, CurrentID=0, Status='live', Symbol='Trp53-ps', Aliases='p53', description='transformation related protein 53, pseudogene', other_designations='p53-variant', map_location='17 28.3 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=54419597, end_position_on_the_genomic_accession=54421236, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22061, CurrentID=0, Status='live', Symbol='Trp63', Aliases='AI462811, Ket, P51/P63, P63, P73l, Tp63, Trp53rp1', description='transformation related protein 63', other_designations='transformation-related protein 63', map_location='16 17.37 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=25683763, end_position_on_the_genomic_accession=25892103, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  22062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22062, CurrentID=0, Status='live', Symbol='Trp73', Aliases='TAp73, Tp73, deltaNp73, p73', description='transformation related protein 73', other_designations='', map_location='4 83.79 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154056249, end_position_on_the_genomic_accession=154140208, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  22063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22063, CurrentID=0, Status='live', Symbol='Trpc1', Aliases='Mtrp1, Trp1, Trrp1', description='transient receptor potential cation channel, subfamily C, member 1', other_designations='TRP-1|transient receptor protein 1|trp-related protein 1', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95705072, end_position_on_the_genomic_accession=95750390, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  22064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22064, CurrentID=0, Status='live', Symbol='Trpc2', Aliases='3010009O07Rik, AI115608, TRPC2a, TRPC2b, Trrp2, mTrp2, smTRPC2, trp2', description='transient receptor potential cation channel, subfamily C, member 2', other_designations='transient receptor potential 2-like protein', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102083116, end_position_on_the_genomic_accession=102096864, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  22065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22065, CurrentID=0, Status='live', Symbol='Trpc3', Aliases='Mwk, Trcp3, Trp3, Trrp3', description='transient receptor potential cation channel, subfamily C, member 3', other_designations='TRP-3|mTrp3|receptor-activated cation channel TRP3|transient receptor protein 3|trp-related protein 3', map_location='3 17.93 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36620482, end_position_on_the_genomic_accession=36690229, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22066, CurrentID=0, Status='live', Symbol='Trpc4', Aliases='CCE1, STRPC4, Trp4, Trrp4', description='transient receptor potential cation channel, subfamily C, member 4', other_designations='capacitative calcium entry channel Trp4|putative capacitative calcium entry  channel|receptor-activated cation channel TRP4|transient receptor protein 4', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54156057, end_position_on_the_genomic_accession=54318471, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  22067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22067, CurrentID=0, Status='live', Symbol='Trpc5', Aliases='CCE2, TRP-5, TRP5, Trrp5', description='transient receptor potential cation channel, subfamily C, member 5', other_designations='capacitative calcium entry channel 2|transient receptor protein 5|trp-related protein 5|vtransient receptor potential cation channel, subfamily C, member 5', map_location='X 65.49 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=144377614, end_position_on_the_genomic_accession=144688180, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22068, CurrentID=0, Status='live', Symbol='Trpc6', Aliases='AV025995, LLHWJM002, LLHWJM003, LLHWJM004, TRP-6, Trrp6, mtrp6', description='transient receptor potential cation channel, subfamily C, member 6', other_designations='calcium entry channel|transient receptor protein 6', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=8543932, end_position_on_the_genomic_accession=8680752, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  22069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22069, CurrentID=0, Status='live', Symbol='n-Ts2', Aliases='Trnau1, Trsp, tRNA-Sec', description='nuclear encoded tRNA selenocysteine 2 (anticodon TCA)', other_designations='opal suppressor|selenocysteine tRNA|tRNA Sec|tRNA phosphoserine|transfer RNA selenocysteine 1 (anticodon UCA)', map_location='7 9.56 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19301244, end_position_on_the_genomic_accession=19301329, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22070, CurrentID=0, Status='live', Symbol='Tpt1', Aliases='TCTP, Trt, p21, p23', description='tumor protein, translationally-controlled 1', other_designations='21 kDa polypeptide under translational control|fortilin|translationally regulated transcript (21 kDa)', map_location='14 40.29 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75845256, end_position_on_the_genomic_accession=75848303, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22072, CurrentID=0, Status='live', Symbol='Prss2', Aliases='TRY8, TRYP, Ta, Tesp4, Try2', description='protease, serine 2', other_designations='anionic trypsin II|pretrypsinogen II|protease, serine, 2|serine protease 2|trypsin 2|trypsinogen 20', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41521776, end_position_on_the_genomic_accession=41525079, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22073, CurrentID=0, Status='live', Symbol='Prss3', Aliases='MTG, TRY4, Tb, Try3', description='protease, serine 3', other_designations='protease, serine, 3|trypsin 3|trypsinogen 15', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41373759, end_position_on_the_genomic_accession=41377613, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22074, CurrentID=0, Status='live', Symbol='Try4', Aliases='0910001B19Rik, Td', description='trypsin 4', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41302272, end_position_on_the_genomic_accession=41305533, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  331360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331360, CurrentID=0, Status='live', Symbol='Zf12', Aliases='', description='Zf12 pseudogene', other_designations='zinc finger protein 191 pseudogene', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22083, CurrentID=0, Status='live', Symbol='Ctr9', Aliases='AA409336, Sh2bp1, Tsbp, Tsp, mKIAA0155', description='CTR9 homolog, Paf1/RNA polymerase II complex component', other_designations='Ctr9, Paf1/RNA polymerase II complex component, homolog|SH2 domain binding protein 1 (tetratricopeptide repeat containing)|SH2 domain-binding protein 1|TPR-containing, SH2-binding phosphoprotein of 150 kDa|p150TSP|tetratricopeptide repeat-containing, SH2-binding phosphoprotein of 150 kDa', map_location='7 57.98 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=111028951, end_position_on_the_genomic_accession=111056377, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  22084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22084, CurrentID=0, Status='live', Symbol='Tsc2', Aliases='Nafld, Tcs2', description='tuberous sclerosis 2', other_designations='tuberous sclerosis 2 protein homolog', map_location='17 12.41 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24595816, end_position_on_the_genomic_accession=24632627, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  22088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22088, CurrentID=0, Status='live', Symbol='Tsg101', Aliases='AI255943, CC2', description='tumor susceptibility gene 101', other_designations='ESCRT-I complex subunit TSG101', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46889027, end_position_on_the_genomic_accession=46919974, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22092, CurrentID=0, Status='live', Symbol='Rsph1', Aliases='MCA, Tsga2', description='radial spoke head 1 homolog (Chlamydomonas)', other_designations='TSGA2 epididymal isoform|male meiotic metaphase chromosome-associated acidic protein|meichroacidin|testis specific gene A2|testis-specific gene A2 protein', map_location='17 15.94 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31255019, end_position_on_the_genomic_accession=31277356, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22094, CurrentID=0, Status='live', Symbol='Tshb', Aliases='', description='thyroid stimulating hormone, beta subunit', other_designations='TSH-B|TSH-beta|thyroid-stimulating hormone subunit beta|thyrotropin beta chain', map_location='3 45.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102773624, end_position_on_the_genomic_accession=102803490, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22095, CurrentID=0, Status='live', Symbol='Tshr', Aliases='AI481368, hypothroid, hyt, pet', description='thyroid stimulating hormone receptor', other_designations='TSH-R|thyroid-stimulating hormone receptor', map_location='12 44.51 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=91400993, end_position_on_the_genomic_accession=91540509, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  22097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22097, CurrentID=0, Status='live', Symbol='Tsix', Aliases='', description='X (inactive)-specific transcript, opposite strand', other_designations='X (inactive)-specific transcript, antisense|antisense of Xist', map_location='X 46.13 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103431517, end_position_on_the_genomic_accession=103484957, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22099, CurrentID=0, Status='live', Symbol='Tsn', Aliases='2610034C24Rik, AU040286, C3PO, TB-RBP', description='translin', other_designations='component 3 of promoter of RISC|testis/brain RNA-binding protein|testis/brain-RNA-binding protein', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=118298517, end_position_on_the_genomic_accession=118311132, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22109, CurrentID=0, Status='live', Symbol='Tspy-ps', Aliases='', description='testis specific protein-Y encoded, pseudogene', other_designations='', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=1055764, end_position_on_the_genomic_accession=1058868, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22110, CurrentID=0, Status='live', Symbol='Tspyl1', Aliases='Tspyl', description='testis-specific protein, Y-encoded-like 1', other_designations='TSPY-like 1|TSPY-like protein 1', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34282190, end_position_on_the_genomic_accession=34284885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  22113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22113, CurrentID=0, Status='live', Symbol='Phlda2', Aliases='Ipl, Tssc3', description='pleckstrin homology-like domain, family A, member 2', other_designations='imprinted in placenta and liver protein|protein 50C15|tumor-suppressing subchromosomal transferable fragment 3', map_location='7 88.23 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143501545, end_position_on_the_genomic_accession=143503145, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22114, CurrentID=0, Status='live', Symbol='Tssk1', Aliases='Stk22a, TSK-1, Tsk1, Tsskb, Tssk1', description='testis-specific serine kinase 1', other_designations='TSSK-1|serine/threonine kinase 22A (spermiogenesis associated)|serine/threonine-protein kinase 22A|testis-specific kinase 1|testis-specific serine/threonine kinase', map_location='16 11.09 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17894203, end_position_on_the_genomic_accession=17895653, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  22115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22115, CurrentID=0, Status='live', Symbol='Tssk2', Aliases='DGS-G, SPOGA2, Stk22b, Tsk2', description='testis-specific serine kinase 2', other_designations='TSK-2|TSSK-2|serine/threonine kinase 22B (spermiogenesis associated)|serine/threonine-protein kinase 22B|testis specific serine/ threonine kinase 2|testis-specific kinase 2', map_location='16 11.09 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17898637, end_position_on_the_genomic_accession=17900024, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  22116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22116, CurrentID=0, Status='live', Symbol='Tsks', Aliases='Stk22s1, Tssks1', description='testis-specific serine kinase substrate', other_designations='STK22 substrate 1|clone 4|serine/threonine kinase 22 substrate 1|testis-specific kinase substrate|testis-specific serine/threonine kinase substrate 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44929577, end_position_on_the_genomic_accession=44958035, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  22117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22117, CurrentID=0, Status='live', Symbol='Tst', Aliases='Rhodanese', description='thiosulfate sulfurtransferase, mitochondrial', other_designations='', map_location='15 37.47 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78399556, end_position_on_the_genomic_accession=78405859, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22121, CurrentID=0, Status='live', Symbol='Rpl13a', Aliases='1810026N22Rik, Tstap198-7, tum-antigen', description='ribosomal protein L13A', other_designations='tissue specific transplantation antigen P198-7|transplantation antigen P198|tum-P198 antigen|tum-transplantation antigen P198', map_location='7 29.14 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45125563, end_position_on_the_genomic_accession=45128745, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22122, CurrentID=0, Status='live', Symbol='Tsta3', Aliases='AI256181, FX, Tstap35b', description='tissue specific transplantation antigen P35B', other_designations='GDP-4-keto-6-deoxy-D-mannose-3,5-epimerase-4-reductase|GDP-L-fucose synthetase|GDP-fucose synthase|protein FX|red cell NADP(H)-binding protein|transplantation antigen P35B|tum-P35B antigen', map_location='15 35.1 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75924669, end_position_on_the_genomic_accession=75929821, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22123, CurrentID=0, Status='live', Symbol='Psmd3', Aliases='AI255837, AntP91a, D11Bwg1349e, Psd3, Tstap91a', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 3', other_designations='26S proteasome regulatory subunit RPN3|26S proteasome regulatory subunit S3|proteasome subunit p58|tissue specific transplantation antigen P91A|transplantation antigen P91A|tum-P91A antigen', map_location='11 62.39 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98682554, end_position_on_the_genomic_accession=98695978, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  22127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22127, CurrentID=0, Status='live', Symbol='Tsx', Aliases='AV207184', description='testis specific X-linked gene', other_designations='testis-specific X-linked protein', map_location='X 46.12 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103409101, end_position_on_the_genomic_accession=103424588, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22129, CurrentID=0, Status='live', Symbol='Ttc3', Aliases='2610202A04Rik, AA409221, D16Ium21, D16Ium21e, TPRD, mKIAA4119', description='tetratricopeptide repeat domain 3', other_designations='TPR repeat protein D|mtprd', map_location='16 C3.3-4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94370739, end_position_on_the_genomic_accession=94469222, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  22130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22130, CurrentID=0, Status='live', Symbol='Ttf1', Aliases='AV245725, TTF-1, TTF-I', description='transcription termination factor, RNA polymerase I', other_designations='RNA polymerase I termination factor|transcription termination factor I', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29060199, end_position_on_the_genomic_accession=29087656, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  22134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22134, CurrentID=0, Status='live', Symbol='Tgoln1', Aliases='AU015504, D6Ertd384e, TGN38, TGN38A, Ttgn1', description='trans-golgi network protein', other_designations='trans-golgi network protein 1', map_location='6 32.27 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72608420, end_position_on_the_genomic_accession=72617000, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22135, CurrentID=0, Status='live', Symbol='Tgoln2', Aliases='TGN38B, TGN41, TGN46, Ttgn2', description='trans-golgi network protein 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22137, CurrentID=0, Status='live', Symbol='Ttk', Aliases='AL022661, Esk1, Mps1, PYT, esk', description='Ttk protein kinase', other_designations='', map_location='9 E2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=83834651, end_position_on_the_genomic_accession=83872390, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  22138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22138, CurrentID=0, Status='live', Symbol='Ttn', Aliases='1100001C23Rik, 2310036G12Rik, 2310057K23Rik, 2310074I15Rik, AF006999, AV006427, D330041I19Rik, D830007G01Rik, L56, mdm, shru', description='titin', other_designations='connectin|ttn', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76703980, end_position_on_the_genomic_accession=76982557, orientation='minus', exon_count=349, OMIM='', no_hdr0=''),
  22139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22139, CurrentID=0, Status='live', Symbol='Ttr', Aliases='AA408768, AI787086, D17860, prealbumin', description='transthyretin', other_designations='', map_location='18 11.47 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20665250, end_position_on_the_genomic_accession=20674326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22141, CurrentID=0, Status='live', Symbol='Tub', Aliases='rd5, tub', description='tubby candidate gene', other_designations='retinal degeneration 5', map_location='7 57.21 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109010880, end_position_on_the_genomic_accession=109034459, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  22142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22142, CurrentID=0, Status='live', Symbol='Tuba1a', Aliases='Tuba-1, Tuba1', description='tubulin, alpha 1A', other_designations='Talpha1 alpha-tubulin|alpha-tubulin 1|alpha-tubulin isotype M-alpha-1|tubulin alpha-1 chain|tubulin, alpha 1', map_location='15 55.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98949847, end_position_on_the_genomic_accession=98953501, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22143, CurrentID=0, Status='live', Symbol='Tuba1b', Aliases='Tuba2', description='tubulin, alpha 1B', other_designations='alpha-tubulin 2|alpha-tubulin isotype M-alpha-2|tubulin alpha-2 chain|tubulin, alpha 2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98931427, end_position_on_the_genomic_accession=98934560, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22144, CurrentID=0, Status='live', Symbol='Tuba3a', Aliases='AI528779, M[a]3, Tuba3', description='tubulin, alpha 3A', other_designations='alpha-tubulin 3/7|alpha-tubulin isotype M-alpha-3/7|tubulin, alpha 3', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125278274, end_position_on_the_genomic_accession=125286042, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22145, CurrentID=0, Status='live', Symbol='Tuba4a', Aliases='M[a]4, Tuba4', description='tubulin, alpha 4A', other_designations='alpha-tubulin isotype M-alpha-4|tubulin alpha-4 chain', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75214974, end_position_on_the_genomic_accession=75219860, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22146, CurrentID=0, Status='live', Symbol='Tuba1c', Aliases='M[a]6, Tuba6', description='tubulin, alpha 1C', other_designations='alpha-tubulin 6|alpha-tubulin isotype M-alpha-6|tubulin alpha-6 chain|tubulin, alpha 6', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99029891, end_position_on_the_genomic_accession=99038105, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22147, CurrentID=0, Status='live', Symbol='Tuba3b', Aliases='M[a]7, Tuba7', description='tubulin, alpha 3B', other_designations='alpha-tubulin 3/7|alpha-tubulin isotype M-alpha-3/7|tubulin alpha-3/alpha-7 chain|tubulin, alpha 7', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145615963, end_position_on_the_genomic_accession=145621477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22151, CurrentID=0, Status='live', Symbol='Tubb2a', Aliases='M(beta)2, Tubb2', description='tubulin, beta 2A class IIA', other_designations='tubulin, beta polypeptide', map_location='13 14.03 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34074278, end_position_on_the_genomic_accession=34078008, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22152, CurrentID=0, Status='live', Symbol='Tubb3', Aliases='3200002H15Rik, M(beta)3, M(beta)6', description='tubulin, beta 3 class III', other_designations='betaIII-tubulin', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123411564, end_position_on_the_genomic_accession=123422010, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22153, CurrentID=0, Status='live', Symbol='Tubb4a', Aliases='AI325297, M(beta)4, Tubb, Tubb4', description='tubulin, beta 4A class IVA', other_designations='tubulin beta-4 chain|tubulin, beta 4', map_location='17 29.66 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57080066, end_position_on_the_genomic_accession=57087782, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22154, CurrentID=0, Status='live', Symbol='Tubb5', Aliases='AA408537, AI596182, B130022C14Rik, M(beta)5', description='tubulin, beta 5 class I', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35833920, end_position_on_the_genomic_accession=35838301, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22156, CurrentID=0, Status='live', Symbol='Tuft1', Aliases='', description='tuftelin 1', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94612753, end_position_on_the_genomic_accession=94658872, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  22157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22157, CurrentID=0, Status='live', Symbol='Tulp1', Aliases='Tulp1l', description='tubby like protein 1', other_designations='tubby like protein 1 like|tubby-like protein 1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28351515, end_position_on_the_genomic_accession=28365211, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  22158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22158, CurrentID=0, Status='live', Symbol='Tulp3', Aliases='2310022L06Rik, AI316887', description='tubby-like protein 3', other_designations='', map_location='6 62.96 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128321161, end_position_on_the_genomic_accession=128355872, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22160, CurrentID=0, Status='live', Symbol='Twist1', Aliases='M-Twist, Pde, Ska10, Ska<m10Jus>, Twist, bHLHa38, pdt', description='twist basic helix-loop-helix transcription factor 1', other_designations='Pluridigite|charlie chaplin|polydactyly EMS|twist gene homolog 1|twist homolog 1', map_location='12 B-C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=33957671, end_position_on_the_genomic_accession=33959831, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  22163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22163, CurrentID=0, Status='live', Symbol='Tnfrsf4', Aliases='ACT35, CD134, Ly-70, Ox40, TXGP1L, Txgp1', description='tumor necrosis factor receptor superfamily, member 4', other_designations='OX40 antigen|OX40L receptor|tax-transcriptionally activated glycoprotein 1', map_location='4 87.68 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156013695, end_position_on_the_genomic_accession=156016613, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22164, CurrentID=0, Status='live', Symbol='Tnfsf4', Aliases='Ath-1, Ath1, CD134L, OX-40L, Ox40l, TXGP1, Tnlg2b, Txgp1l, gp34', description='tumor necrosis factor (ligand) superfamily, member 4', other_designations='OX40 ligand|atherosclerosis 1|tax-transcriptionally activated glycoprotein 1 ligand|tumor necrosis factor ligand 2b', map_location='1 69.75 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161395438, end_position_on_the_genomic_accession=161418206, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22165, CurrentID=0, Status='live', Symbol='Txk', Aliases='A130089B16Rik, Btkl, PTK-RL-18, PTK4, Rlk', description='TXK tyrosine kinase', other_designations='resting lymphocyte kinase', map_location='5 38.44 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72695978, end_position_on_the_genomic_accession=72752992, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  22166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22166, CurrentID=0, Status='live', Symbol='Txn1', Aliases='ADF, AW550880, Trx1, Txn', description='thioredoxin 1', other_designations='ATL-derived factor|trx', map_location='4 31.87 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=57943373, end_position_on_the_genomic_accession=57956411, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22167, CurrentID=0, Status='live', Symbol='Txn-ps1', Aliases='', description='thioredoxin, pseudogene 1', other_designations='', map_location='1 23.63 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44463868, end_position_on_the_genomic_accession=44464293, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  22169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22169, CurrentID=0, Status='live', Symbol='Cmpk2', Aliases='1200004E04Rik, TDKI, Tyki', description='cytidine monophosphate (UMP-CMP) kinase 2, mitochondrial', other_designations='nucleoside-diphosphate kinase|thymidylate kinase LPS-inducible member|thymidylate kinase homologue', map_location='12 9.69 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=26469059, end_position_on_the_genomic_accession=26479837, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22171, CurrentID=0, Status='live', Symbol='Tyms', Aliases='Ts', description='thymidylate synthase', other_designations='Tsase', map_location='5 15.81 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30058200, end_position_on_the_genomic_accession=30073625, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22173, CurrentID=0, Status='live', Symbol='Tyr', Aliases='Oca1, albino, c, skc35', description='tyrosinase', other_designations='albino locus protein|monophenol monooxygenase', map_location='7 49.01 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=87425276, end_position_on_the_genomic_accession=87493442, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22174, CurrentID=0, Status='live', Symbol='Tyro3', Aliases='AI323366, Brt, Dtk, Etk-2, Rse, Sky, TK19-2, Tif, etk2/tyro3, tk19-1', description='TYRO3 protein tyrosine kinase 3', other_designations='receptor tyrosine kinase|tyrosine-protein kinase DTK|tyrosine-protein kinase RSE|tyrosine-protein kinase TIF', map_location='2 F', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119797740, end_position_on_the_genomic_accession=119818104, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  22177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22177, CurrentID=0, Status='live', Symbol='Tyrobp', Aliases='DAP12, KARAP, Ly83', description='TYRO protein tyrosine kinase binding protein', other_designations='DNAX-activation protein 12|KAR-associated protein|killer cell activating receptor associated protein|killer-activating receptor-associated protein', map_location='7 B', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30413788, end_position_on_the_genomic_accession=30417582, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22178, CurrentID=0, Status='live', Symbol='Tyrp1', Aliases='Oca3, TRP-1, TRP1, Tyrp, b, brown, isa', description='tyrosinase-related protein 1', other_designations='DHICA oxidase|brown locus protein|catalase B|iris stromal atrophy', map_location='4 37.89 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=80834208, end_position_on_the_genomic_accession=80851736, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22183, CurrentID=0, Status='live', Symbol='Zrsr1', Aliases='D11Ncvs75, Irlgs2, SP2, U2af1-rs1, U2afbp-rs', description='zinc finger (CCCH type), RNA binding motif and serine/arginine rich 1', other_designations='CCCH type zinc finger, RNA-binding motif and serine/arginine rich protein 1|U2 small nuclear ribonucleoprotein auxiliary factor (U2AF) 1, related sequence 1|U2 small nuclear ribonucleoprotein auxiliary factor (U2AF), 35 kDa, related sequence 1|U2 small nuclear ribonucleoprotein auxiliary factor (U2AF), related sequence 1|U2(RNU2) small nuclear RNA auxiliary factor 1-like 1', map_location='11 14.24 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=22972005, end_position_on_the_genomic_accession=22976496, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  22184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22184, CurrentID=0, Status='live', Symbol='Zrsr2', Aliases='35kDa, 5031411E02Rik, A230052C13Rik, C77286, U2af1-rs2, URP', description='zinc finger (CCCH type), RNA binding motif and serine/arginine rich 2', other_designations='U2 small nuclear ribonucleoprotein auxiliary factor (U2AF), 35 kDa, related sequence 2|U2(RNU2) small nuclear RNA auxiliary factor 1-like 2|U2AF35-related protein', map_location='X 75.99 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=163935443, end_position_on_the_genomic_accession=163958820, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  22185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22185, CurrentID=0, Status='live', Symbol='U2af2', Aliases='65kDa', description='U2 small nuclear ribonucleoprotein auxiliary factor (U2AF) 2', other_designations='U2 auxiliary factor 65 kDa subunit|U2 small nuclear ribonucleoprotein auxiliary factor (U2AF), 65 kDa|U2 snRNP auxiliary factor large subunit', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5060989, end_position_on_the_genomic_accession=5079945, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  22186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22186, CurrentID=0, Status='live', Symbol='Uba52', Aliases='D8Ertd21e, Gm1863, Rps27a, Ubb, Ubc', description='ubiquitin A-52 residue ribosomal protein fusion product 1', other_designations='60S ribosomal protein L40|CEP52|ubiquitin/60S ribosomal fusion protein', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70508266, end_position_on_the_genomic_accession=70510367, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22187, CurrentID=0, Status='live', Symbol='Ubb', Aliases='AL033289, Rps27a, Uba522, Ubc, Ubb', description='ubiquitin B', other_designations='', map_location='11 38.46 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62551504, end_position_on_the_genomic_accession=62553213, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22190, CurrentID=0, Status='live', Symbol='Ubc', Aliases='2700054O04Rik, AI194771, Rps27a, TI-225, Uba52, Ubb', description='ubiquitin C', other_designations='polyubiquitin C', map_location='5 64.18 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=125385965, end_position_on_the_genomic_accession=125390017, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22192, CurrentID=0, Status='live', Symbol='Ube2m', Aliases='2510040H03Rik, UBC12, Ubc-rs2', description='ubiquitin-conjugating enzyme E2M', other_designations='NEDD8 carrier protein|NEDD8 protein ligase|UBC12 homolog, yeast|ubiquitin C, related sequence 2', map_location='7 7.73 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13035120, end_position_on_the_genomic_accession=13038275, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22193, CurrentID=0, Status='live', Symbol='Ube2e3', Aliases='Ubce4, ubcM2', description='ubiquitin-conjugating enzyme E2E 3', other_designations='E2 ubiquitin-conjugating enzyme E3|ubiquitin carrier protein E3|ubiquitin-conjugating enzyme 4|ubiquitin-conjugating enzyme E2-23 kDa|ubiquitin-conjugating enzyme E2E 3, UBC4/5 homolog|ubiquitin-protein ligase E3', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=78868243, end_position_on_the_genomic_accession=78920720, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22194, CurrentID=0, Status='live', Symbol='Ube2e1', Aliases='UbcM3, Ubce5, ubcM2', description='ubiquitin-conjugating enzyme E2E 1', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme E1|E2 ubiquitin-conjugating enzyme E1|ubiquitin carrier protein E1|ubiquitin-conjugating enzyme 5|ubiquitin-conjugating enzyme E2E 1, UBC4/5 homolog|ubiquitin-protein ligase E1', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=18282726, end_position_on_the_genomic_accession=18331844, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22195, CurrentID=0, Status='live', Symbol='Ube2l3', Aliases='C79827, UbcM4, Ubce7', description='ubiquitin-conjugating enzyme E2L 3', other_designations='E2 ubiquitin-conjugating enzyme L3|ubiquitin carrier protein L3|ubiquitin-conjugating enzyme 7|ubiquitin-protein ligase L3', map_location='16 10.63 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17152015, end_position_on_the_genomic_accession=17201498, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22196, CurrentID=0, Status='live', Symbol='Ube2i', Aliases='5830467E05Rik, F830028O17Rik, UBC9, Ubce2i, Ubce9', description='ubiquitin-conjugating enzyme E2I', other_designations='SUMO-protein ligase|ubiquitin carrier protein 9|ubiquitin carrier protein I|ubiquitin-protein ligase I', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25260510, end_position_on_the_genomic_accession=25274311, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  546487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546487, CurrentID=0, Status='live', Symbol='Olfr766-ps1', Aliases='', description='olfactory receptor 766, pseudogene 1', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129063028, end_position_on_the_genomic_accession=129064675, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22200, CurrentID=0, Status='live', Symbol='Uba3', Aliases='A830034N06Rik, AI256736, AI848246, AW546539, Ube1c', description='ubiquitin-like modifier activating enzyme 3', other_designations='NEDD8-activating enzyme E1C|ubiquitin activating enzyme 3|ubiquitin-activating enzyme 3|ubiquitin-activating enzyme E1C|ubiquitin-like modifier-activating enzyme 3', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=97183632, end_position_on_the_genomic_accession=97205647, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  22201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22201, CurrentID=0, Status='live', Symbol='Uba1', Aliases='A1S9, Sbx, Ube-1, Ube1x', description='ubiquitin-like modifier activating enzyme 1', other_designations='ubiquitin-activating enzyme E1, Chr X', map_location='X A2-A3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20658302, end_position_on_the_genomic_accession=20683179, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  22202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22202, CurrentID=0, Status='live', Symbol='Uba1y', Aliases='A1s9Y-1, Sby, Ube-2, Ube1ay, Ube1y, Ube1y-1, Ube1y1', description='ubiquitin-activating enzyme, Chr Y', other_designations='ubiquitin-activating enzyme E1 Y|ubiquitin-activating enzyme E1, Chr Y 1', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=818612, end_position_on_the_genomic_accession=844224, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  22203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22203, CurrentID=0, Status='live', Symbol='Uba1y-ps1', Aliases='A1s9Y-2, Sby, Ube-3, Ube1y-1ps1, Ube1y-2, Ube1y1-ps1', description='ubiquitin-activating enzyme, Chr Y, pseudogene 1', other_designations='ubiquitin activating enzyme E1, Chr Y, pseudogene 1|ubiquitin-activating enzyme E1, Chr Y, pseudogene 1', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=2625440, end_position_on_the_genomic_accession=2649547, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  22209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22209, CurrentID=0, Status='live', Symbol='Ube2a', Aliases='HHR6A, HR6A, Mhr6a', description='ubiquitin-conjugating enzyme E2A', other_designations='E2 ubiquitin-conjugating enzyme A|RAD6 homolog A|ubiquitin carrier protein A|ubiquitin-conjugating enzyme E2A (RAD6 homolog)|ubiquitin-conjugating enzyme E2A, RAD6 homolog|ubiquitin-conjugating enzyme HR6A|ubiquitin-protein ligase A', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36874366, end_position_on_the_genomic_accession=36884222, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22210, CurrentID=0, Status='live', Symbol='Ube2b', Aliases='2610301N02Rik, E2-14k, HR6B, Rad6b, mHR6B', description='ubiquitin-conjugating enzyme E2B', other_designations='E2 ubiquitin-conjugating enzyme B|E214K|RAD6 homolog B|ubiquitin carrier protein B|ubiquitin-conjugating enzyme E2B, RAD6 homology|ubiquitin-protein ligase B', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51985146, end_position_on_the_genomic_accession=52002087, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22213, CurrentID=0, Status='live', Symbol='Ube2g2', Aliases='1110003O05Rik, D10Xrf369, UBC7, Ubc7p', description='ubiquitin-conjugating enzyme E2G 2', other_designations='E2 ubiquitin-conjugating enzyme G2|ubiquitin carrier protein G2|ubiquitin-conjugating enzyme 7 homolog|ubiquitin-protein ligase G2', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77622321, end_position_on_the_genomic_accession=77645990, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22214, CurrentID=0, Status='live', Symbol='Ube2h', Aliases='1500009C23Rik, AI181839, AI326965, AI462483, AW227540, E2-20K, Gid3, N28148, Ubc8', description='ubiquitin-conjugating enzyme E2H', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme H|E2 ubiquitin-conjugating enzyme H|UBCH2|ubiquitin carrier protein H|ubiquitin-conjugating enzyme E2-20K|ubiquitin-protein ligase H', map_location='6 12.52 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30211289, end_position_on_the_genomic_accession=30304539, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22215, CurrentID=0, Status='live', Symbol='Ube3a', Aliases='4732496B02, 5830462N02Rik, A130086L21Rik, Hpve6a, mKIAA4216', description='ubiquitin protein ligase E3A', other_designations='E6-AP ubiquitin protein ligase|oncogenic protein-associated protein E6-AP|ubiquitin conjugating enzyme E3A', map_location='7 33.95 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=59227814, end_position_on_the_genomic_accession=59311536, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  22217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22217, CurrentID=0, Status='live', Symbol='Usp12', Aliases='Ubh1', description='ubiquitin specific peptidase 12', other_designations='deubiquitinating enzyme 12|ubiquitin hydrolyzing enzyme 1|ubiquitin specific protease 12|ubiquitin thioesterase 12|ubiquitin thiolesterase 12|ubiquitin-hydrolyzing enzyme 1|ubiquitin-specific-processing protease 12', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146734812, end_position_on_the_genomic_accession=146794956, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22218, CurrentID=0, Status='live', Symbol='Sumo1', Aliases='GMP1, PIC1, SENTRIN, SMT3, SMT3H3, SMTP3, SUMO-1, Smt3C, Ubl1', description='small ubiquitin-like modifier 1', other_designations='SMT3 homolog 3|SMT3 suppressor of mif two 3 homolog 1|SUMO-1 related peptidase|ubiquitin-homology domain protein PIC1|ubiquitin-like 1|ubiquitin-like protein SMT3C', map_location='1 C2-C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59639434, end_position_on_the_genomic_accession=59670834, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22221, CurrentID=0, Status='live', Symbol='Ubp1', Aliases='Cp2b, LBP-1a, LBP-1b, NF2d9', description='upstream binding protein 1', other_designations='nuclear factor 2d9', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=113930934, end_position_on_the_genomic_accession=113977202, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  22222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22222, CurrentID=0, Status='live', Symbol='Ubr1', Aliases='AI504731', description='ubiquitin protein ligase E3 component n-recognin 1', other_designations='E3 alpha|N-recognin-1|ubiquitin-protein ligase E3-alpha-1|ubiquitin-protein ligase E3-alpha-I|ubiquitin-protein ligase e3 componen n-recognin', map_location='2 60.37 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120860272, end_position_on_the_genomic_accession=120970729, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  22223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22223, CurrentID=0, Status='live', Symbol='Uchl1', Aliases='AW822034, C88048, PGP 9.5, PGP9.5, R75593, UCH-L1, UCHL-1, gad', description='ubiquitin carboxy-terminal hydrolase L1', other_designations='gracile axonal dystrophy|neuron cytoplasmic protein 9.5|protein gene product 9.5|ubiquitin thioesterase L1', map_location='5 35.95 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=66676121, end_position_on_the_genomic_accession=66687234, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22224, CurrentID=0, Status='live', Symbol='Usp10', Aliases='2610014N07Rik, UBPO, Uchrp, mKIAA0190', description='ubiquitin specific peptidase 10', other_designations='deubiquitinating enzyme 10|ubiquintin c-terminal hydrolase related polypeptide|ubiquitin specific protease 10|ubiquitin thioesterase 10|ubiquitin thiolesterase 10|ubiquitin-specific peptidase 10|ubiquitin-specific-processing protease 10', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119910633, end_position_on_the_genomic_accession=119957560, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  22225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22225, CurrentID=0, Status='live', Symbol='Usp5', Aliases='AA407472, ISOT, ISOT-1, Ucht', description='ubiquitin specific peptidase 5 (isopeptidase T)', other_designations='deubiquitinating enzyme 5|isopeptidase T|ubiquitin carboxy terminal hydrolase T|ubiquitin specific protease 5|ubiquitin thioesterase 5|ubiquitin thiolesterase 5|ubiquitin-specific-processing protease 5', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124815019, end_position_on_the_genomic_accession=124829487, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  22226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22226, CurrentID=0, Status='live', Symbol='Ucn', Aliases='Mpv171, Ucn', description='urocortin', other_designations='', map_location='5 17.11 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31137989, end_position_on_the_genomic_accession=31138895, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22227, CurrentID=0, Status='live', Symbol='Ucp1', Aliases='AI385626, Slc25a7, Ucp', description='uncoupling protein 1 (mitochondrial, proton carrier)', other_designations='UCP 1|solute carrier family 25 member 7|thermogenin|uncoupling protein 1, mitochondrial|uncoupling protein, mitochondrial', map_location='8 39.65 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83290348, end_position_on_the_genomic_accession=83298456, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22228, CurrentID=0, Status='live', Symbol='Ucp2', Aliases='Slc25a8', description='uncoupling protein 2 (mitochondrial, proton carrier)', other_designations='UCP 2|UCPH|solute carrier family 25 member 8|uncoupling protein 2, mitochondrial', map_location='7 54.36 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100493358, end_position_on_the_genomic_accession=100499630, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22229, CurrentID=0, Status='live', Symbol='Ucp3', Aliases='AI645527, Slc25a9, UCP-3', description='uncoupling protein 3 (mitochondrial, proton carrier)', other_designations='UCP 3|solute carrier family 25 member 9|uncoupling protein 3, mitochondrial', map_location='7 54.36 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100472991, end_position_on_the_genomic_accession=100486432, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22230, CurrentID=0, Status='live', Symbol='Ufd1l', Aliases='Ufd1', description='ubiquitin fusion degradation 1 like', other_designations='UB fusion protein 1', map_location='16 B1-B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18811801, end_position_on_the_genomic_accession=18835261, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  546519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546519, CurrentID=0, Status='live', Symbol='Tmem235', Aliases='Cldn27, Gm12581, OTTMUSG00000007589', description='transmembrane protein 235', other_designations='claudin 27|claudin-27|transmembrane protein ENSP00000364084 homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117859199, end_position_on_the_genomic_accession=117865543, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22232, CurrentID=0, Status='live', Symbol='Slc35a2', Aliases='AI327289, Had-1, Had1, Sfc8, UGT, Ugalt', description='solute carrier family 35 (UDP-galactose transporter), member A2', other_designations='UDP-Gal-Tr|UDP-galactose transporter|mUGT1|solute carrier family (UDP-N-acetylglucosamine transporter), member 2|solute carrier family 35 (UDP-galactose transporter) member 2|solute carrier family 35 (UDP-galactose transporter), member 2|solute carrier family 35 member A2', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7884034, end_position_on_the_genomic_accession=7894492, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22234, CurrentID=0, Status='live', Symbol='Ugcg', Aliases='AU043821, C80537, Epcs21, GlcT-1l, Ugcg', description='UDP-glucose ceramide glucosyltransferase', other_designations='GCS|UDP-glucose:N-acylsphingosine D-glucosyltransferase|ectoplacental cone, invasive trophoblast giant cells, extraembryonic ectoderm and chorion sequence 21|glucosylceramide synthase', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59189204, end_position_on_the_genomic_accession=59222833, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22235, CurrentID=0, Status='live', Symbol='Ugdh', Aliases='Udpgdh', description='UDP-glucose dehydrogenase', other_designations='UDP-Glc dehydrogenase|UDP-GlcDH', map_location='5 33.67 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65413207, end_position_on_the_genomic_accession=65435842, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22236, CurrentID=0, Status='live', Symbol='Ugt1a2', Aliases='UDPGT 1-2, Ugt1', description='UDP glucuronosyltransferase 1 family, polypeptide A2', other_designations='UDP-glucuronosyltransferase 1 family, member 2|UDP-glucuronosyltransferase 1 family, polypeptide A2|UDP-glucuronosyltransferase 1A2|UDPGT|UGT1*0|UGT1-02|UGT1.2|bilirubin specific|bilirubin-specific UDPGT', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88200611, end_position_on_the_genomic_accession=88220002, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22238, CurrentID=0, Status='live', Symbol='Ugt2b5', Aliases='AI118071, M-1, Udpgt-3, Udpgt2b5, Ugt2b17', description='UDP glucuronosyltransferase 2 family, polypeptide B5', other_designations='UDP-glucuronosyltransferase 2 family, member 5|UDP-glucuronosyltransferase 2B5', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87124946, end_position_on_the_genomic_accession=87140340, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22239, CurrentID=0, Status='live', Symbol='Ugt8a', Aliases='AI850488, AW455908, Cgt, Ugt8, mCerGT', description='UDP galactosyltransferase 8A', other_designations='UDP-galactose-ceramide galactosyltransferase 8A|UDP-glucuronosyltransferase 8|ceramide UDP-galactosyltransferase|cerebroside synthase', map_location='3 E3-F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=125865322, end_position_on_the_genomic_accession=125938550, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22240, CurrentID=0, Status='live', Symbol='Dpysl3', Aliases='CRMP-4, DRP-3, TUC4, ULIP-1, Ulip, Ulip1', description='dihydropyrimidinase-like 3', other_designations='collapsin response mediator protein 4|unc-33-like phosphoprotein 1', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=43320979, end_position_on_the_genomic_accession=43438286, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  22241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22241, CurrentID=0, Status='live', Symbol='Ulk1', Aliases='AU041434, Unc51.1, mKIAA0722', description='unc-51 like kinase 1', other_designations='serine/threonine-protein kinase Unc51.1|unc-51-like kinase 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110784489, end_position_on_the_genomic_accession=110810118, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  22242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22242, CurrentID=0, Status='live', Symbol='Umod', Aliases='THP, Urehd1, urehr4', description='uromodulin', other_designations='Tamm-Horsfall urinary glycoprotein|uromucoid', map_location='7 F1-F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119462708, end_position_on_the_genomic_accession=119479262, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22245, CurrentID=0, Status='live', Symbol='Uck1', Aliases='URK1, Umpk', description='uridine-cytidine kinase 1', other_designations='UCK 1|cytidine monophosphokinase 1|uridine monophosphate kinase|uridine monophosphokinase 1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32255002, end_position_on_the_genomic_accession=32260246, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22247, CurrentID=0, Status='live', Symbol='Umps', Aliases='1700095D23Rik, AA408257, AL033308, BB164745', description='uridine monophosphate synthetase', other_designations='UMP synthase', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33954784, end_position_on_the_genomic_accession=33967302, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22248, CurrentID=0, Status='live', Symbol='Unc119', Aliases='D11Bhm52, HRG4, MRG4, Rg4, Rtg4h, Unc119', description='unc-119 homolog (C. elegans)', other_designations='retinal gene 4|retinal protein 4', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78343443, end_position_on_the_genomic_accession=78349164, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22249, CurrentID=0, Status='live', Symbol='Unc13b', Aliases='Munc13-1, Munc13-2, Unc13a, Unc13h1, Unc13h2', description='unc-13 homolog B (C. elegans)', other_designations='UNC-13 homolog 1|munc13|unc-13 homolog A|unc13 homolog 1', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43058862, end_position_on_the_genomic_accession=43264887, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  22253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22253, CurrentID=0, Status='live', Symbol='Unc5c', Aliases='B130051O18Rik, Unc5h3, rcm', description='unc-5 homolog C (C. elegans)', other_designations='UNC-5 homolog 3|rostral cerebellar malformation protein|unc5 homolog 3', map_location='3 65.57 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=141465564, end_position_on_the_genomic_accession=141834924, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  22255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22255, CurrentID=0, Status='live', Symbol='Uncx', Aliases='Chx44.1, Uncx', description='UNC homeobox', other_designations='Unc4.1 homeobox|homeobox protein Uncx4.1', map_location='5 78.75 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139543898, end_position_on_the_genomic_accession=139548179, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22256, CurrentID=0, Status='live', Symbol='Ung', Aliases='UNG1, UNG2', description='uracil DNA glycosylase', other_designations='UDG|uracil-DNA glycosylase nuclear isoform', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114130360, end_position_on_the_genomic_accession=114139321, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  546546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546546, CurrentID=0, Status='live', Symbol='Serpina3h', Aliases='6C28, Spi2', description='serine (or cysteine) peptidase inhibitor, clade A, member 3H', other_designations='alpha-1 antiproteinase, antitrypsin|serine (or cysteine) proteinase inhibitor, clade A, member 3H', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104247896, end_position_on_the_genomic_accession=104254405, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22259, CurrentID=0, Status='live', Symbol='Nr1h3', Aliases='AU018371, LXR, RLD1, Unr1', description='nuclear receptor subfamily 1, group H, member 3', other_designations='LXR alpha|liver X receptor alpha|ubiquitously-expressed nuclear receptor 1', map_location='2 50.52 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91184061, end_position_on_the_genomic_accession=91202834, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22260, CurrentID=0, Status='live', Symbol='Nr1h2', Aliases='AI194859, LXR, LXRB, LXRBSV, LXRbeta, NER1, OR-1, RIP15, UR, Unr, Unr2', description='nuclear receptor subfamily 1, group H, member 2', other_designations='LXR beta|liver X receptor beta|retinoid X receptor-interacting protein No.15|ubiquitously-expressed nuclear receptor 2', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44549616, end_position_on_the_genomic_accession=44553969, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22262, CurrentID=0, Status='live', Symbol='Uox', Aliases='AI663847', description='urate oxidase', other_designations='', map_location='3 72.09 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146597149, end_position_on_the_genomic_accession=146631483, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22264, CurrentID=0, Status='live', Symbol='Prap1', Aliases='Upa', description='proline-rich acidic protein 1', other_designations='pregnancy-specific uterine protein|uterine-specific proline-rich acidic protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140093396, end_position_on_the_genomic_accession=140097203, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22268, CurrentID=0, Status='live', Symbol='Upk1b', Aliases='AI413235, Tspan20, UPIb, Upk1', description='uroplakin 1B', other_designations='UP1b|uroplakin Ib', map_location='16 B5-C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38773184, end_position_on_the_genomic_accession=38800345, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22269, CurrentID=0, Status='live', Symbol='Upk2', Aliases='AW491716, UPII', description='uroplakin 2', other_designations='UP2|uroplakin II', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44452715, end_position_on_the_genomic_accession=44454767, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22270, CurrentID=0, Status='live', Symbol='Upk3a', Aliases='1110017C07Rik, UP3a, UPIII, Upk3', description='uroplakin 3A', other_designations='proplakin IIIa|uroplakin 3|uroplakin III', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85017141, end_position_on_the_genomic_accession=85022560, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22271, CurrentID=0, Status='live', Symbol='Upp1', Aliases='AI325217, UPase, UdRPase, Up, Upp', description='uridine phosphorylase 1', other_designations='UPase 1|urdPase 1', map_location='11 A1-2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=9118008, end_position_on_the_genomic_accession=9136170, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  22272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22272, CurrentID=0, Status='live', Symbol='Uqcrq', Aliases='1100001F06Rik, 1500040F11Rik, 5830407L17Rik, 9.5kDa, AA959903, QP-C, Qpc, Uqcrb, c1502', description='ubiquinol-cytochrome c reductase, complex III subunit VII', other_designations='complex III subunit 8|complex III subunit VII|complex III subunit VIII|low molecular mass ubiquinone-binding protein (Qpc-pending)|ubiquinol-cytochrome c reductase binding protein|ubiquinol-cytochrome c reductase complex 9.5 kDa protein|ubiquinol-cytochrome c reductase complex ubiquinone-binding protein QP-C', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53428948, end_position_on_the_genomic_accession=53430874, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22273, CurrentID=0, Status='live', Symbol='Uqcrc1', Aliases='1110032G10Rik', description='ubiquinol-cytochrome c reductase core protein 1', other_designations='complex III subunit 1|core protein I|ubiquinol-cytochrome-c reductase complex core protein 1', map_location='9 59.63 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108936648, end_position_on_the_genomic_accession=108949641, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  22275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22275, CurrentID=0, Status='live', Symbol='Urod', Aliases='AI323803, UPD, Uro-d', description='uroporphyrinogen decarboxylase', other_designations='', map_location='4 53.41 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116990212, end_position_on_the_genomic_accession=116994436, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22276, CurrentID=0, Status='live', Symbol='Uros', Aliases='AI415298, UROIIIS3, Uros', description='uroporphyrinogen III synthase', other_designations='URO-synthase|hydroxymethylbilane hydrolyase (cyclizing)|uroporphyrinogen-III cosynthase|uroporphyrinogen-III cosynthetase', map_location='7 77.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133686142, end_position_on_the_genomic_accession=133710519, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22278, CurrentID=0, Status='live', Symbol='Usf1', Aliases='bHLHb11', description='upstream transcription factor 1', other_designations='Upstream stimulatory factor 1 (Major late transcription factor 1)|major late transcription factor 1', map_location='1 H', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171411304, end_position_on_the_genomic_accession=171418759, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  22282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22282, CurrentID=0, Status='live', Symbol='Usf2', Aliases='Usf-2, bHLHb12', description='upstream transcription factor 2', other_designations='major late transcription factor 2', map_location='7 A2-B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30945248, end_position_on_the_genomic_accession=30957250, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22283, CurrentID=0, Status='live', Symbol='Ush2a', Aliases='A930011D15Rik, A930037M10Rik, Gm676, Gm983, Mush2a, Usherin', description='Usher syndrome 2A (autosomal recessive, mild)', other_designations='Usher syndrome 2A (autosomal recessive, mild) homolog|usher syndrome type IIa protein homolog|usher syndrome type-2A protein homolog', map_location='1 92.29 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=188262017, end_position_on_the_genomic_accession=188965497, orientation='plus', exon_count=73, OMIM='', no_hdr0=''),
  22284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22284, CurrentID=0, Status='live', Symbol='Usp9x', Aliases='5730589N07Rik, AA407302, AA407699, AL022658, AL022749, Dffrx, FAF-X, Fafl', description='ubiquitin specific peptidase 9, X chromosome', other_designations='deubiquitinating enzyme FAF-X|fat facets homolog|fat facets protein-related, X-linked|fats facets protein related, X|ubiquitin carboxyl-terminal hydrolase FAM|ubiquitin specific protease 9 SSSRF- isoform|ubiquitin specific protease 9, X chromosome|ubiquitin thioesterase FAF-X|ubiquitin thiolesterase FAF-X|ubiquitin-specific protease 9, X chromosome|ubiquitin-specific-processing protease FAF-X', map_location='X 7.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=13070997, end_position_on_the_genomic_accession=13173328, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  22286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22286, CurrentID=0, Status='live', Symbol='Utf1', Aliases='AI505934', description='undifferentiated embryonic cell transcription factor 1', other_designations='utf1', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139943856, end_position_on_the_genomic_accession=139945112, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  22287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22287, CurrentID=0, Status='live', Symbol='Scgb1a1', Aliases='CC10, CC16, CCSP, PCB-BP, UG, UGB, Utg', description='secretoglobin, family 1A, member 1 (uteroglobin)', other_designations='Blastokinin|CCPBP|PCB-binding protein|clara cell 17 kDa protein|clara cell phospholipid-binding protein|clara cell secretory protein|clara cells 10 kDa secretory protein|secretoglobin family 1A member 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=9083642, end_position_on_the_genomic_accession=9087956, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22288, CurrentID=0, Status='live', Symbol='Utrn', Aliases='AA589569, DRP, Dmdl', description='utrophin', other_designations='G-utrophin|utrophin short isoform, Up109', map_location='10 3.77 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=12382187, end_position_on_the_genomic_accession=12869685, orientation='minus', exon_count=80, OMIM='', no_hdr0=''),
  22289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22289, CurrentID=0, Status='live', Symbol='Kdm6a', Aliases='Utx', description='lysine (K)-specific demethylase 6A', other_designations='histone demethylase UTX|ubiquitously transcribed TPR protein on the X chromosome|ubiquitously transcribed X chromosome tetratricopeptide repeat protein|ubiquitously transcribed tetratricopeptide repeat gene, X chromosome', map_location='X 13.45 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=18162533, end_position_on_the_genomic_accession=18279791, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  22290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22290, CurrentID=0, Status='live', Symbol='Uty', Aliases='Hydb, mKIAA4057', description='ubiquitously transcribed tetratricopeptide repeat gene, Y chromosome', other_designations='male-specific histocompatibility antigen H-YDB|ubiquitously transcribed TPR protein on the Y chromosome|ubiquitously transcribed Y chromosome tetratricopeptide repeat protein', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=1085187, end_position_on_the_genomic_accession=1245754, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  22293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22293, CurrentID=0, Status='live', Symbol='Slc45a2', Aliases='Aim-1, Aim1, Dbr, Matp, Oca4, blanc-sale, bls, uw', description='solute carrier family 45, member 2', other_designations='B/AIM-1-like protein|dominant brown|melanoma antigen AIM1|membrane associated transporter protein|protein AIM-1|protein underwhite|underwhite', map_location='15 5.4 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10995693, end_position_on_the_genomic_accession=11029233, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22294, CurrentID=0, Status='live', Symbol='Uxt', Aliases='0910002B17Rik', description='ubiquitously expressed transcript', other_designations='ubiquitously expressed transcript protein', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20951665, end_position_on_the_genomic_accession=20962035, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22295, CurrentID=0, Status='live', Symbol='Cdh23', Aliases='4930542A03Rik, USH1D, ahl, ahl1, bob, bus, mdfw, nmf112, nmf181, nmf252, sals, v', description='cadherin 23 (otocadherin)', other_designations='age related hearing loss 1|bobby|bustling|modifier of deaf waddler|otocadherin|waltzer', map_location='10 30.11 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60302414, end_position_on_the_genomic_accession=60696515, orientation='minus', exon_count=72, OMIM='', no_hdr0=''),
  22296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22296, CurrentID=0, Status='live', Symbol='Vmn1r51', Aliases='V1R, V1r1, V1ra1, VN12, mV1R1', description='vomeronasal 1 receptor 51', other_designations='pheromone receptor 1|vomeronasal 1 receptor Vmn1r51|vomeronasal 1 receptor, A1|vomeronasal 1, receptor 1|vomeronasal organ family 1, receptor 1|vomeronasal receptor V1RA1|vomeronasal type 1 receptor A1|vomeronasal type-1 receptor A1|vomeronasal type-1 receptor A8', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90122643, end_position_on_the_genomic_accession=90130645, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22297, CurrentID=0, Status='live', Symbol='Vmn1r45', Aliases='V1RA9, V1r2, V1ra11, V1ra2, mV1R2', description='vomeronasal 1 receptor 45', other_designations='pheromone receptor 2|vomeronasal 1 receptor Vmn1r45|vomeronasal 1 receptor, A2|vomeronasal 1, receptor 2|vomeronasal organ family 1, receptor 2|vomeronasal receptor 1 A9|vomeronasal receptor V1RA11|vomeronasal type-1 receptor A2|vomeronasal type-1 receptor A9', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89931650, end_position_on_the_genomic_accession=89940507, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22305, CurrentID=0, Status='live', Symbol='Vmn2r37', Aliases='V2r14', description='vomeronasal 2, receptor 37', other_designations='vomeronasal organ family 2, receptor, 14|vomeronasal receptor Vmn2r37', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=9205549, end_position_on_the_genomic_accession=9223653, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22306, CurrentID=0, Status='live', Symbol='Vmn2r30', Aliases='ENSMUSG00000070847, V2r15', description='vomeronasal 2, receptor 30', other_designations='vomeronasal organ family 2, receptor, 15|vomeronasal receptor Vmn2r30', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7312274, end_position_on_the_genomic_accession=7337493, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22307, CurrentID=0, Status='live', Symbol='Vmn2r10', Aliases='V2r16, VR16', description='vomeronasal 2, receptor 10', other_designations='vomeronasal organ family 2, receptor, 16|vomeronasal receptor Vmn2r10', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108995539, end_position_on_the_genomic_accession=109006436, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22308, CurrentID=0, Status='live', Symbol='Vmn2r122', Aliases='V2r12, V2r2, Vmn2r-ps158', description='vomeronasal 2, receptor, 122', other_designations='vomeronasal 2, receptor, 12|vomeronasal 2, receptor, 2|vomeronasal organ family 2, receptor, 12|vomeronasal organ family 2, receptor, 2', map_location='4', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22310, CurrentID=0, Status='live', Symbol='Vmn2r42', Aliases='V2r4', description='vomeronasal 2, receptor 42', other_designations='vomeronasal 2, receptor, 4|vomeronasal organ family 2, receptor, 4|vomeronasal receptor Vmn2r42', map_location='7 5.8 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=8183265, end_position_on_the_genomic_accession=8200320, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22311, CurrentID=0, Status='live', Symbol='Vmn2r32', Aliases='V2r5', description='vomeronasal 2, receptor 32', other_designations='vomeronasal 2, receptor, 5|vomeronasal organ family 2, receptor, 5|vomeronasal receptor Vmn2r32', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7463969, end_position_on_the_genomic_accession=7479973, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22312, CurrentID=0, Status='live', Symbol='Vmn2r107', Aliases='V2r6, Vmn107', description='vomeronasal 2, receptor 107', other_designations='vomeronasal 2, receptor, 6|vomeronasal organ family 2, receptor, 6|vomeronasal receptor Vmn2r107', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20345425, end_position_on_the_genomic_accession=20375772, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22313, CurrentID=0, Status='live', Symbol='Vmn2r104', Aliases='V2r7', description='vomeronasal 2, receptor 104', other_designations='vomeronasal 2, receptor, 7|vomeronasal organ family 2, receptor, 7|vomeronasal receptor Vmn2r104', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20029425, end_position_on_the_genomic_accession=20048205, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22317, CurrentID=0, Status='live', Symbol='Vamp1', Aliases='Syb-1, Syb1, VAMP-1, lew', description='vesicle-associated membrane protein 1', other_designations='synaptobrevin-1', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125215581, end_position_on_the_genomic_accession=125222306, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22318, CurrentID=0, Status='live', Symbol='Vamp2', Aliases='Syb-2, Syb2, sybII', description='vesicle-associated membrane protein 2', other_designations='VAMP-2|synaptobrevin II|synaptobrevin-2|vesicle-associated membrane 2', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69088528, end_position_on_the_genomic_accession=69092381, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22319, CurrentID=0, Status='live', Symbol='Vamp3', Aliases='D130027G05Rik, VAMP-3, ceb', description='vesicle-associated membrane protein 3', other_designations='cellubrevin|synaptobrevin-3', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=151047305, end_position_on_the_genomic_accession=151057953, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22320, CurrentID=0, Status='live', Symbol='Vamp8', Aliases='AU041171, Edb, endobrevin', description='vesicle-associated membrane protein 8', other_designations='VAMP-8', map_location='6 32.27 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72385221, end_position_on_the_genomic_accession=72390667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22321, CurrentID=0, Status='live', Symbol='Vars', Aliases='Bat6, D17H6S56E, G7a2, Vars', description='valyl-tRNA synthetase', other_designations='protein G7a|valRS', map_location='17 18.54 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35000907, end_position_on_the_genomic_accession=35016329, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  22323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22323, CurrentID=0, Status='live', Symbol='Vasp', Aliases='', description='vasodilator-stimulated phosphoprotein', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19256929, end_position_on_the_genomic_accession=19271854, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22324, CurrentID=0, Status='live', Symbol='Vav1', Aliases='Vav, vav-T', description='vav 1 oncogene', other_designations='p95vav|vav oncogene', map_location='17 29.76 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57279079, end_position_on_the_genomic_accession=57329236, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  22325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22325, CurrentID=0, Status='live', Symbol='Vav2', Aliases='2810040F13Rik, AI847175, Vav-2', description='vav 2 oncogene', other_designations='', map_location='2 19.36 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27262104, end_position_on_the_genomic_accession=27427608, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  22326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22326, CurrentID=0, Status='live', Symbol='Vax1', Aliases='', description='ventral anterior homeobox 1', other_designations='ventral anterior homeobox containing gene 1', map_location='19 54.61 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=59164480, end_position_on_the_genomic_accession=59170029, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  331401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331401, CurrentID=0, Status='live', Symbol='Thoc2', Aliases='6330441O12Rik, D130005M13Rik, Gm1139, Gm1793, Tho2', description='THO complex 2', other_designations='transcription/export complex subunit 2', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=41794992, end_position_on_the_genomic_accession=41911924, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  22329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22329, CurrentID=0, Status='live', Symbol='Vcam1', Aliases='CD106, Vcam-1', description='vascular cell adhesion molecule 1', other_designations='V-CAM 1', map_location='3 50.17 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116110020, end_position_on_the_genomic_accession=116129688, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22330, CurrentID=0, Status='live', Symbol='Vcl', Aliases='9430097D22, AA571387, AI462105, AW545629', description='vinculin', other_designations='metavinculin', map_location='14 11.53 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20929378, end_position_on_the_genomic_accession=21033676, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  22333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22333, CurrentID=0, Status='live', Symbol='Vdac1', Aliases='AL033343, Vdac5, mVDAC1, mVDAC5', description='voltage-dependent anion channel 1', other_designations='outer mitochondrial membrane protein porin 1|plasmalemmal porin|voltage-dependent anion-selective channel protein 5', map_location='11 31.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52361115, end_position_on_the_genomic_accession=52389396, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22334, CurrentID=0, Status='live', Symbol='Vdac2', Aliases='Vdac6, mVDAC2, mVDAC6', description='voltage-dependent anion channel 2', other_designations='VDAC-2|VDAC-6|outer mitochondrial membrane protein porin 2|voltage-dependent anion-selective channel protein 6', map_location='14 11.91 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21831561, end_position_on_the_genomic_accession=21845879, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  22335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22335, CurrentID=0, Status='live', Symbol='Vdac3', Aliases='', description='voltage-dependent anion channel 3', other_designations='VDAC-3|channel protein VDAC3|mVDAC3|mitochondrial outer membrane protein|outer mitochondrial membrane protein porin 3', map_location='8 11.42 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22577075, end_position_on_the_genomic_accession=22593813, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22337, CurrentID=0, Status='live', Symbol='Vdr', Aliases='Nr1i1', description='vitamin D receptor', other_designations='1,25-dihydroxyvitamin D3 receptor|nuclear receptor subfamily 1 group I member 1', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97854427, end_position_on_the_genomic_accession=97908296, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22339, CurrentID=0, Status='live', Symbol='Vegfa', Aliases='Vegf, Vpf', description='vascular endothelial growth factor A', other_designations='vascular permeability factor', map_location='17 22.79 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46016993, end_position_on_the_genomic_accession=46032377, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22340, CurrentID=0, Status='live', Symbol='Vegfb', Aliases='VEGF-B, Vrf', description='vascular endothelial growth factor B', other_designations='VEGF-related factor', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6982472, end_position_on_the_genomic_accession=6987651, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22341, CurrentID=0, Status='live', Symbol='Vegfc', Aliases='AW228853, VEGF-C', description='vascular endothelial growth factor C', other_designations='VRP|flt4 ligand|flt4-L|vascular endothelial growth factor C isoform 129|vascular endothelial growth factor C isoform 184|vascular endothelial growth factor-related protein', map_location='8 B3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=54077532, end_position_on_the_genomic_accession=54186454, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22342, CurrentID=0, Status='live', Symbol='Lin7b', Aliases='LIN-7B, MALS-2, Veli2', description='lin-7 homolog B (C. elegans)', other_designations='lin 7 homolog b|mammalian lin-seven protein 2|veli-2|vertebrate homolog of C. elegans Lin-7 type 2|vertebrate lin-7 homolog 2', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45367891, end_position_on_the_genomic_accession=45370564, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22343, CurrentID=0, Status='live', Symbol='Lin7c', Aliases='9130007B12Rik, AI303698, AU019331, AW125731, D2Ertd520e, LIN-7-C, LIN-7C, MALS-3, Veli3', description='lin-7 homolog C (C. elegans)', other_designations='lin 7 homolog c|mLin7C|mammalian lin-seven protein 3|veli-3|vertebrate homolog of C. elegans Lin-7 type 3|vertebrate lin-7 homolog 3', map_location='2 56.65 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=109890878, end_position_on_the_genomic_accession=109900937, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22344, CurrentID=0, Status='live', Symbol='Vezf1', Aliases='AI848691, AW046909, db1', description='vascular endothelial zinc finger 1', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88068256, end_position_on_the_genomic_accession=88084729, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22346, CurrentID=0, Status='live', Symbol='Vhl', Aliases='Vhlh', description='von Hippel-Lindau tumor suppressor', other_designations='pVHL|von Hippel-Lindau syndrome homolog|von Hippel-Lindau syndrome protein homolog', map_location='6 52.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113624021, end_position_on_the_genomic_accession=113631633, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22348, CurrentID=0, Status='live', Symbol='Slc32a1', Aliases='R75019, VGAT, Viaat', description='solute carrier family 32 (GABA vesicular transporter), member 1', other_designations='GABA and glycine transporter|mVGAT|mVIAAT|solute carrier family 32 member 1|vasicular inhibitory amino acid transporter 10D|vesicular GABA transporter|vesicular GABA transporter a form', map_location='2 H1-3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158610758, end_position_on_the_genomic_accession=158615747, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  22349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22349, CurrentID=0, Status='live', Symbol='Vil1', Aliases='Vil', description='villin 1', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74409191, end_position_on_the_genomic_accession=74435560, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  22350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22350, CurrentID=0, Status='live', Symbol='Ezr', Aliases='AW146364, R75297, Vil2, p81', description='ezrin', other_designations='cytovillin|villin 2|villin-2', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6738131, end_position_on_the_genomic_accession=6782780, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22351, CurrentID=0, Status='live', Symbol='Vill', Aliases='Villp', description='villin-like', other_designations='EF-6|vill', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119052729, end_position_on_the_genomic_accession=119071525, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  22352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22352, CurrentID=0, Status='live', Symbol='Vim', Aliases='', description='vimentin', other_designations='', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=13574311, end_position_on_the_genomic_accession=13582826, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22353, CurrentID=0, Status='live', Symbol='Vip', Aliases='', description='vasoactive intestinal polypeptide', other_designations='PHI, peptide histidine isoleucine', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=5639193, end_position_on_the_genomic_accession=5647617, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22354, CurrentID=0, Status='live', Symbol='Vipr1', Aliases='AV071699, VIP-R1, VPAC1', description='vasoactive intestinal peptide receptor 1', other_designations='PACAP type II receptor|PACAP-R-2|PACAP-R2|VIP receptor subtype 1|VIP-R-1|pituitary adenylate cyclase-activating polypeptide type II receptor', map_location='9 72.5 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121642716, end_position_on_the_genomic_accession=121672954, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  22355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22355, CurrentID=0, Status='live', Symbol='Vipr2', Aliases='VPAC2, VPAC2R, Vip2', description='vasoactive intestinal peptide receptor 2', other_designations='PACAP type III receptor|PACAP-R-3|PACAP-R3|VIP receptor subtype 2|VIP-R-2|VPAC2 receptor|pituitary adenylate cyclase-activating polypeptide type III receptor', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=116077726, end_position_on_the_genomic_accession=116159604, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  546644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546644, CurrentID=0, Status='live', Symbol='Ly6g', Aliases='Gr-1, Gr1, Ly-6G', description='lymphocyte antigen 6 complex, locus G', other_designations='ly-6G.1', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75155261, end_position_on_the_genomic_accession=75159128, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22359, CurrentID=0, Status='live', Symbol='Vldlr', Aliases='AA408956, AI451093, AW047288', description='very low density lipoprotein receptor', other_designations='VLDL receptor|VLDL-R', map_location='19 21.77 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=27216489, end_position_on_the_genomic_accession=27254231, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  22360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22360, CurrentID=0, Status='live', Symbol='Nrsn1', Aliases='Neurensin-1, Neuro-p24, Vmp', description='neurensin 1', other_designations='nrsn1|vesicular membrane protein of 24 kDa|vesicular membrane protein p24', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=25252039, end_position_on_the_genomic_accession=25270070, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22361, CurrentID=0, Status='live', Symbol='Vnn1', Aliases='V-1', description='vanin 1', other_designations='pantetheine hydrolase|vanin-1|vascular non-inflammatory molecule 1', map_location='10 A1-B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23894688, end_position_on_the_genomic_accession=23905344, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22362, CurrentID=0, Status='live', Symbol='Vpreb1', Aliases='CD179a, Vpreb-1', description='pre-B lymphocyte gene 1', other_designations='', map_location='16 10.46 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16868401, end_position_on_the_genomic_accession=16869255, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22363, CurrentID=0, Status='live', Symbol='Vpreb2', Aliases='CD179b, VpreB1, Vpreb-2', description='pre-B lymphocyte gene 2', other_designations='Lambda 5', map_location='16 11.14 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17979452, end_position_on_the_genomic_accession=17981274, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22364, CurrentID=0, Status='live', Symbol='Vpreb3', Aliases='8HS-20, AI528709, Vpreb-3', description='pre-B lymphocyte gene 3', other_designations='', map_location='10 C', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75943025, end_position_on_the_genomic_accession=75949651, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22365, CurrentID=0, Status='live', Symbol='Vps45', Aliases='AI462172, AW554165, mVps45', description='vacuolar protein sorting 45 (yeast)', other_designations='vacuolar protein sorting protein 45', map_location='3 41.65 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95999832, end_position_on_the_genomic_accession=96058455, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  22367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22367, CurrentID=0, Status='live', Symbol='Vrk1', Aliases='51PK', description='vaccinia related kinase 1', other_designations='serine/threonine-protein kinase 51PK|vaccinia-related kinase 1', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=106010163, end_position_on_the_genomic_accession=106077421, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  22368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22368, CurrentID=0, Status='live', Symbol='Trpv2', Aliases='GRC, OTRPC2, VRL-1, Vrl1', description='transient receptor potential cation channel, subfamily V, member 2', other_designations='growth factor regulated channel|growth factor-regulated calcium channel|osm-9-like TRP channel 2|vanilloid receptor-like protein 1', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62574486, end_position_on_the_genomic_accession=62600305, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  22370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22370, CurrentID=0, Status='live', Symbol='Vtn', Aliases='AI256434, Vn', description='vitronectin', other_designations='S-protein|serum-spreading factor', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78499120, end_position_on_the_genomic_accession=78502325, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22371, CurrentID=0, Status='live', Symbol='Vwf', Aliases='6820430P06Rik, AI551257, B130011O06Rik, C630030D09, F8VWF, VWD', description='Von Willebrand factor homolog', other_designations='vwf', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125552935, end_position_on_the_genomic_accession=125686679, orientation='plus', exon_count=53, OMIM='', no_hdr0=''),
  22372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22372, CurrentID=0, Status='live', Symbol='wal', Aliases='', description='waved alopecia', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22373, CurrentID=0, Status='live', Symbol='Wap', Aliases='', description='whey acidic protein', other_designations='', map_location='11 4.61 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6635483, end_position_on_the_genomic_accession=6638649, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22375, CurrentID=0, Status='live', Symbol='Wars', Aliases='TrpRS, WRS', description='tryptophanyl-tRNA synthetase', other_designations='tryptophanyl-tRNA synthetase, cytoplasmic', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108860030, end_position_on_the_genomic_accession=108894174, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22376, CurrentID=0, Status='live', Symbol='Was', Aliases='U42471p, Was', description='Wiskott-Aldrich syndrome homolog (human)', other_designations='', map_location='X 3.65 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8081466, end_position_on_the_genomic_accession=8090491, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22377, CurrentID=0, Status='live', Symbol='Wbp1', Aliases='AI255152', description='WW domain binding protein 1', other_designations='WBP-1', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83119044, end_position_on_the_genomic_accession=83121663, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22378, CurrentID=0, Status='live', Symbol='Wbp2', Aliases='', description='WW domain binding protein 2', other_designations='WBP-2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116078573, end_position_on_the_genomic_accession=116086997, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22379, CurrentID=0, Status='live', Symbol='Fmnl3', Aliases='2700073B04Rik, FBP11, Wbp3, mKIAA2014', description='formin-like 3', other_designations='WW domain binding protein 3|formin-like 3 protein', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99317210, end_position_on_the_genomic_accession=99370503, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  22380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22380, CurrentID=0, Status='live', Symbol='Wbp4', Aliases='AW545037, BB101031, FBP21', description='WW domain binding protein 4', other_designations='WBP-4|WW domain-containing-binding protein 4|formin binding protein 21|formin-binding protein 21', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79459937, end_position_on_the_genomic_accession=79481268, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22381, CurrentID=0, Status='live', Symbol='Wbp5', Aliases='AL024233', description='WW domain binding protein 5', other_designations='WBP-5|schlafen 8-interacting protein', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136245080, end_position_on_the_genomic_accession=136247139, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22384, CurrentID=0, Status='live', Symbol='Eif4h', Aliases='AU018978, D5Ertd355e, E430026L18Rik, Ef4h, Wbscr1, Wscr1, eIF-4H', description='eukaryotic translation initiation factor 4H', other_designations='williams-Beuren syndrome chromosomal region 1 protein homolog', map_location='5 74.71 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134619876, end_position_on_the_genomic_accession=134639494, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22385, CurrentID=0, Status='live', Symbol='Baz1b', Aliases='C87820, WSTF, Wbscr9', description='bromodomain adjacent to zinc finger domain, 1B', other_designations='Williams-Beuren syndrome chromosome region 9 homolog|bromodomain adjacent to zinc finger domain protein 1B|bromodomain adjacent to zinc finger domain, 1B; Williams-Beuren syndrome chromosome region 9 homolog|williams syndrome transcription factor homolog|williams-Beuren syndrome chromosomal region 9 protein homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135187323, end_position_on_the_genomic_accession=135246129, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  22387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22387, CurrentID=0, Status='live', Symbol='wd', Aliases='', description='waddler', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22388, CurrentID=0, Status='live', Symbol='Wdr1', Aliases='Aip1, D5Wsu185e, rede', description='WD repeat domain 1', other_designations='WD40 repeat protein 1|actin-interacting protein 1', map_location='5 20.63 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38526813, end_position_on_the_genomic_accession=38561800, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  22389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22389, CurrentID=0, Status='live', Symbol='we', Aliases='', description='wellhaarig', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22390, CurrentID=0, Status='live', Symbol='Wee1', Aliases='Wee1A', description='WEE 1 homolog 1 (S. pombe)', other_designations='WEE1 tyrosine kinase|wee1A kinase', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110121700, end_position_on_the_genomic_accession=110143299, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  22393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22393, CurrentID=0, Status='live', Symbol='Wfs1', Aliases='AI481085, wolframin', description='Wolfram syndrome 1 homolog (human)', other_designations='Wolfram syndrome 1 protein homolog', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36966104, end_position_on_the_genomic_accession=36988982, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  546683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546683, CurrentID=0, Status='live', Symbol='Gm5966', Aliases='EG546683', description='predicted gene 5966', other_designations='', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17920258, end_position_on_the_genomic_accession=17921480, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  22401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22401, CurrentID=0, Status='live', Symbol='Zmat3', Aliases='Wig1', description='zinc finger matrin type 3', other_designations='p53-activated gene 608 protein|wild-type p53-induced gene 1 protein|zinc finger protein WIG-1|zmat3', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32334792, end_position_on_the_genomic_accession=32366008, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22402, CurrentID=0, Status='live', Symbol='Wisp1', Aliases='AW146261, CCN4, Elm1', description='WNT1 inducible signaling pathway protein 1', other_designations='CCN family member 4|ELM-1|WISP-1', map_location='15 29.3 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=66891325, end_position_on_the_genomic_accession=66923205, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22403, CurrentID=0, Status='live', Symbol='Wisp2', Aliases='Ccn5, Crgr4, Ctgfl, Rcop1', description='WNT1 inducible signaling pathway protein 2', other_designations='CCN family member 5|CTGF-L|WISP-2|connective tissue growth factor-like protein', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163820778, end_position_on_the_genomic_accession=163833146, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22404, CurrentID=0, Status='live', Symbol='Wiz', Aliases='', description='widely-interspaced zinc finger motifs', other_designations='widely-interspaced zinc finger-containing protein', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32354047, end_position_on_the_genomic_accession=32389439, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  22406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22406, CurrentID=0, Status='live', Symbol='Wld', Aliases='Wlds', description='wallerian degeneration', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22408, CurrentID=0, Status='live', Symbol='Wnt1', Aliases='Int-1, Wnt-1, sw, swaying', description='wingless-type MMTV integration site family, member 1', other_designations='proto-oncogene Int-1|proto-oncogene protein Wnt-1|wingless-related MMTV integration site 1', map_location='15 F1-F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98788074, end_position_on_the_genomic_accession=98793837, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22409, CurrentID=0, Status='live', Symbol='Wnt10a', Aliases='', description='wingless-type MMTV integration site family, member 10A', other_designations='wingless related MMTV integration site 10a', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74792019, end_position_on_the_genomic_accession=74804176, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22410, CurrentID=0, Status='live', Symbol='Wnt10b', Aliases='Wnt12', description='wingless-type MMTV integration site family, member 10B', other_designations='wingless related MMTV integration site 10b|wnt-12', map_location='15 54.65 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98770712, end_position_on_the_genomic_accession=98778211, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22411, CurrentID=0, Status='live', Symbol='Wnt11', Aliases='', description='wingless-type MMTV integration site family, member 11', other_designations='wingless-related MMTV integration site 11', map_location='7 53.9 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98834921, end_position_on_the_genomic_accession=98854747, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22412, CurrentID=0, Status='live', Symbol='Wnt9b', Aliases='Wnt14b, Wnt15, clf, clf1, wnt-14b, wnt-15', description='wingless-type MMTV integration site family, member 9B', other_designations='protein Wnt-14b|wingless-type MMTV integration site 15|wingless-type MMTV integration site 9B|wingless-type MMTV integration site family, member 15', map_location='11 67.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103727363, end_position_on_the_genomic_accession=103749821, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22413, CurrentID=0, Status='live', Symbol='Wnt2', Aliases='2610510E18Rik, Int1l1, Irp, Mirp, Wnt-2a, Wnt2', description='wingless-type MMTV integration site family, member 2', other_designations='INT-1-related protein|wingless-related MMTV integration site 2', map_location='6 8.08 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=17985474, end_position_on_the_genomic_accession=18032751, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22414, CurrentID=0, Status='live', Symbol='Wnt2b', Aliases='Wnt13', description='wingless-type MMTV integration site family, member 2B', other_designations='wingless related MMTV integration site 2b|wnt-13', map_location='3 45.88 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104944805, end_position_on_the_genomic_accession=104961709, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22415, CurrentID=0, Status='live', Symbol='Wnt3', Aliases='Int-4, Wnt-3', description='wingless-type MMTV integration site family, member 3', other_designations='proto-oncogene Int-4|proto-oncogene protein Wnt-3|wingless-related MMTV integration site 3', map_location='11 67.5 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103774175, end_position_on_the_genomic_accession=103818021, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22416, CurrentID=0, Status='live', Symbol='Wnt3a', Aliases='Wnt-3a, vt', description='wingless-type MMTV integration site family, member 3A', other_designations='vestigial tail|wingless-related MMTV integration site 3A', map_location='11 37.17 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59248036, end_position_on_the_genomic_accession=59290751, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22417, CurrentID=0, Status='live', Symbol='Wnt4', Aliases='Wnt-4', description='wingless-type MMTV integration site family, member 4', other_designations='signal molecule|wingless-related MMTV integration site 4', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137277635, end_position_on_the_genomic_accession=137299501, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22418, CurrentID=0, Status='live', Symbol='Wnt5a', Aliases='8030457G12Rik, Wnt-5a', description='wingless-type MMTV integration site family, member 5A', other_designations='wingless-related MMTV integration site 5A', map_location='14 16.8 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=28505473, end_position_on_the_genomic_accession=28527448, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22419, CurrentID=0, Status='live', Symbol='Wnt5b', Aliases='AW545702, Wnt-5b', description='wingless-type MMTV integration site family, member 5B', other_designations='wingless-related MMTV integration site 5B', map_location='6 56.86 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119432531, end_position_on_the_genomic_accession=119544824, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22420, CurrentID=0, Status='live', Symbol='Wnt6', Aliases='AA409270, Wnt-6', description='wingless-type MMTV integration site family, member 6', other_designations='wingless-related MMTV integration site 6', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74754344, end_position_on_the_genomic_accession=74785322, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22421, CurrentID=0, Status='live', Symbol='Wnt7a', Aliases='AI849442, Wnt-7a, px, tw', description='wingless-type MMTV integration site family, member 7A', other_designations='postaxial hemimelia|wingless-related MMTV integration site 7A', map_location='6 40.45 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91363983, end_position_on_the_genomic_accession=91411369, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22422, CurrentID=0, Status='live', Symbol='Wnt7b', Aliases='Wnt-7b', description='wingless-type MMTV integration site family, member 7B', other_designations='wingless-related MMTV integration site 7B', map_location='15 40.39 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85535437, end_position_on_the_genomic_accession=85581821, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22423, CurrentID=0, Status='live', Symbol='Wnt8b', Aliases='', description='wingless-type MMTV integration site family, member 8B', other_designations='wingless related MMTV integration site 8b', map_location='19 37.98 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44493472, end_position_on_the_genomic_accession=44514273, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22427, CurrentID=0, Status='live', Symbol='Wrn', Aliases='AI846146', description='Werner syndrome homolog (human)', other_designations='exonuclease WRN', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33234372, end_position_on_the_genomic_accession=33385545, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  22428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22428, CurrentID=0, Status='live', Symbol='Dctn6', Aliases='AU044699, WS-3, p27', description='dynactin 6', other_designations='RGD containing protein|dynactin subunit p27|p27 dynactin subunit|protein WS-3', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=34090420, end_position_on_the_genomic_accession=34108712, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22431, CurrentID=0, Status='live', Symbol='Wt1', Aliases='D630046I19Rik, Wt-1', description='Wilms tumor 1 homolog', other_designations='', map_location='2 E', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=105126529, end_position_on_the_genomic_accession=105173614, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  22433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22433, CurrentID=0, Status='live', Symbol='Xbp1', Aliases='D11Ertd39e, TREB-5, TREB5, XBP-1', description='X-box binding protein 1', other_designations='tax-responsive element-binding protein 5 homolog', map_location='11 3.61 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5520641, end_position_on_the_genomic_accession=5525993, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22436, CurrentID=0, Status='live', Symbol='Xdh', Aliases='XO, Xor, Xox-1, Xox1', description='xanthine dehydrogenase', other_designations='xanthine oxidase|xanthine oxidoreductase', map_location='17 45.25 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=73883895, end_position_on_the_genomic_accession=73950196, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  22437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22437, CurrentID=0, Status='live', Symbol='Xirp1', Aliases='AI415219, Cmya1', description='xin actin-binding repeat containing 1', other_designations='Xin alpha|cardiomyopathy associated 1|cardiomyopathy-associated protein 1|mXin alpha', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120013755, end_position_on_the_genomic_accession=120023598, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  546726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546726, CurrentID=0, Status='live', Symbol='Cyp26c1', Aliases='EG546726', description='cytochrome P450, family 26, subfamily c, polypeptide 1', other_designations='', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37685680, end_position_on_the_genomic_accession=37693307, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22439, CurrentID=0, Status='live', Symbol='Xk', Aliases='1810038K19Rik, XRG1h, Xk', description='X-linked Kx blood group', other_designations='Kell blood group precursor (McLeod phenotype) homolog|McLeod syndrome gene homolog|XK homolog|XK-related protein 1', map_location='X 4.22 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9272784, end_position_on_the_genomic_accession=9313245, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  546729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546729, CurrentID=0, Status='live', Symbol='Calhm1', Aliases='EG546729', description='calcium homeostasis modulator 1', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47141035, end_position_on_the_genomic_accession=47144174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22445, CurrentID=0, Status='live', Symbol='Xlr3a', Aliases='Xlr3', description='X-linked lymphocyte-regulated 3A', other_designations='X-linked B-cell surface antigen 3a|X-linked lymphocyte regulated 3', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73086293, end_position_on_the_genomic_accession=73097095, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22446, CurrentID=0, Status='live', Symbol='Xlr3c', Aliases='Xlr3b', description='X-linked lymphocyte-regulated 3C', other_designations='X-linked B-cell surface antigen 3B|X-linked lymphocyte-regulated 3B|XLR-related protein A12', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73254538, end_position_on_the_genomic_accession=73265571, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  546747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546747, CurrentID=0, Status='live', Symbol='Olfr220', Aliases='EG546747, MOR103-13P, MOR103-17-ps1, Olfr413-ps1, Olfr220', description='olfactory receptor 220', other_designations='GA_x6K02SYWY4V-595-239|GA_x6K02T2P20D-20731742-20730694|olfactory receptor 220, pseudogene 1|olfactory receptor 413, pseudogene 1|olfactory receptor MOR103-13P', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174448625, end_position_on_the_genomic_accession=174449602, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  546769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546769, CurrentID=0, Status='live', Symbol='Olfr1174-ps', Aliases='MOR174-18, OTTMUSG00000013944, Olfr1174', description='olfactory receptor 1174, pseudogene', other_designations='GA_x6K02T2Q125-49796506-49795568', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88310832, end_position_on_the_genomic_accession=88314908, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  677884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=677884, CurrentID=0, Status='live', Symbol='Pakap', Aliases='AF064781, Gm20459, Palm2-Akap2, Palm2Akap2, Palm2akap2', description='paralemmin A kinase anchor protein', other_designations='Palm2-Akap2 readthrough transcript', map_location='4', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22526, CurrentID=0, Status='live', Symbol='Gm4836', Aliases='', description='predicted gene 4836', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=28270710, end_position_on_the_genomic_accession=28293632, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22532, CurrentID=0, Status='live', Symbol='Xmv14', Aliases='Xmv-14', description='xenotropic murine leukemia virus 14', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22564, CurrentID=0, Status='live', Symbol='Xmv44', Aliases='Xmv-44', description='xenotropic murine leukemia virus 44', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22586, CurrentID=0, Status='live', Symbol='Xmv8', Aliases='Xmv-8', description='xenotropic murine leukemia virus 8', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22587, CurrentID=0, Status='live', Symbol='Xmv9', Aliases='Xmv-9', description='xenotropic murine leukemia virus 9', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  22589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22589, CurrentID=0, Status='live', Symbol='Atrx', Aliases='4833408C14Rik, AI447451, ATR2, DXHXS6677E, HP1-BP38, Hp1bp2, Hp1bp38, MRXS3, RAD54L, Rad54, XH2, Xnp, ZNF-HX', description='alpha thalassemia/mental retardation syndrome X-linked', other_designations='ATP-dependent helicase ATRX|HP1 alpha-interacting protein|X-linked nuclear protein|alpha thalassemia/mental retardation syndrome (X-linked)|alpha thalassemia/mental retardation syndrome X-linked homolog|heterochromatin protein 2', map_location='X 47.26 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=105797615, end_position_on_the_genomic_accession=105929408, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  22590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22590, CurrentID=0, Status='live', Symbol='Xpa', Aliases='AI573865c, Xpa', description='xeroderma pigmentosum, complementation group A', other_designations='xeroderma pigmentosum group A-complementing protein homolog', map_location='4 C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46174746, end_position_on_the_genomic_accession=46196349, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22591, CurrentID=0, Status='live', Symbol='Xpc', Aliases='', description='xeroderma pigmentosum, complementation group C', other_designations='p125|xeroderma pigmentosum group C-complementing protein homolog', map_location='6 D', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91489305, end_position_on_the_genomic_accession=91515888, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  22592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22592, CurrentID=0, Status='live', Symbol='Ercc5', Aliases='Xpg', description='excision repair cross-complementing rodent repair deficiency, complementation group 5', other_designations='DNA excision repair protein ERCC-5|xeroderma pigmentosum group G-complementing protein homolog|xeroderma pigmentosum, complementation group G', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44147578, end_position_on_the_genomic_accession=44181260, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  22594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22594, CurrentID=0, Status='live', Symbol='Xrcc1', Aliases='Xrcc-1', description='X-ray repair complementing defective repair in Chinese hamster cells 1', other_designations='X-ray repair cross-complementing protein 1', map_location='7 A3-B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24546300, end_position_on_the_genomic_accession=24573438, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  22596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22596, CurrentID=0, Status='live', Symbol='Xrcc5', Aliases='AI314015, Ku80, Ku86', description='X-ray repair complementing defective repair in Chinese hamster cells 5', other_designations='ATP-dependent DNA helicase 2 subunit 2|ATP-dependent DNA helicase II 80 kDa subunit|CTC box-binding factor 85 kDa subunit|CTC85|CTCBF|DNA repair protein XRCC5|Ku p80|ku autoantigen protein p86 homolog|nuclear factor IV', map_location='1 36.5 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72307417, end_position_on_the_genomic_accession=72394953, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  22598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22598, CurrentID=0, Status='live', Symbol='Slc6a18', Aliases='B0AT3, D630001K16Rik, Xt2, Xtrp2', description='solute carrier family 6 (neurotransmitter transporter), member 18', other_designations='X transporter protein 2|b(0)AT3|sodium and chloride dependent transporter Xtrp2|sodium- and chloride-dependent transporter XTRP2|solute carrier family 6 member 18|system B 0 neutral amino acid transporter AT3|system B(0) neutral amino acid transporter AT3', map_location='13 40.13 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73661750, end_position_on_the_genomic_accession=73678023, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22599, CurrentID=0, Status='live', Symbol='Slc6a20b', Aliases='AA986306, Sit1, Slc6a20, XT3, Xtrp3', description='solute carrier family 6 (neurotransmitter transporter), member 20B', other_designations='IMINO-K|X transporter protein 3|solute carrier family 6 (neurotransmitter transporter), member 20|solute carrier family 6 member 20B|solute carrier family 6, member 20B', map_location='9 74.19 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123593820, end_position_on_the_genomic_accession=123632565, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22601, CurrentID=0, Status='live', Symbol='Yap1', Aliases='AI325207, Yap, Yap65, Yki, Yorkie', description='yes-associated protein 1', other_designations='65 kDa Yes-associated protein|protein yorkie homolog|yes-associated protein YAP65 homolog|yes-associated protein, 65 kDa|yorkie homolog', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7931999, end_position_on_the_genomic_accession=8004596, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22602, CurrentID=0, Status='live', Symbol='YB10', Aliases='', description='Y-specific B10-derived sequences', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  546896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546896, CurrentID=0, Status='live', Symbol='Olfr455', Aliases='EG546896, MOR0-17_p, MOR0-2P, MORO-2-ps1, Olfr455', description='olfactory receptor 455', other_designations='GA_x6K02T2P3E9-5021913-5022918|olfactory receptor MOR0-2P', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42538067, end_position_on_the_genomic_accession=42539020, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22612, CurrentID=0, Status='live', Symbol='Yes1', Aliases='Yes, p61-Yes', description='YES proto-oncogene 1, Src family tyrosine kinase', other_designations='Yamaguchi sarcoma viral (v-yes) oncogene homolog 1|proto-oncogene c-Yes|proto-oncogene tyrosine-protein kinase Yes|viral oncogene yes homolog', map_location='5 17.33 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=32611171, end_position_on_the_genomic_accession=32687066, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  22619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22619, CurrentID=0, Status='live', Symbol='Siae', Aliases='LSE, Ysg2', description='sialic acid acetylesterase', other_designations='clone 165|sialic acid-specific 9-O-acetylesterase|yolk sac gene 2|yolk sac protein 2', map_location='9 20.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37555637, end_position_on_the_genomic_accession=37649054, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  22625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22625, CurrentID=0, Status='live', Symbol='Map3k19', Aliases='Ysk4', description='mitogen-activated protein kinase kinase kinase 19', other_designations='SPS1/STE20-related protein kinase YSK4|Yeast Sps1/Ste20-related kinase 4', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=127814962, end_position_on_the_genomic_accession=127855613, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  22626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22626, CurrentID=0, Status='live', Symbol='Slc23a3', Aliases='SVCT3, Yspl1', description='solute carrier family 23 (nucleobase transporters), member 3', other_designations='Na(+)/L-ascorbic acid transporter 3|YSPL-1|sodium-dependent vitamin C transporter 3|yolk sac permease-like molecule 1', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75125541, end_position_on_the_genomic_accession=75135223, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22627, CurrentID=0, Status='live', Symbol='Ywhae', Aliases='AU019196', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, epsilon polypeptide', other_designations='14-3-3 epsilon|14-3-3E|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activatioprotein, epsilon polypeptide', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75732887, end_position_on_the_genomic_accession=75765841, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22628, CurrentID=0, Status='live', Symbol='Ywhag', Aliases='14-3-3gamma, D7Bwg1348e', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, gamma polypeptide', other_designations='14-3-3 gamma|14-3-3 protein gamma subtype|3-monooxgenase/tryptophan 5-monooxgenase activation protein, gamma polypeptide|3-monooxgenase/tryptophan 5-monooxygenase activation protein, gamma polypeptide|3-monooxygenase/tryptophan 5-monooxygenase activation protein, gamma polypeptide', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135908379, end_position_on_the_genomic_accession=135934641, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  22629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22629, CurrentID=0, Status='live', Symbol='Ywhah', Aliases='', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, eta polypeptide', other_designations='14-3-3 eta', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33018816, end_position_on_the_genomic_accession=33027966, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  22630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22630, CurrentID=0, Status='live', Symbol='Ywhaq', Aliases='2700028P07Rik, AA409740, AU021156, R74690', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, theta polypeptide', other_designations='14-3-3 protein T-cell|14-3-3 protein tau|14-3-3 tau|14-3-3 theta', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=21390329, end_position_on_the_genomic_accession=21417436, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22631, CurrentID=0, Status='live', Symbol='Ywhaz', Aliases='1110013I11Rik, 14-3-3zeta, AI596267, AL022924, AU020854', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, zeta polypeptide', other_designations='14-3-3 zeta|KCIP-1|SEZ-2|protein kinase C inhibitor protein 1', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36770261, end_position_on_the_genomic_accession=36802984, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22632, CurrentID=0, Status='live', Symbol='Yy1', Aliases='AW488674, NF-E1', description='YY1 transcription factor', other_designations='UCR-motif DNA-binding protein|UCRBP transcription factor|YY-1|Yin Yang 1|delta transcription factor|yin and yang 1', map_location='12 59.58 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108793269, end_position_on_the_genomic_accession=108816632, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22634, CurrentID=0, Status='live', Symbol='Plagl1', Aliases='Lot1, Zac1', description='pleiomorphic adenoma gene-like 1', other_designations='zinc finger protein regulator of apoptosis and cell cycle arrest', map_location='10 4.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13090788, end_position_on_the_genomic_accession=13131694, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22635, CurrentID=0, Status='live', Symbol='Zan', Aliases='Zan-test', description='zonadhesin', other_designations='', map_location='5 76.4 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137378637, end_position_on_the_genomic_accession=137477064, orientation='minus', exon_count=87, OMIM='', no_hdr0=''),
  22637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22637, CurrentID=0, Status='live', Symbol='Zap70', Aliases='Srk, ZAP-70, mrtle, mur', description='zeta-chain (TCR) associated protein kinase', other_designations='70 kDa zeta-chain associated protein|syk-related tyrosine kinase', map_location='1 A4-C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36761798, end_position_on_the_genomic_accession=36782821, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  22639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22639, CurrentID=0, Status='live', Symbol='Zfa-ps', Aliases='Zfa, Zfx', description='zinc finger protein, autosomal, pseudogene', other_designations='zinc finger protein X-linked pseudogene|zinc finger protein ZFX', map_location='10 26.71 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52542320, end_position_on_the_genomic_accession=52545739, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22640, CurrentID=0, Status='live', Symbol='Zfp1', Aliases='Fnp-1, Zfp-1, mkr-1', description='zinc finger protein 1', other_designations='protein mKR1', map_location='8 57.98 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111643443, end_position_on_the_genomic_accession=111671108, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22642, CurrentID=0, Status='live', Symbol='Zbtb17', Aliases='AA589413, Lp-1, Miz1, Zfp100, mZ13', description='zinc finger and BTB domain containing 17', other_designations='polyomavirus late initiator promoter-binding protein|zfp-100|zinc finger protein 100|zinc finger protein 151|zinc finger protein Z13', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141444659, end_position_on_the_genomic_accession=141467937, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  22643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22643, CurrentID=0, Status='live', Symbol='Zfp101', Aliases='AI316498, AI325003', description='zinc finger protein 101', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33380174, end_position_on_the_genomic_accession=33394637, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22644, CurrentID=0, Status='live', Symbol='Rnf103', Aliases='AW146237, AW212918, Zfp103, kf-1', description='ring finger protein 103', other_designations='zinc finger protein 103', map_location='6 32.16 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71493894, end_position_on_the_genomic_accession=71510881, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22646, CurrentID=0, Status='live', Symbol='Zfp105', Aliases='AW557864', description='zinc finger protein 105', other_designations='', map_location='9 73.23 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122923040, end_position_on_the_genomic_accession=122931042, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22648, CurrentID=0, Status='live', Symbol='Zfp11', Aliases='6720465D24, Krox-6, Zfp-11', description='zinc finger protein 11', other_designations='protein Krox-6.1A/6.1B+/6.1B-|zinc finger protein Krox-6.1A/6.1B+/6.1B-', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129654595, end_position_on_the_genomic_accession=129670088, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22652, CurrentID=0, Status='live', Symbol='Mkrn3', Aliases='D7H15S9-1, Zfp127', description='makorin, ring finger protein, 3', other_designations='zinc finger protein 127', map_location='7 34.37 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=62417593, end_position_on_the_genomic_accession=62420139, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22654, CurrentID=0, Status='live', Symbol='Zfp13', Aliases='4933429B21, AI835008, Krox-8, Rhit, Zfp-13', description='zinc finger protein 13', other_designations='Kruppel-related protein|repressor of heat-inducible transcription|zinc finger protein Krox-8', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23575844, end_position_on_the_genomic_accession=23599516, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22658, CurrentID=0, Status='live', Symbol='Pcgf2', Aliases='Mel18, Rnf110, Zfp144, mel-18', description='polycomb group ring finger 2', other_designations='DNA-binding protein Mel-18|melanoma nuclear protein 18|ring finger protein 110|zfp-144|zinc finger protein 144', map_location='11 C1-C3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97688821, end_position_on_the_genomic_accession=97701344, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22661, CurrentID=0, Status='live', Symbol='Zfp148', Aliases='2210405J08Rik, AI480666, AW045217, BERF-1, BFCOL1, ZBP-89, Znf148', description='zinc finger protein 148', other_designations='G-rich box-binding protein|beta enolase repressor factor 1|transcription factor BFCOL1|transcription factor ZBP-89|zinc finger DNA-binding protein 89', map_location='16 23.93 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33380775, end_position_on_the_genomic_accession=33503903, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  22666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22666, CurrentID=0, Status='live', Symbol='Zbtb14', Aliases='ZF5, Zfp161, b2b1982Clo', description='zinc finger and BTB domain containing 14', other_designations='zfp-161|zinc finger protein 161|zinc finger protein 5', map_location='17 40.42 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=69383181, end_position_on_the_genomic_accession=69390544, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22668, CurrentID=0, Status='live', Symbol='Sf1', Aliases='BBP, MZFM, WBP4, Zfp162', description='splicing factor 1', other_designations='mammalian branch point-binding protein|transcription factor ZFM1|zinc finger gene in MEN1 locus|zinc finger protein 162', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6363690, end_position_on_the_genomic_accession=6378038, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  22670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22670, CurrentID=0, Status='live', Symbol='Trim26', Aliases='AI462198, Zfp173, Zfp1736', description='tripartite motif-containing 26', other_designations='tripartite motif protein 26|tripartite motif protein TRIM26|zinc finger protein 173', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36837124, end_position_on_the_genomic_accession=36859398, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  22671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22671, CurrentID=0, Status='live', Symbol='Rnf112', Aliases='ZNF179, Zfp179, bfp', description='ring finger protein 112', other_designations='brain finger protein|zinc finger protein 179', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61448417, end_position_on_the_genomic_accession=61454188, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  22673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22673, CurrentID=0, Status='live', Symbol='Zfp185', Aliases='P1, Znf185', description='zinc finger protein 185', other_designations='LIM domain protein Zfp185|P1-A', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72987253, end_position_on_the_genomic_accession=73031543, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  22678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22678, CurrentID=0, Status='live', Symbol='Zfp2', Aliases='9930007F06Rik, Fnp-2, Zfp-2, mKR2, mkr-2, mszf87', description='zinc finger protein 2', other_designations='', map_location='11 30.94 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50898712, end_position_on_the_genomic_accession=50916176, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22680, CurrentID=0, Status='live', Symbol='Zfp207', Aliases='8430401D15Rik, BuGZ, Zep, Znf207', description='zinc finger protein 207', other_designations='49 kDa zinc finger protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80383279, end_position_on_the_genomic_accession=80405992, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  22682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22682, CurrentID=0, Status='live', Symbol='Zfand5', Aliases='2310057A04Rik, 5830475F03Rik, AA415485, Za20d2, Zfp216', description='zinc finger, AN1-type domain 5', other_designations='zinc finger A20 domain-containing protein 2|zinc finger protein 216|zinc finger, A20 domain containing 2', map_location='19 14.32 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=21271279, end_position_on_the_genomic_accession=21286840, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22685, CurrentID=0, Status='live', Symbol='Zfp239', Aliases='Mok-2, Mok2, Znf239', description='zinc finger protein 239', other_designations='Kruppel zinc finger protein 2|zfp-239|zinc finger protein MOK-2', map_location='6 55.62 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=117862081, end_position_on_the_genomic_accession=117872766, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69317, CurrentID=0, Status='live', Symbol='Hmgb4', Aliases='1700001F22Rik, AI326135', description='high-mobility group box 4', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128260212, end_position_on_the_genomic_accession=128260895, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22688, CurrentID=0, Status='live', Symbol='Zfp26', Aliases='5033428C05Rik, KRAB15, Zfp-26, Zfp70, Zfp81-rs1, mKIAA4196, mkr-3, mszf14, mszf52', description='zinc finger protein 26', other_designations='protein mKR3|zinc finger protein 81, related sequence 1', map_location='9 7.55 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20428318, end_position_on_the_genomic_accession=20460169, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  22689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22689, CurrentID=0, Status='live', Symbol='Zfp27', Aliases='Zfp-27, mkr-4, mszf76', description='zinc finger protein 27', other_designations='protein mKR4', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29893337, end_position_on_the_genomic_accession=29906567, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22690, CurrentID=0, Status='live', Symbol='Zfp28', Aliases='2810438M17Rik, Zfp-28, mkr-5', description='zinc finger protein 28', other_designations='protein mKR5', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6383318, end_position_on_the_genomic_accession=6396637, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22691, CurrentID=0, Status='live', Symbol='Zscan2', Aliases='Zfp-29, Zfp29', description='zinc finger and SCAN domain containing 2', other_designations='zinc finger protein 29', map_location='7 45.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80860737, end_position_on_the_genomic_accession=80878128, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22693, CurrentID=0, Status='live', Symbol='Zfp30', Aliases='2610306P15Rik, Zfp-30', description='zinc finger protein 30', other_designations='', map_location='7 17.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29784790, end_position_on_the_genomic_accession=29794540, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22694, CurrentID=0, Status='live', Symbol='Zfp35', Aliases='Zfp-35, Znf271, mszf13, mszf20-1, mszf28, mszf68', description='zinc finger protein 35', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=23989634, end_position_on_the_genomic_accession=24005371, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  105504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105504, CurrentID=0, Status='live', Symbol='Exoc5', Aliases='AI447711, AI448003, Gm76, PRO1912, SEC10, Sec10l1', description='exocyst complex component 5', other_designations='SEC10-like 1|exocyst complex component Sec10', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49011922, end_position_on_the_genomic_accession=49066667, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  22696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22696, CurrentID=0, Status='live', Symbol='Zfp37', Aliases='Tzn, Zfp-37', description='zinc finger protein 37', other_designations='male germ cell-specific zinc finger protein|zinc finger protein 37 homolog', map_location='4 33.04 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62189537, end_position_on_the_genomic_accession=62208729, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22697, CurrentID=0, Status='live', Symbol='Zscan21', Aliases='AI326272, CTfin51, RU49, Zfp-38, Zfp38, Zipro1', description='zinc finger and SCAN domain containing 21', other_designations='transcription factor RU49|zinc finger proliferation 1|zinc finger protein 38', map_location='5 76.96 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138116899, end_position_on_the_genomic_accession=138134265, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22698, CurrentID=0, Status='live', Symbol='Zfp39', Aliases='CTfin33, Zfp-39', description='zinc finger protein 39', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58888153, end_position_on_the_genomic_accession=58904225, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22700, CurrentID=0, Status='live', Symbol='Zfp40', Aliases='CTfin12, NTfin12, Zfp-40', description='zinc finger protein 40', other_designations='', map_location='17 11.56 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23173869, end_position_on_the_genomic_accession=23193228, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  546989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546989, CurrentID=0, Status='live', Symbol='Olfr607', Aliases='EG546989, MOR33-3P', description='olfactory receptor 607', other_designations='GA_x6K02T2PBJ9-6182881-6181898|olfactory receptor MOR33-3P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103460223, end_position_on_the_genomic_accession=103461110, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  22702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22702, CurrentID=0, Status='live', Symbol='Zfp42', Aliases='Rex-1, Rex1, Zfp-42', description='zinc finger protein 42', other_designations='mREX-1|reduced expression protein 1', map_location='8 23.89 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=43295067, end_position_on_the_genomic_accession=43307009, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22704, CurrentID=0, Status='live', Symbol='Zfp46', Aliases='AI426962, Mlz-4, Zfp-46, Znf436', description='zinc finger protein 46', other_designations='zinc finger protein MLZ-4', map_location='4 68.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136284168, end_position_on_the_genomic_accession=136293942, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22709, CurrentID=0, Status='live', Symbol='Zfp51', Aliases='Zfp-51, zfec12', description='zinc finger protein 51', other_designations='', map_location='17 11.4 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21450352, end_position_on_the_genomic_accession=21465591, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22710, CurrentID=0, Status='live', Symbol='Zfp52', Aliases='KRAB11, Zfp-52, Zfp76, mszf23-2, mszf24, mszf34, mszf75, zfec29', description='zinc finger protein 52', other_designations='KRAB-zinc finger protein 76', map_location='17 11.4 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21516295, end_position_on_the_genomic_accession=21562601, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  22712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22712, CurrentID=0, Status='live', Symbol='Zfp54', Aliases='2810034D10Rik, KRAB10, Zfp-54, Zfp76, mszf83', description='zinc finger protein 54', other_designations='clone 18|zinc finger protein 76', map_location='17 11.4 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21423227, end_position_on_the_genomic_accession=21435640, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22715, CurrentID=0, Status='live', Symbol='Zfp57', Aliases='G19, Zfp-57', description='zinc finger protein 57', other_designations='', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37001163, end_position_on_the_genomic_accession=37010729, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22717, CurrentID=0, Status='live', Symbol='Zfp59', Aliases='Mfg-2, Mfg2', description='zinc finger protein 59', other_designations='zfp-59|zinc finger protein Mfg-2', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27838540, end_position_on_the_genomic_accession=27856793, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22718, CurrentID=0, Status='live', Symbol='Zfp60', Aliases='6330516O17Rik, AI426106, AI449354, Mfg-3, Mfg3', description='zinc finger protein 60', other_designations='zfp-60|zinc finger protein Mfg-3', map_location='7 15.97 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27731409, end_position_on_the_genomic_accession=27751689, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  22719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22719, CurrentID=0, Status='live', Symbol='Zfp61', Aliases='', description='zinc finger protein 61', other_designations='', map_location='7 10.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24291045, end_position_on_the_genomic_accession=24299549, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22720, CurrentID=0, Status='live', Symbol='Zfp62', Aliases='', description='zinc finger protein 62', other_designations='ZT3', map_location='11 29.23 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49202960, end_position_on_the_genomic_accession=49218816, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  22722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22722, CurrentID=0, Status='live', Symbol='Zfp64', Aliases='', description='zinc finger protein 64', other_designations='zfp-64', map_location='2 88.99 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168925361, end_position_on_the_genomic_accession=168955587, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22724, CurrentID=0, Status='live', Symbol='Zbtb7b', Aliases='Thpok, Zfp67, c-Krox', description='zinc finger and BTB domain containing 7B', other_designations='T helper-inducing POZ/Krueppel factor|T-helper-inducing POZ/Krueppel-like factor|Vzinc finger and BTB domain containing 7B|krueppel-related zinc finger protein cKrox|zfp-67|zinc finger protein 67|zinc finger protein Th-POK', map_location='3 A1-A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89377644, end_position_on_the_genomic_accession=89394814, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22746, CurrentID=0, Status='live', Symbol='Zfp85', Aliases='KRAB19, Rslcan2, Zfp71-rs1, Zfp85', description='zinc finger protein 85', other_designations='KRAB-zinc finger protein 71|regulator of sex-limitation candidate 2|zinc finger protein 85, related sequence 1', map_location='13 34.74 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67744464, end_position_on_the_genomic_accession=67755134, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22750, CurrentID=0, Status='live', Symbol='Zfp9', Aliases='1810048F22Rik, AU016736, Krox-4, Zfp-9, mszf79-2', description='zinc finger protein 9', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118461950, end_position_on_the_genomic_accession=118479527, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  22751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22751, CurrentID=0, Status='live', Symbol='Zfp90', Aliases='6430515L01Rik, KRAB17, NK10, Zfp64, Zfp83, mKIAA1954', description='zinc finger protein 90', other_designations='Nk10 expressed protein|zfp-90|zinc finger protein 83|zinc finger protein 90 homolog|zinc finger protein NK10', map_location='8 53.14 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106415284, end_position_on_the_genomic_accession=106425889, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22754, CurrentID=0, Status='live', Symbol='Zfp92', Aliases='', description='zinc finger protein 92', other_designations='zfp-92', map_location='X 37.31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73410978, end_position_on_the_genomic_accession=73429813, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  22755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22755, CurrentID=0, Status='live', Symbol='Zfp93', Aliases='Znf235', description='zinc finger protein 93', other_designations='zfp-93', map_location='7 10.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24270418, end_position_on_the_genomic_accession=24277794, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22756, CurrentID=0, Status='live', Symbol='Zfp94', Aliases='', description='zinc finger protein 94', other_designations='', map_location='7 10.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24299586, end_position_on_the_genomic_accession=24320262, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22757, CurrentID=0, Status='live', Symbol='Zkscan5', Aliases='AI132486, AI326970, Zfp95, hKraba1', description='zinc finger with KRAB and SCAN domains 5', other_designations='zfp-95|zinc finger protein 95', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145204463, end_position_on_the_genomic_accession=145221752, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  22758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22758, CurrentID=0, Status='live', Symbol='Zscan12', Aliases='2510038J07Rik, FPM315, Zfp96, mKIAA0426, zfp-96', description='zinc finger and SCAN domain containing 12', other_designations='zinc finger protein 96', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21362820, end_position_on_the_genomic_accession=21372302, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22759, CurrentID=0, Status='live', Symbol='Zfp97', Aliases='NGD16-4', description='zinc finger protein 97', other_designations='', map_location='17 A3.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17121383, end_position_on_the_genomic_accession=17146878, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  22761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22761, CurrentID=0, Status='live', Symbol='Zfpm1', Aliases='FOG, Fog1', description='zinc finger protein, multitype 1', other_designations='FOG-1|Friend of GATA 1|Friend of GATA-1|friend of GATA protein 1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122282141, end_position_on_the_genomic_accession=122337862, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  22762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22762, CurrentID=0, Status='live', Symbol='Zfpm2', Aliases='B330005D23Rik, FOG-2, FOG2', description='zinc finger protein, multitype 2', other_designations='Friend of GATA 2|friend of GATA protein 2|friend of GATA-1 2|friend of GATA-2|mFOG-2', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=40655042, end_position_on_the_genomic_accession=41104592, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22763, CurrentID=0, Status='live', Symbol='Zfr', Aliases='C920030H05Rik', description='zinc finger RNA binding protein', other_designations='', map_location='15 A', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=12117793, end_position_on_the_genomic_accession=12185449, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  22764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22764, CurrentID=0, Status='live', Symbol='Zfx', Aliases='Zfx55,6, Zfx6, Zfx', description='zinc finger protein X-linked', other_designations='X-linked zinc finger protein|zinc finger protein X-linked 5|zinc finger protein X-linked 6', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94074631, end_position_on_the_genomic_accession=94150643, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22767, CurrentID=0, Status='live', Symbol='Zfy1', Aliases='Zfy-1', description='zinc finger protein 1, Y-linked', other_designations='Y-linked zinc finger protein 1|zinc finger protein 1, Y linked', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=725207, end_position_on_the_genomic_accession=797409, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22768, CurrentID=0, Status='live', Symbol='Zfy2', Aliases='Zfy-2', description='zinc finger protein 2, Y-linked', other_designations='Y-linked zinc finger protein 2|zinc finger protein 2, Y linked', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=2106015, end_position_on_the_genomic_accession=2170409, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22770, CurrentID=0, Status='live', Symbol='Zhx1', Aliases='', description='zinc fingers and homeoboxes 1', other_designations='', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58047003, end_position_on_the_genomic_accession=58076508, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  22771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22771, CurrentID=0, Status='live', Symbol='Zic1', Aliases='ZIC, ZNF201', description='zinc finger protein of the cerebellum 1', other_designations='Zic family member 1 (odd-paired homolog, Drosophila)|odd-paired homolog', map_location='9 48.26 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=91358058, end_position_on_the_genomic_accession=91365826, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  22772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22772, CurrentID=0, Status='live', Symbol='Zic2', Aliases='HPE5, Ku', description='zinc finger protein of the cerebellum 2', other_designations='GENA 29|Zic family member 2 (odd-paired homolog, Drosophila)|kumba|odd-paired homolog', map_location='14 65.97 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122475384, end_position_on_the_genomic_accession=122480328, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22773, CurrentID=0, Status='live', Symbol='Zic3', Aliases='Bn, Ka', description='zinc finger protein of the cerebellum 3', other_designations='Zic family member 3 (odd-paired homolog, Drosophila)', map_location='X 32.56 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=58021511, end_position_on_the_genomic_accession=58041735, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22774, CurrentID=0, Status='live', Symbol='Zic4', Aliases='', description='zinc finger protein of the cerebellum 4', other_designations='Zic family member 4 (odd-paired homolog, Drosophila)', map_location='9 48.26 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=91361777, end_position_on_the_genomic_accession=91389349, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  22775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22775, CurrentID=0, Status='live', Symbol='Zik1', Aliases='', description='zinc finger protein interacting with K protein 1', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=10487222, end_position_on_the_genomic_accession=10495415, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  22776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22776, CurrentID=0, Status='live', Symbol='Zim1', Aliases='', description='zinc finger, imprinted 1', other_designations='', map_location='7 3.87 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6674444, end_position_on_the_genomic_accession=6696538, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  22778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22778, CurrentID=0, Status='live', Symbol='Ikzf1', Aliases='5832432G11Rik, Ikaros, LyF-1, Zfpn1a1, Znfn1a1, hlk-1, mKIAA4227', description='IKAROS family zinc finger 1', other_designations='ikaros family zinc finger protein 1|lymphoid transcription factor LyF-1|zinc finger protein subfamily 1A, 1|zinc finger protein, subfamily 1A, 1 (Ikaros)', map_location='11 7.02 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=11684969, end_position_on_the_genomic_accession=11772926, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  22779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22779, CurrentID=0, Status='live', Symbol='Ikzf2', Aliases='A730095J18Rik, Helios, Zfpn1a2, Znfn1a2', description='IKAROS family zinc finger 2', other_designations='ikaros family zinc finger protein 2|zinc finger protein, subfamily 1A, 2 (Helios)', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=69531207, end_position_on_the_genomic_accession=69687351, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22780, CurrentID=0, Status='live', Symbol='Ikzf3', Aliases='5830411O07Rik, Aiolos, Zfpn1a3, Znfn1a3', description='IKAROS family zinc finger 3', other_designations='ikaros family zinc finger protein 3|zinc finger protein, subfamily 1A, 3 (Aiolos)', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98464610, end_position_on_the_genomic_accession=98546380, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  22781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22781, CurrentID=0, Status='live', Symbol='Ikzf4', Aliases='A630026H08Rik, Eos, Zfpn1a4, Znfn1a4', description='IKAROS family zinc finger 4', other_designations='ikaros family zinc finger protein 4|zinc finger protein, subfamily 1A, 4 (Eos)', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128630839, end_position_on_the_genomic_accession=128669358, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22782, CurrentID=0, Status='live', Symbol='Slc30a1', Aliases='AI839647, C130040I11Rik, Znt1', description='solute carrier family 30 (zinc transporter), member 1', other_designations='solute carrier family 30 member 1|znT-1', map_location='1 97.01 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191906686, end_position_on_the_genomic_accession=191913247, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  22784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22784, CurrentID=0, Status='live', Symbol='Slc30a3', Aliases='Znt3', description='solute carrier family 30 (zinc transporter), member 3', other_designations='solute carrier family 30 member 3|znT-3', map_location='5 16.97 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31086106, end_position_on_the_genomic_accession=31108244, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  22785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22785, CurrentID=0, Status='live', Symbol='Slc30a4', Aliases='Znt4, lm, znT-4', description='solute carrier family 30 (zinc transporter), member 4', other_designations='lethal milk protein|solute carrier family 30 (zinc transporter), member 4; zinc transporter 4|solute carrier family 30 member 4', map_location='2 60.65 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122681233, end_position_on_the_genomic_accession=122702663, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  22786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22786, CurrentID=0, Status='live', Symbol='Zp1', Aliases='', description='zona pellucida glycoprotein 1', other_designations='zp-1', map_location='19 7.36 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10914296, end_position_on_the_genomic_accession=10920601, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  22787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22787, CurrentID=0, Status='live', Symbol='Zp2', Aliases='Zp-2', description='zona pellucida glycoprotein 2', other_designations='zona pellucida glycoprotein ZP2|zona pellucida protein A', map_location='7 64.38 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120126772, end_position_on_the_genomic_accession=120145754, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  22788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22788, CurrentID=0, Status='live', Symbol='Zp3', Aliases='Zp-3', description='zona pellucida glycoprotein 3', other_designations='sperm receptor|zona pellucida glycoprotein ZP3|zona pellucida protein C', map_location='5 75.62 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135980105, end_position_on_the_genomic_accession=135988624, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  22789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22789, CurrentID=0, Status='live', Symbol='Zp3r', Aliases='AV264399, SP56', description='zona pellucida 3 receptor', other_designations='sperm fertilization protein 56', map_location='1 56.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130544620, end_position_on_the_genomic_accession=130629634, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  22791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22791, CurrentID=0, Status='live', Symbol='Dnajc2', Aliases='AU020218, MIDA1, Zrf1, Zrf2', description='DnaJ heat shock protein family (Hsp40) member C2', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 2|Id associate 1|zuotin related factor 1|zuotin related factor 2|zuotin-related factor 1', map_location='5 9.97 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21757264, end_position_on_the_genomic_accession=21785248, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  22793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=22793, CurrentID=0, Status='live', Symbol='Zyx', Aliases='9530098H06Rik, R75157', description='zyxin', other_designations='', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42349621, end_position_on_the_genomic_accession=42360213, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  547097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=547097, CurrentID=0, Status='live', Symbol='B020011L13Rik', Aliases='', description='RIKEN cDNA B020011L13 gene', other_designations='', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=117781135, end_position_on_the_genomic_accession=117803359, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  547109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=547109, CurrentID=0, Status='live', Symbol='Trim43a', Aliases='EG547109, Gm6021', description='tripartite motif-containing 43A', other_designations='trim43a', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88580750, end_position_on_the_genomic_accession=88588823, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69349, CurrentID=0, Status='live', Symbol='1700008O03Rik', Aliases='', description='RIKEN cDNA 1700008O03 gene', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44358843, end_position_on_the_genomic_accession=44385692, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  547176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=547176, CurrentID=0, Status='live', Symbol='Zc3h12b', Aliases='A130028J20Rik, CXorf32, Gm373', description='zinc finger CCCH-type containing 12B', other_designations='', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=95711529, end_position_on_the_genomic_accession=95932656, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  547210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=547210, CurrentID=0, Status='live', Symbol='Gm6026', Aliases='H2al2y', description='predicted gene 6026', other_designations='', map_location='Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=2599099, end_position_on_the_genomic_accession=2599434, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  547253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=547253, CurrentID=0, Status='live', Symbol='Parp14', Aliases='1600029O10Rik, ARTD8, BC021340, CoaSt6, mKIAA1268', description='poly (ADP-ribose) polymerase family, member 14', other_designations='ADP-ribosyltransferase diphtheria toxin-like 8|PARP-14|collaborator of STAT6', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35832859, end_position_on_the_genomic_accession=35871514, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  69380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69380, CurrentID=0, Status='live', Symbol='1700013G24Rik', Aliases='AI507018', description='RIKEN cDNA 1700013G24 gene', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137453296, end_position_on_the_genomic_accession=137455461, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  331531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331531, CurrentID=0, Status='live', Symbol='AV320801', Aliases='4932408M16', description='expressed sequence AV320801', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135494552, end_position_on_the_genomic_accession=135504690, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  547428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=547428, CurrentID=0, Status='live', Symbol='Trav19', Aliases='Gm13893, MVA5T, OTTMUSG00000015025', description='T cell receptor alpha variable 19', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53845322, end_position_on_the_genomic_accession=53845827, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  547431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=547431, CurrentID=0, Status='live', Symbol='Btnl2', Aliases='BTL-II, BTLN2, Btl2, Gm315, NG9', description='butyrophilin-like 2', other_designations='butyrophylin-like MHC class II associated', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34354789, end_position_on_the_genomic_accession=34369492, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69479, CurrentID=0, Status='live', Symbol='1700029J07Rik', Aliases='', description='RIKEN cDNA 1700029J07 gene', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45939368, end_position_on_the_genomic_accession=45975454, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69481, CurrentID=0, Status='live', Symbol='Actl9', Aliases='1700029I08Rik, AV259599', description='actin-like 9', other_designations='actin-like 7A; actin-like 7-alpha', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33432899, end_position_on_the_genomic_accession=33434268, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  432879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432879, CurrentID=0, Status='live', Symbol='Kbtbd6', Aliases='EG432879, Gm5465, Zbtbd6', description='kelch repeat and BTB (POZ) domain containing 6', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79451835, end_position_on_the_genomic_accession=79454816, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  23789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23789, CurrentID=0, Status='live', Symbol='Coro1b', Aliases='', description='coronin, actin binding protein 1B', other_designations='coronin 1b|coronin 2|coronin-2', map_location='19 3.85 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4148612, end_position_on_the_genomic_accession=4154035, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23790, CurrentID=0, Status='live', Symbol='Coro1c', Aliases='AL022675, AW455561, AW548837', description='coronin, actin binding protein 1C', other_designations='coronin 1c|coronin 3|coronin-3', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113842439, end_position_on_the_genomic_accession=113908706, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  23792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23792, CurrentID=0, Status='live', Symbol='Adam23', Aliases='AW046396, MDC3', description='a disintegrin and metallopeptidase domain 23', other_designations='a disintegrin and metalloprotease domain 23|metalloproteinase-like, disintegrin-like, and cysteine-rich protein 3', map_location='1 32.31 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63445885, end_position_on_the_genomic_accession=63596515, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  23793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23793, CurrentID=0, Status='live', Symbol='Adam25', Aliases='', description='a disintegrin and metallopeptidase domain 25 (testase 2)', other_designations='ADAM 25|a disintegrin and metalloprotease domain 25 (testase 2)|testase 2|testase-2', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40752208, end_position_on_the_genomic_accession=40756176, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23794, CurrentID=0, Status='live', Symbol='Adamts5', Aliases='9530092O11Rik, ADAM-TS5, ADAMTS1, ADAMTS11, AI481094, ASMP-2', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 5 (aggrecanase-2)', other_designations='ADAM-TS 5|ADAMTS-5|ADMP-2|aggrecanase-2|implantin', map_location='16 48.34 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=85858157, end_position_on_the_genomic_accession=85902374, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23795, CurrentID=0, Status='live', Symbol='Agr2', Aliases='Agr2h, Gob-4, HAG-2, XAG-2, mAG-2', description='anterior gradient 2', other_designations='AG-2|anterior gradient 2 homolog|protein Gob-4|secreted cement gland protein XAG-2 homolog', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=35992925, end_position_on_the_genomic_accession=36004081, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  23796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23796, CurrentID=0, Status='live', Symbol='Aplnr', Aliases='APJ, Agtrl1, msr/apj', description='apelin receptor', other_designations='G-protein coupled receptor APJ|MSR|angiotensin receptor-like 1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85136360, end_position_on_the_genomic_accession=85139923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  23797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23797, CurrentID=0, Status='live', Symbol='Akt3', Aliases='AI851531, D930002M15Rik, Nmf350', description='thymoma viral proto-oncogene 3', other_designations='PKB gamma|RAC-PK-gamma|protein kinase Akt-3|protein kinase B gamma', map_location='1 H4-H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=177020076, end_position_on_the_genomic_accession=177258289, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  23801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23801, CurrentID=0, Status='live', Symbol='Aloxe3', Aliases='e-LOX-3, eLOX-3', description='arachidonate lipoxygenase 3', other_designations='epidermal LOX-3|epidermis-type lipoxygenase 3|hydroperoxy icosatetraenoate dehydratase', map_location='11 42.38 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69125674, end_position_on_the_genomic_accession=69149115, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  23802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23802, CurrentID=0, Status='live', Symbol='Amfr', Aliases='gp78', description='autocrine motility factor receptor', other_designations='AMF receptor', map_location='8 45.96 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93971588, end_position_on_the_genomic_accession=94012640, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  23805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23805, CurrentID=0, Status='live', Symbol='Apc2', Aliases='AI852447, APCL, R75424', description='adenomatosis polyposis coli 2', other_designations='', map_location='10 B5-C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80295957, end_position_on_the_genomic_accession=80318257, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  23806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23806, CurrentID=0, Status='live', Symbol='Arih1', Aliases='AU021774, Ari, Ari1, Hari, Hhari, Ubch7bp, Uip77', description='ariadne RBR E3 ubiquitin protein ligase 1', other_designations='ARI-1|UbcH7-binding protein|UbcM4-interacting protein 77|ariadne 1|ariadne ubiquitin-conjugating enzyme E2 binding protein homolog 1|protein ariadne-1 homolog|ubiquitin-conjugating enzyme E2-binding protein 1', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59388554, end_position_on_the_genomic_accession=59486660, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  23807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23807, CurrentID=0, Status='live', Symbol='Arih2', Aliases='AI843547, ARI2, TRIAD1, UIP48', description='ariadne RBR E3 ubiquitin protein ligase 2', other_designations='ARI-2|ariadne 2|ariadne homolog 2|protein ariadne-2 homolog|ubcM4-interacting protein 48', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108602942, end_position_on_the_genomic_accession=108654719, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  23808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23808, CurrentID=0, Status='live', Symbol='Ash2l', Aliases='', description='ash2 (absent, small, or homeotic)-like (Drosophila)', other_designations='ASH2-like protein', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25816000, end_position_on_the_genomic_accession=25847694, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  23821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23821, CurrentID=0, Status='live', Symbol='Bace1', Aliases='ASP2, Bace, C76936', description='beta-site APP cleaving enzyme 1', other_designations='APP beta-secretase|asp 2|aspartyl protease 2|beta-site amyloid precursor protein cleaving enzyme 1|memapsin-2|membrane-associated aspartic protease 2', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45838529, end_position_on_the_genomic_accession=45862484, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  23825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23825, CurrentID=0, Status='live', Symbol='Banf1', Aliases='Baf, Bcrp1, L2bp1/Baf', description='barrier to autointegration factor 1', other_designations='LAP2-binding protein 1|breakpoint cluster region protein 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5364633, end_position_on_the_genomic_accession=5366876, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23827, CurrentID=0, Status='live', Symbol='Bpnt1', Aliases='BPntase', description="bisphosphate 3'-nucleotidase 1", other_designations='PAP-inositol 1,4-phosphatase|PAP-inositol-1,4-phosphatase|PIP', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=185331965, end_position_on_the_genomic_accession=185357769, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  23828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23828, CurrentID=0, Status='live', Symbol='Bves', Aliases='Pop1, Popdc1, mBVES', description='blood vessel epicardial substance', other_designations='popeye 1|popeye domain-containing protein 1|popeye protein 1', map_location='10 23.47 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=45335762, end_position_on_the_genomic_accession=45372484, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  23829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23829, CurrentID=0, Status='live', Symbol='C1ql1', Aliases='C1qrf, CRF, CTRP14, gliacolin', description='complement component 1, q subcomponent-like 1', other_designations='C1q and tumor necrosis factor-related protein 14|C1q related factor|C1q/TNF-related protein 14', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102939264, end_position_on_the_genomic_accession=102946461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  23830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23830, CurrentID=0, Status='live', Symbol='Capn10', Aliases='AW049679, Capn8, mKIAA1845', description='calpain 10', other_designations='CANP 10|calcium-activated neutral proteinase 10|calpain 8', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92934408, end_position_on_the_genomic_accession=92947948, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23831, CurrentID=0, Status='live', Symbol='Car14', Aliases='AW536446, Ca14', description='carbonic anhydrase 14', other_designations='CA XIV|CA-XIV|carbonate dehydratase XIV|carbonic anhydrase XIV', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95897768, end_position_on_the_genomic_accession=95905118, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  23832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23832, CurrentID=0, Status='live', Symbol='Xcr1', Aliases='Ccxcr1, Gpr5, mXcr1', description='chemokine (C motif) receptor 1', other_designations='C motif-1/lymphotactin receptor|SCM1 receptor|XC chemokine receptor 1|chemokine (C motif) XC receptor 1|lymphotactin receptor', map_location='9 74.71 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123852315, end_position_on_the_genomic_accession=123862029, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  23833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23833, CurrentID=0, Status='live', Symbol='Cd52', Aliases='AI463198, B7, B7-Ag, CAMPATH-1, CLS1, MB7', description='CD52 antigen', other_designations='lymphocyte differentiation antigen B7', map_location='4 66.5 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134093538, end_position_on_the_genomic_accession=134095073, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  23834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23834, CurrentID=0, Status='live', Symbol='Cdc6', Aliases='CDC18L', description='cell division cycle 6', other_designations='CDC18(S.pombe)|CDC6-related protein|cell division cycle 6 homolog|p62(cdc6)', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98907702, end_position_on_the_genomic_accession=98923942, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  23836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23836, CurrentID=0, Status='live', Symbol='Cdh20', Aliases='Cdh7', description='cadherin 20', other_designations='cadherin 7|cadherin-7', map_location='1 49.42 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=104768819, end_position_on_the_genomic_accession=104995481, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23837, CurrentID=0, Status='live', Symbol='Cfdp1', Aliases='AA408409, Bcnt, Bucentaur, Cfdp, cp27', description='craniofacial development protein 1', other_designations='27 kDa craniofacial protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111768473, end_position_on_the_genomic_accession=111854310, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23844, CurrentID=0, Status='live', Symbol='Clca1', Aliases='Clca2, Clca3, gob-5, gob5', description='chloride channel accessory 1', other_designations='calcium-activated chloride channel family member 3|chloride channel calcium activated 1|chloride channel calcium activated 2|chloride channel calcium activated 3|mCLCA3|protein gob-5', map_location='3 69.26 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145004537, end_position_on_the_genomic_accession=145032776, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  23845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23845, CurrentID=0, Status='live', Symbol='Clec5a', Aliases='Clecsf5, Ly100, MDL-1, Mdl1', description='C-type lectin domain family 5, member a', other_designations='C-type (calcium dependent, carbohydrate-recognition domain) lectin, superfamily member 5|C-type lectin superfamily member 5|myeloid DAP12-associating lectin 1|myeloid DAP12-associating lectin-1|type II transmembrane protein MDL-1', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40574898, end_position_on_the_genomic_accession=40585805, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23849, CurrentID=0, Status='live', Symbol='Klf6', Aliases='AI448727, BCD1, C86813, CPBP, Copeb, FM2, FM6, Ierepo1, Ierepo3, R75280, Zf9', description='Kruppel-like factor 6', other_designations='core promoter element binding protein|core promoter element-binding protein|immediate early response, erythropoietin 1', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=5861489, end_position_on_the_genomic_accession=5870393, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  23850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23850, CurrentID=0, Status='live', Symbol='Pappa2', Aliases='PAPP-A2, PLAC3, Pappe', description='pappalysin 2', other_designations='placenta-specific 3|pregnancy-associated plasma preproprotein-A2|pregnancy-associated plasma protein-E', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=158711727, end_position_on_the_genomic_accession=158980523, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  23853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23853, CurrentID=0, Status='live', Symbol='Def6', Aliases='2410003F05Rik, 6430538D02Rik, AV094905, Ibp, Slat, Slat2, Slat6', description='differentially expressed in FDCP 6', other_designations='DEF-6|IRF-4-binding protein|IRF4-binding protein|SWAP-70-like adapter of T-cells', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28207778, end_position_on_the_genomic_accession=28228608, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23854, CurrentID=0, Status='live', Symbol='Def8', Aliases='AI449518, D8Ertd713e, DEF-8', description='differentially expressed in FDCP 8', other_designations='', map_location='8 72.1 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123442956, end_position_on_the_genomic_accession=123463899, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  23855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23855, CurrentID=0, Status='live', Symbol='Defa17', Aliases='AU014719, Cryp17, Defcr17', description='defensin, alpha, 17', other_designations='crypt defensin 17|cryptdin 17|defensin-related cryptdin-17', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21655767, end_position_on_the_genomic_accession=21656736, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23856, CurrentID=0, Status='live', Symbol='Dido1', Aliases='6720461J16Rik, D130048F08Rik, DATF-1, DIO-1, Datf1, dido', description='death inducer-obliterator 1', other_designations='death inducer-obliterator-2|death inducer-obliterator-3|death-associated transcription factor 1', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180657963, end_position_on_the_genomic_accession=180710035, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  23857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23857, CurrentID=0, Status='live', Symbol='Dmtf1', Aliases='Dimp, Dmp1', description='cyclin D binding myb-like transcription factor 1', other_designations='D-interacting myb-like protein|cyclin D binding myb-like transcription factor 1D-interacting myb-like protein|cyclin-D-interacting Myb-like protein 1|mDmp1', map_location='5 A2-A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=9118868, end_position_on_the_genomic_accession=9161776, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  23859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23859, CurrentID=0, Status='live', Symbol='Dlg2', Aliases='A330103J02Rik, B230218P12Rik, B330007M19Rik, Dlgh2, Gm1197, PSD93', description='discs, large homolog 2 (Drosophila)', other_designations='channel-associated protein of synapse-110|chapsyn-110|postsynaptic density protein PSD-93', map_location='7 51.07 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90476183, end_position_on_the_genomic_accession=92449246, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  23863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23863, CurrentID=0, Status='live', Symbol='Dand5', Aliases='Cerl-2, Cerr2, Dte, coco', description='DAN domain family member 5, BMP antagonist', other_designations='DAN domain family, member 5|cerberus-like protein 2|dante', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84815405, end_position_on_the_genomic_accession=84832265, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23871, CurrentID=0, Status='live', Symbol='Ets1', Aliases='AI196000, AI448617, D230050P06, Ets-1, Tpl1', description="E26 avian leukemia oncogene 1, 5' domain", other_designations='avian erythroblastosis virus E26 oncogene 1|p54|tumor progression locus 1', map_location='9 17.97 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=32636210, end_position_on_the_genomic_accession=32757820, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23872, CurrentID=0, Status='live', Symbol='Ets2', Aliases='AU022856, Ets-2', description="E26 avian leukemia oncogene 2, 3' domain", other_designations='', map_location='16 C3-qter', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=95702407, end_position_on_the_genomic_accession=95721049, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  23873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23873, CurrentID=0, Status='live', Symbol='Faim', Aliases='', description='Fas apoptotic inhibitory molecule', other_designations='', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98986373, end_position_on_the_genomic_accession=99002019, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  23874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23874, CurrentID=0, Status='live', Symbol='Farsb', Aliases='C76708, Farsa, Farsl, Farslb, Frsb', description='phenylalanyl-tRNA synthetase, beta subunit', other_designations='PheRS alpha|phenylalanine tRNA synthetase alpha|phenylalanine--tRNA ligase beta chain|phenylalanine-tRNA synthetase-like, beta subunit|phenylalanyl-tRNA synthetase beta chain', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=78417958, end_position_on_the_genomic_accession=78488897, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  23876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23876, CurrentID=0, Status='live', Symbol='Fbln5', Aliases='A55, DANCE, EVEC', description='fibulin 5', other_designations='FIBL-5|developmental arteries and neural crest EGF-like protein', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=101746565, end_position_on_the_genomic_accession=101819130, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  23877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23877, CurrentID=0, Status='live', Symbol='Fiz1', Aliases='AI790204, mFLJ00416', description='Flt3 interacting zinc finger protein 1', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5007056, end_position_on_the_genomic_accession=5014728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  23879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23879, CurrentID=0, Status='live', Symbol='Fxr2', Aliases='Fxr2h', description='fragile X mental retardation, autosomal homolog 2', other_designations='fragile X mental retardation gene 2, autosomal homolog|fragile X mental retardation-related protein 2', map_location='11 42.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69632971, end_position_on_the_genomic_accession=69653297, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  23880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23880, CurrentID=0, Status='live', Symbol='Fyb', Aliases='ADAP, B630013F22Rik, FYB-120/130, Slap130, p120/130, p120/p130', description='FYN binding protein', other_designations='FYN binding protein FYB-130|FYN-T-binding protein|SLP-76-associated phosphoprotein|adhesion and degranulation promoting adaptor protein', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=6522797, end_position_on_the_genomic_accession=6665608, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  23881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23881, CurrentID=0, Status='live', Symbol='G3bp2', Aliases='AA409541, E430034L04Rik, G3BP, mKIAA0660', description='GTPase activating protein (SH3 domain) binding protein 2', other_designations='Ras-GTPase-activating protein (GAP120) SH3-domain binding protein 2', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92052145, end_position_on_the_genomic_accession=92137735, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  23882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23882, CurrentID=0, Status='live', Symbol='Gadd45g', Aliases='AI327420, C86281, CR6, DDIT2, OIG37', description='growth arrest and DNA-damage-inducible 45 gamma', other_designations='cytokine-responsive protein CR6|growth arrest and DNA-damage-inducible protein GADD45 gamma|growth arrest and DNA-damage-inducible, gamma', map_location='13 A5-B', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51846675, end_position_on_the_genomic_accession=51848475, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  23885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23885, CurrentID=0, Status='live', Symbol='Gmcl1', Aliases='2810049L19Rik, Gcl, mglc-1', description='germ cell-less, spermatogenesis associated 1', other_designations='DIP|DP-interacting protein|germ cell-less homolog 1|mGcl-1', map_location='6 37.75 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86691768, end_position_on_the_genomic_accession=86733378, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  23886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23886, CurrentID=0, Status='live', Symbol='Gdf15', Aliases='MIC-1, NAG-1, SBF', description='growth differentiation factor 15', other_designations='GDF-15|macrophage inhibiting cytokine-1', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70629393, end_position_on_the_genomic_accession=70632469, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23887, CurrentID=0, Status='live', Symbol='Ggt5', Aliases='AI551243, GGL, GGT-REL, Ggtla1', description='gamma-glutamyltransferase 5', other_designations='GGT 5|gamma-glutamyl leukotrienase|gamma-glutamyltransferase-like activity 1|gamma-glutamyltranspeptidase 5|glutathione hydrolase 5|leukotriene-C4 hydrolase', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75589327, end_position_on_the_genomic_accession=75616968, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23888, CurrentID=0, Status='live', Symbol='Gpc6', Aliases='', description='glypican 6', other_designations='', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=116921001, end_position_on_the_genomic_accession=117979529, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23890, CurrentID=0, Status='live', Symbol='Gpr34', Aliases='Lypsr1', description='G protein-coupled receptor 34', other_designations='G-protein-coupled receptor', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=13623417, end_position_on_the_genomic_accession=13640859, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  23892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23892, CurrentID=0, Status='live', Symbol='Grem1', Aliases='Cktsf1b1, Drm, Grem, ld', description='gremlin 1, DAN family BMP antagonist', other_designations='cysteine knot superfamily 1, BMP antagonist 1|down-regulated in Mos-transformed cells protein|down-regulated in v-mos-transformed cells|gremlin 1', map_location='2 57.43 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=113748674, end_position_on_the_genomic_accession=113758788, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  23893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23893, CurrentID=0, Status='live', Symbol='Grem2', Aliases='Gremlin2, Prdc', description='gremlin 2, DAN family BMP antagonist', other_designations='Gremlin1|cysteine knot superfamily 1, BMP antagonist 2|gremlin 2 homolog, cysteine knot superfamily|protein related to DAN and cerberus', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174833785, end_position_on_the_genomic_accession=174921915, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23894, CurrentID=0, Status='live', Symbol='Gtf2h2', Aliases='44kDa, BTF2 p44, BTF2-p44, Btf2p44, p44', description='general transcription factor II H, polypeptide 2', other_designations='TFIIH basal transcription factor complex p44 subunit|basal transcription factor 2, p44 subunit|basic transcription factor 2 44 kDa subunit|general transcription factor II H, polypeptide 2 (44 kDa subunit)|general transcription factor IIH, polypeptide 2 (44 kDa subunit)', map_location='13 53.21 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100459786, end_position_on_the_genomic_accession=100492610, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  23897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23897, CurrentID=0, Status='live', Symbol='Hax1', Aliases='HAX-1, HSP1BP-1, Hs1bp1, SIG-111, Silg111, mHAX-1s', description='HCLS1 associated X-1', other_designations='HS1 binding protein|HS1-associated protein X-1|HS1-associating protein X-1', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89995446, end_position_on_the_genomic_accession=89998716, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  23900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23900, CurrentID=0, Status='live', Symbol='Hcst', Aliases='DAP10, KAP10', description='hematopoietic cell signal transducer', other_designations='DNAX-activation protein 10|membrane protein DAP10|transmembrane adapter protein KAP10', map_location='7 17.45 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30417712, end_position_on_the_genomic_accession=30419854, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23908, CurrentID=0, Status='live', Symbol='Hs2st1', Aliases='AW214369, Hs2st, mKIAA0448', description='heparan sulfate 2-O-sulfotransferase 1', other_designations='2-O-sulfotransferase|2-OST|2OST', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144431107, end_position_on_the_genomic_accession=144570216, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  23912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23912, CurrentID=0, Status='live', Symbol='Rhof', Aliases='AI845056, AV026554, Arhf, Ifld1, Rif', description='ras homolog gene family, member f', other_designations='induced in fatty liver dystrophy 1|ras homolog gene family, member f (in filopodia)', map_location='5 62.77 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123118180, end_position_on_the_genomic_accession=123132629, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  23917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23917, CurrentID=0, Status='live', Symbol='Impdh1', Aliases='B930086D20Rik, IMPDH-I', description="inosine 5'-phosphate dehydrogenase 1", other_designations='IMP dehydrogenase 1|IMPD 1|IMPDH 1', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29200437, end_position_on_the_genomic_accession=29216364, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  23918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23918, CurrentID=0, Status='live', Symbol='Impdh2', Aliases='IMPD', description="inosine 5'-phosphate dehydrogenase 2", other_designations='IMP dehydrogenase 2|IMP dehydrogenase type II|IMPD 2|IMPDH 2|IMPDH-II', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108560291, end_position_on_the_genomic_accession=108565584, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  23919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23919, CurrentID=0, Status='live', Symbol='Insl5', Aliases='RIF2', description='insulin-like 5', other_designations='relaxin/insulin-like factor 2|relaxin/insulin-like protein', map_location='4 C5-C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=103017872, end_position_on_the_genomic_accession=103026842, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  23920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23920, CurrentID=0, Status='live', Symbol='Insrr', Aliases='Irr', description='insulin receptor-related receptor', other_designations='IR-related receptor', map_location='3 38.7 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87796908, end_position_on_the_genomic_accession=87816101, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  23921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23921, CurrentID=0, Status='live', Symbol='Sh2b2', Aliases='Aps', description='SH2B adaptor protein 2', other_designations='SH2 and PH domain-containing adapter protein APS|adapter protein with pleckstrin homology and Src homology 2 domains|adaptor protein with pleckstrin homology and src homology 2 domains|insulin receptor substrate 5', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136218147, end_position_on_the_genomic_accession=136246612, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  23922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23922, CurrentID=0, Status='live', Symbol='Jtb', Aliases='Gm622', description='jumping translocation breakpoint', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90231597, end_position_on_the_genomic_accession=90235840, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23923, CurrentID=0, Status='live', Symbol='Aadat', Aliases='AI875679, Aadt, KATII, Kat2, mKat-2', description='aminoadipate aminotransferase', other_designations='2-aminoadipate aminotransferase|2-aminoadipate transaminase|KAT/AadAT|alpha-aminoadipate aminotransferase|kynurenine aminotransferase II|kynurenine--oxoglutarate aminotransferase II|kynurenine--oxoglutarate transaminase 2|kynurenine--oxoglutarate transaminase II', map_location='8 30.85 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=60505870, end_position_on_the_genomic_accession=60545677, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  23924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23924, CurrentID=0, Status='live', Symbol='Katna1', Aliases='', description='katanin p60 (ATPase-containing) subunit A1', other_designations='katanin p60 subunit A1|lipotransin|p60 katanin', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7726000, end_position_on_the_genomic_accession=7763150, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23925, CurrentID=0, Status='live', Symbol='Kel', Aliases='CD238, Ece3', description='Kell blood group', other_designations='Kel blood group protein|Kell blood group, metallo-endopeptidase|Kell protein', map_location='6 19.86 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41686329, end_position_on_the_genomic_accession=41704334, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  23927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23927, CurrentID=0, Status='live', Symbol='Krtap14', Aliases='AA589630, Pmg1, mKAP13', description='keratin associated protein 14', other_designations='anagen-specific protein KAP13|krtap14|pubertal mammary gland-specific protein 1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88825291, end_position_on_the_genomic_accession=88826145, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23928, CurrentID=0, Status='live', Symbol='Lamc3', Aliases='AI562206, AW240805', description='laminin gamma 3', other_designations='laminin-12 subunit gamma|laminin-14 subunit gamma|laminin-15 subunit gamma', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31887281, end_position_on_the_genomic_accession=31946535, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  23934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23934, CurrentID=0, Status='live', Symbol='Ly6h', Aliases='', description='lymphocyte antigen 6 complex, locus H', other_designations='ly-6H', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75564745, end_position_on_the_genomic_accession=75567249, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  23936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23936, CurrentID=0, Status='live', Symbol='Lynx1', Aliases='AI838844, SLURP-2', description='Ly6/neurotoxin 1', other_designations='GC26|secreted Ly-6/uPAR domain-containing protein 2|secreted Ly-6/uPAR-related protein 2', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74747856, end_position_on_the_genomic_accession=74752979, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23937, CurrentID=0, Status='live', Symbol='Mab21l2', Aliases='', description='mab-21-like 2 (C. elegans)', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=86545581, end_position_on_the_genomic_accession=86548283, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23938, CurrentID=0, Status='live', Symbol='Map2k5', Aliases='AI324775, AI428457, MEK5, Mapkk5, Prkmk5', description='mitogen-activated protein kinase kinase 5', other_designations='MAP kinase kinase 5|MAPK/ERK kinase 5|MAPK/ERK kinese 5|MAPKK 5|MEK 5|mitogen activated protein kinase kinase 5', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=63163768, end_position_on_the_genomic_accession=63378438, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  23939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23939, CurrentID=0, Status='live', Symbol='Mapk7', Aliases='BMK-1, BMK1, ERK-5, ERK5, Erk5-T, PRKM7', description='mitogen-activated protein kinase 7', other_designations='MAPK 7|big MAP kinase 1|extracellular signal regulated kinase 5|extracellular signal-regulated kinase 5|mitogen activated protein kinase 7', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61488812, end_position_on_the_genomic_accession=61494266, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23942, CurrentID=0, Status='live', Symbol='Mta2', Aliases='AW550797, Mata1l1, Mta1l1, mmta2', description='metastasis-associated gene family, member 2', other_designations='metastais-associated gene family, member 2|metastasis associated 1-like 1|metastasis-associated 1-like 1|metastasis-associated protein 2', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8941920, end_position_on_the_genomic_accession=8952300, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  23943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23943, CurrentID=0, Status='live', Symbol='Esyt1', Aliases='Fam62a, Mbc2, vp115', description='extended synaptotagmin-like protein 1', other_designations='E-Syt1|family with sequence similarity 62 (C2 domain containing), member A|membrane bound C2 domain containing protein|membrane-bound C2 domain-containing protein', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128510248, end_position_on_the_genomic_accession=128525859, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  23945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23945, CurrentID=0, Status='live', Symbol='Mgll', Aliases='AA589436, Magl, Mgl', description='monoglyceride lipase', other_designations='monoacylglycerol lipase', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88724412, end_position_on_the_genomic_accession=88828360, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23947, CurrentID=0, Status='live', Symbol='Mid2', Aliases='FXY2, Trim1', description='midline 2', other_designations='midline defect 2|midline-2|tripartite motif-containing protein 1', map_location='X 61.35 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140664303, end_position_on_the_genomic_accession=140767715, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23948, CurrentID=0, Status='live', Symbol='Mmp17', Aliases='MT4-MMP, MT4MMP, MTMMP4', description='matrix metallopeptidase 17', other_designations='membrane-type matrix metalloproteinase 4', map_location='5 68.18 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129584174, end_position_on_the_genomic_accession=129608211, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  23950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23950, CurrentID=0, Status='live', Symbol='Dnajb6', Aliases='Mrj, mDj4', description='DnaJ heat shock protein family (Hsp40) member B6', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 6|HSJ-2|heat shock protein J2|mammalian relative of DnaJ', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=29735687, end_position_on_the_genomic_accession=29786478, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  23954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23954, CurrentID=0, Status='live', Symbol='Nek3', Aliases='', description='NIMA (never in mitosis gene a)-related expressed kinase 3', other_designations='NIMA-related expressed kinase 3|never in mitosis A-related kinase 3|nimA-related protein kinase 3', map_location='8 11.07 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22128283, end_position_on_the_genomic_accession=22167004, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  23955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23955, CurrentID=0, Status='live', Symbol='Nek4', Aliases='Stk2', description='NIMA (never in mitosis gene a)-related expressed kinase 4', other_designations='serine/threonine-protein kinase 2', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30951382, end_position_on_the_genomic_accession=30988476, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  23956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23956, CurrentID=0, Status='live', Symbol='Neu2', Aliases='MBS, MSS, MTS', description='neuraminidase 2', other_designations='N-acetyl-alpha-neuraminidase 2|brain sialidase|cystolic sialidase|cytosolic sialidase|murine thymic sialidase|skeletal muscle sialidase', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87574027, end_position_on_the_genomic_accession=87597845, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  23957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23957, CurrentID=0, Status='live', Symbol='Nr0b2', Aliases='SHP, SHP-1, Shp1', description='nuclear receptor subfamily 0, group B, member 2', other_designations='orphan nuclear receptor SHP|small heterodimer partner', map_location='4 66.25 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133553390, end_position_on_the_genomic_accession=133556536, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  23958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23958, CurrentID=0, Status='live', Symbol='Nr2e3', Aliases='A930035N01Rik, PNR, RNR, rd7', description='nuclear receptor subfamily 2, group E, member 3', other_designations='retina-specific nuclear receptor|retinal degeneration 7', map_location='9 32.35 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59942771, end_position_on_the_genomic_accession=59960834, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  23959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23959, CurrentID=0, Status='live', Symbol='Nt5e', Aliases='2210401F01Rik, AI447961, CD73, NT, Nt5, eNT', description="5' nucleotidase, ecto", other_designations="5'-NT|ecto-5'-nucleotidase", map_location='9 E3.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88327609, end_position_on_the_genomic_accession=88372089, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  23960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23960, CurrentID=0, Status='live', Symbol='Oas1g', Aliases='AI449562, L2, Mmu-L, Mmu-L2, Oas1b, Oias-1, Oias1', description="2'-5' oligoadenylate synthetase 1G", other_designations="2'-5' oligoadenylate synthetase 1A|2-5-oligoadenylate synthetase 1A ((2-5)oligo(A) synthetase 1A) (2-5A synthetase 1A) (p42 OAS)", map_location='5 60.65 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120876142, end_position_on_the_genomic_accession=120887613, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23961, CurrentID=0, Status='live', Symbol='Oas1b', Aliases='Flv, L1, Mmu-L1, Oas1, Oias-2, Oias2, Wnv', description="2'-5' oligoadenylate synthetase 1B", other_designations="(2-5')oligo(A) synthase 1B|(2-5')oligo(A) synthetase 1b|2'-5'-oligoadenylate synthase-like protein 1|2'-5'-oligoadenylate synthetase-like protein 1|2-5A synthase 1B|2-5A synthetase 1B|Flavivirus resistance", map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120812635, end_position_on_the_genomic_accession=120824163, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  23962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23962, CurrentID=0, Status='live', Symbol='Oasl2', Aliases='M1204, Mmu-OASL, Oasl', description="2'-5' oligoadenylate synthetase-like 2", other_designations="54 kDa 2'-5'-oligoadenylate synthase-like protein 2|54 kDa 2'-5'-oligoadenylate synthetase-like protein 2|p54 OASL|p54OASL", map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114896934, end_position_on_the_genomic_accession=114912245, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  23963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23963, CurrentID=0, Status='live', Symbol='Tenm1', Aliases='Odz1, Odz2, TCAP-1, Ten-m1, ten-1', description='teneurin transmembrane protein 1', other_designations='protein Odd Oz/ten-m homolog 1|tenascin-M1', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=42527866, end_position_on_the_genomic_accession=43429156, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  23964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23964, CurrentID=0, Status='live', Symbol='Tenm2', Aliases='Odz2, Odz3, Ten-2, Ten-m2', description='teneurin transmembrane protein 2', other_designations='odd Oz/ten-m homolog 2|protein Odd Oz/ten-m homolog 2|tenascin-M2', map_location='11 21.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=36006656, end_position_on_the_genomic_accession=37236808, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  23965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23965, CurrentID=0, Status='live', Symbol='Tenm3', Aliases='2610100B16Rik, Odz1, Odz3, Ten-m3, mKIAA1455', description='teneurin transmembrane protein 3', other_designations='odd Oz/ten-m homolog 3|protein Odd Oz/ten-m homolog 3|ten-3|tenascin-M3', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=48225665, end_position_on_the_genomic_accession=48935162, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  23966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23966, CurrentID=0, Status='live', Symbol='Tenm4', Aliases='Doc4, ELM2, Odz4, R75022, Ten-m4, l(7)-3Rn, l7Rn3, mKIAA1302', description='teneurin transmembrane protein 4', other_designations='downstream of CHOP4|odd Oz/ten-m homolog 4|protein Odd Oz/ten-m homolog 4|ten-4|tenascin-M4', map_location='7 52.53 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=96171237, end_position_on_the_genomic_accession=96911093, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  23967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23967, CurrentID=0, Status='live', Symbol='Osr1', Aliases='Odd1, Osr', description='odd-skipped related 1 (Drosophila)', other_designations='', map_location='12 4.05 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=9574442, end_position_on_the_genomic_accession=9581500, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  23968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23968, CurrentID=0, Status='live', Symbol='Nlrp5', Aliases='Mater, Nalp5, Op1, PAN11', description='NLR family, pyrin domain containing 5', other_designations='NACHT, leucine rich repeat and PYD containing 5|maternal antigen that embryos require|ooplasm-specific protein 1', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23385897, end_position_on_the_genomic_accession=23441921, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  23969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23969, CurrentID=0, Status='live', Symbol='Pacsin1', Aliases='A830061D09Rik, H74, mKIAA1379, syndapin', description='protein kinase C and casein kinase substrate in neurons 1', other_designations='Syndapin I|syndapin 1|syndapin-1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27655591, end_position_on_the_genomic_accession=27711118, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  23970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23970, CurrentID=0, Status='live', Symbol='Pacsin2', Aliases='AI197433', description='protein kinase C and casein kinase substrate in neurons 2', other_designations='Syndapin II|syndapin 2|syndapin-2|syndapin-II', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83375607, end_position_on_the_genomic_accession=83464606, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  23971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23971, CurrentID=0, Status='live', Symbol='Papss1', Aliases='AI325286, Asapk, SK1', description="3'-phosphoadenosine 5'-phosphosulfate synthase 1", other_designations='PAPS synthase 1|PAPSS 1|SK 1|sulfurylase kinase 1', map_location='3 61.05 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=131564768, end_position_on_the_genomic_accession=131643671, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  23972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23972, CurrentID=0, Status='live', Symbol='Papss2', Aliases='1810018P12Rik, AI159688, AtpsU2, Atpsk2, Sk2, bm', description="3'-phosphoadenosine 5'-phosphosulfate synthase 2", other_designations='PAPS synthase 2|PAPSS 2|SK 2|brachymorphic|sulfurylase kinase 2', map_location='19 27.46 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=32595642, end_position_on_the_genomic_accession=32667187, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  23980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23980, CurrentID=0, Status='live', Symbol='Pebp1', Aliases='HCNP, Pbp, Pbp1, Pbpr, Rkip', description='phosphatidylethanolamine binding protein 1', other_designations='HCNPpp|PEBP-1|Raf-1 inhibitor protein|Raf-1 kinase inhibitor protein', map_location='5 56.88 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117282651, end_position_on_the_genomic_accession=117287564, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  23983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23983, CurrentID=0, Status='live', Symbol='Pcbp1', Aliases='WBP17, [a]CP-1, alphaCP-1', description='poly(rC) binding protein 1', other_designations='alpha-CP1|heterogeneous nuclear ribonucleoprotein E1|hnRNP E1|hnRNP-E1', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86524494, end_position_on_the_genomic_accession=86526171, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  23984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23984, CurrentID=0, Status='live', Symbol='Pde10a', Aliases='', description='phosphodiesterase 10A', other_designations='PDE10A2 brain-specific isoform|dual-substrate cAMP and cGMP, cyclic nucleotide phosphodiesterase', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8525187, end_position_on_the_genomic_accession=8986648, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  23985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23985, CurrentID=0, Status='live', Symbol='Slc26a4', Aliases='Pds, pendrin', description='solute carrier family 26, member 4', other_designations="Pendred syndrome homolog|Pendred's syndrome|sodium-independent chloride/iodide transporter", map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31519814, end_position_on_the_genomic_accession=31560637, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  23986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23986, CurrentID=0, Status='live', Symbol='Eci2', Aliases='Acbd2, Drs1, Hca88, Peci', description='enoyl-Coenzyme A delta isomerase 2', other_designations='D3,D2-enoyl-CoA isomerase|delta(3),delta(2)-enoyl-CoA isomerase|dodecenoyl-CoA isomerase|peroxisomal 3,2-trans-enoyl-CoA isomerase|peroxisomal delta3, delta2-enoyl-Coenzyme A isomerase', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34977748, end_position_on_the_genomic_accession=35027089, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  23988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23988, CurrentID=0, Status='live', Symbol='Pin1', Aliases='0610025L01Rik, D9Bwg1161e', description='protein (peptidyl-prolyl cis/trans isomerase) NIMA-interacting 1', other_designations='PPIase Pin1|peptidyl-prolyl cis-trans isomerase Pin1|rotamase Pin1', map_location='9 7.6 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20652130, end_position_on_the_genomic_accession=20666584, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  23989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23989, CurrentID=0, Status='live', Symbol='Med24', Aliases='911GSE, AU040102, AW547152, D11Ertd307e, DRIP100, Gse2, Pparb2, Pparbp2, R75526, Thrap4, Trap100', description='mediator complex subunit 24', other_designations='genetic suppressor element 2|mTRAP100|peroxisome proliferation activation receptor binding protein 2|thyroid hormone receptor associated protein 4|thyroid hormone receptor-associated protein 100 kDa|thyroid hormone receptor-associated protein 4|thyroid hormone receptor-associated protein complex 100 kDa component', map_location='11 62.46 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98704591, end_position_on_the_genomic_accession=98729475, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  23991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23991, CurrentID=0, Status='live', Symbol='Cib1', Aliases='Cibkip, Kip, Prkdcip', description='calcium and integrin binding 1 (calmyrin)', other_designations='CIB|DNA-PKcs-interacting protein|calcium and integrin binding, protein kinase interacting protein|calmyrin|kinase-interacting protein|protein kinase, DNA activated, catalytic polypeptide interacting protein', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80227156, end_position_on_the_genomic_accession=80232805, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  23992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23992, CurrentID=0, Status='live', Symbol='Prkra', Aliases='AV120107, PRK, Pact, RAX, lear', description='protein kinase, interferon inducible double stranded RNA dependent activator', other_designations='PKR-associated protein X|PKR-associating protein X|interferon-inducible double stranded RNA-dependent protein kinase activator A|protein activator of the interferon-induced protein kinase|protein kinase, interferon-inducible double stranded RNA-dependent activator|protein kinase, interferon-inducible double-stranded RNA-dependent activator', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76609215, end_position_on_the_genomic_accession=76648199, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  23993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23993, CurrentID=0, Status='live', Symbol='Klk7', Aliases='Prss6, SCCE', description='kallikrein related-peptidase 7 (chymotryptic, stratum corneum)', other_designations='kallikrein 7 (chymotryptic, stratum corneum)|protease, serine, 6|serine protease 6|stratum corneum chymotryptic enzyme|thymopsin', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43811444, end_position_on_the_genomic_accession=43816359, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  23994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23994, CurrentID=0, Status='live', Symbol='Dazap2', Aliases='AI314727, Brbp, Gcap28, Prtb, gt6-12, mKIAA0058', description='DAZ associated protein 2', other_designations='deleted in azoospermia-associated protein 2|granule cell antiserum positive 28|proline rich protein expressed in brain|proline-rich protein expressed in brain', map_location='15 56.36 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100615662, end_position_on_the_genomic_accession=100620761, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  23996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23996, CurrentID=0, Status='live', Symbol='Psmc4', Aliases='CIP21, MIP224', description='proteasome (prosome, macropain) 26S subunit, ATPase, 4', other_designations='26S proteasome AAA-ATPase subunit RPT3|CAR interacing protein 21|CAR interacting protein 21|MB67-interacting protein|TAT-binding protein 7|TBP-7|proteasome 26S ATPase subunit 4|proteasome 26S subunit ATPase 4', map_location='7 16.11 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28041702, end_position_on_the_genomic_accession=28050092, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  23997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23997, CurrentID=0, Status='live', Symbol='Psmd13', Aliases='S11', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 13', other_designations='26S proteasome regulatory subunit RPN9|26S proteasome regulatory subunit S11|26S proteasome regulatory subunit p40.5|26S proteasome subunit p40.5', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140882394, end_position_on_the_genomic_accession=140898642, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  23999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=23999, CurrentID=0, Status='live', Symbol='Twf2', Aliases='A6-related, AU014993, Ptk9l, Ptk9r', description='twinfilin, actin-binding protein, homolog 2 (Drosophila)', other_designations='PTK9 protein typrotein tyrosine kinase 9-like (A6-related protein)|mA6RP|protein tyrosine kinase 9 related protein|protein tyrosine kinase 9-like (A6-related protein)|twinfilin-1-like protein', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106203108, end_position_on_the_genomic_accession=106215387, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  24000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24000, CurrentID=0, Status='live', Symbol='Ptpn21', Aliases='PTPD1, PTPRL10', description='protein tyrosine phosphatase, non-receptor type 21', other_designations='protein-tyrosine phosphatase PTP-RL10', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98676741, end_position_on_the_genomic_accession=98737618, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  24001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24001, CurrentID=0, Status='live', Symbol='Tiam2', Aliases='3000002F19Rik, STEF, mKIAA2016', description='T cell lymphoma invasion and metastasis 2', other_designations='RAS-related C3 botulinum substrate 1, guanine nucleotide exchange factor 1|SIF and TIAM1-like exchange factor|T-cell lymphoma invasion and metastasis 2|T-cell lymphoma invasion and metastasis 2; RAS-related C3 botulinum substrate 1, guanine nucleotide exchange factor 1|TIAM-2', map_location='17 1.99 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=3326300, end_position_on_the_genomic_accession=3519397, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  24004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24004, CurrentID=0, Status='live', Symbol='Rai2', Aliases='3f8', description='retinoic acid induced 2', other_designations='', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=161717036, end_position_on_the_genomic_accession=161779494, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  24010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24010, CurrentID=0, Status='live', Symbol='Ik', Aliases='MuRED', description='IK cytokine', other_designations='IK factor|arginine/glutamic acid/aspartic acid repeat containing protein|cytokine IK', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36744656, end_position_on_the_genomic_accession=36757639, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  24012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24012, CurrentID=0, Status='live', Symbol='Rgs7', Aliases='', description='regulator of G protein signaling 7', other_designations='Regulator of G-protein signaling 7 (RGS7)', map_location='1 81.11 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175059076, end_position_on_the_genomic_accession=175493238, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  24013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24013, CurrentID=0, Status='live', Symbol='Grk1', Aliases='Gprk1, Rhok, Rk', description='G protein-coupled receptor kinase 1', other_designations='G-protein receptor kinase 1', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13405081, end_position_on_the_genomic_accession=13421951, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  24014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24014, CurrentID=0, Status='live', Symbol='Rnasel', Aliases='E230029I04Rik', description="ribonuclease L (2', 5'-oligoisoadenylate synthetase-dependent)", other_designations='2-5A-dependent RNAase|2-5A-dependent RNase|RNase L|ribonuclease 4', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153749119, end_position_on_the_genomic_accession=153768353, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  24015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24015, CurrentID=0, Status='live', Symbol='Abce1', Aliases='C79080, Oabp, RLI, RNS41, Rnaseli', description='ATP-binding cassette, sub-family E (OABP), member 1', other_designations="ATP-binding cassette, subfamily E, member 1|RNS4I|RNS4l (Eye)|RNase L inhibitor|ribonuclease 4 inhibitor|ribonuclease L (2', 5'-oligoisoadenylate synthetase-dependent) inhibitor", map_location='8 C', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79683442, end_position_on_the_genomic_accession=79711740, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  24017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24017, CurrentID=0, Status='live', Symbol='Rnf13', Aliases='2010001H16Rik, Rzf', description='ring finger protein 13', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=57736043, end_position_on_the_genomic_accession=57835233, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  24018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24018, CurrentID=0, Status='live', Symbol='Rngtt', Aliases='AU020997, HCE, MCE1', description="RNA guanylyltransferase and 5'-phosphatase", other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=33310297, end_position_on_the_genomic_accession=33502610, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  24030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24030, CurrentID=0, Status='live', Symbol='Mrps12', Aliases='AI327385, Rpms12, rps12', description='mitochondrial ribosomal protein S12', other_designations='MRP-S12|MT-RPS12|S12mt|mitoribosomal protein s12|ribosomal protein, mitochondrial, S12', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28739641, end_position_on_the_genomic_accession=28741890, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  24044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24044, CurrentID=0, Status='live', Symbol='Scamp2', Aliases='AI875466, Sc2', description='secretory carrier membrane protein 2', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57560928, end_position_on_the_genomic_accession=57588798, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  24045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24045, CurrentID=0, Status='live', Symbol='Scamp3', Aliases='Sc3', description='secretory carrier membrane protein 3', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89177485, end_position_on_the_genomic_accession=89182770, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  24046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24046, CurrentID=0, Status='live', Symbol='Scn11a', Aliases='NSS2, NaN, NaT, NaV1.9, SNS2', description='sodium channel, voltage-gated, type XI, alpha', other_designations='sensory neuron sodium channel 2|sodium channel protein type XI subunit alpha|sodium channel, voltage-gated, type XI, alpha polypeptide|voltage-gated sodium channel NAV1.9b|voltage-gated sodium channel Nav1.9 variant a|voltage-gated sodium channel Nav1.9 variant b|voltage-gated sodium channel Nav1.9 variant c|voltage-gated sodium channel subunit alpha Nav1.9', map_location='9 F3-F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119753730, end_position_on_the_genomic_accession=119825456, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  24047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24047, CurrentID=0, Status='live', Symbol='Ccl19', Aliases='CKb11, ELC, MIP3B, Scya19, exodus-3', description='chemokine (C-C motif) ligand 19', other_designations='EBI-1 ligand chemokine|EBI1 ligand chemokine|EBI1-ligand chemokine|EBV-induced molecule 1 ligand chemokine|chemokine CCL19|epstein-Barr virus-induced molecule 1 ligand chemokine|small inducible cytokine A19|small-inducible cytokine A19', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42754525, end_position_on_the_genomic_accession=42756518, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  24050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24050, CurrentID=0, Status='live', Symbol='Sept3', Aliases='3110018K01Rik, AV154067, B530002E20Rik, Sep3', description='septin 3', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82272451, end_position_on_the_genomic_accession=82294574, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  24051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24051, CurrentID=0, Status='live', Symbol='Sgcb', Aliases='43kDa, AI747103, AI844814, beta-SG', description='sarcoglycan, beta (dystrophin-associated glycoprotein)', other_designations='43 kDa dystrophin-associated glycoprotein|43DAG|sarcoglycan, beta (43kD dystrophin-associated glycoprotein)', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73632749, end_position_on_the_genomic_accession=73647731, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  24052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24052, CurrentID=0, Status='live', Symbol='Sgcd', Aliases='35kDa, delta-SG', description='sarcoglycan, delta (dystrophin-associated glycoprotein)', other_designations='35 kDa dystrophin-associated glycoprotein|35DAG|SG-delta|sarcoglycan, delta (35kD dystrophin-associated glycoprotein)', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46971077, end_position_on_the_genomic_accession=47988385, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  24053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24053, CurrentID=0, Status='live', Symbol='Sgcg', Aliases='35kDa, 5430420E18Rik, AI642964, gamma-SG', description='sarcoglycan, gamma (dystrophin-associated glycoprotein)', other_designations='35 kDa dystrophin-associated glycoprotein|35DAG|sarcoglycan, gamma (35kD dystrophin-associated glycoprotein)', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61219106, end_position_on_the_genomic_accession=61274858, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  24055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24055, CurrentID=0, Status='live', Symbol='Sh3bp2', Aliases='3BP2', description='SH3-domain binding protein 2', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34525784, end_position_on_the_genomic_accession=34563639, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  24056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24056, CurrentID=0, Status='live', Symbol='Sh3bp5', Aliases='AI606498, SH3BP-5, Sab', description='SH3-domain binding protein 5 (BTK-associated)', other_designations='SH3 domain-binding protein that preferentially associates with BTK', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31373774, end_position_on_the_genomic_accession=31436033, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  24057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24057, CurrentID=0, Status='live', Symbol='Sh3yl1', Aliases='AI314953, Ray, YSC84', description='Sh3 domain YSC-like 1', other_designations='', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=30911669, end_position_on_the_genomic_accession=30960160, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  24058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24058, CurrentID=0, Status='live', Symbol='Sigirr', Aliases='AI256711, TIR8', description='single immunoglobulin and toll-interleukin 1 receptor (TIR) domain', other_designations='single Ig IL-1 receptor related protein|single Ig IL-1R-related molecule|single Ig IL-1R-related protein|single immunoglobulin domain-containing IL1R-related protein|toll/interleukin-1 receptor 8', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141091175, end_position_on_the_genomic_accession=141100610, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  24059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24059, CurrentID=0, Status='live', Symbol='Slco2a1', Aliases='2310021C19Rik, Pgt, Slc21a2, mPgt', description='solute carrier organic anion transporter family, member 2a1', other_designations='prostaglandin transporter|solute carrier family 21 (prostaglandin transporter), member 2|solute carrier family 21 member 2', map_location='9 54.72 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103007668, end_position_on_the_genomic_accession=103087849, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  24060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24060, CurrentID=0, Status='live', Symbol='Slc35a1', Aliases='AA408150, AI314851', description='solute carrier family 35 (CMP-sialic acid transporter), member 1', other_designations='CMP-SA-Tr|CMP-Sia-Tr|solute carrier family 35 member A1', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34663257, end_position_on_the_genomic_accession=34687812, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  24061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24061, CurrentID=0, Status='live', Symbol='Smc1a', Aliases='5830426I24Rik, SMC-1A, Sb1.8, Smc1lpha, Smc1l1, Smcb, mKIAA0178, Smc1a', description='structural maintenance of chromosomes 1A', other_designations='SMC (segregation of mitotic chromosomes 1)-like 1|SMC protein 1A|SMC-1-alpha|chromosome segregation protein SmcB|segregation of mitotic chromosomes b', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152016428, end_position_on_the_genomic_accession=152061973, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  24063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24063, CurrentID=0, Status='live', Symbol='Spry1', Aliases='sprouty1, spry-1', description='sprouty homolog 1 (Drosophila)', other_designations='inhibitor of receptor tyrosine kinases', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37639846, end_position_on_the_genomic_accession=37644599, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  24064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24064, CurrentID=0, Status='live', Symbol='Spry2', Aliases='sprouty2', description='sprouty homolog 2 (Drosophila)', other_designations='spry-2', map_location='14 56.16 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=105891947, end_position_on_the_genomic_accession=105896819, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  24066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24066, CurrentID=0, Status='live', Symbol='Spry4', Aliases='A030006O18Rik, sprouty4', description='sprouty homolog 4 (Drosophila)', other_designations='spry-4', map_location='18 20.5 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38586265, end_position_on_the_genomic_accession=38601268, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  24067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24067, CurrentID=0, Status='live', Symbol='Srp54a', Aliases='54kDa, Srp54', description='signal recognition particle 54A', other_designations='54K subunit of signal recognition particle', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55080496, end_position_on_the_genomic_accession=55115367, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  24068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24068, CurrentID=0, Status='live', Symbol='Sra1', Aliases='AA959952, Sra, Srap, Straa1, Strra1', description='steroid receptor RNA activator 1', other_designations='sra1|steroid receptor RNA activator protein', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36667187, end_position_on_the_genomic_accession=36670311, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  24069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24069, CurrentID=0, Status='live', Symbol='Sufu', Aliases='Su(fu)', description='suppressor of fused homolog (Drosophila)', other_designations='', map_location='19 38.85 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46396896, end_position_on_the_genomic_accession=46488804, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  24070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24070, CurrentID=0, Status='live', Symbol='Mpdu1', Aliases='LEC35, SL15, Supl15h', description='mannose-P-dolichol utilization defect 1', other_designations='suppressor of Lec15 and Lec35 glycosylation mutation homolog|suppressor of Lec15 homolog (C.griseus)', map_location='11 42.86 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69656697, end_position_on_the_genomic_accession=69662649, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  24071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24071, CurrentID=0, Status='live', Symbol='Synj2bp', Aliases='AA409442, ARIP2, ActRIP4, D12Wsu118e, OMP25', description='synaptojanin 2 binding protein', other_designations='activin receptor interacting protein 2|activin receptor-interacting protein 4|mitochondrial outer membrane protein 25|outer membrane protein 25', map_location='12 37.73 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81492194, end_position_on_the_genomic_accession=81532907, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  24074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24074, CurrentID=0, Status='live', Symbol='Taf7', Aliases='55kDa, AI607308, BB007175, TAFII55, Taf2f', description='TATA-box binding protein associated factor 7', other_designations='RNA polymerase II TBP-associated factor subunit F|TAF(II)55|TAF7 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF7 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 55 kDa|TAFII-55|TATA box binding protein (TBP) associated factor, RNA polymerase II, F, 55 kDa|transcription initiation factor TFIID 55 kDa subunit', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37640491, end_position_on_the_genomic_accession=37644204, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  24075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24075, CurrentID=0, Status='live', Symbol='Taf10', Aliases='30kDa, AU041226, TAFII30, Taf2h', description='TATA-box binding protein associated factor 10', other_designations='TAF(II)30|TAF10 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF10 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 30 kDa|TAFII-30|TATA box binding protein (TBP)-associated factor, RNA polymerase II, H, 30 kDa|mTAFII30|transcription initiation factor TFIID 30 kDa subunit', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105742894, end_position_on_the_genomic_accession=105744375, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  24082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24082, CurrentID=0, Status='live', Symbol='Map2k3os', Aliases='F3-1, Gm16516, Gtlf3a', description='mitogen-activated protein kinase kinase 3, opposite strand', other_designations='Map2k3 opposite strand transcript|gene trap locus F3a|transcript expressed during hematopoiesis 1', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60920941, end_position_on_the_genomic_accession=60931867, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  24083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24083, CurrentID=0, Status='live', Symbol='Natd1', Aliases='AI256713, F3-2, Gm16515, Gtlf3b', description='N-acetyltransferase domain containing 1', other_designations='N-acetyltransferase domain-containing protein 1|gene trap locus F3b|gene trap locus protein F3-2|gene trap locus protein F3b|protein GTLF3B', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60902246, end_position_on_the_genomic_accession=60913792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  24084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24084, CurrentID=0, Status='live', Symbol='Tekt2', Aliases='AU014974, tektin-t', description='tektin 2', other_designations='testicular tektin', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126322121, end_position_on_the_genomic_accession=126325199, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  24086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24086, CurrentID=0, Status='live', Symbol='Tlk2', Aliases='4933403M19Rik, PKU-alpha, PKUalpha, Tlk', description='tousled-like kinase 2 (Arabidopsis)', other_designations='protein kinase U-alpha', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105167521, end_position_on_the_genomic_accession=105283959, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  24087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24087, CurrentID=0, Status='live', Symbol='Tll2', Aliases='', description='tolloid-like 2', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41080940, end_position_on_the_genomic_accession=41210765, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  24088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24088, CurrentID=0, Status='live', Symbol='Tlr2', Aliases='Ly105', description='toll-like receptor 2', other_designations='', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=83836272, end_position_on_the_genomic_accession=83841825, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  24099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24099, CurrentID=0, Status='live', Symbol='Tnfsf13b', Aliases='BAFF, BLyS, D8Ertd387e, TALL-1, TALL1, THANK, TNFSF20, Tnlg7a, zTNF4', description='tumor necrosis factor (ligand) superfamily, member 13b', other_designations='B-cell-activating factor|b cell-activating factor|tumor necrosis factor ligand 7a', map_location='8 4.55 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=10006192, end_position_on_the_genomic_accession=10119693, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  24100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24100, CurrentID=0, Status='live', Symbol='Tpra1', Aliases='40kDa, Gpr175, Tpra40', description='transmembrane protein, adipocyte asscociated 1', other_designations='G protein-coupled receptor 175|integral membrane protein GPR175|transmembrane domain protein of 40 kDa regulated in adipocytes|transmembrane domain protein regulated in adipocytes 40 kDa', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88902251, end_position_on_the_genomic_accession=88912240, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  24102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24102, CurrentID=0, Status='live', Symbol='Trex2', Aliases='', description='three prime repair exonuclease 2', other_designations="3'-5' exonuclease TREX2", map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73433705, end_position_on_the_genomic_accession=73435343, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  24105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24105, CurrentID=0, Status='live', Symbol='Rbck1', Aliases='AL033326, HOIL-1, HOIL-1L, UIP28, Ubce7ip3', description='RanBP-type and C3HC4-type zinc finger containing 1', other_designations='heme-oxidized IRP2 ubiquitin ligase 1 homolog|novel zinc finger, C3HC4 type (RING finger) protein|ubcM4-interacting protein 28|ubiquitin conjugating enzyme 7 interacting protein 3|ubiquitin-conjugating enzyme 7-interacting protein 3', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152316334, end_position_on_the_genomic_accession=152332643, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  24108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24108, CurrentID=0, Status='live', Symbol='Ubd', Aliases='FAT10', description='ubiquitin D', other_designations='Diubiquitin|ubiquitin-like protein FAT10', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37193892, end_position_on_the_genomic_accession=37196101, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  24109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24109, CurrentID=0, Status='live', Symbol='Ubl3', Aliases='AW108023, HCG', description='ubiquitin-like 3', other_designations='MUB|membrane-anchored ubiquitin-fold protein|mmMUB|protein HCG-1', map_location='5 G2-3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=148504631, end_position_on_the_genomic_accession=148552788, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  24110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24110, CurrentID=0, Status='live', Symbol='Usp18', Aliases='1110058H21Rik, AW047653, UBP43, Ubp15', description='ubiquitin specific peptidase 18', other_designations='43 kDa ISG15-specific protease|ISG15-specific-processing protease|ubiquitin specific protease 15|ubiquitin specific protease 18|ubl thioesterase 18|ubl thiolesterase 18', map_location='6 F', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121245906, end_position_on_the_genomic_accession=121270917, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  24111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24111, CurrentID=0, Status='live', Symbol='Uts2', Aliases='UII, Ucn2, prepro-UII', description='urotensin 2', other_designations='U-II|urocortin 2|urotensin II|urotensin-II', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150997097, end_position_on_the_genomic_accession=151001810, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  24112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24112, CurrentID=0, Status='live', Symbol='Vmn1r49', Aliases='V1r5, V1rb2, VRi2', description='vomeronasal 1, receptor 49', other_designations='vomeronasal 1 receptor Vmn1r49|vomeronasal 1, receptor B2|vomeronasal organ family 1, receptor 5|vomeronasal organ family 1, receptor B2|vomeronasal receptor VRi2|vomeronasal type 1 receptor B2|vomeronasal type-1 receptor B2', map_location='6 40.02 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90072086, end_position_on_the_genomic_accession=90078374, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  24113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24113, CurrentID=0, Status='live', Symbol='Vax2', Aliases='Dres93', description='ventral anterior homeobox 2', other_designations='ventral anterior homeobox containing gene 2|ventral retina homeodomain protein', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83711264, end_position_on_the_genomic_accession=83739229, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  24115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24115, CurrentID=0, Status='live', Symbol='Best1', Aliases='Bmd, Vmd2, mBest1', description='bestrophin 1', other_designations='best macular dystrophy|vitelliform macular dystrophy 2 homolog|vitelliform macular dystrophy protein 2 homolog', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=9985008, end_position_on_the_genomic_accession=10001653, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  24116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24116, CurrentID=0, Status='live', Symbol='Nelfa', Aliases='Whsc2, Whsc2h', description='negative elongation factor complex member A, Whsc2', other_designations='Wolf-Hirschhorn syndrome candidate 2 homolog|mWHSC2', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33898179, end_position_on_the_genomic_accession=33936258, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  24117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24117, CurrentID=0, Status='live', Symbol='Wif1', Aliases='AW107799, WIF-1', description='Wnt inhibitory factor 1', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121034004, end_position_on_the_genomic_accession=121100642, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  24127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24127, CurrentID=0, Status='live', Symbol='Xrn1', Aliases='Dhm2, exo, mXrn1', description="5'-3' exoribonuclease 1", other_designations='protein Dhm2|strand-exchange protein 1 homolog', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95954760, end_position_on_the_genomic_accession=96057800, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  24128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24128, CurrentID=0, Status='live', Symbol='Xrn2', Aliases='', description="5'-3' exoribonuclease 2", other_designations='protein Dhm1', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=147012631, end_position_on_the_genomic_accession=147078002, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  24131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24131, CurrentID=0, Status='live', Symbol='Ldb3', Aliases='AW742271, PDLIM6, ZASP', description='LIM domain binding 3', other_designations='PDZ-LIM domain protein|Z-band alternatively spliced PDZ-motif protein|cypher|ldb3|oracle protein|protein cypher|protein oracle', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34526699, end_position_on_the_genomic_accession=34588698, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  24132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24132, CurrentID=0, Status='live', Symbol='Zfp53', Aliases='D030067O06Rik, KRAZ1, Zfp-53, Zfp118, zfas8', description='zinc finger protein 53', other_designations='Zinc finger protein 118', map_location='17 11.4 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21488988, end_position_on_the_genomic_accession=21510477, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  24135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24135, CurrentID=0, Status='live', Symbol='Zfp68', Aliases='KRAB3, KRAZ2, Zfp70, mszf22, mszf49', description='zinc finger protein 68', other_designations='zinc finger protein 70', map_location='5 77.1 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138603652, end_position_on_the_genomic_accession=138619743, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  24136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=24136, CurrentID=0, Status='live', Symbol='Zeb2', Aliases='9130203F04Rik, D130016B08Rik, SIP1, Zfhx1b, Zfx1b, Zfxh1b', description='zinc finger E-box binding homeobox 2', other_designations='smad-interacting protein 1|zinc finger homeobox protein 1b', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=44983512, end_position_on_the_genomic_accession=45113279, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100294583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100294583, CurrentID=0, Status='live', Symbol='Fam150b', Aliases='6230419C23Rik', description='family with sequence similarity 150, member B', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=30884324, end_position_on_the_genomic_accession=30893854, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100294660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100294660, CurrentID=0, Status='live', Symbol='Defa2', Aliases='CR2, Defcr-2, Defcr2', description='defensin, alpha, 2', other_designations='alpha-defensin, 2|cryptdin-1|defensin-related cryptdin 2|defensin-related cryptdin-2', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21378517, end_position_on_the_genomic_accession=21379495, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69655, CurrentID=0, Status='live', Symbol='Cd164l2', Aliases='2310047N01Rik', description='CD164 sialomucin-like 2', other_designations='D164 sialomucin-like 2', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133220809, end_position_on_the_genomic_accession=133224554, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69661, CurrentID=0, Status='live', Symbol='2310061N02Rik', Aliases='', description='RIKEN cDNA 2310061N02 gene', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88707171, end_position_on_the_genomic_accession=88707962, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  286940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=286940, CurrentID=0, Status='live', Symbol='Flnb', Aliases='AL024016, Fln-b', description='filamin, beta', other_designations='ABP-280-like protein|actin binding protein 278|actin-binding-like protein|beta-filamin', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=7817957, end_position_on_the_genomic_accession=7951588, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  286942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=286942, CurrentID=0, Status='live', Symbol='Kif19a', Aliases='Kif19', description='kinesin family member 19A', other_designations='KIF19A-like protein|N-8 kinesin|oocyte-G1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114764501, end_position_on_the_genomic_accession=114791118, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  286951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=286951, CurrentID=0, Status='live', Symbol='Eif2c5', Aliases='Ago5', description='eukaryotic translation initiation factor 2C, 5', other_designations='argonaute 5', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140776238, end_position_on_the_genomic_accession=140796880, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69731, CurrentID=0, Status='live', Symbol='Gemin7', Aliases='2400008I04Rik, AI120175, AI461688', description='gem (nuclear organelle) associated protein 7', other_designations='gemin-7', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19564949, end_position_on_the_genomic_accession=19573343, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69773, CurrentID=0, Status='live', Symbol='1810026J23Rik', Aliases='', description='RIKEN cDNA 1810026J23 gene', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21592722, end_position_on_the_genomic_accession=21595970, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100033459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100033459, CurrentID=0, Status='live', Symbol='Pydc3', Aliases='9830148g24rik, E430029J22Rik', description='pyrin domain containing 3', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173673628, end_position_on_the_genomic_accession=173698392, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  407814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407814, CurrentID=0, Status='live', Symbol='BC053393', Aliases='', description='cDNA sequence BC053393', other_designations='hepatitis A virus cellular receptor 1-like family member', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46571513, end_position_on_the_genomic_accession=46589232, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  18313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18313, CurrentID=0, Status='live', Symbol='Olfr16', Aliases='MOR23, MOR26713', description='olfactory receptor 16', other_designations='', map_location='1 H4-H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172956768, end_position_on_the_genomic_accession=172957817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  637898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=637898, CurrentID=0, Status='live', Symbol='Vmn2r60', Aliases='Casr-rs3, EG637898, Gprc2a-rs3', description='vomeronasal 2, receptor 60', other_designations='G protein coupled receptor, family C, group 2, member A, related sequence 3|G protein-coupled receptor, family C, group 2, member A, related sequence 3|calcium sensing receptor, related sequence 3|vomeronasal receptor Vmn2r60', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=42116471, end_position_on_the_genomic_accession=42195776, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  550619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=550619, CurrentID=0, Status='live', Symbol='Arid3c', Aliases='OTTMUSG00000006683', description='AT rich interactive domain 3C (BRIGHT-like)', other_designations='ARID domain-containing protein 3C|AT rich interactive domain 3C (Bright like)', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41723831, end_position_on_the_genomic_accession=41731867, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26356, CurrentID=0, Status='live', Symbol='Ing1', Aliases='2610028J21Rik, AA407184, AI875420, mING1h, p33Ing1, p37Ing1b', description='inhibitor of growth family, member 1', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11555762, end_position_on_the_genomic_accession=11563251, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  26357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26357, CurrentID=0, Status='live', Symbol='Abcg2', Aliases='ABC15, ABCP, AI428558, BCRP, Bcrp1, MXR, MXR1', description='ATP-binding cassette, sub-family G (WHITE), member 2', other_designations='breast cancer resistance protein 1 homolog|mitoxantrone resistance protein 1|urate exporter', map_location='6 27.82 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58584491, end_position_on_the_genomic_accession=58692451, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  26358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26358, CurrentID=0, Status='live', Symbol='Aldh1a7', Aliases='AI987940, Ahd2-like, Aldh-pb, Aldh1a4, Aldhpb', description='aldehyde dehydrogenase family 1, subfamily A7', other_designations='ALDH class 1|ALDH-E1|ALHDII|Aldehyde dehydrogenase 1 (phenobarbitol inducible)|aldehyde dehydrogenase family 1 member A7|aldehyde dehydrogenase family 1, subfamily A4|aldehyde dehydrogenase phenobarbital-inducible|aldh1a7', map_location='19 13.95 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=20692953, end_position_on_the_genomic_accession=20727556, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  26359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26359, CurrentID=0, Status='live', Symbol='Anxa10', Aliases='', description='annexin A10', other_designations='annexin-10', map_location='8 31.55 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=62057042, end_position_on_the_genomic_accession=62123193, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  26360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26360, CurrentID=0, Status='live', Symbol='Angptl2', Aliases='AI593246, AW260363, Arp2', description='angiopoietin-like 2', other_designations='angiopoietin related protein 2|angiopoietin-like protein 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33215951, end_position_on_the_genomic_accession=33247725, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26361, CurrentID=0, Status='live', Symbol='Avpr1b', Aliases='AVPR3, V3/V1b, VIBR, VPR3', description='arginine vasopressin receptor 1B', other_designations='AVPR V1b|AVPR V3|V3/V1b pituitary vasopressin receptor|antidiuretic hormone receptor 1b|arginine vasopressin type 1b receptor|vasopressin V3 receptor', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131599114, end_position_on_the_genomic_accession=131612000, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26362, CurrentID=0, Status='live', Symbol='Axl', Aliases='AI323647, Ark, Tyro7, Ufo', description='AXL receptor tyrosine kinase', other_designations='adhesion-related kinase|ufo oncogene homolog', map_location='7 A3-B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25756500, end_position_on_the_genomic_accession=25788733, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  26363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26363, CurrentID=0, Status='live', Symbol='Btd', Aliases='', description='biotinidase', other_designations='biotinase', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31640999, end_position_on_the_genomic_accession=31668197, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26364, CurrentID=0, Status='live', Symbol='Adgre5', Aliases='AA409984, Cd97, TM7LN1', description='adhesion G protein-coupled receptor E5', other_designations='EGF-TM7 receptor', map_location='8 40.22 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83723250, end_position_on_the_genomic_accession=83741311, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  26365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26365, CurrentID=0, Status='live', Symbol='Ceacam1', Aliases='Bgp, Bgp1, C-CAM, CD66a, Cc1, Cea-1, Cea-7, Cea1, Cea7, Hv-2, Hv2, MHVR, MHVR1, Mhv-1, bb-1, mCEA1, mmCGM1, mmCGM1a, mmCGM2', description='carcinoembryonic antigen-related cell adhesion molecule 1', other_designations='BGP-1|CEA-related cell adhesion molecule 1|MHV-R|biliary glycoprotein 1|biliary glycoprotein D|carcinoembryonic antigen 1|carcinoembryonic antigen 7|hepatitis virus (MHV-4) susceptibility|hepatitis virus receptor|murine hepatitis virus receptor', map_location='7 13.84 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25461702, end_position_on_the_genomic_accession=25477625, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26366, CurrentID=0, Status='live', Symbol='Ceacam10', Aliases='Bgp3, Cea10', description='carcinoembryonic antigen-related cell adhesion molecule 10', other_designations='CEA-related cell adhesion molecule 10|carcinoembryonic antigen 10', map_location='7 12.78 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24777156, end_position_on_the_genomic_accession=24784656, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26367, CurrentID=0, Status='live', Symbol='Ceacam2', Aliases='AI267129, Bgp2-L, Ceacam2', description='carcinoembryonic antigen-related cell adhesion molecule 2', other_designations='BGP-2|CEA-related cell adhesion molecule 2|biliary glycoprotein 2|carcinoembryonic antigen-related cell adhesion molecule 2- Long', map_location='7 13.87 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25516042, end_position_on_the_genomic_accession=25540004, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26368, CurrentID=0, Status='live', Symbol='Ceacam9', Aliases='AA410097, AW545709, Cea-5, Cea5, mmCGM8', description='carcinoembryonic antigen-related cell adhesion molecule 9', other_designations='CEA-related cell adhesion molecule 9|carcinoembryonic antigen 5', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16721929, end_position_on_the_genomic_accession=16726110, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26369, CurrentID=0, Status='live', Symbol='Cetn1', Aliases='caltractin', description='centrin 1', other_designations='testis-specific murine centrin', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=9618419, end_position_on_the_genomic_accession=9619469, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  26370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26370, CurrentID=0, Status='live', Symbol='Cetn2', Aliases='1110034A02Rik, AI326150, Calt, caltractin', description='centrin 2', other_designations='caltractin, 20 kDa calcium binding protein', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72913565, end_position_on_the_genomic_accession=72918344, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26371, CurrentID=0, Status='live', Symbol='Ciao1', Aliases='AW210570, Wdr39', description='cytosolic iron-sulfur protein assembly 1', other_designations='WD repeat domain 39|WD repeat-containing protein 39|WD40 protein Ciao1|ciao1|cytosolic iron-sulfur protein assembly 1 homolog', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127240938, end_position_on_the_genomic_accession=127247816, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26372, CurrentID=0, Status='live', Symbol='Clcn6', Aliases='AI850629, Clc6', description='chloride channel, voltage-sensitive 6', other_designations='chloride channel 6|chloride channel protein 6|clC-6', map_location='4 78.67 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148004259, end_position_on_the_genomic_accession=148038858, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  26373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26373, CurrentID=0, Status='live', Symbol='Clcn7', Aliases='ClC-7, D17Wsu51e', description='chloride channel, voltage-sensitive 7', other_designations='chloride channel 7 alpha subunit|chloride channel protein 7', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25133379, end_position_on_the_genomic_accession=25162100, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  26374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26374, CurrentID=0, Status='live', Symbol='Rfwd2', Aliases='AI316802, C80879, Cop1', description='ring finger and WD repeat domain 2', other_designations='RING finger and WD repeat domain protein 2|constitutive photomorphogenesis protein 1 homolog|constitutive photomorphogenic protein 1|mCOP1', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=159232174, end_position_on_the_genomic_accession=159347638, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  26377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26377, CurrentID=0, Status='live', Symbol='Dapp1', Aliases='Bam32', description='dual adaptor for phosphotyrosine and 3-phosphoinositides 1', other_designations='B-cell adapter molecule of 32 kDa|b lymphocyte adapter protein Bam32', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=137930998, end_position_on_the_genomic_accession=137981576, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26378, CurrentID=0, Status='live', Symbol='Decr2', Aliases='', description='2-4-dienoyl-Coenzyme A reductase 2, peroxisomal', other_designations='2,4-dienoyl-CoA reductase 2', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26081208, end_position_on_the_genomic_accession=26091517, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26379, CurrentID=0, Status='live', Symbol='Esrra', Aliases='ERRalpha, Err1, Estrra, Nr3b1', description='estrogen related receptor, alpha', other_designations='ERR-alpha|Steroid hormone receptor ERR1 (Estrogen-related receptor, alpha) (ERR-alpha) (Estrogen receptor-like 1)|estrogen receptor related 1|estrogen receptor-like 1|estrogen-related receptor alpha|nuclear receptor subfamily 3 group B member 1', map_location='19 5.08 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6910977, end_position_on_the_genomic_accession=6922048, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26380, CurrentID=0, Status='live', Symbol='Esrrb', Aliases='Err2, Errb, Estrrb, Nr3b2', description='estrogen related receptor, beta', other_designations='ERR-beta|estrogen receptor related 2|estrogen receptor-like 2|estrogen-related receptor beta|nuclear receptor subfamily 3 group B member 2', map_location='12 40.49 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86361117, end_position_on_the_genomic_accession=86521628, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  26381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26381, CurrentID=0, Status='live', Symbol='Esrrg', Aliases='ERR3, Errg, NR3B3, mKIAA0832', description='estrogen-related receptor gamma', other_designations='estrogen receptor-related protein 3|estrogen-related receptor 3|nuclear receptor subfamily 3 group B member 3', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=187608792, end_position_on_the_genomic_accession=188214884, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  26382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26382, CurrentID=0, Status='live', Symbol='Fgd2', Aliases='Tcd-2, Tcd2, ZFYVE4, tcs-2, tcs2', description='FYVE, RhoGEF and PH domain containing 2', other_designations='Fgd1 family, member 2|Fgd2-like protein|faciogenital dysplasia homolog 2|faciogenital dysplasia protein 2|t-complex distorter 2', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29360908, end_position_on_the_genomic_accession=29379535, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  26383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26383, CurrentID=0, Status='live', Symbol='Fto', Aliases='AW743446, mKIAA1752', description='fat mass and obesity associated', other_designations='fat mass and obesity-associated protein|protein fatso', map_location='8 44.34 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=91313491, end_position_on_the_genomic_accession=91668433, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26384, CurrentID=0, Status='live', Symbol='Gnpda1', Aliases='GNPDA, Gnp1, Gnpi, oscillin', description='glucosamine-6-phosphate deaminase 1', other_designations='GNPDA 1|glcN6P deaminase 1|glucose-6-phosphate isomerase', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38327537, end_position_on_the_genomic_accession=38338993, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26385, CurrentID=0, Status='live', Symbol='Grk6', Aliases='Gprk6', description='G protein-coupled receptor kinase 6', other_designations='g protein-coupled receptor kinase GRK6', map_location='13 B1-3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55445072, end_position_on_the_genomic_accession=55460927, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  26386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26386, CurrentID=0, Status='live', Symbol='Hsf4', Aliases='HSF 4, HSTF 4, HSTF4, ldis1, mHSF4', description='heat shock transcription factor 4', other_designations='DNA binding protein|binds heat shock DNA elements|heat shock factor 4', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105269748, end_position_on_the_genomic_accession=105275845, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  26388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26388, CurrentID=0, Status='live', Symbol='Ifi202b', Aliases='Ifbip-1, Ifi202, Ifi202a, p202', description='interferon activated gene 202B', other_designations='Interferon-activatable protein 202a (Ifi-202a) (Interferon-inducible protein p202a)|ifi-202|interferon activated gene 202A|interferon-inducible protein p202|lupus susceptibility protein p202', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173962569, end_position_on_the_genomic_accession=173982844, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26390, CurrentID=0, Status='live', Symbol='Mapkbp1', Aliases='2810483F24Rik, AW123212, JNKBP-1, Jnkbp1, mKIAA0596', description='mitogen-activated protein kinase binding protein 1', other_designations='JNK-binding protein 1|Jun N-terminal kinase binding protein 1|Jun N-terminal kinase-binding protein 1|mitogen activated protein kinase binding protein 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119972388, end_position_on_the_genomic_accession=120027403, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  26394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26394, CurrentID=0, Status='live', Symbol='Lypla2', Aliases='LysoII', description='lysophospholipase 2', other_designations='APT-2|LPL-II|lysoPLA II|lysophospholipase II|mLyso II', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135968224, end_position_on_the_genomic_accession=135972594, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26395, CurrentID=0, Status='live', Symbol='Map2k1', Aliases='MAPKK1, MEKK1, Mek1, Prkmk1', description='mitogen-activated protein kinase kinase 1', other_designations='ERK activator kinase 1|MAP kinase kinase 1|MAPK/ERK kinase 1|MAPKK 1|MEK 1|mitogen activated protein kinase kinase 1|protein kinase, mitogen activated, kinase 1, p45', map_location='9 34.55 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64185770, end_position_on_the_genomic_accession=64253605, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  26396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26396, CurrentID=0, Status='live', Symbol='Map2k2', Aliases='AA589381, MEK2, MK2, Prkmk2', description='mitogen-activated protein kinase kinase 2', other_designations='ERK activator kinase 2|MAP kinase kinase 2|MAP kinase/Erk kinase|MAPK/ERK kinase 2|MAPKK 2|MEK 2|mitogen activated protein kinase kinase 2|protein kinase, mitogen activated, kinase 2, p45', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81105912, end_position_on_the_genomic_accession=81124697, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  26397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26397, CurrentID=0, Status='live', Symbol='Map2k3', Aliases='AW212142, MEK3, MKK3, Prkmk3, mMKK3b', description='mitogen-activated protein kinase kinase 3', other_designations='MAP kinase kinase 3|MAPK/ERK kinase 3|MAPKK 3|MEK 3|mitogen activated protein kinase kinase 3|protein kinase, mitogen-activated, kinase 3', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60932057, end_position_on_the_genomic_accession=60952803, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  26398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26398, CurrentID=0, Status='live', Symbol='Map2k4', Aliases='JNKK1, MEK4, MKK4, PRKMK4, Sek1, Serk1', description='mitogen-activated protein kinase kinase 4', other_designations='C-JUN N-terminal kinase kinase 1|JNK kinase 1|JNK-activating kinase 1|JNKK 1|MAP kinase kinase 4|MAPK/ERK kinase 4|MAPKK 4|MEK 4|SAPK/ERK kinase 1|SAPK/Erk/kinase 1|mitogen activated protein kinase kinase 4, variant 1', map_location='11 40.53 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=65688244, end_position_on_the_genomic_accession=65788354, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  26399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26399, CurrentID=0, Status='live', Symbol='Map2k6', Aliases='MEK6, MKK6, Prkmk6, SAPKK3', description='mitogen-activated protein kinase kinase 6', other_designations='MAP kinase kinase 6|MAPK/ERK kinase 6|MAPKK 6|MEK 6|mitogen activated protein kinase kinase 6|protein kinase, mitogen activated, kinase 6', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=110399102, end_position_on_the_genomic_accession=110516212, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  26400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26400, CurrentID=0, Status='live', Symbol='Map2k7', Aliases='5930412N11Rik, JNKK 2, Jnkk2, MAPKK 7, MEK 7, Mapkk7, Mek7, Mkk7, Prkmk7, sek2', description='mitogen-activated protein kinase kinase 7', other_designations='JNK kinase 2|JNK-activating kinase 2|MAP kinase kinase 7|MAPK/ERK kinase 7|c-Jun N-terminal kinase kinase 2|map2k7|mitogen activated protein kinase kinase 7|protein kinase, mitogen activated, kinase 7', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4238740, end_position_on_the_genomic_accession=4247897, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  26401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26401, CurrentID=0, Status='live', Symbol='Map3k1', Aliases='MAPKKK1, MEKK1, Mekk', description='mitogen-activated protein kinase kinase kinase 1', other_designations='MAPK/ERK kinase kinase 1|MEK kinase 1|MEKK 1|mitogen activated protein kinase kinase kinase 1', map_location='13 63.36 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=111746433, end_position_on_the_genomic_accession=111818269, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  26403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26403, CurrentID=0, Status='live', Symbol='Map3k11', Aliases='2610017K16Rik, Mlk3, PTK1, SPRK', description='mitogen-activated protein kinase kinase kinase 11', other_designations='mitogen activated protein kinase kinase kinase 11|mixed lineage kinase 3', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5688740, end_position_on_the_genomic_accession=5702865, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  26404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26404, CurrentID=0, Status='live', Symbol='Map3k12', Aliases='DLK, MUK, Zpk', description='mitogen-activated protein kinase kinase kinase 12', other_designations='MAPK-upstream kinase|dual leucine zipper bearing kinase|leucine-zipper protein kinase|mitogen activated protein kinase kinase kinase 12|mixed lineage kinase|zipper (leucine) protein kinase', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102497647, end_position_on_the_genomic_accession=102518053, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  26405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26405, CurrentID=0, Status='live', Symbol='Map3k2', Aliases='9630061B06Rik, AI585793, Mekk2, Mekk2b', description='mitogen-activated protein kinase kinase kinase 2', other_designations='MAP/ERK kinase kinase 2|MAPK/ERK kinase kinase 2|MEK kinase 2|MEKK 2|mitogen activated protein kinase kinase kinase 2|protein kinase MEKK2b', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32163089, end_position_on_the_genomic_accession=32236751, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  26406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26406, CurrentID=0, Status='live', Symbol='Map3k3', Aliases='AW548911, MAPKKK3, Mekk3, mKIAA4031', description='mitogen-activated protein kinase kinase kinase 3', other_designations='MAPK/ERK kinase kinase 3|MEK kinase 3|MEKK 3|mitogen activated protein kinase kinase kinase 3', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106084902, end_position_on_the_genomic_accession=106156656, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  26407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26407, CurrentID=0, Status='live', Symbol='Map3k4', Aliases='D17Rp17, D17Rp17e, MAPKKK4, MEKK 4, MTK1, Mek4b, Mekk4, RP17, Rp17a, Tas, mKIAA0213', description='mitogen-activated protein kinase kinase kinase 4', other_designations='MAPK/ERK kinase kinase 4|MEK kinase 4|T-associated sex reversal|mitogen activated protein kinase kinase kinase 4', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12227621, end_position_on_the_genomic_accession=12318806, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  26408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26408, CurrentID=0, Status='live', Symbol='Map3k5', Aliases='7420452D20Rik, ASK, ASK1, MAPKKK5, Mekk5', description='mitogen-activated protein kinase kinase kinase 5', other_designations='ASK-1|MAPK/ERK kinase kinase 5|MEK kinase 5|MEKK 5|apoptosis signal-regulating kinase 1|mitogen activated protein kinase kinase kinase 5', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19934513, end_position_on_the_genomic_accession=20142750, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  26409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26409, CurrentID=0, Status='live', Symbol='Map3k7', Aliases='B430101B05, C87327, Tak1', description='mitogen-activated protein kinase kinase kinase 7', other_designations='TGF-beta activated kinase 1|TGF-beta-activated kinase 1|mitogen activated protein kinase kinase kinase 7|transforming growth factor beta-activated kinase 1|transforming growth factor-beta-activated kinase 1', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=31962296, end_position_on_the_genomic_accession=32023470, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  26410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26410, CurrentID=0, Status='live', Symbol='Map3k8', Aliases='Cot, Cot/Tpl2, Est, Estf, Tpl-2, Tpl2, c-COT', description='mitogen-activated protein kinase kinase kinase 8', other_designations='COT proto-oncogene serine/threonine-protein kinase|cancer Osaka thyroid oncogene|cancer Osaka thyroid, oncogene|mitogen activated protein kinase kinase kinase 8|proto-oncogene c-Cot|serine/threonine-protein kinase cot|tumor progression locus 2', map_location='18 2.73 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=4331325, end_position_on_the_genomic_accession=4353481, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26411, CurrentID=0, Status='live', Symbol='Map4k1', Aliases='Hpk1, mHPK1', description='mitogen-activated protein kinase kinase kinase kinase 1', other_designations='HPK|MAPK/ERK kinase kinase kinase 1|MEK kinase kinase 1|MEKKK 1|hematopoietic progenitor kinase 1|mitogen activated protein kinase kinase kinase kinase 1', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28982050, end_position_on_the_genomic_accession=29003279, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  26412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26412, CurrentID=0, Status='live', Symbol='Map4k2', Aliases='AI385662, BL44, GCK, Rab8ip', description='mitogen-activated protein kinase kinase kinase kinase 2', other_designations='MAPK/ERK kinase kinase kinase 2|MEK kinase kinase 2|MEKKK 2|RAB8 interacting protein|germinal center kinase|mitogen activated protein kinase kinase kinase kinase 2|rab8-interacting protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6341116, end_position_on_the_genomic_accession=6355619, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  26413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26413, CurrentID=0, Status='live', Symbol='Mapk1', Aliases='9030612K14Rik, AA407128, AU018647, C78273, ERK, Erk2, MAPK2, PRKM2, Prkm1, p41mapk, p42mapk', description='mitogen-activated protein kinase 1', other_designations='ERK-2|ERT1|MAP kinase 1|MAP kinase 2|MAP kinase isoform p42|MAPK 1|MAPK 2|extracellular signal-regulated kinase 2|mitogen activated protein kinase 1|mitogen-activated protein kinase 2|p42-MAPK', map_location='16 10.53 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16983382, end_position_on_the_genomic_accession=17047453, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26414, CurrentID=0, Status='live', Symbol='Mapk10', Aliases='C230008H04Rik, JNK3, JNK3B1, JNK3B2, SAPK(beta), Serk2, p493F12, p54bSAPK', description='mitogen-activated protein kinase 10', other_designations='JNK3 beta1 protein kinase|JNK3 beta2 protein kinase|MAP kinase 10|MAP kinase p49 3F12|SAPK/Erk/kinase 2|c-Jun N-terminal kinase 3|stress-activated protein kinase JNK3', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=102907948, end_position_on_the_genomic_accession=103211339, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  26415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26415, CurrentID=0, Status='live', Symbol='Mapk13', Aliases='SAPK4, Serk4', description='mitogen-activated protein kinase 13', other_designations='MAP kinase 13|MAP kinase p38 delta|MAPK 13|SAPK/Erk/kinase 4|mitogen activated protein kinase 13|mitogen-activated protein kinase p38 delta|p38 delta MAP kinase|stress-activated protein kinase 4', map_location='17 A3-B', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28769317, end_position_on_the_genomic_accession=28778704, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  26416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26416, CurrentID=0, Status='live', Symbol='Mapk14', Aliases='CSBP2, Crk1, Csbp1, Mxi2, PRKM14, PRKM15, p38, p38-alpha, p38MAPK, p38a, p38alpha', description='mitogen-activated protein kinase 14', other_designations='MAP kinase 14|MAP kinase p38 alpha|MAPK 14|cytokine suppressive anti-inflammatory drug binding protein 1|mitogen activated protein kinase 14|mitogen-activated protein kinase p38 alpha|p38 MAP kinase alpha|p38 MAPK|p38 alpha|tRNA synthetase cofactor p38', map_location='17 14.85 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28691342, end_position_on_the_genomic_accession=28748405, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  26417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26417, CurrentID=0, Status='live', Symbol='Mapk3', Aliases='Erk-1, Erk1, Ert2, Esrk1, Mnk1, Mtap2k, Prkm3, p44, p44erk1, p44mapk', description='mitogen-activated protein kinase 3', other_designations='MAP kinase 3|extracellular signal-regulated kinase 1|insulin-stimulated MAP2 kinase|microtubule-associated protein 2 kinase|p44 MAP kinase|pp42/MAP kinase', map_location='7 69.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126759626, end_position_on_the_genomic_accession=126765816, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26419, CurrentID=0, Status='live', Symbol='Mapk8', Aliases='AI849689, JNK, JNK1, Prkm8, SAPK1', description='mitogen-activated protein kinase 8', other_designations='JNK1 beta1 protein kinase|MAP kinase 8|MAPK 8|c-Jun N-terminal kinase 1|mitogen activated protein kinase 8|protein kinase mitogen-activated 8|stress-activated protein kinase JNK1', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33377898, end_position_on_the_genomic_accession=33447214, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  26420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26420, CurrentID=0, Status='live', Symbol='Mapk9', Aliases='AI851083, JNK2, Prkm9, p54aSAPK', description='mitogen-activated protein kinase 9', other_designations='JNK/SAPK alpha|Jun kinase|MAP kinase 9|MAPK 9|c-Jun N-terminal kinase 2|mitogen activated protein kinase 9|protein kinase, mitogen-activated 9|stress-activated protein kinase JNK2', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49846751, end_position_on_the_genomic_accession=49886421, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  26421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26421, CurrentID=0, Status='live', Symbol='Prl2c4', Aliases='MRP-3, Mrpplf3, PLF-3, mrp/plf3', description='prolactin family 2, subfamily c, member 4', other_designations='mitogen regulated protein, proliferin 3', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.4', start_position_on_the_genomic_accession=12845940, end_position_on_the_genomic_accession=12855187, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26422, CurrentID=0, Status='live', Symbol='Nbea', Aliases='Lyst2, mKIAA1544', description='neurobeachin', other_designations='lysosomal trafficking regulator 2|lysosomal-trafficking regulator 2', map_location='3 26.61 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=55625195, end_position_on_the_genomic_accession=56183701, orientation='minus', exon_count=63, OMIM='', no_hdr0=''),
  26423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26423, CurrentID=0, Status='live', Symbol='Nr5a1', Aliases='Ad4BP, ELP, ELP-3, Ftz-F1, Ftzf1, SF-1, SF1, STF-1', description='nuclear receptor subfamily 5, group A, member 1', other_designations='adrenal 4-binding protein|embryonal LTR-binding protein|embryonal long terminal repeat-binding protein|fushi tarazu 1 factor homolog|fushi tarazu factor homolog 1|steroid hormone receptor Ad4BP|steroid hydroxylase positive regulator', map_location='2 24.42 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38692656, end_position_on_the_genomic_accession=38714560, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26424, CurrentID=0, Status='live', Symbol='Nr5a2', Aliases='AU020803, D1Ertd308e, Ftf, LRH-1, UF2-H3B', description='nuclear receptor subfamily 5, group A, member 2', other_designations='fetoprotein transcription factor|liver receptor homolog 1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136842562, end_position_on_the_genomic_accession=136960448, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26425, CurrentID=0, Status='live', Symbol='Nubp1', Aliases='', description='nucleotide binding protein 1', other_designations='', map_location='16 5.61 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10411878, end_position_on_the_genomic_accession=10424428, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  26426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26426, CurrentID=0, Status='live', Symbol='Nubp2', Aliases='D17Wsu11e', description='nucleotide binding protein 2', other_designations='', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24882611, end_position_on_the_genomic_accession=24886367, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26427, CurrentID=0, Status='live', Symbol='Creb3l1', Aliases='Oasis', description='cAMP responsive element binding protein 3-like 1', other_designations='BBF-2 homolog|cAMP-responsive element-binding protein 3-like protein 1|old astrocyte specifically induced substance|old astrocyte specifically-induced substance', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91982328, end_position_on_the_genomic_accession=92024170, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  26428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26428, CurrentID=0, Status='live', Symbol='Orc4', Aliases='Orc4Pl, mMmORC4, Orc4', description='origin recognition complex, subunit 4', other_designations='', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=48902754, end_position_on_the_genomic_accession=48950272, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  26429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26429, CurrentID=0, Status='live', Symbol='Orc5', Aliases='AL033327, MmORC5l, Orc5', description='origin recognition complex, subunit 5', other_designations='origin recognition complex 5|origin recognition complex, subunit 5 homolog', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=22486484, end_position_on_the_genomic_accession=22550367, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  26430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26430, CurrentID=0, Status='live', Symbol='Parg', Aliases='AI413217', description='poly (ADP-ribose) glycohydrolase', other_designations='cytosolic poly(ADP-ribose) glycohydrolase|mitochondrial poly(ADP-ribose) glycohydrolase 63 kDa isoform', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32201949, end_position_on_the_genomic_accession=32298714, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  26431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26431, CurrentID=0, Status='live', Symbol='Git2', Aliases='1500036H07Rik, 5830420E16Rik, 6430510B20Rik, 9630056M03Rik, B230104M05Rik, CAT2, mKIAA0148', description='G protein-coupled receptor kinase-interactor 2', other_designations='ARF GAP GIT2|ARF GTPase activating protein 2|GRK-interacting protein 2|cool-interacting tyrosine-phosphorylated protein 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114727408, end_position_on_the_genomic_accession=114773520, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  26432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26432, CurrentID=0, Status='live', Symbol='Plod2', Aliases='D530025C14Rik, LH2, Plod-2', description='procollagen lysine, 2-oxoglutarate 5-dioxygenase 2', other_designations='lysyl hydroxylase 2', map_location='9 48.4 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=92542223, end_position_on_the_genomic_accession=92608428, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  26433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26433, CurrentID=0, Status='live', Symbol='Plod3', Aliases='AI414586, LH3', description='procollagen-lysine, 2-oxoglutarate 5-dioxygenase 3', other_designations='lysyl hydroxylase 2|lysyl hydroxylase 3', map_location='5 76.09 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136987019, end_position_on_the_genomic_accession=136996646, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  26434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26434, CurrentID=0, Status='live', Symbol='Prnd', Aliases='AI450264, Dpl, PrPLP, doppel', description='prion protein dublet', other_designations='doppelganger|prion protein-like protein', map_location='2 64.07 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131950861, end_position_on_the_genomic_accession=131956131, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26436, CurrentID=0, Status='live', Symbol='Psg16', Aliases='Cea11, PSG, bCEA', description='pregnancy specific glycoprotein 16', other_designations='carcinoembryonic antigen 11', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17074013, end_position_on_the_genomic_accession=17133450, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  26437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26437, CurrentID=0, Status='live', Symbol='Psg17', Aliases='Cea-2, Cea2, mmCGM5', description='pregnancy specific glycoprotein 17', other_designations='carcinoembryonic antigen 2', map_location='7 9.33 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18813937, end_position_on_the_genomic_accession=18821591, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26438, CurrentID=0, Status='live', Symbol='Psg18', Aliases='Cea-3, Cea3, mmCGM6', description='pregnancy specific glycoprotein 18', other_designations='carcinoembryonic antigen 3|pregnancy specific glycoprotein 18; carcinoembryonic antigen 3', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18345744, end_position_on_the_genomic_accession=18355038, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26439, CurrentID=0, Status='live', Symbol='Psg19', Aliases='CGM7, Cea-4, Cea4', description='pregnancy specific glycoprotein 19', other_designations='carcinoembryonic antigen 4', map_location='7 9.31 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18789125, end_position_on_the_genomic_accession=18798542, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26440, CurrentID=0, Status='live', Symbol='Psma1', Aliases='C2, HC2, Pros-30, alpha-type', description='proteasome (prosome, macropain) subunit, alpha type 1', other_designations='20S proteasome subunit C2|macropain subunit C2|multicatalytic endopeptidase complex subunit C2|proteasome component C2|proteasome nu chain', map_location='7 59.32 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114264550, end_position_on_the_genomic_accession=114276116, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  26441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26441, CurrentID=0, Status='live', Symbol='Psma4', Aliases='C9', description='proteasome (prosome, macropain) subunit, alpha type 4', other_designations='macropain subunit C9|multicatalytic endopeptidase complex subunit C9|proteasome component C9|proteasome subunit L', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54950815, end_position_on_the_genomic_accession=54958042, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26442, CurrentID=0, Status='live', Symbol='Psma5', Aliases='ZETA', description='proteasome (prosome, macropain) subunit, alpha type 5', other_designations='macropain zeta chain|multicatalytic endopeptidase complex zeta chain|proteasome zeta chain', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108256926, end_position_on_the_genomic_accession=108279952, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26443, CurrentID=0, Status='live', Symbol='Psma6', Aliases='IOTA', description='proteasome (prosome, macropain) subunit, alpha type 6', other_designations='macropain iota chain|multicatalytic endopeptidase complex iota chain|proteasome iota chain', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55384217, end_position_on_the_genomic_accession=55418459, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  26444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26444, CurrentID=0, Status='live', Symbol='Psma7', Aliases='C6-I', description='proteasome (prosome, macropain) subunit, alpha type 7', other_designations='proteasome subunit RC6-1', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180036366, end_position_on_the_genomic_accession=180042464, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  26445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26445, CurrentID=0, Status='live', Symbol='Psmb2', Aliases='AU045357, AW108089, C7-I, D4Wsu33e, HC7-I', description='proteasome (prosome, macropain) subunit, beta type 2', other_designations='macropain subunit C7-I|multicatalytic endopeptidase complex subunit C7-I|proteasome component C7-I', map_location='4 60.9 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126677637, end_position_on_the_genomic_accession=126709715, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26446, CurrentID=0, Status='live', Symbol='Psmb3', Aliases='AL033320, C10-II', description='proteasome (prosome, macropain) subunit, beta type 3', other_designations='proteasome chain 13|proteasome component C10-II|proteasome theta chain', map_location='11 61.07 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97703434, end_position_on_the_genomic_accession=97713500, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26447, CurrentID=0, Status='live', Symbol='Poli', Aliases='Rad30b', description='polymerase (DNA directed), iota', other_designations='RAD30 homolog B', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=70508680, end_position_on_the_genomic_accession=70530334, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  26448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26448, CurrentID=0, Status='live', Symbol='Mok', Aliases='RAGE1, Rage, Stk30', description='MOK protein kinase', other_designations='MAPK/MAK/MRK/ overlapping kinase|T/STK 30|renal tumor antigen|serine/threonine kinase 30', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110807798, end_position_on_the_genomic_accession=110840957, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  26450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26450, CurrentID=0, Status='live', Symbol='Rbbp9', Aliases='Bog', description='retinoblastoma binding protein 9', other_designations='B5T overexpressed gene protein|B5T-overexpressed gene protein|RBBP-9|retinoblastoma-binding protein 9', map_location='2 G1-H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144542265, end_position_on_the_genomic_accession=144550859, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26451, CurrentID=0, Status='live', Symbol='Rpl27a', Aliases="L27', L27A, L29", description='ribosomal protein L27A', other_designations='ribosomal protein L29 homolog', map_location='7 E2-F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109519195, end_position_on_the_genomic_accession=109522369, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  26456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26456, CurrentID=0, Status='live', Symbol='Sema4g', Aliases='AI507908, AW554132', description='sema domain, immunoglobulin domain (Ig), transmembrane domain (TM) and short cytoplasmic domain, (semaphorin) 4G', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44988966, end_position_on_the_genomic_accession=45003397, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  26457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26457, CurrentID=0, Status='live', Symbol='Slc27a1', Aliases='FATP1, Fatp', description='solute carrier family 27 (fatty acid transporter), member 1', other_designations='FATP-1|fatty acid transport protein 1|solute carrier family 27 member 1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71568676, end_position_on_the_genomic_accession=71587304, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  26458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26458, CurrentID=0, Status='live', Symbol='Slc27a2', Aliases='ACSVL1, FATP2, VLCS, Vlac, Vlacs', description='solute carrier family 27 (fatty acid transporter), member 2', other_designations='FATP-2|THCA-CoA ligase|fatty acid transport protein 2|fatty-acid-coenzyme A ligase, very long-chain 1|long-chain-fatty-acid--CoA ligase|solute carrier family 27 member 2|very long-chain acyl-Coenzyme A dehydrogenase synthase|very long-chain-fatty-acid-CoA ligase', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126553024, end_position_on_the_genomic_accession=126588243, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26459, CurrentID=0, Status='live', Symbol='Slc27a5', Aliases='FACVL3, FATP5, VLCS-H2, VLCSH2, Vlacsr', description='solute carrier family 27 (fatty acid transporter), member 5', other_designations='BA-CoA ligase|BACS|BAL|FATP-5|VLACS-related|bile acid-CoA ligase|cholate--CoA ligase|fatty acid transport protein 5|solute carrier family 27 member 5|very long-chain acyl-CoA synthetase-related protein|very long-chain acyl-Coenzyme A dehydrogenase synthase related', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12988346, end_position_on_the_genomic_accession=12998192, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  26462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26462, CurrentID=0, Status='live', Symbol='Txnrd2', Aliases='AA118373, ESTM573010, TGR, Tr3, Trxr2, Trxrd2', description='thioredoxin reductase 2', other_designations='TR beta|thioredoxin reductase TR3', map_location='16 11.41 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18426417, end_position_on_the_genomic_accession=18479073, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  26464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26464, CurrentID=0, Status='live', Symbol='Vnn3', Aliases='', description='vanin 3', other_designations='vanin-3', map_location='10 A2-B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23851462, end_position_on_the_genomic_accession=23869843, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  26465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26465, CurrentID=0, Status='live', Symbol='Zfp146', Aliases='OZF, Znf146', description='zinc finger protein 146', other_designations='only zinc finger protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30160487, end_position_on_the_genomic_accession=30169779, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26466, CurrentID=0, Status='live', Symbol='Zfp260', Aliases='AI480997, Ozrf1, PEX1, Zfp63, Znf260', description='zinc finger protein 260', other_designations='zfp-260', map_location='7 17.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30094782, end_position_on_the_genomic_accession=30107622, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26549, CurrentID=0, Status='live', Symbol='Itgb1bp2', Aliases='Chordc3', description='integrin beta 1 binding protein 2', other_designations='melusin', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101448690, end_position_on_the_genomic_accession=101466122, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  26554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26554, CurrentID=0, Status='live', Symbol='Cul3', Aliases='KIAA0617', description='cullin 3', other_designations='', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=80265610, end_position_on_the_genomic_accession=80340762, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  26556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26556, CurrentID=0, Status='live', Symbol='Homer1', Aliases='PSD-Zip45, SYN47, Ves-1, homer-1, vesl-1', description='homer scaffolding protein 1', other_designations='VASP/Ena-related gene up-regulated during seizure and LTP 1|homer homolog 1|homer, neuronal immediate early gene, 1|immediate early gene protein Homer1A', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93304495, end_position_on_the_genomic_accession=93405129, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  26557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26557, CurrentID=0, Status='live', Symbol='Homer2', Aliases='9330120H11Rik, AW539445, CPD, Vesl-2', description='homer scaffolding protein 2', other_designations='VASP/Ena-related gene up-regulated during seizure and LTP 2|cupidin|homer homolog 2|homer, neuronal immediate early gene, 2|homer-2', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81600481, end_position_on_the_genomic_accession=81706925, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26558, CurrentID=0, Status='live', Symbol='Homer3', Aliases='AW146114', description='homer scaffolding protein 3', other_designations='homer homolog 3|homer, neuronal immediate early gene, 3|homer-3', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70282891, end_position_on_the_genomic_accession=70294361, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  26559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26559, CurrentID=0, Status='live', Symbol='Hunk', Aliases='Bstk1, Mak-v', description='hormonally upregulated Neu-associated kinase', other_designations='serine/threonine-protein kinase MAK-V', map_location='16 51.71 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90385797, end_position_on_the_genomic_accession=90499555, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  26560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26560, CurrentID=0, Status='live', Symbol='Krtap15', Aliases='Krtap15-1, Pmg-2, Pmg2', description='keratin associated protein 15', other_designations='keratin-associated protein 15|krtap15-1|pubertal mammary gland-specific protein 2', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88829009, end_position_on_the_genomic_accession=88829844, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26561, CurrentID=0, Status='live', Symbol='Mmp23', Aliases='CA-MMP', description='matrix metallopeptidase 23', other_designations='CAMP metalloproteinase|cysteine array matrix metalloproteinase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155650655, end_position_on_the_genomic_accession=155654481, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26562, CurrentID=0, Status='live', Symbol='Ncdn', Aliases='AU042419, MMS10-AE, Ms10ae, mKIAA0607, norbin', description='neurochondrin', other_designations='M-Sema F-associating protein of 75 kDa|minisatellite 10ae detected by probe MMS10|neurochondrin-1|neurochondrin-2', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126743750, end_position_on_the_genomic_accession=126753566, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26563, CurrentID=0, Status='live', Symbol='Ror1', Aliases='2810404D04Rik, Ntrkr1', description='receptor tyrosine kinase-like orphan receptor 1', other_designations='neurotrophic tyrosine kinase, receptor-related 1|tyrosine-protein kinase transmembrane receptor ROR1', map_location='4 45.71 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=100095791, end_position_on_the_genomic_accession=100444797, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  26564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26564, CurrentID=0, Status='live', Symbol='Ror2', Aliases='Ntrkr2, mRor2', description='receptor tyrosine kinase-like orphan receptor 2', other_designations='neurotrophic tyrosine kinase, receptor related 2|neurotrophic tyrosine kinase, receptor-related 2', map_location='13 27.68 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=53109317, end_position_on_the_genomic_accession=53286109, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  26565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26565, CurrentID=0, Status='live', Symbol='Pla2g10', Aliases='GX sPLA2, PLA2GX, sPLA2-X', description='phospholipase A2, group X', other_designations='group X secretory phospholipase A2|phosphatidylcholine 2-acylhydrolase 10', map_location='16 9.5 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13715057, end_position_on_the_genomic_accession=13730983, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26568, CurrentID=0, Status='live', Symbol='Slc27a3', Aliases='Acsvl3, FATP-3, FATP3, Vlcs-3', description='solute carrier family 27 (fatty acid transporter), member 3', other_designations='mFLJ00207 protein|solute carrier family 27 member 3|very long-chain acyl-CoA synthetase homolog 3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90385228, end_position_on_the_genomic_accession=90390042, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  26569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26569, CurrentID=0, Status='live', Symbol='Slc27a4', Aliases='BB144259, FATP4', description='solute carrier family 27 (fatty acid transporter), member 4', other_designations='FATP-4|fatty acid transport protein 4|solute carrier family 27 member 4', map_location='2 20.64 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29802680, end_position_on_the_genomic_accession=29817522, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  26570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26570, CurrentID=0, Status='live', Symbol='Slc7a11', Aliases='9930009M05Rik, AI451155, sut, xCT', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 11', other_designations='amino acid transport system xc-|cysteine/glutamate transporter|sodium independent anionic amino acid transport system|solute carrier family 7 member 11', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=50364936, end_position_on_the_genomic_accession=50499087, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  26572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26572, CurrentID=0, Status='live', Symbol='Cops3', Aliases='Csn3, Sgn3', description='COP9 signalosome subunit 3', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 3|COP9 (constitutive photomorphogenic), subunit 3|COP9 complex S3|JAB1-containing signalosome subunit 3|signalosome subunit 3', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59817795, end_position_on_the_genomic_accession=59839844, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  26611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26611, CurrentID=0, Status='live', Symbol='Rcn2', Aliases='AA408742, Tcbp49', description='reticulocalbin 2', other_designations='taipoxin-associated calcium binding protein 49|taipoxin-associated calcium-binding protein 49', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56041845, end_position_on_the_genomic_accession=56059083, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69987, CurrentID=0, Status='live', Symbol='1700026L06Rik', Aliases='MAST', description='RIKEN cDNA 1700026L06 gene', other_designations='acrosome and sperm tail protein', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28692080, end_position_on_the_genomic_accession=28699651, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26754, CurrentID=0, Status='live', Symbol='Cops5', Aliases='AI303502, CSN5, Jab1, Mov34, Sgn5', description='COP9 signalosome subunit 5', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 5|COP9 complex S5|Jun activation domain-binding protein 1|Jun coactivator|Kip1 C-terminus-interacting protein 2', map_location='1 2.29 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=10024600, end_position_on_the_genomic_accession=10038159, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  26757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26757, CurrentID=0, Status='live', Symbol='Dpysl4', Aliases='AI173505, CRMP-3, Crmp3, DPY4, Drp-4, Ulip4', description='dihydropyrimidinase-like 4', other_designations='ULIP-4|UNC33-like phosphoprotein 4|collapsin response mediator protein 3|unc-33-like phosphoprotein 4', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139085985, end_position_on_the_genomic_accession=139101795, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100034726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034726, CurrentID=0, Status='live', Symbol='Gm15772', Aliases='ENSMUSG00000062353, OTTMUSG00000026108', description='predicted gene 15772', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3236390, end_position_on_the_genomic_accession=3236928, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100034784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034784, CurrentID=0, Status='live', Symbol='Hrtfm2', Aliases='Hfm2', description='heart failure modifier 2', other_designations='', map_location='3 72.28 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100034793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034793, CurrentID=0, Status='live', Symbol='Hdlq11', Aliases='', description='HDL QTL 11', other_designations='', map_location='6 48.93 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26874, CurrentID=0, Status='live', Symbol='Abcd2', Aliases='ABC39, ALDL1, ALDR, ALDRP', description='ATP-binding cassette, sub-family D (ALD), member 2', other_designations='adrenoleukodystrophy related|adrenoleukodystrophy-related protein', map_location='15 E-F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=91145871, end_position_on_the_genomic_accession=91192504, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  26875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26875, CurrentID=0, Status='live', Symbol='Pclo', Aliases='Acz, Pico', description='piccolo (presynaptic cytomatrix protein)', other_designations='aczonin|brain-derived HLMN protein|multidomain presynaptic cytomatrix protein', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=14514918, end_position_on_the_genomic_accession=14863465, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  26876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26876, CurrentID=0, Status='live', Symbol='Adh4', Aliases='Adh2', description='alcohol dehydrogenase 4 (class II), pi polypeptide', other_designations='alcohol dehydrogenase II|alcohol dehydrogenase class II|class II type ADH', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138415497, end_position_on_the_genomic_accession=138430892, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26877, CurrentID=0, Status='live', Symbol='B3galt1', Aliases='beta3Gal-T1', description='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 1', other_designations='UDP-galactose:beta-N-acetylglucosamine beta1, 3-galactosyltransferase 1|beta-1,3-GalTase 1', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=67565962, end_position_on_the_genomic_accession=68122882, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  26878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26878, CurrentID=0, Status='live', Symbol='B3galt2', Aliases='', description='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 2', other_designations='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase-II|beta-1,3-GalTase 2|beta3Gal-T2|beta3GalT2', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=143640697, end_position_on_the_genomic_accession=143649937, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  26879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26879, CurrentID=0, Status='live', Symbol='B3galnt1', Aliases='B3galt3, b3GT3', description='UDP-GalNAc:betaGlcNAc beta 1,3-galactosaminyltransferase, polypeptide 1', other_designations='Brainiac 1|Globoside blood group|Mbrn 1|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase III|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 3|UDP-N-acetylgalactosamine:globotriaosylceramide beta-1,3-N-acetylgalactosaminyltransferase|b3Gal-T3|beta-1,3-GalNAc-T1|beta-1,3-GalTase 3|beta-1,3-galactosyltransferase 3|beta-3-Gx-T3|beta3GalT3|galactosylgalactosylglucosylceramide beta-D-acetyl-galactosaminyltransferase|globoside synthase', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69573919, end_position_on_the_genomic_accession=69598960, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26885, CurrentID=0, Status='live', Symbol='Casp8ap2', Aliases='D4Ertd659e, FLASH', description='caspase 8 associated protein 2', other_designations='FLICE-associated huge protein', map_location='4 14.27 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=32615473, end_position_on_the_genomic_accession=32653271, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  26886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26886, CurrentID=0, Status='live', Symbol='Cenph', Aliases='1700021I11Rik, 2410018A12Rik, 2610042E16Rik, 2810046K12Rik, AU044255, CENP-H, ENP', description='centromere protein H', other_designations='cenph|centromere autoantigen H', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100759686, end_position_on_the_genomic_accession=100775899, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  26887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26887, CurrentID=0, Status='live', Symbol='Chst4', Aliases='GST-3, Gn6st-2, HEC-GlcNAc6ST', description='carbohydrate (chondroitin 6/keratan) sulfotransferase 4', other_designations='L-selectin ligand sulfotransferase|LSST|N-acetylglucosamine 6-O-sulfotransferase 2|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 3|glcNAc6ST-2|high endothelial cell GlcNAC-6-sulphotransferase|high endothelial cells N-acetylglucosamine 6-O-sulfotransferase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110029075, end_position_on_the_genomic_accession=110039375, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  26888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26888, CurrentID=0, Status='live', Symbol='Clec4a2', Aliases='Clec4a, Clecsf6, DCIR, Dcir1', description='C-type lectin domain family 4, member a2', other_designations='C-type (calcium dependent, carbohydrate recognition domain) lectin, superfamily member 6|C-type lectin superfamily member 6|dendritic cell immunoreceptor', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=123122690, end_position_on_the_genomic_accession=123143999, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  26889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26889, CurrentID=0, Status='live', Symbol='Cln8', Aliases='mnd', description='ceroid-lipofuscinosis, neuronal 8', other_designations='motor neuron degeneration', map_location='8 7.59 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=14881335, end_position_on_the_genomic_accession=14901720, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  26891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26891, CurrentID=0, Status='live', Symbol='Cops4', Aliases='AW208976, D5Ertd774e', description='COP9 signalosome subunit 4', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 4|COP9 (constitutive photomorphogenic), subunit 4|COP9 complex S4|JAB1-containing signalosome subunit 4|SGN4|signalosome subunit 4', map_location='5 48.49 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100518309, end_position_on_the_genomic_accession=100547802, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  26893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26893, CurrentID=0, Status='live', Symbol='Cops6', Aliases='Sgn3, VIP/MOV34', description='COP9 signalosome subunit 6', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 6|COP9 (constitutive photomorphogenic), subunit 6|COP9 complex S6|JAB1-containing signalosome subunit 6|SGN6|signalosome subunit 6', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138161102, end_position_on_the_genomic_accession=138163984, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26894, CurrentID=0, Status='live', Symbol='Cops7a', Aliases='D6Ertd35e', description='COP9 signalosome subunit 7A', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 7a|COP9 (constitutive photomorphogenic), subunit 7a|COP9 complex S7a|JAB1-containing signalosome subunit 7a|SGN7a|signalosome subunit 7a', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124958411, end_position_on_the_genomic_accession=124965529, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26895, CurrentID=0, Status='live', Symbol='Cops7b', Aliases='D1Wsu66e, E130114M23', description='COP9 signalosome subunit 7B', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 7b|COP9 complex S7b|JAB1-containing signalosome subunit 7b|SGN7b|signalosome subunit 7b', map_location='1 43.97 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86582902, end_position_on_the_genomic_accession=86606500, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  26896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26896, CurrentID=0, Status='live', Symbol='Med14', Aliases='9930001L01Rik, AU041628, Crsp2, ENSMUSG00000073278, Gm641, ORF1, Trap170', description='mediator complex subunit 14', other_designations='CRSP complex subunit 2|cofactor required for Sp1 transcriptional activation subunit 2 (150 kDa)|cofactor required for Sp1 transcriptional activation, subunit 2|thyroid hormone receptor-associated protein complex 170 kDa component', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=12675368, end_position_on_the_genomic_accession=12762042, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  26897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26897, CurrentID=0, Status='live', Symbol='Acot1', Aliases='ACH2, CTE-1, CTE-I, Cte1, D12Ucla1', description='acyl-CoA thioesterase 1', other_designations='acyl-CoA thioesterase 1, cytosolic|cytosolic acyl-CoA thioesterase 1|inducible cytosolic acyl-coenzyme A thioester hydrolase|long chain acyl-CoA hydrolase|long chain acyl-CoA thioester hydrolase', map_location='12 39.0 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84009502, end_position_on_the_genomic_accession=84017670, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26898, CurrentID=0, Status='live', Symbol='Ctsj', Aliases='CATP, CatRLP, Ctsp', description='cathepsin J', other_designations='Cat P|cathepsin L-related protein|cathepsin P|catlrp-p|gene/Cathepsin L-related protein', map_location='13 32.62 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61000277, end_position_on_the_genomic_accession=61005958, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  26900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26900, CurrentID=0, Status='live', Symbol='Ddx3y', Aliases='8030469F12Rik, D1Pas1-rs1, Dby', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 3, Y-linked', other_designations='D1Pas1 related sequence 1|D1Pas1-related sequence 1|DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide, Y chromosome|DEAD box protein 3, Y-chromosomal|DEAD-box RNA helicase DEAD2|mDEAD2', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=1260715, end_position_on_the_genomic_accession=1286628, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  26901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26901, CurrentID=0, Status='live', Symbol='Deb1', Aliases='1110020I04Rik, AI463209, Ss18l2', description='differentially expressed in B16F10 1', other_designations='Band 47B', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121710389, end_position_on_the_genomic_accession=121712921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  26903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26903, CurrentID=0, Status='live', Symbol='Dysf', Aliases='2310004N10Rik, AI604795, D6Pas3', description='dysferlin', other_designations='dysferlin variant a|dysferlin_a|dystrophy-associated fer-1-like protein|fer-1-like protein 1', map_location='6 36.14 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=84008316, end_position_on_the_genomic_accession=84211060, orientation='plus', exon_count=63, OMIM='', no_hdr0=''),
  26904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26904, CurrentID=0, Status='live', Symbol='Sh2d1b1', Aliases='EAT-2, EAT-2A, Eat2, Eat2a, Sh2d1b', description='SH2 domain containing 1B1', other_designations='EWS/FLI1 activated transcript 2|EWS/FLI1-activated transcript 2|SH2 domain protein 1B1', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170277324, end_position_on_the_genomic_accession=170285662, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  26905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26905, CurrentID=0, Status='live', Symbol='Eif2s3x', Aliases='AA409828, AA547477, AI314668, Eif-2gx', description='eukaryotic translation initiation factor 2, subunit 3, structural gene X-linked', other_designations='eIF-2-gamma X|eukaryotic translation initiation factor 2 subunit gamma, X-linked', map_location='X C-D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94188709, end_position_on_the_genomic_accession=94212651, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  26908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26908, CurrentID=0, Status='live', Symbol='Eif2s3y', Aliases='Eif-2gy, Spy, Tfy', description='eukaryotic translation initiation factor 2, subunit 3, structural gene Y-linked', other_designations='eIF-2-gamma Y|eukaryotic translation initiation factor 2 subunit gamma, Y-linked|spermatogenesis Y', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=1010578, end_position_on_the_genomic_accession=1046811, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  26909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26909, CurrentID=0, Status='live', Symbol='Exo1', Aliases='5730442G03Rik, Msa', description='exonuclease 1', other_designations='exonuclease I|mExo1', map_location='1 81.9 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175880761, end_position_on_the_genomic_accession=175911396, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  26910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26910, CurrentID=0, Status='live', Symbol='Figla', Aliases='FIGalpha, bHLHc8', description='folliculogenesis specific basic helix-loop-helix', other_designations='FIG alpha|transcription factor FIGa', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86010320, end_position_on_the_genomic_accession=86020996, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26912, CurrentID=0, Status='live', Symbol='Gcat', Aliases='AI526977, Kbl', description='glycine C-acetyltransferase (2-amino-3-ketobutyrate-coenzyme A ligase)', other_designations='AKB ligase|aminoacetone synthase|glycine acetyltransferase', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79030874, end_position_on_the_genomic_accession=79042531, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26913, CurrentID=0, Status='live', Symbol='Gprin1', Aliases='GRIN1, Z16', description='G protein-regulated inducer of neurite outgrowth 1', other_designations='', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54736671, end_position_on_the_genomic_accession=54750150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26914, CurrentID=0, Status='live', Symbol='H2afy', Aliases='H2AF12M, mH2a1, macroH2A1', description='H2A histone family, member Y', other_designations='H2A.y|H2A/y|MACROH2A1.2|histone macroH2A1', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56073622, end_position_on_the_genomic_accession=56135626, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100034853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034853, CurrentID=0, Status='live', Symbol='Alaa1', Aliases='', description='alopecia areata', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26918, CurrentID=0, Status='live', Symbol='Ern2', Aliases='Ern1, Ire1, Ire1b, ire1-beta, mIre1', description='endoplasmic reticulum (ER) to nucleus signalling 2', other_designations='endoplasmic reticulum-to-nucleus signaling 2|inositol-requiring 1 (Yeast homologue)|inositol-requiring 1 beta|inositol-requiring protein 2', map_location='7 65.53 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122169745, end_position_on_the_genomic_accession=122186226, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  26919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26919, CurrentID=0, Status='live', Symbol='Zfp346', Aliases='Jaz, Znf346', description='zinc finger protein 346', other_designations='just another zinc finger protein', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55105271, end_position_on_the_genomic_accession=55135071, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  26920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26920, CurrentID=0, Status='live', Symbol='Cntrl', Aliases='6720467O09Rik, AA968343, Cep1, Cep110, IB3/5, Ma2a8, b2b1468.1Clo, b2b1468Clo', description='centriolin', other_designations='110 kDa centrosomal protein|antigen identified by monoclonal antibody 2A8|centrosomal protein 110|centrosomal protein of 110 kDa', map_location='2 B-C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35109492, end_position_on_the_genomic_accession=35178822, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  26921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26921, CurrentID=0, Status='live', Symbol='Map4k4', Aliases='9430080K19Rik, AU043147, AU045934, AW046177, HGK, Nik', description='mitogen-activated protein kinase kinase kinase kinase 4', other_designations='HPK/GCK-like kinase HGK|MAPK/ERK kinase kinase kinase 4|MEK kinase kinase 4|MEKKK 4|NCK interacting kinase|mitogen activated protein kinase kinase kinase kinase 4|nck-interacting kinase', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39900912, end_position_on_the_genomic_accession=40026310, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  26922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26922, CurrentID=0, Status='live', Symbol='Mecr', Aliases='AI195831, Nrbf1', description='mitochondrial trans-2-enoyl-CoA reductase', other_designations='nuclear receptor binding factor 1', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=131843447, end_position_on_the_genomic_accession=131867787, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  26926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26926, CurrentID=0, Status='live', Symbol='Aifm1', Aliases='AIF, AIFsh2, Hq, Pdcd8', description='apoptosis-inducing factor, mitochondrion-associated 1', other_designations='harlequin|programmed cell death protein 8', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48474944, end_position_on_the_genomic_accession=48513563, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  26927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26927, CurrentID=0, Status='live', Symbol='Foxl2', Aliases='AU045128, BPES, PINTO, Pfrk', description='forkhead box L2', other_designations='P-Frk|pituitary forkhead factor|putative transcription factor foxl2', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98955607, end_position_on_the_genomic_accession=98958126, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  26931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26931, CurrentID=0, Status='live', Symbol='Ppp2r5c', Aliases='2610043M05Rik, 2700063L20Rik, AI060890, AW545884, C85228, D12Bwg0916e, mKIAA0044', description="protein phosphatase 2, regulatory subunit B', gamma", other_designations="B56/PP2A gamma|Band 8A|PP2A B subunit isoform B'-alpha-3|PP2A B subunit isoform B'-gamma|PP2A B subunit isoform B56-gamma|PP2A B subunit isoform PR61-gamma|PP2A B subunit isoform R5-gamma|protein phosphatase 2, regulatory subunit B (B56), gamma isoform", map_location='12 60.56 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110447179, end_position_on_the_genomic_accession=110583075, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100034868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034868, CurrentID=0, Status='live', Symbol='Sgp3', Aliases='', description='serum gp70 production 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26934, CurrentID=0, Status='live', Symbol='Racgap1', Aliases='AI227039, AI327394, Band25, GTPase, MgcRacGAP, gtl11, mKIAA1478', description='Rac GTPase-activating protein 1', other_designations='male germ cell RacGap|racgap1', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99620496, end_position_on_the_genomic_accession=99651656, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  26936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26936, CurrentID=0, Status='live', Symbol='Mprip', Aliases='9530046C02, AA536749, AI647711, C76423, RIP3, Rhoip3, mKIAA0864, p116Rip', description='myosin phosphatase Rho interacting protein', other_designations='Rho interacting protein 3|Rho-interacting protein 3 (p116RIP) (RIP3)|p116 Rho interacting protein', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59662473, end_position_on_the_genomic_accession=59780860, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  26938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26938, CurrentID=0, Status='live', Symbol='St6galnac5', Aliases='AI851940, Siat7e', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 5', other_designations='GD1 alpha synthase|SIAT7-E|ST6GalNAc V|ST6GalNAcV|galNAc alpha-2,6-sialyltransferase V|sialyltransferase 7 ((alpha-N-acetylneuraminyl 2,3-betagalactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialyltransferase) E|sialyltransferase 7E', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152820709, end_position_on_the_genomic_accession=152982207, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26939, CurrentID=0, Status='live', Symbol='Polr3e', Aliases='RPC5, Sin', description='polymerase (RNA) III (DNA directed) polypeptide E', other_designations='RNA polymerase III polypeptide E|RNA polymerase III subunit 5|RNA polymerase III subunit C5|sex-lethal interactor homolog', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120917739, end_position_on_the_genomic_accession=120947432, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  26940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26940, CurrentID=0, Status='live', Symbol='Ecsit', Aliases='Sitpec', description='ECSIT signalling integrator', other_designations='ECSIT homolog|signaling intermediate in Toll pathway-evolutionarily conserved', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22071108, end_position_on_the_genomic_accession=22085443, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  26941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26941, CurrentID=0, Status='live', Symbol='Slc9a3r1', Aliases='EBP-50, NHE-RF, NHERF-1, NHERF1', description='solute carrier family 9 (sodium/hydrogen exchanger), member 3 regulator 1', other_designations='EBP50|ezrin-radixin-moesin-binding phosphoprotein 50|regulatory cofactor of Na(+)/H(+) exchanger|sodium-hydrogen exchanger regulatory factor 1|solute carrier family 9 (sodium/hydrogen exchanger), isoform 3 regulator 1|solute carrier family 9 isoform A3 regulatory factor 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115163312, end_position_on_the_genomic_accession=115181183, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26942, CurrentID=0, Status='live', Symbol='Spag1', Aliases='tpis', description='sperm associated antigen 1', other_designations='TPR-containing protein involved in spermatogenesis|infertility-related sperm protein Spag-1', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36175483, end_position_on_the_genomic_accession=36235621, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  26943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26943, CurrentID=0, Status='live', Symbol='Serinc3', Aliases='AA959945, AA960047, AIGP1, DIFF33, TMS-1, Tde1', description='serine incorporator 3', other_designations='AIGP-1|axotomy-induced glyco/Golgi protein 1|axotomy-induced glycoprotein 1|membrane protein TMS-1|tumor differentially expressed 1|tumor differentially expressed protein 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163623273, end_position_on_the_genomic_accession=163645151, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26944, CurrentID=0, Status='live', Symbol='Tinag', Aliases='AI452335, TIN-ag', description='tubulointerstitial nephritis antigen', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=76951687, end_position_on_the_genomic_accession=77045806, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  26945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26945, CurrentID=0, Status='live', Symbol='Tpsg1', Aliases='Prss31, Tmt', description='tryptase gamma 1', other_designations='transmembrane tryptase', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25369216, end_position_on_the_genomic_accession=25374442, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  26946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26946, CurrentID=0, Status='live', Symbol='Trpc7', Aliases='TRP7, Trrp8', description='transient receptor potential cation channel, subfamily C, member 7', other_designations='transient receptor potential protein homologue TRP7|transient receptor protein 7|transient receptor protein 8', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56773098, end_position_on_the_genomic_accession=56895974, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  26949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26949, CurrentID=0, Status='live', Symbol='Vat1', Aliases='VAT-1', description='vesicle amine transport protein 1 homolog (T californica)', other_designations='membrane protein of cholinergic synaptic vesicles', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101458748, end_position_on_the_genomic_accession=101466199, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  26950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26950, CurrentID=0, Status='live', Symbol='Vsnl1', Aliases='VILIP, Vnsl1', description='visinin-like 1', other_designations='NVL-1|NVP-1|neural visinin-like protein 1', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=11325245, end_position_on_the_genomic_accession=11436649, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  26951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26951, CurrentID=0, Status='live', Symbol='Zw10', Aliases='6330566F14Rik, MmZw10', description='zw10 kinetochore protein', other_designations='ZW10 homolog, centromere/kinetochore protein', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=49055581, end_position_on_the_genomic_accession=49078775, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100034888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034888, CurrentID=0, Status='live', Symbol='Thcr', Aliases='', description='T helper cell response', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26961, CurrentID=0, Status='live', Symbol='Rpl8', Aliases='', description='ribosomal protein L8', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76904071, end_position_on_the_genomic_accession=76906318, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  26965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26965, CurrentID=0, Status='live', Symbol='Cul1', Aliases='', description='cullin 1', other_designations='CUL-1', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47452273, end_position_on_the_genomic_accession=47526139, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  26968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26968, CurrentID=0, Status='live', Symbol='Islr', Aliases='AW553146', description='immunoglobulin superfamily containing leucine-rich repeat', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58156246, end_position_on_the_genomic_accession=58159239, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  26970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26970, CurrentID=0, Status='live', Symbol='Pla2g2e', Aliases='', description='phospholipase A2, group IIE', other_designations='GIIE sPLA2|mGIIEsPLA2s|phosphatidylcholine 2-acylhydrolase 2E|sPLA2-IIE', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138877911, end_position_on_the_genomic_accession=138882817, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  26971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26971, CurrentID=0, Status='live', Symbol='Pla2g2f', Aliases='', description='phospholipase A2, group IIF', other_designations='GIIF sPLA2|mGIIFsPLA2s|phosphatidylcholine 2-acylhydrolase 2F|sPLA2-IIF', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138750533, end_position_on_the_genomic_accession=138757598, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  26972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26972, CurrentID=0, Status='live', Symbol='Spo11', Aliases='AI449549', description='SPO11 meiotic protein covalently bound to DSB', other_designations='SPO11 meiotic protein covalently bound to DSB homolog|sporulation protein, meiosis-specific, SPO11 homolog (S. cervisiae)', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172977222, end_position_on_the_genomic_accession=172993576, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  26987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26987, CurrentID=0, Status='live', Symbol='Eif4e2', Aliases='2700069E09Rik, AI036339, AV129531, D0H0S6743E, Eif4el3', description='eukaryotic translation initiation factor 4E member 2', other_designations='eIF-4E type 2|eIF4E type 2|eIF4E-2|eIF4E-like protein 4E-LP|eukaryotic translation initiation factor 4E like 3|eukaryotic translation initiation factor 4E-like 3|mRNA cap-binding protein type 2', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87213910, end_position_on_the_genomic_accession=87240488, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  26992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26992, CurrentID=0, Status='live', Symbol='Brd7', Aliases='BP75, CELTIX1, Ptpn13ip', description='bromodomain containing 7', other_designations='75 kDa bromodomain protein|bromodomain protein 75 kDa|bromodomain-containing 7|protein tyrosine phosphatase, non-receptor type 13 interacting protein', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88330598, end_position_on_the_genomic_accession=88362255, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  27007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27007, CurrentID=0, Status='live', Symbol='Klrk1', Aliases='D6H12S2489E, NKG2-D, Nkg2d', description='killer cell lectin-like receptor subfamily K, member 1', other_designations='NK cell receptor D|NKG2-D-activating NK receptor|natural killer cell group 2D', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129610323, end_position_on_the_genomic_accession=129623864, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  27008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27008, CurrentID=0, Status='live', Symbol='Micall1', Aliases='AI448246, D15Mit260, D15N2e, EST573322, EST820961, MIRab13, mKIAA1668', description='microtubule associated monooxygenase, calponin and LIM domain containing -like 1', other_designations='microtubule associated monoxygenase, calponin and LIM domain containing -like 1|molecule interacting with Rab13', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79108477, end_position_on_the_genomic_accession=79136900, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  27009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27009, CurrentID=0, Status='live', Symbol='D17Mit232.1', Aliases='D17Mit232', description='DNA Segment, Chr 17, Massachusetts Institute of Technology 232.1', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27015, CurrentID=0, Status='live', Symbol='Polk', Aliases='Dinb1', description='polymerase (DNA directed), kappa', other_designations='DINB protein|DINP|DNA damage-inducible protein b|DNA-directed DNA polymerase kappa|DinB homolog 1', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=96480689, end_position_on_the_genomic_accession=96542648, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  27027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27027, CurrentID=0, Status='live', Symbol='Tspan32', Aliases='AW208513, Art-1, BB235973, D7Wsu37e, Phemx, Tssc6', description='tetraspanin 32', other_designations='AML1-regulated transmembrane protein 1|pan hematopoietic expression|protein Phemx|tspan-32|tumor-suppressing subchromosomal transferable fragment 6', map_location='7 88.09 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143004902, end_position_on_the_genomic_accession=143019486, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27028, CurrentID=0, Status='live', Symbol='Ermap', Aliases='AA409279, AI666418', description='erythroblast membrane-associated protein', other_designations='', map_location='4 55.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119175457, end_position_on_the_genomic_accession=119190013, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  27029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27029, CurrentID=0, Status='live', Symbol='Sgsh', Aliases='4632406A19Rik', description='N-sulfoglucosamine sulfohydrolase (sulfamidase)', other_designations='sulphamidase', map_location='11 83.36 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119343489, end_position_on_the_genomic_accession=119355510, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100034970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034970, CurrentID=0, Status='live', Symbol='Hyplip2', Aliases='', description='hyperlipidemia 2', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100034974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034974, CurrentID=0, Status='live', Symbol='Mastr', Aliases='', description='modifier of astrocytoma', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27041, CurrentID=0, Status='live', Symbol='G3bp1', Aliases='AI849976, B430204O07, C87777, G3bp, mKIAA4115', description='GTPase activating protein (SH3 domain) binding protein 1', other_designations='ATP-dependent DNA helicase VIII|G3BP-1|GAP SH3 binding protein|GAP SH3 domain-binding protein 1|HDH-VIII|Ras-GTPase-activating protein SH3-domain binding protein 1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55469752, end_position_on_the_genomic_accession=55500888, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  27045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27045, CurrentID=0, Status='live', Symbol='Nit1', Aliases='AI255805, ESTM30, W57327', description='nitrilase 1', other_designations='', map_location='1 79.35 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171342237, end_position_on_the_genomic_accession=171345645, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27047, CurrentID=0, Status='live', Symbol='Omd', Aliases='OSAD, SLRR2C', description='osteomodulin', other_designations='KSPG osteomodulin|keratan sulfate proteoglycan osteomodulin|osteoadherin', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49579754, end_position_on_the_genomic_accession=49592612, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  27049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27049, CurrentID=0, Status='live', Symbol='Etv3', Aliases='AI414410, METS, Pe1', description='ets variant 3', other_designations='ETS domain transcriptional repressor PE1|ETS-domain protein|ETS-domain transcriptional repressor|ets variant gene 3|mitogenic Ets transcriptional suppressor METS', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87525577, end_position_on_the_genomic_accession=87540158, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  27050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27050, CurrentID=0, Status='live', Symbol='Rps3', Aliases='D7Ertd795e, Rs_3', description='ribosomal protein S3', other_designations='', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99477896, end_position_on_the_genomic_accession=99483709, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27052, CurrentID=0, Status='live', Symbol='Aoah', Aliases='4930433E13Rik', description='acyloxyacyl hydrolase', other_designations='', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=20794103, end_position_on_the_genomic_accession=21036617, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  27053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27053, CurrentID=0, Status='live', Symbol='Asns', Aliases='', description='asparagine synthetase', other_designations='glutamine-dependent asparagine synthetase', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=7675165, end_position_on_the_genomic_accession=7693182, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  27054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27054, CurrentID=0, Status='live', Symbol='Sec23b', Aliases='', description='SEC23 homolog B, COPII coat complex component', other_designations='SEC23-related protein B|SEC23B', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144556229, end_position_on_the_genomic_accession=144590753, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  27055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27055, CurrentID=0, Status='live', Symbol='Fkbp9', Aliases='1810014L23, AW549641, FKBP-63, FKBP-9, FKBP60, FKBP63, FKBP65RS', description='FK506 binding protein 9', other_designations='63 kDa FK506-binding protein|63 kDa FKBP|FK506-binding protein 9|PPIase FKBP9|rotamase', map_location='6 27.74 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56832059, end_position_on_the_genomic_accession=56879360, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27056, CurrentID=0, Status='live', Symbol='Irf5', Aliases='AW491843, mirf5', description='interferon regulatory factor 5', other_designations='interferon regulatory factor 5 bone marrow', map_location='6 12.36 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29526130, end_position_on_the_genomic_accession=29541851, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  27057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27057, CurrentID=0, Status='live', Symbol='Ncoa4', Aliases='ARA70, Rfg', description='nuclear receptor coactivator 4', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32159887, end_position_on_the_genomic_accession=32179855, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  27058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27058, CurrentID=0, Status='live', Symbol='Srp9', Aliases='9kDa', description='signal recognition particle 9', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182124737, end_position_on_the_genomic_accession=182132415, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  27059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27059, CurrentID=0, Status='live', Symbol='Sh3d19', Aliases='AW011754, Kryn', description='SH3 domain protein D19', other_designations='SH3 domain containing protein D 19', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=85970956, end_position_on_the_genomic_accession=86130522, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  27060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27060, CurrentID=0, Status='live', Symbol='Tcirg1', Aliases='ATP6N1C, ATP6a3, Atp6i, OC-116, OPTB1, Stv1, TIRC7, Vph1, oc', description='T cell, immune regulator 1, ATPase, H+ transporting, lysosomal V0 protein A3', other_designations='ATPase, H+ transporting, lysosomal I|T-cell, immune regulator 1, ATPase, H+ transporting, lysosomal V0 protein A3|V-ATPase a3|osteosclerotic', map_location='19 3.62 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3896050, end_position_on_the_genomic_accession=3907133, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  27061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27061, CurrentID=0, Status='live', Symbol='Bcap31', Aliases='Bap31', description='B cell receptor associated protein 31', other_designations='BCR-associated protein 31|accessory protein BAP31|p28', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73686178, end_position_on_the_genomic_accession=73716466, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  27062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27062, CurrentID=0, Status='live', Symbol='Cadps', Aliases='AU067781, CAPS, CAPS1, mKIAA1121', description='Ca2+-dependent secretion activator', other_designations='CAPS-1|Ca2+-dependent activator protein for secretion|Ca<2+>dependent activator protein for secretion|calcium-dependent activator protein for secretion 1', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=12372560, end_position_on_the_genomic_accession=12823314, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  27078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27078, CurrentID=0, Status='live', Symbol='B9d1', Aliases='AW045994, B9, Eppb9', description='B9 protein domain 1', other_designations='endothelial precursor cells protein B9|endothelial precursor protein B9', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61505168, end_position_on_the_genomic_accession=61512929, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  27081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27081, CurrentID=0, Status='live', Symbol='Zfp275', Aliases='5430431L06Rik, AI593314, DXBay20, DXHXS52, DXPas8', description='zinc finger protein 275', other_designations='', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73342583, end_position_on_the_genomic_accession=73359080, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27083, CurrentID=0, Status='live', Symbol='Xlr4b', Aliases='Xlr4', description='X-linked lymphocyte-regulated 4B', other_designations='X-linked lymphocyte regulated gene 4', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73214323, end_position_on_the_genomic_accession=73222453, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  27084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27084, CurrentID=0, Status='live', Symbol='Xlr5c', Aliases='Xlr5', description='X-linked lymphocyte-regulated 5C', other_designations='', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73283361, end_position_on_the_genomic_accession=73293970, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27096, CurrentID=0, Status='live', Symbol='Trappc3', Aliases='1110058K12Rik, AI835942, AU019197, Bet3', description='trafficking protein particle complex 3', other_designations='Bet3 homolog', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126262405, end_position_on_the_genomic_accession=126275883, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  27103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27103, CurrentID=0, Status='live', Symbol='Eif2ak4', Aliases='2610011M03, GCN2, MGCN2', description='eukaryotic translation initiation factor 2 alpha kinase 4', other_designations='GCN2 eIF2alpha kinase|GCN2-like protein', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118388578, end_position_on_the_genomic_accession=118475234, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  27140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27140, CurrentID=0, Status='live', Symbol='Tlx3', Aliases='Hox11l2, Rnx, Tlx1l2', description='T cell leukemia, homeobox 3', other_designations='T-cell leukemia, homeobox 3|homeo box 11-like 2|homeobox TLX-3|homeobox protein Hox-11L2|respiratory neuron homeobox protein', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=33200752, end_position_on_the_genomic_accession=33203588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27176, CurrentID=0, Status='live', Symbol='Rpl7a', Aliases='Surf3', description='ribosomal protein L7A', other_designations='surfeit 3|surfeit locus protein 3', map_location='2 19.1 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26910807, end_position_on_the_genomic_accession=26913311, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100035114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035114, CurrentID=0, Status='live', Symbol='Sle16', Aliases='', description='systematic lupus erythematosus susceptibility 16', other_designations='', map_location='1 76.73 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100035115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035115, CurrentID=0, Status='live', Symbol='Sle18', Aliases='', description='systematic lupus erythematosus susceptibility 18', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27204, CurrentID=0, Status='live', Symbol='Syn3', Aliases='', description='synapsin III', other_designations='synapsin 3|synapsin IIIa', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=86048746, end_position_on_the_genomic_accession=86499019, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  27205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27205, CurrentID=0, Status='live', Symbol='Podxl', Aliases='AW121214, Ly102, PC, PCLP-1, Pclp1', description='podocalyxin-like', other_designations='podocalyxin-like protein 1', map_location='6 12.57 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=31519493, end_position_on_the_genomic_accession=31563937, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  27206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27206, CurrentID=0, Status='live', Symbol='Nrk', Aliases='AA407323, AA409510, AI225895, AV168563, AW538185, Nesk', description='Nik related kinase', other_designations='NIK-like embryo-specific kinase|nck-interacting kinase-like embryo specific kinase', map_location='X 61.15 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=138914312, end_position_on_the_genomic_accession=139010532, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  27207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27207, CurrentID=0, Status='live', Symbol='Rps11', Aliases='', description='ribosomal protein S11', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45122388, end_position_on_the_genomic_accession=45124389, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  27209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27209, CurrentID=0, Status='live', Symbol='Snord32a', Aliases='Rnu32, Rnu32a, U32, U32A', description='small nucleolar RNA, C/D box 32A', other_designations='RNA, U32 small nucleolar', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45127383, end_position_on_the_genomic_accession=45127463, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27211, CurrentID=0, Status='live', Symbol='Snord35a', Aliases='Rnu35a, U35, U35A', description='small nucleolar RNA, C/D box 35A', other_designations='RNA, U35a small nucleolar', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45126347, end_position_on_the_genomic_accession=45126435, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27214, CurrentID=0, Status='live', Symbol='Dbf4', Aliases='AA545217, Ask', description='DBF4 zinc finger', other_designations='DBF4 homolog|activator of S phase kinase', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8396969, end_position_on_the_genomic_accession=8422722, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  27215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27215, CurrentID=0, Status='live', Symbol='Azi2', Aliases='AA410145, AZ2', description='5-azacytidine induced gene 2', other_designations='NF-kappa-B-activating kinase-associated protein 1|nak-associated protein 1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=118040401, end_position_on_the_genomic_accession=118069798, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  27216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27216, CurrentID=0, Status='live', Symbol='Olfr154', Aliases='912-93, MOR175-1, OR93, Olfr4-3', description='olfactory receptor 154', other_designations='GA_x6K02T2Q125-47143827-47142871|olfactory receptor 175-1|olfactory receptor 4 cluster, gene 3|olfactory receptor MOR175-1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85663377, end_position_on_the_genomic_accession=85664455, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27217, CurrentID=0, Status='live', Symbol='Mixl1', Aliases='Mm1, Mml', description='Mix1 homeobox-like 1 (Xenopus laevis)', other_designations='MIX1 homeobox-like protein 1|homeodomain protein MIX|mMix|mix.1 homeobox-like protein', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180693043, end_position_on_the_genomic_accession=180709055, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27218, CurrentID=0, Status='live', Symbol='Slamf1', Aliases='4933415F16, AA177906, CD150, CDw150, ESTM51, IPO-3, Slam', description='signaling lymphocytic activation molecule family member 1', other_designations='SLAM family member 1|signaling lymphocyte activation molecule', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171767132, end_position_on_the_genomic_accession=171801184, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  27219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27219, CurrentID=0, Status='live', Symbol='Sgk2', Aliases='AI098171, AW146006, Sgkl', description='serum/glucocorticoid regulated kinase 2', other_designations='serum/glucocorticoid-regulated kinase 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=162987302, end_position_on_the_genomic_accession=163014139, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  27220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27220, CurrentID=0, Status='live', Symbol='Cartpt', Aliases='Cart', description='CART prepropeptide', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=99898483, end_position_on_the_genomic_accession=99900683, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27221, CurrentID=0, Status='live', Symbol='Chaf1a', Aliases='AL023013, AL024058, CAF-1, Cac1p', description='chromatin assembly factor 1, subunit A (p150)', other_designations='CAF-1 subunit A|CAF-I 150 kDa subunit|CAF-I p150|CAF-Ip150|chromatin assembly factor I p150 subunit', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56040416, end_position_on_the_genomic_accession=56068026, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  27222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27222, CurrentID=0, Status='live', Symbol='Atp1a4', Aliases='', description='ATPase, Na+/K+ transporting, alpha 4 polypeptide', other_designations='Na(+)/K(+) ATPase alpha-4 subunit|Na+/K+ ATPase alpha 4 subunit|sodium pump subunit alpha-4|sodium/potassium-transporting ATPase alpha-4 chain', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172223508, end_position_on_the_genomic_accession=172258424, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  27223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27223, CurrentID=0, Status='live', Symbol='Trp53bp1', Aliases='53BP1, Tp53bp1, m53BP1, p53BP1', description='transformation related protein 53 binding protein 1', other_designations='murine p53-binding protein|p53-binding protein 1|transformation related protein 53 binding-protein 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121194835, end_position_on_the_genomic_accession=121290320, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  27224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27224, CurrentID=0, Status='live', Symbol='Tceb3', Aliases='110kDa, AA408125a, Tceb3', description='transcription elongation factor B (SIII), polypeptide 3', other_designations='Elongin A|RNA polymerase II transcription factor SIII subunit A1|SIII p110|eloA|elongin 110 kDa subunit|elongin-A', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136003370, end_position_on_the_genomic_accession=136021649, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27225, CurrentID=0, Status='live', Symbol='Ddx24', Aliases='1700055J08Rik, 2510027P10Rik, AI649272', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 24', other_designations='DEAD box protein 24|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 13 (RNA helicase A)', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103407976, end_position_on_the_genomic_accession=103425867, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  27226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27226, CurrentID=0, Status='live', Symbol='Pla2g7', Aliases='R75400', description='phospholipase A2, group VII (platelet-activating factor acetylhydrolase, plasma)', other_designations='1-alkyl-2-acetylglycerophosphocholine esterase|2-acetyl-1-alkylglycerophosphocholine esterase|LDL-PLA(2)|LDL-associated phospholipase A2|PAF 2-acylhydrolase|PAF acetylhydrolase', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43568082, end_position_on_the_genomic_accession=43612202, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  27260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27260, CurrentID=0, Status='live', Symbol='Plek2', Aliases='', description='pleckstrin 2', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=78888691, end_position_on_the_genomic_accession=78907182, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  27261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27261, CurrentID=0, Status='live', Symbol='Dok3', Aliases='AI450713, Dokl', description='docking protein 3', other_designations='DOK-L|Dok-like protein|downstream of tyrosine kinase 3|p62(dok)-like protein|p62Dok-like protein', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55523235, end_position_on_the_genomic_accession=55528538, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  27262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27262, CurrentID=0, Status='live', Symbol='Smok1', Aliases='Smok, Tcr-1, Tcr1', description='sperm motility kinase 1', other_designations='t-complex responder 1', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27263, CurrentID=0, Status='live', Symbol='Smok2a', Aliases='Smok2', description='sperm motility kinase 2A', other_designations='putative protein kinase|sperm motility kinase 2', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13220998, end_position_on_the_genomic_accession=13230264, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  27267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27267, CurrentID=0, Status='live', Symbol='Cars', Aliases='CA3', description='cysteinyl-tRNA synthetase', other_designations='cysRS|cysteinyl-tRNA synthetase, cytoplasmic', map_location='7 88.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143557230, end_position_on_the_genomic_accession=143600090, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  27273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27273, CurrentID=0, Status='live', Symbol='Pdk4', Aliases='AV005916', description='pyruvate dehydrogenase kinase, isoenzyme 4', other_designations='[Pyruvate dehydrogenase [lipoamide]] kinase isozyme 4, mitochondrial|pyruvate dehydrogenase kinase 4', map_location='6 2.06 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=5483351, end_position_on_the_genomic_accession=5496278, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27274, CurrentID=0, Status='live', Symbol='Zfp354b', Aliases='Kid2, Tcf17l1, Znf354b', description='zinc finger protein 354B', other_designations='kidney, ischemia, and developmentally-regulated protein 2|transcription factor 17-like 1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50921786, end_position_on_the_genomic_accession=50931635, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  27275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27275, CurrentID=0, Status='live', Symbol='Nufip1', Aliases='Nufip', description='nuclear fragile X mental retardation protein interacting protein 1', other_designations='nuclear FMRP interacting protein|nuclear FMRP-interacting protein 1', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=76110891, end_position_on_the_genomic_accession=76137379, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27276, CurrentID=0, Status='live', Symbol='Plekhb1', Aliases='AI315628, PHR1, Phret1, evt-1', description='pleckstrin homology domain containing, family B (evectins) member 1', other_designations='PH domain containing protein in retina 1|PH domain-containing family B member 1|PH domain-containing protein in retina 1|evectin-1|pleckstrin homology domain retinal protein 1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100643896, end_position_on_the_genomic_accession=100662394, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27277, CurrentID=0, Status='live', Symbol='Golga5', Aliases='Ret-II', description='golgi autoantigen, golgin subfamily a, 5', other_designations='golgi autoantigen golgin subfamily a5|golgin-84|protein Sumiko', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102469134, end_position_on_the_genomic_accession=102497911, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  27278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27278, CurrentID=0, Status='live', Symbol='Clnk', Aliases='MIST', description='cytokine-dependent hematopoietic cell linker', other_designations='mast cell immunoreceptor signal transducer', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38705588, end_position_on_the_genomic_accession=38876858, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  27279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27279, CurrentID=0, Status='live', Symbol='Tnfrsf12a', Aliases='AI255180, C87282, Fn14, HPIP, TWEAK-R, TweakR', description='tumor necrosis factor receptor superfamily, member 12a', other_designations='FGF-inducible 14|fibroblast growth factor regulated protein 2|fibroblast growth factor-inducible immediate-early response protein 14|fibroblast growth factor-regulated protein 2|tweak-receptor|type I transmembrane protein Fn14', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23675445, end_position_on_the_genomic_accession=23677449, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  27280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27280, CurrentID=0, Status='live', Symbol='Phlda3', Aliases='Tih1', description='pleckstrin homology-like domain, family A, member 3', other_designations='TDAG/lpl homolog 1|TDAG51/Ipl homolog 1', map_location='1 59.26 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135766085, end_position_on_the_genomic_accession=135769134, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  27281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27281, CurrentID=0, Status='live', Symbol='Hrasls', Aliases='2810012B06Rik, A-C1, HRSL1, Hrasrs', description='HRAS-like suppressor', other_designations='HRAS-like suppressor 1|Harvey rat sarcoma virus oncogene related suppressor|phospholipase A', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=29209695, end_position_on_the_genomic_accession=29230531, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  27282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27282, CurrentID=0, Status='live', Symbol='Erv4', Aliases='MuERV-L', description='endogenous retroviral sequence 4 (with leucine tRNA primer)', other_designations='endogenous retroviral sequence 4 (with leucine t-RNA primer)', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27354, CurrentID=0, Status='live', Symbol='Nbn', Aliases='Nbs1', description='nibrin', other_designations='cell cycle regulatory protein p95|nijmegen breakage syndrome protein 1 homolog', map_location='4 A', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=15957967, end_position_on_the_genomic_accession=15992589, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  27355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27355, CurrentID=0, Status='live', Symbol='Pald1', Aliases='MMPAL, Pald, mKIAA1274', description='phosphatase domain containing, paladin 1', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61319656, end_position_on_the_genomic_accession=61383533, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  27356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27356, CurrentID=0, Status='live', Symbol='Insl6', Aliases='RIF1', description='insulin-like 6', other_designations='insulin-like peptide 6|relaxin/insulin-like factor 1', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29321354, end_position_on_the_genomic_accession=29325318, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  27357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27357, CurrentID=0, Status='live', Symbol='Gyg', Aliases='AU0176671, Gyg', description='glycogenin', other_designations='GN-1|GN1|glycogenin 1', map_location='3 6.19 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=20122084, end_position_on_the_genomic_accession=20155266, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  27358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27358, CurrentID=0, Status='live', Symbol='Defb3', Aliases='BD-3', description='defensin beta 3', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19293361, end_position_on_the_genomic_accession=19295339, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  27359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27359, CurrentID=0, Status='live', Symbol='Sytl4', Aliases='Gph, Slp4', description='synaptotagmin-like 4', other_designations='exophilin-2|granuphilin-a|granuphilin-b', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=133936385, end_position_on_the_genomic_accession=133981822, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  27360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27360, CurrentID=0, Status='live', Symbol='Add3', Aliases='AI463285, R75380', description='adducin 3 (gamma)', other_designations='adducin-like protein 70|cytoskeletal protein', map_location='19 47.18 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53140443, end_position_on_the_genomic_accession=53247326, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  27361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27361, CurrentID=0, Status='live', Symbol='Msrb1', Aliases='D17Wsu82e, SELX, SelR, Sepr, Sepx1', description='methionine sulfoxide reductase B1', other_designations='selenoprotein R|selenoprotein X 1', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24736642, end_position_on_the_genomic_accession=24742778, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  27362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27362, CurrentID=0, Status='live', Symbol='Dnajb9', Aliases='AA408011, AA673251, AA673481, AW556981, ERdj4, Mdg1, mDj7', description='DnaJ heat shock protein family (Hsp40) member B9', other_designations='DNA j protein b9|DnaJ (Hsp40) homolog, subfamily B, member 9|ER-resident protein ERdj4|endoplasmic reticulum DNA J domain-containing protein 4|microvascular endothelial differentiation gene 1', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=44205897, end_position_on_the_genomic_accession=44210068, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27364, CurrentID=0, Status='live', Symbol='Srr', Aliases='M100034, Rgsc34, Srs', description='serine racemase', other_designations='D-serine ammonia-lyase|D-serine dehydratase|L-serine ammonia-lyase|L-serine dehydratase', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74905047, end_position_on_the_genomic_accession=74925798, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27366, CurrentID=0, Status='live', Symbol='Txnl4a', Aliases='D18Wsu98e, Dim1, ENSMUSG00000057130, Txnl4, U5-15kD, U5-15kDa', description='thioredoxin-like 4A', other_designations='DIM1 protein homolog|spliceosomal U5 snRNP-specific 15 kDa protein|thioredoxin-like 4|thioredoxin-like U5 snRNP protein U5-15kD', map_location='18 53.34 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80206794, end_position_on_the_genomic_accession=80225851, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27367, CurrentID=0, Status='live', Symbol='Rpl3', Aliases='F2, J1', description='ribosomal protein L3', other_designations='ribosomal protein L3 homolog', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80077781, end_position_on_the_genomic_accession=80083406, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27368, CurrentID=0, Status='live', Symbol='Tbl2', Aliases='C76179, WS-bTRP', description='transducin (beta)-like 2', other_designations='', map_location='5 75.02 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135149426, end_position_on_the_genomic_accession=135163876, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  27369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27369, CurrentID=0, Status='live', Symbol='Dguok', Aliases='dGK, dGK1, dGK2, dGK3', description='deoxyguanosine kinase', other_designations='', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83480214, end_position_on_the_genomic_accession=83506969, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  27370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27370, CurrentID=0, Status='live', Symbol='Rps26', Aliases='', description='ribosomal protein S26', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128624529, end_position_on_the_genomic_accession=128626506, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  27371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27371, CurrentID=0, Status='live', Symbol='Sh2d2a', Aliases='Lad, Ribp, Tsad', description='SH2 domain containing 2A', other_designations='Lck-associated adapter protein|Rlk/Itk-binding protein|SH2 domain protein 2A|T cell-specific adapter protein|p56Lck-associated adapter protein', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87846755, end_position_on_the_genomic_accession=87855722, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  27372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27372, CurrentID=0, Status='live', Symbol='Prl3c1', Aliases='Ghd7, PLP-J, Prlpi, Prlpj, decidualin', description='prolactin family 3, subfamily c, member 1', other_designations='PLP I|PLP-I|PRL-like protein I|PRL-like protein J|growth hormone d7|placental prolactin-like protein J|prolactin-like protein I|prolactin-like protein J', map_location='13 12.32 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27196660, end_position_on_the_genomic_accession=27203749, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27373, CurrentID=0, Status='live', Symbol='Csnk1e', Aliases='AI426939, AI551861, AW457082, CK1epsilon, CKIe, KC1epsilon, tau', description='casein kinase 1, epsilon', other_designations='CKI, epsilon|CKI-epsilon|KC1 epsilon', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79417852, end_position_on_the_genomic_accession=79455579, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  27374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27374, CurrentID=0, Status='live', Symbol='Prmt5', Aliases='Jbp1, Skb1', description='protein arginine N-methyltransferase 5', other_designations='Jak-binding protein 1|SKB1 homolog|histone-arginine N-methyltransferase PRMT5|shk1 kinase-binding protein 1 homolog', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54507182, end_position_on_the_genomic_accession=54517470, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  27375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27375, CurrentID=0, Status='live', Symbol='Tjp3', Aliases='', description='tight junction protein 3', other_designations='zona occludens protein 3|zonula occludens protein 3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81273188, end_position_on_the_genomic_accession=81291581, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  27376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27376, CurrentID=0, Status='live', Symbol='Slc25a10', Aliases='Dic', description='solute carrier family 25 (mitochondrial carrier, dicarboxylate transporter), member 10', other_designations='dicarboxylate ion carrier|solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 10|solute carrier family 25 member 10', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120491837, end_position_on_the_genomic_accession=120501161, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  27377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27377, CurrentID=0, Status='live', Symbol='Yme1l1', Aliases='FtsH1, Ftsh', description='YME1-like 1 (S. cerevisiae)', other_designations='ATP-dependent metalloprotease FtsH1|YME1-like protein 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=23156505, end_position_on_the_genomic_accession=23199260, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  27378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27378, CurrentID=0, Status='live', Symbol='Tcl1b3', Aliases='', description='T cell leukemia/lymphoma 1B, 3', other_designations='T-cell leukemia/lymphoma 1B, 3', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105191045, end_position_on_the_genomic_accession=105195399, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  27379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27379, CurrentID=0, Status='live', Symbol='Tcl1b1', Aliases='C88214', description='T cell leukemia/lymphoma 1B, 1', other_designations='T-cell leukemia/lymphoma 1B, 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105159704, end_position_on_the_genomic_accession=105166625, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  27380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27380, CurrentID=0, Status='live', Symbol='Tcl1b4', Aliases='AA675342', description='T cell leukemia/lymphoma 1B, 4', other_designations='T-cell leukemia/lymphoma 1B, 4', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105201957, end_position_on_the_genomic_accession=105206993, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  27381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27381, CurrentID=0, Status='live', Symbol='Tcl1b2', Aliases='', description='T cell leukemia/lymphoma 1B, 2', other_designations='T-cell leukemia/lymphoma 1B, 2', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105147012, end_position_on_the_genomic_accession=105155225, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  27382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27382, CurrentID=0, Status='live', Symbol='Tcl1b5', Aliases='D12Ertd644e', description='T cell leukemia/lymphoma 1B, 5', other_designations='T-cell leukemia/lymphoma 1B, 5', map_location='12 55.23 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105176358, end_position_on_the_genomic_accession=105181145, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  27384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27384, CurrentID=0, Status='live', Symbol='Akr1c13', Aliases='mKIAA4014', description='aldo-keto reductase family 1, member C13', other_designations='aldo-keto reductase AKR1C13', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4191184, end_position_on_the_genomic_accession=4205604, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27385, CurrentID=0, Status='live', Symbol='Magel2', Aliases='Mage-l2, NDNL1, nM15, ns7', description='melanoma antigen, family L, 2', other_designations='melanoma antigen-like gene 2|necdin-like 1|protein nS7', map_location='7 34.37 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=62376979, end_position_on_the_genomic_accession=62381640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  27386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27386, CurrentID=0, Status='live', Symbol='Npas3', Aliases='4930423H22Rik, bHLHe12', description='neuronal PAS domain protein 3', other_designations='basic-helix-loop-helix-PAS protein MOP6|member of PAS protein 6|neuronal PAS3', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=53248003, end_position_on_the_genomic_accession=54072175, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  27387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27387, CurrentID=0, Status='live', Symbol='Sh2d3c', Aliases='Chat, Nsp3, Shep1', description='SH2 domain containing 3C', other_designations='Cas/HEF1-associated signal transducer|SH2 domain-containing Eph receptor-binding protein 1|novel SH2-containing protein 3', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32721055, end_position_on_the_genomic_accession=32755007, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  27388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27388, CurrentID=0, Status='live', Symbol='Ptdss2', Aliases='AI481134, AI988017, PSS2', description='phosphatidylserine synthase 2', other_designations='PSS-2|ptdSer synthase 2|serine-exchange enzyme II', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141131286, end_position_on_the_genomic_accession=141156158, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  27389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27389, CurrentID=0, Status='live', Symbol='Dusp13', Aliases='DUSP13A, DUSP13B, Gm1203, LMW-DSP6, MDSP, TMDP, TS-DSP6', description='dual specificity phosphatase 13', other_designations='Dsp13-like protein|dual specificity protein phosphatase isoform MDSP|dual specificity tyrosine phosphatase TS-DSP6|dual-specificity tyrosine phosphatase TS-DSP6|muscle-restricted DSP|testis and skeletal muscle-specific dual specificity phosphatase|testis- and skeletal muscle-specific DSP', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21733395, end_position_on_the_genomic_accession=21748622, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27390, CurrentID=0, Status='live', Symbol='Mmel1', Aliases='Mell1, NEP2, NEPII, Nl1, SEP', description='membrane metallo-endopeptidase-like 1', other_designations='NEP2(m)|NEPLP alpha|NEPLP beta|NEPLP gamma|NL-1|NL2|mel transforming oncogene-like 1|neprilysin 2|neprilysin II|neprilysin-2|neprilysin-like 1|neprilysin-like peptidase|soluble secreted endopeptidase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154869585, end_position_on_the_genomic_accession=154895530, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  27392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27392, CurrentID=0, Status='live', Symbol='Pign', Aliases='Gm20308, PIG-N', description='phosphatidylinositol glycan anchor biosynthesis, class N', other_designations='phosphatidylinositol glycan, class N|phosphatidylinositol-glycan biosynthesis class N protein', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=105518422, end_position_on_the_genomic_accession=105663722, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  27393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27393, CurrentID=0, Status='live', Symbol='Mrpl39', Aliases='C21orf8, MRP-L5, ORF22, Rpml5', description='mitochondrial ribosomal protein L39', other_designations='L39mt|MRP-L39|ribosomal protein, mitochondrial, L5', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=84717580, end_position_on_the_genomic_accession=84735302, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  27395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27395, CurrentID=0, Status='live', Symbol='Mrpl15', Aliases='HSPC145, MRP-L7, Rpml7', description='mitochondrial ribosomal protein L15', other_designations='L15mt|MRP-L15|ribosomal protein, mitochondrial, L7', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=4773198, end_position_on_the_genomic_accession=4785745, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27397, CurrentID=0, Status='live', Symbol='Mrpl17', Aliases='MRP-L26, Rpml26', description='mitochondrial ribosomal protein L17', other_designations='L17mt|MRP-L17|ribosomal protein, mitochondrial, L26', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105803782, end_position_on_the_genomic_accession=105811087, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27398, CurrentID=0, Status='live', Symbol='Mrpl2', Aliases='CGI-22, L2mt, MRP-L14, MRP-L2, Rpml14', description='mitochondrial ribosomal protein L2', other_designations='ribosomal protein, mitochondrial, L14', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46646171, end_position_on_the_genomic_accession=46650132, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  27399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27399, CurrentID=0, Status='live', Symbol='Ip6k1', Aliases='1200016D08Rik, AU022776, Ihpk1, InsP6, InsP6k1, Itpk6', description='inositol hexaphosphate kinase 1', other_designations='inositol hexakisphosphate kinase 6|insP6 kinase 1', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108002648, end_position_on_the_genomic_accession=108048782, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27400, CurrentID=0, Status='live', Symbol='Hsd17b6', Aliases='17betaHSD9, Hsd17b9, Rdh8', description='hydroxysteroid (17-beta) dehydrogenase 6', other_designations='17-beta-HSD 6|17-beta-HSD6|17-beta-HSD9|3-alpha->beta-HSE|3-alpha->beta-hydroxysteroid epimerase|hydroxysteroid (17-beta) dehydrogenase 9|hydroxysteroid 17-beta dehydrogenase 9|oxidative 3-alpha hydroxysteroid dehydrogenase', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127990933, end_position_on_the_genomic_accession=128007636, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27401, CurrentID=0, Status='live', Symbol='Skp2', Aliases='4930500A04Rik, FBXL1, FWD1, p45', description='S-phase kinase-associated protein 2 (p45)', other_designations='F-box protein Skp2|F-box/WD-40 protein 1|cyclin A/CDK2-associated protein p45', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9111982, end_position_on_the_genomic_accession=9155425, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  27402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27402, CurrentID=0, Status='live', Symbol='Pdhx', Aliases='AI481367, E3bp, Pdx1', description='pyruvate dehydrogenase complex, component X', other_designations='dihydrolipoamide dehydrogenase-binding protein of pyruvate dehydrogenase complex (E3-binding protein) (proX)|lipoyl-containing pyruvate dehydrogenase complex component X', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103021055, end_position_on_the_genomic_accession=103073513, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27403, CurrentID=0, Status='live', Symbol='Abca7', Aliases='ABCX, Abc51', description='ATP-binding cassette, sub-family A (ABC1), member 7', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79996453, end_position_on_the_genomic_accession=80015572, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  27404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27404, CurrentID=0, Status='live', Symbol='Abca8b', Aliases='Abca8', description='ATP-binding cassette, sub-family A (ABC1), member 8b', other_designations='', map_location='11 72.88 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109933406, end_position_on_the_genomic_accession=109995846, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  27405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27405, CurrentID=0, Status='live', Symbol='Abcg3', Aliases='Abcp2, Mxr2', description='ATP-binding cassette, sub-family G (WHITE), member 3', other_designations='ATP-binding cassette, subfamily G, member 3', map_location='5 50.68 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104935057, end_position_on_the_genomic_accession=104982802, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  27406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27406, CurrentID=0, Status='live', Symbol='Abcf3', Aliases='AI326318, AU016058, BB119416', description='ATP-binding cassette, sub-family F (GCN20), member 3', other_designations='', map_location='16 12.46 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20548556, end_position_on_the_genomic_accession=20561603, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  27407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27407, CurrentID=0, Status='live', Symbol='Abcf2', Aliases='0710005O05Rik, Drr3, E430001O06', description='ATP-binding cassette, sub-family F (GCN20), member 2', other_designations='abcf2|developmentally regulated repeat element-containing transcript 3', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24565341, end_position_on_the_genomic_accession=24577467, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  27409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27409, CurrentID=0, Status='live', Symbol='Abcg5', Aliases='cmp, sterolin-1, trac', description='ATP-binding cassette, sub-family G (WHITE), member 5', other_designations='sterolin 1', map_location='17 55.02 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84658234, end_position_on_the_genomic_accession=84682923, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  27410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27410, CurrentID=0, Status='live', Symbol='Abca3', Aliases='1810036E22Rik, ABC-C, Abc3', description='ATP-binding cassette, sub-family A (ABC1), member 3', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24351969, end_position_on_the_genomic_accession=24410204, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  27411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27411, CurrentID=0, Status='live', Symbol='Slc14a2', Aliases='UT-A1, UT-A3', description='solute carrier family 14 (urea transporter), member 2', other_designations='solute carrier family 14 member 2|urea transporter, kidney', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=78146144, end_position_on_the_genomic_accession=78596950, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  27412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27412, CurrentID=0, Status='live', Symbol='Peg12', Aliases='Frat3', description='paternally expressed 12', other_designations='frequently rearranged in advanced T-cell lymphomas 3', map_location='7 34.38 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=62461871, end_position_on_the_genomic_accession=62464510, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27413, CurrentID=0, Status='live', Symbol='Abcb11', Aliases='ABC16, Bsep, Lith1, PFIC2, PGY4, SPGP', description='ATP-binding cassette, sub-family B (MDR/TAP), member 11', other_designations='ATP-binding cassette sub-family B member 11|ATP-binding cassette, sub-family B, member 11|sister of P-glycoprotein', map_location='2 39.69 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69238282, end_position_on_the_genomic_accession=69342842, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  27414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27414, CurrentID=0, Status='live', Symbol='Sergef', Aliases='DelGEF, Gef, Gnef, Gnefr', description='secretion regulating guanine nucleotide exchange factor', other_designations='deafness locus-associated putative guanine nucleotide exchange factor|guanine nucleotide exchange factor (RCC1 related)|guanine nucleotide exchange factor-related protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46443159, end_position_on_the_genomic_accession=46639807, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  27416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27416, CurrentID=0, Status='live', Symbol='Abcc5', Aliases='2900011L11Rik, AI132311a, Abcc5b, Mrp5, Abcc5', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 5', other_designations='ATP-binding cassette sub-family C member 5|ATP-binding cassette, sub-family C (CFTR/MRP), member 5a|ATP-binding cassette, sub-family C (CFTR/MRP), member 5b|ATP-binding cassette, sub-family C, member 5|MOAT-C|SMRP|multi-specific organic anion transporter C', map_location='16 12.41 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20331303, end_position_on_the_genomic_accession=20426394, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  27418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27418, CurrentID=0, Status='live', Symbol='Mkln1', Aliases='A130067F06Rik, AU015903, BB154892', description='muskelin 1, intracellular mediator containing kelch motifs', other_designations='', map_location='6 B1-B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=31398828, end_position_on_the_genomic_accession=31511795, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  27419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27419, CurrentID=0, Status='live', Symbol='Naglu', Aliases='', description='alpha-N-acetylglucosaminidase (Sanfilippo disease IIIB)', other_designations='N-acetyl-glucosaminidase|alpha-N-acetylglucosaminidase, lysosomal', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101070094, end_position_on_the_genomic_accession=101077672, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27421, CurrentID=0, Status='live', Symbol='Abcc6', Aliases='Abcc1b, DCC, Dyscalc1, Mrp6, dyscalc', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 6', other_designations='ATP-binding cassette sub-family C member 6|ATP-binding cassette, sub-family C, member 6|multidrug resistance-associated protein-6', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45974115, end_position_on_the_genomic_accession=46030486, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  27423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27423, CurrentID=0, Status='live', Symbol='Klra15', Aliases='Ly49O', description='killer cell lectin-like receptor, subfamily A, member 15', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27424, CurrentID=0, Status='live', Symbol='Klra16', Aliases='Klra24, Ly-49P, Ly49P, Ly49x', description='killer cell lectin-like receptor, subfamily A, member 16', other_designations='Ly49X receptor|natural killer cell receptor Ly-49P3', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27425, CurrentID=0, Status='live', Symbol='Atp5l', Aliases='4933437C06Rik', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit G', other_designations='ATPase subunit g|F1F0-ATP synthase g subunit', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44913248, end_position_on_the_genomic_accession=44920742, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27426, CurrentID=0, Status='live', Symbol='Nagpa', Aliases='AI596180, UCE', description='N-acetylglucosamine-1-phosphodiester alpha-N-acetylglucosaminidase', other_designations='alpha-GlcNAcase|mannose 6-phosphate-uncovering enzyme|phosphodiester alpha-GlcNAcase', map_location='16 A2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5195280, end_position_on_the_genomic_accession=5204040, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  27428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27428, CurrentID=0, Status='live', Symbol='Shroom3', Aliases='AL022960, D5Ertd287e, Shrm, Shrm3', description='shroom family member 3', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92683435, end_position_on_the_genomic_accession=92965759, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100035370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035370, CurrentID=0, Status='live', Symbol='Alpq1', Aliases='', description='alcohol preference QTL 1', other_designations='', map_location='9 28.62 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27475, CurrentID=0, Status='live', Symbol='20S', Aliases='', description='DNA segment, 20S', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100034361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034361, CurrentID=0, Status='live', Symbol='Mfap1b', Aliases='', description='microfibrillar-associated protein 1B', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121461666, end_position_on_the_genomic_accession=121474023, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27493, CurrentID=0, Status='live', Symbol='A230006K03Rik', Aliases='CAG-3, CAG8, D0Kist2, Gm7414', description='RIKEN cDNA A230006K03 gene', other_designations='', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=61010264, end_position_on_the_genomic_accession=61091109, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  27494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27494, CurrentID=0, Status='live', Symbol='Amot', Aliases='CAG-2, D0Kist1, Sii6', description='angiomotin', other_designations='', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=145446424, end_position_on_the_genomic_accession=145505541, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  244886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244886, CurrentID=0, Status='live', Symbol='Tmem266', Aliases='9630029F15, AI118078, HVRP1', description='transmembrane protein 266', other_designations='Hv1 related protein 1|transmembrane protein C15orf27 homolog', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55326919, end_position_on_the_genomic_accession=55438345, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  27528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27528, CurrentID=0, Status='live', Symbol='Nrep', Aliases='AI325076, D0H4S114, Harp, P311, PTZ17, SEZ17', description='neuronal regeneration related protein', other_designations='high-altitude regulated protein|neuronal protein 3.1|protein p311', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=33437019, end_position_on_the_genomic_accession=33464029, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100035533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035533, CurrentID=0, Status='live', Symbol='Hdlq5', Aliases='', description='HDL QTL 5', other_designations='', map_location='1 72.38 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27608, CurrentID=0, Status='live', Symbol='D17Leh89', Aliases='', description='DNA segment, Chr 17, Lehrach 89', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27620, CurrentID=0, Status='live', Symbol='D17Tu49', Aliases='', description='DNA segment, Chr 17, Tubingen 49', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27631, CurrentID=0, Status='live', Symbol='D17Leh525', Aliases='', description='DNA segment, Chr 17, Lehrach 525', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27632, CurrentID=0, Status='live', Symbol='Nelfe', Aliases='D17H6S45, NELF-E, RD, Rdbp', description='negative elongation factor complex member E, Rdbp', other_designations='RD RNA-binding protein|RNA-binding protein RD', map_location='17 18.4 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34850391, end_position_on_the_genomic_accession=34856372, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  27642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27642, CurrentID=0, Status='live', Symbol='DXHXS1104', Aliases='DXBay10', description='DNA segment, Chr X, human DXS1104', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27643, CurrentID=0, Status='live', Symbol='Ubl4a', Aliases='DXHXS254E, DXS254E, DXS254Eh, Gdx, Ubl4', description='ubiquitin-like 4A', other_designations='housekeeping protein DXS254E|ubiquitin-like 4|ubiquitin-like protein GDX', map_location='X 38.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74365718, end_position_on_the_genomic_accession=74368548, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  27660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27660, CurrentID=0, Status='live', Symbol='1700088E04Rik', Aliases='ESTJ0827E04', description='RIKEN cDNA 1700088E04 gene', other_designations='', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79134655, end_position_on_the_genomic_accession=79141251, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  27681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27681, CurrentID=0, Status='live', Symbol='Snf8', Aliases='D11Moh34', description='SNF8, ESCRT-II complex subunit, homolog (S. cerevisiae)', other_designations='ESCRT-II complex subunit VPS22', map_location='11 59.24 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96034890, end_position_on_the_genomic_accession=96047430, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  27756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27756, CurrentID=0, Status='live', Symbol='Lsm2', Aliases='D17H6S56E-2, D17H6S56E2, Dmapl, Dmpkap, G7b, Sm-X5, SmX5, snRNP', description='LSM2 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='G7b alternative form|LSM2 homolog, U6 small nuclear RNA associated|dystrophia myotonica-protein kinase associated protein|small ribonuclear protein G7b|snRNP core Sm-like protein Sm-x5|snRNP core protein SMX5', map_location='17 18.52 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34981854, end_position_on_the_genomic_accession=34985891, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100035693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035693, CurrentID=0, Status='live', Symbol='Dbm4', Aliases='', description='diabetes modifier 4', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27762, CurrentID=0, Status='live', Symbol='Vwa7', Aliases='C6orf27, D17H6S56E-3, G7c, NG37', description='von Willebrand factor A domain containing 7', other_designations='protein G7c', map_location='17 18.55 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35016357, end_position_on_the_genomic_accession=35026741, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100035719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035719, CurrentID=0, Status='live', Symbol='Mnic1', Aliases='', description='macronutrient intake, carbohydrate 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27784, CurrentID=0, Status='live', Symbol='Commd8', Aliases='C77770, D5Buc26e', description='COMM domain containing 8', other_designations='', map_location='5 38.44 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72156576, end_position_on_the_genomic_accession=72168183, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  27790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27790, CurrentID=0, Status='live', Symbol='DXPas19', Aliases='', description='DNA segment, Chr X, Pasteur Institute 19', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27801, CurrentID=0, Status='live', Symbol='Zdhhc8', Aliases='D16H22S1738E, DHHC-8, E330009O14Rik, Op53c05, ZDHHCL1', description='zinc finger, DHHC domain containing 8', other_designations='DHHC-containing protein 8|zinc finger DHHC domain-containing protein 8', map_location='16 11.29 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18220748, end_position_on_the_genomic_accession=18235136, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  791069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=791069, CurrentID=0, Status='live', Symbol='Mir301b', Aliases='Mirn301b, mir-301b', description='microRNA 301b', other_designations='mmu-mir-301b', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17124400, end_position_on_the_genomic_accession=17124496, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  27843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27843, CurrentID=0, Status='live', Symbol='DXYHgu1', Aliases='', description='DNA segment, Chr XY, Human Genetics Unit 1', other_designations='', map_location='XY 74.0 cM', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27844, CurrentID=0, Status='live', Symbol='DXSmh141', Aliases='DXOrl1, DXWas68', description="DNA segment, Chr X, St. Mary's Hospital 141", other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100035794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035794, CurrentID=0, Status='live', Symbol='Tabw2', Aliases='', description='tally ho associated body weight 2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100035795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035795, CurrentID=0, Status='live', Symbol='Gluchos1', Aliases='', description='glucose homeostasis QTL 1', other_designations='', map_location='13 65.32 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27878, CurrentID=0, Status='live', Symbol='Tada1', Aliases='2900026B15Rik, D1Ertd251e, Staf42l, Tada1', description='transcriptional adaptor 1', other_designations='transcriptional adapter 1-like protein|transcriptional adaptor 1 (HFI1 homolog, yeast) like', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166379167, end_position_on_the_genomic_accession=166393621, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  27883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27883, CurrentID=0, Status='live', Symbol='Tango2', Aliases='D16H22S680E, T10', description='transport and golgi organization 2', other_designations='ser/Thr-rich protein T10 in DGCR region|transport and golgi organization 2 homolog', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18300825, end_position_on_the_genomic_accession=18348098, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  27886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27886, CurrentID=0, Status='live', Symbol='Dgcr14', Aliases='AI462402, D16H22S1269E, Dgcr1, Dgsi, ES2, Es2el', description='DiGeorge syndrome critical region gene 14', other_designations='DiGeorge syndrome critical region protein 14|diGeorge syndrome critical region 14 homolog', map_location='16 B1-B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17900709, end_position_on_the_genomic_accession=17911364, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  27933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27933, CurrentID=0, Status='live', Symbol='D1Nds2', Aliases='', description='DNA segment, Chr 1, Nuffield Department of Surgery 2', other_designations='', map_location='1 50.84 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27934, CurrentID=0, Status='live', Symbol='D2Nds1', Aliases='', description='DNA segment, Chr 2, Nuffield Department of Surgery 1', other_designations='', map_location='2 51.62 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27935, CurrentID=0, Status='live', Symbol='D3Nds1', Aliases='', description='DNA segment, Chr 3, Nuffield Department of Surgery 1', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27939, CurrentID=0, Status='live', Symbol='D4Nds2', Aliases='', description='DNA segment, Chr 4, Nuffield Department of Surgery 2', other_designations='', map_location='4 57.9 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27941, CurrentID=0, Status='live', Symbol='D5Nds2', Aliases='', description='DNA segment, Chr 5, Nuffield Department of Surgery 2', other_designations='', map_location='5 38.44 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27942, CurrentID=0, Status='live', Symbol='D7Nds1', Aliases='', description='DNA segment, Chr 7, Nuffield Department of Surgery 1', other_designations='', map_location='7 43.46 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27943, CurrentID=0, Status='live', Symbol='D7Nds2', Aliases='', description='DNA segment, Chr 7, Nuffield Department of Surgery 2', other_designations='', map_location='7 43.68 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27944, CurrentID=0, Status='live', Symbol='D9Nds2', Aliases='', description='DNA segment, Chr 9, Nuffield Department of Surgery 2', other_designations='', map_location='9 43.68 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27947, CurrentID=0, Status='live', Symbol='D10Nds1', Aliases='', description='DNA segment, Chr 10, Nuffield Department of Surgery 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27948, CurrentID=0, Status='live', Symbol='D10Nds2', Aliases='', description='DNA segment, Chr 10, Nuffield Department of Surgery 2', other_designations='', map_location='10 55.13 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27949, CurrentID=0, Status='live', Symbol='D11Nds1', Aliases='', description='DNA segment, Chr 11, Nuffield Department of Surgery 1', other_designations='', map_location='11 45.76 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27950, CurrentID=0, Status='live', Symbol='D13Nds1', Aliases='', description='DNA segment, Chr 13, Nuffield Department of Surgery 1', other_designations='', map_location='13 27.68 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27952, CurrentID=0, Status='live', Symbol='D15Nds1', Aliases='', description='DNA segment, Chr 15, Nuffield Department of Surgery 1', other_designations='', map_location='15 32.17 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  27965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27965, CurrentID=0, Status='live', Symbol='Spg21', Aliases='ACP33, BM-019, C78576, D9Wsu18e, GL010, MAST', description='spastic paraplegia 21 homolog (human)', other_designations='acid cluster protein 33|spastic paraplegia 21 autosomal recessive Mast syndrome protein homolog', map_location='9 35.32 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65460937, end_position_on_the_genomic_accession=65488470, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  27966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27966, CurrentID=0, Status='live', Symbol='Rrp9', Aliases='55kDa, D19435, D9Wsu10e, Rnu3ip2, U3-55k', description='RRP9, small subunit (SSU) processome component, homolog (yeast)', other_designations='RNA, U3 small nucleolar interacting protein 2|RRP9 homolog|U3 small nucleolar ribonucleoprotein-associated 55 kDa protein|U3 snoRNP-associated 55 kDa protein|U3 snoRNP-associated 55-kDa protein|U3 snoRNP-associated protein', map_location='9 57.59 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106477309, end_position_on_the_genomic_accession=106485415, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  27967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27967, CurrentID=0, Status='live', Symbol='Cherp', Aliases='5730408I11Rik, D8Wsu96e, DAN16, Scaf6', description='calcium homeostasis endoplasmic reticulum protein', other_designations='SR-related CTD associated factor 6|SR-related CTD-associated factor 6|cherp', map_location='8 35.08 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72460483, end_position_on_the_genomic_accession=72476675, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  27973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27973, CurrentID=0, Status='live', Symbol='Vkorc1', Aliases='D7Wsu86e', description='vitamin K epoxide reductase complex, subunit 1', other_designations='phylloquinone epoxide reductase|vitamin K 2,3-epoxide reductase complex subunit 1|vitamin K1 2,3-epoxide reductase subunit 1|vitamin K1 epoxide reductase (warfarin-sensitive)', map_location='7 69.81 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127893063, end_position_on_the_genomic_accession=127895617, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  27979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27979, CurrentID=0, Status='live', Symbol='Eif3b', Aliases='AL033316, AL033334, AL033369, AW208965, D5Wsu45e, EIF3-ETA, EIF3-P110, EIF3-P116, Eif3s9, PRT1', description='eukaryotic translation initiation factor 3, subunit B', other_designations='eIF-3-eta|eIF3 p116|eIF3b|eukaryotic translation initiation factor 3 subunit 9|eukaryotic translation initiation factor 3, subunit 9 (eta)', map_location='5 78.97 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140419305, end_position_on_the_genomic_accession=140443358, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  27981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27981, CurrentID=0, Status='live', Symbol='Rsrp1', Aliases='2700043I21Rik, D4Ucla2, D4Wsu53e', description='arginine/serine rich protein 1', other_designations='UPF0471 protein C1orf63 homolog', map_location='4 67.15 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134923625, end_position_on_the_genomic_accession=134927370, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  27984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27984, CurrentID=0, Status='live', Symbol='Efhd2', Aliases='2600015J22Rik, AA408606, D4Wsu27e', description='EF hand domain containing 2', other_designations='swiprosin 1|swiprosin-1', map_location='4 74.75 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141858142, end_position_on_the_genomic_accession=141874920, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  27993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27993, CurrentID=0, Status='live', Symbol='Imp4', Aliases='AA409888, AV031295, D1Wsu40e', description='IMP4, U3 small nucleolar ribonucleoprotein', other_designations='IMP4, U3 small nucleolar ribonucleoprotein, homolog|U3 snoRNP protein 4 homolog|U3 snoRNP protein IMP4', map_location='1 13.22 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34439829, end_position_on_the_genomic_accession=34445747, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  27998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27998, CurrentID=0, Status='live', Symbol='Exosc5', Aliases='D7Wsu180e', description='exosome component 5', other_designations='exosome complex exonuclease RRP46|ribosomal RNA-processing protein 46', map_location='7 13.96 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25659153, end_position_on_the_genomic_accession=25668032, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  27999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=27999, CurrentID=0, Status='live', Symbol='Fam3c', Aliases='D6Wsu176e, Ilei', description='family with sequence similarity 3, member C', other_designations='interleukin-like EMT inducer', map_location='6 9.24 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=22306520, end_position_on_the_genomic_accession=22356081, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  28000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28000, CurrentID=0, Status='live', Symbol='Prpf19', Aliases='AA617263, AL024362, D19Wsu55e, NMP200, PSO4, Prp19, Snev', description='pre-mRNA processing factor 19', other_designations='PRP19/PSO4 homolog|PRP19/PSO4 pre-mRNA processing factor 19 homolog|nuclear matrix protein 200|nuclear matrix protein SNEV|prp19 beta|senescence evasion factor', map_location='19 7.33 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10895231, end_position_on_the_genomic_accession=10909559, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  28006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28006, CurrentID=0, Status='live', Symbol='Fam21', Aliases='A130095H06, C530005J20Rik, D6Wsu116e', description='family with sequence similarity 21', other_designations='', map_location='6 53.76 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116208026, end_position_on_the_genomic_accession=116262671, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  28010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28010, CurrentID=0, Status='live', Symbol='Miip', Aliases='AA553001, AI853621, D4Wsu114e, IIP45', description='migration and invasion inhibitory protein', other_designations='invasion inhibitory protein 45|invasion-inhibitory protein 45', map_location='4 78.56 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147860778, end_position_on_the_genomic_accession=147868719, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100035948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100035948, CurrentID=0, Status='live', Symbol='T2dm2sa', Aliases='', description='type 2 diabetes mellitus 2 in SMXA RI mice', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28015, CurrentID=0, Status='live', Symbol='Polr2m', Aliases='AA407243, AA408799, AI303746, AW553238, D9Wsu138e, Grinl1a, mKIAA4185', description='polymerase (RNA) II (DNA directed) polypeptide M', other_designations='DNA-directed RNA polymerase II subunit M|glutamate receptor, ionotropic, N-methyl D-aspartate-like 1A|glutamate receptor-like protein 1A|protein GRINL1A', map_location='9 39.85 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=71478437, end_position_on_the_genomic_accession=71485983, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  28018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28018, CurrentID=0, Status='live', Symbol='Ubfd1', Aliases='AI467302, D7Wsu105e, D7Wsu128e, Ubph', description='ubiquitin family domain containing 1', other_designations='ubiquitin-binding protein homolog', map_location='7 65.35 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122067198, end_position_on_the_genomic_accession=122082199, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  28019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28019, CurrentID=0, Status='live', Symbol='Ing4', Aliases='D6Wsu147e, D6Xrf92, p29ING4', description='inhibitor of growth family, member 4', other_designations='ING1-like protein', map_location='6 59.17 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125039848, end_position_on_the_genomic_accession=125049264, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  28028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28028, CurrentID=0, Status='live', Symbol='Mrpl50', Aliases='AA408971, AI317351, D4Wsu125e', description='mitochondrial ribosomal protein L50', other_designations='L50mt|MRP-L50', map_location='4 26.54 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49512597, end_position_on_the_genomic_accession=49521083, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  28030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28030, CurrentID=0, Status='live', Symbol='Gfm1', Aliases='AW545374, D3Wsu133e, Gfm', description='G elongation factor, mitochondrial 1', other_designations='EF-Gmt|G elongation factor 1|elongation factor G 1, mitochondrial|elongation factor G1|mEF-G 1|mitochondrial', map_location='3 30.96 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=67430115, end_position_on_the_genomic_accession=67475068, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  28035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28035, CurrentID=0, Status='live', Symbol='Usp39', Aliases='AA408960, AI894154, CGI-21, D6Wsu157e, SAD1', description='ubiquitin specific peptidase 39', other_designations='SnRNP assembly defective 1 homolog|inactive ubiquitin-specific peptidase 39|ubiquitin specific protease 39', map_location='6 32.27 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72318676, end_position_on_the_genomic_accession=72345217, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  28036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28036, CurrentID=0, Status='live', Symbol='Larp7', Aliases='C330027G06Rik, D3Wsu161e', description='La ribonucleoprotein domain family, member 7', other_designations='', map_location='3 56.54 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127536714, end_position_on_the_genomic_accession=127553349, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  28040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28040, CurrentID=0, Status='live', Symbol='D6Wsu163e', Aliases='C12orf4', description='DNA segment, Chr 6, Wayne State University 163, expressed', other_designations='', map_location='6 61.92 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126939966, end_position_on_the_genomic_accession=126975704, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  28042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28042, CurrentID=0, Status='live', Symbol='Ept1', Aliases='4933402G07Rik, AI448296, AI452230, C79563, D5Wsu178e, SELI, mKIAA1724', description='ethanolaminephosphotransferase 1 (CDP-ethanolamine-specific)', other_designations='selenoprotein I', map_location='5 16.22 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30232618, end_position_on_the_genomic_accession=30272430, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  28064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28064, CurrentID=0, Status='live', Symbol='Yipf3', Aliases='C6orf109, D17Wsu94e, KLIP1', description='Yip1 domain family, member 3', other_designations='YIP1 family member 3|killer lineage protein 1|natural killer cell-specific antigen KLIP1', map_location='17 22.9 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46248070, end_position_on_the_genomic_accession=46252548, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  28071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28071, CurrentID=0, Status='live', Symbol='Twistnb', Aliases='2410173G11, 2810024J17Rik, D16Wsu83e', description='twist basic helix-loop-helix transcription factor 1 neighbor', other_designations='twist neighbor protein', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=33429624, end_position_on_the_genomic_accession=33439380, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  28075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28075, CurrentID=0, Status='live', Symbol='Desi1', Aliases='AI427858, AI850401, D15Wsu75e, DeSI-1, Fam152b, Pppde2', description='desumoylating isopeptidase 1', other_designations='PPPDE peptidase domain containing 2|PPPDE peptidase domain-containing protein 2|family with sequence similarity 152, member B', map_location='15 38.34 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81992523, end_position_on_the_genomic_accession=82016221, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  28077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28077, CurrentID=0, Status='live', Symbol='Med10', Aliases='AA959813, AI385605, C78613, D13Wsu50e', description='mediator complex subunit 10', other_designations='mediator of RNA polymerase II transcription, subunit 10 homolog (NUT2, S. cerevisiae)', map_location='13 35.55 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=69802348, end_position_on_the_genomic_accession=69816104, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  28078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28078, CurrentID=0, Status='live', Symbol='Prl5a1', Aliases='1600013P04Rik, D13Wsu14e, Ghd24, PLP-L, Prlpl', description='prolactin family 5, subfamily a, member 1', other_designations='PRL-like protein L|growth hormone d24|placental prolactin-like protein L|prolactin-like protein L', map_location='13 12.69 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=28142484, end_position_on_the_genomic_accession=28151595, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  28080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28080, CurrentID=0, Status='live', Symbol='Atp5o', Aliases='ATPO, D12Wsu28e, OSCP', description='ATP synthase, H+ transporting, mitochondrial F1 complex, O subunit', other_designations='ATP synthase, H+ transporting, mitochondrial F1 complex, O subunit; DNA segment, Chr 12, Wayne State University 28, expressed|oligomycin sensitivity conferral protein', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91925223, end_position_on_the_genomic_accession=91931630, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  28081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28081, CurrentID=0, Status='live', Symbol='Fam104a', Aliases='D11Wsu99e', description='family with sequence similarity 104, member A', other_designations='', map_location='11 79.12 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113661319, end_position_on_the_genomic_accession=113684186, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28084, CurrentID=0, Status='live', Symbol='Vps25', Aliases='1110020N13Rik, AW107467, D11Wsu68e', description='vacuolar protein sorting 25 (yeast)', other_designations='ESCRT-II complex subunit VPS25', map_location='11 64.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101253611, end_position_on_the_genomic_accession=101259547, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  28088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28088, CurrentID=0, Status='live', Symbol='Rtcb', Aliases='AI255213, AI463255, D10Wsu52e, FAAP, HSPC117', description="RNA 2',3'-cyclic phosphate and 5'-OH ligase", other_designations='focal adhesion-associated protein', map_location='10 41.81 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85938637, end_position_on_the_genomic_accession=85957793, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  28105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28105, CurrentID=0, Status='live', Symbol='Trim36', Aliases='D18Wsu100e, haprin', description='tripartite motif-containing 36', other_designations='acrosome RBCC protein|tripartite motif-containing protein 36', map_location='18 24.41 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46165300, end_position_on_the_genomic_accession=46212607, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  28106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28106, CurrentID=0, Status='live', Symbol='Mydgf', Aliases='D17Wsu104e, Il25, Ly6elg', description='myeloid derived growth factor', other_designations='UPF0556 protein C19orf10 homolog', map_location='17 29.2 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56176541, end_position_on_the_genomic_accession=56183920, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  28109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28109, CurrentID=0, Status='live', Symbol='D10Wsu102e', Aliases='AI747614, C430041I18Rik', description='DNA segment, Chr 10, Wayne State University 102, expressed', other_designations='', map_location='10 41.16 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=83360221, end_position_on_the_genomic_accession=83368835, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  28113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28113, CurrentID=0, Status='live', Symbol='Tinf2', Aliases='AW552114, D14Wsu146e, Tin2', description='Terf1 (TRF1)-interacting nuclear factor 2', other_designations='TRF1-interacting nuclear factor 2|TRF1-interacting nuclear protein 2', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55676208, end_position_on_the_genomic_accession=55682624, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  28114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28114, CurrentID=0, Status='live', Symbol='Nsun2', Aliases='D13Wsu123e, Misu', description='NOL1/NOP2/Sun domain family member 2', other_designations='Myc-induced SUN domain containing protein|NOL1/NOP2/Sun domain family 2|myc-induced SUN domain-containing protein|tRNA (cytosine-5-)-methyltransferase NSUN2', map_location='13 35.55 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=69612016, end_position_on_the_genomic_accession=69635780, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  100036059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036059, CurrentID=0, Status='live', Symbol='T2dm1sa', Aliases='Nfbg', description='type 2 diabetes mellitus 1 in SMXA RI mice', other_designations='non-fasting blood glucose', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28126, CurrentID=0, Status='live', Symbol='Nop16', Aliases='AA409471, D13Wsu177e', description='NOP16 nucleolar protein', other_designations='NOP16 nucleolar protein homolog', map_location='13 28.83 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54584191, end_position_on_the_genomic_accession=54590074, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  28135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28135, CurrentID=0, Status='live', Symbol='Cep63', Aliases='4921501M07, AL450317.13gm1, AW107703, CD20R, D9Mgc41, D9Mgc48e, ET2', description='centrosomal protein 63', other_designations='centrosome protein Cep63', map_location='9 54.61 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=102586578, end_position_on_the_genomic_accession=102626353, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  28146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28146, CurrentID=0, Status='live', Symbol='Serp1', Aliases='D3Ucla1, Ramp4', description='stress-associated endoplasmic reticulum protein 1', other_designations='ribosome associated membrane protein 4|ribosome-attached membrane protein 4', map_location='3 28.58 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=58521969, end_position_on_the_genomic_accession=58525884, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28169, CurrentID=0, Status='live', Symbol='Agpat3', Aliases='AW061257, AW493985, D10Jhu12e, lpaat3', description='1-acylglycerol-3-phosphate O-acyltransferase 3', other_designations='1-AGP acyltransferase 3|1-AGPAT 3|LPAAT-gamma|lysophosphatidic acid acyltransferase 3|lysophosphatidic acid acyltransferase gamma', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78269174, end_position_on_the_genomic_accession=78352484, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  28185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28185, CurrentID=0, Status='live', Symbol='Tomm70a', Aliases='2610044B22Rik, D16Ium22, D16Ium22e, D16Wsu109e, Tom70, mKIAA0719', description='translocase of outer mitochondrial membrane 70 homolog A (yeast)', other_designations='mitochondrial precursor proteins import receptor|translocase of outer membrane 70 kDa subunit', map_location='16 34.22 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57121714, end_position_on_the_genomic_accession=57154530, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  28193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28193, CurrentID=0, Status='live', Symbol='Reep3', Aliases='D10Ucla1', description='receptor accessory protein 3', other_designations='', map_location='10 34.83 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=67009189, end_position_on_the_genomic_accession=67096988, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  28194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28194, CurrentID=0, Status='live', Symbol='Apon', Aliases='AI195275, D10Ucla2', description='apolipoprotein N', other_designations='', map_location='10 76.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128254131, end_position_on_the_genomic_accession=128255901, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  28199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28199, CurrentID=0, Status='live', Symbol='Dcaf11', Aliases='0710008A13Rik, C76035, D14Ucla1, Dacf11, GLO14, Wdr23', description='DDB1 and CUL4 associated factor 11', other_designations='WD repeat domain 23|WD repeat-containing protein 23', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55559937, end_position_on_the_genomic_accession=55570065, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  28200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28200, CurrentID=0, Status='live', Symbol='Dhrs4', Aliases='AI043103, AI790593, CR, D14Ucla2, NDRD, PHCR, PSCD, RRD, mouNRDR', description='dehydrogenase/reductase (SDR family) member 4', other_designations='NADPH-dependent carbonyl reductase/NADP-retinol dehydrogenase|NADPH-dependent retinol dehydrogenase/reductase|peroxisomal short-chain alcohol dehydrogenase', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55478603, end_position_on_the_genomic_accession=55490340, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100036140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036140, CurrentID=0, Status='live', Symbol='Mobe1', Aliases='Moo1', description='modifier of obesity 1', other_designations='', map_location='2 38.54 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28240, CurrentID=0, Status='live', Symbol='Trpm2', Aliases='9830168K16Rik, C79133, LTRPC2, TRPC7, Trp7, Trrp7', description='transient receptor potential cation channel, subfamily M, member 2', other_designations='LTrpC-2|long transient receptor potential channel 2|transient receptor potential channel 7|transient receptor protein 7', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77907722, end_position_on_the_genomic_accession=77969906, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  28248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28248, CurrentID=0, Status='live', Symbol='Slco1a1', Aliases='A530084B21, OATP-1, Oatp1, Oatp1a1, Slc21a1', description='solute carrier organic anion transporter family, member 1a1', other_designations='sodium-independent organic anion-transporting polypeptide 1|solute carrier family 21 (organic anion transporter), member 1|solute carrier family 21 member 1', map_location='6 A3-A5', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=141907281, end_position_on_the_genomic_accession=141946965, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100036186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036186, CurrentID=0, Status='live', Symbol='Ahl3', Aliases='', description='age related hearing loss 3', other_designations='', map_location='17 38.15 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28253, CurrentID=0, Status='live', Symbol='Slco1b2', Aliases='7330442B20Rik, OATP-C, OATP2, Oatp1b2, Oatp4, Slc21a10, Slc21a6, lst-1, mlst-1', description='solute carrier organic anion transporter family, member 1b2', other_designations='liver-specific organic anion transporter 1|solute carrier family (organic anion transporter) member 10|solute carrier family 21 (organic anion transporter), member 10|solute carrier family 21 (organic anion transporter), member 6|solute carrier family 21 member 10', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=141629452, end_position_on_the_genomic_accession=141686646, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  28254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28254, CurrentID=0, Status='live', Symbol='Slco1a6', Aliases='4930422F19Rik, AI790453, Oatp-5, Oatp5, Slc21a13', description='solute carrier organic anion transporter family, member 1a6', other_designations='kidney-specific organic anion-transporting polypeptide 5|organic anion-transporting polypeptide|solute carrier family (organic anion transporter) member 13|solute carrier family 21 (organic anion transporter), member 13|solute carrier family 21 member 13', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142085761, end_position_on_the_genomic_accession=142186178, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  100036204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036204, CurrentID=0, Status='live', Symbol='Pbrgcsf1', Aliases='G-CSF', description='peripheral blood stem cell response to granulocyte colony stimulating factor 1', other_designations='', map_location='2 60.69 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036230, CurrentID=0, Status='live', Symbol='Tb2r1', Aliases='Lskcn', description='TGF-beta2 responsiveness 1', other_designations='LSK cell number', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  28295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28295, CurrentID=0, Status='live', Symbol='D10Jhu81e', Aliases='C21orf33, ES1', description='DNA segment, Chr 10, Johns Hopkins University 81 expressed', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78162067, end_position_on_the_genomic_accession=78169768, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70252, CurrentID=0, Status='live', Symbol='Jpx', Aliases='2010000I03Rik, 2510040I06Rik, Enox', description='Jpx transcript, Xist activator (non-protein coding)', other_designations='expressed neighbor of Xist', map_location='X 46.17 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103493576, end_position_on_the_genomic_accession=103506425, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100036458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036458, CurrentID=0, Status='live', Symbol='Bmd1b', Aliases='BMD1-2', description='bone mineral density 1b', other_designations='', map_location='1 80.13 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100036521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036521, CurrentID=0, Status='live', Symbol='Umad1', Aliases='Gm16039, OTTMUSG00000029336', description='UMAP1-MVP12 associated (UMA) domain containing 1', other_designations='', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=8259288, end_position_on_the_genomic_accession=8428767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  552899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=552899, CurrentID=0, Status='live', Symbol='Ugt2a2', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A2', other_designations='UDPGT 2A2', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87459493, end_position_on_the_genomic_accession=87482258, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100036768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100036768, CurrentID=0, Status='live', Symbol='Gm11696', Aliases='OTTMUSG00000003456', description='predicted gene 11696', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109354778, end_position_on_the_genomic_accession=109363654, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  553127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=553127, CurrentID=0, Status='live', Symbol='Cxx1b', Aliases='Mart8c', description='CAAX box 1B', other_designations='CAAX box 1 homolog B|mammalian retrotransposon derived 8c', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53669177, end_position_on_the_genomic_accession=53670408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100037258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100037258, CurrentID=0, Status='live', Symbol='Dnajc3', Aliases='AA408985, AU067833a, Dnajc3b, Prkri, mp58, p58, p58IPK, Dnajc3', description='DnaJ heat shock protein family (Hsp40) member C3', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 3|DnaJ (Hsp40) homolog, subfamily C, member 3A|DnaJ (Hsp40) homolog, subfamily C, member 3B|interferon-induced, double-stranded RNA-activated protein kinase inhibitor|protein kinase inhibitor of 58 kDa|protein kinase inhibitor p58|protein kinase, interferon inducible double stranded RNA dependent inhibitor', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118937932, end_position_on_the_genomic_accession=118981702, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100037278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100037278, CurrentID=0, Status='live', Symbol='Fam129c', Aliases='Bcnp1', description='family with sequence similarity 129, member C', other_designations='BCNP1 homolog', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71596366, end_position_on_the_genomic_accession=71608149, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100037282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100037282, CurrentID=0, Status='live', Symbol='Rsph3b', Aliases='Rshl2b', description='radial spoke 3B homolog (Chlamydomonas)', other_designations='A-kinase anchor protein RSPH3B|radial spoke head-like protein 2B|radial spokehead-like 2B', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6904716, end_position_on_the_genomic_accession=6948356, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100037283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100037283, CurrentID=0, Status='live', Symbol='Rnaset2a', Aliases='0610007O07Rik, 4833423A10Rik, 4930532K22Rik, C79750, RNASE6PL, Rnaset2', description='ribonuclease T2A', other_designations='ribonuclease 6', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8128592, end_position_on_the_genomic_accession=8147827, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70435, CurrentID=0, Status='live', Symbol='Inf2', Aliases='2610204M08Rik, AA589465, AW125550, EG629699', description='inverted formin, FH2 and WH2 domain containing', other_designations='formin, inverted', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112588746, end_position_on_the_genomic_accession=112615557, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  29805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29805, CurrentID=0, Status='live', Symbol='Znhit2', Aliases='C11orf5, Fon, ORF6-ps, Znhit2', description='zinc finger, HIT domain containing 2', other_designations='zinc finger, HIT domain containing 2, pseudogene', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6061207, end_position_on_the_genomic_accession=6062468, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  29806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29806, CurrentID=0, Status='live', Symbol='Limd1', Aliases='AW822033, D9Ertd192e', description='LIM domains containing 1', other_designations='', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123478661, end_position_on_the_genomic_accession=123521552, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  29807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29807, CurrentID=0, Status='live', Symbol='Tpk1', Aliases='', description='thiamine pyrophosphokinase', other_designations='mTPK1|thiamine pyrophosphokinase 1', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43344999, end_position_on_the_genomic_accession=43666484, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  29808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29808, CurrentID=0, Status='live', Symbol='Mga', Aliases='AV312082, C130042M01Rik, C80739, D030062C11Rik, Mad5', description='MAX gene associated', other_designations='MAX-interacting protein|Max dimerization protein 5', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119871646, end_position_on_the_genomic_accession=119970129, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  29809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29809, CurrentID=0, Status='live', Symbol='Rabgap1l', Aliases='5830411O09Rik, 8430421H08Rik, 9630005B12Rik, AW049894, HHL, Hh1, mKIAA0471', description='RAB GTPase activating protein 1-like', other_designations='rab6 GTPase activating protein-like|rabgap1l', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160219174, end_position_on_the_genomic_accession=160793476, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  29810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29810, CurrentID=0, Status='live', Symbol='Bag3', Aliases='AA407278, Bis, mg638', description='BCL2-associated athanogene 3', other_designations='BAG-3|Bcl-2-binding protein Bis|Bcl-2-interacting death suppressor|bcl-2-associated athanogene 3', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128523583, end_position_on_the_genomic_accession=128546981, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  29811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29811, CurrentID=0, Status='live', Symbol='Ndrg2', Aliases='AI182517, AU040374, Ndr2, SYLD', description='N-myc downstream regulated gene 2', other_designations='N-myc downstream regulated 2|N-myc downstream-regulated gene 2 protein|protein Ndr2', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51905271, end_position_on_the_genomic_accession=51914092, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  29812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29812, CurrentID=0, Status='live', Symbol='Ndrg3', Aliases='4833415O14Rik, AI747193, Ndr3', description='N-myc downstream regulated gene 3', other_designations='N-myc downstream regulated 3|N-myc downstream-regulated gene 3 protein|protein Ndr3', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156927342, end_position_on_the_genomic_accession=156992111, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  29813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29813, CurrentID=0, Status='live', Symbol='Zfp385a', Aliases='Hzf, Zfp385, Znf385a', description='zinc finger protein 385A', other_designations='hematopoietic zinc finger protein|zinc finger protein 385', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103313895, end_position_on_the_genomic_accession=103340086, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  29815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29815, CurrentID=0, Status='live', Symbol='Bcar3', Aliases='AI131758, AND-34', description='breast cancer anti-estrogen resistance 3', other_designations='p130Cas-binding protein AND-34', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122419780, end_position_on_the_genomic_accession=122530189, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  29816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29816, CurrentID=0, Status='live', Symbol='Hip1r', Aliases='AA410023, mKIAA0655', description='huntingtin interacting protein 1 related', other_designations='HIP1-related protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123973607, end_position_on_the_genomic_accession=124003215, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  29817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29817, CurrentID=0, Status='live', Symbol='Igfbp7', Aliases='AGM, Fstl2, Mac25', description='insulin-like growth factor binding protein 7', other_designations='IBP-7|IGF-binding protein 7|IGFBP-7|angiomodulin', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77349240, end_position_on_the_genomic_accession=77408045, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  29818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29818, CurrentID=0, Status='live', Symbol='Hspb7', Aliases='27kDa, Hsp25-2, cvHsp', description='heat shock protein family, member 7 (cardiovascular)', other_designations='cardiovascular heat shock protein|heat shock 27kDa protein family, member 7 (cardiovascular)|heat shock protein 25 kDa 2 (cardiovascular)|protein p19/6.8', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141420779, end_position_on_the_genomic_accession=141425310, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  29819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29819, CurrentID=0, Status='live', Symbol='Stau2', Aliases='', description='staufen (RNA binding protein) homolog 2 (Drosophila)', other_designations='staufen 2', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16228803, end_position_on_the_genomic_accession=16520181, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  29820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29820, CurrentID=0, Status='live', Symbol='Tnfrsf19', Aliases='AL023044, AW123854, TAJ, TAJ-ALPHA, TRADE, Troy', description='tumor necrosis factor receptor superfamily, member 19', other_designations='Toxicity and JNK inducer', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=60963834, end_position_on_the_genomic_accession=61046855, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  29845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29845, CurrentID=0, Status='live', Symbol='Olfr155', Aliases='GA_x5J8B7W5BNN-979337-980296, MOR262-14, OR37A, Olfr37a, mOR37a', description='olfactory receptor 155', other_designations='GA_x6K02T2N78B-16092200-16091241|olfactory receptor 37a', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43854292, end_position_on_the_genomic_accession=43855463, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  29846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29846, CurrentID=0, Status='live', Symbol='Olfr156', Aliases='MOR262-6, OR37B, Olfr37b, mOR37b', description='olfactory receptor 156', other_designations='GA_x6K02T2N78B-16125144-16126100|olfactory receptor 37b|olfactory receptor MOR262-6', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43820266, end_position_on_the_genomic_accession=43821433, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  29849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29849, CurrentID=0, Status='live', Symbol='Olfr159', Aliases='MOR262-5, Olfr37e, mOR37e', description='olfactory receptor 159', other_designations='GA_x6K02T2N78B-16165641-16166600|olfactory receptor 37e|olfactory receptor MOR262-5', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43770050, end_position_on_the_genomic_accession=43771009, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  29856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29856, CurrentID=0, Status='live', Symbol='Smtn', Aliases='smsmo', description='smoothelin', other_designations='', map_location='11 A2-A3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3517522, end_position_on_the_genomic_accession=3539654, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  29857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29857, CurrentID=0, Status='live', Symbol='Mapk12', Aliases='AW123708, Erk6, P38gamma, Prkm12, Sapk3', description='mitogen-activated protein kinase 12', other_designations='ERK-6|MAP kinase 12|MAP kinase p38 gamma|MAPK 12|extracellular signal-regulated kinase 6|mitogen activated protein kinase 12|mitogen-activated protein kinase p38 gamma|stress activated protein kinase 3|stress-activated protein kinase 3', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89130584, end_position_on_the_genomic_accession=89141593, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  29858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29858, CurrentID=0, Status='live', Symbol='Pmm1', Aliases='C77612', description='phosphomannomutase 1', other_designations='PMM 1|Secp53 homolog', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81951106, end_position_on_the_genomic_accession=81960930, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  29859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29859, CurrentID=0, Status='live', Symbol='Sult4a1', Aliases='2400007A17Rik, AI853543, BR-STL-1, ST4A1, Sultx3', description='sulfotransferase family 4A, member 1', other_designations='NST|brain sulfotransferase-like protein|mBR-STL|nervous system sulfotransferase|sulfotransferase-related protein', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84076097, end_position_on_the_genomic_accession=84105754, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  29861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29861, CurrentID=0, Status='live', Symbol='Dpf1', Aliases='Neud4', description='D4, zinc and double PHD fingers family 1', other_designations='BAF45B|BRG1-associated factor 45B|neuro-d4|neuronal d4 domain family member', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29303959, end_position_on_the_genomic_accession=29317592, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  29862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29862, CurrentID=0, Status='live', Symbol='Ninj2', Aliases='', description='ninjurin 2', other_designations='nerve injury-induced protein 2', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120093380, end_position_on_the_genomic_accession=120200339, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  29863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29863, CurrentID=0, Status='live', Symbol='Pde7b', Aliases='', description='phosphodiesterase 7B', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=20397052, end_position_on_the_genomic_accession=20725068, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  29864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29864, CurrentID=0, Status='live', Symbol='Rnf11', Aliases='', description='ring finger protein 11', other_designations='NEDD4 WW domain-binding protein 2|sid 1669', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109452473, end_position_on_the_genomic_accession=109484723, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  29865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29865, CurrentID=0, Status='live', Symbol='Cabp5', Aliases='', description='calcium binding protein 5', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13383295, end_position_on_the_genomic_accession=13408880, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  29866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29866, CurrentID=0, Status='live', Symbol='Cabp2', Aliases='', description='calcium binding protein 2', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4081469, end_position_on_the_genomic_accession=4087340, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  29867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29867, CurrentID=0, Status='live', Symbol='Cabp1', Aliases='caldendrin', description='calcium binding protein 1', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115168691, end_position_on_the_genomic_accession=115194376, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  29869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29869, CurrentID=0, Status='live', Symbol='Ulk2', Aliases='A830085I22Rik, AU015340, Unc51.2, mKIAA0623', description='unc-51 like kinase 2', other_designations='serine/threonine-protein kinase Unc51.2|unc-51-like kinase 2', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61775598, end_position_on_the_genomic_accession=61855321, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  29870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29870, CurrentID=0, Status='live', Symbol='Gtse1', Aliases='B99, Gtse-1', description='G two S phase expressed protein 1', other_designations='', map_location='15 40.42 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85859707, end_position_on_the_genomic_accession=85876573, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  29871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29871, CurrentID=0, Status='live', Symbol='Scmh1', Aliases='AI315320, AI851618, Scml1, Scml3', description='sex comb on midleg homolog 1', other_designations='scmh1|sex comb on midleg 1|sex comb on midleg-like 1|sex comb on midleg-like 3', map_location='4 D1-D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120405281, end_position_on_the_genomic_accession=120530199, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  29873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29873, CurrentID=0, Status='live', Symbol='Cspg5', Aliases='Caleb, Ngc', description='chondroitin sulfate proteoglycan 5', other_designations='acidic leucine-rich EGF-like domain-containing brain protein|neuroglycan C', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110243783, end_position_on_the_genomic_accession=110262576, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  29875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29875, CurrentID=0, Status='live', Symbol='Iqgap1', Aliases='AA682088, D7Ertd237e, D7Ertd257e, mKIAA0051', description='IQ motif containing GTPase activating protein 1', other_designations='Cdc42-Rac1 effector protein', map_location='7 45.68 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80711583, end_position_on_the_genomic_accession=80815545, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  29876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29876, CurrentID=0, Status='live', Symbol='Clic4', Aliases='D0Jmb3, TU-74, mc3s5, mtCLIC', description='chloride intracellular channel 4 (mitochondrial)', other_designations='mc3s5/mtCLIC', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135213969, end_position_on_the_genomic_accession=135272760, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  29877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=29877, CurrentID=0, Status='live', Symbol='Hdgfrp3', Aliases='2700022B06Rik, HRP-3', description='hepatoma-derived growth factor, related protein 3', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81881256, end_position_on_the_genomic_accession=81934459, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  554292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=554292, CurrentID=0, Status='live', Symbol='Methig1', Aliases='Mettl7a2-Higd1c, Mettl7a2Higd1c, Mettl7a2higd1c, UbiE-YGHL1', description='methyltransferase hypoxia inducible domain containing 1', other_designations='Mettl7a2-Higd1c readthrough transcript|UbiE-YGHL1 fusion protein|UbiE2-Hig1-4 fusion protein', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100353200, end_position_on_the_genomic_accession=100384435, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  30007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30007, CurrentID=0, Status='live', Symbol='D10Mit174', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 174', other_designations='', map_location='10 38.72 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  554327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=554327, CurrentID=0, Status='live', Symbol='2610042L04Rik', Aliases='Tksn, takusan', description='RIKEN cDNA 2610042L04 gene', other_designations='alpha42-takusan|alpha46-takusan', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=4334380, end_position_on_the_genomic_accession=4354694, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  30044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30044, CurrentID=0, Status='live', Symbol='Opn4', Aliases='1110007J02Rik, Gm533', description='opsin 4 (melanopsin)', other_designations='opsin-4', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34590618, end_position_on_the_genomic_accession=34600142, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  30045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30045, CurrentID=0, Status='live', Symbol='Dnajc12', Aliases='Jdp1, mJDP1', description='DnaJ heat shock protein family (Hsp40) member C12', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 12|J domain protein 1|j domain-containing protein 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63382443, end_position_on_the_genomic_accession=63408840, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  30046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30046, CurrentID=0, Status='live', Symbol='Zfp292', Aliases='5730450D02Rik, 5830493J20Rik, 9430062L07Rik, AI449016, Krox-10, Zfp-15, Zfp15, Zn-15, Zn-16, mKIAA0530', description='zinc finger protein 292', other_designations='zfp-15|zfp292|zinc finger protein 15', map_location='4 16.86 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34803110, end_position_on_the_genomic_accession=34884933, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  30049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30049, CurrentID=0, Status='live', Symbol='Scd3', Aliases='4930513N16Rik', description='stearoyl-coenzyme A desaturase 3', other_designations='delta(9)-desaturase 3|delta-9 desaturase 3|fatty acid desaturase 3|palmitoyl-CoA desaturase|stearoyl-CoA desaturase 3', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44203288, end_position_on_the_genomic_accession=44244016, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  30050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30050, CurrentID=0, Status='live', Symbol='Fbxw2', Aliases='2700071L08Rik, FBW2, Fwd2, MD6', description='F-box and WD-40 domain protein 2', other_designations='F-box and WD-40 domain-containing protein 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34804364, end_position_on_the_genomic_accession=34826273, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  30051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30051, CurrentID=0, Status='live', Symbol='Spdef', Aliases='PDEF, Pse', description='SAM pointed domain containing ets transcription factor', other_designations='prostate epithelium-specific Ets transcription factor|prostate specific ets transcription factor|prostate-derived Ets factor|prostate-specific Ets', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27714446, end_position_on_the_genomic_accession=27729283, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  30052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30052, CurrentID=0, Status='live', Symbol='Pcsk1n', Aliases='AI848336, KEP, PEN, PEN19, PEN20, Pan3, SAAS, SAASCT, bLEN, lLEN, proSAAS', description='proprotein convertase subtilisin/kexin type 1 inhibitor', other_designations='IA-4|basic protein IA-4|big PEN LEN|big SAAS|granin-like neuroendocrine peptide|little LEN|little SAAS|pancreas protein 3|pro-SAAS|proprotein convertase 1 inhibitor|protein serine alanine alanine serine', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7919822, end_position_on_the_genomic_accession=7924410, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  30053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30053, CurrentID=0, Status='live', Symbol='Reg3d', Aliases='INGAP, Ingaprp', description='regenerating islet-derived 3 delta', other_designations='Reg III delta|islet neogenesis associated protein-related protein', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=78375874, end_position_on_the_genomic_accession=78378865, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  30054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30054, CurrentID=0, Status='live', Symbol='Rnf17', Aliases='MMIP-2, Mmip2, TDRD4', description='ring finger protein 17', other_designations='Max dimerization protein member-interacting protein 2|mad member-interacting protein 2', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56402634, end_position_on_the_genomic_accession=56525032, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  30055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30055, CurrentID=0, Status='live', Symbol='Timm13', Aliases='D10Ertd378e, Tim9, Timm9', description='translocase of inner mitochondrial membrane 13', other_designations='translocase of inner mitochondrial membrane 13 homolog|translocase of inner mitochondrial membrane 9 homolog', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80899450, end_position_on_the_genomic_accession=80900969, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  30056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30056, CurrentID=0, Status='live', Symbol='Timm9', Aliases='2810011L15Rik, Tim10a, Timm10', description='translocase of inner mitochondrial membrane 9', other_designations='translocase of inner mitochondrial membrane 10 homolog|translocase of inner mitochondrial membrane 9 homolog', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=71123172, end_position_on_the_genomic_accession=71136675, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  30057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30057, CurrentID=0, Status='live', Symbol='Timm8b', Aliases='Tim8b', description='translocase of inner mitochondrial membrane 8B', other_designations='deafness dystonia protein 2 homolog|translocase of inner mitochondrial membrane 8 homolog b', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50603901, end_position_on_the_genomic_accession=50605320, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  30058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30058, CurrentID=0, Status='live', Symbol='Timm8a1', Aliases='DXHXS1274E, Ddp1, Fci-12, Tim8a, Timm8a', description='translocase of inner mitochondrial membrane 8A1', other_designations='deafness dystonia protein 1 homolog|translocase of inner mitochondrial membrane 8 homolog a|translocase of inner mitochondrial membrane 8 homolog a1', map_location='X 56.18 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134537258, end_position_on_the_genomic_accession=134541629, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  30059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30059, CurrentID=0, Status='live', Symbol='Timm10', Aliases='Tim13, Timm13a', description='translocase of inner mitochondrial membrane 10', other_designations='translocase of inner mitochondrial membrane 10 homolog|translocase of inner mitochondrial membrane 13 homolog a', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84827021, end_position_on_the_genomic_accession=84830213, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  30060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30060, CurrentID=0, Status='live', Symbol='Mfi2', Aliases='CD228, MTf', description='antigen p97 (melanoma associated) identified by monoclonal antibodies 133.2 and 96.5', other_designations='membrane-bound transferrin-like protein p97', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31878501, end_position_on_the_genomic_accession=31899020, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  594844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=594844, CurrentID=0, Status='live', Symbol='Tceal3', Aliases='1100001D19Rik', description='transcription elongation factor A (SII)-like 3', other_designations='TCEA-like protein 3|transcription elongation factor S-II protein-like 3', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136590705, end_position_on_the_genomic_accession=136668377, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100038347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038347, CurrentID=0, Status='live', Symbol='Fam174b', Aliases='', description='family with sequence similarity 174, member B', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=73740307, end_position_on_the_genomic_accession=73776919, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100034858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100034858, CurrentID=0, Status='live', Symbol='Fob3b1', Aliases='Fob3b', description='F-line obesity QTL 3b1', other_designations='F-line obesity QTL 3b', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100038417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038417, CurrentID=0, Status='live', Symbol='Spink13', Aliases='ENSMUSG00000073551, Gm10534, Spink5l3', description='serine peptidase inhibitor, Kazal type 13', other_designations='serine protease inhibitor Kazal-type 5-like 3', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=62605898, end_position_on_the_genomic_accession=62741387, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100038456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038456, CurrentID=0, Status='live', Symbol='Plast1b', Aliases='Plast14b', description='plasma plant sterol 1b', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  26932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=26932, CurrentID=0, Status='live', Symbol='Ppp2r5e', Aliases='4633401M22Rik, AI449017, B56beta', description="protein phosphatase 2, regulatory subunit B', epsilon", other_designations="PP2A B subunit isoform B'-epsilon|PP2A B subunit isoform B56-epsilon|PP2A B subunit isoform PR61-epsilon|PP2A B subunit isoform R5-epsilon|protein phosphatase 2, regulatory subunit B (B56), epsilon isoform|protein phosphatase 2A subunit beta", map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=75449463, end_position_on_the_genomic_accession=75596250, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100038489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038489, CurrentID=0, Status='live', Symbol='Apela', Aliases='ENSMUSG00000074303, Ela, Ende, Gm10664', description='apelin receptor early endogenous ligand', other_designations='Protein Elabela|protein Toddler', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=65028417, end_position_on_the_genomic_accession=65037336, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100038514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038514, CurrentID=0, Status='live', Symbol='Gm11837', Aliases='OTTMUSG00000004551', description='predicted gene 11837', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=14929829, end_position_on_the_genomic_accession=14953031, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  245404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245404, CurrentID=0, Status='live', Symbol='Dcaf12l1', Aliases='9330180J20, A230038L21Rik, Decaf12l1, Wdr40b', description='DDB1 and CUL4 associated factor 12-like 1', other_designations='WD repeat domain 40B|WD repeat-containing protein 40B', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=44786567, end_position_on_the_genomic_accession=44790197, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100038570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038570, CurrentID=0, Status='live', Symbol='Gm11744', Aliases='OTTMUSG00000003947, Prcd', description='predicted gene 11744', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116657108, end_position_on_the_genomic_accession=116668389, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100038638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038638, CurrentID=0, Status='live', Symbol='Fmr1os', Aliases='Asfmr1, ENSMUSG00000073146, Gm10474', description='fragile X mental retardation syndrome 1, opposite strand', other_designations='antisense fragile X mental retardation', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=68667514, end_position_on_the_genomic_accession=68678399, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  30785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30785, CurrentID=0, Status='live', Symbol='Cttnbp2', Aliases='3010022N24Rik, 4732477G22Rik, 6430526E05, 9130022E09Rik, AU040881, Cortbp2, ORF4, mKIAA1758', description='cortactin binding protein 2', other_designations='CTTNBP2 intron retention form|cortactin-binding protein 2 intron retention form', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=18366471, end_position_on_the_genomic_accession=18514907, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  100038725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038725, CurrentID=0, Status='live', Symbol='Cep85l', Aliases='ENSMUSG00000038594, Gm9766', description='centrosomal protein 85-like', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=53276799, end_position_on_the_genomic_accession=53380701, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  30791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30791, CurrentID=0, Status='live', Symbol='Slc39a1', Aliases='Zip1, Zirtl', description='solute carrier family 39 (zinc transporter), member 1', other_designations='ZIP-1|Zrt/Irt-like protein 1|mZIP1|solute carrier family 39 member 1|zrt- and Irt-like protein 1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90248192, end_position_on_the_genomic_accession=90253613, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  30794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30794, CurrentID=0, Status='live', Symbol='Pdlim4', Aliases='Ril', description='PDZ and LIM domain 4', other_designations='LIM protein RIL|reversion induced LIM|reversion-induced LIM protein', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54054928, end_position_on_the_genomic_accession=54069032, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  30795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30795, CurrentID=0, Status='live', Symbol='Fkbp3', Aliases='25kDa, FKBP-3, FKBP25', description='FK506 binding protein 3', other_designations='25 kDa FK506-binding protein|25 kDa FKBP|FK506-binding protein 3 (25kD)|FKBP-25|PPIase FKBP3|immunophilin FKBP25|rapamycin-selective 25 kDa immunophilin|rotamase', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=65062432, end_position_on_the_genomic_accession=65073938, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  30800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30800, CurrentID=0, Status='live', Symbol='Mmp20', Aliases='', description='matrix metallopeptidase 20 (enamelysin)', other_designations='enamel metalloproteinase|matrix metalloproteinase 20 (enamelysin)', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7628231, end_position_on_the_genomic_accession=7674968, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  30805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30805, CurrentID=0, Status='live', Symbol='Slc22a4', Aliases='Octn1', description='solute carrier family 22 (organic cation transporter), member 4', other_designations='LSTP-Like 2|organic cation transporter 1|organic cation/carnitine transporter 1|solute carrier family (organic cation transporter), member 4', map_location='11 32.07 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53983123, end_position_on_the_genomic_accession=54028662, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  30806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30806, CurrentID=0, Status='live', Symbol='Adamts8', Aliases='METH-2, METH2', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 8', other_designations='ADAM-TS 8|ADAM-TS8|ADAMTS-8 precursor (A disintegrin and metalloproteinase with thrombospondin motifs 8) (ADAM-TS 8) (ADAM-TS8) (METH-2)|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 8', map_location='9 16.49 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=30942563, end_position_on_the_genomic_accession=30962858, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  30838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30838, CurrentID=0, Status='live', Symbol='Fbxw4', Aliases='Dac, Fbw4, SHFM3, SHSF3, dactylin, dactylyn', description='F-box and WD-40 domain protein 4', other_designations='F-box and WD-40 domain-containing protein 4|F-box protein Fbw4|dactylaplasia|protein hagoromo', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45578254, end_position_on_the_genomic_accession=45660654, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  30839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30839, CurrentID=0, Status='live', Symbol='Fbxw5', Aliases='AI159739, Fbw5', description='F-box and WD-40 domain protein 5', other_designations='F-box and WD-40 domain-containing protein 5|F-box protein Fbw5', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25500767, end_position_on_the_genomic_accession=25505471, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  30840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30840, CurrentID=0, Status='live', Symbol='Fbxl6', Aliases='AU021795, Fbl6', description='F-box and leucine-rich repeat protein 6', other_designations='F-box protein FBL6|F-box protein containing leucine repeats 6', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76535723, end_position_on_the_genomic_accession=76538746, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  30841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30841, CurrentID=0, Status='live', Symbol='Kdm2b', Aliases='Cxxc2, Fbl10, Fbxl10, Jhdm1b, PCCX2', description='lysine (K)-specific demethylase 2B', other_designations='F-box and leucine-rich repeat protein 10|F-box protein FBL10|F-box protein containing leucine repeats 10|F-box/LRR-repeat protein 10|[Histone-H3]-lysine-36 demethylase 1B|jmjC domain-containing histone demethylation protein 1B', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122870662, end_position_on_the_genomic_accession=122989127, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  30843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30843, CurrentID=0, Status='live', Symbol='Fbxl12', Aliases='3110048D16Rik, Fbl12', description='F-box and leucine-rich repeat protein 12', other_designations='F-box protein FBL12|F-box protein containing leucine repeats 12', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20637749, end_position_on_the_genomic_accession=20644767, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  30853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30853, CurrentID=0, Status='live', Symbol='Mlf2', Aliases='AI256856, AU018944, C77284', description='myeloid leukemia factor 2', other_designations='Myelodysplasia-myeloid leukemia factor 2|myeloid leukemia factor 2 homolog', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124931388, end_position_on_the_genomic_accession=124936149, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100169864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100169864, CurrentID=0, Status='live', Symbol='Gm44504', Aliases='Gm38419, Slc10a3-ubl4, SlcUbl4a, Tu11', description='predicted readthrough transcript (NMD candidate), 44504', other_designations='Slc10a3-Ubl4 readthrough|Slc10a3-Ubl4a readthrough transcript (NMD candidate)|ubiquitin-like protein GDX', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74367447, end_position_on_the_genomic_accession=74373349, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100038806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038806, CurrentID=0, Status='live', Symbol='Sicd1', Aliases='', description='seizure-induced cell death 1', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  30877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30877, CurrentID=0, Status='live', Symbol='Gnl3', Aliases='Ns', description='guanine nucleotide binding protein-like 3 (nucleolar)', other_designations='nucleolar GTP-binding protein 3|nucleostemin', map_location='14 19.09 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31012433, end_position_on_the_genomic_accession=31019131, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  30878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30878, CurrentID=0, Status='live', Symbol='Apln', Aliases='6030430G11Rik, Apel', description='apelin', other_designations='APJ endogenous ligand|preproapelin', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48025146, end_position_on_the_genomic_accession=48034852, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100038859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038859, CurrentID=0, Status='live', Symbol='Olfr55', Aliases='IF6, MOR267-17, MOR267-20', description='olfactory receptor 55', other_designations='GA_x6K02T2MG8L-16493-15546|olfactory receptor MOR267-17', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33176416, end_position_on_the_genomic_accession=33177363, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100038860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038860, CurrentID=0, Status='live', Symbol='Olfr239', Aliases='ENSMUSG00000079308, MOR267-10', description='olfactory receptor 239', other_designations='GA_x6K02T2KN0P-2543-1596|olfactory receptor MOR267-10', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33199062, end_position_on_the_genomic_accession=33200009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  30925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30925, CurrentID=0, Status='live', Symbol='Slamf6', Aliases='KAL1, KAL1b, Ly108, NTB-A, NTBA, SF2000', description='SLAM family member 6', other_designations='lymphocyte antigen 108', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171917199, end_position_on_the_genomic_accession=171953170, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100038862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038862, CurrentID=0, Status='live', Symbol='Btnl1', Aliases='Btnl3, Gm316, Gm33', description='butyrophilin-like 1', other_designations='NG10|butyrophilin-like 3', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34377123, end_position_on_the_genomic_accession=34386162, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  30927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30927, CurrentID=0, Status='live', Symbol='Snai3', Aliases='AI643946, Smuc, Zfp293', description='snail family zinc finger 3', other_designations='protein snail homolog 3|snail homolog 3|snail-related gene from muscle cells|snail-related protein 2|zinc finger protein 293', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122454204, end_position_on_the_genomic_accession=122460927, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  30928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30928, CurrentID=0, Status='live', Symbol='Zbtb18', Aliases='RP58, Zfp238, Znf238', description='zinc finger and BTB domain containing 18', other_designations='58 kDa repressor protein|transcriptional repressor RP58|zfp-238|zinc finger protein 238', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=177442356, end_position_on_the_genomic_accession=177450764, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  30930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30930, CurrentID=0, Status='live', Symbol='Vps26a', Aliases='AA407240, HB58, Vps26', description='vacuolar protein sorting 26 homolog A (yeast)', other_designations='H beta 58|h<beta>58 protein|mVPS26|vacuole protein sorting 26|vesicle protein sorting 26A', map_location='10 B3-B5', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62454843, end_position_on_the_genomic_accession=62486805, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  30931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30931, CurrentID=0, Status='live', Symbol='Tor1a', Aliases='DQ2, Dyt1, torsinA', description='torsin family 1, member A (torsin A)', other_designations='dystonia 1 protein|torsin ATPase 1|torsin family 1 member A', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30960561, end_position_on_the_genomic_accession=30967918, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  30932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30932, CurrentID=0, Status='live', Symbol='Zfp330', Aliases='BC008086, Noa36, Znf330', description='zinc finger protein 330', other_designations='NOA 36|nucleolar autoantigen 36|nucleolar cysteine-rich protein', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=82763620, end_position_on_the_genomic_accession=82774126, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  30933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30933, CurrentID=0, Status='live', Symbol='Tor2a', Aliases='Prosalusin', description='torsin family 2, member A', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32757231, end_position_on_the_genomic_accession=32762244, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  30934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30934, CurrentID=0, Status='live', Symbol='Tor1b', Aliases='2610016F05Rik, DQ1, torsinB', description='torsin family 1, member B', other_designations='torsin ATPase-1B', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30952938, end_position_on_the_genomic_accession=30959015, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  30935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30935, CurrentID=0, Status='live', Symbol='Tor3a', Aliases='Adir', description='torsin family 3, member A', other_designations='ATP-dependant interferon responsive|ATP-dependent interferon-responsive protein', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156653617, end_position_on_the_genomic_accession=156674339, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  30936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30936, CurrentID=0, Status='live', Symbol='Slc46a2', Aliases='5430429N04Rik, Ly110, TSO-1C12, Tscot', description='solute carrier family 46, member 2', other_designations='putative thymic stromal co-transporter TSCOT', map_location='4 32.81 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59905899, end_position_on_the_genomic_accession=59915056, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  30937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30937, CurrentID=0, Status='live', Symbol='Lmcd1', Aliases='AW455500, AW553074, dyxin', description='LIM and cysteine-rich domains 1', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=112273758, end_position_on_the_genomic_accession=112330425, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  30938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30938, CurrentID=0, Status='live', Symbol='Fgd3', Aliases='5830461L01Rik, ZFYVE5', description='FYVE, RhoGEF and PH domain containing 3', other_designations='Fgd1 family, member 3|faciogenital dysplasia homolog 3', map_location='13 25.18 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49263110, end_position_on_the_genomic_accession=49320464, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  30939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30939, CurrentID=0, Status='live', Symbol='Pttg1', Aliases='AW555095, C87862, Pttg, Pttg3', description='pituitary tumor-transforming gene 1', other_designations='pituitary tumor-transforming 1|pituitary tumor-transforming gene 1 protein', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43420248, end_position_on_the_genomic_accession=43426248, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  30940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30940, CurrentID=0, Status='live', Symbol='Usp25', Aliases='', description='ubiquitin specific peptidase 25', other_designations='deubiquitinating enzyme 25|mUSP25|ubiquitin specific protease 25|ubiquitin thioesterase 25|ubiquitin thiolesterase 25|ubiquitin-specific-processing protease 25', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=77013789, end_position_on_the_genomic_accession=77116783, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  30941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30941, CurrentID=0, Status='live', Symbol='Usp21', Aliases='ESTM28, Usp16, Usp23, W53272', description='ubiquitin specific peptidase 21', other_designations='deubiquitinating enzyme 21|ubiquitin specific protease 21|ubiquitin specific protease 23|ubiquitin thioesterase 21|ubiquitin thiolesterase 21|ubiquitin-specific-processing protease 21', map_location='1 79.3 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171281949, end_position_on_the_genomic_accession=171287961, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  30942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30942, CurrentID=0, Status='live', Symbol='Hnf4g', Aliases='NR2A2', description='hepatocyte nuclear factor 4, gamma', other_designations='HNF-4-gamma|nuclear receptor subfamily 2 group A member 2', map_location='3 A', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=3508030, end_position_on_the_genomic_accession=3659800, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  30943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30943, CurrentID=0, Status='live', Symbol='Prss30', Aliases='2010015P21Rik, Disp, Disp1, Tmprss8', description='protease, serine 30', other_designations='distal intestinal serine protease|protease, serine, 30|transmembrane protease, serine 8 (intestinal)|transmembrane serine protease 8', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23972127, end_position_on_the_genomic_accession=23980572, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  30944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30944, CurrentID=0, Status='live', Symbol='Zfp354c', Aliases='5330421P20Rik, AJ18, AW743325, Kid3, Tcf17l2, Znf354c, mKIAA4218', description='zinc finger protein 354C', other_designations='kidney, ischemia, and developmentally-regulated protein 3|transcription factor 17-like 2', map_location='11 30.93 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50811081, end_position_on_the_genomic_accession=50827731, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  30945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30945, CurrentID=0, Status='live', Symbol='Rnf19a', Aliases='AA032313, Dorfin, Rnf19, UIP117, Ubce7ip2, XYbp', description='ring finger protein 19A', other_designations='UBCM4-interacting protein 117|XY body protein|double ring-finger protein|gametogenesis-expressed protein GEG-154|ring finger protein (C3HC4 type) 19|ubiquitin conjugating enzyme 7 interacting protein 2', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36239933, end_position_on_the_genomic_accession=36283723, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  30946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30946, CurrentID=0, Status='live', Symbol='Abt1', Aliases='', description='activator of basal transcription 1', other_designations='', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23418361, end_position_on_the_genomic_accession=23423866, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  30947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30947, CurrentID=0, Status='live', Symbol='Adat1', Aliases='MMADAT1, mADAT1', description='adenosine deaminase, tRNA-specific 1', other_designations='tRNA-specific adenosine-37 deaminase', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111966897, end_position_on_the_genomic_accession=111992307, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  30948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30948, CurrentID=0, Status='live', Symbol='Bin1', Aliases='ALP-1, Amphl, BRAMP-2, SH3P9', description='bridging integrator 1', other_designations='SH3 domain-containing protein 9|amphiphysin 2|amphiphysin II|amphiphysin-like protein|myc box dependent interacting protein 1', map_location='18 18.01 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32377217, end_position_on_the_genomic_accession=32435740, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  30949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30949, CurrentID=0, Status='live', Symbol='Lcmt1', Aliases='AL033290, LCMT-1, Lcmt', description='leucine carboxyl methyltransferase 1', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=123377803, end_position_on_the_genomic_accession=123430358, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  30951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30951, CurrentID=0, Status='live', Symbol='Cbx8', Aliases='Pc3', description='chromobox 8', other_designations='chromobox homolog 8|mPc3|polycomb 3 homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119038436, end_position_on_the_genomic_accession=119040913, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  30952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30952, CurrentID=0, Status='live', Symbol='Cngb3', Aliases='CCNC2, CNG6, Cngbeta2', description='cyclic nucleotide gated channel beta 3', other_designations='CNG channel beta-3|cone photoreceptor cGMP-gated channel subunit beta|cyclic nucleotide gated channel beta 6|cyclic nucleotide-gated cation channel modulatory subunit|cyclic nucleotide-gated channel beta-3|cyclic nucleotide-gated channel subunit CNG6', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=19280850, end_position_on_the_genomic_accession=19510623, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  30953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30953, CurrentID=0, Status='live', Symbol='Schip1', Aliases='Nf2ip, Schip-1', description='schwannomin interacting protein 1', other_designations='neurofibromatosis 2 interacting protein', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=68064802, end_position_on_the_genomic_accession=68626482, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  30954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30954, CurrentID=0, Status='live', Symbol='Siva1', Aliases='CD27bp, Siva', description='SIVA1, apoptosis-inducing factor', other_designations='CD27-binding protein|Cd27 binding protein', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112644828, end_position_on_the_genomic_accession=112649152, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  30955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30955, CurrentID=0, Status='live', Symbol='Pik3cg', Aliases='5830428L06Rik, PI3Kgamma, p110gamma', description='phosphoinositide-3-kinase, catalytic, gamma polypeptide', other_designations='PI3-kinase subunit gamma|PI3K-gamma|p120-PI3K|phosphatidylinositol 4,5-bisphosphate 3-kinase 110 kDa catalytic subunit gamma|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit gamma|phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit gamma isoform|ptdIns-3-kinase subunit gamma|ptdIns-3-kinase subunit p110-gamma|serine/threonine protein kinase PIK3CG', map_location='12 B', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=32173129, end_position_on_the_genomic_accession=32208677, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  30956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30956, CurrentID=0, Status='live', Symbol='Aass', Aliases='LKR, LKR/SDH, LOR, LOR/SDH, Lorsdh, SDH', description='aminoadipate-semialdehyde synthase', other_designations='lysine oxoglutarate reductase|saccharopine dehydrogenase', map_location='6 10.27 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=23072173, end_position_on_the_genomic_accession=23132986, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  30957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30957, CurrentID=0, Status='live', Symbol='Mapk8ip3', Aliases='BB120594, D17Wsu15e, JIP-3, JSAP1, JSAP1a, JSAP1b, JSAP1c, JSAP1d, Jip3, Syd2, mKIAA1066', description='mitogen-activated protein kinase 8 interacting protein 3', other_designations='JNK MAP kinase scaffold protein 3|JNK-interacting protein 3|JNK/SAPK-associated protein 1|JNK/SAPK-associated protein 1c|JUN/SAPK-associated protein 1|c-Jun NH2-terminal kinase (JNK)/stress-activated protein kinase-associated protein 1|mitogen activated protein kinase 8 interacting protein 3|mitogen-activated protein kinase 8-interacting protein 3|sunday driver 2', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24897506, end_position_on_the_genomic_accession=24936977, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  30959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30959, CurrentID=0, Status='live', Symbol='Ddx25', Aliases='AW047046, GRTH', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 25', other_designations='DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 25|DEAD box protein 25|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 25|gonadotropin-regulated testicular RNA helicase', map_location='9 A3-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35541848, end_position_on_the_genomic_accession=35558470, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  30960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30960, CurrentID=0, Status='live', Symbol='Vapa', Aliases='33kDa, VAP33', description='vesicle-associated membrane protein, associated protein A', other_designations='33 kDa VAMP-associated protein|VAMP-A|VAMP-associated protein A|VAP-33|VAP-A|synaptobrevin-associated protein of 33kDa', map_location='17 E1.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65580053, end_position_on_the_genomic_accession=65613937, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  30962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30962, CurrentID=0, Status='live', Symbol='Slc7a9', Aliases='CSNU3', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 9', other_designations='b(0,+)AT|b<o>, + amino acid transporter|b<o>, +AT|glycoprotein-associated amino acid transporter  b0,+AT1|glycoprotein-associated amino acid transporter b0,+AT1|solute carrier family 7 member 9', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35448728, end_position_on_the_genomic_accession=35466034, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  30963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30963, CurrentID=0, Status='live', Symbol='Hacd1', Aliases='Ptpla', description='3-hydroxyacyl-CoA dehydratase 1', other_designations='protein tyrosine phosphatase-like (proline instead of catalytic arginine), member a|protein-tyrosine phosphatase-like member A|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 1', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14026366, end_position_on_the_genomic_accession=14056035, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  30964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30964, CurrentID=0, Status='live', Symbol='Rps12-ps1', Aliases='EG668406, Gm9153, Rps12-ps25, pS12X', description='ribosomal protein S12, pseudogene 1', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103489436, end_position_on_the_genomic_accession=103489936, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100038941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038941, CurrentID=0, Status='live', Symbol='Vmn2r121', Aliases='EG625699', description='vomeronasal 2, receptor 121', other_designations='vomeronasal receptor Vmn2r121', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=124127339, end_position_on_the_genomic_accession=124135910, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100038948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038948, CurrentID=0, Status='live', Symbol='Mup9', Aliases='Gm14076, Mup2, Mup6, OTTMUSG00000015595', description='major urinary protein 9', other_designations='major urinary protein 2|major urinary protein 6', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60418046, end_position_on_the_genomic_accession=60421961, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039008, CurrentID=0, Status='live', Symbol='Mup10', Aliases='2610016E04Rik, Mup8', description='major urinary protein 10', other_designations='major urinary protein 8', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60578260, end_position_on_the_genomic_accession=60582155, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039028, CurrentID=0, Status='live', Symbol='Mup11', Aliases='Gm12549, OTTMUSG00000007431', description='major urinary protein 11', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60658466, end_position_on_the_genomic_accession=60662395, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039042, CurrentID=0, Status='live', Symbol='Gm2016', Aliases='', description='predicted gene 2016', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87874072, end_position_on_the_genomic_accession=87877859, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100039052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039052, CurrentID=0, Status='live', Symbol='Gm2022', Aliases='', description='predicted pseudogene 2022', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87892834, end_position_on_the_genomic_accession=87896722, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100039054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039054, CurrentID=0, Status='live', Symbol='Mup12', Aliases='Gm2024', description='major urinary protein 12', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60737381, end_position_on_the_genomic_accession=60741281, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039060, CurrentID=0, Status='live', Symbol='0610010B08Rik', Aliases='Gm2026', description='RIKEN cDNA 0610010B08 gene', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=175965355, end_position_on_the_genomic_accession=175981734, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100039087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039087, CurrentID=0, Status='live', Symbol='Gm2042', Aliases='', description='predicted gene 2042', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87954480, end_position_on_the_genomic_accession=87960682, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100039089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039089, CurrentID=0, Status='live', Symbol='Mup13', Aliases='Gm13513, Mup11, OTTMUSG00000012492', description='major urinary protein 13', other_designations='major urinary protein (Mup) gene family member|major urinary protein 11', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61224307, end_position_on_the_genomic_accession=61228286, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039116, CurrentID=0, Status='live', Symbol='Mup14', Aliases='Gm13514, Mup12, OTTMUSG00000012493', description='major urinary protein 14', other_designations='major urinary protein (Mup) gene family member|major urinary protein 12', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61300037, end_position_on_the_genomic_accession=61303998, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039150, CurrentID=0, Status='live', Symbol='Mup15', Aliases='Gm2068, Mup13', description='major urinary protein 15', other_designations='major urinary protein 13', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61435787, end_position_on_the_genomic_accession=61439714, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039177, CurrentID=0, Status='live', Symbol='Mup16', Aliases='', description='major urinary protein 16', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61515592, end_position_on_the_genomic_accession=61519503, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039192, CurrentID=0, Status='live', Symbol='Tmem254c', Aliases='ENSMUSG00000072676, Gm10395', description='transmembrane protein 254c', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26063751, end_position_on_the_genomic_accession=26067425, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100039206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039206, CurrentID=0, Status='live', Symbol='Mup17', Aliases='Gm12557, MUP 17, OTTMUSG00000007480', description='major urinary protein 17', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61591929, end_position_on_the_genomic_accession=61595853, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100039227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039227, CurrentID=0, Status='live', Symbol='Duxbl2', Aliases='ENSMUSG00000072675, Gm10394', description='double homeobox B-like 2', other_designations='doubl homeobox B-like 2', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26118690, end_position_on_the_genomic_accession=26129785, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100039246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039246, CurrentID=0, Status='live', Symbol='Plac9b', Aliases='ENSMUSG00000072674, Gm10393', description='placenta specific 9b', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26167316, end_position_on_the_genomic_accession=26182351, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100039257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039257, CurrentID=0, Status='live', Symbol='Tmem254b', Aliases='ENSMUSG00000021867, Gm9746', description='transmembrane protein 254b', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26203366, end_position_on_the_genomic_accession=26207025, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100039293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039293, CurrentID=0, Status='live', Symbol='Duxbl3', Aliases='ENSMUSG00000072672, Gm10391', description='double homeobox B-like 3', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26258303, end_position_on_the_genomic_accession=26269434, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100039315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039315, CurrentID=0, Status='live', Symbol='Gm10436', Aliases='ENSMUSG00000072907', description='predicted gene 10436', other_designations='', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=88175589, end_position_on_the_genomic_accession=88210945, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100039319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039319, CurrentID=0, Status='live', Symbol='Gm14920', Aliases='ENSMUSG00000067441, Gm1539, H2A.Bbd1, H2A.Lap1, OTTMUSG00000018481', description='predicted pseudogene 14920', other_designations='H2A histone family member', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=117014757, end_position_on_the_genomic_accession=117015104, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100039550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039550, CurrentID=0, Status='live', Symbol='Gm10486', Aliases='ENSMUSG00000073247', description='predicted gene 10486', other_designations='Xlr-related, meiosis regulated Xmr', map_location='X A3.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=35067384, end_position_on_the_genomic_accession=35090316, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100039551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039551, CurrentID=0, Status='live', Symbol='Tgif2lx2', Aliases='', description='TGFB-induced factor homeobox 2-like, X-linked 2', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=118427227, end_position_on_the_genomic_accession=118428258, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100039596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039596, CurrentID=0, Status='live', Symbol='Tcf24', Aliases='ENSMUSG00000073738, Gm10567, Gm2330', description='transcription factor 24', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9960163, end_position_on_the_genomic_accession=9968436, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100039674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039674, CurrentID=0, Status='live', Symbol='Gm10634', Aliases='ENSMUSG00000074149', description='predicted gene 10634', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88745262, end_position_on_the_genomic_accession=88830919, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  100039681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039681, CurrentID=0, Status='live', Symbol='Zcwpw2', Aliases='4930430K04Rik', description='zinc finger, CW type with PWWP domain 2', other_designations='', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=117912007, end_position_on_the_genomic_accession=118040247, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100039691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039691, CurrentID=0, Status='live', Symbol='Lncenc1', Aliases='ENSMUSG00000078952, Gm11019, Gm2373, Lincenc1, Platr18, linc-Enc1', description='long non-coding RNA, embryonic stem cells expressed 1', other_designations='long intergenic non-coding RNA, embryonic stem cells expressed 1', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=97434970, end_position_on_the_genomic_accession=97497664, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100039706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039706, CurrentID=0, Status='live', Symbol='Gm2381', Aliases='', description='predicted gene 2381', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=42816827, end_position_on_the_genomic_accession=42867234, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100039707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039707, CurrentID=0, Status='live', Symbol='Mthfsl', Aliases='Gm2382', description='5, 10-methenyltetrahydrofolate synthetase-like', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88688605, end_position_on_the_genomic_accession=88720418, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100039781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039781, CurrentID=0, Status='live', Symbol='Hrct1', Aliases='3830408D24Rik', description='histidine rich carboxyl terminus 1', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43727198, end_position_on_the_genomic_accession=43728110, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  114535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114535, CurrentID=0, Status='live', Symbol='Bbaa1', Aliases='Bb1', description='B.burgdorferi-associated arthritis 1', other_designations='', map_location='4 43.3 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100039790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039790, CurrentID=0, Status='live', Symbol='Trbv3', Aliases='Gm2426, Tcrb-V16', description='T cell receptor beta, variable 3', other_designations='T cell receptor beta variable 3|T-cell receptor beta, variable 16|T-cell receptor beta, variable 3', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41048394, end_position_on_the_genomic_accession=41048824, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100039795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039795, CurrentID=0, Status='live', Symbol='Ildr2', Aliases='2810478N18Rik, 3110063L10Rik, AI852300, D1Ertd471e, Dbsm1, ENSMUSG00000040612, Ll', description='immunoglobulin-like domain containing receptor 2', other_designations='', map_location='1 73.61 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166254094, end_position_on_the_genomic_accession=166316832, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100039796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039796, CurrentID=0, Status='live', Symbol='Tgtp2', Aliases='OTTMUSG00000005537', description='T cell specific GTPase 2', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49057196, end_position_on_the_genomic_accession=49064212, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100039801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039801, CurrentID=0, Status='live', Symbol='Cldn24', Aliases='ENSMUSG00000061974, Gm10107', description='claudin 24', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47822143, end_position_on_the_genomic_accession=47822805, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100039863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039863, CurrentID=0, Status='live', Symbol='Gm13306', Aliases='OTTMUSG00000011352', description='predicted gene 13306', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42153445, end_position_on_the_genomic_accession=42158846, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100039913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039913, CurrentID=0, Status='live', Symbol='Rhox2b', Aliases='OTTMUSG00000017138, Rhox2.2, Rhox2a', description='reproductive homeobox 2B', other_designations='reproductive homeobox 2A', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37411736, end_position_on_the_genomic_accession=37416844, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100039939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039939, CurrentID=0, Status='live', Symbol='Gm2506', Aliases='', description='predicted gene 2506', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100039948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039948, CurrentID=0, Status='live', Symbol='Rhox2c', Aliases='OTTMUSG00000017149, Rhox2.3, Rhox2b', description='reproductive homeobox 2C', other_designations='reproductive homeobox 2B', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37453480, end_position_on_the_genomic_accession=37458417, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100039953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039953, CurrentID=0, Status='live', Symbol='Gfy', Aliases='Gm2515, Gm581, Goofy', description='golgi-associated olfactory signaling regulator', other_designations='protein Goofy', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45176349, end_position_on_the_genomic_accession=45179597, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100039968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100039968, CurrentID=0, Status='live', Symbol='Gm12942', Aliases='OTTMUSG00000009332, Zmym6nb', description='predicted gene 12942', other_designations='ZMYM6 neighbor', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127126043, end_position_on_the_genomic_accession=127129660, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  245638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245638, CurrentID=0, Status='live', Symbol='Tbc1d8b', Aliases='4921505D17Rik, 9030605E16Rik', description='TBC1 domain family, member 8B', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139684960, end_position_on_the_genomic_accession=139754221, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  100040016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040016, CurrentID=0, Status='live', Symbol='Rhox2e', Aliases='ENSMUSG00000063709, OTTMUSG00000017141, Rhox2.5, Rhox2d', description='reproductive homeobox 2E', other_designations='reproductive homeobox 2D', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37530322, end_position_on_the_genomic_accession=37541809, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100040048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040048, CurrentID=0, Status='live', Symbol='Ccl27b', Aliases='Ctack, ENSMUSG00000066202, Scya27b', description='chemokine (C-C motif) ligand 27b', other_designations='small inducible cytokine A27b', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42650604, end_position_on_the_genomic_accession=42656005, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70891, CurrentID=0, Status='live', Symbol='Spdya', Aliases='4921517J08Rik, 4930548B21Rik, GS4, MLZ-465, Spdy1', description='speedy/RINGO cell cycle regulator family, member A', other_designations='RINGO A|mSpy/Ringo A|protein expressed in male leptotene and zygotene spermatocytes 465|rapid inducer of G2/M progression in oocytes A|spdya|speedy A|speedy A1|speedy A2|speedy homolog 1|speedy homolog A|speedy-1|spermatogenesis related protein Gs4|spy1', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71526937, end_position_on_the_genomic_accession=71589535, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  223825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223825, CurrentID=0, Status='live', Symbol='Mroh2b', Aliases='4930455B06Rik, Heatr7b2, SPIF', description='maestro heat-like repeat family member 2B', other_designations='HEAT repeat-containing protein 7B2|XVHEAT repeat family member 7B2|sperm PKA interacting factor|truncated sperm PKA interacting factor', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=4898234, end_position_on_the_genomic_accession=4962207, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  100040223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040223, CurrentID=0, Status='live', Symbol='Gm20831', Aliases='', description='predicted gene, 20831', other_designations='Ssty1 family member', map_location='Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=17965160, end_position_on_the_genomic_accession=17967601, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  140741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=140741, CurrentID=0, Status='live', Symbol='Gpr6', Aliases='AI852874, Gm233', description='G protein-coupled receptor 6', other_designations='sphingosine 1-phosphate receptor GPR6', map_location='10 22.08 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41069974, end_position_on_the_genomic_accession=41072698, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100040268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040268, CurrentID=0, Status='live', Symbol='Olfr157', Aliases='Mor262-12, Olfr37c, Or37c', description='olfactory receptor 157', other_designations='GA_x6K02T2N78B-16110014-16110970|olfactory receptor 37c|olfactory receptor MOR262-12', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43834752, end_position_on_the_genomic_accession=43837643, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100040294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040294, CurrentID=0, Status='live', Symbol='Gm2694', Aliases='C230036H16Rik, linc1582', description='predicted gene 2694', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=87472812, end_position_on_the_genomic_accession=87525554, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105827, CurrentID=0, Status='live', Symbol='Amigo2', Aliases='AI415330, AMIGO-2, AW208913, Ali1', description='adhesion molecule with Ig like domain 2', other_designations='alivin 1|alivin-1|amigo2|amphoterin induced gene and ORF 2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97244074, end_position_on_the_genomic_accession=97247287, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100302593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100302593, CurrentID=0, Status='live', Symbol='Snord14d', Aliases='Gm23710', description='small nucleolar RNA, C/D box 14D', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40803995, end_position_on_the_genomic_accession=40804081, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100302594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100302594, CurrentID=0, Status='live', Symbol='Snord14e', Aliases='', description='small nucleolar RNA, C/D box 14E', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40804748, end_position_on_the_genomic_accession=40804838, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100040462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040462, CurrentID=0, Status='live', Symbol='Mndal', Aliases='Gm2785', description='myeloid nuclear differentiation antigen like', other_designations='mndal', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173857220, end_position_on_the_genomic_accession=173880260, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100040531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040531, CurrentID=0, Status='live', Symbol='Dynlt1f', Aliases='100040631, Dynlt1e', description='dynein light chain Tctex-type 1F', other_designations='dynein light chain Tctex-type 1E', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6646606, end_position_on_the_genomic_accession=6655939, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100040563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040563, CurrentID=0, Status='live', Symbol='Dynlt1c', Aliases='100040603, Dynlt1d', description='dynein light chain Tctex-type 1C', other_designations='dynein light chain Tctex-type 1D', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6601779, end_position_on_the_genomic_accession=6607891, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100302730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100302730, CurrentID=0, Status='live', Symbol='5830417I10Rik', Aliases='3632445O20Rik, AA387005', description='RIKEN cDNA 5830417I10 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88777057, end_position_on_the_genomic_accession=88829371, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100040591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040591, CurrentID=0, Status='live', Symbol='Kcnj13', Aliases='', description='potassium inwardly-rectifying channel, subfamily J, member 13', other_designations='inward rectifier K(+) channel Kir7.1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87386284, end_position_on_the_genomic_accession=87394729, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100040599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040599, CurrentID=0, Status='live', Symbol='Gm15319', Aliases='OTTMUSG00000020118', description='predicted gene 15319', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=20337664, end_position_on_the_genomic_accession=20363202, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100040608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040608, CurrentID=0, Status='live', Symbol='Fancf', Aliases='A730016A17', description='Fanconi anemia, complementation group F', other_designations='', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=51860577, end_position_on_the_genomic_accession=51862267, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100040671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040671, CurrentID=0, Status='live', Symbol='Gm2897', Aliases='ENSMUSG00000079410', description='predicted gene 2897', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=3049285, end_position_on_the_genomic_accession=3085663, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100040724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040724, CurrentID=0, Status='live', Symbol='Mirg', Aliases='2810474H01Rik, AI852992, Gm2928, Meg9, mir-410', description='miRNA containing gene', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109734981, end_position_on_the_genomic_accession=109749458, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100040766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040766, CurrentID=0, Status='live', Symbol='Mroh2a', Aliases='ENSMUSG00000044873, Heatr7b1, OTTMUSG00000020804', description='maestro heat-like repeat family member 2A', other_designations='HEAT repeat containing 7B1|HEAT repeat-containing protein 7B1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88226988, end_position_on_the_genomic_accession=88262289, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  100040894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040894, CurrentID=0, Status='live', Symbol='Gm16430', Aliases='EG666120', description='predicted gene 16430', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=54622211, end_position_on_the_genomic_accession=54639016, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100040926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040926, CurrentID=0, Status='live', Symbol='Gm3045', Aliases='', description='predicted gene 3045', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56425679, end_position_on_the_genomic_accession=56434099, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100040972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100040972, CurrentID=0, Status='live', Symbol='Tceal7', Aliases='1110018P05Rik', description='transcription elongation factor A (SII)-like 7', other_designations='TCEA-like protein 7|transcription elongation factor S-II protein-like 7', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136214914, end_position_on_the_genomic_accession=136226100, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100041085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041085, CurrentID=0, Status='live', Symbol='Mfsd4b3', Aliases='G630090E17Rik, Naglt1c', description='major facilitator superfamily domain containing 4B3', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39946912, end_position_on_the_genomic_accession=39960153, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100041102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041102, CurrentID=0, Status='live', Symbol='Gm3139', Aliases='', description='predicted gene 3139', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=94478018, end_position_on_the_genomic_accession=94538367, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100041179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041179, CurrentID=0, Status='live', Symbol='Gm3183', Aliases='', description='predicted gene 3183', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=94812316, end_position_on_the_genomic_accession=94877205, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100041194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041194, CurrentID=0, Status='live', Symbol='Ahnak2', Aliases='AI450948, Gm1185, Gm72', description='AHNAK nucleoprotein 2', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112772906, end_position_on_the_genomic_accession=112802660, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100041253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041253, CurrentID=0, Status='live', Symbol='Gm16513', Aliases='', description='predicted gene, Gm16513', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=95139278, end_position_on_the_genomic_accession=95195342, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100041294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041294, CurrentID=0, Status='live', Symbol='Supt4b', Aliases='Gm3258, Supt4h1b, Supt4h2', description='suppressor of Ty 4B', other_designations='DRB sensitivity-inducing factor small subunit 2|DSIF small subunit 2|suppressor of Ty 4 homolog 2|transcription elongation factor SPT4 2', map_location='10 17.4 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=31413726, end_position_on_the_genomic_accession=31414435, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100041352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041352, CurrentID=0, Status='live', Symbol='Tcp10c', Aliases='D17Leh66C, D17Leh66ca, Gm9880, T66C-a, Tcp-10c', description='t-complex protein 10c', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13354572, end_position_on_the_genomic_accession=13377223, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100041354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041354, CurrentID=0, Status='live', Symbol='Gm3286', Aliases='', description='predicted gene 3286', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=95510742, end_position_on_the_genomic_accession=95522402, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100041375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041375, CurrentID=0, Status='live', Symbol='Cyp3a41b', Aliases='', description='cytochrome P450, family 3, subfamily a, polypeptide 41B', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145558664, end_position_on_the_genomic_accession=145584730, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100041379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041379, CurrentID=0, Status='live', Symbol='Gm13242', Aliases='OTTMUSG00000011070', description='predicted gene 13242', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=145670697, end_position_on_the_genomic_accession=145704437, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100041420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041420, CurrentID=0, Status='live', Symbol='Gm3325', Aliases='', description='predicted gene 3325', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=62015654, end_position_on_the_genomic_accession=62052311, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100041433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041433, CurrentID=0, Status='live', Symbol='Gm13247', Aliases='OTTMUSG00000011075', description='predicted gene 13247', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=146502000, end_position_on_the_genomic_accession=146539395, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100041449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041449, CurrentID=0, Status='live', Symbol='Cyp3a59', Aliases='ENSMUSG00000061292', description='cytochrome P450, family 3, subfamily a, polypeptide 59', other_designations='cytochrome P450, subfamily 3A, polypeptide 59', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146079209, end_position_on_the_genomic_accession=146113791, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100041505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041505, CurrentID=0, Status='live', Symbol='Gm3376', Aliases='Rbmy1b', description='predicted gene 3376', other_designations='RNA binding motif protein, Y chromosome, family 1 member|RNA-binding motif protein 1', map_location='Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=3771673, end_position_on_the_genomic_accession=3783271, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  100041546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041546, CurrentID=0, Status='live', Symbol='Ly6c2', Aliases='Ly-6C.2, Ly-6C2', description='lymphocyte antigen 6 complex, locus C2', other_designations='lymphocyte antigen 6C precursor (Ly-6C)|lymphocyte differentiation antigen', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75108154, end_position_on_the_genomic_accession=75112004, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100041554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041554, CurrentID=0, Status='live', Symbol='Gm3404', Aliases='', description='predicted gene 3404', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146525783, end_position_on_the_genomic_accession=146528553, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100041574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041574, CurrentID=0, Status='live', Symbol='9030025P20Rik', Aliases='D17Zt9e', description='RIKEN cDNA 9030025P20 gene', other_designations='', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14978864, end_position_on_the_genomic_accession=14991577, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100041579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041579, CurrentID=0, Status='live', Symbol='Gm20878', Aliases='', description='predicted gene, 20878', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41941604, end_position_on_the_genomic_accession=41942633, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100041581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041581, CurrentID=0, Status='live', Symbol='Zkscan16', Aliases='XM_143800, Zfp483', description='zinc finger with KRAB and SCAN domains 16', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58943568, end_position_on_the_genomic_accession=58962624, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100041585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041585, CurrentID=0, Status='live', Symbol='Amd2', Aliases='AdoMetDC, Amd1, Amd3, ENSMUSG00000063953', description='S-adenosylmethionine decarboxylase 2', other_designations='SAMDC 2|adenosylmethionine decarboxylase 2|adoMetDC 2', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=35708691, end_position_on_the_genomic_accession=35711891, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100041586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041586, CurrentID=0, Status='live', Symbol='Gm3417', Aliases='', description='predicted gene 3417', other_designations='', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14964189, end_position_on_the_genomic_accession=14978812, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100041596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041596, CurrentID=0, Status='live', Symbol='Gm9994', Aliases='ENSMUSG00000056307, Gal3ST-2, Gal3st2', description='predicted gene 9994', other_designations="beta-galactose-3-O-sulfotransferase 2|gal-beta-1, 3-GalNAc 3'-sulfotransferase 2", map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93918437, end_position_on_the_genomic_accession=93942645, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100303744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100303744, CurrentID=0, Status='live', Symbol='Sprr2a2', Aliases='', description='small proline-rich protein 2A2', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92250647, end_position_on_the_genomic_accession=92257304, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100041621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041621, CurrentID=0, Status='live', Symbol='Gm3435', Aliases='', description='predicted gene 3435', other_designations='', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15009840, end_position_on_the_genomic_accession=15022449, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100041658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041658, CurrentID=0, Status='live', Symbol='Mup7', Aliases='Gm12546, Mup3, OTTMUSG00000007428', description='major urinary protein 7', other_designations='major urinary protein gene family member 3', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60066469, end_position_on_the_genomic_accession=60070481, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100041677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041677, CurrentID=0, Status='live', Symbol='Gm13157', Aliases='OTTMUSG00000010679', description='predicted gene 13157', other_designations='novel KRAB box domain containing protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147753974, end_position_on_the_genomic_accession=147810843, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100041678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041678, CurrentID=0, Status='live', Symbol='Gm3500', Aliases='100041758, Gm3401', description='predicted gene 3500', other_designations='alpha11-takusan|alpha12-takusan', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=5365792, end_position_on_the_genomic_accession=5389224, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100041687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041687, CurrentID=0, Status='live', Symbol='Mup8', Aliases='Gm12809, Mup5, OTTMUSG00000008509', description='major urinary protein 8', other_designations='major urinary protein (Mup) gene family member|major urinary protein 5', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60218620, end_position_on_the_genomic_accession=60222599, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  224090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224090, CurrentID=0, Status='live', Symbol='Tmem44', Aliases='1700007N03Rik, 9330161C17Rik, B230220N21', description='transmembrane protein 44', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30511855, end_position_on_the_genomic_accession=30550834, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100041774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041774, CurrentID=0, Status='live', Symbol='Gm10413', Aliases='100038846, 1700001E04Rik, ENSMUSG00000072749, Gm1973', description='predicted gene 10413', other_designations='', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=3395093, end_position_on_the_genomic_accession=3403534, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  245944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245944, CurrentID=0, Status='live', Symbol='Vps54', Aliases='Hcc8l, mSLP8, wr, Vps54', description='vacuolar protein sorting 54 (yeast)', other_designations='tumor antigen SLP-8p homolog', map_location='11 13.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=21239032, end_position_on_the_genomic_accession=21321133, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  245945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245945, CurrentID=0, Status='live', Symbol='Rbm47', Aliases='9530077J19Rik', description='RNA binding motif protein 47', other_designations='RNA-binding motif protein 47', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=66016549, end_position_on_the_genomic_accession=66151954, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100041932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041932, CurrentID=0, Status='live', Symbol='Gm3579', Aliases='H3097D08', description='predicted gene 3579', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133095962, end_position_on_the_genomic_accession=133101555, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100041953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100041953, CurrentID=0, Status='live', Symbol='Gm10094', Aliases='ENSMUSG00000061104', description='predicted gene 10094', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95825356, end_position_on_the_genomic_accession=95826156, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100042150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042150, CurrentID=0, Status='live', Symbol='Nrg2', Aliases='Don1, NTAK', description='neuregulin 2', other_designations='pro-NRG2', map_location='18 19.38 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36016758, end_position_on_the_genomic_accession=36199047, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100042165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042165, CurrentID=0, Status='live', Symbol='BC005561', Aliases='', description='cDNA sequence BC005561', other_designations='', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104508352, end_position_on_the_genomic_accession=104522383, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100042437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042437, CurrentID=0, Status='live', Symbol='Vmn1r79', Aliases='Gm9807', description='vomeronasal 1 receptor 79', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12176193, end_position_on_the_genomic_accession=12177113, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100042450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042450, CurrentID=0, Status='live', Symbol='Smim17', Aliases='Gm16532', description='small integral membrane protein 17', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6415175, end_position_on_the_genomic_accession=6431086, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100042480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042480, CurrentID=0, Status='live', Symbol='Nhsl2', Aliases='1110062M06Rik, 6430511F03, AI225852, AW488865, EG621083, ENSMUSG00000073037, Gm10456', description='NHS-like 2', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101848669, end_position_on_the_genomic_accession=102092055, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100042493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042493, CurrentID=0, Status='live', Symbol='Ccl21b', Aliases='6CKBAC1, 6Ckine, AI159701, CKb9, SLC, Scya21, Scya21a, TCA4', description='chemokine (C-C motif) ligand 21B (leucine)', other_designations='beta chemokine exodus-2|beta-chemokine exodus-2|exodus-2|secondary lymphoid organ chemokine|small inducible cytokine A21|small-inducible cytokine A21a|small-inducible cytokine A21b|thymus-derived chemotactic agent 4', map_location='4 22.12 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42255731, end_position_on_the_genomic_accession=42256825, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100042782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042782, CurrentID=0, Status='live', Symbol='Fthl17d', Aliases='Fthl17L3, Gm14458, OTTMUSG00000016694', description='ferritin, heavy polypeptide-like 17, member D', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8985940, end_position_on_the_genomic_accession=8986672, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100042784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042784, CurrentID=0, Status='live', Symbol='Prdm11', Aliases='8030443D09Rik, AA163972, Gm707', description='PR domain containing 11', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92974906, end_position_on_the_genomic_accession=93046144, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100042840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042840, CurrentID=0, Status='live', Symbol='Gm16501', Aliases='ENSMUSG00000069034, H2al2y', description='predicted gene 16501', other_designations='', map_location='Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=2720339, end_position_on_the_genomic_accession=2720674, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100042856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100042856, CurrentID=0, Status='live', Symbol='Gm4070', Aliases='VLIG-1', description='predicted gene 4070', other_designations='very large inducible GTPase-1|very large-inducible GTPase-1', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105895118, end_position_on_the_genomic_accession=105954002, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100043034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043034, CurrentID=0, Status='live', Symbol='Rex2', Aliases='Gm13138, OTTMUSG00000010655', description='reduced expression 2', other_designations='novel KRAB box domain containing protein', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147021850, end_position_on_the_genomic_accession=147060799, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100043057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043057, CurrentID=0, Status='live', Symbol='Tmem207', Aliases='Gm540', description='transmembrane protein 207', other_designations='tmem207', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=26472713, end_position_on_the_genomic_accession=26526839, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100043188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043188, CurrentID=0, Status='live', Symbol='Gm10696', Aliases='ENSMUSG00000074424', description='predicted gene 10696', other_designations='', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94174412, end_position_on_the_genomic_accession=94178193, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100043194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043194, CurrentID=0, Status='live', Symbol='Sult2a2', Aliases='C730007P19Rik, ST, ST2A2, Sth2, mSTa2', description='sulfotransferase family 2A, dehydroepiandrosterone (DHEA)-preferring, member 2', other_designations='hydroxysteroid sulfotransferase|sulfotransferase 2A2|sulfotransferase, hydroxysteroid preferring 2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13733506, end_position_on_the_genomic_accession=13779637, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100043207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043207, CurrentID=0, Status='live', Symbol='Amy2a1', Aliases='Amy2-1, OTTMUSG00000022432, mAmy2-6', description='amylase 2a1', other_designations='amylase 2-1, pancreatic|amylase 2a1, pancreatic', map_location='3 49.35 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113528095, end_position_on_the_genomic_accession=113532428, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100043214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043214, CurrentID=0, Status='live', Symbol='Trav14d-2', Aliases='EG667553, Gm16454, Tcra-V2.1', description='T cell receptor alpha variable 14D-2', other_designations='T-cell receptor alpha chain variable regions 14D-2|T-cell receptor alpha variable 14D-2|T-cell receptor alpha, variable 2.1', map_location='14 27.16 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52948668, end_position_on_the_genomic_accession=52949169, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100043254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043254, CurrentID=0, Status='live', Symbol='Nps', Aliases='ENSMUSG00000073804', description='neuropeptide S', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=135268619, end_position_on_the_genomic_accession=135272942, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100043314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043314, CurrentID=0, Status='live', Symbol='Tigit', Aliases='ENSMUSG00000071552, Vstm3', description='T cell immunoreceptor with Ig and ITIM domains', other_designations='V-set and transmembrane domain containing 3|V-set and transmembrane domain-containing protein 3|tigit', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=43648861, end_position_on_the_genomic_accession=43664184, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100043322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043322, CurrentID=0, Status='live', Symbol='Trav12-2', Aliases='', description='T cell receptor alpha variable 12-2', other_designations='T-cell receptor alpha chain V region 12-2', map_location='14 27.44 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53616397, end_position_on_the_genomic_accession=53616914, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100043326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043326, CurrentID=0, Status='live', Symbol='Scgb2b3', Aliases='Abpbg3, Gm4362', description='secretoglobin, family 2B, member 3', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31359038, end_position_on_the_genomic_accession=31362072, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100043332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043332, CurrentID=0, Status='live', Symbol='Ankrd66', Aliases='AU041848, D730048J04Rik, Gm4368', description='ankyrin repeat domain 66', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43534173, end_position_on_the_genomic_accession=43570729, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71449, CurrentID=0, Status='live', Symbol='Mettl13', Aliases='5630401D24Rik', description='methyltransferase like 13', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162533671, end_position_on_the_genomic_accession=162548345, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100043456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043456, CurrentID=0, Status='live', Symbol='Gm10681', Aliases='ENSMUSG00000074380', description='predicted gene 10681', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98552542, end_position_on_the_genomic_accession=98564009, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100043461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043461, CurrentID=0, Status='live', Symbol='Gm4450', Aliases='', description='predicted gene 4450', other_designations='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase-like', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98445675, end_position_on_the_genomic_accession=98457068, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100043468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043468, CurrentID=0, Status='live', Symbol='Zfp955b', Aliases='A430003O12Rik, C430039G02Rik, C530035A18Rik, Gm4455, Zfp883b', description='zinc finger protein 955B', other_designations='zinc finger protein 883B', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33289517, end_position_on_the_genomic_accession=33304689, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  333605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333605, CurrentID=0, Status='live', Symbol='Frmpd4', Aliases='Gm196, PKAP1, Pdzd10, Pdzk10, Preso1, preso', description='FERM and PDZ domain containing 4', other_designations='PDZ domain-containing protein 10|PSD-95-interacting regulator of spine morphogenesis', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=167471301, end_position_on_the_genomic_accession=168577244, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  100043497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043497, CurrentID=0, Status='live', Symbol='Nat8f7', Aliases='Gm4477', description='N-acetyltransferase 8 (GCN5-related) family member 7', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85706707, end_position_on_the_genomic_accession=85713199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100043597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043597, CurrentID=0, Status='live', Symbol='Srcap', Aliases='B930091H02Rik, D030022P06Rik, F630004O05Rik', description='Snf2-related CREBBP activator protein', other_designations='Snf2-related CBP activator protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127511983, end_position_on_the_genomic_accession=127561219, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  100043645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043645, CurrentID=0, Status='live', Symbol='Gm4567', Aliases='EG381936, Gm5156', description='predicted gene 4567', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=22713598, end_position_on_the_genomic_accession=22719710, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100043684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043684, CurrentID=0, Status='live', Symbol='Amy2a4', Aliases='ENSMUSG00000074267, OTTMUSG00000022433, mAmy2-3', description='amylase 2a4', other_designations='amylase 2a4, pancreatic', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113279837, end_position_on_the_genomic_accession=113291449, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100043686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043686, CurrentID=0, Status='live', Symbol='Amy2a3', Aliases='OTTMUSG00000022430, mAmy2-4', description='amylase 2a3', other_designations='amylase 2a3, pancreatic', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113312449, end_position_on_the_genomic_accession=113324052, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100043688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043688, CurrentID=0, Status='live', Symbol='Amy2a2', Aliases='ENSMUSG00000074265, OTTMUSG00000022431, mAmy2-5', description='amylase 2a2', other_designations='amylase 2a2, pancreatic', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113345049, end_position_on_the_genomic_accession=113356658, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100043757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043757, CurrentID=0, Status='live', Symbol='Zfp831', Aliases='ENSMUSG00000050600, OTTMUSG00000017459', description='zinc finger protein 831', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174598687, end_position_on_the_genomic_accession=174710832, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100043836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043836, CurrentID=0, Status='live', Symbol='Scgb2b7', Aliases='Abpbg7, Gm4684', description='secretoglobin, family 2B, member 7', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31703779, end_position_on_the_genomic_accession=31705754, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100043842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043842, CurrentID=0, Status='live', Symbol='Scgb2b11', Aliases='Abpbg11, Gm4687', description='secretoglobin, family 2B, member 11', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=32209335, end_position_on_the_genomic_accession=32211123, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100043856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043856, CurrentID=0, Status='live', Symbol='Scgb2b18', Aliases='Abpbg18, Gm4692', description='secretoglobin, family 2B, member 18', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33172054, end_position_on_the_genomic_accession=33173842, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100043860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043860, CurrentID=0, Status='live', Symbol='Scgb1b24', Aliases='Abpa24, Gm12769, OTTMUSG00000008385', description='secretoglobin, family 1B, member 24', other_designations='androgen-binding protein|novel androgen binding protein (Abp) family member', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33743799, end_position_on_the_genomic_accession=33745103, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100043861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043861, CurrentID=0, Status='live', Symbol='Klrb1', Aliases='4930431A04Rik, Gm4696g, Klrb6, Ly-55, Ly55, NKR-P1G, Nkrp-1e, Nkrp1g, Klrb1', description='killer cell lectin-like receptor subfamily B member 1', other_designations='immunoreceptor NKR-P1G|killer cell lectin-like receptor subfamily A member 1 complex|killer cell lectin-like receptor subfamily B member 1 complex|killer cell lectin-like receptor subfamily B member 1G|lectin-like transmembrane protein Nkrp1g|lymphocyte antigen 55|natural killer cell surface protein NKR-P1G|natural killer lectin-like receptor 1E|natural killer lectin-like receptor protein 1 e', map_location='6 63.08 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128706093, end_position_on_the_genomic_accession=128764561, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100043868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043868, CurrentID=0, Status='live', Symbol='Scgb1b30', Aliases='Abpa30, Gm12776, OTTMUSG00000008394', description='secretoglobin, family 1B, member 30', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34095431, end_position_on_the_genomic_accession=34100837, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100043902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043902, CurrentID=0, Status='live', Symbol='Six3os1', Aliases='D17Mgi26, E130112H22Rik, Rncr1, Six3os', description='SIX homeobox 3, opposite strand 1', other_designations='Six3 antisense|Six3 opposite strand transcript 1', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=85601932, end_position_on_the_genomic_accession=85618396, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100043915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043915, CurrentID=0, Status='live', Symbol='Gm4724', Aliases='', description='predicted gene 4724', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=175419391, end_position_on_the_genomic_accession=175435777, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100043989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100043989, CurrentID=0, Status='live', Symbol='Vh7183.a34psi.56', Aliases='IGHV5-7-3', description='immunoglobulin variable segment pseudogene', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100568459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100568459, CurrentID=0, Status='live', Symbol='Bc1', Aliases='Gm24367', description='brain cytoplasmic RNA 1', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144914470, end_position_on_the_genomic_accession=144914623, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100044236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100044236, CurrentID=0, Status='live', Symbol='Copg2os2', Aliases='AW049845, Copg2as2, Lb9, Mit1', description='coatomer protein complex, subunit gamma 2, opposite strand 2', other_designations='Mest-linked imprinted transcript 1|coatomer protein complex, subunit gamma 2, antisense 2', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=692152, CurrentID=0, Status='live', Symbol='Igkv13-84', Aliases='Igk-V33, gm33', description='immunoglobulin kappa chain variable 13-84', other_designations='immunoglobulin kappa chain variable 33 (V33)', map_location='6 31.34 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68939602, end_position_on_the_genomic_accession=68940067, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  692161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=692161, CurrentID=0, Status='live', Symbol='Igkv19-93', Aliases='Igk-V38, Vx38C', description='immunoglobulin kappa chain variable 19-93', other_designations='immunoglobulin kappa chain variable 38(V38)|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68736297, end_position_on_the_genomic_accession=68736764, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  692169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=692169, CurrentID=0, Status='live', Symbol='Igkv15-103', Aliases='Igk-V32', description='immunoglobulin kappa chain variable 15-103', other_designations='immunoglobulin kappa chain variable 32 (V32)|immunoglobulin light chain variable region', map_location='6 31.18 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68437468, end_position_on_the_genomic_accession=68437925, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  692179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=692179, CurrentID=0, Status='live', Symbol='Igk-V21', Aliases='', description='immunoglobulin kappa chain variable 21 (V21)', other_designations='Ig kappa chain|rheumatoid factor RFA28', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  692182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=692182, CurrentID=0, Status='live', Symbol='Igkv9-129', Aliases='', description='immunoglobulin kappa variable 9-129', other_designations='Ig kappa|immunoglobulin kappa chain variable 9-129|immunoglobulin kappa light chain', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67839793, end_position_on_the_genomic_accession=67840266, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  224742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224742, CurrentID=0, Status='live', Symbol='Abcf1', Aliases='AU041969, Abc50, D17Wsu166e, GCN20', description='ATP-binding cassette, sub-family F (GCN20), member 1', other_designations='', map_location='17 18.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35956816, end_position_on_the_genomic_accession=35969800, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  93681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93681, CurrentID=0, Status='live', Symbol='Zkscan8', Aliases='2510038J07Rik, D430019P06Rik, LD5-1, Zfp192', description='zinc finger with KRAB and SCAN domains 8', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21513221, end_position_on_the_genomic_accession=21531114, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  93684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93684, CurrentID=0, Status='live', Symbol='Sep15', Aliases='9430015P09Rik', description='selenoprotein 15', other_designations='selenoprotein', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144570427, end_position_on_the_genomic_accession=144597677, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100045778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100045778, CurrentID=0, Status='live', Symbol='Rnf223', Aliases='9930005F22Rik', description='ring finger 223', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156127371, end_position_on_the_genomic_accession=156133422, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  224761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224761, CurrentID=0, Status='live', Symbol='H2-M10.5', Aliases='6.9H', description='histocompatibility 2, M region locus 10.5', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36772910, end_position_on_the_genomic_accession=36776235, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  93690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93690, CurrentID=0, Status='live', Symbol='Gpr45', Aliases='9230112G11Rik, AW046772', description='G protein-coupled receptor 45', other_designations='PSP24-1|PSP24-alpha|PSP24alpha', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=42952872, end_position_on_the_genomic_accession=43035451, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100045792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100045792, CurrentID=0, Status='live', Symbol='Ect2l', Aliases='C330021H03Rik, ENSMUSG00000071392, Gm10331', description='epithelial cell transforming sequence 2 oncogene-like', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18129022, end_position_on_the_genomic_accession=18210890, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  93722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93722, CurrentID=0, Status='live', Symbol='Pcdhga10', Aliases='', description='protocadherin gamma subfamily A, 10', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37747188, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  224796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224796, CurrentID=0, Status='live', Symbol='Clic5', Aliases='5730531E12Rik, B330005L24, Gm322, jbg', description='chloride intracellular channel 5', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=44134678, end_position_on_the_genomic_accession=44280168, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  93735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93735, CurrentID=0, Status='live', Symbol='Wnt16', Aliases='E130309I19Rik', description='wingless-type MMTV integration site family, member 16', other_designations='wingless-related MMTV integration site 16', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=22288165, end_position_on_the_genomic_accession=22298522, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  93742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93742, CurrentID=0, Status='live', Symbol='Pard3', Aliases='AA960621, AI256638, Asip, D8Ertd580e, Par-3, Par3, Pard-3a, Phip, Pard3', description='par-3 family cell polarity regulator', other_designations='atypical PKC isotype-specific-interacting protein|ephrin-interacting protein|par-3 (partitioning defective 3) homolog', map_location='8 74.66 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=127064056, end_position_on_the_genomic_accession=127612286, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  93757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93757, CurrentID=0, Status='live', Symbol='Immp2l', Aliases='AI853880, IMP2', description='IMP2 inner mitochondrial membrane peptidase-like (S. cerevisiae)', other_designations='IMP2-like protein', map_location='12 B3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=41628764, end_position_on_the_genomic_accession=41955588, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  93761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93761, CurrentID=0, Status='live', Symbol='Smarca1', Aliases='5730494M04Rik, Snf2l', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 1', other_designations='ATP-dependent helicase SMARCA1|DNA-dependent ATPase SNF2L|nucleosome-remodeling factor subunit SNF2L', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=47809370, end_position_on_the_genomic_accession=47892931, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  268527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=268527, CurrentID=0, Status='live', Symbol='Greb1', Aliases='5730583K22Rik, 9130004E13, AF180470, AU023194, mKIAA0575', description='gene regulated by estrogen in breast cancer protein', other_designations='', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=16670615, end_position_on_the_genomic_accession=16801397, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  101488143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101488143, CurrentID=0, Status='live', Symbol='Hbb-bt', Aliases='Beta-t', description='hemoglobin, beta adult t chain', other_designations='beta t', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103812524, end_position_on_the_genomic_accession=103813923, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  28250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=28250, CurrentID=0, Status='live', Symbol='Slco1a4', Aliases='AI785519, Oatp1a4, Oatp2, Slc21a5', description='solute carrier organic anion transporter family, member 1a4', other_designations='organic anion transporting polypeptide 2|sodium-independent organic anion-transporting polypeptide 2|solute carrier family 21 (organic anion transporter), member 5|solute carrier family 21 member 5', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=141805438, end_position_on_the_genomic_accession=141856283, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  101488212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101488212, CurrentID=0, Status='live', Symbol='Gm21975', Aliases='Evi2a-evi2b', description='predicted gene 21975', other_designations='Evi2a-Evi2b readthrough|ecotropic viral integration site 2B protein', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79513385, end_position_on_the_genomic_accession=79530609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  93835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93835, CurrentID=0, Status='live', Symbol='Amn', Aliases='', description='amnionless', other_designations='type I transmembrane protein', map_location='12 60.94 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111271111, end_position_on_the_genomic_accession=111276426, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  93840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93840, CurrentID=0, Status='live', Symbol='Vangl2', Aliases='C530001F03Rik, Lp, Lpp1, Ltap, Vang1l2, loop-tail, ska17, ska<m17Jus>, stbm, strabismus', description='vang-like 2 (van gogh, Drosophila)', other_designations='loop tail associated protein|loop-tail protein 1|van Gogh-like protein 2', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172004360, end_position_on_the_genomic_accession=172027314, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  93871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93871, CurrentID=0, Status='live', Symbol='Brwd1', Aliases='5330419I02Rik, D530019K20Rik, G1-403-16, Wdr9, repro5', description='bromodomain and WD repeat domain containing 1', other_designations='WD repeat domain 9|WD repeat-containing protein 9', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=95992089, end_position_on_the_genomic_accession=96082428, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  432442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432442, CurrentID=0, Status='live', Symbol='Akap7', Aliases='6430401D08, AI662165, AKAP15, Akap18, BB170514', description='A kinase (PRKA) anchor protein 7', other_designations='A-kinase anchor protein 18|A-kinase anchor protein 9 kDa|AKAP 9|AKAP-18|AKAP-7 isoform alpha|AKAP-7 isoform gamma|PRKA7 isoform alpha|PRKA7 isoform gamma|akap7|protein kinase A anchoring protein 7|protein kinase A-anchoring protein 7 isoform alpha|protein kinase A-anchoring protein 7 isoform gamma', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=25169090, end_position_on_the_genomic_accession=25299574, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  432450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432450, CurrentID=0, Status='live', Symbol='Nkain2', Aliases='6330571D19Rik, AW455467, Tcba1', description='Na+/K+ transporting ATPase interacting 2', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 2|T-cell lymphoma breakpoint associated target 1|T-cell lymphoma breakpoint-associated target protein 1', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=31689310, end_position_on_the_genomic_accession=32893997, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  432467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432467, CurrentID=0, Status='live', Symbol='Hnrnph3', Aliases='AA693301, AI666703, Hnrph3', description='heterogeneous nuclear ribonucleoprotein H3', other_designations='hnRNP 2H9', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63014650, end_position_on_the_genomic_accession=63023889, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  432478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432478, CurrentID=0, Status='live', Symbol='Tmprss9', Aliases='Serase-1B', description='transmembrane protease, serine 9', other_designations='polyserase-1|polyserase-I|polyserine protease 1|polyserine protease-1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80871117, end_position_on_the_genomic_accession=80899495, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  432479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432479, CurrentID=0, Status='live', Symbol='4930404N11Rik', Aliases='2210409H10Rik, 5730551K18Rik, AI854616', description='RIKEN cDNA 4930404N11 gene', other_designations='Uncharacterized protein ENSP00000327950 homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81364024, end_position_on_the_genomic_accession=81365820, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  432486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432486, CurrentID=0, Status='live', Symbol='Gnptab', Aliases='EG432486, mKIAA1208', description='N-acetylglucosamine-1-phosphate transferase, alpha and beta subunits', other_designations='UDP-N-acetylglucosamine-1-phosphotransferase subunits alpha/beta|glcNAc-1-phosphotransferase subunits alpha/beta|stealth protein GNPTAB', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88379137, end_position_on_the_genomic_accession=88447330, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  432508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432508, CurrentID=0, Status='live', Symbol='Cpsf6', Aliases='4733401N12Rik, AI256641, CFIM, CFIM68, HPBRII-4, HPBRII-7', description='cleavage and polyadenylation specific factor 6', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117344667, end_position_on_the_genomic_accession=117376986, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  432516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432516, CurrentID=0, Status='live', Symbol='Myo1a', Aliases='BBM-I, Myhl', description='myosin IA', other_designations='BBMI|MIHC|brush border myosin 1|brush border myosin I|myosin I heavy chain|myosin, heavy polypeptide-like (110kD)|myosin-Ia', map_location='10 74.62 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127703440, end_position_on_the_genomic_accession=127720940, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  432530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432530, CurrentID=0, Status='live', Symbol='Adcy1', Aliases='AC1, D11Bwg1392e, I-AC, brl, mKIAA4070', description='adenylate cyclase 1', other_designations='ATP pyrophosphate-lyase 1|Ca(2+)/calmodulin-activated adenylyl cyclase|adenylate cyclase type I|adenylyl cyclase 1|barrelless', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=7063489, end_position_on_the_genomic_accession=7178506, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  432536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432536, CurrentID=0, Status='live', Symbol='Pom121l12', Aliases='1700054H16Rik', description='POM121 membrane glycoprotein-like 12', other_designations='', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=14599240, end_position_on_the_genomic_accession=14600316, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  432552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432552, CurrentID=0, Status='live', Symbol='Fam71b', Aliases='Gari-l3, OTTMUSG00000005491', description='family with sequence similarity 71, member B', other_designations='Golgi-associated Rab2B interactor-like 3', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46404728, end_position_on_the_genomic_accession=46407985, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  432572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432572, CurrentID=0, Status='live', Symbol='Specc1', Aliases='2810012G08Rik, B230396K10Rik, Cytsb', description='sperm antigen with calponin homology and coiled-coil domains 1', other_designations='spectrin domain with coiled-coils 1', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61956763, end_position_on_the_genomic_accession=62223013, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  432582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432582, CurrentID=0, Status='live', Symbol='Ccdc92b', Aliases='E130309D14Rik', description='coiled-coil domain containing 92B', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74619605, end_position_on_the_genomic_accession=74641516, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  170439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170439, CurrentID=0, Status='live', Symbol='Elovl6', Aliases='C77826, FAE, LCE', description='ELOVL family member 6, elongation of long chain fatty acids (yeast)', other_designations='3-keto acyl-CoA synthase Elovl6|ELOVL FA elongase 6|ELOVL fatty acid elongase 6|fatty acyl-CoA elongase|long chain fatty acid elongase|long chain fatty acyl elongase|myelin-associated SUR4 protein|very long chain 3-ketoacyl-CoA synthase 6|very long chain 3-oxoacyl-CoA synthase 6|very-long-chain 3-oxoacyl-CoA synthase 6', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129532386, end_position_on_the_genomic_accession=129638495, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  170441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170441, CurrentID=0, Status='live', Symbol='Slc2a10', Aliases='AA450473, Glut10', description='solute carrier family 2 (facilitated glucose transporter), member 10', other_designations='glucose transporter type 10', map_location='2 85.66 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165503897, end_position_on_the_genomic_accession=165519917, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  170442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170442, CurrentID=0, Status='live', Symbol='Bbox1', Aliases='AI874739', description='butyrobetaine (gamma), 2-oxoglutarate dioxygenase 1 (gamma-butyrobetaine hydroxylase)', other_designations='gamma-BBH|gamma-butyrobetaine hydroxylase|gamma-butyrobetaine,2-oxoglutarate dioxygenase', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=110264056, end_position_on_the_genomic_accession=110314814, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  170458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170458, CurrentID=0, Status='live', Symbol='Gpha2', Aliases='GPA2, zsig51', description='glycoprotein hormone alpha 2', other_designations='putative secreted protein Zsig51|thyrostimulin subunit alpha', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6226401, end_position_on_the_genomic_accession=6227768, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  170459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170459, CurrentID=0, Status='live', Symbol='Stard4', Aliases='4632419C16Rik, 9030213J02Rik, AA517649, AW324468', description='StAR-related lipid transfer (START) domain containing 4', other_designations='START domain-containing protein 4', map_location='18 18.18 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=33199140, end_position_on_the_genomic_accession=33213879, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  170460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170460, CurrentID=0, Status='live', Symbol='Stard5', Aliases='2310058G22Rik, AI607880, D7Ertd152e', description='StAR-related lipid transfer (START) domain containing 5', other_designations='18B7-T7(GS)|START domain-containing protein 5|StAR-related protein 1-4E|StAR-related protein p3-15E/p3-16E', map_location='7 47.43 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=83632017, end_position_on_the_genomic_accession=83642328, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  170461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170461, CurrentID=0, Status='live', Symbol='Stard6', Aliases='1700011K09Rik, 4833424I06Rik, 4933429L05Rik, AI644424', description='StAR-related lipid transfer (START) domain containing 6', other_designations='START domain containing protein 6|START domain-containing protein 6', map_location='18 44.44 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=70472454, end_position_on_the_genomic_accession=70505254, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  432611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432611, CurrentID=0, Status='live', Symbol='Dnaic2', Aliases='C030015H18Rik, Dnai2, Gm728', description='dynein, axonemal, intermediate chain 2', other_designations='axonemal dynein intermediate chain 2|dynein, axonemal, intermediate polypeptide 2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114722210, end_position_on_the_genomic_accession=114757889, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  432613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432613, CurrentID=0, Status='live', Symbol='Trim80', Aliases='4933422H20Rik, Trim47l', description='tripartite motif-containing 80', other_designations='novel zinc finger, C3HC4 type (RING finger) and SPRY domain containing protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115440545, end_position_on_the_genomic_accession=115448268, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  170472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170472, CurrentID=0, Status='live', Symbol='Recql5', Aliases='RecQ5, Recq5bb, Recql5', description='RecQ protein-like 5', other_designations='DNA helicase, RecQ-like type 5|RECQL5beta', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115892595, end_position_on_the_genomic_accession=115933531, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  170483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170483, CurrentID=0, Status='live', Symbol='Grin3b', Aliases='GluN3B, NR3B', description='glutamate receptor, ionotropic, NMDA3B', other_designations='N-methyl-D-aspartate receptor subunit NR3B|NMDA receptor 4|NMDAR3B|glutamate [NMDA] receptor subunit 3B|nr4', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79970397, end_position_on_the_genomic_accession=79977190, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  170484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170484, CurrentID=0, Status='live', Symbol='Nphs2', Aliases='PDCN, SRN1', description='nephrosis 2, podocin', other_designations='nephrosis 2 homolog, podocin', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156310544, end_position_on_the_genomic_accession=156328035, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  432677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432677, CurrentID=0, Status='live', Symbol='Vrtn', Aliases='7420416P09Rik', description='vertebrae development associated', other_designations='vertebrae development homolog (pig)', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84641019, end_position_on_the_genomic_accession=84651455, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  170571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170571, CurrentID=0, Status='live', Symbol='Cntnap4', Aliases='Caspr4, E130114F09Rik', description='contactin associated protein-like 4', other_designations='cell recognition molecule Caspr4|cell recognition protein CASPR4|contactin associated protein 4', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=112569965, end_position_on_the_genomic_accession=112882707, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  170574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170574, CurrentID=0, Status='live', Symbol='Sp7', Aliases='6430578P22Rik, C22, Osx', description='Sp7 transcription factor 7', other_designations='osterix|trans-acting transcription factor 7|zinc finger protein osterix', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102356606, end_position_on_the_genomic_accession=102367033, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  432720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432720, CurrentID=0, Status='live', Symbol='Akr1c19', Aliases='1810010N06Rik', description='aldo-keto reductase family 1, member C19', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4233504, end_position_on_the_genomic_accession=4248361, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  432729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432729, CurrentID=0, Status='live', Symbol='Tcrg-C', Aliases='A130095M15Rik', description='T cell receptor gamma, constant region', other_designations='T-cell receptor gamma, constant region|Tcell receptor gamma chain', map_location='13 6.89 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.5', start_position_on_the_genomic_accession=19431694, end_position_on_the_genomic_accession=19444634, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  432731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432731, CurrentID=0, Status='live', Symbol='Zscan26', Aliases='BC068174, Zfp187, Znf187', description='zinc finger and SCAN domain containing 26', other_designations='zinc finger protein 187', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21442175, end_position_on_the_genomic_accession=21453762, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  432763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432763, CurrentID=0, Status='live', Symbol='Prr7', Aliases='XM_484263', description='proline rich 7 (synaptic)', other_designations='synaptic proline-rich membrane protein', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55464267, end_position_on_the_genomic_accession=55473155, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  170625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170625, CurrentID=0, Status='live', Symbol='Snx18', Aliases='Snag1', description='sorting nexin 18', other_designations='sorting nexin associated golgi protein 1|sorting nexin-associated Golgi protein 1', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113592179, end_position_on_the_genomic_accession=113618564, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  432779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432779, CurrentID=0, Status='live', Symbol='Lrrc14b', Aliases='AI595366', description='leucine rich repeat containing 14B', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74359582, end_position_on_the_genomic_accession=74364000, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  170638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170638, CurrentID=0, Status='live', Symbol='Hpcal4', Aliases='AI846570, NVP-2', description='hippocalcin-like 4', other_designations='neural visinin-like protein 2', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123183504, end_position_on_the_genomic_accession=123194699, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  170639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170639, CurrentID=0, Status='live', Symbol='Olfr78', Aliases='4633402A21Rik, MOL2.3, MOR18-2, Or51e2, PSGR, RA1c', description='olfactory receptor 78', other_designations='GA_x6K02T2PBJ9-5459657-5458695|olfactory receptor MOR18-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102740721, end_position_on_the_genomic_accession=102759471, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  170643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170643, CurrentID=0, Status='live', Symbol='Kirrel', Aliases='6720469N11Rik1, Neph1, Kirrel', description='kin of IRRE like (Drosophila)', other_designations='kin of IRRE like 1|kin of irregular chiasm-like protein 1|nephrin 1|nephrin-like protein 1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87078592, end_position_on_the_genomic_accession=87174747, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  170644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170644, CurrentID=0, Status='live', Symbol='Ubn1', Aliases='1110029L11Rik, 2610108L02Rik, AA673476, AW124741', description='ubinuclein 1', other_designations='ubiquitously expressed nuclear protein', map_location='16 2.49 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5050068, end_position_on_the_genomic_accession=5086289, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  170646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170646, CurrentID=0, Status='live', Symbol='Noa', Aliases='', description='Naruto Otsuka atrichia', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170648, CurrentID=0, Status='live', Symbol='Olfr138', Aliases='MOR256-40P, Olfr89', description='olfactory receptor 138', other_designations='GA_x6K02T2PSCP-2718585-2719523|olfactory receptor 89', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38274773, end_position_on_the_genomic_accession=38275711, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  170650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170650, CurrentID=0, Status='live', Symbol='Trm1', Aliases='', description='tRNA methionine initiator 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170651, CurrentID=0, Status='live', Symbol='Krtap19-2', Aliases='Krtap16-1', description='keratin associated protein 19-2', other_designations='keratin associated protein 16-1|keratin-associated protein 16.1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88873665, end_position_on_the_genomic_accession=88874269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  170656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170656, CurrentID=0, Status='live', Symbol='Krtap21-1', Aliases='1110017B05Rik, AI506979, KAP21.1, Krtap16-7', description='keratin associated protein 21-1', other_designations='keratin associated protein 16-7', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=89403027, end_position_on_the_genomic_accession=89403774, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  170676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170676, CurrentID=0, Status='live', Symbol='Peg10', Aliases='AA407948, Edr, HB-1, MEF3L, Mar2, Mart2, MyEF-3', description='paternally expressed 10', other_designations='MEF3-like protein 1|MyEF-3 like|embryonal carcinoma differentiation regulated protein|mammalian retrotransposon-derived protein 2|myelin expression factor 3|myelin expression factor 3-like protein 1|paternally expressed gene 10 protein|retrotransposon gag domain-containing protein 3|retrotransposon-derived gag-like polyprotein|ty3/Gypsy-like protein', map_location='6 1.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=4747306, end_position_on_the_genomic_accession=4760517, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  170677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170677, CurrentID=0, Status='live', Symbol='Cdhr1', Aliases='Pcdh21, Prcad, mKIAA1775', description='cadherin-related family member 1', other_designations='photoreceptor cadherin|protocadherin 21|protocadherin-21', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=37077849, end_position_on_the_genomic_accession=37098311, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  432839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432839, CurrentID=0, Status='live', Symbol='Gprin2', Aliases='C130040D06Rik, C230073P13, Gm286, mKIAA0514', description='G protein regulated inducer of neurite outgrowth 2', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34185688, end_position_on_the_genomic_accession=34201692, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  170706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170706, CurrentID=0, Status='live', Symbol='Tmem37', Aliases='AI173373, AW540203, Pr, Pr1', description='transmembrane protein 37', other_designations='neuronal voltage-gated calcium channel gamma-like subunit|protein distantly related to to the gamma subunit family|voltage-dependent calcium channel gamma subunit-like protein', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120067377, end_position_on_the_genomic_accession=120074069, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  170707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170707, CurrentID=0, Status='live', Symbol='Usp48', Aliases='2810449C13Rik, AI115503, BC021769, D330022K21Rik, Usp31', description='ubiquitin specific peptidase 48', other_designations='deubiquitinating enzyme 48|ubiquitin specific protease 31|ubiquitin specific protease 48|ubiquitin thioesterase 48|ubiquitin thiolesterase 48|ubiquitin-specific-processing protease 48', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137593757, end_position_on_the_genomic_accession=137658537, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  432855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432855, CurrentID=0, Status='live', Symbol='Zfhx2os', Aliases='Zfhx2as', description='zinc finger homeobox 2, opposite strand', other_designations='Zfh-5 antisense|zinc finger homeobox 2, antisense', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55073091, end_position_on_the_genomic_accession=55075874, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  170716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170716, CurrentID=0, Status='live', Symbol='Cyp4f13', Aliases='0610030I10Rik, Cypf13', description='cytochrome P450, family 4, subfamily f, polypeptide 13', other_designations='P450 CYP4F13|cytochrome P450 CYP4F13|cytochrome P450,subfamily IVF, polypeptide 13|leukotriene B4 omega hydroxylase', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32924688, end_position_on_the_genomic_accession=32947415, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  170718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170718, CurrentID=0, Status='live', Symbol='Idh3b', Aliases='C78231', description='isocitrate dehydrogenase 3 (NAD+) beta', other_designations='N14A tumor-related protein|isocitrate dehydrogenase 3 beta', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130279309, end_position_on_the_genomic_accession=130284542, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  170719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170719, CurrentID=0, Status='live', Symbol='Oxr1', Aliases='2210416C20Rik, C7, C7B', description='oxidation resistance 1', other_designations='nucleolar protein C7|protein C7', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=41447482, end_position_on_the_genomic_accession=41861048, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  170720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170720, CurrentID=0, Status='live', Symbol='Card14', Aliases='Bimp2, CARMA2', description='caspase recruitment domain family, member 14', other_designations='bcl10-interacting MAGUK protein 2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119307096, end_position_on_the_genomic_accession=119345375, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  170721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170721, CurrentID=0, Status='live', Symbol='Papln', Aliases='E030033C16Rik', description='papilin, proteoglycan-like sulfated glycoprotein', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83762992, end_position_on_the_genomic_accession=83794875, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  170722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170722, CurrentID=0, Status='live', Symbol='Nxf7', Aliases='Nx7, Nxf6, nfx-a1, nxf-a, nxf-a2', description='nuclear RNA export factor 7', other_designations='nuclear RNA export factor 6|nuclear export factor a', map_location='X 57.25 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135579555, end_position_on_the_genomic_accession=135598777, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  432867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432867, CurrentID=0, Status='live', Symbol='Defb48', Aliases='Defb42, EG432867', description='defensin beta 48', other_designations='beta-defensin 42', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62977524, end_position_on_the_genomic_accession=62984510, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  170725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170725, CurrentID=0, Status='live', Symbol='Capn8', Aliases="nCL-2, nCL-2'", description='calpain 8', other_designations='new calpain 2|stomach-specific M-type calpain|stomach-specific calpain (nCL-2)', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182564748, end_position_on_the_genomic_accession=182634456, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  170728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170728, CurrentID=0, Status='live', Symbol='Rtn4ip1', Aliases='D10Ertd690e, NIMP', description='reticulon 4 interacting protein 1', other_designations='NOGO-interacting mitochondrial protein|reticulon 4 interacting protein 1, mitochondrial', map_location='10 23.12 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43901807, end_position_on_the_genomic_accession=43947862, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  170729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170729, CurrentID=0, Status='live', Symbol='Scrt1', Aliases='', description='scratch homolog 1, zinc finger protein (Drosophila)', other_designations='SCRT|mScrt|scratch 1|scratch homolog 1 zinc finger protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76516203, end_position_on_the_genomic_accession=76522129, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  170731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170731, CurrentID=0, Status='live', Symbol='Mfn2', Aliases='D630023P19Rik, Fzo', description='mitofusin 2', other_designations='HSG protein|hypertension related protein 1|hypertension-related protein 1|mitochondrial assembly regulatory factor|transmembrane GTPase MFN2', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147873586, end_position_on_the_genomic_accession=147904909, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  170732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170732, CurrentID=0, Status='live', Symbol='Trhr2', Aliases='TRH-R2', description='thyrotropin releasing hormone receptor 2', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122356065, end_position_on_the_genomic_accession=122362910, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  170733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170733, CurrentID=0, Status='live', Symbol='Klra17', Aliases='Ly49q, Ly49q1', description='killer cell lectin-like receptor, subfamily A, member 17', other_designations='NK receptor Ly-49Q', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129831154, end_position_on_the_genomic_accession=129876672, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  170734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170734, CurrentID=0, Status='live', Symbol='Zscan5b', Aliases='Zfg1, Zfp371', description='zinc finger and SCAN domain containing 5B', other_designations='zinc finger gonad protein 1|zinc finger protein 371', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6222278, end_position_on_the_genomic_accession=6239417, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  170735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170735, CurrentID=0, Status='live', Symbol='Arr3', Aliases='Arr4, Car, Carfl, Carr', description='arrestin 3, retinal', other_designations='X-arrestin|arrestin 4|retinal cone arrestin-3', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100605480, end_position_on_the_genomic_accession=100618953, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  170736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170736, CurrentID=0, Status='live', Symbol='Parvb', Aliases='AI595373, AW742462, D15Gsk1, affixin', description='parvin, beta', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84232043, end_position_on_the_genomic_accession=84315609, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  170737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170737, CurrentID=0, Status='live', Symbol='Znrf1', Aliases='B830022L21Rik, Rnf42, Zrfp1, nin283', description='zinc and ring finger 1', other_designations='nerve injury-induced gene 283 protein|peripheral nerve injury protein nin283|ring finger protein 42|zinc ring finger protein 1|zinc/RING finger protein 1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111536640, end_position_on_the_genomic_accession=111626030, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  170738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170738, CurrentID=0, Status='live', Symbol='Kcnh7', Aliases='9330137I11Rik, Kv11.3, erg3', description='potassium voltage-gated channel, subfamily H (eag-related), member 7', other_designations='ERG-3|eag-related protein 3|ether-a-go-go-related gene potassium channel 3|ether-a-go-go-related protein 3|potassium channel protein erg3|voltage-gated potassium channel subunit Kv11.3', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=62699587, end_position_on_the_genomic_accession=63184287, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  170740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170740, CurrentID=0, Status='live', Symbol='Zfp287', Aliases='B230333C16Rik, SKAT-2, Skat2, Znf287, mszf16, mszf74', description='zinc finger protein 287', other_designations='zfp-287|zinc finger protein SKAT-2|zinc finger protein SKAT2', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62692444, end_position_on_the_genomic_accession=62731900, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  170741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170741, CurrentID=0, Status='live', Symbol='Pilrb1', Aliases='FDFACT, Fdact, Pilrb', description='paired immunoglobin-like type 2 receptor beta 1', other_designations='activating receptor PILR-beta|cell surface receptor FDFACT|immunoglobulin-like cell surface receptor FDFACT, activating counterpart', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137852141, end_position_on_the_genomic_accession=137858527, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  170742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170742, CurrentID=0, Status='live', Symbol='Sertad3', Aliases='AW124805, D1Mgi51, Rbt1, Sei3', description='SERTA domain containing 3', other_designations='RPA-binding trans-activator|replication protein binding trans-activator|replication protein-binding trans-activator', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27473840, end_position_on_the_genomic_accession=27477364, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  170743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170743, CurrentID=0, Status='live', Symbol='Tlr7', Aliases='', description='toll-like receptor 7', other_designations='Toll like receptor 7', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=167304926, end_position_on_the_genomic_accession=167330574, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  170744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170744, CurrentID=0, Status='live', Symbol='Tlr8', Aliases='', description='toll-like receptor 8', other_designations='', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=167241123, end_position_on_the_genomic_accession=167264329, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  170745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170745, CurrentID=0, Status='live', Symbol='Xpnpep2', Aliases='9030008G12Rik, mAPP', description='X-prolyl aminopeptidase (aminopeptidase P) 2, membrane-bound', other_designations='membrane-bound aminopeptidase P', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48106913, end_position_on_the_genomic_accession=48136982, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  170748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170748, CurrentID=0, Status='live', Symbol='Smco4', Aliases='Fn5', description='single-pass membrane protein with coiled-coil domains 4', other_designations='UPF0443 protein C11orf75 homolog', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15468440, end_position_on_the_genomic_accession=15545260, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  170749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170749, CurrentID=0, Status='live', Symbol='Mtmr4', Aliases='AA596759, ESTM44, FYVE-DSP2, ZFYVE11, mKIAA0647', description='myotubularin related protein 4', other_designations='FYVE zinc finger phosphatase', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87592135, end_position_on_the_genomic_accession=87616302, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  170750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170750, CurrentID=0, Status='live', Symbol='Xpnpep1', Aliases='D230045I08Rik, sAMP', description='X-prolyl aminopeptidase (aminopeptidase P) 1, soluble', other_designations='X-Pro aminopeptidase 1|aminoacylproline aminopeptidase|cytosolic aminopeptidase P|xpnpep1', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=52932227, end_position_on_the_genomic_accession=53038654, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  170752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170752, CurrentID=0, Status='live', Symbol='Bco2', Aliases='B-diox-II, Bcdo2, Bcmo2, CMO2', description='beta-carotene oxygenase 2', other_designations="b,b-carotene-9',10'-dioxygenase|beta-carotene 9', 10'-dioxygenase 2|beta-carotene 9',10'-dioxygenase 2|beta-carotene dioxygenase 2|beta-diox-II", map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50533087, end_position_on_the_genomic_accession=50555220, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  170753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170753, CurrentID=0, Status='live', Symbol='Zfp704', Aliases='C030026M03, C030034L19Rik, Gig1, Znf704', description='zinc finger protein 704', other_designations='glucocorticoid induced gene 1|glucocorticoid-induced gene 1 protein', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=9427010, end_position_on_the_genomic_accession=9610085, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  170755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170755, CurrentID=0, Status='live', Symbol='Sgk3', Aliases='2510015P22Rik, A330005P07Rik, Cisk, fy, fz', description='serum/glucocorticoid regulated kinase 3', other_designations='cytokine-independent survival kinase|serum/glucocorticoid-regulated kinase 3', map_location='1 2.08 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9798130, end_position_on_the_genomic_accession=9902570, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  170756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170756, CurrentID=0, Status='live', Symbol='Slc8b1', Aliases='AF261233, NCKX6, NCLX, Slc24a6', description='solute carrier family 8 (sodium/lithium/calcium exchanger), member B1', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 6|sodium/calcium exchanger protein, mitochondrial|solute carrier family 24 (sodium/potassium/calcium exchanger), member 6|solute carrier family 24 member 6|solute carrier family 8 member B1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120511168, end_position_on_the_genomic_accession=120534024, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  170757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170757, CurrentID=0, Status='live', Symbol='Adgrl4', Aliases='1110033N21Rik, ETL1, Eltd1, Etl', description='adhesion G protein-coupled receptor L4', other_designations='EGF, latrophilin seven transmembrane domain containing 1|EGF, latrophilin seven transmembrane domain-containing protein 1|EGF-TM7 receptor', map_location='3 H3-H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=151437882, end_position_on_the_genomic_accession=151545081, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  170758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170758, CurrentID=0, Status='live', Symbol='Rac3', Aliases='Rac1B', description='RAS-related C3 botulinum substrate 3', other_designations='p21-Rac3', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120721468, end_position_on_the_genomic_accession=120723969, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  170759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170759, CurrentID=0, Status='live', Symbol='Atp13a1', Aliases='Atp13a, Cgi152, catp', description='ATPase type 13A1', other_designations='cation-transporting ATPase 2|probable cation-transporting ATPase 13A1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69791163, end_position_on_the_genomic_accession=69807749, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  170760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170760, CurrentID=0, Status='live', Symbol='Acbd3', Aliases='60kDa, 8430407O11Rik, D1Ertd10e, GCP60, GOLPH1, Gocap1, Pap7', description='acyl-Coenzyme A binding domain containing 3', other_designations='PBR associated protein|PBR- and PKA-associated protein 7|acyl-CoA-binding domain-containing protein 3|golgi complex associated protein 1, 60 kDa|golgi complex-associated protein 1|golgi phosphoprotein 1|peripheral benzodiazepine receptor-associated protein PAP7|peripherial benzodiazepine receptor associated protein', map_location='1 84.46 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180726043, end_position_on_the_genomic_accession=180754204, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  170761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170761, CurrentID=0, Status='live', Symbol='Pdzd3', Aliases='NaPi-Cap2, Pdzk2', description='PDZ domain containing 3', other_designations='NHERF-4|PDZ domain containing 2|PDZ domain-containing protein 2|PDZ domain-containing protein 3|na/Pi cotransporter C-terminal-associated protein 2|natrium-phosphate cotransporter IIa C-terminal-associated protein 2|sodium-hydrogen exchanger regulatory factor 4|sodium-phosphate cotransporter IIa C-terminal-associated protein 2', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44247312, end_position_on_the_genomic_accession=44251464, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  170762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170762, CurrentID=0, Status='live', Symbol='Nup155', Aliases='D930027M19Rik, mKIAA0791', description='nucleoporin 155', other_designations='155 kDa nucleoporin|nucleoporin Nup155', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=8109156, end_position_on_the_genomic_accession=8159859, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  170763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170763, CurrentID=0, Status='live', Symbol='Zfp87', Aliases='2210039O17Rik, 4631412O18Rik, KRAB4, Mzf22, Zfp72', description='zinc finger protein 87', other_designations='KRAB zinc finger protein (Mzf22)|zinc finger protein 72', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67515781, end_position_on_the_genomic_accession=67526231, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  170765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170765, CurrentID=0, Status='live', Symbol='Ripply3', Aliases='Dscr6', description='ripply3 homolog (zebrafish)', other_designations='Down syndrome critical region homolog 6|down syndrome critical region protein 6|homolog of human Down syndrome critical region gene 6 DSCR6', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94328422, end_position_on_the_genomic_accession=94336935, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  170767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170767, CurrentID=0, Status='live', Symbol='Rfxap', Aliases='5730495K23Rik', description='regulatory factor X-associated protein', other_designations='RFX-associated protein', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54803115, end_position_on_the_genomic_accession=54807791, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  170768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170768, CurrentID=0, Status='live', Symbol='Pfkfb3', Aliases='E330010H22Rik, iPFK-2, uPFK-2', description='6-phosphofructo-2-kinase/fructose-2,6-biphosphatase 3', other_designations='inducible 6-phosphofructo-2-kinase', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=11471430, end_position_on_the_genomic_accession=11554071, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  170770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170770, CurrentID=0, Status='live', Symbol='Bbc3', Aliases='PUMA, PUMA/JFY1', description='BCL2 binding component 3', other_designations='BH3-only|Bcl-2 binding component 3|p53 up-regulated modulator of apoptosis', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16308321, end_position_on_the_genomic_accession=16318334, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  170771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170771, CurrentID=0, Status='live', Symbol='Khdrbs2', Aliases='6330586C16Rik, Slim1, Slm1', description='KH domain containing, RNA binding, signal transduction associated 2', other_designations='Sam68-like mammalian protein 1|mSLM-1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=32172688, end_position_on_the_genomic_accession=32705224, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  170772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170772, CurrentID=0, Status='live', Symbol='Glcci1', Aliases='2310047L21Rik, A130036A18Rik, Fam117c, Gig18, Tssn1', description='glucocorticoid induced transcript 1', other_designations='glucocorticoid-induced gene 18 protein|testhymin|thymocyte/spermatocyte selection protein 1', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=8509600, end_position_on_the_genomic_accession=8597549, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  170776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170776, CurrentID=0, Status='live', Symbol='Cd209c', Aliases='SIGNR2, mSIGNR2', description='CD209c antigen', other_designations='DC-SIGN-related protein 2|DC-SIGNR2', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3940221, end_position_on_the_genomic_accession=3955947, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  170779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170779, CurrentID=0, Status='live', Symbol='Cd209d', Aliases='SIGN-R3, SIGNR3, mSIGNR3', description='CD209d antigen', other_designations='DC-SIGN-related protein 3|DC-SIGNR3', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3871824, end_position_on_the_genomic_accession=3878548, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  170780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170780, CurrentID=0, Status='live', Symbol='Cd209e', Aliases='SIGNR4, mSIGNR4', description='CD209e antigen', other_designations='DC-SIGN-related protein 4|DC-SIGNR4', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3847973, end_position_on_the_genomic_accession=3854286, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  170786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170786, CurrentID=0, Status='live', Symbol='Cd209a', Aliases='CD209, CDSIGN, CIRE, DC-SIGN, DC-SIGN1, Dcsign, SIGN-R1, SIGNR5', description='CD209a antigen', other_designations='dendritic cell-specific ICAM-3-grabbing non-integrin', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3743395, end_position_on_the_genomic_accession=3748984, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  170787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170787, CurrentID=0, Status='live', Symbol='Hdac10', Aliases='AW548891, Hd10', description='histone deacetylase 10', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89123303, end_position_on_the_genomic_accession=89128700, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  170788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170788, CurrentID=0, Status='live', Symbol='Crb1', Aliases='7530426H14Rik, A930008G09Rik', description='crumbs family member 1, photoreceptor morphogenesis associated', other_designations='crumbs homolog 1', map_location='1 60.87 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=139197546, end_position_on_the_genomic_accession=139379118, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  170789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170789, CurrentID=0, Status='live', Symbol='Acot8', Aliases='PTE-2, Pte1', description='acyl-CoA thioesterase 8', other_designations='PTE-1|choloyl-coenzyme A thioesterase|peroxisomal acyl-CoA thioesterase 1|peroxisomal acyl-CoA thioesterase 2|peroxisomal acyl-coenzyme A thioester hydrolase 1|peroxisomal long-chain acyl-CoA thioesterase 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164792768, end_position_on_the_genomic_accession=164820208, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  170790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170790, CurrentID=0, Status='live', Symbol='Mlc1', Aliases='AW048630, BB074274, Kiaa0027-hp, LVM, MLC, VL, WKL1, mKIAA0027', description='megalencephalic leukoencephalopathy with subcortical cysts 1 homolog (human)', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88955884, end_position_on_the_genomic_accession=88982687, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  170791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170791, CurrentID=0, Status='live', Symbol='Rbm39', Aliases='1500012C14Rik, 2310040E03Rik, B330012G18Rik, C79248, R75070, Rnpc2, caper', description='RNA binding motif protein 39', other_designations='Caper alpha|RNA-binding region (RNP1, RRM) containing 2|RNA-binding region-containing protein 2|coactivator of AP-1 and ERs|coactivator of activating protein-1 and estrogen receptors|transcription coactivator CAPER', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156147239, end_position_on_the_genomic_accession=156180238, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  432940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432940, CurrentID=0, Status='live', Symbol='Otulin', Aliases='C79097, Fam105b, m3Sapc, m7-1Sapc', description='OTU deubiquitinase with linear linkage specificity', other_designations='OTU domain-containing deubiquitinase with linear linkage specificity|deubiquitinating enzyme otulin|family with sequence similarity 105, member B|gumby|protein FAM105B|ubiquitin thioesterase Gumby', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=27605917, end_position_on_the_genomic_accession=27630980, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  170799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170799, CurrentID=0, Status='live', Symbol='Rtkn2', Aliases='B130039D23Rik, Mbf, Plekhk1', description='rhotekin 2', other_designations='PH domain-containing family K member 1|membrane bound factor|pleckstrin homology domain containing, family K member 1|pleckstrin homology domain-containing family K member 1', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=67979566, end_position_on_the_genomic_accession=68043864, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  170800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170800, CurrentID=0, Status='live', Symbol='Cdcs1', Aliases='', description='cytokine deficiency colitis susceptibility 1', other_designations='', map_location='3 56.07 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170801, CurrentID=0, Status='live', Symbol='Idd5.1', Aliases='Idd5a', description='insulin dependent diabetes susceptibility 5.1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170802, CurrentID=0, Status='live', Symbol='Idd5.2', Aliases='Idd5b', description='insulin dependent diabetes susceptibility 5.2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170812, CurrentID=0, Status='live', Symbol='Ahsp', Aliases='EDRF, Eraf', description='alpha hemoglobin stabilizing protein', other_designations='erythroid associated factor|erythroid differentiation-related factor|erythroid-associated factor', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170813, CurrentID=0, Status='live', Symbol='Ms4a3', Aliases='HTm4', description='membrane-spanning 4-domains, subfamily A, member 3', other_designations='haematopoietic cell-specific transmembrane-4|hematopoietic-specific transmembrane protein 4', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11629499, end_position_on_the_genomic_accession=11640851, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  432964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432964, CurrentID=0, Status='live', Symbol='K230010J24Rik', Aliases='', description='RIKEN cDNA K230010J24 gene', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76009858, end_position_on_the_genomic_accession=76046981, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  170822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170822, CurrentID=0, Status='live', Symbol='Usp33', Aliases='9830169D19Rik, AA409780, Vdu1', description='ubiquitin specific peptidase 33', other_designations='VHL-interacting deubiquitinating enzyme 1|Vhlh-interacting deubiquitinating enzyme 1|deubiquitinating enzyme 33|pVHL-interacting deubiquitinating enzyme 1|ubiquitin specific protease 33|ubiquitin thioesterase 33|ubiquitin thiolesterase 33|ubiquitin-specific-processing protease 33', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152346478, end_position_on_the_genomic_accession=152393617, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  170823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170823, CurrentID=0, Status='live', Symbol='Glmn', Aliases='9330160J16Rik, AW227515, Fap48, Fap68', description='glomulin, FKBP associated protein', other_designations='FAP|FK506-binding protein-associated protein|FKBP-associated protein', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107548967, end_position_on_the_genomic_accession=107597888, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  170826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170826, CurrentID=0, Status='live', Symbol='Ppargc1b', Aliases='4631412G21Rik, Perc', description='peroxisome proliferative activated receptor, gamma, coactivator 1 beta', other_designations='ERR ligand 1|PGC-1-beta|PGC-1beta|PGC-1beta/ERRL1|PPAR gamma coactivator-1beta protein|PPAR-gamma coactivator 1-beta|PPARGC-1-beta|peroxisome proliferator-activated receptor gamma coactivator 1 beta', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61291445, end_position_on_the_genomic_accession=61401462, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  170828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170828, CurrentID=0, Status='live', Symbol='Vgll1', Aliases='Tondu', description='vestigial like 1 homolog (Drosophila)', other_designations='vestigial-related factor|vgl-1', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=57088079, end_position_on_the_genomic_accession=57106539, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  170829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170829, CurrentID=0, Status='live', Symbol='Tram2', Aliases='C330003D03Rik', description='translocating chain-associating membrane protein 2', other_designations='', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=21001397, end_position_on_the_genomic_accession=21079225, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  170833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170833, CurrentID=0, Status='live', Symbol='Hook2', Aliases='A630054I03Rik', description='hook microtubule tethering protein 2', other_designations='mHK2', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84989501, end_position_on_the_genomic_accession=85003364, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  170834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170834, CurrentID=0, Status='live', Symbol='Oosp1', Aliases='AA536827', description='oocyte secreted protein 1', other_designations='initiate factor 3|oocyte secretory protein 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11667460, end_position_on_the_genomic_accession=11691051, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  170835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170835, CurrentID=0, Status='live', Symbol='Inpp5j', Aliases='Pib5pa, Pipp', description='inositol polyphosphate 5-phosphatase J', other_designations='phosphatidylinositol (4,5) bisphosphate 5-phosphatase, A|proline-rich inositol polyphosphate 5-phosphatase', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3494272, end_position_on_the_genomic_accession=3505567, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  433016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433016, CurrentID=0, Status='live', Symbol='Gm5483', Aliases='EG433016', description='predicted gene 5483', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36184212, end_position_on_the_genomic_accession=36188111, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  433022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433022, CurrentID=0, Status='live', Symbol='Plcxd2', Aliases='AW494639, EG433022', description='phosphatidylinositol-specific phospholipase C, X domain containing 2', other_designations='', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45959261, end_position_on_the_genomic_accession=46010413, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  433031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433031, CurrentID=0, Status='live', Symbol='Olfr194', Aliases='MOR182-13, MOR182-7P', description='olfactory receptor 194', other_designations='GA_x54KRFPKG5P-55348161-55347241|olfactory receptor MOR182-7P', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59119148, end_position_on_the_genomic_accession=59120068, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  433053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433053, CurrentID=0, Status='live', Symbol='LOC433053', Aliases='', description='Ig lambda1 chain c region', other_designations='immunoglobulin lambda chain C1 region', map_location='16', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  170930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170930, CurrentID=0, Status='live', Symbol='Sumo2', Aliases='SUMO-2, Smt3A, Smt3b, Smt3h2', description='small ubiquitin-like modifier 2', other_designations='SMT3 homolog 2|SMT3 suppressor of mif two 3 homolog 2|SMT3 supressor of mif two 3 homolog 2|sentrin-2|ubiquitin-like protein SMT3A|ubiquitin-like protein SMT3B', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115523109, end_position_on_the_genomic_accession=115536230, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  170935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170935, CurrentID=0, Status='live', Symbol='Grid2ip', Aliases='', description='glutamate receptor, ionotropic, delta 2 (Grid2) interacting protein 1', other_designations='GluR-delta2 philic-protein|L-delphilin|glutamate receptor, ionotropic, delta 2-interacting protein 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143357338, end_position_on_the_genomic_accession=143391798, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  170936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170936, CurrentID=0, Status='live', Symbol='Zfp369', Aliases='B930030B22Rik, D230020H11Rik, NRIF2', description='zinc finger protein 369', other_designations='neurotrophin receptor interacting factor 2', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=65276668, end_position_on_the_genomic_accession=65297804, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  170938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170938, CurrentID=0, Status='live', Symbol='Zfp617', Aliases='Zfps11-6', description='zinc finger protein 617', other_designations='Zinc finger protein s11-6', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71922825, end_position_on_the_genomic_accession=71934630, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  170939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170939, CurrentID=0, Status='live', Symbol='Krtap19-9b', Aliases='Krtap16-10b, Krtap16.10L', description='keratin associated protein 19-9B', other_designations='high-glycine/tyrosine protein type I E5|keratin associated protein 16-10B|keratin-associated protein 16-10b', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88931804, end_position_on_the_genomic_accession=88932264, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  170942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170942, CurrentID=0, Status='live', Symbol='Erdr1', Aliases='edr', description='erythroid differentiation regulator 1', other_designations='', map_location='Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=90785442, end_position_on_the_genomic_accession=90816465, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  170947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170947, CurrentID=0, Status='live', Symbol='Myoz3', Aliases='4833419K08Rik, Fatz-3', description='myozenin 3', other_designations='FATZ-related protein 3|calsarcin-3|myoz3', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60573717, end_position_on_the_genomic_accession=60591716, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  170952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=170952, CurrentID=0, Status='live', Symbol='Prima1', Aliases='B230212M13Rik, BB120497, PRiMA', description='proline rich membrane anchor 1', other_designations='acetylcholinesterase membrane anchor', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103194453, end_position_on_the_genomic_accession=103242720, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  433102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433102, CurrentID=0, Status='live', Symbol='Sfta2', Aliases='EG433102', description='surfactant associated 2', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35649708, end_position_on_the_genomic_accession=35650569, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72185, CurrentID=0, Status='live', Symbol='Dbndd1', Aliases='2810427I04Rik, D8Ertd590e', description='dysbindin (dystrobrevin binding protein 1) domain containing 1', other_designations='', map_location='8 72.1 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123504718, end_position_on_the_genomic_accession=123518041, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  433178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433178, CurrentID=0, Status='live', Symbol='Spink14', Aliases='EG433178, Gm5505', description='serine peptidase inhibitor, Kazal type 14', other_designations='Kazal type serine protease inhibitor 5-like 2', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44027869, end_position_on_the_genomic_accession=44032208, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  433180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433180, CurrentID=0, Status='live', Symbol='Spink6', Aliases='EG433180', description='serine peptidase inhibitor, Kazal type 6', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44071393, end_position_on_the_genomic_accession=44083610, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  433181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433181, CurrentID=0, Status='live', Symbol='Spink11', Aliases='9230112K01Rik', description='serine peptidase inhibitor, Kazal type 11', other_designations='serine protease inhibitor Kazal type 11', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44190045, end_position_on_the_genomic_accession=44196177, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  433182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433182, CurrentID=0, Status='live', Symbol='Eno1b', Aliases='EG433182, Eno1, Gm5506', description='enolase 1B, retrotransposed', other_designations='enolase 1, alpha non-neuron', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=48045335, end_position_on_the_genomic_accession=48048379, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  94043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94043, CurrentID=0, Status='live', Symbol='Tm2d1', Aliases='2310026L18Rik, AA990549, Bbp', description='TM2 domain containing 1', other_designations='beta-amyloid binding protein|beta-amyloid-binding protein|mBBP', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=98355370, end_position_on_the_genomic_accession=98383281, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  433215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433215, CurrentID=0, Status='live', Symbol='BC048609', Aliases='', description='cDNA sequence BC048609', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6077867, end_position_on_the_genomic_accession=6080788, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  171095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171095, CurrentID=0, Status='live', Symbol='Il17rc', Aliases='1110025H02Rik, Gm19850, IL-17RC, IL17-RC, IL17-RL, Il17rl', description='interleukin 17 receptor C', other_designations='IL-17 receptor C|IL-17RL|interleukin-17 receptor-like protein|zcytoR14', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113471455, end_position_on_the_genomic_accession=113483163, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  433247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433247, CurrentID=0, Status='live', Symbol='Cyp2c68', Aliases='9030012A22Rik', description='cytochrome P450, family 2, subfamily c, polypeptide 68', other_designations='cytochrome P450 2C40 family', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39688836, end_position_on_the_genomic_accession=39741101, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  433256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433256, CurrentID=0, Status='live', Symbol='Acsl5', Aliases='1700030F05Rik, ACS2, ACS5, Facl5', description='acyl-CoA synthetase long-chain family member 5', other_designations='LACS 5|fatty acid Coenzyme A ligase, long chain 5|long-chain acyl-CoA synthetase 5', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=55251862, end_position_on_the_genomic_accession=55296628, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  433278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433278, CurrentID=0, Status='live', Symbol='Khdc1c', Aliases='Khdc1b', description='KH domain containing 1C', other_designations='KH homology domain containing 1C|oocyte-specific Khdc1b', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=21368331, end_position_on_the_genomic_accession=21369743, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  433292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433292, CurrentID=0, Status='live', Symbol='Nms', Aliases='AB164466', description='neuromedin S', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=38939149, end_position_on_the_genomic_accession=38950276, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  433294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433294, CurrentID=0, Status='live', Symbol='Mettl21c', Aliases='A530098C11Rik', description='methyltransferase like 21C', other_designations='methyltransferase-like protein 21C', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44009399, end_position_on_the_genomic_accession=44042626, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  171166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171166, CurrentID=0, Status='live', Symbol='Mcoln3', Aliases='6720490O21Rik, TRPML3, Va', description='mucolipin 3', other_designations='varitint-waddler', map_location='3 71.03 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146093277, end_position_on_the_genomic_accession=146141808, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  171167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171167, CurrentID=0, Status='live', Symbol='Fut10', Aliases='fuc-TX, fut10A', description='fucosyltransferase 10', other_designations='fucosyltransferase X|galactoside 3-L-fucosyltransferase 10', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=31187331, end_position_on_the_genomic_accession=31261924, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  171168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171168, CurrentID=0, Status='live', Symbol='Acer1', Aliases='2310024P18Rik, AI662009, Alkcdase1, Asah3, Cer1', description='alkaline ceramidase 1', other_designations='N-acylsphingosine amidohydrolase 3|acylsphingosine deacylase 3|alkCDase 1|alkaline cdase 1|maCER1', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56953490, end_position_on_the_genomic_accession=56982126, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  171170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171170, CurrentID=0, Status='live', Symbol='Mbnl3', Aliases='A530038J18Rik, AI661274, CHCR, E430034C16Rik, MBLX, MBLX39, MBXL', description='muscleblind-like 3 (Drosophila)', other_designations='cys3His CCG1-required protein|muscleblind-like X-linked protein', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=51113494, end_position_on_the_genomic_accession=51206035, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  171171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171171, CurrentID=0, Status='live', Symbol='Ntng2', Aliases='2610016D08Rik, Lmnt2', description='netrin G2', other_designations='laminet 2', map_location='2 19.57 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29194821, end_position_on_the_genomic_accession=29248040, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  433323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433323, CurrentID=0, Status='live', Symbol='Sgpp2', Aliases='SPPase2, Spp2', description='sphingosine-1-phosphate phosphotase 2', other_designations='sphingosine-1-phosphatase 2', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=78309501, end_position_on_the_genomic_accession=78420289, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  171180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171180, CurrentID=0, Status='live', Symbol='Syt12', Aliases='', description='synaptotagmin XII', other_designations='synaptotagmin 12|sytXII', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4445908, end_position_on_the_genomic_accession=4477143, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  171183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171183, CurrentID=0, Status='live', Symbol='Vmn1r37', Aliases='V1rc10', description='vomeronasal 1 receptor 37', other_designations='vomernasal 1 receptor Vmn1r37|vomeronasal 1 receptor, C10|vomeronasal receptor V1RC10', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66731392, end_position_on_the_genomic_accession=66732300, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171184, CurrentID=0, Status='live', Symbol='Vmn1r36', Aliases='V1rc11', description='vomeronasal 1 receptor 36', other_designations='vomernasal 1 receptor Vmn1r36|vomeronasal 1 receptor, C11', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66715972, end_position_on_the_genomic_accession=66716889, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171185, CurrentID=0, Status='live', Symbol='Vmn1r35', Aliases='V1rc12', description='vomeronasal 1 receptor 35', other_designations='vomernasal 1 receptor Vmn1r35|vomeronasal 1 receptor, C12', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66678794, end_position_on_the_genomic_accession=66679684, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171186, CurrentID=0, Status='live', Symbol='Vmn1r38', Aliases='V1rc13', description='vomeronasal 1 receptor 38', other_designations='vomernasal 1 receptor Vmn1r38|vomeronasal 1 receptor, C13', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66776222, end_position_on_the_genomic_accession=66777130, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171187, CurrentID=0, Status='live', Symbol='Vmn1r33', Aliases='V1rc14', description='vomeronasal 1 receptor 33', other_designations='vomernasal 1 receptor Vmn1r33|vomeronasal 1 receptor, C14', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66611651, end_position_on_the_genomic_accession=66612568, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171188, CurrentID=0, Status='live', Symbol='Vmn1r32', Aliases='V1rc15', description='vomeronasal 1 receptor 32', other_designations='vomernasal 1 receptor Vmn1r32|vomeronasal 1 receptor, C15', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66552183, end_position_on_the_genomic_accession=66559708, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  171189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171189, CurrentID=0, Status='live', Symbol='Vmn1r17', Aliases='V1rc16', description='vomeronasal 1 receptor 17', other_designations='vomernasal 1 receptor Vmn1r17|vomeronasal 1 receptor, C16', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57360467, end_position_on_the_genomic_accession=57361378, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171191, CurrentID=0, Status='live', Symbol='Vmn1r24', Aliases='V1rc18', description='vomeronasal 1 receptor 24', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57955641, end_position_on_the_genomic_accession=57956531, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171192, CurrentID=0, Status='live', Symbol='Vmn1r5', Aliases='V1rc19', description='vomeronasal 1 receptor 5', other_designations='vomernasal 1 receptor Vmn1r5|vomeronasal 1 receptor 6 pseudogene', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56985342, end_position_on_the_genomic_accession=56986289, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171193, CurrentID=0, Status='live', Symbol='Vmn1r6', Aliases='V1rc20', description='vomeronasal 1 receptor 6', other_designations='vomernasal 1 receptor Vmn1r6|vomeronasal 1 receptor, C20', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57002355, end_position_on_the_genomic_accession=57003266, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171194, CurrentID=0, Status='live', Symbol='Vmn1r4', Aliases='C230065D10Rik, V1rc21', description='vomeronasal 1 receptor 4', other_designations='vomernasal 1 receptor Vmn1r4|vomeronasal 1 receptor, C21|vomeronasal type 1 receptor C21', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56924058, end_position_on_the_genomic_accession=56957608, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  171195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171195, CurrentID=0, Status='live', Symbol='Vmn1r30', Aliases='V1rc22, V1rc9', description='vomeronasal 1 receptor 30', other_designations='vomernasal 1 receptor Vmn1r30|vomeronasal 1 receptor, C22|vomeronasal 1 receptor, C9', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58434848, end_position_on_the_genomic_accession=58521728, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  171196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171196, CurrentID=0, Status='live', Symbol='Vmn1r22', Aliases='V1rc23', description='vomeronasal 1 receptor 22', other_designations='vomernasal 1 receptor Vmn1r22|vomeronasal 1 receptor, C23', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57900082, end_position_on_the_genomic_accession=57900990, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171197, CurrentID=0, Status='live', Symbol='Vmn1r23', Aliases='V1rc24', description='vomeronasal 1 receptor 23', other_designations='vomernasal 1 receptor Vmn1r23|vomeronasal 1 receptor, C24', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57925883, end_position_on_the_genomic_accession=57926791, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171198, CurrentID=0, Status='live', Symbol='Vmn1r28', Aliases='V1rc25', description='vomeronasal 1 receptor 28', other_designations='vomernasal 1 receptor Vmn1r28|vomeronasal 1 receptor, C25', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58262294, end_position_on_the_genomic_accession=58266082, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  171199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171199, CurrentID=0, Status='live', Symbol='Vmn1r18', Aliases='V1rc26', description='vomeronasal 1 receptor 18', other_designations='vomernasal 1 receptor Vmn1r18|vomeronasal 1 receptor, C26', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57389668, end_position_on_the_genomic_accession=57390567, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171200, CurrentID=0, Status='live', Symbol='Vmn1r19', Aliases='V1rc27', description='vomeronasal 1 receptor 19', other_designations='vomernasal 1 receptor Vmn1r19|vomeronasal 1 receptor, C27|vomeronasal type 1 receptor C27', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57400820, end_position_on_the_genomic_accession=57405477, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  171201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171201, CurrentID=0, Status='live', Symbol='Vmn1r21', Aliases='V1rc28', description='vomeronasal 1 receptor 21', other_designations='vomernasal 1 receptor Vmn1r21|vomeronasal 1 receptor, C28|vomeronasal type 1 receptor C28', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57843564, end_position_on_the_genomic_accession=57844457, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171202, CurrentID=0, Status='live', Symbol='Vmn1r16', Aliases='V1rc29', description='vomeronasal 1 receptor 16', other_designations='vomernasal 1 receptor Vmn1r16', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57322724, end_position_on_the_genomic_accession=57323635, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171203, CurrentID=0, Status='live', Symbol='Vmn1r9', Aliases='V1rc30', description='vomeronasal 1 receptor 9', other_designations='vomernasal 1 receptor Vmn1r9|vomeronasal type 1 receptor C30', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57070895, end_position_on_the_genomic_accession=57071945, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171205, CurrentID=0, Status='live', Symbol='Vmn1r8', Aliases='V1rc32', description='vomeronasal 1 receptor 8', other_designations='vomernasal 1 receptor Vmn1r8|vomeronasal 1 receptor, C30', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57032341, end_position_on_the_genomic_accession=57037125, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  171206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171206, CurrentID=0, Status='live', Symbol='Vmn1r27', Aliases='V1rc33', description='vomeronasal 1 receptor 27', other_designations='vomernasal 1 receptor Vmn1r27|vomeronasal 1 receptor, C33', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58215106, end_position_on_the_genomic_accession=58216017, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171207, CurrentID=0, Status='live', Symbol='Arhgap4', Aliases='A130009C12Rik, A530015A20Rik, Rgc1, c1, mKIAA0131', description='Rho GTPase activating protein 4', other_designations='GTPase-activating protein p115', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73894352, end_position_on_the_genomic_accession=73911346, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  171209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171209, CurrentID=0, Status='live', Symbol='Asic3', Aliases='AW742291, Accn3, DRASIC, SLNAC1, TNAC1', description='acid-sensing (proton-gated) ion channel 3', other_designations='amiloride-sensitive cation channel 3|dorsal root ASIC', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24410289, end_position_on_the_genomic_accession=24417834, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  171210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171210, CurrentID=0, Status='live', Symbol='Acot2', Aliases='AA571646, MTE-I, Mte1', description='acyl-CoA thioesterase 2', other_designations='acyl coenzyme A thioester hydrolase|mitochondrial acyl-CoA thioesterase 1|very-long-chain acyl-CoA thioesterase', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83987861, end_position_on_the_genomic_accession=83993877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  171211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171211, CurrentID=0, Status='live', Symbol='Edaradd', Aliases='1810032E07Rik, 5830469M23Rik, A630043P06, cr', description='EDAR (ectodysplasin-A receptor)-associated death domain', other_designations='EDAR-associated death domain protein|crinkled', map_location='13 4.77 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12471209, end_position_on_the_genomic_accession=12520704, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  171212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171212, CurrentID=0, Status='live', Symbol='Galnt10', Aliases='AU018154, C330012K04Rik, GalNAc-T10, GalNAc-T9, Galnt9, ppGaNTase', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 10', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 10|UDP-N-acetyl-alpha-D-galactosamine: polypeptide N-acetylgalactosaminyltransferase 9|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 9|polypeptide GalNAc transferase 10|pp-GaNTase 10|protein-UDP acetylgalactosaminyltransferase 10', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=57622407, end_position_on_the_genomic_accession=57787501, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  433365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433365, CurrentID=0, Status='live', Symbol='Teddm1b', Aliases='EG433365, Gm5531', description='transmembrane epididymal protein 1B', other_designations='', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153874345, end_position_on_the_genomic_accession=153876871, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171224, CurrentID=0, Status='live', Symbol='Vmn1r229', Aliases='V1re1', description='vomeronasal 1 receptor 229', other_designations='vomernasal 1 receptor Vmn1r229|vomeronasal 1 receptor, E1|vomeronasal type 1 receptor E1', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20814495, end_position_on_the_genomic_accession=20815415, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171225, CurrentID=0, Status='live', Symbol='Vmn1r226', Aliases='V1re2', description='vomeronasal 1 receptor 226', other_designations='vomernasal 1 receptor Vmn1r226|vomeronasal 1 receptor, E2|vomeronasal type 1 receptor E2', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20687508, end_position_on_the_genomic_accession=20688404, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171226, CurrentID=0, Status='live', Symbol='Vmn1r228', Aliases='V1re3, Vldr1', description='vomeronasal 1 receptor 228', other_designations='vomernasal 1 receptor Vmn1r228|vomeronasal 1 receptor, E3|vomeronasal receptor 1 D1|vomeronasal type 1 receptor E3', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20776059, end_position_on_the_genomic_accession=20777501, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171227, CurrentID=0, Status='live', Symbol='Vmn1r232', Aliases='V1re4', description='vomeronasal 1 receptor 232', other_designations='vomernasal 1 receptor Vmn1r232|vomeronasal type 1 receptor E4', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20913205, end_position_on_the_genomic_accession=20914363, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  433375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433375, CurrentID=0, Status='live', Symbol='Creg1', Aliases='AA755314, Creg', description='cellular repressor of E1A-stimulated genes 1', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165763732, end_position_on_the_genomic_accession=165775304, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  171233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171233, CurrentID=0, Status='live', Symbol='Vmn1r235', Aliases='V1rf2', description='vomeronasal 1 receptor 235', other_designations='vomernasal 1 receptor Vmn1r235|vomeronasal type 1 receptor F2', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21260427, end_position_on_the_genomic_accession=21262863, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  171234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171234, CurrentID=0, Status='live', Symbol='Vmn1r237', Aliases='V1rf3, Vfrf3', description='vomeronasal 1 receptor 237', other_designations='vomernasal 1 receptor Vmn1r237|vomeronasal 1 receptor, F3|vomeronasal type 1 receptor F3', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21314017, end_position_on_the_genomic_accession=21314886, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171236, CurrentID=0, Status='live', Symbol='Vmn1r233', Aliases='V1rf5', description='vomeronasal 1 receptor 233', other_designations='vomernasal 1 receptor Vmn1r233|vomeronasal 1 receptor, F5|vomeronasal receptor V1RF5|vomeronasal type 1 receptor F5', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20993727, end_position_on_the_genomic_accession=20994686, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171237, CurrentID=0, Status='live', Symbol='Vmn1r73', Aliases='V1rg2', description='vomeronasal 1 receptor 73', other_designations='vomernasal 1 receptor Vmn1r73|vomeronasal 1 receptor, G2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11756257, end_position_on_the_genomic_accession=11757168, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171240, CurrentID=0, Status='live', Symbol='Vmn1r74', Aliases='V1rg5', description='vomeronasal 1 receptor 74', other_designations='vomernasal 1 receptor Vmn1r74|vomeronasal 1 receptor, G5', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11846775, end_position_on_the_genomic_accession=11847689, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171241, CurrentID=0, Status='live', Symbol='Vmn1r75', Aliases='V1rg6', description='vomeronasal 1 receptor 75', other_designations='vomernasal 1 receptor Vmn1r75|vomeronasal 1 receptor, G6|vomeronasal type 1 receptor G6', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11880343, end_position_on_the_genomic_accession=11881260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  433386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433386, CurrentID=0, Status='live', Symbol='Ccdc185', Aliases='4922505E12Rik', description='coiled-coil domain containing 185', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182747126, end_position_on_the_genomic_accession=182749180, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171243, CurrentID=0, Status='live', Symbol='Vmn1r83', Aliases='V1rg8', description='vomeronasal 1 receptor 83', other_designations='vomernasal 1 receptor Vmn1r83|vomeronasal 1 receptor, G8', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12321208, end_position_on_the_genomic_accession=12322128, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171244, CurrentID=0, Status='live', Symbol='Vmn1r81', Aliases='V1rg9', description='vomeronasal 1 receptor 81', other_designations='vomernasal 1 receptor Vmn1r81|vomeronasal 1 receptor, G9|vomeronasal type 1 receptor G9', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12259759, end_position_on_the_genomic_accession=12260679, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171245, CurrentID=0, Status='live', Symbol='Vmn1r-ps103', Aliases='V1rh1', description='vomeronasal 1 receptor, pseudogene 103', other_designations='vomernasal 1 receptor Vmn1r-ps103', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22441272, end_position_on_the_genomic_accession=22442243, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171246, CurrentID=0, Status='live', Symbol='Vmn1r200', Aliases='V1rh3', description='vomeronasal 1 receptor 200', other_designations='vomernasal 1 receptor Vmn1r200|vomeronasal 1 receptor, H3', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22395029, end_position_on_the_genomic_accession=22395967, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171247, CurrentID=0, Status='live', Symbol='Vmn1r199', Aliases='V1rh4', description='vomeronasal 1 receptor 199', other_designations='vomernasal 1 receptor Vmn1r199|vomeronasal 1 receptor, H4|vomeronasal type 1 receptor H4', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22382538, end_position_on_the_genomic_accession=22383641, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171248, CurrentID=0, Status='live', Symbol='Vmn1r214', Aliases='V1rh5', description='vomeronasal 1 receptor 214', other_designations='vomernasal 1 receptor Vmn1r214|vomeronasal 1 receptor, H5', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23030577, end_position_on_the_genomic_accession=23035441, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  171249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171249, CurrentID=0, Status='live', Symbol='Vmn1r213', Aliases='V1rh6', description='vomeronasal 1 receptor 213', other_designations='vomernasal 1 receptor Vmn1r213|vomeronasal 1 receptor, H6', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23011249, end_position_on_the_genomic_accession=23012403, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171250, CurrentID=0, Status='live', Symbol='Vmn1r206', Aliases='V1rh7', description='vomeronasal 1 receptor 206', other_designations='vomernasal 1 receptor Vmn1r206|vomeronasal 1 receptor, H7', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22619433, end_position_on_the_genomic_accession=22632462, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  171251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171251, CurrentID=0, Status='live', Symbol='Vmn1r205', Aliases='V1rh8', description='vomeronasal 1 receptor 205', other_designations='vomernasal 1 receptor Vmn1r205|vomeronasal 1 receptor, H8', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22591980, end_position_on_the_genomic_accession=22592930, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171253, CurrentID=0, Status='live', Symbol='Vmn1r215', Aliases='V1ri2', description='vomeronasal 1 receptor 215', other_designations='vomernasal 1 receptor Vmn1r215|vomeronasal 1 receptor, I2', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23075792, end_position_on_the_genomic_accession=23076694, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171254, CurrentID=0, Status='live', Symbol='Vmn1r198', Aliases='V1ri3', description='vomeronasal 1 receptor 198', other_designations='vomernasal 1 receptor Vmn1r198|vomeronasal 1 receptor, I3', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22350555, end_position_on_the_genomic_accession=22355248, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  171255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171255, CurrentID=0, Status='live', Symbol='Vmn1r201', Aliases='V1ri4', description='vomeronasal 1 receptor 201', other_designations='vomernasal 1 receptor Vmn1r201|vomeronasal 1 receptor, I4', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22474618, end_position_on_the_genomic_accession=22475520, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171256, CurrentID=0, Status='live', Symbol='Vmn1r218', Aliases='V1ri5', description='vomeronasal 1 receptor 218', other_designations='vomernasal 1 receptor Vmn1r218|vomeronasal 1 receptor, I5', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23136485, end_position_on_the_genomic_accession=23137381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171257, CurrentID=0, Status='live', Symbol='Vmn1r195', Aliases='V1ri6', description='vomeronasal 1 receptor 195', other_designations='vomernasal 1 receptor Vmn1r195|vomeronasal type 1 receptor I6', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22278317, end_position_on_the_genomic_accession=22279404, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171258, CurrentID=0, Status='live', Symbol='Vmn1r202', Aliases='V1ri7', description='vomeronasal 1 receptor 202', other_designations='vomernasal 1 receptor Vmn1r202|vomeronasal 1 receptor, I7|vomeronasal type 1 receptor I7', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22501337, end_position_on_the_genomic_accession=22502245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171259, CurrentID=0, Status='live', Symbol='Vmn1r193', Aliases='V1ri9', description='vomeronasal 1 receptor 193', other_designations='vomernasal 1 receptor Vmn1r193|vomeronasal 1 receptor, I9', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22218861, end_position_on_the_genomic_accession=22219820, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171260, CurrentID=0, Status='live', Symbol='Vmn1r89', Aliases='V1RE8, V1rj2', description='vomeronasal 1 receptor 89', other_designations='vomernasal 1 receptor Vmn1r89|vomeronasal 1 receptor, J2|vomeronasal receptor 1 E8|vomeronasal receptor V1RJ1|vomeronasal type 1 receptor J2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13219339, end_position_on_the_genomic_accession=13220289, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171261, CurrentID=0, Status='live', Symbol='Vmn1r87', Aliases='V1rk1', description='vomeronasal 1 receptor 87', other_designations='vomernasal 1 receptor Vmn1r87|vomeronasal 1 receptor, K1|vomeronasal type 1 receptor K1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13131471, end_position_on_the_genomic_accession=13140087, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  171262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171262, CurrentID=0, Status='live', Symbol='Vmn1r70', Aliases='V1rd2, V1rl1', description='vomeronasal 1 receptor 70', other_designations='vomernasal 1 receptor Vmn1r70|vomeronasal 1 receptor, L1|vomeronasal receptor 1 D2|vomeronasal type 1 receptor L1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=10633587, end_position_on_the_genomic_accession=10634483, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171263, CurrentID=0, Status='live', Symbol='Vmn1r67', Aliases='V1re10', description='vomeronasal 1 receptor 67', other_designations='vomernasal 1 receptor Vmn1r67', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=10446789, end_position_on_the_genomic_accession=10447787, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171265, CurrentID=0, Status='live', Symbol='Vmn1r185', Aliases='V1re12', description='vomeronasal 1 receptor 185', other_designations='vomernasal 1 receptor Vmn1r185|vomeronasal 1 receptor, E12', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26611119, end_position_on_the_genomic_accession=26612078, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171267, CurrentID=0, Status='live', Symbol='Vmn1r84', Aliases='V1rg11', description='vomeronasal 1 receptor 84', other_designations='vomernasal 1 receptor Vmn1r84|vomeronasal 1 receptor, G11|vomeronasal type 1 receptor G11', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12361808, end_position_on_the_genomic_accession=12362764, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171269, CurrentID=0, Status='live', Symbol='Vmn1r210', Aliases='V1rh10', description='vomeronasal 1 receptor 210', other_designations='vomernasal 1 receptor Vmn1r210|vomeronasal 1 receptor, H10', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22827194, end_position_on_the_genomic_accession=22828114, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171270, CurrentID=0, Status='live', Symbol='Vmn1r203', Aliases='V1rh11', description='vomeronasal 1 receptor 203', other_designations='vomernasal 1 receptor Vmn1r203|vomeronasal 1 receptor, H11', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22524051, end_position_on_the_genomic_accession=22524986, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171271, CurrentID=0, Status='live', Symbol='Vmn1r220', Aliases='V1rh12', description='vomeronasal 1 receptor 220', other_designations='vomernasal 1 receptor Vmn1r220|vomeronasal 1 receptor, H12', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23183628, end_position_on_the_genomic_accession=23187867, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  433416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433416, CurrentID=0, Status='live', Symbol='Gm13547', Aliases='OTTMUSG00000012686', description='predicted gene 13547', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29761528, end_position_on_the_genomic_accession=29764529, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  171273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171273, CurrentID=0, Status='live', Symbol='Vmn1r217', Aliases='V1rh14', description='vomeronasal 1 receptor 217', other_designations='vomernasal 1 receptor Vmn1r217|vomeronasal 1 receptor, H14', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23113834, end_position_on_the_genomic_accession=23114730, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171274, CurrentID=0, Status='live', Symbol='Vmn1r222', Aliases='V1rh16', description='vomeronasal 1 receptor 222', other_designations='vomernasal 1 receptor Vmn1r222|vomeronasal 1 receptor, H16', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23232115, end_position_on_the_genomic_accession=23233041, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171275, CurrentID=0, Status='live', Symbol='Vmn1r212', Aliases='V1rh18', description='vomeronasal 1 receptor 212', other_designations='vomernasal 1 receptor Vmn1r212|vomeronasal 1 receptor, H18', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22883088, end_position_on_the_genomic_accession=22884161, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  171277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171277, CurrentID=0, Status='live', Symbol='Vmn1r211', Aliases='V1rh20', description='vomeronasal 1 receptor 211', other_designations='vomernasal 1 receptor Vmn1r211|vomeronasal 1 receptor, H20|vomeronasal type 1 receptor H20', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22851599, end_position_on_the_genomic_accession=22854501, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  171278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171278, CurrentID=0, Status='live', Symbol='Vmn1r197', Aliases='V1rh21', description='vomeronasal 1 receptor 197', other_designations='vomernasal 1 receptor Vmn1r197|vomeronasal 1 receptor, H21', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22324611, end_position_on_the_genomic_accession=22328807, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  171279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171279, CurrentID=0, Status='live', Symbol='Vmn1r216', Aliases='V1ri10', description='vomeronasal 1 receptor 216', other_designations='vomernasal 1 receptor Vmn1r216|vomeronasal 1 receptor, I10', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23099149, end_position_on_the_genomic_accession=23100045, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  433424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433424, CurrentID=0, Status='live', Symbol='Zeb2os', Aliases='Gm13476, OTTMUSG00000012358, Zeb2nat', description='zinc finger E-box binding homeobox 2, opposite strand', other_designations='Zeb2 natural antisense transcript', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=45111138, end_position_on_the_genomic_accession=45114087, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  171281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171281, CurrentID=0, Status='live', Symbol='Acot3', Aliases='AW108394, PTE-Ia, Pte2a', description='acyl-CoA thioesterase 3', other_designations='peroxisomal acyl-CoA thioesterase 2A|peroxisomal acyl-coenzyme A thioester hydrolase 2a|peroxisomal long-chain acyl-CoA thioesterase 2', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84051997, end_position_on_the_genomic_accession=84060400, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  171282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171282, CurrentID=0, Status='live', Symbol='Acot4', Aliases='B430212I04Rik, PTE-Ib, Pte1b, Pte2b', description='acyl-CoA thioesterase 4', other_designations='PTE-2b|peroxisomal acyl coenzyme A thioester hydrolase Ib|peroxisomal acyl-CoA thioesterase 2B|peroxisomal long-chain acyl-CoA thioesterase Ib', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84038379, end_position_on_the_genomic_accession=84044723, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  171283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171283, CurrentID=0, Status='live', Symbol='Havcr1', Aliases='AI503787, KIM-1, TIM-1, Tim1, Timd1', description='hepatitis A virus cellular receptor 1', other_designations='HAVcr-1|T-cell immunoglobulin and mucin domain containing 1|T-cell immunoglobulin mucin receptor 1|TIMD-1|kidney injury molecule 1|t cell immunoglobulin and mucin domain-containing protein 1|t cell membrane protein 1', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46739823, end_position_on_the_genomic_accession=46779578, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  171284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171284, CurrentID=0, Status='live', Symbol='Timd2', Aliases='AA816106, AU019358, AU019662, C78111, G9D3, TIM-2, Tim2', description='T cell immunoglobulin and mucin domain containing 2', other_designations='T-cell immunoglobulin and mucin domain containing 2|T-cell immunoglobulin mucin receptor 2|T-cell membrane protein 2|TIMD-2', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46668960, end_position_on_the_genomic_accession=46707104, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  171285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171285, CurrentID=0, Status='live', Symbol='Havcr2', Aliases='TIM-3, Tim3, Timd3', description='hepatitis A virus cellular receptor 2', other_designations='HAVcr-2|T-cell immunoglobulin and mucin domain containing 3|T-cell immunoglobulin and mucin domain-containing protein 3|T-cell immunoglobulin mucin receptor 3|T-cell membrane protein 3|TIMD-3', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46454547, end_position_on_the_genomic_accession=46481255, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  171286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171286, CurrentID=0, Status='live', Symbol='Slc12a8', Aliases='CCC9, E330020C02Rik', description='solute carrier family 12 (potassium/chloride transporters), member 8', other_designations='cation-chloride cotransporter 9', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33517333, end_position_on_the_genomic_accession=33664135, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  433449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433449, CurrentID=0, Status='live', Symbol='Olfr1197', Aliases='MOR225-10P, MOR225-14', description='olfactory receptor 1197', other_designations='GA_x6K02T2Q125-50202854-50201910', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88728653, end_position_on_the_genomic_accession=88729597, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  225160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225160, CurrentID=0, Status='live', Symbol='Thoc1', Aliases='3110002N20Rik, AW107452, NMP-84', description='THO complex 1', other_designations='nuclear matrix protein p84|tho1', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=9958180, end_position_on_the_genomic_accession=9995484, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  433470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433470, CurrentID=0, Status='live', Symbol='AA467197', Aliases='Nmes1, mir-147', description='expressed sequence AA467197', other_designations='normal mucosa of esophagus specific 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122636971, end_position_on_the_genomic_accession=122641087, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  225164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225164, CurrentID=0, Status='live', Symbol='Mib1', Aliases='DIP-1, E430019M12Rik, Mib, mKIAA1323', description='mindbomb E3 ubiquitin protein ligase 1', other_designations='DAPK-interacting protein 1|DAPK-interacting protein-1|mind bomb homolog 1|mind bomb-1|mindbomb homolog 1|skeletrophin', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10725625, end_position_on_the_genomic_accession=10818578, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  433485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433485, CurrentID=0, Status='live', Symbol='Syndig1', Aliases='AW123113, DSPC2, Gm14134, H2C20orf39, OTTMUSG00000015750, Tmem90b', description='synapse differentiation inducing 1', other_designations='dispanin subfamily C member 2|transmembrane protein 90B', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=149830053, end_position_on_the_genomic_accession=150004392, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  433490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433490, CurrentID=0, Status='live', Symbol='Defb45', Aliases='Defb27, OTTMUSG00000015859', description='defensin beta 45', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152593191, end_position_on_the_genomic_accession=152596485, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  433492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433492, CurrentID=0, Status='live', Symbol='Bpifb9b', Aliases='5430413K10Rik, AI413838, OTTMUSG00000015915', description='BPI fold containing family B, member 9B', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154307244, end_position_on_the_genomic_accession=154320644, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  433502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433502, CurrentID=0, Status='live', Symbol='Wfdc6b', Aliases='WAP6b, Wfdc6', description='WAP four-disulfide core domain 6B', other_designations='WAP four-disulfide core 6-like 1|putative protease inhibitor WAP6B', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164613578, end_position_on_the_genomic_accession=164618213, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  171382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171382, CurrentID=0, Status='live', Symbol='Trpm8', Aliases='CMR1, LTRPC6, LTrpC-6, TRPP8, Trp-p8', description='transient receptor potential cation channel, subfamily M, member 8', other_designations='TRP ion channel TRPM8|long transient receptor potential channel 6|transient receptor potential p8|transient receptor potential-p8', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88306712, end_position_on_the_genomic_accession=88388853, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  171388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171388, CurrentID=0, Status='live', Symbol='Bnipl', Aliases='1700128A13Rik, BNIP-S, BNIPL1, BNIPL2, PP73, PP753', description='BCL2/adenovirus E1B 19kD interacting protein like', other_designations='BNIP-2 similar', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95241293, end_position_on_the_genomic_accession=95251193, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  171395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171395, CurrentID=0, Status='live', Symbol='Pkd1l1', Aliases='AA444596', description='polycystic kidney disease 1 like 1', other_designations='PC1-like 1 protein|polycystin-1L1|protein rikishi', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=8813839, end_position_on_the_genomic_accession=8974367, orientation='minus', exon_count=51, OMIM='', no_hdr0=''),
  171405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171405, CurrentID=0, Status='live', Symbol='Slc22a27', Aliases='Oat9, mOAT6-L, mOAT6-S', description='solute carrier family 22, member 27', other_designations='mOAT6 related protein|organic anion transporter 6', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7864388, end_position_on_the_genomic_accession=7966595, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  171429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171429, CurrentID=0, Status='live', Symbol='Slc26a6', Aliases='B930010B04Rik, CFEX, Pat-1, Pat1', description='solute carrier family 26, member 6', other_designations='anion exchange transporter|chloride-formate exchanger|pendrin-L1|pendrin-like protein 1|putative anion transporter-1', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108912410, end_position_on_the_genomic_accession=108913049, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  433586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433586, CurrentID=0, Status='live', Symbol='Maml3', Aliases='AV234550, BC049812, Mam-2, mKIAA1816', description='mastermind like 3 (Drosophila)', other_designations='mastermind-2|mastermind-like 3', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51687320, end_position_on_the_genomic_accession=52105085, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  171463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171463, CurrentID=0, Status='live', Symbol='Il17rd', Aliases='2810004A10Rik, AI428510, Sef, Sef-S', description='interleukin 17 receptor D', other_designations='IL-17 receptor D|IL-17RD|interleukin-17 receptor-like protein|mSef|sef homolog', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=27038801, end_position_on_the_genomic_accession=27107286, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  171469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171469, CurrentID=0, Status='live', Symbol='Gpr37l1', Aliases='AW047233, CAG-18, D0Kist8', description='G protein-coupled receptor 37-like 1', other_designations='ETBR-LP-2|G protein-coupled receptor 37-lke 1|G-protein coupled receptor 37-like 1|endothelin B receptor-like protein 2|endothelin B receptor-like protein-2|orphan Gpr37-like protein 1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135160250, end_position_on_the_genomic_accession=135167681, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  433619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433619, CurrentID=0, Status='live', Symbol='Kprp', Aliases='1110001M24Rik, AA589586', description='keratinocyte expressed, proline-rich', other_designations='kprp', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92823074, end_position_on_the_genomic_accession=92827247, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  171486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171486, CurrentID=0, Status='live', Symbol='Cd99l2', Aliases='AW548191, Mic2l1, Xap89', description='CD99 antigen-like 2', other_designations='MIC2 (monoclonal Imperial Cancer Research Fund 2)-like 1|MIC2 like-1|MIC2-like protein 1', map_location='X A6-B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71420060, end_position_on_the_genomic_accession=71492869, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  433638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433638, CurrentID=0, Status='live', Symbol='I830077J02Rik', Aliases='', description='RIKEN cDNA I830077J02 gene', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105925851, end_position_on_the_genomic_accession=105932931, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  171504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171504, CurrentID=0, Status='live', Symbol='Apobr', Aliases='Apob-48r, Apob48r', description='apolipoprotein B receptor', other_designations='apolipoprotein B-100 receptor|apolipoprotein B-48 receptor|apolipoprotein B48 receptor', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126584912, end_position_on_the_genomic_accession=126589125, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  171506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171506, CurrentID=0, Status='live', Symbol='H1foo', Aliases='C86609, H1.8, H1fo, H1oo', description='H1 histone family, member O, oocyte-specific', other_designations='oocyte-specific histone H1|oocyte-specific linker histone H1', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115944294, end_position_on_the_genomic_accession=115950242, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  171508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171508, CurrentID=0, Status='live', Symbol='Creld1', Aliases='AI843811', description='cysteine-rich with EGF-like domains 1', other_designations='i11E7 protein', map_location='6 52.77 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113483297, end_position_on_the_genomic_accession=113493338, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  433667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433667, CurrentID=0, Status='live', Symbol='Ankrd13c', Aliases='AI505652, AU022220', description='ankyrin repeat domain 13c', other_designations='', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=157946668, end_position_on_the_genomic_accession=158006837, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  171530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171530, CurrentID=0, Status='live', Symbol='Ucn2', Aliases='Ucn II, Ucn-2, ucn-II', description='urocortin 2', other_designations='Urocortin II', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108986163, end_position_on_the_genomic_accession=108987164, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  171531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171531, CurrentID=0, Status='live', Symbol='Mlph', Aliases='2210418F23Rik, 5031433I09Rik, AW228792, D1Wsu84e, Slac-2a, l(1)-3Rk, l1Rk3, ln', description='melanophilin', other_designations='exophilin-3|leaden protein|slaC2-a|slp homolog lacking C2 domains a|synaptotagmin-like protein 2a', map_location='1 45.73 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90915022, end_position_on_the_genomic_accession=90951142, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  171543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171543, CurrentID=0, Status='live', Symbol='Bmf', Aliases='AW260063, AW411872', description='BCL2 modifying factor', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118528757, end_position_on_the_genomic_accession=118549678, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  433693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433693, CurrentID=0, Status='live', Symbol='Akirin2', Aliases='2700059D21Rik, AA114675, AA522011, AU019887, Akirin-2', description='akirin 2', other_designations='akirin2', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34550615, end_position_on_the_genomic_accession=34566911, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  433700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433700, CurrentID=0, Status='live', Symbol='Spag8', Aliases='MH-SPAG8', description='sperm associated antigen 8', other_designations='sperm associated antigen 8-like protein', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43651330, end_position_on_the_genomic_accession=43654124, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  433702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433702, CurrentID=0, Status='live', Symbol='Ncbp1', Aliases='AU014645, AW538051, CBP80', description='nuclear cap binding protein subunit 1', other_designations='80 kDa nuclear cap-binding protein|NCBP 80 kDa subunit', map_location='4 24.49 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46138511, end_position_on_the_genomic_accession=46172403, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  171567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171567, CurrentID=0, Status='live', Symbol='Nme7', Aliases='D530024H21Rik, Nm23-M7, Nm23-r7', description='NME/NM23 family member 7', other_designations='NDK 7|NDP kinase 7|non-metastatic cells 7, protein expressed in (nucleoside-diphosphate kinase)|nucleoside-diphosphate kinase', map_location='1 H2.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164307662, end_position_on_the_genomic_accession=164437300, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  171580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171580, CurrentID=0, Status='live', Symbol='Mical1', Aliases='MICAL, MICAL-1, Nical', description='microtubule associated monooxygenase, calponin and LIM domain containing 1', other_designations='CasL interacting molecule MICAL|NEDD9 interacting protein with calponin homology and LIM domains|NEDD9-interacting protein with calponin homology and LIM domains|mMical1|microtubule associated monoxygenase, calponin and LIM domain containing 1|molecule interacting with CasL protein 1', map_location='10 B1-B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41475860, end_position_on_the_genomic_accession=41487030, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  433752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433752, CurrentID=0, Status='live', Symbol='AA415398', Aliases='', description='expressed sequence AA415398', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119530308, end_position_on_the_genomic_accession=119538769, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  433759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433759, CurrentID=0, Status='live', Symbol='Hdac1', Aliases='HD1-ps, MommeD5, RPD3, Hdac1', description='histone deacetylase 1', other_designations='', map_location='4 63.26 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129516104, end_position_on_the_genomic_accession=129542646, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  433766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433766, CurrentID=0, Status='live', Symbol='Trim63', Aliases='MuRF1, RF1, Rnf28', description='tripartite motif-containing 63', other_designations='MURF-1|muscle ring finger protein 1|muscle-specific RING finger protein 1|ring finger protein 28|tripartite motif-containing protein 63', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134315088, end_position_on_the_genomic_accession=134329629, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  433771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433771, CurrentID=0, Status='live', Symbol='Minos1', Aliases='2310028O11Rik', description='mitochondrial inner membrane organizing system 1', other_designations='UPF0327 protein C1orf151 homolog|mitochondrial inner membrane organizing system protein 1|novel DUF543 domain containing protein', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139101814, end_position_on_the_genomic_accession=139131113, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  433801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433801, CurrentID=0, Status='live', Symbol='Gm13212', Aliases='OTTMUSG00000010965', description='predicted gene 13212', other_designations='novel KRAB box containing protein', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=145585172, end_position_on_the_genomic_accession=145625345, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  433804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433804, CurrentID=0, Status='live', Symbol='Gm13154', Aliases='OTTMUSG00000010673', description='predicted gene 13154', other_designations='novel protein similar to Rex2', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147553219, end_position_on_the_genomic_accession=147585201, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  433809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433809, CurrentID=0, Status='live', Symbol='Rnf207', Aliases='D330010C22Rik, Gm143', description='ring finger protein 207', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152306134, end_position_on_the_genomic_accession=152319369, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  433813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433813, CurrentID=0, Status='live', Symbol='Pusl1', Aliases='2810021I11Rik, AI853843, D18346', description='pseudouridylate synthase-like 1', other_designations='tRNA pseudouridylate synthase-like 1|tRNA-uridine isomerase-like 1', map_location='4 87.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155888857, end_position_on_the_genomic_accession=155891774, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100048534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100048534, CurrentID=0, Status='live', Symbol='Cfap43', Aliases='4632415N18Rik, 4930428C11Rik, 4930463G05Rik, AI429486, D19Ertd652e, Wdr96', description='cilia and flagella associated protein 43', other_designations='WD repeat domain 96', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47735267, end_position_on_the_genomic_accession=47837836, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  433845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433845, CurrentID=0, Status='live', Symbol='Gm10481', Aliases='ENSMUSG00000073212', description='predicted gene 10481', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=13042242, end_position_on_the_genomic_accession=13043452, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  433855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433855, CurrentID=0, Status='live', Symbol='AI506816', Aliases='', description='expressed sequence AI506816', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23692261, end_position_on_the_genomic_accession=23712667, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  433864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433864, CurrentID=0, Status='live', Symbol='Nom1', Aliases='D5Kng1, Gm1040', description='nucleolar protein with MIF4G domain 1', other_designations='SGD1 homolog', map_location='5 15.0 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=29434659, end_position_on_the_genomic_accession=29452169, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  433899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433899, CurrentID=0, Status='live', Symbol='Grxcr1', Aliases='Tg(Eno2-Gabrb3)0370Brll, pi, tde', description='glutaredoxin, cysteine rich 1', other_designations='glutaredoxin cysteine-rich 1 protein', map_location='5 36.59 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=68031835, end_position_on_the_genomic_accession=68166398, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  433904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433904, CurrentID=0, Status='live', Symbol='Ociad2', Aliases='1810027I20Rik', description='OCIA domain containing 2', other_designations='', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73322199, end_position_on_the_genomic_accession=73339355, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  433926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433926, CurrentID=0, Status='live', Symbol='Lrrc8b', Aliases='R75581, Ta-lrrp, mKIAA0231', description='leucine rich repeat containing 8 family, member B', other_designations='T-cell activation leucine repeat-rich protein|leucine-rich repeat-containing protein 8B', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105415775, end_position_on_the_genomic_accession=105490071, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  433931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433931, CurrentID=0, Status='live', Symbol='Pigg', Aliases='Gpi7', description='phosphatidylinositol glycan anchor biosynthesis, class G', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108312783, end_position_on_the_genomic_accession=108349359, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  433938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433938, CurrentID=0, Status='live', Symbol='Mn1', Aliases='AA003644, AA009236', description='meningioma 1', other_designations='probable tumor suppressor protein MN1', map_location='5 53.86 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=111418166, end_position_on_the_genomic_accession=111457025, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  433940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433940, CurrentID=0, Status='live', Symbol='Fam222a', Aliases='', description='family with sequence similarity 222, member A', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114568251, end_position_on_the_genomic_accession=114613220, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  433956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433956, CurrentID=0, Status='live', Symbol='Dnaaf5', Aliases='BC053401, C76907, Heatr2', description='dynein, axonemal assembly factor 5', other_designations='HEAT repeat containing 2|HEAT repeat-containing protein 2|dynein, axonemal, assembly factor 5', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139150223, end_position_on_the_genomic_accession=139186505, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100310872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100310872, CurrentID=0, Status='live', Symbol='Dynlt1a', Aliases='', description='dynein light chain Tctex-type 1A', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6306344, end_position_on_the_genomic_accession=6317474, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  434016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434016, CurrentID=0, Status='live', Symbol='Vmn1r7', Aliases='EG434016, Gm5568, V1rc31', description='vomeronasal 1 receptor 7', other_designations='vomernasal 1 receptor Vmn1r7|vomeronasal 1 receptor, C31', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57024338, end_position_on_the_genomic_accession=57025273, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  434025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434025, CurrentID=0, Status='live', Symbol='Igkv9-120', Aliases='EG434025, Gm5571', description='immunoglobulin kappa chain variable 9-120', other_designations='IgG3/kappa antibody|anti-DNA antibody light chain variable region|anti-poly(dC) monoclonal antibody kappa light chain|monoclonal antibody kappa light chain', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68049983, end_position_on_the_genomic_accession=68050454, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  434031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434031, CurrentID=0, Status='live', Symbol='Igkv10-95', Aliases='EG434031, Gm5573, Vk10c', description='immunoglobulin kappa variable 10-95', other_designations='immunoglobulin kappa chain variable 10-95', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68680379, end_position_on_the_genomic_accession=68680848, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  434033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434033, CurrentID=0, Status='live', Symbol='Igkv4-91', Aliases='ENSMUSG00000056850, ENSMUSG00000076532, Gm10879, kf4', description='immunoglobulin kappa chain variable 4-91', other_designations='Fv fragment variable light chain|anti-myosin immunoglobulin light chain variable region|immunoglobulin light chain variable region', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68768555, end_position_on_the_genomic_accession=68769103, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  434041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434041, CurrentID=0, Status='live', Symbol='Igkv8-27', Aliases='Gm16944', description='immunoglobulin kappa chain variable 8-27', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70171809, end_position_on_the_genomic_accession=70172270, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  434121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434121, CurrentID=0, Status='live', Symbol='Sult2a4', Aliases='EG434121, Gm5584', description='sulfotransferase family 2A, dehydroepiandrosterone (DHEA)-preferring, member 4', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13909677, end_position_on_the_genomic_accession=13989591, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  434128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434128, CurrentID=0, Status='live', Symbol='Pnmal2', Aliases='EG434128', description='PNMA-like 2', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16944682, end_position_on_the_genomic_accession=16948828, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  434130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434130, CurrentID=0, Status='live', Symbol='Ccdc8', Aliases='ENSMUSG00000041117', description='coiled-coil domain containing 8', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16994588, end_position_on_the_genomic_accession=16996645, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  434147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434147, CurrentID=0, Status='live', Symbol='D930028M14Rik', Aliases='EG434147', description='RIKEN cDNA D930028M14 gene', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25152457, end_position_on_the_genomic_accession=25156630, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  434156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434156, CurrentID=0, Status='live', Symbol='Eid2b', Aliases='3010005C08Rik', description='EP300 interacting inhibitor of differentiation 2B', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28277706, end_position_on_the_genomic_accession=28280129, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100048884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100048884, CurrentID=0, Status='live', Symbol='Mup18', Aliases='100039228, Gm12561, Mup16, Mup6, Mup9, OTTMUSG00000007486', description='major urinary protein 18', other_designations='MUP 6|Major urinary protein 6|alpha-2U-globulin|group 1, BS6|major urinary protein 16|major urinary protein 9|novel member of the major urinary protein (Mup) gene family', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61670177, end_position_on_the_genomic_accession=61674098, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100048885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100048885, CurrentID=0, Status='live', Symbol='Gm2083', Aliases='', description='major urinary protein LOC100048885', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  434171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434171, CurrentID=0, Status='live', Symbol='Gm5591', Aliases='EG434171', description='predicted gene 5591', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=38518139, end_position_on_the_genomic_accession=38528193, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  434172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434172, CurrentID=0, Status='live', Symbol='Gm5592', Aliases='EG434172', description='predicted gene 5592', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=41284327, end_position_on_the_genomic_accession=41290183, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  434179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434179, CurrentID=0, Status='live', Symbol='Gm5595', Aliases='EG434179, mszf88', description='predicted gene 5595', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=42660105, end_position_on_the_genomic_accession=42692732, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  434197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434197, CurrentID=0, Status='live', Symbol='Fam169b', Aliases='EG434197', description='family with sequence similarity 169, member B', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=68273839, end_position_on_the_genomic_accession=68363089, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  434203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434203, CurrentID=0, Status='live', Symbol='Slc28a1', Aliases='Cnt1', description='solute carrier family 28 (sodium-coupled nucleoside transporter), member 1', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81113269, end_position_on_the_genomic_accession=81170416, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  434204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434204, CurrentID=0, Status='live', Symbol='Whamm', Aliases='BB081391, Whdc1, mKIAA1971', description='WAS protein homolog associated with actin, golgi membranes and microtubules', other_designations='WAS protein homology region 2 domain containing 1|WAS protein homology region 2 domain-containing protein 1|WH2 domain-containing protein 1', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81571170, end_position_on_the_genomic_accession=81596837, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  434215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434215, CurrentID=0, Status='live', Symbol='Lrrc32', Aliases='AI426318, D11S833Eh, D7H11S833E, EG434215, Garp', description='leucine rich repeat containing 32', other_designations='glycoprotein A repetitions predominant', map_location='7 53.86 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98489301, end_position_on_the_genomic_accession=98501831, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  434218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434218, CurrentID=0, Status='live', Symbol='Trim34b', Aliases='Gm15134, OTTMUSG00000019188, Trim34-2', description='tripartite motif-containing 34B', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104329471, end_position_on_the_genomic_accession=104336617, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  434223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434223, CurrentID=0, Status='live', Symbol='Gm1966', Aliases='', description='predicted gene 1966', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106595549, end_position_on_the_genomic_accession=106644646, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  94216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94216, CurrentID=0, Status='live', Symbol='Col4a6', Aliases='BB116301', description='collagen, type IV, alpha 6', other_designations='procollagen type IV alpha 6|procollagen, type IV, alpha 6', map_location='X F1-F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=141165403, end_position_on_the_genomic_accession=141474076, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  94217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94217, CurrentID=0, Status='live', Symbol='Lrp1b', Aliases='9630004P12Rik, LRP-DIT', description='low density lipoprotein-related protein 1B (deleted in tumors)', other_designations='LRP-1B|low-density lipoprotein receptor-related protein-deleted in tumor', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=40595290, end_position_on_the_genomic_accession=42653631, orientation='minus', exon_count=94, OMIM='', no_hdr0=''),
  434232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434232, CurrentID=0, Status='live', Symbol='Iqck', Aliases='A230094G09Rik', description='IQ motif containing K', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118855755, end_position_on_the_genomic_accession=118972652, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  434234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434234, CurrentID=0, Status='live', Symbol='2610020H08Rik', Aliases='', description='RIKEN cDNA 2610020H08 gene', other_designations='exonuclease NEF-sp', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119794130, end_position_on_the_genomic_accession=119882695, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  434246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434246, CurrentID=0, Status='live', Symbol='Trim72', Aliases='BC067209, MG53', description='tripartite motif-containing 72', other_designations='mitsugumin 53|mitsugumin-53|trim72', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128003426, end_position_on_the_genomic_accession=128011393, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  434341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434341, CurrentID=0, Status='live', Symbol='Nlrc5', Aliases='AI451557, AK220210, NOD27', description='NLR family, CARD domain containing 5', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94434354, end_position_on_the_genomic_accession=94527272, orientation='plus', exon_count=54, OMIM='', no_hdr0=''),
  434377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434377, CurrentID=0, Status='live', Symbol='Zfp560', Aliases='2310030G09Rik, Znf560', description='zinc finger protein 560', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20345136, end_position_on_the_genomic_accession=20385158, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  434402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434402, CurrentID=0, Status='live', Symbol='Gm5617', Aliases='EG434402', description='predicted gene 5617', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48495343, end_position_on_the_genomic_accession=48495975, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  434423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434423, CurrentID=0, Status='live', Symbol='Dppa5a', Aliases='AA536857, Dppa5, Esg1, ecat2', description='developmental pluripotency associated 5A', other_designations='ES cell-associated transcript 2 protein|ESG-1|developmental pluripotency associated 5|embryonal stem cell specific gene 1|embryonal stem cell-specific gene 1 protein|mDppa5|protein pH 34', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78367052, end_position_on_the_genomic_accession=78368199, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  434437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434437, CurrentID=0, Status='live', Symbol='Amt', Aliases='EG434437', description='aminomethyltransferase', other_designations='GCVT|glycine cleavage system T protein', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108296922, end_position_on_the_genomic_accession=108301722, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  434439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434439, CurrentID=0, Status='live', Symbol='BC048562', Aliases='', description='cDNA sequence BC048562', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108436482, end_position_on_the_genomic_accession=108446083, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  434440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434440, CurrentID=0, Status='live', Symbol='Fbxw20', Aliases='Fbxw12', description='F-box and WD-40 domain protein 20', other_designations='F-box and WD-40 domain protein 12', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109217432, end_position_on_the_genomic_accession=109234754, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  225326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225326, CurrentID=0, Status='live', Symbol='Pik3c3', Aliases='5330434F23Rik, Vps34', description='phosphoinositide-3-kinase, class 3', other_designations='PI3-kinase type 3|PI3K type 3|ptdIns-3-kinase type 3', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=30272094, end_position_on_the_genomic_accession=30348126, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  434674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434674, CurrentID=0, Status='live', Symbol='Slc22a28', Aliases='EG434674, Gm5631', description='solute carrier family 22, member 28', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8062209, end_position_on_the_genomic_accession=8131982, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  434676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434676, CurrentID=0, Status='live', Symbol='Scgb1b19', Aliases='Abpa19, EG434676, Gm5632', description='secretoglobin, family 1B, member 19', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33287307, end_position_on_the_genomic_accession=33288491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  434729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434729, CurrentID=0, Status='live', Symbol='Fthl17f', Aliases='EG434729, Fthl17L5, Gm5635', description='ferritin, heavy polypeptide-like 17, member F', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9063090, end_position_on_the_genomic_accession=9063822, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  434756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434756, CurrentID=0, Status='live', Symbol='Akap14', Aliases='', description='A kinase (PRKA) anchor protein 14', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37150698, end_position_on_the_genomic_accession=37168842, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  434758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434758, CurrentID=0, Status='live', Symbol='Rhox3h', Aliases='EG434758, Rhox3.8, Rhox3g', description='reproductive homeobox 3H', other_designations='reproductive homeobox 3G', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37657688, end_position_on_the_genomic_accession=37668764, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  434759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434759, CurrentID=0, Status='live', Symbol='Rhox4c', Aliases='Rhox4.3, Rhox4b', description='reproductive homeobox 4C', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37480337, end_position_on_the_genomic_accession=37485124, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  434760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434760, CurrentID=0, Status='live', Symbol='Rhox2d', Aliases='EG434760, Rhox2.4, Rhox2c', description='reproductive homeobox 2D', other_designations='reproductive homeobox 2C', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37493410, end_position_on_the_genomic_accession=37497715, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  434764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434764, CurrentID=0, Status='live', Symbol='Rhox2f', Aliases='EG434764, Rhox2.6, Rhox2e', description='reproductive homeobox 2F', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37571421, end_position_on_the_genomic_accession=37576159, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  434766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434766, CurrentID=0, Status='live', Symbol='Rhox2g', Aliases='EG434766, Rhox2.7, Rhox2f', description='reproductive homeobox 2G', other_designations='reproductive homeobox 2F', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37639112, end_position_on_the_genomic_accession=37643470, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  434768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434768, CurrentID=0, Status='live', Symbol='Rhox8', Aliases='Tox', description='reproductive homeobox 8', other_designations='homeoprotein TOX', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37874776, end_position_on_the_genomic_accession=37890766, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  434769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434769, CurrentID=0, Status='live', Symbol='Rhox10', Aliases='', description='reproductive homeobox 10', other_designations='reproductive homeobox on X chromosome 10', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38066475, end_position_on_the_genomic_accession=38071688, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  434778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434778, CurrentID=0, Status='live', Symbol='Ccdc160', Aliases='6330534C20Rik', description='coiled-coil domain containing 160', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52791200, end_position_on_the_genomic_accession=52799474, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  434784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434784, CurrentID=0, Status='live', Symbol='Ldoc1', Aliases='Gm366, Mar7, Mart7', description='leucine zipper, down-regulated in cancer 1', other_designations='breast cancer, up-regulated 1', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=61709616, end_position_on_the_genomic_accession=61710950, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  434794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434794, CurrentID=0, Status='live', Symbol='Xlr4a', Aliases='', description='X-linked lymphocyte-regulated 4A', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73074345, end_position_on_the_genomic_accession=73082507, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  434795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434795, CurrentID=0, Status='live', Symbol='Gm5639', Aliases='EG434795', description='predicted pseudogene 5639', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73297610, end_position_on_the_genomic_accession=73299979, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100049549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049549, CurrentID=0, Status='live', Symbol='Mir181d', Aliases='Mirn181d, mir-181d', description='microRNA 181d', other_designations='mmu-mir-181d', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84178716, end_position_on_the_genomic_accession=84178787, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100049584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049584, CurrentID=0, Status='live', Symbol='Sbmd5', Aliases='Pmsbmd1', description='spinal bone mineral density 5', other_designations='post-maturity spinal BMD 1', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  434881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434881, CurrentID=0, Status='live', Symbol='Samt2', Aliases='EG434881, Gm5647', description='spermatogenesis associated multipass transmembrane protein 2', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154575228, end_position_on_the_genomic_accession=154579330, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  434903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434903, CurrentID=0, Status='live', Symbol='Mageb4', Aliases='CN716893, Mage-b4, mMage-b4', description='melanoma antigen, family B, 4', other_designations='', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=86250256, end_position_on_the_genomic_accession=86256219, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100049710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049710, CurrentID=0, Status='live', Symbol='Mir290a', Aliases='Mir290, Mirn290, mir-290a, mmu-miR-290, mmu-mir-290a', description='microRNA 290a', other_designations='microRNA 290|microRNA mir-290', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3218626, end_position_on_the_genomic_accession=3218708, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100049711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049711, CurrentID=0, Status='live', Symbol='Mir292', Aliases='Mirn292, mir-292a, mmu-mir-292, mmu-mir-292a', description='microRNA 292', other_designations='microRNA mir-292', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3219189, end_position_on_the_genomic_accession=3219270, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100049712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049712, CurrentID=0, Status='live', Symbol='Mir294', Aliases='Mirn294, mir-294, mmu-miR-294', description='microRNA 294', other_designations='microRNA mir-294', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3220641, end_position_on_the_genomic_accession=3220724, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100049713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049713, CurrentID=0, Status='live', Symbol='Mir295', Aliases='Mirn295, mir-295, mmu-miR-295', description='microRNA 295', other_designations='microRNA mir-295', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3220773, end_position_on_the_genomic_accession=3220841, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100049714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049714, CurrentID=0, Status='live', Symbol='Mir293', Aliases='Mirn293, mir-293, mmu-mir-293', description='microRNA 293', other_designations='microRNA mir-293', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3220343, end_position_on_the_genomic_accession=3220422, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100049715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100049715, CurrentID=0, Status='live', Symbol='Mir291a', Aliases='Mirn291a, mir-291a', description='microRNA 291a', other_designations='microRNA mir-291a|mmu-miR-291a', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3218919, end_position_on_the_genomic_accession=3219000, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  435145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435145, CurrentID=0, Status='live', Symbol='Shisa8', Aliases='Orf26', description='shisa family member 8', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82206952, end_position_on_the_genomic_accession=82212815, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  435206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435206, CurrentID=0, Status='live', Symbol='Taar7d', Aliases='taR-7d', description='trace amine-associated receptor 7D', other_designations='mTaar7d', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24027222, end_position_on_the_genomic_accession=24028298, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  435207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435207, CurrentID=0, Status='live', Symbol='Taar7f', Aliases='taR-7f', description='trace amine-associated receptor 7F', other_designations='mTaar7f', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24049510, end_position_on_the_genomic_accession=24050586, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  435337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435337, CurrentID=0, Status='live', Symbol='Gm5662', Aliases='EG435337', description='predicted gene 5662', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=88270635, end_position_on_the_genomic_accession=88274497, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  435350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435350, CurrentID=0, Status='live', Symbol='Serpinb6e', Aliases='Gm11396, OTTMUSG00000000720, SPI3B, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 6e', other_designations='novel serine (or cysteine) proteinase inhibitor, clade B (Serpinb) member|serine (or cysteine) proteinase inhibitor, clade B, member 6e', map_location='13 13.98 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33832292, end_position_on_the_genomic_accession=33843408, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  435391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435391, CurrentID=0, Status='live', Symbol='Dupd1', Aliases='EG435391', description='dual specificity phosphatase and pro isomerase domain containing 1', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21676209, end_position_on_the_genomic_accession=21714576, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  435529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435529, CurrentID=0, Status='live', Symbol='Adgrf2', Aliases='Gpr111, PGR20', description='adhesion G protein-coupled receptor F2', other_designations='G protein-coupled receptor 111|G protein-coupled receptor PGR20|G-protein coupled receptor 111|G-protein coupled receptor PGR20|probable G-protein coupled receptor 111', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=42694028, end_position_on_the_genomic_accession=42749145, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100312484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100312484, CurrentID=0, Status='live', Symbol='Vmn1r196', Aliases='V1rh19', description='vomeronasal 1 receptor 196', other_designations='vomeronasal 1 receptor Vmn1r196|vomeronasal type 1 receptor H19', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22293193, end_position_on_the_genomic_accession=22294098, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  435626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435626, CurrentID=0, Status='live', Symbol='Rufy4', Aliases='F930048N03Rik', description='RUN and FYVE domain containing 4', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74124618, end_position_on_the_genomic_accession=74148230, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  435653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435653, CurrentID=0, Status='live', Symbol='Fcrlb', Aliases='FREB-2, FREB2, FcRL2, Fcrlm2, Fcry, mFCRL2', description='Fc receptor-like B', other_designations='Fc receptor homolog expressed in B-cells protein 2|Fc receptor related Y|Fc receptor-like mucin-like 2|fc receptor homolog expressed in B cells protein 2|fc receptor-like and mucin-like protein 2|fc receptor-like protein 2|fc receptor-related protein Y', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170906852, end_position_on_the_genomic_accession=170913860, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  435684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435684, CurrentID=0, Status='live', Symbol='Shf', Aliases='', description='Src homology 2 domain containing F', other_designations='shf', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122348892, end_position_on_the_genomic_accession=122369169, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  435743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435743, CurrentID=0, Status='live', Symbol='Aqp10-ps', Aliases='AQP10p, EG668102, Gm8977', description='aquaporin 10, pseudogene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89963444, end_position_on_the_genomic_accession=89969734, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  435766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435766, CurrentID=0, Status='live', Symbol='Tnni3k', Aliases='Cark, D830019J24Rik', description='TNNI3 interacting kinase', other_designations='TNNI3-interacting kinase|cardiac ankyrin repeat kinase|tnni3k', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=154786291, end_position_on_the_genomic_accession=155055407, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  435791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435791, CurrentID=0, Status='live', Symbol='Gm13271', Aliases='Ifnz, OTTMUSG00000011262', description='predicted gene 13271', other_designations='interferon zeta', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88754868, end_position_on_the_genomic_accession=88755416, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  435804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435804, CurrentID=0, Status='live', Symbol='Olfr1335', Aliases='MOR259-12, MOR259-5, Olfr1334-ps1-ps1, Olfr1335', description='olfactory receptor 1335', other_designations='GA_x6K02T2QD9B-18723799-18724749|GA_x6K02T2QD9B-18726774-18727577|olfactory receptor 1334, pseudogene 1|olfactory receptor MOR259-5', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118808855, end_position_on_the_genomic_accession=118809862, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  435811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435811, CurrentID=0, Status='live', Symbol='Ldlrad2', Aliases='', description='low density lipoprotein receptor A domain containing 2', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137570876, end_position_on_the_genomic_accession=137582385, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  435845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435845, CurrentID=0, Status='live', Symbol='Tmprss11c', Aliases='Tmprss13', description='transmembrane protease, serine 11c', other_designations='neurobin', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86231481, end_position_on_the_genomic_accession=86289308, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  435921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435921, CurrentID=0, Status='live', Symbol='Clec2f', Aliases='Clec2f-ps, Clr-c, Clrc, Clrz', description='C-type lectin domain family 2, member f', other_designations='C lectin-related protein c|C-type lectin domain family 2, member f, pseudogene|C-type lectin-related protein C', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129008803, end_position_on_the_genomic_accession=129021342, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  435965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=435965, CurrentID=0, Status='live', Symbol='Lrp3', Aliases='', description='low density lipoprotein receptor-related protein 3', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35200878, end_position_on_the_genomic_accession=35215493, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  436003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436003, CurrentID=0, Status='live', Symbol='Hbb-bh2', Aliases='EG436003, Gm5736', description='hemoglobin beta, bh2', other_designations='hemoglobin beta, pseudogene bh2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103838801, end_position_on_the_genomic_accession=103840521, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  436022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436022, CurrentID=0, Status='live', Symbol='Dnaaf3', Aliases='6030429G01Rik, b2b1739Clo', description='dynein, axonemal assembly factor 3', other_designations='UPF0470 protein C19orf51 homolog', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4522948, end_position_on_the_genomic_accession=4532488, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  436062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436062, CurrentID=0, Status='live', Symbol='Fam92b', Aliases='1700120B06Rik', description='family with sequence similarity 92, member B', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120166397, end_position_on_the_genomic_accession=120177486, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  436090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436090, CurrentID=0, Status='live', Symbol='Gpr62', Aliases='4933402E03Rik', description='G protein-coupled receptor 62', other_designations='G protein-coupled receptor 61 pseudogene|G protein-coupled receptor GPR62', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106463960, end_position_on_the_genomic_accession=106466612, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100312987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100312987, CurrentID=0, Status='live', Symbol='Gm17252', Aliases='Pate', description='predicted gene, 17252', other_designations='prostate and testis expressed', map_location='9 20.39 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35685079, end_position_on_the_genomic_accession=35687371, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  436199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436199, CurrentID=0, Status='live', Symbol='Btg1-ps1', Aliases='4930525M21Rik', description='B cell translocation gene 1, anti-proliferative, pseudogene 1', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38013173, end_position_on_the_genomic_accession=38026092, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  436230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436230, CurrentID=0, Status='live', Symbol='BC065397', Aliases='R75336, parl7', description='cDNA sequence BC065397', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136742957, end_position_on_the_genomic_accession=136803364, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  436240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436240, CurrentID=0, Status='live', Symbol='Foxr2', Aliases='EG436240, Foxn6', description='forkhead box R2', other_designations='forkhead box protein N6', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153016795, end_position_on_the_genomic_accession=153132861, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  436440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436440, CurrentID=0, Status='live', Symbol='Gpr31b', Aliases='12-HETER, GPR31c(t), Gpr31, Gpr31c', description='G protein-coupled receptor 31, D17Leh66b region', other_designations='12-(S)-HETE receptor|G-protein coupled receptor 31', map_location='17 8.76 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13051321, end_position_on_the_genomic_accession=13052280, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  436522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436522, CurrentID=0, Status='live', Symbol='Try10', Aliases='', description='trypsin 10', other_designations='trypsinogen 10', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41354105, end_position_on_the_genomic_accession=41357944, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  436579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=436579, CurrentID=0, Status='live', Symbol='D5Mit286', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 286', other_designations='', map_location='5 89.3 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100313531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100313531, CurrentID=0, Status='live', Symbol='Nkx2-2os', Aliases='Gm14112, Nkx2-2as, Nkx2.2AS, OTTMUSG00000015691', description='NK2 homeobox 2, opposite strand', other_designations='NK2 transcription factor related, locus 2, antisense', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=147184083, end_position_on_the_genomic_accession=147331681, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  116748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116748, CurrentID=0, Status='live', Symbol='Lsm10', Aliases='', description='U7 snRNP-specific Sm-like protein LSM10', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126096577, end_position_on_the_genomic_accession=126098584, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101101691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101101691, CurrentID=0, Status='live', Symbol='Hm629797', Aliases='Mrhl', description='cDNA sequence HM629797', other_designations='meiotic recombination hotspot locus RNA', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  226115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226115, CurrentID=0, Status='live', Symbol='Opalin', Aliases='Tmem10, Tmp10', description='oligodendrocytic myelin paranodal and inner loop protein', other_designations='transmembrane protein 10|transmembrane protein TMP10', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41063420, end_position_on_the_genomic_accession=41077113, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100316665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100316665, CurrentID=0, Status='live', Symbol='Mir466i', Aliases='Mirn466i, mmu-mir-466i', description='microRNA 466i', other_designations='microRNA mir-466i', map_location='13 6.05 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=17747473, end_position_on_the_genomic_accession=17747593, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100316669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100316669, CurrentID=0, Status='live', Symbol='Mir1188', Aliases='Mirn1188, mir-1188, mmu-mir-1188', description='microRNA 1188', other_designations='microRNA mir-1188', map_location='12 60.43 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109611822, end_position_on_the_genomic_accession=109611941, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100316739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100316739, CurrentID=0, Status='live', Symbol='Mir1224', Aliases='Mirn1224, mir-1224, mmu-mir-1224', description='microRNA 1224', other_designations='microRNA mir-1224', map_location='16 12.48 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20604452, end_position_on_the_genomic_accession=20604536, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100316769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100316769, CurrentID=0, Status='live', Symbol='Mir466l', Aliases='Mirn466l, mmu-mir-466l', description='microRNA 466l', other_designations='microRNA mir-466l', map_location='2 8.22 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10516097, end_position_on_the_genomic_accession=10516217, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100316783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100316783, CurrentID=0, Status='live', Symbol='Mir1b', Aliases='Mir1-2-as, Mir1-2as, Mirn1-2as, mir-1b, mmu-mir-1-2-as, mmu-mir-1b', description='microRNA 1b', other_designations='microRNA 1-2, antisense|microRNA mir-1-2-as', map_location='18 5.4 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10785446, end_position_on_the_genomic_accession=10785567, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100316814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100316814, CurrentID=0, Status='live', Symbol='Mir1306', Aliases='mmu-mir-1306', description='microRNA 1306', other_designations='microRNA mir-1306', map_location='16 11.33 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18284239, end_position_on_the_genomic_accession=18284317, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  445007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=445007, CurrentID=0, Status='live', Symbol='Nup85', Aliases='Pcnt1, frount', description='nucleoporin 85', other_designations='85 kDa nucleoporin|nucleoporin Nup85|pericentrin-1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115564432, end_position_on_the_genomic_accession=115583924, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  442801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=442801, CurrentID=0, Status='live', Symbol='Arhgef15', Aliases='D130071N09, D530030K12Rik, E5', description='Rho guanine nucleotide exchange factor (GEF) 15', other_designations='ephexin-5', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68927611, end_position_on_the_genomic_accession=68956934, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  442809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=442809, CurrentID=0, Status='live', Symbol='4932416K20Rik', Aliases='', description='RIKEN cDNA 4932416K20 gene', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104795237, end_position_on_the_genomic_accession=104798792, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  442827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=442827, CurrentID=0, Status='live', Symbol='Rab44', Aliases='9830134C10Rik', description='RAB44, member RAS oncogene family', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29114128, end_position_on_the_genomic_accession=29148980, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  442829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=442829, CurrentID=0, Status='live', Symbol='Ccin', Aliases='4933417A14Rik', description='calicin', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43983504, end_position_on_the_genomic_accession=43985533, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  228607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228607, CurrentID=0, Status='live', Symbol='Mavs', Aliases='D430028G21Rik, IPS-1, Visa, cardif', description='mitochondrial antiviral signaling protein', other_designations='CARD adapter inducing interferon beta|IFN-beta promoter stimulator-1|interferon beta promoter stimulator protein 1|mitochondrial anti-viral signaling protein|virus-induced signaling adapter', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131234063, end_position_on_the_genomic_accession=131248025, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  442834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=442834, CurrentID=0, Status='live', Symbol='D830031N03Rik', Aliases='Kiaa0754, mKIAA0754', description='RIKEN cDNA D830031N03 gene', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123403601, end_position_on_the_genomic_accession=123411911, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  442835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=442835, CurrentID=0, Status='live', Symbol='Defb22', Aliases='9230002F21Rik', description='defensin beta 22', other_designations='beta-defensin 22', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152485666, end_position_on_the_genomic_accession=152490138, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100188919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100188919, CurrentID=0, Status='live', Symbol='Israa', Aliases='', description='immune system released activating agent', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25603547, end_position_on_the_genomic_accession=25603924, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574402, CurrentID=0, Status='live', Symbol='Gpr17', Aliases='AI853548', description='G protein-coupled receptor 17', other_designations='G-protein coupled receptor 17|UDP/CysLT receptor|gpr17', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31942999, end_position_on_the_genomic_accession=31949636, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  574403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574403, CurrentID=0, Status='live', Symbol='Fam196b', Aliases='EG574403, Gm6041', description='family with sequence similarity 196, member B', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=34314822, end_position_on_the_genomic_accession=34422642, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  574405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574405, CurrentID=0, Status='live', Symbol='DXBay18', Aliases='', description='DNA segment, Chr X, Baylor 18', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73137223, end_position_on_the_genomic_accession=73149450, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  574417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574417, CurrentID=0, Status='live', Symbol='Tas2r137', Aliases='T2R37, Tas2r3, mt2r41', description='taste receptor, type 2, member 137', other_designations='T2R137|T2R3|candidate taste receptor mt2r41|taste receptor T2R37', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40491238, end_position_on_the_genomic_accession=40492239, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  574418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574418, CurrentID=0, Status='live', Symbol='Serinc4', Aliases='', description='serine incorporator 4', other_designations='ortholog of serine incorporator 4 SERINC4', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121451177, end_position_on_the_genomic_accession=121456764, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  574428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574428, CurrentID=0, Status='live', Symbol='Zmynd15', Aliases='', description='zinc finger, MYND-type containing 15', other_designations='zinc finger, MYND domain containing 15', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70459622, end_position_on_the_genomic_accession=70466207, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  574429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574429, CurrentID=0, Status='live', Symbol='Psg26', Aliases='EG574429, cea14', description='pregnancy-specific glycoprotein 26', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18474577, end_position_on_the_genomic_accession=18484197, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  574437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574437, CurrentID=0, Status='live', Symbol='Xlr3b', Aliases='EG574437, Xlr3', description='X-linked lymphocyte-regulated 3B', other_designations='xlr3b', map_location='X A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73192152, end_position_on_the_genomic_accession=73202930, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  574438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574438, CurrentID=0, Status='live', Symbol='Xlr5a', Aliases='Xlr5d', description='X-linked lymphocyte-regulated 5A', other_designations='X-linked lymphocyte-regulated 5D', map_location='X A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73107621, end_position_on_the_genomic_accession=73119524, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  574519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=574519, CurrentID=0, Status='live', Symbol='Vax2os', Aliases='Vas2os1, Vax2os2, Vax2os', description='ventral anterior homeobox 2, opposite strand', other_designations='Vax2 antisense transcript Vax2OS|Vax2 opposite strand transcript 1|Vax2 opposite strand transcript 2|ventral anterior homeobox antisense', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83692806, end_position_on_the_genomic_accession=83712201, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  50490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50490, CurrentID=0, Status='live', Symbol='Nox4', Aliases='AI648021', description='NADPH oxidase 4', other_designations='KOX-1|kidney oxidase-1|kidney superoxide-producing NADPH oxidase|renal NAD(P)H-oxidase|superoxide-generating NADPH oxidase 4', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=87246405, end_position_on_the_genomic_accession=87398710, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  50492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50492, CurrentID=0, Status='live', Symbol='Thop1', Aliases='AI131655, AI327041, EP24.15', description='thimet oligopeptidase 1', other_designations='endopeptidase 24.15', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81070083, end_position_on_the_genomic_accession=81082360, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  50493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50493, CurrentID=0, Status='live', Symbol='Txnrd1', Aliases='TR, TR1, TrxR1', description='thioredoxin reductase 1', other_designations='TR alpha|thioredoxin reductase TR1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82833951, end_position_on_the_genomic_accession=82897724, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  50496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50496, CurrentID=0, Status='live', Symbol='E2f6', Aliases='AI462434, E2F6b, EMA', description='E2F transcription factor 6', other_designations='E2F-6|E2F-binding site-modulating activity protein|transcriptional repressor E2F6', map_location='12 8.04 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=16810903, end_position_on_the_genomic_accession=16826752, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  50497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50497, CurrentID=0, Status='live', Symbol='Hspa14', Aliases='70kDa, HSP70L1, Hsp70-4, NST-1, hsr.1', description='heat shock protein 14', other_designations='heat shock 70kDa protein 14|heat shock protein 4|heat shock protein, 70 kDa 4', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3488849, end_position_on_the_genomic_accession=3512936, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  50498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50498, CurrentID=0, Status='live', Symbol='Ebi3', Aliases='EBI-3, IL-27', description='Epstein-Barr virus induced gene 3', other_designations='IL-27 subunit beta|IL-27B|epstein-Barr virus-induced gene 3 protein homolog', map_location='17 29.08 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55952623, end_position_on_the_genomic_accession=55957022, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50500, CurrentID=0, Status='live', Symbol='Ttpa', Aliases='alpha-TTP', description='tocopherol (alpha) transfer protein', other_designations='alpha TTP', map_location='4 7.76 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=20008089, end_position_on_the_genomic_accession=20030785, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  50501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50501, CurrentID=0, Status='live', Symbol='Prok2', Aliases='Bv8, PK2, Prok1', description='prokineticin 2', other_designations='Bombina variegata 8 kD protein|Bombina variegata 8kDa protein|Bv 8 homolog (Bombina variegata)|prokineticin 1|protein Bv8 homolog', map_location='6 46.29 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=99711299, end_position_on_the_genomic_accession=99726392, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  50505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50505, CurrentID=0, Status='live', Symbol='Ercc4', Aliases='AI606920, Xpf', description='excision repair cross-complementing rodent repair deficiency, complementation group 4', other_designations='DNA excision repair protein ERCC-4|DNA repair endonuclease subunit', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13109689, end_position_on_the_genomic_accession=13152009, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  50510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50510, CurrentID=0, Status='live', Symbol='Igk-Eik', Aliases='', description='immunoglobulin kappa chain, enhancer intronic', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  50518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50518, CurrentID=0, Status='live', Symbol='a', Aliases='A<y>, ASIP, ASP, As', description='nonagouti', other_designations='agouti coat color protein|agouti signal protein|agouti switch protein|non-agouti', map_location='2 76.83 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154950162, end_position_on_the_genomic_accession=155051012, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  50523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50523, CurrentID=0, Status='live', Symbol='Lats2', Aliases='4932411G09Rik, AV277261, AW228608', description='large tumor suppressor 2', other_designations='kinase phosphorylated during mitosis protein|large tumor suppressor homolog 2|lats2|serine/threonine-protein kinase kpm', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57689662, end_position_on_the_genomic_accession=57746123, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  50524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50524, CurrentID=0, Status='live', Symbol='Sall2', Aliases='AI225809, AW559097, Msal-2, Sal-2, mKIAA0360, p150(Sal2)', description='sal-like 2 (Drosophila)', other_designations='spalt-like protein 2|zinc finger protein Spalt-2', map_location='14 B-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52311177, end_position_on_the_genomic_accession=52328670, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  50525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50525, CurrentID=0, Status='live', Symbol='Spag6', Aliases='PF16', description='sperm associated antigen 6', other_designations='axoneme central apparatus protein|axoneme protein|protein PF16 homolog', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16753016, end_position_on_the_genomic_accession=16829363, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  50527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50527, CurrentID=0, Status='live', Symbol='Ero1l', Aliases='ERO1-L, Ero1a', description='ERO1-like (S. cerevisiae)', other_designations='ERO1-L-alpha|endoplasmic oxidoreductin-1-like protein|endoplasmic reticulum oxidation 1|endoplasmic reticulum oxidoreductase alpha|endoplasmic reticulum oxidoreductin-1-like protein|oxidoreductin-1-L-alpha', map_location='14 C-D', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45283087, end_position_on_the_genomic_accession=45318572, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  50528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50528, CurrentID=0, Status='live', Symbol='Tmprss2', Aliases='D16Ertd61e', description='transmembrane protease, serine 2', other_designations='epitheliasin|plasmic transmembrane protein X', map_location='16 57.53 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97564682, end_position_on_the_genomic_accession=97659538, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  50529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50529, CurrentID=0, Status='live', Symbol='Mrps7', Aliases='MRP-S7, Rpms7', description='mitchondrial ribosomal protein S7', other_designations='S7mt|ribosomal protein, mitochondrial, S7', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115604151, end_position_on_the_genomic_accession=115607624, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50530, CurrentID=0, Status='live', Symbol='Mfap5', Aliases='MAGP-2', description='microfibrillar associated protein 5', other_designations='MFAP-5|microfibril-associated glycoprotein 2', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122513380, end_position_on_the_genomic_accession=122529290, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  50540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50540, CurrentID=0, Status='live', Symbol='Igbp1b', Aliases='alpha4-b', description='immunoglobulin (CD79A) binding protein 1b', other_designations='CD79a-binding protein 1b', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=138657089, end_position_on_the_genomic_accession=138658444, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100189605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100189605, CurrentID=0, Status='live', Symbol='Mup19', Aliases='100039247, Gm12552, Mup11, Mup14, Mup17, Mup8, OTTMUSG00000007472', description='major urinary protein 19', other_designations='MUP11 and MUP8|Major urinary proteins 11 and 8|major urinary protein 14|major urinary protein 17', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=61778324, end_position_on_the_genomic_accession=61782236, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  50701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50701, CurrentID=0, Status='live', Symbol='Elane', Aliases='Ela2, F430011M15Rik, NE', description='elastase, neutrophil expressed', other_designations='elastase 2, neutrophil|elastase-2|leukocyte elastase', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79886312, end_position_on_the_genomic_accession=79888216, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50702, CurrentID=0, Status='live', Symbol='Cfhr1', Aliases='AI194696, CFHRB, Cfhl1', description='complement factor H-related 1', other_designations='complement component factor h-like 1|complement factor H-related protein', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=139547064, end_position_on_the_genomic_accession=139560222, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  50706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50706, CurrentID=0, Status='live', Symbol='Postn', Aliases='A630052E07Rik, AI747096, OSF-2, Osf2, PLF, PN', description='periostin, osteoblast specific factor', other_designations='osteoblast specific factor 2 (fasciclin I-like)|osteoblast-specific factor 2|periostin-like factor', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54361096, end_position_on_the_genomic_accession=54391041, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  50708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50708, CurrentID=0, Status='live', Symbol='Hist1h1c', Aliases='0610008C09Rik, H1.2, H1c, H1f2, H1var1, His1a', description='histone cluster 1, H1c', other_designations='H1 VAR.1|H1 histone family, member 2|histone 1, H1c|histone H1', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23738807, end_position_on_the_genomic_accession=23739531, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50709, CurrentID=0, Status='live', Symbol='Hist1h1e', Aliases='H1-4, H1.4, H1e, H1f4, H1s-4, H1var2', description='histone cluster 1, H1e', other_designations='H1 VAR.2|H1 histone family, member 4|histone 1, H1e|histone H1e', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23621777, end_position_on_the_genomic_accession=23622558, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  50720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50720, CurrentID=0, Status='live', Symbol='Sacs', Aliases='A230052M14, DNAJC29, E130115J16Rik, mKIAA0730', description='sacsin', other_designations='dnaJ homolog subfamily C member 29', map_location='14 32.13 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61138427, end_position_on_the_genomic_accession=61240693, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  50721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50721, CurrentID=0, Status='live', Symbol='Sirt6', Aliases='2810449N18Rik, AI043036, Sir2l6', description='sirtuin 6', other_designations='NAD-dependent deacetylase sirtuin-6|SIR2-like protein 6|regulatory protein SIR2 homolog 6', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81621786, end_position_on_the_genomic_accession=81627792, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  50722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50722, CurrentID=0, Status='live', Symbol='Dkkl1', Aliases='AI414495, SGY, SGY-1, SGY1, Soggy, mSgy', description='dickkopf-like 1', other_designations='protein soggy-1|soggy 1', map_location='7 29.19 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45207525, end_position_on_the_genomic_accession=45212602, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  50723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50723, CurrentID=0, Status='live', Symbol='Icosl', Aliases='AU044799, B7-H2, B7RP-1, B7h, BG071784, GI50, GL50, GL50-B, ICOS-Lg, LICOS, Ly115l, mKIAA0653, Icosl', description='icos ligand', other_designations='B7 homolog 2|B7-like protein Gl50|B7-related protein 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78069368, end_position_on_the_genomic_accession=78079525, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  50724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50724, CurrentID=0, Status='live', Symbol='Sap30l', Aliases='2310079P12Rik, AF006998, L55', description='SAP30-like', other_designations='sin3 corepressor complex subunit SAP30L|sin3-associated protein p30-like|sin3A-associated protein p30-like protein', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=57801637, end_position_on_the_genomic_accession=57810615, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  50753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50753, CurrentID=0, Status='live', Symbol='Fbxo8', Aliases='Fbx8', description='F-box protein 8', other_designations='', map_location='8 B3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=56551134, end_position_on_the_genomic_accession=56593939, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  50754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50754, CurrentID=0, Status='live', Symbol='Fbxw7', Aliases='1110001A17Rik, AGO, Cdc4, Fbw7, Fbwd6, Fbx30, Fbxo30, Fbxw6, SEL-10', description='F-box and WD-40 domain protein 7', other_designations='F-box and WD-40 domain protein 7, archipelago homolog|F-box and WD-40 domain-containing protein 7|F-box-WD40 repeat protein 6|f-box only protein 30', map_location='3 37.7 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84815577, end_position_on_the_genomic_accession=84979198, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  50755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50755, CurrentID=0, Status='live', Symbol='Fbxo18', Aliases='AU015756, Fbh1, Fbx18', description='F-box protein 18', other_designations='F-box only protein 18', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=11742573, end_position_on_the_genomic_accession=11777585, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  50757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50757, CurrentID=0, Status='live', Symbol='Fbxw14', Aliases='AU045528, E330009N23Rik, Fbx12, Fbxo12', description='F-box and WD-40 domain protein 14', other_designations='F-box only protein 12|F-box protein 12', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109270789, end_position_on_the_genomic_accession=109287729, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  50758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50758, CurrentID=0, Status='live', Symbol='Fbxl17', Aliases='6330576B01Rik, AI452053, BB073797, C130023C01Rik, Fbx13, Fbxo13', description='F-box and leucine-rich repeat protein 17', other_designations='F-box only protein 13', map_location='17 E1.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=63045949, end_position_on_the_genomic_accession=63504042, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  50759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50759, CurrentID=0, Status='live', Symbol='Fbxo16', Aliases='4932435C24Rik, Fbx16', description='F-box protein 16', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65266123, end_position_on_the_genomic_accession=65323655, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  50760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50760, CurrentID=0, Status='live', Symbol='Fbxo17', Aliases='Fbg4, Fbx17, Fbxo26', description='F-box protein 17', other_designations='F-box only protein 26', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28716748, end_position_on_the_genomic_accession=28738144, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  50762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50762, CurrentID=0, Status='live', Symbol='Fbxo6', Aliases='AA408845, FBG2, Fbs2, Fbx6bb, Fbxo6', description='F-box protein 6', other_designations='F-box only protein 6b|F-box protein that recognizes sugar chains 2|F-box/G-domain protein 2', map_location='4 78.67 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148145716, end_position_on_the_genomic_accession=148152135, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  50764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50764, CurrentID=0, Status='live', Symbol='Fbxo15', Aliases='AU019763, Fbx15, ecat3', description='F-box protein 15', other_designations='', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84934552, end_position_on_the_genomic_accession=84981476, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  50765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50765, CurrentID=0, Status='live', Symbol='Tfr2', Aliases='Trfr2', description='transferrin receptor 2', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137569822, end_position_on_the_genomic_accession=137588081, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  50766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50766, CurrentID=0, Status='live', Symbol='Crim1', Aliases='AU015004', description='cysteine rich transmembrane BMP regulator 1 (chordin like)', other_designations='CRIM-1', map_location='17 E2-E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78200185, end_position_on_the_genomic_accession=78376592, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  50767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50767, CurrentID=0, Status='live', Symbol='Pnpla6', Aliases='AI661849, MSws, Nte', description='patatin-like phospholipase domain containing 6', other_designations='Swiss cheese|Swiss-cheese|patatin-like phospholipase domain-containing protein 6', map_location='8 1.92 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3515340, end_position_on_the_genomic_accession=3544267, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  50768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50768, CurrentID=0, Status='live', Symbol='Dlc1', Aliases='A730069N07Rik, Arhgap7, HP, STARD12, dlc-1', description='deleted in liver cancer 1', other_designations='START domain-containing protein 12|p122-RhoGAP|rho-type GTPase-activating protein 7|stAR-related lipid transfer protein 12', map_location='8 A4-B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=36567739, end_position_on_the_genomic_accession=36953165, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  50769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50769, CurrentID=0, Status='live', Symbol='Atp8a2', Aliases='AI415030, Atpc1b, agil, wl', description='ATPase, aminophospholipid transporter-like, class I, type 8A, member 2', other_designations='ATPase 8A2, class I|ATPase 8A2, p type|ATPase class I type 8A member 2|ATPase, class 1, member a|P4-ATPase flippase complex alpha subunit ATP8A2|probable phospholipid-transporting ATPase IB', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59647320, end_position_on_the_genomic_accession=60197079, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  50770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50770, CurrentID=0, Status='live', Symbol='Atp11a', Aliases='4930558F19Rik, AU040868, Atpc1h, Ih', description='ATPase, class VI, type 11A', other_designations='9130422H11Rik|ATPase 11A, class VI|ATPase 11A, p type|ATPase, class 1, member h|P4-ATPase flippase complex alpha subunit ATP11A', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=12757016, end_position_on_the_genomic_accession=12868728, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  50771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50771, CurrentID=0, Status='live', Symbol='Atp9b', Aliases='AA934181, Atpc2b, IIb, MMR', description='ATPase, class II, type 9B', other_designations='ATPase 9B, class II|ATPase 9B, p type|ATPase, class 2, member b|macrophage MHC receptor', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80734141, end_position_on_the_genomic_accession=80934077, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  50772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50772, CurrentID=0, Status='live', Symbol='Mapk6', Aliases='2610021I23Rik, D130053K17Rik, Erk3, Mapk43, Prkm4, Prkm6, Mapk6', description='mitogen-activated protein kinase 6', other_designations='Erk-3 related|MAP kinase 6|extracellular signal-regulated kinase 3|mitogen activated protein kinase 4|mitogen activated protein kinase 6|protein kinase, mitogen activated kinase 4', map_location='9 42.3 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75386782, end_position_on_the_genomic_accession=75410018, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  50773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50773, CurrentID=0, Status='live', Symbol='Nt5c', Aliases='Dnt, Dnt1, Umph-2, Umph2', description="5',3'-nucleotidase, cytosolic", other_designations="5' nucleotidase, deoxy (pyrimidine), cytosolic type C|cytosolic 5',3'-pyrimidine nucleotidase|dNT-1|deoxy-5'-nucleotidase 1|uridine monophosphatase 2", map_location='11 E', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115490425, end_position_on_the_genomic_accession=115491876, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  50775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50775, CurrentID=0, Status='live', Symbol='Krtap5-4', Aliases='', description='keratin associated protein 5-4', other_designations='UHS-Ser-2|krtap5-4|ultra high sulfur serine protein 2', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142303502, end_position_on_the_genomic_accession=142304503, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50776, CurrentID=0, Status='live', Symbol='Polg2', Aliases='', description='polymerase (DNA directed), gamma 2, accessory subunit', other_designations='DNA polymerase gamma accessory 55 kDa subunit|mitochondrial DNA polymerase accessory subunit|mitochondrial polymerase gamma 2|mtPolB|p55|polG-beta|polymerase, gamma 2', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106754821, end_position_on_the_genomic_accession=106780010, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  50778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50778, CurrentID=0, Status='live', Symbol='Rgs1', Aliases='BL34', description='regulator of G-protein signaling 1', other_designations='', map_location='1 62.56 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=144244669, end_position_on_the_genomic_accession=144249104, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  50779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50779, CurrentID=0, Status='live', Symbol='Rgs6', Aliases='', description='regulator of G-protein signaling 6', other_designations='', map_location='12 38.14 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=82616176, end_position_on_the_genomic_accession=83158815, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  50780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50780, CurrentID=0, Status='live', Symbol='Rgs3', Aliases='4930506N09Rik, C2PA-RGS3, C2pa, PDZ-RGS3, RGS3S', description='regulator of G-protein signaling 3', other_designations='C2 membrane binding, PDZ protein/protein interaction, ATP/GTP binding domains', map_location='4 33.19 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62559838, end_position_on_the_genomic_accession=62703022, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  50781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50781, CurrentID=0, Status='live', Symbol='Dkk3', Aliases='AW061014, C87148', description='dickkopf WNT signaling pathway inhibitor 3', other_designations='dickkopf homolog 3|dickkopf-3|dkk-3|mDkk-3', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=112116019, end_position_on_the_genomic_accession=112159121, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  50782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50782, CurrentID=0, Status='live', Symbol='Rgs11', Aliases='C78048', description='regulator of G-protein signaling 11', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26202962, end_position_on_the_genomic_accession=26211324, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  50783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50783, CurrentID=0, Status='live', Symbol='Lsm4', Aliases='', description='LSM4 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM4 homolog, U6 small nuclear RNA associated|U6 snRNA-associated SM-like protein 4', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70673231, end_position_on_the_genomic_accession=70678752, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50784, CurrentID=0, Status='live', Symbol='Plpp2', Aliases='Lpp2, PAP2-G, PAP2-gamma, PAP2c, Ppap2c', description='phospholipid phosphatase 2', other_designations='lipid phosphate phosphohydrolase 2|phosphatidate phosphohydrolase type 2c|phosphatidic acid phosphatase 2c|phosphatidic acid phosphatase type 2C', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79526424, end_position_on_the_genomic_accession=79533787, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  50785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50785, CurrentID=0, Status='live', Symbol='Hs6st1', Aliases='6Ost1', description='heparan sulfate 6-O-sulfotransferase 1', other_designations='HS6ST-1|mHS6ST-1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36068400, end_position_on_the_genomic_accession=36106446, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50786, CurrentID=0, Status='live', Symbol='Hs6st2', Aliases='6OST2', description='heparan sulfate 6-O-sulfotransferase 2', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=51386637, end_position_on_the_genomic_accession=51681760, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  50787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50787, CurrentID=0, Status='live', Symbol='Hs6st3', Aliases='6OST3, HS6ST-3', description='heparan sulfate 6-O-sulfotransferase 3', other_designations='', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=119137660, end_position_on_the_genomic_accession=119872871, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50788, CurrentID=0, Status='live', Symbol='Fbxl8', Aliases='FBL8', description='F-box and leucine-rich repeat protein 8', other_designations='F-box protein FBL8', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105264648, end_position_on_the_genomic_accession=105269326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  50789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50789, CurrentID=0, Status='live', Symbol='Fbxl3', Aliases='AU041772, AW212966, FBK, Fbl3aa, Ovtm, Play68, Fbxl3', description='F-box and leucine-rich repeat protein 3', other_designations='F-box and leucine-rich repeat protein 3a|F-box/LRR-repeat protein 3A|protein after-hours|protein overtime', map_location='14 E2.2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=103079730, end_position_on_the_genomic_accession=103099944, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  50790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50790, CurrentID=0, Status='live', Symbol='Acsl4', Aliases='9430020A05Rik, ACS4, AU018108, Facl4, Lacs4', description='acyl-CoA synthetase long-chain family member 4', other_designations='LACS 4|fatty acid-Coenzyme A ligase, long chain 4|long-chain acyl-CoA synthetase 4|mACS4', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=142317993, end_position_on_the_genomic_accession=142390798, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  50791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50791, CurrentID=0, Status='live', Symbol='Magi2', Aliases='AIP-1, Acvri1, Acvrinp1, Acvrip1, Magi-2, S-SCAM, mKIAA0705', description='membrane associated guanylate kinase, WW and PDZ domain containing 2', other_designations='activin receptor interacting protein 1|activin receptor-interacting protein 1|atrophin-1-interacting protein 1|membrane-associated guanylate kinase inverted 2', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=19223283, end_position_on_the_genomic_accession=20704792, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  50793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50793, CurrentID=0, Status='live', Symbol='Orc3', Aliases='Orc3l', description='origin recognition complex, subunit 3', other_designations='origin of replication 3 homolog|origin recognition complex subunit Latheo', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34566781, end_position_on_the_genomic_accession=34614942, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  50794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50794, CurrentID=0, Status='live', Symbol='Klf13', Aliases='0610043C13Rik, 9430029L20Rik, Bteb3, FKLF-2, FKLF2, NSLP1, RFLAT-1, RFLAT1', description='Kruppel-like factor 13', other_designations='BTE-binding protein 3|C2-H2 zinc finger protein|RANTES factor of late activated T-lymphocytes 1|basic transcription element binding protein 3|basic transcription element-binding protein 3|erythroid transcription factor FKLF-2|transcription factor BTEB3', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=63886351, end_position_on_the_genomic_accession=63938915, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  50795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50795, CurrentID=0, Status='live', Symbol='Sh3bgr', Aliases='5430437A18Rik', description='SH3-binding domain glutamic acid-rich protein', other_designations='', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96200470, end_position_on_the_genomic_accession=96228937, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  50796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50796, CurrentID=0, Status='live', Symbol='Dmrt1', Aliases='', description='doublesex and mab-3 related transcription factor 1', other_designations='', map_location='19 C2-C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=25505618, end_position_on_the_genomic_accession=25604328, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  50797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50797, CurrentID=0, Status='live', Symbol='Copb2', Aliases='AI256832', description='coatomer protein complex, subunit beta 2 (beta prime)', other_designations="beta'-COP|beta'-coat protein|p102", map_location='9 E3-F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98563731, end_position_on_the_genomic_accession=98588375, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  50798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50798, CurrentID=0, Status='live', Symbol='Gne', Aliases='2310066H07Rik, DMRV, GLCNE, IBM2, NM, Uae1', description='glucosamine (UDP-N-acetyl)-2-epimerase/N-acetylmannosamine kinase', other_designations='UDP-GlcNAc-2-epimerase/ManAc kinase|UDP-N-acetylglucosamine 2-epimerase/N-acetylmannosamine kinase|UDP-N-acetylglucosamine-2-epimerase/N-acetylmannosamine kinase', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44034075, end_position_on_the_genomic_accession=44084177, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  50799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50799, CurrentID=0, Status='live', Symbol='Slc25a13', Aliases='AI785475, Ctrn', description='solute carrier family 25 (mitochondrial carrier, adenine nucleotide translocator), member 13', other_designations='citrin|mitochondrial aspartate glutamate carrier 2|solute carrier family 25 member 13|solute carrier family 25, member 13', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=6041218, end_position_on_the_genomic_accession=6217233, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  50800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50800, CurrentID=0, Status='live', Symbol='Nr1h2-ps', Aliases='', description='nuclear receptor subfamily 1, group H, member 2, pseudogene', other_designations='LXR-beta pseudogene|nuclear oxysterol receptor|psiLXR beta', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=74364601, end_position_on_the_genomic_accession=74365017, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  50817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50817, CurrentID=0, Status='live', Symbol='Capn15', Aliases='Solh', description='calpain 15', other_designations='small optic lobes homolog', map_location='17 A3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25958364, end_position_on_the_genomic_accession=25985791, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  50849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50849, CurrentID=0, Status='live', Symbol='Rnf10', Aliases='AA675014, RIE2, mKIAA0262', description='ring finger protein 10', other_designations='sid 2705', map_location='5 56.01 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115241770, end_position_on_the_genomic_accession=115272895, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  50850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50850, CurrentID=0, Status='live', Symbol='Spast', Aliases='Spg4, mKIAA1083', description='spastin', other_designations='spastic paraplegia 4 homolog', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74337454, end_position_on_the_genomic_accession=74391115, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  50868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50868, CurrentID=0, Status='live', Symbol='Keap1', Aliases='INRF2, mKIAA0132', description='kelch-like ECH-associated protein 1', other_designations='NRF2 cytosolic inhibitor|cytosolic inhibitor of Nrf2|ring canal protein', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21229730, end_position_on_the_genomic_accession=21239332, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  50873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50873, CurrentID=0, Status='live', Symbol='Park2', Aliases='PRKN', description='Parkinson disease (autosomal recessive, juvenile) 2, parkin', other_designations='parkin protein', map_location='17 7.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=10840384, end_position_on_the_genomic_accession=12063370, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  50874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50874, CurrentID=0, Status='live', Symbol='Tmod4', Aliases='MTMOD, Sk-Tmod', description='tropomodulin 4', other_designations='skeletal muscle tropomodulin|skeletal tropomodulin', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95124117, end_position_on_the_genomic_accession=95129208, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  50875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50875, CurrentID=0, Status='live', Symbol='Tmod3', Aliases='U-Tmod, UTMOD', description='tropomodulin 3', other_designations='tropomodulin 3; ubiquitous tropomodulin|ubiquitous tropomodulin', map_location='9 42.3 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75497784, end_position_on_the_genomic_accession=75559657, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  50876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50876, CurrentID=0, Status='live', Symbol='Tmod2', Aliases='N-Tmod, NTMOD', description='tropomodulin 2', other_designations='neural tropomodulin|neuronal tropomodulin', map_location='9 42.31 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75565621, end_position_on_the_genomic_accession=75611348, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  50877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50877, CurrentID=0, Status='live', Symbol='Neu3', Aliases='', description='neuraminidase 3', other_designations='N-acetyl-alpha-neuraminidase 3|Sialidase 3 (Membrane sialidase) (Ganglioside sialidase) (N-acetyl-alpha-neuraminidase 3)|ganglioside sialidase|membrane sialidase', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99811439, end_position_on_the_genomic_accession=99828525, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  50878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50878, CurrentID=0, Status='live', Symbol='Stag3', Aliases='SA-2', description='stromal antigen 3', other_designations='SCC3 homolog 3|stromalin 3|stromalin-3', map_location='5 77.01 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138272589, end_position_on_the_genomic_accession=138312397, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  50880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50880, CurrentID=0, Status='live', Symbol='Scly', Aliases='9830169H08, A930015N15Rik, SCL1, Scly2, Scly', description='selenocysteine lyase', other_designations='mSCL|selenocysteine lyase; Selenocysteine reductase|selenocysteine reductase', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91298295, end_position_on_the_genomic_accession=91321080, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  50883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50883, CurrentID=0, Status='live', Symbol='Chek2', Aliases='CHK2, Cds1, HUCDS1, Rad53', description='checkpoint kinase 2', other_designations='CHK2 checkpoint homolog|Cds1 homolog|Rad53 homolog|protein kinase Chk2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110839791, end_position_on_the_genomic_accession=110874140, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  50884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50884, CurrentID=0, Status='live', Symbol='Nckap1', Aliases='C79304, H19, Hem-2, Hem2, Nap1, mKIAA0587, mh19, p125Nap1', description='NCK-associated protein 1', other_designations='brain protein H19|membrane-associated protein HEM-2', map_location='2 48.21 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=80500512, end_position_on_the_genomic_accession=80581259, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  50887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50887, CurrentID=0, Status='live', Symbol='Hmgn5', Aliases='GARP45, NBP-45, Nsbp1', description='high-mobility group nucleosome binding domain 5', other_designations='nucleosome binding protein 1|nucleosome binding protein 45', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=109004537, end_position_on_the_genomic_accession=109013380, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  50905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50905, CurrentID=0, Status='live', Symbol='Il17rb', Aliases='Evi27, IL-17ER, IL-17Rh1, IL17RH1, Il17br', description='interleukin 17 receptor B', other_designations='IL-17 receptor B|IL-17 receptor homolog 1|IL-17B receptor|IL-17RB|ecotropic viral integration site 27|interleukin 17B receptor|interleukin-17B receptor', map_location='14 18.39 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=29995506, end_position_on_the_genomic_accession=30008896, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  50907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50907, CurrentID=0, Status='live', Symbol='Preb', Aliases='C85705', description='prolactin regulatory element binding', other_designations='mammalian guanine nucleotide exchange factor mSec12', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30951667, end_position_on_the_genomic_accession=30960386, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  50908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50908, CurrentID=0, Status='live', Symbol='C1s1', Aliases='AA959438, AI255193, AI327365, C1s, C1sa', description='complement component 1, s subcomponent 1', other_designations='C1 esterase|complement component 1 subcomponent s-A|complement component C1SA|serine protease', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124530344, end_position_on_the_genomic_accession=124542390, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  50909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50909, CurrentID=0, Status='live', Symbol='C1ra', Aliases='AI132558, C1r, mC1rA', description='complement component 1, r subcomponent A', other_designations='complement component 1 subcomponent r-A|complement component 1, r-A subcomponent', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124512621, end_position_on_the_genomic_accession=124523440, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  50911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50911, CurrentID=0, Status='live', Symbol='Exosc9', Aliases='PM/Scl-75, Pmscl1, RRP45, p5, p6', description='exosome component 9', other_designations='P75 polymyositis-scleroderma overlap syndrome-associated autoantigen|autoantigen PM/Scl 1|exosome complex exonuclease RRP45|polymyositis/scleroderma autoantigen 1|polymyositis/scleroderma autoantigen 75 kDa', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36552606, end_position_on_the_genomic_accession=36565727, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  50912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50912, CurrentID=0, Status='live', Symbol='Exosc10', Aliases='PM-Scl, PM/Scl-100, Pmscl2, RRP6, p2, p3, p4', description='exosome component 10', other_designations='autoantigen PM/Scl 2 homolog|polymyositis/scleroderma autoantigen 2 homolog', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148558424, end_position_on_the_genomic_accession=148582401, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  50913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50913, CurrentID=0, Status='live', Symbol='Olig2', Aliases='AI604895, Bhlhb1, Olg-2, Oligo2, RK17, bHLHe19', description='oligodendrocyte transcription factor 2', other_designations='', map_location='16 52.6 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91225550, end_position_on_the_genomic_accession=91228677, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50914, CurrentID=0, Status='live', Symbol='Olig1', Aliases='AI836478, AW494459, Bhlhb6, Olg-1, Oligo1, bHLHe21', description='oligodendrocyte transcription factor 1', other_designations='oligodendrocyte-specific bHLH transcription factor 1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91269769, end_position_on_the_genomic_accession=91271939, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  50915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50915, CurrentID=0, Status='live', Symbol='Grb14', Aliases='AI505286', description='growth factor receptor bound protein 14', other_designations='GRB14 adapter protein', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=64912476, end_position_on_the_genomic_accession=65022766, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  50916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50916, CurrentID=0, Status='live', Symbol='Irx4', Aliases='', description='Iroquois related homeobox 4 (Drosophila)', other_designations='homeodomain protein IRXA3|iroquois homeobox protein 4', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73259590, end_position_on_the_genomic_accession=73269622, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  50917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50917, CurrentID=0, Status='live', Symbol='Galns', Aliases='mFLJ00319', description='galactosamine (N-acetyl)-6-sulfate sulfatase', other_designations='GalNAc6S sulfatase|N-acetylgalactosamine-6-sulfate sulfatase|chondroitinase|chondroitinsulfatase|galactose-6-sulfate sulfatase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122578237, end_position_on_the_genomic_accession=122611503, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  50918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50918, CurrentID=0, Status='live', Symbol='Myadm', Aliases='D7Wsu62e', description='myeloid-associated differentiation marker', other_designations='myeloid up-regulated protein', map_location='7 1.92 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3289038, end_position_on_the_genomic_accession=3299355, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50926, CurrentID=0, Status='live', Symbol='Hnrnpdl', Aliases='AA407431, AA959857, D5Ertd650e, D5Wsu145e1, Hnrpdl, JKTBP, hnRNP DL, hnRNP-DL, Hnrnpdl', description='heterogeneous nuclear ribonucleoprotein D-like', other_designations='JKT41-binding protein|hnRNP D-like|hnRPD-like protein', map_location='5 48.46 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100033577, end_position_on_the_genomic_accession=100039950, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  50927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50927, CurrentID=0, Status='live', Symbol='Nasp', Aliases='5033430J04Rik, AI131596, AI317140, D4Ertd767e, Epcs32-T, Nasp', description='nuclear autoantigenic sperm protein (histone-binding)', other_designations='nasp|somatic histone binding protein NASP', map_location='4 53.24 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116601052, end_position_on_the_genomic_accession=116627951, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  50928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50928, CurrentID=0, Status='live', Symbol='Klrg1', Aliases='2F1-Ag, MAFA, MAFA-L', description='killer cell lectin-like receptor subfamily G, member 1', other_designations='mast cell function-associated antigen 2F1', map_location='6 57.52 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122270596, end_position_on_the_genomic_accession=122282833, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  50929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50929, CurrentID=0, Status='live', Symbol='Il22', Aliases='IL-22, IL-22a, ILTIFa, Iltif', description='interleukin 22', other_designations='IL-10-related T-cell-derived-inducible factor|IL-TIF alpha|ILTIF alpha|interleukin 10-related T cell-derived inducible factor|interleukin-22a', map_location='10 66.67 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=118204586, end_position_on_the_genomic_accession=118210047, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  50930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50930, CurrentID=0, Status='live', Symbol='Tnfsf14', Aliases='HVEM-L, HVEML, LIGHT, LTg, Ly113, Tnlg1d', description='tumor necrosis factor (ligand) superfamily, member 14', other_designations='tumor necrosis factor ligand 1d', map_location='17 D-E1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57189474, end_position_on_the_genomic_accession=57194181, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  50931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50931, CurrentID=0, Status='live', Symbol='Il27ra', Aliases='CRL1, IL-27R, Tccr, WSX-1, Wsx1, zcytor1', description='interleukin 27 receptor, alpha', other_designations='IL-27 receptor subunit alpha|IL-27R subunit alpha|IL-27R-alpha|IL-27RA|T cell cytokine receptor|T cell cytokine receptor; cytokine receptor family, class 1 (WSXWS), member 1|cytokine receptor family, class 1 (WSXWS), member 1|type I T-cell cytokine receptor', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84030286, end_position_on_the_genomic_accession=84042629, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  50932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50932, CurrentID=0, Status='live', Symbol='Mink1', Aliases='B55, MINK, Map4k6, Ysk2', description='misshapen-like kinase 1 (zebrafish)', other_designations='GCK family kinase MiNK|MAPK/ERK kinase kinase kinase 6|MEK kinase kinase 6|MEKKK 6|Misshapen/NIKs-related kinase|Yeast Sps1/Ste20-related kinase 2|mink1|misshapen/NIK-related kinase', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70562594, end_position_on_the_genomic_accession=70614482, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  50933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50933, CurrentID=0, Status='live', Symbol='Uchl3', Aliases='', description='ubiquitin carboxyl-terminal esterase L3 (ubiquitin thiolesterase)', other_designations='UCH-L3|ubiquitin thioesterase L3|ubiquitin thiolesterase', map_location='14 50.9 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=101653880, end_position_on_the_genomic_accession=101696125, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  50934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50934, CurrentID=0, Status='live', Symbol='Slc7a8', Aliases='AA408822, LAT2', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 8', other_designations='L-type amino acid transporter 2|mLAT2|solute carrier family 7 member 8|solute carrier family 8 (cationic amino acid transporter, y+ system), member 7', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54722215, end_position_on_the_genomic_accession=54781886, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  50935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50935, CurrentID=0, Status='live', Symbol='St6galnac6', Aliases='ST6GalNAcVI, Siat7f', description='ST6 (alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1,3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 6', other_designations='(alpha-N-acetyl-neuraminyl-2,3-beta-galactosyl-1, 3)-N-acetylgalactosaminide alpha-2,6-sialyltransferase 6|galNAc alpha-2,6-sialyltransferase VI|sialytransferase 7 ((alpha-N-acetylneuraminyl 2,3-betagalactosyl-1,3)-N-acetyl galactosaminide alpha-2,6-sialytransferase) F', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32599709, end_position_on_the_genomic_accession=32620809, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  50994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50994, CurrentID=0, Status='live', Symbol='Mtag2', Aliases='MAG', description='metastasis associated gene 2', other_designations='metastasis associated 2|metastasis associated gene product', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45366163, end_position_on_the_genomic_accession=45367948, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  50995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50995, CurrentID=0, Status='live', Symbol='Uba2', Aliases='AA986091, Arx, Sae2, UBA1, Ubl1a2, Uble1b', description='ubiquitin-like modifier activating enzyme 2', other_designations='SUMO1 activating enzyme subunit 2|Sumo-1 activating enzyme subunit 2|anthracycline-associated resistance ARX|ubiquitin-like 1 (sentrin) activating enzyme E1B|ubiquitin-like 1 (sentrin) activating enzyme subunit 2|ubiquitin-like 1-activating enzyme E1B|ubiquitin-like modifier-activating enzyme 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34140697, end_position_on_the_genomic_accession=34168529, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  50996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50996, CurrentID=0, Status='live', Symbol='Pdcd7', Aliases='C80112, ES18', description='programmed cell death 7', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65346068, end_position_on_the_genomic_accession=65359643, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  50997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=50997, CurrentID=0, Status='live', Symbol='Mpp2', Aliases='D11Bwg0652e, Dlg2, Dlgh2, Pals4', description='membrane protein, palmitoylated 2 (MAGUK p55 subfamily member 2)', other_designations='discs large homolog 2|p55 subfamily member of MAGUK family', map_location='11 65.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102057016, end_position_on_the_genomic_accession=102088636, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  723976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723976, CurrentID=0, Status='live', Symbol='Mir652', Aliases='Mirn652, miRX, mmu-mir-652', description='microRNA 652', other_designations='mir 652', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=142739000, end_position_on_the_genomic_accession=142739097, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  233038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233038, CurrentID=0, Status='live', Symbol='Nccrp1', Aliases='1190020J12Rik, Gm163, NCCRP-1', description='non-specific cytotoxic cell receptor protein 1 homolog (zebrafish)', other_designations='NCC receptor protein 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28543596, end_position_on_the_genomic_accession=28547254, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  51786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51786, CurrentID=0, Status='live', Symbol='Cpsf2', Aliases='100kDa, 2610024B04Rik, AI662483, Cpsf, MCPSF, mKIAA1367', description='cleavage and polyadenylation specific factor 2', other_designations='CPSF 100 kDa subunit|cleavage and polyadenylation specific factor 2, 100kD subunit|cleavage and polyadenylation specificity factor 100 kDa subunit', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=101975765, end_position_on_the_genomic_accession=102006430, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  51788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51788, CurrentID=0, Status='live', Symbol='H2afz', Aliases='H2A.Z, H2a.z-1', description='H2A histone family, member Z', other_designations='H2A/z', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=137864487, end_position_on_the_genomic_accession=137866922, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  51789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51789, CurrentID=0, Status='live', Symbol='Tnk2', Aliases='Ack, Ack-1, Ack1, Cdgip, Pyk1', description='tyrosine kinase, non-receptor, 2', other_designations='Cdc42 GTPase-inhibiting protein|P21cdc42Hs kinase|non-receptor protein tyrosine kinase ACK|tyrosine kinase non-receptor protein 2', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32644099, end_position_on_the_genomic_accession=32683493, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  51791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51791, CurrentID=0, Status='live', Symbol='Rgs14', Aliases='RPIP1', description='regulator of G-protein signaling 14', other_designations='RAP1/RAP2-interacting protein|rap1/rap2 interacting protein', map_location='13 29.8 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55369712, end_position_on_the_genomic_accession=55384691, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  51792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51792, CurrentID=0, Status='live', Symbol='Ppp2r1a', Aliases='6330556D22Rik, PP2A, PR65', description='protein phosphatase 2, regulatory subunit A, alpha', other_designations='PP2A subunit A isoform PR65-alpha|PP2A subunit A isoform R1-alpha|protein phosphatase 2 (formerly 2A), regulatory subunit A (PR 65), alpha isoform|protein phosphatase PP2A|serine/threonine protein phosphatase A subunit type 2A', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20945454, end_position_on_the_genomic_accession=20965905, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  51793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51793, CurrentID=0, Status='live', Symbol='Ddah2', Aliases='1110003M04Rik, AU019324, AW413173, DDAH-2, DDAHII, Ddah, G6a, NG30', description='dimethylarginine dimethylaminohydrolase 2', other_designations='Clone 7u|dimethylargininase-2', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35059035, end_position_on_the_genomic_accession=35062099, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  51795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51795, CurrentID=0, Status='live', Symbol='Srpx', Aliases='drs, drs-1, drs-2, mDRS-1, mDRS-2', description='sushi-repeat-containing protein', other_designations='sushi-repeat-containing protein, X chromosome|sushi-repeat-containing protein, X-chromosome', map_location='X A1.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=10037977, end_position_on_the_genomic_accession=10117661, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  51796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51796, CurrentID=0, Status='live', Symbol='Srrm1', Aliases='AA407769, POP101, Srm160', description='serine/arginine repetitive matrix 1', other_designations='Ser/Arg-related nuclear matrix protein|plenty-of-prolines 101|plenty-of-prolines-101', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135320484, end_position_on_the_genomic_accession=135353254, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  51797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51797, CurrentID=0, Status='live', Symbol='Ctps', Aliases='Ctps1', description="cytidine 5'-triphosphate synthase", other_designations='CTP synthetase 1|UTP--ammonia ligase 1', map_location='4 56.52 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120539866, end_position_on_the_genomic_accession=120570276, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  51798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51798, CurrentID=0, Status='live', Symbol='Ech1', Aliases='AA617331', description='enoyl coenzyme A hydratase 1, peroxisomal', other_designations='dienoyl-CoA isomerase|peroxisomal/mitochondrial dienoyl-CoA isomerase', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28824804, end_position_on_the_genomic_accession=28832248, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  51799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51799, CurrentID=0, Status='live', Symbol='Rundc3a', Aliases='RPIP-8, Rap2ip, Rpip8', description='RUN domain containing 3A', other_designations='Rap2 interacting protein 8|rap2-interacting protein 8', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102393403, end_position_on_the_genomic_accession=102402939, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  51800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51800, CurrentID=0, Status='live', Symbol='Bok', Aliases='matador, mtd', description='BCL2-related ovarian killer', other_designations='apoptosis activator Mtd', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93685694, end_position_on_the_genomic_accession=93695762, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  51801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51801, CurrentID=0, Status='live', Symbol='Ramp1', Aliases='9130218E19Rik', description='receptor (calcitonin) activity modifying protein 1', other_designations='', map_location='1 46.08 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91179822, end_position_on_the_genomic_accession=91225196, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  51810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51810, CurrentID=0, Status='live', Symbol='Hnrnpu', Aliases='AA408410, AI256620, AL024194, AL024437, AW557595, C86794, Hnrpu, SAFA, Sp120, hnRNP U', description='heterogeneous nuclear ribonucleoprotein U', other_designations='SAF-A|heterogenous nuclear ribonucleoprotein U|nuclear matrix protein sp120|scaffold attachment factor A', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=178321108, end_position_on_the_genomic_accession=178337810, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  51811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51811, CurrentID=0, Status='live', Symbol='Clec4f', Aliases='Clecsf13, D18063, KUCR_MOUSE, Kclr', description='C-type lectin domain family 4, member f', other_designations='C-type (calcium dependent, carbohydrate recognition domain) lectin, superfamily member 13|C-type lectin 13|C-type lectin superfamily member 13|Kupffer cell c-type lectin receptor|kupffer cell receptor', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83644542, end_position_on_the_genomic_accession=83656116, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51812, CurrentID=0, Status='live', Symbol='Mcrs1', Aliases='C78274, ICP22BP, MSP58, P78', description='microspherule protein 1', other_designations='58 kDa microspherule protein|nucleolar protein', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99242817, end_position_on_the_genomic_accession=99251961, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  51813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51813, CurrentID=0, Status='live', Symbol='Ccnc', Aliases='AI451004, AU020987, CG1C', description='cyclin C', other_designations='dJ199J3.4 (cyclin C)', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=21727701, end_position_on_the_genomic_accession=21759922, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  51818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51818, CurrentID=0, Status='live', Symbol='D17Leh66E', Aliases='', description='DNA segment, Chr 17, Lehrach 66', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  51869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51869, CurrentID=0, Status='live', Symbol='Rif1', Aliases='5730435J01Rik, 6530403D07Rik, AU016181, AW549474, D2Ertd145e', description='Rap1 interacting factor 1 homolog (yeast)', other_designations='mRif1|rap1-interacting factor 1 homolog', map_location='2 29.98 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=52072811, end_position_on_the_genomic_accession=52126769, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  51875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51875, CurrentID=0, Status='live', Symbol='Tmem141', Aliases='1110065P14Rik, AI663999, D2Ertd217e', description='transmembrane protein 141', other_designations='', map_location='2 17.57 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25620066, end_position_on_the_genomic_accession=25622039, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  51885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51885, CurrentID=0, Status='live', Symbol='Tubgcp4', Aliases='4932441P04Rik, 76p, D2Ertd435e, GCP-4', description='tubulin, gamma complex associated protein 4', other_designations='gamma tubulin ring complex protein', map_location='2 60.37 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121170454, end_position_on_the_genomic_accession=121198770, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  51886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51886, CurrentID=0, Status='live', Symbol='Fubp1', Aliases='9530027K12Rik, D3Ertd330e, FBP, Fubp, Fubp4', description='far upstream element (FUSE) binding protein 1', other_designations='FUSE-binding protein 1|far upstream element (FUSE) binding protein 4', map_location='3 77.05 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152210427, end_position_on_the_genomic_accession=152236830, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  51897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51897, CurrentID=0, Status='live', Symbol='Atg13', Aliases='1110053A20Rik, D2Ertd391e, Harbi1', description='autophagy related 13', other_designations='ATG13 autophagy related 13 homolog', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91674612, end_position_on_the_genomic_accession=91710598, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  51902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51902, CurrentID=0, Status='live', Symbol='Rnf24', Aliases='2810473M14Rik, 4930505A13Rik, AI317164, C86507, D2Ertd504e', description='ring finger protein 24', other_designations='', map_location='2 63.32 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131298488, end_position_on_the_genomic_accession=131352892, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  51938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51938, CurrentID=0, Status='live', Symbol='Ccdc39', Aliases='4921507O14Rik, AI447426, AI844750, D3Ertd789e, b2b1304Clo, b2b1735Clo, b2b2025.1Clo', description='coiled-coil domain containing 39', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=33810757, end_position_on_the_genomic_accession=33844322, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  51944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51944, CurrentID=0, Status='live', Symbol='Knstrn', Aliases='1700025D04Rik, C15orf23, D2Ertd750e, SKAP, Traf4af1', description='kinetochore-localized astrin/SPAG5 binding', other_designations='TRAF4 associated factor 1 homolog|TRAF4-associated factor 1|kinastrin|kinetochore-localized astrin-binding protein|kinetochore-localized astrin/SPAG5-binding protein|putative TRAF4-associated factor 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118813990, end_position_on_the_genomic_accession=118837210, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  51960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=51960, CurrentID=0, Status='live', Symbol='Kctd18', Aliases='4932411A20Rik, 6530404F10Rik', description='potassium channel tetramerisation domain containing 18', other_designations='', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=57955101, end_position_on_the_genomic_accession=57970084, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  52004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52004, CurrentID=0, Status='live', Symbol='Cdk2ap2', Aliases='5830466O21Rik, D19Ertd144e, Doc-1r', description='CDK2-associated protein 2', other_designations='DOC-1 related protein|DOC-1-related protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4096843, end_position_on_the_genomic_accession=4099019, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  52009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52009, CurrentID=0, Status='live', Symbol='Hn1l', Aliases='2810430B18Rik, AL022629, C16orf34, D17Ertd441e', description='hematological and neurological expressed 1-like', other_designations='HN1-like protein|hematopoietic- and neurologic-expressed sequence 1-like', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24942470, end_position_on_the_genomic_accession=24960623, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  52013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52013, CurrentID=0, Status='live', Symbol='R3hcc1l', Aliases='1700036B12Rik, AI450607, C10orf28, D19Ertd386e, GIDRP86, Gidrp88', description='R3H domain and coiled-coil containing 1 like', other_designations='growth inhibition and differentiation related protein 86|growth inhibition and differentiation-related protein 88 homolog', map_location='19 36.47 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42518312, end_position_on_the_genomic_accession=42592256, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  52014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52014, CurrentID=0, Status='live', Symbol='Nus1', Aliases='1600027K07Rik, AU019165, AW538011, BC003223, D10Ertd438e', description='nuclear undecaprenyl pyrophosphate synthase 1 homolog (S. cerevisiae)', other_designations='Nogo-B receptor|di-trans,poly-cis-decaprenylcistransferase|ngBR', map_location='10 26.64 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52417547, end_position_on_the_genomic_accession=52440192, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  52020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52020, CurrentID=0, Status='live', Symbol='Umodl1', Aliases='D17Ertd488e', description='uromodulin-like 1', other_designations='olfactorin', map_location='17 15.8 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=30954640, end_position_on_the_genomic_accession=31010710, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  52023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52023, CurrentID=0, Status='live', Symbol='Pibf1', Aliases='', description='progesterone immunomodulatory binding factor 1', other_designations='progesterone-induced blocking factor 1', map_location='14 49.27 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=99099424, end_position_on_the_genomic_accession=99254495, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  52024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52024, CurrentID=0, Status='live', Symbol='Ankrd22', Aliases='5430429D21Rik, D19Ertd675e', description='ankyrin repeat domain 22', other_designations='', map_location='19 29.33 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34122549, end_position_on_the_genomic_accession=34166041, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  52028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52028, CurrentID=0, Status='live', Symbol='Bbs1', Aliases='AI451249, D19Ertd609e', description='Bardet-Biedl syndrome 1 (human)', other_designations='Bardet-Biedl syndrome 1 homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4886878, end_position_on_the_genomic_accession=4906627, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  100191037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100191037, CurrentID=0, Status='live', Symbol='Krtap10-4', Aliases='675238, A030009A09Rik, AA589524', description='keratin associated protein 10-4', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77826169, end_position_on_the_genomic_accession=77827070, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  52033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52033, CurrentID=0, Status='live', Symbol='Pbk', Aliases='2810434B10Rik, AW538537, D14Ertd732e, TOPK', description='PDZ binding kinase', other_designations='MAPKK-like protein kinase|PDZ-binding kinase|T-LAK cell-originated protein kinase', map_location='14 34.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65805911, end_position_on_the_genomic_accession=65817822, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  52036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52036, CurrentID=0, Status='live', Symbol='Ppp6r3', Aliases='4930528G08Rik, 9130026N02Rik, D19Bwg1430e, D19Ertd703e, Pp6r3, Ppcs3, Pptcs3, Sapl, Saps3, mKIAA1558', description='protein phosphatase 6, regulatory subunit 3', other_designations='SAPS domain family, member 3|pluripotent cell-specific 3|sporulation-induced transcript 4-associated protein', map_location='19 3.21 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3454928, end_position_on_the_genomic_accession=3575749, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  52040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52040, CurrentID=0, Status='live', Symbol='Ppp1r10', Aliases='2610025H06Rik, Cat53, D17Ertd808e, Fb19, Pnuts', description='protein phosphatase 1, regulatory subunit 10', other_designations='HLA-C adjacent transcript 53|MHC class I region proline-rich protein CAT53|protein phosphatase I', map_location='17 18.77 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35916512, end_position_on_the_genomic_accession=35952996, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  52055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52055, CurrentID=0, Status='live', Symbol='Rab11fip5', Aliases='9130206P09Rik, C75969, D6Ertd32e, GAF1, RIP11', description='RAB11 family interacting protein 5 (class I)', other_designations='rab11-FIP5|rab11-interacting protein Rip11', map_location='6 37.35 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85334962, end_position_on_the_genomic_accession=85374648, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  52064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52064, CurrentID=0, Status='live', Symbol='Coq5', Aliases='1810014G04Rik, D5Ertd33e', description='coenzyme Q5 methyltransferase', other_designations='coenzyme Q5 homolog, methyltransferase|ubiquinone biosynthesis methyltransferase COQ5, mitochondrial', map_location='5 56.03 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115279666, end_position_on_the_genomic_accession=115296974, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  52065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52065, CurrentID=0, Status='live', Symbol='Mfhas1', Aliases='2310066G09Rik, D8Ertd91e', description='malignant fibrous histiocytoma amplified sequence 1', other_designations='', map_location='8 21.16 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=35587777, end_position_on_the_genomic_accession=35679449, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  52076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52076, CurrentID=0, Status='live', Symbol='Tmem38b', Aliases='1600017F22Rik, AA437809, AV051057, D4Ertd89e, TRIC-B, mg33b', description='transmembrane protein 38B', other_designations='Mitsugumin33B|TRICB|mitsugumin-33B', map_location='4 28.76 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=53826045, end_position_on_the_genomic_accession=53862019, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  52118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52118, CurrentID=0, Status='live', Symbol='Pvr', Aliases='3830421F03Rik, CD155, D7Ertd458e, HVED, PVS, Taa1, Tage4, mE4, necl-5', description='poliovirus receptor', other_designations='tumor-associated antigen 1', map_location='7 9.95 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19903578, end_position_on_the_genomic_accession=19921143, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  52120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52120, CurrentID=0, Status='live', Symbol='Hgsnat', Aliases='9430010M12Rik, AW208455, D8Ertd354e, Tmem76', description='heparan-alpha-glucosaminide N-acetyltransferase', other_designations='transmembrane protein 76', map_location='8 14.22 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25944459, end_position_on_the_genomic_accession=25976744, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  52123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52123, CurrentID=0, Status='live', Symbol='Agpat5', Aliases='1110013A05Rik, D8Ertd319e', description='1-acylglycerol-3-phosphate O-acyltransferase 5 (lysophosphatidic acid acyltransferase, epsilon)', other_designations='1-AGP acyltransferase 5|1-AGPAT 5|LPAAT-epsilon|lysophosphatidic acid acyltransferase epsilon', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=18846279, end_position_on_the_genomic_accession=18884413, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  52132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52132, CurrentID=0, Status='live', Symbol='Ccdc97', Aliases='1200014H14Rik, 2810446P04Rik, AA545218, D7Ertd462e', description='coiled-coil domain containing 97', other_designations='', map_location='7 13.99 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25711112, end_position_on_the_genomic_accession=25719817, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  52150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52150, CurrentID=0, Status='live', Symbol='Kcnk6', Aliases='D7Ertd764e, Toss, Twik2', description='potassium inwardly-rectifying channel, subfamily K, member 6', other_designations='TWIK-2 two-pore-domain K+ channel', map_location='7 16.94 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29221928, end_position_on_the_genomic_accession=29232522, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  52153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52153, CurrentID=0, Status='live', Symbol='D11Jkn1e', Aliases='D11Jkn1', description='DNA segment, Chr 11, Ian Jackson-1, expressed', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  52163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52163, CurrentID=0, Status='live', Symbol='Camk1', Aliases='AI505105, CaMKIalpha, D6Ertd263e', description='calcium/calmodulin-dependent protein kinase I', other_designations='caM kinase I alpha|caM-KI|caMKI-alpha', map_location='6 52.75 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113334124, end_position_on_the_genomic_accession=113343922, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  52174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52174, CurrentID=0, Status='live', Symbol='Tmem222', Aliases='5730406H10Rik, AA409341, D4Ertd196e', description='transmembrane protein 222', other_designations='', map_location='4 66.25 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133266045, end_position_on_the_genomic_accession=133277790, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  233057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233057, CurrentID=0, Status='live', Symbol='Zfp940', Aliases='', description='zinc finger protein 940', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29843936, end_position_on_the_genomic_accession=29853748, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  52184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52184, CurrentID=0, Status='live', Symbol='Odf2l', Aliases='4733401D09Rik, 9630045K08Rik, D3Ertd250e', description='outer dense fiber of sperm tails 2-like', other_designations='', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145118574, end_position_on_the_genomic_accession=145153915, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  52186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52186, CurrentID=0, Status='live', Symbol='D8Ertd158e', Aliases='', description='DNA segment, Chr 8, ERATO Doi 158, expressed', other_designations='', map_location='8 56.08 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  52187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52187, CurrentID=0, Status='live', Symbol='Rragd', Aliases='5730543C08Rik, AI467523, C030003H22Rik, D4Ertd174e', description='Ras-related GTP binding D', other_designations='rag D|ragD|rragd', map_location='4 14.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=32982998, end_position_on_the_genomic_accession=33022180, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  52202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52202, CurrentID=0, Status='live', Symbol='Rbm34', Aliases='4930547K05Rik, 6330444B11, D8Ertd233e', description='RNA binding motif protein 34', other_designations='RNA-binding motif protein 34', map_location='8 74.64 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=126947172, end_position_on_the_genomic_accession=126971126, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  52206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52206, CurrentID=0, Status='live', Symbol='Anapc4', Aliases='2610306D21Rik, APC4, D5Ertd249e', description='anaphase promoting complex subunit 4', other_designations='cyclosome subunit 4', map_location='5 28.81 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52834126, end_position_on_the_genomic_accession=52866742, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  52231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52231, CurrentID=0, Status='live', Symbol='Ankzf1', Aliases='1300008P06Rik, 2810025E10Rik, D1Ertd161e', description='ankyrin repeat and zinc finger domain containing 1', other_designations='', map_location='1 38.62 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75192160, end_position_on_the_genomic_accession=75199387, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  52245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52245, CurrentID=0, Status='live', Symbol='Commd2', Aliases='1190017G07Rik, AI047739, D3Ertd176e, HSPC042', description='COMM domain containing 2', other_designations='', map_location='3 27.94 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=57644349, end_position_on_the_genomic_accession=57651684, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  52250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52250, CurrentID=0, Status='live', Symbol='Reep1', Aliases='D6Ertd253e', description='receptor accessory protein 1', other_designations='', map_location='6 32.2 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71707681, end_position_on_the_genomic_accession=71810710, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  52276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52276, CurrentID=0, Status='live', Symbol='Cdca8', Aliases='4831429J16Rik, AU044747, BOR, Borealin, D4Ertd421e, DasraB, MESRGP', description='cell division cycle associated 8', other_designations='MESrg|cell division cycle-associated protein 8|embryonic stem cell-related', map_location='4 57.93 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124918465, end_position_on_the_genomic_accession=124936917, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  52323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52323, CurrentID=0, Status='live', Symbol='Klhl7', Aliases='2700038B03Rik, D5Ertd363e, SBBI26', description='kelch-like 7', other_designations='BTB/POZ domain containing protein', map_location='5 10.67 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24100574, end_position_on_the_genomic_accession=24163365, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  52331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52331, CurrentID=0, Status='live', Symbol='Stbd1', Aliases='D530019K15Rik, D5Ertd593e', description='starch binding domain 1', other_designations='genethonin 1|genethonin-1|glycophagy cargo receptor stbd1', map_location='5 47.21 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92603051, end_position_on_the_genomic_accession=92606579, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  52335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52335, CurrentID=0, Status='live', Symbol='Atxn1l', Aliases='AW558552, Boat, C330011L24Rik, D6Mgi37, D8Ertd587e, ENSMUSG00000069895', description='ataxin 1-like', other_designations='brother of ataxin-1', map_location='8 57.22 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109726451, end_position_on_the_genomic_accession=109737796, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  620393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620393, CurrentID=0, Status='live', Symbol='Fam150a', Aliases='EG620393', description='family with sequence similarity 150, member A', other_designations='', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=6359331, end_position_on_the_genomic_accession=6394731, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  52348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52348, CurrentID=0, Status='live', Symbol='Vps37a', Aliases='2210018P21Rik, 4930592A21Rik, AW261445, D8Ertd531e', description='vacuolar protein sorting 37A (yeast)', other_designations='ESCRT-I complex subunit VPS37A', map_location='8 B1.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40511783, end_position_on_the_genomic_accession=40551134, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  52357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52357, CurrentID=0, Status='live', Symbol='Wwc2', Aliases='5330431C14Rik, AU022508, C85457, D8Ertd594e', description='WW, C2 and coiled-coil domain containing 2', other_designations='WW domain-containing protein 2', map_location='8 26.91 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47827595, end_position_on_the_genomic_accession=47990551, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  52377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52377, CurrentID=0, Status='live', Symbol='Rcn3', Aliases='6030455P07Rik, D530026G20Rik, D7Ertd671e, RLP49', description='reticulocalbin 3, EF-hand calcium binding domain', other_designations='', map_location='7 29.11 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45082914, end_position_on_the_genomic_accession=45092257, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  52389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52389, CurrentID=0, Status='live', Symbol='Adgra1', Aliases='2900059M17Rik, D7Ertd680e, Gpr123', description='adhesion G protein-coupled receptor A1', other_designations='G protein-coupled receptor 123|G-protein coupled receptor 123|probable G-protein coupled receptor 123', map_location='7 84.89 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139834084, end_position_on_the_genomic_accession=139878088, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  52392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52392, CurrentID=0, Status='live', Symbol='D1Ertd622e', Aliases='AI987691, AW047450', description='DNA segment, Chr 1, ERATO Doi 622, expressed', other_designations='UPF0684 protein C5orf30 homolog', map_location='1 47.76 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=97643902, end_position_on_the_genomic_accession=97662018, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  52397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52397, CurrentID=0, Status='live', Symbol='Zfp644', Aliases='1110068L01Rik, AA408882, BM-005, D5Ertd689e, Zep-2', description='zinc finger protein 644', other_designations='zinc finger motif enhancer binding protein 2', map_location='5 51.43 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=106616739, end_position_on_the_genomic_accession=106696830, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  52398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52398, CurrentID=0, Status='live', Symbol='Sept11', Aliases='6230410I01Rik, AW548875, D5Ertd606e', description='septin 11', other_designations='', map_location='5 47.29 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=93091879, end_position_on_the_genomic_accession=93179054, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  52428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52428, CurrentID=0, Status='live', Symbol='Rhpn2', Aliases='1300002E07Rik, AA536890, C85860, D7Ertd784e', description='rhophilin, Rho GTPase binding protein 2', other_designations='GTP-Rho-binding protein 2', map_location='7 21.36 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35334180, end_position_on_the_genomic_accession=35392287, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  52430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52430, CurrentID=0, Status='live', Symbol='Echdc2', Aliases='1300017C12Rik, 2610009M20Rik, D4Ertd765e', description='enoyl Coenzyme A hydratase domain containing 2', other_designations='', map_location='4 50.3 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108165436, end_position_on_the_genomic_accession=108179308, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  52432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52432, CurrentID=0, Status='live', Symbol='Ppp2r2d', Aliases='1300017E19Rik, D7Ertd753e, MDS026', description='protein phosphatase 2, regulatory subunit B, delta', other_designations='PP2A subunit B isoform B55-delta|PP2A subunit B isoform PR55-delta|PP2A subunit B isoform R2-delta|PP2A subunit B isoform delta|protein phosphatase 2, regulatory subunit B, delta isoform|protein phosphatase 2A, regulatory subunit B, delta isoform', map_location='7 82.95 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=138846386, end_position_on_the_genomic_accession=138883057, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  52440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52440, CurrentID=0, Status='live', Symbol='Tax1bp1', Aliases='1200003J11Rik, 1700069J21Rik, AA930106, D6Ertd404e, D6Ertd772e, T6bp, TXBP151', description='Tax1 (human T cell leukemia virus type I) binding protein 1', other_designations='TRAF6-binding protein homolog|Tax1 (human T-cell leukemia virus type I) binding protein 1', map_location='6 25.7 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52713575, end_position_on_the_genomic_accession=52766780, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  52443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52443, CurrentID=0, Status='live', Symbol='Mrpl48', Aliases='1810030E20Rik, 2610028L11Rik, CGI-118, D4Ertd786e', description='mitochondrial ribosomal protein L48', other_designations='L48mt|MRP-L48', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100548706, end_position_on_the_genomic_accession=100607996, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  52463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52463, CurrentID=0, Status='live', Symbol='Tet1', Aliases='2510010B09Rik, AA517754, BB001228, Cxxc6, D10Ertd17e, LCX, mKIAA1676', description='tet methylcytosine dioxygenase 1', other_designations='CXXC finger 6|CXXC-type zinc finger protein 6|ten-eleven translocation 1 gene protein homolog|tet oncogene 1', map_location='10 32.48 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62804570, end_position_on_the_genomic_accession=62887580, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  52466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52466, CurrentID=0, Status='live', Symbol='Slc46a1', Aliases='1110002C08Rik, D11Ertd18e, HCP1, Pcft', description='solute carrier family 46, member 1', other_designations='PCFT/HCP1|heme carrier protein 1', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78465701, end_position_on_the_genomic_accession=78471945, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  52468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52468, CurrentID=0, Status='live', Symbol='Ctdsp2', Aliases='AI586070, D10Ertd73e, OS-4, OS4, SCP2', description='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase 2', other_designations='nuclear LIM interactor-interacting factor 2 (NLI-interacting factor 2) (Protein OS-4)|small C-terminal domain phosphatase 2', map_location='10 74.35 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=126978717, end_position_on_the_genomic_accession=126999975, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  52469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52469, CurrentID=0, Status='live', Symbol='Coa3', Aliases='1810033A19Rik, Ccdc56, D11Ertd99e, HSPC009', description='cytochrome C oxidase assembly factor 3', other_designations='coiled-coil domain containing 56|coiled-coil domain-containing protein 56|cytochrome C oxidase assembly factor 3 homolog, mitochondrial|cytochrome c oxidase assembly protein 3 homolog, mitochondrial', map_location='11 64.56 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101277970, end_position_on_the_genomic_accession=101278948, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  52477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52477, CurrentID=0, Status='live', Symbol='Angel2', Aliases='2610307I21Rik, 5730410O10Rik, AI845692, D1Ertd396e, D1Ertd654e', description='angel homolog 2 (Drosophila)', other_designations='', map_location='1 96.28 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=190925108, end_position_on_the_genomic_accession=190946962, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  52480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52480, CurrentID=0, Status='live', Symbol='Snhg14', Aliases='AU018661, D7Ertd715e, Lncat', description='small nucleolar RNA host gene 14', other_designations='', map_location='7 34.04 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=59969577, end_position_on_the_genomic_accession=59974431, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  52486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52486, CurrentID=0, Status='live', Symbol='D19Rp19e', Aliases='D19Rp19', description='DNA segment, Chr 19, Roswell Park 19, expressed', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  52502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52502, CurrentID=0, Status='live', Symbol='Carhsp1', Aliases='1200011K09Rik, Crhsp-24, D16Ertd465e', description='calcium regulated heat stable protein 1', other_designations='calcineurin substrate CRHSP-24|calcium-regulated heat-stable protein of 24 kDa', map_location='16 4.26 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8658580, end_position_on_the_genomic_accession=8672153, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  52504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52504, CurrentID=0, Status='live', Symbol='Cenpo', Aliases='2810429O05Rik, 8430427C03Rik, D12Ertd482e', description='centromere protein O', other_designations='CENP-O', map_location='12 2.02 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4197058, end_position_on_the_genomic_accession=4234294, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  52513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52513, CurrentID=0, Status='live', Symbol='Ddx56', Aliases='2600001H07Rik, D11Ertd619e, Noh61', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 56', other_designations='ATP-dependent 61 kDa nucleolar RNA helicase|DEAD box protein 56|DEAD-box RNA helicase|nucleolar RNA helicase homolog', map_location='11 3.94 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6257545, end_position_on_the_genomic_accession=6267729, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  52521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52521, CurrentID=0, Status='live', Symbol='Zfp622', Aliases='1110033B05Rik, D15Ertd806e, ZPR9, Znf622', description='zinc finger protein 622', other_designations='', map_location='15 9.71 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=25984366, end_position_on_the_genomic_accession=25998482, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  52530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52530, CurrentID=0, Status='live', Symbol='Nhp2', Aliases='2410130M07Rik, D11Ertd175e, Nola2', description='NHP2 ribonucleoprotein', other_designations='H/ACA small nucleolar RNPs|NHP2 ribonucleoprotein homolog|nucleolar protein family A member 2|nucleolar protein family A, member 2|snoRNP protein NHP2', map_location='11 31.14 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51619773, end_position_on_the_genomic_accession=51623714, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  52535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52535, CurrentID=0, Status='live', Symbol='Mettl17', Aliases='2310032K15Rik, D14Ertd209e, Mett11d1', description='methyltransferase like 17', other_designations='methyltransferase 11 domain containing 1|methyltransferase 11 domain-containing protein 1|protein RSM22 homolog, mitochondrial', map_location='14 26.71 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51884842, end_position_on_the_genomic_accession=51891868, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  52538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52538, CurrentID=0, Status='live', Symbol='Acaa2', Aliases='0610011L04Rik, AI255831, AI265397, D18Ertd240e', description='acetyl-Coenzyme A acyltransferase 2 (mitochondrial 3-oxoacyl-Coenzyme A thiolase)', other_designations='acetyl-CoA acyltransferase|beta-ketothiolase|mitochondrial 3-oxoacyl-CoA thiolase', map_location='18 50.76 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74779212, end_position_on_the_genomic_accession=74806207, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  52551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52551, CurrentID=0, Status='live', Symbol='Sgta', Aliases='5330427H01Rik, AI194281, D10Ertd190e, Sgt, Stg', description='small glutamine-rich tetratricopeptide repeat (TPR)-containing, alpha', other_designations='alpha-SGT|small glutamine-rich tetratricopeptide repeat (TPR) containing protein', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81044071, end_position_on_the_genomic_accession=81060232, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  52552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52552, CurrentID=0, Status='live', Symbol='Parp8', Aliases='2810430O08Rik, ARTD16, D13Ertd275e', description='poly (ADP-ribose) polymerase family, member 8', other_designations='ADP-ribosyltransferase diphtheria toxin-like 16|PARP-8', map_location='13 65.32 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=116850842, end_position_on_the_genomic_accession=117025913, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  52563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52563, CurrentID=0, Status='live', Symbol='Cdc23', Aliases='6030435O18, D18Ertd243e', description='CDC23 cell division cycle 23', other_designations='APC8|CDC23 (cell division cycle 23, yeast, homolog)|anaphase-promoting complex subunit 8|cyclosome subunit 8', map_location='18 18.69 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34630946, end_position_on_the_genomic_accession=34651745, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  52570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52570, CurrentID=0, Status='live', Symbol='Ccdc69', Aliases='2210021E03Rik, D11Ertd461e', description='coiled-coil domain containing 69', other_designations='', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55049733, end_position_on_the_genomic_accession=55078645, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  52575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52575, CurrentID=0, Status='live', Symbol='Trmt10c', Aliases='1300018J16Rik, D16Ertd454e, Rg9mtd1, Rnmtd1', description='tRNA methyltransferase 10C', other_designations='RNA (guanine-9-) methyltransferase domain containing 1|RNA (guanine-9-)-methyltransferase domain-containing protein 1|mitochondrial RNase P protein 1|tRNA methyltransferase 10 homolog C', map_location='16 33.79 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56033720, end_position_on_the_genomic_accession=56037774, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  52585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52585, CurrentID=0, Status='live', Symbol='Dhrs1', Aliases='1110029G07Rik, AW112170, D14Ertd484e', description='dehydrogenase/reductase (SDR family) member 1', other_designations='', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55739020, end_position_on_the_genomic_accession=55745684, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  52588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52588, CurrentID=0, Status='live', Symbol='Tspan14', Aliases='AA958793, D14Ertd226e, Tm4sf14, tspan-14', description='tetraspanin 14', other_designations='transmembrane 4 superfamily member 14', map_location='14 22.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=40906427, end_position_on_the_genomic_accession=40966807, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  52589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52589, CurrentID=0, Status='live', Symbol='Ncald', Aliases='AI848120, D030020D09Rik, D15Ertd412e', description='neurocalcin delta', other_designations='', map_location='15 14.94 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=37366175, end_position_on_the_genomic_accession=37792519, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  52592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52592, CurrentID=0, Status='live', Symbol='Brms1l', Aliases='0710008O11Rik, AI159718, BRMS1, D12Ertd407e', description='breast cancer metastasis-suppressor 1-like', other_designations='', map_location='12 24.18 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55836243, end_position_on_the_genomic_accession=55869736, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  52609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52609, CurrentID=0, Status='live', Symbol='Cbx7', Aliases='1600014J01Rik, AI851678, D15Ertd417e', description='chromobox 7', other_designations='chromobox homolog 7', map_location='15 37.85 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79915804, end_position_on_the_genomic_accession=79971149, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  52614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52614, CurrentID=0, Status='live', Symbol='Adgre4', Aliases='D17Ertd479e, Egf-tm7, Emr4, Fire, Gpr127', description='adhesion G protein-coupled receptor E4', other_designations='EGF-like module containing, mucin-like, hormone receptor-like sequence 4|EGF-like module receptor 4|EGF-like module-containing mucin-like hormone receptor-like 4|F4/80-like-receptor|seven-span membrane protein FIRE', map_location='17 28.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55749984, end_position_on_the_genomic_accession=55853662, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  52615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52615, CurrentID=0, Status='live', Symbol='Suz12', Aliases='2610028O16Rik, AI195385, AU016842, AW536442, D11Ertd530e, mKIAA0160', description='suppressor of zeste 12 homolog (Drosophila)', other_designations='suppressor of zeste 12 protein homolog', map_location='11 47.36 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79993106, end_position_on_the_genomic_accession=80034123, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  52626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52626, CurrentID=0, Status='live', Symbol='Cdkn2aipnl', Aliases='A430101B06Rik, AA409260, D11Ertd497e, MNCb-1520', description='CDKN2A interacting protein N-terminal like', other_designations='CDKN2A-interacting protein N-terminal-like protein', map_location='11 31.78 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51967631, end_position_on_the_genomic_accession=51977336, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  52633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52633, CurrentID=0, Status='live', Symbol='Nit2', Aliases='1190017B19Rik, D16Ertd502e', description='nitrilase family, member 2', other_designations='Nit protein 2|nitrilase homolog 2', map_location='16 34.22 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57156665, end_position_on_the_genomic_accession=57167332, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  52635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52635, CurrentID=0, Status='live', Symbol='Esyt2', Aliases='2410017M09Rik, 4921504I16Rik, CHR2SYT, D12Ertd551e, Fam62b', description='extended synaptotagmin-like protein 2', other_designations='E-Syt2|chr2 synaptotagmin|family with sequence similarity 62 (C2 domain containing), member B|family with sequence similarity 62, member B', map_location='12 62.65 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=116279680, end_position_on_the_genomic_accession=116390935, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  52637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52637, CurrentID=0, Status='live', Symbol='Cisd1', Aliases='AU043990, AW743335, D10Ertd214e, Zcd1, mitoNEET', description='CDGSH iron sulfur domain 1', other_designations='CDGSH iron sulfur domain-containing protein 1|zinc finger, CDGSH-type domain 1', map_location='10 36.88 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=71330494, end_position_on_the_genomic_accession=71344849, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  52639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52639, CurrentID=0, Status='live', Symbol='Wipi1', Aliases='4930533H01Rik, AW411817, D11Ertd498e', description='WD repeat domain, phosphoinositide interacting 1', other_designations='WD40 repeat protein Interacting with phosphoInositides of 49kDa|WD40 repeat protein interacting with phosphoinositides of 49 kDa|WIPI 49 kDa|WIPI-1', map_location='11 72.18 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109573521, end_position_on_the_genomic_accession=109611389, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  52653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52653, CurrentID=0, Status='live', Symbol='Nudcd2', Aliases='', description='NudC domain containing 2', other_designations='', map_location='11 24.41 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=40733661, end_position_on_the_genomic_accession=40740046, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  52662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52662, CurrentID=0, Status='live', Symbol='Ldlrad4', Aliases='8230401C20Rik, A430083H02, A430108L08Rik, C18orf1, D18Ertd653e, D330030L18Rik', description='low density lipoprotein receptor class A domain containing 4', other_designations='', map_location='18 40.14 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67932245, end_position_on_the_genomic_accession=68260550, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  52665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52665, CurrentID=0, Status='live', Symbol='Echdc1', Aliases='1700028A24Rik, AI314462, AI930038, D10Ertd667e, MMCD', description='enoyl Coenzyme A hydratase domain containing 1', other_designations='enoyl-CoA hydratase domain-containing protein 1|methylmalonyl-CoA decarboxylase', map_location='10 16.59 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=29311507, end_position_on_the_genomic_accession=29346661, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  52666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52666, CurrentID=0, Status='live', Symbol='Arhgef25', Aliases='2410008H17Rik, D10Ertd610e, GEFT', description='Rho guanine nucleotide exchange factor (GEF) 25', other_designations='guanine nucleotide exchange factor GEFT|p63RhoGEF|rac/Cdc42/Rho exchange factor GEFT|rhoA/Rac/Cdc42 guanine nucleotide exchange factor GEFT', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127182521, end_position_on_the_genomic_accession=127190088, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  52668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52668, CurrentID=0, Status='live', Symbol='Ifi27', Aliases='1110013J02Rik, 2900026P10Rik, D12Ertd647e, ISG12al1, Ifi27', description='interferon, alpha-inducible protein 27', other_designations='', map_location='12 52.93 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103434189, end_position_on_the_genomic_accession=103440245, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  52670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52670, CurrentID=0, Status='live', Symbol='Cpsf4l', Aliases='0610010C04Rik, 1500000C01Rik, D11Ertd636e', description='cleavage and polyadenylation specific factor 4-like', other_designations='', map_location='11 79.17 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113696078, end_position_on_the_genomic_accession=113711425, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  52679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52679, CurrentID=0, Status='live', Symbol='E2f7', Aliases='A630014C11Rik, D10Ertd739e', description='E2F transcription factor 7', other_designations='E2F-7', map_location='10 57.74 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=110745427, end_position_on_the_genomic_accession=110787384, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  52683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52683, CurrentID=0, Status='live', Symbol='Ncaph2', Aliases='', description='non-SMC condensin II complex, subunit H2', other_designations='kleisin beta', map_location='15 44.84 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89355719, end_position_on_the_genomic_accession=89372850, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  52685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52685, CurrentID=0, Status='live', Symbol='Cd300lg', Aliases='2310016B05Rik, Clm9, D11Ertd736e', description='CD300 antigen like family member G', other_designations='CD300 antigen-like family member G|CLM-9|CMRF-35-like molecule-9|nepmucin', map_location='11 65.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102041511, end_position_on_the_genomic_accession=102055617, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  52686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52686, CurrentID=0, Status='live', Symbol='Mettl2', Aliases='2810438F06Rik, AW493404, C130031G21Rik, D11Ertd768e, METL, PSENIP1', description='methyltransferase like 2', other_designations='', map_location='11 68.61 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105126425, end_position_on_the_genomic_accession=105141146, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  52690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52690, CurrentID=0, Status='live', Symbol='Setd3', Aliases='2610102I01Rik, 2610305M23Rik, 6030490B17Rik, D12Ertd771e', description='SET domain containing 3', other_designations='SET domain-containing protein 3|endothelial differentiation inhibitory protein D10', map_location='12 59.23 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108106431, end_position_on_the_genomic_accession=108179381, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  52696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52696, CurrentID=0, Status='live', Symbol='Zwint', Aliases='2010007E07Rik, 2600001N01Rik, D10Ertd749e-1, Zwint', description='ZW10 interactor', other_designations='ZW10-interacting protein 1', map_location='10 37.15 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=72654846, end_position_on_the_genomic_accession=72674964, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  52700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52700, CurrentID=0, Status='live', Symbol='Txndc17', Aliases='4831443O22Rik, D11Ertd672e, TRP14, Txnl5', description='thioredoxin domain containing 17', other_designations='14 kDa thioredoxin-related protein|protein 42-9-9|putative 42-9-9 protein|thioredoxin-like 5|thioredoxin-like protein 5', map_location='11 43.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72207554, end_position_on_the_genomic_accession=72210487, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  52705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52705, CurrentID=0, Status='live', Symbol='Krr1', Aliases='2610511F02Rik, AI255219, AI428520, D10Ertd773e, Hrb2', description='KRR1, small subunit (SSU) processome component, homolog (yeast)', other_designations='HIV-1 Rev binding protein 2|HIV-1 Rev-binding protein 2 homolog|KRR-R motif-containing protein 1|KRR1, small subunit processome component', map_location='10 59.97 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=111972695, end_position_on_the_genomic_accession=111988430, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  52708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52708, CurrentID=0, Status='live', Symbol='Zfp410', Aliases='Apa1, D12Ertd748e, Znf410', description='zinc finger protein 410', other_designations='another partner for ARF 1|zinc finger protein APA-1', map_location='12 39.15 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84315928, end_position_on_the_genomic_accession=84344439, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  52710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52710, CurrentID=0, Status='live', Symbol='Slc52a2', Aliases='2010003P03Rik, D15Ertd747e, GPCR42, Gpr172b, PAR2, mRFT1', description='solute carrier protein 52, member 2', other_designations='G protein-coupled receptor 172B|PERV-A receptor 2 homolog|porcine endogenous retrovirus A receptor 2 homolog|riboflavin transporter 1', map_location='15 36.04 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76538828, end_position_on_the_genomic_accession=76542130, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  52712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52712, CurrentID=0, Status='live', Symbol='Zkscan6', Aliases='1700128E15Rik, D11Ertd714e, KOX11, Zfp535, Znf18', description='zinc finger with KRAB and SCAN domains 6', other_designations='zinc finger protein 535|zinc finger protein with KRAB and SCAN domains 6', map_location='11 40.53 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=65807175, end_position_on_the_genomic_accession=65829239, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  52713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52713, CurrentID=0, Status='live', Symbol='Ccdc59', Aliases='2300004H16Rik, D10Ertd718e', description='coiled-coil domain containing 59', other_designations='TTF-1-associated protein 26|coiled-coil domain-containing protein 59', map_location='10 55.12 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=105841479, end_position_on_the_genomic_accession=105847510, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  52715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52715, CurrentID=0, Status='live', Symbol='Ccdc43', Aliases='2610001E01Rik, D11Ertd707e', description='coiled-coil domain containing 43', other_designations='', map_location='11 66.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102684684, end_position_on_the_genomic_accession=102697943, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  52717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52717, CurrentID=0, Status='live', Symbol='Anapc16', Aliases='2310005G07Rik, APC16, D10Ertd641e', description='anaphase promoting complex subunit 16', other_designations='cyclosome subunit 16', map_location='10 29.66 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59987409, end_position_on_the_genomic_accession=60003131, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  52793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52793, CurrentID=0, Status='live', Symbol='Fam3b', Aliases='2-21, 9030624C24Rik, D16Jhu19e, ORF9, Pander', description='family with sequence similarity 3, member B', other_designations='cytokine-like protein 2-21|pancreatic-derived factor', map_location='16 57.47 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97471086, end_position_on_the_genomic_accession=97504936, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  52808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52808, CurrentID=0, Status='live', Symbol='Tspyl2', Aliases='CDA1, CINAP, DENTT, DXBwg1396e, DXHXS1008E, E130307F10Rik', description='TSPY-like 2', other_designations='CASK interacting nucleosome assembly protein|CASK-interacting nucleosome assembly protein|TSPY-like protein 2|cell division autoantigen 1 nucleolar protein|differentially expressed nucleolar TGF-beta1 target|differentially-expressed nucleolar TGF-beta1 target protein|nucleolar TGF-beta1 target protein', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152336852, end_position_on_the_genomic_accession=152342506, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  52815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52815, CurrentID=0, Status='live', Symbol='Ldhd', Aliases='4733401P21Rik, D8Bwg1320e', description='lactate dehydrogenase D', other_designations='DLD|ldhd', map_location='8 57.98 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111626271, end_position_on_the_genomic_accession=111630322, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  52822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52822, CurrentID=0, Status='live', Symbol='Rufy3', Aliases='2810428M05Rik, 6330416M07Rik, AW455998, AW538594, D5Bwg0860e, Ripx, Rpipx', description='RUN and FYVE domain containing 3', other_designations='rap2 interacting protein x|singar|singar1|single axon-regulated protein|single axon-regulated protein 1', map_location='5 43.77 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88564780, end_position_on_the_genomic_accession=88651417, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  52829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52829, CurrentID=0, Status='live', Symbol='Lurap1l', Aliases='1110029A09Rik, AV077978, AV175137, D4Bwg0951e, WI-13707, bM350F23.1', description='leucine rich adaptor protein 1-like', other_designations='amplimer WI-13707, mouse homolog', map_location='4 37.91 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=80910686, end_position_on_the_genomic_accession=80954301, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  52830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52830, CurrentID=0, Status='live', Symbol='Pnrc2', Aliases='0610011E17Rik, D4Bwg0593e', description='proline-rich nuclear receptor coactivator 2', other_designations='', map_location='4 68.01 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135870917, end_position_on_the_genomic_accession=135873846, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  52837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52837, CurrentID=0, Status='live', Symbol='Tmx4', Aliases='2810417D04Rik, 4930500L08Rik, AI843224, AW046784, D2Bwg1356e, Txndc13, mKIAA1162', description='thioredoxin-related transmembrane protein 4', other_designations='thioredoxin domain containing 13|thioredoxin domain-containing protein 13', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=134594501, end_position_on_the_genomic_accession=134644121, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  52838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52838, CurrentID=0, Status='live', Symbol='Dnlz', Aliases='1110034G11Rik, D2Bwg1335e, HEP1', description='DNL-type zinc finger', other_designations='hsp70-escort protein 1|mtHsp70-escort protein', map_location='2 18.87 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26348118, end_position_on_the_genomic_accession=26352110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  52840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52840, CurrentID=0, Status='live', Symbol='Dbndd2', Aliases='1110017A21Rik, 2900022J10Rik, AU041050, AW048677, D2Bwg0891e, Nkip, R74724', description='dysbindin (dystrobrevin binding protein 1) domain containing 2', other_designations='SCF apoptosis response protein 1|likely orthologue of H. sapiens chromosome 20 open reading frame 35', map_location='2 85.17 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164472235, end_position_on_the_genomic_accession=164493323, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  52846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52846, CurrentID=0, Status='live', Symbol='Cnot11', Aliases='2410015L18Rik, C40, D1Bwg0212e', description='CCR4-NOT transcription complex, subunit 11', other_designations='UPF0760 protein C2orf29 homolog', map_location='1 18.31 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39535802, end_position_on_the_genomic_accession=39546881, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  52850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52850, CurrentID=0, Status='live', Symbol='Sgsm1', Aliases='2410098H20Rik, AA408781, D5Bwg1524e, E130107M15, NuIP, Rutbc2', description='small G protein signaling modulator 1', other_designations='RUN and TBC1 domain containing 2|RUN and TBC1 domain-containing protein 2|nurr1-interacting protein', map_location='5 55.45 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113243220, end_position_on_the_genomic_accession=113310787, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  52855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52855, CurrentID=0, Status='live', Symbol='Lair1', Aliases='5133400O11Rik, BB115266, D7Bwg0421e, Lair-1', description='leukocyte-associated Ig-like receptor 1', other_designations='', map_location='7 2.31 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4007073, end_position_on_the_genomic_accession=4073507, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  52856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52856, CurrentID=0, Status='live', Symbol='Mtg2', Aliases='1810011P19Rik, 2900056P18Rik, D2Bwg0647e, Gtpbp5', description='mitochondrial ribosome associated GTPase 2', other_designations='GTP binding protein 5', map_location='2 102.59 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180070593, end_position_on_the_genomic_accession=180085902, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  52857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52857, CurrentID=0, Status='live', Symbol='Gramd1a', Aliases='1300003M23Rik, AI415002, AW547240, D7Bwg0611e', description='GRAM domain containing 1A', other_designations='', map_location='7 19.3 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31130127, end_position_on_the_genomic_accession=31155935, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  52858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52858, CurrentID=0, Status='live', Symbol='Cdipt', Aliases='9530042F15Rik, D7Bwg0575e, Pis, Pis1', description='CDP-diacylglycerol--inositol 3-phosphatidyltransferase (phosphatidylinositol synthase)', other_designations='PI synthase|phosphatidylinositol synthase|ptdIns synthase', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126975914, end_position_on_the_genomic_accession=126980501, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  52864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52864, CurrentID=0, Status='live', Symbol='Slx4', Aliases='AI256635, AI426760, Btbd12, D16Bwg1016e', description='SLX4 structure-specific endonuclease subunit homolog (S. cerevisiae)', other_designations='BTB/POZ domain-containing protein 12', map_location='16 2.34 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3979105, end_position_on_the_genomic_accession=4003829, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  52874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52874, CurrentID=0, Status='live', Symbol='Pum3', Aliases='1110069H02Rik, AA675048, D19Bwg1357e, Kiaa0020', description='pumilio RNA-binding family member 3', other_designations='', map_location='19 21.96 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=27388702, end_position_on_the_genomic_accession=27429820, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  52882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52882, CurrentID=0, Status='live', Symbol='Rgs7bp', Aliases='A930030I01Rik, AI842293, D13Bwg1146e, R7bp', description='regulator of G-protein signalling 7 binding protein', other_designations='R7 binding protein|R7 family-binding protein', map_location='13 56.9 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104945904, end_position_on_the_genomic_accession=105054930, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  446099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=446099, CurrentID=0, Status='live', Symbol='Nlrp4e', Aliases='4930406H16Rik, Nalp-epsilon, Nalp4e, Nlrp4', description='NLR family, pyrin domain containing 4E', other_designations='NACHT, leucine rich repeat and PYD containing 4E|NALP, epsilon|NLR family, pyrin domain containing 4', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23301192, end_position_on_the_genomic_accession=23400240, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  52884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52884, CurrentID=0, Status='live', Symbol='Mgll-rs2', Aliases='D13Bwg0938e', description='monoglyceride lipase, related sequence 2', other_designations='', map_location='6 39.59 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  446101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=446101, CurrentID=0, Status='live', Symbol='Xrra1', Aliases='AI449753', description='X-ray radiation resistance associated 1', other_designations='xrra1', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99859218, end_position_on_the_genomic_accession=99917827, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  52892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52892, CurrentID=0, Status='live', Symbol='Sco1', Aliases='2610001C07Rik, D11Bwg1310e', description='SCO cytochrome oxidase deficient homolog 1 (yeast)', other_designations='', map_location='11 40.59 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67052670, end_position_on_the_genomic_accession=67067444, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  52897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52897, CurrentID=0, Status='live', Symbol='Rbfox3', Aliases='Fox-3, Hrnbp3, NeuN, Neuna60', description='RNA binding protein, fox-1 homolog (C. elegans) 3', other_designations='fox-1 homolog C|hexaribonucleotide binding protein 3|neuronal nuclear antigen A60|neuronal nuclei', map_location='11 83.22 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118489760, end_position_on_the_genomic_accession=118911597, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  52898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52898, CurrentID=0, Status='live', Symbol='Rnasek', Aliases='2310033H11Rik, AA571397, D11Bwg0434e', description='ribonuclease, RNase K', other_designations='RNase K|RNase kappa', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70238123, end_position_on_the_genomic_accession=70239852, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  52902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52902, CurrentID=0, Status='live', Symbol='D10Bwg1070e', Aliases='5730421K10Rik', description="DNA segment, Chr 10, Brigham & Women's Genetics 1070 expressed", other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  52906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52906, CurrentID=0, Status='live', Symbol='Ahi1', Aliases='1700015F03Rik, Ahi-1, D10Bwg0629e', description='Abelson helper integration site 1', other_designations='abelson helper integration site 1 protein', map_location='10 9.75 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=20952489, end_position_on_the_genomic_accession=21080429, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  52915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52915, CurrentID=0, Status='live', Symbol='Zmiz2', Aliases='Zimp7', description='zinc finger, MIZ-type containing 2', other_designations='PIAS-like protein Zimp7', map_location='11 3.95 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6389064, end_position_on_the_genomic_accession=6406162, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  52919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=52919, CurrentID=0, Status='live', Symbol='D5Mnl25e', Aliases='D5Mnl25, D5Sc25', description='DNA segment, Chr 5, Massachusetts Neurogenetics Laboratory 25, expressed', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53310, CurrentID=0, Status='live', Symbol='Dlg3', Aliases='Dlgh3, SAP102, mKIAA1232', description='discs, large homolog 3 (Drosophila)', other_designations='synapse-associated protein 102', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100767586, end_position_on_the_genomic_accession=100818410, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  53311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53311, CurrentID=0, Status='live', Symbol='Mybph', Aliases='AI385643, MyBP-H', description='myosin binding protein H', other_designations='H-protein', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134193417, end_position_on_the_genomic_accession=134201234, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  53312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53312, CurrentID=0, Status='live', Symbol='Nub1', Aliases='4931404D21Rik, 6330412F12Rik, BS4, NY-REN-18', description='negative regulator of ubiquitin-like proteins 1', other_designations='NY-REN-18 antigen', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24685537, end_position_on_the_genomic_accession=24710555, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  53313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53313, CurrentID=0, Status='live', Symbol='Atp2a3', Aliases='SERCA3b, Serca3', description='ATPase, Ca++ transporting, ubiquitous', other_designations='SR Ca(2+)-ATPase 3|calcium pump 3|sarco /endoplasmic reticulum Ca2+ -ATPase|sarcoendoplasmic reticulum Ca2+ ATPase type 3', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72961169, end_position_on_the_genomic_accession=72993043, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  53314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53314, CurrentID=0, Status='live', Symbol='Batf', Aliases='B-ATF, SFA-2', description='basic leucine zipper transcription factor, ATF-like', other_designations='B-cell activating transcription factor|B-cell-activating transcription factor|activating transcription factor B|basic leucine zipper transcription factor-like', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85686720, end_position_on_the_genomic_accession=85709087, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53315, CurrentID=0, Status='live', Symbol='Sult1d1', Aliases='5033411P13Rik, AI987815, ST1d1, SULT-N, Sultn', description='sulfotransferase family 1D, member 1', other_designations='N-sulfotransferase|amine N-sulfotransferase|dopamine sulfotransferase Sult1d1|tyrosine-ester sulfotransferase', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87554650, end_position_on_the_genomic_accession=87569006, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  53317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53317, CurrentID=0, Status='live', Symbol='Plrg1', Aliases='AA958940, C80566, Tango4', description='pleiotropic regulator 1', other_designations='pleiotropic regulator 1, PRL1 homolog', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=83055538, end_position_on_the_genomic_accession=83072291, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  53318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53318, CurrentID=0, Status='live', Symbol='Pdlim3', Aliases='AI463105, ALP, Actn2lp', description='PDZ and LIM domain 3', other_designations='actinin alpha 2 associated LIM protein|actinin-associated LIM protein|alpha-actinin-2-associated LIM protein', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45885453, end_position_on_the_genomic_accession=45919548, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  53319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53319, CurrentID=0, Status='live', Symbol='Nxf1', Aliases='Mex67, Mvb1, Tap', description='nuclear RNA export factor 1', other_designations='Mex 67 homolog|mRNA export factor TAP|modifier of vibrator 1|nuclear RNA export factor 1 homolog|tip associated protein|tip associating protein|tip-associated protein|tip-associating protein', map_location='19 5.5 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8757103, end_position_on_the_genomic_accession=8770910, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  53320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53320, CurrentID=0, Status='live', Symbol='Folh1', Aliases='GCP2, mopsm', description='folate hydrolase 1', other_designations='FGCP|GCPII|N-acetylated-alpha-linked acidic dipeptidase I|NAALADase I|folylpoly-gamma-glutamate carboxypeptidase|glutamate carboxypeptidase II|mGCP|membrane glutamate carboxypeptidase|prostate-specific membrane antigen homolog|pteroylpoly-gamma-glutamate carboxypeptidase', map_location='7 D1-D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=86718781, end_position_on_the_genomic_accession=86776055, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  53321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53321, CurrentID=0, Status='live', Symbol='Cntnap1', Aliases='AI841080, Caspr, NCP1, Nrxn4, p190, shm', description='contactin associated protein-like 1', other_designations='MHDNIV|caspr1|contactin associated protein 1|neurexin 4 (contactin associated protein)|neurexin IV|neurexin-4|paranodin', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101174629, end_position_on_the_genomic_accession=101190727, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  53322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53322, CurrentID=0, Status='live', Symbol='Nucb2', Aliases='AI607786, Calnuc, Nefa, Nesfatin-1', description='nucleobindin 2', other_designations='DNA-binding protein NEFA|prepronefastin|prepronesfatin', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=116504353, end_position_on_the_genomic_accession=116540588, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  53323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53323, CurrentID=0, Status='live', Symbol='Ube2k', Aliases='AW492011, D5Ertd601e, E2-25k, HIP-2, Hip2, Hypg, Lig', description='ubiquitin-conjugating enzyme E2K', other_designations='E2 ubiquitin-conjugating enzyme K|huntingtin interacting protein 2|ubiquitin carrier protein|ubiquitin-conjugating enzyme E2-25 kDa|ubiquitin-conjugating enzyme E2K (UBC1 homolog, yeast)|ubiquitin-protein ligase', map_location='5 33.72 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65537238, end_position_on_the_genomic_accession=65598989, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  53324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53324, CurrentID=0, Status='live', Symbol='Nptx2', Aliases='Narp, np2', description='neuronal pentraxin 2', other_designations='NP-II|neuro activity regulated petaxin|neuronal pentraxin II', map_location='5 84.58 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144545887, end_position_on_the_genomic_accession=144557478, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  53325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53325, CurrentID=0, Status='live', Symbol='Banp', Aliases='AA408158, SMAR1', description='BTG3 associated nuclear protein', other_designations='scaffold/matrix-associated region-1-binding protein', map_location='8 70.82 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121950492, end_position_on_the_genomic_accession=122029260, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  53328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53328, CurrentID=0, Status='live', Symbol='Pgrmc1', Aliases='AA415812, HPR6.6, PPMR, Vema', description='progesterone receptor membrane component 1', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36598225, end_position_on_the_genomic_accession=36606079, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  53330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53330, CurrentID=0, Status='live', Symbol='Vamp4', Aliases='D1Ertd147e', description='vesicle-associated membrane protein 4', other_designations='VAMP-4|vesicle associated membrane protein 4', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162570404, end_position_on_the_genomic_accession=162599086, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  53331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53331, CurrentID=0, Status='live', Symbol='Stx7', Aliases='AI315064, AI317144, AW107239, Syn7', description='syntaxin 7', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24149290, end_position_on_the_genomic_accession=24188961, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  53332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53332, CurrentID=0, Status='live', Symbol='Mtmr1', Aliases='AW049210', description='myotubularin related protein 1', other_designations='phosphatidylinositol-3-phosphate phosphatase', map_location='X A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71364737, end_position_on_the_genomic_accession=71419197, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  53333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53333, CurrentID=0, Status='live', Symbol='Tomm40', Aliases='AW539759, Mom35, Tom40', description='translocase of outer mitochondrial membrane 40 homolog (yeast)', other_designations='haymaker protein|mitochondrial outer membrane protein 35|mitochondrial outer membrane protein of 35 kDa|translocase of outer membrane 40 kDa subunit homolog', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19701313, end_position_on_the_genomic_accession=19715429, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  53334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53334, CurrentID=0, Status='live', Symbol='Gosr1', Aliases='AI414660, AI426320, BB145494, GOS-28, GOSRI, GS28', description='golgi SNAP receptor complex member 1', other_designations='28 kDa Golgi SNARE protein|28 kDa cis-Golgi SNARE p28|Cis-Golgi SNARE', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76726602, end_position_on_the_genomic_accession=76763622, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  53356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53356, CurrentID=0, Status='live', Symbol='Eif3g', Aliases='44kDa, D0Jmb4, Eif3s4, TU-189B2, eIF3-p44, p44', description='eukaryotic translation initiation factor 3, subunit G', other_designations='eIF-3 RNA-binding subunit|eIF-3-delta|eIF3 p42|eIF3 p44|eIF3g|eiF3 delta|eukaryotic translation initiation factor 3 RNA-binding subunit|eukaryotic translation initiation factor 3 subunit 4|eukaryotic translation initiation factor 3, subunit 4 (delta)|eukaryotic translation initiation factor 3, subunit 4 (delta, 44 kDa)|eukaryotic translation initiation factor 3, subunit 4 (delta, 44kD); eukaryotic translation initiation factor 3 subunit p42|translation initiation factor eIF3-p44', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20894349, end_position_on_the_genomic_accession=20898590, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  53357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53357, CurrentID=0, Status='live', Symbol='Pla2g6', Aliases='BB112799, PNPLA9, iPLA(2)beta, iPLA2, iPLA2beta', description='phospholipase A2, group VI', other_designations='85 kDa calcium-independent phospholipase A2|GVI PLA2|caI-PLA2|group VI phospholipase A2|iPLA2-beta|intracellular membrane-associated calcium-independent phospholipase A2 beta|patatin-like phospholipase domain-containing protein 9', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79286228, end_position_on_the_genomic_accession=79328530, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  53374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53374, CurrentID=0, Status='live', Symbol='Chst3', Aliases='C6ST, C6ST-1, GST-0', description='carbohydrate (chondroitin 6/keratan) sulfotransferase 3', other_designations='chondroitin 6-O-sulfotransferase 1|chondroitin 6-sulfotransferase|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 0', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60181527, end_position_on_the_genomic_accession=60219260, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  53375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53375, CurrentID=0, Status='live', Symbol='Mtx2', Aliases='1500012G02Rik', description='metaxin 2', other_designations='mitochondrial outer membrane import complex protein 2', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74825812, end_position_on_the_genomic_accession=74876748, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  53376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53376, CurrentID=0, Status='live', Symbol='Usp2', Aliases='B930035K21Rik, Ubp41', description='ubiquitin specific peptidase 2', other_designations='41 kDa ubiquitin-specific protease|deubiquitinating enzyme 2|ubiquitin specific protease 2|ubiquitin thioesterase 2|ubiquitin thiolesterase 2|ubiquitin-specific-processing protease 2', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44067021, end_position_on_the_genomic_accession=44095627, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  53378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53378, CurrentID=0, Status='live', Symbol='Sdcbp', Aliases='MDA-9, Sycl, syntenin-1', description='syndecan binding protein', other_designations='scaffold protein Pbp1|syndecan interacting protein|syndecan-binding protein 1|syntenin', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=6365680, end_position_on_the_genomic_accession=6396122, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  53379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53379, CurrentID=0, Status='live', Symbol='Hnrnpa2b1', Aliases='9130414A06Rik, Hnrpa2, Hnrpa2b1, hnrnp-A', description='heterogeneous nuclear ribonucleoprotein A2/B1', other_designations='heterogenous nuclear ribonucleoprotein A2/B1|hnRNP A2/B1', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=51460434, end_position_on_the_genomic_accession=51470492, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  53380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53380, CurrentID=0, Status='live', Symbol='Psmd10', Aliases='AW554874', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 10', other_designations='26S proteasome regulatory subunit p28|gankyrin|proteasome subunit p28|proteosome (prosome, macropain) 26S subunit, non-ATPase, 10|proteosome subunit p28', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140948425, end_position_on_the_genomic_accession=140956711, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  53381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53381, CurrentID=0, Status='live', Symbol='Prdx4', Aliases='AOE372, Prx-iv, Prx4, TRANK', description='peroxiredoxin 4', other_designations='Prx IV|antioxidant enzyme AOE372|peroxiredoxin IV|thioredoxin peroxidase AO372|thioredoxin-dependent peroxide reductase A0372', map_location='X 72.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155323918, end_position_on_the_genomic_accession=155340749, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  53382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53382, CurrentID=0, Status='live', Symbol='Txnl1', Aliases='32kDa, TRP32, Txnl', description='thioredoxin-like 1', other_designations='32 kDa thioredoxin-related protein|thioredoxin-like (32kD)|thioredoxin-related protein', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=63662801, end_position_on_the_genomic_accession=63692359, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  53404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53404, CurrentID=0, Status='live', Symbol='Atoh7', Aliases='Math5, bHLHa13', description='atonal bHLH transcription factor 7', other_designations='atonal homolog 7|helix-loop-helix protein mATH-5', map_location='10 32.54 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63098020, end_position_on_the_genomic_accession=63101284, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53412, CurrentID=0, Status='live', Symbol='Ppp1r3c', Aliases='PTG, Ppp1r5', description='protein phosphatase 1, regulatory (inhibitor) subunit 3C', other_designations='PP1 subunit R5|ppp1r3c|protein phosphatase 1 binding protein PTG|protein phosphatase 1 regulatory subunit 5|protein phosphatase 1, regulatory (inhibitor) subunit 5|protein targeting to glicogen|protein targeting to glycogen', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36731731, end_position_on_the_genomic_accession=36736604, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53413, CurrentID=0, Status='live', Symbol='Exoc7', Aliases='Exo70, sec70', description='exocyst complex component 7', other_designations='exocyst complex component Exo70|exocyst component protein 70 kDa homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116287998, end_position_on_the_genomic_accession=116306749, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  53414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53414, CurrentID=0, Status='live', Symbol='Bysl', Aliases='Bys, Enp1', description='bystin-like', other_designations='', map_location='17 23.63 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47599331, end_position_on_the_genomic_accession=47611492, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  53415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53415, CurrentID=0, Status='live', Symbol='Htatip2', Aliases='AW111545, Cc3, Tip30', description='HIV-1 Tat interactive protein 2', other_designations='HIV-1 tat interactive protein 2, 30 kDa homolog|HIV-1 tat interactive protein 2, homolog|htatip2|tat-interacting protein 30 kDa', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=49759045, end_position_on_the_genomic_accession=49778364, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  53416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53416, CurrentID=0, Status='live', Symbol='Stk39', Aliases='AW227544, AW556857, DCHT, RF005, Rnl5, SPAK', description='serine/threonine kinase 39', other_designations='Ste-20 related kinase|kidney-specific SPAK|serine/threonine kinase 39, STE20/SPS1 homolog|serine/threonine-protein kinase 39|ste-20-related kinase', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=68210445, end_position_on_the_genomic_accession=68472965, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  53417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53417, CurrentID=0, Status='live', Symbol='Hif3a', Aliases='Ipas, MOP7, NEPAS, bHLHe17', description='hypoxia inducible factor 3, alpha subunit', other_designations='HIF-3-alpha|HIF3-alpha|HIF3-alpha-1|basic-helix-loop-helix-PAS protein MOP7|hypoxia inducible factor three alpha|inhibitory PAS domain protein|member of PAS protein 7|neonatal and embryonic PAS protein', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17030993, end_position_on_the_genomic_accession=17062427, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  53418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53418, CurrentID=0, Status='live', Symbol='B4galt2', Aliases='Ggtb2', description='UDP-Gal:betaGlcNAc beta 1,4- galactosyltransferase, polypeptide 2', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 2|UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 2|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 2|b4Gal-T2|beta-1,4-GalTase 2|beta-1,4-galactosyltransferase II|beta4Gal-T2', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117873000, end_position_on_the_genomic_accession=117883482, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  53419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53419, CurrentID=0, Status='live', Symbol='Corin', Aliases='AV273130, Lrp4', description='corin', other_designations='low density lipoprotein receptor-related protein 4|low density lipoprotein-related protein 4|pro-ANP-converting enzyme', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72299948, end_position_on_the_genomic_accession=72504540, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  53420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53420, CurrentID=0, Status='live', Symbol='Syt5', Aliases='SytV', description='synaptotagmin V', other_designations='synaptotagmin 5|synaptotagmin IX', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4539765, end_position_on_the_genomic_accession=4547594, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  53421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53421, CurrentID=0, Status='live', Symbol='Sec61a1', Aliases='AA408394, AA410007, Sec61a', description='Sec61 alpha 1 subunit (S. cerevisiae)', other_designations='SEC61, alpha subunit|protein transport protein Sec61 subunit alpha|sec61 alpha-1|transport protein sec61 alpha subunit', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88503607, end_position_on_the_genomic_accession=88518800, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  53422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53422, CurrentID=0, Status='live', Symbol='Ybx2', Aliases='Msy2', description='Y box protein 2', other_designations='FRGY2 homolog|Y-box protein MSY2|germ cell-specific Y-box-binding protein', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69935835, end_position_on_the_genomic_accession=69941604, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  53424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53424, CurrentID=0, Status='live', Symbol='Tsnax', Aliases='R74833, Trax', description='translin-associated factor X', other_designations='', map_location='8 73.1 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125012997, end_position_on_the_genomic_accession=125034192, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  53598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53598, CurrentID=0, Status='live', Symbol='Dctn3', Aliases='p24', description='dynactin 3', other_designations='dctn3|dynactin light chain (P24)|dynactin light chain p24|p24 dynactin light chain', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41714798, end_position_on_the_genomic_accession=41723163, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  53599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53599, CurrentID=0, Status='live', Symbol='Cd164', Aliases='A115, A24, AA589639, AW540279, MGC-24, MSSP', description='CD164 antigen', other_designations='MGC-24v|MUC-24|c-Myc binding protein|endolyn|multi-glycosylated core protein 24', map_location='10 22.46 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41519500, end_position_on_the_genomic_accession=41531042, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  53600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53600, CurrentID=0, Status='live', Symbol='Timm23', Aliases='Tim23', description='translocase of inner mitochondrial membrane 23', other_designations='translocase of inner mitochondrial membrane 23 homolog', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32180166, end_position_on_the_genomic_accession=32201891, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  53601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53601, CurrentID=0, Status='live', Symbol='Pcdh12', Aliases='Pcdh14, VE-cad-2', description='protocadherin 12', other_designations='VE-cadherin-2|protocadherin 14|vascular cadherin-2|vascular endothelial cadherin-2', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38267092, end_position_on_the_genomic_accession=38284401, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  53602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53602, CurrentID=0, Status='live', Symbol='Hpcal1', Aliases='NVP-3, Nvp3, VILIP3, Vnsl3, Vsnl3', description='hippocalcin-like 1', other_designations='NVL-3|VILIP-3|hippocalcin-like 1; visinin like 3; neural visinin-like 3; neural visinin-like protein 3|neural visinin-like 3|neural visinin-like protein 3|visinin like 3|visinin-like protein 3', map_location='12 A1.1-A1.2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=17690750, end_position_on_the_genomic_accession=17791944, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  53603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53603, CurrentID=0, Status='live', Symbol='Tslp', Aliases='', description='thymic stromal lymphopoietin', other_designations='thymic stroma-derived lymphopoietin', map_location='18 18.12 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32815383, end_position_on_the_genomic_accession=32819799, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  53604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53604, CurrentID=0, Status='live', Symbol='Zpbp', Aliases='4930486K01Rik, Iam38, Sp38', description='zona pellucida binding protein', other_designations='', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=11279624, end_position_on_the_genomic_accession=11463630, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  53605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53605, CurrentID=0, Status='live', Symbol='Nap1l1', Aliases='AA407126, AI256722, D10Ertd68e, NAP-1', description='nucleosome assembly protein 1-like 1', other_designations='NAP-1-related protein|brain protein DN38|nucleosome assembly protein-1', map_location='10 58.67 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=111473192, end_position_on_the_genomic_accession=111498150, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  53607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53607, CurrentID=0, Status='live', Symbol='Snrpa', Aliases='C430021M15Rik, Rnu1a-1, Rnu1a1, U1-A, U1A', description='small nuclear ribonucleoprotein polypeptide A', other_designations='U1 snRNP A|protein A of the U1 small nuclear RNA', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27187006, end_position_on_the_genomic_accession=27196271, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  53608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53608, CurrentID=0, Status='live', Symbol='Map3k6', Aliases='Ask2, MAPKKK6, MEKK6', description='mitogen-activated protein kinase kinase kinase 6', other_designations='apoptosis signal-regulating kinase 2|map3k6|mitogen activated protein kinase kinase kinase 6', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133239669, end_position_on_the_genomic_accession=133252929, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  53609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53609, CurrentID=0, Status='live', Symbol='Clasrp', Aliases='Clasp, Sfrs16, Srsf16, Swap2', description='CLK4-associating serine/arginine rich protein', other_designations='Clk4 associating SR-related protein|serine/arginine-rich splicing factor 16|splicing factor, arginine/serine-rich 16 (suppressor-of-white-apricot homolog, Drosophila)|suppressor of white-apricot homolog 2', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19581038, end_position_on_the_genomic_accession=19604517, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  53610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53610, CurrentID=0, Status='live', Symbol='Nono', Aliases='AA407051, AV149256, NRB54, P54NRB, nonA', description='non-POU-domain-containing, octamer binding protein', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101429560, end_position_on_the_genomic_accession=101448593, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  53611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53611, CurrentID=0, Status='live', Symbol='Vti1a', Aliases='1110014F16Rik, 1110018K19Rik, 4921537J05Rik, MVti1, MVti1a, Vti1, Vti1-rp2', description='vesicle transport through interaction with t-SNAREs 1A', other_designations='vesicle transport through interaction with t-SNAREs 1 homolog|vesicle transport v-SNARE protein Vti1-like 2', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=55316057, end_position_on_the_genomic_accession=55627461, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  53612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53612, CurrentID=0, Status='live', Symbol='Vti1b', Aliases='AU015348, GES30, MVti1b, SNARE, Vti1-rp1', description='vesicle transport through interaction with t-SNAREs 1B', other_designations='A novel SNARE|Soluble N-ethylmaleimide-sensitive factor-attachment protein receptor|vesicle transport through interaction with t-SNAREs 1B homolog|vesicle transport v-SNARE protein Vti1-like 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79156017, end_position_on_the_genomic_accession=79172688, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  53614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53614, CurrentID=0, Status='live', Symbol='Reck', Aliases='St15, mRECK', description='reversion-inducing-cysteine-rich protein with kazal motifs', other_designations='Reversion-inducing cysteine-rich protein with Kazal motifs precursor (mRECK)|membrane-anchored glycoprotein RECK|suppression of tumorigenicity 15', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43875530, end_position_on_the_genomic_accession=43944806, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  53617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53617, CurrentID=0, Status='live', Symbol='Krt35', Aliases='AI662400, Ha5, Krt1-24, Krtha5', description='keratin 35', other_designations='K35|hair keratin acidic 5|hair keratin, type I Ha5|keratin complex 1, acidic, gene 24|keratin complex 1, gene 24|keratin, hair, acidic, 5|keratin-35|krt35', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100092192, end_position_on_the_genomic_accession=100096224, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  53618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53618, CurrentID=0, Status='live', Symbol='Fut8', Aliases='', description='fucosyltransferase 8', other_designations='GDP-L-Fuc:N-acetyl-beta-D-glucosaminide alpha1,6-fucosyltransferase|GDP-fucose--glycoprotein fucosyltransferase|alpha (1,6) fucosyltransferase|alpha-1,6-fucosyltransferase|alpha1-6FucT|glycoprotein 6-alpha-L-fucosyltransferase', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=77237437, end_position_on_the_genomic_accession=77475998, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  53619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53619, CurrentID=0, Status='live', Symbol='Blcap', Aliases='AI462828, Bc10', description='bladder cancer associated protein', other_designations='Blcapbladder cancer associated protein homolog|bladder cancer 10 kDa protein|bladder cancer associated protein homolog|bladder cancer related protein', map_location='2 78.4 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157556362, end_position_on_the_genomic_accession=157566952, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  53620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53620, CurrentID=0, Status='live', Symbol='Vamp5', Aliases='AF119384, Camp', description='vesicle-associated membrane protein 5', other_designations='VAMP-5|myobrevin', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72368049, end_position_on_the_genomic_accession=72380468, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  53621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53621, CurrentID=0, Status='live', Symbol='Cnot4', Aliases='Not4, Not4h, Not4hp', description='CCR4-NOT transcription complex, subunit 4', other_designations='CCR4-associated factor 4|E3 ubiquitin-protein ligase CNOT4|negative regulator of transcription 4|potential transcriptional repressor NOT4Hp|transcriptional repressor NOT4Hp', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=35022065, end_position_on_the_genomic_accession=35133737, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  53622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53622, CurrentID=0, Status='live', Symbol='Krt85', Aliases='AA589597, BB007925, Hb5, Krt2-18', description='keratin 85', other_designations='K85|keratin complex 2, basic, gene 18|keratin-85|type II hair keratin Hb5|type-II keratin Kb25', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53623, CurrentID=0, Status='live', Symbol='Gria3', Aliases='2900064I19Rik, GluA3, GluR-C, GluR-K3, Glur-3, Glur3, Gluralpha3', description='glutamate receptor, ionotropic, AMPA3 (alpha 3)', other_designations='AMPA-selective glutamate receptor 3|glutamate receptor channel alpha3 subunit', map_location='X 23.19 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=41401119, end_position_on_the_genomic_accession=41678601, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  53624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53624, CurrentID=0, Status='live', Symbol='Cldn7', Aliases='', description='claudin 7', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69964779, end_position_on_the_genomic_accession=69967886, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  53625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53625, CurrentID=0, Status='live', Symbol='B3gnt2', Aliases='AA408337, AA409404, AW260308, B3Galt6, B3gnt1, BGnT-1, BGnT-2, BGnT1, BGnT2, Beta3gnt', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 2', other_designations='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 6|UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 1|beta-1,3-Gn-T1|beta-1,3-Gn-T2|beta-1,3-N-acetylglucosaminyltransferase 1|beta-1,3-N-acetylglucosaminyltransferase 2|beta3Gn-T1|beta3Gn-T2', map_location='11 14.22 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=22834739, end_position_on_the_genomic_accession=22860336, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  53626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53626, CurrentID=0, Status='live', Symbol='Insm1', Aliases='IA-1', description='insulinoma-associated 1', other_designations='insulinoma-associated protein (IA-1)|zinc finger protein IA-1', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=146221997, end_position_on_the_genomic_accession=146225020, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  53627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53627, CurrentID=0, Status='live', Symbol='Porcn', Aliases='2410004O13Rik, AW045557, DXHXS7465e, Mg61, Mporc, Ppn, mMg61, porc', description='porcupine homolog (Drosophila)', other_designations='probable protein-cysteine N-palmitoyltransferase porcupine', map_location='X 3.7 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8193848, end_position_on_the_genomic_accession=8206551, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  53674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53674, CurrentID=0, Status='live', Symbol='D6Nds5', Aliases='D4Nds1', description='DNA segment, Chr 6, Nuffield Department of Surgery 5', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53693, CurrentID=0, Status='live', Symbol='D2Mit1', Aliases='D11S14h', description='DNA segment, Chr 2, Massachusetts Institute of Technology 1', other_designations='', map_location='2 2.23 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53695, CurrentID=0, Status='live', Symbol='D3Tu51', Aliases='D0Tu51, D17Tu51', description='DNA segment, Chr 3, Tubingen 51', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53697, CurrentID=0, Status='live', Symbol='D6Nds4', Aliases='D0Nds27', description='DNA segment, Chr 6, Nuffield Department of Surgery 4', other_designations='', map_location='6 20.9 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53704, CurrentID=0, Status='live', Symbol='D12Nyu1', Aliases='D12-1', description='DNA segment, Chr 12, NYU 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53705, CurrentID=0, Status='live', Symbol='D12Nyu2', Aliases='D12-2', description='DNA segment, Chr 12, NYU 2', other_designations='', map_location='12 10.96 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53706, CurrentID=0, Status='live', Symbol='D12Nyu3', Aliases='D12-3', description='DNA segment, Chr 12, NYU 3', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53707, CurrentID=0, Status='live', Symbol='D12Nyu4', Aliases='D12-4', description='DNA segment, Chr 12, NYU 4', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53708, CurrentID=0, Status='live', Symbol='D12Nyu5', Aliases='D12-5', description='DNA segment, Chr 12, NYU 5', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53716, CurrentID=0, Status='live', Symbol='D4Smh6b', Aliases='D4Pas1', description="DNA segment, Chr 4, St. Mary's Hospital 6b", other_designations='', map_location='4 86.17 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53718, CurrentID=0, Status='live', Symbol='D5H4S76', Aliases='D4S76h', description='DNA segment, Chr 5, human D4S76', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53722, CurrentID=0, Status='live', Symbol='D7Cwr18', Aliases='D7Cw18, D7Tm18', description='DNA segment, Chr 7, Case Western Reserve 18', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53726, CurrentID=0, Status='live', Symbol='DXHXS32', Aliases='DXPas7, DXS32h', description='DNA segment, Chr X, human DXS32', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53732, CurrentID=0, Status='live', Symbol='D15Nds2', Aliases='D1Nds3', description='DNA segment, Chr 15, Nuffield Department of Surgery 2', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53733, CurrentID=0, Status='live', Symbol='D17Aus9', Aliases='D17Au9', description='DNA segment, Chr 17, Austin 9', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53738, CurrentID=0, Status='live', Symbol='DXPas23', Aliases='DXPas16, DXPas17', description='DNA segment, Chr X, Pasteur Institute 23', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53749, CurrentID=0, Status='live', Symbol='D12Nds1', Aliases='D0Nds19', description='DNA segment, Chr 12, Nuffield Department of Surgery 1', other_designations='', map_location='12 25.97 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53752, CurrentID=0, Status='live', Symbol='D9Ncvs53', Aliases='Irlgs3', description='DNA segment, Chr 9, National Cardiovascular Center, Shionogi 53', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53757, CurrentID=0, Status='live', Symbol='D17Boy1', Aliases='B144, D17Slk1', description='DNA segment, Chr 17, Boyse 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53760, CurrentID=0, Status='live', Symbol='D13Mit31', Aliases='D0Mit2', description='DNA segment, Chr 13, Massachusetts Institute of Technology 31', other_designations='', map_location='13 64.12 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53761, CurrentID=0, Status='live', Symbol='Prrc2a', Aliases='3110039B05Rik, Bat-2, Bat2, D17H6S51E, G2, Wbp12, mKIAA4178', description='proline-rich coiled-coil 2A', other_designations='HLA-B associated transcript 2|HLA-B-associated transcript 2|large proline-rich protein BAT2|proline-rich and coiled-coil-containing protein 2A', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35149076, end_position_on_the_genomic_accession=35165162, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  53766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53766, CurrentID=0, Status='live', Symbol='DXHXS101', Aliases='C131B2R, DXPas20', description='DNA segment, Chr X, human DXS101', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53777, CurrentID=0, Status='live', Symbol='D17Leh26', Aliases='D17Tu26', description='DNA segment, Chr 17, Lehrach 26', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53778, CurrentID=0, Status='live', Symbol='D17Leh27', Aliases='D17Tu27', description='DNA segment, Chr 17, Lehrach 27', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53782, CurrentID=0, Status='live', Symbol='DXHXF34', Aliases='DXF34h, DXF34h-rs1, DXF34h-rs2', description='DNA segment, Chr X, human DXF34', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53791, CurrentID=0, Status='live', Symbol='Tlr5', Aliases='', description='toll-like receptor 5', other_designations='', map_location='1 86.12 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182954788, end_position_on_the_genomic_accession=182976044, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  53792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53792, CurrentID=0, Status='live', Symbol='D16H21S16', Aliases='D21S16h', description='DNA segment, Chr 16, human D21S16', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53799, CurrentID=0, Status='live', Symbol='DXHXS144E', Aliases='DXPas6, DXS144Eh', description='DNA segment, Chr X, human DXS144E', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53808, CurrentID=0, Status='live', Symbol='D7H19F11S1', Aliases='D19F11S1h', description='DNA segment, Chr 7, human D19F11S1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53814, CurrentID=0, Status='live', Symbol='Oaz3', Aliases='AZ3, Az-3, Oaz-t', description='ornithine decarboxylase antizyme 3', other_designations='ODC-Az 3|antizyme 3|testis specific ornithine decarboxylase antizyme', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94433387, end_position_on_the_genomic_accession=94436651, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  53817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53817, CurrentID=0, Status='live', Symbol='Ddx39b', Aliases='0610030D10Rik, AI428441, Bat-1, Bat1, Bat1a, D17H6S81E, D17H6S81E-1, D6S81Eh', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39B', other_designations='56 kDa U2AF65-associated protein|DEAD box protein UAP56|HLA-B-associated transcript 1A|nuclear RNA helicase Bat1|spliceosome RNA helicase Bat1', map_location='17 18.6 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35241746, end_position_on_the_genomic_accession=35253707, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  53856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53856, CurrentID=0, Status='live', Symbol='Prg3', Aliases='MBP2, MBPH', description='proteoglycan 3', other_designations='eosinophil major basic protein 2|major basic protein 2|prg3', map_location='2 49.45 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84988196, end_position_on_the_genomic_accession=84996559, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  53857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53857, CurrentID=0, Status='live', Symbol='Tuba8', Aliases='', description='tubulin, alpha 8', other_designations='alpha tubulin 8', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121210770, end_position_on_the_genomic_accession=121226097, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  53858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53858, CurrentID=0, Status='live', Symbol='Rwdd2b', Aliases='ORF5', description='RWD domain containing 2B', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87433331, end_position_on_the_genomic_accession=87440592, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  53859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53859, CurrentID=0, Status='live', Symbol='Map3k14', Aliases='Nik, aly', description='mitogen-activated protein kinase kinase kinase 14', other_designations='NF kappa B-inducing kinase|NF-kappa-beta-inducing kinase|Nfkb inducing kinase|alymphoplasia|protein kinase|serine/threonine-protein kinase NIK', map_location='11 66.8 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103219764, end_position_on_the_genomic_accession=103267401, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  53860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53860, CurrentID=0, Status='live', Symbol='Sept9', Aliases='MSF1, Msf, PNUTL4, Sint1', description='septin 9', other_designations='MLL septin-like fusion|SL3-3 integration site 1 protein', map_location='11 82.61 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117199661, end_position_on_the_genomic_accession=117362325, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  53861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53861, CurrentID=0, Status='live', Symbol='Zranb2', Aliases='AI227013, Zfp265, Zis, Znf265', description='zinc finger, RAN-binding domain containing 2', other_designations='zinc finger protein 265|zinc finger, splicing', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=157534397, end_position_on_the_genomic_accession=157548339, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  53867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53867, CurrentID=0, Status='live', Symbol='Col5a3', Aliases='', description='collagen, type V, alpha 3', other_designations='Pro-alpha3(V)|procollagen, type V, alpha 3', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20770050, end_position_on_the_genomic_accession=20815117, orientation='minus', exon_count=68, OMIM='', no_hdr0=''),
  53868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53868, CurrentID=0, Status='live', Symbol='Rab25', Aliases='AV012978', description='RAB25, member RAS oncogene family', other_designations='', map_location='3 E3-F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88542028, end_position_on_the_genomic_accession=88548279, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  53869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53869, CurrentID=0, Status='live', Symbol='Rab11a', Aliases='', description='RAB11A, member RAS oncogene family', other_designations='rab-11', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64715300, end_position_on_the_genomic_accession=64737756, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  53870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53870, CurrentID=0, Status='live', Symbol='Cntn6', Aliases='NB-3', description='contactin 6', other_designations='mNB-3|neural recognition molecule NB-3', map_location='6 E2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=104492704, end_position_on_the_genomic_accession=104863406, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  53871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53871, CurrentID=0, Status='live', Symbol='Pkd2l2', Aliases='TRPP5', description='polycystic kidney disease 2-like 2', other_designations='Polycystin - L2|polycystin-2L2|polycystin-L2', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34409152, end_position_on_the_genomic_accession=34442789, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  53872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53872, CurrentID=0, Status='live', Symbol='Caprin1', Aliases='AL022980, Caprin-1, Gpiap, Gpiap1, Mmgpip137, P137, rng105', description='cell cycle associated protein 1', other_designations='GPI-anchored membrane protein 1|GPI-anchored protein p137|GPI-p137|RNA granule protein 105|cytoplasmic activation- and proliferation-associated protein 1|cytoplasmic activation/proliferation-associated protein 1|membrane component chromosome 11 surface marker 1|p137GPI', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103762945, end_position_on_the_genomic_accession=103797640, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  53873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53873, CurrentID=0, Status='live', Symbol='Ear7', Aliases='mR7', description='eosinophil-associated, ribonuclease A family, member 7', other_designations='eosinophil-associated ribonuclease 7|murine ribonuclease 7', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53878, CurrentID=0, Status='live', Symbol='Svs2', Aliases='BB115391, Semg1, SvsII', description='seminal vesicle secretory protein 2', other_designations='semenoclotin|seminal vesicle protein, secretion 2', map_location='2 84.82 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164235929, end_position_on_the_genomic_accession=164238341, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  53880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53880, CurrentID=0, Status='live', Symbol='Naip7', Aliases='Birc1g, Naip-rs4B, Naip6', description='NLR family, apoptosis inhibitory protein 7', other_designations='neuronal apoptosis inhibitory protein 6|neuronal apoptosis inhibitory protein 7', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.4', start_position_on_the_genomic_accession=101383372, end_position_on_the_genomic_accession=101417945, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  53881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53881, CurrentID=0, Status='live', Symbol='Slc5a3', Aliases='AA623876, BF642829, Smit1', description='solute carrier family 5 (inositol transporters), member 3', other_designations='Na(+)/myo-inositol cotransporter|sodium myo-inositol transporter 1|sodium/myo-inositol transporter 1|solute carrier family 5 member 3|solute carrier family 5, member 3|solute carrier family 5, member alpha 3', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92058322, end_position_on_the_genomic_accession=92087473, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  53883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53883, CurrentID=0, Status='live', Symbol='Celsr2', Aliases='Adgrc2, EGFL2, Flamingo1, flamingo, mKIAA0279, mfmi1', description='cadherin, EGF LAG seven-pass G-type receptor 2', other_designations='cadherin, EGF LAG seven-pass G-type receptor 2 (flamingo homolog, Drosophila)|flamingo homolog', map_location='3 47.02 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108390848, end_position_on_the_genomic_accession=108416026, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  53885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53885, CurrentID=0, Status='live', Symbol='Nphp1', Aliases='', description='nephronophthisis 1 (juvenile) homolog (human)', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127740732, end_position_on_the_genomic_accession=127788914, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  53886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53886, CurrentID=0, Status='live', Symbol='Cdkl2', Aliases='5330436L21Rik, AI505225, KKIAMRE, Kkm', description='cyclin-dependent kinase-like 2 (CDC2-related kinase)', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92006074, end_position_on_the_genomic_accession=92043906, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  53890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53890, CurrentID=0, Status='live', Symbol='Sart3', Aliases='AU045857, mKIAA0156', description='squamous cell carcinoma antigen recognized by T cells 3', other_designations='SART-3|mSART-3|tumor-rejection antigen SART3', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113742444, end_position_on_the_genomic_accession=113772505, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  53892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53892, CurrentID=0, Status='live', Symbol='Ppm1d', Aliases='AV338790, Wip1', description='protein phosphatase 1D magnesium-dependent, delta isoform', other_designations='PP2C-delta|p53-induced protein phosphatase 1|protein phosphatase 2C isoform delta|protein phosphatase magnesium-dependent 1 delta', map_location='11 51.34 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=85311254, end_position_on_the_genomic_accession=85347071, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  53893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53893, CurrentID=0, Status='live', Symbol='Nudt5', Aliases='', description='nudix (nucleoside diphosphate linked moiety X)-type motif 5', other_designations='8-oxo-dGDP phosphatase|nucleoside diphosphate-linked moiety X motif 5|nudix motif 5', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5845034, end_position_on_the_genomic_accession=5868745, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  53895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53895, CurrentID=0, Status='live', Symbol='Clpp', Aliases='AU019820, D17Wsu160e', description='caseinolytic mitochondrial matrix peptidase proteolytic subunit', other_designations='ClpP caseinolytic peptidase, ATP-dependent, proteolytic subunit|caseinolytic peptidase, ATP-dependent, proteolytic subunit homolog|caseinolytic protease, ATP-dependent, proteolytic subunit homolog|endopeptidase Clp|putative ATP-dependent Clp protease proteolytic subunit, mitochondrial', map_location='17 29.61 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56990264, end_position_on_the_genomic_accession=56996371, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  53896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53896, CurrentID=0, Status='live', Symbol='Slc7a10', Aliases='AL024237, Asc-1, D7Bwg0847e', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 10', other_designations='D-serine transporter|solute carrier family 7 member 10', map_location='7 B1-B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35186357, end_position_on_the_genomic_accession=35201114, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  53897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53897, CurrentID=0, Status='live', Symbol='Gal3st1', Aliases='Cst, Gcst', description='galactose-3-O-sulfotransferase 1', other_designations="3'-phosphoadenosine-5'-phosphosulfate:Galcer sulfotransferase|3'-phosphoadenylylsulfate-galactosylceramide 3'-sulfotransferase|Galcer sulfotransferase|cerebroside sulfotransferase", map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3983636, end_position_on_the_genomic_accession=3999328, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  53901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53901, CurrentID=0, Status='live', Symbol='Rcan2', Aliases='Csp2, Dscr1l1, MCIP2, ZAKI-4', description='regulator of calcineurin 2', other_designations='Down syndrome critical region gene 1-like 1|calcineurin inhibitory protein ZAKI-4|down syndrome candidate region 1-like protein 1|myocyte-enriched calcineurin interactin protein 2|myocyte-enriched calcineurin-interacting protein 2', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43801281, end_position_on_the_genomic_accession=44039516, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  53902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53902, CurrentID=0, Status='live', Symbol='Rcan3', Aliases='AU041093, Csp3, Dscr1l2', description='regulator of calcineurin 3', other_designations='Down syndrome critical region gene 1-like 2|MCIP3|down syndrome candidate region 1-like protein 2|myocyte-enriched calcineurin-interacting protein 3', map_location='4 67.59 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135412309, end_position_on_the_genomic_accession=135433805, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  53914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53914, CurrentID=0, Status='live', Symbol='Tallbcr', Aliases='', description='T cell acute lymphoblastic leukemia breakpoint cluster region', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  53945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53945, CurrentID=0, Status='live', Symbol='Slc40a1', Aliases='Dusg, Fpn1, IREG1, MTP, MTP1, Ol5, Pcm, Slc11a3, Slc39a1', description='solute carrier family 40 (iron-regulated transporter), member 1', other_designations='SLC11A3 iron transporter|duodenal-specific|ferroportin 1|ferroportin-1|iron-regulated transporter 1|metal transporter protein 1|metal transporting protein 1|polycythaemia|solute carrier family 11 (proton-coupled divalent metal ion transporters), member 3|solute carrier family 39 (iron-regulated transporter), member 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=45908070, end_position_on_the_genomic_accession=45926535, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  53951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53951, CurrentID=0, Status='live', Symbol='Gpatch11', Aliases='2310002B06Rik, C80922, Ccdc75, L26697', description='G patch domain containing 11', other_designations='coiled-coil domain containing 75|coiled-coil domain-containing protein 75', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78835516, end_position_on_the_genomic_accession=78848308, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  53970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53970, CurrentID=0, Status='live', Symbol='Rfx5', Aliases='', description='regulatory factor X, 5 (influences HLA class II expression)', other_designations='regulatory factor X 5', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94951957, end_position_on_the_genomic_accession=94961561, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  53972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53972, CurrentID=0, Status='live', Symbol='Ngef', Aliases='Tims2, ephexin', description='neuronal guanine nucleotide exchange factor', other_designations='eph-interacting exchange protein|transforming immortalised mammary epithelial sequence 2', map_location='1 44.42 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87476829, end_position_on_the_genomic_accession=87573870, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  53973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53973, CurrentID=0, Status='live', Symbol='Cyp3a41a', Aliases='Cyp3a41', description='cytochrome P450, family 3, subfamily a, polypeptide 41A', other_designations='cytochrome P450, family 3, subfamily a, polypeptide 41|cytochrome P450, steroid inducible 3a41|steroid inducible', map_location='5 85.17 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145694049, end_position_on_the_genomic_accession=145720136, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  53975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53975, CurrentID=0, Status='live', Symbol='Ddx20', Aliases='GEMIN3, dp103', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 20', other_designations='DEAD box protein 20|DEAD box protein DP 103|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 20|ROSF-1|component of gems 3|gemin-3|regulator of steroidogenic factor 1', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105678270, end_position_on_the_genomic_accession=105688113, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  53978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=53978, CurrentID=0, Status='live', Symbol='Lpar2', Aliases='Edg4, IPA2, LPA2', description='lysophosphatidic acid receptor 2', other_designations='LPA receptor 2|LPA-2|endothelial differentiation, lysophosphatidic acid G-protein-coupled receptor 4|lysophosphatidic acid receptor Edg-4', map_location='8 33.91 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69822565, end_position_on_the_genomic_accession=69831102, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54003, CurrentID=0, Status='live', Symbol='Nell2', Aliases='A330108N19Rik, R75516, mel91', description='NEL-like 2', other_designations='NEL-like protein 2|nel-like 2 homolog', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=95219440, end_position_on_the_genomic_accession=95528758, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  54004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54004, CurrentID=0, Status='live', Symbol='Diaph2', Aliases='Dia3, Diap2, Drf2, E430022I22Rik', description='diaphanous related formin 2', other_designations='diaphanous homolog 2|diaphanous-related formin-2|mDia3 isoform', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=129749668, end_position_on_the_genomic_accession=130465834, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  54006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54006, CurrentID=0, Status='live', Symbol='Deaf1', Aliases='AU042387, C230009B13Rik, NUDR', description='deformed epidermal autoregulatory factor 1 (Drosophila)', other_designations='nuclear DEAF-1-related transcriptional regulator|suppressin', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141297176, end_position_on_the_genomic_accession=141338870, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  54120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54120, CurrentID=0, Status='live', Symbol='Gipc2', Aliases='2200002N01Rik, AU021850, Semcap2', description='GIPC PDZ domain containing family, member 2', other_designations='SEMCAP-2|semaF cytoplasmic domain associated protein 2|semaF cytoplasmic domain-associated protein 2', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152093839, end_position_on_the_genomic_accession=152165900, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54122, CurrentID=0, Status='live', Symbol='Uevld', Aliases='8430408E05Rik, Attp', description='UEV and lactate/malate dehyrogenase domains', other_designations='EV and lactate/malate dehydrogenase domain-containing protein|UEV-3|signaling molecule ATTP', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46923216, end_position_on_the_genomic_accession=46958530, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  54123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54123, CurrentID=0, Status='live', Symbol='Irf7', Aliases='', description='interferon regulatory factor 7', other_designations='IRF-7', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141263183, end_position_on_the_genomic_accession=141266499, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54124, CurrentID=0, Status='live', Symbol='Cks1b', Aliases='2410005G18Rik, 2610005D03Rik, AA407784, Cks1, sid1334', description='CDC28 protein kinase 1b', other_designations='CKS-1|cyclin-dependent kinase regulatory subunit 1|sid 1334', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89415472, end_position_on_the_genomic_accession=89418291, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54125, CurrentID=0, Status='live', Symbol='Polm', Aliases='AI462166, B230309I03Rik, Tdt-N', description='polymerase (DNA directed), mu', other_designations='DNA polymerase mu|DNA-directed DNA polymerase mu|pol Mu|terminal transferase', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5827860, end_position_on_the_genomic_accession=5838341, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  54126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54126, CurrentID=0, Status='live', Symbol='Arhgef7', Aliases='Cool, PIX, Pak3bp, betaPix, betaPix-b, betaPix-c, cool-1, mKIAA0142, p85Cool1, p85SPR', description='Rho guanine nucleotide exchange factor (GEF7)', other_designations='PAK-interacting exchange factor beta|Pak3 binding protein|beta-Pix', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11728105, end_position_on_the_genomic_accession=11835219, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  54127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54127, CurrentID=0, Status='live', Symbol='Rps28', Aliases='', description='ribosomal protein S28', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33823036, end_position_on_the_genomic_accession=33824557, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54128, CurrentID=0, Status='live', Symbol='Pmm2', Aliases='AI585868, C86848', description='phosphomannomutase 2', other_designations='PMM 2', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8637707, end_position_on_the_genomic_accession=8657552, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54130, CurrentID=0, Status='live', Symbol='Actr1a', Aliases='Arp1, alpha-Arp1', description='ARP1 actin-related protein 1A, centractin alpha', other_designations='ARP1 actin-related protein 1 homolog A, centractin alpha|Arp1 centractin alpha|Arp1 contractin alpha|actin-RPV|actin-related protein 1 alpha|actin-related protein 1 homolog A|centractin|centrosome-associated actin homolog', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46376814, end_position_on_the_genomic_accession=46395735, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  54131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54131, CurrentID=0, Status='live', Symbol='Irf3', Aliases='C920001K05Rik, IRF-3', description='interferon regulatory factor 3', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44997648, end_position_on_the_genomic_accession=45002848, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54132, CurrentID=0, Status='live', Symbol='Pdlim1', Aliases='CLP36, Clim1, mClim1', description='PDZ and LIM domain 1 (elfin)', other_designations='C-terminal LIM domain protein 1|LIM domain protein CLP-36|carboxyl terminal LIM domain protein 1|elfin', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40222239, end_position_on_the_genomic_accession=40271616, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54135, CurrentID=0, Status='live', Symbol='Lsr', Aliases='ILDR3, Lisch7', description='lipolysis stimulated lipoprotein receptor', other_designations='lipolysis stimulated receptor|lipolysis-stimulated receptor|liver-specific bHLH-Zip transcription factor|liver-specific gene on mouse chromosome 7 protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30957770, end_position_on_the_genomic_accession=30973469, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54137, CurrentID=0, Status='live', Symbol='Acrbp', Aliases='OY-TES-1, sp32', description='proacrosin binding protein', other_designations='acrbp|proacrosin-binding protein (sp32)|proacrosin-binding protein sp32', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125049927, end_position_on_the_genomic_accession=125063265, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54138, CurrentID=0, Status='live', Symbol='Atxn10', Aliases='AI325283, C77170, E46, Sca10, TEG-169, Tex169', description='ataxin 10', other_designations='brain protein E46|spinocerebellar ataxia 10|spinocerebellar ataxia type 10 protein homolog|testis expressed gene 169', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85336381, end_position_on_the_genomic_accession=85463836, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  54139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54139, CurrentID=0, Status='live', Symbol='Irf6', Aliases='AI876454, E230028I05Rik, mirf6', description='interferon regulatory factor 6', other_designations='IRF-6', map_location='1 97.6 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193153112, end_position_on_the_genomic_accession=193172036, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54140, CurrentID=0, Status='live', Symbol='Avpr1a', Aliases='AVPR, Avpr1, V1a, V1aR', description='arginine vasopressin receptor 1A', other_designations='AVPR V1a|V1a arginine vasopressin receptor|antidiuretic hormone receptor 1a|vascular/hepatic-type arginine vasopressin receptor', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=122448499, end_position_on_the_genomic_accession=122453453, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54141, CurrentID=0, Status='live', Symbol='Spag5', Aliases='AI874642, D11Bhm180e, Deepest, MAP126, Mastrin, S17', description='sperm associated antigen 5', other_designations='astrin|microtubule-associated protein|mitotic spindle-associated protein p126', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78301591, end_position_on_the_genomic_accession=78322454, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  54150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54150, CurrentID=0, Status='live', Symbol='Rdh7', Aliases='AI194929, CRAD2', description='retinol dehydrogenase 7', other_designations='CRAD-2|cis-retinol/3alpha-hydroxysterol short-chain dehydrogenase isozyme 2|cis-retinol/androgen dehydrogenase type 2', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127884027, end_position_on_the_genomic_accession=127889247, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54151, CurrentID=0, Status='live', Symbol='Cyhr1', Aliases='1110031M01Rik, AU042374, Chrp', description='cysteine and histidine rich 1', other_designations='cysteine and histidine-rich cytoplasmic protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76643395, end_position_on_the_genomic_accession=76660208, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  54152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54152, CurrentID=0, Status='live', Symbol='Dnal4', Aliases='D15Ertd424e, Dnalc4', description='dynein, axonemal, light chain 4', other_designations='dynein, axon, light chain 4|outer arm dynein light chain 4', map_location='15 37.85 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79761449, end_position_on_the_genomic_accession=79774469, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54153, CurrentID=0, Status='live', Symbol='Rasa4', Aliases='AA793972, AW112107, BB079060, CAPRI, GAPL', description='RAS p21 protein activator 4', other_designations='Ca2+promoted Ras inactivator|calcium-promoted RAS inactivator|rasGAP-activating-like protein 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136083916, end_position_on_the_genomic_accession=136111860, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  54156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54156, CurrentID=0, Status='live', Symbol='Egfl6', Aliases='Maeg', description='EGF-like-domain, multiple 6', other_designations='EGF-like protein 6|MAM and EGF domains-containing gene protein', map_location='X 77.66 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=166523007, end_position_on_the_genomic_accession=166585716, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54159, CurrentID=0, Status='live', Symbol='Rnase2b', Aliases='Ear5, Raf4, mR5', description='ribonuclease, RNase A family, 2B (liver, eosinophil-derived neurotoxin)', other_designations='MR-5|eosinophil cationic-type ribonuclease 5|eosinophil-associated ribonuclease 5|eosinophil-associated, ribonuclease A family, member 5|murine ribonuclease 5|ribonuclease A f4', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51162260, end_position_on_the_genomic_accession=51163018, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54160, CurrentID=0, Status='live', Symbol='Copg2', Aliases='AW227625, gamma-2-COP', description='coatomer protein complex, subunit gamma 2', other_designations='gamma-2-coat protein', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30809559, end_position_on_the_genomic_accession=30896794, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  54161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54161, CurrentID=0, Status='live', Symbol='Copg1', Aliases='AU019265, BC056168, Copg, D6Ertd71e, D6Wsu16e', description='coatomer protein complex, subunit gamma 1', other_designations='coatomer subunit gamma|gamma-1-COP|gamma-1-coat protein|gamma-COP|gamma-coat protein', map_location='6 39.13 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87887819, end_position_on_the_genomic_accession=87913595, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  54167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54167, CurrentID=0, Status='live', Symbol='Icos', Aliases='AILIM, CCLP, CRP-1, H4, Ly115', description='inducible T cell co-stimulator', other_designations='CD28 and CTLA-4-like protein|CD28-related protein 1|activation-inducible lymphocyte immunomediatory molecule|inducible T-cell co-stimulator', map_location='1 30.6 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60960773, end_position_on_the_genomic_accession=61000322, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54169, CurrentID=0, Status='live', Symbol='Kat6b', Aliases='AI507552, B130044K16Rik, MYST-4, Morf, Myst4, mKIAA0383, qkf, querkopf', description='K(lysine) acetyltransferase 6B', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 4|MYST histone acetyltransferase monocytic leukemia 4|histone acetyltransferase MYST4|histone acetyltransferase querkopf|monocytic leukemia|protein querkopf', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21495310, end_position_on_the_genomic_accession=21672478, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  54170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54170, CurrentID=0, Status='live', Symbol='Rragc', Aliases='AU041672, Gtr2, RAGC, TIB929, YGR163W', description='Ras-related GTP binding C', other_designations='GTPase-interacting protein 2|rag C|rragc|small GTPase homolog|small GTPase, homolog', map_location='4 D', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123917433, end_position_on_the_genomic_accession=123936997, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54188, CurrentID=0, Status='live', Symbol='Cpsf4', Aliases='30kDa, C79664', description='cleavage and polyadenylation specific factor 4', other_designations='CPSF 30 kDa subunit|cleavage and polyadenylation specific factor 4, 30kD subunit|cleavage and polyadenylation specificity factor 30 kDa subunit|clipper homolog|clipper/CPSF 30K|musculus clipper/cleavage and polyadenylation specificity factor 30 kDa subunit', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145167213, end_position_on_the_genomic_accession=145182041, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  54189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54189, CurrentID=0, Status='live', Symbol='Rabep1', Aliases='Rab5ep, rabaptin-5, rabaptin-5alpha', description='rabaptin, RAB GTPase binding effector protein 1', other_designations='RAB5 effector protein|neurocrescin|rabaptin 5', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70844723, end_position_on_the_genomic_accession=70943105, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  54192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54192, CurrentID=0, Status='live', Symbol='Pbsn', Aliases='PB, Prbs', description='probasin', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=77837898, end_position_on_the_genomic_accession=77853500, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54194, CurrentID=0, Status='live', Symbol='Akap8l', Aliases='HAP95, Nakap95', description='A kinase (PRKA) anchor protein 8-like', other_designations='AKAP8-like protein|neighbor of A-kinase anchoring protein 95|neighbor of A-kinase-anchoring protein 95|neighbor of AKAP95', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32321424, end_position_on_the_genomic_accession=32350623, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  54195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54195, CurrentID=0, Status='live', Symbol='Gucy1b3', Aliases='GC-S-beta-1, GCbeta1', description='guanylate cyclase 1, soluble, beta 3', other_designations='GCS-beta-1|GCS-beta-3|NO-sensitive guanylyl cyclase beta 1 subunit|beta 1 sGC|guanylate cyclase soluble subunit beta-3|soluble guanylate cyclase beta-1 subunit|soluble guanylate cyclase small subunit', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=82032004, end_position_on_the_genomic_accession=82074711, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  54196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54196, CurrentID=0, Status='live', Symbol='Pabpn1', Aliases='PAB2, Pabp3, mPABII', description='poly(A) binding protein, nuclear 1', other_designations='PABII|PABP-2|nuclear poly(A)-binding protein 1|poly(A) binding protein II|poly(A)-binding protein 2|poly(A)-binding protein II|polyadenylate-binding nuclear protein 1', map_location='14 27.98 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54894143, end_position_on_the_genomic_accession=54898927, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  54197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54197, CurrentID=0, Status='live', Symbol='Rnf5', Aliases='2410131O05Rik, AA407576, NG2', description='ring finger protein 5', other_designations='', map_location='17 18.18 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34601099, end_position_on_the_genomic_accession=34603561, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54198, CurrentID=0, Status='live', Symbol='Snx3', Aliases='SDP3', description='sorting nexin 3', other_designations='', map_location='10 22.89 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=42502054, end_position_on_the_genomic_accession=42535369, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54199, CurrentID=0, Status='live', Symbol='Ccrl2', Aliases='1810047I05Rik, Ackr5, CCR11, Cmkbr1l2, E01, L-CCR', description='chemokine (C-C motif) receptor-like 2', other_designations='G-protein coupled beta chemokine receptor|chemokine (C-C) receptor 1,-like 2|chemokine (C-C) receptor 1-like 2|chemokine receptor CCR11|lipopolysaccharide inducible C-C chemokine receptor related|lipopolysaccharide-inducible C-C chemokine receptor|lipopolysacharide-inducible C-C chemokine receptor', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111054834, end_position_on_the_genomic_accession=111057518, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54200, CurrentID=0, Status='live', Symbol='Sult2b1', Aliases='AI326997, BB173635, ST2B1, SULT2B', description='sulfotransferase family, cytosolic, 2B, member 1', other_designations='alcohol sulfotransferase|hydroxysteroid sulfotransferase 2|sulfotransferase 2B|sulfotransferase 2B1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45721174, end_position_on_the_genomic_accession=45759555, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54201, CurrentID=0, Status='live', Symbol='Zfp316', Aliases='Emzf1, Mzf3, Znf316', description='zinc finger protein 316', other_designations='kruppel-related zinc finger protein|transcription factor MZF-3', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143249695, end_position_on_the_genomic_accession=143272983, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54204, CurrentID=0, Status='live', Symbol='Sept1', Aliases='Diff6, Pnutl3', description='septin 1', other_designations='differentiation 6 (deoxyguanosine triphosphate triphosphohydrolase)|differentiation protein 6|peanut-like protein 3|protein Diff6', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127214442, end_position_on_the_genomic_accession=127232647, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54208, CurrentID=0, Status='live', Symbol='Arl6ip1', Aliases='AIP-6, AL022945, ARMER, AU042858, Aip-1, Arl6ip, C85138, mKIAA0069', description='ADP-ribosylation factor-like 6 interacting protein 1', other_designations='ADP-ribosylation-like factor 6 interacting protein|ARL-6-interacting protein 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118118890, end_position_on_the_genomic_accession=118129625, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54214, CurrentID=0, Status='live', Symbol='Golga4', Aliases='AI225887, AU019508, Olp-1', description='golgi autoantigen, golgin subfamily a, 4', other_designations='golgin-245|tGolgin-1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=118506195, end_position_on_the_genomic_accession=118582519, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  54215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54215, CurrentID=0, Status='live', Symbol='Cd160', Aliases='AU045688, By55', description='CD160 antigen', other_designations='natural killer cell BY55|natural killer cell receptor BY55', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96796883, end_position_on_the_genomic_accession=96829351, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54216, CurrentID=0, Status='live', Symbol='Pcdh7', Aliases='', description='protocadherin 7', other_designations='BH-protocadherin (brain-heart)', map_location='5 C3-D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=57716456, end_position_on_the_genomic_accession=58133235, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54217, CurrentID=0, Status='live', Symbol='Rpl36', Aliases='', description='ribosomal protein L36', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56613395, end_position_on_the_genomic_accession=56614246, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54218, CurrentID=0, Status='live', Symbol='B3galt4', Aliases='Gal-T2, Galt2', description='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 4', other_designations='UDP-Gal:betaGlcNAc beta 13-galactosyltransferase polypeptide 4|UDP-galactose:beta-N-acetyl-galactosamine-beta-1,3-galactosyltransferase|b3Gal-T4|beta-1,3-GalTase 4|beta3Gal-T4|beta3GalT4|ganglioside galactosyltransferase', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33949912, end_position_on_the_genomic_accession=33951488, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54219, CurrentID=0, Status='live', Symbol='Cd320', Aliases='425O18-1, 8D6, D17Ertd716e, NG29, TCblR, VLDL', description='CD320 antigen', other_designations='putative VLDL lipoprotein receptor|transcobalamin receptor', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33843091, end_position_on_the_genomic_accession=33849774, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54324, CurrentID=0, Status='live', Symbol='Arhgef5', Aliases='2210412D05Rik, AA717842, AW495314, Tim1', description='Rho guanine nucleotide exchange factor (GEF) 5', other_designations='', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43265644, end_position_on_the_genomic_accession=43289320, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  54325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54325, CurrentID=0, Status='live', Symbol='Elovl1', Aliases='AA407424, BB151133, Ssc1', description='elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 1', other_designations='3-keto acyl-CoA synthase Elovl1|ELOVL FA elongase 1|ELOVL fatty acid elongase 1|sequence similarity to Cig30 1|very long chain 3-ketoacyl-CoA synthase 1|very long chain 3-oxoacyl-CoA synthase 1|very-long-chain 3-oxoacyl-CoA synthase 1', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118428081, end_position_on_the_genomic_accession=118432953, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54326, CurrentID=0, Status='live', Symbol='Elovl2', Aliases='AI317360, Ssc2', description='elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 2', other_designations='3-keto acyl-CoA synthase Elovl2|ELOVL FA elongase 2|ELOVL family member 2|ELOVL fatty acid elongase 2|sequence similarity to Cig30 2|very long chain 3-ketoacyl-CoA synthase 2|very long chain 3-oxoacyl-CoA synthase 2|very-long-chain 3-oxoacyl-CoA synthase 2', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41182382, end_position_on_the_genomic_accession=41220403, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  54338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54338, CurrentID=0, Status='live', Symbol='Slc23a2', Aliases='AI844736, NBTL1, SVCT2, Slc23a1, YSPL2, YSPL3, mKIAA0238', description='solute carrier family 23 (nucleobase transporters), member 2', other_designations='Na(+)/L-ascorbic acid transporter 2|SVCT-2|mSVCT2|sodium-dependent vitamin C transporter 2|solute carrier family 23 (nucleobase transporters), member 1|yolk sac permease-like molecule 2', map_location='2 G2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132052496, end_position_on_the_genomic_accession=132145108, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  54342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54342, CurrentID=0, Status='live', Symbol='Gnpnat1', Aliases='AU017428, AU040593, EMeg32, Gpnat1, Gsnpat', description='glucosamine-phosphate N-acetyltransferase 1', other_designations='glucosamine-6-phosphate acetyltransferase|glucosamine-phosphate N-acetyltransferase 1; glucosamine-6-phosphate acetyltransferase; glucosamine-phosphate N-acetyltransferase|phosphoglucosamine acetylase|phosphoglucosamine transacetylase', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45376421, end_position_on_the_genomic_accession=45388956, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54343, CurrentID=0, Status='live', Symbol='Atf7ip', Aliases='2610204M12Rik, AM, Mcaf1', description='activating transcription factor 7 interacting protein', other_designations='ATFa-associated Modulator|MBD1-containing chromatin-associated factor 1|mAM', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136518789, end_position_on_the_genomic_accession=136610850, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  54351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54351, CurrentID=0, Status='live', Symbol='Elp5', Aliases='Derp6, Rai12', description='elongator acetyltransferase complex subunit 5', other_designations='dermal papilla-derived protein 6 homolog|retinoic acid induced 12|retinoic acid-induced protein 12', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69968224, end_position_on_the_genomic_accession=69981294, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  54352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54352, CurrentID=0, Status='live', Symbol='Irx5', Aliases='', description='Iroquois related homeobox 5 (Drosophila)', other_designations='homeodomain protein IRXB2|iroquois homeobox protein 5', map_location='8 44.93 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=92357796, end_position_on_the_genomic_accession=92361456, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54353, CurrentID=0, Status='live', Symbol='Skap2', Aliases='2610021A10Rik, AA960083, BB137539, RA70, SKAP-HOM, Saps, Scap2, mSKAP55R', description='src family associated phosphoprotein 2', other_designations='SKAP55 homolog|SKAP55 homologue|Src-associated adaptor protein|pyk2/RAFTK-associated protein|src family-associated phosphoprotein 2|src kinase-associated phosphoprotein 55-related protein|src-associated adapter protein with PH and SH3 domains', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=51859165, end_position_on_the_genomic_accession=52012549, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  54354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54354, CurrentID=0, Status='live', Symbol='Rassf5', Aliases='1300019G20Rik, Maxp1, Nore1, Nore1A, Nore1B, Rapl', description='Ras association (RalGDS/AF-6) domain family member 5', other_designations='Rap1-binding protein|new ras effector 1|novel ras effector 1|regulator for cell adhesion and polarization enriched in lymphoid tissues', map_location='1 56.91 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131176410, end_position_on_the_genomic_accession=131245178, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54357, CurrentID=0, Status='live', Symbol='Epb41l4b', Aliases='6430543G08Rik, AA589614, AU021054, BB007528, D4Ertd346e, Ehm2, Epb4.1l4b, Lulu2', description='erythrocyte membrane protein band 4.1 like 4b', other_designations='erythrocyte protein band 4.1 like 4b|erythrocyte protein band 4.1-like 4b|erythrocyte protein band 4.1-like 4b; expressed in high-metastatic cells|expressed in high-metastatic cells', map_location='4 31.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56992043, end_position_on_the_genomic_accession=57143168, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  54364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54364, CurrentID=0, Status='live', Symbol='Rpp30', Aliases='Rnasep2, TSG15', description='ribonuclease P/MRP 30 subunit', other_designations='RNase P subunit 2|RNaseP protein p30|ribonuclease P/MRP 30kDa subunit|ribonuclease P2', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36083689, end_position_on_the_genomic_accession=36104773, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54366, CurrentID=0, Status='live', Symbol='Ctnnal1', Aliases='ACRP, AI616177, AW545119, C86009, Catnal1', description='catenin (cadherin associated protein), alpha-like 1', other_designations='alpha-catenin related protein|alpha-catenin-related protein|catenin alpha-like 1|catenin alpha-like protein 1', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56810935, end_position_on_the_genomic_accession=56865211, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  54367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54367, CurrentID=0, Status='live', Symbol='Zfp326', Aliases='5730470H14Rik, ZAN75, Znf326', description='zinc finger protein 326', other_designations='zinc finger protein interacting with mRNPs|zinc finger protein-associated with nuclear matrix of 75 kDa', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105876568, end_position_on_the_genomic_accession=105915818, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  54368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54368, CurrentID=0, Status='live', Symbol='Gp9', Aliases='Cd42, GPIX', description='glycoprotein 9 (platelet)', other_designations='GP-IX', map_location='6 39.13 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87776122, end_position_on_the_genomic_accession=87779786, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54369, CurrentID=0, Status='live', Symbol='Nme6', Aliases='nm23-M6', description='NME/NM23 nucleoside diphosphate kinase 6', other_designations='NDK 6|NDP kinase 6|expressed in non-metastatic cells 6, protein (nucleoside diphosphate kinase)|nm23/nucleoside diphosphate kinase 6|non-metastatic cells 6, protein expressed in (nucleoside-diphosphate kinase)|nucleoside diphosphate kinase type 6', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109832711, end_position_on_the_genomic_accession=109842974, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  54371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54371, CurrentID=0, Status='live', Symbol='Chst2', Aliases='AI428561, AW121776, C130041E03Rik, Chts2, GST-2, GlcNAc6ST, Gn6st', description='carbohydrate sulfotransferase 2', other_designations='N-acetylglucosamine 6-O-sulfotransferase 1|N-acetylglucosamine-6-O-sulfotransferase|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 2|glcNAc6ST-1|gn6st-1', map_location='9 E3.2-3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95400926, end_position_on_the_genomic_accession=95407270, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54373, CurrentID=0, Status='live', Symbol='Prss16', Aliases='AI448615, TSSP', description='protease, serine 16 (thymus)', other_designations='protease, serine, 16 (thymus)|serine protease 16', map_location='13 A3-A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22002176, end_position_on_the_genomic_accession=22009741, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54375, CurrentID=0, Status='live', Symbol='Azin1', Aliases='1700085L02Rik, AZI, Oazi, Oazin', description='antizyme inhibitor 1', other_designations='ODC antizyme inhibitor|ornithine decarboxylase antizyme inhibitor', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=38487430, end_position_on_the_genomic_accession=38519287, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  54376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54376, CurrentID=0, Status='live', Symbol='Cacng3', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 3', other_designations='TARP gamma-3|neuronal voltage-gated calcium channel gamma-3 subunit|transmembrane AMPAR regulatory protein gamma-3', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122671744, end_position_on_the_genomic_accession=122769393, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54377, CurrentID=0, Status='live', Symbol='Cacng4', Aliases='AI413107, AW491861', description='calcium channel, voltage-dependent, gamma subunit 4', other_designations='TARP gamma-4|neuronal voltage-gated calcium channel gamma-4 subunit|transmembrane AMPAR regulatory protein gamma-4', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107734780, end_position_on_the_genomic_accession=107794464, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54378, CurrentID=0, Status='live', Symbol='Cacng6', Aliases='2310033H20Rik', description='calcium channel, voltage-dependent, gamma subunit 6', other_designations='neuronal voltage-gated calcium channel gamma-6 subunit', map_location='7 2.0 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3423580, end_position_on_the_genomic_accession=3435667, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54380, CurrentID=0, Status='live', Symbol='Smarcal1', Aliases='6030401P21Rik, AI851716, Mharp', description='SWI/SNF related matrix associated, actin dependent regulator of chromatin, subfamily a-like 1', other_designations='hepA-related protein|sucrose nonfermenting protein 2-like 1', map_location='1 36.72 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72536687, end_position_on_the_genomic_accession=72636790, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  54381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54381, CurrentID=0, Status='live', Symbol='Cpq', Aliases='1190003P12Rik, 2610034C17Rik, Hls2, Lal-1, Pgcp', description='carboxypeptidase Q', other_designations='hematopoietic lineage switch 2|plasma glutamate carboxypeptidase', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=33083129, end_position_on_the_genomic_accession=33594552, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54383, CurrentID=0, Status='live', Symbol='Phc2', Aliases='A3galt2, AA415044, D130050K19Rik, D4Ertd810e, Edr2, Mph2, p36', description='polyhomeotic-like 2 (Drosophila)', other_designations='early development regulator 2|early development regulatory protein 2|polyhomeotic 2 protein', map_location='4 62.12 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128654673, end_position_on_the_genomic_accession=128752881, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  54384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54384, CurrentID=0, Status='live', Symbol='Mtmr7', Aliases='', description='myotubularin related protein 7', other_designations='', map_location='8 B1.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40550159, end_position_on_the_genomic_accession=40634792, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  54387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54387, CurrentID=0, Status='live', Symbol='Mcm3ap', Aliases='GANP, mKIAA0572', description='minichromosome maintenance complex component 3 associated protein', other_designations='80 kDa MCM3-associated protein|GC-associated DNA primase|minichromosome maintenance deficient 3 (S. cerevisiae) associated protein|minichromosome maintenance deficient 3-associated protein|nuclear protein GANP', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76468971, end_position_on_the_genomic_accession=76515859, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  54388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54388, CurrentID=0, Status='live', Symbol='Hils1', Aliases='H1.9, TISP64', description='histone H1-like protein in spermatids 1', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94967661, end_position_on_the_genomic_accession=94968456, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54390, CurrentID=0, Status='live', Symbol='Sit1', Aliases='Sit', description='suppression inducing transmembrane adaptor 1', other_designations='SHP2 interacting transmembrane adaptor protein|SHP2-interacting transmembrane adapter protein|suppression-inducing transmembrane adapter 1', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43482083, end_position_on_the_genomic_accession=43483709, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54391, CurrentID=0, Status='live', Symbol='Rfk', Aliases='0610038L10Rik, AF031381, KOI-4', description='riboflavin kinase', other_designations="ATP:riboflavin 5'-phosphotransferase|flavokinase", map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=17394043, end_position_on_the_genomic_accession=17401349, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54392, CurrentID=0, Status='live', Symbol='Ncapg', Aliases='5730507H05Rik, Hcapg, MFT.M05.13', description='non-SMC condensin I complex, subunit G', other_designations='chromosome condensation protein G|on-SMC condensin I complex, subunit G', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=45669873, end_position_on_the_genomic_accession=45700547, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  54393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54393, CurrentID=0, Status='live', Symbol='Gabbr1', Aliases='GABAB1, GABAbR1, bM573K1.1', description='gamma-aminobutyric acid (GABA) B receptor, 1', other_designations='GABA-B receptor 1|GABA-B-R1|GABA-BR1|GABA-BR1b receptor|bM573K1.1.1 (gamma-aminobutyric acid (GABA) B receptor, 1a)|gamma-aminobutyric acid (GABA-B) receptor, 1|gb1', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37045966, end_position_on_the_genomic_accession=37074305, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  54394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54394, CurrentID=0, Status='live', Symbol='Crlf3', Aliases='BB164954, Creme9, Cytor4', description='cytokine receptor-like factor 3', other_designations='cytokine receptor-like molecule 9|cytokine receptor-related factor 4', map_location='11 47.43 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80046493, end_position_on_the_genomic_accession=80080991, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54396, CurrentID=0, Status='live', Symbol='Irgm2', Aliases='AI481100, GTPI, Ifggd2, Iigp2', description='immunity-related GTPase family M member 2', other_designations='interferon-g induced GTPase|interferon-gamma-inducible GTPase Ifggd2 protein', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58214977, end_position_on_the_genomic_accession=58222783, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  54397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54397, CurrentID=0, Status='live', Symbol='Ppt2', Aliases='0610007M19Rik, AA672937', description='palmitoyl-protein thioesterase 2', other_designations='', map_location='17 18.19 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34616662, end_position_on_the_genomic_accession=34638024, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  54399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54399, CurrentID=0, Status='live', Symbol='Bet1l', Aliases='2610021K23Rik, Gs15', description='Bet1 golgi vesicular membrane trafficking protein like', other_designations='GOS-15|bet1l|blocked early in transport 1 homolog (S. cerevisiae)-like|golgi SNARE with a size of 15 kDa|golgi SNARE, 15 kD|soluble N-ethylmaleimide-sensitive factor attachment protein  (SNARE) 15|vesicle transport protein GOS15', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140853384, end_position_on_the_genomic_accession=140856383, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54401, CurrentID=0, Status='live', Symbol='Ywhab', Aliases='1300003C17Rik', description='tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, beta polypeptide', other_designations='14-3-3 protein beta|KCIP-1|protein kinase C inhibitor protein 1|tyrosine 3/tryptophan 5 -monooxygenase activation protein, beta polypeptide', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163995197, end_position_on_the_genomic_accession=164018588, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54402, CurrentID=0, Status='live', Symbol='Stk19', Aliases='G11, RP1', description='serine/threonine kinase 19', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34823993, end_position_on_the_genomic_accession=34836980, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54403, CurrentID=0, Status='live', Symbol='Slc4a4', Aliases='AI835705, NBC, NBC1', description='solute carrier family 4 (anion exchanger), member 4', other_designations='Na(+)/HCO3(-) cotransporter|electrogenic sodium bicarbonate cotransporter NBCe1 variant D|pancreas sodium bicarbonate cotransporter|sodium bicarbonate cotransporter|solute carrier family 4 member 4|solute carrier family 4, sodium bicarbonate cotransporter, member 4', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88887247, end_position_on_the_genomic_accession=89239656, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  54405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54405, CurrentID=0, Status='live', Symbol='Ndufa1', Aliases='1810049F12Rik, MWFE', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 1', other_designations='CI-MWFE|NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 1 (7.5kD, MWFE)|NADH dehydrogenase 1 alpha subcomplex, 1|NADH ubiquinone oxidoreductase subunit MWFE|NADH-ubiquinone oxidoreductase MWFE subunit|complex I-MWFE', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37187588, end_position_on_the_genomic_accession=37191238, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54409, CurrentID=0, Status='live', Symbol='Ramp2', Aliases='', description='receptor (calcitonin) activity modifying protein 2', other_designations='', map_location='11 64.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101246334, end_position_on_the_genomic_accession=101248250, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54411, CurrentID=0, Status='live', Symbol='Atp6ap1', Aliases='16A, AC45, AI316502, AW108110, Atp6ip1, Atp6s1, C7-1, CF2, VATPS1, XAP-3, mFLJ00383', description='ATPase, H+ transporting, lysosomal accessory protein 1', other_designations='ATPase, H+ transporting, lysosomal (vacuolar proton pump), subunit 1|ATPase, H+ transporting, lysosomal interacting protein 1|V-ATPase Ac45 subunit|V-ATPase S1 accessory protein|V-ATPase subunit Ac45|V-ATPase subunit S1|protein C7-1|vacuolar H+ ATPase accessory subunit 1|vacuolar proton pump subunit S1', map_location='X 37.96 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74297097, end_position_on_the_genomic_accession=74304721, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54418, CurrentID=0, Status='live', Symbol='Fmn2', Aliases='AU024104', description='formin 2', other_designations='', map_location='1 81.04 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174501825, end_position_on_the_genomic_accession=174822729, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  54419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54419, CurrentID=0, Status='live', Symbol='Cldn6', Aliases='AL024037', description='claudin 6', other_designations='skullin|tight junction molecule claudin 6', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23679348, end_position_on_the_genomic_accession=23682446, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54420, CurrentID=0, Status='live', Symbol='Cldn8', Aliases='AI648025', description='claudin 8', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88560826, end_position_on_the_genomic_accession=88563183, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54422, CurrentID=0, Status='live', Symbol='Barhl1', Aliases='Dres115, MBH2', description='BarH-like 1 (Drosophila)', other_designations='bar-class homeodomain protein MBH2|barH-related homeobox protein 1', map_location='2 19.38 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28907680, end_position_on_the_genomic_accession=28918415, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54426, CurrentID=0, Status='live', Symbol='Hgfac', Aliases='HGFA', description='hepatocyte growth factor activator', other_designations='HGF activator', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35041514, end_position_on_the_genomic_accession=35048436, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  54427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54427, CurrentID=0, Status='live', Symbol='Dnmt3l', Aliases='D6Ertd14e, ecat7', description='DNA (cytosine-5-)-methyltransferase 3-like', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78042287, end_position_on_the_genomic_accession=78063622, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  54445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54445, CurrentID=0, Status='live', Symbol='Unc93b1', Aliases='Unc-93B1, Unc93b', description='unc-93 homolog B1 (C. elegans)', other_designations='unc-93 homolog B', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3935186, end_position_on_the_genomic_accession=3949340, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54446, CurrentID=0, Status='live', Symbol='Nfat5', Aliases='AI225870, B130038B15Rik, CAG-8, CAG80, NFATL1, OREBP, TonEBP, mKIAA0827, nfatz', description='nuclear factor of activated T cells 5', other_designations='T-cell transcription factor NFAT5|rel domain-containing transcription factor NFAT5|tonicity-responsive enhancer binding protein', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107293000, end_position_on_the_genomic_accession=107379517, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  54447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54447, CurrentID=0, Status='live', Symbol='Asah2', Aliases='AI585898', description='N-acylsphingosine amidohydrolase 2', other_designations='N-CDase|NCDase|acylsphingosine deacylase 2|neutral/alkaline ceramidase', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=31982597, end_position_on_the_genomic_accession=32108233, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  54448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54448, CurrentID=0, Status='live', Symbol='Il1f6', Aliases='Fil1, IL-1H1, IL1RP2, Il1f9, Il36a', description='interleukin 1 family, member 6', other_designations='FIL1 epsilon|Fil1epsilon|IL-1 epsilon|IL-1F6|interleukin 1 family, member 6 (epsilon)|interleukin 1 family, member 9|interleukin-1 epsilon|interleukin-1 family member 6|interleukin-1 homolog 1', map_location='2 16.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24215417, end_position_on_the_genomic_accession=24225701, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54450, CurrentID=0, Status='live', Symbol='Il1f5', Aliases='AI413231, Fil1delta, IL-36Ra, IL36RN, Il-1h3, Il1hy1', description='interleukin 1 family, member 5 (delta)', other_designations='IL-1 delta|IL-1F5|IL-1HY1|IL-1L1|interleukin 1 receptor antagonist homolog 1|interleukin-1 HY1|interleukin-1 delta|interleukin-1 family member 5|interleukin-1 homolog 3|interleukin-1-like protein 1', map_location='2 16.31 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24276954, end_position_on_the_genomic_accession=24282432, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  54451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54451, CurrentID=0, Status='live', Symbol='Cpsf3', Aliases='', description='cleavage and polyadenylation specificity factor 3', other_designations="CPSF 73 kDa subunit|cleavage and polyadenylation specificity factor 73 kDa subunit|mRNA 3'-end-processing endonuclease CPSF-73", map_location='12 A1.2-A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=21286297, end_position_on_the_genomic_accession=21315056, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  30923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=30923, CurrentID=0, Status='live', Symbol='Foxe3', Aliases='FREAC8, dyl, rct', description='forkhead box E3', other_designations='Rinshoken cataract|dysgenetic lens', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=114925147, end_position_on_the_genomic_accession=114926013, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54473, CurrentID=0, Status='live', Symbol='Tollip', Aliases='4930403G24Rik, 4931428G15Rik', description='toll interacting protein', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141881582, end_position_on_the_genomic_accession=141902465, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54483, CurrentID=0, Status='live', Symbol='Mefv', Aliases='FMF, TRIM20, pyrin', description='Mediterranean fever', other_designations='familial mediterranean fever|marenostrin', map_location='16 2.18 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3707215, end_position_on_the_genomic_accession=3718210, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54484, CurrentID=0, Status='live', Symbol='Mkrn1', Aliases='RFP', description='makorin, ring finger protein, 1', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39397821, end_position_on_the_genomic_accession=39420369, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54485, CurrentID=0, Status='live', Symbol='Dll4', Aliases='Delta4', description='delta-like 4 (Drosophila)', other_designations='drosophila Delta homolog 4', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119325784, end_position_on_the_genomic_accession=119335666, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54486, CurrentID=0, Status='live', Symbol='Hpgds', Aliases='H-PGDS, Ptgds2', description='hematopoietic prostaglandin D synthase', other_designations='GST class-sigma|glutathione S-transferase|glutathione-dependent PGD synthase|glutathione-dependent PGD synthetase|glutathione-requiring prostaglandin D synthase|prostaglandin D2 synthase 2, hematopoietic|prostaglandin-H2 D-isomerase', map_location='6 D-E', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=65117293, end_position_on_the_genomic_accession=65145024, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54519, CurrentID=0, Status='live', Symbol='Apbb1ip', Aliases='9930118P07Rik, Prp48', description='amyloid beta (A4) precursor protein-binding, family B, member 1 interacting protein', other_designations='APBB1-interacting protein 1|PREL-1|proline-rich EVH1 ligand 1|proline-rich protein 48', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=22755989, end_position_on_the_genomic_accession=22875653, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  54524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54524, CurrentID=0, Status='live', Symbol='Syt6', Aliases='3110037A08Rik, AW048930, sytVI', description='synaptotagmin VI', other_designations='synaptotagmin 6', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103575242, end_position_on_the_genomic_accession=103645567, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  54525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54525, CurrentID=0, Status='live', Symbol='Syt7', Aliases='AI851541, B230112P13Rik', description='synaptotagmin VII', other_designations='synaptotagmin 7|sytVII', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10388902, end_position_on_the_genomic_accession=10453181, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  54526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54526, CurrentID=0, Status='live', Symbol='Syt10', Aliases='', description='synaptotagmin X', other_designations='synaptotagmin 10|sytX', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89773714, end_position_on_the_genomic_accession=89841979, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54561, CurrentID=0, Status='live', Symbol='Nap1l3', Aliases='MB20', description='nucleosome assembly protein 1-like 3', other_designations='brain-specific protein MB20', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=122394560, end_position_on_the_genomic_accession=122397385, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54562, CurrentID=0, Status='live', Symbol='Lrrc6', Aliases='LRTP', description='leucine rich repeat containing 6 (testis)', other_designations='leucine-rich repeat-containing 6 (testis)|leucine-rich repeat-containing protein 6|leucine-rich testis-specific protein|protein TILB homolog|testis-specific leucine-rich repeat protein', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=66379858, end_position_on_the_genomic_accession=66500910, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54563, CurrentID=0, Status='live', Symbol='Nup210', Aliases='9830001L10, AI836801, Pom210, gp190, gp210', description='nucleoporin 210', other_designations='nuclear envelope pore membrane protein POM 210|nuclear pore membrane protein 210|nuclear pore protein gp210|nucleoporin 210; nuclear pore membrane glycoprotein 210; nuclear pore membrane protein 210|nucleoporin Nup210|pore membrane protein of 210 kDa', map_location='6 40.3 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91013067, end_position_on_the_genomic_accession=91117003, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  54598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54598, CurrentID=0, Status='live', Symbol='Calcrl', Aliases='AV071593, CRLR', description='calcitonin receptor-like', other_designations='CGRP type 1 receptor|calcitonin receptor-like receptor', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84330626, end_position_on_the_genomic_accession=84425411, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  54601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54601, CurrentID=0, Status='live', Symbol='Foxo4', Aliases='Afxh, Mllt7, afx', description='forkhead box O4', other_designations='fork head domain transcription factor AFX1|forkhead protein|myeloid/lymphoid or mixed lineage-leukemia translocation to 7 homolog', map_location='X 43.89 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101254259, end_position_on_the_genomic_accession=101260873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54604, CurrentID=0, Status='live', Symbol='Pcnx', Aliases='2900024E21Rik, 3526401J03Rik, AF096286, AI327143, AI413187l1, Pcnx', description='pecanex homolog (Drosophila)', other_designations='pecanex 1', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81860028, end_position_on_the_genomic_accession=82000924, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  100038882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100038882, CurrentID=0, Status='live', Symbol='Isg15', Aliases='G1p2, IGI15, IP17, Irfp, UCRP', description='ISG15 ubiquitin-like modifier', other_designations='interferon, alpha-inducible protein|interferon-induced 15 kDa protein|interferon-induced 15-KDa protein|interferon-induced 17 kDa protein|interferon-stimulated protein (15 kDa)|interferon-stimulated protein 15|ubiquitin cross-reactive protein', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156199424, end_position_on_the_genomic_accession=156200818, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54607, CurrentID=0, Status='live', Symbol='Socs6', Aliases='1500012M23Rik, 5830401B18Rik, AI447482, Cis4, Cish4, HSPC060, SOCS-4, SOCS-6, SSI4, STAI4, STATI4, Socs4', description='suppressor of cytokine signaling 6', other_designations='CIS-4|SH2 domain containing SOCS box protein SOCS6|cytokine inducible SH2-containing protein 4|cytokine inducible SH2-containing protein CIS4|cytokine-inducible SH2 protein 4|suppressor of cytokine signaling 4', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=88865169, end_position_on_the_genomic_accession=88894350, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54608, CurrentID=0, Status='live', Symbol='Abhd2', Aliases='2210009N18Rik, LABH2, Labh-2', description='abhydrolase domain containing 2', other_designations='alpha/beta hydrolase fold protein|lung alpha/beta hydrolase 2|mmLABH2', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79273266, end_position_on_the_genomic_accession=79361601, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  54609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54609, CurrentID=0, Status='live', Symbol='Ubqln2', Aliases='Chap1, Dsk2, HRIHFB2157, Plic-2, Plic2', description='ubiquilin 2', other_designations='Ubiquitin-related protein|protein linking IAP with cytoskeleton 2|ubiquitin-like product Chap1/Dsk2', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153498232, end_position_on_the_genomic_accession=153501562, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  54610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54610, CurrentID=0, Status='live', Symbol='Tbc1d8', Aliases='AD3, HBLP1, VRP', description='TBC1 domain family, member 8', other_designations='BUB2-like protein 1|GRAM domain|vascular Rab-GAP/TBC-containing protein', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39371495, end_position_on_the_genomic_accession=39478779, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  54611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54611, CurrentID=0, Status='live', Symbol='Pde3a', Aliases='A930022O17Rik, C87899', description='phosphodiesterase 3A, cGMP inhibited', other_designations='CGI-PDE A|cGMP-inhibited phosphodiesterase|cyclic GMP-inhibited phosphodiesterase A', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=141249269, end_position_on_the_genomic_accession=141506863, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  54612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54612, CurrentID=0, Status='live', Symbol='Sfrp5', Aliases='AI605071, SARP3', description='secreted frizzled-related sequence protein 5', other_designations='sFRP-5', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42197971, end_position_on_the_genomic_accession=42202252, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  54613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54613, CurrentID=0, Status='live', Symbol='St3gal6', Aliases='1700023B24Rik, AI930218, AW552396, Siat10, St3galVI', description='ST3 beta-galactoside alpha-2,3-sialyltransferase 6', other_designations='CMP-NeuAc:beta-galactoside alpha-2,3-sialyltransferase VI|ST3Gal VI|alpha-2,3-sialyltransferase VI|sialyltransferase 10 (alpha-2,3-sialyltransferase VI)', map_location='16 C1.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58468135, end_position_on_the_genomic_accession=58524255, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  54614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54614, CurrentID=0, Status='live', Symbol='Prpf40b', Aliases='2610317D23Rik, Hypc', description='pre-mRNA processing factor 40B', other_designations='PRP40 pre-mRNA processing factor 40 homolog A|PRP40 pre-mRNA processing factor 40 homolog B|huntingtin yeast partner C|huntingtin-interacting protein C|huntington yeast partner C', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99295087, end_position_on_the_genomic_accession=99317018, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  54615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54615, CurrentID=0, Status='live', Symbol='Npff', Aliases='', description='neuropeptide FF-amide peptide precursor', other_designations='FMRFamide-related peptides', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102523839, end_position_on_the_genomic_accession=102524942, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54616, CurrentID=0, Status='live', Symbol='Extl3', Aliases='2900009G18Rik, AU040105, Ext1l, mKIAA0519', description='exostoses (multiple)-like 3', other_designations='exostoses-like 3|exostosin|glucuronyl-galactosyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|multiple exostosis-like protein 3', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65052059, end_position_on_the_genomic_accession=65149899, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  54624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54624, CurrentID=0, Status='live', Symbol='Paf1', Aliases='5730511K23Rik, AA675045', description='Paf1, RNA polymerase II associated factor, homolog (S. cerevisiae)', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28392996, end_position_on_the_genomic_accession=28399383, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  54630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54630, CurrentID=0, Status='live', Symbol='Prickle3', Aliases='Lmo6, Sfc16', description='prickle homolog 3', other_designations='LIM domain only protein 6|triple LIM domain protein 6', map_location='X 3.45 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7657379, end_position_on_the_genomic_accession=7668186, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54631, CurrentID=0, Status='live', Symbol='Nphs1', Aliases='NephrinB, nephrin', description='nephrosis 1, nephrin', other_designations='nephrin 1|nephrosis 1 homolog, nephrin|renal glomerulus-specific cell adhesion receptor', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30458206, end_position_on_the_genomic_accession=30488609, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  54632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54632, CurrentID=0, Status='live', Symbol='Ftsj1', Aliases='AI931847, Ftsj, Ftsjl, Sfc12', description='FtsJ RNA methyltransferase homolog 1 (E. coli)', other_designations='', map_location='X 3.73 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8238668, end_position_on_the_genomic_accession=8252406, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  54633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54633, CurrentID=0, Status='live', Symbol='Pqbp1', Aliases='PQBP-1, Sfc2, npw38', description='polyglutamine binding protein 1', other_designations='38 kDa nuclear protein containing a WW domain|polyglutamine tract-binding protein 1|scurfy candidate 2', map_location='X 3.56 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7894519, end_position_on_the_genomic_accession=7899269, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  54634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54634, CurrentID=0, Status='live', Symbol='Magix', Aliases='Pdzx, Sfc18', description='MAGI family member, X-linked', other_designations='PDZ domain containing X chromosome', map_location='X 3.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7673166, end_position_on_the_genomic_accession=7681251, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  54635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54635, CurrentID=0, Status='live', Symbol='Pdgfc', Aliases='1110064L01Rik, AI647969, PDGF-C', description='platelet-derived growth factor, C polypeptide', other_designations='SCDGF|VEGF-E|fallotein|spinal cord-derived growth factor', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=81035885, end_position_on_the_genomic_accession=81214040, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  54636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54636, CurrentID=0, Status='live', Symbol='Wdr45', Aliases='C79260, DXImx38e, JM5, Sfc19, WIPI-4, Wdrx1', description='WD repeat domain 45', other_designations='WD repeat domain, X-linked 1|WD repeat-containing protein 45', map_location='X 3.48 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7721995, end_position_on_the_genomic_accession=7728201, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  54637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54637, CurrentID=0, Status='live', Symbol='Praf2', Aliases='AA986329, C78013, DXImx39e, JM4, Sfc20', description='PRA1 domain family 2', other_designations='', map_location='X 3.48 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7728571, end_position_on_the_genomic_accession=7731063, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  54638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54638, CurrentID=0, Status='live', Symbol='Ccdc22', Aliases='AI481216, DXImx40e, Ppp1r3f, Sfc22', description='coiled-coil domain containing 22', other_designations='', map_location='X 3.42 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7593809, end_position_on_the_genomic_accession=7605420, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  54644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54644, CurrentID=0, Status='live', Symbol='Otud5', Aliases='AA407879, AI553596, BB114028, DUBA, DXImx46e, Sfc7', description='OTU domain containing 5', other_designations='deubiquinating enzyme A|deubiquitinating enzyme A', map_location='X 3.54 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7841831, end_position_on_the_genomic_accession=7876626, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  54645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54645, CurrentID=0, Status='live', Symbol='Gripap1', Aliases='AI854681, DXImx47e, GRASP-1, Sfc10, mKIAA1167', description='GRIP1 associated protein 1', other_designations='HCMV-interacting protein', map_location='X 3.51 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7789961, end_position_on_the_genomic_accession=7820567, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  54646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54646, CurrentID=0, Status='live', Symbol='Ppp1r3f', Aliases='DXImx48e, R3F, RF3, Sfc15', description='protein phosphatase 1, regulatory (inhibitor) subunit 3F', other_designations='', map_location='X 3.4 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7558562, end_position_on_the_genomic_accession=7574281, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54648, CurrentID=0, Status='live', Symbol='Ccdc120', Aliases='AA030924, DXImx50e, Jm11, Sfc25', description='coiled-coil domain containing 120', other_designations='', map_location='X 3.48 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7731714, end_position_on_the_genomic_accession=7741324, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54650, CurrentID=0, Status='live', Symbol='Sfmbt1', Aliases='4930442N21Rik, 9330180L21Rik, AA536974, Sfmbt, Smr', description='Scm-like with four mbt domains 1', other_designations='Scm-related gene containing four mbt domains 1', map_location='14 18.99 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30714473, end_position_on_the_genomic_accession=30822721, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  54651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54651, CurrentID=0, Status='live', Symbol='Usp27x', Aliases='Sfc11, Usp27', description='ubiquitin specific peptidase 27, X chromosome', other_designations='X-linked ubiquitin carboxyl-terminal hydrolase 27|X-linked ubiquitin thioesterase 27|deubiquitinating enzyme 27|ubiquitin specific protease 27, X chromosome|ubiquitin thioesterase 27|ubiquitin thiolesterase 27|ubiquitin-specific-processing protease 27', map_location='X 3.31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7372591, end_position_on_the_genomic_accession=7375830, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  54652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54652, CurrentID=0, Status='live', Symbol='Cacna1f', Aliases='A930034B14, Cav1.4, Sfc17, nerg1, nob2', description='calcium channel, voltage-dependent, alpha 1F subunit', other_designations='L-type dihydropyridine-sensitive calcium channel alpha-1f subunit|voltage-gated calcium channel subunit alpha Cav1.4', map_location='X 3.42 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7606982, end_position_on_the_genomic_accession=7635196, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  54667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54667, CurrentID=0, Status='live', Symbol='Atp8b2', Aliases='Id', description='ATPase, class I, type 8B, member 2', other_designations='Atpase 8B2, p type|P4-ATPase flippase complex alpha subunit ATP8B2|probable phospholipid-transporting ATPase ID', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89939481, end_position_on_the_genomic_accession=89963332, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  54670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54670, CurrentID=0, Status='live', Symbol='Atp8b1', Aliases='AI451886, FIC1, Ic', description='ATPase, class I, type 8B, member 1', other_designations='ATPase 8B1, p type|ATPase B1, class I|P4-ATPase flippase complex alpha subunit ATP8B1|aminophospholipid translocase|familial intrahepatic cholestasis type I|probable phospholipid-transporting ATPase IC', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=64528979, end_position_on_the_genomic_accession=64661860, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  54672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54672, CurrentID=0, Status='live', Symbol='Adgrg3', Aliases='A030001G24Rik, AW214383, Gpr97, Pb1, Pb99', description='adhesion G protein-coupled receptor G3', other_designations='G 97|G protein-coupled receptor 97|G-protein coupled receptor 97|pre B-cell transcription factor 1|probable G-protein coupled receptor 97', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95017114, end_position_on_the_genomic_accession=95045250, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  54673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54673, CurrentID=0, Status='live', Symbol='Sh3glb1', Aliases='AA409932, AI314629, AU015566, Bif-1, mKIAA0491', description='SH3-domain GRB2-like B1 (endophilin)', other_designations='SH3 domain-containing GRB2-like protein B1', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144683678, end_position_on_the_genomic_accession=144720398, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  54678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54678, CurrentID=0, Status='live', Symbol='Zfp108', Aliases='', description='zinc finger protein 108', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24254794, end_position_on_the_genomic_accession=24262445, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  54683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54683, CurrentID=0, Status='live', Symbol='Prdx5', Aliases='AOEB166, AOPP, Pmp20, Prdx6, PrxV', description='peroxiredoxin 5', other_designations='PLP|antioxidant enzyme B166|liver tissue 2D-page spot 2D-0014IV|peroxiredoxin 6|peroxiredoxin V|peroxisomal antioxidant enzyme|peroxisomal membrane protein 20|prx-V|thioredoxin peroxidase PMP20|thioredoxin reductase', map_location='19 5.08 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6906819, end_position_on_the_genomic_accession=6909645, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  54698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54698, CurrentID=0, Status='live', Symbol='Crtam', Aliases='', description='cytotoxic and regulatory T cell molecule', other_designations='class I-restricted T cell-associated molecule|class-I MHC-restricted T-cell-associated molecule|crtam', map_location='9 21.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40972798, end_position_on_the_genomic_accession=41004628, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54709, CurrentID=0, Status='live', Symbol='Eif3i', Aliases='36kDa, D4Ertd632e, Eif3s2, Trip1', description='eukaryotic translation initiation factor 3, subunit I', other_designations='TGF-beta receptor binding protein|TGF-beta receptor-interacting protein 1|TRIP-1|eIF3i|eukaryotic translation initiation factor 3, subunit 1|eukaryotic translation initiation factor 3, subunit 2 (beta)|eukaryotic translation initiation factor 3, subunit 2 (beta, 36kD)', map_location='4 63.26 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129591974, end_position_on_the_genomic_accession=129600648, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  54710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54710, CurrentID=0, Status='live', Symbol='Hs3st3b1', Aliases='3-OST-3B, 3OST3B1, 3Ost3b, AU041882, AW536289, Hs3st3b, m3-OST-3B', description='heparan sulfate (glucosamine) 3-O-sulfotransferase 3B1', other_designations='D-glycosaminyl 3-O-sulfotransferase-3B|heparan sulfate 3-O-sulfotransferase 3B1|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 3B1', map_location='11 39.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=63884693, end_position_on_the_genomic_accession=63922284, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  54711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54711, CurrentID=0, Status='live', Symbol='Plagl2', Aliases='AU018672, AW552839, mKIAA0198', description='pleiomorphic adenoma gene-like 2', other_designations='pleomorphic adenoma gene like 2', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153227767, end_position_on_the_genomic_accession=153241358, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  54712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54712, CurrentID=0, Status='live', Symbol='Plxnc1', Aliases='2510048K12Rik, AW742158, CD232, Vespr', description='plexin C1', other_designations='viral-encoded semaphorin protein receptor|virus-encoded semaphorin protein receptor', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=94790866, end_position_on_the_genomic_accession=94945156, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  54713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54713, CurrentID=0, Status='live', Symbol='Fezf2', Aliases='AI451466, AI852056, Fez, Fezl, Zfp312', description='Fez family zinc finger 2', other_designations='forebrain embryonic zinc finger|forebrain embryonic zinc finger-like protein 2|zinc finger protein 312', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=12341892, end_position_on_the_genomic_accession=12345865, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54720, CurrentID=0, Status='live', Symbol='Rcan1', Aliases='Adapt78, CALP1L, CSP1, DSC1, Dscr1, MCIP1, RCN1', description='regulator of calcineurin 1', other_designations='calcineurin inhibitor|down syndrome critical region protein 1 homolog|myocyte-enriched calcineurin-interacting protein 1', map_location='16 53.6 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92391951, end_position_on_the_genomic_accession=92466169, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  54721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54721, CurrentID=0, Status='live', Symbol='Tyk2', Aliases='JTK1', description='tyrosine kinase 2', other_designations='tyrosine kinase TYK2', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21104068, end_position_on_the_genomic_accession=21131275, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  54722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54722, CurrentID=0, Status='live', Symbol='Dfna5', Aliases='2310037D07Rik, 4932441K13Rikh, EG14210, Fin15, Dfna5', description='deafness, autosomal dominant 5 (human)', other_designations='deafness, autosomal dominant 5 homolog|nonsyndromic hearing impairment protein', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=50207403, end_position_on_the_genomic_accession=50262862, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  54723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54723, CurrentID=0, Status='live', Symbol='Tfip11', Aliases='2810002G02Rik, AF097181, AW046167, Srr1, TIP33, Tip39', description='tuftelin interacting protein 11', other_designations='STIP-1|septin and tuftelin-interacting protein 1|tuftelin-interacting protein 33|tuftelin-interacting protein 39|tuftelin-interacting protein, 39 kD', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112326369, end_position_on_the_genomic_accession=112338073, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  54725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=54725, CurrentID=0, Status='live', Symbol='Cadm1', Aliases='2900073G06Rik, 3100001I08Rik, AI987920, Bl2, Igsf4, Igsf4a, Necl2, RA175, RA175A, RA175B, RA175C, RA175N, ST17, SgIGSF, SynCam, Tslc1', description='cell adhesion molecule 1', other_designations='NECL-2|TSLC-1|cadm1|cell adhesion molecule 1 transcript variant 8abc|immunoglobulin superfamily member 4|immunoglobulin superfamily member 4A|immunosuperfamily protein Bl2|nectin-like 2|nectin-like protein 2|spermatogenic immunoglobulin superfamily|synaptic cell adhesion molecule|tumor suppressor in lung cancer 1', map_location='9 B-C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=47530239, end_position_on_the_genomic_accession=47853385, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  620913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620913, CurrentID=0, Status='live', Symbol='Gm12185', Aliases='OTTMUSG00000005523', description='predicted gene 12185', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48904656, end_position_on_the_genomic_accession=48927182, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  317652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317652, CurrentID=0, Status='live', Symbol='Klk15', Aliases='', description='kallikrein related-peptidase 15', other_designations='kallikrein 15', map_location='7 28.41 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43929405, end_position_on_the_genomic_accession=43939852, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  317653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317653, CurrentID=0, Status='live', Symbol='Klk14', Aliases='GK14', description='kallikrein related-peptidase 14', other_designations='glandular kallikrein KLK14', map_location='7 28.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43680879, end_position_on_the_genomic_accession=43695535, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  317677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317677, CurrentID=0, Status='live', Symbol='C1s2', Aliases='C1sb, EG317677, Gm5077', description='complement component 1, s subcomponent 2', other_designations='C1 esterase|complement component 1 subcomponent s-B|complement component 1, s-B subcomponent|complement component C1SB', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124624625, end_position_on_the_genomic_accession=124636085, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  317717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317717, CurrentID=0, Status='live', Symbol='Sec22a', Aliases='1810005C06Rik, Sec22l2', description='SEC22 homolog A, vesicle trafficking protein', other_designations='SEC22 vesicle trafficking protein homolog A|SEC22 vesicle trafficking protein homologe A|SEC22 vesicle trafficking protein-like 2|SEC22 vesicle trafficking protein-like A|SEC22 vesicle-trafficking protein homolog A|SEC22 vesicle-trafficking protein-like 2|sec22a', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35311131, end_position_on_the_genomic_accession=35363934, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  317750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317750, CurrentID=0, Status='live', Symbol='Slc24a5', Aliases='F630045L20Rik, JSX, NCKX5, NCX5, Oca6', description='solute carrier family 24, member 5', other_designations='K-independent Na+/Ca2+ exchanger JSX|Na(+)/K(+)/Ca(2+)-exchange protein 5|Na-Ca exchanger 5|sodium/potassion/calcium exchanger 5', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125068127, end_position_on_the_genomic_accession=125088677, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  317755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317755, CurrentID=0, Status='live', Symbol='Zar1', Aliases='', description='zygote arrest 1', other_designations='oocyte-specific maternal effect factor', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72577113, end_position_on_the_genomic_accession=72581361, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  317757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317757, CurrentID=0, Status='live', Symbol='Gimap5', Aliases='D630024P16, E230026N22Rik', description='GTPase, IMAP family member 5', other_designations='immunity-associated nucleotide 4-like 1 protein|immunity-associated protein 3', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48746137, end_position_on_the_genomic_accession=48755235, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  317758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=317758, CurrentID=0, Status='live', Symbol='Gimap9', Aliases='A630002K24, BB145400', description='GTPase, IMAP family member 9', other_designations='immune associated nucleotide family member', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48676121, end_position_on_the_genomic_accession=48679116, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  448850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=448850, CurrentID=0, Status='live', Symbol='Znhit3', Aliases='Myohd1, Trip3', description='zinc finger, HIT type 3', other_designations='TR-interacting protein 3|TRIP-3|myosin head domain containing 1|thyroid hormone receptor interactor 3|thyroid receptor-interacting protein 3', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84910950, end_position_on_the_genomic_accession=84916356, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  448986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=448986, CurrentID=0, Status='live', Symbol='Pnet-ps', Aliases='Et20, Pnet1', description='prenatal ethanol induced mRNA, pseudogene', other_designations='prenatal ethanol induced mRNA 1', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  448987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=448987, CurrentID=0, Status='live', Symbol='Fbxl7', Aliases='AL023057, D230018M15Rik, FBL7, Fbl6', description='F-box and leucine-rich repeat protein 7', other_designations='', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=26540454, end_position_on_the_genomic_accession=26895564, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  449000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=449000, CurrentID=0, Status='live', Symbol='Zfp960', Aliases='', description='zinc finger protein 960', other_designations='', map_location='17 A3.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17064103, end_position_on_the_genomic_accession=17089631, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  55925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55925, CurrentID=0, Status='live', Symbol='Syt8', Aliases='', description='synaptotagmin VIII', other_designations='synaptotagmin 8|sytVIII', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142434462, end_position_on_the_genomic_accession=142440399, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  55927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55927, CurrentID=0, Status='live', Symbol='Hes6', Aliases='AI326893, bHLHb41', description='hairy and enhancer of split 6', other_designations='hes6', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91411482, end_position_on_the_genomic_accession=91413877, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  55932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55932, CurrentID=0, Status='live', Symbol='Gbp3', Aliases='AW228655, GBP-3, GBP-4, Gbp4', description='guanylate binding protein 3', other_designations='GTP-binding protein 3|GTP-binding protein 4|guanine nucleotide-binding protein 4|guanylate nucleotide binding protein 3|guanylate nucleotide binding protein 4|guanylate-binding protein 3|putative purine nucleotide binding protein', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142560016, end_position_on_the_genomic_accession=142573213, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  55934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55934, CurrentID=0, Status='live', Symbol='Rp9', Aliases='PAP-1h, Rp9', description='retinitis pigmentosa 9 (human)', other_designations='Pim-1 associated protein|pim-1-associated protein|retinitis pigmentosa 9 homolog', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22448311, end_position_on_the_genomic_accession=22468356, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  55935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55935, CurrentID=0, Status='live', Symbol='Fnbp4', Aliases='FBP30, Fnbp30, mKIAA1014', description='formin binding protein 4', other_designations='formin binding protein 30|formin-binding protein 30', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90745370, end_position_on_the_genomic_accession=90781021, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55936, CurrentID=0, Status='live', Symbol='Ctps2', Aliases='A830031M15Rik, AI326475, Ctpsh', description="cytidine 5'-triphosphate synthase 2", other_designations='CTP synthetase 2|CTP synthetase homolog|UTP--ammonia ligase 2|cytidine triphosphate synthase II', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=162901243, end_position_on_the_genomic_accession=163034541, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  55938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55938, CurrentID=0, Status='live', Symbol='Apom', Aliases='1190010O19Rik, G3a, NG20', description='apolipoprotein M', other_designations='apo-M', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35128997, end_position_on_the_genomic_accession=35131752, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  55940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55940, CurrentID=0, Status='live', Symbol='D2Mit19', Aliases='D2Mit25', description='DNA segment, Chr 2, Massachusetts Institute of Technology 19', other_designations='', map_location='2 65.43 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  55942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55942, CurrentID=0, Status='live', Symbol='Sertad1', Aliases='1110032C13Rik, AV000455, AW456625, Sei-1, Sei1, TRIP-Br1', description='SERTA domain containing 1', other_designations='CDK4-binding protein p34SEI1|PHD zinc finger- and bromodomain-interacting protein 1|p34SEI-1|transcriptional regulator interacting with the PHD-bromodomain 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27486953, end_position_on_the_genomic_accession=27490316, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  55943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55943, CurrentID=0, Status='live', Symbol='Stx8', Aliases='0610007H08Rik, 1110002H11Rik, 4930571E13Rik, CARB', description='syntaxin 8', other_designations='syntaxin-like protein 3I35', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67966411, end_position_on_the_genomic_accession=68207148, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  55944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55944, CurrentID=0, Status='live', Symbol='Eif3d', Aliases='66/67kDa, AA407891, Eif3s7, eIF3p66', description='eukaryotic translation initiation factor 3, subunit D', other_designations='eIF-3-zeta|eIF3 p66|eIF3d|eukaryotic translation initiation factor 3 subunit 7|eukaryotic translation initiation factor 3, subunit 7 (zeta)|translation initiation factor eIF3 p66', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77958998, end_position_on_the_genomic_accession=77971088, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  55946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55946, CurrentID=0, Status='live', Symbol='Ap3m1', Aliases='1200013D09Rik, C78982, R75378', description='adaptor-related protein complex 3, mu 1 subunit', other_designations='AP-3 adapter complex mu3A subunit|AP-3 adaptor complex mu3A subunit|adapter-related protein complex 3 mu-1 subunit|adapter-related protein complex 3 subunit mu-1|adaptor-related protein complex 3 subunit mu-1|adaptor-related protein complex AP-3, mu 1 subunit|clathrin adaptor protein mu3A|mu-adaptin 3A|mu3A-adaptin', map_location='14 11.58 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21033742, end_position_on_the_genomic_accession=21052535, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  55947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55947, CurrentID=0, Status='live', Symbol='Dclre1a', Aliases='2810043H12Rik, AU022226, Pso2, Smn1a, Snm1, mKIAA0086', description='DNA cross-link repair 1A', other_designations='DNA cross-link repair 1A, PSO2 homolog|SNM1 homolog A', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56529161, end_position_on_the_genomic_accession=56548284, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  55948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55948, CurrentID=0, Status='live', Symbol='Sfn', Aliases='Er, Mme1, Ywhas', description='stratifin', other_designations='14-3-3 sigma protein|tyrosine 3-monooxygenase/tryptophan 5-monooxygenase activation protein, sigma polypeptide', map_location='4 66.25 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133600556, end_position_on_the_genomic_accession=133602168, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  55949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55949, CurrentID=0, Status='live', Symbol='Eef1b2', Aliases='2810017J07Rik, Eef1b', description='eukaryotic translation elongation factor 1 beta 2', other_designations='EF-1-beta|eukaryotic translation elongation factor 1-beta homolog', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63176831, end_position_on_the_genomic_accession=63180486, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55950, CurrentID=0, Status='live', Symbol='Bri3', Aliases='I3', description='brain protein I3', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144244437, end_position_on_the_genomic_accession=144446757, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55951, CurrentID=0, Status='live', Symbol='Mpc1', Aliases='0610006G08Rik, 3830411I18Rik, Brp44l', description='mitochondrial pyruvate carrier 1', other_designations='apoptosis-regulating basic protein|brain protein 44-like protein', map_location='17 4.92 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8283813, end_position_on_the_genomic_accession=8297667, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55960, CurrentID=0, Status='live', Symbol='Ebag9', Aliases='AI835379, Rcas1', description='estrogen receptor-binding fragment-associated gene 9', other_designations='cancer-associated surface antigen RCAS1|estrogen receptor-binding fragment-associated gene 9 protein', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=44619169, end_position_on_the_genomic_accession=44641542, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  55961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55961, CurrentID=0, Status='live', Symbol='Slc13a1', Aliases='AI314066, AI987826, NaSi-1, Nas1', description='solute carrier family 13 (sodium/sulfate symporters), member 1', other_designations='Na(+)/sulfate cotransporter|renal sodium/sulfate cotransporter|solute carrier family 13 (sodium/sulphate symporters), member 1', map_location='6 11.22 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24088283, end_position_on_the_genomic_accession=24168119, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  55963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55963, CurrentID=0, Status='live', Symbol='Slc1a4', Aliases='ASCT-1, ASCT1, AW045657, SATT', description='solute carrier family 1 (glutamate/neutral amino acid transporter), member 4', other_designations='alanine/serine/cysteine/threonine transporter 1|solute carrier family 1 member 4', map_location='11 12.97 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=20302180, end_position_on_the_genomic_accession=20332713, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  55978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55978, CurrentID=0, Status='live', Symbol='Ift20', Aliases='0610009H04Rik, AU015496', description='intraflagellar transport 20', other_designations='intraflagellar transport 20 homolog|mIFT20|uterine protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78536436, end_position_on_the_genomic_accession=78541473, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  55979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55979, CurrentID=0, Status='live', Symbol='Agpat1', Aliases='1-AGP, 1-AGPAT, AW047140', description='1-acylglycerol-3-phosphate O-acyltransferase 1 (lysophosphatidic acid acyltransferase, alpha)', other_designations='1-AGP acyltransferase 1|1-AGPAT 1|Lpaat-alpha|lysophosphatidic acid acyltransferase alpha', map_location='17 18.18 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34604278, end_position_on_the_genomic_accession=34613449, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  55980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55980, CurrentID=0, Status='live', Symbol='Impa1', Aliases='2610002K09Rik, 2900059K10Rik, AI325909', description='inositol (myo)-1(or 4)-monophosphatase 1', other_designations='D-galactose 1-phosphate phosphatase|IMP 1|IMPase 1|inositol-1(or 4)-monophosphatase 1|lithium-sensitive myo-inositol monophosphatase A1', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10313540, end_position_on_the_genomic_accession=10331439, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55981, CurrentID=0, Status='live', Symbol='Pigb', Aliases='Pig-b', description='phosphatidylinositol glycan anchor biosynthesis, class B', other_designations='GPI mannosyltransferase III|GPI-MT-III|involved in GPI-anchor biosynthesis|phosphatidylinositol glycan, class B|phosphatidylinositol-glycan biosynthesis class B protein', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=73013726, end_position_on_the_genomic_accession=73039699, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  55982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55982, CurrentID=0, Status='live', Symbol='Paxip1', Aliases='D5Ertd149e, PTIP', description='PAX interacting (with transcription-activation domain) protein 1', other_designations='PAX transactivation activation domain-interacting protein', map_location='5 13.23 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=27740080, end_position_on_the_genomic_accession=27792079, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  55983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55983, CurrentID=0, Status='live', Symbol='Pdzrn3', Aliases='1110020C07Rik, AI429718, AL023082, LNX3, SEMACAP3, Semcap3', description='PDZ domain containing RING finger 3', other_designations='PDZ domain containing RING finger 3A|PDZ domain-containing RING finger protein 3|semaphorin cytoplasmic domain-associated protein 3|semaphorin cytoplasmic domain-associated protein 3A', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=101149607, end_position_on_the_genomic_accession=101377897, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  55984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55984, CurrentID=0, Status='live', Symbol='Camkk1', Aliases='AI846603, CaMKKalpha, Camkk', description='calcium/calmodulin-dependent protein kinase kinase 1, alpha', other_designations='caM-KK 1|caM-KK alpha|caM-kinase IV kinase|caM-kinase kinase 1|caM-kinase kinase alpha|caMKK 1|caMKK alpha|calcium calmodulin dependent protein kinase kinase alpha|calcium/calmodulin-dependent protein kinase kinase alpha', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73015306, end_position_on_the_genomic_accession=73042073, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  55985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55985, CurrentID=0, Status='live', Symbol='Cxcl13', Aliases='4631412M08Rik, ANGIE2, Angie, BCA-1, BLC, BLR1L, Scyb13', description='chemokine (C-X-C motif) ligand 13', other_designations='B lymphocyte chemoattractant|CXC chemokine BLC|small inducible cytokine subfamily B (Cys-X-Cys), member 13|small-inducible cytokine B13', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=95956939, end_position_on_the_genomic_accession=95961068, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  55987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55987, CurrentID=0, Status='live', Symbol='Cpxm2', Aliases='4632435C11Rik, CPX-2, Cpx2', description='carboxypeptidase X 2 (M14 family)', other_designations='carboxypeptidase X2|metallocarboxypeptidase 2', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132032687, end_position_on_the_genomic_accession=132154768, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  55988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55988, CurrentID=0, Status='live', Symbol='Snx12', Aliases='2610001F05Rik, AW045757, SDP8', description='sorting nexin 12', other_designations='', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101089367, end_position_on_the_genomic_accession=101222563, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  55989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55989, CurrentID=0, Status='live', Symbol='Nop58', Aliases='MSSP, Nol5, SIK, nop5', description='NOP58 ribonucleoprotein', other_designations='NOP58 ribonucleoprotein homolog|SIK similar protein|SIK-similar protein|nucleolar protein 5', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59684951, end_position_on_the_genomic_accession=59711510, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  55990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55990, CurrentID=0, Status='live', Symbol='Fmo2', Aliases='2310008D08Rik, 2310042I22Rik, AW107733', description='flavin containing monooxygenase 2', other_designations='FMO 2|dimethylaniline oxidase 2|pulmonary flavin-containing monooxygenase 2', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162874321, end_position_on_the_genomic_accession=162898712, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  55991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55991, CurrentID=0, Status='live', Symbol='Panx1', Aliases='AI847747', description='pannexin 1', other_designations='', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15002132, end_position_on_the_genomic_accession=15045521, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  55992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55992, CurrentID=0, Status='live', Symbol='Trim3', Aliases='BERP1, HAC1, Rnf22', description='tripartite motif-containing 3', other_designations='RING finger protein HAC1|ring finger protein 22|tripartite motif protein 3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105604463, end_position_on_the_genomic_accession=105633571, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  55993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55993, CurrentID=0, Status='live', Symbol='Msh4', Aliases='4930485C04Rik, AV144863, mMsh4', description='mutS homolog 4 (E. coli)', other_designations='MutS homolog 4 variant iota 1|MutS homolog 4 variant theta 1', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=153857141, end_position_on_the_genomic_accession=153906138, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  55994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=55994, CurrentID=0, Status='live', Symbol='Smad9', Aliases='MADH6, Madh8, Madh9, Smad8', description='SMAD family member 9', other_designations='MAD homolog 9|mothers against DPP homolog 9', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54755457, end_position_on_the_genomic_accession=54801741, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56009, CurrentID=0, Status='live', Symbol='Alyref2', Aliases='C130042O11Rik, REF2, Refbp2', description='Aly/REF export factor 2', other_designations='RNA and export factor binding protein 2|RNA and export factor binding protein 2-I|RNA and export factor-binding protein 2|alyref', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171503478, end_position_on_the_genomic_accession=171504750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  56012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56012, CurrentID=0, Status='live', Symbol='Pgam2', Aliases='', description='phosphoglycerate mutase 2', other_designations='BPG-dependent PGAM 2|PGAM-M|muscle-specific phosphoglycerate mutase|phosphoglycerate mutase isozyme M|phosphoglycerate mutase muscle-specific subunit', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5801637, end_position_on_the_genomic_accession=5803796, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56013, CurrentID=0, Status='live', Symbol='Srcin1', Aliases='P140, SNIP, mKIAA1684, p140Cap', description='SRC kinase signaling inhibitor 1', other_designations='Snap-25-interacting protein|p130Cas-associated protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97509339, end_position_on_the_genomic_accession=97575126, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  56014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56014, CurrentID=0, Status='live', Symbol='Olfr70', Aliases='MOR262-10, mOR6', description='olfactory receptor 70', other_designations='GA_x6K02T2N78B-16239704-16240654|olfactory receptor MOR262-10', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43695000, end_position_on_the_genomic_accession=43700807, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56015, CurrentID=0, Status='live', Symbol='Olfr71', Aliases='MOR262-4, mOR17', description='olfactory receptor 71', other_designations='GA_x6K02T2N78B-16230286-16231224|olfactory receptor MOR262-4', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43705628, end_position_on_the_genomic_accession=43706566, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56016, CurrentID=0, Status='live', Symbol='Hebp2', Aliases='SOUL', description='heme binding protein 2', other_designations='putative heme-binding protein', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18540123, end_position_on_the_genomic_accession=18546076, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56017, CurrentID=0, Status='live', Symbol='Slc2a8', Aliases='D2Ertd44e, GLUT8, GlutX1', description='solute carrier family 2, (facilitated glucose transporter), member 8', other_designations='GLUT-8|glucose transporter X1|glucose transporter type 8|glucose transporter type X1|intestinal GLUT8 transporter', map_location='2 22.09 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32972981, end_position_on_the_genomic_accession=32982113, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56018, CurrentID=0, Status='live', Symbol='Stard10', Aliases='AV048538, CGI-52, NY-C0-28, PC-TP2, PCTP2, Pctpl, Sdccag28, SdccagG28, TISP-81', description='START domain containing 10', other_designations='PCTP-L|START domain-containing protein 10|serologically defined colon cancer antigen 10|serologically defined colon cancer antigen 28 homolog|stAR-related lipid transfer protein 10', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101317091, end_position_on_the_genomic_accession=101346312, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56030, CurrentID=0, Status='live', Symbol='Tmem131', Aliases='2610524E03Rik, CC28, D1Bwg0491e, Neg, RW1, YR-23, mKIAA0257', description='transmembrane protein 131', other_designations='', map_location='1 15.42 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36792189, end_position_on_the_genomic_accession=36939637, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  56031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56031, CurrentID=0, Status='live', Symbol='Ppie', Aliases='2010010D16Rik, Cyp33', description='peptidylprolyl isomerase E (cyclophilin E)', other_designations='PPIase E|cyclophilin E|cyclophilin-33|rotamase E', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123127114, end_position_on_the_genomic_accession=123140000, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56032, CurrentID=0, Status='live', Symbol='Nprl2', Aliases='2810446G01Rik, G21, NPR2L, Tusc4', description='nitrogen permease regulator-like 2', other_designations='NPR2-like protein|gene 21 protein|nitrogen permase homolog|tumor suppressor candidate 4', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107542206, end_position_on_the_genomic_accession=107545706, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56036, CurrentID=0, Status='live', Symbol='Ccnl2', Aliases='1700010A01Rik, 1810019L15Rik, 2010319M22Rik, AA409936, AW261738, Ania-6b, HCLA-150, Pcee, SB138', description='cyclin L2', other_designations='cyclin Ania-6b|paneth cell enhanced expression|paneth cell-enhanced expression protein', map_location='4 87.47 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155812084, end_position_on_the_genomic_accession=155824543, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56040, CurrentID=0, Status='live', Symbol='Rplp1', Aliases='2410042H16Rik, Arpp1, C430017H15Rik', description='ribosomal protein, large, P1', other_designations='acidic ribosomal phosphoprotein P1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=61913283, end_position_on_the_genomic_accession=61914510, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56041, CurrentID=0, Status='live', Symbol='Uso1', Aliases='115kDa, TAP, Vdp', description='USO1 vesicle docking factor', other_designations='USO1 homolog, vesicle docking protein|protein USO1 homolog|transcytosis associated protein p115|transcytosis-associated protein|vesicle docking protein, 115 kDa|vesicle-docking protein', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92137909, end_position_on_the_genomic_accession=92202800, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  56043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56043, CurrentID=0, Status='live', Symbol='Akr1e1', Aliases='1810061I10Rik, Akr1e2', description='aldo-keto reductase family 1, member E1', other_designations='AF reductase|aldo-keto reductase family 1 member C-like protein 2|aldo-keto reductase family 1 member E2', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4592490, end_position_on_the_genomic_accession=4609164, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56044, CurrentID=0, Status='live', Symbol='Rala', Aliases='3010001O15Rik, AW322615, Ral, Rasl1', description='v-ral simian leukemia viral oncogene homolog A (ras related)', other_designations='RAS-like, family 1|ral-A protein', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=17880575, end_position_on_the_genomic_accession=17944217, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56045, CurrentID=0, Status='live', Symbol='Samhd1', Aliases='E330031J07Rik, Mg11', description='SAM domain and HD domain, 1', other_designations='IFN-gamma induced|SAM domain and HD domain-containing protein 1|dNTPase|interferon-gamma-inducible protein Mg11', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157097529, end_position_on_the_genomic_accession=157135277, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  56046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56046, CurrentID=0, Status='live', Symbol='Uqcc1', Aliases='2310079L17Rik, 2410003P15Rik, 3110038N19Rik, Bfzb, Bfzp, Cbp3, Uqcc', description='ubiquinol-cytochrome c reductase complex assembly factor 1', other_designations='basic FGF repressed, Zic binding protein|basic FGF repressed, zinc binding protein|basic FGF-repressed Zic-binding protein|ubiquinol-cytochrome c reductase complex chaperone CBP3 homolog|ubiquinol-cytochrome c reductase complex chaperone, CBP3 homolog', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155846886, end_position_on_the_genomic_accession=155930343, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56047, CurrentID=0, Status='live', Symbol='Msln', Aliases='MPF', description='mesothelin', other_designations='megakaryocyte potentiating factor|pre-pro-megakaryocyte-potentiating factor', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25748613, end_position_on_the_genomic_accession=25754437, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  56048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56048, CurrentID=0, Status='live', Symbol='Lgals8', Aliases='1200015E08Rik, AI326142, D13Ertd524e, Lgals-8', description='lectin, galactose binding, soluble 8', other_designations='galectin 8|galectin-8L|lectin, galactose binding, soluble 5', map_location='13 4.64 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12439402, end_position_on_the_genomic_accession=12464976, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56050, CurrentID=0, Status='live', Symbol='Cyp39a1', Aliases='mCYP39A1', description='cytochrome P450, family 39, subfamily a, polypeptide 1', other_designations='cytochrome P450, 39a1 (oxysterol 7alpha-hydroxylase)|oxysterol 7-alpha-hydroxylase', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43667372, end_position_on_the_genomic_accession=43751431, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56055, CurrentID=0, Status='live', Symbol='Gtpbp2', Aliases='nmf205', description='GTP binding protein 2', other_designations='GTP-binding-like protein 2|gtpbp2', map_location='17 C-D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46159583, end_position_on_the_genomic_accession=46169370, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  56057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56057, CurrentID=0, Status='live', Symbol='Btg4', Aliases='C86116, PC3B', description='B cell translocation gene 4', other_designations='B-cell translocation gene 4|BTG family member 4|protein PC3b', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51102265, end_position_on_the_genomic_accession=51119700, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56066, CurrentID=0, Status='live', Symbol='Cxcl11', Aliases='Cxc11, H174, I-tac, Ip9, Itac, Scyb11, Scyb9b, b-R1, betaR1', description='chemokine (C-X-C motif) ligand 11', other_designations='interferon-inducible T-cell alpha chemoattractant|small inducible cytokine subfamily B (Cys-X-Cys), member 11|small-inducible cytokine B11', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92359544, end_position_on_the_genomic_accession=92363296, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56068, CurrentID=0, Status='live', Symbol='Ammecr1', Aliases='6230420G18Rik', description='Alport syndrome, mental retardation, midface hypoplasia and elliptocytosis chromosomal region gene 1', other_designations='Alport syndrome, mental retardation, midface hypoplasia and elliptocytosis chromosomal region gene 1 homolog', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=142853474, end_position_on_the_genomic_accession=142966728, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56069, CurrentID=0, Status='live', Symbol='Il17b', Aliases='1110006O16Rik, 1700006N07Rik, Zcyto7', description='interleukin 17B', other_designations='IL-17B|cytokine CX1|cytokine-like protein ZCYTO7|gene for cytokine-like protein|neuronal interleukin-17-related factor', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61687624, end_position_on_the_genomic_accession=61692538, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56070, CurrentID=0, Status='live', Symbol='Tcerg1', Aliases='2410022J09Rik, 2900090C16Rik, AI428505, CA150b, FBP23, FBP28, Taf2s, ca150, p144', description='transcription elongation regulator 1 (CA150)', other_designations='TATA box binding protein (TBP)-associated factor, RNA polymerase II, S, 150kD|TATA box-binding protein-associated factor 2S|coactivator of 150 kD|formin-binding protein 28|transcription factor CA150|transcription factor CA150b', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42511066, end_position_on_the_genomic_accession=42575785, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  56072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56072, CurrentID=0, Status='live', Symbol='Lgals12', Aliases='AW987437, Grip1', description='lectin, galactose binding, soluble 12', other_designations='gal-12|galectin-related inhibitor of proliferation 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7596656, end_position_on_the_genomic_accession=7610466, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56075, CurrentID=0, Status='live', Symbol='Pdss1', Aliases='2610203G20Rik, 2700031G06Rik, TPT, TPT 1, Tprt, mDLP1, mSPS1', description='prenyl (solanesyl) diphosphate synthase, subunit 1', other_designations='all-trans-decaprenyl-diphosphate synthase subunit 1|decaprenyl pyrophosphate synthase subunit 1|decaprenyl pyrophosphate synthetase subunit 1|solanesyl-diphosphate synthase subunit 1|subunit 1 of solanesyl diphosphate synthase|trans-prenyltransferase 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=22893560, end_position_on_the_genomic_accession=22940266, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56077, CurrentID=0, Status='live', Symbol='Dgke', Aliases='C87606, DAGK6, DGK', description='diacylglycerol kinase, epsilon', other_designations='DAG kinase epsilon|DGK-epsilon|diglyceride kinase epsilon', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=89037582, end_position_on_the_genomic_accession=89060763, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56078, CurrentID=0, Status='live', Symbol='Car5b', Aliases='7330410H16Rik, CAVB, Ca5b, CarVb, D730005F19Rik', description='carbonic anhydrase 5b, mitochondrial', other_designations='CA-VB|carbonate dehydratase VB|carbonic anhydrase VB', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=163976822, end_position_on_the_genomic_accession=164028010, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56079, CurrentID=0, Status='live', Symbol='Astn2', Aliases='1d8, Astnl, bM452J22.1', description='astrotactin 2', other_designations='bM243L4.1 (astrotactin 2)|bM264P11.1 (astrotactin 2)|bM383P24.1 (astrotactin 2)|bM423P8.1 (astrotactin 2)|bM452J22.1 (novel protein)|bM468K2.1 (astrotactin 2)', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=65380803, end_position_on_the_genomic_accession=66404530, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  56085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56085, CurrentID=0, Status='live', Symbol='Ubqln1', Aliases='1110046H03Rik, 1810030E05Rik, AU019746, C77538, D13Ertd372e, Da41, Dsk2, Plic-1, Plic1, Xdrp1', description='ubiquilin 1', other_designations='protein linking IAP with cytoskeleton 1', map_location='13 30.95 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58176156, end_position_on_the_genomic_accession=58215653, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56086, CurrentID=0, Status='live', Symbol='Set', Aliases='2610030F17Rik, 5730420M11Rik, AA407739, I-2PP2A, StF-IT-1, TAF-I', description='SET nuclear oncogene', other_designations='SET translocation|phosphatase 2A inhibitor I2PP2A|template-activating factor I', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30061996, end_position_on_the_genomic_accession=30072577, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56087, CurrentID=0, Status='live', Symbol='Dnah10', Aliases='Dnahc10', description='dynein, axonemal, heavy chain 10', other_designations='dynein, axon, heavy chain 10', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124725018, end_position_on_the_genomic_accession=124834827, orientation='plus', exon_count=80, OMIM='', no_hdr0=''),
  56088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56088, CurrentID=0, Status='live', Symbol='Psmg1', Aliases='AW552102, Dscr2', description='proteasome (prosome, macropain) assembly chaperone 1', other_designations='Down syndrome critical region homolog 2|down syndrome critical region protein 2 homolog|psmg1', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=95979935, end_position_on_the_genomic_accession=95990903, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56089, CurrentID=0, Status='live', Symbol='Ramp3', Aliases='AI850306', description='receptor (calcitonin) activity modifying protein 3', other_designations='', map_location='11 4.61 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6658533, end_position_on_the_genomic_accession=6677476, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56092, CurrentID=0, Status='live', Symbol='Cts7', Aliases='AA408230, AA408952, CTS1, Epcs24, Epcs71', description='cathepsin 7', other_designations='cathepsin 1|cts7|ectoplacental cone, invasive trophoblast giant cells, extraembryonic ectoderm and chorion sequence 24', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61352461, end_position_on_the_genomic_accession=61358290, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56093, CurrentID=0, Status='live', Symbol='Pfpl', Aliases='Epcs5, Epcs50', description='pore forming protein-like', other_designations='ectoplacental cone, invasive trophoblast giant cells, extraembryonic ectoderm and chorion sequence 50', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12427905, end_position_on_the_genomic_accession=12432110, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  56094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56094, CurrentID=0, Status='live', Symbol='Cts8', Aliases='CTS2, Epcs68, Epcs70', description='cathepsin 8', other_designations='ectoplacental cone, invasive trophoblast giant cells, extraembryonic ectoderm and chorion sequence 68', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61246747, end_position_on_the_genomic_accession=61255348, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56095, CurrentID=0, Status='live', Symbol='Ftsj3', Aliases='AA537063, AU045295, C79843, D11Ertd400e, Epcs3', description='FtsJ RNA methyltransferase homolog 3 (E. coli)', other_designations="2'-O-ribose RNA methyltransferase SPB1 homolog|FtsJ homolog 3|ectoplacental cone, invasive trophoblast giant cells, extraembryonic ectoderm and chorion sequence 3|protein ftsJ homolog 3|putative rRNA methyltransferase 3|rRNA (uridine-2'-O-)-methyltransferase 3", map_location='11 68.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106249144, end_position_on_the_genomic_accession=106255802, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  56096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56096, CurrentID=0, Status='live', Symbol='Plac1', Aliases='DXWsu72e, Epcs26, Oosp2l', description='placental specific protein 1', other_designations='ectoplacental cone, invasive trophoblast giant cells, extraembryonic ectoderm and chorion sequence 26|plac1', map_location='X 29.31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53069993, end_position_on_the_genomic_accession=53240122, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56149, CurrentID=0, Status='live', Symbol='Grasp', Aliases='tamalin', description='GRP1 (general receptor for phosphoinositides 1)-associated scaffold protein', other_designations='GRP1-associated scaffold protein', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101224207, end_position_on_the_genomic_accession=101232756, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56150, CurrentID=0, Status='live', Symbol='Mad2l1', Aliases='AA673185, MAD2', description='MAD2 mitotic arrest deficient-like 1', other_designations='MAD2 (mitotic arrest deficient, homolog)-like 1|MAD2-like protein 1|mitotic arrest deficient 2-like protein 1', map_location='6 30.56 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=66535249, end_position_on_the_genomic_accession=66541109, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56173, CurrentID=0, Status='live', Symbol='Cldn14', Aliases='AI851731', description='claudin 14', other_designations='', map_location='16 C3-4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93919031, end_position_on_the_genomic_accession=94008837, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56174, CurrentID=0, Status='live', Symbol='Nagk', Aliases='Gnk', description='N-acetylglucosamine kinase', other_designations='GlcNAc kinase', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83795031, end_position_on_the_genomic_accession=83802556, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56175, CurrentID=0, Status='live', Symbol='Bace2', Aliases='1110059C24Rik, AEPLC, AI850424, ALP56, ARP1, ASP21, BAE2, CDA13, CEAP1, DRAP', description='beta-site APP-cleaving enzyme 2', other_designations='ASP1|asp 1|aspartyl protease 1|beta-site APP cleaving enzyme 2|beta-site amyloid precursor protein cleaving enzyme 2|memapsin-1|membrane-associated aspartic protease 1|theta-secretase', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97356728, end_position_on_the_genomic_accession=97439012, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56176, CurrentID=0, Status='live', Symbol='Pigp', Aliases='Dcrc, Dscr5', description='phosphatidylinositol glycan anchor biosynthesis, class P', other_designations='Down syndrome critical region gene c|Down syndrome critical region homolog 5|PIG-P|down syndrome critical region protein 5 homolog|phosphatidylinositol-glycan biosynthesis class P protein', map_location='16 C3-C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94358763, end_position_on_the_genomic_accession=94371015, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56177, CurrentID=0, Status='live', Symbol='Olfm1', Aliases='AMY, AW742568, Noe1, OlfA, Pancortin, Pancortin3', description='olfactomedin 1', other_designations='Noelin 1|Noelin 2|Pancortin 1-4|neuronal olfactomedin-related ER localized protein|olfactomedin related ER localized protein|olfactomedin-1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28193093, end_position_on_the_genomic_accession=28230736, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56183, CurrentID=0, Status='live', Symbol='Nmu', Aliases='', description='neuromedin U', other_designations='', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76333495, end_position_on_the_genomic_accession=76363777, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56184, CurrentID=0, Status='live', Symbol='Msgn1', Aliases='Msgn, pMsgn1', description='mesogenin 1', other_designations='pMesogenin1|paraxial mesoderm-specific expression and regulatory capacities|paraxial mesoderm-specific mesogenin1', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=11208382, end_position_on_the_genomic_accession=11208948, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56185, CurrentID=0, Status='live', Symbol='Hao2', Aliases='AI325478, Hao-2, Hao3, Haox3', description='hydroxyacid oxidase 2', other_designations='(S)-2-hydroxy-acid oxidase, peroxisomal|HAOX2|hydroxyacid oxidase (glycolate oxidase) 3|hydroxyacid oxidase 2, kidney|medium chain alpha-hydroxy acid oxidase|medium-chain 2-hydroxy acid oxidase HAOX3|medium-chain L-2-hydroxy acid oxidase', map_location='3 42.9 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98874519, end_position_on_the_genomic_accession=98895377, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56187, CurrentID=0, Status='live', Symbol='Rabggta', Aliases='gm', description='Rab geranylgeranyl transferase, a subunit', other_designations='geranylgeranyl transferase type II subunit alpha|gunmetal|rab GG transferase alpha|rab GGTase alpha|rab geranyl-geranyltransferase subunit alpha|rab geranylgeranyltransferase subunit alpha', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55715877, end_position_on_the_genomic_accession=55722368, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  56188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56188, CurrentID=0, Status='live', Symbol='Fxyd1', Aliases='0610012C17Rik, 1110006M24Rik, Plm, Pml', description='FXYD domain-containing ion transport regulator 1', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31051678, end_position_on_the_genomic_accession=31055877, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56189, CurrentID=0, Status='live', Symbol='Prodh2', Aliases='2510028N04Rik, 2510038B11Rik, MmPOX, MmPOX1, POX1', description='proline dehydrogenase (oxidase) 2', other_designations='PO|kidney and liver proline oxidase 1|probable proline oxidase 2|proline oxidase 1|proline oxidase-like protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30493586, end_position_on_the_genomic_accession=30513402, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56190, CurrentID=0, Status='live', Symbol='Rbm38', Aliases='Rnpc1, Seb4, Seb4l', description='RNA binding motif protein 38', other_designations='RNA-binding motif protein 38|RNA-binding region (RNP1, RRM) containing 1|RNA-binding region-containing protein 1|seb4 mRNA|ssDNA-binding protein SEB4', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173020503, end_position_on_the_genomic_accession=173034734, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56191, CurrentID=0, Status='live', Symbol='Tro', Aliases='AA409408, Maged3, Maged3l, Tnnl, magphinin, Tro', description='trophinin', other_designations='melanoma antigen, family D, 3|necdin and trophinin like|trophinin-2', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=150644710, end_position_on_the_genomic_accession=150657583, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  56193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56193, CurrentID=0, Status='live', Symbol='Plek', Aliases='2010300B13Rik', description='pleckstrin', other_designations='', map_location='11 9.41 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=16971206, end_position_on_the_genomic_accession=17008718, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56194, CurrentID=0, Status='live', Symbol='Prpf40a', Aliases='2810012K09Rik, FBP11, Fnbp11, Fnbp3', description='pre-mRNA processing factor 40A', other_designations='FBP-11|PRP40 pre-mRNA processing factor 40 homolog A|formin binding protein 11|formin binding protein 3|formin-binding protein 11|formin-binding protein 3|pre-mRNA processing factor 40 homolog A', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=53134706, end_position_on_the_genomic_accession=53191279, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  56195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56195, CurrentID=0, Status='live', Symbol='Ptbp2', Aliases='Ptb2, brPTB, nPTB', description='polypyrimidine tract binding protein 2', other_designations='RRM-type RNA-binding protein brPTB|brain-enriched PTB|brain-enriched polypyrimidine tract-binding protein|neural polypyrimidine tract-binding protein', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=119718741, end_position_on_the_genomic_accession=119783799, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56196, CurrentID=0, Status='live', Symbol='Tdp2', Aliases='D13Ertd656e, Ttrap', description='tyrosyl-DNA phosphodiesterase 2', other_designations="5'-Tyr-DNA phosphodiesterase|5'-tyrosyl-DNA phosphodiesterase|TRAF and TNF receptor associated protein|TRAF and TNF receptor-associated protein|tyr-DNA phosphodiesterase 2", map_location='13 10.7 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24831647, end_position_on_the_genomic_accession=24842153, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56198, CurrentID=0, Status='live', Symbol='Heyl', Aliases='Hey3, Hrt3, bHLHb33, hesr3', description='hairy/enhancer-of-split related with YRPW motif-like', other_designations='HRT-3|hairy and enhance of split related 3|hairy and enhancer of split-related protein 3|hairy-related transcription factor 3|mHRT3', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123233556, end_position_on_the_genomic_accession=123249870, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56199, CurrentID=0, Status='live', Symbol='Abcb10', Aliases='Abc-me, Abcb12', description='ATP-binding cassette, sub-family B (MDR/TAP), member 10', other_designations='ABC-mitochondrial erythroid protein|ATP-binding cassette transporter 10', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123952459, end_position_on_the_genomic_accession=123983122, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56200, CurrentID=0, Status='live', Symbol='Ddx21', Aliases='AI255159, AL022742, D10Ertd645e, D10Wsu42e', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 21', other_designations='DEAD box protein 21|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 21 (RNA helicase II/Gu)|RH II/Gu|RH-II/Gualpha|gu-alpha|nucleolar RNA helicase Gu|nucleolar RNA helicase II', map_location='10 B4-B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62580247, end_position_on_the_genomic_accession=62602298, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56205, CurrentID=0, Status='live', Symbol='Ensa', Aliases='1700020C18Rik, 2610007F17Rik, AI451924', description='endosulfine alpha', other_designations='ARPP-19e', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95624980, end_position_on_the_genomic_accession=95632117, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56207, CurrentID=0, Status='live', Symbol='Uchl5', Aliases='5830413B11Rik, Uch37', description='ubiquitin carboxyl-terminal esterase L5', other_designations='UCH-L5|ubiquitin C-terminal hydrolase 37|ubiquitin C-terminal hydrolase UCH37|ubiquitin thioesterase L5', map_location='1 62.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=143777221, end_position_on_the_genomic_accession=143807466, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56208, CurrentID=0, Status='live', Symbol='Becn1', Aliases='4921513J16Rik, 5430417M23Rik, Atg6', description='beclin 1, autophagy related', other_designations='Bcl-2-interacting protein beclin|coiled-coil myosin-like BCL2-interacting protein|coiled-coil, myosin-like BCL2-interacting protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101288262, end_position_on_the_genomic_accession=101302288, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56209, CurrentID=0, Status='live', Symbol='Gde1', Aliases='1200003M13Rik, Mir16, RGS16', description='glycerophosphodiester phosphodiesterase 1', other_designations='membrane interacting protein of RGS16|membrane-interacting protein of RGS16', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118688558, end_position_on_the_genomic_accession=118705738, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56210, CurrentID=0, Status='live', Symbol='Rev1', Aliases='1110027I23Rik, AU022044l, Rev1', description='REV1 homolog (S. cerevisiae)', other_designations='rev1-like terminal deoxycytidyl transferase', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=38052786, end_position_on_the_genomic_accession=38129919, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  56212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56212, CurrentID=0, Status='live', Symbol='Rhog', Aliases='2810426G09Rik, Arhg, Sid10750', description='ras homolog gene family, member G', other_designations='aplysia ras-related homolog G (RhoG)|ras homolog G (RhoG)|sid 10750', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102239123, end_position_on_the_genomic_accession=102250118, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56213, CurrentID=0, Status='live', Symbol='Htra1', Aliases='AI429470, HTRA, L56, Prss11, RSPP11', description='HtrA serine peptidase 1', other_designations='high-temperature requirement A serine peptidase 1|insulin-like growth factor binding protein 5 protease|protease, serine, 11 (Igf binding)|serine protease 11', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=130936203, end_position_on_the_genomic_accession=130985658, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56214, CurrentID=0, Status='live', Symbol='Scamp4', Aliases='2410022D05Rik, AI448940, AU022392, AW742369, Sc4', description='secretory carrier membrane protein 4', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80602882, end_position_on_the_genomic_accession=80615783, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56215, CurrentID=0, Status='live', Symbol='Acin1', Aliases='2610036I19Rik, 2610510L13Rik, Acinus, Acn, C79325, acinusL, acinusS, mKIAA0670', description='apoptotic chromatin condensation inducer 1', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54642161, end_position_on_the_genomic_accession=54686931, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  56216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56216, CurrentID=0, Status='live', Symbol='Stx1b', Aliases='Stx1b2l, Syn1b, Stx1b', description='syntaxin 1B', other_designations='syntaxin 1B1|syntaxin 1B2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127803900, end_position_on_the_genomic_accession=127824531, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56217, CurrentID=0, Status='live', Symbol='Mpp5', Aliases='3830420B02Rik, AI255216, AI644496, Pals1', description='membrane protein, palmitoylated 5 (MAGUK p55 subfamily member 5)', other_designations='membrane protein, palmitoylated 3 (MAGUK p55 subfamily member 5)|protein associated with Lin-7 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=78748912, end_position_on_the_genomic_accession=78840877, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  56218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56218, CurrentID=0, Status='live', Symbol='Patz1', Aliases='8430401L15Rik, AI447519, Mazr, Patz, Zfp278', description='POZ (BTB) and AT hook containing zinc finger 1', other_designations='POZ-AT hook-zinc finger protein|transcription factor (MAZ-related) MAZR|zinc finger protein 278', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3289131, end_position_on_the_genomic_accession=3309083, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56219, CurrentID=0, Status='live', Symbol='Extl1', Aliases='D430033M16Rik', description='exostoses (multiple)-like 1', other_designations='Exostosin-like 1 (Exostosin-L) (Multiple exostosis-like protein)|exostoses-like 1|exostosin-L|glucuronosyl-N-acetylglucosaminyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase|multiple exostoses-like 1|multiple exostosis-like protein', map_location='4 66.63 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134356373, end_position_on_the_genomic_accession=134372547, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56220, CurrentID=0, Status='live', Symbol='Zfp386', Aliases='Kzf1, mKIAA4205', description='zinc finger protein 386 (Kruppel-like)', other_designations='Kruppel associated box (KRAB) zinc finger 1|kruppel-like protein', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=116047707, end_position_on_the_genomic_accession=116063209, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56221, CurrentID=0, Status='live', Symbol='Ccl24', Aliases='CKb-6, MPIF-2, Scya24', description='chemokine (C-C motif) ligand 24', other_designations='CC chemokine CCL24|eosinophil chemotactic protein 2|eotaxin-2|small inducible cytokine A24|small-inducible cytokine A24', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135569937, end_position_on_the_genomic_accession=135573122, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56222, CurrentID=0, Status='live', Symbol='Cited4', Aliases='MRG-2, Mrg2', description='Cbp/p300-interacting transactivator, with Glu/Asp-rich carboxy-terminal domain, 4', other_designations='cited4|msg1-related protein 2', map_location='4 56.52 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120666563, end_position_on_the_genomic_accession=120667820, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  56223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56223, CurrentID=0, Status='live', Symbol='Fscn3', Aliases='', description='fascin actin-bundling protein 3', other_designations='fascin (testis)|fascin homolog 3, actin-bundling protein, testicular|testis fascin', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=28427870, end_position_on_the_genomic_accession=28439398, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56224, CurrentID=0, Status='live', Symbol='Tspan5', Aliases='2810455A09Rik, 4930505M03Rik, NET-4, Tm4sf9', description='tetraspanin 5', other_designations='tetraspan NET-4|tetraspanin Tspan-5|transmembrane 4 superfamily member 9', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138742208, end_position_on_the_genomic_accession=138904433, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56226, CurrentID=0, Status='live', Symbol='Espn', Aliases='je', description='espin', other_designations='actin cytoskeletal regulatory protein|ectoplasmic specialization protein|jerker', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152120331, end_position_on_the_genomic_accession=152152420, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  56228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56228, CurrentID=0, Status='live', Symbol='Ube2j1', Aliases='0710008M05Rik, 1110030I22Rik, NCUBE-1, Ncube, Ncube1, Ubc6p', description='ubiquitin-conjugating enzyme E2J 1', other_designations='E2 ubiquitin-conjugating enzyme J1|non-canonical ubiquitin conjugating enzyme 1|non-canonical ubiquitin-conjugating enzyme 1|ubiquitin-conjugating enzyme E2, J1', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=33031392, end_position_on_the_genomic_accession=33052364, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56229, CurrentID=0, Status='live', Symbol='Thsd1', Aliases='4833423O18Rik, Tmtsp', description='thrombospondin, type I, domain 1', other_designations='transmembrane molecule with thrombospondin module', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22225858, end_position_on_the_genomic_accession=22261334, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56233, CurrentID=0, Status='live', Symbol='Hdac7', Aliases='5830434K02Rik, HD7, HD7aa, mFLJ00062, Hdac7', description='histone deacetylase 7', other_designations='histone deacetylase 7A', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97792664, end_position_on_the_genomic_accession=97844502, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  56248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56248, CurrentID=0, Status='live', Symbol='Ak3', Aliases='AA407498, AI506714, AK-3l, Ak3l1, Akl3l, Ak3', description='adenylate kinase 3', other_designations='AK 3|GTP:AMP phosphotransferase mitochondrial|GTP:AMP phosphotransferase, mitochondrial|adenylate kinase 3 alpha-like 1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=28995380, end_position_on_the_genomic_accession=29048672, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56249, CurrentID=0, Status='live', Symbol='Actr8', Aliases='5730542K05Rik, ARP8', description='ARP8 actin-related protein 8', other_designations='ARP8 actin-related protein 8 homolog|actin-related protein 8 homolog', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=29978330, end_position_on_the_genomic_accession=29993221, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56258, CurrentID=0, Status='live', Symbol='Hnrnph2', Aliases="DXHXS1271E, Ftp-3, Ftp3, H', HNRNP, Hnrph2", description='heterogeneous nuclear ribonucleoprotein H2', other_designations="heterogeneous nuclear ribonucleoprotein H'|hnRNP H'|hnRNP H2", map_location='X 56.2 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134601184, end_position_on_the_genomic_accession=134607060, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56264, CurrentID=0, Status='live', Symbol='Cpxm1', Aliases='AA986902, Cpx-1, Cpx1, Cpxm', description='carboxypeptidase X 1 (M14 family)', other_designations='carboxypeptidase X (M14 family)|metallocarboxypeptidase 1|metallocarboxypeptidase CPX-1', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130390775, end_position_on_the_genomic_accession=130397629, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56273, CurrentID=0, Status='live', Symbol='Pex14', Aliases='Pex14p, R75137', description='peroxisomal biogenesis factor 14', other_designations='PTS1 receptor-docking protein|peroxin-14|peroxisomal membrane anchor protein PEX14|peroxisomal protein Pex14', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148960535, end_position_on_the_genomic_accession=149099812, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56274, CurrentID=0, Status='live', Symbol='Stk3', Aliases='0610042I06Rik, MST, Mst2, Mst3, Ste20, mess1', description='serine/threonine kinase 3', other_designations='MST-2|STE20-like kinase MST2|mammalian STE20-like protein kinase 2|protein kinase homolog', map_location='15 B3.3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34875496, end_position_on_the_genomic_accession=35155806, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56275, CurrentID=0, Status='live', Symbol='Rbm14', Aliases='1300007E16Rik, PSP2, Sytip, p16, p16K', description='RNA binding motif protein 14', other_designations='synaptotagmin interacting protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4800566, end_position_on_the_genomic_accession=4811634, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56277, CurrentID=0, Status='live', Symbol='Tmem45a', Aliases='C630002M10Rik, M32486, p19.5', description='transmembrane protein 45a', other_designations='19.5 mRNA|dermal papilla-derived protein 7 homolog', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56805161, end_position_on_the_genomic_accession=56886458, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56278, CurrentID=0, Status='live', Symbol='Gkap1', Aliases='42kDa, 4933400B15Rik, D13Ertd340e, Gkap42', description='G kinase anchoring protein 1', other_designations='42 kD cGMP-dependent protein kinase anchoring protein|cGMP-dependent protein kinase anchoring protein|cGMP-dependent protein kinase-anchoring protein of 42 kDa', map_location='13 30.97 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58233349, end_position_on_the_genomic_accession=58274876, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56279, CurrentID=0, Status='live', Symbol='Fam69b', Aliases='AB030186, AW209000, B230317C12Rik, PIP49', description='family with sequence similarity 69, member B', other_designations='pancreatitis-induced protein 49', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26609855, end_position_on_the_genomic_accession=26636497, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56280, CurrentID=0, Status='live', Symbol='Mrpl37', Aliases='2300004O14Rik, AI132596, MRP-L2, Rpml2', description='mitochondrial ribosomal protein L37', other_designations='L37mt|MRP-L37|ribosomal protein, mitochondrial, L2', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107055874, end_position_on_the_genomic_accession=107066866, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56282, CurrentID=0, Status='live', Symbol='Mrpl12', Aliases='0610034O11Rik, 1500031N16Rik, MRP-L12, Rpml12', description='mitochondrial ribosomal protein L12', other_designations='L12mt', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120484669, end_position_on_the_genomic_accession=120488754, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56284, CurrentID=0, Status='live', Symbol='Mrpl19', Aliases='9030416F12Rik, 9130412E02Rik, D6Ertd157e, MRP-L15, RLX1, Rpml15', description='mitochondrial ribosomal protein L19', other_designations='L19mt|MRP-L19|ribosomal protein, mitochondrial, L15', map_location='6 35.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=81957826, end_position_on_the_genomic_accession=81965949, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56289, CurrentID=0, Status='live', Symbol='Rassf1', Aliases='123F2, AA536941, AU044980, NORE2A, RDA32, REH3P21A, Rassf1B, Rassf1C, Rassf1', description='Ras association (RalGDS/AF-6) domain family member 1', other_designations='123F protein|RASSF1A tumor suppressor|Ras association (RalGDS/AF-6) domain family 1|protein 123F2|rassf1', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107551555, end_position_on_the_genomic_accession=107562267, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56291, CurrentID=0, Status='live', Symbol='Styx', Aliases='0610039A20Rik, hStyxb', description='serine/threonine/tyrosine interaction protein', other_designations='phosphoserine/threonine/tyrosine interaction protein|protein tyrosine phosphatase-like protein|protein tyrosine phosphatase-like unspliced c-terminal product and spliced c-terminal end STYX', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45351186, end_position_on_the_genomic_accession=45376884, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56292, CurrentID=0, Status='live', Symbol='Naa10', Aliases='2310039H09Rik, Ard1, Ard1a, Te2', description='N(alpha)-acetyltransferase 10, NatA catalytic subunit', other_designations='ARD1 homolog A, N-acetyltransferase|N(alpha)-acetyltransferase 10, NatA catalytic subunitNalpha acetyltransferase 10|N-alpha-acetyltransferase 10, NatA catalytic subunit|N-terminal acetyltransferase complex ARD1 subunit homolog A|Nalpha acetyltransferase 10|alpha-N-acetyltransferase 1A|natA catalytic subunit Naa10', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73916870, end_position_on_the_genomic_accession=73921944, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56293, CurrentID=0, Status='live', Symbol='Slc35g3', Aliases='Amac1', description='solute carrier family 35, member G3', other_designations='acyl-malonyl condensing enzyme 1|acyl-malonyl-condensing enzyme 1|protein AMAC1|testicular condensing enzyme', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69759884, end_position_on_the_genomic_accession=69761844, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56294, CurrentID=0, Status='live', Symbol='Ptpn9', Aliases='MEG2', description='protein tyrosine phosphatase, non-receptor type 9', other_designations='PTPase MEG2|Protein-tyrosine phosphatase, non-receptor type 9 (Protein-tyrosine phosphatase MEG2) (PTPase-MEG2)|protein-tyrosine phosphatase MEG2', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56994968, end_position_on_the_genomic_accession=57062808, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56295, CurrentID=0, Status='live', Symbol='Higd1a', Aliases='2210020B17Rik, 7420700H20Rik, AI303276, AW049839, HIMP1, Hig1', description='HIG1 domain family, member 1A', other_designations='hypoxia induced gene 1|hypoxia-inducible gene 1 protein', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121848560, end_position_on_the_genomic_accession=121858357, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56296, CurrentID=0, Status='live', Symbol='Dmrtb1', Aliases='Dmrt6, Prp13', description='DMRT-like family B with proline-rich C-terminal, 1', other_designations='doublesex and mab-3 related transcription factor 6|doublesex- and mab-3-related transcription factor 6|proline-rich protein 13', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107676288, end_position_on_the_genomic_accession=107684243, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56297, CurrentID=0, Status='live', Symbol='Arl6', Aliases='1110018H24Rik, 2210411E14Rik, BBS3', description='ADP-ribosylation factor-like 6', other_designations='ADP-ribosylation-like 6|ADP-ribosylation-like factor homolog ARL6', map_location='16 C1.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59613321, end_position_on_the_genomic_accession=59639405, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56298, CurrentID=0, Status='live', Symbol='Atl2', Aliases='2010110I21Rik, AA407293, AV334690, Aip-2, Arl6ip2', description='atlastin GTPase 2', other_designations='ADP-ribosylation factor-like 6 interacting protein 2|ADP-ribosylation factor-like protein 6-interacting protein 2|ADP-ribosylation-like factor 6 interacting protein 2|ARL-6 interacting protein-2|ARL-6-interacting protein 2', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=79848352, end_position_on_the_genomic_accession=79896123, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56299, CurrentID=0, Status='live', Symbol='Fkbpl', Aliases='DIR1, NG7, Ppiase-X, Wisp39', description='FK506 binding protein-like', other_designations='WAF-1/CIP1 stabilizing protein 39|immunophilin-like protein NG7', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34639574, end_position_on_the_genomic_accession=34646327, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56304, CurrentID=0, Status='live', Symbol='Ce9', Aliases='', description='immunoglobulin light chain variable region', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  56305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56305, CurrentID=0, Status='live', Symbol='Pitpnb', Aliases='PI-TP-beta', description='phosphatidylinositol transfer protein, beta', other_designations='Pitp beta|phosphotidylinositol transfer protein, beta|ptdIns transfer protein beta|ptdInsTP beta', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=111330697, end_position_on_the_genomic_accession=111388361, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56306, CurrentID=0, Status='live', Symbol='Fam60a', Aliases='Ppcs1, Pptcs1, Tera', description='family with sequence similarity 60, member A', other_designations='pluripotent cell-specific 1|teratocarcinoma expressed, serine rich', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148921045, end_position_on_the_genomic_accession=148946432, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56307, CurrentID=0, Status='live', Symbol='Metap2', Aliases='4930584B20Rik, A930035J23Rik, AI047573, AL024412, AU014659, Amp2, Mnpep, p67, p67eIF2', description='methionine aminopeptidase 2', other_designations='MAP 2|eIF-2-associated p67|initiation factor 2-associated 67 kDa glycoprotein|initiation factor 2-associated protein (p67)|metAP 2|p67|p67eIF2|peptidase M 2', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93858489, end_position_on_the_genomic_accession=93897090, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56309, CurrentID=0, Status='live', Symbol='Mycbp', Aliases='5730488M09Rik, AW552132, AW742590, Amy-1', description='c-myc binding protein', other_designations='associate of Myc 1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123905013, end_position_on_the_genomic_accession=123912250, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56310, CurrentID=0, Status='live', Symbol='Gps2', Aliases='AI505953', description='G protein pathway suppressor 2', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69913980, end_position_on_the_genomic_accession=69917150, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56312, CurrentID=0, Status='live', Symbol='Nupr1', Aliases='2310032H04Rik, Com1, p8', description='nuclear protein transcription regulator 1', other_designations='protein p8', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126623246, end_position_on_the_genomic_accession=126625470, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56314, CurrentID=0, Status='live', Symbol='Zfp113', Aliases='4732456B05Rik, mKIAA4229', description='zinc finger protein 113', other_designations='', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138139702, end_position_on_the_genomic_accession=138155744, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56315, CurrentID=0, Status='live', Symbol='Rhcg', Aliases='BB065800', description='Rhesus blood group-associated C glycoprotein', other_designations='Rh type C glycoprotein|rh family type C glycoprotein|rhesus blood group family type C glycoprotein|rhesus-associated C glycoprotein', map_location='7 45.09 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79593363, end_position_on_the_genomic_accession=79617657, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56316, CurrentID=0, Status='live', Symbol='Ggcx', Aliases='', description='gamma-glutamyl carboxylase', other_designations='peptidyl-glutamate 4-carboxylase|vitamin K gamma glutamyl carboxylase', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72414296, end_position_on_the_genomic_accession=72431106, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  56317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56317, CurrentID=0, Status='live', Symbol='Anapc7', Aliases='APC7, AW545589', description='anaphase promoting complex subunit 7', other_designations='cyclosome subunit 7|prediabetic NOD sera-reactive autoantigen', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122422444, end_position_on_the_genomic_accession=122444912, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56318, CurrentID=0, Status='live', Symbol='Acpp', Aliases="5'-NT, A030005E02Rik, FRAP, Lap, PAP, Ppal", description='acid phosphatase, prostate', other_designations="5'-nucleotidase|TMPase|ecto-5'-nucleotidase|fluoride-resistant acid phosphatase|lysosomal acid phosphatase|thiamine monophosphatase", map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=104288240, end_position_on_the_genomic_accession=104337722, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56320, CurrentID=0, Status='live', Symbol='Dbn1', Aliases='', description='drebrin 1', other_designations='developmentally-regulated brain protein|drebrin A|drebrin E2', map_location='13 30.06 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55473428, end_position_on_the_genomic_accession=55488163, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56321, CurrentID=0, Status='live', Symbol='Aatf', Aliases='4933415H02Rik, 5830465M17Rik, Che-1, Trb', description='apoptosis antagonizing transcription factor', other_designations='apoptosis-antagonizing transcription factor|rb-binding protein Che-1|traube protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84422856, end_position_on_the_genomic_accession=84513570, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  56322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56322, CurrentID=0, Status='live', Symbol='Timm22', Aliases='Tim22', description='translocase of inner mitochondrial membrane 22', other_designations='preprotein translocase|translocase of inner mitochondrial membrane 22 homolog', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76406925, end_position_on_the_genomic_accession=76416313, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56323, CurrentID=0, Status='live', Symbol='Dnajb5', Aliases='1110058L06Rik, AI462558, Hsc40, Hsp40-3', description='DnaJ heat shock protein family (Hsp40) member B5', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 5|heat shock protein cognate 40|heat shock protein hsp40-3', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42949827, end_position_on_the_genomic_accession=42959425, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56324, CurrentID=0, Status='live', Symbol='Stam2', Aliases='1200004O12Rik, 5730456G07Rik, BB145015, Hbp', description='signal transducing adaptor molecule (SH3 domain and ITAM motif) 2', other_designations='Hrs binding protein|STAM-2|hrs-binding protein', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=52692198, end_position_on_the_genomic_accession=52742149, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56325, CurrentID=0, Status='live', Symbol='Abcb9', Aliases='TAPL, mKIAA1520', description='ATP-binding cassette, sub-family B (MDR/TAP), member 9', other_designations='ABC transporter 9 protein|ATP-binding cassette transporter 9|ATP-binding cassette, sub-family B, member 9|TAP-like protein|mABCB9', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124061857, end_position_on_the_genomic_accession=124095966, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56327, CurrentID=0, Status='live', Symbol='Arl2', Aliases='2610009M23Rik, AI115441, AW553335', description='ADP-ribosylation factor-like 2', other_designations='ADP-ribosylation-like 2|arf-like protein 2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6134389, end_position_on_the_genomic_accession=6141137, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56330, CurrentID=0, Status='live', Symbol='Pdcd5', Aliases='2200003D22Rik, AA408513, Tfar19', description='programmed cell death 5', other_designations='TF-1 cell apoptosis-related protein 19', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35641985, end_position_on_the_genomic_accession=35647482, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56332, CurrentID=0, Status='live', Symbol='Amotl2', Aliases='AW549739, C79691, Lccp, MASCOT, mKIAA0989', description='angiomotin-like 2', other_designations='Leman coiled-coil protein|angiomotin like 2', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=102717294, end_position_on_the_genomic_accession=102733418, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56334, CurrentID=0, Status='live', Symbol='Tmed2', Aliases='1110032D12Rik, 1810020N21Rik, Rnp24, Sid394, p24beta1', description='transmembrane emp24 domain trafficking protein 2', other_designations='COPI-coated vesicle membrane protein p24|Sid394p|coated vesicle membrane protein|membrane protein p24A|p24 family protein beta-1|sid 394', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124540791, end_position_on_the_genomic_accession=124550503, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56335, CurrentID=0, Status='live', Symbol='Mettl3', Aliases='2310024F18Rik, M6A, Spo8', description='methyltransferase like 3', other_designations='MT-A70|N6-adenosine-methyltransferase 70 kDa subunit|m6a methyltransferase|methyltransferase-like 3|methyltransferase-like protein 3', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52294841, end_position_on_the_genomic_accession=52305124, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56336, CurrentID=0, Status='live', Symbol='B4galt5', Aliases='9430078I07Rik, AW049941, AW539721', description='UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 5', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 5|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 5|b4Gal-T5|beta-1,4-GalT II|beta-1,4-GalTase 5|beta-1,4-galactosyltransferase V|beta4Gal-T5', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167298444, end_position_on_the_genomic_accession=167349178, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56338, CurrentID=0, Status='live', Symbol='Txnip', Aliases='1200008J08Rik, AA682105, Hyplip1, THIF, Tbp-2, VDUP1', description='thioredoxin interacting protein', other_designations='hyperlipidemia 1|thioredoxin binding protein-2|upregulated by 1,25-dihydroxyvitamin D-3|vitamin D3 up-regulated protein 1', map_location='3 41.93 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96557957, end_position_on_the_genomic_accession=96561857, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56347, CurrentID=0, Status='live', Symbol='Eif3c', Aliases='110kDa, 3230401O13Rik, Eif3s8, NIPIL(A3), NipilA3, Xs, Xsl', description='eukaryotic translation initiation factor 3, subunit C', other_designations='NIPI-like protein|eIF3 p110|eIF3c|eukaryotic translation initiation factor 3 subunit 8|eukaryotic translation initiation factor 3, subunit 8 (110 kDa)|extra-toes spotting', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126546911, end_position_on_the_genomic_accession=126566366, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  56348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56348, CurrentID=0, Status='live', Symbol='Hsd17b12', Aliases='2610510O05Rik, AI172963, KIK-I, Kik1', description='hydroxysteroid (17-beta) dehydrogenase 12', other_designations='17-beta-HSD 12|17-beta-hydroxysteroid dehydrogenase 12|3-ketoacyl-CoA reductase|KAR|estradiol 17-beta-dehydrogenase 12|keratoadhesin|keratonectin|steroid dehydrogenase', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=94032697, end_position_on_the_genomic_accession=94157909, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56349, CurrentID=0, Status='live', Symbol='Net1', Aliases='Net1a, mNET1', description='neuroepithelial cell transforming gene 1', other_designations='Net1 homolog|guanine nucleotide regulatory protein (oncogene)|rho guanine nucleotide exchange factor 8', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3882018, end_position_on_the_genomic_accession=3918220, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56350, CurrentID=0, Status='live', Symbol='Arl3', Aliases='', description='ADP-ribosylation factor-like 3', other_designations='ADP-ribosylation-like 3', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46531105, end_position_on_the_genomic_accession=46573181, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56351, CurrentID=0, Status='live', Symbol='Ptges3', Aliases='5730442A20Rik, Ptges, Tebp, cPGES, p23, sid3177', description='prostaglandin E synthase 3 (cytosolic)', other_designations='cytosolic prostaglandin E2 synthase|hsp90 co-chaperone|p23 cochaperone|progesterone receptor complex p23|sid 3177|telomerase binding protein, p23|telomerase-binding protein p23', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128058982, end_position_on_the_genomic_accession=128077254, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56353, CurrentID=0, Status='live', Symbol='Rybp', Aliases='2410018J24Rik, DEDAF, YEAF1', description='RING1 and YY1 binding protein', other_designations='DED-associated factor|Ring1 interactor Rybp|death effector domain-associated factor', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=100228565, end_position_on_the_genomic_accession=100287358, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56354, CurrentID=0, Status='live', Symbol='Dnajc7', Aliases='2010003F24Rik, 2010004G07Rik, CCRP, Ttc2, mDj11, mTpr2', description='DnaJ heat shock protein family (Hsp40) member C7', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 7|TPR repeat protein 2|cytoplasmic CAR retention protein|tetratricopeptide repeat domain 2|tetratricopeptide repeat protein 2', map_location='11 63.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100582818, end_position_on_the_genomic_accession=100620168, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56356, CurrentID=0, Status='live', Symbol='Gltp', Aliases='1110001F24Rik, C76925, C77564', description='glycolipid transfer protein', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114669490, end_position_on_the_genomic_accession=114690935, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56357, CurrentID=0, Status='live', Symbol='Ivd', Aliases='1300016K07Rik, 6720455E18Rik, AI463340', description='isovaleryl coenzyme A dehydrogenase', other_designations='isovaleryl dehydrogenase', map_location='2 E4-E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118861959, end_position_on_the_genomic_accession=118882907, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56358, CurrentID=0, Status='live', Symbol='Copz2', Aliases='1110012D12Rik, zeta2-COP', description='coatomer protein complex, subunit zeta 2', other_designations='nonclathrin coat protein zeta2-COP|zeta-2 COP|zeta-2-coat protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96849876, end_position_on_the_genomic_accession=96861203, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56360, CurrentID=0, Status='live', Symbol='Acot9', Aliases='0610041P13Rik, Acate2, C76421, MT-ACT48, MTE-2, U8, p48', description='acyl-CoA thioesterase 9', other_designations='acyl coenzyme A thioester hydrolase 2|acyl-CoA thioester hydrolase 9|acyl-Coenzyme A thioesterase 2, mitochondrial|mitochondrial 48 kDa acyl-CoA thioester hydrolase 1|mitochondrial acyl-CoA thioesterase|protein U8', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155262431, end_position_on_the_genomic_accession=155297651, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  56361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56361, CurrentID=0, Status='live', Symbol='Pus1', Aliases='A730013B20Rik, MPUS1, mPus1p', description='pseudouridine synthase 1', other_designations='tRNA pseudouridine(38-40) synthase|tRNA pseudouridylate synthase I|tRNA-uridine isomerase I', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110773667, end_position_on_the_genomic_accession=110781056, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56362, CurrentID=0, Status='live', Symbol='Sult1b1', Aliases='ST1B1, St2b2', description='sulfotransferase family 1B, member 1', other_designations='dopa/tyrosine sulfotransferase|sulfotransferase 1B1', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87513339, end_position_on_the_genomic_accession=87538195, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56363, CurrentID=0, Status='live', Symbol='Tmeff2', Aliases='4832418D20Rik, 7630402F16Rik, TR-2', description='transmembrane protein with EGF-like and two follistatin-like domains 2', other_designations='', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=50909962, end_position_on_the_genomic_accession=51187270, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  56364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56364, CurrentID=0, Status='live', Symbol='Zmym3', Aliases='9030216B10Rik, AW122925, DXS6673El, Zfp261', description='zinc finger, MYM-type 3', other_designations='zinc finger protein 261', map_location='X C-D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101404383, end_position_on_the_genomic_accession=101421184, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  56365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56365, CurrentID=0, Status='live', Symbol='Clcnkb', Aliases='CLCKB, Clc-Ka, Clck2, Clcnk1l', description='chloride channel, voltage-sensitive Kb', other_designations='chloride channel K1-like; putative basolateral cTAL chloride channel ClC-Ka|chloride channel Kb|clC-K2|putative basolateral mTAL chloride channel ClC-Ka', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141404357, end_position_on_the_genomic_accession=141416013, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  56367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56367, CurrentID=0, Status='live', Symbol='Scoc', Aliases='0610011C06Rik, 2810002J11Rik, AA415030, SCOCO', description='short coiled-coil protein', other_designations='', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83434492, end_position_on_the_genomic_accession=83458437, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56368, CurrentID=0, Status='live', Symbol='Cyb561d2', Aliases='AW049681, Tsp10', description='cytochrome b-561 domain containing 2', other_designations='101F6 protein|gene for 101F6 protein|putative tumor suppressor protein 101F6|tumor suppressor, region 10', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107539005, end_position_on_the_genomic_accession=107542188, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56369, CurrentID=0, Status='live', Symbol='Apip', Aliases='APIP2, CGI-29, Mmrp19', description='APAF1 interacting protein', other_designations='APAF1-interacting protein|MTRu-1-P dehydratase|monocyte macrophage 19|monocyte/macrophage protein 19|probable methylthioribulose-1-phosphate dehydratase', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103073500, end_position_on_the_genomic_accession=103092649, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56370, CurrentID=0, Status='live', Symbol='Tagln3', Aliases='2700038H05Rik, 2900005O10Rik, AI426007, Np25', description='transgelin 3', other_designations='neuronal protein Np25', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45711230, end_position_on_the_genomic_accession=45724531, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56371, CurrentID=0, Status='live', Symbol='Fzr1', Aliases='AW108046, Cdh1, FZR, FZR2, Fyr, HCDH, HCDH1', description='fizzy/cell division cycle 20 related 1 (Drosophila)', other_designations='cdh1/Hct1 homolog|cell division cycle 20 related 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81366879, end_position_on_the_genomic_accession=81378370, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56372, CurrentID=0, Status='live', Symbol='1110004F10Rik', Aliases='Smacp, Smap, sid2057, sid2057p', description='RIKEN cDNA 1110004F10 gene', other_designations='sid 2057|small acidic protein sid2057p', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=116093380, end_position_on_the_genomic_accession=116105210, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56373, CurrentID=0, Status='live', Symbol='Cpb2', Aliases='1110032P04Rik, 4930405E17Rik, AI255929, CPR, Cpu, TAFI', description='carboxypeptidase B2 (plasma)', other_designations='carboxypeptidase R|carboxypeptidase U|thrombin-activable fibrinolysis inhibitor|thrombin-activatable fibrinolysis inhibitor', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75242287, end_position_on_the_genomic_accession=75283555, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56374, CurrentID=0, Status='live', Symbol='Tmem59', Aliases='1110001M20Rik, 3110046P06Rik, AI256529, D4Ertd20e, MTDCF1, ORF18', description='transmembrane protein 59', other_designations='thymic dendritic cell-derived factor 1', map_location='4 50.12 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107178630, end_position_on_the_genomic_accession=107200996, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56375, CurrentID=0, Status='live', Symbol='B4galt4', Aliases='9130402O08Rik, B4galt-IV, b4Gal-T4, beta4Gal-T4, beta4GalT-IV', description='UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 4', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 4|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 4|beta-1,4-GalTase 4|beta-1,4-galactosyltransferase IV', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38742103, end_position_on_the_genomic_accession=38769054, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56376, CurrentID=0, Status='live', Symbol='Pdlim5', Aliases='1110001A05Rik, AI987914, C87059, Enh, Enh1, Enh2, Enh3, LIM', description='PDZ and LIM domain 5', other_designations='enigma homolog|enigma-like PDZ and LIM domains protein', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142239585, end_position_on_the_genomic_accession=142395696, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  56378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56378, CurrentID=0, Status='live', Symbol='Arpc3', Aliases='1110006A04Rik, p21-ARC, p21-Ar, p21Arc', description='actin related protein 2/3 complex, subunit 3', other_designations='Arp2/3 complex subunit p21-Arc|arp2/3 complex 21 kDa subunit', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122391928, end_position_on_the_genomic_accession=122406178, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56379, CurrentID=0, Status='live', Symbol='Kcnj1', Aliases='Kir1.1, ROMK, Romk2', description='potassium inwardly-rectifying channel, subfamily J, member 1', other_designations='ATP-regulated potassium channel ROM-K|inward rectifier K(+) channel Kir1.1|inwardly rectifying potassium channel ROMK-2|potassium channel, inwardly rectifying subfamily J member 1', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=32372418, end_position_on_the_genomic_accession=32399197, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56380, CurrentID=0, Status='live', Symbol='Arid3b', Aliases='Bdp, Dri2', description='AT rich interactive domain 3B (BRIGHT-like)', other_designations='ARID domain-containing protein 3B|AT rich interactive domain 3B (Bright like)|bright and dead ringer gene product homologous protein Bdp|bright and dead ringer protein|bright-like protein|dead ringer homolog 2', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57790353, end_position_on_the_genomic_accession=57836847, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56381, CurrentID=0, Status='live', Symbol='Spen', Aliases='Mint, mKIAA0929', description='SPEN homolog, transcriptional regulator (Drosophila)', other_designations='Msx2 interacting nuclear target protein|SMART/HDAC1-associated repressor protein', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141467885, end_position_on_the_genomic_accession=141538635, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  56382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56382, CurrentID=0, Status='live', Symbol='Rab9', Aliases='2410064E05Rik, AI195561a, Sid6061p, Rab9', description='RAB9, member RAS oncogene family', other_designations='SID 99|rab-9A', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=166457251, end_position_on_the_genomic_accession=166479867, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56384, CurrentID=0, Status='live', Symbol='Letm1', Aliases='', description='leucine zipper-EF-hand containing transmembrane protein 1', other_designations='leucine zipper-EF-hand-containing transmembrane protein 1', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33739675, end_position_on_the_genomic_accession=33782704, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56386, CurrentID=0, Status='live', Symbol='B4galt6', Aliases='AA536803, AU022389', description='UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 6', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 6|UDP-Gal:glucosylceramide beta-1,4-galactosyltransferase|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 6|b4Gal-T6|beta-1, 4-galactosyltransferase VI|beta-1,4-GalTase 6|beta4Gal-T6', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20684599, end_position_on_the_genomic_accession=20746404, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56388, CurrentID=0, Status='live', Symbol='Cyp3a25', Aliases='AI327008', description='cytochrome P450, family 3, subfamily a, polypeptide 25', other_designations='CYPIIIA25|cytochrome P450, 3a25', map_location='5 85.17 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145977194, end_position_on_the_genomic_accession=146009639, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56389, CurrentID=0, Status='live', Symbol='Stx5a', Aliases='0610031F24Rik, D19Ertd627e, Stx5', description='syntaxin 5A', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8741424, end_position_on_the_genomic_accession=8755642, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56390, CurrentID=0, Status='live', Symbol='Sssca1', Aliases='1500016H19Rik, C184L', description="Sjogren's syndrome/scleroderma autoantigen 1 homolog (human)", other_designations='autoantigen p27 homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5730306, end_position_on_the_genomic_accession=5731732, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56392, CurrentID=0, Status='live', Symbol='Shoc2', Aliases='AU017287, Sur8, mKIAA0862', description='soc-2 (suppressor of clear) homolog (C. elegans)', other_designations='Ras-binding protein SUR-8|protein soc-2 homolog|protein sur-8 homolog|soc-2-like protein|suppressor of clear, C. elegans, homolog of', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53944232, end_position_on_the_genomic_accession=54033280, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56395, CurrentID=0, Status='live', Symbol='Tmem115', Aliases='C78915, Pl6, Pp6', description='transmembrane protein 115', other_designations='placental protein 6|protein PL6 homolog', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107533945, end_position_on_the_genomic_accession=107538656, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  56397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56397, CurrentID=0, Status='live', Symbol='Morf4l2', Aliases='2410017O14Rik, Mrgx, Sid393p, mKIAA0026', description='mortality factor 4 like 2', other_designations='MORF-related gene X protein|sid 393|transcription factor-like protein Mrgx', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136732942, end_position_on_the_genomic_accession=136743686, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56398, CurrentID=0, Status='live', Symbol='Chp1', Aliases='1500003O03Rik, AA960066, AI046351, Cahp, Chp, Sid470p, p24, vac', description='calcineurin-like EF hand protein 1', other_designations='EF-hand calcium-binding domain-containing protein p22|calcineurin B-like protein|calcineurin homologous protein|calcium binding protein P22|calcium-binding protein CHP|calcium-binding protein p22|sid 470', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119547707, end_position_on_the_genomic_accession=119587024, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56399, CurrentID=0, Status='live', Symbol='Akap8', Aliases='1200016A02Rik, AA673585, AKAP-8, AKAP95, AU015639', description='A kinase (PRKA) anchor protein 8', other_designations='A-kinase anchor protein 95 kDa', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32303672, end_position_on_the_genomic_accession=32321327, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56401, CurrentID=0, Status='live', Symbol='P3h1', Aliases='2410024C15Rik, Gros1, Lepre1', description='prolyl 3-hydroxylase 1', other_designations='growth suppressor 1|leprecan 1|leprecan-1|leucine- and proline-enriched proteoglycan 1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119232915, end_position_on_the_genomic_accession=119248977, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  56403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56403, CurrentID=0, Status='live', Symbol='Syncrip', Aliases='2610109K23Rik, 4632417O19Rik, GRY-RBP, Nsap1, Nsap1l, pp68', description='synaptotagmin binding, cytoplasmic RNA interacting protein', other_designations='NS1-associated protein 1|RRM RNA binding protein GRY-RBP|glycine- and tyrosine-rich RNA-binding protein|hnRNP Q|hnRNP-Q|synaptotagmin-binding, cytoplasmic RNA-interacting protein', map_location='9 E3.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88449364, end_position_on_the_genomic_accession=88482592, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56404, CurrentID=0, Status='live', Symbol='Trip4', Aliases='4930558E03Rik, ASC-1, Asc1, BB191711', description='thyroid hormone receptor interactor 4', other_designations='TR-interacting protein 4|TRIP-4|thyroid receptor-interacting protein 4|transcription coactivator Asc1', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65828926, end_position_on_the_genomic_accession=65908809, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  56405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56405, CurrentID=0, Status='live', Symbol='Dusp14', Aliases='1110014C10Rik, 2310042C07Rik, D11Ertd395e, MKP-L, Mkp6', description='dual specificity phosphatase 14', other_designations='MAP kinase phosphatase 6|MAP kinase phsophatase 6|MKP-6|Mkp-1 like protein tyrosine phosphatase|mitogen-activated protein kinase phosphatase 6', map_location='11 51.29 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84048033, end_position_on_the_genomic_accession=84069256, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56406, CurrentID=0, Status='live', Symbol='Ncoa6', Aliases='AIB3, ASC-2, ASC2, NRC, Ncoa7, PRIP, RAP250, mKIAA0181', description='nuclear receptor coactivator 6', other_designations='PPAR interacting protein PRIP|PPAR-interacting protein|activating signal cointegrator 2|amplified in breast cancer protein 3|cancer-amplified transcriptional coactivator ASC-2|nuclear receptor coactivator RAP250|nuclear receptor-activating protein 250|nuclear receptor-activating protein, 250 kDa|peroxisome proliferator-activated receptor interacting protein,PRIP|peroxisome proliferator-activated receptor-interacting protein|thyroid hormone receptor binding protein|thyroid hormone receptor-binding protein', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155390656, end_position_on_the_genomic_accession=155473894, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56407, CurrentID=0, Status='live', Symbol='Trpc4ap', Aliases='4833429F06Rik, D2Ertd113e, Trrp4ap, mFLJ00177, truss', description='transient receptor potential cation channel, subfamily C, member 4 associated protein', other_designations='TNF-receptor ubiquitous scaffolding/signaling protein|Trp4-associated protein TAP1|rabex-5/Rin2-interacting protein|transient receptor protein 4, associated protein|trpc4-associated protein', map_location='2 77.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155634277, end_position_on_the_genomic_accession=155692384, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  56409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56409, CurrentID=0, Status='live', Symbol='Nudt3', Aliases='1110011B09Rik, AA960325, Dipp, Dipp1', description='nudix (nucleotide diphosphate linked moiety X)-type motif 3', other_designations="DIPP-1|diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 1", map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27579382, end_position_on_the_genomic_accession=27623452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56410, CurrentID=0, Status='live', Symbol='Cbln3', Aliases='', description='cerebellin 3 precursor protein', other_designations='precerebellin 3', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55878920, end_position_on_the_genomic_accession=55884256, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56412, CurrentID=0, Status='live', Symbol='Noa1', Aliases='2610024G14Rik, AU022345, AtNOS1, mAtNOS1', description='nitric oxide associated 1', other_designations='', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77294169, end_position_on_the_genomic_accession=77310158, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56417, CurrentID=0, Status='live', Symbol='Adar', Aliases='AV2424511, Adar1p110, Adar1p150, mZaADAR, Adar', description='adenosine deaminase, RNA-specific', other_designations='DRADA|RNA adenosine deaminase 1|RNA-specific adenosine deaminase p110 form|RNA-specific adenosine deaminase p150 form', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89715022, end_position_on_the_genomic_accession=89753455, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  56418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56418, CurrentID=0, Status='live', Symbol='Ykt6', Aliases='0610042I15Rik, 1810013M05Rik, AW105923', description='YKT6 v-SNARE homolog (S. cerevisiae)', other_designations='YKT6 homolog|prenylated SNARE protein (Ykt6)', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5955758, end_position_on_the_genomic_accession=5967780, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56419, CurrentID=0, Status='live', Symbol='Diaph3', Aliases='4930417P13Rik, Dia2, Diap3, Drf3, mDia2, mKIAA4117, p134MDia2', description='diaphanous related formin 3', other_designations='diaphanous homolog 3|diaphanous-related formin|diaphanous-related formin-3', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=86655361, end_position_on_the_genomic_accession=87141248, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  56420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56420, CurrentID=0, Status='live', Symbol='Ppp4c', Aliases='1110002D08Rik, AU016079, Ppx', description='protein phosphatase 4, catalytic subunit', other_designations='PP-X|PP4C|pp4|protein phosphatase X', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126785866, end_position_on_the_genomic_accession=126792471, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56421, CurrentID=0, Status='live', Symbol='Pfkp', Aliases='1200015H23Rik, 9330125N24Rik, ATP-PFK, PFK-C, PFK-P', description='phosphofructokinase, platelet', other_designations='6-phosphofructokinase type C|ATP-PFK|ATP:D-fructose 6-phosphate 1-phosphotransferase|phosphofructo-1-kinase isozyme C|phosphofructokinase 1|phosphohexokinase', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=6579874, end_position_on_the_genomic_accession=6648789, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  56422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56422, CurrentID=0, Status='live', Symbol='Hbs1l', Aliases='2810035F15Rik, AI326327, eRFS', description='Hbs1-like (S. cerevisiae)', other_designations='eukaryotic release factor 3', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=21295979, end_position_on_the_genomic_accession=21368889, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  56424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56424, CurrentID=0, Status='live', Symbol='Stub1', Aliases='0610033N24Rik, 2210017D18Rik, 2310040B03Rik, AW046544, Chip', description='STIP1 homology and U-Box containing protein 1', other_designations='E3 ubiquitin-protein ligase CHIP|carboxy terminus of Hsp70-interacting protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25830636, end_position_on_the_genomic_accession=25833361, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56426, CurrentID=0, Status='live', Symbol='Pdcd10', Aliases='2410003B13Rik, Ccm3, Tfa15, Tfar15', description='programmed cell death 10', other_designations='TF-1 cell apoptosis-related protein 15', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=75516490, end_position_on_the_genomic_accession=75556852, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56427, CurrentID=0, Status='live', Symbol='Tubd1', Aliases='4930550G19Rik, Tubd', description='tubulin, delta 1', other_designations='delta-tubulin', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86544991, end_position_on_the_genomic_accession=86567422, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56428, CurrentID=0, Status='live', Symbol='Mtch2', Aliases='2310034D24Rik, 4930539J07Rik, HSPC032', description='mitochondrial carrier homolog 2 (C. elegans)', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90847155, end_position_on_the_genomic_accession=90866634, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56429, CurrentID=0, Status='live', Symbol='Dpt', Aliases='1810032B19Rik, 5033416F05Rik, EQ-1, Eq1', description='dermatopontin', other_designations='TRAMP|early quiescence protein 1|early quiescence protein-1|tyrosine-rich acidic matrix protein', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164796732, end_position_on_the_genomic_accession=164824266, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56430, CurrentID=0, Status='live', Symbol='Clip1', Aliases='1110007I12Rik, 4631429H07Rik, AV017631, C81039, CLIP-170, CLIP170, CYLN1, Clip50, Rsn, mKIAA4046, restin', description='CAP-GLY domain containing linker protein 1', other_designations='Clip 170|Reed-Steinberg cell-espressed intermediate filament-associated protein|cytoplasmic linker protein 50|restin (Reed-Steinberg cell-expressed intermediate filament-associated protein)', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123577794, end_position_on_the_genomic_accession=123684527, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  56431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56431, CurrentID=0, Status='live', Symbol='Dstn', Aliases='2610043P17Rik, ADF, AU042046, Dsn, corn1, sid23p', description='destrin', other_designations='actin-depolymerizing factor|corneal disease 1|sid 23', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=143915331, end_position_on_the_genomic_accession=143943324, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56433, CurrentID=0, Status='live', Symbol='Vps29', Aliases='2010015D08Rik, AW049835, PEP11', description='vacuolar protein sorting 29 (S. pombe)', other_designations='vacuolar sorting protein 29|vesicle protein sorting 29', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122354377, end_position_on_the_genomic_accession=122364981, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56434, CurrentID=0, Status='live', Symbol='Tspan3', Aliases='1700055K04Rik, TM4-A, Tm4sf8', description='tetraspanin 3', other_designations='H41D4 protein|OAP-1|OSP-associated protein 1|tetraspanin TM4-A homolog|transmembrane 4 superfamily member 8', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56135884, end_position_on_the_genomic_accession=56161070, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56436, CurrentID=0, Status='live', Symbol='Adrm1', Aliases='1110063P18Rik, 2510006J17Rik, AA408205, ARM-1, AU043535, Arm1, Gp110, Rpn13', description='adhesion regulating molecule 1', other_designations='110 kDa cell membrane glycoprotein|adhesion-regulating molecule 1|glycoprotein 110|rpn13 homolog', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180171490, end_position_on_the_genomic_accession=180176286, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56437, CurrentID=0, Status='live', Symbol='Rrad', Aliases='REM3, Rad', description='Ras-related associated with diabetes', other_designations='Ras-like GTP-binding protein rad', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104628066, end_position_on_the_genomic_accession=104631367, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56438, CurrentID=0, Status='live', Symbol='Rbx1', Aliases='1500002P15Rik, AA517855, ROC1', description='ring-box 1', other_designations='RING finger protein 75|ring-box protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81466316, end_position_on_the_genomic_accession=81476369, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56440, CurrentID=0, Status='live', Symbol='Snx1', Aliases='', description='sorting nexin 1', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66088127, end_position_on_the_genomic_accession=66124886, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56441, CurrentID=0, Status='live', Symbol='Nat6', Aliases='AI225910, Fus2', description='N-acetyltransferase 6', other_designations='fusion 2|protein fus-2|protein fusion-2', map_location='9 58.18 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107580170, end_position_on_the_genomic_accession=107584048, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  56442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56442, CurrentID=0, Status='live', Symbol='Serinc1', Aliases='1500011D18Rik, AI315070, AIGP-2, TMS-2, Tde1l, Tde2, Tms2, mKIAA1253', description='serine incorporator 1', other_designations='axotomy induced glycoprotein 2|axotomy-induced glyco/Golgi protein 2|membrane protein TMS-2|tumor differentially expressed 1, like|tumor differentially expressed 2|tumor differentially expressed protein 2', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=57515774, end_position_on_the_genomic_accession=57532529, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56443, CurrentID=0, Status='live', Symbol='Arpc1a', Aliases='0610010H08Rik, 1110030K07Rik, 41kDa, AA407347, Sid32, Sid329', description='actin related protein 2/3 complex, subunit 1A', other_designations='SOP2-like protein|sid 329', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145083869, end_position_on_the_genomic_accession=145108756, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56444, CurrentID=0, Status='live', Symbol='Actr10', Aliases='Actr11, Arp11', description='ARP10 actin-related protein 10', other_designations='ARP10 actin-related protein 10 homolog', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70937857, end_position_on_the_genomic_accession=70964718, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56445, CurrentID=0, Status='live', Symbol='Dnaja2', Aliases='1500017M13Rik, 2010206B19Rik, DNAJ, DNJ3, Dnaj3, HIRIP4, PRO3015, mDj3', description='DnaJ heat shock protein family (Hsp40) member A2', other_designations='DNA J protein|DnaJ (Hsp40) homolog, subfamily A, member 2', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85537640, end_position_on_the_genomic_accession=85555271, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56447, CurrentID=0, Status='live', Symbol='Copz1', Aliases='5930435A22Rik, AA407760, D4Ertd360e', description='coatomer protein complex, subunit zeta 1', other_designations='nonclathrin coat protein zeta1-COP|zeta-1 COP|zeta-1-coat protein', map_location='15 58.74 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103272834, end_position_on_the_genomic_accession=103299865, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56448, CurrentID=0, Status='live', Symbol='Cyp2d22', Aliases='2D22', description='cytochrome P450, family 2, subfamily d, polypeptide 22', other_designations='cytochrome P450, 2d22', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82370527, end_position_on_the_genomic_accession=82380260, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56449, CurrentID=0, Status='live', Symbol='Ybx3', Aliases='Csda, Dpba, MSY3, MSY4, Yb2, dbpA, oxyR', description='Y box protein 3', other_designations='DNA-binding protein A|Y-box protein 3|cold shock domain protein A|cold shock domain-containing protein A', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131364858, end_position_on_the_genomic_accession=131388450, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56451, CurrentID=0, Status='live', Symbol='Suclg1', Aliases='1500000I01Rik, Sucla1', description='succinate-CoA ligase, GDP-forming, alpha subunit', other_designations='SCS-alpha|succinyl-CoA ligase [GDP-forming] subunit alpha, mitochondrial|succinyl-CoA synthetase subunit alpha', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=73248505, end_position_on_the_genomic_accession=73276907, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56452, CurrentID=0, Status='live', Symbol='Orc6', Aliases='6720420I10Rikl, Orc6', description='origin recognition complex, subunit 6', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85299632, end_position_on_the_genomic_accession=85308279, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56453, CurrentID=0, Status='live', Symbol='Mbtps1', Aliases='0610038M03Rik, AV003995, S1P, SKI-1, Ski1, mKIAA0091', description='membrane-bound transcription factor peptidase, site 1', other_designations='S1P endopeptidase|endopeptidase S1P|membrane-bound transcription factor protease, site 1|site-1 protease|sterol-regulated luminal protease|subtilisin/kexin isozyme 1|subtilisin/kexin isozyme SKI-1|subtilisin/kexin isozyme-1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119508151, end_position_on_the_genomic_accession=119558825, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  56454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56454, CurrentID=0, Status='live', Symbol='Aldh18a1', Aliases='2810433K04Rik, AI429789, Pycs', description='aldehyde dehydrogenase 18 family, member A1', other_designations='P5CS|aldehyde dehydrogenase family 18 member A1|delta-1-pyrroline-5-carboxylate synthetase|glutamate gamma-semialdehyde synthetase|pyrroline-5-carboxylate synthetase (glutamate gamma-semialdehyde synthetase)', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40550257, end_position_on_the_genomic_accession=40588463, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  56455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56455, CurrentID=0, Status='live', Symbol='Dynll1', Aliases='Dlc8, Dnclc1, Pin', description='dynein light chain LC8-type 1', other_designations='8 kDa dynein light chain|8kD LC|8kDa LC|dynein, cytoplasmic, light chain 1|dynein, cytoplasmic, light peptide|mPIN|protein inhibitor of neuronal nitric oxide synthase|protein inhibitor of nitric oxide synthase', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115297110, end_position_on_the_genomic_accession=115300990, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56456, CurrentID=0, Status='live', Symbol='Actl6a', Aliases='2810432C06Rik, AI851094, ARP4, Actl6, Baf53a, C79802', description='actin-like 6A', other_designations='53 kDa BRG1-associated factor A|BRG1-associated factor 53A|BRG1/brm-associated factor 53A|actin-like 6|actin-related protein Baf53a', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32708546, end_position_on_the_genomic_accession=32726971, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  56457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56457, CurrentID=0, Status='live', Symbol='Clptm1', Aliases='HS9, N14', description='cleft lip and palate associated transmembrane protein 1', other_designations='thymic epithelial cell surface antigen', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19631580, end_position_on_the_genomic_accession=19665030, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56458, CurrentID=0, Status='live', Symbol='Foxo1', Aliases='AI876417, Afxh, FKHR, Fkhr1a, Foxo1', description='forkhead box O1', other_designations='forkhead box O1a|forkhead box protein O1A|forkhead in rhabdomyosarcoma|forkhead protein 1', map_location='3 23.19 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=52268337, end_position_on_the_genomic_accession=52350109, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56459, CurrentID=0, Status='live', Symbol='Sae1', Aliases='2400010M20Rik, 2610044L12Rik, AL033372, AOS1, AW743391, D7Ertd177e, HSPC140, Sua1, Uble1a', description='SUMO1 activating enzyme subunit 1', other_designations='SUMO-1 activating enzyme subunit 1|ubiquitin-like (sentrin) activating enzyme E1A|ubiquitin-like 1 (sentrin) activating enzyme E1A|ubiquitin-like 1 (sentrin) activating enzyme subunit 1|ubiquitin-like 1-activating enzyme E1A', map_location='7 8.88 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16320236, end_position_on_the_genomic_accession=16387896, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56460, CurrentID=0, Status='live', Symbol='Pkp3', Aliases='2310056L12Rik', description='plakophilin 3', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141078229, end_position_on_the_genomic_accession=141090510, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  56461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56461, CurrentID=0, Status='live', Symbol='Kcnip3', Aliases='4933407H12Rik, AI413860, Csen, DREAM, KChIP3, R74849', description='Kv channel interacting protein 3, calsenilin', other_designations='A-type potassium channel modulatory protein 3|DRE-antagonist modulator|calsenilin, presenilin binding protein, EF hand transcription factor|kv channel-interacting protein 3|potassium channel interacting protein 3', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127456498, end_position_on_the_genomic_accession=127522094, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56462, CurrentID=0, Status='live', Symbol='Mtch1', Aliases='2310034O17Rik, AI255158, AU018396, C77849, PSAP', description='mitochondrial carrier homolog 1 (C. elegans)', other_designations='mitochondrial carrier-like protein 1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29332076, end_position_on_the_genomic_accession=29347927, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56463, CurrentID=0, Status='live', Symbol='Snd1', Aliases='AL033314, Tudor-SN', description='staphylococcal nuclease and tudor domain containing 1', other_designations='100 kDa coactivator|p100 co-activator', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=28480315, end_position_on_the_genomic_accession=28935162, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  56464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56464, CurrentID=0, Status='live', Symbol='Ctsf', Aliases='AI481912', description='cathepsin F', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4855129, end_position_on_the_genomic_accession=4860912, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  56468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56468, CurrentID=0, Status='live', Symbol='Socs5', Aliases='1810018L08Rik, Cish5, mKIAA0671', description='suppressor of cytokine signaling 5', other_designations='SOCS-5|cytokine inducible SH2-containing protein 5|cytokine-inducible SH2-containing protein 5', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87107407, end_position_on_the_genomic_accession=87137647, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56469, CurrentID=0, Status='live', Symbol='Pias1', Aliases='2900068C24Rik, Ddxbp1, GBP', description='protein inhibitor of activated STAT 1', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box binding protein 1|DEAD/H box-binding protein 1|protein inhibitor of activated STAT protein 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62878366, end_position_on_the_genomic_accession=62980879, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56470, CurrentID=0, Status='live', Symbol='Rgs19', Aliases='2610042F04Rik, AI324841, AW547781, GAIP', description='regulator of G-protein signaling 19', other_designations='G-alpha interacting protein', map_location='2 103.72 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181688419, end_position_on_the_genomic_accession=181693977, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56471, CurrentID=0, Status='live', Symbol='Stmn4', Aliases='Lagl, R75359, RB3', description='stathmin-like 4', other_designations='stathmin-like protein B3|stathmin-like protein RB3', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66344303, end_position_on_the_genomic_accession=66361680, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56473, CurrentID=0, Status='live', Symbol='Fads2', Aliases='2900042M13Rik, Fadsd2', description='fatty acid desaturase 2', other_designations='D6D|delta(6) desaturase|delta(6) fatty acid desaturase|delta-6 desaturase|delta-6 fatty acid desaturase', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10064164, end_position_on_the_genomic_accession=10101503, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56480, CurrentID=0, Status='live', Symbol='Tbk1', Aliases='1200008B05Rik, AI462036, AW048562', description='TANK-binding kinase 1', other_designations='T2K|serine/threonine protein kinase TBK1', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121546455, end_position_on_the_genomic_accession=121586794, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  56484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56484, CurrentID=0, Status='live', Symbol='Foxo3', Aliases='1110048B16Rik, 2010203A17Rik, C76856, FKHRL1, Fkhr2a, Foxo3', description='forkhead box O3', other_designations='forkhead box O3a|forkhead protein FKHR2', map_location='10 22.79 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=42181845, end_position_on_the_genomic_accession=42276746, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56485, CurrentID=0, Status='live', Symbol='Slc2a5', Aliases='AI526984, Glut5, Slc5a', description='solute carrier family 2 (facilitated glucose transporter), member 5', other_designations='GLUT-5|fructose transporter|glucose transporter type 5, small intestine', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150119344, end_position_on_the_genomic_accession=150144168, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  56486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56486, CurrentID=0, Status='live', Symbol='Gabarap', Aliases='', description='gamma-aminobutyric acid receptor associated protein', other_designations='GABA(A) receptor-associated protein|GABA-A receptor-associated protein|gamma-aminobutyric acid reseptor associated protein', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69991370, end_position_on_the_genomic_accession=69994949, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56488, CurrentID=0, Status='live', Symbol='Nxt1', Aliases='1110001N02Rik', description='NTF2-related export protein 1', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148672615, end_position_on_the_genomic_accession=148676026, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56489, CurrentID=0, Status='live', Symbol='Ikbke', Aliases='AW558201, IKK-i, IKKepsilon, Ikki', description='inhibitor of kappaB kinase epsilon', other_designations='I-kappa-B kinase epsilon|IKK-E|IKK-epsilon|inducible I kappa-B kinase|inducible IKappaB kinase|inhibitor of kappa light polypeptide gene enhancer in B-cells, kinase epsilon', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131254600, end_position_on_the_genomic_accession=131279625, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  56490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56490, CurrentID=0, Status='live', Symbol='Zbtb20', Aliases='1300017A20Rik, 7330412A13Rik, A930017C21Rik, D16Wsu73e, DPZF, HOF, ODA-8S, Oda8, Zfp288', description='zinc finger and BTB domain containing 20', other_designations='BTB/POZ domain zinc finger factor HOF|POZ/zinc finger transcription factor ODA-8|zinc finger protein 288', map_location='16 28.44 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=42875853, end_position_on_the_genomic_accession=43621944, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  56491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56491, CurrentID=0, Status='live', Symbol='Vapb', Aliases='AI225786, AI840687, AI848259, D2Abb2e, R75548, VAP33b, Vamp33b', description='vesicle-associated membrane protein, associated protein B and C', other_designations='VAMP-B|VAMP-associated protein 33b|VAMP-associated protein B|VAP-B', map_location='2 97.37 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173737571, end_position_on_the_genomic_accession=173784336, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56492, CurrentID=0, Status='live', Symbol='Cldn18', Aliases='', description='claudin 18', other_designations='', map_location='9 E3-F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99690796, end_position_on_the_genomic_accession=99717267, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56494, CurrentID=0, Status='live', Symbol='Gosr2', Aliases='2310032N09Rik, C76855, Gs27, SNARE, membrin', description='golgi SNAP receptor complex member 2', other_designations='27 kDa Golgi SNARE protein|golgi SNARE', map_location='11 67.43 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103676849, end_position_on_the_genomic_accession=103697738, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56495, CurrentID=0, Status='live', Symbol='Asna1', Aliases='1810048H22Rik, ArsA', description='arsA arsenite transporter, ATP-binding, homolog 1 (bacterial)', other_designations='arsenical pump-driving ATPase|arsenite-stimulated ATPase|arsenite-translocating ATPase', map_location='8 C3-D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85017931, end_position_on_the_genomic_accession=85025278, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56496, CurrentID=0, Status='live', Symbol='Tspan6', Aliases='6720473L21Rik, AI316786, T245, TSPAN-6, Tm4sf, Tm4sf6', description='tetraspanin 6', other_designations='tetraspan TM4SF|transmembrane 4 superfamily member 6', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=133891070, end_position_on_the_genomic_accession=133898429, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56501, CurrentID=0, Status='live', Symbol='Elf4', Aliases='AV314029, BC042423, ENSMUSG00000053512, Gm9907, Mef', description='E74-like factor 4 (ets domain transcription factor)', other_designations='E74-like factor 4; myeloid elf-1-like factor; myeloid elf-1 like factor|elf4|myeloid elf-1 like factor|myeloid elf-1-like factor', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48411049, end_position_on_the_genomic_accession=48463132, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  56503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56503, CurrentID=0, Status='live', Symbol='Ankrd49', Aliases='AW060934, C86108, Gbif', description='ankyrin repeat domain 49', other_designations='fetal globin inducing factor|fetal globin-increasing factor|fetal globin-inducing factor|globin inducing factor, fetal', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14779782, end_position_on_the_genomic_accession=14784601, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  56504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56504, CurrentID=0, Status='live', Symbol='Srpk3', Aliases='Mssk1, Stk23', description='serine/arginine-rich protein specific kinase 3', other_designations='SR-protein-specific kinase 3|muscle-specific serine kinase 1|serine/arginine-rich protein-specific kinase 3|serine/threonine kinase 23, muscle-specific serine kinase 1 70|serine/threonine-protein kinase 23|serine/threonine-protein kinase SRPK3', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73773261, end_position_on_the_genomic_accession=73778925, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  56505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56505, CurrentID=0, Status='live', Symbol='Ruvbl1', Aliases='2510009G06Rik, Pontin52, Tip49a', description='RuvB-like protein 1', other_designations='49 kDa TATA box-binding protein-interacting protein|49 kDa TBP-interacting protein|DNA helicase p50|Pontin 52', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88465423, end_position_on_the_genomic_accession=88497566, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56506, CurrentID=0, Status='live', Symbol='Cib2', Aliases='2810434I23Rik, AI449053', description='calcium and integrin binding family member 2', other_designations='KIP 2|calcium binding protein Kip 2|calcium binding protein Kip2|kinase interacting protein 2|kinase-interacting protein 2', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54545352, end_position_on_the_genomic_accession=54560079, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56508, CurrentID=0, Status='live', Symbol='Rapgef4', Aliases='1300003D15Rik, 5730402K07Rik, 6330581N18Rik, EPAC 2, Epac-2, Epac2, mKIAA4040', description='Rap guanine nucleotide exchange factor (GEF) 4', other_designations='cAMP-GEFII|cAMP-dependent Rap1 guanine-nucleotide exchange factor|cAMP-regulated guanine nucleotide exchange factor IIB|exchange factor directly activated by cAMP 2|exchange protein directly activated by cAMP 2', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71981215, end_position_on_the_genomic_accession=72257474, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  56513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56513, CurrentID=0, Status='live', Symbol='Pard6a', Aliases='0710008C04Rik, 2610010A15Rik, PAR-6A, PAR6alpha, Par-6, Par6, Par6c, TAX40, Tip-40', description='par-6 family cell polarity regulator alpha', other_designations='PAR-6 alpha|Tax interaction protein 40|par-6 (partitioning defective 6,) homolog alpha', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105701148, end_position_on_the_genomic_accession=105703494, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56515, CurrentID=0, Status='live', Symbol='Rnf138', Aliases='2410015A17Rik, 2810480D20Rik, STRIN, Trif, Trif-d', description='ring finger protein 138', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=21001246, end_position_on_the_genomic_accession=21028224, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56516, CurrentID=0, Status='live', Symbol='Rbms2', Aliases='2610315E04Rik, Scr3', description='RNA binding motif, single stranded interacting protein 2', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128129469, end_position_on_the_genomic_accession=128180297, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  56517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56517, CurrentID=0, Status='live', Symbol='Slc22a21', Aliases='Octn3, Slc22a9', description='solute carrier family 22 (organic cation transporter), member 21', other_designations='organic cation/carnitine transporter 3|solute carrier family 22 (organic cation transporter), member 9|solute carrier family 22 member 9', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53950187, end_position_on_the_genomic_accession=53980818, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56519, CurrentID=0, Status='live', Symbol='Defb4', Aliases='2310001F05Rik, BD-4', description='defensin beta 4', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19198704, end_position_on_the_genomic_accession=19201547, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  56520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56520, CurrentID=0, Status='live', Symbol='Nme4', Aliases='2610027N22Rik, 2810024O08Rik, 5730493H09Rik, NM23-M4, Nm23M4', description='NME/NM23 nucleoside diphosphate kinase 4', other_designations='NDK|NDP kinase, mitochondrial|NDPKD|expressed in non-metastatic cells 4, protein|non-metastatic cells 4, protein expressed in|nucleoside diphosphate kinase D', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26091734, end_position_on_the_genomic_accession=26095527, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  271564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=271564, CurrentID=0, Status='live', Symbol='Vps13a', Aliases='4930425F11, 4930516E05Rik, 4930543C13Rik, 9930023P20, Chac, Chorein, D330038K10Rik, mKIAA0986', description='vacuolar protein sorting 13A (yeast)', other_designations='chorea-acanthocytosis protein homolog|vps13a', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=16612194, end_position_on_the_genomic_accession=16781139, orientation='minus', exon_count=74, OMIM='', no_hdr0=''),
  56522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56522, CurrentID=0, Status='live', Symbol='Papolb', Aliases='PAP, Paplob, Papola-ps, Papt, Plap-ps, Tpap', description='poly (A) polymerase beta (testis specific)', other_designations='PAP-beta|polynucleotide adenylyltransferase beta|testis-specific poly(A) polymerase', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142527740, end_position_on_the_genomic_accession=142530076, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56523, CurrentID=0, Status='live', Symbol='Pmfbp1', Aliases='1700016D22Rik, F77, Stap', description='polyamine modulated factor 1 binding protein 1', other_designations='PMF-1-binding protein|sperm tail associated protein|sperm tail-associated protein', map_location='8 57.05 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109377815, end_position_on_the_genomic_accession=109542642, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  56524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56524, CurrentID=0, Status='live', Symbol='Mpp6', Aliases='P55t, Pals2', description='membrane protein, palmitoylated 6 (MAGUK p55 subfamily member 6)', other_designations='Fibroblast growth factor-inducible protein 15|dlgh4 protein|membrane protein, palmitoylated 3 (MAGUK p55 subfamily member 6)|protein associated with Lin-7 2', map_location='6 24.13 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=50110241, end_position_on_the_genomic_accession=50198598, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  56525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56525, CurrentID=0, Status='live', Symbol='Zfp235', Aliases='0610030O19Rik, AI428448', description='zinc finger protein 235', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24134163, end_position_on_the_genomic_accession=24143244, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56526, CurrentID=0, Status='live', Symbol='Sept6', Aliases='2810035H17Rik, C920001C06Rik, Sep6, mKIAA0128', description='septin 6', other_designations='', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36911272, end_position_on_the_genomic_accession=36989695, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56527, CurrentID=0, Status='live', Symbol='Mast1', Aliases='9430008B02Rik, SAST, SAST170', description='microtubule associated serine/threonine kinase 1', other_designations='mast1|syntrophin associated serine/threonine kinase|syntrophin-associated serine-threonine protein kinase|syntrophin-associated serine/threonine-protein kinase', map_location='8 41.32 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84911853, end_position_on_the_genomic_accession=84946129, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  56529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56529, CurrentID=0, Status='live', Symbol='Sec11a', Aliases='1810012E07Rik, 18kDa, Sec11l1, Sid2895p, Spc18, sid2895', description='SEC11 homolog A, signal peptidase complex subunit', other_designations='SEC11 homolog A|SEC11-like protein 1|SPase 18 kDa subunit|Sec11-like 1|endopeptidase SP18|microsomal signal peptidase 18 kDa subunit|sid 2895|signal peptidase complex (18kD)', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80915379, end_position_on_the_genomic_accession=80947550, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56530, CurrentID=0, Status='live', Symbol='Cnpy2', Aliases='5330432A10Rik, AW229003, D10Bwg1546e, Tmem4, Zsig9', description='canopy FGF signaling regulator 2', other_designations='MIR-interacting saposin-like protein|canopy 2 homolog|putative secreted protein ZSIG9|transmembrane protein 4', map_location='10 76.52 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128322431, end_position_on_the_genomic_accession=128327187, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56531, CurrentID=0, Status='live', Symbol='Ylpm1', Aliases='A930013E17Rik, ZAP, Zap3', description='YLP motif containing 1', other_designations='nuclear protein ZAP|nuclear protein ZAP3', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84996285, end_position_on_the_genomic_accession=85074533, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  56532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56532, CurrentID=0, Status='live', Symbol='Ripk3', Aliases='2610528K09Rik, AW107945, Rip3', description='receptor-interacting serine-threonine kinase 3', other_designations='RIP-3|RIP-like protein kinase 3|mRIP3|receptor interacting protein 3|receptor-interacting protein 3', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55784995, end_position_on_the_genomic_accession=55788857, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56533, CurrentID=0, Status='live', Symbol='Rgs17', Aliases='6430507P11Rik, Rgsz2', description='regulator of G-protein signaling 17', other_designations='regulator of G-protein signaling Z2|regulator of Gz-selective protein signaling 2', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=5825663, end_position_on_the_genomic_accession=5922400, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56534, CurrentID=0, Status='live', Symbol='Hspb3', Aliases='2310035K17Rik, AI844863, Hsbp3, spb3', description='heat shock protein 3', other_designations='heat shock 27kD protein 3|small heat shock protein B3', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113662896, end_position_on_the_genomic_accession=113663676, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56535, CurrentID=0, Status='live', Symbol='Pex3', Aliases='1700014F15Rik, 2810027F19Rik, 2900010N04Rik', description='peroxisomal biogenesis factor 3', other_designations='peroxin-3|peroxisomal assembly protein 3|peroxisomal assembly protein PEX3', map_location='10 A', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13523842, end_position_on_the_genomic_accession=13553149, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56538, CurrentID=0, Status='live', Symbol='Klk11', Aliases='2310015I08Rik, Prss20, TLSP', description='kallikrein related-peptidase 11', other_designations='hippostasin|kallikrein 11|protease, serine, 20|serine protease 20', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43774617, end_position_on_the_genomic_accession=43779263, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56541, CurrentID=0, Status='live', Symbol='Habp4', Aliases='4933413D03Rik, 4933428J01Rik', description='hyaluronic acid binding protein 4', other_designations='IHABP-4|IHABP4|habp4|hyaluronic acid-binding protein 4 (intracellular)|intracellular hyaluronan-binding protein p57', map_location='13 33.26 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64161866, end_position_on_the_genomic_accession=64186537, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56542, CurrentID=0, Status='live', Symbol='Ick', Aliases='2210420N10Rik, AI848300, Mrk', description='intestinal cell kinase', other_designations='MAK-related kinase|mICK', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78109192, end_position_on_the_genomic_accession=78172114, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  56543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56543, CurrentID=0, Status='live', Symbol='Kcnd3', Aliases='AW045978, Kncd3, Kv4.3', description='potassium voltage-gated channel, Shal-related family, member 3', other_designations='potassium channel Kv4.3L|potassium channel Kv4.3M|voltage-gated potassium channel subunit Kv4.3', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105452093, end_position_on_the_genomic_accession=105674002, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56544, CurrentID=0, Status='live', Symbol='Vmn2r1', Aliases='Casrl1, EG56544, V2r83', description='vomeronasal 2, receptor 1', other_designations='V2Rx|calcium-sensing receptor like 1|pheromone receptor V2Rx|putative pheromone receptor V2R1|putative pheromone receptor V2R2|tissue-type vomeronasal neurons putative pheromone receptor V2|vomeronasal receptor Vmn2r1', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=64079774, end_position_on_the_genomic_accession=64220847, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  56546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56546, CurrentID=0, Status='live', Symbol='Sec1', Aliases='Fut10, Fut3', description='secretory blood group 1', other_designations='GDP-L-fucose:beta-D-galactoside 2-alpha-L-fucosyltransferase 3|GDP-L-fucose:beta-D-galactoside 2-alpha-L-fucosyltransferase FUT-III|alpha(1,2)FT 3|fucosyltransferase 10', map_location='7 29.43 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45677686, end_position_on_the_genomic_accession=45694402, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56550, CurrentID=0, Status='live', Symbol='Ube2d2a', Aliases='1500034D03Rik, Ubc2e, Ube2d2, ubc4', description='ubiquitin-conjugating enzyme E2D 2A', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme D2|E2 ubiquitin-conjugating enzyme D2|E2(17)KB 2|ubiquitin carrier protein D2|ubiquitin conjugating enzyme 2e|ubiquitin-conjugating enzyme E2(17)KB 2|ubiquitin-conjugating enzyme E2-17 kDa 2|ubiquitin-conjugating enzyme E2D 2|ubiquitin-protein ligase D2', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35771559, end_position_on_the_genomic_accession=35807172, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56551, CurrentID=0, Status='live', Symbol='Txn2', Aliases='2510006J11Rik, AI788873, Trx2', description='thioredoxin 2', other_designations='MTRX|mt-Trx|thioredoxin nuclear gene encoding mitochondrial protein|thioredoxin-2', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77915051, end_position_on_the_genomic_accession=77928994, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56552, CurrentID=0, Status='live', Symbol='Vmn2r26', Aliases='V2R1, V2r1b', description='vomeronasal 2, receptor 26', other_designations='putative pheromone receptor V2R1b|tissue-type vomeronasal neurons putative pheromone receptor V2R1|vomeronasal 2, receptor, 1b|vomeronasal receptor Vmn2r26|vomeronasal type-2 receptor 1b', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124024758, end_position_on_the_genomic_accession=124062035, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56554, CurrentID=0, Status='live', Symbol='Raet1d', Aliases='RAE-1delta', description='retinoic acid early transcript delta', other_designations='RAE-1-delta|retinoic acid early transcript 1D', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=22361865, end_position_on_the_genomic_accession=22374143, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56612, CurrentID=0, Status='live', Symbol='Pfdn5', Aliases='1190001O17Rik, 1700010A06Rik, D15Ertd697e, Eig1, MM-1', description='prefoldin 5', other_designations='C-myc-binding protein Mm-1|EIG-1|c-myc binding protein MM-1|myc modulator 1', map_location='15 57.48 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102326116, end_position_on_the_genomic_accession=102331490, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56613, CurrentID=0, Status='live', Symbol='Rps6ka4', Aliases='1110069D02Rik, 90kDa, AI848992, Msk2, mMSK2', description='ribosomal protein S6 kinase, polypeptide 4', other_designations='90 kDa ribosomal protein S6 kinase 4|RLSK|RSK-like protein kinase|S6K-alpha-4|mitogen- and stress-activated protein kinase-2|nuclear mitogen- and stress-activated protein kinase 2|ribosomal protein S6 kinase, 90kD, polypeptide 4', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6829080, end_position_on_the_genomic_accession=6841045, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  56615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56615, CurrentID=0, Status='live', Symbol='Mgst1', Aliases='1500002K10Rik, Gst', description='microsomal glutathione S-transferase 1', other_designations='microsomal GST-1|microsomal GST-I', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=138140537, end_position_on_the_genomic_accession=138156755, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56619, CurrentID=0, Status='live', Symbol='Clec4e', Aliases='C86253, Clecsf9, Mincle', description='C-type lectin domain family 4, member e', other_designations='C-type (calcium dependent, carbohydrate recognition domain) lectin, superfamily member 9|C-type lectin superfamily member 9|C-type lectin, superfamily member 9|macrophage-inducible C-type lectin', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=123281789, end_position_on_the_genomic_accession=123289871, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56620, CurrentID=0, Status='live', Symbol='Clec4n', Aliases='Clec6a, Clecsf10, Nkcl', description='C-type lectin domain family 4, member n', other_designations='C-type (calcium dependent, carbohydrate recognition domain) lectin, superfamily member 10|C-type lectin superfamily member 10|DC-associated C-type lectin 2|dectin-2|dendritic cell-associated C-type lectin 2|natural killer complex lectin', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=123229843, end_position_on_the_genomic_accession=123247024, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56622, CurrentID=0, Status='live', Symbol='Adam21', Aliases='Adam31', description='a disintegrin and metallopeptidase domain 21', other_designations='ADAM 21|ADAM 31|a disintegrin and metalloprotease domain (ADAM) 21|a disintegrin and metalloprotease domain 21|disintegrin and metalloproteinase domain-containing protein 31', map_location='12 37.79 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81558584, end_position_on_the_genomic_accession=81568474, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56626, CurrentID=0, Status='live', Symbol='Poll', Aliases='1110003P06Rik, AV007317', description='polymerase (DNA directed), lambda', other_designations='DNA polymerase kappa|pol Lambda', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45552275, end_position_on_the_genomic_accession=45560543, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  56629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56629, CurrentID=0, Status='live', Symbol='Dnase2b', Aliases='AI526873, Dlad, DnaseIIb', description='deoxyribonuclease II beta', other_designations='DNase II beta|DNase II-like acid DNase|DNase2 like acid DNase|DNase2-like acid DNase|deoxyribonuclease DLAD|endonuclease DLAD', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146569733, end_position_on_the_genomic_accession=146618500, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56631, CurrentID=0, Status='live', Symbol='Trim17', Aliases='Rnf16, terf', description='tripartite motif-containing 17', other_designations='ring finger protein 16|tripartite motif protein 17|tripartite motif protein TRIM17|tripartite motif-containing protein 17', map_location='11 B1.2-B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58955310, end_position_on_the_genomic_accession=58971737, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56632, CurrentID=0, Status='live', Symbol='Sphk2', Aliases='C76851', description='sphingosine kinase 2', other_designations='SK 2|SPK 2|Spk2, Sk2', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45709463, end_position_on_the_genomic_accession=45718002, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56635, CurrentID=0, Status='live', Symbol='Prl2a1', Aliases='Ghd21, PLP-M, Prlpm', description='prolactin family 2, subfamily a, member 1', other_designations='PRL-like protein M|growth hormone d21|placental prolactin-like protein M|prolactin-like protein M', map_location='13 12.62 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27801655, end_position_on_the_genomic_accession=27808716, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56636, CurrentID=0, Status='live', Symbol='Fgf21', Aliases='', description='fibroblast growth factor 21', other_designations='FGF-21', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45613890, end_position_on_the_genomic_accession=45619350, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56637, CurrentID=0, Status='live', Symbol='Gsk3b', Aliases='7330414F15Rik, 8430431H08Rik, C86142, GSK-3, GSK-3beta, GSK3', description='glycogen synthase kinase 3 beta', other_designations='GSK-3 beta|serine/threonine-protein kinase GSK3B', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38089001, end_position_on_the_genomic_accession=38246080, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56640, CurrentID=0, Status='live', Symbol='Klk4', Aliases='ESMP1, KLK-L1, PSTS, Prss17', description='kallikrein related-peptidase 4 (prostase, enamel matrix, prostate)', other_designations='kallikrein 4 (prostase, enamel matrix, prostate)|protease, serine 17 (enamel matrix, prostate)', map_location='7 28.35 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43881172, end_position_on_the_genomic_accession=43885804, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56642, CurrentID=0, Status='live', Symbol='Ankrd2', Aliases='Arpp, mArpp', description='ankyrin repeat domain 2 (stretch responsive muscle)', other_designations='skeletal muscle ankyrin repeat protein', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42035656, end_position_on_the_genomic_accession=42045111, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56643, CurrentID=0, Status='live', Symbol='Slc15a1', Aliases='D630032F02, PECT1, PEPT1', description='solute carrier family 15 (oligopeptide transporter), member 1', other_designations='intestinal H(+)/peptide cotransporter|intestinal peptide transporter PEPT1|oligopeptide transporter, small intestine isoform|peptide transporter 1|proton-coupled dipeptide cotransporter', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=121459621, end_position_on_the_genomic_accession=121505254, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  56644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56644, CurrentID=0, Status='live', Symbol='Clec7a', Aliases='BGR, Clecsf12, beta-GR', description='C-type lectin domain family 7, member a', other_designations='C-type (calcium dependent, carbohydrate recognition domain) lectin, superfamily member 12|C-type lectin superfamily member 12|DC-associated C-type lectin 1|beta-glucan receptor|dectin-1|dendritic cell-associated C-type lectin 1', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129461591, end_position_on_the_genomic_accession=129472777, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56690, CurrentID=0, Status='live', Symbol='Mlycd', Aliases='AI324784, Mcd', description='malonyl-CoA decarboxylase', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119393977, end_position_on_the_genomic_accession=119434222, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56691, CurrentID=0, Status='live', Symbol='Dnajb8', Aliases='1700016F14Rik, 1700031F10Rik, AI326313, mDj6', description='DnaJ heat shock protein family (Hsp40) member B8', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 8|DnaJ homolog 6', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88222268, end_position_on_the_genomic_accession=88223256, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  56692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56692, CurrentID=0, Status='live', Symbol='Lamtor3', Aliases='AW556229, Map2k1ip1, Mapbp, Mapksp1, Mp1', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 3', other_designations='MAPK scaffold protein 1|MEK binding partner 1|MEK-binding partner 1|late endosomal/lysosomal adaptor and MAPK and MTOR activator 3|mitogen activated protein, binding protein|mitogen-activated protein kinase kinase 1 interacting protein 1|mitogen-activated protein kinase kinase 1-interacting protein 1|mitogen-activated protein kinase scaffold protein 1', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=137918555, end_position_on_the_genomic_accession=137928762, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  56693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56693, CurrentID=0, Status='live', Symbol='Crtap', Aliases='5730529N23Rik, CASP, Leprel3, P3h5', description='cartilage associated protein', other_designations='', map_location='9 F3-F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114375131, end_position_on_the_genomic_accession=114390712, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56695, CurrentID=0, Status='live', Symbol='Pnkd', Aliases='2210013N15Rik, 2810403H05Rik, AI854243, Brp17, MNCb-5687, MR-1, Tahccp2', description='paroxysmal nonkinesiogenic dyskinesia', other_designations='brain protein 17|myofibrillogenesis regulator 1|paroxysmal nonkinesiogenic dyskinesia protein|trans-activated by hepatitis C virus core protein 2', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74284597, end_position_on_the_genomic_accession=74353692, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56696, CurrentID=0, Status='live', Symbol='Gpr132', Aliases='G2a', description='G protein-coupled receptor 132', other_designations='G protein-coupled receptor G2A|G2 accumulation protein', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112850873, end_position_on_the_genomic_accession=112868418, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56697, CurrentID=0, Status='live', Symbol='Akap10', Aliases='1500031L16Rik, B130049N18Rik, D-AKAP-2, D-AKAP2, PRKA10', description='A kinase (PRKA) anchor protein 10', other_designations='dual specificity A kinase-anchoring protein 2|protein kinase A-anchoring protein 10', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61871307, end_position_on_the_genomic_accession=61930307, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  56698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56698, CurrentID=0, Status='live', Symbol='Phax', Aliases='2810055C14Rik, 4933427L19Rik, AU018701, AU018854, D18Ertd65e, Rnuxa, p55', description='phosphorylated adaptor for RNA export', other_designations='RNA U small nuclear RNA export adapter protein|RNA U, small nuclear RNA export adaptor|phosphorylation regulated', map_location='18 30.63 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=56562569, end_position_on_the_genomic_accession=56587712, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56699, CurrentID=0, Status='live', Symbol='Cdc42ep4', Aliases='1500041M20Rik, Borg4, CEP4', description='CDC42 effector protein (Rho GTPase binding) 4', other_designations='binder of Rho GTPase 4|binder of Rho GTPases 4', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113726850, end_position_on_the_genomic_accession=113752543, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56700, CurrentID=0, Status='live', Symbol='Glmp', Aliases='0610031J06Rik, AB027141, NCU-G1', description='glycosylated lysosomal membrane protein', other_designations='lysosomal protein NCU-G1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88325023, end_position_on_the_genomic_accession=88328631, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56702, CurrentID=0, Status='live', Symbol='Hist1h1b', Aliases='H1.5, H1B, H1f5, H1s-3', description='histone cluster 1, H1b', other_designations='H1 VAR.5|H1 histone family, member 5|histone 1, H1b|histone 1, family 5', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21779832, end_position_on_the_genomic_accession=21780625, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56703, CurrentID=0, Status='live', Symbol='Pigo', Aliases='PIG-O', description='phosphatidylinositol glycan anchor biosynthesis, class O', other_designations='phosphatidylinositol-glycan biosynthesis class O protein', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43017637, end_position_on_the_genomic_accession=43027696, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56705, CurrentID=0, Status='live', Symbol='Ranbp9', Aliases='IBAP-1, Ibap1, RanBPM', description='RAN binding protein 9', other_designations='B cell antigen receptor Ig beta associated protein 1|B-cell antigen receptor Ig beta-associated protein 1|ran-binding protein M', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43402673, end_position_on_the_genomic_accession=43481283, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56706, CurrentID=0, Status='live', Symbol='Ccnl1', Aliases='2610030E23Rik, AU018493, Ccnl, ania-6a', description='cyclin L1', other_designations='cyclin ania-6a|cyclin-L', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=65945912, end_position_on_the_genomic_accession=65958971, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56707, CurrentID=0, Status='live', Symbol='Zfp111', Aliases='', description='zinc finger protein 111', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24193215, end_position_on_the_genomic_accession=24208149, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56708, CurrentID=0, Status='live', Symbol='Clcf1', Aliases='BSF-3, Bsf3, CLC, NNT-1', description='cardiotrophin-like cytokine factor 1', other_designations='B-cell stimulating factor 3|neurotrophin-1/B-cell stimulating factor-3|novel neurotrophin-1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4214392, end_position_on_the_genomic_accession=4222615, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56709, CurrentID=0, Status='live', Symbol='Dnajb12', Aliases='Dj10, mDj10', description='DnaJ heat shock protein family (Hsp40) member B12', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 12|DnaJ homolog 10', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59879569, end_position_on_the_genomic_accession=59899296, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56710, CurrentID=0, Status='live', Symbol='Brinp1', Aliases='BRINP, Dbc1, Dbccr1, Fam5a', description='bone morphogenic protein/retinoic acid inducible neural specific 1', other_designations='BMP/retinoic acid-inducible neural-specific protein (BRINP)|BMP/retinoic acid-inducible neural-specific protein-1|deleted in bladder cancer chromosome region candidate 1|deleted in bladder cancer protein 1 homolog', map_location='4 C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=68761372, end_position_on_the_genomic_accession=68954397, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56711, CurrentID=0, Status='live', Symbol='Plag1', Aliases='', description='pleiomorphic adenoma gene 1', other_designations='C2H2 zinc finger protein PLAG1|pleiomorphic adenoma gene 1 protein', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3898769, end_position_on_the_genomic_accession=3938434, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56715, CurrentID=0, Status='live', Symbol='Rabgef1', Aliases='Rab5ef, Rabex5, Rin2', description='RAB guanine nucleotide exchange factor (GEF) 1', other_designations='Rab5 exchange factor|Ras negative regulator Rabex-5|Ras negative regulator Rabex-5/Rin2|rabex-5', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130171791, end_position_on_the_genomic_accession=130214342, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56716, CurrentID=0, Status='live', Symbol='Mlst8', Aliases='0610033N12Rik, AA409454, AI505104, AI851821, Gbl', description='MTOR associated protein, LST8 homolog (S. cerevisiae)', other_designations='TORC subunit LST8|mammalian lethal with SEC13 protein 8|protein GbetaL|transducin (beta)-like 4', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24473550, end_position_on_the_genomic_accession=24479114, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56717, CurrentID=0, Status='live', Symbol='Mtor', Aliases='2610315D21Rik, AI327068, FRAP, FRAP2, Frap1, RAFT1, RAPT1, flat', description='mechanistic target of rapamycin (serine/threonine kinase)', other_designations='FK506 binding protein 12-rapamycin associated protein 1|FK506-binding protein 12-rapamycin complex-associated protein 1|FKBP-rapamycin associated protein (FRAP)|FKBP-rapamycin-associated protein FRAP|FKBP12-rapamycin complex-associated protein|angiopoietin-like factor CDT6|mammalian target of rapamycin|rapamycin target protein 1', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148448582, end_position_on_the_genomic_accession=148557685, orientation='plus', exon_count=58, OMIM='', no_hdr0=''),
  56720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56720, CurrentID=0, Status='live', Symbol='Tdo2', Aliases='AA407491, TDO, TO, chky', description='tryptophan 2,3-dioxygenase', other_designations='TO|TRPO|tdo variant1|tdo variant2|tryptamin 2,3-dioxygenase|tryptophan oxygenase|tryptophan pyrrolase|tryptophan-2,3-dioxygenase|tryptophanase', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=81958412, end_position_on_the_genomic_accession=81975728, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56722, CurrentID=0, Status='live', Symbol='Litaf', Aliases='3222402J11Rik, C85531, N4WBP3, TBX1', description='LPS-induced TN factor', other_designations='LITAF-like protein|LPS-induced TNF-alpha factor homolog|Nedd4 WW domain-binding protein 3|estrogen-enhanced transcript protein|mEET', map_location='16 B1-B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10959273, end_position_on_the_genomic_accession=10993121, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56724, CurrentID=0, Status='live', Symbol='Cript', Aliases='1200020A08Rik, AA589469', description='cysteine-rich PDZ-binding protein', other_designations='cysteine-rich interactor of PDZ three|cysteine-rich interactor of PDZ3|postsynaptic protein Cript', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87025561, end_position_on_the_genomic_accession=87035808, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  56726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56726, CurrentID=0, Status='live', Symbol='Sh3bgrl', Aliases='1190008F14Rik', description='SH3-binding domain glutamic acid-rich protein like', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=109095407, end_position_on_the_genomic_accession=109162467, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56727, CurrentID=0, Status='live', Symbol='Miox', Aliases='0610009I10Rik, AI314022, Aldrl6, C85427, RSOR', description='myo-inositol oxygenase', other_designations='MI oxygenase|aldehyde reductase (aldose reductase)-like 6|aldehyde reductase 6, renal|aldehyde reductase-like 6|renal-specific oxido-reducatse|renal-specific oxidoreductase', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89334473, end_position_on_the_genomic_accession=89337007, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56734, CurrentID=0, Status='live', Symbol='Tulp2', Aliases='Pdet', description='tubby-like protein 2', other_designations='protein P4-6|testis specific phosphodiesterase', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45482026, end_position_on_the_genomic_accession=45523712, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  56735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56735, CurrentID=0, Status='live', Symbol='Krt71', Aliases='AA589543, Ca, Cal4, Cu, Krt2-6g, mK6irs, mK6irs1', description='keratin 71', other_designations='CK-6G|CK-71|K6G|K71|caracul|cytokeratin-6G|cytokeratin-71|keratin complex 2, basic, gene 6g|keratin complex 2, gene 6g|keratin-6G|keratin-71|mK6irs1/Krt2-6g|type II inner root sheath-specific keratin-K6irs1|type-II keratin Kb34', map_location='15 57.0 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101733949, end_position_on_the_genomic_accession=101743097, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56736, CurrentID=0, Status='live', Symbol='Rnf14', Aliases='2310075C09Rik, 2610005D23Rik, AA986456, AU041447, D18Ertd188e, D7Bwg0165e, Triad2', description='ring finger protein 14', other_designations='TRIAD2 type I|androgen receptor-associated protein 54|protein Triad2', map_location='18 20.2 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38296635, end_position_on_the_genomic_accession=38317852, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  56737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56737, CurrentID=0, Status='live', Symbol='Alg2', Aliases='1110018A23Rik, 1300013N08Rik, ALPG2, CDGIi, MNCb-5081', description='asparagine-linked glycosylation 2 (alpha-1,3-mannosyltransferase)', other_designations='GDP-Man:Man(1)GlcNAc(2)-PP-Dol alpha-1,3-mannosyltransferase|GDP-Man:Man(1)GlcNAc(2)-PP-dolichol mannosyltransferase|GDP-Man:Man(2)GlcNAc(2)-PP-Dol alpha-1,6-mannosyltransferase|alpha-1,3-mannosyltransferase ALG2|asparagine-linked glycosylation 2 homolog (yeast, alpha-1,3-mannosyltransferase)|asparagine-linked glycosylation protein 2 homolog', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=47469833, end_position_on_the_genomic_accession=47474374, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56738, CurrentID=0, Status='live', Symbol='Mocs1', Aliases='3110045D15Rik, AI043170, AW536397', description='molybdenum cofactor synthesis 1', other_designations='molybdenum cofactor synthesis-step 1 proteins A and B', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=49428364, end_position_on_the_genomic_accession=49455435, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  56739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56739, CurrentID=0, Status='live', Symbol='Rec8', Aliases='Rec8L1, mrec', description='REC8 meiotic recombination protein', other_designations='REC8 homolog|REC8-like 1|cohesin Rec8p|meiotic cohesion Rec8', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55618166, end_position_on_the_genomic_accession=55625395, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  56741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56741, CurrentID=0, Status='live', Symbol='Igdcc4', Aliases='DDM36, Nope', description='immunoglobulin superfamily, DCC subclass, member 4', other_designations='amplimer WI-16786, mouse homolog|amplimer WI-18508, mouse homolog|neighbor of Punc e11 protein', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65101461, end_position_on_the_genomic_accession=65137947, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  56742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56742, CurrentID=0, Status='live', Symbol='Psrc1', Aliases='5430413I02Rik, AL033312, Dda3', description='proline/serine-rich coiled-coil 1', other_designations='differential display and activated by p53|p53-regulated DDA3', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108383804, end_position_on_the_genomic_accession=108388231, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56743, CurrentID=0, Status='live', Symbol='Lat2', Aliases='AW125574, LAB, NTAL, WSCR5, Wbscr15, Wbscr5', description='linker for activation of T cells family, member 2', other_designations='Williams-Beuren syndrome chromosome region 5 homolog|linker for activation of B-cells|membrane-associated adapter molecule|non-T cell activation linker|williams-Beuren syndrome chromosomal region 15 protein homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134600103, end_position_on_the_genomic_accession=134619783, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  56744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56744, CurrentID=0, Status='live', Symbol='Pf4', Aliases='Cxcl4, Scyb4', description='platelet factor 4', other_designations='C-X-C motif chemokine 4|PF-4|chemokine (C-X-C motif) ligand 4', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90772435, end_position_on_the_genomic_accession=90773383, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56745, CurrentID=0, Status='live', Symbol='C1qtnf1', Aliases='1600017K21Rik, CTRP1, Zsig37', description='C1q and tumor necrosis factor related protein 1', other_designations='putative secreted protein ZSIG37', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118428457, end_position_on_the_genomic_accession=118451782, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56746, CurrentID=0, Status='live', Symbol='Tex101', Aliases='1700008H15Rik, AI429076, TES101RP', description='testis expressed gene 101', other_designations='TES101-reactive protein|testis-expressed protein 101', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24668012, end_position_on_the_genomic_accession=24672050, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56747, CurrentID=0, Status='live', Symbol='Sez6l', Aliases='AI843918, AIG1, AW539724, Acig1, BSRP-B, mKIAA0927', description='seizure related 6 homolog like', other_designations='acupuncture induced gene 1|acupuncture-induced protein 1-L|brain-specific receptor-like protein B', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112419151, end_position_on_the_genomic_accession=112578042, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  56748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56748, CurrentID=0, Status='live', Symbol='Nfu1', Aliases='0610006G17Rik, CGI-33, Hirip5', description='NFU1 iron-sulfur cluster scaffold homolog (S. cerevisiae)', other_designations='HIRA-interacting protein 5|histone cell cycle regulation defective interacting protein 5|mHIRIP5', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87009291, end_position_on_the_genomic_accession=87028461, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56749, CurrentID=0, Status='live', Symbol='Dhodh', Aliases='2810417D19Rik, AI834883', description='dihydroorotate dehydrogenase', other_designations='DHOdehase|dihydroorotate dehydrogenase, mitochondrial|dihydroorotate oxidase', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109593244, end_position_on_the_genomic_accession=109608701, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  56752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56752, CurrentID=0, Status='live', Symbol='Aldh9a1', Aliases='AA139417, Abaldh, ESTM40, TMABA-DH, Tmabadh', description='aldehyde dehydrogenase 9, subfamily A1', other_designations='aldehyde dehydrogenase family 9 member A1|gamma-aminobutyraldehyde dehydrogenase', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167349991, end_position_on_the_genomic_accession=167368530, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56753, CurrentID=0, Status='live', Symbol='Tacstd2', Aliases='C80403, EGP-1, GA733-1, Ly97, TROP2', description='tumor-associated calcium signal transducer 2', other_designations='cell surface glycoprotein Trop-2|lymphocyte antigen 97|tunor-associated calcium signal transducer 2', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67534059, end_position_on_the_genomic_accession=67535822, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56758, CurrentID=0, Status='live', Symbol='Mbnl1', Aliases='Mbnl, mKIAA0428', description='muscleblind-like 1 (Drosophila)', other_designations='triplet-expansion RNA-binding protein', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=60472791, end_position_on_the_genomic_accession=60629750, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  56760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56760, CurrentID=0, Status='live', Symbol='Clec1b', Aliases='1810061I13Rik, Clec-2, Clec2', description='C-type lectin domain family 1, member b', other_designations='C-type lectin-like receptor 2', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129397297, end_position_on_the_genomic_accession=129405413, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56771, CurrentID=0, Status='live', Symbol='Med20', Aliases='1110011O05Rik, 2410115I17Rik, AU018348, Trfp', description='mediator complex subunit 20', other_designations='TRF-proximal protein homolog|Trf (TATA binding protein-related factor)-proximal protein homolog|ubiquitin specific protease 49', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47611596, end_position_on_the_genomic_accession=47624418, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56772, CurrentID=0, Status='live', Symbol='Mllt11', Aliases='AI839562, Af1q, Zfp692', description='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 11', other_designations='ALL1-fused gene from chromosome 1q|zinc finger protein 692', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95218544, end_position_on_the_genomic_accession=95230459, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56773, CurrentID=0, Status='live', Symbol='Chst5', Aliases='AI173964, GST-4, Gn6st-3, I-GlcNAc-6-ST', description='carbohydrate (N-acetylglucosamine 6-O) sulfotransferase 5', other_designations='GST4|I-GlcNAc6ST|N-acetylglucosamine 6-O-sulfotransferase 3|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 4|glcNAc6ST-3|intestinal GlcNAc-6-sulfotransferase|intestinal N-acetylglucosamine-6-O-sulfotransferase|mIGn6ST', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111889135, end_position_on_the_genomic_accession=111910199, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56774, CurrentID=0, Status='live', Symbol='Slc6a14', Aliases='1110007A17Rik, 9030613J17Rik, ATB0plus, CATB0plus', description='solute carrier family 6 (neurotransmitter transporter), member 14', other_designations='amino acid transporter ATB0+|colonic system B0+ amino acid transporter CATB0+|solute carrier family 6 member 14', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=21714900, end_position_on_the_genomic_accession=21742358, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  56784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56784, CurrentID=0, Status='live', Symbol='Ralgapa1', Aliases='2310003F20Rik, 4930400K19Rik, AI563624, GRIPE, Garnl1, Tulip1, mKIAA0884', description='Ral GTPase activating protein, alpha subunit 1', other_designations='GAP-related interacting partner to E12|GAP-related-interacting partner to E12|GTPase activating RANGAP domain-like 1|GTPase-activating RapGAP domain-like 1|p240|ral GTPase-activating protein alpha subunit 1|tuberin-like protein 1|tulip 1 protein', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55602890, end_position_on_the_genomic_accession=55821575, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  56786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56786, CurrentID=0, Status='live', Symbol='Tmem9b', Aliases='2310004K06Rik, AW539847, D7H11orf15, ICRFP703B1614Q5.3', description='TMEM9 domain family, member B', other_designations='ICRFP703N2430Q5.3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109735834, end_position_on_the_genomic_accession=109752870, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  56787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56787, CurrentID=0, Status='live', Symbol='Ascl3', Aliases='Sgn1, bHLHa42', description='achaete-scute family bHLH transcription factor 3', other_designations='ASH-3|ICRFP703B1614Q5.4|ICRFP703N2430Q5.4|achaete-scute complex homolog 3|achaete-scute complex homolog-like 3|bHLH transcriptional regulator Sgn-1|mASH-3|mASH3', map_location='7 57.46 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109727465, end_position_on_the_genomic_accession=109731732, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56788, CurrentID=0, Status='live', Symbol='Scube2', Aliases='4932442O19Rik, Cegf1, Cegp1', description='signal peptide, CUB domain, EGF-like 2', other_designations='CUB domain and EGF-like repeat containing 1', map_location='7 57.53 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109798690, end_position_on_the_genomic_accession=109865679, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  56790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56790, CurrentID=0, Status='live', Symbol='Supt20', Aliases='AA667204, AI450544, D3Ertd300e, Fam48ah, p38IP, Supt20', description='suppressor of Ty 20', other_designations='family with sequence similarity 48, member A|p38-interacting protein|protein FAM48A|transcription factor (p38 interacting protein)', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54692761, end_position_on_the_genomic_accession=54728763, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  56791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56791, CurrentID=0, Status='live', Symbol='Ube2l6', Aliases='2810489I21Rik, RIG-B, UBCH8, Ubce8, Ubcm8', description='ubiquitin-conjugating enzyme E2L 6', other_designations='E2 ubiquitin conjugating enzyme|E2 ubiquitin-conjugating enzyme L6|ubiquitin carrier protein L6|ubiquitin-conjugating enzyme 8|ubiquitin-protein ligase L6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84798828, end_position_on_the_genomic_accession=84810003, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56792, CurrentID=0, Status='live', Symbol='Stap1', Aliases='AI586015, Brdg1, STAP-1', description='signal transducing adaptor family member 1', other_designations='BCR downstream signaling 1|stem cell adaptor protein STAP-1', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86071585, end_position_on_the_genomic_accession=86104784, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  56794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56794, CurrentID=0, Status='live', Symbol='Hacl1', Aliases='1600020H07Rik, Hpcl, Phyh2', description='2-hydroxyacyl-CoA lyase 1', other_designations='2-HPCL|2-hydroxyphytanoyl-CoA lyase|phytanoyl-CoA 2-hydroxylase 2', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31607226, end_position_on_the_genomic_accession=31641286, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  56795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56795, CurrentID=0, Status='live', Symbol='Arl10', Aliases='Arl10a, Arm1', description='ADP-ribosylation factor-like 10', other_designations='ADP-ribosylation factor-like 10A|ADP-ribosylation factor-like membrane-associated protein', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54575000, end_position_on_the_genomic_accession=54582765, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  56802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56802, CurrentID=0, Status='live', Symbol='Nespas', Aliases='Gnas-as, Nespos', description='neuroendocrine secretory protein antisense', other_designations='neuroendocrine secretory protein opposite strand', map_location='2 97.89 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174281237, end_position_on_the_genomic_accession=174295436, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56805, CurrentID=0, Status='live', Symbol='Zbtb33', Aliases='AW260253, E130014G12Rik, Kaiso, Znf-kaiso', description='zinc finger and BTB domain containing 33', other_designations='zinc finger and BTB domain-containing protein 33|zinc finger transcription factor Kaiso', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38189793, end_position_on_the_genomic_accession=38197046, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  56807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56807, CurrentID=0, Status='live', Symbol='Scamp5', Aliases='Sc5', description='secretory carrier membrane protein 5', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57441327, end_position_on_the_genomic_accession=57468099, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  56808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56808, CurrentID=0, Status='live', Symbol='Cacna2d2', Aliases='Cacna2d, a2d2, du, mKIAA0558, td, torpid', description='calcium channel, voltage-dependent, alpha 2/delta subunit 2', other_designations='alpha 2 delta calcium channel subunit|ducky|protein ducky|voltage-gated calcium channel subunit alpha-2/delta-2', map_location='9 58.02 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107399880, end_position_on_the_genomic_accession=107529346, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  56809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56809, CurrentID=0, Status='live', Symbol='Gmeb1', Aliases='1110050A04Rik, AI256615, AI481278', description='glucocorticoid modulatory element binding protein 1', other_designations='GMEB-1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132220376, end_position_on_the_genomic_accession=132261549, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  56811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56811, CurrentID=0, Status='live', Symbol='Dkk2', Aliases='', description='dickkopf WNT signaling pathway inhibitor 2', other_designations='dickkopf 2|dickkopf homolog 1|dickkopf homolog 2|dickkopf-2|dkk-2|mDkk-2', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=132085242, end_position_on_the_genomic_accession=132180304, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  56812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56812, CurrentID=0, Status='live', Symbol='Dnajb2', Aliases='2700059H22Rik, Dnajb10, Hsj1, mDj8', description='DnaJ heat shock protein family (Hsp40) member B2', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 10|DnaJ (Hsp40) homolog, subfamily B, member 2|dnaJ homolog subfamily B member 10|heat shock protein, DNAJ-like 1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75236403, end_position_on_the_genomic_accession=75245692, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  56835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56835, CurrentID=0, Status='live', Symbol='Ctsr', Aliases='AI428350', description='cathepsin R', other_designations='lysosomal cysteine protease', map_location='13 32.67 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61159215, end_position_on_the_genomic_accession=61164188, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  56838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56838, CurrentID=0, Status='live', Symbol='Ccl28', Aliases='CCK1, MEC, Scya28', description='chemokine (C-C motif) ligand 28', other_designations='small inducible cytokine A28|small-inducible cytokine A28', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119610551, end_position_on_the_genomic_accession=119654359, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56839, CurrentID=0, Status='live', Symbol='Lgi1', Aliases='BB130740', description='leucine-rich repeat LGI family, member 1', other_designations='leucine-rich, glioma inactivated 1', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38264782, end_position_on_the_genomic_accession=38308939, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  56843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56843, CurrentID=0, Status='live', Symbol='Trpm5', Aliases='9430099A16Rik, LTrpC-5, Ltrpc5, Mtr1', description='transient receptor potential cation channel, subfamily M, member 5', other_designations='MLSN1- and TRP-related gene 1 protein|long transient receptor potential channel 5|long transient receptor potential-related channel 5|trpm5', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143071529, end_position_on_the_genomic_accession=143094642, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  56844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56844, CurrentID=0, Status='live', Symbol='Tssc4', Aliases='AA241958, ESTM671070', description='tumor-suppressing subchromosomal transferable fragment 4', other_designations='', map_location='7 88.11 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143068035, end_position_on_the_genomic_accession=143071093, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  56846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56846, CurrentID=0, Status='live', Symbol='Necab3', Aliases='2900010M17Rik, AI853434, Apba2bp, Nip1, XB51', description='N-terminal EF-hand calcium binding protein 3', other_designations='X11L-binding protein 51|amyloid beta (A4) precursor protein-binding family A member 1 binding protein|amyloid beta (A4) precursor protein-binding, family A, member 1 binding protein|amyloid beta (A4) precursor protein-binding, family A, member 2 binding protein|amyloid beta A4 protein-binding family A member 2-binding protein|mXB51', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154544399, end_position_on_the_genomic_accession=154559406, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56847, CurrentID=0, Status='live', Symbol='Aldh1a3', Aliases='ALDH6, RALDH3, V1', description='aldehyde dehydrogenase family 1, subfamily A3', other_designations='RALDH-3|aldehyde dehydrogenase 6|aldehyde dehydrogenase-6|retinaldehyde dehydrogenase 3', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66390892, end_position_on_the_genomic_accession=66427477, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  56856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56856, CurrentID=0, Status='live', Symbol='Insm2', Aliases='Mlt1', description='insulinoma-associated 2', other_designations='methylated in liver tumor 1|mlt 1', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55598917, end_position_on_the_genomic_accession=55602018, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  56857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56857, CurrentID=0, Status='live', Symbol='Slc37a2', Aliases='G3PP, Slc37a1, cI-2, ci2', description='solute carrier family 37 (glycerol-3-phosphate transporter), member 2', other_designations='cAMP inducible 2 protein|cAMP inducible gene 2|cAMP-inducible protein 2|solute carrier family 37 (glycerol-3-phosphate transporter), member 1|solute carrier family 37 member 2', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37229149, end_position_on_the_genomic_accession=37255738, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  56858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56858, CurrentID=0, Status='live', Symbol='Olfr749', Aliases='MOR106-1, ORL461, Ors1', description='olfactory receptor 749', other_designations='GA_x6K02T2PMLR-6484046-6483105|odorant receptor S1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50736219, end_position_on_the_genomic_accession=50744324, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  56860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56860, CurrentID=0, Status='live', Symbol='Olfr690', Aliases='MOR31-2, Ors18', description='olfactory receptor 690', other_designations='GA_x6K02T2PBJ9-7959171-7958224|odorant receptor S18|olfactory receptor MOR31-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105329217, end_position_on_the_genomic_accession=105330278, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56861, CurrentID=0, Status='live', Symbol='Olfr480', Aliases='MOR204-32, Ors25, S25', description='olfactory receptor 480', other_designations='GA_x6K02T2PBJ9-10395807-10394869|odorant receptor S25|olfactory receptor 204-32|olfactory receptor MOR204-32', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108065768, end_position_on_the_genomic_accession=108066796, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56863, CurrentID=0, Status='live', Symbol='Cldn9', Aliases='nmf329', description='claudin 9', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23682584, end_position_on_the_genomic_accession=23684026, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  56868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56868, CurrentID=0, Status='live', Symbol='Psg23', Aliases='1620401C02Rik, cea11', description='pregnancy-specific glycoprotein 23', other_designations='', map_location='7 9.18 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18606343, end_position_on_the_genomic_accession=18616501, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  56872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56872, CurrentID=0, Status='live', Symbol='Pate4', Aliases='9530004K16Rik, Pate-B, Svs7', description='prostate and testis expressed 4', other_designations='PATE-like protein B|SVS VII|calcium transport inhibitor|caltrin|seminal vesicle protein 7|seminal vesicle protein VII|seminal vesicle protein, secretion 7|seminal vesicle secretory protein 7', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35607093, end_position_on_the_genomic_accession=35616371, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  56873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56873, CurrentID=0, Status='live', Symbol='Lmbr1', Aliases='1110048D14Rik, AU017641', description='limb region 1', other_designations='', map_location='5 14.81 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=29229802, end_position_on_the_genomic_accession=29378397, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  56874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56874, CurrentID=0, Status='live', Symbol='Rnf32', Aliases='1700009J01Rik, 2700025B22Rik, 4930542N22Rik, Lmbr2', description='ring finger protein 32', other_designations='limb region protein 2', map_location='5 14.78 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=29195170, end_position_on_the_genomic_accession=29225526, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  56876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56876, CurrentID=0, Status='live', Symbol='Nsmf', Aliases='Jacob, Nelf', description='NMDA receptor synaptonuclear signaling and neuronal migration factor', other_designations='juxtasynaptic attractor of caldendrin on dendritic boutons protein|nasal embryonic LHRH factor|nasal embryonic luteinizing hormone-releasing hormone factor', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25054370, end_position_on_the_genomic_accession=25062881, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  56878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=56878, CurrentID=0, Status='live', Symbol='Rbms1', Aliases='2600014B10Rik, AI255215, MSSP-1, MSSP-2, MSSP-3, YC1', description='RNA binding motif, single stranded interacting protein 1', other_designations='single-stranded DNA-binding protein MSSP-1', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=60750193, end_position_on_the_genomic_accession=60963450, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  319146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319146, CurrentID=0, Status='live', Symbol='Ifnz', Aliases='6030405N23Rik, BGIF, ENSMUSG00000073810, Lmtn', description='interferon zeta', other_designations='limitin', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88782131, end_position_on_the_genomic_accession=88783592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  450219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=450219, CurrentID=0, Status='live', Symbol='Gsdma3', Aliases='Bsk, Dfl, Fgn, Gsdm1l, Gsdm3, Rco2, Rim3', description='gasdermin A3', other_designations='bare skin|defolliculated|gasdermin 3|gasdermin-3|gsdma3|recombination induced mutation 3', map_location='11 62.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98626360, end_position_on_the_genomic_accession=98638308, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  319148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319148, CurrentID=0, Status='live', Symbol='Hist1h3c', Aliases='H3-143', description='histone cluster 1, H3c', other_designations='histone1, H3c', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23745042, end_position_on_the_genomic_accession=23745521, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319149, CurrentID=0, Status='live', Symbol='Hist1h3d', Aliases='H3-b', description='histone cluster 1, H3d', other_designations='histone1, H3d', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23575763, end_position_on_the_genomic_accession=23576266, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319150, CurrentID=0, Status='live', Symbol='Hist1h3b', Aliases='H3-53', description='histone cluster 1, H3b', other_designations='histone 1, H3b', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23752380, end_position_on_the_genomic_accession=23752840, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319151, CurrentID=0, Status='live', Symbol='Hist1h3e', Aliases='H3-f, H3.1', description='histone cluster 1, H3e', other_designations='histone 1, H3e', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23561896, end_position_on_the_genomic_accession=23562365, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319152, CurrentID=0, Status='live', Symbol='Hist1h3h', Aliases='H3-291', description='histone cluster 1, H3h', other_designations='histone 1, H3h', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21717628, end_position_on_the_genomic_accession=21718115, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319153, CurrentID=0, Status='live', Symbol='Hist1h3i', Aliases='H3-H', description='histone cluster 1, H3i', other_designations='histone 1, H3i', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21782915, end_position_on_the_genomic_accession=21783397, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319154, CurrentID=0, Status='live', Symbol='Hist2h3b', Aliases='H3-615, H3-616', description='histone cluster 2, H3b', other_designations='histone protein Hist2h2bb', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96268654, end_position_on_the_genomic_accession=96269155, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319155, CurrentID=0, Status='live', Symbol='Hist1h4c', Aliases='', description='histone cluster 1, H4c', other_designations='germinal histone H4|histone 1, H4c', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23698084, end_position_on_the_genomic_accession=23698458, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319156, CurrentID=0, Status='live', Symbol='Hist1h4d', Aliases='', description='histone cluster 1, H4d', other_designations='histone 1, H4d', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23581602, end_position_on_the_genomic_accession=23581969, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319157, CurrentID=0, Status='live', Symbol='Hist1h4f', Aliases='', description='histone cluster 1, H4f', other_designations='histone 1, H4f', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23551286, end_position_on_the_genomic_accession=23551643, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319158, CurrentID=0, Status='live', Symbol='Hist1h4i', Aliases='', description='histone cluster 1, H4i', other_designations='histone 1, H4i', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22040960, end_position_on_the_genomic_accession=22041362, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319159, CurrentID=0, Status='live', Symbol='Hist1h4j', Aliases='', description='histone cluster 1, H4j', other_designations='histone 1, H4j', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21735096, end_position_on_the_genomic_accession=21735407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319160, CurrentID=0, Status='live', Symbol='Hist1h4k', Aliases='', description='histone cluster 1, H4k', other_designations='histone 1, H4k', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21750145, end_position_on_the_genomic_accession=21750553, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319162, CurrentID=0, Status='live', Symbol='Hist3h2a', Aliases='', description='histone cluster 3, H2a', other_designations='histone 3, H2a', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58954685, end_position_on_the_genomic_accession=58955192, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319163, CurrentID=0, Status='live', Symbol='Hist1h2aa', Aliases='Th2a', description='histone cluster 1, H2aa', other_designations='histone 1, H2aa', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23934462, end_position_on_the_genomic_accession=23934913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319164, CurrentID=0, Status='live', Symbol='Hist1h2ac', Aliases='', description='histone cluster 1, H2ac', other_designations='histone 1, H2ac', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23683473, end_position_on_the_genomic_accession=23683959, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319165, CurrentID=0, Status='live', Symbol='Hist1h2ad', Aliases='', description='histone cluster 1, H2ad', other_designations='histone 1, H2ad', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23574381, end_position_on_the_genomic_accession=23574915, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319166, CurrentID=0, Status='live', Symbol='Hist1h2ae', Aliases='1190022L06', description='histone cluster 1, H2ae', other_designations='histone 1, H2ae', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23570663, end_position_on_the_genomic_accession=23571220, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319168, CurrentID=0, Status='live', Symbol='Hist1h2ah', Aliases='', description='histone cluster 1, H2ah', other_designations='histone 1, H2ah', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22035122, end_position_on_the_genomic_accession=22035643, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319169, CurrentID=0, Status='live', Symbol='Hist1h2ak', Aliases='', description='histone cluster 1, H2ak', other_designations='histone 1, H2ak', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21753377, end_position_on_the_genomic_accession=21753912, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319171, CurrentID=0, Status='live', Symbol='Hist1h2ap', Aliases='ENSMUSG00000060262, Gm11276, OTTMUSG00000000448', description='histone cluster 1, H2ap', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21833024, end_position_on_the_genomic_accession=21833475, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319172, CurrentID=0, Status='live', Symbol='Hist1h2ab', Aliases='H2A.1, H2a-53', description='histone cluster 1, H2ab', other_designations='histone 1, H2ab', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23751088, end_position_on_the_genomic_accession=23751592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319173, CurrentID=0, Status='live', Symbol='Hist1h2af', Aliases='H2a-221', description='histone cluster 1, H2af', other_designations='histone 1, H2af|histone H2a (aa 92-128)', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23533911, end_position_on_the_genomic_accession=23534378, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319176, CurrentID=0, Status='live', Symbol='Hist2h2ac', Aliases='H2a-613b', description='histone cluster 2, H2ac', other_designations='histone 2, H2ac', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96220413, end_position_on_the_genomic_accession=96220880, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319177, CurrentID=0, Status='live', Symbol='Hist1h2ba', Aliases='Th2b', description='histone cluster 1, H2ba', other_designations='histone 1, H2ba|histone H2B, testis|testis-specific histone H2B', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23933725, end_position_on_the_genomic_accession=23934156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319178, CurrentID=0, Status='live', Symbol='Hist1h2bb', Aliases='H2b-143', description='histone cluster 1, H2bb', other_designations='histone 1, H2bb', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23746734, end_position_on_the_genomic_accession=23747223, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319179, CurrentID=0, Status='live', Symbol='Hist1h2be', Aliases='', description='histone cluster 1, H2be', other_designations='histone 1, H2be', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23583670, end_position_on_the_genomic_accession=23621124, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319180, CurrentID=0, Status='live', Symbol='Hist1h2bf', Aliases='', description='histone cluster 1, H2bf', other_designations='H2B 291A|histone 1, H2bf', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23573760, end_position_on_the_genomic_accession=23574190, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319181, CurrentID=0, Status='live', Symbol='Hist1h2bg', Aliases='', description='histone cluster 1, H2bg', other_designations='histone 1, H2bg', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23571400, end_position_on_the_genomic_accession=23571863, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319182, CurrentID=0, Status='live', Symbol='Hist1h2bh', Aliases='H2b-221', description='histone cluster 1, H2bh', other_designations='histone 1, H2bh', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23542923, end_position_on_the_genomic_accession=23543444, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319183, CurrentID=0, Status='live', Symbol='Hist1h2bj', Aliases='', description='histone cluster 1, H2bj', other_designations='H2B 291A|histone 1, H2bj', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22043230, end_position_on_the_genomic_accession=22043658, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319184, CurrentID=0, Status='live', Symbol='Hist1h2bk', Aliases='', description='histone cluster 1, H2bk', other_designations='histone 1, H2bk', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22035821, end_position_on_the_genomic_accession=22036320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319185, CurrentID=0, Status='live', Symbol='Hist1h2bl', Aliases='H2b-291a', description='histone cluster 1, H2bl', other_designations='H2B 291A|histone 1, H2bl', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21715713, end_position_on_the_genomic_accession=21716143, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319186, CurrentID=0, Status='live', Symbol='Hist1h2bm', Aliases='H2b-291b', description='histone cluster 1, H2bm', other_designations='H2B 291B|histone 1, H2bm', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21722044, end_position_on_the_genomic_accession=21722526, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319187, CurrentID=0, Status='live', Symbol='Hist1h2bn', Aliases='', description='histone cluster 1, H2bn', other_designations='H2B 291A|histone 1, H2bn', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21754123, end_position_on_the_genomic_accession=21754553, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319188, CurrentID=0, Status='live', Symbol='Hist1h2bp', Aliases='', description='histone cluster 1, H2bp', other_designations='histone 1, H2bp', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21787480, end_position_on_the_genomic_accession=21801561, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  319189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319189, CurrentID=0, Status='live', Symbol='Hist2h2bb', Aliases='H2b-616', description='histone cluster 2, H2bb', other_designations='H2b 616|histone 2, H2bb', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96269700, end_position_on_the_genomic_accession=96270192, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319190, CurrentID=0, Status='live', Symbol='Hist2h2be', Aliases='AV127319, H2b-613, T25626', description='histone cluster 2, H2be', other_designations='H2b 613|histone 2, H2be', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96221121, end_position_on_the_genomic_accession=96223738, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319191, CurrentID=0, Status='live', Symbol='Hist1h2ai', Aliases='H2a-291A', description='histone cluster 1, H2ai', other_designations='histone 1, H2ai|histone 2, H2ai', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21716412, end_position_on_the_genomic_accession=21716859, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  319192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319192, CurrentID=0, Status='live', Symbol='Hist2h2aa2', Aliases='H2a-614, H2a-615', description='histone cluster 2, H2aa2', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96239779, end_position_on_the_genomic_accession=96240374, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319195, CurrentID=0, Status='live', Symbol='Rpl17', Aliases='', description='ribosomal protein L17', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74998978, end_position_on_the_genomic_accession=75003381, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319196, CurrentID=0, Status='live', Symbol='Ankef1', Aliases='AV276460, Ankrd5', description='ankyrin repeat and EF-hand domain containing 1', other_designations='ankyrin repeat domain 5|ankyrin repeat domain-containing protein 5', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=136532321, end_position_on_the_genomic_accession=136555854, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  319197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319197, CurrentID=0, Status='live', Symbol='Gpr4', Aliases='', description='G protein-coupled receptor 4', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19212499, end_position_on_the_genomic_accession=19224176, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  319200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319200, CurrentID=0, Status='live', Symbol='Gpr82', Aliases='', description='G protein-coupled receptor 82', other_designations='G protein-coupled receptor GPR82', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=13658153, end_position_on_the_genomic_accession=13667433, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  319207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319207, CurrentID=0, Status='live', Symbol='Pgbd1', Aliases='4921509E05Rik', description='piggyBac transposable element derived 1', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21421275, end_position_on_the_genomic_accession=21441291, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  319211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319211, CurrentID=0, Status='live', Symbol='Nol4', Aliases='1700013J13Rik, 4930568N03Rik, Gm1262', description='nucleolar protein 4', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=22693152, end_position_on_the_genomic_accession=23041675, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  319217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319217, CurrentID=0, Status='live', Symbol='Vmn2r7', Aliases='4933425M15Rik', description='vomeronasal 2, receptor 7', other_designations='vomeronasal 2, receptor 1 pseudogene|vomeronasal receptor Vmn2r7', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=64690660, end_position_on_the_genomic_accession=64719602, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57080, CurrentID=0, Status='live', Symbol='Gtf2ird1', Aliases='1700012P16Rik, BEN, Cream1, ESTM9, GTF3, Gtf2il, MusTRD1, Tg(Alb1-Myc)166.8Sst, WBSCR11, X83320', description='general transcription factor II I repeat domain-containing 1', other_designations='Alb-c-myc line 166.8|Alb/c-myc line 166.8|GTF2I repeat domain-containing 1|GTF2I repeat domain-containing protein 1|Williams-Beuren syndrome chromosome region 11|binding factor for early enhancer|c-myc line 166.8|general transcription factor 3|muscle TFII-I repeat domain-containing protein 1|transcription factor GTF3 alpha 2|transcription factor GTF3 gamma 2', map_location='5 74.55 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134357661, end_position_on_the_genomic_accession=134456716, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  319229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319229, CurrentID=0, Status='live', Symbol='Sctr', Aliases='6530402O03Rik', description='secretin receptor', other_designations='SCT-R', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120006891, end_position_on_the_genomic_accession=120063549, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  319236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319236, CurrentID=0, Status='live', Symbol='Trim12c', Aliases='9230105E10Rik, Trim12-2', description='tripartite motif-containing 12C', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104338754, end_position_on_the_genomic_accession=104353358, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  319239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319239, CurrentID=0, Status='live', Symbol='Npsr1', Aliases='9330128H10Rik, GPRA, Gpr154, MVTR, PGR14, VRR1', description='neuropeptide S receptor 1', other_designations='G protein-coupled receptor 154|G protein-coupled receptor PGR14|G-protein coupled receptor 154|G-protein coupled receptor PGR14|vasotocin receptor', map_location='9 A3-A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=24097950, end_position_on_the_genomic_accession=24317590, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  319259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319259, CurrentID=0, Status='live', Symbol='Bricd5', Aliases='9930021D14Rik', description='BRICHOS domain containing 5', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24473884, end_position_on_the_genomic_accession=24475469, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  319262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319262, CurrentID=0, Status='live', Symbol='Fchsd1', Aliases='A030002D08Rik', description='FCH and double SH3 domains 1', other_designations='FCH and double SH3 domains protein 1', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37957434, end_position_on_the_genomic_accession=37969814, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  319263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319263, CurrentID=0, Status='live', Symbol='Pcmtd1', Aliases='8430411F12Rik, A030012M09Rik', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase domain containing 1', other_designations='', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=7088891, end_position_on_the_genomic_accession=7173628, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319266, CurrentID=0, Status='live', Symbol='A130010J15Rik', Aliases='', description='RIKEN cDNA A130010J15 gene', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193173469, end_position_on_the_genomic_accession=193177832, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  319269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319269, CurrentID=0, Status='live', Symbol='A130040M12Rik', Aliases='CA782090, H3053C11, VL30', description='RIKEN cDNA A130040M12 gene', other_designations='VL30 retroelement', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86812765, end_position_on_the_genomic_accession=86816129, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319277, CurrentID=0, Status='live', Symbol='A230046K03Rik', Aliases='AA589518, Kiaa1033, mKIAA1033', description='RIKEN cDNA A230046K03 gene', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=83543941, end_position_on_the_genomic_accession=83596486, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  319278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319278, CurrentID=0, Status='live', Symbol='A230050P20Rik', Aliases='', description='RIKEN cDNA A230050P20 gene', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20868592, end_position_on_the_genomic_accession=20874307, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57138, CurrentID=0, Status='live', Symbol='Slc12a5', Aliases='KCC2, mKIAA1176', description='solute carrier family 12, member 5', other_designations='K-Cl cotransporter 2|KCC2a variant 1|KCC2a-S25 variant 1|KCC2b variant 2|electroneutral potassium-chloride cotransporter 2|mKCC2|neuronal K-Cl cotransporter', map_location='2 G2-G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164960818, end_position_on_the_genomic_accession=164999731, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  319293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319293, CurrentID=0, Status='live', Symbol='A530099J19Rik', Aliases='', description='RIKEN cDNA A530099J19 gene', other_designations='', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19727417, end_position_on_the_genomic_accession=19732955, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57170, CurrentID=0, Status='live', Symbol='Dolpp1', Aliases='0610011H20Rik, AB030189, LSFR2', description='dolichyl pyrophosphate phosphatase 1', other_designations='protein 2-23', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30392124, end_position_on_the_genomic_accession=30400529, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  319317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319317, CurrentID=0, Status='live', Symbol='Snhg11', Aliases='A930034L06Rik, AI854265, AL022637, AW319638, E130013N09Rik', description='small nucleolar RNA host gene 11', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158375638, end_position_on_the_genomic_accession=158386145, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  319322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319322, CurrentID=0, Status='live', Symbol='Sf3b2', Aliases='145kDa, 2610311M13Rik, 2810441F20Rik, B230398H18Rik, SAP145, SF3b1, SF3b145, SF3b150', description='splicing factor 3b, subunit 2', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5273921, end_position_on_the_genomic_accession=5295456, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  319352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319352, CurrentID=0, Status='live', Symbol='Pianp', Aliases='AI255183, AU067746, C530028O21Rik, Panp', description='PILR alpha associated neural protein', other_designations='PILR-associating neural protein|brain protein 1|paired immunoglobin-like type 2 receptor-associating neural protein', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124996720, end_position_on_the_genomic_accession=125003096, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  319370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319370, CurrentID=0, Status='live', Symbol='Ubald2', Aliases='1110014K08Rik, D030012E24Rik, Fam100b', description='UBA-like domain containing 2', other_designations='family with sequence similarity 100, member B|protein FAM100B', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116434094, end_position_on_the_genomic_accession=116439077, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57230, CurrentID=0, Status='live', Symbol='Sap30bp', Aliases='2700016D05Rik, Hcngp', description='SAP30 binding protein', other_designations='transcriptional regulator protein (Hcngp)|transcriptional regulator protein HCNGP', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115933578, end_position_on_the_genomic_accession=115965534, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  319387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319387, CurrentID=0, Status='live', Symbol='Adgrl3', Aliases='5430402I23Rik, CIRL-3, D130075K09Rik, Gm1379, LEC3, Lphn3, mKIAA0768', description='adhesion G protein-coupled receptor L3', other_designations='calcium-independent alpha-latrotoxin receptor homolog 3|latrophilin 3|latrophilin-3|lectomedin 3|lectomedin-3', map_location='5 D-E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=81021199, end_position_on_the_genomic_accession=81827825, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  57246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57246, CurrentID=0, Status='live', Symbol='Tbx20', Aliases='9430010M06Rik, AL022859, Tbx12', description='T-box 20', other_designations='T-box protein 20', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=24720812, end_position_on_the_genomic_accession=24774303, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57247, CurrentID=0, Status='live', Symbol='Zfp276', Aliases='AW048709, D8Ertd370e, D8Ertd377e, Znf276', description='zinc finger protein (C2H2 type) 276', other_designations='zfp-276', map_location='8 72.09 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123254195, end_position_on_the_genomic_accession=123270551, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  57248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57248, CurrentID=0, Status='live', Symbol='Ly6i', Aliases='AI789751, Ly-6I, Ly-6M', description='lymphocyte antigen 6 complex, locus I', other_designations='lymphocyte antigen 6M', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74979537, end_position_on_the_genomic_accession=74990574, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  57249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57249, CurrentID=0, Status='live', Symbol='Gabrq', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, subunit theta', other_designations='gamma-aminobutyric acid (GABA-A) receptor, subunit theta', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72825163, end_position_on_the_genomic_accession=72842607, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57250, CurrentID=0, Status='live', Symbol='Olfr653', Aliases='MOR33-1, Ors19', description='olfactory receptor 653', other_designations='GA_x6K02T2PBJ9-7206970-7207923|odorant receptor S19|olfactory receptor MOR33-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104579624, end_position_on_the_genomic_accession=104580669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  57251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57251, CurrentID=0, Status='live', Symbol='Olfr870', Aliases='MOR141-1, Olfr72', description='olfactory receptor 870', other_designations='GA_x6K02T2PVTD-13912679-13911744|olfactory receptor 72|olfactory receptor MOR141-1', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20170634, end_position_on_the_genomic_accession=20171569, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57252, CurrentID=0, Status='live', Symbol='Tas2r105', Aliases='T2R05, T2R9, T2r5, Tas2r5, mGR05, mt2r5', description='taste receptor, type 2, member 105', other_designations='T2R105|candidate taste receptor T2R5|taste receptor T2R5|taste receptor type 2 member 5|taste receptor type 2 member 9', map_location='6 63.53 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131686561, end_position_on_the_genomic_accession=131687463, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57253, CurrentID=0, Status='live', Symbol='Tas2r108', Aliases='T2R108, T2r8, Tas2r4, Tas2r8, mGR08, mt2r8', description='taste receptor, type 2, member 108', other_designations='T2R4|candidate taste receptor T2R8|taste receptor T2R8|taste receptor type 2 member 8', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40493592, end_position_on_the_genomic_accession=40494485, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  57254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57254, CurrentID=0, Status='live', Symbol='Tas2r119', Aliases='T2r19, Tas2r19, mGR19, mt2r19', description='taste receptor, type 2, member 119', other_designations='T2R119|candidate taste receptor T2R19|taste receptor type 2 member 19', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=32177289, end_position_on_the_genomic_accession=32178294, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  57255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57255, CurrentID=0, Status='live', Symbol='Cldn13', Aliases='2500004B14Rik', description='claudin 13', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134914250, end_position_on_the_genomic_accession=134915530, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57256, CurrentID=0, Status='live', Symbol='Prss21', Aliases='1700023E12Rik, TESP5, mT4, testisin', description='protease, serine 21', other_designations='serine protease 21|testicular serine protease 5|tryptase 4', map_location='17 12.11 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23867873, end_position_on_the_genomic_accession=23873113, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57257, CurrentID=0, Status='live', Symbol='Vav3', Aliases='A530094I06Rik, AA986410, Idd18.1', description='vav 3 oncogene', other_designations='VAV-3', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=109340259, end_position_on_the_genomic_accession=109685694, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  57258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57258, CurrentID=0, Status='live', Symbol='Xpo4', Aliases='B430309A01Rik, mKIAA1721', description='exportin 4', other_designations='exp4', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57577520, end_position_on_the_genomic_accession=57664985, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  57259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57259, CurrentID=0, Status='live', Symbol='Tob2', Aliases='2900090N22Rik, 4930545K18Rik, AV071822, mKIAA1663', description='transducer of ERBB2, 2', other_designations='transducer of erbB-2 2', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81848270, end_position_on_the_genomic_accession=81858326, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57260, CurrentID=0, Status='live', Symbol='Ltb4r2', Aliases='5830462O07Rik, BLT2, LTB4-R 2, LTB4-R-2, LTB4-R2', description='leukotriene B4 receptor 2', other_designations='leukotriene B4 receptor BLT2', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55757916, end_position_on_the_genomic_accession=55764066, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57261, CurrentID=0, Status='live', Symbol='Brd4', Aliases='Brd5, HUNK1, MCAP, WI-11513', description='bromodomain containing 4', other_designations='bromodomain-containing 5|fsh-like protein|mitotic chromosome-associated protein', map_location='17 17.39 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32196272, end_position_on_the_genomic_accession=32284722, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  57262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57262, CurrentID=0, Status='live', Symbol='Retnla', Aliases='1810019L16Rik, Fizz-1, Fizz1, HIMF, RELM-alpha, RELMa, RELMalpha, Xcp2', description='resistin like alpha', other_designations='cysteine-rich secreted protein A12-gamma|cysteine-rich secreted protein FIZZ1|found in inflammatory zone 1|hypoxia-induced mitogenic factor|parasite-induced macrophage novel gene 1 protein', map_location='16 30.75 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48842552, end_position_on_the_genomic_accession=48844461, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57263, CurrentID=0, Status='live', Symbol='Retnlb', Aliases='9030012B21Rik, Fizz2, RELMbeta, Relmb, Xcp3', description='resistin like beta', other_designations='cysteine-rich secreted protein A12-beta|cysteine-rich secreted protein FIZZ2|found in inflammatory zone 2|resistin-like molecule beta', map_location='16 30.75 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48816856, end_position_on_the_genomic_accession=48818892, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57264, CurrentID=0, Status='live', Symbol='Retn', Aliases='ADSF, Fizz3, Rstn, Xcp4', description='resistin', other_designations='adipose tissue-specific secretory factor|adipose-specific cysteine-rich secreted protein A12-alpha|cysteine-rich secreted protein FIZZ3|dominant inhibitory adipocyte-specific secretory factor|found in inflammatory zone 3', map_location='8 A1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3655770, end_position_on_the_genomic_accession=3659818, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  57265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57265, CurrentID=0, Status='live', Symbol='Fzd2', Aliases='AL033370, AW456835, Fz10, Fzd10, Mfz10, Mfz10a', description='frizzled class receptor 2', other_designations='frizzled homolog 2|frizzled-10|fz-10|fz-2|mFz2', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102604431, end_position_on_the_genomic_accession=102608058, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  57266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57266, CurrentID=0, Status='live', Symbol='Cxcl14', Aliases='1110031L23Rik, 1200006I23Rik, AI414372, BMAC, BRAK, KS1, Kec, MIP-2g, MIP2gamma, NJAC, Scyb14, bolekine', description='chemokine (C-X-C motif) ligand 14', other_designations='B-cell and monocyte-activating chemokine|chemokine BRAK|kidney-expressed chemokine CXC|musculus CXC chemokine MIP-2gamma|small inducible cytokine subfamily B (Cys-X-Cys), member 14|small-inducible cytokine B14', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56288643, end_position_on_the_genomic_accession=56296792, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  57267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57267, CurrentID=0, Status='live', Symbol='Apba3', Aliases='AW208791, Mint-3, X11gamma, lin-10, mint3', description='amyloid beta (A4) precursor protein-binding, family A, member 3', other_designations='Mint 3|adapter protein X11gamma|neuron-specific X11L2 protein|neuronal munc18-1-interacting protein 3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81266965, end_position_on_the_genomic_accession=81273247, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57269, CurrentID=0, Status='live', Symbol='Olfr1507', Aliases='MOR244-1, Mor28, Or28', description='olfactory receptor 1507', other_designations='GA_x6K02T2RJGY-491851-492792|gene for odorant receptor MOR28|olfactory receptor MOR244-1', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52489935, end_position_on_the_genomic_accession=52495695, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57270, CurrentID=0, Status='live', Symbol='Olfr1508', Aliases='MOR244-2, MOR244-4, Mor10, Or10', description='olfactory receptor 1508', other_designations='GA_x6K02T2RJGY-520647-521579|gene for odorant receptor MOR10|olfactory receptor MOR244-2', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52462721, end_position_on_the_genomic_accession=52467461, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  319415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319415, CurrentID=0, Status='live', Symbol='Hs3st5', Aliases='D930005L05Rik, Gm1151, Hs3ost5', description='heparan sulfate (glucosamine) 3-O-sulfotransferase 5', other_designations='heparan sulfate D-glucosaminyl 3-O-sulfotransferase 5', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=36506804, end_position_on_the_genomic_accession=36834397, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57272, CurrentID=0, Status='live', Symbol='Olfr140', Aliases='A16, MOR235-1, Ora16', description='olfactory receptor 140', other_designations='GA_x6K02T2Q125-51484508-51483600|odorant receptor A16|olfactory receptor 235-1|olfactory receptor 4A', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90051414, end_position_on_the_genomic_accession=90063591, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57274, CurrentID=0, Status='live', Symbol='Slc16a8', Aliases='Mct3', description='solute carrier family 16 (monocarboxylic acid transporters), member 8', other_designations='MCT 3|preferentially expressed in RPE cells and choroid plexus epithelium|proton-coupled monocarboxylate transporter 3|solute carrier family 16 member 8', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79251016, end_position_on_the_genomic_accession=79254748, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  57275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57275, CurrentID=0, Status='live', Symbol='Lenep', Aliases='AU018355, Lep503', description='lens epithelial protein', other_designations='lens epithelial protein 503', map_location='3 39.1 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89401896, end_position_on_the_genomic_accession=89402650, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57276, CurrentID=0, Status='live', Symbol='Vsig2', Aliases='1190004B15Rik, 2210413P10Rik, AU040484, CTX, Ctm', description='V-set and immunoglobulin domain containing 2', other_designations='CT-like protein|CTM|cortical thymocyte receptor (X. laevis CTX) like|cortical thymocyte-like protein', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37538814, end_position_on_the_genomic_accession=37545123, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  57277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57277, CurrentID=0, Status='live', Symbol='Slurp1', Aliases='1110021N19Rik, AI415082, ARS, ArsB, Slurp-1', description='secreted Ly6/Plaur domain containing 1', other_designations='Ars component B|secreted Ly-6/uPAR related protein 1', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74726644, end_position_on_the_genomic_accession=74728026, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57278, CurrentID=0, Status='live', Symbol='Bcam', Aliases='1200005K12Rik, B-CAM, Gplu, Lu', description='basal cell adhesion molecule', other_designations='Auberger b antigen|B-CAM cell surface glycoprotein|Lutheran blood group|lutheran antigen|lutheran glycoprotein', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19756138, end_position_on_the_genomic_accession=19770532, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  57279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57279, CurrentID=0, Status='live', Symbol='Slc25a20', Aliases='1110007P09Rik, C78826, Cact, mCAC', description='solute carrier family 25 (mitochondrial carnitine/acylcarnitine translocase), member 20', other_designations='CAC|carnitine/acylcarnitine translocase|mitochondrial carnitine-acylcarnitine translocase|solute carrier family 25 member 20', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108662098, end_position_on_the_genomic_accession=108684641, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  319430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319430, CurrentID=0, Status='live', Symbol='C5ar2', Aliases='C5L2, E030029A11Rik, Gpr77', description='complement component 5a receptor 2', other_designations='C5a anaphylatoxin chemotactic receptor C5L2|G protein-coupled receptor 77|G-protein coupled receptor 77', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16234585, end_position_on_the_genomic_accession=16244154, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  319433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319433, CurrentID=0, Status='live', Symbol='Serpine3', Aliases='E130113E03Rik', description='serpin peptidase inhibitor, clade E (nexin, plasminogen activator inhibitor type 1), member 3', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62663667, end_position_on_the_genomic_accession=62692243, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57294, CurrentID=0, Status='live', Symbol='Rps27', Aliases='3200001M24Rik', description='ribosomal protein S27', other_designations='rps27', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90212667, end_position_on_the_genomic_accession=90213648, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  57295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57295, CurrentID=0, Status='live', Symbol='Icmt', Aliases='1700008E11Rik, C80758, Gm13095, HSTE14, OTTMUSG00000010406, PPMT, STE14, pcCMT', description='isoprenylcysteine carboxyl methyltransferase', other_designations='prenylated protein carboxyl methyltransferase|prenylcysteine carboxyl methyltransferase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152297214, end_position_on_the_genomic_accession=152307126, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57296, CurrentID=0, Status='live', Symbol='Psmd8', Aliases='6720456J22Rik, AA407360, AL033291, AL033322, AL033323, C76433', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 8', other_designations='26S proteasome regulatory subunit RPN12|26S proteasome regulatory subunit S14|proteasome 26S non-ATPase subunit 8', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29174187, end_position_on_the_genomic_accession=29180673, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  319446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319446, CurrentID=0, Status='live', Symbol='Dpep2', Aliases='F630103D06Rik, MBD-2', description='dipeptidase 2', other_designations='membrane-bound dipeptidase 2|putative membrane-bound dipeptidase 2', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105984945, end_position_on_the_genomic_accession=106006348, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  319448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319448, CurrentID=0, Status='live', Symbol='Fndc3a', Aliases='1700094E19Rik, D14Ertd453e, F730017H24Rik, Fndc3, sys', description='fibronectin type III domain containing 3A', other_designations='fibronectin domain-containing protein 3a', map_location='14 37.62 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=72537948, end_position_on_the_genomic_accession=72710686, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  319455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319455, CurrentID=0, Status='live', Symbol='Pld5', Aliases='B230365F16Rik', description='phospholipase D family, member 5', other_designations='inactive PLD 5|inactive choline phosphatase 5|inactive phosphatidylcholine-hydrolyzing phospholipase D5|phospholipase D5', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175962301, end_position_on_the_genomic_accession=176280017, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  57312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57312, CurrentID=0, Status='live', Symbol='Mrps31', Aliases='1500002D03Rik, AA960481, Imogen38, Imogen44', description='mitochondrial ribosomal protein S31', other_designations='MRP-S31|S31mt|imogen 38|islet mitochondrial antigen, 38 kD', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22410472, end_position_on_the_genomic_accession=22431044, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57314, CurrentID=0, Status='live', Symbol='Nelfcd', Aliases='2410003I03Rik, C77797, Th1, Th1l', description='negative elongation factor complex member C/D, Th1l', other_designations='NELF-D|TH1 homolog|TH1-like homolog|TH1-like protein|trihydrophobin 1 homolog', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174415804, end_position_on_the_genomic_accession=174427503, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  57315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57315, CurrentID=0, Status='live', Symbol='Wdr46', Aliases='2310007I04Rik, Bing4, C78447, C78559', description='WD repeat domain 46', other_designations='WD repeat-containing protein BING4|putative WD repeat domain 46', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33940634, end_position_on_the_genomic_accession=33949695, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  57316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57316, CurrentID=0, Status='live', Symbol='C1d', Aliases='1110036E10Rik, AI875855, SUN-CoR', description='C1D nuclear receptor co-repressor', other_designations='mC1D|nuclear DNA binding protein|nuclear DNA-binding protein|small unique nuclear co-repressor|small unique nuclear receptor corepressor|sunCoR', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=17257618, end_position_on_the_genomic_accession=17269176, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57317, CurrentID=0, Status='live', Symbol='Srsf4', Aliases='5730499P16Rik, AW550192, MNCb-2616, SRp75, Sfrs4', description='serine/arginine-rich splicing factor 4', other_designations='splicing factor, arginine/serine-rich 4 (SRp75)', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=131873611, end_position_on_the_genomic_accession=131901725, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  57319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57319, CurrentID=0, Status='live', Symbol='Smpdl3a', Aliases='0610010C24Rik, AI529588, ASM3A, ASML3, ASML3A', description='sphingomyelin phosphodiesterase, acid-like 3A', other_designations='ASM-like phosphodiesterase 3a', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=57794544, end_position_on_the_genomic_accession=57811830, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57320, CurrentID=0, Status='live', Symbol='Park7', Aliases='DJ-1, Dj1', description='Parkinson disease (autosomal recessive, early onset) 7', other_designations='RNA-binding protein regulatory subunit|parkinson disease protein 7 homolog|protein DJ-1', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150897133, end_position_on_the_genomic_accession=150914432, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57321, CurrentID=0, Status='live', Symbol='Terf2ip', Aliases='Rap1', description='telomeric repeat binding factor 2, interacting protein', other_designations='RAP1 homolog|TERF2-interacting telomeric protein 1|TRF2-interacting telomeric protein 1|TRF2-interacting telomeric protein Rap1|repressor/activator protein 1 homolog', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=112011359, end_position_on_the_genomic_accession=112020528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  319468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319468, CurrentID=0, Status='live', Symbol='Ppm1h', Aliases='A430075L18Rik, ARHCL1, C030002B11Rik', description='protein phosphatase 1H (PP2C domain containing)', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=122678106, end_position_on_the_genomic_accession=122945795, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57329, CurrentID=0, Status='live', Symbol='Otor', Aliases='CDRAP, Fdp, MIA, MIAL', description='otoraplin', other_designations='melanoma inhibitory activity-like protein', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=143078492, end_position_on_the_genomic_accession=143081700, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57330, CurrentID=0, Status='live', Symbol='Gigyf1', Aliases='Perq1, mKIAA4110', description='GRB10 interacting GYF protein 1', other_designations='GRB10 interacting GYF protein 2|GRB10-interacting GYF protein 1|PERQ amino acid rich, with GYF domain 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137510258, end_position_on_the_genomic_accession=137527935, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  319475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319475, CurrentID=0, Status='live', Symbol='Zfp672', Aliases='4930488P06Rik, 4930511N19Rik, Znf672', description='zinc finger protein 672', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58315114, end_position_on_the_genomic_accession=58330346, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  319476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319476, CurrentID=0, Status='live', Symbol='Lrtm1', Aliases='A930016D02Rik', description='leucine-rich repeats and transmembrane domains 1', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=29018208, end_position_on_the_genomic_accession=29033642, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  319477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319477, CurrentID=0, Status='live', Symbol='6030419C18Rik', Aliases='', description='RIKEN cDNA 6030419C18 gene', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58488603, end_position_on_the_genomic_accession=58499742, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  319478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319478, CurrentID=0, Status='live', Symbol='Cxxc4', Aliases='9330210J02Rik, C030003J12Rik, Idax', description='CXXC finger 4', other_designations='inhibition of the Dvl and axin complex protein', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=134236415, end_position_on_the_genomic_accession=134262090, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  319480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319480, CurrentID=0, Status='live', Symbol='Itga11', Aliases='4732459H24Rik', description='integrin alpha 11', other_designations='integrin a11', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62677855, end_position_on_the_genomic_accession=62783986, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  319481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319481, CurrentID=0, Status='live', Symbol='Wdr59', Aliases='5430401O09Rik, Cdw12', description='WD repeat domain 59', other_designations='CUL4- and DDB1-associated WDR protein 12', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111448781, end_position_on_the_genomic_accession=111522107, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  319482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319482, CurrentID=0, Status='live', Symbol='9530053A07Rik', Aliases='FCGBP', description='RIKEN cDNA 9530053A07 gene', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28129466, end_position_on_the_genomic_accession=28164811, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  57339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57339, CurrentID=0, Status='live', Symbol='Jph1', Aliases='ENSMUSG00000054314, JP-1, Jp1', description='junctophilin 1', other_designations='junctophilin type 1|mitsugumin72', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16942853, end_position_on_the_genomic_accession=17097932, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57340, CurrentID=0, Status='live', Symbol='Jph3', Aliases='JP-3, Jp3', description='junctophilin 3', other_designations='junctophilin type 3', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121730559, end_position_on_the_genomic_accession=121791083, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  57342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57342, CurrentID=0, Status='live', Symbol='Parva', Aliases='2010012A22Rik, 5430400F08Rik, AI225929, AU042898, Actp, CH-ILKBP, Parvin', description='parvin, alpha', other_designations='actopaxin', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=112427706, end_position_on_the_genomic_accession=112591688, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  57344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57344, CurrentID=0, Status='live', Symbol='As3mt', Aliases='2310045H08Rik, Cyt19', description='arsenic (+3 oxidation state) methyltransferase', other_designations='S-adenosyl-L-methionine:arsenic(III) methyltransferase|S-adenosylmethionine:arsenic (III) methyltransferase|methylarsonite methyltransferase|methyltransferase Cyt19', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46707359, end_position_on_the_genomic_accession=46741099, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  57349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57349, CurrentID=0, Status='live', Symbol='Ppbp', Aliases='2400003M24Rik, AI854500, CTAP3, CTAPIII, Cxcl7, LA-PF4, LDGF, MDGF, NAP-2, NAP-2-L1, Scyb7, TGB, TGB1, THBGB1, b-TG1, beta-TG', description='pro-platelet basic protein', other_designations='chemokine (C-X-C motif) ligand 7', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90768518, end_position_on_the_genomic_accession=90770060, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  57354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57354, CurrentID=0, Status='live', Symbol='Cramp1l', Aliases='5830477H08Rik, AI256853, AI790734, Cramp1, Tce4, mKIAA1426', description='Crm, cramped-like (Drosophila)', other_designations='cramped chromatin regulator homolog 1|hematological and neurological expressed 1-like protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24961226, end_position_on_the_genomic_accession=25015846, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  57355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57355, CurrentID=0, Status='live', Symbol='BC051019', Aliases='AA571491, D7H11orf16, ICRFP703B1614Q5.5', description='cDNA sequence BC051019', other_designations='ICRFP703B1614Q5.5|ICRFP703N2430Q5.5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109712181, end_position_on_the_genomic_accession=109723932, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  57357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57357, CurrentID=0, Status='live', Symbol='Srd5a3', Aliases='1110025P14Rik, A430076C09, AV364670, AW987574, D730040M03Rik, H5ar, S5AR 3, Srd5a2l', description='steroid 5 alpha-reductase 3', other_designations='3-oxo-5-alpha-steroid 4-dehydrogenase 3|H5AR|SR type 3|probable polyprenol reductase|steroid 5 alpha-reductase 2 like|steroid 5 alpha-reductase 2-like; H5AR gene; steroid 5 alpha-reductase 2 like|steroid 5-alpha-reductase 3', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76140273, end_position_on_the_genomic_accession=76155503, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  319504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319504, CurrentID=0, Status='live', Symbol='Nrcam', Aliases='Bravo, C030017F07Rik, C130076O07Rik, mKIAA0343', description='neuronal cell adhesion molecule', other_designations='mBravo|neuron-glia-CAM-related cell adhesion molecule|neuronal surface protein Bravo|ng-CAM-related|ngCAM-related cell adhesion molecule|nr-CAM', map_location='12 20.71 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=44328832, end_position_on_the_genomic_accession=44605194, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  319508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319508, CurrentID=0, Status='live', Symbol='Syt15', Aliases='CHR10SYT, E230025K04Rik, sytXV', description='synaptotagmin XV', other_designations='synaptotagmin 15', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34220046, end_position_on_the_genomic_accession=34230421, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319513, CurrentID=0, Status='live', Symbol='Pced1a', Aliases='A930025D01Rik, AI480484, Fam113a', description='PC-esterase domain containing 1A', other_designations='family with sequence similarity 113, member A|protein FAM113A', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130417684, end_position_on_the_genomic_accession=130424641, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57370, CurrentID=0, Status='live', Symbol='B4galt3', Aliases='9530061M23Rik, AA104562, AW125175, ESTM26, ESTM6, R74981', description='UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 3', other_designations='N-acetyllactosamine synthase|UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 3|UDP-Gal:betaGlcNAc beta 1,4-galactosyltransferase, polypeptide 3; beta-1,4-galactosyltransferase III; beta4GalT-III; expressed sequence tag mouse EST 6|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 3|b4Gal-T3|beta-1,4-GalTase 3|beta-1,4-galactosyltransferase III|beta4Gal-T3|beta4GalT-III', map_location='1 79.29 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171270328, end_position_on_the_genomic_accession=171276895, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57373, CurrentID=0, Status='live', Symbol='Akip1', Aliases='BCA3, C11orf17, D7H11orf17, D930014E17Rik, ICRFP703B1614Q5.6, ORF27', description='A kinase (PRKA) interacting protein 1', other_designations='ICRFP703N2430Q5.6|PKA-interacting protein|breast cancer associated 3|breast cancer-associated gene 3 protein|proline-rich protein BCA3|protein kinase A-interacting protein 1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109703670, end_position_on_the_genomic_accession=109712186, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319518, CurrentID=0, Status='live', Symbol='Pdpr', Aliases='4930402E16Rik, AU018056, mKIAA1990', description='pyruvate dehydrogenase phosphatase regulatory subunit', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111094599, end_position_on_the_genomic_accession=111139779, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  57376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57376, CurrentID=0, Status='live', Symbol='Smarce1', Aliases='2810417B20Rik, 5830412H02Rik, 9030408N19Rik, Baf27', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily e, member 1', other_designations='BAF57|BRG1-associated factor 57|SWI/SNF-related matrix-associated actin-dependent regulator chromatin subfamily E member 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99209047, end_position_on_the_genomic_accession=99231339, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  57377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57377, CurrentID=0, Status='live', Symbol='Mogs', Aliases='1810017N02Rik, AI181835, Gcs1', description='mannosyl-oligosaccharide glucosidase', other_designations='glucosidase 1|glucosidase I|glycoprotein-processing glucosidase I', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83115506, end_position_on_the_genomic_accession=83118902, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57385, CurrentID=0, Status='live', Symbol='P2ry4', Aliases='P2Y4, P2Y4R', description='pyrimidinergic receptor P2Y, G-protein coupled, 4', other_designations='P2Y4 receptor|purinergic receptor P2Y, G-protein coupled, 4', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100588358, end_position_on_the_genomic_accession=100594927, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  319530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319530, CurrentID=0, Status='live', Symbol='Zfp750', Aliases='A030007D23Rik, Znf750', description='zinc finger protein 750', other_designations='protein ZNF750', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121510976, end_position_on_the_genomic_accession=121519342, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57390, CurrentID=0, Status='live', Symbol='Psors1c2', Aliases='AI448740, Pcg, SPR1', description='psoriasis susceptibility 1 candidate 2 (human)', other_designations='MHC psoriasis candidate', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35532207, end_position_on_the_genomic_accession=35534647, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  319535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319535, CurrentID=0, Status='live', Symbol='Zfp182', Aliases='6330583I20Rik, Znf182', description='zinc finger protein 182', other_designations='zinc finger protein 21', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=21026184, end_position_on_the_genomic_accession=21062038, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57394, CurrentID=0, Status='live', Symbol='Tmem27', Aliases='0610008J07Rik, NX-17, NX17', description='transmembrane protein 27', other_designations='kidney-specific membrane protein', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164088830, end_position_on_the_genomic_accession=164118860, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  319552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319552, CurrentID=0, Status='live', Symbol='Spx', Aliases='B230216G23Rik, Npq', description='spexin hormone', other_designations='neuropeptide Q', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142410186, end_position_on_the_genomic_accession=142419740, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  319554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319554, CurrentID=0, Status='live', Symbol='Idi1', Aliases='4832416K17Rik', description='isopentenyl-diphosphate delta isomerase', other_designations='IPP isomerase 1|IPPI1|isopentenyl pyrophosphate isomerase 1', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=8885606, end_position_on_the_genomic_accession=8892396, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  319555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319555, CurrentID=0, Status='live', Symbol='Nwd1', Aliases='A230063L24Rik', description='NACHT and WD repeat domain containing 1', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72646225, end_position_on_the_genomic_accession=72716880, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  319565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319565, CurrentID=0, Status='live', Symbol='Syne2', Aliases='6820443O06Rik, AW546258, D12Ertd777e, NUA, Nesp2g, mKIAA1011', description='spectrin repeat containing, nuclear envelope 2', other_designations='nesprin 2|nuance|nuclear envelope spectrin repeat protein 2|nucleus and actin connecting element protein|synaptic nuclear envelope 2|synaptic nuclear envelope protein 2', map_location='12 33.11 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=75818318, end_position_on_the_genomic_accession=76110928, orientation='plus', exon_count=116, OMIM='', no_hdr0=''),
  57423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57423, CurrentID=0, Status='live', Symbol='Atp5j2', Aliases='1110019H14Rik', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit F2', other_designations='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit f|hypothetical protein, clone:2-31', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145183706, end_position_on_the_genomic_accession=145191592, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  57426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57426, CurrentID=0, Status='live', Symbol='Scgb1b2', Aliases='Abpa2, Abph, Apbh, LGP, Mja1l, Scgb1b1', description='secretoglobin, family 1B, member 2', other_designations='androgen binding protein eta|lacrimal gland protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31290519, end_position_on_the_genomic_accession=31291816, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57429, CurrentID=0, Status='live', Symbol='Sult5a1', Aliases='Sult-x1, Sultx1', description='sulfotransferase family 5A, member 1', other_designations='sulfotransferase related gene X1|sulfotransferase-related protein SULT-X1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123142843, end_position_on_the_genomic_accession=123158914, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57430, CurrentID=0, Status='live', Symbol='Sult3a1', Aliases='ST3A1, Sult-x2, Sultx2', description='sulfotransferase family 3A, member 1', other_designations='sulfotransferase 3A1|sulfotransferase related gene X2|sulfotransferase-related protein SULT-X2', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=33851020, end_position_on_the_genomic_accession=33879475, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57431, CurrentID=0, Status='live', Symbol='Dnajc4', Aliases='2010301J22Rik, Hspf2, Mcg18', description='DnaJ heat shock protein family (Hsp40) member C4', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 4|heat shock 40kD protein 2|multiple endocrine neoplasia type 1 candidate protein 18|multiple endocrine neoplasia type 1 candidate protein number 18 homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6987911, end_position_on_the_genomic_accession=6992696, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57432, CurrentID=0, Status='live', Symbol='Zc3h8', Aliases='AU020882, E130108N08Rik, Fliz1, Zc3hdc8', description='zinc finger CCCH type containing 8', other_designations='fetal liver zinc finger 1|fetal liver zinc finger 1; putative zinc finger protein FLIZ1|fetal liver zinc finger protein 1|zinc finger CCCH type domain containing 8', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128926268, end_position_on_the_genomic_accession=128944084, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57434, CurrentID=0, Status='live', Symbol='Xrcc2', Aliases='4921524O04Rik, 8030409M04Rik, RAD51, RecA', description='X-ray repair complementing defective repair in Chinese hamster cells 2', other_designations='X-ray repair cross-complementing protein 2|recA/RAD51 family protein', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=25689814, end_position_on_the_genomic_accession=25705797, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  319579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319579, CurrentID=0, Status='live', Symbol='Defb20', Aliases='9230107O10Rik, BD-20, mBD-20', description='defensin beta 20', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152477063, end_position_on_the_genomic_accession=152479934, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  57436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57436, CurrentID=0, Status='live', Symbol='Gabarapl1', Aliases='3110025G09Rik, 9130422N19Rik, AI196471, Apg8l, Atg8l, GECI, MNCb-0091', description='gamma-aminobutyric acid (GABA) A receptor-associated protein-like 1', other_designations='GABA(A) receptor-associated protein like 1|GABA(A) receptor-associated protein-like 1|GEC-1|glandular epithelial cell protein 1', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129533192, end_position_on_the_genomic_accession=129542331, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57437, CurrentID=0, Status='live', Symbol='Golga7', Aliases='AB041568, C130038N16Rik, GCP16, GOLGA3AP1, HSPC041, R75586', description='golgi autoantigen, golgin subfamily a, 7', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=23241326, end_position_on_the_genomic_accession=23257183, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57438, CurrentID=0, Status='live', Symbol='March7', Aliases='Axo, Axot, Gtrgeo17', description='membrane-associated ring finger (C3HC4) 7', other_designations='MARCH-VII|axotrophin|gene trap ROSA b-geo 17|membrane-associated RING finger protein 7|membrane-associated RING-CH protein VII', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=60209892, end_position_on_the_genomic_accession=60249211, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  57439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57439, CurrentID=0, Status='live', Symbol='Tmem183a', Aliases='1300007B12Rik, 1700020N19Rik, Tmem183', description='transmembrane protein 183A', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134346097, end_position_on_the_genomic_accession=134362001, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57440, CurrentID=0, Status='live', Symbol='Ehd3', Aliases='Ehd2', description='EH-domain containing 3', other_designations='EH-domain containing protein 2', map_location='17 45.2 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=73804841, end_position_on_the_genomic_accession=73832093, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57441, CurrentID=0, Status='live', Symbol='Gmnn', Aliases='AW546347, Gem', description='geminin', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24751845, end_position_on_the_genomic_accession=24761937, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57442, CurrentID=0, Status='live', Symbol='Kcne3', Aliases='2210017H05Rik, MiRP2', description='potassium voltage-gated channel, Isk-related subfamily, gene 3', other_designations='minK-related peptide 2|minimum potassium ion channel-related peptide 2|potassium channel subunit beta MiRP2', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100176507, end_position_on_the_genomic_accession=100184869, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  57443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57443, CurrentID=0, Status='live', Symbol='Fbxo3', Aliases='1200002G09Rik, 1700026K02Rik, AI046358, Fba', description='F-box protein 3', other_designations='F-box protein FBA', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104027799, end_position_on_the_genomic_accession=104063237, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57444, CurrentID=0, Status='live', Symbol='Isg20', Aliases='1600023I01Rik, 2010107M23Rik, 20kDa, DnaQL, HEM45', description='interferon-stimulated protein', other_designations='interferon stimulated exonuclease|promyelocytic leukemia nuclear body-associated protein ISG20', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=78913424, end_position_on_the_genomic_accession=78920396, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  319594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319594, CurrentID=0, Status='live', Symbol='Hif1an', Aliases='2310046M24Rik, A830014H24Rik, FIH, FIH1', description='hypoxia-inducible factor 1, alpha subunit inhibitor', other_designations='hypoxia-inducible factor asparagine hydroxylase', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44562854, end_position_on_the_genomic_accession=44576274, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319601, CurrentID=0, Status='live', Symbol='Zfp653', Aliases='E430039K05Rik, Znf653', description='zinc finger protein 653', other_designations='67 kDa zinc finger protein|Zip67|zinc finger protein Zip67', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22055460, end_position_on_the_genomic_accession=22071370, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  319604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319604, CurrentID=0, Status='live', Symbol='Fam168a', Aliases='2610030B18Rik, B930006L02Rik, mKIAA0280', description='family with sequence similarity 168, member A', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100706640, end_position_on_the_genomic_accession=100841646, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  319613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319613, CurrentID=0, Status='live', Symbol='Sybu', Aliases='Golsyn', description='syntabulin (syntaxin-interacting)', other_designations='Golgi-localized syntaphilin-related protein|Golsyn/Syntabulin|syntaxin-1-binding protein', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=44671856, end_position_on_the_genomic_accession=44788063, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  319615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319615, CurrentID=0, Status='live', Symbol='Zfp944', Aliases='6330416L07Rik', description='zinc finger protein 944', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=22337983, end_position_on_the_genomic_accession=22361581, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  319618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319618, CurrentID=0, Status='live', Symbol='Dcp1b', Aliases='B930050E02Rik', description='decapping mRNA 1B', other_designations='DCP1 decapping enzyme homolog B', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119175251, end_position_on_the_genomic_accession=119222737, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  319622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319622, CurrentID=0, Status='live', Symbol='Itpripl2', Aliases='C130081G24, E030018N11Rik', description='inositol 1,4,5-triphosphate receptor interacting protein-like 2', other_designations='inositol 1,4,5-triphosphate receptor-interacting protein-like 2', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118485111, end_position_on_the_genomic_accession=118491975, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  228033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228033, CurrentID=0, Status='live', Symbol='Atp5g3', Aliases='6030447M23', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C3 (subunit 9)', other_designations='ATP synthase lipid-binding protein, mitochondrial|ATP synthase proteolipid P3|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9)|ATPase protein 9|ATPase subunit c', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73908447, end_position_on_the_genomic_accession=73911326, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  319625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319625, CurrentID=0, Status='live', Symbol='Galm', Aliases='A530057M15Rik, AU015645, AU020959', description='galactose mutarotase', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80127458, end_position_on_the_genomic_accession=80185032, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319634, CurrentID=0, Status='live', Symbol='Efcab5', Aliases='4930563A03Rik', description='EF-hand calcium binding domain 5', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77089904, end_position_on_the_genomic_accession=77199943, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  319636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319636, CurrentID=0, Status='live', Symbol='Fsd1l', Aliases='A230072O16Rik, Ccdc10, Csdufd1, Fsd1nl, OTTMUSG00000007002', description='fibronectin type III and SPRY domain containing 1-like', other_designations='FSD1 N-terminal-like protein|coiled-coil domain-containing protein 10|cystatin and DUF19 domain containing 1', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=53629929, end_position_on_the_genomic_accession=53707009, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  319638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319638, CurrentID=0, Status='live', Symbol='Nt5dc1', Aliases='6030401B09Rik, AW987726, Nt5c2l1', description="5'-nucleotidase domain containing 1", other_designations="5'-nucleotidase, cytosolic II-like 1|cytosolic 5'-nucleotidase II-like protein 1", map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34303609, end_position_on_the_genomic_accession=34418589, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  319642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319642, CurrentID=0, Status='live', Symbol='Rab9b', Aliases='9330195C02Rik', description='RAB9B, member RAS oncogene family', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136858151, end_position_on_the_genomic_accession=136868540, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  319651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319651, CurrentID=0, Status='live', Symbol='Usp37', Aliases='4932415L06Rik, C330008N13Rik', description='ubiquitin specific peptidase 37', other_designations='deubiquitinating enzyme 37|ubiquitin specific protease 37|ubiquitin thioesterase 37|ubiquitin thiolesterase 37|ubiquitin-specific-processing protease 37|usp37', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74435510, end_position_on_the_genomic_accession=74544286, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  319653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319653, CurrentID=0, Status='live', Symbol='Slc25a40', Aliases='B230315F11Rik', description='solute carrier family 25, member 40', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8422838, end_position_on_the_genomic_accession=8454840, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  319655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319655, CurrentID=0, Status='live', Symbol='Podxl2', Aliases='D130074J02Rik, PODLX2', description='podocalyxin-like 2', other_designations='endoglycan', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88842557, end_position_on_the_genomic_accession=88875053, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  319660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319660, CurrentID=0, Status='live', Symbol='Agmo', Aliases='A530016O06Rik, AI790538, Tmem195', description='alkylglycerol monooxygenase', other_designations='transmembrane protein 195', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=37241629, end_position_on_the_genomic_accession=37581932, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  319670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319670, CurrentID=0, Status='live', Symbol='Eml5', Aliases='BC027154, C130068M19Rik', description='echinoderm microtubule associated protein like 5', other_designations='EMAP-5', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98786604, end_position_on_the_genomic_accession=98901494, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  319675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319675, CurrentID=0, Status='live', Symbol='Cep295', Aliases='5830418K08Rik, 5832426L23Rik, Gm1131, Kiaa1731', description='centrosomal protein 295', other_designations='', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15316914, end_position_on_the_genomic_accession=15357820, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  319695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319695, CurrentID=0, Status='live', Symbol='Ankar', Aliases='4932422E22Rik', description='ankyrin and armadillo repeat containing', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72642980, end_position_on_the_genomic_accession=72700579, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  319701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319701, CurrentID=0, Status='live', Symbol='Fbxo48', Aliases='A630050E13Rik', description='F-box protein 48', other_designations='novel F-box domain containing protein', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=16951375, end_position_on_the_genomic_accession=16954777, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  319710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319710, CurrentID=0, Status='live', Symbol='Frmd6', Aliases='2610019M19Rik, 4930488L10Rik, AW212977, Willin', description='FERM domain containing 6', other_designations='', map_location='12 C2-C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70825335, end_position_on_the_genomic_accession=70902234, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  319713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319713, CurrentID=0, Status='live', Symbol='Ablim3', Aliases='AW990719, D930036B08Rik', description='actin binding LIM protein family, member 3', other_designations='abLIM-3|actin-binding LIM protein family member 3', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61799392, end_position_on_the_genomic_accession=61911857, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  319719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319719, CurrentID=0, Status='live', Symbol='Simc1', Aliases='4732471D19Rik', description='SUMO-interacting motifs containing 1', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54503739, end_position_on_the_genomic_accession=54551290, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  319734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319734, CurrentID=0, Status='live', Symbol='Cacna2d4', Aliases='5730412N02Rik, BE686333', description='calcium channel, voltage-dependent, alpha 2/delta subunit 4', other_designations='voltage-gated calcium channel subunit alpha-2/delta-4', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119236474, end_position_on_the_genomic_accession=119353392, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  319740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319740, CurrentID=0, Status='live', Symbol='Zfyve27', Aliases='2210011N02Rik, 9530077C24Rik, AI426636, AI593546, AI835681', description='zinc finger, FYVE domain containing 27', other_designations='zinc finger FYVE domain-containing protein 27', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42170567, end_position_on_the_genomic_accession=42194593, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  319742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319742, CurrentID=0, Status='live', Symbol='Mpzl3', Aliases='5430427F17Rik, A530065I17Rik, rc', description='myelin protein zero-like 3', other_designations='Ig-like domain-containing protein UNQ2966', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45055181, end_position_on_the_genomic_accession=45075042, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  319748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319748, CurrentID=0, Status='live', Symbol='Zfp865', Aliases='6430526N21Rik, Znf865', description='zinc finger protein 865', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5020079, end_position_on_the_genomic_accession=5033226, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  319757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319757, CurrentID=0, Status='live', Symbol='Smo', Aliases='E130215L21Rikh, bnb, smoothened, Smo', description='smoothened, frizzled class receptor', other_designations='bent body', map_location='6 12.36 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29735497, end_position_on_the_genomic_accession=29761366, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  319758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319758, CurrentID=0, Status='live', Symbol='Rfx7', Aliases='2510005N23Rik, 9930116O05Rik, D130086K05Rik, Rfxdc2', description='regulatory factor X, 7', other_designations='regulatory factor X domain containing 2 homolog', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72532167, end_position_on_the_genomic_accession=72622949, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  319765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319765, CurrentID=0, Status='live', Symbol='Igf2bp2', Aliases='C330012H03Rik, IMP-2, Imp2, Neilsen', description='insulin-like growth factor 2 mRNA binding protein 2', other_designations='IGF-II mRNA-binding protein 2|IGF2 mRNA-binding protein 2|VICKZ family member 2|insulin-like growth factor 2, binding protein 2', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22059009, end_position_on_the_genomic_accession=22163299, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  319767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319767, CurrentID=0, Status='live', Symbol='Atp10b', Aliases='5930426O13Rik, 9030605H24Rik', description='ATPase, class V, type 10B', other_designations='potential phospholipid-transporting ATPase VB', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=42939557, end_position_on_the_genomic_accession=43262285, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  319772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319772, CurrentID=0, Status='live', Symbol='C130050O18Rik', Aliases='', description='RIKEN cDNA C130050O18 gene', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139403597, end_position_on_the_genomic_accession=139416092, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  319776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319776, CurrentID=0, Status='live', Symbol='Tmem72', Aliases='C230095G01Rik', description='transmembrane protein 72', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116679103, end_position_on_the_genomic_accession=116716978, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  319800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319800, CurrentID=0, Status='live', Symbol='Slc22a30', Aliases='AA986766, C730048C13Rik, UST1R', description='solute carrier family 22, member 30', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8335371, end_position_on_the_genomic_accession=8405111, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  319801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319801, CurrentID=0, Status='live', Symbol='Tigar', Aliases='9630033F20Rik, AA793651, AI595337, C79710, C85509', description='Trp53 induced glycolysis repulatory phosphatase', other_designations='TP53-induced glycolysis and apoptosis regulator|TP53-induced glycolysis regulatory phosphatase|probable fructose-2,6-bisphosphatase TIGAR', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127085116, end_position_on_the_genomic_accession=127109552, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  319804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319804, CurrentID=0, Status='live', Symbol='Glt1d1', Aliases='5730455A04Rik', description='glycosyltransferase 1 domain containing 1', other_designations='', map_location='5 G1.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=127632130, end_position_on_the_genomic_accession=127708750, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  319807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319807, CurrentID=0, Status='live', Symbol='Nwd2', Aliases='3110047P20Rik, B830017A01Rik, Hn1-ps2, mKIAA1239', description='NACHT and WD repeat domain containing 2', other_designations='hematological and neurological expressed 1', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=63649103, end_position_on_the_genomic_accession=63810546, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319817, CurrentID=0, Status='live', Symbol='Rc3h2', Aliases='2900024N03Rik, 9430019J22Rik, D930043C02Rik, Mnab, Rnf164', description='ring finger and CCCH-type zinc finger domains 2', other_designations='membrane associated DNA binding protein|membrane-associated nucleic acid binding protein|ring finger and CCCH-type zinc finger domain-containing protein 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37368861, end_position_on_the_genomic_accession=37423380, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  319822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319822, CurrentID=0, Status='live', Symbol='Smyd4', Aliases='G430029E23Rik', description='SET and MYND domain containing 4', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75348433, end_position_on_the_genomic_accession=75405707, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  319832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319832, CurrentID=0, Status='live', Symbol='Tmem229a', Aliases='6332401O19Rik', description='transmembrane protein 229A', other_designations='', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24951141, end_position_on_the_genomic_accession=24956125, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  319845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319845, CurrentID=0, Status='live', Symbol='Bbs9', Aliases='E130103I17Rik', description='Bardet-Biedl syndrome 9 (human)', other_designations='EST 3159894|bardet-Biedl syndrome 9 protein homolog|parathyroid hormone-responsive B1 gene protein homolog', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22475542, end_position_on_the_genomic_accession=22888283, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  319848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319848, CurrentID=0, Status='live', Symbol='Slc17a4', Aliases='9130214H05Rik', description='solute carrier family 17 (sodium phosphate), member 4', other_designations='putative small intestine sodium-dependent phosphate transport protein|solute carrier family 17 member 4', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23897879, end_position_on_the_genomic_accession=23915009, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  319875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319875, CurrentID=0, Status='live', Symbol='Tmprss11b', Aliases='9930019B18Riknl, Tmprss11b', description='transmembrane protease, serine 11B', other_designations='airway trypsin-like protease 5|transmembrane protease, serine 11b N terminal like', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86659522, end_position_on_the_genomic_accession=86676298, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  319876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319876, CurrentID=0, Status='live', Symbol='Cobll1', Aliases='1810047P18Rik, Coblr1, D430044D16Rik', description='Cobl-like 1', other_designations='cobl-related 1|cobl-related protein 1', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=65088339, end_position_on_the_genomic_accession=65239480, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  319880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319880, CurrentID=0, Status='live', Symbol='Tmcc3', Aliases='AW488095, C630016B22Rik, C88213, Tmcc1', description='transmembrane and coiled coil domains 3', other_designations='transmembrane and coiled coil domains 1', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=94311912, end_position_on_the_genomic_accession=94590954, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  57737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57737, CurrentID=0, Status='live', Symbol='Rhox4b', Aliases='5430432L21Rik, Ehox, Rhox4.2, Rhox4a', description='reproductive homeobox 4B', other_designations='ES cell derived homeobox containing|homeobox protein EHOX|reproductive homeobox on X chromosome 4', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37432494, end_position_on_the_genomic_accession=37437278, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  57738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57738, CurrentID=0, Status='live', Symbol='Slc15a2', Aliases='8430408C16Rik, C78862, Pept2', description='solute carrier family 15 (H+/peptide transporter), member 2', other_designations='Slc15a2 variant A|Slc15a2 variant A2|kidney H(+)/peptide cotransporter|oligopeptide transporter, kidney isoform|peptide transporter 2|proton-dependent high affinity oligopeptide transporter PepT2', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36750164, end_position_on_the_genomic_accession=36785157, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  57740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57740, CurrentID=0, Status='live', Symbol='Stk32c', Aliases='PKE, Pkek, YANK3', description='serine/threonine kinase 32C', other_designations='PKE protein kinase|stk32c', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139103638, end_position_on_the_genomic_accession=139213307, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  319885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319885, CurrentID=0, Status='live', Symbol='Zcchc7', Aliases='4930572I07Rik, D4Wsu132e', description='zinc finger, CCHC domain containing 7', other_designations='TRAMP-like complex RNA-binding factor ZCCHC7', map_location='4 23.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44756559, end_position_on_the_genomic_accession=44932215, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57742, CurrentID=0, Status='live', Symbol='Abhd1', Aliases='LABH-1, LABH1', description='abhydrolase domain containing 1', other_designations='alpha/beta hydrolase-1|lung alpha/beta hydrolase 1|mmLABH1', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30950066, end_position_on_the_genomic_accession=30955095, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57743, CurrentID=0, Status='live', Symbol='Sec61a2', Aliases='', description='Sec61, alpha subunit 2 (S. cerevisiae)', other_designations='Sec61 alpha|protein transport protein Sec61 subunit alpha|sec61 alpha-2', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5870977, end_position_on_the_genomic_accession=5895427, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  319888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319888, CurrentID=0, Status='live', Symbol='Oacyl', Aliases='5330437I02Rik, Acyl3', description='O-acyltransferase like', other_designations='acyltransferase 3', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65697370, end_position_on_the_genomic_accession=65751624, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  57745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57745, CurrentID=0, Status='live', Symbol='Zfp112', Aliases='Znf112, Znf228', description='zinc finger protein 112', other_designations='zfp-112|zinc finger protein 112 homolog|zinc finger protein 112; zinc finger protein ZFP112', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24111898, end_position_on_the_genomic_accession=24127952, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  57746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57746, CurrentID=0, Status='live', Symbol='Piwil2', Aliases='Piwil1l, mili', description='piwi-like RNA-mediated gene silencing 2', other_designations='Miwi like|Miwi likw|piwi like homolog 2|piwi-like homolog 2', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70372477, end_position_on_the_genomic_accession=70436916, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  57748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57748, CurrentID=0, Status='live', Symbol='Jmy', Aliases='', description='junction-mediating and regulatory protein', other_designations='p300 transcriptional cofactor JMY', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93430097, end_position_on_the_genomic_accession=93499808, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  57749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57749, CurrentID=0, Status='live', Symbol='Piwil1', Aliases='MIWI', description='piwi-like RNA-mediated gene silencing 1', other_designations='piwi like homolog 1|piwi-like homolog 1', map_location='5 67.86 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=128736070, end_position_on_the_genomic_accession=128755474, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  57750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57750, CurrentID=0, Status='live', Symbol='Wdr12', Aliases='4933402C23Rik, Ytm1, Ytm1p', description='WD repeat domain 12', other_designations='WD repeat-containing protein 12|nuclear protein Ytm1', map_location='1 C1-C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60076868, end_position_on_the_genomic_accession=60098500, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  57751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57751, CurrentID=0, Status='live', Symbol='Rnf25', Aliases='0610009H16Rik, AO7', description='ring finger protein 25', other_designations='RING finger protein AO7', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74593748, end_position_on_the_genomic_accession=74601431, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  57752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57752, CurrentID=0, Status='live', Symbol='Tacc2', Aliases='mKIAA4180', description='transforming, acidic coiled-coil containing protein 2', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=130553010, end_position_on_the_genomic_accession=130764789, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  57753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57753, CurrentID=0, Status='live', Symbol='Noc3l', Aliases='AF233884, Fad24', description='nucleolar complex associated 3 homolog (S. cerevisiae)', other_designations='NOC3 protein homolog|NOC3-like protein|factor for adipocyte differentiation 24|nucleolar complex-associated protein 3-like protein', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38788128, end_position_on_the_genomic_accession=38819270, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  57754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57754, CurrentID=0, Status='live', Symbol='Cend1', Aliases='1500001H12Rik, AI415214, BM88', description='cell cycle exit and neuronal differentiation 1', other_designations='BM88 antigen', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141426446, end_position_on_the_genomic_accession=141429538, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57755, CurrentID=0, Status='live', Symbol='Dnajb7', Aliases='4933424H20Rik, Dj5, mDj5', description='DnaJ heat shock protein family (Hsp40) member B7', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 7|dnaJ-like 5', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81407088, end_position_on_the_genomic_accession=81408273, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  57756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57756, CurrentID=0, Status='live', Symbol='Fhl5', Aliases='1700027G07Rik, Act', description='four and a half LIM domains 5', other_designations='FHL-5|LIM-only protein ACT|activator of CREM in testis|activator of cAMP-responsive element modulator in testis', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=25199907, end_position_on_the_genomic_accession=25243044, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  319901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319901, CurrentID=0, Status='live', Symbol='Dsel', Aliases='9330132E09Rik, DS-epi2', description='dermatan sulfate epimerase-like', other_designations='', map_location='1 E2.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=111858701, end_position_on_the_genomic_accession=111864918, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57764, CurrentID=0, Status='live', Symbol='Ntn4', Aliases='', description='netrin 4', other_designations='beta-netrin', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93641049, end_position_on_the_genomic_accession=93745972, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  319909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319909, CurrentID=0, Status='live', Symbol='Ism1', Aliases='5430433G21Rik, Ism', description='isthmin 1, angiogenesis inhibitor', other_designations='isthmin 1 homolog', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=139675007, end_position_on_the_genomic_accession=139758581, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  319915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319915, CurrentID=0, Status='live', Symbol='A830049F12Rik', Aliases='', description='RIKEN cDNA A830049F12 gene', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57773, CurrentID=0, Status='live', Symbol='Wdr4', Aliases='AI415180, AI448349, D530049K22Rik', description='WD repeat domain 4', other_designations='WD repeat-containing protein 4|WD40 repeat protein 4|tRNA (guanine-N(7)-)-methyltransferase subunit WDR4', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31494322, end_position_on_the_genomic_accession=31519992, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  57775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57775, CurrentID=0, Status='live', Symbol='Usp29', Aliases='Ocat', description='ubiquitin specific peptidase 29', other_designations='ASL1/Usp29 fusion|deubiquitinating enzyme 29|ossification center-associated transcript|ubiquitin thioesterase 29|ubiquitin thiolesterase 29|ubiquitin-specific-processing protease 29', map_location='7 4.04 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6730561, end_position_on_the_genomic_accession=6967220, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57776, CurrentID=0, Status='live', Symbol='Ttyh1', Aliases='4930459B04Rik, 6330408P11Rik, TTY1, mTTY1, tty', description='tweety homolog 1 (Drosophila)', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4119527, end_position_on_the_genomic_accession=4136241, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  319922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319922, CurrentID=0, Status='live', Symbol='Vwc2', Aliases='A930041G11Rik, G11, PSST739, UNQ739, cradin', description='von Willebrand factor C domain containing 2', other_designations='CR (chordin-like cysteine-rich) domain-containing adhesive protein|brain-specific chordin-like protein|von Willebrand factor C domain-containing protein 2', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=11113859, end_position_on_the_genomic_accession=11269295, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  57780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57780, CurrentID=0, Status='live', Symbol='Fxyd7', Aliases='1110035I01Rik', description='FXYD domain-containing ion transport regulator 7', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31042515, end_position_on_the_genomic_accession=31051454, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  57781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57781, CurrentID=0, Status='live', Symbol='Cd200r1', Aliases='CD200R, Mox2r, OX2R', description='CD200 receptor 1', other_designations='CD200 cell surface glycoprotein receptor|Orexin receptor 2|antigen identified by monoclonal antibody MRC OX-2 receptor|cell surface glycoprotein OX2 receptor 1', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44765736, end_position_on_the_genomic_accession=44794977, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  57782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57782, CurrentID=0, Status='live', Symbol='Rbak', Aliases='', description='RB-associated KRAB zinc finger', other_designations='RB-associated KRAB repressor|zinc finger protein 769', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143172182, end_position_on_the_genomic_accession=143180842, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  57783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57783, CurrentID=0, Status='live', Symbol='Tnip1', Aliases='ABIN, ABIN-1, ABIN1, AU018810, Naf1, Nef, VAN', description='TNFAIP3 interacting protein 1', other_designations='A20-binding inhibitor of NF-kappa B activation|Nef-associated factor 1|virion-associated nuclear shuttling protein', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54910787, end_position_on_the_genomic_accession=54962958, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  57784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57784, CurrentID=0, Status='live', Symbol='Bin3', Aliases='1700015F07Rik', description='bridging integrator 3', other_designations='bridging integrator-3', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70100090, end_position_on_the_genomic_accession=70138206, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  57785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57785, CurrentID=0, Status='live', Symbol='Rangrf', Aliases='2400006H24Rik, Mog1, Rangnrf', description='RAN guanine nucleotide release factor', other_designations='ran-binding protein MOG1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68972484, end_position_on_the_genomic_accession=68975185, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  319930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319930, CurrentID=0, Status='live', Symbol='Ceacam19', Aliases='C130022P09Rik, CEAL1', description='carcinoembryonic antigen-related cell adhesion molecule 19', other_designations='CEA-related cell adhesion molecule 19|carcinoembryonic antigen-like 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19875742, end_position_on_the_genomic_accession=19887990, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  319934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319934, CurrentID=0, Status='live', Symbol='Sbf2', Aliases='4833411B01Rik, AA414977, AI317167, B430219L04Rik, Mtmr13, mMTMH1', description='SET binding factor 2', other_designations='myotubularin related protein 13|myotubularin-related protein 5', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110308011, end_position_on_the_genomic_accession=110614969, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  319939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319939, CurrentID=0, Status='live', Symbol='Tns3', Aliases='BC023928, F830010I22Rik, TEM6, Tens1', description='tensin 3', other_designations='tensin-like SH2 domain containing 1|tensin-like SH2 domain-containing protein 1|tumor endothelial marker 6', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=8431652, end_position_on_the_genomic_accession=8664735, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  319942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319942, CurrentID=0, Status='live', Symbol='A530016L24Rik', Aliases='Nrac', description='RIKEN cDNA A530016L24 gene', other_designations='transmembrane protein C14orf180 homolog', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112489394, end_position_on_the_genomic_accession=112499928, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  319944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319944, CurrentID=0, Status='live', Symbol='Taf2', Aliases='150kDa, 4732460C16Rik, AI425886, CIF150, TAF2B, TAFII-150, TAFII150', description='TATA-box binding protein associated factor 2', other_designations='TAF(II)150|TAF2 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TBP-associated factor 150 kDa|transcription initiation factor TFIID 150 kDa subunit', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=55015129, end_position_on_the_genomic_accession=55072186, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  319945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319945, CurrentID=0, Status='live', Symbol='Flad1', Aliases='A930017E24Rik, Pp591', description='flavin adenine dinucleotide synthetase 1', other_designations='FAD pyrophosphorylase|FAD synthetase|FAD-synthetase|FMN adenylyltransferase|RFad1, flavin adenine dinucleotide synthetase, homolog|flavin adenine dinucleotide synthase|flavin adenine dinucleotide synthetase', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89402673, end_position_on_the_genomic_accession=89411863, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  57808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57808, CurrentID=0, Status='live', Symbol='Rpl35a', Aliases='2810431L15Rik, Rpl35', description='ribosomal protein L35A', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33056453, end_position_on_the_genomic_accession=33060189, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  319953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319953, CurrentID=0, Status='live', Symbol='Ttll1', Aliases='6330444E16Rik, AV014541', description='tubulin tyrosine ligase-like 1', other_designations='PGs3|bM121M7.1 (novel protein (ortholog of human tubulin tyrosine ligase-like 1 (TTLL1)), isoform 1)|p49|tubulin polyglutamylase complex subunit 3|tubulin--tyrosine ligase-like protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83483769, end_position_on_the_genomic_accession=83518917, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  57810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57810, CurrentID=0, Status='live', Symbol='Cdon', Aliases='CDO, Cd0, Orcam', description='cell adhesion molecule-related/down-regulated by oncogenes', other_designations='CAM-related/down-regulated by oncogenes|immunoglobulin superfamily member CDO|oncogene-regulated cell adhesion molecule', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35421129, end_position_on_the_genomic_accession=35507652, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  57811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57811, CurrentID=0, Status='live', Symbol='Rgr', Aliases='', description='retinal G protein coupled receptor', other_designations='RGR opsin', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=37034909, end_position_on_the_genomic_accession=37049014, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  57813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57813, CurrentID=0, Status='live', Symbol='Tk2', Aliases='AU024611, mt-TK', description='thymidine kinase 2, mitochondrial', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104226691, end_position_on_the_genomic_accession=104248574, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  57814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57814, CurrentID=0, Status='live', Symbol='Kcne4', Aliases='AW487937, MiRP3', description='potassium voltage-gated channel, Isk-related subfamily, gene 4', other_designations='minK-related peptide 3|minimum potassium ion channel-related peptide 3|potassium channel subunit beta MiRP3', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=78816949, end_position_on_the_genomic_accession=78820025, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  57815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57815, CurrentID=0, Status='live', Symbol='Spata5', Aliases='2510048F20Rik, C78064, Spaf', description='spermatogenesis associated 5', other_designations='spermatogenesis-associated factor protein', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37420280, end_position_on_the_genomic_accession=37579096, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  57816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57816, CurrentID=0, Status='live', Symbol='Tesc', Aliases='1010001A17Rik, 2410011K10Rik, TE-1', description='tescalcin', other_designations='CHP3|TSC', map_location='5 57.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=118027748, end_position_on_the_genomic_accession=118061878, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  319965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319965, CurrentID=0, Status='live', Symbol='Cc2d1b', Aliases='A830039B04Rik, Freud2', description='coiled-coil and C2 domain containing 1B', other_designations='FRE under dual repression-binding protein 2|five prime repressor element under dual repression-binding protein 2|freud-2', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108619767, end_position_on_the_genomic_accession=108634123, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  319974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319974, CurrentID=0, Status='live', Symbol='Auts2', Aliases='2700063G02Rik, A730011F23Rik, D830032G16Rik', description='autism susceptibility candidate 2', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=131437682, end_position_on_the_genomic_accession=132542343, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  57837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57837, CurrentID=0, Status='live', Symbol='Eral1', Aliases='2610524P08Rik, 9130407C09Rik, AU019798, Era, M-ERA, MERA-S, MERA-W', description='Era (G-protein)-like 1 (E. coli)', other_designations='CEGA|ERA-W|ERA-like GTPase|ERA-like protein 1|GTP-binding protein era homolog|conserved ERA-like GTPase', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78073376, end_position_on_the_genomic_accession=78080383, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  319984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319984, CurrentID=0, Status='live', Symbol='Jph4', Aliases='9330157P13Rik, JP-4, JPHL1', description='junctophilin 4', other_designations='junctophilin-like 1 protein', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55106826, end_position_on_the_genomic_accession=55116935, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  319991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319991, CurrentID=0, Status='live', Symbol='Kif6', Aliases='D130004B10Rik, D130084M03Rik', description='kinesin family member 6', other_designations='kinesin 6', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=49615104, end_position_on_the_genomic_accession=49909847, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  319996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319996, CurrentID=0, Status='live', Symbol='Casc4', Aliases='D130060C09Rik, VGFG2573', description='cancer susceptibility candidate 4', other_designations='cancer susceptibility candidate gene 4 protein homolog', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121866970, end_position_on_the_genomic_accession=121936220, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  319997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319997, CurrentID=0, Status='live', Symbol='A630001G21Rik', Aliases='', description='RIKEN cDNA A630001G21 gene', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85717083, end_position_on_the_genomic_accession=85736656, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  319998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319998, CurrentID=0, Status='live', Symbol='Tmem198', Aliases='A230078I05Rik-1, Tmem198a, Tmem198', description='transmembrane protein 198', other_designations='', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75479532, end_position_on_the_genomic_accession=75485693, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  320007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320007, CurrentID=0, Status='live', Symbol='Sidt1', Aliases='AW045928, B830021E24Rik', description='SID1 transmembrane family, member 1', other_designations='', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44240177, end_position_on_the_genomic_accession=44333328, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  320011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320011, CurrentID=0, Status='live', Symbol='Uggt1', Aliases='0910001L17Rik, A930007H10Rik, AA589501, AI414429, AI448372, C820010P03Rik, GT, UGT1, Ugcgl1', description='UDP-glucose glycoprotein glucosyltransferase 1', other_designations='UDP--Glc:glycoprotein glucosyltransferase|UDP-glucose ceramide glucosyltransferase-like 1|UDP-glucose:glycoprotein glucosyltransferase', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36140028, end_position_on_the_genomic_accession=36244302, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  57869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57869, CurrentID=0, Status='live', Symbol='Adck2', Aliases='AA408112, BB006057', description='aarF domain containing kinase 2', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39573835, end_position_on_the_genomic_accession=39588769, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  57874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57874, CurrentID=0, Status='live', Symbol='Hacd3', Aliases='4930523M17Rik, AW742319, B-ind1, Hcad3, Hspc121, Ptplad1', description='3-hydroxyacyl-CoA dehydratase 3', other_designations='butyrate-induced protein 1|butyrate-induced transcript 1|protein tyrosine phosphatase-like A domain containing 1|protein tyrosine phosphatase-like protein PTPLAD1|protein-tyrosine phosphatase-like A domain-containing protein 1|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 3', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64986983, end_position_on_the_genomic_accession=65021735, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  57875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57875, CurrentID=0, Status='live', Symbol='Angptl4', Aliases='Arp4, Bk89, Fiaf, Hfarp, Ng27, Pgar, Pgarg, Pp1158', description='angiopoietin-like 4', other_designations='425O18-1|angiopoietin-like protein 4|fasting-induced adipose factor|fibrinogen/angiopoietin-related protein|hepatic fibrinogen/angiopoietin-related protein|major histocompatibility complex region NG27|secreted protein Bk89', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33774900, end_position_on_the_genomic_accession=33781575, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  320022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320022, CurrentID=0, Status='live', Symbol='Ccdc79', Aliases='4930532D21Rik, BB085179, Terb1', description='coiled-coil domain containing 79', other_designations='coiled-coil domain-containing protein 79|telomere repeat binding bouquet formation protein 1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104446718, end_position_on_the_genomic_accession=104509887, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  320024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320024, CurrentID=0, Status='live', Symbol='Nceh1', Aliases='Aadacl1, B230106I24Rik, CPO-BP, Nceh', description='neutral cholesterol ester hydrolase 1', other_designations='arylacetamide deacetylase-like 1|chlorpyrifos oxon-binding protein', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=27183004, end_position_on_the_genomic_accession=27245266, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320027, CurrentID=0, Status='live', Symbol='Fstl4', Aliases='B230374F23Rik, SPIG1, mKIAA1061', description='follistatin-like 4', other_designations='follistatin-like protein 4|m-D/Bsp120I 1-1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52764044, end_position_on_the_genomic_accession=53192450, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  57890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57890, CurrentID=0, Status='live', Symbol='Il17re', Aliases='AA589509, IL-17RE, Il25r', description='interleukin 17 receptor E', other_designations='IL-17 receptor E', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113458484, end_position_on_the_genomic_accession=113470758, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  320038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320038, CurrentID=0, Status='live', Symbol='Dlx6os1', Aliases='A230055N17Rik, Dlx6as1, Evf-1, Evf-2, Evf1, Evf1/2, Evf2, Shhrs, mEvf-1', description='distal-less homeobox 6, opposite strand 1', other_designations='Dlx6 antisense RNA 1|Dlx6 opposite strand RNA 1|Dlx6 opposite strand transcript 1|sonic hedgehog regulated sequence', map_location='6 2.68 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=6820546, end_position_on_the_genomic_accession=6869533, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  57895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57895, CurrentID=0, Status='live', Symbol='Ccdc126', Aliases='6330407D12Rik, AA675040, AI465320, AL022711', description='coiled-coil domain containing 126', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49319268, end_position_on_the_genomic_accession=49341589, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  320040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320040, CurrentID=0, Status='live', Symbol='Rnf222', Aliases='9930039A11Rik', description='ring finger protein 222', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68881964, end_position_on_the_genomic_accession=68895779, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  57905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57905, CurrentID=0, Status='live', Symbol='Isy1', Aliases='5830446M03Rik, AI181014, AU020769', description='ISY1 splicing factor homolog', other_designations='', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87818447, end_position_on_the_genomic_accession=87838759, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  320051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320051, CurrentID=0, Status='live', Symbol='Exph5', Aliases='AC079869.22gm5, B130009M24Rik, E030050P12, Kiaa0624, Slac2b, slac2-b', description='exophilin 5', other_designations='exph5|slaC2-b|slp homolog lacking C2 domains b|synaptotagmin-like homologue lacking C2 domains b|synaptotagmin-like protein homolog lacking C2 domains b', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53301519, end_position_on_the_genomic_accession=53381166, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  57908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57908, CurrentID=0, Status='live', Symbol='Zfp318', Aliases='2610034E08Rik, D530032D06Rik, TZF, Znf318', description='zinc finger protein 318', other_designations='testicular zinc finger protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46383503, end_position_on_the_genomic_accession=46420920, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  57911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57911, CurrentID=0, Status='live', Symbol='Gsdma', Aliases='BB149167, Gsdm, Gsdm11, H312E, Gsdma', description='gasdermin A', other_designations='gasdermin 1|gasdermin A1|gasdermin-1|gasdermin-A1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98664337, end_position_on_the_genomic_accession=98677708, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  57912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57912, CurrentID=0, Status='live', Symbol='Cdc42se1', Aliases='1300002M12Rik, AW558204, Cdcse1, SCIP1, Spec1', description='CDC42 small effector 1', other_designations='small protein effector 1 of Cdc42', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95228424, end_position_on_the_genomic_accession=95236424, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  57913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57913, CurrentID=0, Status='live', Symbol='Pidd1', Aliases='1200011D09Rik, AU042446, Lrdd, Pidd', description='p53 induced death domain protein 1', other_designations='leucine-rich and death domain containing|leucine-rich repeat and death domain-containing protein|p53 protein induced, with death domain|p53-induced protein with a death domain', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141438113, end_position_on_the_genomic_accession=141444408, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  57914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57914, CurrentID=0, Status='live', Symbol='Crlf2', Aliases='CRLM2, Ly114, Tpte2, Tslpr', description='cytokine receptor-like factor 2', other_designations='TSLP receptor|cytokine receptor-like molecule 2|lymphocyte antigen 114|thymic stromal lymphopoietin protein receptor|thymic stromal-derived lymphopoietin, receptor|transmembrane phosphoinositide 3-phosphatase and tensin homolog 2|type I cytokine receptor delta 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=109554709, end_position_on_the_genomic_accession=109559017, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57915, CurrentID=0, Status='live', Symbol='Tbc1d1', Aliases='1110062G02Rik, AI385682, AW555803, Nob1, Nobq1, Tbc1, mKIAA1108', description='TBC1 domain family, member 1', other_designations='New Zealand obese QTL 1|tre-2/USP6, BUB2, cdc16) domain family, member 1', map_location='5 32.8 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=64156310, end_position_on_the_genomic_accession=64351486, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  57916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57916, CurrentID=0, Status='live', Symbol='Tnfrsf13b', Aliases='1200009E08Rik, Taci', description='tumor necrosis factor receptor superfamily, member 13b', other_designations='transmembrane activator and CAML interactor|transmembrane activator and calcium-modulating and cyclophilin ligand interactor', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61126742, end_position_on_the_genomic_accession=61149373, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320078, CurrentID=0, Status='live', Symbol='Olfml2b', Aliases='1110018N05Rik, 4832415H08Rik, AI467542', description='olfactomedin-like 2B', other_designations='photomedin-2', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170644423, end_position_on_the_genomic_accession=170682789, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  320080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320080, CurrentID=0, Status='live', Symbol='Zbtb39', Aliases='7030401O21Rik, mKIAA0352', description='zinc finger and BTB domain containing 39', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127739538, end_position_on_the_genomic_accession=127747349, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  320082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320082, CurrentID=0, Status='live', Symbol='Fbxw21', Aliases='E330009P21Rik', description='F-box and WD-40 domain protein 21', other_designations='novel gene (E330009P21Rik)', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109136887, end_position_on_the_genomic_accession=109162022, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  320083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320083, CurrentID=0, Status='live', Symbol='Fbxw16', Aliases='7420402K12Rik', description='F-box and WD-40 domain protein 16', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109432318, end_position_on_the_genomic_accession=109449140, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  320091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320091, CurrentID=0, Status='live', Symbol='Ano4', Aliases='A330096O15Rik, Gm65, Tmem16d', description='anoctamin 4', other_designations='ano4|transmembrane protein 16D (eight membrane-spanning domains)', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88948935, end_position_on_the_genomic_accession=89345087, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  320095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320095, CurrentID=0, Status='live', Symbol='6430550D23Rik', Aliases='', description='RIKEN cDNA 6430550D23 gene', other_designations='likely orthologue of H. sapiens chromosome 20 open reading frame 173 (C20orf173)', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155982970, end_position_on_the_genomic_accession=156007977, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  320100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320100, CurrentID=0, Status='live', Symbol='Relt', Aliases='E430021K24Rik, Tnfrsf19l', description='RELT tumor necrosis factor receptor', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100845847, end_position_on_the_genomic_accession=100863480, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  320106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320106, CurrentID=0, Status='live', Symbol='Slc38a11', Aliases='9330158F14Rik', description='solute carrier family 38, member 11', other_designations='solute carrier family 38 member 11', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=65315799, end_position_on_the_genomic_accession=65364394, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  320111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320111, CurrentID=0, Status='live', Symbol='Prr18', Aliases='9630019K15Rik', description='proline rich 18', other_designations='proline rich region 18', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8340406, end_position_on_the_genomic_accession=8344113, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  320116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320116, CurrentID=0, Status='live', Symbol='Fndc9', Aliases='C030019I05Rik', description='fibronectin type III domain containing 9', other_designations='fibronectin type-III domain-containing protein C5orf40 homolog', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46235557, end_position_on_the_genomic_accession=46239871, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  320118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320118, CurrentID=0, Status='live', Symbol='Fbxl13', Aliases='4921539K22Rik', description='F-box and leucine-rich repeat protein 13', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21483847, end_position_on_the_genomic_accession=21645698, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  320119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320119, CurrentID=0, Status='live', Symbol='Rps6kc1', Aliases='AA682037, B130003F20Rik, C80612, Rpk118', description='ribosomal protein S6 kinase polypeptide 1', other_designations='52 kDa ribosomal protein S6 kinase|S6K-delta-1|ribosomal protein S6 kinase, 52kDa, polypeptide 1', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=190772879, end_position_on_the_genomic_accession=190912139, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  320127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320127, CurrentID=0, Status='live', Symbol='Dgki', Aliases='C130010K08Rik', description='diacylglycerol kinase, iota', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=36837381, end_position_on_the_genomic_accession=37300060, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  407831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407831, CurrentID=0, Status='live', Symbol='Tmem204', Aliases='', description='transmembrane protein 204', other_designations='claudin-like protein 24', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25057702, end_position_on_the_genomic_accession=25083553, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  320129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320129, CurrentID=0, Status='live', Symbol='Adrbk2', Aliases='4833444A01Rik, AI851927, AW551196, Adrbk-2, Bark-2, GRK3', description='adrenergic receptor kinase, beta 2', other_designations='beta ARK2|beta-ARK-2', map_location='5 55.29 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112910478, end_position_on_the_genomic_accession=113015951, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  320135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320135, CurrentID=0, Status='live', Symbol='BC049715', Aliases='', description='cDNA sequence BC049715', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136828160, end_position_on_the_genomic_accession=136841920, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320139, CurrentID=0, Status='live', Symbol='Ptpn7', Aliases='BPTP-4, C920001D21Rik, HEPTP, LC-PTP, LPTP', description='protein tyrosine phosphatase, non-receptor type 7', other_designations='hematopoietic protein-tyrosine phosphatase', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135131840, end_position_on_the_genomic_accession=135145328, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  320145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320145, CurrentID=0, Status='live', Symbol='Sp8', Aliases='D930049B17Rik, mBtd', description='trans-acting transcription factor 8', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=118845840, end_position_on_the_genomic_accession=118852578, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320148, CurrentID=0, Status='live', Symbol='B430306N03Rik', Aliases='Treml6', description='RIKEN cDNA B430306N03 gene', other_designations='triggering receptor expressed on myeloid cells-like 6|triggering receptor expressed on myeloid cells-like 6b', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48315268, end_position_on_the_genomic_accession=48326511, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320150, CurrentID=0, Status='live', Symbol='Zdhhc17', Aliases='A230053P19Rik, BB187739, D130071N24Rik, Hip14', description='zinc finger, DHHC domain containing 17', other_designations='DHHC-17|huntingtin interacting protein 14|huntingtin-interacting protein 14|zinc finger DHHC domain-containing protein 17', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=110941779, end_position_on_the_genomic_accession=111010066, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  320158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320158, CurrentID=0, Status='live', Symbol='Zmat4', Aliases='9630048M01Rik', description='zinc finger, matrin type 4', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=23669547, end_position_on_the_genomic_accession=24063117, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  320159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320159, CurrentID=0, Status='live', Symbol='Fam179a', Aliases='4632412N22Rik', description='family with sequence similarity 179, member A', other_designations='', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71673157, end_position_on_the_genomic_accession=71729669, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  320162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320162, CurrentID=0, Status='live', Symbol='Cep95', Aliases='4732496G21Rik, AI118346, AI448335, Ccdc45, D330027A14Rik, F630025I20Rik', description='centrosomal protein 95', other_designations='coiled-coil domain containing 45|coiled-coil domain-containing protein 45', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106789252, end_position_on_the_genomic_accession=106818861, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  320165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320165, CurrentID=0, Status='live', Symbol='Tacc1', Aliases='4833447E04Rik, AA960202, B230378H13Rik', description='transforming, acidic coiled-coil containing protein 1', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25154552, end_position_on_the_genomic_accession=25241585, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  320181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320181, CurrentID=0, Status='live', Symbol='Fndc7', Aliases='E230011A21Rik', description='fibronectin type III domain containing 7', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108853678, end_position_on_the_genomic_accession=108890008, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  320183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320183, CurrentID=0, Status='live', Symbol='Msrb3', Aliases='D430026P16Rik, I-8-23, MsrB2', description='methionine sulfoxide reductase B3', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=120781098, end_position_on_the_genomic_accession=120899304, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  320184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320184, CurrentID=0, Status='live', Symbol='Lrrc58', Aliases='1810012N18Rik, C330018J07Rik', description='leucine rich repeat containing 58', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37868400, end_position_on_the_genomic_accession=37888857, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  320191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320191, CurrentID=0, Status='live', Symbol='Hook3', Aliases='5830454D03Rik, AI317159, E330005F07Rik', description='hook microtubule tethering protein 3', other_designations='hook homolog 3|mHK3', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=26021421, end_position_on_the_genomic_accession=26119224, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  320202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320202, CurrentID=0, Status='live', Symbol='Lefty2', Aliases='6030463A22Rik, Ebaf, Lefta, Leftb, lefty-2, lefty-B', description='left-right determination factor 2', other_designations='endometrial bleeding associated factor|left-right determination factor B|left-right determination, factor A', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180893119, end_position_on_the_genomic_accession=180899103, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  320203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320203, CurrentID=0, Status='live', Symbol='C130071C03Rik', Aliases='2810449C10Rik', description='RIKEN cDNA C130071C03 gene', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=83728106, end_position_on_the_genomic_accession=83735669, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  320204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320204, CurrentID=0, Status='live', Symbol='Mettl20', Aliases='4833442J19Rik, AI256744', description='methyltransferase like 20', other_designations='methyltransferase-like protein 20', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=149141513, end_position_on_the_genomic_accession=149151171, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  320207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320207, CurrentID=0, Status='live', Symbol='Pik3r5', Aliases='AV230647, F730038I15Rik', description='phosphoinositide-3-kinase, regulatory subunit 5, p101', other_designations='PI3-kinase p101 subunit|PI3-kinase regulatory subunit 5|p101-PI3K|phosphatidylinositol 4,5-bisphosphate 3-kinase regulatory subunit|phosphatidylinositol-4,5-bisphosphate 3-kinase regulatory subunit|ptdIns-3-kinase p101|ptdIns-3-kinase regulatory subunit', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68432125, end_position_on_the_genomic_accession=68497849, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  320208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320208, CurrentID=0, Status='live', Symbol='Tmem91', Aliases='A830041P22Rik, DSPC3', description='transmembrane protein 91', other_designations='dispanin subfamily C member 3|synapse differentiation-induced protein 3', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25669139, end_position_on_the_genomic_accession=25679815, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  320209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320209, CurrentID=0, Status='live', Symbol='Ddx11', Aliases='4732462I11Rik, CHL1, CHLR1, KRG2, essa15a', description='DEAD/H (Asp-Glu-Ala-Asp/His) box helicase 11', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 11 (CHL1-like helicase homolog, S. cerevisiae)|DEAD/H box protein 11|probable ATP-dependent RNA helicase DDX11', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=66123501, end_position_on_the_genomic_accession=66152174, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  320213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320213, CurrentID=0, Status='live', Symbol='Senp5', Aliases='6230429P13Rik, A730063F07Rik, AI851888, BB189556, SMT3IP3', description='SUMO/sentrin specific peptidase 5', other_designations='SUMO/Smt3-specific isopeptidase 3|SUMO/sentrin specific protease 5|sentrin/SUMO-specific protease SENP5', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31959670, end_position_on_the_genomic_accession=32010740, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  320214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320214, CurrentID=0, Status='live', Symbol='Maats1', Aliases='4932425I24Rik, Aat1, Spata26', description='MYCBP-associated, testis expressed 1', other_designations='AAT-1|AMY-1-associating protein expressed in testis 1 homolog|MYCBP/AMY-1-associated testis-expressed protein 1|spermatogenesis associated 26', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38297754, end_position_on_the_genomic_accession=38342253, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  320225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320225, CurrentID=0, Status='live', Symbol='Catsperg1', Aliases='A230107C01Rik, Catsperg', description='catsper channel auxiliary subunit gamma 1', other_designations='cation channel, sperm-associated, gamma 1', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29181532, end_position_on_the_genomic_accession=29215631, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  320226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320226, CurrentID=0, Status='live', Symbol='Ccdc171', Aliases='4930418J05Rik, 4930473A06Rik, A330015D16Rik', description='coiled-coil domain containing 171', other_designations='', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=83525537, end_position_on_the_genomic_accession=83866143, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  320234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320234, CurrentID=0, Status='live', Symbol='Ccdc66', Aliases='E230015L20Rik', description='coiled-coil domain containing 66', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=27481090, end_position_on_the_genomic_accession=27508506, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  97064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97064, CurrentID=0, Status='live', Symbol='Wwtr1', Aliases='2310058J06Rik, 2610021I22Rik, C78399, Taz', description='WW domain containing transcription regulator 1', other_designations='transcriptional coactivator with PDZ binding motif|transcriptional coactivator with PDZ-binding motif', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=57455644, end_position_on_the_genomic_accession=57575910, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  320244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320244, CurrentID=0, Status='live', Symbol='Ttll5', Aliases='1700048H13Rik, 2310009M18Rik, 4930556H18Rik, AI428864, D630041K24Rik, R75373, Stamp, mKIAA0998', description='tubulin tyrosine ligase-like family, member 5', other_designations='polyglutamylase|tubulin tyrosine ligase like family 5|tubulin--tyrosine ligase-like protein 5', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85824950, end_position_on_the_genomic_accession=86053760, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  320253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320253, CurrentID=0, Status='live', Symbol='March3', Aliases='6330411I15Rik, A530081L18Rik, AI843639, MARCH-III', description='membrane-associated ring finger (C3HC4) 3', other_designations='membrane-associated RING finger protein 3|membrane-associated RING-CH protein III', map_location='18 D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=56761716, end_position_on_the_genomic_accession=56925652, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  320256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320256, CurrentID=0, Status='live', Symbol='Dlec1', Aliases='D630005C06Rik', description='deleted in lung and esophageal cancer 1', other_designations='dlec1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119101535, end_position_on_the_genomic_accession=119148246, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  320265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320265, CurrentID=0, Status='live', Symbol='Fam19a1', Aliases='C630007B19Rik, FAM19A1 TAFA-1, TAFA1, Tafa-1', description='family with sequence similarity 19, member A1', other_designations='chemokine-like protein TAFA-1', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=96112775, end_position_on_the_genomic_accession=96657204, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320267, CurrentID=0, Status='live', Symbol='Fubp3', Aliases='A330051M14Rik, AV006371, FBP3, Marta2', description='far upstream element (FUSE) binding protein 3', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31572651, end_position_on_the_genomic_accession=31617526, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  320271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320271, CurrentID=0, Status='live', Symbol='Scai', Aliases='9330112M16, A930041I02Rik, AI662729', description='suppressor of cancer cell invasion', other_designations='suppressor of cancer cell invasion protein', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=39066214, end_position_on_the_genomic_accession=39190730, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  320277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320277, CurrentID=0, Status='live', Symbol='Spef2', Aliases='C230086A09Rik', description='sperm flagellar 2', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9578193, end_position_on_the_genomic_accession=9749044, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  320292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320292, CurrentID=0, Status='live', Symbol='Rasgef1b', Aliases='4732452O09Rik, GPIgamma4, Gpig4', description='RasGEF domain family, member 1B', other_designations='GPI gamma-4|GPI-gamma 4', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=99217420, end_position_on_the_genomic_accession=99252927, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  320299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320299, CurrentID=0, Status='live', Symbol='Iqcb1', Aliases='6820449I09Rik, AV128382, NPHP5', description='IQ calmodulin-binding motif containing 1', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36828363, end_position_on_the_genomic_accession=36872719, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  320302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320302, CurrentID=0, Status='live', Symbol='Glt28d2', Aliases='4732486J07Rik, Glt28d1l', description='glycosyltransferase 28 domain containing 2', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=85869846, end_position_on_the_genomic_accession=85887518, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  320311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320311, CurrentID=0, Status='live', Symbol='Rnf152', Aliases='A930029B02Rik', description='ring finger protein 152', other_designations='', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=105276917, end_position_on_the_genomic_accession=105356710, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  58170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58170, CurrentID=0, Status='live', Symbol='Asic5', Aliases='Accn5, Blinac, Inac', description='acid-sensing (proton-gated) ion channel family member 5', other_designations='amiloride-sensitive Na+ channel|amiloride-sensitive cation channel 5, intestinal|amiloride-sensitive sodium channel|brain-liver-intestine amiloride-sensitive Na(+) channel|brain-liver-intestine amiloride-sensitive sodium channel', map_location='3 E-F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=81982293, end_position_on_the_genomic_accession=82021233, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  58172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58172, CurrentID=0, Status='live', Symbol='Sertad2', Aliases='AB041541, AU021878, Kiaa0127, MNCb-1504, SEI-2, Sei2, TRIP-Br2, mKIAA0127', description='SERTA domain containing 2', other_designations='PHD zinc finger- and bromodomain-interacting protein 2|transcriptional regulator interacting with the PHD-bromodomain 2', map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=20542938, end_position_on_the_genomic_accession=20653023, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  58175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58175, CurrentID=0, Status='live', Symbol='Rgs20', Aliases='2900073E09Rik, Rgsz1', description='regulator of G-protein signaling 20', other_designations='regulator of G-protein signaling Z1', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=4909573, end_position_on_the_genomic_accession=5070308, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  58176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58176, CurrentID=0, Status='live', Symbol='Rhbg', Aliases='', description='Rhesus blood group-associated B glycoprotein', other_designations='Rh type B glycoprotein|rh family type B glycoprotein|rhesus blood group family type B glycoprotein', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88242874, end_position_on_the_genomic_accession=88254743, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  58178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58178, CurrentID=0, Status='live', Symbol='Sorcs1', Aliases='Sorcs, mSorCS', description='sortilin-related VPS10 domain containing receptor 1', other_designations='VPS10 domain receptor protein SORCS 1', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=50143301, end_position_on_the_genomic_accession=50679136, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  58179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58179, CurrentID=0, Status='live', Symbol='Klrc3', Aliases='Klrc2, Nkg2e', description='killer cell lectin-like receptor subfamily C, member 3', other_designations='killer cell lectin-like receptor subfamily C, member 2|natural killer cell group 2E cell receptor', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129636263, end_position_on_the_genomic_accession=129643511, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  58180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58180, CurrentID=0, Status='live', Symbol='Hic2', Aliases='AA409108, HRG22, mKIAA1020', description='hypermethylated in cancer 2', other_designations='hic-2', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17233587, end_position_on_the_genomic_accession=17263430, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  58181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58181, CurrentID=0, Status='live', Symbol='Il20', Aliases='Zcyto10', description='interleukin 20', other_designations='IL-20|cytokine Zcyto10|four alpha helix cytokine ZCYTO10|four helix bundle cytokine 10', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130907192, end_position_on_the_genomic_accession=130911296, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  58182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58182, CurrentID=0, Status='live', Symbol='Prokr1', Aliases='EG-VEGFR1, Gpr73, Pkr1', description='prokineticin receptor 1', other_designations='G protein-coupled receptor 73|G-protein coupled receptor 73|PK-R1', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87578589, end_position_on_the_genomic_accession=87590774, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  58184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58184, CurrentID=0, Status='live', Symbol='Rqcd1', Aliases='2610007F23Rik, AI593551, Fl10', description='rcd1 (required for cell differentiation) homolog 1 (S. pombe)', other_designations='CCR4-NOT transcription complex subunit 9|EPO-induced protein FL10|rcd-1', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74506040, end_position_on_the_genomic_accession=74530842, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  58185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58185, CurrentID=0, Status='live', Symbol='Rsad2', Aliases='2510004L01Rik, Vig1, cig5', description='radical S-adenosyl methionine domain containing 2', other_designations='VHSV-induced-like protein|rsad2|viperin|viral hemorrhagic septicemia virus (VHSV)-induced gene 1|virus inhibitory protein, endoplasmic reticulum-associated, interferon-inducible', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=26442743, end_position_on_the_genomic_accession=26456478, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  58186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58186, CurrentID=0, Status='live', Symbol='Rad18', Aliases='2810024C04Riksc, Rad18', description='RAD18 E3 ubiquitin protein ligase', other_designations='RAD18 homolog|mRAD18Sc|post-replication repair protein RAD18SC|postreplication repair protein RAD18', map_location='6 F', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=112619851, end_position_on_the_genomic_accession=112696689, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  58187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58187, CurrentID=0, Status='live', Symbol='Cldn10', Aliases='6720456I16Rika, Cldn10b, D14Ertd728e, Cldn10', description='claudin 10', other_designations='claudin 10A|claudin 10B|claudin-10A', map_location='14 62.55 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118787871, end_position_on_the_genomic_accession=118874525, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  58188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58188, CurrentID=0, Status='live', Symbol='Vstm2b', Aliases='2900093B09Rik, AB030198', description='V-set and transmembrane domain containing 2B', other_designations='', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=40898297, end_position_on_the_genomic_accession=40930044, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  320333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320333, CurrentID=0, Status='live', Symbol='D830030K20Rik', Aliases='', description='RIKEN cDNA D830030K20 gene', other_designations='alpha30-takusan', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=3208129, end_position_on_the_genomic_accession=3234346, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  58193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58193, CurrentID=0, Status='live', Symbol='Extl2', Aliases='3000001D04Rik, AW146439', description='exostoses (multiple)-like 2', other_designations='EXT-related protein 2|alpha-1,4-N-acetylhexosaminyltransferase EXTL2|alpha-GalNAcT EXTL2|exostoses-like 2|exotoses (multiple)-like 2|glucuronyl-galactosyl-proteoglycan 4-alpha-N-acetylglucosaminyltransferase', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116007446, end_position_on_the_genomic_accession=116029017, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  58194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58194, CurrentID=0, Status='live', Symbol='Sh3kbp1', Aliases='1200007H22Rik, 1700125L08Rik, 5830464D22Rik, AI447724, IN85, Ruk, Seta', description='SH3-domain kinase binding protein 1', other_designations='SH3-containing, expressed in tumorigenic astrocytes|regulator of ubiquitous kinase', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=159627029, end_position_on_the_genomic_accession=159975920, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  58198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58198, CurrentID=0, Status='live', Symbol='Sall1', Aliases='Msal-3', description='sal-like 1 (Drosophila)', other_designations='sal-3|zinc finger protein Spalt-3', map_location='8 D', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=89027239, end_position_on_the_genomic_accession=89044162, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  320343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320343, CurrentID=0, Status='live', Symbol='Lypd6', Aliases='E130115E03Rik', description='LY6/PLAUR domain containing 6', other_designations='', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=50066429, end_position_on_the_genomic_accession=50193569, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  58200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58200, CurrentID=0, Status='live', Symbol='Ppp1r1a', Aliases='0610038N18Rik, I-1', description='protein phosphatase 1, regulatory (inhibitor) subunit 1A', other_designations='IPP-1|inhibitor-1|protein phosphatase inhibitor 1|protein phosphatase inhibitor-1', map_location='15 F', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103530279, end_position_on_the_genomic_accession=103538005, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  58202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58202, CurrentID=0, Status='live', Symbol='Nelfb', Aliases='A730008L03Rik, AB041607, AI663983, Cobra1, Nelf-b', description='negative elongation factor complex member B', other_designations='cofactor of BRCA1|negative elongation factor complex member B, Cobra1|ortholog of human cofactor of BRCA1 (COBRA1)', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25199712, end_position_on_the_genomic_accession=25211489, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  58203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58203, CurrentID=0, Status='live', Symbol='Zbp1', Aliases='2010010H03Rik, Dai, Dlm1, mZaDLM', description='Z-DNA binding protein 1', other_designations='DNA-dependent activator of IFN-regulatory factors|tumor stroma and activated macrophage protein DLM-1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173206612, end_position_on_the_genomic_accession=173218922, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  58205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58205, CurrentID=0, Status='live', Symbol='Pdcd1lg2', Aliases='B7-DC, Btdc, F730015O22Rik, PD-L2', description='programmed cell death 1 ligand 2', other_designations='PD-1 ligand 2|PDCD1 ligand 2|butyrophilin B7-DC|butyrophilin-like protein|programmed death ligand 2', map_location='19 C2-C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29406240, end_position_on_the_genomic_accession=29472927, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  58206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58206, CurrentID=0, Status='live', Symbol='Zbtb32', Aliases='4930524C15Rik, FAXF, FAZF, PLZP, Rog, Tzfp', description='zinc finger and BTB domain containing 32', other_designations='GATA repressor|repressor of GATA|testis zinc finger protein', map_location='7 B2-B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30589680, end_position_on_the_genomic_accession=30602818, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  320351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320351, CurrentID=0, Status='live', Symbol='Tmem251', Aliases='D230037D09Rik', description='transmembrane protein 251', other_designations='UPF0694 transmembrane protein C14orf109 homolog', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102743687, end_position_on_the_genomic_accession=102745406, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  58208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58208, CurrentID=0, Status='live', Symbol='Bcl11b', Aliases='9130430L19Rik, AI604821, B630002E05Rik, BCL-11B, Ctip2, Rit1', description='B cell leukemia/lymphoma 11B', other_designations='B-cell CLL/lymphoma 11B|B-cell leukemia/lymphoma 11B|B-cell lymphoma/leukaemia 11B|COUP-TF interacting protein 2|COUP-TF-interacting protein 2|mRit1|radiation-induced tumor suppressor gene 1 protein', map_location='12 59.1 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=107910403, end_position_on_the_genomic_accession=108004139, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  58210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58210, CurrentID=0, Status='live', Symbol='Sectm1b', Aliases='1810003C24Rik, K12, Sectm1', description='secreted and transmembrane 1B', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121053423, end_position_on_the_genomic_accession=121063569, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  320355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320355, CurrentID=0, Status='live', Symbol='Lipi', Aliases='D930038D03Rik, lpd1', description='lipase, member I', other_designations='', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=75540514, end_position_on_the_genomic_accession=75586072, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  58212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58212, CurrentID=0, Status='live', Symbol='Srrm3', Aliases='2900083I11Rik, AB041661, SRm300-like, Srrm2l', description='serine/arginine repetitive matrix 3', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135806861, end_position_on_the_genomic_accession=135874773, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  58214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58214, CurrentID=0, Status='live', Symbol='Cst10', Aliases='DD72', description='cystatin 10 (chondrocytes)', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=149405249, end_position_on_the_genomic_accession=149410278, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  320360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320360, CurrentID=0, Status='live', Symbol='Ric3', Aliases='E130307J04Rik', description='resistance to inhibitors of cholinesterase 3 homolog (C. elegans)', other_designations='resistant to inhibitor of cholinesterase 3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109034319, end_position_on_the_genomic_accession=109083389, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  58217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58217, CurrentID=0, Status='live', Symbol='Trem1', Aliases='', description='triggering receptor expressed on myeloid cells 1', other_designations='triggering receptor expressed in monocytes 1', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48232367, end_position_on_the_genomic_accession=48246924, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  58218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58218, CurrentID=0, Status='live', Symbol='Trem3', Aliases='BB134760', description='triggering receptor expressed on myeloid cells 3', other_designations='triggering receptor expressed on monocytes 3', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48247729, end_position_on_the_genomic_accession=48258848, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  58220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58220, CurrentID=0, Status='live', Symbol='Pard6b', Aliases='AV025615, Par6b', description='par-6 family cell polarity regulator beta', other_designations='PAR-6 beta|PAR-6B|par-6 (partitioning defective 6) homolog beta|partitioning-defective protein 6, beta', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168081004, end_position_on_the_genomic_accession=168101203, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  320365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320365, CurrentID=0, Status='live', Symbol='Fry', Aliases='13CDNA73, 9330186A19Rik, AF346502, AI462658, C87180, D930046M03, cg003, mKIAA4143', description='FRY microtubule binding protein', other_designations='furry homolog', map_location='5 89.36 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=150118636, end_position_on_the_genomic_accession=150497753, orientation='plus', exon_count=66, OMIM='', no_hdr0=''),
  58222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58222, CurrentID=0, Status='live', Symbol='Rab37', Aliases='B230331O03Rik, B230354I04Rik', description='RAB37, member RAS oncogene family', other_designations='GTPase Rab37|RAB37, member of RAS oncogene family', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115091203, end_position_on_the_genomic_accession=115162773, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  58223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58223, CurrentID=0, Status='live', Symbol='Mmp19', Aliases='', description='matrix metallopeptidase 19', other_designations='matrix metalloproteinase RASI', map_location='10 77.16 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128790910, end_position_on_the_genomic_accession=128800824, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  58226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58226, CurrentID=0, Status='live', Symbol='Cacna1h', Aliases='Cav3.2, MNCb-1209, alpha13.2', description='calcium channel, voltage-dependent, T type, alpha 1H subunit', other_designations='T-type Cav3.2|calcium channel alpha13.2 subunit|low-voltage-activated calcium channel alpha13.2 subunit|voltage-gated calcium channel subunit alpha Cav3.2', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25374287, end_position_on_the_genomic_accession=25433783, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  58227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58227, CurrentID=0, Status='live', Symbol='Fam184b', Aliases='9630031F12Rik, AB041544', description='family with sequence similarity 184, member B', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=45529705, end_position_on_the_genomic_accession=45639501, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  58230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58230, CurrentID=0, Status='live', Symbol='Rnf8', Aliases='3830404E21Rik, AIP37, laXp180', description='ring finger protein 8', other_designations='ActA binding protein 2|actA-interacting protein 37', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29614831, end_position_on_the_genomic_accession=29641659, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  58231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58231, CurrentID=0, Status='live', Symbol='Stk4', Aliases='AI447339, AU020804, Kas-2, Mst1, Ysk3', description='serine/threonine kinase 4', other_designations='MST-1|STE20-like kinase MST1|Yeast Sps1/Ste20-related kinase 3|mammalian STE20-like protein kinase 1|sterile 20-like kinase 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164074109, end_position_on_the_genomic_accession=164155524, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  320376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320376, CurrentID=0, Status='live', Symbol='Bcorl1', Aliases='6720425J07Rik, B930011H20Rik', description='BCL6 co-repressor-like 1', other_designations='BCoR-L1|BCoR-like protein 1', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48341358, end_position_on_the_genomic_accession=48408571, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  58233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58233, CurrentID=0, Status='live', Symbol='Dnaja4', Aliases='1110021L12Rik, AV358213, Dj4, Hsj4, mmDjA4', description='DnaJ heat shock protein family (Hsp40) member A4', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 4|heat shock protein, DNAJ-like 4', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54696590, end_position_on_the_genomic_accession=54716317, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  58234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58234, CurrentID=0, Status='live', Symbol='Shank3', Aliases='AI841104, SHANK3c-3, SHANK3c-4b, Shank3', description='SH3/ankyrin domain gene 3', other_designations='SH3 and multiple ankyrin repeat domains protein 3 variant c-3|SH3 and multiple ankyrin repeat domains protein 3 variant c-4|SH3 and multiple ankyrin repeat domains-like protein 3 variant 1|SH3 and multiple ankyrin repeat domains-like protein 3 variant 2|SH3 and multiple ankyrin repeat domains-like protein 3 variant 3|SH3 and multiple ankyrin repeat domains-like protein 3 variant c-1|SH3 and multiple ankyrin repeat domains-like protein 3 variant c-2|SH3 and multiple ankyrin repeat domains-like protein 3 variant d-1|SH3 and multiple ankyrin repeat domains-like protein 3 variant d-2|SH3 and multiple ankyrin repeat domains-like protein 3 variant d-3|SH3 and multiple ankyrin repeat domains-like protein 3 variant e-1|SH3 and multiple ankyrin repeat domains-like protein 3 variant e-2|SPANK-2|proSAP2|proline-rich synapse-associated protein 2|shank3 splice variant II|shank3 splice variant III|shank3 splice variant IV|shank3e splice variant III|shank3e splice variant IV', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89499707, end_position_on_the_genomic_accession=89560261, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  58235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58235, CurrentID=0, Status='live', Symbol='Pvrl1', Aliases='AI835281, AW549174, Cd111, HIgR, HveC, PRR, PRR1, nectin-1', description='poliovirus receptor-related 1', other_designations='herpes virus entry mediator C|herpesvirus entry mediator C|nectin-1 alpha|nectin-1 delta|poliovirus receptor-related protein 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=43744576, end_position_on_the_genomic_accession=43807461, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  58237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58237, CurrentID=0, Status='live', Symbol='Nkain4', Aliases='AB030182, C030019F02Rik', description='Na+/K+ transporting ATPase interacting 4', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 4', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180934769, end_position_on_the_genomic_accession=180954888, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  58238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58238, CurrentID=0, Status='live', Symbol='Fam181b', Aliases='A830059I20Rik, AB041569', description='family with sequence similarity 181, member B', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=93079879, end_position_on_the_genomic_accession=93081722, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  58239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58239, CurrentID=0, Status='live', Symbol='Dexi', Aliases='1810029J14Rik, AI836170, AW413143, D16Bwg0586e, Myle', description='dexamethasone-induced transcript', other_designations='', map_location='16 5.67 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10530207, end_position_on_the_genomic_accession=10543122, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  58240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58240, CurrentID=0, Status='live', Symbol='Hs1bp3', Aliases='', description='HCLS1 binding protein 3', other_designations='HS1 binding protein 3|HS1-binding protein 3|HSP1BP-3', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8313433, end_position_on_the_genomic_accession=8343824, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  58242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58242, CurrentID=0, Status='live', Symbol='Nudt11', Aliases='DIPP3, DIPP3b', description='nudix (nucleoside diphosphate linked moiety X)-type motif 11', other_designations="DIPP-3-beta|DIPP3 beta|DIPP3-beta|diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 3-beta|diadenosine hexaphosphate hydrolase (AMP-forming)|nucleoside diphosphate-linked moiety X motif 11|nudix motif 11|nudix-type motif 11", map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=6047479, end_position_on_the_genomic_accession=6054751, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  58243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58243, CurrentID=0, Status='live', Symbol='Nap1l5', Aliases='1110020M21Rik', description='nucleosome assembly protein 1-like 5', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58905233, end_position_on_the_genomic_accession=58907126, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  58244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58244, CurrentID=0, Status='live', Symbol='Stx6', Aliases='2310039E05Rik, 2410005I16Rik, AA437559, AI415714, AI845861', description='syntaxin 6', other_designations='', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155158677, end_position_on_the_genomic_accession=155207659, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  58245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58245, CurrentID=0, Status='live', Symbol='Gpr180', Aliases='AB041545, E130016I23Rik, Itr', description='G protein-coupled receptor 180', other_designations='intimal thickness-related receptor', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118137101, end_position_on_the_genomic_accession=118164232, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  58246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58246, CurrentID=0, Status='live', Symbol='Slc35b4', Aliases='4930474D06Rik, AB041549, AW494345', description='solute carrier family 35, member B4', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34155743, end_position_on_the_genomic_accession=34177077, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  58248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58248, CurrentID=0, Status='live', Symbol='1700123O20Rik', Aliases='MNCb-2990', description='RIKEN cDNA 1700123O20 gene', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54686171, end_position_on_the_genomic_accession=54690742, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  58249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58249, CurrentID=0, Status='live', Symbol='Fibp', Aliases='2010004G08Rik, 2010005N21Rik, 3010027N18Rik', description='fibroblast growth factor (acidic) intracellular binding protein', other_designations='FGF-1 intracellular-binding protein|aFGF intracellular binding protein|aFGF intracellular-binding protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5460601, end_position_on_the_genomic_accession=5465052, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  320394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320394, CurrentID=0, Status='live', Symbol='Cenpt', Aliases='G630055P03Rik', description='centromere protein T', other_designations='CENP-T', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105844669, end_position_on_the_genomic_accession=105855038, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  58251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58251, CurrentID=0, Status='live', Symbol='BC100451', Aliases='Cep295nl, Ddc8, FS39, Kiaa1731nl', description='cDNA sequence BC100451', other_designations='KIAA1731 N-terminal like protein|differential display clone 8|protein DDC8|testis specific protein, Ddc8|testis-specific protein DDC8', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118332360, end_position_on_the_genomic_accession=118342568, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  320398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320398, CurrentID=0, Status='live', Symbol='Lrig3', Aliases='9030421L11Rik, 9130004I02Rik, 9430095K15Rik, mKIAA3016', description='leucine-rich repeats and immunoglobulin-like domains 3', other_designations='LIG-3', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=125966219, end_position_on_the_genomic_accession=126015359, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  320404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320404, CurrentID=0, Status='live', Symbol='Itpkb', Aliases='1110033J02Rik, E130307H12Rik', description='inositol 1,4,5-trisphosphate 3-kinase B', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180330476, end_position_on_the_genomic_accession=180424802, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  320405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320405, CurrentID=0, Status='live', Symbol='Cadps2', Aliases='A230044C21Rik, Caps-2, Caps2, Cpd2', description='Ca2+-dependent activator protein for secretion 2', other_designations='cerebellum postnatal development associated protein 2', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=23262773, end_position_on_the_genomic_accession=23839421, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  320407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320407, CurrentID=0, Status='live', Symbol='Klri2', Aliases='A530090P03Rik', description='killer cell lectin-like receptor family I member 2', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129729041, end_position_on_the_genomic_accession=129740484, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  320415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320415, CurrentID=0, Status='live', Symbol='Gchfr', Aliases='2010323F13Rik, GFRP, P35', description='GTP cyclohydrolase I feedback regulator', other_designations='GTP cyclohydrolase I feedback regulatory protein', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119167788, end_position_on_the_genomic_accession=119172389, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  320429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320429, CurrentID=0, Status='live', Symbol='Trank1', Aliases='A230061D21Rik, C030048J01Rik, Gm187, Lba1', description='tetratricopeptide repeat and ankyrin repeat containing 1', other_designations='lupus brain antigen 1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111311590, end_position_on_the_genomic_accession=111395775, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  320435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320435, CurrentID=0, Status='live', Symbol='Rinl', Aliases='5830482F20Rik, 9930116N10Rik', description='Ras and Rab interactor-like', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28788861, end_position_on_the_genomic_accession=28798965, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  320438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320438, CurrentID=0, Status='live', Symbol='Alg6', Aliases='E230028F23Rik', description='asparagine-linked glycosylation 6 (alpha-1,3,-glucosyltransferase)', other_designations='asparagine-linked glycosylation 6 homolog (yeast, alpha-1,3,-glucosyltransferase)|asparagine-linked glycosylation protein 6 homolog|dol-P-Glc:Man(9)GlcNAc(2)-PP-Dol alpha-1,3-glucosyltransferase|dolichyl-P-Glc:Man9GlcNAc2-PP-dolichyl glucosyltransferase', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=99715599, end_position_on_the_genomic_accession=99763460, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  320452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320452, CurrentID=0, Status='live', Symbol='P4ha3', Aliases='D930031A02Rik', description='procollagen-proline, 2-oxoglutarate 4-dioxygenase (proline 4-hydroxylase), alpha polypeptide III', other_designations='4-PH alpha-3|collagen prolyl 4-hydroxylase alpha 3|procollagen-proline,2-oxoglutarate-4-dioxygenase subunit alpha-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100285489, end_position_on_the_genomic_accession=100319701, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  320454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320454, CurrentID=0, Status='live', Symbol='Tmprss11g', Aliases='9930032O22Rik, Desc4', description='transmembrane protease, serine 11g', other_designations='serine protease Desc4', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86485877, end_position_on_the_genomic_accession=86518600, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  320460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320460, CurrentID=0, Status='live', Symbol='Vwc2l', Aliases='A830006F12Rik, Brl', description='von Willebrand factor C domain-containing protein 2-like', other_designations='Brorin-like|cysteine-knot protein F12-c isoform', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=70725221, end_position_on_the_genomic_accession=70886514, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  320472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320472, CurrentID=0, Status='live', Symbol='Ppm1e', Aliases='AW049266, B930008A12Rik, POPX1, PP2CH, mKIAA1072', description='protein phosphatase 1E (PP2C domain containing)', other_designations='ca(2+)/calmodulin-dependent protein kinase phosphatase N|caMKN|caMKP-N|caMKP-nucleus|partner of PIX 1|partner of PIX-alpha|partner of PIXA', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87226906, end_position_on_the_genomic_accession=87358994, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  320473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320473, CurrentID=0, Status='live', Symbol='Heatr5b', Aliases='2010013B10Rik, A230048G03Rik, AV071430, D330050P16Rik, mKIAA1414', description='HEAT repeat containing 5B', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78752906, end_position_on_the_genomic_accession=78835381, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  320478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320478, CurrentID=0, Status='live', Symbol='Sox2ot', Aliases='B230215L15Rik', description='SOX2 overlapping transcript (non-protein coding)', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=34560381, end_position_on_the_genomic_accession=34677993, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  320482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320482, CurrentID=0, Status='live', Symbol='Nron', Aliases='A630071L07Rik, Gm24353', description='non-protein coding RNA, repressor of NFAT', other_designations='noncoding repressor of NFAT', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33805816, end_position_on_the_genomic_accession=33813591, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  320484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320484, CurrentID=0, Status='live', Symbol='Rasal3', Aliases='A430107D22Rik', description='RAS protein activator like 3', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32390661, end_position_on_the_genomic_accession=32403612, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  320487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320487, CurrentID=0, Status='live', Symbol='Heatr5a', Aliases='C230081H03Rik, D930036F22Rik, mKIAA1316', description='HEAT repeat containing 5A', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51875871, end_position_on_the_genomic_accession=51971333, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  320492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320492, CurrentID=0, Status='live', Symbol='A830018L16Rik', Aliases='', description='RIKEN cDNA A830018L16 gene', other_designations='', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=11414105, end_position_on_the_genomic_accession=11975902, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  320495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320495, CurrentID=0, Status='live', Symbol='Ipcef1', Aliases='A130090K04Rik', description='interaction protein for cytohesin exchange factors 1', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=6885779, end_position_on_the_genomic_accession=7023185, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  320500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320500, CurrentID=0, Status='live', Symbol='Tmem215', Aliases='A930001M12Rik', description='transmembrane protein 215', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40472180, end_position_on_the_genomic_accession=40475653, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  320502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320502, CurrentID=0, Status='live', Symbol='Lmod3', Aliases='5430424A14Rik', description='leiomodin 3 (fetal)', other_designations='', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=97238528, end_position_on_the_genomic_accession=97253097, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  320506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320506, CurrentID=0, Status='live', Symbol='Lmbrd2', Aliases='9930036E21Rik', description='LMBR1 domain containing 2', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9140533, end_position_on_the_genomic_accession=9202483, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  320508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320508, CurrentID=0, Status='live', Symbol='Cachd1', Aliases='1190007F10Rik, AI852726, B430218L07Rik, Vwcd1, mKIAA1573', description='cache domain containing 1', other_designations='cache domain-containing protein 1|von Willebrand factor type A and cache domain containing 1', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=100776679, end_position_on_the_genomic_accession=101003749, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  320522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320522, CurrentID=0, Status='live', Symbol='Bhlha9', Aliases='A830053O21Rik, Fingerin', description='basic helix-loop-helix family, member a9', other_designations='bHLHf42|class B basic helix-loop-helix factor 42|novel helix-loop-helix DNA-binding domain containing protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76672470, end_position_on_the_genomic_accession=76673676, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  320528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320528, CurrentID=0, Status='live', Symbol='Vps13c', Aliases='AI552542, AV026040, C230055H22Rik, W51662, chorein', description='vacuolar protein sorting 13C (yeast)', other_designations="Vps13c 5' variant VI|vps13c 5' variant II|vps13c 5' variant III", map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=67840331, end_position_on_the_genomic_accession=67995640, orientation='plus', exon_count=87, OMIM='', no_hdr0=''),
  320534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320534, CurrentID=0, Status='live', Symbol='Tmem104', Aliases='C630005D06Rik, mFLJ00021', description='transmembrane protein 104', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115187487, end_position_on_the_genomic_accession=115247025, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  320538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320538, CurrentID=0, Status='live', Symbol='Ubn2', Aliases='2900060J04Rik, 6030408G03Rik, BC060116, D130059P03Rik, mKIAA2030', description='ubinuclein 2', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38433925, end_position_on_the_genomic_accession=38512763, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  320541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320541, CurrentID=0, Status='live', Symbol='Slc35e2', Aliases='A530082C11Rik, AI957035', description='solute carrier family 35, member E2', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155600592, end_position_on_the_genomic_accession=155623340, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  320549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320549, CurrentID=0, Status='live', Symbol='D5Ertd577e', Aliases='E330023J07Rik', description='DNA segment, Chr 5, ERATO Doi 577, expressed', other_designations='', map_location='5 47.29 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=95453012, end_position_on_the_genomic_accession=95485589, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  320554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320554, CurrentID=0, Status='live', Symbol='Tcp11l1', Aliases='C130096D04Rik', description='t-complex 11 like 1', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104678008, end_position_on_the_genomic_accession=104712169, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  320557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320557, CurrentID=0, Status='live', Symbol='Fam169a', Aliases='B230112C05Rik, SLAP75', description='family with sequence similarity 169, member A', other_designations='protein FAM169A', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=97067287, end_position_on_the_genomic_accession=97129977, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  320558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320558, CurrentID=0, Status='live', Symbol='Sycp2', Aliases='3830402K23Rik, 4930518F03Rik', description='synaptonemal complex protein 2', other_designations='SCP-2|synaptonemal complex lateral element protein', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=178345295, end_position_on_the_genomic_accession=178407658, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  320560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320560, CurrentID=0, Status='live', Symbol='Dennd5b', Aliases='9330160C06Rik, AI852444, D030011O10Rik, ENSMUSG00000072657, Gm17524', description='DENN/MADD domain containing 5B', other_designations='RAB6IP1-like protein', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148988069, end_position_on_the_genomic_accession=149101680, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  320563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320563, CurrentID=0, Status='live', Symbol='Islr2', Aliases='B930052A04Rik, Linx, Mbu-3, mKIAA1465', description='immunoglobulin superfamily containing leucine-rich repeat 2', other_designations='leucine-rich repeat domain and immunoglobulin domain containing axon extension protein|leucine-rich repeat domain and immunoglobulin domain-containing axon extension protein', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58196297, end_position_on_the_genomic_accession=58204319, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  320571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320571, CurrentID=0, Status='live', Symbol='Atp8b5', Aliases='4930417M19Rik, Feta', description='ATPase, class I, type 8B, member 5', other_designations='ATPase class I type 8B member 2-like protein|flippase expressed in testis A|probable phospholipid-transporting ATPase FetA|testis flippase', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43266714, end_position_on_the_genomic_accession=43373851, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  320581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320581, CurrentID=0, Status='live', Symbol='Idi2', Aliases='4833405L16Rik, IPPI2', description='isopentenyl-diphosphate delta isomerase 2', other_designations='IPP isomerase 2|isopentenyl diphosphate delta-isomerase type 2|isopentenyl pyrophosphate isomerase 2|isopentenyl-diphosphate Delta-isomerase 2', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=8952863, end_position_on_the_genomic_accession=8960935, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  320587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320587, CurrentID=0, Status='live', Symbol='Tmem88b', Aliases='A230069A22Rik', description='transmembrane protein 88B', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155781591, end_position_on_the_genomic_accession=155785874, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  320590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320590, CurrentID=0, Status='live', Symbol='Svopl', Aliases='9430071P14Rik', description='SV2 related protein homolog (rat)-like', other_designations='SVOP-like protein', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=37983739, end_position_on_the_genomic_accession=38047014, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  320595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320595, CurrentID=0, Status='live', Symbol='Phf8', Aliases='9830141C09Rik, mKIAA1111', description='PHD finger protein 8', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151520612, end_position_on_the_genomic_accession=151633859, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  320604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320604, CurrentID=0, Status='live', Symbol='Ccdc169', Aliases='A730037C10Rik', description='coiled-coil domain containing 169', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=55137339, end_position_on_the_genomic_accession=55175250, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  320609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320609, CurrentID=0, Status='live', Symbol='Strip2', Aliases='D330017J20Rik, Fam40b', description='striatin interacting protein 2', other_designations='family with sequence similarity 40, member B|protein FAM40B', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29916976, end_position_on_the_genomic_accession=29959681, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  320615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320615, CurrentID=0, Status='live', Symbol='Dopey1', Aliases='B130005I07Rik, BC035275, C130028L17, D9Ertd809e', description='dopey family member 1', other_designations='', map_location='9 46.58 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=86467154, end_position_on_the_genomic_accession=86555806, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  320631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320631, CurrentID=0, Status='live', Symbol='Abca15', Aliases='4930500I12Rik', description='ATP-binding cassette, sub-family A (ABC1), member 15', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120328486, end_position_on_the_genomic_accession=120407687, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  320632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320632, CurrentID=0, Status='live', Symbol='Snrnp200', Aliases='A330064G03Rik, Ascc3l1, BC011390, HELIC2, U5-200-KD, U5-200KD', description='small nuclear ribonucleoprotein 200 (U5)', other_designations='BRR2 homolog|U5 snRNP-specific 200 kDa protein|U5 snRNP-specific protein, 200 kDa (DEXH RNA helicase family)|activating signal cointegrator 1 complex subunit 3-like 1|small nuclear ribonucleoprotein 200kDa (U5)', map_location='2 61.84 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127208404, end_position_on_the_genomic_accession=127240451, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  320633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320633, CurrentID=0, Status='live', Symbol='Zbtb26', Aliases='A630026F21Rik', description='zinc finger and BTB domain containing 26', other_designations='zinc finger protein 483', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37432168, end_position_on_the_genomic_accession=37443121, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  320634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320634, CurrentID=0, Status='live', Symbol='Ocrl', Aliases='9530014D17Rik, BB143339, OCRL1', description='oculocerebrorenal syndrome of Lowe', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=47912456, end_position_on_the_genomic_accession=47966428, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  320635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320635, CurrentID=0, Status='live', Symbol='Cyb5r2', Aliases='D630003K02Rik, b5R.2', description='cytochrome b5 reductase 2', other_designations='cytochrome b5 reductase b5R.2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107748455, end_position_on_the_genomic_accession=107758189, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100328588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100328588, CurrentID=0, Status='live', Symbol='Gm21948', Aliases='Nup62-il4i1', description='predicted gene, 21948', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44816370, end_position_on_the_genomic_accession=44840804, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  320655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320655, CurrentID=0, Status='live', Symbol='Pgap3', Aliases='CAB2, D430035D22Rik, Perld1', description='post-GPI attachment to proteins 3', other_designations='PER1-like domain-containing protein 1|per1-like domain containing 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98388672, end_position_on_the_genomic_accession=98400539, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  320661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320661, CurrentID=0, Status='live', Symbol='D5Ertd579e', Aliases='9030221A05Rik, A930018H20Rik, Kiaa0232, mKIAA0232', description='DNA segment, Chr 5, ERATO Doi 579, expressed', other_designations='', map_location='5 19.18 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36600483, end_position_on_the_genomic_accession=36696732, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  58518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58518, CurrentID=0, Status='live', Symbol='Cts6', Aliases='1600022N02Rik', description='cathepsin 6', other_designations='', map_location='13 32.67 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61195147, end_position_on_the_genomic_accession=61203392, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  58520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58520, CurrentID=0, Status='live', Symbol='0610007P14Rik', Aliases='1190004E09Rik, AU019315, C77855, ORF11', description='RIKEN cDNA 0610007P14 gene', other_designations='c14orf1-like protein', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85815447, end_position_on_the_genomic_accession=85824921, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  58521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58521, CurrentID=0, Status='live', Symbol='Eid1', Aliases='2610002K20Rik, Cri1, EID-1, ORF12, PNAS-22, PTD014', description='EP300 interacting inhibitor of differentiation 1', other_designations='CREBBP/EP300 inhibitory protein 1|E1A-like inhibitor of differentiation 1|eid1', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125673100, end_position_on_the_genomic_accession=125675643, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  58522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58522, CurrentID=0, Status='live', Symbol='Trim54', Aliases='4930486E09Rik, 4930566I02Rik, MURF, MURF-3, MuRF3, Rnf30', description='tripartite motif-containing 54', other_designations='muscle-specific RING finger protein 3|muscle-specific RING-finger protein|ring finger protein 30', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31116612, end_position_on_the_genomic_accession=31137626, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  58523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58523, CurrentID=0, Status='live', Symbol='Elp2', Aliases='AU023723, Epl2, StIP1, Statip1', description='elongator acetyltransferase complex subunit 2', other_designations='STAT3-interacting protein 1|elongation protein 2 homolog|signal transducer and activator of transcription 3 interacting protein 1|signal transducer and activator of transcription interacting protein 1', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24603923, end_position_on_the_genomic_accession=24638830, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  320678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320678, CurrentID=0, Status='live', Symbol='Iffo1', Aliases='4733401N06Rik, A930037G23Rik, HOM-TES-103, Iffo', description='intermediate filament family orphan 1', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125145202, end_position_on_the_genomic_accession=125161782, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  320679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320679, CurrentID=0, Status='live', Symbol='Samd12', Aliases='A830094I09Rik', description='sterile alpha motif domain containing 12', other_designations='SAM domain-containing protein 12', map_location='15 C-D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=53399119, end_position_on_the_genomic_accession=53902536, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  320683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320683, CurrentID=0, Status='live', Symbol='Zfp629', Aliases='9330199A09Rik, Znf629', description='zinc finger protein 629', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127607031, end_position_on_the_genomic_accession=127615786, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  320685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320685, CurrentID=0, Status='live', Symbol='Dctd', Aliases='6030466N05Rik', description='dCMP deaminase', other_designations='', map_location='8 B1.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=48099092, end_position_on_the_genomic_accession=48141667, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  320696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320696, CurrentID=0, Status='live', Symbol='Ccdc158', Aliases='4932413O14Rik', description='coiled-coil domain containing 158', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92608237, end_position_on_the_genomic_accession=92675263, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  320701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320701, CurrentID=0, Status='live', Symbol='Fam19a4', Aliases='C130034I18Rik Tafa-4, Tafa-4, Fam19a4', description='family with sequence similarity 19, member A4', other_designations='TAFA4 protein|chemokine-like protein TAFA-4|family with sequence similarity 19 (chemokine (C-C motif)-like), member A4', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=96831206, end_position_on_the_genomic_accession=97080124, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  320705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320705, CurrentID=0, Status='live', Symbol='Bend6', Aliases='B230209C24Rik', description='BEN domain containing 6', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=33852052, end_position_on_the_genomic_accession=33907767, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  320706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320706, CurrentID=0, Status='live', Symbol='Soga1', Aliases='9830001H06Rik, AI426038, D430036N24Rik, Soga', description='suppressor of glucose, autophagy associated 1', other_designations='SOGA family member 1|suppressor of glucose by autophagy|suppressor of glucose, autophagy-associated protein 1', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157010442, end_position_on_the_genomic_accession=157079265, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  320707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320707, CurrentID=0, Status='live', Symbol='Atp2b3', Aliases='6430519O13Rik, Pmca3', description='ATPase, Ca++ transporting, plasma membrane 3', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73503023, end_position_on_the_genomic_accession=73573275, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  320709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320709, CurrentID=0, Status='live', Symbol='Tmem117', Aliases='B930062P21Rik', description='transmembrane protein 117', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=94629185, end_position_on_the_genomic_accession=95096097, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  320712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320712, CurrentID=0, Status='live', Symbol='Abi3bp', Aliases='5033411B22Rik, AI506287, BG172926, D930038M13Rik, TARSH, eratin', description='ABI gene family, member 3 (NESH) binding protein', other_designations='', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56477801, end_position_on_the_genomic_accession=56690135, orientation='plus', exon_count=60, OMIM='', no_hdr0=''),
  320713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320713, CurrentID=0, Status='live', Symbol='Mysm1', Aliases='C130067A03Rik, C530050H10Rik', description='myb-like, SWIRM and MPN domains 1', other_designations='2A-DUB|myb-like, SWIRM and MPN domain-containing protein 1', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=94942037, end_position_on_the_genomic_accession=94979302, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  320714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320714, CurrentID=0, Status='live', Symbol='Trappc11', Aliases='AW558465, D030016E14Rik, R75422', description='trafficking protein particle complex 11', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47490128, end_position_on_the_genomic_accession=47533470, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  320717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320717, CurrentID=0, Status='live', Symbol='Pptc7', Aliases='9130017A15Rik, AA672638, AI848390, TA-PP2C', description='PTC7 protein phosphatase homolog', other_designations='T-cell activation protein phosphatase 2C', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122284398, end_position_on_the_genomic_accession=122324281, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320718, CurrentID=0, Status='live', Symbol='Slc26a9', Aliases='E030002L01Rik', description='solute carrier family 26, member 9', other_designations='SLC26A9 anion transporter/exchanger|anion transporter/exchanger protein 9|anion transporter/exchanger-9', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131744022, end_position_on_the_genomic_accession=131771504, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  320720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320720, CurrentID=0, Status='live', Symbol='Fastkd1', Aliases='5330408N05Rik, mKIAA1800', description='FAST kinase domains 1', other_designations='', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69686820, end_position_on_the_genomic_accession=69712606, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  320722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320722, CurrentID=0, Status='live', Symbol='A330050F15Rik', Aliases='', description='RIKEN cDNA A330050F15 gene', other_designations='', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=69437591, end_position_on_the_genomic_accession=69489233, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  320727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320727, CurrentID=0, Status='live', Symbol='Ipo8', Aliases='6230418K12Rik, Abcc10, C130009K11Rik, MRP7, OM-1, Om1, Ranbp8', description='importin 8', other_designations='ATP-binding cassette, sub-family C (CFTR/MRP), member 10|imp8|oocyte maturation 1|ran-binding protein 8', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148770683, end_position_on_the_genomic_accession=148831467, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  320736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320736, CurrentID=0, Status='live', Symbol='Vstm4', Aliases='E130203B14Rik', description='V-set and transmembrane domain containing 4', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32856724, end_position_on_the_genomic_accession=32939489, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  320747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320747, CurrentID=0, Status='live', Symbol='Lingo4', Aliases='A530050P17Rik, LERN4, Lrrn6d', description='leucine rich repeat and Ig domain containing 4', other_designations='leucine rich repeat neuronal 6D|leucine-rich repeat neuronal protein 6D', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94399219, end_position_on_the_genomic_accession=94404501, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  320752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320752, CurrentID=0, Status='live', Symbol='Dpy19l2', Aliases='4932443J21Rik, Gm18376', description='dpy-19-like 2 (C. elegans)', other_designations='dpy-19-like protein 2|protein dpy-19 homolog 2', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=24557047, end_position_on_the_genomic_accession=24696941, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  320769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320769, CurrentID=0, Status='live', Symbol='Prdx6b', Aliases='1-cysPrx-P1, 4930414C22Rik, Aop2-rs1, Prdx5-rs1, Prdx6-ps1, Prdx6-rs1, Prdx6-rs1-ps', description='peroxiredoxin 6B', other_designations='anti-oxidant protein 2 related sequence 1|peroxiredoxin 5 related sequence 1|peroxiredoxin 6, pseudogene 1', map_location='2 48.03 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=80292470, end_position_on_the_genomic_accession=80295358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  320772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320772, CurrentID=0, Status='live', Symbol='Mdga2', Aliases='6720489L24Rik, 9330209L04Rik, Adp, Mamdc1', description='MAM domain containing glycosylphosphatidylinositol anchor 2', other_designations='MAM domain containing 1|acrodysplasia', map_location='12 27.63 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=66466059, end_position_on_the_genomic_accession=67222549, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  320782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320782, CurrentID=0, Status='live', Symbol='Tmem154', Aliases='9930117H01Rik', description='transmembrane protein 154', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84666192, end_position_on_the_genomic_accession=84704575, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  320790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320790, CurrentID=0, Status='live', Symbol='Chd7', Aliases='A730019I05Rik, Cycn, Cyn, Dz, Edy, Flo, GENA 47, GENA 60, Gena 52, Lda, Mt, Obt, Todo, WBE1, Whi, metis', description='chromodomain helicase DNA binding protein 7', other_designations='ATP-dependent helicase CHD7', map_location='4 3.68 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=8690406, end_position_on_the_genomic_accession=8868449, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  320795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320795, CurrentID=0, Status='live', Symbol='Pkn1', Aliases='DBK, F730027O18Rik, PAK1, PRK1, Pkn, Prkcl1, Stk3', description='protein kinase N1', other_designations='protein kinase C-like 1|protein kinase C-like PKN|protein-kinase C-related kinase 1|serine-threonine protein kinase N|serine/threonine kinase 3', map_location='8 40.22 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83669762, end_position_on_the_genomic_accession=83699179, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  320799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320799, CurrentID=0, Status='live', Symbol='Zhx3', Aliases='1810059C13Rik, 4932418O04Rik, 9530010N21Rik, Tix1, mKIAA0395', description='zinc fingers and homeoboxes 3', other_designations='triple homeobox 1|triple homeobox protein 1|zinc finger and homeodomain protein 3', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=160770447, end_position_on_the_genomic_accession=160872990, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  320806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320806, CurrentID=0, Status='live', Symbol='Gfm2', Aliases='6530419G12Rik, A930009M04Rik, EF-G2mt, EFG2, MST027, RRF2mt', description='G elongation factor, mitochondrial 2', other_designations='EF-G2mt|RRF2mt', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=97137937, end_position_on_the_genomic_accession=97181196, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  320808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320808, CurrentID=0, Status='live', Symbol='Dcaf5', Aliases='9430020B07Rik, AI325023, AI430035, AI848110, AW124614, BCRG2, BCRP2, Wdr22', description='DDB1 and CUL4 associated factor 5', other_designations='WD repeat domain 22|WD repeat-containing protein 22', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80335847, end_position_on_the_genomic_accession=80436601, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  320816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320816, CurrentID=0, Status='live', Symbol='Ankrd16', Aliases='2810455F06Rik, AI646698, D430029B21Rik', description='ankyrin repeat domain 16', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=11777753, end_position_on_the_genomic_accession=11790327, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  320817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320817, CurrentID=0, Status='live', Symbol='Atad2b', Aliases='1110014E10Rik, BC032887, C79189, D530031C13Rik', description='ATPase family, AAA domain containing 2B', other_designations='', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4916084, end_position_on_the_genomic_accession=5047410, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  320825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320825, CurrentID=0, Status='live', Symbol='Samd5', Aliases='E130306M17Rik', description='sterile alpha motif domain containing 5', other_designations='SAM domain-containing protein 5', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=9627259, end_position_on_the_genomic_accession=9675208, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  320827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320827, CurrentID=0, Status='live', Symbol='C530008M17Rik', Aliases='2810489J07Rik, 4930529A19Rik, 4931405B09Rik, Kiaa1211, mKIAA1211', description='RIKEN cDNA C530008M17 gene', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76657777, end_position_on_the_genomic_accession=76873554, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  320832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320832, CurrentID=0, Status='live', Symbol='Sirpb1a', Aliases='9930027N05Rik, SIRP-beta, Sirpb, Sirpb1', description='signal-regulatory protein beta 1A', other_designations='SIRP beta 1 cell surface protein|SIRP-beta B', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=15371661, end_position_on_the_genomic_accession=15426460, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  320840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320840, CurrentID=0, Status='live', Symbol='Negr1', Aliases='5330422G01Rik, KILON, Ntra', description='neuronal growth regulator 1', other_designations='kindred of IgLON|neurotractin', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=156561794, end_position_on_the_genomic_accession=157316464, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  668158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668158, CurrentID=0, Status='live', Symbol='Ccdc85c', Aliases='EG668158, Gm265, Gm9010', description='coiled-coil domain containing 85C', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108203599, end_position_on_the_genomic_accession=108275671, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  320844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320844, CurrentID=0, Status='live', Symbol='Amigo3', Aliases='E430002N15Rik, ali3, mKIAA1851', description='adhesion molecule with Ig like domain 3', other_designations='AMIGO-3|alivin 3|alivin-3|amigo3|amphoterin induced gene and ORF 3', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108053159, end_position_on_the_genomic_accession=108055701, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  320858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320858, CurrentID=0, Status='live', Symbol='L3mbtl4', Aliases='A730037L19Rik, D930040M24Rik', description='l(3)mbt-like 4 (Drosophila)', other_designations='l(3)mbt-like protein 4', map_location='17 E1.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=68273246, end_position_on_the_genomic_accession=68784078, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  320864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320864, CurrentID=0, Status='live', Symbol='Krt26', Aliases='4732407F15Rik, Krt25b', description='keratin 26', other_designations='CK-26|K26|cytokeratin-26|keratin 25B|keratin-26|krt26|type I inner root sheath-specific keratin-K25irs2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99328484, end_position_on_the_genomic_accession=99338286, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  320865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320865, CurrentID=0, Status='live', Symbol='Cdh18', Aliases='B230220E17Rik, Cdh14l', description='cadherin 18', other_designations='', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=22548899, end_position_on_the_genomic_accession=23474432, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  320869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320869, CurrentID=0, Status='live', Symbol='Spata33', Aliases='4732415M23Rik, AI847264', description='spermatogenesis associated 33', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123212816, end_position_on_the_genomic_accession=123222045, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  320873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320873, CurrentID=0, Status='live', Symbol='Cdh10', Aliases='A830016G23Rik, C030003B10Rik, C030011H18Rik', description='cadherin 10', other_designations='T2-cadherin', map_location='15 8.2 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=18818888, end_position_on_the_genomic_accession=19016206, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  320878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320878, CurrentID=0, Status='live', Symbol='Mical2', Aliases='5330438E18Rik, 9530064J02, MICAL-2, mKIAA0750', description='microtubule associated monooxygenase, calponin and LIM domain containing 2', other_designations='mMical2|microtubule associated monoxygenase, calponin and LIM domain containing 2|molecule interacting with CasL protein 2|protein MICAL-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=112225836, end_position_on_the_genomic_accession=112355194, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  320898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320898, CurrentID=0, Status='live', Symbol='Trav3-3', Aliases='A430107P09Rik, Gm13917, OTTMUSG00000015062', description='T cell receptor alpha variable 3-3', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53666005, end_position_on_the_genomic_accession=53666506, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  320910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320910, CurrentID=0, Status='live', Symbol='Itgb8', Aliases='4832412O06Rik', description='integrin beta 8', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=119158186, end_position_on_the_genomic_accession=119238276, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  320913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320913, CurrentID=0, Status='live', Symbol='Trdv2-2', Aliases='Gm17015', description='T cell receptor delta variable 2-2', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53960993, end_position_on_the_genomic_accession=53961602, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  320916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320916, CurrentID=0, Status='live', Symbol='Wscd2', Aliases='4933413A10Rik, C530024P05Rik, Gm450', description='WSC domain containing 2', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113546849, end_position_on_the_genomic_accession=113589725, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  320923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320923, CurrentID=0, Status='live', Symbol='Mtap7d3', Aliases='4933424A10Rik, Map7d3', description='MAP7 domain containing 3', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56775049, end_position_on_the_genomic_accession=56822490, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  320924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320924, CurrentID=0, Status='live', Symbol='Ccbe1', Aliases='4933426F18Rik, 9430093N24Rik, mKIAA1983', description='collagen and calcium binding EGF domains 1', other_designations='full of fluid protein homolog', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=66056855, end_position_on_the_genomic_accession=66291838, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  320938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320938, CurrentID=0, Status='live', Symbol='Tnpo3', Aliases='5730544L10Rik, C430013M08Rik, C81142, D6Ertd313e, Trn-SR, mKIAA4133', description='transportin 3', other_designations='transportin-SR', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29540827, end_position_on_the_genomic_accession=29610110, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  320940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320940, CurrentID=0, Status='live', Symbol='Atp11c', Aliases='A330005H02Rik, AI315324, Ig', description='ATPase, class VI, type 11C', other_designations='Atpase 11c, p type|P4-ATPase flippase complex alpha subunit ATP11C|probable phospholipid-transporting ATPase 11C', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=60223283, end_position_on_the_genomic_accession=60404859, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  58799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58799, CurrentID=0, Status='live', Symbol='Crbn', Aliases='2610203G15Rik, 2900045O07Rik, AF229032, AW108261, piL', description='cereblon', other_designations='novel lethal|protein PiL', map_location='6 E2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=106778244, end_position_on_the_genomic_accession=106800087, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  58800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58800, CurrentID=0, Status='live', Symbol='Trpm7', Aliases='2310022G15Rik, 4833414K03Rik, 5033407O22Rik, CHAK, CHAK1, LTrpC-7, Ltpr7, Ltrpc7, TRPPLIK', description='transient receptor potential cation channel, subfamily M, member 7', other_designations='channel-kinase 1|long transient receptor potential channel 7|transient receptor potential M7|transient receptor potential-phospholipase C-interacting kinase|transient receptor potential-related protein, ChaK', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126791558, end_position_on_the_genomic_accession=126876261, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  58801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58801, CurrentID=0, Status='live', Symbol='Pmaip1', Aliases='Noxa', description='phorbol-12-myristate-13-acetate-induced protein 1', other_designations='protein Noxa', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=66458604, end_position_on_the_genomic_accession=66465558, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  58802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58802, CurrentID=0, Status='live', Symbol='Kcnmb4', Aliases='1700058G18Rik, 2900045G12Rik', description='potassium large conductance calcium-activated channel, subfamily M, beta member 4', other_designations='BK channel subunit beta-4|BKbeta4|Slowpoke beta 4|calcium-activated potassium channel, subfamily M subunit beta-4|charybdotoxin receptor subunit beta-4|k(VCA)beta-4|large conductance calcium-dependent potassium ion channel beta 4 subunit|maxi K channel subunit beta-4|slo-beta-4', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=116417868, end_position_on_the_genomic_accession=116473882, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  58803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58803, CurrentID=0, Status='live', Symbol='Pga5', Aliases='1110035E17Rik, Pepf', description='pepsinogen 5, group I', other_designations='pepsinogen A5|pepsinogen F', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10668957, end_position_on_the_genomic_accession=10678071, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  58804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58804, CurrentID=0, Status='live', Symbol='Cdc42ep5', Aliases='1700027J19Rik, 2010007O02Rik, Borg3, C85526, CEP5', description='CDC42 effector protein (Rho GTPase binding) 5', other_designations='CRIB-containing BORG3 protein|binder of Rho GTPase 3|binder of Rho GTPases 3', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4151260, end_position_on_the_genomic_accession=4164737, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  58805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58805, CurrentID=0, Status='live', Symbol='Mlxipl', Aliases='ChREBP, WS-bHLH, Wbscr14, bHLHd14', description='MLX interacting protein-like', other_designations='MLX interacting protein-like beta|MLX interactor|MLX-interacting protein-like|Williams-Beuren syndrome chromosome region 14 homolog|putative hepatic transcription factor', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135089891, end_position_on_the_genomic_accession=135138382, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  320951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320951, CurrentID=0, Status='live', Symbol='Pisd', Aliases='9030221M09Rik', description='phosphatidylserine decarboxylase', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=32736303, end_position_on_the_genomic_accession=32785636, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  58809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58809, CurrentID=0, Status='live', Symbol='Rnase4', Aliases='C730049F20Rik, Rab1', description='ribonuclease, RNase A family 4', other_designations='RNase 4|ribonuclease A b1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51091077, end_position_on_the_genomic_accession=51106151, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  58810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58810, CurrentID=0, Status='live', Symbol='Akr1a1', Aliases='2610201A18Rik, Akr1a4', description='aldo-keto reductase family 1, member A1 (aldehyde reductase)', other_designations='alcohol dehydrogenase|aldehyde reductase|aldo-keto reductase family 1 member A1|aldo-keto reductase family 1, member A4 (aldehyde reductase)', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116636510, end_position_on_the_genomic_accession=116651674, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  320974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320974, CurrentID=0, Status='live', Symbol='Lrrn4', Aliases='B430119L13Rik, Nlrr4', description='leucine rich repeat neuronal 4', other_designations='NLRR-4|neuronal leucine rich repeat-4|neuronal leucine-rich repeat protein 4', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132866194, end_position_on_the_genomic_accession=132894009, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  320981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320981, CurrentID=0, Status='live', Symbol='Enpp6', Aliases='4833421B01Rik, B830047L21Rik, D8Ertd514e, E-NPP 6, Npp6', description='ectonucleotide pyrophosphatase/phosphodiesterase 6', other_designations='E-NPP6|GPC-Cpde|NPP-6|choline-specific glycerophosphodiester phosphodiesterase|glycerophosphocholine cholinephosphodiesterase', map_location='8 26.48 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46939199, end_position_on_the_genomic_accession=47096358, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  320982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320982, CurrentID=0, Status='live', Symbol='Arl4c', Aliases='A630084M22Rik, Arl7, LAK', description='ADP-ribosylation factor-like 4C', other_designations='ADP-ribosylation factor-like 7', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88698225, end_position_on_the_genomic_accession=88702191, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  320995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320995, CurrentID=0, Status='live', Symbol='Rfx6', Aliases='4930572O07Rik, Rfxdc1', description='regulatory factor X, 6', other_designations='regulatory factor X domain containing 1|regulatory factor X domain-containing protein 1', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=51677752, end_position_on_the_genomic_accession=51730432, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  320997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320997, CurrentID=0, Status='live', Symbol='Cyp4f39', Aliases='4732474A20Rik', description='cytochrome P450, family 4, subfamily f, polypeptide 39', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32452695, end_position_on_the_genomic_accession=32493320, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  321000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321000, CurrentID=0, Status='live', Symbol='Lrif1', Aliases='2010012G17Rik, 4933421E11Rik, AI450568', description='ligand dependent nuclear receptor interacting factor 1', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106684934, end_position_on_the_genomic_accession=106736577, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  58859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58859, CurrentID=0, Status='live', Symbol='Efemp2', Aliases='0610011K11Rik, Fbln4, MBP1', description='epidermal growth factor-containing fibulin-like extracellular matrix protein 2', other_designations='EGF-containing fibulin-like extracellular matrix protein 2 (Efemp2)|FIBL-4|fibulin 4|fibulin-4|mutant p53-binding protein 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5473960, end_position_on_the_genomic_accession=5481854, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  58860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58860, CurrentID=0, Status='live', Symbol='Adamdec1', Aliases='2210414L24Rik, AI574231, Dcsn', description='ADAM-like, decysin 1', other_designations='ADAM-like protein decysin 1|ADAM-like protein decysin-1|a disintegrin and metalloproteinase domain-like protein decysin 1|a disintegrin and metalloproteinase domain-like protein decysin-1|decysin|disintegrin metalloprotease (decysin)', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=68563387, end_position_on_the_genomic_accession=68582072, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  58861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58861, CurrentID=0, Status='live', Symbol='Cysltr1', Aliases='BB147369, CysLT1R, Cyslt1', description='cysteinyl leukotriene receptor 1', other_designations='LTD4 receptor|cysteinyl leukotriene 1 receptor short isoform|cysteinyl leukotriene D4 receptor', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106576509, end_position_on_the_genomic_accession=106603679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  321006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321006, CurrentID=0, Status='live', Symbol='Vprbp', Aliases='AI447437, B930007L02Rik, mKIAA0800', description='Vpr (HIV-1) binding protein', other_designations='DDB1- and CUL4-associated factor 1|serine/threonine-protein kinase VPRBP', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106814172, end_position_on_the_genomic_accession=106880992, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  321007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321007, CurrentID=0, Status='live', Symbol='Serac1', Aliases='4930511N22Rik, D17Ertd141e', description='serine active site containing 1', other_designations='serine active site-containing protein 1', map_location='17 3.67 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6040571, end_position_on_the_genomic_accession=6079762, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  321008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321008, CurrentID=0, Status='live', Symbol='6330408A02Rik', Aliases='', description='RIKEN cDNA 6330408A02 gene', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13258967, end_position_on_the_genomic_accession=13278721, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  58865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58865, CurrentID=0, Status='live', Symbol='Tdh', Aliases='', description='L-threonine dehydrogenase', other_designations='tdh', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63492347, end_position_on_the_genomic_accession=63509092, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  58866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58866, CurrentID=0, Status='live', Symbol='Treh', Aliases='2210412M19Rik', description='trehalase (brush-border membrane glycoprotein)', other_designations='alpha,alpha-trehalase|alpha,alpha-trehalose glucohydrolase', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44673233, end_position_on_the_genomic_accession=44686305, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  58867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58867, CurrentID=0, Status='live', Symbol='Syngr4', Aliases='1700016O14Rik', description='synaptogyrin 4', other_designations='synaptogyrin 4 protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45886078, end_position_on_the_genomic_accession=45896728, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  58869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58869, CurrentID=0, Status='live', Symbol='Pex5l', Aliases='1700016J08Rik, PXR2, Pex2, TRIP8b, pex5Rp', description='peroxisomal biogenesis factor 5-like', other_designations='PEX2-related protein|PEX5-like protein|peroxin 2|peroxin-5-related protein|peroxisomal targeting signal 1 receptor-like|tetratricopeptide repeat-containing Rab8b-interacting protein', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32949631, end_position_on_the_genomic_accession=33143273, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  321018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321018, CurrentID=0, Status='live', Symbol='Serpina4-ps1', Aliases='KALps, Klst', description='serine (or cysteine) peptidase inhibitor, clade A, member 4, pseudogene 1', other_designations='kallistatin related protein|serine (or cysteine) proteinase inhibitor, clade A, member 4, pseudogene 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104077959, end_position_on_the_genomic_accession=104086920, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  58875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58875, CurrentID=0, Status='live', Symbol='Hibadh', Aliases='6430402H10Rik, AI265272', description='3-hydroxyisobutyrate dehydrogenase', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52546230, end_position_on_the_genomic_accession=52640300, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  321020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321020, CurrentID=0, Status='live', Symbol='Fpr-rs6', Aliases='', description='formyl peptide receptor, related sequence 6', other_designations='', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20182078, end_position_on_the_genomic_accession=20183097, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  321021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321021, CurrentID=0, Status='live', Symbol='Fpr-rs7', Aliases='', description='formyl peptide receptor, related sequence 7', other_designations='', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20113210, end_position_on_the_genomic_accession=20114226, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  321022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321022, CurrentID=0, Status='live', Symbol='Cdv3', Aliases='2510010F10Rik, C230084J24Rik, C79446, TPP36', description='carnitine deficiency-associated gene expressed in ventricle 3', other_designations='carnitine deficiency-associated protein 3|tyrosine-phosphorylated protein 36', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103353102, end_position_on_the_genomic_accession=103365835, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  58887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58887, CurrentID=0, Status='live', Symbol='Repin1', Aliases='AI425994, Ap4, E430037F08Rik, Zfp-464, Zfp464', description='replication initiator 1', other_designations='zinc finger protein 464 (RIP60)', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48593883, end_position_on_the_genomic_accession=48599082, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  58909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58909, CurrentID=0, Status='live', Symbol='Fam13a', Aliases='AI462012, D430015B01Rik, FAM13A1, Precm1', description='family with sequence similarity 13, member A', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58933536, end_position_on_the_genomic_accession=59024502, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  58911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58911, CurrentID=0, Status='live', Symbol='Sumf1', Aliases='AA543204, AI463102, AI851573, FGE', description='sulfatase modifying factor 1', other_designations='C-alpha-formylglycine-generating enzyme 1', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=108107021, end_position_on_the_genomic_accession=108185591, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  58916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58916, CurrentID=0, Status='live', Symbol='Myot', Aliases='5530402I04Rik, Ttid', description='myotilin', other_designations='myofibrillar titin-like Ig domains protein|titan-like protein|titin immunoglobulin domain protein', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44334053, end_position_on_the_genomic_accession=44355740, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  665186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665186, CurrentID=0, Status='live', Symbol='Gm7534', Aliases='EG665186', description='predicted gene 7534', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134190801, end_position_on_the_genomic_accession=134203004, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  58988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58988, CurrentID=0, Status='live', Symbol='Rps6kb2', Aliases='70kDa, S6K2', description='ribosomal protein S6 kinase, polypeptide 2', other_designations='70 kDa ribosomal protein S6 kinase 2|S6 kinase 2|S6K-beta-2|p70 S6 kinase beta|p70 S6K-beta|p70 S6KB|p70 ribosomal S6 kinase beta|ribosomal protein S6 kinase, 70kD, polypeptide 2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4156792, end_position_on_the_genomic_accession=4163383, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  58991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58991, CurrentID=0, Status='live', Symbol='Ghrl', Aliases='2210006E23Rik, Ghr, MTLRP, MTLRPAP, m46', description='ghrelin', other_designations='growth hormone secretagogue|growth hormone-releasing peptide|motilin-related peptide|preproghrelin|proghrelin', map_location='6 52.84 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113716119, end_position_on_the_genomic_accession=113727485, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  58992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58992, CurrentID=0, Status='live', Symbol='F12', Aliases='FXII, HAF', description='coagulation factor XII (Hageman factor)', other_designations='factor XII|hageman factor', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55417966, end_position_on_the_genomic_accession=55426802, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  58994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58994, CurrentID=0, Status='live', Symbol='Smpd3', Aliases='4631433G07Rik, AI427456, AW537966, fro, nSMase2', description='sphingomyelin phosphodiesterase 3, neutral', other_designations='nSMase-2|neutral sphingomyelinase 2|neutral sphingomyelinase II', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106252548, end_position_on_the_genomic_accession=106337988, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  58996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58996, CurrentID=0, Status='live', Symbol='Arhgap23', Aliases='A330041B18Rik', description='Rho GTPase activating protein 23', other_designations='rho-type GTPase-activating protein 23', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97404393, end_position_on_the_genomic_accession=97502402, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  58998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58998, CurrentID=0, Status='live', Symbol='Pvrl3', Aliases='2610301B19Rik, 3000002N23Rik, 4921513D19Rik, AA407785, AU016832, AW538082, CD113', description='poliovirus receptor-related 3', other_designations='poliovirus receptor-related protein 3', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=46394858, end_position_on_the_genomic_accession=46498684, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  59001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59001, CurrentID=0, Status='live', Symbol='Pole3', Aliases='1810034K18Rik, YBL1', description='polymerase (DNA directed), epsilon 3 (p17 subunit)', other_designations='DNA polymerase II subunit 3|DNA polymerase epsilon subunit p17|DNA polymerase epsilon, subunit 3|NF-YB-like protein|YB-like protein 1', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62523798, end_position_on_the_genomic_accession=62525014, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59002, CurrentID=0, Status='live', Symbol='Wrap73', Aliases='2610044M17Rik, 5330425N03Rik, Dd57, Wdr8', description='WD repeat containing, antisense to Trp73', other_designations='WD repeat domain 8|WD repeat-containing protein 8|WD repeat-containing protein antisense to TP73', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154142357, end_position_on_the_genomic_accession=154168259, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  59003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59003, CurrentID=0, Status='live', Symbol='Maea', Aliases='1110030D19Rik, EMP, Gid9', description='macrophage erythroblast attacher', other_designations='erythroblast macrophage protein', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33335572, end_position_on_the_genomic_accession=33373294, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59004, CurrentID=0, Status='live', Symbol='Pias4', Aliases='PIASY, Pias-gamma, Piasg', description='protein inhibitor of activated STAT 4', other_designations='protein inhibitor of activated STAT PIASy|protein inhibitor of activated STAT protein 4|protein inhibitor of activated STAT protein gamma', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81153961, end_position_on_the_genomic_accession=81167720, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  59005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59005, CurrentID=0, Status='live', Symbol='Trappc2l', Aliases='1810017G16Rik, AB030200, Hspc176, Tca17', description='trafficking protein particle complex 2-like', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122611626, end_position_on_the_genomic_accession=122615591, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  59006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59006, CurrentID=0, Status='live', Symbol='Myoz2', Aliases='1110012I24Rik, Fatz-2, Myozl2', description='myozenin 2', other_designations='FATZ-related protein 2|calsarcin-1|myoz2|myozenin-like 2', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=123006206, end_position_on_the_genomic_accession=123035001, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  59007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59007, CurrentID=0, Status='live', Symbol='Ngly1', Aliases='1110002C09Rik, PNGase, Png1', description='N-glycanase 1', other_designations='mPNGase|peptide N-glycanase homolog|peptide:N-glycanase', map_location='14 7.08 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=16249314, end_position_on_the_genomic_accession=16311926, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  59008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59008, CurrentID=0, Status='live', Symbol='Anapc5', Aliases='2510006G12Rik, AA408751, AA536819, AA986414, APC5, Anpc5', description='anaphase-promoting complex subunit 5', other_designations='cyclosome subunit 5', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122787459, end_position_on_the_genomic_accession=122821342, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  59009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59009, CurrentID=0, Status='live', Symbol='Sh3rf1', Aliases='2200003J05Rik, Posh, R75531, Sh3md2', description='SH3 domain containing ring finger 1', other_designations='SH3 domain-containing RING finger protein 1|SH3 multiple domains 2|SH3 multiple domains protein 2|plenty of SH3 domains|plenty of SH3s|putative E3 ubiquitin-protein ligase SH3RF1', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=61224171, end_position_on_the_genomic_accession=61464778, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  59010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59010, CurrentID=0, Status='live', Symbol='Sqrdl', Aliases='0610039J17Rik, 4930557M22Rik, SQOR', description='sulfide quinone reductase-like (yeast)', other_designations='flavo-binding protein', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122764928, end_position_on_the_genomic_accession=122809569, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  59011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59011, CurrentID=0, Status='live', Symbol='Myoz1', Aliases='2310001N11Rik, AV090278, FATZ, Myoz', description='myozenin 1', other_designations='calcineurin-2|calsarcin-2|filamin-, actinin- and telethonin-binding protein|myoz1|skeletal muscle-specific protein', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20649102, end_position_on_the_genomic_accession=20656540, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  59012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59012, CurrentID=0, Status='live', Symbol='Moxd1', Aliases='3230402N08Rik, MNCb-5203', description='monooxygenase, DBH-like 1', other_designations='DBH-related protein|monooxygenase X', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24223517, end_position_on_the_genomic_accession=24302783, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  59013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59013, CurrentID=0, Status='live', Symbol='Hnrnph1', Aliases='AI642080, E430005G16Rik, Hnrnph, Hnrph1', description='heterogeneous nuclear ribonucleoprotein H1', other_designations='hnRNP H', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50377694, end_position_on_the_genomic_accession=50386528, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  59014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59014, CurrentID=0, Status='live', Symbol='Rrs1', Aliases='5730466A07Rik, AA415037, D1Ertd701e, Rrr', description='RRS1 ribosome biogenesis regulator homolog (S. cerevisiae)', other_designations='regulator for ribosome resistance homolog', map_location='1 2.08 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9545408, end_position_on_the_genomic_accession=9547455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  59015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59015, CurrentID=0, Status='live', Symbol='Nup160', Aliases='160kDa, 2810011M03Rik, AA414952, AU020188, Gtl-13, Gtl1-13, Gtl13, mKIAA0197', description='nucleoporin 160', other_designations='160 kDa nucleoporin|gene trap locus 1-13 protein|gene trap locus-13|nucleoporin 160kDa|nucleoporin Nup160', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90677215, end_position_on_the_genomic_accession=90736872, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  59016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59016, CurrentID=0, Status='live', Symbol='Thap11', Aliases='2810036E22Rik, AB041579, CTG-B43a, CTG-B45d, Ronin', description='THAP domain containing 11', other_designations='', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105855103, end_position_on_the_genomic_accession=105856950, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  59020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59020, CurrentID=0, Status='live', Symbol='Pdzk1', Aliases='1700023D20Rik, 2610507N21Rik, 4921513F16Rik, AI267131, AI314638, AL022680, D3Ertd537e, Pdzd1, mPDZK1', description='PDZ domain containing 1', other_designations='CFTR-associated protein of 70 kDa|NHERF-3|Na(+)/H(+) exchanger regulatory factor 3|PDZ domain-containing protein 1|na/Pi cotransporter C-terminal-associated protein 1|naPi-Cap1|sodium-hydrogen exchanger regulatory factor 3', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96830065, end_position_on_the_genomic_accession=96870926, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  59021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59021, CurrentID=0, Status='live', Symbol='Rab2a', Aliases='9330148M11Rik, C80220, Rab2', description='RAB2A, member RAS oncogene family', other_designations='GTP-binding protein|RAB2, member RAS oncogene family', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=8535586, end_position_on_the_genomic_accession=8607702, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  59022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59022, CurrentID=0, Status='live', Symbol='Edf1', Aliases='0610008L11Rik, AA409425', description='endothelial differentiation-related factor 1', other_designations='EDF-1|MBF1|hypothetical protein 1-9|multiprotein-bridging factor 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25557822, end_position_on_the_genomic_accession=25562082, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  59024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59024, CurrentID=0, Status='live', Symbol='Med12', Aliases='230kDa, Mopa, OPA-1, Tnrc11, Trap230', description='mediator complex subunit 12', other_designations='OPA-containing protein 1|mediator of RNA polymerase II transcription, subunit 12 homolog|thyroid hormone receptor-associated protein complex 230 kDa component|trinucleotide repeat containing 11 (THR-associated protein)|trinucleotide repeat containing 11 (THR-associated protein, 230 kDa subunit)|trinucleotide repeat-containing gene 11 protein', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101274091, end_position_on_the_genomic_accession=101298934, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  59025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59025, CurrentID=0, Status='live', Symbol='Usp14', Aliases='2610005K12Rik, 2610037B11Rik, AW107924, C78769, ax, nmf375', description='ubiquitin specific peptidase 14', other_designations='ataxia|dUB-type TGT for deubiquitinating enzyme|deubiquitinating enzyme 14|ubiquitin specific protease 14|ubiquitin thioesterase 14|ubiquitin thiolesterase 14|ubiquitin-specific-processing protease 14', map_location='18 4.91 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=9993615, end_position_on_the_genomic_accession=10030149, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  59026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59026, CurrentID=0, Status='live', Symbol='Huwe1', Aliases='5430439H10Rik, AU041296, Arf-bp1, C430014N20Rik, C80292, Gm1718, Ib772, LASU1, Mule, Ureb1', description='HECT, UBA and WWE domain containing 1', other_designations='E3Histone|HECT, UBA and WWE domain-containing protein 1|Mcl-1 ubiquitin ligase|URE-B1|URE-binding protein 1|upstream regulatory element binding protein 1|upstream regulatory element-binding protein 1', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151800807, end_position_on_the_genomic_accession=151935417, orientation='plus', exon_count=90, OMIM='', no_hdr0=''),
  59027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59027, CurrentID=0, Status='live', Symbol='Nampt', Aliases='1110035O14Rik, AI314458, AI480535, NAmPRTase, Pbef, Pbef1, Visfatin', description='nicotinamide phosphoribosyltransferase', other_designations='pre-B-cell colony-enhancing factor 1 homolog', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=32820335, end_position_on_the_genomic_accession=32853369, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  59028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59028, CurrentID=0, Status='live', Symbol='Rcl1', Aliases='2310040A02Rik, AI789745, C76567, RPCL1, Rnac', description='RNA terminal phosphate cyclase-like 1', other_designations='RNA cyclase homolog', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29101375, end_position_on_the_genomic_accession=29143851, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59029, CurrentID=0, Status='live', Symbol='Psmd14', Aliases='2610312C03Rik, 3200001M20Rik, AA986732, Pad1, Poh1, rpm11', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 14', other_designations='26S proteasome non-ATPase subunit|26S proteasome regulatory subunit rpn11|26S proteasome-associated pad1 homolog|MAD1|proteasome 26S subunit, non-ATPase 14', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=61711694, end_position_on_the_genomic_accession=61800376, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  59030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59030, CurrentID=0, Status='live', Symbol='Mkks', Aliases='1300013E18Rik, AI463362, AI957237, Bbs6', description='McKusick-Kaufman syndrome', other_designations='McKusick-Kaufman syndrome protein|McKusick-Kaufman syndrome protein; MKKS protein|protein Bbs6 homolog', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=136873780, end_position_on_the_genomic_accession=136891505, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  59031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59031, CurrentID=0, Status='live', Symbol='Chst12', Aliases='AI595374, C4S-2, C4ST2', description='carbohydrate sulfotransferase 12', other_designations='chondroitin 4-O-sulfotransferase 2|chondroitin 4-sulfotransferase 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140505580, end_position_on_the_genomic_accession=140525238, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  59032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59032, CurrentID=0, Status='live', Symbol='Ppp2r3c', Aliases='4930511A21Rik, 5730412A08Rik, G4-1, G5pr', description="protein phosphatase 2, regulatory subunit B'', gamma", other_designations='phosphatase subunit gene g4-1|protein phosphatase subunit G5PR', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55280809, end_position_on_the_genomic_accession=55303005, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  59033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59033, CurrentID=0, Status='live', Symbol='Slc4a8', Aliases='AW493845, C230026C11, NDCBE, kNBC-3', description='solute carrier family 4 (anion exchanger), member 8', other_designations='electroneutral Na+-driven Cl-HCO3 exchanger|k-NBC3|slc4a8|sodium bicarbonate cotransporter isoform 3 kNBC-3|solute carrier family 4 member 8', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100761700, end_position_on_the_genomic_accession=100823971, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  59035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59035, CurrentID=0, Status='live', Symbol='Carm1', Aliases='Prmt4', description='coactivator-associated arginine methyltransferase 1', other_designations='protein arginine N-methyltransferase 4|protein arginine methyltransferase', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21546894, end_position_on_the_genomic_accession=21589467, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  59036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59036, CurrentID=0, Status='live', Symbol='Dact1', Aliases='4921528D17Rik, AI115603, DAPPER, DAPPER1, FRODO, Frd1, Frodo1, MDpr1, MTNG3, THYEX3', description='dishevelled-binding antagonist of beta-catenin 1', other_designations='dapper 1|dapper antagonist of catenin 1|dapper homolog 1, antagonist of beta-catenin|frodo homolog|thymus expressed gene 3|thymus-expressed novel gene 3 protein', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=71309884, end_position_on_the_genomic_accession=71320107, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  59038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59038, CurrentID=0, Status='live', Symbol='Pxmp4', Aliases='3010018P03Rik, Pmp24', description='peroxisomal membrane protein 4', other_designations='24 kDa intrinsic membrane protein|24 kDa peroxisomal intrinsic membrane protein', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154587044, end_position_on_the_genomic_accession=154603673, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59040, CurrentID=0, Status='live', Symbol='Rhot1', Aliases='2210403N23Rik, AA415293, AF244542, AI834919, Arht1, C430039G08Rik, Miro1', description='ras homolog gene family, member T1', other_designations='MIRO-1', map_location='11 47.62 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80209005, end_position_on_the_genomic_accession=80267907, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  59041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59041, CurrentID=0, Status='live', Symbol='Stk25', Aliases='1500019J11Rik, AU018434, C86992, Ste20-like, Ysk1', description='serine/threonine kinase 25 (yeast)', other_designations='SOK-1|Ste20-like kinase|Yeast Sps1/Ste20-related kinase 1|serine/threonine kinase 25 (Ste20, yeast homolog)|ste20/oxidant stress response kinase 1|sterile 20/oxidant stress-response kinase 1', map_location='1 47.25 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93620751, end_position_on_the_genomic_accession=93654216, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  59042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59042, CurrentID=0, Status='live', Symbol='Cope', Aliases='1110005D17Rik1, Cope', description='coatomer protein complex, subunit epsilon', other_designations='epsilon-COP|epsilon-coat protein|nonclathrin coat protein epsilon-COP', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70302785, end_position_on_the_genomic_accession=70312990, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59043, CurrentID=0, Status='live', Symbol='Wsb2', Aliases='AA673511, Swip2, WSB-2', description='WD repeat and SOCS box-containing 2', other_designations='SOCS box-containing WD protein SWiP-2|WD-40-repeat-containing protein with a SOCS box 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117357305, end_position_on_the_genomic_accession=117378589, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  59044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59044, CurrentID=0, Status='live', Symbol='Rnf130', Aliases='2510042A13Rik, G1RZFP, G1rp, GOLIATH, GP', description='ring finger protein 130', other_designations='g1-related zinc finger protein|goliath homolog|rnf130', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50025331, end_position_on_the_genomic_accession=50125719, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59045, CurrentID=0, Status='live', Symbol='Stard3', Aliases='Mln64, es64', description='START domain containing 3', other_designations='START domain-containing protein 3|protein ES 64|protein MLN 64|steroidogenic acute regulatory protein related', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98358367, end_position_on_the_genomic_accession=98381112, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  59046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59046, CurrentID=0, Status='live', Symbol='Arpp19', Aliases='19kDa, 2700024H10Rik, ARPP-19, AW559096, Arpp12', description='cAMP-regulated phosphoprotein 19', other_designations='cAMP-regulated phosphoprotein ARPP-19|cyclic AMP phosphoprotein, 19 kDa|cyclic AMP phosphoprotein, 19kD', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75037614, end_position_on_the_genomic_accession=75060313, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  59047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59047, CurrentID=0, Status='live', Symbol='Pnkp', Aliases='1810009G08Rik, PNK', description="polynucleotide kinase 3'- phosphatase", other_designations="DNA 5'-kinase/3'-phosphatase|polynucleotide kinase-3'-phosphatase", map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44857055, end_position_on_the_genomic_accession=44862929, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  59048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59048, CurrentID=0, Status='live', Symbol='C1galt1c1', Aliases='1500002I11Rik, AB030184, C1GalT2, C81205, Cosmc', description='C1GALT1-specific chaperone 1', other_designations='C1Gal-T2|beta1,3-galactosyltransferase 2|core 1 beta1,3-galactosyltransferase 2|core 1 beta3-Gal-T2|core 1 beta3-galactosyltransferase-specific molecular chaperone|mC1Gal-T2', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38630783, end_position_on_the_genomic_accession=38635143, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  59049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59049, CurrentID=0, Status='live', Symbol='Slc22a17', Aliases='1700094C23Rik, 24p3R, AU041908, AW555662, BOIT, Boct, mBOCT', description='solute carrier family 22 (organic cation transporter), member 17', other_designations='24p3 receptor|brain-type organic cation transporter|lipocalin-2 receptor|organic cation transporter BOCT', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54906727, end_position_on_the_genomic_accession=54913139, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  59050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59050, CurrentID=0, Status='live', Symbol='Nsa2', Aliases='5730427N09Rik, LNR42p, Tinp1, Yr29, Nsa2', description='NSA2 ribosome biogenesis homolog (S. cerevisiae)', other_designations='L-name related protein|L-name-related protein 42|TGF-beta-inducible nuclear protein 1', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=97129427, end_position_on_the_genomic_accession=97137926, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  59052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59052, CurrentID=0, Status='live', Symbol='Mettl9', Aliases='0610012D09Rik, AA517660, Drev, MNCb-5680', description='methyltransferase like 9', other_designations='DORA reverse strand protein', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121032794, end_position_on_the_genomic_accession=121076835, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  59053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59053, CurrentID=0, Status='live', Symbol='Hgh1', Aliases='Brp16, D15Ertd741e, Fam203a, MNCb-5873', description='HGH1 homolog', other_designations='brain protein 16|family with sequence similarity 203, member A|protein FAM203A', map_location='15 35.78 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76368860, end_position_on_the_genomic_accession=76371411, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  59054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59054, CurrentID=0, Status='live', Symbol='Mrps30', Aliases='2610020A16Rik, AA968347, PAP, Pdcd9', description='mitochondrial ribosomal protein S30', other_designations='MRP-S30|S30mt|programmed cell death 9', map_location='13 D2.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=118380110, end_position_on_the_genomic_accession=118387252, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  59056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59056, CurrentID=0, Status='live', Symbol='Evc', Aliases='', description='Ellis van Creveld gene syndrome', other_designations='Ellis van Creveld gene homolog|EvC protein (Ellis-van Creveld syndrome)', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37288977, end_position_on_the_genomic_accession=37337242, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  59057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59057, CurrentID=0, Status='live', Symbol='Zfp24', Aliases='3526401F17Rik, 5033419P20Rik, AI480505, AI642085, AL024422, KOX17, ZF-12, Zfp191, Znf191, Znf24, Znp191', description='zinc finger protein 24', other_designations='hypomyelinated CNS protein|zfp-191|zinc finger ZF-12|zinc finger protein 191|zinc finger protein ZF-12', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24009717, end_position_on_the_genomic_accession=24022219, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  59058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59058, CurrentID=0, Status='live', Symbol='Bhlhe22', Aliases='Beta3, Bhlhb5', description='basic helix-loop-helix family, member e22', other_designations='basic helix-loop-helix (bHLH) gene, class B, beta3|basic helix-loop-helix domain containing, class B5|class B basic helix-loop-helix protein 5', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=18054325, end_position_on_the_genomic_accession=18057514, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  59069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59069, CurrentID=0, Status='live', Symbol='Tpm3', Aliases='TM30nm, TMnm, Tm5NM, Tpm-5, Tpm5, Trop-5, gamma-TM, hTM30nm, hTMnm', description='tropomyosin 3, gamma', other_designations='gamma-tropomyosin|skalphaTM.2|tropomyosin 5|tropomyosin-3', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90072651, end_position_on_the_genomic_accession=90100902, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  59079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59079, CurrentID=0, Status='live', Symbol='Erbb2ip', Aliases='1700028E05Rik, Erbin, mKIAA1225', description='Erbb2 interacting protein', other_designations='densin-180-like protein|erbb2-interacting protein', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=103818786, end_position_on_the_genomic_accession=103920586, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  59083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59083, CurrentID=0, Status='live', Symbol='Fetub', Aliases='2310011O17Rik, AI255764, D17980, Gugu', description='fetuin beta', other_designations='IRL685|fetuin-like protein IRL685', map_location='16 B', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22918382, end_position_on_the_genomic_accession=22939768, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  59090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59090, CurrentID=0, Status='live', Symbol='Midn', Aliases='3000003C15Rik', description='midnolin', other_designations='midbrain nucleolar protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80148227, end_position_on_the_genomic_accession=80158368, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  59091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59091, CurrentID=0, Status='live', Symbol='Jph2', Aliases='1110002E14Rik, JP-2, Jp2', description='junctophilin 2', other_designations='junctophilin type 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163336242, end_position_on_the_genomic_accession=163398049, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  59092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59092, CurrentID=0, Status='live', Symbol='Pcbp4', Aliases='1200003L19Rik, AlphaCP-4', description='poly(rC) binding protein 4', other_designations='alpha-CP4', map_location='9 F1-F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106453406, end_position_on_the_genomic_accession=106464012, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  59093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59093, CurrentID=0, Status='live', Symbol='Pcbp3', Aliases='AlphaCP-3', description='poly(rC) binding protein 3', other_designations='alpha-CP3', map_location='10 B5.1-B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76761854, end_position_on_the_genomic_accession=76961953, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  59095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59095, CurrentID=0, Status='live', Symbol='Fxyd6', Aliases='0610030I18Rik, Php', description='FXYD domain-containing ion transport regulator 6', other_designations='PLM-like protein|phosphohippolin', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45370185, end_position_on_the_genomic_accession=45396159, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  59125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59125, CurrentID=0, Status='live', Symbol='Nek7', Aliases='2810460C19Rik, AU020186', description='NIMA (never in mitosis gene a)-related expressed kinase 7', other_designations='NIMA-related expressed kinase 7|NIMA-related kinase 7|never in mitosis A-related kinase 7|nimA-related protein kinase 7', map_location='1 60.82 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=138484106, end_position_on_the_genomic_accession=138619779, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  59126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59126, CurrentID=0, Status='live', Symbol='Nek6', Aliases='1300007C09Rik', description='NIMA (never in mitosis gene a)-related expressed kinase 6', other_designations='NIMA-related expressed kinase 6|never in mitosis A-related kinase 6|nimA-related protein kinase 6', map_location='2 24.41 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38511697, end_position_on_the_genomic_accession=38587490, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  59287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59287, CurrentID=0, Status='live', Symbol='Ncstn', Aliases='9430068N19Rik, AA727311, Aph2, D1Dau13e, Kiaa0253, Nct, mKIAA0253', description='nicastrin', other_designations='', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172066013, end_position_on_the_genomic_accession=172082749, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  59288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59288, CurrentID=0, Status='live', Symbol='Dctn5', Aliases='4930427E12Rik, C78178, b2b315Clo', description='dynactin 5', other_designations='dynactin 4|dynactin subunit p25|p25 dynactin subunit', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122133041, end_position_on_the_genomic_accession=122149044, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  59289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59289, CurrentID=0, Status='live', Symbol='Ackr2', Aliases='AI464239, CCR10, CCR9, Ccbp2, Cmkbr9, D6', description='atypical chemokine receptor 2', other_designations='C-C chemokine receptor D6|D6 beta-chemokine receptor|chemokine binding protein 2|chemokine decoy receptor D6|chemokine-binding protein 2|chemokine-binding protein D6', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121898355, end_position_on_the_genomic_accession=121911071, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  59290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59290, CurrentID=0, Status='live', Symbol='Gpa33', Aliases='2010310L10Rik, 2210401D16Rik, BB116197', description='glycoprotein A33 (transmembrane)', other_designations='A33 antigen|mA33', map_location='1 73.49 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166130460, end_position_on_the_genomic_accession=166166510, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  59308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59308, CurrentID=0, Status='live', Symbol='Emcn', Aliases='0610012K22Rik, AI315669, Muc14', description='endomucin', other_designations='MUC-14|endomucin-1|endomucin-1/2|mucin 14, endothelial|mucin-14', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=137341061, end_position_on_the_genomic_accession=137431069, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  59310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=59310, CurrentID=0, Status='live', Symbol='Myl10', Aliases='1700027I08Rik, Mylc2pl, PLRLC', description='myosin, light chain 10, regulatory', other_designations='myosin light chain 2, precursor lymphocyte-specific|precursor lymphocyte-specific regulatory light chain', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136693146, end_position_on_the_genomic_accession=136701094, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  228359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228359, CurrentID=0, Status='live', Symbol='Arhgap1', Aliases='B230365D05Rik, C76222, Cdc42GAP, p50rhoGAP', description='Rho GTPase activating protein 1', other_designations='rho-type GTPase-activating protein 1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91650011, end_position_on_the_genomic_accession=91672325, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  75553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75553, CurrentID=0, Status='live', Symbol='Zc3h14', Aliases='1010001P15Rik, 1700016A15Rik, 2700069A02Rik, AU014748', description='zinc finger CCCH type containing 14', other_designations='zinc finger CCCH-type containing 14', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98746968, end_position_on_the_genomic_accession=98787774, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  60315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60315, CurrentID=0, Status='live', Symbol='Myg1', Aliases='0610023A07Rik, 2810433J21Rik, 5830429P19Rik, AI325965, Gamm1', description='melanocyte proliferating gene 1', other_designations='protein Gamm1', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102331709, end_position_on_the_genomic_accession=102338139, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  60321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60321, CurrentID=0, Status='live', Symbol='Wbp11', Aliases='2510026P17Rik, D6Wsu113e, Npwbp, SIPP1', description='WW domain binding protein 11', other_designations='Npw38-binding protein|WBP-11|splicing factor that interacts with PQBP-1 and PP1', map_location='6 66.72 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136813654, end_position_on_the_genomic_accession=136828216, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  60322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60322, CurrentID=0, Status='live', Symbol='Chst7', Aliases='2600013M07Rik, C6ST-2, GST5, Gn6st-4, glcNAc6ST-4', description='carbohydrate (N-acetylglucosamino) sulfotransferase 7', other_designations='N-acetylglucosamine 6-O-sulfotransferase 1|N-acetylglucosamine 6-O-sulfotransferase 4|chondroitin 6-sulfotransferase 2|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 5', map_location='X A3.1-A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20059437, end_position_on_the_genomic_accession=20099321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  60344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60344, CurrentID=0, Status='live', Symbol='Fign', Aliases='fi, fidget', description='fidgetin', other_designations='', map_location='2 37.19 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=63971508, end_position_on_the_genomic_accession=64098038, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  60345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60345, CurrentID=0, Status='live', Symbol='Nrip2', Aliases='AW491344, NIX1', description='nuclear receptor interacting protein 2', other_designations='neuronal interacting factor 1|neuronal-interacting factor X 1', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128397625, end_position_on_the_genomic_accession=128409784, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  60361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60361, CurrentID=0, Status='live', Symbol='Ms4a4b', Aliases='AI463180, Chandra, Ly116', description='membrane-spanning 4-domains, subfamily A, member 4B', other_designations='lymphocyte antigen 116', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11443558, end_position_on_the_genomic_accession=11463549, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  60363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60363, CurrentID=0, Status='live', Symbol='Cldn15', Aliases='2210009B08Rik, BB107105', description='claudin 15', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136967869, end_position_on_the_genomic_accession=136975845, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  60364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60364, CurrentID=0, Status='live', Symbol='Donson', Aliases='1110025J21Rik, AI845729, ORF60', description='downstream neighbor of SON', other_designations='protein 3SG', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91666856, end_position_on_the_genomic_accession=91688759, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  60365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60365, CurrentID=0, Status='live', Symbol='Rbm8a', Aliases='2310057C03Rik, AA673428, Rbm8', description='RNA binding motif protein 8a', other_designations='RNA-binding motif protein 8A|ribonucleoprotein RBM8A', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96629928, end_position_on_the_genomic_accession=96633791, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  60367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60367, CurrentID=0, Status='live', Symbol='Il1rapl2', Aliases='', description='interleukin 1 receptor accessory protein-like 2', other_designations='IL-1 receptor accessory protein-like 2|IL-1-RAPL-2|IL-1RAPL-2|IL1RAPL-2|IL1RAPL-2-related protein|TIGIRR-1|three immunoglobulin domain-containing IL-1 receptor-related 1', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=137571549, end_position_on_the_genomic_accession=138846955, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  60406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60406, CurrentID=0, Status='live', Symbol='Sap30', Aliases='30kDa', description='sin3 associated polypeptide', other_designations='30 kDa Sin3-associated polypeptide|Sin3-associated protein|sap30|sin3 associated polypeptide, 30kD|sin3 corepressor complex subunit SAP30|sin3-associated polypeptide p30|sin3-associated polypeptide, 30 kDa', map_location='8 29.85 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=57482702, end_position_on_the_genomic_accession=57487860, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  60409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60409, CurrentID=0, Status='live', Symbol='Trappc4', Aliases='1500017G03Rik, AI303617, HSPC172, PTD009, Sbd, Sbdn, Sdcbp2, TRS23', description='trafficking protein particle complex 4', other_designations='TRS23 homolog|synbindin|syndecan binding protein 2', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44403759, end_position_on_the_genomic_accession=44407563, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  60411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60411, CurrentID=0, Status='live', Symbol='Cenpk', Aliases='B130045K24Rik, BB232189, C530004N04Rik, Cenp-K, Solt, Solzt', description='centromere protein K', other_designations='SoxLZ/Sox6 leucine zipper binding protein in testis', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104228611, end_position_on_the_genomic_accession=104249622, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  60425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60425, CurrentID=0, Status='live', Symbol='Doc2g', Aliases='D830013O18Rik', description='double C2, gamma', other_designations='doc2-gamma', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4003385, end_position_on_the_genomic_accession=4007005, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  60440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60440, CurrentID=0, Status='live', Symbol='Iigp1', Aliases='2900074L10Rik, AI046432, AW111922, Ifgga1, Iigp, Irga6', description='interferon inducible GTPase 1', other_designations='interferon-gamma-inducible GTPase Ifgga1 protein', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60373632, end_position_on_the_genomic_accession=60392634, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  60441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60441, CurrentID=0, Status='live', Symbol='Mrpl38', Aliases='1110036N21Rik, 4733401F03Rik, MRP-L3, R75240, Rpml3', description='mitochondrial ribosomal protein L38', other_designations='L38mt|MRP-L38|ribosomal protein, mitochondrial, L3', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116131817, end_position_on_the_genomic_accession=116138868, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  60455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60455, CurrentID=0, Status='live', Symbol='Tmem8', Aliases='AW490096, M83a, Tmem8', description='transmembrane protein 8 (five membrane-spanning domains)', other_designations='transmembrane protein 6|type I transmembrane protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26112691, end_position_on_the_genomic_accession=26123254, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  60504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60504, CurrentID=0, Status='live', Symbol='Il21r', Aliases='NILR', description='interleukin 21 receptor', other_designations='IL-21 receptor|IL-21R|LR-beta|NR8|lymphocyte receptor beta|novel cytokine receptor 8|novel interleukin receptor', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125603253, end_position_on_the_genomic_accession=125634397, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  191577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=191577, CurrentID=0, Status='live', Symbol='cir', Aliases='', description='circling', other_designations='circling mouse mutation', map_location='9', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  191578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=191578, CurrentID=0, Status='live', Symbol='Helq', Aliases='AU021725, D430018E21Rik, Hel308', description='helicase, POLQ-like', other_designations='DNA helicase Hel308|POLQ-like helicase|helicase, mus308-like|mus308-like helicase', map_location='5 E', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100762144, end_position_on_the_genomic_accession=100798663, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  60507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60507, CurrentID=0, Status='live', Symbol='Qtrt1', Aliases='2610028E17Rik, Tgt', description='queuine tRNA-ribosyltransferase 1', other_designations='guanine insertion enzyme|tRNA-guanine transglycosylase', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21411572, end_position_on_the_genomic_accession=21420279, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  60510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60510, CurrentID=0, Status='live', Symbol='Syt9', Aliases='', description='synaptotagmin IX', other_designations='synaptotagmin 9|synaptotagmin V|sytIX', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107370704, end_position_on_the_genomic_accession=107548656, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  60525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60525, CurrentID=0, Status='live', Symbol='Acss2', Aliases='1110017C11Rik, ACAS, Acas1, Acas2, AceCS1, Acs1', description='acyl-CoA synthetase short-chain family member 2', other_designations='ACS|aceCS|acetate--CoA ligase|acetyl-CoA synthetase 1|acetyl-Coenzyme A synthetase 1 (AMP forming)|acetyl-Coenzyme A synthetase 2 (ADP forming)|acetyl-Coenzyme A synthetase 2 (AMP forming)|acyl-activating enzyme', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155517953, end_position_on_the_genomic_accession=155562746, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  60527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60527, CurrentID=0, Status='live', Symbol='Fads3', Aliases='AI464531', description='fatty acid desaturase 3', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10040855, end_position_on_the_genomic_accession=10059671, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  60530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60530, CurrentID=0, Status='live', Symbol='Fignl1', Aliases='', description='fidgetin-like 1', other_designations='', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=11800288, end_position_on_the_genomic_accession=11808995, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  60531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60531, CurrentID=0, Status='live', Symbol='Npvf', Aliases='Rfrp', description='neuropeptide VF precursor', other_designations='FMRFamide-related peptides|RFamide related peptide|RFamide-related peptide', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=50650871, end_position_on_the_genomic_accession=50654393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  60532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60532, CurrentID=0, Status='live', Symbol='Wtap', Aliases='2810408K05Rik, 9430038B09Rik', description='Wilms tumour 1-associating protein', other_designations="WT1-associated protein|Wilms' tumour 1-associating protein|female-lethal(2)D homolog|wilms tumor 1-associating protein", map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12966798, end_position_on_the_genomic_accession=12992808, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  60533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60533, CurrentID=0, Status='live', Symbol='Cd274', Aliases='A530045L16Rik, B7h1, Pdcd1l1, Pdcd1lg1, Pdl1', description='CD274 antigen', other_designations='B7 homolog 1|PDCD1 ligand 1', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29367438, end_position_on_the_genomic_accession=29388095, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  60534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60534, CurrentID=0, Status='live', Symbol='Fancg', Aliases='AU041407, Xrcc9', description='Fanconi anemia, complementation group G', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43002337, end_position_on_the_genomic_accession=43010568, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  67484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67484, CurrentID=0, Status='live', Symbol='Eepd1', Aliases='2310005P05Rik, AI481005', description='endonuclease/exonuclease/phosphatase family domain containing 1', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=25481526, end_position_on_the_genomic_accession=25604110, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  60594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60594, CurrentID=0, Status='live', Symbol='Capn12', Aliases='CANP 12', description='calpain 12', other_designations='calcium-activated neutral proteinase 12', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28881446, end_position_on_the_genomic_accession=28893585, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  60595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60595, CurrentID=0, Status='live', Symbol='Actn4', Aliases='C77391', description='actinin alpha 4', other_designations='F-actin cross-linking protein|non-muscle alpha-actinin 4', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28893248, end_position_on_the_genomic_accession=28962352, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  60596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60596, CurrentID=0, Status='live', Symbol='Gucy1a3', Aliases='1200016O07Rik, sGC-alpha1', description='guanylate cyclase 1, soluble, alpha 3', other_designations='GCS-alpha-1|GCS-alpha-3|alpha 1 sGC|soluble guanylate cyclase large subunit', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=82092427, end_position_on_the_genomic_accession=82145877, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  60597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60597, CurrentID=0, Status='live', Symbol='Mapk8ip2', Aliases='3230402N03Rik, AI847694, AW049958, IB2, Jip2, R75148', description='mitogen-activated protein kinase 8 interacting protein 2', other_designations='IB-2|JIP-2|JNK MAP kinase scaffold protein 2|JNK-interacting protein 2|islet-brain-2|mitogen activated protein kinase 8 interacting protein 2|mitogen-activated protein kinase 8-interacting protein 2', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89453911, end_position_on_the_genomic_accession=89462447, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  60599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60599, CurrentID=0, Status='live', Symbol='Trp53inp1', Aliases='2700057G22Rik, SIP, SIP18, SIP27, Stinp, Teap', description='transformation related protein 53 inducible nuclear protein 1', other_designations='p53-dependent damage-inducible nuclear protein 1|p53DINP1|stress induced protein|stress-induced protein|thymus expressed acidic protein|thymus-expressed acidic protein', map_location='4 A1-A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11156441, end_position_on_the_genomic_accession=11174377, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  60611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60611, CurrentID=0, Status='live', Symbol='Foxj2', Aliases='Fhx', description='forkhead box J2', other_designations='fork head homologous X', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122820184, end_position_on_the_genomic_accession=122845366, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  60613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60613, CurrentID=0, Status='live', Symbol='Kcnq4', Aliases='', description='potassium voltage-gated channel, subfamily Q, member 4', other_designations='KQT-like 4|potassium channel subunit alpha KvLQT4|voltage-gated potassium channel subunit Kv7.4', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120696134, end_position_on_the_genomic_accession=120747491, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  60690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60690, CurrentID=0, Status='live', Symbol='D4Mit1', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 1', other_designations='', map_location='4 7.09 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60691, CurrentID=0, Status='live', Symbol='D1Mit3', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 3', other_designations='', map_location='1 6.27 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60692, CurrentID=0, Status='live', Symbol='D1Mit7', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 7', other_designations='', map_location='1 38.55 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60693, CurrentID=0, Status='live', Symbol='D2Mit8', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 8', other_designations='', map_location='2 30.84 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60694, CurrentID=0, Status='live', Symbol='D4Mit7', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 7', other_designations='', map_location='4 34.92 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60695, CurrentID=0, Status='live', Symbol='D6Mit9', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 9', other_designations='', map_location='6 37.96 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60696, CurrentID=0, Status='live', Symbol='D8Mit5', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 5', other_designations='', map_location='8 24.7 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60697, CurrentID=0, Status='live', Symbol='DXMit5', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 5', other_designations='', map_location='X 74.91 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60698, CurrentID=0, Status='live', Symbol='D1Mit58', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 58', other_designations='', map_location='1 2.08 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60699, CurrentID=0, Status='live', Symbol='D1Mit64', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 64', other_designations='', map_location='1 3.67 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60702, CurrentID=0, Status='live', Symbol='D1Mit67', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 67', other_designations='', map_location='1 4.12 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60705, CurrentID=0, Status='live', Symbol='D3Mit62', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 62', other_designations='', map_location='3 4.85 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60708, CurrentID=0, Status='live', Symbol='D5Mit72', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 72', other_designations='', map_location='5 10.06 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60709, CurrentID=0, Status='live', Symbol='D6Mit47', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 47', other_designations='', map_location='6 12.88 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60712, CurrentID=0, Status='live', Symbol='D6Mit86', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 86', other_designations='', map_location='6 1.81 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60713, CurrentID=0, Status='live', Symbol='D7Mit21', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 21', other_designations='', map_location='7 1.91 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60714, CurrentID=0, Status='live', Symbol='D7Mit56', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 56', other_designations='', map_location='7 9.15 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60715, CurrentID=0, Status='live', Symbol='D7Mit57', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 57', other_designations='', map_location='7 9.94 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60716, CurrentID=0, Status='live', Symbol='D8Mit58', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 58', other_designations='', map_location='8 2.52 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60723, CurrentID=0, Status='live', Symbol='D9Mit89', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 89', other_designations='', map_location='9 14.79 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60724, CurrentID=0, Status='live', Symbol='D9Mit90', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 90', other_designations='', map_location='9 17.8 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60725, CurrentID=0, Status='live', Symbol='DXMit26', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 26', other_designations='', map_location='X 3.43 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60727, CurrentID=0, Status='live', Symbol='DXMit54', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 54', other_designations='', map_location='X 6.64 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60728, CurrentID=0, Status='live', Symbol='DXMit57', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 57', other_designations='', map_location='X 16.24 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60729, CurrentID=0, Status='live', Symbol='DXMit85', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 85', other_designations='', map_location='X 8.3 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60730, CurrentID=0, Status='live', Symbol='DXMit89', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 89', other_designations='', map_location='X 4.73 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60731, CurrentID=0, Status='live', Symbol='D10Mit1', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 1', other_designations='', map_location='10 4.25 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60732, CurrentID=0, Status='live', Symbol='D1Mit18', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 18', other_designations='', map_location='1 26.99 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60733, CurrentID=0, Status='live', Symbol='D1Mit19', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 19', other_designations='', map_location='1 38.01 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60734, CurrentID=0, Status='live', Symbol='D1Mit24', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 24', other_designations='', map_location='1 38.54 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60736, CurrentID=0, Status='live', Symbol='D1Mit45', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 45', other_designations='', map_location='1 46.74 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60737, CurrentID=0, Status='live', Symbol='D1Mit46', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 46', other_designations='', map_location='1 39.16 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60738, CurrentID=0, Status='live', Symbol='D1Mit49', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 49', other_designations='', map_location='1 44.07 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60739, CurrentID=0, Status='live', Symbol='D1Mit57', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 57', other_designations='', map_location='1 74.07 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60742, CurrentID=0, Status='live', Symbol='D1Mit74', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 74', other_designations='', map_location='1 28.81 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60743, CurrentID=0, Status='live', Symbol='D1Mit76', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 76', other_designations='', map_location='1 33.31 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60744, CurrentID=0, Status='live', Symbol='D1Mit77', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 77', other_designations='', map_location='1 38.01 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60745, CurrentID=0, Status='live', Symbol='D1Mit84', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 84', other_designations='', map_location='1 46.24 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60749, CurrentID=0, Status='live', Symbol='D1Mit94', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 94', other_designations='', map_location='1 54.7 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60751, CurrentID=0, Status='live', Symbol='D2Mit22', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 22', other_designations='', map_location='2 74.83 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60752, CurrentID=0, Status='live', Symbol='D2Mit32', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 32', other_designations='', map_location='2 17.18 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60754, CurrentID=0, Status='live', Symbol='D2Mit41', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 41', other_designations='', map_location='2 51.21 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60755, CurrentID=0, Status='live', Symbol='D2Mit49', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 49', other_designations='', map_location='2 83.83 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60756, CurrentID=0, Status='live', Symbol='D2Mit51', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 51', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60757, CurrentID=0, Status='live', Symbol='D2Mit52', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 52', other_designations='', map_location='2 87.22 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60760, CurrentID=0, Status='live', Symbol='D2Mit56', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 56', other_designations='', map_location='2 41.97 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60761, CurrentID=0, Status='live', Symbol='D2Mit57', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 57', other_designations='', map_location='2 73.42 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60762, CurrentID=0, Status='live', Symbol='D2Mit58', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 58', other_designations='', map_location='2 56.21 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60763, CurrentID=0, Status='live', Symbol='D2Mit61', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 61', other_designations='', map_location='2 34.9 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60764, CurrentID=0, Status='live', Symbol='D2Mit62', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 62', other_designations='', map_location='2 59.34 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60765, CurrentID=0, Status='live', Symbol='D2Mit65', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 65', other_designations='', map_location='2 22.09 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60772, CurrentID=0, Status='live', Symbol='D2Mit92', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 92', other_designations='', map_location='2 42.74 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60773, CurrentID=0, Status='live', Symbol='D2Mit94', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 94', other_designations='', map_location='2 47.93 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60774, CurrentID=0, Status='live', Symbol='D2Mit97', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 97', other_designations='', map_location='2 54.43 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60775, CurrentID=0, Status='live', Symbol='D3Mit10', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 10', other_designations='', map_location='3 43.03 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60776, CurrentID=0, Status='live', Symbol='D3Mit12', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 12', other_designations='', map_location='3 43.71 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60777, CurrentID=0, Status='live', Symbol='D3Mit14', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 14', other_designations='', map_location='3 61.32 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60778, CurrentID=0, Status='live', Symbol='D3Mit17', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 17', other_designations='', map_location='3 67.53 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60779, CurrentID=0, Status='live', Symbol='D3Mit19', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 19', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60785, CurrentID=0, Status='live', Symbol='D3Mit42', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 42', other_designations='', map_location='3 52.94 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60786, CurrentID=0, Status='live', Symbol='D3Mit43', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 43', other_designations='', map_location='3 52.94 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60787, CurrentID=0, Status='live', Symbol='D3Mit44', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 44', other_designations='', map_location='3 72.43 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60788, CurrentID=0, Status='live', Symbol='D3Mit45', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 45', other_designations='', map_location='3 72.43 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60797, CurrentID=0, Status='live', Symbol='D3Mit75', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 75', other_designations='', map_location='3 43.73 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60798, CurrentID=0, Status='live', Symbol='D3Mit76', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 76', other_designations='', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60803, CurrentID=0, Status='live', Symbol='D4Mit14', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 14', other_designations='', map_location='4 80.99 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60804, CurrentID=0, Status='live', Symbol='D4Mit16', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 16', other_designations='', map_location='4 61.67 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60805, CurrentID=0, Status='live', Symbol='D4Mit29', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 29', other_designations='', map_location='4 48.18 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60806, CurrentID=0, Status='live', Symbol='D4Mit31', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 31', other_designations='', map_location='4 50.04 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60807, CurrentID=0, Status='live', Symbol='D4Mit33', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 33', other_designations='', map_location='4 81.43 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60808, CurrentID=0, Status='live', Symbol='D4Mit38', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 38', other_designations='', map_location='4 24.82 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60809, CurrentID=0, Status='live', Symbol='D4Mit39', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 39', other_designations='', map_location='4 16.92 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60811, CurrentID=0, Status='live', Symbol='D4Mit42', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 42', other_designations='', map_location='4 82.64 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60813, CurrentID=0, Status='live', Symbol='D4Mit46', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 46', other_designations='', map_location='4 48.18 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60816, CurrentID=0, Status='live', Symbol='D4Mit54', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 54', other_designations='', map_location='4 70.02 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60817, CurrentID=0, Status='live', Symbol='D4Mit58', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 58', other_designations='', map_location='4 45.66 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60823, CurrentID=0, Status='live', Symbol='D4Mit68', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 68', other_designations='', map_location='4 69.44 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60824, CurrentID=0, Status='live', Symbol='D4Mit69', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 69', other_designations='', map_location='4 68.55 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60825, CurrentID=0, Status='live', Symbol='D4Mit70', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 70', other_designations='', map_location='4 67.01 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60826, CurrentID=0, Status='live', Symbol='D4Mit71', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 71', other_designations='', map_location='4 66.32 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60827, CurrentID=0, Status='live', Symbol='D4Mit72', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 72', other_designations='', map_location='4 62.53 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60831, CurrentID=0, Status='live', Symbol='D4Mit79', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 79', other_designations='', map_location='4 39.4 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60832, CurrentID=0, Status='live', Symbol='D4Mit80', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 80', other_designations='', map_location='4 37.89 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60833, CurrentID=0, Status='live', Symbol='D4Mit81', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 81', other_designations='', map_location='4 37.89 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60834, CurrentID=0, Status='live', Symbol='D4Mit82', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 82', other_designations='', map_location='4 34.06 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60838, CurrentID=0, Status='live', Symbol='D5Mit10', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 10', other_designations='', map_location='5 50.64 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60839, CurrentID=0, Status='live', Symbol='D5Mit11', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 11', other_designations='', map_location='5 26.02 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60840, CurrentID=0, Status='live', Symbol='D5Mit20', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 20', other_designations='', map_location='5 47.77 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60841, CurrentID=0, Status='live', Symbol='D5Mit41', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 41', other_designations='', map_location='5 50.68 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60844, CurrentID=0, Status='live', Symbol='D5Mit58', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 58', other_designations='', map_location='5 38.44 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60845, CurrentID=0, Status='live', Symbol='D5Mit62', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 62', other_designations='', map_location='5 81.9 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60846, CurrentID=0, Status='live', Symbol='D5Mit63', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 63', other_designations='', map_location='5 71.21 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60847, CurrentID=0, Status='live', Symbol='D5Mit65', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 65', other_designations='', map_location='5 62.96 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60848, CurrentID=0, Status='live', Symbol='D5Mit66', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 66', other_designations='', map_location='5 17.27 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60851, CurrentID=0, Status='live', Symbol='D5Mit73', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 73', other_designations='', map_location='5 12.88 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60852, CurrentID=0, Status='live', Symbol='D5Mit75', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 75', other_designations='', map_location='5 19.14 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60853, CurrentID=0, Status='live', Symbol='D5Mit89', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 89', other_designations='', map_location='5 48.47 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60856, CurrentID=0, Status='live', Symbol='D6Mit14', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 14', other_designations='', map_location='6 77.64 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60857, CurrentID=0, Status='live', Symbol='D6Mit15', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 15', other_designations='', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60858, CurrentID=0, Status='live', Symbol='D6Mit22', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 22', other_designations='', map_location='6 37.15 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60859, CurrentID=0, Status='live', Symbol='D6Mit29', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 29', other_designations='', map_location='6 37.75 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60860, CurrentID=0, Status='live', Symbol='D6Mit33', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 33', other_designations='', map_location='6 24.42 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60864, CurrentID=0, Status='live', Symbol='D6Mit54', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 54', other_designations='', map_location='6 52.14 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60865, CurrentID=0, Status='live', Symbol='D6Mit55', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 55', other_designations='', map_location='6 53.05 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60866, CurrentID=0, Status='live', Symbol='D6Mit57', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 57', other_designations='', map_location='6 77.37 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60867, CurrentID=0, Status='live', Symbol='D6Mit58', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 58', other_designations='', map_location='6 73.86 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60868, CurrentID=0, Status='live', Symbol='D6Mit59', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 59', other_designations='', map_location='6 69.36 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60869, CurrentID=0, Status='live', Symbol='D6Mit63', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 63', other_designations='', map_location='6 53.05 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60877, CurrentID=0, Status='live', Symbol='D7Mit15', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 15', other_designations='', map_location='7 88.64 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60878, CurrentID=0, Status='live', Symbol='D7Mit40', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 40', other_designations='', map_location='7 62.01 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60879, CurrentID=0, Status='live', Symbol='D7Mit43', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 43', other_designations='', map_location='7 73.19 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60882, CurrentID=0, Status='live', Symbol='D7Mit62', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 62', other_designations='', map_location='7 48.36 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60884, CurrentID=0, Status='live', Symbol='D7Mit66', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 66', other_designations='', map_location='7 64.3 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60885, CurrentID=0, Status='live', Symbol='D7Mit67', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 67', other_designations='', map_location='7 70.15 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60886, CurrentID=0, Status='live', Symbol='D7Mit69', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 69', other_designations='', map_location='7 31.19 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60887, CurrentID=0, Status='live', Symbol='D7Mit72', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 72', other_designations='', map_location='7 16.93 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60890, CurrentID=0, Status='live', Symbol='D7Mit83', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 83', other_designations='', map_location='7 32.84 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60892, CurrentID=0, Status='live', Symbol='D7Mit98', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 98', other_designations='', map_location='7 60.49 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60893, CurrentID=0, Status='live', Symbol='D8Mit25', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 25', other_designations='', map_location='8 32.3 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60894, CurrentID=0, Status='live', Symbol='D8Mit42', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 42', other_designations='', map_location='8 74.46 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60895, CurrentID=0, Status='live', Symbol='D8Mit45', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 45', other_designations='', map_location='8 42.16 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60897, CurrentID=0, Status='live', Symbol='D8Mit47', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 47', other_designations='', map_location='8 53.24 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60898, CurrentID=0, Status='live', Symbol='D8Mit50', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 50', other_designations='', map_location='8 43.51 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60901, CurrentID=0, Status='live', Symbol='D8Mit56', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 56', other_designations='', map_location='8 76.14 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60902, CurrentID=0, Status='live', Symbol='D8Mit64', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 64', other_designations='', map_location='8 19.01 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60903, CurrentID=0, Status='live', Symbol='D8Mit65', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 65', other_designations='', map_location='8 23.89 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60908, CurrentID=0, Status='live', Symbol='D8Mit80', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 80', other_designations='', map_location='8 43.06 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60910, CurrentID=0, Status='live', Symbol='D8Mit86', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 86', other_designations='', map_location='8 56.18 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60911, CurrentID=0, Status='live', Symbol='D8Mit88', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 88', other_designations='', map_location='8 61.66 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60912, CurrentID=0, Status='live', Symbol='D8Mit91', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 91', other_designations='', map_location='8 72.73 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60915, CurrentID=0, Status='live', Symbol='D8Mit96', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 96', other_designations='', map_location='8 16.6 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60917, CurrentID=0, Status='live', Symbol='D9Mit18', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 18', other_designations='', map_location='9 71.49 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60918, CurrentID=0, Status='live', Symbol='D9Mit19', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 19', other_designations='', map_location='9 71.55 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60919, CurrentID=0, Status='live', Symbol='D9Mit24', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 24', other_designations='', map_location='9 55.07 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60920, CurrentID=0, Status='live', Symbol='D9Mit35', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 35', other_designations='', map_location='9 51.41 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60921, CurrentID=0, Status='live', Symbol='D9Mit36', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 36', other_designations='', map_location='9 52.64 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60923, CurrentID=0, Status='live', Symbol='D9Mit48', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 48', other_designations='', map_location='9 31.63 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60925, CurrentID=0, Status='live', Symbol='D9Mit53', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 53', other_designations='', map_location='9 56.61 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60928, CurrentID=0, Status='live', Symbol='D9Mit67', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 67', other_designations='', map_location='9 20.71 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60929, CurrentID=0, Status='live', Symbol='D9Mit73', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 73', other_designations='', map_location='9 39.85 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60930, CurrentID=0, Status='live', Symbol='D9Mit78', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 78', other_designations='', map_location='9 57.77 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60931, CurrentID=0, Status='live', Symbol='D9Mit80', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 80', other_designations='', map_location='9 56.61 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60936, CurrentID=0, Status='live', Symbol='DXMit28', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 28', other_designations='', map_location='X 73.67 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60938, CurrentID=0, Status='live', Symbol='DXMit31', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 31', other_designations='', map_location='X 76.03 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60940, CurrentID=0, Status='live', Symbol='DXMit33', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 33', other_designations='', map_location='X 73.55 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60941, CurrentID=0, Status='live', Symbol='DXMit37', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 37', other_designations='', map_location='X 59.1 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60942, CurrentID=0, Status='live', Symbol='DXMit38', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 38', other_designations='', map_location='X 57.4 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60944, CurrentID=0, Status='live', Symbol='DXMit41', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 41', other_designations='', map_location='X 43.72 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60947, CurrentID=0, Status='live', Symbol='DXMit48', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 48', other_designations='', map_location='X 25.51 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60948, CurrentID=0, Status='live', Symbol='DXMit49', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 49', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60949, CurrentID=0, Status='live', Symbol='DXMit50', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 50', other_designations='', map_location='X 23.19 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60951, CurrentID=0, Status='live', Symbol='DXMit65', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 65', other_designations='', map_location='X 47.86 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60952, CurrentID=0, Status='live', Symbol='DXMit73', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 73', other_designations='', map_location='X 33.5 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60953, CurrentID=0, Status='live', Symbol='DXMit79', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 79', other_designations='', map_location='X 53.17 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60954, CurrentID=0, Status='live', Symbol='DXMit81', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 81', other_designations='', map_location='X 20.59 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60958, CurrentID=0, Status='live', Symbol='DXMit87', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 87', other_designations='', map_location='X 34.81 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60968, CurrentID=0, Status='live', Symbol='D10Mit3', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 3', other_designations='', map_location='10 16.53 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60970, CurrentID=0, Status='live', Symbol='D12Mit2', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 2', other_designations='', map_location='12 18.94 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60971, CurrentID=0, Status='live', Symbol='D12Mit4', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 4', other_designations='', map_location='12 35.51 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60972, CurrentID=0, Status='live', Symbol='D12Mit6', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 6', other_designations='', map_location='12 44.05 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60973, CurrentID=0, Status='live', Symbol='D12Mit8', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 8', other_designations='', map_location='12 62.1 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60974, CurrentID=0, Status='live', Symbol='D13Mit7', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 7', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60975, CurrentID=0, Status='live', Symbol='D14Mit4', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 4', other_designations='', map_location='14 24.6 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60976, CurrentID=0, Status='live', Symbol='D16Mit1', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 1', other_designations='', map_location='16 12.89 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60977, CurrentID=0, Status='live', Symbol='D16Mit3', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 3', other_designations='', map_location='16 19.88 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60978, CurrentID=0, Status='live', Symbol='D16Mit4', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 4', other_designations='', map_location='16 25.43 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60979, CurrentID=0, Status='live', Symbol='D16Mit7', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 7', other_designations='', map_location='16 48.96 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60980, CurrentID=0, Status='live', Symbol='D17Mit2', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 2', other_designations='', map_location='17 50.97 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60981, CurrentID=0, Status='live', Symbol='D17Mit7', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 7', other_designations='', map_location='17 27.85 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60982, CurrentID=0, Status='live', Symbol='D17Mit9', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 9', other_designations='', map_location='17 26.61 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60984, CurrentID=0, Status='live', Symbol='D18Mit4', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 4', other_designations='', map_location='18 57.53 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60985, CurrentID=0, Status='live', Symbol='D18Mit8', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 8', other_designations='', map_location='18 50.7 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60986, CurrentID=0, Status='live', Symbol='D19Mit1', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 1', other_designations='', map_location='19 50.32 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60987, CurrentID=0, Status='live', Symbol='D19Mit6', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 6', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60989, CurrentID=0, Status='live', Symbol='D19Mit9', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 9', other_designations='', map_location='19 38.97 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  60999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60999, CurrentID=0, Status='live', Symbol='D2Mit74', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 74', other_designations='', map_location='2 103.43 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61004, CurrentID=0, Status='live', Symbol='D4Mit101', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 101', other_designations='', map_location='4 4.29 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61007, CurrentID=0, Status='live', Symbol='D4Mit172', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 172', other_designations='', map_location='4 16.7 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61008, CurrentID=0, Status='live', Symbol='D4Mit181', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 181', other_designations='', map_location='4 4.23 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61009, CurrentID=0, Status='live', Symbol='D4Mit193', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 193', other_designations='', map_location='4 13.99 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61011, CurrentID=0, Status='live', Symbol='D4Mit84', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 84', other_designations='', map_location='4 37.32 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61012, CurrentID=0, Status='live', Symbol='D5Mit179', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 179', other_designations='', map_location='5 10.81 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61027, CurrentID=0, Status='live', Symbol='D7Mit152', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 152', other_designations='', map_location='7 2.71 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61028, CurrentID=0, Status='live', Symbol='D7Mit178', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 178', other_designations='', map_location='7 2.02 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61030, CurrentID=0, Status='live', Symbol='D7Mit191', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 191', other_designations='', map_location='7 9.38 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61032, CurrentID=0, Status='live', Symbol='D8Mit124', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 124', other_designations='', map_location='8 7.59 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192105, CurrentID=0, Status='live', Symbol='Stheal8', Aliases='Sth8', description='soft tissue heal 8', other_designations='', map_location='9 32.76 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192106, CurrentID=0, Status='live', Symbol='Stheal9', Aliases='Sth9', description='soft tissue heal 9', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61039, CurrentID=0, Status='live', Symbol='DXMit103', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 103', other_designations='', map_location='X 7.25 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192113, CurrentID=0, Status='live', Symbol='Atp12a', Aliases='Atp1al1, HKalpha2, cHKA', description='ATPase, H+/K+ transporting, nongastric, alpha polypeptide', other_designations='ATPase H+K+-transporting, alpha 2|non-gastric H(+)/K(+) ATPase subunit alpha|proton pump', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56365068, end_position_on_the_genomic_accession=56388551, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  61042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61042, CurrentID=0, Status='live', Symbol='DXMit161', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 161', other_designations='', map_location='X 4.04 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61044, CurrentID=0, Status='live', Symbol='DXMit136', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 136', other_designations='', map_location='X 4.23 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192118, CurrentID=0, Status='live', Symbol='Lith4', Aliases='', description='lithogenic gene 4', other_designations='', map_location='X 50.14 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192119, CurrentID=0, Status='live', Symbol='Dicer1', Aliases='1110006F08Rik, D12Ertd7e, mKIAA0928', description='dicer 1, ribonuclease type III', other_designations='Dicer1, Dcr-1 homolog|double-strand-specific ribonuclease mDCR-1', map_location='12 E-F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104687740, end_position_on_the_genomic_accession=104751952, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  192120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192120, CurrentID=0, Status='live', Symbol='Bspry', Aliases='', description='B-box and SPRY domain containing', other_designations='zetin 1', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62479995, end_position_on_the_genomic_accession=62497300, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  61052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61052, CurrentID=0, Status='live', Symbol='D8Mit257', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 257', other_designations='', map_location='8 11.22 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61061, CurrentID=0, Status='live', Symbol='D4Mit264', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 264', other_designations='', map_location='4 4.14 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61063, CurrentID=0, Status='live', Symbol='D4Mit235', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 235', other_designations='', map_location='4 3.57 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192136, CurrentID=0, Status='live', Symbol='Sugct', Aliases='5033411D12Rik, D17907', description='succinyl-CoA glutarate-CoA transferase', other_designations='succinylCoA:glutarate-CoA transferase', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=16857472, end_position_on_the_genomic_accession=17694793, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  192140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192140, CurrentID=0, Status='live', Symbol='Tmc2', Aliases='CWEA2', description='transmembrane channel-like gene family 2', other_designations='transmembrane cochlear-expressed protein 2|transmembrane, cochlear expressed 2', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130195194, end_position_on_the_genomic_accession=130264445, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  192156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192156, CurrentID=0, Status='live', Symbol='Mvd', Aliases='C78718, Mpd', description='mevalonate (diphospho) decarboxylase', other_designations='MDDase|mevalonate pyrophosphate decarboxylase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122433596, end_position_on_the_genomic_accession=122443422, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  192157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192157, CurrentID=0, Status='live', Symbol='Socs7', Aliases='2310063P06Rik, C85125, Nap4', description='suppressor of cytokine signaling 7', other_designations='Nck, Ash and phospholipase C binding protein|SH2 domain containing SOCS box protein SOCS7|SOCS-7', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97362296, end_position_on_the_genomic_accession=97398542, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  61086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61086, CurrentID=0, Status='live', Symbol='D5Mit353', Aliases='', description='DNA Segment, Chr 5 Massachusetts Institute of Technology 353', other_designations='', map_location='5 21.25 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192159, CurrentID=0, Status='live', Symbol='Prpf8', Aliases='AU019467, D11Bwg0410e, DBF3/PRP8, Prp8, Sfprp8l', description='pre-mRNA processing factor 8', other_designations='splicing factor Prp8', map_location='11 45.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75486777, end_position_on_the_genomic_accession=75509447, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  192160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192160, CurrentID=0, Status='live', Symbol='Casc3', Aliases='Btz, Mln51', description='cancer susceptibility candidate 3', other_designations='Barentsz|MLN 51 homolog|cancer susceptibility candidate gene 3 protein homolog|mBtz|metastatic lymph node 51|metastatic lymph node gene 51 protein homolog|protein barentsz', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98809808, end_position_on_the_genomic_accession=98833807, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  192161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192161, CurrentID=0, Status='live', Symbol='Pcdha9', Aliases='', description='protocadherin alpha 9', other_designations='PCDH-alpha-9', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36997880, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  192162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192162, CurrentID=0, Status='live', Symbol='Pcdha@', Aliases='', description='protocadherin alpha cluster', other_designations='protocadherin alpha 7', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.5', start_position_on_the_genomic_accession=37089939, end_position_on_the_genomic_accession=37347311, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  192163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192163, CurrentID=0, Status='live', Symbol='Pcdha3', Aliases='mKIAA0345', description='protocadherin alpha 3', other_designations='', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36946207, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  192164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192164, CurrentID=0, Status='live', Symbol='Pcdha12', Aliases='Cnr5, Crnr5, Pcdha13', description='protocadherin alpha 12', other_designations='cadherin-related neuronal receptor 5|protocadherin alpha 13', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37020230, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  192166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192166, CurrentID=0, Status='live', Symbol='Sardh', Aliases='', description='sarcosine dehydrogenase', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27188393, end_position_on_the_genomic_accession=27248332, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  61095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61095, CurrentID=0, Status='live', Symbol='D1Mit478', Aliases='', description='DNA Segment, Chr 1 Massachusetts Institute of Technology 478', other_designations='', map_location='1 26.58 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192169, CurrentID=0, Status='live', Symbol='Ufsp2', Aliases='1810047C23Rik', description='UFM1-specific peptidase 2', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45975528, end_position_on_the_genomic_accession=45996958, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  192170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192170, CurrentID=0, Status='live', Symbol='Eif4a3', Aliases='2400003O03Rik, Ddx48, eIF4A-III, mKIAA0111', description='eukaryotic translation initiation factor 4A3', other_designations='ATP-dependent RNA helicase DDX48|ATP-dependent RNA helicase eIF4A-3|DEAD (Asp-Glu-Ala-Asp) box polypeptide 48|DEAD box protein 48|eIF-4A-III|eukaryotic translation initiation factor 4A', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119288363, end_position_on_the_genomic_accession=119300043, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  192173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192173, CurrentID=0, Status='live', Symbol='Fam195b', Aliases='', description='family with sequence similarity 195, member B', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120542888, end_position_on_the_genomic_accession=120549727, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  192174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192174, CurrentID=0, Status='live', Symbol='Rwdd4a', Aliases='BC016198, Gm1942, Rwdd4', description='RWD domain containing 4A', other_designations='RWD domain-containing protein 4A', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47533645, end_position_on_the_genomic_accession=47552837, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  192176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192176, CurrentID=0, Status='live', Symbol='Flna', Aliases='ABP-280, Dilp2, F730004A14Rik, Fln1, GENA 379, filamin-1', description='filamin, alpha', other_designations='actin-binding protein 280|dilated pupils 2|endothelial actin-binding protein|non-muscle filamin', map_location='X 37.89 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74223461, end_position_on_the_genomic_accession=74249854, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  61110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61110, CurrentID=0, Status='live', Symbol='D2Mit480', Aliases='', description='DNA Segment, Chr 2 Massachusetts Institute of Technology 480', other_designations='', map_location='2 56.65 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192185, CurrentID=0, Status='live', Symbol='Nadk', Aliases='4432404C02Rik, BC004012', description='NAD kinase', other_designations='Putative inorganic polyphosphate/ATP-NAD kinase (Poly(P)/ATP NAD kinase)|poly(P)/ATP NAD kinase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155562392, end_position_on_the_genomic_accession=155591001, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  61115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61115, CurrentID=0, Status='live', Symbol='D2Mit491', Aliases='', description='DNA Segment, Chr 2 Massachusetts Institute of Technology 491', other_designations='', map_location='2 71.0 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61116, CurrentID=0, Status='live', Symbol='D2Mit492', Aliases='', description='DNA Segment, Chr 2 Massachusetts Institute of Technology 492', other_designations='', map_location='2 73.4 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192190, CurrentID=0, Status='live', Symbol='Pkhd1l1', Aliases='AB055648, PKHDL1', description='polycystic kidney and hepatic disease 1-like 1', other_designations='PKHD1-like protein 1|fibrocystin L|polycystic kidney and hepatic disease 1-like protein 1', map_location='15 B3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=44457371, end_position_on_the_genomic_accession=44601376, orientation='plus', exon_count=82, OMIM='', no_hdr0=''),
  192191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192191, CurrentID=0, Status='live', Symbol='Med9', Aliases='BC019367, Med25', description='mediator complex subunit 9', other_designations='mediator of RNA polymerase II transcription, subunit 9 homolog', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59948214, end_position_on_the_genomic_accession=59963306, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  192192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192192, CurrentID=0, Status='live', Symbol='Shkbp1', Aliases='B930062H15Rik, Sb1', description='Sh3kbp1 binding protein 1', other_designations='SETA binding protein 1|SETA-binding protein 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27342128, end_position_on_the_genomic_accession=27356015, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  192193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192193, CurrentID=0, Status='live', Symbol='Edem1', Aliases='A130059K23Rik, EDEM, mKIAA0212', description='ER degradation enhancer, mannosidase alpha-like 1', other_designations='ER degradation ER degradation enhancer, mannosidase alpha-like 1|ER degradation-enhancing alpha-mannosidase-like 1', map_location='6 E2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=108828641, end_position_on_the_genomic_accession=108859356, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  192194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192194, CurrentID=0, Status='live', Symbol='Btnl10', Aliases='BUTR-1, Butr1', description='butyrophilin-like 10', other_designations='butyrophilin related 1|butyrophilin-like protein BUTR-1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58917878, end_position_on_the_genomic_accession=58926965, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  61123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61123, CurrentID=0, Status='live', Symbol='D3Mit351', Aliases='', description='DNA Segment, Chr 3 Massachusetts Institute of Technology 351', other_designations='', map_location='3 64.89 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192196, CurrentID=0, Status='live', Symbol='Luc7l2', Aliases='4930471C18Rik, AA522013, AU015269, CGI59, Cgi74', description='LUC7-like 2 (S. cerevisiae)', other_designations='CGI-74 homolog|CGI-74-like SR-rich', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38551339, end_position_on_the_genomic_accession=38609471, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  61125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61125, CurrentID=0, Status='live', Symbol='D5Mit387', Aliases='', description='DNA Segment, Chr 5 Massachusetts Institute of Technology 387', other_designations='', map_location='5 14.39 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192198, CurrentID=0, Status='live', Symbol='Lrrc4', Aliases='MBAG1, NGL-2, NLG-2, Nag14', description='leucine rich repeat containing 4', other_designations='LIBG-like protein|LRRC4-like protein|brain tumor-associated protein MBAG1|netrin-G2 ligand', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=28828126, end_position_on_the_genomic_accession=28831747, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  192199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192199, CurrentID=0, Status='live', Symbol='Rspo1', Aliases='R-spondin, Rspondin', description='R-spondin 1', other_designations='R-spondin homolog|cristin-3|cysteine-rich and single thrombospondin domain-containing protein 3|mCristin-3|roof plate-specific spondin-1|thrombospondin type 1 domain containing gene Rspondin', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124986400, end_position_on_the_genomic_accession=125009102, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  192200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192200, CurrentID=0, Status='live', Symbol='Wfdc12', Aliases='Swam2, WAP2', description='WAP four-disulfide core domain 12', other_designations='elafin-like protein II|single WAP motif protein 2|wfdc12|whey acidic protein 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164189231, end_position_on_the_genomic_accession=164190558, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  192201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192201, CurrentID=0, Status='live', Symbol='Wfdc15b', Aliases='9230106L14Rik, SWAM1, Wfdc15', description='WAP four-disulfide core domain 15B', other_designations='WAP four-disulfide core domain 15|elafin-like protein I|single WAP motif protein 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164214454, end_position_on_the_genomic_accession=164221663, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  61137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61137, CurrentID=0, Status='live', Symbol='D9Mit328', Aliases='', description='DNA Segment, Chr 9 Massachusetts Institute of Technology 328', other_designations='', map_location='9 23.53 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61138, CurrentID=0, Status='live', Symbol='D9Mit329', Aliases='', description='DNA Segment, Chr 9 Massachusetts Institute of Technology 329', other_designations='', map_location='9 24.33 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192212, CurrentID=0, Status='live', Symbol='Prom2', Aliases='AI507243, Prom-rp', description='prominin 2', other_designations='PROM-2|mPROML2|prominin-like protein 2|prominin-related protein', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127526953, end_position_on_the_genomic_accession=127541417, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  192216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192216, CurrentID=0, Status='live', Symbol='Tmem47', Aliases='3010015F07Rik, AI115536, BCMP1, Tm4sf10', description='transmembrane protein 47', other_designations='transmembrane 4 superfamily member 10', map_location='X A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=81070644, end_position_on_the_genomic_accession=81097875, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  61147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61147, CurrentID=0, Status='live', Symbol='DXMit220', Aliases='', description='DNA Segment, Chr X Massachusetts Institute of Technology 220', other_designations='', map_location='X 73.95 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61150, CurrentID=0, Status='live', Symbol='D10Mit51', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 51', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61154, CurrentID=0, Status='live', Symbol='D11Mit77', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 77', other_designations='', map_location='11 10.44 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61155, CurrentID=0, Status='live', Symbol='D11Mit78', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 78', other_designations='', map_location='11 10.44 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61156, CurrentID=0, Status='live', Symbol='D11Mit8', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 8', other_designations='', map_location='11 47.28 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61157, CurrentID=0, Status='live', Symbol='D12Mit37', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 37', other_designations='', map_location='12 2.59 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192231, CurrentID=0, Status='live', Symbol='Hexim1', Aliases='7330426E13Rik, CLP-1, Clp1, HIS1', description='hexamethylene bis-acetamide inducible 1', other_designations='cardiac lineage protein 1', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103116325, end_position_on_the_genomic_accession=103119725, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  61160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61160, CurrentID=0, Status='live', Symbol='D12Mit82', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 82', other_designations='', map_location='12 7.03 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61162, CurrentID=0, Status='live', Symbol='D13Mit17', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 17', other_designations='', map_location='13 7.73 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61164, CurrentID=0, Status='live', Symbol='D14Mit11', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 11', other_designations='', map_location='14 6.33 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61166, CurrentID=0, Status='live', Symbol='D15Mit10', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 10', other_designations='', map_location='15 5.35 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61167, CurrentID=0, Status='live', Symbol='D15Mit13', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 13', other_designations='', map_location='15 1.84 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61169, CurrentID=0, Status='live', Symbol='D16Mit32', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 32', other_designations='', map_location='16 2.33 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61172, CurrentID=0, Status='live', Symbol='D16Mit73', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 73', other_designations='', map_location='16 8.31 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61173, CurrentID=0, Status='live', Symbol='D16Mit79', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 79', other_designations='', map_location='16 2.59 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61174, CurrentID=0, Status='live', Symbol='D16Mit88', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 88', other_designations='', map_location='16 9.23 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61175, CurrentID=0, Status='live', Symbol='D17Mit25a', Aliases='D17Mit25', description='DNA segment, Chr 17, Massachusetts Institute of Technology 25a', other_designations='', map_location='17 5.07 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61176, CurrentID=0, Status='live', Symbol='D18Mit18', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 18', other_designations='', map_location='18 4.53 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61177, CurrentID=0, Status='live', Symbol='D18Mit21', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 21', other_designations='', map_location='18 8.85 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61182, CurrentID=0, Status='live', Symbol='D19Mit22', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 22', other_designations='', map_location='19 7.16 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61183, CurrentID=0, Status='live', Symbol='D19Mit29', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 29', other_designations='', map_location='19 4.22 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61187, CurrentID=0, Status='live', Symbol='D19Mit68', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 68', other_designations='', map_location='19 3.38 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61190, CurrentID=0, Status='live', Symbol='D1Mit107', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 107', other_designations='', map_location='1 70.19 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61193, CurrentID=0, Status='live', Symbol='D1Mit110', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 110', other_designations='', map_location='1 75.1 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61194, CurrentID=0, Status='live', Symbol='D1Mit115', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 115', other_designations='', map_location='1 82.78 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61197, CurrentID=0, Status='live', Symbol='D1Mit123', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 123', other_designations='', map_location='1 17.67 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61206, CurrentID=0, Status='live', Symbol='D1Mit145', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 145', other_designations='', map_location='1 74.68 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61209, CurrentID=0, Status='live', Symbol='D1Mit171', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 171', other_designations='', map_location='1 15.4 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61212, CurrentID=0, Status='live', Symbol='D1Mit180', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 180', other_designations='', map_location='1 38.2 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61213, CurrentID=0, Status='live', Symbol='D1Mit181', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 181', other_designations='', map_location='1 38.54 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192287, CurrentID=0, Status='live', Symbol='Slc25a36', Aliases='C330005L02Rik', description='solute carrier family 25, member 36', other_designations='', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=97075281, end_position_on_the_genomic_accession=97111041, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  61216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61216, CurrentID=0, Status='live', Symbol='D1Mit194', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 194', other_designations='', map_location='1 63.28 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192289, CurrentID=0, Status='live', Symbol='Tmlhe', Aliases='Bbox2, D430017M14Rik, TMLH', description='trimethyllysine hydroxylase, epsilon', other_designations='TML dioxygenase|TML hydroxylase|TML-alpha-ketoglutarate dioxygenase|TMLD|butyrobetaine (gamma), 2-oxoglutarate dioxygenase 2 (gamma-butyrobetaine hydroxylase)|epsilon-trimethyllysine 2-oxoglutarate dioxygenase', map_location='X', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61218, CurrentID=0, Status='live', Symbol='D1Mit211', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 211', other_designations='', map_location='1 10.59 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  665563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=665563, CurrentID=0, Status='live', Symbol='Mthfd2l', Aliases='1110019K23Rik, C630010D07Rik', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent) 2-like', other_designations='MTHFD2-like|NADP-dependent methylenetetrahydrofolate dehydrogenase 2-like protein', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90931196, end_position_on_the_genomic_accession=91021368, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  61220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61220, CurrentID=0, Status='live', Symbol='D1Mit213', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 213', other_designations='', map_location='1 22.88 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61221, CurrentID=0, Status='live', Symbol='D1Mit215', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 215', other_designations='', map_location='1 39.91 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61223, CurrentID=0, Status='live', Symbol='D1Mit218', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 218', other_designations='', map_location='1 55.76 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61229, CurrentID=0, Status='live', Symbol='D1Mit236', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 236', other_designations='', map_location='1 23.69 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61237, CurrentID=0, Status='live', Symbol='D1Mit270', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 270', other_designations='', map_location='1 77.05 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61244, CurrentID=0, Status='live', Symbol='D1Mit288', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 288', other_designations='', map_location='1 63.51 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61248, CurrentID=0, Status='live', Symbol='D2Mit109', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 109', other_designations='', map_location='2 71.17 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61251, CurrentID=0, Status='live', Symbol='D2Mit124', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 124', other_designations='', map_location='2 37.06 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61253, CurrentID=0, Status='live', Symbol='D2Mit126', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 126', other_designations='', map_location='2 49.35 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61259, CurrentID=0, Status='live', Symbol='D2Mit151', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 151', other_designations='', map_location='2 18.92 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61262, CurrentID=0, Status='live', Symbol='D2Mit159', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 159', other_designations='', map_location='2 46.94 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61269, CurrentID=0, Status='live', Symbol='D2Mit219', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 219', other_designations='', map_location='2 44.69 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61271, CurrentID=0, Status='live', Symbol='D2Mit229', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 229', other_designations='', map_location='2 88.99 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61275, CurrentID=0, Status='live', Symbol='D2Mit242', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 242', other_designations='', map_location='2 31.72 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61277, CurrentID=0, Status='live', Symbol='D2Mit245', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 245', other_designations='', map_location='2 43.35 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61279, CurrentID=0, Status='live', Symbol='D2Mit255', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 255', other_designations='', map_location='2 60.69 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61281, CurrentID=0, Status='live', Symbol='D2Mit258', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 258', other_designations='', map_location='2 63.22 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61290, CurrentID=0, Status='live', Symbol='D2Mit281', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 281', other_designations='', map_location='2 73.53 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61293, CurrentID=0, Status='live', Symbol='D2Mit285', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 285', other_designations='', map_location='2 75.41 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61298, CurrentID=0, Status='live', Symbol='D3Mit101', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 101', other_designations='', map_location='3 41.93 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61299, CurrentID=0, Status='live', Symbol='D3Mit102', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 102', other_designations='', map_location='3 44.3 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61300, CurrentID=0, Status='live', Symbol='D3Mit103', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 103', other_designations='', map_location='3 46.83 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61302, CurrentID=0, Status='live', Symbol='D3Mit106', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 106', other_designations='', map_location='3 49.08 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61310, CurrentID=0, Status='live', Symbol='D3Mit137', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 137', other_designations='', map_location='3 34.91 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61313, CurrentID=0, Status='live', Symbol='D3Mit147', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 147', other_designations='', map_location='3 74.02 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61321, CurrentID=0, Status='live', Symbol='D3Mit173', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 173', other_designations='', map_location='3 28.63 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61323, CurrentID=0, Status='live', Symbol='D3Mit178', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 178', other_designations='', map_location='3 14.27 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61335, CurrentID=0, Status='live', Symbol='D3Mit199', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 199', other_designations='', map_location='3 54.67 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61336, CurrentID=0, Status='live', Symbol='D3Mit200', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 200', other_designations='', map_location='3 72.41 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61341, CurrentID=0, Status='live', Symbol='D3Mit216', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 216', other_designations='', map_location='3 54.48 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  228677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228677, CurrentID=0, Status='live', Symbol='Sptlc3', Aliases='C130053K05Rik', description='serine palmitoyltransferase, long chain base subunit 3', other_designations='LCB 3|LCB2b|SPT 3|long chain base biosynthesis protein 2b|long chain base biosynthesis protein 3|serine-palmitoyl-CoA transferase 3', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=139493913, end_position_on_the_genomic_accession=139637674, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  61347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61347, CurrentID=0, Status='live', Symbol='D4Mit116', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 116', other_designations='', map_location='4 39.46 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61348, CurrentID=0, Status='live', Symbol='D4Mit120', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 120', other_designations='', map_location='4 51.43 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61349, CurrentID=0, Status='live', Symbol='D4Mit127', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 127', other_designations='', map_location='4 80.52 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61351, CurrentID=0, Status='live', Symbol='D4Mit145', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 145', other_designations='', map_location='4 43.3 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61352, CurrentID=0, Status='live', Symbol='D4Mit148', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 148', other_designations='', map_location='4 69.48 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61354, CurrentID=0, Status='live', Symbol='D4Mit152', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 152', other_designations='', map_location='4 39.46 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61356, CurrentID=0, Status='live', Symbol='D4Mit157', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 157', other_designations='', map_location='4 67.76 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61357, CurrentID=0, Status='live', Symbol='D4Mit158', Aliases='Lynf1', description='DNA segment, Chr 4, Massachusetts Institute of Technology 158', other_designations='Lymphocyte infiltration 1', map_location='4 70.02 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61358, CurrentID=0, Status='live', Symbol='D4Mit160', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 160', other_designations='', map_location='4 77.98 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61360, CurrentID=0, Status='live', Symbol='D4Mit175', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 175', other_designations='', map_location='4 45.71 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61361, CurrentID=0, Status='live', Symbol='D4Mit180', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 180', other_designations='', map_location='4 82.6 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61364, CurrentID=0, Status='live', Symbol='D4Mit190', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 190', other_designations='', map_location='4 81.46 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61369, CurrentID=0, Status='live', Symbol='D4Mit203', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 203', other_designations='', map_location='4 63.26 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61370, CurrentID=0, Status='live', Symbol='D4Mit204', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 204', other_designations='', map_location='4 66.25 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61372, CurrentID=0, Status='live', Symbol='D4Mit209', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 209', other_designations='', map_location='4 84.59 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61373, CurrentID=0, Status='live', Symbol='D4Mit214', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 214', other_designations='', map_location='4 24.16 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61375, CurrentID=0, Status='live', Symbol='D4Mit219', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 219', other_designations='', map_location='4 47.21 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61378, CurrentID=0, Status='live', Symbol='D4Mit225', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 225', other_designations='', map_location='4 79.77 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61379, CurrentID=0, Status='live', Symbol='D4Mit226', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 226', other_designations='', map_location='4 81.42 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61385, CurrentID=0, Status='live', Symbol='D5Mit122', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 122', other_designations='', map_location='5 89.28 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61391, CurrentID=0, Status='live', Symbol='D5Mit136', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 136', other_designations='', map_location='5 59.57 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61392, CurrentID=0, Status='live', Symbol='D5Mit138', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 138', other_designations='', map_location='5 64.01 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61396, CurrentID=0, Status='live', Symbol='D5Mit148', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 148', other_designations='', map_location='5 17.33 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61397, CurrentID=0, Status='live', Symbol='D5Mit149', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 149', other_designations='', map_location='5 19.06 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61401, CurrentID=0, Status='live', Symbol='D5Mit158', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 158', other_designations='', map_location='5 55.99 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61403, CurrentID=0, Status='live', Symbol='D5Mit161', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 161', other_designations='', map_location='5 65.34 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61410, CurrentID=0, Status='live', Symbol='D5Mit175', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 175', other_designations='', map_location='5 52.15 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61412, CurrentID=0, Status='live', Symbol='D5Mit180', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 180', other_designations='', map_location='5 11.93 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61413, CurrentID=0, Status='live', Symbol='D5Mit182', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 182', other_designations='', map_location='5 20.2 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61416, CurrentID=0, Status='live', Symbol='D5Mit188', Aliases='Tgfbm1', description='DNA segment, Chr 5, Massachusetts Institute of Technology 188', other_designations='transforming growth factor beta, modifier 1', map_location='5 57.51 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61428, CurrentID=0, Status='live', Symbol='D5Mit233', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 233', other_designations='', map_location='5 28.55 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61430, CurrentID=0, Status='live', Symbol='D5Mit235', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 235', other_designations='', map_location='5 39.55 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61432, CurrentID=0, Status='live', Symbol='D5Mit239', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 239', other_designations='', map_location='5 52.23 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61433, CurrentID=0, Status='live', Symbol='D5Mit242', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 242', other_designations='', map_location='5 61.73 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61436, CurrentID=0, Status='live', Symbol='D6Mit104', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 104', other_designations='', map_location='6 51.53 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61437, CurrentID=0, Status='live', Symbol='D6Mit105', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 105', other_designations='', map_location='6 49.71 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61445, CurrentID=0, Status='live', Symbol='D6Mit123', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 123', other_designations='', map_location='6 27.76 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61449, CurrentID=0, Status='live', Symbol='D6Mit132', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 132', other_designations='', map_location='6 44.78 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61450, CurrentID=0, Status='live', Symbol='D6Mit134', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 134', other_designations='', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61451, CurrentID=0, Status='live', Symbol='D6Mit135', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 135', other_designations='', map_location='6 63.1 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61453, CurrentID=0, Status='live', Symbol='D6Mit150', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 150', other_designations='', map_location='6 53.75 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61458, CurrentID=0, Status='live', Symbol='D6Mit183', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 183', other_designations='', map_location='6 25.82 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61463, CurrentID=0, Status='live', Symbol='D6Mit200', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 200', other_designations='', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61464, CurrentID=0, Status='live', Symbol='D6Mit201', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 201', other_designations='', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61466, CurrentID=0, Status='live', Symbol='D6Mit209', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 209', other_designations='', map_location='6 32.56 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61468, CurrentID=0, Status='live', Symbol='D7Mit101', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 101', other_designations='', map_location='7 69.01 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61470, CurrentID=0, Status='live', Symbol='D7Mit103', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 103', other_designations='', map_location='7 71.55 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61472, CurrentID=0, Status='live', Symbol='D7Mit108', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 108', other_designations='', map_location='7 76.89 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61493, CurrentID=0, Status='live', Symbol='D7Mit193', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 193', other_designations='', map_location='7 31.44 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61502, CurrentID=0, Status='live', Symbol='D7Mit222', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 222', other_designations='', map_location='7 59.13 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61508, CurrentID=0, Status='live', Symbol='D8Mit113', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 113', other_designations='', map_location='8 56.16 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61511, CurrentID=0, Status='live', Symbol='D8Mit125', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 125', other_designations='', map_location='8 21.16 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61521, CurrentID=0, Status='live', Symbol='D8Mit156', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 156', other_designations='', map_location='8 76.09 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61531, CurrentID=0, Status='live', Symbol='D8Mit191', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 191', other_designations='', map_location='8 21.16 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61532, CurrentID=0, Status='live', Symbol='D8Mit192', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 192', other_designations='', map_location='8 25.38 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61534, CurrentID=0, Status='live', Symbol='D8Mit200', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 200', other_designations='', map_location='8 61.37 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61538, CurrentID=0, Status='live', Symbol='D9Mit115', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 115', other_designations='', map_location='9 53.48 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61546, CurrentID=0, Status='live', Symbol='D9Mit140', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 140', other_designations='', map_location='9 24.1 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61549, CurrentID=0, Status='live', Symbol='D9Mit144', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 144', other_designations='', map_location='9 34.43 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61551, CurrentID=0, Status='live', Symbol='D9Mit151', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 151', other_designations='', map_location='9 72.43 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61560, CurrentID=0, Status='live', Symbol='D9Mit182', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 182', other_designations='', map_location='9 53.48 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61562, CurrentID=0, Status='live', Symbol='DXMit100', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 100', other_designations='', map_location='X 78.83 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61569, CurrentID=0, Status='live', Symbol='DXMit121', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 121', other_designations='', map_location='X 73.95 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61576, CurrentID=0, Status='live', Symbol='DXMit174', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 174', other_designations='', map_location='X 57.38 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61577, CurrentID=0, Status='live', Symbol='DXMit170', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 170', other_designations='', map_location='X 45.87 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192650, CurrentID=0, Status='live', Symbol='Cabp7', Aliases='', description='calcium binding protein 7', other_designations='calcium-binding protein CaBP7|calneuron II|calneuron-2|calneuron-II', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4738821, end_position_on_the_genomic_accession=4746778, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  192651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192651, CurrentID=0, Status='live', Symbol='Zfp286', Aliases='BC027023', description='zinc finger protein 286', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62778387, end_position_on_the_genomic_accession=62789483, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  192652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192652, CurrentID=0, Status='live', Symbol='Wdr81', Aliases='BC054822, Gm883, mFLJ00182', description='WD repeat domain 81', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75440943, end_position_on_the_genomic_accession=75454717, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  61581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61581, CurrentID=0, Status='live', Symbol='DXMit140', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 140', other_designations='', map_location='X 33.5 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192654, CurrentID=0, Status='live', Symbol='Pla2g15', Aliases='ACS, C87498, LLPL, Lpla2, Lypla3', description='phospholipase A2, group XV', other_designations='1-O-acylceramide synthase|LCAT-like lysophospholipase|lysophospholipase 3|lysosomal phospholipase A2', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106150351, end_position_on_the_genomic_accession=106164715, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  192656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192656, CurrentID=0, Status='live', Symbol='Ripk2', Aliases='2210420D18Rik, CARD3, CARDIAK, CCK, D4Bwg0615e, RICK, RIP2', description='receptor (TNFRSF)-interacting serine-threonine kinase 2', other_designations='tyrosine-protein kinase RIPK2', map_location='4 6.7 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=16122741, end_position_on_the_genomic_accession=16163679, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  192657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192657, CurrentID=0, Status='live', Symbol='Ell2', Aliases='', description='elongation factor RNA polymerase II 2', other_designations='ELL-related RNA polymerase II, elongation factor|eleven-nineteen lysine-rich leukemia gene 2', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=75707484, end_position_on_the_genomic_accession=75772361, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  192658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192658, CurrentID=0, Status='live', Symbol='Rfpl4', Aliases='D7Ertd486ea, Rfpl4', description='ret finger protein-like 4', other_designations='', map_location='7 2.96 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5109787, end_position_on_the_genomic_accession=5116911, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  61589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61589, CurrentID=0, Status='live', Symbol='D9Mit289', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 289', other_designations='', map_location='9 38.58 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192662, CurrentID=0, Status='live', Symbol='Arhgdia', Aliases='5330430M07Rik, Gdi-1, Rho-GDI, RhoDGI, RhoGDI-1', description='Rho GDP dissociation inhibitor (GDI) alpha', other_designations='Rho GDI alpha|rho GDI 1|rho-GDI alpha', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120577235, end_position_on_the_genomic_accession=120581620, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  192663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192663, CurrentID=0, Status='live', Symbol='Abcg4', Aliases='6430517O04Rik', description='ATP-binding cassette, sub-family G (WHITE), member 4', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44273188, end_position_on_the_genomic_accession=44289102, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  61594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61594, CurrentID=0, Status='live', Symbol='D9Mit279', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 279', other_designations='', map_location='9 69.87 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61600, CurrentID=0, Status='live', Symbol='D9Mit270', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 270', other_designations='', map_location='9 42.48 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61603, CurrentID=0, Status='live', Symbol='D9Mit263', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 263', other_designations='', map_location='9 42.32 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192678, CurrentID=0, Status='live', Symbol='Rassf3', Aliases='AW212023, AW322379', description='Ras association (RalGDS/AF-6) domain family member 3', other_designations='Ras association (RalGDS/AF-6) domain family 3|Ras association domain family 3|novel ras effector 1', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121410350, end_position_on_the_genomic_accession=121476250, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  192682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192682, CurrentID=0, Status='live', Symbol='Pcdhg@', Aliases='', description='protocadherin gamma cluster', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.5', start_position_on_the_genomic_accession=37821599, end_position_on_the_genomic_accession=38001527, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  61613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61613, CurrentID=0, Status='live', Symbol='D9Mit227', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 227', other_designations='', map_location='9 23.36 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61618, CurrentID=0, Status='live', Symbol='D8Mit281', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 281', other_designations='', map_location='8 16.6 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61619, CurrentID=0, Status='live', Symbol='D8Mit280', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 280', other_designations='', map_location='8 74.66 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61623, CurrentID=0, Status='live', Symbol='D8Mit271', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 271', other_designations='', map_location='8 60.86 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61633, CurrentID=0, Status='live', Symbol='D8Mit242', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 242', other_designations='', map_location='8 50.07 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61635, CurrentID=0, Status='live', Symbol='D8Mit240', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 240', other_designations='', map_location='8 44.99 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61640, CurrentID=0, Status='live', Symbol='D8Mit231', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 231', other_designations='', map_location='8 30.79 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61646, CurrentID=0, Status='live', Symbol='D8Mit211', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 211', other_designations='', map_location='8 52.0 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61647, CurrentID=0, Status='live', Symbol='D8Mit205', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 205', other_designations='', map_location='8 28.85 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61650, CurrentID=0, Status='live', Symbol='D7Mit301', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 301', other_designations='', map_location='7 51.2 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61655, CurrentID=0, Status='live', Symbol='D7Mit281', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 281', other_designations='', map_location='7 58.74 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61658, CurrentID=0, Status='live', Symbol='D7Mit270', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 270', other_designations='', map_location='7 25.61 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61659, CurrentID=0, Status='live', Symbol='D7Mit267', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 267', other_designations='', map_location='7 17.09 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61660, CurrentID=0, Status='live', Symbol='D7Mit261', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 261', other_designations='', map_location='7 47.52 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192734, CurrentID=0, Status='live', Symbol='Lrrc75b', Aliases='AI646023, Fam211b', description='leucine rich repeat containing 75B', other_designations='family with sequence similarity 211, member B|leucine-rich repeat-containing protein C22orf36 homolog|leucine-rich repeat-containing protein FAM211B', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75550119, end_position_on_the_genomic_accession=75561469, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  61665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61665, CurrentID=0, Status='live', Symbol='D7Mit230', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 230', other_designations='', map_location='7 31.44 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61666, CurrentID=0, Status='live', Symbol='D7Mit228', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 228', other_designations='', map_location='7 25.61 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61667, CurrentID=0, Status='live', Symbol='D7Mit227', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 227', other_designations='', map_location='7 22.44 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61669, CurrentID=0, Status='live', Symbol='D6Mit304', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 304', other_designations='', map_location='6 78.22 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61670, CurrentID=0, Status='live', Symbol='D6Mit301', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 301', other_designations='', map_location='6 66.48 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61671, CurrentID=0, Status='live', Symbol='D6Mit300', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 300', other_designations='', map_location='6 58.07 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61674, CurrentID=0, Status='live', Symbol='D6Mit294', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 294', other_designations='', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61676, CurrentID=0, Status='live', Symbol='D6Mit291', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 291', other_designations='', map_location='6 66.78 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61677, CurrentID=0, Status='live', Symbol='D6Mit289', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 289', other_designations='', map_location='6 63.1 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61678, CurrentID=0, Status='live', Symbol='D6Mit287', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 287', other_designations='', map_location='6 52.14 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61679, CurrentID=0, Status='live', Symbol='D6Mit284', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 284', other_designations='', map_location='6 41.11 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61692, CurrentID=0, Status='live', Symbol='D6Mit243', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 243', other_designations='', map_location='6 32.27 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61695, CurrentID=0, Status='live', Symbol='D6Mit230', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 230', other_designations='', map_location='6 45.74 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61698, CurrentID=0, Status='live', Symbol='D6Mit223', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 223', other_designations='', map_location='6 21.81 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61701, CurrentID=0, Status='live', Symbol='D5Mit338', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 338', other_designations='', map_location='5 53.23 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192775, CurrentID=0, Status='live', Symbol='Kcnh6', Aliases='m-erg2', description='potassium voltage-gated channel, subfamily H (eag-related), member 6', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106007994, end_position_on_the_genomic_accession=106035595, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  61710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61710, CurrentID=0, Status='live', Symbol='D5Mit317', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 317', other_designations='', map_location='5 54.48 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61712, CurrentID=0, Status='live', Symbol='D5Mit314', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 314', other_designations='', map_location='5 53.25 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61713, CurrentID=0, Status='live', Symbol='D5Mit312', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 312', other_designations='', map_location='5 47.29 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61714, CurrentID=0, Status='live', Symbol='D5Mit309', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 309', other_designations='', map_location='5 42.22 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61732, CurrentID=0, Status='live', Symbol='D5Mit255', Aliases='', description='DNA segment, Chr 5, Massachusetts Institute of Technology 255', other_designations='', map_location='5 30.52 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61739, CurrentID=0, Status='live', Symbol='D4Mit278', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 278', other_designations='', map_location='4 52.81 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61752, CurrentID=0, Status='live', Symbol='D4Mit251', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 251', other_designations='', map_location='4 69.05 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61754, CurrentID=0, Status='live', Symbol='D4Mit249', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 249', other_designations='', map_location='4 59.38 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61758, CurrentID=0, Status='live', Symbol='D4Mit237a', Aliases='D4Mit237', description='DNA segment, Chr 4, Massachusetts Institute of Technology 237a', other_designations='', map_location='4 22.38 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61759, CurrentID=0, Status='live', Symbol='D4Mit236', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 236', other_designations='', map_location='4 20.0 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61765, CurrentID=0, Status='live', Symbol='D3Mit292', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 292', other_designations='', map_location='3 71.03 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61783, CurrentID=0, Status='live', Symbol='D3Mit244', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 244', other_designations='', map_location='3 35.01 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61784, CurrentID=0, Status='live', Symbol='D3Mit241', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 241', other_designations='', map_location='3 30.51 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61792, CurrentID=0, Status='live', Symbol='D2Mit423', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 423', other_designations='', map_location='2 73.57 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61794, CurrentID=0, Status='live', Symbol='D2Mit420', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 420', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61797, CurrentID=0, Status='live', Symbol='D2Mit411', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 411', other_designations='', map_location='2 80.04 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61802, CurrentID=0, Status='live', Symbol='D2Mit398', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 398', other_designations='', map_location='2 61.76 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61803, CurrentID=0, Status='live', Symbol='D2Mit396', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 396', other_designations='', map_location='2 60.68 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61804, CurrentID=0, Status='live', Symbol='D2Mit395', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 395', other_designations='', map_location='2 59.97 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61809, CurrentID=0, Status='live', Symbol='D2Mit380', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 380', other_designations='', map_location='2 40.89 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61817, CurrentID=0, Status='live', Symbol='D2Mit354', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 354', other_designations='', map_location='2 3.11 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61822, CurrentID=0, Status='live', Symbol='D2Mit343', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 343', other_designations='', map_location='2 89.27 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192897, CurrentID=0, Status='live', Symbol='Itgb4', Aliases='AA407042, C230078O20, CD104', description='integrin beta 4', other_designations='', map_location='11 80.91 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115974725, end_position_on_the_genomic_accession=116008412, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  61826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61826, CurrentID=0, Status='live', Symbol='D2Mit337', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 337', other_designations='', map_location='2 60.8 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61827, CurrentID=0, Status='live', Symbol='D2Mit333', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 333', other_designations='', map_location='2 54.85 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61828, CurrentID=0, Status='live', Symbol='D2Mit328', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 328', other_designations='', map_location='2 42.89 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61833, CurrentID=0, Status='live', Symbol='D2Mit311', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 311', other_designations='', map_location='2 86.02 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61836, CurrentID=0, Status='live', Symbol='D2Mit304', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 304', other_designations='', map_location='2 62.49 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61837, CurrentID=0, Status='live', Symbol='D2Mit296', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 296', other_designations='', map_location='2 21.81 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61843, CurrentID=0, Status='live', Symbol='D1Mit425', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 425', other_designations='', map_location='1 67.71 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61845, CurrentID=0, Status='live', Symbol='D1Mit415', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 415', other_designations='', map_location='1 43.94 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61848, CurrentID=0, Status='live', Symbol='D1Mit410', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 410', other_designations='', map_location='1 10.98 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61851, CurrentID=0, Status='live', Symbol='D1Mit399', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 399', other_designations='', map_location='1 70.29 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61861, CurrentID=0, Status='live', Symbol='D1Mit370', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 370', other_designations='', map_location='1 75.22 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61867, CurrentID=0, Status='live', Symbol='D1Mit353', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 353', other_designations='', map_location='1 76.84 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61874, CurrentID=0, Status='live', Symbol='D1Mit334', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 334', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61876, CurrentID=0, Status='live', Symbol='D1Mit332', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 332', other_designations='', map_location='1 39.58 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192950, CurrentID=0, Status='live', Symbol='Nacad', Aliases='D230024G13Rik, mKIAA0363', description='NAC alpha domain containing', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6597814, end_position_on_the_genomic_accession=6606054, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  61884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61884, CurrentID=0, Status='live', Symbol='D1Mit318', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 318', other_designations='', map_location='1 12.82 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192970, CurrentID=0, Status='live', Symbol='Dhrs11', Aliases='', description='dehydrogenase/reductase (SDR family) member 11', other_designations='short-chain dehydrogenase/reductase family 24C member 1', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84820721, end_position_on_the_genomic_accession=84829054, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  61899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61899, CurrentID=0, Status='live', Symbol='D2Mit434', Aliases='', description='DNA Segment, Chr 2, Massachusetts Institute of Technology 434', other_designations='', map_location='2 42.18 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61901, CurrentID=0, Status='live', Symbol='D2Mit436', Aliases='', description='DNA Segment, Chr 2, Massachusetts Institute of Technology 436', other_designations='', map_location='2 47.87 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61903, CurrentID=0, Status='live', Symbol='D2Mit444', Aliases='', description='DNA Segment, Chr 2, Massachusetts Institute of Technology 444', other_designations='', map_location='2 59.46 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  192976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192976, CurrentID=0, Status='live', Symbol='Lrrc75a', Aliases='Fam211a', description='leucine rich repeat containing 75A', other_designations='family with sequence similarity 211, member A|leucine-rich repeat-containing protein C17orf76 homolog|leucine-rich repeat-containing protein FAM211A', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62604884, end_position_on_the_genomic_accession=62648583, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  61914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61914, CurrentID=0, Status='live', Symbol='D4Mit297', Aliases='', description='DNA Segment, Chr 4, Massachusetts Institute of Technology 297', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61923, CurrentID=0, Status='live', Symbol='D4Mit310', Aliases='', description='DNA Segment, Chr 4, Massachusetts Institute of Technology 310', other_designations='', map_location='4 78.87 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61927, CurrentID=0, Status='live', Symbol='D5Mit357', Aliases='', description='DNA Segment, Chr 5, Massachusetts Institute of Technology 357', other_designations='', map_location='5 41.45 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193003, CurrentID=0, Status='live', Symbol='Pirt', Aliases='A530088H08Rik', description='phosphoinositide-interacting regulator of transient receptor potential channels', other_designations='phosphoinositide interacting protein', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=66911910, end_position_on_the_genomic_accession=66929877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  61933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61933, CurrentID=0, Status='live', Symbol='D5Mit369', Aliases='', description='DNA Segment, Chr 5, Massachusetts Institute of Technology 369', other_designations='', map_location='5 63.38 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61943, CurrentID=0, Status='live', Symbol='D6Mit316', Aliases='', description='DNA Segment, Chr 6, Massachusetts Institute of Technology 316', other_designations='', map_location='6 27.41 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61948, CurrentID=0, Status='live', Symbol='D6Mit328', Aliases='', description='DNA Segment, Chr 6, Massachusetts Institute of Technology 328', other_designations='', map_location='6 52.62 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61949, CurrentID=0, Status='live', Symbol='D6Mit329', Aliases='', description='DNA Segment, Chr 6, Massachusetts Institute of Technology 329', other_designations='', map_location='6 53.01 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61952, CurrentID=0, Status='live', Symbol='D6Mit339', Aliases='', description='DNA Segment, Chr 6, Massachusetts Institute of Technology 339', other_designations='', map_location='6 66.59 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61957, CurrentID=0, Status='live', Symbol='D7Mit321', Aliases='', description='DNA Segment, Chr 7, Massachusetts Institute of Technology 321', other_designations='', map_location='7 53.57 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61958, CurrentID=0, Status='live', Symbol='D7Mit326', Aliases='', description='DNA Segment, Chr 7, Massachusetts Institute of Technology 326', other_designations='', map_location='7 54.15 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111176, CurrentID=0, Status='live', Symbol='pwk', Aliases='', description='patchwork', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193034, CurrentID=0, Status='live', Symbol='Trpv1', Aliases='OTRPC1, TRPV1alpha, TRPV1beta, VR-1, Vr1', description='transient receptor potential cation channel, subfamily V, member 1', other_designations='capsaicin receptor|cation channel|deltaN-TRPv1|osm-9-like TRP channel 1|transient receptor potential cation channel V1|vanilloid receptor 1|vanilloid receptor subtype 1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73231024, end_position_on_the_genomic_accession=73261322, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  193043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193043, CurrentID=0, Status='live', Symbol='Zfp3', Aliases='AI854635, B230309J24, Fnp-1, Zfp-3', description='zinc finger protein 3', other_designations='', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70764447, end_position_on_the_genomic_accession=70772928, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  61981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61981, CurrentID=0, Status='live', Symbol='D9Mit297', Aliases='', description='DNA Segment, Chr 9, Massachusetts Institute of Technology 297', other_designations='', map_location='9 18.24 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61989, CurrentID=0, Status='live', Symbol='D9Mit310', Aliases='', description='DNA Segment, Chr 9, Massachusetts Institute of Technology 310', other_designations='', map_location='9 57.34 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61993, CurrentID=0, Status='live', Symbol='DXMit189', Aliases='', description='DNA Segment, Chr X, Massachusetts Institute of Technology 189', other_designations='', map_location='X 73.93 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  61994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61994, CurrentID=0, Status='live', Symbol='D4Mit339', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 339', other_designations='', map_location='4 66.64 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62009, CurrentID=0, Status='live', Symbol='D6Mit374a', Aliases='D6Mit374', description='DNA segment, Chr 6, Massachusetts Institute of Technology 374a', other_designations='', map_location='6 64.6 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62010, CurrentID=0, Status='live', Symbol='D6Mit373', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 373', other_designations='', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62014, CurrentID=0, Status='live', Symbol='D6Mit355', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 355', other_designations='', map_location='6 25.86 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62041, CurrentID=0, Status='live', Symbol='D6Mit384', Aliases='', description='DNA Segment, Chr 6, Massachusetts Institute of Technology 384', other_designations='', map_location='6 27.38 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193116, CurrentID=0, Status='live', Symbol='Slu7', Aliases='AU018913, D11Ertd730e, D3Bwg0878e', description='SLU7 splicing factor homolog (S. cerevisiae)', other_designations='step II splicing factor SLU7', map_location='11 A5-B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43433723, end_position_on_the_genomic_accession=43447982, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  62045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62045, CurrentID=0, Status='live', Symbol='D9Mit355', Aliases='', description='DNA Segment, Chr 9, Massachusetts Institute of Technology 355', other_designations='', map_location='9 51.41 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62057, CurrentID=0, Status='live', Symbol='DXMit249', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 249', other_designations='', map_location='X 76.03 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62058, CurrentID=0, Status='live', Symbol='D10Mit12', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 12', other_designations='', map_location='10 51.15 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62060, CurrentID=0, Status='live', Symbol='D10Mit14', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 14', other_designations='', map_location='10 66.75 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62061, CurrentID=0, Status='live', Symbol='D10Mit16', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 16', other_designations='', map_location='10 9.82 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62062, CurrentID=0, Status='live', Symbol='D10Mit17', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 17', other_designations='', map_location='10 9.78 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62065, CurrentID=0, Status='live', Symbol='D10Mit35a', Aliases='D10Mit35', description='DNA segment, Chr 10, Massachusetts Institute of Technology 35a', other_designations='', map_location='10 70.36 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62066, CurrentID=0, Status='live', Symbol='D10Mit41', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 41', other_designations='', map_location='10 48.79 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62067, CurrentID=0, Status='live', Symbol='D10Mit42', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 42', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62069, CurrentID=0, Status='live', Symbol='D10Mit46', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 46', other_designations='', map_location='10 64.71 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62072, CurrentID=0, Status='live', Symbol='D10Mit54', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 54', other_designations='', map_location='10 23.02 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62073, CurrentID=0, Status='live', Symbol='D10Mit55', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 55', other_designations='', map_location='10 23.07 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62075, CurrentID=0, Status='live', Symbol='D10Mit59', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 59', other_designations='', map_location='10 28.84 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62076, CurrentID=0, Status='live', Symbol='D10Mit61', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 61', other_designations='', map_location='10 34.8 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62080, CurrentID=0, Status='live', Symbol='D10Mit65', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 65', other_designations='', map_location='10 41.33 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62082, CurrentID=0, Status='live', Symbol='D10Mit68', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 68', other_designations='', map_location='10 47.46 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62083, CurrentID=0, Status='live', Symbol='D10Mit69', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 69', other_designations='', map_location='10 49.23 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62084, CurrentID=0, Status='live', Symbol='D10Mit70', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 70', other_designations='', map_location='10 54.72 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62087, CurrentID=0, Status='live', Symbol='D10Mit87', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 87', other_designations='', map_location='10 12.66 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62092, CurrentID=0, Status='live', Symbol='D11Mit11', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 11', other_designations='', map_location='11 75.67 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62093, CurrentID=0, Status='live', Symbol='D11Mit14', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 14', other_designations='', map_location='11 62.4 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62094, CurrentID=0, Status='live', Symbol='D11Mit22', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 22', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62095, CurrentID=0, Status='live', Symbol='D11Mit51', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 51', other_designations='', map_location='11 21.96 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62096, CurrentID=0, Status='live', Symbol='D11Mit61', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 61', other_designations='', map_location='11 75.93 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62097, CurrentID=0, Status='live', Symbol='D11Mit72', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 72', other_designations='', map_location='11 2.94 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62098, CurrentID=0, Status='live', Symbol='D11Mit79', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 79', other_designations='', map_location='11 12.35 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62100, CurrentID=0, Status='live', Symbol='D11Mit84', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 84', other_designations='', map_location='11 19.82 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62101, CurrentID=0, Status='live', Symbol='D11Mit87', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 87', other_designations='', map_location='11 31.86 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62102, CurrentID=0, Status='live', Symbol='D11Mit90', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 90', other_designations='', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62104, CurrentID=0, Status='live', Symbol='D11Mit98', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 98', other_designations='', map_location='11 60.14 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62105, CurrentID=0, Status='live', Symbol='D11Mit99', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 99', other_designations='', map_location='11 62.99 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62107, CurrentID=0, Status='live', Symbol='D12Mit54', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 54', other_designations='', map_location='12 22.76 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62108, CurrentID=0, Status='live', Symbol='D12Mit60', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 60', other_designations='', map_location='12 15.54 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62121, CurrentID=0, Status='live', Symbol='D12Mit97', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 97', other_designations='', map_location='12 50.43 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62124, CurrentID=0, Status='live', Symbol='D13Mit13', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 13', other_designations='', map_location='13 30.06 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62125, CurrentID=0, Status='live', Symbol='D13Mit16', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 16', other_designations='', map_location='13 7.26 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62126, CurrentID=0, Status='live', Symbol='D13Mit18', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 18', other_designations='', map_location='13 14.44 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62127, CurrentID=0, Status='live', Symbol='D13Mit32', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 32', other_designations='', map_location='13 64.06 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62128, CurrentID=0, Status='live', Symbol='D13Mit35', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 35', other_designations='', map_location='13 67.21 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62129, CurrentID=0, Status='live', Symbol='D13Mit38', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 38', other_designations='', map_location='13 17.62 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62131, CurrentID=0, Status='live', Symbol='D13Mit47', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 47', other_designations='', map_location='13 54.93 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62134, CurrentID=0, Status='live', Symbol='D13Mit63', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 63', other_designations='', map_location='13 21.0 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62136, CurrentID=0, Status='live', Symbol='D13Mit66', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 66', other_designations='', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62138, CurrentID=0, Status='live', Symbol='D13Mit70', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 70', other_designations='', map_location='13 54.28 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62140, CurrentID=0, Status='live', Symbol='D13Mit72', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 72', other_designations='', map_location='13 53.81 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62141, CurrentID=0, Status='live', Symbol='D13Mit76', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 76', other_designations='', map_location='13 62.35 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62142, CurrentID=0, Status='live', Symbol='D13Mit78', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 78', other_designations='', map_location='13 67.21 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193217, CurrentID=0, Status='live', Symbol='BC018473', Aliases='', description='cDNA sequence BC018473', other_designations='hitchhiker', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116752167, end_position_on_the_genomic_accession=116757883, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  62146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62146, CurrentID=0, Status='live', Symbol='D13Mit87', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 87', other_designations='', map_location='13 18.41 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62149, CurrentID=0, Status='live', Symbol='D13Mit91', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 91', other_designations='', map_location='13 24.5 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62150, CurrentID=0, Status='live', Symbol='D13Mit94', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 94', other_designations='', map_location='13 24.5 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62155, CurrentID=0, Status='live', Symbol='D14Mit37', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 37', other_designations='', map_location='14 33.21 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62156, CurrentID=0, Status='live', Symbol='D14Mit38', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 38', other_designations='', map_location='14 50.9 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62157, CurrentID=0, Status='live', Symbol='D14Mit39', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 39', other_designations='', map_location='14 35.69 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62158, CurrentID=0, Status='live', Symbol='D14Mit45', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 45', other_designations='', map_location='14 20.81 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62162, CurrentID=0, Status='live', Symbol='D14Mit60', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 60', other_designations='', map_location='14 24.6 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62164, CurrentID=0, Status='live', Symbol='D14Mit62', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 62', other_designations='', map_location='14 26.46 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62178, CurrentID=0, Status='live', Symbol='D15Mit11', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 11', other_designations='', map_location='15 4.39 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62179, CurrentID=0, Status='live', Symbol='D15Mit14', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 14', other_designations='', map_location='15 56.13 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62181, CurrentID=0, Status='live', Symbol='D15Mit22', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 22', other_designations='', map_location='15 15.1 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62183, CurrentID=0, Status='live', Symbol='D15Mit33', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 33', other_designations='', map_location='15 39.43 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62185, CurrentID=0, Status='live', Symbol='D15Mit40', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 40', other_designations='', map_location='15 57.22 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62186, CurrentID=0, Status='live', Symbol='D15Mit41', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 41', other_designations='', map_location='15 53.99 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62187, CurrentID=0, Status='live', Symbol='D15Mit44', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 44', other_designations='', map_location='15 55.99 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62193, CurrentID=0, Status='live', Symbol='D15Mit62', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 62', other_designations='', map_location='15 29.03 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62198, CurrentID=0, Status='live', Symbol='D15Mit71', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 71', other_designations='', map_location='15 37.8 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62199, CurrentID=0, Status='live', Symbol='D15Mit72', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 72', other_designations='', map_location='15 39.98 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62202, CurrentID=0, Status='live', Symbol='D15Mit79', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 79', other_designations='', map_location='15 58.97 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62205, CurrentID=0, Status='live', Symbol='D15Mit96', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 96', other_designations='', map_location='15 41.95 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62206, CurrentID=0, Status='live', Symbol='D16Mit19', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 19', other_designations='', map_location='16 45.36 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62208, CurrentID=0, Status='live', Symbol='D16Mit34', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 34', other_designations='', map_location='16 10.09 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62210, CurrentID=0, Status='live', Symbol='D16Mit38', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 38', other_designations='', map_location='16 28.16 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62211, CurrentID=0, Status='live', Symbol='D16Mit39', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 39', other_designations='', map_location='16 28.44 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193286, CurrentID=0, Status='live', Symbol='BC049762', Aliases='4930503F14', description='cDNA sequence BC049762', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51253645, end_position_on_the_genomic_accession=51265635, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  62215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62215, CurrentID=0, Status='live', Symbol='D16Mit50', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 50', other_designations='', map_location='16 45.41 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62216, CurrentID=0, Status='live', Symbol='D16Mit58', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 58', other_designations='', map_location='16 22.55 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62217, CurrentID=0, Status='live', Symbol='D16Mit59', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 59', other_designations='', map_location='16 26.86 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62218, CurrentID=0, Status='live', Symbol='D16Mit60', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 60', other_designations='', map_location='16 23.27 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62219, CurrentID=0, Status='live', Symbol='D16Mit64', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 64', other_designations='', map_location='16 34.22 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62221, CurrentID=0, Status='live', Symbol='D16Mit84', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 84', other_designations='', map_location='16 29.53 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62222, CurrentID=0, Status='live', Symbol='D16Mit87', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 87', other_designations='', map_location='16 5.34 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62224, CurrentID=0, Status='live', Symbol='D16Mit93', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 93', other_designations='', map_location='16 43.22 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62226, CurrentID=0, Status='live', Symbol='D17Mit29', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 29', other_designations='', map_location='17 14.84 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62228, CurrentID=0, Status='live', Symbol='D17Mit34', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 34', other_designations='', map_location='17 18.31 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62229, CurrentID=0, Status='live', Symbol='D17Mit38', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 38', other_designations='', map_location='17 45.64 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62230, CurrentID=0, Status='live', Symbol='D17Mit39', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 39', other_designations='', map_location='17 45.64 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62231, CurrentID=0, Status='live', Symbol='D17Mit41', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 41', other_designations='', map_location='17 55.02 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62233, CurrentID=0, Status='live', Symbol='D17Mit44', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 44', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62234, CurrentID=0, Status='live', Symbol='D17Mit51', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 51', other_designations='', map_location='17 19.74 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62235, CurrentID=0, Status='live', Symbol='D17Mit52', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 52', other_designations='', map_location='17 19.74 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62237, CurrentID=0, Status='live', Symbol='D17Mit60', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 60', other_designations='', map_location='17 15.64 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62238, CurrentID=0, Status='live', Symbol='D17Mit66', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 66', other_designations='', map_location='17 23.97 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62244, CurrentID=0, Status='live', Symbol='D17Mit76', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 76', other_designations='', map_location='17 55.48 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62248, CurrentID=0, Status='live', Symbol='D18Mit24', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 24', other_designations='', map_location='18 27.17 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62249, CurrentID=0, Status='live', Symbol='D18Mit25', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 25', other_designations='', map_location='18 59.05 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62250, CurrentID=0, Status='live', Symbol='D18Mit33', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 33', other_designations='', map_location='18 43.49 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62251, CurrentID=0, Status='live', Symbol='D18Mit34', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 34', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62252, CurrentID=0, Status='live', Symbol='D18Mit35', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 35', other_designations='', map_location='18 23.86 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62253, CurrentID=0, Status='live', Symbol='D18Mit36', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 36', other_designations='', map_location='18 24.64 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62254, CurrentID=0, Status='live', Symbol='D18Mit37', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 37', other_designations='', map_location='18 22.63 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62258, CurrentID=0, Status='live', Symbol='D18Mit53', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 53', other_designations='', map_location='18 28.28 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62259, CurrentID=0, Status='live', Symbol='D18Mit57', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 57', other_designations='', map_location='18 25.47 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62261, CurrentID=0, Status='live', Symbol='D18Mit88', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 88', other_designations='', map_location='18 17.99 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62266, CurrentID=0, Status='live', Symbol='D19Mit11.1', Aliases='D19Mit11', description='DNA segment, Chr 19, Massachusetts Institute of Technology 11.1', other_designations='', map_location='19 36.26 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62267, CurrentID=0, Status='live', Symbol='D19Mit16', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 16', other_designations='', map_location='19 13.82 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62268, CurrentID=0, Status='live', Symbol='D19Mit17', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 17', other_designations='', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62269, CurrentID=0, Status='live', Symbol='D19Mit19', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 19', other_designations='', map_location='19 34.08 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62270, CurrentID=0, Status='live', Symbol='D19Mit23', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 23', other_designations='', map_location='19 13.85 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62271, CurrentID=0, Status='live', Symbol='D19Mit24', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 24', other_designations='', map_location='19 38.16 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62273, CurrentID=0, Status='live', Symbol='D19Mit27', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 27', other_designations='', map_location='19 38.15 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62274, CurrentID=0, Status='live', Symbol='D19Mit28', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 28', other_designations='', map_location='19 10.85 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62275, CurrentID=0, Status='live', Symbol='D19Mit30', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 30', other_designations='', map_location='19 21.34 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62276, CurrentID=0, Status='live', Symbol='D19Mit35', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 35', other_designations='', map_location='19 51.41 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62278, CurrentID=0, Status='live', Symbol='D19Mit38', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 38', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62279, CurrentID=0, Status='live', Symbol='D19Mit39', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 39', other_designations='', map_location='19 23.68 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62280, CurrentID=0, Status='live', Symbol='D19Mit41', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 41', other_designations='', map_location='19 13.18 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62282, CurrentID=0, Status='live', Symbol='D19Mit46', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 46', other_designations='', map_location='19 29.06 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62285, CurrentID=0, Status='live', Symbol='D19Mit53', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 53', other_designations='', map_location='19 38.46 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62290, CurrentID=0, Status='live', Symbol='D19Mit60', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 60', other_designations='', map_location='19 13.9 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62291, CurrentID=0, Status='live', Symbol='D19Mit63', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 63', other_designations='', map_location='19 30.3 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62292, CurrentID=0, Status='live', Symbol='D19Mit65', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 65', other_designations='', map_location='19 33.55 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62294, CurrentID=0, Status='live', Symbol='D19Mit71', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 71', other_designations='', map_location='19 56.28 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62295, CurrentID=0, Status='live', Symbol='D1Mit1001', Aliases='D19Mit72', description='DNA segment, Chr 1, Massachusetts Institute of Technology 1001', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62304, CurrentID=0, Status='live', Symbol='D1Mit155', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 155', other_designations='', map_location='1 98.2 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62305, CurrentID=0, Status='live', Symbol='D1Mit166', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 166', other_designations='', map_location='1 82.24 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62307, CurrentID=0, Status='live', Symbol='D1Mit210', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 210', other_designations='', map_location='1 97.55 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62308, CurrentID=0, Status='live', Symbol='D1Mit221', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 221', other_designations='', map_location='1 89.46 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62311, CurrentID=0, Status='live', Symbol='D1Mit273', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 273', other_designations='', map_location='1 84.79 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193385, CurrentID=0, Status='live', Symbol='Fam65b', Aliases='', description='family with sequence similarity 65, member B', other_designations='', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24501534, end_position_on_the_genomic_accession=24733806, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  62315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62315, CurrentID=0, Status='live', Symbol='D2Mit113', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 113', other_designations='', map_location='2 96.21 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62317, CurrentID=0, Status='live', Symbol='D2Mit147', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 147', other_designations='', map_location='2 94.85 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62318, CurrentID=0, Status='live', Symbol='D2Mit200', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 200', other_designations='', map_location='2 102.29 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62320, CurrentID=0, Status='live', Symbol='D2Mit266', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 266', other_designations='', map_location='2 103.83 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62322, CurrentID=0, Status='live', Symbol='D6Mit107', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 107', other_designations='', map_location='6 52.31 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62323, CurrentID=0, Status='live', Symbol='D6Mit118', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 118', other_designations='', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62325, CurrentID=0, Status='live', Symbol='D6Mit128', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 128', other_designations='', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62330, CurrentID=0, Status='live', Symbol='D6Mit184', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 184', other_designations='', map_location='6 25.81 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62332, CurrentID=0, Status='live', Symbol='D6Mit256', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 256', other_designations='', map_location='6 60.85 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62335, CurrentID=0, Status='live', Symbol='D19Mit99', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 99', other_designations='', map_location='19 21.25 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62338, CurrentID=0, Status='live', Symbol='D19Mit91', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 91', other_designations='', map_location='19 40.53 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62339, CurrentID=0, Status='live', Symbol='D19Mit86', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 86', other_designations='', map_location='19 20.99 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62347, CurrentID=0, Status='live', Symbol='D1Mit406', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 406', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62350, CurrentID=0, Status='live', Symbol='D1Mit403', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 403', other_designations='', map_location='1 81.63 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62353, CurrentID=0, Status='live', Symbol='D1Mit315', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 315', other_designations='', map_location='1 83.48 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62356, CurrentID=0, Status='live', Symbol='D1Mit459', Aliases='', description='DNA Segment, Chr 1, Massachusetts Institute of Technology 459', other_designations='', map_location='1 91.86 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62357, CurrentID=0, Status='live', Symbol='D1Mit462', Aliases='', description='DNA Segment, Chr 1, Massachusetts Institute of Technology 462', other_designations='', map_location='1 96.27 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62370, CurrentID=0, Status='live', Symbol='D10Mit168', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 168', other_designations='', map_location='10 7.94 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62371, CurrentID=0, Status='live', Symbol='D11Mit133', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 133', other_designations='', map_location='11 8.01 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62372, CurrentID=0, Status='live', Symbol='D11Mit162', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 162', other_designations='', map_location='11 10.57 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62377, CurrentID=0, Status='live', Symbol='D13Mit114', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 114', other_designations='', map_location='13 5.12 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62378, CurrentID=0, Status='live', Symbol='D13Mit134', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 134', other_designations='', map_location='13 4.45 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62379, CurrentID=0, Status='live', Symbol='D13Mit158', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 158', other_designations='', map_location='13 2.74 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62380, CurrentID=0, Status='live', Symbol='D13Mit173', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 173', other_designations='', map_location='13 4.55 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62391, CurrentID=0, Status='live', Symbol='D16Mit29', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 29', other_designations='', map_location='16 12.38 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62392, CurrentID=0, Status='live', Symbol='D16Mit51', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 51', other_designations='', map_location='16 53.78 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62394, CurrentID=0, Status='live', Symbol='D17Mit113', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 113', other_designations='', map_location='17 8.14 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62395, CurrentID=0, Status='live', Symbol='D17Mit16', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 16', other_designations='', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62396, CurrentID=0, Status='live', Symbol='D17Mit62', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 62', other_designations='', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62397, CurrentID=0, Status='live', Symbol='D17Mit63', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 63', other_designations='', map_location='17 17.59 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62399, CurrentID=0, Status='live', Symbol='D18Mit117', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 117', other_designations='', map_location='18 11.24 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62401, CurrentID=0, Status='live', Symbol='D18Mit146', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 146', other_designations='', map_location='18 6.36 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62402, CurrentID=0, Status='live', Symbol='D17Mit196', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 196', other_designations='', map_location='17 3.58 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62411, CurrentID=0, Status='live', Symbol='D16Mit130', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 130', other_designations='', map_location='16 2.65 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62417, CurrentID=0, Status='live', Symbol='D14Mit207', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 207', other_designations='', map_location='14 12.07 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62426, CurrentID=0, Status='live', Symbol='D12Mit182', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 182', other_designations='', map_location='12 5.52 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62436, CurrentID=0, Status='live', Symbol='D19Mit109', Aliases='', description='DNA Segment, Chr 19, Massachusetts Institute of Technology 109', other_designations='', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62439, CurrentID=0, Status='live', Symbol='D12Mit269', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 269', other_designations='', map_location='12 8.04 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62442, CurrentID=0, Status='live', Symbol='D10Mit282', Aliases='', description='DNA Segment, Chr 10 Massachusetts Institute of Technology 282', other_designations='', map_location='10 9.75 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62444, CurrentID=0, Status='live', Symbol='D10Mit286', Aliases='', description='DNA Segment, Chr 10 Massachusetts Institute of Technology 286', other_designations='', map_location='10 29.26 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62452, CurrentID=0, Status='live', Symbol='D11Mit360', Aliases='', description='DNA Segment, Chr 11 Massachusetts Institute of Technology 360', other_designations='', map_location='11 67.13 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62458, CurrentID=0, Status='live', Symbol='D14Mit259', Aliases='', description='DNA Segment, Chr 14 Massachusetts Institute of Technology 259', other_designations='', map_location='14 27.65 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62461, CurrentID=0, Status='live', Symbol='D14Mit265', Aliases='', description='DNA Segment, Chr 14 Massachusetts Institute of Technology 265', other_designations='', map_location='14 50.9 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62473, CurrentID=0, Status='live', Symbol='D19Mit123', Aliases='', description='DNA Segment, Chr 19 Massachusetts Institute of Technology 123', other_designations='', map_location='19 45.06 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  381783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381783, CurrentID=0, Status='live', Symbol='Igkv5-43', Aliases='Igk-V23, Igkv5-45', description='immunoglobulin kappa chain variable 5-43', other_designations='Ig kappa-chain V-region|Vk23 Ig variable region|immunoglobulin kappa chain variable 23 (V23)|immunoglobulin kappa light chain variable region Vk23', map_location='6 31.63 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69823355, end_position_on_the_genomic_accession=69823911, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  62480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62480, CurrentID=0, Status='live', Symbol='D10Mit103', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 103', other_designations='', map_location='10 73.57 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62482, CurrentID=0, Status='live', Symbol='D10Mit106', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 106', other_designations='', map_location='10 11.67 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62483, CurrentID=0, Status='live', Symbol='D10Mit109', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 109', other_designations='', map_location='10 22.53 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62484, CurrentID=0, Status='live', Symbol='D10Mit113', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 113', other_designations='', map_location='10 34.95 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62485, CurrentID=0, Status='live', Symbol='D10Mit114', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 114', other_designations='', map_location='10 36.32 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62486, CurrentID=0, Status='live', Symbol='D10Mit116', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 116', other_designations='', map_location='10 38.51 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62487, CurrentID=0, Status='live', Symbol='D10Mit117', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 117', other_designations='', map_location='10 43.66 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62490, CurrentID=0, Status='live', Symbol='D10Mit130', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 130', other_designations='', map_location='10 34.72 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62491, CurrentID=0, Status='live', Symbol='D10Mit133', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 133', other_designations='', map_location='10 54.72 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62496, CurrentID=0, Status='live', Symbol='D10Mit140', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 140', other_designations='', map_location='10 39.93 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62498, CurrentID=0, Status='live', Symbol='D10Mit145', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 145', other_designations='', map_location='10 72.35 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62500, CurrentID=0, Status='live', Symbol='D10Mit158', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 158', other_designations='', map_location='10 38.58 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62505, CurrentID=0, Status='live', Symbol='D10Mit170', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 170', other_designations='', map_location='10 24.71 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62506, CurrentID=0, Status='live', Symbol='D10Mit175', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 175', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62508, CurrentID=0, Status='live', Symbol='D10Mit178', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 178', other_designations='', map_location='10 51.42 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62509, CurrentID=0, Status='live', Symbol='D10Mit179', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 179', other_designations='', map_location='10 65.36 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62514, CurrentID=0, Status='live', Symbol='D11Mit117', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 117', other_designations='', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62515, CurrentID=0, Status='live', Symbol='D11Mit122', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 122', other_designations='', map_location='11 58.33 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62518, CurrentID=0, Status='live', Symbol='D11Mit128', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 128', other_designations='', map_location='11 79.52 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62519, CurrentID=0, Status='live', Symbol='D11Mit129', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 129', other_designations='', map_location='11 5.51 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62520, CurrentID=0, Status='live', Symbol='D11Mit132', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 132', other_designations='', map_location='11 62.92 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62523, CurrentID=0, Status='live', Symbol='D11Mit139', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 139', other_designations='', map_location='11 29.23 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62524, CurrentID=0, Status='live', Symbol='D11Mit140', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 140', other_designations='', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62526, CurrentID=0, Status='live', Symbol='D11Mit143', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 143', other_designations='', map_location='11 35.4 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62527, CurrentID=0, Status='live', Symbol='D11Mit151', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 151', other_designations='', map_location='11 15.29 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62528, CurrentID=0, Status='live', Symbol='D11Mit152', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 152', other_designations='', map_location='11 14.44 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62529, CurrentID=0, Status='live', Symbol='D11Mit153', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 153', other_designations='', map_location='11 29.93 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62532, CurrentID=0, Status='live', Symbol='D11Mit160', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 160', other_designations='', map_location='11 60.15 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62533, CurrentID=0, Status='live', Symbol='D11Mit164', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 164', other_designations='', map_location='11 34.5 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62535, CurrentID=0, Status='live', Symbol='D11Mit168', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 168', other_designations='', map_location='11 78.74 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62537, CurrentID=0, Status='live', Symbol='D11Mit174', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 174', other_designations='', map_location='11 25.18 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62539, CurrentID=0, Status='live', Symbol='D11Mit179', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 179', other_designations='', map_location='11 54.6 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62540, CurrentID=0, Status='live', Symbol='D11Mit181', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 181', other_designations='', map_location='11 71.88 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62546, CurrentID=0, Status='live', Symbol='D11Mit198', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 198', other_designations='', map_location='11 62.92 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62547, CurrentID=0, Status='live', Symbol='D11Mit199', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 199', other_designations='', map_location='11 65.48 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62548, CurrentID=0, Status='live', Symbol='D11Mit202', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 202', other_designations='', map_location='11 75.23 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62550, CurrentID=0, Status='live', Symbol='D11Mit208', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 208', other_designations='', map_location='11 36.21 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62552, CurrentID=0, Status='live', Symbol='D11Mit212', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 212', other_designations='', map_location='11 54.34 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62553, CurrentID=0, Status='live', Symbol='D11Mit213', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 213', other_designations='', map_location='11 58.9 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62565, CurrentID=0, Status='live', Symbol='D12Mit132', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 132', other_designations='', map_location='12 57.68 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62566, CurrentID=0, Status='live', Symbol='D12Mit133', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 133', other_designations='', map_location='12 60.56 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62567, CurrentID=0, Status='live', Symbol='D12Mit134', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 134', other_designations='', map_location='12 62.45 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62568, CurrentID=0, Status='live', Symbol='D12Mit136', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 136', other_designations='', map_location='12 13.0 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62569, CurrentID=0, Status='live', Symbol='D12Mit141', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 141', other_designations='', map_location='12 60.61 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62574, CurrentID=0, Status='live', Symbol='D12Mit153', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 153', other_designations='', map_location='12 15.71 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62576, CurrentID=0, Status='live', Symbol='D12Mit156', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 156', other_designations='', map_location='12 35.7 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62578, CurrentID=0, Status='live', Symbol='D12Mit158', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 158', other_designations='', map_location='12 38.14 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62580, CurrentID=0, Status='live', Symbol='D12Mit167', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 167', other_designations='', map_location='12 55.11 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62581, CurrentID=0, Status='live', Symbol='D12Mit171', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 171', other_designations='', map_location='12 11.86 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62586, CurrentID=0, Status='live', Symbol='D12Mit181', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 181', other_designations='', map_location='12 59.59 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62589, CurrentID=0, Status='live', Symbol='D13Mit104', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 104', other_designations='', map_location='13 50.13 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62590, CurrentID=0, Status='live', Symbol='D13Mit106', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 106', other_designations='', map_location='13 47.75 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62591, CurrentID=0, Status='live', Symbol='D13Mit108', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 108', other_designations='', map_location='13 49.22 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62592, CurrentID=0, Status='live', Symbol='D13Mit110', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 110', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62594, CurrentID=0, Status='live', Symbol='D13Mit115', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 115', other_designations='', map_location='13 7.34 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62595, CurrentID=0, Status='live', Symbol='D13Mit117', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 117', other_designations='', map_location='13 16.81 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62597, CurrentID=0, Status='live', Symbol='D13Mit126', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 126', other_designations='', map_location='13 45.05 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193670, CurrentID=0, Status='live', Symbol='Rnf185', Aliases='', description='ring finger protein 185', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3415973, end_position_on_the_genomic_accession=3452956, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  62599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62599, CurrentID=0, Status='live', Symbol='D13Mit128', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 128', other_designations='', map_location='13 50.66 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62604, CurrentID=0, Status='live', Symbol='D13Mit139', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 139', other_designations='', map_location='13 26.25 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62607, CurrentID=0, Status='live', Symbol='D13Mit144', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 144', other_designations='', map_location='13 50.5 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62608, CurrentID=0, Status='live', Symbol='D13Mit148', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 148', other_designations='', map_location='13 59.69 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62612, CurrentID=0, Status='live', Symbol='D13Mit160', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 160', other_designations='', map_location='13 49.23 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62613, CurrentID=0, Status='live', Symbol='D13Mit161', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 161', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62616, CurrentID=0, Status='live', Symbol='D13Mit169', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 169', other_designations='', map_location='13 49.22 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62620, CurrentID=0, Status='live', Symbol='D13Mit196', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 196', other_designations='', map_location='13 63.93 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62622, CurrentID=0, Status='live', Symbol='D13Mit202', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 202', other_designations='', map_location='13 47.43 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62625, CurrentID=0, Status='live', Symbol='D14Mit101', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 101', other_designations='', map_location='14 25.69 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62629, CurrentID=0, Status='live', Symbol='D14Mit113', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 113', other_designations='', map_location='14 31.51 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62632, CurrentID=0, Status='live', Symbol='D14Mit121', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 121', other_designations='', map_location='14 25.36 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62637, CurrentID=0, Status='live', Symbol='D14Mit132', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 132', other_designations='', map_location='14 6.03 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62639, CurrentID=0, Status='live', Symbol='D14Mit141', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 141', other_designations='', map_location='14 24.28 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62640, CurrentID=0, Status='live', Symbol='D14Mit142', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 142', other_designations='', map_location='14 27.68 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62643, CurrentID=0, Status='live', Symbol='D14Mit154', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 154', other_designations='', map_location='14 30.51 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62644, CurrentID=0, Status='live', Symbol='D14Mit155', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 155', other_designations='', map_location='14 31.52 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62648, CurrentID=0, Status='live', Symbol='D14Mit160', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 160', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62651, CurrentID=0, Status='live', Symbol='D14Mit166', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 166', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62656, CurrentID=0, Status='live', Symbol='D15Mit100', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 100', other_designations='', map_location='15 19.26 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62657, CurrentID=0, Status='live', Symbol='D15Mit105', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 105', other_designations='', map_location='15 33.42 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62658, CurrentID=0, Status='live', Symbol='D15Mit107', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 107', other_designations='', map_location='15 39.79 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62659, CurrentID=0, Status='live', Symbol='D15Mit108', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 108', other_designations='', map_location='15 47.79 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193736, CurrentID=0, Status='live', Symbol='Zbtb12', Aliases='BC020447, Bat-9, Bat9, G10, NG35', description='zinc finger and BTB domain containing 12', other_designations='HLA-B-associated transcript 9', map_location='17 18.43 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34894559, end_position_on_the_genomic_accession=34896867, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  193740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193740, CurrentID=0, Status='live', Symbol='Hspa1a', Aliases='Hsp70-3, Hsp70.3, Hsp72, hsp68, hsp70A1', description='heat shock protein 1A', other_designations='68 kDa heat shock protein|heat shock 70 kDa protein 3|heat shock 70kDa protein 1A|heat shock protein, 70 kDa 3|inducible heat shock protein 70', map_location='17 18.51 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34969359, end_position_on_the_genomic_accession=34972156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  62670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62670, CurrentID=0, Status='live', Symbol='D15Mit143', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 143', other_designations='', map_location='15 19.62 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62671, CurrentID=0, Status='live', Symbol='D15Mit144', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 144', other_designations='', map_location='15 29.44 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62675, CurrentID=0, Status='live', Symbol='D15Mit159', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 159', other_designations='', map_location='15 41.96 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62677, CurrentID=0, Status='live', Symbol='D15Mit161', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 161', other_designations='', map_location='15 52.78 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62678, CurrentID=0, Status='live', Symbol='D16Mit101', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 101', other_designations='', map_location='16 14.69 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62679, CurrentID=0, Status='live', Symbol='D16Mit103', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 103', other_designations='', map_location='16 22.4 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62686, CurrentID=0, Status='live', Symbol='D16Mit122', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 122', other_designations='', map_location='16 3.43 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62689, CurrentID=0, Status='live', Symbol='D16Mit128', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 128', other_designations='', map_location='16 53.72 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62690, CurrentID=0, Status='live', Symbol='D17Mit101', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 101', other_designations='', map_location='17 15.17 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62693, CurrentID=0, Status='live', Symbol='D17Mit115', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 115', other_designations='', map_location='17 23.56 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62695, CurrentID=0, Status='live', Symbol='D17Mit119', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 119', other_designations='', map_location='17 38.15 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62697, CurrentID=0, Status='live', Symbol='D17Mit123', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 123', other_designations='', map_location='17 60.67 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62698, CurrentID=0, Status='live', Symbol='D17Mit129', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 129', other_designations='', map_location='17 56.83 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62699, CurrentID=0, Status='live', Symbol='D17Mit130', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 130', other_designations='', map_location='17 57.39 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62702, CurrentID=0, Status='live', Symbol='D17Mit138', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 138', other_designations='', map_location='17 22.89 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62703, CurrentID=0, Status='live', Symbol='D17Mit139', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 139', other_designations='', map_location='17 27.4 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62710, CurrentID=0, Status='live', Symbol='D17Mit155', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 155', other_designations='', map_location='17 55.02 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62713, CurrentID=0, Status='live', Symbol='D18Mit103', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 103', other_designations='', map_location='18 44.19 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62715, CurrentID=0, Status='live', Symbol='D18Mit111', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 111', other_designations='', map_location='18 11.96 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62718, CurrentID=0, Status='live', Symbol='D18Mit122', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 122', other_designations='', map_location='18 30.77 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62719, CurrentID=0, Status='live', Symbol='D18Mit124', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 124', other_designations='', map_location='18 32.15 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62722, CurrentID=0, Status='live', Symbol='D18Mit141', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 141', other_designations='', map_location='18 45.63 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193796, CurrentID=0, Status='live', Symbol='Kdm4b', Aliases='4732474L06Rik, Jmjd2b, mKIAA0876', description='lysine (K)-specific demethylase 4B', other_designations='jmjC domain-containing histone demethylation protein 3B|jumonji domain containing 2B|jumonji domain-containing protein 2B', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56326033, end_position_on_the_genomic_accession=56403769, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  62730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62730, CurrentID=0, Status='live', Symbol='D19Mit101', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 101', other_designations='', map_location='19 32.35 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62732, CurrentID=0, Status='live', Symbol='D18Mit189', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 189', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62735, CurrentID=0, Status='live', Symbol='D18Mit184', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 184', other_designations='', map_location='18 39.7 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193813, CurrentID=0, Status='live', Symbol='Mcfd2', Aliases='1810021C21Rik, F5f8d, Lman1ip, Sdnsf', description='multiple coagulation factor deficiency 2', other_designations='neural stem cell derived neuronal survival protein|neural stem cell-derived neuronal survival protein', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87254443, end_position_on_the_genomic_accession=87265947, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  228911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228911, CurrentID=0, Status='live', Symbol='Tshz2', Aliases='2900073F20Rik, B830045G17, Mtsh2, Sdccag33l, Tsh2, Zfp218, mKIAA4248, teashirt2', description='teashirt zinc finger family member 2', other_designations='SDCCAG33-like protein|serologically defined colon cancer antigen 33 like|teashirt 2|zinc finger protein 218', map_location='2 89.47 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=169632496, end_position_on_the_genomic_accession=170071666, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  62753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62753, CurrentID=0, Status='live', Symbol='D17Mit198', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 198', other_designations='', map_location='17 14.59 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62755, CurrentID=0, Status='live', Symbol='D17Mit189', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 189', other_designations='', map_location='17 55.23 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62758, CurrentID=0, Status='live', Symbol='D17Mit185', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 185', other_designations='', map_location='17 39.3 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62761, CurrentID=0, Status='live', Symbol='D17Mit180', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 180', other_designations='', map_location='17 26.71 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62762, CurrentID=0, Status='live', Symbol='D17Mit175', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 175', other_designations='', map_location='17 17.25 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193838, CurrentID=0, Status='live', Symbol='Eme2', Aliases='2810013J18Rik', description='essential meiotic structure-specific endonuclease subunit 2', other_designations='essential meiotic endonuclease 1 homolog 2', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24892152, end_position_on_the_genomic_accession=24895087, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  62770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62770, CurrentID=0, Status='live', Symbol='D16Mit165', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 165', other_designations='', map_location='16 9.66 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62776, CurrentID=0, Status='live', Symbol='D16Mit143', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 143', other_designations='', map_location='16 11.23 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62777, CurrentID=0, Status='live', Symbol='D16Mit140', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 140', other_designations='', map_location='16 40.3 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62778, CurrentID=0, Status='live', Symbol='D16Mit138', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 138', other_designations='', map_location='16 29.93 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62795, CurrentID=0, Status='live', Symbol='D15Mit171', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 171', other_designations='', map_location='15 45.02 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62806, CurrentID=0, Status='live', Symbol='D14Mit203', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 203', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62810, CurrentID=0, Status='live', Symbol='D14Mit193', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 193', other_designations='', map_location='14 37.26 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62816, CurrentID=0, Status='live', Symbol='D13Mit266', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 266', other_designations='', map_location='13 14.44 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62821, CurrentID=0, Status='live', Symbol='D13Mit250', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 250', other_designations='', map_location='13 30.06 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62822, CurrentID=0, Status='live', Symbol='D13Mit249', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 249', other_designations='', map_location='13 27.65 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62823, CurrentID=0, Status='live', Symbol='D13Mit247', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 247', other_designations='', map_location='13 26.12 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62827, CurrentID=0, Status='live', Symbol='D13Mit233', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 233', other_designations='', map_location='13 43.68 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62829, CurrentID=0, Status='live', Symbol='D13Mit231', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 231', other_designations='', map_location='13 40.98 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62837, CurrentID=0, Status='live', Symbol='D12Mit231', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 231', other_designations='', map_location='12 50.87 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62844, CurrentID=0, Status='live', Symbol='D12Mit214', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 214', other_designations='', map_location='12 37.86 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62862, CurrentID=0, Status='live', Symbol='D11Mit294', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 294', other_designations='', map_location='11 13.89 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62865, CurrentID=0, Status='live', Symbol='D11Mit289', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 289', other_designations='', map_location='11 58.98 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62868, CurrentID=0, Status='live', Symbol='D11Mit285', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 285', other_designations='', map_location='11 54.64 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62872, CurrentID=0, Status='live', Symbol='D11Mit271', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 271', other_designations='', map_location='11 27.44 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62875, CurrentID=0, Status='live', Symbol='D11Mit263', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 263', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62876, CurrentID=0, Status='live', Symbol='D11Mit258', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 258', other_designations='', map_location='11 70.29 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62878, CurrentID=0, Status='live', Symbol='D11Mit245', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 245', other_designations='', map_location='11 46.18 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62879, CurrentID=0, Status='live', Symbol='D11Mit236', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 236', other_designations='', map_location='11 27.39 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62883, CurrentID=0, Status='live', Symbol='D11Mit231', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 231', other_designations='', map_location='11 21.92 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62885, CurrentID=0, Status='live', Symbol='D11Mit226', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 226', other_designations='', map_location='11 5.64 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62889, CurrentID=0, Status='live', Symbol='D11Mit219', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 219', other_designations='', map_location='11 44.0 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62896, CurrentID=0, Status='live', Symbol='D10Mit234', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 234', other_designations='', map_location='10 66.75 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62898, CurrentID=0, Status='live', Symbol='D10Mit230', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 230', other_designations='', map_location='10 45.28 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62904, CurrentID=0, Status='live', Symbol='D10Mit214', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 214', other_designations='', map_location='10 12.51 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62905, CurrentID=0, Status='live', Symbol='D10Mit213', Aliases='', description='DNA segment, Chr 10, Massachusetts Institute of Technology 213', other_designations='', map_location='10 9.75 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62922, CurrentID=0, Status='live', Symbol='D10Mit255', Aliases='', description='DNA Segment, Chr 10, Massachusetts Institute of Technology 255', other_designations='', map_location='10 22.83 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62929, CurrentID=0, Status='live', Symbol='D10Mit267', Aliases='', description='DNA Segment, Chr 10, Massachusetts Institute of Technology 267', other_designations='', map_location='10 67.55 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62930, CurrentID=0, Status='live', Symbol='D10Mit271', Aliases='', description='DNA Segment, Chr 10, Massachusetts Institute of Technology 271', other_designations='', map_location='10 72.31 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62931, CurrentID=0, Status='live', Symbol='D10Mit269', Aliases='', description='DNA Segment, Chr 10, Massachusetts Institute of Technology 269', other_designations='', map_location='10 77.2 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62941, CurrentID=0, Status='live', Symbol='D11Mit320', Aliases='', description='DNA Segment, Chr 11, Massachusetts Institute of Technology 320', other_designations='', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62948, CurrentID=0, Status='live', Symbol='D11Mit329', Aliases='', description='DNA Segment, Chr 11, Massachusetts Institute of Technology 329', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62949, CurrentID=0, Status='live', Symbol='D11Mit330', Aliases='', description='DNA Segment, Chr 11, Massachusetts Institute of Technology 330', other_designations='', map_location='11 62.92 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62952, CurrentID=0, Status='live', Symbol='D11Mit336', Aliases='', description='DNA Segment, Chr 11, Massachusetts Institute of Technology 336', other_designations='', map_location='11 73.75 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62960, CurrentID=0, Status='live', Symbol='D12Mit263', Aliases='', description='DNA Segment, Chr 12, Massachusetts Institute of Technology 263', other_designations='', map_location='12 62.11 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62965, CurrentID=0, Status='live', Symbol='D14Mit234', Aliases='', description='DNA Segment, Chr 14, Massachusetts Institute of Technology 234', other_designations='', map_location='14 31.49 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62972, CurrentID=0, Status='live', Symbol='D15Mit239', Aliases='', description='DNA Segment, Chr 15, Massachusetts Institute of Technology 239', other_designations='', map_location='15 37.58 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62973, CurrentID=0, Status='live', Symbol='D15Mit241', Aliases='', description='DNA Segment, Chr 15, Massachusetts Institute of Technology 241', other_designations='', map_location='15 41.4 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62983, CurrentID=0, Status='live', Symbol='D17Mit234', Aliases='', description='DNA Segment, Chr 17, Massachusetts Institute of Technology 234', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62987, CurrentID=0, Status='live', Symbol='D18Mit208', Aliases='', description='DNA Segment, Chr 18, Massachusetts Institute of Technology 208', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62989, CurrentID=0, Status='live', Symbol='D18Mit210', Aliases='', description='DNA Segment, Chr 18, Massachusetts Institute of Technology 210', other_designations='', map_location='18 51.12 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62993, CurrentID=0, Status='live', Symbol='D12Mit280', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 280', other_designations='', map_location='12 60.94 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  62994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62994, CurrentID=0, Status='live', Symbol='D12Mit279', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 279', other_designations='', map_location='12 59.46 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63006, CurrentID=0, Status='live', Symbol='D15Mit270', Aliases='', description='DNA Segment, Chr 15, Massachusetts Institute of Technology 270', other_designations='', map_location='15 27.16 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63018, CurrentID=0, Status='live', Symbol='D11Mit119', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 119', other_designations='', map_location='11 51.23 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63019, CurrentID=0, Status='live', Symbol='D16Mit106', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 106', other_designations='', map_location='16 57.68 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63020, CurrentID=0, Status='live', Symbol='D17Mit100', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 100', other_designations='', map_location='17 13.07 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63022, CurrentID=0, Status='live', Symbol='D17Mit105', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 105', other_designations='', map_location='17 19.58 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63024, CurrentID=0, Status='live', Symbol='D17Mit168', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 168', other_designations='', map_location='17 19.74 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63025, CurrentID=0, Status='live', Symbol='D11Mit260', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 260', other_designations='', map_location='11 37.96 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  194126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194126, CurrentID=0, Status='live', Symbol='Mtmr11', Aliases='AI035306, BC051083, Cra', description='myotubularin related protein 11', other_designations='cisplatin resistance associated', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96161946, end_position_on_the_genomic_accession=96171718, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  194219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194219, CurrentID=0, Status='live', Symbol='Slfnl1', Aliases='4933406A14Rik', description='schlafen like 1', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120531981, end_position_on_the_genomic_accession=120536661, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  194225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194225, CurrentID=0, Status='live', Symbol='Gm13103', Aliases='E330017N17, OTTMUSG00000010433', description='predicted gene 13103', other_designations='novel protein similar to oogenesin 3 (Oog3)', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143846497, end_position_on_the_genomic_accession=143853637, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  194227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194227, CurrentID=0, Status='live', Symbol='Gm13023', Aliases='BC084594, MGC:91194, OTTMUSG00000010009', description='predicted gene 13023', other_designations='novel protein similar to oogenesin (Oog) family', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143789334, end_position_on_the_genomic_accession=143795585, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  194231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194231, CurrentID=0, Status='live', Symbol='Cnksr1', Aliases='BC037758', description='connector enhancer of kinase suppressor of Ras 1', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134228041, end_position_on_the_genomic_accession=134238416, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  97908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97908, CurrentID=0, Status='live', Symbol='Hist1h3g', Aliases='H3.1-221, M32460', description='histone cluster 1, H3g', other_designations='histone 1, H3g', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23535418, end_position_on_the_genomic_accession=23535909, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  194237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194237, CurrentID=0, Status='live', Symbol='Rimkla', Aliases='B930030J24, NAAGS, NAAGS-II, Rimk', description='ribosomal modification protein rimK-like family member A', other_designations='N-acetylaspartyl-glutamate synthetase A|N-acetylaspartylglutamylglutamate synthase A|NAAG synthetase A|ribosomal protein S6 modification-like protein A', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119465279, end_position_on_the_genomic_accession=119493153, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  194268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194268, CurrentID=0, Status='live', Symbol='9930104L06Rik', Aliases='', description='RIKEN cDNA 9930104L06 gene', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124936984, end_position_on_the_genomic_accession=124944661, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  194309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194309, CurrentID=0, Status='live', Symbol='Vps37d', Aliases='Wbscr24', description='vacuolar protein sorting 37D (yeast)', other_designations='ESCRT-I complex subunit VPS37D|Williams-Beuren syndrome chromosome region 24 homolog|Williams-Beuren syndrome critical region protein 24|vps37d|williams-Beuren syndrome region protein 24 homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135072899, end_position_on_the_genomic_accession=135078266, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  194352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194352, CurrentID=0, Status='live', Symbol='Trpv5', Aliases='CAT2, D630033B11, ECAC1', description='transient receptor potential cation channel, subfamily V, member 5', other_designations='OTRPC3|calcium transport protein 2|epithelial calcium channel 1|osm-9-like TRP channel 3', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41651661, end_position_on_the_genomic_accession=41682393, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  194357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194357, CurrentID=0, Status='live', Symbol='Moxd2', Aliases='Dbhl1', description='monooxygenase, DBH-like 2', other_designations='dopamine beta-hydroxylase-like 1|dopamine-beta-hydroxylase|dopamine-beta-hydroxylase-like protein', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40878794, end_position_on_the_genomic_accession=40887494, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  194388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194388, CurrentID=0, Status='live', Symbol='Tet3', Aliases='B430006D22Rik, BC037432, D230004J03Rik', description='tet methylcytosine dioxygenase 3', other_designations='Tet methylcytosine deoxygenase 3 isoform|probable methylcytosine dioxygenase TET3|tet oncogene family member 3', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83362373, end_position_on_the_genomic_accession=83457313, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  194401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194401, CurrentID=0, Status='live', Symbol='Mical3', Aliases='C130040D16Rik, Mical-3, mKIAA1364', description='microtubule associated monooxygenase, calponin and LIM domain containing 3', other_designations='flavoprotein oxidoreductase MICAL3|microtubule associated monoxygenase, calponin and LIM domain containing 3|molecule interacting with CasL protein 3|protein MICAL-3', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120931546, end_position_on_the_genomic_accession=121173757, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  194433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194433, CurrentID=0, Status='live', Symbol='Olfr707', Aliases='MOR260-8P', description='olfactory receptor 707', other_designations='B4 olfactory receptor|B6 olfactory receptor|GA_x6K02T2PBJ9-9276470-9275547|olfactory receptor MOR260-8P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106891184, end_position_on_the_genomic_accession=106892107, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  194590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194590, CurrentID=0, Status='live', Symbol='Reps2', Aliases='POB1', description='RALBP1 associated Eps domain containing protein 2', other_designations='partner of RalBP1|ralBP1-interacting protein 2', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=162411952, end_position_on_the_genomic_accession=162643705, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  194597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194597, CurrentID=0, Status='live', Symbol='Tmprss11a', Aliases='Gm7', description='transmembrane protease, serine 11a', other_designations='DESC-3|airway trypsin-like protease 1|serine protease DESC3', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86410410, end_position_on_the_genomic_accession=86469175, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  194655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194655, CurrentID=0, Status='live', Symbol='Klf11', Aliases='9830142A17, D12Ertd427e, Tieg2, Tieg2b, Tieg3', description='Kruppel-like factor 11', other_designations='TGFB inducible early growth response 2|TGFB inducible early growth response 3|TGFB-inducible early growth response protein 2b|TGFB-inducible early growth response protein 3|TIEG-3|klf11|transforming growth factor-beta-inducible early growth response protein 3', map_location='12 8.49 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24651371, end_position_on_the_genomic_accession=24662782, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76137, CurrentID=0, Status='live', Symbol='Mcur1', Aliases='6230416A05Rik, AU015498, AV136929, AW554392, C88263, Ccdc90a', description='mitochondrial calcium uniporter regulator 1', other_designations='coiled-coil domain containing 90A|coiled-coil domain-containing protein 90A, mitochondrial', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43538406, end_position_on_the_genomic_accession=43560259, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  207214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207214, CurrentID=0, Status='live', Symbol='Larp4', Aliases='DXErtd793, DXErtd793e', description='La ribonucleoprotein domain family, member 4', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99970065, end_position_on_the_genomic_accession=100016358, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  194735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194735, CurrentID=0, Status='live', Symbol='Btg1-ps2', Aliases='4930430D24Rik', description='B cell translocation gene 1, anti-proliferative, pseudogene 2', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38038199, end_position_on_the_genomic_accession=38043114, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  194738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194738, CurrentID=0, Status='live', Symbol='Rhox11', Aliases='BC049711, Gm39', description='reproductive homeobox 11', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38076598, end_position_on_the_genomic_accession=38085139, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  194744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194744, CurrentID=0, Status='live', Symbol='Slc25a43', Aliases='Gm8', description='solute carrier family 25, member 43', other_designations='mitochondrial solute carrier protein', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36743549, end_position_on_the_genomic_accession=36777307, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  408022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408022, CurrentID=0, Status='live', Symbol='Primpol', Aliases='BC065112, Ccdc111', description='primase and polymerase (DNA-directed)', other_designations='coiled-coil domain containing 111|coiled-coil domain-containing protein 111', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46575579, end_position_on_the_genomic_accession=46617234, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  194854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194854, CurrentID=0, Status='live', Symbol='Gm9', Aliases='', description='predicted gene 9', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37208502, end_position_on_the_genomic_accession=37211041, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  194856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194856, CurrentID=0, Status='live', Symbol='Rhox4e', Aliases='Rhox4, Rhox4.5, Rhox4d', description='reproductive homeobox 4E', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37557412, end_position_on_the_genomic_accession=37562282, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  63828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63828, CurrentID=0, Status='live', Symbol='Fn3k', Aliases='2310074G21Rik', description='fructosamine 3 kinase', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121434809, end_position_on_the_genomic_accession=121450491, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  63830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63830, CurrentID=0, Status='live', Symbol='Kcnq1ot1', Aliases='Kvlqt1-as, Lit1, Tssc8', description='KCNQ1 overlapping transcript 1', other_designations='KCNQ1 antisense RNA 2|antisense transcript of Kvlqt1|tumor-suppressing subchromosomal transferable fragment 8', map_location='7 88.15 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143213111, end_position_on_the_genomic_accession=143296547, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  63831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63831, CurrentID=0, Status='live', Symbol='DXMit20', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 20', other_designations='', map_location='X 75.66 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63832, CurrentID=0, Status='live', Symbol='D11Mit54', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 54', other_designations='', map_location='11 59.82 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  63833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63833, CurrentID=0, Status='live', Symbol='D15Mit15', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 15', other_designations='', map_location='15 58.06 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  194908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194908, CurrentID=0, Status='live', Symbol='Pld6', Aliases='4933433K01Rik, Gm10, Zuc, mZuc, mitoPLD', description='phospholipase D family, member 6', other_designations='choline phosphatase 6|mitochondrial phospholipase|novel Phospholipase D Active site motif-containing protein|phosphatidylcholine-hydrolyzing phospholipase D6|phospholipase D6|protein zucchini homolog', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59783893, end_position_on_the_genomic_accession=59787657, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  63856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63856, CurrentID=0, Status='live', Symbol='Taf8', Aliases='AW260255, Tbn', description='TATA-box binding protein associated factor 8', other_designations='TAF8 RNA polymerase II, TATA box binding protein (TBP)-associated factorq|TBP-associated factor 8|protein taube nuss|taube nuss', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47483414, end_position_on_the_genomic_accession=47502321, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  63857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63857, CurrentID=0, Status='live', Symbol='Bco1', Aliases='Bcdo, Bcdo1, Bcmo1, CMO1, Cmoi, beta-CD, betaCMOOX', description='beta-carotene oxygenase 1', other_designations="beta,beta-carotene 15,15'-monooxygenase|beta-carotene 15, 15'-dioxygenase 1|beta-carotene 15,15' dioxygenase|beta-carotene 15,15'-monooxygenase|beta-carotene cleavage enzyme|beta-carotene dioxygenase 1|carotenoid-15,15'-oxygenase", map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=117095851, end_position_on_the_genomic_accession=117133720, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  63859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63859, CurrentID=0, Status='live', Symbol='Impg1', Aliases='A930015H12Rik, IMP150, SPACR', description='interphotoreceptor matrix proteoglycan 1', other_designations='IPM-150|impg1|interphotoreceptor matrix proteoglycan of 150 kDa|sialoprotein associated with cones and rods', map_location='9 43.99 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=80308757, end_position_on_the_genomic_accession=80465654, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  63872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63872, CurrentID=0, Status='live', Symbol='Zfp296', Aliases='2210018A16Rik', description='zinc finger protein 296', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19577287, end_position_on_the_genomic_accession=19580656, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  63873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63873, CurrentID=0, Status='live', Symbol='Trpv4', Aliases='0610033B08Rik, OTRPC4, Trp12, VR-OAC, VRL-2, VROAC', description='transient receptor potential cation channel, subfamily V, member 4', other_designations='osm-9-like TRP channel 4|osmosensitive transient receptor potential channel 4|transient receptor potential protein 12|trpv4 channel|vanilloid receptor-like channel 2|vanilloid receptor-like protein 2|vanilloid receptor-related osmotically-activated channel', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114622152, end_position_on_the_genomic_accession=114658435, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  194952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194952, CurrentID=0, Status='live', Symbol='Jmjd4', Aliases='6430559I23', description='jumonji domain containing 4', other_designations='jumonji domain-containing protein 4', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59450045, end_position_on_the_genomic_accession=59458567, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  194974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=194974, CurrentID=0, Status='live', Symbol='Sun3', Aliases='Sunc1', description='Sad1 and UNC84 domain containing 3', other_designations='Sad1 and UNC84 domain containing 1|sad1/unc-84 domain-containing protein 1', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=9015798, end_position_on_the_genomic_accession=9048991, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  207259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207259, CurrentID=0, Status='live', Symbol='Zbtb7c', Aliases='B230208J24Rik, Kr-pok, Zbtb36', description='zinc finger and BTB domain containing 7C', other_designations='zinc finger and BTB domain containing 36|zinc finger and BTB domain-containing protein 36', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=75820178, end_position_on_the_genomic_accession=76148564, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  63913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63913, CurrentID=0, Status='live', Symbol='Fam129a', Aliases='AI256368, AU019833, Niban', description='family with sequence similarity 129, member A', other_designations='', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151571373, end_position_on_the_genomic_accession=151721939, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  195018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195018, CurrentID=0, Status='live', Symbol='Zzef1', Aliases='8430405D05Rik, BC051585, C130099L13Rik, C85424, NSPA, mKIAA0399', description='zinc finger, ZZ-type with EF hand domain 1', other_designations='KIAA0399-like protein human|neuronal superface P antigen', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72796226, end_position_on_the_genomic_accession=72927120, orientation='plus', exon_count=56, OMIM='', no_hdr0=''),
  63953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63953, CurrentID=0, Status='live', Symbol='Dusp10', Aliases='2610306G15Rik, AI158871, MKP-5, MKP5', description='dual specificity phosphatase 10', other_designations='MAP kinase phosphatase 5|mitogen-activated protein kinase phosphatase 5', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=184034461, end_position_on_the_genomic_accession=184075636, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  63954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63954, CurrentID=0, Status='live', Symbol='Rbp7', Aliases='1110002J23Rik, CRBP-III', description='retinol binding protein 7, cellular', other_designations='CRABP-IV|CRABP4|CRBP4|cellular retinoic acid-binding protein 4|cellular retinoic acid-binding protein IV|cellular retinol binding protein type III|cytosolic binding protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149449687, end_position_on_the_genomic_accession=149466462, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  63955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63955, CurrentID=0, Status='live', Symbol='Cables1', Aliases='Cables, ik3-1', description='CDK5 and Abl enzyme substrate 1', other_designations='cdk3-binding protein ik3-1|interactor with CDK3 1|interactor-1 with cdk3', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=11839274, end_position_on_the_genomic_accession=11945627, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  63958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63958, CurrentID=0, Status='live', Symbol='Ube4b', Aliases='4930551I19Rik, 4933406G05Rik, AU014668, D4Bwg0973e, UFD2, UFD2a, Ufd2p, mKIAA0684', description='ubiquitination factor E4B', other_designations='ubiquitin fusion degradation protein 2|ubiquitination factor E4B, UFD2 homolog', map_location='4 79.08 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149328416, end_position_on_the_genomic_accession=149426637, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  63959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63959, CurrentID=0, Status='live', Symbol='Slc29a1', Aliases='1200014D21Rik, AA407560, ENT1', description='solute carrier family 29 (nucleoside transporters), member 1', other_designations='NBMPR-sensitive equilibrative nucleoside transporter|equilibrative NBMPR-sensitive nucleoside transporter|equilibrative nitrobenzylmercaptopurine riboside-sensitive nucleoside transporter|equilibrative nucleoside transporter 1 variant delta 11|nucleoside transporter, es-type|solute carrier family 29 member 1', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45585200, end_position_on_the_genomic_accession=45599608, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  195040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195040, CurrentID=0, Status='live', Symbol='Tmem199', Aliases='AI316787, AI848469, R74819', description='transmembrane protein 199', other_designations='RE40914p', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78507055, end_position_on_the_genomic_accession=78512168, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  195046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195046, CurrentID=0, Status='live', Symbol='Nlrp1a', Aliases='CARD7, DEFCAP, Gm14, Gm15, NAC, Nalp1, Nalp1a, Nlrp1, PP1044', description='NLR family, pyrin domain containing 1A', other_designations='NACHT, LRR and PYD containing protein 1|NACHT, leucine rich repeat and PYD containing 1|NACHT/LRR/pyrin domain-containing protein 1|NLR family, pyrin domain containing 1|caspase recruitment domain protein|caspase recruitment domain-containing protein 7|death effector filament-forming Ced-4-like apoptosis protein|nucleotide-binding domain and caspase recruitment domain', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=71091197, end_position_on_the_genomic_accession=71144704, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  63985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63985, CurrentID=0, Status='live', Symbol='Gmfb', Aliases='3110001H22Rik, 3110001O16Rik, AI851627, C79176, D14Ertd630e', description='glia maturation factor, beta', other_designations='GMF-beta', map_location='14 24.32 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=46808149, end_position_on_the_genomic_accession=46822263, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  63986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63986, CurrentID=0, Status='live', Symbol='Gmfg', Aliases='0610039G16Rik, 2310057N07Rik, AI324845', description='glia maturation factor, gamma', other_designations='GMF-gamma', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28437447, end_position_on_the_genomic_accession=28448221, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  63993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=63993, CurrentID=0, Status='live', Symbol='Slc5a7', Aliases='CHT1', description='solute carrier family 5 (choline transporter), member 7', other_designations='CHT|hemicholinium-3-sensitive choline transporter|solute carrier family 5 member 7', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=54273590, end_position_on_the_genomic_accession=54299041, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  64008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64008, CurrentID=0, Status='live', Symbol='Aqp9', Aliases='1700020I22Rik, AI266899, AQP-9', description='aquaporin 9', other_designations='aquaglyceroporin-9', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=71110659, end_position_on_the_genomic_accession=71168085, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  64009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64009, CurrentID=0, Status='live', Symbol='Syne1', Aliases='8B, A330049M09Rik, BE692247, C130039F11Rik, CPG2, Myne1, mKIAA1756', description='spectrin repeat containing, nuclear envelope 1', other_designations='enaptin|enaptin165|myne-1|myocyte nuclear envelope protein 1|nesprin 1|nuclear envelope spectrin repeat protein 1|synaptic nuclear envelope 1|synaptic nuclear envelope protein 1', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=5020192, end_position_on_the_genomic_accession=5550829, orientation='minus', exon_count=152, OMIM='', no_hdr0=''),
  64010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64010, CurrentID=0, Status='live', Symbol='Sav1', Aliases='1700040G09Rik, Salv, Sav, WW45, Wwp3, Wwp4', description='salvador homolog 1 (Drosophila)', other_designations='45 kDa WW domain protein|WW domain-containing protein 3|WW domain-containing protein 4', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69965012, end_position_on_the_genomic_accession=69987002, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  64011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64011, CurrentID=0, Status='live', Symbol='Nrgn', Aliases='0710001B06Rik, AI838505, NG, NG/RC3, Pss1, R75334, RC3', description='neurogranin', other_designations='neurogranin (protein kinase C substrate, RC3)|postsynaptic spine 1|protein kinase C substrate', map_location='9 20.78 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37544493, end_position_on_the_genomic_accession=37552745, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76205, CurrentID=0, Status='live', Symbol='Stard3nl', Aliases='0610035N01Rik, 6530409L22Rik, AW124774, MENTHO', description='STARD3 N-terminal like', other_designations='STARD3 N-terminal-like protein', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19357676, end_position_on_the_genomic_accession=19395809, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76206, CurrentID=0, Status='live', Symbol='Gpr165', Aliases='6530406P05Rik', description='G protein-coupled receptor 165', other_designations='G-protein coupled receptor 165', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=96713468, end_position_on_the_genomic_accession=96719388, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  64050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64050, CurrentID=0, Status='live', Symbol='Yeats4', Aliases='4930573H17Rik, B230215M10Rik, GAS41, NuBI-1', description='YEATS domain containing 4', other_designations='glioma-amplified sequence-41', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117215141, end_position_on_the_genomic_accession=117224507, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  64051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64051, CurrentID=0, Status='live', Symbol='Sv2a', Aliases='AI746429, mKIAA0736', description='synaptic vesicle glycoprotein 2 a', other_designations='calcium regulator SV2A|synaptic vesicle protein 2|synaptic vesicle protein 2A', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96181227, end_position_on_the_genomic_accession=96195183, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  64058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64058, CurrentID=0, Status='live', Symbol='Perp', Aliases='1110017A08Rik, KCP1, KRTCAP1, PIGPC1, THW', description='PERP, TP53 apoptosis effector', other_designations='KCP-1|keratinocyte-associated protein 1|p53 apoptosis effector related to Pmp22|p53 apoptosis-associated target', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18845071, end_position_on_the_genomic_accession=18857073, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64059, CurrentID=0, Status='live', Symbol='Oxct2a', Aliases='Oxct, Oxct2, Scot, Scot-t1', description='3-oxoacid CoA transferase 2A', other_designations='3-oxoacid CoA-transferase 2A|3-oxoacid-CoA transferase 2A|haploid germ cell specific succinyl CoA transferase|succinyl-CoA:3-ketoacid-coenzyme A transferase 2A, mitochondrial|succinyl-CoA:3-oxoacid CoA-transferase|testis-specific succinyl CoA:3-oxoacid CoA-transferase 1|testis-specific succinyl-CoA:3-oxoacid CoA-transferase 1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123321875, end_position_on_the_genomic_accession=123323679, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  64074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64074, CurrentID=0, Status='live', Symbol='Smoc2', Aliases='1700056C05Rik, 5430426J21Rikl, Smoc2', description='SPARC related modular calcium binding 2', other_designations='SMOC-2|secreted modular calcium-binding protein 2', map_location='17 8.95 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14279487, end_position_on_the_genomic_accession=14404790, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  64075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64075, CurrentID=0, Status='live', Symbol='Smoc1', Aliases='2600002F22Rik, AI848508, SRG', description='SPARC related modular calcium binding 1', other_designations='SMOC-1|SPARC-related gene protein|SPARC-related protein|secreted modular calcium binding protein 1|secreted modular calcium-binding protein 1', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81026808, end_position_on_the_genomic_accession=81186414, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  64082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64082, CurrentID=0, Status='live', Symbol='Popdc2', Aliases='AV006127, Pop2', description='popeye domain containing 2', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38362164, end_position_on_the_genomic_accession=38378216, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  64085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64085, CurrentID=0, Status='live', Symbol='Clstn2', Aliases='2900042C18Rik, AI448973, CSTN2, Cs2, Cst-2, mKIAA4134', description='calsyntenin 2', other_designations='alc-gamma|alcadein-gamma', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=97444395, end_position_on_the_genomic_accession=98033384, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  64095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64095, CurrentID=0, Status='live', Symbol='Gpr35', Aliases='KPG_007', description='G protein-coupled receptor 35', other_designations='KYNA receptor|kynurenic acid receptor', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92950862, end_position_on_the_genomic_accession=92986391, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  338363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338363, CurrentID=0, Status='live', Symbol='Tmem241', Aliases='6030446N20Rik', description='transmembrane protein 241', other_designations='transmembrane protein C18orf45 homolog', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=11980136, end_position_on_the_genomic_accession=12121543, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  64103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64103, CurrentID=0, Status='live', Symbol='Tnmd', Aliases='1110017I01Rik, Bricd4, ChM1L, TeM', description='tenomodulin', other_designations='chondromodulin-1-like protein|chondromodulin-I-like protein|mChM1L|mTeM|myodulin|tendin', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=133851008, end_position_on_the_genomic_accession=133865577, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  195176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195176, CurrentID=0, Status='live', Symbol='Igh-VX24', Aliases='', description='immunoglobulin heavy chain (X24 family)', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  64113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64113, CurrentID=0, Status='live', Symbol='Moap1', Aliases='AA987038, Map-1, Pnma4', description='modulator of apoptosis 1', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102739830, end_position_on_the_genomic_accession=102743661, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  338367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338367, CurrentID=0, Status='live', Symbol='Myo1d', Aliases='9930104H07Rik, AW544947, D11Ertd9e, myosin-1d', description='myosin ID', other_designations='myo1d|myosin-Id', map_location='11 47.95 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80482126, end_position_on_the_genomic_accession=80780055, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  64136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64136, CurrentID=0, Status='live', Symbol='Sdf2l1', Aliases='', description='stromal cell-derived factor 2-like 1', other_designations='SDF2-like protein 1', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17130138, end_position_on_the_genomic_accession=17132383, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  195209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195209, CurrentID=0, Status='live', Symbol='Gm22', Aliases='', description='predicted gene 22', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122043125, end_position_on_the_genomic_accession=122272650, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  64138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64138, CurrentID=0, Status='live', Symbol='Ctsz', Aliases='AI787083, AU019819, CTSX, D2Wsu143e', description='cathepsin Z', other_designations='cathepsin X', map_location='2 97.94 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174427494, end_position_on_the_genomic_accession=174438992, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  64139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64139, CurrentID=0, Status='live', Symbol='Ctsm', Aliases='1600027J17Rik, AI324100, Catm', description='cathepsin M', other_designations='Cat M', map_location='13 32.74 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61535743, end_position_on_the_genomic_accession=61541839, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  64143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64143, CurrentID=0, Status='live', Symbol='Ralb', Aliases='5730472O18Rik', description='v-ral simian leukemia viral oncogene homolog B (ras related)', other_designations='', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=119470305, end_position_on_the_genomic_accession=119504782, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  64144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64144, CurrentID=0, Status='live', Symbol='Mllt1', Aliases='AA407901, BAM11, ENL, LTG19', description='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 1', other_designations='leukemia-associated protein MLLT1|myeloid/lymphoid or mixed lineage-leukemia translocation to 1 homolog|myeloid/lymphoid or mixed lineage-leukemia translocation to 4 homolog|trithorax', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56892608, end_position_on_the_genomic_accession=56935484, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  195236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195236, CurrentID=0, Status='live', Symbol='Pom121l2', Aliases='Gm24', description='POM121 membrane glycoprotein-like 2 (rat)', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21981181, end_position_on_the_genomic_accession=21988734, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  64176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64176, CurrentID=0, Status='live', Symbol='Sv2b', Aliases='A830038F04Rik, AI850768, mKIAA0735', description='synaptic vesicle glycoprotein 2 b', other_designations='synaptic vesicle protein 2B', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=75114890, end_position_on_the_genomic_accession=75309262, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  64177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64177, CurrentID=0, Status='live', Symbol='Trpv6', Aliases='CAT, CaT1, Cac, Ecac2, Otrpc3', description='transient receptor potential cation channel, subfamily V, member 6', other_designations='calcium transport protein 1|epithelial calcium channel 2|osmosensitive transient receptor potential channel 3', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41620619, end_position_on_the_genomic_accession=41636480, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  64209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64209, CurrentID=0, Status='live', Symbol='Herpud1', Aliases='Herp, Mifl, SUP', description='homocysteine-inducible, endoplasmic reticulum stress-inducible, ubiquitin-like domain member 1', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94386143, end_position_on_the_genomic_accession=94395366, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  64213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64213, CurrentID=0, Status='live', Symbol='St7', Aliases='9430001H04Rik, Fam4a2, HELG, RAY1, SEN4, TSG7', description='suppression of tumorigenicity 7', other_designations='4low density lipoprotein-related protein 12|family with sequence similarity 4, subfamily A, member 2', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=17692940, end_position_on_the_genomic_accession=17943023, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  64214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64214, CurrentID=0, Status='live', Symbol='Rgs18', Aliases='', description='regulator of G-protein signaling 18', other_designations='', map_location='1 62.99 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=144752692, end_position_on_the_genomic_accession=144775421, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  64242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64242, CurrentID=0, Status='live', Symbol='Ngb', Aliases='', description='neuroglobin', other_designations='', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87097531, end_position_on_the_genomic_accession=87102539, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64243, CurrentID=0, Status='live', Symbol='Snord116', Aliases='MBII-85, Pwcr1', description='small nucleolar RNA, C/D box 116', other_designations='Prader-Willi chromosome region 1 homolog|small nucleolar RNA, C/D box 116 cluster|snoRNA MBII-85', map_location='7 29.0 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=59675990, end_position_on_the_genomic_accession=59676079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  195333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195333, CurrentID=0, Status='live', Symbol='Gsc2', Aliases='4930568H22Rik, GSC-2, Gscl', description='goosecoid homebox 2', other_designations='GSC-L', map_location='16 11.1 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17913554, end_position_on_the_genomic_accession=17915059, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  195359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195359, CurrentID=0, Status='live', Symbol='Trim40', Aliases='Gm319, Gm35, Gm865', description='tripartite motif-containing 40', other_designations='probable E3 NEDD8-protein ligase|ring finger RNF35', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36881598, end_position_on_the_genomic_accession=36896743, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64290, CurrentID=0, Status='live', Symbol='Foxb1', Aliases='C43, Fkh5a, Foxb1b, Hfh-e5.1, Mf3, Twh, Foxb1', description='forkhead box B1', other_designations='forkhead box B1a|forkhead box B1b|mesoderm/mesenchyme forkhead 3|transcription factor FKH-5', map_location='9 38.76 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=69757710, end_position_on_the_genomic_accession=69760940, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  64291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64291, CurrentID=0, Status='live', Symbol='Osbpl1a', Aliases='G430090F17Rik, Gm753, ORP-1, Osbpl1b', description='oxysterol binding protein-like 1A', other_designations='OSBP-related protein 1', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12755312, end_position_on_the_genomic_accession=12941875, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  64292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64292, CurrentID=0, Status='live', Symbol='Ptges', Aliases='2410099E23Rik, D2Ertd369e, Pges, mPGES, mPGES-1', description='prostaglandin E synthase', other_designations='microsomal prostaglandin E synthase 1', map_location='2 21.75 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30889471, end_position_on_the_genomic_accession=30903297, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64293, CurrentID=0, Status='live', Symbol='Stk32b', Aliases='2510009F08Rik, STKG6, Stk32, YANK2', description='serine/threonine kinase 32B', other_designations='serine threonine kinase 32|stk32b', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37446825, end_position_on_the_genomic_accession=37717153, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  64294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64294, CurrentID=0, Status='live', Symbol='Itm2c', Aliases='3110038L02Rik, BRI3, Bricd2c, E25, E25C, ITM3', description='integral membrane protein 2C', other_designations='integral membrane protein 3|transmembrane protein BRI3', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85894510, end_position_on_the_genomic_accession=85908698, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64295, CurrentID=0, Status='live', Symbol='Tmub1', Aliases='2010004O20Rik, AB030183, Hops', description='transmembrane and ubiquitin-like domain containing 1', other_designations='hepatocyte odd protein shuttling protein', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24445463, end_position_on_the_genomic_accession=24448331, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64296, CurrentID=0, Status='live', Symbol='Abhd8', Aliases='0910001L24Rik, AB030191', description='abhydrolase domain containing 8', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71456700, end_position_on_the_genomic_accession=71463657, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  64297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64297, CurrentID=0, Status='live', Symbol='Gprc5b', Aliases='AW125761, Raig2', description='G protein-coupled receptor, family C, group 5, member B', other_designations='RAIG-2|hypothetical protein, clone 2-63|retinoic acid-induced gene 2 protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118972040, end_position_on_the_genomic_accession=118995211, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64335, CurrentID=0, Status='live', Symbol='Svs3a', Aliases='9530026M05Rik, BB121242, Semg2, Svp-3, Svp3, Svs3', description='seminal vesicle secretory protein 3A', other_designations='SVS III|semenogelin II|seminal vesicle protein 3|seminal vesicle secretion 3|seminal vesicle secretion 3A', map_location='2 85.07 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164289268, end_position_on_the_genomic_accession=164291500, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64337, CurrentID=0, Status='live', Symbol='Gng13', Aliases='1500031D04Rik, AB030194, Ggamma13', description='guanine nucleotide binding protein (G protein), gamma 13', other_designations='G gamma subunit, clone 2-35|guanine nucleotide binding protein 13, gamma', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25717172, end_position_on_the_genomic_accession=25719102, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  64339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64339, CurrentID=0, Status='live', Symbol='Fndc4', Aliases='2810430J06Rik, 6330410H20Rik, AB030187, AI838506, AW487863, FRCP1, Fnmp1', description='fibronectin type III domain containing 4', other_designations='fibronectin type III repeat containing protein 1|fibronectin type III repeat-containing protein 1', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31292245, end_position_on_the_genomic_accession=31295877, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64340, CurrentID=0, Status='live', Symbol='Dhx38', Aliases='5730550P09Rik, AI325984, AW540902, Ddx38, Prp16, mKIAA0224', description='DEAH (Asp-Glu-Ala-His) box polypeptide 38', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 38|pre-mRNA splicing factor 16', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109548003, end_position_on_the_genomic_accession=109565856, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  76261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76261, CurrentID=0, Status='live', Symbol='0610040J01Rik', Aliases='AI662686', description='RIKEN cDNA 0610040J01 gene', other_designations='', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=63811969, end_position_on_the_genomic_accession=63899626, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  329482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329482, CurrentID=0, Status='live', Symbol='Dcdc5', Aliases='4732421G10Rik, BC048594, EG436559', description='doublecortin domain containing 5', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=106042135, end_position_on_the_genomic_accession=106408603, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  64378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64378, CurrentID=0, Status='live', Symbol='Gpr88', Aliases='AW061286, Strg', description='G-protein coupled receptor 88', other_designations='striatum-specific G-protein coupled receptor', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116249654, end_position_on_the_genomic_accession=116253498, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64379, CurrentID=0, Status='live', Symbol='Irx6', Aliases='', description='Iroquois related homeobox 6 (Drosophila)', other_designations='homeodomain protein IRXB3|iroquois homeobox protein 6', map_location='8 44.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=92674289, end_position_on_the_genomic_accession=92680956, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64380, CurrentID=0, Status='live', Symbol='Ms4a4c', Aliases='5830413L19Rik, AI449185, Ms4a9', description='membrane-spanning 4-domains, subfamily A, member 4C', other_designations='membrane-spanning 4-domains, subfamily A, member 9', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11404717, end_position_on_the_genomic_accession=11427246, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  64381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64381, CurrentID=0, Status='live', Symbol='Ms4a8a', Aliases='2010004L09Rik, AI481240, CD20L5, Ms4a8', description='membrane-spanning 4-domains, subfamily A, member 8A', other_designations='CD20 antigen-like 5|MS4a8B protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11067471, end_position_on_the_genomic_accession=11081103, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  64383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64383, CurrentID=0, Status='live', Symbol='Sirt2', Aliases='5730427M03Rik, SIR2L2, Sir2l', description='sirtuin 2', other_designations='NAD-dependent deacetylase sirtuin-2|SIR2-like protein 2|mSIR2L2|regulatory protein SIR2 homolog 2|silent mating type information regulation 2, (S.cerevisiae, homolog)-like; sirtuin 2|sirtuin-2', map_location='7 B1-2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28766752, end_position_on_the_genomic_accession=28788665, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  64384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64384, CurrentID=0, Status='live', Symbol='Sirt3', Aliases='2310003L23Rik, AI848213, Sir2l3', description='sirtuin 3', other_designations='NAD-dependent deacetylase sirtuin-3|SIR2-like protein 3|SIRT3L mitochondrial|mSIR2L3|mitochondrial protein lysine deacetylase|regulatory protein SIR2 homolog 3|silent mating type information regulation 2, (S.cerevisiae, homolog)-like 3', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140863663, end_position_on_the_genomic_accession=140882309, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64385, CurrentID=0, Status='live', Symbol='Cyp4f14', Aliases='1300014O15Rik, cyp4f-14', description='cytochrome P450, family 4, subfamily f, polypeptide 14', other_designations='cytochrome P450 4F14|cytochrome P450, subfamily IVF, polypeptide 14 (leukotriene B4 omega hydroxylase)|cytochrome P450-LTB-omega|leukotriene-B4 20-monooxygenase 3', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32905070, end_position_on_the_genomic_accession=32917348, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  64406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64406, CurrentID=0, Status='live', Symbol='Sp5', Aliases='', description='trans-acting transcription factor 5', other_designations='', map_location='2 41.49 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=70474923, end_position_on_the_genomic_accession=70477729, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338417, CurrentID=0, Status='live', Symbol='Scgb1c1', Aliases='Ryd5', description='secretoglobin, family 1C, member 1', other_designations='ligand binding protein RYD5|secretoglobin RYD5', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140845565, end_position_on_the_genomic_accession=140846769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64424, CurrentID=0, Status='live', Symbol='Polr1e', Aliases='53kDa, AU042259, D030019D19Rik, Paf53, Praf1', description='polymerase (RNA) I polypeptide E', other_designations='DNA-directed RNA polymerase I subunit E|RNA polymerase I associated factor (Paf53)|RNA polymerase I subunit A49|RNA polymerase I-associated factor 53|polymerase (RNA) I associated factor 1', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45018609, end_position_on_the_genomic_accession=45034279, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  64435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64435, CurrentID=0, Status='live', Symbol='Fcamr', Aliases='', description='Fc receptor, IgA, IgM, high affinity', other_designations='Fc IgA/IgM receptor|Fca/m receptor|fc alpha/mu receptor|mFcamR', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130785260, end_position_on_the_genomic_accession=130819316, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  64436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64436, CurrentID=0, Status='live', Symbol='Inpp5e', Aliases='1200002L24Rik, 72kDa, mKIAA0123', description='inositol polyphosphate-5-phosphatase E', other_designations='inositol polyphosphate-5-phosphatase, 72 kDa|phosphatidylinositol 4,5-bisphosphate 5-phosphatase|phosphatidylinositol polyphosphate 5-phosphatase type IV|phosphatidylinositol-4,5-bisphosphate 5-phosphatase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26396249, end_position_on_the_genomic_accession=26409219, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  195522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195522, CurrentID=0, Status='live', Symbol='Zfp691', Aliases='6430544H17Rik, Znf691', description='zinc finger protein 691', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119169516, end_position_on_the_genomic_accession=119174195, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  64451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64451, CurrentID=0, Status='live', Symbol='Dip2a', Aliases='4931420H10Rik, AI426328, Dip2, Kiaa0184-hp, mKIAA0184', description='disco interacting protein 2 homolog A', other_designations='DIP2 disco-interacting protein 2 homolog A|DIP2 homolog A|disco interacting protein 2 homolog', map_location='10 38.76 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76263049, end_position_on_the_genomic_accession=76345301, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  64452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64452, CurrentID=0, Status='live', Symbol='Slc5a4a', Aliases='SGLT3-a', description='solute carrier family 5, member 4a', other_designations='Na(+)/glucose cotransporter 3|sodium-dependent glucose cotransporter|sodium/glucose cotransporter 3|solute carrier family 5 member 4|solute carrier family 5, member alpha 4a', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76147451, end_position_on_the_genomic_accession=76189265, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  64453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64453, CurrentID=0, Status='live', Symbol='Zfp280b', Aliases='D10Jhu82e, Suhw2', description='zinc finger protein 280B', other_designations='suppressor of hairy wing homolog 2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76032401, end_position_on_the_genomic_accession=76042970, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64454, CurrentID=0, Status='live', Symbol='Slc5a4b', Aliases='2010104G07Rik, SAAT1, SGLT3-b, SGLT3b, pSGLT2', description='solute carrier family 5 (neutral amino acid transporters, system A), member 4b', other_designations='low affinity sodium-glucose cotransporter', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76058618, end_position_on_the_genomic_accession=76111018, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  195531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195531, CurrentID=0, Status='live', Symbol='Gm13152', Aliases='OTTMUSG00000010671', description='predicted gene 13152', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147500686, end_position_on_the_genomic_accession=147513420, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  326618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=326618, CurrentID=0, Status='live', Symbol='Tpm4', Aliases='2610528G24Rik.2, Tpm4.2cy, Tpm4', description='tropomyosin 4', other_designations='tropomyosin-4', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72135292, end_position_on_the_genomic_accession=72153129, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  326619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=326619, CurrentID=0, Status='live', Symbol='Hist1h4a', Aliases='', description='histone cluster 1, H4a', other_designations='histone 1, H4a', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23760795, end_position_on_the_genomic_accession=23761249, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  326620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=326620, CurrentID=0, Status='live', Symbol='Hist1h4b', Aliases='', description='histone cluster 1, H4b', other_designations='histone 1, H4b', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23756937, end_position_on_the_genomic_accession=23757386, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  326622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=326622, CurrentID=0, Status='live', Symbol='Upf2', Aliases='', description='UPF2 regulator of nonsense transcripts homolog (yeast)', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5951447, end_position_on_the_genomic_accession=6056703, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  326623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=326623, CurrentID=0, Status='live', Symbol='Tnfsf15', Aliases='Tl1, Tl1a, Tnlg1b, Vegi', description='tumor necrosis factor (ligand) superfamily, member 15', other_designations='TNF ligand-related molecule 1|bM20K13.3 (tumor necrosis factor (ligand) superfamily, member 15)|tumor necrosis factor ligand 1b|vascular endothelial cell growth inhibitor|vascular endothelial growth inhibitor', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63724603, end_position_on_the_genomic_accession=63745113, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64540, CurrentID=0, Status='live', Symbol='Tspan4', Aliases='AI325509, AI746565, D130042I01Rik, NAG-2, Tm4sf7, Tspan-4', description='tetraspanin 4', other_designations='novel antigen 2|tetraspan TM4SF|transmembrane 4 superfamily member 7', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141475236, end_position_on_the_genomic_accession=141493427, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76294, CurrentID=0, Status='live', Symbol='Asb5', Aliases='1110018D09Rik', description='ankyrin repeat and SOCs box-containing 5', other_designations='ASB-5|ankyrin repeat and SOCs box-containing protein 5', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=54550331, end_position_on_the_genomic_accession=54587842, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  195646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195646, CurrentID=0, Status='live', Symbol='Hs3st2', Aliases='6430516N12Rik, A830061E14Rik, AW491345', description='heparan sulfate (glucosamine) 3-O-sulfotransferase 2', other_designations='heparan sulfate 3-O-sulfotransferase 2|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 2', map_location='7 65.07 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121391610, end_position_on_the_genomic_accession=121501768, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  64580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64580, CurrentID=0, Status='live', Symbol='Ndst4', Aliases='4930439H17Rik', description='N-deacetylase/N-sulfotransferase (heparin glucosaminyl) 4', other_designations='N-HSST 4|N-deacetylase/N-sulfotransferase 4|N-heparan sulfate sulfotransferase 4|NDST-4|glucosaminyl N-deacetylase/N-sulfotransferase 4', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=125404076, end_position_on_the_genomic_accession=125728351, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  64602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64602, CurrentID=0, Status='live', Symbol='Ireb2', Aliases='D9Ertd85e, Irp2', description='iron responsive element binding protein 2', other_designations='IRE-BP 2|iron regulatory protein 2', map_location='9 29.81 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54863667, end_position_on_the_genomic_accession=54912534, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  207375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207375, CurrentID=0, Status='live', Symbol='Fam120c', Aliases='D930001I21Rik, ORF34', description='family with sequence similarity 120, member C', other_designations='family with sequence similarity 120C', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151342932, end_position_on_the_genomic_accession=151474136, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  64652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64652, CurrentID=0, Status='live', Symbol='Nisch', Aliases='1200007D05Rik, 3202002H23Rik, AW494485, I-1', description='nischarin', other_designations='I1R|IR1|imidazoline receptor 1|imidazoline receptor I-1-like protein|imidazoline-1 receptor', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31170928, end_position_on_the_genomic_accession=31206826, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  64654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64654, CurrentID=0, Status='live', Symbol='Fgf23', Aliases='', description='fibroblast growth factor 23', other_designations='FGF-23', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127072902, end_position_on_the_genomic_accession=127082296, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  64655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64655, CurrentID=0, Status='live', Symbol='Mrps22', Aliases='3100002P07Rik, Rpms22', description='mitochondrial ribosomal protein S22', other_designations='MRP-S22|S22mt|ribosomal protein, mitochondrial, S22', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98588730, end_position_on_the_genomic_accession=98601679, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64656, CurrentID=0, Status='live', Symbol='Mrps23', Aliases='D11Bwg1153e, MRP-S23, Rpms23, S23mt', description='mitochondrial ribosomal protein S23', other_designations='mrps23', map_location='11 52.4 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88194106, end_position_on_the_genomic_accession=88211507, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64657, CurrentID=0, Status='live', Symbol='Mrps10', Aliases='1110038B19Rik, AW743084, Rpms10', description='mitochondrial ribosomal protein S10', other_designations='MRP-S10|S10mt', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47368887, end_position_on_the_genomic_accession=47381418, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  64658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64658, CurrentID=0, Status='live', Symbol='Mrps25', Aliases='2810429N01Rik, Rpms25', description='mitochondrial ribosomal protein S25', other_designations='MRP-S25|S25mt', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=92169512, end_position_on_the_genomic_accession=92184023, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64659, CurrentID=0, Status='live', Symbol='Mrps14', Aliases='1810032L21Rik, AL022847, Rpms14', description='mitochondrial ribosomal protein S14', other_designations='MRP-S14|S14mt', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160195260, end_position_on_the_genomic_accession=160201186, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  64660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64660, CurrentID=0, Status='live', Symbol='Mrps24', Aliases='3110030K20Rik, AI414579, Rpms24', description='mitochondrial ribosomal protein S24', other_designations='MRP-S24|S24mt|ribosomal protein, mitochondrial, S24', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5703983, end_position_on_the_genomic_accession=5707717, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  64661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64661, CurrentID=0, Status='live', Symbol='Krtdap', Aliases='Kdap, sk221', description='keratinocyte differentiation associated protein', other_designations='krtdap', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30787905, end_position_on_the_genomic_accession=30791083, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64685, CurrentID=0, Status='live', Symbol='Nmi', Aliases='', description='N-myc (and STAT) interactor', other_designations='N-mcy (and STAT) interactor', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=51948498, end_position_on_the_genomic_accession=51973520, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  64697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64697, CurrentID=0, Status='live', Symbol='Keg1', Aliases='0610008P16Rik, AI316519, GS4059', description='kidney expressed gene 1', other_designations='Acyl-CoA:glycine N-acyltransferase protein Keg1|kidney-expressed gene 1 protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12695790, end_position_on_the_genomic_accession=12719902, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  64704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64704, CurrentID=0, Status='live', Symbol='Htra2', Aliases='AI481710, Omi, Prss25, mnd2', description='HtrA serine peptidase 2', other_designations='high temperature requirement protein A2|motor neuron degeneration 2|omi stress-regulated endoprotease|protease, serine, 25|serine protease 25|serine protease OMI|serine proteinase OMI', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83051266, end_position_on_the_genomic_accession=83054571, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64705, CurrentID=0, Status='live', Symbol='Dpys', Aliases='1200017I10Rik, 1300004I01Rik, DHP, DHPase', description='dihydropyrimidinase', other_designations='dihydropyrimidine amidohydrolase|hydantoinase', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39768485, end_position_on_the_genomic_accession=39857470, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  64706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64706, CurrentID=0, Status='live', Symbol='Scube1', Aliases='7330410C13Rik, A630023E24Rik', description='signal peptide, CUB domain, EGF-like 1', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83602583, end_position_on_the_genomic_accession=83725049, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  64707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64707, CurrentID=0, Status='live', Symbol='Suv39h2', Aliases='4930507K23Rik, AA536750, D030054H19Rik, D2Ertd544e, KMT1B', description='suppressor of variegation 3-9 homolog 2 (Drosophila)', other_designations='H3-K9-HMTase 2|Suv39h histone methyltransferase|histone H3-K9 methyltransferase 2|su(var)3-9 homolog 2', map_location='2 1.95 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3455815, end_position_on_the_genomic_accession=3474986, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  64817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64817, CurrentID=0, Status='live', Symbol='Svep1', Aliases='1110021D17Rik, 4833413O10Rik, AA387071, D430029O09Rik, Polydom', description='sushi, von Willebrand factor type A, EGF and pentraxin domain containing 1', other_designations='polydomain protein (Polydom)', map_location='4 B-C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58042796, end_position_on_the_genomic_accession=58206596, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  64818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64818, CurrentID=0, Status='live', Symbol='Krt81', Aliases='AA589625, Krt2-19', description='keratin 81', other_designations='K81|keratin complex 2, basic, gene 19|krt81|type II hair keratin Hb1', map_location='15 56.9 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101459061, end_position_on_the_genomic_accession=101463765, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  64821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64821, CurrentID=0, Status='live', Symbol='Cigs1', Aliases='CNS-1, Cns1, HSS1, HSS2', description='conserved intergenic sequence 1', other_designations='conserved noncoding sequence 1', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  64833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64833, CurrentID=0, Status='live', Symbol='Acot10', Aliases='Acate3, MT-ACT48, p48', description='acyl-CoA thioesterase 10', other_designations='acot10|acyl-Coenzyme A thioesterase 3, mitochondrial|mitochondrial 48 kDa acyl-CoA thioester hydrolase 2|mt-ACT48.2', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=20665214, end_position_on_the_genomic_accession=20666750, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  64898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64898, CurrentID=0, Status='live', Symbol='Lpin2', Aliases='2610511G02Rik, AI481352, AW742896', description='lipin 2', other_designations='lipin-2', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71183528, end_position_on_the_genomic_accession=71249818, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  64899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64899, CurrentID=0, Status='live', Symbol='Lpin3', Aliases='9130206L11Rik, AA438110, AV236139, mKIAA4023', description='lipin 3', other_designations='lipin-3', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=160880670, end_position_on_the_genomic_accession=160906000, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  64918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64918, CurrentID=0, Status='live', Symbol='Bhmt2', Aliases='C81077, D13Ucla2', description='betaine-homocysteine methyltransferase 2', other_designations='SMM-hcy methyltransferase|betaine--homocysteine S-methyltransferase 2', map_location='13 47.88 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93656097, end_position_on_the_genomic_accession=93674302, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  64929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64929, CurrentID=0, Status='live', Symbol='Scel', Aliases='9230114I02Rik', description='sciellin', other_designations='', map_location='14 E', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=103513060, end_position_on_the_genomic_accession=103613346, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  64930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64930, CurrentID=0, Status='live', Symbol='Tsc1', Aliases='', description='tuberous sclerosis 1', other_designations='tuberous sclerosis 1 protein homolog', map_location='2 B-C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28641190, end_position_on_the_genomic_accession=28691172, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  64931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64931, CurrentID=0, Status='live', Symbol='Izumo1r', Aliases='0910001L11Rik, C86255, Folbp3, Folr4, Juno', description='IZUMO1 receptor, JUNO', other_designations='FR-delta|IZUMO1 receptor protein JUNO|folate receptor 4 (delta)|folate receptor delta|folate-binding protein 3|probable folate receptor delta', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14885814, end_position_on_the_genomic_accession=14903951, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  64933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64933, CurrentID=0, Status='live', Symbol='Ap3m2', Aliases='5830445E16Rik, AP-3B', description='adaptor-related protein complex 3, mu 2 subunit', other_designations='HA1 47 kDa subunit homolog 2|P47B|adapter-related protein complex 3 mu-2 subunit|adapter-related protein complex 3 subunit mu-2|adaptor-related protein complex 3 subunit mu-2|clathrin assembly protein assembly protein complex 1 medium chain homolog 2|clathrin assembly protein assembly protein complex 3 mu-2 medium chain|clathrin coat assembly protein AP47 homolog 2|clathrin coat-associated protein AP47 homolog 2|golgi adaptor AP-1 47 kDa protein homolog 2|m3B|mu3B-adaptin', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22787354, end_position_on_the_genomic_accession=22805654, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  64934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64934, CurrentID=0, Status='live', Symbol='Pes1', Aliases='', description='pescadillo homolog 1, containing BRCT domain (zebrafish)', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3963975, end_position_on_the_genomic_accession=3980004, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  64938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64938, CurrentID=0, Status='live', Symbol='Cigs2', Aliases='CNS-7, Cns7', description='conserved intergenic sequence 2', other_designations='GM-CSF enhancer region|conserved noncoding sequence 7', map_location='11', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  64945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64945, CurrentID=0, Status='live', Symbol='Cldn12', Aliases='', description='claudin 12', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=5505015, end_position_on_the_genomic_accession=5514976, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  65019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65019, CurrentID=0, Status='live', Symbol='Rpl23', Aliases='2810009A01Rik', description='ribosomal protein L23', other_designations='', map_location='11 61.1 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97777526, end_position_on_the_genomic_accession=97782439, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  65020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65020, CurrentID=0, Status='live', Symbol='Zfp110', Aliases='2900024E01Rik, N28112, NRIF, Nrif1', description='zinc finger protein 110', other_designations='neurotrophin receptor interacting factor', map_location='7 7.66 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12834810, end_position_on_the_genomic_accession=12850584, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  65079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65079, CurrentID=0, Status='live', Symbol='Rtn4r', Aliases='NOGOR, NgR, NgR1', description='reticulon 4 receptor', other_designations='Nogo-66 receptor|nogo receptor|nogo receptor; Nogo-66 receptor', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18127706, end_position_on_the_genomic_accession=18152408, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  65086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65086, CurrentID=0, Status='live', Symbol='Lpar3', Aliases='Edg7, lpA3', description='lysophosphatidic acid receptor 3', other_designations='LPA receptor 3|LPA-3|endothelial differentiation, lysophosphatidic acid G-protein-coupled receptor 7|lysophosphatidic acid receptor Edg-7', map_location='3 71.03 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146220961, end_position_on_the_genomic_accession=146286214, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  65098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65098, CurrentID=0, Status='live', Symbol='Zfand6', Aliases='3110005P07Rik, Awp1, Za20d3', description='zinc finger, AN1-type domain 6', other_designations='associated with PRK1 protein|associated with Prkcl1|zinc finger A20 domain-containing protein 3|zinc finger, A20 domain containing 3', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=84615045, end_position_on_the_genomic_accession=84689952, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  65099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65099, CurrentID=0, Status='live', Symbol='Irak1bp1', Aliases='4921528N06Rik, AI851240, Aabp3, Aip70, Simpl', description='interleukin-1 receptor-associated kinase 1 binding protein 1', other_designations='Acta binding protein 3|Acta-binding protein 70|IRAK1-binding protein 1|Plk-interacting protein|signaling molecule that associates with the mouse pelle-like kinase', map_location='9 E2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=82829806, end_position_on_the_genomic_accession=82847688, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  65100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65100, CurrentID=0, Status='live', Symbol='Zic5', Aliases='1700049L20Rik, Opr', description='zinc finger protein of the cerebellum 5', other_designations='Zic family member 5 (odd-paired homolog, Drosophila)|odd-paired related', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122459160, end_position_on_the_genomic_accession=122465716, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  65102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65102, CurrentID=0, Status='live', Symbol='Nif3l1', Aliases='1110030G24Rik', description='Ngg1 interacting factor 3-like 1 (S. pombe)', other_designations='GTP cyclohydrolase I|Ngg1 interacting factor 3 homolog-like 1|putative GTP cyclohydrolase 1 type 2 Nif3l1', map_location='1 C', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58445156, end_position_on_the_genomic_accession=58462276, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  65103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65103, CurrentID=0, Status='live', Symbol='Arl6ip6', Aliases='2310057C01Rik, 2610529A11Rik, Aip-6', description='ADP-ribosylation factor-like 6 interacting protein 6', other_designations='ADP-ribosylation-like factor 6 interacting protein 6|ARL-6-interacting protein 6', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=53192084, end_position_on_the_genomic_accession=53219221, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  65105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65105, CurrentID=0, Status='live', Symbol='Arl6ip4', Aliases='AA408210, AA408365, Aip-4, Srp25', description='ADP-ribosylation factor-like 6 interacting protein 4', other_designations='ADP-ribosylation-like factor 6 interacting protein 4|ARL-6-interacting protein 4|SRp25 nuclear protein|splicing factor SRrp37', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124116108, end_position_on_the_genomic_accession=124118195, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  65106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65106, CurrentID=0, Status='live', Symbol='Arl6ip5', Aliases='5930404D22Rik, AV001879, Aip-5, Gtrap3-18, addiscin', description='ADP-ribosylation factor-like 6 interacting protein 5', other_designations='ADP-ribosylation factor-like protein 6-interacting protein 5|ADP-ribosylation-like factor 6 interacting protein 5|ARL-6-interacting protein 5|addicsin|glutamate transporter EAAC1-interacting protein|prenylated Rab acceptor protein 2|protein JWa', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=97210792, end_position_on_the_genomic_accession=97233315, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  65107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65107, CurrentID=0, Status='live', Symbol='Lrp10', Aliases='Lrp9', description='low-density lipoprotein receptor-related protein 10', other_designations='LRP-10', map_location='14 C2-C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54464147, end_position_on_the_genomic_accession=54470292, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  65111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65111, CurrentID=0, Status='live', Symbol='Dap3', Aliases='4921514D13Rik, DAP-3, MRP-S29, S29mt', description='death associated protein 3', other_designations='death-associated protein 3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88920803, end_position_on_the_genomic_accession=88951266, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  65112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65112, CurrentID=0, Status='live', Symbol='Pmepa1', Aliases='2210418I02Rik, AW455466, Erg1.2, N4wbp4, Stag1, Tmepai', description='prostate transmembrane protein, androgen induced 1', other_designations='NEDD4 WW domain-binding protein 4|Nedd4 WW binding protein 4|prostate transmembrane protein androgen induced 1|transmembrane prostate androgen-induced protein|transmembrane, prostate androgen induced RNA', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173224458, end_position_on_the_genomic_accession=173276533, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  65113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65113, CurrentID=0, Status='live', Symbol='Ndfip1', Aliases='0610010M22Rik, N4wbp5', description='Nedd4 family interacting protein 1', other_designations='NEDD4 WW domain-binding protein 5|Nedd4 WW binding protein 5|Nedd4 WW-binding protein 5', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38418912, end_position_on_the_genomic_accession=38465190, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  65114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65114, CurrentID=0, Status='live', Symbol='Vps35', Aliases='AI647796, Mem3', description='vacuolar protein sorting 35', other_designations='maternal embryonic message 3|maternal-embryonic 3|vesicle protein sorting 35', map_location='8 41.61 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85260392, end_position_on_the_genomic_accession=85299497, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  65115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65115, CurrentID=0, Status='live', Symbol='Bean1', Aliases='Bean', description='brain expressed, associated with Nedd4, 1', other_designations='brain-expressed protein associating with Nedd4|protein BEAN', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104170376, end_position_on_the_genomic_accession=104219098, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  65116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65116, CurrentID=0, Status='live', Symbol='Prrg2', Aliases='1600014G02Rik, 2510041M06Rik, N4WBP1', description='proline-rich Gla (G-carboxyglutamic acid) polypeptide 2', other_designations='NEDD4 WW domain-binding protein 1|proline-rich Gla protein 2|proline-rich gamma-carboxyglutamic acid protein 2', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45053584, end_position_on_the_genomic_accession=45062526, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  65221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65221, CurrentID=0, Status='live', Symbol='Slc15a3', Aliases='Ci1, cI-1', description='solute carrier family 15, member 3', other_designations='cAMP inducible gene 1|cAMP-inducible gene 1 protein|peptide transporter 3|peptide/histidine transporter 2', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10842544, end_position_on_the_genomic_accession=10857915, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  65246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65246, CurrentID=0, Status='live', Symbol='Xpo7', Aliases='4930506C02Rik, BB164534, Ranbp16, mKIAA0745', description='exportin 7', other_designations='RAN binding protein 16|exp7|ran-binding protein 16', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70654246, end_position_on_the_genomic_accession=70766634, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  65247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65247, CurrentID=0, Status='live', Symbol='Asb1', Aliases='1700029O08Rik, 1700054C17Rik, AI426616, AI851378, mKIAA1146', description='ankyrin repeat and SOCS box-containing 1', other_designations='ASB-1|ankyrin repeat and SOCS box-containing protein 1|ankyrin repeat-containing SOCS box protein 1', map_location='1 46.19 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91539459, end_position_on_the_genomic_accession=91559590, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  65254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65254, CurrentID=0, Status='live', Symbol='Dpysl5', Aliases='CRAM, CRMP-5, Crmp5', description='dihydropyrimidinase-like 5', other_designations='DRP-5|collapsin response mediator protein 5', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30711569, end_position_on_the_genomic_accession=30799369, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  65255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65255, CurrentID=0, Status='live', Symbol='Asb4', Aliases='8430401O13Rik, AV117194, AW060255', description='ankyrin repeat and SOCS box-containing 4', other_designations='ASB-4|ankyrin repeat and SOCS box-containing protein 4', map_location='6 1.99 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=5383386, end_position_on_the_genomic_accession=5433022, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  65256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65256, CurrentID=0, Status='live', Symbol='Asb2', Aliases='1110008E15Rik', description='ankyrin repeat and SOCS box-containing 2', other_designations='ASB-2|ankyrin repeat and SOCS box-containing protein 2|ankyrin repeat-containing SOCS box protein 2|asb2', map_location='12 52.9 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103321142, end_position_on_the_genomic_accession=103356001, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  65257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65257, CurrentID=0, Status='live', Symbol='Asb3', Aliases='2400011J03Rik, Asb-3', description='ankyrin repeat and SOCS box-containing 3', other_designations='ankyrin repeat and SOCS box-containing protein 3|ankyrin repeat-containing SOCS box protein 3', map_location='11 18.19 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30954194, end_position_on_the_genomic_accession=31102704, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100073351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100073351, CurrentID=0, Status='live', Symbol='Yy2', Aliases='', description='Yy2 transcription factor', other_designations='YY-2|yin and yang 2', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=157566119, end_position_on_the_genomic_accession=157568985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  327655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327655, CurrentID=0, Status='live', Symbol='Ppip5k1', Aliases='B430315C20Rik, Hisppd2a, mKIAA0377', description='diphosphoinositol pentakisphosphate kinase 1', other_designations='VIP1 homolog|histidine acid phosphatase domain containing 2A|histidine acid phosphatase domain-containing protein 2A|insP6 and PP-IP5 kinase 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121310557, end_position_on_the_genomic_accession=121355636, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  327743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327743, CurrentID=0, Status='live', Symbol='Wisp3', Aliases='CCN6, ENSMUSG00000062074, Gm735, WISP-3', description='WNT1 inducible signaling pathway protein 3', other_designations='CCN family member 6', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39150971, end_position_on_the_genomic_accession=39163794, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  327747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327747, CurrentID=0, Status='live', Symbol='Mettl24', Aliases='9030224M15Rik', description='methyltransferase like 24', other_designations='UPF0624 protein C6orf186 homolog', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40682299, end_position_on_the_genomic_accession=40811086, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  327762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327762, CurrentID=0, Status='live', Symbol='Dna2', Aliases='Dna2l, E130315B21Rik', description='DNA replication helicase/nuclease 2', other_designations='DNA replication ATP-dependent helicase-like homolog|DNA replication helicase 2 homolog|DNA2-like helicase', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62946980, end_position_on_the_genomic_accession=62974188, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  327766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327766, CurrentID=0, Status='live', Symbol='Tmem26', Aliases='3230401A06, AI195372', description='transmembrane protein 26', other_designations='', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=68723746, end_position_on_the_genomic_accession=68782655, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  327799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327799, CurrentID=0, Status='live', Symbol='Usp44', Aliases='E430004F17Rik', description='ubiquitin specific peptidase 44', other_designations='deubiquitinating enzyme 44|inactive deubiquitinating enzyme 44|inactive ubiquitin carboxyl-terminal hydrolase 44|inactive ubiquitin thioesterase 44|inactive ubiquitin-specific-processing protease 44|ubiquitin specific peptidase 49 pseudogene|ubiquitin thioesterase 44|ubiquitin-specific-processing protease 44', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93819968, end_position_on_the_genomic_accession=93860738, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  327814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327814, CurrentID=0, Status='live', Symbol='Ppfia2', Aliases='5330438O12, B230207K17Rik, E130120L08Rik, mKIAA4112', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 2', other_designations='PTPRF interacting protein alpha 2|PTPRF-interacting protein alpha-2|liprin-alpha 2|protein tyrosine phosphatase receptor type f polypeptide-interacting protein alpha-2', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=106468872, end_position_on_the_genomic_accession=106933468, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  327826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327826, CurrentID=0, Status='live', Symbol='Frs2', Aliases='4732458E18, C330018A15Rikalpha, SNT1, Frs2', description='fibroblast growth factor receptor substrate 2', other_designations='FGFR substrate 2|FGFR-signaling adaptor SNT|FRS2-alpha|SNT-1|suc1-associated neurotrophic factor target 1', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117070127, end_position_on_the_genomic_accession=117148474, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  327900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327900, CurrentID=0, Status='live', Symbol='Ubtd2', Aliases='4930571L24, 9630054F20Rik, AI645720', description='ubiquitin domain containing 2', other_designations='', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32455372, end_position_on_the_genomic_accession=32518709, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  327942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327942, CurrentID=0, Status='live', Symbol='Pigl', Aliases='Gm737', description='phosphatidylinositol glycan anchor biosynthesis, class L', other_designations='PIG-L|ortholog of human and rat phosphatidylinositol glycan class L PIGL|phosphatidylinositol glycan, class L|phosphatidylinositol-glycan biosynthesis class L protein', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62458460, end_position_on_the_genomic_accession=62513900, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  327951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327951, CurrentID=0, Status='live', Symbol='Cyb5d1', Aliases='Gm740', description='cytochrome b5 domain containing 1', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69390930, end_position_on_the_genomic_accession=69395346, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  327954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327954, CurrentID=0, Status='live', Symbol='Dnah2', Aliases='2900022L05Rik, 4930564A01, D130094J20, D330014H01Rik, Dnahc2, Dnhd3', description='dynein, axonemal, heavy chain 2', other_designations='axonemal beta dynein heavy chain 2|axonemal dynein heavy chain|ciliary dynein heavy chain 2|dynein heavy chain domain 3|dynein, axon, heavy chain 2', map_location='11 42.63 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69420807, end_position_on_the_genomic_accession=69549108, orientation='minus', exon_count=94, OMIM='', no_hdr0=''),
  327956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327956, CurrentID=0, Status='live', Symbol='Vmo1', Aliases='Gm741', description='vitelline membrane outer layer 1 homolog (chicken)', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70513516, end_position_on_the_genomic_accession=70514616, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  327957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327957, CurrentID=0, Status='live', Symbol='Scimp', Aliases='A430084P05Rik', description='SLP adaptor and CSK interacting membrane protein', other_designations='transmembrane protein C17orf87 homolog', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70790932, end_position_on_the_genomic_accession=70812561, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  327958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327958, CurrentID=0, Status='live', Symbol='Pitpnm3', Aliases='A330068P14Rik, AI848332, Ackr6, Gm880', description='PITPNM family member 3', other_designations='NIR-1|PITPnm 3|phosphatidylinositol transfer protein, membrane-associated 3|pyk2 N-terminal domain-interacting receptor 1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72047528, end_position_on_the_genomic_accession=72135889, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  327959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327959, CurrentID=0, Status='live', Symbol='Xaf1', Aliases='Fbox39', description='XIAP associated factor 1', other_designations='BIRC4-binding protein|F-box protein 39|novel TRAF-type zinc finger protein', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72301436, end_position_on_the_genomic_accession=72313733, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  327978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327978, CurrentID=0, Status='live', Symbol='Slfn5', Aliases='AI956840', description='schlafen 5', other_designations='expressesd sequence AI956840', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82951349, end_position_on_the_genomic_accession=82964850, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  327987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327987, CurrentID=0, Status='live', Symbol='Med13', Aliases='1110067M05Rik, D030023K18, Thrap1, Trap240', description='mediator complex subunit 13', other_designations='Zmediator complex subunit 13|mediator of RNA polymerase II transcription, subunit 13|thyroid hormone receptor associated protein 1|thyroid hormone receptor associated protein, 240 kDa subunit|thyroid hormone receptor-associated protein 1|thyroid hormone receptor-associated protein complex 240 kDa component', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86265715, end_position_on_the_genomic_accession=86357525, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  327992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=327992, CurrentID=0, Status='live', Symbol='Hsf5', Aliases='Gm739, MLZ-220', description='heat shock transcription factor family member 5', other_designations='HSF 5|HSTF 5|heat shock transcription factor 5|protein expressed in male leptotene and zygotene spermatocytes 220', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87616404, end_position_on_the_genomic_accession=87659881, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  328019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328019, CurrentID=0, Status='live', Symbol='Spata32', Aliases='4933400C05Rik, VAD1.2', description='spermatogenesis associated 32', other_designations='acrosome expressed protein 2', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103208127, end_position_on_the_genomic_accession=103218432, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  328035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328035, CurrentID=0, Status='live', Symbol='Fads6', Aliases='BC044804, OTTMUSG00000021749', description='fatty acid desaturase domain family, member 6', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115279622, end_position_on_the_genomic_accession=115297822, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  328059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328059, CurrentID=0, Status='live', Symbol='Slc7a15', Aliases='Arpat', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 15', other_designations='', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8528481, end_position_on_the_genomic_accession=8599066, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76524, CurrentID=0, Status='live', Symbol='Cln6', Aliases='1810065L06Rik, AW743417, D9Bwg1455e, nclf', description='ceroid-lipofuscinosis, neuronal 6', other_designations='neuronal ceroid lipofuscinosis', map_location='9 33.89 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62838787, end_position_on_the_genomic_accession=62852002, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  65945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65945, CurrentID=0, Status='live', Symbol='Clstn1', Aliases='1810034E21Rik, Cst-1, Cstn1', description='calsyntenin 1', other_designations='alc-alpha|alcadein alpha', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149586439, end_position_on_the_genomic_accession=149648899, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  328092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328092, CurrentID=0, Status='live', Symbol='Dtd2', Aliases='4930578F06Rik, 6530401N04Rik, B830049N13Rik', description='D-tyrosyl-tRNA deacylase 2', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51997507, end_position_on_the_genomic_accession=52006501, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  328099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328099, CurrentID=0, Status='live', Symbol='Prps1l3', Aliases='AU021838, AW536152, Gm5081', description='phosphoribosyl pyrophosphate synthetase 1-like 3', other_designations='', map_location='12 24.7 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=57230412, end_position_on_the_genomic_accession=57242168, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  65956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65956, CurrentID=0, Status='live', Symbol='Ccl21c', Aliases='6Ckine, CKb9, Ccl21b, SLC, Scya21, Scya21b, Scya21c, TCA4, exodus-2', description='chemokine (C-C motif) ligand 21C (leucine)', other_designations='C-C motif chemokine 21b|Ccl21-leu|Small-inducible cytokine A21b|beta chemokine exodus-2|beta-chemokine exodus-2|small inducible cytokine A21c|small-inducible cytokine A21c|thymus-derived chemotactic agent 4', map_location='4 22.54 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  65960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65960, CurrentID=0, Status='live', Symbol='Twsg1', Aliases='1810013J15Rik, 9030422N06Rik, AW552143, D17Ertd403e, Tsg, Twg', description='twisted gastrulation homolog 1 (Drosophila)', other_designations='', map_location='17 35.26 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65923065, end_position_on_the_genomic_accession=65951187, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  65961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65961, CurrentID=0, Status='live', Symbol='Utp3', Aliases='2400011K09Rik, AW557551, C87704, Crl1, Crlz1, Sas10', description='UTP3, small subunit (SSU) processome component, homolog (S. cerevisiae)', other_designations='UTP3 homolog|charged amino acid rich leucine zipper 1|charged amino acid-rich leucine zipper 1|crl-1|disrupter of silencing SAS10', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88554483, end_position_on_the_genomic_accession=88556083, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  65962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65962, CurrentID=0, Status='live', Symbol='Slc9a3r2', Aliases='0610011L07Rik, 1200011K07Rik, 2010007A20Rik, E3karp, NHERF-2, Nherf2, Octs2, Sip-1, Sip1, Sryip1, Tka-1', description='solute carrier family 9 (sodium/hydrogen exchanger), member 3 regulator 2', other_designations='NHE3 kinase A regulatory protein E3KARP|SRY-interacting protein 1|Sry interacting protein 1|sodium-hydrogen exchanger regulatory factor 2|sodium/hydrogen exchanger|solute carrier family 9 (sodium/hydrogen exchanger), isoform 3 regulator 2|solute carrier family 9 isoform 3 regulator 2, isoform A; sodium/hydrogen exchanger|solute carrier family 9 isoform A3 regulatory factor 2|tyrosine kinase activator protein 1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24639281, end_position_on_the_genomic_accession=24650340, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  65963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65963, CurrentID=0, Status='live', Symbol='Tmem176b', Aliases='1810009M01Rik, Clast1, Lr8', description='transmembrane protein 176B', other_designations='', map_location='6 23.75 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48833811, end_position_on_the_genomic_accession=48845950, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  65964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65964, CurrentID=0, Status='live', Symbol='Zak', Aliases='AV006891, B230120H23Rik, MLTK, MLTKalpha, MLTKbeta', description='sterile alpha motif and leucine zipper containing kinase AZK', other_designations='MLK-like mitogen-activated protein triple kinase|MLK-related kinase|MRK|leucine zipper- and sterile alpha motif kinase ZAK|mixed lineage kinase-related kinase|sterile alpha motif- and leucine zipper-containing kinase AZK|sterile-alpha motif and leucine zipper containing kinase AZK', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=72285637, end_position_on_the_genomic_accession=72442610, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  328110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328110, CurrentID=0, Status='live', Symbol='Prpf39', Aliases='Srcs1', description='pre-mRNA processing factor 39', other_designations='PRP39 homolog|PRP39 pre-mRNA processing factor 39 homolog|v-src suppressed transcript 1', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=65036309, end_position_on_the_genomic_accession=65063386, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  65967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65967, CurrentID=0, Status='live', Symbol='Eefsec', Aliases='Selb, sec', description='eukaryotic elongation factor, selenocysteine-tRNA-specific', other_designations='MJ0495-like protein SelB|elongation factor sec|mSelB', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88257334, end_position_on_the_genomic_accession=88446560, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  65969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65969, CurrentID=0, Status='live', Symbol='Cubn', Aliases='AA408369, AL022750, D2Wsu88e', description='cubilin (intrinsic factor-cobalamin receptor)', other_designations='intrinsic factor-cobalamin receptor', map_location='2 9.86 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=13276338, end_position_on_the_genomic_accession=13491876, orientation='minus', exon_count=67, OMIM='', no_hdr0=''),
  65970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65970, CurrentID=0, Status='live', Symbol='Lima1', Aliases='Eplin', description='LIM domain and actin binding 1', other_designations='epithelial protein lost in neoplasm', map_location='15 56.13 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99778468, end_position_on_the_genomic_accession=99875456, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  65971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65971, CurrentID=0, Status='live', Symbol='Tbata', Aliases='1700021K02Rik, AI428928, Spatial, Titest', description='thymus, brain and testes associated', other_designations='protein SPATIAL|stromal protein associated with thymii and lymph node|stromal protein associated with thymii and lymph nodes|thymus and testis specific nuclear protein|thymus, brain and testes-associated protein', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61171964, end_position_on_the_genomic_accession=61188841, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  65972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65972, CurrentID=0, Status='live', Symbol='Ifi30', Aliases='GILT, IP30', description='interferon gamma inducible protein 30', other_designations='gamma-interferon-inducible protein IP-30|lysosomal thio reductase IP30|lysosomal thiol reductase IP30', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70762773, end_position_on_the_genomic_accession=70766663, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  65973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=65973, CurrentID=0, Status='live', Symbol='Asph', Aliases='2310005F16Rik, 3110001L23Rik, AI848629, AW261690, AW561901, BAH, C79816, cI-37', description='aspartate-beta-hydroxylase', other_designations='ASP beta-hydroxylase|aspartate beta-hydroxylase|aspartyl beta-hydroxylase|calsequestrin-binding protein|jumbug|junctate|junctin|peptide-aspartate beta-dioxygenase', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=9449085, end_position_on_the_genomic_accession=9669344, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  328133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328133, CurrentID=0, Status='live', Symbol='Slc39a9', Aliases='2010002A02Rik, 2610511I23Rik, 4833420E20Rik, AW209151', description='solute carrier family 39 (zinc transporter), member 9', other_designations='ZIP-9|solute carrier family 39 member 9|zrt- and Irt-like protein 9', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80644215, end_position_on_the_genomic_accession=80683342, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66011, CurrentID=0, Status='live', Symbol='Ranbp17', Aliases='', description='RAN binding protein 17', other_designations='', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=33211794, end_position_on_the_genomic_accession=33514750, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  328162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328162, CurrentID=0, Status='live', Symbol='Trmt61a', Aliases='6720458F09Rik, AI606093, Gcd14, Trm61', description='tRNA methyltransferase 61A', other_designations='tRNA (adenine-N(1)-)-methyltransferase catalytic subunit TRM61|tRNA (adenine-N(1)-)-methyltransferase catalytic subunit TRMT61A|tRNA methyltransferase 61 homolog A|tRNA(m1A58)-methyltransferase subunit TRM61|tRNA(m1A58)-methyltransferase subunit TRMT61A|tRNA(m1A58)MTase subunit TRM61|tRNA(m1A58)MTase subunit TRMT61A', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111678105, end_position_on_the_genomic_accession=111683902, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66039, CurrentID=0, Status='live', Symbol='Tmem254a', Aliases='0610008K04Rik, 0610010A22Rik, AA960404, D14Ertd449e, Tmem254', description='transmembrane protein 254a', other_designations='tranmembrane protein 254a|transmembrane protein C10orf57 homolog', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25923298, end_position_on_the_genomic_accession=25927672, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66042, CurrentID=0, Status='live', Symbol='Sostdc1', Aliases='0610006G05Rik, Sostl, USAG-1, Wise, ectodin', description='sclerostin domain containing 1', other_designations='cystine-knot containing secreted protein|ectodermal BMP inhibitor|sclerostin-like protein|uterine sensitization-associated gene 1 protein|uterine sensitization-associated protein 1', map_location='12 B2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=36314169, end_position_on_the_genomic_accession=36318452, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66043, CurrentID=0, Status='live', Symbol='Atp5d', Aliases='0610008F14Rik, 1500000I11Rik, AA960090, AI876556, AU020773, C85518', description='ATP synthase, H+ transporting, mitochondrial F1 complex, delta subunit', other_designations='F-ATPase delta subunit', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80142280, end_position_on_the_genomic_accession=80145818, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66044, CurrentID=0, Status='live', Symbol='Dtd1', Aliases='0610006H08Rik, DUE-B, Hars2', description='D-tyrosyl-tRNA deacylase 1', other_designations='D-tyrosyl-tRNA deacylase 1 homolog|DNA-unwinding element-binding protein B', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144599953, end_position_on_the_genomic_accession=144768747, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66046, CurrentID=0, Status='live', Symbol='Ndufb5', Aliases='0610007D05Rik, AU015782, CI-SGDH, SGDH', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 5', other_designations='NADH-ubiquinone oxidoreductase SGDH subunit|complex I-SGDH', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32737063, end_position_on_the_genomic_accession=32751560, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66047, CurrentID=0, Status='live', Symbol='Mrpl54', Aliases='0610008M19Rik, D10Sut1e', description='mitochondrial ribosomal protein L54', other_designations='L54mt|MRP-L54', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81264722, end_position_on_the_genomic_accession=81266926, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66048, CurrentID=0, Status='live', Symbol='Emc6', Aliases='0610009E20Rik, 0610025L18Rik, Tmem93', description='ER membrane protein complex subunit 6', other_designations='transmembrane protein 93', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73175503, end_position_on_the_genomic_accession=73177042, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66049, CurrentID=0, Status='live', Symbol='Rogdi', Aliases='0610011C19Rik, AU020118, C76152, Lzf', description='rogdi homolog (Drosophila)', other_designations='leucine zipper domain protein|leucine-zipper-containing LZF', map_location='16 2.49 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5008729, end_position_on_the_genomic_accession=5013805, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66050, CurrentID=0, Status='live', Symbol='0610009B22Rik', Aliases='', description='RIKEN cDNA 0610009B22 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51685380, end_position_on_the_genomic_accession=51689169, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66052, CurrentID=0, Status='live', Symbol='Sdhc', Aliases='0610010E03Rik, AI316496, AU019277', description='succinate dehydrogenase complex, subunit C, integral membrane protein', other_designations='QPs-1|QPs1|integral membrane protein CII-3', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171129157, end_position_on_the_genomic_accession=171150603, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66053, CurrentID=0, Status='live', Symbol='Ppil2', Aliases='0610009L05Rik, 1700016N17Rik, 4921520K19Rik, 4930511F14Rik, AA589416, C130078A06Rik, CYC4', description='peptidylprolyl isomerase (cyclophilin)-like 2', other_designations='PPIase|cyclophilin-60|cyclophilin-like protein Cyp-60|rotamase PPIL2', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17086555, end_position_on_the_genomic_accession=17111263, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  66054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66054, CurrentID=0, Status='live', Symbol='Cndp2', Aliases='0610010E05Rik, C76600, Cn2, Dip-2, Pep-1, Pep1', description='CNDP dipeptidase 2 (metallopeptidase M20 family)', other_designations='cytosolic nonspecific dipeptidase|glutamate carboxypeptidase-like protein 1|peptidase 1|peptidase A', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84667465, end_position_on_the_genomic_accession=84685702, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66055, CurrentID=0, Status='live', Symbol='Sf3b6', Aliases='0610009D07Rik, 6030419K15Rik, AV001342, Sf3b14', description='splicing factor 3B, subunit 6', other_designations='SF3B 14 kDa subunit|pre-mRNA branch site protein p14|splicing factor 3B, 14 kDa subunit', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4817608, end_position_on_the_genomic_accession=4827659, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66056, CurrentID=0, Status='live', Symbol='Zfp524', Aliases='0610012F07Rik, 2300009P13Rik, Znf524', description='zinc finger protein 524', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5015152, end_position_on_the_genomic_accession=5018493, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66058, CurrentID=0, Status='live', Symbol='Tmem176a', Aliases='0610011I04Rik, AU040201, AU041743, Keg2', description='transmembrane protein 176A', other_designations='gene signature 188|kidney expressed gene 2|kidney-expressed gene 2 protein', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48841483, end_position_on_the_genomic_accession=48845364, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66059, CurrentID=0, Status='live', Symbol='Krtcap2', Aliases='0610010I12Rik, Kcp2', description='keratinocyte associated protein 2', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89246300, end_position_on_the_genomic_accession=89249723, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66061, CurrentID=0, Status='live', Symbol='Tctex1d2', Aliases='0610012D17Rik, 2410011K15Rik', description='Tctex1 domain containing 2', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32419686, end_position_on_the_genomic_accession=32429099, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66065, CurrentID=0, Status='live', Symbol='Hsd17b14', Aliases='0610039E24Rik, Dhrs10, retSDR3', description='hydroxysteroid (17-beta) dehydrogenase 14', other_designations='dehydrogenase/reductase (SDR family) member 10', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45554855, end_position_on_the_genomic_accession=45567321, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66066, CurrentID=0, Status='live', Symbol='Gng11', Aliases='0610037B21Rik', description='guanine nucleotide binding protein (G protein), gamma 11', other_designations='', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=4003987, end_position_on_the_genomic_accession=4008445, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66067, CurrentID=0, Status='live', Symbol='Gtpbp8', Aliases='0610037H22Rik', description='GTP-binding protein 8 (putative)', other_designations='', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44738879, end_position_on_the_genomic_accession=44746377, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66069, CurrentID=0, Status='live', Symbol='Snupn', Aliases='0610031A09Rik, Rnut11, Snupn', description='snurportin 1', other_designations='RNA U transporter 1|RNA, U transporter 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56950924, end_position_on_the_genomic_accession=56983206, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66070, CurrentID=0, Status='live', Symbol='Cwc15', Aliases='0610040D20Rik, 2900052N06Rik, Ed1, c11orf5, mED1', description='CWC15 spliceosome-associated protein', other_designations='CWC15 homolog|embryonic development factor 1|nucleoporin|protein CWC15 homolog', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14500606, end_position_on_the_genomic_accession=14510620, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66071, CurrentID=0, Status='live', Symbol='Ethe1', Aliases='0610025L15Rik, Hsco', description='ethylmalonic encephalopathy 1', other_designations='ethylmalonic encephalopathy protein 1 homolog|hepatoma subtracted clone one protein|protein ETHE1, mitochondrial|sulfur dioxygenase ETHE1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24587543, end_position_on_the_genomic_accession=24608925, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66072, CurrentID=0, Status='live', Symbol='Sdhaf2', Aliases='0610038F07Rik, AA407634, AW049997', description='succinate dehydrogenase complex assembly factor 2', other_designations='SDH assembly factor 2|succinate dehydrogenase subunit 5, mitochondrial', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10500512, end_position_on_the_genomic_accession=10525209, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66073, CurrentID=0, Status='live', Symbol='Txndc12', Aliases='0610040B21Rik, ERp16, ERp19', description='thioredoxin domain containing 12 (endoplasmic reticulum)', other_designations='ER protein 19|endoplasmic reticulum protein ERp19|endoplasmic reticulum resident protein 19|ortholog of endoplasmic reticulum thioredoxin superfamily member, 18 kDa TLP19|thioredoxin-like protein p19', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108834678, end_position_on_the_genomic_accession=108862120, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66074, CurrentID=0, Status='live', Symbol='Tmem167', Aliases='0610041E09Rik, 5730424F14Rik, AU041184, AW537806, ENSMUSG00000060669, Gm10085a, Tmem167', description='transmembrane protein 167', other_designations='transmembrane protein 167A', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=90089667, end_position_on_the_genomic_accession=90114921, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66075, CurrentID=0, Status='live', Symbol='Chchd3', Aliases='0610041L09Rik, 1700039J09Rik, AW558177', description='coiled-coil-helix-coiled-coil-helix domain containing 3', other_designations='coiled-coil-helix-coiled-coil-helix domain-containing protein 3, mitochondrial', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=32792217, end_position_on_the_genomic_accession=33060276, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66077, CurrentID=0, Status='live', Symbol='Aurkaip1', Aliases='0610033H09Rik, Aip, Akip, MRP-S38', description='aurora kinase A interacting protein 1', other_designations='28S ribosomal protein S38, mitochondrial|AURKA-interacting protein|Aurora-A kinase interacting protein', map_location='4 87.55 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155831269, end_position_on_the_genomic_accession=155833098, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66078, CurrentID=0, Status='live', Symbol='Tsen34', Aliases='0610027F08Rik, Leng5', description='tRNA splicing endonuclease 34 homolog (S. cerevisiae)', other_designations='leukocyte receptor cluster (LRC) member 5|leukocyte receptor cluster member 5 homolog|tRNA-intron endonuclease Sen34', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3693610, end_position_on_the_genomic_accession=3701035, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66079, CurrentID=0, Status='live', Symbol='Tmem42', Aliases='0610027O18Rik, 4933429E06Rik, AV003444, D9Ertd662e', description='transmembrane protein 42', other_designations='', map_location='9 73.52 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123021038, end_position_on_the_genomic_accession=123023491, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66082, CurrentID=0, Status='live', Symbol='Abhd6', Aliases='0610041D24Rik, AA673485, AV065425', description='abhydrolase domain containing 6', other_designations='2-arachidonoylglycerol hydrolase|abhydrolase domain-containing protein 6', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8002807, end_position_on_the_genomic_accession=8056773, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66083, CurrentID=0, Status='live', Symbol='Setd6', Aliases='0610039J04Rik, 3110004G14Rik, AI413388, C76402', description='SET domain containing 6', other_designations='SET domain-containing protein 6', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95715113, end_position_on_the_genomic_accession=95721499, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66084, CurrentID=0, Status='live', Symbol='Rmnd1', Aliases='0610042C05Rik, AA408137, AI462664, AW536662', description='required for meiotic nuclear division 1 homolog (S. cerevisiae)', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=4403166, end_position_on_the_genomic_accession=4432407, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66085, CurrentID=0, Status='live', Symbol='Eif3f', Aliases='0610037M02Rik, AA409853, Eif3s5', description='eukaryotic translation initiation factor 3, subunit F', other_designations='deubiquitinating enzyme eIF3f|eIF-3-epsilon|eIF3 p47|eIF3f|eukaryotic translation initiation factor 3 subunit 5 epsilon|eukaryotic translation initiation factor 3, subunit 5 (epsilon)', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108934415, end_position_on_the_genomic_accession=108941942, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66086, CurrentID=0, Status='live', Symbol='Fopnl', Aliases='0610037P05Rik', description='Fgfr1op N-terminal like', other_designations='FGFR1OP N-terminal-like protein|FOP-related protein of 20 kDa|lisH domain-containing protein C16orf63 homolog', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14299241, end_position_on_the_genomic_accession=14317926, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66087, CurrentID=0, Status='live', Symbol='Emc3', Aliases='0610039A15Rik, AI225901, AW260416, Pob, Tmem111', description='ER membrane protein complex subunit 3', other_designations='partial optokinetic response b|transmembrane protein 111', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113514887, end_position_on_the_genomic_accession=113531638, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  328232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328232, CurrentID=0, Status='live', Symbol='Gfod1', Aliases='9630032O13, AI850995', description='glucose-fructose oxidoreductase domain containing 1', other_designations='', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43195240, end_position_on_the_genomic_accession=43304172, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66089, CurrentID=0, Status='live', Symbol='Rmnd5b', Aliases='0610039K22Rik, Gid2', description='required for meiotic nuclear division 5 homolog B (S. cerevisiae)', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51623673, end_position_on_the_genomic_accession=51635896, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66090, CurrentID=0, Status='live', Symbol='Ypel3', Aliases='0610043B10Rik, 1190001G19Rik, Suap', description='yippee-like 3 (Drosophila)', other_designations='small ubiquitinated apoptotic protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126776975, end_position_on_the_genomic_accession=126780514, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66091, CurrentID=0, Status='live', Symbol='Ndufa3', Aliases='1010001M12Rik, 1700022J01Rik', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 3', other_designations='CI-B9|Complex I-B9|NADH-ubiquinone oxidoreductase B9 subunit', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3617373, end_position_on_the_genomic_accession=3620161, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66092, CurrentID=0, Status='live', Symbol='Ghitm', Aliases='1010001P14Rik, C77840, MICS1, PTD010, Tmbim5', description='growth hormone inducible transmembrane protein', other_designations='mitochondrial morphology and cristae structure 1', map_location='14 21.29 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=37120444, end_position_on_the_genomic_accession=37135327, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66094, CurrentID=0, Status='live', Symbol='Lsm7', Aliases='0910001B06Rik, 1110033F18Rik', description='LSM7 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM7 homolog, U6 small nuclear RNA associated', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80852823, end_position_on_the_genomic_accession=80855249, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66096, CurrentID=0, Status='live', Symbol='Lamtor4', Aliases='0910001L09Rik, AV006840', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 4', other_designations='UPF0539 protein C7orf59 homolog|late endosomal/lysosomal adaptor and MAPK and MTOR activator 4', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138255482, end_position_on_the_genomic_accession=138259395, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66098, CurrentID=0, Status='live', Symbol='Chchd6', Aliases='0710001P09Rik, 1700021B03Rik', description='coiled-coil-helix-coiled-coil-helix domain containing 6', other_designations='coiled-coil-helix-coiled-coil-helix domain-containing protein 6, mitochondrial', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89383146, end_position_on_the_genomic_accession=89595652, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66101, CurrentID=0, Status='live', Symbol='Ppih', Aliases='1100001J08Rik, 2010111B15Rik, 4833408F11Rik, AI464484, CYP-20, CYPH, D4Wsu43e', description='peptidyl prolyl isomerase H', other_designations='PPIase H|cyclophilin H|rotamase H', map_location='4 D', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119300010, end_position_on_the_genomic_accession=119320523, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66102, CurrentID=0, Status='live', Symbol='Cxcl16', Aliases='0910001K24Rik, AV290116, BB024863, CXCL16v1, CXCL16v2, SR-PSOX, Zmynd15, b2b498Clo', description='chemokine (C-X-C motif) ligand 16', other_designations='Cxc chemokine ligand 16|Mutant line 498|SR-PSOX/CXCL16|scavenger receptor for phosphatidylserine and oxidized low density lipoprotein|small-inducible cytokine B16|transmembrane chemokine CXCL16|zinc finger, MYND-type containing 15', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70454234, end_position_on_the_genomic_accession=70459984, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66104, CurrentID=0, Status='live', Symbol='Tceal6', Aliases='1500026B10Rik', description='transcription elongation factor A (SII)-like 6', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135208686, end_position_on_the_genomic_accession=135210687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66105, CurrentID=0, Status='live', Symbol='Ube2d3', Aliases='1100001F19Rik, 9430029A22Rik, AA414951', description='ubiquitin-conjugating enzyme E2D 3', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme D3|E2 ubiquitin-conjugating enzyme D3|E2(17)KB 3|ubiquitin carrier protein D3|ubiquitin-conjugating enzyme E2(17)KB 3|ubiquitin-conjugating enzyme E2-17 kDa 3|ubiquitin-conjugating enzyme E2D 3 (UBC4/5 homolog, yeast)|ubiquitin-protein ligase D3', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135438244, end_position_on_the_genomic_accession=135467180, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66106, CurrentID=0, Status='live', Symbol='Smpx', Aliases='1010001C09Rik, Csl', description='small muscle protein, X-linked', other_designations='cardiac and skeletal muscle protein|chisel protein|stretch-responsive skeletal muscle protein', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=157698973, end_position_on_the_genomic_accession=157752591, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66107, CurrentID=0, Status='live', Symbol='Wfdc21', Aliases='1100001G20Rik, Wdnm1-like', description='WAP four-disulfide core domain 21', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83746940, end_position_on_the_genomic_accession=83752646, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66108, CurrentID=0, Status='live', Symbol='Ndufa9', Aliases='1010001N11Rik', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 9', other_designations='CI-39kD|NADH-ubiquinone oxidoreductase 39 kDa subunit|complex I-39kD', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126821863, end_position_on_the_genomic_accession=126849144, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66109, CurrentID=0, Status='live', Symbol='Tspan13', Aliases='1100001I23Rik, NET-6, Tm4sf13', description='tetraspanin 13', other_designations='transmembrane 4 superfamily member 13|tspan-13', map_location='12 B2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=36014555, end_position_on_the_genomic_accession=36042478, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66111, CurrentID=0, Status='live', Symbol='Tmed3', Aliases='1200002G13Rik, AW546672, P24b', description='transmembrane emp24 domain containing 3', other_designations='integral type I protein|p24 family protein gamma-4|p24gamma4', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89699203, end_position_on_the_genomic_accession=89705043, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66112, CurrentID=0, Status='live', Symbol='Marc1', Aliases='1300013F15Rik, Mosc1', description='mitochondrial amidoxime reducing component 1', other_designations='MOCO sulphurase C-terminal domain containing 1|MOSC domain-containing protein 1, mitochondrial|moco sulfurase C-terminal domain-containing protein 1|molybdenum cofactor sulfurase C-terminal domain-containing protein 1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=184786767, end_position_on_the_genomic_accession=184811308, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66113, CurrentID=0, Status='live', Symbol='Apoa5', Aliases='1300007O05Rik, Apoav, RAP3', description='apolipoprotein A-V', other_designations='apo-AV|apoA-V|apolipoprotein A5|regeneration-associated protein 3', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46268608, end_position_on_the_genomic_accession=46271919, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66114, CurrentID=0, Status='live', Symbol='Dnajc30', Aliases='1300007M11Rik, Wbscr18', description='DnaJ heat shock protein family (Hsp40) member C30', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 30|Williams-Beuren syndrome chromosome region 18 homolog|Williams-Beuren syndrome critical region 18|williams-Beuren syndrome chromosomal region 18 protein homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135064206, end_position_on_the_genomic_accession=135065365, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66116, CurrentID=0, Status='live', Symbol='Nat8f1', Aliases='1110002I11Rik, AI266962', description='N-acetyltransferase 8 (GCN5-related) family member 1', other_designations='camello-like 1|camello-like protein 1|cml1', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85910154, end_position_on_the_genomic_accession=85915677, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66117, CurrentID=0, Status='live', Symbol='1110001J03Rik', Aliases='Fmc1', description='RIKEN cDNA 1110001J03 gene', other_designations='formation of mitochondrial complexes 1 homolog', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38534861, end_position_on_the_genomic_accession=38539449, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66118, CurrentID=0, Status='live', Symbol='Sarnp', Aliases='1110005A23Rik, Cip29', description='SAP domain containing ribonucleoprotein', other_designations='nuclear protein Hcc-1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128821766, end_position_on_the_genomic_accession=128877638, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  66119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66119, CurrentID=0, Status='live', Symbol='Tomm6', Aliases='1110002E23Rik, AI663970', description='translocase of outer mitochondrial membrane 6 homolog (yeast)', other_designations='overexpressed breast tumor protein homolog|translocase of outer membrane 6 kDa subunit homolog', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47686645, end_position_on_the_genomic_accession=47688386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66120, CurrentID=0, Status='live', Symbol='Fkbp11', Aliases='1110002O23Rik, FKBP-11', description='FK506 binding protein 11', other_designations='19 kDa FK506-binding protein|19 kDa FKBP|FK506-binding protein 11|FKBP-19|PPIase FKBP11|rotamase', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98724366, end_position_on_the_genomic_accession=98728198, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66121, CurrentID=0, Status='live', Symbol='Chchd1', Aliases='1110001O19Rik, 2400010G13Rik, MRP-S37', description='coiled-coil-helix-coiled-coil-helix domain containing 1', other_designations='28S ribosomal protein S37, mitochondrial', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20703027, end_position_on_the_genomic_accession=20704425, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66124, CurrentID=0, Status='live', Symbol='Josd2', Aliases='1110007C05Rik', description='Josephin domain containing 2', other_designations='josephin domain-containing protein 2', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44467980, end_position_on_the_genomic_accession=44471658, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66125, CurrentID=0, Status='live', Symbol='Sf3b5', Aliases='10kDa, 1110005L13Rik, AU043053, Sf3b10', description='splicing factor 3b, subunit 5', other_designations='pre-mRNA-splicing factor SF3b 10 kDa subunit', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13008450, end_position_on_the_genomic_accession=13009183, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66126, CurrentID=0, Status='live', Symbol='Elof1', Aliases='1110011K10Rik', description='ELF1 homolog, elongation factor 1', other_designations='elongation factor 1 homolog (ELF1, S. cerevisiae)', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22112989, end_position_on_the_genomic_accession=22114169, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66128, CurrentID=0, Status='live', Symbol='Mrps36', Aliases='1110018B13Rik, S36mt', description='mitochondrial ribosomal protein S36', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100735940, end_position_on_the_genomic_accession=100744659, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66129, CurrentID=0, Status='live', Symbol='Aaed1', Aliases='1110018J18Rik, AI849003', description='AhpC/TSA antioxidant enzyme domain containing 1', other_designations='UPF0308 protein C9orf21 homolog|ahpC/TSA antioxidant enzyme domain-containing protein 1', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64291836, end_position_on_the_genomic_accession=64312710, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  328274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328274, CurrentID=0, Status='live', Symbol='Zfp459', Aliases='9930025G17Rik, Rslcan-14', description='zinc finger protein 459', other_designations='', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67405712, end_position_on_the_genomic_accession=67443302, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66131, CurrentID=0, Status='live', Symbol='Tipin', Aliases='1110005A05Rik, 1110018P21Rik', description='timeless interacting protein', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64281489, end_position_on_the_genomic_accession=64304794, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66132, CurrentID=0, Status='live', Symbol='1110008L16Rik', Aliases='Kiaa0391, Mrpp3, mKIAA0391', description='RIKEN cDNA 1110008L16 gene', other_designations='mitochondrial RNase P protein 3', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55302637, end_position_on_the_genomic_accession=55382491, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66136, CurrentID=0, Status='live', Symbol='Znrd1', Aliases='1110014N07Rik, Rpa12', description='zinc ribbon domain containing, 1', other_designations='nuclear RNA polymerase I small specific subunit|zinc ribbon domain-containing protein 1', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36954349, end_position_on_the_genomic_accession=36958520, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66138, CurrentID=0, Status='live', Symbol='Wbscr22', Aliases='1110003N24Rik', description='Williams Beuren syndrome chromosome region 22', other_designations='Williams-Beuren syndrome chromosomal region 22 protein homolog|bud site selection protein 23 homolog|ribosome biogenesis methyltransferase WBSCR22', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135052957, end_position_on_the_genomic_accession=135064666, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66139, CurrentID=0, Status='live', Symbol='Tmem8c', Aliases='1110002H13Rik, AI131587', description='transmembrane protein 8C', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27061636, end_position_on_the_genomic_accession=27072161, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66140, CurrentID=0, Status='live', Symbol='Ska2', Aliases='1110001A07Rik, C78640, Fam33a', description='spindle and kinetochore associated complex subunit 2', other_designations='family with sequence similarity 33, member A', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87109261, end_position_on_the_genomic_accession=87122974, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66141, CurrentID=0, Status='live', Symbol='Ifitm3', Aliases='1110004C05Rik, Cd225, Cdw217, DSPA2b, Fgls, IP15, mil-1', description='interferon induced transmembrane protein 3', other_designations='dispanin subfamily A member 2b|fragilis protein|ifitm-like protein 1|interferon-inducible protein 15', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141009590, end_position_on_the_genomic_accession=141010744, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66142, CurrentID=0, Status='live', Symbol='Cox7b', Aliases='1100001F07Rik, 1110004F07Rik, C80563', description='cytochrome c oxidase subunit VIIb', other_designations='cytochrome c oxidase polypeptide VIIb', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106015700, end_position_on_the_genomic_accession=106022450, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66143, CurrentID=0, Status='live', Symbol='Eef1e1', Aliases='1110003A02Rik', description='eukaryotic translation elongation factor 1 epsilon 1', other_designations='elongation factor p18|multisynthase complex auxiliary component p18|multisynthetase complex auxiliary component p18', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38645691, end_position_on_the_genomic_accession=38659028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66144, CurrentID=0, Status='live', Symbol='Atp6v1f', Aliases='1110004G16Rik', description='ATPase, H+ transporting, lysosomal V1 subunit F', other_designations='ATPase, H+ transporting, V1 subunit F|ATPase, vacuolar, 14 kD|V-ATPase 14 kDa subunit|V-ATPase subunit F|vacuolar proton pump subunit F', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29467783, end_position_on_the_genomic_accession=29470509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66146, CurrentID=0, Status='live', Symbol='Tmem57', Aliases='1110007C24Rik, 9230118A01Rik, AI317300, AI606104, C61', description='transmembrane protein 57', other_designations='brain-specific adapter protein C61|brain-specific adaptor protein C61', map_location='4 67.11 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134802759, end_position_on_the_genomic_accession=134853560, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66147, CurrentID=0, Status='live', Symbol='Necap2', Aliases='1110005F07Rik, AA409105, C78898', description='NECAP endocytosis associated 2', other_designations='NECAP endocytosis-associated protein 2|NECAP-2|adaptin-ear-binding coat-associated protein 2|clathrin-coated vesicle coat protein|likely ortholog of H. sapiens adaptin-ear-binding coat-associated protein 2 (NECAP2)', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141066512, end_position_on_the_genomic_accession=141078357, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66148, CurrentID=0, Status='live', Symbol='Dnajc15', Aliases='1110003P16Rik, Dnajd1', description='DnaJ heat shock protein family (Hsp40) member C15', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 15|DnaJ (Hsp40) homolog, subfamily D, member 1', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=77826217, end_position_on_the_genomic_accession=77874917, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66151, CurrentID=0, Status='live', Symbol='Prr13', Aliases='1110020C13Rik, 2010324E22Rik', description='proline rich 13', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102459170, end_position_on_the_genomic_accession=102462806, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66152, CurrentID=0, Status='live', Symbol='Uqcr10', Aliases='1110020P15Rik, AA960494, Ucrc', description='ubiquinol-cytochrome c reductase, complex III subunit X', other_designations='complex III subunit 9|complex III subunit X|cytochrome c1 non-heme 7 kDa protein|ubiquinol-cytochrome c reductase complex (7.2 kD)|ubiquinol-cytochrome c reductase complex 7.2 kDa protein|ubiquinol-cytochrome c reductase complex protein', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4701968, end_position_on_the_genomic_accession=4704362, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66153, CurrentID=0, Status='live', Symbol='Fbxo36', Aliases='0610008D19Rik, 1110020F21Rik, 2410002G19Rik, D1Ertd757e', description='F-box protein 36', other_designations='', map_location='1 43.43 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=84839841, end_position_on_the_genomic_accession=84900487, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66154, CurrentID=0, Status='live', Symbol='Tmem14c', Aliases='1110021D01Rik, AU020193, HSPC194', description='transmembrane protein 14C', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41016250, end_position_on_the_genomic_accession=41022582, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66155, CurrentID=0, Status='live', Symbol='Ufc1', Aliases='1110021H02Rik, AA062310, ESTM29', description='ubiquitin-fold modifier conjugating enzyme 1', other_designations='Ufm1-conjugating enzyme 1', map_location='1 79.31 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171288564, end_position_on_the_genomic_accession=171294982, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66156, CurrentID=0, Status='live', Symbol='Anapc11', Aliases='1110011I19Rik, R75218', description='anaphase promoting complex subunit 11', other_designations='APC11|anaphase promoting complex subunit 11 homolog|cyclosome subunit 11', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120598421, end_position_on_the_genomic_accession=120608198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66158, CurrentID=0, Status='live', Symbol='Cxx1a', Aliases='1110012O05Rik, Mar8.2A/B, Mart8b', description='CAAX box 1A', other_designations='CAAX box 1 homolog A', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53642488, end_position_on_the_genomic_accession=53643763, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66161, CurrentID=0, Status='live', Symbol='Pop4', Aliases='1110023P21Rik, Rpp29', description='processing of precursor 4, ribonuclease P/MRP family, (S. cerevisiae)', other_designations='POP4 (processing of precursor , S. cerevisiae) homolog', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=38262815, end_position_on_the_genomic_accession=38271367, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66162, CurrentID=0, Status='live', Symbol='Bola2', Aliases='1110025L05Rik, BolA-2', description='bolA-like 2 (E. coli)', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126696000, end_position_on_the_genomic_accession=126696693, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66163, CurrentID=0, Status='live', Symbol='Mrpl4', Aliases='1110017G11Rik', description='mitochondrial ribosomal protein L4', other_designations='L4mt|MRP-L4', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21002605, end_position_on_the_genomic_accession=21008839, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66164, CurrentID=0, Status='live', Symbol='Nip7', Aliases='1110017C15Rik, 6330509M23Rik, AA408773, AA410017, PEachy', description='nuclear import 7 homolog (S. cerevisiae)', other_designations='kDa93', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107056877, end_position_on_the_genomic_accession=107060931, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66165, CurrentID=0, Status='live', Symbol='Bccip', Aliases='1110013J05Rik, TOK-1', description='BRCA2 and CDKN1A interacting protein', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133709333, end_position_on_the_genomic_accession=133721145, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66166, CurrentID=0, Status='live', Symbol='S100a14', Aliases='1110013O05Rik, S100a15, S114', description='S100 calcium binding protein A14', other_designations='S100 calcium binding protein A15|S100 calcium-binding protein A14', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90526849, end_position_on_the_genomic_accession=90528837, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66167, CurrentID=0, Status='live', Symbol='Tma7', Aliases='1110017O22Rik, Ccdc72', description='translational machinery associated 7 homolog (S. cerevisiae)', other_designations='coiled-coil domain containing 72|coiled-coil domain-containing protein 72', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109077988, end_position_on_the_genomic_accession=109082381, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66168, CurrentID=0, Status='live', Symbol='Grina', Aliases='1110025J15Rik, Lag, Tmbim3', description='glutamate receptor, ionotropic, N-methyl D-aspartate-associated protein 1 (glutamate binding)', other_designations='NMDA receptor glutamate-binding subunit|glutamate [NMDA] receptor-associated protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76246740, end_position_on_the_genomic_accession=76249911, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66169, CurrentID=0, Status='live', Symbol='Tomm7', Aliases='1110020J08Rik, AW047273, Tom7', description='translocase of outer mitochondrial membrane 7 homolog (yeast)', other_designations='translocase of outer membrane 7 kDa subunit homolog', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23838944, end_position_on_the_genomic_accession=23844145, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66170, CurrentID=0, Status='live', Symbol='Chchd5', Aliases='1110027L01Rik, AW045710', description='coiled-coil-helix-coiled-coil-helix domain containing 5', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129129700, end_position_on_the_genomic_accession=129134112, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66171, CurrentID=0, Status='live', Symbol='Pgls', Aliases='1110030K05Rik, AI447866, Plgs', description='6-phosphogluconolactonase', other_designations='6PGL', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71592184, end_position_on_the_genomic_accession=71596267, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66172, CurrentID=0, Status='live', Symbol='Med11', Aliases='1110030J09Rik, AI465144, AW545069', description='mediator complex subunit 11', other_designations='mediator of RNA polymerase II transcription, subunit 11 homolog', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70451931, end_position_on_the_genomic_accession=70453727, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66174, CurrentID=0, Status='live', Symbol='Nudt14', Aliases='1110030M18Rik, UGPPase', description='nudix (nucleoside diphosphate linked moiety X)-type motif 14', other_designations='UDPG pyrophosphatase|nucleoside diphosphate-linked moiety X motif 14|nudix motif 14', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112934727, end_position_on_the_genomic_accession=112942167, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66175, CurrentID=0, Status='live', Symbol='Mustn1', Aliases='1110028G01Rik, Mustang', description='musculoskeletal, embryonic nuclear protein 1', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30879257, end_position_on_the_genomic_accession=30881610, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66176, CurrentID=0, Status='live', Symbol='Nat9', Aliases='1110028N05Rik', description='N-acetyltransferase 9 (GCN5-related, putative)', other_designations='novel acetyltransferase protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115182831, end_position_on_the_genomic_accession=115187886, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66177, CurrentID=0, Status='live', Symbol='Ubl5', Aliases='1110030M22Rik, beacon', description='ubiquitin-like 5', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20643318, end_position_on_the_genomic_accession=20646790, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66179, CurrentID=0, Status='live', Symbol='Ogfod3', Aliases='1110031I02Rik', description='2-oxoglutarate and iron-dependent oxygenase domain containing 3', other_designations='PKHD domain-containing transmembrane protein C17orf101 homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121177593, end_position_on_the_genomic_accession=121204648, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66180, CurrentID=0, Status='live', Symbol='P3h4', Aliases='1110036O03Rik, AI413214, Leprel4, Sc65', description='prolyl 3-hydroxylase family member 4 (non-enzymatic)', other_designations='SC65 synaptonemal complex protein (Sc65-pending)|leprecan-like 4|leprecan-like protein 4', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100408748, end_position_on_the_genomic_accession=100414819, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66181, CurrentID=0, Status='live', Symbol='Nop10', Aliases='1110036B12Rik, NOP10P, Nola3', description='NOP10 ribonucleoprotein', other_designations='H/ACA small nucleolar RNPs|NOP10 ribonucleoprotein homolog|nucleolar protein 10|nucleolar protein family A member 3|nucleolar protein family A, member 3|snoRNP protein NOP10', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112261926, end_position_on_the_genomic_accession=112262898, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66183, CurrentID=0, Status='live', Symbol='Sptssb', Aliases='1110032A04Rik, ADMP, Sssptb', description='serine palmitoyltransferase, small subunit B', other_designations='androgen down regulated in mouse prostate|small subunit of serine palmitoyltransferase B', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69819538, end_position_on_the_genomic_accession=69859894, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66184, CurrentID=0, Status='live', Symbol='Rps4l', Aliases='1110033J19Rik, AA407494, Rps4y2', description='ribosomal protein S4-like', other_designations='ribosomal protein S4, X-linked pseudogene|ribosomal protein S4, Y-linked 2', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148354656, end_position_on_the_genomic_accession=148355596, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66185, CurrentID=0, Status='live', Symbol='1110037F02Rik', Aliases='4930422M05Rik, Kiaa1429, mKIAA1429', description='RIKEN cDNA 1110037F02 gene', other_designations='', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11485948, end_position_on_the_genomic_accession=11551143, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  66190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66190, CurrentID=0, Status='live', Symbol='Acer3', Aliases='1110057L18Rik, 5430429L08Rik, AV015045, Phca', description='alkaline ceramidase 3', other_designations='aPHC|alkCDase 3|alkaline CDase 3|alkaline phytoceramidase|phytoceramidase, alkaline', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98213660, end_position_on_the_genomic_accession=98309546, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66191, CurrentID=0, Status='live', Symbol='Ier3ip1', Aliases='1110057H19Rik, AI644142, AL022842, AL022863, AV026606', description='immediate early response 3 interacting protein 1', other_designations='ier3ip1', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=76930027, end_position_on_the_genomic_accession=76941614, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66192, CurrentID=0, Status='live', Symbol='Lage3', Aliases='1110049G11Rik, Eso3, Itba2', description='L antigen family, member 3', other_designations='ITBA2 protein homolog', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74352162, end_position_on_the_genomic_accession=74353618, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66193, CurrentID=0, Status='live', Symbol='Pithd1', Aliases='1110049F12Rik, AA408179, Trp26', description='PITH (C-terminal proteasome-interacting domain of thioredoxin-like) domain containing 1', other_designations='AD039|thioredoxin family Trp26', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135975602, end_position_on_the_genomic_accession=135987244, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66194, CurrentID=0, Status='live', Symbol='Pycrl', Aliases='1110058B13Rik, 2700073G24Rik', description='pyrroline-5-carboxylate reductase-like', other_designations='P5C reductase 3|P5CR 3|pyrroline-5-carboxylate reductase-like protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75916463, end_position_on_the_genomic_accession=75921560, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66196, CurrentID=0, Status='live', Symbol='Myo19', Aliases='1110055A02Rik, Myohd1', description='myosin XIX', other_designations='myosin head domain containing 1|myosin head domain-containing protein 1|myosin-XIX', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84879922, end_position_on_the_genomic_accession=84911131, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  66197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66197, CurrentID=0, Status='live', Symbol='Cks2', Aliases='1110038L14Rik, CKSHS2', description='CDC28 protein kinase regulatory subunit 2', other_designations='CKS-2', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51645232, end_position_on_the_genomic_accession=51650662, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66198, CurrentID=0, Status='live', Symbol='Them5', Aliases='1110007B02Rik, 1110038F21Rik', description='thioesterase superfamily member 5', other_designations='acyl-CoA thioesterase THEM5|acyl-coenzyme A thioesterase 15', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94342099, end_position_on_the_genomic_accession=94347352, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66199, CurrentID=0, Status='live', Symbol='Commd4', Aliases='1110039H05Rik', description='COMM domain containing 4', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57155035, end_position_on_the_genomic_accession=57158349, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66200, CurrentID=0, Status='live', Symbol='Commd6', Aliases='1110059J08Rik, 1700063H17Rik', description='COMM domain containing 6', other_designations='', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=101633766, end_position_on_the_genomic_accession=101640471, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66201, CurrentID=0, Status='live', Symbol='Vta1', Aliases='1110001D18Rik, 1110059P08Rik, AU040813, C85340, LIP5, SBP1', description='Vps20-associated 1 homolog (S. cerevisiae)', other_designations='SKD1 binding protein 1|SKD1-binding protein 1|lysosomal trafficking regulator interacting protein-5', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=14654760, end_position_on_the_genomic_accession=14705527, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66202, CurrentID=0, Status='live', Symbol='1110059G10Rik', Aliases='', description='RIKEN cDNA 1110059G10 gene', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122945089, end_position_on_the_genomic_accession=122951000, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  229487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229487, CurrentID=0, Status='live', Symbol='Gatb', Aliases='9430026F02Rik, Pet112, Pet112l', description='glutamyl-tRNA(Gln) amidotransferase, subunit B', other_designations='PET112 homolog|cytochrome c oxidase assembly factor PET112 homolog|cytochrome oxidase assembly factor PET112 homolog|glu-ADT subunit B|glu-AdT subunit B|probable glutamyl-tRNA(Gln) amidotransferase subunit B, mitochondrial', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=85574102, end_position_on_the_genomic_accession=85655660, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  66204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66204, CurrentID=0, Status='live', Symbol='Acyp1', Aliases='1110039O14Rik, AI325944', description='acylphosphatase 1, erythrocyte (common) type', other_designations='acylphosphatase, organ-common type isozyme|acylphosphate phosphohydrolase 1', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85272398, end_position_on_the_genomic_accession=85288196, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66205, CurrentID=0, Status='live', Symbol='Cd302', Aliases='1110055L24Rik, AI159627', description='CD302 antigen', other_designations='C-type lectin domain family 13 member A|type I transmembrane C-type lectin receptor DCL-1', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=60251993, end_position_on_the_genomic_accession=60284484, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66206, CurrentID=0, Status='live', Symbol='1110059E24Rik', Aliases='', description='RIKEN cDNA 1110059E24 gene', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=21597313, end_position_on_the_genomic_accession=21652791, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66208, CurrentID=0, Status='live', Symbol='Nenf', Aliases='1110060M21Rik, SCIRP10, Spuf', description='neuron derived neurotrophic factor', other_designations='SCIRP10-related protein|Spinal cord injury related protein 10|neuron-derived neurotrophic factor|secreted protein of unknown function', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191306797, end_position_on_the_genomic_accession=191318118, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66209, CurrentID=0, Status='live', Symbol='Inip', Aliases='1110054O05Rik, 2610312O17Rik, AA399876, Ssbip1', description='INTS3 and NABP interacting protein', other_designations='INTS3- and NABP-interacting protein|SOSS-C|SSB-interacting protein 1|sensor of single-strand DNA complex subunit C|sensor of ssDNA subunit C|single-stranded DNA-binding protein-interacting protein 1', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59769642, end_position_on_the_genomic_accession=59783855, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66211, CurrentID=0, Status='live', Symbol='Rpl3l', Aliases='1110057H16Rik', description='ribosomal protein L3-like', other_designations='ribosomal protein L3 pseudogene', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24727829, end_position_on_the_genomic_accession=24736149, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  66212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66212, CurrentID=0, Status='live', Symbol='Sec61b', Aliases='1190006C12Rik, AI326121, AW122942', description='Sec61 beta subunit', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=47474661, end_position_on_the_genomic_accession=47483233, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66213, CurrentID=0, Status='live', Symbol='Med7', Aliases='1110063B05Rik, Crsp33, Crsp9', description='mediator complex subunit 7', other_designations='CRSP complex subunit 9|cofactor required for Sp1 transcriptional activation subunit 9|cofactor required for Sp1 transcriptional activation, subunit 9, 33kDa', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46436933, end_position_on_the_genomic_accession=46442721, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66214, CurrentID=0, Status='live', Symbol='Rgcc', Aliases='1190002H23Rik, Rgc-32, Rgc32', description='regulator of cell cycle', other_designations='response gene to complement 32 protein', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79288750, end_position_on_the_genomic_accession=79301635, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66218, CurrentID=0, Status='live', Symbol='Ndufb9', Aliases='1190008J14Rik', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 9', other_designations='CI-B22|NADH-ubiquinone oxidoreductase B22 subunit|complex I-B22', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58933810, end_position_on_the_genomic_accession=58939489, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66220, CurrentID=0, Status='live', Symbol='Zdhhc12', Aliases='1190004A01Rik', description='zinc finger, DHHC domain containing 12', other_designations='DHHC-12|zinc finger DHHC domain-containing protein 12', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30090944, end_position_on_the_genomic_accession=30093635, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  328365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328365, CurrentID=0, Status='live', Symbol='Zmiz1', Aliases='BC065120, E330020C23, ENSMUSG00000072684, Gm10397, I920194n01, Rai17, Zimp10', description='zinc finger, MIZ-type containing 1', other_designations='PIAS-like protein Zimp10|retinoic acid induced 17|retinoic acid-induced protein 17', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25458865, end_position_on_the_genomic_accession=25666747, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  66222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66222, CurrentID=0, Status='live', Symbol='Serpinb1a', Aliases='1190005M04Rik, AI325983, EI, EIA, ELANH2, LEI, M/NEI, MNEI, PI2', description='serine (or cysteine) peptidase inhibitor, clade B, member 1a', other_designations='ovalbumin|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 1a|serine (or cysteine) proteinase inhibitor, clade B, member 1a|serine protease inhibitor EIA|serpin B1a', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=32842092, end_position_on_the_genomic_accession=32851185, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66223, CurrentID=0, Status='live', Symbol='Mrpl35', Aliases='1110066C01Rik', description='mitochondrial ribosomal protein L35', other_designations='L35mt|MRP-L35', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71812997, end_position_on_the_genomic_accession=71823784, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66225, CurrentID=0, Status='live', Symbol='Llph', Aliases='1190005P17Rik', description='LLP homolog, long-term synaptic facilitation (Aplysia)', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=120227060, end_position_on_the_genomic_accession=120232071, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66226, CurrentID=0, Status='live', Symbol='Trappc2', Aliases='1110066L09Rik, 1810064C02Rik, AW496358, MIP-2A, Sedl, TRS20', description='trafficking protein particle complex 2', other_designations='sedlin|spondyloepiphyseal dysplasia tarda protein|spondyloepiphyseal dysplasia, late', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=166440825, end_position_on_the_genomic_accession=166452543, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66229, CurrentID=0, Status='live', Symbol='Rpl7l1', Aliases='1500016H10Rik', description='ribosomal protein L7-like 1', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46773907, end_position_on_the_genomic_accession=46782656, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66230, CurrentID=0, Status='live', Symbol='Mrps28', Aliases='1500012D08Rik, AW061224', description='mitochondrial ribosomal protein S28', other_designations='MRP-S28|S28mt', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=8802146, end_position_on_the_genomic_accession=8923857, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66231, CurrentID=0, Status='live', Symbol='Thoc7', Aliases='Nif3l1bp1', description='THO complex 7 homolog (Drosophila)', other_designations='ngg1-interacting factor 3-like protein 1-binding protein 1', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=13949012, end_position_on_the_genomic_accession=13962042, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66233, CurrentID=0, Status='live', Symbol='Dmap1', Aliases='1500016M21Rik, DNMAP1, DNMTAP1, Dmtap1, mKIAA1425', description='DNA methyltransferase 1-associated protein 1', other_designations='DNMT1 associated protein-1|DNMT1-associated protein 1|MAT1-mediated transcriptional repressor', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117674681, end_position_on_the_genomic_accession=117682321, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66234, CurrentID=0, Status='live', Symbol='Msmo1', Aliases='1500001G16Rik, C78600, DESP4, ERG25, Sc4mol', description='methylsterol monoxygenase 1', other_designations='C-4 methylsterol oxidase|sterol-C4-methyl oxidase-like', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=64718145, end_position_on_the_genomic_accession=64733578, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66235, CurrentID=0, Status='live', Symbol='Eif1ax', Aliases='1500010B24Rik, AI426898, Eif1ay', description='eukaryotic translation initiation factor 1A, X-linked', other_designations='eIF-1A X isoform|eIF-4C|eukaryotic translation initiation factor 1A, Y-linked|eukaryotic translation initiation factor 4C', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=159372195, end_position_on_the_genomic_accession=159385699, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66237, CurrentID=0, Status='live', Symbol='Atp6v1g2', Aliases='1500002D01Rik, Atp6g2, NG38, VAG2', description='ATPase, H+ transporting, lysosomal V1 subunit G2', other_designations='ATPase, H+ transporting, V1 subunit G|ATPase, H+ transporting, lysosomal (vacuolar proton pump)|ATPase, H+ transporting, lysosomal 13kD, V1 subunit G|V-ATPase 13 kDa subunit 2|V-ATPase subunit G 2|lysosomal 13kDa|vacuolar proton pump subunit G 2', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35235166, end_position_on_the_genomic_accession=35238767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66240, CurrentID=0, Status='live', Symbol='Kcne1l', Aliases='1500015C14Rik, Kcne5, Mink', description='potassium voltage-gated channel, Isk-related family, member 1-like, pseudogene', other_designations='minK-like protein|potassium voltage-gated channel subfamily E member 1-like protein', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=142304753, end_position_on_the_genomic_accession=142306198, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66241, CurrentID=0, Status='live', Symbol='Tmem9', Aliases='1500015G18Rik, AW545782', description='transmembrane protein 9', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136008219, end_position_on_the_genomic_accession=136035030, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66242, CurrentID=0, Status='live', Symbol='Mrps16', Aliases='1500011E11Rik, Mprs16', description='mitochondrial ribosomal protein S16', other_designations='MRP-S16|S16mt', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20391231, end_position_on_the_genomic_accession=20393555, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66244, CurrentID=0, Status='live', Symbol='Nemf', Aliases='1500011I12Rik, 4933405E14Rik, Sdccag1', description='nuclear export mediator factor', other_designations='serologically defined colon cancer antigen 1 homolog', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69311543, end_position_on_the_genomic_accession=69357189, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  66245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66245, CurrentID=0, Status='live', Symbol='Hspbp1', Aliases='1500019G21Rik', description='HSPA (heat shock 70kDa) binding protein, cytoplasmic cochaperone 1', other_designations='SPA ) binding protein, cytoplasmic cochaperone 1|heat shock protein-binding protein 1|hsp70-interacting protein 1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4660515, end_position_on_the_genomic_accession=4685121, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66246, CurrentID=0, Status='live', Symbol='Osgep', Aliases='1500019L24Rik, GCPL-1, PRSMG1', description='O-sialoglycoprotein endopeptidase', other_designations='N6-L-threonylcarbamoyladenine synthase|mOsgep|probable O-sialoglycoprotein endopeptidase|probable tRNA threonylcarbamoyladenosine biosynthesis protein Osgep|t(6)A synthase|t(6)A37 threonylcarbamoyladenosine biosynthesis protein Osgep|tRNA threonylcarbamoyladenosine biosynthesis protein Osgep', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50915374, end_position_on_the_genomic_accession=50924893, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66248, CurrentID=0, Status='live', Symbol='Alg5', Aliases='1500026A19Rik, 2600005J22Rik, AV009521, AV170001, AW545564', description='asparagine-linked glycosylation 5 (dolichyl-phosphate beta-glucosyltransferase)', other_designations='Alg5 homolog|asparagine-linked glycosylation 5 homolog (yeast, dolichyl-phosphate beta-glucosyltransferase)|asparagine-linked glycosylation protein 5 homolog|dolP-glucosyltransferase', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54735539, end_position_on_the_genomic_accession=54749795, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66249, CurrentID=0, Status='live', Symbol='Pno1', Aliases='1810003N24Rik, AA407421, AU015468, AU019495, AU042323, Emi3, Imi3', description='partner of NOB1 homolog (S. cerevisiae)', other_designations='expressed during mesenchymal induction 3|induced during mesenchymal induction 3', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=17203198, end_position_on_the_genomic_accession=17211589, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66251, CurrentID=0, Status='live', Symbol='Arfgap3', Aliases='0610009H19Rik, 1810004P07Rik, 1810035F16Rik, 9130416J18Rik, Arfgap1', description='ADP-ribosylation factor GTPase activating protein 3', other_designations='ADP-ribosylation factor GTPase activating protein 1|ARF GAP 3', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83299739, end_position_on_the_genomic_accession=83350344, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  66253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66253, CurrentID=0, Status='live', Symbol='Aig1', Aliases='1500031O19Rik, AV064870, AW413422, CGI-103', description='androgen-induced 1', other_designations='AIG-1', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13647054, end_position_on_the_genomic_accession=13869036, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66254, CurrentID=0, Status='live', Symbol='Dimt1', Aliases='1500031M22Rik, AV111347l, Dimt1', description='DIM1 dimethyladenosine transferase 1-like (S. cerevisiae)', other_designations="18S rRNA (adenine(1779)-N(6)/adenine(1780)-N(6))-dimethyltransferase|18S rRNA dimethylase|DIM1 dimethyladenosine transferase 1 homolog|S-adenosylmethionine-6-N',N'-adenosyl(rRNA) dimethyltransferase|probable 18S rRNA (adenine(1779)-N(6)/adenine(1780)-N(6))-dimethyltransferase|probable 18S rRNA dimethylase|probable S-adenosylmethionine-6-N',N'-adenosyl(rRNA) dimethyltransferase", map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=106947043, end_position_on_the_genomic_accession=106960225, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  66255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66255, CurrentID=0, Status='live', Symbol='Hsbp1l1', Aliases='1810005K13Rik', description='heat shock factor binding protein 1-like 1', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80229571, end_position_on_the_genomic_accession=80250103, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66256, CurrentID=0, Status='live', Symbol='Ssr2', Aliases='1500032E05Rik, AI315033, AU020133, TLAP, TRAPB, TRAPbeta', description='signal sequence receptor, beta', other_designations='SSR-beta|TRAP-beta|signal sequence receptor subunit beta', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88579584, end_position_on_the_genomic_accession=88588420, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66257, CurrentID=0, Status='live', Symbol='Nicn1', Aliases='1500032A17Rik', description='nicolin 1', other_designations='PGs5|p24|tubulin polyglutamylase complex subunit 5', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108290457, end_position_on_the_genomic_accession=108296496, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66258, CurrentID=0, Status='live', Symbol='Mrps17', Aliases='1810006P02Rik, Mprs17', description='mitochondrial ribosomal protein S17', other_designations='MRP-S17|S17mt', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129715441, end_position_on_the_genomic_accession=129718691, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66259, CurrentID=0, Status='live', Symbol='Camk2n1', Aliases='1810006K23Rik, AI848599, AI891706', description='calcium/calmodulin-dependent protein kinase II inhibitor 1', other_designations='CaMKII inhibitor protein alpha|CaMKIIN-alpha|CaMKIINalpha|calcium/calmodulin-dependent protein kinase II inhibitor alpha|mCaMKIINalpha', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138455148, end_position_on_the_genomic_accession=138460126, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66260, CurrentID=0, Status='live', Symbol='Tmem54', Aliases='1810017F10Rik, AI507302', description='transmembrane protein 54', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129105515, end_position_on_the_genomic_accession=129111626, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66261, CurrentID=0, Status='live', Symbol='Tm4sf20', Aliases='1810018L02Rik', description='transmembrane 4 L six family member 20', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=82756649, end_position_on_the_genomic_accession=82768456, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66262, CurrentID=0, Status='live', Symbol='Ing5', Aliases='1700001C14Rik, 1700027H23Rik, 1810018M11Rik, AI225768', description='inhibitor of growth family, member 5', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93802051, end_position_on_the_genomic_accession=93822101, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66264, CurrentID=0, Status='live', Symbol='Ccdc28b', Aliases='1110002E08Rik, 1810010N17Rik', description='coiled coil domain containing 28B', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129619274, end_position_on_the_genomic_accession=129623908, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66266, CurrentID=0, Status='live', Symbol='Eapp', Aliases='1810011O16Rik', description='E2F-associated phosphoprotein', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=54670340, end_position_on_the_genomic_accession=54695896, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66268, CurrentID=0, Status='live', Symbol='Pigyl', Aliases='1810008A14Rik, PIG-Y, Pigy', description='phosphatidylinositol glycan anchor biosynthesis, class Y-like', other_designations='phosphatidylinositol-glycan biosynthesis class Y protein', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22156846, end_position_on_the_genomic_accession=22158355, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66269, CurrentID=0, Status='live', Symbol='Tmed6', Aliases='1810015P03Rik, 1810018I24Rik', description='transmembrane emp24 protein transport domain containing 6', other_designations='p24 family protein gamma-5|p24gamma5', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107061484, end_position_on_the_genomic_accession=107065644, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66270, CurrentID=0, Status='live', Symbol='Fam134b', Aliases='1810015C04Rik, AU015349', description='family with sequence similarity 134, member B', other_designations='protein FAM134B', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=25843298, end_position_on_the_genomic_accession=25973696, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  66271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66271, CurrentID=0, Status='live', Symbol='Tmem126a', Aliases='1810020E01Rik', description='transmembrane protein 126A', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90450712, end_position_on_the_genomic_accession=90457208, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66272, CurrentID=0, Status='live', Symbol='Cox16', Aliases='1810020G14Rik, 1810055I05Rik, BB388670', description='cytochrome c oxidase assembly protein 16', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81470704, end_position_on_the_genomic_accession=81485110, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66273, CurrentID=0, Status='live', Symbol='Aamdc', Aliases='1810020D17Rik, 1810037D19Rik, LI2', description='adipogenesis associated Mth938 domain containing', other_designations='UPF0366 protein C11orf67 homolog', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97550329, end_position_on_the_genomic_accession=97579505, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66274, CurrentID=0, Status='live', Symbol='Lyrm9', Aliases='1810012P15Rik', description='LYR motif containing 9', other_designations='UPF0631 protein C17orf108 homolog', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78826595, end_position_on_the_genomic_accession=78843899, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66276, CurrentID=0, Status='live', Symbol='1810009A15Rik', Aliases='', description='RIKEN cDNA 1810009A15 gene', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8888883, end_position_on_the_genomic_accession=8890759, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66277, CurrentID=0, Status='live', Symbol='Klf15', Aliases='1810013I09Rik, AV048136, AW494632, CKLF, KKLF', description='Kruppel-like factor 15', other_designations='cardiovascular Krueppel-like factor', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90462626, end_position_on_the_genomic_accession=90479170, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66278, CurrentID=0, Status='live', Symbol='Smim20', Aliases='1110067B18Rik, 1810013D10Rik, N28078', description='small integral membrane protein 20', other_designations='', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=53267106, end_position_on_the_genomic_accession=53278540, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66279, CurrentID=0, Status='live', Symbol='Tmem218', Aliases='1810021J13Rik, AV004931', description='transmembrane protein 218', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37208223, end_position_on_the_genomic_accession=37223872, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  328424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328424, CurrentID=0, Status='live', Symbol='Kcnrg', Aliases='Clld4, E030012H22Rik, Gm745', description='potassium channel regulator', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61607457, end_position_on_the_genomic_accession=61612833, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66282, CurrentID=0, Status='live', Symbol='Tma16', Aliases='1810029B16Rik', description='translation machinery associated 16 homolog (S. cerevisiae)', other_designations='UPF0534 protein C4orf43 homolog', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=66473118, end_position_on_the_genomic_accession=66486548, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66283, CurrentID=0, Status='live', Symbol='Gkn1', Aliases='2200002K21Rik, AMP-18, BRICD1, Fov', description='gastrokine 1', other_designations='18 kDa antrum mucosa protein|foveolin|protein CA11 homolog', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87345653, end_position_on_the_genomic_accession=87350915, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66284, CurrentID=0, Status='live', Symbol='Gkn2', Aliases='1810036H07Rik, Bricd1b', description='gastrokine 2', other_designations='blottin', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87373365, end_position_on_the_genomic_accession=87379494, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66286, CurrentID=0, Status='live', Symbol='Sec11c', Aliases='1810029G24Rik, Sec11l3', description='SEC11 homolog C, signal peptidase complex subunit', other_designations='SEC11 homolog C|SEC11-like protein 3|SPC21|SPase 21 kDa subunit|Sec11-like 3|microsomal signal peptidase 21 kDa subunit', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65800499, end_position_on_the_genomic_accession=65817660, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66289, CurrentID=0, Status='live', Symbol='Mptx1', Aliases='1810030J14Rik, AV054416, Mptx', description='mucosal pentraxin 1', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174330518, end_position_on_the_genomic_accession=174332877, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66290, CurrentID=0, Status='live', Symbol='Atp6v1g1', Aliases='1810024D14Rik, AA960677, ATP6J, Atp6g1, VAG1, Vma10', description='ATPase, H+ transporting, lysosomal V1 subunit G1', other_designations='ATPase, H+ transporting, V1 subunit G|ATPase, H+ transporting, lysosomal (vacuolar proton pump)|ATPase, H+ transporting, lysosomal 13kD, V1 subunit G|V-ATPase 13 kDa subunit 1|V-ATPase subunit G 1|lysosomal 13kDa|vacuolar proton pump subunit G 1', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63544765, end_position_on_the_genomic_accession=63550701, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66291, CurrentID=0, Status='live', Symbol='Smim8', Aliases='1810030N24Rik, 2810406B13Rik', description='small integral membrane protein 8', other_designations='UPF0708 protein C6orf162 homolog', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34768664, end_position_on_the_genomic_accession=34778396, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66292, CurrentID=0, Status='live', Symbol='Mrps21', Aliases='1810031B19Rik', description='mitochondrial ribosomal protein S21', other_designations='MRP-S21|S21mt', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95862646, end_position_on_the_genomic_accession=95871604, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66293, CurrentID=0, Status='live', Symbol='1810032O08Rik', Aliases='', description='RIKEN cDNA 1810032O08 gene', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116671660, end_position_on_the_genomic_accession=116675798, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66294, CurrentID=0, Status='live', Symbol='Fam3a', Aliases='1810037C20Rik', description='family with sequence similarity 3, member A', other_designations='', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74384719, end_position_on_the_genomic_accession=74393280, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  328440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328440, CurrentID=0, Status='live', Symbol='Npm2', Aliases='', description='nucleophosmin/nucleoplasmin 2', other_designations='nucleoplasmin 2', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70647296, end_position_on_the_genomic_accession=70660753, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66297, CurrentID=0, Status='live', Symbol='Pantr1', Aliases='2610017I09Rik, Pou3f3os, linc-Brn1a', description='POU domain, class 3, transcription factor 3 adjacent noncoding transcript 1', other_designations='POU domain, class 3, transcription factor 3, opposite strand', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=42648200, end_position_on_the_genomic_accession=42694825, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66298, CurrentID=0, Status='live', Symbol='Defa21', Aliases='2010016B13Rik, Defcr21', description='defensin, alpha, 21', other_designations='defensin-related cryptdin-21', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21025545, end_position_on_the_genomic_accession=21026517, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66300, CurrentID=0, Status='live', Symbol='Inafm1', Aliases='2610014I16Rik, Prr24', description='InaF motif containing 1', other_designations='proline rich 24', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16272013, end_position_on_the_genomic_accession=16273692, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66302, CurrentID=0, Status='live', Symbol='Rmdn1', Aliases='Fam82b, RMD-1', description='regulator of microtubule dynamics 1', other_designations='family with sequence similarity 82, member B|microtubule-associated protein', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=19574862, end_position_on_the_genomic_accession=19606932, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  229504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229504, CurrentID=0, Status='live', Symbol='Isg20l2', Aliases='4930429N24, C80587', description='interferon stimulated exonuclease gene 20-like 2', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87930314, end_position_on_the_genomic_accession=87940686, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66306, CurrentID=0, Status='live', Symbol='Fam53c', Aliases='2810012G03Rik, Gm21179', description='family with sequence similarity 53, member C', other_designations='', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34758830, end_position_on_the_genomic_accession=34773760, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66307, CurrentID=0, Status='live', Symbol='Isoc1', Aliases='2610034N03Rik', description='isochorismatase domain containing 1', other_designations='', map_location='18 D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=58659482, end_position_on_the_genomic_accession=58679570, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66308, CurrentID=0, Status='live', Symbol='Mplkip', Aliases='2810021B07Rik, C330007M08Rik, Ttdn1', description='M-phase specific PLK1 intereacting protein', other_designations='TTD non-photosensitive 1 protein homolog', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=17695413, end_position_on_the_genomic_accession=17699105, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66309, CurrentID=0, Status='live', Symbol='Tmem128', Aliases='2810021O14Rik', description='transmembrane protein 128', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38260190, end_position_on_the_genomic_accession=38269618, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66310, CurrentID=0, Status='live', Symbol='Dpy30', Aliases='2810410M20Rik, C87842', description='dpy-30, histone methyltransferase complex regulatory subunit', other_designations='dosage compensation-related protein DPY30|dpy-30 homolog|dpy-30-like protein|dpy-30L', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74299474, end_position_on_the_genomic_accession=74323944, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66311, CurrentID=0, Status='live', Symbol='Cenpw', Aliases='2610036L11Rik, Cug2', description='centromere protein W', other_designations='CENP-W|cancer-up-regulated gene 2 protein', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=30196009, end_position_on_the_genomic_accession=30200600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66313, CurrentID=0, Status='live', Symbol='Smurf2', Aliases='2810411E22Rik, AI558114, AI649275', description='SMAD specific E3 ubiquitin protein ligase 2', other_designations='SMAD ubiquitination regulatory factor 2|SMAD-specific E3 ubiquitin-protein ligase 2', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106820064, end_position_on_the_genomic_accession=106920715, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  66314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66314, CurrentID=0, Status='live', Symbol='Tpd52l2', Aliases='2810411G23Rik, AU016537, C86069, D54', description='tumor protein D52-like 2', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181497104, end_position_on_the_genomic_accession=181517962, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66315, CurrentID=0, Status='live', Symbol='Senp7', Aliases='2410152H17Rik, 2810413I22Rik, 2900036C23Rik, 6030449K19Rik, AI790676', description='SUMO1/sentrin specific peptidase 7', other_designations='SUMO-1-specific protease 2|SUMO1/sentrin specific protease 7|sentrin/SUMO-specific protease SENP7', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56072059, end_position_on_the_genomic_accession=56190011, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  66317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66317, CurrentID=0, Status='live', Symbol='Wdr61', Aliases='2700038L12Rik, 2810418I05Rik, REC14', description='WD repeat domain 61', other_designations='meiotic recombination protein REC14', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54717153, end_position_on_the_genomic_accession=54734549, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66320, CurrentID=0, Status='live', Symbol='Tmem208', Aliases='1700006C06Rik, 2610030K20Rik, Hspc171', description='transmembrane protein 208', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105326364, end_position_on_the_genomic_accession=105335485, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66322, CurrentID=0, Status='live', Symbol='1700011A15Rik', Aliases='Krt88', description='RIKEN cDNA 1700011A15 gene', other_designations='', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101447738, end_position_on_the_genomic_accession=101453910, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66323, CurrentID=0, Status='live', Symbol='1700001K19Rik', Aliases='', description='RIKEN cDNA 1700001K19 gene', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110667688, end_position_on_the_genomic_accession=110682651, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66326, CurrentID=0, Status='live', Symbol='Dnajc5b', Aliases='1700008A05Rik', description='DnaJ heat shock protein family (Hsp40) member C5 beta', other_designations='CSP-beta|DnaJ (Hsp40) homolog, subfamily C, member 5 beta|beta cysteine string protein|beta-CSP|cysteine string protein beta', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19508591, end_position_on_the_genomic_accession=19610862, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66328, CurrentID=0, Status='live', Symbol='Scp2d1', Aliases='1700010M22Rik', description='SCP2 sterol-binding domain containing 1', other_designations='likely ortholog of H. sapiens chromosome 20 open reading frame 79 (C20orf79)', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144823666, end_position_on_the_genomic_accession=144824415, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66329, CurrentID=0, Status='live', Symbol='Susd3', Aliases='1700017I11Rik, 2810440J20Rik', description='sushi domain containing 3', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49230827, end_position_on_the_genomic_accession=49248706, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66330, CurrentID=0, Status='live', Symbol='1700020L24Rik', Aliases='', description='RIKEN cDNA 1700020L24 gene', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83437694, end_position_on_the_genomic_accession=83441235, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66333, CurrentID=0, Status='live', Symbol='Aqp11', Aliases='1700015P13Rik, AI648923, sjds', description='aquaporin 11', other_designations='AQP-11|CG12251 gene product', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97725490, end_position_on_the_genomic_accession=97738514, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66335, CurrentID=0, Status='live', Symbol='Atp6v1c1', Aliases='1700025B18Rik, U13839', description='ATPase, H+ transporting, lysosomal V1 subunit C1', other_designations='ATPase, H+ transporting, V1 subunit C|V-ATPase subunit C 1|vacuolar proton pump subunit C 1', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=38661904, end_position_on_the_genomic_accession=38692444, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66336, CurrentID=0, Status='live', Symbol='Cenpp', Aliases='1700022C02Rik, 4921518G09Rik', description='centromere protein P', other_designations='CENP-P', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49464025, end_position_on_the_genomic_accession=49652879, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66337, CurrentID=0, Status='live', Symbol='Fam229b', Aliases='1700025K23Rik', description='family with sequence similarity 229, member B', other_designations='UPF0731 protein C6orf225 homolog', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39118804, end_position_on_the_genomic_accession=39133943, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66338, CurrentID=0, Status='live', Symbol='Cdrt4', Aliases='1700019I23Rik', description='CMT1A duplicated region transcript 4', other_designations='', map_location='11 B2-B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62951193, end_position_on_the_genomic_accession=62993095, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  328483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328483, CurrentID=0, Status='live', Symbol='Trav6-3', Aliases='Gm13948, Gm193, Gm4, OTTMUSG00000015160, TCR', description='T cell receptor alpha variable 6-3', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53428761, end_position_on_the_genomic_accession=53430377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  66340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66340, CurrentID=0, Status='live', Symbol='Psenen', Aliases='1700023M09Rik, Pen-2', description='presenilin enhancer gamma secretase subunit', other_designations='presenilin enhancer 2 homolog|presenilin enhancer protein 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30561865, end_position_on_the_genomic_accession=30563184, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66341, CurrentID=0, Status='live', Symbol='Eid3', Aliases='1700027M21Rik, NS4EB', description='EP300 interacting inhibitor of differentiation 3', other_designations='EID-1-like inhibitor of differentiation 3|non-SMC element 4 homolog B|non-structural maintenance of chromosomes element 4 homolog B', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82866626, end_position_on_the_genomic_accession=82867930, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66343, CurrentID=0, Status='live', Symbol='Tmem177', Aliases='2300008B03Rik', description='transmembrane protein 177', other_designations='', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=119907890, end_position_on_the_genomic_accession=119913168, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66344, CurrentID=0, Status='live', Symbol='Lce3b', Aliases='2310007F04Rik, Sprrl6A', description='late cornified envelope 3B', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92932979, end_position_on_the_genomic_accession=92934096, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66346, CurrentID=0, Status='live', Symbol='1700029P11Rik', Aliases='', description='RIKEN cDNA 1700029P11 gene', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81980540, end_position_on_the_genomic_accession=81981563, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66349, CurrentID=0, Status='live', Symbol='Atp5sl', Aliases='2310004L02Rik, C030044E10Rik', description='ATP5S-like', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25619414, end_position_on_the_genomic_accession=25625550, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66350, CurrentID=0, Status='live', Symbol='Pla2g12a', Aliases='2310004B05Rik, GXII, GXII-1, GXII-1-PLA2, Pla2g12, Rossy, mGXII-1', description='phospholipase A2, group XIIA', other_designations='group XII-1 phospholipase A2|group XIIA secreted phospholipase A2|phosphatidylcholine 2-acylhydrolase 12A', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129878606, end_position_on_the_genomic_accession=129895825, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66352, CurrentID=0, Status='live', Symbol='Blzf1', Aliases='1700030G05Rikl, Golgin-45, Jem-1, Blzf1', description='basic leucine zipper nuclear factor 1', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164289800, end_position_on_the_genomic_accession=164307484, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66353, CurrentID=0, Status='live', Symbol='Riiad1', Aliases='2310007A19Rik, Ncrna00166', description='regulatory subunit of type II PKA R-subunit (RIIa) domain containing 1', other_designations='RIIa domain-containing protein C1orf230 homolog', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94464984, end_position_on_the_genomic_accession=94475078, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66354, CurrentID=0, Status='live', Symbol='Snw1', Aliases='2310008B08Rik, AW048543, NCoA-62, SKIP, Skiip', description='SNW domain containing 1', other_designations='SKI interacting protein|nuclear protein SkiP|ski-interacting protein', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87449910, end_position_on_the_genomic_accession=87472299, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66355, CurrentID=0, Status='live', Symbol='Gmpr', Aliases='2310004P21Rik, AV028449, GMPR 1', description='guanosine monophosphate reductase', other_designations="guanosine 5'-monophosphate oxidoreductase 1|guanosine monophosphate reductase 1", map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=45507429, end_position_on_the_genomic_accession=45546386, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66356, CurrentID=0, Status='live', Symbol='Knop1', Aliases='2310008H09Rik, Tsg118', description='lysine rich nucleolar protein 1', other_designations='protein C16orf88 homolog|testis-specific gene 118 protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118842217, end_position_on_the_genomic_accession=118855998, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66357, CurrentID=0, Status='live', Symbol='Ostc', Aliases='2310008M10Rik, 5730557H03Rik', description='oligosaccharyltransferase complex subunit', other_designations='', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=130695917, end_position_on_the_genomic_accession=130709444, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66358, CurrentID=0, Status='live', Symbol='Adprm', Aliases='2310004I24Rik, MDS006', description='ADP-ribose/CDP-alcohol diphosphatase, manganese dependent', other_designations='ADP-ribose/CDP-alcohol pyrophosphatase|ADPRibase-Mn|CDP-choline phosphohydrolase', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67037880, end_position_on_the_genomic_accession=67052623, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66359, CurrentID=0, Status='live', Symbol='Cox20', Aliases='2310005N03Rik, Fam36a', description='COX20 Cox2 chaperone', other_designations='family with sequence similarity 36, member A|protein FAM36A', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=178319153, end_position_on_the_genomic_accession=178322693, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66361, CurrentID=0, Status='live', Symbol='Zfand1', Aliases='2310008M20Rik, AW048890', description='zinc finger, AN1-type domain 1', other_designations='', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10339953, end_position_on_the_genomic_accession=10351338, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66362, CurrentID=0, Status='live', Symbol='Exosc3', Aliases='2310005D06Rik, AI593501, Rrp40', description='exosome component 3', other_designations='exosome complex exonuclease RRP40|ribosomal RNA-processing protein 40', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45316629, end_position_on_the_genomic_accession=45320603, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66365, CurrentID=0, Status='live', Symbol='Ccdc90b', Aliases='2310015N07Rik, C86558', description='coiled-coil domain containing 90B', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=92561149, end_position_on_the_genomic_accession=92582294, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66366, CurrentID=0, Status='live', Symbol='Ergic3', Aliases='2310015B14Rik, AV318804, CGI-54, D2Ucla1, NY-BR-84, Sdbcag84', description='ERGIC and golgi 3', other_designations='serologically defined breast cancer antigen 84|serologically defined breast cancer antigen NY-BR-84 homolog', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156008033, end_position_on_the_genomic_accession=156018279, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  66367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66367, CurrentID=0, Status='live', Symbol='2310022A10Rik', Aliases='', description='RIKEN cDNA 2310022A10 gene', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27553260, end_position_on_the_genomic_accession=27582099, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66368, CurrentID=0, Status='live', Symbol='Rtca', Aliases='2310009A18Rik, AI450277, RPC, Rtcd1', description="RNA 3'-terminal phosphate cyclase", other_designations="RNA cyclase|RNA terminal phosphate cyclase domain 1|RNA terminal phosphate cyclase domain-containing protein 1|RNA-3'-phosphate cyclase", map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116488963, end_position_on_the_genomic_accession=116508175, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66369, CurrentID=0, Status='live', Symbol='Dus2', Aliases='2310016K04Rikl, Dus2', description='dihydrouridine synthase 2', other_designations='', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106011478, end_position_on_the_genomic_accession=106053840, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  66371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66371, CurrentID=0, Status='live', Symbol='Chmp4c', Aliases='2010012P02Rik, 2210015K02Rik, 2310010I16Rik, Shax3, Snf7-3', description='charged multivesicular body protein 4C', other_designations='Snf7 homologue associated with Alix 3|chromatin modifying protein 4C|chromatin-modifying protein 4c', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10366973, end_position_on_the_genomic_accession=10391005, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66373, CurrentID=0, Status='live', Symbol='Lsm5', Aliases='2010208O10Rik, 2310034K10Rik', description='LSM5 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM5 homolog, U6 small nuclear RNA associated', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56701057, end_position_on_the_genomic_accession=56704699, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66374, CurrentID=0, Status='live', Symbol='2310011J03Rik', Aliases='AI452186, AW545358', description='RIKEN cDNA 2310011J03 gene', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80318255, end_position_on_the_genomic_accession=80320548, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66375, CurrentID=0, Status='live', Symbol='Dhrs7', Aliases='2310016E22Rik, 5730564L20Rik, AW061210, Retdsr4, Retsdr4', description='dehydrogenase/reductase (SDR family) member 7', other_designations='retinal short-chain dehydrogenase/reductase 4', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72650353, end_position_on_the_genomic_accession=72664828, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66377, CurrentID=0, Status='live', Symbol='Ndufc1', Aliases='2310016K22Rik, KFYI', description='NADH dehydrogenase (ubiquinone) 1, subcomplex unknown, 1', other_designations='CI-KFYI|NADH-ubiquinone oxidoreductase KFYI subunit|complex I-KFYI', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51405479, end_position_on_the_genomic_accession=51408955, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66379, CurrentID=0, Status='live', Symbol='Cox14', Aliases='2310016M24Rik, AA959804', description='cytochrome c oxidase assembly protein 14', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99725618, end_position_on_the_genomic_accession=99728136, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66380, CurrentID=0, Status='live', Symbol='Krtap3-3', Aliases='2310020D06Rik, KAP3.3, KRTAP3.3', description='keratin associated protein 3-3', other_designations='keratin associated protein 3-2|keratin-associated protein 3-3|krtap3-2|krtap3-3', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99550132, end_position_on_the_genomic_accession=99550863, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66381, CurrentID=0, Status='live', Symbol='Rnf113a2', Aliases='2310020H19Rik, AI426758, Rnf113a', description='ring finger protein 113A2', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84417200, end_position_on_the_genomic_accession=84418578, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66383, CurrentID=0, Status='live', Symbol='Iscu', Aliases='2310020H20Rik, AA407971, Nifu, Nifun', description='iron-sulfur cluster assembly enzyme', other_designations='ISCU2|IscU iron-sulfur cluster scaffold homolog|NifU-like N-terminal domain containing|nifU-like N-terminal domain-containing protein|nifU-like protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113772812, end_position_on_the_genomic_accession=113778282, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66384, CurrentID=0, Status='live', Symbol='Srp19', Aliases='2310020D23Rik', description='signal recognition particle 19', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34331121, end_position_on_the_genomic_accession=34336599, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66385, CurrentID=0, Status='live', Symbol='Ppp1r7', Aliases='2310014J01Rik, SDS22', description='protein phosphatase 1, regulatory (inhibitor) subunit 7', other_designations='protein phosphatase 1 regulatory subunit 22|protein phosphatase-1 regulatory subunit 7', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93343645, end_position_on_the_genomic_accession=93367619, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66387, CurrentID=0, Status='live', Symbol='Nudt8', Aliases='2310039H17Rik', description='nudix (nucleoside diphosphate linked moiety X)-type motif 8', other_designations='nucleoside diphosphate-linked moiety X motif 8, mitochondrial|nudix motif 8|nudix-type motif 8', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4000580, end_position_on_the_genomic_accession=4002102, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66388, CurrentID=0, Status='live', Symbol='Cutc', Aliases='2310039I18Rik, AI326282, CGI-32', description='cutC copper transporter', other_designations='cutC copper transporter homolog', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43753023, end_position_on_the_genomic_accession=43768638, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66390, CurrentID=0, Status='live', Symbol='Slmo2', Aliases='2310042G06Rik, AA407013, Prelid3b', description='slowmo homolog 2 (Drosophila)', other_designations='likely orthologue of H. sapiens chromosome 20 open reading frame 45 (C20orf45)|protein slowmo homolog 2', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174465076, end_position_on_the_genomic_accession=174472991, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66392, CurrentID=0, Status='live', Symbol='Prl2b1', Aliases='2310047B08Rik, Ghd12, PLP-K, Plpk, Prlpk, mPLP-K', description='prolactin family 2, subfamily b, member 1', other_designations='PRL-like protein K|growth hormone d12|placental prolactin-like protein K|prolactin like protein k|prolactin-like protein K', map_location='13 12.43 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27383345, end_position_on_the_genomic_accession=27390846, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66394, CurrentID=0, Status='live', Symbol='Nosip', Aliases='2310061K06Rik, CGI-25', description='nitric oxide synthase interacting protein', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45062429, end_position_on_the_genomic_accession=45078503, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66395, CurrentID=0, Status='live', Symbol='Ahnak', Aliases='DY6', description='AHNAK nucleoprotein (desmoyokin)', other_designations='desmoyokin', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8989125, end_position_on_the_genomic_accession=9076935, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66396, CurrentID=0, Status='live', Symbol='Ccdc82', Aliases='2310043N13Rik', description='coiled-coil domain containing 82', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=13246979, end_position_on_the_genomic_accession=13292353, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66397, CurrentID=0, Status='live', Symbol='Sar1b', Aliases='2310075M17Rik, 2900019I22Rik, CMRD, Sara1b, Sara2, Sarb', description='SAR1 gene homolog B (S. cerevisiae)', other_designations='SAR1a gene homolog 2|SAR1b gene homolog', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51763663, end_position_on_the_genomic_accession=51791953, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66398, CurrentID=0, Status='live', Symbol='Commd5', Aliases='2310065H03Rik, AI854466, D15Ertd81e, Hcarg', description='COMM domain containing 5', other_designations='hypertension-related, calcium regulated', map_location='15 36.28 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76899941, end_position_on_the_genomic_accession=76901297, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66399, CurrentID=0, Status='live', Symbol='Tsfm', Aliases='2310050B20Rik, 9430024O13Rik, EF-TS, EF-Tsmt', description='Ts translation elongation factor, mitochondrial', other_designations='EF-Ts|EF-TsMt|translation elongation factor', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127022332, end_position_on_the_genomic_accession=127030814, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66400, CurrentID=0, Status='live', Symbol='Alkbh7', Aliases='2310045B01Rik, 2510008E23Rik, Abh7, Spata11', description='alkB homolog 7', other_designations='alkB, alkylation repair homolog 7|alkylated DNA repair protein alkB homolog 7|probable alpha-ketoglutarate-dependent dioxygenase ABH7|spermatogenesis associated 11', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56997339, end_position_on_the_genomic_accession=56999340, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66401, CurrentID=0, Status='live', Symbol='Nudt2', Aliases='2310051L06Rik, AA939917, APAH1', description='nudix (nucleoside diphosphate linked moiety X)-type motif 2', other_designations="ap4A hydrolase|ap4Aase|bis(5'-nucleosyl)-tetraphosphatase|diadenosine 5',5'''-P1,P4-tetraphosphate asymmetrical hydrolase|diadenosine tetraphosphatase|nucleoside diphosphate-linked moiety X motif 2|nudix motif 2", map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41465080, end_position_on_the_genomic_accession=41480926, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66402, CurrentID=0, Status='live', Symbol='Sln', Aliases='2310045A07Rik', description='sarcolipin', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53850251, end_position_on_the_genomic_accession=53853849, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66403, CurrentID=0, Status='live', Symbol='Asf1a', Aliases='2310079C17Rik, C79052', description='anti-silencing function 1A histone chaperone', other_designations='ASF1 anti-silencing function 1 homolog A|anti-silencing function protein 1 homolog A', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=53596961, end_position_on_the_genomic_accession=53609225, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66404, CurrentID=0, Status='live', Symbol='Rtfdc1', Aliases='1700067C04Rik, 2410001C21Rik, 5730427M17Rik, AA589417', description='replication termination factor 2 domain containing 1', other_designations='UPF0549 protein C20orf43 homolog|likely orthologue of H. sapiens chromosome 20 open reading frame 43 (C20orf43)|replication termination factor 2 domain-containing protein 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172440578, end_position_on_the_genomic_accession=172469899, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66405, CurrentID=0, Status='live', Symbol='Mcts2', Aliases='2400002F11Rik, MCT-2', description='malignant T cell amplified sequence 2', other_designations='multiple copies T-cell malignancies 2', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152687148, end_position_on_the_genomic_accession=152687946, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66406, CurrentID=0, Status='live', Symbol='Sac3d1', Aliases='2410004C24Rik, AU015683, shd1', description='SAC3 domain containing 1', other_designations='SAC3 homology domain-containing protein 1|Sac3 homology domain 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6116004, end_position_on_the_genomic_accession=6118586, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66407, CurrentID=0, Status='live', Symbol='Mrps15', Aliases='1500003E24Rik, 2410002B11Rik, Mprs15', description='mitochondrial ribosomal protein S15', other_designations='MRP-S15|S15mt', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126046928, end_position_on_the_genomic_accession=126055536, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66408, CurrentID=0, Status='live', Symbol='Aptx', Aliases='2410016G21Rik, AA388047, FHA-HIT', description='aprataxin', other_designations='forkhead-associated domain histidine triad-like protein', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40682078, end_position_on_the_genomic_accession=40721667, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66409, CurrentID=0, Status='live', Symbol='Rsl1d1', Aliases='2410005K20Rik, C76439, pBK1', description='ribosomal L1 domain containing 1', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11193037, end_position_on_the_genomic_accession=11203292, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66410, CurrentID=0, Status='live', Symbol='Mterf3', Aliases='2410017I18Rik, Mterfd1', description='mitochondrial transcription termination factor 3', other_designations='MTERF domain containing 1|mTERF domain-containing protein 1, mitochondrial', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=66905456, end_position_on_the_genomic_accession=66933072, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66411, CurrentID=0, Status='live', Symbol='Tbcb', Aliases='2410007D12Rik, AU041393, CG22, CKAPI, Ckap1', description='tubulin folding cofactor B', other_designations='cytoskeleton-associated protein 1|cytoskeleton-associated protein CKAPI|tubulin-specific chaperone B', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30224130, end_position_on_the_genomic_accession=30232029, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66412, CurrentID=0, Status='live', Symbol='Arrdc4', Aliases='2410003C09Rik, AV216361', description='arrestin domain containing 4', other_designations='', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=68736994, end_position_on_the_genomic_accession=68749238, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66413, CurrentID=0, Status='live', Symbol='Psmd6', Aliases='2400006A19Rik', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 6', other_designations='26S proteasome regulatory subunit RPN7|26S proteasome regulatory subunit S10|p42A|proteasome, 26S, non-ATPase regulatory subunit 6', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=14112185, end_position_on_the_genomic_accession=14120904, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66414, CurrentID=0, Status='live', Symbol='Ndufa12', Aliases='2410011G03Rik, AW112974', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 12', other_designations='13kDa differentiation-associated protein|CI-B17.2|CIB17.2|NADH-ubiquinone oxidoreductase subunit B17.2|complex I-B17.2', map_location='10 F', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=94199009, end_position_on_the_genomic_accession=94220948, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66416, CurrentID=0, Status='live', Symbol='Ndufa7', Aliases='14.5kDa, 2400007M02Rik, CI-B14.5a', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 7 (B14.5a)', other_designations='NADH-ubiquinone oxidoreductase subunit B14.5a|complex I-B14.5a', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33824572, end_position_on_the_genomic_accession=33838316, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  328561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328561, CurrentID=0, Status='live', Symbol='Apol10b', Aliases='9130218O11Rik', description='apolipoprotein L 10B', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77584106, end_position_on_the_genomic_accession=77598542, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66419, CurrentID=0, Status='live', Symbol='Mrpl11', Aliases='2410001P07Rik, L11mt', description='mitochondrial ribosomal protein L11', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4962152, end_position_on_the_genomic_accession=4966995, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66420, CurrentID=0, Status='live', Symbol='Polr2e', Aliases='2410021N14Rik, 25kDa, AW208866, RPB5, XAP4', description='polymerase (RNA) II (DNA directed) polypeptide E', other_designations='DNA-directed RNA polymerase II subunit E|RNA polymerases I, II, and III subunit ABC1|RPB5 homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80035953, end_position_on_the_genomic_accession=80039659, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66421, CurrentID=0, Status='live', Symbol='2410004B18Rik', Aliases='4930572C24Rik', description='RIKEN cDNA 2410004B18 gene', other_designations='', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145938003, end_position_on_the_genomic_accession=145944829, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66422, CurrentID=0, Status='live', Symbol='Dctpp1', Aliases='2410015N17Rik, AI854235, RS21-C6', description='dCTP pyrophosphatase 1', other_designations='XTP3-transactivated gene A protein homolog|dCTPase 1|deoxycytidine-triphosphatase 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127256959, end_position_on_the_genomic_accession=127260667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66423, CurrentID=0, Status='live', Symbol='Coprs', Aliases='1700029I03Rik, 2410022L05Rik, AA409325, AI256813, C85432, Copr5', description='coordinator of PRMT5, differentiation stimulator', other_designations='cooperator of PRMT5', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13884787, end_position_on_the_genomic_accession=13890327, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66425, CurrentID=0, Status='live', Symbol='Pcp4l1', Aliases='1500036F01Rik', description='Purkinje cell protein 4-like 1', other_designations='PCP4-like protein 1', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171173264, end_position_on_the_genomic_accession=171196268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66427, CurrentID=0, Status='live', Symbol='Cyb5b', Aliases='1810044O22Rik, AU015618, Cyb5m', description='cytochrome b5 type B', other_designations='cytochrome b5 outer mitochondrial membrane isoform', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107150661, end_position_on_the_genomic_accession=107187470, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  328572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328572, CurrentID=0, Status='live', Symbol='Ep300', Aliases='A430090G16, A730011L11, KAT3B, p300, p300 HAT', description='E1A binding protein p300', other_designations='E1A-associated protein p300', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81586214, end_position_on_the_genomic_accession=81652077, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  328573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328573, CurrentID=0, Status='live', Symbol='4930407I10Rik', Aliases='Gm747', description='RIKEN cDNA 4930407I10 gene', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82059151, end_position_on_the_genomic_accession=82066538, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66431, CurrentID=0, Status='live', Symbol='Oxld1', Aliases='1810049H13Rik', description='oxidoreductase like domain containing 1', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120456601, end_position_on_the_genomic_accession=120458091, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66432, CurrentID=0, Status='live', Symbol='Slc7a6os', Aliases='2010007L18Rik, 2400002F02Rik', description='solute carrier family 7, member 6 opposite strand', other_designations='protein SLC7A6OS|solute carrier family 7 member 6 opposite strand transcript homolog', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106200438, end_position_on_the_genomic_accession=106210992, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66433, CurrentID=0, Status='live', Symbol='Chchd7', Aliases='1810049H20Rik', description='coiled-coil-helix-coiled-coil-helix domain containing 7', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3938860, end_position_on_the_genomic_accession=3951382, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66435, CurrentID=0, Status='live', Symbol='Uggt2', Aliases='1810064L21Rik, 3110001A05Rik, 3110027P15Rik, A230065J02Rik, AW047562, Ugcgl2', description='UDP-glucose glycoprotein glucosyltransferase 2', other_designations='UDP-glucose ceramide glucosyltransferase-like 2', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118981765, end_position_on_the_genomic_accession=119099437, orientation='minus', exon_count=49, OMIM='', no_hdr0=''),
  328580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328580, CurrentID=0, Status='live', Symbol='Tubgcp6', Aliases='AI843275, D430023H11, mKIAA1669', description='tubulin, gamma complex associated protein 6', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89099097, end_position_on_the_genomic_accession=89123208, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  66437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66437, CurrentID=0, Status='live', Symbol='Fis1', Aliases='2010003O14Rik, Ttc11', description='fission, mitochondrial 1', other_designations='FIS1 homolog|TPR repeat protein 11|fission 1 (mitochondrial outer membrane) homolog|tetratricopeptide repeat domain 11|tetratricopeptide repeat protein 11', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136953264, end_position_on_the_genomic_accession=136966234, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66438, CurrentID=0, Status='live', Symbol='Hamp2', Aliases='1810073K19Rik, HEPC2', description='hepcidin antimicrobial peptide 2', other_designations='', map_location='7 19.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30922372, end_position_on_the_genomic_accession=30924181, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66439, CurrentID=0, Status='live', Symbol='Borcs7', Aliases='2010012O05Rik, 4930569L17Rik, AI413851, AI467257', description='BLOC-1 related complex subunit 7', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46689906, end_position_on_the_genomic_accession=46703382, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66440, CurrentID=0, Status='live', Symbol='Cdc26', Aliases='2010012C09Rik', description='cell division cycle 26', other_designations='CDC26 subunit of anaphase promoting complex|cell division cycle protein 26 homolog', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62393812, end_position_on_the_genomic_accession=62408640, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66441, CurrentID=0, Status='live', Symbol='Magohb', Aliases='2010012C16Rik, Mago, Magoh, Magoh-rs1', description='mago-nashi homolog B (Drosophila)', other_designations='mago nashi protein|protein mago nashi homolog 1-related', map_location='6 C7', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131284388, end_position_on_the_genomic_accession=131293244, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66442, CurrentID=0, Status='live', Symbol='Spc25', Aliases='2600017H08Rik, 2610205L13Rik, Spbc25', description='SPC25, NDC80 kinetochore complex component, homolog (S. cerevisiae)', other_designations='spindle pole body component 25 homolog', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69193895, end_position_on_the_genomic_accession=69206215, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66443, CurrentID=0, Status='live', Symbol='Tnfaip8l1', Aliases='2600017J23Rik, TIPE1', description='tumor necrosis factor, alpha-induced protein 8-like 1', other_designations='TNF alpha-induced protein 8-like protein 1|TNFAIP8-like protein 1|oxidative stress regulated gene-beta|oxy-beta|tumor necrosis factor, alpha-induced protein 8-like protein 1', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56162491, end_position_on_the_genomic_accession=56173955, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66445, CurrentID=0, Status='live', Symbol='Cyc1', Aliases='2610002H19Rik, AA408921', description='cytochrome c-1', other_designations='complex III subunit 4|complex III subunit IV|cytochrome b-c1 complex subunit 4|ubiquinol-cytochrome-c reductase complex cytochrome c1 subunit', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76343523, end_position_on_the_genomic_accession=76345934, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66446, CurrentID=0, Status='live', Symbol='Exosc7', Aliases='2610002K22Rik, AV212732, mKIAA0116', description='exosome component 7', other_designations='ribosomal RNA-processing protein 42', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123113231, end_position_on_the_genomic_accession=123136129, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66447, CurrentID=0, Status='live', Symbol='Mgst3', Aliases='2010012L10Rik, 2010306B17Rik, 2700004G04Rik, AA516734, GST-III', description='microsomal glutathione S-transferase 3', other_designations='microsomal GST-3|microsomal GST-III', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167372384, end_position_on_the_genomic_accession=167393797, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66448, CurrentID=0, Status='live', Symbol='Mrpl20', Aliases='2610008D01Rik, 4930425I20Rik', description='mitochondrial ribosomal protein L20', other_designations='L20mt|MRP-L20', map_location='4 87.44 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155802883, end_position_on_the_genomic_accession=155808831, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66449, CurrentID=0, Status='live', Symbol='Pam16', Aliases='2010110I09Rik, AV006767, CGI-136, Magmas, Tim16, Timm16', description='presequence translocase-asssociated motor 16 homolog (S. cerevisiae)', other_designations='magmas processed gene 1|mitochondria associated granulocyte macrophage CSF signaling molecule|mitochondria-associated granulocyte macrophage CSF-signaling molecule|mitochondria-associated protein involved in granulocyte-macrophage colony-stimulating factor signal transduction|presequence translocated-associated motor subunit PAM16', map_location='16 2.45 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4616466, end_position_on_the_genomic_accession=4624946, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66451, CurrentID=0, Status='live', Symbol='2610528J11Rik', Aliases='TEMP', description='RIKEN cDNA 2610528J11 gene', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118526995, end_position_on_the_genomic_accession=118530217, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66454, CurrentID=0, Status='live', Symbol='Nmnat1', Aliases='2610529L11Rik, 5730441G13Rik, D4Cole1e, nmnat', description='nicotinamide nucleotide adenylyltransferase 1', other_designations='NMN adenylyltransferase 1|NMN/NaMN adenylyltransferase 1|naMN adenylyltransferase 1|nicotinamide mononucleotide adenylyl transferase|nicotinamide mononucleotide adenylyltransferase 1|nicotinamide/nicotinate mononucleotide adenylyltransferase 1|nicotinate-nucleotide adenylyltransferase 1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149468597, end_position_on_the_genomic_accession=149485142, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66455, CurrentID=0, Status='live', Symbol='Cnpy4', Aliases='2610019P18Rik', description='canopy FGF signaling regulator 4', other_designations='canopy 4 homolog|protein associated with Tlr4', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138187535, end_position_on_the_genomic_accession=138193894, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66459, CurrentID=0, Status='live', Symbol='Pyurf', Aliases='2610022G08Rik, AI847956, Prey', description='Pigy upstream reading frame', other_designations='pre-Y homolog', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57684739, end_position_on_the_genomic_accession=57692078, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66460, CurrentID=0, Status='live', Symbol='Sys1', Aliases='2610042O14Rik, AI266964, C87389', description='SYS1 Golgi-localized integral membrane protein homolog (S. cerevisiae)', other_designations='Sys1 homolog', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164456981, end_position_on_the_genomic_accession=164479637, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66461, CurrentID=0, Status='live', Symbol='Ptpmt1', Aliases='1110001D10Rik, 2810004N20Rik, PLIP', description='protein tyrosine phosphatase, mitochondrial 1', other_designations='PTEN-like phosphatase|phosphoinositide lipid phosphatase|protein-tyrosine phosphatase mitochondrial 1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90910713, end_position_on_the_genomic_accession=90918050, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66462, CurrentID=0, Status='live', Symbol='2810428I15Rik', Aliases='', description='RIKEN cDNA 2810428I15 gene', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70504295, end_position_on_the_genomic_accession=70506773, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66464, CurrentID=0, Status='live', Symbol='Taf12', Aliases='20kDa, 2810422D08Rik, AW557038, Taf2J', description='TATA-box binding protein associated factor 12', other_designations='TAF12 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF12 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 20 kDa|TAFII-20|TAFII20|transcription initiation factor TFIID 20 kDa subunits', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132274355, end_position_on_the_genomic_accession=132293329, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66467, CurrentID=0, Status='live', Symbol='Gtf2h5', Aliases='2700017P07Rik, 2810432H05Rik, D17Wsu155e', description='general transcription factor IIH, polypeptide 5', other_designations='TFB5 ortholog|TFIIH basal transcription factor complex TTD-A subunit', map_location='17 3.7 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6079828, end_position_on_the_genomic_accession=6085488, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66468, CurrentID=0, Status='live', Symbol='Ska1', Aliases='2810433K01Rik, AV117428', description='spindle and kinetochore associated complex subunit 1', other_designations='', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74195299, end_position_on_the_genomic_accession=74207818, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66469, CurrentID=0, Status='live', Symbol='Fam213b', Aliases='2810405K02Rik, AI836168, PM/PGFS', description='family with sequence similarity 213, member B', other_designations='prostamide/PG F synthase|prostamide/PGF synthase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154896430, end_position_on_the_genomic_accession=154899043, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66471, CurrentID=0, Status='live', Symbol='Anp32e', Aliases='2810018A15Rik, AI047746, AI326868, CPD1, LANP-L, LANPL, mLANP-L', description='acidic (leucine-rich) nuclear phosphoprotein 32 family, member E', other_designations='LANP-like protein|cerebellar postnatal development protein 1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95929250, end_position_on_the_genomic_accession=95947387, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66473, CurrentID=0, Status='live', Symbol='Ctrb1', Aliases='2200008D09Rik, AI504462, Ctrb, Prt-2', description='chymotrypsinogen B1', other_designations='', map_location='8 57.98 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111686510, end_position_on_the_genomic_accession=111691010, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66475, CurrentID=0, Status='live', Symbol='Rps23', Aliases='2410044J15Rik', description='ribosomal protein S23', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=90923122, end_position_on_the_genomic_accession=90924732, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66477, CurrentID=0, Status='live', Symbol='Usmg5', Aliases='2010301L15Rik', description='upregulated during skeletal muscle growth 5', other_designations='diabetes-associated protein in insulin-sensitive tissues', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47083471, end_position_on_the_genomic_accession=47090625, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66480, CurrentID=0, Status='live', Symbol='Rpl15', Aliases='2510008H07Rik', description='ribosomal protein L15', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=18267823, end_position_on_the_genomic_accession=18271024, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66481, CurrentID=0, Status='live', Symbol='Rps21', Aliases='1810049N11Rik, 2410030A14Rik', description='ribosomal protein S21', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180257373, end_position_on_the_genomic_accession=180258444, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66482, CurrentID=0, Status='live', Symbol='Exoc2', Aliases='2410030I24Rik, AI648199, Sec5, Sec5l1', description='exocyst complex component 2', other_designations='SEC5 homolog|SEC5-like 1|exocyst complex component Sec5', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=30789516, end_position_on_the_genomic_accession=30974273, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  66483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66483, CurrentID=0, Status='live', Symbol='Rpl36al', Aliases='2410038A03Rik, RPL36A', description='ribosomal protein L36A-like', other_designations='', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69182734, end_position_on_the_genomic_accession=69184067, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66488, CurrentID=0, Status='live', Symbol='Fam136a', Aliases='2010309E21Rik, AW124904, Hyccin', description='family with sequence similarity 136, member A', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86365453, end_position_on_the_genomic_accession=86370063, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66489, CurrentID=0, Status='live', Symbol='Rpl35', Aliases='2410039E09Rik', description='ribosomal protein L35', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=39001581, end_position_on_the_genomic_accession=39005131, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66491, CurrentID=0, Status='live', Symbol='Polr2l', Aliases='2510029B14Rik', description='polymerase (RNA) II (DNA directed) polypeptide L', other_designations='DNA-directed RNA polymerase III subunit L|RNA polymerases I, II, and III subunit ABC5|RPB10 homolog', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141471860, end_position_on_the_genomic_accession=141475153, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66492, CurrentID=0, Status='live', Symbol='Zmat2', Aliases='2610510D14Rik, 2900082I05Rik', description='zinc finger, matrin type 2', other_designations='', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36793923, end_position_on_the_genomic_accession=36799660, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66493, CurrentID=0, Status='live', Symbol='Mrpl51', Aliases='2610511M02Rik, CDA09, HSPC241, Mrp64', description='mitochondrial ribosomal protein L51', other_designations='L51mt|MRP-L51|bMRP-64|bMRP64|mitochondrial ribosomal protein 64', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125192200, end_position_on_the_genomic_accession=125194392, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66494, CurrentID=0, Status='live', Symbol='Prelid1', Aliases='2610524G07Rik, Preli', description='PRELI domain containing 1', other_designations='PX19 homolog|protein of relevant evolutionary and lymphoid interest|px19-like protein', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55322055, end_position_on_the_genomic_accession=55325272, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66495, CurrentID=0, Status='live', Symbol='Ndufb3', Aliases='2700033I16Rik, CI-B12', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex 3', other_designations='NADH-ubiquinone oxidoreductase B12 subunit|complex I-B12', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58586599, end_position_on_the_genomic_accession=58595948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66496, CurrentID=0, Status='live', Symbol='Ppdpf', Aliases='0610012G23Rik, 2610317A05Rik, 2700038C09Rik, 3110053G12Rik', description='pancreatic progenitor cell differentiation and proliferation factor', other_designations='exocrine differentiation and proliferation factor', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181187222, end_position_on_the_genomic_accession=181188506, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66497, CurrentID=0, Status='live', Symbol='Cmss1', Aliases='1110001A06Rik, 2610528E23Rik, 4930572F24Rik, AV099812, AV122629', description='cms small ribosomal subunit 1', other_designations='cms1 ribosomal small subunit homolog', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57302000, end_position_on_the_genomic_accession=57606867, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66498, CurrentID=0, Status='live', Symbol='Dda1', Aliases='1500034J01Rik, 1700095J19Rik, R74921', description='DET1 and DDB1 associated 1', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71469171, end_position_on_the_genomic_accession=71476097, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  328643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328643, CurrentID=0, Status='live', Symbol='Vwa5b2', Aliases='BB182430, EG328644, Gm1738, Gm931', description='von Willebrand factor A domain containing 5B2', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20589319, end_position_on_the_genomic_accession=20605378, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  66500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66500, CurrentID=0, Status='live', Symbol='Slc30a7', Aliases='1810059J10Rik, 2610034N15Rik, 4833428C12Rik, AI467242, ZnT-7, ZnT7, Zntl2', description='solute carrier family 30 (zinc transporter), member 7', other_designations='solute carrier family 30 member 7|znt-like transporter 2', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=115938973, end_position_on_the_genomic_accession=116007414, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66501, CurrentID=0, Status='live', Symbol='1700029H14Rik', Aliases='', description='RIKEN cDNA 1700029H14 gene', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13550722, end_position_on_the_genomic_accession=13562554, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66505, CurrentID=0, Status='live', Symbol='Zmynd11', Aliases='2210402G22Rik, BS69', description='zinc finger, MYND domain containing 11', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=9684836, end_position_on_the_genomic_accession=9766014, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  66506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66506, CurrentID=0, Status='live', Symbol='Psmg3', Aliases='1810042K04Rik, 4930403H09Rik, AI303239', description='proteasome (prosome, macropain) assembly chaperone 3', other_designations='PAC-3|mPAC3', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139823594, end_position_on_the_genomic_accession=139827003, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66508, CurrentID=0, Status='live', Symbol='Lamtor1', Aliases='2400001E08Rik, Pdro, p18', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 1', other_designations='RhoA activator C11orf59 homolog|late endosomal/lysosomal adaptor and MAPK and MTOR activator 1|lipid raft adaptor protein p18|p27Kip1-releasing factor from RhoA|p27RF-Rho|ragulator complex protein PDRO', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101905837, end_position_on_the_genomic_accession=101911903, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66510, CurrentID=0, Status='live', Symbol='Rnf181', Aliases='1700047F16Rik, 2500002L14Rik, 6230415N01Rik, C77350', description='ring finger protein 181', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72359714, end_position_on_the_genomic_accession=72366967, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66511, CurrentID=0, Status='live', Symbol='Chtop', Aliases='2500003M10Rik, Fop, Srag', description='chromatin target of PRMT1', other_designations='Friend of PRMT1 protein|small arginine- and glycine-rich protein|small protein rich in arginine and glycine', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90498539, end_position_on_the_genomic_accession=90509888, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66513, CurrentID=0, Status='live', Symbol='Tab1', Aliases='2310012M03Rik, Map3k7ip1, b2b449Clo', description='TGF-beta activated kinase 1/MAP3K7 binding protein 1', other_designations='TGF-beta-activated kinase 1-binding protein 1|Tak1-binding protein 1|beta activated kinase-1 binding protein-1|mitogen activated protein kinase kinase kinase 7 interacting protein 1|mitogen-activated protein kinase kinase kinase 7 interacting protein 1|mitogen-activated protein kinase kinase kinase 7-interacting protein 1', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80133154, end_position_on_the_genomic_accession=80161709, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66514, CurrentID=0, Status='live', Symbol='Asrgl1', Aliases='2410004D18Rik, ALP, ALP1, AU040643, AW060375', description='asparaginase like 1', other_designations='L-asparaginase|L-asparagine amidohydrolase|asparaginase-like protein 1|beta-aspartyl-peptidase|isoaspartyl dipeptidase', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=9111719, end_position_on_the_genomic_accession=9135566, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66515, CurrentID=0, Status='live', Symbol='Cul7', Aliases='2510004L20Rik, AA409809, C230011P08Rik, p185, p193', description='cullin 7', other_designations='CUL-7', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46650338, end_position_on_the_genomic_accession=46664364, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  328660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328660, CurrentID=0, Status='live', Symbol='Bex6', Aliases='B020003O03Rik', description='brain expressed gene 6', other_designations='brain expressed X-linked 6|brain expressed, X-linked 6|brain-expressed X-linked protein 6 homolog', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32179800, end_position_on_the_genomic_accession=32186944, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66521, CurrentID=0, Status='live', Symbol='Rwdd1', Aliases='0710001K08Rik, 2610002D06Rik, 2700069A07Rik, IH1', description='RWD domain containing 1', other_designations='DRG family-regulatory protein 2', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=33996555, end_position_on_the_genomic_accession=34019616, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66522, CurrentID=0, Status='live', Symbol='Pgpep1', Aliases='2810003H13Rik, Gm19458, PAP-I, PGP, PGP-I, PGPI, Pcp', description='pyroglutamyl-peptidase I', other_designations='5-oxoprolyl-peptidase|putative pyroglutamyl-peptidase I|pyroglutamyl aminopeptidase I|pyrrolidone-carboxylate peptidase', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70646436, end_position_on_the_genomic_accession=70659857, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66523, CurrentID=0, Status='live', Symbol='2810004N23Rik', Aliases='3110004G14Rik, Ayu21-55, Gt(pU21)55Imeg, GtAyu21-55', description='RIKEN cDNA 2810004N23 gene', other_designations='Ayu21-55Imeg|Gt(Ayu21)55Imeg', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124839355, end_position_on_the_genomic_accession=124863029, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66525, CurrentID=0, Status='live', Symbol='Timm50', Aliases='2810403L02Rik, AU015082, TIM50L', description='translocase of inner mitochondrial membrane 50', other_designations='translocase of inner mitochondrial membrane 50 homolog', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28305516, end_position_on_the_genomic_accession=28312046, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66526, CurrentID=0, Status='live', Symbol='Tceanc2', Aliases='2210010B22Rik, 2210012G02Rik, 6330404A07Rik, AU015593, Tdeanc2', description='transcription elongation factor A (SII) N-terminal and central domain containing 2', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107134162, end_position_on_the_genomic_accession=107178366, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66528, CurrentID=0, Status='live', Symbol='Smim5', Aliases='2210020M01Rik', description='small integral membrane protein 5', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115899512, end_position_on_the_genomic_accession=115906269, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66530, CurrentID=0, Status='live', Symbol='Ubxn6', Aliases='1200008L11Rik, 2210415J11Rik, AU040909, Ubxd1, Ubxdc2', description='UBX domain protein 6', other_designations='UBX domain containing 1|UBX domain-containing 2|UBX domain-containing protein 1', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56068249, end_position_on_the_genomic_accession=56074989, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66531, CurrentID=0, Status='live', Symbol='Cmc2', Aliases='1110046L09Rik, 2310061C15Rik, DC13', description='COX assembly mitochondrial protein 2', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116888685, end_position_on_the_genomic_accession=116944209, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66532, CurrentID=0, Status='live', Symbol='Rep15', Aliases='2210417D09Rik', description='RAB15 effector protein', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=147032537, end_position_on_the_genomic_accession=147033518, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66533, CurrentID=0, Status='live', Symbol='2310050C09Rik', Aliases='', description='RIKEN cDNA 2310050C09 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92868359, end_position_on_the_genomic_accession=92870205, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66536, CurrentID=0, Status='live', Symbol='Nipsnap3b', Aliases='2700063N13Rik, Nipsnap3a', description='nipsnap homolog 3B (C. elegans)', other_designations='NIPSNAP-related protein|nipsnap homolog 3A|protein NipSnap homolog 3A', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=53011775, end_position_on_the_genomic_accession=53022060, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66537, CurrentID=0, Status='live', Symbol='Pomp', Aliases='2510048O06Rik, Ump1', description='proteasome maturation protein', other_designations='mUMP1|proteassemblin|protein UMP1 homolog', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147860628, end_position_on_the_genomic_accession=147875778, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66538, CurrentID=0, Status='live', Symbol='Rps19bp1', Aliases='2510038A11Rik, AI604923, S19BP', description='ribosomal protein S19 binding protein 1', other_designations='40S ribosomal protein S19-binding protein 1|RPS19-binding protein 1|S19 binding protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80260614, end_position_on_the_genomic_accession=80264306, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66540, CurrentID=0, Status='live', Symbol='Fam107b', Aliases='3110001A13Rik, AA589595, AA959793, AI132312', description='family with sequence similarity 107, member B', other_designations='novel DUF1151 domain containing protein', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3704812, end_position_on_the_genomic_accession=3782142, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66541, CurrentID=0, Status='live', Symbol='Immp1l', Aliases='1500034J20Rik, 2610528O17Rik, AU042386', description='IMP1 inner mitochondrial membrane peptidase-like (S. cerevisiae)', other_designations='IMP1-like protein', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=105904638, end_position_on_the_genomic_accession=105965558, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66548, CurrentID=0, Status='live', Symbol='Adamtsl5', Aliases='2010109H09Rik, Thsd6', description='ADAMTS-like 5', other_designations='thrombospondin, type I domain containing 6', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80339793, end_position_on_the_genomic_accession=80349036, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66549, CurrentID=0, Status='live', Symbol='Aggf1', Aliases='2010009L17Rik, 2310029P06Rik, AW112072, Peg3, VG5Q', description='angiogenic factor with G patch and FHA domains 1', other_designations='angiogenic factor VG5Q', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95350683, end_position_on_the_genomic_accession=95375357, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66552, CurrentID=0, Status='live', Symbol='Sppl2a', Aliases='2010106G01Rik, C130089K23Rik', description='signal peptide peptidase like 2A', other_designations='IMP-3|SPP-like 2A protein|intramembrane protease 3|presenilin-like protein 2|protein SPPL2a', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126890391, end_position_on_the_genomic_accession=126933235, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  328699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328699, CurrentID=0, Status='live', Symbol='Gabrr3', Aliases='', description='gamma-aminobutyric acid (GABA) receptor, rho 3', other_designations='', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59407382, end_position_on_the_genomic_accession=59464099, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66556, CurrentID=0, Status='live', Symbol='Drap1', Aliases='2310074H19Rik, NC2-alpha', description='Dr1 associated protein 1 (negative cofactor 2 alpha)', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5422804, end_position_on_the_genomic_accession=5424979, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66557, CurrentID=0, Status='live', Symbol='Bpifb2', Aliases='2310034L21Rik, 2310069A01Rik, AV089125, Bpil1, LPLUNC2, RYSR', description='BPI fold containing family B, member 2', other_designations='bactericidal/permeability-increasing protein-like 1', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153875045, end_position_on_the_genomic_accession=153895270, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  66559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66559, CurrentID=0, Status='live', Symbol='Metap1d', Aliases='2310066F24Rik, 3110033D18Rik, AV117938, MAP 1D, Map1d, MetAP 1D, Metapl1', description='methionyl aminopeptidase type 1D (mitochondrial)', other_designations='MAP 1D|metAP 1D|methionine aminopeptidase-like 1|methionyl aminopeptidase type 1D, mitochondrial|peptidase M 1D', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71453319, end_position_on_the_genomic_accession=71525196, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66561, CurrentID=0, Status='live', Symbol='Teddm3', Aliases='2310042E22Rik', description='transmembrane epididymal family member 3', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21152659, end_position_on_the_genomic_accession=21153944, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66566, CurrentID=0, Status='live', Symbol='Ntpcr', Aliases='2310079N02Rik, AI449709', description='nucleoside-triphosphatase, cancer-related', other_designations='NTPase|nucleoside triphosphate phosphohydrolase|nucleoside-triphosphatase C1orf57 homolog', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125729976, end_position_on_the_genomic_accession=125748235, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66568, CurrentID=0, Status='live', Symbol='Rwdd3', Aliases='2510027J23Rik, 3110037C01Rik, AI035684, AI662458, Rsume, X2CR1', description='RWD domain containing 3', other_designations='RWD domain-containing sumoylation enhancer', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121155400, end_position_on_the_genomic_accession=121171780, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66569, CurrentID=0, Status='live', Symbol='Gdpd1', Aliases='2610020H15Rik', description='glycerophosphodiester phosphodiesterase domain containing 1', other_designations='glycerophosphodiester phosphodiesterase 4', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87033794, end_position_on_the_genomic_accession=87074953, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66570, CurrentID=0, Status='live', Symbol='Cenpm', Aliases='2610019I03Rik, AI853711, Pane1', description='centromere protein M', other_designations='CENP-M|proliferation associated nuclear element 1|proliferation-associated nuclear element protein 1', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82233762, end_position_on_the_genomic_accession=82244767, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66573, CurrentID=0, Status='live', Symbol='Dzip1', Aliases='2510025K24Rik, 2810422M04Rik', description='DAZ interacting protein 1', other_designations='DAZ-interacting protein 1 homolog', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118875520, end_position_on_the_genomic_accession=118925520, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  66576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66576, CurrentID=0, Status='live', Symbol='Uqcrh', Aliases='2210416J04Rik, 2310021J10Rik, 2610041P16Rik', description='ubiquinol-cytochrome c reductase hinge protein', other_designations='complex III subunit 6|complex III subunit VIII|cytochrome C1, nonheme 11 kDa protein|cytochrome c1 non-heme 11 kDa protein|mitochondrial hinge protein|ubiquinol cytochrome c reductase hinge protein|ubiquinol-cytochrome c reductase complex 11 kDa protein', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116066965, end_position_on_the_genomic_accession=116075070, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66578, CurrentID=0, Status='live', Symbol='Mis18a', Aliases='2610039C10Rik, 2810018N07Rik', description='MIS18 kinetochore protein A', other_designations='MIS18 kinetochore protein homolog A', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90719312, end_position_on_the_genomic_accession=90727371, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66580, CurrentID=0, Status='live', Symbol='Esf1', Aliases='2610101J03Rik, ABTAP, AW545818, C79684', description='ESF1 nucleolar pre-rRNA processing protein homolog', other_designations='ABT1-associated protein|ESF1, nucleolar pre-rRNA processing protein, homolog', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=140119880, end_position_on_the_genomic_accession=140170558, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66583, CurrentID=0, Status='live', Symbol='Exosc1', Aliases='2610035C18Rik, 2610104C07Rik, 2610312F07Rik, AI447561', description='exosome component 1', other_designations="3'-5' exoribonuclease CSL4 homolog", map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41922979, end_position_on_the_genomic_accession=41933418, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66585, CurrentID=0, Status='live', Symbol='Snrnp40', Aliases='0610009C03Rik, Prp8bp, Wdr57', description='small nuclear ribonucleoprotein 40 (U5)', other_designations='U5 snRNP 40 kDa protein|U5 snRNP-specific protein (Prp8-binding)|WD repeat domain 57 (U5 snRNP specific)|WD repeat-containing protein 57|small nuclear ribonucleoprotein 40kDa (U5)', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130360132, end_position_on_the_genomic_accession=130390033, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66586, CurrentID=0, Status='live', Symbol='Crls1', Aliases='0610009I22Rik, 4930557M15Rik, 5730490M08Rik', description='cardiolipin synthase 1', other_designations='CLS', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132846650, end_position_on_the_genomic_accession=132866786, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66587, CurrentID=0, Status='live', Symbol='Fastk', Aliases='0610011K02Rik, 610011K02Rik', description='Fas-activated serine/threonine kinase', other_designations='FAST kinase', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24441039, end_position_on_the_genomic_accession=24445282, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66588, CurrentID=0, Status='live', Symbol='Cmpk1', Aliases='0610011D08Rik, CK, Cmpk, UMP-CMPk', description='cytidine monophosphate (UMP-CMP) kinase 1', other_designations='CK|UMP/CMP kinase|UMP/CMPK|cytidine monophosphate kinase|cytidylate kinase 1|dCMP kinase|deoxycytidylate kinase|likely ortholog of H. sapiens UMP-CMP kinase (UMP-CMPK)|nucleoside-diphosphate kinase|uridine monophosphate kinase|uridine monophosphate/cytidine monophosphate kinase', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=114960613, end_position_on_the_genomic_accession=114987228, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66589, CurrentID=0, Status='live', Symbol='Ube2v1', Aliases='0610011J09Rik, AI256840, CROC-1, CROC1, D7Bwg1382e, UEV-1', description='ubiquitin-conjugating enzyme E2 variant 1', other_designations='', map_location='2 87.44 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167607639, end_position_on_the_genomic_accession=167632047, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66590, CurrentID=0, Status='live', Symbol='Farsa', Aliases='0610012A19Rik, Farsla', description='phenylalanyl-tRNA synthetase, alpha subunit', other_designations='pheRS|phenylalanine--tRNA ligase alpha chain|phenylalanine-tRNA synthetase-like, alpha subunit|phenylalanyl-tRNA synthetase alpha chain', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84856972, end_position_on_the_genomic_accession=84869257, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66591, CurrentID=0, Status='live', Symbol='Mad2l1bp', Aliases='0610009D16Rik, 2510031P20Rik, AU018945, AW550977, C76045, CMT2A, CMT2B, Mad2lbp, mKIAA0110', description='MAD2L1 binding protein', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46147385, end_position_on_the_genomic_accession=46153551, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66592, CurrentID=0, Status='live', Symbol='Stoml2', Aliases='0610038F01Rik, MSLP2, SLP-2', description='stomatin (Epb7.2)-like 2', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43027690, end_position_on_the_genomic_accession=43031384, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66593, CurrentID=0, Status='live', Symbol='Diablo', Aliases='0610041G12Rik, 1700006L01Rik, AU040403, Smac', description='diablo, IAP-binding mitochondrial protein', other_designations='diablo homolog|direct IAP binding protein with low PI|direct IAP-binding protein with low pI|second mitochondria-derived activator of caspase', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123511330, end_position_on_the_genomic_accession=123524164, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66594, CurrentID=0, Status='live', Symbol='Uqcr11', Aliases='0710008D09Rik, AL022707, Uqcr', description='ubiquinol-cytochrome c reductase, complex III subunit XI', other_designations='complex III subunit 10|complex III subunit XI|ubiquinol-cytochrome c reductase (6.4kD) subunit|ubiquinol-cytochrome c reductase complex 6.4 kDa protein|ubiquinol-cytochrome c reductase subunit', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80402997, end_position_on_the_genomic_accession=80406821, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66595, CurrentID=0, Status='live', Symbol='Aste1', Aliases='1100001A21Rik, AV006403', description='asteroid homolog 1 (Drosophila)', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105395294, end_position_on_the_genomic_accession=105408298, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66596, CurrentID=0, Status='live', Symbol='Gtf3a', Aliases='2010015D03Rik, 2610111I01Rik, 5330403M05Rik', description='general transcription factor III A', other_designations='TFIIIA|factor A', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146948657, end_position_on_the_genomic_accession=146955614, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66597, CurrentID=0, Status='live', Symbol='Trim13', Aliases='3110001L12Rik, CAR, LEU5, RNF77, Rfp2', description='tripartite motif-containing 13', other_designations='putative tumor suppressor RFP2|ret finger protein 2|tripartite motif protein 13|tripartite motif-containing protein 13', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61598226, end_position_on_the_genomic_accession=61605946, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66598, CurrentID=0, Status='live', Symbol='3110001I22Rik', Aliases='', description='RIKEN cDNA 3110001I22 gene', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13672020, end_position_on_the_genomic_accession=13678385, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66599, CurrentID=0, Status='live', Symbol='Rdm1', Aliases='2410008M22Rik, AW212028, Rad52b', description='RAD52 motif 1', other_designations='RAD52 homolog B|RAD52B homolog', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101627242, end_position_on_the_genomic_accession=101636091, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66601, CurrentID=0, Status='live', Symbol='Tmigd1', Aliases='2010002A20Rik, Tmigd', description='transmembrane and immunoglobulin domain containing 1', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76891246, end_position_on_the_genomic_accession=76916587, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66602, CurrentID=0, Status='live', Symbol='1700020I14Rik', Aliases='1190006L01Rik, 4933437I03Rik, AI451541, AI585852, AW538376, AW558897, Cyrano', description='RIKEN cDNA 1700020I14 gene', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119594296, end_position_on_the_genomic_accession=119600746, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66603, CurrentID=0, Status='live', Symbol='Gemin2', Aliases='Sip1, gemin-2', description='gem (nuclear organelle) associated protein 2', other_designations='SMN-interacting protein 1|component of gems 2|survival of motor neuron protein-interacting protein 1|survivor of motor neuron protein interacting protein 1', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=59013393, end_position_on_the_genomic_accession=59027988, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66605, CurrentID=0, Status='live', Symbol='1700017N19Rik', Aliases='', description='RIKEN cDNA 1700017N19 gene', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=100590511, end_position_on_the_genomic_accession=100618391, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  66606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66606, CurrentID=0, Status='live', Symbol='Lrrc57', Aliases='2810002D13Rik, AA407405', description='leucine rich repeat containing 57', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120604238, end_position_on_the_genomic_accession=120609508, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66607, CurrentID=0, Status='live', Symbol='Ms4a4d', Aliases='2210417J23Rik', description='membrane-spanning 4-domains, subfamily A, member 4D', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11536849, end_position_on_the_genomic_accession=11558466, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66609, CurrentID=0, Status='live', Symbol='Cryzl1', Aliases='2210407J23Rik, 2410006O11Rik, QOH-1', description='crystallin, zeta (quinone reductase)-like 1', other_designations='quinone oxidoreductase homolog 1|zeta-crystallin homolog', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91689322, end_position_on_the_genomic_accession=91728802, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66610, CurrentID=0, Status='live', Symbol='Abi3', Aliases='2210414K06Rik, AI987680, NESH', description='ABI gene family, member 3', other_designations='new molecule including SH3', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95830072, end_position_on_the_genomic_accession=95842476, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66611, CurrentID=0, Status='live', Symbol='Ribc1', Aliases='2610028I09Rik, W08639', description='RIB43A domain with coiled-coils 1', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152004580, end_position_on_the_genomic_accession=152018219, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66612, CurrentID=0, Status='live', Symbol='Ormdl3', Aliases='2810011N17Rik, AI315223', description='ORM1-like 3 (S. cerevisiae)', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98581294, end_position_on_the_genomic_accession=98587245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66614, CurrentID=0, Status='live', Symbol='Gpatch4', Aliases='2610029K21Rik, Gpatc4', description='G patch domain containing 4', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88043106, end_position_on_the_genomic_accession=88055994, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66615, CurrentID=0, Status='live', Symbol='Atg4b', Aliases='2510009N07Rik, AW048066, Apg4bl, Autl1, Atg4b', description='autophagy related 4B, cysteine peptidase', other_designations='APG4 (ATG4) autophagy-related homolog B|AUT-like 1 cysteine endopeptidase|AUT-like 1, cysteine endopeptidase|autophagin 1|autophagin-1|autophagy-related 4B|autophagy-related cysteine endopeptidase 1|autophagy-related protein 4 homolog B|cysteine protease involved in autophagy APG4-B', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93755033, end_position_on_the_genomic_accession=93789529, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  66616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66616, CurrentID=0, Status='live', Symbol='Snx9', Aliases='2700073N08Rik, SDP1, SH3PX1', description='sorting nexin 9', other_designations='', map_location='17 3.51 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=5841380, end_position_on_the_genomic_accession=5930711, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  66617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66617, CurrentID=0, Status='live', Symbol='Ntmt1', Aliases='2610205E22Rik, AL033331, AL033332, Mettl11a, NTM1A', description='N-terminal Xaa-Pro-Lys N-methyltransferase 1', other_designations='X-Pro-Lys N-terminal protein methyltransferase 1A|alpha N-terminal protein methyltransferase 1A|methyltransferase like 11A|methyltransferase-like protein 11A', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30807780, end_position_on_the_genomic_accession=30823025, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66618, CurrentID=0, Status='live', Symbol='Snrnp27', Aliases='2610209M04Rik, AI449063, AU022368, Ry1', description='small nuclear ribonucleoprotein 27 (U4/U6.U5)', other_designations='U4/U6.U5 snRNP 27 kDa protein|U4/U6.U5 tri-snRNP-associated protein 3|U4/U6.U5-27K|putative nucleic acid binding protein RY-1', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86675169, end_position_on_the_genomic_accession=86684491, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66622, CurrentID=0, Status='live', Symbol='Ubr7', Aliases='5730410I19Rik, AA589405, AW557761', description='ubiquitin protein ligase E3 component n-recognin 7 (putative)', other_designations='N-recognin-7', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102757975, end_position_on_the_genomic_accession=102777701, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66624, CurrentID=0, Status='live', Symbol='Spcs2', Aliases='5730406I15Rik, AA408713, mKIAA0102', description='signal peptidase complex subunit 2 homolog (S. cerevisiae)', other_designations='SPase 25 kDa subunit|microsomal signal peptidase 25 kDa subunit', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99837569, end_position_on_the_genomic_accession=99870341, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66625, CurrentID=0, Status='live', Symbol='Pnisr', Aliases='5730406M06Rik, AA673174, Pinsr, Sfrs18, Srsf18', description='PNN interacting serine/arginine-rich', other_designations='SRrp130|serine-arginine-rich-splicing regulatory protein 130|serine/arginine-rich splicing factor 18|serine/arginine-rich-splicing regulatory protein 130|splicing factor, arginine/serine-rich 130|splicing factor, arginine/serine-rich 18', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=21847583, end_position_on_the_genomic_accession=21876475, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66626, CurrentID=0, Status='live', Symbol='Cdip1', Aliases='2700048O17Rik, 5730403B10Rik, AU018521, AW742692, CDIP', description='cell death inducing Trp53 target 1', other_designations='LITAF-like protein', map_location='16 2.47 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4765461, end_position_on_the_genomic_accession=4790287, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66627, CurrentID=0, Status='live', Symbol='Ogfod2', Aliases='1300006G11Rik, 5730405M13Rik, AU041636, AW552395', description='2-oxoglutarate and iron-dependent oxygenase domain containing 2', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124112058, end_position_on_the_genomic_accession=124115480, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66628, CurrentID=0, Status='live', Symbol='Thg1l', Aliases='', description='tRNA-histidine guanylyltransferase 1-like (S. cerevisiae)', other_designations='interphase cyctoplasmic foci protein 45', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=45945305, end_position_on_the_genomic_accession=45955503, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66629, CurrentID=0, Status='live', Symbol='Golph3', Aliases='4733401N08Rik, 5730410D03Rik, AW413496', description='golgi phosphoprotein 3', other_designations='coat protein GPP34', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=12321197, end_position_on_the_genomic_accession=12351601, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66631, CurrentID=0, Status='live', Symbol='Mfsd14b', Aliases='5730414C17Rik, AU045608, AW539692, Hiatl1', description='major facilitator superfamily domain containing 14B', other_designations='hippocampus abundant transcript-like 1', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=65065030, end_position_on_the_genomic_accession=65113000, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66632, CurrentID=0, Status='live', Symbol='Dph6', Aliases='5730421E18Rik, Atpbd4', description='diphthamine biosynthesis 6', other_designations='ATP binding domain 4|ATP-binding domain-containing protein 4|diphthamide synthase|diphthamide synthetase|protein DPH6 homolog', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=114516416, end_position_on_the_genomic_accession=114655052, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  328778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328778, CurrentID=0, Status='live', Symbol='Rab26', Aliases='A830020M03Rik', description='RAB26, member RAS oncogene family', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24529054, end_position_on_the_genomic_accession=24533747, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  328779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328779, CurrentID=0, Status='live', Symbol='Hs3st6', Aliases='3-OST-6', description='heparan sulfate (glucosamine) 3-O-sulfotransferase 6', other_designations='heparan sulfate 3-O-sulfotransferase 6|heparan sulfate D-glucosaminyl 3-O-sulfotransferase 6', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24752950, end_position_on_the_genomic_accession=24758690, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  328780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328780, CurrentID=0, Status='live', Symbol='Prss34', Aliases='C130091P12, Mcp11, mMcp-11', description='protease, serine 34', other_designations='mast cell protease 11|protease, serine, 34', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25298394, end_position_on_the_genomic_accession=25300168, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66637, CurrentID=0, Status='live', Symbol='Tsen15', Aliases='5730449L18Rik, AL023077, Sen15', description='tRNA splicing endonuclease 15 homolog (S. cerevisiae)', other_designations='tRNA-intron endonuclease Sen15', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152370735, end_position_on_the_genomic_accession=152386682, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  328783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328783, CurrentID=0, Status='live', Symbol='Mslnl', Aliases='4732467B22', description='mesothelin-like', other_designations='pre-pro-megakaryocyte-potentiating-factor-like', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25733503, end_position_on_the_genomic_accession=25748530, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  66641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66641, CurrentID=0, Status='live', Symbol='Sike1', Aliases='2810005O12Rik, 5730470L24Rik, AI450236, AI839862, Sike, Sikeb', description='suppressor of IKBKE 1', other_designations='suppressor of IKK epsilon|suppressor of IKK-epsilon', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102995740, end_position_on_the_genomic_accession=103003914, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66642, CurrentID=0, Status='live', Symbol='Ctnnbl1', Aliases='5730471K09Rik, NAP, NYD-SP19, P14L', description='catenin, beta like 1', other_designations='nuclear protein NAP|nuclear-associated protein', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157737401, end_position_on_the_genomic_accession=157891903, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  66643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66643, CurrentID=0, Status='live', Symbol='Lix1', Aliases='5730466L18Rik', description='limb and CNS expressed 1', other_designations='limb expression 1 homolog', map_location='17 A3.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17402686, end_position_on_the_genomic_accession=17459387, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  328788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328788, CurrentID=0, Status='live', Symbol='Clpsl2', Aliases='Gm749', description='colipase-like 2', other_designations='colipase-like protein C6orf126 homolog', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28549487, end_position_on_the_genomic_accession=28552618, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66645, CurrentID=0, Status='live', Symbol='Pspc1', Aliases='5730470C09Rik, AI327109, AI449052, PSP1', description='paraspeckle protein 1', other_designations='mPSP1', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56722441, end_position_on_the_genomic_accession=56778316, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66646, CurrentID=0, Status='live', Symbol='Rpe', Aliases='2810429B02Rik, 5730518J08Rik', description='ribulose-5-phosphate-3-epimerase', other_designations='ribulose-5-phosphate-epimerase', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=66700797, end_position_on_the_genomic_accession=66719805, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66647, CurrentID=0, Status='live', Symbol='Ndnl2', Aliases='5730494G16Rik, AI642138, BB044375, HCA4, Mageg1, mage-g1', description='necdin-like 2', other_designations='MAGE-G1 antigen|melanoma antigen, family G, 1|necdin-like protein 2', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64871650, end_position_on_the_genomic_accession=64873040, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66648, CurrentID=0, Status='live', Symbol='Tpgs2', Aliases='5730437P09Rik, 5730494M16Rik, AI666318, Pgs2', description='tubulin polyglutamylase complex subunit 2', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=25127223, end_position_on_the_genomic_accession=25169007, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  404324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404324, CurrentID=0, Status='live', Symbol='Olfr1051', Aliases='GA_x5J8B7W3KVV-25848-25992, MOR187-5', description='olfactory receptor 1051', other_designations='GA_x6K02T2Q125-47756192-47755266|olfactory receptor GA_x5J8B7W3KVV-25848-25992', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86275559, end_position_on_the_genomic_accession=86276485, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66650, CurrentID=0, Status='live', Symbol='Nepn', Aliases='5730521E12Rik, Npn, periolin', description='nephrocan', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52388870, end_position_on_the_genomic_accession=52404611, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  328795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328795, CurrentID=0, Status='live', Symbol='Ubash3a', Aliases='5830413C03Rik, C330001M22, Sts-2, TULA, TULA-1', description='ubiquitin associated and SH3 domain containing, A', other_designations='T-cell ubiquitin ligand 1|suppressor of T-cell receptor signaling 2', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31207713, end_position_on_the_genomic_accession=31253419, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  66653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66653, CurrentID=0, Status='live', Symbol='Brf2', Aliases='2700059M06Rik, 5730512K07Rik, AU018504, AU022602, BRFU, TFIIIB50', description='BRF2, RNA polymerase III transcription initiation factor 50kDa subunit', other_designations='B-related factor 2|BRF-2|BRF2, subunit of RNA polymerase III transcription initiation factor, BRF1-like', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27123355, end_position_on_the_genomic_accession=27128699, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66654, CurrentID=0, Status='live', Symbol='Tex12', Aliases='5730518K06Rik', description='testis expressed gene 12', other_designations='testis exprtessed gene 12', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50557148, end_position_on_the_genomic_accession=50561268, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66656, CurrentID=0, Status='live', Symbol='Eef1d', Aliases='EF-1-delta', description='eukaryotic translation elongation factor 1 delta (guanine nucleotide exchange protein)', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75894796, end_position_on_the_genomic_accession=75909556, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  328801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328801, CurrentID=0, Status='live', Symbol='Zfp414', Aliases='0610030H11Rik, Znf414', description='zinc finger protein 414', other_designations='zinc finger, C2H2 type containing protein', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33629092, end_position_on_the_genomic_accession=33631716, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66658, CurrentID=0, Status='live', Symbol='Ccdc51', Aliases='5730568A12Rik, AI551049', description='coiled-coil domain containing 51', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109082496, end_position_on_the_genomic_accession=109093363, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66659, CurrentID=0, Status='live', Symbol='Acp6', Aliases='5730559A09Rik, ACPL1, AU022842, mPACPL1', description='acid phosphatase 6, lysophosphatidic', other_designations='PACPL1|acid phosphatase like 1|acid phosphatase-like protein 1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97158752, end_position_on_the_genomic_accession=97176576, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66660, CurrentID=0, Status='live', Symbol='Sltm', Aliases='5730455C01Rik, 5730555F13Rik, 9130215G10Rik, Met', description='SAFB-like, transcription modulator', other_designations='SAF-like transcription modulator|modulator of estrogen induced transcription|modulator of estrogen-induced transcription', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70541796, end_position_on_the_genomic_accession=70592237, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  66661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66661, CurrentID=0, Status='live', Symbol='Srp72', Aliases='5730576P14Rik, 72kDa, AI132477, BC019196, C77589', description='signal recognition particle 72', other_designations='signal recognition particle 72 kDa', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76974701, end_position_on_the_genomic_accession=76999935, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  66663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66663, CurrentID=0, Status='live', Symbol='Uba5', Aliases='5730525G14Rik, AW240750, Ube1dc1', description='ubiquitin-like modifier activating enzyme 5', other_designations='UFM1-activating enzyme|ubiquitin-activating enzyme 5|ubiquitin-activating enzyme E1 domain-containing protein 1|ubiquitin-activating enzyme E1-domain containing 1', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=104046588, end_position_on_the_genomic_accession=104063176, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66664, CurrentID=0, Status='live', Symbol='Tmem41a', Aliases='2900010K02Rik, 5730578N08Rik', description='transmembrane protein 41a', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21934327, end_position_on_the_genomic_accession=21947552, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66665, CurrentID=0, Status='live', Symbol='Msantd3', Aliases='2410046L14Rik, 2810429P03Rik, 5730528L13Rik, Mysantd3', description='Myb/SANT-like DNA-binding domain containing 3', other_designations='UPF0439 protein C9orf30 homolog', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48539832, end_position_on_the_genomic_accession=48565892, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66667, CurrentID=0, Status='live', Symbol='Hspbap1', Aliases='3830421G21Rik', description='Hspb associated protein 1', other_designations='27 KdA heat shock protein-associated protein 1|protein associated with small stress protein 1', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35769663, end_position_on_the_genomic_accession=35828462, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66671, CurrentID=0, Status='live', Symbol='Ccnh', Aliases='6330408H09Rik, AI661354, AV102684, AW538719', description='cyclin H', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=85189377, end_position_on_the_genomic_accession=85213726, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66673, CurrentID=0, Status='live', Symbol='Sorcs3', Aliases='6330404A12Rik, AW045579', description='sortilin-related VPS10 domain containing receptor 3', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=48206025, end_position_on_the_genomic_accession=48805505, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  66674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66674, CurrentID=0, Status='live', Symbol='Spryd7', Aliases='6330409N04Rik, Clld6', description='SPRY domain containing 7', other_designations='CLL deletion region gene 6 protein homolog|chronic lymphocytic leukemia deletion region gene 6 protein homolog', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61533879, end_position_on_the_genomic_accession=61556886, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66676, CurrentID=0, Status='live', Symbol='Tmed7', Aliases='3930401E15Rik, 5830493P14Rik, AU044611, AW545765, TRAM', description='transmembrane emp24 protein transport domain containing 7', other_designations='', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46585928, end_position_on_the_genomic_accession=46597535, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66679, CurrentID=0, Status='live', Symbol='Rae1', Aliases='3230401I12Rik, D2Ertd342e, MNRP, MNRP41', description='RAE1 RNA export 1 homolog (S. pombe)', other_designations='mRNA-associated protein mrnp 41|rae1 protein homolog', map_location='2 95.66 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173000112, end_position_on_the_genomic_accession=173015739, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66680, CurrentID=0, Status='live', Symbol='Oser1', Aliases='3230401D17Rik, AI256769, Perit1', description='oxidative stress responsive serine rich 1', other_designations='oxidative stress-responsive protein 1|peroxide-inducible transcript 1 protein', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163405822, end_position_on_the_genomic_accession=163419470, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66681, CurrentID=0, Status='live', Symbol='Pgm1', Aliases='3230402E02Rik, Pgm-1, Pgm2', description='phosphoglucomutase 1', other_designations='glucose phosphomutase 2|phosphodeoxyribomutase|phosphoglucomutase 2|phosphoglucomutase-2|phosphopentomutase', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=64092831, end_position_on_the_genomic_accession=64128355, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  66682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66682, CurrentID=0, Status='live', Symbol='Trappc5', Aliases='4021401A16Rik, TRS31', description='trafficking protein particle complex 5', other_designations='', map_location='8 A1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3676477, end_position_on_the_genomic_accession=3680921, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66684, CurrentID=0, Status='live', Symbol='Tceal8', Aliases='3930402F23Rik', description='transcription elongation factor A (SII)-like 8', other_designations='TCEA-like protein 8|transcription elongation factor S-II protein-like 8', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136168984, end_position_on_the_genomic_accession=136172251, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  328829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328829, CurrentID=0, Status='live', Symbol='9830107B12Rik', Aliases='Trem5', description='RIKEN cDNA 9830107B12 gene', other_designations='triggering receptor expressed in myeloid cells 5|triggering receptor expressed on myeloid cells 5', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48098750, end_position_on_the_genomic_accession=48146307, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  328830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328830, CurrentID=0, Status='live', Symbol='A530064D06Rik', Aliases='Trem4, pdctrem', description='RIKEN cDNA A530064D06 gene', other_designations='plasmacytoid dendritic cell-specific receptor|triggering receptor expressed on myeloid cells 4', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48151896, end_position_on_the_genomic_accession=48167257, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66687, CurrentID=0, Status='live', Symbol='Tbc1d15', Aliases='4432405K22Rik, AU022358, AU023730, AW557774, Ly6dl, Rab7-GAP', description='TBC1 domain family, member 15', other_designations='GAP for RAB7|GTPase-activating protein RAB7|lymphocyte antigen 6, locus D ligand', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115197871, end_position_on_the_genomic_accession=115251493, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  328833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328833, CurrentID=0, Status='live', Symbol='Treml2', Aliases='AW049306, Gm750, Tlt2', description='triggering receptor expressed on myeloid cells-like 2', other_designations='TLT-2|triggering receptor expressed on myeloid cells-like 2a dominant|triggering receptor expressed on myeloid cells-like protein 2', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48299387, end_position_on_the_genomic_accession=48312534, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66690, CurrentID=0, Status='live', Symbol='Tmem186', Aliases='2810440M13Rik, 4432406C05Rik, AI648175, AW210586', description='transmembrane protein 186', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8633731, end_position_on_the_genomic_accession=8637701, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66691, CurrentID=0, Status='live', Symbol='Gapvd1', Aliases='2010005B09Rik, 4432404J10Rik, AW108497, Gapex-5, RAP6, RME-6, mKIAA1521', description='GTPase activating protein and VPS9 domains 1', other_designations='rab5-activating protein 6', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34676179, end_position_on_the_genomic_accession=34755263, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  66694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66694, CurrentID=0, Status='live', Symbol='Uqcrfs1', Aliases='4430402G14Rik, AI875505', description='ubiquinol-cytochrome c reductase, Rieske iron-sulfur polypeptide 1', other_designations='RISP|complex III subunit 5|cytochrome b-c1 complex subunit 5|rieske iron-sulfur protein|ubiquinol-cytochrome c reductase iron-sulfur subunit', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=30540312, end_position_on_the_genomic_accession=30545316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66695, CurrentID=0, Status='live', Symbol='Aspn', Aliases='4631401G09Rik, AA986886, Plap1, Slrr1c', description='asporin', other_designations='periodontal ligament-associated protein 1', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49544443, end_position_on_the_genomic_accession=49567565, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66696, CurrentID=0, Status='live', Symbol='Snx31', Aliases='4631426E05Rik', description='sorting nexin 31', other_designations='', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36504062, end_position_on_the_genomic_accession=36555614, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66700, CurrentID=0, Status='live', Symbol='Chmp3', Aliases='25.1, 4921505F14Rik, 9130011K15Rik, CGI-49, D6Ertd286e, NEDF, Vps24', description='charged multivesicular body protein 3', other_designations='chromatin-modifying protein 3|neuroendocrine differentiation factor|vacuolar protein sorting 24|vacuolar protein sorting-associated protein 24|vacuolar protein-sorting-associated protein 24', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71543802, end_position_on_the_genomic_accession=71582609, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66701, CurrentID=0, Status='live', Symbol='Spryd4', Aliases='4633402N23Rik, AA114801', description='SPRY domain containing 4', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128209908, end_position_on_the_genomic_accession=128211794, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66704, CurrentID=0, Status='live', Symbol='Rbm4b', Aliases='4921506I22Rik, AI504630, AI506404, Lark2, Neosin', description='RNA binding motif protein 4B', other_designations='RNA-binding motif protein 30|RNA-binding motif protein 4B|RNA-binding protein 30', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4755146, end_position_on_the_genomic_accession=4765940, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66705, CurrentID=0, Status='live', Symbol='Dnase1l2', Aliases='4733401H14Rik', description='deoxyribonuclease 1-like 2', other_designations='DNase I-like 2|deoxyribonuclease I-like 2', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24440087, end_position_on_the_genomic_accession=24443765, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66706, CurrentID=0, Status='live', Symbol='Ndufaf3', Aliases='4733401H18Rik', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 3', other_designations='nuclear protein E3-3|protein 2P1', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108565865, end_position_on_the_genomic_accession=108567342, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66707, CurrentID=0, Status='live', Symbol='Nkapl', Aliases='4921504I05Rik', description='NFKB activating protein-like', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21467047, end_position_on_the_genomic_accession=21468501, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66708, CurrentID=0, Status='live', Symbol='Krtap3-2', Aliases='2310020D06Rik, 4733401L16Rik, KAP3.2, KRTAP3.2', description='keratin associated protein 3-2', other_designations='krtap3-2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99555823, end_position_on_the_genomic_accession=99556853, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66711, CurrentID=0, Status='live', Symbol='Sbds', Aliases='4733401P19Rik, AI836084, CGI-97, SDS', description='Shwachman-Bodian-Diamond syndrome homolog (human)', other_designations='protein 22A3|shwachman-Bodian-Diamond syndrome protein homolog', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130245731, end_position_on_the_genomic_accession=130255462, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66712, CurrentID=0, Status='live', Symbol='Spesp1', Aliases='4921508E09Rik, AI428149', description='sperm equatorial segment protein 1', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62270729, end_position_on_the_genomic_accession=62282179, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66713, CurrentID=0, Status='live', Symbol='Actr2', Aliases='4921510D23Rik, AA409782, Arp2, D6Ertd746e', description='ARP2 actin-related protein 2', other_designations='ARP2 actin-related protein 2 homolog|actin-like protein 2', map_location='11 12.88 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=20062304, end_position_on_the_genomic_accession=20112951, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66714, CurrentID=0, Status='live', Symbol='4921524J17Rik', Aliases='', description='RIKEN cDNA 4921524J17 gene', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85408759, end_position_on_the_genomic_accession=85432841, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66715, CurrentID=0, Status='live', Symbol='Henmt1', Aliases='4921515J06Rik, Hen1, mHEN1', description='HEN1 methyltransferase homolog 1 (Arabidopsis)', other_designations='', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108938542, end_position_on_the_genomic_accession=108960776, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66716, CurrentID=0, Status='live', Symbol='Ccer1', Aliases='4921510H08Rik, AV259683', description='coiled-coil glutamate-rich protein 1', other_designations='coiled coil glutamate rich protein 1', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=97693059, end_position_on_the_genomic_accession=97694926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66717, CurrentID=0, Status='live', Symbol='Ccdc96', Aliases='4921513E08Rik', description='coiled-coil domain containing 96', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36484588, end_position_on_the_genomic_accession=36488171, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66720, CurrentID=0, Status='live', Symbol='Klhl10', Aliases='4921517C11Rik', description='kelch-like 10', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100441924, end_position_on_the_genomic_accession=100457027, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66722, CurrentID=0, Status='live', Symbol='Spag16', Aliases='4921511D23Rik, 4930524F24Rik, 4930585K05Rik, AV261009, Pf20, Wdr29', description='sperm associated antigen 16', other_designations='pf20 protein homolog|sperm-associated WD repeat protein', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=69826872, end_position_on_the_genomic_accession=70725132, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  66724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66724, CurrentID=0, Status='live', Symbol='Tab3', Aliases='4921526G09Rik, Map3k7ip3, mKIAA4135', description='TGF-beta activated kinase 1/MAP3K7 binding protein 3', other_designations='TAB-3|TAK1-binding protein 3|TGF-beta-activated kinase 1-binding protein 3|mitogen activated protein kinase kinase kinase 7 interacting protein 3|mitogen-activated protein kinase kinase kinase 7 interacting protein 3|mitogen-activated protein kinase kinase kinase 7-interacting protein 3', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=85573851, end_position_on_the_genomic_accession=85634469, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66725, CurrentID=0, Status='live', Symbol='Lrrk2', Aliases='4921513O20Rik, 9330188B09Rik, AW561911, D630001M17Rik, Gm927, cI-46', description='leucine-rich repeat kinase 2', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=91672821, end_position_on_the_genomic_accession=91816124, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  66727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66727, CurrentID=0, Status='live', Symbol='Hrasls5', Aliases='1700055A14Rik, 1700086E08Rik, 4921526K24Rik, HRSL5, iNAT', description='HRAS-like suppressor family, member 5', other_designations='Ca-independent N-acyltransferase|H-rev107-like protein 5|HRAS-like suppressor 5', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7612569, end_position_on_the_genomic_accession=7639238, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66729, CurrentID=0, Status='live', Symbol='Ankrd61', Aliases='4921520G13Rik', description='ankyrin repeat domain 61', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143890741, end_position_on_the_genomic_accession=143895152, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66732, CurrentID=0, Status='live', Symbol='4921530L21Rik', Aliases='', description='RIKEN cDNA 4921530L21 gene', other_designations='testis highly expressed protein 1', map_location='14 E2.1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=95881225, end_position_on_the_genomic_accession=95882780, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66733, CurrentID=0, Status='live', Symbol='Kcng4', Aliases='4921535I01Rik, AW049024, KV6.3, KV6.4', description='potassium voltage-gated channel, subfamily G, member 4', other_designations='voltage-gated potassium channel subunit Kv6.3|voltage-gated potassium channel subunit Kv6.4', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119623854, end_position_on_the_genomic_accession=119636353, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66734, CurrentID=0, Status='live', Symbol='Map1lc3a', Aliases='1010001H21Rik, 4922501H04Rik, LC3, LC3a', description='microtubule-associated protein 1 light chain 3 alpha', other_designations='MAP1 light chain 3-like protein 1|MAP1A/MAP1B LC3 A|MAP1A/MAP1B light chain 3 A|autophagy-related protein LC3 A|autophagy-related ubiquitin-like modifier LC3 A', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155276364, end_position_on_the_genomic_accession=155278073, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66736, CurrentID=0, Status='live', Symbol='Emc2', Aliases='4921531G14Rik, AV060620, AW209495, Ttc35', description='ER membrane protein complex subunit 2', other_designations='TPR repeat protein 35|tetratricopeptide repeat domain 35|tetratricopeptide repeat protein 35', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=43477214, end_position_on_the_genomic_accession=43527777, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66740, CurrentID=0, Status='live', Symbol='4931417E11Rik', Aliases='', description='RIKEN cDNA 4931417E11 gene', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=73468573, end_position_on_the_genomic_accession=73469667, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66742, CurrentID=0, Status='live', Symbol='Cypt1', Aliases='4931420D14Rik, Ckt1r3, Cypt5, Cypt6, TISP55', description='cysteine-rich perinuclear theca 1', other_designations='basic protein CKT1R3|cysteine-rich perinuclear theca 6|cysteine-rich postacrosomal PT protein|cysteine-rich postacrosomal perinuclear theca protein', map_location='X A1.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=16522855, end_position_on_the_genomic_accession=16523691, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66743, CurrentID=0, Status='live', Symbol='Rnf220', Aliases='4732477A13, 4931406I20Rik, 5730503K05Rik, AV302541', description='ring finger protein 220', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117271464, end_position_on_the_genomic_accession=117497047, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  66745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66745, CurrentID=0, Status='live', Symbol='Trpd52l3', Aliases='4931412G03Rik', description='tumor protein D52-like 3', other_designations='', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=30003790, end_position_on_the_genomic_accession=30006020, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66748, CurrentID=0, Status='live', Symbol='Erich2', Aliases='4933404M02Rik', description='glutamate rich 2', other_designations='', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=70508630, end_position_on_the_genomic_accession=70540885, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  66753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66753, CurrentID=0, Status='live', Symbol='Erlec1', Aliases='4933407N01Rik', description='endoplasmic reticulum lectin 1', other_designations='ER lectin|erlectin', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30865632, end_position_on_the_genomic_accession=30954173, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  66755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66755, CurrentID=0, Status='live', Symbol='4933415F23Rik', Aliases='', description='RIKEN cDNA 4933415F23 gene', other_designations='', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=23100473, end_position_on_the_genomic_accession=23102253, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66756, CurrentID=0, Status='live', Symbol='Cfap97', Aliases='1110068E21Rik, 4933411K20Rik, AW049273, Kiaa1430, mKIAA1430', description='cilia and flagella associated protein 97', other_designations='', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46151788, end_position_on_the_genomic_accession=46195590, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66757, CurrentID=0, Status='live', Symbol='Adat2', Aliases='4933426M09Rik, Deadc1', description='adenosine deaminase, tRNA-specific 2', other_designations='adenosine deaminase, tRNA-specific 2, TAD2 homolog|deaminase domain containing 1|deaminase domain-containing protein 1|tRNA-specific adenosine-34 deaminase subunit ADAT2', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13552907, end_position_on_the_genomic_accession=13563376, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66758, CurrentID=0, Status='live', Symbol='Zfp474', Aliases='4933409D10Rik, Znf474', description='zinc finger protein 474', other_designations='TSZFP|testis-specific zinc finger protein', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=52615899, end_position_on_the_genomic_accession=52639832, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66761, CurrentID=0, Status='live', Symbol='4933417A18Rik', Aliases='', description='RIKEN cDNA 4933417A18 gene', other_designations='', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34924405, end_position_on_the_genomic_accession=34956218, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66763, CurrentID=0, Status='live', Symbol='4933425L06Rik', Aliases='', description='RIKEN cDNA 4933425L06 gene', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=105082122, end_position_on_the_genomic_accession=105121782, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66765, CurrentID=0, Status='live', Symbol='4933411K16Rik', Aliases='', description='RIKEN cDNA 4933411K16 gene', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42052228, end_position_on_the_genomic_accession=42053628, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66766, CurrentID=0, Status='live', Symbol='Tmem239', Aliases='4933425O20Rik', description='transmembrane 239', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130406522, end_position_on_the_genomic_accession=130407794, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66768, CurrentID=0, Status='live', Symbol='Pacrgl', Aliases='4933428G09Rik', description='PARK2 co-regulated-like', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=48372392, end_position_on_the_genomic_accession=48388756, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66771, CurrentID=0, Status='live', Symbol='Gid4', Aliases='4933439F18Rik', description='GID complex subunit 4, VID24 homolog', other_designations='vacuolar import and degradation protein 24 homolog', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60417145, end_position_on_the_genomic_accession=60445277, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66772, CurrentID=0, Status='live', Symbol='Asb17', Aliases='4933436O18Rik', description='ankyrin repeat and SOCS box-containing 17', other_designations='ASB-17|ankyrin repeat and SOCS box-containing protein 17', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=153844247, end_position_on_the_genomic_accession=153853615, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66773, CurrentID=0, Status='live', Symbol='Gm17019', Aliases='', description='predicted gene 17019', other_designations='', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=15028950, end_position_on_the_genomic_accession=15033527, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  328918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328918, CurrentID=0, Status='live', Symbol='Zscan30', Aliases='C230097I24Rik, Gm754, Zfp397os', description='zinc finger and SCAN domain containing 30', other_designations='zinc finger protein 397 opposite strand', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=23932540, end_position_on_the_genomic_accession=23987348, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66775, CurrentID=0, Status='live', Symbol='Hacd4', Aliases='4933428I03Rik, Hcad4, Ptplad2', description='3-hydroxyacyl-CoA dehydratase 4', other_designations='protein tyrosine phosphatase-like A domain containing 2|protein tyrosine phosphatase-like protein PTPLAD2|protein-tyrosine phosphatase-like A domain-containing protein 2|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 4', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88407687, end_position_on_the_genomic_accession=88438935, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66776, CurrentID=0, Status='live', Symbol='Pisd-ps3', Aliases='4933439C20Rik', description='phosphatidylserine decarboxylase, pseudogene 3', other_designations='phosphatidylserine decarboxylase pseudogene', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  66784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66784, CurrentID=0, Status='live', Symbol='Fam187a', Aliases='4933439F11Rik', description='family with sequence similarity 187, member A', other_designations='ig-like V-type domain-containing protein ENSP00000329499 homolog', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102885169, end_position_on_the_genomic_accession=102886731, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66786, CurrentID=0, Status='live', Symbol='Olfr701', Aliases='4932441H21Rik, 4933433E02Rik, MOR283-1', description='olfactory receptor 701', other_designations='GA_x6K02T2PBJ9-9195805-9196755|olfactory receptor MOR283-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106814144, end_position_on_the_genomic_accession=106820709, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66787, CurrentID=0, Status='live', Symbol='Gskip', Aliases='4933433P14Rik', description='GSK3B interacting protein', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105685352, end_position_on_the_genomic_accession=105703058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66789, CurrentID=0, Status='live', Symbol='Alg14', Aliases='5430428G01Rik, AI854024', description='asparagine-linked glycosylation 14', other_designations='asparagine-linked glycosylation 14 homolog', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121291716, end_position_on_the_genomic_accession=121362980, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66790, CurrentID=0, Status='live', Symbol='Grtp1', Aliases='5430401C05Rik, C81211, Tbc1d6', description='GH regulated TBC protein 1', other_designations='TBC1 domain family member 6', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13176869, end_position_on_the_genomic_accession=13200654, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66793, CurrentID=0, Status='live', Symbol='Efcab1', Aliases='4930449A16Rik, 5430404L10Rik', description='EF hand calcium binding domain 1', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14906257, end_position_on_the_genomic_accession=14969061, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  66795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66795, CurrentID=0, Status='live', Symbol='Atg10', Aliases='5330424L23Rik, 5430428K15Rik, AI852123, APG10, Agp10, Apg10l, Apg10pl, Atg10', description='autophagy related 10', other_designations='autophagy-related 10|autophagy-related protein 10|mAPG10', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=90935348, end_position_on_the_genomic_accession=91224792, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66797, CurrentID=0, Status='live', Symbol='Cntnap2', Aliases='5430425M22Rik, Caspr2, mKIAA0868', description='contactin associated protein-like 2', other_designations='cell recognition molecule Caspr2', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=45060061, end_position_on_the_genomic_accession=47301388, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  66799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66799, CurrentID=0, Status='live', Symbol='Ube2w', Aliases='6130401J04Rik', description='ubiquitin-conjugating enzyme E2W (putative)', other_designations='E2 ubiquitin-conjugating enzyme W|N-terminal E2 ubiquitin-conjugating enzyme|N-terminus-conjugating E2|probable ubiquitin-conjugating enzyme E2 W|ubiquitin carrier protein W|ubiquitin-protein ligase W', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16540788, end_position_on_the_genomic_accession=16619420, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66801, CurrentID=0, Status='live', Symbol='Prkrip1', Aliases='8430424D23Rik, C114', description='Prkr interacting protein 1 (IL11 inducible)', other_designations='', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136180357, end_position_on_the_genomic_accession=136198954, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  328949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328949, CurrentID=0, Status='live', Symbol='Mcc', Aliases='D18Ertd451e, E330037C19', description='mutated in colorectal cancers', other_designations='', map_location='18 23.74 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44425060, end_position_on_the_genomic_accession=44812182, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  66808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66808, CurrentID=0, Status='live', Symbol='9030624G23Rik', Aliases='', description='RIKEN cDNA 9030624G23 gene', other_designations='', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24043202, end_position_on_the_genomic_accession=24097269, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66809, CurrentID=0, Status='live', Symbol='Krt20', Aliases='9030623C06Rik, Ck-20, Ck20, K20, Krt21', description='keratin 20', other_designations='cytokeratin 20', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99428403, end_position_on_the_genomic_accession=99438153, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66810, CurrentID=0, Status='live', Symbol='Rbm22', Aliases='8430430L24Rik', description='RNA binding motif protein 22', other_designations='RNA-binding motif protein 22', map_location='18 D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60560786, end_position_on_the_genomic_accession=60572729, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66811, CurrentID=0, Status='live', Symbol='Duoxa2', Aliases='9030623N16Rik', description='dual oxidase maturation factor 2', other_designations='Numb-interacting protein 2', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122298900, end_position_on_the_genomic_accession=122302885, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66812, CurrentID=0, Status='live', Symbol='Ppcdc', Aliases='1810057I13Rik, 8430432M10Rik', description='phosphopantothenoylcysteine decarboxylase', other_designations='PPC-DC|coaC', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57385095, end_position_on_the_genomic_accession=57440175, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66813, CurrentID=0, Status='live', Symbol='Bcl2l14', Aliases='4930452K23Rik, 4933405K19Rik, 9030625M01Rik, AI429211, Bclg', description='BCL2-like 14 (apoptosis facilitator)', other_designations='apoptosis regulator BCL-G|bcl2-L-14', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134396329, end_position_on_the_genomic_accession=134438741, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66815, CurrentID=0, Status='live', Symbol='Ccdc109b', Aliases='9030408N13Rik, MCUb', description='coiled-coil domain containing 109B', other_designations='coiled-coil domain-containing protein 109B|mitochondrial calcium uniporter regulatory subunit MCUb', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129914960, end_position_on_the_genomic_accession=129970214, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66816, CurrentID=0, Status='live', Symbol='Thap2', Aliases='2900040O07Rik, 9030625G08Rik, AI450385, AI649097', description='THAP domain containing, apoptosis associated protein 2', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115369966, end_position_on_the_genomic_accession=115384435, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66817, CurrentID=0, Status='live', Symbol='Tmem170', Aliases='9030409E16Rik, AA403674, D8Bwg1414ea, Tmem170', description='transmembrane protein 170', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111863038, end_position_on_the_genomic_accession=111876863, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66818, CurrentID=0, Status='live', Symbol='Smim7', Aliases='9130011J15Rik', description='small integral membrane protein 7', other_designations='UPF0608 protein C19orf42 homolog', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72565198, end_position_on_the_genomic_accession=72571048, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66821, CurrentID=0, Status='live', Symbol='Bcs1l', Aliases='9130022O19Rik', description='BCS1-like (yeast)', other_designations='BCS1-like protein', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74588294, end_position_on_the_genomic_accession=74592443, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66822, CurrentID=0, Status='live', Symbol='Fbxo25', Aliases='9130015I06Rik, AI649137, Fbx25', description='F-box protein 25', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13907740, end_position_on_the_genomic_accession=13940544, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  328967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328967, CurrentID=0, Status='live', Symbol='Arhgef37', Aliases='4933429F08Rik', description='Rho guanine nucleotide exchange factor (GEF) 37', other_designations='probable guanine nucleotide exchange factor FLJ41603 homolog', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61493625, end_position_on_the_genomic_accession=61537044, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66824, CurrentID=0, Status='live', Symbol='Pycard', Aliases='9130417A21Rik, Asc, CARD5, TMS-1, TNS1, masc', description='PYD and CARD domain containing', other_designations='PYD and CARD domain-containing protein', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127991373, end_position_on_the_genomic_accession=127993867, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66825, CurrentID=0, Status='live', Symbol='Rnf186', Aliases='9130020G10Rik', description='ring finger protein 186', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138967112, end_position_on_the_genomic_accession=138968366, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66826, CurrentID=0, Status='live', Symbol='Taz', Aliases='5031411C02Rik, 9130012G04Rik, AW107266, AW552613, G4.5', description='tafazzin', other_designations='', map_location='X 37.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74281900, end_position_on_the_genomic_accession=74295319, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  66827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66827, CurrentID=0, Status='live', Symbol='Ttc1', Aliases='4833412C19Rik, C79328, TPR1', description='tetratricopeptide repeat domain 1', other_designations='TPR repeat protein 1', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43730006, end_position_on_the_genomic_accession=43747973, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66830, CurrentID=0, Status='live', Symbol='Nacc1', Aliases='2010001H03Rik, 4930511N13Rik, Btbd14b, Nac1', description='nucleus accumbens associated 1, BEN and BTB (POZ) domain containing', other_designations='BTB (POZ) domain containing 14B|BTB/POZ domain-containing protein 14B|NAC-1|transcriptional repressor NAC1', map_location='8 41.02 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84670476, end_position_on_the_genomic_accession=84687985, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66832, CurrentID=0, Status='live', Symbol='Rsph3a', Aliases='1700012G05Rik, 4930524H12Rik, R74860, Rshl2, Rshl2a', description='radial spoke 3A homolog (Chlamydomonas)', other_designations='A-kinase anchor protein RSPH3A|radial spoke head-like protein 2A|radial spokehead-like 2|radial spokehead-like 2A', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=7945614, end_position_on_the_genomic_accession=7979556, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  328977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=328977, CurrentID=0, Status='live', Symbol='Zfp532', Aliases='C530030I18Rik, Znf532, mKIAA1629', description='zinc finger protein 532', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65580230, end_position_on_the_genomic_accession=65689443, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  66834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66834, CurrentID=0, Status='live', Symbol='Acot13', Aliases='0610006O17Rik, Them2', description='acyl-CoA thioesterase 13', other_designations='thioesterase superfamily member 2', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24817948, end_position_on_the_genomic_accession=24831564, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66836, CurrentID=0, Status='live', Symbol='Tmem223', Aliases='0610006I08Rik', description='transmembrane protein 223', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8770996, end_position_on_the_genomic_accession=8772475, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66839, CurrentID=0, Status='live', Symbol='0610009O20Rik', Aliases='2700004E22Rik, Dele, Kiaa0141', description='RIKEN cDNA 0610009O20 gene', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38250249, end_position_on_the_genomic_accession=38262629, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66840, CurrentID=0, Status='live', Symbol='Wdr45b', Aliases='0610008N23Rik, AA408735, D16Bwg0193e, WIPI-3, Wdr45l', description='WD repeat domain 45B', other_designations='WD repeat-containing protein 45-like|WD repeat-containing protein 45B|WDR45-like protein|Wdr45 like', map_location='11 85.24 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121327203, end_position_on_the_genomic_accession=121354447, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66841, CurrentID=0, Status='live', Symbol='Etfdh', Aliases='0610010I20Rik, AV001013', description='electron transferring flavoprotein, dehydrogenase', other_designations='ETF dehydrogenase|ETF-QO|ETF-ubiquinone oxidoreductase|electron-transferring-flavoprotein dehydrogenase', map_location='3 34.91 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79603788, end_position_on_the_genomic_accession=79628767, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66844, CurrentID=0, Status='live', Symbol='Ormdl2', Aliases='0610012C09Rik', description='ORM1-like 2 (S. cerevisiae)', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128817457, end_position_on_the_genomic_accession=128821632, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66845, CurrentID=0, Status='live', Symbol='Mrpl33', Aliases='0610009M10Rik, L33mt, MRP-L33', description='mitochondrial ribosomal protein L33', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31613951, end_position_on_the_genomic_accession=31622644, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66847, CurrentID=0, Status='live', Symbol='Hint3', Aliases='0610010I17Rik, AV009015, HINT-3, HINT-4, HINT4', description='histidine triad nucleotide binding protein 3', other_designations='histidine triad protein 4', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=30608207, end_position_on_the_genomic_accession=30629401, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66848, CurrentID=0, Status='live', Symbol='Fuca2', Aliases='0610025O11Rik, 5530401P20Rik', description='fucosidase, alpha-L- 2, plasma', other_designations='alpha-L-fucosidase 2|alpha-L-fucoside fucohydrolase 2', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13501006, end_position_on_the_genomic_accession=13517275, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66849, CurrentID=0, Status='live', Symbol='Ppp1r2', Aliases='0610025N14Rik, 2310007G06Rik, 4930440J04Rik, 5430408E15Rik, D16Ertd248e, IPP-2', description='protein phosphatase 1, regulatory (inhibitor) subunit 2', other_designations='', map_location='16 21.41 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31251541, end_position_on_the_genomic_accession=31275277, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66853, CurrentID=0, Status='live', Symbol='Pnpla2', Aliases='0610039C21Rik, 1110001C14Rik, Atgl, TTS-2.2', description='patatin-like phospholipase domain containing 2', other_designations='adipose triglyceride lipase|desnutrin|transport-secretion protein|triglyceride hydrolase', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141455188, end_position_on_the_genomic_accession=141460743, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66854, CurrentID=0, Status='live', Symbol='Trim35', Aliases='0710005M05Rik, A430106H13Rik, AW046487, HLS5, Mair, NC8, mKIAA1098', description='tripartite motif-containing 35', other_designations='hemopoietic lineage switch protein 5|macrophage-derived apoptosis-inducing RBCC protein|protein Nc8', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66297025, end_position_on_the_genomic_accession=66311424, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66855, CurrentID=0, Status='live', Symbol='Tcf25', Aliases='1100001J13Rik, 1810041K11Rik, D8Ertd325e, Nulp1, mKIAA1049', description='transcription factor 25 (basic helix-loop-helix)', other_designations='nuclear localized protein 1', map_location='8 72.1 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123373635, end_position_on_the_genomic_accession=123404174, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  66857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66857, CurrentID=0, Status='live', Symbol='Plbd1', Aliases='1100001H23Rik', description='phospholipase B domain containing 1', other_designations='LAMA-like protein 1|lamina ancestor homolog 1|phospholipase B domain-containing protein 1|putative phospholipase B-like 1', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136612071, end_position_on_the_genomic_accession=136662269, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  329002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329002, CurrentID=0, Status='live', Symbol='Zfp236', Aliases='AI447957, Gm335', description='zinc finger protein 236', other_designations='', map_location='18 56.26 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=82593593, end_position_on_the_genomic_accession=82693627, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  329003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329003, CurrentID=0, Status='live', Symbol='Zfp516', Aliases='C330029B10Rik, D230016L03, Zfp26l, Znf516, mKIAA0222', description='zinc finger protein 516', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=82910137, end_position_on_the_genomic_accession=83005314, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  66860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66860, CurrentID=0, Status='live', Symbol='Tanc1', Aliases='1200003E16Rik, Tanc, mKIAA1728', description='tetratricopeptide repeat, ankyrin repeat and coiled-coil containing 1', other_designations='TPR domain, ankyrin-repeat and coiled-coil-containing protein|tetratricopeptide repeat, ankyrin repeat and coiled-coil domain-containing protein 1', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=59611848, end_position_on_the_genomic_accession=59846213, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  66861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66861, CurrentID=0, Status='live', Symbol='Dnajc10', Aliases='1200006L06Rik, D2Ertd706e, ERdj5, JPDI', description='DnaJ heat shock protein family (Hsp40) member C10', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 10|ER-resident protein ERdj5|endoplasmic reticulum DNA J domain-containing protein 5|endoplasmic reticulum DnaJ-PDI fusion protein 1|j domain-containing PDI-like protein|j domain-containing protein disulfide isomerase-like protein', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=80315466, end_position_on_the_genomic_accession=80354055, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  66863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66863, CurrentID=0, Status='live', Symbol='Lztr1', Aliases='1200003E21Rik, AI591627, AW550890', description='leucine-zipper-like transcriptional regulator, 1', other_designations='LZTR-1', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17508541, end_position_on_the_genomic_accession=17526333, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  66864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66864, CurrentID=0, Status='live', Symbol='Clec14a', Aliases='1200003C23Rik, AI642649', description='C-type lectin domain family 14, member a', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=58264720, end_position_on_the_genomic_accession=58269258, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66865, CurrentID=0, Status='live', Symbol='Pmpca', Aliases='1200002L24Rik, 4933435E07Rik, Alpha-MPP, P-55', description='peptidase (mitochondrial processing) alpha', other_designations='mitochondrial matrix processing protease, alpha subunit', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26389348, end_position_on_the_genomic_accession=26397121, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  66866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66866, CurrentID=0, Status='live', Symbol='Nhlrc2', Aliases='1200003G01Rik, AI835049, AV002846, AW496455', description='NHL repeat containing 2', other_designations='', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56548261, end_position_on_the_genomic_accession=56603503, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66867, CurrentID=0, Status='live', Symbol='Hmg20a', Aliases='1200004E06Rik, 5730490E10Rik, Hmgxb1, Ibraf', description='high mobility group 20A', other_designations='HMG box-containing protein 20A|HMG domain-containing protein HMGX1|inhibitor of BRAF35', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56418628, end_position_on_the_genomic_accession=56496936, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66868, CurrentID=0, Status='live', Symbol='Mfsd1', Aliases='1200003O06Rik', description='major facilitator superfamily domain containing 1', other_designations='', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=67582768, end_position_on_the_genomic_accession=67604231, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  66869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66869, CurrentID=0, Status='live', Symbol='Zfp869', Aliases='1200003I07Rik', description='zinc finger protein 869', other_designations='U 2-7-1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69705137, end_position_on_the_genomic_accession=69716989, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66870, CurrentID=0, Status='live', Symbol='Serbp1', Aliases='1200009K13Rik, 9330147J08Rik, AL022786, Pairbp1', description='serpine1 mRNA binding protein 1', other_designations='PAI-1 mRNA-binding protein|PAI-RBP1|PAI1 RNA-binding protein 1|SERPINE1 mRNA-binding protein 1', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67266979, end_position_on_the_genomic_accession=67289302, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66871, CurrentID=0, Status='live', Symbol='Cpne8', Aliases='1200003E11Rik, 1500031E20Rik', description='copine VIII', other_designations='copine 8', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=90487481, end_position_on_the_genomic_accession=90680800, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  66873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66873, CurrentID=0, Status='live', Symbol='Tril', Aliases='1200009O22Rik, AI316813, mKIAA0644', description='TLR4 interactor with leucine-rich repeats', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=53815468, end_position_on_the_genomic_accession=53820825, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66874, CurrentID=0, Status='live', Symbol='Ncbp3', Aliases='1200014J11Rik, C130061O14Rik, C78393', description='nuclear cap binding subunit 3', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73047867, end_position_on_the_genomic_accession=73083579, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66875, CurrentID=0, Status='live', Symbol='Swt1', Aliases='1200016B10Rik, C1orf26', description='SWT1 RNA endoribonuclease homolog (S. cerevisiae)', other_designations='', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151367698, end_position_on_the_genomic_accession=151428450, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  66877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66877, CurrentID=0, Status='live', Symbol='Crnkl1', Aliases='1200013P10Rik, 5730590A01Rik, C80326, crn', description='Crn, crooked neck-like 1 (Drosophila)', other_designations='crooked neck homolog', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=145917482, end_position_on_the_genomic_accession=145934700, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66878, CurrentID=0, Status='live', Symbol='Riok3', Aliases='1200013N13Rik, D18Ertd331e, E130306C24Rik, Sudd', description='RIO kinase 3', other_designations='sudD, suppressor of bimD6 homolog', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12128850, end_position_on_the_genomic_accession=12157367, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66880, CurrentID=0, Status='live', Symbol='Rsrc1', Aliases='1200013F24Rik, SRrp53', description='arginine/serine-rich coiled-coil 1', other_designations='arginine/serine-rich coiled-coil protein 1', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=66985579, end_position_on_the_genomic_accession=67358403, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  66881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66881, CurrentID=0, Status='live', Symbol='Pcyox1', Aliases='1200015P13Rik, AI115532, AI448340, PCL1, Pcly', description='prenylcysteine oxidase 1', other_designations='prenylcysteine lyase', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86386006, end_position_on_the_genomic_accession=86397158, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66882, CurrentID=0, Status='live', Symbol='Bzw1', Aliases='1200015E15Rik, mKIAA0005', description='basic leucine zipper and W2 domains 1', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58393136, end_position_on_the_genomic_accession=58407354, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66884, CurrentID=0, Status='live', Symbol='Appbp2', Aliases='1300003O07Rik, AI465480, PAT1', description='amyloid beta precursor protein (cytoplasmic tail) binding protein 2', other_designations='APP-BP2|amyloid beta precursor protein-binding protein 2', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=85191308, end_position_on_the_genomic_accession=85235120, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66885, CurrentID=0, Status='live', Symbol='Acadsb', Aliases='1300003O09Rik, 2-MEBCAD, BB066609, SBCAD', description='acyl-Coenzyme A dehydrogenase, short/branched chain', other_designations='2-methyl branched chain acyl-CoA dehydrogenase|2-methylbutyryl-coenzyme A dehydrogenase', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131410601, end_position_on_the_genomic_accession=131446211, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66887, CurrentID=0, Status='live', Symbol='Lonp2', Aliases='1300002A08Rik, AU015403, Lonp', description='lon peptidase 2, peroxisomal', other_designations='lon protease 2|lon protease-like protein 2|peroxisomal Lon protease homolog 2', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=86624043, end_position_on_the_genomic_accession=86716636, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  66889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66889, CurrentID=0, Status='live', Symbol='Rnf128', Aliases='1300002C13Rik, AI987883, GRAIL, Greul1', description='ring finger protein 128', other_designations='gene related to anergy in lymphocytes protein|goliath-related E3 ubiquitin-protein ligase 1', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139563340, end_position_on_the_genomic_accession=139673145, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66890, CurrentID=0, Status='live', Symbol='Lman2', Aliases='1110003H06Rik, 1300009F09Rik, AA408240, AL023023, AU040819, GP36B, VIP36', description='lectin, mannose-binding 2', other_designations='vesicular integral-membrane protein 36', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55343833, end_position_on_the_genomic_accession=55362783, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66892, CurrentID=0, Status='live', Symbol='Eif4e3', Aliases='1300018P11Rik, AI451927, eIF4E-3', description='eukaryotic translation initiation factor 4E member 3', other_designations='eIF-4E type 3|eIF-4E3|eIF4E type 3|eukaryotic translation initiation factor-3', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=99625137, end_position_on_the_genomic_accession=99666771, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  66894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66894, CurrentID=0, Status='live', Symbol='Wwp2', Aliases='1300010O06Rik, AA690238, AIP2, AW554328', description='WW domain containing E3 ubiquitin protein ligase 2', other_designations='WW domain-containing protein 2|WW domain-containing protein 4', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107436368, end_position_on_the_genomic_accession=107558597, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  66895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66895, CurrentID=0, Status='live', Symbol='Pxdc1', Aliases='1300014I06Rik', description='PX domain containing 1', other_designations='PX domain-containing protein C6orf145 homolog', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34627840, end_position_on_the_genomic_accession=34653133, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66897, CurrentID=0, Status='live', Symbol='Naa16', Aliases='1300019C06Rik, Narg1l', description='N(alpha)-acetyltransferase 16, NatA auxiliary subunit', other_designations='NARG1-like protein|NMDA receptor-regulated 1-like protein|Nalpha acetyltransferase 16', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79334507, end_position_on_the_genomic_accession=79390777, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  66898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66898, CurrentID=0, Status='live', Symbol='Baiap2l1', Aliases='1300006M19Rik, AI585895, IRTKS', description='BAI1-associated protein 2-like 1', other_designations='BAI1-associated protein 2-like protein 1|insulin receptor tyrosine kinase substrate', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144264525, end_position_on_the_genomic_accession=144358112, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66899, CurrentID=0, Status='live', Symbol='Fip1l1', Aliases='1300019H17Rik, Rje', description='FIP1 like 1 (S. cerevisiae)', other_designations='FIP1-like 1 protein', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=74535356, end_position_on_the_genomic_accession=74598790, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  66901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66901, CurrentID=0, Status='live', Symbol='Proz', Aliases='1300015B06Rik', description='protein Z, vitamin K-dependent plasma glycoprotein', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13060582, end_position_on_the_genomic_accession=13076026, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66902, CurrentID=0, Status='live', Symbol='Mtap', Aliases='1300019I21Rik, MSAP', description='methylthioadenosine phosphorylase', other_designations="5'-methylthioadenosine phosphorylase|MTA phosphorylase|MTAP|MTAPase", map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=89137370, end_position_on_the_genomic_accession=89181090, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66904, CurrentID=0, Status='live', Symbol='Pccb', Aliases='1300012P06Rik, AI314687, R74805', description='propionyl Coenzyme A carboxylase, beta polypeptide', other_designations='PCCase subunit beta|propanoyl-CoA:carbon dioxide ligase subunit beta', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=100982032, end_position_on_the_genomic_accession=101034940, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66905, CurrentID=0, Status='live', Symbol='Plin3', Aliases='1300012C15Rik, M6prbp1, Tip47', description='perilipin 3', other_designations='cargo selection protein TIP47|mannose-6-phosphate receptor binding protein 1|mannose-6-phosphate receptor-binding protein 1', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56278962, end_position_on_the_genomic_accession=56290511, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66910, CurrentID=0, Status='live', Symbol='Tmem107', Aliases='1110004B13Rik, 2810049P21Rik', description='transmembrane protein 107', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69070809, end_position_on_the_genomic_accession=69073293, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66911, CurrentID=0, Status='live', Symbol='Nudt16l1', Aliases='1110001K21Rik, 5330437I08Rik, Sdos', description='nudix (nucleoside diphosphate linked moiety X)-type motif 16-like 1', other_designations='NUDT16-like protein 1', map_location='16 2.48 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4939111, end_position_on_the_genomic_accession=4941021, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66912, CurrentID=0, Status='live', Symbol='Bzw2', Aliases='1110001I24Rik, Bdm2, HSPC028', description='basic leucine zipper and W2 domains 2', other_designations='brain development-related molecule 2', map_location='12 B2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=36091835, end_position_on_the_genomic_accession=36156825, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  66913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66913, CurrentID=0, Status='live', Symbol='Kdelr2', Aliases='1110007A14Rik', description='KDEL (Lys-Asp-Glu-Leu) endoplasmic reticulum protein retention receptor 2', other_designations='ER lumen protein retaining receptor 2|KDEL endoplasmic reticulum protein retention receptor 2|KDEL receptor 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143403820, end_position_on_the_genomic_accession=143421904, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66914, CurrentID=0, Status='live', Symbol='Vps28', Aliases='1110014J03Rik, CIIA, D730005C08Rik', description='vacuolar protein sorting 28 (yeast)', other_designations='ESCRT-I complex subunit VPS28|caspase-activated DNase inhibitor that interacts with ASK1', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76622088, end_position_on_the_genomic_accession=76626027, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66915, CurrentID=0, Status='live', Symbol='Myeov2', Aliases='1110002M09Rik', description='myeloma overexpressed 2', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92637145, end_position_on_the_genomic_accession=92641985, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  66916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66916, CurrentID=0, Status='live', Symbol='Ndufb7', Aliases='1110002H15Rik', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 7', other_designations='CI-B18|NADH-ubiquinone oxidoreductase B18 subunit|complex I-B18', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83566758, end_position_on_the_genomic_accession=83571623, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  66917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66917, CurrentID=0, Status='live', Symbol='Chordc1', Aliases='1110001O09Rik, AA409036, Chp-1, morgana', description='cysteine and histidine-rich domain (CHORD)-containing, zinc-binding protein 1', other_designations='CHORD domain-containing protein 1|CHORD-containing protein 1|protein morgana', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18292233, end_position_on_the_genomic_accession=18314000, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  329064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329064, CurrentID=0, Status='live', Symbol='Pkd2l1', Aliases='B830002B15, BC046386, PCL, PKD2L, Pkdl, TRPP3', description='polycystic kidney disease 2-like 1', other_designations='polycystin-2 homolog|polycystin-L', map_location='19 36.91 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44147637, end_position_on_the_genomic_accession=44248354, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  329065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329065, CurrentID=0, Status='live', Symbol='Scd4', Aliases='ACOD4', description='stearoyl-coenzyme A desaturase 4', other_designations='delta(9)-desaturase 4|delta-9 desaturase 4|fatty acid desaturase 4|stearoyl-CoA desaturase 4', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44333326, end_position_on_the_genomic_accession=44346743, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  66922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66922, CurrentID=0, Status='live', Symbol='Rras2', Aliases='2610016H24Rik, C86394, TC21', description='related RAS viral (r-ras) oncogene homolog 2', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114046782, end_position_on_the_genomic_accession=114117781, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66923, CurrentID=0, Status='live', Symbol='Pbrm1', Aliases='2610016F04Rik, AI507524, BAF180, Pb1', description='polybromo 1', other_designations='2310032M22Rik|BRG1-associated factor 180', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31019138, end_position_on_the_genomic_accession=31121592, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  66925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66925, CurrentID=0, Status='live', Symbol='Sdhd', Aliases='3110001M13Rik, AVLL5809, C78570, PRO19626', description='succinate dehydrogenase complex, subunit D, integral membrane protein', other_designations='CII-4|QPs3|cybS|succinate dehydrogenase complex subunit D|succinate-ubiquinone oxidoreductase cytochrome b small subunit|succinate-ubiquinone reductase membrane anchor subunit', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50596340, end_position_on_the_genomic_accession=50603849, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66926, CurrentID=0, Status='live', Symbol='Trmt6', Aliases='3300001M20Rik, AU016412, AW492787, CGI-09, mKIAA1153', description='tRNA methyltransferase 6', other_designations='tRNA (adenine-N(1)-)-methyltransferase non-catalytic subunit TRM6|tRNA methyltransferase 6 homolog|tRNA(m1A58)-methyltransferase subunit TRM6|tRNA(m1A58)MTase subunit TRM6', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132804207, end_position_on_the_genomic_accession=132816054, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66928, CurrentID=0, Status='live', Symbol='Tmem261', Aliases='1700027K24Rik, 3110001D03Rik', description='transmembrane protein 261', other_designations='transmembrane protein C9orf123 homolog', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=75277354, end_position_on_the_genomic_accession=75278286, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  66929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66929, CurrentID=0, Status='live', Symbol='Asf1b', Aliases='1700003K02Rik, AA409591', description='anti-silencing function 1B histone chaperone', other_designations='ASF1 anti-silencing function 1 homolog B|anti-silencing function protein 1 homolog B|mCIA-II', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83955694, end_position_on_the_genomic_accession=83970195, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66930, CurrentID=0, Status='live', Symbol='Fank1', Aliases='1700007B22Rik, AI850911', description='fibronectin type 3 and ankyrin repeat domains 1', other_designations='GSG1|germ cell-specific 1|germ cell-specific gene 1 protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133776856, end_position_on_the_genomic_accession=133881536, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66931, CurrentID=0, Status='live', Symbol='1700010I14Rik', Aliases='', description='RIKEN cDNA 1700010I14 gene', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8988320, end_position_on_the_genomic_accession=9008319, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66932, CurrentID=0, Status='live', Symbol='Rexo1', Aliases='1700021P10Rik, 2610511M11Rik, AW050347, EloA-BP1, Tceb3bp1', description='REX1, RNA exonuclease 1 homolog (S. cerevisiae)', other_designations='elongin A binding protein 1|transcription elongation factor B polypeptide 3 binding protein 1|transcription elongation factor B polypeptide 3-binding protein 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80540926, end_position_on_the_genomic_accession=80563381, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  66934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66934, CurrentID=0, Status='live', Symbol='Dsn1', Aliases='1700022L09Rik, AW552447', description='DSN1 homolog, MIS12 kinetochore complex component', other_designations='DSN1, MIND kinetochore complex component, homolog', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156995062, end_position_on_the_genomic_accession=157007075, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  66935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66935, CurrentID=0, Status='live', Symbol='Cir1', Aliases='1700023B02Rik, 2810021A19Rik, Cicr, Cir', description='corepressor interacting with RBPJ, 1', other_designations='CBF1 interacting corepressor|CBF1-interacting corepressor', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73283872, end_position_on_the_genomic_accession=73312592, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66938, CurrentID=0, Status='live', Symbol='Sh3d21', Aliases='1700029G01Rik, AU019788', description='SH3 domain containing 21', other_designations='SH3 domain-containing protein C1orf113 homolog', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126150602, end_position_on_the_genomic_accession=126163548, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  66939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66939, CurrentID=0, Status='live', Symbol='Aagab', Aliases='2310007F21Rik, p34', description='alpha- and gamma-adaptin binding protein', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=63602629, end_position_on_the_genomic_accession=63641897, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66940, CurrentID=0, Status='live', Symbol='Shisa5', Aliases='2310008D10Rik, 6430628I05Rik, Scotin, mShisa5', description='shisa family member 5', other_designations='shisa homolog 5', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109038567, end_position_on_the_genomic_accession=109057792, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66942, CurrentID=0, Status='live', Symbol='Ddx18', Aliases='2310005B10Rik', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 18', other_designations='DEAD box protein 18', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=121553830, end_position_on_the_genomic_accession=121567980, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66943, CurrentID=0, Status='live', Symbol='Pqlc1', Aliases='2310009N05Rik, 4933425L21Rik, 5730564E11Rik, C78974', description='PQ loop repeat containing 1', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80255241, end_position_on_the_genomic_accession=80292724, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66945, CurrentID=0, Status='live', Symbol='Sdha', Aliases='1500032O14Rik, 2310034D06Rik, 4921513A11, C81073, FP, SDH2, SDHF', description='succinate dehydrogenase complex, subunit A, flavoprotein (Fp)', other_designations='flavoprotein subunit of complex II', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74322255, end_position_on_the_genomic_accession=74350240, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66948, CurrentID=0, Status='live', Symbol='Acad8', Aliases='2310016C19Rik, AI786953', description='acyl-Coenzyme A dehydrogenase family, member 8', other_designations='ACAD-8|acyl-CoA dehydrogenase family member 8', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=26974138, end_position_on_the_genomic_accession=26999576, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66949, CurrentID=0, Status='live', Symbol='Trim59', Aliases='2310035M22Rik, 2700022F13Rik, Mrf1, TSBF1', description='tripartite motif-containing 59', other_designations='RING finger 1|RING finger protein 1|tumor suppressor TSBF1 homolog', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69035292, end_position_on_the_genomic_accession=69046404, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  66950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66950, CurrentID=0, Status='live', Symbol='Tmem206', Aliases='2310028N02Rik, AI118576', description='transmembrane protein 206', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191325965, end_position_on_the_genomic_accession=191352925, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66952, CurrentID=0, Status='live', Symbol='2310030G06Rik', Aliases='2610024D04Rik, AV017521', description='RIKEN cDNA 2310030G06 gene', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50739691, end_position_on_the_genomic_accession=50746521, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  66953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66953, CurrentID=0, Status='live', Symbol='Cdca7', Aliases='2310021G01Rik, AU044109, JPO1', description='cell division cycle associated 7', other_designations='c-Myc target JPO1', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=72476189, end_position_on_the_genomic_accession=72486892, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66957, CurrentID=0, Status='live', Symbol='Serpinb11', Aliases='2310046M08Rik, AU015399', description='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 11', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 11', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107362314, end_position_on_the_genomic_accession=107380475, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  66958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66958, CurrentID=0, Status='live', Symbol='Tmx2', Aliases='2310042M24Rik, AA589631, Txndc14', description='thioredoxin-related transmembrane protein 2', other_designations='thioredoxin domain containing 14|thioredoxin domain-containing protein 14', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84671311, end_position_on_the_genomic_accession=84678828, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66959, CurrentID=0, Status='live', Symbol='Dusp26', Aliases='2310043K02Rik, Skrp3', description='dual specificity phosphatase 26 (putative)', other_designations='dual specificity phosphatase SKRP3|dual-specificity phosphatase SKRP3', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=31089662, end_position_on_the_genomic_accession=31097047, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66960, CurrentID=0, Status='live', Symbol='Fam188a', Aliases='1810041E18Rik, 2310047O13Rik, 5830410F13Rik, AI447827, AW111958', description='family with sequence similarity 188, member A', other_designations='caspase recruitment domain containing protein', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=12347263, end_position_on_the_genomic_accession=12419490, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  66961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66961, CurrentID=0, Status='live', Symbol='Neat1', Aliases='2310043N10Rik, VINC', description='nuclear paraspeckle assembly transcript 1 (non-protein coding)', other_designations='virus-inducible ncRNA', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5842302, end_position_on_the_genomic_accession=5845478, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  66962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66962, CurrentID=0, Status='live', Symbol='Swsap1', Aliases='2310047B19Rik', description='SWIM type zinc finger 7 associated protein 1', other_designations='SWIM-type zinc finger 7 associated protein 1', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21955753, end_position_on_the_genomic_accession=21958270, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66964, CurrentID=0, Status='live', Symbol='Golt1b', Aliases='2310061A22Rik, AA407874, AU024291, CGI-141', description='golgi transport 1B', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142387209, end_position_on_the_genomic_accession=142410084, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  66965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66965, CurrentID=0, Status='live', Symbol='Ctu2', Aliases='2310061F22Rik, C16orf84, Ncs2', description='cytosolic thiouridylase subunit 2', other_designations='cytosolic thiouridylase subunit 2 homolog', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122476137, end_position_on_the_genomic_accession=122483092, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  66966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66966, CurrentID=0, Status='live', Symbol='Trit1', Aliases='2310075G14Rik, AI314189, AI314635, AV099619, IPT, MOD5', description='tRNA isopentenyltransferase 1', other_designations='IPP transferase|IPPT|IPTase|isopentenyl-diphosphate:tRNA isopentenyltransferase|tRNA isopentenyltransferase, mitochondrial', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123016530, end_position_on_the_genomic_accession=123054940, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66967, CurrentID=0, Status='live', Symbol='Edem3', Aliases='2310050N11Rik', description='ER degradation enhancer, mannosidase alpha-like 3', other_designations='ER degradation-enhancing alpha-mannosidase-like 3|alpha-1,2-mannosidase EDEM3', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151755344, end_position_on_the_genomic_accession=151822328, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  66968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66968, CurrentID=0, Status='live', Symbol='Plin5', Aliases='2310076L09Rik, AI415325, AW109675, Lsdp5, MLDP, PAT-1', description='perilipin 5', other_designations='lipid droplet associated protein|lipid droplet-associated protein PAT-1|lipid storage droplet protein 5|myocardial LD protein', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56111597, end_position_on_the_genomic_accession=56117606, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  66970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66970, CurrentID=0, Status='live', Symbol='Ssbp2', Aliases='1500004K09Rik, 2310079I02Rik, 9330163K02Rik, A830008M03Rik, AU067692, AW558684, Hspc116, SSDP2', description='single-stranded DNA binding protein 2', other_designations='sequence-specific single-stranded-DNA-binding protein 2', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=91460979, end_position_on_the_genomic_accession=91706175, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  66971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66971, CurrentID=0, Status='live', Symbol='Cdk5rap1', Aliases='2310066P17Rik', description='CDK5 regulatory subunit associated protein 1', other_designations='CDK5 activator-binding protein C42', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154289809, end_position_on_the_genomic_accession=154372850, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66972, CurrentID=0, Status='live', Symbol='Slc25a23', Aliases='2310067G05Rik, SCaMC-3', description='solute carrier family 25 (mitochondrial carrier; phosphate carrier), member 23', other_designations='small calcium-binding mitochondrial carrier protein 3|solute carrier family 25 member 23', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57043711, end_position_on_the_genomic_accession=57059880, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66973, CurrentID=0, Status='live', Symbol='Mrps18b', Aliases='2400002C15Rik', description='mitochondrial ribosomal protein S18B', other_designations='28S ribosomal protein S18-2, mitochondrial|MRP-S18-2|MRP-S18-b|S18mt-b|mrps18-b', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35910374, end_position_on_the_genomic_accession=35916385, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  66975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66975, CurrentID=0, Status='live', Symbol='Trappc13', Aliases='2410002O22Rik, 2610524F24Rik, AI451892, AU017229', description='trafficking protein particle complex 13', other_designations='UPF0533 protein C5orf44 homolog', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104142153, end_position_on_the_genomic_accession=104178703, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  66977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66977, CurrentID=0, Status='live', Symbol='Nuf2', Aliases='2410003C07Rik, AU044112, C85691, Cdca1, NUF2R', description='NUF2, NDC80 kinetochore complex component, homolog (S. cerevisiae)', other_designations='cell division cycle associated 1|cell division cycle-associated protein 1', map_location='1 H2.3-H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=169497934, end_position_on_the_genomic_accession=169531464, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  66978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66978, CurrentID=0, Status='live', Symbol='Luc7l', Aliases='1810045C04Rik, 2410018D03Rik', description='Luc7-like', other_designations='Luc7 homolog (S. cerevisiae)-like', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26245364, end_position_on_the_genomic_accession=26285510, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  66979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66979, CurrentID=0, Status='live', Symbol='Pole4', Aliases='2400007P05Rik, 5830430F06Rik', description='polymerase (DNA-directed), epsilon 4 (p12 subunit)', other_designations='DNA polymerase II subunit 4|DNA polymerase epsilon subunit p12', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=82618925, end_position_on_the_genomic_accession=82652936, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  66980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66980, CurrentID=0, Status='live', Symbol='Zdhhc6', Aliases='2400007G07Rik, 5930409M18Rik, AI462345, C77369', description='zinc finger, DHHC domain containing 6', other_designations='DHHC-6|H4 homolog|probable palmitoyltransferase ZDHHC6|zinc finger DHHC domain-containing protein 6', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=55298293, end_position_on_the_genomic_accession=55316032, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  66983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66983, CurrentID=0, Status='live', Symbol='Zfp830', Aliases='2410003C20Rik, AV301118, AW048207, Ccdc16, Omcg1, Znf830', description='zinc finger protein 830', other_designations='coiled-coil domain containing 16|coiled-coil domain-containing protein 16', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82764345, end_position_on_the_genomic_accession=82767622, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  66985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66985, CurrentID=0, Status='live', Symbol='Rassf7', Aliases='2400009B11Rik, AW210608', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 7', other_designations='Ras association (RalGDS/AF-6) domain family 7', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141215560, end_position_on_the_genomic_accession=141218658, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66988, CurrentID=0, Status='live', Symbol='Lap3', Aliases='2410015L10Rik, AA410100, LAP-3, Lap, Lapep, Pep-7, Pep-S, Pep7, Peps', description='leucine aminopeptidase 3', other_designations='leucyl aminopeptidase|peptidase 7|peptidase S|proline aminopeptidase|prolyl aminopeptidase', map_location='5 24.98 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=45493374, end_position_on_the_genomic_accession=45512691, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  66989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66989, CurrentID=0, Status='live', Symbol='Kctd20', Aliases='2410004N11Rik, AI451943, AW541186, D17Ertd562e', description='potassium channel tetramerisation domain containing 20', other_designations='potassium channel tetramerization domain containing 20', map_location='17 15.03 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28952448, end_position_on_the_genomic_accession=28969549, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  66990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66990, CurrentID=0, Status='live', Symbol='Tmem134', Aliases='2410001H17Rik, AI463452', description='transmembrane protein 134', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4125960, end_position_on_the_genomic_accession=4132307, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  66991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66991, CurrentID=0, Status='live', Symbol='Khdc3', Aliases='2410004A20Rik, AI467128, FILIA, OEEP48', description='KH domain containing 3, subcortical maternal complex member', other_designations='protein Filia', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=73101809, end_position_on_the_genomic_accession=73106631, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  66993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66993, CurrentID=0, Status='live', Symbol='Smarcd3', Aliases='1500001J14Rik, 2210409C08Rik, BAF60C', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily d, member 3', other_designations='60 kDa BRG-1/Brm-associated factor subunit C|BRG1-associated factor 60C|mBAF60c', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24592622, end_position_on_the_genomic_accession=24624456, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  66994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66994, CurrentID=0, Status='live', Symbol='Cep19', Aliases='1500031L02Rik, AI428934, AL022620', description='centrosomal protein 19', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32099802, end_position_on_the_genomic_accession=32108054, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66995, CurrentID=0, Status='live', Symbol='Zcchc18', Aliases='1500031H04Rik, Sizn2', description='zinc finger, CCHC domain containing 18', other_designations='SMAD-interacting zinc finger protein 2', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136992781, end_position_on_the_genomic_accession=136996923, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  66996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66996, CurrentID=0, Status='live', Symbol='Ceacam11', Aliases='1600015D01Rik', description='carcinoembryonic antigen-related cell adhesion molecule 11', other_designations='Ceacam11-C1|Ceacam11-C3|Ceacam11-C4|carcinoembryonic antigen related protein', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17972167, end_position_on_the_genomic_accession=17978556, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  66997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66997, CurrentID=0, Status='live', Symbol='Psmd12', Aliases='1500002F15Rik, AI480719, P55', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 12', other_designations='26S proteasome regulatory subunit RPN5|26S proteasome regulatory subunit p55', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107479528, end_position_on_the_genomic_accession=107498036, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  66998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66998, CurrentID=0, Status='live', Symbol='Psmd5', Aliases='1500032A03Rik, AW475925, S5b, W91691, mKIAA0072', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 5', other_designations='26S protease subunit S5 basic|26S proteasome subunit S5B', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34852089, end_position_on_the_genomic_accession=34870962, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  66999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=66999, CurrentID=0, Status='live', Symbol='Med28', Aliases='1500003D12Rik, AI451633, AU045690, Eg1, FKSG20, magicin', description='mediator complex subunit 28', other_designations='endothelial-derived|mediator of RNA polymerase II transcription, subunit 28 homolog', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=45520229, end_position_on_the_genomic_accession=45529284, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67000, CurrentID=0, Status='live', Symbol='Prl3a1', Aliases='1600016E11Rik, Ghd9, PLP-H, PLP-I, Plpi, Prlpi', description='prolactin family 3, subfamily a, member 1', other_designations='growth hormone d9|prolactin like protein I|prolactin-like protein I', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27259489, end_position_on_the_genomic_accession=27276660, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67003, CurrentID=0, Status='live', Symbol='Uqcrc2', Aliases='1500004O06Rik', description='ubiquinol cytochrome c reductase core protein 2', other_designations='complex III subunit 2|core protein II|ubiquinol-cytochrome-c reductase complex core protein 2', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120635189, end_position_on_the_genomic_accession=120659523, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67005, CurrentID=0, Status='live', Symbol='Polr3k', Aliases='12.3kDa, 1500004O14Rik, AI196849, AU014893, C11, RPC10, RPC11', description='polymerase (RNA) III (DNA directed) polypeptide K', other_designations='DNA-directed RNA polymerase III subunit K|RNA polymerase III subunit C10|RNA polymerase III subunit C11', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181864360, end_position_on_the_genomic_accession=181870830, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67006, CurrentID=0, Status='live', Symbol='Cisd2', Aliases='1500009M05Rik, 1500026J14Rik, 1500031D15Rik, AI848398, B630006A20Rik, Miner1, Noxp70, Zcd2', description='CDGSH iron sulfur domain 2', other_designations='CDGSH iron sulfur domain-containing protein 2|mitoNEET-related 1 protein|nervous system over-expressed protein 70|nervous system overexpressed protein 70|zinc finger, CDGSH-type domain 2', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135406412, end_position_on_the_genomic_accession=135424442, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  329152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329152, CurrentID=0, Status='live', Symbol='Hecw2', Aliases='A730039N16Rik, D030049F17Rik, Gm971, Nedl2, mKIAA1301', description='HECT, C2 and WW domain containing E3 ubiquitin protein ligase 2', other_designations='HECT, C2 and WW domain-containing protein 2|NEDD4-like E3 ubiquitin-protein ligase 2|NEDD4-related E3 ubiquitin ligase NEDL2', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53806872, end_position_on_the_genomic_accession=54195034, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  67009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67009, CurrentID=0, Status='live', Symbol='Ttc23', Aliases='1600012K10Rik', description='tetratricopeptide repeat domain 23', other_designations='TPR repeat protein 23', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=67647410, end_position_on_the_genomic_accession=67730060, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67010, CurrentID=0, Status='live', Symbol='Rbm7', Aliases='1200007M24Rik, 1500011D06Rik, AU041934, AW554393', description='RNA binding motif protein 7', other_designations='RNA-binding motif protein 7', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48488697, end_position_on_the_genomic_accession=48495341, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67011, CurrentID=0, Status='live', Symbol='Mettl6', Aliases='1600013P15Rik, AI467521, AU020711', description='methyltransferase like 6', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31460360, end_position_on_the_genomic_accession=31495021, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67013, CurrentID=0, Status='live', Symbol='Oma1', Aliases='2010001O09Rik, MPRP-1, ZMPOMA1', description='OMA1 homolog, zinc metallopeptidase (S. cerevisiae)', other_designations='overlapping with the m-AAA protease 1 homolog', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=103313209, end_position_on_the_genomic_accession=103366438, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67014, CurrentID=0, Status='live', Symbol='Mina', Aliases='1810047J07Rik, 2410057H13Rik, 3830408E23Rik, AI449204', description='myc induced nuclear antigen', other_designations='histone lysine demethylase MINA|mina|myc-induced nuclear antigen', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59471769, end_position_on_the_genomic_accession=59492461, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67015, CurrentID=0, Status='live', Symbol='Ccdc91', Aliases='1700086G08Rik, 1810060J02Rik, AU045683, AW555238, p56', description='coiled-coil domain containing 91', other_designations='GGA-binding partner', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=147475871, end_position_on_the_genomic_accession=147632613, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  67016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67016, CurrentID=0, Status='live', Symbol='Tbc1d2b', Aliases='1810061M12Rik, AI480956, AV021536, AV023399, AW491493, mKIAA1055', description='TBC1 domain family, member 2B', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=90202047, end_position_on_the_genomic_accession=90270769, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67017, CurrentID=0, Status='live', Symbol='Fam210b', Aliases='2010011I20Rik, AI430961, AV108736', description='family with sequence similarity 210, member B', other_designations='transmembrane protein C20orf108 homolog', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172345570, end_position_on_the_genomic_accession=172355749, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67019, CurrentID=0, Status='live', Symbol='Actr6', Aliases='2010200J04Rik, AU044443, Arp6, ArpX, CDA12', description='ARP6 actin-related protein 6', other_designations='ARP6 actin-related protein 6 homolog|mArp6', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=89711966, end_position_on_the_genomic_accession=89732295, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67020, CurrentID=0, Status='live', Symbol='Tmem88', Aliases='2600017H02Rik', description='transmembrane protein 88', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69396516, end_position_on_the_genomic_accession=69398234, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  329165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329165, CurrentID=0, Status='live', Symbol='Abi2', Aliases='8430425M24Rik, AI839867, C130078H13', description='abl-interactor 2', other_designations='abelson interactor 2|abi-2', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60408988, end_position_on_the_genomic_accession=60481945, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67023, CurrentID=0, Status='live', Symbol='Use1', Aliases='2010315L10Rik, 5730403H22Rik, AV002165, D12, Ed2, Q-snare, p31', description='unconventional SNARE in the ER 1 homolog (S. cerevisiae)', other_designations='USE1-like protein|embryonic development factor 2', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71366848, end_position_on_the_genomic_accession=71369732, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67025, CurrentID=0, Status='live', Symbol='Rpl11', Aliases='2010203J19Rik', description='ribosomal protein L11', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136049947, end_position_on_the_genomic_accession=136053371, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67026, CurrentID=0, Status='live', Symbol='Thap4', Aliases='2010320B01Rik', description='THAP domain containing 4', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93705390, end_position_on_the_genomic_accession=93754838, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67027, CurrentID=0, Status='live', Symbol='Mkrn2', Aliases='2610002L04Rik, C81377', description='makorin, ring finger protein, 2', other_designations='Makorin RING zinc-finger protein 2', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115601938, end_position_on_the_genomic_accession=115618670, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67028, CurrentID=0, Status='live', Symbol='2610002M06Rik', Aliases='8430408J07Rik, AI876413, Chmp1b2', description='RIKEN cDNA 2610002M06 gene', other_designations='CHMP1b-2|chromatin-modifying protein 1b-2', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107782751, end_position_on_the_genomic_accession=107816334, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67030, CurrentID=0, Status='live', Symbol='Fancl', Aliases='2010322C19Rik, AW554273, B230118H11Rik, Phf9, Pog, gcd', description='Fanconi anemia, complementation group L', other_designations='PHD finger protein 9|fanconi anemia group L protein homolog|germ cell deficient|proliferation of germ cells protein', map_location='11 A2-A3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=26387084, end_position_on_the_genomic_accession=26471883, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67031, CurrentID=0, Status='live', Symbol='Upf3a', Aliases='2600001C03Rik, 4930546M19Rik, RENT3A, UPF3', description='UPF3 regulator of nonsense transcripts homolog A (yeast)', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13785215, end_position_on_the_genomic_accession=13798538, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  329178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329178, CurrentID=0, Status='live', Symbol='Unc80', Aliases='C030018G13Rik, C230061B10Rik, UNC-80', description='unc-80 homolog (C. elegans)', other_designations='NALCN channel associated protein|mUNC-80', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=66468295, end_position_on_the_genomic_accession=66699148, orientation='plus', exon_count=65, OMIM='', no_hdr0=''),
  67035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67035, CurrentID=0, Status='live', Symbol='Dnajb4', Aliases='1700029A20Rik, 2010306G19Rik, 5730460G06Rik', description='DnaJ heat shock protein family (Hsp40) member B4', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 4', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152183871, end_position_on_the_genomic_accession=152210333, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67036, CurrentID=0, Status='live', Symbol='Mrpl45', Aliases='2600005P05Rik, AU043542, R75241', description='mitochondrial ribosomal protein L45', other_designations='L45mt|MRP-L45', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97315716, end_position_on_the_genomic_accession=97329920, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67037, CurrentID=0, Status='live', Symbol='Pmf1', Aliases='2600009M07Rik, AL033286, AW060657', description='polyamine-modulated factor 1', other_designations='PMF-1|pmf1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88394138, end_position_on_the_genomic_accession=88410408, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67038, CurrentID=0, Status='live', Symbol='2010109I03Rik', Aliases='', description='RIKEN cDNA 2010109I03 gene', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74878336, end_position_on_the_genomic_accession=74881704, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67039, CurrentID=0, Status='live', Symbol='Rbm25', Aliases='2600011C06Rik, 2610015J01Rik, A130095G20Rik, AI159652, AL023075, AU043498, RNPC7', description='RNA binding motif protein 25', other_designations='RNA-binding motif protein 25', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83632204, end_position_on_the_genomic_accession=83683127, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  67040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67040, CurrentID=0, Status='live', Symbol='Ddx17', Aliases='2610007K22Rik, A430025E01Rik, AI047725, C80929, Gm926, p72', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 17', other_designations='DEAD box protein 17', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79527696, end_position_on_the_genomic_accession=79546764, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67041, CurrentID=0, Status='live', Symbol='Oxct1', Aliases='2610008O03Rik, Oxct, Oxct2a, Scot-s', description='3-oxoacid CoA transferase 1', other_designations='3-oxoacid CoA-transferase 1|3-oxoacid-CoA transferase 1|somatic-type succinyl CoA:3-oxoacid CoA-transferase|somatic-type succinyl-CoA:3-oxoacid CoA-transferase|succinyl-CoA:3-ketoacid CoA transferase|succinyl-CoA:3-ketoacid-coenzyme A transferase 1, mitochondrial', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=4026428, end_position_on_the_genomic_accession=4155344, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  67042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67042, CurrentID=0, Status='live', Symbol='Ift27', Aliases='2600013G09Rik, Rabl4', description='intraflagellar transport 27', other_designations='RAB, member of RAS oncogene family-like 4|intraflagellar transport 27 homolog|rab-like protein 4', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78159463, end_position_on_the_genomic_accession=78174108, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67043, CurrentID=0, Status='live', Symbol='Syap1', Aliases='2010110O17Rik, AW011796', description='synapse associated protein 1', other_designations='', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=162856843, end_position_on_the_genomic_accession=162888462, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67044, CurrentID=0, Status='live', Symbol='Higd2a', Aliases='2010110M21Rik, AA690237', description='HIG1 domain family, member 2A', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54590231, end_position_on_the_genomic_accession=54591147, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67045, CurrentID=0, Status='live', Symbol='Riok2', Aliases='2010110K24Rik, 2410085M17Rik', description='RIO kinase 2 (yeast)', other_designations='', map_location='17 A3.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17369865, end_position_on_the_genomic_accession=17394899, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67046, CurrentID=0, Status='live', Symbol='Tbc1d7', Aliases='2610009C09Rik', description='TBC1 domain family, member 7', other_designations='', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43151740, end_position_on_the_genomic_accession=43171501, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67048, CurrentID=0, Status='live', Symbol='Vma21', Aliases='2610030H06Rik, AI840175', description='VMA21 vacuolar H+-ATPase homolog (S. cerevisiae)', other_designations='vacuolar ATPase assembly integral membrane protein VMA21', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71812752, end_position_on_the_genomic_accession=71824706, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67049, CurrentID=0, Status='live', Symbol='Pus3', Aliases='2610020J05Rik, 5730412F04Rik', description='pseudouridine synthase 3', other_designations='tRNA pseudouridine synthase 3|tRNA pseudouridylate synthase 3|tRNA-uridine isomerase 3', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35558597, end_position_on_the_genomic_accession=35567400, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67050, CurrentID=0, Status='live', Symbol='Nkap', Aliases='2610020O08Rik, AA987160, AI849147, AL024236', description='NFKB activating protein', other_designations='nuclear NF-kappaB activating protein', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37126763, end_position_on_the_genomic_accession=37150746, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67052, CurrentID=0, Status='live', Symbol='Ndc80', Aliases='2610020P18Rik, AV139730, HEC, HEC1, Kntc2', description='NDC80 homolog, kinetochore complex component (S. cerevisiae)', other_designations='highly expressed in cancer, rich in leucine heptad repeats|kinetochore associated 2|kinetochore protein Hec1|kinetochore-associated protein 2', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71496100, end_position_on_the_genomic_accession=71526857, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  67053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67053, CurrentID=0, Status='live', Symbol='Rpp14', Aliases='2610511E03Rik, AA682089', description='ribonuclease P 14 subunit', other_designations='ribonuclease P 14kDa subunit', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8063265, end_position_on_the_genomic_accession=8092127, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67054, CurrentID=0, Status='live', Symbol='Paics', Aliases='2610511I09Rik, ADE2H1, AIRC, PAIS', description='phosphoribosylaminoimidazole carboxylase, phosphoribosylaminoribosylaminoimidazole, succinocarboxamide synthetase', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76951315, end_position_on_the_genomic_accession=76967507, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67057, CurrentID=0, Status='live', Symbol='Yaf2', Aliases='2810021M11Rik', description='YY1 associated factor 2', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=93283833, end_position_on_the_genomic_accession=93336935, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67059, CurrentID=0, Status='live', Symbol='Ola1', Aliases='2510025G09Rik, 2810405J23Rik, 2810409H07Rik, Gtpbp9, PTD004', description='Obg-like ATPase 1', other_designations='GTP-binding protein 9 (putative)|GTP-binding protein PTD004', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73092801, end_position_on_the_genomic_accession=73214447, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67062, CurrentID=0, Status='live', Symbol='Slc25a53', Aliases='2310046F18Rik, 2810402A17Rik, AI448995, AI841128, Mcart6', description='solute carrier family 25, member 53', other_designations='mitochondrial carrier triple repeat 6', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136981116, end_position_on_the_genomic_accession=137038302, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  329207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329207, CurrentID=0, Status='live', Symbol='Rbm44', Aliases='Gm817', description='RNA binding motif protein 44', other_designations='RNA-binding motif protein 44', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91144973, end_position_on_the_genomic_accession=91170799, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  67064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67064, CurrentID=0, Status='live', Symbol='Chmp1b', Aliases='2810405I11Rik1, Chmp1b', description='charged multivesicular body protein 1B', other_designations='chromatin modifying protein 1B|chromatin-modifying protein 1b-1', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67205359, end_position_on_the_genomic_accession=67207887, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67065, CurrentID=0, Status='live', Symbol='Polr3d', Aliases='2810426M17Rik, 44kDa, AI326118, AW489084, BN51T, RPC4, TSBN51', description='polymerase (RNA) III (DNA directed) polypeptide D', other_designations='BN51 (BHK21) temperature sensitivity complementing|DNA-directed RNA polymerase III subunit C4|DNA-directed RNA polymerase III subunit D|RNA polymerase III subunit C4', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70438748, end_position_on_the_genomic_accession=70443471, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67067, CurrentID=0, Status='live', Symbol='Romo1', Aliases='2010100O12Rik, AI853864', description='reactive oxygen species modulator 1', other_designations='ROS modulator 1|protein MGR2 homolog', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156144153, end_position_on_the_genomic_accession=156145794, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67068, CurrentID=0, Status='live', Symbol='Dynlrb1', Aliases='2010012N15Rik, 2010320M17Rik, 9430076K19Rik, AV124457, DNLC2A, Dncl2a, km23-1', description='dynein light chain roadblock-type 1', other_designations='dynein light chain 2A, cytoplasmic|dynein, cytoplasmic, light chain 2A|dynein, cytoplasmic, light polypeptide 2A', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155236533, end_position_on_the_genomic_accession=155250277, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67070, CurrentID=0, Status='live', Symbol='Lsm14a', Aliases='2700023B17Rik, AA407828, AU017544, Tral', description='LSM14A mRNA processing body assembly factor', other_designations='LSM14 homolog A|RNA-associated protein 55A|mRAP55A', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34344637, end_position_on_the_genomic_accession=34389637, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67071, CurrentID=0, Status='live', Symbol='Rps6ka6', Aliases='2610524K04Rik, RSK4, S6K-alpha-6', description='ribosomal protein S6 kinase polypeptide 6', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=111387837, end_position_on_the_genomic_accession=111538097, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  67072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67072, CurrentID=0, Status='live', Symbol='Cdc37l1', Aliases='2700033A15Rik, BB159850, Cdc37l, Harc', description='cell division cycle 37-like 1', other_designations='Hsp90-associating relative of Cdc37|cell division cycle 37 homolog-like 1', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=28990325, end_position_on_the_genomic_accession=29026682, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67073, CurrentID=0, Status='live', Symbol='Pi4k2b', Aliases='2610042N09Rik, 4933409G22Rik', description='phosphatidylinositol 4-kinase type 2 beta', other_designations='phosphatidylinositol 4-kinase type II-beta', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52741557, end_position_on_the_genomic_accession=52769344, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67074, CurrentID=0, Status='live', Symbol='Mon2', Aliases='2610528O22Rik, AW495628, Sf21, mKIAA1040', description='MON2 homolog (yeast)', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=122992060, end_position_on_the_genomic_accession=123076505, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  67075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67075, CurrentID=0, Status='live', Symbol='Magt1', Aliases='2410001C15Rik, 2610529C04Rik, 2810482I07Rik, IAG2, IAP', description='magnesium transporter 1', other_designations='implantation-associated protein', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=105968084, end_position_on_the_genomic_accession=106011899, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67077, CurrentID=0, Status='live', Symbol='Tex40', Aliases='1700019N12Rik, A430107B04Rik, MLZ-622', description='testis expressed 40', other_designations='protein expressed in male leptotene and zygotene spermatocytes 622', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6922426, end_position_on_the_genomic_accession=6925380, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67078, CurrentID=0, Status='live', Symbol='Pgp', Aliases='1700012G19Rik, AI481330, AUM', description='phosphoglycolate phosphatase', other_designations='PGPase|aspartate-based ubiquitous Mg(2+)-dependent phosphatase', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24470473, end_position_on_the_genomic_accession=24471602, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67080, CurrentID=0, Status='live', Symbol='1700019D03Rik', Aliases='2410136C24Rik, AA410151, smAKAP', description='RIKEN cDNA 1700019D03 gene', other_designations='UPF0733 protein C2orf88 homolog|small membrane AKAP', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52922324, end_position_on_the_genomic_accession=52991343, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67082, CurrentID=0, Status='live', Symbol='1700011H14Rik', Aliases='', description='RIKEN cDNA 1700011H14 gene', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49224249, end_position_on_the_genomic_accession=49245656, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67084, CurrentID=0, Status='live', Symbol='Ceacam14', Aliases='1600021E03Rik, 1600025E09Rik, AV033953', description='carcinoembryonic antigen-related cell adhesion molecule 14', other_designations='CEA-related cell adhesion molecule 14', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17812682, end_position_on_the_genomic_accession=17815627, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67085, CurrentID=0, Status='live', Symbol='1700024G13Rik', Aliases='C10orf53', description='RIKEN cDNA 1700024G13 gene', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32376502, end_position_on_the_genomic_accession=32388373, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67087, CurrentID=0, Status='live', Symbol='Ctnnbip1', Aliases='1110008O09Rik, 2310001I19Rik, AW457332, Catnbip1, Icat', description='catenin beta interacting protein 1', other_designations='inhibitor of beta-catenin and Tcf-4', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149518241, end_position_on_the_genomic_accession=149566437, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67088, CurrentID=0, Status='live', Symbol='Cand2', Aliases='2210404G23Rik, Tp120b, mKIAA0667', description='cullin-associated and neddylation-dissociated 2 (putative)', other_designations='TBP-interacting protein 120B|TBP-interacting protein b|TBP-interacting protein of 120 kDa B|cullin-associated and neddylation-dissociated protein 2|p120 CAND2', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115774557, end_position_on_the_genomic_accession=115805555, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67089, CurrentID=0, Status='live', Symbol='Psmc6', Aliases='2300001E01Rik, AI451058', description='proteasome (prosome, macropain) 26S subunit, ATPase, 6', other_designations='26S protease regulatory subunit S10B|26S proteasome AAA-ATPase subunit RPT4|proteasome 26S subunit ATPase 6|proteasome subunit p42', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45329824, end_position_on_the_genomic_accession=45349071, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67091, CurrentID=0, Status='live', Symbol='Trappc6a', Aliases='1810073E21Rik, 4930519D19Rik, AI480686, TRS33, mhyp', description='trafficking protein particle complex 6A', other_designations='TRAPP complex subunit 6A', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19507406, end_position_on_the_genomic_accession=19516145, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67092, CurrentID=0, Status='live', Symbol='Gatm', Aliases='1810003P21Rik, AI314789, AT', description='glycine amidinotransferase (L-arginine:glycine amidinotransferase)', other_designations='L-arginine:glycine amidinotransferase|transamidinase', map_location='2 60.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122594473, end_position_on_the_genomic_accession=122611277, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67095, CurrentID=0, Status='live', Symbol='Trak1', Aliases='2310001H13Rik, AI413908, AI467545, hyrt, mKIAA1042', description='trafficking protein, kinesin binding 1', other_designations='', map_location='9 72.41 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121297507, end_position_on_the_genomic_accession=121474919, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  67096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67096, CurrentID=0, Status='live', Symbol='Mmachc', Aliases='1810037K07Rik', description='methylmalonic aciduria cblC type, with homocystinuria', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116702434, end_position_on_the_genomic_accession=116708385, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67097, CurrentID=0, Status='live', Symbol='Rps10', Aliases='2210402A09Rik', description='ribosomal protein S10', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27630420, end_position_on_the_genomic_accession=27635480, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67099, CurrentID=0, Status='live', Symbol='Mettl21a', Aliases='2310038H17Rik, AI464204, Fam119a', description='methyltransferase like 21A', other_designations='family with sequence similarity 119, member A|methyltransferase-like protein 21A|protein-lysine methyltransferase METTL21A', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=64606473, end_position_on_the_genomic_accession=64620421, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  329244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329244, CurrentID=0, Status='live', Symbol='Il19', Aliases='', description='interleukin 19', other_designations='IL-19', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130932656, end_position_on_the_genomic_accession=130945457, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67101, CurrentID=0, Status='live', Symbol='2310039H08Rik', Aliases='', description='RIKEN cDNA 2310039H08 gene', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46772635, end_position_on_the_genomic_accession=46773407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67102, CurrentID=0, Status='live', Symbol='D16Ertd472e', Aliases='1700010I10Rik, 2310009O17Rik, E330003K22Rik, Eurl', description='DNA segment, Chr 16, ERATO Doi 472, expressed', other_designations='', map_location='16 45.36 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=78540336, end_position_on_the_genomic_accession=78576688, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67103, CurrentID=0, Status='live', Symbol='Ptgr1', Aliases='2510002C21Rik, Ltb4dh, Zadh3', description='prostaglandin reductase 1', other_designations='15-oxoprostaglandin 13-reductase|NADP-dependent leukotriene B4 12-hydroxydehydrogenase|PRG-1|leukotriene B4 12-hydroxydehydrogenase', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58965590, end_position_on_the_genomic_accession=58987131, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67105, CurrentID=0, Status='live', Symbol='Timm21', Aliases='1700034H14Rik, 2700002I20Rik', description='tranlocase of inner mitochondrial membrane 21', other_designations='TIM21-like protein, mitochondrial', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84947294, end_position_on_the_genomic_accession=84951738, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67106, CurrentID=0, Status='live', Symbol='Zbtb8os', Aliases='2010001H09Rik, 2310028N13Rik, AI851226, Arch, Archm', description='zinc finger and BTB domain containing 8 opposite strand', other_designations='archease (Arch)', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129335560, end_position_on_the_genomic_accession=129347083, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  329251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329251, CurrentID=0, Status='live', Symbol='Ppp1r12b', Aliases='1810037O03Rik, 9530009M10Rik, AI132431, Mypt2', description='protein phosphatase 1, regulatory (inhibitor) subunit 12B', other_designations='myosin phosphatase target subunit 2|myosin phosphatase-targeting subunit 2', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134754665, end_position_on_the_genomic_accession=134955965, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  329252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329252, CurrentID=0, Status='live', Symbol='Lgr6', Aliases='A530037C04Rik, D830026M09', description='leucine-rich repeat-containing G protein-coupled receptor 6', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134983300, end_position_on_the_genomic_accession=135105374, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  67109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67109, CurrentID=0, Status='live', Symbol='Zfp787', Aliases='2210018M03Rik, Znf787', description='zinc finger protein 787', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6131489, end_position_on_the_genomic_accession=6159564, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67111, CurrentID=0, Status='live', Symbol='Naaa', Aliases='2210023K21Rik, 3830414F09Rik, Asahl', description='N-acylethanolamine acid amidase', other_designations='ASAH-like protein', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92257660, end_position_on_the_genomic_accession=92278181, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67112, CurrentID=0, Status='live', Symbol='Fgf22', Aliases='2210414E06Rik, FGF-22', description='fibroblast growth factor 22', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79754575, end_position_on_the_genomic_accession=79757572, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67115, CurrentID=0, Status='live', Symbol='Rpl14', Aliases='3100001N19Rik, AA407502, AL022816, CAG-ISL-7, CTG-B33, L14', description='ribosomal protein L14', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120571517, end_position_on_the_genomic_accession=120574653, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67116, CurrentID=0, Status='live', Symbol='Cuedc2', Aliases='3010002G01Rik', description='CUE domain containing 2', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46329812, end_position_on_the_genomic_accession=46338660, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67117, CurrentID=0, Status='live', Symbol='Dynlt3', Aliases='2310075M16Rik, AU042796, Tcte1l', description='dynein light chain Tctex-type 3', other_designations='protein 91/23', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9654270, end_position_on_the_genomic_accession=9662983, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67118, CurrentID=0, Status='live', Symbol='Bfar', Aliases='3010001A07Rik, AI666707, AW107665, Bar, Rnf47', description='bifunctional apoptosis regulator', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13671858, end_position_on_the_genomic_accession=13703612, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67119, CurrentID=0, Status='live', Symbol='Ccdc159', Aliases='2510048L02Rik', description='coiled-coil domain containing 159', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21927471, end_position_on_the_genomic_accession=21936256, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67120, CurrentID=0, Status='live', Symbol='Ttc14', Aliases='2700016E08Rik, 4930434D01Rik, 4931403I22Rik, 4933402I15Rik, AI662468, AU014779, AW561908, cI-44, mKIAA1980', description='tetratricopeptide repeat domain 14', other_designations='TPR repeat protein 14', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=33800022, end_position_on_the_genomic_accession=33814860, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67121, CurrentID=0, Status='live', Symbol='Mastl', Aliases='2700091H24Rik, C88295, GW, GWL, MAST-L, THC2', description='microtubule associated serine/threonine kinase-like', other_designations='microtubule-associated serine/threonine-protein kinase-like', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=23116544, end_position_on_the_genomic_accession=23156024, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67122, CurrentID=0, Status='live', Symbol='Nrarp', Aliases='2700054M22Rik', description='Notch-regulated ankyrin repeat protein', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25180758, end_position_on_the_genomic_accession=25183339, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67123, CurrentID=0, Status='live', Symbol='Ubap1', Aliases='2700092A01Rik, NAG20, UBAP-1, Ubap', description='ubiquitin-associated protein 1', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41348996, end_position_on_the_genomic_accession=41389766, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67125, CurrentID=0, Status='live', Symbol='Tspan31', Aliases='2700085A14Rik, AW742554, Sas', description='tetraspanin 31', other_designations='sarcoma amplified sequence|sarcoma-amplified sequence homolog|tspan-31', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127067290, end_position_on_the_genomic_accession=127070261, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67126, CurrentID=0, Status='live', Symbol='Atp5e', Aliases='2410043G19Rik, ATPE, AV000645', description='ATP synthase, H+ transporting, mitochondrial F1 complex, epsilon subunit', other_designations='ATP synthase epsilon chain, mitochondrial|ATPase subunit epsilon|mitochondrial ATP synthase epsilon subunit', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174461075, end_position_on_the_genomic_accession=174464101, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67128, CurrentID=0, Status='live', Symbol='Ube2g1', Aliases='2700059C12Rik, AI256795, AU014992, AW552068, D130023C12Rik', description='ubiquitin-conjugating enzyme E2G 1', other_designations='E2 ubiquitin-conjugating enzyme G1|E217K|UBC7|ubiquitin carrier protein G1|ubiquitin-conjugating enzyme E2G 1 (UBC7 homolog, C. elegans)|ubiquitin-protein ligase G1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72607261, end_position_on_the_genomic_accession=72686481, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67130, CurrentID=0, Status='live', Symbol='Ndufa6', Aliases='14kDa, 2700038D15Rik, B230217P19Rik', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 6 (B14)', other_designations='CI-B14|NADH-ubiquinone oxidoreductase B14 subunit|complex I-B14', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82350139, end_position_on_the_genomic_accession=82354291, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67131, CurrentID=0, Status='live', Symbol='Acbd4', Aliases='2010009P05Rik, 2010015A21Rik, AI849317', description='acyl-Coenzyme A binding domain containing 4', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103101688, end_position_on_the_genomic_accession=103112199, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67133, CurrentID=0, Status='live', Symbol='Gp2', Aliases='2310037I18Rik, AV060639', description='glycoprotein 2 (zymogen granule membrane)', other_designations='pancreatic zymogen granule membrane protein GP-2', map_location='7 F1-F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119442542, end_position_on_the_genomic_accession=119459272, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  329278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329278, CurrentID=0, Status='live', Symbol='Tnn', Aliases='Tnw, tenascin-W', description='tenascin N', other_designations='TN-N|tenasin W', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160085029, end_position_on_the_genomic_accession=160153692, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  67135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67135, CurrentID=0, Status='live', Symbol='Bpifa5', Aliases='2310021H06Rik, 2310074B19Rik, Pluncl, Splunc5, Tpl', description='BPI fold containing family A, member 5', other_designations='BPI fold-containing family A member Pluncl|palate lung and nasal carcinoma-like protein|short palate lung and nasal clone protein 5|tongue plunc-like protein', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154162607, end_position_on_the_genomic_accession=154168446, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67136, CurrentID=0, Status='live', Symbol='Kbtbd4', Aliases='2510026C23Rik', description='kelch repeat and BTB (POZ) domain containing 4', other_designations='BTB and kelch domain-containing protein 4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90904724, end_position_on_the_genomic_accession=90910561, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67138, CurrentID=0, Status='live', Symbol='Herc6', Aliases='1700121D12Rik, 2510038N07Rik, 4930427L17Rik, AI451296, CEB1, Herc5', description='hect domain and RLD 6', other_designations='hect domain and RLD 5', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57580992, end_position_on_the_genomic_accession=57665136, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  67139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67139, CurrentID=0, Status='live', Symbol='Mis12', Aliases='2510025F08Rik, AU044651, Misc12', description='MIS12 kinetochore complex component', other_designations='MIS12 homolog', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=71019611, end_position_on_the_genomic_accession=71027371, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67141, CurrentID=0, Status='live', Symbol='Fbxo5', Aliases='2510044I10Rik, C85305, Emi1, Fbxo31', description='F-box protein 5', other_designations='early mitotic inhibitor 1|f-box only protein 31|fbxo5', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=5799158, end_position_on_the_genomic_accession=5805465, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67143, CurrentID=0, Status='live', Symbol='Ikzf5', Aliases='2610034F18Rik, PEGASUS, Zfpn1a5', description='IKAROS family zinc finger 5', other_designations='ikaros family zinc finger protein 5|zinc finger protein, subfamily 1A, 5', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131388649, end_position_on_the_genomic_accession=131410506, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67144, CurrentID=0, Status='live', Symbol='Lrrc40', Aliases='2610040E16Rik, AI837674', description='leucine rich repeat containing 40', other_designations='', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=158036682, end_position_on_the_genomic_accession=158067090, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  67145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67145, CurrentID=0, Status='live', Symbol='Tomm34', Aliases='2610100K07Rik, TOM34', description='translocase of outer mitochondrial membrane 34', other_designations='translocase of outer membrane 34 kDa subunit', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164053538, end_position_on_the_genomic_accession=164071177, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67148, CurrentID=0, Status='live', Symbol='Fam103a1', Aliases='2410047I02Rik, 2610204K14Rik, RAM', description='family with sequence similarity 103, member A1', other_designations='protein FAM103A1', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81762953, end_position_on_the_genomic_accession=81769490, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67149, CurrentID=0, Status='live', Symbol='Nkain1', Aliases='2610200G18Rik, 2810426C15Rik', description='Na+/K+ transporting ATPase interacting 1', other_designations='Na(+)/K(+)-transporting ATPase subunit beta-1-interacting protein 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130530131, end_position_on_the_genomic_accession=130574036, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67150, CurrentID=0, Status='live', Symbol='Rnf141', Aliases='2610110L04Rik, AA792898, AU022812, ZFP36, ZNF230', description='ring finger protein 141', other_designations='zinc finger protein 230', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110799469, end_position_on_the_genomic_accession=110844452, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67151, CurrentID=0, Status='live', Symbol='Psmd9', Aliases='1500011J20Rik, 2610202L11Rik, Bridge-1, P27', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 9', other_designations='26S proteasome regulatory subunit p27', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123228190, end_position_on_the_genomic_accession=123250126, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67153, CurrentID=0, Status='live', Symbol='Rnaseh2b', Aliases='1110019N06Rik, 2610207P08Rik, Dleu8', description='ribonuclease H2, subunit B', other_designations='RNase H2 subunit B|deleted in lymphocytic leukemia 8 homolog|ribonuclease HI subunit B', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62332068, end_position_on_the_genomic_accession=62392708, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  67154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67154, CurrentID=0, Status='live', Symbol='Mtdh', Aliases='2610103J23Rik, 3D3, 3D3/Lyric, AEG-1, AV353288, D8Bwg1112e, Lyric', description='metadherin', other_designations='lysine-rich CEACAM1 co-isolated protein|metastasis adhesion protein', map_location='15 B3.3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34082719, end_position_on_the_genomic_accession=34143537, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67155, CurrentID=0, Status='live', Symbol='Smarca2', Aliases='2610209L14Rik, SNF2alpha, Snf2l2, brahma, brm', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 2', other_designations='ATP-dependent helicase SMARCA2|BAF190B|BRG1-associated factor 190B|SNF2-alpha|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily A member 2|protein brahma homolog|subunit of the SWI/SNF chromatin remodeling complex', map_location='19 21.17 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=26605089, end_position_on_the_genomic_accession=26778321, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  67157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67157, CurrentID=0, Status='live', Symbol='2610301B20Rik', Aliases='AI428449', description='RIKEN cDNA 2610301B20 gene', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=10874249, end_position_on_the_genomic_accession=10899425, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67158, CurrentID=0, Status='live', Symbol='Sft2d3', Aliases='2610206C24Rik, AI847965, AW046173', description='SFT2 domain containing 3', other_designations='SFT2 domain-containing protein 3|sft2d3', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31909094, end_position_on_the_genomic_accession=31911903, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  67160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67160, CurrentID=0, Status='live', Symbol='Eef1g', Aliases='2610301D06Rik, AA407312, EF1G', description='eukaryotic translation elongation factor 1 gamma', other_designations='EF-1-gamma|eEF-1B gamma', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8967041, end_position_on_the_genomic_accession=8978180, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67161, CurrentID=0, Status='live', Symbol='Sclt1', Aliases='2610207F23Rik, 4931421F20Rik', description='sodium channel and clathrin linker 1', other_designations='sodium channel-associated protein 1', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=41626705, end_position_on_the_genomic_accession=41742530, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  67163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67163, CurrentID=0, Status='live', Symbol='Ccdc47', Aliases='2610204L23Rik, C88307, asp4, calumin', description='coiled-coil domain containing 47', other_designations='adipocyte-specific protein 4', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106199354, end_position_on_the_genomic_accession=106216367, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  67164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67164, CurrentID=0, Status='live', Symbol='Lipt2', Aliases='2610209A20Rik', description='lipoyl(octanoyl) transferase 2 (putative)', other_designations='lipoate-protein ligase B|lipoyl/octanoyl transferase|octanoyl-[acyl-carrier-protein]-protein N-octanoyltransferase|putative octanoyltransferase, mitochondrial', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100159277, end_position_on_the_genomic_accession=100160931, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67166, CurrentID=0, Status='live', Symbol='Arl8b', Aliases='2610313E07Rik, 3100002J04Rik, Arl10c, gie1', description='ADP-ribosylation factor-like 8B', other_designations='ADP-ribosylation factor-like 10C|ADP-ribosylation factor-like protein 10C|novel small G protein indispensable for equal chromosome segregation 1|small G protein indispensable for equal chromosome segregation 1', map_location='6 E2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=108783059, end_position_on_the_genomic_accession=108823723, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67168, CurrentID=0, Status='live', Symbol='Lpar6', Aliases='2610302I02Rik, P2ry5, P2y5', description='lysophosphatidic acid receptor 6', other_designations='LPA receptor 6|LPA-6|P2Y purinoceptor 5|oleoyl-L-alpha-lysophosphatidic acid receptor|purinergic receptor (family A group 5)|purinergic receptor 5|purinergic receptor P2Y, G-protein coupled, 5', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73237891, end_position_on_the_genomic_accession=73240358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67169, CurrentID=0, Status='live', Symbol='Nradd', Aliases='2610311B09Rik, BB104149, NRDD, NRH2, PLAIDD', description='neurotrophin receptor associated death domain', other_designations='death domain containing membrane protein NRADD|neurotrophin receptor homolog-2|neurotrophin receptor-alike death domain protein', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110621131, end_position_on_the_genomic_accession=110624393, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67171, CurrentID=0, Status='live', Symbol='Dram2', Aliases='2010305N14Rik, 2610318G18Rik, AI647667, Tmem77', description='DNA-damage regulated autophagy modulator 2', other_designations='VDNA-damage regulated autophagy modulator 2|transmembrane protein 77', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106547798, end_position_on_the_genomic_accession=106575341, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67177, CurrentID=0, Status='live', Symbol='Cdt1', Aliases='2610318F11Rik, AW545653, C76791, Ris2', description='chromatin licensing and DNA replication factor 1', other_designations='DUP|cdt1|double parked homolog|retroviral insertion site 2 protein|retroviral integration site 1|retroviral integration site 2', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122568015, end_position_on_the_genomic_accession=122573130, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67178, CurrentID=0, Status='live', Symbol='Zmat5', Aliases='2610510L01Rik, D11Bwg0572e, D11Bwg1548e', description='zinc finger, matrin type 5', other_designations='U11/U12 small nuclear ribonucleoprotein 20 kDa protein|U11/U12 snRNP 20 kDa protein', map_location='11 2.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4704606, end_position_on_the_genomic_accession=4737669, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67179, CurrentID=0, Status='live', Symbol='Ccdc25', Aliases='2610528H13Rik, NSrp70', description='coiled-coil domain containing 25', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65837302, end_position_on_the_genomic_accession=65866604, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67180, CurrentID=0, Status='live', Symbol='Yipf5', Aliases='2610311I19Rik, AA408236, Yip1a', description='Yip1 domain family, member 5', other_designations='YIP1 family member 5|YPT-interacting protein 1 A', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=40204865, end_position_on_the_genomic_accession=40219399, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67181, CurrentID=0, Status='live', Symbol='Ctdnep1', Aliases='2610507E10Rik, Dullard', description='CTD nuclear envelope phosphatase 1', other_designations='serine/threonine-protein phosphatase dullard', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69981168, end_position_on_the_genomic_accession=69990601, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67182, CurrentID=0, Status='live', Symbol='Pdzk1ip1', Aliases='0610007F13Rik, 2700030M23Rik, AU046042, Map17', description='PDZK1 interacting protein 1', other_designations='17 kDa membrane-associated protein|membrane-associated protein 17 (Map17)', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115088708, end_position_on_the_genomic_accession=115093894, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67184, CurrentID=0, Status='live', Symbol='Ndufa13', Aliases='2700054G14Rik, AU022060, CDA016, CGI-39, GRIM-19, Grim19', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 13', other_designations='CI-B16.6|NADH-ubiquinone oxidoreductase B16.6 subunit|cell death regulatory protein GRIM-19|complex I-B16.6|gene associated with retinoic and IFN-induced mortality 19 protein|gene associated with retinoic and interferon-induced mortality 19 protein|gene associated with retinoic-interferon-induced mortality 19 protein|genes associated with retinoid-IFN-induced mortality 19', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69887788, end_position_on_the_genomic_accession=69902558, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67186, CurrentID=0, Status='live', Symbol='Rplp2', Aliases='2700049I22Rik', description='ribosomal protein, large P2', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141447464, end_position_on_the_genomic_accession=141451350, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67187, CurrentID=0, Status='live', Symbol='Zmynd19', Aliases='2700064H14Rik, AA536891, AU014915, mizip', description='zinc finger, MYND domain containing 19', other_designations='MCH-R1-interacting zinc finger protein|melanin concentrating hormone receptor interacting zinc finger protein|melanin-concentrating hormone receptor 1-interacting zinc finger protein', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24949802, end_position_on_the_genomic_accession=24960200, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67188, CurrentID=0, Status='live', Symbol='2700046G09Rik', Aliases='', description='RIKEN cDNA 2700046G09 gene', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=32389216, end_position_on_the_genomic_accession=32391184, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67194, CurrentID=0, Status='live', Symbol='2700038G22Rik', Aliases='', description='RIKEN cDNA 2700038G22 gene', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23850597, end_position_on_the_genomic_accession=23855033, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67196, CurrentID=0, Status='live', Symbol='Ube2t', Aliases='2700084L22Rik, C80607', description='ubiquitin-conjugating enzyme E2T', other_designations='E2 ubiquitin-conjugating enzyme T|ubiquitin carrier protein T|ubiquitin-conjugating enzyme E2T (putative)|ubiquitin-protein ligase T', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134962565, end_position_on_the_genomic_accession=134974180, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67197, CurrentID=0, Status='live', Symbol='Zcrb1', Aliases='2700088M22Rik, Madp-1', description='zinc finger CCHC-type and RNA binding motif 1', other_designations='U11/U12 small nuclear ribonucleoprotein 31 kDa protein|U11/U12 snRNP 31 kDa protein', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=93386106, end_position_on_the_genomic_accession=93398329, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67198, CurrentID=0, Status='live', Symbol='Spats2l', Aliases='2810022L02Rik, A230104H11Rik, AA987173, AI480531, AI842453, AW413586', description='spermatogenesis associated, serine-rich 2-like', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=57774162, end_position_on_the_genomic_accession=57948398, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  67199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67199, CurrentID=0, Status='live', Symbol='Pfdn1', Aliases='2700086I23Rik, AA408327, AU044714', description='prefoldin 1', other_designations='', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36403679, end_position_on_the_genomic_accession=36454495, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67200, CurrentID=0, Status='live', Symbol='Ccdc77', Aliases='2400002C23Rik, 2700091N06Rik, AV168274', description='coiled-coil domain containing 77', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120324322, end_position_on_the_genomic_accession=120364369, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  67201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67201, CurrentID=0, Status='live', Symbol='Glod4', Aliases='1700082G03Rik, 2700085E05Rik, C81254', description='glyoxalase domain containing 4', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76220395, end_position_on_the_genomic_accession=76243699, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67203, CurrentID=0, Status='live', Symbol='Nde1', Aliases='2810027M15Rik, AU042936, AW822251, Nude, mNudE', description='nuclear distribution gene E homolog 1 (A nidulans)', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14163275, end_position_on_the_genomic_accession=14192928, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67204, CurrentID=0, Status='live', Symbol='Eif2s2', Aliases='2810026E11Rik, 38kDa, AA408636, AA571381, AA986487, AW822225, D2Ertd303e, EIF2, EIF2B', description='eukaryotic translation initiation factor 2, subunit 2 (beta)', other_designations='eIF-2-beta|eukaryotic translation initiation factor 2 subunit beta|eukaryotic translation initiation factor 2, subunit 2 (beta, 38kDa)', map_location='2 76.89 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154871410, end_position_on_the_genomic_accession=154892906, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67205, CurrentID=0, Status='live', Symbol='Utp11l', Aliases='2700082D03Rik, CGI-94, Cgi94', description='UTP11-like, U3 small nucleolar ribonucleoprotein, (yeast)', other_designations='U3 snoRNA-associated protein 11|UTP11-like protein', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124678232, end_position_on_the_genomic_accession=124693612, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67207, CurrentID=0, Status='live', Symbol='Lsm1', Aliases='2810025O06Rik, CASM', description='LSM1 homolog, mRNA degradation associated', other_designations='LSM1 homolog, U6 small nuclear RNA associated', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25785288, end_position_on_the_genomic_accession=25803975, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67210, CurrentID=0, Status='live', Symbol='Gatad1', Aliases='2310031E19Rik, 2810047M21Rik, 8430439A17Rik, 9130430G15Rik, B330017N08Rik, ENSMUSG00000073241, Odag', description='GATA zinc finger domain containing 1', other_designations='ocular development associated|ocular development-associated gene protein', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3639961, end_position_on_the_genomic_accession=3647936, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67211, CurrentID=0, Status='live', Symbol='Armc10', Aliases='2810037C14Rik, Svh', description='armadillo repeat containing 10', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21645984, end_position_on_the_genomic_accession=21662592, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67212, CurrentID=0, Status='live', Symbol='Mrpl55', Aliases='2810038N09Rik, L55mt, MRP-L55', description='mitochondrial ribosomal protein L55', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59202486, end_position_on_the_genomic_accession=59206137, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67213, CurrentID=0, Status='live', Symbol='Cmtm6', Aliases='2810051A14Rik, AA536733, Cklfsf6', description='CKLF-like MARVEL transmembrane domain containing 6', other_designations='chemokine-like factor super family 6|chemokine-like factor superfamily 6|chemokine-like factor superfamily member 6', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114731157, end_position_on_the_genomic_accession=114749347, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67216, CurrentID=0, Status='live', Symbol='Mboat2', Aliases='2810049G06Rik, AU022889, AW547221, LPCAT4, Moact2, Oact2', description='membrane bound O-acyltransferase domain containing 2', other_designations='1-acylglycerophosphocholine O-acyltransferase|1-acylglycerophosphoethanolamine O-acyltransferase|LPC acyltransferase 4|LPCAT|LPEAT|LPLAT 2|O-acyltransferase (membrane bound) domain containing 2|O-acyltransferase domain-containing protein 2|lyso-PC acyltransferase 4|lyso-PE acyltransferase|lysophosphatidylcholine acyltransferase 4|lysophosphatidylethanolamine acyltransferase|mLPCAT4|membrane-bound O-acyltransferase domain-containing protein 2', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24831487, end_position_on_the_genomic_accession=24964400, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  67217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67217, CurrentID=0, Status='live', Symbol='L3hypdh', Aliases='2810055F11Rik', description='L-3-hydroxyproline dehydratase (trans-)', other_designations='probable proline racemase|trans-3-hydroxy-l-proline dehydratase', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72071660, end_position_on_the_genomic_accession=72085433, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67219, CurrentID=0, Status='live', Symbol='Med18', Aliases='2810046C01Rik', description='mediator complex subunit 18', other_designations='mediator of RNA polymerase II transcription, subunit 18 homolog', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132458729, end_position_on_the_genomic_accession=132463921, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67220, CurrentID=0, Status='live', Symbol='Plekho1', Aliases='2810052M02Rik, CKIP-1, Ckip1, JZA-20, Jza2', description='pleckstrin homology domain containing, family O member 1', other_designations='CK2 interacting protein 1|Jun zipper-associated 2|PH domain-containing family O member 1|TNF intracellular domain-interacting protein', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95988809, end_position_on_the_genomic_accession=95995863, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67222, CurrentID=0, Status='live', Symbol='Srfbp1', Aliases='2810036K01Rik, p49/STRAP', description='serum response factor binding protein 1', other_designations='SRF-dependent transcription regulation associated protein|SRF-dependent transcription regulation-associated protein|srfbp1', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=52465672, end_position_on_the_genomic_accession=52490738, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67223, CurrentID=0, Status='live', Symbol='Rrp15', Aliases='2810430M08Rik, 5430437H21Rik', description='ribosomal RNA processing 15 homolog (S. cerevisiae)', other_designations='ribosomal RNA-processing protein 15', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=186721086, end_position_on_the_genomic_accession=186749358, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67224, CurrentID=0, Status='live', Symbol='Med29', Aliases='2810405O22Rik, AU020902, Ixl', description='mediator complex subunit 29', other_designations='intersex-like protein', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28385247, end_position_on_the_genomic_accession=28392690, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67225, CurrentID=0, Status='live', Symbol='Rnpc3', Aliases='2810441O16Rik, AI447568, C030014B17Rik', description='RNA-binding region (RNP1, RRM) containing 3', other_designations='RNA-binding motif protein 40|RNA-binding region-containing protein 3', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113605067, end_position_on_the_genomic_accession=113630149, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67226, CurrentID=0, Status='live', Symbol='Tmem19', Aliases='2810428F02Rik, AW125171', description='transmembrane protein 19', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115340739, end_position_on_the_genomic_accession=115362524, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67228, CurrentID=0, Status='live', Symbol='Dph7', Aliases='2810443J12Rik, AW060693, Wdr85', description='diphthamine biosynethesis 7', other_designations='WD repeat domain 85|WD repeat-containing protein 85|WD40 repeat domain 85|diphthamide biosynthesis protein 7', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24962171, end_position_on_the_genomic_accession=24973471, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67229, CurrentID=0, Status='live', Symbol='Prpf18', Aliases='2810441A10Rik', description='pre-mRNA processing factor 18', other_designations='HPRP18 homolog|PRP18 homolog|PRP18 pre-mRNA processing factor 18 homolog', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=4622058, end_position_on_the_genomic_accession=4652142, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67230, CurrentID=0, Status='live', Symbol='Zfp329', Aliases='2810439M05Rik, 4632409L22Rik, ZNF329', description='zinc finger protein 329', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12803778, end_position_on_the_genomic_accession=12818895, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67231, CurrentID=0, Status='live', Symbol='Tbc1d20', Aliases='1110028I04Rik, 2810442O16Rik, AI414693, bs', description='TBC1 domain family, member 20', other_designations='blind-sterile', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152293872, end_position_on_the_genomic_accession=152312590, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67235, CurrentID=0, Status='live', Symbol='Zkscan14', Aliases='2310046C23Rik, 2810437E14Rik, Zfp99, Znf394', description='zinc finger with KRAB and SCAN domains 14', other_designations='zfp-94|zinc finger protein 94|zinc finger protein 99|zinc finger protein with KRAB and SCAN domains 14', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145194946, end_position_on_the_genomic_accession=145204164, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67236, CurrentID=0, Status='live', Symbol='Cinp', Aliases='1810047K05Rik, 2810452K22Rik', description='cyclin-dependent kinase 2 interacting protein', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110872610, end_position_on_the_genomic_accession=110889172, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67238, CurrentID=0, Status='live', Symbol='Fam220a', Aliases='1500026F15Rik, 2810453I06Rik, AW047776, Acpin1, Sipar', description='family with sequence similarity 220, member A', other_designations='STAT3-interacting protein as a repressor|acrosomal protein ACPIN1|salivary gland specific protein SAGSIN1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143548706, end_position_on_the_genomic_accession=143564525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67239, CurrentID=0, Status='live', Symbol='Rpf2', Aliases='2810470K21Rik, AU040229, AU043242, Bxdc1', description='ribosome production factor 2 homolog (S. cerevisiae)', other_designations='brix domain containing 1|brix domain-containing protein 1|ribosome biogenesis protein RPF2 homolog', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40223246, end_position_on_the_genomic_accession=40247039, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  329384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329384, CurrentID=0, Status='live', Symbol='Ptrh1', Aliases='2210013M04Rik, AW046744', description='peptidyl-tRNA hydrolase 1 homolog', other_designations='PTH', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32775738, end_position_on_the_genomic_accession=32777593, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67241, CurrentID=0, Status='live', Symbol='Smc6', Aliases='2810489L22Rik, 3830418C19Rik, AA990493, AU018782, AW742439, SMC-6l1, mKIAA4103, Smc6', description='structural maintenance of chromosomes 6', other_designations='SMC protein 6|SMC6 structural maintenance of chromosomes 6-like 1|mSMC6', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=11265886, end_position_on_the_genomic_accession=11319785, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  67242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67242, CurrentID=0, Status='live', Symbol='Gemin6', Aliases='2610019B15Rik, 2810470M17Rik', description='gem (nuclear organelle) associated protein 6', other_designations='gemin-6', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80207704, end_position_on_the_genomic_accession=80228497, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67245, CurrentID=0, Status='live', Symbol='Peli1', Aliases='2810468L03Rik, A930031K15Rik, AA409794, AI586297, D11Ertd676e', description='pellino 1', other_designations='pellino-1|protein pellino homolog 1', map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=21091324, end_position_on_the_genomic_accession=21150327, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67246, CurrentID=0, Status='live', Symbol='2810474O19Rik', Aliases='6720435I21Rik, AA536804, AU020969, GET, Kiaa1551', description='RIKEN cDNA 2810474O19 gene', other_designations='likely orthologue of H. sapiens chromosome 12 open reading frame 35 (C12orf35)|uncharacterized protein C12orf35 homolog', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=149309414, end_position_on_the_genomic_accession=149335663, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67247, CurrentID=0, Status='live', Symbol='Marc2', Aliases='2810484M10Rik, Mosc2, mARC1', description='mitochondrial amidoxime reducing component 2', other_designations='MOCO sulphurase C-terminal domain containing 2|MOSC domain-containing protein 2, mitochondrial|moco sulfurase C-terminal domain-containing protein 2|molybdenum cofactor sulfurase C-terminal domain-containing protein 2', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=184813068, end_position_on_the_genomic_accession=184846446, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67248, CurrentID=0, Status='live', Symbol='Rpl39', Aliases='2810465O16Rik', description='ribosomal protein L39', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37082520, end_position_on_the_genomic_accession=37085184, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67249, CurrentID=0, Status='live', Symbol='Tbc1d19', Aliases='2810453K03Rik', description='TBC1 domain family, member 19', other_designations='', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=53809581, end_position_on_the_genomic_accession=53909221, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  67252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67252, CurrentID=0, Status='live', Symbol='Cap2', Aliases='2810452G09Rik, AV045235', description='CAP, adenylate cyclase-associated protein, 2 (yeast)', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=46501903, end_position_on_the_genomic_accession=46650281, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67254, CurrentID=0, Status='live', Symbol='2900011O08Rik', Aliases='MINP', description='RIKEN cDNA 2900011O08 gene', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13986637, end_position_on_the_genomic_accession=14101494, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67255, CurrentID=0, Status='live', Symbol='Zfp422', Aliases='2900028O21Rik, AU041373, KOX15, Krox-25-2, Krox-26, Krox25, Krox26, Znf22', description='zinc finger protein 422', other_designations='Kruppel type zinc finger protein Krox26|krox-25 protein|zinc finger protein Krox-25|zinc finger protein Krox-26', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116624016, end_position_on_the_genomic_accession=116628999, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67260, CurrentID=0, Status='live', Symbol='Cers4', Aliases='2900019C14Rik, Lass4, Trh1', description='ceramide synthase 4', other_designations='LAG1 homolog, ceramide synthase 4|LAG1 longevity assurance homolog 4|TRAM homolog 1|longevity assurance homolog 4|translocating chain-associating membrane protein homolog 1', map_location='8 A1.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4492987, end_position_on_the_genomic_accession=4526308, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67263, CurrentID=0, Status='live', Symbol='Zswim6', Aliases='2900036G02Rik, mKIAA1577', description='zinc finger SWIM-type containing 6', other_designations='zinc finger, SWIM domain containing 6', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=107724617, end_position_on_the_genomic_accession=107890064, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67264, CurrentID=0, Status='live', Symbol='Ndufb8', Aliases='2900010I05Rik, AI987932', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex 8', other_designations='CI-ASHI|NADH-ubiquinone oxidoreductase ASHI subunit|complex I-ASHI', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44548576, end_position_on_the_genomic_accession=44555446, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67266, CurrentID=0, Status='live', Symbol='Fam69a', Aliases='2900024C23Rik, AI315274', description='family with sequence similarity 69, member A', other_designations='', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107908042, end_position_on_the_genomic_accession=107987077, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67267, CurrentID=0, Status='live', Symbol='Uqcc2', Aliases='2900010M23Rik, Cbp6, M19, Mnf1', description='ubiquinol-cytochrome c reductase complex assembly factor 2', other_designations='mitochondrial nucleoid factor 1|mitochondrial protein M19', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27122665, end_position_on_the_genomic_accession=27133891, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67268, CurrentID=0, Status='live', Symbol='Myl12a', Aliases='2900073G15Rik, AI838292, C80200', description='myosin, light chain 12A, regulatory, non-sarcomeric', other_designations='NMDA receptor-interacting protein|brain specific myosin regulatory light chain', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=70993793, end_position_on_the_genomic_accession=71002533, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67269, CurrentID=0, Status='live', Symbol='Agtpbp1', Aliases='1700020N17Rik, 2310001G17Rik, 2900054O13Rik, 4930445M19Rik, 5730402G09Rik, CCP1, Nna1, nmf243, pcd', description='ATP/GTP binding protein 1', other_designations='Purkinje cell degeneration|nervous system nuclear protein induced by axotomy protein 1|nuclear ATP/GTP-binding protein', map_location='13 31.87 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59449534, end_position_on_the_genomic_accession=59557364, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  67270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67270, CurrentID=0, Status='live', Symbol='Mrpl42', Aliases='2700009F22Rik, 2900055D03Rik, D10Ertd322e, HSPC204, L31mt, MRP-L31, PTD007, Rpml31', description='mitochondrial ribosomal protein L42', other_designations='28S ribosomal protein L42, mitochondrial|28S ribosomal protein S32, mitochondrial|39S ribosomal protein L31, mitochondrial|L42mt|MRP-L42|MRP-S32|S32mt|mitochndrial ribosomal protein L42', map_location='10 49.39 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=95480805, end_position_on_the_genomic_accession=95501992, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  329416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329416, CurrentID=0, Status='live', Symbol='Nostrin', Aliases='Daip2, mDaIP2', description='nitric oxide synthase trafficker', other_designations='Dab2-interacting protein 2|disabled homolog 2-interacting protein 2|eNOS trafficking inducer|eNOS-trafficking inducer|probable ortholog of human nitric oxide synthase trafficker (NOSTRIN)', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69135800, end_position_on_the_genomic_accession=69189330, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67273, CurrentID=0, Status='live', Symbol='Ndufa10', Aliases='2900053E13Rik', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex 10', other_designations='CI-42kD|NADH-ubiquinone oxidoreductase 42 kDa subunit|complex I-42kD', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92439708, end_position_on_the_genomic_accession=92473758, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67276, CurrentID=0, Status='live', Symbol='Eri1', Aliases="3'hexo, 3110010F15Rik, Thex1, eri-1", description='exoribonuclease 1', other_designations="3'-5' exonuclease ERI1|enhanced RNAi protein 1|eri-1 homolog|histone mRNA 3' end-specific exonuclease|histone mRNA 3'-exonuclease 1|three prime histone mRNA exonuclease 1", map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=35465265, end_position_on_the_genomic_accession=35495533, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  329421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329421, CurrentID=0, Status='live', Symbol='Myo3b', Aliases='A430065P19Rik', description='myosin IIIB', other_designations='myosin IIIB1|myosin IIIB2', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=70039029, end_position_on_the_genomic_accession=70429195, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  67278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67278, CurrentID=0, Status='live', Symbol='Pagr1a', Aliases='2900092E17Rik, C77040, PA1, PAGR1', description='PAXIP1 associated glutamate rich protein 1A', other_designations='PAXIP1-associated glutamate-rich protein 1|PAXIP1-associated protein 1|PTIP-associated 1|PTIP-associated protein 1', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127015051, end_position_on_the_genomic_accession=127017352, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67279, CurrentID=0, Status='live', Symbol='Med31', Aliases='3110004H13Rik, CGI-125, l11Jus15', description='mediator complex subunit 31', other_designations='mediator complex subunit SOH1|mediator of RNA polymerase II transcription, subunit 31 homolog', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72211724, end_position_on_the_genomic_accession=72215592, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67281, CurrentID=0, Status='live', Symbol='Rpl37', Aliases='3110005M08Rik', description='ribosomal protein L37', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=5116613, end_position_on_the_genomic_accession=5119140, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67282, CurrentID=0, Status='live', Symbol='Ccdc53', Aliases='2900091E11Rik, 5730495F03Rik', description='coiled-coil domain containing 53', other_designations='coiled-coil domain-containing protein 53', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88201097, end_position_on_the_genomic_accession=88260817, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67283, CurrentID=0, Status='live', Symbol='Slc25a19', Aliases='2900089E13Rik, DNC, MUP1, TPC', description='solute carrier family 25 (mitochondrial thiamine pyrophosphate carrier), member 19', other_designations='solute carrier family 25 (mitochondrial deoxynucleotide carrier), member 19|solute carrier family 25 member 19', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115614170, end_position_on_the_genomic_accession=115628295, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67285, CurrentID=0, Status='live', Symbol='Cwc27', Aliases='3110009E13Rik, NY-CO-10, Sdccag10', description='CWC27 spliceosome-associated protein', other_designations='CWC27 spliceosome-associated protein homolog|PPIase CWC27|PPIase SDCCAG10|peptidyl-prolyl cis-trans isomerase SDCCAG10|serologically defined colon cancer antigen 10', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104631322, end_position_on_the_genomic_accession=104816964, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  67286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67286, CurrentID=0, Status='live', Symbol='Ift22', Aliases='3110017O03Rik, AI835745, AI847427, Rabl5', description='intraflagellar transport 22', other_designations='RAB, member of RAS oncogene family-like 5|rab-like protein 5', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136908150, end_position_on_the_genomic_accession=136914509, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67287, CurrentID=0, Status='live', Symbol='Parp6', Aliases='1700119G14Rik, 2310028P13Rik, 3110038K10Rik, ARTD17, C030013N01Rik, PARP-6', description='poly (ADP-ribose) polymerase family, member 6', other_designations='ADP-ribosyltransferase diphtheria toxin-like 17', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59617284, end_position_on_the_genomic_accession=59650290, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  67288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67288, CurrentID=0, Status='live', Symbol='Srek1ip1', Aliases='3110031B13Rik, AI462702, AV348305, AW121048, Sfrs12ip1, Srsf12ip1', description='splicing regulatory glutamine/lysine-rich protein 1interacting protein 1', other_designations='SFRS12-interacting protein 1|SREK1-interacting protein 1|protein SFRS12IP1|serine/arginine-rich splicing factor 12, interacting protein 1', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104817239, end_position_on_the_genomic_accession=104838659, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67290, CurrentID=0, Status='live', Symbol='3110040N11Rik', Aliases='AI851475', description='RIKEN cDNA 3110040N11 gene', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81780283, end_position_on_the_genomic_accession=81790462, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67291, CurrentID=0, Status='live', Symbol='Ccdc137', Aliases='3110023B02Rik', description='coiled-coil domain containing 137', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120458129, end_position_on_the_genomic_accession=120464358, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67292, CurrentID=0, Status='live', Symbol='Pigc', Aliases='3110030E07Rik, AW212108', description='phosphatidylinositol glycan anchor biosynthesis, class C', other_designations='PIG-C|phosphatidylinositol-glycan biosynthesis class C protein', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161969188, end_position_on_the_genomic_accession=161973435, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67295, CurrentID=0, Status='live', Symbol='Rab3c', Aliases='2700062I01Rik, 3110015B08Rik, 3110037E15Rik, AI850886', description='RAB3C, member RAS oncogene family', other_designations='small GTP-binding protein Rab3C', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=110054187, end_position_on_the_genomic_accession=110281041, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67296, CurrentID=0, Status='live', Symbol='Socs4', Aliases='3110032M18Rik, A730004F22Rik, AI427843, AU015727, AU015859, Socs7', description='suppressor of cytokine signaling 4', other_designations='SOCS-4|SOCS-7|suppressor of cytokine signaling 7|suppressor of cytokine signalling 4', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47277143, end_position_on_the_genomic_accession=47291591, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67298, CurrentID=0, Status='live', Symbol='Gprasp1', Aliases='2210415K24Rik, 3110031O14Rik, C87852, GASP, GASP1', description='G protein-coupled receptor associated sorting protein 1', other_designations='GASP-1', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135742692, end_position_on_the_genomic_accession=135803468, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67299, CurrentID=0, Status='live', Symbol='Dock7', Aliases='3110056M06Rik, Gm430, m', description='dedicator of cytokinesis 7', other_designations='protein moonlight', map_location='4 45.6 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=98936659, end_position_on_the_genomic_accession=99122605, orientation='minus', exon_count=51, OMIM='', no_hdr0=''),
  67300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67300, CurrentID=0, Status='live', Symbol='Cltc', Aliases='3110065L21Rik, CHC, R74732', description='clathrin, heavy polypeptide (Hc)', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86694610, end_position_on_the_genomic_accession=86757742, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  67302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67302, CurrentID=0, Status='live', Symbol='Zc3h13', Aliases='2600010B19Rik, 3110050K21Rik, 4930570G11Rik, C87618', description='zinc finger CCCH type containing 13', other_designations='', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75284308, end_position_on_the_genomic_accession=75344433, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  67305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67305, CurrentID=0, Status='live', Symbol='Gpx7', Aliases='3110050F08Rik, AI327032, GPX6', description='glutathione peroxidase 7', other_designations='GPx-7|GSHPx-7', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108400217, end_position_on_the_genomic_accession=108406713, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67306, CurrentID=0, Status='live', Symbol='Zc2hc1a', Aliases='3110050N22Rik, AI790358, AU023959, Fam164a', description='zinc finger, C2HC-type containing 1A', other_designations='family with sequence similarity 164, member A|protein FAM164A', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=7503415, end_position_on_the_genomic_accession=7553848, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67307, CurrentID=0, Status='live', Symbol='Pbld2', Aliases='3110049J23Rik, Mawbp', description='phenazine biosynthesis-like protein domain containing 2', other_designations='MAWD binding protein homolog 2', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63024049, end_position_on_the_genomic_accession=63058813, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67308, CurrentID=0, Status='live', Symbol='Mrpl46', Aliases='3110052F15Rik, C15orf4, C80107, LIECG2, P2ECSL', description='mitochondrial ribosomal protein L46', other_designations='L46mt|MRP-L46', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=78775341, end_position_on_the_genomic_accession=78783089, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67310, CurrentID=0, Status='live', Symbol='Prl8a9', Aliases='1600019F10Rik, Ghd15, PLP-C2, PLP-Cbeta, Prlpc2', description='prolactin family8, subfamily a, member 9', other_designations='PLP C-beta|PRL-like protein C2|growth hormone d15|placental prolactin-like protein C2|prolactin-like protein C 2|prolactin-like protein C-beta', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27558001, end_position_on_the_genomic_accession=27564616, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67311, CurrentID=0, Status='live', Symbol='Nanp', Aliases='1600031M04Rik, Hdhd4', description='N-acetylneuraminic acid phosphatase', other_designations='haloacid dehalogenase-like hydrolase domain containing 4|haloacid dehalogenase-like hydrolase domain-containing protein 4|neu5Ac-9-Pase', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151029685, end_position_on_the_genomic_accession=151039379, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67313, CurrentID=0, Status='live', Symbol='5730559C18Rik', Aliases='1700034M08Rik, 4933426C09Rik, AI586180, D1Mgi54', description='RIKEN cDNA 5730559C18 gene', other_designations='', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136213522, end_position_on_the_genomic_accession=136234289, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67315, CurrentID=0, Status='live', Symbol='Ceacam12', Aliases='1600031J20Rik-C1, Ceacam12-C3, Ceacam12C1, Ceacam12C3, Ceacam12', description='carcinoembryonic antigen-related cell adhesion molecule 12', other_designations='CEA-related cell adhesion molecule 12', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18065929, end_position_on_the_genomic_accession=18077986, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67317, CurrentID=0, Status='live', Symbol='1700022I11Rik', Aliases='', description='RIKEN cDNA 1700022I11 gene', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42969946, end_position_on_the_genomic_accession=42974325, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67320, CurrentID=0, Status='live', Symbol='Iqcf4', Aliases='1700042N06Rik, mtIQ1', description='IQ motif containing F4', other_designations='testis-specific IQ motif containing protein 1', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106568319, end_position_on_the_genomic_accession=106570967, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67323, CurrentID=0, Status='live', Symbol='1700042G07Rik', Aliases='', description='RIKEN cDNA 1700042G07 gene', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116173372, end_position_on_the_genomic_accession=116174297, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67326, CurrentID=0, Status='live', Symbol='1700037H04Rik', Aliases='', description='RIKEN cDNA 1700037H04 gene', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131146315, end_position_on_the_genomic_accession=131160086, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67328, CurrentID=0, Status='live', Symbol='Lyzl1', Aliases='1700038F02Rik, Lyc2, Lyzd1', description='lysozyme-like 1', other_designations='lysozyme d1|lysozyme-like protein 2', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=4165832, end_position_on_the_genomic_accession=4182236, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67331, CurrentID=0, Status='live', Symbol='Atp8b3', Aliases='1700042F02Rik, 1700056N23Rik, ATPIK, SAPLT', description='ATPase, class I, type 8B, member 3', other_designations='sperm aminophospholipid transporter|spermatozoan aminophospholipid translocase', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80519584, end_position_on_the_genomic_accession=80539150, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  67332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67332, CurrentID=0, Status='live', Symbol='Snrpd3', Aliases='1700043E15Rik, 2310009E13Rik, AW046420, SMD3', description='small nuclear ribonucleoprotein D3', other_designations='sm-D3|snRNP core protein D3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75518042, end_position_on_the_genomic_accession=75535440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67333, CurrentID=0, Status='live', Symbol='Stk35', Aliases='1700054C12Rik, CLIK1l1, Stk35', description='serine/threonine kinase 35', other_designations='CLP-36 interacting kinase|serine/threonine-protein kinase 35 L1', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129800517, end_position_on_the_genomic_accession=129832287, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67334, CurrentID=0, Status='live', Symbol='Hypm', Aliases='1700054O13Rik, H2al3', description='huntingtin interacting protein M', other_designations='huntingtin yeast partner M', map_location='X A1.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9846947, end_position_on_the_genomic_accession=9847500, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67337, CurrentID=0, Status='live', Symbol='Cstf1', Aliases='1700057K18Rik, AI788832', description="cleavage stimulation factor, 3' pre-RNA, subunit 1", other_designations='CF-1 50 kDa subunit|CSTF 50 kDa subunit|cleavage stimulation factor 50 kDa subunit|cstF-50', map_location='2 94.84 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172370703, end_position_on_the_genomic_accession=172382449, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67338, CurrentID=0, Status='live', Symbol='Rffl', Aliases='1700051E09Rik, 4930516L10Rik, BG080975, Carp2', description='ring finger and FYVE like domain containing protein', other_designations='RING finger and FYVE-like domain-containing protein 1|fring|rififylin', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82803819, end_position_on_the_genomic_accession=82871239, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67341, CurrentID=0, Status='live', Symbol='Ascl4', Aliases='1700054F16Rik, bHLHa44', description='achaete-scute family bHLH transcription factor 4', other_designations='achaete-scute complex homolog 4', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85928491, end_position_on_the_genomic_accession=85929647, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67343, CurrentID=0, Status='live', Symbol='Tex43', Aliases='1700065I17Rik, Tseg7', description='testis expressed 43', other_designations='testis specific expressed protein 7', map_location='18 D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=56587613, end_position_on_the_genomic_accession=56594783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67344, CurrentID=0, Status='live', Symbol='Tctex1d1', Aliases='1700055O19Rik', description='Tctex1 domain containing 1', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=102978388, end_position_on_the_genomic_accession=103005594, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67345, CurrentID=0, Status='live', Symbol='Herc4', Aliases='1700056O17Rik, 4921531D01Rik, 9530080M15Rik, mKIAA1593', description='hect domain and RLD 4', other_designations='HECT domain and RCC1-like domain-containing protein 4', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63243782, end_position_on_the_genomic_accession=63317881, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  67356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67356, CurrentID=0, Status='live', Symbol='Tmco5', Aliases='1700095F04Rika, Tmco5', description='transmembrane and coiled-coil domains 5', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=116878691, end_position_on_the_genomic_accession=116892500, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67358, CurrentID=0, Status='live', Symbol='1700093K21Rik', Aliases='', description='RIKEN cDNA 1700093K21 gene', other_designations='', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23516203, end_position_on_the_genomic_accession=23521317, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  329504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329504, CurrentID=0, Status='live', Symbol='Lcmt2', Aliases='D330024M17, Tyw4', description='leucine carboxyl methyltransferase 2', other_designations='p21WAF1/CIP1 promoter-interacting protein|tRNA wybutosine-synthesizing protein 4 homolog|tRNA yW-synthesizing protein 4 homolog|tRNA(Phe) (7-(3-amino-3-(methoxycarbonyl)propyl)wyosine(37)-N)-methoxycarbonyltransferase|tRNA(Phe) (7-(3-amino-3-carboxypropyl)wyosine(37)-O)-methyltransferase', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121137292, end_position_on_the_genomic_accession=121140698, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  329506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329506, CurrentID=0, Status='live', Symbol='Ctdspl2', Aliases='D2Ertd485e', description='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase like 2', other_designations='CTDSP-like 2', map_location='2 60.46 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121956001, end_position_on_the_genomic_accession=122015326, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  67365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67365, CurrentID=0, Status='live', Symbol='Hdhd1a', Aliases='1700121L12Rik, GS1, Hdhd1, Pudp', description='haloacid dehalogenase-like hydrolase domain containing 1A', other_designations="5'-PsiMPase|haloacid dehalogenase-like hydrolase domain-containing protein 1|haloacid dehalogenase-like hydrolase domain-containing protein 1A|pseudouridine-5'-monophosphatase", map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=50567656, end_position_on_the_genomic_accession=50568699, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  67366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67366, CurrentID=0, Status='live', Symbol='Tmed11', Aliases='1810008K16Rik', description='transmembrane emp24 protein transport domain containing', other_designations='GP25L|glycoprotein 25L|p24 family protein alpha-1|p24alpha1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108777235, end_position_on_the_genomic_accession=108795363, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67367, CurrentID=0, Status='live', Symbol='Paxbp1', Aliases='1810007M14Rik, C21orf66, Gcfc, Gcfc1, Pax3/7bp', description='PAX3 and PAX7 binding protein 1', other_designations='GC-rich sequence DNA binding factor 1|GC-rich sequence DNA-binding factor 1|GC-rich sequence DNA-binding factor homolog', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91014037, end_position_on_the_genomic_accession=91044734, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  67369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67369, CurrentID=0, Status='live', Symbol='Qpctl', Aliases='1810019P04Rik, BB101812, gQC', description='glutaminyl-peptide cyclotransferase-like', other_designations='golgi-resident glutaminyl-peptide cyclotransferase|isoQC', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19140217, end_position_on_the_genomic_accession=19149196, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67370, CurrentID=0, Status='live', Symbol='Zfp606', Aliases='2410022M24Rik, AW494727', description='zinc finger protein 606', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12478160, end_position_on_the_genomic_accession=12496235, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67371, CurrentID=0, Status='live', Symbol='Gtf3c6', Aliases='2410016F19Rik, AU019813', description='general transcription factor IIIC, polypeptide 6, alpha', other_designations='TFIIIC 35 kDa subunit|TFIIIC35|gtf3c6|transcription factor IIIC 35 kDa subunit|transcription factor IIIC subunit 6', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40247701, end_position_on_the_genomic_accession=40258158, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67373, CurrentID=0, Status='live', Symbol='2210010C04Rik', Aliases='AV072249', description='RIKEN cDNA 2210010C04 gene', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41030268, end_position_on_the_genomic_accession=41035509, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67374, CurrentID=0, Status='live', Symbol='Jam2', Aliases='1110002N23Rik, 2410030G21Rik, 2410167M24Rik, JAM-2, JAM-B, Jcam2, VE-JAM', description='junction adhesion molecule 2', other_designations='junction adhesion molecule B|junction cell adhesion molecule 2|junctional adhesion molecule 2|vascular endothelial junction-associated molecule', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=84774123, end_position_on_the_genomic_accession=84826380, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67375, CurrentID=0, Status='live', Symbol='Qprt', Aliases='2410027J01Rik, AI647766, QPRTase', description='quinolinate phosphoribosyltransferase', other_designations='QAPRTase|nicotinate-nucleotide pyrophosphorylase', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127107770, end_position_on_the_genomic_accession=127122029, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67378, CurrentID=0, Status='live', Symbol='Bbs2', Aliases='2410125H22Rik, AI447581', description='Bardet-Biedl syndrome 2 (human)', other_designations='Bardet-Biedl syndrome 2 homolog', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94067954, end_position_on_the_genomic_accession=94099370, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  67379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67379, CurrentID=0, Status='live', Symbol='Dedd2', Aliases='2410050E11Rik, FLAME-3', description='death effector domain-containing DNA binding protein 2', other_designations='DED-containing protein FLAME-3|FADD-like anti-apoptotic molecule 3|death effector domain-containing DNA-binding protein 2', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25202838, end_position_on_the_genomic_accession=25219873, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67381, CurrentID=0, Status='live', Symbol='Med4', Aliases='2410046H15Rik, DRIP36, HSPC126, TRAP36, Vdirp, Vdrip', description='mediator complex subunit 4', other_designations='mediator of RNA polymerase II transcription, subunit 4 homolog|p36 TRAP/SMCC/PC2 subunit|vitamin D receptor interacting protein', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73510049, end_position_on_the_genomic_accession=73518545, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67382, CurrentID=0, Status='live', Symbol='Brd3', Aliases='2410084F24Rik, AW060456, Fsrg2, ORFX, RINGL3', description='bromodomain containing 3', other_designations='bromodomain-containing 3|bromodomain-containing FSH-like protein FSRG2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27445579, end_position_on_the_genomic_accession=27507657, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  67383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67383, CurrentID=0, Status='live', Symbol='Carnmt1', Aliases='2410127L17Rik', description='carnosine N-methyltransferase 1', other_designations='UPF0586 protein C9orf41 homolog', map_location='19 E1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=18670736, end_position_on_the_genomic_accession=18706674, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67384, CurrentID=0, Status='live', Symbol='Bag4', Aliases='2410112I15Rik, SODD', description='BCL2-associated athanogene 4', other_designations='BAG-4|bcl-2-associated athanogene 4|silencer of death domains', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25764538, end_position_on_the_genomic_accession=25785209, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67387, CurrentID=0, Status='live', Symbol='Unc50', Aliases='1110002A21Rik, GMH1, UNCL, URP', description='unc-50 homolog (C. elegans)', other_designations='PDLs22|periodontal ligament-specific protein 22', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37430144, end_position_on_the_genomic_accession=37439124, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67388, CurrentID=0, Status='live', Symbol='1110008F13Rik', Aliases='AI316789', description='RIKEN cDNA 1110008F13 gene', other_designations='novel transcript', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156863122, end_position_on_the_genomic_accession=156873563, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67389, CurrentID=0, Status='live', Symbol='Fam132a', Aliases='1110035L05Rik, C1qdc2, CTRP12, alipolin', description='family with sequence similarity 132, member A', other_designations='C1q domain containing 2|adipose-derived insulin-sensitizing factor|complement C1q tumor necrosis factor-related protein 12|protein FAM132A', map_location='4 87.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155962312, end_position_on_the_genomic_accession=155966629, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67390, CurrentID=0, Status='live', Symbol='Rnmtl1', Aliases='4833420N02Rik, AI255153, HC90', description='RNA methyltransferase like 1', other_designations="16S rRNA (guanosine(1370)-2'-O)-methyltransferase|16S rRNA [Gm1370] 2'-O-methyltransferase|RNA methyltransferase-like protein 1", map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76243736, end_position_on_the_genomic_accession=76250625, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67391, CurrentID=0, Status='live', Symbol='Fundc2', Aliases='4833415N24Rik, DC44, HCBP6', description='FUN14 domain containing 2', other_designations='hepatitis C virus core-binding protein 6', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75382399, end_position_on_the_genomic_accession=75397158, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67392, CurrentID=0, Status='live', Symbol='4833420G17Rik', Aliases='2210017J06Rik, 5730533D17Rik, AW107518, C85193', description='RIKEN cDNA 4833420G17 gene', other_designations='', map_location='13 D2.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119462015, end_position_on_the_genomic_accession=119486120, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67393, CurrentID=0, Status='live', Symbol='Cxxc5', Aliases='4930415K17Rik', description='CXXC finger 5', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35828924, end_position_on_the_genomic_accession=35861693, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  329540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329540, CurrentID=0, Status='live', Symbol='Nol4l', Aliases='4931410H15, 8430427H17Rik, AI429157, Gm1636', description='nucleolar protein 4-like', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153407461, end_position_on_the_genomic_accession=153529971, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  67397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67397, CurrentID=0, Status='live', Symbol='Erp29', Aliases='1200015M03Rik, 2810446M09Rik, AW209030, Erp28, Erp31, PDI-Db', description='endoplasmic reticulum protein 29', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121444753, end_position_on_the_genomic_accession=121452474, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67398, CurrentID=0, Status='live', Symbol='Srpr', Aliases='1300011P19Rik, D11Mgi27', description="signal recognition particle receptor ('docking protein')", other_designations='DP-alpha|SR-alpha|docking protein alpha', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35211203, end_position_on_the_genomic_accession=35217003, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67399, CurrentID=0, Status='live', Symbol='Pdlim7', Aliases='LMP', description='PDZ and LIM domain 7', other_designations='LIM mineralization protein|protein enigma', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55497487, end_position_on_the_genomic_accession=55513473, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67402, CurrentID=0, Status='live', Symbol='Txndc8', Aliases='4930429J24Rik, Sptrx-3, Sptrx3, Trx6', description='thioredoxin domain containing 8', other_designations='spermatid-specific thioredoxin-3|thioredoxin-6', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=57984021, end_position_on_the_genomic_accession=58009135, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  329547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329547, CurrentID=0, Status='live', Symbol='Bpi', Aliases='9230105K17Rikfd1, Bpi', description='bactericidal permeablility increasing protein', other_designations='bactericidal/permeability-increasing protein', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158258091, end_position_on_the_genomic_accession=158284533, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  67405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67405, CurrentID=0, Status='live', Symbol='Nts', Aliases='5033428E16Rik, NMN-125, NN, NT, NT/N, NTS1', description='neurotensin', other_designations='neuromedin N', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=102481756, end_position_on_the_genomic_accession=102490418, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67407, CurrentID=0, Status='live', Symbol='Cylc1', Aliases='4930401M13Rik', description='cylicin, basic protein of sperm head cytoskeleton 1', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=111110418, end_position_on_the_genomic_accession=111125961, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67410, CurrentID=0, Status='live', Symbol='4930449I24Rik', Aliases='', description='RIKEN cDNA 4930449I24 gene', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146502343, end_position_on_the_genomic_accession=146505183, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67412, CurrentID=0, Status='live', Symbol='Soga3', Aliases='6330407J23Rik', description='SOGA family member 3', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=29143274, end_position_on_the_genomic_accession=29199630, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  329557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329557, CurrentID=0, Status='live', Symbol='Svs3b', Aliases='9530004A22Rik', description='seminal vesicle secretory protein 3B', other_designations='SVS III', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164254363, end_position_on_the_genomic_accession=164256643, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67414, CurrentID=0, Status='live', Symbol='Mfn1', Aliases='2310002F04Rik, 6330416C07Rik, D3Ertd265e, HR2, mKIAA4032', description='mitofusin 1', other_designations='transmembrane GTPase MFN1', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32529464, end_position_on_the_genomic_accession=32579225, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  329559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329559, CurrentID=0, Status='live', Symbol='Zfp335', Aliases='1810045J01Rik, NIF-1, Nif1, Znf335', description='zinc finger protein 335', other_designations='NRC-interacting factor 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164891882, end_position_on_the_genomic_accession=164911750, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  67416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67416, CurrentID=0, Status='live', Symbol='Armcx2', Aliases='3230401N03Rik, AI043003, ALEX2', description='armadillo repeat containing, X-linked 2', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134804142, end_position_on_the_genomic_accession=134809221, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67417, CurrentID=0, Status='live', Symbol='Ears2', Aliases='3230401I01Rik, mKIAA1970', description='glutamyl-tRNA synthetase 2, mitochondrial', other_designations='gluRS|glutamate--tRNA ligase|glutamyl-tRNA synthetase 2 (mitochondrial)(putative)|probable glutamyl-tRNA synthetase, mitochondrial', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122038654, end_position_on_the_genomic_accession=122067185, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67418, CurrentID=0, Status='live', Symbol='Ppil4', Aliases='3732410E19Rik, 3830425H19Rik, AI788954, AW146233, PPIase', description='peptidylprolyl isomerase (cyclophilin)-like 4', other_designations='cyclophilin-like protein PPIL4|rotamase PPIL4', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7792891, end_position_on_the_genomic_accession=7822563, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67419, CurrentID=0, Status='live', Symbol='3632451O06Rik', Aliases='AU067705', description='RIKEN cDNA 3632451O06 gene', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49681664, end_position_on_the_genomic_accession=49783411, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67420, CurrentID=0, Status='live', Symbol='Far1', Aliases='2600011M19Rik, 2900034E22Rik, 3732409C05Rik, AI850429, Mlstd2', description='fatty acyl CoA reductase 1', other_designations='male sterility domain containing 2', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=113513834, end_position_on_the_genomic_accession=113571506, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67422, CurrentID=0, Status='live', Symbol='Dhdds', Aliases='3222401G21Rik, CIT, DS, HDS, W91638', description='dehydrodolichyl diphosphate synthase', other_designations='cis-IPTase|cis-isoprenyltransferase|dedol-PP synthase', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133969035, end_position_on_the_genomic_accession=134000956, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67425, CurrentID=0, Status='live', Symbol='Eps8l1', Aliases='2310051G19Rik, 4632407K17Rik, AW060268, DRC3, EPS8R1', description='EPS8-like 1', other_designations='EPS8-like protein 1|EPS8-related protein 1|epidermal growth factor receptor pathway substrate 8-related protein 1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4460631, end_position_on_the_genomic_accession=4480487, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  67426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67426, CurrentID=0, Status='live', Symbol='Adck3', Aliases='4632432J16Rik, AI462003, Cabc1, mKIAA0451', description='aarF domain containing kinase 3', other_designations='aarF domain-containing protein kinase 3|chaperone activity of bc1 complex-like, mitochondrial|chaperone, ABC1 activity of bc1 complex like', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180165238, end_position_on_the_genomic_accession=180196020, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  67427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67427, CurrentID=0, Status='live', Symbol='Rps20', Aliases='4632426K06Rik, Dsk4', description='ribosomal protein S20', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3834473, end_position_on_the_genomic_accession=3835600, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67429, CurrentID=0, Status='live', Symbol='Nudcd1', Aliases='4921532K09Rik, AA407246, AW260430, AW556235, CML66-L, Cml66', description='NudC domain containing 1', other_designations='chronic myelogenous leukemia tumor antigen 66|nudcd1', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=44375227, end_position_on_the_genomic_accession=44428307, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67430, CurrentID=0, Status='live', Symbol='4921536K21Rik', Aliases='AV043724', description='RIKEN cDNA 4921536K21 gene', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3886088, end_position_on_the_genomic_accession=3895126, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  329575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329575, CurrentID=0, Status='live', Symbol='Gm14325', Aliases='OTTMUSG00000016325', description='predicted gene 14325', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=177828991, end_position_on_the_genomic_accession=177840318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67432, CurrentID=0, Status='live', Symbol='Hoga1', Aliases='0610010D20Rik, Dhdpsl, Npl2', description='4-hydroxy-2-oxoglutarate aldolase 1', other_designations='DHDPS-like protein|N-acetylneuraminate pyruvate lyase 2 (putative)|dihydrodipicolinate synthase-like, mitochondrial|probable 2-keto-4-hydroxyglutarate aldolase|probable 4-hydroxy-2-oxoglutarate aldolase, mitochondrial|probable KHG-aldolase', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42045851, end_position_on_the_genomic_accession=42070939, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67433, CurrentID=0, Status='live', Symbol='Ccdc127', Aliases='0610011N22Rik, 5430420E17Rik, A430066A18, AI256405, AW050060, C130002N06', description='coiled-coil domain containing 127', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74350317, end_position_on_the_genomic_accession=74365783, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67434, CurrentID=0, Status='live', Symbol='Ankrd33b', Aliases='0610012A05Rik, 3021401C12Rik, 5730557B15Rik', description='ankyrin repeat domain 33B', other_designations='', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=31291227, end_position_on_the_genomic_accession=31368894, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  329581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329581, CurrentID=0, Status='live', Symbol='Birc7', Aliases='E130019N06, KIAP, Livin, ML-IAP', description='baculoviral IAP repeat-containing 7 (livin)', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180928152, end_position_on_the_genomic_accession=180935028, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67439, CurrentID=0, Status='live', Symbol='Xab2', Aliases='0610041O14Rik, AV025587', description='XPA binding protein 2', other_designations='XPA-binding protein 2', map_location='8 1.92 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3610087, end_position_on_the_genomic_accession=3621334, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  67440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67440, CurrentID=0, Status='live', Symbol='Mtpap', Aliases='0610027A18Rik, AW551379, Papd1', description='mitochondrial poly(A) polymerase', other_designations='PAP associated domain containing 1|PAP-associated domain-containing protein 1|polynucleotide adenylyltransferase', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=4375592, end_position_on_the_genomic_accession=4398800, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67441, CurrentID=0, Status='live', Symbol='Isoc2b', Aliases='0610042E07Rik, AI553388, Isoc2', description='isochorismatase domain containing 2b', other_designations='isochorismatase domain-containing protein 2B, mitochondrial', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4844959, end_position_on_the_genomic_accession=4866179, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67442, CurrentID=0, Status='live', Symbol='Retsat', Aliases='0610039N19Rik, C80029, MMT-7, Ppsig', description='retinol saturase (all trans retinol 13,14 reductase)', other_designations='PPAR-alpha-regulated and starvation-induced gene protein|all-trans-13,14-dihydroretinol saturase', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72598628, end_position_on_the_genomic_accession=72607488, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67443, CurrentID=0, Status='live', Symbol='Map1lc3b', Aliases='1010001C15Rik, Atg8, LC3b, Map1lc3', description='microtubule-associated protein 1 light chain 3 beta', other_designations='MAP1 light chain 3-like protein 2|MAP1A/MAP1B LC3 B|MAP1A/MAP1B light chain 3 B|autophagy-related protein LC3 B|autophagy-related ubiquitin-like modifier LC3 B', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121590468, end_position_on_the_genomic_accession=121598047, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67444, CurrentID=0, Status='live', Symbol='Ilkap', Aliases='0710007A14Rik, 1600009O09Rik, PP2C-DELTA', description='integrin-linked kinase-associated serine/threonine phosphatase 2C', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91373846, end_position_on_the_genomic_accession=91398841, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67445, CurrentID=0, Status='live', Symbol='C1qtnf4', Aliases='0710001E10Rik, 9430004J15Rik, CTRP4', description='C1q and tumor necrosis factor related protein 4', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90885383, end_position_on_the_genomic_accession=90890526, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67446, CurrentID=0, Status='live', Symbol='Dusp28', Aliases='0710001B24Rik, AV005521', description='dual specificity phosphatase 28', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92906989, end_position_on_the_genomic_accession=92908620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67448, CurrentID=0, Status='live', Symbol='Plxdc2', Aliases='1200007L24Rik, 5430431D22Rik, AI646728, AU022916, AV231634, Tem7r', description='plexin domain containing 2', other_designations='tumor endothelial marker 7 related|tumor endothelial marker 7-related protein', map_location='2 A2-A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=16356210, end_position_on_the_genomic_accession=16755839, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67451, CurrentID=0, Status='live', Symbol='Pkp2', Aliases='1200008D14Rik, 1200012P04Rik, AA516617l, Pkp2', description='plakophilin 2', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16213345, end_position_on_the_genomic_accession=16272712, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67452, CurrentID=0, Status='live', Symbol='Pnpla8', Aliases='1200006O19Rik, AI467579, Ipla2(gamma)', description='patatin-like phospholipase domain containing 8', other_designations='iPLA2 gamma|iPLA2-gamma|intracellular membrane-associated calcium-independent phospholipase A2 gamma|patatin-like phospholipase domain-containing protein 8', map_location='12 B3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=44269128, end_position_on_the_genomic_accession=44315935, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67453, CurrentID=0, Status='live', Symbol='Slc25a46', Aliases='1200007B05Rik, AI325987', description='solute carrier family 25, member 46', other_designations='', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31580168, end_position_on_the_genomic_accession=31609902, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67454, CurrentID=0, Status='live', Symbol='Ikbip', Aliases='1200009F10Rik, 1700023M03Rik, D18354, Ikip', description='IKBKB interacting protein', other_designations='IKBKB-interacting protein|IKK-interacting protein|i kappa-B kinase interacting protein|i kappa-B kinase-interacting protein', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=91083039, end_position_on_the_genomic_accession=91102613, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67455, CurrentID=0, Status='live', Symbol='Klhl13', Aliases='1200009K10Rik, AI451128, Bklhd2, D130072F20Rik', description='kelch-like 13', other_designations='BTB and kelch domain-containing protein 2', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=23219271, end_position_on_the_genomic_accession=23365107, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  67456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67456, CurrentID=0, Status='live', Symbol='Ergic2', Aliases='1200009B18Rik, 4930572C01Rik, AA408438, Ptx1', description='ERGIC and golgi 2', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148179318, end_position_on_the_genomic_accession=148212374, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  67457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67457, CurrentID=0, Status='live', Symbol='Frmd8', Aliases='1200004M23Rik, 2310035N23Rik, 4931429L16Rik, AU018809', description='FERM domain containing 8', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5850974, end_position_on_the_genomic_accession=5875270, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67458, CurrentID=0, Status='live', Symbol='Ergic1', Aliases='1200007D18Rik, maa-136', description='endoplasmic reticulum-golgi intermediate compartment (ERGIC) 1', other_designations='ER-Golgi intermediate compartment 32 kDa protein|ERGIC-32', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26560762, end_position_on_the_genomic_accession=26656934, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67459, CurrentID=0, Status='live', Symbol='Nvl', Aliases='1200009I24Rik', description='nuclear VCP-like', other_designations='', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181087140, end_position_on_the_genomic_accession=181144215, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  67460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67460, CurrentID=0, Status='live', Symbol='Decr1', Aliases='1200012F07Rik, Decr, Nadph', description='2,4-dienoyl CoA reductase 1, mitochondrial', other_designations='4-enoyl-CoA reductase', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=15917240, end_position_on_the_genomic_accession=15945507, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67463, CurrentID=0, Status='live', Symbol='Poc5', Aliases='1200014M14Rik, AV144706', description='POC5 centriolar protein', other_designations='POC5 centriolar protein homolog|protein of centriole 5', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=96388294, end_position_on_the_genomic_accession=96415587, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67464, CurrentID=0, Status='live', Symbol='Entpd4', Aliases='1200014F22Rik, 4930513O14Rik, A430102M06Rik, AI414893, Lysal1, mKIAA0392', description='ectonucleoside triphosphate diphosphohydrolase 4', other_designations='NTPDase 4|UDPase|lysosomal apyrase-like 1|lysosomal apyrase-like protein of 70 kDa|uridine-diphosphatase', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69337149, end_position_on_the_genomic_accession=69366742, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67465, CurrentID=0, Status='live', Symbol='Sf3a1', Aliases='1200014H24Rik, 5930416L09Rik, AI159724', description='splicing factor 3a, subunit 1', other_designations='SF3a120|splicing factor 3 subunit 1', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4160354, end_position_on_the_genomic_accession=4182541, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67466, CurrentID=0, Status='live', Symbol='Pdcl', Aliases='1200011E13Rik, AW108059, PhLP1', description='phosducin-like', other_designations='PHLP', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37350074, end_position_on_the_genomic_accession=37359820, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67467, CurrentID=0, Status='live', Symbol='Gpalpp1', Aliases='1200011I18Rik, Kiaa1704', description='GPALPP motifs containing 1', other_designations='', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=76086232, end_position_on_the_genomic_accession=76110815, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67468, CurrentID=0, Status='live', Symbol='Mmd', Aliases='1200017E07Rik, 1810073C06Rik, AA690185', description='monocyte to macrophage differentiation-associated', other_designations='monocyte to macrophage differentiation protein|progestin and adipoQ receptor family member XI', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=90187687, end_position_on_the_genomic_accession=90278574, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67469, CurrentID=0, Status='live', Symbol='Abhd5', Aliases='1300003D03Rik, 2010002J10Rik, CDS, CGI-58, IECN5, NCIE2', description='abhydrolase domain containing 5', other_designations='abhydrolase domain-containing protein 5|lipid droplet-binding protein CGI-58', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122351539, end_position_on_the_genomic_accession=122381524, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67470, CurrentID=0, Status='live', Symbol='Abcg8', Aliases='1300003C16Rik, AI114946, sterolin-2', description='ATP-binding cassette, sub-family G (WHITE), member 8', other_designations='sterolin 2', map_location='17 55.02 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84676302, end_position_on_the_genomic_accession=84700333, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67471, CurrentID=0, Status='live', Symbol='Gpatch1', Aliases='1300003A17Rik, ECGP, Gpatc1', description='G patch domain containing 1', other_designations='evolutionarily conserved G-patch domain containing', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35276544, end_position_on_the_genomic_accession=35318440, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  67472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67472, CurrentID=0, Status='live', Symbol='Mtfr1', Aliases='1300002C08Rik, 1700080D04Rik, 2810026O10Rik, 4930579E05Rik, Chppr, Fam54a2', description='mitochondrial fission regulator 1', other_designations='chondrocyte protein with a poly-proline region', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19187326, end_position_on_the_genomic_accession=19220817, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67473, CurrentID=0, Status='live', Symbol='Slc47a1', Aliases='1300013J15Rik, AI036982, MATE1, mMATE1', description='solute carrier family 47, member 1', other_designations='MATE-1|mMATE-1', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61343400, end_position_on_the_genomic_accession=61378781, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  67474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67474, CurrentID=0, Status='live', Symbol='Snap29', Aliases='1300018G05Rik, AI891940, AU020222, BB131856, Gs32', description='synaptosomal-associated protein 29', other_designations='SNAP-29|golgi SNARE of 32 kDa|soluble 29 kDa NSF attachment protein|vesicle-membrane fusion protein SNAP-29', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17406000, end_position_on_the_genomic_accession=17430826, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67475, CurrentID=0, Status='live', Symbol='Ero1lb', Aliases='1300013B24Rik, 1700065B09Rik, AI447560, Ero1b, ero1-beta', description='ERO1-like beta (S. cerevisiae)', other_designations='ERO1-L-beta|endoplasmic oxidoreductase 1 beta|endoplasmic oxidoreductin-1-like protein B|endoplasmic reticulum oxidoreductase beta|endoplasmic reticulum oxidoreductin-1-like protein B|oxidoreductin-1-L-beta', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12565809, end_position_on_the_genomic_accession=12609530, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  67477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67477, CurrentID=0, Status='live', Symbol='Abhd15', Aliases='1300007F04Rik', description='abhydrolase domain containing 15', other_designations='alpha/beta hydrolase', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77515117, end_position_on_the_genomic_accession=77520628, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67480, CurrentID=0, Status='live', Symbol='Cwc25', Aliases='1300013D05Rik, AA473636, Ccdc49, R75228', description='CWC25 spliceosome-associated protein', other_designations='CWC25 spliceosome-associated protein homolog|coiled-coil domain containing 49|coiled-coil domain-containing protein 49|spliceosome-associated protein homolog CWC25', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97745470, end_position_on_the_genomic_accession=97766613, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  67483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67483, CurrentID=0, Status='live', Symbol='1700028P14Rik', Aliases='AV209940', description='RIKEN cDNA 1700028P14 gene', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=23558760, end_position_on_the_genomic_accession=23652812, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  329628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329628, CurrentID=0, Status='live', Symbol='Fat4', Aliases='6030410K14Rik, 9430004M15', description='FAT atypical cadherin 4', other_designations='FAT tumor suppressor homolog 4|fat-like cadherin protein FAT-J', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=38886487, end_position_on_the_genomic_accession=39011987, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  67486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67486, CurrentID=0, Status='live', Symbol='Polr3g', Aliases='2310047G20Rik, AV275904, RPC32, RPC7', description='polymerase (RNA) III (DNA directed) polypeptide G', other_designations='DNA-directed RNA polymerase III subunit G|RNA polymerase III subunit C7', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=81673837, end_position_on_the_genomic_accession=81711013, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67487, CurrentID=0, Status='live', Symbol='Dhx40', Aliases='2410016C14Rik, ARG147, DDX40, PAD', description='DEAH (Asp-Glu-Ala-His) box polypeptide 40', other_designations='DEAH box protein 40', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86768849, end_position_on_the_genomic_accession=86807660, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  67488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67488, CurrentID=0, Status='live', Symbol='Calcoco1', Aliases='1810009B06Rik, Cocoa, Gcap11, mKIAA1536', description='calcium binding and coiled coil domain 1', other_designations='GRIP1-interacting protein|coiled-coil coactivator protein|coiled-coil transcriptional coactivator|granule cell antiserum positive 11', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102706777, end_position_on_the_genomic_accession=102722178, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67489, CurrentID=0, Status='live', Symbol='Ap4b1', Aliases='1810038H16Rik, AV004952, Ap4b4', description='adaptor-related protein complex AP-4, beta 1', other_designations='AP-4 adapter complex subunit beta|AP-4 adaptor complex subunit beta|AP-4 beta-4|adapter-related protein complex 4 subunit beta-1|adaptor-related protein complex 4 subunit beta-1|beta subunit of AP-4|beta4-adaptin', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103809517, end_position_on_the_genomic_accession=103822025, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67490, CurrentID=0, Status='live', Symbol='Ufl1', Aliases='1810074P20Rik, AI429228, Kiaa0776, Maxer, Rcad, mKIAA0776', description='UFM1 specific ligase 1', other_designations='', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=25247820, end_position_on_the_genomic_accession=25281926, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  67492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67492, CurrentID=0, Status='live', Symbol='Zfand4', Aliases='Anubl1', description='zinc finger, AN1-type domain 4', other_designations='AN1, ubiquitin-like, homolog|AN1-type zinc finger and ubiquitin domain-containing protein-like 1', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116264209, end_position_on_the_genomic_accession=116330379, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67493, CurrentID=0, Status='live', Symbol='Mettl16', Aliases='2610100D03Rik, 2810013M15Rik, A830095F14Rik, AI846133, Mett10d', description='methyltransferase like 16', other_designations='methyltransferase 10 domain containing|methyltransferase 10 domain-containing protein|putative methyltransferase METT10D', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74770836, end_position_on_the_genomic_accession=74818694, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67495, CurrentID=0, Status='live', Symbol='Tmem167b', Aliases='2010200O16Rik', description='transmembrane protein 167B', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108556425, end_position_on_the_genomic_accession=108562466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  329641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329641, CurrentID=0, Status='live', Symbol='Sertm1', Aliases='6030405A18Rik', description='serine rich and transmembrane domain containing 1', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54897068, end_position_on_the_genomic_accession=54915887, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67498, CurrentID=0, Status='live', Symbol='Kcnv1', Aliases='2700023A03Rik, vibe', description='potassium channel, subfamily V, member 1', other_designations='neuronal potassium channel alpha subunit|potassium channel Kv8.1 homolog|voltage-gated potassium channel subunit Kv8.1', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=45106237, end_position_on_the_genomic_accession=45116488, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67500, CurrentID=0, Status='live', Symbol='Ccar1', Aliases='2610511G16Rik, 9430036H15Rik, Carp1', description='cell division cycle and apoptosis regulator 1', other_designations='CARP-1|cell cycle and apoptosis regulatory protein 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62743928, end_position_on_the_genomic_accession=62792368, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  67501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67501, CurrentID=0, Status='live', Symbol='Ccdc50', Aliases='C3orf6', description='coiled-coil domain containing 50', other_designations='protein Ymer', map_location='16 18.98 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=27388765, end_position_on_the_genomic_accession=27452218, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67504, CurrentID=0, Status='live', Symbol='Rnf151', Aliases='1700010O16Rik, AV047371', description='ring finger protein 151', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24715839, end_position_on_the_genomic_accession=24718057, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67505, CurrentID=0, Status='live', Symbol='Prl7c1', Aliases='1600017N11Rik, Ghd20, PLP-O, Prlpo, Prplo', description='prolactin family 7, subfamily c, member 1', other_designations='PRL-like protein O|growth hormone d20|placental prolactin-like protein O|prolactin like protein O|prolactin-like protein O', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27773495, end_position_on_the_genomic_accession=27780843, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  329650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329650, CurrentID=0, Status='live', Symbol='Med12l', Aliases='A130035F20, A630079M02, AU044581, TRALPUSH', description='mediator complex subunit 12-like', other_designations='mediator complex subunit 12-like protein|mediator of RNA polymerase II transcription, subunit 12 homolog-like|no opposite paired repeat protein', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=59005595, end_position_on_the_genomic_accession=59318416, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  67507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67507, CurrentID=0, Status='live', Symbol='1700019N19Rik', Aliases='', description='RIKEN cDNA 1700019N19 gene', other_designations='', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58785803, end_position_on_the_genomic_accession=58794414, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67509, CurrentID=0, Status='live', Symbol='Saysd1', Aliases='1810063B07Rik, 4930488P03Rik', description='SAYSVFN motif domain containing 1', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20075636, end_position_on_the_genomic_accession=20083172, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  67510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67510, CurrentID=0, Status='live', Symbol='Tvp23b', Aliases='1810036I24Rik, Fam18b, Fam18b1', description='trans-golgi network vesicle protein 23B', other_designations='family with sequence similarity 18, member B|ortholog of human family with sequence similarity 18 member B FAM18B|protein FAM18B1', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62879437, end_position_on_the_genomic_accession=62895186, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67511, CurrentID=0, Status='live', Symbol='Tmed9', Aliases='2400003B06Rik', description='transmembrane emp24 protein transport domain containing 9', other_designations='glycoprotein 25L2|p24 family protein alpha-2|p24alpha2', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55593135, end_position_on_the_genomic_accession=55597697, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67512, CurrentID=0, Status='live', Symbol='Agpat2', Aliases='2510002J07Rik, AV000834, BSCL, BSCL1, LPAAB, LPAAT-beta', description='1-acylglycerol-3-phosphate O-acyltransferase 2 (lysophosphatidic acid acyltransferase, beta)', other_designations='1-AGP acyltransferase 2|1-AGPAT 2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26593057, end_position_on_the_genomic_accession=26604417, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67513, CurrentID=0, Status='live', Symbol='Faap20', Aliases='2610002J02Rik', description='Fanconi anemia core complex associated protein 20', other_designations='FANCA-associated protein of 20 kDa|Fanconi anemia-associated protein of 20 kDa|likely orthologue of H. sapiens chromosome 1 open reading frame 86 (C1orf86)', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155246759, end_position_on_the_genomic_accession=155256687, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67515, CurrentID=0, Status='live', Symbol='Ttc33', Aliases='2410099M07Rik, 2900001O04Rik, AI507072, AW538342, Osrf', description='tetratricopeptide repeat domain 33', other_designations='TPR repeat protein 33|osmosis responsive factor', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=5185485, end_position_on_the_genomic_accession=5218336, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67516, CurrentID=0, Status='live', Symbol='Kctd4', Aliases='2210017A09Rik, AU017169', description='potassium channel tetramerisation domain containing 4', other_designations='', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75955003, end_position_on_the_genomic_accession=75965217, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67525, CurrentID=0, Status='live', Symbol='Trim62', Aliases='6330414G21Rik, AI450348, Dear1', description='tripartite motif-containing 62', other_designations='tripartite motif-containing protein 62', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128880832, end_position_on_the_genomic_accession=128911332, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67526, CurrentID=0, Status='live', Symbol='Atg12', Aliases='4931423H11Rik, A330058M13Rik, Apg12ll, Atg12', description='autophagy related 12', other_designations='autophagy-related 12|autophagy-related protein 12', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46732417, end_position_on_the_genomic_accession=46741839, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67528, CurrentID=0, Status='live', Symbol='Nudt7', Aliases='1300007B24Rik, 2210404C19Rik', description='nudix (nucleoside diphosphate linked moiety X)-type motif 7', other_designations='coenzyme A diphosphatase|nucleoside diphosphate-linked moiety X motif 7|nudix motif 7', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=114133574, end_position_on_the_genomic_accession=114154739, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67529, CurrentID=0, Status='live', Symbol='Fgfr1op2', Aliases='1500031J01Rik', description='FGFR1 oncogene partner 2', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146577908, end_position_on_the_genomic_accession=146599198, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67530, CurrentID=0, Status='live', Symbol='Uqcrb', Aliases='2210415M14Rik, QP-C, QPC, UQBC, UQBP, UQPC', description='ubiquinol-cytochrome c reductase binding protein', other_designations='complex III subunit 7|complex III subunit VII|ubiquinol-cytochrome c reductase complex 14 kDa protein', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=66900621, end_position_on_the_genomic_accession=66905350, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67532, CurrentID=0, Status='live', Symbol='Mfap1a', Aliases='4432409M24Rik, Mfap1', description='microfibrillar-associated protein 1A', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121492681, end_position_on_the_genomic_accession=121506656, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67533, CurrentID=0, Status='live', Symbol='Ppfibp1', Aliases='4632409B19Rik, AW214454, AW261454', description='PTPRF interacting protein, binding protein 1 (liprin beta 1)', other_designations='PTPRF-interacting protein-binding protein 1|protein tyrosine phosphatase receptor type f polypeptide-interacting protein-binding protein 1', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146888080, end_position_on_the_genomic_accession=147032025, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  67534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67534, CurrentID=0, Status='live', Symbol='Ttll4', Aliases='4632407P03Rik, AI451681, mKIAA0173', description='tubulin tyrosine ligase-like family, member 4', other_designations='polyglutamylase|tubulin--tyrosine ligase-like protein 4', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74661645, end_position_on_the_genomic_accession=74701832, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  329679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329679, CurrentID=0, Status='live', Symbol='Fnip2', Aliases='D630023B12Rik, mKIAA1450', description='folliculin interacting protein 2', other_designations='FNIP1-like protein|O6-methylguanine-induced apoptosis 1 protein', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79455971, end_position_on_the_genomic_accession=79567879, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  67537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67537, CurrentID=0, Status='live', Symbol='Glipr1l2', Aliases='4921508O11Rik', description='GLI pathogenesis-related 1 like 2', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=112083344, end_position_on_the_genomic_accession=112108098, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67538, CurrentID=0, Status='live', Symbol='Zswim3', Aliases='4921517A06Rik, C86566', description='zinc finger SWIM-type containing 3', other_designations='zinc finger, SWIM domain containing 3', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164805114, end_position_on_the_genomic_accession=164822127, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67542, CurrentID=0, Status='live', Symbol='Cog6', Aliases='4933405E16Rik, AU018618, mKIAA1134', description='component of oligomeric golgi complex 6', other_designations='COG complex subunit 6|conserved oligomeric Golgi complex protein 6', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=52982123, end_position_on_the_genomic_accession=53017223, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  67543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67543, CurrentID=0, Status='live', Symbol='Pabpc6', Aliases='4932702K14Rik, AI428050, Pabpc3', description='poly(A) binding protein, cytoplasmic 6', other_designations='poly A binding protein, cytoplasmic 1 pseudogene', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=9666497, end_position_on_the_genomic_accession=9669704, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  67544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67544, CurrentID=0, Status='live', Symbol='Fam120b', Aliases='4932442K08Rik, AW049283, AW413338, CCPG, Pgcc1, mKIAA1838', description='family with sequence similarity 120, member B', other_designations='PPARgamma constitutive coactivator 1|constitutive coactivator of PPAR-gamma|constitutive coactivator of PPARG|family with sequence similarity 120B', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15396157, end_position_on_the_genomic_accession=15433581, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67547, CurrentID=0, Status='live', Symbol='Slc39a8', Aliases='4933419D20Rik, AA986696, BIGM103, ZIP-8, Zip8', description='solute carrier family 39 (metal ion transporter), member 8', other_designations='Zrt- and Irt-like protein 8', map_location='3 63.04 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135816883, end_position_on_the_genomic_accession=135888572, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  329693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329693, CurrentID=0, Status='live', Symbol='Fcrl5', Aliases='Fcrh3, mBXMH2', description='Fc receptor-like 5', other_designations='BXMAS1-like protein 2|Fc receptor-like protein 3|fc receptor homolog 3|fcR-like protein 5|moFcRH3', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87435782, end_position_on_the_genomic_accession=87500678, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  67553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67553, CurrentID=0, Status='live', Symbol='Gstcd', Aliases='4933434L15Rik', description='glutathione S-transferase, C-terminal domain containing', other_designations='', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=132981752, end_position_on_the_genomic_accession=133091971, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67554, CurrentID=0, Status='live', Symbol='Slc25a30', Aliases='4933433D23Rik, AV025504, AV221431, AW319655, KMCP1', description='solute carrier family 25, member 30', other_designations='VH186.2-D-J-C mu', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75757536, end_position_on_the_genomic_accession=75787047, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67555, CurrentID=0, Status='live', Symbol='Mgat4d', Aliases='4933434I20Rik, GnT1IP, GnT1IP-L', description='MGAT4 family, member C', other_designations='GlcNAcT-I inhibitory protein|glycosyltransferase 54 domain-containing protein', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83339402, end_position_on_the_genomic_accession=83382299, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67556, CurrentID=0, Status='live', Symbol='Pigm', Aliases='4933437L05Rik, AI644422, AV302080, C920011G20Rik', description='phosphatidylinositol glycan anchor biosynthesis, class M', other_designations='GPI mannosyltransferase I|GPI-MT-I|PIG-M|phosphatidylinositol glycan, class M|phosphatidylinositol-glycan biosynthesis class M protein', map_location='1 79.73 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172376531, end_position_on_the_genomic_accession=172384099, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67557, CurrentID=0, Status='live', Symbol='Larp6', Aliases='5430431G03Rik, AI552438, Achn', description='La ribonucleoprotein domain family, member 6', other_designations='acheron', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=60713121, end_position_on_the_genomic_accession=60738801, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67561, CurrentID=0, Status='live', Symbol='Wdr48', Aliases='8430408H12Rik, Uaf1, mKIAA1449', description='WD repeat domain 48', other_designations='USP1-associated factor 1', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119894895, end_position_on_the_genomic_accession=119926586, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  67563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67563, CurrentID=0, Status='live', Symbol='Narfl', Aliases='9030612I22Rik, AI504405, PRN', description='nuclear prelamin A recognition factor-like', other_designations='IOP1|iron-only hydrogenase-like protein 1|nuclear prelamin A recognition factor-like protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25773759, end_position_on_the_genomic_accession=25783332, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67564, CurrentID=0, Status='live', Symbol='Tmem35', Aliases='9030603L14Rik, AI841526, PMP52', description='transmembrane protein 35', other_designations='peroxisomal membrane protein 52', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134295225, end_position_on_the_genomic_accession=134305969, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67568, CurrentID=0, Status='live', Symbol='Mrfap1', Aliases='9130413I22Rik, AL022827, Pam14', description='Morf4 family associated protein 1', other_designations='mrfap1|protein associated with MRG of 14 kDa|protein associated with MRG, 14 kDa', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36794867, end_position_on_the_genomic_accession=36796754, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  67569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67569, CurrentID=0, Status='live', Symbol='Mgat4c', Aliases='9130411I17Rik, GntIVh', description='MGAT4 family, member C', other_designations='N-acetylglucosaminyltransferase IVc|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase IVc|UDP-N-acetylglucosamine: alpha-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IVc|gnT-IVc|mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase, isozyme C (putative)', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=101681487, end_position_on_the_genomic_accession=102391851, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67573, CurrentID=0, Status='live', Symbol='Loxl4', Aliases='4833426I20Rik, Loxc', description='lysyl oxidase-like 4', other_designations='lysyl oxidase-like protein 4|lysyl oxidase-related protein C', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42592279, end_position_on_the_genomic_accession=42612906, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  67574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67574, CurrentID=0, Status='live', Symbol='Alg13', Aliases='2810046O15Rik, 4833435D08Rik, Glt28d1, MDS031, OTTMUSG00000018917', description='asparagine-linked glycosylation 13', other_designations='UDP-N-acetylglucosamine transferase subunit ALG13 homolog|asparagine-linked glycosylation 13 homolog|glycosyltransferase 28 domain containing 1|glycosyltransferase 28 domain-containing protein 1', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=144317966, end_position_on_the_genomic_accession=144374450, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  67575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67575, CurrentID=0, Status='live', Symbol='4930430A15Rik', Aliases='', description='RIKEN cDNA 4930430A15 gene', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111101478, end_position_on_the_genomic_accession=111229602, orientation='minus', exon_count=53, OMIM='', no_hdr0=''),
  67578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67578, CurrentID=0, Status='live', Symbol='Patl2', Aliases='4930424G05Rik, Pat1a', description='protein associated with topoisomerase II homolog 2 (yeast)', other_designations='PAT1-like protein 2|protein PAT1 homolog a', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122120103, end_position_on_the_genomic_accession=122186189, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  67579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67579, CurrentID=0, Status='live', Symbol='Cpeb4', Aliases='Cpe-bp4, Cpeb-4, mCPEB-4', description='cytoplasmic polyadenylation element binding protein 4', other_designations='CPE-binding protein 4', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=31870884, end_position_on_the_genomic_accession=31935635, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67580, CurrentID=0, Status='live', Symbol='Lrrc18', Aliases='4930442L21Rik, MTLR1', description='leucine rich repeat containing 18', other_designations='testis-specific LRR protein', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32991382, end_position_on_the_genomic_accession=33015292, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67581, CurrentID=0, Status='live', Symbol='Tbc1d23', Aliases='4930451A13Rik, AU015720, AU043671, AU043778, D030022P07Rik, W51689', description='TBC1 domain family, member 23', other_designations='', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57168858, end_position_on_the_genomic_accession=57231532, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  67582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67582, CurrentID=0, Status='live', Symbol='Slc25a26', Aliases='4930433D19Rik, 4933433F13Rik, AW557176, D6Bwg0781e, Slc25a6', description='solute carrier family 25 (mitochondrial carrier, phosphate carrier), member 26', other_designations='mitochondrial S-adenosylmethionine transporter|solute carrier family 25 member 26', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=94500140, end_position_on_the_genomic_accession=94604655, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  329727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329727, CurrentID=0, Status='live', Symbol='Dennd2c', Aliases='A930010I20Rik', description='DENN/MADD domain containing 2C', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103102654, end_position_on_the_genomic_accession=103169738, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  67586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67586, CurrentID=0, Status='live', Symbol='Ubxn11', Aliases='4930506L07Rik, D4Bwg1540e, Soc, Soci, Ubxd5', description='UBX domain protein 11', other_designations='UBX domain containing 5|UBX domain-containing protein 5|socius', map_location='4 66.5 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134102512, end_position_on_the_genomic_accession=134127621, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  329731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329731, CurrentID=0, Status='live', Symbol='Fam19a3', Aliases='7530404M11Rik, TAFA3, Tafa-3', description='family with sequence similarity 19, member A3', other_designations='chemokine-like protein TAFA-3', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104767405, end_position_on_the_genomic_accession=104777547, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67588, CurrentID=0, Status='live', Symbol='Rnf41', Aliases='2210404G21Rik, 4930511A05Rik, 4933415P08Rik, D10Ertd722e, FLRF, Nrdp1', description='ring finger protein 41', other_designations='', map_location='10 76.55 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128411616, end_position_on_the_genomic_accession=128441441, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67590, CurrentID=0, Status='live', Symbol='Tctn3', Aliases='4930521E07Rik, AI197391, Tect3', description='tectonic family member 3', other_designations='tectonic 3', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40596446, end_position_on_the_genomic_accession=40612281, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67591, CurrentID=0, Status='live', Symbol='Ubl4b', Aliases='4930522D07Rik', description='ubiquitin-like 4B', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107553698, end_position_on_the_genomic_accession=107555073, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  67592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67592, CurrentID=0, Status='live', Symbol='4930524B15Rik', Aliases='', description='RIKEN cDNA 4930524B15 gene', other_designations='', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=31965554, end_position_on_the_genomic_accession=31979655, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67593, CurrentID=0, Status='live', Symbol='4930519G04Rik', Aliases='1700014B12Rik', description='RIKEN cDNA 4930519G04 gene', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114852411, end_position_on_the_genomic_accession=114883880, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  329738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329738, CurrentID=0, Status='live', Symbol='Aknad1', Aliases='4921525H12Rik', description='AKNA domain containing 1', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108739560, end_position_on_the_genomic_accession=108782309, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  329739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329739, CurrentID=0, Status='live', Symbol='Fam102b', Aliases='1600010D10Rik, B430201A12Rik', description='family with sequence similarity 102, member B', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108970997, end_position_on_the_genomic_accession=109027607, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67596, CurrentID=0, Status='live', Symbol='Tespa1', Aliases='5830405N20Rik, A430001F24Rik, AI450549', description='thymocyte expressed, positive selection associated 1', other_designations='thymocyte-expressed positive selection-associated protein 1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130322763, end_position_on_the_genomic_accession=130363469, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67602, CurrentID=0, Status='live', Symbol='Necap1', Aliases='1200016B17Rik, AI747569, mFLJ00061', description='NECAP endocytosis associated 1', other_designations='NECAP endocytosis-associated protein 1|NECAP-1|adaptin-ear-binding coat-associated protein 1|brain enriched NECAP2 similar protein|clathrin-coated vesicle coat protein|snoRNA MBI-129', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122874557, end_position_on_the_genomic_accession=122888941, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67603, CurrentID=0, Status='live', Symbol='Dusp6', Aliases='1300019I03Rik, MKP-3, MKP3, PYST1', description='dual specificity phosphatase 6', other_designations='MAP kinase phosphatase 3|mitogen-activated protein kinase phosphatase 3', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=99263231, end_position_on_the_genomic_accession=99267489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67604, CurrentID=0, Status='live', Symbol='Get4', Aliases='1110007L15Rik, AW412535, Cee', description='golgi to ER traffic protein 4', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139252324, end_position_on_the_genomic_accession=139270050, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67605, CurrentID=0, Status='live', Symbol='Akt1s1', Aliases='1110012J22Rik, AI227026, Lobe, Lobel, PRAS40', description='AKT1 substrate 1 (proline-rich)', other_designations='akt1s1|proline-rich AKT substrate', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44849133, end_position_on_the_genomic_accession=44855421, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67606, CurrentID=0, Status='live', Symbol='Fibin', Aliases='1110018M03Rik', description='fin bud initiation factor homolog (zebrafish)', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=110360925, end_position_on_the_genomic_accession=110362993, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  67607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67607, CurrentID=0, Status='live', Symbol='Zfp788', Aliases='2810426N06Rik, AI256342, AI605193', description='zinc finger protein 788', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=41633531, end_position_on_the_genomic_accession=41651534, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67608, CurrentID=0, Status='live', Symbol='Narf', Aliases='4430402O11Rik', description='nuclear prelamin A recognition factor', other_designations='IOP2|iron-only hydrogenase-like protein 2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121237236, end_position_on_the_genomic_accession=121255856, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67609, CurrentID=0, Status='live', Symbol='4930453N24Rik', Aliases='AI649104', description='RIKEN cDNA 4930453N24 gene', other_designations='', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=64762152, end_position_on_the_genomic_accession=64771156, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67610, CurrentID=0, Status='live', Symbol='Rspry1', Aliases='4930470D19Rik, AI608258', description='ring finger and SPRY domain containing 1', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94601941, end_position_on_the_genomic_accession=94660277, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  67615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67615, CurrentID=0, Status='live', Symbol='Ube2r2', Aliases='1200003M11Rik, Cdc34b, Ubc3b', description='ubiquitin-conjugating enzyme E2R 2', other_designations='E2 ubiquitin-conjugating enzyme R2|E2-CDC34B|ubiquitin carrier protein R2|ubiquitin-conjugating enzyme E2-CDC34B|ubiquitin-protein ligase R2', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41136021, end_position_on_the_genomic_accession=41193370, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67618, CurrentID=0, Status='live', Symbol='Aasdhppt', Aliases='2010309J24Rik, 2810407B07Rik, AASD-PPT, CGI-80, LYS2, LYS5', description='aminoadipate-semialdehyde dehydrogenase-phosphopantetheinyl transferase', other_designations="4'-phosphopantetheinyl transferase|alpha-aminoadipic semialdehyde dehydrogenase-phosphopantetheinyl transferase", map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=4294793, end_position_on_the_genomic_accession=4309509, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67619, CurrentID=0, Status='live', Symbol='Nob1', Aliases='1700021I09Rik, ART-4p, Psmd8bp1, Nob1', description='NIN1/RPN12 binding protein 1 homolog (S. cerevisiae)', other_designations='PSMD8 binding protein 1|nin one binding protein', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107412489, end_position_on_the_genomic_accession=107425038, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67620, CurrentID=0, Status='live', Symbol='Lrp2bp', Aliases='1700113N17Rik, 2310006J04Rik, 4930479L12Rik, MegBP', description='Lrp2 binding protein', other_designations='megalin-binding protein', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46010598, end_position_on_the_genomic_accession=46029478, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67621, CurrentID=0, Status='live', Symbol='Bend5', Aliases='2310026E23Rik', description='BEN domain containing 5', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=111414540, end_position_on_the_genomic_accession=111460298, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67622, CurrentID=0, Status='live', Symbol='Mxra7', Aliases='1810057P16Rik, E130302J09Rik', description='matrix-remodelling associated 7', other_designations='transmembrane anchor protein 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116803400, end_position_on_the_genomic_accession=116828066, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67623, CurrentID=0, Status='live', Symbol='Tm7sf3', Aliases='2010003B14Rik, NARP1', description='transmembrane 7 superfamily member 3', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146602276, end_position_on_the_genomic_accession=146634592, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67628, CurrentID=0, Status='live', Symbol='Anp32b', Aliases='2410015B15Rik, PAL31, PHAPI2a, Ssp29', description='acidic (leucine-rich) nuclear phosphoprotein 32 family, member B', other_designations='acidic nuclear phosphoprotein 32 family, member B|acidic protein rich in leucines|proliferation related acidic leucine rich protein PAL31|proliferation-related acidic leucine-rich protein PAL31', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46451117, end_position_on_the_genomic_accession=46472657, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67629, CurrentID=0, Status='live', Symbol='Spc24', Aliases='2410030K01Rik, AV109292, Spbc24', description='SPC24, NDC80 kinetochore complex component, homolog (S. cerevisiae)', other_designations='spindle pole body component 24 homolog', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21755442, end_position_on_the_genomic_accession=21760452, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67630, CurrentID=0, Status='live', Symbol='Samd8', Aliases='1110053F04Rik, 1700010P07Rik, SMSr', description='sterile alpha motif domain containing 8', other_designations='CPE synthase|GrailEXP prediction|SAM domain-containing protein 8|ceramide phosphoethanolamine synthase|sterile alpha motif domain-containing 8|sterile alpha motif domain-containing protein 8', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21750531, end_position_on_the_genomic_accession=21798725, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  329777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329777, CurrentID=0, Status='live', Symbol='Pigk', Aliases='3000001O05Rik', description='phosphatidylinositol glycan anchor biosynthesis, class K', other_designations='GPI transamidase|PIG-K|phosphatidylinositol-glycan biosynthesis class K protein', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152714100, end_position_on_the_genomic_accession=152789015, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67634, CurrentID=0, Status='live', Symbol='Ftmt', Aliases='4930447C24Rik, FerH, Fth3, MtF', description='ferritin mitochondrial', other_designations='ferritin H|ferritin heavy chain 3', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=52331536, end_position_on_the_genomic_accession=52332996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67636, CurrentID=0, Status='live', Symbol='Lyrm5', Aliases='4930469P12Rik, AL022899, Ghiso', description='LYR motif containing 5', other_designations='growth hormone-inducible soluble protein', map_location='6 77.37 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145211134, end_position_on_the_genomic_accession=145216542, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67645, CurrentID=0, Status='live', Symbol='Armc12', Aliases='4930511I11Rik, AV041658', description='armadillo repeat containing 12', other_designations='testis specific expressed protein 3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28527534, end_position_on_the_genomic_accession=28539110, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67646, CurrentID=0, Status='live', Symbol='4930522H14Rik', Aliases='1700007J23Rik', description='RIKEN cDNA 4930522H14 gene', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109505332, end_position_on_the_genomic_accession=109531297, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67647, CurrentID=0, Status='live', Symbol='4930523C07Rik', Aliases='', description='RIKEN cDNA 4930523C07 gene', other_designations='', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160044368, end_position_on_the_genomic_accession=160078587, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  329795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329795, CurrentID=0, Status='live', Symbol='Tmem67', Aliases='5330408M12Rik, B230117O07, b2b1163.1Clo, b2b1291.1Clo', description='transmembrane protein 67', other_designations='meckel syndrome type 3 protein homolog', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=12039355, end_position_on_the_genomic_accession=12089365, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  67652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67652, CurrentID=0, Status='live', Symbol='Spaca1', Aliases='1700124L11Rik, 4930540L03Rik', description='sperm acrosome associated 1', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34024872, end_position_on_the_genomic_accession=34050067, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67653, CurrentID=0, Status='live', Symbol='4930544G11Rik', Aliases='', description='RIKEN cDNA 4930544G11 gene', other_designations='testis specific expressed protein 5', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=65952571, end_position_on_the_genomic_accession=65954014, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67655, CurrentID=0, Status='live', Symbol='Ctdp1', Aliases='4930563P03Rik, AW553592', description='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) phosphatase, subunit 1', other_designations='CTD (carboxy-terminal domina, RNA polymerase II, polypeptide A) phosphatase, subunit 1|TFIIF-associating CTD phosphatase', map_location='18 53.48 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80407959, end_position_on_the_genomic_accession=80470612, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67657, CurrentID=0, Status='live', Symbol='Rabl3', Aliases='4930553C05Rik, AA958834', description='RAB, member RAS oncogene family-like 3', other_designations='RAB, member of RAS oncogene family-like 3', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37539858, end_position_on_the_genomic_accession=37572385, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67661, CurrentID=0, Status='live', Symbol='Ift172', Aliases='4930553F24Rik, Slb, avc1, wim', description='intraflagellar transport 172', other_designations='intraflagellar transport 172 homolog|intraflagellar transport 172 protein|protein wimple|selective LIM-binding factor Wimple', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31253279, end_position_on_the_genomic_accession=31292160, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  67663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67663, CurrentID=0, Status='live', Symbol='4930549C01Rik', Aliases='', description='RIKEN cDNA 4930549C01 gene', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136606944, end_position_on_the_genomic_accession=136613277, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67664, CurrentID=0, Status='live', Symbol='Rnf125', Aliases='4930553F04Rik', description='ring finger protein 125', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20944625, end_position_on_the_genomic_accession=20987355, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67665, CurrentID=0, Status='live', Symbol='Dctn4', Aliases='1110001K06Rik, 4930547K17Rik, C130039E17Rik, p62', description='dynactin 4', other_designations='dynactin subunit p62', map_location='18 D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60526159, end_position_on_the_genomic_accession=60558762, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67666, CurrentID=0, Status='live', Symbol='Hapln3', Aliases='4930554N11Rik, Lpr3', description='hyaluronan and proteoglycan link protein 3', other_designations='link protein 3', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79117027, end_position_on_the_genomic_accession=79133176, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67667, CurrentID=0, Status='live', Symbol='Alkbh8', Aliases='4930562C03Rik, 8030431D03Rik, 9430088N01Rik, Abh8', description='alkB homolog 8, tRNA methyltransferase', other_designations='S-adenosyl-L-methionine-dependent tRNA methyltransferase ABH8|alkB, alkylation repair homolog 8|probable alpha-ketoglutarate-dependent dioxygenase ABH8|tRNA (carboxymethyluridine(34)-5-O)-methyltransferase ABH8', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=3335231, end_position_on_the_genomic_accession=3387313, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67669, CurrentID=0, Status='live', Symbol='l7Rn6', Aliases='0610007P06Rik, 1110002N09Rik, Hikeshi, l(7)6Rn', description='lethal, Chr 7, Rinchik 6', other_designations='lethal gene on chromosome 7 Rinchik 6 protein', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=89918685, end_position_on_the_genomic_accession=89941221, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67671, CurrentID=0, Status='live', Symbol='Rpl38', Aliases='0610025G13Rik, Rbt, Ts, Tss', description='ribosomal protein L38', other_designations='', map_location='11 G1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114668543, end_position_on_the_genomic_accession=114672331, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67673, CurrentID=0, Status='live', Symbol='Tceb2', Aliases='0610040H15Rik', description='transcription elongation factor B (SIII), polypeptide 2', other_designations='18 kDa, elongin B|RNA polymerase II transcription factor SIII subunit B|SIII p18|eloB|elongin 18 kDa subunit|elongin B|elongin-B|transcription elongation factor B (SIII), polypeptide 2 (18 kDa, elongin B)', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23824740, end_position_on_the_genomic_accession=23829109, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67674, CurrentID=0, Status='live', Symbol='Trmt112', Aliases='0610038D11Rik, Trm112p', description='tRNA methyltransferase 11-2', other_designations='TRM112-like protein|tRNA methyltransferase 11-2 homolog|tRNA methyltransferase 112 homolog', map_location='19 E1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6909698, end_position_on_the_genomic_accession=6911028, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67675, CurrentID=0, Status='live', Symbol='Cuta', Aliases='0610039D01Rik, 1810022E02Rik, 1810060C03Rik, 2700094G22Rik, AI326454', description='cutA divalent cation tolerance homolog', other_designations='brain acetylcholinesterase putative membrane anchor|divalent cation tolerant protein CUTA', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26937972, end_position_on_the_genomic_accession=26939815, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67676, CurrentID=0, Status='live', Symbol='Rpp21', Aliases='0610037N01Rik', description='ribonuclease P 21 subunit', other_designations='RNaseP protein p21|ribonuclease P 21kDa subunit|ribonucleoprotein V', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36255673, end_position_on_the_genomic_accession=36257846, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67678, CurrentID=0, Status='live', Symbol='Lsm3', Aliases='1010001J12Rik, 2610005D18Rik, 6030401D18Rik, SMX4, USS2', description='LSM3 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM3 homolog, U6 small nuclear RNA associated', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91516035, end_position_on_the_genomic_accession=91522620, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67680, CurrentID=0, Status='live', Symbol='Sdhb', Aliases='0710008N11Rik', description='succinate dehydrogenase complex, subunit B, iron sulfur (Ip)', other_designations='ip|iron-sulfur subunit of complex II|succinate dehydrogenase complex subunit B', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140961271, end_position_on_the_genomic_accession=140979197, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67681, CurrentID=0, Status='live', Symbol='Mrpl18', Aliases='1010001C05Rik, C79879, HSPC071, MRP-L18', description='mitochondrial ribosomal protein L18', other_designations='L18mt', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12911345, end_position_on_the_genomic_accession=12916478, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67683, CurrentID=0, Status='live', Symbol='Pbdc1', Aliases='2610029G23Rik', description='polysaccharide biosynthesis domain containing 1', other_designations='UPF0368 protein Cxorf26 homolog', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=105079756, end_position_on_the_genomic_accession=105117094, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67684, CurrentID=0, Status='live', Symbol='Luc7l3', Aliases='3300001P08Rik, Luc7a', description='LUC7-like 3 (S. cerevisiae)', other_designations='cisplatin resistance-associated overexpressed protein|cisplatin resistance-associated-overexpressed protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94291139, end_position_on_the_genomic_accession=94321983, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67685, CurrentID=0, Status='live', Symbol='Dyx1c1', Aliases='1700010I24Rik, EKN1, b2b811.1Clo, b2b811Clo', description='dyslexia susceptibility 1 candidate 1', other_designations='dyslexia susceptibility 1 candidate 1 homolog', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72957365, end_position_on_the_genomic_accession=72973689, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  329831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329831, CurrentID=0, Status='live', Symbol='Fam166b', Aliases='4833436C18Rik', description='family with sequence similarity 166, member B', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43426675, end_position_on_the_genomic_accession=43431593, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67689, CurrentID=0, Status='live', Symbol='Aldh3b1', Aliases='1700001N19Rik, ALDH4, ALDH7', description='aldehyde dehydrogenase 3 family, member B1', other_designations='aldehyde dehydrogenase 7', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3913491, end_position_on_the_genomic_accession=3930817, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67690, CurrentID=0, Status='live', Symbol='Prss37', Aliases='1700016G05Rik, Tryx2', description='protease, serine 37', other_designations='probable inactive trypsin-X2|protease, serine, 37', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40514820, end_position_on_the_genomic_accession=40519542, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67693, CurrentID=0, Status='live', Symbol='Hypk', Aliases='2310003F16Rik, HSPC136', description='huntingtin interacting protein K', other_designations='huntingtin yeast partner K', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121457088, end_position_on_the_genomic_accession=121458440, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67694, CurrentID=0, Status='live', Symbol='Ift74', Aliases='1700029H06Rik, Ccdc2, Cmg1, b2b796Clo', description='intraflagellar transport 74', other_designations='capillary morphogenesis gene 1 protein|capillary morphogenesis protein 1|coiled-coil domain-containing protein 2', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=94603266, end_position_on_the_genomic_accession=94693233, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  67695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67695, CurrentID=0, Status='live', Symbol='Ost4', Aliases='2310016E02Rik', description='oligosaccharyltransferase 4 homolog (S. cerevisiae)', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30906516, end_position_on_the_genomic_accession=30907807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67698, CurrentID=0, Status='live', Symbol='Fam174a', Aliases='2310044D20Rik, Fam174, Naa6, Tmem157', description='family with sequence similarity 174, member A', other_designations='transmembrane protein 157', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=95313628, end_position_on_the_genomic_accession=95335284, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67701, CurrentID=0, Status='live', Symbol='Wfdc2', Aliases='1600023A02Rik, HE4, WAP5', description='WAP four-disulfide core domain 2', other_designations='WAP domain-containing protein HE4', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164562563, end_position_on_the_genomic_accession=164568510, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67702, CurrentID=0, Status='live', Symbol='Rnf149', Aliases='1600023E10Rik, AV037261, C76763, Gm15832, Greul4, OTTMUSG00000026591', description='ring finger protein 149', other_designations='goliath-related E3 ubiquitin-protein ligase 4', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39551294, end_position_on_the_genomic_accession=39577441, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67703, CurrentID=0, Status='live', Symbol='Kirrel3', Aliases='1500010O20Rik, 2900036G11Rik, NEPH2, SST4, mKIAA1867', description='kin of IRRE like 3 (Drosophila)', other_designations='X kin of IRRE like 3|kin of irregular chiasm-like protein 3|membrane protein mKirre', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=34485852, end_position_on_the_genomic_accession=35036315, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  67704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67704, CurrentID=0, Status='live', Symbol='1810037I17Rik', Aliases='5730414I01Rik', description='RIKEN cDNA 1810037I17 gene', other_designations='', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122924397, end_position_on_the_genomic_accession=122926194, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67706, CurrentID=0, Status='live', Symbol='Tmem179b', Aliases='0610031L02Rik, 1500003K22Rik, 1810059G22Rik', description='transmembrane protein 179B', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8772522, end_position_on_the_genomic_accession=8774467, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67707, CurrentID=0, Status='live', Symbol='Mrpl24', Aliases='2010005E08Rik, 2810470K06Rik, 6720473G22Rik, AA407670', description='mitochondrial ribosomal protein L24', other_designations='L24mt|MRP-L24', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87919074, end_position_on_the_genomic_accession=87923434, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67708, CurrentID=0, Status='live', Symbol='Pcnxl4', Aliases='1810048J11Rik, 4932413O10, AU017192', description='pecanex-like 4 (Drosophila)', other_designations='pecanex-like protein C14orf135 homolog', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72536146, end_position_on_the_genomic_accession=72580213, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67709, CurrentID=0, Status='live', Symbol='Reg4', Aliases='2010002L15Rik, GISP, RELP', description='regenerating islet-derived family, member 4', other_designations='REG-4', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98222156, end_position_on_the_genomic_accession=98236748, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67710, CurrentID=0, Status='live', Symbol='Polr2g', Aliases='2410046K11Rik, A230108L04Rik, C76415, RBP7, Rpo2-7l', description='polymerase (RNA) II (DNA directed) polypeptide G', other_designations='DNA-directed RNA polymerase II subunit G|RNA polymerase II subunit B7', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8793129, end_position_on_the_genomic_accession=8798557, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67711, CurrentID=0, Status='live', Symbol='Nsmce1', Aliases='2510027N19Rik', description='non-SMC element 1 homolog (S. cerevisiae)', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125467640, end_position_on_the_genomic_accession=125491576, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67712, CurrentID=0, Status='live', Symbol='Slc25a37', Aliases='1700020E22Rik, 4930513O14Rik, 4930526G11Rik, AI848481, C330015G08Rik, Mfrn, Mfrn1, Mscp, frascati, mitoferrin', description='solute carrier family 25, member 37', other_designations='mitochondrial iron transporter 1|mitochondrial solute carrier protein|mitoferrin 1', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69241848, end_position_on_the_genomic_accession=69285103, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67713, CurrentID=0, Status='live', Symbol='Dnajc19', Aliases='1810055D05Rik, AA959924, Tim14', description='DnaJ heat shock protein family (Hsp40) member C19', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 19', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=34057280, end_position_on_the_genomic_accession=34081354, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67715, CurrentID=0, Status='live', Symbol='2010106E10Rik', Aliases='', description='RIKEN cDNA 2010106E10 gene', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=112495274, end_position_on_the_genomic_accession=112558343, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67717, CurrentID=0, Status='live', Symbol='Lipf', Aliases='2310051B21Rik, AV082900', description='lipase, gastric', other_designations='GL|gastric lipase', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=33919441, end_position_on_the_genomic_accession=33976813, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67718, CurrentID=0, Status='live', Symbol='Lce1h', Aliases='2310066F03Rik, AI426284, Sprrl9', description='late cornified envelope 1H', other_designations='small proline rich-like 9', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92763215, end_position_on_the_genomic_accession=92765065, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  67719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67719, CurrentID=0, Status='live', Symbol='2310057J18Rik', Aliases='2210414P04Rik, Leg1', description='RIKEN cDNA 2310057J18 gene', other_designations='UPF0762 protein C6orf58 homolog', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=28972288, end_position_on_the_genomic_accession=28986306, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67722, CurrentID=0, Status='live', Symbol='Actl11', Aliases='4921517D21Rik', description='actin-like 11', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107928469, end_position_on_the_genomic_accession=107932461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67723, CurrentID=0, Status='live', Symbol='Cep83os', Aliases='4932415G12Rik, Ccdc41os1', description='centrosomal protein 83, opposite strand', other_designations='Ccdc41 opposite strand 1|coiled-coil domain containing 41, opposite strand', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=94673493, end_position_on_the_genomic_accession=94688613, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67724, CurrentID=0, Status='live', Symbol='Pop1', Aliases='4932434G09Rik', description='processing of precursor 1, ribonuclease P/MRP family, (S. cerevisiae)', other_designations='processing of precursors 1|ribonucleases P/MRP protein subunit homolog', map_location='15 B3.3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34495311, end_position_on_the_genomic_accession=34530653, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  67725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67725, CurrentID=0, Status='live', Symbol='Nudt13', Aliases='4933433B15Rik', description='nudix (nucleoside diphosphate linked moiety X)-type motif 13', other_designations='nudix motif 13', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20294679, end_position_on_the_genomic_accession=20317584, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67726, CurrentID=0, Status='live', Symbol='Fam114a2', Aliases='1810073G14Rik, 9030624B09Rik', description='family with sequence similarity 114, member A2', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=57482990, end_position_on_the_genomic_accession=57518656, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67727, CurrentID=0, Status='live', Symbol='Stx17', Aliases='4833418L03Rik, 6330411F21Rik, 9030425C21Rik, AW048351', description='syntaxin 17', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48124915, end_position_on_the_genomic_accession=48186507, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  329872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329872, CurrentID=0, Status='live', Symbol='Frem1', Aliases='BC037594, D430009N09, D630008K06, crf11, eye<m02Jus>, eyes2, heb, qbrick', description='Fras1 related extracellular matrix protein 1', other_designations='', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=82897920, end_position_on_the_genomic_accession=83052548, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  67729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67729, CurrentID=0, Status='live', Symbol='Mansc1', Aliases='9130403P13Rik', description='MANSC domain containing 1', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134609207, end_position_on_the_genomic_accession=134632488, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67731, CurrentID=0, Status='live', Symbol='Fbxo32', Aliases='4833442G10Rik, AI430017, ATROGIN1, Gm20361, MAFbx', description='F-box protein 32', other_designations='atrogin 1|muscle atrophy F-box protein', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58175879, end_position_on_the_genomic_accession=58214892, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67732, CurrentID=0, Status='live', Symbol='Iah1', Aliases='1500019E20Rik, 4833421E05Rik', description='isoamyl acetate-hydrolyzing esterase 1 homolog', other_designations='', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=21316360, end_position_on_the_genomic_accession=21323607, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  329877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329877, CurrentID=0, Status='live', Symbol='Dennd4c', Aliases='1700065A05Rik, AA420392', description='DENN/MADD domain containing 4C', other_designations='likely ortholog of H. sapiens chromosome 9 open reading frame 55 (C9orf55)', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86748555, end_position_on_the_genomic_accession=86850603, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  67736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67736, CurrentID=0, Status='live', Symbol='Ccdc130', Aliases='4930527D15Rik, C77430', description='coiled-coil domain containing 130', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84257767, end_position_on_the_genomic_accession=84270458, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67737, CurrentID=0, Status='live', Symbol='Ttc39d', Aliases='4930560E09Rik', description='tetratricopeptide repeat domain 39D', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80215914, end_position_on_the_genomic_accession=80217936, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67738, CurrentID=0, Status='live', Symbol='Ppid', Aliases='4930564J03Rik, CYP-40l, Ppif, Ppid', description='peptidylprolyl isomerase D (cyclophilin D)', other_designations='40 kDa peptidyl-prolyl cis-trans isomerase|PPIase D|cyclophilin 40|cyclophilin-40|cytoplasmic cyclophilin D|rotamase D', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79591389, end_position_on_the_genomic_accession=79603650, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67739, CurrentID=0, Status='live', Symbol='Slc48a1', Aliases='4930570C03Rik', description='solute carrier family 48 (heme transporter), member 1', other_designations='heme-responsive gene 1 protein homolog|solute carrier family 48 member 1', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97784365, end_position_on_the_genomic_accession=97792692, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67741, CurrentID=0, Status='live', Symbol='4930579F01Rik', Aliases='', description='RIKEN cDNA 4930579F01 gene', other_designations='', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138164078, end_position_on_the_genomic_accession=138194242, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67742, CurrentID=0, Status='live', Symbol='Samsn1', Aliases='4930571B16Rik, 930571B16Rik, Hacs1, SLy2', description='SAM domain, SH3 domain and nuclear localization signals, 1', other_designations='SAM domain, SH3 domain and nuclear localisation signals, 1|SAM domain, SH3 domain and nuclear localization signals protein 1|SH3 protein expressed in lymphocytes 2|SH3-lymphocyte protein 2', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=75858794, end_position_on_the_genomic_accession=75909266, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67747, CurrentID=0, Status='live', Symbol='Ribc2', Aliases='4930579A10Rik, Trib', description='RIB43A domain with coiled-coils 2', other_designations='testis-specific ribbon protein', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85132060, end_position_on_the_genomic_accession=85144569, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67749, CurrentID=0, Status='live', Symbol='Mgarp', Aliases='4930583H14Rik, AI195347, CESP-1, HUMMR, Osap, Qsap', description='mitochondria localized glutamic acid rich protein', other_designations='corneal endothelium-specific protein 1|hypoxia up-regulated mitochondrial movement regulator protein|mitochondria-localized glutamic acid-rich protein|ovary-specific acidic protein', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51388412, end_position_on_the_genomic_accession=51396547, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67750, CurrentID=0, Status='live', Symbol='4930578I06Rik', Aliases='AI429111', description='RIKEN cDNA 4930578I06 gene', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63971121, end_position_on_the_genomic_accession=63987780, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67752, CurrentID=0, Status='live', Symbol='Ppp1r32', Aliases='4930579J09Rik, AU015816, IIIG9, IIIG9L, IIIG9S', description='protein phosphatase 1, regulatory subunit 32', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10474235, end_position_on_the_genomic_accession=10491452, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  67753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67753, CurrentID=0, Status='live', Symbol='Eqtn', Aliases='1700028B15Rik, 4930579C15Rik, Afaf, C81058, Eqtl, Eqts', description='equatorin, sperm acrosome associated', other_designations='MN9 antigen|acrosome formation associated factor|sperm-specific single transmembrane domain protein', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=94875696, end_position_on_the_genomic_accession=94929950, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67755, CurrentID=0, Status='live', Symbol='Ddx47', Aliases='4930588A18Rik, C77285', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 47', other_designations='DEAD box protein 47', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135011612, end_position_on_the_genomic_accession=135023776, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67758, CurrentID=0, Status='live', Symbol='Aadac', Aliases='5033417E09Rik, AI265437, Aada', description='arylacetamide deacetylase (esterase)', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=60031788, end_position_on_the_genomic_accession=60040157, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67759, CurrentID=0, Status='live', Symbol='Plgrkt', Aliases='1110007H22Rik, 5033414D02Rik, AI852040, Plg-R(KT), Plg-RKT', description='plasminogen receptor, C-terminal lysine transmembrane protein', other_designations='transmembrane protein C9orf46 homolog', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29348599, end_position_on_the_genomic_accession=29361885, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67760, CurrentID=0, Status='live', Symbol='Slc38a2', Aliases='5033402L14Rik, AI316867, ATA2, SAT2, mKIAA1382', description='solute carrier family 38, member 2', other_designations='amino acid system A transporter|amino acid transporter A2|system A amino acid transporter 2|system A transporter 1|system N amino acid transporter 2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=96687386, end_position_on_the_genomic_accession=96699724, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  67763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67763, CurrentID=0, Status='live', Symbol='Prpsap1', Aliases='5730409F23Rik, PAP39', description='phosphoribosyl pyrophosphate synthetase-associated protein 1', other_designations='39 kDa phosphoribosypyrophosphate synthase-associated protein|39 kDa phosphoribosypyrophosphate synthetase-associated protein|PRPP synthase-associated protein 1|PRPP synthetase-associated protein 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116470816, end_position_on_the_genomic_accession=116494652, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  329908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329908, CurrentID=0, Status='live', Symbol='Usp24', Aliases='2700066K03Rik, 2810030C21Rik, AI414051, B130021E18, C79851', description='ubiquitin specific peptidase 24', other_designations='deubiquitinating enzyme 24|ubiquitin specific protease 24|ubiquitin thioesterase 24|ubiquitin thiolesterase 24|ubiquitin-specific-processing protease 24', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106315899, end_position_on_the_genomic_accession=106446572, orientation='plus', exon_count=70, OMIM='', no_hdr0=''),
  329909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329909, CurrentID=0, Status='live', Symbol='Tmem61', Aliases='4732441L20Rik', description='transmembrane protein 61', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106498813, end_position_on_the_genomic_accession=106506335, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  329910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329910, CurrentID=0, Status='live', Symbol='Acot11', Aliases='1110020M10Rik, 2010309H15Rik, AW060409, BFIT, BFIT1, Thea, Them1, mKIAA0707', description='acyl-CoA thioesterase 11', other_designations='acyl-CoA thioester hydrolase 11|adipose-associated thioesterase|brown fat-inducible thioesterase|thioesterase, adipose associated', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106744555, end_position_on_the_genomic_accession=106807180, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  67767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67767, CurrentID=0, Status='live', Symbol='Jagn1', Aliases='5830427H10Rik', description='jagunal homolog 1', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113442517, end_position_on_the_genomic_accession=113448229, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67768, CurrentID=0, Status='live', Symbol='N6amt1', Aliases='5830445C04Rik, Hemk2, Pred28', description='N-6 adenine-specific DNA methyltransferase 1 (putative)', other_designations='M-N6DNA MTase A|N6-DNA methyltransferase A', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87354185, end_position_on_the_genomic_accession=87368649, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67769, CurrentID=0, Status='live', Symbol='Gpatch2', Aliases='5830433G22Rik, 5830436K05Rik, AI427714, AI447508, AW107440, AW491060, Gpatc2', description='G patch domain containing 2', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=187215393, end_position_on_the_genomic_accession=187369480, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  67770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67770, CurrentID=0, Status='live', Symbol='Caap1', Aliases='5830433M19Rik, AI536248, CAAP', description='caspase activity and apoptosis inhibitor 1', other_designations='conserved anti-apoptotic protein|likely ortholog of H. sapiens chromosome 9 open reading frame 82 (C9orf82)', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=94460277, end_position_on_the_genomic_accession=94557076, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67771, CurrentID=0, Status='live', Symbol='Arpc5', Aliases='5830443F10Rik, p16-Arc', description='actin related protein 2/3 complex, subunit 5', other_designations='arp2/3 complex 16 kDa subunit', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152766542, end_position_on_the_genomic_accession=152775580, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67772, CurrentID=0, Status='live', Symbol='Chd8', Aliases='5830451P18Rik, AU015341, Duplin, HELSNF1, mKIAA1564', description='chromodomain helicase DNA binding protein 8', other_designations='ATP-dependent helicase CHD8|CHD-8|axis duplication inhibitor|helicase with SNF2 domain 1', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52198149, end_position_on_the_genomic_accession=52257820, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  67773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67773, CurrentID=0, Status='live', Symbol='Kat8', Aliases='2010203C02Rik, 5830450F21Rik, D7Ertd629e, MOF, Myst1', description='K(lysine) acetyltransferase 8', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 1|MYST histone acetyltransferase 1|MYST-1|histone acetyltransferase Myst1|lysine acetyltransferase 8|probable histone acetyltransferase MYST1', map_location='7 69.83 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127912517, end_position_on_the_genomic_accession=127925763, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67774, CurrentID=0, Status='live', Symbol='Borcs5', Aliases='5830457J20Rik, AW124643, Loh12cr1', description='BLOC-1 related complex subunit 5', other_designations='loss of heterozygosity, 12, chromosomal region 1 homolog', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134639512, end_position_on_the_genomic_accession=134711184, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  329919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329919, CurrentID=0, Status='live', Symbol='Skint2', Aliases='A430070M19, B7S3, OTTMUSG00000008540, Skint-2', description='selection and upkeep of intraepithelial T cells 2', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=112613609, end_position_on_the_genomic_accession=112652248, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67776, CurrentID=0, Status='live', Symbol='Vwa5a', Aliases='5830475I06Rik, AW552491, BCSC-1, E130119J22, Loh11cr2a', description='von Willebrand factor A domain containing 5A', other_designations='loss of heterozygosity, 11, chromosomal region 2, gene A homolog', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38718268, end_position_on_the_genomic_accession=38743337, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  67778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67778, CurrentID=0, Status='live', Symbol='Zfp639', Aliases='6230400O18Rik, ANC-2H01, ZASC1, Znf639', description='zinc finger protein 639', other_designations='zinc finger protein ZASC1', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32510550, end_position_on_the_genomic_accession=32520833, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67781, CurrentID=0, Status='live', Symbol='Ilf2', Aliases='6230405A16Rik, TEG-267, Tex261, Tex267', description='interleukin enhancer binding factor 2', other_designations='nuclear factor of activated T-cells 45 kDa|testis expressed gene 267', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90476201, end_position_on_the_genomic_accession=90488379, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67784, CurrentID=0, Status='live', Symbol='Plxnd1', Aliases='6230425C21Rik, b2b1863Clo, b2b553Clo', description='plexin D1', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115954811, end_position_on_the_genomic_accession=115995005, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  67785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67785, CurrentID=0, Status='live', Symbol='Zmym4', Aliases='6330503C17Rik, AI480785, AW493829, CDIR, D630001M21, MYM, Zfp262, Znf262, mKIAA0425', description='zinc finger, MYM-type 4', other_designations='cell death inhibiting RNA|zinc finger protein 262', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126861819, end_position_on_the_genomic_accession=126969370, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  67788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67788, CurrentID=0, Status='live', Symbol='Sfr1', Aliases='6330577E15Rik, Meir5', description='SWI5 dependent recombination repair 1', other_designations='DNA repair protein Sfr1|meiosis protein 5 homolog', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47731733, end_position_on_the_genomic_accession=47735588, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67789, CurrentID=0, Status='live', Symbol='Dalrd3', Aliases='6330580J24Rik, C77829, mir-425', description='DALR anticodon binding domain containing 3', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108567496, end_position_on_the_genomic_accession=108572773, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  329934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329934, CurrentID=0, Status='live', Symbol='Foxo6', Aliases='', description='forkhead box O6', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120267078, end_position_on_the_genomic_accession=120287261, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67792, CurrentID=0, Status='live', Symbol='Rgs8', Aliases='6530413N01Rik', description='regulator of G-protein signaling 8', other_designations='', map_location='1 65.41 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153652953, end_position_on_the_genomic_accession=153697668, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67795, CurrentID=0, Status='live', Symbol='Rnls', Aliases='6530404N21Rik, AI452315, AW060440, C10orf59', description='renalase, FAD-dependent amine oxidase', other_designations='MAO-C|alpha-NAD(P)H oxidase/anomerase|mMAO-C|monoamine oxidase-C|rnls', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=33137744, end_position_on_the_genomic_accession=33392295, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  329941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329941, CurrentID=0, Status='live', Symbol='Col8a2', Aliases='AI429819', description='collagen, type VIII, alpha 2', other_designations='endothelial collagen|procollagen type VIII alpha 2|procollagen, type VIII, alpha 2', map_location='4 60.29 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126286756, end_position_on_the_genomic_accession=126314330, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  329942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329942, CurrentID=0, Status='live', Symbol='Csmd2', Aliases='Gm139', description='CUB and Sushi multiple domains 2', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127988016, end_position_on_the_genomic_accession=128567666, orientation='plus', exon_count=71, OMIM='', no_hdr0=''),
  67800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67800, CurrentID=0, Status='live', Symbol='Dgat2', Aliases='0610010B06Rik, ARAT, DGAT-2', description='diacylglycerol O-acyltransferase 2', other_designations='acyl-CoA retinol O-fatty-acyltransferase|diacylglycerol acyltransferase 2|diglyceride acyltransferase 2|retinol O-fatty-acyltransferase', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99153663, end_position_on_the_genomic_accession=99182713, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67801, CurrentID=0, Status='live', Symbol='Pllp', Aliases='0610010I06Rik, AV001002, Plapi, Tm4sf11', description='plasma membrane proteolipid', other_designations='transmembrane 4 superfamily member 11', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94674895, end_position_on_the_genomic_accession=94696242, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67803, CurrentID=0, Status='live', Symbol='Limd2', Aliases='0610025L06Rik, AI413966', description='LIM domain containing 2', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106156256, end_position_on_the_genomic_accession=106161039, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67804, CurrentID=0, Status='live', Symbol='Snx2', Aliases='0610030A03Rik', description='sorting nexin 2', other_designations='', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=53176287, end_position_on_the_genomic_accession=53220860, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67808, CurrentID=0, Status='live', Symbol='Tprgl', Aliases='1200015A19Rik, Svap30, Tprg1l, mover', description='transformation related protein 63 regulated like', other_designations='mossy fiber terminal-associated vertebrate-specific presynaptic protein|mossy-fiber terminal-associated vertebrate-specific presynaptic protein|synaptic vesicle associated protein of 30 kD', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154157485, end_position_on_the_genomic_accession=154160712, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67809, CurrentID=0, Status='live', Symbol='Rmdn3', Aliases='1200015F23Rik, AI131757, Fam82a2, Ptpip51, RMD-3, Rmd3', description='regulator of microtubule dynamics 3', other_designations='family with sequence similarity 82, member A2|microtubule-associated protein|protein tyrosine phosphatase-interacting protein 51|tyrosine phosphatase interacting protein 51', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119136998, end_position_on_the_genomic_accession=119157034, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  329954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329954, CurrentID=0, Status='live', Symbol='Catsper4', Aliases='4933412L05', description='cation channel, sperm associated 4', other_designations='cation channel of sperm 4|channel, sperm associated 4', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134211968, end_position_on_the_genomic_accession=134227434, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  67811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67811, CurrentID=0, Status='live', Symbol='Poldip2', Aliases='1300003F06Rik, Pdip38', description='polymerase (DNA-directed), delta interacting protein 2', other_designations='mitogenin 1|mitogenin I|polymerase delta interacting protein 38', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78512296, end_position_on_the_genomic_accession=78522736, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67812, CurrentID=0, Status='live', Symbol='Ubxn4', Aliases='1300013G12Rik, Ubxd2, mKIAA0242', description='UBX domain protein 4', other_designations='UBX domain containing 2|UBX domain-containing protein 2|erasin', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=128243898, end_position_on_the_genomic_accession=128279378, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67815, CurrentID=0, Status='live', Symbol='Sec14l2', Aliases='1300013M05Rik, Spf, TAP', description='SEC14-like lipid binding 2', other_designations='SEC14-like 2|alpha-tocopherol-associated protein', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4097039, end_position_on_the_genomic_accession=4118729, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  111371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111371, CurrentID=0, Status='live', Symbol='Emv1', Aliases='C3v, Cv-1, Emv-1, Inc-1, Rjv-2', description='endogenous ecotropic MuLV 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  67819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67819, CurrentID=0, Status='live', Symbol='Derl1', Aliases='1110021N07Rik, AI195141, AW551338, Derlin-1', description='Der1-like domain family, member 1', other_designations='degradation in endoplasmic reticulum protein 1|der1-like protein 1', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=57869502, end_position_on_the_genomic_accession=57892418, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67821, CurrentID=0, Status='live', Symbol='Atp1b4', Aliases='1110020B02Rik', description='ATPase, (Na+)/K+ transporting, beta 4 polypeptide', other_designations='x,K-ATPase subunit beta-m|x/potassium-transporting ATPase subunit beta-m', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38316122, end_position_on_the_genomic_accession=38336784, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67824, CurrentID=0, Status='live', Symbol='Nmral1', Aliases='1110025F24Rik, AI256624', description='NmrA-like family domain containing 1', other_designations='', map_location='16 2.46 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4711317, end_position_on_the_genomic_accession=4719356, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67826, CurrentID=0, Status='live', Symbol='Snap47', Aliases='1110031B06Rik, SNAP-47', description='synaptosomal-associated protein, 47', other_designations='synaptosomal-associated 47 kDa protein', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59407141, end_position_on_the_genomic_accession=59450029, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  329972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329972, CurrentID=0, Status='live', Symbol='Spata21', Aliases='4933414G08', description='spermatogenesis associated 21', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141088322, end_position_on_the_genomic_accession=141112760, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  67830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67830, CurrentID=0, Status='live', Symbol='Rer1', Aliases='1110060F11Rik, 5830454N22Rik, AU043380', description='RER1 retention in endoplasmic reticulum 1 homolog (S. cerevisiae)', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155074109, end_position_on_the_genomic_accession=155086382, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67832, CurrentID=0, Status='live', Symbol='Brix1', Aliases='1110064N10Rik, Bxdc2, C76935', description='BRX1, biogenesis of ribosomes', other_designations='BRX1, biogenesis of ribosomes, homolog|brix domain containing 2|brix domain-containing protein 2|ribosome biogenesis protein Brix', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10474779, end_position_on_the_genomic_accession=10485937, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  329977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329977, CurrentID=0, Status='live', Symbol='Fhad1', Aliases='2900090M10Rik, B230311F11', description='forkhead-associated (FHA) phosphopeptide binding domain 1', other_designations='FHA domain-containing protein 1', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141890430, end_position_on_the_genomic_accession=142015124, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  67834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67834, CurrentID=0, Status='live', Symbol='Idh3a', Aliases='1110003P10Rik, 1500012E04Rik, AA407078, AI316514', description='isocitrate dehydrogenase 3 (NAD+) alpha', other_designations='NAD(+)-specific ICDH subunit alpha|isocitric dehydrogenase subunit alpha', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54586511, end_position_on_the_genomic_accession=54604662, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67836, CurrentID=0, Status='live', Symbol='Wdr83', Aliases='1500041N16Rik, Morg1', description='WD repeat domain containing 83', other_designations='MAPK cascade scaffold protein|MAPK organizer 1|mitogen-activated protein kinase organizer 1', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85075035, end_position_on_the_genomic_accession=85080746, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67838, CurrentID=0, Status='live', Symbol='Dnajb11', Aliases='1810031F23Rik, ABBP-2, AL024055, Dj9, ERdj3, ERj3p', description='DnaJ heat shock protein family (Hsp40) member B11', other_designations='APOBEC1-binding protein 2|DnaJ (Hsp40) homolog, subfamily B, member 11|ER-associated Hsp40 co-chaperone|ER-associated dnaJ protein 3|ER-resident protein ERdj3|endoplasmic reticulum DNA J domain-containing protein 3', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22857845, end_position_on_the_genomic_accession=22879634, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67839, CurrentID=0, Status='live', Symbol='Gpsm1', Aliases='1810037C22Rik, AW107933, Ags3', description='G-protein signalling modulator 1 (AGS3-like, C. elegans)', other_designations='activator of G-protein signaling 3', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26314606, end_position_on_the_genomic_accession=26348237, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  67840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67840, CurrentID=0, Status='live', Symbol='Mrpl57', Aliases='2610025P17Rik, AV005697, Mrp63', description='mitochondrial ribosomal protein L53', other_designations='mMRP63|mitochondrial ribosomal protein 63|mitochondrial ribosomal protein L57', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57826228, end_position_on_the_genomic_accession=57829087, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67841, CurrentID=0, Status='live', Symbol='Atg3', Aliases='2610016C12Rik, APG3, Apg3ll, PC3-96, Atg3', description='autophagy related 3', other_designations='autophagy-related 3|autophagy-related protein 3', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45158829, end_position_on_the_genomic_accession=45188538, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67842, CurrentID=0, Status='live', Symbol='Nop9', Aliases='2610027L16Rik', description='NOP9 nucleolar protein', other_designations='pumilio domain-containing protein C14orf21 homolog|pumilio domain-containing protein NOP9', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55745693, end_position_on_the_genomic_accession=55755634, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67843, CurrentID=0, Status='live', Symbol='Slc35a4', Aliases='2610030J16Rik', description='solute carrier family 35, member A4', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36679215, end_position_on_the_genomic_accession=36683862, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67844, CurrentID=0, Status='live', Symbol='Rab32', Aliases='2810011A17Rik, AU022057', description='RAB32, member RAS oncogene family', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=10545039, end_position_on_the_genomic_accession=10558207, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67845, CurrentID=0, Status='live', Symbol='Rnf115', Aliases='2610028E05Rik, AU042696, Zfp364', description='ring finger protein 115', other_designations='Rab7-interacting RING finger protein|Rabring7|rabring 7|zinc finger protein 364', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96727591, end_position_on_the_genomic_accession=96791155, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67846, CurrentID=0, Status='live', Symbol='Tmem39a', Aliases='2610033C09Rik', description='transmembrane protein 39a', other_designations='', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38558698, end_position_on_the_genomic_accession=38592162, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67847, CurrentID=0, Status='live', Symbol='Sncaip', Aliases='2810407O15Rik, 4933427B05Rik, BB104594, SYPH1', description='synuclein, alpha interacting protein (synphilin)', other_designations='alpha-synuclein-interacting protein', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=52767811, end_position_on_the_genomic_accession=52915935, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67848, CurrentID=0, Status='live', Symbol='Ddx55', Aliases='2810021H22Rik, mKIAA1595', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 55', other_designations='DEAD box protein 55', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124552544, end_position_on_the_genomic_accession=124569660, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67849, CurrentID=0, Status='live', Symbol='Cdca5', Aliases='2610036L13Rik, AL024086, AW536684, C85404', description='cell division cycle associated 5', other_designations='cell division cycle-associated protein 5|sororin p35', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6085097, end_position_on_the_genomic_accession=6091773, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67851, CurrentID=0, Status='live', Symbol='1700021F05Rik', Aliases='', description='RIKEN cDNA 1700021F05 gene', other_designations='', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43525121, end_position_on_the_genomic_accession=43541041, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67854, CurrentID=0, Status='live', Symbol='Slco6b1', Aliases='1700022M03Rik', description='solute carrier organic anion transporter family, member 6b1', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=96906176, end_position_on_the_genomic_accession=96997560, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67855, CurrentID=0, Status='live', Symbol='Asprv1', Aliases='2300003P22Rik, AA986851, SASP, SASPase, Taps', description='aspartic peptidase, retroviral-like 1', other_designations='TPA-inducible aspartic proteinase-like protein|asprv1|skin aspartic protease|skin-specific retroviral-like aspartic protease', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86628174, end_position_on_the_genomic_accession=86629710, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  67856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67856, CurrentID=0, Status='live', Symbol='Echdc3', Aliases='2310005D12Rik, AI662097', description='enoyl Coenzyme A hydratase domain containing 3', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=6188465, end_position_on_the_genomic_accession=6212994, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67857, CurrentID=0, Status='live', Symbol='Ppp6c', Aliases='2310003C10Rik', description='protein phosphatase 6, catalytic subunit', other_designations='PP6C', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=39194369, end_position_on_the_genomic_accession=39226338, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67859, CurrentID=0, Status='live', Symbol='Cysrt1', Aliases='2310002J15Rik, AV023762', description='cysteine rich tail 1', other_designations='UPF0574 protein C9orf169 homolog', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25238819, end_position_on_the_genomic_accession=25239897, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  67860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67860, CurrentID=0, Status='live', Symbol='S100a16', Aliases='2300002L21Rik, AI325039, AI663996, DT1P1A7, S100F', description='S100 calcium binding protein A16', other_designations='', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90537257, end_position_on_the_genomic_accession=90543151, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67861, CurrentID=0, Status='live', Symbol='Akr1b10', Aliases='2310005E10Rik, Akr1b16', description='aldo-keto reductase family 1, member B10 (aldose reductase)', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34384206, end_position_on_the_genomic_accession=34396954, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67862, CurrentID=0, Status='live', Symbol='2310033P09Rik', Aliases='Mmtag2', description='RIKEN cDNA 2310033P09 gene', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59208361, end_position_on_the_genomic_accession=59210736, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67863, CurrentID=0, Status='live', Symbol='Slc25a11', Aliases='2310022P18Rik, 2oxoc', description='solute carrier family 25 (mitochondrial carrier oxoglutarate carrier), member 11', other_designations='OGCP|solute carrier family 25 member 11', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70644019, end_position_on_the_genomic_accession=70647423, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67864, CurrentID=0, Status='live', Symbol='Yipf4', Aliases='2310034L04Rik', description='Yip1 domain family, member 4', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74489493, end_position_on_the_genomic_accession=74500277, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67865, CurrentID=0, Status='live', Symbol='Rgs10', Aliases='2310010N19Rik', description='regulator of G-protein signalling 10', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128373621, end_position_on_the_genomic_accession=128419061, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67866, CurrentID=0, Status='live', Symbol='Wfdc1', Aliases='2310058A03Rik, ps20', description='WAP four-disulfide core domain 1', other_designations='prostate stromal protein ps20|ps20 growth inhibitor', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119666365, end_position_on_the_genomic_accession=119688226, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67867, CurrentID=0, Status='live', Symbol='Lrrc28', Aliases='1300004K21Rik, 2210012C09Rik, 2310058O11Rik', description='leucine rich repeat containing 28', other_designations='', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=67513410, end_position_on_the_genomic_accession=67645331, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  67868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67868, CurrentID=0, Status='live', Symbol='Cela3b', Aliases='0910001F22Rik, 2310074F01Rik, AI504000, Ela3, Ela3b', description='chymotrypsin-like elastase family, member 3B', other_designations='elastase 3, pancreatic|elastase 3B, pancreatic|elastase IIIB|elastase-3B|protease E', map_location='4 69.9 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137421008, end_position_on_the_genomic_accession=137430520, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67869, CurrentID=0, Status='live', Symbol='Paip2', Aliases='2310050K10Rik, AU045972', description='polyadenylate-binding protein-interacting protein 2', other_designations='PAIP-2|Pabp-interacting protein 2|poly(A)-binding protein-interacting protein 2', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35598602, end_position_on_the_genomic_accession=35617185, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67870, CurrentID=0, Status='live', Symbol='Enoph1', Aliases='2310057D15Rik, BB183658, C81437', description='enolase-phosphatase 1', other_designations='2,3-diketo-5-methylthio-1-phosphopentane phosphatase|MASA homolog', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100039994, end_position_on_the_genomic_accession=100068925, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67871, CurrentID=0, Status='live', Symbol='Mrrf', Aliases='2400002D02Rik, MRFF, MTRRF, RRF', description='mitochondrial ribosome recycling factor', other_designations='ribosome-releasing factor, mitochondrial', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36136398, end_position_on_the_genomic_accession=36190284, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67872, CurrentID=0, Status='live', Symbol='Nsmce4a', Aliases='2410003A14Rik', description='non-SMC element 4 homolog A (S. cerevisiae)', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=130532526, end_position_on_the_genomic_accession=130547381, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67873, CurrentID=0, Status='live', Symbol='Mri1', Aliases='2410018C20Rik', description='methylthioribose-1-phosphate isomerase homolog (S. cerevisiae)', other_designations='M1Pi|MTR-1-P isomerase|S-methyl-5-thioribose-1-phosphate isomerase|translation initiation factor eIF-2B subunit alpha/beta/delta-like protein', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84249629, end_position_on_the_genomic_accession=84257337, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67874, CurrentID=0, Status='live', Symbol='Rprm', Aliases='2410012A13Rik, Reprimo', description='reprimo, TP53 dependent G2 arrest mediator candidate', other_designations='reprimo, TP53 dependant G2 arrest mediator candidate', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=54084093, end_position_on_the_genomic_accession=54085552, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  67876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67876, CurrentID=0, Status='live', Symbol='Coq10b', Aliases='1500041J02Rik, 1700030I21Rik, 9530077A17Rik, AV002237, AW742641', description='coenzyme Q10B', other_designations='coenzyme Q10 homolog B', map_location='1 C1.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55052770, end_position_on_the_genomic_accession=55072702, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67877, CurrentID=0, Status='live', Symbol='Naa20', Aliases='1500004D14Rik, 2900026I01Rik, AU041458, D2Ertd186e, Nat5', description='N(alpha)-acetyltransferase 20, NatB catalytic subunit', other_designations='N-acetyltransferase 5 (ARD1 homolog, S. cerevisiae)|N-alpha-acetyltransferase 20, NatB catalytic subunit|N-terminal acetyltransferase B complex catalytic subunit NAA20|N-terminal acetyltransferase B complex catalytic subunit NAT5|methionine N-acetyltransferase|natB catalytic subunit|natB complex subunit NAT5', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=145903206, end_position_on_the_genomic_accession=145916425, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67878, CurrentID=0, Status='live', Symbol='Tmem33', Aliases='1110006G02Rik, 1600019D15Rik, 2410089A21Rik, 2700052H22Rik, 5430406L04Rik, AA388285, AI314185, AI426615, db83', description='transmembrane protein 33', other_designations='', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=67260565, end_position_on_the_genomic_accession=67291461, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100862247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100862247, CurrentID=0, Status='live', Symbol='Gm21586', Aliases='', description='predicted gene, 21586', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42438945, end_position_on_the_genomic_accession=42439974, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67880, CurrentID=0, Status='live', Symbol='Dcxr', Aliases='0610038K04Rik, 1810027P18Rik', description='dicarbonyl L-xylulose reductase', other_designations='XR|diacetyl/L-xylulose reductase', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120725373, end_position_on_the_genomic_accession=120727322, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67881, CurrentID=0, Status='live', Symbol='Mdp1', Aliases='1810034K20Rik, AI035604', description='magnesium-dependent phosphatase 1', other_designations='MDP-1', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55657879, end_position_on_the_genomic_accession=55660508, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67883, CurrentID=0, Status='live', Symbol='Uxs1', Aliases='1600025I13Rik, AI451869, AI649125, AW550562', description='UDP-glucuronate decarboxylase 1', other_designations='UGD|UXS-1', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=43748459, end_position_on_the_genomic_accession=43827821, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  67884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67884, CurrentID=0, Status='live', Symbol='1810043G02Rik', Aliases='AV026620, D10Jhu13e', description='RIKEN cDNA 1810043G02 gene', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77978530, end_position_on_the_genomic_accession=77985438, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67885, CurrentID=0, Status='live', Symbol='1500011K16Rik', Aliases='', description='RIKEN cDNA 1500011K16 gene', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127791377, end_position_on_the_genomic_accession=127792488, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  67886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67886, CurrentID=0, Status='live', Symbol='Camsap2', Aliases='1600013L13Rik, 4930541M15Rik, Camsap1l1, mKIAA1078', description='calmodulin regulated spectrin-associated protein family, member 2', other_designations='calmodulin regulated spectrin-associated protein 1-like 1|calmodulin-regulated spectrin-associated protein 1-like protein 1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136268123, end_position_on_the_genomic_accession=136346131, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  67887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67887, CurrentID=0, Status='live', Symbol='Saraf', Aliases='1810045K07Rik, Tmem66', description='store-operated calcium entry-associated regulatory factor', other_designations='SOCE-associated regulatory factor|transmembrane protein 66', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=34154563, end_position_on_the_genomic_accession=34170847, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67888, CurrentID=0, Status='live', Symbol='Tmem100', Aliases='1810057C19Rik, AV011897', description='transmembrane protein 100', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=90030348, end_position_on_the_genomic_accession=90036508, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67889, CurrentID=0, Status='live', Symbol='Rbm18', Aliases='2010004P11Rik, AU015530', description='RNA binding motif protein 18', other_designations='RNA-binding motif protein 18', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36116079, end_position_on_the_genomic_accession=36136704, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67890, CurrentID=0, Status='live', Symbol='Ufm1', Aliases='1810045K17Rik, AI132708, AI463323, ENSMUSG00000074598, Gm10726', description='ubiquitin-fold modifier 1', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=53853376, end_position_on_the_genomic_accession=53863807, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67891, CurrentID=0, Status='live', Symbol='Rpl4', Aliases='2010004J23Rik', description='ribosomal protein L4', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64173387, end_position_on_the_genomic_accession=64178562, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67892, CurrentID=0, Status='live', Symbol='Coa6', Aliases='1810063B05Rik, AI447995', description='cytochrome c oxidase assembly factor 6', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=126422501, end_position_on_the_genomic_accession=126425435, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67893, CurrentID=0, Status='live', Symbol='Tmem86a', Aliases='1810054O13Rik, AI414959, AW742698', description='transmembrane protein 86A', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=47050640, end_position_on_the_genomic_accession=47054776, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67894, CurrentID=0, Status='live', Symbol='Fam45a', Aliases='1810055E12Rik', description='family with sequence similarity 45, member A', other_designations='', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60811441, end_position_on_the_genomic_accession=60836235, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67895, CurrentID=0, Status='live', Symbol='Ppa1', Aliases='2010317E03Rik, Pyp', description='pyrophosphatase (inorganic) 1', other_designations='PPase|pyrophosphate phospho-hydrolase', map_location='10 32.25 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61648621, end_position_on_the_genomic_accession=61674165, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  67896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67896, CurrentID=0, Status='live', Symbol='Ccdc80', Aliases='2610001E17Rik, DRO1, Ssg1, Urb', description='coiled-coil domain containing 80', other_designations='steroid sensitive gene 1|up-regulated in BRS-3 deficient mouse', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45093407, end_position_on_the_genomic_accession=45127924, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67897, CurrentID=0, Status='live', Symbol='Rnmt', Aliases='2610002P10Rik, AI848273, Rg7mt1, mKIAA0398', description='RNA (guanine-7-) methyltransferase', other_designations='mRNA (guanine-N(7)-)-methyltransferase|mRNA cap methyltransferase', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=68300338, end_position_on_the_genomic_accession=68324852, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67898, CurrentID=0, Status='live', Symbol='Pef1', Aliases='2600002E23Rik, Peflin', description='penta-EF hand domain containing 1', other_designations='PEF protein with a long N-terminal hydrophobic domain|penta-EF hand domain-containing protein 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130107556, end_position_on_the_genomic_accession=130128134, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67899, CurrentID=0, Status='live', Symbol='Cmc1', Aliases='2010110K16Rik, 2010312P05Rik, AI596001', description='COX assembly mitochondrial protein 1', other_designations='cmc1p', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=118063983, end_position_on_the_genomic_accession=118150196, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67900, CurrentID=0, Status='live', Symbol='Mtfp1', Aliases='1700020C11Rik, 2610507A21Rik, AV005788, Mtp18', description='mitochondrial fission process 1', other_designations='mitochondrial 18 kDa protein|mitochondrial fission protein MTP18', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4091481, end_position_on_the_genomic_accession=4095431, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67902, CurrentID=0, Status='live', Symbol='Sumf2', Aliases='2610040F05Rik', description='sulfatase modifying factor 2', other_designations='C-alpha-formylglycine-generating enzyme 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129846990, end_position_on_the_genomic_accession=129863951, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67903, CurrentID=0, Status='live', Symbol='Gipc1', Aliases='GIPC, Glut1CIP, Rgs19ip1, Semcap1, TIP-2, TaxIP2', description='GIPC PDZ domain containing family, member 1', other_designations='GAIP C-terminus-interacting protein|RGS-GAIP interacting protein GIPC|RGS-GAIP-interacting protein|RGS19-interacting protein 1|SEMCAP-1|Tax interacting protein, clone 2|neurophilin1-IP|regulator of G-protein signaling 19 interacting protein 1|semaF cytoplasmic domain associated protein|semaF cytoplasmic domain-associated protein 1|synectin', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83652678, end_position_on_the_genomic_accession=83664789, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67905, CurrentID=0, Status='live', Symbol='Ppm1m', Aliases='2810423O19Rik, AW610647, C77250, PP2Ceta', description='protein phosphatase 1M', other_designations='PP2C eta|PP2C-eta|PP2CE|protein phosphatase 2C eta|protein phosphatase 2C eta-2|protein phosphatase 2C isoform eta', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106194953, end_position_on_the_genomic_accession=106199233, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  330050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330050, CurrentID=0, Status='live', Symbol='Fam185a', Aliases='4631428I10, AI847670', description='family with sequence similarity 185, member A', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21424861, end_position_on_the_genomic_accession=21488271, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67909, CurrentID=0, Status='live', Symbol='Galntl5', Aliases='1700021B12Rik, Galnt15', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 5', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 15|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 15|galNAc-T15|polypeptide GalNAc transferase 15|polypeptide N-acetylgalactosaminyltransferase-like 5|pp-GaNTase 15|protein-UDP acetylgalactosaminyltransferase 15|putative polypeptide N-acetylgalactosaminyltransferase-like protein 5', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=25181440, end_position_on_the_genomic_accession=25220297, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67911, CurrentID=0, Status='live', Symbol='Zfp169', Aliases='1700025J14Rik, 4930429A13Rik, BE688704', description='zinc finger protein 169', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48484092, end_position_on_the_genomic_accession=48513478, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67912, CurrentID=0, Status='live', Symbol='1600012H06Rik', Aliases='', description='RIKEN cDNA 1600012H06 gene', other_designations='', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14943184, end_position_on_the_genomic_accession=14945939, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  67914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67914, CurrentID=0, Status='live', Symbol='Coq9', Aliases='2310005O14Rik, C78387', description='coenzyme Q9', other_designations='coenzyme Q9 homolog', map_location='8 C5-D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94838417, end_position_on_the_genomic_accession=94854895, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  67916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67916, CurrentID=0, Status='live', Symbol='Plpp3', Aliases='1110003O22Rik, 2610002D05Rik, AV025606, D4Bwg0538e, D4Bwg1535e, Lpp3, Ppab2b, Ppap2b', description='phospholipid phosphatase 3', other_designations='PAP-2b|PAP2-beta|PAP2b|lipid phosphate phosphohydrolase 3|phosphatidate phosphohydrolase type 2b|phosphatidic acid phosphatase 2b|phosphatidic acid phosphatase type 2B', map_location='4 49.18 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=105157347, end_position_on_the_genomic_accession=105232767, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67917, CurrentID=0, Status='live', Symbol='Zcchc3', Aliases='2810406K24Rik', description='zinc finger, CCHC domain containing 3', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152411956, end_position_on_the_genomic_accession=152415044, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  330064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330064, CurrentID=0, Status='live', Symbol='Slc5a6', Aliases='E430023I20', description='solute carrier family 5 (sodium-dependent vitamin transporter), member 6', other_designations='Na(+)-dependent multivitamin transporter|solute carrier family 5 member 6', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31036036, end_position_on_the_genomic_accession=31048924, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  67921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67921, CurrentID=0, Status='live', Symbol='Ube2f', Aliases='2510010F15Rik, AI851109', description='ubiquitin-conjugating enzyme E2F (putative)', other_designations='NEDD8 carrier protein UBE2F|NEDD8 protein ligase UBE2F|NEDD8-conjugating enzyme 2|ubiquitin-conjugating enzyme E2 F', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91250306, end_position_on_the_genomic_accession=91286516, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67922, CurrentID=0, Status='live', Symbol='Fam32a', Aliases='2510049I19Rik, AW822030, OTAG-12, Otag12', description='family with sequence similarity 32, member A', other_designations='ovarian tumor associated gene 12|ovarian tumor associated gene-12', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72219730, end_position_on_the_genomic_accession=72223775, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67923, CurrentID=0, Status='live', Symbol='Tceb1', Aliases='2610043E24Rik, 2610301I15Rik, AA407206, AI987979, AW049146', description='transcription elongation factor B (SIII), polypeptide 1', other_designations='RNA polymerase II transcription factor SIII subunit C|SIII p15|eloC|elongin 15 kDa subunit|elongin C|elongin-C|stromal membrane-associated protein SMAP1B homolog', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16641725, end_position_on_the_genomic_accession=16657097, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67926, CurrentID=0, Status='live', Symbol='Spert', Aliases='1700086N05Rik, AI596460, Nurit', description='spermatid associated', other_designations='novel leucine zipper testicular protein|protein nurit|testis-specific leucine zipper protein nurit', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75582834, end_position_on_the_genomic_accession=75593177, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67928, CurrentID=0, Status='live', Symbol='Abca14', Aliases='1700110B15Rik, 4930539G24Rik', description='ATP-binding cassette, sub-family A (ABC1), member 14', other_designations='ATP-binding cassette transporter sub-family A member 14', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120203895, end_position_on_the_genomic_accession=120325352, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  67929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67929, CurrentID=0, Status='live', Symbol='Ccdc70', Aliases='1700112P19Rik', description='coiled-coil domain containing 70', other_designations='', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21969741, end_position_on_the_genomic_accession=21974054, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67931, CurrentID=0, Status='live', Symbol='Serpini2', Aliases='Spi14', description='serine (or cysteine) peptidase inhibitor, clade I, member 2', other_designations='pancreas-specific serpin|serine protease inhibitor 14', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=75242357, end_position_on_the_genomic_accession=75270078, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67932, CurrentID=0, Status='live', Symbol='1700129C05Rik', Aliases='', description='RIKEN cDNA 1700129C05 gene', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59133040, end_position_on_the_genomic_accession=59142930, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67933, CurrentID=0, Status='live', Symbol='Hcfc2', Aliases='1700129L13Rik, AU019838', description='host cell factor C2', other_designations='C2 factor|HCF-2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82697497, end_position_on_the_genomic_accession=82795865, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  67935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67935, CurrentID=0, Status='live', Symbol='Ces5a', Aliases='1700081L16Rik, 1700122C07Rik, BB081581, Ces7, Gm503, cauxin', description='carboxylesterase 5A', other_designations='carboxylesterase 7|carboxylesterase-like urinary excreted protein homolog', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93499063, end_position_on_the_genomic_accession=93535881, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  67936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67936, CurrentID=0, Status='live', Symbol='Wdr55', Aliases='2410080P20Rik, C80692', description='WD repeat domain 55', other_designations='', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36759739, end_position_on_the_genomic_accession=36763708, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67937, CurrentID=0, Status='live', Symbol='Tmem59l', Aliases='5330410G16Rik', description='transmembrane protein 59-like', other_designations='brain-specific membrane-anchored protein', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70483863, end_position_on_the_genomic_accession=70487468, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67938, CurrentID=0, Status='live', Symbol='Myl12b', Aliases='1500001M02Rik, C77744, Mylc2b, RLC-B', description='myosin, light chain 12B, regulatory', other_designations='MLC20|myosin light chain, regulatory B|myosin regulatory light chain 2-B, smooth muscle isoform|myosin regulatory light chain 20 kDa|myosin regulatory light chain MRLC2', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=70973963, end_position_on_the_genomic_accession=70990516, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67939, CurrentID=0, Status='live', Symbol='Prorsd1', Aliases='2010316F05Rik, Ncrna00117, Prdxdd1', description='prolyl-tRNA synthetase domain containing 1', other_designations='PrdX-deacylase domain 1|prdX deacylase domain-containing protein 1|prdX-deacylase domain-containing protein 1', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29511757, end_position_on_the_genomic_accession=29515033, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67941, CurrentID=0, Status='live', Symbol='Rps27l', Aliases='1810034D23Rik', description='ribosomal protein S27-like', other_designations='40S ribosomal protein S27-like protein|rps27l', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66946076, end_position_on_the_genomic_accession=66949520, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  67942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67942, CurrentID=0, Status='live', Symbol='Atp5g2', Aliases='1810041M08Rik', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit C2 (subunit 9)', other_designations='ATP synthase lipid-binding protein, mitochondrial|ATP synthase proteolipid P2|ATP synthase, H+ transporting, mitochondrial F0 complex, subunit c (subunit 9)|ATPase protein 9|ATPase subunit c', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102662864, end_position_on_the_genomic_accession=102671134, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67943, CurrentID=0, Status='live', Symbol='Mesdc2', Aliases='2210015O11Rik, AW537813, mKIAA0081, mesd, msd', description='mesoderm development candidate 2', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=83892000, end_position_on_the_genomic_accession=83901532, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  67944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67944, CurrentID=0, Status='live', Symbol='Tex13a', Aliases='1700025D03Rik', description='testis expressed 13A', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=138208158, end_position_on_the_genomic_accession=138209591, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67945, CurrentID=0, Status='live', Symbol='Rpl41', Aliases='1810055P16Rik, 2210411K19Rik', description='ribosomal protein L41', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128548110, end_position_on_the_genomic_accession=128549168, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67946, CurrentID=0, Status='live', Symbol='Spata6', Aliases='1700062C23Rik, AI790763, Hash, KRP, Mash', description='spermatogenesis associated 6', other_designations='kinesin-related protein', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=111719941, end_position_on_the_genomic_accession=111829137, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  67948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67948, CurrentID=0, Status='live', Symbol='Fbxo28', Aliases='4833428J17Rik, 5730505P19Rik, D1Ertd578e, Fbx28, mKIAA0483', description='F-box protein 28', other_designations='', map_location='1 84.93 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182313102, end_position_on_the_genomic_accession=182341606, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67949, CurrentID=0, Status='live', Symbol='Nifk', Aliases='AI852665, C130020J04Rik, Mki67ip', description='nucleolar protein interacting with the FHA domain of MKI67', other_designations='Mki67 (FHA domain) interacting nucleolar phosphoprotein|mNIFK|nucleolar protein interacting with the FHA domain of pKI-67', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=118321843, end_position_on_the_genomic_accession=118333831, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  67951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67951, CurrentID=0, Status='live', Symbol='Tubb6', Aliases='2310057H16Rik, BB220206', description='tubulin, beta 6 class V', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67390731, end_position_on_the_genomic_accession=67402749, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67952, CurrentID=0, Status='live', Symbol='Tomm20', Aliases='1810060K07Rik, BB284719, Gm19268, MAS20, MOM19, TOM20, mKIAA0016', description='translocase of outer mitochondrial membrane 20 homolog (yeast)', other_designations='mitochondrial 20 kDa outer membrane protein|outer mitochondrial membrane receptor Tom20', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=126930664, end_position_on_the_genomic_accession=126945921, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  67955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67955, CurrentID=0, Status='live', Symbol='Sugt1', Aliases='2410174K12Rik, SGT1', description='SGT1, suppressor of G2 allele of SKP1 (S. cerevisiae)', other_designations='suppressor of G2 allele of SKP1 homolog', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79587691, end_position_on_the_genomic_accession=79629755, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  67956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67956, CurrentID=0, Status='live', Symbol='Setd8', Aliases='2410195B05Rik, AA617402, AW536475, PR-SET7, PR/SET07', description='SET domain containing (lysine methyltransferase) 8', other_designations='H4-K20-HMTase SETD8|PR/SET domain-containing protein 07|SET domain-containing protein 8|histone-lysine N-methyltransferase SETD8', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124439930, end_position_on_the_genomic_accession=124462308, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  67958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67958, CurrentID=0, Status='live', Symbol='U2surp', Aliases='2610101N10Rik, AU023006, Sr140', description='U2 snRNP-associated SURP domain containing', other_designations='140 kDa Ser/Arg-rich domain protein|U2-associated protein SR140', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95456894, end_position_on_the_genomic_accession=95513208, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  67959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67959, CurrentID=0, Status='live', Symbol='Puf60', Aliases='2410104I19Rik, 2810454F19Rik, SIAHBP1', description='poly-U binding splicing factor 60', other_designations='60 kDa poly(U)-binding-splicing factor|fuse-binding protein-interacting repressor', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76070182, end_position_on_the_genomic_accession=76080950, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  67963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67963, CurrentID=0, Status='live', Symbol='Npc2', Aliases='2700012J19Rik, AA408070, AU045843, HE1', description='Niemann-Pick type C2', other_designations='mE1|niemann Pick type C2 protein homolog', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84754559, end_position_on_the_genomic_accession=84773112, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  67966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67966, CurrentID=0, Status='live', Symbol='Zcchc10', Aliases='2410141K03Rik, AA675321, D11Ertd416e', description='zinc finger, CCHC domain containing 10', other_designations='', map_location='11 31.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53324689, end_position_on_the_genomic_accession=53333301, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67967, CurrentID=0, Status='live', Symbol='Pold3', Aliases='2410142G14Rik, C85233, P66, P68', description='polymerase (DNA-directed), delta 3, accessory subunit', other_designations='DNA polymerase delta subunit p66|GC12|polymerase (DNA directed), delta 3', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100082111, end_position_on_the_genomic_accession=100121560, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  67968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67968, CurrentID=0, Status='live', Symbol='Ooep', Aliases='2410146L05Rik, Floped, Moep19, Sddr', description='oocyte expressed protein', other_designations='STAT3 downstream gene and differentiation regulator|factor located in oocytes permitting embryonic development|oocyte and embryo specific protein|oocyte expressed protein homolog (dog)|oocyte- and embryo-specific protein 19', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78376903, end_position_on_the_genomic_accession=78378588, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  67971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67971, CurrentID=0, Status='live', Symbol='Tppp3', Aliases='2700055K07Rik, CGI-38, Ceacam9, mmCGM8', description='tubulin polymerization-promoting protein family member 3', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105467492, end_position_on_the_genomic_accession=105471422, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67972, CurrentID=0, Status='live', Symbol='Atp2b1', Aliases='2810442I22Rik, E130111D10Rik, Pmca1', description='ATPase, Ca++ transporting, plasma membrane 1', other_designations='plasma membrane calcium ATPase|plasma membrane calcium pump', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=98914359, end_position_on_the_genomic_accession=99026143, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  67973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67973, CurrentID=0, Status='live', Symbol='Mphosph10', Aliases='2810453H10Rik, 5730405D16Rik, AI326008', description='M-phase phosphoprotein 10 (U3 small nucleolar ribonucleoprotein)', other_designations='m phase phosphoprotein 10', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64376541, end_position_on_the_genomic_accession=64392236, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67974, CurrentID=0, Status='live', Symbol='Ccny', Aliases='1700025H17Rik, 3110050L10Rik, 4631402G10Rik, 5730405I09Rik', description='cyclin Y', other_designations='cyclin fold protein 1', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=9312304, end_position_on_the_genomic_accession=9450150, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  330119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330119, CurrentID=0, Status='live', Symbol='Adamts3', Aliases='1100001H14Rik, 6330442E02Rik', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 3', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 3', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=89673841, end_position_on_the_genomic_accession=89883669, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  67976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67976, CurrentID=0, Status='live', Symbol='Trabd', Aliases='5730502D15Rik, AL023039', description='TraB domain containing', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89076050, end_position_on_the_genomic_accession=89087078, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  330122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330122, CurrentID=0, Status='live', Symbol='Cxcl3', Aliases='Dcip1, Gm1960', description='chemokine (C-X-C motif) ligand 3', other_designations='cxcl3|dendritic cell inflammatory protein 1', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90786101, end_position_on_the_genomic_accession=90788093, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  67979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67979, CurrentID=0, Status='live', Symbol='Atad1', Aliases='4921525H23Rik, AW107648, Thorase', description='ATPase family, AAA domain containing 1', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=32672558, end_position_on_the_genomic_accession=32739766, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  67980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67980, CurrentID=0, Status='live', Symbol='Gnpda2', Aliases='4921523I18Rik, 4933412A11Rik, BB189630, GNPDA, Gnp2, Sb52, mKIAA4008', description='glucosamine-6-phosphate deaminase 2', other_designations='GNPDA 2|glcN6P deaminase 2|oscillin', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=69574998, end_position_on_the_genomic_accession=69592332, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  67981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67981, CurrentID=0, Status='live', Symbol='Hormad1', Aliases='4921522K05Rik, Nohma', description='HORMA domain containing 1', other_designations='HORMA domain containing protein|newborn ovary HORMA protein', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95559640, end_position_on_the_genomic_accession=95587940, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67983, CurrentID=0, Status='live', Symbol='Pdzd9', Aliases='4930408O21Rik', description='PDZ domain containing 9', other_designations='PDZ domain-containing protein C16orf65 homolog', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120659104, end_position_on_the_genomic_accession=120677562, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  67985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67985, CurrentID=0, Status='live', Symbol='Ssxb1', Aliases='4930414C09Rik', description='synovial sarcoma, X member B, breakpoint 1', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8413305, end_position_on_the_genomic_accession=8422157, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  67988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67988, CurrentID=0, Status='live', Symbol='Tmx3', Aliases='6430411B10Rik, A730024F05Rik, AV259382, Txndc10, mKIAA1830', description='thioredoxin-related transmembrane protein 3', other_designations='thioredoxin domain containing 10|thioredoxin domain-containing protein 10', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=90510154, end_position_on_the_genomic_accession=90543267, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  67991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67991, CurrentID=0, Status='live', Symbol='Nacc2', Aliases='0610020I02Rik, AI448087, Btbd14a, C030048H19Rik', description='nucleus accumbens associated 2, BEN and BTB (POZ) domain containing', other_designations='BTB (POZ) domain containing 14A|BTB/POZ domain-containing protein 14A|NAC-2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26055535, end_position_on_the_genomic_accession=26122811, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  67993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67993, CurrentID=0, Status='live', Symbol='Nudt12', Aliases='0610016O18Rik', description='nudix (nucleoside diphosphate linked moiety X)-type motif 12', other_designations='nucleoside diphosphate-linked moiety X motif 12|nudix (nucleoside diphosphnudix (nucleoside diphosphate linked moiety X)-type motif 12ate linked moiety X)-type motif 12|nudix motif 12', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=59000107, end_position_on_the_genomic_accession=59013367, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  67994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67994, CurrentID=0, Status='live', Symbol='Mrps11', Aliases='0710005I03Rik, C79873', description='mitochondrial ribosomal protein S11', other_designations='MRP-S11|S11mt', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=78783131, end_position_on_the_genomic_accession=78792989, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67996, CurrentID=0, Status='live', Symbol='Srsf6', Aliases='1210001E11Rik, AI314910, AW146126, Sfrs6', description='serine/arginine-rich splicing factor 6', other_designations='pre-mRNA-splicing factor SRP55|splicing factor, arginine/serine-rich 6', map_location='2 83.87 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=162931508, end_position_on_the_genomic_accession=162937121, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  67997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67997, CurrentID=0, Status='live', Symbol='Ddx59', Aliases='1210002B07Rik, 4833411G06Rik', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 59', other_designations='DEAD box protein 59', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136415271, end_position_on_the_genomic_accession=136456190, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  67998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=67998, CurrentID=0, Status='live', Symbol='Fam134c', Aliases='1300010M03Rik, 4933404C01Rik, AI551748', description='family with sequence similarity 134, member C', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101096322, end_position_on_the_genomic_accession=101119894, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68001, CurrentID=0, Status='live', Symbol='1110004E09Rik', Aliases='', description='RIKEN cDNA 1110004E09 gene', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90925811, end_position_on_the_genomic_accession=90934849, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68002, CurrentID=0, Status='live', Symbol='Sdhaf4', Aliases='1110058L19Rik, 1700001E18Rik, C6orf57', description='succinate dehydrogenase complex assembly factor 4', other_designations='SDH assembly factor 4|UPF0369 protein C6orf57 homolog', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=23995939, end_position_on_the_genomic_accession=24005640, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  330149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330149, CurrentID=0, Status='live', Symbol='Hfm1', Aliases='A330009G12Rik, Gm1046, Mer3, Sec63d1', description='HFM1, ATP-dependent DNA helicase homolog', other_designations='SEC63 domain containing 1', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=106840193, end_position_on_the_genomic_accession=106926571, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  68009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68009, CurrentID=0, Status='live', Symbol='Defa20', Aliases='2010016F14Rik, Defcr20', description='defensin, alpha, 20', other_designations='cryptdin-4|cryptdin4|defensin-related cryptdin-20', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21509258, end_position_on_the_genomic_accession=21510237, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68010, CurrentID=0, Status='live', Symbol='Bambi', Aliases='2610003H06Rik', description='BMP and activin membrane-bound inhibitor', other_designations='BMP and activin membrane-bound inhibitor, homolog|putative transmembrane protein NMA', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=3507957, end_position_on_the_genomic_accession=3516404, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68011, CurrentID=0, Status='live', Symbol='Snrpg', Aliases='2810024K17Rik, AL022803, SMG, sm-G, snRNP-G', description='small nuclear ribonucleoprotein polypeptide G', other_designations='sm protein G', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86371540, end_position_on_the_genomic_accession=86378902, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68014, CurrentID=0, Status='live', Symbol='Zwilch', Aliases='2310031L18Rik', description='zwilch kinetochore protein', other_designations='Zwilch, kinetochore associated, homolog', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64137144, end_position_on_the_genomic_accession=64173104, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  68015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68015, CurrentID=0, Status='live', Symbol='Trap1', Aliases='2410002K23Rik, HSP75', description='TNF receptor-associated protein 1', other_designations='HSP 75|TNFR-associated protein 1|TRAP-1|tumor necrosis factor type 1 receptor-associated protein', map_location='16 2.38 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4039977, end_position_on_the_genomic_accession=4077810, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  68016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68016, CurrentID=0, Status='live', Symbol='Murc', Aliases='2310039E09Rik', description='muscle-related coiled-coil protein', other_designations='cavin 4|muscle-restricted coiled-coil protein', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48663514, end_position_on_the_genomic_accession=48673492, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68017, CurrentID=0, Status='live', Symbol='Ftsj2', Aliases='2310037B18Rik', description='FtsJ RNA methyltransferase homolog 2 (E. coli)', other_designations="16S rRNA (uridine(1369)-2'-O)-methyltransferase|16S rRNA [Um1369] 2'-O-methyltransferase|FtsJ homolog 2|protein ftsJ homolog 2|putative ribosomal RNA methyltransferase 2|rRNA (uridine-2'-O-)-methyltransferase", map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140327674, end_position_on_the_genomic_accession=140331898, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68018, CurrentID=0, Status='live', Symbol='Col4a3bp', Aliases='2810404O15Rik, 9230101K08Rik, AU016711, CERT, GPBP', description='collagen, type IV, alpha 3 (Goodpasture antigen) binding protein', other_designations='START domain-containing protein 11|ceramide transfer protein|ceramide transport protein|goodpasture antigen-binding protein|procollagen, type IV, alpha 3 (Goodpasture antigen) binding protein|stAR-related lipid transfer protein 11|stARD11', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=96542735, end_position_on_the_genomic_accession=96640167, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  68020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68020, CurrentID=0, Status='live', Symbol='Apopt1', Aliases='1700081D05Rik, 2810002N01Rik, Apop-1, Apop1', description='apoptogenic, mitochondrial 1', other_designations='UPF0671 protein C14orf153 homolog', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111713269, end_position_on_the_genomic_accession=111755055, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68021, CurrentID=0, Status='live', Symbol='Bphl', Aliases='2010012D11Rik, 5730533B08Rik, AI115341', description='biphenyl hydrolase-like (serine hydrolase, breast epithelial mucin-associated antigen)', other_designations='VACVase|biphenyl hydrolase-like protein|valacyclovirase', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34037593, end_position_on_the_genomic_accession=34074074, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  330166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330166, CurrentID=0, Status='live', Symbol='Miat', Aliases='3632434I06, A230057G18Rik, Rncr2, gomafu', description='myocardial infarction associated transcript (non-protein coding)', other_designations='retinal noncoding RNA 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112213228, end_position_on_the_genomic_accession=112228948, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68023, CurrentID=0, Status='live', Symbol='Pdf', Aliases='2610019N19Rik', description='peptide deformylase (mitochondrial)', other_designations='peptide deformylase-like protein', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107046289, end_position_on_the_genomic_accession=107048614, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68024, CurrentID=0, Status='live', Symbol='Hist1h2bc', Aliases='2610022J01Rik, H2bfs, R74621', description='histone cluster 1, H2bc', other_designations='H2B histone family, member S|histone 1, H2bc', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23684199, end_position_on_the_genomic_accession=23692480, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68026, CurrentID=0, Status='live', Symbol='2810417H13Rik', Aliases='AA409629, Ns5apt9, Ns5atp9, PAF15, Paf, mKIAA0101, p15(PAF)', description='RIKEN cDNA 2810417H13 gene', other_designations='HCV NS5A-transactivated protein 9 homolog|PCNA-associated factor of 15 kDa|p15PAF', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65890300, end_position_on_the_genomic_accession=65903553, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  330171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330171, CurrentID=0, Status='live', Symbol='Kctd10', Aliases='AW536343, C87062, mBACURD3', description='potassium channel tetramerisation domain containing 10', other_designations='BTB/POZ domain-containing protein KCTD10|potassium channel tetramerization domain-containing 10', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114363572, end_position_on_the_genomic_accession=114380532, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68028, CurrentID=0, Status='live', Symbol='Rpl22l1', Aliases='3110001N18Rik, AU041196', description='ribosomal protein L22 like 1', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=28805444, end_position_on_the_genomic_accession=28807424, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  330173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330173, CurrentID=0, Status='live', Symbol='2610524H06Rik', Aliases='', description='RIKEN cDNA 2610524H06 gene', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114821937, end_position_on_the_genomic_accession=114823468, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68031, CurrentID=0, Status='live', Symbol='Rnf146', Aliases='2610509H23Rik, Iduna', description='ring finger protein 146', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=29344176, end_position_on_the_genomic_accession=29362442, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68032, CurrentID=0, Status='live', Symbol='Emc4', Aliases='2610318K02Rik, Tmem85', description='ER membrane protein complex subunit 4', other_designations='transmembrane protein 85', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112363019, end_position_on_the_genomic_accession=112368027, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  330177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330177, CurrentID=0, Status='live', Symbol='Taok3', Aliases='2900006A08Rik, A130052D22, A430105I05Rik', description='TAO kinase 3', other_designations='thousand and one amino acid protein 3', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117120112, end_position_on_the_genomic_accession=117275219, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  68034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68034, CurrentID=0, Status='live', Symbol='Fam122a', Aliases='2410124L17Rik, 2900009I07Rik, ENSMUSG00000051082, Gm9849', description='family with sequence similarity 122, member A', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24475779, end_position_on_the_genomic_accession=24477474, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68035, CurrentID=0, Status='live', Symbol='Rbm42', Aliases='1700003D06Rik, 3100004P22Rik', description='RNA binding motif protein 42', other_designations='RNA-binding motif protein 42', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30640995, end_position_on_the_genomic_accession=30650228, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68036, CurrentID=0, Status='live', Symbol='Zfp706', Aliases='3110006P09Rik, AA793007, AW536850, C77232, Znf706', description='zinc finger protein 706', other_designations='', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36997027, end_position_on_the_genomic_accession=37007402, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68038, CurrentID=0, Status='live', Symbol='Chid1', Aliases='3110023E09Rik, AI841869', description='chitinase domain containing 1', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141493136, end_position_on_the_genomic_accession=141539888, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  68039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68039, CurrentID=0, Status='live', Symbol='Nmb', Aliases='3110023K12Rik', description='neuromedin B', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80902227, end_position_on_the_genomic_accession=80911430, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68040, CurrentID=0, Status='live', Symbol='Zfp593', Aliases='3110024A21Rik, AV062409, E130106C14Rik, ZNF593', description='zinc finger protein 593', other_designations='zinc finger protein T86', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134243306, end_position_on_the_genomic_accession=134245591, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68041, CurrentID=0, Status='live', Symbol='Mid1ip1', Aliases='3110038L01Rik, Mig12, S14R', description='Mid1 interacting protein 1 (gastrulation specific G12-like (zebrafish))', other_designations='gastrulation-specific G12-like protein|mid1-interacting G12-like protein|protein STRAIT11499 homolog|spot 14-R|spot 14-related protein', map_location='X A1.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=10717365, end_position_on_the_genomic_accession=10719702, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68043, CurrentID=0, Status='live', Symbol='Eef1akmt1', Aliases='2510005D08Rik, AW045965, Ayu21-96, Gt(pU21)96Imeg, GtAyu21-96, N6amt2', description='eukaryotic translation elongation factor 1 alpha lysine methyltransferase 1', other_designations='Gt(Ayu21)96Imeg|N-6 adenine-specific DNA methyltransferase 2 (putative)|n(6)-adenine-specific DNA methyltransferase 2', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57549596, end_position_on_the_genomic_accession=57571569, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68044, CurrentID=0, Status='live', Symbol='Chac2', Aliases='', description='ChaC, cation transport regulator 2', other_designations='ChaC, cation transport regulator homolog 2|cation transport regulator-like protein 2|gamma-GCG 2', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30976707, end_position_on_the_genomic_accession=30986365, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68045, CurrentID=0, Status='live', Symbol='2700060E02Rik', Aliases='', description='RIKEN cDNA 2700060E02 gene', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=19811402, end_position_on_the_genomic_accession=19823823, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68046, CurrentID=0, Status='live', Symbol='2700062C07Rik', Aliases='AI195775, C87515', description='RIKEN cDNA 2700062C07 gene', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24470871, end_position_on_the_genomic_accession=24477767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68047, CurrentID=0, Status='live', Symbol='Mpnd', Aliases='E130307M08Rik', description='MPN domain containing', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56009157, end_position_on_the_genomic_accession=56016790, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  330192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330192, CurrentID=0, Status='live', Symbol='Vps37b', Aliases='2300007F24Rik, AI415429, BC026744', description='vacuolar protein sorting 37B (yeast)', other_designations='ESCRT-I complex subunit VPS37B', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124004641, end_position_on_the_genomic_accession=124032260, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68050, CurrentID=0, Status='live', Symbol='Akirin1', Aliases='6330407G11Rik', description='akirin 1', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123735195, end_position_on_the_genomic_accession=123750299, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68051, CurrentID=0, Status='live', Symbol='Nutf2', Aliases='2700067I02Rik, AI115459, AI413596, AW546000, Ntf2, Ntf3, PP15', description='nuclear transport factor 2', other_designations='NTF-2', map_location='8 53.05 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105860634, end_position_on_the_genomic_accession=105880402, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68052, CurrentID=0, Status='live', Symbol='Rps13', Aliases='2700063M04Rik', description='ribosomal protein S13', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=116331507, end_position_on_the_genomic_accession=116334190, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68053, CurrentID=0, Status='live', Symbol='Ubxn2b', Aliases='3110003A22Rik, 6430407D20Rik, AI451665, p37', description='UBX domain protein 2B', other_designations='NSFL1 cofactor p37|p97 cofactor p37', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=6191105, end_position_on_the_genomic_accession=6219788, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68054, CurrentID=0, Status='live', Symbol='Serpina12', Aliases='4632419J12Rik', description='serine (or cysteine) peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 12', other_designations='serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 12|vaspin|visceral adipose tissue-derived serine protease inhibitor|visceral adipose-specific SERPIN', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104028769, end_position_on_the_genomic_accession=104044443, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68055, CurrentID=0, Status='live', Symbol='Atp5s', Aliases='1110015E18Rik, ATPW', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit S', other_designations='ATP synthase-coupling factor B|facyor B|mitochondrial ATP synthase regulatory component factor B', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69724930, end_position_on_the_genomic_accession=69744665, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68058, CurrentID=0, Status='live', Symbol='Chd1l', Aliases='4432404A22Rik, Alc1, Snf2p', description='chromodomain helicase DNA binding protein 1-like', other_designations='amplified in liver cancer 1', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97560748, end_position_on_the_genomic_accession=97610190, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  68059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68059, CurrentID=0, Status='live', Symbol='Tm9sf2', Aliases='1500001N15Rik, AA536814, D14Ertd64e, P76', description='transmembrane 9 superfamily member 2', other_designations='', map_location='14 65.86 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122107082, end_position_on_the_genomic_accession=122159603, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  76880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76880, CurrentID=0, Status='live', Symbol='6430411K18Rik', Aliases='PEG11as', description='RIKEN cDNA 6430411K18 gen', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109591537, end_position_on_the_genomic_accession=109592851, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68066, CurrentID=0, Status='live', Symbol='Slc25a39', Aliases='3010027G13Rik, D11Ertd333e', description='solute carrier family 25, member 39', other_designations='', map_location='11 66.29 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102402969, end_position_on_the_genomic_accession=102408675, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68067, CurrentID=0, Status='live', Symbol='3010026O09Rik', Aliases='2900016G09Rik, AI851514, AW475975', description='RIKEN cDNA 3010026O09 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50174851, end_position_on_the_genomic_accession=50200115, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68070, CurrentID=0, Status='live', Symbol='Pdzd2', Aliases='4930537L06Rik, A930022H17Rik, Gm82, Pdzk3, mKIAA0300', description='PDZ domain containing 2', other_designations='PDZ domain containing 3', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=12357054, end_position_on_the_genomic_accession=12739973, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  330216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330216, CurrentID=0, Status='live', Symbol='Mblac1', Aliases='', description='metallo-beta-lactamase domain containing 1', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138194314, end_position_on_the_genomic_accession=138195621, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  330217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330217, CurrentID=0, Status='live', Symbol='Gal3st4', Aliases='1500031A01Rik', description='galactose-3-O-sulfotransferase 4', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138264918, end_position_on_the_genomic_accession=138273038, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68075, CurrentID=0, Status='live', Symbol='Lurap1', Aliases='1520402A15Rik, Lrp35a', description='leucine rich adaptor protein 1', other_designations='NF-kappa-B activator C1orf190 homolog|leucine repeat adapter protein 35A', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116136668, end_position_on_the_genomic_accession=116144616, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68077, CurrentID=0, Status='live', Symbol='Gltscr2', Aliases='5330430H08Rik, 9430097C02Rik, AU041936, AW536441, PICT-1, R74911', description='glioma tumor suppressor candidate region gene 2', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15937835, end_position_on_the_genomic_accession=15946108, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  330222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330222, CurrentID=0, Status='live', Symbol='Sdk1', Aliases='5330440E10, 6720466O15Rik', description='sidekick homolog 1 (chicken)', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=141241497, end_position_on_the_genomic_accession=142215586, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  68079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68079, CurrentID=0, Status='live', Symbol='Pdcd2l', Aliases='6030457N17Rik', description='programmed cell death 2-like', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34184497, end_position_on_the_genomic_accession=34196714, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68080, CurrentID=0, Status='live', Symbol='Gpn3', Aliases='A930018B01Rik, Atpbd1c, D5Ertd708e', description='GPN-loop GTPase 3', other_designations='ATP binding domain 1 family, member C|ATP-binding domain 1 family member C|protein x 0004', map_location='5 62.33 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122372508, end_position_on_the_genomic_accession=122382771, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68082, CurrentID=0, Status='live', Symbol='Dusp19', Aliases='5930436K22Rik, C79103, SKRP1, TS-DSP1', description='dual specificity phosphatase 19', other_designations='protein phosphatase SKRP1|stress-activated protein kinase pathway-regulating phosphatase 1', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=80617214, end_position_on_the_genomic_accession=80631661, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68083, CurrentID=0, Status='live', Symbol='Pak1ip1', Aliases='5830431I15Rik, 5930415H02Rik, AA419825, AI314040, AW556169, Gdpd1, PIP1', description='PAK1 interacting protein 1', other_designations='PAK1-interacting protein 1|glycerophosphodiester phosphodiesterase domain containing 1|putative PAK inhibitor Skb15', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41001010, end_position_on_the_genomic_accession=41013033, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  330228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330228, CurrentID=0, Status='live', Symbol='Spdye4b', Aliases='4933411G11Rik', description='speedy/RINGO cell cycle regulator family, member E4B', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143180924, end_position_on_the_genomic_accession=143205069, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68087, CurrentID=0, Status='live', Symbol='Dcakd', Aliases='3010024O21Rik, 6720485C15Rik, AI849483', description='dephospho-CoA kinase domain containing', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102994047, end_position_on_the_genomic_accession=103028239, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68089, CurrentID=0, Status='live', Symbol='Arpc4', Aliases='20kDa, 5330419I20Rik, AI327076, p20-Arc', description='actin related protein 2/3 complex, subunit 4', other_designations='arp2/3 complex 20 kDa subunit', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113378113, end_position_on_the_genomic_accession=113390447, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68090, CurrentID=0, Status='live', Symbol='Yif1a', Aliases='5330422J23Rik, 54TM, Yif1, Yif1p', description='Yip1 interacting factor homolog A (S. cerevisiae)', other_designations='Yip1p-interacting factor|putative transmembrane protein; homolog of yeast Golgi membrane protein Yif1p (Yip1p-interacting factor)', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5088538, end_position_on_the_genomic_accession=5092879, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68092, CurrentID=0, Status='live', Symbol='Ncbp2', Aliases='20kDa, 5930413E18Rik, AA536802, AI843301, C79367', description='nuclear cap binding protein subunit 2', other_designations='20 kDa nuclear cap-binding protein|CBP20|NCBP 20 kDa subunit|nuclear cap binding protein subunit 2, 20kDa', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31948546, end_position_on_the_genomic_accession=31958472, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68094, CurrentID=0, Status='live', Symbol='Smarcc2', Aliases='5930405J04Rik', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily c, member 2', other_designations='BAF170|BRG1-associated factor 170|SWI/SNF complex 170 kDa subunit|SWI/SNF-related matrix-associated actin-dependent regulator of chromatin subfamily C member 2', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128459236, end_position_on_the_genomic_accession=128490583, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  68095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68095, CurrentID=0, Status='live', Symbol='Ociad1', Aliases='6030432N09Rik, AI481327, AW557942, Asrij, B230209J16Rik, BB021357, Emi2, Imi2, OCIA, TPA018', description='OCIA domain containing 1', other_designations='expressed during mesenchymal induction 2|induced during mesenchymal induction 2', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73292772, end_position_on_the_genomic_accession=73314077, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68097, CurrentID=0, Status='live', Symbol='Dynll2', Aliases='1700064A15Rik, 6720463E02Rik, C87222, DLC8, DLC8b, Dlc2', description='dynein light chain LC8-type 2', other_designations='8 kDa dynein light chain b', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87979525, end_position_on_the_genomic_accession=87998298, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68098, CurrentID=0, Status='live', Symbol='Rchy1', Aliases='ARNIP, CHIMP, Pirh2, Zfp363, Znf363', description='ring finger and CHY zinc finger domain containing 1', other_designations='CH-rich-interacting match with PLAG1|E3 ubiquitin-protein ligase Pirh2|androgen receptor N-terminal-interacting protein|zinc finger protein 363', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=91948842, end_position_on_the_genomic_accession=91963224, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  68099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68099, CurrentID=0, Status='live', Symbol='Fam92a', Aliases='6720467C03Rik1, Fam92a', description='family with sequence similarity 92, member A', other_designations='', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=12153722, end_position_on_the_genomic_accession=12172022, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  76886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76886, CurrentID=0, Status='live', Symbol='Fam81a', Aliases='6430514L14Rik', description='family with sequence similarity 81, member A', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70089167, end_position_on_the_genomic_accession=70142555, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  68106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68106, CurrentID=0, Status='live', Symbol='Nt5c3b', Aliases='2610037D24Rik, AI841843, C330027I04Rik, Nt5c3l', description="5'-nucleotidase, cytosolic IIIB", other_designations="5'-nucleotidase, cytosolic III-like|7-methylguanosine nucleotidase|N(7)-methylguanylate 5'-phosphatase|cN-III-like protein|cytosolic 5'-nucleotidase 3B|cytosolic 5'-nucleotidase III-like protein", map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100422321, end_position_on_the_genomic_accession=100441827, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68107, CurrentID=0, Status='live', Symbol='Cntd1', Aliases='1700051C09Rik, 4921513J16Rik, 5430417M23Rik, Cntd', description='cyclin N-terminal domain containing 1', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101279203, end_position_on_the_genomic_accession=101288701, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68108, CurrentID=0, Status='live', Symbol='Snhg17', Aliases='9430008C03Rik', description='small nucleolar RNA host gene 17', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158353700, end_position_on_the_genomic_accession=158361522, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68112, CurrentID=0, Status='live', Symbol='Sdccag3', Aliases='C330016H24Rik, C630038K21Rik', description='serologically defined colon cancer antigen 3', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26382798, end_position_on_the_genomic_accession=26389316, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68114, CurrentID=0, Status='live', Symbol='Mum1', Aliases='2610019J09, 9430059D04Rik, UBE-1C2', description='melanoma associated antigen (mutated) 1', other_designations='MUM-1|mutated melanoma-associated antigen 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80226434, end_position_on_the_genomic_accession=80245144, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  68115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68115, CurrentID=0, Status='live', Symbol='9430016H08Rik', Aliases='AA409999, C78988', description='RIKEN cDNA 9430016H08 gene', other_designations='', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=57406548, end_position_on_the_genomic_accession=57415958, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  330260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330260, CurrentID=0, Status='live', Symbol='Pon2', Aliases='6330405I24Rik, AI481612', description='paraoxonase 2', other_designations='A-esterase 2|PON 2|aromatic esterase 2|serum aryldialkylphosphatase 2', map_location='6 1.99 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=5264620, end_position_on_the_genomic_accession=5298373, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68117, CurrentID=0, Status='live', Symbol='Apool', Aliases='6720473G16Rik, 9430083G14Rik, E130106L15Rik', description='apolipoprotein O-like', other_designations='', map_location='X E2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=112311408, end_position_on_the_genomic_accession=112372755, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68118, CurrentID=0, Status='live', Symbol='Atg101', Aliases='9430023L20Rik, AA407441, AU022967', description='autophagy related 101', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101284301, end_position_on_the_genomic_accession=101290947, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68119, CurrentID=0, Status='live', Symbol='Cmtm3', Aliases='9430096L06Rik, AI413895, BNAS2, Cklfsf3', description='CKLF-like MARVEL transmembrane domain containing 3', other_designations='chemokine-like factor super family 3|chemokine-like factor superfamily member 3', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104339383, end_position_on_the_genomic_accession=104347672, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68121, CurrentID=0, Status='live', Symbol='Cep70', Aliases='6720484E09Rik, Bite, C030018L16Rik', description='centrosomal protein 70', other_designations='p10-binding protein', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99183595, end_position_on_the_genomic_accession=99302065, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  330267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330267, CurrentID=0, Status='live', Symbol='Thsd7a', Aliases='Gm837', description='thrombospondin, type I, domain containing 7A', other_designations='', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=12311608, end_position_on_the_genomic_accession=12749253, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  68126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68126, CurrentID=0, Status='live', Symbol='Fahd2a', Aliases='1500003K14Rik, B430104H02Rik, Cgi-105, Fahd2', description='fumarylacetoacetate hydrolase domain containing 2A', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127436215, end_position_on_the_genomic_accession=127444565, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  330277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330277, CurrentID=0, Status='live', Symbol='Fam71f1', Aliases='GARI-L1, Gm764, NYD-SP18', description='family with sequence similarity 71, member F1', other_designations='Golgi-associated Rab2B interactor-like 1|testes development-related NYD-SP18', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29313482, end_position_on_the_genomic_accession=29336022, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68134, CurrentID=0, Status='live', Symbol='Upf3b', Aliases='5730594O13Rik, AI317193, AW541158, RENT3B, UPF3X', description='UPF3 regulator of nonsense transcripts homolog B (yeast)', other_designations='hUPF3B', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37091678, end_position_on_the_genomic_accession=37110396, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  68135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68135, CurrentID=0, Status='live', Symbol='Eif3h', Aliases='1110008A16Rik, 40kD, 9430017H16Rik, EIF3-P40, EIF3-gamma, Eif3s3', description='eukaryotic translation initiation factor 3, subunit H', other_designations='eIF-3-gamma|eIF3 p40 subunit|eIF3h|eukaryotic translation initiation factor 3 subunit 3|eukaryotic translation initiation factor 3, subunit 3 (gamma)|eukaryotic translation initiation factor 3, subunit 3 (gamma, 40kD)', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=51786560, end_position_on_the_genomic_accession=51865515, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68137, CurrentID=0, Status='live', Symbol='Kdelr1', Aliases='8030486F04Rik, AW215843, Erd2.1', description='KDEL (Lys-Asp-Glu-Leu) endoplasmic reticulum protein retention receptor 1', other_designations='ER lumen protein retaining receptor 1|KDEL endoplasmic reticulum protein retention receptor 1|KDEL receptor 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45872840, end_position_on_the_genomic_accession=45883726, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68140, CurrentID=0, Status='live', Symbol='Tigd2', Aliases='3632410O17Rik', description='tigger transposable element derived 2', other_designations='tigger transposable element derived 2 homolog', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=59208870, end_position_on_the_genomic_accession=59212033, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68142, CurrentID=0, Status='live', Symbol='Ino80', Aliases='2310079N15Rik, 4632409L19Rik, Inoc1', description='INO80 complex subunit', other_designations='INO80 complex homolog 1|INO80 complex subunit A|INO80 homolog|putative DNA helicase INO80 complex homolog 1|yeast INO80-like protein', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119373042, end_position_on_the_genomic_accession=119477629, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  207965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207965, CurrentID=0, Status='live', Symbol='Vcpkmt', Aliases='Gm71, Mettl21d, Mettl221d, VCP-KMT', description='valosin containing protein lysine (K) methyltransferase', other_designations='VCP lysine methyltransferase|methyltransferase like 21D|methyltransferase-like protein 21D|valosin-containing protein lysine methyltransferase', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69576730, end_position_on_the_genomic_accession=69583109, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68145, CurrentID=0, Status='live', Symbol='Etaa1', Aliases='5730466H23Rik, AA672646', description='Ewing tumor-associated antigen 1', other_designations='', map_location='11 A3.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=17938749, end_position_on_the_genomic_accession=17953875, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68146, CurrentID=0, Status='live', Symbol='Arl13b', Aliases='A530097K21Rik, A930014M17Rik, Arl2l1, C530009C10Rik, hnn', description='ADP-ribosylation factor-like 13B', other_designations='ADP-ribosylation factor-like 2-like 1|ADP-ribosylation factor-like protein 2-like 1|ARL2-like protein 1|protein hennin', map_location='16 C1.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=62793678, end_position_on_the_genomic_accession=62847055, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68147, CurrentID=0, Status='live', Symbol='Gar1', Aliases='AA409823, AI326794, C430047J18Rik, Nola1', description='GAR1 ribonucleoprotein', other_designations='GAR1 ribonucleoprotein homolog|H/ACA small nucleolar RNPs|nucleolar protein family A member 1|nucleolar protein family A, member 1 (H/ACA small nucleolar RNPs)|snoRNP protein GAR1', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129824912, end_position_on_the_genomic_accession=129831396, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68149, CurrentID=0, Status='live', Symbol='Otub2', Aliases='2010015L18Rik, 4930586I02Rik, AI413508, AW557219, OTB2, OTU2', description='OTU domain, ubiquitin aldehyde binding 2', other_designations='OTU domain-containing ubiquitin aldehyde-binding protein 2|deubiquitinating enzyme OTUB2|otubain-2|ubiquitin-specific-processing protease OTUB2', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103376894, end_position_on_the_genomic_accession=103406351, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68151, CurrentID=0, Status='live', Symbol='Wls', Aliases='5031439A09Rik, AI173978, AI987742, Gpr177', description='wntless homolog (Drosophila)', other_designations='EVI|G protein-coupled receptor 177|integral membrane protein GPR177|protein evenness interrupted homolog', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=159839677, end_position_on_the_genomic_accession=159938670, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  68152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68152, CurrentID=0, Status='live', Symbol='Fam133b', Aliases='2900022K02Rik, 5830415L20Rik', description='family with sequence similarity 133, member B', other_designations='', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3543801, end_position_on_the_genomic_accession=3570548, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68153, CurrentID=0, Status='live', Symbol='Gtf2e2', Aliases='34kDa, AI462509, C330006J08Rik', description='general transcription factor II E, polypeptide 2 (beta subunit)', other_designations='TFIIE-beta|general transcription factor II E, polypeptide 2 (beta subunit, 34kD)|general transcription factor IIE, polypeptide 2 (beta subunit, 34kD)|transcription initiation factor IIE subunit beta', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33731914, end_position_on_the_genomic_accession=33777173, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  330301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330301, CurrentID=0, Status='live', Symbol='Zfp786', Aliases='A730012O14Rik, Znf786', description='zinc finger protein 786', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47819263, end_position_on_the_genomic_accession=47835920, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68159, CurrentID=0, Status='live', Symbol='Stx19', Aliases='A030009B12Rik', description='syntaxin 19', other_designations='syntaxin 9', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=62814676, end_position_on_the_genomic_accession=62822722, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68165, CurrentID=0, Status='live', Symbol='Fdx1l', Aliases='B230118G17Rik', description='ferredoxin 1-like', other_designations='ferredoxin-1-like protein', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21067520, end_position_on_the_genomic_accession=21073514, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68166, CurrentID=0, Status='live', Symbol='Spire1', Aliases='6030430B19Rik, AI415299, AU022898, AW550622, Spir-1', description='spire homolog 1 (Drosophila)', other_designations='spire1', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67488209, end_position_on_the_genomic_accession=67610902, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  68169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68169, CurrentID=0, Status='live', Symbol='Ndnf', Aliases='A930038C07Rik, epidermacan', description='neuron-derived neurotrophic factor', other_designations='fibronectin type-III domain-containing protein C4orf31 homolog', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=65671611, end_position_on_the_genomic_accession=65712329, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68170, CurrentID=0, Status='live', Symbol='B230118H07Rik', Aliases='NWC', description='RIKEN cDNA B230118H07 gene', other_designations='Rag1/Nwc fusion protein|Rag1/Rag2/Nwc fusion protein', map_location='2 53.83 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=101560781, end_position_on_the_genomic_accession=101629105, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68171, CurrentID=0, Status='live', Symbol='D730048I06Rik', Aliases='Pate-C, mANLP2', description='RIKEN cDNA D730048I06 gene', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35788050, end_position_on_the_genomic_accession=35790112, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68172, CurrentID=0, Status='live', Symbol='Rpl39l', Aliases='4930517K11Rik', description='ribosomal protein L39-like', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10170192, end_position_on_the_genomic_accession=10174911, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  330319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330319, CurrentID=0, Status='live', Symbol='Wipf3', Aliases='9030621G03, BQ952480', description='WAS/WASL interacting protein family, member 3', other_designations='corticosteroids and regional expression protein 16 homolog', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54429218, end_position_on_the_genomic_accession=54503768, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68176, CurrentID=0, Status='live', Symbol='Fam212a', Aliases='6230427J02Rik, Inka1', description='family with sequence similarity 212, member A', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107984223, end_position_on_the_genomic_accession=107985916, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68177, CurrentID=0, Status='live', Symbol='Ebpl', Aliases='5730442K12Rik, EBRP', description='emopamil binding protein-like', other_designations='emopamil binding related protein, delta8-delta7 sterol isomerase related protein|emopamil-binding-related protein', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61339763, end_position_on_the_genomic_accession=61360445, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68178, CurrentID=0, Status='live', Symbol='Cgnl1', Aliases='4933421H10Rik, 9930020M10Rik, AI503810, Jacop', description='cingulin-like 1', other_designations='junction-associated coiled-coil protein|paracingulin', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=71626507, end_position_on_the_genomic_accession=71809300, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  330323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330323, CurrentID=0, Status='live', Symbol='Fam188b', Aliases='C330043M08Rik, Gm1406', description='family with sequence similarity 188, member B', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55203347, end_position_on_the_genomic_accession=55320222, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  68180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68180, CurrentID=0, Status='live', Symbol='Hyi', Aliases='2700033B16Rik, 6430559E15Rik, HT036', description='hydroxypyruvate isomerase (putative)', other_designations='hydroxypyruvate isomerase homolog', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118359999, end_position_on_the_genomic_accession=118362744, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68183, CurrentID=0, Status='live', Symbol='Bcas2', Aliases='6430539P16Rik, AI132645, C76366, C80030', description='breast carcinoma amplified sequence 2', other_designations='DNA amplified in mammary carcinoma 1 protein|breast carcinoma-amplified sequence 2 homolog', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103171660, end_position_on_the_genomic_accession=103179163, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68184, CurrentID=0, Status='live', Symbol='Denr', Aliases='1500003K04Rik', description='density-regulated protein', other_designations='DRP', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123907269, end_position_on_the_genomic_accession=123928832, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68185, CurrentID=0, Status='live', Symbol='Coa4', Aliases='5330414O08Rik, AI115477, AW047259, Chchd8, E2ig2', description='cytochrome c oxidase assembly factor 4', other_designations='coiled-coil-helix-coiled-coil-helix domain containing 8|coiled-coil-helix-coiled-coil-helix domain-containing protein 8', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100537100, end_position_on_the_genomic_accession=100539812, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68187, CurrentID=0, Status='live', Symbol='Fam135a', Aliases='4921533L14Rik, AI256732, AI835491', description='family with sequence similarity 135, member A', other_designations='', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=24010758, end_position_on_the_genomic_accession=24100460, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  68188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68188, CurrentID=0, Status='live', Symbol='Sympk', Aliases='1500016F02Rik, 4632415H16Rik, AA125406, AI449890, SPK, SYM', description='symplekin', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19024352, end_position_on_the_genomic_accession=19054623, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  68190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68190, CurrentID=0, Status='live', Symbol='Dubr', Aliases='5330426P16Rik, Dum', description='Dppa2 upstream binding RNA', other_designations='developmental pluripotency-associated 2 upstream binding muscle lncRNA', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=50726751, end_position_on_the_genomic_accession=50732773, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68192, CurrentID=0, Status='live', Symbol='Leprotl1', Aliases='1110067H13Rik, 1520402O14Rik, AI854296', description='leptin receptor overlapping transcript-like 1', other_designations='endospanin-2', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=34135572, end_position_on_the_genomic_accession=34146739, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68193, CurrentID=0, Status='live', Symbol='Rpl24', Aliases='0610008L05Rik, Bst', description='ribosomal protein L24', other_designations='belly spot and tail', map_location='16 C1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=55966275, end_position_on_the_genomic_accession=55971437, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68194, CurrentID=0, Status='live', Symbol='Ndufb4', Aliases='0610006N12Rik, 1300010H20Rik', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex 4', other_designations='CI-B15|NADH-ubiquinone oxidoreductase B15 subunit|NADH:ubiquinone oxidoreductase B15 subunit|complex I-B15', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37647602, end_position_on_the_genomic_accession=37654368, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68195, CurrentID=0, Status='live', Symbol='Rnaset2b', Aliases='0610007O07Rik, 4833423A10Rik, 4930532K22Rik, C79750, RNASE6PL, Rnaset2', description='ribonuclease T2B', other_designations='ribonuclease T2', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6978860, end_position_on_the_genomic_accession=6998193, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68196, CurrentID=0, Status='live', Symbol='Hsbp1', Aliases='0610007A03Rik, AA408769, Hsp25', description='heat shock factor binding protein 1', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119344538, end_position_on_the_genomic_accession=119348929, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68197, CurrentID=0, Status='live', Symbol='Ndufc2', Aliases='1810004I06Rik, 2010300P09Rik, G1, R75094', description='NADH dehydrogenase (ubiquinone) 1, subcomplex unknown, 2', other_designations='14.5b subunit of the NADH-ubiquinone dehydrogenase complex|CI-B14.5b|NADH-ubiquinone oxidoreductase subunit B14.5b|complex I-B14.5b', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97400003, end_position_on_the_genomic_accession=97407800, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68198, CurrentID=0, Status='live', Symbol='Ndufb2', Aliases='1810011O01Rik, 8kDa, AGGG, AI325567, CI-AGGG', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 2', other_designations='NADH-ubiquinone oxidoreductase AGGG subunit|complex I-AGGG', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39592551, end_position_on_the_genomic_accession=39599472, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68201, CurrentID=0, Status='live', Symbol='Ccdc34', Aliases='2810027O19Rik, 4930522P10Rik, AI596202', description='coiled-coil domain containing 34', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=110017510, end_position_on_the_genomic_accession=110173355, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68202, CurrentID=0, Status='live', Symbol='Ndufa5', Aliases='2900002J19Rik', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 5', other_designations='CI-13kD-B|NADH-ubiquinone oxidoreductase 13 kDa-B subunit|complex I subunit B13|complex I-13kD-B', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24518662, end_position_on_the_genomic_accession=24527687, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68203, CurrentID=0, Status='live', Symbol='Diras2', Aliases='2900052J15Rik, AI414999', description='DIRAS family, GTP-binding RAS-like 2', other_designations='distinct subgroup of the Ras family member 2', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=52504375, end_position_on_the_genomic_accession=52531242, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68205, CurrentID=0, Status='live', Symbol='Urm1', Aliases='2900073H19Rik', description='ubiquitin related modifier 1 homolog (S. cerevisiae)', other_designations='ubiquitin-related modifier 1 homolog', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29827389, end_position_on_the_genomic_accession=29844996, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68209, CurrentID=0, Status='live', Symbol='Rnaseh2c', Aliases='1500026D16Rik, AYP1', description='ribonuclease H2, subunit C', other_designations='RNase H2 subunit C|ribonuclease HI subunit C', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5601873, end_position_on_the_genomic_accession=5602959, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  330355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330355, CurrentID=0, Status='live', Symbol='Dnah6', Aliases='9830168K20, A730004I20Rik, Dnahc6, mKIAA1697, mdhc6', description='dynein, axonemal, heavy chain 6', other_designations='axonemal dynein heavy chain|dynein, axon, heavy chain 6', map_location='6 32.37 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=73017606, end_position_on_the_genomic_accession=73222400, orientation='minus', exon_count=79, OMIM='', no_hdr0=''),
  68212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68212, CurrentID=0, Status='live', Symbol='Tmbim4', Aliases='0610007H07Rik, AU022431', description='transmembrane BAX inhibitor motif containing 4', other_designations='Z-protein|transmembrane BAX inhibitor motif-containing protein 4', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=120208826, end_position_on_the_genomic_accession=120224897, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68214, CurrentID=0, Status='live', Symbol='Gsto2', Aliases='1700020F09Rik, 4930425C18Rik, GSTO 2-2', description='glutathione S-transferase omega 2', other_designations='GSTO-2|MMA(V) reductase|glutathione S-transferase omega 2-2|glutathione-dependent dehydroascorbate reductase|monomethylarsonic acid reductase', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47865545, end_position_on_the_genomic_accession=47886308, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68215, CurrentID=0, Status='live', Symbol='Fam98b', Aliases='2610510H03Rik', description='family with sequence similarity 98, member B', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=117249711, end_position_on_the_genomic_accession=117271540, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  330361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330361, CurrentID=0, Status='live', Symbol='Gcfc2', Aliases='A130099G21, AW146020, GCF2, Tcf9', description='GC-rich sequence DNA binding factor 2', other_designations='GCF|TCF-9|transcription factor 9', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=81923636, end_position_on_the_genomic_accession=81959098, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  68219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68219, CurrentID=0, Status='live', Symbol='Nudt21', Aliases='25kDa, 3110048P04Rik, 5730530J16Rik, AU014860, AW549947, Cpsf5', description='nudix (nucleoside diphosphate linked moiety X)-type motif 21', other_designations='cleavage and polyadenylation specific factor 5, 25 kD subunit|nucleoside diphosphate-linked moiety X motif 21|nudix motif 21', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94019403, end_position_on_the_genomic_accession=94037039, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68221, CurrentID=0, Status='live', Symbol='Wfdc15a', Aliases='1700049M11Rik, Wfdcl1', description='WAP four-disulfide core domain 15A', other_designations='wfdc15a', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164198872, end_position_on_the_genomic_accession=164200117, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68222, CurrentID=0, Status='live', Symbol='Fam166a', Aliases='4931415M17', description='family with sequence similarity 166, member A', other_designations='4931415M17 mRNA', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25214724, end_position_on_the_genomic_accession=25222281, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68223, CurrentID=0, Status='live', Symbol='Fam24a', Aliases='1700063I17Rik', description='family with sequence similarity 24, member A', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131334622, end_position_on_the_genomic_accession=131336716, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68224, CurrentID=0, Status='live', Symbol='1700067P10Rik', Aliases='', description='RIKEN cDNA 1700067P10 gene', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48089318, end_position_on_the_genomic_accession=48090920, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  330369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330369, CurrentID=0, Status='live', Symbol='Fbxo41', Aliases='9630017H13, D6Ertd538e', description='F-box protein 41', other_designations='', map_location='6 37.48 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85469574, end_position_on_the_genomic_accession=85503028, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68226, CurrentID=0, Status='live', Symbol='Efcab2', Aliases='1700073K01Rik, D830011E08Rik', description='EF-hand calcium binding domain 2', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=178405881, end_position_on_the_genomic_accession=178483250, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68229, CurrentID=0, Status='live', Symbol='AI846148', Aliases='1700105P06Rik', description='expressed sequence AI846148', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7356463, end_position_on_the_genomic_accession=7383026, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68231, CurrentID=0, Status='live', Symbol='H2afb1', Aliases='1700113O17Rik, H2A.Bbd, H2al2, TSEG-1, Tseg1', description='H2A histone family, member B1', other_designations='H2A Barr body-deficient|histone variant H2al2|testis-specific expressed gene 1 protein|testis-specific expressed protein 1', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=17996422, end_position_on_the_genomic_accession=17996946, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68233, CurrentID=0, Status='live', Symbol='Fam229a', Aliases='1700125D06Rik', description='family with sequence similarity 229, member A', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129491190, end_position_on_the_genomic_accession=129491956, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68235, CurrentID=0, Status='live', Symbol='Mturn', Aliases='2410066E13Rik, AI448550, B230212L03Rik', description='maturin, neural progenitor differentiation regulator homolog (Xenopus)', other_designations='UPF0452 protein C7orf41 homolog|maturin neural progenitor differentiation regulator protein homolog', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54681624, end_position_on_the_genomic_accession=54703855, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68236, CurrentID=0, Status='live', Symbol='Gtsf1l', Aliases='2410116G06Rik', description='gametocyte specific factor 1-like', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163087031, end_position_on_the_genomic_accession=163089601, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68239, CurrentID=0, Status='live', Symbol='Krt42', Aliases='2410039E07Rik, K17n, Ka22, ecat6', description='keratin 42', other_designations='CK-42|K42|cytokeratin-42|keratin 17n|keratin-17n|keratin-42|krt42|predicted ortholog of type I keratin KA22|type I keratin KA22', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100262882, end_position_on_the_genomic_accession=100272565, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68240, CurrentID=0, Status='live', Symbol='Rpa3', Aliases='14kDa, C330026P08Rik', description='replication protein A3', other_designations='RF-A protein 3|RP-A p14|replication factor A protein 3', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=8255936, end_position_on_the_genomic_accession=8259141, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68241, CurrentID=0, Status='live', Symbol='Fam195a', Aliases='9530058B02Rik', description='family with sequence similarity 195, member A', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25863698, end_position_on_the_genomic_accession=25868786, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68243, CurrentID=0, Status='live', Symbol='A930018P22Rik', Aliases='', description='RIKEN cDNA A930018P22 gene', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104122769, end_position_on_the_genomic_accession=104124746, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68246, CurrentID=0, Status='live', Symbol='1700112J05Rik', Aliases='1700110A09Rik', description='RIKEN cDNA 1700112J05 gene', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77006442, end_position_on_the_genomic_accession=77018863, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68250, CurrentID=0, Status='live', Symbol='Fam96a', Aliases='5730536A07Rik', description='family with sequence similarity 96, member A', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66126611, end_position_on_the_genomic_accession=66138968, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68251, CurrentID=0, Status='live', Symbol='Babam1', Aliases='5430437P03Rik, Merit40', description='BRISC and BRCA1 A complex member 1', other_designations='BRCA1-A complex subunit MERIT40|mediator of RAP80 interactions and targeting subunit of 40 kDa|new component of the BRCA1-A complex', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71396844, end_position_on_the_genomic_accession=71404772, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68255, CurrentID=0, Status='live', Symbol='Tmem86b', Aliases='C330014O21Rik', description='transmembrane protein 86B', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4628040, end_position_on_the_genomic_accession=4631395, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  330401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330401, CurrentID=0, Status='live', Symbol='Tmcc1', Aliases='3632431M01Rik, mKIAA0779', description='transmembrane and coiled coil domains 1', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116018611, end_position_on_the_genomic_accession=116193379, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  68259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68259, CurrentID=0, Status='live', Symbol='Ift80', Aliases='4921524P20Rik, Wdr56, mKIAA1374', description='intraflagellar transport 80', other_designations='WD repeat domain 56|WD repeat-containing protein 56|intraflagellar transport 80 homolog', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=68892499, end_position_on_the_genomic_accession=69004653, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  68260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68260, CurrentID=0, Status='live', Symbol='Trmt12', Aliases='4632406N01Rik, AI449275, Tyw2', description='tRNA methyltranferase 12', other_designations='alpha-amino-alpha-carboxypropyl transferase TYW2|tRNA methyltranferase 12 homolog|tRNA methyltransferase 12|tRNA(Phe) (4-demethylwyosine(37)-C(7)) aminocarboxypropyltransferase|tRNA-yW-synthesizing protein 2', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58872649, end_position_on_the_genomic_accession=58876781, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  623046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623046, CurrentID=0, Status='live', Symbol='Fscb', Aliases='EG623046', description='fibrous sheath CABYR binding protein', other_designations='PKA phosphorylated calcium and CABYR-binding protein|PKA-phosphorylated calcium and CABYR-binding protein|fscb', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=64471333, end_position_on_the_genomic_accession=64474898, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68262, CurrentID=0, Status='live', Symbol='Agpat4', Aliases='1500003P24Rik', description='1-acylglycerol-3-phosphate O-acyltransferase 4 (lysophosphatidic acid acyltransferase, delta)', other_designations='1-AGP acyltransferase 4|1-AGPAT 4|1-acylglycerol-3-phosphate O-acyltransferase 1 (lysophosphatidic acid acyltransferase, delta)|LPAAT-delta|lysophosphatidic acid acyltransferase delta|lysophosphatidic acid acyltransferase-delta', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12118709, end_position_on_the_genomic_accession=12219649, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68263, CurrentID=0, Status='live', Symbol='Pdhb', Aliases='2610103L06Rik, AL024199, C81408', description='pyruvate dehydrogenase (lipoamide) beta', other_designations='PDHE1-B', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8165991, end_position_on_the_genomic_accession=8172992, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  330409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330409, CurrentID=0, Status='live', Symbol='Cecr2', Aliases='2610101O16Rik, 2810409N01Rik, Gtl4, mKIAA1740', description='cat eye syndrome chromosome region, candidate 2', other_designations='cat eye syndrome chromosome region, candidate 2 homolog|gene trap locus 4', map_location='6 56.97 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120666320, end_position_on_the_genomic_accession=120771194, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  68267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68267, CurrentID=0, Status='live', Symbol='Slc25a22', Aliases='1300006L01Rik, AI060884, Gc1', description='solute carrier family 25 (mitochondrial carrier, glutamate), member 22', other_designations='glutamate/H(+) symporter 1|solute carrier family 25 member 22', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141429749, end_position_on_the_genomic_accession=141437887, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  68268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68268, CurrentID=0, Status='live', Symbol='Zdhhc21', Aliases='9130404H11Rik, AL024349, D130004H04Rik, dep', description='zinc finger, DHHC domain containing 21', other_designations='DHHC-21|GABA-A receptor-associated membrane protein 3|zinc finger DHHC domain-containing protein 21', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=82798738, end_position_on_the_genomic_accession=82859891, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  68270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68270, CurrentID=0, Status='live', Symbol='Dnaaf1', Aliases='4930457P18Rik, Lrrc50', description='dynein, axonemal assembly factor 1', other_designations='leucine rich repeat containing 50|leucine-rich repeat-containing protein 50', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119575235, end_position_on_the_genomic_accession=119598454, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68271, CurrentID=0, Status='live', Symbol='Zfp85os', Aliases='4930441O14Rik', description='zinc finger protein 85, opposite strand', other_designations='Zfp85 opposite strand transcript', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67729262, end_position_on_the_genomic_accession=67756838, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68272, CurrentID=0, Status='live', Symbol='Rbm28', Aliases='2810480G15Rik, AI503051', description='RNA binding motif protein 28', other_designations='RNA-binding motif protein 28', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29123573, end_position_on_the_genomic_accession=29164724, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  68273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68273, CurrentID=0, Status='live', Symbol='Pomgnt1', Aliases='0610016I07Rik, 4930467B06Rik', description='protein O-linked mannose beta 1,2-N-acetylglucosaminyltransferase', other_designations='O-mannosyl N-acetylglucosaminyltransferase', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116150498, end_position_on_the_genomic_accession=116159844, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  68274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68274, CurrentID=0, Status='live', Symbol='Toporsl', Aliases='4930547C10Rik', description='topoisomerase I binding, arginine/serine-rich like', other_designations='', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=52596274, end_position_on_the_genomic_accession=52612424, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68275, CurrentID=0, Status='live', Symbol='Rpa1', Aliases='5031405K23Rik, 70kDa, AA589576, AW557552, RF-A, RP-A, Rpa', description='replication protein A1', other_designations='RF-A protein 1|RP-A p70|replication factor A protein 1', map_location='11 45.79 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75300259, end_position_on_the_genomic_accession=75348383, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  68276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68276, CurrentID=0, Status='live', Symbol='Toe1', Aliases='4930584N22Rik, 4933424D16Rik, AI413517', description='target of EGR1, member 1 (nuclear)', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116804003, end_position_on_the_genomic_accession=116807635, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68277, CurrentID=0, Status='live', Symbol='2310057M21Rik', Aliases='3110040E10Rik, AI429544', description='RIKEN cDNA 2310057M21 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131342710, end_position_on_the_genomic_accession=131362843, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68278, CurrentID=0, Status='live', Symbol='Ddx39', Aliases='2610307C23Rik, BAT1, DDXLa, URH49, Ddx39', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 39', other_designations='ATP-dependent RNA helicase DDX39|DEAD box protein 39|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 39|UAP56-related helicase, 49 kDa', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83715177, end_position_on_the_genomic_accession=83723351, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  68279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68279, CurrentID=0, Status='live', Symbol='Mcoln2', Aliases='3300002C04Rik, AI549968, C86638, TRPML2', description='mucolipin 2', other_designations='mucolipidin 2', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146149796, end_position_on_the_genomic_accession=146195514, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  68281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68281, CurrentID=0, Status='live', Symbol='4930430F08Rik', Aliases='4930571E09Rik, A730088G13Rik, AW551239, C430008C19Rik', description='RIKEN cDNA 4930430F08 gene', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=100572274, end_position_on_the_genomic_accession=100589390, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68283, CurrentID=0, Status='live', Symbol='9530077C05Rik', Aliases='1110003N12Rik, Kiaa0895, mKIAA0895', description='RIKEN cDNA 9530077C05 gene', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22387732, end_position_on_the_genomic_accession=22444684, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  330428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330428, CurrentID=0, Status='live', Symbol='Tmem52b', Aliases='D630042F21Rik', description='transmembrane protein 52B', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129512546, end_position_on_the_genomic_accession=129519227, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68291, CurrentID=0, Status='live', Symbol='Mto1', Aliases='2310039H01Rik, 5730419A02Rik', description='mitochondrial translation optimization 1 homolog (S. cerevisiae)', other_designations='', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78448015, end_position_on_the_genomic_accession=78474152, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68292, CurrentID=0, Status='live', Symbol='Stt3b', Aliases='1300006C19Rik, Simp', description='STT3, subunit of the oligosaccharyltransferase complex, homolog B (S. cerevisiae)', other_designations='B6dom1 antigen|STT3-B|oligosaccharyl transferase subunit STT3B|source of immunodominant MHC-associated peptides', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=115242581, end_position_on_the_genomic_accession=115310421, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  68294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68294, CurrentID=0, Status='live', Symbol='Mfsd10', Aliases='0610009O03Rik, Tetran', description='major facilitator superfamily domain containing 10', other_designations='tetracycline transporter-like protein', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34633644, end_position_on_the_genomic_accession=34637215, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  68295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68295, CurrentID=0, Status='live', Symbol='Aar2', Aliases='0610011L14Rik', description='AAR2 splicing factor homolog', other_designations='novel transcript', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156547558, end_position_on_the_genomic_accession=156568972, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  330440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330440, CurrentID=0, Status='live', Symbol='Gm766', Aliases='', description='predicted gene 766', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142785781, end_position_on_the_genomic_accession=142804466, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68298, CurrentID=0, Status='live', Symbol='Ncapd2', Aliases='2810406C15Rik, 2810465G24Rik, CAP-D2, CNAP1, mKIAA0159', description='non-SMC condensin I complex, subunit D2', other_designations='XCAP-D2 homolog|chromosome condensation-related SMC-associated protein 1|chromosome-associated protein D2|mCAP-D2', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125168007, end_position_on_the_genomic_accession=125191668, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  68299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68299, CurrentID=0, Status='live', Symbol='Vps53', Aliases='2010002A08Rik, 2310040I21Rik, 3100002B05Rik, Hccs1', description='vacuolar protein sorting 53 (yeast)', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76046226, end_position_on_the_genomic_accession=76179630, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  68303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68303, CurrentID=0, Status='live', Symbol='Fam114a1', Aliases='1190001N04Rik, 9130005N14Rik, Noxp20', description='family with sequence similarity 114, member A1', other_designations='nervous system over-expressed protein 20|nervous system overexpressed protein 20', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=64970075, end_position_on_the_genomic_accession=65041901, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  68304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68304, CurrentID=0, Status='live', Symbol='Kdelc2', Aliases='2010004J24Rik, 4833410J10Rik, AW549401', description='KDEL (Lys-Asp-Glu-Leu) containing 2', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53383826, end_position_on_the_genomic_accession=53401867, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  330450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330450, CurrentID=0, Status='live', Symbol='Far2', Aliases='A230046P18, AW048109, BC055759, Mlstd1', description='fatty acyl CoA reductase 2', other_designations='male sterility domain-containing protein 1', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148047003, end_position_on_the_genomic_accession=148182760, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  68307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68307, CurrentID=0, Status='live', Symbol='Lrriq4', Aliases='4930558O21Rik', description='leucine-rich repeats and IQ motif containing 4', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30644506, end_position_on_the_genomic_accession=30672431, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68310, CurrentID=0, Status='live', Symbol='Zmym1', Aliases='5830412B09Rik, AW209050', description='zinc finger, MYM domain containing 1', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127047094, end_position_on_the_genomic_accession=127061171, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68311, CurrentID=0, Status='live', Symbol='Lypd2', Aliases='0610005K03Rik, Lypdc2, VLL', description='Ly6/Plaur domain containing 2', other_designations='H8C4 protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74732246, end_position_on_the_genomic_accession=74734795, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68312, CurrentID=0, Status='live', Symbol='Gstm7', Aliases='0610005A07Rik, AI894236, Cd203c', description='glutathione S-transferase, mu 7', other_designations='GST class-mu 7', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107926332, end_position_on_the_genomic_accession=107931873, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  330460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330460, CurrentID=0, Status='live', Symbol='Tmem150b', Aliases='A630041N19, Tmem224', description='transmembrane protein 150B', other_designations='transmembrane protein 224', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4706832, end_position_on_the_genomic_accession=4725251, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68318, CurrentID=0, Status='live', Symbol='Aph1c', Aliases='0610008A10Rik, APH-1b, Aph1b', description='aph1 homolog C, gamma secretase subunit', other_designations='anterior pharynx defective 1b|anterior pharynx defective 1c homolog', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66814994, end_position_on_the_genomic_accession=66834779, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  330463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330463, CurrentID=0, Status='live', Symbol='Zfp78', Aliases='A330079D15, KRAB12, Zfp77', description='zinc finger protein 78', other_designations='', map_location='7 3.69 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6362640, end_position_on_the_genomic_accession=6382606, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68323, CurrentID=0, Status='live', Symbol='Nudt22', Aliases='0610006K04Rik, AI851793, AW545547', description='nudix (nucleoside diphosphate linked moiety X)-type motif 22', other_designations='nudix motif 22', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6993019, end_position_on_the_genomic_accession=6996115, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  330470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330470, CurrentID=0, Status='live', Symbol='Bsph1', Aliases='Gm767', description='binder of sperm protein homolog 1', other_designations='bovine seminal plasma protein homolog 1|bovine seminal plasma protein-like 1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13450841, end_position_on_the_genomic_accession=13473450, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68327, CurrentID=0, Status='live', Symbol='Tsr3', Aliases='0610007P22Rik, 1110014B11Rik, AL022759', description='TSR3 20S rRNA accumulation', other_designations='probable ribosome biogenesis protein C16orf42 homolog', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25240170, end_position_on_the_genomic_accession=25242799, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68328, CurrentID=0, Status='live', Symbol='Rab13', Aliases='0610007N03Rik, B230212B15Rik', description='RAB13, member RAS oncogene family', other_designations='', map_location='3 39.21 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90213715, end_position_on_the_genomic_accession=90226387, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  330474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330474, CurrentID=0, Status='live', Symbol='Zc3h4', Aliases='Bwq1, Gm768, Kiaa1064-hp, mKIAA1064', description='zinc finger CCCH-type containing 4', other_designations='', map_location='7 9.03 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16399572, end_position_on_the_genomic_accession=16437698, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  68332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68332, CurrentID=0, Status='live', Symbol='Sdhaf1', Aliases='0610010E21Rik, AI430885, AW490662, Lyrm8', description='succinate dehydrogenase complex assembly factor 1', other_designations='LYR motif-containing protein 8|SDH assembly factor 1', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30321409, end_position_on_the_genomic_accession=30322375, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68337, CurrentID=0, Status='live', Symbol='Crip2', Aliases='0610010I23Rik, AW743261, C77570, CRP2, CRP4, Crp, ESP1, Hlp', description='cysteine rich protein 2', other_designations='CRP-2|LIM only protein HLP|heart LIM protein', map_location='12 61.57 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113140236, end_position_on_the_genomic_accession=113145506, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68338, CurrentID=0, Status='live', Symbol='Golt1a', Aliases='0610012C01Rik, Kiss1, metastatin', description='golgi transport 1A', other_designations='golt1a', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133309823, end_position_on_the_genomic_accession=133323026, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  330483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330483, CurrentID=0, Status='live', Symbol='Ceacam16', Aliases='Gm769', description='carcinoembryonic antigen-related cell adhesion molecule 16', other_designations='CEA-related cell adhesion molecule 16', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19852097, end_position_on_the_genomic_accession=19871449, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  330485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330485, CurrentID=0, Status='live', Symbol='Tmem145', Aliases='B930076A02Rik', description='transmembrane protein 145', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25305995, end_position_on_the_genomic_accession=25316199, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  68342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68342, CurrentID=0, Status='live', Symbol='Ndufb10', Aliases='0610011B04Rik, 22kDa, PDSW', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 10', other_designations='CI-PDSW|NADH-ubiquinone oxidoreductase PDSW subunit|complex I-PDSW', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24722067, end_position_on_the_genomic_accession=24724388, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68344, CurrentID=0, Status='live', Symbol='Tmem174', Aliases='0610009B10Rik, AA986094', description='transmembrane protein 174', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98634978, end_position_on_the_genomic_accession=98637410, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68346, CurrentID=0, Status='live', Symbol='Sirt5', Aliases='0610012J09Rik, 1500032M05Rik, AV001953', description='sirtuin 5', other_designations='NAD-dependent deacetylase sirtuin-5|NAD-dependent lysine demalonylase and desuccinylase sirtuin-5, mitochondrial', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43365439, end_position_on_the_genomic_accession=43395203, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  68347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68347, CurrentID=0, Status='live', Symbol='0610011F06Rik', Aliases='', description='RIKEN cDNA 0610011F06 gene', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25875500, end_position_on_the_genomic_accession=25877163, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68348, CurrentID=0, Status='live', Symbol='Serpina1f', Aliases='0610012A11Rik, AI327419, Serpina2, epserin', description='serine (or cysteine) peptidase inhibitor, clade A, member 1F', other_designations='alpha-1 protease inhibitor 6|alpha-1-antiproteinase|epididymal serine protease inhibitor|epsin|serine protease inhibitor 1-6|serine protease inhibitor A1f|serpin A1f|serpina1f', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103688044, end_position_on_the_genomic_accession=103695529, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68349, CurrentID=0, Status='live', Symbol='Ndufs3', Aliases='0610010M09Rik, CI-30kD', description='NADH dehydrogenase (ubiquinone) Fe-S protein 3', other_designations='NADH-ubiquinone oxidoreductase 30 kDa subunit|complex I-30kD', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90894636, end_position_on_the_genomic_accession=90904721, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68350, CurrentID=0, Status='live', Symbol='Mul1', Aliases='0610009K11Rik, AV000801, Gide', description='mitochondrial ubiquitin ligase activator of NFKB 1', other_designations='E3 ubiquitin ligase|E3 ubiquitin-protein ligase MUL1|growth inhibition and death E3 ligase', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138434661, end_position_on_the_genomic_accession=138442265, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  330496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330496, CurrentID=0, Status='live', Symbol='Ifnl2', Aliases='EG330496, IL-28A, Il28a', description='interferon lambda 2', other_designations='IFN-lambda-2|IFN-lambda2|IL-28A|interferon-lambda2|interleukin 28A|interleukin-28A', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28508784, end_position_on_the_genomic_accession=28510532, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68355, CurrentID=0, Status='live', Symbol='2010204K13Rik', Aliases='0610010M13Rik, Cip7', description='RIKEN cDNA 2010204K13 gene', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7411817, end_position_on_the_genomic_accession=7422988, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  330502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330502, CurrentID=0, Status='live', Symbol='Zfp82', Aliases='A030010D05, Krab16, Mszf3, Zfp-82', description='zinc finger protein 82', other_designations='zinc finger protein 545', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30056034, end_position_on_the_genomic_accession=30072823, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68364, CurrentID=0, Status='live', Symbol='0610030E20Rik', Aliases='1500031N04Rik, 2810411C16Rik, AI461894', description='RIKEN cDNA 0610030E20 gene', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72347023, end_position_on_the_genomic_accession=72353162, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68365, CurrentID=0, Status='live', Symbol='Rab14', Aliases='0610030G24Rik, 2810475J17Rik, A830021G03Rik, AI314285, AI649155, D030017L14Rik', description='RAB14, member RAS oncogene family', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35180205, end_position_on_the_genomic_accession=35201431, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68366, CurrentID=0, Status='live', Symbol='Tmem129', Aliases='0610039G24Rik, 2410015J15Rik, AA409575', description='transmembrane protein 129', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33653216, end_position_on_the_genomic_accession=33657911, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  330513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330513, CurrentID=0, Status='live', Symbol='Gm5114', Aliases='4921507A12, EG330513', description='predicted gene 5114', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=39407294, end_position_on_the_genomic_accession=39413268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68371, CurrentID=0, Status='live', Symbol='Pbld1', Aliases='0610038K03Rik, Mawbp, Pbld', description='phenazine biosynthesis-like protein domain containing 1', other_designations='MAWD binding protein homolog 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63060626, end_position_on_the_genomic_accession=63077964, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68375, CurrentID=0, Status='live', Symbol='Ndufa8', Aliases='0610033L03Rik, AW261656', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, 8', other_designations='CI-19kD|CI-PGIV|NADH-ubiquinone oxidoreductase 19 kDa subunit|complex I-19kD|complex I-PGIV', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36036334, end_position_on_the_genomic_accession=36049292, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68379, CurrentID=0, Status='live', Symbol='Ciz1', Aliases='0610038H21Rik, 2900056O04Rik, LSFR1, ZNF356', description='CDKN1A interacting zinc finger protein 1', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32363006, end_position_on_the_genomic_accession=32378313, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  68385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68385, CurrentID=0, Status='live', Symbol='Tlcd1', Aliases='', description='TLC domain containing 1', other_designations='TLC domain-containing protein 1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78176716, end_position_on_the_genomic_accession=78180820, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68393, CurrentID=0, Status='live', Symbol='Mogat1', Aliases='0610030A14Rik, 1110064N14Rik, Dgat2l, Dgat2l1, MGAT1, WI1-2612I11.1, mDC2', description='monoacylglycerol O-acyltransferase 1', other_designations='acyl CoA:monoacylglycerol acyltransferase 1|acyl-CoA:monoacylglycerol acyltransferase 1|diacylglycerol O-acyltransferase 2-like 1|diacylglycerol acyltransferase 2-like protein 1', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=78511060, end_position_on_the_genomic_accession=78538173, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68394, CurrentID=0, Status='live', Symbol='Ccdc163', Aliases='0610037D15Rik, 4933430J04Rik', description='coiled-coil domain containing 163', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116708555, end_position_on_the_genomic_accession=116715104, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68396, CurrentID=0, Status='live', Symbol='Nat8', Aliases='0610037O16Rik, CCNAT, Cml4', description='N-acetyltransferase 8 (GCN5-related)', other_designations='ATase2|N-acetyltransferase 8 (GCN5-related, putative)|acetyltransferase 2|camello-like 4|camello-like protein 4|cysteinyl-conjugate N-acetyltransferase|probable N-acetyltransferase CML4', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85830387, end_position_on_the_genomic_accession=85836636, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68401, CurrentID=0, Status='live', Symbol='G6pc3', Aliases='0710001K01Rik, AU019276, AU045429, AV128920, AW545836, UGRP', description='glucose 6 phosphatase, catalytic, 3', other_designations='G-6-Pase 3|G6Pase 3|glucose-6-phosphatase catalytic subunit-related|ubiquitous glucose-6-phosphatase catalytic subunit-related protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102189699, end_position_on_the_genomic_accession=102194081, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68404, CurrentID=0, Status='live', Symbol='Nrn1', Aliases='0710008J23Rik, Cpg15, Nrn', description='neuritin 1', other_designations='candidate plasticity gene 15 protein|candidate plasticity-related gene 15', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=36725622, end_position_on_the_genomic_accession=36734477, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  330554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330554, CurrentID=0, Status='live', Symbol='Fan1', Aliases='6030441H18Rik, Mtmr15, mFAN1', description='FANCD2/FANCI-associated nuclease 1', other_designations='coiled-coil domain-containing protein MTMR15|myotubularin related protein 15|myotubularin-related protein 15', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64346166, end_position_on_the_genomic_accession=64374139, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  68416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68416, CurrentID=0, Status='live', Symbol='Sycn', Aliases='0910001K16Rik, 1810038B08Rik, SYL', description='syncollin', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28540885, end_position_on_the_genomic_accession=28542210, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68420, CurrentID=0, Status='live', Symbol='Ankrd13a', Aliases='1100001D10Rik, ANKRD13, AU046136', description='ankyrin repeat domain 13a', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114774713, end_position_on_the_genomic_accession=114805821, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  68421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68421, CurrentID=0, Status='live', Symbol='Lmbrd1', Aliases='0910001K20Rik, AV347960', description='LMBR1 domain containing 1', other_designations='LMBR1 domain-containing protein 1|protein N90b', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=24678544, end_position_on_the_genomic_accession=24766301, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  68423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68423, CurrentID=0, Status='live', Symbol='Ankrd13d', Aliases='0710001P18Rik, AI430801', description='ankyrin repeat domain 13 family, member D', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4270180, end_position_on_the_genomic_accession=4283482, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  68427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68427, CurrentID=0, Status='live', Symbol='Slc39a13', Aliases='1100001L14Rik, AA387663, ZIP-13, ZIP13', description='solute carrier family 39 (metal ion transporter), member 13', other_designations='solute carrier family 39 member 13|zrt- and Irt-like protein 13', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91061781, end_position_on_the_genomic_accession=91070315, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  68428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68428, CurrentID=0, Status='live', Symbol='Steap3', Aliases='1010001D01Rik, Tsap6, pHyde', description='STEAP family member 3', other_designations='dudulin 2|dudulin-2|protein nm1054|six-transmembrane epithelial antigen of prostate 3|tumor suppressor-activated pathway protein 6', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120226414, end_position_on_the_genomic_accession=120271082, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68431, CurrentID=0, Status='live', Symbol='Fbxl15', Aliases='0710008C12Rik, Fbxo37', description='F-box and leucine-rich repeat protein 15', other_designations='F-box only protein 37', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46328059, end_position_on_the_genomic_accession=46330446, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  330577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330577, CurrentID=0, Status='live', Symbol='Saxo2', Aliases='1700129I04Rik, D030069K18, Fam154b', description='stablizer of axonemal microtubules 2', other_designations='family with sequence similarity 154, member B|protein FAM154B', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=82632960, end_position_on_the_genomic_accession=82648563, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68436, CurrentID=0, Status='live', Symbol='Rpl34', Aliases='1100001I22Rik', description='ribosomal protein L34', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=130726827, end_position_on_the_genomic_accession=130730398, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68440, CurrentID=0, Status='live', Symbol='Dusp23', Aliases='1300005N15Rik, LDP-3', description='dual specificity phosphatase 23', other_designations='low molecular mass dual specificity phosphatase 3|low-molecular-mass dual-specificity phosphatase', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172630769, end_position_on_the_genomic_accession=172632974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68441, CurrentID=0, Status='live', Symbol='Rraga', Aliases='1300010C19Rik, AI255374, FIP-1, RAGA', description='Ras-related GTP binding A', other_designations='rag A|rraga', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86575673, end_position_on_the_genomic_accession=86577285, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68444, CurrentID=0, Status='live', Symbol='Cyp2d13', Aliases='1300007K12Rik', description='cytochrome P450, family 2, subfamily d, polypeptide 13', other_designations='cytochrome P450, 2d13', map_location='15 38.67 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82636750, end_position_on_the_genomic_accession=82642045, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68449, CurrentID=0, Status='live', Symbol='Tbc1d10b', Aliases='1110003P22Rik, C87963', description='TBC1 domain family, member 10b', other_designations='protein wz3-85', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127197459, end_position_on_the_genomic_accession=127208468, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68453, CurrentID=0, Status='live', Symbol='Gpihbp1', Aliases='1110002J19Rik, GPI-HBP1', description='GPI-anchored HDL-binding protein 1', other_designations='gpihbp1|high density lipoprotein-binding protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75596658, end_position_on_the_genomic_accession=75598213, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  723814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723814, CurrentID=0, Status='live', Symbol='Mir223', Aliases='Mirn223, miR-223, mmu-mir-223', description='microRNA 223', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=96242817, end_position_on_the_genomic_accession=96242926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723815, CurrentID=0, Status='live', Symbol='Mir128-2', Aliases='Mirn128-2, Mirn128b, mir-128-2, mir-128b', description='microRNA 128-2', other_designations='microRNA 128b|mir 128b|mmu-mir-128-2|mmu-mir-128b', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=112118636, end_position_on_the_genomic_accession=112118711, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723816, CurrentID=0, Status='live', Symbol='Mir130b', Aliases='Mirn130b, mir-130b', description='microRNA 130b', other_designations='mir 130b|mmu-mir-130b', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17124061, end_position_on_the_genomic_accession=17124142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723817, CurrentID=0, Status='live', Symbol='Mir133b', Aliases='Mirn133b, mir-133b', description='microRNA 133b', other_designations='mir 133b|mmu-mir-133b', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20682769, end_position_on_the_genomic_accession=20682887, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68458, CurrentID=0, Status='live', Symbol='Ppp1r14a', Aliases='1110001M11Rik, Cpi17', description='protein phosphatase 1, regulatory (inhibitor) subunit 14A', other_designations='17 kDa PKC-potentiated inhibitory protein of PP1|CPI-17|protein kinase C-potentiated inhibitor protein of 17 kDa', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29289320, end_position_on_the_genomic_accession=29293390, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  723819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723819, CurrentID=0, Status='live', Symbol='Mir181c', Aliases='Mirn181c, mir-181c', description='microRNA 181c', other_designations='mir 181c|mmu-mir-181c', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84178873, end_position_on_the_genomic_accession=84178961, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723820, CurrentID=0, Status='live', Symbol='Mir196b', Aliases='Mirn196b, mir-196b', description='microRNA 196b', other_designations='mir 196b|mmu-mir-196b', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52230081, end_position_on_the_genomic_accession=52230165, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723821, CurrentID=0, Status='live', Symbol='Mir199a-2', Aliases='Mirn199a-2, mir-199a-2', description='microRNA 199a-2', other_designations='mir 199a-2|mmu-mir-199a-2', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162217814, end_position_on_the_genomic_accession=162217923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723822, CurrentID=0, Status='live', Symbol='Mir218-1', Aliases='Mirn218-1, mir-218-1', description='microRNA 218-1', other_designations='mir 218-1|mmu-mir-218-1', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=48223942, end_position_on_the_genomic_accession=48224051, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68463, CurrentID=0, Status='live', Symbol='Mrpl14', Aliases='1110006I11Rik, AI846816, MRP-L32, Rpml32', description='mitochondrial ribosomal protein L14', other_designations='L14mt|MRP-L14|mitochondrial ribosomal protein L32', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45685694, end_position_on_the_genomic_accession=45698495, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  723824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723824, CurrentID=0, Status='live', Symbol='Mir103-1', Aliases='Mirn103-1, mir-103-1', description='microRNA 103-1', other_designations='mir 103-1|mmu-mir-103-1', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=35782396, end_position_on_the_genomic_accession=35782481, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68465, CurrentID=0, Status='live', Symbol='Adipor2', Aliases='1110001I14Rik, ADCR2, AI115388, AW554121, D6Ucla1e, Paqr2', description='adiponectin receptor 2', other_designations='progestin and adipoQ receptor family member II', map_location='6 56.78 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119353150, end_position_on_the_genomic_accession=119417483, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  723826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723826, CurrentID=0, Status='live', Symbol='Mir107', Aliases='Mirn107, mir-107, mmu-mir-107', description='microRNA 107', other_designations='mir 107', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34820687, end_position_on_the_genomic_accession=34820773, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723827, CurrentID=0, Status='live', Symbol='Mir221', Aliases='Mirn221, mir-221, mmu-mir-221', description='microRNA 221', other_designations='mir 221', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=19146294, end_position_on_the_genomic_accession=19146388, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723828, CurrentID=0, Status='live', Symbol='Mir222', Aliases='Mirn222, mir-222, mmu-mir-222', description='microRNA 222', other_designations='mir 222', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=19146893, end_position_on_the_genomic_accession=19146971, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723829, CurrentID=0, Status='live', Symbol='Mir106a', Aliases='Mirn106a, mir-106a', description='microRNA 106a', other_designations='mir 106a|mmu-mir-106a', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52742503, end_position_on_the_genomic_accession=52742567, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723830, CurrentID=0, Status='live', Symbol='Mir28a', Aliases='Mir28, Mirn28, mmu-mir-28, mmu-mir-28a', description='microRNA 28a', other_designations='microRNA 28|mir 28', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=24827855, end_position_on_the_genomic_accession=24827940, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68472, CurrentID=0, Status='live', Symbol='Tmem126b', Aliases='1110001A23Rik', description='transmembrane protein 126B', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90468833, end_position_on_the_genomic_accession=90475995, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68473, CurrentID=0, Status='live', Symbol='Mob1b', Aliases='1110003E08Rik, AU015450, B230364F10, Mobkl1a', description='MOB kinase activator 1B', other_designations='MOB1, Mps One Binder kinase activator-like 1A|mob1 homolog 1A|mps one binder kinase activator-like 1A', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88720871, end_position_on_the_genomic_accession=88758455, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  723834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723834, CurrentID=0, Status='live', Symbol='Mir301', Aliases='Mirn301, mir-301a, mmu-mir-301', description='microRNA 301', other_designations='mir 301|mmu-mir-301a', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87113004, end_position_on_the_genomic_accession=87113089, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68475, CurrentID=0, Status='live', Symbol='Ssna1', Aliases='1110003H09Rik, 1190004J23Rik, AA934184, N14, NA14', description="Sjogren's syndrome nuclear autoantigen 1", other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25271039, end_position_on_the_genomic_accession=25272418, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  723836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723836, CurrentID=0, Status='live', Symbol='Mir30e', Aliases='Mirn30e, mir-30e, mmu-mir-30e', description='microRNA 30e', other_designations='mir 30e', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120772606, end_position_on_the_genomic_accession=120772697, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723837, CurrentID=0, Status='live', Symbol='Mir32', Aliases='Mirn32, mir-32, mmu-mir-32', description='microRNA 32', other_designations='mir 32', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56895229, end_position_on_the_genomic_accession=56895298, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723838, CurrentID=0, Status='live', Symbol='Mir320', Aliases='Mirn320, mir-320, mmu-mir-320', description='microRNA 320', other_designations='mir 320', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70443510, end_position_on_the_genomic_accession=70443591, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68479, CurrentID=0, Status='live', Symbol='Phf5a', Aliases='1110007B08Rik', description='PHD finger protein 5A', other_designations='PHD finger-like domain protein 5A|SF3b14b|splicing factor 3B-associated 14 kDa protein', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81864516, end_position_on_the_genomic_accession=81871892, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68480, CurrentID=0, Status='live', Symbol='Card19', Aliases='1110007C09Rik, AI851695, Bincard', description='caspase recruitment domain family, member 19', other_designations='Bcl10-interacting protein with Card|bcl10-interacting CARD protein', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49202950, end_position_on_the_genomic_accession=49216067, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68481, CurrentID=0, Status='live', Symbol='Mpzl1', Aliases='1110007A10Rik, PZR', description='myelin protein zero-like 1', other_designations='protein zero-related', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165592181, end_position_on_the_genomic_accession=165634558, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  723842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723842, CurrentID=0, Status='live', Symbol='Mir329', Aliases='Mirn329, mir-329, mmu-mir-329', description='microRNA 329', other_designations='microRNA mir-329|mir 329', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109713481, end_position_on_the_genomic_accession=109713577, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  330627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330627, CurrentID=0, Status='live', Symbol='Trim66', Aliases='Kiaa0298-hp, Tif1d', description='tripartite motif-containing 66', other_designations='TIF1delta|transcriptional intermediary factor 1 delta', map_location='7 57.21 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109449001, end_position_on_the_genomic_accession=109516763, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  723844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723844, CurrentID=0, Status='live', Symbol='Mir338', Aliases='Mirn338, mir-338, mmu-mir-338', description='microRNA 338', other_designations='mir 338', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120014765, end_position_on_the_genomic_accession=120014862, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723847, CurrentID=0, Status='live', Symbol='Mir346', Aliases='Mirn346, mir-346, mmu-mir-346', description='microRNA 346', other_designations='mir 346', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34894609, end_position_on_the_genomic_accession=34894706, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723848, CurrentID=0, Status='live', Symbol='Mir34a', Aliases='Mirn34a, mir-34a, mmu-mir-34a', description='microRNA 34a', other_designations='mir 34a', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150068454, end_position_on_the_genomic_accession=150068555, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723849, CurrentID=0, Status='live', Symbol='Mir34b', Aliases='Mirn34b, mir-34b, mmu-mir-34b', description='microRNA 34b', other_designations='mir 34b', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51103562, end_position_on_the_genomic_accession=51103645, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723850, CurrentID=0, Status='live', Symbol='Mir361', Aliases='Mirn361, mir-361, mmu-mir-361', description='microRNA 361', other_designations='mir 361', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=113074824, end_position_on_the_genomic_accession=113074893, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723851, CurrentID=0, Status='live', Symbol='Mir362', Aliases='Mirn362, mir-362, mmu-mir-362', description='microRNA 362', other_designations='mir 362', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7241982, end_position_on_the_genomic_accession=7242046, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723852, CurrentID=0, Status='live', Symbol='Mir363', Aliases='Mirn363, mir-363, mmu-mir-363', description='microRNA 363', other_designations='mir 363', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52741693, end_position_on_the_genomic_accession=52741767, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723853, CurrentID=0, Status='live', Symbol='Mir365-2', Aliases='Mirn365-2, mir-365-2', description='microRNA 365-2', other_designations='mir 365-2|mmu-mir-365-2', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79726400, end_position_on_the_genomic_accession=79726511, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723854, CurrentID=0, Status='live', Symbol='Mir370', Aliases='Mirn370, mir-370, mmu-mir-370', description='microRNA 370', other_designations='mir 370', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109618258, end_position_on_the_genomic_accession=109618336, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723855, CurrentID=0, Status='live', Symbol='Mir376a', Aliases='Mirn376a, mir-376a', description='microRNA 376a', other_designations='mir 376a|mmu-mir-376a', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109723781, end_position_on_the_genomic_accession=109723848, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723856, CurrentID=0, Status='live', Symbol='Mir376c', Aliases='Mirn376c, mir-376c', description='microRNA 376c', other_designations='mir 376c|mmu-mir-376c', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109722718, end_position_on_the_genomic_accession=109722803, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723857, CurrentID=0, Status='live', Symbol='Mir377', Aliases='Mirn377, mir-377, mmu-mir-377', description='microRNA 377', other_designations='mir 377', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109740510, end_position_on_the_genomic_accession=109740577, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723858, CurrentID=0, Status='live', Symbol='Mir379', Aliases='Mirn379, mir-379, mmu-mir-379', description='microRNA 379', other_designations='mir 379', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109709060, end_position_on_the_genomic_accession=109709125, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723859, CurrentID=0, Status='live', Symbol='Mir380', Aliases='Mirn380, mir-380, mmu-mir-380', description='microRNA 380', other_designations='mir 380', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109711803, end_position_on_the_genomic_accession=109711863, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723860, CurrentID=0, Status='live', Symbol='Mir383', Aliases='Mirn383, mir-383, mmu-mir-383', description='microRNA 383', other_designations='mir 383', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=38252133, end_position_on_the_genomic_accession=38252202, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723861, CurrentID=0, Status='live', Symbol='Mir384', Aliases='Mirn384, mir-384, mmu-mir-384', description='microRNA 384', other_designations='mir 384', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=105344282, end_position_on_the_genomic_accession=105344369, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723862, CurrentID=0, Status='live', Symbol='Mir409', Aliases='Mirn409, mir-409, mmu-mir-409', description='microRNA 409', other_designations='mir 409', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109743158, end_position_on_the_genomic_accession=109743236, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723863, CurrentID=0, Status='live', Symbol='Mir410', Aliases='Mirn410, mmu-mir-410', description='microRNA 410', other_designations='mir 410', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109743715, end_position_on_the_genomic_accession=109743795, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723864, CurrentID=0, Status='live', Symbol='Mir425', Aliases='Mirn425, mmu-mir-425', description='microRNA 425', other_designations='mir 425', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108568777, end_position_on_the_genomic_accession=108568861, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723865, CurrentID=0, Status='live', Symbol='Mir429', Aliases='Mirn429, mir-429, mmu-mir-429', description='microRNA 429', other_designations='mir 429', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156053905, end_position_on_the_genomic_accession=156053987, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723866, CurrentID=0, Status='live', Symbol='Mir431', Aliases='Mirn431, mir-431, mmu-mir-431', description='microRNA 431', other_designations='mir 431', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109590447, end_position_on_the_genomic_accession=109590537, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68507, CurrentID=0, Status='live', Symbol='Ppfia4', Aliases='100042382, 1110008G13Rik, AI448359, AI852265, C81506, Gm3812', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 4', other_designations='Liprin-alpha4', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134296783, end_position_on_the_genomic_accession=134344750, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  723868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723868, CurrentID=0, Status='live', Symbol='Mir449a', Aliases='Mirn449, Mirn449a, mir-449a, mmu-mir-449', description='microRNA 449a', other_designations='microRNA 449|mir 449|mmu-mir-449a', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113037534, end_position_on_the_genomic_accession=113037624, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  11418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11418, CurrentID=0, Status='live', Symbol='Asic2', Aliases='ACIC2, Accn1, BNC1, BNaC1, BNaC1a, Mdeg', description='acid-sensing (proton-gated) ion channel 2', other_designations='amiloride-sensitive brain sodium channel|amiloride-sensitive cation channel 1, neuronal (degenerin)|brain sodium channel 1|degenerin', map_location='11 48.43 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80880163, end_position_on_the_genomic_accession=81968396, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  723870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723870, CurrentID=0, Status='live', Symbol='Mir451a', Aliases='Mir451, Mirn451, mir-451a, mmu-mir-451, mmu-mir-451a', description='microRNA 451a', other_designations='microRNA 451|mir 451', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78073170, end_position_on_the_genomic_accession=78073241, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68511, CurrentID=0, Status='live', Symbol='Dcdc2c', Aliases='1110015M06Rik', description='doublecortin domain containing 2C', other_designations='putative doublecortin domain-containing protein ENSP00000382097 homolog', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28437795, end_position_on_the_genomic_accession=28552942, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  68512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68512, CurrentID=0, Status='live', Symbol='Tomm5', Aliases='1110019J04Rik, Tom5', description='translocase of outer mitochondrial membrane 5 homolog (yeast)', other_designations='tomm5', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45105209, end_position_on_the_genomic_accession=45108113, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  723873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723873, CurrentID=0, Status='live', Symbol='Mir470', Aliases='Mirn470, mir-470, mmu-mir-470', description='microRNA 470', other_designations='mir 470', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=66813951, end_position_on_the_genomic_accession=66814025, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723874, CurrentID=0, Status='live', Symbol='Mir483', Aliases='Mirn483, mir-483, mmu-mir-483', description='microRNA 483', other_designations='mir 483', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142654924, end_position_on_the_genomic_accession=142654996, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68515, CurrentID=0, Status='live', Symbol='Myadml2', Aliases='1110012N22Rik', description='myeloid-associated differentiation marker-like 2', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120646031, end_position_on_the_genomic_accession=120648337, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  330660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330660, CurrentID=0, Status='live', Symbol='Btbd16', Aliases='E330040A16Rik', description='BTB (POZ) domain containing 16', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=130774069, end_position_on_the_genomic_accession=130825899, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  723877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723877, CurrentID=0, Status='live', Symbol='Mir489', Aliases='Mirn489, mir-489, mmu-mir-489', description='microRNA 489', other_designations='mir 489', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3721897, end_position_on_the_genomic_accession=3722003, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  330662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330662, CurrentID=0, Status='live', Symbol='Dock1', Aliases='9130006G06Rik, AI854900, D630004B07Rik', description='dedicator of cytokinesis 1', other_designations='180 kDa protein downstream of CRK|DOCK180|dedicator of cyto-kinesis 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=134670608, end_position_on_the_genomic_accession=135173647, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  723879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723879, CurrentID=0, Status='live', Symbol='Mir503', Aliases='Mirn503, mmu-mir-503', description='microRNA 503', other_designations='mir 503', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53053984, end_position_on_the_genomic_accession=53054054, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68520, CurrentID=0, Status='live', Symbol='Zfyve21', Aliases='1110013H04Rik, AW558780, C85416', description='zinc finger, FYVE domain containing 21', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111814170, end_position_on_the_genomic_accession=111828388, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68521, CurrentID=0, Status='live', Symbol='Fam189b', Aliases='1110013L07Rik', description='family with sequence similarity 189, member B', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89183142, end_position_on_the_genomic_accession=89189291, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  723883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723883, CurrentID=0, Status='live', Symbol='Mir7b', Aliases='Mirn7b, mir-7b, mmu-mir-7b', description='microRNA 7b', other_designations='mir 7b', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56242988, end_position_on_the_genomic_accession=56243098, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723884, CurrentID=0, Status='live', Symbol='Mir7-2', Aliases='Mir7a2, Mirn7-2, mir-7a-2, mir-7a2, mmu-mir-7-2, mmu-mir-7a2', description='microRNA 7-2', other_designations='mir 7-2|mmu-mir-7a-2', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=78888277, end_position_on_the_genomic_accession=78888373, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723885, CurrentID=0, Status='live', Symbol='Mir93', Aliases='Mirn93, mir-93, mmu-mir-93', description='microRNA 93', other_designations='mir 93', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138165523, end_position_on_the_genomic_accession=138165610, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68526, CurrentID=0, Status='live', Symbol='Gpr155', Aliases='1110017O10Rik, DEPDC3, F730029F15Rik, PGR22', description='G protein-coupled receptor 155', other_designations='GPR155 variant 3|GPR155 variant 6', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73341506, end_position_on_the_genomic_accession=73386851, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  68527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68527, CurrentID=0, Status='live', Symbol='Ucma', Aliases='1110017I16Rik, AW121955, Grp', description='upper zone of growth plate and cartilage matrix associated', other_designations='Gla-rich protein|upper zone of growth plate and cartilage matrix associated protein', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=4976122, end_position_on_the_genomic_accession=4985748, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  723889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723889, CurrentID=0, Status='live', Symbol='Mir378a', Aliases='Mir378, Mirn378, mmu-mir-378, mmu-mir-378a', description='microRNA 378a', other_designations='microRNA 378', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61397835, end_position_on_the_genomic_accession=61397900, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723890, CurrentID=0, Status='live', Symbol='Mir181b-1', Aliases='Mirn181b-1, mir-181b-1', description='microRNA 181b-1', other_designations='mir 181b-1|mmu-mir-181b-1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=137966639, end_position_on_the_genomic_accession=137966718, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723891, CurrentID=0, Status='live', Symbol='Mir19a', Aliases='Mirn19a, mir-19a, mmu-mir-19a', description='microRNA 19a', other_designations='mir 19a', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115044000, end_position_on_the_genomic_accession=115044081, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723892, CurrentID=0, Status='live', Symbol='Mir100', Aliases='Mirn100, mir-100, mmu-mir-100', description='microRNA 100', other_designations='mir 100', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=41531425, end_position_on_the_genomic_accession=41531504, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723893, CurrentID=0, Status='live', Symbol='Mir10a', Aliases='Mirn10a, miR-10a, mmu-mir-10a', description='microRNA 10a', other_designations='MicroRNA-10a|mir 10a', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96317165, end_position_on_the_genomic_accession=96317274, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723894, CurrentID=0, Status='live', Symbol='Mir224', Aliases='Mirn224, mir-224, mmu-mir-224', description='microRNA 224', other_designations='mir 224', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72261031, end_position_on_the_genomic_accession=72261112, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723895, CurrentID=0, Status='live', Symbol='Mir31', Aliases='Mirn31, miR-31, mmu-mir-31', description='microRNA 31', other_designations='mir 31', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88910557, end_position_on_the_genomic_accession=88910662, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723896, CurrentID=0, Status='live', Symbol='Mir324', Aliases='Mirn324, mir-324, mmu-mir-324', description='microRNA 324', other_designations='mir 324', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70012043, end_position_on_the_genomic_accession=70012131, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723897, CurrentID=0, Status='live', Symbol='Mir33', Aliases='Mirn33, mir-33, mmu-mir-33', description='microRNA 33', other_designations='mir 33', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82198122, end_position_on_the_genomic_accession=82198190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723898, CurrentID=0, Status='live', Symbol='Mir339', Aliases='Mirn339, mir-339, mmu-mir-339', description='microRNA 339', other_designations='mir 339', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139369650, end_position_on_the_genomic_accession=139369745, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68539, CurrentID=0, Status='live', Symbol='Tmem109', Aliases='1110006I15Rik, AA958958, C77652', description='transmembrane protein 109', other_designations='mg23|mitsugumin-23|mitsugumin23', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10870657, end_position_on_the_genomic_accession=10881743, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  723900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723900, CurrentID=0, Status='live', Symbol='Mir375', Aliases='Mirn375, mir-375, mmu-mir-375', description='microRNA 375', other_designations='mir 375', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74900658, end_position_on_the_genomic_accession=74900721, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723901, CurrentID=0, Status='live', Symbol='Mir542', Aliases='Mirn542, mir-542, mmu-mir-542', description='microRNA 542', other_designations='mir 542', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53049403, end_position_on_the_genomic_accession=53049487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723902, CurrentID=0, Status='live', Symbol='Mir7-1', Aliases='Mir7a1, Mirn7-1, miR-7a1, mir-7a-1, mmu-miR-7a1, mmu-mir-7-1', description='microRNA 7-1', other_designations='mir 7-1|mmu-mir-7a-1', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58392779, end_position_on_the_genomic_accession=58392886, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723903, CurrentID=0, Status='live', Symbol='Mir181b-2', Aliases='Mirn181b-2, mir-181b-2', description='microRNA 181b-2', other_designations='mir 181b-2|mmu-mir-181b-2', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38853830, end_position_on_the_genomic_accession=38853918, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68544, CurrentID=0, Status='live', Symbol='2310036O22Rik', Aliases='1110002H14Rik, AI854688', description='RIKEN cDNA 2310036O22 gene', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85026814, end_position_on_the_genomic_accession=85030286, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  723905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723905, CurrentID=0, Status='live', Symbol='Mir17', Aliases='Mirn17, mir-17, mmu-mir-17', description='microRNA 17', other_designations='mir 17', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115043671, end_position_on_the_genomic_accession=115043754, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723906, CurrentID=0, Status='live', Symbol='Mir296', Aliases='Mirn296, mir-296, mmu-mir-296', description='microRNA 296', other_designations='mir 296', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174267047, end_position_on_the_genomic_accession=174267125, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723907, CurrentID=0, Status='live', Symbol='Mir322', Aliases='Mirn322, mmu-mir-322', description='microRNA 322', other_designations='mir 322', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53054255, end_position_on_the_genomic_accession=53054349, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723908, CurrentID=0, Status='live', Symbol='Mir331', Aliases='Mirn331, mmu-mir-331', description='microRNA 331', other_designations='mir 331', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93963768, end_position_on_the_genomic_accession=93963863, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68549, CurrentID=0, Status='live', Symbol='Sgol2a', Aliases='1110007N04Rik, 4932411A20Rik, 5730576N04Rik, AU018846, AV008062, D1Ertd8e, Sgol2, Tripin', description='shugoshin-like 2a (S. pombe)', other_designations='sgo2|shugoshin-2', map_location='1 28.86 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=57985340, end_position_on_the_genomic_accession=58026277, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  723910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723910, CurrentID=0, Status='live', Symbol='Mir351', Aliases='Mirn351, mmu-mir-351', description='microRNA 351', other_designations='mir 351', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53053255, end_position_on_the_genomic_accession=53053353, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  330695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330695, CurrentID=0, Status='live', Symbol='Ctxn1', Aliases='BC028881, Ctxn', description='cortexin 1', other_designations='', map_location='8 1.99 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4257646, end_position_on_the_genomic_accession=4259274, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  723912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723912, CurrentID=0, Status='live', Symbol='Mir382', Aliases='Mirn382, mir-382, mmu-mir-382', description='microRNA 382', other_designations='mir 382', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109733771, end_position_on_the_genomic_accession=109733846, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68553, CurrentID=0, Status='live', Symbol='Col6a4', Aliases='1110001D15Rik, AI413310, AU023415, Dvwa, EG235580, Vwa6', description='collagen, type VI, alpha 4', other_designations='double von Willebrand factor A domains|dual von Willebrand factor A domains', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105990318, end_position_on_the_genomic_accession=106096691, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  723914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723914, CurrentID=0, Status='live', Symbol='Mir448', Aliases='Mirn448, mir-448, mmu-mir-448', description='microRNA 448', other_designations='mir 448', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=147158210, end_position_on_the_genomic_accession=147158321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68556, CurrentID=0, Status='live', Symbol='Uckl1', Aliases='1110007H10Rik, Urkl1', description='uridine-cytidine kinase 1-like 1', other_designations='uridine kinase-like 1', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181569149, end_position_on_the_genomic_accession=181581973, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  723917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723917, CurrentID=0, Status='live', Symbol='Mir539', Aliases='Mirn539, mir-539, mmu-mir-539', description='microRNA 539', other_designations='mir 539', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109728129, end_position_on_the_genomic_accession=109728202, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68558, CurrentID=0, Status='live', Symbol='Ankra2', Aliases='1110004M18Rik, Ankra', description='ankyrin repeat, family A (RFXANK-like), 2', other_designations='RFXANK-like protein 2', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98263076, end_position_on_the_genomic_accession=98274754, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68559, CurrentID=0, Status='live', Symbol='Pdrg1', Aliases='1110004D19Rik, AU020945, AV006589, Pdrg', description='p53 and DNA damage regulated 1', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153008890, end_position_on_the_genomic_accession=153015383, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  723920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723920, CurrentID=0, Status='live', Symbol='Mir302a', Aliases='Mirn302, Mirn302a, mir-302a, mmu-mir-302', description='microRNA 302a', other_designations='microRNA 302|mir 302|mmu-mir-302a', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127545496, end_position_on_the_genomic_accession=127545564, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723921, CurrentID=0, Status='live', Symbol='Mir350', Aliases='Mirn350, mir-350, mmu-mir-350', description='microRNA 350', other_designations='mir 350', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=176772325, end_position_on_the_genomic_accession=176772423, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723922, CurrentID=0, Status='live', Symbol='Mir466', Aliases='Mirn466, mmu-mir-466', description='microRNA 466', other_designations='mir 466|mmu-mir-466a', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10507918, end_position_on_the_genomic_accession=10507990, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723923, CurrentID=0, Status='live', Symbol='Mir20b', Aliases='Mirn20b, mir-20b, mmu-mir-20b', description='microRNA 20b', other_designations='mir 20b', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52742113, end_position_on_the_genomic_accession=52742192, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68564, CurrentID=0, Status='live', Symbol='Nufip2', Aliases='1110001M19Rik, 9530056D24Rik, AI256692, AI461731, PIG1, mKIAA1321', description='nuclear fragile X mental retardation protein interacting protein 2', other_designations='82 kDa FMRP-interacting protein|82-FIP|FMRP-interacting protein 2|proliferation-inducing gene 1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77685650, end_position_on_the_genomic_accession=77717966, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  723925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723925, CurrentID=0, Status='live', Symbol='Mir106b', Aliases='Mirn106b, mir-106b', description='microRNA 106b', other_designations='mir 106b|mmu-mir-106b', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138165737, end_position_on_the_genomic_accession=138165818, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68566, CurrentID=0, Status='live', Symbol='Caly', Aliases='0710001P07Rik, 1110004A22Rik, Calcyon, Drd1ip', description='calcyon neuron-specific vesicular protein', other_designations='calcyon; D1 dopamine receptor-interacting protein|dopamine receptor D1 interacting protein', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140069880, end_position_on_the_genomic_accession=140082575, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68567, CurrentID=0, Status='live', Symbol='Cgref1', Aliases='1110004G24Rik, AW045201, Cgr11', description='cell growth regulator with EF hand domain 1', other_designations='cell growth regulatory gene 11 protein|cell growth regulatory with EF-hand domain|hydrophobestin', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30933143, end_position_on_the_genomic_accession=30945480, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  723928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723928, CurrentID=0, Status='live', Symbol='Mir302d', Aliases='Mirn302d, mir-302d', description='microRNA 302d', other_designations='mir 302d|mmu-mir-302d', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127545624, end_position_on_the_genomic_accession=127545689, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723929, CurrentID=0, Status='live', Symbol='Mir325', Aliases='Mirn325, mmu-mir-325', description='microRNA 325', other_designations='mir 325', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=105379082, end_position_on_the_genomic_accession=105379179, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723930, CurrentID=0, Status='live', Symbol='Mir335', Aliases='Mirn335, mir-335, mmu-mir-335', description='microRNA 335', other_designations='mir 335', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30741299, end_position_on_the_genomic_accession=30741396, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723931, CurrentID=0, Status='live', Symbol='Mir344', Aliases='Mirn344, mir-344-1, mmu-mir-344', description='microRNA 344', other_designations='mir 344|mmu-mir-344-1', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=61877770, end_position_on_the_genomic_accession=61877864, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723932, CurrentID=0, Status='live', Symbol='Mir34c', Aliases='Mirn34c, mir-34c, mmu-mir-34c', description='microRNA 34c', other_designations='mir 34c', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51103034, end_position_on_the_genomic_accession=51103110, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723933, CurrentID=0, Status='live', Symbol='Mir369', Aliases='Mirn369, mir-369, mmu-mir-369', description='microRNA 369', other_designations='mir 369', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109743418, end_position_on_the_genomic_accession=109743496, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723934, CurrentID=0, Status='live', Symbol='Mir376b', Aliases='Mirn376b, mir-376b', description='microRNA 376b', other_designations='mir 376b|mmu-mir-376b', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109723458, end_position_on_the_genomic_accession=109723539, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723935, CurrentID=0, Status='live', Symbol='Mir381', Aliases='Mirn381, mir-381, mmu-mir-381', description='microRNA 381', other_designations='mir 381', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109726822, end_position_on_the_genomic_accession=109726896, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68576, CurrentID=0, Status='live', Symbol='Lamtor5', Aliases='1110003H18Rik, Hbxip, XIP', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 5', other_designations='HBV X-interacting protein homolog|HBX-interacting protein homolog|hepatitis B virus X-interacting protein homolog|hepatitis B virus x interacting protein|late endosomal/lysosomal adaptor and MAPK and MTOR activator 5', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107278858, end_position_on_the_genomic_accession=107284082, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  723937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723937, CurrentID=0, Status='live', Symbol='Mir433', Aliases='Mirn433, mir-433, mmu-mir-433', description='microRNA 433', other_designations='mir 433', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109591715, end_position_on_the_genomic_accession=109591838, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  330723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330723, CurrentID=0, Status='live', Symbol='Htra4', Aliases='B430206E18Rik', description='HtrA serine peptidase 4', other_designations='high-temperature requirement factor A4', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25024928, end_position_on_the_genomic_accession=25038962, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68581, CurrentID=0, Status='live', Symbol='Tmed10', Aliases='1110014C03Rik, Tmp21, p23, p24delta1', description='transmembrane emp24-like trafficking protein 10 (yeast)', other_designations='21 kDa transmembrane-trafficking protein|p24 family protein delta-1|transmembrane protein Tmp21|transmembrane trafficking protein', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85340614, end_position_on_the_genomic_accession=85374717, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  723942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723942, CurrentID=0, Status='live', Symbol='Mir92-2', Aliases='Mirn92-2, mir-92a-2', description='microRNA 92-2', other_designations='mir 92-2|mmu-mir-92-2|mmu-mir-92a-2', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52741838, end_position_on_the_genomic_accession=52741928, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723944, CurrentID=0, Status='live', Symbol='Mir200c', Aliases='Mirn200c, mir-200c', description='microRNA 200c', other_designations='mir 200c|mmu-mir-200c', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124718322, end_position_on_the_genomic_accession=124718390, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68585, CurrentID=0, Status='live', Symbol='Rtn4', Aliases='1110020G17Rik, AA407876, AA409940, AA960376, ASY, C130026I10Rik, NOGO, NSP-CL, NgA, Nogo-A, Nogo-B, Nogo-C, mKIAA0886, mKIAA4153', description='reticulon 4', other_designations='neurite outgrowth inhibitor', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29692898, end_position_on_the_genomic_accession=29744414, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  723946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723946, CurrentID=0, Status='live', Symbol='Mir345', Aliases='Mirn345, mir-345, mmu-mir-345', description='microRNA 345', other_designations='mir 345', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108836973, end_position_on_the_genomic_accession=108837068, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723947, CurrentID=0, Status='live', Symbol='Mir98', Aliases='Mirn98, mir-98, mmu-mir-98', description='microRNA 98', other_designations='mir 98', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151913214, end_position_on_the_genomic_accession=151913321, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68588, CurrentID=0, Status='live', Symbol='Cthrc1', Aliases='1110014B07Rik', description='collagen triple helix repeat containing 1', other_designations='', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39076851, end_position_on_the_genomic_accession=39087121, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  723949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723949, CurrentID=0, Status='live', Symbol='Mir16-2', Aliases='Mirn16-2, mir-16-2', description='microRNA 16-2', other_designations='mmu-mir-16-2', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69009902, end_position_on_the_genomic_accession=69009996, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723950, CurrentID=0, Status='live', Symbol='Mir124a-2', Aliases='Mirn124a-2, mir-124-2, mir-124a-2', description='microRNA 124a-2', other_designations='mmu-mir-124-2|mmu-mir-124a-2', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=17795662, end_position_on_the_genomic_accession=17795770, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723951, CurrentID=0, Status='live', Symbol='Mir124a-3', Aliases='Mirn124a-3, mir-124-3, mir-124a-3', description='microRNA 124a-3', other_designations='mmu-mir-124-3|mmu-mir-124a-3', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180894040, end_position_on_the_genomic_accession=180894107, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723952, CurrentID=0, Status='live', Symbol='Mir125b-2', Aliases='Mirn125b-2, mir-125b-2', description='microRNA 125b-2', other_designations='mmu-mir-125b-2', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=77646273, end_position_on_the_genomic_accession=77646343, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723953, CurrentID=0, Status='live', Symbol='Mir129-2', Aliases='Mirn129-2, mir-129-2', description='microRNA 129-2', other_designations='mmu-mir-129-2', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=94241364, end_position_on_the_genomic_accession=94241453, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723954, CurrentID=0, Status='live', Symbol='Mir133a-2', Aliases='Mirn133a-2, mir-133a-2', description='microRNA 133a-2', other_designations='mir 133a-2|mmu-mir-133a-2', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180398379, end_position_on_the_genomic_accession=180398482, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723955, CurrentID=0, Status='live', Symbol='Mir135a-2', Aliases='Mirn135a-2, mir-135a-2', description='microRNA 135a-2', other_designations='mir 135a-2|mmu-mir-135a-2', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=92072086, end_position_on_the_genomic_accession=92072185, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723956, CurrentID=0, Status='live', Symbol='Mir138-2', Aliases='Mirn138-2, mir-138-2', description='microRNA 138-2', other_designations='mir 138-2|mircoRNA 138-2|mmu-mir-138-2', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94324311, end_position_on_the_genomic_accession=94324381, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723957, CurrentID=0, Status='live', Symbol='Mir194-2', Aliases='Mirn194-2, mir-194-2', description='microRNA 194-2', other_designations='mir 194-2|mmu-mir-194-2', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6264643, end_position_on_the_genomic_accession=6264728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723958, CurrentID=0, Status='live', Symbol='Mir196a-2', Aliases='Mirn196a-2, mir-196a-2', description='microRNA 196a-2', other_designations='mir 196a-2|mmu-mir-196a-2', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102973350, end_position_on_the_genomic_accession=102973434, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723959, CurrentID=0, Status='live', Symbol='Mir1a-2', Aliases='Mir1-2, Mirn1-2, mir-1a-2, mmu-mir-1-2, mmu-mir-1a-2', description='microRNA 1a-2', other_designations='microRNA 1-2|mir 1-2', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10785481, end_position_on_the_genomic_accession=10785552, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  723960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723960, CurrentID=0, Status='live', Symbol='Mir24-2', Aliases='Mirn24-2, mir-24-2', description='microRNA 24-2', other_designations='mir 24-2|mmu-mir-24-2', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84208815, end_position_on_the_genomic_accession=84208921, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723962, CurrentID=0, Status='live', Symbol='Mir26a-2', Aliases='Mirn26a-2, mir-26a-2', description='microRNA 26a-2', other_designations='mir 26a-2|mmu-mir-26a-2', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=126995530, end_position_on_the_genomic_accession=126995613, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723963, CurrentID=0, Status='live', Symbol='Mir29b-2', Aliases='Mirn29b-2, mir-29b-2', description='microRNA 29b-2', other_designations='mir 29b-2|mmu-mir-29b-2', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=195037040, end_position_on_the_genomic_accession=195037120, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723964, CurrentID=0, Status='live', Symbol='Mir30c-2', Aliases='Mirn30c-2, mir-30c-2', description='microRNA 30c-2', other_designations='mir 30c-2|mmu-mir-30c-2', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=23291701, end_position_on_the_genomic_accession=23291784, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723965, CurrentID=0, Status='live', Symbol='Mirlet7a-2', Aliases='Mirnlet7a-2, let-7a-2', description='microRNA let7a-2', other_designations='let 7a-2|mmu-let-7a-2', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=41536716, end_position_on_the_genomic_accession=41536811, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723966, CurrentID=0, Status='live', Symbol='Mirlet7c-2', Aliases='Mirnlet7c-2, let-7c-2', description='microRNA let7c-2', other_designations='let 7c-2|mmu-let-7c-2', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85706603, end_position_on_the_genomic_accession=85706697, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723967, CurrentID=0, Status='live', Symbol='Mir9-2', Aliases='Mirn9-2, mir-9-2, mmu-mir-9-2', description='microRNA 9-2', other_designations='mir 9-2', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=83738814, end_position_on_the_genomic_accession=83738885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  723968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723968, CurrentID=0, Status='live', Symbol='Mir9-3', Aliases='Mirn9-3, mir-9-3, mmu-mir-9-3', description='microRNA 9-3', other_designations='mir 9-3', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79505264, end_position_on_the_genomic_accession=79505353, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68611, CurrentID=0, Status='live', Symbol='Mrpl28', Aliases='1110015G04Rik, L28mt, MAAT1, MRP-L28, p15', description='mitochondrial ribosomal protein L28', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26123503, end_position_on_the_genomic_accession=26126613, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68612, CurrentID=0, Status='live', Symbol='Ube2c', Aliases='1110015A16Rik, D2Ertd695e', description='ubiquitin-conjugating enzyme E2C', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme C|E2 ubiquitin-conjugating enzyme C|ubcH10|ubiquitin carrier protein C|ubiquitin-protein ligase C', map_location='2 85.27 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164769929, end_position_on_the_genomic_accession=164772902, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68614, CurrentID=0, Status='live', Symbol='Letmd1', Aliases='1110019O13Rik, AI593524, BB130465, BB235638, HCCR-2, HCCR1, MCC-32, Mccr', description='LETM1 domain containing 1', other_designations='cervical cancer receptor', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100468165, end_position_on_the_genomic_accession=100479252, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68616, CurrentID=0, Status='live', Symbol='Gdpd3', Aliases='1110015E22Rik', description='glycerophosphodiester phosphodiesterase domain containing 3', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126766414, end_position_on_the_genomic_accession=126775645, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68617, CurrentID=0, Status='live', Symbol='Mtcl1', Aliases='1110012J17Rik, Ccdc165, D130071O13, Kiaa0802, Soga2, mKIAA0802, t8219b25', description='microtubule crosslinking factor 1', other_designations='PAR-1-interacting protein|SOGA family member 2|coiled-coil domain-containing protein 165|protein SOGA2', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=66336980, end_position_on_the_genomic_accession=66449808, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  68618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68618, CurrentID=0, Status='live', Symbol='1110012L19Rik', Aliases='', description='RIKEN cDNA 1110012L19 gene', other_designations='', map_location='X A7.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=70385913, end_position_on_the_genomic_accession=70389416, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68625, CurrentID=0, Status='live', Symbol='Cfap57', Aliases='1110020C03Rik, AV009650, C130004B06Rik, Gm1363, Wdr65', description='cilia and flagella associated protein 57', other_designations='WD repeat domain 65|WD repeat-containing protein 65', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118554550, end_position_on_the_genomic_accession=118620766, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  68626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68626, CurrentID=0, Status='live', Symbol='Elac2', Aliases='1110017O07Rik, D11Wsu80e, Hpc2', description='elaC ribonuclease Z 2', other_designations='RNase Z 2|elaC homolog 2|elaC homolog protein 2|putative prostate cancer susceptibility protein|ribonuclease Z 2|tRNA 3 endonuclease 2|tRNase Z 2|tRNase Z(L)', map_location='11 40.42 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=64979023, end_position_on_the_genomic_accession=65002076, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  68628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68628, CurrentID=0, Status='live', Symbol='Fbxw9', Aliases='1110017H11Rik, Fbw9', description='F-box and WD-40 domain protein 9', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85060025, end_position_on_the_genomic_accession=85067121, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68631, CurrentID=0, Status='live', Symbol='Cryl1', Aliases='1110025H08Rik, A230106J09Rik, C85932, gul3dh', description='crystallin, lambda 1', other_designations='L-gulonate 3-dehydrogenase|crystallin, lamda 1', map_location='14 30.1 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57275034, end_position_on_the_genomic_accession=57398483, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68632, CurrentID=0, Status='live', Symbol='Myct1', Aliases='1110020B04Rik, AI225941, AI642973, Mtmc1', description='myc target 1', other_designations='myc target in myeloid cells protein 1', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=5593728, end_position_on_the_genomic_accession=5606789, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68634, CurrentID=0, Status='live', Symbol='Tm2d3', Aliases='1110025I09Rik, 5930422O05Rik, Blp2', description='TM2 domain containing 3', other_designations='BBP-like protein 2', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=65693417, end_position_on_the_genomic_accession=65701913, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68636, CurrentID=0, Status='live', Symbol='Fahd1', Aliases='1110025H10Rik, AV002085', description='fumarylacetoacetate hydrolase domain containing 1', other_designations='OAA decarboxylase|fumarylacetoacetate hydrolase domain-containing protein 1|oxaloacetate decarboxylase', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24848896, end_position_on_the_genomic_accession=24850302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68642, CurrentID=0, Status='live', Symbol='Tmem216', Aliases='1110017C22Rik, 2810441K11Rik, 4921533J23Rik, A930021F15Rik, AI482550', description='transmembrane protein 216', other_designations='thymus atrophy-related protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10539326, end_position_on_the_genomic_accession=10556303, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68644, CurrentID=0, Status='live', Symbol='Abhd14a', Aliases='1110013B16Rik, AW558221, Dorz1', description='abhydrolase domain containing 14A', other_designations='abhydrolase domain-containing protein 14A|down-regulated in Zic-1-mutant protein|down-regulated in Zic1 deficient cerebellar primordium|downregulated in Zic1 deficient cerebellum', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106440051, end_position_on_the_genomic_accession=106451384, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  330790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330790, CurrentID=0, Status='live', Symbol='Hapln4', Aliases='9330174O11, Bral2, Lpr4', description='hyaluronan and proteoglycan link protein 4', other_designations='brain link protein 2|brain linked protein 2|link protein 4', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70083529, end_position_on_the_genomic_accession=70090862, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68652, CurrentID=0, Status='live', Symbol='Tab2', Aliases='1110030N06Rik, A530078N03Rik, Map3k7ip2, mKIAA0733', description='TGF-beta activated kinase 1/MAP3K7 binding protein 2', other_designations='TAB-2|TAK1-binding protein 2|TGF-beta-activated kinase 1-binding protein 2|Tak1 binding protein 2|mitogen activated protein kinase kinase kinase 7 interacting protein 2|mitogen-activated protein kinase kinase kinase 7 interacting protein 2|mitogen-activated protein kinase kinase kinase 7-interacting protein 2', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7905648, end_position_on_the_genomic_accession=7956225, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68653, CurrentID=0, Status='live', Symbol='Samm50', Aliases='1110030L07Rik', description='sorting and assembly machinery component 50 homolog (S. cerevisiae)', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84192233, end_position_on_the_genomic_accession=84214303, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  68659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68659, CurrentID=0, Status='live', Symbol='Fam198b', Aliases='1110032E23Rik, 2210419I08Rik, AV011458, Ened', description='family with sequence similarity 198, member B', other_designations='expressed in nerve and epithelium during development|protein ENED', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79884475, end_position_on_the_genomic_accession=79946278, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68662, CurrentID=0, Status='live', Symbol='Scgb3a1', Aliases='HIN-1, LuLeu2, Pnsp2, UGRP2, Utgrp2', description='secretoglobin, family 3A, member 1', other_designations='cytokine HIN-1|high in normal 1|pneumo secretory protein 2|uteroglobin related protein 2|uteroglobin-related protein 2', map_location='11 29.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49663595, end_position_on_the_genomic_accession=49665118, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68666, CurrentID=0, Status='live', Symbol='Svop', Aliases='1110030H18Rik, AI415691, msvop', description='SV2 related protein', other_designations='SV2-related protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114026913, end_position_on_the_genomic_accession=114091423, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  68667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68667, CurrentID=0, Status='live', Symbol='Trpm4', Aliases='1110030C19Rik, AW047689, LTRPC4, LTrpC-4, TRPM4B', description='transient receptor potential cation channel, subfamily M, member 4', other_designations='calcium-activated non-selective cation channel 1|long transient receptor potential channel 4|transient receptor potential ion channel melastatin subgroup member 4', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45303155, end_position_on_the_genomic_accession=45334070, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  330812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330812, CurrentID=0, Status='live', Symbol='Rnf150', Aliases='A630007N06Rik, C030044C12Rik, Greul5, mKIAA1214', description='ring finger protein 150', other_designations='', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=82863018, end_position_on_the_genomic_accession=83091271, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  330814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330814, CurrentID=0, Status='live', Symbol='Adgrl1', Aliases='2900070I05Rik, AI182398, CLIBA, Lec2, Lphn1, mKIAA0821', description='adhesion G protein-coupled receptor L1', other_designations='CIRL-1|calcium-independent alpha-latrotoxin receptor 1|latrophilin 1|latrophilin-1|lectomedin-2', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83900098, end_position_on_the_genomic_accession=83941954, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  68671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68671, CurrentID=0, Status='live', Symbol='Pcyt2', Aliases='1110033E03Rik, ET', description='phosphate cytidylyltransferase 2, ethanolamine', other_designations='CTP:phosphoethanolamine cytidylyltransferase|phosphorylethanolamine transferase', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120610086, end_position_on_the_genomic_accession=120617949, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  330817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330817, CurrentID=0, Status='live', Symbol='Dhps', Aliases='Dhs', description='deoxyhypusine synthase', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85069718, end_position_on_the_genomic_accession=85075162, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68675, CurrentID=0, Status='live', Symbol='Fam172a', Aliases='1110033M05Rik, 2610318O14Rik, 53-E6, 9430037D06Rik, AF064782, pEN87', description='family with sequence similarity 172, member A', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=77708678, end_position_on_the_genomic_accession=78166240, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  330820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330820, CurrentID=0, Status='live', Symbol='4933402J07Rik', Aliases='', description='RIKEN cDNA 4933402J07 gene', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=87563907, end_position_on_the_genomic_accession=87589197, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68678, CurrentID=0, Status='live', Symbol='Smtnl1', Aliases='1110030K22Rik, CHASM', description='smoothelin-like 1', other_designations='calponin homology-associated smooth muscle protein', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84811176, end_position_on_the_genomic_accession=84826934, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68680, CurrentID=0, Status='live', Symbol='Fitm1', Aliases='1110028A07Rik, AV010547, Cg10671, Fit1', description='fat storage-inducing transmembrane protein 1', other_designations='fat-inducing protein 1|fat-inducing transcript 1', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55575674, end_position_on_the_genomic_accession=55576952, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  98828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98828, CurrentID=0, Status='live', Symbol='Cdc123', Aliases='AA959893, G431001I09Rik', description='cell division cycle 123', other_designations='cell division cycle 123 homolog', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5794294, end_position_on_the_genomic_accession=5844968, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68682, CurrentID=0, Status='live', Symbol='Slc44a2', Aliases='1110028E10Rik, CTL2', description='solute carrier family 44, member 2', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21320719, end_position_on_the_genomic_accession=21355028, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  330830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330830, CurrentID=0, Status='live', Symbol='Drc7', Aliases='Ccdc135, Gm770, SRG-L, SRG-S', description='dynein regulatory complex subunit 7', other_designations='coiled-coil domain containing 135|coiled-coil domain-containing protein 135|coiled-coil domain-containing protein lobo homolog|spermatogenesis-related gene in late stages of spermatogenesis cells protein', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95055103, end_position_on_the_genomic_accession=95078141, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  68691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68691, CurrentID=0, Status='live', Symbol='Kansl1l', Aliases='1110028C15Rik, 1190030G24, AI427598, C430010P07Rik, mKIAA4189', description='KAT8 regulatory NSL complex subunit 1-like', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=66719243, end_position_on_the_genomic_accession=66817596, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  330836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330836, CurrentID=0, Status='live', Symbol='Slc7a6', Aliases='AI643885, LAT-2, LAT3', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 6', other_designations='slc7a6|solute carrier family 7 member 6|solute carriersolute carrier family 7 (cationic amino acid transporter, y+ system), member 6|y(+)L-type amino acid transporter 2|y+LAT-2|y+LAT2', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106168830, end_position_on_the_genomic_accession=106198708, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68693, CurrentID=0, Status='live', Symbol='Hnrnpul2', Aliases='1110031M08Rik, AI465155, Hnrpul2', description='heterogeneous nuclear ribonucleoprotein U-like 2', other_designations='MLF1-associated nuclear protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8819401, end_position_on_the_genomic_accession=8834142, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  68695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68695, CurrentID=0, Status='live', Symbol='Hddc3', Aliases='1110033O09Rik, C86475, MESH1', description='HD domain containing 3', other_designations="(ppGpp)ase|HD domain-containing protein 3|metazoan SpoT homolog 1|metazoan SpoT homolog-1|penta-phosphate guanosine-3'-pyrophosphohydrolase", map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80343083, end_position_on_the_genomic_accession=80346097, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68701, CurrentID=0, Status='live', Symbol='Ppp1r27', Aliases='1110033I14Rik, Dysfip1, toonin', description='protein phosphatase 1, regulatory subunit 27', other_designations='dysferlin interacting protein 1|dysferlin-interacting protein 1|ortholog of human dysferlin interacting protein 1 (toonin) DYSFP1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120549975, end_position_on_the_genomic_accession=120551132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  724062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=724062, CurrentID=0, Status='live', Symbol='Mir101b', Aliases='Mirn101b, mir-101b', description='microRNA 101b', other_designations='mmu-mir-101b', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29135279, end_position_on_the_genomic_accession=29135375, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68703, CurrentID=0, Status='live', Symbol='Rere', Aliases='1110033A15Rik, AI414665, ARG, ARP, ATN1L, AW742570, Atr2, DNB1, eye<m03Jus>, eyes3, mKIAA0458', description='arginine glutamic acid dipeptide (RE) repeats', other_designations='atrophin-2', map_location='4 81.34 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150281549, end_position_on_the_genomic_accession=150621966, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  68705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68705, CurrentID=0, Status='live', Symbol='Gtf2f2', Aliases='1110031C13Rik, AI326315', description='general transcription factor IIF, polypeptide 2', other_designations='ATP-dependent helicase GTF2F2|TFIIF-beta|transcription initiation factor IIF subunit beta', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75896937, end_position_on_the_genomic_accession=76010865, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68708, CurrentID=0, Status='live', Symbol='Rabl2', Aliases='1110031N17Rik, AW549303a, Rabl2', description='RAB, member RAS oncogene family-like 2', other_designations='RAB, member of RAS oncogene family-like 2|RAB, member of RAS oncogene family-like 2A', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89582527, end_position_on_the_genomic_accession=89592547, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68709, CurrentID=0, Status='live', Symbol='Cilp2', Aliases='1110031K21Rik, AA407300, CLIP-2', description='cartilage intermediate layer protein 2', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69880366, end_position_on_the_genomic_accession=69887392, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  623121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623121, CurrentID=0, Status='live', Symbol='Pydc4', Aliases='E030037K03Rik, E430029J22Rik', description='pyrin domain containing 4', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173566595, end_position_on_the_genomic_accession=173599302, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68713, CurrentID=0, Status='live', Symbol='Ifitm1', Aliases='1110036C17Rik, DSPA2a, Mil-2, Mil2', description='interferon induced transmembrane protein 1', other_designations='dispanin subfamily A member 2a|fragilis protein 2|fragilis2|ifitm-like protein 2|interferon induced transmembrane protein 2 like', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140967429, end_position_on_the_genomic_accession=140969828, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68718, CurrentID=0, Status='live', Symbol='Rnf166', Aliases='1110031E24Rik, AW555115, Zfp313l', description='ring finger protein 166', other_designations='zinc finger protein 313; zinc finger protein 228', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122466145, end_position_on_the_genomic_accession=122476064, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  330863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330863, CurrentID=0, Status='live', Symbol='Trim67', Aliases='D130049O21Rik, TNL', description='tripartite motif-containing 67', other_designations='TRIM9-like protein TNL', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124793019, end_position_on_the_genomic_accession=124834704, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68720, CurrentID=0, Status='live', Symbol='Lce1b', Aliases='1110029C13Rik, Sprrl5', description='late cornified envelope 1B', other_designations='small proline rich-like 5', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92655650, end_position_on_the_genomic_accession=92656926, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  68721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68721, CurrentID=0, Status='live', Symbol='1110032A03Rik', Aliases='4930500N06Rik, c11orf1', description='RIKEN cDNA 1110032A03 gene', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50762022, end_position_on_the_genomic_accession=50768174, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68723, CurrentID=0, Status='live', Symbol='Hrnr', Aliases='1110033K19Rik, A530063N20Rik, S100a18', description='hornerin', other_designations='S100 calcium binding protein A18', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93319541, end_position_on_the_genomic_accession=93333572, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68724, CurrentID=0, Status='live', Symbol='Arl8a', Aliases='1110033P22Rik, Arl10b, gie2', description='ADP-ribosylation factor-like 8A', other_designations='ADP-ribosylation factor-like 10B|ADP-ribosylation factor-like protein 10B|novel small G protein indispensable for equal chromosome segregation 2|small G protein indispensable for equal chromosome segregation 2', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135146834, end_position_on_the_genomic_accession=135156268, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68725, CurrentID=0, Status='live', Symbol='1110032F04Rik', Aliases='AI115547', description='RIKEN cDNA 1110032F04 gene', other_designations='', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=68869586, end_position_on_the_genomic_accession=68872163, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68728, CurrentID=0, Status='live', Symbol='Trp53inp2', Aliases='1110029F20Rik, Tp53inp2', description='transformation related protein 53 inducible nuclear protein 2', other_designations='diabetes and obesity-regulated protein|tumor protein p53 inducible nuclear protein 2', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155381808, end_position_on_the_genomic_accession=155389847, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68729, CurrentID=0, Status='live', Symbol='Trim37', Aliases='1110032A10Rik, 2810004E07Rik, AI848587, AU043018, MUL, TEF3, mKIAA0898', description='tripartite motif-containing 37', other_designations='TRAF encompassing factor 3|tripartite motif-containing protein 37', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87127051, end_position_on_the_genomic_accession=87250874, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  68730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68730, CurrentID=0, Status='live', Symbol='Dus1l', Aliases='1110032N12Rik', description='dihydrouridine synthase 1-like (S. cerevisiae)', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120789201, end_position_on_the_genomic_accession=120796427, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68731, CurrentID=0, Status='live', Symbol='Rbfa', Aliases='1110032A13Rik, AI595940', description='ribosome binding factor A', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80192264, end_position_on_the_genomic_accession=80200619, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68732, CurrentID=0, Status='live', Symbol='Lrrc16a', Aliases='1110037D04Rik, AI425970, CARMIL, D130057M20, Lrrc16', description='leucine rich repeat containing 16A', other_designations='CARMIL homolog|leucine rich repeat containing 16', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24012481, end_position_on_the_genomic_accession=24280790, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  68734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68734, CurrentID=0, Status='live', Symbol='Smek1', Aliases='1110034C04Rik, BC064465, Ppp4r3a, mKIAA2010', description='SMEK homolog 1, suppressor of mek1 (Dictyostelium)', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=101039409, end_position_on_the_genomic_accession=101083702, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  68735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68735, CurrentID=0, Status='live', Symbol='Mrps18c', Aliases='1110037D14Rik', description='mitochondrial ribosomal protein S18C', other_designations='28S ribosomal protein S18-1, mitochondrial|MRP-S18-1|MRP-S18-c|S18mt-c|mrps18-c', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100798759, end_position_on_the_genomic_accession=100804467, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68736, CurrentID=0, Status='live', Symbol='Tyw5', Aliases='1110034B05Rik', description='tRNA-yW synthesizing protein 5', other_designations='jmjC domain-containing protein C2orf60 homolog|tRNA yW-synthesizing enzyme 5|tRNA(Phe) (7-(3-amino-3-carboxypropyl)wyosine(37)-C(2))-hydroxylase', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=57387899, end_position_on_the_genomic_accession=57406674, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68737, CurrentID=0, Status='live', Symbol='Angel1', Aliases='1110030H02Rik, mKIAA0759', description='angel homolog 1 (Drosophila)', other_designations='angel1', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86700502, end_position_on_the_genomic_accession=86734187, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68738, CurrentID=0, Status='live', Symbol='Acss1', Aliases='1110032O15Rik, AI788978, Acas2, Acas2l, AceCS2', description='acyl-CoA synthetase short-chain family member 1', other_designations='acetate--CoA ligase 2|acetyl-CoA synthetase 2', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150618111, end_position_on_the_genomic_accession=150668240, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68742, CurrentID=0, Status='live', Symbol='Tmem219', Aliases='1110032O16Rik, 2900045G02Rik, IGFBP-3R', description='transmembrane protein 219', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126886219, end_position_on_the_genomic_accession=126928680, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68743, CurrentID=0, Status='live', Symbol='Anln', Aliases='1110037A17Rik, 2900037I21Rik, C78101, Scraps', description='anillin, actin binding protein', other_designations='anillin, actin binding protein (scraps homolog, Drosophila)', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22331214, end_position_on_the_genomic_accession=22389696, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  68744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68744, CurrentID=0, Status='live', Symbol='Zfp740', Aliases='1110034O07Rik, AW548228, Znf740', description='zinc finger protein 740', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102203645, end_position_on_the_genomic_accession=102215610, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  330890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330890, CurrentID=0, Status='live', Symbol='Piwil4', Aliases='9230101H05Rik, Miwi2', description='piwi-like RNA-mediated gene silencing 4', other_designations='mAgo5|piwi-like 4|piwi-like homolog 4', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14701782, end_position_on_the_genomic_accession=14740765, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  68750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68750, CurrentID=0, Status='live', Symbol='Rreb1', Aliases='1110037N09Rik, AA414966, B930013M22Rik', description='ras responsive element binding protein 1', other_designations='RAS-responsive zinc finger transcription factor RREB|RREB-1', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=37778382, end_position_on_the_genomic_accession=37952005, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  68753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68753, CurrentID=0, Status='live', Symbol='Mybphl', Aliases='1110037P11Rik', description='myosin binding protein H-like', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108364911, end_position_on_the_genomic_accession=108380057, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68755, CurrentID=0, Status='live', Symbol='Cgrrf1', Aliases='1110038G02Rik, 1810009H17Rik, CGR19', description='cell growth regulator with ring finger domain 1', other_designations='cell growth regulatory gene 19 protein', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=46832093, end_position_on_the_genomic_accession=46854200, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68758, CurrentID=0, Status='live', Symbol='Abhd11', Aliases='1110054D16Rik, A630008N09Rik, Wbscr21', description='abhydrolase domain containing 11', other_designations='Williams Beuren syndrome chromosome region 21|Williams-Beuren syndrome chromosomal region 21 protein homolog|Williams-Beuren syndrome critical region 21|abhydrolase domain-containing protein 11', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135009152, end_position_on_the_genomic_accession=135012175, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68760, CurrentID=0, Status='live', Symbol='Synpo2l', Aliases='1110054M18Rik, Chap', description='synaptopodin 2-like', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20658945, end_position_on_the_genomic_accession=20668349, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68763, CurrentID=0, Status='live', Symbol='1110038B12Rik', Aliases='0610009C20Rik, 0910001G04Rik', description='RIKEN cDNA 1110038B12 gene', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34950235, end_position_on_the_genomic_accession=34952471, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  330908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330908, CurrentID=0, Status='live', Symbol='Opcml', Aliases='2900075O15Rik, 3732419F12, AI844366, B930023M13Rik, C230027C17, Gm181, Obcam', description='opioid binding protein/cell adhesion molecule-like', other_designations='opioid binding cell adhesion molecule', map_location='9 13.73 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=27790952, end_position_on_the_genomic_accession=28925410, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68767, CurrentID=0, Status='live', Symbol='Wash1', Aliases='1110049F14Rik, ORF19, Wash', description='WAS protein family homolog 1', other_designations='', map_location='17 F3-F4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=66111546, end_position_on_the_genomic_accession=66120503, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68768, CurrentID=0, Status='live', Symbol='Krtap4-6', Aliases='1110054P19Rik', description='keratin associated protein 4-6', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99665043, end_position_on_the_genomic_accession=99665960, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  330914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330914, CurrentID=0, Status='live', Symbol='Arhgap32', Aliases='3426406O18Rik, Gc-gap, Grit, Px-rics, Rics, mKIAA0712, p200Rhogap, p250Gap', description='Rho GTPase activating protein 32', other_designations='GAB-associated CDC42|GAB-associated Cdc42/Rac GTPase-activating protein|GTPase-activating protein for Cdc42 and Rac1|Rho GTPase activating protein 33 pseudogene|brain-specific Rho GTPase-activating protein|rac GTPase activating protein|rho-type GTPase-activating protein 32|rho/Cdc42/Rac GTPase-activating protein RICS|rhoGAP involved in the beta-catenin-N-cadherin and NMDA receptor signaling', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=32016882, end_position_on_the_genomic_accession=32268443, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  68774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68774, CurrentID=0, Status='live', Symbol='Ms4a6d', Aliases='1110058E16Rik, AI447446, Ms4a11', description='membrane-spanning 4-domains, subfamily A, member 6D', other_designations='CD20 antigen-like 8|membrane-spanning 4-domains, subfamily A, member 11', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11586606, end_position_on_the_genomic_accession=11604804, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68775, CurrentID=0, Status='live', Symbol='Atp6v1c2', Aliases='1110038G14Rik', description='ATPase, H+ transporting, lysosomal V1 subunit C2', other_designations='ATPase, H+ transporting, V1 subunit C|V-ATPase subunit C 2|vacuolar proton pump subunit C 2', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=17284719, end_position_on_the_genomic_accession=17325371, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  68776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68776, CurrentID=0, Status='live', Symbol='Taf11', Aliases='1110038O14Rik, 28kDa, AI115401, AL024207, AW552398, PRO214, TAF2I, TAFII28', description='TATA-box binding protein associated factor 11', other_designations='TAF(II)28|TAF11 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAFII-28|TFIID subunit p30-beta|transcription initiation factor TFIID 28 kDa subunit', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27901128, end_position_on_the_genomic_accession=27909242, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68777, CurrentID=0, Status='live', Symbol='Tmem53', Aliases='1110038M16Rik, 1500001P22Rik, 1700012A05Rik', description='transmembrane protein 53', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117251951, end_position_on_the_genomic_accession=117268588, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68778, CurrentID=0, Status='live', Symbol='Gucd1', Aliases='1110038D17Rik', description='guanylyl cyclase domain containing 1', other_designations='guanylyl cyclase domain-containing protein 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75506814, end_position_on_the_genomic_accession=75517967, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68790, CurrentID=0, Status='live', Symbol='Fendrr', Aliases='1110050K14Rik, LL66, linc-Foxf1a, lincFoxf1', description='Foxf1 adjacent non-coding developmental regulatory RNA', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121054882, end_position_on_the_genomic_accession=121083110, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68792, CurrentID=0, Status='live', Symbol='Srpx2', Aliases='1110039C07Rik, SRP, SRPUL', description='sushi-repeat-containing protein, X-linked 2', other_designations='sushi-repeat containing protein', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=133908416, end_position_on_the_genomic_accession=133932448, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68794, CurrentID=0, Status='live', Symbol='Flnc', Aliases='1110055E19Rik, ABP-280, ABPL, Fln2', description='filamin C, gamma', other_designations='ABP-280-like protein|actin binding protein 280|actin-binding protein 280|actin-binding-like protein|filamin, gamma|filamin-2|gamma-filamin', map_location='6 12.36 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29433127, end_position_on_the_genomic_accession=29461888, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  68795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68795, CurrentID=0, Status='live', Symbol='Ubr3', Aliases='1110059H15Rik, 4833421P10Rik, A130030D10Rik, AA409735, AA414972, AA422631, AI646861, AU016126, Zfp650, Znf650', description='ubiquitin protein ligase E3 component n-recognin 3', other_designations='N-recognin-3|ubiquitin-protein ligase E3-alpha-3|ubiquitin-protein ligase E3-alpha-III|zinc finger protein 650', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69896969, end_position_on_the_genomic_accession=70024010, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  68796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68796, CurrentID=0, Status='live', Symbol='Tmem214', Aliases='1110039B18Rik, 4921530J21Rik, AI841294, AU018742, C76280', description='transmembrane protein 214', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30869647, end_position_on_the_genomic_accession=30877469, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  68797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68797, CurrentID=0, Status='live', Symbol='Pdgfrl', Aliases='1110039P19Rik, AV013190', description='platelet-derived growth factor receptor-like', other_designations='PDGFR-like protein', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40926233, end_position_on_the_genomic_accession=40990782, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68799, CurrentID=0, Status='live', Symbol='Rgmb', Aliases='1110059F19Rik, DRAGON', description='repulsive guidance molecule family member B', other_designations='DRG11-responsive axonal guidance and outgrowth of neurite|RGM domain family, member B|repulsive guidance molecule B', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15804661, end_position_on_the_genomic_accession=15831329, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68800, CurrentID=0, Status='live', Symbol='Prr32', Aliases='1110059M19Rik', description='proline rich 32', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=45090904, end_position_on_the_genomic_accession=45092791, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68801, CurrentID=0, Status='live', Symbol='Elovl5', Aliases='1110059L23Rik, AI747313, AU043003, HELO1', description='ELOVL family member 5, elongation of long chain fatty acids (yeast)', other_designations='3-keto acyl-CoA synthase Elovl5|ELOVL FA elongase 5|ELOVL fatty acid elongase 5|fatty acid elongase 1|homolog of yeast long chain polyunsaturated fatty acid elongation enzyme|very long chain 3-ketoacyl-CoA synthase 5|very long chain 3-oxoacyl-CoA synthase 5|very-long-chain 3-oxoacyl-CoA synthase 5', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=77917365, end_position_on_the_genomic_accession=77984519, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68802, CurrentID=0, Status='live', Symbol='Mypn', Aliases='1110056A04Rik, 9830112J03, AI853556, mKIAA4170', description='myopalladin', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63115795, end_position_on_the_genomic_accession=63203952, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  330953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330953, CurrentID=0, Status='live', Symbol='Hcn4', Aliases='Bcng3, Hcn3', description='hyperpolarization-activated, cyclic nucleotide-gated K+ 4', other_designations='BCNG-3|brain cyclic nucleotide-gated channel 3|hyperpolarization-activated, cyclic nucleotide-gated K+ 3', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58823512, end_position_on_the_genomic_accession=58860955, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68810, CurrentID=0, Status='live', Symbol='Nexn', Aliases='1110046H09Rik, AA553326, NELIN', description='nexilin', other_designations='F-actin binding protein|F-actin-binding protein|nF actin binding protein|nexilin alpha|nexilin beta|nexilin gamma|nexn', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152234785, end_position_on_the_genomic_accession=152266349, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  68813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68813, CurrentID=0, Status='live', Symbol='Dock5', Aliases='1110060D06Rik, AI666732, AI956923, BC016533, E130320D18, lr2, rlc', description='dedicator of cytokinesis 5', other_designations='dock5|lens rupture 2|lens rupture protein 2|rupture of lens cataract protein', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=67751928, end_position_on_the_genomic_accession=67933581, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  68815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68815, CurrentID=0, Status='live', Symbol='Btbd10', Aliases='1110056N09Rik, Gmrp1', description='BTB (POZ) domain containing 10', other_designations='K+ channel tetramerization protein|glucose metabolism-related protein 1', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=113315631, end_position_on_the_genomic_accession=113370531, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  68816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68816, CurrentID=0, Status='live', Symbol='Ppil1', Aliases='1110060O10Rik, AI327391, Cypl1', description='peptidylprolyl isomerase (cyclophilin)-like 1', other_designations='PPIase|rotamase PPIL1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29250835, end_position_on_the_genomic_accession=29263971, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68817, CurrentID=0, Status='live', Symbol='Ddi2', Aliases='1110056G13Rik, 1700027M01Rik, 9130022E05Rik, AI604911, AU040698', description='DNA-damage inducible protein 2', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141683563, end_position_on_the_genomic_accession=141723419, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68818, CurrentID=0, Status='live', Symbol='Zfand2b', Aliases='1110060O18Rik, C81256', description='zinc finger, AN1 type domain 2B', other_designations='AIRAP-like protein|arsenite-inducible RNA-associated protein-like protein', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75168646, end_position_on_the_genomic_accession=75171626, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  330963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330963, CurrentID=0, Status='live', Symbol='Ankdd1a', Aliases='A730067D02, A730067D02Rik, EG330963, Gm1473', description='ankyrin repeat and death domain containing 1A', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65488464, end_position_on_the_genomic_accession=65529899, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  68828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68828, CurrentID=0, Status='live', Symbol='Sync', Aliases='1110057H03Rik, SNIP4', description='syncoilin', other_designations='syncoilin intermediate filament 1|syncoilin-1', map_location='4 63.26 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129287536, end_position_on_the_genomic_accession=129309386, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68832, CurrentID=0, Status='live', Symbol='Ldah', Aliases='1110057K04Rik, mLDAH', description='lipid droplet associated hydrolase', other_designations='LD-associated hydrolase|UPF0554 protein C2orf43 homolog|lipid droplet-associated serine hydrolase', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8208100, end_position_on_the_genomic_accession=8299566, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68833, CurrentID=0, Status='live', Symbol='Pdcl3', Aliases='1110061A19Rik, C80025, Viaf1', description='phosducin-like 3', other_designations='IAP-associated factor VIAF1|VIAF-1|viral IAP-associated factor 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=38987814, end_position_on_the_genomic_accession=38997236, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  68836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68836, CurrentID=0, Status='live', Symbol='Mrpl52', Aliases='1110047B07Rik', description='mitochondrial ribosomal protein L52', other_designations='L52mt|MRP-L52', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54426909, end_position_on_the_genomic_accession=54429750, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68837, CurrentID=0, Status='live', Symbol='Foxk2', Aliases='1110054H05Rik, 5730434B08Rik, 6230415M23Rik, ILF, Ilf1', description='forkhead box K2', other_designations='cellular transcription factor ILF-1|foxk2|interleukin enhancer binding factor 1|interleukin enhancer-binding factor 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121259987, end_position_on_the_genomic_accession=121309896, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68839, CurrentID=0, Status='live', Symbol='Ankrd46', Aliases='1110054N06Rik, AI314978, AI987733', description='ankyrin repeat domain 46', other_designations='ANK-S|ankyrin repeat small protein', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36477668, end_position_on_the_genomic_accession=36503426, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68842, CurrentID=0, Status='live', Symbol='Tulp4', Aliases='1110057P05Rik, 2210038L17Rik, Tusp', description='tubby like protein 4', other_designations='tubby super-family protein|tubby superfamily protein|tubby-like protein 4', map_location='17 3.72 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6106827, end_position_on_the_genomic_accession=6240637, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  68845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68845, CurrentID=0, Status='live', Symbol='Pih1d1', Aliases='1110061L23Rik, 4933413A04Rik, Nop17', description='PIH1 domain containing 1', other_designations='nucleolar protein 17 homolog', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45153835, end_position_on_the_genomic_accession=45160065, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  68846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68846, CurrentID=0, Status='live', Symbol='Rnf208', Aliases='1110061N23Rik', description='ring finger protein 208', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25242929, end_position_on_the_genomic_accession=25244261, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68852, CurrentID=0, Status='live', Symbol='Lrrn4cl', Aliases='1110067I12Rik', description='LRRN4 C-terminal like', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8850194, end_position_on_the_genomic_accession=8853914, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68854, CurrentID=0, Status='live', Symbol='Asb11', Aliases='1110067L12Rik, 1600009D24Rik', description='ankyrin repeat and SOCS box-containing 11', other_designations='ASB-11|ankyrin repeat and SOCS box-containing protein 11', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164436957, end_position_on_the_genomic_accession=164459170, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68857, CurrentID=0, Status='live', Symbol='Dtwd2', Aliases='1190002H09Rik, 8030470C17Rik, AI414673, BB115449', description='DTW domain containing 2', other_designations='', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=49696145, end_position_on_the_genomic_accession=49755667, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68859, CurrentID=0, Status='live', Symbol='Smim1', Aliases='0610011H04Rik, 1190007F08Rik, 2700009I11Rik', description='small integral membrane protein 1', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154020470, end_position_on_the_genomic_accession=154026044, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  331004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331004, CurrentID=0, Status='live', Symbol='Slc9a9', Aliases='5730527A11Rik, 9930105B05, AI854429, Nhe9', description='solute carrier family 9 (sodium/hydrogen exchanger), member 9', other_designations='Na(+)/H(+) exchanger 9|solute carrier family 9 member 9', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=94669892, end_position_on_the_genomic_accession=95230452, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  68861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68861, CurrentID=0, Status='live', Symbol='1190002N15Rik', Aliases='GoPro49, HASF', description='RIKEN cDNA 1190002N15 gene', other_designations='Golgi Protein of 49 kDa|Golgi protein GoPro49|UPF0672 protein C3orf58 homolog|hypoxia and AKT-induced stem cell factor', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=94517864, end_position_on_the_genomic_accession=94538081, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  68865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68865, CurrentID=0, Status='live', Symbol='Arv1', Aliases='1110067L22Rik, AI461928, AW121084', description='ARV1 homolog, fatty acid homeostasis modulator', other_designations='ARV1 homolog', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124722138, end_position_on_the_genomic_accession=124740648, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68867, CurrentID=0, Status='live', Symbol='Rnf122', Aliases='1110063C11Rik', description='ring finger protein 122', other_designations='ring finger protein 122 homolog', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=31111666, end_position_on_the_genomic_accession=31131758, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68870, CurrentID=0, Status='live', Symbol='Ak8', Aliases='1190002A17Rik, AK 8', description='adenylate kinase 8', other_designations='ATP-AMP transphosphorylase 8|Adenylate kinase isoenzyme 4, mitochondrial (ATP-AMP transphosphorylase)|putative adenylate kinase-like protein C9orf98 homolog', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28698226, end_position_on_the_genomic_accession=28813165, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  68874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68874, CurrentID=0, Status='live', Symbol='Klhdc9', Aliases='1190002J23Rik, AA087357, ESTM31', description='kelch domain containing 9', other_designations='', map_location='1 79.36 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171358449, end_position_on_the_genomic_accession=171360798, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68875, CurrentID=0, Status='live', Symbol='Tmcc2', Aliases='1110063G11Rik', description='transmembrane and coiled-coil domains 2', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132356315, end_position_on_the_genomic_accession=132391807, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68876, CurrentID=0, Status='live', Symbol='Xrcc6bp1', Aliases='1110068E08Rik, 2410012H02Rik', description='XRCC6 binding protein 1', other_designations='XRCC6-binding protein 1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=126868427, end_position_on_the_genomic_accession=126901711, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68877, CurrentID=0, Status='live', Symbol='Maf1', Aliases='1110068E11Rik, AU042856', description='MAF1 homolog, negative regulator of RNA polymerase III', other_designations='MAF1 homolog', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76351294, end_position_on_the_genomic_accession=76354378, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68879, CurrentID=0, Status='live', Symbol='Prpf6', Aliases='1190003A07Rik, 2610031L17Rik, ANT-1, U5-102K', description='pre-mRNA splicing factor 6', other_designations='PRP6 homolog|PRP6 pre-mRNA splicing factor 6 homolog|U5 snRNP-associated 102 kDa protein|U5-102 kDa protein', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181601319, end_position_on_the_genomic_accession=181655661, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  331026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331026, CurrentID=0, Status='live', Symbol='Gmppb', Aliases='AI317178, E430010H19', description='GDP-mannose pyrophosphorylase B', other_designations='GTP-mannose-1-phosphate guanylyltransferase beta', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108049229, end_position_on_the_genomic_accession=108051936, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68888, CurrentID=0, Status='live', Symbol='Gkn3', Aliases='1190003M12Rik', description='gastrokine 3', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87383319, end_position_on_the_genomic_accession=87388935, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68889, CurrentID=0, Status='live', Symbol='Ubac2', Aliases='1190008A14Rik, AW554959, Phgdhl1', description='ubiquitin associated domain containing 2', other_designations='UBA domain-containing protein 2|phosphoglycerate dehydrogenase like 1|phosphoglycerate dehydrogenase-like protein 1', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=121878556, end_position_on_the_genomic_accession=122021035, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68891, CurrentID=0, Status='live', Symbol='Cd177', Aliases='1190003K14Rik, Pdp3', description='CD177 antigen', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24743983, end_position_on_the_genomic_accession=24760334, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  68895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68895, CurrentID=0, Status='live', Symbol='Rasl11a', Aliases='1110065D03Rik, AI852031', description='RAS-like, family 11, member A', other_designations='rasl11a', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146845071, end_position_on_the_genomic_accession=146847726, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68897, CurrentID=0, Status='live', Symbol='Disp1', Aliases='1190008H24Rik, DispA', description='dispatched homolog 1 (Drosophila)', other_designations='dispatched A|mdispa|twelve-pass transmembrane protein', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=183086264, end_position_on_the_genomic_accession=183221536, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  331046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331046, CurrentID=0, Status='live', Symbol='Tgm4', Aliases='9530008N10Rik, Eapa1, TGP', description='transglutaminase 4 (prostate)', other_designations='TGase-4|experimental autoimmune prostatitis antigen 1|tgm4|transglutaminase-4', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123034741, end_position_on_the_genomic_accession=123067558, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  68904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68904, CurrentID=0, Status='live', Symbol='Abhd13', Aliases='1110065L07Rik, AI463703, AI788994', description='abhydrolase domain containing 13', other_designations='abhydrolase domain-containing protein 13', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=9977717, end_position_on_the_genomic_accession=9992155, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  68910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68910, CurrentID=0, Status='live', Symbol='Zfp467', Aliases='1190001I08Rik, AB041616, AI426952, AV017623, EZI, MNCb-3350, Znf467', description='zinc finger protein 467', other_designations='endothelial cell-derived zinc finger protein|zinc finger protein EZI', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48427692, end_position_on_the_genomic_accession=48445998, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  68911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68911, CurrentID=0, Status='live', Symbol='Pygo2', Aliases='1190004M21Rik', description='pygopus 2', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89430132, end_position_on_the_genomic_accession=89435131, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68915, CurrentID=0, Status='live', Symbol='Vars2', Aliases='1190004I24Rikl, mKIAA1885, Vars2', description='valyl-tRNA synthetase 2, mitochondrial', other_designations='valRS|valyl-tRNA synthetase 2, mitochondrial (putative)|valyl-tRNA synthetase, mitochondrial', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35655634, end_position_on_the_genomic_accession=35667639, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  68916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68916, CurrentID=0, Status='live', Symbol='Cdkal1', Aliases='1190005B03Rik, 6620401C13Rik', description='CDK5 regulatory subunit associated protein 1-like 1', other_designations='CDK5 regulatory subunit-associated protein 1-like 1|tRNA-t(6)A37 methylthiotransferase', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=29191751, end_position_on_the_genomic_accession=29855684, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  68917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68917, CurrentID=0, Status='live', Symbol='Hint2', Aliases='1190005L05Rik', description='histidine triad nucleotide binding protein 2', other_designations='HINT-2|HINT-3', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43654227, end_position_on_the_genomic_accession=43656486, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68918, CurrentID=0, Status='live', Symbol='1190005I06Rik', Aliases='', description='RIKEN cDNA 1190005I06 gene', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120608602, end_position_on_the_genomic_accession=120634382, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  331063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331063, CurrentID=0, Status='live', Symbol='Gsdmc2', Aliases='9130023H10, AI987692', description='gasdermin C2', other_designations='gsdmc2', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=63824328, end_position_on_the_genomic_accession=63845196, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68922, CurrentID=0, Status='live', Symbol='Dnaic1', Aliases='1110066F04Rik, BB124644, Dnai1, b2b1526Clo', description='dynein, axonemal, intermediate chain 1', other_designations='axonemal dynein intermediate chain 1', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41569663, end_position_on_the_genomic_accession=41638158, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  68925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68925, CurrentID=0, Status='live', Symbol='Rpap1', Aliases='1190005L06Rik, A730023M06Rik, AU041788, AW107702, mKIAA1403', description='RNA polymerase II associated protein 1', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119763959, end_position_on_the_genomic_accession=119787537, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  68926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68926, CurrentID=0, Status='live', Symbol='Ubap2', Aliases='1190005K07Rik, AA408600, AU045235, UBAP-2, mKIAA1491', description='ubiquitin-associated protein 2', other_designations='MLig-1|protein lingerer homolog 1|ubap2', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41194313, end_position_on_the_genomic_accession=41275321, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  68927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68927, CurrentID=0, Status='live', Symbol='Ptcd2', Aliases='1190005P08Rik', description='pentatricopeptide repeat domain 2', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=99319649, end_position_on_the_genomic_accession=99344678, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68929, CurrentID=0, Status='live', Symbol='Mospd3', Aliases='1190005J19Rik, 5133401H10Rik, Gtig2, R124', description='motile sperm domain containing 3', other_designations='gene trap integration gene 2', map_location='5 76.59 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137596645, end_position_on_the_genomic_accession=137601083, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68936, CurrentID=0, Status='live', Symbol='Smim11', Aliases='1190017O12Rik, C21orf51, Fam165ba, Smim11', description='small integral membrane protein 11', other_designations='family with sequence similarity 165, member B|protein FAM165B', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92301303, end_position_on_the_genomic_accession=92313041, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68938, CurrentID=0, Status='live', Symbol='Aspscr1', Aliases='1190006K01Rik, ASPC, ASPCR1, ASPL, ASPS, RCC17, TUG', description='alveolar soft part sarcoma chromosome region, candidate 1 (human)', other_designations='alveolar soft part sarcoma chromosomal region candidate gene 1 protein homolog|tether, containing a UBX domain, for GLUT4', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120672439, end_position_on_the_genomic_accession=120709447, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  68939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68939, CurrentID=0, Status='live', Symbol='Rasl11b', Aliases='1190017B18Rik', description='RAS-like, family 11, member B', other_designations='rasl11b', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=74195326, end_position_on_the_genomic_accession=74199477, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68942, CurrentID=0, Status='live', Symbol='Chmp2b', Aliases='1190006E07Rik', description='charged multivesicular body protein 2B', other_designations='chromatin modifying protein 2B|chromatin-modifying protein 2b', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=65539133, end_position_on_the_genomic_accession=65562697, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  68943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68943, CurrentID=0, Status='live', Symbol='Pink1', Aliases='1190006F07Rik, AU042772, AW557854, BRPK, mFLJ00387', description='PTEN induced putative kinase 1', other_designations='PTEN-induced putative kinase protein 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138313409, end_position_on_the_genomic_accession=138326296, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68944, CurrentID=0, Status='live', Symbol='Tmco1', Aliases='1190006A08Rik, 4930403O06Rik, AA109065, AU022572, ESTM39', description='transmembrane and coiled-coil domains 1', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167308670, end_position_on_the_genomic_accession=167333978, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68947, CurrentID=0, Status='live', Symbol='Chst8', Aliases='1500011J21Rik, AI426009', description='carbohydrate (N-acetylgalactosamine 4-0) sulfotransferase 8', other_designations='GalNAc4ST-1|N-acetylgalactosamine-4-O-sulfotransferase 1|galNAc-4-O-sulfotransferase 1|galNAc-4-ST1', map_location='7 B3-B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34674468, end_position_on_the_genomic_accession=34812711, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68948, CurrentID=0, Status='live', Symbol='Fam216a', Aliases='1500011H22Rik', description='family with sequence similarity 216, member A', other_designations='uncharacterized protein C12orf24 homolog', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122364584, end_position_on_the_genomic_accession=122371963, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  68949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68949, CurrentID=0, Status='live', Symbol='Zfos1', Aliases='1500012F01Rik, Zfas1', description='zinc finger, NFX1-type containing 1, opposite strand RNA 1', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167062934, end_position_on_the_genomic_accession=167065862, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68952, CurrentID=0, Status='live', Symbol='Fam57b', Aliases='1500016O10Rik, A330104J06Rik, AI413816, AW060769', description='family with sequence similarity 57, member B', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126812688, end_position_on_the_genomic_accession=126830219, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68953, CurrentID=0, Status='live', Symbol='Chmp2a', Aliases='1500016L11Rik', description='charged multivesicular body protein 2A', other_designations='chromatin modifying protein 2A|chromatin-modifying protein 2a|mVps2|vacuolar protein sorting-associated protein 2|vacuolar protein-sorting-associated protein 2', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13032006, end_position_on_the_genomic_accession=13034777, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  68955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68955, CurrentID=0, Status='live', Symbol='Srrm4', Aliases='1500001A10Rik, B230202K19Rik, bv, mKIAA1853, nSR100', description='serine/arginine repetitive matrix 4', other_designations='neural-specific SR-related protein of 100 kDa|neural-specific serine/arginine repetitive splicing factor of 100 kDa', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=116439272, end_position_on_the_genomic_accession=116592562, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  68957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68957, CurrentID=0, Status='live', Symbol='Paqr6', Aliases='1500001B10Rik', description='progestin and adipoQ receptor family member VI', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88364072, end_position_on_the_genomic_accession=88368541, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68961, CurrentID=0, Status='live', Symbol='Phkg2', Aliases='1500017I02Rik', description='phosphorylase kinase, gamma 2 (testis)', other_designations='PHK-gamma-LT|PHK-gamma-T|phosphorylase b kinase gamma catalytic chain, testis/liver isoform|phosphorylase kinase subunit gamma-2|serine/threonine-protein kinase PHKG2', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127573344, end_position_on_the_genomic_accession=127583307, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68964, CurrentID=0, Status='live', Symbol='Ctc1', Aliases='1500010J02Rik, AAF-132, AAF132', description='CTS telomere maintenance complex component 1', other_designations='alpha accessory factor 132|alpha-accessory factor of 132 kDa|conserved telomere maintenance component 1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69015904, end_position_on_the_genomic_accession=69036473, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  68966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68966, CurrentID=0, Status='live', Symbol='Ngdn', Aliases='1500001L15Rik, Ngd, neuroguidin', description='neuroguidin, EIF4E binding protein', other_designations='EIF4E-binding protein', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55015454, end_position_on_the_genomic_accession=55024138, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68968, CurrentID=0, Status='live', Symbol='Cdan1', Aliases='1500015A01Rik, AI448026, AW492297, CDA-I, CDA1, CDAI', description='congenital dyserythropoietic anemia, type I (human)', other_designations='congenital dyserythropoietic anemia type I', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120716154, end_position_on_the_genomic_accession=120731517, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  68969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68969, CurrentID=0, Status='live', Symbol='Eif1b', Aliases='1500010M16Rik, 3110001N14Rik, Gc20', description='eukaryotic translation initiation factor 1B', other_designations='translation factor sui1 homolog', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120492606, end_position_on_the_genomic_accession=120495327, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68970, CurrentID=0, Status='live', Symbol='Dcaf12', Aliases='1500001L20Rik, 5830424K06Rik, AA420338, AI851081, Wdr40a', description='DDB1 and CUL4 associated factor 12', other_designations='WD repeat domain 40A|WD repeat-containing protein 40A', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41291300, end_position_on_the_genomic_accession=41316726, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  68971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68971, CurrentID=0, Status='live', Symbol='Tamm41', Aliases='1500001M20Rik, TAM41', description='TAM41, mitochondrial translocator assembly and maintenance protein, homolog (S. cerevisiae)', other_designations='CDP-DAG synthase|CDP-diacylglycerol synthase|MMP37-like protein, mitochondrial|mitochondrial translocator assembly and maintenance protein 41 homolog', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115004381, end_position_on_the_genomic_accession=115037908, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  68972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68972, CurrentID=0, Status='live', Symbol='Tatdn3', Aliases='1500010M24Rik', description='TatD DNase domain containing 3', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191045826, end_position_on_the_genomic_accession=191063010, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68975, CurrentID=0, Status='live', Symbol='Med27', Aliases='1500015J03Rik, 2310042P07Rik, AA682045, Crsp8, D2Ertd434e', description='mediator complex subunit 27', other_designations='CRSP complex subunit 8|cofactor required for Sp1 transcriptional activation subunit 8|cofactor required for Sp1 transcriptional activation, subunit 8|transcriptional co-activator CRSP8 homolog', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29346810, end_position_on_the_genomic_accession=29524790, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  68977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68977, CurrentID=0, Status='live', Symbol='Haghl', Aliases='1500017E18Rik, 2810014I23Rik, C330022E15Rik', description='hydroxyacylglutathione hydrolase-like', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25779843, end_position_on_the_genomic_accession=25785669, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  68979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68979, CurrentID=0, Status='live', Symbol='Nol11', Aliases='1500002M01Rik, AU015220, AU019874', description='nucleolar protein 11', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107166660, end_position_on_the_genomic_accession=107189381, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  68980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68980, CurrentID=0, Status='live', Symbol='Wdr53', Aliases='1500002B03Rik, AI848860', description='WD repeat domain 53', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32247178, end_position_on_the_genomic_accession=32257085, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68981, CurrentID=0, Status='live', Symbol='Snrpa1', Aliases='1500015N06Rik', description="small nuclear ribonucleoprotein polypeptide A'", other_designations="U2 snRNP A'|U2 snRNP-A'", map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66060336, end_position_on_the_genomic_accession=66074587, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  68988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68988, CurrentID=0, Status='live', Symbol='Prpf31', Aliases='1500019O16Rik, 2810404O06Rik, AW554706, PRP31, RP11', description='pre-mRNA processing factor 31', other_designations='PRP31 pre-mRNA processing factor 31 homolog|U4/U6 snRNP 61 kDa protein|pre-mRNA-processing factor 31|protein 61K', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3629985, end_position_on_the_genomic_accession=3642485, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  68991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68991, CurrentID=0, Status='live', Symbol='Ssu72', Aliases='1190002E22Rik, 1500011L16Rik, 2610101M12Rik', description='Ssu72 RNA polymerase II CTD phosphatase homolog (yeast)', other_designations='CTD phosphatase SSU72', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155704805, end_position_on_the_genomic_accession=155733879, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68992, CurrentID=0, Status='live', Symbol='Zfp580', Aliases='1500015O20Rik, AI838225, Znf580', description='zinc finger protein 580', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5051532, end_position_on_the_genomic_accession=5053723, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  68995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68995, CurrentID=0, Status='live', Symbol='Mcts1', Aliases='1500019M23Rik, MCT-1', description='malignant T cell amplified sequence 1', other_designations='malignant T cell-amplified sequence 1|multiple copies T-cell malignancies 1', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38600444, end_position_on_the_genomic_accession=38613524, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  68999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68999, CurrentID=0, Status='live', Symbol='Anapc10', Aliases='1500026N15Rik, A830003M23Rik, APC10', description='anaphase promoting complex subunit 10', other_designations='cyclosome subunit 10', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79711755, end_position_on_the_genomic_accession=79777321, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69008, CurrentID=0, Status='live', Symbol='Cab39l', Aliases='1500031K13Rik, 2810425O13Rik, 4930520C08Rik, AA589432, MO2L', description='calcium binding protein 39-like', other_designations='MO25beta|mo25-like protein', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59440964, end_position_on_the_genomic_accession=59548904, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69009, CurrentID=0, Status='live', Symbol='Thap7', Aliases='1810004B07Rik', description='THAP domain containing 7', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17527558, end_position_on_the_genomic_accession=17531512, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69010, CurrentID=0, Status='live', Symbol='Anapc13', Aliases='1810004D07Rik, APC13, SWM1', description='anaphase promoting complex subunit 13', other_designations='cyclosome subunit 13|cylosome subunit 13', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=102626296, end_position_on_the_genomic_accession=102634244, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69017, CurrentID=0, Status='live', Symbol='Prrt2', Aliases='1500031I19Rik, AI195361, DSPB3', description='proline-rich transmembrane protein 2', other_designations='dispanin subfamily B member 3', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127017542, end_position_on_the_genomic_accession=127021226, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69019, CurrentID=0, Status='live', Symbol='Spcs1', Aliases='1810004F21Rik', description='signal peptidase complex subunit 1 homolog (S. cerevisiae)', other_designations='SPase 12 kDa subunit|microsomal signal peptidase 12 kDa subunit', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30999826, end_position_on_the_genomic_accession=31001666, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69020, CurrentID=0, Status='live', Symbol='Zfp707', Aliases='1500031N24Rik, AI451708, AW495315', description='zinc finger protein 707', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75969127, end_position_on_the_genomic_accession=75975865, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69024, CurrentID=0, Status='live', Symbol='Snx15', Aliases='1500032B08Rik, E130013C21Rik', description='sorting nexin 15', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6119399, end_position_on_the_genomic_accession=6133783, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69028, CurrentID=0, Status='live', Symbol='Mitd1', Aliases='1500032H18Rik', description='MIT, microtubule interacting and transport, domain containing 1', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37878890, end_position_on_the_genomic_accession=37890428, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69029, CurrentID=0, Status='live', Symbol='Smdt1', Aliases='1500032L24Rik, Emre', description='single-pass membrane protein with aspartate rich tail 1', other_designations='UPF0466 protein C22orf32 homolog, mitochondrial|single-pass membrane protein with aspartate-rich tail 1, mitochondrial', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82346046, end_position_on_the_genomic_accession=82349062, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69032, CurrentID=0, Status='live', Symbol='Lyzl4', Aliases='1810009N24Rik, AI504271, Lyc4, Lyza', description='lysozyme-like 4', other_designations='lysozyme a|lysozyme-4', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121577841, end_position_on_the_genomic_accession=121641566, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69034, CurrentID=0, Status='live', Symbol='Nupr1l', Aliases='1810010E01Rik, 4930579G22Rik', description='nuclear protein transcriptional regulator 1 like', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129908540, end_position_on_the_genomic_accession=129913789, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69035, CurrentID=0, Status='live', Symbol='Zdhhc3', Aliases='1110020O22Rik, 1810006O10Rik, 2210017C02Rik, DHHC-3, GODZ, Gramp1, Zfp373', description='zinc finger, DHHC domain containing 3', other_designations='GABA-A receptor-associated membrane protein 1|golgi-specific DHHC zinc finger protein|zinc finger DHHC domain-containing protein 3', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123066802, end_position_on_the_genomic_accession=123113205, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  69036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69036, CurrentID=0, Status='live', Symbol='Zg16', Aliases='1810010M01Rik, AI593689, ZG16p', description='zymogen granule protein 16', other_designations='secretory lectin ZG16', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127050156, end_position_on_the_genomic_accession=127052701, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69038, CurrentID=0, Status='live', Symbol='Tmem258', Aliases='1810006K21Rik', description='transmembrane protein 258', other_designations='UPF0197 transmembrane protein C11orf10 homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10204202, end_position_on_the_genomic_accession=10207824, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69046, CurrentID=0, Status='live', Symbol='Isca1', Aliases='1810010A06Rik, Hbld2', description='iron-sulfur cluster assembly 1', other_designations='HESB like domain containing 2|HESB-like domain-containing protein 2|iron sulfur assembly protein IscA|iron-sulfur assembly protein IscA|iron-sulfur cluster assembly 1 homolog', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59755415, end_position_on_the_genomic_accession=59769789, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69047, CurrentID=0, Status='live', Symbol='Atp2c2', Aliases='1810010G06Rik, Spca2, mKIAA0703', description='ATPase, Ca++ transporting, type 2C, member 2', other_designations='ATPase 2C2|Secretory pathway Ca(2+)-ATPase 2|secretory pathway Ca2+-ATPase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119700009, end_position_on_the_genomic_accession=119757718, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  69048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69048, CurrentID=0, Status='live', Symbol='Slc30a5', Aliases='1810010K08Rik, AF233321, ZNT5, ZTL1, ZnT-5, Zntl1', description='solute carrier family 30 (zinc transporter), member 5', other_designations='solute carrier family 30 member 5|zinc transporter like 1', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100802645, end_position_on_the_genomic_accession=100833563, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  69049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69049, CurrentID=0, Status='live', Symbol='Nat8f5', Aliases='1810018F03Rik', description='N-acetyltransferase 8 (GCN5-related) family member 5', other_designations='camello-like 5|camello-like protein 5|cml5', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85817218, end_position_on_the_genomic_accession=85820972, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  331195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331195, CurrentID=0, Status='live', Symbol='A430089I19Rik', Aliases='Gm16522', description='RIKEN cDNA A430089I19 gene', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=94302519, end_position_on_the_genomic_accession=94306796, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69053, CurrentID=0, Status='live', Symbol='1810013L24Rik', Aliases='1110017P05Rik, 3010002C02Rik, AA792997, AW124177, Gm19906, R74903', description='RIKEN cDNA 1810013L24 gene', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8830100, end_position_on_the_genomic_accession=8858924, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  208117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208117, CurrentID=0, Status='live', Symbol='Aph1b', Aliases='2310057K14Rik, Aph1c', description='aph1 homolog B, gamma secretase subunit', other_designations='APH-1b|APH1B homolog|anterior pharynx defective 1b homolog|anterior pharynx defective 1c homolog|aph-1beta', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66775202, end_position_on_the_genomic_accession=66797293, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69060, CurrentID=0, Status='live', Symbol='Pnlip', Aliases='1810007A24Rik, PL, PTL', description='pancreatic lipase', other_designations='pancreatic triglyceride lipase', map_location='19 54.46 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58670365, end_position_on_the_genomic_accession=58681788, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  69064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69064, CurrentID=0, Status='live', Symbol='Fuom', Aliases='Fucu, Le51', description='fucose mutarotase', other_designations='protein fucU homolog', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140097815, end_position_on_the_genomic_accession=140103070, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69065, CurrentID=0, Status='live', Symbol='Chac1', Aliases='1810008K03Rik', description='ChaC, cation transport regulator 1', other_designations='ChaC, cation transport regulator-like 1|blocks Notch protein|botch|cation transport regulator-like protein 1|gamma-GCG 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119351242, end_position_on_the_genomic_accession=119354326, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69066, CurrentID=0, Status='live', Symbol='1810010H24Rik', Aliases='OTTMUSG00000003581', description='RIKEN cDNA 1810010H24 gene', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107023268, end_position_on_the_genomic_accession=107030446, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69068, CurrentID=0, Status='live', Symbol='1810011O10Rik', Aliases='1110065B09Rik, AW121743, AW321058', description='RIKEN cDNA 1810011O10 gene', other_designations='thyroid cancer-1', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24437616, end_position_on_the_genomic_accession=24438946, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69069, CurrentID=0, Status='live', Symbol='1810011H11Rik', Aliases='4733401I05Rik', description='RIKEN cDNA 1810011H11 gene', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32785885, end_position_on_the_genomic_accession=32817969, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69071, CurrentID=0, Status='live', Symbol='Tmem97', Aliases='1810014L12Rik, AI115531, AL022956, D11Bhm182e', description='transmembrane protein 97', other_designations='', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78541817, end_position_on_the_genomic_accession=78550735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69072, CurrentID=0, Status='live', Symbol='Ebna1bp2', Aliases='1810014B19Rik, AA537058, B830003A16Rik, C81321, Ebp2, Nobp, p40', description='EBNA1 binding protein 2', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118620799, end_position_on_the_genomic_accession=118627776, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69073, CurrentID=0, Status='live', Symbol='Kdf1', Aliases='1810019J16Rik', description='keratinocyte differentiation factor 1', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133518963, end_position_on_the_genomic_accession=133530790, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69076, CurrentID=0, Status='live', Symbol='Triap1', Aliases='1810015M01Rik, AU020874, AU043831, P53csv, Wf-1', description='TP53 regulated inhibitor of apoptosis 1', other_designations='p53-inducible cell-survival factor|protein 15E1.1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115341247, end_position_on_the_genomic_accession=115343552, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69077, CurrentID=0, Status='live', Symbol='Psmd11', Aliases='1700089D09Rik, 1810019E17Rik, 2610024G20Rik, 2810055C24Rik, C78232, P44.5, S9', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 11', other_designations='26S proteasome regulatory subunit RPN6|26S proteasome regulatory subunit S9|26S proteasome regulatory subunit p44.5', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80428615, end_position_on_the_genomic_accession=80473248, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  69080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69080, CurrentID=0, Status='live', Symbol='Gmppa', Aliases='1810012N01Rik', description='GDP-mannose pyrophosphorylase A', other_designations='GTP-mannose-1-phosphate guanylyltransferase alpha', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75435912, end_position_on_the_genomic_accession=75443176, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  69082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69082, CurrentID=0, Status='live', Symbol='Zc3h15', Aliases='1700006A17Rik, 1810012H02Rik, 2610312B22Rik, FM22, Ierepo4', description='zinc finger CCCH-type containing 15', other_designations='DRG family-regulatory protein 1|epo-immediate response gene protein FM22', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=83644578, end_position_on_the_genomic_accession=83664617, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69083, CurrentID=0, Status='live', Symbol='Sult1c2', Aliases='1810008N17Rik, ST1C1, ST1C2, Sult1c1', description='sulfotransferase family, cytosolic, 1C, member 2', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=53829637, end_position_on_the_genomic_accession=53845958, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69085, CurrentID=0, Status='live', Symbol='Zcchc9', Aliases='1810019C21Rik', description='zinc finger, CCHC domain containing 9', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=91796520, end_position_on_the_genomic_accession=91807997, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69089, CurrentID=0, Status='live', Symbol='Oxa1l', Aliases='1810020M02Rik, AI894287', description='oxidase assembly 1-like', other_designations='OXA1-like protein|oxidase assembly 1-like protein', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54360841, end_position_on_the_genomic_accession=54369669, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69090, CurrentID=0, Status='live', Symbol='Ascc1', Aliases='1810015P09Rik, AI550520, ASC1p50, CGI-18', description='activating signal cointegrator 1 complex subunit 1', other_designations='ASC-1 complex subunit p50|trip4 complex subunit p50', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60002493, end_position_on_the_genomic_accession=60099990, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  69091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69091, CurrentID=0, Status='live', Symbol='Vps26b', Aliases='1810012I05Rik, 2310075A12Rik, AI848392', description='vacuolar protein sorting 26 homolog B (yeast)', other_designations='vesicle protein sorting 26B', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=27004502, end_position_on_the_genomic_accession=27030094, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69094, CurrentID=0, Status='live', Symbol='Tmem160', Aliases='1810008O21Rik', description='transmembrane protein 160', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16452779, end_position_on_the_genomic_accession=16455490, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69097, CurrentID=0, Status='live', Symbol='Trim15', Aliases='1810012B10Rik', description='tripartite motif-containing 15', other_designations='tripartite motif protein 15', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36860690, end_position_on_the_genomic_accession=36868085, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69101, CurrentID=0, Status='live', Symbol='Ydjc', Aliases='1810015A11Rik, 4930521M19Rik', description='YdjC homolog (bacterial)', other_designations='UPF0249 protein ydjC homolog', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17146882, end_position_on_the_genomic_accession=17148857, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69104, CurrentID=0, Status='live', Symbol='March5', Aliases='1810015H18Rik, 2310008I22Rik, 2700055A20Rik, 5730499H23Rik, E130202O05Rik, MARCH-V, MITOL, Marchv, Rnf153', description='membrane-associated ring finger (C3HC4) 5', other_designations='membrane-associated RING finger protein 5|membrane-associated RING-CH protein V|ring finger protein 153', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37207545, end_position_on_the_genomic_accession=37224457, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69106, CurrentID=0, Status='live', Symbol='Stoml1', Aliases='1810015E19Rik, AI847431, SLP1, STORP, UNC24, WPB72', description='stomatin-like 1', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58253164, end_position_on_the_genomic_accession=58262524, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69109, CurrentID=0, Status='live', Symbol='Fam58b', Aliases='1810009O10Rik, Fam58a', description='family with sequence similarity 58, member B', other_designations='CDK10-activating cyclin|cyclin M|cyclin-M|cyclin-related protein FAM58A', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78750506, end_position_on_the_genomic_accession=78751729, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69113, CurrentID=0, Status='live', Symbol='Alkbh3', Aliases='1700108H04Rik, 1810020C19Rik, Abh3, mABH3', description='alkB homolog 3, alpha-ketoglutarate-dependent dioxygenase', other_designations='alkB, alkylation repair homolog 3|alkylated DNA repair protein alkB homolog 3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93980634, end_position_on_the_genomic_accession=94010823, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  69116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69116, CurrentID=0, Status='live', Symbol='Ubr4', Aliases='1810009A16Rik, A930005E13Rik, D930005K06Rik, Gm1032, Gm1666, N28143, RBAF600, Zubr1, mKIAA0462, p600', description='ubiquitin protein ligase E3 component n-recognin 4', other_designations='N-recognin-4|retinoblastoma-associated factor 600|zinc finger UBR1-type protein 1|zinc finger, UBR1 type 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139380659, end_position_on_the_genomic_accession=139489533, orientation='plus', exon_count=106, OMIM='', no_hdr0=''),
  69117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69117, CurrentID=0, Status='live', Symbol='Adh6a', Aliases='1810009C16Rik, Adh5a', description='alcohol dehydrogenase 6A (class V)', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138313286, end_position_on_the_genomic_accession=138331134, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69121, CurrentID=0, Status='live', Symbol='Chrdl2', Aliases='1810022C01Rik, Chl2', description='chordin-like 2', other_designations='chordin like 2', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100006404, end_position_on_the_genomic_accession=100034728, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  69123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69123, CurrentID=0, Status='live', Symbol='Eci3', Aliases='1810022C23Rik, Kseg2', description='enoyl-Coenzyme A delta isomerase 3', other_designations='D3,D2-enoyl-CoA isomerase|delta(3),delta(2)-enoyl-CoA isomerase|dodecenoyl-CoA isomerase|kidney specific expressed protein 2', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34946614, end_position_on_the_genomic_accession=34963884, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  69125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69125, CurrentID=0, Status='live', Symbol='Cnot8', Aliases='1500015I04Rik, 1810022F04Rik, AA536816, AU015770, AU043059', description='CCR4-NOT transcription complex, subunit 8', other_designations='CCR4-associated factor 8', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58092799, end_position_on_the_genomic_accession=58118596, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  69126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69126, CurrentID=0, Status='live', Symbol='1810022K09Rik', Aliases='AI746582, AL023036', description='RIKEN cDNA 1810022K09 gene', other_designations='', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14606289, end_position_on_the_genomic_accession=14611256, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69129, CurrentID=0, Status='live', Symbol='Pex11g', Aliases='1810022F11Rik, 1810049N02Rik, Pex11camma, Pex11g', description='peroxisomal biogenesis factor 11 gamma', other_designations='PEX11-gamma|peroxin-11C|peroxisomal biogenesis factor 11c|protein PEX11 homolog gamma', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3457105, end_position_on_the_genomic_accession=3467839, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69131, CurrentID=0, Status='live', Symbol='Cdk12', Aliases='1810022J16Rik, AI646528, Crk7, Crkrs, D11Ertd752e, Pksc', description='cyclin-dependent kinase 12', other_designations='CDC2-related kinase 7|CDC2-related kinase, arginine/serine-rich|CDC2-related protein kinase 7|cell division cycle 2-related protein kinase 7|cell division protein kinase 12|protein kinase for splicing component', map_location='11 61.75 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98202326, end_position_on_the_genomic_accession=98281101, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  69134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69134, CurrentID=0, Status='live', Symbol='Fam25c', Aliases='2200001I15Rik, Fam25', description='family with sequence similarity 25, member C', other_designations='protein FAM25', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34351882, end_position_on_the_genomic_accession=34355393, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69136, CurrentID=0, Status='live', Symbol='Tusc1', Aliases='2200001D17Rik, TSG-9', description='tumor suppressor candidate 1', other_designations='tumour suppressor candidate 1', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=93334148, end_position_on_the_genomic_accession=93335511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69137, CurrentID=0, Status='live', Symbol='Vstm5', Aliases='2200002K05Rik', description='V-set and transmembrane domain containing 5', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15238919, end_position_on_the_genomic_accession=15259421, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69142, CurrentID=0, Status='live', Symbol='Cd209f', Aliases='1810029C22Rik, AI647041, SIGNR8', description='CD209f antigen', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4102784, end_position_on_the_genomic_accession=4105838, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69146, CurrentID=0, Status='live', Symbol='Gsdmd', Aliases='1810036L03Rik, AW558049, DF5L, Dfna5lc1, M2-4, Gsdmd', description='gasdermin D', other_designations='gasdermin domain containing 1|gasdermin domain-containing protein 1', map_location='15 D3-E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75862142, end_position_on_the_genomic_accession=75867408, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  69147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69147, CurrentID=0, Status='live', Symbol='2200002J24Rik', Aliases='', description='RIKEN cDNA 2200002J24 gene', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30698886, end_position_on_the_genomic_accession=30710296, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69149, CurrentID=0, Status='live', Symbol='Kbtbd3', Aliases='2200003A07Rik, Bklhd3', description='kelch repeat and BTB (POZ) domain containing 3', other_designations='BTB and kelch domain containing 3|BTB and kelch domain-containing protein 3', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=4309743, end_position_on_the_genomic_accession=4331829, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69150, CurrentID=0, Status='live', Symbol='Snx4', Aliases='1810036H14Rik, AI037066', description='sorting nexin 4', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33251456, end_position_on_the_genomic_accession=33299562, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  69151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69151, CurrentID=0, Status='live', Symbol='Lzic', Aliases='1810030J04Rik, AW047580', description='leucine zipper and CTNNBIP1 domain containing', other_designations='leucine zipper and CTNNBIP1 domain-containing protein|leucine zipper and ICAT homologous domain-containing protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149485210, end_position_on_the_genomic_accession=149496668, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69155, CurrentID=0, Status='live', Symbol='1810030O07Rik', Aliases='', description='RIKEN cDNA 1810030O07 gene', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=12654879, end_position_on_the_genomic_accession=12674940, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69156, CurrentID=0, Status='live', Symbol='Comtd1', Aliases='1810030M08Rik, MT773', description='catechol-O-methyltransferase domain containing 1', other_designations='O-methyltransferase-containing protein|catechol-O-methyltransferase domain-containing protein 1', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21845861, end_position_on_the_genomic_accession=21848910, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69159, CurrentID=0, Status='live', Symbol='Rhebl1', Aliases='1810036J22Rik, Rheb2', description='Ras homolog enriched in brain like 1', other_designations='Ras homolog enriched in brain-like 1|ras homolog enriched in brain-like protein 1|rheb-like protein 1', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98877760, end_position_on_the_genomic_accession=98881497, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69161, CurrentID=0, Status='live', Symbol='Manbal', Aliases='1810024K12Rik, AI836500', description='mannosidase, beta A, lysosomal-like', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157367594, end_position_on_the_genomic_accession=157396763, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69162, CurrentID=0, Status='live', Symbol='Sec31a', Aliases='1810024J13Rik, ABP125, ABP130, AU042160, HSPC275, Sec31l1', description='Sec31 homolog A (S. cerevisiae)', other_designations='SEC31-like 1|SEC31-like protein 1|SEC31-related protein A', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100361649, end_position_on_the_genomic_accession=100416234, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  69163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69163, CurrentID=0, Status='live', Symbol='Mrpl44', Aliases='1810030E18Rik, 5730593H20Rik, AI463302', description='mitochondrial ribosomal protein L44', other_designations='L44mt|MRP-L44', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=79776018, end_position_on_the_genomic_accession=79781445, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69165, CurrentID=0, Status='live', Symbol='Cd209b', Aliases='1810030I22Rik, DC-SIGNR1, OtB7, SIGNR1, mSIGNR1', description='CD209b antigen', other_designations='DC-SIGN-related protein 1', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3917165, end_position_on_the_genomic_accession=3926841, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  69168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69168, CurrentID=0, Status='live', Symbol='Bola1', Aliases='1810037G04Rik', description='bolA-like 1 (E. coli)', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96196583, end_position_on_the_genomic_accession=96200408, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69169, CurrentID=0, Status='live', Symbol='Fcmr', Aliases='1810037B05Rik, Faim3, FcmuR, Toso', description='Fc fragment of IgM receptor', other_designations='IgM Fc fragment receptor|regulator of Fas-induced apoptosis Toso', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130865777, end_position_on_the_genomic_accession=130880790, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69171, CurrentID=0, Status='live', Symbol='Cnppd1', Aliases='1810031K17Rik, AU043960', description='cyclin Pas1/PHO80 domain containing 1', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75135215, end_position_on_the_genomic_accession=75142733, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  69178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69178, CurrentID=0, Status='live', Symbol='Snx5', Aliases='0910001N05Rik, 1810032P22Rik, AU019504, D2Ertd52e', description='sorting nexin 5', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144250123, end_position_on_the_genomic_accession=144270902, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69179, CurrentID=0, Status='live', Symbol='Tmem110', Aliases='1810038N08Rik, 2310014H19Rik, AW554125, STIMATE', description='transmembrane protein 110', other_designations='STIM-activating enhancer encoded by TMEM110', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30825594, end_position_on_the_genomic_accession=30877210, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69181, CurrentID=0, Status='live', Symbol='Dyrk2', Aliases='1810038L18Rik', description='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 2', other_designations='dyrk2', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=118855602, end_position_on_the_genomic_accession=118889246, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69183, CurrentID=0, Status='live', Symbol='C1qtnf2', Aliases='1810033K05Rik, CTRP2', description='C1q and tumor necrosis factor related protein 2', other_designations='', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43474284, end_position_on_the_genomic_accession=43491525, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69185, CurrentID=0, Status='live', Symbol='Dtwd1', Aliases='1810033A06Rik', description='DTW domain containing 1', other_designations='', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126152113, end_position_on_the_genomic_accession=126165277, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69186, CurrentID=0, Status='live', Symbol='Tmem256', Aliases='1810027O10Rik, 3110009M16Rik', description='transmembrane protein 256', other_designations='UPF0451 protein C17orf61 homolog', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69838525, end_position_on_the_genomic_accession=69839558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69187, CurrentID=0, Status='live', Symbol='Erp27', Aliases='1810033M07Rik, 1810047B09Rik', description='endoplasmic reticulum protein 27', other_designations='ER protein 27|endoplasmic reticulum resident protein ERp27', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136907314, end_position_on_the_genomic_accession=136922461, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69188, CurrentID=0, Status='live', Symbol='Kmt2e', Aliases='1810033J14Rik, 9530077A04Rik, D230038D11Rik, Mll5, NKp44L', description='lysine (K)-specific methyltransferase 2E', other_designations='histone-lysine N-methyltransferase MLL5|lysine N-methyltransferase 2E|myeloid/lymphoid or mixed-lineage leukemia 5|myeloid/lymphoid or mixed-lineage leukemia protein 5 homolog', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23434429, end_position_on_the_genomic_accession=23504229, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  69189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69189, CurrentID=0, Status='live', Symbol='Mcemp1', Aliases='1810033B17Rik', description='mast cell expressed membrane protein 1', other_designations='mast cell-expressed membrane protein 1 homolog', map_location='8 A1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3665762, end_position_on_the_genomic_accession=3668917, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69190, CurrentID=0, Status='live', Symbol='Dym', Aliases='1810041M12Rik, 4933427L07Rik, C030019K18Rik', description='dymeclin', other_designations='', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=75018699, end_position_on_the_genomic_accession=75286966, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  69191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69191, CurrentID=0, Status='live', Symbol='Pdia2', Aliases='1810041F13Rik, AI661267, Pdip, Pdipl', description='protein disulfide isomerase associated 2', other_designations='PDIp|pdia2|protein disulfide isomerase (pancreas) like|protein disulfide isomerase, pancreatic', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26195999, end_position_on_the_genomic_accession=26199132, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  69192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69192, CurrentID=0, Status='live', Symbol='Dhx16', Aliases='2410006N22Rik, DBP2, Ddx16, mKIAA0577', description='DEAH (Asp-Glu-Ala-His) box polypeptide 16', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 16', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35879764, end_position_on_the_genomic_accession=35892670, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  69195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69195, CurrentID=0, Status='live', Symbol='Tmem121', Aliases='2410008J05Rik, Hole', description='transmembrane protein 121', other_designations='protein hole', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113185903, end_position_on_the_genomic_accession=113189523, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69202, CurrentID=0, Status='live', Symbol='Ptms', Aliases='2610009E16Rik', description='parathymosin', other_designations='', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124913675, end_position_on_the_genomic_accession=124917946, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69207, CurrentID=0, Status='live', Symbol='Srsf11', Aliases='0610009J05Rik, 2610019N13Rik, BF642805, Sfrs11', description='serine/arginine-rich splicing factor 11', other_designations='splicing factor, arginine/serine-rich 11', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=158010493, end_position_on_the_genomic_accession=158036650, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  69215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69215, CurrentID=0, Status='live', Symbol='Sat2', Aliases='2610016A03Rik, SSAT-2, SSAT2', description='spermidine/spermine N1-acetyl transferase 2', other_designations='polyamine N-acetyltransferase 2|spermidine/spermine N(1)-acetyltransferase 2', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69621993, end_position_on_the_genomic_accession=69624034, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69216, CurrentID=0, Status='live', Symbol='Svbp', Aliases='2410005K17Rik, AI851162, Ccdc23', description='small vasohibin binding protein', other_designations='coiled coil domain-containing protein 23|coiled-coil domain containing 23|coiled-coil domain-containing protein 23', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119195310, end_position_on_the_genomic_accession=119201298, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69217, CurrentID=0, Status='live', Symbol='Plekha4', Aliases='2410005C22Rik, C88169, Pepp1', description='pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 4', other_designations='PEPP-1|PH domain-containing family A member 4|phosphoinositol 3-phosphate binding protein-1|phosphoinositol 3-phosphate-binding protein 1|pleckstrin homology domain-containing, family A member 4', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45526330, end_position_on_the_genomic_accession=45554229, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  69219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69219, CurrentID=0, Status='live', Symbol='Ddah1', Aliases='2410006N07Rik, 2510015N06Rik, AI987801, AW050362', description='dimethylarginine dimethylaminohydrolase 1', other_designations='DDAH-1|DDAHI|dimethylargininase-1', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145758692, end_position_on_the_genomic_accession=145894277, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69221, CurrentID=0, Status='live', Symbol='2410006H16Rik', Aliases='', description='RIKEN cDNA 2410006H16 gene', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62602877, end_position_on_the_genomic_accession=62604806, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69225, CurrentID=0, Status='live', Symbol='Carkd', Aliases='0710008K08Rik, 2810407E01Rik', description='carbohydrate kinase domain containing', other_designations='ATP-dependent NAD(P)HX dehydratase|carbohydrate kinase domain-containing protein', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11497506, end_position_on_the_genomic_accession=11513287, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69226, CurrentID=0, Status='live', Symbol='Snx24', Aliases='2810011K15Rik, 5730433I16Rik, AI451341', description='sorting nexing 24', other_designations='', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=53244514, end_position_on_the_genomic_accession=53390837, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69227, CurrentID=0, Status='live', Symbol='Selt', Aliases='2810407C02Rik, 5730408P04Rik', description='selenoprotein T', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=58576658, end_position_on_the_genomic_accession=58593544, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69228, CurrentID=0, Status='live', Symbol='Zfp746', Aliases='2810407L07Rik, AI317225, Znf746', description='zinc finger protein 746', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48062381, end_position_on_the_genomic_accession=48086990, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  331374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331374, CurrentID=0, Status='live', Symbol='Dgkk', Aliases='Gm360', description='diacylglycerol kinase kappa', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=6873484, end_position_on_the_genomic_accession=6948363, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  69232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69232, CurrentID=0, Status='live', Symbol='Qrich1', Aliases='2610028H07Rik', description='glutamine-rich 1', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108517087, end_position_on_the_genomic_accession=108560167, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  69234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69234, CurrentID=0, Status='live', Symbol='Zfp688', Aliases='2810407K09Rik, AI326029', description='zinc finger protein 688', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127418966, end_position_on_the_genomic_accession=127423640, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69237, CurrentID=0, Status='live', Symbol='Gtpbp4', Aliases='2610028C09Rik, Crfg, Gtpbp3, NGB, Nog1', description='GTP binding protein 4', other_designations='GTP binding protein 3|GTP-binding protein NGB|Nucleolar GTP-binding protein 1 (Chronic renal failure gene protein) (GTP-binding protein NGB)|chronic renal failure gene protein|chronic renal failure protein', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=8972464, end_position_on_the_genomic_accession=8996012, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  69239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69239, CurrentID=0, Status='live', Symbol='Pdzph1', Aliases='2610034M16Rik, AV114557', description='PDZ and pleckstrin homology domains 1', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=58878808, end_position_on_the_genomic_accession=58991375, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  69241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69241, CurrentID=0, Status='live', Symbol='Polr2d', Aliases='2310002G05Rik, 2610028L19Rik, HSRBP4, RBP4', description='polymerase (RNA) II (DNA directed) polypeptide D', other_designations='DNA-directed RNA polymerase II subunit D|RNA polymerase II subunit B4', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31789159, end_position_on_the_genomic_accession=31796701, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69253, CurrentID=0, Status='live', Symbol='Hspb2', Aliases='27kDa, 2810021G24Rik, HSP27, MKBP', description='heat shock protein 2', other_designations='heat shock 27kD protein 2', map_location='9 27.75 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50751072, end_position_on_the_genomic_accession=50752354, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69256, CurrentID=0, Status='live', Symbol='Zfp397', Aliases='2810411K16Rik, 6720480F11Rik, AI461664', description='zinc finger protein 397', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=23954662, end_position_on_the_genomic_accession=23964671, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69257, CurrentID=0, Status='live', Symbol='Elf2', Aliases='2610036A20Rik, A230104O07Rik, AW111824, BB183398, EU32, NERF, NERF-1A, NERF-1B, NERF-2', description='E74-like factor 2', other_designations='elf2|ets domain transcription factor|ets family transcription factor ELF2A2|new ETS-related factor', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51252720, end_position_on_the_genomic_accession=51340644, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69259, CurrentID=0, Status='live', Symbol='Kctd5', Aliases='2610030N08Rik, mKIAA0176', description='potassium channel tetramerisation domain containing 5', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24047720, end_position_on_the_genomic_accession=24073485, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69260, CurrentID=0, Status='live', Symbol='Ing2', Aliases='2810011M06Rik, Ing1lb, P33ING2, Ing2', description='inhibitor of growth family, member 2', other_designations='inhibitor of growth 1-like protein', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47667178, end_position_on_the_genomic_accession=47675609, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69263, CurrentID=0, Status='live', Symbol='Rfc3', Aliases='2810416I22Rik, 38kDa, AU022547, Recc3', description='replication factor C (activator 1) 3', other_designations='A1 38 kDa subunit|RF-C 38 kDa subunit|RFC38|activator 1 38 kDa subunit|activator 1 subunit 3|replication factor C 3|replication factor C 38 kDa subunit|replication factor C, 38 kDa', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=151642824, end_position_on_the_genomic_accession=151651208, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  11544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11544, CurrentID=0, Status='live', Symbol='Adprh', Aliases='Arh1', description='ADP-ribosylarginine hydrolase', other_designations='ADP-ribose-L-arginine cleaving enzyme|[Protein ADP-ribosylarginine] hydrolase', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38445399, end_position_on_the_genomic_accession=38452689, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69269, CurrentID=0, Status='live', Symbol='Scnm1', Aliases='3110001I17Rik-ps, Scnm1', description='sodium channel modifier 1', other_designations='sodium channel modifier 1, pseudogene', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95129719, end_position_on_the_genomic_accession=95134012, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69270, CurrentID=0, Status='live', Symbol='Gins1', Aliases='2810418N01Rik, Gins4, PSF1, mKIAA0186', description='GINS complex subunit 1 (Psf1 homolog)', other_designations='GINS complex subunit 4 (Sld5 homolog)', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150909557, end_position_on_the_genomic_accession=150931280, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  331416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331416, CurrentID=0, Status='live', Symbol='Gm773', Aliases='', description='predicted gene 773', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56189827, end_position_on_the_genomic_accession=56212886, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69274, CurrentID=0, Status='live', Symbol='Ctdspl', Aliases='2810418J22Rik, AI426263, HYA22, SCP3', description='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase-like', other_designations='CTDSP-like|NIF-like protein|NLI-interacting factor 1|carboxy-terminal domain RNA polymerase II polypeptide A small phosphatase 3|nuclear LIM interactor-interacting factor 1|small C-terminal domain phosphatase 3|small CTD phosphatase 3', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=118926536, end_position_on_the_genomic_accession=119044119, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69276, CurrentID=0, Status='live', Symbol='Sec62', Aliases='3100002M17Rik, AI844545, AW545092, Dtrp1, HTP1, Tloc1', description='SEC62 homolog (S. cerevisiae)', other_designations='TP-1|translocation protein 1', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30792774, end_position_on_the_genomic_accession=30821263, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69277, CurrentID=0, Status='live', Symbol='3300002I08Rik', Aliases='2900016H11Rik', description='RIKEN cDNA 3300002I08 gene', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150306363, end_position_on_the_genomic_accession=150362935, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69281, CurrentID=0, Status='live', Symbol='Spata4', Aliases='1700001N01Rik, Srg2, TSARG2', description='spermatogenesis associated 4', other_designations='spermatogenesis-related gene 2 protein|testis and spermatogenesis cell-related protein 2|testis spermatocyte apoptosis-related gene 2 protein', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=54600781, end_position_on_the_genomic_accession=54610098, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69282, CurrentID=0, Status='live', Symbol='1700001J03Rik', Aliases='', description='RIKEN cDNA 1700001J03 gene', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146182450, end_position_on_the_genomic_accession=146185304, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69286, CurrentID=0, Status='live', Symbol='Glipr1l1', Aliases='1700011E04Rik', description='GLI pathogenesis-related 1 like 1', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=112060189, end_position_on_the_genomic_accession=112078510, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69287, CurrentID=0, Status='live', Symbol='Odf3', Aliases='1700011O04Rik, SHIPPO1', description='outer dense fiber of sperm tails 3', other_designations='outer dense fiber of sperm tails protein 3|sperm tail protein SHIPPO 1', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140847916, end_position_on_the_genomic_accession=140850925, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69288, CurrentID=0, Status='live', Symbol='Rhobtb1', Aliases='1700008H16Rik, 3110048G13Rik, AI173445, AI573858, AV350930', description='Rho-related BTB domain containing 1', other_designations='', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=69151279, end_position_on_the_genomic_accession=69291791, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  69294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69294, CurrentID=0, Status='live', Symbol='Cst13', Aliases='1700006C19Rik, Cymg1, Mtcg1', description='cystatin 13', other_designations='cst13|cystatin T|cystatin-T', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148820099, end_position_on_the_genomic_accession=148830410, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69295, CurrentID=0, Status='live', Symbol='Far1os', Aliases='1700006P19Rik', description='fatty acyl CoA reductase 1, opposite strand', other_designations='Far1 opposite strand transcript', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=113538026, end_position_on_the_genomic_accession=113564987, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69297, CurrentID=0, Status='live', Symbol='Lrrc46', Aliases='1700006D24Rik', description='leucine rich repeat containing 46', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97034602, end_position_on_the_genomic_accession=97041369, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69299, CurrentID=0, Status='live', Symbol='Asb9', Aliases='1700011M07Rik', description='ankyrin repeat and SOCS box-containing 9', other_designations='ASB-9|ankyrin repeat and SOCS box-containing protein 9', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164497903, end_position_on_the_genomic_accession=164539752, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69301, CurrentID=0, Status='live', Symbol='Tescl', Aliases='1700008P20Rik', description='tescalcin-like', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24333080, end_position_on_the_genomic_accession=24333965, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69305, CurrentID=0, Status='live', Symbol='Dcps', Aliases='1700001E16Rik, AA408441', description='decapping enzyme, scavenger', other_designations='DCS-1|HINT-5|decapping scavenger enzyme|hint-related 7meGMP-directed hydrolase|histidine triad nucleotide-binding protein 5|histidine triad protein member 5|mRNA decapping enzyme|scavenger mRNA-decapping enzyme DcpS', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35124414, end_position_on_the_genomic_accession=35175987, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69306, CurrentID=0, Status='live', Symbol='Efcab9', Aliases='1700007I06Rik, 4930403C08Rik', description='EF-hand calcium binding domain 9', other_designations='', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32522733, end_position_on_the_genomic_accession=32533271, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69307, CurrentID=0, Status='live', Symbol='Pxt1', Aliases='1700001G18Rik, Stepp', description='peroxisomal, testis specific 1', other_designations='small testis-specific peroxisomal protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28933986, end_position_on_the_genomic_accession=28942262, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69309, CurrentID=0, Status='live', Symbol='Slc16a13', Aliases='1700007D07Rik', description='solute carrier family 16 (monocarboxylic acid transporters), member 13', other_designations='MCT 13|solute carrier family 16 member 13', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70216789, end_position_on_the_genomic_accession=70222794, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69310, CurrentID=0, Status='live', Symbol='Pacrg', Aliases='1700008H23Rik', description='PARK2 co-regulated', other_designations='PARK2 coregulated gene protein|hypertension-related protein 1-like protein|parkin co-regulated gene protein', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=10403012, end_position_on_the_genomic_accession=10840191, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69312, CurrentID=0, Status='live', Symbol='Ppp1r42', Aliases='1700011J18Rik, 4930418G15Rik, Lrrc67, TLRR', description='protein phosphatase 1, regulatory subunit 42', other_designations='leucine rich repeat containing 67|testis leucine-rich repeat protein', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9968622, end_position_on_the_genomic_accession=10013541, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  69314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69314, CurrentID=0, Status='live', Symbol='Izumo3', Aliases='1700011H22Rik', description='IZUMO family member 3', other_designations='transmembrane protein C9orf134', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=92144318, end_position_on_the_genomic_accession=92147221, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69315, CurrentID=0, Status='live', Symbol='1700001L19Rik', Aliases='', description='RIKEN cDNA 1700001L19 gene', other_designations='Y regulated sperm protein', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=68597439, end_position_on_the_genomic_accession=68614231, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  331461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331461, CurrentID=0, Status='live', Symbol='Il1rapl1', Aliases='6330532G10Rik, C030039N24, IL1R8, IL1RAPL, MRX34, OPHN4, TIGIRR-2', description='interleukin 1 receptor accessory protein-like 1', other_designations='IL-1-RAPL-1|IL-1RAPL-1|IL1RAPL-1|X-linked interleukin-1 receptor accessory protein-like 1', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=86740867, end_position_on_the_genomic_accession=88115623, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  69318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69318, CurrentID=0, Status='live', Symbol='1700007K09Rik', Aliases='', description='RIKEN cDNA 1700007K09 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131325930, end_position_on_the_genomic_accession=131331434, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69324, CurrentID=0, Status='live', Symbol='1700012B07Rik', Aliases='1700069D21Rik', description='RIKEN cDNA 1700012B07 gene', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109788268, end_position_on_the_genomic_accession=109828079, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69325, CurrentID=0, Status='live', Symbol='1700012B09Rik', Aliases='1700007P19Rik', description='RIKEN cDNA 1700012B09 gene', other_designations='', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14756592, end_position_on_the_genomic_accession=14771037, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69327, CurrentID=0, Status='live', Symbol='1700007K13Rik', Aliases='Rbest47', description='RIKEN cDNA 1700007K13 gene', other_designations='p53-induced expression in RB-null cells protein 1|pierce 1|pierce1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28462001, end_position_on_the_genomic_accession=28466324, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69329, CurrentID=0, Status='live', Symbol='Cfap206', Aliases='1700003M02Rik, AU024269', description='cilia and flagella associated protein 206', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34688574, end_position_on_the_genomic_accession=34730237, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  331474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331474, CurrentID=0, Status='live', Symbol='Rgag4', Aliases='6430402L03Rik, Mar5, Mart5, mKIAA2001', description='retrotransposon gag domain containing 4', other_designations='gag-like protein', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102066542, end_position_on_the_genomic_accession=102071307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69339, CurrentID=0, Status='live', Symbol='Ccdc54', Aliases='1700007N18Rik, AI644412, BB013989, NYD-SP17', description='coiled-coil domain containing 54', other_designations='testes development-related NYD-SP17', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=50589859, end_position_on_the_genomic_accession=50591154, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  331487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331487, CurrentID=0, Status='live', Symbol='Uprt', Aliases='Gm774', description='uracil phosphoribosyltransferase (FUR1) homolog (S. cerevisiae)', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=104482782, end_position_on_the_genomic_accession=104506262, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  331493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331493, CurrentID=0, Status='live', Symbol='Gm5127', Aliases='B130007I08, EG331493', description='predicted gene 5127', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106583151, end_position_on_the_genomic_accession=106710557, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69351, CurrentID=0, Status='live', Symbol='Smim23', Aliases='1700008A04Rik', description='small integral membrane protein 23', other_designations='testis specific expressed protein 4', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32820376, end_position_on_the_genomic_accession=32824594, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69352, CurrentID=0, Status='live', Symbol='Necab1', Aliases='1700003H21Rik, Efcbp1, STIP-1', description='N-terminal EF-hand calcium binding protein 1', other_designations='EF hand calcium binding protein 1|EF-hand calcium binding protein 1|EF-hand calcium-binding protein 1', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=14952245, end_position_on_the_genomic_accession=15149804, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69354, CurrentID=0, Status='live', Symbol='Slc38a4', Aliases='1110012E16Rik, 1700012A18Rik, Ata3, mATA3, mNAT3', description='solute carrier family 38, member 4', other_designations='Na(+)-coupled neutral amino acid transporter 4|amino acid transporter A3|system A amino acid transporter 3|system-N amino acid transporter 3', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=96994816, end_position_on_the_genomic_accession=97055956, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  69358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69358, CurrentID=0, Status='live', Symbol='Lrrc51', Aliases='1700008D07Rik, Lrtomt', description='leucine rich repeat containing 51', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101912989, end_position_on_the_genomic_accession=101933917, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69361, CurrentID=0, Status='live', Symbol='Cypt3', Aliases='1700006H02Rik, Ckt1r1', description='cysteine-rich perinuclear theca 3', other_designations='basic protein CKT1R1|cysteine-rich perinuclear theca protein 3', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153558574, end_position_on_the_genomic_accession=153559436, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69362, CurrentID=0, Status='live', Symbol='Cst12', Aliases='1700006F03Rik, CRES3', description='cystatin 12', other_designations='cystatin TE-1|cystatin-related epididymal spermatogenic protein 3|cystatin-related gene expressed in testis and epididymis', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148789361, end_position_on_the_genomic_accession=148793437, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69363, CurrentID=0, Status='live', Symbol='Spaca4', Aliases='1700008E09Rik, AV043694, Samp14', description='sperm acrosome associated 4', other_designations='sperm acrosomal membrane 14', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45725107, end_position_on_the_genomic_accession=45725816, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69364, CurrentID=0, Status='live', Symbol='1700012P22Rik', Aliases='', description='RIKEN cDNA 1700012P22 gene', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144418189, end_position_on_the_genomic_accession=144438793, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69367, CurrentID=0, Status='live', Symbol='Glrx2', Aliases='1700010P22Rik, AI645710, Grx2', description='glutaredoxin 2 (thioltransferase)', other_designations='', map_location='1 62.53 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=143739224, end_position_on_the_genomic_accession=143749678, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69368, CurrentID=0, Status='live', Symbol='Wdfy1', Aliases='1700013B03Rik, 1700120F24Rik, FENS-1, Jr1, WDF1, ZFYVE17, mKIAA1435', description='WD repeat and FYVE domain containing 1', other_designations='WD40 and FYVE domain containing 1|WD40- and FYVE domain-containing protein 1', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=79702262, end_position_on_the_genomic_accession=79776047, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  69371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69371, CurrentID=0, Status='live', Symbol='Smco2', Aliases='1700023A16Rik', description='single-pass membrane protein with coiled-coil domains 2', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146850093, end_position_on_the_genomic_accession=146871407, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69372, CurrentID=0, Status='live', Symbol='Mocs3', Aliases='1700020H17Rik, Uba4', description='molybdenum cofactor synthesis 3', other_designations='MPT synthase sulfurylase|molybdopterin synthase sulfurylase', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168230622, end_position_on_the_genomic_accession=168232303, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  69376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69376, CurrentID=0, Status='live', Symbol='Zpbp2', Aliases='1700017D11Rik, 2610022C02Rik, ZPBPL', description='zona pellucida binding protein 2', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98551097, end_position_on_the_genomic_accession=98558665, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69379, CurrentID=0, Status='live', Symbol='C8g', Aliases='', description='complement component 8, gamma polypeptide', other_designations='complement component 8, gamma subunit|novel lipocalin protein', map_location='2 17.31 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25498650, end_position_on_the_genomic_accession=25501926, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  331524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331524, CurrentID=0, Status='live', Symbol='Xkrx', Aliases='A630026F04', description='X-linked Kx blood group related, X-linked', other_designations='Kell blood group complex subunit-related, X-linked|X Kell blood group precursor related X linked|x Kell blood group-related, X-linked', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134149045, end_position_on_the_genomic_accession=134185509, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69382, CurrentID=0, Status='live', Symbol='1700024P04Rik', Aliases='H2bl1, Subh2bv', description='RIKEN cDNA 1700024P04 gene', other_designations='histone H2B subacrosomal|histone variant H2bl1', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98984090, end_position_on_the_genomic_accession=98984565, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69384, CurrentID=0, Status='live', Symbol='Tmem89', Aliases='1700024B07Rik', description='transmembrane protein 89', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108914619, end_position_on_the_genomic_accession=108915563, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  331529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331529, CurrentID=0, Status='live', Symbol='Gm5128', Aliases='4930481M05, EG331529', description='predicted gene 5128', other_designations='preferentially expressed antigen in melanoma-like 3', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135373249, end_position_on_the_genomic_accession=135383393, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69386, CurrentID=0, Status='live', Symbol='Hist1h4h', Aliases='1700024H08Rik', description='histone cluster 1, H4h', other_designations='H4 histone family, member A|histone 1, H4h', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23531044, end_position_on_the_genomic_accession=23531478, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  69387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69387, CurrentID=0, Status='live', Symbol='Dnajb13', Aliases='1700014P03Rik, Tsarg', description='DnaJ heat shock protein family (Hsp40) member B13', other_designations='DnaJ (Hsp40) related, subfamily B, member 13|spermatogenesis apoptosis-related protein|testis and spermatogenesis cell-related protein 6|testis spermatocyte apoptosis-related gene 6 protein|testis spermatogenesis apoptosis-related gene 3 protein|testis spermatogenesis apoptosis-related gene 6 protein', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100503011, end_position_on_the_genomic_accession=100514815, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  331532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331532, CurrentID=0, Status='live', Symbol='Tceal5', Aliases='6430401A05Rik', description='transcription elongation factor A (SII)-like 5', other_designations='TCEA-like protein 5|transcription elongation factor S-II protein-like 5', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136200948, end_position_on_the_genomic_accession=136203851, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69389, CurrentID=0, Status='live', Symbol='H2bfm', Aliases='1700014N06Rik, H2bfw, H2bl2', description='H2B histone family, member M', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136927325, end_position_on_the_genomic_accession=136928373, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  331535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331535, CurrentID=0, Status='live', Symbol='Serpina7', Aliases='C730040N12Rik, Tbg', description='serine (or cysteine) peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 7', other_designations='T4-binding globulin|serine (or cysteine) peptidase inhibitor, clade A (alpha-1 antipeptidase, antitrypsin), member 7|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 7|serpin A7', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139079245, end_position_on_the_genomic_accession=139085270, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  331537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331537, CurrentID=0, Status='live', Symbol='Pih1h3b', Aliases='E230019M04Rik', description='PIH1 domain containing 3B', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140062784, end_position_on_the_genomic_accession=140107014, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69396, CurrentID=0, Status='live', Symbol='1700018F24Rik', Aliases='AV045857', description='RIKEN cDNA 1700018F24 gene', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145042990, end_position_on_the_genomic_accession=145045678, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69398, CurrentID=0, Status='live', Symbol='Cdhr4', Aliases='Cdh29', description='cadherin-related family member 4', other_designations='cadherin-like 29', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107992492, end_position_on_the_genomic_accession=107999683, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  69399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69399, CurrentID=0, Status='live', Symbol='1700025G04Rik', Aliases='2610510E02Rik, 2900042O04Rik, ENSMUSG00000073540', description='RIKEN cDNA 1700025G04 gene', other_designations='', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151884524, end_position_on_the_genomic_accession=152090320, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69401, CurrentID=0, Status='live', Symbol='Plac8l1', Aliases='1700009N18Rik, 1700015M15Rik', description='PLAC8-like 1', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42178675, end_position_on_the_genomic_accession=42196709, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69408, CurrentID=0, Status='live', Symbol='Dnajc17', Aliases='1700025B16Rik, C87112, D9Bwg1371e', description='DnaJ heat shock protein family (Hsp40) member C17', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 17', map_location='2 59.97 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119172500, end_position_on_the_genomic_accession=119208795, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  69416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69416, CurrentID=0, Status='live', Symbol='1700025F22Rik', Aliases='', description='RIKEN cDNA 1700025F22 gene', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11139662, end_position_on_the_genomic_accession=11165628, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69428, CurrentID=0, Status='live', Symbol='1700016C15Rik', Aliases='', description='RIKEN cDNA 1700016C15 gene', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=177728122, end_position_on_the_genomic_accession=177753306, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69439, CurrentID=0, Status='live', Symbol='Mroh4', Aliases='1700016M24Rik', description='maestro heat-like repeat family member 4', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74606026, end_position_on_the_genomic_accession=74636868, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  69440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69440, CurrentID=0, Status='live', Symbol='Dennd6b', Aliases='1700027J05Rik, AI414881, Fam116b', description='DENN/MADD domain containing 6B', other_designations='DENN domain-containing protein 6B|family with sequence similarity 116, member B|protein FAM116B', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89182213, end_position_on_the_genomic_accession=89196474, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  69441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69441, CurrentID=0, Status='live', Symbol='1700023F06Rik', Aliases='', description='RIKEN cDNA 1700023F06 gene', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103198943, end_position_on_the_genomic_accession=103208548, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  69444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69444, CurrentID=0, Status='live', Symbol='Lyzl6', Aliases='1700023H08Rik, Lyc1, Lyzb', description='lysozyme-like 6', other_designations='lysozyme b|lysozyme-like protein 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103631069, end_position_on_the_genomic_accession=103638889, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69453, CurrentID=0, Status='live', Symbol='Prss56', Aliases='1700027L20Rik', description='protease, serine 56', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87183094, end_position_on_the_genomic_accession=87188405, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  69454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69454, CurrentID=0, Status='live', Symbol='Clic3', Aliases='2300003G24Rik', description='chloride intracellular channel 3', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25456843, end_position_on_the_genomic_accession=25458776, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69456, CurrentID=0, Status='live', Symbol='Commd10', Aliases='2310003A05Rik, AU019438, DRWMS2', description='COMM domain containing 10', other_designations='down-regulated in W/WV mouse stomach 2|mDRWMS2', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46958785, end_position_on_the_genomic_accession=47087995, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69457, CurrentID=0, Status='live', Symbol='Tmem45a2', Aliases='2310005G13Rik', description='transmembrane protein 45A2', other_designations='', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57036967, end_position_on_the_genomic_accession=57072092, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69459, CurrentID=0, Status='live', Symbol='Ubl7', Aliases='2300004C15Rik', description='ubiquitin-like 7 (bone marrow stromal cell-derived)', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57910980, end_position_on_the_genomic_accession=57929970, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69462, CurrentID=0, Status='live', Symbol='2300005B03Rik', Aliases='SLURP-2, Slurp2', description='RIKEN cDNA 2300005B03 gene', other_designations='secreted Ly6/uPAR related protein 2|secreted Ly6/uPAR related protein-2', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74742839, end_position_on_the_genomic_accession=74746687, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69464, CurrentID=0, Status='live', Symbol='Krtap4-13', Aliases='2300006N05Rik', description='keratin associated protein 4-13', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99809078, end_position_on_the_genomic_accession=99809896, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69469, CurrentID=0, Status='live', Symbol='Tmco2', Aliases='1700028F16Rik', description='transmembrane and coiled-coil domains 2', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=121105651, end_position_on_the_genomic_accession=121109226, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69470, CurrentID=0, Status='live', Symbol='Tmem127', Aliases='2310003P10Rik, AI314202, AI317350', description='transmembrane protein 127', other_designations='', map_location='2 61.87 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127247975, end_position_on_the_genomic_accession=127260764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69473, CurrentID=0, Status='live', Symbol='Krtap3-1', Aliases='2310003K17Rik, KAP3.1, KRTAP3.1', description='keratin associated protein 3-1', other_designations='High-sulfur keratin protein|high-sulfur keratin protein|krtap3-1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99566027, end_position_on_the_genomic_accession=99566630, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69478, CurrentID=0, Status='live', Symbol='2300009A05Rik', Aliases='', description='RIKEN cDNA 2300009A05 gene', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=63394447, end_position_on_the_genomic_accession=63399244, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  331623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=331623, CurrentID=0, Status='live', Symbol='Bend3', Aliases='mKIAA1553', description='BEN domain containing 3', other_designations='', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43479140, end_position_on_the_genomic_accession=43515417, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69480, CurrentID=0, Status='live', Symbol='Ttc9', Aliases='', description='tetratricopeptide repeat domain 9', other_designations='TPR repeat protein 9A', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81629771, end_position_on_the_genomic_accession=81664941, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  208188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208188, CurrentID=0, Status='live', Symbol='Ghsr', Aliases='C530020I22Rik', description='growth hormone secretagogue receptor', other_designations='G-protein coupled receptor|GH-releasing peptide receptor|GHRP|GHS-R|ghrelin receptor', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=27371351, end_position_on_the_genomic_accession=27378010, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69482, CurrentID=0, Status='live', Symbol='Nup35', Aliases='2310006I24Rik, 35kDa, 5330402E05Rik, MP44, NO44', description='nucleoporin 35', other_designations='35 kDa nucleoporin|mitotic phosphoprotein 44|nuclear pore complex protein Nup53', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=80638812, end_position_on_the_genomic_accession=80660071, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69487, CurrentID=0, Status='live', Symbol='Ndufaf5', Aliases='2310003L22Rik', description='NADH dehydrogenase (ubiquinone) complex I, assembly factor 5', other_designations='probable methyltransferase C20orf7 homolog, mitochondrial', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=140170631, end_position_on_the_genomic_accession=140205252, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69496, CurrentID=0, Status='live', Symbol='Dydc1', Aliases='1700029M23Rik', description='DPY30 domain containing 1', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=41072911, end_position_on_the_genomic_accession=41092197, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69499, CurrentID=0, Status='live', Symbol='Tsr2', Aliases='2310007F12Rik', description='TSR2 20S rRNA accumulation', other_designations='TSR2, 20S rRNA accumulation, homolog', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151087094, end_position_on_the_genomic_accession=151096543, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69501, CurrentID=0, Status='live', Symbol='Etd', Aliases='1700030K02Rik', description='embryonic testis differentiation', other_designations='embryonic testis differentiation-associated', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53434918, end_position_on_the_genomic_accession=53443576, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69504, CurrentID=0, Status='live', Symbol='Zfp932', Aliases='2310001H12Rik, Znf431', description='zinc finger protein 932', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=109996527, end_position_on_the_genomic_accession=110010411, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69511, CurrentID=0, Status='live', Symbol='Klk12', Aliases='2310008B01Rik, AI324852, KLK-L5', description='kallikrein related-peptidase 12', other_designations='kallikrein 12', map_location='7 28.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43768664, end_position_on_the_genomic_accession=43774704, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69514, CurrentID=0, Status='live', Symbol='Lce3e', Aliases='2310001H18Rik', description='late cornified envelope 3E', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92967061, end_position_on_the_genomic_accession=92968281, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69519, CurrentID=0, Status='live', Symbol='Rwdd2a', Aliases='1700030C20Rik, AI848608, Rwdd2', description='RWD domain containing 2A', other_designations='RWD domain containing 2', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=86571988, end_position_on_the_genomic_accession=86574899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69520, CurrentID=0, Status='live', Symbol='Lce3f', Aliases='2310002A05Rik', description='late cornified envelope 3F', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92992226, end_position_on_the_genomic_accession=92993426, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69524, CurrentID=0, Status='live', Symbol='Esam', Aliases='2310008D05Rik1, W117m, Esam', description='endothelial cell-specific adhesion molecule', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37528055, end_position_on_the_genomic_accession=37538319, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69527, CurrentID=0, Status='live', Symbol='Mrps9', Aliases='2310002A08Rik, AW322516, Mrp-S9, S9mt, mg637', description='mitochondrial ribosomal protein S9', other_designations='muscle protein 637', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=42851233, end_position_on_the_genomic_accession=42905683, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  69528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69528, CurrentID=0, Status='live', Symbol='1700030J22Rik', Aliases='', description='RIKEN cDNA 1700030J22 gene', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116969594, end_position_on_the_genomic_accession=116979001, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69533, CurrentID=0, Status='live', Symbol='Krtap26-1', Aliases='2310002B14Rik, AI507462', description='keratin associated protein 26-1', other_designations='krtap26-1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88646824, end_position_on_the_genomic_accession=88647796, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69534, CurrentID=0, Status='live', Symbol='Avpi1', Aliases='2310008N12Rik, AA410041, VIT32, esu, mVIT32', description='arginine vasopressin-induced 1', other_designations='AVP-induced protein 1|Esau protein|VIP32|arginine vasopressin-induced transcript 32 protein', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42123275, end_position_on_the_genomic_accession=42128993, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69535, CurrentID=0, Status='live', Symbol='Ten1', Aliases='2310004N24Rik', description='TEN1 telomerase capping complex subunit', other_designations='protein telomeric pathways with STN1 homolog|telomere length regulation protein TEN1 homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116198855, end_position_on_the_genomic_accession=116215318, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69536, CurrentID=0, Status='live', Symbol='Hemk1', Aliases='2310008M14Rik, AW049265', description='HemK methyltransferase family member 1', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107327082, end_position_on_the_genomic_accession=107340411, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  69537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69537, CurrentID=0, Status='live', Symbol='Dnase1l1', Aliases='2310005K03Rik, AI324885, Dnase1ll, Dnasex, Dnl1ll, G4.8', description='deoxyribonuclease 1-like 1', other_designations='DNase I-like 1|DNase X|deoxyribonuclease I-like 1', map_location='X 37.94 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74273217, end_position_on_the_genomic_accession=74282333, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69538, CurrentID=0, Status='live', Symbol='Antxr1', Aliases='2310008J16Rik, 2810405N18Rik, Antrx1, Tem8', description='anthrax toxin receptor 1', other_designations='tumor endothelial marker 8', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87133853, end_position_on_the_genomic_accession=87335809, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  69539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69539, CurrentID=0, Status='live', Symbol='Trnp1', Aliases='2300002D11Rik, Trnp', description='TMF1-regulated nuclear protein 1', other_designations='mexneurin', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133491097, end_position_on_the_genomic_accession=133499137, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69540, CurrentID=0, Status='live', Symbol='Klk10', Aliases='2300002A13Rik, NES1, PRSSL1', description='kallikrein related-peptidase 10', other_designations='kallikrein 10', map_location='7 28.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43781054, end_position_on_the_genomic_accession=43785410, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69541, CurrentID=0, Status='live', Symbol='Lyg1', Aliases='2300002O18Rik', description='lysozyme G-like 1', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37946738, end_position_on_the_genomic_accession=37957759, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69542, CurrentID=0, Status='live', Symbol='2300002M23Rik', Aliases='Stg, emprin', description='RIKEN cDNA 2300002M23 gene', other_designations='extracellular matrix protein with prion homology', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35567485, end_position_on_the_genomic_accession=35568945, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69543, CurrentID=0, Status='live', Symbol='Capns2', Aliases='2310005G05Rik, 30K-2', description='calpain, small subunit 2', other_designations='CSS2|calcium-dependent protease small subunit 2', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=92901395, end_position_on_the_genomic_accession=92902409, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  69544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69544, CurrentID=0, Status='live', Symbol='Wdr5b', Aliases='2310009C03Rik, AI606931', description='WD repeat domain 5B', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36041190, end_position_on_the_genomic_accession=36042974, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  69546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69546, CurrentID=0, Status='live', Symbol='Mapk1ip1', Aliases='2310009E07Rik, AU043776, MISS', description='mitogen-activated protein kinase 1 interacting protein 1', other_designations='mitogen activated protein kinase 1 interacting protein 1|mitogen-activated protein kinase 1-interacting protein 1', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=138835818, end_position_on_the_genomic_accession=138846267, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69547, CurrentID=0, Status='live', Symbol='Nkpd1', Aliases='2310015G09Rik', description='NTPase, KAP family P-loop domain containing 1', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19516236, end_position_on_the_genomic_accession=19525057, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69549, CurrentID=0, Status='live', Symbol='2310009B15Rik', Aliases='', description='RIKEN cDNA 2310009B15 gene', other_designations='', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=138851979, end_position_on_the_genomic_accession=138856854, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69550, CurrentID=0, Status='live', Symbol='Bst2', Aliases='2310015I10Rik, Bst-2, C87040, CD317, DAMP-1, GREG', description='bone marrow stromal cell antigen 2', other_designations='DAMP-1 protein homolog|HM1.24 antigen|tetherin', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71534259, end_position_on_the_genomic_accession=71537437, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69551, CurrentID=0, Status='live', Symbol='2310022B05Rik', Aliases='', description='RIKEN cDNA 2310022B05 gene', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124635756, end_position_on_the_genomic_accession=124663369, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69553, CurrentID=0, Status='live', Symbol='Fam65c', Aliases='2310033K02Rik, Gm124', description='family with sequence similarity 65, member C', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167980161, end_position_on_the_genomic_accession=168050887, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  69554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69554, CurrentID=0, Status='live', Symbol='Klhdc2', Aliases='2310022K15Rik, D12Ertd522e, HCLP-1, LCP', description='kelch domain containing 2', other_designations='endothelial differentiation inhibitory protein TNG', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69296681, end_position_on_the_genomic_accession=69310687, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  69556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69556, CurrentID=0, Status='live', Symbol='Bod1', Aliases='2310022M17Rik, BOD2.1, Fam44b', description='biorientation of chromosomes in cell division 1', other_designations='biorientation defective protein 1|bone length and organs 1|family with sequence similarity 44, member B', map_location='11 18.55 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=31665150, end_position_on_the_genomic_accession=31671862, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69562, CurrentID=0, Status='live', Symbol='Cdk13', Aliases='2310015O17Rik, Cdc2l5', description='cyclin-dependent kinase 13', other_designations='CDC2-related protein kinase 5|cell division cycle 2-like 5 (cholinesterase-related cell division controller)|cell division cycle 2-like protein kinase 5|cell division protein kinase 13', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=17715962, end_position_on_the_genomic_accession=17805332, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  69564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69564, CurrentID=0, Status='live', Symbol='Nmrk2', Aliases='2310015C21Rik, Itgb1bp3, Mibp', description='nicotinamide riboside kinase 2', other_designations='NRK 2|RNK 2|integrin beta 1 binding protein 3|integrin beta-1-binding protein 3|muscle integrin-binding protein|muscle-specific beta 1 integrin binding protein|nicotinic acid riboside kinase 2|nmR-K 2|ribosylnicotinamide kinase 2|ribosylnicotinic acid kinase 2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81198170, end_position_on_the_genomic_accession=81202212, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69568, CurrentID=0, Status='live', Symbol='Vkorc1l1', Aliases='2310024K08Rik', description='vitamin K epoxide reductase complex, subunit 1-like 1', other_designations='VKORC1-like protein 1', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129942109, end_position_on_the_genomic_accession=129986692, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69572, CurrentID=0, Status='live', Symbol='Mfsd3', Aliases='2310010G13Rik', description='major facilitator superfamily domain containing 3', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76701542, end_position_on_the_genomic_accession=76704239, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69573, CurrentID=0, Status='live', Symbol='Hilpda', Aliases='2310016C08Rik, AW049004, Hig2', description='hypoxia inducible lipid droplet associated', other_designations='hypoxia-inducible gene 2 protein', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29272488, end_position_on_the_genomic_accession=29275449, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69574, CurrentID=0, Status='live', Symbol='Cmbl', Aliases='2310016A09Rik', description='carboxymethylenebutenolidase-like (Pseudomonas)', other_designations='', map_location='15 B2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=31565374, end_position_on_the_genomic_accession=31590119, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69576, CurrentID=0, Status='live', Symbol='Smco1', Aliases='2310010M20Rik', description='single-pass membrane protein with coiled-coil domains 1', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32271609, end_position_on_the_genomic_accession=32274779, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69577, CurrentID=0, Status='live', Symbol='Fastkd3', Aliases='2310010B21Rik, AW048587', description='FAST kinase domains 3', other_designations='', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=68582163, end_position_on_the_genomic_accession=68592338, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69581, CurrentID=0, Status='live', Symbol='Rhou', Aliases='2310026M05Rik, AI182090, Arhu, CDC42L1, G28K, WRCH-1, WRCH1, mG28K', description='ras homolog gene family, member U', other_designations='Ras-like protein member U|Wnt1 responsive Cdc42 homolog|rho GTPase-like protein ARHU|rhou|wnt-1 responsive Cdc42 homolog 1', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123653929, end_position_on_the_genomic_accession=123663880, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69582, CurrentID=0, Status='live', Symbol='Plekhm2', Aliases='2310034J19Rik, AI854247, mKIAA0842', description='pleckstrin homology domain containing, family M (with RUN domain) member 2', other_designations='PH domain-containing family M member 2', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141625732, end_position_on_the_genomic_accession=141664151, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  69583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69583, CurrentID=0, Status='live', Symbol='Tnfsf13', Aliases='2310026N09Rik, April, Tall2, Tnlg7b, Trdl1', description='tumor necrosis factor (ligand) superfamily, member 13', other_designations='a proliferation-inducing ligand|tumor necrosis factor ligand 7b', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69682577, end_position_on_the_genomic_accession=69685554, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69585, CurrentID=0, Status='live', Symbol='Hfe2', Aliases='2310035L15Rik, 5230400G09Rik, AI414844, AI789733, DL-M, HJV, Rgmc, hemojuvelin', description='hemochromatosis type 2 (juvenile)', other_designations='RGM domain family member C|hemochromatosis type 2 (juvenile) (human homolog)|hemochromatosis type 2 protein homolog|repulsive guidance molecule C', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96525163, end_position_on_the_genomic_accession=96529216, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69587, CurrentID=0, Status='live', Symbol='Pcgf3', Aliases='2310035N15Rik, AI662857, D630042K08Rik, DONG1, E430039C14, RNF3A, Rnf3', description='polycomb group ring finger 3', other_designations='RING finger protein 3A|ring finger protein 3', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108461215, end_position_on_the_genomic_accession=108503099, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69590, CurrentID=0, Status='live', Symbol='Gpx8', Aliases='2310016C16Rik, AU017063', description='glutathione peroxidase 8 (putative)', other_designations='GPx-8|GSHPx-8', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113042763, end_position_on_the_genomic_accession=113046388, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  69592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69592, CurrentID=0, Status='live', Symbol='Odam', Aliases='2310011G06Rik, APIN', description='odontogenic, ameloblast asssociated', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87885025, end_position_on_the_genomic_accession=87894174, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69596, CurrentID=0, Status='live', Symbol='Ap5s1', Aliases='0610038L13Rik, 2310035K24Rik, AW060727, sigma5', description='adaptor-related protein 5 complex, sigma 1 subunit', other_designations='adapter-related protein complex 5 sigma subunit|adaptor-related protein complex 5 sigma subunit|protein C20orf29 homolog', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131210360, end_position_on_the_genomic_accession=131213514, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69597, CurrentID=0, Status='live', Symbol='Afg3l2', Aliases='2310036I02Rik, AW260507, Emv66, par', description='AFG3-like AAA ATPase 2', other_designations='AFG3(ATPase family gene 3)-like 2', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67404760, end_position_on_the_genomic_accession=67449136, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  69601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69601, CurrentID=0, Status='live', Symbol='Dab2ip', Aliases='2310011D08Rik, AI480459, Aip1, mKIAA1743', description='disabled 2 interacting protein', other_designations='ASK-interacting protein 1|ASK1-interacting protein 1|DAB2-interacting protein|DOC-2/DAB-2 interactive protein|disabled 2 interacting protein long form', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35558458, end_position_on_the_genomic_accession=35730994, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  69602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69602, CurrentID=0, Status='live', Symbol='Otop3', Aliases='2310011E08Rik', description='otopetrin 3', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115334731, end_position_on_the_genomic_accession=115346927, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69605, CurrentID=0, Status='live', Symbol='Lnp', Aliases='2310011O18Rik, 4921514L11Rik, 9530051D01Rik, AI666268, Ul, ulnaless', description='limb and neural patterns', other_designations='lunapark|protein ulnaless', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74514836, end_position_on_the_genomic_accession=74579466, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  69606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69606, CurrentID=0, Status='live', Symbol='Mtfmt', Aliases='2310020P08Rik', description='mitochondrial methionyl-tRNA formyltransferase', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65435562, end_position_on_the_genomic_accession=65453054, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69608, CurrentID=0, Status='live', Symbol='Sec24d', Aliases='2310020L09Rik, Gm1349', description='Sec24 related gene family, member D (S. cerevisiae)', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=123267496, end_position_on_the_genomic_accession=123365641, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  69612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69612, CurrentID=0, Status='live', Symbol='Kansl2', Aliases='2310037I24Rik, B930009D17', description='KAT8 regulatory NSL complex subunit 2', other_designations='NSL complex protein NSL2|non-specific lethal 2 homolog', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98517658, end_position_on_the_genomic_accession=98534732, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  69617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69617, CurrentID=0, Status='live', Symbol='Pitrm1', Aliases='2310012C15Rik, AA410010, MP-1, MP1, Ntup1, PreP, mKIAA1104', description='pitrilysin metallepetidase 1', other_designations='nuclear transplantation upregulated protein 1|pitrilysin metalloprotease 1|pitrilysin metalloproteinase 1', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=6548125, end_position_on_the_genomic_accession=6580158, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  69623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69623, CurrentID=0, Status='live', Symbol='Zfp33b', Aliases='2310012P17Rik', description='zinc finger protein 33B', other_designations='', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104570873, end_position_on_the_genomic_accession=104584620, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69627, CurrentID=0, Status='live', Symbol='Fam89a', Aliases='2310031A18Rik', description='family with sequence similarity 89, member A', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124740257, end_position_on_the_genomic_accession=124751809, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69632, CurrentID=0, Status='live', Symbol='Arhgef12', Aliases='2310014B11Rik, AU019857, Larg, mKIAA0382', description='Rho guanine nucleotide exchange factor (GEF) 12', other_designations='leukemia-associated Rho guanine-nucleotide exchange factor|leukemia-associated RhoGEF', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=42963842, end_position_on_the_genomic_accession=43105718, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  69634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69634, CurrentID=0, Status='live', Symbol='Clybl', Aliases='0610033J05Rik, 2310014M14Rik, AI256068, Clb', description='citrate lyase beta like', other_designations='beta-methylmalate synthase|citrate lyase beta-like|malate synthase', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122171283, end_position_on_the_genomic_accession=122402234, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69635, CurrentID=0, Status='live', Symbol='Dapk1', Aliases='D13Ucla1, DAP-Kinase', description='death associated protein kinase 1', other_designations='DAP kinase 1', map_location='13 32.53 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60601947, end_position_on_the_genomic_accession=60763191, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  69638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69638, CurrentID=0, Status='live', Symbol='Enho', Aliases='1110065P19Rik, 2310040A07Rik', description='energy homeostasis associated', other_designations='energy homeostasis-associated protein', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41638144, end_position_on_the_genomic_accession=41640302, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69639, CurrentID=0, Status='live', Symbol='Exosc8', Aliases='2310032N20Rik, CIP3, mKIAA4013', description='exosome component 8', other_designations='exosome complex exonuclease RRP43|ribosomal RNA-processing protein 43', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=54728679, end_position_on_the_genomic_accession=54735364, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  69640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69640, CurrentID=0, Status='live', Symbol='Fam83g', Aliases='2310040C09Rik, AI506990, wly', description='family with sequence similarity 83, member G', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61684082, end_position_on_the_genomic_accession=61709950, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69641, CurrentID=0, Status='live', Symbol='Wdr20', Aliases='2310040A13Rik, AI549910a, Wdr20', description='WD repeat domain 20', other_designations='WD repeat domain 20A', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110737231, end_position_on_the_genomic_accession=110804283, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69642, CurrentID=0, Status='live', Symbol='Mlip', Aliases='2310046A06Rik, AI931782, CIP', description='muscular LMNA-interacting protein', other_designations='cardiac Isl1-interacting protein|muscular-enriched A-type laminin-interacting protein', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=77102084, end_position_on_the_genomic_accession=77347870, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  69654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69654, CurrentID=0, Status='live', Symbol='Dctn2', Aliases='2310042E05Rik, C130077D06Rik, C87049, DCTN-50, GMP23-48K, RBP50, p50', description='dynactin 2', other_designations='50 kDa dynein-associated polypeptide|dynactin complex 50 kDa subunit|growth cone membrane protein 23-48K|p50 dynamitin', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127266262, end_position_on_the_genomic_accession=127281959, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  142681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=142681, CurrentID=0, Status='live', Symbol='Slc34a3', Aliases='AI649385, Npt2c, Npt2c-v1, NptIIc, naPi-2c', description='solute carrier family 34 (sodium phosphate), member 3', other_designations='Na(+)-dependent phosphate cotransporter 2C|Na(+)/Pi cotransporter 2C|sodium-phosphate cotransporter IIC|sodium/phosphate cotransporter 2C|type IIc Na+/Pi-cotransporter|type IIc NaPi-cotransporter-v1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25228897, end_position_on_the_genomic_accession=25234271, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69656, CurrentID=0, Status='live', Symbol='Pir', Aliases='2310042L19Riknl, Pir', description='pirin', other_designations='iron-binding nuclear protein|pirin like|probable quercetin 2,3-dioxygenase PIR|probable quercetinase', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164269431, end_position_on_the_genomic_accession=164373013, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69660, CurrentID=0, Status='live', Symbol='Tmbim1', Aliases='2310061B02Rik, AA960455, AU024746, C78899, RECS1, Tmbib1, mKIAA4161', description='transmembrane BAX inhibitor motif containing 1', other_designations='responsive to centrifugal force and shear stress gene 1 protein|transmembrane BAX inhibitor motif-containing protein 1', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74288247, end_position_on_the_genomic_accession=74304336, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  11610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11610, CurrentID=0, Status='live', Symbol='Agtrap', Aliases='3300002E14Rik, AT1R, Atrap, D4Wsu124e', description='angiotensin II, type I receptor-associated protein', other_designations='AT1 receptor|AT1 receptor-associated protein|angiotensin II receptor, type I', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148077061, end_position_on_the_genomic_accession=148088064, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69662, CurrentID=0, Status='live', Symbol='2310061I04Rik', Aliases='', description='RIKEN cDNA 2310061I04 gene', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35892677, end_position_on_the_genomic_accession=35897378, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69663, CurrentID=0, Status='live', Symbol='Ddx51', Aliases='2310061O04Rik, AA537002, C81583, R75505', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 51', other_designations='DEAD box protein 51', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110653434, end_position_on_the_genomic_accession=110660496, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  69664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69664, CurrentID=0, Status='live', Symbol='Krtap1-5', Aliases='2310043L02Rik, AI592983', description='keratin associated protein 1-5', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99579977, end_position_on_the_genomic_accession=99581016, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69665, CurrentID=0, Status='live', Symbol='Upk3bl', Aliases='2310043J07Rik', description='uroplakin 3B-like', other_designations='uroplakin-3-like protein', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136057253, end_position_on_the_genomic_accession=136076695, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69666, CurrentID=0, Status='live', Symbol='Psmg4', Aliases='PAC-4, mPAC4', description='proteasome (prosome, macropain) assembly chaperone 4', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34162964, end_position_on_the_genomic_accession=34178172, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69668, CurrentID=0, Status='live', Symbol='Ccdc115', Aliases='2310061I09Rik, Ccp1', description='coiled-coil domain containing 115', other_designations='coiled-coil protein 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34436670, end_position_on_the_genomic_accession=34439672, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69671, CurrentID=0, Status='live', Symbol='Tmem52', Aliases='2310043I08Rik', description='transmembrane protein 52', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155469114, end_position_on_the_genomic_accession=155470858, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69672, CurrentID=0, Status='live', Symbol='Txndc15', Aliases='2310047H23Rik, AI854086, C5orf14, UNQ335', description='thioredoxin domain containing 15', other_designations='disulfide isomerase', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55714650, end_position_on_the_genomic_accession=55726227, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69674, CurrentID=0, Status='live', Symbol='Mif4gd', Aliases='1110014L05Rik, 2310075G12Rik', description='MIF4G domain containing', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115607918, end_position_on_the_genomic_accession=115612973, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69675, CurrentID=0, Status='live', Symbol='Pxdn', Aliases='2310075M15Rik, C85409, E330004E07, VPO1, mKIAA0230', description='peroxidasin', other_designations='', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=29937877, end_position_on_the_genomic_accession=30017658, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  69677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69677, CurrentID=0, Status='live', Symbol='Il1f8', Aliases='2310043N20Rik, Il36b', description='interleukin 1 family, member 8', other_designations='IL-1F8|interleukin-1 family member 8', map_location='2 16.21 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24153172, end_position_on_the_genomic_accession=24160103, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69681, CurrentID=0, Status='live', Symbol='Cdk3-ps', Aliases='2310050O10Rik, Cdk3', description='cyclin-dependent kinase 3, pseudogene', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116216002, end_position_on_the_genomic_accession=116220287, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69683, CurrentID=0, Status='live', Symbol='Emc10', Aliases='2310044H10Rik, 5430410O10Rik, Inm02, Mirta22', description='ER membrane protein complex subunit 10', other_designations='UPF0510 protein INM02|hematopoietic signal peptide-containing membrane domain-containing protein 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44489938, end_position_on_the_genomic_accession=44496524, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69684, CurrentID=0, Status='live', Symbol='Aarsd1', Aliases='1110069E20Rik, 2310044P18Rik, AA589600, AlaX, AlaXp-II', description='alanyl-tRNA synthetase domain containing 1', other_designations='alanyl-tRNA deacylase alaX|alanyl-tRNA synthetase domain-containing protein 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101406665, end_position_on_the_genomic_accession=101417433, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  69692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69692, CurrentID=0, Status='live', Symbol='Hddc2', Aliases='2310057G13Rik', description='HD domain containing 2', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=31313362, end_position_on_the_genomic_accession=31328086, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69693, CurrentID=0, Status='live', Symbol='Pof1b', Aliases='2310066B14Rik', description='premature ovarian failure 1B', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=112638427, end_position_on_the_genomic_accession=112698711, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  69694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69694, CurrentID=0, Status='live', Symbol='Tatdn1', Aliases='2310079P03Rik, CDA11', description='TatD DNase domain containing 1', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58890153, end_position_on_the_genomic_accession=58933781, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  69696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69696, CurrentID=0, Status='live', Symbol='2310057N15Rik', Aliases='Krtap13-1', description='RIKEN cDNA 2310057N15 gene', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88773181, end_position_on_the_genomic_accession=88774206, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69697, CurrentID=0, Status='live', Symbol='Camsap3', Aliases='2310057J16Rik, Kiaa1543, Nezha', description='calmodulin regulated spectrin-associated protein family, member 3', other_designations='marshalin-La|marshalin-Lb|marshalin-Lc|marshalin-Ld|marshalin-Sa|marshalin-Sb|marshalin-Sc|marshalin-Sd|protein Nezha', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3587299, end_position_on_the_genomic_accession=3609075, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  69698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69698, CurrentID=0, Status='live', Symbol='Slc52a3', Aliases='2310046K01Rik, RFT2', description='solute carrier protein family 52, member 3', other_designations='riboflavin transporter 2', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151996511, end_position_on_the_genomic_accession=152009258, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69699, CurrentID=0, Status='live', Symbol='2310079G19Rik', Aliases='', description='RIKEN cDNA 2310079G19 gene', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88626788, end_position_on_the_genomic_accession=88627666, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  69700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69700, CurrentID=0, Status='live', Symbol='Col22a1', Aliases='2310067L16Rik, C80743', description='collagen, type XXII, alpha 1', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=71798476, end_position_on_the_genomic_accession=72150980, orientation='minus', exon_count=68, OMIM='', no_hdr0=''),
  69702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69702, CurrentID=0, Status='live', Symbol='Ndufaf1', Aliases='2410001M24Rik, CGI-65, CIA30', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 1', other_designations='NADH dehydrogenase (ubiquinone) 1 assembly factor 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119655446, end_position_on_the_genomic_accession=119662885, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69706, CurrentID=0, Status='live', Symbol='Lrr1', Aliases='2410005L11Rik, AU021780, LRR-1, Ppil5', description='leucine rich repeat protein 1', other_designations='peptidylprolyl isomerase (cyclophilin) like 5', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69168814, end_position_on_the_genomic_accession=69179010, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69707, CurrentID=0, Status='live', Symbol='Iqcg', Aliases='2400003L07Rik, G1-374-12, esgd12d, repro1, stubby12d', description='IQ motif containing G', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33014270, end_position_on_the_genomic_accession=33056341, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69709, CurrentID=0, Status='live', Symbol='Ptrhd1', Aliases='2410017P09Rik', description='peptidyl-tRNA hydrolase domain containing 1', other_designations='peptidyl-tRNA hydrolase domain-containing protein 1', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4234027, end_position_on_the_genomic_accession=4240123, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  69710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69710, CurrentID=0, Status='live', Symbol='Arap1', Aliases='2410002L19Rik, Centd2, mKIAA0782', description='ArfGAP with RhoGAP domain, ankyrin repeat and PH domain 1', other_designations='arap1|centaurin, delta 2|centaurin-delta-2|cnt-d2', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101348069, end_position_on_the_genomic_accession=101412586, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  69713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69713, CurrentID=0, Status='live', Symbol='Pin4', Aliases='2410002I22Rik, EPVH, Par14', description='protein (peptidyl-prolyl cis/trans isomerase) NIMA-interacting, 4 (parvulin)', other_designations='PPIase Pin4|parvulin-14|peptidyl-prolyl cis-trans isomerase Pin4|rotamase Pin4', map_location='X B2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102119465, end_position_on_the_genomic_accession=102127673, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69714, CurrentID=0, Status='live', Symbol='Tfpt', Aliases='2400004F01Rik, AI450389, Amida, FB1', description='TCF3 (E2A) fusion partner', other_designations='protein amida', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3620324, end_position_on_the_genomic_accession=3629929, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69716, CurrentID=0, Status='live', Symbol='Trip13', Aliases='2410002G23Rik, D13Ertd328e', description='thyroid hormone receptor interactor 13', other_designations='TR-interacting protein 13|TRIP-13|thyroid receptor-interacting protein 13', map_location='13 40.15 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73912462, end_position_on_the_genomic_accession=73937767, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  69717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69717, CurrentID=0, Status='live', Symbol='Gm10499', Aliases='ENSMUSG00000073403', description='predicted gene 10499', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  69718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69718, CurrentID=0, Status='live', Symbol='Ipmk', Aliases='2410017C19Rik, AA408208, Impk', description='inositol polyphosphate multikinase', other_designations='inositol 1,3,4,6-tetrakisphosphate 5-kinase', map_location='10 B5.2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=71347765, end_position_on_the_genomic_accession=71385885, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69719, CurrentID=0, Status='live', Symbol='Cad', Aliases='2410008J01Rik, AU018859, Cpad', description='carbamoyl-phosphate synthetase 2, aspartate transcarbamylase, and dihydroorotase', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31054636, end_position_on_the_genomic_accession=31078479, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  69721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69721, CurrentID=0, Status='live', Symbol='Nkiras1', Aliases='2400004O09Rik, KBRAS1, kappaB-Ras1', description='NFKB inhibitor interacting Ras-like protein 1', other_designations='I-kappa-B-interacting Ras-like protein 1|kappa B-Ras protein 1', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=18271142, end_position_on_the_genomic_accession=18284003, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69723, CurrentID=0, Status='live', Symbol='Rpain', Aliases='2400006N03Rik, Rip', description='RPA interacting protein', other_designations='RPA interacting protein delta 2|RPA interacting protein epsilon', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70970200, end_position_on_the_genomic_accession=70977933, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69724, CurrentID=0, Status='live', Symbol='Rnaseh2a', Aliases='2400006P09Rik, RNASEHI, RNHIA, RNHL', description='ribonuclease H2, large subunit', other_designations='RNase H2 subunit A|RNase HI large subunit|ribonuclease HI large subunit|ribonuclease HI subunit A', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84956610, end_position_on_the_genomic_accession=84969375, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  69726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69726, CurrentID=0, Status='live', Symbol='Smyd3', Aliases='2410008A19Rik, Zmynd1', description='SET and MYND domain containing 3', other_designations='SET and MYND domain-containing protein 3|zinc finger MYND domain-containing protein 1|zinc finger, MYND domain containing 1', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=178951960, end_position_on_the_genomic_accession=179518052, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  69727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69727, CurrentID=0, Status='live', Symbol='Usp46', Aliases='1190009E20Rik, 2410018I08Rik, AI451644', description='ubiquitin specific peptidase 46', other_designations='deubiquitinating enzyme 46|ubiquitin specific protease 46|ubiquitin thioesterase 46|ubiquitin thiolesterase 46|ubiquitin-specific-processing protease 46', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73998726, end_position_on_the_genomic_accession=74068411, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  99003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99003, CurrentID=0, Status='live', Symbol='Qser1', Aliases='4732486I23Rik, AI502974', description='glutamine and serine rich 1', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104754793, end_position_on_the_genomic_accession=104816841, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  69736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69736, CurrentID=0, Status='live', Symbol='Nup37', Aliases='2410003L22Rik, 2810039M17Rik', description='nucleoporin 37', other_designations='nup107-160 subcomplex subunit Nup37', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88146976, end_position_on_the_genomic_accession=88178395, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69737, CurrentID=0, Status='live', Symbol='Ttl', Aliases='2410003M22Rik, 2700049H19Rik, AI848570', description='tubulin tyrosine ligase', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129065947, end_position_on_the_genomic_accession=129096283, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69740, CurrentID=0, Status='live', Symbol='Dph5', Aliases='2410012M04Rik, AU045680, C80186', description='diphthamide biosynthesis 5', other_designations='DPH5 homolog|diphthamide biosynthesis methyltransferase|diphthine synthase', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=115888165, end_position_on_the_genomic_accession=115929323, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69742, CurrentID=0, Status='live', Symbol='Tm2d2', Aliases='2410018G23Rik, Blp1', description='TM2 domain containing 2', other_designations='BBP-like protein 1', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25017211, end_position_on_the_genomic_accession=25023260, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69743, CurrentID=0, Status='live', Symbol='Casz1', Aliases='2410019P08Rik, AV096627, BC040081, Cst, D4Ertd432e', description='castor zinc finger 1', other_designations='castor homolog 1, zinc finger|castor-related protein', map_location='4 78.87 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148804367, end_position_on_the_genomic_accession=148954893, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  69745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69745, CurrentID=0, Status='live', Symbol='Pold4', Aliases='2410012M21Rik, AI463381, AW060307, Polds, p12', description='polymerase (DNA-directed), delta 4', other_designations='DNA polymerase delta smallest subunit p12', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4231899, end_position_on_the_genomic_accession=4233634, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69747, CurrentID=0, Status='live', Symbol='Zswim7', Aliases='2410012H22Rik', description='zinc finger SWIM-type containing 7', other_designations='SWIM domain-containing and Srs2-interacting protein 1 homolog|zinc finger, SWIM-type containing 7', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62267224, end_position_on_the_genomic_accession=62281395, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69748, CurrentID=0, Status='live', Symbol='Aldh16a1', Aliases='2410004H02Rik, AI452234, mFLJ00209', description='aldehyde dehydrogenase 16 family, member A1', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45141833, end_position_on_the_genomic_accession=45154538, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  69752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69752, CurrentID=0, Status='live', Symbol='Zfp511', Aliases='2410004P19Rik, C86579, Znf511', description='zinc finger protein 511', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140036391, end_position_on_the_genomic_accession=140040606, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69754, CurrentID=0, Status='live', Symbol='Fbxo7', Aliases='2410015K21Rik, A230052G17Rik, AI117679, AW260462, C85118', description='F-box protein 7', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=86021597, end_position_on_the_genomic_accession=86048336, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69757, CurrentID=0, Status='live', Symbol='Leng1', Aliases='1500001K17Rik', description='leukocyte receptor cluster (LRC) member 1', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3660107, end_position_on_the_genomic_accession=3665840, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69761, CurrentID=0, Status='live', Symbol='1600015I10Rik', Aliases='AI429442, Doxl1', description='RIKEN cDNA 1600015I10 gene', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48929896, end_position_on_the_genomic_accession=48933687, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69769, CurrentID=0, Status='live', Symbol='Tnfaip8l2', Aliases='1810019A08Rik, AW610835, TIPE2', description='tumor necrosis factor, alpha-induced protein 8-like 2', other_designations='TNF alpha-induced protein 8-like protein 2|TNFAIP8-like protein 2|tumor necrosis factor, alpha-induced protein 8-like protein 2', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95139521, end_position_on_the_genomic_accession=95142448, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69770, CurrentID=0, Status='live', Symbol='1600002K03Rik', Aliases='', description='RIKEN cDNA 1600002K03 gene', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80172944, end_position_on_the_genomic_accession=80175119, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69772, CurrentID=0, Status='live', Symbol='Bdh2', Aliases='1810026B04Rik, Dhrs6', description='3-hydroxybutyrate dehydrogenase, type 2', other_designations='R-beta-hydroxybutyrate dehydrogenase|dehydrogenase/reductase (SDR family) member 6|dehydrogenase/reductase SDR family member 6', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135281221, end_position_on_the_genomic_accession=135304426, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  99010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99010, CurrentID=0, Status='live', Symbol='Lpcat4', Aliases='AI505034, Agpat7, Aytl3, LPEAT2', description='lysophosphatidylcholine acyltransferase 4', other_designations='-acylglycerol-3-phosphate O-acyltransferase 7 (lysophosphatidic acid acyltransferase, eta)|1-AGP acyltransferase 7|1-AGPAT 7|1-acylglycerol-3-phosphate O-acyltransferase 7 (lysophosphatidic acid acyltransferase, eta)|1-acylglycerophosphocholine O-acyltransferase|1-acylglycerophosphoserine O-acyltransferase|1-alkenylglycerophosphoethanolamine O-acyltransferase|1-alkylglycerophosphocholine O-acetyltransferase|acyltransferase like 3|acyltransferase-like 3|lysophosphatidylethanolamine acyltransferase 2|plasmalogen synthase', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112239841, end_position_on_the_genomic_accession=112247111, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  69774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69774, CurrentID=0, Status='live', Symbol='Ms4a6b', Aliases='1810027D10Rik', description='membrane-spanning 4-domains, subfamily A, member 6B', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11518475, end_position_on_the_genomic_accession=11530403, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69780, CurrentID=0, Status='live', Symbol='Smap2', Aliases='1810031K02Rik, AA408614, Smap1l', description='small ArfGAP 2', other_designations='stromal membrane-associated GTPase-activating protein 2', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120968317, end_position_on_the_genomic_accession=121017289, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  69784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69784, CurrentID=0, Status='live', Symbol='1500009L16Rik', Aliases='2700050C19Rik, OCC-1', description='RIKEN cDNA 1500009L16 gene', other_designations='Putative overexpressed in colon carcinoma 1 protein homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=83722666, end_position_on_the_genomic_accession=83762762, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69785, CurrentID=0, Status='live', Symbol='Ceacam13', Aliases='1600012K03Rik, 1600025J19Rik', description='carcinoembryonic antigen-related cell adhesion molecule 13', other_designations='CEA-related cell adhesion molecule 13|Ceacam13-C1|Ceacam13-C2', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18009889, end_position_on_the_genomic_accession=18019220, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69786, CurrentID=0, Status='live', Symbol='Tprkb', Aliases='0610033G21Rik, 1810034M08Rik, Cgi-121, DRWMS1, mDRWMS1', description='Tp53rk binding protein', other_designations='Prpk (p53-related protein kinase)-binding protein|TP53RK-binding protein', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85915278, end_position_on_the_genomic_accession=85930284, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69787, CurrentID=0, Status='live', Symbol='Anxa13', Aliases='1810034H17Rik, AV055219', description='annexin A13', other_designations='annexin XIII|annexin-13', map_location='15 24.66 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58341464, end_position_on_the_genomic_accession=58389274, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69790, CurrentID=0, Status='live', Symbol='Med30', Aliases='1810038N03Rik, 2510044J04Rik, TRAP25, Thrap6', description='mediator complex subunit 30', other_designations='TRAP/Mediator complex component TRAP25|thyroid hormone receptor associated protein 6|thyroid hormone receptor-associated protein 6|thyroid hormone receptor-associated protein complex 25 kDa component', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=52712445, end_position_on_the_genomic_accession=52730431, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69792, CurrentID=0, Status='live', Symbol='Med6', Aliases='1500012F11Rik, AU019660, AV213303', description='mediator complex subunit 6', other_designations='mediator of RNA polymerase II transcription, subunit 6 homolog', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81573564, end_position_on_the_genomic_accession=81595071, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69802, CurrentID=0, Status='live', Symbol='Cox11', Aliases='2010004I09Rik', description='cytochrome c oxidase assembly protein 11', other_designations='COX11 homolog, cytochrome c oxidase assembly protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=90638184, end_position_on_the_genomic_accession=90645977, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69804, CurrentID=0, Status='live', Symbol='Tmem147', Aliases='2010004E11Rik, 5033425B17Rik, Nifie14', description='transmembrane protein 147', other_designations='seven transmembrane domain protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30727701, end_position_on_the_genomic_accession=30729534, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69806, CurrentID=0, Status='live', Symbol='Slc39a11', Aliases='1810074D23Rik', description='solute carrier family 39 (metal ion transporter), member 11', other_designations='ZIP-11|solute carrier family 39 member 11|zrt- and Irt-like protein 11', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113244851, end_position_on_the_genomic_accession=113650752, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  69807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69807, CurrentID=0, Status='live', Symbol='Trim32', Aliases='1810045E12Rik, 3f3, BBS11, Zfp117', description='tripartite motif-containing 32', other_designations='bM468K2.2 (tripartite motif protein 32)|tripartite motif protein 32|tripartite motif-containing protein 32|zinc finger protein 117', map_location='4 34.43 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=65604986, end_position_on_the_genomic_accession=65616240, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69809, CurrentID=0, Status='live', Symbol='1810046K07Rik', Aliases='', description='RIKEN cDNA 1810046K07 gene', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51289686, end_position_on_the_genomic_accession=51328924, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69810, CurrentID=0, Status='live', Symbol='Clec4b1', Aliases='1810046I24Rik, Aplra2, Clec4b, DCAR, DCARbeta, Dcar2', description='C-type lectin domain family 4, member b1', other_designations='antigen presenting cell lectin-like receptor A2|dendritic cell immunoactivating receptor', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=123049962, end_position_on_the_genomic_accession=123071555, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69812, CurrentID=0, Status='live', Symbol='Abhd11os', Aliases='2010001M06Rik, AI462243, Wbscr26, Wnscr26', description='abhydrolase domain containing 11, opposite strand', other_designations='Abhd11 opposite strand transcript|Williams Beuren syndrome chromosome region 26', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135012122, end_position_on_the_genomic_accession=135013157, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69814, CurrentID=0, Status='live', Symbol='Prss32', Aliases='2010001P08Rik, mT5, tryptase-5', description='protease, serine 32', other_designations='eosinophil serine protease|testisin|tryptase 5', map_location='17 12.09 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23853772, end_position_on_the_genomic_accession=23859776, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69815, CurrentID=0, Status='live', Symbol='Krtcap3', Aliases='2010001C09Rik, AW321088, Kcp3', description='keratinocyte associated protein 3', other_designations='KCP-3|keratinocytes associated protein 3', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31251706, end_position_on_the_genomic_accession=31253197, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  69816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69816, CurrentID=0, Status='live', Symbol='Mzb1', Aliases='2010001M09Rik, AV064572, PACAP, pERp1', description='marginal zone B and B1 cell-specific protein 1', other_designations='plasma cell-induced ER protein 1|plasma cell-induced resident ER protein|plasma cell-induced resident endoplasmic reticulum protein|proapoptotic caspase adapter protein', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35647265, end_position_on_the_genomic_accession=35649367, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69821, CurrentID=0, Status='live', Symbol='Mterf4', Aliases='1810059A23Rik, 4933412H03Rik, Mterfd2', description='mitochondrial transcription termination factor 4', other_designations='MTERF domain containing 2|mTERF domain-containing protein 2', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93299211, end_position_on_the_genomic_accession=93305965, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69823, CurrentID=0, Status='live', Symbol='Fyttd1', Aliases='2010005M05Rik, 3830411L18Rik, 3830421F13Rik, 4933423A17Rik, AI098218, AW539567, C85190', description='forty-two-three domain containing 1', other_designations='forty-two-three domain-containing protein 1|protein 40-2-3', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32877581, end_position_on_the_genomic_accession=32908963, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  69824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69824, CurrentID=0, Status='live', Symbol='Glod5', Aliases='2010001H14Rik, AV061890', description='glyoxalase domain containing 5', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8004200, end_position_on_the_genomic_accession=8018492, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69826, CurrentID=0, Status='live', Symbol='Ms4a10', Aliases='2010001N17Rik', description='membrane-spanning 4-domains, subfamily A, member 10', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10962289, end_position_on_the_genomic_accession=10974725, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  69833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69833, CurrentID=0, Status='live', Symbol='Polr2f', Aliases='1810060D16Rik, RPB6', description='polymerase (RNA) II (DNA directed) polypeptide F', other_designations='DNA-directed RNA polymerase II subunit F|RNA Pol II|RNA polymerases I, II, and III subunit ABC2', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79141367, end_position_on_the_genomic_accession=79151767, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69834, CurrentID=0, Status='live', Symbol='Rab43', Aliases='1810048P08Rik, 2500004H21Rik, AW490415', description='RAB43, member RAS oncogene family', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87788853, end_position_on_the_genomic_accession=87812164, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69836, CurrentID=0, Status='live', Symbol='Pla2g12b', Aliases='2010002E04Rik, Fksg71, Pla2g13, hlb218', description='phospholipase A2, group XIIB', other_designations='GXIIB|group XIIB phospholipase A2|group XIII secretory phospholipase A2-like protein|phospholipase A2, group XIII|sPLA2-GXIIB', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59403420, end_position_on_the_genomic_accession=59421976, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69837, CurrentID=0, Status='live', Symbol='Pcgf1', Aliases='2010002K04Rik, AU024121, Nspc1', description='polycomb group ring finger 1', other_designations='RING finger protein 68|nervous system Polycomb-1|nervous system polycomb 1', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83078390, end_position_on_the_genomic_accession=83080855, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69847, CurrentID=0, Status='live', Symbol='Wnk4', Aliases='2010002J11Rik, Pha2b, Prkwnk4', description='WNK lysine deficient protein kinase 4', other_designations='protein kinase lysine-deficient 4|protein kinase with no lysine 4|protein kinase, lysine deficient 4', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101260537, end_position_on_the_genomic_accession=101277409, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  69852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69852, CurrentID=0, Status='live', Symbol='Tcf23', Aliases='2010002O16Rik, Out, bHLHa24', description='transcription factor 23', other_designations='TCF-23|basic helix-loop-helix transcription factor OUT|ovary, uterus and testis protein', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30968601, end_position_on_the_genomic_accession=30988065, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69860, CurrentID=0, Status='live', Symbol='Eif1ad', Aliases='2010003J03Rik, AI131644', description='eukaryotic translation initiation factor 1A domain containing', other_designations='eukaryotic translation initiation factor 1A domain-containing protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5366813, end_position_on_the_genomic_accession=5371511, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69861, CurrentID=0, Status='live', Symbol='2010003K11Rik', Aliases='', description='RIKEN cDNA 2010003K11 gene', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4496788, end_position_on_the_genomic_accession=4498583, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69863, CurrentID=0, Status='live', Symbol='Ttc39b', Aliases='1810054D07Rik, 9130422G05Rik, AI450603, AV103077, AW215585', description='tetratricopeptide repeat domain 39B', other_designations='TPR repeat protein 39B', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=83220301, end_position_on_the_genomic_accession=83324270, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  69864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69864, CurrentID=0, Status='live', Symbol='1810065E05Rik', Aliases='AV047390', description='RIKEN cDNA 1810065E05 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58421111, end_position_on_the_genomic_accession=58426023, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69865, CurrentID=0, Status='live', Symbol='A1cf', Aliases='1810073H04Rik, ACF64, ACF65, ASP, Acf', description='APOBEC1 complementation factor', other_designations='APOBEC-1 stimulating protein|APOBEC1-stimulating protein|apobec-1 complementation factor', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=31868755, end_position_on_the_genomic_accession=31952365, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69870, CurrentID=0, Status='live', Symbol='Polr3gl', Aliases='2010003I19Rik', description='polymerase (RNA) III (DNA directed) polypeptide G like', other_designations='DNA-directed RNA polymerase III subunit G-like|RNA polymerase III subunit C7-like', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96577872, end_position_on_the_genomic_accession=96594181, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69871, CurrentID=0, Status='live', Symbol='Ppp1r35', Aliases='2010007H12Rik, 2010011D20Rik, AL024004', description='protein phosphatase 1, regulatory subunit 35', other_designations='UPF0683 protein C7orf47 homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137778904, end_position_on_the_genomic_accession=137782201, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69875, CurrentID=0, Status='live', Symbol='Ndufa11', Aliases='2010012C24Rik, AV006275, B14.7', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex 11', other_designations='CI-B14.7|NADH-ubiquinone oxidoreductase subunit B14.7|complex I-B14.7', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56717762, end_position_on_the_genomic_accession=56724248, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  69876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69876, CurrentID=0, Status='live', Symbol='Thap3', Aliases='2010013E08Rik, 2210418H06Rik', description='THAP domain containing, apoptosis associated protein 3', other_designations='snoRNA MBI-15', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=151982638, end_position_on_the_genomic_accession=151988986, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69878, CurrentID=0, Status='live', Symbol='Snrpf', Aliases='2010013O18Rik, SMF, sm-F, snRNP-F', description='small nuclear ribonucleoprotein polypeptide F', other_designations='sm protein F', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93583029, end_position_on_the_genomic_accession=93589658, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69882, CurrentID=0, Status='live', Symbol='Vwa9', Aliases='2010321M09Rik', description='von Willebrand factor A domain containing 9', other_designations='UPF0464 protein C15orf44 homolog', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64960832, end_position_on_the_genomic_accession=64986981, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  69885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69885, CurrentID=0, Status='live', Symbol='Aunip', Aliases='2610002D18Rik, Clhc1', description='aurora kinase A and ninein interacting protein', other_designations='aurora A-binding|clathrin heavy chain linker domain containing 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134510999, end_position_on_the_genomic_accession=134523922, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69888, CurrentID=0, Status='live', Symbol='Cyp2c66', Aliases='2010301M18Rik', description='cytochrome P450, family 2, subfamily c, polypeptide 66', other_designations='cytochrome P450, 2c66', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39113898, end_position_on_the_genomic_accession=39186756, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  69890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69890, CurrentID=0, Status='live', Symbol='Zfp219', Aliases='2010302A17Rik', description='zinc finger protein 219', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52006075, end_position_on_the_genomic_accession=52020733, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69893, CurrentID=0, Status='live', Symbol='Coa7', Aliases='2010305A19Rik, AI451324, D4Ertd796e, Selrc1', description='cytochrome c oxidase assembly factor 7', other_designations='Sel1 repeat containing 1|beta-lactamase hcp-like protein|hcp beta-lactamase-like protein C1orf163 homolog|respiratory chain assembly factor 1|sel1 repeat-containing protein 1', map_location='4 50.44 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108328152, end_position_on_the_genomic_accession=108340718, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  69894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69894, CurrentID=0, Status='live', Symbol='2010107G23Rik', Aliases='AA545165', description='RIKEN cDNA 2010107G23 gene', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62107656, end_position_on_the_genomic_accession=62111074, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  69895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69895, CurrentID=0, Status='live', Symbol='Snhg8', Aliases='1110065M07Rik, 2010109N14Rik', description='small nucleolar RNA host gene 8', other_designations='small nucleolar RNA host gene (non-protein coding) 8', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=123507552, end_position_on_the_genomic_accession=123508336, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  99031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99031, CurrentID=0, Status='live', Symbol='Osbpl6', Aliases='1110062M20Rik, AI596402, ORP-6, mKIAA4128', description='oxysterol binding protein-like 6', other_designations='OSBP-related protein 6', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76406508, end_position_on_the_genomic_accession=76600647, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  69900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69900, CurrentID=0, Status='live', Symbol='Mfsd11', Aliases='2600014M03Rik', description='major facilitator superfamily domain containing 11', other_designations='major facilitator superfamily domain-containing protein 11|protein ET', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116852283, end_position_on_the_genomic_accession=116875660, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  69902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69902, CurrentID=0, Status='live', Symbol='Mrto4', Aliases='2610012O22Rik, Mg684, Mrt4', description='mRNA turnover 4, homolog (S. cerevisiae)', other_designations='MRT4, mRNA turnover 4, homolog|muscle protein684|ribosome assembly factor Mrto4', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139347440, end_position_on_the_genomic_accession=139352576, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69903, CurrentID=0, Status='live', Symbol='Rasip1', Aliases='1110025D03Rik, 2610025P08Rik, AI853551, Rain', description='Ras interacting protein 1', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45627537, end_position_on_the_genomic_accession=45639092, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69906, CurrentID=0, Status='live', Symbol='Slc25a32', Aliases='2610043O12Rik, Mftc', description='solute carrier family 25, member 32', other_designations='', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39094191, end_position_on_the_genomic_accession=39112768, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  69908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69908, CurrentID=0, Status='live', Symbol='Rab3b', Aliases='2610528C18Rik', description='RAB3B, member RAS oncogene family', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108879070, end_position_on_the_genomic_accession=108943324, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69912, CurrentID=0, Status='live', Symbol='Nup43', Aliases='2610016K01Rik, 2610529I12Rik, AA409950, p42', description='nucleoporin 43', other_designations='nucleoporin 43kDa|nup107-160 subcomplex subunit Nup43', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7667504, end_position_on_the_genomic_accession=7678886, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  69917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69917, CurrentID=0, Status='live', Symbol='Nabp2', Aliases='2610036N15Rik, Obfc2b', description='nucleic acid binding protein 2', other_designations='SOSS-B1|nucleic acid-binding protein 2|oligonucleotide/oligosaccharide-binding fold containing 2B|oligonucleotide/oligosaccharide-binding fold-containing protein 2B|sensor of single-strand DNA complex subunit B1|sensor of ssDNA subunit B1|single-stranded DNA-binding protein 1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128401394, end_position_on_the_genomic_accession=128411487, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  69920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69920, CurrentID=0, Status='live', Symbol='Polr2i', Aliases='2810002B19Rik', description='polymerase (RNA) II (DNA directed) polypeptide I', other_designations='DNA-directed RNA polymerase II subunit I|RNA polymerase II 14.5 kDa subunit|RNA polymerase II subunit B9|RPB14.5', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30232074, end_position_on_the_genomic_accession=30233387, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  69922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69922, CurrentID=0, Status='live', Symbol='Vrk2', Aliases='2810003O05Rik, AI447698', description='vaccinia related kinase 2', other_designations='vaccinia virus BIR kinase related kinase 2|vaccinia-related kinase 2', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=26471401, end_position_on_the_genomic_accession=26594014, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  69923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69923, CurrentID=0, Status='live', Symbol='Agk', Aliases='2610037M15Rik, 6720408I04Rik, AI465370, Mulk', description='acylglycerol kinase', other_designations='multi-substrate lipid kinase|multiple substrate lipid kinase', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40325155, end_position_on_the_genomic_accession=40397562, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  69926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69926, CurrentID=0, Status='live', Symbol='Dnah17', Aliases='2810003K23Rik, Dnahc17, Dnahcl1, Gm1178, mKIAA3028', description='dynein, axonemal, heavy chain 17', other_designations='axonemal beta dynein heavy chain 17|ciliary dynein heavy chain 17|dynein, axonemal, heavy like 1|dynein, axonemal, heavy polypeptide 17', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118021723, end_position_on_the_genomic_accession=118130961, orientation='minus', exon_count=83, OMIM='', no_hdr0=''),
  69928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69928, CurrentID=0, Status='live', Symbol='Apitd1', Aliases='2610040C18Rik, 2810407L01Rik, CENP-S', description='apoptosis-inducing, TAF9-like domain 1', other_designations='FANCM-interacting histone fold protein 1|apoptosis-inducing TAF9-like domain-containing protein 1 homolog|ortholog of human apoptosis-inducing, TAF9-like domain 1 APITD1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149128349, end_position_on_the_genomic_accession=149137600, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  69930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69930, CurrentID=0, Status='live', Symbol='Zfp715', Aliases='2610041B18Rik, mszf15', description='zinc finger protein 715', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43296522, end_position_on_the_genomic_accession=43313311, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69932, CurrentID=0, Status='live', Symbol='2810004I08Rik', Aliases='D10Wsu1e', description='RIKEN cDNA 2810004I08 gene', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  69934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69934, CurrentID=0, Status='live', Symbol='Trmt10b', Aliases='2610042J10Rik, Rg9mtd3, Rnmtd3', description='tRNA methyltransferase 10B', other_designations='RNA (guanine-9-) methyltransferase domain containing 3|RNA (guanine-9-)-methyltransferase domain-containing protein 3', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45297071, end_position_on_the_genomic_accession=45316131, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  69938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69938, CurrentID=0, Status='live', Symbol='Scrn1', Aliases='2810019K23Rik, 6330535A03Rik, AI852905, SES1, mKIAA0193', description='secernin 1', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54472300, end_position_on_the_genomic_accession=54566499, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  69940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69940, CurrentID=0, Status='live', Symbol='Exoc1', Aliases='2810407P21Rik, A730011E05Rik, SEC3, Sec3l1, Sec3p', description='exocyst complex component 1', other_designations='SEC3-like 1|exocyst complex component Sec3', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76529311, end_position_on_the_genomic_accession=76570298, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  69942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69942, CurrentID=0, Status='live', Symbol='Rnf113a1', Aliases='2810428C21Rik, Rnf113a', description='ring finger protein 113A1', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37191343, end_position_on_the_genomic_accession=37192467, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  69944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69944, CurrentID=0, Status='live', Symbol='2810021J22Rik', Aliases='AI449137', description='RIKEN cDNA 2810021J22 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58867240, end_position_on_the_genomic_accession=58884338, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  69955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69955, CurrentID=0, Status='live', Symbol='Fars2', Aliases='Fars1, pheRS', description='phenylalanine-tRNA synthetase 2 (mitochondrial)', other_designations='phenylalanine-tRNA synthetase 1 (mitochondrial)', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=36117643, end_position_on_the_genomic_accession=36537595, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  69956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69956, CurrentID=0, Status='live', Symbol='Ptcd3', Aliases='2610034F17Rik, 2810422B04Rik, AA589622, AU045708, MRP-S39', description='pentatricopeptide repeat domain 3', other_designations='28S ribosomal protein S39, mitochondrial|pentatricopeptide repeat-containing protein 3, mitochondrial', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71880638, end_position_on_the_genomic_accession=71908772, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  69957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69957, CurrentID=0, Status='live', Symbol='Cdc16', Aliases='2700071J12Rik, 2810431D22Rik, APC6', description='CDC16 cell division cycle 16', other_designations='CDC16 cell division cycle 16 homolog|anaphase-promoting complex subunit 6|cyclosome subunit 6', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13757646, end_position_on_the_genomic_accession=13781938, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  69961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69961, CurrentID=0, Status='live', Symbol='Rpp25l', Aliases='2810432D09Rik, AW987258, Rppl25', description='ribonuclease P/MRP 25 subunit-like', other_designations='RNase P protein subunit-like p25|alba-like protein C9orf23 homolog|rpp25-like protein', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41712026, end_position_on_the_genomic_accession=41713819, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  69962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69962, CurrentID=0, Status='live', Symbol='Mettl18', Aliases='2810422O20Rik, C78571', description='methyltransferase like 18', other_designations='methyltransferase-like protein 18', map_location='1 H2.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=163994891, end_position_on_the_genomic_accession=163997243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  332110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332110, CurrentID=0, Status='live', Symbol='Mapk15', Aliases='BC048082', description='mitogen-activated protein kinase 15', other_designations='ERK-7|MAP kinase 15|MAPK 15|extracellular signal-regulated kinase 7', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75993731, end_position_on_the_genomic_accession=75999153, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  69976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69976, CurrentID=0, Status='live', Symbol='Galk2', Aliases='2810017M24Rik, AI327030, Gk2', description='galactokinase 2', other_designations='galNAc kinase', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125859109, end_position_on_the_genomic_accession=125984298, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  69981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69981, CurrentID=0, Status='live', Symbol='Tmem30a', Aliases='2010200I23Rik, AW540225, Cdc50a, D9Wsu20e', description='transmembrane protein 30A', other_designations='P4-ATPase flippase complex beta subunit TMEM30A', map_location='9 43.82 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=79760276, end_position_on_the_genomic_accession=79794028, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  69982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69982, CurrentID=0, Status='live', Symbol='Spink2', Aliases='1700007F22Rik, AV038945, HUSI-II', description='serine peptidase inhibitor, Kazal type 2', other_designations='acrosin-trypsin inhibitor', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77205106, end_position_on_the_genomic_accession=77216260, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  69983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69983, CurrentID=0, Status='live', Symbol='Sis', Aliases='2010204N08Rik, SI, Si-s', description='sucrase isomaltase (alpha-glucosidase)', other_designations='alpha-glucosidase|sucrase isomaltase, structural|sucrase-isomaltase', map_location='3 32.59 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=72888554, end_position_on_the_genomic_accession=72967864, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  332131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332131, CurrentID=0, Status='live', Symbol='Krt78', Aliases='2310030B04Rik, Kb40', description='keratin 78', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101946001, end_position_on_the_genomic_accession=101996609, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  69993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69993, CurrentID=0, Status='live', Symbol='Chn2', Aliases='1700026N20Rik, 4930557O16Rik, ARHGAP3, Bch', description='chimerin 2', other_designations='Rho GTPase-activating protein 3|beta chimaerin|chimerin (chimaerin) 2', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54039932, end_position_on_the_genomic_accession=54301812, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  69994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=69994, CurrentID=0, Status='live', Symbol='Rsc1a1', Aliases='1700027M01Rik, OTTMUSG00000010365, Rs1', description='regulatory solute carrier protein, family 1, member 1', other_designations='regulatory subunit of SGLT1|transporter regulator RS1', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141683563, end_position_on_the_genomic_accession=141685716, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  70008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70008, CurrentID=0, Status='live', Symbol='Ace2', Aliases='2010305L05Rik', description='angiotensin I converting enzyme (peptidyl-dipeptidase A) 2', other_designations='ACE-related carboxypeptidase|angiotensin I converting enzyme 2', map_location='X 76.12 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164139342, end_position_on_the_genomic_accession=164188418, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  70009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70009, CurrentID=0, Status='live', Symbol='Ssty2', Aliases='1700029H17Rik, pc11', description='spermiogenesis specific transcript on the Y 2', other_designations='', map_location='Y D', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=67045743, end_position_on_the_genomic_accession=67048072, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  70012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70012, CurrentID=0, Status='live', Symbol='Cep85', Aliases='2410030J07Rik, AI173272, AI173504, Ccdc21', description='centrosomal protein 85', other_designations='coiled-coil domain containing 21|coiled-coil domain-containing protein 21', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134129858, end_position_on_the_genomic_accession=134187085, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  70020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70020, CurrentID=0, Status='live', Symbol='Ino80b', Aliases='2510009I23Rik, HMG1YL4, Hmga1l4, Papa1, Znhit4', description='INO80 complex subunit B', other_designations='PAP-1 binding protein|PAP-1-associated protein 1|PAPA-1|high mobility group AT-hook 1-like 4|zinc finger HIT domain-containing protein 4|zinc finger, HIT type 4', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83121828, end_position_on_the_genomic_accession=83125029, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70021, CurrentID=0, Status='live', Symbol='Nt5dc2', Aliases='2510015F01Rik', description="5'-nucleotidase domain containing 2", other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31131058, end_position_on_the_genomic_accession=31139124, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70024, CurrentID=0, Status='live', Symbol='Mcm10', Aliases='2410041F14Rik, AU018508, C330019M07Rik, C79164', description='minichromosome maintenance 10 replication initiation factor', other_designations='minichromosome maintenance deficient 10', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=4990724, end_position_on_the_genomic_accession=5012848, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  70025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70025, CurrentID=0, Status='live', Symbol='Acot7', Aliases='2410041A17Rik, Ach1, Act, Bach, CTE-IIa, Cte-II, Lach1', description='acyl-CoA thioesterase 7', other_designations='50-kDa brain acyl-CoA hydrolase|brain acyl-CoA hydrolase|long chain acyl-CoA thioester hydrolase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152178100, end_position_on_the_genomic_accession=152271855, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70026, CurrentID=0, Status='live', Symbol='Tspo2', Aliases='2510027D20Rik, Bzrpl1', description='translocator protein 2', other_designations='benzodiazapine receptor, peripheral-like 1|peripheral-type benzodiazepine receptor-like protein 1', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48444411, end_position_on_the_genomic_accession=48451635, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70028, CurrentID=0, Status='live', Symbol='Dopey2', Aliases='0610038M01Rik, 2610510B01Rik, AI503246, C21orf5, mKIAA0933', description='dopey family member 2', other_designations='', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93711907, end_position_on_the_genomic_accession=93810588, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  70031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70031, CurrentID=0, Status='live', Symbol='Cmtm8', Aliases='2700018N07Rik, AA408515, Cklfsf8', description='CKLF-like MARVEL transmembrane domain containing 8', other_designations='chemokine-like factor super family 8|chemokine-like factor superfamily member 8', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114789343, end_position_on_the_genomic_accession=114844152, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70036, CurrentID=0, Status='live', Symbol='Dancr', Aliases='2700023E23Rik', description='differentiation antagonizing non-protein coding RNA', other_designations='', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=74093083, end_position_on_the_genomic_accession=74094336, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  70044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70044, CurrentID=0, Status='live', Symbol='Tut1', Aliases='2700038E08Rik, PAPD2, Rbm21, TUTase6', description='terminal uridylyl transferase 1, U6 snRNA-specific', other_designations='RNA binding motif protein 21|RNA-binding motif protein 21|RNA-binding protein 21|U6 snRNA-specific terminal uridylyltransferase 1|U6-TUTase|star-PAP', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8953850, end_position_on_the_genomic_accession=8966210, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70045, CurrentID=0, Status='live', Symbol='2610528A11Rik', Aliases='AW494830', description='RIKEN cDNA 2610528A11 gene', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=37102140, end_position_on_the_genomic_accession=37110101, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70047, CurrentID=0, Status='live', Symbol='Trnt1', Aliases='2410043H24Rik, 2610044E04Rik, 9830143O18Rik, AA408384, C76540, CGI-47, mt-Trat', description='tRNA nucleotidyl transferase, CCA-adding, 1', other_designations='mitochondrial tRNA nucleotidyl transferase, CCA-adding|mt CCA-adding enzyme|mt tRNA CCA-diphosphorylase|mt tRNA CCA-pyrophosphorylase|mt tRNA adenylyltransferase|tRNA adenylyltransferase, mitochondrial|tRNA-nucleotidyltransferase 1, mitochondrial', map_location='6 E2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=106769130, end_position_on_the_genomic_accession=106782474, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70052, CurrentID=0, Status='live', Symbol='Prpf4', Aliases='1600015H11Rik, AI874830, AW047464, bN189G18.1', description='pre-mRNA processing factor 4', other_designations='U4/U6 snRNP 60 kDa protein|WD splicing factor Prp4', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62408767, end_position_on_the_genomic_accession=62426990, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70054, CurrentID=0, Status='live', Symbol='Ccdc89', Aliases='1700019B01Rik, AU022100, BOIP', description='coiled-coil domain containing 89', other_designations='bc8 orange-interacting protein|ccdc89', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90426312, end_position_on_the_genomic_accession=90428664, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  70059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70059, CurrentID=0, Status='live', Symbol='Degs2', Aliases='2210008A03Rik, AI852933, DES2', description='delta(4)-desaturase, sphingolipid 2', other_designations='degenerative spermatocyte homolog 2 (Drosophila), lipid desaturase|degs2|sphingolipid 4-desaturase|sphingolipid C4-monooxygenase|sphingolipid delta 4 desaturase/C-4 hydroxylase|sphingolipid delta(4)-desaturase/C4-hydroxylase DES2', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108686792, end_position_on_the_genomic_accession=108702306, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70060, CurrentID=0, Status='live', Symbol='Spata3', Aliases='1700011N12Rik, 1700029H01Rik, 4930424D10Rik, Mtsarg1, Tsarg1', description='spermatogenesis associated 3', other_designations='testis and spermatogenesis cell apoptosis related gene 1|testis and spermatogenesis cell apoptosis related protein 1|testis and spermatogenesis cell-related protein 1|testis spermatocyte apoptosis-related protein 1', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86017840, end_position_on_the_genomic_accession=86029958, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70061, CurrentID=0, Status='live', Symbol='Sdr9c7', Aliases='1810054F20Rik, Rdh20, Rdhs, SDR-O, Sdro', description='4short chain dehydrogenase/reductase family 9C, member 7', other_designations='RDH-S|orphan short chain dehydrogenase/reductase|orphan short-chain dehydrogenase/reductase|retinol dehydrogenase 20|short chain dehydrogenase/reductase family 9C member 7', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127898535, end_position_on_the_genomic_accession=127911761, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  70062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70062, CurrentID=0, Status='live', Symbol='Ctag2', Aliases='1700030B21Rik', description='cancer/testis antigen 2', other_designations='', map_location='X A7.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=65047644, end_position_on_the_genomic_accession=65049017, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  70065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70065, CurrentID=0, Status='live', Symbol='Ankrd60', Aliases='1700019A24Rik, 1700030G11Rik', description='ankyrin repeat domain 60', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173568661, end_position_on_the_genomic_accession=173578341, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70069, CurrentID=0, Status='live', Symbol='H1fnt', Aliases='1700026P10Rik, H1.7, H1t2, Hanp1, mH1t2', description='H1 histone family, member N, testis-specific', other_designations='expressed in haploid spermatids during spermiogenesis, specifically at the apical pole of round and elongating spermatids|haploid germ cell-specific nuclear protein 1|histone H1t2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98255982, end_position_on_the_genomic_accession=98257307, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  70073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70073, CurrentID=0, Status='live', Symbol='Zdhhc25', Aliases='1700030J15Rik', description='zinc finger, DHHC domain containing 25', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88600302, end_position_on_the_genomic_accession=88601669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  332221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332221, CurrentID=0, Status='live', Symbol='Zscan10', Aliases='BC068284, Zfp206, Zkscan10, Znf206', description='zinc finger and SCAN domain containing 10', other_designations='zinc finger and SCAN domains 10|zinc finger protein 206|zinc finger with KRAB and SCAN domains 10', map_location='17 A3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23599751, end_position_on_the_genomic_accession=23611019, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70078, CurrentID=0, Status='live', Symbol='Nol7', Aliases='2210008F15Rik, 5730556I21Rik, NOP27, RARG-1', description='nucleolar protein 7', other_designations='', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43398376, end_position_on_the_genomic_accession=43402858, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70080, CurrentID=0, Status='live', Symbol='Igsf23', Aliases='2210010C17Rik', description='immunoglobulin superfamily, member 23', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19937305, end_position_on_the_genomic_accession=19950765, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70081, CurrentID=0, Status='live', Symbol='2210404O09Rik', Aliases='', description='RIKEN cDNA 2210404O09 gene', other_designations='', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21878907, end_position_on_the_genomic_accession=21909926, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70082, CurrentID=0, Status='live', Symbol='Lysmd2', Aliases='2210402C18Rik, AV033872, AW538442', description='LysM, putative peptidoglycan-binding, domain containing 2', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75625732, end_position_on_the_genomic_accession=75637773, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  70083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70083, CurrentID=0, Status='live', Symbol='Metrn', Aliases='1810034B16Rik, Hyrac', description='meteorin, glial cell differentiation regulator', other_designations='hypoxia/reoxygenation regulatory factor in astrocytes', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25794571, end_position_on_the_genomic_accession=25797094, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70086, CurrentID=0, Status='live', Symbol='Cysltr2', Aliases='2300001H05Rik, CYSLT2R, Cltr2, Cyslt2', description='cysteinyl leukotriene receptor 2', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73026075, end_position_on_the_genomic_accession=73049114, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70088, CurrentID=0, Status='live', Symbol='Meaf6', Aliases='2310005N01Rik, 2810036M01Rik', description='MYST/Esa1-associated factor 6', other_designations='esa1-associated factor 6 homolog|protein EAF6 homolog', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=125085074, end_position_on_the_genomic_accession=125110070, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70093, CurrentID=0, Status='live', Symbol='Ube2q1', Aliases='1110002C01Rik, 2310012M18Rik, AI854014, Gtap, NICE-5, PRO3094, Ube2q', description='ubiquitin-conjugating enzyme E2Q family member 1', other_designations='E2 ubiquitin-conjugating enzyme Q1|galactosyl transferase-associated protein|ubiquitin carrier protein Q1|ubiquitin-conjugating enzyme E2Q (putative) 1|ubiquitin-protein ligase Q1', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89773609, end_position_on_the_genomic_accession=89783997, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  70097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70097, CurrentID=0, Status='live', Symbol='Sash1', Aliases='1100001C18Rik, 2500002E12Rik, A330076K04Rik, mKIAA0790', description='SAM and SH3 domain containing 1', other_designations='putative adapter and scaffold protein 1', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=8722217, end_position_on_the_genomic_accession=8956137, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  70099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70099, CurrentID=0, Status='live', Symbol='Smc4', Aliases='2500002A22Rik, C79747, SMC-4l1, Smc4', description='structural maintenance of chromosomes 4', other_designations='SMC protein 4|SMC4 structural maintenance of chromosomes 4-like 1|XCAP-C homolog|chromosome-associated polypeptide C', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69004972, end_position_on_the_genomic_accession=69034623, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  70101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70101, CurrentID=0, Status='live', Symbol='Cyp4f16', Aliases='2310021J05Rik, AI746515', description='cytochrome P450, family 4, subfamily f, polypeptide 16', other_designations='cytochrome P450, 4f16', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32536545, end_position_on_the_genomic_accession=32551798, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70103, CurrentID=0, Status='live', Symbol='Znhit1', Aliases='2700001K05Rik, ZNFN4A1', description='zinc finger, HIT domain containing 1', other_designations='zinc finger protein, subfamily 4A (HIT domain containing), member 1', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136982194, end_position_on_the_genomic_accession=136987882, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77220, CurrentID=0, Status='live', Symbol='Tmem200a', Aliases='C030003D03Rik', description='transmembrane protein 200A', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=25991182, end_position_on_the_genomic_accession=26079408, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  70110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70110, CurrentID=0, Status='live', Symbol='Ifi35', Aliases='2010008K16Rik, AW986054, IFP35, ifi-35', description='interferon-induced protein 35', other_designations='IFP 35', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101448412, end_position_on_the_genomic_accession=101458701, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70113, CurrentID=0, Status='live', Symbol='Odf3b', Aliases='2010001J22Rik', description='outer dense fiber of sperm tails 3B', other_designations='outer dense fiber protein 3-like protein 3', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89377447, end_position_on_the_genomic_accession=89379681, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70118, CurrentID=0, Status='live', Symbol='Srrd', Aliases='2810002G02Rik, Srr1', description='SRR1 domain containing', other_designations='SRR1 domain-containing protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112337391, end_position_on_the_genomic_accession=112343115, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70120, CurrentID=0, Status='live', Symbol='Yars2', Aliases='2210023C10Rik, tyrRS', description='tyrosyl-tRNA synthetase 2 (mitochondrial)', other_designations='', map_location='16 A2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16302965, end_position_on_the_genomic_accession=16309640, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  70122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70122, CurrentID=0, Status='live', Symbol='Mllt3', Aliases='2210011H10Rik, 2610012I03Rik, 3830408D16Rik, Af9, D4Ertd321e', description='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 3', other_designations='myeloid/lymphoid or mixed lineage-leukemia translocation to 3 homolog|myeloid/lymphoid or mixed-lineage leukemia translocated to chromosome 3 protein homolog', map_location='4 41.06 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=87769925, end_position_on_the_genomic_accession=88033407, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  70127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70127, CurrentID=0, Status='live', Symbol='Dpf3', Aliases='2810403B03Rik, 6530402L11Rik, BAF45C, C78788, CERD4, cer-d4', description='D4, zinc and double PHD fingers, family 3', other_designations='BRG1-associated factor 45C|D4 zinc and double PHD finger family 3|cerebellum D4|zinc finger protein cer-d4', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83213751, end_position_on_the_genomic_accession=83487736, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70129, CurrentID=0, Status='live', Symbol='Slc44a4', Aliases='2210409B01Rik, NG22', description='solute carrier family 44, member 4', other_designations='', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34914224, end_position_on_the_genomic_accession=34930436, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  70144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70144, CurrentID=0, Status='live', Symbol='Lrch3', Aliases='2210409B11Rik, AW215594, Gm1742', description='leucine-rich repeats and calponin homology (CH) domain containing 3', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32914092, end_position_on_the_genomic_accession=33016029, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  70152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70152, CurrentID=0, Status='live', Symbol='Mettl7a1', Aliases='2210414H16Rik, 3300001H21Rik, Aam-B, Mettl7a, UbiE1', description='methyltransferase like 7A1', other_designations='methyltransferase like 7A', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100304145, end_position_on_the_genomic_accession=100328665, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  70153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70153, CurrentID=0, Status='live', Symbol='2210016F16Rik', Aliases='', description='RIKEN cDNA 2210016F16 gene', other_designations='', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58380046, end_position_on_the_genomic_accession=58385225, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70155, CurrentID=0, Status='live', Symbol='Ogfrl1', Aliases='2210417C17Rik', description='opioid growth factor receptor-like 1', other_designations='', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=23366424, end_position_on_the_genomic_accession=23383210, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70160, CurrentID=0, Status='live', Symbol='Vps36', Aliases='1700010A24Rik, 2210415M20Rik, 2810408E15Rik, Eap45', description='vacuolar protein sorting 36 (yeast)', other_designations='ESCRT-II complex subunit VPS36', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22192860, end_position_on_the_genomic_accession=22218597, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70163, CurrentID=0, Status='live', Symbol='Lypd8', Aliases='2210415F13Rik', description='LY6/PLAUR domain containing 8', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58379032, end_position_on_the_genomic_accession=58390544, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  332309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332309, CurrentID=0, Status='live', Symbol='Grxcr2', Aliases='Gm851', description='glutaredoxin, cysteine rich 2', other_designations='GRXCR1-like protein|glutaredoxin domain-containing cysteine-rich protein 1-like protein', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=41986201, end_position_on_the_genomic_accession=41999049, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70166, CurrentID=0, Status='live', Symbol='Lipn', Aliases='2210418G03Rik, Lipl4', description='lipase, family member N', other_designations='lipase-like abhydrolase domain-containing protein 4|lipase-like, ab-hydrolase domain containing 4', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34067058, end_position_on_the_genomic_accession=34085946, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70178, CurrentID=0, Status='live', Symbol='Abhd17c', Aliases='2210412D01Rik, AL023007, Fam108c, Fam108c1', description='abhydrolase domain containing 17C', other_designations='abhydrolase domain-containing protein FAM108C1|family with sequence similarity 108, member C', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=84109356, end_position_on_the_genomic_accession=84151893, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70186, CurrentID=0, Status='live', Symbol='Fam162a', Aliases='2310056P07Rik, HGTD-P', description='family with sequence similarity 162, member A', other_designations='E2-induced gene 5 protein homolog|growth and transformation-dependent protein', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36043844, end_position_on_the_genomic_accession=36071515, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70192, CurrentID=0, Status='live', Symbol='Cd209g', Aliases='2310066I10Rik, AI481039, SIGNR7', description='CD209g antigen', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4134246, end_position_on_the_genomic_accession=4137707, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70202, CurrentID=0, Status='live', Symbol='Ctsll3', Aliases='2310051M13Rik', description='cathepsin L-like 3', other_designations='', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60798250, end_position_on_the_genomic_accession=60802844, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  70207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70207, CurrentID=0, Status='live', Symbol='Taco1', Aliases='2310066I18Rik, Ccdc44', description='translational activator of mitochondrially encoded cytochrome c oxidase I', other_designations='coiled-coil domain containing 44|coiled-coil domain-containing protein 44|translational activator of mitochondrially-encoded cytochrome c oxidase 1', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106066039, end_position_on_the_genomic_accession=106073612, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  70208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70208, CurrentID=0, Status='live', Symbol='Med23', Aliases='130kDa, 3000002A17Rik, Crsp3, ESTM7, Sur2, X83317, mKIAA1216', description='mediator complex subunit 23', other_designations='CRSP complex subunit 3|cofactor required for Sp1 transcriptional activation subunit 3|cofactor required for Sp1 transcriptional activation, subunit 3 (130kD)|mSur-2|protein sur-2 homolog', map_location='10 12.26 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24869986, end_position_on_the_genomic_accession=24913529, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  70209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70209, CurrentID=0, Status='live', Symbol='Tmem143', Aliases='2310076O21Rik, AI481604', description='transmembrane protein 143', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45896990, end_position_on_the_genomic_accession=45917418, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70218, CurrentID=0, Status='live', Symbol='Kif18b', Aliases='3000004C01Rik', description='kinesin family member 18B', other_designations='N-8 kinesin', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102905518, end_position_on_the_genomic_accession=102925134, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  70223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70223, CurrentID=0, Status='live', Symbol='Nars', Aliases='3010001M15Rik, AA960128, ASNRS, C78150', description='asparaginyl-tRNA synthetase', other_designations='asparaginyl-tRNA synthetase, cytoplasmic', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=64499647, end_position_on_the_genomic_accession=64516557, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  70225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70225, CurrentID=0, Status='live', Symbol='Ppil3', Aliases='2310076N22Rik, 2510026K04Rik, Cyp10l', description='peptidylprolyl isomerase (cyclophilin)-like 3', other_designations='PPIase|cyclophilin-like 3|cyclophilin-like protein PPIL3|rotamase PPIL3', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58430993, end_position_on_the_genomic_accession=58445527, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  70227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70227, CurrentID=0, Status='live', Symbol='Zfp619', Aliases='3000002G13Rik', description='zinc finger protein 619', other_designations='', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=39517766, end_position_on_the_genomic_accession=39540415, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  70231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70231, CurrentID=0, Status='live', Symbol='Gorasp2', Aliases='0610011A07Rik, 2410043M02Rik, 5730520M13Rik, 9430094F20Rik, AW552058, GOLPH2, GRASP55, GRS2, p59', description='golgi reassembly stacking protein 2', other_designations='golgi reassembly-stacking protein of 55 kDa', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=70661489, end_position_on_the_genomic_accession=70691725, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70233, CurrentID=0, Status='live', Symbol='Cd2bp2', Aliases='1500011B02Rik, 2410024K20Rik, AW048183', description='CD2 antigen (cytoplasmic tail) binding protein 2', other_designations='CD2 cytoplasmic domain-binding protein 2|CD2 tail-binding protein 2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127191660, end_position_on_the_genomic_accession=127196077, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70235, CurrentID=0, Status='live', Symbol='Poc1a', Aliases='2510040D07Rik, Wdr51a, cha', description='POC1 centriolar protein A', other_designations='WD repeat domain 51A|WD repeat-containing protein 51A', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106281024, end_position_on_the_genomic_accession=106349891, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  70237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70237, CurrentID=0, Status='live', Symbol='Bhlhb9', Aliases='2700087I09Rik', description='basic helix-loop-helix domain containing, class B9', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135841639, end_position_on_the_genomic_accession=135891081, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70238, CurrentID=0, Status='live', Symbol='Rnf168', Aliases='3110001H15Rik', description='ring finger protein 168', other_designations='ring fnger protein 168', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32277461, end_position_on_the_genomic_accession=32301442, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  70239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70239, CurrentID=0, Status='live', Symbol='Gtf3c5', Aliases='2700084A09Rik, TFIIIC63, TFiiiC2-63', description='general transcription factor IIIC, polypeptide 5', other_designations='TF3C-epsilon|TFIIIC 63 kDa subunit|TFIIICepsilon|general transcription factor III polypeptide 5|transcription factor IIIC 63 kDa subunit|transcription factor IIIC subunit epsilon', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28566245, end_position_on_the_genomic_accession=28583746, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  70240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70240, CurrentID=0, Status='live', Symbol='Ufsp1', Aliases='2700038N03Rik, D5Ertd655e', description='UFM1-specific peptidase 1', other_designations='', map_location='5 76.33 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137294669, end_position_on_the_genomic_accession=137295664, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  70247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70247, CurrentID=0, Status='live', Symbol='Psmd1', Aliases='2410026J11Rik, P112, S1', description='proteasome (prosome, macropain) 26S subunit, non-ATPase, 1', other_designations='26S proteasome regulatory subunit RPN2|26S proteasome regulatory subunit S1', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86064619, end_position_on_the_genomic_accession=86139295, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  70248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70248, CurrentID=0, Status='live', Symbol='Dazap1', Aliases='2410042M16Rik, mPrrp', description='DAZ associated protein 1', other_designations='deleted in azoospermia-associated protein 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80261471, end_position_on_the_genomic_accession=80288413, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  332396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332396, CurrentID=0, Status='live', Symbol='Kcnk18', Aliases='Gm781, Tresk, Tresk-2, Trik', description='potassium channel, subfamily K, member 18', other_designations='2P K ion channel TRESK|TWIK-related individual K+ channel|TWIK-related spinal cord K+ channel|two-pore-domain potassium channel TRESK', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=59219648, end_position_on_the_genomic_accession=59237370, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  332397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332397, CurrentID=0, Status='live', Symbol='Nanos1', Aliases='NOS-1', description='nanos homolog 1 (Drosophila)', other_designations='nanos 1', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60755987, end_position_on_the_genomic_accession=60759914, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  70257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70257, CurrentID=0, Status='live', Symbol='2010107E04Rik', Aliases='9430003J03Rik, AU043134, AV124504, Mp68', description='RIKEN cDNA 2010107E04 gene', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111961375, end_position_on_the_genomic_accession=111966977, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70261, CurrentID=0, Status='live', Symbol='Chp2', Aliases='2010110P09Rik, Cbhp2', description='calcineurin-like EF hand protein 2', other_designations='hepatocellular carcinoma-associated antigen 520 homolog', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122208621, end_position_on_the_genomic_accession=122222539, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70266, CurrentID=0, Status='live', Symbol='Ccbl1', Aliases='2010009K05Rik, AI182306, Gtk, Kat1, KatI', description='cysteine conjugate-beta lyase 1', other_designations='cysteine-S-conjugate beta-lyase|cytoplasmic (glutamine transaminase K, kyneurenine aminotransferase)|glutamine transaminase K|glutamine--phenylpyruvate transaminase|kynurenine aminotransferase I|kynurenine--oxoglutarate transaminase I', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30185124, end_position_on_the_genomic_accession=30206295, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  70274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70274, CurrentID=0, Status='live', Symbol='Ly6g6e', Aliases='2310011I02Rik, G6e', description='lymphocyte antigen 6 complex, locus G6E', other_designations='', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35076819, end_position_on_the_genomic_accession=35078804, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  332427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332427, CurrentID=0, Status='live', Symbol='Lyg2', Aliases='Gm776', description='lysozyme G-like 2', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37905583, end_position_on_the_genomic_accession=37916493, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70285, CurrentID=0, Status='live', Symbol='Rpf1', Aliases='2210420E24Rik, 2310066N05Rik, Bxdc5', description='ribosome production factor 1 homolog (S. cerevisiae)', other_designations='RNA processing factor 1|brix domain containing 5|brix domain-containing protein 5|ribosome biogenesis protein RPF1', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146506344, end_position_on_the_genomic_accession=146521431, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  70291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70291, CurrentID=0, Status='live', Symbol='Mkrn2os', Aliases='2510049J12Rik', description='makorin, ring finger protein 2, opposite strand', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115583547, end_position_on_the_genomic_accession=115592576, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70292, CurrentID=0, Status='live', Symbol='Afap1', Aliases='2600003E23Rik, 9630044L16Rik, AI848729, Afap, mKIAA3018', description='actin filament associated protein 1', other_designations='110 kDa actin filament-associated protein|AFAP-110', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35893317, end_position_on_the_genomic_accession=36003923, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  70294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70294, CurrentID=0, Status='live', Symbol='Rnf126', Aliases='2610010O19Rik', description='ring finger protein 126', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79758515, end_position_on_the_genomic_accession=79766952, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70296, CurrentID=0, Status='live', Symbol='Tbc1d13', Aliases='2600014A06Rik, BC025586', description='TBC1 domain family, member 13', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30133751, end_position_on_the_genomic_accession=30152013, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  70297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70297, CurrentID=0, Status='live', Symbol='Gcc2', Aliases='0610043A03Rik, 2210420P05Rik, 2600014C01Rik, AW112121, mKIAA0336', description='GRIP and coiled-coil domain containing 2', other_designations='185 kDa Golgi coiled-coil protein|GCC185', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=58254296, end_position_on_the_genomic_accession=58305599, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  70300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70300, CurrentID=0, Status='live', Symbol='Fuz', Aliases='2600013E07Rik, b2b1273Clo', description='fuzzy planar cell polarity protein', other_designations='fuzzy homolog', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44895265, end_position_on_the_genomic_accession=44900624, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70310, CurrentID=0, Status='live', Symbol='Plscr3', Aliases='2210403O21Rik, 2610037N06Rik, ESTM3, X83310', description='phospholipid scramblase 3', other_designations='Ca 2+ dependent phospholipid scramblase 3|PL scramblase 3|ca(2+)-dependent phospholipid scramblase 3', map_location='11 42.9 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69846251, end_position_on_the_genomic_accession=69852058, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70312, CurrentID=0, Status='live', Symbol='Cactin', Aliases='2510012J08Rik', description='cactin, spliceosome C complex subunit', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81321103, end_position_on_the_genomic_accession=81326251, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70314, CurrentID=0, Status='live', Symbol='Rabep2', Aliases='2610011A08Rik, AI462746, Fra', description='rabaptin, RAB GTPase binding effector protein 2', other_designations='fos-related antigen|rabaptin-5beta', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126428416, end_position_on_the_genomic_accession=126445907, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  70315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70315, CurrentID=0, Status='live', Symbol='Hdac8', Aliases='2610007D20Rik', description='histone deacetylase 8', other_designations='HD8', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102284640, end_position_on_the_genomic_accession=102509686, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  70316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70316, CurrentID=0, Status='live', Symbol='Ndufab1', Aliases='2210401F17Rik, 2310039H15Rik, 2610003B19Rik, 8kDa, 9130423F15Rik, ACP, CI-SDAP, SDAP', description='NADH dehydrogenase (ubiquinone) 1, alpha/beta subcomplex, 1', other_designations='NADH-ubiquinone oxidoreductase 9.6 kDa subunit', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122088044, end_position_on_the_genomic_accession=122101877, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  70317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70317, CurrentID=0, Status='live', Symbol='Arl16', Aliases='2600005N12Rik', description='ADP-ribosylation factor-like 16', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120464326, end_position_on_the_genomic_accession=120467600, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70325, CurrentID=0, Status='live', Symbol='Pigw', Aliases='2610044A17Rik, Gwt1', description='phosphatidylinositol glycan anchor biosynthesis, class W', other_designations='PIG-W|phosphatidylinositol glycan, class W', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84876312, end_position_on_the_genomic_accession=84880285, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70333, CurrentID=0, Status='live', Symbol='Cd3eap', Aliases='1190028F09, 2610103M17Rik, ASE-1, Ase1, CAST, PAF49', description='CD3E antigen, epsilon polypeptide associated protein', other_designations='A34.5|CD3 epsilon associated protein; antisense to Ercc1|CD3-epsilon-associated protein|CD3E-associated protein|RNA polymerase I-associated factor PAF49|anti-sense to ERCC-1 protein', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19356007, end_position_on_the_genomic_accession=19359483, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70335, CurrentID=0, Status='live', Symbol='Reep6', Aliases='0610011M24Rik, Dp1l1', description='receptor accessory protein 6', other_designations='TB2 protein-like 1|deleted in polyposis 1-like 1|polyposis locus protein 1-like 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80330003, end_position_on_the_genomic_accession=80336441, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70337, CurrentID=0, Status='live', Symbol='Iyd', Aliases='0610009A07Rik, AI265638, Dehal1', description='iodotyrosine deiodinase', other_designations='IYD-1|iodotyrosine dehalogenase 1 protein', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=3540279, end_position_on_the_genomic_accession=3554877, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  70348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70348, CurrentID=0, Status='live', Symbol='Ube2cbp', Aliases='2610018I03Rik, Ube3d', description='ubiquitin-conjugating enzyme E2C binding protein', other_designations='UbcH10 binding protein with a hect-like domain|ubcH10-binding protein with a HECT-like domain|ubiquitin-conjugating enzyme E2C-binding protein', map_location='9 E3.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=86307234, end_position_on_the_genomic_accession=86465181, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  70349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70349, CurrentID=0, Status='live', Symbol='Copb1', Aliases='2610019B04Rik', description='coatomer protein complex, subunit beta 1', other_designations='beta-COP|beta-coat protein', map_location='7 59.31 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114215559, end_position_on_the_genomic_accession=114254680, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  70350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70350, CurrentID=0, Status='live', Symbol='Basp1', Aliases='2610024P12Rik, CAP-23, CAP23, Ckap3, NAP-22, NAP22', description='brain abundant, membrane attached signal protein 1', other_designations='22 kDa neuronal tissue-enriched acidic protein|cytoskeleton associated protein 3|neuronal axonal membrane protein NAP-22', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=25363277, end_position_on_the_genomic_accession=25413764, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  70351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70351, CurrentID=0, Status='live', Symbol='Ppp4r1', Aliases='3110001J10Rik, Pp4r1', description='protein phosphatase 4, regulatory subunit 1', other_designations='', map_location='17 E1.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65782398, end_position_on_the_genomic_accession=65841926, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  70354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70354, CurrentID=0, Status='live', Symbol='Secisbp2l', Aliases='3110001I20Rik, AI504340, C630011I23', description='SECIS binding protein 2-like', other_designations='SECIS-binding protein 2-like', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125736986, end_position_on_the_genomic_accession=125782870, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  70355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70355, CurrentID=0, Status='live', Symbol='Gprc5c', Aliases='1110028I06Rik, 3200002M13Rik, Raig3', description='G protein-coupled receptor, family C, group 5, member C', other_designations='RAIG-3|retinoic acid inducible protein 3|retinoic acid-induced gene 3 protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114851146, end_position_on_the_genomic_accession=114872617, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70356, CurrentID=0, Status='live', Symbol='St13', Aliases='1110007I03Rik, 3110002K08Rik, AW555194, HIP, HOP, HSPABP, HSPABP1, PRO0786, SNC6, p48', description='suppression of tumorigenicity 13', other_designations='Hsp70 interacting protein|protein ST13 homolog', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81365044, end_position_on_the_genomic_accession=81399694, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70357, CurrentID=0, Status='live', Symbol='Kcnip1', Aliases='KCHIP1, Kchip1.2', description='Kv channel-interacting protein 1', other_designations='A-type potassium channel modulatory protein 1|A-type potassium channel modulatory protein 1.2|potassium channel-interacting protein 1', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=33629339, end_position_on_the_genomic_accession=33993193, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70358, CurrentID=0, Status='live', Symbol='Steap1', Aliases='2410007B19Rik, Prss24, Steap', description='six transmembrane epithelial antigen of the prostate 1', other_designations='six-transmembrane epithelial antigen of prostate 1', map_location='5 3.2 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=5736322, end_position_on_the_genomic_accession=5749317, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70359, CurrentID=0, Status='live', Symbol='Gtpbp3', Aliases='2410009F13Rik, AI607903', description='GTP binding protein 3', other_designations='GTP-binding protein 3', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71487870, end_position_on_the_genomic_accession=71499623, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70361, CurrentID=0, Status='live', Symbol='Lman1', Aliases='2610020P13Rik, AI326273, AU043785, C730041J05, ERGIC53, F5F8D, MCFD1, MR60, P58, gp58', description='lectin, mannose-binding, 1', other_designations='ER-Golgi intermediate compartment 53 kDa protein', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65980739, end_position_on_the_genomic_accession=66002635, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  70362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70362, CurrentID=0, Status='live', Symbol='Actl10', Aliases='1700007I08Rik', description='actin-like 10', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154551776, end_position_on_the_genomic_accession=154553276, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  70363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70363, CurrentID=0, Status='live', Symbol='Fam135b', Aliases='1700010C24Rik, A830008O07Rik', description='family with sequence similarity 135, member B', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=71438951, end_position_on_the_genomic_accession=71728436, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  70365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70365, CurrentID=0, Status='live', Symbol='Speer5-ps1', Aliases='1700007C10Rik, SPEER-5, Speer5-ps', description='spermatogenesis associated glutamate (E)-rich protein 5, pseudogene 1', other_designations='sperm-associated glutamate (E)-rich protein 5, pseudogene|spermatogenic-associated glutamate (E)-rich protein 5, pseudogene', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=44170447, end_position_on_the_genomic_accession=44219585, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70369, CurrentID=0, Status='live', Symbol='Bag5', Aliases='1600025G07Rik, 4930405J06Rik', description='BCL2-associated athanogene 5', other_designations='BAG-5|BAG-family molecular chaperone regulator-5 (BAG-5)|bcl-2-associated athanogene 5', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111709488, end_position_on_the_genomic_accession=111713256, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70370, CurrentID=0, Status='live', Symbol='Fbln7', Aliases='1600015H20Rik, AI464334, TM14', description='fibulin 7', other_designations='FIBL-7|extracellular matrix protein TM14', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128863862, end_position_on_the_genomic_accession=128897034, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70373, CurrentID=0, Status='live', Symbol='Gpatch2l', Aliases='1700020O03Rik, 4732487G21Rik, A630090I05Rik, Kiaa1152', description='G patch domain containing 2 like', other_designations='', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86241838, end_position_on_the_genomic_accession=86291793, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70375, CurrentID=0, Status='live', Symbol='Ica1l', Aliases='1700030B17Rik, Als2cr15', description='islet cell autoantigen 1-like', other_designations='Ica69-related protein|amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 15|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 15 protein homolog', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59989068, end_position_on_the_genomic_accession=60043166, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  70377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70377, CurrentID=0, Status='live', Symbol='Derl3', Aliases='1810006I20Rik, 1810063P04Rik, IZP6, derlin-3', description='Der1-like domain family, member 3', other_designations='degradation in endoplasmic reticulum protein 3|der1-like protein 3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75893080, end_position_on_the_genomic_accession=75895941, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70380, CurrentID=0, Status='live', Symbol='Mospd1', Aliases='1810018L05Rik', description='motile sperm domain containing 1', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53344594, end_position_on_the_genomic_accession=53370562, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70381, CurrentID=0, Status='live', Symbol='Tecpr1', Aliases='2210010N04Rik, mKIAA1358', description='tectonin beta-propeller repeat containing 1', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144194442, end_position_on_the_genomic_accession=144223608, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  70382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70382, CurrentID=0, Status='live', Symbol='Kctd2', Aliases='2310012I15Rik', description='potassium channel tetramerisation domain containing 2', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115420126, end_position_on_the_genomic_accession=115431274, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70383, CurrentID=0, Status='live', Symbol='Cox10', Aliases='2410004F01Rik, AU042636', description='cytochrome c oxidase assembly protein 10', other_designations='COX10 homolog, cytochrome c oxidase assembly protein, heme A: farnesyltransferase|heme O synthase', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=63962627, end_position_on_the_genomic_accession=64079472, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  70385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70385, CurrentID=0, Status='live', Symbol='Spdl1', Aliases='1700018I02Rik, 2600001J17Rik, 2810049B11Rik, AA409762, Ccdc99', description='spindle apparatus coiled-coil protein 1', other_designations='coiled-coil domain containing 99|coiled-coil domain-containing protein 99|spindle apparatus coiled-coil domain-containing protein 1', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=34809185, end_position_on_the_genomic_accession=34833664, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  70387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70387, CurrentID=0, Status='live', Symbol='Ttc9c', Aliases='2210019E14Rik, 6330408J23Rik', description='tetratricopeptide repeat domain 9C', other_designations='TPR repeat protein 9C', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8809066, end_position_on_the_genomic_accession=8819780, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70392, CurrentID=0, Status='live', Symbol='Asb12', Aliases='2310036C05Rik, Asb-12', description='ankyrin repeat and SOCS box-containing 12', other_designations='ankyrin repeat and SOCS box-containing protein 12|ankyrin repeat domain-containing SOCS box protein 12', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=95470199, end_position_on_the_genomic_accession=95557993, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  70394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70394, CurrentID=0, Status='live', Symbol='Kptn', Aliases='2310042D10Rik, 2E4, C030013F01Rik', description='kaptin', other_designations='actin-binding protein', map_location='7 8.75 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16119876, end_position_on_the_genomic_accession=16127516, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  70396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70396, CurrentID=0, Status='live', Symbol='Asnsd1', Aliases='2210409M21Rik', description='asparagine synthetase domain containing 1', other_designations='', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53344617, end_position_on_the_genomic_accession=53352752, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  70397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70397, CurrentID=0, Status='live', Symbol='Tmem70', Aliases='1110020A09Rik, 2210416J16Rik', description='transmembrane protein 70', other_designations='', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16665191, end_position_on_the_genomic_accession=16678275, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  70405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70405, CurrentID=0, Status='live', Symbol='Calml3', Aliases='2310068O22Rik, AI326174', description='calmodulin-like 3', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3802893, end_position_on_the_genomic_accession=3804318, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  70408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70408, CurrentID=0, Status='live', Symbol='Polr3f', Aliases='2810411G20Rik, 3010019O03Rik, 3110032A07Rik, RPC39, RPC6', description='polymerase (RNA) III (DNA directed) polypeptide F', other_designations='DNA-directed RNA polymerase III subunit F|RNA polymerase III subunit C6', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144527745, end_position_on_the_genomic_accession=144541779, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70415, CurrentID=0, Status='live', Symbol='Stk26', Aliases='2610018G03Rik, Mst4', description='serine/threonine kinase 26', other_designations='Mst3 and SOK1-related kinase|STE20-like kinase MST4|mammalian STE20-like protein kinase 4|serine/threonine protein kinase MST4|serine/threonine-protein kinase MST4', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=50829694, end_position_on_the_genomic_accession=50893098, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70417, CurrentID=0, Status='live', Symbol='Megf10', Aliases='3000002B06Rik, Gm331', description='multiple EGF-like-domains 10', other_designations='multiple EGF-like domains protein 10', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=57133090, end_position_on_the_genomic_accession=57297467, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  70419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70419, CurrentID=0, Status='live', Symbol='2810408A11Rik', Aliases='1700019L01Rik, 4930470J18Rik', description='RIKEN cDNA 2810408A11 gene', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69897358, end_position_on_the_genomic_accession=69901721, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70420, CurrentID=0, Status='live', Symbol='Arpin', Aliases='2610034B18Rik', description='actin-related protein 2/3 complex inhibitor', other_designations='UPF0552 protein C15orf38 homolog|arp2/3 inhibition protein', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79925359, end_position_on_the_genomic_accession=79935264, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  70422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70422, CurrentID=0, Status='live', Symbol='Ints2', Aliases='2810417D08Rik, AA408260, AI987735, mKIAA1287', description='integrator complex subunit 2', other_designations='int2', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86210681, end_position_on_the_genomic_accession=86257612, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  70423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70423, CurrentID=0, Status='live', Symbol='Tspan15', Aliases='1110036D12Rik, 2700063A19Rik, AW048364, Tm4sf15', description='tetraspanin 15', other_designations='transmembrane 4 superfamily member 15|tspan-15', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62185396, end_position_on_the_genomic_accession=62231277, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70425, CurrentID=0, Status='live', Symbol='Csnk1g3', Aliases='3300002K07Rik, C330049O21Rik', description='casein kinase 1, gamma 3', other_designations='CKI-gamma 3', map_location='18 D1-D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=53862082, end_position_on_the_genomic_accession=53955684, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70426, CurrentID=0, Status='live', Symbol='Tekt5', Aliases='3300001K11Rik', description='tektin 5', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10357952, end_position_on_the_genomic_accession=10395465, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70427, CurrentID=0, Status='live', Symbol='Mier2', Aliases='2700087H15Rik, mKIAA1193', description='MIER family member 2', other_designations='mesoderm induction early response 1, family member 2|mi-er2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79540245, end_position_on_the_genomic_accession=79555631, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  70428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70428, CurrentID=0, Status='live', Symbol='Polr3b', Aliases='2700078H01Rik, A330032P03Rik, C85372, RPC2', description='polymerase (RNA) III (DNA directed) polypeptide B', other_designations='C128|DNA-directed RNA polymerase III 127.6 kDa polypeptide|DNA-directed RNA polymerase III B|DNA-directed RNA polymerase III subunit B|RNA polymerase III subunit C2|RNA polymerase III subunit RPC2 homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=84621456, end_position_on_the_genomic_accession=84727178, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  70430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70430, CurrentID=0, Status='live', Symbol='Tbce', Aliases='2610206D02Rik, C530005D02Rik, pmn', description='tubulin-specific chaperone E', other_designations='progressive motor neuropathy|tubulin-folding cofactor E', map_location='13 5.29 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=13997949, end_position_on_the_genomic_accession=14039638, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  70432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70432, CurrentID=0, Status='live', Symbol='Rufy2', Aliases='2610111M19Rik, AI844453, Denn, LZ-FYVE, ZFYVE13', description='RUN and FYVE domain-containing 2', other_designations='leucine zipper FYVE-finger protein', map_location='10 32.51 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62980223, end_position_on_the_genomic_accession=63018742, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  70433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70433, CurrentID=0, Status='live', Symbol='Draxin', Aliases='2610109H07Rik, neucrin', description='dorsal inhibitory axon guidance protein', other_designations='dorsal repulsive axon guidance protein', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148098437, end_position_on_the_genomic_accession=148132927, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  332578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332578, CurrentID=0, Status='live', Symbol='Lcn10', Aliases='9230112J07Rik', description='lipocalin 10', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25682726, end_position_on_the_genomic_accession=25686081, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  332579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332579, CurrentID=0, Status='live', Symbol='Card9', Aliases='Gm782', description='caspase recruitment domain family, member 9', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26352236, end_position_on_the_genomic_accession=26360979, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  70439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70439, CurrentID=0, Status='live', Symbol='Taf15', Aliases='2610111C21Rik, 68kDa, TAFII68, Taf2n', description='TATA-box binding protein associated factor 15', other_designations='TAF15 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF15 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 68 kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase II, N, 68kD (RNA binding protein 56)', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83473071, end_position_on_the_genomic_accession=83506743, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  70445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70445, CurrentID=0, Status='live', Symbol='Cd248', Aliases='2610111G01Rik, AI842296, Cd164l1, Tem1', description='CD248 antigen, endosialin', other_designations='CD164 sialomucin-like 1|tumor endothelial marker 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5068078, end_position_on_the_genomic_accession=5070640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  70450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70450, CurrentID=0, Status='live', Symbol='Unc13d', Aliases='2610108D09Rik, Jinx, Munc13-4, mFLJ00067', description='unc-13 homolog D (C. elegans)', other_designations='likely ortholog of H. sapiens unc-13 homolog D (UNC13D)', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116062095, end_position_on_the_genomic_accession=116078195, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  70451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70451, CurrentID=0, Status='live', Symbol='Dhrs13', Aliases='2610209N15Rik', description='dehydrogenase/reductase (SDR family) member 13', other_designations='androgen-regulated short-chain dehydrogenase/reductase 1; prostate short-chain dehydrogenase reductase 1; likely ortholog of mouse cell line MC/9.IL4 derived transcript 1; CGI-82 protein; HCV core-binding protein|short chain dehydrogenase/reductase family 7C member 5', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78032270, end_position_on_the_genomic_accession=78037866, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  70454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70454, CurrentID=0, Status='live', Symbol='Cenpl', Aliases='2610300B10Rik, AW121806, AW550697, CENP-L', description='centromere protein L', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161070718, end_position_on_the_genomic_accession=161086343, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  70456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70456, CurrentID=0, Status='live', Symbol='Mpc2', Aliases='0610006C01Rik, 2010002I07Rik, 2610205H19Rik, AA108335, Brp44, ESTM43', description='mitochondrial pyruvate carrier 2', other_designations='brain protein 44', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165461208, end_position_on_the_genomic_accession=165481214, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  70458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70458, CurrentID=0, Status='live', Symbol='2610318N02Rik', Aliases='AL033325', description='RIKEN cDNA 2610318N02 gene', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17113398, end_position_on_the_genomic_accession=17125197, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  70461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70461, CurrentID=0, Status='live', Symbol='Crtc3', Aliases='2610312F20Rik, 6332415K15Rik, AI429792, BC011210, TORC3', description='CREB regulated transcription coactivator 3', other_designations='TORC-3|transducer of CREB protein 3|transducer of regulated cAMP response element-binding protein 3', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80586627, end_position_on_the_genomic_accession=80688942, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  70465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70465, CurrentID=0, Status='live', Symbol='Wdr77', Aliases='2610003I18Rik, 2610312E17Rik, C79984, p44/MEP50', description='WD repeat domain 77', other_designations='MEP-50|WD repeat-containing protein 77', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105959498, end_position_on_the_genomic_accession=105969760, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70466, CurrentID=0, Status='live', Symbol='Ckap2l', Aliases='2010016H04Rik, 2610318C08Rik, AV070319, radmis, rdm', description='cytoskeleton associated protein 2-like', other_designations='radial fiber and mitotic spindle protein', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129268210, end_position_on_the_genomic_accession=129297230, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70470, CurrentID=0, Status='live', Symbol='Rprd1b', Aliases='2610304G08Rik, 2810446G03Rik, Crept', description='regulation of nuclear pre-mRNA domain containing 1B', other_designations='cell cycle-related and expression-elevated protein in tumor', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158028489, end_position_on_the_genomic_accession=158078207, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70472, CurrentID=0, Status='live', Symbol='Atad2', Aliases='2610509G12Rik', description='ATPase family, AAA domain containing 2', other_designations='', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58094047, end_position_on_the_genomic_accession=58135139, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  70478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70478, CurrentID=0, Status='live', Symbol='Mipep', Aliases='5730405E07Rik, MIP', description='mitochondrial intermediate peptidase', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=60784562, end_position_on_the_genomic_accession=60905481, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  70481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70481, CurrentID=0, Status='live', Symbol='Pnma1', Aliases='5730402C15Rik', description='paraneoplastic antigen MA1', other_designations='', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84146131, end_position_on_the_genomic_accession=84148489, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  70484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70484, CurrentID=0, Status='live', Symbol='Slc35d2', Aliases='5730408I21Rik, SQV7L, UGTrel8, hfrc, mUGTrel8', description='solute carrier family 35, member D2', other_designations='UDP-galactose transporter-related protein 8|UDP-glucuronic acid/UDP-N-acetylgalactosamine transporter', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64096308, end_position_on_the_genomic_accession=64129381, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70495, CurrentID=0, Status='live', Symbol='Atp6ap2', Aliases='(P)RR, 5730403E06Rik, APT6M8-9, ATP6M8-9, Atp6ip2, M8-9', description='ATPase, H+ transporting, lysosomal accessory protein 2', other_designations='(pro)renin receptor|ATPase H(+)-transporting lysosomal-interacting protein 2|ATPase, H+ transporting, lysosomal interacting protein 1|ATPase, H+ transporting, lysosomal interacting protein 2|renin/prorenin receptor', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=12587759, end_position_on_the_genomic_accession=12617051, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70497, CurrentID=0, Status='live', Symbol='Arhgap17', Aliases='5730403H17Rik, Nadrin, Nadrin2, Rich1, Wbp15', description='Rho GTPase activating protein 17', other_designations='WW domain-binding protein 15|neuron-associated developmentally regulated protein|rho-type GTPase-activating protein 17', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=123279144, end_position_on_the_genomic_accession=123372511, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  70503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70503, CurrentID=0, Status='live', Symbol='Ddo', Aliases='5330420D20Rik, 5730402C02Rik, AI467244', description='D-aspartate oxidase', other_designations='DASOX', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40629976, end_position_on_the_genomic_accession=40674915, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70508, CurrentID=0, Status='live', Symbol='Bbx', Aliases='5530401J07Rik, 5730403O13Rik', description='bobby sox homolog (Drosophila)', other_designations='HMG box-containing protein 2', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=50191844, end_position_on_the_genomic_accession=50432461, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  70510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70510, CurrentID=0, Status='live', Symbol='Rnf167', Aliases='0610010G05Rik, 5730408C10Rik, AU019305, AV328608', description='ring finger protein 167', other_designations='novel PA domain and C3HC4 type zinc finger (RING finger) containing protein', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70647589, end_position_on_the_genomic_accession=70651421, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70511, CurrentID=0, Status='live', Symbol='Eef2kmt', Aliases='5730409G15Rik, Fam86, Fam86a, eEF2-KMT', description='eukaryotic elongation factor 2 lysine methyltransferase', other_designations='eEF2-lysine methyltransferase|family with sequence similarity 86|protein FAM86A|putative protein N-methyltransferase FAM86A', map_location='16 A2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5244155, end_position_on_the_genomic_accession=5255956, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  70527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70527, CurrentID=0, Status='live', Symbol='Stambp', Aliases='5330424L14Rik, 5730422L11Rik, AW107289, Amsh, mKIAA4198', description='STAM binding protein', other_designations='associated molecule with the SH3 domain of STAM', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83543206, end_position_on_the_genomic_accession=83575770, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70530, CurrentID=0, Status='live', Symbol='Lrfn2', Aliases='5730420O05Rik, SALM1, mKIAA1246', description='leucine rich repeat and fibronectin type III domain containing 2', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48931566, end_position_on_the_genomic_accession=49097590, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  70533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70533, CurrentID=0, Status='live', Symbol='Btf3l4', Aliases='4632412E09Rik, 5730434I03Rik, AI316863, AU022207, AW552446', description='basic transcription factor 3-like 4', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108814295, end_position_on_the_genomic_accession=108834297, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70536, CurrentID=0, Status='live', Symbol='Qpct', Aliases='5730422A13Rik', description='glutaminyl-peptide cyclotransferase (glutaminyl cyclase)', other_designations='QC|glutaminyl cyclase|glutaminyl-tRNA cyclotransferase', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=79051906, end_position_on_the_genomic_accession=79090268, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70544, CurrentID=0, Status='live', Symbol='Tmem242', Aliases='1110008A10Rik, 2310046K16Rik, 5730437N04Rik', description='transmembrane protein 242', other_designations='UPF0463 transmembrane protein C6orf35 homolog', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=5410864, end_position_on_the_genomic_accession=5440260, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70546, CurrentID=0, Status='live', Symbol='Zdhhc2', Aliases='5730415P04Rik, 6430583A19Rik', description='zinc finger, DHHC domain containing 2', other_designations='DHHC-2|zinc finger DHHC domain-containing protein 2', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40423000, end_position_on_the_genomic_accession=40489592, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70549, CurrentID=0, Status='live', Symbol='Tln2', Aliases='5730421P04Rik, AI507121, AI787438, AL118320, mKIAA0320', description='talin 2', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=67217085, end_position_on_the_genomic_accession=67634036, orientation='minus', exon_count=62, OMIM='', no_hdr0=''),
  70551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70551, CurrentID=0, Status='live', Symbol='Tmtc4', Aliases='4930403J22Rik, 5730419O14Rik', description='transmembrane and tetratricopeptide repeat containing 4', other_designations='', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122918971, end_position_on_the_genomic_accession=122983283, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  70552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70552, CurrentID=0, Status='live', Symbol='Lrrc56', Aliases='5730427C23Rik, BB110509, mFLJ00101', description='leucine rich repeat containing 56', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141194116, end_position_on_the_genomic_accession=141210055, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  70556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70556, CurrentID=0, Status='live', Symbol='Slc25a33', Aliases='5730438N18Rik, Pnc1', description='solute carrier family 25, member 33', other_designations='mitochondrial carrier protein|novel mitochondrial carrier protein domains containing protein', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149744036, end_position_on_the_genomic_accession=149774267, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70560, CurrentID=0, Status='live', Symbol='Wars2', Aliases='5730427B17Rik, 9430020O07Rik, AI413375, TrpRS', description='tryptophanyl tRNA synthetase 2 (mitochondrial)', other_designations='(Mt)TrpRS|tryptophanyl-tRNA synthetase, mitochondrial', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=99140618, end_position_on_the_genomic_accession=99220203, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70561, CurrentID=0, Status='live', Symbol='Txndc16', Aliases='5730420B22Rik, C77647', description='thioredoxin domain containing 16', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45134448, end_position_on_the_genomic_accession=45219553, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  70564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70564, CurrentID=0, Status='live', Symbol='Fam213a', Aliases='5730469M10Rik, AU040822, Pamm', description='family with sequence similarity 213, member A', other_designations='UPF0765 protein C10orf58 homolog|peroxiredoxin-like 2 activated in M-CSF stimulated monocytes|redox-regulatory protein PAMM', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=40993732, end_position_on_the_genomic_accession=41013829, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70567, CurrentID=0, Status='live', Symbol='Fra10ac1', Aliases='5730455O13Rik, AI448293', description='FRA10AC1 homolog (human)', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38188477, end_position_on_the_genomic_accession=38224513, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  70568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70568, CurrentID=0, Status='live', Symbol='Cpne3', Aliases='5430428M23Rik, 5730450C07Rik, C85108, CPN3', description='copine III', other_designations='', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=19519252, end_position_on_the_genomic_accession=19570104, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  332713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332713, CurrentID=0, Status='live', Symbol='BC051628', Aliases='1700026M20Rik', description='cDNA sequence BC051628', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181213978, end_position_on_the_genomic_accession=181222856, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  70571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70571, CurrentID=0, Status='live', Symbol='Tcerg1l', Aliases='5730476P14Rik', description='transcription elongation regulator 1-like', other_designations='transcription elongation regulator 1-like protein-like', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=138208972, end_position_on_the_genomic_accession=138397730, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70572, CurrentID=0, Status='live', Symbol='Ipo5', Aliases='1110011C18Rik, 5730478E03Rik, AA409333, C76941, IMB3, Kpnb3, Ranbp5', description='importin 5', other_designations='RAN binding protein 5|imp5|importin subunit beta-3|karyopherin (importin) beta 3|karyopherin beta-3|ran-binding protein 5', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=120896839, end_position_on_the_genomic_accession=120948046, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  70573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70573, CurrentID=0, Status='live', Symbol='Tbccd1', Aliases='5730478M09Rik, AU015083, BB165076', description='TBCC domain containing 1', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22813215, end_position_on_the_genomic_accession=22857701, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70574, CurrentID=0, Status='live', Symbol='Cpm', Aliases='1110060I01Rik, 5730456K23Rik, AA589379, E030045M14Rik', description='carboxypeptidase M', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117629500, end_position_on_the_genomic_accession=117687352, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70575, CurrentID=0, Status='live', Symbol='Gfod2', Aliases='5730466C23Rik', description='glucose-fructose oxidoreductase domain containing 2', other_designations='', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105716110, end_position_on_the_genomic_accession=105758668, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70579, CurrentID=0, Status='live', Symbol='Zc3h11a', Aliases='1110003F06Rik, 5730454B08Rik, G630041M05Rik, Zc3hdc11a, Zc3hh11a, mKIAA0663', description='zinc finger CCCH type containing 11A', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133619868, end_position_on_the_genomic_accession=133661399, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  70584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70584, CurrentID=0, Status='live', Symbol='Pak4', Aliases='5730488L07Rik, AW555722, mKIAA1142', description='p21 protein (Cdc42/Rac)-activated kinase 4', other_designations='PAK-4|p21-activated kinase 4', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28558819, end_position_on_the_genomic_accession=28598186, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70591, CurrentID=0, Status='live', Symbol='5730455P16Rik', Aliases='', description='RIKEN cDNA 5730455P16 gene', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80360492, end_position_on_the_genomic_accession=80378052, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  70598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70598, CurrentID=0, Status='live', Symbol='Filip1', Aliases='5730485H21Rik, AI843021', description='filamin A interacting protein 1', other_designations='FILIP', map_location='9 E2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=79805018, end_position_on_the_genomic_accession=79978050, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70599, CurrentID=0, Status='live', Symbol='Ssfa2', Aliases='5730488C15Rik, CS-1, CS1, KRAP, SPAG13, mKIAA1927', description='sperm specific antigen 2', other_designations='ki-ras-induced actin-interacting protein', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=79635373, end_position_on_the_genomic_accession=79672966, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  70601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70601, CurrentID=0, Status='live', Symbol='Ecd', Aliases='5730461K03Rik, Hsgt1', description='ecdysoneless homolog (Drosophila)', other_designations='protein SGT1 homolog|suppressor of S. cerevisiae gcr2', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20319859, end_position_on_the_genomic_accession=20348121, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  70603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70603, CurrentID=0, Status='live', Symbol='Mutyh', Aliases='5730495A01Rika, Mutyhalpha, Mutyhb, Mutyhbeta, Mutyhc, Myh, Mutyh', description='mutY homolog (E. coli)', other_designations='a/G-specific adenine DNA glycosylase|adenine-DNA glycosylase|mMYH|mutY homolog alpha', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116807682, end_position_on_the_genomic_accession=116819439, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  70604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70604, CurrentID=0, Status='live', Symbol='Dnajb14', Aliases='5730496F10Rik', description='DnaJ heat shock protein family (Hsp40) member B14', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 14', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=137867675, end_position_on_the_genomic_accession=137916563, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70605, CurrentID=0, Status='live', Symbol='Zdhhc24', Aliases='5730496N17Rik, Leng4', description='zinc finger, DHHC domain containing 24', other_designations='DHHC-24|leukocyte receptor cluster (LRC) member 4|zinc finger DHHC domain-containing protein 24', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4878668, end_position_on_the_genomic_accession=4885397, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  70611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70611, CurrentID=0, Status='live', Symbol='Fbxo33', Aliases='5730501N20Rik, AI642135', description='F-box protein 33', other_designations='AIG30-12-1', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=59200655, end_position_on_the_genomic_accession=59219483, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70612, CurrentID=0, Status='live', Symbol='Tmem230', Aliases='5730494N06Rik, AA407821', description='transmembrane protein 230', other_designations='UPF0414 transmembrane protein C20orf30 homolog', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132239492, end_position_on_the_genomic_accession=132248415, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70615, CurrentID=0, Status='live', Symbol='Ankrd24', Aliases='4631433D01Rik, 5730519E19Rik, D10Bur2e', description='ankyrin repeat domain 24', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81627795, end_position_on_the_genomic_accession=81647612, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  70616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70616, CurrentID=0, Status='live', Symbol='Sugp1', Aliases='5730496N02Rik, Sf4', description='SURP and G patch domain containing 1', other_designations='splicing factor 4', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70042785, end_position_on_the_genomic_accession=70071964, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70617, CurrentID=0, Status='live', Symbol='5730508B09Rik', Aliases='2700063A17Rik, C80506', description='RIKEN cDNA 5730508B09 gene', other_designations='', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127869688, end_position_on_the_genomic_accession=127896323, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  70620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70620, CurrentID=0, Status='live', Symbol='Ube2v2', Aliases='1110021H13Rik, 4632410D19Rik, 5730524P06Rik, 6820402M05, AI848315, C81524, MMS2', description='ubiquitin-conjugating enzyme E2 variant 2', other_designations='methyl methanesulfonate sensitive 2|ubc-like protein MMS2', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=15550986, end_position_on_the_genomic_accession=15594518, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70625, CurrentID=0, Status='live', Symbol='Med26', Aliases='5730493L18Rik, AI414941, AW495270, Crsp7', description='mediator complex subunit 26', other_designations='CRSP complex subunit 7|cofactor required for Sp1 transcriptional activation subunit 7|cofactor required for Sp1 transcriptional activation, subunit 7, 70kDa', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72493453, end_position_on_the_genomic_accession=72548310, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70638, CurrentID=0, Status='live', Symbol='Fam189a1', Aliases='5730507A09Rik, AW047451', description='family with sequence similarity 189, member A1', other_designations='', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64756091, end_position_on_the_genomic_accession=65156564, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  70640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70640, CurrentID=0, Status='live', Symbol='Dcp2', Aliases='2410015D23Rik, 5730537H01Rik, AL118268', description='decapping mRNA 2', other_designations='DCP2 decapping enzyme homolog|mRNA-decapping enzyme 2', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44380500, end_position_on_the_genomic_accession=44424969, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70645, CurrentID=0, Status='live', Symbol='Oip5', Aliases='5730547N13Rik, Lint-25, Ms18b', description='Opa interacting protein 5', other_designations='opa-interacting protein 5 homolog', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119609532, end_position_on_the_genomic_accession=119618505, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  70646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70646, CurrentID=0, Status='live', Symbol='Naa30', Aliases='4930487N19Rik, 5730533P17Rik, AI447922, AW322455, Nat12', description='N(alpha)-acetyltransferase 30, NatC catalytic subunit', other_designations='N-acetyltransferase 12|N-acetyltransferase MAK3 homolog|N-alpha-acetyltransferase 30, NatC catalytic subunit|Nalpha acetyltransferase 30|natC catalytic subunit', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49172133, end_position_on_the_genomic_accession=49191031, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  70650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70650, CurrentID=0, Status='live', Symbol='Zcchc8', Aliases='5730565F05Rik', description='zinc finger, CCHC domain containing 8', other_designations='TRAMP-like complex RNA-binding factor ZCCHC8', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123698302, end_position_on_the_genomic_accession=123721044, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  70652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70652, CurrentID=0, Status='live', Symbol='Tmem144', Aliases='1110057I03Rik, 5730537D05Rik', description='transmembrane protein 144', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79812564, end_position_on_the_genomic_accession=79852787, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  70661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70661, CurrentID=0, Status='live', Symbol='Sik3', Aliases='5730525O22Rik, 9030204A07Rik, AI447252, AI846254, AI849445, Qsk, SIK-3, mKIAA0999', description='SIK family kinase 3', other_designations='salt-inducible kinase 3|serine/threonine-protein kinase QSK', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46012820, end_position_on_the_genomic_accession=46224194, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  70673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70673, CurrentID=0, Status='live', Symbol='Prdm16', Aliases='5730557K01Rik, csp1, mel1', description='PR domain containing 16', other_designations='MDS1/EVI1-like gene 1|PR domain-containing protein 16|line 27|transcription factor MEL1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154316122, end_position_on_the_genomic_accession=154637457, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  70675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70675, CurrentID=0, Status='live', Symbol='Vcpip1', Aliases='4932442A08, 5730421J18Rik, 5730538E15Rik, VCIP135, mKIAA1850', description='valosin containing protein (p97)/p47 complex interacting protein 1', other_designations='valosin-containing protein p97/p47 complex-interacting protein 1|valosin-containing protein p97/p47 complex-interacting protein p135', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9718622, end_position_on_the_genomic_accession=9748382, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70676, CurrentID=0, Status='live', Symbol='Gulp1', Aliases='3110030A04Rik, 5730529O06Rik, CED-6, Ced6, GULP, Gulp-2', description='GULP, engulfment adaptor PTB domain containing 1', other_designations='CED 6|PTB domain adapter protein CED-6|cell death protein 6 homolog', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44551496, end_position_on_the_genomic_accession=44796838, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70681, CurrentID=0, Status='live', Symbol='Fam175a', Aliases='3830405G04Rik, 5630400M01Rik, AI506069, AL024423, AV118690, Ccdc98', description='family with sequence similarity 175, member A', other_designations='coiled-coil domain containing 98|coiled-coil domain-containing protein 98', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100804802, end_position_on_the_genomic_accession=100845216, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70683, CurrentID=0, Status='live', Symbol='Utp20', Aliases='3830408P06Rik, AA617408, Drim, mDRIM, mDRIN', description='UTP20, small subunit (SSU) processome component, homolog (yeast)', other_designations='down-regulated in metastasis protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88746607, end_position_on_the_genomic_accession=88826814, orientation='minus', exon_count=62, OMIM='', no_hdr0=''),
  70686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70686, CurrentID=0, Status='live', Symbol='Dusp16', Aliases='3830417M17Rik, AW558566, D6Ertd213e, MKP-7, MKP7, Mkpm', description='dual specificity phosphatase 16', other_designations='MAP kinase phosphatase-7|map kinase phosphatase-M', map_location='6 65.77 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134715468, end_position_on_the_genomic_accession=134792628, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70691, CurrentID=0, Status='live', Symbol='3830403N18Rik', Aliases='Xlrl', description='RIKEN cDNA 3830403N18 gene', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=55839252, end_position_on_the_genomic_accession=56153496, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  70693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70693, CurrentID=0, Status='live', Symbol='Adgra3', Aliases='3830613O22Rik, AU044632, Gpr125, Tem5-like', description='adhesion G protein-coupled receptor A3', other_designations='G protein-coupled receptor 125|G-protein coupled receptor 125|probable G-protein coupled receptor 125', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=49959951, end_position_on_the_genomic_accession=50058996, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  70696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70696, CurrentID=0, Status='live', Symbol='3830417A13Rik', Aliases='', description='RIKEN cDNA 3830417A13 gene', other_designations='', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=64173548, end_position_on_the_genomic_accession=64178778, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  70699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70699, CurrentID=0, Status='live', Symbol='Nup205', Aliases='3830404O05Rik, AV248391, mKIAA0225', description='nucleoporin 205', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=35177408, end_position_on_the_genomic_accession=35247599, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  70701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70701, CurrentID=0, Status='live', Symbol='Nipal1', Aliases='3830408G10Rik, Npal1', description='NIPA-like domain containing 1', other_designations='NIPA-like protein 1|non-imprinted in Prader-Willi/Angelman syndrome region protein 3 homolog', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72647796, end_position_on_the_genomic_accession=72671078, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  70713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70713, CurrentID=0, Status='live', Symbol='Gpr137c', Aliases='6330416L11Rik, Gm908, TM7SF1L2', description='G protein-coupled receptor 137C', other_designations='G protein-coupled receptor TM7SF1L2', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45219686, end_position_on_the_genomic_accession=45282735, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70717, CurrentID=0, Status='live', Symbol='Medag', Aliases='6330406I15Rik, Awms3, MEDA-4, Meda4, mAWMS3', description='mesenteric estrogen dependent adipogenesis', other_designations='activated in W/Wv mouse stomach 3|mesenteric estrogen-dependent adipose 4', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149405343, end_position_on_the_genomic_accession=149431703, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70719, CurrentID=0, Status='live', Symbol='Hmha1', Aliases='6330406L22Rik, AW539505, Ha-1', description='histocompatibility (minor) HA-1', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80016672, end_position_on_the_genomic_accession=80031471, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  70726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70726, CurrentID=0, Status='live', Symbol='Angptl6', Aliases='6330404E11Rik, AGF, ARP3, Arp5', description='angiopoietin-like 6', other_designations='Angiopoietin-related growth factor|angiopoietin-like protein 6|angiopoietin-related protein 5', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20873808, end_position_on_the_genomic_accession=20879816, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70727, CurrentID=0, Status='live', Symbol='Rasgef1a', Aliases='6330404M18Rik, AI835194', description='RasGEF domain family, member 1A', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118011466, end_position_on_the_genomic_accession=118091546, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  70729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70729, CurrentID=0, Status='live', Symbol='Nos1ap', Aliases='6330408P19Rik, Capon', description='nitric oxide synthase 1 (neuronal) adaptor protein', other_designations='C-terminal PDZ domain ligand of neuronal nitric oxide synthase|C-terminal PDZ ligand of neuronal nitric oxide synthase protein|nitric oxide synthase 1 adaptor protein', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170317414, end_position_on_the_genomic_accession=170589849, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70737, CurrentID=0, Status='live', Symbol='Cgn', Aliases='6330408J11Rik', description='cingulin', other_designations='', map_location='3 40.74 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94760069, end_position_on_the_genomic_accession=94787724, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  70747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70747, CurrentID=0, Status='live', Symbol='Tspan2', Aliases='6330415F13Rik, B230119D02Rik, tspan-2', description='tetraspanin 2', other_designations='tetraspan 2', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102734770, end_position_on_the_genomic_accession=102772310, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70750, CurrentID=0, Status='live', Symbol='Kdsr', Aliases='6330410P18Rik, 9430079B08Rik, Fvt1', description='3-ketodihydrosphingosine reductase', other_designations='3-dehydrosphinganine reductase|KDS reductase|follicular lymphoma variant translocation 1|follicular variant translocation protein 1 homolog', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=106720410, end_position_on_the_genomic_accession=106759830, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70757, CurrentID=0, Status='live', Symbol='Hacd2', Aliases='6330408J20Rik, AI255777, AI481689, Hcad2, Ptplb', description='3-hydroxyacyl-CoA dehydratase 2', other_designations='protein tyrosine phosphatase-like (proline instead of catalytic arginine), member b|protein-tyrosine phosphatase-like member B|very-long-chain (3R)-3-hydroxyacyl-[acyl-carrier protein] dehydratase 2', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35022421, end_position_on_the_genomic_accession=35109190, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  70762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70762, CurrentID=0, Status='live', Symbol='Dclk2', Aliases='6330415M09Rik, AU044875, CL2, CLICK2, Click-II, Dcamkl2', description='doublecortin-like kinase 2', other_designations='caMK-like CREB regulatory kinase 2|doublecortin and CaM kinase-like 2|doublecortin-like and CAM kinase-like 2', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=86786150, end_position_on_the_genomic_accession=86920884, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  70767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70767, CurrentID=0, Status='live', Symbol='Prpf3', Aliases='3632413F13Rik', description='pre-mRNA processing factor 3', other_designations='PRP3 pre-mRNA processing factor 3 homolog|pre-mRNA-splicing factor 3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95830615, end_position_on_the_genomic_accession=95855880, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  70769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70769, CurrentID=0, Status='live', Symbol='Nolc1', Aliases='NOPP130, NOPP140, P130', description='nucleolar and coiled-body phosphoprotein 1', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46075847, end_position_on_the_genomic_accession=46085543, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  70771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70771, CurrentID=0, Status='live', Symbol='Gpr173', Aliases='3230401K02Rik, Sreb3', description='G-protein coupled receptor 173', other_designations='G protein-coupled receptor SREB3|super conserved receptor expressed in brain 3', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152343598, end_position_on_the_genomic_accession=152368968, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70772, CurrentID=0, Status='live', Symbol='Ggnbp1', Aliases='0610031G08Rik, mdf', description='gametogenetin binding protein 1', other_designations='deubiquitinating enzyme Usp15|gametogenin binding protein 1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27018035, end_position_on_the_genomic_accession=27036378, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70779, CurrentID=0, Status='live', Symbol='Prdm5', Aliases='4432417F03Rik, 6530401I24Rik, AI197291, E130112L17Rik, PFM2', description='PR domain containing 5', other_designations='PR domain-containing protein 5', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=65778754, end_position_on_the_genomic_accession=65937138, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  70784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70784, CurrentID=0, Status='live', Symbol='Rasl12', Aliases='4631404I11Rik', description='RAS-like, family 12', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65398485, end_position_on_the_genomic_accession=65415060, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  70785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70785, CurrentID=0, Status='live', Symbol='Dennd1c', Aliases='4432409M07Rik', description='DENN/MADD domain containing 1C', other_designations='connecdenn 3', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57066051, end_position_on_the_genomic_accession=57078510, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  70788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70788, CurrentID=0, Status='live', Symbol='Klhl30', Aliases='4631423F02Rik', description='kelch-like 30', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91351073, end_position_on_the_genomic_accession=91362404, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70789, CurrentID=0, Status='live', Symbol='Kynu', Aliases='4432411A05Rik', description='kynureninase (L-kynurenine hydrolase)', other_designations='L-kynurenine hydrolase', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=43555258, end_position_on_the_genomic_accession=43683048, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  70790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70790, CurrentID=0, Status='live', Symbol='Ubr5', Aliases='Edd, Edd1', description='ubiquitin protein ligase E3 component n-recognin 5', other_designations='E3 ubiquitin-protein ligase, HECT domain-containing 1|extraembryonic development|hyperplastic discs protein homolog', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=37967328, end_position_on_the_genomic_accession=38078854, orientation='minus', exon_count=60, OMIM='', no_hdr0=''),
  70791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70791, CurrentID=0, Status='live', Symbol='Hars2', Aliases='4631412B19Rik, HARSR, HO3, Harsl', description='histidyl-tRNA synthetase 2', other_designations='hisRS|histidine--tRNA ligase|histidyl-tRNA synthetase 2, mitochondrial (putative)|probable histidyl-tRNA synthetase, mitochondrial', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36783232, end_position_on_the_genomic_accession=36792562, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  332937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332937, CurrentID=0, Status='live', Symbol='Tfap2e', Aliases='AI505200, AP-2e, Tcfap2e', description='transcription factor AP-2, epsilon', other_designations='AP2-epsilon|activating enhancer-binding protein 2 epsilon|activating enhancer-binding protein 2-epsilon', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126716003, end_position_on_the_genomic_accession=126736269, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70796, CurrentID=0, Status='live', Symbol='Zdhhc1', Aliases='4432412D04Rik', description='zinc finger, DHHC domain containing 1', other_designations='DHHC-1|zinc finger DHHC domain-containing protein 1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105472419, end_position_on_the_genomic_accession=105497105, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  70797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70797, CurrentID=0, Status='live', Symbol='Ankib1', Aliases='2310061P20Rik, 4631416I11Rik, AW494740, C80642, mKIAA1386', description='ankyrin repeat and IBR domain containing 1', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3689999, end_position_on_the_genomic_accession=3803124, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  332942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332942, CurrentID=0, Status='live', Symbol='Gm853', Aliases='OTTMUSG00000009560', description='predicted gene 853', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130209023, end_position_on_the_genomic_accession=130222427, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70799, CurrentID=0, Status='live', Symbol='Cep192', Aliases='4631422C13Rik, D430014P18Rik', description='centrosomal protein 192', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67800101, end_position_on_the_genomic_accession=67885170, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  70802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70802, CurrentID=0, Status='live', Symbol='Pwwp2a', Aliases='4631424J17Rik, AI891583, AU024105, D930040F23Rik', description='PWWP domain containing 2A', other_designations='', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43681082, end_position_on_the_genomic_accession=43722633, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  70804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70804, CurrentID=0, Status='live', Symbol='Pgrmc2', Aliases='4631434O19Rik, 5730409G06Rik, DG6, PMBP', description='progesterone receptor membrane component 2', other_designations='progesterone membrane binding protein', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=41066326, end_position_on_the_genomic_accession=41083046, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70807, CurrentID=0, Status='live', Symbol='Arrdc2', Aliases='4632416I05Rik, Ilad1', description='arrestin domain containing 2', other_designations='induced by lysergic acid diethylamide-1', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70835129, end_position_on_the_genomic_accession=70839720, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70809, CurrentID=0, Status='live', Symbol='Clec2g', Aliases='4632413B12Rik, Clr-c, Clr-d, Clrc, Clrd, Clrx, Ddv10, Ocilrp1', description='C-type lectin domain family 2, member g', other_designations='C-type lectin-related protein C|C-type lectin-related protein x variant 1|ocil-related protein 1|osteoclast inhibitory lectin related protein|osteoclast inhibitory lectin-related protein 1', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128934381, end_position_on_the_genomic_accession=128984707, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70810, CurrentID=0, Status='live', Symbol='Krt25', Aliases='4631426H08Rik, Ka38a, mIRSa1, Krt25', description='keratin 25', other_designations='CK-25|K25|cytokeratin-25|keratin 25A|keratin-25|krt25|type I inner root sheath-specific keratin-K25irs1|type I keratin KA38', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99315516, end_position_on_the_genomic_accession=99322941, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70821, CurrentID=0, Status='live', Symbol='4921507P07Rik', Aliases='TISP74', description='RIKEN cDNA 4921507P07 gene', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=50573302, end_position_on_the_genomic_accession=50596912, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70823, CurrentID=0, Status='live', Symbol='Hmgxb4', Aliases='4733401K04Rik, AI316803, Hmgb2l1', description='HMG box domain containing 4', other_designations='high mobility group box 2-like 1', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=74993575, end_position_on_the_genomic_accession=75031978, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  70827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70827, CurrentID=0, Status='live', Symbol='Trak2', Aliases='2900022D04Rik, 4733401O11Rik, AI480836, AW046271, Als2cr3, CALS-C, GRIF-1, GRIF1, OIP98, mKIAA0549', description='trafficking protein, kinesin binding 2', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 3', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58900449, end_position_on_the_genomic_accession=58973482, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  70829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70829, CurrentID=0, Status='live', Symbol='Ccdc93', Aliases='4633402D15Rik, 9230102M16Rik', description='coiled-coil domain containing 93', other_designations='', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=121431067, end_position_on_the_genomic_accession=121506460, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  70831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70831, CurrentID=0, Status='live', Symbol='Krtap31-1', Aliases='4733401H21Rik, AI506529', description='keratin associated protein 31-1', other_designations='keratin-associated protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99907920, end_position_on_the_genomic_accession=99908890, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  70834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70834, CurrentID=0, Status='live', Symbol='Spag9', Aliases='3110018C07Rik, 4733401I23Rik, 4831406C20Rik, AW552012, JLP, JSAP2, JSAP2a, Jip4, Mapk8ip4, syd1', description='sperm associated antigen 9', other_designations='JIP-4|JNK-associated leucine-zipper protein|JNK-interacting leucine zipper|JNK-interacting protein 4|JNK/SAPK-associated protein 2|mitogen-activated protein kinase 8-interacting protein 4|sperm-associated antigen 9', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=93996058, end_position_on_the_genomic_accession=94126087, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  70835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70835, CurrentID=0, Status='live', Symbol='Prss22', Aliases='4733401N09Rik, BSSP-4, SP001LA', description='protease, serine 22', other_designations='protease, serine, 22|serine protease 22|serine protease 26|tryptase epsilon', map_location='17 12.18 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23993534, end_position_on_the_genomic_accession=24000735, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70839, CurrentID=0, Status='live', Symbol='P2ry12', Aliases='2900079B22Rik, 4921504D23Rik, P2Y12', description='purinergic receptor P2Y, G-protein coupled 12', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=59216271, end_position_on_the_genomic_accession=59263005, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70840, CurrentID=0, Status='live', Symbol='Slc22a16', Aliases='4921504E14Rik, CT2, FLIPT2, OCT6, OKB1', description='solute carrier family 22 (organic cation transporter), member 16', other_designations='carnitine transporter 2', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40570321, end_position_on_the_genomic_accession=40604135, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70843, CurrentID=0, Status='live', Symbol='Krt28', Aliases='4733401L19Rik, Krt25d', description='keratin 28', other_designations='CK-28|K25D|K28|cytokeratin-28|keratin 25D|keratin-25D|keratin-28|krt28|type I inner root sheath-specific keratin-K25irs4', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99364861, end_position_on_the_genomic_accession=99379930, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  70846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70846, CurrentID=0, Status='live', Symbol='Ttc6', Aliases='4921506M07Rik, EG639426, ENSMUSG00000046782, Gm69, Gm9813', description='tetratricopeptide repeat domain 6', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=57564115, end_position_on_the_genomic_accession=57622105, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  595136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=595136, CurrentID=0, Status='live', Symbol='Ndufs5', Aliases='AI256693', description='NADH dehydrogenase (ubiquinone) Fe-S protein 5', other_designations='CI-15 kDa|NADH-ubiquinone oxidoreductase 15 kDa subunit|complex I-15 kDa', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123712710, end_position_on_the_genomic_accession=123718186, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70853, CurrentID=0, Status='live', Symbol='Vwa3b', Aliases='4921511C04Rik, A230074B11Rik, Gm555', description='von Willebrand factor A domain containing 3B', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37026591, end_position_on_the_genomic_accession=37187608, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  70859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70859, CurrentID=0, Status='live', Symbol='Lrrc63', Aliases='4921509B22Rik', description='leucine rich repeat containing 63', other_designations='', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75084303, end_position_on_the_genomic_accession=75130962, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70861, CurrentID=0, Status='live', Symbol='Akr1cl', Aliases='4921521F21Rik, AI451536, AI503553', description='aldo-keto reductase family 1, member C-like', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65013078, end_position_on_the_genomic_accession=65032770, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  70862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70862, CurrentID=0, Status='live', Symbol='Spata16', Aliases='4921511F01Rik, 4930503K02Rik, Nyd-sp12', description='spermatogenesis associated 16', other_designations='spermatogenesis-related protein', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=26637631, end_position_on_the_genomic_accession=26983212, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70864, CurrentID=0, Status='live', Symbol='Fam47c', Aliases='4921509A18Rik, Fam47a, Fam47a-ps', description='family with sequence similarity 47, member C', other_designations='family with sequence similarity 47, member A, pseudogene', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=78737763, end_position_on_the_genomic_accession=78739410, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  70866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70866, CurrentID=0, Status='live', Symbol='Slco6d1', Aliases='4921511I05Rik', description='solute carrier organic anion transporter family, member 6d1', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=98421085, end_position_on_the_genomic_accession=98516985, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  70873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70873, CurrentID=0, Status='live', Symbol='Cnbd2', Aliases='4921517L17Rik, 5430421B09Rik, Cris', description='cyclic nucleotide binding domain containing 2', other_designations='cyclic nucleotide-receptor involved in sperm function|novel transcript', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156312400, end_position_on_the_genomic_accession=156375638, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  70881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70881, CurrentID=0, Status='live', Symbol='Nt5c1b', Aliases='4921514H13Rik, AIRP, CN-IB, cN1B', description="5'-nucleotidase, cytosolic IB", other_designations="5'nucleotidase|autoimmune infertility-related protein|cytosolic 5'-nucleotidase IB", map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=10369971, end_position_on_the_genomic_accession=10390175, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70882, CurrentID=0, Status='live', Symbol='Armc3', Aliases='4921513G22Rik', description='armadillo repeat containing 3', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=19198865, end_position_on_the_genomic_accession=19310243, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  70884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70884, CurrentID=0, Status='live', Symbol='Ccdc81', Aliases='4921513D09Rik', description='coiled-coil domain containing 81', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=89866148, end_position_on_the_genomic_accession=89903629, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  70885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70885, CurrentID=0, Status='live', Symbol='Ints10', Aliases='4921521J11Rik, AI462004', description='integrator complex subunit 10', other_designations='int10', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=68793928, end_position_on_the_genomic_accession=68829410, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  70887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70887, CurrentID=0, Status='live', Symbol='Dmrtc1a', Aliases='1700030H14Rik, 4921520P21Rik, DMRTC1, Dmrt8.1', description='DMRT-like family C1a', other_designations='doublesex and mab-3 related transcription factor 8.1|doublesex- and mab-3-related transcription factor 8.1|likely ortologue to human doublesex-and mab-3-related transcription factor C1 (Dmrtc1)', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102903216, end_position_on_the_genomic_accession=102934003, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  242409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242409, CurrentID=0, Status='live', Symbol='Tmem8b', Aliases='4930500O05Rik', description='transmembrane protein 8B', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43668971, end_position_on_the_genomic_accession=43692668, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  70892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70892, CurrentID=0, Status='live', Symbol='Ttll7', Aliases='1110049N09Rik, 4921517B04Rik, C630030B20, mTTLL7', description='tubulin tyrosine ligase-like family, member 7', other_designations='polyglutamylase|tubulin--tyrosine ligase-like protein 7', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146852136, end_position_on_the_genomic_accession=146984009, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  70893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70893, CurrentID=0, Status='live', Symbol='Glb1l3', Aliases='4921509F24Rik', description='galactosidase, beta 1 like 3', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=26817953, end_position_on_the_genomic_accession=26862100, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  70894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70894, CurrentID=0, Status='live', Symbol='Efcab3', Aliases='4921510J17Rik', description='EF-hand calcium binding domain 3', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104934702, end_position_on_the_genomic_accession=105117537, orientation='plus', exon_count=85, OMIM='', no_hdr0=''),
  70897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70897, CurrentID=0, Status='live', Symbol='Fam71d', Aliases='4921509E07Rik, 4930516C23Rik, GARI-L2', description='family with sequence similarity 71, member D', other_designations='Golgi-associated Rab2B interactor-like 2', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=78691488, end_position_on_the_genomic_accession=78734519, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  70900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70900, CurrentID=0, Status='live', Symbol='4921517D22Rik', Aliases='', description='RIKEN cDNA 4921517D22 gene', other_designations='', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59687402, end_position_on_the_genomic_accession=59694103, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  70901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70901, CurrentID=0, Status='live', Symbol='4921524L21Rik', Aliases='', description='RIKEN cDNA 4921524L21 gene', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=6603633, end_position_on_the_genomic_accession=6640435, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  70902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70902, CurrentID=0, Status='live', Symbol='Lpcat2b', Aliases='4921521K07Rik, Aytl1b', description='lysophosphatidylcholine acyltransferase 2B', other_designations='acyltransferase like 1B|acyltransferase-like 1-B|lysophosphatidylcholine acyltransferase 2-B', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107431549, end_position_on_the_genomic_accession=107435039, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  333048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333048, CurrentID=0, Status='live', Symbol='Tmem211', Aliases='Gm854', description='transmembrane protein 211', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113226909, end_position_on_the_genomic_accession=113240223, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  333050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333050, CurrentID=0, Status='live', Symbol='Ksr2', Aliases='', description='kinase suppressor of ras 2', other_designations='testes-kinase suppressor of ras 2 isoform', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117413770, end_position_on_the_genomic_accession=117775004, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  70909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70909, CurrentID=0, Status='live', Symbol='4921504E06Rik', Aliases='4921522E24Rik', description='RIKEN cDNA 4921504E06 gene', other_designations='', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=19462837, end_position_on_the_genomic_accession=19553910, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  70911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70911, CurrentID=0, Status='live', Symbol='Phyhipl', Aliases='4921522K17Rik, AI267048', description='phytanoyl-CoA hydroxylase interacting protein-like', other_designations='', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=70557686, end_position_on_the_genomic_accession=70599291, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  70918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70918, CurrentID=0, Status='live', Symbol='Nsun7', Aliases='4921525L17Rik', description='NOL1/NOP2/Sun domain family, member 7', other_designations='', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=66259840, end_position_on_the_genomic_accession=66299945, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70920, CurrentID=0, Status='live', Symbol='4921511H03Rik', Aliases='', description='RIKEN cDNA 4921511H03 gene', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=7302505, end_position_on_the_genomic_accession=7311491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  70925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70925, CurrentID=0, Status='live', Symbol='Cdkn2aip', Aliases='4921511I16Rik, AW208986, CARF', description='CDKN2A interacting protein', other_designations='collaborates/cooperates with ARF (alternate reading frame) protein|collaborator of ARF', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47709344, end_position_on_the_genomic_accession=47714735, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70928, CurrentID=0, Status='live', Symbol='Trim69', Aliases='4921519C19Rik, Rnf36, Trif, Trimless', description='tripartite motif-containing 69', other_designations='RFP-like (B30.2) domain protein|RING finger B-box coiled-coil transcription factor|ring finger protein 36|testis-specific RING finger protein|tripartite motif-containing protein 69', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122160700, end_position_on_the_genomic_accession=122179027, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70930, CurrentID=0, Status='live', Symbol='Nol8', Aliases='4921532D18Rik, 5730412B09Rik', description='nucleolar protein 8', other_designations='nol8', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49653078, end_position_on_the_genomic_accession=49679016, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  70935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70935, CurrentID=0, Status='live', Symbol='Speer4f1', Aliases='4922502J04Rik, SPEER-4F, Speer4f', description='spermatogenesis associated glutamate (E)-rich protein 4F1', other_designations='sperm-associated glutamate (E)-rich protein 4f|spermatogenesis associated glutamate (E)-rich protein 4f|spermatogenic-associated glutamate (E)-rich protein 4f', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=17476098, end_position_on_the_genomic_accession=17480936, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  70941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70941, CurrentID=0, Status='live', Symbol='4921539E11Rik', Aliases='', description='RIKEN cDNA 4921539E11 gene', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=103230445, end_position_on_the_genomic_accession=103290953, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  333088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333088, CurrentID=0, Status='live', Symbol='Kcp', Aliases='AW060220, CRIM-2, Crim2, Gm793, KCP-1', description='kielin/chordin-like protein', other_designations='cysteine rich BMP regulator 2 (chordin like)|cysteine-rich BMP regulator 2|cysteine-rich motor neuron 2 protein|kielin/chordin-like protein 1', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29473162, end_position_on_the_genomic_accession=29508011, orientation='minus', exon_count=49, OMIM='', no_hdr0=''),
  70945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70945, CurrentID=0, Status='live', Symbol='Mmrn1', Aliases='4921530G03Rik, Emilin4, Mmrn', description='multimerin 1', other_designations='mmrn1', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=60944317, end_position_on_the_genomic_accession=60989378, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  70950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70950, CurrentID=0, Status='live', Symbol='Ccdc178', Aliases='4921528I01Rik', description='coiled coil domain containing 178', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=21810891, end_position_on_the_genomic_accession=22171442, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  70951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70951, CurrentID=0, Status='live', Symbol='Spata1', Aliases='4921536I21Rik, AV266943, SP-2', description='spermatogenesis associated 1', other_designations='', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146457200, end_position_on_the_genomic_accession=146499753, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  70952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70952, CurrentID=0, Status='live', Symbol='Poteg', Aliases='4921537P18Rik, 4930456F22Rik, Gsarp1, Gsarp2', description='POTE ankyrin domain family, member G', other_designations='gonad specific ankyrin repeat protein', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27447529, end_position_on_the_genomic_accession=27582715, orientation='plus', exon_count=56, OMIM='', no_hdr0=''),
  70956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70956, CurrentID=0, Status='live', Symbol='Tex19.2', Aliases='4921530G04Rik, Tex19b', description='testis expressed gene 19.2', other_designations='testis-expressed protein 19B|testis-expressed sequence 19B', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121116215, end_position_on_the_genomic_accession=121118677, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  70967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70967, CurrentID=0, Status='live', Symbol='Eva1c', Aliases='1700092M14Rik, 4931408A02Rik, Fam176c', description='eva-1 homolog C (C. elegans)', other_designations='protein FAM176C', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90826719, end_position_on_the_genomic_accession=90904885, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70974, CurrentID=0, Status='live', Symbol='Pgm2l1', Aliases='4931406N15Rik, AI553438, BM32A', description='phosphoglucomutase 2-like 1', other_designations='phosphoglucomutase-2-like 1', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100227283, end_position_on_the_genomic_accession=100278872, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  70976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70976, CurrentID=0, Status='live', Symbol='Ccdc105', Aliases='4931413A09Rik', description='coiled-coil domain containing 105', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78746924, end_position_on_the_genomic_accession=78753067, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70977, CurrentID=0, Status='live', Symbol='Cabs1', Aliases='4931407G18Rik, AI428159, Clph', description='calcium binding protein, spermatid specific 1', other_designations='casein-like phosphoprotein', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87979451, end_position_on_the_genomic_accession=87981541, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  70979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70979, CurrentID=0, Status='live', Symbol='Fancd2os', Aliases='4931417G12Rik', description='Fancd2 opposite strand', other_designations='Fanconi anemia group D2 protein opposite strand transcript protein', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113596761, end_position_on_the_genomic_accession=113600715, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  70980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70980, CurrentID=0, Status='live', Symbol='4931431F19Rik', Aliases='', description='RIKEN cDNA 4931431F19 gene', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104127913, end_position_on_the_genomic_accession=104129823, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  70981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70981, CurrentID=0, Status='live', Symbol='4931423N10Rik', Aliases='', description='RIKEN cDNA 4931423N10 gene', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=23203438, end_position_on_the_genomic_accession=23267129, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  70984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70984, CurrentID=0, Status='live', Symbol='4931406C07Rik', Aliases='', description='RIKEN cDNA 4931406C07 gene', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15283337, end_position_on_the_genomic_accession=15306448, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  70989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70989, CurrentID=0, Status='live', Symbol='4931429I11Rik', Aliases='BB239335, jhy', description='RIKEN cDNA 4931429I11 gene', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40894849, end_position_on_the_genomic_accession=40966864, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  70993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70993, CurrentID=0, Status='live', Symbol='Prss54', Aliases='4931432M23Rik, Klkbl4', description='protease, serine 54', other_designations='plasma kallikrein-like protein 4|protease, serine, 54', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95559067, end_position_on_the_genomic_accession=95576330, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  70997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70997, CurrentID=0, Status='live', Symbol='Spef1', Aliases='4931426K16Rik, Clamp', description='sperm flagellar 1', other_designations='calponin-homology and microtubule-associated protein', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131170261, end_position_on_the_genomic_accession=131174810, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  70998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70998, CurrentID=0, Status='live', Symbol='Phf6', Aliases='2700007B13Rik, 4931428F02Rik, mKIAA1823', description='PHD finger protein 6', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52912214, end_position_on_the_genomic_accession=52956961, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  70999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=70999, CurrentID=0, Status='live', Symbol='Naa40', Aliases='4931433E08Rik, AU023197, Nat11', description='N(alpha)-acetyltransferase 40, NatD catalytic subunit, homolog (S. cerevisiae)', other_designations='N-acetyltransferase 11|N-alpha-acetyltransferase 40, NatD catalytic subunit|Nalpha acetyltransferase 40|natD catalytic subunit', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7225668, end_position_on_the_genomic_accession=7241222, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71001, CurrentID=0, Status='live', Symbol='Mgat4e', Aliases='4931440L10Rik', description='MGAT4 family, member E', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134540941, end_position_on_the_genomic_accession=134549682, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71003, CurrentID=0, Status='live', Symbol='Prss41', Aliases='4931440B09Rik, Tessp-1, Tessp1', description='protease, serine 41', other_designations='protease, serine, 41|testis serine protease 1|testis-specific serine protease 1', map_location='17 12.09 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23836768, end_position_on_the_genomic_accession=23844374, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71007, CurrentID=0, Status='live', Symbol='D7Ertd443e', Aliases='4933400E14Rik, FATS, Gm497', description='DNA segment, Chr 7, ERATO Doi 443, expressed', other_designations='fragile-site associated tumor suppressor', map_location='7 78.4 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=134266262, end_position_on_the_genomic_accession=134520141, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  71020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71020, CurrentID=0, Status='live', Symbol='Spats1', Aliases='1700011H05Rik, 4933400B06Rik, Daip2, Ddip, Srsp1', description='spermatogenesis associated, serine-rich 1', other_designations='Dvl2-DEP domain interacting protein|dvl2-DEP domain-interacting protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45445877, end_position_on_the_genomic_accession=45475194, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71026, CurrentID=0, Status='live', Symbol='Speer3', Aliases='4933405P08Rik, SPEER-3', description='spermatogenesis associated glutamate (E)-rich protein 3', other_designations='sperm-associated glutamate (E)-rich protein 3|spermatogenic-associated glutamate (E)-rich protein 3', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=13791619, end_position_on_the_genomic_accession=13796819, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71027, CurrentID=0, Status='live', Symbol='Tmem30c', Aliases='4933401B01Rik, 4933409A18Rik, Cdc50c', description='transmembrane protein 30C', other_designations='', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57266138, end_position_on_the_genomic_accession=57301893, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71030, CurrentID=0, Status='live', Symbol='4933403O08Rik', Aliases='', description='RIKEN cDNA 4933403O08 gene', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=112171438, end_position_on_the_genomic_accession=112249350, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71037, CurrentID=0, Status='live', Symbol='Prss55', Aliases='4933401F05Rik', description='protease, serine 55', other_designations='probable serine protease UNQ9391/PRO34284 homolog|protease, serine, 55', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64075435, end_position_on_the_genomic_accession=64085433, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  333182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333182, CurrentID=0, Status='live', Symbol='Cox6b2', Aliases='1700067P11Rik, BC048670, COXVIB2', description='cytochrome c oxidase subunit VIb polypeptide 2', other_designations='cytochrome c oxidase subunit VIb, testis-specific isoform', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4751792, end_position_on_the_genomic_accession=4753094, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71041, CurrentID=0, Status='live', Symbol='Pcgf6', Aliases='4933407A11Rik, AI604840, MBLR, Rnf134', description='polycomb group ring finger 6', other_designations='Mel18 and Bmi1-like RING finger protein|Mel18- and Bmi1-like RING finger protein|ring finger protein 134', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47033617, end_position_on_the_genomic_accession=47051098, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71046, CurrentID=0, Status='live', Symbol='4933405L10Rik', Aliases='', description='RIKEN cDNA 4933405L10 gene', other_designations='', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105708303, end_position_on_the_genomic_accession=105710245, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  333193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333193, CurrentID=0, Status='live', Symbol='Proser3', Aliases='', description='proline and serine rich 3', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30539129, end_position_on_the_genomic_accession=30552297, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71059, CurrentID=0, Status='live', Symbol='Hexim2', Aliases='4933402L21Rik', description='hexamethylene bis-acetamide inducible 2', other_designations='hexamthylene bis-acetamide inducible 2', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103127255, end_position_on_the_genomic_accession=103139908, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71062, CurrentID=0, Status='live', Symbol='Tekt3', Aliases='4933407G07Rik', description='tektin 3', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=63061659, end_position_on_the_genomic_accession=63094960, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71063, CurrentID=0, Status='live', Symbol='Zfp597', Aliases='4933407K12Rik, AI462383, AI666357', description='zinc finger protein 597', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3861544, end_position_on_the_genomic_accession=3884563, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71066, CurrentID=0, Status='live', Symbol='Hsfy2', Aliases='4933413G11Rik, Hspy2l', description='heat shock transcription factor, Y-linked 2', other_designations='heat shock transcription factor, Y linked 2 like', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=56636044, end_position_on_the_genomic_accession=56637451, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71069, CurrentID=0, Status='live', Symbol='Stox2', Aliases='4933409N07Rik, AI449080, AI835114, mKIAA1392', description='storkhead box 2', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47180048, end_position_on_the_genomic_accession=47413081, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71078, CurrentID=0, Status='live', Symbol='Adam30', Aliases='4933424D07Rik, svph4', description='a disintegrin and metallopeptidase domain 30', other_designations='a disintegrin and metalloproteinase domain 30', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98160812, end_position_on_the_genomic_accession=98163173, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71083, CurrentID=0, Status='live', Symbol='Dmrtc1c1', Aliases='4933424F23Rik, Dmrt8.3, Dmrtc1c', description='DMRT-like family C1c1', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102797880, end_position_on_the_genomic_accession=102809945, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71085, CurrentID=0, Status='live', Symbol='Arhgap19', Aliases='4933411B03Rik, AI120128, AW546831, mFLJ00368', description='Rho GTPase activating protein 19', other_designations='rho-type GTPase-activating protein 19', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41766588, end_position_on_the_genomic_accession=41802084, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  71088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71088, CurrentID=0, Status='live', Symbol='4933412E24Rik', Aliases='', description='RIKEN cDNA 4933412E24 gene', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=60014866, end_position_on_the_genomic_accession=60016618, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71089, CurrentID=0, Status='live', Symbol='Sept12', Aliases='1700028G04Rik, 4933413B09Rik, Septin12', description='septin 12', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4986858, end_position_on_the_genomic_accession=4997852, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71091, CurrentID=0, Status='live', Symbol='Cdkl1', Aliases='4933411O17Rik', description='cyclin-dependent kinase-like 1 (CDC2-related kinase)', other_designations='', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69746846, end_position_on_the_genomic_accession=69791219, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71093, CurrentID=0, Status='live', Symbol='Atoh8', Aliases='4933425C05Rik, Hath6, Math6, bHLHa21, okadin', description='atonal bHLH transcription factor 8', other_designations='atonal homolog 8|basic helix-loop-helix transcription factor 6|helix-loop-helix protein mATH-6', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72206174, end_position_on_the_genomic_accession=72235780, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71096, CurrentID=0, Status='live', Symbol='Sntg1', Aliases='4933426D16Rik, G1SYN, SYN4', description='syntrophin, gamma 1', other_designations='syntrophin-4', map_location='1 A1-A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=8359739, end_position_on_the_genomic_accession=9301061, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  71099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71099, CurrentID=0, Status='live', Symbol='Tssk4', Aliases='1700020B19Rik, 4933424F08Rik, TSK-4, TSSK-4', description='testis-specific serine kinase 4', other_designations='testis-specific kinase 4|testis-specific serine/threonine protein kinase 5 variant alpha|testis-specific serine/threonine protein kinase 5 variant beta|testis-specific serine/threonine protein kinase 5 variant delta|testis-specific serine/threonine protein kinase 5 variant gamma', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55643873, end_position_on_the_genomic_accession=55652539, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71101, CurrentID=0, Status='live', Symbol='Uvssa', Aliases='4933407H18Rik, D330017J19Rik, Kiaa1530, mKIAA1530', description='UV stimulated scaffold protein A', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33378655, end_position_on_the_genomic_accession=33419754, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  71103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71103, CurrentID=0, Status='live', Symbol='Glt6d1', Aliases='4933411C14Rik, GT6m7', description='glycosyltransferase 6 domain containing 1', other_designations='novel glycosyltransferase family 6 protein', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25771820, end_position_on_the_genomic_accession=25815940, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71111, CurrentID=0, Status='live', Symbol='Gpr39', Aliases='4933415E13Rik', description='G protein-coupled receptor 39', other_designations='', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=125676996, end_position_on_the_genomic_accession=125873862, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  71116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71116, CurrentID=0, Status='live', Symbol='Stx18', Aliases='1810035L21Rik, 4933425D03Rik, AU041562', description='syntaxin 18', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38038622, end_position_on_the_genomic_accession=38137770, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  667214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667214, CurrentID=0, Status='live', Symbol='9930111J21Rik1', Aliases='9930111J21Rik, Iigp', description='RIKEN cDNA 9930111J21 gene 1', other_designations='', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48946150, end_position_on_the_genomic_accession=48979381, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71130, CurrentID=0, Status='live', Symbol='Sh2d6', Aliases='', description='SH2 domain containing 6', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72513652, end_position_on_the_genomic_accession=72534207, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  71131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71131, CurrentID=0, Status='live', Symbol='Zfp689', Aliases='4933416E05Rik, Znf689', description='zinc finger protein 689', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127442136, end_position_on_the_genomic_accession=127449158, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  71132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71132, CurrentID=0, Status='live', Symbol='Cabyr', Aliases='1700016C01Rik, 4933421A18Rik, CBP86, FSP-2', description='calcium-binding tyrosine-(Y)-phosphorylation regulated (fibrousheathin 2)', other_designations='calcium-binding protein 86|testis-specific calcium-binding protein CBP86', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12741307, end_position_on_the_genomic_accession=12755151, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71137, CurrentID=0, Status='live', Symbol='Rfx4', Aliases='4933412G19Rik, NYD-sp10', description='regulatory factor X, 4 (influences HLA class II expression)', other_designations='regulatory factor X 4 protein|winged-helix transcription factor RFX4', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=84756048, end_position_on_the_genomic_accession=84906538, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  71138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71138, CurrentID=0, Status='live', Symbol='Tmem217', Aliases='4933413N12Rik, EG622644', description='transmembrane protein 217', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29519143, end_position_on_the_genomic_accession=29549689, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71145, CurrentID=0, Status='live', Symbol='Scara5', Aliases='4932433F15Rik, 4933425F03Rik, AV278087, Tesr', description='scavenger receptor class A, member 5', other_designations='scavenger receptor class A, member 5 (putative)|testis expressed scavenger receptor', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65666403, end_position_on_the_genomic_accession=65764826, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  71146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71146, CurrentID=0, Status='live', Symbol='Golga7b', Aliases='4933417O08Rik, AI839934', description='golgi autoantigen, golgin subfamily a, 7B', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42247578, end_position_on_the_genomic_accession=42270348, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71147, CurrentID=0, Status='live', Symbol='Oxsm', Aliases='4933425A18Rik, C80494', description='3-oxoacyl-ACP synthase, mitochondrial', other_designations='beta-ketoacyl-ACP synthase', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=16238659, end_position_on_the_genomic_accession=16249808, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71148, CurrentID=0, Status='live', Symbol='Mier1', Aliases='4933425I22Rik, 5830411K19Rik, er1', description='MEIR1 treanscription regulator', other_designations='early response 1|mesoderm induction early response 1 homolog (Xenopus laevis', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=103114390, end_position_on_the_genomic_accession=103165754, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  71149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71149, CurrentID=0, Status='live', Symbol='4933413G19Rik', Aliases='', description='RIKEN cDNA 4933413G19 gene', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128375510, end_position_on_the_genomic_accession=128385144, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  71151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71151, CurrentID=0, Status='live', Symbol='Eri2', Aliases='4933424N09Rik, Exod1, mKIAA1504', description='exoribonuclease 2', other_designations='exonuclease domain containing 1|exonuclease domain-containing protein 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119783826, end_position_on_the_genomic_accession=119794058, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71156, CurrentID=0, Status='live', Symbol='Lrrc72', Aliases='1700108M19Rik, 4933421E18Rik', description='leucine rich repeat containing 72', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=36208345, end_position_on_the_genomic_accession=36256396, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71159, CurrentID=0, Status='live', Symbol='4933416I08Rik', Aliases='', description='RIKEN cDNA 4933416I08 gene', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53577073, end_position_on_the_genomic_accession=53580003, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71162, CurrentID=0, Status='live', Symbol='4933421I07Rik', Aliases='', description='RIKEN cDNA 4933421I07 gene', other_designations='', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=42445462, end_position_on_the_genomic_accession=42448128, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  333307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333307, CurrentID=0, Status='live', Symbol='Trim75', Aliases='Gm794', description='tripartite motif-containing 75', other_designations='', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=64980588, end_position_on_the_genomic_accession=64987669, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  71164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71164, CurrentID=0, Status='live', Symbol='Zdhhc11', Aliases='4933421L13Rik, DHHC-11, Zdhhc20', description='zinc finger, DHHC domain containing 11', other_designations='DHHC-containing protein 10|zinc finger DHHC domain-containing protein 11|zinc finger, DHHC domain containing 20', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73963847, end_position_on_the_genomic_accession=73992985, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71169, CurrentID=0, Status='live', Symbol='Nbas', Aliases='4933425L03Rik, Nag', description='neuroblastoma amplified sequence', other_designations='neuroblastoma-amplified protein', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=13269101, end_position_on_the_genomic_accession=13637380, orientation='plus', exon_count=53, OMIM='', no_hdr0=''),
  333315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333315, CurrentID=0, Status='live', Symbol='Frem3', Aliases='Gm795, NV2', description='Fras1 related extracellular matrix protein 3', other_designations='NV domain containing protein NV2|NV domain-containing protein 2|embryonic blastocoelar extracellular matrix protein', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=80611039, end_position_on_the_genomic_accession=80695557, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71175, CurrentID=0, Status='live', Symbol='Nipbl', Aliases='Idn3', description='Nipped-B homolog (Drosophila)', other_designations='SCC2 homolog|delangin homolog', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=8289824, end_position_on_the_genomic_accession=8465310, orientation='minus', exon_count=49, OMIM='', no_hdr0=''),
  71176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71176, CurrentID=0, Status='live', Symbol='Fbxo24', Aliases='4933422D21Rik, Fbx24', description='F-box protein 24', other_designations='', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137612505, end_position_on_the_genomic_accession=137625584, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71177, CurrentID=0, Status='live', Symbol='Asun', Aliases='4933424B01Rik, Spata30', description='asunder, spermatogenesis regulator', other_designations='cell cycle regulator Mat89Bb homolog|likely orthologue of H. sapiens chromosome 12 open reading frame 11 (C12orf11)|spermatogenesis associated 30|spermatogenesis-associated protein 30', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146549632, end_position_on_the_genomic_accession=146577835, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  71183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71183, CurrentID=0, Status='live', Symbol='Clec12b', Aliases='4933425B16Rik', description='C-type lectin domain family 12, member B', other_designations='macrophage antigen h', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129375513, end_position_on_the_genomic_accession=129427634, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  333329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333329, CurrentID=0, Status='live', Symbol='Cngb1', Aliases='Cngb1b, Gm1959', description='cyclic nucleotide gated channel beta 1', other_designations='cyclic nucleotide-gated channel beta subunit 1', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95239043, end_position_on_the_genomic_accession=95306585, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  71198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71198, CurrentID=0, Status='live', Symbol='Otud1', Aliases='4933428L19Rik', description='OTU domain containing 1', other_designations='likely ortholog of H. sapiens OTU domain containing 1 (OTUD1)', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=19658062, end_position_on_the_genomic_accession=19660590, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71200, CurrentID=0, Status='live', Symbol='Dydc2', Aliases='4933428D01Rik', description='DPY30 domain containing 2', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=41048957, end_position_on_the_genomic_accession=41074049, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71206, CurrentID=0, Status='live', Symbol='Katnal2', Aliases='3110023G01Rik, 4933439B08Rik', description='katanin p60 subunit A-like 2', other_designations='Katanin-like protein 2|p60 katanin-like 2', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=76977147, end_position_on_the_genomic_accession=77048225, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  71207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71207, CurrentID=0, Status='live', Symbol='Nudt4', Aliases='4933436C10Rik, DIPP2, DIPP2alpha, DIPP2beta, HDCMB47P', description='nudix (nucleoside diphosphate linked moiety X)-type motif 4', other_designations="DIPP-2|diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 2|nucleoside diphosphate-linked moiety X motif 4|nudix motif 4", map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=95547007, end_position_on_the_genomic_accession=95564167, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71213, CurrentID=0, Status='live', Symbol='Cage1', Aliases='4933427I01Rik, Ctag3', description='cancer antigen 1', other_designations='CAGE-1|CT3 homolog|cancer/testis antigen 3 homolog', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=37994314, end_position_on_the_genomic_accession=38036937, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71223, CurrentID=0, Status='live', Symbol='Gpr15', Aliases='4933439K08Rik', description='G protein-coupled receptor 15', other_designations='', map_location='16 C1.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58717435, end_position_on_the_genomic_accession=58718724, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71228, CurrentID=0, Status='live', Symbol='Dlg5', Aliases='4933429D20Rik, T25557, mKIAA0583', description='discs, large homolog 5 (Drosophila)', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=24133953, end_position_on_the_genomic_accession=24245920, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  71233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71233, CurrentID=0, Status='live', Symbol='Enkur', Aliases='4933434I06Rik, AU016229', description='enkurin, TRPC channel interacting protein', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=21180731, end_position_on_the_genomic_accession=21205365, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71236, CurrentID=0, Status='live', Symbol='Rsph14', Aliases='4933431K05Rik, Rtdr1', description='radial spoke head homolog 14 (Chlamydomonas)', other_designations='rhabdoid tumor deletion region gene 1|rhabdoid tumor deletion region protein 1', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=74957477, end_position_on_the_genomic_accession=75034314, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71238, CurrentID=0, Status='live', Symbol='Sdhaf3', Aliases='0610005M07Rik, 4933430A16Rik, Acn9', description='succinate dehydrogenase complex assembly factor 3', other_designations='ACN9 homolog|SDH assembly factor 3|protein ACN9 homolog, mitochondrial', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=6956018, end_position_on_the_genomic_accession=7039220, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  71240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71240, CurrentID=0, Status='live', Symbol='Osbpl7', Aliases='4933437E18Rik', description='oxysterol binding protein-like 7', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97050344, end_position_on_the_genomic_accession=97068904, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  71241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71241, CurrentID=0, Status='live', Symbol='Dmrtc2', Aliases='4933432E21Rik, Dmrt7', description='doublesex and mab-3 related transcription factor like family C2', other_designations='doublesex and mab-3 related transcription factor 7', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24869876, end_position_on_the_genomic_accession=24877655, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71242, CurrentID=0, Status='live', Symbol='Spata24', Aliases='2700012K08Rik, 4930583E11Rik, 5133400G04Rik, AU016220, TIPT, TIPT2', description='spermatogenesis associated 24', other_designations='TATA-binding protein-like factor-interacting protein|TLF-interacting protein|TRF2-interacting protein in testis|testis protein T6441 homolog', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35656689, end_position_on_the_genomic_accession=35662186, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71254, CurrentID=0, Status='live', Symbol='Naif1', Aliases='2310007O20Rik, 4933440H19Rik', description='nuclear apoptosis inducing factor 1', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32450457, end_position_on_the_genomic_accession=32455476, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  71263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71263, CurrentID=0, Status='live', Symbol='Mro', Aliases='4930507C04Rik, 4933435E20Rik', description='maestro', other_designations='male-specific transcription in the developing reproductive organs', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=73851827, end_position_on_the_genomic_accession=73881334, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71268, CurrentID=0, Status='live', Symbol='Lrrfip2', Aliases='5133400F20Rik, AI850587', description='leucine rich repeat (in FLII) interacting protein 2', other_designations='LRR FLII-interacting protein 2', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111118111, end_position_on_the_genomic_accession=111225668, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  71275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71275, CurrentID=0, Status='live', Symbol='Noxred1', Aliases='4933437F05Rik', description='NADP+ dependent oxidoreductase domain containing 1', other_designations='pyrroline-5-carboxylate reductase-like protein C14orf148 homolog', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87221123, end_position_on_the_genomic_accession=87238774, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71276, CurrentID=0, Status='live', Symbol='Ccdc57', Aliases='4933434G05Rik', description='coiled-coil domain containing 57', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120826529, end_position_on_the_genomic_accession=120932902, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  71279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71279, CurrentID=0, Status='live', Symbol='Slc29a3', Aliases='4933435C21Rik, AW987637, Ent3', description='solute carrier family 29 (nucleoside transporters), member 3', other_designations='mENT3|solute carrier family 29 member 3', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60712072, end_position_on_the_genomic_accession=60752850, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  333424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333424, CurrentID=0, Status='live', Symbol='A4gnt', Aliases='AV080780, Alpha4gnt, Gm798', description='alpha-1,4-N-acetylglucosaminyltransferase', other_designations='', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99591175, end_position_on_the_genomic_accession=99622367, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71281, CurrentID=0, Status='live', Symbol='Apobec4', Aliases='4933431M11Rik', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide-like 4 (putative)', other_designations='apolipoprotein B mRNA-editing enzyme catalytic polypeptide-like 4', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152750551, end_position_on_the_genomic_accession=152757544, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  71287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71287, CurrentID=0, Status='live', Symbol='Cpvl', Aliases='4933436L16Rik, HVLP', description='carboxypeptidase, vitellogenic-like', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=53873279, end_position_on_the_genomic_accession=53978673, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  333433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333433, CurrentID=0, Status='live', Symbol='Gpd1l', Aliases='2210409H23Rik, D9Ertd660e', description='glycerol-3-phosphate dehydrogenase 1-like', other_designations='', map_location='9 65.47 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114899339, end_position_on_the_genomic_accession=114933987, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71296, CurrentID=0, Status='live', Symbol='Crnde', Aliases='4933436C20Rik', description='colorectal neoplasia differentially expressed (non-protein coding)', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=92326031, end_position_on_the_genomic_accession=92356120, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71300, CurrentID=0, Status='live', Symbol='Rnf148', Aliases='4933432M07Rik, Greul3', description='ring finger protein 148', other_designations='goliath-related E3 ubiquitin-protein ligase 3', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=23653895, end_position_on_the_genomic_accession=23655136, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71302, CurrentID=0, Status='live', Symbol='Arhgap26', Aliases='1810044B20Rik, 2610010G17Rik, 4933432P15Rik, AI853435, mKIAA0621', description='Rho GTPase activating protein 26', other_designations='rho-type GTPase-activating protein 26', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38992968, end_position_on_the_genomic_accession=39376285, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  71306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71306, CurrentID=0, Status='live', Symbol='Mfap3l', Aliases='4933428A15Rik, 5430405D20Rik, AI461995, AW125052, NYD-sp9, mKIAA0626', description='microfibrillar-associated protein 3-like', other_designations='', map_location='8 B3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=60632825, end_position_on_the_genomic_accession=60676731, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71310, CurrentID=0, Status='live', Symbol='Tbc1d9', Aliases='4933431N12Rik, AI847101, AW490653, C76116', description='TBC1 domain family, member 9', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83165352, end_position_on_the_genomic_accession=83272940, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  71313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71313, CurrentID=0, Status='live', Symbol='Fsip1', Aliases='1700012M13Rik, 4933432K11Rik', description='fibrous sheath-interacting protein 1', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118130261, end_position_on_the_genomic_accession=118256987, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71323, CurrentID=0, Status='live', Symbol='Rassf8', Aliases='5133400D11Rik, AA815821, AW123240, Hoj1, mHoj-1', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 8', other_designations='Ras association (RalGDS/AF-6) domain family 8|carcinoma-associated protein HOJ-1 homolog', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145746562, end_position_on_the_genomic_accession=145821079, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71325, CurrentID=0, Status='live', Symbol='Tchhl1', Aliases='5430400H23Rik, AA589420, S100a17, Thhl1', description='trichohyalin-like 1', other_designations='S100 calcium binding protein A17', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93468734, end_position_on_the_genomic_accession=93471980, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  71326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71326, CurrentID=0, Status='live', Symbol='Treml1', Aliases='5430401J17Rik, TLT-1', description='triggering receptor expressed on myeloid cells-like 1', other_designations='triggering receptor expressed on myeloid cells-like 1c|triggering receptor expressed on myeloid cells-like 1d|triggering receptor expressed on myeloid cells-like 1e|triggering receptor expressed on myeloid cells-like protein 1', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48359916, end_position_on_the_genomic_accession=48367176, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  333473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333473, CurrentID=0, Status='live', Symbol='Zfp36l3', Aliases='AY661338', description='zinc finger protein 36, C3H type-like 3', other_designations='zinc finger protein 36-like 3', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53772686, end_position_on_the_genomic_accession=53776394, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71330, CurrentID=0, Status='live', Symbol='Rcbtb1', Aliases='5430409I18Rik, AW111883, CLLD7, CLLL7', description='regulator of chromosome condensation (RCC1) and BTB (POZ) domain containing protein 1', other_designations='regulator of chromosome condensation and BTB domain-containing protein 1', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59201027, end_position_on_the_genomic_accession=59237267, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  71336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71336, CurrentID=0, Status='live', Symbol='Rbks', Aliases='5230400M11Rik, AI326806, RBSK', description='ribokinase', other_designations='rbks', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31624439, end_position_on_the_genomic_accession=31697698, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71338, CurrentID=0, Status='live', Symbol='Tprg', Aliases='5430420C16Rik1, Tprg', description='transformation related protein 63 regulated', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=25286817, end_position_on_the_genomic_accession=25423588, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71340, CurrentID=0, Status='live', Symbol='Riok1', Aliases='3110046C13Rik, 5430416A05Rik, Ad034', description='RIO kinase 1 (yeast)', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38036989, end_position_on_the_genomic_accession=38061433, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  71345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71345, CurrentID=0, Status='live', Symbol='Ano9', Aliases='5430425C04Rik, Tmem16j, Tp53i5, Trp53i5', description='anoctamin 9', other_designations='Trp53 inducible protein 5|ano9|transformation related protein 53 inducible protein 5|transmembrane protein 16J|tumor protein p53 inducible protein 5', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141101218, end_position_on_the_genomic_accession=141117806, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  71351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71351, CurrentID=0, Status='live', Symbol='5430402E10Rik', Aliases='', description='RIKEN cDNA 5430402E10 gene', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=77919786, end_position_on_the_genomic_accession=77925062, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71354, CurrentID=0, Status='live', Symbol='Wdr31', Aliases='5430402I10Rik, spWD', description='WD repeat domain 31', other_designations='spermatid WD repeat-containing protein|spermatid WD-repeat protein', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62448653, end_position_on_the_genomic_accession=62470964, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71355, CurrentID=0, Status='live', Symbol='Col24a1', Aliases='5430404K19Rik', description='collagen, type XXIV, alpha 1', other_designations='procollagen, type XXIV, alpha 1', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145292472, end_position_on_the_genomic_accession=145552013, orientation='plus', exon_count=60, OMIM='', no_hdr0=''),
  71361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71361, CurrentID=0, Status='live', Symbol='Aifm2', Aliases='5430437E11Rik, Amid, D730001I10Rik, PRG3', description='apoptosis-inducing factor, mitochondrion-associated 2', other_designations='apoptosis-inducing factor (AIF)-like mitochondrion-associated inducer of death|apoptosis-inducing factor homologous mitochondrion-associated inducer of death|apoptosis-inducing factor-like mitchondrion-associated inducer of death|apoptosis-inducing factor-like mitochondrion-associated inducer of death', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61715263, end_position_on_the_genomic_accession=61739260, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71363, CurrentID=0, Status='live', Symbol='Krtap7-1', Aliases='5430433J05Rik', description='keratin associated protein 7-1', other_designations='high glycine-tyrosine keratin-associated protein 7.1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=89507703, end_position_on_the_genomic_accession=89508323, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71365, CurrentID=0, Status='live', Symbol='Pdss2', Aliases='5430420P03Rik, Gm60, Plmp, kd, mDLP1', description='prenyl (solanesyl) diphosphate synthase, subunit 2', other_designations='all-trans-decaprenyl-diphosphate synthase subunit 2|decaprenyl pyrophosphate synthase subunit 2|decaprenyl pyrophosphate synthetase subunit 2|prenyltransferase-like mitochondrial protein|solanesyl-diphosphate synthase subunit 2', map_location='10 22.91 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43220310, end_position_on_the_genomic_accession=43464882, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  71367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71367, CurrentID=0, Status='live', Symbol='Chst9', Aliases='5430438D01Rik', description='carbohydrate (N-acetylgalactosamine 4-0) sulfotransferase 9', other_designations='GalNAc4ST-2|N-acetylgalactosamine-4-O-sulfotransferase 2|galNAc-4-O-sulfotransferase 2|galNAc-4-ST2', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=15451773, end_position_on_the_genomic_accession=15760415, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71369, CurrentID=0, Status='live', Symbol='Krtap16-3', Aliases='5430404J20Rik, Krtap16-10', description='keratin associated protein 16-3', other_designations='keratin-associated protein 16.3', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88962304, end_position_on_the_genomic_accession=88962873, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71371, CurrentID=0, Status='live', Symbol='Arid5b', Aliases='4930580B11, 5430435G07Rik, AI467247, Desrt, Mrf2, Mrf2alpha, Mrf2beta', description='AT rich interactive domain 5B (MRF1-like)', other_designations='ARID class DNA binding protein|ARID domain-containing protein 5B|AT rich interactive domain 5B (Mrf1 like)|MRF-2|developmentally and sexually retarded with transient immune abnormalities protein|modulator recognition factor 2|modulator recognition factor protein 2', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=68092523, end_position_on_the_genomic_accession=68278726, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71373, CurrentID=0, Status='live', Symbol='Prr16', Aliases='5430406M13Rik, AI607429', description='proline rich 16', other_designations='proline-rich protein 16', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=51117898, end_position_on_the_genomic_accession=51305129, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  71375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71375, CurrentID=0, Status='live', Symbol='Foxn3', Aliases='5430426H20Rik, AA589593, AW556347, Ches1, Ches1l, HTLFL1', description='forkhead box N3', other_designations='checkpoint suppressor 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=99190081, end_position_on_the_genomic_accession=99563625, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71382, CurrentID=0, Status='live', Symbol='Pex1', Aliases='5430414H02Rik, E330005K07Rik, ZWS1', description='peroxisomal biogenesis factor 1', other_designations='peroxin-1', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3596066, end_position_on_the_genomic_accession=3637230, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  71389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71389, CurrentID=0, Status='live', Symbol='Chd6', Aliases='5430439G14Rik, 6330406J24Rik, CHD-6', description='chromodomain helicase DNA binding protein 6', other_designations='ATP-dependent helicase CHD6', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=160946978, end_position_on_the_genomic_accession=161109102, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  71392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71392, CurrentID=0, Status='live', Symbol='5430401F13Rik', Aliases='', description='RIKEN cDNA 5430401F13 gene', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131543762, end_position_on_the_genomic_accession=131553757, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71393, CurrentID=0, Status='live', Symbol='Kctd6', Aliases='5430433B02Rik, AU044285', description='potassium channel tetramerisation domain containing 6', other_designations='', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8214148, end_position_on_the_genomic_accession=8223570, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71395, CurrentID=0, Status='live', Symbol='5430419D17Rik', Aliases='', description='RIKEN cDNA 5430419D17 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131174402, end_position_on_the_genomic_accession=131282194, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  71398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71398, CurrentID=0, Status='live', Symbol='5430427O19Rik', Aliases='', description='RIKEN cDNA 5430427O19 gene', other_designations='', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=85776957, end_position_on_the_genomic_accession=85891499, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71405, CurrentID=0, Status='live', Symbol='Fam83c', Aliases='5530400B04Rik', description='family with sequence similarity 83, member C', other_designations='ortholog of family with sequence similarity 83, member C FAM83C (5530400B04Rik)', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155829183, end_position_on_the_genomic_accession=155834854, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71409, CurrentID=0, Status='live', Symbol='Fmnl2', Aliases='5430425K04Rik, Man', description='formin-like 2', other_designations='formin-like domain containing protein MAN|protein Man', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=52857716, end_position_on_the_genomic_accession=53134202, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  71412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71412, CurrentID=0, Status='live', Symbol='Dhrs2', Aliases='5430405K24Rik, Hep27', description='dehydrogenase/reductase member 2', other_designations='SDR family member', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55124154, end_position_on_the_genomic_accession=55241440, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  333564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333564, CurrentID=0, Status='live', Symbol='Fndc3c1', Aliases='Fndc3c, Gm784', description='fibronectin type III domain containing 3C1', other_designations='fibronectin domain-containing protein 3c|fibronectin type III domain containing protein 3C', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106420041, end_position_on_the_genomic_accession=106485576, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  71421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71421, CurrentID=0, Status='live', Symbol='Amtn', Aliases='5430427O21Rik', description='amelotin', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88376108, end_position_on_the_genomic_accession=88385916, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71425, CurrentID=0, Status='live', Symbol='Bpifb9a', Aliases='4833413D08Rik, AI413838', description='BPI fold containing family B, member 9A', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154253979, end_position_on_the_genomic_accession=154271245, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  71435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71435, CurrentID=0, Status='live', Symbol='Arhgap21', Aliases='5530401C11Rik, AA416458, ARHGAP10', description='Rho GTPase activating protein 21', other_designations='rho GTPase-activating protein 10|rho-type GTPase-activating protein 21', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=20847919, end_position_on_the_genomic_accession=20968638, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  71436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71436, CurrentID=0, Status='live', Symbol='Flrt3', Aliases='5530600M07Rik, C430047I10Rik, mKIAA1469', description='fibronectin leucine rich transmembrane protein 3', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=140658198, end_position_on_the_genomic_accession=140671476, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  71446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71446, CurrentID=0, Status='live', Symbol='Wrb', Aliases='5530402J05Rik, C030018G21Rik, Chd5', description='tryptophan rich basic protein', other_designations='tryptophan-rich basic protein|tryptophan-rich protein', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96145354, end_position_on_the_genomic_accession=96189247, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71448, CurrentID=0, Status='live', Symbol='Tmem80', Aliases='5530601I19Rik', description='transmembrane protein 80', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141327743, end_position_on_the_genomic_accession=141337156, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  142980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=142980, CurrentID=0, Status='live', Symbol='Tlr3', Aliases='AI957183', description='toll-like receptor 3', other_designations='', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45395665, end_position_on_the_genomic_accession=45411949, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71452, CurrentID=0, Status='live', Symbol='Ankrd40', Aliases='1110011C06Rik, 5530600A18Rik, Gcap15', description='ankyrin repeat domain 40', other_designations='granule cell antiserum positive 15', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94328001, end_position_on_the_genomic_accession=94341847, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71453, CurrentID=0, Status='live', Symbol='Krtap2-4', Aliases='5530401P19Rik, KAP2.4, KRTAP2.4', description='keratin associated protein 2-4', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99614017, end_position_on_the_genomic_accession=99614846, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71458, CurrentID=0, Status='live', Symbol='Bcor', Aliases='5830466J11Rik, 8430401K06RikR, D930024N20Rik, mKIAA1575, Bcor', description='BCL6 interacting corepressor', other_designations='', map_location='X A1.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=12036737, end_position_on_the_genomic_accession=12160355, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  71461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71461, CurrentID=0, Status='live', Symbol='Ptk7', Aliases='8430404F20Rik, chz, mPTK7/CCK4', description='PTK7 protein tyrosine kinase 7', other_designations='protein chuzhoi|protein-tyrosine kinase 7|pseudo tyrosine kinase receptor 7|receptor tyrosine kinase-like protein|tyrosine-protein kinase-like 7', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46564451, end_position_on_the_genomic_accession=46630049, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  71468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71468, CurrentID=0, Status='live', Symbol='Obox1', Aliases='7420700M11Rik, AU015708, AU024152', description='oocyte specific homeobox 1', other_designations='', map_location='7 8.5 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15547257, end_position_on_the_genomic_accession=15556846, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71472, CurrentID=0, Status='live', Symbol='Usp19', Aliases='8430421I07Rik, AI047774, Zmynd9', description='ubiquitin specific peptidase 19', other_designations='deubiquitinating enzyme 19|ubiquitin specific protease 19|ubiquitin thioesterase 19|ubiquitin thiolesterase 19|ubiquitin-specific processing protease 19|ubiquitin-specific-processing protease 19', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108490607, end_position_on_the_genomic_accession=108502337, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  71474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71474, CurrentID=0, Status='live', Symbol='Ppp6r2', Aliases='1110033O10Rik, 8430411H09Rik, B230107H12Rik, Pp6r2, Saps2', description='protein phosphatase 6, regulatory subunit 2', other_designations='SAPS domain family, member 2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89211521, end_position_on_the_genomic_accession=89287015, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  71481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71481, CurrentID=0, Status='live', Symbol='Alpk1', Aliases='8430410J10Rik, LAK', description='alpha-kinase 1', other_designations='lymphocyte alpha-kinase|lymphocyte alpha-protein kinase', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127670308, end_position_on_the_genomic_accession=127780553, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  71492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71492, CurrentID=0, Status='live', Symbol='Bbs7', Aliases='8430406N16Rik', description='Bardet-Biedl syndrome 7 (human)', other_designations='BBS2-like protein 1', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36573140, end_position_on_the_genomic_accession=36613489, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  333639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333639, CurrentID=0, Status='live', Symbol='Mamld1', Aliases='G630014P10Rik', description='mastermind-like domain containing 1', other_designations='', map_location='X A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71049082, end_position_on_the_genomic_accession=71156056, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  333654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333654, CurrentID=0, Status='live', Symbol='Ppp1r13l', Aliases='AV260376, IASPP, Nkip1, wa3', description='protein phosphatase 1, regulatory (inhibitor) subunit 13 like', other_designations='NFkB interacting protein 1|NFkB-interacting protein 1|PPP1R13B-like protein|inhibitor of ASPP protein|ppp1r13l|protein iASPP', map_location='7 9.62 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19359714, end_position_on_the_genomic_accession=19378540, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  71514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71514, CurrentID=0, Status='live', Symbol='Sfpq', Aliases='1110004P21Rik, 2810416M14Rik, 5730453G22Rik, 9030402K04Rik, AU021830, D4Ertd314e, Gm12940, OTTMUSG00000009329, PSF, REP1', description='splicing factor proline/glutamine rich (polypyrimidine tract binding protein associated)', other_designations='DNA-binding p52/p100 complex, 100 kDa subunit|PTB-associated splicing factor|PTB-associated-splicing factor|polypyrimidine tract-binding protein-associated-splicing factor', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127020560, end_position_on_the_genomic_accession=127037016, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71517, CurrentID=0, Status='live', Symbol='9030624J02Rik', Aliases='2610029J22Rik', description='RIKEN cDNA 9030624J02 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118740197, end_position_on_the_genomic_accession=118841491, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  71519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71519, CurrentID=0, Status='live', Symbol='Cyp2u1', Aliases='8430436A10Rik', description='cytochrome P450, family 2, subfamily u, polypeptide 1', other_designations='cytochrome P450, 2u1', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=131290491, end_position_on_the_genomic_accession=131303227, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71520, CurrentID=0, Status='live', Symbol='Grap', Aliases='8430435N19Rik, BB233594', description='GRB2-related adaptor protein', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61653321, end_position_on_the_genomic_accession=61672777, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71521, CurrentID=0, Status='live', Symbol='Pds5a', Aliases='9030416H16Rik, E230024D05Rik, mKIAA0648', description='PDS5, regulator of cohesion maintenance, homolog A (S. cerevisiae)', other_designations='', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65608124, end_position_on_the_genomic_accession=65698163, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  71522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71522, CurrentID=0, Status='live', Symbol='Ggt6', Aliases='9030405D14Rik, AI427679, AW490214', description='gamma-glutamyltransferase 6', other_designations='GGT 6|gamma-glutamyltranspeptidase 6|glutathione hydrolase 6', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72435243, end_position_on_the_genomic_accession=72438566, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  333669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333669, CurrentID=0, Status='live', Symbol='Gm5134', Aliases='EG333669', description='predicted gene 5134', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75954343, end_position_on_the_genomic_accession=76009591, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  333670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333670, CurrentID=0, Status='live', Symbol='Gm867', Aliases='Gm736', description='predicted gene 867', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75937574, end_position_on_the_genomic_accession=75940685, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  667281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667281, CurrentID=0, Status='live', Symbol='H60b', Aliases='EG667281', description='histocompatibility 60b', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=22273474, end_position_on_the_genomic_accession=22289144, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71529, CurrentID=0, Status='live', Symbol='Kazn', Aliases='2310007B04Rik, 9030409G11Rik, AI256607, AI666764, AW061099, Kaz, Kazrin', description='kazrin, periplakin interacting protein', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=142102390, end_position_on_the_genomic_accession=142239401, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71532, CurrentID=0, Status='live', Symbol='Fam217b', Aliases='9030418K01Rik, AI851938', description='family with sequence similarity 217, member B', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=178414534, end_position_on_the_genomic_accession=178422181, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71538, CurrentID=0, Status='live', Symbol='Fbxo9', Aliases='9030401P18Rik, AA986398', description='f-box protein 9', other_designations='', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78081499, end_position_on_the_genomic_accession=78109065, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  71544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71544, CurrentID=0, Status='live', Symbol='Arhgap42', Aliases='9030420J04Rik, BE136619', description='Rho GTPase activating protein 42', other_designations='VRho GTPase activating protein 42|arhgap42', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=8994331, end_position_on_the_genomic_accession=9239164, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  71562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71562, CurrentID=0, Status='live', Symbol='Afmid', Aliases='9030621K19Rik, Ammd, FKF, KFA, Kf, formylase', description='arylformamidase', other_designations='FKF|KFA|KFase|N-formylkynurenine formamidase|formylkynureninase|probable arylformamidase', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117825919, end_position_on_the_genomic_accession=117839908, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  71564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71564, CurrentID=0, Status='live', Symbol='Izumo4', Aliases='9030607L17Rik, AI550407', description='IZUMO family member 4', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80702199, end_position_on_the_genomic_accession=80705382, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  71566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71566, CurrentID=0, Status='live', Symbol='Clmp', Aliases='9030425E11Rik, ACAM, ASP5, AW557819', description='CXADR-like membrane protein', other_designations='CAR-like membrane protein|adipocyte adhesion molecule|adipocyte-specific protein 5|coxsackie- and adenovirus receptor-like membrane protein', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40685964, end_position_on_the_genomic_accession=40785553, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71567, CurrentID=0, Status='live', Symbol='Mcm9', Aliases='9030408O17Rik, Gm235, Mcmdc1', description='minichromosome maintenance 9 homologous recombination repair factor', other_designations='DNA replication licensing factor MCM9|mini-chromosome maintenance deficient 9|mini-chromosome maintenance deficient domain-containing protein 1|minichromosome maintenance 9|minichromosome maintenance complex component 9|minichromosome maintenance deficient domain containing 1', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=53536315, end_position_on_the_genomic_accession=53630439, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  333715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333715, CurrentID=0, Status='live', Symbol='H2-M10.2', Aliases='4.7H', description='histocompatibility 2, M region locus 10.2', other_designations='major histocompatibility complex class Ib M10.1', map_location='17 19.01 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36284281, end_position_on_the_genomic_accession=36286421, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71578, CurrentID=0, Status='live', Symbol='Sval1', Aliases='9130404C18Rik, SLP-C', description='seminal vesicle antigen-like 1', other_designations='SVA-like protein in the colon', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41951628, end_position_on_the_genomic_accession=41956098, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71583, CurrentID=0, Status='live', Symbol='9130008F23Rik', Aliases='', description='RIKEN cDNA 9130008F23 gene', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40875482, end_position_on_the_genomic_accession=40880558, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71584, CurrentID=0, Status='live', Symbol='Gdpd2', Aliases='9130017L10Rik, Gde3, Obdpf', description='glycerophosphodiester phosphodiesterase domain containing 2', other_designations='glycerophosphodiester phosphodiesterase 3|osteoblast differentiation promoting factor', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100729836, end_position_on_the_genomic_accession=100738900, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  71586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71586, CurrentID=0, Status='live', Symbol='Ifih1', Aliases='9130009C22Rik, Helicard, Hlcd, MDA5, RLR-2', description='interferon induced with helicase C domain 1', other_designations='RIG-I-like receptor 2|helicase with 2 CARD domains|interferon induced with helicase C domain protein 1|melanoma differentiation-associated protein 5', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=62595612, end_position_on_the_genomic_accession=62646255, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71591, CurrentID=0, Status='live', Symbol='Zfp251', Aliases='9130001M19Rik, AV339307', description='zinc finger protein 251', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76852131, end_position_on_the_genomic_accession=76871435, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71592, CurrentID=0, Status='live', Symbol='Pogk', Aliases='9130401E23Rik, BASS2, mKIAA1513', description='pogo transposable element with KRAB domain', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166393610, end_position_on_the_genomic_accession=166409858, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71597, CurrentID=0, Status='live', Symbol='Isx', Aliases='9130012O13Rik, Pix-1', description='intestine specific homeobox', other_designations='pancreas-intestine homeodomain transcription factor', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=74873174, end_position_on_the_genomic_accession=74893506, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71599, CurrentID=0, Status='live', Symbol='Senp8', Aliases='9130010J17Rik, AU020827, Den1, Nedp1, Prsc2', description='SUMO/sentrin specific peptidase 8', other_designations='NEDD8-specific protease 1|SUMO/sentrin specific protease 8|SUMO/sentrin specific protease family member 8|deneddylase-1|protease, cysteine, 2 (NEDD8 specific)|sentrin/SUMO-specific protease SENP8', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59734259, end_position_on_the_genomic_accession=59750649, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71601, CurrentID=0, Status='live', Symbol='Ceacam20', Aliases='9130012D09Rik', description='carcinoembryonic antigen-related cell adhesion molecule 20', other_designations='CEA-related cell adhesion molecule 20', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19965352, end_position_on_the_genomic_accession=19991113, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  71602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71602, CurrentID=0, Status='live', Symbol='Myo1e', Aliases='2310020N23Rik, 9130023P14Rik, AA407778, myosin-1e, myr 3', description='myosin IE', other_designations='molecular motor|unconventional myosin 1E', map_location='9 39.41 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70207350, end_position_on_the_genomic_accession=70400067, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  71607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71607, CurrentID=0, Status='live', Symbol='Snx20', Aliases='9130017C17Rik', description='sorting nexin 20', other_designations='', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88626561, end_position_on_the_genomic_accession=88636133, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71609, CurrentID=0, Status='live', Symbol='Tradd', Aliases='9130005N23Rik, AA930854', description='TNFRSF1A-associated via death domain', other_designations='TNF receptor 1 associated signal transducer|TNFR1-associated DEATH domain protein|tradd', map_location='8 53.04 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105258575, end_position_on_the_genomic_accession=105264594, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71617, CurrentID=0, Status='live', Symbol='9130011E15Rik', Aliases='AI431055, AW050122', description='RIKEN cDNA 9130011E15 gene', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45817364, end_position_on_the_genomic_accession=45998593, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  71619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71619, CurrentID=0, Status='live', Symbol='Arl14', Aliases='9130014L17Rik, Arf7', description='ADP-ribosylation factor-like 14', other_designations='ADP-ribosylation factor 7', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69222419, end_position_on_the_genomic_accession=69223618, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71640, CurrentID=0, Status='live', Symbol='Zfp949', Aliases='4930422I07Rik', description='zinc finger protein 949', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88548020, end_position_on_the_genomic_accession=88571086, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71643, CurrentID=0, Status='live', Symbol='Zgrf1', Aliases='4930422G04Rik, AI448607', description='zinc finger, GRF-type containing 1', other_designations='GRF-type zinc finger domain-containing protein 1', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127553436, end_position_on_the_genomic_accession=127618023, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  333789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333789, CurrentID=0, Status='live', Symbol='N4bp2', Aliases='B3bp, E430014I16, Gm1791, Gm868', description='NEDD4 binding protein 2', other_designations='Bcl3 binding protein', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65761726, end_position_on_the_genomic_accession=65830107, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  71648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71648, CurrentID=0, Status='live', Symbol='Optn', Aliases='4930441O07Rik, FIP2, HYPL, NRP', description='optineurin', other_designations='TFIIIA-INTP', map_location='2 3.15 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5020520, end_position_on_the_genomic_accession=5064112, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71653, CurrentID=0, Status='live', Symbol='Shtn1', Aliases='4930506M07Rik, Kiaa1598, Shootin1, mKIAA1598', description='shootin 1', other_designations='', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58973358, end_position_on_the_genomic_accession=59076122, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  71660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71660, CurrentID=0, Status='live', Symbol='Rarres2', Aliases='0610007L05Rik, AI303516', description='retinoic acid receptor responder (tazarotene induced) 2', other_designations='chemerin', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48569696, end_position_on_the_genomic_accession=48572844, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71664, CurrentID=0, Status='live', Symbol='Mettl7b', Aliases='0610006F02Rik, AI266817', description='methyltransferase like 7B', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128958277, end_position_on_the_genomic_accession=128960988, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71665, CurrentID=0, Status='live', Symbol='Fuca1', Aliases='0610006A03Rik, 9530055J05Rik, Afuc, Fuca', description='fucosidase, alpha-L- 1, tissue', other_designations='alpha-L-fucosidase 1|alpha-L-fucosidase I|alpha-L-fucoside fucohydrolase 1|fucosidase, alpha-L-1, tissue', map_location='4 68.01 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135920726, end_position_on_the_genomic_accession=135940300, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71667, CurrentID=0, Status='live', Symbol='Tmem248', Aliases='0610007L01Rik, A930023A16Rik, AW557951, G430067H08Rik', description='transmembrane protein 248', other_designations='UPF0458 protein C7orf42 homolog', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130219744, end_position_on_the_genomic_accession=130243765, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  99326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99326, CurrentID=0, Status='live', Symbol='Garnl3', Aliases='AW120551', description='GTPase activating RANGAP domain-like 3', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32986366, end_position_on_the_genomic_accession=33131888, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  71670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71670, CurrentID=0, Status='live', Symbol='Acy3', Aliases='0610006H10Rik, AA3, AAIII, AW107362, Acy-3, HCBP1', description='aspartoacylase (aminoacylase) 3', other_designations='acylase III|aminoacylase III|aminoacylase-3|aspartoacylase-2|hepatitis C virus core-binding protein 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3986570, end_position_on_the_genomic_accession=3990007, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71673, CurrentID=0, Status='live', Symbol='Rnf215', Aliases='0610009J22Rik, AW557025, C77903', description='ring finger protein 215', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4135153, end_position_on_the_genomic_accession=4141193, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71675, CurrentID=0, Status='live', Symbol='0610010F05Rik', Aliases='Kiaa1841, mKIAA1841', description='RIKEN cDNA 0610010F05 gene', other_designations='', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23564949, end_position_on_the_genomic_accession=23634992, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  71678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71678, CurrentID=0, Status='live', Symbol='Brox', Aliases='0610010K06Rik', description='BRO1 domain and CAAX motif containing', other_designations='BRO1 domain- and CAAX motif-containing protein', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=183276342, end_position_on_the_genomic_accession=183297274, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  71679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71679, CurrentID=0, Status='live', Symbol='Atp5h', Aliases='0610009D10Rik', description='ATP synthase, H+ transporting, mitochondrial F0 complex, subunit D', other_designations='ATPase subunit d', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115415691, end_position_on_the_genomic_accession=115419919, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71682, CurrentID=0, Status='live', Symbol='Wdr27', Aliases='0610012K18Rik', description='WD repeat domain 27', other_designations='', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14784832, end_position_on_the_genomic_accession=14943151, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  71683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71683, CurrentID=0, Status='live', Symbol='Gypc', Aliases='0610037F22Rik, Cd236r', description='glycophorin C', other_designations='Gerbich antigen', map_location='18 18.05 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32526129, end_position_on_the_genomic_accession=32560126, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  71684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71684, CurrentID=0, Status='live', Symbol='Rbm43', Aliases='0610033I05Rik', description='RNA binding motif protein 43', other_designations='RNA-binding motif protein 43', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=51924448, end_position_on_the_genomic_accession=51936264, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71685, CurrentID=0, Status='live', Symbol='Galnt14', Aliases='0610033M06Rik', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 14', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 14|galNAc-T14|polypeptide GalNAc transferase 14|pp-GaNTase 14|protein-UDP acetylgalactosaminyltransferase 14', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=73493207, end_position_on_the_genomic_accession=73710461, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71687, CurrentID=0, Status='live', Symbol='Tmem25', Aliases='0610039J01Rik, AI429491', description='transmembrane protein 25', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44793772, end_position_on_the_genomic_accession=44799273, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71690, CurrentID=0, Status='live', Symbol='Esm1', Aliases='0610042H23Rik, AV004503, ESM-1', description='endothelial cell-specific molecule 1', other_designations='', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113209659, end_position_on_the_genomic_accession=113218104, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  71691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71691, CurrentID=0, Status='live', Symbol='Pnmal1', Aliases='0710005I19Rik', description='PNMA-like 1', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16949837, end_position_on_the_genomic_accession=16964607, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71693, CurrentID=0, Status='live', Symbol='Colec11', Aliases='CL-K1', description='collectin sub-family member 11', other_designations='collectin kidney protein 1', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28594172, end_position_on_the_genomic_accession=28623454, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71699, CurrentID=0, Status='live', Symbol='Slc41a3', Aliases='1010001P06Rik, AI480742, SLC41A1-L2', description='solute carrier family 41, member 3', other_designations='', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90604729, end_position_on_the_genomic_accession=90646412, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71701, CurrentID=0, Status='live', Symbol='Pnpt1', Aliases='1200003F12Rik, Old35, PNPase, Pnptl1', description='polyribonucleotide nucleotidyltransferase 1', other_designations="3'-5' RNA exonuclease OLD35|PNPase 1|PNPase old-35|polynucleotide phosphorylase 1|polynucleotide phosphorylase-like protein", map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29130215, end_position_on_the_genomic_accession=29162010, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  71702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71702, CurrentID=0, Status='live', Symbol='Cdc5l', Aliases='1200002I02Rik, AA408004, PCDC5RP', description='cell division cycle 5-like (S. pombe)', other_designations='cdc5-like protein|cell division cycle 5-related protein', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45391887, end_position_on_the_genomic_accession=45433707, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71703, CurrentID=0, Status='live', Symbol='Armcx3', Aliases='1200004E24Rik, AI450003, ALEX3', description='armadillo repeat containing, X-linked 3', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134756568, end_position_on_the_genomic_accession=134761457, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71704, CurrentID=0, Status='live', Symbol='Arhgef3', Aliases='1200004I24Rik, 9830169H03Rik, C76747, Xpln', description='Rho guanine nucleotide exchange factor (GEF) 3', other_designations='', map_location='14 16.09 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=27114823, end_position_on_the_genomic_accession=27403916, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  71706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71706, CurrentID=0, Status='live', Symbol='Slc46a3', Aliases='1200006F02Rik', description='solute carrier family 46, member 3', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147877846, end_position_on_the_genomic_accession=147894864, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71707, CurrentID=0, Status='live', Symbol='Ubiad1', Aliases='1200002M06Rik, AI426463, AW320947, Tere1', description='UbiA prenyltransferase domain containing 1', other_designations='transitional epithelia response protein', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148434497, end_position_on_the_genomic_accession=148444751, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71709, CurrentID=0, Status='live', Symbol='Syde1', Aliases='1200008N06Rik, BB043000, mSYD1A', description='synapse defective 1, Rho GTPase, homolog 1 (C. elegans)', other_designations='protein syd-1 homolog 1|synapse defective protein 1 homolog 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78584503, end_position_on_the_genomic_accession=78591964, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71710, CurrentID=0, Status='live', Symbol='Lrrcc1', Aliases='1200008A14Rik, 4932441F23Rik, AI195358, AI447421', description='leucine rich repeat and coiled-coil domain containing 1', other_designations='', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14533788, end_position_on_the_genomic_accession=14572658, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  71711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71711, CurrentID=0, Status='live', Symbol='Mus81', Aliases='1200008A18Rik, AI182501, AW045863', description='MUS81 endonuclease homolog (yeast)', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5482836, end_position_on_the_genomic_accession=5488409, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71712, CurrentID=0, Status='live', Symbol='Dram1', Aliases='1200002N14Rik, Dram', description='DNA-damage regulated autophagy modulator 1', other_designations='damage-regulated autophagy modulator', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88322801, end_position_on_the_genomic_accession=88364383, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71713, CurrentID=0, Status='live', Symbol='Cdc40', Aliases='1200003H23Rik, EHB3, PRP17', description='cell division cycle 40', other_designations='PRP17 homolog|cell division cycle 40 homolog', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40831621, end_position_on_the_genomic_accession=40883143, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71715, CurrentID=0, Status='live', Symbol='Dhx35', Aliases='1200009D07Rik, Ddx35', description='DEAH (Asp-Glu-Ala-His) box polypeptide 35', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158794770, end_position_on_the_genomic_accession=158858220, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  71718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71718, CurrentID=0, Status='live', Symbol='Telo2', Aliases='1200003M09Rik, AI415602, Tel2, mKIAA0683', description='TEL2, telomere maintenance 2, homolog (S. cerevisiae)', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25099569, end_position_on_the_genomic_accession=25115971, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  71720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71720, CurrentID=0, Status='live', Symbol='Osbpl3', Aliases='1200014M06Rik, 6720421I08Rik, A530055M08, ORP3, OSBP3', description='oxysterol binding protein-like 3', other_designations='ORP-3|OSBP-related protein 3', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=50293326, end_position_on_the_genomic_accession=50456355, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  71721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71721, CurrentID=0, Status='live', Symbol='Fam13c', Aliases='1200015N20Rik, C030038O19Rik, mKIAA1796', description='family with sequence similarity 13, member C', other_designations='', map_location='10 B5.2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=70440668, end_position_on_the_genomic_accession=70558732, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  71722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71722, CurrentID=0, Status='live', Symbol='Cic', Aliases='1200010B10Rik, mKIAA0306', description='capicua transcriptional repressor', other_designations='capicua homolog', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25267657, end_position_on_the_genomic_accession=25294159, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  71723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71723, CurrentID=0, Status='live', Symbol='Dhx34', Aliases='1200013B07Rik, 1810012L18Rik, Ddx34, mKIAA0134', description='DEAH (Asp-Glu-Ala-His) box polypeptide 34', other_designations='DEAH box protein 34', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16197220, end_position_on_the_genomic_accession=16222078, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  71724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71724, CurrentID=0, Status='live', Symbol='Aox3', Aliases='1200011D03Rik, AI326299, Aoh1, Aoh2', description='aldehyde oxidase 3', other_designations='aldehyde oxidase 1|aldehyde oxidase homolog 1|aldehyde oxidase structural homolog 2|azaheterocycle hydroxylase 3', map_location='1 28.96 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58113136, end_position_on_the_genomic_accession=58200452, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  71726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71726, CurrentID=0, Status='live', Symbol='Smug1', Aliases='1200013B09Rik, A930006H09Rik, C85220', description='single-strand selective monofunctional uracil DNA glycosylase', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103153290, end_position_on_the_genomic_accession=103168895, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71728, CurrentID=0, Status='live', Symbol='Stk11ip', Aliases='1200014D22Rik, BB131189, LIP1, LKB1IP', description='serine/threonine kinase 11 interacting protein', other_designations='LKB1-interacting protein 1|serine/threonine kinase 11-interacting protein', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75521529, end_position_on_the_genomic_accession=75537335, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  71729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71729, CurrentID=0, Status='live', Symbol='Rgs12', Aliases='1200016K18Rik, 4632412M04Rik, AI481290, E130309H11', description='regulator of G-protein signaling 12', other_designations='', map_location='5 17.96 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34949387, end_position_on_the_genomic_accession=35039642, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  71732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71732, CurrentID=0, Status='live', Symbol='Vps11', Aliases='1200011A11Rik', description='vacuolar protein sorting 11 (yeast)', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44347881, end_position_on_the_genomic_accession=44361675, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71733, CurrentID=0, Status='live', Symbol='Susd2', Aliases='1200011D11Rik', description='sushi domain containing 2', other_designations='', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75636619, end_position_on_the_genomic_accession=75644008, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71735, CurrentID=0, Status='live', Symbol='Lrwd1', Aliases='1200011O22Rik, AU042569, AW548074, Orca', description='leucine-rich repeats and WD repeat domain containing 1', other_designations='ORC-associated protein|origin recognition complex-associated protein', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136123066, end_position_on_the_genomic_accession=136136074, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71738, CurrentID=0, Status='live', Symbol='Mamdc2', Aliases='1200015L10Rik, AI591564, mamcan', description='MAM domain containing 2', other_designations='MAM domain-containing proteoglycan', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=23302609, end_position_on_the_genomic_accession=23448500, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  333883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=333883, CurrentID=0, Status='live', Symbol='Cd59b', Aliases='', description='CD59b antigen', other_designations='MAC-IP|MAC-inhibitory protein|MACIF|membrane attack complex inhibition factor|protectin', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104069780, end_position_on_the_genomic_accession=104085790, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71740, CurrentID=0, Status='live', Symbol='Pvrl4', Aliases='1200017F15Rik, Prr4', description='poliovirus receptor-related 4', other_designations='ig superfamily receptor LNIR|nectin 4|poliovirus receptor-related protein 4', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171370173, end_position_on_the_genomic_accession=171388287, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71742, CurrentID=0, Status='live', Symbol='Ulk3', Aliases='1200015E14Rik', description='unc-51-like kinase 3', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57589137, end_position_on_the_genomic_accession=57596233, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  71743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71743, CurrentID=0, Status='live', Symbol='Coasy', Aliases='1300003G02Rik, Dpck, Ppat, Ukr1', description='Coenzyme A synthase', other_designations="4'-phosphopantetheine adenylyltransferase and dephospho-CoA kinase|CoA synthase", map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101082562, end_position_on_the_genomic_accession=101086619, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71745, CurrentID=0, Status='live', Symbol='Cul2', Aliases='1300003D18Rik, 4932411N15Rik, AI327301, mKIAA4106', description='cullin 2', other_designations='CUL-2', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=3382975, end_position_on_the_genomic_accession=3436700, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  71746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71746, CurrentID=0, Status='live', Symbol='Rgl3', Aliases='1300003D20Rik, mFLJ00153', description='ral guanine nucleotide dissociation stimulator-like 3', other_designations='RPM|Ral guanine-nucleotide exchange factor|RalGDS-like protein 3|ral guanine nucleotide dissociation stimulator like 3|ralGDS-like 3|ralGDS-related effector protein of M-Ras|ras pathway modulator', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21971179, end_position_on_the_genomic_accession=21989465, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  71750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71750, CurrentID=0, Status='live', Symbol='R3hdm2', Aliases='1300003K24Rik, AU041262, mKIAA1002', description='R3H domain containing 2', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127380332, end_position_on_the_genomic_accession=127499384, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  71751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71751, CurrentID=0, Status='live', Symbol='Map3k13', Aliases='C130026N12Rik, LZK', description='mitogen-activated protein kinase kinase kinase 13', other_designations='mitogen activated protein kinase kinase kinase 13', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21825171, end_position_on_the_genomic_accession=21933439, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  71752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71752, CurrentID=0, Status='live', Symbol='Gtf3c2', Aliases='1300004C11Rik, 2610510G03Rik, AI225816, AU041069, TFIIIC110, mKIAA0011', description='general transcription factor IIIC, polypeptide 2, beta', other_designations='TF3C-beta|TFIIIC 110 kDa subunit|TFIIIC-BETA|transcription factor IIIC 110 kDa subunit|transcription factor IIIC subunit beta', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31156005, end_position_on_the_genomic_accession=31180658, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  71753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71753, CurrentID=0, Status='live', Symbol='Tmprss6', Aliases='1300008A22Rik', description='transmembrane serine protease 6', other_designations='matriptase 2|matriptase-2|transmembrane protease, serine 6|type II transmembrane serine protease 6', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78439667, end_position_on_the_genomic_accession=78468648, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  71754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71754, CurrentID=0, Status='live', Symbol='Cyp2d40', Aliases='1300013D18Rik, CYPIID3, P450-DB3', description='cytochrome P450, family 2, subfamily d, polypeptide 40', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82759833, end_position_on_the_genomic_accession=82764122, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71755, CurrentID=0, Status='live', Symbol='Dhdh', Aliases='1300018L09Rik, AA591799, Xld-1, Xld1', description='dihydrodiol dehydrogenase (dimeric)', other_designations='D-xylose 1-dehydrogenase|D-xylose-NADP dehydrogenase|dimeric dihydrodiol dehydrogenase|xylose dehydrogenase 1', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45473563, end_position_on_the_genomic_accession=45488873, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71756, CurrentID=0, Status='live', Symbol='Cpn2', Aliases='1300018K11Rik', description='carboxypeptidase N, polypeptide 2', other_designations='carboxypeptidase N 83 kDa chain|carboxypeptidase N large subunit|carboxypeptidase N regulatory subunit|carboxypeptidase N, polypeptide 2 homolog', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30256378, end_position_on_the_genomic_accession=30267614, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71760, CurrentID=0, Status='live', Symbol='Etnppl', Aliases='1300019H02Rik, AI195447, Agxt2l1', description='ethanolamine phosphate phospholyase', other_designations='alanine--glyoxylate aminotransferase 2-like 1|alanine-glyoxylate aminotransferase 2-like 1', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=130617448, end_position_on_the_genomic_accession=130635750, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  71761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71761, CurrentID=0, Status='live', Symbol='Amdhd1', Aliases='1300019J08Rik', description='amidohydrolase domain containing 1', other_designations='amidohydrolase domain-containing protein 1', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93523338, end_position_on_the_genomic_accession=93540043, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71764, CurrentID=0, Status='live', Symbol='C2cd2l', Aliases='1300006O23Rik, Tmem24', description='C2 calcium-dependent domain containing 2-like', other_designations='transmembrane protein 24', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44309237, end_position_on_the_genomic_accession=44320324, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  71765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71765, CurrentID=0, Status='live', Symbol='Klhdc3', Aliases='1300011D16Rik, Peas', description='kelch domain containing 3', other_designations='protein Peas|testis intracellular mediator protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46674550, end_position_on_the_genomic_accession=46680930, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  71766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71766, CurrentID=0, Status='live', Symbol='Raver1', Aliases='1300006N24Rik, AA050359, mKIAA1978', description='ribonucleoprotein, PTB-binding 1', other_designations='protein raver-1', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21074158, end_position_on_the_genomic_accession=21091988, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  71767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71767, CurrentID=0, Status='live', Symbol='Tysnd1', Aliases='1300019N10Rik', description='trypsin domain containing 1', other_designations='trypsin domain-containing protein 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61695497, end_position_on_the_genomic_accession=61702773, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71768, CurrentID=0, Status='live', Symbol='Vwce', Aliases='1300015B04Rik', description='von Willebrand factor C and EGF domains', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10634229, end_position_on_the_genomic_accession=10665210, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  71769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71769, CurrentID=0, Status='live', Symbol='Bbs10', Aliases='1300007O09Rik, AI452285', description='Bardet-Biedl syndrome 10 (human)', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=111298679, end_position_on_the_genomic_accession=111301736, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  71770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71770, CurrentID=0, Status='live', Symbol='Ap2b1', Aliases='1300012O03Rik, AI788979', description='adaptor-related protein complex 2, beta 1 subunit', other_designations='AP105B|adapter-related protein complex 2 beta subunit|adapter-related protein complex 2 subunit beta|adaptor protein complex AP-2 subunit beta|adaptor-related protein complex 2 subunit beta|beta-2-adaptin|beta-adaptin|beta2-adaptin|clathrin assembly protein complex 2 beta large chain|plasma membrane adaptor HA2/AP2 adaptin beta subunit', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83298969, end_position_on_the_genomic_accession=83405035, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  71772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71772, CurrentID=0, Status='live', Symbol='Plbd2', Aliases='1300012G16Rik, AU019810, P76', description='phospholipase B domain containing 2', other_designations='66.3 kDa protein|76 kDa protein|LAMA-like protein 2|lamina ancestor homolog 2|phospholipase B domain-containing protein 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120483893, end_position_on_the_genomic_accession=120503623, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71773, CurrentID=0, Status='live', Symbol='Ugt2b1', Aliases='1300012D20Rik', description='UDP glucuronosyltransferase 2 family, polypeptide B1', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86916639, end_position_on_the_genomic_accession=86926503, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71774, CurrentID=0, Status='live', Symbol='Shroom1', Aliases='Apx, Shrm1', description='shroom family member 1', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53456709, end_position_on_the_genomic_accession=53467766, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71775, CurrentID=0, Status='live', Symbol='1300017J02Rik', Aliases='Ica, mICA', description='RIKEN cDNA 1300017J02 gene', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103250523, end_position_on_the_genomic_accession=103288297, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71776, CurrentID=0, Status='live', Symbol='Tha1', Aliases='1300017K07Rik, GLY1', description='threonine aldolase 1', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117867945, end_position_on_the_genomic_accession=117873526, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71777, CurrentID=0, Status='live', Symbol='Ing3', Aliases='1300013A07Rik, P47ING3', description='inhibitor of growth family, member 3', other_designations='p47 regulator protein', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=21949217, end_position_on_the_genomic_accession=21976037, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71778, CurrentID=0, Status='live', Symbol='Klhl5', Aliases='1300013C10Rik', description='kelch-like 5', other_designations='', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65107527, end_position_on_the_genomic_accession=65168186, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  71779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71779, CurrentID=0, Status='live', Symbol='March8', Aliases='1300017E09Rik, MARCH-VIII, Mir', description='membrane-associated ring finger (C3HC4) 8', other_designations='c-mir, cellular modulator of immune recognition|cellular modulator of immune recognition|membrane-associated RING finger protein 8|membrane-associated RING-CH protein VIII', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116337884, end_position_on_the_genomic_accession=116409540, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71780, CurrentID=0, Status='live', Symbol='Isyna1', Aliases='1300017C10Rik, AU018670', description='myo-inositol 1-phosphate synthase A1', other_designations='IPS 1|MI-1-P synthase|MIP synthase|myo-inositol-1-phosphate synthase', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70594481, end_position_on_the_genomic_accession=70597290, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71781, CurrentID=0, Status='live', Symbol='Slc16a14', Aliases='1110004H10Rik', description='solute carrier family 16 (monocarboxylic acid transporters), member 14', other_designations='MCT 14|solute carrier family 16 member 14', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=84905457, end_position_on_the_genomic_accession=84935281, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71782, CurrentID=0, Status='live', Symbol='Ankle2', Aliases='1110001J12Rik, AI661024, D5Ertd585e', description='ankyrin repeat and LEM domain containing 2', other_designations='LEM domain-containing protein 4', map_location='5 53.4 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110231004, end_position_on_the_genomic_accession=110256651, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71785, CurrentID=0, Status='live', Symbol='Pdgfd', Aliases='1110003I09Rik', description='platelet-derived growth factor, D polypeptide', other_designations='PDGF-D|SCDGF-B|spinal cord-derived growth factor B', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=6168362, end_position_on_the_genomic_accession=6536014, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  71787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71787, CurrentID=0, Status='live', Symbol='Trnau1ap', Aliases='1110007F05Rik, AA387999, Secp43, Trspap1, WI1-2643G14.2', description='tRNA selenocysteine 1 associated protein 1', other_designations='tRNA selenocysteine associated protein 1|tRNA selenocysteine-associated protein 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132311763, end_position_on_the_genomic_accession=132329545, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71790, CurrentID=0, Status='live', Symbol='Anxa9', Aliases='1110003P15Rik, 2310069F17Rik', description='annexin A9', other_designations='annexin XXXI|annexin-31|annexin-9', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95296093, end_position_on_the_genomic_accession=95314891, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71791, CurrentID=0, Status='live', Symbol='Cpa4', Aliases='1110019K20Rik, AV009555', description='carboxypeptidase A4', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30568376, end_position_on_the_genomic_accession=30591747, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71793, CurrentID=0, Status='live', Symbol='Ints12', Aliases='1110020M19Rik, 2810027J24Rik, 4930529N21Rik, A230056J18Rik, AI666778, Phf22', description='integrator complex subunit 12', other_designations='PHD finger protein 22|int12', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=133091953, end_position_on_the_genomic_accession=133110988, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71795, CurrentID=0, Status='live', Symbol='Pitpnc1', Aliases='1110020B03Rik, 5830436L09Rik, AI662802, AI851387, C330017I21Rik, Dnr411, RDGBB, RDGBB1, rdgB-beta', description='phosphatidylinositol transfer protein, cytoplasmic 1', other_designations='M-rdgB beta|diabetic nephropathy-related protein|mM-rdgBbeta|mammalian rdgB homolog beta|mrdgBbeta|rdgBbeta|retinal degeneration B beta|retinal degeneration B homolog beta', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107207892, end_position_on_the_genomic_accession=107470892, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71797, CurrentID=0, Status='live', Symbol='Chst13', Aliases='1110067M19Rik, C4ST-3', description='carbohydrate (chondroitin 4) sulfotransferase 13', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90308351, end_position_on_the_genomic_accession=90325185, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  71799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71799, CurrentID=0, Status='live', Symbol='Ptcd1', Aliases='1110069M14Rik', description='pentatricopeptide repeat domain 1', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145147378, end_position_on_the_genomic_accession=145167119, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71801, CurrentID=0, Status='live', Symbol='Plekhf2', Aliases='1110070J07Rik, AA673237, ZFYVE18', description='pleckstrin homology domain containing, family F (with FYVE domain) member 2', other_designations='PH domain-containing family F member 2', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=10988662, end_position_on_the_genomic_accession=11007619, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71803, CurrentID=0, Status='live', Symbol='Slc25a18', Aliases='1500015I14Rik, AW125787', description='solute carrier family 25 (mitochondrial carrier), member 18', other_designations='GC-2|glutamate/H(+) symporter 2|solute carrier family 25 member 18', map_location='6 57.0 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120773510, end_position_on_the_genomic_accession=120794169, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71804, CurrentID=0, Status='live', Symbol='Mtfr2', Aliases='2610016C23Rik, 4933412C16Rik, Dufd1, Fam54a', description='mitochondrial fission regulator 2', other_designations='DUF729 domain containing 1|DUF729 domain-containing protein 1|family with sequence similarity 54, member A|protein FAM54A', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=20346989, end_position_on_the_genomic_accession=20361669, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71805, CurrentID=0, Status='live', Symbol='Nup93', Aliases='2410008G02Rik', description='nucleoporin 93', other_designations='93 kDa nucleoporin|CBP-interacting protein 4|nucleoporin Nup93', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94214541, end_position_on_the_genomic_accession=94315066, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  71807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71807, CurrentID=0, Status='live', Symbol='Tars2', Aliases='2610024N01Rik, AI429208, Tarsl1', description='threonyl-tRNA synthetase 2, mitochondrial (putative)', other_designations='thrRS|threonyl-tRNA synthetase, mitochondrial|threonyl-tRNA synthetase-like 1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95739974, end_position_on_the_genomic_accession=95754977, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  71810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71810, CurrentID=0, Status='live', Symbol='Ranbp3', Aliases='2610024N24Rik, AA408221', description='RAN binding protein 3', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56673225, end_position_on_the_genomic_accession=56711769, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  71816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71816, CurrentID=0, Status='live', Symbol='Rnf180', Aliases='3110001E11Rik, Rines', description='ring finger protein 180', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=105130567, end_position_on_the_genomic_accession=105294890, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  71817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71817, CurrentID=0, Status='live', Symbol='Tmem50a', Aliases='3200001F09Rik, CAM, Smp1', description='transmembrane protein 50A', other_designations='small membrane protein 1', map_location='4 67.14 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134897849, end_position_on_the_genomic_accession=134915030, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71819, CurrentID=0, Status='live', Symbol='Kif23', Aliases='3110001D19Rik, C87313, CHO1, Knsl5, MKLP-1, MKLP1', description='kinesin family member 23', other_designations='kinesin-like 5 (mitotic kinesin-like protein 1)', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=61917277, end_position_on_the_genomic_accession=61946812, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  71820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71820, CurrentID=0, Status='live', Symbol='Wdr34', Aliases='3200002I06Rik', description='WD repeat domain 34', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30031551, end_position_on_the_genomic_accession=30048879, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71824, CurrentID=0, Status='live', Symbol='1700006A11Rik', Aliases='', description='RIKEN cDNA 1700006A11 gene', other_designations='', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=124400984, end_position_on_the_genomic_accession=124426028, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  71826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71826, CurrentID=0, Status='live', Symbol='1700001F09Rik', Aliases='Gm10377', description='RIKEN cDNA 1700001F09 gene', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=43342384, end_position_on_the_genomic_accession=43347811, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71827, CurrentID=0, Status='live', Symbol='Lrrc34', Aliases='1700007J06Rik, SPATA34', description='leucine rich repeat containing 34', other_designations='spermatogenesis associated SPATA34', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30624267, end_position_on_the_genomic_accession=30647818, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71828, CurrentID=0, Status='live', Symbol='Gtf2a1l', Aliases='1700011N16Rik, Alff, Gtf2a1l', description='general transcription factor IIA, 1-like', other_designations='TfIIa-alpha/beta-like factor|general transcription factor II A, 1-like factor|general transcription factor IIA, 1-like factor', map_location='17 E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=88665502, end_position_on_the_genomic_accession=88715157, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  71829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71829, CurrentID=0, Status='live', Symbol='Ddi1', Aliases='1700011N24Rik', description='DNA-damage inducible 1', other_designations='DDI1, DNA-damage inducible 1, homolog 1', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=6265028, end_position_on_the_genomic_accession=6266547, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71830, CurrentID=0, Status='live', Symbol='Pdilt', Aliases='1700007B13Rik', description='protein disulfide isomerase-like, testis expressed', other_designations='protein disulfide isomerase-like protein of the testis', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119486586, end_position_on_the_genomic_accession=119528621, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  71831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71831, CurrentID=0, Status='live', Symbol='1700007B14Rik', Aliases='', description='RIKEN cDNA 1700007B14 gene', other_designations='', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=75448694, end_position_on_the_genomic_accession=75984508, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  71832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71832, CurrentID=0, Status='live', Symbol='Csl', Aliases='1700007H16Rik', description='citrate synthase like', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=99757705, end_position_on_the_genomic_accession=99759658, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71833, CurrentID=0, Status='live', Symbol='Dcaf7', Aliases='1700012F10Rik, 2610037L01Rik, C86529, HAN11, Wdr68', description='DDB1 and CUL4 associated factor 7', other_designations='WD repeat domain 68|WD repeat-containing protein 68|WD repeat-containing protein An11 homolog|WD-repeat protein', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106036872, end_position_on_the_genomic_accession=106059324, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71834, CurrentID=0, Status='live', Symbol='Zbtb43', Aliases='1700010E06Rik, Zfp297b, mKIAA0414', description='zinc finger and BTB domain containing 43', other_designations='ortholog of zinc finger protein 297B ZNF297B|zinc finger protein 297B', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33450287, end_position_on_the_genomic_accession=33468655, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71835, CurrentID=0, Status='live', Symbol='Lancl2', Aliases='1700003F10Rik', description='LanC (bacterial lantibiotic synthetase component C)-like 2', other_designations='testis-specific adriamycin sensitivity protein', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57702455, end_position_on_the_genomic_accession=57739449, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71836, CurrentID=0, Status='live', Symbol='Shcbp1l', Aliases='1700012A16Rik', description='Shc SH2-domain binding protein 1-like', other_designations='', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153425098, end_position_on_the_genomic_accession=153452574, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71837, CurrentID=0, Status='live', Symbol='1700003E16Rik', Aliases='AU041756', description='RIKEN cDNA 1700003E16 gene', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83156404, end_position_on_the_genomic_accession=83162975, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71838, CurrentID=0, Status='live', Symbol='Phf7', Aliases='1700006H01Rik, 1700010P14Rik, AI427892, AW555949', description='PHD finger protein 7', other_designations='testis development protein NYD-SP6', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31237690, end_position_on_the_genomic_accession=31251375, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71839, CurrentID=0, Status='live', Symbol='Osgin1', Aliases='1700012B18Rik, Okl38', description='oxidative stress induced growth inhibitor 1', other_designations='pregnancy-induced growth inhibitor', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119434150, end_position_on_the_genomic_accession=119446257, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71840, CurrentID=0, Status='live', Symbol='Tekt4', Aliases='1700010L19Rik', description='tektin 4', other_designations='testicular microtubules-related protein 4', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25433645, end_position_on_the_genomic_accession=25476602, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71841, CurrentID=0, Status='live', Symbol='Tcp11x2', Aliases='1700008I05Rik, C77164, TCP11, Tcp11l3', description='t-complex 11 family, X-linked 2', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135654698, end_position_on_the_genomic_accession=135693790, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71843, CurrentID=0, Status='live', Symbol='R3hcc1', Aliases='1700020M16Rik, AA407796', description='R3H domain and coiled-coil containing 1', other_designations='', map_location='14 36.08 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69697304, end_position_on_the_genomic_accession=69714809, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71844, CurrentID=0, Status='live', Symbol='Nupl1', Aliases='1700017F11Rik, Nup58, mKIAA0410', description='nucleoporin like 1', other_designations='58 kDa nucleoporin|nucleoporin-like protein 1', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=60205610, end_position_on_the_genomic_accession=60251502, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  71846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71846, CurrentID=0, Status='live', Symbol='Syce2', Aliases='1700013H19Rik, AA407907, Cesc1', description='synaptonemal complex central element protein 2', other_designations='central element synaptonemal complex 1', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84872111, end_position_on_the_genomic_accession=84887446, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71847, CurrentID=0, Status='live', Symbol='Gmcl1l', Aliases='1700017G21Rik, Gclh, Gmcl2', description='germ cell-less, spermatogenesis associated-like', other_designations='germ cell-less homolog 1 (Drosophila)-like|germ cell-less protein-like 2', map_location='X A3.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=32560055, end_position_on_the_genomic_accession=32562009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71853, CurrentID=0, Status='live', Symbol='Pdia6', Aliases='1700015E05Rik, AL023058, C77895, CaBP5, P5, Txndc7', description='protein disulfide isomerase associated 6', other_designations='protein disulfide isomerase-related protein|thioredoxin domain containing 7|thioredoxin domain-containing protein 7', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=17266595, end_position_on_the_genomic_accession=17284770, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  71854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71854, CurrentID=0, Status='live', Symbol='Dpep3', Aliases='1700018F16Rik, AI507039, MBD-3, MLZ-136', description='dipeptidase 3', other_designations='membrane-bound dipeptidase 3|protein expressed in male leptotene and zygotene spermatocytes 136|putative membrane-bound dipeptidase 3', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105973520, end_position_on_the_genomic_accession=105979419, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71856, CurrentID=0, Status='live', Symbol='Wfdc3', Aliases='1700015L13Rik, 1700127F16Rik', description='WAP four-disulfide core domain 3', other_designations='putative protease inhibitor WAP14', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164731226, end_position_on_the_genomic_accession=164745910, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71860, CurrentID=0, Status='live', Symbol='Cfap52', Aliases='1700019F09Rik, 4933417B11Rik, WDRPUH, Wdr16', description='cilia and flagella associated protein 52', other_designations='WD repeat domain 16|WD repeat-containing protein 16|WD40-repear protein upregulated in HCC', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67924806, end_position_on_the_genomic_accession=67965929, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  71861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71861, CurrentID=0, Status='live', Symbol='Zswim2', Aliases='1700025P14Rik, 4933437F18Rik, MEX', description='zinc finger SWIM-type containing 2', other_designations='MEKK1-related protein X|ZZ-type zinc finger-containing protein 2|zinc finger SWIM domain-containing protein 2|zinc finger, SWIM domain containing 2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=83915079, end_position_on_the_genomic_accession=83941254, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  71862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71862, CurrentID=0, Status='live', Symbol='Gpr160', Aliases='1700025D19Rik, GPCR150', description='G protein-coupled receptor 160', other_designations='', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30855930, end_position_on_the_genomic_accession=30897194, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71863, CurrentID=0, Status='live', Symbol='1700019O17Rik', Aliases='', description='RIKEN cDNA 1700019O17 gene', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86426329, end_position_on_the_genomic_accession=86428049, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71864, CurrentID=0, Status='live', Symbol='Fam217a', Aliases='1700026J04Rik', description='family with sequence similarity 217, member A', other_designations='uncharacterized protein C6orf146 homolog', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34909959, end_position_on_the_genomic_accession=34924358, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71865, CurrentID=0, Status='live', Symbol='Fbxo30', Aliases='1700026A16Rik, MUSA1', description='F-box protein 30', other_designations='muscle ubiquitin ligase of SCF complex in atrophy-1', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=11281330, end_position_on_the_genomic_accession=11298048, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71868, CurrentID=0, Status='live', Symbol='Thegl', Aliases='1700023E05Rik', description='theg spermatid protein like', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77016023, end_position_on_the_genomic_accession=77061541, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  71869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71869, CurrentID=0, Status='live', Symbol='Serpinb12', Aliases='2300003F07Rik, 4833409F13Rik', description='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 12', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 12', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=106934449, end_position_on_the_genomic_accession=106957080, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71870, CurrentID=0, Status='live', Symbol='Cfap45', Aliases='1700028D05Rik, Ccdc19, Nesg1', description='cilia and flagella associated protein 45', other_designations='coiled-coil domain containing 19|nasopharyngeal epithelium specific protein 1', map_location='1 79.91 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172521130, end_position_on_the_genomic_accession=172545870, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  321003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321003, CurrentID=0, Status='live', Symbol='Xpnpep3', Aliases='APP3, E430012M05Rik', description='X-prolyl aminopeptidase 3, mitochondrial', other_designations='X-Pro aminopeptidase 3|X-prolyl aminopeptidase (aminopeptidase P) 3, putative|aminopeptidase P3', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81400128, end_position_on_the_genomic_accession=81457476, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71872, CurrentID=0, Status='live', Symbol='Aox4', Aliases='2310003G12Rik, Aoh2', description='aldehyde oxidase 4', other_designations='aldehyde oxidase homolog 2|aldehyde oxidase structural homolog 2|azaheterocycle hydroxylase 4|retinal oxidase', map_location='1 28.97 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58210397, end_position_on_the_genomic_accession=58268600, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  71874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71874, CurrentID=0, Status='live', Symbol='2310007B03Rik', Aliases='', description='RIKEN cDNA 2310007B03 gene', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93151354, end_position_on_the_genomic_accession=93160973, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71876, CurrentID=0, Status='live', Symbol='Cenpu', Aliases='1700029A22Rik, Mlf1ip', description='centromere protein U', other_designations='CENP-U|MLF1-interacting protein|myeloid leukemia factor 1 interacting protein', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46552019, end_position_on_the_genomic_accession=46581144, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  71877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71877, CurrentID=0, Status='live', Symbol='Efhc1', Aliases='1700029F22Rik, mRib72-1, myoclonin1', description='EF-hand domain (C-terminal) containing 1', other_designations='myoclonin-1', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20951551, end_position_on_the_genomic_accession=20990841, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71878, CurrentID=0, Status='live', Symbol='Fam83d', Aliases='2310007D09Rik, BB104611', description='family with sequence similarity 83, member D', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158768099, end_position_on_the_genomic_accession=158786637, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71881, CurrentID=0, Status='live', Symbol='Apmap', Aliases='2310001A20Rik, AI314817', description='adipocyte plasma membrane associated protein', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150583080, end_position_on_the_genomic_accession=150608557, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71883, CurrentID=0, Status='live', Symbol='Coq2', Aliases='2310002F18Rik, PHB:PPT', description='coenzyme Q2 4-hydroxybenzoate polyprenyltransferase', other_designations='4-HB polyprenyltransferase|PHB:polyprenyltransferase|coenzyme Q2 homolog, prenyltransferase|para-hydroxybenzoate--polyprenyltransferase, mitochondrial', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100654723, end_position_on_the_genomic_accession=100674285, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71884, CurrentID=0, Status='live', Symbol='Chit1', Aliases='2300002L19Rik', description='chitinase 1 (chitotriosidase)', other_designations='chitinase-1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134111242, end_position_on_the_genomic_accession=134151541, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  71885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71885, CurrentID=0, Status='live', Symbol='Faap100', Aliases='2310003H01Rik', description='Fanconi anemia core complex associated protein 100', other_designations='fanconi anemia-associated protein of 100 kDa', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120369562, end_position_on_the_genomic_accession=120378746, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71886, CurrentID=0, Status='live', Symbol='2310002L09Rik', Aliases='2310003M01Rik, AV083437', description='RIKEN cDNA 2310002L09 gene', other_designations='', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=73939371, end_position_on_the_genomic_accession=73950856, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  71887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71887, CurrentID=0, Status='live', Symbol='Ppm1j', Aliases='2310008J22Rik, PP2Czeta, Ppp2cz', description='protein phosphatase 1J', other_designations='PP2C-zeta|protein phosphatase 2C isoform zeta|protein phosphatase 2C zeta|protein phosphatase 2a, catalytic subunit, zeta isoform', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104781006, end_position_on_the_genomic_accession=104786019, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  71888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71888, CurrentID=0, Status='live', Symbol='Krt33a', Aliases='2310015J09Rik', description='keratin 33A', other_designations='K33A|hair keratin, type I Ha3-I|keratin-33A|krt33a|ortholog of keratin, hair, acidic, 3A KRTHA3A', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100011199, end_position_on_the_genomic_accession=100016212, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71889, CurrentID=0, Status='live', Symbol='Epn3', Aliases='2310022G12Rik', description='epsin 3', other_designations='EPS-15-interacting protein 3', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94489599, end_position_on_the_genomic_accession=94499974, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71890, CurrentID=0, Status='live', Symbol='Mad2l2', Aliases='2310033C13Rik, G1-453-4, MAD2B, REV7, repro22', description='MAD2 mitotic arrest deficient-like 2', other_designations='MAD2-like protein 2|mitotic arrest deficient 2-like protein 2', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148130255, end_position_on_the_genomic_accession=148145699, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  71891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71891, CurrentID=0, Status='live', Symbol='Cdadc1', Aliases='2310010M10Rik, AI449174, AI463398, BB130816, NYD-SP15', description='cytidine and dCMP deaminase domain containing 1', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59559388, end_position_on_the_genomic_accession=59597984, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  71893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71893, CurrentID=0, Status='live', Symbol='Noxo1', Aliases='2310034C04Rik, P41NOX, P41NOXA, P41NOXB, P41NOXC, Snx28, hslt', description='NADPH oxidase organizer 1', other_designations='nox organizer 1|sorting nexin-28', map_location='17 12.49 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24696234, end_position_on_the_genomic_accession=24700530, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  71897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71897, CurrentID=0, Status='live', Symbol='Lypd6b', Aliases='2310010M24Rik, AW049525', description='LY6/PLAUR domain containing 6B', other_designations='', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=49787353, end_position_on_the_genomic_accession=49948849, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  71898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71898, CurrentID=0, Status='live', Symbol='Apol9b', Aliases='2310016F22Rik', description='apolipoprotein L 9b', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77729072, end_position_on_the_genomic_accession=77736382, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71900, CurrentID=0, Status='live', Symbol='Tmem106b', Aliases='2310036D22Rik, 5830455K21Rik, 6430519M21Rik, AI428776, AI661344', description='transmembrane protein 106B', other_designations='', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=13069746, end_position_on_the_genomic_accession=13089269, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71901, CurrentID=0, Status='live', Symbol='Fam219a', Aliases='2310028H24Rik', description='family with sequence similarity 219, member A', other_designations='likely ortholog of H. sapiens chromosome 9 open reading frame 25 (C9orf25)|uncharacterized protein C9orf25 homolog', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41517691, end_position_on_the_genomic_accession=41569533, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71902, CurrentID=0, Status='live', Symbol='Cand1', Aliases='2310038O07Rik, 6330512O03Rik, AI195005, AI846556, D10Ertd516e, Tp120a, mKIAA0829', description='cullin associated and neddylation disassociated 1', other_designations='TBP-interacting protein|cullin-associated and neddylation-dissociated protein 1|p120 CAND1', map_location='10 67.08 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=119198812, end_position_on_the_genomic_accession=119240055, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  71903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71903, CurrentID=0, Status='live', Symbol='Ces2f', Aliases='2310038E17Rik, A430088E12', description='carboxylesterase 2F', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104947355, end_position_on_the_genomic_accession=104961130, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  71904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71904, CurrentID=0, Status='live', Symbol='Paqr7', Aliases='2310021M12Rik, AV001994, BB007237, Mpra, PGLP, mPR', description='progestin and adipoQ receptor family member VII', other_designations='PPAR-gamma-induced liver protein|mPR alpha', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134496761, end_position_on_the_genomic_accession=134510237, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71907, CurrentID=0, Status='live', Symbol='Serpina9', Aliases='2310014L03Rik, AW049635, Centerin', description='serine (or cysteine) peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 9', other_designations='serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 9', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103996613, end_position_on_the_genomic_accession=104013776, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71908, CurrentID=0, Status='live', Symbol='Cldn23', Aliases='2310014B08Rik', description='claudin 23', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=35824709, end_position_on_the_genomic_accession=35826559, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71909, CurrentID=0, Status='live', Symbol='Haus5', Aliases='2310022K01Rik', description='HAUS augmin-like complex, subunit 5', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30653708, end_position_on_the_genomic_accession=30664994, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  71910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71910, CurrentID=0, Status='live', Symbol='Plpp5', Aliases='1810019D05Rik, 2310022A04Rik, AW045471, Ppapdc1, Ppapdc1b', description='phospholipid phosphatase 5', other_designations='phosphatidate phosphatase PPAPDC1B|phosphatidic acid phosphatase type 2 domain containing 1|phosphatidic acid phosphatase type 2 domain containing 1B|phosphatidic acid phosphatase type 2 domain-containing protein 1B', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25720043, end_position_on_the_genomic_accession=25724887, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  71911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71911, CurrentID=0, Status='live', Symbol='Bdh1', Aliases='2310032J20Rik, AI327223, Bdh', description='3-hydroxybutyrate dehydrogenase, type 1', other_designations='3-hydroxybutyrate dehydrogenase (heart, mitochondrial)|b-hydroxybutyrate dehydrogenase', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31422297, end_position_on_the_genomic_accession=31458901, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71912, CurrentID=0, Status='live', Symbol='Jsrp1', Aliases='2300003C06Rik, 2310032K21Rik, JP-45, JP45', description='junctional sarcoplasmic reticulum protein 1', other_designations='junctional-face membrane protein of 45 kDa homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80808494, end_position_on_the_genomic_accession=80813593, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71913, CurrentID=0, Status='live', Symbol='Tmem79', Aliases='2310042N02Rik, 2310074C17Rik, ma', description='transmembrane protein 79', other_designations='matted|mattrin', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88328653, end_position_on_the_genomic_accession=88334433, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  71914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71914, CurrentID=0, Status='live', Symbol='Antxr2', Aliases='2310046B19Rik, AW561899, CMG-2, CMG2, cI-35', description='anthrax toxin receptor 2', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=97884688, end_position_on_the_genomic_accession=98030962, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71916, CurrentID=0, Status='live', Symbol='Dus4l', Aliases='2310069P03Rik, 2700089B10Rik, AI482040, Pp35', description='dihydrouridine synthase 4-like (S. cerevisiae)', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31640050, end_position_on_the_genomic_accession=31654826, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  71918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71918, CurrentID=0, Status='live', Symbol='Zcchc24', Aliases='2310047A01Rik', description='zinc finger, CCHC domain containing 24', other_designations='zcchc24', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25711640, end_position_on_the_genomic_accession=25768856, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  71919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71919, CurrentID=0, Status='live', Symbol='Rpap3', Aliases='2310042P20Rik, D15Ertd682e', description='RNA polymerase II associated protein 3', other_designations='', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97675097, end_position_on_the_genomic_accession=97705822, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71920, CurrentID=0, Status='live', Symbol='Epgn', Aliases='2310069M11Rik, epigen', description='epithelial mitogen', other_designations='EPG', map_location='', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=91027517, end_position_on_the_genomic_accession=91035212, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71923, CurrentID=0, Status='live', Symbol='Borcs6', Aliases='2310047M10Rik', description='BLOC-1 related complex subunit 6', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69059775, end_position_on_the_genomic_accession=69061576, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71924, CurrentID=0, Status='live', Symbol='Tube1', Aliases='2310061K05Rik, AI551343, Tube', description='epsilon-tubulin 1', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39133979, end_position_on_the_genomic_accession=39151058, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  71927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71927, CurrentID=0, Status='live', Symbol='Itfg1', Aliases='2310047C21Rik, AI314457, Cda08, D8Wsu49e', description='integrin alpha FG-GAP repeat containing 1', other_designations='integrin-alpha FG-GAP repeat-containing protein 1|linkin', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85717557, end_position_on_the_genomic_accession=85840949, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  71929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71929, CurrentID=0, Status='live', Symbol='Tmem123', Aliases='2310075C12Rik', description='transmembrane protein 123', other_designations='pro-oncosis receptor inducing membrane injury', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7764077, end_position_on_the_genomic_accession=7794333, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71932, CurrentID=0, Status='live', Symbol='Ephx3', Aliases='2310063B19Rik, Abhd9', description='epoxide hydrolase 3', other_designations='abhydrolase domain containing 9|abhydrolase domain-containing protein 9', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32183770, end_position_on_the_genomic_accession=32189850, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71934, CurrentID=0, Status='live', Symbol='Car13', Aliases='2310075C21Rik, Ca13', description='carbonic anhydrase 13', other_designations='CA-XIII|carbonate dehydratase XIII|carbonic anhydrase XIII', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=14641727, end_position_on_the_genomic_accession=14663002, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71939, CurrentID=0, Status='live', Symbol='Apol6', Aliases='2310076O14Rik', description='apolipoprotein L 6', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77044734, end_position_on_the_genomic_accession=77057108, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71941, CurrentID=0, Status='live', Symbol='Cars2', Aliases='2310051N18Rik, 2410044A07Rik, D530030H10Rik', description='cysteinyl-tRNA synthetase 2 (mitochondrial)(putative)', other_designations='cysRS|cysteine--tRNA ligase|cysteinyl-tRNA synthetase 2, mitochondrial|probable cysteinyl-tRNA synthetase, mitochondrial', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11514012, end_position_on_the_genomic_accession=11550778, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  71943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71943, CurrentID=0, Status='live', Symbol='Tom1l1', Aliases='2310045L10Rik, C80573, Srcasm', description='target of myb1-like 1 (chicken)', other_designations='Jerry protein|Src activating and signaling molecule|src-activating and signaling molecule protein|target of Myb-like protein 1', map_location='11 55.47 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=90645690, end_position_on_the_genomic_accession=90687661, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  71946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71946, CurrentID=0, Status='live', Symbol='Endod1', Aliases='2210414F18Rik, 2310067E08Rik, C85344', description='endonuclease domain containing 1', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14353990, end_position_on_the_genomic_accession=14381242, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71947, CurrentID=0, Status='live', Symbol='Tmem94', Aliases='2310067B10Rik, Kiaa0195, mKIAA0195', description='transmembrane protein 94', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115765433, end_position_on_the_genomic_accession=115799033, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  71949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71949, CurrentID=0, Status='live', Symbol='Cers5', Aliases='2310081H14Rik, AU045339, AW544927, Lass5, Trh4', description='ceramide synthase 5', other_designations='LAG1 homolog, ceramide synthase 5|LAG1 longevity assurance homolog 5|TRAM homolog 4|longevity assurance homolog 5|translocating chain-associating membrane protein homolog 4', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99735592, end_position_on_the_genomic_accession=99772515, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71950, CurrentID=0, Status='live', Symbol='Nanog', Aliases='2410002E02Rik, ENK, ecat4', description='Nanog homeobox', other_designations='ES cell-associated protein 4|early embryo specific expression NK family|early embryo specific expression NK-type homeobox protein|homeobox transcription factor Nanog', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122706483, end_position_on_the_genomic_accession=122714633, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71951, CurrentID=0, Status='live', Symbol='Gpc2', Aliases='2410016G05Rik', description='glypican 2 (cerebroglycan)', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138273657, end_position_on_the_genomic_accession=138279958, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  71952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71952, CurrentID=0, Status='live', Symbol='2410016O06Rik', Aliases='MAPJD, NO66', description='RIKEN cDNA 2410016O06 gene', other_designations='histone lysine demethylase NO66|lysine-specific demethylase NO66|myc associated protein with JmjC domain|myc-associated protein with JmjC domain|nucleolar protein 66', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83950608, end_position_on_the_genomic_accession=83952953, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71954, CurrentID=0, Status='live', Symbol='Suds3', Aliases='2400003N08Rik, 2410008L21Rik, AU067672, mSds3', description='suppressor of defective silencing 3 homolog (S. cerevisiae)', other_designations='suppressor of defective silencing 3 protein homolog', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117091678, end_position_on_the_genomic_accession=117115993, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  71955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71955, CurrentID=0, Status='live', Symbol='Ist1', Aliases='2400003C14Rik, AW536298, mKIAA0174', description='increased sodium tolerance 1 homolog (yeast)', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109671321, end_position_on_the_genomic_accession=109693294, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71956, CurrentID=0, Status='live', Symbol='Rnf135', Aliases='0610037N03Rik, 2410006N06Rik', description='ring finger protein 135', other_designations='Riplet|U 2-3-0', map_location='11 47.59 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80183825, end_position_on_the_genomic_accession=80199754, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  71957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71957, CurrentID=0, Status='live', Symbol='Cpsf3l', Aliases='2410006F12Rik, AU044035', description='cleavage and polyadenylation specific factor 3-like', other_designations='CPSF3-like protein|cleavage and polyadenylation-specific factor 3-like protein|int11', map_location='4 87.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155869548, end_position_on_the_genomic_accession=155889103, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  71960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71960, CurrentID=0, Status='live', Symbol='Myh14', Aliases='2400004E04Rik, II-C, NHMCII, NMHC II-C', description='myosin, heavy polypeptide 14', other_designations='myosin heavy chain 14|myosin heavy chain, non-muscle IIc|myosin, heavy chain 14|non-muscle myosin heavy chain IIc|nonmuscle myosin heavy chain', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44605803, end_position_on_the_genomic_accession=44670843, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  71962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71962, CurrentID=0, Status='live', Symbol='Gatsl3', Aliases='2410008K03Rik', description='GATS protein-like 3', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4218162, end_position_on_the_genomic_accession=4222409, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  71963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71963, CurrentID=0, Status='live', Symbol='Cdca4', Aliases='2410018C03Rik, HEPP, SEI-3', description='cell division cycle associated 4', other_designations='hematopoietic progenitor protein', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112820229, end_position_on_the_genomic_accession=112829516, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  71966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71966, CurrentID=0, Status='live', Symbol='Nkiras2', Aliases='2410003M04Rik, 4930527H08Rik, D630018G21Rik, KBRAS2', description='NFKB inhibitor interacting Ras-like protein 2', other_designations='I-kappa-B-interacting Ras-like protein 2|kappa B-Ras protein 2|kappaB-Ras2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100622955, end_position_on_the_genomic_accession=100627602, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71967, CurrentID=0, Status='live', Symbol='Mageb16', Aliases='2410003J06Rik', description='melanoma antigen family B, 16', other_designations='MAGE-B16 antigen|Novel MAGE family protein', map_location='X A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=79623231, end_position_on_the_genomic_accession=79671435, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  71968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71968, CurrentID=0, Status='live', Symbol='Wdr73', Aliases='1200011I23Rik, 2410008B13Rik, AI848574, C85352', description='WD repeat domain 73', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80890723, end_position_on_the_genomic_accession=80901269, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  71970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71970, CurrentID=0, Status='live', Symbol='Zbed5', Aliases='2410018M08Rik, Scand3', description='zinc finger, BED type containing 5', other_designations='ZSCAN domain containing 3|zinc finger, BED-type containing 5', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129895723, end_position_on_the_genomic_accession=129903623, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  71971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71971, CurrentID=0, Status='live', Symbol='Zswim1', Aliases='2410003H12Rik, AI850991, BB046916', description='zinc finger SWIM-type containing 1', other_designations='zinc finger, SWIM domain containing 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164822686, end_position_on_the_genomic_accession=164826867, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  71972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71972, CurrentID=0, Status='live', Symbol='Dnmbp', Aliases='2410003L07Rik, 2410003M15Rik, TUBA', description='dynamin binding protein', other_designations='scaffold protein Tuba', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43846815, end_position_on_the_genomic_accession=43940318, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  71973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71973, CurrentID=0, Status='live', Symbol='Rbpms2', Aliases='2400008B06Rik, AI316523', description='RNA binding protein with multiple splicing 2', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65629336, end_position_on_the_genomic_accession=65660536, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  71974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71974, CurrentID=0, Status='live', Symbol='Prmt3', Aliases='2010005E20Rik, 2410018A17Rik, AL033309, Hrmt1l3', description='protein arginine N-methyltransferase 3', other_designations='heterogeneous nuclear ribonucleoprotein methyltransferase-like 3|heterogeneous nuclear ribonucleoprotein methyltransferase-like protein 3|protein arginine methyltransferase 3', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=49778358, end_position_on_the_genomic_accession=49858265, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  71978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71978, CurrentID=0, Status='live', Symbol='Ppp2r2a', Aliases='2410004D02Rik, mKIAA1541', description='protein phosphatase 2, regulatory subunit B, alpha', other_designations='PP2A subunit B isoform B55-alpha|PP2A subunit B isoform PR55-alpha|PP2A subunit B isoform R2-alpha|PP2A subunit B isoform alpha|protein phosphatase 2 (formerly 2A), regulatory subunit B (PR 52), alpha isoform', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=67014056, end_position_on_the_genomic_accession=67072471, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71979, CurrentID=0, Status='live', Symbol='Prr23a3', Aliases='2410012M07Rik', description='proline rich 23A, member 3', other_designations='proline-rich protein 23', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98864767, end_position_on_the_genomic_accession=98866580, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  71981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71981, CurrentID=0, Status='live', Symbol='Tdrd12', Aliases='2410004F06Rik, 2410070K17Rik, ECAT8, EG434165, G1-476-14, repro23', description='tudor domain containing 12', other_designations='ES cell associated transcript 8|ES cell-associated transcript 8 protein|tudor domain-containing protein 12', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35466941, end_position_on_the_genomic_accession=35538156, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  71982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71982, CurrentID=0, Status='live', Symbol='Snx10', Aliases='2410004M09Rik', description='sorting nexin 10', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=51523902, end_position_on_the_genomic_accession=51590679, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  71983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71983, CurrentID=0, Status='live', Symbol='Tmco6', Aliases='2410015B03Rik', description='transmembrane and coiled-coil domains 6', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36735018, end_position_on_the_genomic_accession=36742400, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  71984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71984, CurrentID=0, Status='live', Symbol='Sars2', Aliases='2410015F05Rik, D7Ertd353e, SerRS, SerRSmt', description='seryl-aminoacyl-tRNA synthetase 2', other_designations='seryl-tRNA synthetase, mitochondrial|seryl-tRNA(Ser/Sec) synthetase', map_location='7 16.91 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28741934, end_position_on_the_genomic_accession=28753879, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  71985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71985, CurrentID=0, Status='live', Symbol='Acad10', Aliases='2410021P16Rik', description='acyl-Coenzyme A dehydrogenase family, member 10', other_designations='ACAD-10', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121621026, end_position_on_the_genomic_accession=121660510, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  71986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71986, CurrentID=0, Status='live', Symbol='Ddx28', Aliases='2410004K13Rik, AI449652, Mddx28', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 28', other_designations='ATP-dependent RNA helicase Ddx28|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 28|mitochondrial DEAD box protein 28', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106009616, end_position_on_the_genomic_accession=106011486, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  71988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71988, CurrentID=0, Status='live', Symbol='Esco2', Aliases='2410004I17Rik, D030072L07Rik', description='establishment of sister chromatid cohesion N-acetyltransferase 2', other_designations='ECO1 homolog 2|establishment of cohesion 1 homolog 2', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65819027, end_position_on_the_genomic_accession=65833969, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  71989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71989, CurrentID=0, Status='live', Symbol='Rpusd4', Aliases='2410001E19Rik', description='RNA pseudouridylate synthase domain containing 4', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35267881, end_position_on_the_genomic_accession=35275957, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71990, CurrentID=0, Status='live', Symbol='Ddx54', Aliases='2410015A15Rik, AI414901, DP97', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 54', other_designations='DEAD box protein 54', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120613107, end_position_on_the_genomic_accession=120628592, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  71991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71991, CurrentID=0, Status='live', Symbol='Ercc8', Aliases='2410022P04Rik, 2810431L23Rik, 4631412O06Rik, B130065P18Rik, Ckn1, Csa', description='excision repaiross-complementing rodent repair deficiency, complementation group 8', other_designations='Cockayne syndrome 1 homolog|cockayne syndrome WD repeat protein CSA homolog', map_location='13 D', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=108158694, end_position_on_the_genomic_accession=108199284, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  71994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71994, CurrentID=0, Status='live', Symbol='Cnn3', Aliases='1600014M03Rik, C85854, Calpo3', description='calponin 3, acidic', other_designations='calponin, acidic isoform', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121426541, end_position_on_the_genomic_accession=121458205, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  71996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71996, CurrentID=0, Status='live', Symbol='Fthl17a', Aliases='1600014K23Rik, AI324104', description='ferritin, heavy polypeptide-like 17, member A', other_designations='novel protein similar to ferritin, heavy polypeptide 1 FTH1', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=85249677, end_position_on_the_genomic_accession=85270291, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  71997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71997, CurrentID=0, Status='live', Symbol='Smg9', Aliases='1500002O20Rik, AI429693, N28092', description='smg-9 homolog, nonsense mediated mRNA decay factor (C. elegans)', other_designations='protein smg-9 homolog', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24398910, end_position_on_the_genomic_accession=24422777, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  71998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71998, CurrentID=0, Status='live', Symbol='Slc25a35', Aliases='1810012H11Rik', description='solute carrier family 25, member 35', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68968108, end_position_on_the_genomic_accession=68972516, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  71999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=71999, CurrentID=0, Status='live', Symbol='Fbxo22', Aliases='0610033L19Rik, 1600016C16Rik', description='F-box protein 22', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55208935, end_position_on_the_genomic_accession=55224433, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72000, CurrentID=0, Status='live', Symbol='Lmntd2', Aliases='1600016N20Rik', description='lamin tail domain containing 2', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141209990, end_position_on_the_genomic_accession=141215195, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  72002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72002, CurrentID=0, Status='live', Symbol='Slc39a5', Aliases='1810013D05Rik, 2010205A06Rik, Zip5', description='solute carrier family 39 (metal ion transporter), member 5', other_designations='ZIP-5|solute carrier family 39 member 5|zrt- and Irt-like protein 5', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128395931, end_position_on_the_genomic_accession=128401276, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72003, CurrentID=0, Status='live', Symbol='Synpr', Aliases='1500003F20Rik, SPO', description='synaptoporin', other_designations='synaptorin', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=13284780, end_position_on_the_genomic_accession=13615469, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72007, CurrentID=0, Status='live', Symbol='Fndc3b', Aliases='1600019O04Rik, AW550168, Fad104, mKIAA4164', description='fibronectin type III domain containing 3B', other_designations='HCV NS5A-binding protein 37|factor for adipocyte differentiation 104|fibronectin domain-containing protein 3b', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=27416161, end_position_on_the_genomic_accession=27711662, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  72008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72008, CurrentID=0, Status='live', Symbol='Zfyve19', Aliases='1500041L05Rik, ANCHR', description='zinc finger, FYVE domain containing 19', other_designations='zinc finger FYVE domain-containing protein 19', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119208720, end_position_on_the_genomic_accession=119217050, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72012, CurrentID=0, Status='live', Symbol='1600020E01Rik', Aliases='', description='RIKEN cDNA 1600020E01 gene', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86527330, end_position_on_the_genomic_accession=86564449, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72014, CurrentID=0, Status='live', Symbol='Btbd17', Aliases='1500005I02Rik, Tango10a', description='BTB (POZ) domain containing 17', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114790669, end_position_on_the_genomic_accession=114802640, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72016, CurrentID=0, Status='live', Symbol='1600002H07Rik', Aliases='', description='RIKEN cDNA 1600002H07 gene', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24215054, end_position_on_the_genomic_accession=24220938, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72017, CurrentID=0, Status='live', Symbol='Cyb5r1', Aliases='1500005G05Rik, B5R.1, C80155, Nqo3a2', description='cytochrome b5 reductase 1', other_designations='NAD(P)H:quinone oxidoreductase type 3 polypeptide A2|NAD(P)H:quinone oxidoreductase type 3, polypeptide A2', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134405561, end_position_on_the_genomic_accession=134411740, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72018, CurrentID=0, Status='live', Symbol='Fundc1', Aliases='1500005J14Rik, 1810033P05Rik', description='FUN14 domain containing 1', other_designations='', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=17519967, end_position_on_the_genomic_accession=17572330, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72020, CurrentID=0, Status='live', Symbol='Zfp654', Aliases='1600021C16Rik, 1810008K20Rik, Znf654', description='zinc finger protein 654', other_designations='', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=64780347, end_position_on_the_genomic_accession=64851905, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  72022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72022, CurrentID=0, Status='live', Symbol='Slc35f2', Aliases='1500009K05Rik, AU019213', description='solute carrier family 35, member F2', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53771226, end_position_on_the_genomic_accession=53818161, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72023, CurrentID=0, Status='live', Symbol='Cyb561d1', Aliases='1600010M23Rik, AI605489, AW457162', description='cytochrome b-561 domain containing 1', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108195771, end_position_on_the_genomic_accession=108202226, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72026, CurrentID=0, Status='live', Symbol='Trmu', Aliases='1110005N20Rik, 1600025P05Rik, AI314320, Trmt1', description='tRNA 5-methylaminomethyl-2-thiouridylate methyltransferase', other_designations='tRNA (5-methylaminomethyl-2-thiouridylate)-methyltransferase 1', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85879309, end_position_on_the_genomic_accession=85897394, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72027, CurrentID=0, Status='live', Symbol='Slc39a4', Aliases='1600025H15Rik, AU041686, AWMS2, ZIP4', description='solute carrier family 39 (zinc transporter), member 4', other_designations='ZIP-4|activated in W/Wv mouse stomach 2|mAWMS2|solute carrier family 39 member 4|zrt- and Irt-like protein 4', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76612383, end_position_on_the_genomic_accession=76617379, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72029, CurrentID=0, Status='live', Symbol='Cnpy3', Aliases='1600025D17Rik, 2410050O22Rik, AI413153, CAG4A, ERDA5, PRAT4A, Tnrc5', description='canopy FGF signaling regulator 3', other_designations='canopy 3 homolog|protein associated with Tlr4|protein associated with toll-like receptor 4|trinucleotide repeat-containing gene 5 protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46735713, end_position_on_the_genomic_accession=46752293, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  72033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72033, CurrentID=0, Status='live', Symbol='Tsc22d2', Aliases='1810043J12Rik, 5530402M19Rik', description='TSC22 domain family, member 2', other_designations='TSC22 domain family 2', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=58415531, end_position_on_the_genomic_accession=58466787, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72039, CurrentID=0, Status='live', Symbol='Mccc1', Aliases='1810045E08Rik, 2310058B18Rik, MCCalpha, Mcca, R75106', description='methylcrotonoyl-Coenzyme A carboxylase 1 (alpha)', other_designations='3-methylcrotonyl-CoA carboxylase 1|3-methylcrotonyl-CoA carboxylase biotin-containing subunit|3-methylcrotonyl-CoA:carbon dioxide ligase subunit alpha|MCCase subunit alpha', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=35959296, end_position_on_the_genomic_accession=36000678, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  72040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72040, CurrentID=0, Status='live', Symbol='Cdhr5', Aliases='1810074H01Rik, AI481143, Mucdhl, Mupcdh', description='cadherin-related family member 5', other_designations='mu-protocadherin|mucin and cadherin-like protein|mucin-like protocadherin', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141269080, end_position_on_the_genomic_accession=141276794, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  72041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72041, CurrentID=0, Status='live', Symbol='Alkbh4', Aliases='2010004B12Rik, Abj4', description='alkB homolog 4, lysine demethylase', other_designations='alkB, alkylation repair homolog 4|alkylated DNA repair protein alkB homolog 4|probable alpha-ketoglutarate-dependent dioxygenase ABH4', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136136145, end_position_on_the_genomic_accession=136141802, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72042, CurrentID=0, Status='live', Symbol='Cotl1', Aliases='1810074P22Rik, 2010004C08Rik, Clp', description='coactosin-like 1 (Dictyostelium)', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119809214, end_position_on_the_genomic_accession=119840579, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72043, CurrentID=0, Status='live', Symbol='Sulf2', Aliases='2010004N24Rik, AU020235, MSulf-2, mKIAA1247', description='sulfatase 2', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=166073899, end_position_on_the_genomic_accession=166155683, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  72045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72045, CurrentID=0, Status='live', Symbol='Mfsd4b2', Aliases='2010001E11Rik, Naglt1b', description='major facilitator superfamily domain containing 4B2', other_designations='sodium-dependent glucose transporter 1', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39920382, end_position_on_the_genomic_accession=39926923, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72046, CurrentID=0, Status='live', Symbol='Urgcp', Aliases='2010005J08Rik, 9130001I21Rik, AW060359, Urg4, mKIAA1507', description='upregulator of cell proliferation', other_designations='HBV X protein up-regulated gene 4 protein homolog|HBxAg up-regulated gene 4 protein homolog|protein URG4', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5713417, end_position_on_the_genomic_accession=5762436, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72047, CurrentID=0, Status='live', Symbol='Ddx42', Aliases='1810047H21Rik, AW319508, AW556242, B430002H05Rik, RHELP, RNAHP, SF3b125', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 42', other_designations='DEAD box protein 42|RNA helicase-related protein', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106216463, end_position_on_the_genomic_accession=106249140, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  72049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72049, CurrentID=0, Status='live', Symbol='Tnfrsf13c', Aliases='2010006P15Rik, BAFF-R, Baffr, Bcmd, Bcmd-1, Bcmd1, Lvis22', description='tumor necrosis factor receptor superfamily, member 13c', other_designations='B-cell maturation defect 1|B-cell-activating factor receptor|BAFF receptor|BLyS receptor 3|b cell-activating factor receptor', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82221742, end_position_on_the_genomic_accession=82224747, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72050, CurrentID=0, Status='live', Symbol='Kdelc1', Aliases='1810049A15Rik, 5730416C13Rik, EP58, Kdel1', description='KDEL (Lys-Asp-Glu-Leu) containing 1', other_designations='ER protein 58|ERp58|KDEL containing protein 1|endoplasmic reticulum resident protein 58|endoplasmic reticulum-resident KDEL protein', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44106546, end_position_on_the_genomic_accession=44118857, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  72053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72053, CurrentID=0, Status='live', Symbol='Tmub2', Aliases='2010008E23Rik', description='transmembrane and ubiquitin-like domain containing 2', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102284820, end_position_on_the_genomic_accession=102289427, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72054, CurrentID=0, Status='live', Symbol='Cyp4f18', Aliases='1810054N16Rik, Cyp4f3, Cypf18', description='cytochrome P450, family 4, subfamily f, polypeptide 18', other_designations='CYPIVF3|cyp4f3|cytochrome P450 4F3|cytochrome P450 CYP4F18|cytochrome P450, 4f18|cytochrome P450-LTB-omega|leukotriene-B(4) 20-monooxygenase 2', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71988482, end_position_on_the_genomic_accession=72009685, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72055, CurrentID=0, Status='live', Symbol='Slc38a10', Aliases='1810073N04Rik', description='solute carrier family 38, member 10', other_designations='novel amino acid transporter protein|solute carrier family 38 member 10', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120103951, end_position_on_the_genomic_accession=120151351, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  72056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72056, CurrentID=0, Status='live', Symbol='1810055G02Rik', Aliases='AI256204', description='RIKEN cDNA 1810055G02 gene', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3708298, end_position_on_the_genomic_accession=3717882, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72057, CurrentID=0, Status='live', Symbol='Phf10', Aliases='1810055P05Rik, AV024533', description='PHD finger protein 10', other_designations='BAF45a|BRG1-associated factor 45a', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14944994, end_position_on_the_genomic_accession=14961638, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72058, CurrentID=0, Status='live', Symbol='Igsf5', Aliases='2010003D20Rik, JCAM, Jam4', description='immunoglobulin superfamily, member 5', other_designations='JAM-4|cell adhesion molecule JCAM|junction adhesion molecule 4|junctional adhesion molecule 4', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96360467, end_position_on_the_genomic_accession=96465572, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72061, CurrentID=0, Status='live', Symbol='2010111I01Rik', Aliases='2300006M17Rik, AP-O, Aopep, mir-23b, mir-24-1, mir-27b', description='RIKEN cDNA 2010111I01 gene', other_designations='', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=62964893, end_position_on_the_genomic_accession=63326096, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  72065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72065, CurrentID=0, Status='live', Symbol='Rap2c', Aliases='2010200P20Rik, AI194294, AL022976', description='RAP2C, member of RAS oncogene family', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=51003912, end_position_on_the_genomic_accession=51020092, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72068, CurrentID=0, Status='live', Symbol='Cnot2', Aliases='2600016M12Rik, 2810470K03Rik, AA537049, AA959607, AW557563, C79650', description='CCR4-NOT transcription complex, subunit 2', other_designations='CCR4-associated factor 2', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=116485161, end_position_on_the_genomic_accession=116581535, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  72074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72074, CurrentID=0, Status='live', Symbol='Anks4b', Aliases='2010013E14Rik, AI956809, Harp', description='ankyrin repeat and sterile alpha motif domain containing 4B', other_designations='harmonin-interacting ankyrin repeat-containing protein|harmonin-interacting ankyrin-repeat containing', map_location='7 64.45 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120173858, end_position_on_the_genomic_accession=120183716, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  72075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72075, CurrentID=0, Status='live', Symbol='Ogfr', Aliases='2010013E17Rik', description='opioid growth factor receptor', other_designations='zeta-type opioid receptor', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180589407, end_position_on_the_genomic_accession=180595837, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72077, CurrentID=0, Status='live', Symbol='Gcnt3', Aliases='2010013H22Rik, 2210021I22Rik, 2210401J11Rik', description='glucosaminyl (N-acetyl) transferase 3, mucin type', other_designations='C2GnT-M|C2GnT-mucin type|core 2 beta-1,6-N-acetylglucosaminyltransferase II|mucus-type core 2 beta-1,6-N-acetylglucosaminyltransferase', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70031496, end_position_on_the_genomic_accession=70039887, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12013, CurrentID=0, Status='live', Symbol='Bach1', Aliases='6230421P05Rik, AI323795', description='BTB and CNC homology 1, basic leucine zipper transcription factor 1', other_designations='BTB and CNC homolog 1|BTB and CNC homology 1 transcription factor', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87698954, end_position_on_the_genomic_accession=87733346, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72080, CurrentID=0, Status='live', Symbol='Sapcd2', Aliases='2010317E24Rik, 6030458L21Rik, AL033337, ang', description='suppressor APC domain containing 2', other_designations='protein Ang|protein C9orf140 homolog', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25372035, end_position_on_the_genomic_accession=25378213, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72082, CurrentID=0, Status='live', Symbol='Cyp2c55', Aliases='2010318C06Rik', description='cytochrome P450, family 2, subfamily c, polypeptide 55', other_designations='CYPIIC55|cyp2c55|cytochrome P450, 2c55', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39006938, end_position_on_the_genomic_accession=39042687, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72083, CurrentID=0, Status='live', Symbol='Mzt2', Aliases='2410018G20Rik, 2610001E06Rik, Fam128, Fam128b', description='mitotic spindle organizing protein 2', other_designations='family with sequence similarity 128, member B|mitotic-spindle organizing protein associated with a ring of gamma-tubulin 2|mitotic-spindle organizing protein associated with a ring of gamma-tubulin 2B', map_location='16 A2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=15848441, end_position_on_the_genomic_accession=15863364, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72084, CurrentID=0, Status='live', Symbol='Pigx', Aliases='2010319C14Rik, PIG-X', description='phosphatidylinositol glycan anchor biosynthesis, class X', other_designations='phosphatidylinositol glycan, class X', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32084416, end_position_on_the_genomic_accession=32099727, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72085, CurrentID=0, Status='live', Symbol='Osgepl1', Aliases='2610001M19Rik, AA416452', description='O-sialoglycoprotein endopeptidase-like 1', other_designations='N6-L-threonylcarbamoyladenine synthase|O-sialoglycoprotein endopeptidase-like protein 1|probable O-sialoglycoprotein endopeptidase 2|probable tRNA threonylcarbamoyladenosine biosynthesis protein Osgepl1|t(6)A synthase|t(6)A37 threonylcarbamoyladenosine biosynthesis protein Osgepl1|tRNA threonylcarbamoyladenosine biosynthesis protein Osgepl1', map_location='1 26.99 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53312897, end_position_on_the_genomic_accession=53326343, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72088, CurrentID=0, Status='live', Symbol='Ush1c', Aliases='2010016F01Rik, harmonin', description='Usher syndrome 1C', other_designations='PDZ domain-containing protein|Usher syndrome 1C homolog|usher syndrome type-1C protein homolog', map_location='7 29.66 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46195351, end_position_on_the_genomic_accession=46238490, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  72090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72090, CurrentID=0, Status='live', Symbol='Entpd8', Aliases='', description='ectonucleoside triphosphate diphosphohydrolase 8', other_designations='E-NTPDase 8|NTPDase 8|NTPDase8', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25080146, end_position_on_the_genomic_accession=25085719, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  72091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72091, CurrentID=0, Status='live', Symbol='Snhg7', Aliases='2610002F03Rik, AI427062', description='small nucleolar RNA host gene 7', other_designations='small nucleolar RNA host gene (non-protein coding) 7', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26637176, end_position_on_the_genomic_accession=26640244, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72094, CurrentID=0, Status='live', Symbol='Ugt2a3', Aliases='2010321J07Rik', description='UDP glucuronosyltransferase 2 family, polypeptide A3', other_designations='UDPGT 2A3', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87324972, end_position_on_the_genomic_accession=87337195, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72096, CurrentID=0, Status='live', Symbol='Mettl10', Aliases='2010208K18Rik', description='methyltransferase like 10', other_designations='methyltransferase-like protein 10', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132827232, end_position_on_the_genomic_accession=132852880, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  208624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208624, CurrentID=0, Status='live', Symbol='Alg3', Aliases='D16Ertd36e, not', description='asparagine-linked glycosylation 3 (alpha-1,3-mannosyltransferase)', other_designations='alpha-1,3-mannosyltransferase ALG3|asparagine-linked glycosylation 3 homolog (yeast, alpha-1,3-mannosyltransferase)|asparagine-linked glycosylation protein 3 homolog|dol-P-Man-dependent alpha(1-3)-mannosyltransferase|dolichyl-P-Man:Man(5)GlcNAc(2)-PP-dolichyl mannosyltransferase|dolichyl-phosphate-mannose--glycolipid alpha-mannosyltransferase|not56-like protein', map_location='16 12.48 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20605458, end_position_on_the_genomic_accession=20610753, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72098, CurrentID=0, Status='live', Symbol='Tmem68', Aliases='2010300G19Rik, AA408325, AW123402', description='transmembrane protein 68', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3549041, end_position_on_the_genomic_accession=3574848, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  72102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72102, CurrentID=0, Status='live', Symbol='Dusp11', Aliases='2010300F21Rik, AI481497', description='dual specificity phosphatase 11 (RNA/RNP complex 1-interacting)', other_designations='phosphatase that interacts with RNA/RNP complex 1', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85942268, end_position_on_the_genomic_accession=85962045, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  72103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72103, CurrentID=0, Status='live', Symbol='Aplf', Aliases='2010301N04Rik, AI452191', description='aprataxin and PNKP like factor', other_designations='apurinic-apyrimidinic endonuclease APLF', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87628429, end_position_on_the_genomic_accession=87672168, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72106, CurrentID=0, Status='live', Symbol='Jmjd8', Aliases='2610003J06Rik', description='jumonji domain containing 8', other_designations='jumonji domain-containing protein 8', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25829043, end_position_on_the_genomic_accession=25831843, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72107, CurrentID=0, Status='live', Symbol='Dscc1', Aliases='2010006I05Rik, 2600005O03Rik, Dcc1', description='DNA replication and sister chromatid cohesion 1', other_designations='defective in sister chromatid cohesion 1 homolog|defective in sister chromatid cohesion protein 1 homolog', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=55076093, end_position_on_the_genomic_accession=55092232, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72108, CurrentID=0, Status='live', Symbol='Ddhd2', Aliases='2010305K11Rik, SAMWD1, mKIAA0725', description='DDHD domain containing 2', other_designations='DDHD domain-containing protein 2|SAM, WWE and DDHD domain-containing protein 1', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25725324, end_position_on_the_genomic_accession=25754590, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  72112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72112, CurrentID=0, Status='live', Symbol='Ppp1r14d', Aliases='2010107K19Rik, GBPI-1, Ppp1r14', description='protein phosphatase 1, regulatory (inhibitor) subunit 14D', other_designations='PKC-dependent PP1 inhibitory protein subunit 14|gastrointestinal and brain-specific PP1-inhibitory protein 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119218119, end_position_on_the_genomic_accession=119230067, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  72113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72113, CurrentID=0, Status='live', Symbol='Adck1', Aliases='2610005A10Rik', description='aarF domain containing kinase 1', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=88360491, end_position_on_the_genomic_accession=88461726, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  72114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72114, CurrentID=0, Status='live', Symbol='Zbed3', Aliases='0610037K01Rik, 2610005H11Rik, AU018975, AU024588', description='zinc finger, BED type containing 3', other_designations='axin-interacting protein|zinc finger, BED domain containing 3', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95323605, end_position_on_the_genomic_accession=95337842, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72117, CurrentID=0, Status='live', Symbol='Naa50', Aliases='2600005K24Rik, 2810441M03Rik, AW112078, Mak3, Mak3p, Nat13, Nat5, San', description='N(alpha)-acetyltransferase 50, NatE catalytic subunit', other_designations='Mak3 homolog|Mak3p homolog|N-acetyltransferase 13|N-acetyltransferase NAT13|N-alpha-acetyltransferase 50, NatE catalytic subunit|Nalpha acetyltransferase 50|natE catalytic subunit', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44139809, end_position_on_the_genomic_accession=44163366, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72119, CurrentID=0, Status='live', Symbol='Tpx2', Aliases='2610005B21Rik, DIL2, REPP86, p100', description='TPX2, microtubule-associated protein homolog (Xenopus laevis)', other_designations='', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152847964, end_position_on_the_genomic_accession=152895321, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  72121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72121, CurrentID=0, Status='live', Symbol='Dennd2d', Aliases='2010308M01Rik, AI451421', description='DENN/MADD domain containing 2D', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106482431, end_position_on_the_genomic_accession=106503030, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  72123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72123, CurrentID=0, Status='live', Symbol='Ccdc71l', Aliases='2010109K11Rik, AV375874', description='coiled-coil domain containing 71 like', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=32378789, end_position_on_the_genomic_accession=32382943, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  72124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72124, CurrentID=0, Status='live', Symbol='Seh1l', Aliases='2610007A16Rik, AW540070, SEC13L, SEH1A, SEH1B, Seh1', description='SEH1-like (S. cerevisiae', other_designations='nup107-160 subcomplex subunit SEH1|sec13-like protein', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67774798, end_position_on_the_genomic_accession=67795487, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72125, CurrentID=0, Status='live', Symbol='Amer2', Aliases='2600011E07Rik, Fam123a', description='APC membrane recruitment 2', other_designations='family with sequence similarity 123, member A|protein FAM123A', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=60373293, end_position_on_the_genomic_accession=60388197, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72128, CurrentID=0, Status='live', Symbol='2610008E11Rik', Aliases='', description='RIKEN cDNA 2610008E11 gene', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79064374, end_position_on_the_genomic_accession=79097625, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72129, CurrentID=0, Status='live', Symbol='Pex13', Aliases='2610008O20Rik', description='peroxisomal biogenesis factor 13', other_designations='peroxin-13', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23646479, end_position_on_the_genomic_accession=23665935, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72133, CurrentID=0, Status='live', Symbol='Trub1', Aliases='2610009I02Rik, 9030425C13Rik, AI448235, BB118943, C76870', description='TruB pseudouridine (psi) synthase homolog 1 (E. coli)', other_designations='', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=57452891, end_position_on_the_genomic_accession=57491005, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72135, CurrentID=0, Status='live', Symbol='Pygo1', Aliases='2600014C22Rik', description='pygopus 1', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72925650, end_position_on_the_genomic_accession=72946015, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72136, CurrentID=0, Status='live', Symbol='Chst14', Aliases='2600016L03Rik, D4ST-1, D4st1', description='carbohydrate (N-acetylgalactosamine 4-0) sulfotransferase 14', other_designations='dermatan 4 sulfotransferase 1|dermatan 4-sulfotransferase 1|dermatan-4-sulfotransferase-1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118926497, end_position_on_the_genomic_accession=118928585, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  72137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72137, CurrentID=0, Status='live', Symbol='Wdsub1', Aliases='1700048E19Rik, 2610014F08Rik', description='WD repeat, SAM and U-box domain containing 1', other_designations='', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=59852364, end_position_on_the_genomic_accession=59882625, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72140, CurrentID=0, Status='live', Symbol='Cep89', Aliases='2610507L03Rik, Ccdc123', description='centrosomal protein 89', other_designations='coiled-coil domain containing 123|coiled-coil domain-containing protein 123, mitochondrial', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35397000, end_position_on_the_genomic_accession=35439828, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  72141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72141, CurrentID=0, Status='live', Symbol='Adpgk', Aliases='2610017G09Rik, Adp-gk, C77577', description='ADP-dependent glucokinase', other_designations='ATP-dependent glucokinase', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59291545, end_position_on_the_genomic_accession=59316200, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72144, CurrentID=0, Status='live', Symbol='Slc37a3', Aliases='2610507O21Rik, AU044904', description='solute carrier family 37 (glycerol-3-phosphate transporter), member 3', other_designations='solute carrier family 37 member 3', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39334770, end_position_on_the_genomic_accession=39377725, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  72145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72145, CurrentID=0, Status='live', Symbol='Wdfy3', Aliases='2610509D04Rik, ALFY, AW319683, B930017C24, BWF1, Bchs, D5Ertd66e, Ggtb3, ZFYVE25, mKIAA0993', description='WD repeat and FYVE domain containing 3', other_designations='beach domain, WD repeat and FYVE domain-containing protein 1|galactosyltransferase 3 beta 1, 4', map_location='5 48.95 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=101832953, end_position_on_the_genomic_accession=102072124, orientation='minus', exon_count=70, OMIM='', no_hdr0=''),
  72147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72147, CurrentID=0, Status='live', Symbol='Zbtb46', Aliases='2610019F01Rik, 4933406L05Rik, BZEL, Btbd4', description='zinc finger and BTB domain containing 46', other_designations='BTB (POZ) domain containing 4|BTB-ZF protein expressed in effector lymphocytes|BTB/POZ domain-containing protein 4', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181371261, end_position_on_the_genomic_accession=181492982, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72148, CurrentID=0, Status='live', Symbol='Tdrp', Aliases='2610019F03Rik', description='testis development related protein', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13952008, end_position_on_the_genomic_accession=13975372, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72149, CurrentID=0, Status='live', Symbol='Strada', Aliases='2610019A05Rik, 6030402H20Rik, AI480680, E130112C09Rik', description='STE20-related kinase adaptor alpha', other_designations='STE20-related adapter protein|STRAD alpha|protein kinase LYK5', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106162974, end_position_on_the_genomic_accession=106193604, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72151, CurrentID=0, Status='live', Symbol='Rfc5', Aliases='2610020K06Rik, 2610209F07Rik, 36.5kDa, 36kDa, Recc5', description='replication factor C (activator 1) 5', other_designations='A1 36 kDa subunit|RF-C 36 kDa subunit|RFC36|activator 1 36 kDa subunit|activator 1 subunit 5|replication factor C 36 kDa subunit|replication factor C 5|replication factor C, 36 kDa', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117379145, end_position_on_the_genomic_accession=117389023, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72154, CurrentID=0, Status='live', Symbol='Zfp157', Aliases='2610020C11Rik, A630094N24Rik, AI327407, mszf12, mszf23-1', description='zinc finger protein 157', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138441467, end_position_on_the_genomic_accession=138460694, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72155, CurrentID=0, Status='live', Symbol='Cenpn', Aliases='2610510J17Rik, AI426416, AW545024', description='centromere protein N', other_designations='CENP-N', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116921601, end_position_on_the_genomic_accession=116941504, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72157, CurrentID=0, Status='live', Symbol='Pgm2', Aliases='2610020G18Rik, AA407108, AI098105, Pgm-2, Pgm1', description='phosphoglucomutase 2', other_designations='glucose phosphomutase 1|phosphoglucomutase 1|phosphoglucomutase-1', map_location='4 45.71 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=99929451, end_position_on_the_genomic_accession=99987294, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  72160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72160, CurrentID=0, Status='live', Symbol='Tmem163', Aliases='2610024A01Rik, SV31', description='transmembrane protein 163', other_designations='synaptic vesicle membrane protein of 31 kDa', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=127490342, end_position_on_the_genomic_accession=127679619, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  72162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72162, CurrentID=0, Status='live', Symbol='Dhx36', Aliases='2810407E23Rik, AI452301, AU022184, Ddx36, mKIAA1488', description='DEAH (Asp-Glu-Ala-His) box polypeptide 36', other_designations='DEAH box protein 36|MLE-like protein 1|RNA helicase associated with AU-rich element ARE|probable ATP-dependent RNA helicase DHX36', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=62468015, end_position_on_the_genomic_accession=62507032, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  72167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72167, CurrentID=0, Status='live', Symbol='Thumpd2', Aliases='2810025A12Rik', description='THUMP domain containing 2', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=81026327, end_position_on_the_genomic_accession=81065127, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  72168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72168, CurrentID=0, Status='live', Symbol='Aifm3', Aliases='2810401C16Rik, AI840249, Aifl', description='apoptosis-inducing factor, mitochondrion-associated 3', other_designations='apoptosis-inducing factor like|apoptosis-inducing factor-like protein', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17489616, end_position_on_the_genomic_accession=17507485, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  72169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72169, CurrentID=0, Status='live', Symbol='Trim29', Aliases='1110047J21Rik, 2810431N19Rik, 4732461M22Rik, AI119726', description='tripartite motif-containing 29', other_designations='tripartite motif protein 29', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=43310715, end_position_on_the_genomic_accession=43336125, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72170, CurrentID=0, Status='live', Symbol='Chchd4', Aliases='2410012P20Rik, 2810014D17Rik, AI838740', description='coiled-coil-helix-coiled-coil-helix domain containing 4', other_designations='coiled-coil-helix-coiled-coil-helix domain-containing protein 4', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91464276, end_position_on_the_genomic_accession=91473423, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72171, CurrentID=0, Status='live', Symbol='Shq1', Aliases='2810403P18Rik, Grim-1', description='SHQ1 homolog (S. cerevisiae)', other_designations='', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=100572544, end_position_on_the_genomic_accession=100671175, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72175, CurrentID=0, Status='live', Symbol='Mfsd8', Aliases='2810423E13Rik, AI836898, AV142426, Cln7', description='major facilitator superfamily domain containing 8', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40818112, end_position_on_the_genomic_accession=40846856, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72179, CurrentID=0, Status='live', Symbol='Fbxl2', Aliases='2810423A21Rik, Fbl3', description='F-box and leucine-rich repeat protein 2', other_designations='', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=113958757, end_position_on_the_genomic_accession=114026787, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  72180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72180, CurrentID=0, Status='live', Symbol='Zfp661', Aliases='2810405K07Rik, A1-5, Znf2', description='zinc finger protein 661', other_designations='zfp-661|zinc finger protein 2 (A1-5)', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127574617, end_position_on_the_genomic_accession=127587135, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72181, CurrentID=0, Status='live', Symbol='Nsun4', Aliases='2310010O12Rik, 2810405F18Rik, AI507900, Shtap', description='NOL1/NOP2/Sun domain family, member 4', other_designations='sperm head and tail associated protein', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116031770, end_position_on_the_genomic_accession=116053881, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72183, CurrentID=0, Status='live', Symbol='Snx6', Aliases='2010006G21Rik, 2610032J07Rik, 2810425K19Rik, AU018928, C85963', description='sorting nexin 6', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=54746353, end_position_on_the_genomic_accession=54795749, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72184, CurrentID=0, Status='live', Symbol='Klhl35', Aliases='2810406K13Rik', description='kelch-like 35', other_designations='', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99464216, end_position_on_the_genomic_accession=99474022, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  230484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230484, CurrentID=0, Status='live', Symbol='Usp1', Aliases='', description='ubiquitin specific peptidase 1', other_designations='deubiquitinating enzyme 1|ubiquitin specific protease 1|ubiquitin thioesterase 1|ubiquitin thiolesterase 1|ubiquitin-specific-processing protease 1|usp1', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=98923622, end_position_on_the_genomic_accession=98935543, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72190, CurrentID=0, Status='live', Symbol='2510009E07Rik', Aliases='AI194952', description='RIKEN cDNA 2510009E07 gene', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21649042, end_position_on_the_genomic_accession=21694665, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72193, CurrentID=0, Status='live', Symbol='Scaf11', Aliases='1110061H03Rik, 2610510E10Rik, AI462454, CASP11, SIP1, SRRP129, Sfrs2ip, Srsf2ip, mKIAA3013', description='SR-related CTD-associated factor 11', other_designations='serine/arginine-rich splicing factor 2, interacting protein|splicing factor, arginine/serine-rich 2, interacting protein', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=96411698, end_position_on_the_genomic_accession=96460843, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  72194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72194, CurrentID=0, Status='live', Symbol='Fbxl20', Aliases='2610511F20Rik, 4632423N09Rik, AI849362, AL117906, C86145, Fbl2, Scr, mKIAA4147', description='F-box and leucine-rich repeat protein 20', other_designations='scrapper', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98082554, end_position_on_the_genomic_accession=98150440, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  72195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72195, CurrentID=0, Status='live', Symbol='Supt7l', Aliases='2610524B01Rik, 6030455L14Rik', description='suppressor of Ty 7-like', other_designations='SPTF-associated factor 65 gamma|STAF65gamma', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31514569, end_position_on_the_genomic_accession=31526762, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72198, CurrentID=0, Status='live', Symbol='Skiv2l2', Aliases='2610528A15Rik, mKIAA0052', description='superkiller viralicidic activity 2-like 2 (S. cerevisiae)', other_designations='ATP-dependent helicase SKIV2L2|TRAMP-like complex helicase', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112867780, end_position_on_the_genomic_accession=112927380, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  72199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72199, CurrentID=0, Status='live', Symbol='Mms19', Aliases='2410001K24Rik, 2610042O15Rik, AI316855, C79368, C86341l, Mms19', description='MMS19 (MET18 S. cerevisiae)', other_designations='MET18 homolog|MMS19-like protein', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41943707, end_position_on_the_genomic_accession=41981159, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  72201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72201, CurrentID=0, Status='live', Symbol='Otud6b', Aliases='2600013N14Rik, AU015433', description='OTU domain containing 6B', other_designations='ortholog of human CGI-77', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=14809505, end_position_on_the_genomic_accession=14826618, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72205, CurrentID=0, Status='live', Symbol='Eml2', Aliases='1600029N02Rik', description='echinoderm microtubule associated protein like 2', other_designations='EMAP-2', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19176266, end_position_on_the_genomic_accession=19206482, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  72215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72215, CurrentID=0, Status='live', Symbol='1700001P01Rik', Aliases='', description='RIKEN cDNA 1700001P01 gene', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97771481, end_position_on_the_genomic_accession=97775918, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72219, CurrentID=0, Status='live', Symbol='Spata31d1a', Aliases='1700013B16Rik, Fam75d1a, Fam75d3', description='spermatogenesis associated 31 subfamily D, member 1A', other_designations='family with sequence similarity 75, member D1A|family with sequence similarity 75, member D3', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59700083, end_position_on_the_genomic_accession=59706589, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72221, CurrentID=0, Status='live', Symbol='1700021F07Rik', Aliases='', description='RIKEN cDNA 1700021F07 gene', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=173522592, end_position_on_the_genomic_accession=173528502, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72230, CurrentID=0, Status='live', Symbol='Zfp558', Aliases='1700007A21Rik, 4932704I17Rik-ps, Zfp558', description='zinc finger protein 558', other_designations='zinc finger protein 558, pseudogene', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18453558, end_position_on_the_genomic_accession=18478301, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  72236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72236, CurrentID=0, Status='live', Symbol='Tsnaxip1', Aliases='1700016K08Rik, TXI1', description='translin-associated factor X (Tsnax) interacting protein 1', other_designations='trax-interacting protein 1|tsnaxip1', map_location='8 D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105827711, end_position_on_the_genomic_accession=105844676, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  72238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72238, CurrentID=0, Status='live', Symbol='Tbc1d5', Aliases='1600014N05Rik', description='TBC1 domain family, member 5', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=50733124, end_position_on_the_genomic_accession=51180198, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  72242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72242, CurrentID=0, Status='live', Symbol='Psg21', Aliases='1600019C01Rik, 1600025N01Rik, 1600026N13Rik, cea8', description='pregnancy-specific glycoprotein 21', other_designations='pregnancy-specific glrcoprotein 21', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18646654, end_position_on_the_genomic_accession=18656725, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72244, CurrentID=0, Status='live', Symbol='1600014C10Rik', Aliases='AI428873', description='RIKEN cDNA 1600014C10 gene', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=38183217, end_position_on_the_genomic_accession=38197568, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  208650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208650, CurrentID=0, Status='live', Symbol='Cblb', Aliases='AI429560, AI851073, Cbl-b', description='Casitas B-lineage lymphoma b', other_designations='SH3-binding protein CBL-B|casitas B-lineage lymphoma proto-oncogene b|signal transduction protein CBL-B', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=52031230, end_position_on_the_genomic_accession=52208047, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  72254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72254, CurrentID=0, Status='live', Symbol='1700030K09Rik', Aliases='', description='RIKEN cDNA 1700030K09 gene', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72443880, end_position_on_the_genomic_accession=72460541, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72258, CurrentID=0, Status='live', Symbol='Kcnk10', Aliases='1700024D23Rik, 3010005K24Rik, Trek2', description='potassium channel, subfamily K, member 10', other_designations='TREK-2 two-pore-domain K+ channel|outward rectifying potassium channel TREK2', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98429432, end_position_on_the_genomic_accession=98577940, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72265, CurrentID=0, Status='live', Symbol='Tram1', Aliases='1810049E02Rik, TRAMP', description='translocating chain-associating membrane protein 1', other_designations='', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=13564693, end_position_on_the_genomic_accession=13589910, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72267, CurrentID=0, Status='live', Symbol='Lrrc8e', Aliases='1810049O03Rik, C87354', description='leucine rich repeat containing 8 family, member E', other_designations='leucine-rich repeat-containing protein 8E', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4224280, end_position_on_the_genomic_accession=4237470, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72269, CurrentID=0, Status='live', Symbol='Cda', Aliases='2210401N16Rik', description='cytidine deaminase', other_designations='cytidine aminohydrolase', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138338528, end_position_on_the_genomic_accession=138368312, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72273, CurrentID=0, Status='live', Symbol='Smim24', Aliases='2210404O07Rik', description='small integral membrane protein 24', other_designations='transmembrane protein C19orf77 homolog|transmembrane protein HSPC323 homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81393064, end_position_on_the_genomic_accession=81395079, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72275, CurrentID=0, Status='live', Symbol='2200002D01Rik', Aliases='2010005J02Rik, H2RSP, IMUP, IMUP-1, IMUP-2', description='RIKEN cDNA 2200002D01 gene', other_designations='HAI-2 related small protein', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29247521, end_position_on_the_genomic_accession=29248467, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72278, CurrentID=0, Status='live', Symbol='Ccpg1', Aliases='1700030B06Rik, 1810073J13Rik, 9430028F23Rik, AI426686, AI875170, CPR8, D9Ertd392e', description='cell cycle progression 1', other_designations='cell cycle progression 8 protein', map_location='9 40.08 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72985427, end_position_on_the_genomic_accession=73016340, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72281, CurrentID=0, Status='live', Symbol='Sh2d4a', Aliases='2210402M20Rik, SH2A', description='SH2 domain containing 4A', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=68276528, end_position_on_the_genomic_accession=68347704, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72284, CurrentID=0, Status='live', Symbol='Oraov1', Aliases='2210010N10Rik, TAOS1', description='oral cancer overexpressed 1', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144915097, end_position_on_the_genomic_accession=144930979, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72287, CurrentID=0, Status='live', Symbol='Plekhf1', Aliases='1810013P09Rik, APPD, LAPF, PHAFIN1, ZFYVE15', description='pleckstrin homology domain containing, family F (with FYVE domain) member 1', other_designations='PH domain-containing family F member 1|lysosome-associated apoptosis-inducing protein containing PH and FYVE domains', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=38220654, end_position_on_the_genomic_accession=38227994, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  72289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72289, CurrentID=0, Status='live', Symbol='Malat1', Aliases='2210401K01Rik, 9430072K23Rik, AI647968, Neat2', description='metastasis associated lung adenocarcinoma transcript 1 (non-coding RNA)', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5795690, end_position_on_the_genomic_accession=5802671, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  72290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72290, CurrentID=0, Status='live', Symbol='Lsm11', Aliases='2210404M20Rik, AV079530', description='U7 snRNP-specific Sm-like protein LSM11', other_designations='', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=45928265, end_position_on_the_genomic_accession=45944998, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72293, CurrentID=0, Status='live', Symbol='Nkd2', Aliases='2210403L10Rik, AW212591', description='naked cuticle 2 homolog (Drosophila)', other_designations='mNkd2|naked cuticle 1 homolog|naked-2', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73818529, end_position_on_the_genomic_accession=73847631, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  72296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72296, CurrentID=0, Status='live', Symbol='Rusc1', Aliases='2210403N08Rik, AA408288, NESCA', description='RUN and SH3 domain containing 1', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89083975, end_position_on_the_genomic_accession=89093363, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72297, CurrentID=0, Status='live', Symbol='B3gnt3', Aliases='2210008L19Rik', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 3', other_designations='BGnT-3|beta-1,3-Gn-T3|beta-1,3-N-acetylglucosaminyltransferase 3|beta3Gn-T3|core 1 beta-1,3-N-acetylglucosaminyltransferase|core 1 extending beta-1,3-N-acetylglucosaminyltransferase|core1-beta3GlcNAcT', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71691373, end_position_on_the_genomic_accession=71702627, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72301, CurrentID=0, Status='live', Symbol='1810041L15Rik', Aliases='Kiaa1644', description='RIKEN cDNA 1810041L15 gene', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84379203, end_position_on_the_genomic_accession=84482844, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72303, CurrentID=0, Status='live', Symbol='Cyp2c65', Aliases='2210009K14Rik', description='cytochrome P450, family 2, subfamily c, polypeptide 65', other_designations='cytochrome P450, 2c65', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=39061006, end_position_on_the_genomic_accession=39093948, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72306, CurrentID=0, Status='live', Symbol='Zfp777', Aliases='2500002G23Rik', description='zinc finger protein 777', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48017660, end_position_on_the_genomic_accession=48050236, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72307, CurrentID=0, Status='live', Symbol='2510002D24Rik', Aliases='2410043I01Rik, AA163535, AA409624, DXMgi27', description='RIKEN cDNA 2510002D24 gene', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18836580, end_position_on_the_genomic_accession=18840113, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72308, CurrentID=0, Status='live', Symbol='Brf1', Aliases='2510002F24Rik, GTF3B, TAF3C, TAFIII90, TFIIIB90, mTFIIIB90', description='BRF1, RNA polymerase III transcription initiation factor 90 kDa subunit', other_designations='B-related factor 1|BRF-1|BRF1 homolog, subunit of RNA polymerase III transcription initiation factor IIIB|hTFIIIB90', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112955631, end_position_on_the_genomic_accession=113000707, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  72309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72309, CurrentID=0, Status='live', Symbol='Tmem158', Aliases='2310037P21Rik, Ris1', description='transmembrane protein 158', other_designations='40 kDa BINP-binding protein|BINP binding protein p40BBP|Ras-induced senescence 1|brain injury-derived neurotrophic peptide (BINP) binding protein|p40BBP|ras-induced senescence protein 1', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123259057, end_position_on_the_genomic_accession=123260789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  72310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72310, CurrentID=0, Status='live', Symbol='Nkg7', Aliases='2500004F03Rik', description='natural killer cell group 7 sequence', other_designations='natural killer cell protein 7', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43437138, end_position_on_the_genomic_accession=43438246, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72313, CurrentID=0, Status='live', Symbol='Fryl', Aliases='2010313D22Rik, 2310004H21Rik, 2510002A14Rik, 9030227G01Rik, mKIAA0826', description='FRY like transcription coactivator', other_designations='furry homolog-like', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73020191, end_position_on_the_genomic_accession=73256645, orientation='minus', exon_count=70, OMIM='', no_hdr0=''),
  72315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72315, CurrentID=0, Status='live', Symbol='Ccdc74a', Aliases='2310015A05Rik', description='coiled-coil domain containing 74A', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17645903, end_position_on_the_genomic_accession=17650739, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72318, CurrentID=0, Status='live', Symbol='Cyth4', Aliases='2510004M07Rik, 5830469K17Rik, AI467541, Pscd4, mFLJ00017', description='cytohesin 4', other_designations='PH, SEC7 and coiled-coil domain-containing protein 4|pleckstrin homology, Sec7 and coiled/coil domains 4', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78597047, end_position_on_the_genomic_accession=78622019, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72320, CurrentID=0, Status='live', Symbol='Kif1bp', Aliases='0710007C18Rik, 2510003E04Rik, KBP, mKIAA1279', description='KIF1 binding protein', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62558470, end_position_on_the_genomic_accession=62578464, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72322, CurrentID=0, Status='live', Symbol='Xpo5', Aliases='2410004H11Rik, 2700038C24Rik, AI648907, AW549301, Exp5, RanBp21, mKIAA1291', description='exportin 5', other_designations='ran-binding protein 21', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46202855, end_position_on_the_genomic_accession=46242299, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  72323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72323, CurrentID=0, Status='live', Symbol='Asb6', Aliases='2510004M11Rik, AA409356', description='ankyrin repeat and SOCS box-containing 6', other_designations='ASB-6|ankyrin repeat and SOCS box-containing protein 6', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30823097, end_position_on_the_genomic_accession=30828300, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72324, CurrentID=0, Status='live', Symbol='Plxdc1', Aliases='2410003I07Rik, AI848450, Tem7', description='plexin domain containing 1', other_designations='TEM7 related', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97923237, end_position_on_the_genomic_accession=97986667, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  72325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72325, CurrentID=0, Status='live', Symbol='Vps9d1', Aliases='1300018I17Rik, 2410004N05Rik, 5DOS1, Vsp9d1', description='VPS9 domain containing 1', other_designations='5-day ovary-specific transcript 1 protein|VSP9 domain containing 1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123242356, end_position_on_the_genomic_accession=123254217, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  72330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72330, CurrentID=0, Status='live', Symbol='Klhl40', Aliases='2310024D23Rik, Kbtbd5', description='kelch-like 40', other_designations='kelch repeat and BTB (POZ) domain containing 5|kelch repeat and BTB domain-containing protein 5', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121777101, end_position_on_the_genomic_accession=121791002, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72333, CurrentID=0, Status='live', Symbol='Palld', Aliases='2410003B16Rik, 6030492A02', description='palladin, cytoskeletal associated protein', other_designations='90kDa actin-associated protein palladin|palladin isoform', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=61511429, end_position_on_the_genomic_accession=61902700, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  72338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72338, CurrentID=0, Status='live', Symbol='Wdr89', Aliases='2600001A11Rik', description='WD repeat domain 89', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=75630594, end_position_on_the_genomic_accession=75669537, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  72341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72341, CurrentID=0, Status='live', Symbol='Elp6', Aliases='2610001P13Rik, 2610002I17Rik, AV159994, Tmem103', description='elongator acetyltransferase complex subunit 6', other_designations='UPF0405 protein C3orf75 homolog|elongator complex protein 6 homolog|transmembrane protein 103', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110305192, end_position_on_the_genomic_accession=110322102, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  208665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208665, CurrentID=0, Status='live', Symbol='Akr1d1', Aliases='', description='aldo-keto reductase family 1, member D1', other_designations='delta(4)-3-ketosteroid 5-beta-reductase|delta(4)-3-oxosteroid 5-beta-reductase', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=37529978, end_position_on_the_genomic_accession=37568815, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72344, CurrentID=0, Status='live', Symbol='Usp36', Aliases='2700002L06Rik, mKIAA1453', description='ubiquitin specific peptidase 36', other_designations='deubiquitinating enzyme 36|ubiquitin specific protease 36|ubiquitin thioesterase 36|ubiquitin thiolesterase 36|ubiquitin-specific-processing protease 36', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118259651, end_position_on_the_genomic_accession=118290244, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  72345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72345, CurrentID=0, Status='live', Symbol='Amer1', Aliases='2810002O09Rik, AW492303, Fam123b, Wtx', description='APC membrane recruitment 1', other_designations='family with sequence similarity 123, member B|protein Fam123b', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=95420313, end_position_on_the_genomic_accession=95444840, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  72349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72349, CurrentID=0, Status='live', Symbol='Dusp3', Aliases='2210015O03Rik, 5031436O03Rik, T-DSP11, VHR', description='dual specificity phosphatase 3 (vaccinia virus phosphatase VH1-related)', other_designations='vaccinia H1-related phosphatase', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101971141, end_position_on_the_genomic_accession=101987005, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72350, CurrentID=0, Status='live', Symbol='Zc2hc1c', Aliases='2810002I04Rik, AV046379, Fam164c', description='zinc finger, C2HC-type containing 1C', other_designations='family with sequence similarity 164, member C|protein FAM164C', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85274388, end_position_on_the_genomic_accession=85299361, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72351, CurrentID=0, Status='live', Symbol='Ptar1', Aliases='1700084D09Rik, 2010013L17Rik, 4930428J16Rik, AI449677', description='protein prenyltransferase alpha subunit repeat containing 1', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=23686898, end_position_on_the_genomic_accession=23723092, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72354, CurrentID=0, Status='live', Symbol='Ttc4', Aliases='2810002P21Rik, AF007005, L62', description='tetratricopeptide repeat domain 4', other_designations='TPR repeat protein 4', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106655014, end_position_on_the_genomic_accession=106678967, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72355, CurrentID=0, Status='live', Symbol='Cdpf1', Aliases='2210021J22Rik', description='cysteine rich, DPF motif domain containing 1', other_designations='UPF0595 protein C22orf40 homolog', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85806972, end_position_on_the_genomic_accession=85811697, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72357, CurrentID=0, Status='live', Symbol='2210016L21Rik', Aliases='', description='RIKEN cDNA 2210016L21 gene', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114942163, end_position_on_the_genomic_accession=114948544, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72361, CurrentID=0, Status='live', Symbol='Ces2g', Aliases='2210023G05Rik, Ces2l', description='carboxylesterase 2G', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104961611, end_position_on_the_genomic_accession=104969537, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72368, CurrentID=0, Status='live', Symbol='Borcs8', Aliases='2310045N01Rik, Mef2bnb', description='BLOC-1 related complex subunit 8', other_designations='MEF2B neighbor gene protein homolog', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70139713, end_position_on_the_genomic_accession=70147261, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72371, CurrentID=0, Status='live', Symbol='2210408I21Rik', Aliases='', description='RIKEN cDNA 2210408I21 gene', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=77135536, end_position_on_the_genomic_accession=77613679, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  72373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72373, CurrentID=0, Status='live', Symbol='Psca', Aliases='2210408B04Rik', description='prostate stem cell antigen', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74714839, end_position_on_the_genomic_accession=74717065, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72383, CurrentID=0, Status='live', Symbol='Cnfn', Aliases='2210418J09Rik', description='cornifelin', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25367616, end_position_on_the_genomic_accession=25369724, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72388, CurrentID=0, Status='live', Symbol='Ripk4', Aliases='2310069J12Rik, AI552420, ANKK2, Ankrd3, DIk, PKK, RIP4', description='receptor-interacting serine-threonine kinase 4', other_designations='PKC-associated protein kinase|PKC-regulated protein kinase|ankyrin repeat domain 3|ankyrin repeat domain-containing protein 3', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97741933, end_position_on_the_genomic_accession=97763755, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72391, CurrentID=0, Status='live', Symbol='Cdkn3', Aliases='2410006H10Rik, KAP', description='cyclin-dependent kinase inhibitor 3', other_designations='CDK2-associated dual-specificity phosphatase|cdkn3|kinase-associated phosphatase', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=46760464, end_position_on_the_genomic_accession=46771529, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72392, CurrentID=0, Status='live', Symbol='Tmem175', Aliases='3010001K23Rik, AI504381, mTMEM175', description='transmembrane protein 175', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108629801, end_position_on_the_genomic_accession=108647770, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72393, CurrentID=0, Status='live', Symbol='Faim2', Aliases='2900002L20Rik, Lfg, Lfg2, NMP25, Nmp35, Tmbim2', description='Fas apoptotic inhibitory molecule 2', other_designations='neural membrane protein 35', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99497005, end_position_on_the_genomic_accession=99534878, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  72397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72397, CurrentID=0, Status='live', Symbol='Rbm12b1', Aliases='3000004N20Rik, Rbm12b, Rbm12ba', description='RNA binding motif protein 12 B1', other_designations='RNA binding motif protein 12B|RNA-binding motif protein 12B-A', map_location='', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=12138957, end_position_on_the_genomic_accession=12146745, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72399, CurrentID=0, Status='live', Symbol='Brap', Aliases='3010002G07Rik, BRAP2, IMP', description='BRCA1 associated protein', other_designations='brap|impedes mitogenic signal propagation', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121660563, end_position_on_the_genomic_accession=121687249, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72400, CurrentID=0, Status='live', Symbol='Pinx1', Aliases='2210403I16Rik, 2610028A01Rik, 67-11-3, AU024023, LPTS, LPTS1', description='PIN2/TERF1 interacting, telomerase inhibitor 1', other_designations='PIN2-interacting protein X1|PIN2/TRF1-interacting protein|TRF1-interacting protein 1|liver-related putative tumor suppressor', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63860301, end_position_on_the_genomic_accession=63919862, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72401, CurrentID=0, Status='live', Symbol='Slc43a1', Aliases='2610016F07Rik, AA986141, Lat3, PB39, Pov1, R00504', description='solute carrier family 43, member 1', other_designations='L-type amino acid transporter 3|prostate cancer overexpressed gene 1|slc43a1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84838888, end_position_on_the_genomic_accession=84863586, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  72404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72404, CurrentID=0, Status='live', Symbol='Wdr44', Aliases='2610034K17Rik, RAB11BP, RPH11', description='WD repeat domain 44', other_designations='rabphilin-11', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=23692926, end_position_on_the_genomic_accession=23806026, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  72413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72413, CurrentID=0, Status='live', Symbol='Kcnmb2', Aliases='2700049B16Rik, 3110031N04Rik', description='potassium large conductance calcium-activated channel, subfamily M, beta member 2', other_designations='BK channel subunit beta-2|BKbeta2|calcium-activated potassium channel, subfamily M subunit beta-2|charybdotoxin receptor subunit beta-2|k(VCA)beta-2|maxi K channel subunit beta-2|slo-beta-2', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=31902132, end_position_on_the_genomic_accession=32200180, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72415, CurrentID=0, Status='live', Symbol='Sgol1', Aliases='3300001M08Rik, C81037, Sgo1', description='shugoshin-like 1 (S. pombe)', other_designations='shugoshin 1A protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=53674786, end_position_on_the_genomic_accession=53689846, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72416, CurrentID=0, Status='live', Symbol='Lrpprc', Aliases='3110001K13Rik, C76645, Gp130, Lrp130, Lsfc', description='leucine-rich PPR-motif containing', other_designations='130 kDa leucine-rich protein|leucine rich protein LRP130', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84705247, end_position_on_the_genomic_accession=84790786, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  72421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72421, CurrentID=0, Status='live', Symbol='Ttc30b', Aliases='2510042P03Rik, Tcc30b', description='tetratricopeptide repeat domain 30B', other_designations='TPR repeat protein 30B|novel tetratrico peptide repeat (TPR) containing protein', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=75935849, end_position_on_the_genomic_accession=75938462, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  72425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72425, CurrentID=0, Status='live', Symbol='Katnbl1', Aliases='2410042D21Rik, AA166487, AA473307, AI467442', description='katanin p80 subunit B like 1', other_designations='katanin p80 subunit B-like 1', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112377107, end_position_on_the_genomic_accession=112414645, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72429, CurrentID=0, Status='live', Symbol='Dnajc25', Aliases='2010109C08Rik, 2010203O07Rik', description='DnaJ heat shock protein family (Hsp40) member C25', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 25', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58994810, end_position_on_the_genomic_accession=59023398, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72431, CurrentID=0, Status='live', Symbol='Ceacam18', Aliases='2010110O04Rik', description='carcinoembryonic antigen-related cell adhesion molecule 18', other_designations='CEA-related cell adhesion molecule 1|carcinoembryonic antigen-related cell adhesion molecule 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43634670, end_position_on_the_genomic_accession=43650228, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72432, CurrentID=0, Status='live', Symbol='Spink5', Aliases='2310065D10Rik, AV238912, LEKT1, LETK1, VAKT1', description='serine peptidase inhibitor, Kazal type 5', other_designations='Netherton syndrome|serine protease inhibitor, Kazal type 5', map_location='18 23.74 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=43959545, end_position_on_the_genomic_accession=44022487, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  72433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72433, CurrentID=0, Status='live', Symbol='Rab38', Aliases='2310011F14Rik, AU043391, cht', description='RAB38, member RAS oncogene family', other_designations='RAB38, member of RAS oncogene family|chocolate', map_location='7 49.19 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=88430273, end_position_on_the_genomic_accession=88491572, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72434, CurrentID=0, Status='live', Symbol='Lypd3', Aliases='2310061G07Rik, C4.4a', description='Ly6/Plaur domain containing 3', other_designations='GPI-anchored metastasis-associated protein C4.4A homolog|GPI-anchored metastasis-associated protein homolog', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24636570, end_position_on_the_genomic_accession=24641118, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72446, CurrentID=0, Status='live', Symbol='Prr5l', Aliases='2600010E01Rik, 4833411O04Rik, Protor-2, Protor2', description='proline rich 5 like', other_designations='protein observed with Rictor-2', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=101714280, end_position_on_the_genomic_accession=101883074, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  72454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72454, CurrentID=0, Status='live', Symbol='Ccdc71', Aliases='2600016J21Rik', description='coiled-coil domain containing 71', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108452571, end_position_on_the_genomic_accession=108465945, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72459, CurrentID=0, Status='live', Symbol='Htatsf1', Aliases='1600023H17Rik, 2600017A12Rik, 2700077B20Rik, TAT-SF1', description='HIV TAT specific factor 1', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=57053570, end_position_on_the_genomic_accession=57067183, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72461, CurrentID=0, Status='live', Symbol='Prcp', Aliases='2510048K03Rik, 2610104A14Rik, AI451719, HUMPCP, PCP', description='prolylcarboxypeptidase (angiotensinase C)', other_designations='proline carboxypeptidase', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=92875253, end_position_on_the_genomic_accession=92934581, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72462, CurrentID=0, Status='live', Symbol='Rrp1b', Aliases='2600005C20Rik, D030064A17, Kiaa0179, mKIAA0179', description='ribosomal RNA processing 1 homolog B (S. cerevisiae)', other_designations='RRP1-like protein B', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32036162, end_position_on_the_genomic_accession=32062865, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  72465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72465, CurrentID=0, Status='live', Symbol='Zfp131', Aliases='2610109I01Rik, AW558871, Znf131', description='zinc finger protein 131', other_designations='putative transcription factor ZNF131', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119765186, end_position_on_the_genomic_accession=119791714, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72469, CurrentID=0, Status='live', Symbol='Plcd3', Aliases='2610205J15Rik, mKIAA1964', description='phospholipase C, delta 3', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase delta-3|PLC-delta-3|phosphoinositide phospholipase C-delta-3|phospholipase C-delta-3', map_location='11 66.71 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103069236, end_position_on_the_genomic_accession=103101658, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  72472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72472, CurrentID=0, Status='live', Symbol='Slc16a10', Aliases='2610103N14Rik, 9830169E08, Mct10, PRO0813, TAT1', description='solute carrier family 16 (monocarboxylic acid transporters), member 10', other_designations='MCT 10|T-type amino acid transporter 1|aromatic amino acid transporter 1|solute carrier family 16 member 10', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40033532, end_position_on_the_genomic_accession=40142264, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72475, CurrentID=0, Status='live', Symbol='Ssbp3', Aliases='2610021L12Rik, 2610200M23Rik, 5730488C10Rik, AI854733, AW551939, LAST, SSDP1', description='single-stranded DNA binding protein 3', other_designations='Lck-associated signal transducer|sequence-specific single-stranded-DNA-binding protein', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106911470, end_position_on_the_genomic_accession=107049694, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  72477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72477, CurrentID=0, Status='live', Symbol='Tmem87b', Aliases='2610301K12Rik, 2810431I02Rik, AU014804', description='transmembrane protein 87B', other_designations='novel lung seven transmembrane receptor domain containing protein', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128818072, end_position_on_the_genomic_accession=128854261, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  72479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72479, CurrentID=0, Status='live', Symbol='Hsdl2', Aliases='2610207I16Rik', description='hydroxysteroid dehydrogenase like 2', other_designations='', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59580458, end_position_on_the_genomic_accession=59618694, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72480, CurrentID=0, Status='live', Symbol='Tspyl4', Aliases='2610102M01Rik, B230210I21Rik, D10Bwg0791e', description='TSPY-like 4', other_designations='TSPY-like protein 4', map_location='10 18.85 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34297421, end_position_on_the_genomic_accession=34301320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  72482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72482, CurrentID=0, Status='live', Symbol='Acbd6', Aliases='0610010G04Rik, 2610100E10Rik', description='acyl-Coenzyme A binding domain containing 6', other_designations='', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155558098, end_position_on_the_genomic_accession=155687316, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  72486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72486, CurrentID=0, Status='live', Symbol='Rnf219', Aliases='2610206B13Rik, 2810449K13Rik, AI451544', description='ring finger protein 219', other_designations='', map_location='14 E2.2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=104477534, end_position_on_the_genomic_accession=104522711, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  208691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208691, CurrentID=0, Status='live', Symbol='Eif5a2', Aliases='9630038B20, eIF5AII', description='eukaryotic translation initiation factor 5A2', other_designations='eIF-5A-2|eIF-5A2|eukaryotic initiation factor 5A', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=28781311, end_position_on_the_genomic_accession=28798846, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72500, CurrentID=0, Status='live', Symbol='Ier5l', Aliases='2610524G09Rik', description='immediate early response 5-like', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30472631, end_position_on_the_genomic_accession=30474199, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  72502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72502, CurrentID=0, Status='live', Symbol='Cwf19l1', Aliases='2610528C06Rik, AI854304, AV336991', description='CWF19-like 1, cell cycle control (S. pombe)', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44108637, end_position_on_the_genomic_accession=44135878, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  72503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72503, CurrentID=0, Status='live', Symbol='2610507B11Rik', Aliases='AU022687, D11Bhm178e, D11Bhm179e, E1, Kiaa0100, mKIAA0100', description='RIKEN cDNA 2610507B11 gene', other_designations='', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78261703, end_position_on_the_genomic_accession=78290625, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  72504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72504, CurrentID=0, Status='live', Symbol='Taf4b', Aliases='105kDa, 2610524B04Rik, 4932409F03Rik, AW987595, TAFII-105, TAFII105, Taf2c2', description='TATA-box binding protein associated factor 4b', other_designations='TAF(II)105|TAF4B RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF4B RNA polymerase II, TATA box binding protein (TBP)-associated factor, 105 kDa|TATA box binding protein (TBP)-associated factor, RNA polymerase II, C2, 105kD|transcription initiation factor TFIID 105 kDa subunit', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=14783201, end_position_on_the_genomic_accession=14960424, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  72507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72507, CurrentID=0, Status='live', Symbol='Dzip1l', Aliases='2610524A10Rik, AI663976', description='DAZ interacting protein 1-like', other_designations='DAZ-interacting protein 1-like protein', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99629501, end_position_on_the_genomic_accession=99669263, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  72508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72508, CurrentID=0, Status='live', Symbol='Rps6kb1', Aliases='2610318I15Rik, 4732464A07Rik, 70kDa, AA959758, AI256796, AI314060, S6K1, p70/85s6k, p70s6k', description='ribosomal protein S6 kinase, polypeptide 1', other_designations='70 kDa ribosomal protein S6 kinase 1|P70S6K1|S6K|S6K-beta-1|p70 S6 kinase alpha|p70 S6K-alpha|p70 S6KA|p70 ribosomal S6 kinase alpha|p70-S6K 1|ribosomal protein S6 kinase I|ribosomal protein S6 kinase, 70kD, polypeptide 1', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86498869, end_position_on_the_genomic_accession=86544807, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  72512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72512, CurrentID=0, Status='live', Symbol='Tmem173', Aliases='2610307O08Rik, ERIS, MPYS, Mita, STING', description='transmembrane protein 173', other_designations='endoplasmic reticulum interferon stimulator|mediator of IRF3 activation|mitochondrial mediator of IRF3 activation', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35733678, end_position_on_the_genomic_accession=35740554, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72514, CurrentID=0, Status='live', Symbol='Fgfbp3', Aliases='2610306H15Rik, Fgf-bp3', description='fibroblast growth factor binding protein 3', other_designations='FGF-binding protein 3|FGFBP-3', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36917550, end_position_on_the_genomic_accession=36919599, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  72515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72515, CurrentID=0, Status='live', Symbol='Wdr43', Aliases='2610318G08Rik, AU020887, AV024208, Wrd43, mKIAA0007', description='WD repeat domain 43', other_designations='', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71615900, end_position_on_the_genomic_accession=71659031, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  72519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72519, CurrentID=0, Status='live', Symbol='Tmem55a', Aliases='2610319K07Rik, AI315591, AV001360', description='transmembrane protein 55A', other_designations='ptdIns-4,5-P2 4-Ptase II|type 2 PtdIns-4,5-P2 4-Ptase|type II phosphatidylinositol 4,5-bisphosphate 4-phosphatase', map_location='4 A1-A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=14864219, end_position_on_the_genomic_accession=14915260, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72522, CurrentID=0, Status='live', Symbol='Atxn7l2', Aliases='2610528J18Rik', description='ataxin 7-like 2', other_designations='mFLJ00381 protein', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108202222, end_position_on_the_genomic_accession=108210527, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  72535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72535, CurrentID=0, Status='live', Symbol='Aldh1b1', Aliases='2700007F14Rik', description='aldehyde dehydrogenase 1 family, member B1', other_designations='aldehyde dehydrogenase family 1 member B1', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45799022, end_position_on_the_genomic_accession=45804608, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  72536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72536, CurrentID=0, Status='live', Symbol='Tagap', Aliases='2610315E15Rik', description='T cell activation Rho GTPase activating protein', other_designations='T-cell activation GTPase-activating protein 1', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=7926000, end_position_on_the_genomic_accession=7934897, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72542, CurrentID=0, Status='live', Symbol='Pgam5', Aliases='2610528A17Rik', description='phosphoglycerate mutase family member 5', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110259135, end_position_on_the_genomic_accession=110269899, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72543, CurrentID=0, Status='live', Symbol='Mvb12b', Aliases='2610200O14Rik, 2610528K11Rik, AI414895, Fam125b', description='multivesicular body subunit 12B', other_designations='ESCRT-I complex subunit MVB12B|family with sequence similarity 125, member B', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33729953, end_position_on_the_genomic_accession=33887946, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  72544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72544, CurrentID=0, Status='live', Symbol='Exosc6', Aliases='2610510N21Rik, C76919, Mtr3', description='exosome component 6', other_designations='Mtr3 (mRNA transport regulator 3)-homolog|exosome complex exonuclease MTR3', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111056339, end_position_on_the_genomic_accession=111057664, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  72549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72549, CurrentID=0, Status='live', Symbol='Reep4', Aliases='2700029E10Rik', description='receptor accessory protein 4', other_designations='', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70545218, end_position_on_the_genomic_accession=70548935, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72552, CurrentID=0, Status='live', Symbol='Hsdl1', Aliases='2700067E09Rik', description='hydroxysteroid dehydrogenase like 1', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119561978, end_position_on_the_genomic_accession=119575200, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72554, CurrentID=0, Status='live', Symbol='Utp14a', Aliases='2700066J21Rik, JsdX, NY-CO-16, SDCCAG16, T25628, mKIAA0266', description='UTP14, U3 small nucleolar ribonucleoprotein, homolog A (yeast)', other_designations='jsd-like X-linked protein|juvenile spermatogonial depletion-like X-linked protein', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48256934, end_position_on_the_genomic_accession=48282449, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  72555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72555, CurrentID=0, Status='live', Symbol='Shisa9', Aliases='2700045P11Rik, Ckamp44', description='shisa family member 9', other_designations='cystine-knot AMPAR modulating protein of 44 kDa|shisa homolog 9', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11984113, end_position_on_the_genomic_accession=12270904, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72556, CurrentID=0, Status='live', Symbol='Zfp566', Aliases='2700043M03Rik, mszf4', description='zinc finger protein 566', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30077337, end_position_on_the_genomic_accession=30090510, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72560, CurrentID=0, Status='live', Symbol='Naalad2', Aliases='2700022G20Rik, D9Ertd285e, Folh1b, GCP3, GCPIII, NAADALASE2, NAALADASE2', description='N-acetylated alpha-linked acidic dipeptidase 2', other_designations='N-acetylaspartylglutamate peptidase II|N-acetylated-alpha-linked acidic dipeptidase II|NAAG-peptidase II|NAALADase II|glutamate carboxypeptidase III', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18321954, end_position_on_the_genomic_accession=18397538, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  72562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72562, CurrentID=0, Status='live', Symbol='Pcbd2', Aliases='2700061N24Rik, Dcoh2, Dcohm', description='pterin 4 alpha carbinolamine dehydratase/dimerization cofactor of hepatocyte nuclear factor 1 alpha (TCF1) 2', other_designations='4-alpha-hydroxy-tetrahydropterin dehydratase 2|HNF-1-alpha dimerization cofactor|PHS 2|dcoH-like protein DCoHm|dimerization cofactor of hepatocyte nuclear factor 1 (HNF1) from muscle|dimerization cofactor of hepatocyte nuclear factor 1 from muscle', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55727124, end_position_on_the_genomic_accession=55776830, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72565, CurrentID=0, Status='live', Symbol='Uaca', Aliases='2700059D02Rik, nucling', description='uveal autoantigen with coiled-coil domains and ankyrin repeats', other_designations='nuclear membrane binding protein NUCLING|nuclear membrane-binding protein', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=60793206, end_position_on_the_genomic_accession=60880370, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  72567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72567, CurrentID=0, Status='live', Symbol='Bclaf1', Aliases='2610102K23Rik, 2700025J07Rik, 2810454G14Rik, 5730534O06Rik, AI450190, AW556225, Btf, mKIAA0164', description='BCL2-associated transcription factor 1', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=20311461, end_position_on_the_genomic_accession=20342503, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  72568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72568, CurrentID=0, Status='live', Symbol='Lin9', Aliases='2700022J23Rik, Bara, Lin-9, TGS, TGS1, mLin-9', description='lin-9 homolog (C. elegans)', other_designations='TUDOR gene similar 1 protein|type I interferon receptor beta chain-associated protein', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180641150, end_position_on_the_genomic_accession=180690687, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  72569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72569, CurrentID=0, Status='live', Symbol='Bbs5', Aliases='1700049I01Rik, 2700023J09Rik', description='Bardet-Biedl syndrome 5 (human)', other_designations='novel DUF1448 protein', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69646730, end_position_on_the_genomic_accession=69667569, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  72572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72572, CurrentID=0, Status='live', Symbol='Spats2', Aliases='2700012F11Rik, 59kDa, Scr59, p59', description='spermatogenesis associated, serine-rich 2', other_designations='p59scr|serine-rich spermatocytes and round spermatid 59 kDa protein|serine-rich spermatocytes and round spermatid protein, 59kDa', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99125926, end_position_on_the_genomic_accession=99213215, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  72580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72580, CurrentID=0, Status='live', Symbol='Zufsp', Aliases='2700019D07Rik', description='zinc finger with UFM1-specific peptidase domain', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=33919142, end_position_on_the_genomic_accession=33951347, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  72584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72584, CurrentID=0, Status='live', Symbol='Cul4b', Aliases='2700050M05Rik, AA409770, CUL-4B, mKIAA0695', description='cullin 4B', other_designations='CUL-4B|cul4b', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38531616, end_position_on_the_genomic_accession=38576196, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  72585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72585, CurrentID=0, Status='live', Symbol='Lypd1', Aliases='2700050C12Rik, AI853408, C530008O16Rik, Lynx2, Lypdc1', description='Ly6/Plaur domain containing 1', other_designations='', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=125867635, end_position_on_the_genomic_accession=125913186, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72587, CurrentID=0, Status='live', Symbol='Pan3', Aliases='2700050F09Rik, A430027N15Rik, AU014670', description='PAN3 polyA specific ribonuclease subunit homolog (S. cerevisiae)', other_designations='PAB-dependent poly(A)-specific ribonuclease subunit 3|Pabp-dependent poly(A) nuclease 3|Pabp1-dependent poly A-specific ribonuclease subunit Pan3', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147430162, end_position_on_the_genomic_accession=147548502, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  72590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72590, CurrentID=0, Status='live', Symbol='Ppme1', Aliases='1110069N17Rik, 2700017M01Rik, PME-1, Pme1', description='protein phosphatase methylesterase 1', other_designations='', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100326737, end_position_on_the_genomic_accession=100371896, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72599, CurrentID=0, Status='live', Symbol='Pdia5', Aliases='2700053F16Rik, AU015525, Pdir', description='protein disulfide isomerase associated 5', other_designations='protein disulfide isomerase-associated 5|protein disulfide isomerase-related protein', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35397308, end_position_on_the_genomic_accession=35490898, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  72605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72605, CurrentID=0, Status='live', Symbol='Car10', Aliases='2700029L05Rik, BB085816, Ca10', description='carbonic anhydrase 10', other_designations='CA-RP X|CARP X|carbonic anhydrase-related protein X', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=93098142, end_position_on_the_genomic_accession=93601751, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72607, CurrentID=0, Status='live', Symbol='Usp13', Aliases='2700071E21Rik, AI848077, ISOT3, IsoT-3', description='ubiquitin specific peptidase 13 (isopeptidase T-3)', other_designations='deubiquitinating enzyme 13|ubiquitin specific protease 13 (isopeptidase T-3)|ubiquitin thioesterase 13|ubiquitin-specific-processing protease 13', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32817019, end_position_on_the_genomic_accession=32938075, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  72611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72611, CurrentID=0, Status='live', Symbol='Zfp655', Aliases='2700038I16Rik, 9030409O18Rik, mKIAA4222', description='zinc finger protein 655', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145231428, end_position_on_the_genomic_accession=145247306, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72612, CurrentID=0, Status='live', Symbol='2700029M09Rik', Aliases='C230006B22Rik', description='RIKEN cDNA 2700029M09 gene', other_designations='', map_location='8 B3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=60890393, end_position_on_the_genomic_accession=60907582, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72614, CurrentID=0, Status='live', Symbol='Pih1d2', Aliases='2700059L22Rik', description='PIH1 domain containing 2', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50617321, end_position_on_the_genomic_accession=50625000, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72615, CurrentID=0, Status='live', Symbol='Anks3', Aliases='2700067D09Rik, C81345, mKIAA1977', description='ankyrin repeat and sterile alpha motif domain containing 3', other_designations='', map_location='16 2.48 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4941415, end_position_on_the_genomic_accession=4964327, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  72621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72621, CurrentID=0, Status='live', Symbol='Pdzd11', Aliases='1810012H22Rik, 2310079D11Rik, 2700099C19Rik, AI854765, Pdzk11', description='PDZ domain containing 11', other_designations='', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100622880, end_position_on_the_genomic_accession=100625974, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  72630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72630, CurrentID=0, Status='live', Symbol='Hspa12b', Aliases='2700081N06Rik', description='heat shock protein 12B', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131127412, end_position_on_the_genomic_accession=131145985, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  72632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72632, CurrentID=0, Status='live', Symbol='Smim18', Aliases='2700090O03Rik, AI835723', description='small integral membrane protein 18', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33742112, end_position_on_the_genomic_accession=33748016, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  72634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72634, CurrentID=0, Status='live', Symbol='Tdrkh', Aliases='2700091C21Rik, Tdrd2', description='tudor and KH domain containing protein', other_designations='tudor domain-containing protein 2', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94413239, end_position_on_the_genomic_accession=94432975, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72635, CurrentID=0, Status='live', Symbol='Lins1', Aliases='2700083B01Rik, AW547219, Lins, Lins2, Wins2', description='lines homolog 1', other_designations='Wins2 protein with Drosophila Lin homologous domain|lines homolog (Drosophila) 1|lines homolog 2|protein Lines homolog 2|protein with Drosophila Lines homologous domain|wnt-signaling molecule Lines homolog 1|wnt-signaling molecule Lines homolog 2', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66689767, end_position_on_the_genomic_accession=66717256, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72640, CurrentID=0, Status='live', Symbol='Mex3a', Aliases='2700083E18Rik, Gm411, Rkhd4', description='mex3 RNA binding family member A', other_designations='mex3 homolog A|ring finger and KH domain containing 4', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88532395, end_position_on_the_genomic_accession=88541394, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  72649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72649, CurrentID=0, Status='live', Symbol='Tmem209', Aliases='2700094F01Rik, AI428435', description='transmembrane protein 209', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30481022, end_position_on_the_genomic_accession=30509773, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  72650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72650, CurrentID=0, Status='live', Symbol='2810006K23Rik', Aliases='', description='RIKEN cDNA 2810006K23 gene', other_designations='Probable peptide chain release factor C12orf65 homolog, mitochondrial', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124328089, end_position_on_the_genomic_accession=124341852, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72654, CurrentID=0, Status='live', Symbol='Ccdc12', Aliases='2700094L05Rik, C76605', description='coiled-coil domain containing 12', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110656503, end_position_on_the_genomic_accession=110711593, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72655, CurrentID=0, Status='live', Symbol='Snhg5', Aliases='2810026P18Rik, 4933428I09Rik, mU50HG-b', description='small nucleolar RNA host gene 5', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88521052, end_position_on_the_genomic_accession=88522897, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72656, CurrentID=0, Status='live', Symbol='Ints8', Aliases='2810013E07Rik, AV063769, D130008D20Rik', description='integrator complex subunit 8', other_designations='int8', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11199158, end_position_on_the_genomic_accession=11254266, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  72657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72657, CurrentID=0, Status='live', Symbol='2700094K13Rik', Aliases='Selh', description='RIKEN cDNA 2700094K13 gene', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84669221, end_position_on_the_genomic_accession=84670708, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  72658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72658, CurrentID=0, Status='live', Symbol='2700097O09Rik', Aliases='AW552125', description='RIKEN cDNA 2700097O09 gene', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55045661, end_position_on_the_genomic_accession=55080238, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72661, CurrentID=0, Status='live', Symbol='Serp2', Aliases='2810032E02Rik', description='stress-associated endoplasmic reticulum protein family member 2', other_designations='ribosome-associated membrane protein RAMP4-2', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=76532812, end_position_on_the_genomic_accession=76557919, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  72662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72662, CurrentID=0, Status='live', Symbol='Dis3', Aliases='2810028N01Rik', description="DIS3 homolog, exosome endoribonuclease and 3'-5' exoribonuclease", other_designations='DIS3 mitotic control homolog|protein DIS3 homolog|ribosomal RNA-processing protein 44', map_location='14 E2.2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=99076634, end_position_on_the_genomic_accession=99099778, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  72667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72667, CurrentID=0, Status='live', Symbol='Zfp444', Aliases='2810031J10Rik, 6230401O10Rik, mFLJ00134', description='zinc finger protein 444', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6172462, end_position_on_the_genomic_accession=6193104, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72668, CurrentID=0, Status='live', Symbol='Skida1', Aliases='2810030E01Rik, 5730507N06Rik, BB088113', description='SKI/DACH domain containing 1', other_designations='protein DLN-1', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=18040225, end_position_on_the_genomic_accession=18049036, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72672, CurrentID=0, Status='live', Symbol='Zfp518a', Aliases='2810401C22Rik, 6330417C12Rik, AI426493, Zfp518, Znf518, Znf518a, mKIAA0335', description='zinc finger protein 518A', other_designations='zinc finger protein 518', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40893930, end_position_on_the_genomic_accession=40917947, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72674, CurrentID=0, Status='live', Symbol='Adipor1', Aliases='2810031L11Rik, ACDCR1, CGI-45, Paqr1', description='adiponectin receptor 1', other_designations='progestin and adipoQ receptor family member I', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134415384, end_position_on_the_genomic_accession=134432352, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72685, CurrentID=0, Status='live', Symbol='Dnajc6', Aliases='2810027M23Rik, mKIAA0473', description='DnaJ heat shock protein family (Hsp40) member C6', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 6|DnaJ (Hsp40) homolog, subfamily C, member 6|auxilin|dnaJ homolog subfamily C member 6', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101496022, end_position_on_the_genomic_accession=101645004, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  72690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72690, CurrentID=0, Status='live', Symbol='Grrp1', Aliases='2810043G13Rik, Fam110d', description='glycine/arginine rich protein 1', other_designations='glycine/arginine-rich protein 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134251102, end_position_on_the_genomic_accession=134266197, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72691, CurrentID=0, Status='live', Symbol='Calhm2', Aliases='2810048G17Rik, Fam26b', description='calcium homeostasis modulator 2', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47132232, end_position_on_the_genomic_accession=47138294, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72692, CurrentID=0, Status='live', Symbol='Hnrnpll', Aliases='2510028H02Rik, 2810036L13Rik, AI256697, AI852082, Hnrpll', description='heterogeneous nuclear ribonucleoprotein L-like', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80029487, end_position_on_the_genomic_accession=80062513, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72693, CurrentID=0, Status='live', Symbol='Zcchc12', Aliases='2810028A01Rik, AV136720, Sizn1', description='zinc finger, CCHC domain containing 12', other_designations='smad-interacting zinc finger protein 1', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36195904, end_position_on_the_genomic_accession=36199158, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72699, CurrentID=0, Status='live', Symbol='Lime1', Aliases='2810038K19Rik, LIME', description='Lck interacting transmembrane adaptor 1', other_designations='lck-interacting molecule', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181381235, end_position_on_the_genomic_accession=181383628, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72701, CurrentID=0, Status='live', Symbol='Zfp618', Aliases='2810031P15Rik, 2810040O04Rik, AI415580, AI505139, D430033D05Rik, Nedd10, Znf618, bM257P7.1', description='zinc finger protein 618', other_designations='neural precursor cell expressed, developmentally down-regulated gene 10', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62965555, end_position_on_the_genomic_accession=63139709, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  72709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72709, CurrentID=0, Status='live', Symbol='C1qtnf6', Aliases='2810036M19Rik, CTRP6', description='C1q and tumor necrosis factor related protein 6', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78523346, end_position_on_the_genomic_accession=78532191, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72713, CurrentID=0, Status='live', Symbol='Angptl1', Aliases='2810039D03Rik, ANG3, ANGPT3, ANGY, ARP1', description='angiopoietin-like 1', other_designations='angiopoietin-like protein 1', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156839023, end_position_on_the_genomic_accession=156861078, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72720, CurrentID=0, Status='live', Symbol='Zfp248', Aliases='2810037F07Rik, E130106N01Rik', description='zinc finger protein 248', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118427319, end_position_on_the_genomic_accession=118455506, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72722, CurrentID=0, Status='live', Symbol='Fam98a', Aliases='2810405J04Rik, AL024114', description='family with sequence similarity 98, member A', other_designations='', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=75537086, end_position_on_the_genomic_accession=75551946, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72723, CurrentID=0, Status='live', Symbol='Zfp74', Aliases='2810054M15Rik, KRAB8, Zfp66, Znf569, mszf21, mszf77', description='zinc finger protein 74', other_designations='zfp-74', map_location='7 17.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29931711, end_position_on_the_genomic_accession=29955843, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72726, CurrentID=0, Status='live', Symbol='Tbcc', Aliases='2810055C19Rik', description='tubulin-specific chaperone C', other_designations='CFC|tubulin-folding cofactor C', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46890621, end_position_on_the_genomic_accession=46892463, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  72727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72727, CurrentID=0, Status='live', Symbol='B3gat3', Aliases='2810405M13Rik', description='beta-1,3-glucuronyltransferase 3 (glucuronosyltransferase I)', other_designations='UDP-GlcUA:Gal beta-1,3-Gal-R glucuronyltransferase|glcAT-I|glcUAT-I|glucuronosyltransferase I|glucuronosyltransferase-I|glucuronyltransferase I', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8920393, end_position_on_the_genomic_accession=8927236, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72729, CurrentID=0, Status='live', Symbol='Cdc42se2', Aliases='2810404F18Rik, AA408783, AA536669, SPEC2', description='CDC42 small effector 2', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54717415, end_position_on_the_genomic_accession=54787703, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72735, CurrentID=0, Status='live', Symbol='Eldr', Aliases='2810442I21Rik, 4930500E12Rik, Fabl', description='Egfr long non-coding downstream RNA', other_designations='', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=16934709, end_position_on_the_genomic_accession=16951282, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72736, CurrentID=0, Status='live', Symbol='Tmx1', Aliases='2810425A04Rik, AI465319, Txndc1', description='thioredoxin-related transmembrane protein 1', other_designations='thioredoxin domain containing 1|thioredoxin domain-containing protein 1', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70453154, end_position_on_the_genomic_accession=70467624, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72739, CurrentID=0, Status='live', Symbol='Zkscan3', Aliases='2810435N07Rik, AA408594, AI132359, Skz1, Zf47, Zfp-47, Zfp306, Zfp307, mszf35', description='zinc finger with KRAB and SCAN domains 3', other_designations='SCAN-KRAB-zinc finger gene 1|SCAN-KRAB-zinc finger protein|zinc finger protein 306|zinc finger protein 307|zinc finger protein 47 homolog', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21387004, end_position_on_the_genomic_accession=21402755, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72745, CurrentID=0, Status='live', Symbol='Tmem161b', Aliases='2810446P07Rik, AI843389', description='transmembrane protein 161B', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=84222187, end_position_on_the_genomic_accession=84295966, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72747, CurrentID=0, Status='live', Symbol='Ttc39c', Aliases='1700008N02Rik, 2810439F02Rik', description='tetratricopeptide repeat domain 39C', other_designations='TPR repeat protein 39C', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12643533, end_position_on_the_genomic_accession=12737052, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  72748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72748, CurrentID=0, Status='live', Symbol='Hdhd3', Aliases='2810435D12Rik', description='haloacid dehalogenase-like hydrolase domain containing 3', other_designations='', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62499054, end_position_on_the_genomic_accession=62502200, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  72749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72749, CurrentID=0, Status='live', Symbol='Tonsl', Aliases='2810439M11Rik, Nfkbil2', description='tonsoku-like, DNA repair protein', other_designations='I-kappa-B-related protein|NF-kappa-B inhibitor-like protein 2|ikappaBR|inhibitor of kappa B-related protein|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor-like 2', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76626237, end_position_on_the_genomic_accession=76639929, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  72750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72750, CurrentID=0, Status='live', Symbol='Fam117b', Aliases='2810425F24Rik, 6330416D14Rik, Als2cr13', description='family with sequence similarity 117, member B', other_designations='amyotrophic lateral sclerosis 2 chromosomal region candidate gene 13 protein homolog', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59913006, end_position_on_the_genomic_accession=59985348, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72754, CurrentID=0, Status='live', Symbol='Arhgef10l', Aliases='2810441C07Rik', description='Rho guanine nucleotide exchange factor (GEF) 10-like', other_designations='GrinchGEF', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140514485, end_position_on_the_genomic_accession=140665905, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  72759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72759, CurrentID=0, Status='live', Symbol='Tmem135', Aliases='2810439K08Rik, AW319712, PMP52', description='transmembrane protein 135', other_designations='peroxisomal membrane protein 52', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=89139721, end_position_on_the_genomic_accession=89404488, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  72772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72772, CurrentID=0, Status='live', Symbol='Rint1', Aliases='1500019C06Rik, 2810450M21Rik, Rint-1', description='RAD50 interactor 1', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23787681, end_position_on_the_genomic_accession=23820369, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  72774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72774, CurrentID=0, Status='live', Symbol='Neil1', Aliases='2810450N13Rik, Nei1', description='nei endonuclease VIII-like 1 (E. coli)', other_designations='DNA glycosylase/AP lyase Neil1|DNA-(apurinic or apyrimidinic site) lyase Neil1|NEH1|endonuclease VIII|endonuclease VIII-like 1|nei homolog 1|nei-like protein 1', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57142788, end_position_on_the_genomic_accession=57148675, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72775, CurrentID=0, Status='live', Symbol='Fance', Aliases='2810451D06Rik, AI415634, AW209126', description='Fanconi anemia, complementation group E', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28313530, end_position_on_the_genomic_accession=28326574, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72776, CurrentID=0, Status='live', Symbol='Sass6', Aliases='2810453L12Rik', description='spindle assembly 6 homolog (C. elegans)', other_designations='', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116594958, end_position_on_the_genomic_accession=116630986, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  72778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72778, CurrentID=0, Status='live', Symbol='Dnajc22', Aliases='2810451A06Rik, AI506245', description='DnaJ heat shock protein family (Hsp40) member C22', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 3; heat shock protein, DNAJ-like 3|DnaJ (Hsp40) homolog, subfamily C, member 22', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99099134, end_position_on_the_genomic_accession=99104737, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  72780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72780, CurrentID=0, Status='live', Symbol='Rspo3', Aliases='2810459H04Rik, AW742308, Cristin1, Thsd2', description='R-spondin 3', other_designations='R-spondin 3 homolog|cabriolet|cristin 1|cristin-1|cysteine-rich and single thrombospondin domain-containing protein 1|nucleopondin|roof plate-specific spondin-3|thrombospondin, type I, domain 2', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=29453106, end_position_on_the_genomic_accession=29535867, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  230584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230584, CurrentID=0, Status='live', Symbol='Yipf1', Aliases='C030002N13Rik', description='Yip1 domain family, member 1', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107314254, end_position_on_the_genomic_accession=107359823, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72787, CurrentID=0, Status='live', Symbol='Ndc1', Aliases='2810475A17Rik, AI450313, Tmem48, sks', description='NDC1 transmembrane nucleoporin', other_designations='skeletal fusions with sterility|transmembrane protein 48', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107367784, end_position_on_the_genomic_accession=107416346, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  72789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72789, CurrentID=0, Status='live', Symbol='Veph1', Aliases='2810471M23Rik, Veph', description='ventricular zone expressed PH domain-containing 1', other_designations='protein melted homolog', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=66053558, end_position_on_the_genomic_accession=66297066, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  72792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72792, CurrentID=0, Status='live', Symbol='2810459M11Rik', Aliases='', description='RIKEN cDNA 2810459M11 gene', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86045863, end_position_on_the_genomic_accession=86055456, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  72795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72795, CurrentID=0, Status='live', Symbol='Ttc19', Aliases='2010204O13Rik, 2810460C24Rik, AI505442', description='tetratricopeptide repeat domain 19', other_designations='TPR repeat protein 19', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62281427, end_position_on_the_genomic_accession=62316441, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72805, CurrentID=0, Status='live', Symbol='Zfp839', Aliases='2810455K09Rik', description='zinc finger protein 839', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110850250, end_position_on_the_genomic_accession=110869998, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72807, CurrentID=0, Status='live', Symbol='Zfp429', Aliases='2810487A22Rik, AI929863, Rsl2', description='zinc finger protein 429', other_designations='regulator of sex-limitation 2', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67389304, end_position_on_the_genomic_accession=67399844, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  72821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72821, CurrentID=0, Status='live', Symbol='Scn2b', Aliases='2810451E09Rik, AI840361, Gm183', description='sodium channel, voltage-gated, type II, beta', other_designations='sodium channel, voltage-gated, type II, beta polypeptide', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45117876, end_position_on_the_genomic_accession=45130070, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72823, CurrentID=0, Status='live', Symbol='Pard3b', Aliases='1810008K04Rik, 2010002N16Rik, 2810455B10Rik, Als2cr19, PAR3B, PAR3L, PAR3beta', description='par-3 family cell polarity regulator beta', other_designations='DN-276O13.2|PAR3-L protein|PAR3-beta|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 19 protein homolog|par-3 partitioning defective 3 homolog B|partitioning defective 3-like protein', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=61638667, end_position_on_the_genomic_accession=62642284, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  72825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72825, CurrentID=0, Status='live', Symbol='Mon1a', Aliases='2810468K17Rik', description='MON1 homolog A (yeast)', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107888112, end_position_on_the_genomic_accession=107903139, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72826, CurrentID=0, Status='live', Symbol='Fam76b', Aliases='2810485I05Rik, C78303', description='family with sequence similarity 76, member B', other_designations='', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=13827702, end_position_on_the_genomic_accession=13854687, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72828, CurrentID=0, Status='live', Symbol='Ubash3b', Aliases='2810457I06Rik, BB125008, TULA-2, p70', description='ubiquitin associated and SH3 domain containing, B', other_designations='STS-1|T-cell ubiquitin ligand 2|cbl-interacting protein p70|suppressor of T-cell receptor signaling 1|tyrosine-protein phosphatase STS1/TULA2', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=41010760, end_position_on_the_genomic_accession=41161692, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  72831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72831, CurrentID=0, Status='live', Symbol='Dhx30', Aliases='2810477H02Rik, C130058C04Rik, Ddx30, HELG, Ret-CoR', description='DEAH (Asp-Glu-Ala-His) box polypeptide 30', other_designations='DEAH box protein 30', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110084319, end_position_on_the_genomic_accession=110117616, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  72832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72832, CurrentID=0, Status='live', Symbol='Crtac1', Aliases='2810454P21Rik, AW047536B, Lotus, Crtac1', description='cartilage acidic protein 1', other_designations='68 kDa chondrocyte-expressed protein|ASPIC|CEP-68|protein CRTAC1-B', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42283036, end_position_on_the_genomic_accession=42431820, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  72836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72836, CurrentID=0, Status='live', Symbol='Pot1b', Aliases='2810458H16Rik', description='protection of telomeres 1B', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55651190, end_position_on_the_genomic_accession=55712647, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  72843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72843, CurrentID=0, Status='live', Symbol='Prdm4', Aliases='1700031E19Rik, 2810470D21Rik, AW552272, SC-1', description='PR domain containing 4', other_designations='PR domain-containing protein 4', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85891964, end_position_on_the_genomic_accession=85917152, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72844, CurrentID=0, Status='live', Symbol='Kctd17', Aliases='2900008M13Rik, AA414907, AW742389, N28155', description='potassium channel tetramerisation domain containing 17', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78428559, end_position_on_the_genomic_accession=78439303, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  72852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72852, CurrentID=0, Status='live', Symbol='Mblac2', Aliases='2900024O10Rik', description='metallo-beta-lactamase domain containing 2', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=81711417, end_position_on_the_genomic_accession=81753275, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  72865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72865, CurrentID=0, Status='live', Symbol='Cxx1c', Aliases='2900027G03Rik, Mar8.1, Mart8a', description='CAAX box 1C', other_designations='mammalian retrotransposon derived 8a', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53607922, end_position_on_the_genomic_accession=53609135, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  99526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99526, CurrentID=0, Status='live', Symbol='Usp53', Aliases='AA939927, Phxr3, Sp6, mKIAA1350', description='ubiquitin specific peptidase 53', other_designations='inactive ubiquitin-specific peptidase 53|per-hexamer repeat gene 3|per-hexamer repeat protein 3|ubiquitin specific protease 53', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122931486, end_position_on_the_genomic_accession=122984509, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  72873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72873, CurrentID=0, Status='live', Symbol='Bbof1', Aliases='2900006K08Rik, AI427784, Ccdc176', description='basal body orientation factor 1', other_designations='coiled-coil domain containing 176|coiled-coil domain-containing protein 176', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84409068, end_position_on_the_genomic_accession=84441763, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100862126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100862126, CurrentID=0, Status='live', Symbol='Gm21498', Aliases='', description='predicted gene, 21498', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21427428, end_position_on_the_genomic_accession=21428386, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  72881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72881, CurrentID=0, Status='live', Symbol='Zdhhc4', Aliases='1810021D01Rik, 2900029I10Rik', description='zinc finger, DHHC domain containing 4', other_designations='DHHC-4|zinc finger DHHC domain-containing protein 4', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143316489, end_position_on_the_genomic_accession=143329252, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72886, CurrentID=0, Status='live', Symbol='Ccdc94', Aliases='2900016D05Rik, AI413813', description='coiled-coil domain containing 94', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55959187, end_position_on_the_genomic_accession=55968285, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72891, CurrentID=0, Status='live', Symbol='Xlr4c', Aliases='2900036K24Rik', description='X-linked lymphocyte-regulated 4C', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73234072, end_position_on_the_genomic_accession=73243130, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  72895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72895, CurrentID=0, Status='live', Symbol='Setd5', Aliases='2900045N06Rik, C330007C20, C85544, mKIAA1757', description='SET domain containing 5', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113077624, end_position_on_the_genomic_accession=113153426, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  72898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72898, CurrentID=0, Status='live', Symbol='Asphd2', Aliases='2900006N09Rik, 9230106G13Rik', description='aspartate beta-hydroxylase domain containing 2', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112384354, end_position_on_the_genomic_accession=112395379, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  72899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72899, CurrentID=0, Status='live', Symbol='Macrod2', Aliases='1110033L15Rik, 2610107G07Rik, 2900006F19Rik', description='MACRO domain containing 2', other_designations='MACRO domain-containing protein 2|[Protein ADP-ribosylglutamate] hydrolase', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=140395356, end_position_on_the_genomic_accession=142392966, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  72900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72900, CurrentID=0, Status='live', Symbol='Ndufv2', Aliases='2900010C23Rik', description='NADH dehydrogenase (ubiquinone) flavoprotein 2', other_designations='NADH-ubiquinone oxidoreductase 24 kDa subunit', map_location='17 E1.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=66078795, end_position_on_the_genomic_accession=66101559, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72902, CurrentID=0, Status='live', Symbol='Spock3', Aliases='2900045C01Rik, AI428471, mKIAA4039', description='sparc/osteonectin, cwcv and kazal-like domains proteoglycan 3', other_designations='testican 3', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=62951090, end_position_on_the_genomic_accession=63357101, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  72925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72925, CurrentID=0, Status='live', Symbol='March1', Aliases='2900024D24Rik, BB085186', description='membrane-associated ring finger (C3HC4) 1', other_designations='MARCH-I|membrane-associated RING finger protein 1|membrane-associated RING-CH protein I', map_location='8 B3.1-B3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=65617922, end_position_on_the_genomic_accession=66471637, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  72927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72927, CurrentID=0, Status='live', Symbol='Hepacam', Aliases='2900042E01Rik, Glialcam, Hepn1', description='hepatocyte cell adhesion molecule', other_designations='hepacam|protein hepaCAM', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37367606, end_position_on_the_genomic_accession=37386572, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72930, CurrentID=0, Status='live', Symbol='Ppp2r2b', Aliases='2900026H06Rik, 6330404L05Rik, E130009M08Rik, PP2A-PR55B, PR55-BETA, SCA12', description='protein phosphatase 2, regulatory subunit B, beta', other_designations='PP2A subunit B isoform B55-beta|PP2A subunit B isoform PR55-beta|PP2A subunit B isoform R2-beta|PP2A subunit B isoform beta|protein phosphatase 2 (formerly 2A), regulatory subunit B (PR 52), beta isoform', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42645221, end_position_on_the_genomic_accession=43059471, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72931, CurrentID=0, Status='live', Symbol='Swi5', Aliases='1500019F05Rik, 2900010J23Rik, AI046671', description='SWI5 recombination repair homolog (yeast)', other_designations='protein SAE3 homolog', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32278805, end_position_on_the_genomic_accession=32288136, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  72935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72935, CurrentID=0, Status='live', Symbol='Ddx41', Aliases='2900024F02Rik, AA958953, ABS, AI324246', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 41', other_designations='DEAD box protein 41|DEAD-box protein abstrakt', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55530410, end_position_on_the_genomic_accession=55536658, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  72938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72938, CurrentID=0, Status='live', Symbol='Hspb11', Aliases='2900042B11Rik, IFT25', description='heat shock protein family B (small), member 11', other_designations='PP25|heat shock protein beta-11|placental protein 25 homolog', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107253598, end_position_on_the_genomic_accession=107284497, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  72946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72946, CurrentID=0, Status='live', Symbol='Lrrc47', Aliases='2900010D03Rik, AA960559, mKIAA1185', description='leucine rich repeat containing 47', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154011803, end_position_on_the_genomic_accession=154021512, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72947, CurrentID=0, Status='live', Symbol='Phykpl', Aliases='2900006B13Rik, Agxt2l2', description='5-phosphohydroxy-L-lysine phospholyase', other_designations='alanine--glyoxylate aminotransferase 2-like 2|alanine-glyoxylate aminotransferase 2-like 2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51583983, end_position_on_the_genomic_accession=51603269, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  72948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72948, CurrentID=0, Status='live', Symbol='Tppp', Aliases='2900041A09Rik, AI849835', description='tubulin polymerization promoting protein', other_designations='brain-specific protein p25 alpha', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74009419, end_position_on_the_genomic_accession=74035753, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  72949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72949, CurrentID=0, Status='live', Symbol='Ccnt2', Aliases='2900041I18Rik, C81304, CycT2, CycT2a, CycT2b', description='cyclin T2', other_designations='cyclin T2a|cyclin T2b', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=127774164, end_position_on_the_genomic_accession=127804837, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  72960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72960, CurrentID=0, Status='live', Symbol='Top1mt', Aliases='2900052H09Rik', description='DNA topoisomerase 1, mitochondrial', other_designations='mitochondrial topoisomerase I', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75657029, end_position_on_the_genomic_accession=75678822, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72961, CurrentID=0, Status='live', Symbol='Slc17a7', Aliases='2900052E22Rik, AI851913, Vglut1', description='solute carrier family 17 (sodium-dependent inorganic phosphate cotransporter), member 7', other_designations='brain-specific Na(+)-dependent inorganic phosphate cotransporter|solute carrier family 17 member 7', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45163921, end_position_on_the_genomic_accession=45176139, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  72962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72962, CurrentID=0, Status='live', Symbol='Tymp', Aliases='2900072D10Rik, Ecgf1, PD-ECGF, PDECGF, Pdgfec', description='thymidine phosphorylase', other_designations='TP|endothelial cell growth factor 1 (platelet-derived)|gliostatin|platelet derived growth factor, endothelial cell|tdRPase', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89371931, end_position_on_the_genomic_accession=89377036, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  72972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72972, CurrentID=0, Status='live', Symbol='Ccser2', Aliases='1700012P13Rik, 2900054P12Rik, AI035535, AW108503, Fam190b, Gcap14', description='coiled-coil serine rich 2', other_designations='coiled-coil serine-rich protein 2|granule cell antiserum positive 14|granule cell antiserum positive protein 14|protein FAM190B', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=36874936, end_position_on_the_genomic_accession=36968764, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72973, CurrentID=0, Status='live', Symbol='Fbxo47', Aliases='2900052P03Rik, Gm1562', description='F-box protein 47', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97851558, end_position_on_the_genomic_accession=97884390, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  72978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72978, CurrentID=0, Status='live', Symbol='Cnih3', Aliases='2900075G08Rik, CNIH-3', description='cornichon family AMPA receptor auxiliary protein 3', other_designations='cornichon homolog 3', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181352628, end_position_on_the_genomic_accession=181460641, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  72981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72981, CurrentID=0, Status='live', Symbol='Prkrir', Aliases='2900052B10Rik, Dap4, Rpkrir', description='protein-kinase, interferon-inducible double stranded RNA dependent inhibitor, repressor of (P58 repressor)', other_designations='58 kDa interferon-induced protein kinase-interacting protein|P52rIPK|THAP domain-containing protein 0|interferon-inducible, double stranded RNA dependent inhibitor, protein-kinase|p58IPK-interacting protein', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98703230, end_position_on_the_genomic_accession=98718062, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  72982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72982, CurrentID=0, Status='live', Symbol='Tmem138', Aliases='1700113I01Rik, 2900055D14Rik', description='transmembrane protein 138', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10570478, end_position_on_the_genomic_accession=10577386, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  72993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72993, CurrentID=0, Status='live', Symbol='Appl1', Aliases='2900057D21Rik, 7330406P05Rik, AI585782, AW209077, BB022931, C88264, DIP13', description='adaptor protein, phosphotyrosine interaction, PH domain and leucine zipper containing 1', other_designations='DIP13 alpha|adapter protein containing PH domain, PTB domain and leucine zipper motif 1|adaptor protein APPL like|dip13-alpha', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26918986, end_position_on_the_genomic_accession=26971205, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  72997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72997, CurrentID=0, Status='live', Symbol='Kantr', Aliases='2810454L23Rik, 2900056M20Rik, Spasm', description='Kdm5c adjacent non-coding transcript', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152294824, end_position_on_the_genomic_accession=152327493, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  72999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72999, CurrentID=0, Status='live', Symbol='Insig2', Aliases='2900053I11Rik, C730043J18Rik, Insig-2', description='insulin induced gene 2', other_designations='INSIG-2 membrane protein', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=121304353, end_position_on_the_genomic_accession=121333654, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73010, CurrentID=0, Status='live', Symbol='Gpr22', Aliases='2900068K05Rik, AW061316', description='G protein-coupled receptor 22', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31706867, end_position_on_the_genomic_accession=31713926, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  208777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208777, CurrentID=0, Status='live', Symbol='Sned1', Aliases='6720455I24Rik, AI197264, AI642697, D430044C15Rik, SST3, Snep', description='sushi, nidogen and EGF-like domains 1', other_designations='secreted nidogen domain protein|secreted protein SST-3|secreted protein SST3|stromal nidogen extracellular matrix protein', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93235815, end_position_on_the_genomic_accession=93301065, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  73016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73016, CurrentID=0, Status='live', Symbol='Kremen2', Aliases='2900054E04Rik, Krm2', description='kringle containing transmembrane protein 2', other_designations='dickkopf receptor 2|kringle domain-containing transmembrane protein 2|kringle-containing protein marking the eye and the nose', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23741199, end_position_on_the_genomic_accession=23745829, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  73024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73024, CurrentID=0, Status='live', Symbol='Emc7', Aliases='2900064A13Rik, AI451465, ORF3, c11orf3', description='ER membrane protein complex subunit 7', other_designations='UPF0480 protein C15orf24 homolog', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112455014, end_position_on_the_genomic_accession=112472155, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73032, CurrentID=0, Status='live', Symbol='Ttc9b', Aliases='2900074C18Rik', description='tetratricopeptide repeat domain 9B', other_designations='TPR repeat protein 9B', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27653924, end_position_on_the_genomic_accession=27656207, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73046, CurrentID=0, Status='live', Symbol='Glrx5', Aliases='', description='glutaredoxin 5', other_designations='glutaredoxin 5 homolog|monothiol glutaredoxin-5', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105032689, end_position_on_the_genomic_accession=105040910, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  73047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73047, CurrentID=0, Status='live', Symbol='Camk2n2', Aliases='2900075A18Rik', description='calcium/calmodulin-dependent protein kinase II inhibitor 2', other_designations='caM-KII inhibitory protein|caM-KIIN', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20619215, end_position_on_the_genomic_accession=20621278, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  73061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73061, CurrentID=0, Status='live', Symbol='Cldn34c1', Aliases='', description='claudin 34C1', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=123103025, end_position_on_the_genomic_accession=123145654, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73062, CurrentID=0, Status='live', Symbol='Ppp1r16a', Aliases='2900084E10Rik, Mypt3, R75527', description='protein phosphatase 1, regulatory (inhibitor) subunit 16A', other_designations='mFLJ00392 protein|myosin phosphatase targeting subunit 3|myosin phosphatase, target subunit 3', map_location='15 36.25 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76671620, end_position_on_the_genomic_accession=76694919, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  73067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73067, CurrentID=0, Status='live', Symbol='Tmem192', Aliases='3110005G23Rik', description='transmembrane protein 192', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=64947160, end_position_on_the_genomic_accession=64969036, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73068, CurrentID=0, Status='live', Symbol='Fut11', Aliases='3110009G03Rik, AI317183', description='fucosyltransferase 11', other_designations='fuc-TXI|fucT-XI|fucosyltransferase XI|galactoside 3-L-fucosyltransferase 11', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20694968, end_position_on_the_genomic_accession=20700197, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73072, CurrentID=0, Status='live', Symbol='Prr36', Aliases='', description='proline rich 36', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4209478, end_position_on_the_genomic_accession=4217548, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73075, CurrentID=0, Status='live', Symbol='Ppil6', Aliases='2900084F20Rik', description='peptidylprolyl isomerase (cyclophilin)-like 6', other_designations='PPIase|cyclophilin-like protein PPIL6|rotamase PPIL6', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41490439, end_position_on_the_genomic_accession=41514295, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  73078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73078, CurrentID=0, Status='live', Symbol='Pmpcb', Aliases='3110004O18Rik, MPP11, MPPB, MPPP52', description='peptidase (mitochondrial processing) beta', other_designations='P-52|beta-MPP', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21737160, end_position_on_the_genomic_accession=21757152, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  73086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73086, CurrentID=0, Status='live', Symbol='Rps6ka5', Aliases='3110005L17Rik, 6330404E13Rik, AI854034, MSK1, MSPK1, RLPK', description='ribosomal protein S6 kinase, polypeptide 5', other_designations='90 kDa ribosomal protein S6 kinase 5|RLSK|RSK-like protein kinase|S6K-alpha-5|nuclear mitogen- and stress-activated protein kinase 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100548433, end_position_on_the_genomic_accession=100725382, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  73090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73090, CurrentID=0, Status='live', Symbol='2900092C05Rik', Aliases='', description='RIKEN cDNA 2900092C05 gene', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12512190, end_position_on_the_genomic_accession=12560756, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73094, CurrentID=0, Status='live', Symbol='Sgip1', Aliases='3110007P09Rik, SGIP1alpha, gag', description='SH3-domain GRB2-like (endophilin) interacting protein 1', other_designations='SH3-domain GRB2-like interacting protein 1 alpha|endophilin-3-interacting protein', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=102759167, end_position_on_the_genomic_accession=102972701, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  73095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73095, CurrentID=0, Status='live', Symbol='Slc25a42', Aliases='2900084M01Rik', description='solute carrier family 25, member 42', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70184340, end_position_on_the_genomic_accession=70212281, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73102, CurrentID=0, Status='live', Symbol='Slc22a23', Aliases='3110004L20Rik', description='solute carrier family 22, member 23', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34179158, end_position_on_the_genomic_accession=34345487, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  73103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73103, CurrentID=0, Status='live', Symbol='3110009E18Rik', Aliases='', description='RIKEN cDNA 3110009E18 gene', other_designations='', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120121115, end_position_on_the_genomic_accession=120188190, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  73106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73106, CurrentID=0, Status='live', Symbol='Prss57', Aliases='2900092M14Rik, GLGL782, Prssl1, UNQ782', description='protease, serine 57', other_designations='protease, serine-like 1|serine protease 1-like protein 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79781474, end_position_on_the_genomic_accession=79795605, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  73112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73112, CurrentID=0, Status='live', Symbol='Abracl', Aliases='1700028I04Rik, 3110003A17Rik', description='ABRA C-terminal like', other_designations='ABRA C-terminal-like protein', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18011260, end_position_on_the_genomic_accession=18023363, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73121, CurrentID=0, Status='live', Symbol='Fam101a', Aliases='3110032G18Rik, cfm, cfm2', description='family with sequence similarity 101, member A', other_designations='protein FAM101A|refilinA|regulator of filamin protein A', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=125003475, end_position_on_the_genomic_accession=125012547, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73122, CurrentID=0, Status='live', Symbol='Tgfbrap1', Aliases='3110018K12Rik, AU024090, AU042859, AW742551, Trap1', description='transforming growth factor, beta receptor associated protein 1', other_designations='TGF beta receptor associated protein 1|TGF-beta receptor-associated protein 1|TRAP-1|tansforming growth factor, beta receptor associated protein 1', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=43047265, end_position_on_the_genomic_accession=43098622, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  73124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73124, CurrentID=0, Status='live', Symbol='Golim4', Aliases='3110027H23Rik, GIMPC, GPP130, Golph4, P138', description='golgi integral membrane protein 4', other_designations='decapacitation factor 10|golgi phosphoprotein 4', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=75876183, end_position_on_the_genomic_accession=75956949, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  73130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73130, CurrentID=0, Status='live', Symbol='Tmed5', Aliases='3110020O18Rik, 4432412D15Rik', description='transmembrane emp24 protein transport domain containing 5', other_designations='p24 family protein gamma-2|p24gamma2', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108121627, end_position_on_the_genomic_accession=108132625, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73132, CurrentID=0, Status='live', Symbol='Slc25a16', Aliases='3110021G18Rik, GDA, GDC, HGT.1, ML7', description='solute carrier family 25 (mitochondrial carrier, Graves disease autoantigen), member 16', other_designations='mitochondrial solute carrier protein homolog|solute carrier family 25 member 16', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62920633, end_position_on_the_genomic_accession=62946499, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  73133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73133, CurrentID=0, Status='live', Symbol='3110021N24Rik', Aliases='2610002D03Rik', description='RIKEN cDNA 3110021N24 gene', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108719649, end_position_on_the_genomic_accession=108781904, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73137, CurrentID=0, Status='live', Symbol='Prrc1', Aliases='1190002C06Rik, 2310058D16Rik, 3110038B19Rik, 9430085A19Rik', description='proline-rich coiled-coil 1', other_designations='proline-rich and coiled-coil-containing protein 1', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=57354733, end_position_on_the_genomic_accession=57392956, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  73139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73139, CurrentID=0, Status='live', Symbol='Cenpv', Aliases='3110013H01Rik, AA671303, Prr6', description='centromere protein V', other_designations='CENP-V|proline-rich polypeptide 6|proline-rich protein 6', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62524944, end_position_on_the_genomic_accession=62539502, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  73144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73144, CurrentID=0, Status='live', Symbol='3110039I08Rik', Aliases='2610203C20Rik, AI848448, D230004N17Rik, D3Mgi9, Mir100hg', description='RIKEN cDNA 3110039I08 gene', other_designations='mir-100-let-7a-2 cluster host', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=41581339, end_position_on_the_genomic_accession=41592487, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  73149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73149, CurrentID=0, Status='live', Symbol='Clec4a3', Aliases='3110037K17Rik, Dcir3', description='C-type lectin domain family 4, member a3', other_designations='', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122952515, end_position_on_the_genomic_accession=122969878, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  73158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73158, CurrentID=0, Status='live', Symbol='Larp1', Aliases='1810024J12Rik, 3110040D16Rik, Larp, mKIAA0731', description='La ribonucleoprotein domain family, member 1', other_designations='la related protein', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58009064, end_position_on_the_genomic_accession=58062034, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  73162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73162, CurrentID=0, Status='live', Symbol='Otud3', Aliases='3110030K17Rik', description='OTU domain containing 3', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138891063, end_position_on_the_genomic_accession=138913947, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73166, CurrentID=0, Status='live', Symbol='Tm7sf2', Aliases='3110041O18Rik, ANG1, C14SR, Dhcr14', description='transmembrane 7 superfamily member 2', other_designations='C-14 sterol reductase|delta-14-SR|sterol C14-reductase', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6062821, end_position_on_the_genomic_accession=6067850, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  73167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73167, CurrentID=0, Status='live', Symbol='Arhgap8', Aliases='3110043J09Rik', description='Rho GTPase activating protein 8', other_designations='rho-type GTPase-activating protein 8', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84720052, end_position_on_the_genomic_accession=84772857, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  73172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73172, CurrentID=0, Status='live', Symbol='Exo5', Aliases='3110037I16Rik, AV297100, Dem1, Exo V, mExo5', description='exonuclease 5', other_designations='defects in morphology 1 homolog|defects in morphology protein 1 homolog|probable exonuclease V', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120921202, end_position_on_the_genomic_accession=120925005, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  73173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73173, CurrentID=0, Status='live', Symbol='Pcdh18', Aliases='3110038E07Rik, BB095589, PCDH68L', description='protocadherin 18', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=49743291, end_position_on_the_genomic_accession=49757382, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  73174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73174, CurrentID=0, Status='live', Symbol='Tbkbp1', Aliases='3110043L15Rik, ProSAPiP2, SINTBAD', description='TBK1 binding protein 1', other_designations='TBK1-binding protein 1|tbkbp1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97136171, end_position_on_the_genomic_accession=97151490, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  73178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73178, CurrentID=0, Status='live', Symbol='Wasl', Aliases='2900021I12Rik, 3110031I02Rik, N-WASP', description='Wiskott-Aldrich syndrome-like (human)', other_designations='', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24613810, end_position_on_the_genomic_accession=24664995, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  73181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73181, CurrentID=0, Status='live', Symbol='Nfatc4', Aliases='3110041H08Rik, AW107667, AW546455, Nfat3', description='nuclear factor of activated T cells, cytoplasmic, calcineurin dependent 4', other_designations='NF-AT3|NF-ATc4|NFATc4|T-cell transcription factor NFAT3|nfatc4|nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 4|transcription complex subunit NF-ATc4', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55824795, end_position_on_the_genomic_accession=55833943, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  73182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73182, CurrentID=0, Status='live', Symbol='Pear1', Aliases='3110045G13Rik, Jedi-1, Megf12', description='platelet endothelial aggregation receptor 1', other_designations='Jedi|jagged and Delta protein|mPEAR1|multiple EGF-like domains protein 12|multiple epidermal growth factor-like domains protein 12|protein Jedi', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87749097, end_position_on_the_genomic_accession=87769111, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  73191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73191, CurrentID=0, Status='live', Symbol='Fezf1', Aliases='3110069A13Rik, Fez, Zfp312-like, fez-like', description='Fez family zinc finger 1', other_designations='', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=23245044, end_position_on_the_genomic_accession=23255368, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  73192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73192, CurrentID=0, Status='live', Symbol='Xpot', Aliases='1110004L07Rik, 3110065H13Rik, AI452076, C79645, EXPORTIN-T', description='exportin, tRNA (nuclear export receptor for tRNAs)', other_designations='exportin(tRNA)|tRNA exportin', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121587380, end_position_on_the_genomic_accession=121626316, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  73205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73205, CurrentID=0, Status='live', Symbol='3110043O21Rik', Aliases='AI840585', description='RIKEN cDNA 3110043O21 gene', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=35191282, end_position_on_the_genomic_accession=35226018, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  73212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73212, CurrentID=0, Status='live', Symbol='3110082I17Rik', Aliases='AW411904', description='RIKEN cDNA 3110082I17 gene', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139359739, end_position_on_the_genomic_accession=139460534, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  73218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73218, CurrentID=0, Status='live', Symbol='Sppl2b', Aliases='3110056O03Rik, AW550292, PSL1', description='signal peptide peptidase like 2B', other_designations='SPP-like 2B|protein SPP-like 2B|protein SPPL2b', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80855248, end_position_on_the_genomic_accession=80868708, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  73225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73225, CurrentID=0, Status='live', Symbol='Fam118a', Aliases='', description='family with sequence similarity 118, member A', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85037015, end_position_on_the_genomic_accession=85116933, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  73230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73230, CurrentID=0, Status='live', Symbol='Bmper', Aliases='3110056H04Rik, CV-2, Crim3, Cv2', description='BMP-binding endothelial regulator', other_designations='bone morphogenetic protein-binding endothelial cell precursor-derived regulator|crossveinless-2', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=23223076, end_position_on_the_genomic_accession=23485215, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  73233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73233, CurrentID=0, Status='live', Symbol='Zfp942', Aliases='3110048L19Rik', description='zinc finger protein 942', other_designations='', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21926960, end_position_on_the_genomic_accession=21962464, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  73234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73234, CurrentID=0, Status='live', Symbol='3110079O15Rik', Aliases='Snorc', description='RIKEN cDNA 3110079O15 gene', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87470216, end_position_on_the_genomic_accession=87475482, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73242, CurrentID=0, Status='live', Symbol='Atat1', Aliases='0610011P08Rik, 2610008K08Rik, 2610110G12Rik, 3110080J08Rik, Mec17, TAT', description='alpha tubulin acetyltransferase 1', other_designations='Novel DUF738 containing protein (2610110G12Rik)|TAT|acetyltransferase mec-17 homolog|alpha-TAT|alpha-TAT1', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35897595, end_position_on_the_genomic_accession=35910093, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  73244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73244, CurrentID=0, Status='live', Symbol='Prl8a1', Aliases='1600017L04Rik, 3830403L08Rik, Ghd16, PLP-Cd, Plpcd, Prlpc4', description='prolactin family 8, subfamily a, member 1', other_designations='growth hormone d16|placental prolactin-like protein C delta|prolactin like protein C delta|prolactin-like protein C 4', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27573922, end_position_on_the_genomic_accession=27582171, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73246, CurrentID=0, Status='live', Symbol='Rassf6', Aliases='1600016B17Rik', description='Ras association (RalGDS/AF-6) domain family member 6', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90601875, end_position_on_the_genomic_accession=90640737, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  73247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73247, CurrentID=0, Status='live', Symbol='Mrgbp', Aliases='1600027N09Rik, AW060503, C80444', description='MRG/MORF4L binding protein', other_designations='MRG-binding protein', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180581222, end_position_on_the_genomic_accession=180586304, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73250, CurrentID=0, Status='live', Symbol='Ceacam5', Aliases='1600029H12Rik, Psg30', description='carcinoembryonic antigen-related cell adhesion molecule 5', other_designations='pregnancy-specific glycoprotein 30', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17713216, end_position_on_the_genomic_accession=17761154, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  73251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73251, CurrentID=0, Status='live', Symbol='Setd7', Aliases='1600028F23Rik, H3K4MT, KMT7, Set7, Set7/9, mKIAA1717', description='SET domain containing (lysine methyltransferase) 7', other_designations='H3-K4 methyltransferase|H3-K4-HMTase SETD7|SET domain-containing protein 7|histone H3-K4 methyltransferase SETD7', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51515318, end_position_on_the_genomic_accession=51560823, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73254, CurrentID=0, Status='live', Symbol='Ccdc18', Aliases='1700021E15Rik, 4932411G06Rik, NY-SAR-41', description='coiled-coil domain containing 18', other_designations='sarcoma antigen NY-SAR-41', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108132619, end_position_on_the_genomic_accession=108232951, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  73259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73259, CurrentID=0, Status='live', Symbol='Cib4', Aliases='1700041E20Rik', description='calcium and integrin binding family member 4', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30485583, end_position_on_the_genomic_accession=30545837, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  73261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73261, CurrentID=0, Status='live', Symbol='1700037C18Rik', Aliases='2810022L20Rik', description='RIKEN cDNA 1700037C18 gene', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3905798, end_position_on_the_genomic_accession=3908690, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73274, CurrentID=0, Status='live', Symbol='Gpbp1', Aliases='1700034P14Rik, AU019836, D230035M11Rik, Gpbp, mGPBP', description='GC-rich promoter binding protein 1', other_designations='GC-rich promoter-binding protein 1', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=111425680, end_position_on_the_genomic_accession=111490123, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  73284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73284, CurrentID=0, Status='live', Symbol='Ddit4l', Aliases='1700037B15Rik, 1700108M02Rik, REDD2, RTP801L, Smhs1', description='DNA-damage-inducible transcript 4-like', other_designations='DNA-damage-inducible transcript 4-like protein|HIF-1 responsive protein RTP801-like|skeletal muscle hindlimb suspension 1|soleus muscle atrophied after hindlimb suspension protein 1', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=137623672, end_position_on_the_genomic_accession=137628332, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73287, CurrentID=0, Status='live', Symbol='1700040L02Rik', Aliases='', description='RIKEN cDNA 1700040L02 gene', other_designations='', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=68430956, end_position_on_the_genomic_accession=68541919, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73288, CurrentID=0, Status='live', Symbol='Vps50', Aliases='1700034M03Rik, 8430415E05Rik, BLM, C87205, Ccdc132, mKIAA1861', description='VPS50 EARP/GARPII complex subunit', other_designations='EARP/GARPII complex subunit VPS50|bcl2-like protein blm|coiled-coil domain containing 132|coiled-coil domain-containing protein 132', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3498393, end_position_on_the_genomic_accession=3603531, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  111346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111346, CurrentID=0, Status='live', Symbol='Cm', Aliases='', description='coloboma deletion region', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  73293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73293, CurrentID=0, Status='live', Symbol='Ccdc103', Aliases='1700039D13Rik', description='coiled-coil domain containing 103', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102881244, end_position_on_the_genomic_accession=102885215, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73296, CurrentID=0, Status='live', Symbol='Rhobtb3', Aliases='1700040C17Rik, 2610033K01Rik, 4930503C18Rik, AI317148, AW208826, mKIAA0878', description='Rho-related BTB domain containing 3', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=75869537, end_position_on_the_genomic_accession=75943937, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  73301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73301, CurrentID=0, Status='live', Symbol='Ttc29', Aliases='1700031F13Rik', description='tetratricopeptide repeat domain 29', other_designations='TPR repeat protein 29', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=78213302, end_position_on_the_genomic_accession=78394326, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  73314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73314, CurrentID=0, Status='live', Symbol='Lrrc69', Aliases='1700034K16Rik', description='leucine rich repeat containing 69', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=14623622, end_position_on_the_genomic_accession=14801320, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73316, CurrentID=0, Status='live', Symbol='Calr3', Aliases='1700031L01Rik, 6330586I20Rik, Crt2, cspn', description='calreticulin 3', other_designations='calreticulin-2|calspeculin|calsperin', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72424176, end_position_on_the_genomic_accession=72443791, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  73318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73318, CurrentID=0, Status='live', Symbol='Ube2d2b', Aliases='1700013N18Rik', description='ubiquitin-conjugating enzyme E2D 2B', other_designations='E2 ubiquitin-conjugating enzyme D2B|ubiquitin carrier protein D2B|ubiquitin-protein ligase D2B', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107830162, end_position_on_the_genomic_accession=107831777, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  73324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73324, CurrentID=0, Status='live', Symbol='Clhc1', Aliases='1700034F02Rik', description='clathrin heavy chain linker domain containing 1', other_designations='', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29547950, end_position_on_the_genomic_accession=29578360, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  73327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73327, CurrentID=0, Status='live', Symbol='Pradc1', Aliases='1700040I03Rik, Pap21', description='protease-associated domain containing 1', other_designations='protease-associated domain-containing protein of 21 kDa', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85446785, end_position_on_the_genomic_accession=85451970, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  73329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73329, CurrentID=0, Status='live', Symbol='1700040F15Rik', Aliases='', description='RIKEN cDNA 1700040F15 gene', other_designations='', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=60241398, end_position_on_the_genomic_accession=60297384, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73332, CurrentID=0, Status='live', Symbol='Ccdc30', Aliases='1700001O02Rik, 1700041C02Rik, 1700111D19Rik, 4930445I03Rik', description='coiled-coil domain containing 30', other_designations='prefoldin subunit 6-like protein', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119322893, end_position_on_the_genomic_accession=119415521, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  73333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73333, CurrentID=0, Status='live', Symbol='Slc25a31', Aliases='1700034J06Rik, Ant4, Sfec', description='solute carrier family 25 (mitochondrial carrier; adenine nucleotide translocator), member 31', other_designations='ADP,ATP carrier protein 4|ANT 4|adenine nucleotide translocase 4|adenine nucleotide translocator 4|solute carrier family 25 member 31|sperm flagellar energy carrier protein', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40708871, end_position_on_the_genomic_accession=40726094, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73336, CurrentID=0, Status='live', Symbol='Prss44', Aliases='1700036D21Rik, 79H19A, AU043574, AU043597, TESSP-4, Tessp4', description='protease, serine 44', other_designations='protease, serine, 44|serine protease-like 1|testis serine protease 4|testis specific serine protease 4|testis-specific serine protease 4', map_location='9 60.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110813900, end_position_on_the_genomic_accession=110817999, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73338, CurrentID=0, Status='live', Symbol='Itpripl1', Aliases='1700041B20Rik', description='inositol 1,4,5-triphosphate receptor interacting protein-like 1', other_designations='inositol 1,4,5-triphosphate receptor-interacting protein-like 1', map_location='2 61.79 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127138769, end_position_on_the_genomic_accession=127143488, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73340, CurrentID=0, Status='live', Symbol='Nptxr', Aliases='1200009K17Rik, 1700036C17Rik, 5730406O18Rik, AI452369, AI785356, D15Bwg0580e, NPR, Npcd', description='neuronal pentraxin receptor', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79786351, end_position_on_the_genomic_accession=79804709, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73341, CurrentID=0, Status='live', Symbol='Arhgef6', Aliases='1600028C08Rik, 1700038J06Rik, 4930592P22Rik, alpha-PIX', description='Rac/Cdc42 guanine nucleotide exchange factor (GEF) 6', other_designations='Rac/Cdc42 guanine exchange factor (GEF) 6|alpha PIX Rho GTPase guanine nucleotide exchange factor|rac/Cdc42 guanine nucleotide exchange factor 6', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=57231485, end_position_on_the_genomic_accession=57338756, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  73344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73344, CurrentID=0, Status='live', Symbol='1700034J05Rik', Aliases='', description='RIKEN cDNA 1700034J05 gene', other_designations='', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146951301, end_position_on_the_genomic_accession=146955058, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73347, CurrentID=0, Status='live', Symbol='Cldn34b2', Aliases='1700042B14Rik', description='claudin 34B2', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155124917, end_position_on_the_genomic_accession=155147849, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73353, CurrentID=0, Status='live', Symbol='Actrt2', Aliases='1700052K15Rik, Arp-T2, Arpm2', description='actin-related protein T2', other_designations='actin related protein M2 (Arpm2)|actin-related protein M2', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154666428, end_position_on_the_genomic_accession=154667867, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  73360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73360, CurrentID=0, Status='live', Symbol='Actrt1', Aliases='1700061J02Rik, Arp-T1', description='actin-related protein T1', other_designations='', map_location='X A3.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=46329007, end_position_on_the_genomic_accession=46330345, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  73363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73363, CurrentID=0, Status='live', Symbol='1700056E22Rik', Aliases='', description='RIKEN cDNA 1700056E22 gene', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=184033032, end_position_on_the_genomic_accession=184033998, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  73368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73368, CurrentID=0, Status='live', Symbol='Col20a1', Aliases='1700051I12Rik', description='collagen, type XX, alpha 1', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180983054, end_position_on_the_genomic_accession=181018363, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  73373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73373, CurrentID=0, Status='live', Symbol='Phospho2', Aliases='1700048E23Rik, AI661517, AU021728, AV006103, Phos2', description='phosphatase, orphan 2', other_designations='', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69789628, end_position_on_the_genomic_accession=69797223, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73376, CurrentID=0, Status='live', Symbol='Tex33', Aliases='1700061J05Rik, Ean57', description='testis expressed 33', other_designations='protein EAN57', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78378400, end_position_on_the_genomic_accession=78395912, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  73379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73379, CurrentID=0, Status='live', Symbol='Dcbld2', Aliases='1700055P21Rik, AW146002, CLCP1, Esdn', description='discoidin, CUB and LCCL domain containing 2', other_designations='endothelial and smooth muscle cell-derived neuropilin-like protein', map_location='16 C1.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58408426, end_position_on_the_genomic_accession=58469745, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  73381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73381, CurrentID=0, Status='live', Symbol='Cmtm2a', Aliases='1700001K04Rik, 1700041N15Rik, 1700063K20Rik, AI449770, ARR19, C32, CKLF3, CKLF4, CKLF5, Cklf, Cklf1, Cklfsf2-1b, Cklfsf2a, UCK-1', description='CKLF-like MARVEL transmembrane domain containing 2A', other_designations='chemokine-like factor super family 2A|chemokine-like factor superfamily 2-1b|chemokine-like factor superfamily member 2A', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104281042, end_position_on_the_genomic_accession=104293181, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73382, CurrentID=0, Status='live', Symbol='Prss52', Aliases='1700049K14Rik, Tesp3', description='protease, serine 52', other_designations='protease, serine, 52|testicular serine protease 3|testicular-specific serine protease 3', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64104323, end_position_on_the_genomic_accession=64113751, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73385, CurrentID=0, Status='live', Symbol='Fam177a', Aliases='1700047I17Rik, 1700047I17Rik1, C14orf24, D730001C10Rik1, Fam177a', description='family with sequence similarity 177, member A', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55124468, end_position_on_the_genomic_accession=55142104, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73388, CurrentID=0, Status='live', Symbol='Bpifa3', Aliases='1700058C13Rik, Splunc3', description='BPI fold containing family A, member 3', other_designations='Novel LBP / BPI / CETP family, N-terminal domain containing protein|short palate, lung and nasal epithelium carcinoma-associated protein 3 homolog', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154130304, end_position_on_the_genomic_accession=154138359, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  73389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73389, CurrentID=0, Status='live', Symbol='Hbp1', Aliases='1700058O05Rik, BE629963, C330012F01Rik, C86454', description='high mobility group box transcription factor 1', other_designations='HMG box transcription factor 1', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31926455, end_position_on_the_genomic_accession=31950535, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  73390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73390, CurrentID=0, Status='live', Symbol='Msl3l2', Aliases='1700060H10Rik', description='male-specific lethal 3-like 2 (Drosophila)', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=56083827, end_position_on_the_genomic_accession=56116880, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73398, CurrentID=0, Status='live', Symbol='Nipsnap3a', Aliases='1700054F22Rik, AI429120', description='nipsnap homolog 3A (C. elegans)', other_designations='', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=52989284, end_position_on_the_genomic_accession=53000854, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  73407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73407, CurrentID=0, Status='live', Symbol='Tepp', Aliases='1700055M20Rik', description='testis, prostate and placenta expressed', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95311570, end_position_on_the_genomic_accession=95326321, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  73412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73412, CurrentID=0, Status='live', Symbol='Nme8', Aliases='1700056P15Rik, Sptrx-2, Txndc3', description='NME/NM23 family member 8', other_designations='spermatid-specific thioredoxin-2|thioredoxin domain containing 3 (spermatozoa)', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19645061, end_position_on_the_genomic_accession=19697760, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  73419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73419, CurrentID=0, Status='live', Symbol='Armt1', Aliases='1700052N19Rik, AW320013, AW536799', description='acidic residue methyltransferase 1', other_designations='UPF0364 protein C6orf211 homolog', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=4429208, end_position_on_the_genomic_accession=4455140, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73420, CurrentID=0, Status='live', Symbol='Ccsap', Aliases='1700054N08Rik, AI853424', description='centriole, cilia and spindle associated protein', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123839973, end_position_on_the_genomic_accession=123860209, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  73422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73422, CurrentID=0, Status='live', Symbol='Prox2', Aliases='1700058C01Rik', description='prospero homeobox 2', other_designations='PROX-2|homeobox prospero-like protein PROX2', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85085810, end_position_on_the_genomic_accession=85110770, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73435, CurrentID=0, Status='live', Symbol='Tex35', Aliases='1700057K13Rik', description='testis expressed 35', other_designations='testis-specific conserved protein of 24 kDa', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=157099141, end_position_on_the_genomic_accession=157110753, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  73442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73442, CurrentID=0, Status='live', Symbol='Hspa12a', Aliases='1700063D12Rik, AI118035, AI840429, AW048913, AW556406, D5Mgi40, mKIAA0417', description='heat shock protein 12A', other_designations='heat shock 70kDa protein 12A', map_location='19 D2-D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58795750, end_position_on_the_genomic_accession=58943545, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  73447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73447, CurrentID=0, Status='live', Symbol='Wdr13', Aliases='1700060B08Rik, 5730411P10Rik, DXHXS7467e, W51679, mMg21', description='WD repeat domain 13', other_designations='memory-related protein', map_location='X 3.67 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8123301, end_position_on_the_genomic_accession=8132858, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73451, CurrentID=0, Status='live', Symbol='Zfp763', Aliases='1700065O13Rik, AI839635', description='zinc finger protein 763', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33016864, end_position_on_the_genomic_accession=33033381, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  73453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73453, CurrentID=0, Status='live', Symbol='1700067K01Rik', Aliases='', description='RIKEN cDNA 1700067K01 gene', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84000496, end_position_on_the_genomic_accession=84004770, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73456, CurrentID=0, Status='live', Symbol='Izumo1', Aliases='1700058F15Rik, Izumo', description='izumo sperm-egg fusion 1', other_designations='OBF|oocyte binding/fusion factor|sperm-specific protein Izumo', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45621811, end_position_on_the_genomic_accession=45627242, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  73458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73458, CurrentID=0, Status='live', Symbol='Aldh3b3', Aliases='1700055N04Rik', description='aldehyde dehydrogenase 3 family, member B3', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3958733, end_position_on_the_genomic_accession=3972419, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  73463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73463, CurrentID=0, Status='live', Symbol='Als2cr11', Aliases='1700052H20Rik, 4930408G06Rik', description='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 11 (human)', other_designations='', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59050506, end_position_on_the_genomic_accession=59095015, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  73466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73466, CurrentID=0, Status='live', Symbol='Ms4a13', Aliases='1700060E18Rik, Tetm4.2', description='membrane-spanning 4-domains, subfamily A, member 13', other_designations='testis-expressed transmembrane 4|testis-expressed transmembrane protein 4.2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11169418, end_position_on_the_genomic_accession=11196755, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73467, CurrentID=0, Status='live', Symbol='1700066M21Rik', Aliases='AI428479', description='RIKEN cDNA 1700066M21 gene', other_designations='', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=57377594, end_position_on_the_genomic_accession=57385423, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73469, CurrentID=0, Status='live', Symbol='Rnf38', Aliases='1700065B19Rik, 2610202O07Rik, AA673263, Oip1', description='ring finger protein 38', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44126212, end_position_on_the_genomic_accession=44233643, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  73470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73470, CurrentID=0, Status='live', Symbol='Kif2b', Aliases='1700063D03Rik', description='kinesin family member 2B', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=91575286, end_position_on_the_genomic_accession=91577555, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  73472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73472, CurrentID=0, Status='live', Symbol='Spata18', Aliases='1700067I02Rik', description='spermatogenesis associated 18', other_designations='spermatogenesis associated 18 homolog|spermatogenesis-associated protein 18', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73651380, end_position_on_the_genomic_accession=73679484, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  73473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73473, CurrentID=0, Status='live', Symbol='Iws1', Aliases='1700069O15Rik, 3000008H23, AL117927, AW214353', description='IWS1 homolog (S. cerevisiae)', other_designations='IWS1-like protein', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32067325, end_position_on_the_genomic_accession=32104331, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  73481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73481, CurrentID=0, Status='live', Symbol='1700074P13Rik', Aliases='AV047902, Tryx5', description='RIKEN cDNA 1700074P13 gene', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40920437, end_position_on_the_genomic_accession=40940973, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  73490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73490, CurrentID=0, Status='live', Symbol='Mipol1', Aliases='1700081O04Rik, 6030439O22Rik, AU021838, AW536152, D12Ertd19e', description='mirror-image polydactyly 1', other_designations='mirror-image polydactyly gene 1 homolog', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=57230425, end_position_on_the_genomic_accession=57457241, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  73495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73495, CurrentID=0, Status='live', Symbol='Samt3', Aliases='1700072E05Rik', description='spermatogenesis associated multipass transmembrane protein 3', other_designations='', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=86044199, end_position_on_the_genomic_accession=86047313, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73503, CurrentID=0, Status='live', Symbol='Mbd3l1', Aliases='1700070G05Rik, 1700095H13Rik, Mbd3l', description='methyl-CpG binding domain protein 3-like 1', other_designations='MBD3-like protein 1', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18478330, end_position_on_the_genomic_accession=18485298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73523, CurrentID=0, Status='live', Symbol='Pebp4', Aliases='PEBP-4', description='phosphatidylethanolamine binding protein 4', other_designations='', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69840407, end_position_on_the_genomic_accession=70059918, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  73526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73526, CurrentID=0, Status='live', Symbol='Speer4b', Aliases='1700081O22Rik, SPEER-4B', description='spermatogenesis associated glutamate (E)-rich protein 4B', other_designations='sperm-associated glutamate (E)-rich protein 4b|spermatogenic-associated glutamate (E)-rich protein 4b', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=27495809, end_position_on_the_genomic_accession=27501392, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73532, CurrentID=0, Status='live', Symbol='1700080E11Rik', Aliases='AI429110, Tseg8', description='RIKEN cDNA 1700080E11 gene', other_designations='testis specific expressed protein 8', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105143344, end_position_on_the_genomic_accession=105145082, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73542, CurrentID=0, Status='live', Symbol='Tssk5', Aliases='1700091F14Rik', description='testis-specific serine kinase 5', other_designations='TSK-5|TSSK-5|testis-specific kinase 5|tssk5', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76371958, end_position_on_the_genomic_accession=76375191, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73545, CurrentID=0, Status='live', Symbol='1700094D03Rik', Aliases='', description='RIKEN cDNA 1700094D03 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90062796, end_position_on_the_genomic_accession=90068347, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73547, CurrentID=0, Status='live', Symbol='Dusp21', Aliases='1700094E07Rik, DSP21', description='dual specificity phosphatase 21', other_designations='novel dual specificity phosphatase, catalytic domain containing protein', map_location='X A1.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=18145870, end_position_on_the_genomic_accession=18146694, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  73569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73569, CurrentID=0, Status='live', Symbol='Vgll3', Aliases='1700110N18Rik, 4832416J22, C80713, Vgl-3, Vito-2', description='vestigial like 3 (Drosophila)', other_designations='vgll3', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=65815017, end_position_on_the_genomic_accession=65866368, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73582, CurrentID=0, Status='live', Symbol='Camkmt', Aliases='1700106N22Rik, AI480743, AV099781, CLNMT, CaM KMT', description='calmodulin-lysine N-methyltransferase', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=85090592, end_position_on_the_genomic_accession=85458589, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  73598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73598, CurrentID=0, Status='live', Symbol='1700001O22Rik', Aliases='1700113K14Rik, 4930549G18Rik', description='RIKEN cDNA 1700001O22 gene', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30795563, end_position_on_the_genomic_accession=30803724, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73603, CurrentID=0, Status='live', Symbol='Trp53tg5', Aliases='1700126L10Rik', description='transformation related protein 53 target 5', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164467903, end_position_on_the_genomic_accession=164478812, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73608, CurrentID=0, Status='live', Symbol='Marveld3', Aliases='1810006A16Rik, AI642133, Marvd3, Mrvldc3', description='MARVEL (membrane-associating) domain containing 3', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109947909, end_position_on_the_genomic_accession=109962205, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  73610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73610, CurrentID=0, Status='live', Symbol='Zfp433', Aliases='1700123A16Rik', description='zinc finger protein 433', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81704194, end_position_on_the_genomic_accession=81721976, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73614, CurrentID=0, Status='live', Symbol='Rhox13', Aliases='1700123J19Rik', description='reproductive homeobox 13', other_designations='novel homeobox domain-containing protein|reproductive homeobox protein 13', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38120840, end_position_on_the_genomic_accession=38129966, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73616, CurrentID=0, Status='live', Symbol='Pldi', Aliases='1700125F08Rik, Poldi', description='polymorphic derived intron containing', other_designations='polymorphic derived intron-containing', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60928226, end_position_on_the_genomic_accession=60938132, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73626, CurrentID=0, Status='live', Symbol='1810009J06Rik', Aliases='', description='RIKEN cDNA 1810009J06 gene', other_designations='trypsinogen 5', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40964772, end_position_on_the_genomic_accession=40968427, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73634, CurrentID=0, Status='live', Symbol='1700125H20Rik', Aliases='', description='RIKEN cDNA 1700125H20 gene', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=85140088, end_position_on_the_genomic_accession=85181164, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73635, CurrentID=0, Status='live', Symbol='Ptges3l', Aliases='1110069E20Rik, 1700113I22Rik', description='prostaglandin E synthase 3 (cytosolic)-like', other_designations='alanyl-tRNA synthetase domain containing 1|prostaglandin E synthase 3-like', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101418801, end_position_on_the_genomic_accession=101425333, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73647, CurrentID=0, Status='live', Symbol='Capn9', Aliases='2200003B16Rik, GC36, nCL-4, nCL4', description='calpain 9', other_designations='digestive tract-specific calpain|new calpain 4', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124576062, end_position_on_the_genomic_accession=124618731, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  73649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73649, CurrentID=0, Status='live', Symbol='Cybrd1', Aliases='2210407P13Rik, Dcytb', description='cytochrome b reductase 1', other_designations='duodenal cytochrome b', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71118054, end_position_on_the_genomic_accession=71142926, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73656, CurrentID=0, Status='live', Symbol='Ms4a6c', Aliases='2200009H22Rik, 2210417N07Rik', description='membrane-spanning 4-domains, subfamily A, member 6C', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11469041, end_position_on_the_genomic_accession=11487069, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  73658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73658, CurrentID=0, Status='live', Symbol='Spns1', Aliases='2210013K02Rik, Spin1, Spinl', description='spinster homolog 1', other_designations='spinster-like protein', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126370060, end_position_on_the_genomic_accession=126377934, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  73660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73660, CurrentID=0, Status='live', Symbol='Cabp4', Aliases='2410038D05Rik', description='calcium binding protein 4', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4135423, end_position_on_the_genomic_accession=4144080, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  73666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73666, CurrentID=0, Status='live', Symbol='Thoc3', Aliases='2410044K02Rik, AL033344', description='THO complex 3', other_designations='tho3', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54458837, end_position_on_the_genomic_accession=54468840, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  73667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73667, CurrentID=0, Status='live', Symbol='2410004P03Rik', Aliases='', description='RIKEN cDNA 2410004P03 gene', other_designations='', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=17004958, end_position_on_the_genomic_accession=17011734, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73668, CurrentID=0, Status='live', Symbol='Ttc21b', Aliases='2410066K11Rik, Thm1, aln, mKIAA1992', description='tetratricopeptide repeat domain 21B', other_designations='TPR repeat protein 21B|line 158|tetratricopeptide repeat-containing hedgehog modulator 1', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=66184327, end_position_on_the_genomic_accession=66256675, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  73670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73670, CurrentID=0, Status='live', Symbol='Defb30', Aliases='2410125J01Rik, 4930449O14Rik, BD-30, mBD-30', description='defensin beta 30', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63034087, end_position_on_the_genomic_accession=63037846, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73671, CurrentID=0, Status='live', Symbol='Sult6b1', Aliases='2410078J06Rik, AV101767, ST6B1', description='sulfotransferase family, cytosolic, 6B, member 1', other_designations='Thyroxine sulfotransferase', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78883938, end_position_on_the_genomic_accession=78906992, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73673, CurrentID=0, Status='live', Symbol='Rec114', Aliases='2410076I21Rik, 4930527A11Rik', description='REC114 meiotic recombination protein', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58652689, end_position_on_the_genomic_accession=58741563, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73674, CurrentID=0, Status='live', Symbol='Wdr75', Aliases='1300003A18Rik, 2410118I19Rik, C77608', description='WD repeat domain 75', other_designations='', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=45795153, end_position_on_the_genomic_accession=45823619, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  73677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73677, CurrentID=0, Status='live', Symbol='Psma8', Aliases='2410072D24Rik', description='proteasome (prosome, macropain) subunit, alpha type, 8', other_designations='proteasome alpha 8 subunit', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=14706135, end_position_on_the_genomic_accession=14762301, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73679, CurrentID=0, Status='live', Symbol='Tex19.1', Aliases='2410081M02Rik, Tex19, Tex19a, mTex19.1', description='testis expressed gene 19.1', other_designations='testis expressed gene 19|testis-expressed protein 19A|testis-expressed sequence 19A', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121146143, end_position_on_the_genomic_accession=121148314, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73680, CurrentID=0, Status='live', Symbol='Zbtb8a', Aliases='2410081M15Rik, AI480941', description='zinc finger and BTB domain containing 8a', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129353632, end_position_on_the_genomic_accession=129378028, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73681, CurrentID=0, Status='live', Symbol='Trmt11', Aliases='2410075D05Rik, 3110045I18Rik, AW213713', description='tRNA methyltransferase 11', other_designations="tRNA guanosine-2'-O-methyltransferase TRM11 homolog|tRNA methyltransferase 11 homolog", map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=30534225, end_position_on_the_genomic_accession=30602147, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  73683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73683, CurrentID=0, Status='live', Symbol='Atg16l2', Aliases='2410118P20Rik', description='autophagy related 16-like 2 (S. cerevisiae)', other_designations='APG16-like 2|Atg16L2 alpha|Atg16L2 beta|autophagy related 16 like 2|autophagy-related 16-like 2 alpha|autophagy-related 16-like 2 beta|mFLJ00012 protein', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101286313, end_position_on_the_genomic_accession=101302266, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  230734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230734, CurrentID=0, Status='live', Symbol='Yrdc', Aliases='AV303379, BC023823, IRIP, ITIP', description='yrdC domain containing (E.coli)', other_designations='Ischemia/reperfusion-inducible protein|mIRIP', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124850759, end_position_on_the_genomic_accession=124855242, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73689, CurrentID=0, Status='live', Symbol='Bloc1s2', Aliases='2410089B13Rik, BLOS2a, Bloc1s2', description='biogenesis of lysosomal organelles complex-1, subunit 2', other_designations='BLOC-1 subunit 2|biogenesis of lysosome-related organelles complex-1, subunit 2|biogenesis of lysosome-related organelles complex-1, subunit 2A', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44139246, end_position_on_the_genomic_accession=44146470, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  73690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73690, CurrentID=0, Status='live', Symbol='Glipr1', Aliases='2410114O14Rik, RTVP-1, RTVP1, mRTVP-1', description='GLI pathogenesis-related 1 (glioma)', other_designations='gliPR 1|related to testes-specific, vespid, and pathogenesis proteins', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=111984792, end_position_on_the_genomic_accession=112002619, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  230735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230735, CurrentID=0, Status='live', Symbol='Epha10', Aliases='A330090H18', description='Eph receptor A10', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124881785, end_position_on_the_genomic_accession=124917800, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  73692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73692, CurrentID=0, Status='live', Symbol='2410089E03Rik', Aliases='4732468D17Rik, Hug, Jbts17, b2b012Clo', description='RIKEN cDNA 2410089E03 gene', other_designations='transmembrane protein ENSP00000382582 homolog', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=8169026, end_position_on_the_genomic_accession=8271158, orientation='plus', exon_count=54, OMIM='', no_hdr0=''),
  73693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73693, CurrentID=0, Status='live', Symbol='Dppa4', Aliases='2410091M23Rik, C76608, ECAT15-1', description='developmental pluripotency associated 4', other_designations='embryonic stem cell-associated transcript 15-1 protein', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48283735, end_position_on_the_genomic_accession=48294292, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73694, CurrentID=0, Status='live', Symbol='Ndufaf7', Aliases='2410091C18Rik, AL033374, PRO1853', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex assembly factor 7', other_designations='protein midA homolog, mitochondrial', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78937106, end_position_on_the_genomic_accession=78948052, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  73699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73699, CurrentID=0, Status='live', Symbol='Ppp2r1b', Aliases='2410091N08Rik, AI790395', description='protein phosphatase 2, regulatory subunit A, beta', other_designations='PP2A subunit A isoform PR65-beta|PP2A subunit A isoform R1-beta|protein phosphatase 2 (formerly 2A), regulatory subunit A (PR 65), beta isoform', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50856615, end_position_on_the_genomic_accession=50899313, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  73703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73703, CurrentID=0, Status='live', Symbol='Dppa2', Aliases='2410088E07Rik, C80932, D19Mgi18, ECAT15-2', description='developmental pluripotency associated 2', other_designations='developmental pluripotency-associated 2|embryonic stem cell-associated transcript 15-2 protein|pluripotent embryonic stem cell-related gene 1 protein', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48310274, end_position_on_the_genomic_accession=48319513, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  73707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73707, CurrentID=0, Status='live', Symbol='Gucy2g', Aliases='2410077I05Rik, GC-G', description='guanylate cyclase 2g', other_designations='guanylyl cyclase receptor G', map_location='19 51.18 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=55198098, end_position_on_the_genomic_accession=55242429, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  73708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73708, CurrentID=0, Status='live', Symbol='Dppa3', Aliases='2410075G02Rik, PCG7, PGC7, Stella', description='developmental pluripotency-associated 3', other_designations='compaction-associated protein 1|primordial germ cell 7|primordial germ cell protein 7', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122626424, end_position_on_the_genomic_accession=122630271, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73710, CurrentID=0, Status='live', Symbol='Tubb2b', Aliases='2410129E14Rik, brdp', description='tubulin, beta 2B class IIB', other_designations='', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34127008, end_position_on_the_genomic_accession=34130354, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  73711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73711, CurrentID=0, Status='live', Symbol='Mvb12a', Aliases='1110012M11Rik, Fam125a', description='multivesicular body subunit 12A', other_designations='ESCRT-I complex subunit MVB12A|family with sequence similarity 125, member A', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71542880, end_position_on_the_genomic_accession=71548026, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  73712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73712, CurrentID=0, Status='live', Symbol='Dmkn', Aliases='1110014F24Rik, AW561900, C130074A08, SK30, SK89, cI-36', description='dermokine', other_designations='epidermis-specific secreted protein SK30/SK89', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30763723, end_position_on_the_genomic_accession=30781066, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  73713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73713, CurrentID=0, Status='live', Symbol='Rbm20', Aliases='1110018J23Rik, 2010003H22Rik, AI646761', description='RNA binding motif protein 20', other_designations='RNA-binding motif protein 20|probable RNA-binding protein 20', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53660989, end_position_on_the_genomic_accession=53867083, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  73720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73720, CurrentID=0, Status='live', Symbol='Cst6', Aliases='1110017E11Rik, N28197, ichq', description='cystatin E/M', other_designations='cystatin M|harlequin ichthyosis|harlequin ichthysosis', map_location='19 4.3 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5344705, end_position_on_the_genomic_accession=5349574, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73721, CurrentID=0, Status='live', Symbol='1110017D15Rik', Aliases='Cbe1, Smrp1', description='RIKEN cDNA 1110017D15 gene', other_designations='ciliated bronchial epithelial protein 1|ciliated bronchial epithelium 1|likely ortholog of H. sapiens chromosome 9 open reading frame 24 (C9orf24)|spermatid specific manchette related protein 1|testes development-related NYD-SP22', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41505005, end_position_on_the_genomic_accession=41517333, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73722, CurrentID=0, Status='live', Symbol='Lce1a2', Aliases='1110004E04Rik, Sprrl2', description='late cornified envelope 1A2', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92668613, end_position_on_the_genomic_accession=92670315, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  73723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73723, CurrentID=0, Status='live', Symbol='Sh3bgrl3', Aliases='1110004L05Rik, AU022188', description='SH3 domain binding glutamic acid-rich protein-like 3', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134127406, end_position_on_the_genomic_accession=134128753, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  73724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73724, CurrentID=0, Status='live', Symbol='Mcee', Aliases='1110007A04Rik', description='methylmalonyl CoA epimerase', other_designations='DL-methylmalonyl-CoA racemase', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64392612, end_position_on_the_genomic_accession=64412121, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  73728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73728, CurrentID=0, Status='live', Symbol='Psd', Aliases='1110007H17Rik, A930015K15, Efa6, Efa6al, mKIAA2011, Psd', description='pleckstrin and Sec7 domain containing', other_designations='exchange factor for ARF6|pleckstrin and Sec7 domain containing homolog|pleckstrin homology and SEC7 domain-containing protein 1', map_location='19 38.8 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46312085, end_position_on_the_genomic_accession=46327184, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  73729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73729, CurrentID=0, Status='live', Symbol='Zfp383', Aliases='1110003H10Rik', description='zinc finger protein 383', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29908281, end_position_on_the_genomic_accession=29916813, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73732, CurrentID=0, Status='live', Symbol='Muc16', Aliases='1110008I14Rik, Ca125, Gm1480', description='mucin 16', other_designations='', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18495445, end_position_on_the_genomic_accession=18674821, orientation='minus', exon_count=89, OMIM='', no_hdr0=''),
  73736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73736, CurrentID=0, Status='live', Symbol='Fcf1', Aliases='1110008B24Rik', description='FCF1 rRNA processing protein', other_designations='FCF1 small subunit', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84970930, end_position_on_the_genomic_accession=84983303, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73737, CurrentID=0, Status='live', Symbol='1110008P14Rik', Aliases='C79326', description='RIKEN cDNA 1110008P14 gene', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32379101, end_position_on_the_genomic_accession=32381915, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  73738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73738, CurrentID=0, Status='live', Symbol='Haus7', Aliases='1110020L19Rik, Uchl5ip, Uip1', description='HAUS augmin-like complex, subunit 7', other_designations='26S proteasome-associated UCH37-interacting protein 1|UCH37-interacting protein 1|UCHL5 interacting protein|UCHL5-interacting protein', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73437315, end_position_on_the_genomic_accession=73459029, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  73739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73739, CurrentID=0, Status='live', Symbol='Cby1', Aliases='1110014P06Rik, CBY, PGEA14, Pgea1, arb1', description='chibby homolog 1 (Drosophila)', other_designations='PIGEA-14|PKD2 interactor, Golgi and endoplasmic reticulum-associated 1|PKD2 interactor, golgi and endoplasmic reticulum associated 1|cytosolic leucine-rich protein', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79659166, end_position_on_the_genomic_accession=79667662, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  73744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73744, CurrentID=0, Status='live', Symbol='Man2c1', Aliases='1110025H24Rik', description='mannosidase, alpha, class 2C, member 1', other_designations='alpha-D-mannoside mannohydrolase|cytosolic alpha-D-mannosidase', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57130618, end_position_on_the_genomic_accession=57142213, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  73747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73747, CurrentID=0, Status='live', Symbol='1110034G24Rik', Aliases='', description='RIKEN cDNA 1110034G24 gene', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132686875, end_position_on_the_genomic_accession=132751055, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73748, CurrentID=0, Status='live', Symbol='Gadl1', Aliases='1110027M19Rik, ADC, CSADC', description='glutamate decarboxylase-like 1', other_designations='aspartate 1-decarboxylase|cysteine sulfinic acid decarboxylase|glutamate decarboxylase-like protein 1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=115884644, end_position_on_the_genomic_accession=116076176, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  73750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73750, CurrentID=0, Status='live', Symbol='Whrn', Aliases='1110035G07Rik, AW122018, AW742671, C430046P22Rik, Dfnb31, wi', description='whirlin', other_designations='whirler', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63414909, end_position_on_the_genomic_accession=63496078, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  73754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73754, CurrentID=0, Status='live', Symbol='Thap1', Aliases='4833431A01Rik, AW490810', description='THAP domain containing, apoptosis associated protein 1', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=26158169, end_position_on_the_genomic_accession=26164151, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73779, CurrentID=0, Status='live', Symbol='Prr27', Aliases='4930432K09Rik', description='proline rich 27', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87825697, end_position_on_the_genomic_accession=87846386, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  230753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230753, CurrentID=0, Status='live', Symbol='Thrap3', Aliases='9330151F09Rik, B230333E16Rik, Trap150', description='thyroid hormone receptor associated protein 3', other_designations='thyroid hormone receptor-associated protein complex 150 kDa component', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126164082, end_position_on_the_genomic_accession=126202710, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  208908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208908, CurrentID=0, Status='live', Symbol='Ccdc62', Aliases='AI661708, Gm150', description='coiled-coil domain containing 62', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123930328, end_position_on_the_genomic_accession=123969897, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  73804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73804, CurrentID=0, Status='live', Symbol='Kif2c', Aliases='4930402F02Rik, ESTM5, Knsl6, MCAK, X83316', description='kinesin family member 2C', other_designations='kinesin-like 6|mitotic centromere-associated kinesin', map_location='4 53.42 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117159633, end_position_on_the_genomic_accession=117182624, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  73808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73808, CurrentID=0, Status='live', Symbol='Tex36', Aliases='4930404H21Rik', description='testis expressed 36', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133587024, end_position_on_the_genomic_accession=133602115, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  73809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73809, CurrentID=0, Status='live', Symbol='Satl1', Aliases='4930404K22Rik', description='spermidine/spermine N1-acetyl transferase-like 1', other_designations='', map_location='X E2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=112383207, end_position_on_the_genomic_accession=112406779, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  73813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73813, CurrentID=0, Status='live', Symbol='Fam83e', Aliases='4930403C10Rik', description='family with sequence similarity 83, member E', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45721220, end_position_on_the_genomic_accession=45729492, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73822, CurrentID=0, Status='live', Symbol='Mfsd12', Aliases='C19orf28, F630110N24Rik, Wdt1', description='major facilitator superfamily domain containing 12', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81357469, end_position_on_the_genomic_accession=81366225, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  73825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73825, CurrentID=0, Status='live', Symbol='Ppp1r21', Aliases='1110018J12Rik, AI426045, AW550781, Ccdc128, Klraq1', description='protein phosphatase 1, regulatory subunit 21', other_designations='KLRAQ motif containing 1|KLRAQ motif-containing protein 1|coiled-coil domain containing 128|coiled-coil domain-containing protein 128', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=88530124, end_position_on_the_genomic_accession=88588367, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  73826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73826, CurrentID=0, Status='live', Symbol='Poldip3', Aliases='1110008P04Rik, AA408269, AL022852, C77954, PDIP46, SKAR, mKIAA1649', description='polymerase (DNA-directed), delta interacting protein 3', other_designations='S6K1 Aly/REF-like target|polymerase delta interacting protein 46', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83125976, end_position_on_the_genomic_accession=83149373, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73827, CurrentID=0, Status='live', Symbol='Tmem198b', Aliases='1110012D08Rik', description='transmembrane protein 198b', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128798445, end_position_on_the_genomic_accession=128805461, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73828, CurrentID=0, Status='live', Symbol='Dcaf4', Aliases='1110018E21Rik, Wdr21', description='DDB1 and CUL4 associated factor 4', other_designations='WD repeat domain 21', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83520457, end_position_on_the_genomic_accession=83541994, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  73830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73830, CurrentID=0, Status='live', Symbol='Eif3k', Aliases='1200009C21Rik, Eif3s12', description='eukaryotic translation initiation factor 3, subunit K', other_designations='eIF-3 p25|eIF3k|eukaryotic translation initiation factor 3 subunit 12|eukaryotic translation initiation factor 3, subunit 12', map_location='7 A3-B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28971372, end_position_on_the_genomic_accession=28981889, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73833, CurrentID=0, Status='live', Symbol='Fam98c', Aliases='1110006G06Rik, AI836081, AV026483, B230110F21Rik', description='family with sequence similarity 98, member C', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29152510, end_position_on_the_genomic_accession=29156359, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73834, CurrentID=0, Status='live', Symbol='Atp6v1d', Aliases='1110004P10Rik, Atp6m, VATD, Vma8', description='ATPase, H+ transporting, lysosomal V1 subunit D', other_designations='ATPase, H+ transporting lysosomal (vacuolar proton pump)|ATPase, H+ transporting, V1 subunit D|ATPase, H+ transporting, lysosomal 34kD, V1 subunit D|V-ATPase 28 kDa accessory protein|V-ATPase subunit D|lysosomal 34kDa|vacuolar proton pump subunit D', map_location='12 35.51 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=78842982, end_position_on_the_genomic_accession=78861638, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  73835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73835, CurrentID=0, Status='live', Symbol='Ifitm5', Aliases='1110003J06Rik, AW213665, Bril, DSPA1, Hrmp1, Hrtm1', description='interferon induced transmembrane protein 5', other_designations='bone-restricted interferon induced transmembrane protein-like protein|bone-restricted interferon-induced transmembrane protein-like protein|dispanin subfamily A member 1|fragilis family member 4|fragilis4|haemopoiesis related membrane protein 1', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140948962, end_position_on_the_genomic_accession=140950239, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  73836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73836, CurrentID=0, Status='live', Symbol='Slc35b2', Aliases='1110003M08Rik, AA407792, AI842275, PAPST1, Slc35b1', description='solute carrier family 35, member B2', other_designations='PAPS transporter 1|solute carrier family 35, member B1', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45563879, end_position_on_the_genomic_accession=45567671, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  667666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667666, CurrentID=0, Status='live', Symbol='Zfp600', Aliases='Gm13164, OTTMUSG00000010694', description='zinc finger protein 600', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=146156827, end_position_on_the_genomic_accession=146198756, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  73844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73844, CurrentID=0, Status='live', Symbol='Ankrd45', Aliases='4933409K03Rik', description='ankyrin repeat domain 45', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161142547, end_position_on_the_genomic_accession=161170507, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  73845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73845, CurrentID=0, Status='live', Symbol='Ankrd42', Aliases='4931426M20, 4933417L02Rik, Ikbn', description='ankyrin repeat domain 42', other_designations='', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=92584175, end_position_on_the_genomic_accession=92637143, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  73847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73847, CurrentID=0, Status='live', Symbol='Fam110a', Aliases='1700008J10Rik, 5430432M24Rik, AI747080', description='family with sequence similarity 110, member A', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151946975, end_position_on_the_genomic_accession=151980366, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73852, CurrentID=0, Status='live', Symbol='D3Ertd751e', Aliases='2810009O15Rik, 4930415G15Rik', description='DNA segment, Chr 3, ERATO Doi 751, expressed', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=41742606, end_position_on_the_genomic_accession=41803292, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  73862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73862, CurrentID=0, Status='live', Symbol='4930415F15Rik', Aliases='', description='RIKEN cDNA 4930415F15 gene', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=11463724, end_position_on_the_genomic_accession=11515192, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  73863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73863, CurrentID=0, Status='live', Symbol='4930415O20Rik', Aliases='', description='RIKEN cDNA 4930415O20 gene', other_designations='', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98571004, end_position_on_the_genomic_accession=98589588, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  73866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73866, CurrentID=0, Status='live', Symbol='Fam122c', Aliases='4930432H15Rik, 4930527G05Rik', description='family with sequence similarity 122, member C', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53273419, end_position_on_the_genomic_accession=53331186, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  73873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73873, CurrentID=0, Status='live', Symbol='Fam161a', Aliases='4930430E16Rik', description='family with sequence similarity 161, member A', other_designations='', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23007317, end_position_on_the_genomic_accession=23045743, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  230767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230767, CurrentID=0, Status='live', Symbol='Iqcc', Aliases='BC026978', description='IQ motif containing C', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129615617, end_position_on_the_genomic_accession=129619213, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  73884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73884, CurrentID=0, Status='live', Symbol='Zdbf2', Aliases='4930431J08Rik, 9330107J05Rik', description='zinc finger, DBF-type containing 2', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63200358, end_position_on_the_genomic_accession=63314576, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  73893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73893, CurrentID=0, Status='live', Symbol='Tmem202', Aliases='4930425N13Rik', description='transmembrane protein 202', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59518685, end_position_on_the_genomic_accession=59525501, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105980076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105980076, CurrentID=0, Status='live', Symbol='2310045n01rik-mef2b', Aliases='', description='2310045N01Rik-mef2b readthrough', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  73902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73902, CurrentID=0, Status='live', Symbol='Psmb11', Aliases='5830406J20Rik, beta5t', description='proteasome (prosome, macropain) subunit, beta type, 11', other_designations='proteasome beta 11 subunit|proteasome subunit beta-5t|proteasome subunit beta5t', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54625310, end_position_on_the_genomic_accession=54629556, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  73910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73910, CurrentID=0, Status='live', Symbol='Arhgap18', Aliases='4833419J07Rik, AI314906, AI843501', description='Rho GTPase activating protein 18', other_designations='arhgap18|rho-type GTPase-activating protein 18', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=26695312, end_position_on_the_genomic_accession=26918648, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  73914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73914, CurrentID=0, Status='live', Symbol='Irak3', Aliases='4833428C18Rik, AI563835, IRAK-M', description='interleukin-1 receptor-associated kinase 3', other_designations='IL-1 receptor-associated kinase M|IRAK-3|irak3', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=120141648, end_position_on_the_genomic_accession=120201597, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  73916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73916, CurrentID=0, Status='live', Symbol='Ift57', Aliases='4833420A15Rik, Esrrbl1, Hippi, MHS4R2', description='intraflagellar transport 57', other_designations='HIP1-interacting protein|Vestrogen-related receptor beta like 1|estrogen-related receptor beta like 1|huntingtin-interacting protein-1 protein interactor|intraflagellar transport 57 homolog', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=49699294, end_position_on_the_genomic_accession=49765126, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  73919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73919, CurrentID=0, Status='live', Symbol='Lyrm1', Aliases='1110065L10Rik, 2310004B22Rik, 4930404J24Rik', description='LYR motif containing 1', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119895807, end_position_on_the_genomic_accession=119956203, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  73934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73934, CurrentID=0, Status='live', Symbol='Cldn34c4', Aliases='4930412D23Rik', description='claudin 34C4', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=127721175, end_position_on_the_genomic_accession=127736554, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  73936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73936, CurrentID=0, Status='live', Symbol='Ccdc175', Aliases='4930403N07Rik', description='coiled-coil domain containing 175', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72100973, end_position_on_the_genomic_accession=72185029, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  12323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12323, CurrentID=0, Status='live', Symbol='Camk2b', Aliases='', description='calcium/calmodulin-dependent protein kinase II, beta', other_designations='CaMK II|caM kinase II subunit beta|caMK-II subunit beta', map_location='11 3.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5969642, end_position_on_the_genomic_accession=6065748, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  73940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73940, CurrentID=0, Status='live', Symbol='Hapln2', Aliases='4930401E20Rik, Bral1', description='hyaluronan and proteoglycan link protein 2', other_designations='brain link protein 1', map_location='3 38.78 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88020558, end_position_on_the_genomic_accession=88027511, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  73942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73942, CurrentID=0, Status='live', Symbol='Fam151b', Aliases='4930405M20Rik', description='family with sequence similarity 151, member B', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92449620, end_position_on_the_genomic_accession=92484290, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  73945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73945, CurrentID=0, Status='live', Symbol='Otud4', Aliases='4930431L18Rik, AI449692, D8Ertd69e, mKIAA1046', description='OTU domain containing 4', other_designations='otud4', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79639676, end_position_on_the_genomic_accession=79677755, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  667686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667686, CurrentID=0, Status='live', Symbol='Mif-ps1', Aliases='EG667686, Gm8762', description='macrophage migration inhibitory factor, pseudogene 1', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136789068, end_position_on_the_genomic_accession=136789572, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  73968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73968, CurrentID=0, Status='live', Symbol='4930444F02Rik', Aliases='1700010L24Rik, mpiR', description='RIKEN cDNA 4930444F02 gene', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18797299, end_position_on_the_genomic_accession=18831930, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77864, CurrentID=0, Status='live', Symbol='Ypel2', Aliases='6430570G24, E130113K08Rik', description='yippee-like 2 (Drosophila)', other_designations='ortholog of human yippee-like 2 YPEL2', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86936425, end_position_on_the_genomic_accession=86993849, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  99712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99712, CurrentID=0, Status='live', Symbol='Cept1', Aliases='9930118K05Rik, mCEPT1', description='choline/ethanolaminephosphotransferase 1', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106502260, end_position_on_the_genomic_accession=106547802, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  73988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73988, CurrentID=0, Status='live', Symbol='4930438A08Rik', Aliases='', description='RIKEN cDNA 4930438A08 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58274733, end_position_on_the_genomic_accession=58294290, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  73991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73991, CurrentID=0, Status='live', Symbol='Atl1', Aliases='4930435M24Rik, Adfsp, Fsp1, Spg3, Spg3a', description='atlastin GTPase 1', other_designations='spastic paraplegia 3A homolog', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69893105, end_position_on_the_genomic_accession=69964085, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  73998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=73998, CurrentID=0, Status='live', Symbol='Herc3', Aliases='5730409F18Rik, AI646603, mKIAA0032', description='hect domain and RLD 3', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58831477, end_position_on_the_genomic_accession=58920398, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  74002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74002, CurrentID=0, Status='live', Symbol='Psd2', Aliases='6330404E20Rik, AW125584, EFA6C', description='pleckstrin and Sec7 domain containing 2', other_designations='pleckstrin homology and SEC7 domain-containing protein 2', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35963881, end_position_on_the_genomic_accession=36014715, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74004, CurrentID=0, Status='live', Symbol='Jakmip3', Aliases='6330417G02Rik, AI593226, mKIAA4091', description='janus kinase and microtubule interacting protein 3', other_designations='neuroendocrine long coiled-coil protein 2', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139062607, end_position_on_the_genomic_accession=139083976, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  74006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74006, CurrentID=0, Status='live', Symbol='Dnm1l', Aliases='6330417M19Rik, AI450666, Dlp1, Dnmlp1, Drp1, python', description='dynamin 1-like', other_designations='dymple|dynamin family member proline-rich carboxyl-terminal domain less|dynamin-related protein 1', map_location='16 A2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16312228, end_position_on_the_genomic_accession=16359046, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  74007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74007, CurrentID=0, Status='live', Symbol='Btbd11', Aliases='6330404E16Rik', description='BTB (POZ) domain containing 11', other_designations='BTB/POZ domain-containing protein 11', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85386776, end_position_on_the_genomic_accession=85660292, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  74008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74008, CurrentID=0, Status='live', Symbol='Arsg', Aliases='6330406P08Rik, AI846872', description='arylsulfatase G', other_designations='ASG', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109450845, end_position_on_the_genomic_accession=109573330, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  74011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74011, CurrentID=0, Status='live', Symbol='Slc25a27', Aliases='3632410G24Rik, 9430092A03Rik, D530043E16Rik, Ucp4', description='solute carrier family 25, member 27', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43638210, end_position_on_the_genomic_accession=43667194, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74012, CurrentID=0, Status='live', Symbol='Rap2b', Aliases='4021402C18Rik, AA408554', description='RAP2B, member of RAS oncogene family', other_designations='', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=61364507, end_position_on_the_genomic_accession=61368703, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  74013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74013, CurrentID=0, Status='live', Symbol='Rftn2', Aliases='2700010E02Rik, 3222401M22Rik', description='raftlin family member 2', other_designations='raft-linking protein 2', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55170159, end_position_on_the_genomic_accession=55227065, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  74015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74015, CurrentID=0, Status='live', Symbol='Fcho1', Aliases='3322402E17Rik, N28152', description='FCH domain only 1', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71708387, end_position_on_the_genomic_accession=71725704, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  74016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74016, CurrentID=0, Status='live', Symbol='Phf19', Aliases='3110009G19Rik, 3321402G02Rik', description='PHD finger protein 19', other_designations='polycomb-like protein 3', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34893752, end_position_on_the_genomic_accession=34914077, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74018, CurrentID=0, Status='live', Symbol='Als2', Aliases='3222402C23Rik, 9430073A21Rikcr6, Alsin, mKIAA1563, Als2', description='amyotrophic lateral sclerosis 2 (juvenile)', other_designations='amyotrophic lateral sclerosis 2 (juvenile) homolog|amyotrophic lateral sclerosis 2 protein homolog|amyotrophic lateral sclerosis protein 2 homolog', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59162756, end_position_on_the_genomic_accession=59237231, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  74019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74019, CurrentID=0, Status='live', Symbol='Traf3ip1', Aliases='3930402D05Rik, AU041749, MIP-T3', description='TRAF3 interacting protein 1', other_designations='TNF receptor-associated factor 3 interacting protein 1|intraflagellar transport protein 54 homolog|microtubule-interacting protein associated with TRAF3|microtubule-interacting protein that associates with TRAF3', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91494590, end_position_on_the_genomic_accession=91529307, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  74020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74020, CurrentID=0, Status='live', Symbol='Cpne4', Aliases='3632411M23Rik, 4933406O10Rik', description='copine IV', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=104547291, end_position_on_the_genomic_accession=105034549, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  74022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74022, CurrentID=0, Status='live', Symbol='Glyr1', Aliases='2810419J22Rik, 3930401K13Rik, AW545332, Npac', description='glyoxylate reductase 1 homolog (Arabidopsis)', other_designations='cytokine-like nuclear factor n-pac|nuclear protein NP60', map_location='16 2.49 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5013902, end_position_on_the_genomic_accession=5049910, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74023, CurrentID=0, Status='live', Symbol='Rd3', Aliases='3322402L07Rik, rd-3, rd3', description='retinal degeneration 3', other_designations='', map_location='1 97.09 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191954198, end_position_on_the_genomic_accession=191988282, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74025, CurrentID=0, Status='live', Symbol='Nphp3', Aliases='3632410F03Rik, AI550417, C230078J01, D330020E01Rik, pcy', description='nephronophthisis 3 (adolescent)', other_designations='', map_location='9 56.11 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=104002544, end_position_on_the_genomic_accession=104043811, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  74026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74026, CurrentID=0, Status='live', Symbol='Msl1', Aliases='2810017F12Rik, 4121402D02Rik, 4930463F05Rik, AA682082, Msl-1', description='male-specific lethal 1 homolog (Drosophila)', other_designations='MSL1-like 1|hampin|male-specific lethal 1-like 1|male-specific lethal-1 homolog 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98795468, end_position_on_the_genomic_accession=98807862, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74030, CurrentID=0, Status='live', Symbol='Rin2', Aliases='2010003K16Rik, 4632403N06Rik, AW821980, RASSF4', description='Ras and Rab interactor 2', other_designations='ras interaction/interference protein 2', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=145771727, end_position_on_the_genomic_accession=145887616, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74032, CurrentID=0, Status='live', Symbol='Sdr42e1', Aliases='4632417N05Rik, AI117603', description='short chain dehydrogenase/reductase family 42E, member 1', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=117661399, end_position_on_the_genomic_accession=117673718, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74035, CurrentID=0, Status='live', Symbol='Nol9', Aliases='4632412I24Rik, 6030462G04Rik, AI449622, AW490720', description='nucleolar protein 9', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152039274, end_position_on_the_genomic_accession=152061494, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  74038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74038, CurrentID=0, Status='live', Symbol='Brip1os', Aliases='4632419I22Rik', description='BRCA1 interacting protein C-terminal helicase 1, opposite strand', other_designations='BRCA1 interacting protein C-terminal helicase 1 opposite strand', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  74039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74039, CurrentID=0, Status='live', Symbol='Nfam1', Aliases='4921501M20Rik, Cnaip', description='Nfat activating molecule with ITAM motif 1', other_designations='calcineurin/NFAT-activating ITAM-containing protein', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82996736, end_position_on_the_genomic_accession=83033397, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  74041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74041, CurrentID=0, Status='live', Symbol='Ddias', Aliases='4632434I11Rik, noxin', description='DNA damage-induced apoptosis suppressor', other_designations='nitric oxide-inducible gene protein', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=92857525, end_position_on_the_genomic_accession=92874286, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74042, CurrentID=0, Status='live', Symbol='4921501E09Rik', Aliases='PHF8', description='RIKEN cDNA 4921501E09 gene', other_designations='KIAA1111-like protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33064143, end_position_on_the_genomic_accession=33068058, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  74043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74043, CurrentID=0, Status='live', Symbol='Pex26', Aliases='4632428M11Rik', description='peroxisomal biogenesis factor 26', other_designations='peroxin-26|peroxisome biogenesis factor 26', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121183657, end_position_on_the_genomic_accession=121197906, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74044, CurrentID=0, Status='live', Symbol='Ttf2', Aliases='4632434F22Rik, 8030447N19, AV218430', description='transcription termination factor, RNA polymerase II', other_designations='RNA polymerase II termination factor|transcription release factor 2', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100938860, end_position_on_the_genomic_accession=100969725, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  74048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74048, CurrentID=0, Status='live', Symbol='Vsir', Aliases='4632428N05Rik, Dies1, PD-1H, VISTA', description='V-set immunoregulatory receptor', other_designations='differentiation of ESCs 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60346851, end_position_on_the_genomic_accession=60428381, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74051, CurrentID=0, Status='live', Symbol='Steap2', Aliases='4921538B17Rik, AI930049, AW045895, IPCA-1, IPCA1, PCANAP1, STAMP1, STMP', description='six transmembrane epithelial antigen of prostate 2', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=5664829, end_position_on_the_genomic_accession=5694568, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  74052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74052, CurrentID=0, Status='live', Symbol='Ttc21a', Aliases='4921538N17Rik, AA387572, Thm2', description='tetratricopeptide repeat domain 21A', other_designations='TPR repeat protein 21A|tetratricopeptide repeat-containing hedgehog modulator 2', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119937606, end_position_on_the_genomic_accession=119967795, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  74053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74053, CurrentID=0, Status='live', Symbol='Grip1', Aliases='4931400F03Rik, GRIP, eb', description='glutamate receptor interacting protein 1', other_designations='eye blebs', map_location='10 67.33 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=119453835, end_position_on_the_genomic_accession=120087267, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  74054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74054, CurrentID=0, Status='live', Symbol='4931406B18Rik', Aliases='4930538L19Rik', description='RIKEN cDNA 4931406B18 gene', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43492044, end_position_on_the_genomic_accession=43505969, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74055, CurrentID=0, Status='live', Symbol='Plce1', Aliases='4933403A21Rik, PLCepsilon, Plce, mKIAA1516', description='phospholipase C, epsilon 1', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase epsilon-1|PLC-epsilon-1|phosphoinositide phospholipase C-epsilon-1|phospholipase C-epsilon-1', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38481114, end_position_on_the_genomic_accession=38785105, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  74068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74068, CurrentID=0, Status='live', Symbol='Asz1', Aliases='4933400N19Rik, Gasz, ORF3', description='ankyrin repeat, SAM and basic leucine zipper domain containing 1', other_designations='asz1|germ cell-specific ankyrin, SAM and basic leucine zipper domain containing protein|germ cell-specific ankyrin, SAM and basic leucine zipper domain-containing protein', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=18050964, end_position_on_the_genomic_accession=18109061, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74069, CurrentID=0, Status='live', Symbol='Serpina3a', Aliases='', description='serine (or cysteine) peptidase inhibitor, clade A, member 3A', other_designations='serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 3A|serpin A3A', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104106704, end_position_on_the_genomic_accession=104121897, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74071, CurrentID=0, Status='live', Symbol='Ifltd1', Aliases='4933403M22Rik, AI482300, Lmna-rs1, Lmna1, Lmntd1', description='intermediate filament tail domain containing 1', other_designations='intermediate filament tail domain-containing protein 1|lamin A-related sequence 1|lamin-A-related sequence 1 protein', map_location='6 71.3 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145365142, end_position_on_the_genomic_accession=145450766, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74075, CurrentID=0, Status='live', Symbol='Syce1', Aliases='4933406J07Rik', description='synaptonemal complex central element protein 1', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140777229, end_position_on_the_genomic_accession=140787854, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  74080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74080, CurrentID=0, Status='live', Symbol='Nmnat3', Aliases='4933408N02Rik, PNAT3', description='nicotinamide nucleotide adenylyltransferase 3', other_designations='NMN adenylyltransferase 3|NMN/NaMN adenylyltransferase 3|naMN adenylyltransferase 1|naMN adenylyltransferase 3|nicotinamide mononucleotide adenylyltransferase 3|nicotinamide-nucleotide adenylyltransferase 3|nicotinate-nucleotide adenylyltransferase 1|nicotinate-nucleotide adenylyltransferase 3', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98287432, end_position_on_the_genomic_accession=98420442, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74081, CurrentID=0, Status='live', Symbol='Cep350', Aliases='4933409L06Rik, 6430546F08Rik', description='centrosomal protein 350', other_designations='', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155844964, end_position_on_the_genomic_accession=155974435, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  74087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74087, CurrentID=0, Status='live', Symbol='Slc7a13', Aliases='0610009O04Rik, AGT-1, AGT1, XAT2', description='solute carrier family 7, (cationic amino acid transporter, y+ system) member 13', other_designations='X-amino acid transporter 2|aspartate/glutamate transporter 1|sodium-independent aspartate/glutamate transporter 1', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=19818727, end_position_on_the_genomic_accession=19842213, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74088, CurrentID=0, Status='live', Symbol='Them7', Aliases='0610012H03Rik', description='thioesterase superfamily member 7', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=105224254, end_position_on_the_genomic_accession=105379860, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74090, CurrentID=0, Status='live', Symbol='Paqr5', Aliases='0610010I15Rik, AV002411, BB115488, mPRg', description='progestin and adipoQ receptor family member V', other_designations='mPR gamma', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=61953738, end_position_on_the_genomic_accession=62026790, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  74091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74091, CurrentID=0, Status='live', Symbol='Npl', Aliases='0610033B02Rik', description='N-acetylneuraminate pyruvate lyase', other_designations='N-acetylneuraminate pyruvate-lyase|N-acetylneuraminic acid aldolase|NALase|sialate lyase|sialate-pyruvate lyase|sialic acid aldolase|sialic acid lyase', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153503016, end_position_on_the_genomic_accession=153549714, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74094, CurrentID=0, Status='live', Symbol='Tjap1', Aliases='0610041D19Rik, AI415281, AW121008, Pilt, Tjp4', description='tight junction associated protein 1', other_designations='hdlg-binding protein Pilt|protein incorporated later into tight junctions|tight junction protein 4 (peripheral)', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46257851, end_position_on_the_genomic_accession=46283065, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74096, CurrentID=0, Status='live', Symbol='Hvcn1', Aliases='0610039P13Rik, AI450555, BTS, HV1, Vsop., mVSOP', description='hydrogen voltage-gated channel 1', other_designations='voltage sensor domain-only protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122206807, end_position_on_the_genomic_accession=122242297, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74097, CurrentID=0, Status='live', Symbol='Pop7', Aliases='0610037N12Rik, AI852017, Rpp20', description='processing of precursor 7, ribonuclease P family, (S. cerevisiae)', other_designations='Pop7 homolog|RNaseP protein p20', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137501439, end_position_on_the_genomic_accession=137502429, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  74098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74098, CurrentID=0, Status='live', Symbol='0610037L13Rik', Aliases='1110008H16Rik', description='RIKEN cDNA 0610037L13 gene', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107889786, end_position_on_the_genomic_accession=107897802, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  74100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74100, CurrentID=0, Status='live', Symbol='Arpp21', Aliases='0710001E13Rik, AI853636, ARPP-21, D9Bwg1012e, R3hdm3, Tarpp', description='cyclic AMP-regulated phosphoprotein, 21', other_designations='protein phosphatase 1, regulatory (inhibitory) subunit 1C|regulator of calmodulin signaling|thymocyte ARPP|thymocyte cAMP-regulated phosphoprotein', map_location='9 62.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=112065091, end_position_on_the_genomic_accession=112236022, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  74102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74102, CurrentID=0, Status='live', Symbol='Slc35a5', Aliases='1010001J06Rik, AU021179, BB097433, D16Ertd450e, D730043G07Rik', description='solute carrier family 35, member A5', other_designations='', map_location='16 29.44 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45139573, end_position_on_the_genomic_accession=45158673, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74103, CurrentID=0, Status='live', Symbol='Nebl', Aliases='1200007O21Rik, A630080F05Rik, BB140644, D830029A09Rik, Lnebl', description='nebulette', other_designations='actin-binding Z-disk protein', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=17343022, end_position_on_the_genomic_accession=17731068, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  74104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74104, CurrentID=0, Status='live', Symbol='Abcb6', Aliases='1200005B17Rik', description='ATP-binding cassette, sub-family B (MDR/TAP), member 6', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75171640, end_position_on_the_genomic_accession=75180283, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  74105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74105, CurrentID=0, Status='live', Symbol='Gga2', Aliases='1200007E24Rik, mKIAA1080', description='golgi associated, gamma adaptin ear containing, ARF binding protein 2', other_designations='gamma-adaptin-related protein 2|golgi-localized, gamma ear-containing, ARF-binding protein 2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121986722, end_position_on_the_genomic_accession=122021198, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74106, CurrentID=0, Status='live', Symbol='Dcaf6', Aliases='1200006M05Rik, Iqwd1, PC326', description='DDB1 and CUL4 associated factor 6', other_designations='IQ motif and WD repeat-containing protein 1|IQ motif and WD repeats 1|NRIP|nuclear receptor interaction protein', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165329500, end_position_on_the_genomic_accession=165460553, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  74107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74107, CurrentID=0, Status='live', Symbol='Cep55', Aliases='1200008O12Rik, 2700032M20Rik', description='centrosomal protein 55', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38054981, end_position_on_the_genomic_accession=38074425, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74108, CurrentID=0, Status='live', Symbol='Parn', Aliases='1200003I18Rik, DAN', description='poly(A)-specific ribonuclease (deadenylation nuclease)', other_designations='polyadenylate-specific ribonuclease', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13537960, end_position_on_the_genomic_accession=13668196, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  74111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74111, CurrentID=0, Status='live', Symbol='Rbm19', Aliases='1200009A02Rik, AV302714, NPO', description='RNA binding motif protein 19', other_designations='RNA-binding motif protein 19', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120116513, end_position_on_the_genomic_accession=120198971, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  74112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74112, CurrentID=0, Status='live', Symbol='Usp16', Aliases='1200004E02Rik, 2810483I07Rik, 6330514E22Rik, UBPM', description='ubiquitin specific peptidase 16', other_designations='deubiquitinating enzyme 16|ubiquitin specific protease 16|ubiquitin thioesterase 16|ubiquitin thiolesterase 16|ubiquitin-specific-processing protease 16', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87454708, end_position_on_the_genomic_accession=87483515, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  74114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74114, CurrentID=0, Status='live', Symbol='Crot', Aliases='1200003H03Rik', description='carnitine O-octanoyltransferase', other_designations='COT', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8966023, end_position_on_the_genomic_accession=8997146, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  74116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74116, CurrentID=0, Status='live', Symbol='Pi16', Aliases='1200009H11Rik, Cripi, PI-16', description='peptidase inhibitor 16', other_designations='cysteine-rich protease inhibitor|protease inhibitor 16', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29318796, end_position_on_the_genomic_accession=29328903, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  74117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74117, CurrentID=0, Status='live', Symbol='Actr3', Aliases='1200003A09Rik, Arp3', description='ARP3 actin-related protein 3', other_designations='ARP3 actin-related protein 3 homolog|actin-like protein 3', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=125392905, end_position_on_the_genomic_accession=125435727, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  74120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74120, CurrentID=0, Status='live', Symbol='Zfp263', Aliases='1200014J04Rik, AI326880, AU020888, NT2, mFPM315', description='zinc finger protein 263', other_designations='kruppel-associated box-zinc finger protein NT2', map_location='16 2.2 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3744099, end_position_on_the_genomic_accession=3750790, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74121, CurrentID=0, Status='live', Symbol='Acoxl', Aliases='1200014P05Rik, AV025673', description='acyl-Coenzyme A oxidase-like', other_designations='acyl-CoA oxidase-like protein', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127838797, end_position_on_the_genomic_accession=128126252, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  74122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74122, CurrentID=0, Status='live', Symbol='Tmem43', Aliases='1200015A22Rik, LUMA', description='transmembrane protein 43', other_designations='', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91473751, end_position_on_the_genomic_accession=91488458, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  74123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74123, CurrentID=0, Status='live', Symbol='Foxp4', Aliases='1200010K03Rik, 2310007G05Rik, mFKHLA', description='forkhead box P4', other_designations='fork head-related protein like A|fork head-related protein-like A|foxp4', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47867133, end_position_on_the_genomic_accession=47924645, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74125, CurrentID=0, Status='live', Symbol='Armc8', Aliases='1200015K23Rik, Gid5, HSPC056', description='armadillo repeat containing 8', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99478385, end_position_on_the_genomic_accession=99568936, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  74126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74126, CurrentID=0, Status='live', Symbol='Syvn1', Aliases='1200010C09Rik, AW211966, C85322, D530017H19Rik, Hrd1', description='synovial apoptosis inhibitor 1, synoviolin', other_designations='synoviolin 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6045030, end_position_on_the_genomic_accession=6053718, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74127, CurrentID=0, Status='live', Symbol='Krt80', Aliases='1200016G03Rik, 2310041I20Rik, Kb20', description='keratin 80', other_designations='CK-80|K80|cytokeratin-80|keratin-80|type-II keratin Kb20', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101348315, end_position_on_the_genomic_accession=101370125, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74129, CurrentID=0, Status='live', Symbol='Dmgdh', Aliases='1200014D15Rik, AI787269', description='dimethylglycine dehydrogenase precursor', other_designations='ME2GLYDH', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93674436, end_position_on_the_genomic_accession=93752823, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  74131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74131, CurrentID=0, Status='live', Symbol='Sash3', Aliases='1200013B08Rik, AW413946, SLY1', description='SAM and SH3 domain containing 3', other_designations='Novel SH3 and SAM domain containing protein|SH3 protein expressed in lymphocytes', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48146451, end_position_on_the_genomic_accession=48161565, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  74132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74132, CurrentID=0, Status='live', Symbol='Rnf6', Aliases='1200013I08Rik, 5730419H05Rik, AA537053', description='ring finger protein (C3H2C3 type) 6', other_designations='RLIM-like protein', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146209193, end_position_on_the_genomic_accession=146222191, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74133, CurrentID=0, Status='live', Symbol='Smg8', Aliases='1200011M11Rik', description='smg-8 homolog, nonsense mediated mRNA decay factor (C. elegans)', other_designations='CG6729 gene product|protein smg-8 homolog', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87077732, end_position_on_the_genomic_accession=87086777, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74134, CurrentID=0, Status='live', Symbol='Cyp2s1', Aliases='1200011C15Rik, AU041727, C79779', description='cytochrome P450, family 2, subfamily s, polypeptide 1', other_designations='CYPIIS1|cytochrome P450, 2s1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25802476, end_position_on_the_genomic_accession=25816530, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  74136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74136, CurrentID=0, Status='live', Symbol='Sec14l1', Aliases='1200017E04Rik, 2810012L19Rik, Naa-35', description='SEC14-like lipid binding 1', other_designations='SEC14-like 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117115172, end_position_on_the_genomic_accession=117159268, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  74137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74137, CurrentID=0, Status='live', Symbol='Nuak2', Aliases='1200013B22Rik, Omphk2, Snark, mKIAA0537', description='NUAK family, SNF1-like kinase, 2', other_designations='SNF1/AMP-activated protein kinase|nuak2|omphalocele kinase 2', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132316125, end_position_on_the_genomic_accession=132333488, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74140, CurrentID=0, Status='live', Symbol='Tm9sf1', Aliases='1200014D02Rik, AI893436, MP70', description='transmembrane 9 superfamily member 1', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55635965, end_position_on_the_genomic_accession=55643806, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  99738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99738, CurrentID=0, Status='live', Symbol='Kcnc4', Aliases='AI850292, KSHIIIC, Kcr2-4, Kv3.4', description='potassium voltage gated channel, Shaw-related subfamily, member 4', other_designations='voltage-gated potassium channel subunit Kv3.4', map_location='3 46.83 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107438296, end_position_on_the_genomic_accession=107459952, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74142, CurrentID=0, Status='live', Symbol='Lonp1', Aliases='1200017E13Rik, LON, Prss15', description='lon peptidase 1, mitochondrial', other_designations='LONP|lon protease-like protein|mitochondrial ATP-dependent protease Lon|protease, serine, 15|serine protease 15', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56614297, end_position_on_the_genomic_accession=56626903, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74143, CurrentID=0, Status='live', Symbol='Opa1', Aliases='1200011N24Rik, AI225888, AI847218, lilr3, mKIAA0567', description='optic atrophy 1', other_designations='large GTP-binding protein|largeG|optic atrophy 1 homolog|optic atrophy protein 1 homolog', map_location='16 20.65 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=29579306, end_position_on_the_genomic_accession=29663122, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  74144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74144, CurrentID=0, Status='live', Symbol='Robo4', Aliases='1200012D01Rik, AI593217', description='roundabout homolog 4 (Drosophila)', other_designations='Magic roundabout', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37401902, end_position_on_the_genomic_accession=37414023, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74145, CurrentID=0, Status='live', Symbol='F13a1', Aliases='1200014I03Rik, AI462306, F13a', description='coagulation factor XIII, A1 subunit', other_designations='coagulation factor XIII, alpha subunit|coagulation factor XIIIa|protein-glutamine gamma-glutamyltransferase A chain|transglutaminase A chain', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=36867178, end_position_on_the_genomic_accession=37050244, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  74147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74147, CurrentID=0, Status='live', Symbol='Ehhadh', Aliases='1300002P22Rik, HD, L-PBE, LBFP, LBP, MFP, MFP1, PBFE', description='enoyl-Coenzyme A, hydratase/3-hydroxyacyl Coenzyme A dehydrogenase', other_designations='L-bifunctional enzyme|L-specific multifunctional beta-oxdiation protein|PBE', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21761285, end_position_on_the_genomic_accession=21787834, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74148, CurrentID=0, Status='live', Symbol='Cluh', Aliases='1300001I01Rik, Kiaa0664, mKIAA0664', description='clustered mitochondria (cluA/CLU1) homolog', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74649495, end_position_on_the_genomic_accession=74670848, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  74150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74150, CurrentID=0, Status='live', Symbol='Slc35f5', Aliases='1300003P13Rik, AI646727', description='solute carrier family 35, member F5', other_designations='', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=125560414, end_position_on_the_genomic_accession=125595780, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  74152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74152, CurrentID=0, Status='live', Symbol='Stra6l', Aliases='1300002K09Rik, Rbpr2', description='STRA6-like', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45823811, end_position_on_the_genomic_accession=45887009, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  74153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74153, CurrentID=0, Status='live', Symbol='Uba7', Aliases='1300004C08Rik, Ube1l', description='ubiquitin-like modifier activating enzyme 7', other_designations='', map_location='9 59.07 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107975567, end_position_on_the_genomic_accession=107984056, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  74154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74154, CurrentID=0, Status='live', Symbol='Unkl', Aliases='1300004G08Rik', description='unkempt-like (Drosophila)', other_designations='Grailexp predicted', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25188344, end_position_on_the_genomic_accession=25234443, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  74155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74155, CurrentID=0, Status='live', Symbol='Errfi1', Aliases='1300002F13Rik, AI788755, Mig-6, Mig6, RALT', description='ERBB receptor feedback inhibitor 1', other_designations='mitogen-inducible gene 6 protein homolog', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150853801, end_position_on_the_genomic_accession=150868889, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74156, CurrentID=0, Status='live', Symbol='Acot12', Aliases='1300004O04Rik, 4930449F15Rik, AV027244, Cach, mCACH-1', description='acyl-CoA thioesterase 12', other_designations='CACH-1|acyl-CoA thioester hydrolase 12|cytoplasmic acetyl-CoA hydrolase 1|cytosolic acetyl-CoA hydrolase', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=91741511, end_position_on_the_genomic_accession=91786148, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74157, CurrentID=0, Status='live', Symbol='Cmtr1', Aliases='1300018I05Rik, AI451264, Ftsjd2, MTr1', description='cap methyltransferase 1', other_designations="FtsJ methyltransferase domain-containing protein 2|S-adenosyl-L-methionine-dependent methyltransferase FTSJD2|cap1 2'O-ribose methyltransferase 1", map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29660601, end_position_on_the_genomic_accession=29703359, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  74158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74158, CurrentID=0, Status='live', Symbol='Josd1', Aliases='1300006C06Rik, mKIAA0063', description='Josephin domain containing 1', other_designations='josephin domain-containing protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79674250, end_position_on_the_genomic_accession=79687872, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74159, CurrentID=0, Status='live', Symbol='Acbd5', Aliases='1300014E15Rik', description='acyl-Coenzyme A binding domain containing 5', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=23067700, end_position_on_the_genomic_accession=23114513, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74164, CurrentID=0, Status='live', Symbol='Nfx1', Aliases='1300017N15Rik, 3000003M19Rik, NFX.1, TEG-42, Tex42', description='nuclear transcription factor, X-box binding 1', other_designations='nuclear transcription factor, X box-binding protein 1|testis expressed gene 42', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40970644, end_position_on_the_genomic_accession=41025992, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  74165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74165, CurrentID=0, Status='live', Symbol='Fbxl22', Aliases='1110004B15Rik', description='F-box and leucine-rich repeat protein 22', other_designations='F-box/LRR-repeat protein 22', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66508459, end_position_on_the_genomic_accession=66514593, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  74166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74166, CurrentID=0, Status='live', Symbol='Tmem38a', Aliases='1110001E17Rik, AI413399, TRIC-A, mg33a', description='transmembrane protein 38A', other_designations='27 kDa sarcoplasmic reticulum protein|Mitsugumin33A|SPR-27|TRICA|mitsugumin-33A|sarcoplasmic reticulum protein 27', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72571998, end_position_on_the_genomic_accession=72587284, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74167, CurrentID=0, Status='live', Symbol='Nudt9', Aliases='1190002C07Rik, ADPR-PPase, AI462474', description='nudix (nucleoside diphosphate linked moiety X)-type motif 9', other_designations='ADP-ribose diphosphatase|ADP-ribose phosphohydrolase|adenosine diphosphoribose pyrophosphatase|nucleoside diphosphate-linked moiety X motif 9|nudix motif 9', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104046514, end_position_on_the_genomic_accession=104065369, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74168, CurrentID=0, Status='live', Symbol='Zdhhc16', Aliases='1500015N03Rik, Aph2', description='zinc finger, DHHC domain containing 16', other_designations='Abl-philin 2|DHHC-16|ablphilin-2|zinc finger DHHC domain-containing protein 16', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41933472, end_position_on_the_genomic_accession=41944104, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74173, CurrentID=0, Status='live', Symbol='Rab10os', Aliases='1700012B15Rik, 1810036A22Rik', description='RAB10, member RAS oncogene family, opposite strand', other_designations='Rab10 opposite strand transcript', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3235791, end_position_on_the_genomic_accession=3250374, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74174, CurrentID=0, Status='live', Symbol='Gtsf1', Aliases='1700006H03Rik, Cue110', description='gametocyte specific factor 1', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103402458, end_position_on_the_genomic_accession=103430448, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74175, CurrentID=0, Status='live', Symbol='Crct1', Aliases='2300002G24Rik, AA589624, Nice-1', description='cysteine-rich C-terminal 1', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93014205, end_position_on_the_genomic_accession=93015686, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  74176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74176, CurrentID=0, Status='live', Symbol='Tgm5', Aliases='2310007C07Rik, TGMX, TGx', description='transglutaminase 5', other_designations='TGase 5|TGase-5|transglutaminase-5', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121046111, end_position_on_the_genomic_accession=121085759, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  74178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74178, CurrentID=0, Status='live', Symbol='Stk40', Aliases='2310004N11Rik, AI428792', description='serine/threonine kinase 40', other_designations='ortholog of Ser/Thr-like protein kinase Lyk4|serine/threonine-protein kinase lyk4', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126103874, end_position_on_the_genomic_accession=126141029, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  74180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74180, CurrentID=0, Status='live', Symbol='Muc5b', Aliases='2300002I04Rik, A130042M24, AV085033, MUC5, MUC9, mucin 5b', description='mucin 5, subtype B, tracheobronchial', other_designations='', map_location='7 87.29 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141839012, end_position_on_the_genomic_accession=141881952, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  74182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74182, CurrentID=0, Status='live', Symbol='Gpcpd1', Aliases='2310004G06Rik, 2310032D16Rik, AU015213, Gde5, Prei4, mKIAA1434', description='glycerophosphocholine phosphodiesterase 1', other_designations='glycerophosphocholine phosphodiesterase GDE1 homolog|glycerophosphodiester phosphodiesterase 5|preimplantation protein 4|putative glycerophosphocholine phosphodiesterase GPCPD1|putative glycerophosphodiester phosphodiesterase 5', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132529082, end_position_on_the_genomic_accession=132578282, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  74183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74183, CurrentID=0, Status='live', Symbol='Perm1', Aliases='2310042D19Rik', description='PPARGC1 and ESRR induced regulator, muscle 1', other_designations='PGC-1 and ERR-induced regulator in muscle 1|PPARGC1 and ESRR-induced regulator in muscle 1|peroxisome proliferator-activated receptor gamma coactivator 1 and estrogen-related receptor-induced regulator in muscle 1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156203268, end_position_on_the_genomic_accession=156221307, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74185, CurrentID=0, Status='live', Symbol='Gbe1', Aliases='2310045H19Rik, 2810426P10Rik, D16Ertd536e', description='glucan (1,4-alpha-), branching enzyme 1', other_designations='brancher enzyme|glycogen-branching enzyme', map_location='16 40.19 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=70313949, end_position_on_the_genomic_accession=70569720, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  74186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74186, CurrentID=0, Status='live', Symbol='Ccdc3', Aliases='2310045O21Rik', description='coiled-coil domain containing 3', other_designations='fat/vessel-derived secretory protein|favine', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5137776, end_position_on_the_genomic_accession=5230871, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  74187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74187, CurrentID=0, Status='live', Symbol='Katnb1', Aliases='2410003J24Rik, KAT', description='katanin p80 (WD40-containing) subunit B 1', other_designations='katanin p80 subunit B1|p80 katanin', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95081173, end_position_on_the_genomic_accession=95099874, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  74188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74188, CurrentID=0, Status='live', Symbol='Prl8a8', Aliases='1600032B14Rik, Ghd14, PLP-Cgamma, Prlpc3', description='prolactin family 8, subfamily a, member 81', other_designations='PLP C-gamma|PLP-C gamma2|PLP-C3|PRL-like protein C3|growth hormone d14|placental prolactin-like protein C3|prolactin-like protein C 3|prolactin-like protein C-gamma', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27507032, end_position_on_the_genomic_accession=27513228, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74189, CurrentID=0, Status='live', Symbol='Phactr3', Aliases='1500003N10Rik, 4930415A02Rik, H17739, Scapin1, Scapinin, mKIAA4224', description='phosphatase and actin regulator 3', other_designations='scaffold-associated PP1-inhibiting protein', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=178118975, end_position_on_the_genomic_accession=178338492, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74190, CurrentID=0, Status='live', Symbol='Exoc3l4', Aliases='1200009I06Rik, 1600013K19Rik', description='exocyst complex component 3-like 4', other_designations='SEC6-like protein C14orf73 homolog', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111417430, end_position_on_the_genomic_accession=111431680, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74191, CurrentID=0, Status='live', Symbol='P2ry13', Aliases='2010001L06Rik, GPCR1, GPR94, Gpr86, P2Y13, P2yr13, SP174', description='purinergic receptor P2Y, G-protein coupled 13', other_designations='G protein-coupled receptor 68|G protein-coupled receptor 86|G-protein coupled receptor 86|Vpurinergic receptor P2Y, G-protein coupled 13|purinergic receptor P2Y13', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=59207892, end_position_on_the_genomic_accession=59210882, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  74192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74192, CurrentID=0, Status='live', Symbol='Arpc5l', Aliases='2010015J01Rik, AI852867, ARC16-2, AW555592, AW742746', description='actin related protein 2/3 complex, subunit 5-like', other_designations='arp2/3 complex 16 kDa subunit 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=39005272, end_position_on_the_genomic_accession=39015878, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74194, CurrentID=0, Status='live', Symbol='Rnd3', Aliases='2610017M01Rik, AI661404, Arhe, Rhoe', description='Rho family GTPase 3', other_designations='ras homolog gene family, member E', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=51130439, end_position_on_the_genomic_accession=51149111, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74195, CurrentID=0, Status='live', Symbol='Elp3', Aliases='2610507P14Rik, KAT9', description='elongator acetyltransferase complex subunit 3', other_designations='elongation protein 3 homolog', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65530446, end_position_on_the_genomic_accession=65593112, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74196, CurrentID=0, Status='live', Symbol='Ttc27', Aliases='2610511O17Rik', description='tetratricopeptide repeat domain 27', other_designations='TPR repeat protein 27', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74717736, end_position_on_the_genomic_accession=74863570, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  74197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74197, CurrentID=0, Status='live', Symbol='Gtf2e1', Aliases='2610024P03Rik, 56kDa, AI303800, FE, TFIIE-A', description='general transcription factor II E, polypeptide 1 (alpha subunit)', other_designations='TFIIE-alpha|general transcription factor II E, polypeptide 1 (alpha subunit, 56kDa)|general transcription factor IIE 56 kDa subunit|general transcription factor IIE, polypeptide 1 (alpha subunit, 56kDa)|transcription initiation factor IIE subunit alpha', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37509790, end_position_on_the_genomic_accession=37539769, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74198, CurrentID=0, Status='live', Symbol='Dtx2', Aliases='2610524D08Rik, AA408415, AU022494, Deltex2', description='deltex 2, E3 ubiquitin ligase', other_designations='deltex 2 homolog|mDTX2|protein deltex-2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135994800, end_position_on_the_genomic_accession=136032881, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74199, CurrentID=0, Status='live', Symbol='Vit', Aliases='1700052E02Rik, 1700110E08Rik, 2810429K11Rik', description='vitrin', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78508051, end_position_on_the_genomic_accession=78627409, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74200, CurrentID=0, Status='live', Symbol='2810403A07Rik', Aliases='A430106P18Rik, AI256352, AI451678, Blom7, Kiaa0907', description='RIKEN cDNA 2810403A07 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88685780, end_position_on_the_genomic_accession=88712933, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74201, CurrentID=0, Status='live', Symbol='Cep97', Aliases='2810403B08Rik, 4932439K18Rik, AA437768, E130116N02Rik, Lrriq2', description='centrosomal protein 97', other_designations='leucine-rich repeat and IQ domain-containing protein 2|leucine-rich repeats and IQ motif containing 2', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=55899888, end_position_on_the_genomic_accession=55934848, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74202, CurrentID=0, Status='live', Symbol='Fblim1', Aliases='2410043F08Rik, Cal, Fblp1, Gt10, migfilin, migfilin(s)', description='filamin binding LIM protein 1', other_designations='CSX-associated LIM|FBLP-1', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141576062, end_position_on_the_genomic_accession=141606052, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  74203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74203, CurrentID=0, Status='live', Symbol='Eif4enif1', Aliases='2610509L04Rik, A930019J01Rik, AA410001, AU021239, Clast4, D11Ertd166e', description='eukaryotic translation initiation factor 4E nuclear import factor 1', other_designations='4E-T|eIF4E transporter', map_location='11 2.24 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3201598, end_position_on_the_genomic_accession=3244588, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  74204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74204, CurrentID=0, Status='live', Symbol='Xpo6', Aliases='2610005L19Rik, AL022631, C230091E20Rik, R75304, Ranbp20, mKIAA0370', description='exportin 6', other_designations='RAN binding protein 20|exp6|ran-binding protein 20', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126101718, end_position_on_the_genomic_accession=126200496, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  74205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74205, CurrentID=0, Status='live', Symbol='Acsl3', Aliases='2610510B12Rik, Acs3, C85929, Facl3, Pro2194', description='acyl-CoA synthetase long-chain family member 3', other_designations='LACS 3|fatty acid Coenzyme A ligase, long chain 3|long-chain acyl-CoA synthetase 3', map_location='1 40.84 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=78657825, end_position_on_the_genomic_accession=78707743, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  74206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74206, CurrentID=0, Status='live', Symbol='Sipa1l3', Aliases='2610511M17Rik', description='signal-induced proliferation-associated 1 like 3', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29320372, end_position_on_the_genomic_accession=29518943, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  74211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74211, CurrentID=0, Status='live', Symbol='1700017B05Rik', Aliases='AA682102, D9Ertd278e', description='RIKEN cDNA 1700017B05 gene', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57252322, end_position_on_the_genomic_accession=57262599, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74213, CurrentID=0, Status='live', Symbol='Rbm26', Aliases='1700009P03Rik, AI447318, C230097K14Rik, C86225, MLZ-393, Pro1777, Se70-2', description='RNA binding motif protein 26', other_designations='RNA-binding motif protein 26|cutaneous T-cell lymphoma tumor antigen se70-2|protein expressed in male leptotene and zygotene spermatocytes 393', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=105114519, end_position_on_the_genomic_accession=105177327, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  74218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74218, CurrentID=0, Status='live', Symbol='1700016H13Rik', Aliases='', description='RIKEN cDNA 1700016H13 gene', other_designations='', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=103648583, end_position_on_the_genomic_accession=103655863, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74221, CurrentID=0, Status='live', Symbol='Tex37', Aliases='1700011F03Rik', description='testis expressed 37', other_designations='Tsc21|protein TSC21|testis-specific conserved protein of 21 kDa', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70913089, end_position_on_the_genomic_accession=70951150, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74222, CurrentID=0, Status='live', Symbol='Sept14', Aliases='1700016K13Rik', description='septin 14', other_designations='', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129682266, end_position_on_the_genomic_accession=129708511, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74229, CurrentID=0, Status='live', Symbol='Paqr8', Aliases='1700019B16Rik, 3110001D06Rik', description='progestin and adipoQ receptor family member VIII', other_designations='mPR beta', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20890622, end_position_on_the_genomic_accession=20939650, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74230, CurrentID=0, Status='live', Symbol='1700016K19Rik', Aliases='Liat1', description='RIKEN cDNA 1700016K19 gene', other_designations='ligand of ATE1 protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75999912, end_position_on_the_genomic_accession=76003569, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  74237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74237, CurrentID=0, Status='live', Symbol='Tubgcp2', Aliases='1700022B05Rik', description='tubulin, gamma complex associated protein 2', other_designations='GCP-2', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139995955, end_position_on_the_genomic_accession=140036674, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74238, CurrentID=0, Status='live', Symbol='Mterf2', Aliases='1700007D05Rik, Mterfd3', description='mitochondrial transcription termination factor 2', other_designations='MTERF domain containing 3|mTERF domain-containing protein 3, mitochondrial', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85119433, end_position_on_the_genomic_accession=85128069, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74239, CurrentID=0, Status='live', Symbol='Iqce', Aliases='1700028P05Rik, mKIAA1023', description='IQ motif containing E', other_designations='', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140662553, end_position_on_the_genomic_accession=140702436, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  74241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74241, CurrentID=0, Status='live', Symbol='Chpf', Aliases='1700028N03Rik, AI414328, D1Bwg1363e', description='chondroitin polymerizing factor', other_designations='N-acetylgalactosaminyl-proteoglycan 3-beta-glucuronosyltransferase II|N-acetylgalactosaminyltransferase 2|chondroitin glucuronyltransferase 2|chondroitin-polymerizing factor|glucuronosyl-N-acetylgalactosaminyl-proteoglycan 4-beta-N-acetylgalactosaminyltransferase II', map_location='1 39.14 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75474569, end_position_on_the_genomic_accession=75479471, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74243, CurrentID=0, Status='live', Symbol='Slx4ip', Aliases='2210009G21Rik, 2410004I22Rik, C85661, Slxip', description='SLX4 interacting protein', other_designations='SLX4-interacting protein|UPF0492 protein C20orf94 homolog', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=136891218, end_position_on_the_genomic_accession=137082176, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74244, CurrentID=0, Status='live', Symbol='Atg7', Aliases='1810013K23Rik, Agp7, Apg7ll, Gm21553, Atg7', description='autophagy related 7', other_designations='ATG12-activating enzyme E1 ATG7|autophagy-related protein 7|ubiquitin-activating enzyme E1-like protein', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=114643097, end_position_on_the_genomic_accession=114860614, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  74245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74245, CurrentID=0, Status='live', Symbol='Ctbs', Aliases='2210401K11Rik, CTB', description='chitobiase, di-N-acetyl-', other_designations='', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146450467, end_position_on_the_genomic_accession=146465849, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74246, CurrentID=0, Status='live', Symbol='Gale', Aliases='2310002A12Rik, AI323962', description='galactose-4-epimerase, UDP', other_designations='UDP-GalNAc 4-epimerase|UDP-GlcNAc 4-epimerase|UDP-N-acetylglucosamine 4-epimerase|UDP-galactosamine 4-epimerase|UDP-galactose 4-epimerase|galactowaldenase', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135963681, end_position_on_the_genomic_accession=135968178, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74249, CurrentID=0, Status='live', Symbol='Lrrc2', Aliases='2400002D05Rik, 4933431K03Rik', description='leucine rich repeat containing 2', other_designations='leucine-rich repeat-containing 2', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110951545, end_position_on_the_genomic_accession=110984066, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74251, CurrentID=0, Status='live', Symbol='Ankrd9', Aliases='2500003O20Rik', description='ankyrin repeat domain 9', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110975353, end_position_on_the_genomic_accession=110979021, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74252, CurrentID=0, Status='live', Symbol='Armc1', Aliases='2310016N05Rik, 2900046P06Rik, 3110009G21Rik, AA409258, AW048872, Arcp, C330014L16Rik', description='armadillo repeat containing 1', other_designations='armadillo repeat containing protein', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19132144, end_position_on_the_genomic_accession=19163090, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74253, CurrentID=0, Status='live', Symbol='Klrg2', Aliases='2310020F24Rik', description='killer cell lectin-like receptor subfamily G, member 2', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38621414, end_position_on_the_genomic_accession=38637706, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74254, CurrentID=0, Status='live', Symbol='Gpn1', Aliases='2410004J02Rik, AI449615, MBDIN, NTPBP, Xab1', description='GPN-loop GTPase 1', other_designations='MBD2-interacting protein|XPA binding protein 1|XPA-binding protein 1|methyl-CpG binding domain protein 2 interacting protein', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31494752, end_position_on_the_genomic_accession=31511631, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  74255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74255, CurrentID=0, Status='live', Symbol='Smu1', Aliases='2600001O03Rik, 2610203K23Rik, AB044414, AI845086, AW556129, Bwd', description='smu-1 suppressor of mec-8 and unc-52 homolog (C. elegans)', other_designations='smu-1 suppressor of mec-8 and unc-52 protein homolog', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40735649, end_position_on_the_genomic_accession=40757885, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74256, CurrentID=0, Status='live', Symbol='Cyld', Aliases='2010013M14Rik, 2900009M21Rik, C130039D01Rik, CDMT, CYLD1, EAC, mKIAA0849', description='cylindromatosis (turban tumor syndrome)', other_designations='deubiquitinating enzyme CYLD|ubiquitin thioesterase CYLD|ubiquitin thiolesterase CYLD|ubiquitin-specific-processing protease CYLD', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88697001, end_position_on_the_genomic_accession=88751946, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  74257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74257, CurrentID=0, Status='live', Symbol='Tspan17', Aliases='2210021G21Rik, AI047581, Fbxo23, Tm4sf17', description='tetraspanin 17', other_designations='F-box only protein 23|tetraspan protein SB134|transmembrane 4 superfamily member 17|tspan-17', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54789352, end_position_on_the_genomic_accession=54796896, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74264, CurrentID=0, Status='live', Symbol='Rnf138rt1', Aliases='', description='ring finger protein 138, retrogene 1', other_designations='ring finger protein 138 pseudogene|ring finger protein 138 retrogene 1', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=163760139, end_position_on_the_genomic_accession=163761332, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  74267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74267, CurrentID=0, Status='live', Symbol='Iqcf1', Aliases='1700055J15Rik', description='IQ motif containing F1', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106499967, end_position_on_the_genomic_accession=106502250, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  74268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74268, CurrentID=0, Status='live', Symbol='Aven', Aliases='1700013A01Rik, 1700056A21Rik, mAven-L, mAven-S', description='apoptosis, caspase activation inhibitor', other_designations='', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112492964, end_position_on_the_genomic_accession=112631253, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74270, CurrentID=0, Status='live', Symbol='Usp20', Aliases='1700055M05Rik, AI467231, Vdu2', description='ubiquitin specific peptidase 20', other_designations='VHL-interacting deubiquitinating enzyme 2|deubiquitinating enzyme 20|ubiquitin specific protease 20|ubiquitin thioesterase 20|ubiquitin thiolesterase 20|ubiquitin-specific-processing protease 20', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30982279, end_position_on_the_genomic_accession=31022655, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  74271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74271, CurrentID=0, Status='live', Symbol='Mageb5', Aliases='1700056A17Rik, mMage-b5', description='melanoma antigen, family B, 5', other_designations='', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=91779559, end_position_on_the_genomic_accession=91782976, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  74276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74276, CurrentID=0, Status='live', Symbol='Cldnd2', Aliases='1700071E18Rik', description='claudin domain containing 2', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43440816, end_position_on_the_genomic_accession=43443320, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74277, CurrentID=0, Status='live', Symbol='Chic2', Aliases='1700081B18Rik, 4930502K01Rik, BTL', description='cysteine-rich hydrophobic domain 2', other_designations='cysteine-rich hydrophobic domain 2 protein', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=75006424, end_position_on_the_genomic_accession=75044626, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74281, CurrentID=0, Status='live', Symbol='Spatc1', Aliases='1700084J23Rik, speriolin', description='spermatogenesis and centriole associated 1', other_designations='spermatogenesis and centriole-associated protein 1|spermatogenic cell-specific Cdc20-binding protein', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76267835, end_position_on_the_genomic_accession=76292577, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74284, CurrentID=0, Status='live', Symbol='1700086L19Rik', Aliases='1700006L05Rik, AU016007', description='RIKEN cDNA 1700086L19 gene', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=74284276, end_position_on_the_genomic_accession=74295950, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74286, CurrentID=0, Status='live', Symbol='Tbc1d21', Aliases='1700095K08Rik, MgcRabGAP', description='TBC1 domain family, member 21', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58359714, end_position_on_the_genomic_accession=58370479, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  74287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74287, CurrentID=0, Status='live', Symbol='Kcmf1', Aliases='1700094M07Rik, Debt91, Pmcf', description='potassium channel modulatory factor 1', other_designations='clone DEBT-91|debt-91|differentially expressed in branching tubulogenesis 91|potassium channel modulatory factor DEBT-91', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72841114, end_position_on_the_genomic_accession=72899979, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74288, CurrentID=0, Status='live', Symbol='Spem1', Aliases='1700095G12Rik', description='sperm maturation 1', other_designations='spermatid maturation 1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69820871, end_position_on_the_genomic_accession=69822165, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  77918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77918, CurrentID=0, Status='live', Symbol='Krtap19-3', Aliases='A030006N06Rik, Krtap16-5', description='keratin associated protein 19-3', other_designations='keratin associated protein 16-5|keratin-associated protein 16-5|keratin-associated protein 16.5', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88877513, end_position_on_the_genomic_accession=88878038, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  74297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74297, CurrentID=0, Status='live', Symbol='Ccdc182', Aliases='1700106J16Rik', description='coiled-coil domain containing 182', other_designations='coiled-coil domain-containing protein ENSP00000299415 homolog|novel tropomyosin domain containing protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88294043, end_position_on_the_genomic_accession=88295261, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  74302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74302, CurrentID=0, Status='live', Symbol='Mtmr3', Aliases='1700092A20Rik, AI255150, AW557713, FYVE-DSP1, ZFYVE10, mKIAA0371', description='myotubularin related protein 3', other_designations='phosphatidylinositol-3,5-bisphosphate 3-phosphatase|phosphatidylinositol-3-phosphate phosphatase', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4480868, end_position_on_the_genomic_accession=4594815, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  74306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74306, CurrentID=0, Status='live', Symbol='Prss46', Aliases='1700112C13Rik, 79H19B, Tessp6', description='protease, serine 46', other_designations='protease, serine, 46|testis serine protease 6', map_location='9 60.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110844506, end_position_on_the_genomic_accession=110856522, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74309, CurrentID=0, Status='live', Symbol='Osbp2', Aliases='1700095P05Rik, C630001G20Rik, Gm244, ORP-4, OSBPL1', description='oxysterol binding protein 2', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3703731, end_position_on_the_genomic_accession=3863907, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  74315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74315, CurrentID=0, Status='live', Symbol='Rnf145', Aliases='3732413I11Rik, TMRF1', description='ring finger protein 145', other_designations='', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=44518680, end_position_on_the_genomic_accession=44565520, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74316, CurrentID=0, Status='live', Symbol='Isca2', Aliases='0710001C05Rik, 5730594E03Rik, Hbld1', description='iron-sulfur cluster assembly 2', other_designations='HESB like domain containing 1|HESB-like domain-containing protein 1|iron-sulfur cluster assembly 2 homolog', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84773228, end_position_on_the_genomic_accession=84775089, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74318, CurrentID=0, Status='live', Symbol='Hopx', Aliases='1110018K11Rik, 1200015P04Rik, 2300002F06Rik, AI848177, AW490897, Cameo, Hdop, Hod, Hop, Ob1, Obl, Toto', description='HOP homeobox', other_designations='homeobox only domain|homeobox-only protein|homeodomain only protein|mOB1|odd homeobox 1|odd homeobox protein 1', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77086986, end_position_on_the_genomic_accession=77115123, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74319, CurrentID=0, Status='live', Symbol='Mettl23', Aliases='1110005A03Rik, 1500035B17Rik, 4933424L15Rik', description='methyltransferase like 23', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116843515, end_position_on_the_genomic_accession=116849740, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74320, CurrentID=0, Status='live', Symbol='Wdr33', Aliases='1110001N06Rik, 2310011G05Rik, 2810021O11Rik, 8430413N20Rik, WDC146', description='WD repeat domain 33', other_designations='WD repeat-containing protein 33|WD repeat-containing protein WDC146', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31804057, end_position_on_the_genomic_accession=31908987, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  74322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74322, CurrentID=0, Status='live', Symbol='Cxxc1', Aliases='2410002I16Rik, 5830420C16Rik, AI426635, Cfp1, Cgbp, PHF18', description='CXXC finger 1 (PHD domain)', other_designations='CpG binding protein|PHD finger and CXXC domain-containing protein 1|cpG-binding protein', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74216093, end_position_on_the_genomic_accession=74221496, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74325, CurrentID=0, Status='live', Symbol='Cltb', Aliases='2310046E19Rik', description='clathrin, light polypeptide (Lcb)', other_designations='lcb', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54592939, end_position_on_the_genomic_accession=54611397, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74326, CurrentID=0, Status='live', Symbol='Hnrnpr', Aliases='2610003J05Rik, 2610528B01Rik, Hnrpr', description='heterogeneous nuclear ribonucleoprotein R', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136310942, end_position_on_the_genomic_accession=136345981, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  74330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74330, CurrentID=0, Status='live', Symbol='Dnajc14', Aliases='5730551F12Rik, DNAJ, DRIP78, HDJ3, LIP6', description='DnaJ heat shock protein family (Hsp40) member C14', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 14', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128805468, end_position_on_the_genomic_accession=128819446, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  74334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74334, CurrentID=0, Status='live', Symbol='Ranbp10', Aliases='4432417N03Rik', description='RAN binding protein 10', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105768308, end_position_on_the_genomic_accession=105833844, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74335, CurrentID=0, Status='live', Symbol='Xrcc3', Aliases='', description='X-ray repair complementing defective repair in Chinese hamster cells 3', other_designations='X-ray repair cross-complementing protein 3', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111803192, end_position_on_the_genomic_accession=111813841, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74337, CurrentID=0, Status='live', Symbol='Palm3', Aliases='4432412L15Rik, AL024300', description='paralemmin 3', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84021474, end_position_on_the_genomic_accession=84030295, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74338, CurrentID=0, Status='live', Symbol='Slc6a19', Aliases='4632401C08Rik, B<0>AT1', description='solute carrier family 6 (neurotransmitter transporter), member 19', other_designations='b(0)AT1|solute carrier family 6 member 19|system B(0) neutral amino acid transporter AT1', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73679745, end_position_on_the_genomic_accession=73700695, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  74340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74340, CurrentID=0, Status='live', Symbol='Ahcyl2', Aliases='4631427C17Rik, AI227036, Adohcyase3, mKIAA0828', description='S-adenosylhomocysteine hydrolase-like 2', other_designations='S-adenosyl-L-homocysteine hydrolase 3|S-adenosylhomocysteine hydrolase-like protein 2|adenosylhomocysteinase 3|clone MNCb-5555|long-IRBIT', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29768379, end_position_on_the_genomic_accession=29912310, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  74342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74342, CurrentID=0, Status='live', Symbol='Lrrtm1', Aliases='4632401D06Rik, AW125451', description='leucine rich repeat transmembrane neuronal 1', other_designations='leucine-rich repeat transmembrane neuronal 1', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=77242480, end_position_on_the_genomic_accession=77259915, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  74343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74343, CurrentID=0, Status='live', Symbol='Crtc2', Aliases='4632407F12Rik, Torc2, mTORC2', description='CREB regulated transcription coactivator 2', other_designations='TORC-2|transducer of CREB protein 2|transducer of regulated cAMP response element-binding protein 2', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90254168, end_position_on_the_genomic_accession=90264125, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74347, CurrentID=0, Status='live', Symbol='Tldc1', Aliases='4632415K11Rik, Kiaa1609', description='TBC/LysM associated domain containing 1', other_designations='TBC/LysM-associated domain-containing protein 1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119760063, end_position_on_the_genomic_accession=119778416, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74349, CurrentID=0, Status='live', Symbol='Fam160a2', Aliases='4632419K20Rik, 6530415H11Rik, FHIP', description='family with sequence similarity 160, member A2', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105371211, end_position_on_the_genomic_accession=105400054, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  74351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74351, CurrentID=0, Status='live', Symbol='Ddx23', Aliases='3110082M05Rik, 4921506D17Rik', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 23', other_designations='', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98645379, end_position_on_the_genomic_accession=98662936, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74352, CurrentID=0, Status='live', Symbol='Zfp84', Aliases='2210410P13Rik, 4633401C23Rik, AI415447, BB081605, C86188, KRAB18, Zfp69', description='zinc finger protein 84', other_designations='zinc finger protein 69', map_location='7 17.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29768552, end_position_on_the_genomic_accession=29781419, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74354, CurrentID=0, Status='live', Symbol='Lrguk', Aliases='4921528H16Rik', description='leucine-rich repeats and guanylate kinase domain containing', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34028994, end_position_on_the_genomic_accession=34134034, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  74355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74355, CurrentID=0, Status='live', Symbol='Smchd1', Aliases='4931400A14Rik, AW554188, MommeD1, mKIAA0650', description='SMC hinge domain containing 1', other_designations='', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71344489, end_position_on_the_genomic_accession=71475343, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  74356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74356, CurrentID=0, Status='live', Symbol='4931428F04Rik', Aliases='AI426165, Kiaa0895l', description='RIKEN cDNA 4931428F04 gene', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105280409, end_position_on_the_genomic_accession=105289817, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74359, CurrentID=0, Status='live', Symbol='4931414P19Rik', Aliases='', description='RIKEN cDNA 4931414P19 gene', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54583662, end_position_on_the_genomic_accession=54606554, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74360, CurrentID=0, Status='live', Symbol='Cep57', Aliases='3110002L15Rik, 4921510P06Rik, 4931428M20Rik, AI467480, Tsp57, mKIAA0092', description='centrosomal protein 57', other_designations='testis-specific protein 57|translokin', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=13807788, end_position_on_the_genomic_accession=13827555, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  74361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74361, CurrentID=0, Status='live', Symbol='4931429L15Rik', Aliases='', description='RIKEN cDNA 4931429L15 gene', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46300913, end_position_on_the_genomic_accession=46319986, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74362, CurrentID=0, Status='live', Symbol='Spag17', Aliases='4931427F14Rik, PF6-ps, Spag17', description='sperm associated antigen 17', other_designations='projection protein PF6 homolog|sperm associated antigen 17, pseudogene', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=99885417, end_position_on_the_genomic_accession=100143322, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  74365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74365, CurrentID=0, Status='live', Symbol='Lonrf3', Aliases='4932412G04Rik, 5730439E01Rik, A830039N02Rik, AU023707, Rnf127', description='LON peptidase N-terminal domain and ring finger 3', other_designations='ring finger protein 127', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36328358, end_position_on_the_genomic_accession=36366857, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  74369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74369, CurrentID=0, Status='live', Symbol='Mei1', Aliases='4932408F18Rik, mei1', description='meiotic double-stranded break formation protein 1', other_designations='mei1|meiosis defective 1|meiosis defective protein 1', map_location='15 38.41 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82070059, end_position_on_the_genomic_accession=82126817, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  74370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74370, CurrentID=0, Status='live', Symbol='Rptor', Aliases='4932417H02Rik, Rap, Raptor, mKIAA1303', description='regulatory associated protein of MTOR, complex 1', other_designations='p150 target of rapamycin (TOR)-scaffold protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119602995, end_position_on_the_genomic_accession=119899591, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  74374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74374, CurrentID=0, Status='live', Symbol='Clec16a', Aliases='4932416N17Rik, curt, mKIAA0350', description='C-type lectin domain family 16, member A', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10545339, end_position_on_the_genomic_accession=10744878, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  74375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74375, CurrentID=0, Status='live', Symbol='Gcc1', Aliases='4932417P04Rik', description='golgi coiled coil 1', other_designations='Golgi coiled coil protein 1|golgi coiled-coil protein 1', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=28416603, end_position_on_the_genomic_accession=28421724, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  74376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74376, CurrentID=0, Status='live', Symbol='Myo18b', Aliases='4932408L24Rik, 4933411E19Rik, Gm448', description='myosin XVIIIb', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112688876, end_position_on_the_genomic_accession=112896459, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  74377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74377, CurrentID=0, Status='live', Symbol='Hsf2bp', Aliases='4932437G14Rik', description='heat shock transcription factor 2 binding protein', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31944769, end_position_on_the_genomic_accession=32036349, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  74383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74383, CurrentID=0, Status='live', Symbol='Ubap2l', Aliases='3110083O19Rik, 4932431F02Rik, A430103N23Rik, C77168, Nice-4, mKIAA0144', description='ubiquitin-associated protein 2-like', other_designations='Nice-4 protein homolog|ubiquitin associated protein 2-like', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89999589, end_position_on_the_genomic_accession=90052623, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  74385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74385, CurrentID=0, Status='live', Symbol='Ap5m1', Aliases='4932432K03Rik, Mu5, Mudeng', description='adaptor-related protein complex 5, mu 1 subunit', other_designations='MHD domain-containing death-inducing protein|MU-2/AP1M2 domain containing, death-inducing|adapter-related protein complex 5 mu subunit|adapter-related protein complex 5 subunit mu-1|adaptor-related protein complex 5 subunit mu-1|mu-2-related death-inducing protein|muD', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49066495, end_position_on_the_genomic_accession=49094114, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74386, CurrentID=0, Status='live', Symbol='Rmi1', Aliases='4932432N11Rik, C79893', description='RMI1, RecQ mediated genome instability 1', other_designations='RMI1, RecQ mediated genome instability 1, homolog', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58402248, end_position_on_the_genomic_accession=58411149, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74387, CurrentID=0, Status='live', Symbol='4932438H23Rik', Aliases='4931406G06Rik, AI550392', description='RIKEN cDNA 4932438H23 gene', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91053935, end_position_on_the_genomic_accession=91130260, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74388, CurrentID=0, Status='live', Symbol='Dpp8', Aliases='2310004I03Rik, 4932434F09Rik, AI666706, DPP VIII', description='dipeptidylpeptidase 8', other_designations='DP8|dipeptidyl peptidase VIII', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65032458, end_position_on_the_genomic_accession=65084503, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  74392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74392, CurrentID=0, Status='live', Symbol='Specc1l', Aliases='4930470P14Rik, 4932439K10Rik, 9530057A13Rik, AI317249, Cytsa, mKIAA0376', description='sperm antigen with calponin homology and coiled-coil domains 1-like', other_designations='SPECC1-like protein|cytospin A', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75212073, end_position_on_the_genomic_accession=75312400, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  74393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74393, CurrentID=0, Status='live', Symbol='Map10', Aliases='4933403G14Rik, AI835009, mKIAA1383', description='microtubule-associated protein 10', other_designations='microtubule regulator of 120 KDa', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125669818, end_position_on_the_genomic_accession=125673365, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  74400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74400, CurrentID=0, Status='live', Symbol='Zfp819', Aliases='4930427I11Rik, 4933405K07Rik', description='zinc finger protein 819', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43607169, end_position_on_the_genomic_accession=43618279, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74401, CurrentID=0, Status='live', Symbol='4933406J08Rik', Aliases='TERB2', description='RIKEN cDNA 4933406J08 gene', other_designations='uncharacterized protein C15orf43 homolog', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122186272, end_position_on_the_genomic_accession=122206393, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74405, CurrentID=0, Status='live', Symbol='Efhc2', Aliases='4933407D04Rik, mRib72-2', description='EF-hand domain (C-terminal) containing 2', other_designations='', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=17132049, end_position_on_the_genomic_accession=17320461, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  74407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74407, CurrentID=0, Status='live', Symbol='Ttc25', Aliases='4933404O19Rik', description='tetratricopeptide repeat domain 25', other_designations='TPR repeat protein 25', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100545613, end_position_on_the_genomic_accession=100572566, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  74409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74409, CurrentID=0, Status='live', Symbol='Hyal6', Aliases='4932701A20Rik, Hyal-ps1, Hyalp1', description='hyaluronoglucosaminidase 6', other_designations='hyaluronoglucosaminidase, pseudogene 1', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24733245, end_position_on_the_genomic_accession=24745452, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74410, CurrentID=0, Status='live', Symbol='Ttll11', Aliases='4932702F08Rik, 4933424A20Rik, AU043211, D2Ertd624e', description='tubulin tyrosine ligase-like family, member 11', other_designations='polyglutamylase|ttll11|tubulin--tyrosine ligase-like protein 11', map_location='2 23.9 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35751226, end_position_on_the_genomic_accession=35979624, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74411, CurrentID=0, Status='live', Symbol='Plpp6', Aliases='4932443D16Rik, Ppapdc2', description='phospholipid phosphatase 6', other_designations='phosphatidic acid phosphatase type 2 domain containing 2|phosphatidic acid phosphatase type 2 domain-containing protein 2|presqualene diphosphate phosphatase', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=28963920, end_position_on_the_genomic_accession=28966801, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  74412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74412, CurrentID=0, Status='live', Symbol='Gle1', Aliases='4933405K21Rik, AA553313l, Gle1', description='GLE1 RNA export mediator (yeast)', other_designations='GLE1-like protein', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29935409, end_position_on_the_genomic_accession=29959432, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  74413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74413, CurrentID=0, Status='live', Symbol='Tc2n', Aliases='4933406D09Rik, Mtac2d1, Tac2-N', description='tandem C2 domains, nuclear', other_designations='membrane targeting (tandem) C2 domain containing 1|membrane targeting tandem C2 domain-containing protein 1|tandem C2 protein in nucleus', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=101645446, end_position_on_the_genomic_accession=101719000, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74414, CurrentID=0, Status='live', Symbol='Polr3c', Aliases='4933407E01Rik, RPC3, RPC62', description='polymerase (RNA) III (DNA directed) polypeptide C', other_designations='DNA-directed RNA polymerase III subunit C|RNA polymerase III subunit C3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96711520, end_position_on_the_genomic_accession=96727480, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74419, CurrentID=0, Status='live', Symbol='Tktl2', Aliases='4933401I19Rik', description='transketolase-like 2', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=66511740, end_position_on_the_genomic_accession=66519199, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  74424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74424, CurrentID=0, Status='live', Symbol='Tmc5', Aliases='4932443L08Rik, AU016085', description='transmembrane channel-like gene family 5', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118597297, end_position_on_the_genomic_accession=118675086, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  74427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74427, CurrentID=0, Status='live', Symbol='Eaf1', Aliases='4933403C17Rik, AI666536', description='ELL associated factor 1', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31495079, end_position_on_the_genomic_accession=31509858, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74430, CurrentID=0, Status='live', Symbol='4930452B06Rik', Aliases='4933406L09Rik, AI854000', description='RIKEN cDNA 4930452B06 gene', other_designations='', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8431171, end_position_on_the_genomic_accession=8666502, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74434, CurrentID=0, Status='live', Symbol='Sohlh2', Aliases='4933406N12Rik, AI413378', description='spermatogenesis and oogenesis specific basic helix-loop-helix 2', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=55182044, end_position_on_the_genomic_accession=55209957, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74435, CurrentID=0, Status='live', Symbol='Lrriq3', Aliases='4930438B07Rik, 4930511J15Rik, 4933403H06Rik, BB020367, Lrrc44', description='leucine-rich repeats and IQ motif containing 3', other_designations='leucine rich repeat containing 44|leucine-rich repeat-containing protein 44', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=155093434, end_position_on_the_genomic_accession=155194278, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74437, CurrentID=0, Status='live', Symbol='4933402E13Rik', Aliases='', description='RIKEN cDNA 4933402E13 gene', other_designations='Novel transcript similar to melanoma-associated antigen 11 (MAGE-11 antigen)', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=62282961, end_position_on_the_genomic_accession=62292084, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74438, CurrentID=0, Status='live', Symbol='Clvs1', Aliases='4933402J24Rik, Clvl1, Rlbp1l1', description='clavesin 1', other_designations='retinaldehyde binding protein 1-like 1|retinaldehyde-binding protein 1-like 1', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=9144756, end_position_on_the_genomic_accession=9451691, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74440, CurrentID=0, Status='live', Symbol='Cmip', Aliases='4933407C03Rik, 5830471E12Rik, AI314065, AI464126', description='c-Maf inducing protein', other_designations='c-Mip', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=117257019, end_position_on_the_genomic_accession=117461506, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  74441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74441, CurrentID=0, Status='live', Symbol='Slco6c1', Aliases='4933404A18Rik, C79179', description='solute carrier organic anion transporter family, member 6c1', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=97059447, end_position_on_the_genomic_accession=97128340, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74442, CurrentID=0, Status='live', Symbol='Sgms2', Aliases='4933405A16Rik, 5133401H06Rik, AI854299', description='sphingomyelin synthase 2', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=131318985, end_position_on_the_genomic_accession=131491192, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74443, CurrentID=0, Status='live', Symbol='P4htm', Aliases='4933406E20Rik, AI853847, BB128974, Ph-4', description='prolyl 4-hydroxylase, transmembrane (endoplasmic reticulum)', other_designations='HIF-PH4|HIF-prolyl hydroxylase 4|HPH-4|P4H-TM|hypoxia-inducible factor prolyl hydroxylase 4', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108578826, end_position_on_the_genomic_accession=108597876, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74446, CurrentID=0, Status='live', Symbol='Slc9b1', Aliases='1700094G20Rik, 4933424B12Rik, 4933425K02Rik, Nhedc1, mtsNHE', description='solute carrier family 9, subfamily B (NHA1, cation proton antiporter 1), member 1', other_designations='NHE domain-containing protein 1|Na(+)/H(+) exchanger-like domain-containing protein 1|Na+/H+ exchanger domain containing 1|sodium/hydrogen exchanger-like domain-containing protein 1|solute carrier family 9 subfamily B member 1', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135347994, end_position_on_the_genomic_accession=135397827, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74448, CurrentID=0, Status='live', Symbol='Arl13a', Aliases='4933422M21Rik, Arl13', description='ADP-ribosylation factor-like 13A', other_designations='ADP-ribosylation factor-like 13', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134187501, end_position_on_the_genomic_accession=134208030, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74450, CurrentID=0, Status='live', Symbol='Pank2', Aliases='4933409I19Rik, AI642621', description='pantothenate kinase 2', other_designations='pantothenate kinase 2 (Hallervorden-Spatz syndrome)', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131262011, end_position_on_the_genomic_accession=131299188, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74451, CurrentID=0, Status='live', Symbol='Pgs1', Aliases='2610019F11Rik, 4933424M23Rik, SAF', description='phosphatidylglycerophosphate synthase 1', other_designations='PGP synthase 1|silencer-associated factor', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117986297, end_position_on_the_genomic_accession=118024011, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74453, CurrentID=0, Status='live', Symbol='Cfap53', Aliases='4933415I03Rik, Ccdc11', description='cilia and flagella associated protein 53', other_designations='coiled-coil domain containing 11|coiled-coil domain-containing protein 11', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=74283100, end_position_on_the_genomic_accession=74359984, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74455, CurrentID=0, Status='live', Symbol='Nsun6', Aliases='4933403D21Rik, 4933414E04Rik, NOPD1', description='NOL1/NOP2/Sun domain family member 6', other_designations='NOL1/NOP2/Sun domain family 6', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14991522, end_position_on_the_genomic_accession=15054872, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  74463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74463, CurrentID=0, Status='live', Symbol='Exoc3l2', Aliases='4933417E01Rik, AU042655', description='exocyst complex component 3-like 2', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19463148, end_position_on_the_genomic_accession=19496762, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  74464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74464, CurrentID=0, Status='live', Symbol='Zswim5', Aliases='4933426E21Rik, AI503093, mKIAA1511', description='zinc finger SWIM-type containing 5', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116877402, end_position_on_the_genomic_accession=116989105, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74466, CurrentID=0, Status='live', Symbol='4933427G17Rik', Aliases='', description='RIKEN cDNA 4933427G17 gene', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120982442, end_position_on_the_genomic_accession=121015011, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  74467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74467, CurrentID=0, Status='live', Symbol='Pus10', Aliases='2810013G11Rik, 4933435A13Rik, AU014648, C77560, Ccdc139', description='pseudouridylate synthase 10', other_designations='coiled-coil domain containing 139|coiled-coil domain-containing protein 139|psi55 synthase|tRNA pseudouridine 55 synthase|tRNA pseudouridylate synthase|tRNA-uridine isomerase', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23665354, end_position_on_the_genomic_accession=23732876, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  74468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74468, CurrentID=0, Status='live', Symbol='Hyal5', Aliases='4933439A12Rik', description='hyaluronoglucosaminidase 5', other_designations='hyaluronidase Hyal-5', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24857962, end_position_on_the_genomic_accession=24891958, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74469, CurrentID=0, Status='live', Symbol='Taf7l', Aliases='4933438I11Rik, Taf2q', description='TATA-box binding protein associated factor 7 like', other_designations='TAF7-like RNA polymerase II, TATA box binding protein (TBP)-associated factor|TATA box binding protein (TBP)-associated factor, RNA polymerase II, Q|TATA box binding protein-associated factor RNA polymerase II subunit Q|TATA box-binding protein-associated factor 50 kDa|transcription initiation factor TFIID 50 kDa subunit', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134460116, end_position_on_the_genomic_accession=134476526, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74470, CurrentID=0, Status='live', Symbol='Cep72', Aliases='2610029E11Rik, 4933440J22Rik', description='centrosomal protein 72', other_designations='centrosomal protein 72 kDa', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74036495, end_position_on_the_genomic_accession=74062314, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74472, CurrentID=0, Status='live', Symbol='4933433C11Rik', Aliases='', description='RIKEN cDNA 4933433C11 gene', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25212207, end_position_on_the_genomic_accession=25218703, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  12412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12412, CurrentID=0, Status='live', Symbol='Cbx1', Aliases='Cbx, Cbx-rs2, E430007M08Rik, HP1B, Hp1beta, M31, MOD1', description='chromobox 1', other_designations='HP1 beta|chromobox homolog 1|heterochromatin protein 1 homolog beta|heterochromatin protein p25|modifier 1 protein', map_location='11 60.11 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96788679, end_position_on_the_genomic_accession=96808982, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74477, CurrentID=0, Status='live', Symbol='4933427D14Rik', Aliases='C85113, Kiaa0753, MNR', description='RIKEN cDNA 4933427D14 gene', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72154097, end_position_on_the_genomic_accession=72207473, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  74478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74478, CurrentID=0, Status='live', Symbol='Snx29', Aliases='4933437K13Rik, ENSMUSG00000079738, Gm11170, Gm1737, Gm930, Rundc2a', description='sorting nexin 29', other_designations='RUN domain containing 2A', map_location='', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11321256, end_position_on_the_genomic_accession=11755931, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  74479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74479, CurrentID=0, Status='live', Symbol='Snx11', Aliases='4933439F10Rik, A930041K09Rik, AI117694', description='sorting nexin 11', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96767546, end_position_on_the_genomic_accession=96777697, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  74480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74480, CurrentID=0, Status='live', Symbol='Samd4', Aliases='1700024G08Rik, 1700111L17Rik, 4933436G17Rika, Smaug, Smaug1, sunk, Samd4', description='sterile alpha motif domain containing 4', other_designations='smaug 1|sterile alpha motif domain-containing protein 4A', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=46882965, end_position_on_the_genomic_accession=47105817, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74481, CurrentID=0, Status='live', Symbol='Batf2', Aliases='4933430F08Rik', description='basic leucine zipper transcription factor, ATF-like 2', other_designations='B-ATF-2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6164392, end_position_on_the_genomic_accession=6172476, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  74482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74482, CurrentID=0, Status='live', Symbol='Ifitm7', Aliases='4933438K12Rik, Mil4', description='interferon induced transmembrane protein 7', other_designations='interferon induced transmembrane protein 4 like|interferon induced transmembrane protein like 4', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13981479, end_position_on_the_genomic_accession=13986867, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  74484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74484, CurrentID=0, Status='live', Symbol='Rbm31y', Aliases='4933434H11Rik', description='RNA binding motif 31, Y-linked', other_designations='', map_location='Y', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=17400761, end_position_on_the_genomic_accession=17402718, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  74485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74485, CurrentID=0, Status='live', Symbol='Lrrc71', Aliases='4933430H15Rik', description='leucine rich repeat containing 71', other_designations='leucine-rich repeat-containing protein C10orf92 homolog', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87736923, end_position_on_the_genomic_accession=87748980, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74486, CurrentID=0, Status='live', Symbol='Osbpl10', Aliases='4933433D06Rik, C820004B04Rik, OPR-10, ORP-10', description='oxysterol binding protein-like 10', other_designations='OSBP-related protein 10', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114978583, end_position_on_the_genomic_accession=115233273, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  74488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74488, CurrentID=0, Status='live', Symbol='Lrrc15', Aliases='5430427N11Rik, LIB', description='leucine rich repeat containing 15', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30269302, end_position_on_the_genomic_accession=30283254, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  74490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74490, CurrentID=0, Status='live', Symbol='Mamstr', Aliases='2810022D01Rik, 5430432N15Rik, AW743872, Mastr', description='MEF2 activating motif and SAP domain containing transcriptional regulator', other_designations='MEF2-activating SAP transcriptional regulatory protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45639200, end_position_on_the_genomic_accession=45648206, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74492, CurrentID=0, Status='live', Symbol='Kbtbd13', Aliases='5430433E21Rik', description='kelch repeat and BTB (POZ) domain containing 13', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65388684, end_position_on_the_genomic_accession=65391652, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  74493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74493, CurrentID=0, Status='live', Symbol='Tnks2', Aliases='5430432P15Rik, AA517131, AI662480, ARTD6, TNKS-2, Tank2', description='tankyrase, TRF1-interacting ankyrin-related ADP-ribose polymerase 2', other_designations='ADP-ribosyltransferase diphtheria toxin-like 6|TRF1-interacting ankyrin-related ADP-ribose polymerase 2|tankyrase 2|tankyrase II|tankyrase-like protein|tankyrase-related protein', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36834220, end_position_on_the_genomic_accession=36893477, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  74498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74498, CurrentID=0, Status='live', Symbol='Gorasp1', Aliases='5430411C10Rik, GOLPH5, GRASP65, P65', description='golgi reassembly stacking protein 1', other_designations='golgi peripheral membrane protein p65|golgi reassembly-stacking protein of 65 kDa', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119925673, end_position_on_the_genomic_accession=119937558, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  74499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74499, CurrentID=0, Status='live', Symbol='Sost', Aliases='5430411E23Rik', description='sclerostin', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101962458, end_position_on_the_genomic_accession=101967015, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  74504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74504, CurrentID=0, Status='live', Symbol='Fam53a', Aliases='2410018C17Rik, 5430419M09Rik, Dntnp', description='family with sequence similarity 53, member A', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33600351, end_position_on_the_genomic_accession=33629647, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74511, CurrentID=0, Status='live', Symbol='Lrrc17', Aliases='37kDa, 4833425M04Rik, 6130400C22Rik, P37nb', description='leucine rich repeat containing 17', other_designations='37 kDa leucine-rich repeat (LRR) protein|leucine-rich repeat (LRR) protein', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21543527, end_position_on_the_genomic_accession=21575911, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74513, CurrentID=0, Status='live', Symbol='Neto2', Aliases='5530601C23Rik, AI854825, BTCL2', description='neuropilin (NRP) and tolloid (TLL)-like 2', other_designations='brain-specific transmembrane protein containing 2 CUB and 1 LDL-receptor class A domains protein 2', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85636588, end_position_on_the_genomic_accession=85700919, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74519, CurrentID=0, Status='live', Symbol='Cyp2j9', Aliases='8430417E17Rik', description='cytochrome P450, family 2, subfamily j, polypeptide 9', other_designations='cytochrome P450, 2j9', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=96568429, end_position_on_the_genomic_accession=96591719, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74521, CurrentID=0, Status='live', Symbol='Ppp4r4', Aliases='8430415E04Rik, mKIAA1622', description='protein phosphatase 4, regulatory subunit 4', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103532289, end_position_on_the_genomic_accession=103613839, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  74522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74522, CurrentID=0, Status='live', Symbol='Morc2a', Aliases='8430403M08Rik, Zcwcc1', description='microrchidia 2A', other_designations='zinc finger CW-type coiled-coil domain protein 1|zinc finger, CW-type with coiled-coil domain 1', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3649494, end_position_on_the_genomic_accession=3690477, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  74525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74525, CurrentID=0, Status='live', Symbol='Fam234b', Aliases='8430419L09Rik, Kiaa1467, mKIAA1467', description='family with sequence similarity 234, member B', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135197983, end_position_on_the_genomic_accession=135236242, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74528, CurrentID=0, Status='live', Symbol='Mgme1', Aliases='8430406I07Rik, AI426476', description='mitochondrial genome maintainance exonuclease 1', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144270856, end_position_on_the_genomic_accession=144281232, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  74533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74533, CurrentID=0, Status='live', Symbol='Gzf1', Aliases='8430437G08Rik, Zfp336, mGZF1', description='GDNF-inducible zinc finger protein 1', other_designations='zinc finger protein 336', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148679385, end_position_on_the_genomic_accession=148692949, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74548, CurrentID=0, Status='live', Symbol='Gsdmc4', Aliases='9030605I04Rik', description='gasdermin C4', other_designations='gsdmc4', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=63891257, end_position_on_the_genomic_accession=63912316, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74549, CurrentID=0, Status='live', Symbol='Mau2', Aliases='9130404D08Rik, A930019L04Rik, C77863, C79014, Mau-2, mKIAA0892', description='MAU2 sister chromatid cohesion factor', other_designations='Scc4-like protein|cohesin loading complex subunit SCC4 homolog|protein MAU-2', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70014527, end_position_on_the_genomic_accession=70042734, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  74551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74551, CurrentID=0, Status='live', Symbol='Pck2', Aliases='1810010O14Rik, 9130022B02Rik, PEPCK-M', description='phosphoenolpyruvate carboxykinase 2 (mitochondrial)', other_designations='phosphoenolpyruvate carboxylase', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55540263, end_position_on_the_genomic_accession=55550777, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74552, CurrentID=0, Status='live', Symbol='Nipal3', Aliases='9130020G22Rik, AI480660, Npal3', description='NIPA-like domain containing 3', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135447136, end_position_on_the_genomic_accession=135494806, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74556, CurrentID=0, Status='live', Symbol='Themis3', Aliases='9130404H23Rik', description='thymocyte selection associated family member 3', other_designations='thymocyte-expressed molecule involved in selection protein 3', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=66555252, end_position_on_the_genomic_accession=66594622, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74558, CurrentID=0, Status='live', Symbol='Gvin1', Aliases='9130002C22Rik, 9830104F22Rik, Iigs1, VLIG, VLIG-1', description='GTPase, very large interferon inducible 1', other_designations='interferon inducible GTPase 1|very large-inducible GTPase-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106156535, end_position_on_the_genomic_accession=106215377, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74559, CurrentID=0, Status='live', Symbol='Elovl7', Aliases='9130013K24Rik, AI840082', description='ELOVL family member 7, elongation of long chain fatty acids (yeast)', other_designations='3-keto acyl-CoA synthase Elovl7|ELOVL FA elongase 7|ELOVL fatty acid elongase 7|elongation of very long chain fatty acids-like 7|very long chain 3-ketoacyl-CoA synthase 7|very long chain 3-oxoacyl-CoA synthase 7|very-long-chain 3-oxoacyl-CoA synthase 7', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=108214328, end_position_on_the_genomic_accession=108287109, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  74561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74561, CurrentID=0, Status='live', Symbol='Nkx6-3', Aliases='9130417I07Rik, Nkx6.3', description='NK6 homeobox 3', other_designations='NK6 transcription factor related, locus 3', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=23153271, end_position_on_the_genomic_accession=23158948, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  74562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74562, CurrentID=0, Status='live', Symbol='Fer1l4', Aliases='9130402C12Rik', description='fer-1-like 4 (C. elegans)', other_designations='otoferlin-like protein', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156019139, end_position_on_the_genomic_accession=156053018, orientation='minus', exon_count=47, OMIM='', no_hdr0=''),
  74563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74563, CurrentID=0, Status='live', Symbol='Rasgef1c', Aliases='9130006A14Rik', description='RasGEF domain family, member 1C', other_designations='', map_location='11 B1.2-B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49901816, end_position_on_the_genomic_accession=49980223, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  74568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74568, CurrentID=0, Status='live', Symbol='Mlkl', Aliases='9130019I15Rik', description='mixed lineage kinase domain-like', other_designations='mlkl', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111311681, end_position_on_the_genomic_accession=111338449, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74569, CurrentID=0, Status='live', Symbol='Ttc17', Aliases='9130020K17Rik, D2Bwg1005e', description='tetratricopeptide repeat domain 17', other_designations='TPR repeat protein 17', map_location='2 51.69 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=94300766, end_position_on_the_genomic_accession=94406689, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  74570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74570, CurrentID=0, Status='live', Symbol='Zkscan1', Aliases='5930429A01Rik, 9130423L19Rik, 9230118B16Rik, AI646829, AI788588, C87323, KOX18, PHZ-37, ZNF139', description='zinc finger with KRAB and SCAN domains 1', other_designations='zinc finger protein 36 (KOX 18)', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138085076, end_position_on_the_genomic_accession=138107822, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  74571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74571, CurrentID=0, Status='live', Symbol='Kcnk16', Aliases='4731413G05Rik, TALK-1, TALK1', description='potassium channel, subfamily K, member 16', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20262756, end_position_on_the_genomic_accession=20269162, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74574, CurrentID=0, Status='live', Symbol='Lvrn', Aliases='4833403I15Rik, APQ, Aqpep', description='laeverin', other_designations='AP-Q', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46850039, end_position_on_the_genomic_accession=46907246, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  74577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74577, CurrentID=0, Status='live', Symbol='Glb1l', Aliases='4833408P15Rik', description='galactosidase, beta 1-like', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75198235, end_position_on_the_genomic_accession=75210816, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74580, CurrentID=0, Status='live', Symbol='Pyroxd2', Aliases='3830409H07Rik, 4833409A17Rik', description='pyridine nucleotide-disulphide oxidoreductase domain 2', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42725857, end_position_on_the_genomic_accession=42752825, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  74585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74585, CurrentID=0, Status='live', Symbol='Sppl3', Aliases='4833416I09Rik, Psl4, Usmg3', description='signal peptide peptidase 3', other_designations='IMP-2|SPP-like 3 protein|intramembrane protease 2|presenilin-like protein 4|upregulated during skeletal muscle growth 3', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115011524, end_position_on_the_genomic_accession=115098790, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74589, CurrentID=0, Status='live', Symbol='Kbtbd12', Aliases='4833415F11Rik, 4933428M03Rik, Klhdc6', description='kelch repeat and BTB (POZ) domain containing 12', other_designations='kelch domain containing 6', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88547722, end_position_on_the_genomic_accession=88627448, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74591, CurrentID=0, Status='live', Symbol='Abca12', Aliases='4832428G11Rik, 4833417A11Rik', description='ATP-binding cassette, sub-family A (ABC1), member 12', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=71243090, end_position_on_the_genomic_accession=71414910, orientation='minus', exon_count=53, OMIM='', no_hdr0=''),
  74596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74596, CurrentID=0, Status='live', Symbol='Cds1', Aliases='4833409J18Rik, AI314024, AW125888', description='CDP-diacylglycerol synthase 1', other_designations='CDP-DAG synthase 1|CDP-DG synthase 1|CDP-diglyceride pyrophosphorylase 1|CDP-diglyceride synthase 1|CDS 1|CTP:phosphatidate cytidylyltransferase 1', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=101765130, end_position_on_the_genomic_accession=101823858, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  74600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74600, CurrentID=0, Status='live', Symbol='Mrpl47', Aliases='4833424P18Rik, CGI-204, ENSMUSG00000051761, Gm9859, L47mt, MRP-L47, MTF/L47, NCM1', description='mitochondrial ribosomal protein L47', other_designations='metanephros-derived tubulogenic factor', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=32727497, end_position_on_the_genomic_accession=32736755, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74603, CurrentID=0, Status='live', Symbol='Cd200r3', Aliases='4733401I18Rik, 4833409J19Rik, AI505817, BB106877, mCD200RLb', description='CD200 receptor 3', other_designations='CD200 cell surface glycoprotein receptor 3|CD200 cell surface glycoprotein receptor-like 3|CD200 cell surface glycoprotein receptor-like b|CD200 receptor-like 3|CD200RLb|cell surface glycoprotein OX2 receptor 3', map_location='16 B4-B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44943596, end_position_on_the_genomic_accession=45058365, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  74610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74610, CurrentID=0, Status='live', Symbol='Abcb8', Aliases='4833412N02Rik, AA409895', description='ATP-binding cassette, sub-family B (MDR/TAP), member 8', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24393681, end_position_on_the_genomic_accession=24409947, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  74614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74614, CurrentID=0, Status='live', Symbol='Ocstamp', Aliases='4833422F24Rik, OC-STAMP', description='osteoclast stimulatory transmembrane protein', other_designations='transmembrane protein C20orf123 homolog', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165395450, end_position_on_the_genomic_accession=165400394, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  74616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74616, CurrentID=0, Status='live', Symbol='Scrn3', Aliases='4833415E20Rik, AI427019, AV002033, Ses3', description='secernin 3', other_designations='', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73312652, end_position_on_the_genomic_accession=73337814, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74617, CurrentID=0, Status='live', Symbol='Scpep1', Aliases='2410018F01Rik, 4833411K15Rik, AI957256, Risc', description='serine carboxypeptidase 1', other_designations='retinoid-inducible serine caroboxypeptidase|retinoid-inducible serine caroboxypetidase|serine caroboxypeptidase 1', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88924020, end_position_on_the_genomic_accession=88955442, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  74626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74626, CurrentID=0, Status='live', Symbol='Tmem81', Aliases='4930429O20Rik, AI465369', description='transmembrane protein 81', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132506230, end_position_on_the_genomic_accession=132508639, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  209047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209047, CurrentID=0, Status='live', Symbol='Gipc3', Aliases='Ahl5, Rgs19ip3', description='GIPC PDZ domain containing family, member 3', other_designations='age related hearing loss 5|regulator of G-protein signaling 19 interacting protein 3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81336540, end_position_on_the_genomic_accession=81343744, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74637, CurrentID=0, Status='live', Symbol='Shpk', Aliases='4930431K22Rik, AI194947, AW260459, Carkl', description='sedoheptulokinase', other_designations='SHK|carbohydrate kinase-like protein', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73199224, end_position_on_the_genomic_accession=73224506, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  74645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74645, CurrentID=0, Status='live', Symbol='Fam46c', Aliases='4930431B09Rik, AI449797', description='family with sequence similarity 46, member C', other_designations='DN-384L23.3', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100468057, end_position_on_the_genomic_accession=100489192, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  74646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74646, CurrentID=0, Status='live', Symbol='Spsb1', Aliases='1110014L01Rik, 4930422J18Rik, AI596360, AI854583, SSB-1, SSB1', description='splA/ryanodine receptor domain and SOCS box containing 1', other_designations='SPRY domain-containing SOCS box protein SSB-1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149896283, end_position_on_the_genomic_accession=149955006, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74648, CurrentID=0, Status='live', Symbol='S100pbp', Aliases='4930429A08Rik, AI851343r, S100pbp', description='S100P binding protein', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129148002, end_position_on_the_genomic_accession=129189742, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74649, CurrentID=0, Status='live', Symbol='Cpa5', Aliases='4930430M09Rik', description='carboxypeptidase A5', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30611005, end_position_on_the_genomic_accession=30631740, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  74653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74653, CurrentID=0, Status='live', Symbol='Pomk', Aliases='4930444A02Rik, Sgk196', description='protein-O-mannose kinase', other_designations='probable inactive protein kinase-like protein SgK196|protein kinase-like protein SgK196|sugen kinase 196', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25980604, end_position_on_the_genomic_accession=25994121, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12443, CurrentID=0, Status='live', Symbol='Ccnd1', Aliases='AI327039, CycD1, Cyl-1, PRAD1, bcl-1, cD1', description='cyclin D1', other_designations='', map_location='7 88.92 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144929931, end_position_on_the_genomic_accession=144939937, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74663, CurrentID=0, Status='live', Symbol='4930435E12Rik', Aliases='TSCPA', description='RIKEN cDNA 4930435E12 gene', other_designations='testis-specific conserved, cAMP dependent type PK anchoring protein', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38812204, end_position_on_the_genomic_accession=38828772, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74665, CurrentID=0, Status='live', Symbol='Lrrc48', Aliases='4930449E07Rik, Drc3', description='leucine rich repeat containing 48', other_designations='leucine-rich repeat-containing protein 48|novel Leucine Rich Repeat domain-containing protein', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60353206, end_position_on_the_genomic_accession=60394342, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  74666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74666, CurrentID=0, Status='live', Symbol='4930432K21Rik', Aliases='', description='RIKEN cDNA 4930432K21 gene', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84147963, end_position_on_the_genomic_accession=84172597, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  74670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74670, CurrentID=0, Status='live', Symbol='Zfp943', Aliases='4930432O21Rik', description='zinc finger prtoein 943', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21962559, end_position_on_the_genomic_accession=21994366, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74673, CurrentID=0, Status='live', Symbol='Spdye4a', Aliases='4930445A17Rik, 4930451F05Rik, RINGO E4A, Spdy2, Spdye4, Spy/Ringo B, spdyb', description='speedy/RINGO cell cycle regulator family, member E4A', other_designations='mSpy/Ringo B|mSpy/Ringo E4A|rapid inducer of G2/M progression in oocytes B|rapid inducer of G2/M progression in oocytes E4A|speedy homolog 2|speedy homolog E4|speedy protein B', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143216316, end_position_on_the_genomic_accession=143226902, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74675, CurrentID=0, Status='live', Symbol='Ptchd3', Aliases='4930451E13Rik, 4933440L20Rik, RNDEu-3', description='patched domain containing 3', other_designations='RND-type protein RNDEu-3', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121830212, end_position_on_the_genomic_accession=121857143, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74682, CurrentID=0, Status='live', Symbol='Wdr35', Aliases='4930459M12Rik, 4931430C06, mKIAA1336', description='WD repeat domain 35', other_designations='', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8973992, end_position_on_the_genomic_accession=9028848, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  74685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74685, CurrentID=0, Status='live', Symbol='Lrrc74b', Aliases='4930451C15Rik, mCG12354, mCP17390', description='leucine rich repeat containing 74B', other_designations='leucine-rich repeat-containing protein LOC400891 homolog', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17544465, end_position_on_the_genomic_accession=17561368, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74686, CurrentID=0, Status='live', Symbol='Slc25a54', Aliases='4930443G12Rik, SCaMC-1like', description='solute carrier family 25, member 54', other_designations='solute carrier family 25 member 24', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=109080499, end_position_on_the_genomic_accession=109120284, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74691, CurrentID=0, Status='live', Symbol='Tdrd9', Aliases='4930441E05Rik', description='tudor domain containing 9', other_designations='tudor domain-containing protein 9', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111971474, end_position_on_the_genomic_accession=112068854, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  74694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74694, CurrentID=0, Status='live', Symbol='Tbc1d30', Aliases='4930505D03Rik', description='TBC1 domain family, member 30', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121263819, end_position_on_the_genomic_accession=121311189, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  74703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74703, CurrentID=0, Status='live', Symbol='Ccdc7a', Aliases='4930517G15Rik, 4930540C21Rik, Biot2, Ccdc7', description='coiled-coil domain containing 7A', other_designations='coiled-coil domain containing 7|putative cancer/testis antigen|testis-specific antigen', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=128961395, end_position_on_the_genomic_accession=129065564, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74708, CurrentID=0, Status='live', Symbol='Pih1d3', Aliases='4930521A18Rik', description='PIH1 domain containing 3', other_designations='PIH1 domain-containing protein 3', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=31222839, end_position_on_the_genomic_accession=31224287, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  74711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74711, CurrentID=0, Status='live', Symbol='Ttll9', Aliases='1700016F23Rik, 4930509O20Rik, AW046572', description='tubulin tyrosine ligase-like family, member 9', other_designations='putative polyglutamylase|tubulin tyrosine ligase like family 9|tubulin--tyrosine ligase-like protein 9', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152950811, end_position_on_the_genomic_accession=153008482, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74716, CurrentID=0, Status='live', Symbol='Wbp2nl', Aliases='4930521I23Rik, PAWP', description='WBP2 N-terminal like', other_designations='WW domain-binding protein 2-like', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82298984, end_position_on_the_genomic_accession=82314558, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74717, CurrentID=0, Status='live', Symbol='Spata17', Aliases='1700065F16Rik, 4930504I07Rik, 4930513F16Rik, MSRG-11', description='spermatogenesis associated 17', other_designations='spermatogenesis-related protein 11', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=187009614, end_position_on_the_genomic_accession=187215462, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74718, CurrentID=0, Status='live', Symbol='Snx16', Aliases='4930522N22Rik, AI117577, AI747675', description='sorting nexin 16', other_designations='', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10417817, end_position_on_the_genomic_accession=10440131, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  74720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74720, CurrentID=0, Status='live', Symbol='Tmem114', Aliases='4930511J11Rik, Cldn26', description='transmembrane protein 114', other_designations='claudin 26|claudin-26', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8409276, end_position_on_the_genomic_accession=8425136, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  74729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74729, CurrentID=0, Status='live', Symbol='Setmar', Aliases='5830404F24Rik, Etet2', description='SET domain without mariner transposase fusion', other_designations='SET domain and mariner transposase fusion gene-containing protein homolog|SET domain without mariner transposase fusion protein|epidermal tumor expressed transcript 2', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=108065045, end_position_on_the_genomic_accession=108077127, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  74732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74732, CurrentID=0, Status='live', Symbol='Stx11', Aliases='5830405C08Rik', description='syntaxin 11', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=12939983, end_position_on_the_genomic_accession=12964500, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74734, CurrentID=0, Status='live', Symbol='Rhoh', Aliases='5830400A04Rik, AU019774, Arhh', description='ras homolog gene family, member H', other_designations='', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65861399, end_position_on_the_genomic_accession=65896700, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74735, CurrentID=0, Status='live', Symbol='Trim14', Aliases='5830400N10Rik, pub', description='tripartite motif-containing 14', other_designations='PU.1 binding protein Pub', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46505072, end_position_on_the_genomic_accession=46536141, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74737, CurrentID=0, Status='live', Symbol='Pcf11', Aliases='2500001H09Rik, 5730417B17Rik, C77803', description='cleavage and polyadenylation factor subunit homolog (S. cerevisiae)', other_designations='', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=92643458, end_position_on_the_genomic_accession=92670087, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  74741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74741, CurrentID=0, Status='live', Symbol='C2cd5', Aliases='5730419I09Rik, C030008B15Rik, CDP138', description='C2 calcium-dependent domain containing 5', other_designations='138 kDa C2 domain-containing phosphoprotein', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=143010920, end_position_on_the_genomic_accession=143100603, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  74747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74747, CurrentID=0, Status='live', Symbol='Ddit4', Aliases='5830413E08Rik, AA415483, REDD1, Rtp801, dig2', description='DNA-damage-inducible transcript 4', other_designations='DNA-damage-inducible transcript 4 protein|HIF-1 responsive RTP801|HIF-1 responsive protein RTP801|REDD-1|dexamethasone-induced gene 2 protein|protein regulated in development and DNA damage response 1', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59949675, end_position_on_the_genomic_accession=59951770, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  74748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74748, CurrentID=0, Status='live', Symbol='Slamf8', Aliases='5830408F06Rik, Blame, SBBI42', description='SLAM family member 8', other_designations='B lymphocyte activator macrophage expressed|B-lymphocyte activator macrophage expressed', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172581377, end_position_on_the_genomic_accession=172591191, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  74753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74753, CurrentID=0, Status='live', Symbol='Trmo', Aliases='5830415F09Rik, AV014846, Nap1', description='tRNA methyltransferase O', other_designations='likely ortholog of H. sapiens chromosome 9 open reading frame 156 (C9orf156)|thioesterase NAP1', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46376504, end_position_on_the_genomic_accession=46393557, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74754, CurrentID=0, Status='live', Symbol='Dhcr24', Aliases='2310076D10Rik, 5830417J06Rik, mKIAA0018', description='24-dehydrocholesterol reductase', other_designations='3-beta-hydroxysterol delta-24 reductase|3-beta-hydroxysterol delta-24-reductase|seladin-1', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106561038, end_position_on_the_genomic_accession=106589113, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74760, CurrentID=0, Status='live', Symbol='Rab3il1', Aliases='1200014K04Rik, AI115013, C76746, Rab3ail1', description='RAB3A interacting protein (rabin3)-like 1', other_designations='guanine nucleotide exchange factor for Rab3A|rab-3A-interacting-like protein 1|rab3A-interacting-like protein 1|rabin3-like 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10001674, end_position_on_the_genomic_accession=10038380, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  74761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74761, CurrentID=0, Status='live', Symbol='Mxra8', Aliases='1200013A08Rik, 1700095D18Rik, AI131686, Asp3, Dicam', description='matrix-remodelling associated 8', other_designations='adipocyte-specific protein 3|dual Ig domain containing cell adhesion molecule|limitrin', map_location='4 87.58 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155839680, end_position_on_the_genomic_accession=155844102, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  74762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74762, CurrentID=0, Status='live', Symbol='Mdga1', Aliases='1200011I03Rik, GPIM, Mamdc3', description='MAM domain containing glycosylphosphatidylinositol anchor 1', other_designations='mdga1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29827956, end_position_on_the_genomic_accession=29888429, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74763, CurrentID=0, Status='live', Symbol='Naa60', Aliases='1200013P24Rik, AI315146, Nat15', description='N(alpha)-acetyltransferase 60, NatF catalytic subunit', other_designations='N-acetyltransferase 15 (GCN5-related, putative)|natF catalytic subunit', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3872501, end_position_on_the_genomic_accession=3904781, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  74764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74764, CurrentID=0, Status='live', Symbol='Klc4', Aliases='1200014P03Rik, AA408085, AW146303, Knsl8', description='kinesin light chain 4', other_designations='KLC 4|kinesin-like 8|kinesin-like protein 8', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46630624, end_position_on_the_genomic_accession=46646167, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  74766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74766, CurrentID=0, Status='live', Symbol='Yipf2', Aliases='1300010K09Rik', description='Yip1 domain family, member 2', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21588682, end_position_on_the_genomic_accession=21592831, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  74769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74769, CurrentID=0, Status='live', Symbol='Pik3cb', Aliases='1110001J02Rik, AI447572, p110beta', description='phosphatidylinositol 3-kinase, catalytic, beta polypeptide', other_designations='PI3-kinase subunit beta|PI3K-beta|PI3Kbeta|phosphatidylinositol 4,5-bisphosphate 3-kinase 110 kDa catalytic subunit beta|phosphatidylinositol-4,5-bisphosphate 3-kinase 110 kDa catalytic subunit beta|phosphatidylinositol-4,5-bisphosphate 3-kinase catalytic subunit beta isoform|ptdIns-3-kinase subunit beta|ptdIns-3-kinase subunit p110-beta', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99038401, end_position_on_the_genomic_accession=99140211, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  74770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74770, CurrentID=0, Status='live', Symbol='Hhatl', Aliases='1110011D13Rik, Gup1', description='hedgehog acyltransferase-like', other_designations='glycerol uptake/transporter homolog|hedgehog acyltransferase-like protein', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121784016, end_position_on_the_genomic_accession=121792525, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74772, CurrentID=0, Status='live', Symbol='Atp13a2', Aliases='1110012E06Rik, AA589443', description='ATPase type 13A2', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140986865, end_position_on_the_genomic_accession=141007701, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  74775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74775, CurrentID=0, Status='live', Symbol='Lmbr1l', Aliases='1110013E13Rik, AA415653, D15Ertd735e, Limr', description='limb region 1 like', other_designations='lipocalin-interacting membrane receptor|uteroglobin receptor', map_location='15 55.1 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98903914, end_position_on_the_genomic_accession=98918550, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  74776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74776, CurrentID=0, Status='live', Symbol='Ppa2', Aliases='1110013G13Rik, Sid6306', description='pyrophosphatase (inorganic) 2', other_designations='PPase 2|pyrophosphate phospho-hydrolase 2', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=133309159, end_position_on_the_genomic_accession=133378235, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  74777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74777, CurrentID=0, Status='live', Symbol='Sepn1', Aliases='1110019I12Rik, AI414492', description='selenoprotein N, 1', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134537892, end_position_on_the_genomic_accession=134552166, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74778, CurrentID=0, Status='live', Symbol='Rrp7a', Aliases='1110014J01Rik, AA408146, Kheg1', description='ribosomal RNA processing 7 homolog A (S. cerevisiae)', other_designations='gastric cancer antigen Zg14 homolog|kidney highly expressed protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83115846, end_position_on_the_genomic_accession=83122826, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74781, CurrentID=0, Status='live', Symbol='Wipi2', Aliases='1110018O08Rik, 2510001I10Rik', description='WD repeat domain, phosphoinositide interacting 2', other_designations='WIPI-2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142629554, end_position_on_the_genomic_accession=142669672, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  74782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74782, CurrentID=0, Status='live', Symbol='Glt8d2', Aliases='1110021D20Rik', description='glycosyltransferase 8 domain containing 2', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82650428, end_position_on_the_genomic_accession=82694113, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  493061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493061, CurrentID=0, Status='live', Symbol='Hsl1', Aliases='', description='hyperoxia susceptibility locus 1', other_designations='', map_location='2 48.5 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  74838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74838, CurrentID=0, Status='live', Symbol='Naa15', Aliases='5730450D16Rik, 6330400I15, ASTBDN, Narg1, Tbdn-1, mNAT1', description='N(alpha)-acetyltransferase 15, NatA auxiliary subunit', other_designations='N-terminal acetyltransferase 1|N-terminal aceyltransferase 1|NMDA receptor-regulated gene 1|NMDA receptor-regulated protein 1|protein tubedown-1|tubedown', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51416016, end_position_on_the_genomic_accession=51475985, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  74840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74840, CurrentID=0, Status='live', Symbol='Manf', Aliases='3230402M22Rik, AA407711, AA408789, AA673178, Armet, D17914, D18Mgi17', description='mesencephalic astrocyte-derived neurotrophic factor', other_designations='arginine-rich protein|arginine-rich, mutated in early stage tumors', map_location='9 57.98 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106868310, end_position_on_the_genomic_accession=106891985, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74841, CurrentID=0, Status='live', Symbol='Usp38', Aliases='4631402N15Rik, 4833420O05Rik, AA536967, AU044701, AW544820, mKIAA1891', description='ubiquitin specific peptidase 38', other_designations='deubiquitinating enzyme 38|ubiquitin specific protease 38|ubiquitin thioesterase 38|ubiquitin thiolesterase 38|ubiquitin-specific-processing protease 38', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=80980724, end_position_on_the_genomic_accession=81015031, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74843, CurrentID=0, Status='live', Symbol='Mss51', Aliases='4833444M15Rik, Zmynd17', description='MSS51 mitochondrial translational activator', other_designations='zinc finger MYND domain-containing protein 17|zinc finger, MYND domain containing 17', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20482862, end_position_on_the_genomic_accession=20498131, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  74847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74847, CurrentID=0, Status='live', Symbol='Meikin', Aliases='4930404A10Rik', description='meiotic kinetochore factor', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54370652, end_position_on_the_genomic_accession=54426790, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  74851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74851, CurrentID=0, Status='live', Symbol='Spin2-ps1', Aliases='Gm14354, OTTMUSG00000016414', description='spindlin family, member 2, pseudogene 1', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=3835879, end_position_on_the_genomic_accession=3836656, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  74854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74854, CurrentID=0, Status='live', Symbol='4930402F06Rik', Aliases='', description='RIKEN cDNA 4930402F06 gene', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35375562, end_position_on_the_genomic_accession=35397174, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74855, CurrentID=0, Status='live', Symbol='Fam228a', Aliases='4930417G10Rik', description='family with sequence similarity 228, member A', other_designations='UPF0638 protein C2orf84 homolog', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4697272, end_position_on_the_genomic_accession=4738468, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  74868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74868, CurrentID=0, Status='live', Symbol='Tmem65', Aliases='2610029O13Rik, 4930438D12Rik', description='transmembrane protein 65', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58782269, end_position_on_the_genomic_accession=58823659, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  12478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12478, CurrentID=0, Status='live', Symbol='Cd19', Aliases='AW495831', description='CD19 antigen', other_designations='differentiation antigen CD19', map_location='7 F3-F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126408448, end_position_on_the_genomic_accession=126415140, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  74890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74890, CurrentID=0, Status='live', Symbol='Morn3', Aliases='4930438O03Rik', description='MORN repeat containing 3', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123033392, end_position_on_the_genomic_accession=123047036, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  74895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74895, CurrentID=0, Status='live', Symbol='Ccdc181', Aliases='4930455F23Rik, AI427889', description='coiled-coil domain containing 181', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164275598, end_position_on_the_genomic_accession=164287847, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74901, CurrentID=0, Status='live', Symbol='Kbtbd11', Aliases='2900016B01Rik, 4930465M17Rik, mKIAA0711', description='kelch repeat and BTB (POZ) domain containing 11', other_designations='', map_location='8 A1.1-2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=15011011, end_position_on_the_genomic_accession=15033333, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  74910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74910, CurrentID=0, Status='live', Symbol='4930480E11Rik', Aliases='', description='RIKEN cDNA 4930480E11 gene', other_designations='', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=78369643, end_position_on_the_genomic_accession=78371128, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  74914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74914, CurrentID=0, Status='live', Symbol='Cylc2', Aliases='4930488P18Rik', description='cylicin, basic protein of sperm head cytoskeleton 2', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=51216667, end_position_on_the_genomic_accession=51229963, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  74915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74915, CurrentID=0, Status='live', Symbol='Atp6v1e2', Aliases='4930500C14Rik, Atp6e1, E1', description='ATPase, H+ transporting, lysosomal V1 subunit E2', other_designations='ATPase, H+ transporting, V1 subunit E-like 2|ATPase, H+ transporting, lysosomal 31kDa, V1 subunit E-like 2|V-ATPase subunit E 2|lysosomal 31kDa|vacuolar proton pump subunit E 2', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=86944109, end_position_on_the_genomic_accession=86947887, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  74918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74918, CurrentID=0, Status='live', Symbol='Iqca', Aliases='4930465P12Rik, 4930585L22Rik, BB2885051, Iqca', description='IQ motif containing with AAA domain', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90042132, end_position_on_the_genomic_accession=90153401, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  74919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74919, CurrentID=0, Status='live', Symbol='Slc35f6', Aliases='4930471M23Rik, 5730478O19Rik, ANT2BP, P40, Tango9', description='solute carrier family 35, member F6', other_designations='ANT2-binding protein|transmembrane protein C2orf18 homolog|transport and Golgi organization 9 homolog', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30647939, end_position_on_the_genomic_accession=30659729, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  74934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74934, CurrentID=0, Status='live', Symbol='Armc4', Aliases='4930463I21Rik, b2b227.1Clo, b2b643Clo', description='armadillo repeat containing 4', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=7088233, end_position_on_the_genomic_accession=7301128, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  74954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74954, CurrentID=0, Status='live', Symbol='4930503E14Rik', Aliases='', description='RIKEN cDNA 4930503E14 gene', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=44163061, end_position_on_the_genomic_accession=44172602, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  74978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74978, CurrentID=0, Status='live', Symbol='Lrriq1', Aliases='4930503E15Rik, Gm1557', description='leucine-rich repeats and IQ motif containing 1', other_designations='IQ domain-containing protein LRRIQ1|eucine-rich repeats and IQ motif containing 1', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=103044377, end_position_on_the_genomic_accession=103236331, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  74987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74987, CurrentID=0, Status='live', Symbol='Cldn34d', Aliases='4930468A15Rik', description='claudin 34D', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=76582610, end_position_on_the_genomic_accession=76622754, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  74996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74996, CurrentID=0, Status='live', Symbol='Usp47', Aliases='4930502N04Rik, A630020C16Rik', description='ubiquitin specific peptidase 47', other_designations='deubiquitinating enzyme 47|ubiquitin specific protease 47|ubiquitin thioesterase 47|ubiquitin thiolesterase 47|ubiquitin-specific-processing protease 47', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=112022808, end_position_on_the_genomic_accession=112111386, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  74998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=74998, CurrentID=0, Status='live', Symbol='Rab11fip2', Aliases='4930470G04Rik, A830046J09Rik, AW558126, Nrip11', description='RAB11 family interacting protein 2 (class I)', other_designations='', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=59760441, end_position_on_the_genomic_accession=59943666, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75002, CurrentID=0, Status='live', Symbol='Tmprss12', Aliases='4930478A21Rik', description='transmembrane (C-terminal) protease, serine 12', other_designations='transmembrane protease, serine 12', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100280837, end_position_on_the_genomic_accession=100293053, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75007, CurrentID=0, Status='live', Symbol='Fam63a', Aliases='1810005H09Rik, 4930504E06Rik, AA562754, AA960384, AW561904, cI-40, mKIAA1390, ni', description='family with sequence similarity 63, member A', other_designations='NF-E2 inducible protein', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95281590, end_position_on_the_genomic_accession=95296164, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  75010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75010, CurrentID=0, Status='live', Symbol='Tmbim7', Aliases='4930403J02Rik, 4930500J03Rik, 4930511M11Rik, Lgf5, Tmbim1b', description='transmembrane BAX inhibitor motif containing 7', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3657004, end_position_on_the_genomic_accession=3680677, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  75011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75011, CurrentID=0, Status='live', Symbol='4930488N24Rik', Aliases='', description='RIKEN cDNA 4930488N24 gene', other_designations='', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14101817, end_position_on_the_genomic_accession=14109145, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75013, CurrentID=0, Status='live', Symbol='4930502E18Rik', Aliases='Ct55', description='RIKEN cDNA 4930502E18 gene', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53724831, end_position_on_the_genomic_accession=53738565, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75015, CurrentID=0, Status='live', Symbol='4930503B20Rik', Aliases='1700013A07Rik, 1700095G21Rik, 4930483N21Rik, Gm1652, Samd13', description='RIKEN cDNA 4930503B20 gene', other_designations='SAM domain-containing protein 13', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146645189, end_position_on_the_genomic_accession=146686801, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  75019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75019, CurrentID=0, Status='live', Symbol='Rnase10', Aliases='4930474F22Rik, Rah1', description='ribonuclease, RNase A family, 10 (non-active)', other_designations='RNase 10|protein train A|ribonuclease A family, member 10|ribonuclease A h1|ribonuclease-like protein 10', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51007751, end_position_on_the_genomic_accession=51010761, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75029, CurrentID=0, Status='live', Symbol='Purg', Aliases='4930486B15Rik', description='purine-rich element binding protein G', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33385614, end_position_on_the_genomic_accession=33417472, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12505, CurrentID=0, Status='live', Symbol='Cd44', Aliases='AU023126, AW121933, AW146109, HERMES, Ly-24, Pgp-1', description='CD44 antigen', other_designations='ECMR-III|GP90 lymphocyte homing/adhesion receptor|HUTCH-I|PGP-I|extracellular matrix receptor III|hermes antigen|hyaluronate receptor|lymphocyte antigen 24|phagocytic glycoprotein 1|phagocytic glycoprotein I', map_location='2 54.13 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=102811141, end_position_on_the_genomic_accession=102901669, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  75033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75033, CurrentID=0, Status='live', Symbol='Mei4', Aliases='4930486G11Rik', description='meiotic double-stranded break formation protein 4', other_designations='UPF0623 protein|meiosis-specific, MEI4 homolog', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=81863647, end_position_on_the_genomic_accession=82206007, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  230959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230959, CurrentID=0, Status='live', Symbol='Ajap1', Aliases='Gm573', description='adherens junction associated protein 1', other_designations='likely ortholog of H. sapiens transmembrane protein SHREW1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=153373221, end_position_on_the_genomic_accession=153484013, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75040, CurrentID=0, Status='live', Symbol='Efcab10', Aliases='4930504H06Rik', description='EF-hand calcium binding domain 10', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=33394854, end_position_on_the_genomic_accession=33401269, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  75050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75050, CurrentID=0, Status='live', Symbol='Kif27', Aliases='4930517I18Rik', description='kinesin family member 27', other_designations='kinesin-related protein KIF27', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58287516, end_position_on_the_genomic_accession=58364745, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  75051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75051, CurrentID=0, Status='live', Symbol='Ccdc173', Aliases='4930525K21Rik, 4930578N16Rik', description='coiled-coil domain containing 173', other_designations='coiled -coil domain containing 173', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69757392, end_position_on_the_genomic_accession=69789579, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75062, CurrentID=0, Status='live', Symbol='Sf3a3', Aliases='4930512K19Rik, 60kDa', description='splicing factor 3a, subunit 3', other_designations='SAP 61|SF3a60|spliceosome-associated protein 61', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124714861, end_position_on_the_genomic_accession=124732423, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  75064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75064, CurrentID=0, Status='live', Symbol='Zcchc13', Aliases='4930513O09Rik, Cnbp2', description='zinc finger, CCHC domain containing 13', other_designations='cellular nucleic acid binding protein 2|cellular nucleic acid binding-like protein', map_location='X 46.26 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103630586, end_position_on_the_genomic_accession=103631664, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75071, CurrentID=0, Status='live', Symbol='4930524N10Rik', Aliases='', description='RIKEN cDNA 4930524N10 gene', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154339225, end_position_on_the_genomic_accession=154343392, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75079, CurrentID=0, Status='live', Symbol='Zbtb49', Aliases='4930518A03Rik, Zfp509, Znf509', description='zinc finger and BTB domain containing 49', other_designations='zinc finger protein 509', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38189732, end_position_on_the_genomic_accession=38220449, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  75083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75083, CurrentID=0, Status='live', Symbol='Usp50', Aliases='1700086G18Rik, 4930511O11Rik', description='ubiquitin specific peptidase 50', other_designations='deubiquitinating enzyme 50|ubiquitin specific protease 50|ubiquitin thioesterase 50|ubiquitin thiolesterase 50|ubiquitin-specific-processing protease 50', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126761050, end_position_on_the_genomic_accession=126783460, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75087, CurrentID=0, Status='live', Symbol='4930505A04Rik', Aliases='', description='RIKEN cDNA 4930505A04 gene', other_designations='', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30426006, end_position_on_the_genomic_accession=30471829, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75089, CurrentID=0, Status='live', Symbol='Uhrf1bp1l', Aliases='2010319N22Rik, 4930506D01Rik, AA589495, C80672, E030041M21Rik, mKIAA0701', description='UHRF1 (ICBP90) binding protein 1-like', other_designations='', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=89744991, end_position_on_the_genomic_accession=89819891, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  75097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75097, CurrentID=0, Status='live', Symbol='Ube2dnl2', Aliases='4930524E20Rik', description='ubiquitin-conjugating enzyme E2D N-terminal like 2', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=114907582, end_position_on_the_genomic_accession=114908510, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75099, CurrentID=0, Status='live', Symbol='Lysmd4', Aliases='4930506D23Rik', description='LysM, putative peptidoglycan-binding, domain containing 4', other_designations='', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=67222544, end_position_on_the_genomic_accession=67228468, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  75104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75104, CurrentID=0, Status='live', Symbol='Mmd2', Aliases='4930518M15Rik, C88001', description='monocyte to macrophage differentiation-associated 2', other_designations='progestin and adipoQ receptor family member X', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142563474, end_position_on_the_genomic_accession=142608814, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75106, CurrentID=0, Status='live', Symbol='4930519F16Rik', Aliases='Ppnx', description='RIKEN cDNA 4930519F16 gene', other_designations='novel gene likely coding for a serine proteases, trypsin family/Chymotrypsin serine protease family (S1) containing protein', map_location='X 45.97 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103182991, end_position_on_the_genomic_accession=103256632, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  99899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99899, CurrentID=0, Status='live', Symbol='Ifi44', Aliases='A430056A10Rik, AW261460, MTAP44, p44', description='interferon-induced protein 44', other_designations='microtubule-associated protein 44', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=151730922, end_position_on_the_genomic_accession=151749959, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75122, CurrentID=0, Status='live', Symbol='Zc2hc1b', Aliases='4930519B02Rik, Fam164b', description='zinc finger, C2HC-type containing 1B', other_designations='family with sequence similarity 164, member B|protein FAM164B', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13143095, end_position_on_the_genomic_accession=13178038, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  75124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75124, CurrentID=0, Status='live', Symbol='Nxnl2', Aliases='4930519N16Rik, Rdcvf2', description='nucleoredoxin-like 2', other_designations='rod-derived cone viability factor 2', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51170114, end_position_on_the_genomic_accession=51175191, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  12522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12522, CurrentID=0, Status='live', Symbol='Cd83', Aliases='', description='CD83 antigen', other_designations='mCD83', map_location='13 A4-5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=43785112, end_position_on_the_genomic_accession=43803133, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75136, CurrentID=0, Status='live', Symbol='Rsph10b', Aliases='4930526H21Rik2, Rsph10b', description='radial spoke head 10 homolog B (Chlamydomonas)', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143933107, end_position_on_the_genomic_accession=143985719, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  75137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75137, CurrentID=0, Status='live', Symbol='Rprd2', Aliases='2810036A19Rik, 4930535B03Rik, 6720469I21Rik, AL022841, AL022940, AU021304, BB077382, mKIAA0460', description='regulation of nuclear pre-mRNA domain containing 2', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95759873, end_position_on_the_genomic_accession=95819723, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75140, CurrentID=0, Status='live', Symbol='Slxl1', Aliases='4930527E24Rik, OTTMUSG00000017407', description='Slx-like 1', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=55226876, end_position_on_the_genomic_accession=55243736, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75141, CurrentID=0, Status='live', Symbol='Rasd2', Aliases='4930526B11Rik, AU045414, Rhes, TEM-2, TEM2', description='RASD family, member 2', other_designations='ras homolog enriched in striatum', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=75213944, end_position_on_the_genomic_accession=75224113, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75146, CurrentID=0, Status='live', Symbol='Tmem180', Aliases='4930449A08Rik, 4930538D17Rik', description='transmembrane protein 180', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46356877, end_position_on_the_genomic_accession=46375254, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  75172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75172, CurrentID=0, Status='live', Symbol='Ccdc146', Aliases='4930528G09Rik, mKIAA1505', description='coiled-coil domain containing 146', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21292961, end_position_on_the_genomic_accession=21424677, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  75173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75173, CurrentID=0, Status='live', Symbol='Tex38', Aliases='4930544O15Rik, Theg4', description='testis expressed 38', other_designations='ATPAF1 antisense RNA 1|ATPAF1 antisense gene protein 1|testis highly expressed protein 4', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115779831, end_position_on_the_genomic_accession=115782759, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75178, CurrentID=0, Status='live', Symbol='Meiob', Aliases='4930528F23Rik, MLZ-675', description='meiosis specific with OB domains', other_designations='protein expressed in male leptotene and zygotene spermatocytes 675', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24804301, end_position_on_the_genomic_accession=24839788, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  75180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75180, CurrentID=0, Status='live', Symbol='4930538K18Rik', Aliases='', description='RIKEN cDNA 4930538K18 gene', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119205055, end_position_on_the_genomic_accession=119225722, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  75185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75185, CurrentID=0, Status='live', Symbol='Samt4', Aliases='4930542N07Rik', description='spermatogenesis associated multipass transmembrane protein 4', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154482002, end_position_on_the_genomic_accession=154484682, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75196, CurrentID=0, Status='live', Symbol='Ankrd7', Aliases='4930532L20Rik, AV278427', description='ankyrin repeat domain 7', other_designations='', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=18861634, end_position_on_the_genomic_accession=18879586, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  75199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75199, CurrentID=0, Status='live', Symbol='Rhox2a', Aliases='4930539I12Rik, Rhox2, Rhox2.1', description='reproductive homeobox 2A', other_designations='reproductive homeobox on X chromosome, 2', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37244402, end_position_on_the_genomic_accession=37249690, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75202, CurrentID=0, Status='live', Symbol='Spaca6', Aliases='4930546H06Rik, 9530079D04Rik, B230206P06Rik, Ncrna00085, SPACA6A, let-7e, mir-125a, mir-99b', description='sperm acrosome associated 6', other_designations='immunoglobulin domain-containing protein', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17824821, end_position_on_the_genomic_accession=17843010, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  75204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75204, CurrentID=0, Status='live', Symbol='Orly', Aliases='4930529H12Rik, 4933412F11Rik, Asty(rec)', description='oppositely-transcribed, rearranged locus on the Y', other_designations='oppositely-transcribed ressorted locus Y Chr', map_location='', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=78564791, end_position_on_the_genomic_accession=78575409, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75209, CurrentID=0, Status='live', Symbol='Sv2c', Aliases='4930527L09Rik, AI841971', description='synaptic vesicle glycoprotein 2c', other_designations='synaptic vesicle protein 2C', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95959443, end_position_on_the_genomic_accession=96132577, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  75210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75210, CurrentID=0, Status='live', Symbol='Prr3', Aliases='4930540G07Rik, Cat56', description='proline-rich polypeptide 3', other_designations='MHC class I region proline-rich protein Cat56', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35972539, end_position_on_the_genomic_accession=35979825, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75212, CurrentID=0, Status='live', Symbol='Rnf121', Aliases='4930544L10Rik', description='ring finger protein 121', other_designations='', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102019145, end_position_on_the_genomic_accession=102072428, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75216, CurrentID=0, Status='live', Symbol='Cep128', Aliases='4930534B04Rik, 5430424K18Rik, AI449708, AI480494, E230024F20', description='centrosomal protein 128', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=90998490, end_position_on_the_genomic_accession=91384416, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  75219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75219, CurrentID=0, Status='live', Symbol='Dusp18', Aliases='4930527G07Rik, AI430870, BB028064, DSP18, DUSP20, LMWDSP20', description='dual specificity phosphatase 18', other_designations='low molecular weight dual specificity phosphatase 20', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3895139, end_position_on_the_genomic_accession=3901299, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75221, CurrentID=0, Status='live', Symbol='Dpp3', Aliases='4930533O14Rik, C86324', description='dipeptidylpeptidase 3', other_designations='DPP III|dipeptidyl aminopeptidase III|dipeptidyl arylamidase III|dipeptidyl peptidase III|dipeptidyl-peptidase 3|dipeptidyl-peptidase III|dipeptidylpeptidase III|enkephalinase B', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4907229, end_position_on_the_genomic_accession=4928287, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  75234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75234, CurrentID=0, Status='live', Symbol='Rnf19b', Aliases='4930534K13Rik, 4930555L03Rik, Ibrdc3', description='ring finger protein 19B', other_designations='IBR domain containing 3|IBR domain-containing protein 3|natural killer lytic-associated molecule', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129058174, end_position_on_the_genomic_accession=129085886, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75258, CurrentID=0, Status='live', Symbol='4930563M21Rik', Aliases='Rfpl3s', description='RIKEN cDNA 4930563M21 gene', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55962587, end_position_on_the_genomic_accession=56026518, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  75266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75266, CurrentID=0, Status='live', Symbol='Tomm20l', Aliases='4930553D19Rik', description='translocase of outer mitochondrial membrane 20 homolog (yeast)-like', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=71111428, end_position_on_the_genomic_accession=71123222, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75272, CurrentID=0, Status='live', Symbol='Erich6b', Aliases='4930564B18Rik', description='glutamate rich 6B', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75642259, end_position_on_the_genomic_accession=75667137, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75273, CurrentID=0, Status='live', Symbol='Pelp1', Aliases='4930563C04Rik, MNAR', description='proline, glutamic acid and leucine rich protein 1', other_designations='modulator of non-genomic activity of estrogen receptor', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70392881, end_position_on_the_genomic_accession=70410031, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  75275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75275, CurrentID=0, Status='live', Symbol='Tmco5b', Aliases='4930563P21Rik', description='transmembrane and coiled-coil domains 5B', other_designations='', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=113285095, end_position_on_the_genomic_accession=113297190, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  75276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75276, CurrentID=0, Status='live', Symbol='Ppp1r1c', Aliases='4930565M23Rik, 4930579P15Rik', description='protein phosphatase 1, regulatory (inhibitor) subunit 1C', other_designations='', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=79707780, end_position_on_the_genomic_accession=79818546, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75284, CurrentID=0, Status='live', Symbol='Bcdin3d', Aliases='4930556P03Rik, AV138748', description='BCDIN3 domain containing', other_designations='BCDIN3 domain-containing protein|probable methyltransferase BCDIN3D', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99470084, end_position_on_the_genomic_accession=99474730, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  75288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75288, CurrentID=0, Status='live', Symbol='Slc35f4', Aliases='4930550L21Rik', description='solute carrier family 35, member F4', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49298519, end_position_on_the_genomic_accession=49525837, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  75291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75291, CurrentID=0, Status='live', Symbol='Zbtb3', Aliases='4930563M09Rik', description='zinc finger and BTB domain containing 3', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8802223, end_position_on_the_genomic_accession=8804854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75292, CurrentID=0, Status='live', Symbol='Prkd3', Aliases='4930557O20Rik, 5730497N19Rik, PKD3, Pkcnu, Prkcn', description='protein kinase D3', other_designations='nPKC-nu|protein kinase C nu type|protein kinase C, nu', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78949405, end_position_on_the_genomic_accession=79020816, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  493146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493146, CurrentID=0, Status='live', Symbol='Mnotch', Aliases='', description='modifier of Notch', other_designations='', map_location='1 81.08 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  75296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75296, CurrentID=0, Status='live', Symbol='Fgfr1op', Aliases='4930553O10Rik, FOP', description='Fgfr1 oncogene partner', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8165485, end_position_on_the_genomic_accession=8196477, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  231003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231003, CurrentID=0, Status='live', Symbol='Klhl17', Aliases='AL022703, BC058738', description='kelch-like 17', other_designations='actinfilin', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156229044, end_position_on_the_genomic_accession=156234857, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  75302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75302, CurrentID=0, Status='live', Symbol='Asxl2', Aliases='4930556B16Rik, mKIAA1685', description='additional sex combs like 2 (Drosophila)', other_designations='additional sex combs-like protein 2|polycomb group protein ASXH2 homolog', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3426857, end_position_on_the_genomic_accession=3506852, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  75304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75304, CurrentID=0, Status='live', Symbol='4930563E22Rik', Aliases='', description='RIKEN cDNA 4930563E22 gene', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72215138, end_position_on_the_genomic_accession=72218450, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75305, CurrentID=0, Status='live', Symbol='Ankrd53', Aliases='4930564N15Rik', description='ankyrin repeat domain 53', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83762646, end_position_on_the_genomic_accession=83768326, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75311, CurrentID=0, Status='live', Symbol='4930550C14Rik', Aliases='AU017961', description='RIKEN cDNA 4930550C14 gene', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53401888, end_position_on_the_genomic_accession=53438266, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  75316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75316, CurrentID=0, Status='live', Symbol='Taf1d', Aliases='2810003M17Rik, 4930553M18Rik, 5930426I11Rik, Josd3, TAF(I)41, TAFI41', description='TATA-box binding protein associated factor, RNA polymerase I, D', other_designations='Josephin domain containing 3|TATA box binding protein (Tbp)-associated factor, RNA polymerase I, D|TATA box binding protein-associated factor, RNA polymerase I, D|TATA box-binding protein-associated factor 1D|TBP-associated factor 1D|transcription initiation factor SL1/TIF-IB subunit D', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15306202, end_position_on_the_genomic_accession=15316997, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  75317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75317, CurrentID=0, Status='live', Symbol='Parpbp', Aliases='4930547N16Rik, AI503644, PARI', description='PARP1 binding protein', other_designations='PARP-1 binding protein|PARP1-binding protein', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88091398, end_position_on_the_genomic_accession=88146945, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  75320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75320, CurrentID=0, Status='live', Symbol='Etnk1', Aliases='1110061E11Rik, 4930555L11Rik, AI195356, AI841245, C80956, D6Ertd3e, EKI1', description='ethanolamine kinase 1', other_designations='EKI 1', map_location='6 75.41 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=143166866, end_position_on_the_genomic_accession=143208546, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75328, CurrentID=0, Status='live', Symbol='4930563D23Rik', Aliases='', description='RIKEN cDNA 4930563D23 gene', other_designations='uncharacterized protein ENSP00000386791 homolog', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92318763, end_position_on_the_genomic_accession=92321441, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  75329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75329, CurrentID=0, Status='live', Symbol='Atf7ip2', Aliases='4930558K11Rik, BC018510, Get-1, PSM2', description='activating transcription factor 7 interacting protein 2', other_designations='MBD1-containing chromatin-associated factor 2|protein similar to MCAF2', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10192640, end_position_on_the_genomic_accession=10249669, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  75338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75338, CurrentID=0, Status='live', Symbol='Ccdc83', Aliases='4930549K11Rik, 4930554C01Rik, 4932423M01Rik', description='coiled-coil domain containing 83', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90222539, end_position_on_the_genomic_accession=90265876, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  75339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75339, CurrentID=0, Status='live', Symbol='Mphosph8', Aliases='1500035L22Rik, 4930548G07Rik, AU040635', description='M-phase phosphoprotein 8', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56668240, end_position_on_the_genomic_accession=56697430, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  75341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75341, CurrentID=0, Status='live', Symbol='4930564C03Rik', Aliases='', description='RIKEN cDNA 4930564C03 gene', other_designations='', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=44879704, end_position_on_the_genomic_accession=44880885, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  75345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75345, CurrentID=0, Status='live', Symbol='Slamf7', Aliases='19A, 19A24, 4930560D03Rik, CRACC, CS1', description='SLAM family member 7', other_designations='leukocyte cell-surface antigen|novel Ly9', map_location='1 H2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171632403, end_position_on_the_genomic_accession=171653037, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75352, CurrentID=0, Status='live', Symbol='4930550L24Rik', Aliases='mage-k1', description='RIKEN cDNA 4930550L24 gene', other_designations='novel MAGE family domain containing protein', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=58911461, end_position_on_the_genomic_accession=58920304, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75368, CurrentID=0, Status='live', Symbol='4930558K02Rik', Aliases='', description='RIKEN cDNA 4930558K02 gene', other_designations='', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161942089, end_position_on_the_genomic_accession=161982371, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  12564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12564, CurrentID=0, Status='live', Symbol='Cdh8', Aliases='AI851472, cad8', description='cadherin 8', other_designations='', map_location='8 49.4 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=99024471, end_position_on_the_genomic_accession=99416485, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  75387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75387, CurrentID=0, Status='live', Symbol='Sirt4', Aliases='4930596O17Rik', description='sirtuin 4', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115478006, end_position_on_the_genomic_accession=115490050, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75388, CurrentID=0, Status='live', Symbol='Boll', Aliases='4930554P13Rik, 4930597B14Rik', description='bol, boule-like (Drosophila)', other_designations='', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55239385, end_position_on_the_genomic_accession=55364058, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  75395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75395, CurrentID=0, Status='live', Symbol='Hoxb5os', Aliases='0610040B09Rik, AV302770, Hoxb5/6as', description='homeobox B5 and homeobox B6, opposite strand', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96290039, end_position_on_the_genomic_accession=96310226, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75396, CurrentID=0, Status='live', Symbol='Spp2', Aliases='0610038O04Rik, pp-24, spp24', description='secreted phosphoprotein 2', other_designations='secreted phosphoprotein 2, 24kDa', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88406967, end_position_on_the_genomic_accession=88428300, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  75398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75398, CurrentID=0, Status='live', Symbol='Mrpl32', Aliases='0610033O15Rik', description='mitochondrial ribosomal protein L32', other_designations='L32mt|MRP-L32|heart-expressed gene 1 protein', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=14610301, end_position_on_the_genomic_accession=14613037, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75400, CurrentID=0, Status='live', Symbol='Defb29', Aliases='0610030I09Rik, BD-29, mBD-29', description='defensin beta 29', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152538710, end_position_on_the_genomic_accession=152540247, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75404, CurrentID=0, Status='live', Symbol='Arhgap36', Aliases='1100001E04Rik, rho-gap', description='Rho GTPase activating protein 36', other_designations='Novel RhoGAP domain containing protein|Rho-GTPase-activating protein like protein', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=49448328, end_position_on_the_genomic_accession=49500250, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  75406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75406, CurrentID=0, Status='live', Symbol='Ndufs7', Aliases='1010001M04Rik', description='NADH dehydrogenase (ubiquinone) Fe-S protein 7', other_designations='CI-20kD|NADH-ubiquinone oxidoreductase 20 kDa subunit|complex I-20kD', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80249121, end_position_on_the_genomic_accession=80256793, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  75409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75409, CurrentID=0, Status='live', Symbol='Slitrk5', Aliases='2610019D03Rik', description='SLIT and NTRK-like family, member 5', other_designations='slit and trk like gene 5', map_location='14 E3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=111675115, end_position_on_the_genomic_accession=111683144, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75410, CurrentID=0, Status='live', Symbol='Kmt2b', Aliases='2610014H22Rik, Mll2, Wbp7, mKIAA0304', description='lysine (K)-specific methyltransferase 2B', other_designations='WBP-7|WW domain binding protein 7|WW domain-binding protein 7|histone-lysine N-methyltransferase MLL4|lysine N-methyltransferase 2B|mixed lineage leukemia 2|myeloid/lymphoid or mixed-lineage leukemia protein 4 homolog|trithorax homolog 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30568855, end_position_on_the_genomic_accession=30588726, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  75415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75415, CurrentID=0, Status='live', Symbol='Arhgap12', Aliases='2810011M08Rik', description='Rho GTPase activating protein 12', other_designations='rho-type GTPase-activating protein 12', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=6024448, end_position_on_the_genomic_accession=6136098, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  75416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75416, CurrentID=0, Status='live', Symbol='Nop14', Aliases='2610033H07Rik, Nol14', description='NOP14 nucleolar protein', other_designations='NOP14 nucleolar protein homolog|nucleolar complex protein 14|probable nucleolar complex protein 14', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34638536, end_position_on_the_genomic_accession=34660148, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  75420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75420, CurrentID=0, Status='live', Symbol='Secisbp2', Aliases='2210413N07Rik, 2810012K13Rik, SBP2', description='SECIS binding protein 2', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=51650829, end_position_on_the_genomic_accession=51684044, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  75422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75422, CurrentID=0, Status='live', Symbol='Mettl5', Aliases='2810410A08Rik', description='methyltransferase like 5', other_designations='', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69871195, end_position_on_the_genomic_accession=69885984, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75423, CurrentID=0, Status='live', Symbol='Arl5a', Aliases='2410015N24Rik, 2810410P22Rik, AA408731, AW610751, Arl5, T25534', description='ADP-ribosylation factor-like 5A', other_designations='ADP-ribosylation factor-like 5', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=52397949, end_position_on_the_genomic_accession=52424874, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75425, CurrentID=0, Status='live', Symbol='Tti1', Aliases='2610036D13Rik, AI449463', description='TELO2 interacting protein 1', other_designations='TEL2-interacting protein 1 homolog|Tel2 interacting protein 1 homolog', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157981803, end_position_on_the_genomic_accession=158028477, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  75426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75426, CurrentID=0, Status='live', Symbol='Igfbpl1', Aliases='2810011G06Rik, 2810453O06Rik, IGFBPL', description='insulin-like growth factor binding protein-like 1', other_designations='IGFBP-like protein|IGFBP-rP4|IGFBP-related protein 10|insulin-like growth factor-binding-related protein 4', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45809507, end_position_on_the_genomic_accession=45826827, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75429, CurrentID=0, Status='live', Symbol='Fam183b', Aliases='3100002J23Rik, Theg6', description='family with sequence similarity 183, member B', other_designations='testis highly expressed protein 6', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58783892, end_position_on_the_genomic_accession=58802237, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75430, CurrentID=0, Status='live', Symbol='Anapc15', Aliases='3200002M19Rik, 6330414C15Rik, APC15', description='anaphase prompoting complex C subunit 15', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101881300, end_position_on_the_genomic_accession=101901849, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  75434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75434, CurrentID=0, Status='live', Symbol='1700001C02Rik', Aliases='1700047H18Rik', description='RIKEN cDNA 1700001C02 gene', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30466051, end_position_on_the_genomic_accession=30484088, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75437, CurrentID=0, Status='live', Symbol='1700006E09Rik', Aliases='', description='RIKEN cDNA 1700006E09 gene', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101987056, end_position_on_the_genomic_accession=101992264, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75444, CurrentID=0, Status='live', Symbol='1700011I03Rik', Aliases='4930583I06Rik', description='RIKEN cDNA 1700011I03 gene', other_designations='', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=57533798, end_position_on_the_genomic_accession=57731067, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76709, CurrentID=0, Status='live', Symbol='Arpc2', Aliases='2210023N03Rik, 34kDa, p34-Arc', description='actin related protein 2/3 complex, subunit 2', other_designations='arp2/3 complex 34 kDa subunit', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74236504, end_position_on_the_genomic_accession=74268213, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75452, CurrentID=0, Status='live', Symbol='Ascc2', Aliases='1700011I11Rik, 2610034L15Rik, AI482016, ASC1p100, AW046480', description='activating signal cointegrator 1 complex subunit 2', other_designations='ASC-1 complex subunit P100|trip4 complex subunit p100', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4637696, end_position_on_the_genomic_accession=4685699, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  75453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75453, CurrentID=0, Status='live', Symbol='Ccdc7b', Aliases='1700008F21Rik', description='coiled-coil domain containing 7B', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=129067064, end_position_on_the_genomic_accession=129183732, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  75454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75454, CurrentID=0, Status='live', Symbol='Phpt1', Aliases='1700008C22Rik, Php14', description='phosphohistidine phosphatase 1', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25573431, end_position_on_the_genomic_accession=25574871, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75456, CurrentID=0, Status='live', Symbol='Prps1l1', Aliases='1700011K15Rik', description='phosphoribosyl pyrophosphate synthetase 1-like 1', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=34984761, end_position_on_the_genomic_accession=34986436, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75458, CurrentID=0, Status='live', Symbol='Cklf', Aliases='1700001C14Rik, 1810018M11Rik, C32, CKLF1, CKLF3, CKLF4, CKLF52, Cklf6, HSPC224, UCK-1, Cklf', description='chemokine-like factor', other_designations='chemokine-like factor 2|chemokine-like factor 4|chemokine-like factor 6', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104250861, end_position_on_the_genomic_accession=104264936, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75459, CurrentID=0, Status='live', Symbol='Syce3', Aliases='1700007E06Rik, TSEG-2, TSEG2', description='synaptonemal complex central element protein 3', other_designations='testis-specific expressed protein 2', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89390174, end_position_on_the_genomic_accession=89410503, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  75462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75462, CurrentID=0, Status='live', Symbol='1700001C19Rik', Aliases='', description='RIKEN cDNA 1700001C19 gene', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47412734, end_position_on_the_genomic_accession=47437375, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75465, CurrentID=0, Status='live', Symbol='Dynlrb2', Aliases='1700009A04Rik, DNLC2B, Dncl2b', description='dynein light chain roadblock-type 2', other_designations='dynein light chain 2B, cytoplasmic|dynein, cytoplasmic, light chain 2B|dynein, cytoplasmic, light polypeptide 2B', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116505015, end_position_on_the_genomic_accession=116515915, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  75467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75467, CurrentID=0, Status='live', Symbol='1700011E24Rik', Aliases='GSE', description='RIKEN cDNA 1700011E24 gene', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87389570, end_position_on_the_genomic_accession=87428081, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  75469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75469, CurrentID=0, Status='live', Symbol='Spata19', Aliases='1700001L23Rik, AV047838, spergen-1', description='spermatogenesis associated 19', other_designations='spermatogenic cell-specific gene 1 protein', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=27396821, end_position_on_the_genomic_accession=27401711, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  75470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75470, CurrentID=0, Status='live', Symbol='Iqcf5', Aliases='1700007L12Rik', description='IQ motif containing F5', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106514573, end_position_on_the_genomic_accession=106516003, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75471, CurrentID=0, Status='live', Symbol='1700009N14Rik', Aliases='AI429145', description='RIKEN cDNA 1700009N14 gene', other_designations='novel Ras protein', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=39450293, end_position_on_the_genomic_accession=39451780, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75472, CurrentID=0, Status='live', Symbol='Cfap126', Aliases='1700009P17Rik, AV282292, Flattop, Fltp', description='cilia and flagella associated protein 126', other_designations='UPF0740 protein C1orf192 homolog|cilia- and flagella-associated protein 126', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171113744, end_position_on_the_genomic_accession=171126967, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  75475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75475, CurrentID=0, Status='live', Symbol='Oplah', Aliases='1700010G02Rik', description='5-oxoprolinase (ATP-hydrolysing)', other_designations='5-OPase|5-oxo-L-prolinase|pyroglutamase', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76296603, end_position_on_the_genomic_accession=76331101, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  75477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75477, CurrentID=0, Status='live', Symbol='Pfn3', Aliases='1700012P12Rik', description='profilin 3', other_designations='profilin III', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55414688, end_position_on_the_genomic_accession=55415232, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  75480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75480, CurrentID=0, Status='live', Symbol='1700003F12Rik', Aliases='', description='RIKEN cDNA 1700003F12 gene', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154548904, end_position_on_the_genomic_accession=154550050, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  12580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12580, CurrentID=0, Status='live', Symbol='Cdkn2c', Aliases='C77269, INK4c, p18, p18-INK4c, p18-INK6, p18INK4c', description='cyclin-dependent kinase inhibitor 2C (p18, inhibits CDK4)', other_designations='cyclin-dependent kinase 6 inhibitor', map_location='4 51.32 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109660876, end_position_on_the_genomic_accession=109666756, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75482, CurrentID=0, Status='live', Symbol='Hspb9', Aliases='1700007H20Rik', description='heat shock protein, alpha-crystallin-related, B9', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100713850, end_position_on_the_genomic_accession=100714575, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75483, CurrentID=0, Status='live', Symbol='Cox8c', Aliases='1700007F21Rik, COX8-3, COXVIII-3', description='cytochrome c oxidase subunit VIIIc', other_designations='COX VIII-3|cytochrome c oxidase polypeptide 8|cytochrome c oxidase polypeptide VIII|cytochrome c oxidase subunit 8-3', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102899306, end_position_on_the_genomic_accession=102900534, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75495, CurrentID=0, Status='live', Symbol='Morn5', Aliases='1700010A17Rik', description='MORN repeat containing 5', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36049473, end_position_on_the_genomic_accession=36079709, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75497, CurrentID=0, Status='live', Symbol='Fabp12', Aliases='1700008G05Rik', description='fatty acid binding protein 12', other_designations='intracellular fatty acid-binding protein FABP12|probable fatty acid-binding protein ENSP00000353650 homolog', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=10244209, end_position_on_the_genomic_accession=10301526, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  75502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75502, CurrentID=0, Status='live', Symbol='Cmtm2b', Aliases='1700013O04Rik, Cklfsf22, Cklfsf2b', description='CKLF-like MARVEL transmembrane domain containing 2B', other_designations='chemokine-like factor super family 2B|chemokine-like factor superfamily member 2B|chemokine-like factor superfamily protein 2-2', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104322205, end_position_on_the_genomic_accession=104330764, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  75504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75504, CurrentID=0, Status='live', Symbol='1700013F07Rik', Aliases='', description='RIKEN cDNA 1700013F07 gene', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108537584, end_position_on_the_genomic_accession=108544697, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75507, CurrentID=0, Status='live', Symbol='Pou5f2', Aliases='1700013G10Rik, SPRM-1, Sprm1', description='POU domain class 5, transcription factor 2', other_designations='sperm 1 POU domain transcription factor|sperm-associated POU domain protein 1', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=78024902, end_position_on_the_genomic_accession=78026296, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75509, CurrentID=0, Status='live', Symbol='1700020N15Rik', Aliases='', description='RIKEN cDNA 1700020N15 gene', other_designations='', map_location='X A7.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=69945281, end_position_on_the_genomic_accession=69945980, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75512, CurrentID=0, Status='live', Symbol='Gpx6', Aliases='1700020G18Rik, Ry2d1, olfactory', description='glutathione peroxidase 6', other_designations='GPx-6|GSHPx-6|odorant-metabolizing protein RY2D1', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21295368, end_position_on_the_genomic_accession=21319625, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  75514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75514, CurrentID=0, Status='live', Symbol='1700013H16Rik', Aliases='Slx2, Xlr6', description='RIKEN cDNA 1700013H16 gene', other_designations='SYCP3-like X-linked 2', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53742901, end_position_on_the_genomic_accession=53757970, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  75516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75516, CurrentID=0, Status='live', Symbol='Ttc32', Aliases='1700013G20Rik', description='tetratricopeptide repeat domain 32', other_designations='TPR repeat protein 32', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=9029912, end_position_on_the_genomic_accession=9045118, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75524, CurrentID=0, Status='live', Symbol='1700018C11Rik', Aliases='bM230C2.3', description='RIKEN cDNA 1700018C11 gene', other_designations='bM230C2.3 (novel protein (1700018C11Rik))', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63607085, end_position_on_the_genomic_accession=63622448, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75526, CurrentID=0, Status='live', Symbol='Eppin', Aliases='1700024E17Rik, Spinlw1', description='epididymal peptidase inhibitor', other_designations='epididymal protease inhibitor|serine protease inhibitor-like with Kunitz and WAP domains 1|serine protease inhibitor-like, with Kunitz and WAP domains 1 (eppin)', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164588343, end_position_on_the_genomic_accession=164593571, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  75530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75530, CurrentID=0, Status='live', Symbol='Lyrm7', Aliases='1700024C24Rik, 9330147L21Rik', description='LYR motif containing 7', other_designations='LYR motif-containing protein 7', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54826866, end_position_on_the_genomic_accession=54860907, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75533, CurrentID=0, Status='live', Symbol='Nme5', Aliases='1700019D05Rik, Nm23-M5', description='NME/NM23 family member 5', other_designations='NDK-H 5|NDP kinase homolog 5|expressed in non-metastatic cells 5|non-metastatic cells 5, protein expressed in (nucleoside-diphosphate kinase)', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34562636, end_position_on_the_genomic_accession=34579152, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75538, CurrentID=0, Status='live', Symbol='Fam71e1', Aliases='0610007G24Rik, 1700021N13Rik, 1700021P22Rik, GARI-L5', description='family with sequence similarity 71, member E1', other_designations='Golgi-associated Rab2B interactor-like 5', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44496584, end_position_on_the_genomic_accession=44501681, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  75540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75540, CurrentID=0, Status='live', Symbol='Fpgt', Aliases='1700016E03Rik', description='fucose-1-phosphate guanylyltransferase', other_designations='', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=155084915, end_position_on_the_genomic_accession=155093378, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75541, CurrentID=0, Status='live', Symbol='Nat8f4', Aliases='1700019G17Rik', description='N-acetyltransferase 8 (GCN5-related) family member 4', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85899050, end_position_on_the_genomic_accession=85909424, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75547, CurrentID=0, Status='live', Symbol='Akap13', Aliases='1700026G02Rik, 5730522G15Rik, 5830460E08Rik, AKAP-Lbc, BRX, Ht31, LBC, PROTO-LB, PROTO-LBC', description='A kinase (PRKA) anchor protein 13', other_designations='type II cAMP-dependent protein kinase anchoring protein Ht31', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=75455508, end_position_on_the_genomic_accession=75754609, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  75552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75552, CurrentID=0, Status='live', Symbol='Paqr9', Aliases='1700020G04Rik', description='progestin and adipoQ receptor family member IX', other_designations='', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95559817, end_position_on_the_genomic_accession=95562121, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  209200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209200, CurrentID=0, Status='live', Symbol='Dtx3l', Aliases='AU042200, BC023741', description='deltex 3-like, E3 ubiquitin ligase', other_designations='deltex 3-like', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35926515, end_position_on_the_genomic_accession=35939027, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75555, CurrentID=0, Status='live', Symbol='1700020D05Rik', Aliases='mage-g2', description='RIKEN cDNA 1700020D05 gene', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5502767, end_position_on_the_genomic_accession=5503787, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  75556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75556, CurrentID=0, Status='live', Symbol='Cfap161', Aliases='1700026D08Rik, AW047638, AW551802', description='cilia and flagella associated protein 161', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=83774101, end_position_on_the_genomic_accession=83795384, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75558, CurrentID=0, Status='live', Symbol='Spata45', Aliases='1700022P22Rik', description='spermatogenesis associated 45', other_designations='UPF0732 protein C1orf227 homolog', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191036822, end_position_on_the_genomic_accession=191042941, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75560, CurrentID=0, Status='live', Symbol='Ep400', Aliases='1700020J09Rik, AU023439, mDomino, mKIAA1498, p400', description='E1A binding protein p400', other_designations='domino homolog|p400 kDa SWI2/SNF2-related protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110664373, end_position_on_the_genomic_accession=110770749, orientation='minus', exon_count=56, OMIM='', no_hdr0=''),
  75563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75563, CurrentID=0, Status='live', Symbol='Dnali1', Aliases='1700023A09Rik, AW049135', description='dynein, axonemal, light intermediate polypeptide 1', other_designations='axonemal dynein light chain (hp28) homolog|inner dynein arm light chain, axonemal', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=125055339, end_position_on_the_genomic_accession=125065657, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75564, CurrentID=0, Status='live', Symbol='Rsph9', Aliases='1700027N10Rik', description='radial spoke head 9 homolog (Chlamydomonas)', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46122038, end_position_on_the_genomic_accession=46144237, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75565, CurrentID=0, Status='live', Symbol='Sgf29', Aliases='1700023O11Rik, 9530025I05Rik, Ccdc101', description='SAGA complex associated factor 29', other_designations='SAGA complex-associated factor 29|SAGA-associated factor 29 homolog|coiled-coil domain containing 101|coiled-coil domain-containing protein 101', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126649309, end_position_on_the_genomic_accession=126672779, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75568, CurrentID=0, Status='live', Symbol='Capsl', Aliases='1700028N11Rik, AW125205', description='calcyphosine-like', other_designations='Calcyphosine', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9436015, end_position_on_the_genomic_accession=9466039, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75570, CurrentID=0, Status='live', Symbol='Nhej1', Aliases='1700029B21Rik, XLF, cernunnos', description='nonhomologous end-joining factor 1', other_designations='XRCC4-like factor|protein cernunnos', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74967346, end_position_on_the_genomic_accession=75046653, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75571, CurrentID=0, Status='live', Symbol='Spata9', Aliases='1700030K01Rik, 4930599C08Rik, A930023H06Rik', description='spermatogenesis associated 9', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=75960003, end_position_on_the_genomic_accession=75998987, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  75572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75572, CurrentID=0, Status='live', Symbol='Acyp2', Aliases='2310004B09Rik', description='acylphosphatase 2, muscle type', other_designations='acylphosphatase, muscle type isozyme|acylphosphate phosphohydrolase 2', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30505992, end_position_on_the_genomic_accession=30649396, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  75573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75573, CurrentID=0, Status='live', Symbol='Prr29', Aliases='2310007L24Rik', description='proline rich 29', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106365479, end_position_on_the_genomic_accession=106377114, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  75577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75577, CurrentID=0, Status='live', Symbol='Dynap', Aliases='2310002L13Rik', description='dynactin associated protein', other_designations='', map_location='18 44.21 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=70240429, end_position_on_the_genomic_accession=70244584, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75578, CurrentID=0, Status='live', Symbol='Fggy', Aliases='2310009E04Rik', description='FGGY carbohydrate kinase domain containing', other_designations='', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=95557485, end_position_on_the_genomic_accession=95926939, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  75580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75580, CurrentID=0, Status='live', Symbol='Zbtb4', Aliases='2310026P19Rik, 9230111I22Rik, mKIAA1538', description='zinc finger and BTB domain containing 4', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69765912, end_position_on_the_genomic_accession=69784026, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  75581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75581, CurrentID=0, Status='live', Symbol='Yipf7', Aliases='2310016N21Rik, Yip1b', description='Yip1 domain family, member 7', other_designations='', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=69516670, end_position_on_the_genomic_accession=69542647, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75586, CurrentID=0, Status='live', Symbol='Krtap9-3', Aliases='2310040M23Rik', description='keratin associated protein 9-3', other_designations='keratin-associated protein 9.3|krtap9-3|ultrahigh sulfur keratin-associated protein 9.3', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99597348, end_position_on_the_genomic_accession=99598106, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  75590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75590, CurrentID=0, Status='live', Symbol='Dusp9', Aliases='Dusp4, Mpk4, Pyst3', description='dual specificity phosphatase 9', other_designations='dual-specificity MAP kinase phosphatase 4', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73639389, end_position_on_the_genomic_accession=73643514, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75593, CurrentID=0, Status='live', Symbol='Malsu1', Aliases='2410003K15Rik, Matsu1', description='mitochondrial assembly of ribosomal large subunit 1', other_designations='protein C7orf30 homolog', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49073795, end_position_on_the_genomic_accession=49139144, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  75596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75596, CurrentID=0, Status='live', Symbol='Prl7b1', Aliases='1600014J19Rik, Ghd17, PLP-N, Prlpn', description='prolactin family 7, subfamily b, member 1', other_designations='PRL-like protein N|growth hormone d17|placental prolactin-like protein N|prolactin-like protein N', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27601812, end_position_on_the_genomic_accession=27610582, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75597, CurrentID=0, Status='live', Symbol='Ndufaf2', Aliases='1810058I14Rik, C86051, Ndufa12l, mimitin', description='NADH dehydrogenase (ubiquinone) 1 alpha subcomplex, assembly factor 2', other_designations='MMTN|Myc-induced mitochondria protein|NDUFA12-like protein|myc-induced mitochondrial protein', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=108052589, end_position_on_the_genomic_accession=108158654, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75599, CurrentID=0, Status='live', Symbol='Pcdh1', Aliases='2010005A06Rik, AI585920', description='protocadherin 1', other_designations='', map_location='18 20.08 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=38185914, end_position_on_the_genomic_accession=38212048, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  75600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75600, CurrentID=0, Status='live', Symbol='Calml4', Aliases='2010002G05Rik', description='calmodulin-like 4', other_designations='CALM-Rel|calmodulin-related', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62857481, end_position_on_the_genomic_accession=62875921, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75604, CurrentID=0, Status='live', Symbol='Tm4sf5', Aliases='2010003F10Rik, AI323987', description='transmembrane 4 superfamily member 5', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70505274, end_position_on_the_genomic_accession=70511183, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75605, CurrentID=0, Status='live', Symbol='Kdm5b', Aliases='2010009J12Rik, 2210016I17Rik, AW556288, D1Ertd202e, Jarid1b, PLU-1, PUT1, Plu1, RBBP2H1A, Rb-Bp2, mKIAA4034', description='lysine (K)-specific demethylase 5B', other_designations='histone demethylase JARID1B|jumonji, AT rich interactive domain 1B (Rbp2 like)|jumonji/ARID domain-containing protein 1B|putative DNA/chromatin binding motif 1', map_location='1 58.24 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134560178, end_position_on_the_genomic_accession=134632883, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  75607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75607, CurrentID=0, Status='live', Symbol='Wnk2', Aliases='1810073P09Rik, AW122246, ESTM15, X83337, mKIAA1760', description='WNK lysine deficient protein kinase 2', other_designations='protein kinase lysine-deficient 2|protein kinase with no lysine 2', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49036302, end_position_on_the_genomic_accession=49148574, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  75608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75608, CurrentID=0, Status='live', Symbol='Chmp4b', Aliases='2010012F05Rik, C76846, Snf7-2', description='charged multivesicular body protein 4B', other_designations='chromatin modifying protein 4B|chromatin-modifying protein 4b', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154657026, end_position_on_the_genomic_accession=154694783, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75612, CurrentID=0, Status='live', Symbol='Gns', Aliases='2610016K11Rik, AU042285, C87209, G6S, N28088', description='glucosamine (N-acetyl)-6-sulfatase', other_designations='glucosamine-6-sulfatase', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121365067, end_position_on_the_genomic_accession=121397253, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  75613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75613, CurrentID=0, Status='live', Symbol='Med25', Aliases='2610034E13Rik, 2610529E18Rik, ESTM2', description='mediator complex subunit 25', other_designations='mediator of RNA polymerase II transcription, subunit 9 homolog', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44876765, end_position_on_the_genomic_accession=44892664, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  75616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75616, CurrentID=0, Status='live', Symbol='Smim15', Aliases='2810008M24Rik', description='small integral membrane protein 15', other_designations='UPF0542 protein C5orf43 homolog', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=108044474, end_position_on_the_genomic_accession=108049175, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75617, CurrentID=0, Status='live', Symbol='Rps25', Aliases='2810009D21Rik', description='ribosomal protein S25', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44407714, end_position_on_the_genomic_accession=44410406, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75619, CurrentID=0, Status='live', Symbol='Fastkd2', Aliases='2810421I24Rik', description='FAST kinase domains 2', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63730599, end_position_on_the_genomic_accession=63753385, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  75620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75620, CurrentID=0, Status='live', Symbol='Kxd1', Aliases='0610030B01Rik, 2810422J05Rik, C78305', description='KxDL motif containing 1', other_designations='UPF0459 protein C19orf50 homolog', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70513396, end_position_on_the_genomic_accession=70523180, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75622, CurrentID=0, Status='live', Symbol='Spaca3', Aliases='1700025M08Rik, ALLP17, Lyc3, SLLP1, mSLLP1', description='sperm acrosome associated 3', other_designations='acrosomal sperm protein|lysozyme-like protein 3|sperm c-type lysozyme-like protein|sperm lysozyme-like protein 1 precusor', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80858389, end_position_on_the_genomic_accession=80867814, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75623, CurrentID=0, Status='live', Symbol='Tex30', Aliases='1700029F09Rik, 3110030D08Rik', description='testis expressed 30', other_designations='', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44086613, end_position_on_the_genomic_accession=44102497, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  75624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75624, CurrentID=0, Status='live', Symbol='Metap1', Aliases='1700029C17Rik, AW047992, mKIAA0094', description='methionyl aminopeptidase 1', other_designations='MAP 1|Putative methionine aminopeptidase 1 (MetAP 1) (Peptidase M 1)|metAP 1|peptidase M 1', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138458960, end_position_on_the_genomic_accession=138489382, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75625, CurrentID=0, Status='live', Symbol='Mageh1', Aliases='2010107K23Rik, APR-1, APR1', description='melanoma antigen, family H, 1', other_designations='MAGE-H1 antigen', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153036166, end_position_on_the_genomic_accession=153037563, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  75627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75627, CurrentID=0, Status='live', Symbol='Snapc1', Aliases='2700033G17Rik, AU015787', description='small nuclear RNA activating complex, polypeptide 1', other_designations='SNAPc 43 kDa subunit|SNAPc subunit 1|small nuclear RNA-activating complex polypeptide 1|snRNA-activating protein complex 43 kDa subunit', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73964477, end_position_on_the_genomic_accession=73985925, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  75632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75632, CurrentID=0, Status='live', Symbol='1700003O11Rik', Aliases='', description='RIKEN cDNA 1700003O11 gene', other_designations='spermatocyte-specific protein', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=76014210, end_position_on_the_genomic_accession=76016974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  75641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75641, CurrentID=0, Status='live', Symbol='1700029I15Rik', Aliases='1700021K07Rik', description='RIKEN cDNA 1700029I15 gene', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92382918, end_position_on_the_genomic_accession=92383603, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75642, CurrentID=0, Status='live', Symbol='Spata25', Aliases='1700020C07Rik, Tsg23', description='spermatogenesis associated 25', other_designations='testis-specific gene 23 protein', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164827382, end_position_on_the_genomic_accession=164828534, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  75645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75645, CurrentID=0, Status='live', Symbol='Ccdc172', Aliases='1700011F14Rik', description='coiled-coil domain containing 172', other_designations='UPF0628 protein C10orf96 homolog', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58511999, end_position_on_the_genomic_accession=58553104, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  75646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75646, CurrentID=0, Status='live', Symbol='Rai14', Aliases='1700008J19Rik, 1700020L11Rik, Ankycorbin, Norpeg, mKIAA1334', description='retinoic acid induced 14', other_designations='NORPEG-like protein|ankyrin repeat and coiled-coil structure-containing protein|novel retinal pigment epithelial cell protein|p125|retinoic acid-induced protein 14', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10568971, end_position_on_the_genomic_accession=10719267, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  75647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75647, CurrentID=0, Status='live', Symbol='Ssmem1', Aliases='1700016K02Rik, 1700025E21Rik', description='serine-rich single-pass membrane protein 1', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30509849, end_position_on_the_genomic_accession=30520253, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  229474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229474, CurrentID=0, Status='live', Symbol='Fhdc1', Aliases='6330505N24Rik, Gm126', description='FH2 domain containing 1', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84442191, end_position_on_the_genomic_accession=84480439, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  75656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75656, CurrentID=0, Status='live', Symbol='1700020A23Rik', Aliases='', description='RIKEN cDNA 1700020A23 gene', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130405259, end_position_on_the_genomic_accession=130406074, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75657, CurrentID=0, Status='live', Symbol='Speer4a', Aliases='1700027N01Rik, SPEER-4A', description='spermatogenesis associated glutamate (E)-rich protein 4A', other_designations='sperm-associated glutamate (E)-rich protein 4a|spermatogenic-associated glutamate (E)-rich protein 4a', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=26017738, end_position_on_the_genomic_accession=26042295, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75659, CurrentID=0, Status='live', Symbol='Wdr54', Aliases='1700030E05Rik', description='WD repeat domain 54', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83152710, end_position_on_the_genomic_accession=83156406, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75660, CurrentID=0, Status='live', Symbol='Lin37', Aliases='1810054G18Rik', description='lin-37 homolog (C. elegans)', other_designations='antolefinin|antolefinine', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30555441, end_position_on_the_genomic_accession=30559719, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  75665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75665, CurrentID=0, Status='live', Symbol='Ccdc64', Aliases='2210403N09Rik, BICDR-1', description='coiled-coil domain containing 64', other_designations='BICD-related protein 1|bicaudal-D-related protein 1|coiled-coil domain-containing protein 64A', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115648174, end_position_on_the_genomic_accession=115731643, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  75667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75667, CurrentID=0, Status='live', Symbol='Tmem225', Aliases='1700030E15Rik, Tnen225', description='transmembrane protein 225', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40148100, end_position_on_the_genomic_accession=40150901, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  75668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75668, CurrentID=0, Status='live', Symbol='Rasl10a', Aliases='2210403B10Rik, AI852688', description='RAS-like, family 10, member A', other_designations='Ras-like protein RRP22 (RAS-related protein on chromosome 22)|ras-related protein on chromosome 22 homolog', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5057981, end_position_on_the_genomic_accession=5060383, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75669, CurrentID=0, Status='live', Symbol='Pik3r4', Aliases='2210010O15Rik, C730038E05Rik, C85833, D9Ertd418e', description='phosphatidylinositol 3 kinase, regulatory subunit, polypeptide 4, p150', other_designations='PI3-kinase regulatory subunit 4', map_location='9 56.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105642995, end_position_on_the_genomic_accession=105687655, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  75671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75671, CurrentID=0, Status='live', Symbol='Tex22', Aliases='1700028O09Rik, AV210422, Tep22', description='testis expressed gene 22', other_designations='testis-expressed protein of 22 kDa', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113074469, end_position_on_the_genomic_accession=113088917, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  75677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75677, CurrentID=0, Status='live', Symbol='Cldn22', Aliases='2210404A22Rik, 5330431C14Rik, 9530051B05Rik', description='claudin 22', other_designations='', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=47824482, end_position_on_the_genomic_accession=47825475, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75678, CurrentID=0, Status='live', Symbol='Ippk', Aliases='1810043M15Rik, InsP6', description='inositol 1,3,4,5,6-pentakisphosphate 2-kinase', other_designations='ins(1,3,4,5,6)P5 2-kinase|insP5 2-kinase', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49421223, end_position_on_the_genomic_accession=49463060, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  75686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75686, CurrentID=0, Status='live', Symbol='Nudt16', Aliases='2310041H06Rik, 2810047L04Rik, 2900006H04Rik, AI851783', description='nudix (nucleoside diphosphate linked moiety X)-type motif 16', other_designations='IDP phosphatase|IDPase|inosine diphosphate phosphatase|m7GpppN-mRNA hydrolase|nucleoside diphosphate-linked moiety X motif 16|nudix motif 16', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105129338, end_position_on_the_genomic_accession=105131841, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75687, CurrentID=0, Status='live', Symbol='Fam65a', Aliases='2310066E14Rik, AI837433', description='family with sequence similarity 65, member A', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105594087, end_position_on_the_genomic_accession=105622219, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  75689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75689, CurrentID=0, Status='live', Symbol='Higd1b', Aliases='2310056K19Rik', description='HIG1 domain family, member 1B', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102835756, end_position_on_the_genomic_accession=102844133, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75690, CurrentID=0, Status='live', Symbol='Vsig10l', Aliases='', description='V-set and immunoglobulin domain containing 10 like', other_designations='ZV-set and immunoglobulin domain containing 10 like', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43463061, end_position_on_the_genomic_accession=43472023, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75691, CurrentID=0, Status='live', Symbol='Anks6', Aliases='2210417J20Rik, Ankrd14, Gm635, SamCystin, Samd6, b2b1801.1Clo', description='ankyrin repeat and sterile alpha motif domain containing 6', other_designations='SAM domain-containing protein 6|ankyrin repeat domain 14|sterile alpha motif domain containing 6|sterile alpha motif domain-containing protein 6', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=47015365, end_position_on_the_genomic_accession=47057335, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  75692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75692, CurrentID=0, Status='live', Symbol='Nr2c2ap', Aliases='2310073E15Rik', description='nuclear receptor 2C2-associated protein', other_designations='TR4 orphan receptor-associated 16 kDa protein', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70131327, end_position_on_the_genomic_accession=70133752, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75695, CurrentID=0, Status='live', Symbol='Rilpl1', Aliases='2900002H16Rik, 6330559I19Rik, GOSPEL', description='Rab interacting lysosomal protein-like 1', other_designations='rab-interacting lysosomal-like protein 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124493080, end_position_on_the_genomic_accession=124531391, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  75697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75697, CurrentID=0, Status='live', Symbol='C2cd4b', Aliases='3300001A09Rik, Fam148b', description='C2 calcium-dependent domain containing 4B', other_designations='family with sequence similarity 148, member B', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=67716224, end_position_on_the_genomic_accession=67760933, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  75698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75698, CurrentID=0, Status='live', Symbol='Fam35a', Aliases='3110001K24Rik', description='family with sequence similarity 35, member A', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34236948, end_position_on_the_genomic_accession=34310503, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  624286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624286, CurrentID=0, Status='live', Symbol='Cfhr3', Aliases='EG624286', description='complement factor H-related 3', other_designations='', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=139575184, end_position_on_the_genomic_accession=139660986, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  75704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75704, CurrentID=0, Status='live', Symbol='Ces1h', Aliases='2310039D24Rik', description='carboxylesterase 1H', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93351842, end_position_on_the_genomic_accession=93386990, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  75705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75705, CurrentID=0, Status='live', Symbol='Eif4b', Aliases='2310046H11Rik, AL024095, C85189, Eif4a2', description='eukaryotic translation initiation factor 4B', other_designations='eIF-4B', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102073773, end_position_on_the_genomic_accession=102097173, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  75706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75706, CurrentID=0, Status='live', Symbol='Krt24', Aliases='2310058N18Rik, 2310075C18Rik', description='keratin 24', other_designations='CK-24|K24|cytokeratin-24|keratin-24|ortholog of keratin 24 KRT24|type I keratin-24', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99280093, end_position_on_the_genomic_accession=99285238, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75710, CurrentID=0, Status='live', Symbol='Rbm12', Aliases='5730420G12Rik, 9430070C08Rik, AI852903, SWAN, mKIAA0765', description='RNA binding motif protein 12', other_designations='RNA-binding motif protein 12|SH3/WW domain anchor protein in the nucleus', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156094882, end_position_on_the_genomic_accession=156111965, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75712, CurrentID=0, Status='live', Symbol='Tmem14a', Aliases='5730496E24Rik, C85783, PTD011', description='transmembrane protein 14A', other_designations='', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=21166436, end_position_on_the_genomic_accession=21230167, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75717, CurrentID=0, Status='live', Symbol='Cul5', Aliases='4921514I20Rik, 8430423K24Rik, AI852817, C030032G03Rik, C330021I08Rik, VACM-1, VACM1', description='cullin 5', other_designations='CUL-5|vasopressin-activated calcium-mobilizing receptor protein', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53614582, end_position_on_the_genomic_accession=53667507, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  75718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75718, CurrentID=0, Status='live', Symbol='Vwa5b1', Aliases='4931403E03Rik', description='von Willebrand factor A domain containing 5B1', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138564229, end_position_on_the_genomic_accession=138635937, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  75721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75721, CurrentID=0, Status='live', Symbol='4932414N04Rik', Aliases='', description='RIKEN cDNA 4932414N04 gene', other_designations='', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=68656791, end_position_on_the_genomic_accession=68748526, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  75723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75723, CurrentID=0, Status='live', Symbol='Amotl1', Aliases='2310010G08Rik, 2310067L22Rik, 4932416D09Rik, JEAP, mFLJ00155', description='angiomotin-like 1', other_designations='junction-enriched and -associated protein', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14541966, end_position_on_the_genomic_accession=14662960, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  75725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75725, CurrentID=0, Status='live', Symbol='Phf14', Aliases='1110001C23Rik, 4932409F11Rik, 5730446A07Rik, AA623952, AV297001, mKIAA0783', description='PHD finger protein 14', other_designations='', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=11907599, end_position_on_the_genomic_accession=12099859, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  75729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75729, CurrentID=0, Status='live', Symbol='Fam227a', Aliases='4933432B09Rik', description='family with sequence similarity 227, member A', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79609576, end_position_on_the_genomic_accession=79658970, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  75731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75731, CurrentID=0, Status='live', Symbol='Idnk', Aliases='5133401N09Rik', description='idnK gluconokinase homolog (E. coli)', other_designations='gluconate kinase', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58157649, end_position_on_the_genomic_accession=58164693, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  75732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75732, CurrentID=0, Status='live', Symbol='Iqcd', Aliases='4933433C09Rik, AI662646', description='IQ motif containing D', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120586060, end_position_on_the_genomic_accession=120607243, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75734, CurrentID=0, Status='live', Symbol='Mff', Aliases='5230400G24Rik, AI314724', description='mitochondrial fission factor', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=82724895, end_position_on_the_genomic_accession=82752390, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  75735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75735, CurrentID=0, Status='live', Symbol='Pank1', Aliases='Pankb, Pank1', description='pantothenate kinase 1', other_designations='pantothenic acid kinase 1', map_location='19 G1-G3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34806935, end_position_on_the_genomic_accession=34882233, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75736, CurrentID=0, Status='live', Symbol='Bcl2l12', Aliases='2810475P17Rik, 5430429M05Rik, Bcl-L12, Bcl2-L12', description='BCL2-like 12 (proline rich)', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44991222, end_position_on_the_genomic_accession=44997643, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75739, CurrentID=0, Status='live', Symbol='Mpp7', Aliases='1110068J02Rik, 2810038M04Rik, 5430426E14Rik, AI415104, Gm955', description='membrane protein, palmitoylated 7 (MAGUK p55 subfamily member 7)', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=7347962, end_position_on_the_genomic_accession=7626894, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  75740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75740, CurrentID=0, Status='live', Symbol='Egfem1', Aliases='6130401L20Rik, Efcab11', description='EGF-like and EMI domain containing 1', other_designations='EF-hand calcium binding domain 11', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=29080745, end_position_on_the_genomic_accession=29691058, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  75744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75744, CurrentID=0, Status='live', Symbol='Svip', Aliases='1700006C13Rik, 6620401M08Rik', description='small VCP/p97-interacting protein', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=51997161, end_position_on_the_genomic_accession=52006018, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75745, CurrentID=0, Status='live', Symbol='Rian', Aliases='5530401N18Rik, C130089L09Rik, Meg8', description='RNA imprinted and accumulated in nucleus', other_designations='', map_location='12 60.41 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109603945, end_position_on_the_genomic_accession=109661716, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  75746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75746, CurrentID=0, Status='live', Symbol='Morc4', Aliases='1600017G11Rik, 5630401M14Rik, AA117449, Zcwcc2', description='microrchidia 4', other_designations='zinc finger CW-type coiled-coil domain protein 2|zinc finger, CW-type with coiled-coil domain 2', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139821635, end_position_on_the_genomic_accession=139871692, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  75747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75747, CurrentID=0, Status='live', Symbol='Sesn3', Aliases='5630400E15Rik, BC003348, SEST3', description='sestrin 3', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14276301, end_position_on_the_genomic_accession=14326134, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75750, CurrentID=0, Status='live', Symbol='Slc10a6', Aliases='8430417G17Rik, C78479, Soat', description='solute carrier family 10 (sodium/bile acid cotransporter family), member 6', other_designations='sodium-dependent organic anion transporter', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=103605707, end_position_on_the_genomic_accession=103629403, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75751, CurrentID=0, Status='live', Symbol='Ipo4', Aliases='8430408O15Rik, AA409693, Imp4a, RanBP4', description='importin 4', other_designations='imp4|importin 4a|importin-4a|ran-binding protein 4', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55625629, end_position_on_the_genomic_accession=55635952, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  75753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75753, CurrentID=0, Status='live', Symbol='Klf17', Aliases='7420700M05Rik, AA420409, AU043468, C85123, D4Ertd561e, Gzf, Zfp393', description='Kruppel-like factor 17', other_designations='germ cell specific zinc finger protein|germ cell-specific zinc finger protein|zinc finger protein 393', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117756659, end_position_on_the_genomic_accession=117765666, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  75758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75758, CurrentID=0, Status='live', Symbol='9130401M01Rik', Aliases='AI849328', description='RIKEN cDNA 9130401M01 gene', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58022270, end_position_on_the_genomic_accession=58034294, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75761, CurrentID=0, Status='live', Symbol='Apol7a', Aliases='9130022K13Rik, Apol3', description='apolipoprotein L 7a', other_designations='apolipoprotein L 3', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77388217, end_position_on_the_genomic_accession=77399110, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75763, CurrentID=0, Status='live', Symbol='Dcaf17', Aliases='2810055O12Rik, 4833418A01Rik, A030004A10Rik, A930009G19Rik, AI448937', description='DDB1 and CUL4 associated factor 17', other_designations='', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71055742, end_position_on_the_genomic_accession=71114514, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  75764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75764, CurrentID=0, Status='live', Symbol='Slx1b', Aliases='1110030E23Rik, 2410170E21Rik, 4833422P03Rik, AI853643, Giyd1, Giyd2', description='SLX1 structure-specific endonuclease subunit homolog B (S. cerevisiae)', other_designations='GIY-YIG domain containing 2|GIY-YIG domain-containing protein 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126688927, end_position_on_the_genomic_accession=126695783, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75766, CurrentID=0, Status='live', Symbol='Dcstamp', Aliases='4833414I07Rik, DC-STAMP, FIND, Tm7sf4, mDC-STAMP', description='dentrocyte expressed seven transmembrane protein', other_designations='dendrocyte-expressed seven transmembrane protein|transmembrane 7 superfamily member 4', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39745930, end_position_on_the_genomic_accession=39760938, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  75767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75767, CurrentID=0, Status='live', Symbol='Rab11fip1', Aliases='2010200K21Rik, 4833414G05Rik, Rcp', description='RAB11 family interacting protein 1 (class I)', other_designations='Rab coupling protein|rab-coupling protein|rab11-FIP1', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27138773, end_position_on_the_genomic_accession=27174646, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75769, CurrentID=0, Status='live', Symbol='Plppr5', Aliases='4833424O15Rik, Lppr5, PRG-5, Pap2d', description='phospholipid phosphatase related 5', other_designations='phosphatidic acid phosphatase type 2d|plasticity-related gene 5 protein', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=117573063, end_position_on_the_genomic_accession=117689507, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  75770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75770, CurrentID=0, Status='live', Symbol='Brsk2', Aliases='4833424K13Rik, SAD-A, SADA', description='BR serine/threonine kinase 2', other_designations='BR serine/threonine-protein kinase 2|brain-specific serine/threonine-protein kinase 2|serine/threonine-protein kinase SAD-A', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141949751, end_position_on_the_genomic_accession=142004244, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  75772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75772, CurrentID=0, Status='live', Symbol='Pnpla5', Aliases='4833426H19Rik, GS2L', description='patatin-like phospholipase domain containing 5', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84112356, end_position_on_the_genomic_accession=84127992, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  75773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75773, CurrentID=0, Status='live', Symbol='Adad2', Aliases='4930403J07Rik', description='adenosine deaminase domain containing 2', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119612710, end_position_on_the_genomic_accession=119616926, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75775, CurrentID=0, Status='live', Symbol='4930402K13Rik', Aliases='', description='RIKEN cDNA 4930402K13 gene', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9104562, end_position_on_the_genomic_accession=9106342, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75777, CurrentID=0, Status='live', Symbol='Ttc23l', Aliases='4930401A09Rik', description='tetratricopeptide repeat domain 23-like', other_designations='', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10499155, end_position_on_the_genomic_accession=10559261, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  75778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75778, CurrentID=0, Status='live', Symbol='Them4', Aliases='2700077M13Rik, 4921507I02Rik', description='thioesterase superfamily member 4', other_designations='acyl-CoA thioesterase THEM4|carboxyl-terminal modulator protein', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94310106, end_position_on_the_genomic_accession=94332540, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  337924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=337924, CurrentID=0, Status='live', Symbol='Cyp3a44', Aliases='', description='cytochrome P450, family 3, subfamily a, polypeptide 44', other_designations='cytochrome P450, 3a44', map_location='5 85.17 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145773983, end_position_on_the_genomic_accession=145805891, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  75782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75782, CurrentID=0, Status='live', Symbol='Lca5', Aliases='4930431B11Rik, 5730406O13Rik, AV274874, ORF64', description='Leber congenital amaurosis 5 (human)', other_designations='leber congenital amaurosis 5 protein homolog', map_location='9 E2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=83391355, end_position_on_the_genomic_accession=83441733, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75784, CurrentID=0, Status='live', Symbol='1700007G11Rik', Aliases='4930428O21Rik', description='RIKEN cDNA 1700007G11 gene', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=98329045, end_position_on_the_genomic_accession=98802053, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  75785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75785, CurrentID=0, Status='live', Symbol='Klhl24', Aliases='1110046J11Rik, 4930429H24Rik, 6530402O20Rik, C85082', description='kelch-like 24', other_designations='KRIP6|kainate receptor-interacting protein for GluR6', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20097457, end_position_on_the_genomic_accession=20129226, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  75786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75786, CurrentID=0, Status='live', Symbol='Ckap5', Aliases='3110043H24Rik, 4930432B04Rik, D730027C18Rik, T25636, mKIAA0097', description='cytoskeleton associated protein 5', other_designations='ckap5', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91526747, end_position_on_the_genomic_accession=91620679, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  75788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75788, CurrentID=0, Status='live', Symbol='Smurf1', Aliases='4930431E10Rik, mKIAA1625', description='SMAD specific E3 ubiquitin protein ligase 1', other_designations='SMAD-specific E3 ubiquitin-protein ligase 1|Smad ubiquitination regulatory factor 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144876495, end_position_on_the_genomic_accession=144965830, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12632, CurrentID=0, Status='live', Symbol='Cfl2', Aliases='', description='cofilin 2, muscle', other_designations='cofilin, muscle isoform', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=54858815, end_position_on_the_genomic_accession=54862877, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75796, CurrentID=0, Status='live', Symbol='Cdyl2', Aliases='1700029M19Rik, 4930453I21Rik, Gm20194', description='chromodomain protein, Y chromosome-like 2', other_designations='CDY-like 2', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116568724, end_position_on_the_genomic_accession=116732991, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  75799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75799, CurrentID=0, Status='live', Symbol='4930444P10Rik', Aliases='', description='RIKEN cDNA 4930444P10 gene', other_designations='', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16065974, end_position_on_the_genomic_accession=16093325, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75801, CurrentID=0, Status='live', Symbol='4930447C04Rik', Aliases='4921504I02Rik, A930035O15Rik, Six6OS, Six6as, Six6os1', description='RIKEN cDNA 4930447C04 gene', other_designations='Six6 antisense|Six6 opposite strand transcript 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72880983, end_position_on_the_genomic_accession=72940083, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  75805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75805, CurrentID=0, Status='live', Symbol='Nln', Aliases='4930472G13Rik, C79345', description='neurolysin (metallopeptidase M3 family)', other_designations='MEP|microsomal endopeptidase|mitochondrial oligopeptidase M|neurotensin endopeptidase', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104023050, end_position_on_the_genomic_accession=104109660, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  75811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75811, CurrentID=0, Status='live', Symbol='Saxo1', Aliases='4930500O09Rik, Fam154a', description='stablizer of axonemal microtubules 1', other_designations='family with sequence similarity 154, member A|likely ortholog of H. sapiens chromosome 9 open reading frame 138 (C9orf138)', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86444698, end_position_on_the_genomic_accession=86558303, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  75812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75812, CurrentID=0, Status='live', Symbol='Tasp1', Aliases='4930485D02Rik, AW986064', description='taspase, threonine aspartase 1', other_designations='taspase-1', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=139833479, end_position_on_the_genomic_accession=140067666, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  75820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75820, CurrentID=0, Status='live', Symbol='Wdr64', Aliases='4930415O10Rik, 4930511H01Rik', description='WD repeat domain 64', other_designations='', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175698569, end_position_on_the_genomic_accession=175815734, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  75823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75823, CurrentID=0, Status='live', Symbol='Fam227b', Aliases='4930525F21Rik', description='family with sequence similarity 227, member B', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125983483, end_position_on_the_genomic_accession=126152087, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  75826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75826, CurrentID=0, Status='live', Symbol='Senp2', Aliases='2310007L05Rik, 4930538C18Rik, AI646780, AW554757, Smt3ip2, SuPr-1, mKIAA1331', description='SUMO/sentrin specific peptidase 2', other_designations='SUMO-1 protease 1|SUMO-1 protease-1|SUMO-1/Smt3-specific isopeptidase 2|SUMO/sentrin specific protease 2|axam2|sentrin/SUMO-specific protease SENP2', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22009432, end_position_on_the_genomic_accession=22049269, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  75828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75828, CurrentID=0, Status='live', Symbol='Hormad2', Aliases='4930529M09Rik', description='HORMA domain containing 2', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4345801, end_position_on_the_genomic_accession=4442459, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  75829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75829, CurrentID=0, Status='live', Symbol='Prame', Aliases='4930534P07Rik', description='preferentially expressed antigen in melanoma', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135612957, end_position_on_the_genomic_accession=135627705, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  75835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75835, CurrentID=0, Status='live', Symbol='Vmn2r-ps54', Aliases='1700108E19Rik, 4930562N12Rik, Casr-rs5, ENSMUSG00000078527, Gm10948, Gprc2a-rs5', description='vomeronasal 2, receptor, pseudogene 54', other_designations='G protein-coupled receptor, family C, group 2, member A, related sequence 5|calcium sensing receptor, related sequence 5|vomeronasal 2, receptor 57 pseudogene', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=41668014, end_position_on_the_genomic_accession=41727193, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  75841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75841, CurrentID=0, Status='live', Symbol='Rnf139', Aliases='4930555P18Rik', description='ring finger protein 139', other_designations='rnf139|translocation in renal carcinoma on chromosome 8 protein', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58889229, end_position_on_the_genomic_accession=58902390, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75847, CurrentID=0, Status='live', Symbol='Ispd', Aliases='4930579E17Rik, AV040780', description='isoprenoid synthase domain containing', other_designations='2-C-methyl-D-erythritol 4-phosphate cytidylyltransferase-like protein|notch1-induced protein', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=36380585, end_position_on_the_genomic_accession=36689503, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  75859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75859, CurrentID=0, Status='live', Symbol='4930568D16Rik', Aliases='Ggta1l1', description='RIKEN cDNA 4930568D16 gene', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35354218, end_position_on_the_genomic_accession=35367729, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  75860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75860, CurrentID=0, Status='live', Symbol='Tex26', Aliases='4930588N13Rik', description='testis expressed 26', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149439517, end_position_on_the_genomic_accession=149470620, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  75861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75861, CurrentID=0, Status='live', Symbol='4930571K23Rik', Aliases='AI428146', description='RIKEN cDNA 4930571K23 gene', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125368861, end_position_on_the_genomic_accession=125371013, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75863, CurrentID=0, Status='live', Symbol='Clec4g', Aliases='4930572L20Rik', description='C-type lectin domain family 4, member g', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3707069, end_position_on_the_genomic_accession=3721183, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  75869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75869, CurrentID=0, Status='live', Symbol='Arl5b', Aliases='4930587A11Rik, AW146430, Arl8', description='ADP-ribosylation factor-like 5B', other_designations='ADP-ribosylation factor-like 8|ADP-ribosylation factor-like protein 8|ADP-ribosylation-like factor 8', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=15055362, end_position_on_the_genomic_accession=15079191, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  75870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75870, CurrentID=0, Status='live', Symbol='Tcam1', Aliases='4930570F09Rik, AI429105', description='testicular cell adhesion molecule 1', other_designations='', map_location='11 68.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106273079, end_position_on_the_genomic_accession=106288158, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  75871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75871, CurrentID=0, Status='live', Symbol='Zfp821', Aliases='4930566A11Rik, Znf821', description='zinc finger protein 821', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109705492, end_position_on_the_genomic_accession=109724932, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75886, CurrentID=0, Status='live', Symbol='Gstt4', Aliases='4930583C14Rik, AI507011', description='glutathione S-transferase, theta 4', other_designations='GST class-theta-4', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75814943, end_position_on_the_genomic_accession=75822563, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  75894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75894, CurrentID=0, Status='live', Symbol='Adal', Aliases='4930578F03Rik', description='adenosine deaminase-like', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121140428, end_position_on_the_genomic_accession=121156680, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  75901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75901, CurrentID=0, Status='live', Symbol='Dcp1a', Aliases='1110066A22Rik, 4930568L04Rik, AU019772, D14Ertd817e, Mitc1, SMIF', description='decapping mRNA 1A', other_designations='DCP1 decapping enzyme homolog A|MAD homolog 4 interacting transcription coactivator 1|MAD homolog 4-interacting transcription coactivator 1|Smad4-interacting transcriptional co-activator|decapping enzyme|transcription factor SMIF', map_location='14 18.75 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30479522, end_position_on_the_genomic_accession=30527059, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  75905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75905, CurrentID=0, Status='live', Symbol='4930578C19Rik', Aliases='', description='RIKEN cDNA 4930578C19 gene', other_designations='UPF0672 protein CXorf36 homolog', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=18414167, end_position_on_the_genomic_accession=18461421, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  75906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75906, CurrentID=0, Status='live', Symbol='Fam184a', Aliases='3110012E06Rik, 4930438C08Rik, 4930589M24Rik', description='family with sequence similarity 184, member A', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=53633128, end_position_on_the_genomic_accession=53751231, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  75909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75909, CurrentID=0, Status='live', Symbol='Vmp1', Aliases='3110098I04Rik, 4930579A11Rik, AI787464, Tango5, Tmem49, mir-21a, ni-2', description='vacuole membrane protein 1', other_designations='NF-E2 induceble protein-2|NF-E2-inducible protein 2|protein ni-2|transmembrane protein 49', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86583852, end_position_on_the_genomic_accession=86683822, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  75914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75914, CurrentID=0, Status='live', Symbol='Exoc6b', Aliases='4930569O18Rik, G430127E12Rik, Sec15b, Sec15l2, mKIAA0919', description='exocyst complex component 6B', other_designations='SEC15-like 2|SEC15-like protein 2|exocyst complex component Sec15B', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=84618486, end_position_on_the_genomic_accession=85070259, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  75939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75939, CurrentID=0, Status='live', Symbol='4930579G24Rik', Aliases='', description='RIKEN cDNA 4930579G24 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79629079, end_position_on_the_genomic_accession=79632820, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  75953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75953, CurrentID=0, Status='live', Symbol='Samd7', Aliases='4930597A01Rik', description='sterile alpha motif domain containing 7', other_designations='SAM domain-containing protein 7', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30746293, end_position_on_the_genomic_accession=30767174, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  75956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75956, CurrentID=0, Status='live', Symbol='Srrm2', Aliases='5033413A03Rik, AA410130, SRm300, mKIAA0324', description='serine/arginine repetitive matrix 2', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23790857, end_position_on_the_genomic_accession=23824741, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  75957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75957, CurrentID=0, Status='live', Symbol='Mir17hg', Aliases='5033413D16Rik, Mirhg1, OTTMUSG00000033261, mir-17-92', description='Mir17 host gene (non-protein coding)', other_designations='Mir17 host gene 1 (non-protein coding)|microRNA host gene 1 (non-protein coding)', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115044390, end_position_on_the_genomic_accession=115046728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  75964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75964, CurrentID=0, Status='live', Symbol='Trappc8', Aliases='5033403J15Rik, AI845423, D030074E01Rik, Trs85, mKIAA1012', description='trafficking protein particle complex 8', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20817223, end_position_on_the_genomic_accession=20896386, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  75965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75965, CurrentID=0, Status='live', Symbol='Zdhhc20', Aliases='5033406L14Rik, AI448102, B230110O18Rik, HK11', description='zinc finger, DHHC domain containing 20', other_designations='DHHC-20|DHHC-containing protein 20|zinc finger DHHC domain-containing protein 20', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57832702, end_position_on_the_genomic_accession=57890834, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  75973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75973, CurrentID=0, Status='live', Symbol='Ccdc162', Aliases='5033413D22Rikp, Gm61, Ccdc162', description='coiled-coil domain containing 162', other_designations='coiled-coil domain-containing protein 162', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41536409, end_position_on_the_genomic_accession=41716568, orientation='minus', exon_count=55, OMIM='', no_hdr0=''),
  75974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75974, CurrentID=0, Status='live', Symbol='Dock11', Aliases='5033414A21Rik, 8030476J24, Zizimin2, sph238', description='dedicator of cytokinesis 11', other_designations='ACG|activated Cdc42-associated guanine nucleotide exchange factor|spleen specific 238 kDa protein with PH domain|zizimin-2', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=35888627, end_position_on_the_genomic_accession=36076570, orientation='plus', exon_count=55, OMIM='', no_hdr0=''),
  75985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75985, CurrentID=0, Status='live', Symbol='Rab30', Aliases='5033421K01Rik, AI323892, Rsb30', description='RAB30, member RAS oncogene family', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=92741714, end_position_on_the_genomic_accession=92844535, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  75986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75986, CurrentID=0, Status='live', Symbol='Agmat', Aliases='5033405N08Rik', description='agmatine ureohydrolase (agmatinase)', other_designations='AUH', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141746675, end_position_on_the_genomic_accession=141759263, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  75991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=75991, CurrentID=0, Status='live', Symbol='Slain2', Aliases='5033405K12Rik, 8030444K12Rik, AI596370', description='SLAIN motif family, member 2', other_designations='', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72914362, end_position_on_the_genomic_accession=72978841, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76002, CurrentID=0, Status='live', Symbol='Ms4a18', Aliases='5033428B15Rik', description='membrane-spanning 4-domains, subfamily A, member 18', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10997024, end_position_on_the_genomic_accession=11018026, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  76007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76007, CurrentID=0, Status='live', Symbol='Zmym2', Aliases='5830413P05Rik, FIM, MYM, RAMP, SCLL, Zfp198, Znf198', description='zinc finger, MYM-type 2', other_designations='zinc finger protein 198', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56886420, end_position_on_the_genomic_accession=56962579, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  76014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76014, CurrentID=0, Status='live', Symbol='Zc3h18', Aliases='1190001B23Rik, 5830416A07Rik, Nhn1', description='zinc finger CCCH-type containing 18', other_designations='conserved nuclear protein Nhn1|nuclear protein NHN1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122376614, end_position_on_the_genomic_accession=122417360, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  76022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76022, CurrentID=0, Status='live', Symbol='Gon4l', Aliases='1500041I23Rik, 2610100B20Rik, 3632445O20Rik, 5430404N14Rik, AA387005', description='gon-4-like (C.elegans)', other_designations='GON-4 homolog', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88830168, end_position_on_the_genomic_accession=88910099, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  76025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76025, CurrentID=0, Status='live', Symbol='Cant1', Aliases='5830420C20Rik, Apy1h, D11Bwg0554e, Entpd8, SCAN-1, Shapy', description='calcium activated nucleotidase 1', other_designations='Ca2+-dependent endoplasmic reticulum nucleoside diphosphatase|apyrase 1, homolog (C. lectularius)|apyrase homolog', map_location='11 83.15 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118406289, end_position_on_the_genomic_accession=118419134, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  76041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76041, CurrentID=0, Status='live', Symbol='Ccdc125', Aliases='5830436D01Rik, AL023012', description='coiled-coil domain containing 125', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100668934, end_position_on_the_genomic_accession=100697240, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76044, CurrentID=0, Status='live', Symbol='Ncapg2', Aliases='5830426I05Rik, Luzp5, Mtb, mCAP-G2', description='non-SMC condensin II complex, subunit G2', other_designations='CAP-G2|chromosome-associated protein G2|leucine zipper protein 5|more than blood protein', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=116405357, end_position_on_the_genomic_accession=116463532, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  76051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76051, CurrentID=0, Status='live', Symbol='Ganc', Aliases='5830445O15Rik, 9330160A12, mFLJ00088', description='glucosidase, alpha; neutral C', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120403896, end_position_on_the_genomic_accession=120460866, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  76055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76055, CurrentID=0, Status='live', Symbol='Mgea5', Aliases='2810009A20Rik, 4833427O07Rik, 5830447M11Rik, AA408215, Hy5, Ncoat, OGA, mKIAA0679', description='meningioma expressed antigen 5 (hyaluronidase)', other_designations='N-acetyl-beta-D-glucosaminidase|N-acetyl-beta-glucosaminidase|beta-N-acetylhexosaminidase|beta-hexosaminidase|bifunctional protein NCOAT|meningioma-expressed antigen 5|nuclear cytoplasmic O-GlcNAcase and acetyltransferase', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45750255, end_position_on_the_genomic_accession=45783832, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  76071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76071, CurrentID=0, Status='live', Symbol='Jakmip1', Aliases='5830437M04Rik, C330021K24Rik, Gababrbp, Marlin-1', description='janus kinase and microtubule interacting protein 1', other_designations='GABA-B receptor-binding protein|gamma-aminobutyric acid (GABA-B) receptor binding protein|multiple alpha-helices and RNA-linker protein 1', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37028282, end_position_on_the_genomic_accession=37150614, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  76072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76072, CurrentID=0, Status='live', Symbol='Rnf183', Aliases='5830442J12Rik', description='ring finger protein 183', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62427536, end_position_on_the_genomic_accession=62434726, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76073, CurrentID=0, Status='live', Symbol='Pcgf5', Aliases='0610009F02Rik, 1110054A01Rik, 5830406C17Rik, 5830443C21Rik, 9530023M17Rik, AI324127', description='polycomb group ring finger 5', other_designations='RING finger protein 159', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36348084, end_position_on_the_genomic_accession=36460970, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  207151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207151, CurrentID=0, Status='live', Symbol='Slc22a19', Aliases='D630043A20Rik, OAT-5, Oat5, Slc22a9', description='solute carrier family 22 (organic anion transporter), member 19', other_designations='solute carrier family 22 (organic cation transporter), member 19', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7672654, end_position_on_the_genomic_accession=7711310, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76080, CurrentID=0, Status='live', Symbol='Ttpal', Aliases='3110080A02Rik, 5830472M02Rik', description='tocopherol (alpha) transfer protein-like', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163602281, end_position_on_the_genomic_accession=163619013, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76088, CurrentID=0, Status='live', Symbol='Dock8', Aliases='1200017A24Rik, 5830472H07Rik, A130095G14Rik, AI461977', description='dedicator of cytokinesis 8', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24999511, end_position_on_the_genomic_accession=25202432, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  76089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76089, CurrentID=0, Status='live', Symbol='Rapgef2', Aliases='5830453M24Rik, CNRasGEF, Pdzgef1, RA-GEF-1, mKIAA0313, nRapGEP', description='Rap guanine nucleotide exchange factor (GEF) 2', other_designations='PDZ domain-containing guanine nucleotide exchange factor 1|cyclic nucleotide ras GEF|neural RAP guanine nucleotide exchange protein|ras/Rap1-associating GEF-1', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79062516, end_position_on_the_genomic_accession=79287244, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  207165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207165, CurrentID=0, Status='live', Symbol='Bptf', Aliases='9430093H17Rik, Falz', description='bromodomain PHD finger transcription factor', other_designations='fetal Alzheimer antigen', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107033081, end_position_on_the_genomic_accession=107131922, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  207175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207175, CurrentID=0, Status='live', Symbol='Cetn4', Aliases='', description='centrin 4', other_designations='centrin4|cetn4', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37308627, end_position_on_the_genomic_accession=37312446, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76108, CurrentID=0, Status='live', Symbol='Rap2a', Aliases='5830461H18Rik', description='RAS related protein 2a', other_designations='', map_location='14 64.72 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=120478461, end_position_on_the_genomic_accession=120507194, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  207181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207181, CurrentID=0, Status='live', Symbol='Rbms3', Aliases='6720477E09Rik, 8430436O14Rik', description='RNA binding motif, single stranded interacting protein', other_designations='RNA binding motif, single stranded interacting protein 3|RNA-binding protein RBMS3', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=116572746, end_position_on_the_genomic_accession=118040292, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  207182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207182, CurrentID=0, Status='live', Symbol='Ggt7', Aliases='1110017C11Rik, 6330563L03Rik, Ggtl3', description='gamma-glutamyltransferase 7', other_designations='GGT 7|gamma-glutamyltransferase-like 3|gamma-glutamyltranspeptidase 7|glutathione hydrolase 7|mFLJ00311 protein', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155490379, end_position_on_the_genomic_accession=155514860, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  76113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76113, CurrentID=0, Status='live', Symbol='Lpo', Aliases='5830499B15Rik', description='lactoperoxidase', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87806428, end_position_on_the_genomic_accession=87826135, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  76117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76117, CurrentID=0, Status='live', Symbol='Arhgap15', Aliases='5830480G12Rik', description='Rho GTPase activating protein 15', other_designations='rho-type GTPase-activating protein 15', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=43748783, end_position_on_the_genomic_accession=44395953, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  102443350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102443350, CurrentID=0, Status='live', Symbol='Xndc1', Aliases='Xndr', description='Xrcc1 N-terminal domain containing 1', other_designations='', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102065491, end_position_on_the_genomic_accession=102083762, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  102443351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102443351, CurrentID=0, Status='live', Symbol='Xntrpc', Aliases='Xndr-trpc2', description='Xndc1-transient receptor potential cation channel, subfamily C, member 2 readthrough', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102065491, end_position_on_the_genomic_accession=102096864, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  76123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76123, CurrentID=0, Status='live', Symbol='Gpsm2', Aliases='6230410J09Rik, LGN, Pins', description='G-protein signalling modulator 2 (AGS3-like, C. elegans)', other_designations='pins homolog (Drosophia)', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108678638, end_position_on_the_genomic_accession=108722330, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  76130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76130, CurrentID=0, Status='live', Symbol='Las1l', Aliases='1810030A06Rik, 5830482G23Rik', description='LAS1-like (S. cerevisiae)', other_designations='protein LAS1 homolog', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=95920045, end_position_on_the_genomic_accession=95956974, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  76131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76131, CurrentID=0, Status='live', Symbol='Depdc1a', Aliases='5830484J08Rik, Depdc1', description='DEP domain containing 1a', other_designations='', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=159495426, end_position_on_the_genomic_accession=159529955, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76132, CurrentID=0, Status='live', Symbol='Faxc', Aliases='6230409E13Rik, AW742931', description='failed axon connections homolog', other_designations='', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=21930001, end_position_on_the_genomic_accession=22001461, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  207209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207209, CurrentID=0, Status='live', Symbol='Ccdc154', Aliases='Ccdc157, Gm317', description='coiled-coil domain containing 154', other_designations='Grailexp predicted|golgin-160-like protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25162168, end_position_on_the_genomic_accession=25171913, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  76138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76138, CurrentID=0, Status='live', Symbol='Ccdc138', Aliases='6230424H07Rik, BC042726', description='coiled-coil domain containing 138', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=58497701, end_position_on_the_genomic_accession=58576244, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  207212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207212, CurrentID=0, Status='live', Symbol='Arhgef17', Aliases='8030463K16, AI428794, AW558066, BC035332, mKIAA0337', description='Rho guanine nucleotide exchange factor (GEF) 17', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100869746, end_position_on_the_genomic_accession=100932161, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  207213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207213, CurrentID=0, Status='live', Symbol='Tdpoz1', Aliases='2cpoz56, Mapp2, Spopl1', description='TD and POZ domain containing 1', other_designations='MAPP family protein 2|MATH- and POZ-encompassing protein 2|POZ 56 protein|tdpoz1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93669333, end_position_on_the_genomic_accession=93676283, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  76142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76142, CurrentID=0, Status='live', Symbol='Ppp1r14c', Aliases='6330514J04Rik, KEPI', description='protein phosphatase 1, regulatory (inhibitor) subunit 14c', other_designations='PKC-potentiated PP1 inhibitory protein|kinase-enhanced PP1 inhibitor', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=3366071, end_position_on_the_genomic_accession=3464979, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  207215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207215, CurrentID=0, Status='live', Symbol='Fbxo40', Aliases='9830003A13Rik, mKIAA1195', description='F-box protein 40', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36963461, end_position_on_the_genomic_accession=36990535, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  207227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207227, CurrentID=0, Status='live', Symbol='Stxbp5l', Aliases='9430094L02, A830015P08Rik, LLGL4, T2dm1, t2md1, tomosyn-2', description='syntaxin binding protein 5-like', other_designations='insulin level locus 1|lethal(2) giant larvae protein homolog 4|tomosyn 2', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37107307, end_position_on_the_genomic_accession=37384988, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  76156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76156, CurrentID=0, Status='live', Symbol='Fam131b', Aliases='6330503C03Rik, 6530406I18Rik, AV277466', description='family with sequence similarity 131, member B', other_designations='', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42315305, end_position_on_the_genomic_accession=42324640, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76157, CurrentID=0, Status='live', Symbol='Slc35d3', Aliases='6230421J19Rik, Frcl1', description='solute carrier family 35, member D3', other_designations='frc, fringe-like 1|fringe connection-like protein 1', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19847917, end_position_on_the_genomic_accession=19852361, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76161, CurrentID=0, Status='live', Symbol='Lamp5', Aliases='3110035N03Rik, 6330527O06Rik, BAD-LAMP, LAMP-5', description='lysosomal-associated membrane protein family, member 5', other_designations='LAMP family protein C20orf103 homolog|Protein C20orf103|brain and dendritic cell-associated LAMP|brain-associated LAMP-like protein|lysosome-associated membrane protein 5', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=136051959, end_position_on_the_genomic_accession=136069917, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76166, CurrentID=0, Status='live', Symbol='Rsg1', Aliases='6330545A04Rik, Gm723', description='REM2 and RAB-like small GTPase 1', other_designations='miro domain-containing protein C1orf89 homolog', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141213666, end_position_on_the_genomic_accession=141220365, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76167, CurrentID=0, Status='live', Symbol='Snrnp35', Aliases='6330548G22Rik', description='small nuclear ribonucleoprotein 35 (U11/U12)', other_designations='U1 snRNP-binding protein homolog|U11/U12 snRNP 35 kDa protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124483080, end_position_on_the_genomic_accession=124491323, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  338320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338320, CurrentID=0, Status='live', Symbol='Mia2', Aliases='', description='melanoma inhibitory activity 2', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=59095800, end_position_on_the_genomic_accession=59109130, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76178, CurrentID=0, Status='live', Symbol='Coa5', Aliases='1700001A24Rik, 6330578E17Rik, Pet191', description='cytochrome C oxidase assembly factor 5', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37417085, end_position_on_the_genomic_accession=37430103, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76179, CurrentID=0, Status='live', Symbol='Usp31', Aliases='6330567E21Rik, mKIAA1203', description='ubiquitin specific peptidase 31', other_designations='ubiquitin specific protease 31', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121642021, end_position_on_the_genomic_accession=121707344, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  76183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76183, CurrentID=0, Status='live', Symbol='Celf6', Aliases='6330569O16Rik, Brunol6', description='CUGBP, Elav-like family member 6', other_designations='CELF-6|CUG-BP- and ETR-3-like factor 6|RNA-binding protein BRUNOL-6|bruno-like 6, RNA binding protein|bruno-like protein 6', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59578286, end_position_on_the_genomic_accession=59607293, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  76184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76184, CurrentID=0, Status='live', Symbol='Abca6', Aliases='6330565N06Rik', description='ATP-binding cassette, sub-family A (ABC1), member 6', other_designations='ATP-binding cassette transporter sub-family A member 6', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=110174638, end_position_on_the_genomic_accession=110251776, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  76187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76187, CurrentID=0, Status='live', Symbol='Adhfe1', Aliases='6330565B14Rik, AI043035, Adh8', description='alcohol dehydrogenase, iron containing, 1', other_designations='Fe-containing alcohol dehydrogenase 1|HOT|alcohol dehydrogenase 8|alcohol dehydrogenase iron-containing protein 1', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9547969, end_position_on_the_genomic_accession=9577968, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  76192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76192, CurrentID=0, Status='live', Symbol='Abhd12', Aliases='1500011G07Rik, 6330583M11Rik, AI431047, AW547313', description='abhydrolase domain containing 12', other_designations='2-arachidonoylglycerol hydrolase|abhydrolase domain-containing protein 12', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150832493, end_position_on_the_genomic_accession=150904745, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  338337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338337, CurrentID=0, Status='live', Symbol='Cog3', Aliases='E430004N23', description='component of oligomeric golgi complex 3', other_designations='COG complex subunit 3|component of oligomeric golgi complex 2', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=75702351, end_position_on_the_genomic_accession=75754578, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  76199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76199, CurrentID=0, Status='live', Symbol='Med13l', Aliases='2210413I17Rik, 6330591G05Rik, 9030618F05Rik, Thrap2, Trap240L, mKIAA1025', description='mediator complex subunit 13-like', other_designations='thyroid hormone receptor associated protein 2|thyroid hormone receptor-associated protein 2|thyroid hormone receptor-associated protein complex 240 kDa component-like', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=118560706, end_position_on_the_genomic_accession=118765438, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  338346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338346, CurrentID=0, Status='live', Symbol='Gpr21', Aliases='C230004C13Rik', description='G protein-coupled receptor 21', other_designations='G protein-coupled receptor GPR21', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37516626, end_position_on_the_genomic_accession=37519281, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  338348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338348, CurrentID=0, Status='live', Symbol='Ttc16', Aliases='1200002K10Rik', description='tetratricopeptide repeat domain 16', other_designations='TPR repeat protein 16', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32757026, end_position_on_the_genomic_accession=32775663, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  338349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338349, CurrentID=0, Status='live', Symbol='Cntln', Aliases='AI448092, B430108F07Rik, D530005L17Rik', description='centlein, centrosomal protein', other_designations='cntln', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=84883512, end_position_on_the_genomic_accession=85131923, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  207278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207278, CurrentID=0, Status='live', Symbol='Fchsd2', Aliases='BC034086, R74866, Sh3md3, mKIAA0769', description='FCH and double SH3 domains 2', other_designations='FCH and double SH3 domains protein 2|SH3 multiple domains 3|SH3 multiple domains protein 3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101108711, end_position_on_the_genomic_accession=101284405, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  338351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338351, CurrentID=0, Status='live', Symbol='Akap17b', Aliases='AA407619, AKAP-17B, B230333C21Rik, C230056F04Rik, PRKA17B, Sfrs17b, Srsf17b, Talia', description='A kinase (PRKA) anchor protein 17B', other_designations='protein Talia|protein kinase A-anchoring protein 17B|serine/arginine-rich splicing factor 17B|splicing factor, arginine/serine-rich 17b', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36608173, end_position_on_the_genomic_accession=36645471, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  338352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338352, CurrentID=0, Status='live', Symbol='Nell1', Aliases='B230343H07Rik, l7R6', description='NEL-like 1', other_designations='NEL-like protein 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=49975350, end_position_on_the_genomic_accession=50863289, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  338354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338354, CurrentID=0, Status='live', Symbol='Zfp780b', Aliases='B230208L21Rik', description='zinc finger protein 780B', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27959135, end_position_on_the_genomic_accession=27979192, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  338355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338355, CurrentID=0, Status='live', Symbol='Fkbp15', Aliases='BB131447, C430014M02Rik, FKBO133, FKBP-133, FKBP-15, FKBP133, mKIAA0674', description='FK506 binding protein 15', other_designations='133 kDa FK506-binding protein|133 kDa FKBP|FK506-binding protein 133|WAFL', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62300342, end_position_on_the_genomic_accession=62360591, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  338359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338359, CurrentID=0, Status='live', Symbol='Supv3l1', Aliases='6330443E10Rik', description='suppressor of var1, 3-like 1 (S. cerevisiae)', other_designations='SUV3-like protein 1|suppressor of var1 3-like protein 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62429329, end_position_on_the_genomic_accession=62451294, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  76217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76217, CurrentID=0, Status='live', Symbol='Jakmip2', Aliases='6430702L21Rik, AI850334, D930046L20Rik', description='janus kinase and microtubule interacting protein 2', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=43527096, end_position_on_the_genomic_accession=43687773, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  338362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338362, CurrentID=0, Status='live', Symbol='Ust', Aliases='D930010O20Rik, UA2OST', description='uronyl-2-sulfotransferase', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=8204753, end_position_on_the_genomic_accession=8518825, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  76219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76219, CurrentID=0, Status='live', Symbol='Arxes1', Aliases='6530401D17Rik, Spc22/23, Spcs3', description='adipocyte-related X-chromosome expressed sequence 1', other_designations='SPase 22 kDa subunit|SPase 22/23 kDa subunit|microsomal signal peptidase 22/23 kDa subunit|microsomal signal peptidase 23 kDa subunit', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136033367, end_position_on_the_genomic_accession=136034946, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  338364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338364, CurrentID=0, Status='live', Symbol='Trim65', Aliases='4732463G12Rik', description='tripartite motif-containing 65', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116124708, end_position_on_the_genomic_accession=116131128, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  338365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338365, CurrentID=0, Status='live', Symbol='Slc41a2', Aliases='A230035L05Rik, SLC41A1-L1', description='solute carrier family 41, member 2', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=83230848, end_position_on_the_genomic_accession=83338329, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  338366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338366, CurrentID=0, Status='live', Symbol='Mia3', Aliases='9130229H14Rik, A930039G15Rik, AU022327, B230399H06Rik, C80126, Gm1525, TANGO1, Tango', description='melanoma inhibitory activity 3', other_designations='transport and Golgi organization protein 1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=183326236, end_position_on_the_genomic_accession=183369565, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  76223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76223, CurrentID=0, Status='live', Symbol='Agbl3', Aliases='2900053G10Rik, 4930431N21Rik, 6530406M24Rik, CCP3', description='ATP/GTP binding protein-like 3', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34780432, end_position_on_the_genomic_accession=34862184, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  338368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338368, CurrentID=0, Status='live', Symbol='Fam109b', Aliases='C920005C14Rik, IPIP27B, Ses2', description='family with sequence similarity 109, member B', other_designations='27 kDa inositol polyphosphate phosphatase interacting protein B', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82338947, end_position_on_the_genomic_accession=82345710, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  338369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338369, CurrentID=0, Status='live', Symbol='Tmem220', Aliases='', description='transmembrane protein 220', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67025154, end_position_on_the_genomic_accession=67035312, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  338370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338370, CurrentID=0, Status='live', Symbol='Nalcn', Aliases='A530023G15Rik, AI849508, Vgcnl1', description='sodium leak channel, non-selective', other_designations='voltage gated channel like 1|voltage gated channel-like protein 1', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=123276634, end_position_on_the_genomic_accession=123627363, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  338371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338371, CurrentID=0, Status='live', Symbol='Endov', Aliases='A730011L01Rik', description='endonuclease V', other_designations='novel endonuclease V protein|putative endonuclease FLJ39025 homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119491323, end_position_on_the_genomic_accession=119511465, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  338372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338372, CurrentID=0, Status='live', Symbol='Map3k9', Aliases='E130314H24Rik, Mlk1, Prke1', description='mitogen-activated protein kinase kinase kinase 9', other_designations='mitogen activated protein kinase kinase kinase 9', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81714950, end_position_on_the_genomic_accession=81782549, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  76229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76229, CurrentID=0, Status='live', Symbol='Vmn2r29', Aliases='6430701C03Rik', description='vomeronasal 2, receptor 29', other_designations='vomeronasal receptor Vmn2r29', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7212996, end_position_on_the_genomic_accession=7278289, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  338374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338374, CurrentID=0, Status='live', Symbol='Ifnl3', Aliases='IFL-1, IL-28B, INF-alpha, INF-lambda, Il28, Il28b', description='interferon lambda 3', other_designations='IFN-lambda-3|interferon alpha|interleukin 28|interleukin 28B|interleukin-28B', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28522833, end_position_on_the_genomic_accession=28524401, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  338375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338375, CurrentID=0, Status='live', Symbol='Atp6v1g3', Aliases='', description='ATPase, H+ transporting, lysosomal V1 subunit G3', other_designations='ATPase, H+ transporting, V1 subunit G|V-ATPase 13 kDa subunit 3|V-ATPase subunit G 3|vacuolar proton pump subunit G 3', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=138273738, end_position_on_the_genomic_accession=138289462, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  207304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207304, CurrentID=0, Status='live', Symbol='Hectd1', Aliases='A630086P08Rik, AI844876, b2b327Clo, opm', description='HECT domain containing 1', other_designations='HECT domain-containing protein 1|protein open mind', map_location='12 22.11 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51743721, end_position_on_the_genomic_accession=51829605, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  76233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76233, CurrentID=0, Status='live', Symbol='Dnttip1', Aliases='6430706C13Rik', description='deoxynucleotidyltransferase, terminal, interacting protein 1', other_designations='tdIF1|tdT-interacting factor 1|terminal deoxynucleotidyltransferase-interacting factor 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164745992, end_position_on_the_genomic_accession=164768220, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76237, CurrentID=0, Status='live', Symbol='6430628N08Rik', Aliases='', description='RIKEN cDNA 6430628N08 gene', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115890353, end_position_on_the_genomic_accession=115899228, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  76238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76238, CurrentID=0, Status='live', Symbol='Grhpr', Aliases='1110059D05Rik, 6430629L09Rik, Glxr', description='glyoxylate reductase/hydroxypyruvate reductase', other_designations='', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44981387, end_position_on_the_genomic_accession=44992085, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76246, CurrentID=0, Status='live', Symbol='Rtf1', Aliases='2900005O08Rik, 6530416A09Rik, AI853581, AU043144, AW553985, Gtl7, cbp82', description='Rtf1, Paf1/RNA polymerase II complex component, homolog (S. cerevisiae)', other_designations='Rtf1 Paf1/RNA polymerase II complex component homolog|gene trap locus 7', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119675068, end_position_on_the_genomic_accession=119735407, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  76251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76251, CurrentID=0, Status='live', Symbol='Ercc6l2', Aliases='0610007P08Rik, 1700019D06Rik, 9330134C04Rik, Rad26l, Sr278', description='excision repair cross-complementing rodent repair deficiency, complementation group 6 like 2', other_designations='DNA repair and recombination protein RAD26-like|putative repair and recombination helicase RAD26L|stretch responsive protein 278', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=63813415, end_position_on_the_genomic_accession=63900302, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  76252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76252, CurrentID=0, Status='live', Symbol='Atp6v0e2', Aliases='0610006O14Rik, NM9.2', description='ATPase, H+ transporting, lysosomal V0 subunit E2', other_designations='ATPase, H+ transporting, lysosomal, V0 subunit E|V-ATPase subunit e 2|vacuolar proton pump subunit e 2', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48537569, end_position_on_the_genomic_accession=48541801, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100090, CurrentID=0, Status='live', Symbol='Zbtb48', Aliases='0610011D15Rik, AI327031, Hkr3', description='zinc finger and BTB domain containing 48', other_designations='GLI-Krueppel family member HKR3|GLI-Kruppel family member HKR3|krueppel-related zinc finger protein 3', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152019773, end_position_on_the_genomic_accession=152027671, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  76257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76257, CurrentID=0, Status='live', Symbol='Slc38a3', Aliases='0610012J02Rik, D9Ucla2, Nat1, Slc38-3, Sn1, Snat3, mNAT', description='solute carrier family 38, member 3', other_designations='N system amino acids transporter NAT-1|N-system amino acid transporter 1|Na(+)-coupled neutral amino acid transporter 3|slc38a3|system N amino acid transporter 1', map_location='9 58.69 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107651154, end_position_on_the_genomic_accession=107668968, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  338403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338403, CurrentID=0, Status='live', Symbol='Cndp1', Aliases='AI746433, Cn1', description='carnosine dipeptidase 1 (metallopeptidase M20 family)', other_designations='CNDP dipeptidase 1|carnosinase 1', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84610384, end_position_on_the_genomic_accession=84650113, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  76260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76260, CurrentID=0, Status='live', Symbol='Ttc8', Aliases='0610012F22Rik, AV001447, BBS8', description='tetratricopeptide repeat domain 8', other_designations='TPR repeat protein 8|bardet-Biedl syndrome 8 protein homolog', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98920574, end_position_on_the_genomic_accession=98983250, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  209318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209318, CurrentID=0, Status='live', Symbol='Gps1', Aliases='Cops1, Csn1, R75577, Sgn1', description='G protein pathway suppressor 1', other_designations='JAB1-containing signalosome subunit 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120784260, end_position_on_the_genomic_accession=120789107, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  76263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76263, CurrentID=0, Status='live', Symbol='Gstk1', Aliases='0610025I19Rik, AW260476, DsbA-L', description='glutathione S-transferase kappa 1', other_designations='GST 13-13|GST class-kappa|GSTK1-1|glutathione S-transferase class kappa|glutathione S-transferase subunit 13|mGSTK1', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42245935, end_position_on_the_genomic_accession=42250441, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  76265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76265, CurrentID=0, Status='live', Symbol='Tsen54', Aliases='0610034P02Rik', description='tRNA splicing endonuclease 54 homolog (S. cerevisiae)', other_designations='likely ortholog of H. sapiens SEN54L|tRNA-intron endonuclease Sen54', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115813743, end_position_on_the_genomic_accession=115823102, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76267, CurrentID=0, Status='live', Symbol='Fads1', Aliases='0710001O03Rik, A930006B21Rik, AI317215, DSD', description='fatty acid desaturase 1', other_designations='D5D|delta(5) desaturase|delta(5) fatty acid desaturase|delta-5 desaturase', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10182888, end_position_on_the_genomic_accession=10196874, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  76273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76273, CurrentID=0, Status='live', Symbol='Ndfip2', Aliases='0710001O20Rik, 2810436B12Rik, 9130207N19Rik, N4wbp5a, mKIAA1165', description='Nedd4 family interacting protein 2', other_designations='NEDD4 WW domain-binding protein 5A', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=105258673, end_position_on_the_genomic_accession=105309298, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  76279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76279, CurrentID=0, Status='live', Symbol='Cyp2d26', Aliases='1300006E06Rik', description='cytochrome P450, family 2, subfamily d, polypeptide 26', other_designations='CYPIID26|cyp2d26|cytochrome P450, 2d26', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82790107, end_position_on_the_genomic_accession=82794452, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  207352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207352, CurrentID=0, Status='live', Symbol='Sec23ip', Aliases='D7Ertd373e, p125', description='Sec23 interacting protein', other_designations='', map_location='7 70.51 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128744862, end_position_on_the_genomic_accession=128784836, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  76281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76281, CurrentID=0, Status='live', Symbol='Tax1bp3', Aliases='1300011C24Rik, TIP-1', description='Tax1 (human T cell leukemia virus type I) binding protein 3', other_designations='Tax1 (human T-cell leukemia virus type I) binding protein 3|tax interaction protein 1|tax1bp3', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73177083, end_position_on_the_genomic_accession=73182046, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76282, CurrentID=0, Status='live', Symbol='Gpt', Aliases='1300007J06Rik, 2310022B03Rik, ALT, ALT1-1, Gpt1, Gpt', description='glutamic pyruvic transaminase, soluble', other_designations='GPT 1|glutamate pyruvate transaminase 1|glutamic pyruvic transaminase 1, soluble|glutamic--alanine transaminase 1|glutamic--pyruvic transaminase 1', map_location='15 36.28 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76696764, end_position_on_the_genomic_accession=76699675, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76286, CurrentID=0, Status='live', Symbol='1110006E14Rik', Aliases='', description='RIKEN cDNA 1110006E14 gene', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  76293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76293, CurrentID=0, Status='live', Symbol='Mfap4', Aliases='1110007F23Rik', description='microfibrillar-associated protein 4', other_designations='', map_location='11 37.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61485378, end_position_on_the_genomic_accession=61488706, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  338438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338438, CurrentID=0, Status='live', Symbol='Hymai', Aliases='', description='hydatidiform mole associated and imprinted transcript', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  76295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76295, CurrentID=0, Status='live', Symbol='Atp11b', Aliases='1110019I14Rik, ATPIF, ATPIR, mKIAA0956', description='ATPase, class VI, type 11B', other_designations='', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=35754062, end_position_on_the_genomic_accession=35856276, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  76299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76299, CurrentID=0, Status='live', Symbol='Erp44', Aliases='1110001E24Rik, AI849526, AL033348, Txndc4', description='endoplasmic reticulum protein 44', other_designations='ER protein 44|endoplasmic reticulum resident protein 44kDa|endoplasmic reticulum resident protein ERp44|thioredoxin domain containing 4 (endoplasmic reticulum)|thioredoxin domain-containing protein 4', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48193331, end_position_on_the_genomic_accession=48279601, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  76302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76302, CurrentID=0, Status='live', Symbol='Pcnp', Aliases='1110018D06Rik, AI647035', description='PEST proteolytic signal containing nuclear protein', other_designations='PEST-containing nuclear protein', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56015506, end_position_on_the_genomic_accession=56030177, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  76303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76303, CurrentID=0, Status='live', Symbol='Osbp', Aliases='1110018F06Rik, AW559088, mKIAA4220', description='oxysterol binding protein', other_designations='', map_location='19 8.58 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11965844, end_position_on_the_genomic_accession=11994114, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  76306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76306, CurrentID=0, Status='live', Symbol='Slc18b1', Aliases='1110021L09Rik', description='solute carrier family 18, subfamily B, member 1', other_designations='MFS-type transporter C6orf192 homolog|solute carrier family 18 member B1', map_location='10 A3-A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23796654, end_position_on_the_genomic_accession=23827969, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  76308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76308, CurrentID=0, Status='live', Symbol='Rab1b', Aliases='1110011F09Rik', description='RAB1B, member RAS oncogene family', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5099207, end_position_on_the_genomic_accession=5106996, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  207393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207393, CurrentID=0, Status='live', Symbol='Elfn2', Aliases='6330514E13, AW048948, BC094219, Lrrc62, Ppp1r29', description='leucine rich repeat and fibronectin type III, extracellular 2', other_designations='5leucine rich repeat containing 62|extracellular leucine-rich repeat and fibronectin type III domain-containing protein 2|leucine rich repeat containing 62|leucine-rich repeat and fibronectin type-III domain-containing protein 6|leucine-rich repeat-containing protein 62', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78667331, end_position_on_the_genomic_accession=78718873, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  338467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338467, CurrentID=0, Status='live', Symbol='Morc3', Aliases='1110051N18Rik, AI452146, BF318192, D16Jhu32e, NXP2, Zcwcc3', description='microrchidia 3', other_designations='nuclear matrix protein 2|zinc finger CW-type coiled-coil domain protein 3|zinc finger, CW-type with coiled-coil domain 3', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93832121, end_position_on_the_genomic_accession=93876077, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  76332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76332, CurrentID=0, Status='live', Symbol='Cog2', Aliases='1190002B08Rik, 2700012E02Rik, C86089, Ldlc', description='component of oligomeric golgi complex 2', other_designations='COG complex subunit 2|low density lipoprotein receptor defect C complementing|low density lipoprotein receptor defect C-complementing protein', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124520767, end_position_on_the_genomic_accession=124552008, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  76338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76338, CurrentID=0, Status='live', Symbol='Rab2b', Aliases='1500012D09Rik, 4930528G15Rik, A230002G14, D530043M21Rik', description='RAB2B, member RAS oncogene family', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52261759, end_position_on_the_genomic_accession=52279395, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  207425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207425, CurrentID=0, Status='live', Symbol='Wdr11', Aliases='2900055P10Rik, AW489876, Brwd2, mKIAA1351', description='WD repeat domain 11', other_designations='bromodomain and WD repeat domain containing 2|bromodomain and WD repeat-containing protein 2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=129591714, end_position_on_the_genomic_accession=129635738, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  76355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76355, CurrentID=0, Status='live', Symbol='Tgds', Aliases='2610017J16Rik, 2610025M23Rik, AI648925', description='TDP-glucose 4,6-dehydratase', other_designations='', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118111911, end_position_on_the_genomic_accession=118132765, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  76357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76357, CurrentID=0, Status='live', Symbol='Trmt5', Aliases='2610027O18Rik, mKIAA1393', description='TRM5 tRNA methyltransferase 5', other_designations='M1G-methyltransferase|TRM5 tRNA methyltransferase 5 homolog|tRNA (guanine-N(1)-)-methyltransferase|tRNA [GM37] methyltransferase|tRNA methyltransferase 5 homolog', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73268515, end_position_on_the_genomic_accession=73286787, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76365, CurrentID=0, Status='live', Symbol='Tbx18', Aliases='2810012F10Rik, 2810404D13Rik', description='T-box18', other_designations='T-box protein 18', map_location='9 E3.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=87702800, end_position_on_the_genomic_accession=87731260, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76366, CurrentID=0, Status='live', Symbol='Mtif3', Aliases='2810012L14Rik, AI414549', description='mitochondrial translational initiation factor 3', other_designations='IF-3(Mt)|IF-3Mt|IF3(mt)|IF3mt', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146951573, end_position_on_the_genomic_accession=146963805, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76367, CurrentID=0, Status='live', Symbol='Trp53rkb', Aliases='2810408M09Rik, 4933401B08Rik, 5630401H01Rik, Nori-2, PRPK, Tp53rk, Trp53rk, Trp53tk, mNori-2p', description='transformation related protein 53 regulating kinase B', other_designations='EKC/KEOPS complex subunit Tp53rk|TP53 regulating kinase|Trp53 regulating kinase|atypical serine/threonine protein kinase Tp53rk|p53-related protein kinase', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=166792518, end_position_on_the_genomic_accession=166799492, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  76373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76373, CurrentID=0, Status='live', Symbol='Zfp773', Aliases='2810409K11Rik', description='zinc finger protein 773', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7127073, end_position_on_the_genomic_accession=7136785, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76375, CurrentID=0, Status='live', Symbol='Det1', Aliases='2610034H20Rik', description='de-etiolated homolog 1 (Arabidopsis)', other_designations='de-etiolated-1 homolog', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=78822959, end_position_on_the_genomic_accession=78847382, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76376, CurrentID=0, Status='live', Symbol='Slc24a2', Aliases='2810021B17Rik, 6330417K15Rik, AI847460, Nckx2', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 2', other_designations='', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86983124, end_position_on_the_genomic_accession=87230554, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  338521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338521, CurrentID=0, Status='live', Symbol='Fa2h', Aliases='FAAH, Faxdc1, G630055L08Rik', description='fatty acid 2-hydroxylase', other_designations='fatty acid alpha-hydroxylase|fatty acid hydroxylase domain containing 1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111345138, end_position_on_the_genomic_accession=111393821, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  76378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76378, CurrentID=0, Status='live', Symbol='Ropn1', Aliases='1700008N21Rik, ODF6, RHPNAP1, Rppn, ropporin', description='ropporin, rhophilin associated protein 1', other_designations="rhophilin-associated 'oppo' protein|rhophilin-associated oppo protein|rhophilin-associated protein 1", map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=34649842, end_position_on_the_genomic_accession=34678610, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  338523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=338523, CurrentID=0, Status='live', Symbol='Kdm7a', Aliases='A630082K20Rik, BB041802, ENSMUSG00000073143, Jhdm1d, mKIAA1718', description='lysine (K)-specific demethylase 7A', other_designations='jmjC domain-containing histone demethylation protein 1D|jumonji C domain-containing histone demethylase 1 homolog D|lysine-specific demethylase 7', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39136620, end_position_on_the_genomic_accession=39206773, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  76380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76380, CurrentID=0, Status='live', Symbol='Cep112', Aliases='1700001M19Rik, 1700029K01Rik, 8430407H02Rik, AV043680, AV207351, Ccdc46, Macoco', description='centrosomal protein 112', other_designations='coiled-coil domain containing 46|coiled-coil domain-containing protein 46', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=108425068, end_position_on_the_genomic_accession=108860615, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  76383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76383, CurrentID=0, Status='live', Symbol='1700012L04Rik', Aliases='H2al1', description='RIKEN cDNA 1700012L04 gene', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9283764, end_position_on_the_genomic_accession=9284288, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  76389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76389, CurrentID=0, Status='live', Symbol='Ankrd36', Aliases='1700008J08Rik, 1700012M14Rik, Ankrd48', description='ankyrin repeat domain 36', other_designations='GC3|ankyrin repeat domain 48', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5569680, end_position_on_the_genomic_accession=5693796, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  76390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76390, CurrentID=0, Status='live', Symbol='Zfp735', Aliases='1700012C15Rik', description='zinc finger protein 735', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73688778, end_position_on_the_genomic_accession=73713808, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76399, CurrentID=0, Status='live', Symbol='Il31', Aliases='1700013B14Rik', description='interleukin 31', other_designations='IL-31', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123480149, end_position_on_the_genomic_accession=123482634, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76400, CurrentID=0, Status='live', Symbol='Pbp2', Aliases='1700023A18Rik, AI326308, Pebp-2, Pebp2', description='phosphatidylethanolamine binding protein 2', other_designations='phosphatidylethanolamine binding protein-2 variant 1', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135309129, end_position_on_the_genomic_accession=135310384, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  207474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207474, CurrentID=0, Status='live', Symbol='Kctd12b', Aliases='9530071H11Rik, B430319H24Rik', description='potassium channel tetramerisation domain containing 12b', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153685154, end_position_on_the_genomic_accession=153696386, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76405, CurrentID=0, Status='live', Symbol='1700018B08Rik', Aliases='1700084O22Rik', description='RIKEN cDNA 1700018B08 gene', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121530776, end_position_on_the_genomic_accession=121544351, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76406, CurrentID=0, Status='live', Symbol='1700019B03Rik', Aliases='', description='RIKEN cDNA 1700019B03 gene', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3468886, end_position_on_the_genomic_accession=3487183, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76407, CurrentID=0, Status='live', Symbol='Sun5', Aliases='1700021O15Rik, SPAG4L-2, SRG4, Spag4l', description='Sad1 and UNC84 domain containing 5', other_designations='sperm-associated antigen 4-like protein|spermatogenesis-related gene 4', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153856188, end_position_on_the_genomic_accession=153871095, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  76408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76408, CurrentID=0, Status='live', Symbol='Abcc3', Aliases='1700019L09Rik, ABC31, MLP2, MOAT-D, MRP3', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 3', other_designations='ATP-binding cassette protein C3|ATP-binding cassette sub-family C member 3|ATP-binding cassette transporter|multidrug resistance-associated protein 3|transporter', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94343295, end_position_on_the_genomic_accession=94393022, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  76411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76411, CurrentID=0, Status='live', Symbol='Ift43', Aliases='1700019E19Rik, R75064', description='intraflagellar transport 43', other_designations='intraflagellar transport 43 homolog', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86082533, end_position_on_the_genomic_accession=86162459, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76415, CurrentID=0, Status='live', Symbol='Fam187b', Aliases='1700020B09Rik, Tmem162', description='family with sequence similarity 187, member B', other_designations='family with sequence similarity 187, member B pseudogene|transmembrane protein 162', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30973804, end_position_on_the_genomic_accession=30989726, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76416, CurrentID=0, Status='live', Symbol='Znrd1as', Aliases='1700022C21Rik, AI746295, Znrd1-as', description='zinc ribbon domain containing 1, antisense', other_designations='T-complex testis-expressed protein 4|ZNRD1 antisense RNA 1|ZNRD1 antisense gene protein 1|Znrd1 antisense', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36958592, end_position_on_the_genomic_accession=36965625, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76421, CurrentID=0, Status='live', Symbol='1700028K03Rik', Aliases='', description='RIKEN cDNA 1700028K03 gene', other_designations='', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107534711, end_position_on_the_genomic_accession=107551542, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  207495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207495, CurrentID=0, Status='live', Symbol='Baiap2l2', Aliases='BC048937', description='BAI1-associated protein 2-like 2', other_designations='BAI1-associated protein 2-like protein 2|pinkbar|planar intestinal- and kidney-specific BAR domain protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79258195, end_position_on_the_genomic_accession=79285532, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  76425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76425, CurrentID=0, Status='live', Symbol='Gid8', Aliases='2310003C23Rik, 4833420G11Rik, AI451474, Twa1', description='GID complex subunit 8', other_designations='GID complex subunit 8 homolog|protein C20orf11 homolog|two hybrid-associated protein 1 with RanBPM', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180710226, end_position_on_the_genomic_accession=180721599, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  76426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76426, CurrentID=0, Status='live', Symbol='Fam209', Aliases='1700029J11Rik', description='family with sequence similarity 209', other_designations='likely orthologue of H. sapiens chromosome 20 open reading frame 106 (C20orf106)', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172472554, end_position_on_the_genomic_accession=172474316, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  76429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76429, CurrentID=0, Status='live', Symbol='Lhpp', Aliases='2310007H09Rik', description='phospholysine phosphohistidine inorganic pyrophosphate phosphatase', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132610615, end_position_on_the_genomic_accession=132706420, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76438, CurrentID=0, Status='live', Symbol='Rftn1', Aliases='2310015N21Rik, AI853526', description='raftlin lipid raft linker 1', other_designations='raft-linking protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=49993258, end_position_on_the_genomic_accession=50190497, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76441, CurrentID=0, Status='live', Symbol='Daam2', Aliases='2310016D11Rik, AI843643, AW557870', description='dishevelled associated activator of morphogenesis 2', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=49456022, end_position_on_the_genomic_accession=49564491, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  76444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76444, CurrentID=0, Status='live', Symbol='Krtap4-7', Aliases='2310037K05Rik, KAP4.7, KRTAP4.7', description='keratin associated protein 4-7', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99643112, end_position_on_the_genomic_accession=99644089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  12741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12741, CurrentID=0, Status='live', Symbol='Cldn5', Aliases='AI854493, MBEC1, Tmvcf', description='claudin 5', other_designations='brain endothelial cell clone 1 protein|lung-specific membrane protein|transmembrane protein deleted in VCFS', map_location='16 11.63 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18776847, end_position_on_the_genomic_accession=18778262, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  76448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76448, CurrentID=0, Status='live', Symbol='Ppp1r18', Aliases='2310014H01Rik, AA407849, AI450394', description='protein phosphatase 1, regulatory subunit 18', other_designations='protein phosphatase 1 F-actin cytoskeleton-targeting subunit|protein phosphatase 1 regulatory subunit 18', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35865495, end_position_on_the_genomic_accession=35875596, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  207521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207521, CurrentID=0, Status='live', Symbol='Dtx4', Aliases='AI449438, BC044798, D930042H13, RNF155', description='deltex 4, E3 ubiquitin ligase', other_designations='deltex 4 homolog|deltex4|protein deltex-4', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12466329, end_position_on_the_genomic_accession=12502201, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76453, CurrentID=0, Status='live', Symbol='Prss23', Aliases='2310046G15Rik, 4930479H08Rik, AI790574, Spuve', description='protease, serine 23', other_designations='protease, serine, 23', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=89507785, end_position_on_the_genomic_accession=89527254, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  76454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76454, CurrentID=0, Status='live', Symbol='Fbxo31', Aliases='1110003O08Rik, 2310046N15Rik, Fbx14, Fbxo14', description='F-box protein 31', other_designations='F-box only protein 14', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121549443, end_position_on_the_genomic_accession=121578863, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76457, CurrentID=0, Status='live', Symbol='Ccdc134', Aliases='2310042L06Rik, AW208859', description='coiled-coil domain containing 134', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82127901, end_position_on_the_genomic_accession=82142203, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76459, CurrentID=0, Status='live', Symbol='Car12', Aliases='2310047E01Rik, AI314958, CA-XII, Ca12', description='carbonic anyhydrase 12', other_designations='CA XII|carbonate dehydratase XII|carbonic anhydrase XII', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66713679, end_position_on_the_genomic_accession=66766845, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76464, CurrentID=0, Status='live', Symbol='Casc5', Aliases='2310043D08Rik, 5730505K17Rik', description='cancer susceptibility candidate 5', other_designations='kinetochore-null protein 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119047038, end_position_on_the_genomic_accession=119104829, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  76467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76467, CurrentID=0, Status='live', Symbol='Msrb2', Aliases='2310050L06Rik, Mrsb, Msrb, Pilb', description='methionine sulfoxide reductase B2', other_designations='pilin-like transcription factor', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=19371636, end_position_on_the_genomic_accession=19394976, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  76469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76469, CurrentID=0, Status='live', Symbol='Cmya5', Aliases='2310076E16Rik, 2310076E21Rik, AA420382, AI504003, Srfsd, Tims, sr553', description='cardiomyopathy associated 5', other_designations='TRIM-like protein|heart and skeletal muscle-specific and sprouty domain-containing|myospryn|stretch response protein 553|stretch-response protein 553|stretch-responsive fibronectin protein type 3', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93040713, end_position_on_the_genomic_accession=93145863, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  76477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76477, CurrentID=0, Status='live', Symbol='Pcolce2', Aliases='2400001O18Rik, Pcpe2', description='procollagen C-endopeptidase enhancer 2', other_designations='PCPE-2|procollagen C-proteinase enhancer 2|procollagen COOH-terminal proteinase enhancer 2', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95637628, end_position_on_the_genomic_accession=95695551, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76478, CurrentID=0, Status='live', Symbol='Haus8', Aliases='2410004L22Rik, Hice1', description='4HAUS augmin-like complex, subunit 8', other_designations='HEC1/NDC80 interacting, centrosome associated 1|HEC1/NDC80-interacting centrosome-associated protein 1|sarcoma antigen NY-SAR-48 homolog', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71251124, end_position_on_the_genomic_accession=71272590, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76479, CurrentID=0, Status='live', Symbol='Smndc1', Aliases='2410004J23Rik, 4933440I19Rik', description='survival motor neuron domain containing 1', other_designations='30 kDa splicing factor SMNrp|SMN-related protein|survival motor neuron domain-containing protein 1', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53379214, end_position_on_the_genomic_accession=53391819, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  76482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76482, CurrentID=0, Status='live', Symbol='3110002H16Rik', Aliases='2400010D15Rik, Mic-1, Mic1', description='RIKEN cDNA 3110002H16 gene', other_designations='colon cancer-associated protein Mic1 homolog', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12168691, end_position_on_the_genomic_accession=12198497, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  76483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76483, CurrentID=0, Status='live', Symbol='Lmf1', Aliases='2400010G15Rik, AW822050, Tmem112, cld', description='lipase maturation factor 1', other_designations='transmembrane protein 112', map_location='17 12.7 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25579174, end_position_on_the_genomic_accession=25662826, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  76484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76484, CurrentID=0, Status='live', Symbol='Kndc1', Aliases='2410012C07Rik, B830014K08Rik, VKIND, very-kind', description='kinase non-catalytic C-lobe domain (KIND) containing 1', other_designations='KIND domain-containing protein 1|kinase non-catalytic C-lobe domain-containing protein 1|ras-GEF domain-containing family member 2', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139894696, end_position_on_the_genomic_accession=139941540, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  76485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76485, CurrentID=0, Status='live', Symbol='Glt8d1', Aliases='2410004H05Rik, 5430414N14Rik, AI450005', description='glycosyltransferase 8 domain containing 1', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31001347, end_position_on_the_genomic_accession=31012441, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  76486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76486, CurrentID=0, Status='live', Symbol='Ly6k', Aliases='2410015A16Rik, 3110035B01Rik, mLy-6K', description='lymphocyte antigen 6 complex, locus K', other_designations='ly-6K', map_location='', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74796874, end_position_on_the_genomic_accession=74799968, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76487, CurrentID=0, Status='live', Symbol='Ppp1r3g', Aliases='1600032L17Rik', description='protein phosphatase 1, regulatory (inhibitor) subunit 3G', other_designations='putative protein phosphatase 1 regulatory inhibitor subunit 3G', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=35958799, end_position_on_the_genomic_accession=35970388, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76491, CurrentID=0, Status='live', Symbol='Abhd14b', Aliases='1810013B01Rik', description='abhydrolase domain containing 14b', other_designations='CCG1-interacting factor B|abhydrolase domain-containing protein 14B', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106447737, end_position_on_the_genomic_accession=106452923, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  207565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207565, CurrentID=0, Status='live', Symbol='Camkk2', Aliases='6330570N16Rik, AW061083, mKIAA0787', description='calcium/calmodulin-dependent protein kinase kinase 2, beta', other_designations='caM-KK 2|caM-KK beta|caM-kinase kinase 2|caM-kinase kinase beta|caMKK 2|caMKK beta|calcium/calmodulin-dependent protein kinase kinase beta', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122718580, end_position_on_the_genomic_accession=122779476, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  76497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76497, CurrentID=0, Status='live', Symbol='Ppp1r11', Aliases='1500041B02Rik, AV117883, Hcgv, Tcte5, Tctex-5, Tctex5', description='protein phosphatase 1, regulatory (inhibitor) subunit 11', other_designations='T-complex testis expressed protein 5|t-complex testis-expressed 5', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36948355, end_position_on_the_genomic_accession=36951792, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76498, CurrentID=0, Status='live', Symbol='Paqr4', Aliases='1500004C10Rik', description='progestin and adipoQ receptor family member IV', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23736186, end_position_on_the_genomic_accession=23740898, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76499, CurrentID=0, Status='live', Symbol='Clasp2', Aliases='1500004F14Rik, 8030404L10Rik, C77448, CLASP2beta, mKIAA0627', description='CLIP associating protein 2', other_designations='CLASP2alpha|CLASP2gamma|CLIP-associating protein CLASP2|cytoplasmic linker-associated protein 2', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=113740026, end_position_on_the_genomic_accession=113919697, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  76500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76500, CurrentID=0, Status='live', Symbol='Ip6k2', Aliases='1500005N04Rik, AW050208, Ihpk2', description='inositol hexaphosphate kinase 2', other_designations='insP6 kinase 2|p(i)-uptake stimulator|piUS', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108783764, end_position_on_the_genomic_accession=108806333, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76501, CurrentID=0, Status='live', Symbol='Commd9', Aliases='1810029F08Rik, AA409293', description='COMM domain containing 9', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=101886262, end_position_on_the_genomic_accession=101901639, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  76507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76507, CurrentID=0, Status='live', Symbol='Aoc1', Aliases='1600012D06Rik, Abp1, DAO', description='amine oxidase, copper-containing 1', other_designations='ABP|amiloride binding protein 1 (amine oxidase, copper-containing)|amiloride-binding protein 1|amiloride-sensitive amine oxidase|amine oxidase copper domain-containing protein 1|diamine oxidase|histaminase', map_location='6 23.78 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48895255, end_position_on_the_genomic_accession=48909188, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76509, CurrentID=0, Status='live', Symbol='Plet1', Aliases='0610037B23Rik, 1600029D21Rik, AA407923, AU043048', description='placenta expressed transcript 1', other_designations='antigen mAgK114', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50494525, end_position_on_the_genomic_accession=50505639, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76510, CurrentID=0, Status='live', Symbol='Trappc9', Aliases='1810044A24Rik, 2900005P22Rik, 4632408O18Rik, Ibp, Nibp, TRS130, mKIAA1882', description='trafficking protein particle complex 9', other_designations='Ikk2 binding protein|NIK and IKK(beta) binding protein|NIK- and IKBKB-binding protein', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=72589620, end_position_on_the_genomic_accession=73061243, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  207592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207592, CurrentID=0, Status='live', Symbol='Tbc1d16', Aliases='B930087K01, BC026530, Tcd1d16', description='TBC1 domain family, member 16', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119143043, end_position_on_the_genomic_accession=119228499, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  76522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76522, CurrentID=0, Status='live', Symbol='Lsm8', Aliases='2010003I05Rik, AW214405, Naa38', description='LSM8 homolog, U6 small nuclear RNA associated', other_designations='U6 snRNA-associated Sm-like protein LSm8', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=18848635, end_position_on_the_genomic_accession=18854052, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  207596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207596, CurrentID=0, Status='live', Symbol='Thsd4', Aliases='ADAMTSL-6, ADAMTSL6, AI585901, Adamtsl4, B230114P05Rik', description='thrombospondin, type I, domain containing 4', other_designations='A disintegrin and metalloproteinase with thrombospondin motifs-like protein 6|ADAMTS-like protein 4|ADAMTS-like protein 6|ADAMTSL-4', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59966931, end_position_on_the_genomic_accession=60522171, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  76527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76527, CurrentID=0, Status='live', Symbol='Il34', Aliases='2010004A03Rik, AI593503', description='interleukin 34', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110741829, end_position_on_the_genomic_accession=110805950, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  207607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207607, CurrentID=0, Status='live', Symbol='Ccdc40', Aliases='B930008I02Rik', description='coiled-coil domain containing 40', other_designations='protein links', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119228572, end_position_on_the_genomic_accession=119265248, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  76539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76539, CurrentID=0, Status='live', Symbol='Fam204a', Aliases='2310065H12Rik, 2610015K05Rik, D19Ertd737e', description='family with sequence similarity 204, member A', other_designations='', map_location='19 56.52 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60198586, end_position_on_the_genomic_accession=60226726, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  207615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207615, CurrentID=0, Status='live', Symbol='Wdr37', Aliases='', description='WD repeat domain 37', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=8802966, end_position_on_the_genomic_accession=8871904, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  76547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76547, CurrentID=0, Status='live', Symbol='Tmem101', Aliases='2610511E22Rik', description='transmembrane protein 101', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102152547, end_position_on_the_genomic_accession=102156407, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76551, CurrentID=0, Status='live', Symbol='Ccdc6', Aliases='2810012H18Rik, AA536681, AA960498, AW061011', description='coiled-coil domain containing 6', other_designations='', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=70096876, end_position_on_the_genomic_accession=70193200, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76559, CurrentID=0, Status='live', Symbol='Atg2b', Aliases='2410024A21Rik, AI047755, AI503411, AW558123, C030004M05Rik, C630028L02Rik, mKIAA4067', description='autophagy related 2B', other_designations='ATG2 autophagy related 2 homolog B', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105613536, end_position_on_the_genomic_accession=105685241, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  76560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76560, CurrentID=0, Status='live', Symbol='Prss8', Aliases='2410039E18Rik, AI313909, C79772, CAP1, fr, mCAP1', description='protease, serine 8 (prostasin)', other_designations='channel activating protease 1|channel-activating protease 1|protease, serine, 8 (prostasin)|serine protease 8', map_location='7 69.84 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127925717, end_position_on_the_genomic_accession=127930142, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  76561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76561, CurrentID=0, Status='live', Symbol='Snx7', Aliases='2510028H01Rik', description='sorting nexin 7', other_designations='', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=117781497, end_position_on_the_genomic_accession=117868936, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  76563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76563, CurrentID=0, Status='live', Symbol='Qrsl1', Aliases='2700038P16Rik, C80053, GatA', description='glutaminyl-tRNA synthase (glutamine-hydrolyzing)-like 1', other_designations='glu-AdT subunit A|glutaminyl-tRNA synthase-like protein 1|glutamyl-tRNA(Gln) amidotransferase subunit A homolog', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43874188, end_position_on_the_genomic_accession=43901739, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76566, CurrentID=0, Status='live', Symbol='Fam101b', Aliases='1500005K14Rik, RefilinB, cfm', description='family with sequence similarity 101, member B', other_designations='protein FAM101B|regulator of filamin protein B', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76019194, end_position_on_the_genomic_accession=76027782, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76568, CurrentID=0, Status='live', Symbol='Ift46', Aliases='1500035H01Rik, AA408110, AI663846, R75047, c11orf2', description='intraflagellar transport 46', other_designations='intraflagellar transport 46 homolog', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44772938, end_position_on_the_genomic_accession=44792714, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  76571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76571, CurrentID=0, Status='live', Symbol='Styxl1', Aliases='1700011C14Rik, Dusp24, Mkstyx', description='serine/threonine/tyrosine interacting-like 1', other_designations='dual specificity phosphatase 24 (putative)|map kinase phosphatase-like protein MK-STYX', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135747220, end_position_on_the_genomic_accession=135778410, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76572, CurrentID=0, Status='live', Symbol='Rbmxl2', Aliases='1700012H05Rik, Hnrnpg-t', description='RNA binding motif protein, X-linked-like 2', other_designations='testes-specific heterogenous nuclear ribonucleoprotein G-T', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107209445, end_position_on_the_genomic_accession=107210916, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  76573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76573, CurrentID=0, Status='live', Symbol='Spatc1l', Aliases='1700022B01Rik, 1700027D21Rik, AA062323', description='spermatogenesis and centriole associated 1 like', other_designations='spermatogenesis and centriole-associated protein 1-like protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76557552, end_position_on_the_genomic_accession=76570201, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  76574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76574, CurrentID=0, Status='live', Symbol='Mfsd2a', Aliases='1700018O18Rik, Mfsd2, NLS1', description='major facilitator superfamily domain containing 2A', other_designations='major facilitator superfamily domain containing 2|major facilitator superfamily domain-containing protein 2A|sodium-dependent LPC symporter 1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=122946850, end_position_on_the_genomic_accession=122961195, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  76577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76577, CurrentID=0, Status='live', Symbol='Faf2', Aliases='2210404D11Rik, AI462440, Ubxd8, mKIAA0887', description='Fas associated factor family member 2', other_designations='UBX domain containing 8|UBX domain-containing protein 8|protein expressed in T-cells and eosinophils in atopic dermatitis', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54621775, end_position_on_the_genomic_accession=54664063, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76580, CurrentID=0, Status='live', Symbol='Mib2', Aliases='2210008I11Rik, Zzank1, skd', description='mindbomb E3 ubiquitin protein ligase 2', other_designations='dyslike|dystrophin-like protein|mind bomb-2|mindbomb homolog 2|skeletrophin|zinc finger, ZZ type with ankyrin repeat domain 1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155654470, end_position_on_the_genomic_accession=155669254, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  76582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76582, CurrentID=0, Status='live', Symbol='Ipo11', Aliases='1700081H05Rik, 2510001A17Rik, AI314624, AW555235, E330021B14Rik, Ranbp11', description='importin 11', other_designations='imp11|ran-binding protein 11', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=106794439, end_position_on_the_genomic_accession=106936959, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  76589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76589, CurrentID=0, Status='live', Symbol='Unc5cl', Aliases='2510009H09Rik, BB155270, Zud', description='unc-5 homolog C (C. elegans)-like', other_designations='ZU5 and death domain-containing inhibitor of NF-kB|ZU5 and death domain-containing protein|protein unc-5 homolog C-like', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48453839, end_position_on_the_genomic_accession=48539709, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  76594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76594, CurrentID=0, Status='live', Symbol='Dnajc18', Aliases='2700075B01Rik, AU041129', description='DnaJ heat shock protein family (Hsp40) member C18', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 18', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35671103, end_position_on_the_genomic_accession=35708421, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  207667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207667, CurrentID=0, Status='live', Symbol='Skor1', Aliases='AV273001, C230094B15Rik, Corl1, Lbxcor1', description='SKI family transcriptional corepressor 1', other_designations='Lbx1 corepressor 1|fussel-15 homolog|ladybird homeobox 1 homolog corepressor 1|ladybird homeobox corepressor 1', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=63138164, end_position_on_the_genomic_accession=63148961, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76606, CurrentID=0, Status='live', Symbol='1700034O15Rik', Aliases='', description='RIKEN cDNA 1700034O15 gene', other_designations='', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41684346, end_position_on_the_genomic_accession=41685717, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  76608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76608, CurrentID=0, Status='live', Symbol='Hectd3', Aliases='1700064K09Rik, AI467540, AW743312', description='HECT domain containing 3', other_designations='HECT domain-containing protein 3|probable E3 ubiquitin-protein ligase HECTD3', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116995295, end_position_on_the_genomic_accession=117005277, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  207683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207683, CurrentID=0, Status='live', Symbol='Igsf11', Aliases='1700025L02Rik, BT-IgSF, Igsf13', description='immunoglobulin superfamily, member 11', other_designations='brain and testis-specific immunoglobulin superfamily protein|immunoglobulin superfamily, member 13', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38892476, end_position_on_the_genomic_accession=39027159, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  76612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76612, CurrentID=0, Status='live', Symbol='Lrrc27', Aliases='1700071K18Rik, 2310044E02Rik', description='leucine rich repeat containing 27', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139212869, end_position_on_the_genomic_accession=139242979, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  76614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76614, CurrentID=0, Status='live', Symbol='Immt', Aliases='1700082C19Rik, D830041H16Rik, HMP, P87, P87/89, P89', description='inner membrane protein, mitochondrial', other_designations='mitochondrial inner membrane protein|mitofilin', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71831061, end_position_on_the_genomic_accession=71877381, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  76615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76615, CurrentID=0, Status='live', Symbol='Got1l1', Aliases='1700083M11Rik', description='glutamic-oxaloacetic transaminase 1-like 1', other_designations='glutamate oxaloacetate transaminase 1-like protein 1|transaminase A-like protein 1', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27197459, end_position_on_the_genomic_accession=27223848, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78306, CurrentID=0, Status='live', Symbol='Tsga10ip', Aliases='1810013N20Rik, Fam161c', description='testis specific 10 interacting protein', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5390047, end_position_on_the_genomic_accession=5402534, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76626, CurrentID=0, Status='live', Symbol='Msi2', Aliases='1700105C15Rik, AI451722, AI563628, AW489193h, Msi2', description='musashi RNA-binding protein 2', other_designations='Musashi2|musashi-2', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88339382, end_position_on_the_genomic_accession=88718267, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  76627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76627, CurrentID=0, Status='live', Symbol='Prr30', Aliases='1700110M21Rik', description='proline rich 30', other_designations='', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=101197690, end_position_on_the_genomic_accession=101200404, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76629, CurrentID=0, Status='live', Symbol='Wbscr28', Aliases='1700111I05Rik', description='Williams-Beuren syndrome chromosome region 28 (human)', other_designations='Williams-Beuren syndrome critical region 28', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134901593, end_position_on_the_genomic_accession=134906733, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76630, CurrentID=0, Status='live', Symbol='Stambpl1', Aliases='1700095N21Rik, 8230401J17Rik, ALMalpha, AMSH-FP', description='STAM binding protein like 1', other_designations='AMSH family protein|AMSH-LP|AMSH-family protein|STAM-binding protein-like 1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34192190, end_position_on_the_genomic_accession=34240657, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  207704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207704, CurrentID=0, Status='live', Symbol='Gtpbp10', Aliases='4930545J22Rik, BC034507', description='GTP-binding protein 10 (putative)', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=5537450, end_position_on_the_genomic_accession=5559774, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76633, CurrentID=0, Status='live', Symbol='1700112E06Rik', Aliases='2700009F18Rik', description='RIKEN cDNA 1700112E06 gene', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=22019503, end_position_on_the_genomic_accession=23056088, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76637, CurrentID=0, Status='live', Symbol='Zim2', Aliases='1700116N21Rik', description='zinc finger, imprinted 2', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6651130, end_position_on_the_genomic_accession=6658385, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76640, CurrentID=0, Status='live', Symbol='1700113H08Rik', Aliases='', description='RIKEN cDNA 1700113H08 gene', other_designations='', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=87040467, end_position_on_the_genomic_accession=87230599, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76645, CurrentID=0, Status='live', Symbol='Pkd1l2', Aliases='1700126L06Rik', description='polycystic kidney disease 1 like 2', other_designations='PC1-like 2 protein|polycystic kidney disease 1-like 2|polycystin-1L2', map_location='8 C5-D2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116995218, end_position_on_the_genomic_accession=117082488, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  76646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76646, CurrentID=0, Status='live', Symbol='Wdr38', Aliases='1700123D08Rik', description='WD repeat domain 38', other_designations='novel WD domain, G-beta repeat domain containing protein (1700123D08Rik)', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38998309, end_position_on_the_genomic_accession=39001588, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76650, CurrentID=0, Status='live', Symbol='Srxn1', Aliases='1700127B04Rik, AI854065, AW488194, Npn3, Srx, Srx1, TX01', description='sulfiredoxin 1 homolog (S. cerevisiae)', other_designations='neoplastic progression 3|neoplastic progression protein 3', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152105524, end_position_on_the_genomic_accession=152111376, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  76652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76652, CurrentID=0, Status='live', Symbol='Actrt3', Aliases='1700119I24Rik, ARP-T3, Arpm1', description='actin related protein T3', other_designations='actin related protein M1|actin-related protein M1 homolog', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30597073, end_position_on_the_genomic_accession=30599870, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  76653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76653, CurrentID=0, Status='live', Symbol='Cby3', Aliases='1700121K02Rik', description='chibby homolog 3 (Drosophila)', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50357715, end_position_on_the_genomic_accession=50360017, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  76654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76654, CurrentID=0, Status='live', Symbol='Upp2', Aliases='1700124F02Rik, AI266885, UDRPASE2, UP2, UPASE2', description='uridine phosphorylase 2', other_designations='L-UrdPase|liver-specific uridine phosphorylase', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=58567225, end_position_on_the_genomic_accession=58793042, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  207728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207728, CurrentID=0, Status='live', Symbol='Pde2a', Aliases='CGS-PDE, cGSPDE', description='phosphodiesterase 2A, cGMP-stimulated', other_designations='cyclic GMP-stimulated phosphodiesterase', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101421691, end_position_on_the_genomic_accession=101512829, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  76658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76658, CurrentID=0, Status='live', Symbol='1700123K08Rik', Aliases='', description='RIKEN cDNA 1700123K08 gene', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138559773, end_position_on_the_genomic_accession=138565561, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  76668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76668, CurrentID=0, Status='live', Symbol='Mdh1b', Aliases='1700124B08Rik, AV255588', description='malate dehydrogenase 1B, NAD (soluble)', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63687573, end_position_on_the_genomic_accession=63730547, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  76670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76670, CurrentID=0, Status='live', Symbol='Cfap70', Aliases='Ttc18', description='cilia and flagella associated protein 70', other_designations='tetratricopeptide repeat domain 18', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20394053, end_position_on_the_genomic_accession=20452264, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  76681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76681, CurrentID=0, Status='live', Symbol='Trim12a', Aliases='2310043C01Rik, Trim12', description='tripartite motif-containing 12A', other_designations='tripartite motif protein 12|tripartite motif-containing 12|tripartite motif-containing protein 12', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104299893, end_position_on_the_genomic_accession=104315849, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76686, CurrentID=0, Status='live', Symbol='Clip3', Aliases='1500005P14Rik, AI844915, Clipr-59', description='CAP-GLY domain containing linker protein 3', other_designations='CLIP-170-related 59 kDa protein|clip-170 related protein 59 kDa|cytoplasmic linker protein 170-related 59 kDa protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30291688, end_position_on_the_genomic_accession=30308367, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  76687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76687, CurrentID=0, Status='live', Symbol='Spcs3', Aliases='1810011E08Rik', description='signal peptidase complex subunit 3 homolog (S. cerevisiae)', other_designations='', map_location='8 B1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=54520433, end_position_on_the_genomic_accession=54530739, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76688, CurrentID=0, Status='live', Symbol='Arfrp1', Aliases='1500006I01Rik, AI480700', description='ADP-ribosylation factor related protein 1', other_designations='ARF-related protein 1', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181357690, end_position_on_the_genomic_accession=181365404, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  207686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207686, CurrentID=0, Status='live', Symbol='Cfap69', Aliases='4921525K03, A330021E22Rik, AI427898', description='cilia and flagella associated protein 69', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=5579278, end_position_on_the_genomic_accession=5664256, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  76701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76701, CurrentID=0, Status='live', Symbol='Ctrc', Aliases='1810044E12Rik', description='chymotrypsin C (caldecrin)', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141838239, end_position_on_the_genomic_accession=141846359, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76703, CurrentID=0, Status='live', Symbol='Cpb1', Aliases='0910001A18Rik, 1810063F02Rik, 2210008M23Rik, AI504870', description='carboxypeptidase B1 (tissue)', other_designations='', map_location='3 6.29 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=20249484, end_position_on_the_genomic_accession=20275729, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  207777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207777, CurrentID=0, Status='live', Symbol='Bzrap1', Aliases='D230016K05, PRAX-1, mKIAA0612', description='benzodiazepine receptor associated protein 1', other_designations='PBR-IP|RIM-BP1|RIMS-binding protein 1|benzodiazapine receptor associated protein 1|peripheral benzodiazepine receptor-associated protein 1|peripheral benzodiazepine receptor-interacting protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87759066, end_position_on_the_genomic_accession=87785928, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  76707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76707, CurrentID=0, Status='live', Symbol='Clasp1', Aliases='1700030C23Rik, 5730583A19Rik, B130045P17Rik, CLASP1alpha', description='CLIP associating protein 1', other_designations='cytoplasmic linker-associated protein 1', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=118389058, end_position_on_the_genomic_accession=118609462, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  207781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207781, CurrentID=0, Status='live', Symbol='C2cd2', Aliases='5730563M15, 5830404H04Rik, AI314146, AI449533, BF168443, ORF25, mKIAA4045', description='C2 calcium-dependent domain containing 2', other_designations='', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97855209, end_position_on_the_genomic_accession=97922651, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  76713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76713, CurrentID=0, Status='live', Symbol='1700039E15Rik', Aliases='Gm19188', description='RIKEN cDNA 1700039E15 gene', other_designations='solute carrier family 6 (neurotransmitter transporter), member 20B pseudogene', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45277331, end_position_on_the_genomic_accession=45288996, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76718, CurrentID=0, Status='live', Symbol='Catsperg2', Aliases='1700067C01Rik, CATSPERG', description='catsper channel auxiliary subunit gamma 2', other_designations='CATSPER gamma subunit protein|cation channel, sperm-associated, gamma 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29697217, end_position_on_the_genomic_accession=29727048, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  76719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76719, CurrentID=0, Status='live', Symbol='Kansl1', Aliases='1700081L11Rik, 9430041J06Rik, mKIAA1267', description='KAT8 regulatory NSL complex subunit 1', other_designations='NSL complex protein NSL1|non-specific lethal 1 homolog', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104333229, end_position_on_the_genomic_accession=104469507, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  207792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207792, CurrentID=0, Status='live', Symbol='BC034090', Aliases='6820428L09, mKIAA1614', description='cDNA sequence BC034090', other_designations='', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155212471, end_position_on_the_genomic_accession=155246220, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  76722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76722, CurrentID=0, Status='live', Symbol='Ckmt2', Aliases='2300008A19Rik, ScCKmit', description='creatine kinase, mitochondrial 2', other_designations='S-MtCK|basic-type mitochondrial creatine kinase|mib-CK|sarcomeric mitochondrial creatine kinase', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=91853387, end_position_on_the_genomic_accession=91876923, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  207798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207798, CurrentID=0, Status='live', Symbol='Gramd1c', Aliases='4921521N14Rik', description='GRAM domain containing 1C', other_designations='', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=43980350, end_position_on_the_genomic_accession=44065372, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  207806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207806, CurrentID=0, Status='live', Symbol='Usf3', Aliases='5530400K22Rik, Gm1779, Gm608, mKIAA2018', description='upstream transcription factor family member 3', other_designations='', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44173236, end_position_on_the_genomic_accession=44227465, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76737, CurrentID=0, Status='live', Symbol='Creld2', Aliases='5730592L21Rik, C85758', description='cysteine-rich with EGF-like domains 2', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88819646, end_position_on_the_genomic_accession=88826681, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  76740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76740, CurrentID=0, Status='live', Symbol='Efr3a', Aliases='A130089M23Rik, BB071175, C76891, C920006C10Rik, D030063F01Rik, mKIAA0143', description='EFR3 homolog A', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=65787020, end_position_on_the_genomic_accession=65873812, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  76742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76742, CurrentID=0, Status='live', Symbol='Snx27', Aliases='5730552M22Rik, ESTM45, ESTM47, R75405', description='sorting nexin family member 27', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94497542, end_position_on_the_genomic_accession=94582716, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  76743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76743, CurrentID=0, Status='live', Symbol='Gje1', Aliases='AEY12, Cx23, D230044M03Rik, Gjf1, Gsfaey12', description='gap junction protein, epsilon 1', other_designations='connexin 23|connexin-23|putative gap junction protein connexin Cx43.4', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=14715626, end_position_on_the_genomic_accession=14718214, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  207818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207818, CurrentID=0, Status='live', Symbol='Smagp', Aliases='', description='small cell adhesion glycoprotein', other_designations='small cell transmembrane and glycosylated protein', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100621339, end_position_on_the_genomic_accession=100636940, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76747, CurrentID=0, Status='live', Symbol='Dapl1', Aliases='2310032F03Rik, EEDA', description='death associated protein-like 1', other_designations='early epithelial differentiation-associated protein', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=59484653, end_position_on_the_genomic_accession=59505020, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76757, CurrentID=0, Status='live', Symbol='Trdn', Aliases='2310045H21Rik, EG432451, TDN', description='triadin', other_designations='cardiac triadin', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=33083483, end_position_on_the_genomic_accession=33476709, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  76758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76758, CurrentID=0, Status='live', Symbol='Gsdma2', Aliases='2200001G21Rik, 2210006M16Rik, 2210009F20Rik, 2210411P14Rik, Gsdm2, Gsdml2', description='gasdermin A2', other_designations='gasdermin 2|gasdermin-2|gsdma2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98646734, end_position_on_the_genomic_accession=98657957, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76763, CurrentID=0, Status='live', Symbol='Mospd2', Aliases='2410013I23Rik', description='motile sperm domain containing 2', other_designations='', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164936169, end_position_on_the_genomic_accession=164980375, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  207839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207839, CurrentID=0, Status='live', Symbol='Galnt6', Aliases='4632410F13, AW047994, GalNAc-T6', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 6', other_designations='N-acetylgalactosaminyltransferase T-6|Protein-UDP acetylgalactosaminyltransferase 6|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 6|polypeptide GalNAc transferase 6|pp-GaNTase 6', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100689915, end_position_on_the_genomic_accession=100729929, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  76768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76768, CurrentID=0, Status='live', Symbol='Alpi', Aliases='2010001C14Rik, Akp6, gIAP', description='alkaline phosphatase, intestinal', other_designations='alkaline phosphatase 6', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87098002, end_position_on_the_genomic_accession=87101606, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76770, CurrentID=0, Status='live', Symbol='2010005H15Rik', Aliases='', description='RIKEN cDNA 2010005H15 gene', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36221562, end_position_on_the_genomic_accession=36257451, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  76773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76773, CurrentID=0, Status='live', Symbol='Wdyhv1', Aliases='2410187C16Rik, AU014961, AW550036, Ntaq1', description='WDYHV motif containing 1', other_designations='N-terminal Gln amidase|WDYHV motif-containing protein 1|nt(Q)-amidase|protein NH2-terminal glutamine deamidase', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58141334, end_position_on_the_genomic_accession=58158654, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76775, CurrentID=0, Status='live', Symbol='Slc10a7', Aliases='2410193C02Rik, P7', description='solute carrier family 10 (sodium/bile acid cotransporter family), member 7', other_designations='Na(+)/bile acid cotransporter 7', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=78509328, end_position_on_the_genomic_accession=78734012, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  76779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76779, CurrentID=0, Status='live', Symbol='Cluap1', Aliases='2310030D15Rik, 2610111M03Rik', description='clusterin associated protein 1', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3907659, end_position_on_the_genomic_accession=3941147, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  76781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76781, CurrentID=0, Status='live', Symbol='Mettl4', Aliases='2410198H06Rik, A730091E08Rik, AV296509, HsT661', description='methyltransferase like 4', other_designations='', map_location='17 E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=94722658, end_position_on_the_genomic_accession=94750126, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  207854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207854, CurrentID=0, Status='live', Symbol='Fmr1nb', Aliases='3830422N12Rik, NY-SAR-35, NYSAR35', description='fragile X mental retardation 1 neighbor', other_designations='protein mNY-SAR-35', map_location='X A7.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=68761839, end_position_on_the_genomic_accession=68804561, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  76784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76784, CurrentID=0, Status='live', Symbol='Mtif2', Aliases='2310038D14Rik, 2410112O06Rik, IF-2(Mt), IF-2mt, IF2(mt)', description='mitochondrial translational initiation factor 2', other_designations='', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29526379, end_position_on_the_genomic_accession=29545452, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  76787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76787, CurrentID=0, Status='live', Symbol='Ppfia3', Aliases='2410127E16Rik', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 3', other_designations='Liprin-alpha3|PTPRF-interacting protein alpha-3|protein tyrosine phosphatase receptor type f polypeptide-interacting protein alpha-3', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45339126, end_position_on_the_genomic_accession=45367019, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  76788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76788, CurrentID=0, Status='live', Symbol='Klhdc10', Aliases='2410127E18Rik, AI593221, mKIAA0265', description='kelch domain containing 10', other_designations='', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30401517, end_position_on_the_genomic_accession=30455179, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  76789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76789, CurrentID=0, Status='live', Symbol='Mzt1', Aliases='2410129H14Rik, AI120158, Mozart1', description='mitotic spindle organizing protein 1', other_designations='mitotic-spindle organizing protein associated with a ring of gamma-tubulin 1', map_location='14 E2.2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=99034544, end_position_on_the_genomic_accession=99046463, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76792, CurrentID=0, Status='live', Symbol='2410131K14Rik', Aliases='', description='RIKEN cDNA 2410131K14 gene', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=118245227, end_position_on_the_genomic_accession=118263114, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  76793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76793, CurrentID=0, Status='live', Symbol='Snip1', Aliases='2410133M08Rik', description='Smad nuclear interacting protein 1', other_designations='Smad nuclear interacting protein 1 Snip1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=125066679, end_position_on_the_genomic_accession=125074054, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76795, CurrentID=0, Status='live', Symbol='Tbc1d9b', Aliases='', description='TBC1 domain family, member 9B', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50131057, end_position_on_the_genomic_accession=50172785, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  76799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76799, CurrentID=0, Status='live', Symbol='Tmem234', Aliases='1500002D11Rik, 2510006D16Rik, 4933407D05Rik', description='transmembrane protein 234', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129600691, end_position_on_the_genomic_accession=129608528, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76800, CurrentID=0, Status='live', Symbol='Usp42', Aliases='2410140K03Rik, 3110031A07Rik, A630018G05Rik, D5Ertd591e', description='ubiquitin specific peptidase 42', other_designations='deubiquitinating enzyme 42|ubiquitin specific protease 42|ubiquitin thioesterase 42|ubiquitin thiolesterase 42|ubiquitin-specific-processing protease 42', map_location='5 82.5 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143710325, end_position_on_the_genomic_accession=143759808, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  76803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76803, CurrentID=0, Status='live', Symbol='2410141K09Rik', Aliases='4930482J15Rik, Ayu21-35, Gt(pU21)35Imeg', description='RIKEN cDNA 2410141K09 gene', other_designations='Gt(Ayu21)35Imeg', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=66418114, end_position_on_the_genomic_accession=66441118, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76804, CurrentID=0, Status='live', Symbol='Kdm4c', Aliases='2410141F18Rik, AA517467, Gasc1, Jmjd2c', description='lysine (K)-specific demethylase 4C', other_designations='gene amplified in squamous cell carcinoma 1|jmjC domain-containing histone demethylation protein 3C|jumonji domain containing 2C|jumonji domain-containing protein 2C', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=74242497, end_position_on_the_genomic_accession=74405864, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  76808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76808, CurrentID=0, Status='live', Symbol='Rpl18a', Aliases='2510019J09Rik', description='ribosomal protein L18A', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70894722, end_position_on_the_genomic_accession=70897443, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76809, CurrentID=0, Status='live', Symbol='Bri3bp', Aliases='2410150I18Rik, AI841257, AW742481', description='Bri3 binding protein', other_designations='I3-binding protein|bri3bp', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=125441538, end_position_on_the_genomic_accession=125460885, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76813, CurrentID=0, Status='live', Symbol='Armc6', Aliases='2410153K17Rik, AW554412', description='armadillo repeat containing 6', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70220172, end_position_on_the_genomic_accession=70234422, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76815, CurrentID=0, Status='live', Symbol='Calcoco2', Aliases='2410154J16Rik, C77254, Ndp52, Ndp52l1', description='calcium binding and coiled-coil domain 2', other_designations='nuclear domain 10 protein 52 like 1|nuclear domain 10 protein NDP52', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96098916, end_position_on_the_genomic_accession=96124367, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  76816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76816, CurrentID=0, Status='live', Symbol='Sdccag8', Aliases='2700048G21Rik, 5730470G24Rik, Cccap', description='serologically defined colon cancer antigen 8', other_designations='centrosomal colon cancer autoantigen protein', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=176814526, end_position_on_the_genomic_accession=177020437, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  76820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76820, CurrentID=0, Status='live', Symbol='Fam49a', Aliases='2410157M17Rik, 9630047E15, D12Ertd553e', description='family with sequence similarity 49, member A', other_designations='family with sequence similarity 49, member B', map_location='12 5.63 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=12262129, end_position_on_the_genomic_accession=12379058, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  76824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76824, CurrentID=0, Status='live', Symbol='Mtfr1l', Aliases='2410166I05Rik, AI480750, Fam54b', description='mitochondrial fission regulator 1-like', other_designations='family with sequence similarity 54, member B|protein FAM54B', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134525555, end_position_on_the_genomic_accession=134535387, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76826, CurrentID=0, Status='live', Symbol='Nubpl', Aliases='2410170E07Rik', description='nucleotide binding protein-like', other_designations='nucleotide-binding protein-like', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=52097746, end_position_on_the_genomic_accession=52310970, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76829, CurrentID=0, Status='live', Symbol='Dok5', Aliases='2700055C10Rik', description='docking protein 5', other_designations='downstream of tyrosine kinase 5', map_location='2 92.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=170731807, end_position_on_the_genomic_accession=170879775, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  76832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76832, CurrentID=0, Status='live', Symbol='Hyls1', Aliases='3010015K02Rik', description='hydrolethalus syndrome 1', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35560821, end_position_on_the_genomic_accession=35570069, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  207911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207911, CurrentID=0, Status='live', Symbol='Mchr1', Aliases='AW049955, Gpr24, Mch1r', description='melanin-concentrating hormone receptor 1', other_designations='G protein coupled receptor 24|G-protein coupled receptor 24|Gpr24-9 protein coupled receptor|MCH receptor 1|MCHR|SLC-1|somatostatin receptor-like protein', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81235499, end_position_on_the_genomic_accession=81238964, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  76843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76843, CurrentID=0, Status='live', Symbol='Dtl', Aliases='2810047L02Rik, 5730564G15Rik, Ramp', description='denticleless E3 ubiquitin protein ligase', other_designations='Meth A ramp|denticleless homolog|lethal(2) denticleless protein homolog|meth A retinoic acid-regulated nuclear matrix-associated protein', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191537365, end_position_on_the_genomic_accession=191575534, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  76846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76846, CurrentID=0, Status='live', Symbol='Rps9', Aliases='3010033P07Rik, AL022771, AL022885', description='ribosomal protein S9', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3703953, end_position_on_the_genomic_accession=3706899, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  207920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207920, CurrentID=0, Status='live', Symbol='Esrp1', Aliases='2210008M09Rik, A630065D16, BC031468, Rbm35a', description='epithelial splicing regulatory protein 1', other_designations='RNA binding motif protein 35A', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11331933, end_position_on_the_genomic_accession=11386783, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  207921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207921, CurrentID=0, Status='live', Symbol='Fam228b', Aliases='A830093I24Rik', description='family with sequence similarity 228, member B', other_designations='UPF0638 protein B', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4738397, end_position_on_the_genomic_accession=4769267, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  76850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76850, CurrentID=0, Status='live', Symbol='Ago4', Aliases='5730550L01Rik, AI481660, Eif2c4', description='argonaute RISC catalytic subunit 4', other_designations='Piwi/Argonaute family protein meIF2C4|argonaute 4|argonaute RISC catalytic component 4|eukaryotic translation initiation factor 2C, 4|mAgo4', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126489487, end_position_on_the_genomic_accession=126533588, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  76854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76854, CurrentID=0, Status='live', Symbol='Gper1', Aliases='6330420K13Rik, CMKRL2, Ceprl, FEG-1, GPCR-Br, Gper, Gpr30', description='G protein-coupled estrogen receptor 1', other_designations='G protein-coupled receptor 30|G-protein coupled receptor 30|chemoattractant receptor-like 2|constitutively expressed peptide-like receptor like|mER|membrane estrogen receptor', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139422972, end_position_on_the_genomic_accession=139427800, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  76856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76856, CurrentID=0, Status='live', Symbol='Catsper3', Aliases='4921522D01Rik, CACRC', description='cation channel, sperm associated 3', other_designations='', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55784579, end_position_on_the_genomic_accession=55824942, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76857, CurrentID=0, Status='live', Symbol='Spopl', Aliases='4921517N04Rik, AU014935, BB233739, E430033K04Rik', description='speckle-type POZ protein-like', other_designations='HIB homolog 2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=23506220, end_position_on_the_genomic_accession=23572128, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  76858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76858, CurrentID=0, Status='live', Symbol='Nlrp14', Aliases='4921520L01Rik, GC-LRR, Nalp-iota, Nalp14, Nalp14l', description='NLR family, pyrin domain containing 14', other_designations='NACHT, leucine rich repeat and PYD containing 14|RNI-like protein|germ cell specific leucine-rich repeat NTPase|germ-cell specific leucine rich repeat NTPase', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107127432, end_position_on_the_genomic_accession=107198103, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  207932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207932, CurrentID=0, Status='live', Symbol='Urb1', Aliases='4921511H13Rik, 5730405K23Rik, AA670781, AI645590, mKIAA0539', description='URB1 ribosome biogenesis 1 homolog (S. cerevisiae)', other_designations='nucleolar preribosomal-associated protein 1', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90751527, end_position_on_the_genomic_accession=90810465, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  76863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76863, CurrentID=0, Status='live', Symbol='Dcun1d5', Aliases='3110001A18Rik, 4833420K19Rik, AW060460, D430047L21Rik', description='DCN1, defective in cullin neddylation 1, domain containing 5 (S. cerevisiae)', other_designations='DCUN1 domain-containing protein 5|defective in cullin neddylation protein 1-like protein 5', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7183021, end_position_on_the_genomic_accession=7207042, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76866, CurrentID=0, Status='live', Symbol='Morn1', Aliases='2900057D20Rik, 4930417P05Rik, AU042805', description='MORN repeat containing 1', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155086240, end_position_on_the_genomic_accession=155145507, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  76867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76867, CurrentID=0, Status='live', Symbol='Rhbdd1', Aliases='4930418P06Rik, RRP4, mRHBDD1', description='rhomboid domain containing 1', other_designations='rhomboid domain-containing protein 1|rhomboid-like protein 4', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=82316457, end_position_on_the_genomic_accession=82445366, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76872, CurrentID=0, Status='live', Symbol='Ccdc116', Aliases='4930432J16Rik', description='coiled-coil domain containing 116', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17139064, end_position_on_the_genomic_accession=17147146, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  76877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76877, CurrentID=0, Status='live', Symbol='Rab36', Aliases='6330593L16Rik, AW047545', description='RAB36, member RAS oncogene family', other_designations='', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75036895, end_position_on_the_genomic_accession=75054748, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  12813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12813, CurrentID=0, Status='live', Symbol='Col10a1', Aliases='Col10, Col10a-1', description='collagen, type X, alpha 1', other_designations='procollagen, type X, alpha 1', map_location='10 18.85 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34389981, end_position_on_the_genomic_accession=34397085, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  207952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207952, CurrentID=0, Status='live', Symbol='Klhl25', Aliases='2810402K13Rik, BC027373, Enc-1', description='kelch-like 25', other_designations='ENC-2|ectoderm-neural cortex protein 2|ectoderm-neural cortex-1', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=75848338, end_position_on_the_genomic_accession=75874136, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  76884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76884, CurrentID=0, Status='live', Symbol='Cyfip2', Aliases='1500004I01Rik, 6430511D02Rik, AA930218, AU022376, Pir121, mKIAA1168', description='cytoplasmic FMR1 interacting protein 2', other_designations='cyfip2|p53-inducible protein 121', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46193849, end_position_on_the_genomic_accession=46312859, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  207958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207958, CurrentID=0, Status='live', Symbol='Alg11', Aliases='AI849156, AW492253, B230397C21, Mmat-4(5)', description='asparagine-linked glycosylation 11 (alpha-1,2-mannosyltransferase)', other_designations='asparagine-linked glycosylation 11 homolog (yeast, alpha-1,2-mannosyltransferase)|asparagine-linked glycosylation protein 11 homolog|defective mannosyltransferase IV(V)|glycolipid 2-alpha-mannosyltransferase|mannosyltransferase IV(V)', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22060721, end_position_on_the_genomic_accession=22071627, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  76889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76889, CurrentID=0, Status='live', Symbol='Adck4', Aliases='0610012P18Rik', description='aarF domain containing kinase 4', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27228823, end_position_on_the_genomic_accession=27257949, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  76890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76890, CurrentID=0, Status='live', Symbol='Memo1', Aliases='0610016J10Rik, D930048L02Rik', description='mediator of cell motility 1', other_designations='mediator of ErbB2-driven cell motility 1|memo-1|protein memo', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74198398, end_position_on_the_genomic_accession=74295544, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  76892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76892, CurrentID=0, Status='live', Symbol='Rnft1', Aliases='0610013E23Rik, AV007605', description='ring finger protein, transmembrane 1', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86484657, end_position_on_the_genomic_accession=86499007, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76893, CurrentID=0, Status='live', Symbol='Cers2', Aliases='0610013I17Rik, AI225939, Lass2, TRH3', description='ceramide synthase 2', other_designations='LAG1 homolog, ceramide synthase 2|LAG1 longevity assurance homolog 2|TRAM homolog 3|longevity assurance homolog 2|translocating chain-associating membrane protein homolog 3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95314805, end_position_on_the_genomic_accession=95323588, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  76894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76894, CurrentID=0, Status='live', Symbol='Mettl15', Aliases='0610027B03Rik, AI303697, AW049952, Mett5d1', description='methyltransferase like 15', other_designations='methyltransferase 5 domain containing 1|methyltransferase 5 domain-containing protein 1|probable S-adenosyl-L-methionine-dependent methyltransferase METT5D1|putative S-adenosyl-L-methionine-dependent methyltransferase METT5D1', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=109092297, end_position_on_the_genomic_accession=109280743, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  76895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76895, CurrentID=0, Status='live', Symbol='Bicd2', Aliases='', description='bicaudal D homolog 2 (Drosophila)', other_designations='bic-D 2', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49341549, end_position_on_the_genomic_accession=49387026, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76897, CurrentID=0, Status='live', Symbol='Ralyl', Aliases='0710005M24Rik', description='RALY RNA binding protein-like', other_designations='', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=13471655, end_position_on_the_genomic_accession=14182287, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  76898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76898, CurrentID=0, Status='live', Symbol='B3gat1', Aliases='0710007K08Rik, AI846286, GlcAT-P, Glcatp, Hnk1', description='beta-1,3-glucuronyltransferase 1 (glucuronosyltransferase P)', other_designations='UDP-GlcUA:glycoprotein beta-1,3-glucuronyltransferase|glcUAT-P|glucuronosyltransferase P|glucuronosyltransferase-P', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=26733783, end_position_on_the_genomic_accession=26761342, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  76899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76899, CurrentID=0, Status='live', Symbol='Golga1', Aliases='0710001G09Rik, 2210418B03Rik, AW107649, Golgi97, awag, golgin-97', description='golgi autoantigen, golgin subfamily a, 1', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=39016155, end_position_on_the_genomic_accession=39065579, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  76900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76900, CurrentID=0, Status='live', Symbol='Ssbp4', Aliases='1210002E11Rik, AW743380, Sspb4', description='single stranded DNA binding protein 4', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70597485, end_position_on_the_genomic_accession=70608380, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  76901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76901, CurrentID=0, Status='live', Symbol='Jade2', Aliases='1200017K05Rik, AI480685, Phf15, mKIAA0239', description='jade family PHD finger 2', other_designations='PHD finger protein 15|jade family PHD finger protein 2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51813451, end_position_on_the_genomic_accession=51859506, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  76905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76905, CurrentID=0, Status='live', Symbol='Lrg1', Aliases='1300008B03Rik, 2310031E04Rik, Lrg, Lrhg', description='leucine-rich alpha-2-glycoprotein 1', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56119678, end_position_on_the_genomic_accession=56121946, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  76915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76915, CurrentID=0, Status='live', Symbol='Mnd1', Aliases='2610034E18Rik, AI591601, Gaj', description='meiotic nuclear divisions 1 homolog (S. cerevisiae)', other_designations='mnd1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84087933, end_position_on_the_genomic_accession=84156861, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  76916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76916, CurrentID=0, Status='live', Symbol='Timmdc1', Aliases='2810021C21Rik, 4930455C21Rik, AV135763', description='translocase of inner mitochondrial membrane domain containing 1', other_designations='TIMM domain containing-protein 1|translocase of inner mitochondrial membrane domain-containing protein 1|transmembrane protein C3orf1 homolog', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38497843, end_position_on_the_genomic_accession=38522685, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  76917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76917, CurrentID=0, Status='live', Symbol='Flywch2', Aliases='2810417J12Rik, 3010014L17Rik, AW548203', description='FLYWCH family member 2', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23776918, end_position_on_the_genomic_accession=23786077, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76920, CurrentID=0, Status='live', Symbol='Arrdc5', Aliases='1700013E09Rik', description='arrestin domain containing 5', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56294111, end_position_on_the_genomic_accession=56300286, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76921, CurrentID=0, Status='live', Symbol='1700013D24Rik', Aliases='', description='RIKEN cDNA 1700013D24 gene', other_designations='', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124347593, end_position_on_the_genomic_accession=124357086, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  76925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76925, CurrentID=0, Status='live', Symbol='1700015E13Rik', Aliases='', description='RIKEN cDNA 1700015E13 gene', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170308861, end_position_on_the_genomic_accession=170312125, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  76927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76927, CurrentID=0, Status='live', Symbol='Tsacc', Aliases='1700021C14Rik, SIP', description='TSSK6 activating co-chaperone', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88282752, end_position_on_the_genomic_accession=88297009, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  76932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76932, CurrentID=0, Status='live', Symbol='Arfip2', Aliases='2310002N04Rik, POR, Por1', description='ADP-ribosylation factor interacting protein 2', other_designations='ADP-ribosylation factor-interacting protein 2|Arfaptin 2|partner of RAC1', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105634201, end_position_on_the_genomic_accession=105640416, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  76933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76933, CurrentID=0, Status='live', Symbol='Ifi27l2a', Aliases='2310061N23Rik, Ifi27, Isg12, Isg12(b1)', description='interferon, alpha-inducible protein 27 like 2A', other_designations='interferon stimulated gene 12(b1)|interferon-stimulated gene 12 protein', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103433873, end_position_on_the_genomic_accession=103443680, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  76936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76936, CurrentID=0, Status='live', Symbol='Hnrnpm', Aliases='2610023M21Rik, AA409009, Hnrpm, mKIAA4193', description='heterogeneous nuclear ribonucleoprotein M', other_designations='hnRNP M', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33646233, end_position_on_the_genomic_accession=33685458, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  76938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76938, CurrentID=0, Status='live', Symbol='Rbm17', Aliases='2700027J02Rik', description='RNA binding motif protein 17', other_designations='45 kDa-splicing factor|RNA-binding motif protein 17', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=11585439, end_position_on_the_genomic_accession=11603199, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  76942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76942, CurrentID=0, Status='live', Symbol='Lypd5', Aliases='2210003I03Rik', description='Ly6/Plaur domain containing 5', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24349224, end_position_on_the_genomic_accession=24353833, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  76943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76943, CurrentID=0, Status='live', Symbol='Psapl1', Aliases='2310020A21Rik', description='prosaposin-like 1', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36204021, end_position_on_the_genomic_accession=36206568, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  76947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76947, CurrentID=0, Status='live', Symbol='Ndufaf6', Aliases='2310030N02Rik', description='NADH dehydrogenase (ubiquinone) complex I, assembly factor 6', other_designations='UPF0551 protein C8orf38 homolog, mitochondrial', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11051042, end_position_on_the_genomic_accession=11076235, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  76952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76952, CurrentID=0, Status='live', Symbol='Nt5c2', Aliases='2010002I23Rik, CnII, ENSMUSG00000025039, Gm9751, Gmp, Nt5b, Pnt5, cN-II', description="5'-nucleotidase, cytosolic II", other_designations="cytosolic 5'-nucleotidase II", map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46885390, end_position_on_the_genomic_accession=47015189, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  76954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76954, CurrentID=0, Status='live', Symbol='St5', Aliases='2010004M01Rik, 2610305K15Rik', description='suppression of tumorigenicity 5', other_designations='DENN domain-containing protein 2B', map_location='7 57.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109523911, end_position_on_the_genomic_accession=109703621, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  76959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76959, CurrentID=0, Status='live', Symbol='Chmp5', Aliases='2210412K09Rik, AW545668', description='charged multivesicular body protein 5', other_designations='SNF7 domain-containing protein 2|chromatin modifying protein 5|chromatin-modifying protein 5|likely ortholog of H. sapiens SNF7 domain containing 2', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40948553, end_position_on_the_genomic_accession=40965302, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  76960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76960, CurrentID=0, Status='live', Symbol='Bcas1', Aliases='2210416M21Rik, 9030223A09Rik, AI841227, NABC1', description='breast carcinoma amplified sequence 1', other_designations='amplified in breast cancer 1|novel amplified in breast cancer 1 homolog', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=170344714, end_position_on_the_genomic_accession=170427911, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  76964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76964, CurrentID=0, Status='live', Symbol='2610028H24Rik', Aliases='ORF67', description='RIKEN cDNA 2610028H24 gene', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76449023, end_position_on_the_genomic_accession=76461218, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76965, CurrentID=0, Status='live', Symbol='Slitrk1', Aliases='3200001I04Rik', description='SLIT and NTRK-like family, member 1', other_designations='slit and trk like gene 1', map_location='14 E3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=108909988, end_position_on_the_genomic_accession=108914239, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  76967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76967, CurrentID=0, Status='live', Symbol='2700049A03Rik', Aliases='Talpid3, mKIAA0586', description='RIKEN cDNA 2700049A03 gene', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=71136848, end_position_on_the_genomic_accession=71309964, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  76969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76969, CurrentID=0, Status='live', Symbol='Chst1', Aliases='2610008E20Rik, AW125896, C6ST, GST-1, Gst1, KSGAL6ST', description='carbohydrate (keratan sulfate Gal-6) sulfotransferase 1', other_designations='KS6ST|KSST|N-acetylglucosamine-6-O-sulfotransferase 1|galactose/N-acetylglucosamine/N-acetylglucosamine 6-O-sulfotransferase 1|keratan sulfate Gal-6 sulfotransferase', map_location='2 D-E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92599632, end_position_on_the_genomic_accession=92615252, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  208043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208043, CurrentID=0, Status='live', Symbol='Setd1b', Aliases='AA516740, BC035291, KMT2G, mKIAA1076', description='SET domain containing 1B', other_designations='SET domain-containing protein 1B', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123140787, end_position_on_the_genomic_accession=123168634, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  76974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76974, CurrentID=0, Status='live', Symbol='Urah', Aliases='1190003J15Rik, 2810420C16Rik', description='urate (5-hydroxyiso-) hydrolase', other_designations='HIU hydrolase|HIUHase|transthyretin-related protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140835023, end_position_on_the_genomic_accession=140837968, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  76976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76976, CurrentID=0, Status='live', Symbol='Arxes2', Aliases='2900062L11Rik, Spc22/23, Spcs3', description='adipocyte-related X-chromosome expressed sequence 2', other_designations='SPase 22 kDa subunit|SPase 22/23 kDa subunit|microsomal signal peptidase 22/23 kDa subunit|microsomal signal peptidase 23 kDa subunit', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135993820, end_position_on_the_genomic_accession=135995359, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  76980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76980, CurrentID=0, Status='live', Symbol='Ube2ql1', Aliases='3110006E14Rik', description='ubiquitin-conjugating enzyme E2Q family-like 1', other_designations='E2Q-like ubiquitin-conjugating enzyme 1', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=69702832, end_position_on_the_genomic_accession=69739897, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  76982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76982, CurrentID=0, Status='live', Symbol='3110035E14Rik', Aliases='AI415019, AV164974, R75066', description='RIKEN cDNA 3110035E14 gene', other_designations='', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9601267, end_position_on_the_genomic_accession=9627142, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  76983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76983, CurrentID=0, Status='live', Symbol='Scfd1', Aliases='3110021P21Rik, RA410, STXBP1L2', description='Sec1 family domain containing 1', other_designations='syntaxin-binding protein 1-like 2', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51377515, end_position_on_the_genomic_accession=51450096, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  76987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76987, CurrentID=0, Status='live', Symbol='Hdhd2', Aliases='0610039H12Rik, 3110052N05Rik', description='haloacid dehalogenase-like hydrolase domain containing 2', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=76944115, end_position_on_the_genomic_accession=76972171, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  76998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=76998, CurrentID=0, Status='live', Symbol='Fbxw27', Aliases='1700124P09Rik', description='F-box and WD-40 domain protein 27', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109765655, end_position_on_the_genomic_accession=109790653, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  208076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208076, CurrentID=0, Status='live', Symbol='Pknox2', Aliases='D230005H23Rik, Prep2', description='Pbx/knotted 1 homeobox 2', other_designations='PBX/knotted homeobox 2|homeobox protein PREP-2|homeodomain containing transcription factor PREP2', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=36890979, end_position_on_the_genomic_accession=37147341, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  77006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77006, CurrentID=0, Status='live', Symbol='Ddrgk1', Aliases='1110001I20Rik, 2600009E05Rik, AI326138, Ufbp1', description='DDRGK domain containing 1', other_designations='UFM1-binding and PCI domain-containing protein 1|UFM1-binding protein 1 containing a PCI domain', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130654083, end_position_on_the_genomic_accession=130665600, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  77011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77011, CurrentID=0, Status='live', Symbol='Ticrr', Aliases='5730590G19Rik, BC028450, Lrrk1', description='TOPBP1-interacting checkpoint and replication regulator', other_designations='leucine-rich repeat kinase 1|topBP1-interacting, replication-stimulating protein', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79660196, end_position_on_the_genomic_accession=79698148, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  208084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208084, CurrentID=0, Status='live', Symbol='Pif1', Aliases='4631410M14, AI449441', description="PIF1 5'-to-3' DNA helicase homolog (S. cerevisiae)", other_designations='DNA helicase-like protein|DNA repair and recombination helicase PIF1|PIF1 homolog|Pif1/Rrm3 DNA helicase-like protein', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65587147, end_position_on_the_genomic_accession=65595967, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  77015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77015, CurrentID=0, Status='live', Symbol='Mpped2', Aliases='239Fb, 2700082O15Rik, AV354767, AW060960', description='metallophosphoesterase domain containing 2', other_designations='metallophosphoesterase domain-containing protein 2', map_location='2 55.8 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=106693459, end_position_on_the_genomic_accession=106868360, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  77018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77018, CurrentID=0, Status='live', Symbol='Col25a1', Aliases='2700062B08Rik, CLAC-P', description='collagen, type XXV, alpha 1', other_designations='col25a1|collagen-like Alzheimer amyloid plaque component|procollagen, type XXV, alpha 1', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=130180339, end_position_on_the_genomic_accession=130599884, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  208092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208092, CurrentID=0, Status='live', Symbol='Chmp6', Aliases='2400004G01Rik', description='charged multivesicular body protein 6', other_designations='chromatin modifying protein 6|chromatin-modifying protein 6|novel SNF7 protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119913446, end_position_on_the_genomic_accession=119919552, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  208098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208098, CurrentID=0, Status='live', Symbol='Panx3', Aliases='3230401P04, 4833413G11Rik', description='pannexin 3', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37659902, end_position_on_the_genomic_accession=37669222, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  77031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77031, CurrentID=0, Status='live', Symbol='Slc9a8', Aliases='1200006P13Rik, 6430709P13Rik, AI182282, NHE-8, NHE8', description='solute carrier family 9 (sodium/hydrogen exchanger), member 8', other_designations='Na(+)/H(+) exchanger 8|Na-H exchanger isoform NHE8|sodium/proton exchanger NHE8', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167420056, end_position_on_the_genomic_accession=167477000, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  208104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208104, CurrentID=0, Status='live', Symbol='Mlxip', Aliases='AW228700, Mir, bHLHe36', description='MLX interacting protein', other_designations='Mlx interactor|mlxip|transcriptional activator MondoA', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123394798, end_position_on_the_genomic_accession=123457932, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  77034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77034, CurrentID=0, Status='live', Symbol='2510039O18Rik', Aliases='Kiaa2013', description='RIKEN cDNA 2510039O18 gene', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147940855, end_position_on_the_genomic_accession=147947316, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  77035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77035, CurrentID=0, Status='live', Symbol='Kdm8', Aliases='3110005O21Rik, Jmjd5', description='lysine (K)-specific demethylase 8', other_designations='jmjC domain-containing protein 5|jumonji domain containing 5|jumonji domain-containing protein 5', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125444620, end_position_on_the_genomic_accession=125462269, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  77036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77036, CurrentID=0, Status='live', Symbol='1700109H08Rik', Aliases='', description='RIKEN cDNA 1700109H08 gene', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3571013, end_position_on_the_genomic_accession=3584423, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  77037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77037, CurrentID=0, Status='live', Symbol='Mrap', Aliases='1110025G12Rik, Falp, ORF61', description='melanocortin 2 receptor accessory protein', other_designations='fat cell-specific low molecular weight protein|fat tissue-specific low MW protein', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90738324, end_position_on_the_genomic_accession=90749776, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  77038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77038, CurrentID=0, Status='live', Symbol='Arfgap2', Aliases='2310032E02Rik, Zfp289', description='ADP-ribosylation factor GTPase activating protein 2', other_designations='ARF GAP 2|GTPase-activating protein ZNF289|zinc finger protein 289', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91265115, end_position_on_the_genomic_accession=91277371, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  208111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208111, CurrentID=0, Status='live', Symbol='9830147E19Rik', Aliases='C330019L16Rik', description='RIKEN cDNA 9830147E19 gene', other_designations='', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=42609526, end_position_on_the_genomic_accession=42642604, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  77040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77040, CurrentID=0, Status='live', Symbol='Atg16l1', Aliases='1500009K01Rik, Apg16l, Atg16l, WDR30', description='autophagy related 16-like 1 (S. cerevisiae)', other_designations='APG16-like 1|Atg16L1 gamma|autophagy-related 16-like 1 gamma', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87756006, end_position_on_the_genomic_accession=87792428, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  77041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77041, CurrentID=0, Status='live', Symbol='Arsk', Aliases='2810429K17Rik, 4833414G15Rik', description='arylsulfatase K', other_designations='ASK', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=76060422, end_position_on_the_genomic_accession=76098667, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  77042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77042, CurrentID=0, Status='live', Symbol='Hyal4', Aliases='4632428M18Rik, CSHY', description='hyaluronoglucosaminidase 4', other_designations='chondroitin sulfate endo-beta-N-acetylgalactosaminidase|chondroitin sulfate hydrolase|hyal-4|hyaluronoglucosaminidase-4', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24748367, end_position_on_the_genomic_accession=24766519, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  77044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77044, CurrentID=0, Status='live', Symbol='Arid2', Aliases='1700124K17Rik, 4432409D24Rik, zipzap/p200', description='AT rich interactive domain 2 (ARID, RFX-like)', other_designations='AT rich interactive domain 2 (Arid-rfx like)|zipzap', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=96287522, end_position_on_the_genomic_accession=96405463, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  77045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77045, CurrentID=0, Status='live', Symbol='Bcl7a', Aliases='4432415N06Rik, AI448316', description='B cell CLL/lymphoma 7A', other_designations='B-cell CLL/lymphoma 7A', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123344448, end_position_on_the_genomic_accession=123374083, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  77048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77048, CurrentID=0, Status='live', Symbol='Cep83', Aliases='2600001G24Rik, 4921537D05Rik, 5730513H21Rik, Ccdc41', description='centrosomal protein 83', other_designations='coiled-coil domain containing 41|coiled-coil domain-containing protein 41', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=94688495, end_position_on_the_genomic_accession=94790339, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  77049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77049, CurrentID=0, Status='live', Symbol='4921528I07Rik', Aliases='', description='RIKEN cDNA 4921528I07 gene', other_designations='', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114237695, end_position_on_the_genomic_accession=114317064, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  77053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77053, CurrentID=0, Status='live', Symbol='Sun1', Aliases='4632417G13Rik, 5730434D03Rik, Unc84a, mKIAA0810', description='Sad1 and UNC84 domain containing 1', other_designations='protein unc-84 homolog A|sad1/unc-84 protein-like 1|unc-84 homolog A', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139200558, end_position_on_the_genomic_accession=139249839, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  77055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77055, CurrentID=0, Status='live', Symbol='Krt76', Aliases='2310001L23Rik', description='keratin 76', other_designations='K76|keratin-76|krt76|type-II keratin Kb9', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101884351, end_position_on_the_genomic_accession=101892920, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  77056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77056, CurrentID=0, Status='live', Symbol='Tmco4', Aliases='4632413C14Rik, AI430825', description='transmembrane and coiled-coil domains 4', other_designations='novel DUF726 domain containing protein', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138972901, end_position_on_the_genomic_accession=139059173, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  77057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77057, CurrentID=0, Status='live', Symbol='Ston1', Aliases='4921524J06Rik, C86231, Salf, Sblf', description='stonin 1', other_designations='stoned B-like factor|stoned B/TFIIA-alpha/beta-like factor', map_location='17 E5', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=88597635, end_position_on_the_genomic_accession=88645724, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  77058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77058, CurrentID=0, Status='live', Symbol='Ccdc183', Aliases='4921530D09Rik, Cccd183, Kiaa1984', description='coiled-coil domain containing 183', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25608629, end_position_on_the_genomic_accession=25617678, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  77065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77065, CurrentID=0, Status='live', Symbol='Ints7', Aliases='5930412E23Rik, int7', description='integrator complex subunit 7', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191575641, end_position_on_the_genomic_accession=191622676, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  208144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208144, CurrentID=0, Status='live', Symbol='Dhx37', Aliases='Gm1050, Gm451, mKIAA1517', description='DEAH (Asp-Glu-Ala-His) box polypeptide 37', other_designations='', map_location='5 G1.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=125413747, end_position_on_the_genomic_accession=125434139, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  208146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208146, CurrentID=0, Status='live', Symbol='Yeats2', Aliases='BC042768, mKIAA1197', description='YEATS domain containing 2', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20141059, end_position_on_the_genomic_accession=20232574, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  77080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77080, CurrentID=0, Status='live', Symbol='9230110F15Rik', Aliases='Pate-A, mANLP3', description='RIKEN cDNA 9230110F15 gene', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35838329, end_position_on_the_genomic_accession=35844150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  208154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208154, CurrentID=0, Status='live', Symbol='Btla', Aliases='A630002H24', description='B and T lymphocyte associated', other_designations='B and T lymphocyte attenuator|B- and T-lymphocyte-associated protein', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45223547, end_position_on_the_genomic_accession=45252895, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  208158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208158, CurrentID=0, Status='live', Symbol='Map6d1', Aliases='A330066I15Rik', description='MAP6 domain containing 1', other_designations='21 kDa STOP-like protein|SL21', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20233309, end_position_on_the_genomic_accession=20241358, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  77087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77087, CurrentID=0, Status='live', Symbol='Ankrd11', Aliases='2410104C19Rik, 3010027A04Rik, 6330578C09Rik, 9530048I21Rik, AA930108, Gm176, Yod', description='ankyrin repeat domain 11', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122883822, end_position_on_the_genomic_accession=123042284, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  668208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668208, CurrentID=0, Status='live', Symbol='Gm13288', Aliases='OTTMUSG00000011289', description='predicted gene 13288', other_designations='interferon zeta family Ifnz member', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88805531, end_position_on_the_genomic_accession=88808380, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  77090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77090, CurrentID=0, Status='live', Symbol='Ocel1', Aliases='9430098E02Rik, AI326022, AI593544', description='occludin/ELL domain containing 1', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71371298, end_position_on_the_genomic_accession=71373689, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  208164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208164, CurrentID=0, Status='live', Symbol='Fam180a', Aliases='B230314O19', description='family with sequence similarity 180, member A', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=35312746, end_position_on_the_genomic_accession=35326141, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  208166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208166, CurrentID=0, Status='live', Symbol='Gm609', Aliases='Gm1749, isec1', description='predicted gene 609', other_designations='913020L22Rik|iSEC1 protein, intestinal secretory cell protein 1', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45412005, end_position_on_the_genomic_accession=45493004, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77097, CurrentID=0, Status='live', Symbol='Tanc2', Aliases='3526402J09Rik, 5730590C14Rik', description='tetratricopeptide repeat, ankyrin repeat and coiled-coil containing 2', other_designations='tetratricopeptide repeat, ankyrin repeat and coiled-coil domain-containing protein 2', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105589986, end_position_on_the_genomic_accession=105929304, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  208171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208171, CurrentID=0, Status='live', Symbol='Tmprss7', Aliases='B230219I23Rik, Gm1748', description='transmembrane serine protease 7', other_designations='matriptase-3|transmembrane protease, serine 7', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45656309, end_position_on_the_genomic_accession=45698553, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  208177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208177, CurrentID=0, Status='live', Symbol='Phldb2', Aliases='AV253284, C820004H04Rik, LL5b, LL5beta', description='pleckstrin homology-like domain, family B, member 2', other_designations='protein LL5-beta', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45746230, end_position_on_the_genomic_accession=45953701, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  77106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77106, CurrentID=0, Status='live', Symbol='Tmem181a', Aliases='5930418K15Rik, C76977, Gpr178, Tmem181, mKIAA1423', description='transmembrane protein 181A', other_designations='G protein-coupled receptor 178', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6256865, end_position_on_the_genomic_accession=6308314, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  77110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77110, CurrentID=0, Status='live', Symbol='Gpbp1l1', Aliases='5330440M15Rik, BC002292', description='GC-rich promoter binding protein 1-like 1', other_designations='GC-rich promoter-binding protein 1-like 1', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116557157, end_position_on_the_genomic_accession=116593882, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  77113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77113, CurrentID=0, Status='live', Symbol='Klhl2', Aliases='6030411N21Rik, ABP-KELCH, AU020744, Mav', description='kelch-like 2, Mayven', other_designations='klhl2', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=64811678, end_position_on_the_genomic_accession=64849924, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  77116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77116, CurrentID=0, Status='live', Symbol='Mtmr2', Aliases='6030445P13Rik', description='myotubularin related protein 2', other_designations='phosphatidylinositol-3,5-bisphosphate 3-phosphatase|phosphatidylinositol-3-phosphate phosphatase', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=13748232, end_position_on_the_genomic_accession=13808896, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  208194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208194, CurrentID=0, Status='live', Symbol='Exog', Aliases='AW557704, ENGL-B, ENGL-a, Endogl1, Endogl2, Engl, Engla, Englb', description="endo/exonuclease (5'-3'), endonuclease G-like", other_designations='endo G-like 1|endonuclease G-like 1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119444923, end_position_on_the_genomic_accession=119465518, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  77124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77124, CurrentID=0, Status='live', Symbol='9130221H12Rik', Aliases='AU016887', description='RIKEN cDNA 9130221H12 gene', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25228723, end_position_on_the_genomic_accession=25230264, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  77125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77125, CurrentID=0, Status='live', Symbol='Il33', Aliases='9230117N10Rik, Il-33, Il1f11, NF-HEV', description='interleukin 33', other_designations='nuclear factor from high endothelial venules', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29925114, end_position_on_the_genomic_accession=29960715, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  208198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208198, CurrentID=0, Status='live', Symbol='Btbd2', Aliases='2610037C03Rik, 4930512K17Rik', description='BTB (POZ) domain containing 2', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80642617, end_position_on_the_genomic_accession=80657118, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  77128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77128, CurrentID=0, Status='live', Symbol='Crebrf', Aliases='A930001N09Rik, LRF', description='CREB3 regulatory factor', other_designations='UPF0474 protein C5orf41 homolog|luman recruitment factor', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26714711, end_position_on_the_genomic_accession=26780985, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  732489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=732489, CurrentID=0, Status='live', Symbol='Mir674', Aliases='Mirn674, mir-674, mmu-mir-674', description='microRNA 674', other_designations='mir 674', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=117185127, end_position_on_the_genomic_accession=117185226, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  77134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77134, CurrentID=0, Status='live', Symbol='Hnrnpa0', Aliases='1110055B05Rik, 3010025E17Rik, Hnrpa0, hnRNP A0', description='heterogeneous nuclear ribonucleoprotein A0', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58125879, end_position_on_the_genomic_accession=58128556, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  208211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208211, CurrentID=0, Status='live', Symbol='Alg1', Aliases='HMAT1, HMT1', description='asparagine-linked glycosylation 1 (beta-1,4-mannosyltransferase)', other_designations='GDP-Man:GlcNAc2-PP-dolichol mannosyltransferase|GDP-mannose-dolichol diphosphochitobiose mannosyltransferase|asparagine-linked glycosylation 1 homolog (yeast, beta-1,4-mannosyltransferase)|asparagine-linked glycosylation protein 1 homolog|beta-1,4 mannosyltransferase|beta-1,4-mannosyltransferase', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5233621, end_position_on_the_genomic_accession=5244907, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  208213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208213, CurrentID=0, Status='live', Symbol='Tmem132c', Aliases='2810482M11Rik, 3230401P16, 4632425D07Rik', description='transmembrane protein 132C', other_designations='', map_location='5 G1.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=127241826, end_position_on_the_genomic_accession=127565790, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  12858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12858, CurrentID=0, Status='live', Symbol='Cox5a', Aliases='AA959768, CcOX', description='cytochrome c oxidase subunit Va', other_designations='cytochrome c oxidase polypeptide Va', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57521232, end_position_on_the_genomic_accession=57532427, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  208228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208228, CurrentID=0, Status='live', Symbol='Mob3a', Aliases='5330417K06Rik, A630029F06, AV218468, Mobkl2a', description='MOB kinase activator 3A', other_designations='MOB1, Mps One Binder kinase activator-like 2A|mob1 homolog 2A|mps one binder kinase activator-like 2A', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80685253, end_position_on_the_genomic_accession=80701896, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  208258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208258, CurrentID=0, Status='live', Symbol='Ankrd33', Aliases='A930021G21Rik, Panky, Panky-a', description='ankyrin repeat domain 33', other_designations='photorecptor ankyrin repeat protein A', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101115755, end_position_on_the_genomic_accession=101120407, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  208263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208263, CurrentID=0, Status='live', Symbol='Tor1aip1', Aliases='LAP1, Lap1B', description='torsin A interacting protein 1', other_designations='lamina-associated polypeptide 1B', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156004599, end_position_on_the_genomic_accession=156036480, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  208266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208266, CurrentID=0, Status='live', Symbol='Dot1l', Aliases='A630076O07, AW907654, Dot1, KMT4, mDot1', description='DOT1-like, histone H3 methyltransferase (S. cerevisiae)', other_designations='DOT1-like histone H3 methyltransferase|histone H3 methyltransferase DOT1|histone methyltransferase DOT1L', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80755129, end_position_on_the_genomic_accession=80795461, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  77219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77219, CurrentID=0, Status='live', Symbol='Ptgr2', Aliases='1810016I24Rik, 9130222H03Rik, 9630002F03Rik, AI838763, B830026H24Rik, PGR-2, PRG-2, Zadh1', description='prostaglandin reductase 2', other_designations='15-oxoprostaglandin 13-reductase|zinc binding alcohol dehydrogenase, domain containing 1|zinc-binding alcohol dehydrogenase domain-containing protein 1', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84285237, end_position_on_the_genomic_accession=84315832, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  208292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208292, CurrentID=0, Status='live', Symbol='Zfp871', Aliases='4732483N19, 9030612M13Rik, AI118577, AI265322, mKIAA3006', description='zinc finger protein 871', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32765496, end_position_on_the_genomic_accession=32788289, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  77252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77252, CurrentID=0, Status='live', Symbol='9430038I01Rik', Aliases='', description='RIKEN cDNA 9430038I01 gene', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=137375569, end_position_on_the_genomic_accession=137411900, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  77254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77254, CurrentID=0, Status='live', Symbol='Yif1b', Aliases='9430029K10Rik, Yip1b', description='Yip1 interacting factor homolog B (S. cerevisiae)', other_designations='YIP1-interacting factor homolog B', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29238308, end_position_on_the_genomic_accession=29247532, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  102642252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102642252, CurrentID=0, Status='live', Symbol='LOC102642252', Aliases='IgH, IgVH', description='uncharacterized LOC102642252', other_designations='immunoglobulin heavy chain V-region', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  77264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77264, CurrentID=0, Status='live', Symbol='Zfp142', Aliases='9330177B18Rik, BB154236, mKIAA0236', description='zinc finger protein 142', other_designations='', map_location='1 38.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74565128, end_position_on_the_genomic_accession=74588211, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  78416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78416, CurrentID=0, Status='live', Symbol='Rnase6', Aliases='9530043P15Rik, BB119466', description='ribonuclease, RNase A family, 6', other_designations='RNase K6|ribonuclease A family, member 6', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51123913, end_position_on_the_genomic_accession=51132187, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  77286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77286, CurrentID=0, Status='live', Symbol='Nkrf', Aliases='9430034D17Rik, AW124770, C80287', description='NF-kappaB repressing factor', other_designations='NFkB-repressing factor|transcription factor NRF', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36886440, end_position_on_the_genomic_accession=36903110, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  208366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208366, CurrentID=0, Status='live', Symbol='Rpp40', Aliases='D8Bwg1265e, Rnasep1', description='ribonuclease P 40 subunit', other_designations='RNaseP protein p40|ribonuclease P 40kDa subunit|ribonuclease P1', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=35895104, end_position_on_the_genomic_accession=35911565, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  77296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77296, CurrentID=0, Status='live', Symbol='Fam162b', Aliases='9430073N08Rik', description='family with sequence similarity 162, member B', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=51585414, end_position_on_the_genomic_accession=51590592, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  77300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77300, CurrentID=0, Status='live', Symbol='Raph1', Aliases='9430025M21Rik, C730009O10Rik, Lpd', description='Ras association (RalGDS/AF-6) and pleckstrin homology domains 1', other_designations='lamellipodin', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60483185, end_position_on_the_genomic_accession=60567074, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  77305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77305, CurrentID=0, Status='live', Symbol='Wdr82', Aliases='9430077D24Rik, CDW5/WDR82', description='WD repeat domain containing 82', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106170929, end_position_on_the_genomic_accession=106191706, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  77318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77318, CurrentID=0, Status='live', Symbol='Ankrd55', Aliases='C030011J08Rik', description='ankyrin repeat domain 55', other_designations='', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112288451, end_position_on_the_genomic_accession=112384002, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  12889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12889, CurrentID=0, Status='live', Symbol='Cplx1', Aliases='921-S', description='complexin 1', other_designations='CPX I|complexin I|synaphin-2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108518554, end_position_on_the_genomic_accession=108550027, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  77337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77337, CurrentID=0, Status='live', Symbol='Akr1c21', Aliases='9430025F20Rik, AI315367', description='aldo-keto reductase family 1, member C21', other_designations='17-alpha-HSD|17-alpha-hydroxysteroid dehydrogenase|3(17)alpha-hydroxysteroid dehydrogenase|3(or 17)-alpha-hydroxysteroid dehydrogenase|3-alpha-hydroxysteroid dehydrogenase|DD1|DD3|dihydrodiol dehydrogenase type 1|dihydrodiol dehydrogenase type 3', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4574075, end_position_on_the_genomic_accession=4586543, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  77352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77352, CurrentID=0, Status='live', Symbol='Axdnd1', Aliases='9430070O13Rik, Gm979', description='axonemal dynein light chain domain containing 1', other_designations='', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156330415, end_position_on_the_genomic_accession=156420915, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  208426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208426, CurrentID=0, Status='live', Symbol='Iqcj', Aliases='D230050A05, EG208426', description='IQ motif containing J', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=67892220, end_position_on_the_genomic_accession=68056593, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  208431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208431, CurrentID=0, Status='live', Symbol='Shroom4', Aliases='D430043L16Rik, Gm724, Shrm4', description='shroom family member 4', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=6399938, end_position_on_the_genomic_accession=6637452, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  208439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208439, CurrentID=0, Status='live', Symbol='Klhl29', Aliases='A230106N14Rik, Gm68, Kbtbd9, mKIAA1921', description='kelch-like 29', other_designations='kelch repeat and BTB (POZ) domain containing 9|kelch repeat and BTB domain-containing protein 9', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=5077465, end_position_on_the_genomic_accession=5375682, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  208440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208440, CurrentID=0, Status='live', Symbol='Dip2c', Aliases='2900024P20Rik, 9630044M06, mKIAA0934', description='disco interacting protein 2 homolog C', other_designations='DIP2 disco-interacting protein 2 homolog C', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=9276503, end_position_on_the_genomic_accession=9668928, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  77371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77371, CurrentID=0, Status='live', Symbol='Sec24a', Aliases='', description='Sec24 related gene family, member A (S. cerevisiae)', other_designations='SEC24-related protein A', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51692263, end_position_on_the_genomic_accession=51763627, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  208449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208449, CurrentID=0, Status='live', Symbol='Sgms1', Aliases='9530058O11Rik, AI841905, C80702, Mob, Sms1, Sor1, Tmem23', description='sphingomyelin synthase 1', other_designations='protein Mob|transmembrane protein 23', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=32122727, end_position_on_the_genomic_accession=32389705, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  77397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77397, CurrentID=0, Status='live', Symbol='9530003J23Rik', Aliases='Lyzf1', description='RIKEN cDNA 9530003J23 gene', other_designations='lysozyme f1', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117231047, end_position_on_the_genomic_accession=117238681, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  77407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77407, CurrentID=0, Status='live', Symbol='Rab35', Aliases='9530019H02Rik, AU040256, H-ray, RAB1C, RAY', description='RAB35, member RAS oncogene family', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115631882, end_position_on_the_genomic_accession=115647158, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  77411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77411, CurrentID=0, Status='live', Symbol='Esrp2', Aliases='9530027K23Rik, Rbm35b', description='epithelial splicing regulatory protein 2', other_designations='RNA binding motif protein 35b|RNA-binding motif protein 35B|RNA-binding protein 35B', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106130403, end_position_on_the_genomic_accession=106137528, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  77422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77422, CurrentID=0, Status='live', Symbol='C330018D20Rik', Aliases='AU018666', description='RIKEN cDNA C330018D20 gene', other_designations='glutaredoxin-like protein YDR286C homolog', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=56955828, end_position_on_the_genomic_accession=56975409, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  77424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77424, CurrentID=0, Status='live', Symbol='Spinkl', Aliases='9530002K18Rik', description='serine protease inhibitor, Kazal type-like', other_designations='SPINKL, minor form', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44166358, end_position_on_the_genomic_accession=44176707, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  208501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208501, CurrentID=0, Status='live', Symbol='1810043H04Rik', Aliases='', description='RIKEN cDNA 1810043H04 gene', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120098934, end_position_on_the_genomic_accession=120100424, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  77432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77432, CurrentID=0, Status='live', Symbol='9530002B09Rik', Aliases='Aump, Vpp1, Vppl', description='RIKEN cDNA 9530002B09 gene', other_designations='palmitoyl protein thioesterase-like protein|ventral prostrate predominant 1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=122689259, end_position_on_the_genomic_accession=122705135, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  208518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208518, CurrentID=0, Status='live', Symbol='Cep78', Aliases='5730599I05Rik, D030027P05', description='centrosomal protein 78', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=15955768, end_position_on_the_genomic_accession=15985016, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  329918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329918, CurrentID=0, Status='live', Symbol='Skint9', Aliases='A030013N09Rik, Skint-9', description='selection and upkeep of intraepithelial T cells 9', other_designations='skint 9', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=112385969, end_position_on_the_genomic_accession=112433985, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77462, CurrentID=0, Status='live', Symbol='Tmem116', Aliases='4930406A18Rik, 4930513P12Rik, C030022K24Rik', description='transmembrane protein 116', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121452709, end_position_on_the_genomic_accession=121495421, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  77480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77480, CurrentID=0, Status='live', Symbol='Kidins220', Aliases='3110039L19Rik, AI194387, AI316525, C330002I19Rik, mKIAA1250', description='kinase D-interacting substrate 220', other_designations='truncated Kidins220', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24974876, end_position_on_the_genomic_accession=25063152, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  77485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77485, CurrentID=0, Status='live', Symbol='Stk31', Aliases='C330007K24Rik', description='serine threonine kinase 31', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49395483, end_position_on_the_genomic_accession=49469506, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  77505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77505, CurrentID=0, Status='live', Symbol='Dnhd1', Aliases='8030491N06Rik', description='dynein heavy chain domain 1', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105650813, end_position_on_the_genomic_accession=105721800, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  208583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208583, CurrentID=0, Status='live', Symbol='Nek11', Aliases='4932416N14Rik', description='NIMA (never in mitosis gene a)-related expressed kinase 11', other_designations='NIMA-related kinase 11|nek11|never in mitosis A-related kinase 11|nimA-related protein kinase 11', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105162156, end_position_on_the_genomic_accession=105395524, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  77519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77519, CurrentID=0, Status='live', Symbol='Zfp266', Aliases='5330440G10Rik, 5730601F06Rik, AW552317', description='zinc finger protein 266', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20495068, end_position_on_the_genomic_accession=20521619, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  77521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77521, CurrentID=0, Status='live', Symbol='Mtus2', Aliases='5730592G18Rik, A730013O20Rik, AU067634, C130038G02Rik, Cazip, Gm763, mKIAA0774', description='microtubule associated tumor suppressor candidate 2', other_designations='cardiac zipper protein|microtubule plus-end tracking protein TIP150|tracking protein of 150 kDa', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147956987, end_position_on_the_genomic_accession=148316065, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  77522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77522, CurrentID=0, Status='live', Symbol='Tmem213', Aliases='D630002J15Rik', description='transmembrane protein 213', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38109353, end_position_on_the_genomic_accession=38115806, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  208595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208595, CurrentID=0, Status='live', Symbol='Mterf1b', Aliases='ENSMUSG00000053178, Gm9897', description='mitochondrial transcription termination factor 1b', other_designations='transcription termination factor, mitochondrial', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=4192367, end_position_on_the_genomic_accession=4197651, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  77531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77531, CurrentID=0, Status='live', Symbol='Anks1b', Aliases='AIDA-1b, C030032C09Rik, E530015N03Rik, ENSMUSG00000078431, Gm10937, Gm1555', description='ankyrin repeat and sterile alpha motif domain containing 1B', other_designations='AIDA-1|E2A-PBX1-associated protein|EB-1|amyloid-beta protein intracellular domain-associated protein 1', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=89873509, end_position_on_the_genomic_accession=90973300, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  77532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77532, CurrentID=0, Status='live', Symbol='Jrkl', Aliases='C030035D04Rik, HHMJG', description='Jrk-like', other_designations='jerky homolog-like', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=13242790, end_position_on_the_genomic_accession=13245741, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  208606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208606, CurrentID=0, Status='live', Symbol='Rsrc2', Aliases='1500011J06Rik', description='arginine/serine-rich coiled-coil 2', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123728425, end_position_on_the_genomic_accession=123749418, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  208613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208613, CurrentID=0, Status='live', Symbol='Tmem212', Aliases='E030011K20Rik, Gm564', description='transmembrane protein 212', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=27866065, end_position_on_the_genomic_accession=27896391, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  208618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208618, CurrentID=0, Status='live', Symbol='Etl4', Aliases='6620402G01Rik, 9430077C05Rik, BC026657, E330027G05Rik, Etl-4, Skt', description='enhancer trap locus 4', other_designations='', map_location='2 14.27 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=19909763, end_position_on_the_genomic_accession=20810535, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  77552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77552, CurrentID=0, Status='live', Symbol='Shisa4', Aliases='9330132O05Rik, Tmem58, mShisa4', description='shisa family member 4', other_designations='shisa homolog 4|transmembrane protein 58', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135371057, end_position_on_the_genomic_accession=135375277, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  208628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208628, CurrentID=0, Status='live', Symbol='Kntc1', Aliases='D330012D13, jgl', description='kinetochore associated 1', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123749689, end_position_on_the_genomic_accession=123821593, orientation='plus', exon_count=66, OMIM='', no_hdr0=''),
  77559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77559, CurrentID=0, Status='live', Symbol='Agl', Aliases='1110061O17Rik, 9430004C13Rik, 9630046L06Rik, AI850929, C77197', description='amylo-1,6-glucosidase, 4-alpha-glucanotransferase', other_designations='', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116739999, end_position_on_the_genomic_accession=116808166, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  208634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208634, CurrentID=0, Status='live', Symbol='Tspan10', Aliases='Ocsp', description='tetraspanin 10', other_designations='oculospanin|tspan-10', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120442631, end_position_on_the_genomic_accession=120447321, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  208638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208638, CurrentID=0, Status='live', Symbol='Slc25a38', Aliases='AV019094, BC010801', description='solute carrier family 25, member 38', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120110391, end_position_on_the_genomic_accession=120124342, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  77569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77569, CurrentID=0, Status='live', Symbol='Limch1', Aliases='3732412D22Rik, mKIAA1102', description='LIM and calponin homology domains 1', other_designations='', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=66745840, end_position_on_the_genomic_accession=67057159, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  208643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208643, CurrentID=0, Status='live', Symbol='Eif4g1', Aliases='E030015G23Rik, eIF-4-gamma 1, eIF-4G 1, eIF-4G1, eIF4GI', description='eukaryotic translation initiation factor 4, gamma 1', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20672720, end_position_on_the_genomic_accession=20692883, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  77573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77573, CurrentID=0, Status='live', Symbol='Vps33a', Aliases='3830421M04Rik, AI503300, AW048546, AW554476, bf', description='vacuolar protein sorting 33A (yeast)', other_designations='buff|vacuolar sorting protein 33a', map_location='5 63.03 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123527304, end_position_on_the_genomic_accession=123573043, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  77574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77574, CurrentID=0, Status='live', Symbol='Tcaf1', Aliases='2810407D09Rik, 3321401G04Rik, A230020K05Rik, AI848529, Fam115a', description='TRPM8 channel-associated factor 1', other_designations='TRP channel-associated factor 1|family with sequence similarity 115, member A|protein FAM115A', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42667999, end_position_on_the_genomic_accession=42710110, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  208647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208647, CurrentID=0, Status='live', Symbol='Creb3l2', Aliases='BBF2H7, C530025K05Rik, SCIRR69', description='cAMP responsive element binding protein 3-like 2', other_designations='cAMP-responsive element-binding protein 3-like protein 2', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=37327255, end_position_on_the_genomic_accession=37442186, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  77577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77577, CurrentID=0, Status='live', Symbol='Spns3', Aliases='9830002I17Rik', description='spinster homolog 3', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72498153, end_position_on_the_genomic_accession=72550508, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  77578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77578, CurrentID=0, Status='live', Symbol='Bcl9', Aliases='2610202E01Rik, 8030475K17Rik, A330041G23Rik, Gm130', description='B cell CLL/lymphoma 9', other_designations='B-cell lymphoma 9 protein|bcl-9|nuclear co-factor of beta-catenin signalling', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97203658, end_position_on_the_genomic_accession=97298083, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  77579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77579, CurrentID=0, Status='live', Symbol='Myh10', Aliases='5730504C04Rik, 9330167F11Rik, Fltn, Myhn-2, Myhn2, NMHC II-B, NMHC-B, NMHCII-B, NMMHC II-b, NMMHC-B, NMMHC-IIB, SMemb, mKIAA3005', description='myosin, heavy polypeptide 10, non-muscle', other_designations='cellular myosin heavy chain, type B|flectin|myosin IIB|myosin heavy chain 10, non-muscle|myosin heavy chain, non-muscle IIb|myosin heavy chain, nonmuscular, type B|non-muscle myosin heavy chain B|non-muscle myosin heavy chain IIb|nonmuscle myosin heavy chain II-B', map_location='11 41.95 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68691915, end_position_on_the_genomic_accession=68816625, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  77582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77582, CurrentID=0, Status='live', Symbol='Mboat7', Aliases='5730589L02Rik, BB1, Leng4, Lpiat, Lpiat1, mBB1', description='membrane bound O-acyltransferase domain containing 7', other_designations='LPLAT 7|O-acyltransferase domain-containing protein 7|bladder and breast carcinoma-overexpressed gene 1 protein|leukocyte receptor cluster (LRC) member 4|leukocyte receptor cluster member 4|lysophosphatidylinositol acyltransferase|m-mboa-7|membrane-bound O-acyltransferase domain-containing protein 7', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3677789, end_position_on_the_genomic_accession=3693525, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  77583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77583, CurrentID=0, Status='live', Symbol='Notum', Aliases='5730593N15Rik', description='notum pectinacetylesterase homolog (Drosophila)', other_designations='protein notum homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120653788, end_position_on_the_genomic_accession=120661745, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  208659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208659, CurrentID=0, Status='live', Symbol='Fam20a', Aliases='AI606893', description='family with sequence similarity 20, member A', other_designations='ortholog of human family with sequence similarity 20, member A FAM20A|protein FAM20A', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109669492, end_position_on_the_genomic_accession=109722510, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  77590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77590, CurrentID=0, Status='live', Symbol='Chst15', Aliases='4631426J05Rik, BRAG, GalNAcS-6ST, MAd5, mKIAA0598', description='carbohydrate (N-acetylgalactosamine 4-sulfate 6-O) sulfotransferase 15', other_designations='B-cell RAG-associated gene protein|GalNAc 4-sulfate 6-O-sulfotransferase|GalNAc4S-6ST|N-acetylgalactosamine 4-sulfate 6-O-sulfotransferase', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132235779, end_position_on_the_genomic_accession=132317155, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  77591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77591, CurrentID=0, Status='live', Symbol='Ddx10', Aliases='4632415A01Rik, AI646054', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 10', other_designations='DEAD box protein 10|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 10 (RNA helicase)', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53098454, end_position_on_the_genomic_accession=53248170, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  77593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77593, CurrentID=0, Status='live', Symbol='Usp45', Aliases='3110003C05Rik, 4930550B20Rik, AI843191, Gcap7', description='ubiquitin specific petidase 45', other_designations='deubiquitinating enzyme 45|granule cell antiserum positive 7|ubiquitin specific protease 45|ubiquitin thioesterase 45|ubiquitin thiolesterase 45|ubiquitin-specific-processing protease 45', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=21767161, end_position_on_the_genomic_accession=21837872, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  208666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208666, CurrentID=0, Status='live', Symbol='Diras1', Aliases='Di-Ras1, Gbts1, RIG', description='DIRAS family, GTP-binding RAS-like 1', other_designations='distinct subgroup of the Ras family member 1|small GTP-binding tumor suppressor 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81019589, end_position_on_the_genomic_accession=81025821, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  77595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77595, CurrentID=0, Status='live', Symbol='Nup210l', Aliases='4930548O11Rik, R26-EGFP, Tg(Gt(ROSA)26Sor-EGFP)130910Eps', description='nucleoporin 210-like', other_designations='nucleoporin 210 kDa-like|nucleoporin Nup210-like', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90101569, end_position_on_the_genomic_accession=90212151, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  77596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77596, CurrentID=0, Status='live', Symbol='Adgrf1', Aliases='5031409J19Rik, Gpr110, KPG_009', description='adhesion G protein-coupled receptor F1', other_designations='G protein-coupled receptor 110|G-protein coupled receptor 110|internal gene name of KIRIN laboratory:M5', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43270276, end_position_on_the_genomic_accession=43325876, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  77604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77604, CurrentID=0, Status='live', Symbol='Rbm12b2', Aliases='AV299215, C430048L16Rik, Rbm12bb', description='RNA binding motif protein 12 B2', other_designations='RNA-binding motif protein 12B-B', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=12089370, end_position_on_the_genomic_accession=12096271, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  77605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77605, CurrentID=0, Status='live', Symbol='H2afv', Aliases='C530002L11Rik, H2a.z-2, H2av', description='H2A histone family, member V', other_designations='H2A.F/Z|histone H2A.F/Z', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6427226, end_position_on_the_genomic_accession=6444443, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  77609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77609, CurrentID=0, Status='live', Symbol='Ccdc151', Aliases='AI644415, C330001K17Rik, b2b1885Clo', description='coiled-coil domain containing 151', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21989871, end_position_on_the_genomic_accession=22002972, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  77610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77610, CurrentID=0, Status='live', Symbol='C330002G04Rik', Aliases='', description='RIKEN cDNA C330002G04 gene', other_designations='', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=22996890, end_position_on_the_genomic_accession=23075906, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  77613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77613, CurrentID=0, Status='live', Symbol='Prss36', Aliases='C330007D15Rik', description='protease, serine 36', other_designations='polyserase-2 protein|polyserine protease 2|protease, serine, 36|serine protease 36', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127932638, end_position_on_the_genomic_accession=127946725, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  77619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77619, CurrentID=0, Status='live', Symbol='Prelid2', Aliases='1700003A01Rik, C330008K14Rik', description='PRELI domain containing 2', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=41875696, end_position_on_the_genomic_accession=41951194, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77622, CurrentID=0, Status='live', Symbol='Apex2', Aliases='C430040P13Rik, ape2', description='apurinic/apyrimidinic endonuclease 2', other_designations='AP endonuclease 2|APEX nuclease 2|apurinic-apyrimidinic endonuclease 2', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=150503629, end_position_on_the_genomic_accession=150588191, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  77626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77626, CurrentID=0, Status='live', Symbol='Smpd4', Aliases='4122402O22Rik, mKIAA1418', description='sphingomyelin phosphodiesterase 4', other_designations='nSMase-3|nSMase3|neutral membrane (neutral sphingomyelinase-3)|neutral sphingomyelinase 3|neutral sphingomyelinase III', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17619354, end_position_on_the_genomic_accession=17644830, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  77627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77627, CurrentID=0, Status='live', Symbol='Efcab6', Aliases='4931407K02Rik, 4932408N08Rik, bM150J22.2, mKIAA1672', description='EF-hand calcium binding domain 6', other_designations='DJ-1-binding protein|DJBP|bM150J22.2 (novel protein (4931407K02RIK, 4932408N08RIK))', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83866712, end_position_on_the_genomic_accession=84075025, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  77629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77629, CurrentID=0, Status='live', Symbol='Sphkap', Aliases='4930544G21Rik, A930009L15Rik, AI852220, SKIP, mKIAA1678', description='SPHK1 interactor, AKAP domain containing', other_designations='SPHK1-interactor and AKAP domain-containing protein', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=83255442, end_position_on_the_genomic_accession=83408388, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  77630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77630, CurrentID=0, Status='live', Symbol='Prdm8', Aliases='4930565F05Rik, PFM5', description='PR domain containing 8', other_designations='PR domain-containing protein 8', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=98167198, end_position_on_the_genomic_accession=98187454, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  77632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77632, CurrentID=0, Status='live', Symbol='Pramef12', Aliases='4930569K13Rik, AU023001, C86578', description='PRAME family member 12', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144391674, end_position_on_the_genomic_accession=144408467, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77634, CurrentID=0, Status='live', Symbol='Snapc3', Aliases='1810020H02Rik, 4930558A07Rik, 5031401C21Rik, AI414457, E030018J20Rik', description='small nuclear RNA activating complex, polypeptide 3', other_designations='SNAPc subunit 3|small nuclear RNA-activating complex polypeptide 3', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=83417701, end_position_on_the_genomic_accession=83467658, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  208715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208715, CurrentID=0, Status='live', Symbol='Hmgcs1', Aliases='B130032C06Rik', description='3-hydroxy-3-methylglutaryl-Coenzyme A synthase 1', other_designations='3-hydroxy-3-methylglutaryl coenzyme A synthase|HMG-CoA synthase', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119690351, end_position_on_the_genomic_accession=119708260, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  77644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77644, CurrentID=0, Status='live', Symbol='C330007P06Rik', Aliases='', description='RIKEN cDNA C330007P06 gene', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=36814763, end_position_on_the_genomic_accession=36864303, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  208718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208718, CurrentID=0, Status='live', Symbol='Dis3l2', Aliases='4930429A22Rik, 8030493P09Rik', description="DIS3 like 3'-5' exoribonuclease 2", other_designations='DIS3 mitotic control homolog (S. cerevisiae)-like 2', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86703779, end_position_on_the_genomic_accession=87050097, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  77647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77647, CurrentID=0, Status='live', Symbol='Trat1', Aliases='C030046M14Rik, Tcrim, Trim', description='T cell receptor associated transmembrane adaptor 1', other_designations='T cell receptor interacting molecule|T-cell receptor interacting molecule|T-cell receptor-interacting molecule', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48731914, end_position_on_the_genomic_accession=48810520, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  77652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77652, CurrentID=0, Status='live', Symbol='Zfp955a', Aliases='AI842447, Zfp883a', description='zinc finger protein 955A', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33239507, end_position_on_the_genomic_accession=33255169, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  208727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208727, CurrentID=0, Status='live', Symbol='Hdac4', Aliases='4932408F19Rik, HD4', description='histone deacetylase 4', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91928779, end_position_on_the_genomic_accession=92180385, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  668303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668303, CurrentID=0, Status='live', Symbol='Kif26a', Aliases='mKIAA1236', description='kinesin family member 26A', other_designations='N-11 kinesin', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112143968, end_position_on_the_genomic_accession=112181747, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  77669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77669, CurrentID=0, Status='live', Symbol='Arhgef38', Aliases='9130221D24Rik, D630013G24Rik', description='Rho guanine nucleotide exchange factor (GEF) 38', other_designations='putative Rho guanine nucleotide exchange factor FLJ20184 homolog', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=133159529, end_position_on_the_genomic_accession=133234889, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  77673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77673, CurrentID=0, Status='live', Symbol='Defb41', Aliases='9230102D03Rik, Bd-41, Defb16, Gm15386, OTTMUSG00000020946, OTTMUSG00000020948', description='defensin beta 41', other_designations='beta-defensin 16', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=18250978, end_position_on_the_genomic_accession=18265138, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  77674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77674, CurrentID=0, Status='live', Symbol='Defb12', Aliases='9230103N16Rik, BD-12', description='defensin beta 12', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19111931, end_position_on_the_genomic_accession=19114833, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  208748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208748, CurrentID=0, Status='live', Symbol='Prrg3', Aliases='Gm368', description='proline rich Gla (G-carboxyglutamic acid) 3 (transmembrane)', other_designations='proline-rich Gla protein 3|proline-rich gamma-carboxyglutamic acid protein 3', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=71962156, end_position_on_the_genomic_accession=71972722, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  77683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77683, CurrentID=0, Status='live', Symbol='Ehmt1', Aliases='9230102N17Rik, D330003E03, Eu-HMTase1, GLP, GLP1, KMT1D, mKIAA1876', description='euchromatic histone methyltransferase 1', other_designations='G9a-like protein 1|euchromatic histone-lysine N-methyltransferase 1|lysine N-methyltransferase 1D', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24790769, end_position_on_the_genomic_accession=24919643, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  77684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77684, CurrentID=0, Status='live', Symbol='Bsph2', Aliases='9230107M04Rika, Bsph2', description='binder of sperm protein homolog 2', other_designations='bovine seminal plasma protein-like 2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13554866, end_position_on_the_genomic_accession=13571067, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  208760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208760, CurrentID=0, Status='live', Symbol='Aqp12', Aliases='AB084104, mAQP12', description='aquaporin 12', other_designations='AQP-12', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93006334, end_position_on_the_genomic_accession=93012269, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  208768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208768, CurrentID=0, Status='live', Symbol='Sde2', Aliases='', description='SDE2 telomere maintenance homolog (S. pombe)', other_designations='UPF0667 protein C1orf55 homolog', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180851151, end_position_on_the_genomic_accession=180868113, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  77697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77697, CurrentID=0, Status='live', Symbol='Mmab', Aliases='9130222L19Rik, ATR', description='methylmalonic aciduria (cobalamin deficiency) type B homolog (human)', other_designations='ATP:Cob(I)alamin Adenosyltransferase|cob(I)alamin adenosyltransferase|methylmalonic aciduria type B homolog', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114431034, end_position_on_the_genomic_accession=114444073, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  77701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77701, CurrentID=0, Status='live', Symbol='Lcn12', Aliases='9230102M18Rik', description='lipocalin 12', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25490845, end_position_on_the_genomic_accession=25494061, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  77704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77704, CurrentID=0, Status='live', Symbol='Lcn9', Aliases='9230102I19Rik', description='lipocalin 9', other_designations='MUP-like lipocalin', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25823094, end_position_on_the_genomic_accession=25825542, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  77705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77705, CurrentID=0, Status='live', Symbol='9230104L09Rik', Aliases='mCST E2', description='RIKEN cDNA 9230104L09 gene', other_designations='Cystatin 12', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148844620, end_position_on_the_genomic_accession=148850962, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  77706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77706, CurrentID=0, Status='live', Symbol='Abcb5', Aliases='9230106F14Rik', description='ATP-binding cassette, sub-family B (MDR/TAP), member 5', other_designations='ABCB5 P-gp|P-glycoprotein ABCB5', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=118867486, end_position_on_the_genomic_accession=118966421, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  77721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77721, CurrentID=0, Status='live', Symbol='Mrps5', Aliases='1620401I16Rik, AI850294', description='mitochondrial ribosomal protein S5', other_designations='MRP-S5|S5mt', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127587426, end_position_on_the_genomic_accession=127606829, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  208795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208795, CurrentID=0, Status='live', Symbol='Tmem63a', Aliases='BC014795, Tmem64a', description='transmembrane protein 63a', other_designations='', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180941225, end_position_on_the_genomic_accession=180975106, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  77727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77727, CurrentID=0, Status='live', Symbol='6030468B19Rik', Aliases='', description='RIKEN cDNA 6030468B19 gene', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117797627, end_position_on_the_genomic_accession=117807308, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  77733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77733, CurrentID=0, Status='live', Symbol='Rnf170', Aliases='6720407G21Rik, AI481227', description='ring finger protein 170', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=26119368, end_position_on_the_genomic_accession=26143874, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  77739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77739, CurrentID=0, Status='live', Symbol='Adamtsl1', Aliases='5930437A14Rik, 6720426B09Rik, ADAMTSL-1', description='ADAMTS-like 1', other_designations='punctin-1', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=85513234, end_position_on_the_genomic_accession=86428390, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  77744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77744, CurrentID=0, Status='live', Symbol='Bora', Aliases='6720463M24Rik, AI317232', description='bora, aurora kinase A activator', other_designations='aurora borealis', map_location='14 E2.2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=99046247, end_position_on_the_genomic_accession=99076450, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  208820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208820, CurrentID=0, Status='live', Symbol='Triqk', Aliases='C130086A10, Gm11818, OTTMUSG00000004461', description='triple QxxK/R motif containing', other_designations='protein TRIQK|triple repetitive-sequence of QXXK/R protein homolog', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=12906837, end_position_on_the_genomic_accession=12981485, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  12960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12960, CurrentID=0, Status='live', Symbol='Crybb1', Aliases='3110006K12Rik, BB1CRY', description='crystallin, beta B1', other_designations='', map_location='5 54.63 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112254477, end_position_on_the_genomic_accession=112269582, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  77763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77763, CurrentID=0, Status='live', Symbol='Gm9736', Aliases='', description='predicted gene 9736', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77750514, end_position_on_the_genomic_accession=77751420, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  208836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208836, CurrentID=0, Status='live', Symbol='Fanci', Aliases='', description='Fanconi anemia, complementation group I', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79392312, end_position_on_the_genomic_accession=79450265, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  77766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77766, CurrentID=0, Status='live', Symbol='Elp4', Aliases='A330107A17Rik, Paxneb', description='elongator acetyltransferase complex subunit 4', other_designations='PAX6 neighbor gene protein|elongation protein 4 homolog', map_location='2 55.33 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=105700109, end_position_on_the_genomic_accession=105904500, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  77767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77767, CurrentID=0, Status='live', Symbol='Ermn', Aliases='A330104H05Rik, AI854460, ermin, mKIAA1189', description='ermin, ERM-like protein', other_designations='JN|juxtanodin', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=58045115, end_position_on_the_genomic_accession=58052752, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  77771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77771, CurrentID=0, Status='live', Symbol='Csrnp3', Aliases='A330102K23Rik, CSRNP-3, Csnrp3, Mbu1, taip-2', description='cysteine-serine-rich nuclear protein 3', other_designations='TGF-beta induced apoptosis protein 2|TGF-beta-induced apoptosis protein 2', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=65845767, end_position_on_the_genomic_accession=66031546, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  77772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77772, CurrentID=0, Status='live', Symbol='Dcst1', Aliases='A330106H01Rik', description='DC-STAMP domain containing 1', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89350235, end_position_on_the_genomic_accession=89365239, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  208846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208846, CurrentID=0, Status='live', Symbol='Daam1', Aliases='', description='dishevelled associated activator of morphogenesis 1', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=71831051, end_position_on_the_genomic_accession=71992376, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  77777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77777, CurrentID=0, Status='live', Symbol='Ulbp1', Aliases='A430108B07Rik, MULT1', description='UL16 binding protein 1', other_designations='UL16-binding protein-like transcript 1', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7444873, end_position_on_the_genomic_accession=7473637, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  77781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77781, CurrentID=0, Status='live', Symbol='Epm2aip1', Aliases='A930003G21Rik, mKIAA0766', description='EPM2A (laforin) interacting protein 1', other_designations='laforin-interacting protein', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111271845, end_position_on_the_genomic_accession=111279093, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  77782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77782, CurrentID=0, Status='live', Symbol='Polq', Aliases='A430110D14Rik', description='polymerase (DNA directed), theta', other_designations='DNA polymerase eta|chromosome aberrations occurring spontaneously protein 1', map_location='16 26.32 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37011742, end_position_on_the_genomic_accession=37095419, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  77794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77794, CurrentID=0, Status='live', Symbol='Adamtsl2', Aliases='A930008K15Rik, tcp-1', description='ADAMTS-like 2', other_designations='ADAMTSL-2|TSP1-repeat-containing protein 1|TSP1-repeats containing protein|TSP1-repeats-containing protein 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27079342, end_position_on_the_genomic_accession=27108613, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  208869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208869, CurrentID=0, Status='live', Symbol='Dock3', Aliases='MOCA, PBP, mKIAA0299', description='dedicator of cyto-kinesis 3', other_designations='modifier of cell adhesion|presenilin-binding protein', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106892825, end_position_on_the_genomic_accession=107231909, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  77799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77799, CurrentID=0, Status='live', Symbol='Sla2', Aliases='A930009E21Rik, AI430952, SLAP-2, SLAP2', description='Src-like-adaptor 2', other_designations='modulator of antigen receptor signaling MARS|src-like adapter protein 2', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156872922, end_position_on_the_genomic_accession=156887241, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  77803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77803, CurrentID=0, Status='live', Symbol='Fam159b', Aliases='A930021C24Rik', description='family with sequence similarity 159, member B', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104845283, end_position_on_the_genomic_accession=104863893, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  77805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77805, CurrentID=0, Status='live', Symbol='Esco1', Aliases='A930014I12Rik', description='establishment of sister chromatid cohesion N-acetyltransferase 1', other_designations='ECO1 homolog 1|establishment of cohesion 1 homolog 1|establisment of cohesion 1 homolog 1', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10566507, end_position_on_the_genomic_accession=10610685, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  77809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77809, CurrentID=0, Status='live', Symbol='Lrrc42', Aliases='A930011F22Rik, AA536996', description='leucine rich repeat containing 42', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107233514, end_position_on_the_genomic_accession=107253533, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  208884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208884, CurrentID=0, Status='live', Symbol='Zdhhc9', Aliases='6430508G22, 9530098M12Rik', description='zinc finger, DHHC domain containing 9', other_designations='DHHC-9|DHHC9|zinc finger DHHC domain-containing protein 9', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48171967, end_position_on_the_genomic_accession=48208702, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  208890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208890, CurrentID=0, Status='live', Symbol='Slc26a7', Aliases='', description='solute carrier family 26, member 7', other_designations='anion exchanger Slc26a7|slc26a7', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=14502430, end_position_on_the_genomic_accession=14621971, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  12970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12970, CurrentID=0, Status='live', Symbol='Crygs', Aliases='AI327013, Opj, rncat', description='crystallin, gamma S', other_designations='opacity due to poor secondary fiber cell junction|recessive nuclear cataract', map_location='16 13.66 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22805203, end_position_on_the_genomic_accession=22811410, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  208898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208898, CurrentID=0, Status='live', Symbol='Unc13c', Aliases='1500037O19Rik, AU019458, D9Ertd414e, Munc13-3, Unc13h3', description='unc-13 homolog C (C. elegans)', other_designations='unc13 homolog 3', map_location='9 40.8 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=73479422, end_position_on_the_genomic_accession=74008153, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  77827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77827, CurrentID=0, Status='live', Symbol='Krba1', Aliases='A930040G15Rik, AI448780', description='KRAB-A domain containing 1', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48395586, end_position_on_the_genomic_accession=48419855, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  77832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77832, CurrentID=0, Status='live', Symbol='Tchp', Aliases='A930031F18Rik, AV279199', description='trichoplein, keratin filament binding', other_designations='mitochondrial protein with oncostatic activity|mitostatin|tchp', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114707765, end_position_on_the_genomic_accession=114722327, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  77836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77836, CurrentID=0, Status='live', Symbol='Mlana', Aliases='A930034P04Rik, Mart1', description='melan-A', other_designations='melanoma associated antigen recognized by T-cells 1', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29697548, end_position_on_the_genomic_accession=29708447, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  208922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208922, CurrentID=0, Status='live', Symbol='Cpeb3', Aliases='4831444O18Rik, CPE-BP3, mKIAA0940', description='cytoplasmic polyadenylation element binding protein 3', other_designations='CPE-binding protein 3', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37019341, end_position_on_the_genomic_accession=37207619, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  77853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77853, CurrentID=0, Status='live', Symbol='Msl2', Aliases='E130103E02Rikl1, Rnf184, Msl2', description='male-specific lethal 2 homolog (Drosophila)', other_designations='MSL2-like 1|male-specific lethal 2-like 1|male-specific lethal-2 homolog 1|ring finger protein 184', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=101100570, end_position_on_the_genomic_accession=101104800, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  77862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77862, CurrentID=0, Status='live', Symbol='Thyn1', Aliases='D730042P09Rik, HSPC144, MDS012, MY105, Thy28', description='thymocyte nuclear protein 1', other_designations='mThy28|thymocyte protein thy28', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=26999659, end_position_on_the_genomic_accession=27007334, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  208936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208936, CurrentID=0, Status='live', Symbol='Adamts18', Aliases='9630038L21, ADAMTS21, E130314N14Rik', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 18', other_designations='ADAM-TS 18|ADAM-TS18|ADAMTS-18|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 18', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=113697123, end_position_on_the_genomic_accession=113855050, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  77866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77866, CurrentID=0, Status='live', Symbol='E130102H24Rik', Aliases='', description='RIKEN cDNA E130102H24 gene', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101346524, end_position_on_the_genomic_accession=101356248, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  78514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78514, CurrentID=0, Status='live', Symbol='Arhgap10', Aliases='A930033B01Rik, PSGAP-m, PSGAP-s', description='Rho GTPase activating protein 10', other_designations='PH and SH3 domain-containing rhoGAP protein|PS-GAP|PSGAP|rho-type GTPase-activating protein 10', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=77250364, end_position_on_the_genomic_accession=77517953, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  208943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208943, CurrentID=0, Status='live', Symbol='Myo5c', Aliases='9130003O20Rik, BB085560, C330029I24', description='myosin VC', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75231725, end_position_on_the_genomic_accession=75306263, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  77877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77877, CurrentID=0, Status='live', Symbol='6030458C11Rik', Aliases='', description='RIKEN cDNA 6030458C11 gene', other_designations='', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=12808177, end_position_on_the_genomic_accession=12824688, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  77883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77883, CurrentID=0, Status='live', Symbol='6030498E09Rik', Aliases='', description='RIKEN cDNA 6030498E09 gene', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38772660, end_position_on_the_genomic_accession=38969711, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  77889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77889, CurrentID=0, Status='live', Symbol='Lbh', Aliases='1810009F10Rik, 6720416L16Rik', description='limb-bud and heart', other_designations='limb bud and heart-expressed protein', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=72918305, end_position_on_the_genomic_accession=72941946, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  77891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77891, CurrentID=0, Status='live', Symbol='Ube2s', Aliases='0910001J09Rik, 6720465F12Rik, AA409170, E2-EPF', description='ubiquitin-conjugating enzyme E2S', other_designations='E2 ubiquitin-conjugating enzyme S|E2-EPF5|ubiquitin carrier protein S|ubiquitin-conjugating enzyme E2-24 kDa|ubiquitin-conjugating enzyme E2-EPF5|ubiquitin-protein ligase S', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4808014, end_position_on_the_genomic_accession=4812340, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  208967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208967, CurrentID=0, Status='live', Symbol='Thnsl1', Aliases='AW413632', description='threonine synthase-like 1 (bacterial)', other_designations='TSH1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=21205385, end_position_on_the_genomic_accession=21215009, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  208968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208968, CurrentID=0, Status='live', Symbol='Zfp280c', Aliases='BC028839, Suhw3, Zfp633, Znf280c, mKIAA1584', description='zinc finger protein 280C', other_designations='suppressor of hairy wing homolog 3', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48541626, end_position_on_the_genomic_accession=48594646, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  208982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208982, CurrentID=0, Status='live', Symbol='Hmgcll1', Aliases='BC037381', description='3-hydroxymethyl-3-methylglutaryl-Coenzyme A lyase-like 1', other_designations='3-hydroxymethyl-3-methylglutaryl-CoA lyase-like protein 1|probable 3-hydroxymethyl-3-methylglutaryl-CoA lyase 2', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=76014650, end_position_on_the_genomic_accession=76136350, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  208990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208990, CurrentID=0, Status='live', Symbol='Npb', Aliases='', description='neuropeptide B', other_designations='mPPL7|preproneuropeptide B|preproprotein L7', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120608477, end_position_on_the_genomic_accession=120609100, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  208994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208994, CurrentID=0, Status='live', Symbol='Fam83b', Aliases='C530008M07Rik, Gm516', description='family with sequence similarity 83, member B', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=76490054, end_position_on_the_genomic_accession=76567546, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  77929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77929, CurrentID=0, Status='live', Symbol='Yipf6', Aliases='A430107J06Rik', description='Yip1 domain family, member 6', other_designations='YIP1 family member 6', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=98936321, end_position_on_the_genomic_accession=98949021, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  209003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209003, CurrentID=0, Status='live', Symbol='Rbmx2', Aliases='2700089A17Rik, 2810411A03Rik, CGI-79', description='RNA binding motif protein, X-linked 2', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48695004, end_position_on_the_genomic_accession=48710719, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  209005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209005, CurrentID=0, Status='live', Symbol='Gm595', Aliases='', description='predicted gene 595', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48841466, end_position_on_the_genomic_accession=48877807, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  77938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77938, CurrentID=0, Status='live', Symbol='Fam53b', Aliases='A930008G19Rik, mKIAA0140', description='family with sequence similarity 53, member B', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132712078, end_position_on_the_genomic_accession=132813196, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  209011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209011, CurrentID=0, Status='live', Symbol='Sirt7', Aliases='', description='sirtuin 7', other_designations='NAD-dependent deacetylase sirtuin-7|SIR2-like protein 7|regulatory protein SIR2 homolog 7', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120618372, end_position_on_the_genomic_accession=120625002, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  209012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209012, CurrentID=0, Status='live', Symbol='Ulk4', Aliases='4932415A06Rik, A730098P15', description='unc-51-like kinase 4', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120960243, end_position_on_the_genomic_accession=121286186, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  77945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77945, CurrentID=0, Status='live', Symbol='Rpgrip1', Aliases='0610005A07Rik, 4930401L23Rik, 4930505G06Rik, A930002K18Rik, AA415034, nmf247', description='retinitis pigmentosa GTPase regulator interacting protein 1', other_designations='RPGR-interacting protein 1', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52104106, end_position_on_the_genomic_accession=52161339, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  209018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209018, CurrentID=0, Status='live', Symbol='Vps8', Aliases='AI315068, AU040738, mKIAA0804', description='vacuolar protein sorting 8 homolog (S. cerevisiae)', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21423118, end_position_on_the_genomic_accession=21644681, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  77951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77951, CurrentID=0, Status='live', Symbol='Cyp20a1', Aliases='A930011N14Rik, Cypm', description='cytochrome P450, family 20, subfamily a, polypeptide 1', other_designations='cytochrome P450 monooxygenase|cytochrome P450, 20a1', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60342800, end_position_on_the_genomic_accession=60388060, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  209027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209027, CurrentID=0, Status='live', Symbol='Pycr1', Aliases='', description='pyrroline-5-carboxylate reductase 1', other_designations='P5C reductase 1|P5CR 1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120635712, end_position_on_the_genomic_accession=120643782, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  209032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209032, CurrentID=0, Status='live', Symbol='Zc3hav1l', Aliases='B130055L09Rik, E430016P22Rik', description='zinc finger CCCH-type, antiviral 1-like', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38287394, end_position_on_the_genomic_accession=38299259, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  77963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77963, CurrentID=0, Status='live', Symbol='Hook1', Aliases='A930033L17Rik, azh', description='hook microtubule tethering protein 1', other_designations='abnormal spermatozoon head shape|hook homolog 1|mHK1', map_location='4 44.17 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=95967187, end_position_on_the_genomic_accession=96024277, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  209039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209039, CurrentID=0, Status='live', Symbol='Tns2', Aliases='C1-ten, Tenc1, nep, nph', description='tensin 2', other_designations='C1 domain-containing phosphatase and tensin homolog|C1 domain-containing phosphatase and tensin-like protein|tensin like C1 domain-containing phosphatase|tensin-like C1 domain-containing phosphatase', map_location='15 57.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102100367, end_position_on_the_genomic_accession=102116401, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  77974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77974, CurrentID=0, Status='live', Symbol='Rdh12', Aliases='', description='retinol dehydrogenase 12', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79208914, end_position_on_the_genomic_accession=79222661, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  77975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77975, CurrentID=0, Status='live', Symbol='Tmem50b', Aliases='AU015466, AU019872, B230114J08Rik', description='transmembrane protein 50B', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91574508, end_position_on_the_genomic_accession=91597680, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77976, CurrentID=0, Status='live', Symbol='Nuak1', Aliases='AU014801, AW494241, B230104P22Rik, Omphk1', description='NUAK family, SNF1-like kinase, 1', other_designations='AMPK-related protein kinase 5|ZNUAK family, SNF1-like kinase, 1|omphalocele kinase 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=84370900, end_position_on_the_genomic_accession=84440471, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  77980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77980, CurrentID=0, Status='live', Symbol='Sbf1', Aliases='2610510A08Rik, B230113C15Rik, Mtmr5, mKIAA3020', description='SET binding factor 1', other_designations='SET-binding factor 1', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89288236, end_position_on_the_genomic_accession=89315311, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  77987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77987, CurrentID=0, Status='live', Symbol='Ascc3', Aliases='ASC1p200, B630009I04Rik, BC023451, D430001L07Rik, D630041L21, Helic1, RNAH', description='activating signal cointegrator 1 complex subunit 3', other_designations='helicase, ATP binding 1', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=50592669, end_position_on_the_genomic_accession=50851204, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  77996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77996, CurrentID=0, Status='live', Symbol='Cutal', Aliases='D730039F16Rik', description='cutA divalent cation tolerance homolog-like', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34874365, end_position_on_the_genomic_accession=34892133, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  77998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77998, CurrentID=0, Status='live', Symbol='Grifin', Aliases='E130101O17Rik', description='galectin-related inter-fiber protein', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140563117, end_position_on_the_genomic_accession=140567628, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  78004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78004, CurrentID=0, Status='live', Symbol='Prr15', Aliases='E130201N16Rik, G90', description='proline rich 15', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54327012, end_position_on_the_genomic_accession=54330200, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  13001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13001, CurrentID=0, Status='live', Symbol='Csnk2b', Aliases='CK II beta', description='casein kinase 2, beta polypeptide', other_designations='phosvitin', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35116195, end_position_on_the_genomic_accession=35122053, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  72097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=72097, CurrentID=0, Status='live', Symbol='2010300C02Rik', Aliases='', description='RIKEN cDNA 2010300C02 gene', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37611676, end_position_on_the_genomic_accession=37720092, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  13002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13002, CurrentID=0, Status='live', Symbol='Dnajc5', Aliases='2610314I24Rik, AU018536, Csp', description='DnaJ heat shock protein family (Hsp40) member C5', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 5|cysteine string protein', map_location='2 103.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181520305, end_position_on_the_genomic_accession=181552886, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  209086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209086, CurrentID=0, Status='live', Symbol='Samd9l', Aliases='AA175286, ESTM25, mKIAA2005', description='sterile alpha motif domain containing 9-like', other_designations='SAM domain-containing protein 9-like|v-myc myelocytomatosis viral related oncogene, neuroblastoma derived', map_location='6 A1-A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3372257, end_position_on_the_genomic_accession=3399571, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  78016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78016, CurrentID=0, Status='live', Symbol='Ccdc150', Aliases='4930511H11Rik', description='coiled-coil domain containing 150', other_designations='', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=54250661, end_position_on_the_genomic_accession=54368727, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  209091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209091, CurrentID=0, Status='live', Symbol='Ccnb3', Aliases='', description='cyclin B3', other_designations='', map_location='X A2-A3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=6979652, end_position_on_the_genomic_accession=7041635, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  13004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13004, CurrentID=0, Status='live', Symbol='Ncan', Aliases='C230035B04, Cspg3, Cspg3-rs, Tgfbit', description='neurocan', other_designations='TGFbeta-induced transcript, brain|chondroitin sulfate proteoglycan 3, related sequence', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70093082, end_position_on_the_genomic_accession=70120870, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  78038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78038, CurrentID=0, Status='live', Symbol='Mccc2', Aliases='4930552N12Rik, MCCB', description='methylcrotonoyl-Coenzyme A carboxylase 2 (beta)', other_designations='3-methylcrotonyl-CoA carboxylase 2|3-methylcrotonyl-CoA carboxylase non-biotin-containing subunit|3-methylcrotonyl-CoA:carbon dioxide ligase subunit beta|MCCase subunit beta', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=99948532, end_position_on_the_genomic_accession=100015639, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  78052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78052, CurrentID=0, Status='live', Symbol='Tmem190', Aliases='4930572D21Rik', description='transmembrane protein 190', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4781776, end_position_on_the_genomic_accession=4784491, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  209131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209131, CurrentID=0, Status='live', Symbol='Snx30', Aliases='4732481H14Rik, C030041J06Rik', description='sorting nexin family member 30', other_designations='novel PX domain containing protein', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59805126, end_position_on_the_genomic_accession=59904740, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  78070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78070, CurrentID=0, Status='live', Symbol='Cpt1c', Aliases='6530437J22Rik, 9630004I06Rik', description='carnitine palmitoyltransferase 1c', other_designations='CPT I-C|CPT IC|CPT1-B|CPTI-B|carnitine O-palmitoyltransferase I, brain isoform|carnitine palmitoyltransferase 1, brain', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44958483, end_position_on_the_genomic_accession=44974851, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  78076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78076, CurrentID=0, Status='live', Symbol='Lcn8', Aliases='9230106L18Rik, EP17, Lcn5, mEP17', description='lipocalin 8', other_designations='epididymal 17 kDa lipocalin|lipocalin 5', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25653118, end_position_on_the_genomic_accession=25656217, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  78081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78081, CurrentID=0, Status='live', Symbol='Crisp4', Aliases='9230112K08Rik', description='cysteine-rich secretory protein 4', other_designations='', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=18114905, end_position_on_the_genomic_accession=18145967, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78088, CurrentID=0, Status='live', Symbol='Sowahb', Aliases='5730467H21Rik, Ankrd56', description='sosondowah ankyrin repeat domain family member B', other_designations='ankyrin repeat domain 56|ankyrin repeat domain-containing protein 56|protein sosondowah homolog B', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=93041123, end_position_on_the_genomic_accession=93045022, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  78092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78092, CurrentID=0, Status='live', Symbol='Samt1', Aliases='Gm15144, OTTMUSG00000019268', description='spermatogenesis associated multipass transmembrane protein 1', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=154001590, end_position_on_the_genomic_accession=154012643, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  78100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78100, CurrentID=0, Status='live', Symbol='Msantd4', Aliases='8430410K20Rik, Mysantd4', description='Myb/SANT-like DNA-binding domain containing 4 with coiled-coils', other_designations='', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=4376413, end_position_on_the_genomic_accession=4386890, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  209176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209176, CurrentID=0, Status='live', Symbol='Ido2', Aliases='AI265623, C230043N17Rik, Ido-2, Indol1', description='indoleamine 2,3-dioxygenase 2', other_designations='indoleamine 2,3-dioxygenase-like protein 1|indoleamine-pyrrole 2,3 dioxygenase-like 1', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24531892, end_position_on_the_genomic_accession=24576404, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  209186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209186, CurrentID=0, Status='live', Symbol='Acnat2', Aliases='C730036D15Rik', description='acyl-coenzyme A amino acid N-acyltransferase 2', other_designations='acnat2|acyl-CoA:amino acid N-acyltransferase 2', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49375341, end_position_on_the_genomic_accession=49408151, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  668379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668379, CurrentID=0, Status='live', Symbol='Scgb2b12', Aliases='Abpbg12, EG668379, Gm9138', description='secretoglobin, family 2B, member 12', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=32325286, end_position_on_the_genomic_accession=32327245, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  78118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78118, CurrentID=0, Status='live', Symbol='4930451I11Rik', Aliases='', description='RIKEN cDNA 4930451I11 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126830468, end_position_on_the_genomic_accession=126831540, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  209195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209195, CurrentID=0, Status='live', Symbol='Clic6', Aliases='5730466J16Rik, AL022908, AW045520, CLIC1L', description='chloride intracellular channel 6', other_designations='', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92498147, end_position_on_the_genomic_accession=92541241, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  78124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78124, CurrentID=0, Status='live', Symbol='Spata31', Aliases='4930458L03Rik, Aep1, Fam75aa, Vad1.3, Spata31', description='spermatogenesis associated 31', other_designations='acrosome expressed protein 1|acrosome-expressed protein 1|acrosome-specific protein VAD1.3|family with sequence similarity 75, member A', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64917406, end_position_on_the_genomic_accession=64923195, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78128, CurrentID=0, Status='live', Symbol='Spag11a', Aliases='9230111C08Rik, Bin1b, EP2Q, EP2e, Spag11', description='sperm associated antigen 11A', other_designations='antimicrobial peptide Bin1b|beta defensin EP2e|sperm associated antigen 11', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19157214, end_position_on_the_genomic_accession=19159920, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78558, CurrentID=0, Status='live', Symbol='Htra3', Aliases='2210021K23Rik, 9530081K03Rik, Prsp, Tasp', description='HtrA serine peptidase 3', other_designations='high-temperature requirement factor A3|pregnancy-related serine protease|probable serine protease HTRA3|toll-associated serine protease', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35652026, end_position_on_the_genomic_accession=35679813, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  78134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78134, CurrentID=0, Status='live', Symbol='Lpar4', Aliases='5730485F04Rik, Gpr23, LPA4, p2y9', description='lysophosphatidic acid receptor 4', other_designations='G protein-coupled receptor 23|G-protein coupled receptor 23|LPA receptor 4|LPA-4|P2Y purinoceptor 9|purinergic receptor 9', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106920625, end_position_on_the_genomic_accession=106933899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  209212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209212, CurrentID=0, Status='live', Symbol='Osgin2', Aliases='BC010311, C230027H09Rik', description='oxidative stress induced growth inhibitor family member 2', other_designations='', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=15997121, end_position_on_the_genomic_accession=16014654, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  209224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209224, CurrentID=0, Status='live', Symbol='Enox2', Aliases='APK1, Cova1, tNOX', description='ecto-NOX disulfide-thiol exchanger 2', other_designations='APK1 antigen|cytosolic ovarian carcinoma antigen 1|tumor-associated hydroquinone oxidase', map_location='X A4-A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=49009707, end_position_on_the_genomic_accession=49288245, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  209225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209225, CurrentID=0, Status='live', Symbol='Zfp710', Aliases='5430400N05Rik, 9830165K13, AI324798, Znf710', description='zinc finger protein 710', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80024528, end_position_on_the_genomic_accession=80092751, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  209232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209232, CurrentID=0, Status='live', Symbol='Wfdc5', Aliases='BC019734, Prg5, Wap1', description='WAP four-disulfide core domain 5', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164176325, end_position_on_the_genomic_accession=164182742, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  209239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209239, CurrentID=0, Status='live', Symbol='Gan', Aliases='A330045G18, gigaxonin', description='giant axonal neuropathy', other_designations='gan', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=117158135, end_position_on_the_genomic_accession=117205186, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  78174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78174, CurrentID=0, Status='live', Symbol='Cox7b2', Aliases='4930503B16Rik', description='cytochrome c oxidase subunit VIIb2', other_designations='', map_location='5 C3.1-C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=71442823, end_position_on_the_genomic_accession=71548205, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  78177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78177, CurrentID=0, Status='live', Symbol='Ninl', Aliases='4930519N13Rik, Gm1004, Gm1634, mKIAA0980', description='ninein-like', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150934519, end_position_on_the_genomic_accession=151009398, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  209268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209268, CurrentID=0, Status='live', Symbol='Igsf1', Aliases='5330413N23, 5530402E03, AI747649, InhBP/p120, mKIAA0364', description='immunoglobulin superfamily, member 1', other_designations='betaglycan and inhibin binding protein|inhBP|inhibin-binding protein', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=49782536, end_position_on_the_genomic_accession=49797906, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  668395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668395, CurrentID=0, Status='live', Symbol='Ighv5-12', Aliases='Gm16592', description='immunoglobulin heavy variable 5-12', other_designations='immunoglobulin heavy chain variable region|immunoglobulin heavy variable V5-12', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113702126, end_position_on_the_genomic_accession=113702419, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  78217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78217, CurrentID=0, Status='live', Symbol='Tmem210', Aliases='4930571C24Rik', description='transmembrane protein 210', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25288145, end_position_on_the_genomic_accession=25289187, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  209294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209294, CurrentID=0, Status='live', Symbol='Csta1', Aliases='Csta', description='cystatin A1', other_designations='cystatin A|cystatin-A1|stefin-A', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36119946, end_position_on_the_genomic_accession=36131189, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  78232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78232, CurrentID=0, Status='live', Symbol='Trappc6b', Aliases='5830498C14Rik, C79212', description='trafficking protein particle complex 6B', other_designations='', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=59043092, end_position_on_the_genomic_accession=59061489, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  78239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78239, CurrentID=0, Status='live', Symbol='Spint4', Aliases='9230105I15Rik', description='serine protease inhibitor, Kunitz type 4', other_designations='epididymal trypsin inhibitor Spint4', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164698501, end_position_on_the_genomic_accession=164702448, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78240, CurrentID=0, Status='live', Symbol='Cst11', Aliases='9230101F08Rik, BB067629, CRES2, mCSTE1', description='cystatin 11', other_designations='cystatin E1|cystatin-related epididymal spermatogenic protein 2|mCST E1', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148768618, end_position_on_the_genomic_accession=148771497, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  78242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78242, CurrentID=0, Status='live', Symbol='Spink12', Aliases='9230117E20Rik', description='serine peptidase inhibitor, Kazal type 12', other_designations='serine peptidase inhibitor, Kazal type 11|serine protease inhibitor Kazal type 12', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44104523, end_position_on_the_genomic_accession=44108543, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  68073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68073, CurrentID=0, Status='live', Symbol='Fam173b', Aliases='A930016P21Rik, AA987072', description='family with sequence similarity 173, member B', other_designations='', map_location='15 B2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=31601941, end_position_on_the_genomic_accession=31617537, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  78244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78244, CurrentID=0, Status='live', Symbol='Dnajc21', Aliases='4930461P20Rik, 9930116P15Rik', description='DnaJ heat shock protein family (Hsp40) member C21', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 21', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10446763, end_position_on_the_genomic_accession=10470516, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  78245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78245, CurrentID=0, Status='live', Symbol='Acbd7', Aliases='9230116B18Rik', description='acyl-Coenzyme A binding domain containing 7', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3336168, end_position_on_the_genomic_accession=3340997, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78246, CurrentID=0, Status='live', Symbol='Phf23', Aliases='JUNE-1', description='PHD finger protein 23', other_designations='PHD zinc finger containing protein JUNE1|novel PHD-finger protein JUNE1', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69995766, end_position_on_the_genomic_accession=70000011, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  78248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78248, CurrentID=0, Status='live', Symbol='Armcx1', Aliases='3010033I09Rik, ALEX1', description='armadillo repeat containing, X-linked 1', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134717912, end_position_on_the_genomic_accession=134721917, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  78249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78249, CurrentID=0, Status='live', Symbol='Adgrf4', Aliases='4632435A09Rik, AV239010, Gpr115', description='adhesion G protein-coupled receptor F4', other_designations='G protein-coupled receptor 115|G-protein coupled receptor 115|probable G-protein coupled receptor 115', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=42656887, end_position_on_the_genomic_accession=42692284, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78250, CurrentID=0, Status='live', Symbol='Iqch', Aliases='4921504K03Rik', description='IQ motif containing H', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=63421015, end_position_on_the_genomic_accession=63602547, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  78251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78251, CurrentID=0, Status='live', Symbol='Zfp712', Aliases='4921504N20Rik, mszf31, mszf89', description='zinc finger protein 712', other_designations='', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67038594, end_position_on_the_genomic_accession=67061170, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  78252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78252, CurrentID=0, Status='live', Symbol='Nxpe2', Aliases='4432416J03Rik, AI323362, Fam55b', description='neurexophilin and PC-esterase domain family, member 2', other_designations='family with sequence similarity 55, member B|protein FAM55B', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48318003, end_position_on_the_genomic_accession=48353453, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  78255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78255, CurrentID=0, Status='live', Symbol='Ralgps2', Aliases='1810020P17Rik, 2210408F11Rik, 4921528G01Rik, 9130014M22Rik, AU043409, AW046161', description='Ral GEF with PH domain and SH3 binding motif 2', other_designations='ral GEF with PH domain and SH3-binding motif 2|ralA exchange factor RalGPS2', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156804166, end_position_on_the_genomic_accession=156941508, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  78257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78257, CurrentID=0, Status='live', Symbol='Lrrc9', Aliases='4921529O18Rik, 4930432K16Rik', description='leucine rich repeat containing 9', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72433940, end_position_on_the_genomic_accession=72530174, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  78258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78258, CurrentID=0, Status='live', Symbol='Mroh9', Aliases='4921528O07Rik, Armc11', description='maestro heat-like repeat family member 9', other_designations='armadillo repeat containing 11', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=163024302, end_position_on_the_genomic_accession=163085670, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  209334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209334, CurrentID=0, Status='live', Symbol='Gen1', Aliases='5830483C08Rik', description="GEN1, Holliday junction 5' flap endonuclease", other_designations='Gen homolog 1, endonuclease', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=11239718, end_position_on_the_genomic_accession=11266029, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  78266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78266, CurrentID=0, Status='live', Symbol='Zfp687', Aliases='4931408L03Rik, Znf687, mKIAA1441', description='zinc finger protein 687', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95006702, end_position_on_the_genomic_accession=95015452, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  78267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78267, CurrentID=0, Status='live', Symbol='Klhdc8b', Aliases='4931406O17Rik', description='kelch domain containing 8B', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108447636, end_position_on_the_genomic_accession=108461581, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  209351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209351, CurrentID=0, Status='live', Symbol='Wfdc6a', Aliases='Gm122', description='WAP four-disulfide core domain 6A', other_designations='likely orthologue of H. sapiens WAP four-disulfide core domain 6 (WFDC6)|putative protease inhibitor WAP6A', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164579519, end_position_on_the_genomic_accession=164585845, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  209354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209354, CurrentID=0, Status='live', Symbol='Eif2b1', Aliases='D5Ertd406e, EIF2BA', description='eukaryotic translation initiation factor 2B, subunit 1 (alpha)', other_designations='eIF-2B GDP-GTP exchange factor subunit alpha|eukaryotic translation initiation factor 2B, subunit 1 (alpha, 26kDa)', map_location='5 63.67 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124570214, end_position_on_the_genomic_accession=124579057, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78283, CurrentID=0, Status='live', Symbol='Map7d2', Aliases='1600028E09Rik, 2900002G04Rik, 5330432J06Rik, C030036K04Rik, Mtap7d2', description='MAP7 domain containing 2', other_designations='novel E-MAP-115 family protein', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=159414527, end_position_on_the_genomic_accession=159498961, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  78284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78284, CurrentID=0, Status='live', Symbol='Creb3l4', Aliases='1700012K17Rik, 5330432F22Rik, AIBZIP, AV040530, AV258827, CREB4, JAL, TISP40, acre1', description='cAMP responsive element binding protein 3-like 4', other_designations='androgen-induced basic leucine zipper|attaching to CRE-like 1|cAMP-responsive element-binding protein 3-like protein 4|transcript induced in spermiogenesis protein 40', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90237500, end_position_on_the_genomic_accession=90243524, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  209357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209357, CurrentID=0, Status='live', Symbol='Gtf2h3', Aliases='34kDa, 5033417D07Rik, BTF2, BTF2 p34, C730029A10, D5Ertd679e, TFIIH', description='general transcription factor IIH, polypeptide 3', other_designations='BTF2-p34|TFIIH basal transcription factor complex p34 subunit|basic transcription factor 2 34 kDa subunit|general transcription factor IIH, polypeptide 3, 34kDa', map_location='5 63.67 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124579148, end_position_on_the_genomic_accession=124597680, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78286, CurrentID=0, Status='live', Symbol='Nav2', Aliases='5330421F07Rik, AU067780, BQ960892, HELAD1, POMFIL2, RAINB1, RAINB2, Unc53H2, mKIAA3015', description='neuron navigator 2', other_designations='retinoic acid inducible in neuroblastoma', map_location='7 B3-B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48959066, end_position_on_the_genomic_accession=49610088, orientation='plus', exon_count=51, OMIM='', no_hdr0=''),
  78287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78287, CurrentID=0, Status='live', Symbol='Rbsn', Aliases='5330426D11Rik, Zfyve20', description='rabenosyn, RAB effector', other_designations='FYVE finger-containing Rab5 effector protein rabenosyn-5|RAB effector RBSN|zinc finger FYVE domain-containing protein 20|zinc finger, FYVE domain containing 20', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=92186712, end_position_on_the_genomic_accession=92214901, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  209361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209361, CurrentID=0, Status='live', Symbol='Taf3', Aliases='140kDa, 4933439M23Rik, AW539625, TAF140, TAFII-140, TAFII140, mTAFII140', description='TATA-box binding protein associated factor 3', other_designations='140 kDa TATA box-binding protein-associated factor|RNA polymerase II transcription factor, TFIID subunit|TAF(II)140|TAF3 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF3 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 140kDa|TBP-associated factor 3|transcription initiation factor TFIID 140 kDa subunit', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=9914552, end_position_on_the_genomic_accession=10048609, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  78294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78294, CurrentID=0, Status='live', Symbol='Rps27a', Aliases='0610006J14Rik, Uba52, Ubb, Ubc', description='ribosomal protein S27A', other_designations='ubiquitin carboxyl extension protein 80', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29545842, end_position_on_the_genomic_accession=29548040, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  78303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78303, CurrentID=0, Status='live', Symbol='Hist3h2ba', Aliases='1500011O09Rik, AI413321', description='histone cluster 3, H2ba', other_designations='histone 3, H2ba', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58948911, end_position_on_the_genomic_accession=58949372, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  78304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78304, CurrentID=0, Status='live', Symbol='Naa38', Aliases='1500034E06Rik, Lsmd1', description='N(alpha)-acetyltransferase 38, NatC auxiliary subunit', other_designations='LSM domain containing 1|LSM domain-containing protein 1|MAK31-like protein|N-alpha-acetyltransferase 38, NatC auxiliary subunit', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69395791, end_position_on_the_genomic_accession=69396671, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  209378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209378, CurrentID=0, Status='live', Symbol='Itih5', Aliases='4631408O11Rik, 5430408M01Rik, E130106B02', description='inter-alpha (globulin) inhibitor H5', other_designations='ITI heavy chain H5|ITI-HC5|inter-alpha trypsin inhibitor heavy chain precursor 5|inter-alpha-inhibitor heavy chain 5', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10153536, end_position_on_the_genomic_accession=10256529, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  78308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78308, CurrentID=0, Status='live', Symbol='Gpr108', Aliases='1810015L19Rik, AA589464, C79132, Lustr2', description='G protein-coupled receptor 108', other_designations='lung seven transmembrane receptor 2', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57234914, end_position_on_the_genomic_accession=57248441, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  78309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78309, CurrentID=0, Status='live', Symbol='Cul9', Aliases='1810035I07Rik, Cul-9, Parc, mKIAA0708', description='cullin 9', other_designations='p53-associated parkin-like cytoplasmic protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46500609, end_position_on_the_genomic_accession=46546388, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  207740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207740, CurrentID=0, Status='live', Symbol='Ubald1', Aliases='1500031H01Rik, BC013706, Fam100a', description='UBA-like domain containing 1', other_designations='family with sequence similarity 100, member A|protein FAM100A', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4874779, end_position_on_the_genomic_accession=4879851, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  209387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209387, CurrentID=0, Status='live', Symbol='Trim30d', Aliases='AI451617, TRIM30-3, Trim79', description='tripartite motif-containing 30D', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104470014, end_position_on_the_genomic_accession=104507849, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  78317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78317, CurrentID=0, Status='live', Symbol='Ccdc88b', Aliases='2610041P08Rik, Ccdc88', description='coiled-coil domain containing 88B', other_designations='coiled-coil domain containing 88|hkRP3|hook-related protein 3', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6844623, end_position_on_the_genomic_accession=6858328, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  78321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78321, CurrentID=0, Status='live', Symbol='Ankrd23', Aliases='1110058D09Rik, 2310041C05Rik, AI853437, Darp, MARP3', description='ankyrin repeat domain 23', other_designations='diabetes-realted ankyrin repeat protein|diabetes-related ankyrin repeat protein', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36530188, end_position_on_the_genomic_accession=36536883, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78323, CurrentID=0, Status='live', Symbol='Fam219b', Aliases='2310046O06Rik, AI594866', description='family with sequence similarity 219, member B', other_designations='uncharacterized protein C15orf17 homolog', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57537528, end_position_on_the_genomic_accession=57543188, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  78330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78330, CurrentID=0, Status='live', Symbol='Ndufv3', Aliases='1500032D16Rik', description='NADH dehydrogenase (ubiquinone) flavoprotein 3', other_designations='CI-9kD|NADH-ubiquinone oxidoreductase 9 kDa subunit|NADH-ubiquinone oxidoreductase flavoprotein 3|complex I-9kD', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31520115, end_position_on_the_genomic_accession=31531326, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78334, CurrentID=0, Status='live', Symbol='Cdk19', Aliases='2700084L06Rik, AI845279, AW228747, CDK11, Cdc2l6, mKIAA1028', description='cyclin-dependent kinase 19', other_designations='CDC2-related protein kinase 6|cell division cycle 2-like 6 (CDK8-like)|cell division cycle 2-like protein kinase 6|cell division protein kinase 19|cyclin-dependent kinase (CDC2-like) 11|cyclin-dependent kinase 11', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40349308, end_position_on_the_genomic_accession=40483818, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78339, CurrentID=0, Status='live', Symbol='Ttyh3', Aliases='2900029G13Rik, AI414930, TTY3, mTTY3', description='tweety homolog 3 (Drosophila)', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140620578, end_position_on_the_genomic_accession=140649031, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  209416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209416, CurrentID=0, Status='live', Symbol='Gpkow', Aliases='4432411B18, DXErtd697e, DXImx41e, T54', description='G patch domain and KOW motifs', other_designations='protein MOS2 homolog', map_location='X 3.47 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7697134, end_position_on_the_genomic_accession=7710261, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  78353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78353, CurrentID=0, Status='live', Symbol='2500002B13Rik', Aliases='', description='RIKEN cDNA 2500002B13 gene', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=57487963, end_position_on_the_genomic_accession=57510262, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  78354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78354, CurrentID=0, Status='live', Symbol='2210407C18Rik', Aliases='Ep1', description='RIKEN cDNA 2210407C18 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58608204, end_position_on_the_genomic_accession=58613492, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  78365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78365, CurrentID=0, Status='live', Symbol='Lhx1os', Aliases='1500016L03Rik, Lhx1as', description='LIM homeobox 1, opposite strand', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84525660, end_position_on_the_genomic_accession=84535831, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78369, CurrentID=0, Status='live', Symbol='Icam4', Aliases='1810015M19Rik, Cd242, ICAM-4', description='intercellular adhesion molecule 4, Landsteiner-Wiener blood group', other_designations='LW glycoprotein', map_location='9 7.69 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21029373, end_position_on_the_genomic_accession=21030531, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78372, CurrentID=0, Status='live', Symbol='Snrnp25', Aliases='3300001G02Rik, AL033324, C16ORF33', description='small nuclear ribonucleoprotein 25 (U11/U12)', other_designations='U11/U12 snRNP 25 kDa protein', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32205415, end_position_on_the_genomic_accession=32208996, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  78373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78373, CurrentID=0, Status='live', Symbol='Nudt17', Aliases='2410015C20Rik', description='nudix (nucleoside diphosphate linked moiety X)-type motif 17', other_designations='nudix motif 17', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96705892, end_position_on_the_genomic_accession=96708929, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  209446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209446, CurrentID=0, Status='live', Symbol='Tfe3', Aliases='F830016E06Rik, Tcfe3, Tfe-3, bHLHe33', description='transcription factor E3', other_designations='', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7762538, end_position_on_the_genomic_accession=7775202, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78376, CurrentID=0, Status='live', Symbol='Sapcd1', Aliases='2810436B06Rik, G7d, Ng23', description='suppressor APC domain containing 1', other_designations='protein G7d', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35025959, end_position_on_the_genomic_accession=35028571, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  209456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209456, CurrentID=0, Status='live', Symbol='Trp53bp2', Aliases='53BP2, AI746547, ASPP2, PPP1R13A, Tp53bp2, X98550', description='transformation related protein 53 binding protein 2', other_designations='p53-binding protein 2|p53BP2|tumor protein p53 binding protein, 2|tumor suppressor p53-binding protein 2', map_location='1 84.93 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182409167, end_position_on_the_genomic_accession=182462436, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  78388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78388, CurrentID=0, Status='live', Symbol='Mvp', Aliases='2310009M24Rik, LRP, VAULT1', description='major vault protein', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126986860, end_position_on_the_genomic_accession=127014594, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  78390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78390, CurrentID=0, Status='live', Symbol='Pla2g4d', Aliases='2610311B01Rik, Pla2delta, cPLA2delta', description='phospholipase A2, group IVD', other_designations='cPLA2-delta', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120265866, end_position_on_the_genomic_accession=120294260, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  78394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78394, CurrentID=0, Status='live', Symbol='Ddx52', Aliases='2700029C06Rik, ROK1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 52', other_designations='DEAD box protein 52|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 52', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83942090, end_position_on_the_genomic_accession=83963088, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  78405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78405, CurrentID=0, Status='live', Symbol='Ntf5', Aliases='2900040K06Rik, AI462899, NT-4, NT-5, NT4, NT4/5, Ntf-5, Ntf4', description='neurotrophin 5', other_designations='neurotrophin-5|neutrophic factor 4', map_location='7 29.3 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45413695, end_position_on_the_genomic_accession=45417179, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  209478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209478, CurrentID=0, Status='live', Symbol='Tbc1d12', Aliases='BC033574', description='TBC1D12: TBC1 domain family, member 12', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38836579, end_position_on_the_genomic_accession=38919923, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  78408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78408, CurrentID=0, Status='live', Symbol='Fam131a', Aliases='2900046G09Rik', description='family with sequence similarity 131, member A', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20693096, end_position_on_the_genomic_accession=20703036, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  78412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78412, CurrentID=0, Status='live', Symbol='3110062M04Rik', Aliases='AL023049, Mri', description='RIKEN cDNA 3110062M04 gene', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34871771, end_position_on_the_genomic_accession=34879019, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  78414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78414, CurrentID=0, Status='live', Symbol='1700034E13Rik', Aliases='', description='RIKEN cDNA 1700034E13 gene', other_designations='', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=52646185, end_position_on_the_genomic_accession=52663734, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  209488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209488, CurrentID=0, Status='live', Symbol='Hsh2d', Aliases='ALX, Hsh2', description='hematopoietic SH2 domain containing', other_designations='adaptor in lymphocytes of unknown function X|adaptor protein ALx2|hematopoietic SH2 protein', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72189668, end_position_on_the_genomic_accession=72200958, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  209497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209497, CurrentID=0, Status='live', Symbol='Tmem164', Aliases='AI316850, AI852450, AW547186, F730011B02', description='transmembrane protein 164', other_designations='', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=142680347, end_position_on_the_genomic_accession=142843494, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78428, CurrentID=0, Status='live', Symbol='Pym1', Aliases='A030010B05Rik, Pym, Wibg', description='PYM homolog 1, exon junction complex associated factor', other_designations='protein wibg homolog|within bgcn homolog', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128747879, end_position_on_the_genomic_accession=128766568, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  209512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209512, CurrentID=0, Status='live', Symbol='Taar2', Aliases='Gm224, Gm225, Gpr58, Tarr2', description='trace amine-associated receptor 2', other_designations='G protein-coupled receptor 58|G-protein coupled receptor 58|mTaar2|taR-2|trace amine associated receptor 2|trace amine receptor 2', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23938525, end_position_on_the_genomic_accession=23960552, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  209513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209513, CurrentID=0, Status='live', Symbol='Taar4', Aliases='Gm226, Ta2, taR-4', description='trace amine-associated receptor 4', other_designations='2-phenylethylamine receptor|mTaar4|trace amine associated receptor 4|trace amine receptor 2|trace amine receptor 4', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23960494, end_position_on_the_genomic_accession=23961537, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  78444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78444, CurrentID=0, Status='live', Symbol='Pgpep1l', Aliases='C330024D12Rik', description='pyroglutamyl-peptidase I-like', other_designations='', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=68236450, end_position_on_the_genomic_accession=68264484, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  209517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209517, CurrentID=0, Status='live', Symbol='Taar7b', Aliases='Gm698, taR-7b', description='trace amine-associated receptor 7B', other_designations='mTaar7b', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23999939, end_position_on_the_genomic_accession=24001015, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  78455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78455, CurrentID=0, Status='live', Symbol='Helz', Aliases='3110078M01Rik, 9430093I07Rik, 9630002H22Rik, AI851979, DRHC', description='helicase with zinc finger domain', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=107546958, end_position_on_the_genomic_accession=107693826, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  78459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78459, CurrentID=0, Status='live', Symbol='1700057G04Rik', Aliases='', description='RIKEN cDNA 1700057G04 gene', other_designations='novel Plsrc protein', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=92309327, end_position_on_the_genomic_accession=92357876, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  78465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78465, CurrentID=0, Status='live', Symbol='Ccdc190', Aliases='1700084C01Rik', description='coiled-coil domain containing 190', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=169928597, end_position_on_the_genomic_accession=169935318, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  209540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209540, CurrentID=0, Status='live', Symbol='Rgag1', Aliases='Gm385, Mar9, Mart9', description='retrotransposon gag domain containing 1', other_designations='sushi-XF2', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=143057943, end_position_on_the_genomic_accession=143105330, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  78469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78469, CurrentID=0, Status='live', Symbol='Tmem247', Aliases='1700090G07Rik', description='transmembrane protein 247', other_designations='transmembrane protein ENSP00000343375 homolog', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=86917348, end_position_on_the_genomic_accession=86922482, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78473, CurrentID=0, Status='live', Symbol='Skap1', Aliases='1700091G21Rik, Scap1, Skap-55', description='src family associated phosphoprotein 1', other_designations='ortholog of human src family associated phosphoprotein 1 SCAP1|src family-associated phosphoprotein 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96464576, end_position_on_the_genomic_accession=96759622, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78478, CurrentID=0, Status='live', Symbol='Tmsb15a', Aliases='1700129I15Rik, Tb15a, Tmsbl1, Tmsl8', description='thymosin beta 15a', other_designations='thymosin beta-like protein 1|thymosin-like 8', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135718667, end_position_on_the_genomic_accession=135725261, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  209558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209558, CurrentID=0, Status='live', Symbol='Enpp3', Aliases='AI876438, CD203c', description='ectonucleotide pyrophosphatase/phosphodiesterase 3', other_designations='E-NPP 3|PD-Ibeta|phosphodiesterase I beta|phosphodiesterase I/nucleotide pyrophosphatase 3', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24773814, end_position_on_the_genomic_accession=24836195, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  78506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78506, CurrentID=0, Status='live', Symbol='Micu3', Aliases='2900075B16Rik, Efha2', description='mitochondrial calcium uptake family, member 3', other_designations='EF-hand domain family, member A2|EF-hand domain-containing family member A2', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40307840, end_position_on_the_genomic_accession=40386312, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  209584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209584, CurrentID=0, Status='live', Symbol='Tyw3', Aliases='5230400J09Rik', description='tRNA-yW synthesizing protein 3 homolog (S. cerevisiae)', other_designations='tRNA(Phe) 7-((3-amino-3-carboxypropyl)-4-demethylwyosine(37)-N(4))-methyltransferase|tRNA-yW-synthesizing protein 3', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=154576520, end_position_on_the_genomic_accession=154597098, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  209586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209586, CurrentID=0, Status='live', Symbol='Nudcd3', Aliases='AI427847, BC024322, mKIAA1068', description='NudC domain containing 3', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6105691, end_position_on_the_genomic_accession=6200862, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  209588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209588, CurrentID=0, Status='live', Symbol='Sectm1a', Aliases='BC010462', description='secreted and transmembrane 1A', other_designations='sectm1a', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121067395, end_position_on_the_genomic_accession=121097213, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  209590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209590, CurrentID=0, Status='live', Symbol='Il23r', Aliases='IL-23R', description='interleukin 23 receptor', other_designations='IL-23 receptor', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67422932, end_position_on_the_genomic_accession=67491855, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  78521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78521, CurrentID=0, Status='live', Symbol='B230219D22Rik', Aliases='AA119856, D530037I19Rik', description='RIKEN cDNA B230219D22 gene', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55693124, end_position_on_the_genomic_accession=55703500, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  78523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78523, CurrentID=0, Status='live', Symbol='Mrpl9', Aliases='8030480E20Rik, AA409733, C330013D18Rik, L9mt', description='mitochondrial ribosomal protein L9', other_designations='MRP-L9', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94443272, end_position_on_the_genomic_accession=94448708, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  209601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209601, CurrentID=0, Status='live', Symbol='Erich3', Aliases='1700013E18Rik, 4922501L14Rik, 5031409G23Rik, AI596448, Gm133', description='glutamate rich 3', other_designations='', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=154663387, end_position_on_the_genomic_accession=154767790, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  78541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78541, CurrentID=0, Status='live', Symbol='Asb8', Aliases='4930539L19Rik, AI788835, C430011H06Rik', description='ankyrin repeat and SOCS box-containing 8', other_designations='ASB-8|ankyrin repeat and SOCS box-containing protein 8', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98134637, end_position_on_the_genomic_accession=98165602, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  209630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209630, CurrentID=0, Status='live', Symbol='Frmd4a', Aliases='2700017I06Rik, 5330420J21, AI594879, C230040M21Rik, Frmd4, Gm13190, OTTMUSG00000010806, mKIAA1294', description='FERM domain containing 4A', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=4017695, end_position_on_the_genomic_accession=4614043, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  78560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78560, CurrentID=0, Status='live', Symbol='Adgra2', Aliases='8430414O08Rik, 9530074E10Rik, Gpr124, Tem5, mKIAA1531', description='adhesion G protein-coupled receptor A2', other_designations='G protein-coupled receptor 124|tumor endothelial marker 5', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27085822, end_position_on_the_genomic_accession=27123436, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  209645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209645, CurrentID=0, Status='live', Symbol='Bend7', Aliases='1110017O21Rik, E130319B15Rik', description='BEN domain containing 7', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=4717831, end_position_on_the_genomic_accession=4802146, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  78581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78581, CurrentID=0, Status='live', Symbol='Utp23', Aliases='1700010I21Rik, AI662478, BB238373, D530033C11Rik', description='UTP23, small subunit (SSU) processome component, homolog (yeast)', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=51877441, end_position_on_the_genomic_accession=51884622, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78586, CurrentID=0, Status='live', Symbol='Srbd1', Aliases='AI461933, C85414, D530025C17Rik', description='S1 RNA binding domain 1', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=85984665, end_position_on_the_genomic_accession=86145216, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  78591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78591, CurrentID=0, Status='live', Symbol='Mir142hg', Aliases='A430104N18Rik, AW909330', description='Mir142 host gene (non-protein coding)', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  78593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78593, CurrentID=0, Status='live', Symbol='Nrip3', Aliases='A330103B05Rik, D7H11orf14, ICRFP703B1614Q5.2', description='nuclear receptor interacting protein 3', other_designations='D7H11orf14|ICRFP703N2430Q5.2', map_location='7 E3-F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109758056, end_position_on_the_genomic_accession=109781545, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  78600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78600, CurrentID=0, Status='live', Symbol='Pde6h', Aliases='A930033D18Rik, PDEgamma', description='phosphodiesterase 6H, cGMP-specific, cone, gamma', other_designations='GMP-PDE gamma|cone phosphodiesterase 6 gamma|cone-like cGMP-phosphodiesterase 6 gamma subunit type 2', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136952221, end_position_on_the_genomic_accession=136963485, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  78609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78609, CurrentID=0, Status='live', Symbol='8030411F24Rik', Aliases='Cst14', description='RIKEN cDNA 8030411F24 gene', other_designations='cystatin SC', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148782009, end_position_on_the_genomic_accession=148785936, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78610, CurrentID=0, Status='live', Symbol='Uvrag', Aliases='9530039D02Rik, AI648770, BB1242051, Uvragl, Uvrag', description='UV radiation resistance associated gene', other_designations='UV radiation resistance-associated gene protein', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98886742, end_position_on_the_genomic_accession=99141168, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  209683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209683, CurrentID=0, Status='live', Symbol='Ttc28', Aliases='2310015L07Rik, 6030435N04, AI428795, AI851761, BC002262, TPRBK', description='tetratricopeptide repeat domain 28', other_designations='TPR repeat protein 28', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110879784, end_position_on_the_genomic_accession=111289780, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  78617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78617, CurrentID=0, Status='live', Symbol='Cstad', Aliases='9530039B02Rik', description='CSA-conditional, T cell activation-dependent protein', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30595044, end_position_on_the_genomic_accession=30608945, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  78618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78618, CurrentID=0, Status='live', Symbol='Acap2', Aliases='4832442G16, 9530039J15Rik, CNT-B2, Centb2, mKIAA0041', description='ArfGAP with coiled-coil, ankyrin repeat and PH domains 2', other_designations='centaurin, beta 2|centaurin-beta-2', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31092412, end_position_on_the_genomic_accession=31202056, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  78619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78619, CurrentID=0, Status='live', Symbol='Zfp449', Aliases='9530049C15Rik', description='zinc finger protein 449', other_designations='', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56346383, end_position_on_the_genomic_accession=56365674, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  209692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209692, CurrentID=0, Status='live', Symbol='Dhtkd1', Aliases='C330018I04Rik', description='dehydrogenase E1 and transketolase domain containing 1', other_designations='dehydrogenase E1 and transketolase domain-containing protein 1', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5898060, end_position_on_the_genomic_accession=5943010, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  78625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78625, CurrentID=0, Status='live', Symbol='1700061G19Rik', Aliases='1700026H02, AI507598', description='RIKEN cDNA 1700061G19 gene', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56875452, end_position_on_the_genomic_accession=56889855, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  78631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78631, CurrentID=0, Status='live', Symbol='Cypt15', Aliases='1700085A12Rik', description='cysteine-rich perinuclear theca 15', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=39346266, end_position_on_the_genomic_accession=39346963, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  78634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78634, CurrentID=0, Status='live', Symbol='Spaca7', Aliases='1700094C09Rik', description='sperm acrosome associated 7', other_designations='sperm acrosome-associated protein 7', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=12572997, end_position_on_the_genomic_accession=12600744, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  209707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209707, CurrentID=0, Status='live', Symbol='Lcorl', Aliases='Mlr1', description='ligand dependent nuclear receptor corepressor-like', other_designations='LCoR-like protein|Mblk1-related protein-1|mblk1-related protein 1|transcription factor MLR1', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=45697184, end_position_on_the_genomic_accession=45860115, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78651, CurrentID=0, Status='live', Symbol='Lsm6', Aliases='1500031N17Rik, 2410088K19Rik, AI747288', description='LSM6 homolog, U6 small nuclear RNA and mRNA degradation associated', other_designations='LSM6 homolog, U6 small nuclear RNA associated', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=78804868, end_position_on_the_genomic_accession=78821152, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  78653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78653, CurrentID=0, Status='live', Symbol='Bola3', Aliases='1810056O20Rik', description='bolA-like 3 (E. coli)', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83349484, end_position_on_the_genomic_accession=83358392, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78655, CurrentID=0, Status='live', Symbol='Eif3j1', Aliases='1810016I04Rik, 2700079K05Rik, AA409117, AA409446, C78575, Eif3j, Eif3j-1, Eif3j-2, Eif3j2, Eif3s1, eIF-3-alpha-A, eIF-3-alpha-B, eIF3-alpha, eIF3-p35, eIF3j-A, eIF3j-B', description='eukaryotic translation initiation factor 3, subunit J1', other_designations='eIF-3-alpha-A|eIF-3-alpha-B|eIF3j-A|eIF3j-B|eukaryotic translation initiation factor 3 subunit 1-A|eukaryotic translation initiation factor 3 subunit 1-B|eukaryotic translation initiation factor 3, subunit 1 alpha|eukaryotic translation initiation factor 3, subunit J', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122028113, end_position_on_the_genomic_accession=122053632, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  78656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78656, CurrentID=0, Status='live', Symbol='Brd8', Aliases='2610007E11Rik, 4432404P07Rik, SMAP', description='bromodomain containing 8', other_designations='', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34598615, end_position_on_the_genomic_accession=34624863, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  78658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78658, CurrentID=0, Status='live', Symbol='Ncapd3', Aliases='2810487N22Rik, 4632407J06Rik, AI195468, AU018739, B130055D15Rik, mKIAA0056', description='non-SMC condensin II complex, subunit D3', other_designations='Jacobsen syndrome region 2', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=27030116, end_position_on_the_genomic_accession=27096522, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  209737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209737, CurrentID=0, Status='live', Symbol='Kif15', Aliases='3110023M17Rik, 3930402I10Rik, D330038N01, HKLP2, Knsl7, b2b1117.1Clo, b2b1117.1aClo', description='kinesin family member 15', other_designations='N-10 kinesin|kinesin 15|kinesin-like 7|kinesin-like protein 2|kinesin-like protein 7', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122951052, end_position_on_the_genomic_accession=123018733, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  78670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78670, CurrentID=0, Status='live', Symbol='Plekhj1', Aliases='9530063M10Rik', description='pleckstrin homology domain containing, family J member 1', other_designations='PH domain-containing family J member 1|guanine nucleotide releasing protein x|guanine nucleotide-releasing protein x', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80796099, end_position_on_the_genomic_accession=80798626, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  209743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209743, CurrentID=0, Status='live', Symbol='AF529169', Aliases='DD1, Kiaa1024', description='cDNA sequence AF529169', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89587210, end_position_on_the_genomic_accession=89623583, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  78672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78672, CurrentID=0, Status='live', Symbol='9530057J20Rik', Aliases='', description='RIKEN cDNA 9530057J20 gene', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19903184, end_position_on_the_genomic_accession=19935330, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  209760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209760, CurrentID=0, Status='live', Symbol='Tmc7', Aliases='1700030H01Rik, C230064B05, C630024K23Rik', description='transmembrane channel-like gene family 7', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118535841, end_position_on_the_genomic_accession=118584790, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  78689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78689, CurrentID=0, Status='live', Symbol='Naa35', Aliases='A330021G12Rik, A330027C19Rik, AI158944, C030004C14Rik, Mak10', description='N(alpha)-acetyltransferase 35, NatC auxiliary subunit', other_designations='MAK10 homolog, amino-acid N-acetyltransferase subunit|Nalpha acetyltransferase 35|embryonic growth-associated protein|protein MAK10 homolog', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59585323, end_position_on_the_genomic_accession=59634798, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  78697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78697, CurrentID=0, Status='live', Symbol='Pus7', Aliases='C330017I15Rik', description='pseudouridylate synthase 7', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23740165, end_position_on_the_genomic_accession=23783711, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  209773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209773, CurrentID=0, Status='live', Symbol='Dennd2a', Aliases='B930096L08Rik, C130031G22, mKIAA1277', description='DENN/MADD domain containing 2A', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39462374, end_position_on_the_genomic_accession=39557864, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  209776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209776, CurrentID=0, Status='live', Symbol='Gpr139', Aliases='GPRg1, Gm495, PGR3', description='G protein-coupled receptor 139', other_designations='G protein-coupled receptor PGR3|G-protein-coupled receptor PGR3|g(q)-coupled orphan receptor GPRg1|orphan G protein-coupled receptor TRHRL2', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119144323, end_position_on_the_genomic_accession=119184374, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  78709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78709, CurrentID=0, Status='live', Symbol='Spink8', Aliases='C630041L24Rik', description='serine peptidase inhibitor, Kazal type 8', other_designations='serine protease inhibitor Kazal type 8', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109812203, end_position_on_the_genomic_accession=109826628, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  78733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78733, CurrentID=0, Status='live', Symbol='Troap', Aliases='AW476063, E130301L11Rik, tastin', description='trophinin associated protein', other_designations='troap protein|trophinin-assisting protein|trophinin-associated protein', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99074973, end_position_on_the_genomic_accession=99083409, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  209815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209815, CurrentID=0, Status='live', Symbol='Tbc1d25', Aliases='6330576A08Rik, DXHXS7927e, Oatl1', description='TBC1 domain family, member 25', other_designations='ornithine aminotransferase-like 1', map_location='X 3.69 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8154470, end_position_on_the_genomic_accession=8176181, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  78748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78748, CurrentID=0, Status='live', Symbol='Rassf10', Aliases='4632411J06Rik', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 10', other_designations='', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=112953962, end_position_on_the_genomic_accession=112957458, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  78749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78749, CurrentID=0, Status='live', Symbol='Filip1l', Aliases='4631422O05Rik, Doc1', description='filamin A interacting protein 1-like', other_designations='downregulated in ovarian cancer 1-like|protein down-regulated in ovarian cancer 1 homolog', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=57353277, end_position_on_the_genomic_accession=57572804, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  78751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78751, CurrentID=0, Status='live', Symbol='Zc3h6', Aliases='4631426G04Rik, 4833425H18Rik, AU024389, AW060453, Zc3hdc6, mKIAA2035', description='zinc finger CCCH type containing 6', other_designations='zinc finger CCCH type domain containing 6', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128967402, end_position_on_the_genomic_accession=129018563, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  209824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209824, CurrentID=0, Status='live', Symbol='Vmn1r183', Aliases='Gm476, V1rd15', description='vomeronasal 1 receptor 183', other_designations='vomernasal 1 receptor Vmn1r183|vomeronasal 1 receptor, D15', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24054774, end_position_on_the_genomic_accession=24055691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  78753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78753, CurrentID=0, Status='live', Symbol='Lipm', Aliases='4632427C23Rik, AW227514, Lipl3', description='lipase, family member M', other_designations='lipase-like abhydrolase domain-containing protein 3|lipase-like, ab-hydrolase domain containing 3', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34100943, end_position_on_the_genomic_accession=34122687, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  78754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78754, CurrentID=0, Status='live', Symbol='Galnt15', Aliases='4631401E18Rik, Galntl2, mpp-GalNAc-T15', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 15', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 2|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 2|galNAc-T-like protein 2|polypeptide GalNAc transferase-like protein 2|polypeptide N-acetylgalactosaminyltransferase-like 2|polypeptide N-acetylgalactosaminyltransferase-like protein 2|pp-GaNTase-like protein 2|protein-UDP acetylgalactosaminyltransferase-like protein 2', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32028651, end_position_on_the_genomic_accession=32062204, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  78755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78755, CurrentID=0, Status='live', Symbol='Fam122b', Aliases='4632404H22Rik, AI256344, SPACIA2', description='family with sequence similarity 122, member B', other_designations='synoviocyte proliferation associated in collagen-induced arthritis 2', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53243415, end_position_on_the_genomic_accession=53269850, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  78757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78757, CurrentID=0, Status='live', Symbol='Rictor', Aliases='4921505C17Rik, 6030405M08Rik, AVO3, AW492497, D530039E11Rik', description='RPTOR independent companion of MTOR, complex 2', other_designations='AVO3 homolog|mAVO3|protein pianissimo|rictor', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=6708349, end_position_on_the_genomic_accession=6800401, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  209837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209837, CurrentID=0, Status='live', Symbol='Slc38a5', Aliases='C81234, E330031E14, JM24, SN2', description='solute carrier family 38, member 5', other_designations='system N transporter 2', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8271381, end_position_on_the_genomic_accession=8280179, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  78767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78767, CurrentID=0, Status='live', Symbol='Efcab11', Aliases='2610021K21Rik, 2610110A13Rik', description='EF-hand calcium binding domain 11', other_designations='EF-hand domain-containing protein C14orf143 homolog', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=99692752, end_position_on_the_genomic_accession=99884445, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  78771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78771, CurrentID=0, Status='live', Symbol='Mctp1', Aliases='2810465F10Rik', description='multiple C2 domains, transmembrane 1', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=76384961, end_position_on_the_genomic_accession=77033284, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  78772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78772, CurrentID=0, Status='live', Symbol='Hhipl2', Aliases='4930507C10Rik, AW208802, Hhip3', description='hedgehog interacting protein-like 2', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=183413929, end_position_on_the_genomic_accession=183437053, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  78777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78777, CurrentID=0, Status='live', Symbol='Enthd2', Aliases='2410002I01Rik', description='ENTH domain containing 2', other_designations='ENTH domain-containing protein 2|protein C17orf56 homolog|tetra-epsin', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120090522, end_position_on_the_genomic_accession=120098780, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  78779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78779, CurrentID=0, Status='live', Symbol='Spata2l', Aliases='2610039E05Rik, AI648233, mKIAA4138', description='spermatogenesis associated 2-like', other_designations='SPATA2-like protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123232255, end_position_on_the_genomic_accession=123236245, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  78781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78781, CurrentID=0, Status='live', Symbol='Zc3hav1', Aliases='1200014N16Rik, 2900058M19Rik, 9130009D18Rik, 9830115L13Rik, ARTD13, D6Bwg1452e, ZAP', description='zinc finger CCCH type, antiviral 1', other_designations='ADP-ribosyltransferase diphtheria toxin-like 13', map_location='6 17.72 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=38305286, end_position_on_the_genomic_accession=38355840, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  78783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78783, CurrentID=0, Status='live', Symbol='Brpf1', Aliases='4833438B11Rik, 4930540D11Rik, Brpf2', description='bromodomain and PHD finger containing, 1', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113306049, end_position_on_the_genomic_accession=113324862, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78784, CurrentID=0, Status='live', Symbol='Celf3', Aliases='4930415M08Rik, BRUNOL1, CAGH4, ERDA4, Tnrc4', description='CUGBP, Elav-like family member 3', other_designations='CELF-3|CUG-BP- and ETR-3-like factor 3|ELAV-type RNA-binding protein 1|ETR-1|RNA-binding protein BRUNOL-1|VCUGBP, Elav-like family member 3|bruno-like protein 1|trinucleotide repeat containing 4|trinucleotide repeat-containing gene 4 protein', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94478285, end_position_on_the_genomic_accession=94492199, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  78785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78785, CurrentID=0, Status='live', Symbol='Clip4', Aliases='1700024K14Rik, 1700074B05Rik, 4833417L20Rik, 5830409B12Rik, Rsnl2', description='CAP-GLY domain containing linker protein family, member 4', other_designations='restin-like protein 2', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71769669, end_position_on_the_genomic_accession=71864210, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  78787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78787, CurrentID=0, Status='live', Symbol='Usp54', Aliases='4930429G18Rik, AI115571, C030002J06Rik', description='ubiquitin specific peptidase 54', other_designations='inactive ubiquitin-specific peptidase 54|ubiquitin specific protease 54', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20548912, end_position_on_the_genomic_accession=20643129, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  78789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78789, CurrentID=0, Status='live', Symbol='Vsig1', Aliases='1700062D20Rik, 4930405J24Rik', description='V-set and immunoglobulin domain containing 1', other_designations='cell surface A33 antigen|glycoprotein A34', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=140907515, end_position_on_the_genomic_accession=140939472, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  78795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78795, CurrentID=0, Status='live', Symbol='Armc9', Aliases='3830422A13Rik, 4831423D23Rik, 4930438O05Rik, 5730415N24Rik', description='armadillo repeat containing 9', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86154732, end_position_on_the_genomic_accession=86278284, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  78796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78796, CurrentID=0, Status='live', Symbol='Zcchc4', Aliases='4930449I23Rik', description='zinc finger, CCHC domain containing 4', other_designations='', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52783041, end_position_on_the_genomic_accession=52824666, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  78797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78797, CurrentID=0, Status='live', Symbol='Ndor1', Aliases='4930447P04Rik, NR1, Ndor', description='NADPH dependent diflavin oxidoreductase 1', other_designations='NADPH-dependent FMN and FAD-containing oxidoreductase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25244813, end_position_on_the_genomic_accession=25255414, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  78798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78798, CurrentID=0, Status='live', Symbol='Eml4', Aliases='4930443C24Rik, AI644019', description='echinoderm microtubule associated protein like 4', other_designations='EMAP-4', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=83350931, end_position_on_the_genomic_accession=83480361, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  78801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78801, CurrentID=0, Status='live', Symbol='Ak7', Aliases='4930502N02Rik, AK 7', description='adenylate kinase 7', other_designations='ATP-AMP transphosphorylase 7|putative adenylate kinase 7', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=105703859, end_position_on_the_genomic_accession=105782447, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  78802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78802, CurrentID=0, Status='live', Symbol='Ttc30a1', Aliases='4930506L13Rik', description='tetratricopeptide repeat domain 30A1', other_designations='TPR repeat protein 30A1|novel tetratrico peptide repeat (TPR) containing protein|tetratricopeptide repeat domain 30A', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=75979106, end_position_on_the_genomic_accession=75981967, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  78803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78803, CurrentID=0, Status='live', Symbol='Fbxo43', Aliases='4930533G20Rik, Emi2', description='F-box protein 43', other_designations='XErp1 homolog|early mitotic inhibitor 2|endogenous meiotic inhibitor 2', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=36150060, end_position_on_the_genomic_accession=36173980, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  78806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78806, CurrentID=0, Status='live', Symbol='Stpg1', Aliases='4930403G18Rik, 4930555I21Rik, Mapo2', description='sperm tail PG rich repeat containing 1', other_designations='O6-methylguanine-induced apoptosis 2|UPF0490 protein C1orf201 homolog|sperm-tail PG-rich repeat-containing protein 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135494833, end_position_on_the_genomic_accession=135537803, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  78808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78808, CurrentID=0, Status='live', Symbol='Stxbp5', Aliases='0710001E20Rik, 4930565N16Rik, AW742610, LGL3, mKIAA1006', description='syntaxin binding protein 5 (tomosyn)', other_designations='lethal(2) giant larvae protein homolog 3|tomosyn 1|tomosyn-1', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=9755547, end_position_on_the_genomic_accession=9901227, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  78809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78809, CurrentID=0, Status='live', Symbol='4930562C15Rik', Aliases='', description='RIKEN cDNA 4930562C15 gene', other_designations='putative uncharacterized protein C16orf96 homolog', map_location='16 2.47 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4835416, end_position_on_the_genomic_accession=4868200, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  78816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78816, CurrentID=0, Status='live', Symbol='Gmip', Aliases='5031419I10Rik', description='Gem-interacting protein', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69808249, end_position_on_the_genomic_accession=69821870, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  78825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78825, CurrentID=0, Status='live', Symbol='Desi2', Aliases='5830417C01Rik, DeSI-2, Fam152a, Pppde1', description='desumoylating isopeptidase 2', other_designations='PPPDE peptidase domain containing 1|PPPDE peptidase domain-containing protein 1|family with sequence similarity 152, member A', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=178187231, end_position_on_the_genomic_accession=178252598, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  78826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78826, CurrentID=0, Status='live', Symbol='P2ry10', Aliases='5830408N17Rik, Lypsr2, P2Y10', description='purinergic receptor P2Y, G-protein coupled 10', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107088497, end_position_on_the_genomic_accession=107104974, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78829, CurrentID=0, Status='live', Symbol='Tsc22d4', Aliases='0610009M14Rik, 1700023B23Rik, AI415410, Thg-1pit, Tilz2', description='TSC22 domain family, member 4', other_designations='TSC22 domain family 4|TSC22-related inducible leucine zipper 2|TSC22-related-inducible leucine zipper protein 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137745969, end_position_on_the_genomic_accession=137768453, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  78830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78830, CurrentID=0, Status='live', Symbol='Slc25a12', Aliases='2610002D09Rik, AI839531, B230107K20Rik, BB129864', description='solute carrier family 25 (mitochondrial carrier, Aralar), member 12', other_designations='mitochondrial aspartate glutamate carrier 1|solute carrier family 25 member 12', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71274270, end_position_on_the_genomic_accession=71376694, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  78832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78832, CurrentID=0, Status='live', Symbol='Cacul1', Aliases='2700078E11Rik, 2810417M16Rik, 9830127L17Rik, AI450346, D130033C15Rik', description='CDK2 associated, cullin domain 1', other_designations='', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60524696, end_position_on_the_genomic_accession=60581023, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78833, CurrentID=0, Status='live', Symbol='Gins3', Aliases='2700085M18Rik, AI616142', description='GINS complex subunit 3 (Psf3 homolog)', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95633559, end_position_on_the_genomic_accession=95645059, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78834, CurrentID=0, Status='live', Symbol='Zfp623', Aliases='2610029D06Rik, AI847036', description='zinc finger protein 623', other_designations='zinc finger CCCH type domain containing 3', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75940952, end_position_on_the_genomic_accession=75949402, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78878, CurrentID=0, Status='live', Symbol='Ftx', Aliases='B230206F22Rik, BB219131, NCRNA00182, Thcytx', description='Ftx transcript, Xist regulator (non-protein coding)', other_designations='thrombocytosis, familial X-linked', map_location='X 46.21 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103560910, end_position_on_the_genomic_accession=103623754, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  78885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78885, CurrentID=0, Status='live', Symbol='Coro7', Aliases='0610011B16Rik, AW048373, AW556392', description='coronin 7', other_designations='70 kDa WD repeat tumor rejection antigen homolog|crn7', map_location='16 2.45 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4626884, end_position_on_the_genomic_accession=4679720, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  78887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78887, CurrentID=0, Status='live', Symbol='Sfi1', Aliases='', description='Sfi1 homolog, spindle assembly associated (yeast)', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3131850, end_position_on_the_genomic_accession=3203343, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  78889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78889, CurrentID=0, Status='live', Symbol='Wsb1', Aliases='1110056B13Rik, 2700038M07Rik', description='WD repeat and SOCS box-containing 1', other_designations='WD-40-repeat-containing protein with a SOCS box 1|WSB-1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79239382, end_position_on_the_genomic_accession=79254700, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78890, CurrentID=0, Status='live', Symbol='Trmt44', Aliases='2310079F23Rik, Mettl19', description='tRNA methyltransferase 44', other_designations='methyltransferase-like protein 19', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35556208, end_position_on_the_genomic_accession=35575070, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  78891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78891, CurrentID=0, Status='live', Symbol='Scyl1', Aliases='2810011O19Rik, C85140, Ntkl, mdf, mfd, p105', description='SCY1-like 1 (S. cerevisiae)', other_designations='105 kDa kinase-like protein|SCY1-like 1 splice variant 13|SCY1-like 1 splice variant 15|SCY1-like protein 1|mitosis-associated kinase-like protein NTKL|ubiquitous protein kinase-like (105 kDa)', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5758427, end_position_on_the_genomic_accession=5771414, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  78892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78892, CurrentID=0, Status='live', Symbol='Crispld2', Aliases='1810049K24Rik, Lcrisp2, Lgl1, coffeecrisp', description='cysteine-rich secretory protein LCCL domain containing 2', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119992443, end_position_on_the_genomic_accession=120052793, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  78893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78893, CurrentID=0, Status='live', Symbol='Cnot10', Aliases='2600001P13Rik', description='CCR4-NOT transcription complex, subunit 10', other_designations='', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114585874, end_position_on_the_genomic_accession=114640216, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  78894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78894, CurrentID=0, Status='live', Symbol='Aacs', Aliases='2210408B16Rik, SUR5', description='acetoacetyl-CoA synthetase', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=125475873, end_position_on_the_genomic_accession=125517403, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  78895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78895, CurrentID=0, Status='live', Symbol='Pus7l', Aliases='3000003F02Rik', description='pseudouridylate synthase 7-like', other_designations='pseudouridylate synthase 7 homolog (S. cerevisiae)-like', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=94522640, end_position_on_the_genomic_accession=94543513, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  78896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78896, CurrentID=0, Status='live', Symbol='1500015O10Rik', Aliases='Ecrg4', description='RIKEN cDNA 1500015O10 gene', other_designations='esophageal cancer related gene 4 protein|esophageal cancer-related gene 4 protein homolog', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=43730602, end_position_on_the_genomic_accession=43742564, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78903, CurrentID=0, Status='live', Symbol='Wrnip1', Aliases='4833444L21Rik, WHIP, Wrnip', description='Werner helicase interacting protein 1', other_designations='Werner helicasae interacting protein 1|Werner syndrome homolog interacting protein|werner helicase-interacting protein 1|wrnip1', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=32802030, end_position_on_the_genomic_accession=32822610, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  78906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78906, CurrentID=0, Status='live', Symbol='Misp', Aliases='9130017N09Rik', description='mitotic spindle positioning', other_designations='mitotic spindle positioning protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79821019, end_position_on_the_genomic_accession=79830452, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  78908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78908, CurrentID=0, Status='live', Symbol='Igsf3', Aliases='1700016K10Rik, 2810035F16Rik, 4833439O17Rik, mKIAA0466', description='immunoglobulin superfamily, member 3', other_designations='', map_location='3 44.3 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=101377011, end_position_on_the_genomic_accession=101463060, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78910, CurrentID=0, Status='live', Symbol='Asb15', Aliases='4930400E23Rik, Asb-15', description='ankyrin repeat and SOCS box-containing 15', other_designations='ankyrin repeat and SOCS box-containing protein 15', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24528144, end_position_on_the_genomic_accession=24573164, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  78911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78911, CurrentID=0, Status='live', Symbol='Trim42', Aliases='4930486B16Rik', description='tripartite motif-containing 42', other_designations='', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=97349562, end_position_on_the_genomic_accession=97369958, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  78912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78912, CurrentID=0, Status='live', Symbol='Sp2', Aliases='4930480I16Rik, mKIAA0048', description='Sp2 transcription factor', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96953332, end_position_on_the_genomic_accession=96982611, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  78688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78688, CurrentID=0, Status='live', Symbol='Nol3', Aliases='ARC, B430311C09Rik, MYC, NOP, Nop30', description='nucleolar protein 3 (apoptosis repressor with CARD domain)', other_designations='apoptosis repressor with CARD', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105275937, end_position_on_the_genomic_accession=105281939, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78914, CurrentID=0, Status='live', Symbol='Nadsyn1', Aliases='9130012B15Rik', description='NAD synthetase 1', other_designations='NAD(+) synthase|NH3-dependent NAD(+) synthetase-like protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143795587, end_position_on_the_genomic_accession=143822864, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  78919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78919, CurrentID=0, Status='live', Symbol='Fndc8', Aliases='4930466G16Rik', description='fibronectin type III domain containing 8', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82892145, end_position_on_the_genomic_accession=82900737, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  78920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78920, CurrentID=0, Status='live', Symbol='Dlst', Aliases='1600017E01Rik, 4632413C10Rik, 4930529O08Rik, DLTS', description='dihydrolipoamide S-succinyltransferase (E2 component of 2-oxo-glutarate complex)', other_designations='2-oxoglutarate dehydrogenase complex component E2|E2K|OGDC-E2|dihydrolipoamide succinyltransferase component of 2-oxoglutarate dehydrogenase complex', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85110824, end_position_on_the_genomic_accession=85134091, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  78921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78921, CurrentID=0, Status='live', Symbol='9130019O22Rik', Aliases='', description='RIKEN cDNA 9130019O22 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127382260, end_position_on_the_genomic_accession=127387166, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  78923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78923, CurrentID=0, Status='live', Symbol='Chsy3', Aliases='4833446K15Rik, AI662215', description='chondroitin sulfate synthase 3', other_designations='N-acetylgalactosaminyl-proteoglycan 3-beta-glucuronosyltransferase 3|N-acetylgalactosaminyltransferase 3|carbohydrate synthase 2|chSy-2|chondroitin glucuronyltransferase 3|chondroitin synthase 2|glucuronosyl-N-acetylgalactosaminyl-proteoglycan 4-beta-N-acetylgalactosaminyltransferase II', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=59175340, end_position_on_the_genomic_accession=59411336, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  78925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78925, CurrentID=0, Status='live', Symbol='Srd5a1', Aliases='0610031P22Rik, 4930435F02Rik, S5AR 1, Srd5a-1', description='steroid 5 alpha-reductase 1', other_designations='SR type 1|steroid 5-alpha-reductase 1|type 1 5alpha-reductase', map_location='13 35.55 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=69573449, end_position_on_the_genomic_accession=69611900, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  78926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78926, CurrentID=0, Status='live', Symbol='Gas2l1', Aliases='4930500E24Rik, D0Jmb1, Gar22, TU-71.1', description='growth arrest-specific 2 like 1', other_designations='GAS2-related protein', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5054131, end_position_on_the_genomic_accession=5065327, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  78928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78928, CurrentID=0, Status='live', Symbol='Pigt', Aliases='2510012P17Rik, 4930534E15Rik, CGI-06, NDAP, Ndap7', description='phosphatidylinositol glycan anchor biosynthesis, class T', other_designations='neuronal development-associated protein 7|phosphatidylinositol-glycan biosynthesis class T protein', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164497511, end_position_on_the_genomic_accession=164508301, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  78929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78929, CurrentID=0, Status='live', Symbol='Polr3h', Aliases='5031409G22Rik, RPC8', description='polymerase (RNA) III (DNA directed) polypeptide H', other_designations='DNA-directed RNA polymerase III subunit H|RNA polymerase III subunit C7', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81915030, end_position_on_the_genomic_accession=81926247, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  210004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210004, CurrentID=0, Status='live', Symbol='B3gntl1', Aliases='6030413G23Rik', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase-like 1', other_designations='BGnT-like protein 1|beta1,3-N-acetylglucosaminyltransferase-like protein 1|beta3Gn-T-like protein 1|beta3GnTL1|ortholog of human UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase-like 1 B3GNTL1', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121616197, end_position_on_the_genomic_accession=121673160, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  78933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78933, CurrentID=0, Status='live', Symbol='Agbl4', Aliases='4930578N11Rik, 4931433A01Rik', description='ATP/GTP binding protein-like 4', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=110397643, end_position_on_the_genomic_accession=111664324, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  78935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78935, CurrentID=0, Status='live', Symbol='Saal1', Aliases='5031425D22Rik, SPACIA1', description='serum amyloid A-like 1', other_designations='synoviocyte proliferation-associated in collagen-induced arthritis 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46686108, end_position_on_the_genomic_accession=46710691, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  78937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78937, CurrentID=0, Status='live', Symbol='Avl9', Aliases='5830411G16Rik, D730049P16Rik, mKIAA0241', description='AVL9 homolog (S. cerevisiase)', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56714905, end_position_on_the_genomic_accession=56761912, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  78938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78938, CurrentID=0, Status='live', Symbol='Fbxo34', Aliases='2900057B08Rik, 5830426G16Rik', description='F-box protein 34', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47472561, end_position_on_the_genomic_accession=47531962, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  78943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78943, CurrentID=0, Status='live', Symbol='Ern1', Aliases='9030414B18Rik, AI225830, C85377, Ire1a, Ire1alpha, Ire1p', description='endoplasmic reticulum (ER) to nucleus signalling 1', other_designations='endoplasmic reticulum-to-nucleus signaling 1|ern1|inositol-requiring 1 alpha|inositol-requiring protein 1|ire1-alpha', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106397620, end_position_on_the_genomic_accession=106487796, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  210027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210027, CurrentID=0, Status='live', Symbol='Slc35f3', Aliases='B230375D17Rik', description='solute carrier family 35, member F3', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=126137601, end_position_on_the_genomic_accession=126395987, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  210029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210029, CurrentID=0, Status='live', Symbol='Metrnl', Aliases='9430048M07Rik, BC019776', description='meteorin, glial cell differentiation regulator-like', other_designations='cometin|subfatin', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121702427, end_position_on_the_genomic_accession=121717389, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  210035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210035, CurrentID=0, Status='live', Symbol='Tmem194', Aliases='Nemp1a, Tmem194', description='transmembrane protein 194', other_designations='transmembrane protein 194A', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127677065, end_position_on_the_genomic_accession=127701049, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  210044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210044, CurrentID=0, Status='live', Symbol='Adcy2', Aliases='mKIAA1060', description='adenylate cyclase 2', other_designations='ATP pyrophosphate-lyase 2|adenylate cyclase type II|adenylyl cyclase 2', map_location='13 35.55 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=68620043, end_position_on_the_genomic_accession=68999541, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  210045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210045, CurrentID=0, Status='live', Symbol='Nlrp4b', Aliases='AU022726, E330024M09, Nalp-gamma, Nalp4b', description='NLR family, pyrin domain containing 4B', other_designations='NACHT, leucine rich repeat and PYD containing 4B|NACHT/LRR/pyrin domain-containing protein 4b', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=10687793, end_position_on_the_genomic_accession=10730168, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  78977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78977, CurrentID=0, Status='live', Symbol='Popdc3', Aliases='AA682104, Pop3', description='popeye domain containing 3', other_designations='popeye 3|popeye protein 3', map_location='10 23.35 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=45289305, end_position_on_the_genomic_accession=45318450, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  668526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668526, CurrentID=0, Status='live', Symbol='Scgb2b17', Aliases='Abpbg17, EG668526, Gm9223', description='secretoglobin, family 2B, member 17', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33054582, end_position_on_the_genomic_accession=33056552, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  210094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210094, CurrentID=0, Status='live', Symbol='Iglon5', Aliases='A230106M20Rik, Gm1431', description='IgLON family member 5', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43472907, end_position_on_the_genomic_accession=43490057, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  210104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210104, CurrentID=0, Status='live', Symbol='Zfp658', Aliases='', description='zinc finger protein 658', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43562370, end_position_on_the_genomic_accession=43576142, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  210105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210105, CurrentID=0, Status='live', Symbol='Zfp719', Aliases='8030458J18, 9430094P17Rik, C630016O21Rik, mszf6', description='zinc finger protein 719', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43579555, end_position_on_the_genomic_accession=43593710, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  210106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210106, CurrentID=0, Status='live', Symbol='Papd7', Aliases='LAK-1, POLK, Pols, TRF4, TRF4-1', description='PAP associated domain containing 7', other_designations='DNA polymerase sigma|PAP-associated domain-containing protein 7|TRAMP-like complex polyadenylate polymerase|polymerase (DNA directed) sigma', map_location='13 35.55 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=69497952, end_position_on_the_genomic_accession=69534236, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  210108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210108, CurrentID=0, Status='live', Symbol='D130043K22Rik', Aliases='4930451E12Rik, Kiaa0319', description='RIKEN cDNA D130043K22 gene', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24843930, end_position_on_the_genomic_accession=24902159, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  79043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79043, CurrentID=0, Status='live', Symbol='Spsb3', Aliases='2310012N15Rik, 3300001M01Rik, SSB3, Tce1, mKIAA4204', description='splA/ryanodine receptor domain and SOCS box containing 3', other_designations='SPRY domain-containing SOCS box protein SSB-3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24886349, end_position_on_the_genomic_accession=24892147, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  79044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79044, CurrentID=0, Status='live', Symbol='Mrps34', Aliases='0610007F04Rik, 5330430D13Rik, AV001970, Tce2', description='mitochondrial ribosomal protein S34', other_designations='MRP-S34|S34mt|T-complex expressed gene 2 protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24895120, end_position_on_the_genomic_accession=24896274, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  210126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210126, CurrentID=0, Status='live', Symbol='Lpp', Aliases='9430020K16Rik, AA959454, AU024130, B130055L10Rik, C79715, D630048H16', description='LIM domain containing preferred translocation partner in lipoma', other_designations='lipoma preferred partner', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=24393350, end_position_on_the_genomic_accession=24992578, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  79059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79059, CurrentID=0, Status='live', Symbol='Nme3', Aliases='1810009F08Rik, AI413736, DR-nm23, Ndk3, Nm23-DR, Nm23-M3', description='NME/NM23 nucleoside diphosphate kinase 3', other_designations='NDK 3|NDP kinase 3|NDPKC|expressed in non-metastatic cells 3|non-metastatic cells 3, protein expressed in|nucleoside diphosphate kinase C|nucleoside diphosphate kinase DR-nm23', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24896500, end_position_on_the_genomic_accession=24897531, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  210135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210135, CurrentID=0, Status='live', Symbol='Zfp180', Aliases='2310040I01Rik, AI852378, D130011P11, HHZ168', description='zinc finger protein 180', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24081897, end_position_on_the_genomic_accession=24107709, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  210145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210145, CurrentID=0, Status='live', Symbol='Irgc1', Aliases='F630044M05Rik, Gm1102, Gm474, Ifgge, Iigp5, Irgc', description='immunity-related GTPase family, cinema 1', other_designations='immunity-related GTPase cinema 1|interferon inducible GTPase 5|interferon inducible GTPase family member 5|interferon-gamma-inducible GTPase Ifgge protein', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24431919, end_position_on_the_genomic_accession=24445682, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  210146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210146, CurrentID=0, Status='live', Symbol='Irgq', Aliases='AF322649, FKSG271, Irgq', description='immunity-related GTPase family, Q', other_designations='immunity-related GTPase family, Q1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24530648, end_position_on_the_genomic_accession=24538600, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  210148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210148, CurrentID=0, Status='live', Symbol='Slc30a6', Aliases='9530029F08Rik, ZnT6', description='solute carrier family 30 (zinc transporter), member 6', other_designations='solute carrier family 30 (zinc transporter), member 4|solute carrier family 30 member 6|zinc transporter-like 3', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74395608, end_position_on_the_genomic_accession=74424229, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  231633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231633, CurrentID=0, Status='live', Symbol='Tmem119', Aliases='AW208946, BC025600, obif', description='transmembrane protein 119', other_designations='osteoblast induction factor', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113793729, end_position_on_the_genomic_accession=113800358, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  210162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210162, CurrentID=0, Status='live', Symbol='Zkscan2', Aliases='9430065N20Rik, AW493225, Zfp694', description='zinc finger with KRAB and SCAN domains 2', other_designations='zinc finger protein 694', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=123477747, end_position_on_the_genomic_accession=123504904, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  210172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210172, CurrentID=0, Status='live', Symbol='Zfp526', Aliases='D030024H03Rik, Znf526', description='zinc finger protein 526', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25219847, end_position_on_the_genomic_accession=25228040, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  210198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210198, CurrentID=0, Status='live', Symbol='Gprc6a', Aliases='', description='G protein-coupled receptor, family C, group 6, member A', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=51614688, end_position_on_the_genomic_accession=51631614, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79196, CurrentID=0, Status='live', Symbol='Osbpl5', Aliases='1110006M06Rik, AI462538, ORP5, Obph1, Osbp2', description='oxysterol binding protein-like 5', other_designations='ORP-5|OSBP-related protein 5|oxysterol binding protein 2|oxysterol binding protein homolog 1|oxysterol-binding protein homolog 1|oxysterol-binding protein-like protein 5', map_location='7 88.29 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143688762, end_position_on_the_genomic_accession=143756980, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  79201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79201, CurrentID=0, Status='live', Symbol='Tnfrsf23', Aliases='SOB, TNFRSFH23, Tnfrh1, Tnfrsf1al1, mDcTRAILR1, mSOB', description='tumor necrosis factor receptor superfamily, member 23', other_designations='TNF receptor family member SOB|TNF receptor homolog 1|decoy TRAIL receptor 1|tumor necrosis factor receptor p60 homolog 1|tumor necrosis factor receptor superfamily, member 1a-like 1', map_location='7 88.28 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143665807, end_position_on_the_genomic_accession=143686423, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  210274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210274, CurrentID=0, Status='live', Symbol='Shank2', Aliases='ProSAP1, mKIAA1022', description='SH3/ankyrin domain gene 2', other_designations='cortBP1|cortactin-binding protein 1', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143978056, end_position_on_the_genomic_accession=144424494, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  79221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79221, CurrentID=0, Status='live', Symbol='Hdac9', Aliases='AV022454, D030072B18Rik, HD7B, HD9, HDRP, Hdac7b, Mitr, mKIAA0744', description='histone deacetylase 9', other_designations='MEF2-interacting transcription repressor MITR|histone deacetylase 7B|histone deacetylase-related protein', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=34047580, end_position_on_the_genomic_accession=34917095, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  210297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210297, CurrentID=0, Status='live', Symbol='Lrch2', Aliases='AI316885, mKIAA1495', description='leucine-rich repeats and calponin homology (CH) domain containing 2', other_designations='', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=147470375, end_position_on_the_genomic_accession=147554103, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  79233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79233, CurrentID=0, Status='live', Symbol='Zfp319', Aliases='AI225750, AI413472, Znf319, mKIAA1388', description='zinc finger protein 319', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95326136, end_position_on_the_genomic_accession=95331950, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  79235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79235, CurrentID=0, Status='live', Symbol='Lrat', Aliases='1300010A18Rik, AI449251', description='lecithin-retinol acyltransferase (phosphatidylcholine-retinol-O-acyltransferase)', other_designations='phosphatidylcholine--retinol O-acyltransferase|phosphatidylcholine-retinol-O-acyltransferase', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=82892582, end_position_on_the_genomic_accession=82903974, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  210321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210321, CurrentID=0, Status='live', Symbol='BC048679', Aliases='', description='cDNA sequence BC048679', other_designations='epididymal protein Av381126', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81494272, end_position_on_the_genomic_accession=81498330, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79263, CurrentID=0, Status='live', Symbol='Trim39', Aliases='1100001D15Rik, E130103K13Rik, RBCC-B30.2, Rnf23, mKIAA4179, tfp', description='tripartite motif-containing 39', other_designations='RING-B box-coiled coil-B30.2|RING-B box-coiled-coil-B30.2|ring finger protein 23|testis-abundant finger protein|tripartite motif protein 39|tripartite motif-containing protein 39', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36258872, end_position_on_the_genomic_accession=36272004, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  79264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79264, CurrentID=0, Status='live', Symbol='Krit1', Aliases='2010007K12Rik, A630036P20Rik, AA432855, AI450393, AI643869, BB155247, BB235701, Ccm1', description='KRIT1, ankyrin repeat containing', other_designations='cerebral cavernous malformations 1 protein homolog|krev interaction trapped 1', map_location='5 2.26 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3803165, end_position_on_the_genomic_accession=3844515, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  210356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210356, CurrentID=0, Status='live', Symbol='Nckap5', Aliases='8430408F21, D130011D22Rik, E030049G20Rik, Erih1, Erih2, Gm1548, Nap5', description='NCK-associated protein 5', other_designations='peripheral clock protein', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=125913636, end_position_on_the_genomic_accession=126830897, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  78752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78752, CurrentID=0, Status='live', Symbol='Csgalnact2', Aliases='4632415D10Rik, Galnact2', description='chondroitin sulfate N-acetylgalactosaminyltransferase 2', other_designations='beta4GalNAcT-2|chondroitin beta-1,4-N-acetylgalactosaminyltransferase 2|chondroitin sulfate GalNAcT-2|galNAcT-2', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118107452, end_position_on_the_genomic_accession=118139140, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  210376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210376, CurrentID=0, Status='live', Symbol='Mtmr9', Aliases='9430075G12Rik, AA516943, AF073881, LIP-STYX, MTMR8, mMTMH3', description='myotubularin related protein 9', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63519651, end_position_on_the_genomic_accession=63544050, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  210417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210417, CurrentID=0, Status='live', Symbol='Thsd7b', Aliases='1700074E13Rik, B930082A18, D130067I03Rik', description='thrombospondin, type I, domain containing 7B', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=129273304, end_position_on_the_genomic_accession=130219278, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  79362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79362, CurrentID=0, Status='live', Symbol='Bhlhe41', Aliases='6430520M22Rik, Bhlhb2l, Bhlhb3, DEC2, Sharp1', description='basic helix-loop-helix family, member e41', other_designations='bHLH transcriptional factor Dec2|basic helix-loop-helix domain containing, class B3|class B basic helix-loop-helix protein 3|differentially expressed in chondrocytes protein 2', map_location='6 G2-G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145858243, end_position_on_the_genomic_accession=145865591, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  210463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210463, CurrentID=0, Status='live', Symbol='Slc22a22', Aliases='oat-pg', description='solute carrier family 22 (organic cation transporter), member 22', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=57243767, end_position_on_the_genomic_accession=57477625, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  79401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79401, CurrentID=0, Status='live', Symbol='Spz1', Aliases='1700027M20Rik, AI429130', description='spermatogenic leucine zipper 1', other_designations='bhlh-zip transcription factor SPZ1|spermatogenic Zip 1', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92574631, end_position_on_the_genomic_accession=92576232, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  79410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79410, CurrentID=0, Status='live', Symbol='Klra23', Aliases='Ly-49W1, Ly-49W2, Ly49W', description='killer cell lectin-like receptor subfamily A, member 23', other_designations='killer cell lectin-like receptor subfamily A, member 17', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  210503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210503, CurrentID=0, Status='live', Symbol='Zfp677', Aliases='A830058L05Rik, Tce8', description='zinc finger protein 677', other_designations='', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21383455, end_position_on_the_genomic_accession=21399265, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  210510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210510, CurrentID=0, Status='live', Symbol='Tdrd6', Aliases='AB097085, Tdr2', description='tudor domain containing 6', other_designations='tudor repeat 2 protein', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43615334, end_position_on_the_genomic_accession=43631350, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  79455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79455, CurrentID=0, Status='live', Symbol='Pdcl2', Aliases='1700010B22Rik, 1700016K07Rik, Mgcphlp, Tphlp', description='phosducin-like 2', other_designations='phosducin-like protein similar 1', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76312127, end_position_on_the_genomic_accession=76331172, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  79456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79456, CurrentID=0, Status='live', Symbol='Recql4', Aliases='RecQ4', description='RecQ protein-like 4', other_designations='DNA helicase, RecQ-like type 4|RecQ helicase protein-like 4', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76703553, end_position_on_the_genomic_accession=76710576, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  210529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210529, CurrentID=0, Status='live', Symbol='Mettl14', Aliases='G430022H21Rik, mKIAA1627', description='methyltransferase like 14', other_designations='methyltransferase-like protein 14|snoRNA MBII-84', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=123368295, end_position_on_the_genomic_accession=123385990, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  210530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210530, CurrentID=0, Status='live', Symbol='P3h2', Aliases='4832416N06, AW553532, Leprel1, Mlat4', description='prolyl 3-hydroxylase 2', other_designations='leprecan-like 1|leprecan-like protein 1|myxoid liposarcoma associated protein 4', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=25960323, end_position_on_the_genomic_accession=26105784, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  79459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79459, CurrentID=0, Status='live', Symbol='Aldoart2', Aliases='4933425L11Rik, Aldo1-ps1, Aldoa-ps1', description='aldolase 1 A, retrogene 2', other_designations='aldolase 1 A retrogene 2|aldolase 1, A isoform, pseudogene 1|aldolase A retroprotein 2', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55565205, end_position_on_the_genomic_accession=55566896, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  210535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210535, CurrentID=0, Status='live', Symbol='Gm12169', Aliases='OTTMUSG00000005497', description='predicted gene 12169', other_designations='', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46524251, end_position_on_the_genomic_accession=46538156, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  79464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79464, CurrentID=0, Status='live', Symbol='Lias', Aliases='2900022L22Rik, 4933425M12Rik, 7a5ex, C77512, LS, lip-syn', description='lipoic acid synthetase', other_designations='lipoate synthase|lipoic acid synthase', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65391497, end_position_on_the_genomic_accession=65409208, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  210544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210544, CurrentID=0, Status='live', Symbol='Tbc1d31', Aliases='4B3, D330013L20Rik, Gm85, Wdr67', description='TBC1 domain family, member 31', other_designations='WD repeat domain 67|WD repeat-containing protein 67|protein 4-B-3', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=57911684, end_position_on_the_genomic_accession=57970068, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  210554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210554, CurrentID=0, Status='live', Symbol='Hus1b', Aliases='', description='HUS1 checkpoint clamp component B', other_designations='Hus1 checkpoint homolog b|Hus1 homolog b', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=30946576, end_position_on_the_genomic_accession=30947761, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  210573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210573, CurrentID=0, Status='live', Symbol='Tmem151b', Aliases='Gm323', description='transmembrane protein 151B', other_designations='', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45541940, end_position_on_the_genomic_accession=45549677, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  210582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210582, CurrentID=0, Status='live', Symbol='Coq10a', Aliases='Gm1', description='coenzyme Q10A', other_designations='coenzyme Q10 homolog A', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128363096, end_position_on_the_genomic_accession=128370037, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  210622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210622, CurrentID=0, Status='live', Symbol='Pamr1', Aliases='5930437L24, E430002G05Rik, Ramp', description='peptidase domain containing associated with muscle regeneration 1', other_designations='peptidase domain-containing protein associated with muscle regeneration 1|regeneration associated muscle protease|regeneration-associated muscle protease homolog', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=102550021, end_position_on_the_genomic_accession=102643041, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  79554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79554, CurrentID=0, Status='live', Symbol='Cptp', Aliases='Gltpd1', description='ceramide-1-phosphate transfer protein', other_designations='glycolipid transfer protein domain containing 1|glycolipid transfer protein domain-containing protein 1', map_location='4 87.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155864723, end_position_on_the_genomic_accession=155869440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  79555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79555, CurrentID=0, Status='live', Symbol='BC005537', Aliases='8030460C05Rik, C78201', description='cDNA sequence BC005537', other_designations='', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24801657, end_position_on_the_genomic_accession=24812899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  79560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79560, CurrentID=0, Status='live', Symbol='Ublcp1', Aliases='4930527B16Rik, 8430435I17Rik, BC002236', description='ubiquitin-like domain containing CTD phosphatase 1', other_designations='nuclear proteasome inhibitor UBLCP1', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=44454571, end_position_on_the_genomic_accession=44470548, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  79565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79565, CurrentID=0, Status='live', Symbol='Wbscr27', Aliases='AW492986, BC002286', description='Williams Beuren syndrome chromosome region 27 (human)', other_designations='Williams-Beuren syndrome chromosome region 27 homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134932310, end_position_on_the_genomic_accession=134942637, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  79566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79566, CurrentID=0, Status='live', Symbol='Sh3bp5l', Aliases='2310074E09Rik, BC003251, mKIAA1720', description='SH3 binding domain protein 5 like', other_designations='SH3BP-5-like', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58330718, end_position_on_the_genomic_accession=58347728, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  210673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210673, CurrentID=0, Status='live', Symbol='Prrt3', Aliases='6330505P20, B230206N24Rik', description='proline-rich transmembrane protein 3', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113493638, end_position_on_the_genomic_accession=113502235, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  210710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210710, CurrentID=0, Status='live', Symbol='Gab3', Aliases='5930433H21Rik', description='growth factor receptor bound protein 2-associated protein 3', other_designations='GRB2-associated binder 3', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74988545, end_position_on_the_genomic_accession=75084905, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  210711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210711, CurrentID=0, Status='live', Symbol='Mcmbp', Aliases='1110007A13Rik, AU017382, BC025641, MCM-BP', description='minichromosome maintenance complex binding protein', other_designations='MCM (minichromosome maintenance deficient) binding protein|MCM-binding protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128696441, end_position_on_the_genomic_accession=128740429, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  210719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210719, CurrentID=0, Status='live', Symbol='Mkx', Aliases='9430023B20Rik, Irxl1', description='mohawk homeobox', other_designations='iroquois homeobox protein-like 1', map_location='18 4.53 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=6934966, end_position_on_the_genomic_accession=7004779, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  207785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207785, CurrentID=0, Status='live', Symbol='Csrnp2', Aliases='CSRNP-2, Csnrp2, taip-12', description='cysteine-serine-rich nuclear protein 2', other_designations='TGF-beta induced apoptosis protein 12|TGF-beta-induced apoptosis protein 12', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100479570, end_position_on_the_genomic_accession=100495239, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  210741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210741, CurrentID=0, Status='live', Symbol='Kcnk12', Aliases='mntk1', description='potassium channel, subfamily K, member 12', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87745821, end_position_on_the_genomic_accession=87797994, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  210757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210757, CurrentID=0, Status='live', Symbol='Themis', Aliases='E430004N04Rik, Gasp, Spot, Tsepa, thylex', description='thymocyte selection associated', other_designations='grb2-associating protein|signaling phosphoprotein specific for T cells|thymocyte selection pathway associated|thymocyte-expressed molecule involved in selection|thymus lymphocyte exclusively expressed', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=28668327, end_position_on_the_genomic_accession=28883820, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  210762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210762, CurrentID=0, Status='live', Symbol='Ppp1r36', Aliases='Gm70', description='protein phosphatase 1, regulatory subunit 36', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76417599, end_position_on_the_genomic_accession=76439492, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  210766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210766, CurrentID=0, Status='live', Symbol='Brcc3', Aliases='C6.1A', description='BRCA1/BRCA2-containing complex, subunit 3', other_designations='6.1A protein|BRCA1-A complex subunit BRCC36|BRCA1/BRCA2-containing complex subunit 36|BRISC complex subunit BRCC36', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75416628, end_position_on_the_genomic_accession=75455702, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  210789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210789, CurrentID=0, Status='live', Symbol='Tbc1d4', Aliases='5930406J04Rik, A930035N22, AV295684, As160', description='TBC1 domain family, member 4', other_designations='akt substrate of 160 kDa', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=101442360, end_position_on_the_genomic_accession=101609228, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  210801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210801, CurrentID=0, Status='live', Symbol='Unc5d', Aliases='D930029E11Rik, Unc5h4, mKIAA1777', description='unc-5 homolog D (C. elegans)', other_designations='protein unc-5 homolog 4|protein unc-5 homolog D|unc-5 homolog 4|unc5 homolog 4', map_location='8 A2-A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=28646717, end_position_on_the_genomic_accession=29219682, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  210808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210808, CurrentID=0, Status='live', Symbol='Lacc1', Aliases='9030625A04Rik', description='laccase (multicopper oxidoreductase) domain containing 1', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=77024201, end_position_on_the_genomic_accession=77036898, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  210853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210853, CurrentID=0, Status='live', Symbol='Zfp947', Aliases='6720451E15, EG210853, Gm4769', description='zinc finger protein 947', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=22144358, end_position_on_the_genomic_accession=22166018, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  210876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210876, CurrentID=0, Status='live', Symbol='Vmn2r111', Aliases='EG210876, V2rp2', description='vomeronasal 2, receptor 111', other_designations='vomeronasal 2 receptor p2', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=22547941, end_position_on_the_genomic_accession=22573273, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  210925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210925, CurrentID=0, Status='live', Symbol='Ints9', Aliases='BC028953, D14Ertd231e', description='integrator complex subunit 9', other_designations='int9', map_location='14 33.81 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64950045, end_position_on_the_genomic_accession=65039835, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  210933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210933, CurrentID=0, Status='live', Symbol='Adgrb3', Aliases='A830096D10Rik, Bai3', description='adhesion G protein-coupled receptor B3', other_designations='G protein-coupled receptor BAI3|brain-specific angiogenesis inhibitor 3', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=25067476, end_position_on_the_genomic_accession=25830206, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  404195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404195, CurrentID=0, Status='live', Symbol='Cyp2c54', Aliases='EG404195', description='cytochrome P450, family 2, subfamily c, polypeptide 54', other_designations='CYPIIC54|cyp2c54', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40037940, end_position_on_the_genomic_accession=40073845, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  210940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210940, CurrentID=0, Status='live', Symbol='4931408C20Rik', Aliases='', description='RIKEN cDNA 4931408C20 gene', other_designations='', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=26681801, end_position_on_the_genomic_accession=26687460, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  735252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735252, CurrentID=0, Status='live', Symbol='Mir181a-1', Aliases='Mirn181a-1, Mirn213, miR-213, mir-181a-1, mmu-mir-213', description='microRNA 181a-1', other_designations='microRNA 213|mir 181a-1|mmu-mir-181a-1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=137966455, end_position_on_the_genomic_accession=137966541, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  735257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735257, CurrentID=0, Status='live', Symbol='Mir467b', Aliases='Mirn467b, mir-467b', description='microRNA 467b', other_designations='microRNA mir-467b|mir 467b|mmu-mir-467b', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10481248, end_position_on_the_genomic_accession=10481320, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  210973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210973, CurrentID=0, Status='live', Symbol='Kbtbd2', Aliases='BC022962, Bklhd1, mKIAA1489', description='kelch repeat and BTB (POZ) domain containing 2', other_designations='BTB and kelch domain containing 1', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56777524, end_position_on_the_genomic_accession=56798285, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  735262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735262, CurrentID=0, Status='live', Symbol='Mir455', Aliases='Mirn455, mir-455, mmu-mir-455', description='microRNA 455', other_designations='mir 455', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63256851, end_position_on_the_genomic_accession=63256932, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  735266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735266, CurrentID=0, Status='live', Symbol='Mir592', Aliases='Mirn592, mir-592, mmu-mir-592', description='microRNA 592', other_designations='mir 592', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=27936655, end_position_on_the_genomic_accession=27936750, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  735267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735267, CurrentID=0, Status='live', Symbol='Mir705', Aliases='Mirn705, mmu-mir-705', description='microRNA 705', other_designations='mir 705', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85336292, end_position_on_the_genomic_accession=85336373, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  210982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210982, CurrentID=0, Status='live', Symbol='Gltscr1l', Aliases='mKIAA0240', description='GLTSCR1-like', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46798116, end_position_on_the_genomic_accession=46890403, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  735271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735271, CurrentID=0, Status='live', Symbol='Mir709', Aliases='Mirn709, mir-709, mmu-mir-709', description='microRNA 709', other_designations='MIR 709', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84086099, end_position_on_the_genomic_accession=84086186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  735275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735275, CurrentID=0, Status='live', Symbol='Mir499', Aliases='Mirn499, mir-499, mmu-mir-499', description='microRNA 499', other_designations='mir 499', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155622880, end_position_on_the_genomic_accession=155622958, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  735279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735279, CurrentID=0, Status='live', Symbol='Mir490', Aliases='Mirn490, mir-490, mmu-mir-490', description='microRNA 490', other_designations='mir 490', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=36421742, end_position_on_the_genomic_accession=36421825, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  210992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210992, CurrentID=0, Status='live', Symbol='Lpcat1', Aliases='2900035H07Rik, Aytl2, BB137372, BC005662, C87117, LPCAT, rd11', description='lysophosphatidylcholine acyltransferase 1', other_designations='1-acylglycerophosphocholine O-acyltransferase|1-alkylglycerophosphocholine O-acetyltransferase|LPC acyltransferase 1|LPCAT-1|acetyl-CoA:lyso-PAF acetyltransferase|acetyl-CoA:lyso-platelet-activating factor acetyltransferase|acyltransferase like 2|acyltransferase-like 2|lyso-PAF acetyltransferase|lysoPAFAT|lysoPC acyltransferase 1|mLPCAT1', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73467204, end_position_on_the_genomic_accession=73516420, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  735283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735283, CurrentID=0, Status='live', Symbol='Mir702', Aliases='Mirn702, mir-702, mmu-mir-702', description='microRNA 702', other_designations='mir 702', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136991433, end_position_on_the_genomic_accession=136991541, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  735284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735284, CurrentID=0, Status='live', Symbol='Mir708', Aliases='Mirn708, mmu-mir-708', description='microRNA 708', other_designations='mir 708', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=96249424, end_position_on_the_genomic_accession=96249532, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  210998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210998, CurrentID=0, Status='live', Symbol='Fam91a1', Aliases='AV220772, BC033609, D15Ertd621e, SKIN, mKIAA0493', description='family with sequence similarity 91, member A1', other_designations='skeletal muscle cells re-entry induced', map_location='15 24.76 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58405908, end_position_on_the_genomic_accession=58457801, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  211006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211006, CurrentID=0, Status='live', Symbol='Sepsecs', Aliases='9130208G10, AA986712, D5Ertd135e, SLA, SecS', description='Sep (O-phosphoserine) tRNA:Sec (selenocysteine) tRNA synthase', other_designations='UGA suppressor tRNA-associated protein|sec synthase|selenocysteine synthase|selenocysteinyl-tRNA(Sec) synthase|sep-tRNA:Sec-tRNA synthase', map_location='5 28.44 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52643407, end_position_on_the_genomic_accession=52669770, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  211007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211007, CurrentID=0, Status='live', Symbol='Trim41', Aliases='AW552703, BC020156, R75223, RINCK', description='tripartite motif-containing 41', other_designations='tripartite motif-containing protein 41', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48806403, end_position_on_the_genomic_accession=48817765, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  57014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57014, CurrentID=0, Status='live', Symbol='Htr3b', Aliases='5-HT3B', description='5-hydroxytryptamine (serotonin) receptor 3B', other_designations='5-HT3 receptor subunit B|5-HT3-B|serotonin receptor 3B', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48935008, end_position_on_the_genomic_accession=48964990, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  735308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735308, CurrentID=0, Status='live', Symbol='Mir216b', Aliases='Mirn216b, mir-216b', description='microRNA 216b', other_designations='mmu-mir-216b', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=28746191, end_position_on_the_genomic_accession=28746276, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  211064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211064, CurrentID=0, Status='live', Symbol='Alkbh1', Aliases='2700073G19Rik, Abh, Alkbh, alkB, hABH', description='alkB homolog 1, histone H2A dioxygenase', other_designations='Alkylated DNA repair protein alkB homolog 1|DNA lyase ABH1|alkB, alkylation repair homolog 1|alpha-ketoglutarate-dependent dioxygenase ABH1', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87428077, end_position_on_the_genomic_accession=87443839, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  57028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57028, CurrentID=0, Status='live', Symbol='Pdxp', Aliases='1600027H05Rik, AB041662, PLPP', description='pyridoxal (pyridoxine, vitamin B6) phosphatase', other_designations='PLP phosphatase|PLP-phosophatase|chronophilin|chronophin', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78913919, end_position_on_the_genomic_accession=78919517, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  225256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225256, CurrentID=0, Status='live', Symbol='Dsg1b', Aliases='Dsg5, dsg1-beta', description='desmoglein 1 beta', other_designations='desmoglein 5', map_location='18 11.32 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20376835, end_position_on_the_genomic_accession=20412180, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  211134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211134, CurrentID=0, Status='live', Symbol='Lzts1', Aliases='F37, FEZ1', description='leucine zipper, putative tumor suppressor 1', other_designations='F37 esophageal cancer-related leucine zipper motif 1|F37/Esophageal cancer-related gene-coding leucine-zipper motif|leucine-zipper protein FEZ1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69132669, end_position_on_the_genomic_accession=69184220, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  211135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211135, CurrentID=0, Status='live', Symbol='D130040H23Rik', Aliases='', description='RIKEN cDNA D130040H23 gene', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69271055, end_position_on_the_genomic_accession=69303384, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  211147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211147, CurrentID=0, Status='live', Symbol='March11', Aliases='9630025C22', description='membrane-associated ring finger (C3HC4) 11', other_designations='MARCH-XI|membrane-associated RING finger protein 11|membrane-associated RING-CH protein XI', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=26308904, end_position_on_the_genomic_accession=26409594, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  211151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211151, CurrentID=0, Status='live', Symbol='Churc1', Aliases='1700030P15Rik, My015, chch', description='churchill domain containing 1', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76765573, end_position_on_the_genomic_accession=76783180, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  211187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211187, CurrentID=0, Status='live', Symbol='Lrtm2', Aliases='A230084J22, AI841794', description='leucine-rich repeats and transmembrane domains 2', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119315133, end_position_on_the_genomic_accession=119332270, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  211208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211208, CurrentID=0, Status='live', Symbol='Gm382', Aliases='', description='predicted gene 382', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=127039972, end_position_on_the_genomic_accession=127063986, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  209966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209966, CurrentID=0, Status='live', Symbol='Pgbd5', Aliases='2900019M05Rik, AI854313', description='piggyBac transposable element derived 5', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124369049, end_position_on_the_genomic_accession=124434462, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  211228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211228, CurrentID=0, Status='live', Symbol='Lrrc25', Aliases='Mapa', description='leucine rich repeat containing 25', other_designations='monocyte and plasmacytoid activated molecule|monocyte and plasmacytoid-activated protein', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70616844, end_position_on_the_genomic_accession=70620850, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  211232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211232, CurrentID=0, Status='live', Symbol='Cpne9', Aliases='A730016F12Rik, mKIAA4217', description='copine family member IX', other_designations='copine IX', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113282266, end_position_on_the_genomic_accession=113305571, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  668723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668723, CurrentID=0, Status='live', Symbol='Amy2-ps1', Aliases='EG668723', description='amylase 2, pseudogene 1', other_designations='', map_location='3 49.35 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113182234, end_position_on_the_genomic_accession=113206694, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  211253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211253, CurrentID=0, Status='live', Symbol='Mtrf1', Aliases='A830062K05Rik, MtRF-1', description='mitochondrial translational release factor 1', other_designations='MRF-1', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79397761, end_position_on_the_genomic_accession=79423650, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  211255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211255, CurrentID=0, Status='live', Symbol='Kbtbd7', Aliases='1110008P08Rik, AU045792, Gm295', description='kelch repeat and BTB (POZ) domain containing 7', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=79426511, end_position_on_the_genomic_accession=79431039, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  211286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211286, CurrentID=0, Status='live', Symbol='Cln5', Aliases='A730075N08Rik', description='ceroid-lipofuscinosis, neuronal 5', other_designations='', map_location='14 51.71 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=103070216, end_position_on_the_genomic_accession=103077630, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  211305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211305, CurrentID=0, Status='live', Symbol='Fbxw13', Aliases='E330017E16', description='F-box and WD-40 domain protein 13', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109179227, end_position_on_the_genomic_accession=109195975, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  211323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211323, CurrentID=0, Status='live', Symbol='Nrg1', Aliases='6030402G23Rik, ARIA, D230005F13Rik, GGF, GGFII, HRG, HRGalpha, Hgl, NDF, SMDF', description='neuregulin 1', other_designations='heregulin|type I neuregulin 1|type III neuregulin 1', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=31807452, end_position_on_the_genomic_accession=32890583, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  211329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211329, CurrentID=0, Status='live', Symbol='Ncoa7', Aliases='9030406N13Rik', description='nuclear receptor coactivator 7', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=30645582, end_position_on_the_genomic_accession=30805977, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  211331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211331, CurrentID=0, Status='live', Symbol='Ighv1-54', Aliases='EG211331, Gm4770, VH', description='immunoglobulin heavy variable V1-54', other_designations='monoclonal antibody heavy chain', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=115193675, end_position_on_the_genomic_accession=115193968, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  78913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=78913, CurrentID=0, Status='live', Symbol='Ltn1', Aliases='4930528H02Rik, AV266914, C87237, Listerin, Rnf160, Zfp294', description='listerin E3 ubiquitin protein ligase 1', other_designations='ring finger protein 160|zfp-294|zinc finger protein 294', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87376651, end_position_on_the_genomic_accession=87432619, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  211347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211347, CurrentID=0, Status='live', Symbol='Pank3', Aliases='', description='pantothenate kinase 3', other_designations='mPanK3|pantothenic acid kinase 3', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=35769495, end_position_on_the_genomic_accession=35791285, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  80280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80280, CurrentID=0, Status='live', Symbol='Cdk5rap3', Aliases='1810007E24Rik, BC002318, C53, C81486, HSF-27, IC53, LZAP, MST016', description='CDK5 regulatory subunit associated protein 3', other_designations='OK/SW-cl.114', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96907763, end_position_on_the_genomic_accession=96916530, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  80281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80281, CurrentID=0, Status='live', Symbol='Cttnbp2nl', Aliases='AA552995, AA589392, AU018624, BC003236, mKIAA1433', description='CTTNBP2 N-terminal like', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105001915, end_position_on_the_genomic_accession=105053146, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  80283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80283, CurrentID=0, Status='live', Symbol='Abtb1', Aliases='AI847549, BC003234, BPOZ, EF1ABP', description='ankyrin repeat and BTB (POZ) domain containing 1', other_designations='abtb1|mFLJ00331 protein', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88835914, end_position_on_the_genomic_accession=88842386, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  80284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80284, CurrentID=0, Status='live', Symbol='Smim12', Aliases='BC003266', description='small integral membrane protein 12', other_designations='UPF0767 protein C1orf212 homolog', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127243784, end_position_on_the_genomic_accession=127247809, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  80285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80285, CurrentID=0, Status='live', Symbol='Parp9', Aliases='ARTD9, AW214463, BC003281, Bagl, Bal, PARP-9', description='poly (ADP-ribose) polymerase family, member 9', other_designations='ADP-ribosyltransferase diphtheria toxin-like 9|b aggressive lymphoma protein homolog', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35938470, end_position_on_the_genomic_accession=35972621, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  80286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80286, CurrentID=0, Status='live', Symbol='Tusc3', Aliases='AU022242, BC003311, N33', description='tumor suppressor candidate 3', other_designations='magnesium uptake/transporter TUSC3|putative prostate cancer tumor suppressor', map_location='8 B1.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=39157260, end_position_on_the_genomic_accession=39165114, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  80287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80287, CurrentID=0, Status='live', Symbol='Apobec3', Aliases='Apobec, Arp3, BC003314, Cem15, Rfv3', description='apolipoprotein B mRNA editing enzyme, catalytic polypeptide 3', other_designations='apolipoprotein B editing complex 3|apolipoprotein B mRNA-editing complex 3|recovery from Friend virus 3', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79892408, end_position_on_the_genomic_accession=79908905, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  80288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80288, CurrentID=0, Status='live', Symbol='Bcl9l', Aliases='B9L, BC003321, BCL9-2, DLNB11', description='B cell CLL/lymphoma 9-like', other_designations='B-cell lymphoma 9-like protein|BCL9-like protein|BCL9-related beta-catenin-binding protein|nuclear co-factor of beta-catenin signalling|protein BCL9-2', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44483113, end_position_on_the_genomic_accession=44511901, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  80289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80289, CurrentID=0, Status='live', Symbol='Lysmd3', Aliases='1110030H10Rik, BC003322', description='LysM, putative peptidoglycan-binding, domain containing 3', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=81657657, end_position_on_the_genomic_accession=81672871, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  80290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80290, CurrentID=0, Status='live', Symbol='Gpr146', Aliases='BC003323, PGR8', description='G protein-coupled receptor 146', other_designations='', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139377698, end_position_on_the_genomic_accession=139396415, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  80291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80291, CurrentID=0, Status='live', Symbol='Rilpl2', Aliases='', description='Rab interacting lysosomal protein-like 2', other_designations='rab-interacting lysosomal-like protein 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124463265, end_position_on_the_genomic_accession=124478235, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  80292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80292, CurrentID=0, Status='live', Symbol='Zxdc', Aliases='A930012H20, B930086F11Rik, BC003332', description='ZXD family zinc finger C', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90369483, end_position_on_the_genomic_accession=90403490, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  80294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80294, CurrentID=0, Status='live', Symbol='Pofut2', Aliases='2310011G23Rik, AI256847, BC003494, C21orf80, FUT13', description='protein O-fucosyltransferase 2', other_designations='O-FucT-2|peptide-O-fucosyltransferase 2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77259300, end_position_on_the_genomic_accession=77269586, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  80297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80297, CurrentID=0, Status='live', Symbol='Sptbn4', Aliases='1700022P15Rik, 5830426A08Rik, ROSA62, SpbIV, Spnb4, dyn, lnd, nmf261, nmf379, qv', description='spectrin beta, non-erythrocytic 4', other_designations='beta-spectrin 4|lumbosacral neuroaxonal dystrophy|neuroaxonal dystrophy|quivering|spectrin beta 4', map_location='7 15.88 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27356383, end_position_on_the_genomic_accession=27447704, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  211383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211383, CurrentID=0, Status='live', Symbol='Amer3', Aliases='9430069J07Rik, Fam123c', description='APC membrane recruitment 3', other_designations='family with sequence similarity 123, member C|protein FAM123C', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34578935, end_position_on_the_genomic_accession=34590948, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  211389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211389, CurrentID=0, Status='live', Symbol='Suox', Aliases='SO', description='sulfite oxidase', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128669887, end_position_on_the_genomic_accession=128673918, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  211401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211401, CurrentID=0, Status='live', Symbol='Mtss1', Aliases='2310003N14Rik, BC024131, D130001D01Rik, Mim, mKIAA0429', description='metastasis suppressor 1', other_designations='actin monomer-binding protein|missing in metastasis protein', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=58941234, end_position_on_the_genomic_accession=59082325, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  80334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80334, CurrentID=0, Status='live', Symbol='Kcnip4', Aliases='AV032399, Calp, Calp250, KChIP4, KChIP4a', description='Kv channel interacting protein 4', other_designations='A-type potassium channel modulatory protein 4|calsenilin-like protein|potassium channel interacting protein 4|potassium channel-interacting protein 4', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=48389503, end_position_on_the_genomic_accession=49525122, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  211429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211429, CurrentID=0, Status='live', Symbol='Pla2g4b', Aliases='A030011C02Rik, Gm1013, cPLA2-beta', description='phospholipase A2, group IVB (cytosolic)', other_designations='phospholipase A2 group IVB', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120033433, end_position_on_the_genomic_accession=120043033, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  211446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211446, CurrentID=0, Status='live', Symbol='Exoc3', Aliases='2810050O03Rik, E430013E20Rik, Sec6, Sec6l1', description='exocyst complex component 3', other_designations='SEC6-like 1|exocyst complex component Sec6', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74169796, end_position_on_the_genomic_accession=74208729, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  80384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80384, CurrentID=0, Status='live', Symbol='Tex21', Aliases='4931406F04Rik, 4931412D23Rik, 4931421K24Rik, tsec-2', description='testis expressed gene 21', other_designations='testis expressed gene 11', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76198687, end_position_on_the_genomic_accession=76246758, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  80385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80385, CurrentID=0, Status='live', Symbol='Tusc2', Aliases='1190001E22Rik, AA407686, Fus1, Lgcc, PAP, Pdap2', description='tumor suppressor candidate 2', other_designations='PDGFA associated protein 2|PDGFA-associated protein 2|fus-1 protein|fusion 1 protein|lung cancer candidiate, FUS', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107563255, end_position_on_the_genomic_accession=107566108, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  211468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211468, CurrentID=0, Status='live', Symbol='Kcnh8', Aliases='C130090D05Rik, ELK, ELK1, ELK3, Kv12.1', description='potassium voltage-gated channel, subfamily H (eag-related), member 8', other_designations='ELK channel 3|ether-a-go-go-like potassium channel 3|potassium voltage-gated channel, subfamily H, member 8|voltage-gated potassium channel subunit Kv12.1', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=52598233, end_position_on_the_genomic_accession=52979864, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  210009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210009, CurrentID=0, Status='live', Symbol='Mtrr', Aliases='4732420G08, MSR', description='5-methyltetrahydrofolate-homocysteine methyltransferase reductase', other_designations='', map_location='13 35.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=68560780, end_position_on_the_genomic_accession=68582121, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  211480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211480, CurrentID=0, Status='live', Symbol='Kcnj14', Aliases='A930026G01Rik, IRK4, Kir2.4', description='potassium inwardly-rectifying channel, subfamily J, member 14', other_designations='IRK-4|inward rectifier K(+) channel Kir2.4|potassium channel, inwardly rectifying subfamily J member 14', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45816467, end_position_on_the_genomic_accession=45824747, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  211482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211482, CurrentID=0, Status='live', Symbol='Efhb', Aliases='4921525D22Rik', description='EF hand domain family, member B', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=53398048, end_position_on_the_genomic_accession=53463334, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  211484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211484, CurrentID=0, Status='live', Symbol='Tsga10', Aliases='4933432N21Rik, AI843383, Gm217, Mtsga10', description='testis specific 10', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37728144, end_position_on_the_genomic_accession=37866739, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  211488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211488, CurrentID=0, Status='live', Symbol='Ado', Aliases='Gm237', description='2-aminoethanethiol (cysteamine) dioxygenase', other_designations='cysteamine dioxygenase', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=67544511, end_position_on_the_genomic_accession=67548955, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  211499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211499, CurrentID=0, Status='live', Symbol='Tmem87a', Aliases='A930025J12Rik', description='transmembrane protein 87A', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120355309, end_position_on_the_genomic_accession=120404116, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  211535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211535, CurrentID=0, Status='live', Symbol='Ccdc114', Aliases='', description='coiled-coil domain containing 114', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45921526, end_position_on_the_genomic_accession=45948963, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  211548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211548, CurrentID=0, Status='live', Symbol='Nomo1', Aliases='D7Ertd156e, Nomo, pM5', description='nodal modulator 1', other_designations='nomo1', map_location='7 29.66 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46033696, end_position_on_the_genomic_accession=46084212, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  211550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211550, CurrentID=0, Status='live', Symbol='Tifa', Aliases='T2bp', description='TRAF-interacting protein with forkhead-associated domain', other_designations='TRAF2-binding protein|Traf2 binding protein', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127788857, end_position_on_the_genomic_accession=127798394, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  211556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211556, CurrentID=0, Status='live', Symbol='Ap1ar', Aliases='AA407621, C78909, Gadkin', description='adaptor-related protein complex 1 associated regulatory protein', other_designations='', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=127807007, end_position_on_the_genomic_accession=127837518, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  211577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211577, CurrentID=0, Status='live', Symbol='Mrgprf', Aliases='BC019711, Mgrc, MrgF', description='MAS-related GPR, member F', other_designations='G protein-coupled receptor RTA|Mas-related G protein-coupled receptor c|mas-related gene F protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=145300909, end_position_on_the_genomic_accession=145309557, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  211578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211578, CurrentID=0, Status='live', Symbol='Mrgprd', Aliases='Gm499, Mgrd, MrgD, TGR7', description='MAS-related GPR, member D', other_designations='G protein-coupled receptor MrgD|G-protein coupled receptor TGR7|Mas-related G protein-coupled receptor d|beta-alanine receptor', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=145314835, end_position_on_the_genomic_accession=145324059, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  80509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80509, CurrentID=0, Status='live', Symbol='Med8', Aliases='2210021A15Rik, AB041805, ARC32', description='mediator complex subunit 8', other_designations='activator-recruited cofactor 32 kDa component|mediator of RNA polymerase II transcription, subunit 8 homolog', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118409298, end_position_on_the_genomic_accession=118415824, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  211586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211586, CurrentID=0, Status='live', Symbol='Tfdp2', Aliases='1110029I05Rik, A330080J22Rik, DP-3, DP3', description='transcription factor Dp 2', other_designations='E2F dimerization partner 2', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96196275, end_position_on_the_genomic_accession=96323646, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  80517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80517, CurrentID=0, Status='live', Symbol='Herpud2', Aliases='5031400M07Rik, AB041580', description='HERPUD family member 2', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=25108124, end_position_on_the_genomic_accession=25151781, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  211612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211612, CurrentID=0, Status='live', Symbol='Ptchd1', Aliases='9630036J22Rik, AI848149, Gm387', description='patched domain containing 1', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=155569736, end_position_on_the_genomic_accession=155623812, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  211623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211623, CurrentID=0, Status='live', Symbol='Plac9a', Aliases='Plac9', description='placenta specific 9a', other_designations='placenta specific 9', map_location='', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25888403, end_position_on_the_genomic_accession=25902893, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  211651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211651, CurrentID=0, Status='live', Symbol='Fancd2', Aliases='2410150O07Rik, AU015151, BB137857, FA-D2, FA4, FACD, FAD, FANCD', description='Fanconi anemia, complementation group D2', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113531661, end_position_on_the_genomic_accession=113597018, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  211652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211652, CurrentID=0, Status='live', Symbol='Wwc1', Aliases='AA408228, AU017197, BC037006, Kibra, mKIAA0869', description='WW, C2 and coiled-coil domain containing 1', other_designations='WW domain-containing protein 1|kidney and brain protein|protein WWC1', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=35839178, end_position_on_the_genomic_accession=35980089, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  211660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211660, CurrentID=0, Status='live', Symbol='Cspp1', Aliases='2310020J12Rik, 4930413O22Rik, AA387550, BC055915, Cspp2, Cspp3', description='centrosome and spindle pole associated protein 1', other_designations='centrosome/spindle pole-associated protein 2|centrosome/spindle pole-associated protein 3', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=10038218, end_position_on_the_genomic_accession=10136771, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  211666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211666, CurrentID=0, Status='live', Symbol='Mgst2', Aliases='GST2, MGST-II', description='microsomal glutathione S-transferase 2', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=51660360, end_position_on_the_genomic_accession=51682675, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  211673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211673, CurrentID=0, Status='live', Symbol='Arfgef1', Aliases='ARFGEP1, BIG1, D130059B05Rik, D730028O18Rik, P200', description='ADP-ribosylation factor guanine nucleotide-exchange factor 1(brefeldin A-inhibited)', other_designations='ADP-ribosylation factor guanine nucleotide-exchange factor 1|brefeldin A-inhibited GEP 1', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=10137507, end_position_on_the_genomic_accession=10232670, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  13435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13435, CurrentID=0, Status='live', Symbol='Dnmt3a', Aliases='MmuIIIA', description='DNA methyltransferase 3A', other_designations='DNA MTase MmuIIIA|DNA methyltransferase MmuIIIA', map_location='12 A2-A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=3804535, end_position_on_the_genomic_accession=3914443, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  211712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211712, CurrentID=0, Status='live', Symbol='Pcdh9', Aliases='', description='protocadherin 9', other_designations='', map_location='14 45.8 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=93013700, end_position_on_the_genomic_accession=93890721, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  211739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211739, CurrentID=0, Status='live', Symbol='Vstm2a', Aliases='Vstm2', description='V-set and transmembrane domain containing 2A', other_designations='', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=16257602, end_position_on_the_genomic_accession=16284551, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  207819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207819, CurrentID=0, Status='live', Symbol='4930539E08Rik', Aliases='9130014G02', description='RIKEN cDNA 4930539E08 gene', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28896395, end_position_on_the_genomic_accession=28915336, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  211770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211770, CurrentID=0, Status='live', Symbol='Trib1', Aliases='A530090O15Rik, TRB-1, Trb1', description='tribbles homolog 1 (Drosophila)', other_designations='trib1', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=59648654, end_position_on_the_genomic_accession=59657099, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  80706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80706, CurrentID=0, Status='live', Symbol='Olfr160', Aliases='M72, MOR171-3, Olfr7b', description='olfactory receptor 160', other_designations='GA_x6K02T2PVTD-31389446-31388517|olfactory receptor 7b|olfactory receptor MOR171-3', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37711348, end_position_on_the_genomic_accession=37712280, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80707, CurrentID=0, Status='live', Symbol='Wwox', Aliases='5330426P09Rik, 9030416C10Rik, WOX1', description='WW domain-containing oxidoreductase', other_designations='WW-domain oxidoreductase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=114439630, end_position_on_the_genomic_accession=115352712, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  80708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80708, CurrentID=0, Status='live', Symbol='Pacsin3', Aliases='4921507A02Rik, 6330413E15Rik, AW413130', description='protein kinase C and casein kinase substrate in neurons 3', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91256098, end_position_on_the_genomic_accession=91264680, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  80718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80718, CurrentID=0, Status='live', Symbol='Rab27b', Aliases='2310021G14Rik, B130064M09Rik', description='RAB27B, member RAS oncogene family', other_designations='', map_location='18 43.89 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=69979131, end_position_on_the_genomic_accession=70141633, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  80719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80719, CurrentID=0, Status='live', Symbol='Igsf6', Aliases='', description='immunoglobulin superfamily, member 6', other_designations='', map_location='7 F2-F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121064067, end_position_on_the_genomic_accession=121074903, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  80720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80720, CurrentID=0, Status='live', Symbol='Pbx4', Aliases='2410015M21Rik, AI429113, AW045266', description='pre B cell leukemia homeobox 4', other_designations='homeobox protein PBX4|pre-B-cell leukemia homeobox 4', map_location='8 C', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69832476, end_position_on_the_genomic_accession=69872292, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  80721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80721, CurrentID=0, Status='live', Symbol='Slc19a3', Aliases='A230084E24Rik, AI788884, ThTr2', description='solute carrier family 19, member 3', other_designations='solute carrier family 19 (sodium/hydrogen exchanger), member 3|thTr-2', map_location='1 42.65 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=83012523, end_position_on_the_genomic_accession=83038859, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  211798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211798, CurrentID=0, Status='live', Symbol='Mfsd9', Aliases='4931419K03Rik', description='major facilitator superfamily domain containing 9', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40772040, end_position_on_the_genomic_accession=40790657, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  80732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80732, CurrentID=0, Status='live', Symbol='Mynn', Aliases='2810011C24Rik, AA415053, AI661031, AW049392, Mynr, SBBIZ1', description='myoneurin', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30602087, end_position_on_the_genomic_accession=30619873, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  80733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80733, CurrentID=0, Status='live', Symbol='Car15', Aliases='AI315043, Ca15, Cals2', description='carbonic anhydrase 15', other_designations='CA-XV|carbonate dehydratase XV|carbonic anhydrase XV|carbonic anhydrase-like sequence 2', map_location='16 11.05 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17835276, end_position_on_the_genomic_accession=17838186, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  80743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80743, CurrentID=0, Status='live', Symbol='Vps16', Aliases='1810074M16Rik', description='vacuolar protein sorting 16 (yeast)', other_designations='mVPS16', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130424304, end_position_on_the_genomic_accession=130444269, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  80744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80744, CurrentID=0, Status='live', Symbol='Cwc22', Aliases='AA684037, AI173004, AL022752, B230213M24, mKIAA1604', description='CWC22 spliceosome-associated protein', other_designations='CWC22 spliceosome-associated protein homolog|nucampholin homolog', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=77895653, end_position_on_the_genomic_accession=77946356, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  80748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80748, CurrentID=0, Status='live', Symbol='BC004004', Aliases='2400006G15Rik', description='cDNA sequence BC004004', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29268788, end_position_on_the_genomic_accession=29302887, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  80749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80749, CurrentID=0, Status='live', Symbol='Lrfn1', Aliases='BC004018, SALM2, Semo1', description='leucine rich repeat and fibronectin type III domain containing 1', other_designations='synaptic adhesion-like molecule 2|synaptic differentiation enhancing molecule 1|synaptic differentiation-enhancing molecule 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28451985, end_position_on_the_genomic_accession=28468242, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  80750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80750, CurrentID=0, Status='live', Symbol='N4bp1', Aliases='AI481586, C81621', description='NEDD4 binding protein 1', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=86841136, end_position_on_the_genomic_accession=86885258, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  80751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80751, CurrentID=0, Status='live', Symbol='Rnf34', Aliases='AW061037, AW536122, BC004042, C88279, RIFF', description='ring finger protein 34', other_designations='RING finger protein RIFF|phafin 1|phafin-1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122850162, end_position_on_the_genomic_accession=122868945, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80752, CurrentID=0, Status='live', Symbol='Fam20c', Aliases='C76981, DMP-4, DMP4, GEF-CK, mKIAA4081', description='family with sequence similarity 20, member C', other_designations='dentin matrix protein 4|golgi-enriched fraction casein kinase', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138754306, end_position_on_the_genomic_accession=138810072, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  80782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80782, CurrentID=0, Status='live', Symbol='Klrb1b', Aliases='Klrb1d, Ly55b, Ly55d, NKR-P1B, NKR-P1D, Nkrp1-b, Nkrp1d', description='killer cell lectin-like receptor subfamily B member 1B', other_designations='CD161 antigen-like family member B|NKR-P1 34|inhibitory receptor NKR-P1B|killer cell lectin-like receptor subfamily B member 1B allele A|killer cell lectin-like receptor subfamily B member 1D|ly-55b|ly-55d|lymphocyte antigen 55 complex, locus B|lymphocyte antigen 55 complex, locus D|lymphocyte antigen 55b|lymphocyte antigen 55d|natural killer cell receptor protein NKR-P1B|natural killer cell receptor-P1|natural killer cell surface protein NKR-P1B allele B6|natural killer cell surface protein NKR-P1B allele SJL/BALB', map_location='6 63.09 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128813706, end_position_on_the_genomic_accession=128826375, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  80794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80794, CurrentID=0, Status='live', Symbol='Cblc', Aliases='2310076I21Rik, 2310079L19Rik, Cbl3', description='Casitas B-lineage lymphoma c', other_designations='CBL-3|SH3-binding protein CBL-3|SH3-binding protein CBL-C|signal transduction protein CBL-C', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19779718, end_position_on_the_genomic_accession=19796809, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  80795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80795, CurrentID=0, Status='live', Symbol='Selk', Aliases='1110001C03Rik, AA673253, HSPC030, Hsp30', description='selenoprotein K', other_designations='heat shock protein, 30 kDa', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=29968380, end_position_on_the_genomic_accession=29975074, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  80796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80796, CurrentID=0, Status='live', Symbol='Calm4', Aliases='2310037J09Rik, AI649108, DD112, Scarf', description='calmodulin 4', other_designations='calcium binding gene, DD112|calcium-binding protein Dd112|skin calmodulin-related factor', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3837757, end_position_on_the_genomic_accession=3838671, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  80797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80797, CurrentID=0, Status='live', Symbol='Clca3a2', Aliases='Clca2', description='chloride channel accessory 3A2', other_designations='chloride channel calcium activated 2', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144796559, end_position_on_the_genomic_accession=144819629, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  211896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211896, CurrentID=0, Status='live', Symbol='Depdc7', Aliases='AV216087', description='DEP domain containing 7', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104721784, end_position_on_the_genomic_accession=104742801, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  80837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80837, CurrentID=0, Status='live', Symbol='Rhoj', Aliases='1110005O19Rik, AW210585, Arhj, TC10L, TCL', description='ras homolog gene family, member J', other_designations='TC10-like Rho GTPase|tc10-like GTP-binding protein TCL', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=75308313, end_position_on_the_genomic_accession=75401455, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  80838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80838, CurrentID=0, Status='live', Symbol='Hist1h1a', Aliases='H1.1, H1a, H1f1, H1var3', description='histone cluster 1, H1a', other_designations='H1 VAR.3|H1 histone family, member 1|histone 1, H1a|histone H1a', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23763668, end_position_on_the_genomic_accession=23764412, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  211914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211914, CurrentID=0, Status='live', Symbol='Asap2', Aliases='6530401G17Rik, Ddef2, Gm1523, Gm592, PAG3, PAP', description='ArfGAP with SH3 domain, ankyrin repeat and PH domain 2', other_designations='development and differentiation enhancing factor 2|development and differentiation-enhancing factor 2|paxillin-associated protein with ARF GAP activity 3|pyk2 C-terminus-associated protein', map_location='12 A1.2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=21111713, end_position_on_the_genomic_accession=21270171, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  211922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211922, CurrentID=0, Status='live', Symbol='Dennd6a', Aliases='Fam116a', description='DENN/MADD domain containing 6A', other_designations='DENN domain-containing protein 6A|family with sequence similarity 116, member A|protein FAM116A', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26573858, end_position_on_the_genomic_accession=26634322, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  211924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211924, CurrentID=0, Status='live', Symbol='Dsg1c', Aliases='Dsg1gamma, Dsg6, dsg1g', description='desmoglein 1 gamma', other_designations='desmoglein 6', map_location='18 11.26 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=20247340, end_position_on_the_genomic_accession=20283923, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  80857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80857, CurrentID=0, Status='live', Symbol='Fgf20', Aliases='', description='fibroblast growth factor 20', other_designations='FGF-20', map_location='8 B1.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40279166, end_position_on_the_genomic_accession=40286953, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  80859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80859, CurrentID=0, Status='live', Symbol='Nfkbiz', Aliases='AA408868, INAP, Mail', description='nuclear factor of kappa light polypeptide gene enhancer in B cells inhibitor, zeta', other_designations='I-kappa-B-zeta|IL-1 inducible nuclear ankyrin-repeat protein|IkappaB-zeta|ikB-zeta|ikappaBzeta|molecule possessing ankyrin repeats induced by lipopolysaccharide|molecule possessing ankyrin-repeats induced by lipopolysaccharide|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, zeta', map_location='16 C1.2-C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=55811377, end_position_on_the_genomic_accession=55838641, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  80860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80860, CurrentID=0, Status='live', Symbol='Ghdc', Aliases='D11Lgp1e', description='GH3 domain containing', other_designations='D11lgp1', map_location='11 63.6 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100766026, end_position_on_the_genomic_accession=100770982, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  80861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80861, CurrentID=0, Status='live', Symbol='Dhx58', Aliases='B430001I08Rik, D11Lgp2e, LPG2, Lgp2, RLR-3', description='DEXH (Asp-Glu-X-His) box polypeptide 58', other_designations='RIG-I-like receptor 3|RIG-I-like receptor Lgp2|RLR|RNA helicase LGP2|probable ATP-dependent helicase LGP2|protein D11Lgp2', map_location='11 63.52 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100694884, end_position_on_the_genomic_accession=100704271, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  211936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211936, CurrentID=0, Status='live', Symbol='Ccdc73', Aliases='2210415I11Rik, 4931417A20, AV080665, BC095976', description='coiled-coil domain containing 73', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104886324, end_position_on_the_genomic_accession=104999737, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  211945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211945, CurrentID=0, Status='live', Symbol='Plekhh1', Aliases='AV308913, C77772, D630024D12Rik, mKIAA1200', description='pleckstrin homology domain containing, family H (with MyTH4 domain) member 1', other_designations='PH domain-containing family H member 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79029163, end_position_on_the_genomic_accession=79081655, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  80876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80876, CurrentID=0, Status='live', Symbol='Ifitm2', Aliases='DSPA2c, Ifitm3l, fragilis3, mil-3', description='interferon induced transmembrane protein 2', other_designations='dispanin subfamily A member 2c|fragilis R|fragilis protein 3|interferon induced transmembrane protein, like', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140954839, end_position_on_the_genomic_accession=140955961, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  80877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80877, CurrentID=0, Status='live', Symbol='Lrba', Aliases='C80285, D3Ertd775e, Lab300, Lba', description='LPS-responsive beige-like anchor', other_designations='beige-like protein', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=86223568, end_position_on_the_genomic_accession=86782705, orientation='plus', exon_count=60, OMIM='', no_hdr0=''),
  80879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80879, CurrentID=0, Status='live', Symbol='Slc16a3', Aliases='Mct3, Mct4', description='solute carrier family 16 (monocarboxylic acid transporters), member 3', other_designations='MCT 4|solute carrier family 16 member 3', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120948484, end_position_on_the_genomic_accession=120959000, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  80880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80880, CurrentID=0, Status='live', Symbol='Kank3', Aliases='0610013D04Rik, Ankrd47, D17Ertd288e, NG28', description='KN motif and ankyrin repeat domains 3', other_designations='ankyrin repeat domain 47|ankyrin repeat domain-containing protein 47|putative KN motif and ankyrin repeat domains 3', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33810523, end_position_on_the_genomic_accession=33822920, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  80883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80883, CurrentID=0, Status='live', Symbol='Ntng1', Aliases='A930010C08Rik, AI853992, Lmnt1', description='netrin G1', other_designations='laminet-1', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=109780050, end_position_on_the_genomic_accession=110143472, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  80884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80884, CurrentID=0, Status='live', Symbol='Maged2', Aliases='4833439A22Rik, Mage-d2, ORF1', description='melanoma antigen, family D, 2', other_designations='', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=150806421, end_position_on_the_genomic_accession=150814340, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  80885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80885, CurrentID=0, Status='live', Symbol='Hcar2', Aliases='Gpr109a, Gpr109b, HM74, Niacr1, PUMA-G, Pumag, mHM74b', description='hydroxycarboxylic acid receptor 2', other_designations='G protein-coupled receptor 109A|G protein-coupled receptor 109B|G-protein coupled receptor 109|G-protein coupled receptor HM74|interferon-gamma inducible gene, Puma-g|niacin receptor 1|nicotinic acid receptor|putative seven transmembrane spanning receptor', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123863570, end_position_on_the_genomic_accession=123865516, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  80886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80886, CurrentID=0, Status='live', Symbol='Senp3', Aliases='AA408656, Smt3ip, Smt3ip1', description='SUMO/sentrin specific peptidase 3', other_designations='SUMO-1-specific protease 3|SUMO/sentrin specific protease 3|sentrin/SUMO-specific protease SENP3|smt3-specific isopeptidase 1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69673110, end_position_on_the_genomic_accession=69682084, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  80888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80888, CurrentID=0, Status='live', Symbol='Hspb8', Aliases='AU018630, AW413033, Cryac, D5Ucla4, E2IG1, H11, H11K, HSP20-like, HSP22', description='heat shock protein 8', other_designations='alpha-crystallin C chain|crystallin, alpha C|heat shock 27kDa protein 8|heat shock protein 20-like protein|small stress protein-like protein HSP22', map_location='5 56.41 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=116408495, end_position_on_the_genomic_accession=116422864, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  211961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211961, CurrentID=0, Status='live', Symbol='Asxl3', Aliases='C230079D11Rik, D430002O22Rik, D930044O18Rik, Gm329, Gm945', description='additional sex combs like 3 (Drosophila)', other_designations='additional sex combs-like protein 3|polycomb group protein', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=22344077, end_position_on_the_genomic_accession=22530227, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  80890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80890, CurrentID=0, Status='live', Symbol='Trim2', Aliases='mKIAA0517, narf', description='tripartite motif-containing 2', other_designations='E3 ubiquitin-protein ligase TRIM2|neural activity-related ring finger protein|tripartite motif protein 2|tripartite motif protein TRIM2', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84160439, end_position_on_the_genomic_accession=84305832, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  80891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80891, CurrentID=0, Status='live', Symbol='Fcrls', Aliases='2810439C17Rik, 9330158F12, FcRH2sc, Fcrh2, IFGP2, IgSR, MMAN-g, Msr2, moFcRH2sc', description='Fc receptor-like S, scavenger receptor', other_designations='immunoglobulin scavenger receptor|macrophage scavenger receptor 2', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87250758, end_position_on_the_genomic_accession=87263765, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  80892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80892, CurrentID=0, Status='live', Symbol='Zfhx4', Aliases='A930021B15, C130041O22Rik, Zfh-4, Zfh4', description='zinc finger homeodomain 4', other_designations='zinc finger homeodomain protein 4', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=5217420, end_position_on_the_genomic_accession=5415861, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  80893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80893, CurrentID=0, Status='live', Symbol='Tmprss5', Aliases='spinesin', description='transmembrane protease, serine 5 (spinesin)', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=49080439, end_position_on_the_genomic_accession=49117591, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  80898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80898, CurrentID=0, Status='live', Symbol='Erap1', Aliases='Arts1, ERAAP, PILSA, PILSAP', description='endoplasmic reticulum aminopeptidase 1', other_designations='A-LAP|ARTS-1|PILS-AP|VEGF-induced aminopeptidase|adipocyte-derived leucine aminopeptidase|aminopeptidase PILS|puromycin-insensitive leucyl-specific aminopeptidase|type 1 tumor necrosis factor receptor shedding aminopeptidase regulator', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74639624, end_position_on_the_genomic_accession=74691931, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  80901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80901, CurrentID=0, Status='live', Symbol='Cxcr6', Aliases='BB217514, BONZO, STRL33', description='chemokine (C-X-C motif) receptor 6', other_designations='CXC-R6|CXCR-6', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123806467, end_position_on_the_genomic_accession=123811756, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  80902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80902, CurrentID=0, Status='live', Symbol='Zfp202', Aliases='C130037E22Rik, Znf202', description='zinc finger protein 202', other_designations='zinc finger 202', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40191169, end_position_on_the_genomic_accession=40213604, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  80903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80903, CurrentID=0, Status='live', Symbol='Fgf16', Aliases='', description='fibroblast growth factor 16', other_designations='FGF-16', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=105764477, end_position_on_the_genomic_accession=105776532, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  80904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80904, CurrentID=0, Status='live', Symbol='Dtx3', Aliases='Deltex3, mg310', description='deltex 3, E3 ubiquitin ligase', other_designations='deltex 3 homolog|mDTX3|protein deltex-3', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127190378, end_position_on_the_genomic_accession=127196412, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  80905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80905, CurrentID=0, Status='live', Symbol='Polh', Aliases='RAD30A, XPV', description='polymerase (DNA directed), eta (RAD 30 related)', other_designations='DNA-directed DNA polymerase eta|RAD30 homolog A|xeroderma pigmentosum variant type protein homolog', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46171355, end_position_on_the_genomic_accession=46202683, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  211978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211978, CurrentID=0, Status='live', Symbol='Zfyve26', Aliases='4930465A13, 9330197E15Rik, A630028O16Rik, Gm893, mKIAA0321', description='zinc finger, FYVE domain containing 26', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79232346, end_position_on_the_genomic_accession=79296310, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  80907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80907, CurrentID=0, Status='live', Symbol='Lactb', Aliases='LACT-1, Lact1, Mrpl56', description='lactamase, beta', other_designations='mitochondrial ribosomal protein L56|serine beta lactamase-like protein', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66955393, end_position_on_the_genomic_accession=66975484, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  80908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80908, CurrentID=0, Status='live', Symbol='Abo', Aliases='NAGAT', description='ABO blood group (transferase A, alpha 1-3-N-acetylgalactosaminyltransferase, transferase B, alpha 1-3-galactosyltransferase)', other_designations='cis-AB transferase|fucosylglycoprotein 3-alpha-galactosyltransferase|fucosylglycoprotein alpha-N-acetylgalactosaminyltransferase|glycoprotein-fucosylgalactoside alpha-N-acetylgalactosaminyltransferase|glycoprotein-fucosylgalactoside alpha-galactosyltransferase|histo-blood group A transferase|histo-blood group B transferase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26842496, end_position_on_the_genomic_accession=26864995, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  80909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80909, CurrentID=0, Status='live', Symbol='Gatsl2', Aliases='7530428J21Rik, AW742413, Gats, Gatsl1', description='GATS protein-like 2', other_designations='GATS-like protein 1|opposite strand transcription unit to Stag3', map_location='5 G1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134099748, end_position_on_the_genomic_accession=134141782, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  80910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80910, CurrentID=0, Status='live', Symbol='Gpr84', Aliases='EX33', description='G protein-coupled receptor 84', other_designations='orphan G protein-coupled receptor 84', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103308233, end_position_on_the_genomic_accession=103311242, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  80911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80911, CurrentID=0, Status='live', Symbol='Acox3', Aliases='BI685180, EST-s59, PCOX', description='acyl-Coenzyme A oxidase 3, pristanoyl', other_designations='BRCACox|branched-chain acyl-CoA oxidase|pristanoyl-CoA oxidase', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35582967, end_position_on_the_genomic_accession=35614210, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  80912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80912, CurrentID=0, Status='live', Symbol='Pum1', Aliases='AA517475, Pumm, mKIAA0099', description='pumilio RNA-binding family member 1', other_designations='pumilio 1', map_location='4 D2.2-D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130663297, end_position_on_the_genomic_accession=130781564, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  80913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80913, CurrentID=0, Status='live', Symbol='Pum2', Aliases='5730503J23Rik, Pumm2', description='pumilio RNA-binding family member 2', other_designations='pumilio 2', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8674134, end_position_on_the_genomic_accession=8752583, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  211986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211986, CurrentID=0, Status='live', Symbol='Tmem18', Aliases='BC024093', description='transmembrane protein 18', other_designations='', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=30584443, end_position_on_the_genomic_accession=30591219, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  80915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80915, CurrentID=0, Status='live', Symbol='Dusp12', Aliases='1190004O14Rik, AA027408, AW049275, ESTM36, LMW-DSP4, T-DSP4, VH1, mVH1', description='dual specificity phosphatase 12', other_designations='dual specificity phosphatase T-DSP4|dual specificity phosphatase VH1', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170874187, end_position_on_the_genomic_accession=170885711, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  474145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=474145, CurrentID=0, Status='live', Symbol='Clec4a4', Aliases='Dcir2', description='C-type lectin domain family 4, member a4', other_designations='dendritic cell inhibitory receptor 2', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122990367, end_position_on_the_genomic_accession=123024421, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  474156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=474156, CurrentID=0, Status='live', Symbol='Zbtb9', Aliases='3930402F13Rik', description='zinc finger and BTB domain containing 9', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26973179, end_position_on_the_genomic_accession=26976203, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  212032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212032, CurrentID=0, Status='live', Symbol='Hk3', Aliases='HK III, HK-III', description='hexokinase 3', other_designations='hexokinase type III', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55005985, end_position_on_the_genomic_accession=55021472, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  80976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80976, CurrentID=0, Status='live', Symbol='Syt13', Aliases='5730409J20Rik, AI549909, mKIAA1427', description='synaptotagmin XIII', other_designations='synaptotagmin 13|sytXIII', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92915101, end_position_on_the_genomic_accession=92956051, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  80978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80978, CurrentID=0, Status='live', Symbol='Mrgprh', Aliases='Gpr90, Mgrb, MrgH', description='MAS-related GPR, member H', other_designations='G protein-coupled receptor 90|G-protein coupled receptor 90|G-protein coupled receptor GPR90|Mas-related G protein-coupled receptor b', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12876034, end_position_on_the_genomic_accession=12877842, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  80979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80979, CurrentID=0, Status='live', Symbol='Slc26a5', Aliases='Pres, prestin', description='solute carrier family 26, member 5', other_designations='outer hair cell motor protein|prestin (motor protein)', map_location='5 9.97 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21809001, end_position_on_the_genomic_accession=21865604, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  80981 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80981, CurrentID=0, Status='live', Symbol='Arl4d', Aliases='1110036H21Rik, ARL6, AW456149, Arf4l, Arfl4, Arl5', description='ADP-ribosylation factor-like 4D', other_designations='ADP-ribosylation factor-like 5|ADP-ribosylation factor-like protein 4L|ADP-ribosylation factor-like protein 5', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101665541, end_position_on_the_genomic_accession=101667832, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  80982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80982, CurrentID=0, Status='live', Symbol='Cemip', Aliases='12H19.01.T7, 6330404C01Rik, 9930013L23Rik, AY007814, Kiaa1199', description='cell migration inducing protein, hyaluronan binding', other_designations='protein 12H19.01.T7', map_location='7 48.35 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=83932857, end_position_on_the_genomic_accession=84086692, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  80985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80985, CurrentID=0, Status='live', Symbol='Trim44', Aliases='Dipb, Mc7', description='tripartite motif-containing 44', other_designations='DIPB|brain cDNA 7|protein Mc7', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=102300119, end_position_on_the_genomic_accession=102400900, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  80986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80986, CurrentID=0, Status='live', Symbol='Ckap2', Aliases='AI461656, AW228814, LB1', description='cytoskeleton associated protein 2', other_designations='ckap2', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22168149, end_position_on_the_genomic_accession=22185819, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  80987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80987, CurrentID=0, Status='live', Symbol='Nckipsd', Aliases='AF3P21, DIP1, ORF1, SPIN90, WASLBP, WISH, Wasbp', description='NCK interacting protein with SH3 domain', other_designations='54 kDa VacA-interacting protein|90 kDa N-WASP-interacting protein|90 kDa SH3 protein interacting with Nck|N-WASP-binding protein|SH3 adapter protein SPIN90|VIP54|WASP-interacting SH3-domain protein|Wiskott-Aldrich syndrome homolog binding protein|mDia interacting protein-1|wiskott-Aldrich syndrome protein-binding protein WISH', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108808347, end_position_on_the_genomic_accession=108818839, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  212070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212070, CurrentID=0, Status='live', Symbol='Clrn3', Aliases='AI649392, Tmem12', description='clarin 3', other_designations='transmembrane protein 12', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=135511456, end_position_on_the_genomic_accession=135528654, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81000, CurrentID=0, Status='live', Symbol='Rad54l2', Aliases='Arip4, D130058C05, G630026H09Rik, Srisnf2l', description='RAD54 like 2 (S. cerevisiae)', other_designations='HCMV-interacting protein|RAD54-like protein 2|androgen receptor-interacting protein 4|steroid receptor-interacting SNF2 domain protein', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106688080, end_position_on_the_genomic_accession=106789213, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  212073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212073, CurrentID=0, Status='live', Symbol='Syne3', Aliases='nesprin-3', description='spectrin repeat containing, nuclear envelope family member 3', other_designations='nuclear envelope spectrin repeat protein 3', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104929830, end_position_on_the_genomic_accession=105010098, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  81003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81003, CurrentID=0, Status='live', Symbol='Trim23', Aliases='6330516O20Rik, AI450195, Arfd1', description='tripartite motif-containing 23', other_designations='ADP-ribosylation factor domain protein 1, 64kD|ADP-ribosylation factor domain-containing protein 1|GTP-binding protein ARD-1|tripartite motif protein 23|tripartite motif protein TRIM23|tripartite motif-containing protein 23', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104178774, end_position_on_the_genomic_accession=104203421, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  81004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81004, CurrentID=0, Status='live', Symbol='Tbl1xr1', Aliases='8030499H02Rik, A630076E03Rik, AW539987, C21, C230089I12Rik, DC42, Ira1, TBLR1', description='transducin (beta)-like 1X-linked receptor 1', other_designations='F-box-like/WD repeat protein TBL1XR1|TBL1-related protein 1|nuclear receptor corepressor/HDAC3 complex subunit TBLR1|transducin beta-like 1X-related protein 1', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=22076593, end_position_on_the_genomic_accession=22216594, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  81006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81006, CurrentID=0, Status='live', Symbol='Gpr63', Aliases='PSP24beta, Psp24-2', description='G protein-coupled receptor 63', other_designations='G protein-coupled receptor PSP24-2|PSP24-beta', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=24966541, end_position_on_the_genomic_accession=25009233, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  81010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81010, CurrentID=0, Status='live', Symbol='Vmn1r172', Aliases='V1rd9, V3R9', description='vomeronasal 1 receptor 172', other_designations='pheromone receptor V3R9|vomeronasal 1 receptor Vmn1r172|vomeronasal 1 receptor, D9', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23658316, end_position_on_the_genomic_accession=23660663, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81011, CurrentID=0, Status='live', Symbol='Vmn1r148', Aliases='V1rd14, V3R8', description='vomeronasal 1 receptor 148', other_designations='pheromone receptor V3R8|vomeronasal 1 receptor Vmn1r148', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=22412231, end_position_on_the_genomic_accession=22414669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81012, CurrentID=0, Status='live', Symbol='Vmn1r171', Aliases='V1rd7, V3R7', description='vomeronasal 1 receptor 171', other_designations='pheromone receptor V3R7|vomeronasal 1 receptor Vmn1r171|vomeronasal 1 receptor, D7|vomeronasal type 1 receptor D7', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23631988, end_position_on_the_genomic_accession=23633568, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81013, CurrentID=0, Status='live', Symbol='Vmn1r65', Aliases='V1rd6, V3R6', description='vomeronasal 1 receptor 65', other_designations='pheromone receptor V3R6|vomeronasal 1 receptor Vmn1r65|vomeronasal type 1 receptor D6', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6007750, end_position_on_the_genomic_accession=6011010, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81014, CurrentID=0, Status='live', Symbol='Vmn1r58', Aliases='V1rd4, V3R4', description='vomeronasal 1 receptor 58', other_designations='pheromone receptor V3R4|vomeronasal 1 receptor Vmn1r58|vomeronasal 1 receptor, D4', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5408887, end_position_on_the_genomic_accession=5413145, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  81015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81015, CurrentID=0, Status='live', Symbol='Vmn1r56', Aliases='V1rd3, V3R3', description='vomeronasal 1 receptor 56', other_designations='pheromone receptor V3R3|vomeronasal 1 receptor Vmn1r56|vomeronasal 1 receptor, D3', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5194916, end_position_on_the_genomic_accession=5196747, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  81016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81016, CurrentID=0, Status='live', Symbol='Vmn1r62', Aliases='V1rd2, V1rd8, V3R2', description='vomeronasal 1 receptor 62', other_designations='pheromone receptor V3R2|vomernasal 1 receptor Vmn1r62|vomeronasal 1 receptor, D8', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5674604, end_position_on_the_genomic_accession=5676599, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81017, CurrentID=0, Status='live', Symbol='Vmn1r63', Aliases='V1R1, V1rd1, V3R1', description='vomeronasal 1 receptor 63', other_designations='pheromone receptor V3R1|vomernasal 1 receptor Vmn1r63|vomeronasal 1 receptor, D1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5802345, end_position_on_the_genomic_accession=5805445, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  212090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212090, CurrentID=0, Status='live', Symbol='Tmem60', Aliases='BC027828', description='transmembrane protein 60', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=20882453, end_position_on_the_genomic_accession=20886870, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  212108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212108, CurrentID=0, Status='live', Symbol='Rln3', Aliases='M3, insl7', description='relaxin 3', other_designations='insulin-like 7|insulin-like peptide 7|insulin-like peptide INSL7|prorelaxin M3', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84043067, end_position_on_the_genomic_accession=84044979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  212111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212111, CurrentID=0, Status='live', Symbol='Inpp5a', Aliases='', description='inositol polyphosphate-5-phosphatase A', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139389109, end_position_on_the_genomic_accession=139579652, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  212114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212114, CurrentID=0, Status='live', Symbol='Nhlrc3', Aliases='4833441N19, 8030451K01Rik, mKIAA4083', description='NHL repeat containing 3', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=53451996, end_position_on_the_genomic_accession=53463258, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  212123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212123, CurrentID=0, Status='live', Symbol='Dcaf15', Aliases='6720484B16', description='DDB1 and CUL4 associated factor 15', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84097072, end_position_on_the_genomic_accession=84104792, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  212127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212127, CurrentID=0, Status='live', Symbol='Proser1', Aliases='2810046L04Rik, 9330161F11', description='proline and serine rich 1', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=53463817, end_position_on_the_genomic_accession=53481755, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  212139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212139, CurrentID=0, Status='live', Symbol='Cc2d1a', Aliases='BC016188, Freud-1, Tape', description='coiled-coil and C2 domain containing 1A', other_designations="5-HT1A receptor|FRE under dual repression-binding protein 1|five prime repressor element under dual repression-binding protein 1|five repressor element under dual repression-binding protein 1|five' repressor element under dual repression binding protein 1|putative NFkB activating protein", map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84132828, end_position_on_the_genomic_accession=84147931, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  212153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212153, CurrentID=0, Status='live', Symbol='Ccdc191', Aliases='2610015P09Rik, 8430422M09Rik, AU040712', description='coiled-coil domain containing 191', other_designations='', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=43889752, end_position_on_the_genomic_accession=43964314, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  212163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212163, CurrentID=0, Status='live', Symbol='8030462N17Rik', Aliases='Akd2, Ark2N', description='RIKEN cDNA 8030462N17 gene', other_designations='arkadia 2', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77631488, end_position_on_the_genomic_accession=77714489, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  212167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212167, CurrentID=0, Status='live', Symbol='Gsap', Aliases='A530088I07Rik, Pion', description='gamma-secretase activating protein', other_designations='pigeon homolog|protein pigeon homolog', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=21176333, end_position_on_the_genomic_accession=21291719, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  212168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212168, CurrentID=0, Status='live', Symbol='Zswim4', Aliases='D330002I24, E130119J17Rik', description='zinc finger SWIM-type containing 4', other_designations='zinc finger, SWIM domain containing 4', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84210942, end_position_on_the_genomic_accession=84237042, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  474332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=474332, CurrentID=0, Status='live', Symbol='Dnm3os', Aliases='6030416H16Rik, Hag2, Npn1, P/L01', description='dynamin 3, opposite strand', other_designations='neoplastic progression 1', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162217623, end_position_on_the_genomic_accession=162225550, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  212190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212190, CurrentID=0, Status='live', Symbol='Ubxn10', Aliases='5730509E04Rik, A830047D02, Ubxd3', description='UBX domain protein 10', other_designations='UBX domain containing 3|UBX domain-containing protein 3', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138709837, end_position_on_the_genomic_accession=138738438, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  212198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212198, CurrentID=0, Status='live', Symbol='Wdr25', Aliases='B930090D16Rik', description='WD repeat domain 25', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108894228, end_position_on_the_genomic_accession=109028452, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  13525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13525, CurrentID=0, Status='live', Symbol='Adam26a', Aliases='Adam26, Dtgn4', description='a disintegrin and metallopeptidase domain 26A (testase 3)', other_designations='ADAM 26A|a disintegrin and metalloprotease domain 26|disintegrin 4|testase 3|testase-3', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=43568273, end_position_on_the_genomic_accession=43576707, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  212276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212276, CurrentID=0, Status='live', Symbol='Zfp748', Aliases='KRAB-O, Rslcan20/21, Zfp208, mszf54', description='zinc finger protein 748', other_designations='KRAB zinc finger protein', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67538641, end_position_on_the_genomic_accession=67553152, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  212281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212281, CurrentID=0, Status='live', Symbol='Zfp729a', Aliases='A530054K11Rik, Rslcan3', description='zinc finger protein 729a', other_designations='regulator of sex-limitation candidate 3', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67616309, end_position_on_the_genomic_accession=67638062, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  212285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212285, CurrentID=0, Status='live', Symbol='Arap2', Aliases='Centd1, Gm148, mKIAA0580', description='ArfGAP with RhoGAP domain, ankyrin repeat and PH domain 2', other_designations='PARX protein|centaurin, delta 1|centaurin-delta-1|cnt-d1', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=62602445, end_position_on_the_genomic_accession=62766177, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  105463122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105463122, CurrentID=0, Status='live', Symbol='Tip3', Aliases='', description='transcribed intergenic polycomb 3', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  212307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212307, CurrentID=0, Status='live', Symbol='Mapre2', Aliases='AI314113, C820009F03Rik, D18Abb1e, EB1, EB2, RP1', description='microtubule-associated protein, RP/EB family, member 2', other_designations='APC-binding protein EB1|APC-binding protein EB2|T-cell activation protein, EB1 family, member 2|end-binding protein 2', map_location='18 12.08 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=23752333, end_position_on_the_genomic_accession=23893861, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  105463124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105463124, CurrentID=0, Status='live', Symbol='Borg', Aliases='', description='BMP/OP-responsive gene', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  212326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212326, CurrentID=0, Status='live', Symbol='Fam149a', Aliases='AW213256, mKIAA4166', description='family with sequence similarity 149, member A', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45336715, end_position_on_the_genomic_accession=45382291, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  212377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212377, CurrentID=0, Status='live', Symbol='Mms22l', Aliases='F730047E07Rik, Gm134', description='MMS22-like, DNA repair protein', other_designations='', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=24496444, end_position_on_the_genomic_accession=24602950, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  212390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212390, CurrentID=0, Status='live', Symbol='Klhl32', Aliases='6430524H05Rik, D4Ertd389e, Gm1356, mKIAA1900', description='kelch-like 32', other_designations='BTB and kelch domain containing 5', map_location='4 10.42 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=24614352, end_position_on_the_genomic_accession=24851110, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  212391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212391, CurrentID=0, Status='live', Symbol='Lcor', Aliases='3110023F06Rik, A630025C20Rik, Mlr2, mKIAA1795', description='ligand dependent nuclear receptor corepressor', other_designations='Mblk1-related protein-2|mblk1-related protein 2|transcription factor MLR2', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41482583, end_position_on_the_genomic_accession=41596161, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  212392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212392, CurrentID=0, Status='live', Symbol='Ccdc110', Aliases='Gm172', description='coiled-coil domain containing 110', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45934649, end_position_on_the_genomic_accession=45945230, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  212398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212398, CurrentID=0, Status='live', Symbol='Frat2', Aliases='', description='frequently rearranged in advanced T cell lymphomas 2', other_designations='FRAT-2|frequently rearranged in advanced T-cell lymphomas 2', map_location='19 35.33 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41845972, end_position_on_the_genomic_accession=41848132, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  13557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13557, CurrentID=0, Status='live', Symbol='E2f3', Aliases='E2F3ba, E2f3', description='E2F transcription factor 3', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=29906575, end_position_on_the_genomic_accession=29987143, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  212439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212439, CurrentID=0, Status='live', Symbol='AA986860', Aliases='E030025M07, Sarg', description='expressed sequence AA986860', other_designations='specifically androgen-regulated protein', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130731976, end_position_on_the_genomic_accession=130744622, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  212442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212442, CurrentID=0, Status='live', Symbol='Lactb2', Aliases='Cgi-83, E430032H21Rik', description='lactamase, beta 2', other_designations='', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=13625893, end_position_on_the_genomic_accession=13660605, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  212448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212448, CurrentID=0, Status='live', Symbol='9330159F19Rik', Aliases='', description='RIKEN cDNA 9330159F19 gene', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=29199192, end_position_on_the_genomic_accession=29230779, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  212483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212483, CurrentID=0, Status='live', Symbol='Fam193b', Aliases='IRIZIO, Kiaa1931', description='family with sequence similarity 193, member B', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55539316, end_position_on_the_genomic_accession=55571120, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  212503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212503, CurrentID=0, Status='live', Symbol='Paox', Aliases='2410012F02Rik, AI118225, Pao', description='polyamine oxidase (exo-N4-amino)', other_designations='exo-N4-amino|peroxisomal N1-acetyl-spermine/spermidine oxidase', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140115534, end_position_on_the_genomic_accession=140134368, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  212508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212508, CurrentID=0, Status='live', Symbol='Mtg1', Aliases='Gm169, Gtpbp7', description='mitochondrial GTPase 1 homolog (S. cerevisiae)', other_designations='GTP binding protein 7 (putative)|GTP-binding protein 7', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140137564, end_position_on_the_genomic_accession=140150786, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  212514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212514, CurrentID=0, Status='live', Symbol='Spice1', Aliases='Ccdc52, D16Ertd480e', description='spindle and centriole associated protein 1', other_designations='coiled-coil domain containing 52|coiled-coil domain-containing protein 52', map_location='16 28.67 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44347395, end_position_on_the_genomic_accession=44388492, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  212516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212516, CurrentID=0, Status='live', Symbol='Efcab12', Aliases='BB124106', description='EF-hand calcium binding domain 12', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115810729, end_position_on_the_genomic_accession=115838791, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  212517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212517, CurrentID=0, Status='live', Symbol='Cfap44', Aliases='6330444M21Rik, D16Ertd642e, Gm1750, Wdr52', description='cilia and flagella associated protein 44', other_designations='WD repeat domain 52', map_location='16 28.72 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44394742, end_position_on_the_genomic_accession=44482694, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  212518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212518, CurrentID=0, Status='live', Symbol='Sprn', Aliases='C630041J07', description='shadow of prion protein', other_designations='Shadoo|protein shadoo', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140150628, end_position_on_the_genomic_accession=140154659, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  212528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212528, CurrentID=0, Status='live', Symbol='Trmt1', Aliases='6720406L13Rik, 6720477L24, D8Ertd812e', description='tRNA methyltransferase 1', other_designations='TRM1 tRNA methyltransferase 1 homolog|n(2),N(2)-dimethylguanosine tRNA methyltransferase|tRNA 2,2-dimethylguanosine-26 methyltransferase|tRNA(guanine-26,N(2)-N(2)) methyltransferase|tRNA(m(2,2)G26)dimethyltransferase', map_location='8 41.02 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84688725, end_position_on_the_genomic_accession=84699808, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  212531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212531, CurrentID=0, Status='live', Symbol='Sh3bgrl2', Aliases='A930014C21Rik, AA408038, B930089I21', description='SH3 domain binding glutamic acid-rich protein like 2', other_designations='', map_location='9 E2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=83548338, end_position_on_the_genomic_accession=83600292, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  212539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212539, CurrentID=0, Status='live', Symbol='Gm266', Aliases='', description='predicted gene 266', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111484609, end_position_on_the_genomic_accession=111485823, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  212541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212541, CurrentID=0, Status='live', Symbol='Rho', Aliases='Noerg1, Opn2, Ops, RP4', description='rhodopsin', other_designations='L opsin|LWS opsin|Long Wavelength Sensitive opsin|Red Opsin|Rod Opsin|opsin 2', map_location='6 53.72 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115926754, end_position_on_the_genomic_accession=115940036, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  212547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212547, CurrentID=0, Status='live', Symbol='Nepro', Aliases='', description='nucleolus and neural progenitor protein', other_designations='neural progenitor protein', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44724301, end_position_on_the_genomic_accession=44737284, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  212555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212555, CurrentID=0, Status='live', Symbol='Pqlc2', Aliases='BC019216', description='PQ loop repeat containing 2', other_designations='PQ-loop repeat-containing protein 2', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139294030, end_position_on_the_genomic_accession=139310711, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  81489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81489, CurrentID=0, Status='live', Symbol='Dnajb1', Aliases='0610007I11Rik, DjB1, HSPF1, Hdj1, Hsp40', description='DnaJ heat shock protein family (Hsp40) member B1', other_designations='DnaJ (Hsp40) homolog, subfamily B, member 1|heat shock 40 kDa protein 1', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83608175, end_position_on_the_genomic_accession=83611903, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  212569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212569, CurrentID=0, Status='live', Symbol='Zfp273', Aliases='6820416H06Rik, Rslcan1', description='zinc finger protein 273', other_designations='regulator of sex-limitation candidate 1', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67813772, end_position_on_the_genomic_accession=67827668, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  81500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81500, CurrentID=0, Status='live', Symbol='Sil1', Aliases='1810057E01Rik, AI042831, wz', description='endoplasmic reticulum chaperone SIL1 homolog (S. cerevisiae)', other_designations='', map_location='18 B1-B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35266396, end_position_on_the_genomic_accession=35499052, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  81535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81535, CurrentID=0, Status='live', Symbol='Sgpp1', Aliases='AI463453, SPP, SPP1, Spph1', description='sphingosine-1-phosphate phosphatase 1', other_designations='SPPase1|sphingosine-1-phosphatase 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=75714248, end_position_on_the_genomic_accession=75735729, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  212627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212627, CurrentID=0, Status='live', Symbol='Prpsap2', Aliases='A230054F23Rik', description='phosphoribosyl pyrophosphate synthetase-associated protein 2', other_designations='41 kDa phosphoribosypyrophosphate synthetase-associated protein|PAP41|PRPP synthase-associated protein 2|PRPP synthetase-associated protein 2', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61729650, end_position_on_the_genomic_accession=61762088, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  212632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212632, CurrentID=0, Status='live', Symbol='Iffo2', Aliases='C79267', description='intermediate filament family orphan 2', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139530548, end_position_on_the_genomic_accession=139620382, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  212647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212647, CurrentID=0, Status='live', Symbol='Aldh4a1', Aliases='A930035F14Rik, Ahd-1, Ahd1, Aldh4, Aldh5a1, E330022C09, P5cd, P5cdh, P5cdhl, P5cdhs, Ssdh1', description='aldehyde dehydrogenase 4 family, member A1', other_designations='L-glutamate gamma-semialdehyde dehydrogenase|P5C dehydrogenase|aldehyde dehydrogenase family 4 member A1', map_location='4 70.79 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139622871, end_position_on_the_genomic_accession=139649691, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  212670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212670, CurrentID=0, Status='live', Symbol='Catsper2', Aliases='', description='cation channel, sperm associated 2', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121393024, end_position_on_the_genomic_accession=121416224, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  81600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81600, CurrentID=0, Status='live', Symbol='Chia1', Aliases='2200003E03Rik, AMCase, Chia, YNL', description='chitinase, acidic 1', other_designations='acidic chitinase', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106113382, end_position_on_the_genomic_accession=106132118, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  81601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81601, CurrentID=0, Status='live', Symbol='Kat5', Aliases='AI839539, CPLA2, Htatip, Htatip1, PLIP, Tip55, Tip60', description='K(lysine) acetyltransferase 5', other_designations='60 kDa Tat-interactive protein|HIV-1 tat interactive protein 1, 60 kDa homolog|HIV-1 tat interactive protein 1, homolog|Tat-interactive protein-60|histone acetyltransferase HTATIP|lysine acetyltransferase 5', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5603014, end_position_on_the_genomic_accession=5610232, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  212679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212679, CurrentID=0, Status='live', Symbol='Mars2', Aliases='C730026E21Rik, MetRS', description='methionine-tRNA synthetase 2 (mitochondrial)', other_designations='methionine--tRNA ligase 2|methionyl-tRNA synthetase 2|methionyl-tRNA synthetase, mitochondrial|mitochondrial methionine--tRNA ligase|mitochondrial methionyl-tRNA synthetase|mtMetRS', map_location='1 C1.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55237177, end_position_on_the_genomic_accession=55240058, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  81630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81630, CurrentID=0, Status='live', Symbol='Zbtb22', Aliases='1110008J20Rik, AI265210, AI415166, Bing1, Zfp297', description='zinc finger and BTB domain containing 22', other_designations='zinc finger protein 297', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33915895, end_position_on_the_genomic_accession=33919325, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  212706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212706, CurrentID=0, Status='live', Symbol='N4bp3', Aliases='C330016O10Rik', description='NEDD4 binding protein 3', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51643071, end_position_on_the_genomic_accession=51651073, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  212712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212712, CurrentID=0, Status='live', Symbol='Satb2', Aliases='mKIAA1034', description='special AT-rich sequence binding protein 2', other_designations='KIAA1034-like DNA binding protein|special AT-rich sequence-binding protein 2|two cut domains-containing homeodomain protein', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=56793981, end_position_on_the_genomic_accession=56978681, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  212728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212728, CurrentID=0, Status='live', Symbol='Gm17296', Aliases='Gm179, Tarbp1', description='predicted gene, 17296', other_designations='TAR (HIV-1) RNA binding protein 1', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=126425314, end_position_on_the_genomic_accession=126475113, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  212733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212733, CurrentID=0, Status='live', Symbol='Ccdc64b', Aliases='BICDR-2', description='coiled-coil domain containing 64B', other_designations='BICD-related protein 2|coiled-coil domain-containing protein 64B', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23660523, end_position_on_the_genomic_accession=23668818, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  668976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=668976, CurrentID=0, Status='live', Symbol='Ssxb6', Aliases='', description='synovial sarcoma, X member B, breakpoint 6', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8542179, end_position_on_the_genomic_accession=8548362, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  212772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212772, CurrentID=0, Status='live', Symbol='Arl14ep', Aliases='2700007P21Rik, 4930448O08Rik, ARF7EP', description='ADP-ribosylation factor-like 14 effector protein', other_designations='ARF7 effector protein', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=106962527, end_position_on_the_genomic_accession=106974447, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  81701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81701, CurrentID=0, Status='live', Symbol='Egfl8', Aliases='Ng3', description='EGF-like domain 8', other_designations='EGF-like protein 8', map_location='17 18.19 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34613200, end_position_on_the_genomic_accession=34616572, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  81702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81702, CurrentID=0, Status='live', Symbol='Ankrd17', Aliases='4933425K22Rik, A130069E23Rik, A930008M01, AA407558, AA516750, AU040470, Foe, Gtar, Mask, mKIAA0697', description='ankyrin repeat domain 17', other_designations='ankyrin repeat domain containing protein FOE|gene trap ankyrin repeat protein', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90227166, end_position_on_the_genomic_accession=90367845, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  81703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81703, CurrentID=0, Status='live', Symbol='Jdp2', Aliases='Jundm2, Jundp2, TIF', description='Jun dimerization protein 2', other_designations='c-Jun dimerization protein 2', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85597229, end_position_on_the_genomic_accession=85639878, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  212862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212862, CurrentID=0, Status='live', Symbol='Chpt1', Aliases='', description='choline phosphotransferase 1', other_designations='diabetic nephropathy-associated gene transcript 1|diacylglycerol cholinephosphotransferase 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88472813, end_position_on_the_genomic_accession=88503970, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  81799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81799, CurrentID=0, Status='live', Symbol='C1qtnf3', Aliases='2310005P21Rik, CORS-26, CORS26, CTRP3, Corcs, Cors', description='C1q and tumor necrosis factor related protein 3', other_designations='cartducin|cartonectin|collagenous repeat-containing sequence 26 kDa protein|secretory protein CORS26', map_location='15 A2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=10952332, end_position_on_the_genomic_accession=10980162, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  212880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212880, CurrentID=0, Status='live', Symbol='Ddx46', Aliases='2200005K02Rik, 8430438J23Rik, AI325430, AI957095, mKIAA0801', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 46', other_designations='DEAD box polypeptide 46|DEAD box protein 46', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55634969, end_position_on_the_genomic_accession=55681267, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  212892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212892, CurrentID=0, Status='live', Symbol='Rsph4a', Aliases='A230081C05, Rshl3', description='radial spoke head 4 homolog A (Chlamydomonas)', other_designations='radial spoke head-like protein 3|radial spokehead-like 3', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=33905111, end_position_on_the_genomic_accession=33916021, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  212898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212898, CurrentID=0, Status='live', Symbol='Dse', Aliases='6030499O08, AI480506, B130024B19Rik, DS-epi1, Sart2', description='dermatan sulfate epimerase', other_designations='DS epimerase|SART-2|chondroitin-glucuronate 5-epimerase|squamous cell carcinoma antigen recognized by T cells 2|squamous cell carcinoma antigen recognized by T-cells 2', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34151386, end_position_on_the_genomic_accession=34208275, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  13639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13639, CurrentID=0, Status='live', Symbol='Efna4', Aliases='EFL-4, Epl4, LERK-4', description='ephrin A4', other_designations='EPH-related receptor tyrosine kinase ligand 4', map_location='3 39.07 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89333393, end_position_on_the_genomic_accession=89338028, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  81840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81840, CurrentID=0, Status='live', Symbol='Sorcs2', Aliases='N28137, mKIAA1329', description='sortilin-related VPS10 domain containing receptor 2', other_designations='VPS10 domain receptor protein SORCS 2', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36017180, end_position_on_the_genomic_accession=36398167, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  81845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81845, CurrentID=0, Status='live', Symbol='Gpank1', Aliases='Bat-4, Bat4, D17H6S54E, G5', description='G patch domain and ankyrin repeats 1', other_designations='G patch domain and ankyrin repeats-containing protein 1|HLA-B associated transcript 4|HLA-B-associated transcript 4|protein BAT4', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35121473, end_position_on_the_genomic_accession=35124815, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  212919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212919, CurrentID=0, Status='live', Symbol='Kctd7', Aliases='4932409E18, 9430010P06Rik', description='potassium channel tetramerisation domain containing 7', other_designations='', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130143749, end_position_on_the_genomic_accession=130155808, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  212933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212933, CurrentID=0, Status='live', Symbol='Pm20d1', Aliases='4732466D17Rik, AI098026', description='peptidase M20 domain containing 1', other_designations='peptidase M20 domain-containing protein 1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131797385, end_position_on_the_genomic_accession=131821419, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  212937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212937, CurrentID=0, Status='live', Symbol='Tifab', Aliases='', description='TRAF-interacting protein with forkhead-associated domain, family member B', other_designations='TIFA like protein|TIFA related protein TIFAB|TIFA-like protein|tifab', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56173702, end_position_on_the_genomic_accession=56179018, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  212943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212943, CurrentID=0, Status='live', Symbol='Fam46a', Aliases='D930050G01Rik', description='family with sequence similarity 46, member A', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=85320439, end_position_on_the_genomic_accession=85327150, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  81877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81877, CurrentID=0, Status='live', Symbol='Tnxb', Aliases='Tn-mhc, Tnx', description='tenascin XB', other_designations='', map_location='17 18.24 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34659788, end_position_on_the_genomic_accession=34795460, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  81879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81879, CurrentID=0, Status='live', Symbol='Tfcp2l1', Aliases='1810030F05Rik, 4932442M07Rik, AA575098, Cp2l1, Crtr-1, D930018N21Rik, LBP-9, Tcfcp2l1', description='transcription factor CP2-like 1', other_designations='CP2-related transcriptional repressor 1|Tcfcp2-related transcriptional repressor 1|transcription repressor CRTR-1', map_location='1 E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=118627923, end_position_on_the_genomic_accession=118685168, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  81896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81896, CurrentID=0, Status='live', Symbol='Ift122', Aliases='C86139, Wdr10, sopb', description='intraflagellar transport 122', other_designations='WD repeat domain 10|WD repeat-containing protein 10|intraflagellar transport 122 homolog', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115853460, end_position_on_the_genomic_accession=115926699, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  81897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81897, CurrentID=0, Status='live', Symbol='Tlr9', Aliases='', description='toll-like receptor 9', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106222598, end_position_on_the_genomic_accession=106226876, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81898, CurrentID=0, Status='live', Symbol='Sf3b1', Aliases='155kDa, 2810001M05Rik, AA409119, Prp10, SAP155, SF3b155, TA-8, Targ4', description='splicing factor 3b, subunit 1', other_designations='SAP 155|pre-mRNA-splicing factor SF3b 155 kDa subunit|spliceosome-associated protein 155|splicing factor 3b, subunit 1, 155 kDa|transforming growth factor alpha regulated gene 4', map_location='1 27.98 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=54985169, end_position_on_the_genomic_accession=55027500, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  212974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212974, CurrentID=0, Status='live', Symbol='Athl1', Aliases='5730511L01Rik, BC023151', description='ATH1, acid trehalase-like 1 (yeast)', other_designations='mFLJ00228 protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140941581, end_position_on_the_genomic_accession=140947658, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  81904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81904, CurrentID=0, Status='live', Symbol='Cacng7', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 7', other_designations='TARP gamma-7|neuronal voltage-gated calcium channel gamma-7 subunit|transmembrane AMPAR regulatory protein gamma-7|voltage dependent calcium channel gamma 7 subunit', map_location='7 1.95 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3332024, end_position_on_the_genomic_accession=3368222, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  81905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81905, CurrentID=0, Status='live', Symbol='Cacng8', Aliases='', description='calcium channel, voltage-dependent, gamma subunit 8', other_designations='TARP gamma-8|neuronal voltage-gated calcium channel gamma-8 subunit|transmembrane AMPAR regulatory protein gamma-8', map_location='7 1.98 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3394117, end_position_on_the_genomic_accession=3415605, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  81906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81906, CurrentID=0, Status='live', Symbol='Cyp4x1', Aliases='A230025G20, CYPIVX1, CYP_a, Cyp4a28-ps', description='cytochrome P450, family 4, subfamily x, polypeptide 1', other_designations='cytochrome P450, 4x1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115108682, end_position_on_the_genomic_accession=115133977, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  81907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81907, CurrentID=0, Status='live', Symbol='Tmem108', Aliases='AI462967, B130017P16Rik, R74726', description='transmembrane protein 108', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103482936, end_position_on_the_genomic_accession=103761872, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  212980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212980, CurrentID=0, Status='live', Symbol='Slc45a3', Aliases='2210413P12Rik, AU023994, AU043764, IPCA-6, PRST, Pcanap6', description='solute carrier family 45, member 3', other_designations='Prostein|prostate cancer-associated protein 6', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131962915, end_position_on_the_genomic_accession=131982972, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  81909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81909, CurrentID=0, Status='live', Symbol='Zfpl1', Aliases='1500015B20Rik', description='zinc finger like protein 1', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6077366, end_position_on_the_genomic_accession=6084939, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  81910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81910, CurrentID=0, Status='live', Symbol='Rrbp1', Aliases='1700087N07Rik, 5730465C04Rik, ES/130, mKIAA1398, mRRp0, mRRp1.8, mRRp10, mRRp15a, mRRp15b, mRRp16.8, mRRp2, mRRp41, mRRp47, mRRp5.4, p180', description='ribosome binding protein 1', other_designations='RRp|mRRp|ribosome receptor protein', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=143947395, end_position_on_the_genomic_accession=144011296, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  212986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212986, CurrentID=0, Status='live', Symbol='Scfd2', Aliases='E430013M20Rik, STXBP1L1', description='Sec1 family domain containing 2', other_designations='neuronal Sec1|syntaxin-binding protein 1-like 1', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=74204816, end_position_on_the_genomic_accession=74531742, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  212989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212989, CurrentID=0, Status='live', Symbol='Best2', Aliases='Vmd2l1', description='bestrophin 2', other_designations='vitelliform macular dystrophy 2-like protein 1', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85007202, end_position_on_the_genomic_accession=85014531, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  212996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212996, CurrentID=0, Status='live', Symbol='Wbscr17', Aliases='AA388321, E330012B09Rik', description='Williams-Beuren syndrome chromosome region 17 homolog (human)', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like 3|UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 3|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase|galNAc-T-like protein 3|polypeptide GalNAc transferase-like protein 3|polypeptide N-acetylgalactosaminyltransferase-like 3|pp-GaNTase-like protein 3|protein-UDP acetylgalactosaminyltransferase-like protein 3|williams-Beuren syndrome chromosomal region 17 protein homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130874951, end_position_on_the_genomic_accession=131307522, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  212998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212998, CurrentID=0, Status='live', Symbol='BC016579', Aliases='', description='cDNA sequence, BC016579', other_designations='', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45626848, end_position_on_the_genomic_accession=45656243, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  212999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212999, CurrentID=0, Status='live', Symbol='Tnpo2', Aliases='1110034O24Rik, AA414969, AI464345, AI852433, IPO3, Knpb2b, Kpnb2b, TRN2', description='transportin 2 (importin 3, karyopherin beta 2b)', other_designations='karyopherin (importin) beta 2b|karyopherin beta-2b', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85036888, end_position_on_the_genomic_accession=85057585, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  213002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213002, CurrentID=0, Status='live', Symbol='Ifitm6', Aliases='A330075D07, Dspa2f, Fgls5, fragilis5', description='interferon induced transmembrane protein 6', other_designations='fragilis 5', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141015812, end_position_on_the_genomic_accession=141016892, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  213006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213006, CurrentID=0, Status='live', Symbol='Mfsd4a', Aliases='A230072B04, A930031D07Rik, AI850289, Mfsd4', description='major facilitator superfamily domain containing 4A', other_designations='major facilitator superfamily domain containing 4', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132022806, end_position_on_the_genomic_accession=132068062, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  213011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213011, CurrentID=0, Status='live', Symbol='Zfp583', Aliases='BC056221, Znf583', description='zinc finger protein 583', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6315144, end_position_on_the_genomic_accession=6331835, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  213012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213012, CurrentID=0, Status='live', Symbol='Abhd10', Aliases='', description='abhydrolase domain containing 10', other_designations='abhydrolase domain-containing protein 10, mitochondrial|alpha/beta hydrolase domain-containing protein 10, mitochondrial', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45729720, end_position_on_the_genomic_accession=45743278, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  213019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213019, CurrentID=0, Status='live', Symbol='Pdlim2', Aliases='4732462F18Rik, Slim', description='PDZ and LIM domain 2', other_designations='PDZ-LIM protein mystique|STAT-interacting LIM protein|mystique', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70164218, end_position_on_the_genomic_accession=70177703, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  213027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213027, CurrentID=0, Status='live', Symbol='Evi5l', Aliases='1700084G18Rik, 2310039H16Rik, 3110007G05Rik, B130050I23Rik', description='ecotropic viral integration site 5 like', other_designations='ecotropic viral integration site 5-like', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4166538, end_position_on_the_genomic_accession=4208050, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  213043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213043, CurrentID=0, Status='live', Symbol='Aox2', Aliases='Aoh3, Aox3l1', description='aldehyde oxidase 2', other_designations='aldehyde oxidase 3-like 1|aldehyde oxidase homolog 3|aldehyde oxidase-like protein 3|azaheterocycle hydroxylase 2', map_location='1 29.0 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58276141, end_position_on_the_genomic_accession=58379264, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  213053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213053, CurrentID=0, Status='live', Symbol='Slc39a14', Aliases='FAD-123, ZIP-14, Zip14, fad123', description='solute carrier family 39 (zinc transporter), member 14', other_designations='Zrt- and Irt-like protein 14|factor for adipocyte differentiation 123', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70303467, end_position_on_the_genomic_accession=70351459, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  213054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213054, CurrentID=0, Status='live', Symbol='Gabpb2', Aliases='1810015F01Rik, 5830427M07, 9430006E19Rik, A430024B14Rik, AV050852-1, Gabpb2', description='GA repeat binding protein, beta 2', other_designations='GA-binding protein beta-2-1|GABP subunit beta-2|GABP subunit beta-2-1|GABPB-2', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95181766, end_position_on_the_genomic_accession=95217951, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  213056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213056, CurrentID=0, Status='live', Symbol='Fam126b', Aliases='C130065N10Rik, D1Ertd53e, D630010C10', description='family with sequence similarity 126, member B', other_designations='', map_location='1 29.12 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58522806, end_position_on_the_genomic_accession=58586333, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  213068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213068, CurrentID=0, Status='live', Symbol='Tmem71', Aliases='AI661017, E430024D12', description='transmembrane protein 71', other_designations='', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=66526212, end_position_on_the_genomic_accession=66561236, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  79202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=79202, CurrentID=0, Status='live', Symbol='Tnfrsf22', Aliases='2810028K06Rik, C130035G06Rik, SOBa, Tnfrh2, Tnfrsf1al2, mDcTrailr2', description='tumor necrosis factor receptor superfamily, member 22', other_designations='TNF receptor family member SOBa|TNF receptor homolog 2|decoy TRAIL receptor 2|tumor necrosis factor receptor p60 homolog 2|tumor necrosis factor receptor superfamily, member 1a-like 2', map_location='7 88.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143634808, end_position_on_the_genomic_accession=143649752, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  213081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213081, CurrentID=0, Status='live', Symbol='Wdr19', Aliases='C330027H04Rik, D330023L08Rik, DYF2, Ift144, PWDMP, mKIAA1638', description='WD repeat domain 19', other_designations='WD repeat membrane protein PWDMP', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65199658, end_position_on_the_genomic_accession=65260415, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  213084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213084, CurrentID=0, Status='live', Symbol='Cdkl3', Aliases='AI852479, B230379H01Rik', description='cyclin-dependent kinase-like 3', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52000858, end_position_on_the_genomic_accession=52089778, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  213109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213109, CurrentID=0, Status='live', Symbol='Phf3', Aliases='2310061N19Rik, mKIAA0244', description='PHD finger protein 3', other_designations='', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=30802342, end_position_on_the_genomic_accession=30873923, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  213119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213119, CurrentID=0, Status='live', Symbol='Itga10', Aliases='A630048L14', description='integrin, alpha 10', other_designations='integrin subunit alpha10', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96645584, end_position_on_the_genomic_accession=96664519, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  213121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213121, CurrentID=0, Status='live', Symbol='Ankrd35', Aliases='4732436F15Rik', description='ankyrin repeat domain 35', other_designations='', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96670131, end_position_on_the_genomic_accession=96691034, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  213171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213171, CurrentID=0, Status='live', Symbol='Prss27', Aliases='CAPH2, Mpn, marapsin', description='protease, serine 27', other_designations='channel-activating protease 2-like protein|pancreasin', map_location='17 12.21 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24038243, end_position_on_the_genomic_accession=24048378, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  210293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=210293, CurrentID=0, Status='live', Symbol='Dock10', Aliases='9330153B10Rik, A630054M16Rik, Jr4, Jr5, R75174, ZIZ3, Zizimin3, mKIAA0694', description='dedicator of cytokinesis 10', other_designations='zizimin-3', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=80501068, end_position_on_the_genomic_accession=80758704, orientation='minus', exon_count=62, OMIM='', no_hdr0=''),
  319520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319520, CurrentID=0, Status='live', Symbol='Dusp4', Aliases='2700078F24Rik, AI844617, BB104621, E130306H24Rik, MKP2', description='dual specificity phosphatase 4', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=34807610, end_position_on_the_genomic_accession=34819894, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  213208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213208, CurrentID=0, Status='live', Symbol='Il20rb', Aliases='AV228068, Fndc6, Gm186, Il20R2', description='interleukin 20 receptor beta', other_designations='fibronectin type III domain containing 6', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=100457719, end_position_on_the_genomic_accession=100486473, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  213211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213211, CurrentID=0, Status='live', Symbol='Rnf26', Aliases='8030450I18Rik', description='ring finger protein 26', other_designations='', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44110781, end_position_on_the_genomic_accession=44113051, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  213233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213233, CurrentID=0, Status='live', Symbol='Tapbpl', Aliases='BC017613, TAPBPL-Rr, Tapbpl', description='TAP binding protein-like', other_designations='TAP-binding protein-like|TAP-binding protein-related protein|TAPASIN-R|TAPBP-R|tapasin-like', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125223922, end_position_on_the_genomic_accession=125231860, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  213234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213234, CurrentID=0, Status='live', Symbol='Zbbx', Aliases='4931432L23', description='zinc finger, B-box domain containing', other_designations='zinc finger B-box domain containing protein 1', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=75037895, end_position_on_the_genomic_accession=75165050, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  213236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213236, CurrentID=0, Status='live', Symbol='Dnd1', Aliases='BC034897, RBMS4, Ter', description='DND microRNA-mediated repression inhibitor 1', other_designations='RNA-binding motif, single-stranded-interacting protein 4|dead end homolog 1|teratoma', map_location='18 19.46 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36763671, end_position_on_the_genomic_accession=36766214, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  213248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213248, CurrentID=0, Status='live', Symbol='Wdr49', Aliases='EG213248', description='WD repeat domain 49', other_designations='', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=75273928, end_position_on_the_genomic_accession=75482165, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  213262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213262, CurrentID=0, Status='live', Symbol='Fstl5', Aliases='9130207J01Rik', description='follistatin-like 5', other_designations='follistatin-like protein 5|m-D/Bsp120I 1-2|ollistatin-like 5', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=76074275, end_position_on_the_genomic_accession=76710010, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  213272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213272, CurrentID=0, Status='live', Symbol='Txndc2', Aliases='AU021712, Sptrx-1, Sptrx1, Trx4', description='thioredoxin domain containing 2 (spermatozoa)', other_designations='spermatid-specific thioredoxin 1|spermatid-specific thioredoxin-1|thioredoxin-4', map_location='17 E1.2-E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65637505, end_position_on_the_genomic_accession=65643550, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  606496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=606496, CurrentID=0, Status='live', Symbol='Gsk3a', Aliases='2700086H06Rik', description='glycogen synthase kinase 3 alpha', other_designations='GSK-3 alpha|serine/threonine-protein kinase GSK3A', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25228258, end_position_on_the_genomic_accession=25237851, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  213311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213311, CurrentID=0, Status='live', Symbol='Fbxl21', Aliases='D630045D17Rik, FBL3B, FBXL3B, Psttm', description='F-box and leucine-rich repeat protein 21', other_designations='F-box and leucine-rich repeat protein 3B|F-box/LRR-repeat protein 3B', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=56522388, end_position_on_the_genomic_accession=56538691, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  213326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213326, CurrentID=0, Status='live', Symbol='Scyl2', Aliases='AU015688, AW536445, BC030932, CVAK104, D10Ertd802e', description='SCY1-like 2 (S. cerevisiae)', other_designations='coated vesicle-associated kinase of 104 kDa', map_location='10 45.05 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=89638720, end_position_on_the_genomic_accession=89686292, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  213350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213350, CurrentID=0, Status='live', Symbol='Pddc1', Aliases='BC023835, D230016J19Rik', description='Parkinson disease 7 domain containing 1', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141408184, end_position_on_the_genomic_accession=141414125, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  213389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213389, CurrentID=0, Status='live', Symbol='Prdm9', Aliases='BC012016, Dsbc1, G1-419-29, Meisetz, PRDM9-B, Rcr1, repro7', description='PR domain containing 9', other_designations='PR domain zinc finger protein 9|PR domain-containing protein 9|PR-domain zinc finger protein|PRDM9 isoform Hst1f|PRDM9-A isoform Hst1f|VPR domain containing 9|histone H3 lysine-4 methyltransferase|hybrid sterility protein 1|meiosis-induced factor containing a PR/SET domain and zinc-finger motif', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15543072, end_position_on_the_genomic_accession=15564379, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  213391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213391, CurrentID=0, Status='live', Symbol='Rassf4', Aliases='3830411C14Rik, AD037, AI586137', description='Ras association (RalGDS/AF-6) domain family member 4', other_designations='Ras association (RalGDS/AF-6) domain family 4', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116632994, end_position_on_the_genomic_accession=116673836, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  213393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213393, CurrentID=0, Status='live', Symbol='8430408G22Rik', Aliases='Depp, Fseg', description='RIKEN cDNA 8430408G22 gene', other_designations='decidual protein induced by progesterone|fat-specific expressed|fat-specific-expressed gene protein', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116650684, end_position_on_the_genomic_accession=116652847, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  213402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213402, CurrentID=0, Status='live', Symbol='Armc2', Aliases='2610018I05Rik', description='armadillo repeat containing 2', other_designations='', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41907755, end_position_on_the_genomic_accession=42018382, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  213409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213409, CurrentID=0, Status='live', Symbol='Lemd1', Aliases='4930540I23Rik, 6430514M23Rik, AI428128, LEMP-1', description='LEM domain containing 1', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132190687, end_position_on_the_genomic_accession=132257382, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  213417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213417, CurrentID=0, Status='live', Symbol='Klhdc8a', Aliases='A630065K24Rik', description='kelch domain containing 8A', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132298490, end_position_on_the_genomic_accession=132307357, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  213435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213435, CurrentID=0, Status='live', Symbol='Mylk3', Aliases='D830007F02Rik, MLCK', description='myosin light chain kinase 3', other_designations='cardiac MLCK|cardiac-MLCK|cardiac-MyBP-C-associated Ca/CaM kinase|cardiac-specific myosin light chain kinase|putative myosin light chain kinase 3', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85324304, end_position_on_the_genomic_accession=85386340, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  213436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213436, CurrentID=0, Status='live', Symbol='Zcchc5', Aliases='D430021I08Rik, Gm375, Mar3, Mart3, ZHC5', description='zinc finger, CCHC domain containing 5', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106837082, end_position_on_the_genomic_accession=106840643, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  213438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213438, CurrentID=0, Status='live', Symbol='A630033H20Rik', Aliases='AI662791, P2ry10', description='RIKEN cDNA A630033H20 gene', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107148927, end_position_on_the_genomic_accession=107173662, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  213439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213439, CurrentID=0, Status='live', Symbol='Gpr174', Aliases='Gm376, Lypsr3', description='G protein-coupled receptor 174', other_designations='putative purinergic receptor FKSG79', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107255878, end_position_on_the_genomic_accession=107296769, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  213449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213449, CurrentID=0, Status='live', Symbol='Fam46d', Aliases='Gm377', description='family with sequence similarity 46, member D', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107792590, end_position_on_the_genomic_accession=107872911, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  213450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213450, CurrentID=0, Status='live', Symbol='Gm732', Aliases='', description='predicted gene 732', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107945735, end_position_on_the_genomic_accession=107948447, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  213452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213452, CurrentID=0, Status='live', Symbol='Dstyk', Aliases='A930019K20Rik, C430014H23Rik, C820013G01, Ripk5', description='dual serine/threonine and tyrosine protein kinase', other_designations='dusty PK|dusty protein kinase|receptor interacting protein kinase 5|receptor-interacting serine/threonine-protein kinase 5', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132417181, end_position_on_the_genomic_accession=132466959, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  213464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213464, CurrentID=0, Status='live', Symbol='Rbbp5', Aliases='4933411J24Rik, C330016J05', description='retinoblastoma binding protein 5', other_designations='RBBP-5', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132477347, end_position_on_the_genomic_accession=132505665, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  213469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213469, CurrentID=0, Status='live', Symbol='Lgi3', Aliases='AI841179, AW049851', description='leucine-rich repeat LGI family, member 3', other_designations='leubrin|leucine-rich glioma-inactivated protein 3', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70530821, end_position_on_the_genomic_accession=70538324, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  213484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213484, CurrentID=0, Status='live', Symbol='Nudt18', Aliases='BC036718', description='nudix (nucleoside diphosphate linked moiety X)-type motif 18', other_designations='2-hydroxy-dADP phosphatase|7,8-dihydro-8-oxoguanine phosphatase|mutT homolog 3|nucleoside diphosphate-linked moiety X motif 18|nudix motif 18', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70577847, end_position_on_the_genomic_accession=70582571, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  213491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213491, CurrentID=0, Status='live', Symbol='Szrd1', Aliases='1110022I03, D4Ertd22e', description='SUZ RNA binding domain containing 1', other_designations='UPF0485 protein C1orf144 homolog', map_location='4 73.38 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141112978, end_position_on_the_genomic_accession=141139799, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  213498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213498, CurrentID=0, Status='live', Symbol='Arhgef11', Aliases='B930073M02, E130307F09, PDZ-RhoGEF, Prg, RhoGEF, mKIAA0380', description='Rho guanine nucleotide exchange factor (GEF) 11', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87617548, end_position_on_the_genomic_accession=87738034, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  213499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213499, CurrentID=0, Status='live', Symbol='Fbxo42', Aliases='6720460I06Rik, mKIAA1332', description='F-box protein 42', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141147907, end_position_on_the_genomic_accession=141204062, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  213522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213522, CurrentID=0, Status='live', Symbol='Plekhg6', Aliases='BC026778, Gm155', description='pleckstrin homology domain containing, family G (with RhoGef domain) member 6', other_designations='PH domain-containing family G member 6', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125362651, end_position_on_the_genomic_accession=125380951, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  213527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213527, CurrentID=0, Status='live', Symbol='Pth2r', Aliases='Pthr2', description='parathyroid hormone 2 receptor', other_designations='PTH2 receptor|parathyroid hormone receptor 2', map_location='1 33.05 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65280282, end_position_on_the_genomic_accession=65389625, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  213539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213539, CurrentID=0, Status='live', Symbol='Bag2', Aliases='2610042A13Rik, BC016230', description='BCL2-associated athanogene 2', other_designations='BAG-2|bcl-2-associated athanogene 2', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=33745484, end_position_on_the_genomic_accession=33757750, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  213541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213541, CurrentID=0, Status='live', Symbol='Ythdf2', Aliases='9430020E02Rik, HGRG8, NY-REN-2', description='YTH domain family 2', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132184916, end_position_on_the_genomic_accession=132212256, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  213550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213550, CurrentID=0, Status='live', Symbol='Dis3l', Aliases='AV340375', description="DIS3 like exosome 3'-5' exoribonuclease", other_designations='DIS3 mitotic control homolog (S. cerevisiae)-like', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64306756, end_position_on_the_genomic_accession=64341423, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  319581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319581, CurrentID=0, Status='live', Symbol='Xkr5', Aliases='5430438H03Rik, XRG5', description='X-linked Kx blood group related 5', other_designations='X Kell blood group precursor-related family, member 5', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=18924153, end_position_on_the_genomic_accession=18952366, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  213556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213556, CurrentID=0, Status='live', Symbol='Plekhh2', Aliases='AI256725, E030001K05, mKIAA2028', description='pleckstrin homology domain containing, family H (with MyTH4 domain) member 2', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84511895, end_position_on_the_genomic_accession=84622142, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  319582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319582, CurrentID=0, Status='live', Symbol='6430573F11Rik', Aliases='Kiaa1456, TRM9L', description='RIKEN cDNA 6430573F11 gene', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=36457290, end_position_on_the_genomic_accession=36513013, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  319583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319583, CurrentID=0, Status='live', Symbol='Lig4', Aliases='5830471N16Rik, tiny', description='ligase IV, DNA, ATP-dependent', other_designations='DNA ligase IV|polydeoxyribonucleotide synthase [ATP] 4', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=9970020, end_position_on_the_genomic_accession=9976323, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  213573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213573, CurrentID=0, Status='live', Symbol='Cracr2b', Aliases='6330520A15Rik, BC026645, Cracr2a, Efcab4, Efcab4a', description='calcium release activated channel regulator 2B', other_designations='EF-hand calcium binding domain 4A|calcium release-activated channel regulator 2B', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141460874, end_position_on_the_genomic_accession=141466613, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  213575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213575, CurrentID=0, Status='live', Symbol='Dync2li1', Aliases='4933404O11Rik, CGI-60, D2lic, LIC3, mD2LIC', description='dynein cytoplasmic 2 light intermediate chain 1', other_designations='dynein 2 light intermediate chain', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84626496, end_position_on_the_genomic_accession=84655568, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  319586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319586, CurrentID=0, Status='live', Symbol='Celf5', Aliases='4930565A21Rik, Brunol5', description='CUGBP, Elav-like family member 5', other_designations='bruno-like 5, RNA binding protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81459227, end_position_on_the_genomic_accession=81482725, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  213582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213582, CurrentID=0, Status='live', Symbol='Map9', Aliases='5033421J10Rik, 5330427D05Rik, ASAP, Mtap9', description='microtubule-associated protein 9', other_designations='aster-associated protein', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=82358028, end_position_on_the_genomic_accession=82395268, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  213603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213603, CurrentID=0, Status='live', Symbol='Slc44a3', Aliases='BC010552', description='solute carrier family 44, member 3', other_designations='', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121459528, end_position_on_the_genomic_accession=121532344, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  213649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213649, CurrentID=0, Status='live', Symbol='Arhgef19', Aliases='6030432F23, 6430573B13Rik, WGEF', description='Rho guanine nucleotide exchange factor (GEF) 19', other_designations='ephexin-2|weakly similar to Rho GEF 5', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141239781, end_position_on_the_genomic_accession=141257915, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  213673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213673, CurrentID=0, Status='live', Symbol='9530068E07Rik', Aliases='C85061, Kct2', description='RIKEN cDNA 9530068E07 gene', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52396428, end_position_on_the_genomic_accession=52408723, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  213696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213696, CurrentID=0, Status='live', Symbol='Duoxa1', Aliases='BC019755, Nip, Nip1', description='dual oxidase maturation factor 1', other_designations='Numb-interacting protein 1|dual oxidase activator 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122303540, end_position_on_the_genomic_accession=122313744, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  213742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213742, CurrentID=0, Status='live', Symbol='Xist', Aliases='A430022B11, AI314753', description='inactive X specific transcripts', other_designations='', map_location='X 46.15 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103460373, end_position_on_the_genomic_accession=103483233, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  213753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213753, CurrentID=0, Status='live', Symbol='Zfp598', Aliases='BC023040, Ntrap, Znf598', description='zinc finger protein 598', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24669693, end_position_on_the_genomic_accession=24682016, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  213760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213760, CurrentID=0, Status='live', Symbol='Prepl', Aliases='2810457N15Rik, 9530014L06Rik, D030028O16Rik, mKIAA0436', description='prolyl endopeptidase-like', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=85062309, end_position_on_the_genomic_accession=85090274, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  213765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213765, CurrentID=0, Status='live', Symbol='Nutm1', Aliases='4932438M10, Nut', description='NUT midline carcinoma, family member 1', other_designations='nuclear protein in testis (Nut)|protein NUT', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112247948, end_position_on_the_genomic_accession=112259291, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  213773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213773, CurrentID=0, Status='live', Symbol='Tbl3', Aliases='9430070M15Rik', description='transducin (beta)-like 3', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24700653, end_position_on_the_genomic_accession=24707653, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  213783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213783, CurrentID=0, Status='live', Symbol='Plekhg1', Aliases='D10Ertd733e, Gm521, mKIAA1209', description='pleckstrin homology domain containing, family G (with RhoGef domain) member 1', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=3740251, end_position_on_the_genomic_accession=3967303, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  213788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213788, CurrentID=0, Status='live', Symbol='Chrm5', Aliases='AChR-M5, M5R, m5', description='cholinergic receptor, muscarinic 5', other_designations='muscarinic acetylcholine receptor 5', map_location='2 57.02 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112478733, end_position_on_the_genomic_accession=112537563, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  213819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213819, CurrentID=0, Status='live', Symbol='Casd1', Aliases='Cas1, Cast1', description='CAS1 domain containing 1', other_designations='O-acetyltransferase', map_location='6 1.81 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=4600563, end_position_on_the_genomic_accession=4643381, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  213827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213827, CurrentID=0, Status='live', Symbol='Arcn1', Aliases='4632432M07Rik, nur17', description='archain 1', other_designations='delta-COP|delta-coat protein|pale coat neuro', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44742143, end_position_on_the_genomic_accession=44767808, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  213895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213895, CurrentID=0, Status='live', Symbol='Bms1', Aliases='AA408648, AU020092, AU043373, BB007109, BC030906l, mKIAA0187, Bms1', description='BMS1, ribosome biogenesis factor', other_designations='BMS1 homolog, ribosome assembly protein|BMS1-like, ribosome assembly protein', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118383381, end_position_on_the_genomic_accession=118420018, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  213945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213945, CurrentID=0, Status='live', Symbol='Col28a1', Aliases='Col28, Gm466', description='collagen, type XXVIII, alpha 1', other_designations='collagen 28', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=7997100, end_position_on_the_genomic_accession=8194223, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  213948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213948, CurrentID=0, Status='live', Symbol='Atg9b', Aliases='Apg912, Apg9l2, Apgdc2, Gm574, Nos3as, eONE, sONE', description='autophagy related 9B', other_designations='APG9-like 2|ATG9 autophagy related 9 homolog B|autophagy 9-like 2 protein|nitric oxide synthase 3 antisense|nitric oxide synthase 3-overlapping antisense gene protein', map_location='5 11.49 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24384181, end_position_on_the_genomic_accession=24392143, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  213956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213956, CurrentID=0, Status='live', Symbol='Fam83f', Aliases='AW544981', description='family with sequence similarity 83, member F', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80671847, end_position_on_the_genomic_accession=80700425, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  213980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213980, CurrentID=0, Status='live', Symbol='Fbxw10', Aliases='Fbw10, HREP, SM25H2, SM2SH2', description='F-box and WD-40 domain protein 10', other_designations='F-box and WD-40 domain-containing protein 10', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62846990, end_position_on_the_genomic_accession=62877464, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  213988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213988, CurrentID=0, Status='live', Symbol='Tnrc6b', Aliases='2700090M07Rik, A730065C02Rik, AI848765, D230019K20Rik', description='trinucleotide repeat containing 6b', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80711313, end_position_on_the_genomic_accession=80941086, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  213989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213989, CurrentID=0, Status='live', Symbol='Tmem82', Aliases='', description='transmembrane protein 82', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141614229, end_position_on_the_genomic_accession=141618633, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  213990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213990, CurrentID=0, Status='live', Symbol='Agap3', Aliases='AGAP-3, AV028425, Centg3, Crag, Crag., MRIP-1, cnt-g3', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 3', other_designations='CRAM-associated GTPase|MR1-interacting protein|centaurin, gamma 3|centaurin-gamma-3|spidin', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24452177, end_position_on_the_genomic_accession=24502047, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  213993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=213993, CurrentID=0, Status='live', Symbol='Ccdc186', Aliases='1810028B20Rik, A630007B06Rik, AI159700, Otg1', description='coiled-coil domain containing 186', other_designations='oocyte-testis gene 1 protein', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56787485, end_position_on_the_genomic_accession=56822482, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  214048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214048, CurrentID=0, Status='live', Symbol='Larp1b', Aliases='1700108L22Rik, 4933421B21Rik, AA690246, Larp2', description='La ribonucleoprotein domain family, member 1B', other_designations='La ribonucleoprotein domain family, member 2', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40950631, end_position_on_the_genomic_accession=40977797, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  214058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214058, CurrentID=0, Status='live', Symbol='Megf11', Aliases='2410080H04Rik, D130061K05', description='multiple EGF-like-domains 11', other_designations='multiple EGF-like domains protein 11', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64385618, end_position_on_the_genomic_accession=64709205, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  214063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214063, CurrentID=0, Status='live', Symbol='Dnajc16', Aliases='2900037O03Rik, 4732437J24Rik, D030060M24, mKIAA0962', description='DnaJ heat shock protein family (Hsp40) member C16', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 16', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141761998, end_position_on_the_genomic_accession=141790644, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  232286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232286, CurrentID=0, Status='live', Symbol='Tmf1', Aliases='7030402D04Rik, Gm153`, Tmf1', description='TATA element modulatory factor 1', other_designations='TMF|androgen receptor coactivator 160 kDa protein|androgen receptor-associated protein of 160 kDa', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=97151950, end_position_on_the_genomic_accession=97179124, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  214084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214084, CurrentID=0, Status='live', Symbol='Slc18a2', Aliases='1110037L13Rik, 9330105E13, Vmat2', description='solute carrier family 18 (vesicular monoamine), member 2', other_designations='VAT2|monoamine transporter|solute carrier family 18 member 2|vesicular amine transporter 2|vesicular monoamine transporter 2', map_location='19 54.64 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=59260746, end_position_on_the_genomic_accession=59296012, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  214105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214105, CurrentID=0, Status='live', Symbol='Sox30', Aliases='', description='SRY (sex determining region Y)-box 30', other_designations='SRY-box containing gene 30', map_location='11 27.49 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=45979363, end_position_on_the_genomic_accession=46017992, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  214106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214106, CurrentID=0, Status='live', Symbol='4933430I17Rik', Aliases='bN189G18.7', description='RIKEN cDNA 4933430I17 gene', other_designations='bN189G18.7 (novel protein (4933430I17Rik))', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62525262, end_position_on_the_genomic_accession=62552684, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  214111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214111, CurrentID=0, Status='live', Symbol='Slc24a1', Aliases='', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 1', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64922861, end_position_on_the_genomic_accession=64951607, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  214112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214112, CurrentID=0, Status='live', Symbol='Nipal4', Aliases='9530066K23Rik', description='NIPA-like domain containing 4', other_designations='NIPA-like protein 4|ichthyin|non-imprinted in Prader-Willi/Angelman syndrome region protein 4 homolog', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46148155, end_position_on_the_genomic_accession=46166359, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  214133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214133, CurrentID=0, Status='live', Symbol='Tet2', Aliases='Ayu17-449, E130014J05Rik, mKIAA1546', description='tet methylcytosine dioxygenase 2', other_designations='probable methylcytosine dioxygenase TET2|tet oncogene 2|tet oncogene family member 2', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=133463677, end_position_on_the_genomic_accession=133545201, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  214137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214137, CurrentID=0, Status='live', Symbol='Arhgap29', Aliases='6720461J18Rik, A830014I19, AU040217, B130017I01Rik, C76601, Parg1', description='Rho GTPase activating protein 29', other_designations='PTPL1-associated RhoGAP 1|rho-type GTPase-activating protein 29', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121953072, end_position_on_the_genomic_accession=122016157, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  214150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214150, CurrentID=0, Status='live', Symbol='Ago3', Aliases='AW048688, C130014L07Rik, Eif2c3', description='argonaute RISC catalytic subunit 3', other_designations='Piwi/Argonaute family protein meIF2C3|argonaute 3|argonaute RISC catalytic component 3|argonaute3|eIF-2C 3|eIF2C 3|eukaryotic translation initiation factor 2C, 3|mAgo3', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126326990, end_position_on_the_genomic_accession=126434830, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  214158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214158, CurrentID=0, Status='live', Symbol='Trim38', Aliases='Gm23', description='tripartite motif-containing 38', other_designations='RING finger protein 15|zinc finger protein RoRet', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23778500, end_position_on_the_genomic_accession=23791731, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  214162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214162, CurrentID=0, Status='live', Symbol='Kmt2a', Aliases='6430520K01, ALL-1, All1, Cxxc7, HRX, HTRX1, Mll, Mll1, mKIAA4050', description='lysine (K)-specific methyltransferase 2A', other_designations='histone-lysine N-methyltransferase MLL|lysine N-methyltransferase 2A|myeloid/lymphoid or mixed-lineage leukemia 1|myeloid/lymphoid or mixed-lineage leukemia protein 1|trithorax Drosophila|zinc finger protein HRX', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44803355, end_position_on_the_genomic_accession=44881359, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  214189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214189, CurrentID=0, Status='live', Symbol='Scgn', Aliases='', description='secretagogin, EF-hand calcium binding protein', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23953456, end_position_on_the_genomic_accession=23991337, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  214191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214191, CurrentID=0, Status='live', Symbol='Ttc24', Aliases='A430025D11Rik', description='tetratricopeptide repeat domain 24', other_designations='TPR repeat protein 24', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88069410, end_position_on_the_genomic_accession=88081987, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  214230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214230, CurrentID=0, Status='live', Symbol='Pak6', Aliases='4732456M09', description='p21 protein (Cdc42/Rac)-activated kinase 6', other_designations='PAK-6|p21-activated kinase 6', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118663027, end_position_on_the_genomic_accession=118698020, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  625530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625530, CurrentID=0, Status='live', Symbol='Usp17le', Aliases='DUB-3, Dub3, Dub6, EG625530, Gm6596, USP17, USP17-E, Usp17l2', description='ubiquitin specific peptidase 17-like E', other_designations='deubiquitinating enzyme 17-like protein 2|deubiquitinating enzyme 3|deubiquitinating enzyme 6|deubiquitinating protein 3|deubiquitinating protein 6|ubiquitin carboxyl-terminal hydrolase 17|ubiquitin carboxyl-terminal hydrolase 17-like protein 2|ubiquitin thioesterase 17-like protein 2|ubiquitin thiolesterase 17-like protein 2|ubiquitin-specific-processing protease 17-like protein 2', map_location='7 55.69 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104767964, end_position_on_the_genomic_accession=104777470, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  214239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214239, CurrentID=0, Status='live', Symbol='A430105I19Rik', Aliases='Gm353', description='RIKEN cDNA A430105I19 gene', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118754145, end_position_on_the_genomic_accession=118762951, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  214240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214240, CurrentID=0, Status='live', Symbol='Disp2', Aliases='AI840597, B230210L08Rik, DispB, mKIAA1742', description='dispatched homolog 2 (Drosophila)', other_designations='dispatched B', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118779702, end_position_on_the_genomic_accession=118795175, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  214253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214253, CurrentID=0, Status='live', Symbol='Etnk2', Aliases='4933417N20Rik, AI197444, EKI 2, Eki2, Tuc1', description='ethanolamine kinase 2', other_designations='ethanolamine kinase-like protein|putative ethanolamine kinase', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133363488, end_position_on_the_genomic_accession=133380318, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  214254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214254, CurrentID=0, Status='live', Symbol='Nudt15', Aliases='6530403O17, A730068G11Rik, MTH2', description='nudix (nucleoside diphosphate linked moiety X)-type motif 15', other_designations='7,8-dihydro-8-oxoguanine-triphosphatase NUDT15|8-oxo-dGTPase NUDT15|MutT homolog 2|mMTH2|nucleoside diphosphate-linked moiety X motif 15|nudix motif 15|probable 7,8-dihydro-8-oxoguanine triphosphatase NUDT15', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73519864, end_position_on_the_genomic_accession=73548242, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  214290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214290, CurrentID=0, Status='live', Symbol='Zcchc6', Aliases='6030448M23Rik, AA420405', description='zinc finger, CCHC domain containing 6', other_designations='TUTase 7|zinc finger CCHC domain-containing protein 6', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59771876, end_position_on_the_genomic_accession=59823167, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  214292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214292, CurrentID=0, Status='live', Symbol='Syna', Aliases='Gm453, Gm52, syncytin-A', description='syncytin a', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134557254, end_position_on_the_genomic_accession=134577057, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  214301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214301, CurrentID=0, Status='live', Symbol='Crygn', Aliases='', description='crystallin, gamma N', other_designations='gamma-N-crystallin', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24751002, end_position_on_the_genomic_accession=24757843, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  214305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214305, CurrentID=0, Status='live', Symbol='Hhipl1', Aliases='1600002O04Rik, AK129450, mKIAA1822', description='hedgehog interacting protein-like 1', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108305610, end_position_on_the_genomic_accession=108328300, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  214321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214321, CurrentID=0, Status='live', Symbol='Gm4787', Aliases='Adam4b, EG214321', description='predicted gene 4787', other_designations='a disintegrin and metallopeptidase domain 4b', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81376991, end_position_on_the_genomic_accession=81379486, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  214345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214345, CurrentID=0, Status='live', Symbol='Lrrc1', Aliases='A430093J20Rik, AU016091, D030059I21, mKIAA4018', description='leucine rich repeat containing 1', other_designations='', map_location='9 D-E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=77430823, end_position_on_the_genomic_accession=77544852, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  214359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214359, CurrentID=0, Status='live', Symbol='Tmem51', Aliases='BC003277', description='transmembrane protein 51', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=142030990, end_position_on_the_genomic_accession=142084954, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  385253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385253, CurrentID=0, Status='live', Symbol='Igkv4-55', Aliases='Gm1524, at4', description='immunoglobulin kappa variable 4-55', other_designations='Ig kappa V-region from NQ2-6.1|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69607275, end_position_on_the_genomic_accession=69607801, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  214384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214384, CurrentID=0, Status='live', Symbol='Myocd', Aliases='BSAC2A, Srfcp', description='myocardin', other_designations='SRF co-factor protein (cardiac and smooth muscle)|SRF cofactor protein|basic SAP coiled-coil transcription activator 2', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=65173608, end_position_on_the_genomic_accession=65269989, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  214424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214424, CurrentID=0, Status='live', Symbol='Parp16', Aliases='ARTD15, BC055447, C79952, PARP-16', description='poly (ADP-ribose) polymerase family, member 16', other_designations='ADP-ribosyltransferase diphtheria toxin-like 15|poly [ADP-ribose] polymerase 16', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65214380, end_position_on_the_genomic_accession=65239220, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  214425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214425, CurrentID=0, Status='live', Symbol='Cilp', Aliases='9830114D16, C130036G17Rik', description='cartilage intermediate layer protein, nucleotide pyrophosphohydrolase', other_designations='CILP-1', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65265180, end_position_on_the_genomic_accession=65280605, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  214444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214444, CurrentID=0, Status='live', Symbol='Cdk5rap2', Aliases='2900018K03Rik, an, mKIAA1633', description='CDK5 regulatory subunit associated protein 2', other_designations='CDK5 activator-binding protein C48', map_location='4 C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=70217690, end_position_on_the_genomic_accession=70410461, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  83379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83379, CurrentID=0, Status='live', Symbol='Klb', Aliases='AV071179, betaKlotho', description='klotho beta', other_designations='BKL|klotho beta-like protein', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65348411, end_position_on_the_genomic_accession=65384020, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  83380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83380, CurrentID=0, Status='live', Symbol='Prp2', Aliases='MP14', description='proline rich protein 2', other_designations='proline-rich protein MP-3', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132595946, end_position_on_the_genomic_accession=132600702, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  83382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83382, CurrentID=0, Status='live', Symbol='Siglece', Aliases='Cd170, Siglec12, Siglec5, Siglec9, Siglecl1, mSiglec-E', description='sialic acid binding Ig-like lectin E', other_designations='MIS|SIGLEC-like 1|myeloid inhibitory siglec|sialic acid binding Ig-like lectin 5|sialic acid-binding Ig-like lectin 5|sialic acid-binding Ig-like lectin E|sialic acid-binding Ig-like lectin-like 1|sialic acid-binding immunoglobulin-like lectin E|siglec-12|siglec-5|siglec-E|siglec-L1', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43651070, end_position_on_the_genomic_accession=43660179, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  83383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83383, CurrentID=0, Status='live', Symbol='Tfap4', Aliases='AI642933, AP-4, D930048N17Rik, Tcfap4, bHLHc41', description='transcription factor AP4', other_designations='transcription factor AP-4 (activating enhancer-binding protein 4)', map_location='16 2.45 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4544661, end_position_on_the_genomic_accession=4560772, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  214459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214459, CurrentID=0, Status='live', Symbol='Fnbp1l', Aliases='2610318I01Rik, AW548221, TOCA1, toca-1', description='formin binding protein 1-like', other_designations='transducer of Cdc42-dependent actin assembly protein 1', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122538719, end_position_on_the_genomic_accession=122619714, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  83395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83395, CurrentID=0, Status='live', Symbol='Sp6', Aliases='1110025J03Rik, AA591031, AI592962, Epfn, Klf14', description='trans-acting transcription factor 6', other_designations='Kruppel-like factor 14|Sp6 transcription factor|epiprofin|krueppel-like factor 14|specificity protein 6', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97013569, end_position_on_the_genomic_accession=97024739, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83396, CurrentID=0, Status='live', Symbol='Glis2', Aliases='Gli5, Klf16, Nkl', description='GLIS family zinc finger 2', other_designations='GLI-Kruppel family member GLI5|GLI-similar 2|Kruppel-like factor 16 (neuronal)|neuronal Krueppel-like protein|zinc finger protein GLI5', map_location='16 A1-B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4594713, end_position_on_the_genomic_accession=4625141, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  214469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214469, CurrentID=0, Status='live', Symbol='Fam168b', Aliases='mKIAA4042', description='family with sequence similarity 168, member B', other_designations='mani|protein FAM168B', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34813218, end_position_on_the_genomic_accession=34843050, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  83398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83398, CurrentID=0, Status='live', Symbol='Ndst3', Aliases='4921531K01Rik, 4930511P15Rik, N-HSST 3, NDST-3', description='N-deacetylase/N-sulfotransferase (heparan glucosaminyl) 3', other_designations='N-heparan sulfate sulfotransferase 3|glucosaminyl N-deacetylase/N-sulfotransferase 3', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=123526166, end_position_on_the_genomic_accession=123690852, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  83408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83408, CurrentID=0, Status='live', Symbol='Gimap3', Aliases='2010110D23Rik, Gimap5, Ian4', description='GTPase, IMAP family member 3', other_designations='IAN-4|immune associated nucleotide 4|immunity-associated nucleotide 4 protein', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48764464, end_position_on_the_genomic_accession=48770851, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  83409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83409, CurrentID=0, Status='live', Symbol='Lamtor2', Aliases='2010111E04Rik, AL022628, Mapbpip, P14, Rab25, Robld3', description='late endosomal/lysosomal adaptor, MAPK and MTOR activator 2', other_designations='endosomal adaptor protein p14|late endosomal/lysosomal MP1 interacting protein|late endosomal/lysosomal Mp1-interacting protein|late endosomal/lysosomal adaptor and MAPK and MTOR activator 2|mitogen activated protein binding protein interacting protein|mitogen-activated protein binding protein interacting protein|mitogen-activated protein-binding protein-interacting protein|roadblock domain containing 3|roadblock domain-containing protein 3', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88549819, end_position_on_the_genomic_accession=88552927, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  83410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83410, CurrentID=0, Status='live', Symbol='Cstf2t', Aliases='64kDa, C77975, tCstF-64, tauCstF-64', description="cleavage stimulation factor, 3' pre-RNA subunit 2, tau", other_designations="CF-1 64 kDa subunit tau|CF-1 64 kDa subunit, tau|CSTF 64 kDa subunit tau|CSTF 64 kDa subunit, tau|cleavage stimulation factor 64 kDa subunit tau|cleavage stimulation factor 64 kDa subunit, tau|cleavage stimulation factor subunit 2 tau|cleavage stimulation factor, 3' pre-RNA subunit 2, 64 kDa, tau", map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=31082841, end_position_on_the_genomic_accession=31086592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  214489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214489, CurrentID=0, Status='live', Symbol='BC003965', Aliases='Ccsmst1', description='cDNA sequence BC003965', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25184561, end_position_on_the_genomic_accession=25187662, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  214498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214498, CurrentID=0, Status='live', Symbol='Cdc73', Aliases='8430414L16Rik, C130030P16Rik, C81219, Hrpt2', description='cell division cycle 73, Paf1/RNA polymerase II complex component', other_designations='cell division cycle protein 73 homolog|hyperparathyroidism 2 protein homolog', map_location='1 62.52 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=143607499, end_position_on_the_genomic_accession=143702888, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  83428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83428, CurrentID=0, Status='live', Symbol='Ucn3', Aliases='', description='urocortin 3', other_designations='Urocortin II|Urocortin III|ucn III', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3940688, end_position_on_the_genomic_accession=3945349, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  83429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83429, CurrentID=0, Status='live', Symbol='Ctns', Aliases='AI195360, AW049661', description='cystinosis, nephropathic', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73183133, end_position_on_the_genomic_accession=73199029, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  83430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83430, CurrentID=0, Status='live', Symbol='Il23a', Aliases='IL-23, p19', description='interleukin 23, alpha subunit p19', other_designations='IL-23 subunit alpha|IL-23-A|IL-23p19|interleukin-23 subunit p19', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128296140, end_position_on_the_genomic_accession=128298084, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  83431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83431, CurrentID=0, Status='live', Symbol='Ndel1', Aliases='2600006O07Rik, MITAP1, NUDEL, mNudel', description='nuclear distribution gene E-like homolog 1 (A. nidulans)', other_designations='mNudE-L|nuclear disribution gene E like|protein Nudel', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68821434, end_position_on_the_genomic_accession=68853131, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  83433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83433, CurrentID=0, Status='live', Symbol='Trem2', Aliases='TREM-2a, Trem2b, Trem2c, Trem2', description='triggering receptor expressed on myeloid cells 2', other_designations='triggering receptor expressed on monocytes 2|triggering receptor expressed on myeloid cells 2a|triggering receptor expressed on myeloid cells 2b|triggering receptor expressed on myeloid cells 2c', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48346401, end_position_on_the_genomic_accession=48352276, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  83434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83434, CurrentID=0, Status='live', Symbol='Rsph6a', Aliases='RSP4, Rshl1', description='radial spoke head 6 homolog A (Chlamydomonas)', other_designations='radial spoke head-like protein 1|radial spokehead-like 1', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19054678, end_position_on_the_genomic_accession=19074447, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  83435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83435, CurrentID=0, Status='live', Symbol='Plekha3', Aliases='FAPP1', description='pleckstrin homology domain-containing, family A (phosphoinositide binding specific) member 3', other_designations='FAPP-1|PH domain-containing family A member 3|phosphatidylinositol-four-phosphate adapter protein 1|phosphoinositol 4-phosphate adapter protein 1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76675310, end_position_on_the_genomic_accession=76697335, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  83436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83436, CurrentID=0, Status='live', Symbol='Plekha2', Aliases='6430512N22Rik, AI851530, AU041791, TAPP2, W91773', description='pleckstrin homology domain-containing, family A (phosphoinositide binding specific) member 2', other_designations='PH domain-containing adaptor PHAD47|PH domain-containing family A member 2|TAPP-2|pleckstrin homology domain-containing, family A (phosphoinositide binding specific) member 2andem PH Domain containing protein 2|tandem PH domain-containing protein 2', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25039144, end_position_on_the_genomic_accession=25104917, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  214523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214523, CurrentID=0, Status='live', Symbol='Tmprss4', Aliases='mCAP2', description='transmembrane protease, serine 4', other_designations='channel-activating protease 2|membrane-type serine protease 2', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45172726, end_position_on_the_genomic_accession=45204075, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  83453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83453, CurrentID=0, Status='live', Symbol='Chrdl1', Aliases='CHL, CHL1, Nrln1, VOPT', description='chordin-like 1', other_designations='neuralin-1|neurogenesin-1|ventroptin-alpha|ventroptin-beta', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=143285674, end_position_on_the_genomic_accession=143394262, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  83454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83454, CurrentID=0, Status='live', Symbol='Nxf2', Aliases='4930455N06Rik, Nxf-b', description='nuclear RNA export factor 2', other_designations='', map_location='X 56.31 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134944525, end_position_on_the_genomic_accession=134964754, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  83456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83456, CurrentID=0, Status='live', Symbol='Mov10l1', Aliases='CHAMP, Csm', description='Moloney leukemia virus 10-like 1', other_designations='MOV10-like protein 1|cardiac helicase activated by MEF2 protein|cardiac-specific RNA helicase|moloney leukemia virus 10-like protein 1', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88982891, end_position_on_the_genomic_accession=89055152, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  83457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83457, CurrentID=0, Status='live', Symbol='Fthl17e', Aliases='Fthl17', description='ferritin, heavy polypeptide-like 17, member E', other_designations='ferritin, heavy polypeptide-like 17', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9033486, end_position_on_the_genomic_accession=9034333, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  214531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214531, CurrentID=0, Status='live', Symbol='Tmprss13', Aliases='BC010843, Tpmrss13', description='transmembrane protease, serine 13', other_designations='membrane-type mosaic serine protease', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45319071, end_position_on_the_genomic_accession=45347581, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  214547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214547, CurrentID=0, Status='live', Symbol='She', Aliases='9430022A14', description='src homology 2 domain-containing transforming protein E', other_designations='Src homology 2 domain containing E', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89831370, end_position_on_the_genomic_accession=89858846, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  214552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214552, CurrentID=0, Status='live', Symbol='Cep164', Aliases='AI450905, BC027092, D030051D21, mKIAA1052', description='centrosomal protein 164', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45766946, end_position_on_the_genomic_accession=45828686, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  83485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83485, CurrentID=0, Status='live', Symbol='Ngrn', Aliases='AW552001', description='neugrin, neurite outgrowth associated', other_designations='FI58Gm|m-Neugrin|neurite outgrowth associated protein|neurite outgrowth-associated protein', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80261215, end_position_on_the_genomic_accession=80265378, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83486, CurrentID=0, Status='live', Symbol='Rbm5', Aliases='D030069N10Rik', description='RNA binding motif protein 5', other_designations='RNA-binding motif protein 5|putative tumor suppressor LUCA15', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107740495, end_position_on_the_genomic_accession=107771008, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  83490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83490, CurrentID=0, Status='live', Symbol='Pik3ap1', Aliases='1810044J04Rik, BCAP', description='phosphoinositide-3-kinase adaptor protein 1', other_designations='B cell phosphoinositide 3-kinase adaptor|B-cell adapter for phosphoinositide 3-kinase|B-cell phosphoinositide 3-kinase adapter protein 1', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41272752, end_position_on_the_genomic_accession=41385070, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  83491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83491, CurrentID=0, Status='live', Symbol='Pramel1', Aliases='', description='preferentially expressed antigen in melanoma-like 1', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143394359, end_position_on_the_genomic_accession=143400160, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  83492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83492, CurrentID=0, Status='live', Symbol='Gsdmc', Aliases='Gsdmc1, Mlze', description='gasdermin C', other_designations='gasdermin C1|melanoma-derived leucine zipper, extra-nuclear factor|melanoma-derived leucine zipper-containing extranuclear factor', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=63775971, end_position_on_the_genomic_accession=63808739, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  83493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83493, CurrentID=0, Status='live', Symbol='Sacm1l', Aliases='SAC1, Sac1p, mKIAA0851', description='SAC1 (suppressor of actin mutations 1, homolog)-like (S. cerevisiae)', other_designations='polyphosphatidylinositol phosphatase|suppressor of actin mutations 1-like protein', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123529882, end_position_on_the_genomic_accession=123592598, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  214568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214568, CurrentID=0, Status='live', Symbol='Gm136', Aliases='', description='predicted gene 136', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34743788, end_position_on_the_genomic_accession=34756259, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  214572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214572, CurrentID=0, Status='live', Symbol='Prmt7', Aliases='4933402B05Rik, BC006705', description='protein arginine N-methyltransferase 7', other_designations='[Myelin basic protein]-arginine N-methyltransferase PRMT7|histone-arginine N-methyltransferase PRMT7', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106211054, end_position_on_the_genomic_accession=106251694, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  214575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214575, CurrentID=0, Status='live', Symbol='Tdrd5', Aliases='Gm103', description='tudor domain containing 5', other_designations='tudor domain containing protein 5', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156255296, end_position_on_the_genomic_accession=156305295, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  214579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214579, CurrentID=0, Status='live', Symbol='Aldh5a1', Aliases='6330403E24Rik, Ahd-1, Ahd1, D630032B01Rik, OTTMUSG00000000613, SSADH, Ssdh1', description='aldhehyde dehydrogenase family 5, subfamily A1', other_designations='NAD(+)-dependent succinic semialdehyde dehydrogenase|aldehyde dehydrogenase 1, mitochondrial|aldehyde dehydrogenase family 5 member A1|aldehyde dehydrogenase family 5, subfamily A1', map_location='13 10.77 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24907579, end_position_on_the_genomic_accession=24937661, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  214580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214580, CurrentID=0, Status='live', Symbol='Pstk', Aliases='5430423O14Rik, 5730458D16Rik, Gm1099, R75284', description='phosphoseryl-tRNA kinase', other_designations='O-phosphoseryl-tRNA(Sec) kinase', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131371113, end_position_on_the_genomic_accession=131392768, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  214585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214585, CurrentID=0, Status='live', Symbol='Spg11', Aliases='6030465E24Rik, A330015I11, C530005A01Rik', description='spastic paraplegia 11', other_designations='spastic paraplegia 11 protein homolog', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122053526, end_position_on_the_genomic_accession=122118415, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  214593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214593, CurrentID=0, Status='live', Symbol='Duox2', Aliases='A430065P05Rik, LNOX2, NOXEF2, P138-TOX, THOX2', description='dual oxidase 2', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122279246, end_position_on_the_genomic_accession=122298742, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  214597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214597, CurrentID=0, Status='live', Symbol='Sidt2', Aliases='B930096O19, BC023957, CGI-40', description='SID1 transmembrane family, member 2', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45937857, end_position_on_the_genomic_accession=45955249, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  214601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214601, CurrentID=0, Status='live', Symbol='Slc10a3', Aliases='DXHXS253E, DXS253Eh, DXSmhG28, P3', description='solute carrier family 10 (sodium/bile acid cotransporter family), member 3', other_designations='protein P3|solute carrier family 10 member 3|solute carrier family 10, member 3', map_location='X 38.0 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74369217, end_position_on_the_genomic_accession=74373349, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  214604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214604, CurrentID=0, Status='live', Symbol='4932411E22Rik', Aliases='', description='RIKEN cDNA 4932411E22 gene', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=89390223, end_position_on_the_genomic_accession=89418948, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  214616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214616, CurrentID=0, Status='live', Symbol='Spata5l1', Aliases='AV141009, C130039A10Rik', description='spermatogenesis associated 5-like 1', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122630625, end_position_on_the_genomic_accession=122636478, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83553, CurrentID=0, Status='live', Symbol='Tktl1', Aliases='C79967', description='transketolase-like 1', other_designations='TK 2|tktl1|transketolase 2', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74177259, end_position_on_the_genomic_accession=74208498, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  83554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83554, CurrentID=0, Status='live', Symbol='Fstl3', Aliases='E030038F23Rik, Flrg', description='follistatin-like 3', other_designations='follistatin-like protein 3|follistatin-related gene protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79777274, end_position_on_the_genomic_accession=79782630, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  214627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214627, CurrentID=0, Status='live', Symbol='Papd5', Aliases='5730445M16Rik, 5830428A09', description='PAP associated domain containing 5', other_designations='PAP-associated domain-containing protein 5|TRF4-2|TUTase 3|terminal uridylyltransferase 3|topoisomerase-related function protein 4-2', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88199213, end_position_on_the_genomic_accession=88262997, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  83556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83556, CurrentID=0, Status='live', Symbol='Tex16', Aliases='OTTMUSG00000018437', description='testis expressed gene 16', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=111894858, end_position_on_the_genomic_accession=112127326, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  83557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83557, CurrentID=0, Status='live', Symbol='Lin28a', Aliases='AL024421, ENSMUSG00000070700, Gm10299, Lin-28, Lin28, Tex17, lin-28A', description='lin-28 homolog A (C. elegans)', other_designations='RNA-binding protein LIN-28|testis expressed gene 17|testis-expressed protein 17', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134003330, end_position_on_the_genomic_accession=134019375, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  83558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83558, CurrentID=0, Status='live', Symbol='Tex11', Aliases='4930565P14Rik, BB024231, ZIP4H', description='testis expressed gene 11', other_designations='protein ZIP4 homolog', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100838648, end_position_on_the_genomic_accession=101059639, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  83560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83560, CurrentID=0, Status='live', Symbol='Tex14', Aliases='C85585', description='testis expressed gene 14', other_designations='testis-expressed protein 14|testis-expressed sequence 14 protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87405050, end_position_on_the_genomic_accession=87555823, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  83561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83561, CurrentID=0, Status='live', Symbol='Tdrd1', Aliases='MTR-1', description='tudor domain containing 1', other_designations='', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56826171, end_position_on_the_genomic_accession=56870012, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  83563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83563, CurrentID=0, Status='live', Symbol='Usp26', Aliases='', description='ubiquitin specific peptidase 26', other_designations='deubiquitinating enzyme 26|ubiquitin specific protease 26|ubiquitin thioesterase 26|ubiquitin thiolesterase 26|ubiquitin-specific-processing protease 26', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=51752667, end_position_on_the_genomic_accession=51801233, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  83564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83564, CurrentID=0, Status='live', Symbol='Nlrp4c', Aliases='Nalp-alpha, Nalp4c, Rnh2', description='NLR family, pyrin domain containing 4C', other_designations='NACHT, leucine rich repeat and PYD containing 4C|VNACHT, leucine rich repeat and PYD containing 4C|ribonuclease/angiogenin inhibitor 2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6045161, end_position_on_the_genomic_accession=6105509, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  83565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83565, CurrentID=0, Status='live', Symbol='Pramel3', Aliases='', description='preferentially expressed antigen in melanoma-like 3', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135302289, end_position_on_the_genomic_accession=135312636, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  214639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214639, CurrentID=0, Status='live', Symbol='4930486L24Rik', Aliases='AW545031', description='RIKEN cDNA 4930486L24 gene', other_designations='testin', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60842612, end_position_on_the_genomic_accession=60864416, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  214642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214642, CurrentID=0, Status='live', Symbol='Cped1', Aliases='6720481P07, A430107O13Rik, AI552584', description='cadherin-like and PC-esterase domain containing 1', other_designations='', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=21985705, end_position_on_the_genomic_accession=22256407, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  214663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214663, CurrentID=0, Status='live', Symbol='Slc25a29', Aliases='C030003J19Rik, CACL, mCACL', description='solute carrier family 25 (mitochondrial carrier, palmitoylcarnitine transporter), member 29', other_designations='carnitine/acylcarnitine translocase-like|mitochondrial carnitine/acylcarnitine carrier protein CACL|mitochondrial ornithine transporter 3|solute carrier family 25 member 29', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108825878, end_position_on_the_genomic_accession=108835876, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  214669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214669, CurrentID=0, Status='live', Symbol='L3mbtl2', Aliases='4732493N06Rik, M4mbt', description='l(3)mbt-like 2 (Drosophila)', other_designations='l(3)mbt-like protein 2', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81663889, end_position_on_the_genomic_accession=81688315, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  83602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83602, CurrentID=0, Status='live', Symbol='Gtf2a1', Aliases='6330549H03Rik, AA536742, AA959775, AW060250, TfIIAa/b, Tfiia1', description='general transcription factor II A, 1', other_designations='TFIIA alpha/beta|general transcription factor IIA subunit 1|general transcription factor IIa, 1 (37kD and 19kD subunits)', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=91555262, end_position_on_the_genomic_accession=91590487, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  83603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83603, CurrentID=0, Status='live', Symbol='Elovl4', Aliases='', description='elongation of very long chain fatty acids (FEN1/Elo2, SUR4/Elo3, yeast)-like 4', other_designations='3-keto acyl-CoA synthase Elovl4|ELOVL FA elongase 4|ELOVL fatty acid elongase 4|very long chain 3-ketoacyl-CoA synthase 4|very long chain 3-oxoacyl-CoA synthase 4|very-long-chain 3-oxoacyl-CoA synthase 4', map_location='9 E2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=83778692, end_position_on_the_genomic_accession=83806305, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  214685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214685, CurrentID=0, Status='live', Symbol='Chadl', Aliases='AY100452, D930017K21Rik', description='chondroadherin-like', other_designations='chondroadherin pseudogene|leucine rich repeat containing 15 pseudogene', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81686166, end_position_on_the_genomic_accession=81697376, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  214704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214704, CurrentID=0, Status='live', Symbol='Iqub', Aliases='4932408B21Rik, Trs4', description='IQ motif and ubiquitin domain containing', other_designations='', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24444865, end_position_on_the_genomic_accession=24518558, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  83669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83669, CurrentID=0, Status='live', Symbol='Wdr6', Aliases='mWDR6', description='WD repeat domain 6', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108572313, end_position_on_the_genomic_accession=108578670, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  214742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214742, CurrentID=0, Status='live', Symbol='Rcor3', Aliases='4921514E24Rik, C730034D20Rik, E130101E15Rik', description='REST corepressor 3', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=192098546, end_position_on_the_genomic_accession=192138067, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  83671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83671, CurrentID=0, Status='live', Symbol='Sytl2', Aliases='AI266830, Slp2, mKIAA1597', description='synaptotagmin-like 2', other_designations='exophilin-4', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90302207, end_position_on_the_genomic_accession=90410719, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  83672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83672, CurrentID=0, Status='live', Symbol='Sytl3', Aliases='A430092N21, Slp3, Slp3-a, Slp3-b, mFLJ00307', description='synaptotagmin-like 3', other_designations='exophilin-6', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6658233, end_position_on_the_genomic_accession=6738070, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  83673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83673, CurrentID=0, Status='live', Symbol='Snhg1', Aliases='', description='small nucleolar RNA host gene 1', other_designations='small nucleolar RNA host gene (non-protein coding) 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8723487, end_position_on_the_genomic_accession=8726326, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  83674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83674, CurrentID=0, Status='live', Symbol='Cnnm1', Aliases='AI846314, Acdp1, CLP-1', description='cyclin M1', other_designations='ancient conserved domain protein 1|ancient conserved domain-containing protein 1|cyclin-M1|cyclin-like protein 1|mACDP1', map_location='19 36.67 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43440425, end_position_on_the_genomic_accession=43497213, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  83675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83675, CurrentID=0, Status='live', Symbol='Bicc1', Aliases='Bic-C, bpk, jcpk', description='BicC family RNA binding protein 1', other_designations='bicaudal C homolog 1|juvenile congenital polycystic kidney disease|polycystic kidney mutation of BALB/c origin', map_location='10 36.75 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=70922837, end_position_on_the_genomic_accession=71159746, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  83679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83679, CurrentID=0, Status='live', Symbol='Pde4dip', Aliases='4732458A06Rik, 9430063L05Rik, C87016, D130016K21Rik, D3Bwg1078e, Usmg4, mKIAA0454', description='phosphodiesterase 4D interacting protein (myomegalin)', other_designations='phosphodiesterase 4D-interacting protein|upregulated during skeletal muscle growth 4', map_location='3 42.28 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97689828, end_position_on_the_genomic_accession=97888707, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  214763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214763, CurrentID=0, Status='live', Symbol='Mb21d1', Aliases='E330016A19Rik, cGAS', description='Mab-21 domain containing 1', other_designations="2'3'-cGAMP synthase|cGAMP synthase|mab-21 domain-containing protein 1", map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78430518, end_position_on_the_genomic_accession=78443524, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  214764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214764, CurrentID=0, Status='live', Symbol='Edrf1', Aliases='2700050L05Rik, AU022667, AW558805', description='erythroid differentiation regulatory factor 1', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133637536, end_position_on_the_genomic_accession=133673017, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  214766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214766, CurrentID=0, Status='live', Symbol='Mmp21', Aliases='b2b2458Clo, b2b873Clo', description='matrix metallopeptidase 21', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133671568, end_position_on_the_genomic_accession=133682438, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  83701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83701, CurrentID=0, Status='live', Symbol='Srrt', Aliases='2810019G02Rik, ASR2A, ASR2B, ASR2C, ASR2D, Ars2, Asr2', description='serrate RNA effector molecule homolog (Arabidopsis)', other_designations='arsenate resistance protein 2|arsenite-resistance protein 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137295704, end_position_on_the_genomic_accession=137307714, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  83702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83702, CurrentID=0, Status='live', Symbol='Akr1c6', Aliases='3alpha-HSD, Akr1c1, Hsd17b5', description='aldo-keto reductase family 1, member C6', other_designations='17-beta-HSD 5|17HSD type 5|aldo-keto reductase family 1, member C1 (aldehyde reductase)|estradiol 17-beta-dehydrogenase (A-specific)|hydroxysteroid 17-beta dehydrogenase 5', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4434311, end_position_on_the_genomic_accession=4457530, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  83703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83703, CurrentID=0, Status='live', Symbol='Dbr1', Aliases='AW018415', description='debranching RNA lariats 1', other_designations='debranching enzyme homolog 1', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99575775, end_position_on_the_genomic_accession=99584461, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  83704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83704, CurrentID=0, Status='live', Symbol='Slc12a9', Aliases='CIP1', description='solute carrier family 12 (potassium/chloride transporters), member 9', other_designations='cation-chloride cotransporter-interacting protein 1|potassium-chloride transporter 9', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137314558, end_position_on_the_genomic_accession=137335688, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  214779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214779, CurrentID=0, Status='live', Symbol='Zfp879', Aliases='Znf879', description='zinc finger protein 879', other_designations='novel zinc finger protein', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50832031, end_position_on_the_genomic_accession=50841563, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  214791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214791, CurrentID=0, Status='live', Symbol='Sertad4', Aliases='4833404A09, C130018M11Rik', description='SERTA domain containing 4', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=192844469, end_position_on_the_genomic_accession=192855752, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  214804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214804, CurrentID=0, Status='live', Symbol='Syde2', Aliases='C430017H16Rik', description='synapse defective 1, Rho GTPase, homolog 2 (C. elegans)', other_designations='', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145987356, end_position_on_the_genomic_accession=146021720, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  214812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214812, CurrentID=0, Status='live', Symbol='Zfp609', Aliases='9830165N24, BC016271, Znf609, mKIAA0295', description='zinc finger protein 609', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65691578, end_position_on_the_genomic_accession=65827564, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  83762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83762, CurrentID=0, Status='live', Symbol='Otof', Aliases='', description='otoferlin', other_designations='fer-1-like protein 2', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30367066, end_position_on_the_genomic_accession=30462669, orientation='minus', exon_count=50, OMIM='', no_hdr0=''),
  83766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83766, CurrentID=0, Status='live', Symbol='Actl6b', Aliases='Actl6, ArpNa, Baf53b', description='actin-like 6B', other_designations='53 kDa BRG1-associated factor B|BRG1-associated factor 53B|actin-like 6|actin-related protein Baf53b|arpN-alpha', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137553525, end_position_on_the_genomic_accession=137569582, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  83767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83767, CurrentID=0, Status='live', Symbol='Wasf1', Aliases='AI195380, AI838537, Scar, WAVE, WAVE-1', description='WAS protein family, member 1', other_designations='WASP family 1|WASP family protein member 1|protein WAVE-1', map_location='10 22.07 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40883534, end_position_on_the_genomic_accession=40938569, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  83768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83768, CurrentID=0, Status='live', Symbol='Dpp7', Aliases='DPPII, Dpp2, QPP', description='dipeptidylpeptidase 7', other_designations='DPP II|dipeptidyl aminopeptidase II|dipeptidyl peptidase II|dipeptidyl-peptidase 2|dipeptidyl-peptidase II|quiescent cell proline dipeptidase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25352290, end_position_on_the_genomic_accession=25357313, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  83770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83770, CurrentID=0, Status='live', Symbol='Tas1r2', Aliases='Gpr71, T1r2, TR2', description='taste receptor, type 1, member 2', other_designations='G protein-coupled receptor 71|G-protein coupled receptor 71|candidate taste receptor T1R2|sweet taste receptor T1R2', map_location='4 70.82 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139653538, end_position_on_the_genomic_accession=139670280, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  83771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83771, CurrentID=0, Status='live', Symbol='Tas1r3', Aliases='Sac, T1r3', description='taste receptor, type 1, member 3', other_designations='saccharin preference protein|sweet taste receptor T1R3', map_location='4 87.65 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155859268, end_position_on_the_genomic_accession=155863353, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  83772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83772, CurrentID=0, Status='live', Symbol='H2-D', Aliases='', description='histocompatibility 2, D region', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  214854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214854, CurrentID=0, Status='live', Symbol='Neurl3', Aliases='2010300P06Rik, Lincr', description='neuralized E3 ubiquitin protein ligase 3', other_designations='E3 ubiquitin-protein ligase LINCR|lung inducible neuralized-related C3HC4 RING finger|lung-inducible neuralized-related C3CH4 RING domain protein|lung-inducible neuralized-related C3CH4 RING domain-containing protein|lung-inducible neuralized-related C3HC4 RING domain protein|neuralized homolog 3 homolog|neuralized-like protein 3', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36264602, end_position_on_the_genomic_accession=36274674, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  214855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214855, CurrentID=0, Status='live', Symbol='Arid5a', Aliases='D430024K22Rik, Mrf1', description='AT rich interactive domain 5A (MRF1-like)', other_designations='ARID domain-containing protein 5A|AT rich interactive domain 5A (Mrf1 like)|modulator recognition factor I', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36307706, end_position_on_the_genomic_accession=36324036, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  83796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83796, CurrentID=0, Status='live', Symbol='Smarcd2', Aliases='AW322457, Baf60b', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily d, member 2', other_designations='60 kDa BRG-1/Brm-associated factor subunit B|BRG1-associated factor 60B', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106263171, end_position_on_the_genomic_accession=106272972, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  83797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83797, CurrentID=0, Status='live', Symbol='Smarcd1', Aliases='AA407987, Baf60a, D15Kz1', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily d, member 1', other_designations='60 kDa BRG-1/Brm-associated factor subunit A|BRG1-associated factor 60A|SWI/SNF complex 60 kDa subunit', map_location='15 56.13 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99702287, end_position_on_the_genomic_accession=99713995, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  83813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83813, CurrentID=0, Status='live', Symbol='Tnk1', Aliases='Kos1a, Tnk1b, Tnk1', description='tyrosine kinase, non-receptor, 1', other_designations='kinase of embryonic stem cells|non-receptor tyrosine kinase|tnk1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69850563, end_position_on_the_genomic_accession=69858730, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  83814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83814, CurrentID=0, Status='live', Symbol='Nedd4l', Aliases='1300012C07Rik, Nedd4-2, Nedd4b', description='neural precursor cell expressed, developmentally down-regulated gene 4-like', other_designations='NEDD4.2|neural precursor cell expressed, developmentally down-regulated gene 4b', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=64887756, end_position_on_the_genomic_accession=65217826, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  83815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83815, CurrentID=0, Status='live', Symbol='Cenpq', Aliases='2610528M18Rik', description='centromere protein Q', other_designations='CENP-Q', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40923055, end_position_on_the_genomic_accession=40934551, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  214895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214895, CurrentID=0, Status='live', Symbol='Lman2l', Aliases='A630028F14Rik, VIP36-like, VIPL', description='lectin, mannose-binding 2-like', other_designations='LMAN2-like protein', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36423183, end_position_on_the_genomic_accession=36445271, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  214897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214897, CurrentID=0, Status='live', Symbol='Csnk1g1', Aliases='9130020E21Rik', description='casein kinase 1, gamma 1', other_designations='CKI-gamma 1|casein kinase I, gamma 1', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65908962, end_position_on_the_genomic_accession=66045015, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  214899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214899, CurrentID=0, Status='live', Symbol='Kdm5a', Aliases='AA409370, C76986, Jarid1a, RBP2, Rbbp2', description='lysine (K)-specific demethylase 5A', other_designations='RBBP-2|histone demethylase JARID1A|jumonji, AT rich interactive domain 1A (Rbp2 like)|jumonji/ARID domain-containing protein 1A|retinoblastoma binding protein 2|retinoblastoma-binding protein 2', map_location='6 56.95 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120364099, end_position_on_the_genomic_accession=120444574, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  214901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214901, CurrentID=0, Status='live', Symbol='Chtf18', Aliases='6030457M03Rik, CTF18, Chl12', description='CTF18, chromosome transmission fidelity factor 18', other_designations='CTF18, chromosome transmission fidelity factor 18 homolog|Chl12 homolog', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25719030, end_position_on_the_genomic_accession=25727440, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  214917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214917, CurrentID=0, Status='live', Symbol='Fam173a', Aliases='', description='family with sequence similarity 173, member A', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25790080, end_position_on_the_genomic_accession=25794027, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100616072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100616072, CurrentID=0, Status='live', Symbol='Snord116', Aliases='MBII-85, Pwcr1', description='small nucleolar RNA, C/D box 116 cluster', other_designations='Prader-Willi chromosome region 1 homolog|snoRNA MBII-85', map_location='7', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  214922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214922, CurrentID=0, Status='live', Symbol='Slc39a2', Aliases='F730005G13Rik, Gm1789, Gm289, Zip2', description='solute carrier family 39 (zinc transporter), member 2', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51891972, end_position_on_the_genomic_accession=51896745, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  214931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214931, CurrentID=0, Status='live', Symbol='Fbxl16', Aliases='BC042620, Scirr1', description='F-box and leucine-rich repeat protein 16', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25809050, end_position_on_the_genomic_accession=25821265, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  214932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214932, CurrentID=0, Status='live', Symbol='Cecr5', Aliases='A930002G03Rik', description='cat eye syndrome chromosome region, candidate 5', other_designations='cat eye syndrome chromosome region, candidate 5 homolog', map_location='6 56.95 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120509494, end_position_on_the_genomic_accession=120531364, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  232430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232430, CurrentID=0, Status='live', Symbol='Crebl2', Aliases='AI046348, B230205M03', description='cAMP responsive element binding protein-like 2', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134828686, end_position_on_the_genomic_accession=134858931, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101359, CurrentID=0, Status='live', Symbol='Prrt4', Aliases='AW455481, D330027H18Rik', description='proline-rich transmembrane protein 4', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29169230, end_position_on_the_genomic_accession=29179584, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  214944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214944, CurrentID=0, Status='live', Symbol='Mob3b', Aliases='8430436F23Rik, A430018A01Rik, Mobkl2b', description='MOB kinase activator 3B', other_designations='MOB1 (Mps One Binder) kinase activator-like 2B|MOB1, Mps One Binder kinase activator-like 2B|mob1 homolog 2b|mps one binder kinase activator-like 2B', map_location='4 16.86 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34949074, end_position_on_the_genomic_accession=35157495, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  214951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214951, CurrentID=0, Status='live', Symbol='Rhbdl1', Aliases='Rhbdl', description='rhomboid, veinlet-like 1 (Drosophila)', other_designations='RRP|rhomboid-like protein 1', map_location='17 12.93 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25834462, end_position_on_the_genomic_accession=25838353, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  214952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214952, CurrentID=0, Status='live', Symbol='Rhot2', Aliases='Arht2, Miro2', description='ras homolog gene family, member T2', other_designations='MIRO-2|mitochondrial Rho 2', map_location='17 12.93 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25838838, end_position_on_the_genomic_accession=25844851, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  83885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83885, CurrentID=0, Status='live', Symbol='Slc25a2', Aliases='Ornt2', description='solute carrier family 25 (mitochondrial carrier, ornithine transporter) member 2', other_designations='mitochondrial ornithine transporter|ornithine transporter 2', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37637378, end_position_on_the_genomic_accession=37638723, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  214968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214968, CurrentID=0, Status='live', Symbol='Sema6d', Aliases='1110067B02Rik, AA409156, D330011G23, mKIAA1479', description='sema domain, transmembrane domain (TM), and cytoplasmic domain, (semaphorin) 6D', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=124089969, end_position_on_the_genomic_accession=124667789, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  214987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214987, CurrentID=0, Status='live', Symbol='Chtf8', Aliases='5830457O10Rik, Ctf8, Derpc', description='CTF8, chromosome transmission fidelity factor 8', other_designations='CTF8, chromosome transmission fidelity factor 8 homolog|chromosome transmission fidelity factor 8|protein DERPC', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=106883863, end_position_on_the_genomic_accession=106893593, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  83921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83921, CurrentID=0, Status='live', Symbol='Tmem2', Aliases='', description='transmembrane protein 2', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=21777508, end_position_on_the_genomic_accession=21858437, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  83922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83922, CurrentID=0, Status='live', Symbol='Cep41', Aliases='1700017E11Rik, 2810431D15Rik, Tsga14', description='centrosomal protein 41', other_designations='testis specific gene A14', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30653457, end_position_on_the_genomic_accession=30710514, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  83924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83924, CurrentID=0, Status='live', Symbol='Gpr137b', Aliases='AI788917, AU016435, AW546472, C730041E01, C80550, C80741, D13Abb1e, Tm7sf1', description='G protein-coupled receptor 137B', other_designations='transmembrane 7 superfamily member 1 protein', map_location='13 5.24 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=13357620, end_position_on_the_genomic_accession=13394236, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  83925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83925, CurrentID=0, Status='live', Symbol='Trps1', Aliases='AI115454, AI447310, D15Ertd586e', description='trichorhinophalangeal syndrome I (human)', other_designations='', map_location='15 19.18 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=50654757, end_position_on_the_genomic_accession=50890747, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  215001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215001, CurrentID=0, Status='live', Symbol='Wfikkn1', Aliases='Gasp2, Wfikkn', description='WAP, FS, Ig, KU, and NTR-containing protein 1', other_designations='GASP-2|growth and differentiation factor-associated serum protein 2|mGASP-2', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25877628, end_position_on_the_genomic_accession=25880858, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  215008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215008, CurrentID=0, Status='live', Symbol='Vezt', Aliases='A630071D13Rik, AI848282', description='vezatin, adherens junctions transmembrane protein', other_designations='', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93961522, end_position_on_the_genomic_accession=94035799, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  215015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215015, CurrentID=0, Status='live', Symbol='Fam20b', Aliases='C530043G21Rik, mKIAA0475', description='family with sequence similarity 20, member B', other_designations='xylose kinase', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156678545, end_position_on_the_genomic_accession=156719027, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  83945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83945, CurrentID=0, Status='live', Symbol='Dnaja3', Aliases='1200003J13Rik, 1810053A11Rik, C81173, Tid-1, Tid1l', description='DnaJ heat shock protein family (Hsp40) member A3', other_designations='DnaJ (Hsp40) homolog, subfamily A, member 3|dnaJ protein Tid-1|mTid-1|tumorous imaginal discs protein Tid56 homolog', map_location='16 2.46 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4684070, end_position_on_the_genomic_accession=4707695, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  83946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83946, CurrentID=0, Status='live', Symbol='Phip', Aliases='2810004D21Rik, 4632404O06Rik, Ndrp, Wdr11', description='pleckstrin homology domain interacting protein', other_designations='IRS-1 PH domain-binding protein|WD repeat domain 11|WD repeat-containing protein 11|neuronal differentiation related protein|neuronal differentiation-related protein', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=82866159, end_position_on_the_genomic_accession=82975700, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  215028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215028, CurrentID=0, Status='live', Symbol='Prl3d2', Aliases='Ghd5, PL-Ib, Plib', description='prolactin family 3, subfamily d, member 1', other_designations='PL-1 beta|growth hormone d5|placental lactogen-I beta', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27121633, end_position_on_the_genomic_accession=27127555, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  215029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215029, CurrentID=0, Status='live', Symbol='Prl3d3', Aliases='PL-Ig, Plig', description='prolactin family 3, subfamily d, member 3', other_designations='PL-1 gamma|placental lactogen I gamma', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=27156770, end_position_on_the_genomic_accession=27162593, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  215031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215031, CurrentID=0, Status='live', Symbol='Vgll2', Aliases='C130057C21Rik, VITO-1, Vito1, vgl-2', description='vestigial like 2 homolog (Drosophila)', other_designations='', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52022502, end_position_on_the_genomic_accession=52028471, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  83961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83961, CurrentID=0, Status='live', Symbol='Nrg4', Aliases='AI552600', description='neuregulin 4', other_designations='pro-NRG4', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55220222, end_position_on_the_genomic_accession=55283625, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  83962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83962, CurrentID=0, Status='live', Symbol='Btbd1', Aliases='1190005H08Rik', description='BTB (POZ) domain containing 1', other_designations='glucose signal-repressing protein', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81792074, end_position_on_the_genomic_accession=81829503, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  83964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83964, CurrentID=0, Status='live', Symbol='Jam3', Aliases='1110002N23Rik, JAM-3, JAM-C, Jcam3', description='junction adhesion molecule 3', other_designations='junction cell adhesion molecule 3|junction cell adhesion molecule C|junctional adhesion molecule 3', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=27097384, end_position_on_the_genomic_accession=27155421, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  83965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83965, CurrentID=0, Status='live', Symbol='Enpp5', Aliases='AW061004, D17Abb1e, E-NPP 5, mKIAA4152', description='ectonucleotide pyrophosphatase/phosphodiesterase 5', other_designations='E-NPP5|NPP-5|enpp5', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=44078796, end_position_on_the_genomic_accession=44086567, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  215051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215051, CurrentID=0, Status='live', Symbol='Bud13', Aliases='AV305342, D030060M11Rik', description='BUD13 homolog', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=46283012, end_position_on_the_genomic_accession=46298783, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  83984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83984, CurrentID=0, Status='live', Symbol='Tssk6', Aliases='Sstk, TSK-6, TSSK-6', description='testis-specific serine kinase 6', other_designations='serine/threonine protein kinase SSTK|serine/threonine-protein kinase SSTK|small serine/threonine kinase|testis-specific kinase 6', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69902215, end_position_on_the_genomic_accession=69903518, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  215061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215061, CurrentID=0, Status='live', Symbol='Trim50', Aliases='', description='tripartite motif-containing 50', other_designations='tripartite motif protein 50|tripartite motif-containing protein 50', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135351872, end_position_on_the_genomic_accession=135367903, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  83993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83993, CurrentID=0, Status='live', Symbol='Tbx19', Aliases='AU019250, D1Ertd754e, Tpit', description='T-box 19', other_designations='T-box factor, pituitary|T-box protein 19', map_location='1 72.47 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165137834, end_position_on_the_genomic_accession=165160773, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  83995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83995, CurrentID=0, Status='live', Symbol='Mmp1a', Aliases='MMP-1a, Mcol-A, Mcola', description='matrix metallopeptidase 1a (interstitial collagenase)', other_designations='collagenase-like A|matrix metalloproteinase 1a (interstitial collagenase)', map_location='9 A1-A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7464141, end_position_on_the_genomic_accession=7476869, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  83996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83996, CurrentID=0, Status='live', Symbol='Mmp1b', Aliases='MMP-1b, Mcol-B, Mcolb', description='matrix metallopeptidase 1b (interstitial collagenase)', other_designations='collagenase-like B|matrix metalloproteinase 1b (interstitial collagenase)', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7367640, end_position_on_the_genomic_accession=7388026, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  83997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83997, CurrentID=0, Status='live', Symbol='Slmap', Aliases='D330001L02Rik, Slap, mKIAA1601', description='sarcolemma associated protein', other_designations='sarcolemmal-associated protein', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26413168, end_position_on_the_genomic_accession=26534624, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  84004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84004, CurrentID=0, Status='live', Symbol='Mcam', Aliases='1-gicerin, AV025631, CD146, CD149, Muc18, s-endo, s-gicerin', description='melanoma cell adhesion molecule', other_designations='gicerin|l-gicerin protein|melanoma adhesion molecule|melanoma-associated antigen MUC18', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44134584, end_position_on_the_genomic_accession=44142727, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  215085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215085, CurrentID=0, Status='live', Symbol='Slc35f1', Aliases='Duf6, E130016L02', description='solute carrier family 35, member F1', other_designations='DUF6 containing protein', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52690501, end_position_on_the_genomic_accession=53111622, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  215090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215090, CurrentID=0, Status='live', Symbol='Maneal', Aliases='Gm50', description='mannosidase, endo-alpha-like', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=124855239, end_position_on_the_genomic_accession=124880938, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  215095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215095, CurrentID=0, Status='live', Symbol='Astl', Aliases='7420700F21, C87576, Sas1b', description='astacin-like metalloendopeptidase (M12 family)', other_designations='astl|oocyte astacin|ovastacin|sperm acrosomal SLLP1|sperm acrosomal SLLP1-binding protein', map_location='2 F', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127338639, end_position_on_the_genomic_accession=127357655, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84035, CurrentID=0, Status='live', Symbol='Kremen1', Aliases='AV002070, Kremen, Krm1', description='kringle containing transmembrane protein 1', other_designations='dickkopf receptor|kringle domain-containing transmembrane protein 1|kringle-coding gene marking the eye and the nose|kringle-containing protein marking the eye and the nose', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5191551, end_position_on_the_genomic_accession=5261636, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  84036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84036, CurrentID=0, Status='live', Symbol='Kcnn1', Aliases='AI854248, KCa2.1, SK1', description='potassium intermediate/small conductance calcium-activated channel, subfamily N, member 1', other_designations='SKCa 1|SKCa1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70842049, end_position_on_the_genomic_accession=70865783, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  215113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215113, CurrentID=0, Status='live', Symbol='Slc43a2', Aliases='7630402D21Rik, BC042513, Lat4', description='solute carrier family 43, member 2', other_designations='L-type amino L-acid transporter 4|L-type amino acid transporter 4', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75530282, end_position_on_the_genomic_accession=75577575, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  215114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215114, CurrentID=0, Status='live', Symbol='Hip1', Aliases='2610109B09Rik, A930014B11Rik, E130315I21Rik, mKIAA4113', description='huntingtin interacting protein 1', other_designations='HIP-I|hip1|huntingtin-interacting protein I', map_location='5 F-G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135406518, end_position_on_the_genomic_accession=135545319, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  215160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215160, CurrentID=0, Status='live', Symbol='Rhbdd2', Aliases='0610011L13Rik, AI480532, Rhbdl7, Usmg1', description='rhomboid domain containing 2', other_designations='rhomboid, veinlet-like 7|rhomboid-like protein 7|upregulated during skeletal muscle growth 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135632654, end_position_on_the_genomic_accession=135646376, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  84092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84092, CurrentID=0, Status='live', Symbol='Usp8', Aliases='AI574262, AW557536, Ubpy, mKIAA0055', description='ubiquitin specific peptidase 8', other_designations='deubiquitinating enzyme 8|putative deubiquitinating enzyme|ubiquitin isopeptidase Y|ubiquitin thioesterase 8|ubiquitin thiolesterase 8|ubiquitin-specific-processing protease 8', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126707202, end_position_on_the_genomic_accession=126759314, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  84094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84094, CurrentID=0, Status='live', Symbol='Plvap', Aliases='MECA32, Pv1', description='plasmalemma vesicle associated protein', other_designations='MECA-32 antigen|plasmalemma vesicle protein 1', map_location='8 34.43 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71497753, end_position_on_the_genomic_accession=71511804, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84095, CurrentID=0, Status='live', Symbol='Pi4k2a', Aliases='Pi4k2', description='phosphatidylinositol 4-kinase type 2 alpha', other_designations='phosphatidylinositol 4-kinase type II-alpha', map_location='19 35.74 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42090146, end_position_on_the_genomic_accession=42122218, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  84111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84111, CurrentID=0, Status='live', Symbol='Gpr87', Aliases='', description='G protein-coupled receptor 87', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=59178904, end_position_on_the_genomic_accession=59195468, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  84112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84112, CurrentID=0, Status='live', Symbol='Sucnr1', Aliases='Gpr91', description='succinate receptor 1', other_designations='G protein-coupled receptor 91|G-protein coupled receptor 91', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=60081869, end_position_on_the_genomic_accession=60087566, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  84113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84113, CurrentID=0, Status='live', Symbol='Ptov1', Aliases='1110030G05Rik, AU041779, Gcap3', description='prostate tumor over expressed gene 1', other_designations='granule cell antiserum positive 3', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44863067, end_position_on_the_genomic_accession=44870411, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  215193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215193, CurrentID=0, Status='live', Symbol='Diexf', Aliases='AA408296, AW045375, Def', description='digestive organ expansion factor homolog (zebrafish)', other_designations='digestive-organ expansion factor homolog', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193104403, end_position_on_the_genomic_accession=193130285, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  215194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215194, CurrentID=0, Status='live', Symbol='Kri1', Aliases='', description='KRI1 homolog', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21273457, end_position_on_the_genomic_accession=21287969, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  215201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215201, CurrentID=0, Status='live', Symbol='Trmt2b', Aliases='4732479N06Rik', description='TRM2 tRNA methyltransferase 2B', other_designations='TRM2 homolog|TRM2 tRNA methyltransferase 2 homolog B|tRNA (uracil-5-)-methyltransferase homolog', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134222955, end_position_on_the_genomic_accession=134277039, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  215210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215210, CurrentID=0, Status='live', Symbol='Tmem120a', Aliases='2010310D06Rik, Tmpit', description='transmembrane protein 120A', other_designations='transmembrane protein induced by tumor necrosis factor alpha', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135735490, end_position_on_the_genomic_accession=135744172, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  14027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14027, CurrentID=0, Status='live', Symbol='Evpl', Aliases='', description='envoplakin', other_designations='210 kDa cornified envelope precursor protein|210-kD protein|210kDa protein|p210', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116220559, end_position_on_the_genomic_accession=116238147, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  215243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215243, CurrentID=0, Status='live', Symbol='Traf3ip3', Aliases='6030423D04Rik, AI115021, T3jam', description='TRAF3 interacting protein 3', other_designations='TRAF3-interacting Jun N-terminal kinase (JNK)-activating modulator|TRAF3-interacting protein 3', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193175504, end_position_on_the_genomic_accession=193201546, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  215257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215257, CurrentID=0, Status='live', Symbol='Il1f9', Aliases='Il36g', description='interleukin 1 family, member 9', other_designations='IL-1F9|interleukin-1 family member 9', map_location='2 16.24 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24186476, end_position_on_the_genomic_accession=24193567, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  215274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215274, CurrentID=0, Status='live', Symbol='Il1f10', Aliases='', description='interleukin 1 family, member 10', other_designations='IL-1F10|interleukin 1 family, member 7', map_location='2 16.32 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24287095, end_position_on_the_genomic_accession=24293820, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  215280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215280, CurrentID=0, Status='live', Symbol='Wipf1', Aliases='AI115543, D2Ertd120e, WIP, Waspip', description='WAS/WASL interacting protein family, member 1', other_designations='WASP interacting protein|Wiskott-Aldrich syndrome protein interacting protein', map_location='2 43.68 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73429610, end_position_on_the_genomic_accession=73529733, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  215303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215303, CurrentID=0, Status='live', Symbol='Camk1g', Aliases='CLICK-III, CaMKIgamma', description='calcium/calmodulin-dependent protein kinase I gamma', other_designations='CLICK III|caM kinase I gamma|caM kinase IG|caM-KI gamma|caMK-like CREB kinase III|caMKI gamma|caMKIG', map_location='1 97.74 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193346346, end_position_on_the_genomic_accession=193370292, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  215332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215332, CurrentID=0, Status='live', Symbol='Slc36a3', Aliases='PAT3, TRAMD2, tramdorin2', description='solute carrier family 36 (proton/amino acid symporter), member 3', other_designations='proton/amino acid transporter 3|solute carrier family 36 member 3|tramdorin 2|tramdorin-2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55124815, end_position_on_the_genomic_accession=55151745, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  215335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215335, CurrentID=0, Status='live', Symbol='Slc36a1', Aliases='5830411H19Rik, AI839897, Pat1', description='solute carrier family 36 (proton/amino acid symporter), member 1', other_designations='proton/amino acid transporter 1|solute carrier family 36 member 1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55204321, end_position_on_the_genomic_accession=55236330, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  215351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215351, CurrentID=0, Status='live', Symbol='Senp6', Aliases='2810017C20Rik, E130319N12Rik, Susp1, mKIAA0797', description='SUMO/sentrin specific peptidase 6', other_designations='SUMO-1-specific protease 1|SUMO/sentrin specific protease 6|sentrin/SUMO-specific protease SENP6', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=80066516, end_position_on_the_genomic_accession=80144782, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  57741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57741, CurrentID=0, Status='live', Symbol='Noc2l', Aliases='AA410003, AF155546, NIR', description='nucleolar complex associated 2 homolog (S. cerevisiae)', other_designations='NOC2-like protein|novel INHAT repressor|protein NOC2 homolog', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156235823, end_position_on_the_genomic_accession=156247616, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  215378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215378, CurrentID=0, Status='live', Symbol='Brinp3', Aliases='B830045N13Rik, Fam5c', description='bone morphogenetic protein/retinoic acid inducible neural specific 3', other_designations='family with sequence similarity 5, member C|protein FAM5C', map_location='1 63.24 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=146495666, end_position_on_the_genomic_accession=146902474, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  215384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215384, CurrentID=0, Status='live', Symbol='Fcgbp', Aliases='A430096B05Rik', description='Fc fragment of IgG binding protein', other_designations='IgG Fc binding protein', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28071236, end_position_on_the_genomic_accession=28120864, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  215387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215387, CurrentID=0, Status='live', Symbol='Ncaph', Aliases='A730011O11Rik, Brrn1, HCAP-H, mKIAA0074', description='non-SMC condensin I complex, subunit H', other_designations='XCAP-H homolog|barren homolog protein 1|chromosome-associated protein H|mCAP-H', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127103809, end_position_on_the_genomic_accession=127133954, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  215418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215418, CurrentID=0, Status='live', Symbol='Csrnp1', Aliases='4931429D10Rik, Axud1, CSRNP-1, URAX1, taip-3', description='cysteine-serine-rich nuclear protein 1', other_designations='AXIN1 up-regulated 1|TGF-beta induced apoptosis protein 3|TGF-beta-induced apoptosis protein 3|axin-1 up-regulated gene 1 protein', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119971163, end_position_on_the_genomic_accession=119985134, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  215436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215436, CurrentID=0, Status='live', Symbol='Slc35e3', Aliases='9330166G04Rik, AI481200', description='solute carrier family 35, member E3', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117733678, end_position_on_the_genomic_accession=117746358, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  215445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215445, CurrentID=0, Status='live', Symbol='Rab11fip3', Aliases='Cart1, D030060O14Rik, Rab11-FIP3, mKIAA0665', description='RAB11 family interacting protein 3 (class II)', other_designations='FIP3-Rab11|cytoplasmic adaptor for RAR and TR', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25989036, end_position_on_the_genomic_accession=26069550, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  215446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215446, CurrentID=0, Status='live', Symbol='Entpd3', Aliases='Cd39l3, HB6, NTPDase-3', description='ectonucleoside triphosphate diphosphohydrolase 3', other_designations='CD39 antigen-like 3', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120539818, end_position_on_the_genomic_accession=120568327, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  215449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215449, CurrentID=0, Status='live', Symbol='Rap1b', Aliases='2810443E11Rik', description='RAS related protein 1b', other_designations='GTP-binding protein smg p21B', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117814597, end_position_on_the_genomic_accession=117845974, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  215456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215456, CurrentID=0, Status='live', Symbol='Gpat2', Aliases='A530057A03Rik, Gm116, xGPAT1', description='glycerol-3-phosphate acyltransferase 2, mitochondrial', other_designations='GPAT-2|glycerol-3-phosphate acyltransferase-like 1', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127410628, end_position_on_the_genomic_accession=127436093, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  319899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319899, CurrentID=0, Status='live', Symbol='Dock6', Aliases='2410095B20Rik, 4931431C02Rik, C330023D02Rik, mKIAA1395', description='dedicator of cytokinesis 6', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21800180, end_position_on_the_genomic_accession=21852719, orientation='minus', exon_count=51, OMIM='', no_hdr0=''),
  57757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57757, CurrentID=0, Status='live', Symbol='Pglyrp2', Aliases='C730002N09Rik, PGRP-L, Pglyrpl, tagL', description='peptidoglycan recognition protein 2', other_designations='peptidoglycan recognition protein long', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32412459, end_position_on_the_genomic_accession=32424183, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  215474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215474, CurrentID=0, Status='live', Symbol='Sec22c', Aliases='4932412K21, 5930407I15Rik, C530046H07, Sec22l3', description='SEC22 homolog C, vesicle trafficking protein', other_designations='SEC22 vesicle trafficking protein homolog C|SEC22 vesicle trafficking protein-like 3|SEC22 vesicle trafficking protein-like C|SEC22 vesicle-trafficking protein homolog C', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121680045, end_position_on_the_genomic_accession=121706736, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  215476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215476, CurrentID=0, Status='live', Symbol='Prr14l', Aliases='6030436E02Rik, C330019G07Rik, Prl14l', description='proline rich 14-like', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=32789206, end_position_on_the_genomic_accession=32854338, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  215493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215493, CurrentID=0, Status='live', Symbol='A3galt2', Aliases='Gm433, iGb3, iGb3S', description='alpha 1,3-galactosyltransferase 2 (isoglobotriaosylceramide synthase)', other_designations='iGb3 synthase', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128759258, end_position_on_the_genomic_accession=128770201, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  215494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215494, CurrentID=0, Status='live', Symbol='Pomgnt2', Aliases='Ago61, C85492, Gtdc2', description='protein O-linked mannose beta 1,4-N-acetylglucosaminyltransferase 2', other_designations='glycosyltransferase|glycosyltransferase-like domain containing 2|glycosyltransferase-like domain-containing protein 2', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121981606, end_position_on_the_genomic_accession=121996026, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  215512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215512, CurrentID=0, Status='live', Symbol='Fam117a', Aliases='5730593F17Rik, AW548096', description='family with sequence similarity 117, member A', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95337018, end_position_on_the_genomic_accession=95381872, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  57765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57765, CurrentID=0, Status='live', Symbol='Tbx21', Aliases='TBT1, Tbet, Tblym', description='T-box 21', other_designations='T-box protein 21|T-cell-specific T-box transcription factor T-bet|transcription factor TBLYM', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97098007, end_position_on_the_genomic_accession=97115331, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  84505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84505, CurrentID=0, Status='live', Symbol='Setdb1', Aliases='AU022152, ESET, KMT1E, mKIAA0067', description='SET domain, bifurcated 1', other_designations='ERG-associated protein with SET domain|SET domain ERG-associated histone methyltransferase', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95323525, end_position_on_the_genomic_accession=95357202, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  84506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84506, CurrentID=0, Status='live', Symbol='Hamp', Aliases='Hamp1, Hepc, Hepc1', description='hepcidin antimicrobial peptide', other_designations='hepcidin antimicrobial peptide 1', map_location='7 19.27 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30942368, end_position_on_the_genomic_accession=30944017, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  57778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57778, CurrentID=0, Status='live', Symbol='Fmnl1', Aliases='8030453N10Rik, AI553564, Fmnl, Fnrl, Frls', description='formin-like 1', other_designations='formin-related gene in leukocytes|formin-related protein|lymphocyte specific formin related protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103170575, end_position_on_the_genomic_accession=103198902, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  319924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319924, CurrentID=0, Status='live', Symbol='Apba1', Aliases='6430513E09Rik, Lin-10, Mint, Mint1, Urop11, X11, X11alpha', description='amyloid beta (A4) precursor protein binding, family A, member 1', other_designations='UROP11-110|adapter protein X11alpha|mint-1|neuron-specific X11 protein|neuronal Munc18-1-interacting protein 1', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=23758876, end_position_on_the_genomic_accession=23949598, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  84543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84543, CurrentID=0, Status='live', Symbol='Sval2', Aliases='SLP-M', description='seminal vesicle antigen-like 2', other_designations='SLP in the mammary gland|SVA-like protein in the mammary gland', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41860339, end_position_on_the_genomic_accession=41864322, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  84544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84544, CurrentID=0, Status='live', Symbol='Cd96', Aliases='1700109I12Rik, Tactile', description='CD96 antigen', other_designations='T cell activation, increased late expression|cell surface antigen CD96|t cell-activated increased late expression protein', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=46035657, end_position_on_the_genomic_accession=46120248, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  215627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215627, CurrentID=0, Status='live', Symbol='Zbtb8b', Aliases='AV220464, BC023839, BOZF1, Zbtb8', description='zinc finger and BTB domain containing 8b', other_designations='zinc finger and BTB domain containing 8', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129425765, end_position_on_the_genomic_accession=129446972, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  215632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215632, CurrentID=0, Status='live', Symbol='Psd4', Aliases='BC046518, EFA6B', description='pleckstrin and Sec7 domain containing 4', other_designations='SEC7 homolog|pleckstrin homology and SEC7 domain-containing protein 4', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24367561, end_position_on_the_genomic_accession=24411223, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  215641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215641, CurrentID=0, Status='live', Symbol='Mageb18', Aliases='B430216B18Rik', description='melanoma antigen family B, 18', other_designations='', map_location='X C2-C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=92118879, end_position_on_the_genomic_accession=92599572, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  215653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215653, CurrentID=0, Status='live', Symbol='Rassf2', Aliases='3830431H01Rik, 9030412M04Rik, AI852669, AW495050', description='Ras association (RalGDS/AF-6) domain family member 2', other_designations='Ras association (RalGDS/AF-6) domain family 2', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131992850, end_position_on_the_genomic_accession=132030287, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  215654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215654, CurrentID=0, Status='live', Symbol='Cdh12', Aliases='Cdhb', description='cadherin 12', other_designations='Br-cadherin', map_location='15 8.25 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=20448879, end_position_on_the_genomic_accession=21597005, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  84585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84585, CurrentID=0, Status='live', Symbol='Rnf123', Aliases='BC003945, Kpc1', description='ring finger protein 123', other_designations='kip1 ubiquitination-promoting complex protein 1|ubiquitin ligase', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108051672, end_position_on_the_genomic_accession=108079773, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  215690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215690, CurrentID=0, Status='live', Symbol='Nav1', Aliases='9530089B19, 9930003A20Rik, C230080M11Rik, Pomfil3, Unc53h1, mKIAA1151, steerin-1', description='neuron navigator 1', other_designations='Steerin1|pore membrane and/or filament-interacting-like protein 3', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135434580, end_position_on_the_genomic_accession=135688116, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  215693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215693, CurrentID=0, Status='live', Symbol='Zmat1', Aliases='A230096A06, AW990744, B930008K04Rik', description='zinc finger, matrin type 1', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134971372, end_position_on_the_genomic_accession=135009209, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  215705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215705, CurrentID=0, Status='live', Symbol='Arrdc1', Aliases='AI957342, BC004091', description='arrestin domain containing 1', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24925352, end_position_on_the_genomic_accession=24935335, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  215707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215707, CurrentID=0, Status='live', Symbol='Ccdc92', Aliases='AW050178, D5Bwg0834e', description='coiled-coil domain containing 92', other_designations='', map_location='5 63.85 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124834432, end_position_on_the_genomic_accession=124862221, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  215708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215708, CurrentID=0, Status='live', Symbol='Fam73a', Aliases='A930038D06, AI852168, C030011O14Rik', description='family with sequence similarity 73, member A', other_designations='', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152273459, end_position_on_the_genomic_accession=152340437, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  101488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101488, CurrentID=0, Status='live', Symbol='Slco2b1', Aliases='AI060904, AI852653, OATP-B, Slc21a9', description='solute carrier organic anion transporter family, member 2b1', other_designations='solute carrier family 21 (organic anion transporter), member 9|solute carrier family 21 member 9', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99657804, end_position_on_the_genomic_accession=99711340, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  84652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84652, CurrentID=0, Status='live', Symbol='Fam126a', Aliases='AB030242, Drctnnb1a', description='family with sequence similarity 126, member A', other_designations='down-regulated by CTNNB1 protein A', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=23915276, end_position_on_the_genomic_accession=24030780, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  84653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84653, CurrentID=0, Status='live', Symbol='Hes7', Aliases='bHLHb37', description='hairy and enhancer of split 7 (Drosophila)', other_designations='bHLH factor Hes7|mHes7', map_location='11 42.38 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69118711, end_position_on_the_genomic_accession=69124054, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  215748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215748, CurrentID=0, Status='live', Symbol='Cnksr3', Aliases='6820402C05, BC024086, Magi1', description='Cnksr family member 3', other_designations='CNK homolog protein 3|CNK3|connector enhancer of KSR 3|maguin-like protein|membrane associated guanylate kinase interacting protein-like 1|membrane-associated guanylate kinase-interacting protein-like 1', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7119061, end_position_on_the_genomic_accession=7212237, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  215751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215751, CurrentID=0, Status='live', Symbol='Ginm1', Aliases='AA589616', description='glycoprotein integral membrane 1', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7767947, end_position_on_the_genomic_accession=7780946, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  84682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84682, CurrentID=0, Status='live', Symbol='Cox4i2', Aliases='Cox4b, CoxIV-2', description='cytochrome c oxidase subunit IV isoform 2', other_designations='Cox IV-2|cytochrome c oxidase, subunit IVb', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152754173, end_position_on_the_genomic_accession=152765037, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  215772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215772, CurrentID=0, Status='live', Symbol='Adgb', Aliases='9130014G24Rik', description='androglobin', other_designations='calpain-7-like protein', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=10335703, end_position_on_the_genomic_accession=10475135, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  84704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=84704, CurrentID=0, Status='live', Symbol='Snurf', Aliases='2410045I01Rik, Snrpn', description='SNRPN upstream reading frame', other_designations='small nuclear ribonucleoprotein N', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=59982501, end_position_on_the_genomic_accession=60005156, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  215789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215789, CurrentID=0, Status='live', Symbol='Phactr2', Aliases='AV158170, BC012871', description='phosphatase and actin regulator 2', other_designations='', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13207717, end_position_on_the_genomic_accession=13474521, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  319955 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=319955, CurrentID=0, Status='live', Symbol='Ercc6', Aliases='4732403I04, C130058G22Rik, CSB', description='excision repair cross-complementing rodent repair deficiency, complementation group 6', other_designations='CS group B correcting', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32513288, end_position_on_the_genomic_accession=32580990, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  215798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215798, CurrentID=0, Status='live', Symbol='Adgrg6', Aliases='1190004A11Rik, AI449247, AW045736, DREG, Gm222, Gpr126', description='adhesion G protein-coupled receptor G6', other_designations='G protein-coupled receptor 126|developmentally regulated G-protein-coupled receptor|probable G-protein coupled receptor 126', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=14402580, end_position_on_the_genomic_accession=14545722, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  215814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215814, CurrentID=0, Status='live', Symbol='Ccdc28a', Aliases='1700009P13Rik, AI480677', description='coiled-coil domain containing 28A', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18213685, end_position_on_the_genomic_accession=18234991, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  215819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215819, CurrentID=0, Status='live', Symbol='Nhsl1', Aliases='5730409E15Rik, A630035H13Rik, A730096F01, BC013565, D10Bwg0940e, mKIAA1357', description='NHS-like 1', other_designations='', map_location='10 7.65 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18318558, end_position_on_the_genomic_accession=18533892, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  215821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215821, CurrentID=0, Status='live', Symbol='Arfgef3', Aliases='9630004D11, B930094H20Rik, BIG3, D10Bwg1379e, mKIAA1244', description='ARFGEF family member 3', other_designations='', map_location='10 7.79 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18581839, end_position_on_the_genomic_accession=18744069, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  215854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215854, CurrentID=0, Status='live', Symbol='Taar5', Aliases='Gm227, taR-5', description='trace amine-associated receptor 5', other_designations='mTaar5|trimethylamine receptor', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23970706, end_position_on_the_genomic_accession=23971719, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  215855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215855, CurrentID=0, Status='live', Symbol='Taar6', Aliases='Gm228, taR-6', description='trace amine-associated receptor 6', other_designations='mTaar6', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23984609, end_position_on_the_genomic_accession=23985646, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  215856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215856, CurrentID=0, Status='live', Symbol='Taar7a', Aliases='Gm229, taR-7a', description='trace amine-associated receptor 7A', other_designations='mTaar7a', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23992405, end_position_on_the_genomic_accession=23993481, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  215859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215859, CurrentID=0, Status='live', Symbol='Taar8a', Aliases='Gm230, taR-8a', description='trace amine-associated receptor 8A', other_designations='mTaar8a', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24076500, end_position_on_the_genomic_accession=24077534, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  215890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215890, CurrentID=0, Status='live', Symbol='Clvs2', Aliases='A330019N05Rik, Rlbp1l2', description='clavesin 2', other_designations='retinaldehyde binding protein 1-like 2|retinaldehyde-binding protein 1-like 2|retinaldehyde-binding protein 1-like protein 2', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=33512286, end_position_on_the_genomic_accession=33624905, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  215895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215895, CurrentID=0, Status='live', Symbol='Gm4794', Aliases='EG215895, Sult3a2', description='predicted gene 4794', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=33766424, end_position_on_the_genomic_accession=33786704, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  215900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215900, CurrentID=0, Status='live', Symbol='Fam26f', Aliases='A630077B13Rik', description='family with sequence similarity 26, member F', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34126067, end_position_on_the_genomic_accession=34127972, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  215928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215928, CurrentID=0, Status='live', Symbol='Mfsd4b5', Aliases='Naglt1c', description='major facilitator superfamily domain containing 4B5', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39958061, end_position_on_the_genomic_accession=39986646, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  215929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215929, CurrentID=0, Status='live', Symbol='Mfsd4b1', Aliases='AI317395, AI482555, Naglt1a', description='major facilitator superfamily domain containing 4B1', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=40001573, end_position_on_the_genomic_accession=40025268, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  215951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215951, CurrentID=0, Status='live', Symbol='Lace1', Aliases='CG8520', description='lactation elevated 1', other_designations='', map_location='10 22.88 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=42312581, end_position_on_the_genomic_accession=42478584, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  229782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229782, CurrentID=0, Status='live', Symbol='Slc35a3', Aliases='2310050P13Rik', description='solute carrier family 35 (UDP-N-acetylglucosamine (UDP-GlcNAc) transporter), member 3', other_designations='golgi UDP-GlcNAc transporter|solute carrier family 35 member A3', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116669469, end_position_on_the_genomic_accession=116713960, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  215999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215999, CurrentID=0, Status='live', Symbol='Mcu', Aliases='2010012O16Rik, AV064928, C10orf42, Ccdc109a, D130073L02Rik, Gm64', description='mitochondrial calcium uniporter', other_designations='coiled-coil domain containing 109A', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59446984, end_position_on_the_genomic_accession=59616692, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  216001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216001, CurrentID=0, Status='live', Symbol='Micu1', Aliases='C730016L05Rik, Calc, Cbara1', description='mitochondrial calcium uptake 1', other_designations='calcium binding atopy-related autoantigen 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59702482, end_position_on_the_genomic_accession=59864134, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  216011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216011, CurrentID=0, Status='live', Symbol='Lrrc20', Aliases='BC036304', description='leucine rich repeat containing 20', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61475276, end_position_on_the_genomic_accession=61582791, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  216019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216019, CurrentID=0, Status='live', Symbol='Hkdc1', Aliases='BC016235', description='hexokinase domain containing 1', other_designations='hexokinase domain-containing protein 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62383137, end_position_on_the_genomic_accession=62422465, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  216021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216021, CurrentID=0, Status='live', Symbol='Stox1', Aliases='4732470K04Rik, Gm63', description='storkhead box 1', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62659043, end_position_on_the_genomic_accession=62726099, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  216028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216028, CurrentID=0, Status='live', Symbol='Lrrtm3', Aliases='9630044H04Rik', description='leucine rich repeat transmembrane neuronal 3', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63928497, end_position_on_the_genomic_accession=64090255, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  216033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216033, CurrentID=0, Status='live', Symbol='Ctnna3', Aliases='4930429L08Rik, 4933408A16, Catna3, Vr22', description='catenin (cadherin associated protein), alpha 3', other_designations='alpha T-catenin|alpha-catenin-like protein|cadherin-associated protein|ctnna3', map_location='10 B4-B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63430098, end_position_on_the_genomic_accession=65005395, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  216049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216049, CurrentID=0, Status='live', Symbol='Zfp365', Aliases='AI839779, AV340892, DBZ, Su48, Znf365, talanin', description='zinc finger protein 365', other_designations='', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=67886105, end_position_on_the_genomic_accession=67912662, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  216080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216080, CurrentID=0, Status='live', Symbol='Ube2d1', Aliases='UBCH5', description='ubiquitin-conjugating enzyme E2D 1', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme D1|E2 ubiquitin-conjugating enzyme D1|E2(17)KB 1|ubiquitin carrier protein D1|ubiquitin-conjugating enzyme E2(17)KB 1|ubiquitin-conjugating enzyme E2-17 kDa 1|ubiquitin-conjugating enzyme E2D 1, UBC4/5 homolog|ubiquitin-protein ligase D1', map_location='10 B5.3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=71254980, end_position_on_the_genomic_accession=71285262, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  85029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85029, CurrentID=0, Status='live', Symbol='Rpph1', Aliases='Gm24821, H1RNA, Rmrp5, Rpr', description='ribonuclease P RNA component H1', other_designations='RNA component of nuclear RNAase P, 5|RNase P', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50807447, end_position_on_the_genomic_accession=50807771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  85030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85030, CurrentID=0, Status='live', Symbol='Tnfrsf25', Aliases='APO-3, DDR3, DR3, LARD, TR3, TRAMP, Tnfrsf12, WSL-1, WSL-LR, Wsl', description='tumor necrosis factor receptor superfamily, member 25', other_designations='tumor necrosis factor receptor superfamily, member 12', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152115528, end_position_on_the_genomic_accession=152120119, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  85031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85031, CurrentID=0, Status='live', Symbol='Pla1a', Aliases='AA986889, Ps-pla1, Pspla1', description='phospholipase A1 member A', other_designations='phosphatidylserine-specific phospholipase A1', map_location='16 26.83 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38396117, end_position_on_the_genomic_accession=38433191, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  85034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85034, CurrentID=0, Status='live', Symbol='D2Mit2', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 2', other_designations='', map_location='2 4.24 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85035, CurrentID=0, Status='live', Symbol='D2Mit4', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 4', other_designations='', map_location='2 6.9 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85036, CurrentID=0, Status='live', Symbol='D4Mit2', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 2', other_designations='', map_location='4 10.61 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85038, CurrentID=0, Status='live', Symbol='D2Mit9', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 9', other_designations='', map_location='2 38.54 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85039, CurrentID=0, Status='live', Symbol='D9Mit9', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 9', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85040, CurrentID=0, Status='live', Symbol='DXMit4', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 4', other_designations='', map_location='X 61.35 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85041, CurrentID=0, Status='live', Symbol='D3Mit60', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 60', other_designations='', map_location='3 1.96 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85042, CurrentID=0, Status='live', Symbol='D4Mit18', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 18', other_designations='', map_location='4 6.01 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85043, CurrentID=0, Status='live', Symbol='D4Mit19', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 19', other_designations='', map_location='4 5.58 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216119, CurrentID=0, Status='live', Symbol='Ybey', Aliases='1110068N23Rik, A130042E20Rik, C21orf57, ORF57, ORF66', description='ybeY metallopeptidase', other_designations='putative metalloprotease C21orf57 homolog|rRNA maturation factor homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=76459567, end_position_on_the_genomic_accession=76469114, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  85048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85048, CurrentID=0, Status='live', Symbol='D7Mit77', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 77', other_designations='', map_location='7 15.94 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85049, CurrentID=0, Status='live', Symbol='DXMit1.1', Aliases='DXMit1', description='DNA segment, Chr X, Massachusetts Institute of Technology 1.1', other_designations='', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85052, CurrentID=0, Status='live', Symbol='D13Mit1', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 1', other_designations='', map_location='13 3.25 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85053, CurrentID=0, Status='live', Symbol='D1Mit20', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 20', other_designations='', map_location='1 11.77 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85054, CurrentID=0, Status='live', Symbol='D1Mit22', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 22', other_designations='', map_location='1 30.44 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85056, CurrentID=0, Status='live', Symbol='D1Mit34', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 34', other_designations='', map_location='1 67.5 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85059, CurrentID=0, Status='live', Symbol='D2Mit13', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 13', other_designations='', map_location='2 55.67 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85060, CurrentID=0, Status='live', Symbol='D2Mit79', Aliases='', description='DNA segment, Chr 2, Massachusetts Institute of Technology 79', other_designations='', map_location='2 15.15 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85062, CurrentID=0, Status='live', Symbol='D3Mit51', Aliases='', description='DNA segment, Chr 3, Massachusetts Institute of Technology 51', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85063, CurrentID=0, Status='live', Symbol='D4Mit49', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 49', other_designations='', map_location='4 79.08 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85064, CurrentID=0, Status='live', Symbol='D4Mit55', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 55', other_designations='', map_location='4 24.28 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85065, CurrentID=0, Status='live', Symbol='D4Mit57', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 57', other_designations='', map_location='4 54.27 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85066, CurrentID=0, Status='live', Symbol='D4Mit89', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 89', other_designations='', map_location='4 24.66 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85070, CurrentID=0, Status='live', Symbol='D6Mit74', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 74', other_designations='', map_location='6 23.7 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85071, CurrentID=0, Status='live', Symbol='D6Mit94', Aliases='', description='DNA segment, Chr 6, Massachusetts Institute of Technology 94', other_designations='', map_location='6 29.56 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216144, CurrentID=0, Status='live', Symbol='Vmn2r81', Aliases='EC1-V2R, EC1-VR2, V2rf2', description='vomeronasal 2, receptor 81', other_designations='EC1-V2R pheromone receptor', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79247777, end_position_on_the_genomic_accession=79294535, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  85074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85074, CurrentID=0, Status='live', Symbol='D7Mit37', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 37', other_designations='', map_location='7 54.85 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216148, CurrentID=0, Status='live', Symbol='Shc2', Aliases='6720466E06, SCK, ShcB, Sli', description='SHC (Src homology 2 domain containing) transforming protein 2', other_designations='SH2 domain protein C2|SHC-transforming protein B|protein Sck|protein Sli|src homology 2 domain-containing transforming protein C2|src homology 2 domain-containing-transforming protein C2', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79617934, end_position_on_the_genomic_accession=79637918, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  85077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85077, CurrentID=0, Status='live', Symbol='D7Mit91', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 91', other_designations='', map_location='7 33.88 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216150, CurrentID=0, Status='live', Symbol='Cdc34', Aliases='AI327276, UBE2R1', description='cell division cycle 34', other_designations='(E3-independent) E2 ubiquitin-conjugating enzyme R1|E2 ubiquitin-conjugating enzyme R1|E2-CDC34|cell division cycle 34 homolog|ubiquitin-conjugating enzyme Cdc34|ubiquitin-conjugating enzyme E2-32 kDa complementing|ubiquitin-conjugating enzyme E2-CDC34|ubiquitin-protein ligase R1', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79682195, end_position_on_the_genomic_accession=79688398, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  85079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85079, CurrentID=0, Status='live', Symbol='D9Mit14', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 14', other_designations='', map_location='9 55.65 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85080, CurrentID=0, Status='live', Symbol='D9Mit76', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 76', other_designations='', map_location='9 50.18 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216154, CurrentID=0, Status='live', Symbol='Med16', Aliases='95kDa, A630083L04, Thrap5, Trap95', description='mediator complex subunit 16', other_designations='DRIP92|thyroid hormone receptor associated protein 5|thyroid hormone receptor-associated protein 5|thyroid hormone receptor-associated protein 95|thyroid hormone receptor-associated protein complex 95 kDa component|thyroid hormone receptor-associated protein, 95-kD subunit|vitamin D3 receptor-interacting protein complex 92 kDa component', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79894707, end_position_on_the_genomic_accession=79908941, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  216156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216156, CurrentID=0, Status='live', Symbol='Wdr18', Aliases='2310012I10Rik, AU044733, AW122032', description='WD repeat domain 18', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79960152, end_position_on_the_genomic_accession=79969246, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  216157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216157, CurrentID=0, Status='live', Symbol='Tmem259', Aliases='AC087114.7gm2, ORF61, R32184_3, mFLJ00277', description='transmembrane protein 259', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79977120, end_position_on_the_genomic_accession=79984330, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  85087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85087, CurrentID=0, Status='live', Symbol='DXMit93', Aliases='', description='DNA segment, Chr X, Massachusetts Institute of Technology 93', other_designations='', map_location='X 39.95 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216161, CurrentID=0, Status='live', Symbol='Sbno2', Aliases='BC019206, F730040C21, Sno, Stno', description='strawberry notch homolog 2 (Drosophila)', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80056997, end_position_on_the_genomic_accession=80106108, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  85090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85090, CurrentID=0, Status='live', Symbol='D1Mit17', Aliases='Tgfbkm2, Tgfbm2', description='DNA segment, Chr 1, Massachusetts Institute of Technology 17', other_designations='transforming growth factor beta, modifier 2', map_location='1 95.03 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85093, CurrentID=0, Status='live', Symbol='D7Mit114', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 114', other_designations='', map_location='7 15.42 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216166, CurrentID=0, Status='live', Symbol='Plk5', Aliases='6330514A18Rik', description='polo-like kinase 5', other_designations='PLK-5|plk5|serine/threonine-protein kinase PLK5', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80356459, end_position_on_the_genomic_accession=80365489, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  216169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216169, CurrentID=0, Status='live', Symbol='Abhd17a', Aliases='1700013O15Rik, BC005632, D10Bwg1364e, Fam108a', description='abhydrolase domain containing 17A', other_designations='abhydrolase domain-containing protein FAM108A|family with sequence similarity 108, member A|mFLJ00358 protein', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80583649, end_position_on_the_genomic_accession=80590341, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  216177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216177, CurrentID=0, Status='live', Symbol='AU041133', Aliases='', description='expressed sequence AU041133', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82128013, end_position_on_the_genomic_accession=82153065, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  85110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85110, CurrentID=0, Status='live', Symbol='D15Mit12', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 12', other_designations='', map_location='15 1.8 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85111, CurrentID=0, Status='live', Symbol='D16Mit10', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 10', other_designations='', map_location='16 24.2 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85112, CurrentID=0, Status='live', Symbol='D17Mit48', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 48', other_designations='', map_location='17 5.25 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85113, CurrentID=0, Status='live', Symbol='D18Mit19', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 19', other_designations='', map_location='18 3.02 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85114, CurrentID=0, Status='live', Symbol='D18Mit20', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 20', other_designations='', map_location='18 8.24 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85116, CurrentID=0, Status='live', Symbol='D19Mit42', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 42', other_designations='', map_location='19 6.39 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216190, CurrentID=0, Status='live', Symbol='Appl2', Aliases='Dip3b', description='adaptor protein, phosphotyrosine interaction, PH domain and leucine zipper containing 2', other_designations='DIP13 beta|Dip3 beta|adapter protein containing PH domain, PTB domain and leucine zipper motif 2|dip13-beta', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=83600033, end_position_on_the_genomic_accession=83648854, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  85122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85122, CurrentID=0, Status='live', Symbol='D4Mit139', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 139', other_designations='', map_location='4 29.65 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216197, CurrentID=0, Status='live', Symbol='Ckap4', Aliases='5630400A09Rik, CLIMP-63, P63', description='cytoskeleton-associated protein 4', other_designations='63 kDa membrane protein|63-kDa cytoskeleton-linking membrane protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=84526305, end_position_on_the_genomic_accession=84533888, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  216198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216198, CurrentID=0, Status='live', Symbol='Tcp11l2', Aliases='E430026E19Rik', description='t-complex 11 (mouse) like 2', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=84576947, end_position_on_the_genomic_accession=84614355, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  85130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85130, CurrentID=0, Status='live', Symbol='D8Mit129', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 129', other_designations='', map_location='8 30.82 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85139, CurrentID=0, Status='live', Symbol='D9Mit260', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 260', other_designations='', map_location='9 37.87 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  77032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77032, CurrentID=0, Status='live', Symbol='Tstd3', Aliases='2610029I01Rik', description='thiosulfate sulfurtransferase (rhodanese)-like domain containing 3', other_designations='rhodanese domain-containing protein 3', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=21757382, end_position_on_the_genomic_accession=21767211, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  216225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216225, CurrentID=0, Status='live', Symbol='Slc5a8', Aliases='Ait, SMCT', description='solute carrier family 5 (iodide transporter), member 8', other_designations='apical iodide transporter , putative|electrogenic sodium monocarboxylate cotransporter|slc5a8|solute carrier family 5 member 8', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88885992, end_position_on_the_genomic_accession=88929515, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  85155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85155, CurrentID=0, Status='live', Symbol='D1Mit308', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 308', other_designations='', map_location='1 50.78 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85161, CurrentID=0, Status='live', Symbol='D4Mit331', Aliases='', description='DNA segment, Chr 4, Massachusetts Institute of Technology 331', other_designations='', map_location='4 47.82 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216238, CurrentID=0, Status='live', Symbol='Eea1', Aliases='A430109M19Rik, B230358H09Rik, ZFYVE2', description='early endosome antigen 1', other_designations='', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=95940660, end_position_on_the_genomic_accession=96045518, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  85169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85169, CurrentID=0, Status='live', Symbol='D11Mit42', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 42', other_designations='', map_location='11 78.7 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85170, CurrentID=0, Status='live', Symbol='D12Mit36', Aliases='', description='DNA segment, Chr 12, Massachusetts Institute of Technology 36', other_designations='', map_location='12 26.43 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85173, CurrentID=0, Status='live', Symbol='D13Mit54', Aliases='', description='DNA segment, Chr 13, Massachusetts Institute of Technology 54', other_designations='', map_location='13 28.71 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85177, CurrentID=0, Status='live', Symbol='D14Mit35', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 35', other_designations='', map_location='14 42.53 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85180, CurrentID=0, Status='live', Symbol='D16Mit15', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 15', other_designations='', map_location='16 30.92 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85181, CurrentID=0, Status='live', Symbol='D16Mit47', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 47', other_designations='', map_location='16 36.36 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85184, CurrentID=0, Status='live', Symbol='D17Mit59', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 59', other_designations='', map_location='17 11.4 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85188, CurrentID=0, Status='live', Symbol='D19Mit34', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 34', other_designations='', map_location='19 50.72 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85191, CurrentID=0, Status='live', Symbol='D19Mit88', Aliases='', description='DNA segment, Chr 19, Massachusetts Institute of Technology 88', other_designations='', map_location='19 32.23 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85195, CurrentID=0, Status='live', Symbol='D18Mit110', Aliases='', description='DNA segment, Chr 18, Massachusetts Institute of Technology 110', other_designations='', map_location='18 5.88 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216274, CurrentID=0, Status='live', Symbol='Cep290', Aliases='BC004690, Nphp6, b2b1454Clo, b2b1752Clo', description='centrosomal protein 290', other_designations='Bardet-Biedl syndrome 14 protein homolog|cep290|nephrocystin-6', map_location='10 51.48 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=100487548, end_position_on_the_genomic_accession=100573655, orientation='plus', exon_count=56, OMIM='', no_hdr0=''),
  85203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85203, CurrentID=0, Status='live', Symbol='D17Mit245', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 245', other_designations='', map_location='17 8.29 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85208, CurrentID=0, Status='live', Symbol='D11Mit124', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 124', other_designations='', map_location='11 62.92 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85211, CurrentID=0, Status='live', Symbol='D14Mit165', Aliases='', description='DNA segment, Chr 14, Massachusetts Institute of Technology 165', other_designations='', map_location='14 56.16 cM', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85212, CurrentID=0, Status='live', Symbol='D15Mit111', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 111', other_designations='', map_location='15 13.02 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216285, CurrentID=0, Status='live', Symbol='Alx1', Aliases='AI314867, C130005I02, Cart1', description='ALX homeobox 1', other_designations='CART-1|cartilage homeo protein 1|cartilage homeoprotein 1', map_location='10 53.56 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=103007847, end_position_on_the_genomic_accession=103029452, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  85219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85219, CurrentID=0, Status='live', Symbol='D17Mit197', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 197', other_designations='', map_location='17 11.4 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216292, CurrentID=0, Status='live', Symbol='Mettl25', Aliases='', description='methyltransferase like 25', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=105763185, end_position_on_the_genomic_accession=105841380, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  85225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85225, CurrentID=0, Status='live', Symbol='D11Mit242', Aliases='', description='DNA segment, Chr 11, Massachusetts Institute of Technology 242', other_designations='', map_location='11 39.47 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85228, CurrentID=0, Status='live', Symbol='D16Mit204', Aliases='', description='DNA Segment, Chr 16, Massachusetts Institute of Technology 204', other_designations='', map_location='16 54.7 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  57896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=57896, CurrentID=0, Status='live', Symbol='Krcc1', Aliases='AA792894', description='lysine-rich coiled-coil 1', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71271569, end_position_on_the_genomic_accession=71285319, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  85234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85234, CurrentID=0, Status='live', Symbol='D16Mit205', Aliases='', description='DNA Segment, Chr 16, Massachusetts Institute of Technology 205', other_designations='', map_location='16 57.43 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  232664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232664, CurrentID=0, Status='live', Symbol='Ccdc136', Aliases='4921511K06Rik, BC006583', description='coiled-coil domain containing 136', other_designations='taraco', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29396569, end_position_on_the_genomic_accession=29426995, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  216343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216343, CurrentID=0, Status='live', Symbol='Tph2', Aliases='AU043594, Ntph', description='tryptophan hydroxylase 2', other_designations='neuronal tryptophan hydroxylase|tryptophan 5-monooxygenase 2', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115078641, end_position_on_the_genomic_accession=115185022, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  216344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216344, CurrentID=0, Status='live', Symbol='Rab21', Aliases='9630024B22', description='RAB21, member RAS oncogene family', other_designations='rab-12', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115289862, end_position_on_the_genomic_accession=115316026, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  216345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216345, CurrentID=0, Status='live', Symbol='Zfc3h1', Aliases='BC033596, Ccdc131, Psrc2', description='zinc finger, C3H1-type containing', other_designations='coiled-coil domain containing 131|proline/serine-rich coiled-coil 2', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115384920, end_position_on_the_genomic_accession=115432772, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  216350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216350, CurrentID=0, Status='live', Symbol='Tspan8', Aliases='C76990, E330007O21Rik, Tm4sf3', description='tetraspanin 8', other_designations='transmembrane 4 superfamily member 3|tspan-8', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=115816817, end_position_on_the_genomic_accession=115849893, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  216363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216363, CurrentID=0, Status='live', Symbol='Rab3ip', Aliases='B230311A06, Gtpat12, Rabin3', description='RAB3A interacting protein', other_designations='RAB3A-interacting protein|SSX2 interacting protein|SSX2-interacting protein|gene trap PAT 12|rab3ip|rabin-3', map_location='10 64.96 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=116905780, end_position_on_the_genomic_accession=116950448, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  85305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85305, CurrentID=0, Status='live', Symbol='Kars', Aliases='AA589550, AL024334, AL033315, AL033367, D8Ertd698e, D8Wsu108e, LysRS, mKIAA0070', description='lysyl-tRNA synthetase', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111993439, end_position_on_the_genomic_accession=112011354, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  85308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85308, CurrentID=0, Status='live', Symbol='Emc9', Aliases='1500005A01Rik, Cgi112, Fam158a', description='ER membrane protein complex subunit 9', other_designations='UPF0172 protein FAM158A|family with sequence similarity 158, member A', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55581519, end_position_on_the_genomic_accession=55585356, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  216393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216393, CurrentID=0, Status='live', Symbol='D930020B18Rik', Aliases='1500017I17Rik', description='RIKEN cDNA D930020B18 gene', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121640811, end_position_on_the_genomic_accession=121694334, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  216395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216395, CurrentID=0, Status='live', Symbol='Tmem5', Aliases='6330415D21Rik', description='transmembrane protein 5', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=122081120, end_position_on_the_genomic_accession=122097366, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  216438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216438, CurrentID=0, Status='live', Symbol='March9', Aliases='AI608492, BC019560', description='membrane-associated ring finger (C3HC4) 9', other_designations='MARCH-IX|membrane-associated RING finger protein 9|membrane-associated RING-CH protein IX', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127056050, end_position_on_the_genomic_accession=127060184, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  216439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216439, CurrentID=0, Status='live', Symbol='Agap2', Aliases='AGAP-2, Centg1, PIKE, cnt-g1, mKIAA0167', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 2', other_designations='centaurin, gamma 1|phosphatidylinositol 3-kinase enhancer', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127075258, end_position_on_the_genomic_accession=127093173, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  216440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216440, CurrentID=0, Status='live', Symbol='Os9', Aliases='4632413K17Rik, AU022351', description='amplified in osteosarcoma', other_designations='osteosarcoma amplified 9', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127094259, end_position_on_the_genomic_accession=127121160, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  216441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216441, CurrentID=0, Status='live', Symbol='Slc26a10', Aliases='C78409, D630010I07', description='solute carrier family 26, member 10', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127172313, end_position_on_the_genomic_accession=127180709, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  216443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216443, CurrentID=0, Status='live', Symbol='Mars', Aliases='Metrs, Mtrns', description='methionine-tRNA synthetase', other_designations='methionine tRNA ligase|methionyl-tRNA synthetase, cytoplasmic', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127296221, end_position_on_the_genomic_accession=127311786, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  216445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216445, CurrentID=0, Status='live', Symbol='Arhgap9', Aliases='AU043488', description='Rho GTPase activating protein 9', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127321959, end_position_on_the_genomic_accession=127329946, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  216453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216453, CurrentID=0, Status='live', Symbol='Rdh19', Aliases='RDH-S, RDH-S2, Rdhs', description='retinol dehydrogenase 19', other_designations='', map_location='10 76.27 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127849928, end_position_on_the_genomic_accession=127861176, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  216454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216454, CurrentID=0, Status='live', Symbol='BC089597', Aliases='CRAD-L', description='cDNA sequence BC089597', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127866476, end_position_on_the_genomic_accession=127877319, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  216456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216456, CurrentID=0, Status='live', Symbol='Gls2', Aliases='A330074B06Rik, AI195532, GA, GLS, Lga', description='glutaminase 2 (liver, mitochondrial)', other_designations='L-glutaminase|L-glutamine amidohydrolase|liver mitochondrial glutaminase', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128194635, end_position_on_the_genomic_accession=128210004, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  216459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216459, CurrentID=0, Status='live', Symbol='Myl6b', Aliases='5730437E04Rik, BC037527, MLC1SA', description='myosin, light polypeptide 6B', other_designations='myl6b|myosin light chain 1 slow a|smooth muscle and nonmuscle myosin light chain alkali 6B', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128494157, end_position_on_the_genomic_accession=128498685, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  216505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216505, CurrentID=0, Status='live', Symbol='Pik3ip1', Aliases='1500004A08Rik, 5830455E04Rik, Crkd, Hgfl', description='phosphoinositide-3-kinase interacting protein 1', other_designations='calcineurin-regulated kringle domain protein|calcineurin-regulated kringle domain-containing protein|kringle domain-containing protein HGFL', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3330013, end_position_on_the_genomic_accession=3342971, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  216516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216516, CurrentID=0, Status='live', Symbol='Ccdc157', Aliases='4930562D19Rik, mKIAA1656', description='coiled-coil domain containing 157', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4141123, end_position_on_the_genomic_accession=4160293, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  216527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216527, CurrentID=0, Status='live', Symbol='Ccm2', Aliases='BC029157', description='cerebral cavernous malformation 2', other_designations='cerebral cavernous malformation 2 homolog|malcavernin|osmosensing scaffold for MEKK3', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6546887, end_position_on_the_genomic_accession=6596761, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  216543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216543, CurrentID=0, Status='live', Symbol='Cep68', Aliases='6030463E10Rik, AI481761, BC027174, Kiaa0582', description='centrosomal protein 68', other_designations='', map_location='11 12.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=20226688, end_position_on_the_genomic_accession=20250127, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  216549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216549, CurrentID=0, Status='live', Symbol='Aftph', Aliases='', description='aftiphilin', other_designations='', map_location='11 A3.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=20685084, end_position_on_the_genomic_accession=20742841, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  216551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216551, CurrentID=0, Status='live', Symbol='Lgalsl', Aliases='1110067D22Rik, A530071M23, Grpa, Hspc159a, Lgalsl', description='lectin, galactoside binding-like', other_designations='galectin-related protein A|lectin galactoside-binding-like protein A', map_location='11 13.53 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=20823354, end_position_on_the_genomic_accession=20890002, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  216558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216558, CurrentID=0, Status='live', Symbol='Ugp2', Aliases='UDPGP, UGPase', description='UDP-glucose pyrophosphorylase 2', other_designations='uridindiphosphoglucosepyrophosphorylase 2', map_location='11 13.89 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=21321126, end_position_on_the_genomic_accession=21371267, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  216560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216560, CurrentID=0, Status='live', Symbol='Wdpcp', Aliases='AV249152', description='WD repeat containing planar cell polarity effector', other_designations='WD repeat-containing protein C2orf86 homolog|homoloc-13|homolog-13|mFrtz', map_location='11 13.9 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=21572281, end_position_on_the_genomic_accession=21898687, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  216565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216565, CurrentID=0, Status='live', Symbol='Ehbp1', Aliases='AF424697, Flj21950, NACSIN', description='EH domain binding protein 1', other_designations='DNA sequence AF424697', map_location='11 14.1 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=22005825, end_position_on_the_genomic_accession=22287078, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  216578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216578, CurrentID=0, Status='live', Symbol='Papolg', Aliases='9630006B20Rik', description='poly(A) polymerase gamma', other_designations="PAP-gamma|SRP 3'-adenylating protein|SRP RNA 3'-adenylating enzyme|polynucleotide adenylyltransferase gamma", map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23860145, end_position_on_the_genomic_accession=23895296, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  68316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68316, CurrentID=0, Status='live', Symbol='Apoo', Aliases='0610008C08Rik, 1110019O03Rik', description='apolipoprotein O', other_designations='MICOS complex subunit Mic23', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94367011, end_position_on_the_genomic_accession=94417093, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  216613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216613, CurrentID=0, Status='live', Symbol='Ccdc85a', Aliases='', description='coiled-coil domain containing 85A', other_designations='', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=28385684, end_position_on_the_genomic_accession=28584441, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  216616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216616, CurrentID=0, Status='live', Symbol='Efemp1', Aliases='', description='epidermal growth factor-containing fibulin-like extracellular matrix protein 1', other_designations='FIBL-3|fibulin-3', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=28853205, end_position_on_the_genomic_accession=28926743, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  216618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216618, CurrentID=0, Status='live', Symbol='Cfap36', Aliases='2300003H10, 4931428D14Rik, Ccdc104', description='cilia and flagella associated protein 36', other_designations='coiled-coil domain containing 104|coiled-coil domain-containing protein 104', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29221536, end_position_on_the_genomic_accession=29247272, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  216635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216635, CurrentID=0, Status='live', Symbol='Hbq1a', Aliases='A230090E05Rik, Gm876, Hbq1, Hbqt2', description='hemoglobin, theta 1A', other_designations="Hbq1 3'-theta|hemoglobin, theta 1", map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32300066, end_position_on_the_genomic_accession=32304090, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  347708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=347708, CurrentID=0, Status='live', Symbol='Dppa1', Aliases='C77727', description='developmental pluripotency associated 1', other_designations='', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=46607010, end_position_on_the_genomic_accession=46632233, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  347710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=347710, CurrentID=0, Status='live', Symbol='Pramel4', Aliases='C79430, D4Ertd792e', description='preferentially expressed antigen in melanoma like 4', other_designations='', map_location='4 77.98 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144059126, end_position_on_the_genomic_accession=144069324, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  347711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=347711, CurrentID=0, Status='live', Symbol='Pramel6', Aliases='C80453, C80515', description='preferentially expressed antigen in melanoma like 6', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87506565, end_position_on_the_genomic_accession=87510868, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  347712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=347712, CurrentID=0, Status='live', Symbol='Pramel7', Aliases='AA675346, C81180', description='preferentially expressed antigen in melanoma like 7', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87489088, end_position_on_the_genomic_accession=87492418, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  216643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216643, CurrentID=0, Status='live', Symbol='Gabrp', Aliases='', description='gamma-aminobutyric acid (GABA) A receptor, pi', other_designations='GABA(A) receptor subunit pi|gamma-aminobutyric acid (GABA-A) receptor, pi', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=33550781, end_position_on_the_genomic_accession=33578978, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  347722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=347722, CurrentID=0, Status='live', Symbol='Agap1', Aliases='Centg2, Ggap1, R75364, mKIAA1099', description='ArfGAP with GTPase domain, ankyrin repeat and PH domain 1', other_designations='AGAP-1|centaurin, gamma 2|centaurin-gamma-2|cnt-g2', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=89454811, end_position_on_the_genomic_accession=89895282, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  216705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216705, CurrentID=0, Status='live', Symbol='Clint1', Aliases='AI642036, AW537616, C530049I24Rik, Enth, Epn4, F630206G17Rik, mKIAA0171', description='clathrin interactor 1', other_designations='ENTH domain containing protein|enthoprotin|epsin 4|epsin-4|epsin-related protein|epsinR', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=45851930, end_position_on_the_genomic_accession=45910625, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  216724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216724, CurrentID=0, Status='live', Symbol='Rufy1', Aliases='3000002E04Rik, Rabip4, ZFYVE12', description='RUN and FYVE domain containing 1', other_designations='FYVE-finger containing protein|Rab4a interacting protein|rab4-interacting protein', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50389286, end_position_on_the_genomic_accession=50431111, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  216725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216725, CurrentID=0, Status='live', Symbol='Adamts2', Aliases='ADAM-TS 2, ADAM-TS2, ADAMTS-2, PC I-NP, PCINP, pNPI', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 2', other_designations='a disintegrin and metalloproteinase with thrombospondin repeats|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 2|procollagen I/II amino propeptide-processing enzyme|procollagen N-endopeptidase|procollagen N-proteinase', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50602085, end_position_on_the_genomic_accession=50807573, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  216739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216739, CurrentID=0, Status='live', Symbol='Acsl6', Aliases='A330035H04Rik, AW050338, Facl6, LACS, Lacsl, mKIAA0837', description='acyl-CoA synthetase long-chain family member 6', other_designations='LACS 6|fatty acid Coenzyme A ligase, long chain 6|long-chain acyl-CoA synthetase 6', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54303824, end_position_on_the_genomic_accession=54364756, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  216742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216742, CurrentID=0, Status='live', Symbol='Fnip1', Aliases='A730024A03Rik, AI838773, AW557298', description='folliculin interacting protein 1', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54438146, end_position_on_the_genomic_accession=54518241, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  216749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216749, CurrentID=0, Status='live', Symbol='Nmur2', Aliases='Nmu2r', description='neuromedin U receptor 2', other_designations='NMU-R2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=56024990, end_position_on_the_genomic_accession=56040987, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  216760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216760, CurrentID=0, Status='live', Symbol='Mfap3', Aliases='2610509F16Rik, 2700079M14Rik, AW536327', description='microfibrillar-associated protein 3', other_designations='novel immunoglobulin domain containing protein', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=57518652, end_position_on_the_genomic_accession=57533817, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  216766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216766, CurrentID=0, Status='live', Symbol='Gemin5', Aliases='AA407055, AA407208, AI451603, BB194447, C330013N08', description='gem (nuclear organelle) associated protein 5', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58120001, end_position_on_the_genomic_accession=58168555, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  216767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216767, CurrentID=0, Status='live', Symbol='Mrpl22', Aliases='E030011D16Rik, HSPC158, MRP-L25, Rpml25', description='mitochondrial ribosomal protein L22', other_designations='L22mt|MRP-L22', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58171604, end_position_on_the_genomic_accession=58179580, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  216781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216781, CurrentID=0, Status='live', Symbol='Trim58', Aliases='Gm16, Gm1783', description='tripartite motif-containing 58', other_designations='tripartite motif-containing protein 58', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58640465, end_position_on_the_genomic_accession=58652404, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  216783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216783, CurrentID=0, Status='live', Symbol='Olfr320', Aliases='GA_x6K02T2NKPP-730312-729392, MOR285-4, MOR285-5, Olfr321-ps1', description='olfactory receptor 320', other_designations='GA_x6K02T2NKPP-733777-732813|olfactory receptor 321, pseudogene 1|olfactory receptor GA_x6K02T2NKPP-730312-729392|olfactory receptor MOR285-4', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58683875, end_position_on_the_genomic_accession=58684795, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  216792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216792, CurrentID=0, Status='live', Symbol='Iba57', Aliases='4930543L23Rik, A230051G13Rik, C1orf69', description='IBA57 homolog, iron-sulfur cluster assembly', other_designations='IBA57, iron-sulfur cluster assembly homolog|iron-sulfur cluster assembly factor homolog|putative transferase C1orf69 homolog, mitochondrial', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59155368, end_position_on_the_genomic_accession=59164591, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  216795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216795, CurrentID=0, Status='live', Symbol='Wnt9a', Aliases='Wnt14', description='wingless-type MMTV integration site family, member 9A', other_designations='wingless-type MMTV integration site 14|wingless-type MMTV integration site 9A|wnt-14', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59306849, end_position_on_the_genomic_accession=59333552, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  216797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216797, CurrentID=0, Status='live', Symbol='Prss38', Aliases='Gm249, Mpn2', description='protease, serine 38', other_designations='marapsin-2|protease, serine, 38|serine protease MPN2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59372669, end_position_on_the_genomic_accession=59376253, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  216799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216799, CurrentID=0, Status='live', Symbol='Nlrp3', Aliases='AGTAVPRL, AII/AVP, Cias1, FCAS, FCU, MWS, Mmig1, NALP3, Pypaf1', description='NLR family, pyrin domain containing 3', other_designations='NACHT/LRR/pyrin domain-containing protein 3|PYRIN-containing APAF1-like protein 1|cold autoinflammatory syndrome 1 homolog|cold autoinflammatory syndrome 1 protein homolog|cryopyrin|mast cell maturation inducible protein 1|mast cell maturation-associated-inducible protein 1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59541569, end_position_on_the_genomic_accession=59566956, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  216805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216805, CurrentID=0, Status='live', Symbol='Flcn', Aliases='AU014660, B430214A04Rik, Bhd, FLCL', description='folliculin', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59791408, end_position_on_the_genomic_accession=59810039, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  216810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216810, CurrentID=0, Status='live', Symbol='Tom1l2', Aliases='2900016I08Rik, A730055F12Rik, AU042072, Srebf1', description='target of myb1-like 2 (chicken)', other_designations='myb1-like protein 2|target of Myb-like protein 2', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60223796, end_position_on_the_genomic_accession=60352908, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  216820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216820, CurrentID=0, Status='live', Symbol='Dhrs7b', Aliases='BC003479, C79874, C80074, PexRAP', description='dehydrogenase/reductase (SDR family) member 7B', other_designations='short-chain dehydrogenase/reductase family 32C member 1', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60830631, end_position_on_the_genomic_accession=60858423, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  216821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216821, CurrentID=0, Status='live', Symbol='Tmem11', Aliases='5730466P16Rik, AA409091', description='transmembrane protein 11', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60864452, end_position_on_the_genomic_accession=60879038, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  216825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216825, CurrentID=0, Status='live', Symbol='Usp22', Aliases='AI427806', description='ubiquitin specific peptidase 22', other_designations='deubiquitinating enzyme 22|ubiquitin specific protease 22|ubiquitin thioesterase 22|ubiquitin thiolesterase 22|ubiquitin-specific-processing protease 22', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61151781, end_position_on_the_genomic_accession=61175059, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  216829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216829, CurrentID=0, Status='live', Symbol='Mmgt2', Aliases='AI852046', description='membrane magnesium transporter 2', other_designations='mmgt2', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62648632, end_position_on_the_genomic_accession=62666359, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  216831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216831, CurrentID=0, Status='live', Symbol='Arhgap44', Aliases='6330543G20, AI840762, AU040829, AW493732, Rich2', description='Rho GTPase activating protein 44', other_designations='RICH-2|Rho GTPase-activating protein RICH2|Rho-type GTPase-activating protein RICH2|RhoGAP interacting with CIP4 homologs protein 2', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=65002039, end_position_on_the_genomic_accession=65163120, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  216835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216835, CurrentID=0, Status='live', Symbol='Usp43', Aliases='', description='ubiquitin specific peptidase 43', other_designations='deubiquitinating enzyme 43|ubiquitin thioesterase 43|ubiquitin thiolesterase 43|ubiquitin-specific-processing protease 43', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67854523, end_position_on_the_genomic_accession=67922153, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  216846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216846, CurrentID=0, Status='live', Symbol='Cntrob', Aliases='9830165K03Rik, Lip8', description='centrobin, centrosomal BRCA2 interacting protein', other_designations='LYST-interacting protein 8|LYST-interacting protein LIP8|centrosomal BRCA2-interacting protein', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69299491, end_position_on_the_genomic_accession=69323893, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  216848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216848, CurrentID=0, Status='live', Symbol='Chd3', Aliases='2600010P09Rik, AF020312, Chd7, Prp7, Prp9-1', description='chromodomain helicase DNA binding protein 3', other_designations='Mi-2 alpha', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69343273, end_position_on_the_genomic_accession=69369426, orientation='minus', exon_count=42, OMIM='', no_hdr0=''),
  216850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216850, CurrentID=0, Status='live', Symbol='Kdm6b', Aliases='1700064E03Rik, BC038313, Jmjd3, mKIAA0346', description='KDM1 lysine (K)-specific demethylase 6B', other_designations='jmjC domain-containing protein 3|jumonji domain containing 3|jumonji domain-containing protein 3', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69398508, end_position_on_the_genomic_accession=69418148, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  216853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216853, CurrentID=0, Status='live', Symbol='Wrap53', Aliases='BC021790, Wdr79', description='WD repeat containing, antisense to Trp53', other_designations='WD repeat containing, antisense to TP53|WD repeat domain 79|WD repeat-containing protein 79|WD repeat-containing protein WRAP53|WD40 repeat-containing protein antisense to TP53|WD40 repeat-containing protein encoded by RNA antisense to p53', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69561754, end_position_on_the_genomic_accession=69580251, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  216856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216856, CurrentID=0, Status='live', Symbol='Nlgn2', Aliases='NL2', description='neuroligin 2', other_designations='ortholog of human and rat neuroligin 2 NLGN2', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69823123, end_position_on_the_genomic_accession=69838377, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  216858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216858, CurrentID=0, Status='live', Symbol='Kctd11', Aliases='AF465352, Ren', description='potassium channel tetramerisation domain containing 11', other_designations='retinoic acid, EGF, and NGF upregulated', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69878264, end_position_on_the_genomic_accession=69880985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  216859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216859, CurrentID=0, Status='live', Symbol='Acap1', Aliases='Centb1', description='ArfGAP with coiled-coil, ankyrin repeat and PH domains 1', other_designations='centaurin, beta 1|centaurin-beta-1|cnt-b1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69881567, end_position_on_the_genomic_accession=69895539, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  216860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216860, CurrentID=0, Status='live', Symbol='Neurl4', Aliases='', description='neuralized E3 ubiquitin protein ligase 4', other_designations='neuralized homolog 4', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69901816, end_position_on_the_genomic_accession=69913824, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  216864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216864, CurrentID=0, Status='live', Symbol='Mgl2', Aliases='CD301b', description='macrophage galactose N-acetyl-galactosamine specific lectin 2', other_designations='macrophage galactose-type C-type lectin 2', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70130125, end_position_on_the_genomic_accession=70137542, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  216867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216867, CurrentID=0, Status='live', Symbol='Slc16a11', Aliases='AB041591, MCT 11, MNCb-2717', description='solute carrier family 16 (monocarboxylic acid transporters), member 11', other_designations='novel RNA polymerase Rpb1 C-terminal repeat domain containing protein|solute carrier family 16 member 11', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70213910, end_position_on_the_genomic_accession=70216414, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  216869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216869, CurrentID=0, Status='live', Symbol='Arrb2', Aliases='AI326910, AW122872', description='arrestin, beta 2', other_designations='beta arr2', map_location='11 42.99 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70432580, end_position_on_the_genomic_accession=70440828, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  216871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216871, CurrentID=0, Status='live', Symbol='Gltpd2', Aliases='C730027E14Rik', description='glycolipid transfer protein domain containing 2', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70519209, end_position_on_the_genomic_accession=70520736, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  216873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216873, CurrentID=0, Status='live', Symbol='Spag7', Aliases='5730443G10, ACRP, FSA-1, Fsa1l', description='sperm associated antigen 7', other_designations='', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70663769, end_position_on_the_genomic_accession=70669416, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  216874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216874, CurrentID=0, Status='live', Symbol='Camta2', Aliases='Kiaa0909-hp, mKIAA0909', description='calmodulin binding transcription activator 2', other_designations='', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70669463, end_position_on_the_genomic_accession=70688108, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  216877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216877, CurrentID=0, Status='live', Symbol='Dhx33', Aliases='3110057P17Rik, 9430096J02Rik, Ddx33', description='DEAH (Asp-Glu-Ala-His) box polypeptide 33', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 33|DEAH box protein 33', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70984091, end_position_on_the_genomic_accession=71004553, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  216881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216881, CurrentID=0, Status='live', Symbol='Wscd1', Aliases='mKIAA0523', description='WSC domain containing 1', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=71749923, end_position_on_the_genomic_accession=71789647, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  216892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216892, CurrentID=0, Status='live', Symbol='Spns2', Aliases='', description='spinster homolog 2', other_designations='spinster-like protein|spinster2', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72451638, end_position_on_the_genomic_accession=72489920, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  15953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15953, CurrentID=0, Status='live', Symbol='Ifi47', Aliases='47kDa, IRG-47, Ifggc1, Igrd, Iigp4, Iipg4, Irgd', description='interferon gamma inducible protein 47', other_designations='interferon gamma inducible protein, 47 kDa|interferon-gamma-inducible GTPase Ifggc1 protein', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49076587, end_position_on_the_genomic_accession=49096987, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  216961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216961, CurrentID=0, Status='live', Symbol='Coro6', Aliases='', description='coronin 6', other_designations='clipin E|clipin-E|coronin 6/clipin E type D|coronin, actin binding protein 6|coronin-like protein E', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77462321, end_position_on_the_genomic_accession=77470484, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  216963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216963, CurrentID=0, Status='live', Symbol='Git1', Aliases='Cat-1, p95Cat', description='G protein-coupled receptor kinase-interactor 1', other_designations='ARF GAP GIT1|G protein-coupled receptor kinase-associated ADP ribosylation factor GTPase-activating protein|GRK-interacting protein 1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77493412, end_position_on_the_genomic_accession=77507775, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  216964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216964, CurrentID=0, Status='live', Symbol='Trp53i13', Aliases='2410019G02Rik, Tp53i13', description='transformation related protein 53 inducible protein 13', other_designations='tumor protein p53 inducible protein 13', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77508099, end_position_on_the_genomic_accession=77515921, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  216965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216965, CurrentID=0, Status='live', Symbol='Taok1', Aliases='2810468K05Rik, AU020252, D130018F14Rik, Map3k16, Markk, Psk2, mKIAA1361', description='TAO kinase 1', other_designations='serine/threonine protein kinase TAO1 homolog|thousand and one amino acid protein 1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77529162, end_position_on_the_genomic_accession=77607860, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  216971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216971, CurrentID=0, Status='live', Symbol='Fam222b', Aliases='9630020I17Rik', description='family with sequence similarity 222, member B', other_designations='uncharacterized protein C17orf63 homolog', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78094666, end_position_on_the_genomic_accession=78157339, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  216974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216974, CurrentID=0, Status='live', Symbol='Proca1', Aliases='4933404M19Rik', description='protein interacting with cyclin A1', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78193392, end_position_on_the_genomic_accession=78205763, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  216976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216976, CurrentID=0, Status='live', Symbol='BC030499', Aliases='BC037104, Sgk494', description='cDNA sequence BC030499', other_designations='sugen kinase 494', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78290732, end_position_on_the_genomic_accession=78296805, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  216984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216984, CurrentID=0, Status='live', Symbol='Evi2b', Aliases='Evi-2, Evi2', description='ecotropic viral integration site 2b', other_designations='EVI-2B|ecotropic viral integration site 2B protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79513385, end_position_on_the_genomic_accession=79523762, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  216987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216987, CurrentID=0, Status='live', Symbol='Utp6', Aliases='4732497O03Rik, HCA66', description='UTP6, small subunit (SSU) processome component, homolog (yeast)', other_designations='hepatocellular carcinoma-associated antigen 66|multiple hat domains protein', map_location='11 47.29 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79932320, end_position_on_the_genomic_accession=79962426, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  216991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216991, CurrentID=0, Status='live', Symbol='Adap2', Aliases='Centa2', description='ArfGAP with dual PH domains 2', other_designations='centaurin, alpha 2|centaurin-alpha-2|cnt-a2', map_location='11 47.56 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80153896, end_position_on_the_genomic_accession=80178827, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  217011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217011, CurrentID=0, Status='live', Symbol='Nle1', Aliases='AL022765, BC018399, Nle, l11Jus1, l11Jus4', description='notchless homolog 1 (Drosophila)', other_designations='Notchless gene homolog|novel WD40 repeat containing protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82900768, end_position_on_the_genomic_accession=82908445, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  217012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217012, CurrentID=0, Status='live', Symbol='Unc45b', Aliases='AA445617, Cmya4, D230041A13Rik, Unc45', description='unc-45 homolog B (C. elegans)', other_designations='UNC-45B|cardiomyopathy associated 4|striated muscle UNC45', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82911253, end_position_on_the_genomic_accession=82943406, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  217026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217026, CurrentID=0, Status='live', Symbol='Heatr6', Aliases='2700008B19Rik, BB135312, BB249351', description='HEAT repeat containing 6', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83753397, end_position_on_the_genomic_accession=83783754, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  217030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217030, CurrentID=0, Status='live', Symbol='Synrg', Aliases='Ap1gbp1, SYNG', description='synergin, gamma', other_designations='AP1 gamma subunit binding protein 1', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83964380, end_position_on_the_genomic_accession=84044578, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  217031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217031, CurrentID=0, Status='live', Symbol='Tada2a', Aliases='AV319371, D030022J10Rik, Tada2l', description='transcriptional adaptor 2A', other_designations='ADA2-like protein', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84078920, end_position_on_the_genomic_accession=84129642, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  217038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217038, CurrentID=0, Status='live', Symbol='Mrm1', Aliases='A530065E19Rik, ENSMUSG00000054212', description='mitochondrial rRNA methyltransferase 1 homolog (S. cerevisiae)', other_designations="16S rRNA (guanosine(1145)-2'-O)-methyltransferase|16S rRNA [Gm1145] 2'-O-methyltransferase|mitochondrial large ribosomal RNA ribose methylase", map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84813061, end_position_on_the_genomic_accession=84819515, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  217039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217039, CurrentID=0, Status='live', Symbol='Ggnbp2', Aliases='AI451896, C75939, D330017P12Rik, DIF-3, Zfp403', description='gametogenetin binding protein 2', other_designations='dioxin inducible factor 3|dioxin-inducible factor 3|gametogenin binding protein 2|zinc finger protein 403', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84832361, end_position_on_the_genomic_accession=84870738, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  217057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217057, CurrentID=0, Status='live', Symbol='Ptrh2', Aliases='A230072I16Rik, Bit1, CGI-147', description='peptidyl-tRNA hydrolase 2', other_designations='PTH 2', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86683983, end_position_on_the_genomic_accession=86692457, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  217069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217069, CurrentID=0, Status='live', Symbol='Trim25', Aliases='AA960166, AL022677, EFP, Zfp147', description='tripartite motif-containing 25', other_designations='RING-type E3 ubiquitin transferase|estrogen-responsive finger protein|tripartite motif protein 25|tripartite motif-containing protein 25|ubiquitin/ISG15-conjugating enzyme TRIM25|zinc finger protein 147', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88999403, end_position_on_the_genomic_accession=89020293, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  217071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217071, CurrentID=0, Status='live', Symbol='Gm525', Aliases='', description='predicted gene 525', other_designations='novel transcript', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=89073841, end_position_on_the_genomic_accession=89093064, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  217082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217082, CurrentID=0, Status='live', Symbol='Hlf', Aliases='E230015K02Rik', description='hepatic leukemia factor', other_designations='', map_location='11 55.03 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=90336535, end_position_on_the_genomic_accession=90390917, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  217109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217109, CurrentID=0, Status='live', Symbol='Utp18', Aliases='6230425C22Rik, 6430513M13, Wdr50', description='UTP18, small subunit (SSU) processome component, homolog (yeast)', other_designations='UTP18, small subunit processome component|WD repeat domain 50|WD repeat-containing protein 50', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=93859243, end_position_on_the_genomic_accession=93885766, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  217116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217116, CurrentID=0, Status='live', Symbol='Spata20', Aliases='Tisp78', description='spermatogenesis associated 20', other_designations='sperm protein SSP411|sperm-specific protein 411|ssp411|transcript increased in spermiogenesis 78 protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94478904, end_position_on_the_genomic_accession=94485310, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  217119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217119, CurrentID=0, Status='live', Symbol='Xylt2', Aliases='E030002B02Rik, XT-II, XT2, xylT-II', description='xylosyltransferase II', other_designations='peptide O-xylosyltransferase 2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94663845, end_position_on_the_genomic_accession=94677518, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  217122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217122, CurrentID=0, Status='live', Symbol='Gm11545', Aliases='A430060F13Rik, C86086, OTTMUSG00000002043', description='predicted gene 11545', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94755129, end_position_on_the_genomic_accession=94761197, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  217124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217124, CurrentID=0, Status='live', Symbol='Ppp1r9b', Aliases='SPL, Spn', description='protein phosphatase 1, regulatory subunit 9B', other_designations='neurabin II|neurabin-II|spinophilin', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94991212, end_position_on_the_genomic_accession=95006898, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  217125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217125, CurrentID=0, Status='live', Symbol='Samd14', Aliases='AI839049, AI854782, BC034054', description='sterile alpha motif domain containing 14', other_designations='SAM domain-containing protein 14', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95009402, end_position_on_the_genomic_accession=95026087, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  217127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217127, CurrentID=0, Status='live', Symbol='Kat7', Aliases='Hbo1, Hboa, Myst2', description='K(lysine) acetyltransferase 7', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 2|MYST histone acetyltransferase 2|histone acetyltransferase MYST2|histone acetyltransferase binding to ORC1|lysine acetyltransferase 7', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95271853, end_position_on_the_genomic_accession=95310246, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  217138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217138, CurrentID=0, Status='live', Symbol='Prr15l', Aliases='Atad4, BC022765, E230006A12Rik', description='proline rich 15-like', other_designations='ATPase family, AAA domain containing 4', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96926179, end_position_on_the_genomic_accession=96935648, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  217140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217140, CurrentID=0, Status='live', Symbol='Scrn2', Aliases='AV001119, D11Moh48, SES2', description='secernin 2', other_designations='', map_location='11 60.62 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97029902, end_position_on_the_genomic_accession=97033960, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  217143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217143, CurrentID=0, Status='live', Symbol='Gpr179', Aliases='5330439C02Rik, Gpr158l1', description='G protein-coupled receptor 179', other_designations='GPR158-like 1 receptor', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97332109, end_position_on_the_genomic_accession=97352394, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  217149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217149, CurrentID=0, Status='live', Symbol='Cisd3', Aliases='Mel-13, Mel13', description='CDGSH iron sulfur domain 3', other_designations='CDGSH iron sulfur domain-containing protein 3, mitochondrial|melanoma nuclear protein 13', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97685504, end_position_on_the_genomic_accession=97688629, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  217151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217151, CurrentID=0, Status='live', Symbol='Arl5c', Aliases='Arl12, BC065791', description='ADP-ribosylation factor-like 5C', other_designations='ADP-ribosylation factor-like 12|ADP-ribosylation factor-like 5C protein|ADP-ribosylation factor-like protein 12', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97989578, end_position_on_the_genomic_accession=97996272, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  217154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217154, CurrentID=0, Status='live', Symbol='Stac2', Aliases='24b2, 24b2/STAC2, AW240854', description='SH3 and cysteine rich domain 2', other_designations='SRC homology 3 and cysteine-rich domain-containing protein 2', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98036624, end_position_on_the_genomic_accession=98053462, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  217166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217166, CurrentID=0, Status='live', Symbol='Nr1d1', Aliases='A530070C09Rik, R75201', description='nuclear receptor subfamily 1, group D, member 1', other_designations='EAR-1|REV-ERBalpha|Rev-ErbA-alpha protein|V-erbA-related protein 1|nuclear receptor protein|orphan nuclear receptor|rev-erb alpha|rev-erbA alpha|rev-erbA(alpha)', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98767932, end_position_on_the_genomic_accession=98775377, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  217169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217169, CurrentID=0, Status='live', Symbol='Tns4', Aliases='9930017A07Rik, AA589547, AU016405, Cten', description='tensin 4', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99065678, end_position_on_the_genomic_accession=99089306, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  217194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217194, CurrentID=0, Status='live', Symbol='Klhl11', Aliases='BC011167', description='kelch-like 11', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100457122, end_position_on_the_genomic_accession=100472810, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100502887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502887, CurrentID=0, Status='live', Symbol='Olfr331', Aliases='MOR275-4', description='olfactory receptor 331', other_designations='GA_x6K02T2NKPP-912840-913784|olfactory receptor MOR275-4', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58501598, end_position_on_the_genomic_accession=58508490, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  217198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217198, CurrentID=0, Status='live', Symbol='Plekhh3', Aliases='BC020025', description='pleckstrin homology domain containing, family H (with MyTH4 domain) member 3', other_designations='PH domain-containing family H member 3', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101162679, end_position_on_the_genomic_accession=101171405, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  217201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217201, CurrentID=0, Status='live', Symbol='Rundc1', Aliases='AW125546, D930024B17', description='RUN domain containing 1', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101425084, end_position_on_the_genomic_accession=101435666, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  217203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217203, CurrentID=0, Status='live', Symbol='Tmem106a', Aliases='0610008L10Rik, AI043106, BC022145', description='transmembrane protein 106A', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101582242, end_position_on_the_genomic_accession=101591788, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  217207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217207, CurrentID=0, Status='live', Symbol='Dhx8', Aliases='Ddx8, mDEAH6, mKIAA4096', description='DEAH (Asp-Glu-Ala-His) box polypeptide 8', other_designations='DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 8|DEAH box protein 8|RNA helicase', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101732956, end_position_on_the_genomic_accession=101767358, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  217212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217212, CurrentID=0, Status='live', Symbol='Pyy', Aliases='', description='peptide YY', other_designations='Peptide tyrosine tyrosine', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102106676, end_position_on_the_genomic_accession=102107859, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  217214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217214, CurrentID=0, Status='live', Symbol='Nags', Aliases='1700120E20Rik, argA', description='N-acetylglutamate synthase', other_designations='amino-acid N-acetyltransferase|amino-acid acetyltransferase', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102145513, end_position_on_the_genomic_accession=102149528, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  217216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217216, CurrentID=0, Status='live', Symbol='BC030867', Aliases='', description='cDNA sequence BC030867', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102248865, end_position_on_the_genomic_accession=102265183, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  217217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217217, CurrentID=0, Status='live', Symbol='Asb16', Aliases='', description='ankyrin repeat and SOCS box-containing 16', other_designations='ASB-16|ankyrin repeat domain-containing SOCS box protein Asb-16', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102268823, end_position_on_the_genomic_accession=102278061, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  217218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217218, CurrentID=0, Status='live', Symbol='Atxn7l3', Aliases='E030022H21Rik', description='ataxin 7-like 3', other_designations='SAGA-associated factor 11 homolog', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102289300, end_position_on_the_genomic_accession=102296643, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  232811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232811, CurrentID=0, Status='live', Symbol='Suv420h2', Aliases='BC024816, KMT5C, Suv4-20h2', description='suppressor of variegation 4-20 homolog 2 (Drosophila)', other_designations='su(var)4-20 homolog 2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4740092, end_position_on_the_genomic_accession=4747514, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  217232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217232, CurrentID=0, Status='live', Symbol='Cdc27', Aliases='AI452358, APC3, BC023187', description='cell division cycle 27', other_designations='cell division cycle 27 homolog', map_location='11 67.79 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104502524, end_position_on_the_genomic_accession=104550620, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  217246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217246, CurrentID=0, Status='live', Symbol='Ace3', Aliases='EG217246', description='angiotensin I converting enzyme (peptidyl-dipeptidase A) 3', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105994418, end_position_on_the_genomic_accession=106007503, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  101744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101744, CurrentID=0, Status='live', Symbol='Acp7', Aliases='AU017076, C330005M16Rik, Papl', description='acid phosphatase 7, tartrate resistant', other_designations='iron/zinc purple acid phospatase-like', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28607634, end_position_on_the_genomic_accession=28631361, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  217258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217258, CurrentID=0, Status='live', Symbol='Abca8a', Aliases='', description='ATP-binding cassette, sub-family A (ABC1), member 8a', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=110025634, end_position_on_the_genomic_accession=110097583, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  217262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217262, CurrentID=0, Status='live', Symbol='Abca9', Aliases='D630040K07Rik', description='ATP-binding cassette, sub-family A (ABC1), member 9', other_designations='ATP-binding cassette transporter sub-family A member 9', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=110100822, end_position_on_the_genomic_accession=110168153, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  217265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217265, CurrentID=0, Status='live', Symbol='Abca5', Aliases='ABC13, B930033A02Rik, mKIAA1888', description='ATP-binding cassette, sub-family A (ABC1), member 5', other_designations='ATP-binding cassette transporter sub-family A member 5', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=110269369, end_position_on_the_genomic_accession=110337724, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  217302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217302, CurrentID=0, Status='live', Symbol='Gpr142', Aliases='PGR2', description='G protein-coupled receptor 142', other_designations='G protein-coupled receptor PGR2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114796965, end_position_on_the_genomic_accession=114806808, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  217303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217303, CurrentID=0, Status='live', Symbol='Cd300a', Aliases='B230315M08Rik, Clm8, LMIR1, MAIR-I, MAIR-Ia, MMAC8, Pigr4, mcpir1', description='CD300A antigen', other_designations='CD300 antigen-like family member A|CLM-8|MAIR-1|leukocyte mono-Ig-like receptor 1|mast cell-derived paired immunoglobulin-like receptor 1|myeloid-associated immunoglobulin-like receptor 1|polymeric immunoglobulin receptor 4', map_location='11 80.54 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114889149, end_position_on_the_genomic_accession=114904651, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  217304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217304, CurrentID=0, Status='live', Symbol='Cd300lb', Aliases='CLM-7, Clm7, Gm253, LMIR5', description='CD300 antigen like family member B', other_designations='CD300 antigen-like family member B|CMRF-35-like molecule-7|IREM-3|immune receptor expressed on myeloid cells 3|leukocyte mono-Ig-like receptor 5', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114922779, end_position_on_the_genomic_accession=114934386, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  217305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217305, CurrentID=0, Status='live', Symbol='Cd300ld', Aliases='4732429D16Rik, CLM5, LMIR4, MAIR-IV, clm-5', description='CD300 molecule-like family member d', other_designations='CD300 antigen like family member D|MAIR-4|leukocyte mono-Ig-like receptor 4|myeloid-associated immunoglobulin (Ig)-like receptor IV|myeloid-associated immunoglobulin-like receptor 4|novel immunoglobulin V-set protein (Clm5)', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114982443, end_position_on_the_genomic_accession=114989886, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  217310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217310, CurrentID=0, Status='live', Symbol='Hid1', Aliases='C630004H02Rik', description='HID1 domain containing', other_designations='HID1 domain-containing protein|UPF0663 transmembrane protein C17orf28 homolog|protein hid-1 homolog', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115347705, end_position_on_the_genomic_accession=115367753, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  217316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217316, CurrentID=0, Status='live', Symbol='Slc16a5', Aliases='A130015N09Rik, MCT 5, MCT 6, MCT5', description='solute carrier family 16 (monocarboxylic acid transporters), member 5', other_designations='monocarboxylate transporter 5|solute carrier family 16 member 5', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115462250, end_position_on_the_genomic_accession=115474398, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  217325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217325, CurrentID=0, Status='live', Symbol='Llgl2', Aliases='9130006H11Rik, Llglh2', description='lethal giant larvae homolog 2 (Drosophila)', other_designations='lethal giant larvae-like protein 2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115824029, end_position_on_the_genomic_accession=115855780, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  217328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217328, CurrentID=0, Status='live', Symbol='Myo15b', Aliases='E330039G21Rik, Gm688, Gm886, mKIAA1783', description='myosin XVB', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115858391, end_position_on_the_genomic_accession=115892594, orientation='plus', exon_count=66, OMIM='', no_hdr0=''),
  217331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217331, CurrentID=0, Status='live', Symbol='Unk', Aliases='B230379M23Rik, Zc3h5, Zc3hdc5, mKIAA1753', description='unkempt homolog (Drosophila)', other_designations='zinc finger CCCH domain-containing protein 5|zinc finger CCCH type containing 5|zinc finger CCCH type domain containing 5', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116030295, end_position_on_the_genomic_accession=116061214, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  217333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217333, CurrentID=0, Status='live', Symbol='Trim47', Aliases='2210023F24Rik', description='tripartite motif-containing 47', other_designations='tripartite motif protein 47', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116105750, end_position_on_the_genomic_accession=116110235, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  217335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217335, CurrentID=0, Status='live', Symbol='Fbf1', Aliases='1110033G01Rik, 5930413D20, mKIAA1863', description='Fas (TNFRSF6) binding factor 1', other_designations='FBF-1|likely ortholog of H. sapiens Fas (TNFRSF6) binding factor 1 (FBF1)', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116142285, end_position_on_the_genomic_accession=116168270, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  217337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217337, CurrentID=0, Status='live', Symbol='Srp68', Aliases='2610024I03Rik', description='signal recognition particle 68', other_designations='likely ortholog of H. sapiens signal recognition particle 68kDa (SRP68)|signal recognition particle 68 kDa protein|signal recognition particle 68kDa', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116245166, end_position_on_the_genomic_accession=116274217, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  217340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217340, CurrentID=0, Status='live', Symbol='Rnf157', Aliases='2610036E23Rik, A130073L17Rik, mKIAA1917', description='ring finger protein 157', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116336345, end_position_on_the_genomic_accession=116413032, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  217341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217341, CurrentID=0, Status='live', Symbol='Qrich2', Aliases='Gm66', description='glutamine rich 2', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116441323, end_position_on_the_genomic_accession=116466269, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  217342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217342, CurrentID=0, Status='live', Symbol='Ube2o', Aliases='9630022H21, B230113M03Rik, E2-230K, mKIAA1734', description='ubiquitin-conjugating enzyme E2O', other_designations='E2/E3 hybrid ubiquitin-protein ligase UBE2O|ubiquitin carrier protein O|ubiquitin-conjugating enzyme E2 of 230 kDa|ubiquitin-conjugating enzyme E2-230K|ubiquitin-protein ligase O|ubiquitinating enzyme E2-230 kDa', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116537740, end_position_on_the_genomic_accession=116581476, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  217344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217344, CurrentID=0, Status='live', Symbol='Rhbdf2', Aliases='4732465I17Rik, Rhbdl6, cub', description='rhomboid 5 homolog 2 (Drosophila)', other_designations='iRhom2|rhomboid family member 2|rhomboid veinlet-like protein 6|rhomboid, veinlet-like 6|rhomboid-like protein 6|rhomboid-related protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116598165, end_position_on_the_genomic_accession=116627019, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  217351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217351, CurrentID=0, Status='live', Symbol='Tnrc6c', Aliases='9930033H14Rik, mKIAA1582', description='trinucleotide repeat containing 6C', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117654289, end_position_on_the_genomic_accession=117763439, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  217353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217353, CurrentID=0, Status='live', Symbol='Tmc6', Aliases='D11Ertd204e, EVER1', description='transmembrane channel-like gene family 6', other_designations='', map_location='11 82.96 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117765985, end_position_on_the_genomic_accession=117782193, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  217356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217356, CurrentID=0, Status='live', Symbol='Tmc8', Aliases='EVIN2, Ever2, mFLJ00400', description='transmembrane channel-like gene family 8', other_designations='epidermodysplasia verruciformis 2|epidermodysplasia verruciformis protein 2 homolog|transmembrane channel-like 8', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=117781598, end_position_on_the_genomic_accession=117793137, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  217364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217364, CurrentID=0, Status='live', Symbol='Engase', Aliases='C130099A03, D230014K01Rik', description='endo-beta-N-acetylglucosaminidase', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118476066, end_position_on_the_genomic_accession=118489212, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  217365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217365, CurrentID=0, Status='live', Symbol='Nploc4', Aliases='AK129375, Npl4, mKIAA1499', description='nuclear protein localization 4 homolog (S. cerevisiae)', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120379798, end_position_on_the_genomic_accession=120437700, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  217366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217366, CurrentID=0, Status='live', Symbol='Lrrc45', Aliases='BC023296', description='leucine rich repeat containing 45', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120713953, end_position_on_the_genomic_accession=120721128, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  217369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217369, CurrentID=0, Status='live', Symbol='Uts2r', Aliases='Gpr14, UR-2-R, UTR, UTR2', description='urotensin 2 receptor', other_designations='G protein-coupled receptor 14|G-protein coupled receptor 14|UR-II-R|urotensin II receptor', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121160271, end_position_on_the_genomic_accession=121161973, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  217370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217370, CurrentID=0, Status='live', Symbol='BC017643', Aliases='', description='cDNA sequence BC017643', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121222588, end_position_on_the_genomic_accession=121229558, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  217371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217371, CurrentID=0, Status='live', Symbol='Rab40b', Aliases='SEC4L, rar', description='Rab40B, member RAS oncogene family', other_designations='RAR SOCS box containing protein|SOCS box-containing protein RAR', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121356121, end_position_on_the_genomic_accession=121388251, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  735280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735280, CurrentID=0, Status='live', Symbol='Mir675', Aliases='Mirn675, mir-675, mmu-mir-675', description='microRNA 675', other_designations='mir 675', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142577064, end_position_on_the_genomic_accession=142577147, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  217378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217378, CurrentID=0, Status='live', Symbol='Dnajc27', Aliases='AI639580, C330021A05Rik, Rabj, Rbj', description='DnaJ heat shock protein family (Hsp40) member C27', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 27|Rab-related GTP-binding protein|rab and DnaJ domain containing|rab and DnaJ domain-containing protein', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4082568, end_position_on_the_genomic_accession=4118013, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  217379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217379, CurrentID=0, Status='live', Symbol='Ubxn2a', Aliases='6330407P03Rik, Ubxd4', description='UBX domain protein 2A', other_designations='UBX domain containing 4|UBX domain-containing protein 4', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4876647, end_position_on_the_genomic_accession=4907781, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  217410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217410, CurrentID=0, Status='live', Symbol='Trib2', Aliases='AW319517, TRB-2, TRB2', description='tribbles homolog 2 (Drosophila)', other_designations='trib2', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=15791725, end_position_on_the_genomic_accession=15816962, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  735286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=735286, CurrentID=0, Status='live', Symbol='Mir467a-1', Aliases='Mir467a, Mirn467, Mirn467a, Mirn467a-1, mir-467a-1, mmu-mir-467, mmu-mir-467a, mmu-mir-467a-1', description='microRNA 467a-1', other_designations='microRNA 467|microRNA 467a|microRNA mir-467a|mir 467|mir 467a', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10476346, end_position_on_the_genomic_accession=10476418, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  217430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217430, CurrentID=0, Status='live', Symbol='Pqlc3', Aliases='C78076, E030024M05Rik', description='PQ loop repeat containing', other_designations='PQ loop repeat containing 3', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=16988648, end_position_on_the_genomic_accession=17000419, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  217431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217431, CurrentID=0, Status='live', Symbol='Nol10', Aliases='Gm67', description='nucleolar protein 10', other_designations='', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=17348493, end_position_on_the_genomic_accession=17430095, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  217449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217449, CurrentID=0, Status='live', Symbol='Trappc12', Aliases='CGI-87, D930014A20Rik, Ttc15', description='trafficking protein particle complex 12', other_designations='TPR repeat protein 15|tetratricopeptide repeat domain 15|tetratricopeptide repeat protein 15', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28690620, end_position_on_the_genomic_accession=28750488, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  217463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217463, CurrentID=0, Status='live', Symbol='Snx13', Aliases='Rgs-px1, mKIAA0713', description='sorting nexin 13', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=35046121, end_position_on_the_genomic_accession=35147480, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  217473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217473, CurrentID=0, Status='live', Symbol='Ankmy2', Aliases='AI035571, ZMYND20', description='ankyrin repeat and MYND domain containing 2', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=36157124, end_position_on_the_genomic_accession=36197291, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  217480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217480, CurrentID=0, Status='live', Symbol='Dgkb', Aliases='6430574F24, 90kda, C630029D13Rik, DAGK2, DGK, DGK-beta, mKIAA0718', description='diacylglycerol kinase, beta', other_designations='DAG kinase beta|diglyceride kinase beta', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=37880168, end_position_on_the_genomic_accession=38634239, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  217517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217517, CurrentID=0, Status='live', Symbol='Stxbp6', Aliases='BC024598, C85317', description='syntaxin binding protein 6 (amisyn)', other_designations='amisyn', map_location='12 B3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=44852484, end_position_on_the_genomic_accession=45074737, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  217558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217558, CurrentID=0, Status='live', Symbol='G2e3', Aliases='6030408C04Rik, 9030416F18, AW046379, D930034K21Rik, mKIAA1333', description='G2/M-phase specific E3 ubiquitin ligase', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51348230, end_position_on_the_genomic_accession=51376986, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  217578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217578, CurrentID=0, Status='live', Symbol='Baz1a', Aliases='Acf1, B930060C03, BC065123, Gtl5, Wcrf180, cbp146', description='bromodomain adjacent to zinc finger domain 1A', other_designations='gene trap locus 5', map_location='12 C', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=54892600, end_position_on_the_genomic_accession=55014371, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  217588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217588, CurrentID=0, Status='live', Symbol='Mbip', Aliases='4933408E06Rik', description='MAP3K12 binding inhibitory protein 1', other_designations='MAPK upstream kinase-binding inhibitory protein|MUK-binding inhibitory protein', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=56328306, end_position_on_the_genomic_accession=56345894, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  217593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217593, CurrentID=0, Status='live', Symbol='Slc25a21', Aliases='9930033G19Rik, A630030I10, BB158148', description='solute carrier family 25 (mitochondrial oxodicarboxylate carrier), member 21', other_designations='mitochondrial solute carrier family 25 member 21|solute carrier family 25 member 21', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=56712634, end_position_on_the_genomic_accession=57197472, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  217615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217615, CurrentID=0, Status='live', Symbol='Ctage5', Aliases='D12Bwg0579e, Mea6, Mgea, Mgea6', description='CTAGE family, member 5', other_designations='cutaneous T-cell lymphoma-associated antigen 5 homolog|meningioma expressed antigen 6 (coiled-coil proline-rich)|meningioma-expressed antigen 6|protein cTAGE-5', map_location='12 26.01 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=59129740, end_position_on_the_genomic_accession=59190220, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  329659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=329659, CurrentID=0, Status='live', Symbol='E130311K13Rik', Aliases='', description='RIKEN cDNA E130311K13 gene', other_designations='', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=63914696, end_position_on_the_genomic_accession=63929385, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  217653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217653, CurrentID=0, Status='live', Symbol='Mis18bp1', Aliases='6720403H11, C79407, Knl2, mKIAA1903', description='MIS18 binding protein 1', other_designations='kinetochore-associated protein KNL-2 homolog', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=65132734, end_position_on_the_genomic_accession=65172580, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  217664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217664, CurrentID=0, Status='live', Symbol='Mgat2', Aliases='AA407964, CDGS2, GLCNACTII, GNT-II, GNT2', description='mannoside acetylglucosaminyltransferase 2', other_designations='N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase II|beta-1,2-N-acetylglucosaminyltransferase II|glcNAc-T II', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69184158, end_position_on_the_genomic_accession=69186773, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  217666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217666, CurrentID=0, Status='live', Symbol='L2hgdh', Aliases='BC016226', description='L-2-hydroxyglutarate dehydrogenase', other_designations='duranin', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69690436, end_position_on_the_genomic_accession=69724874, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  217674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217674, CurrentID=0, Status='live', Symbol='Gphb5', Aliases='OGH, Zlut1', description='glycoprotein hormone beta 5', other_designations='glycoprotein hormone beta subunit|thyrostimulin subunit beta', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=75411720, end_position_on_the_genomic_accession=75416781, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  217682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217682, CurrentID=0, Status='live', Symbol='Plekhd1', Aliases='3830431G21Rik', description='pleckstrin homology domain containing, family D (with coiled-coil domains) member 1', other_designations='PH domain-containing family D member 1', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80692581, end_position_on_the_genomic_accession=80724216, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  217684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217684, CurrentID=0, Status='live', Symbol='Susd6', Aliases='4933426M11Rik, mKIAA0247', description='sushi domain containing 6', other_designations='drago', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80790532, end_position_on_the_genomic_accession=80880833, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  217692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217692, CurrentID=0, Status='live', Symbol='Sipa1l1', Aliases='4931426N11Rik, AW213287, Spar, mKIAA0440', description='signal-induced proliferation-associated 1 like 1', other_designations='SIPA1-like protein 1', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=82169378, end_position_on_the_genomic_accession=82451786, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  68352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68352, CurrentID=0, Status='live', Symbol='Aspdh', Aliases='0610012D14Rik', description='aspartate dehydrogenase domain containing', other_designations='aspartate dehydrogenase domain-containing protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44465435, end_position_on_the_genomic_accession=44467751, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  217695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217695, CurrentID=0, Status='live', Symbol='Zfyve1', Aliases='DFCP1, G630053K23, TAFF1, ZNFN2A1, mKIAA1589', description='zinc finger, FYVE domain containing 1', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=83546556, end_position_on_the_genomic_accession=83597243, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  217698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217698, CurrentID=0, Status='live', Symbol='Acot5', Aliases='BC016076, PTE-Ic', description='acyl-CoA thioesterase 5', other_designations='peroxisomal acyl-CoA thioesterase Ic|peroxisomal acyl-coenzyme A thioester hydrolase Ic', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84069078, end_position_on_the_genomic_accession=84076024, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  217700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217700, CurrentID=0, Status='live', Symbol='Acot6', Aliases='4632408A20Rik, A330054B07, BE688602', description='acyl-CoA thioesterase 6', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84100654, end_position_on_the_genomic_accession=84109783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  217705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217705, CurrentID=0, Status='live', Symbol='Fam161b', Aliases='9330169D17, 9830169C18Rik', description='family with sequence similarity 161, member B', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84345314, end_position_on_the_genomic_accession=84361883, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  217707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217707, CurrentID=0, Status='live', Symbol='Coq6', Aliases='5930427M12Rik, AW742344', description='coenzyme Q6 monooxygenase', other_designations='coenzyme Q10 monooxygenase 6|coenzyme Q6 homolog', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84361756, end_position_on_the_genomic_accession=84373796, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  217708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217708, CurrentID=0, Status='live', Symbol='Lin52', Aliases='5830457H20Rik', description='lin-52 homolog (C. elegans)', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84451123, end_position_on_the_genomic_accession=84546515, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  217715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217715, CurrentID=0, Status='live', Symbol='Eif2b2', Aliases='EIF-2Bbeta, EIF2B', description='eukaryotic translation initiation factor 2B, subunit 2 beta', other_designations='eIF-2B GDP-GTP exchange factor subunit beta', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85219475, end_position_on_the_genomic_accession=85226628, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  217716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217716, CurrentID=0, Status='live', Symbol='Mlh3', Aliases='AV125803, BB126472', description='mutL homolog 3 (E coli)', other_designations='mutL 3', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85234466, end_position_on_the_genomic_accession=85270648, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  217718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217718, CurrentID=0, Status='live', Symbol='Nek9', Aliases='C130021H08Rik', description='NIMA (never in mitosis gene a)-related expressed kinase 9', other_designations='nercc1 kinase|never in mitosis A-related kinase 9|nimA-related protein kinase 9', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85299514, end_position_on_the_genomic_accession=85339362, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  217721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217721, CurrentID=0, Status='live', Symbol='Mfsd7c', Aliases='CCT, Flvcr2', description='major facilitator superfamily domain containing 7C', other_designations='calcium-chelate transporter|feline leukemia virus subgroup C cellular receptor family, member 2', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85746539, end_position_on_the_genomic_accession=85813585, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  217732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217732, CurrentID=0, Status='live', Symbol='Cipc', Aliases='2310044G17Rik, AI842544, AI853744, Kiaa1737, mKIAA1737', description='CLOCK interacting protein, circadian', other_designations='CLOCK-interacting circadian protein', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86947041, end_position_on_the_genomic_accession=86965364, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  217733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217733, CurrentID=0, Status='live', Symbol='Tmem63c', Aliases='4932420N09, 9330187M14Rik', description='transmembrane protein 63c', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87021344, end_position_on_the_genomic_accession=87091824, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  217734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217734, CurrentID=0, Status='live', Symbol='Pomt2', Aliases='', description='protein-O-mannosyltransferase 2', other_designations='dolichyl-phosphate-mannose--protein mannosyltransferase 2', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87106861, end_position_on_the_genomic_accession=87147902, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  217737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217737, CurrentID=0, Status='live', Symbol='Ahsa1', Aliases='BC023857, p38', description='AHA1, activator of heat shock protein ATPase 1', other_designations='AHA1, activator of heat shock 90kDa protein ATPase homolog 1|AHA1, activator of heat shock protein ATPase homolog 1', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87266484, end_position_on_the_genomic_accession=87273998, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  217779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217779, CurrentID=0, Status='live', Symbol='Lysmd1', Aliases='2610022K04Rik, AI415311', description='LysM, putative peptidoglycan-binding, domain containing 1', other_designations='', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95134088, end_position_on_the_genomic_accession=95139525, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  217826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217826, CurrentID=0, Status='live', Symbol='Kcnk13', Aliases='BB085247, F730021E22Rik, Gm1570, Gm1685', description='potassium channel, subfamily K, member 13', other_designations='THIK-1|tandem pore domain halothane-inhibited potassium channel 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=99961755, end_position_on_the_genomic_accession=100062682, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  217827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217827, CurrentID=0, Status='live', Symbol='Nrde2', Aliases='6720454P05Rik', description='nrde-2 necessary for RNA interference, domain containing', other_designations='UPF0614 protein C14orf102 homolog', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100125450, end_position_on_the_genomic_accession=100159931, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  217830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217830, CurrentID=0, Status='live', Symbol='9030617O03Rik', Aliases='', description='RIKEN cDNA 9030617O03 gene', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100779071, end_position_on_the_genomic_accession=100872610, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  217835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217835, CurrentID=0, Status='live', Symbol='Rin3', Aliases='6430500K07, C86933', description='Ras and Rab interactor 3', other_designations='Ras interaction/interference protein 3', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102282830, end_position_on_the_genomic_accession=102390854, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  217837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217837, CurrentID=0, Status='live', Symbol='Itpk1', Aliases='BC031182', description='inositol 1,3,4-triphosphate 5/6 kinase', other_designations='inositol 1,3,4-trisphosphate 5/6-kinase|inositol-triphosphate 5/6-kinase|ins(1,3,4)P(3) 5/6-kinase', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102568583, end_position_on_the_genomic_accession=102704869, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  217843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217843, CurrentID=0, Status='live', Symbol='Unc79', Aliases='9030205A07Rik, Mlca3', description='unc-79 homolog (C. elegans)', other_designations='modifier of locomotor activity 3', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102948329, end_position_on_the_genomic_accession=103184228, orientation='plus', exon_count=54, OMIM='', no_hdr0=''),
  217845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217845, CurrentID=0, Status='live', Symbol='Ifi27l2b', Aliases='1810023F06Rik, ISG12(b2)', description='interferon, alpha-inducible protein 27 like 2B', other_designations='interferon stimulated gene 12(b2)|interferon-stimulated gene 12 protein B2', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103450898, end_position_on_the_genomic_accession=103457223, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  217847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217847, CurrentID=0, Status='live', Symbol='Serpina10', Aliases='PZI, ZPI', description='serine (or cysteine) peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 10', other_designations='PZ-dependent protease inhibitor|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 10', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103616675, end_position_on_the_genomic_accession=103631490, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  217864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217864, CurrentID=0, Status='live', Symbol='Rcor1', Aliases='5730409O11, 6720480E22Rik, AU042633, D12Wsu95e, Rocr1, mKIAA0071', description='REST corepressor 1', other_designations='', map_location='12 60.93 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111038236, end_position_on_the_genomic_accession=111115901, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  217866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217866, CurrentID=0, Status='live', Symbol='Cdc42bpb', Aliases='MRCKb', description='CDC42 binding protein kinase beta', other_designations='CDC42-binding protein kinase beta|DMPK-like beta|MRCK beta|cdc42bpb|myotonic dystrophy kinase related CDC42 binding kinase beta|myotonic dystrophy kinase-related CDC42-binding kinase beta|myotonic dystrophy protein kinase like beta|myotonic dystrophy protein kinase-like beta', map_location='12 60.94 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111292972, end_position_on_the_genomic_accession=111377718, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  217869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217869, CurrentID=0, Status='live', Symbol='Eif5', Aliases='2810011H21Rik, D12Ertd549e', description='eukaryotic translation initiation factor 5', other_designations='eIF-5', map_location='12 61.03 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111538101, end_position_on_the_genomic_accession=111546753, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  217874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217874, CurrentID=0, Status='live', Symbol='Rd3l', Aliases='A930011N13', description='retinal degeneration 3-like', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111979191, end_position_on_the_genomic_accession=111981128, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  217882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217882, CurrentID=0, Status='live', Symbol='Cep170b', Aliases='AI466840, AW555464, Kiaa0284, mKIAA0284', description='centrosomal protein 170B', other_designations='', map_location='12 61.2 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112720591, end_position_on_the_genomic_accession=112746592, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  217887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217887, CurrentID=0, Status='live', Symbol='BC022687', Aliases='AW050009, C130001I08Rik, Flj20080', description='cDNA sequence BC022687', other_designations='', map_location='12 61.2 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112808975, end_position_on_the_genomic_accession=112816245, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  217893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217893, CurrentID=0, Status='live', Symbol='Pacs2', Aliases='6720425G15Rik, AW208651, AW489976, E230011J22, Pacs1l, mKIAA0602', description='phosphofurin acidic cluster sorting protein 2', other_designations='PACS-2|PACS1-like protein', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113014507, end_position_on_the_genomic_accession=113074401, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  217935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217935, CurrentID=0, Status='live', Symbol='Wdr60', Aliases='D430033N04Rik', description='WD repeat domain 60', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=116206261, end_position_on_the_genomic_accession=116263047, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  217944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217944, CurrentID=0, Status='live', Symbol='Rapgef5', Aliases='4932413M22, C86120, D030051B22Rik, GFR, Mrgef, mKIAA0277, mmr-gef', description='Rap guanine nucleotide exchange factor (GEF) 5', other_designations='M-Ras-regulated Rap GEF|MR-GEF|Mras regulated guanine nucleotide exchange factor|guanine nucleotide exchange factor for Rap1', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=117516479, end_position_on_the_genomic_accession=117759737, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  217946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217946, CurrentID=0, Status='live', Symbol='Cdca7l', Aliases='BC006933, JPO2', description='cell division cycle associated 7 like', other_designations='transcription factor RAM2', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=117843475, end_position_on_the_genomic_accession=117878712, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101861, CurrentID=0, Status='live', Symbol='Ints4', Aliases='2610034N24Rik, AA692684, AW493223', description='integrator complex subunit 4', other_designations='int4', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97480950, end_position_on_the_genomic_accession=97541398, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  217980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217980, CurrentID=0, Status='live', Symbol='Larp4b', Aliases='A630096F19, AI256361, D13Wsu64e, Larp5', description='La ribonucleoprotein domain family, member 4B', other_designations='La ribonucleoprotein domain family, member 5|la ribonucleoprotein domain family member 5|la-related protein 5', map_location='13 3.66 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=9093814, end_position_on_the_genomic_accession=9174406, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  217995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217995, CurrentID=0, Status='live', Symbol='Heatr1', Aliases='AA517551, B130016L12Rik, BC019693', description='HEAT repeat containing 1', other_designations='', map_location='13 4.56 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12395375, end_position_on_the_genomic_accession=12438893, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  232941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232941, CurrentID=0, Status='live', Symbol='Ppm1n', Aliases='A930002H20, C79127', description='protein phosphatase, Mg2+/Mn2+ dependent, 1N (putative)', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19276807, end_position_on_the_genomic_accession=19280194, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  218030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218030, CurrentID=0, Status='live', Symbol='Pou6f2', Aliases='D130006K24Rik, RPF-1', description='POU domain, class 6, transcription factor 2', other_designations='retina-derived POU-domain factor-1', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=18124960, end_position_on_the_genomic_accession=18382044, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  218035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218035, CurrentID=0, Status='live', Symbol='Vps41', Aliases='AI317346, Vam2, mVam2', description='vacuolar protein sorting 41 (yeast)', other_designations='VAM2 homolog', map_location='13 6.75 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=18717292, end_position_on_the_genomic_accession=18866811, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  218038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218038, CurrentID=0, Status='live', Symbol='Amph', Aliases='', description='amphiphysin', other_designations='', map_location='13 6.89 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=18948141, end_position_on_the_genomic_accession=19154032, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  320332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320332, CurrentID=0, Status='live', Symbol='Hist4h4', Aliases='B130044J01Rik', description='histone cluster 4, H4', other_designations='histone 4, H4', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136803993, end_position_on_the_genomic_accession=136804431, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  218066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218066, CurrentID=0, Status='live', Symbol='Olfr11', Aliases='M49, MOR256-11, ORL121', description='olfactory receptor 11', other_designations='GA_x6K02T2QHY8-11597382-11598323|odorant receptor M49|olfactory receptor 256-11', map_location='13 7.9 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21638580, end_position_on_the_genomic_accession=21639521, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  218100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218100, CurrentID=0, Status='live', Symbol='Zfp322a', Aliases='Znf322, Znf322a', description='zinc finger protein 322A', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23353103, end_position_on_the_genomic_accession=23369208, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  218103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218103, CurrentID=0, Status='live', Symbol='Slc17a2', Aliases='C730032N17Rik, NPT3', description='solute carrier family 17 (sodium phosphate), member 2', other_designations='Na(+)/PI cotransporter 3|sodium/phosphate cotransporter 3|solute carrier family 17 member 2', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23806941, end_position_on_the_genomic_accession=23825191, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  218121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218121, CurrentID=0, Status='live', Symbol='Mboat1', Aliases='9130215M02Rik, BC023845, LPEAT1, Moact1, Oact1', description='membrane bound O-acyltransferase domain containing 1', other_designations='1-acylglycerophosphoethanolamine O-acyltransferase|1-acylglycerophosphoserine O-acyltransferase|LPE acyltransferase 1|LPEAT|LPLAT 1|LPSAT|O-acyltransferase (membrane bound) domain containing 1|O-acyltransferase domain-containing protein 1|lyso-PE acyltransferase|lyso-PS acyltransferase|lysophosphatidylethanolamine acyltransferase|lysophosphatidylserine acyltransferase|membrane bound O-acyl transferase family, member 1|membrane-bound O-acyltransferase domain-containing protein 1', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=30136449, end_position_on_the_genomic_accession=30246694, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  218138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218138, CurrentID=0, Status='live', Symbol='Gmds', Aliases='BC031788, C87208', description='GDP-mannose 4, 6-dehydratase', other_designations='GDP-D-mannose dehydratase|GMD', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=31819583, end_position_on_the_genomic_accession=32338725, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  218165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218165, CurrentID=0, Status='live', Symbol='Ofcc1', Aliases='opo', description='orofacial cleft 1 candidate 1', other_designations='ojoplano|orofacial clefting chromosomal breakpoint region candidate 1 protein homolog', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=40001834, end_position_on_the_genomic_accession=40361259, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  58207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58207, CurrentID=0, Status='live', Symbol='Slc43a3', Aliases='Eeg1, SEEEG-1', description='solute carrier family 43, member 3', other_designations='embryonic epithelia gene 1 protein|embryonic epithelial gene 1|selectively expressed in embryonic epithelia protein-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84936545, end_position_on_the_genomic_accession=84958509, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  218194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218194, CurrentID=0, Status='live', Symbol='Phactr1', Aliases='9630030F18Rik, Rpel1', description='phosphatase and actin regulator 1', other_designations='RPEL repeat containing 1', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=42680571, end_position_on_the_genomic_accession=43138526, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  218203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218203, CurrentID=0, Status='live', Symbol='Mylip', Aliases='9430057C20Rik, Mir', description='myosin regulatory light chain interacting protein', other_designations='idol|inducible degrader of the LDL-receptor|myosin regulatory light chain-interacting protein', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=45389742, end_position_on_the_genomic_accession=45411940, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  218210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218210, CurrentID=0, Status='live', Symbol='Nup153', Aliases='B130015D15Rik, C88147', description='nucleoporin 153', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=46679902, end_position_on_the_genomic_accession=46727849, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  218214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218214, CurrentID=0, Status='live', Symbol='Kdm1b', Aliases='4632428N09Rik, AI482520, Aof1', description='lysine (K)-specific demethylase 1B', other_designations='amine oxidase, flavin containing 1|flavin-containing amine oxidase domain-containing protein 1|lysine-specific histone demethylase 2', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=47043471, end_position_on_the_genomic_accession=47085279, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  218215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218215, CurrentID=0, Status='live', Symbol='Rnf144b', Aliases='BC025007, E130105P19Rik, Ibrdc2, Pir2', description='ring finger protein 144B', other_designations='IBR domain containing 2|IBR domain-containing protein 2', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=47122720, end_position_on_the_genomic_accession=47247991, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  218232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218232, CurrentID=0, Status='live', Symbol='Ptpdc1', Aliases='AI843923, AW456874, Naa-1, Ptpcd1', description='protein tyrosine phosphatase domain containing 1', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48577869, end_position_on_the_genomic_accession=48625672, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  218236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218236, CurrentID=0, Status='live', Symbol='Fam120a', Aliases='C9orf10, Ossa', description='family with sequence similarity 120, member A', other_designations='family with sequence similarity 120A|oxidative stress-associated Src activator', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48879216, end_position_on_the_genomic_accession=48967877, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  218268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218268, CurrentID=0, Status='live', Symbol='Eif4e1b', Aliases='Eif4eloo, Gm273', description='eukaryotic translation initiation factor 4E family member 1B', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54783927, end_position_on_the_genomic_accession=54788459, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  218271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218271, CurrentID=0, Status='live', Symbol='B4galt7', Aliases='', description='xylosylprotein beta1,4-galactosyltransferase, polypeptide 7 (galactosyltransferase I)', other_designations='UDP-Gal:beta-GlcNAc beta-1,4-galactosyltransferase 7|UDP-galactose:beta-N-acetylglucosamine beta-1,4-galactosyltransferase 7|b4Gal-T7|beta-1,4-GalTase 7|beta4Gal-T7', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=55599901, end_position_on_the_genomic_accession=55610445, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  218275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218275, CurrentID=0, Status='live', Symbol='BC051665', Aliases='', description='cDNA sequence BC051665', other_designations='', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60781887, end_position_on_the_genomic_accession=60786364, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  218294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218294, CurrentID=0, Status='live', Symbol='Cdc14b', Aliases='2810432N10Rik, A530086E13Rik, AA472821, CDC14B3, Cdc14B1', description='CDC14 cell division cycle 14B', other_designations='CDC14 cell division cycle 14 homolog B', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64189271, end_position_on_the_genomic_accession=64275290, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  218311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218311, CurrentID=0, Status='live', Symbol='Zfp455', Aliases='3732412P20Rik, Rslcan-10', description='zinc finger protein 455', other_designations='', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67194506, end_position_on_the_genomic_accession=67209298, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  218314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218314, CurrentID=0, Status='live', Symbol='Zfp595', Aliases='A230042K10Rik, Rslcan23', description='zinc finger protein 595', other_designations='KRAB zinc finger protein|regulator of sex-limitation candidate 23', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67312998, end_position_on_the_genomic_accession=67333704, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  218333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218333, CurrentID=0, Status='live', Symbol='Ice1', Aliases='C77245, mKIAA0947', description='interactor of little elongation complex ELL subunit 1', other_designations='interactor of little elongator complex ELL subunit 1', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=70588688, end_position_on_the_genomic_accession=70637838, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  218335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218335, CurrentID=0, Status='live', Symbol='Clptm1l', Aliases='C130052I12Rik, Crr9', description='CLPTM1-like', other_designations='CLPTM1-like protein|cisplatin resistance related protein CRR9p', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73604002, end_position_on_the_genomic_accession=73620639, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  218341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218341, CurrentID=0, Status='live', Symbol='Rfesd', Aliases='AI256775, D030068K09', description='Rieske (Fe-S) domain containing', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=76001535, end_position_on_the_genomic_accession=76018712, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  218343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218343, CurrentID=0, Status='live', Symbol='Ttc37', Aliases='mKIAA0372', description='tetratricopeptide repeat domain 37', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=76098730, end_position_on_the_genomic_accession=76190323, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  100503085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503085, CurrentID=0, Status='live', Symbol='Klhl3', Aliases='7530408C15Rik, AI430941, EG627648', description='kelch-like 3', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58004439, end_position_on_the_genomic_accession=58123481, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  218397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218397, CurrentID=0, Status='live', Symbol='Rasa1', Aliases='Gap, RasGAP, Rasa', description='RAS p21 protein activator 1', other_designations='GTPase-activating protein|p120-rasGAP', map_location='13 45.05 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=85214699, end_position_on_the_genomic_accession=85289493, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  58250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58250, CurrentID=0, Status='live', Symbol='Chst11', Aliases='1110020P09Rik, C4ST, C4ST-1, C4ST1, C4s', description='carbohydrate sulfotransferase 11', other_designations='C4S-1|chondroitin 4|chondroitin 4-O-sulfotransferase 1|chondroitin 4-sulfotransferase 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82985380, end_position_on_the_genomic_accession=83195907, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  218440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218440, CurrentID=0, Status='live', Symbol='Ankrd34b', Aliases='6430502M16Rik, DP58', description='ankyrin repeat domain 34B', other_designations='cytosolic phosphoprotein DP58|dendritic phosphoprotein 58 kDa|dendritic protein of 58 kDa', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92425641, end_position_on_the_genomic_accession=92441658, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  218441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218441, CurrentID=0, Status='live', Symbol='Zfyve16', Aliases='AI035632, B130024H06Rik, B130031L15, mKIAA0305', description='zinc finger, FYVE domain containing 16', other_designations='endofin|endosomal-associated FYVE domain protein', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92487104, end_position_on_the_genomic_accession=92532710, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  218442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218442, CurrentID=0, Status='live', Symbol='Serinc5', Aliases='A130038L21Rik, AIGP3, C86123, TPO1', description='serine incorporator 5', other_designations='AIGP-3|axotomy-induced glycoprotein 3', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92611138, end_position_on_the_genomic_accession=92711947, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  218454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218454, CurrentID=0, Status='live', Symbol='Lhfpl2', Aliases='6030465B15, AI447312, AW050335, mKIAA0206, vgim', description='lipoma HMGIC fusion partner-like 2', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=94057656, end_position_on_the_genomic_accession=94195410, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  218460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218460, CurrentID=0, Status='live', Symbol='Wdr41', Aliases='B830029I03Rik, MSTP048', description='WD repeat domain 41', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=94976213, end_position_on_the_genomic_accession=95023316, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  218461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218461, CurrentID=0, Status='live', Symbol='Pde8b', Aliases='B230331L10Rik, C030047E14Rik', description='phosphodiesterase 8B', other_designations='cell proliferation-inducing gene 22 protein', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95024105, end_position_on_the_genomic_accession=95250397, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  218476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218476, CurrentID=0, Status='live', Symbol='Gcnt4', Aliases='C2gnt3, Gm279, Gm73', description='glucosaminyl (N-acetyl) transferase 4, core 2 (beta-1,6-N-acetylglucosaminyltransferase)', other_designations='core 2 beta-1,6-N-acetylglucosaminyltransferase 3|core 2-branching enzyme 3|core2-GlcNAc-transferase 3|glucosaminyl (N-acetyl) transferase 3, mucin type pseudogene', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=96924352, end_position_on_the_genomic_accession=96950914, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  407785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407785, CurrentID=0, Status='live', Symbol='Ndufs6', Aliases='BC059730, IP13', description='NADH dehydrogenase (ubiquinone) Fe-S protein 6', other_designations='CI-13kD-A|NADH dehydrogenase Fe-S protein 6|NADH-ubiquinone oxidoreductase 13 kDa-A subunit|complex I-13kD-A', map_location='13 D2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73319876, end_position_on_the_genomic_accession=73328482, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  218490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218490, CurrentID=0, Status='live', Symbol='Btf3', Aliases='1700054E11Rik', description='basic transcription factor 3', other_designations='NAC-beta|RNA polymerase B transcription factor 3|nascent polypeptide-associated complex subunit beta', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98309897, end_position_on_the_genomic_accession=98317006, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  407786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407786, CurrentID=0, Status='live', Symbol='Taf9b', Aliases='BC066223, Taf9l', description='TATA-box binding protein associated factor 9B', other_designations='TAF9-like RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF9B RNA polymerase II, TATA box binding protein (TBP)-associated factor|transcription-associated factor TAFII31L', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=106206874, end_position_on_the_genomic_accession=106221158, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  218503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218503, CurrentID=0, Status='live', Symbol='Fcho2', Aliases='5832424M12Rik, AA387320', description='FCH domain only 2', other_designations='FCH domain only protein 2', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98723406, end_position_on_the_genomic_accession=98815769, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  218506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218506, CurrentID=0, Status='live', Symbol='Mrps27', Aliases='2610028H14Rik', description='mitochondrial ribosomal protein S27', other_designations='MRP-S27|S27mt|mrps27', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=99344767, end_position_on_the_genomic_accession=99415562, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  218518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218518, CurrentID=0, Status='live', Symbol='Marveld2', Aliases='BC003296, MARVD2, Mrvldc2, Tric, Trica, Tricb, Tricc', description='MARVEL (membrane-associating) domain containing 2', other_designations='tricellulin', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100595957, end_position_on_the_genomic_accession=100616971, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  218543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218543, CurrentID=0, Status='live', Symbol='Srek1', Aliases='8430401B01, AI450757, AI462342, SRrp508, SRrp86, Sfrs12, Srsf12', description='splicing regulatory glutamine/lysine-rich protein 1', other_designations='serine/arginine-rich splicing factor 12|serine/arginine-rich-splicing regulatory protein 86', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=103739348, end_position_on_the_genomic_accession=103774603, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  218544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218544, CurrentID=0, Status='live', Symbol='Sgtb', Aliases='C630001O05Rik', description='small glutamine-rich tetratricopeptide repeat (TPR)-containing, beta', other_designations='beta-SGT', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104109724, end_position_on_the_genomic_accession=104142492, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  407800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407800, CurrentID=0, Status='live', Symbol='Ecm2', Aliases='9030618O22Rik, BC065151, tenonectin', description='extracellular matrix protein 2, female organ and adipocyte specific', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49504810, end_position_on_the_genomic_accession=49532789, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  218581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218581, CurrentID=0, Status='live', Symbol='Depdc1b', Aliases='9830132O03, AW260467, XTP1', description='DEP domain containing 1B', other_designations='', map_location='13 D2.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=108316014, end_position_on_the_genomic_accession=108389568, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  218613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218613, CurrentID=0, Status='live', Symbol='Mier3', Aliases='5730509D14, AU015422, D130064H19Rik', description='MIER family member 3', other_designations='mesoderm induction early response 1, family member 3|mi-er3', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=111686157, end_position_on_the_genomic_accession=111718596, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  218624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218624, CurrentID=0, Status='live', Symbol='Il31ra', Aliases='CRL3, GLM-R, GPL, Glmr, NR10', description='interleukin 31 receptor A', other_designations='IL-31 receptor subunit alpha|IL-31R subunit alpha|IL-31R-alpha|IL-31RA|class I cytokine receptor|cytokine receptor NR10|gp130-like monocyte receptor|gp130-like receptor|interleukin 31RA|mGLM-R|novel cytokine receptor 10|zcytoR17', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112521849, end_position_on_the_genomic_accession=112591399, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  218629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218629, CurrentID=0, Status='live', Symbol='Dhx29', Aliases='3732415M03, AU043276, Ddxx, E130202M19Rik', description='DEAH (Asp-Glu-Ala-His) box polypeptide 29', other_designations='DEAH box protein 29|nucleic acid helicase DDXx', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112927530, end_position_on_the_genomic_accession=112969187, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  218630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218630, CurrentID=0, Status='live', Symbol='Ccno', Aliases='C86987, Ccnu, E130318K11, Ung2', description='cyclin O', other_designations='UDG2|cyclin U|cyclin-like uracil-DNA glycosylase|uracil DNA glycosylase 2|uracil-DNA glycosylase 2', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112987802, end_position_on_the_genomic_accession=112990778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  218639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218639, CurrentID=0, Status='live', Symbol='Arl15', Aliases='A430036I03, Arfrp2, C230032K13Rik', description='ADP-ribosylation factor-like 15', other_designations='ADP-ribosylation factor related protein 2|ADP-ribosylation factor-related protein 2|ARF-related protein 2', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=113794506, end_position_on_the_genomic_accession=114157462, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  276742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276742, CurrentID=0, Status='live', Symbol='Taar7e', Aliases='Gm697, taR-7e', description='trace amine-associated receptor 7E', other_designations='mTaar7e', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24037614, end_position_on_the_genomic_accession=24038690, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  407819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407819, CurrentID=0, Status='live', Symbol='BC031181', Aliases='', description='cDNA sequence BC031181', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=75005274, end_position_on_the_genomic_accession=75009934, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  218693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218693, CurrentID=0, Status='live', Symbol='Paip1', Aliases='PAIP-1', description='polyadenylate binding protein-interacting protein 1', other_designations='PABP-interacting protein 1|poly(A)-binding protein-interacting protein 1', map_location='13 D2.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119428271, end_position_on_the_genomic_accession=119460323, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  218695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218695, CurrentID=0, Status='live', Symbol='Gm10044', Aliases='ENSMUSG00000058570', description='predicted gene 10044', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=7718000, end_position_on_the_genomic_accession=7792342, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  218699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218699, CurrentID=0, Status='live', Symbol='Pxk', Aliases='C230080L11Rik, D14Ertd813e, MONaKA', description='PX domain containing serine/threonine kinase', other_designations='modulator of Na,K-ATPase|pxk', map_location='14 4.7 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8098191, end_position_on_the_genomic_accession=8165117, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  407821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407821, CurrentID=0, Status='live', Symbol='Znrf3', Aliases='Gm1167', description='zinc and ring finger 3', other_designations='zinc/RING finger protein 3', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5276329, end_position_on_the_genomic_accession=5444847, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  407823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=407823, CurrentID=0, Status='live', Symbol='Baz2b', Aliases='5830435C13Rik, BC053917, D2Ertd794e', description='bromodomain adjacent to zinc finger domain, 2B', other_designations='', map_location='2 34.18 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=59899363, end_position_on_the_genomic_accession=60209846, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  218734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218734, CurrentID=0, Status='live', Symbol='3830406C13Rik', Aliases='', description='RIKEN cDNA 3830406C13 gene', other_designations='', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=12284153, end_position_on_the_genomic_accession=12322485, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  218739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218739, CurrentID=0, Status='live', Symbol='Sntn', Aliases='A430083B19Rik', description='sentan, cilia apical structure protein', other_designations='', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=13670876, end_position_on_the_genomic_accession=13683148, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101994, CurrentID=0, Status='live', Symbol='Champ1', Aliases='AA675043, AI116001, D8Ertd457e, D8Ertd569e, Zfp828, Znf828, mKIAA1802', description='chromosome alignment maintaining phosphoprotein 1', other_designations='zinc finger protein 828', map_location='8 6.4 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13869641, end_position_on_the_genomic_accession=13881639, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  218756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218756, CurrentID=0, Status='live', Symbol='Slc4a7', Aliases='E430014N10Rik, NBC3, NBCn1, NBCn1-G', description='solute carrier family 4, sodium bicarbonate cotransporter, member 7', other_designations='Na-bicarbonate cotransporter|sodium bicarbonate cotransporter NBCn1-A|sodium bicarbonate cotransporter NBCn1-B|sodium bicarbonate cotransporter NBCn1-D|sodium bicarbonate cotransporter NBCn1-H|sodium bicarbonate cotransporter NBCn1-c', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=14703011, end_position_on_the_genomic_accession=14799944, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  218763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218763, CurrentID=0, Status='live', Symbol='Lrrc3b', Aliases='BC019794, LRP15', description='leucine rich repeat containing 3B', other_designations='leucine-rich repeat protein LRP15', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=15357515, end_position_on_the_genomic_accession=15438987, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  218772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218772, CurrentID=0, Status='live', Symbol='Rarb', Aliases='A830025K23, Hap, Nr1b2', description='retinoic acid receptor, beta', other_designations='RAR-beta|nuclear receptor subfamily 1 group B member 2', map_location='14 A1-A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=16430839, end_position_on_the_genomic_accession=17082331, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  218793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218793, CurrentID=0, Status='live', Symbol='Ube2e2', Aliases='BC016265', description='ubiquitin-conjugating enzyme E2E 2', other_designations='E2 ubiquitin-conjugating enzyme E2|TBC1 domain family, member 12|ubiquitin carrier protein E2|ubiquitin-conjugating enzyme E2E 2 (UBC4/5 homolog, yeast)|ubiquitin-protein ligase E2', map_location='14 A2-A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=18573563, end_position_on_the_genomic_accession=18894262, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  218811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218811, CurrentID=0, Status='live', Symbol='Sec24c', Aliases='2610204K03Rik, mKIAA0079', description='Sec24 related gene family, member C (S. cerevisiae)', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20674285, end_position_on_the_genomic_accession=20694852, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  218820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218820, CurrentID=0, Status='live', Symbol='Zfp503', Aliases='AI181838, B830002A16Rik, Nolz-1, Nolz1, Znf503', description='zinc finger protein 503', other_designations='zinc finger protein Nolz-1|zinc-finger protein NOLZ1', map_location='14 12.0 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=21983962, end_position_on_the_genomic_accession=21989601, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  276770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276770, CurrentID=0, Status='live', Symbol='Eif5a', Aliases='AA410058, D19Wsu54e, Eif4d1, eIF-4D, eIF-5A, eIF-5A-1, eIF-5A1, Eif5a', description='eukaryotic translation initiation factor 5A', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69916712, end_position_on_the_genomic_accession=69921958, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  218832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218832, CurrentID=0, Status='live', Symbol='Polr3a', Aliases='9330175N20Rik, BC053071, RPC1, RPC155', description='polymerase (RNA) III (DNA directed) polypeptide A', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=24448694, end_position_on_the_genomic_accession=24487186, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  218850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218850, CurrentID=0, Status='live', Symbol='Fam208a', Aliases='4732418E12, 4933409E02Rik, D14Abb1e, MommeD6', description='family with sequence similarity 208, member A', other_designations='protein FAM208A|retinoblastoma-associated protein 140|transgene activation suppressor protein', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=27426658, end_position_on_the_genomic_accession=27483555, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  218865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218865, CurrentID=0, Status='live', Symbol='Chdh', Aliases='AU041472, D630034H06Rik', description='choline dehydrogenase', other_designations='CDH|CHD', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=30009000, end_position_on_the_genomic_accession=30040466, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100415785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100415785, CurrentID=0, Status='live', Symbol='Gm11559', Aliases='OTTMUSG00000002191', description='predicted gene 11559', other_designations='', map_location='11 63.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99864476, end_position_on_the_genomic_accession=99865571, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  218877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218877, CurrentID=0, Status='live', Symbol='Sema3g', Aliases='AK129018', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3G', other_designations='Semaphorin 3G', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31217873, end_position_on_the_genomic_accession=31230352, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  218885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218885, CurrentID=0, Status='live', Symbol='Oxnad1', Aliases='2410002F01Rik, AW492292, BC019806', description='oxidoreductase NAD-binding domain containing 1', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32085728, end_position_on_the_genomic_accession=32103203, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  218914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218914, CurrentID=0, Status='live', Symbol='Wapl', Aliases='A530089A20Rik, BC037674, DIF-2, FOE, Wapal', description='WAPL cohesin release factor', other_designations='dioxin-inducible factor 2|friend of EBNA2|wings apart-like homolog', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34671487, end_position_on_the_genomic_accession=34747983, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  218921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218921, CurrentID=0, Status='live', Symbol='4930474N05Rik', Aliases='', description='RIKEN cDNA 4930474N05 gene', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=36094942, end_position_on_the_genomic_accession=36097401, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  242608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242608, CurrentID=0, Status='live', Symbol='Podn', Aliases='9430070G18, Pcan, SLRR5A', description='podocan', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108014793, end_position_on_the_genomic_accession=108032090, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  218952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218952, CurrentID=0, Status='live', Symbol='Fermt2', Aliases='AA960555, Kindlin-2, Mig2, Plekhc1', description='fermitin family member 2', other_designations='mitogen inducible 2|pleckstrin homology domain containing, family C (with FERM domain) member 1|pleckstrin homology domain-containing family C member 1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45458792, end_position_on_the_genomic_accession=45530113, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  218973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218973, CurrentID=0, Status='live', Symbol='Wdhd1', Aliases='A630092E18, AND-1, D630024B06Rik', description='WD repeat and HMG-box DNA binding protein 1', other_designations='acidic nucleoplasmic DNA-binding protein 1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47240944, end_position_on_the_genomic_accession=47276857, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  218975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218975, CurrentID=0, Status='live', Symbol='Mapk1ip1l', Aliases='9130202B12, C130032J12Rik', description='mitogen-activated protein kinase 1 interacting protein 1-like', other_designations='mitogen activated protein kinase 1 interacting protein 1-like|mitogen-activated protein kinase 1-interacting protein 1-like', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47297880, end_position_on_the_genomic_accession=47323091, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  218977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218977, CurrentID=0, Status='live', Symbol='Dlgap5', Aliases='C77459, C86398, Dap-5, Dlg7, Hurp, mKIAA0008', description='discs, large (Drosophila) homolog-associated protein 5', other_designations='discs, large homolog 7|discs, large homolog-associated protein 5|hepatoma up-regulated protein homolog', map_location='14 24.6 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47387779, end_position_on_the_genomic_accession=47418407, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  218989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=218989, CurrentID=0, Status='live', Symbol='Tmem260', Aliases='5930435P15, 6720456H20Rik', description='transmembrane protein 260', other_designations='UPF0679 protein C14orf101 homolog', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=48446128, end_position_on_the_genomic_accession=48520708, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  219019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219019, CurrentID=0, Status='live', Symbol='Olfr743', Aliases='GA_x6K02T2PMLR-6243196-6244132, MOR106-14, MOR106-8P, Olfr264, Olfr264-ps1, Olfr265, Olfr265-ps1-ps1, Olfr743', description='olfactory receptor 743', other_designations='GA_x6K02T2N6FY-2320-2039|GA_x6K02T2N6FY-3870-3385|olfactory receptor 264, pseudogene 1|olfactory receptor 265, pseudogene 1|olfactory receptor 743, pseudogene 1|olfactory receptor MOR106-8P|olfactory receptor Olfr265', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50533414, end_position_on_the_genomic_accession=50534349, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  219022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219022, CurrentID=0, Status='live', Symbol='Ttc5', Aliases='', description='tetratricopeptide repeat domain 5', other_designations='TPR repeat protein 5|stress-responsive activator of p300', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50765409, end_position_on_the_genomic_accession=50785520, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  219024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219024, CurrentID=0, Status='live', Symbol='Tmem55b', Aliases='BC039161', description='transmembrane protein 55b', other_designations='PtdIns-4,5-P(2) 4-phosphatase type I|ptdIns-4,5-P2 4-Ptase I|type 1 PtdIns-4,5-P2 4-Ptase|type I phosphatidylinositol 4,5-bisphosphate 4-phosphatase', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50926070, end_position_on_the_genomic_accession=50930905, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  219026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219026, CurrentID=0, Status='live', Symbol='Eddm3b', Aliases='Fam12, Gm75', description='epididymal protein 3B', other_designations='epididymal secretory protein 3|family with sequence similarity 12', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51114370, end_position_on_the_genomic_accession=51117815, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  219033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219033, CurrentID=0, Status='live', Symbol='Ang4', Aliases='Raa4, Rnase5d', description='angiogenin, ribonuclease A family, member 4', other_designations='ribonuclease A a4', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51763892, end_position_on_the_genomic_accession=51773590, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  219038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219038, CurrentID=0, Status='live', Symbol='Tppp2', Aliases='Gm1790, Gm77', description='tubulin polymerization-promoting protein family member 2', other_designations='p25 beta', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51918425, end_position_on_the_genomic_accession=51920704, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  219065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219065, CurrentID=0, Status='live', Symbol='A630038E17Rik', Aliases='A430070I15', description='RIKEN cDNA A630038E17 gene', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  219072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219072, CurrentID=0, Status='live', Symbol='Haus4', Aliases='9430093H08Rik, AU016300, D14Ertd500e', description='HAUS augmin-like complex, subunit 4', other_designations='', map_location='14 27.83 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54541785, end_position_on_the_genomic_accession=54554361, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  219094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219094, CurrentID=0, Status='live', Symbol='Khnyn', Aliases='6330555F21, 9130227C08Rik, AI605259, Kiaa0323, mKIAA0323', description='KH and NYN domain containing', other_designations='KH and NYN domain-containing protein', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55884964, end_position_on_the_genomic_accession=55898780, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  219103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219103, CurrentID=0, Status='live', Symbol='Cenpj', Aliases='4932437H03Rik, CPAP, Gm81, LAP, LIP1, Sas4', description='centromere protein J', other_designations='CENP-J', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56526761, end_position_on_the_genomic_accession=56575571, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  219105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219105, CurrentID=0, Status='live', Symbol='Zmym5', Aliases='9830124H08Rik', description='zinc finger, MYM-type 5', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56790585, end_position_on_the_genomic_accession=56811896, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  219114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219114, CurrentID=0, Status='live', Symbol='Ska3', Aliases='F630043A04Rik', description='spindle and kinetochore associated complex subunit 3', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57806560, end_position_on_the_genomic_accession=57826190, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  219131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219131, CurrentID=0, Status='live', Symbol='Phf11a', Aliases='4933417L10Rik, Phf11, Phf11l', description='PHD finger protein 11A', other_designations='PHD finger protein 11', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59276913, end_position_on_the_genomic_accession=59297522, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  219132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219132, CurrentID=0, Status='live', Symbol='Phf11d', Aliases='D14Ertd668e, Phf11, Phf11l', description='PHD finger protein 11D', other_designations='PHD finger protein 11 family member', map_location='14 31.48 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59347407, end_position_on_the_genomic_accession=59370520, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  219134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219134, CurrentID=0, Status='live', Symbol='Shisa2', Aliases='9430059P22Rik, MAd2, Tmem46, mShisa, shisa', description='shisa family member 2', other_designations='shisa homolog 2|shisa-like protein 2|transmembrane protein 46', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59625281, end_position_on_the_genomic_accession=59631660, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  219135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219135, CurrentID=0, Status='live', Symbol='Mtmr6', Aliases='4022440C11Rik, AI428804, AU041072', description='myotubularin related protein 6', other_designations='mtmr6', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=60265183, end_position_on_the_genomic_accession=60302371, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  219140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219140, CurrentID=0, Status='live', Symbol='Spata13', Aliases='C79050, ESTM11, X83327, mFLJ00298', description='spermatogenesis associated 13', other_designations='APC-stimulated guanine nucleotide exchange factor 2|asef2', map_location='14 31.77 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=60634164, end_position_on_the_genomic_accession=60764556, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  219144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219144, CurrentID=0, Status='live', Symbol='Arl11', Aliases='ARLTS1, C730007L20Rik', description='ADP-ribosylation factor-like 11', other_designations='ADP-ribosylation factor-like tumor suppressor protein 1', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61306288, end_position_on_the_genomic_accession=61311937, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  219148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219148, CurrentID=0, Status='live', Symbol='Fam167a', Aliases='A030013D21', description='family with sequence similarity 167, member A', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63436394, end_position_on_the_genomic_accession=63465502, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  219149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219149, CurrentID=0, Status='live', Symbol='Xkr6', Aliases='BC024502, Xrg6', description='X Kell blood group precursor related family member 6 homolog', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63606530, end_position_on_the_genomic_accession=63820411, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  219150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219150, CurrentID=0, Status='live', Symbol='Hmbox1', Aliases='AI451877, AI604847, F830020C16Rik', description='homeobox containing 1', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64811600, end_position_on_the_genomic_accession=64949906, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  219151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219151, CurrentID=0, Status='live', Symbol='Scara3', Aliases='APC7, C130058N24Rik, CSR, CSR1, MSLR1, MSRL1', description='scavenger receptor class A, member 3', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65919395, end_position_on_the_genomic_accession=65953744, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  219158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219158, CurrentID=0, Status='live', Symbol='Ccar2', Aliases='2610301G19Rik, mKIAA1967', description='cell cycle activator and apoptosis regulator 2', other_designations='cell division cycle and apoptosis regulator protein 2', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70138168, end_position_on_the_genomic_accession=70153853, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  219170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219170, CurrentID=0, Status='live', Symbol='Fam216b', Aliases='9330158N17, AU021034', description='family with sequence similarity 216, member B', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=78081024, end_position_on_the_genomic_accession=78089007, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  219181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219181, CurrentID=0, Status='live', Symbol='Akap11', Aliases='6330501D17Rik, AKAP220, Gm80, PRKA11, mKIAA0629', description='A kinase (PRKA) anchor protein 11', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=78492246, end_position_on_the_genomic_accession=78536982, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  276829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276829, CurrentID=0, Status='live', Symbol='Smtnl2', Aliases='D130058I21Rik', description='smoothelin-like 2', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72389163, end_position_on_the_genomic_accession=72411713, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  219189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219189, CurrentID=0, Status='live', Symbol='Vwa8', Aliases='1300010F03Rik, 4932416F07Rik, AI957255, Kiaa0564, mKIAA0564', description='von Willebrand factor A domain containing 8', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=78849051, end_position_on_the_genomic_accession=79202310, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  219228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219228, CurrentID=0, Status='live', Symbol='Pcdh17', Aliases='C030033F14Rik, Gm78', description='protocadherin 17', other_designations='protocadherin 68', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=84443563, end_position_on_the_genomic_accession=84537565, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  68673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68673, CurrentID=0, Status='live', Symbol='Krtap4-2', Aliases='1110033F04Rik, AI604442', description='keratin associated protein 4-2', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99634114, end_position_on_the_genomic_accession=99635084, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102075, CurrentID=0, Status='live', Symbol='Plekhg4', Aliases='4931414L13Rik, AI448583, mFLJ00068', description='pleckstrin homology domain containing, family G (with RhoGef domain) member 4', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105369441, end_position_on_the_genomic_accession=105382862, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  219249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219249, CurrentID=0, Status='live', Symbol='Tdrd3', Aliases='4732418C03Rik, 6720468N18', description='tudor domain containing 3', other_designations='', map_location='14 E1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=87416583, end_position_on_the_genomic_accession=87545508, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  219257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=219257, CurrentID=0, Status='live', Symbol='Pcdh20', Aliases='C630015B17Rik, Pcdh13', description='protocadherin 20', other_designations='protocadherin 13', map_location='14 E1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=88464747, end_position_on_the_genomic_accession=88471396, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  276846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276846, CurrentID=0, Status='live', Symbol='Pigs', Aliases='BC058979, Gm393, Gm689', description='phosphatidylinositol glycan anchor biosynthesis, class S', other_designations='phosphatidylinositol glycan, class S homolog|phosphatidylinositol-glycan biosynthesis class S protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78328422, end_position_on_the_genomic_accession=78342781, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  276852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276852, CurrentID=0, Status='live', Symbol='D11Wsu47e', Aliases='', description='DNA segment, Chr 11, Wayne State University 47, expressed', other_designations='', map_location='11 79.15 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113684400, end_position_on_the_genomic_accession=113727444, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  626391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626391, CurrentID=0, Status='live', Symbol='Zfp951', Aliases='C230055K05Rik', description='zinc finger protein 951', other_designations='', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104813168, end_position_on_the_genomic_accession=104860068, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102115, CurrentID=0, Status='live', Symbol='Dohh', Aliases='1110033C18Rik, AI585913, AL022869, Hlrc1', description='deoxyhypusine hydroxylase/monooxygenase', other_designations='DOHH|HEAT-like (PBS lyase) repeat containing 1|deoxyhypusine dioxygenase|deoxyhypusine monooxygenase', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81384428, end_position_on_the_genomic_accession=81388352, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  11674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=11674, CurrentID=0, Status='live', Symbol='Aldoa', Aliases='Aldo-1, Aldo1', description='aldolase A, fructose-bisphosphate', other_designations='aldolase 1, A isoform|muscle-type aldolase', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126795234, end_position_on_the_genomic_accession=126809313, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  626415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626415, CurrentID=0, Status='live', Symbol='4930467E23Rik', Aliases='ENSMUSG00000074453, OTTMUSG00000018948, TSAP', description='RIKEN cDNA 4930467E23 gene', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19729576, end_position_on_the_genomic_accession=19753603, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  14747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14747, CurrentID=0, Status='live', Symbol='Cmklr1', Aliases='ChemR23, DEZ, Gpcr27, mcmklr1', description='chemokine-like receptor 1', other_designations='G-protein coupled receptor 27|G-protein coupled receptor DEZ|chemerin receptor|chemokine receptor-like 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113612355, end_position_on_the_genomic_accession=113650450, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  276905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276905, CurrentID=0, Status='live', Symbol='Armc7', Aliases='A630084C13, BC023126', description='armadillo repeat containing 7', other_designations='A630084C13rik', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115475638, end_position_on_the_genomic_accession=115490468, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  276919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276919, CurrentID=0, Status='live', Symbol='Gemin4', Aliases='4932415L08Rik, AU015359', description='gem (nuclear organelle) associated protein 4', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76210571, end_position_on_the_genomic_accession=76217572, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  276920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276920, CurrentID=0, Status='live', Symbol='Ccdc42', Aliases='A530001H01Rik', description='coiled-coil domain containing 42', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68587037, end_position_on_the_genomic_accession=68597954, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  233231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233231, CurrentID=0, Status='live', Symbol='Mrgprb1', Aliases='MrgB1', description='MAS-related GPR, member B1', other_designations='MrgB1 G protein-coupled receptor', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48444113, end_position_on_the_genomic_accession=48456342, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  276950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=276950, CurrentID=0, Status='live', Symbol='Slfn8', Aliases='', description='schlafen 8', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83002158, end_position_on_the_genomic_accession=83021994, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  320662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320662, CurrentID=0, Status='live', Symbol='Casc1', Aliases='A230084G12Rik, Las1', description='cancer susceptibility candidate 1', other_designations='cancer susceptibility candidate gene 1 protein homolog|cancer susceptibility candidate protein 1|lung adenoma susceptibility 1|lung adenoma susceptibility protein 1', map_location='6 77.37 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=145174872, end_position_on_the_genomic_accession=145210970, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  102209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102209, CurrentID=0, Status='live', Symbol='Snapc2', Aliases='0610007H10Rik, AU015675', description='small nuclear RNA activating complex, polypeptide 2', other_designations='SNAPc 45 kDa subunit|SNAPc subunit 2|small nuclear RNA-activating complex polypeptide 2|snRNA-activating protein complex 45 kDa subunit', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4253102, end_position_on_the_genomic_accession=4256220, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  320664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=320664, CurrentID=0, Status='live', Symbol='Cass4', Aliases='F730031O20Rik', description='Cas scaffolding protein family member 4', other_designations='cass4|likely orthologue of H. sapiens chromosome 20 open reading frame 32 (C20orf32)', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172393683, end_position_on_the_genomic_accession=172433757, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  408058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408058, CurrentID=0, Status='live', Symbol='BC048507', Aliases='', description='cDNA sequence BC048507', other_designations='', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67863326, end_position_on_the_genomic_accession=67863923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  408059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408059, CurrentID=0, Status='live', Symbol='BC049352', Aliases='', description='cDNA sequence BC049352', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45181508, end_position_on_the_genomic_accession=45251370, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  408068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=408068, CurrentID=0, Status='live', Symbol='Zfp738', Aliases='3830402I07Rik, 6720487G11Rik', description='zinc finger protein 738', other_designations='Novel Zinc finger, C2H2 type containing protein (3830402I07Rik)', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67667437, end_position_on_the_genomic_accession=67724322, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  14862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14862, CurrentID=0, Status='live', Symbol='Gstm1', Aliases='Gstb-1, Gstb1', description='glutathione S-transferase, mu 1', other_designations='GST 1-1|GST class-mu 1|glutathione S-transferase GT8.7|glutathione-S-transferase, mu 1|pmGT10', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108012250, end_position_on_the_genomic_accession=108017973, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  277010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277010, CurrentID=0, Status='live', Symbol='Marveld1', Aliases='AI504298, Mrvldc1', description='MARVEL (membrane-associating) domain containing 1', other_designations='putative MARVEL domain-containing protein 1', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42147389, end_position_on_the_genomic_accession=42151703, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  102334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102334, CurrentID=0, Status='live', Symbol='Ankrd10', Aliases='4833425P12Rik, AW549277', description='ankyrin repeat domain 10', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=11611581, end_position_on_the_genomic_accession=11636136, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  89867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=89867, CurrentID=0, Status='live', Symbol='Sec16b', Aliases='Lztr2, Rgpr, Rgpr-p117', description='SEC16 homolog B (S. cerevisiae)', other_designations='leucine zipper transcription regulator 2|protein SEC16 homolog B|regucalcin gene promoter region-related protein p117|regucalcin gene promotor region related protein', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=157506733, end_position_on_the_genomic_accession=157568425, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  14980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=14980, CurrentID=0, Status='live', Symbol='H2-L', Aliases='H-2L', description='histocompatibility 2, D region locus L', other_designations='histocompatibility 2, L region', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  626708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626708, CurrentID=0, Status='live', Symbol='Defa26', Aliases='Defcr26', description='defensin, alpha, 26', other_designations='defensin-related cryptdin-26|testis defensin', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21618168, end_position_on_the_genomic_accession=21618972, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  15039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15039, CurrentID=0, Status='live', Symbol='H2-T22', Aliases='27b, H-2T17, H-2T22, H2-T17, T22b', description='histocompatibility 2, T region locus 22', other_designations='MHC classIb T22|histocompatibility 2, T region locus 17', map_location='17 18.87 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36037133, end_position_on_the_genomic_accession=36042742, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  58807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58807, CurrentID=0, Status='live', Symbol='Slco1c1', Aliases='OATP-14, OATP-F, Oatp2, Oatpf, Slc21a14', description='solute carrier organic anion transporter family, member 1c1', other_designations='organic anion transporter 2|organic anion transporter F|organic anion-transporting polypeptide 14|solute carrier family 21 (organic anion transporter), member 14|thyroxine transporter', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=141524378, end_position_on_the_genomic_accession=141570177, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  68490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68490, CurrentID=0, Status='live', Symbol='Zfp579', Aliases='1110003A17Rik, Znf579', description='zinc finger protein 579', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4992852, end_position_on_the_genomic_accession=4996101, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  353025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353025, CurrentID=0, Status='live', Symbol='Caps2', Aliases='D630005B03Rik', description='calcyphosphine 2', other_designations='calcyphosin 2|calcyphosine 2|calcyphosine-2', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=112163346, end_position_on_the_genomic_accession=112216561, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  353047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353047, CurrentID=0, Status='live', Symbol='Plekhm1', Aliases='AP162, B2, BC038943, D330036J23Rik', description='pleckstrin homology domain containing, family M (with RUN domain) member 1', other_designations='PH domain-containing family M member 1', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103365092, end_position_on_the_genomic_accession=103412664, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  353130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353130, CurrentID=0, Status='live', Symbol='Prss33', Aliases='Eos, mT6, tryptase-6', description='protease, serine 33', other_designations='protease, serine, 33|serine protease EOS|tryptase 6', map_location='17 12.08 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23833338, end_position_on_the_genomic_accession=23836126, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  353148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353148, CurrentID=0, Status='live', Symbol='Tas2r139', Aliases='Tas2r39, mt2r34', description='taste receptor, type 2, member 139', other_designations='T2R39|candidate taste receptor mt2r34', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42140936, end_position_on_the_genomic_accession=42141895, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  68498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68498, CurrentID=0, Status='live', Symbol='Tspan11', Aliases='1110014F12Rik, AU067806', description='tetraspanin 11', other_designations='tspan-11', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127887622, end_position_on_the_genomic_accession=127953031, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  353155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353155, CurrentID=0, Status='live', Symbol='Gjd3', Aliases='Gja11, Gjc1, cx30.2', description='gap junction protein, delta 3', other_designations='connexin 30.2|gap junction alpha-11 protein|gap junction chi-1 protein|gap junction membrane channel protein chi 1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98982180, end_position_on_the_genomic_accession=98983016, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  353156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353156, CurrentID=0, Status='live', Symbol='Egfl7', Aliases='VE-statin, Zneu1', description='EGF-like domain 7', other_designations='EGF-like protein 7|NEU1 protein|NOTCH4-like protein|multiple EGF-like domains protein 7|multiple epidermal growth factor-like domains protein 7|vascular endothelial statin|vascular endothelial-statin', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26581028, end_position_on_the_genomic_accession=26592682, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  353165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353165, CurrentID=0, Status='live', Symbol='Tas2r136', Aliases='Tas2r36, mt2r52', description='taste receptor, type 2, member 136', other_designations='T2R136|T2R36|taste receptor type 2 member 36', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132777179, end_position_on_the_genomic_accession=132778162, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  353166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353166, CurrentID=0, Status='live', Symbol='Tas2r117', Aliases='T2R17, Tas2r17, mGR17, mt2r54', description='taste receptor, type 2, member 117', other_designations='T2R117|tas2r117', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132802901, end_position_on_the_genomic_accession=132803893, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  353167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353167, CurrentID=0, Status='live', Symbol='Tas2r123', Aliases='T2R23, Tas2r23, mGR23, mt2r55, t2r55', description='taste receptor, type 2, member 123', other_designations='STC9-2|T2R123|taste receptor type 2 member 23', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132847142, end_position_on_the_genomic_accession=132848143, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  353169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353169, CurrentID=0, Status='live', Symbol='Slc2a12', Aliases='GLUT-12, Glut12', description='solute carrier family 2 (facilitated glucose transporter), member 12', other_designations='facilitated glucose transporter|glucose transporter type 12|solute carrier family 2, member 12', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=22612723, end_position_on_the_genomic_accession=22703880, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  353170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353170, CurrentID=0, Status='live', Symbol='Txlng', Aliases='4932441K18Rik, Fiat, Lrp', description='taxilin gamma', other_designations='factor inhibiting activating transcription factor 4 (ATF4)-mediated transcription|lipopolysaccharide-responsive gene protein|lipopolysaccharide-specific response protein 5', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=162778917, end_position_on_the_genomic_accession=162829454, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  353172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353172, CurrentID=0, Status='live', Symbol='Gars', Aliases='GENA202, Gena201, Nmf249, Sgrp23', description='glycyl-tRNA synthetase', other_designations='AP-4-A synthetase|diadenosine tetraphosphate synthetase|glyRS|storage granule protein 23', map_location='6 27.29 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55038001, end_position_on_the_genomic_accession=55079504, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  58864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=58864, CurrentID=0, Status='live', Symbol='Tssk3', Aliases='1700014N07Rik, 4930594I21Rik, Stk22c, Stk22d, Tsk-3, Tssk-3b, Tssk3', description='testis-specific serine kinase 3', other_designations='serine/threonine kinase 22C (spermiogenesis associated)|serine/threonine-protein kinase 22C|spermiogenesis associated serine/threonine kinase 22D|testis specific serine kinase-3|testis-specific kinase 3|testis-specific serine/threonine kinase 3b', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129489008, end_position_on_the_genomic_accession=129490770, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  353187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353187, CurrentID=0, Status='live', Symbol='Nr1d2', Aliases='RVR, Rev-erb', description='nuclear receptor subfamily 1, group D, member 2', other_designations='Rev-erb beta|orphan nuclear receptor RVR|rev-erb-beta', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=18204056, end_position_on_the_genomic_accession=18239106, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  353188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353188, CurrentID=0, Status='live', Symbol='Adam32', Aliases='', description='a disintegrin and metallopeptidase domain 32', other_designations='a disintegrin and metalloprotease domain 32', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=24836143, end_position_on_the_genomic_accession=24948917, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  353190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353190, CurrentID=0, Status='live', Symbol='Edc3', Aliases='AA517853, Lsm16, Yjdc', description='enhancer of mRNA decapping 3', other_designations='enhancer of mRNA decapping 3 homolog|yjeF domain-containing protein 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57708534, end_position_on_the_genomic_accession=57750167, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  353204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353204, CurrentID=0, Status='live', Symbol='Aldoart1', Aliases='4921524E03Rik, Aldo1-ps, Aldo1-ps2, Aldoa-ps2', description='aldolase 1 A, retrogene 1', other_designations='aldolase 1, A isoform, pseudogene 2|aldolase 1, A isoform, retrogene 1|aldolase A retroprotein 1', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=72850583, end_position_on_the_genomic_accession=72852634, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  353208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353208, CurrentID=0, Status='live', Symbol='Zfp931', Aliases='2810021G02Rik', description='zinc finger protein 931', other_designations='novel KRAB box and zinc finger, C2H2 type domain containing protein', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=178066878, end_position_on_the_genomic_accession=178078425, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  353211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353211, CurrentID=0, Status='live', Symbol='Prune2', Aliases='6330414G02Rik, A230083H22Rik, A330102H22Rik, Bmcc1, mKIAA0367', description='prune homolog 2 (Drosophila)', other_designations='BNIP2 motif-containing molecule at the C-terminal region 1|neuronal protein', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=16955670, end_position_on_the_genomic_accession=17223958, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  353234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353234, CurrentID=0, Status='live', Symbol='Pcdha2', Aliases='', description='protocadherin alpha 2', other_designations='', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36939205, end_position_on_the_genomic_accession=37187661, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  353235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353235, CurrentID=0, Status='live', Symbol='Pcdha8', Aliases='', description='protocadherin alpha 8', other_designations='', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36992467, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  353236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353236, CurrentID=0, Status='live', Symbol='Pcdhac1', Aliases='CNRc1', description='protocadherin alpha subfamily C, 1', other_designations='protocadherin alpha C1', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37090136, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  353237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353237, CurrentID=0, Status='live', Symbol='Pcdhac2', Aliases='CNRc2', description='protocadherin alpha subfamily C, 2', other_designations='protocadherin alpha C2', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37143969, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  353242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353242, CurrentID=0, Status='live', Symbol='Mrpl21', Aliases='BC028768', description='mitochondrial ribosomal protein L21', other_designations='L21mt|MRP-L21', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3282920, end_position_on_the_genomic_accession=3292837, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  277328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277328, CurrentID=0, Status='live', Symbol='Trpa1', Aliases='Anktm1, TRPA1b', description='transient receptor potential cation channel, subfamily A, member 1', other_designations='ANKTM1|ankyrin-like with transmembrane domains protein 1', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=14872648, end_position_on_the_genomic_accession=14919014, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  321019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=321019, CurrentID=0, Status='live', Symbol='Gpr183', Aliases='Ebi2', description='G protein-coupled receptor 183', other_designations='EBV-induced G-protein coupled receptor 2 homolog|Epstein-Barr virus induced gene 2|epstein-Barr virus-induced G-protein coupled receptor 2', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=121952331, end_position_on_the_genomic_accession=121965193, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  353258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353258, CurrentID=0, Status='live', Symbol='Ltv1', Aliases='2610020N02Rik', description='LTV1 ribosome biogenesis factor', other_designations='LTV1 homolog', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=13178597, end_position_on_the_genomic_accession=13193137, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  353282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353282, CurrentID=0, Status='live', Symbol='Sfmbt2', Aliases='D2Wsu23e, D330030P06Rik', description='Scm-like with four mbt domains 2', other_designations='Scm-related gene containing four mbt domain 2', map_location='2 7.81 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10370437, end_position_on_the_genomic_accession=10595253, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  353283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353283, CurrentID=0, Status='live', Symbol='Eras', Aliases='HRAS2, HRasp, Ha-Ras2, ecat5', description='ES cell-expressed Ras', other_designations='E-Ras|embryonic stem cell-expressed Ras|small GTPase E-Ras', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7924276, end_position_on_the_genomic_accession=7928607, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  353287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353287, CurrentID=0, Status='live', Symbol='Clec18a', Aliases='Mrcl', description='C-type lectin domain family 18, member A', other_designations='mannose receptor-like protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111069496, end_position_on_the_genomic_accession=111092426, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  353310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353310, CurrentID=0, Status='live', Symbol='Zfp703', Aliases='1110032O19Rik, AI430822, AL022941, Csmn1, End2, Zeppo1, Znf703, Zpo1', description='zinc finger protein 703', other_designations='ES neuronal differentiation 2|zinc finger elbow-related proline domain protein 1', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=26977336, end_position_on_the_genomic_accession=26981462, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  353326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353326, CurrentID=0, Status='live', Symbol='Rtl1', Aliases='6430411K18Rik, Mar, Mart1, Mor1, Peg11', description='retrotransposon-like 1', other_designations='mammalian retrotransposon derived protein 1|paternally expressed gene 11 protein|retrotransposon-derived protein PEG11', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109590169, end_position_on_the_genomic_accession=109595403, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  353327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353327, CurrentID=0, Status='live', Symbol='Mkrn1-ps1', Aliases='', description='makorin, ring finger protein 1, pseudogene 1', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88837799, end_position_on_the_genomic_accession=88839399, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  353328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353328, CurrentID=0, Status='live', Symbol='Muc6', Aliases='', description='mucin 6, gastric', other_designations='MUC-6|gastric mucin-6|gastric mucin-like protein|secreted gel-forming mucin-6', map_location='7 87.03 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141634049, end_position_on_the_genomic_accession=141655308, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  277343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277343, CurrentID=0, Status='live', Symbol='Wfdc8', Aliases='Gm706', description='WAP four-disulfide core domain 8', other_designations='WAP-type serine protease inhibitor|putative protease inhibitor WAP8C', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164596458, end_position_on_the_genomic_accession=164613626, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  208169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208169, CurrentID=0, Status='live', Symbol='Slc9c1', Aliases='Gm610, NHE-10, Slc9a10, sNHE, spermNHE', description='solute carrier family 9, subfamily C (Na+-transporting carboxylic acid decarboxylase), member 1', other_designations='Na(+)/H(+) exchanger 10|solute carrier family 10 member 10|solute carrier family 9|solute carrier family 9 member C1|solute carrier family 9, member 10|sperm-specific Na(+)/H(+) exchanger|sperm-specific sodium proton exchanger', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=45535266, end_position_on_the_genomic_accession=45610396, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  353342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353342, CurrentID=0, Status='live', Symbol='Peg13', Aliases='', description='paternally expressed 13', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=72805604, end_position_on_the_genomic_accession=72810324, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  353344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353344, CurrentID=0, Status='live', Symbol='Opn5', Aliases='Gpr136, Neuropsin, PGR12, TMEM13', description='opsin 5', other_designations='G protein-coupled receptor 136|G-protein coupled receptor 136|G-protein coupled receptor PGR12', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=42556779, end_position_on_the_genomic_accession=42620470, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  353346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353346, CurrentID=0, Status='live', Symbol='Gpr141', Aliases='PGR13', description='G protein-coupled receptor 141', other_designations='G-protein coupled receptor PGR13', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19749680, end_position_on_the_genomic_accession=19824337, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  277345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277345, CurrentID=0, Status='live', Symbol='Wfdc16', Aliases='BC043944, WAP16', description='WAP four-disulfide core domain 16', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164634701, end_position_on_the_genomic_accession=164638808, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  353371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353371, CurrentID=0, Status='live', Symbol='Oxct2b', Aliases='Scot-t2', description='3-oxoacid CoA transferase 2B', other_designations='3-oxoacid CoA-transferase 2B|3-oxoacid-CoA transferase 2B|succinyl CoA:3-oxoacid CoA-transferase|succinyl-CoA:3-ketoacid-coenzyme A transferase 2B, mitochondrial|testis-specific succinyl CoA:3-oxoacid CoA-transferase 2|testis-specific succinyl-CoA:3-oxoacid CoA-transferase 2', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123116248, end_position_on_the_genomic_accession=123118000, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15209, CurrentID=0, Status='live', Symbol='Hesx1', Aliases='HES-1, Rpx', description='homeobox gene expressed in ES cells', other_designations='anterior-restricted homeobox protein|homeo box gene expressed in ES cells|homeobox protein ANF|rathke pouch homeo box', map_location='14 A3-B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26994420, end_position_on_the_genomic_accession=27002329, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  15216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15216, CurrentID=0, Status='live', Symbol='Hfe', Aliases='MR2', description='hemochromatosis', other_designations='', map_location='13 A2-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23703564, end_position_on_the_genomic_accession=23710986, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  353499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353499, CurrentID=0, Status='live', Symbol='Tmc4', Aliases='', description='transmembrane channel-like gene family 4', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3665754, end_position_on_the_genomic_accession=3677553, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  353502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353502, CurrentID=0, Status='live', Symbol='Hcfc1r1', Aliases='HPIP', description='host cell factor C1 regulator 1 (XPO1-dependent)', other_designations='HCF-1 beta-propeller-interacting protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23673608, end_position_on_the_genomic_accession=23675524, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  353504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=353504, CurrentID=0, Status='live', Symbol='Dio3os', Aliases='AW556460, Dio3as', description='deiodinase, iodothyronine type III, opposite strand', other_designations='deiodinase, iodothyronine type III, antisense', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110275384, end_position_on_the_genomic_accession=110278068, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  68514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68514, CurrentID=0, Status='live', Symbol='Micu2', Aliases='1110008L20Rik, 4833427E09Rik, AU041871, Efha1, Smhs2', description='mitochondrial calcium uptake 2', other_designations='EF hand domain family A1|EF-hand domain-containing family member A1', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57916261, end_position_on_the_genomic_accession=57999336, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  15270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15270, CurrentID=0, Status='live', Symbol='H2afx', Aliases='AW228881, H2A.X, H2ax, Hist5-2ax, gammaH2ax', description='H2A histone family, member X', other_designations='H2a/x|histone 5 protein 2ax|histone H2A.x', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44334715, end_position_on_the_genomic_accession=44336077, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  15288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=15288, CurrentID=0, Status='live', Symbol='Hmbs', Aliases='PBGD, T25658, Ups, Uros1', description='hydroxymethylbilane synthase', other_designations='PBG-D|pre-uroporphyrinogen synthase', map_location='9 24.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44336348, end_position_on_the_genomic_accession=44344228, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  68524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68524, CurrentID=0, Status='live', Symbol='Wipf2', Aliases='1110014J05Rik, 5730509C05Rik, AA407487, Gm1176, Wich, Wire', description='WAS/WASL interacting protein family, member 2', other_designations='WASP-binding protein|WASP-interacting protein-related protein|WIP-related protein', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98863598, end_position_on_the_genomic_accession=98905578, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  233752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233752, CurrentID=0, Status='live', Symbol='Insc', Aliases='3830422K02Rik, C730020D23', description='inscuteable homolog (Drosophila)', other_designations='minsc', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114743677, end_position_on_the_genomic_accession=114850380, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  233765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233765, CurrentID=0, Status='live', Symbol='Plekha7', Aliases='A430081P20Rik, AI414265', description='pleckstrin homology domain containing, family A member 7', other_designations='PH domain-containing family A member 7|heart adapter protein 1', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=116124118, end_position_on_the_genomic_accession=116308381, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  211948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211948, CurrentID=0, Status='live', Symbol='Pde12', Aliases="2'-PDE, E430028B21Rik", description='phosphodiesterase 12', other_designations="2'-phosphodiesterase|2-PDE|mitochondrial deadenylase", map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26664117, end_position_on_the_genomic_accession=26669846, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  211949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=211949, CurrentID=0, Status='live', Symbol='Spsb4', Aliases='D030068E18Rik, SSB-4, Ssb4', description='splA/ryanodine receptor domain and SOCS box containing 4', other_designations='SPRY domain-containing SOCS box 4|SPRY domain-containing SOCS box protein SSB-4', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96943482, end_position_on_the_genomic_accession=97018355, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  277496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=277496, CurrentID=0, Status='live', Symbol='Lkaaear1', Aliases='4930526D03Rik, Gm692', description='LKAAEAR motif containing 1 (IKAAEAR murine motif)', other_designations='LKAAEAR motif-containing protein 1', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181696793, end_position_on_the_genomic_accession=181698468, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  80889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80889, CurrentID=0, Status='live', Symbol='Mesdc1', Aliases='AW061151', description='mesoderm development candidate 1', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=83880495, end_position_on_the_genomic_accession=83884341, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100503890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503890, CurrentID=0, Status='live', Symbol='Pet100', Aliases='2900053A13Rik', description='PET100 homolog (S. cerevisiae)', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3621551, end_position_on_the_genomic_accession=3624235, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  223227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223227, CurrentID=0, Status='live', Symbol='Sox21', Aliases='Sox25', description='SRY (sex determining region Y)-box 21', other_designations='DNA-binding protein SOX21|HMG-box protein SOX21|SRY-box containing gene 21', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118233233, end_position_on_the_genomic_accession=118237030, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  68537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68537, CurrentID=0, Status='live', Symbol='Mrpl13', Aliases='1110002D09Rik', description='mitochondrial ribosomal protein L13', other_designations='L13mt|MRP-L13', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=55534095, end_position_on_the_genomic_accession=55557312, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  223254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223254, CurrentID=0, Status='live', Symbol='Farp1', Aliases='AW228844, BC030329, Cdep', description='FERM, RhoGEF (Arhgef) and pleckstrin domain protein 1 (chondrocyte-derived)', other_designations='FERMRhoGEF (Arhgef) and pleckstrin domain protein 1', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=121034469, end_position_on_the_genomic_accession=121283744, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  223255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223255, CurrentID=0, Status='live', Symbol='Stk24', Aliases='1810013H02Rik, C76483, MST-3, Mst3, STE20', description='serine/threonine kinase 24', other_designations='STE20-like kinase MST3|mammalian STE20-like protein kinase 3|mammalian sterile twenty 3 kinase|serine/threonine kinase 24 (STE20 homolog, yeast)', map_location='14 64.95 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=121284187, end_position_on_the_genomic_accession=121380006, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  223262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223262, CurrentID=0, Status='live', Symbol='Timm8a2', Aliases='', description='translocase of inner mitochondrial membrane 8A2', other_designations='translocase of inner mitochondrial membrane 8 homolog a2', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122028538, end_position_on_the_genomic_accession=122038422, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  223267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223267, CurrentID=0, Status='live', Symbol='Ggact', Aliases='A2ld1', description='gamma-glutamylamine cyclotransferase', other_designations='AIG2-like domain 1|AIG2-like domain-containing protein 1', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122890860, end_position_on_the_genomic_accession=122913165, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  223272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223272, CurrentID=0, Status='live', Symbol='Itgbl1', Aliases='B930011D01Rik', description='integrin, beta-like 1', other_designations='with EGF-like repeat domains', map_location='14 E5', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=123659938, end_position_on_the_genomic_accession=123974079, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  80906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80906, CurrentID=0, Status='live', Symbol='Kcnip2', Aliases='KChIP2', description='Kv channel-interacting protein 2', other_designations='A-type potassium channel modulatory protein 2|potassium channel-interacting protein 2', map_location='19 38.75 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45792346, end_position_on_the_genomic_accession=45816063, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  223332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223332, CurrentID=0, Status='live', Symbol='Ranbp3l', Aliases='C130037N17Rik', description='RAN binding protein 3-like', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9055679, end_position_on_the_genomic_accession=9067335, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  223337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223337, CurrentID=0, Status='live', Symbol='Ugt3a2', Aliases='AI313915', description='UDP glycosyltransferases 3 family, polypeptide A2', other_designations='UDP glucuronosyltransferase 3 family, polypeptide A2|UDPGT 3A2', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9335551, end_position_on_the_genomic_accession=9370955, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  80914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=80914, CurrentID=0, Status='live', Symbol='Uck2', Aliases='AA407809, AI481316, AU018180, AU020720, TSA903, UK, UMK, Umpk', description='uridine-cytidine kinase 2', other_designations='UCK 2|cytidine monophosphokinase 2|uridine monophosphate kinase 2|uridine monophosphokinase 2', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167222895, end_position_on_the_genomic_accession=167285127, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  627049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=627049, CurrentID=0, Status='live', Symbol='Zfp800', Aliases='AA407452, Znf800', description='zinc finger protein 800', other_designations='', map_location='6 A3.2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=28210548, end_position_on_the_genomic_accession=28398053, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  223433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223433, CurrentID=0, Status='live', Symbol='Fam105a', Aliases='9830126M18', description='family with sequence similarity 105, member A', other_designations='protein FAM105A', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=27655071, end_position_on_the_genomic_accession=27681596, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  223435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223435, CurrentID=0, Status='live', Symbol='Trio', Aliases='6720464I07Rik, AA408740, Solo', description='triple functional domain (PTPRF interacting)', other_designations='', map_location='15 B1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=27730649, end_position_on_the_genomic_accession=28025953, orientation='minus', exon_count=60, OMIM='', no_hdr0=''),
  223453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223453, CurrentID=0, Status='live', Symbol='Dap', Aliases='4921531N22Rik, AI789753, AU040360', description='death-associated protein', other_designations='DAP-1', map_location='15 B2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=31224385, end_position_on_the_genomic_accession=31274338, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  223455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223455, CurrentID=0, Status='live', Symbol='March6', Aliases='3830408G03, F830029L24Rik, mKIAA0597', description='membrane-associated ring finger (C3HC4) 6', other_designations='MARCH-VI|membrane-associated RING finger protein 6|membrane-associated RING-CH protein VI', map_location='15 B2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=31455894, end_position_on_the_genomic_accession=31531048, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  223473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223473, CurrentID=0, Status='live', Symbol='Nipal2', Aliases='9330161F08Rik, Npal2', description='NIPA-like domain containing 2', other_designations='', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34572799, end_position_on_the_genomic_accession=34679437, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  223499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223499, CurrentID=0, Status='live', Symbol='Dcaf13', Aliases='Gm83, Wdsof1', description='DDB1 and CUL4 associated factor 13', other_designations='WD repeat and SOF domain-containing protein 1|WD repeats and SOF domain containing 1', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39112874, end_position_on_the_genomic_accession=39146856, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  223513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223513, CurrentID=0, Status='live', Symbol='Abra', Aliases='C130068O12Rik, STARS', description='actin-binding Rho activating protein', other_designations='abra|striated muscle activator of Rho-dependent signaling', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=41865293, end_position_on_the_genomic_accession=41869720, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  223527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223527, CurrentID=0, Status='live', Symbol='Eny2', Aliases='1810057B09Rik, 6720481I12, DC6, Ey2', description='enhancer of yellow 2 homolog (Drosophila)', other_designations='e(y)2 homolog|e(y)2 protein|enhancer of yellow 2 transcription factor homolog|enhancer of yellow eye homolog', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=44428081, end_position_on_the_genomic_accession=44438256, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102791, CurrentID=0, Status='live', Symbol='Tcta', Aliases='9130410M22Rik, AW553637, C85065l, Tcta', description='T cell leukemia translocation altered gene', other_designations='T-cell leukemia translocation altered', map_location='9 59.2 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108302958, end_position_on_the_genomic_accession=108306125, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  627081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=627081, CurrentID=0, Status='live', Symbol='Xlr5b', Aliases='OTTMUSG00000017646', description='X-linked lymphocyte-regulated 5B', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73146976, end_position_on_the_genomic_accession=73158886, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  223593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223593, CurrentID=0, Status='live', Symbol='E430025E21Rik', Aliases='AL022848, C76463, Kiaa0196, mKIAA0196, strumpellin', description='RIKEN cDNA E430025E21 gene', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=59331997, end_position_on_the_genomic_accession=59374194, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  223601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223601, CurrentID=0, Status='live', Symbol='Fam49b', Aliases='0910001A06Rik, AW122079', description='family with sequence similarity 49, member B', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=63929094, end_position_on_the_genomic_accession=64060452, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  223604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223604, CurrentID=0, Status='live', Symbol='Kcnk9', Aliases='Task3', description='potassium channel, subfamily K, member 9', other_designations='TWIK-related acid-sensitive K(+) channel 3|acid-sensitive potassium channel protein TASK-3', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=72502927, end_position_on_the_genomic_accession=72547853, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  223626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223626, CurrentID=0, Status='live', Symbol='Them6', Aliases='4930572J05Rik', description='thioesterase superfamily member 6', other_designations='UPF0670 protein C8orf55 homolog|UPF0670 protein THEM6 homolog|mesenchymal stem cell protein DSCD75 homolog', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74721234, end_position_on_the_genomic_accession=74724373, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  223642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223642, CurrentID=0, Status='live', Symbol='Zc3h3', Aliases='BC049953, Kiaa0150, Smicl, Zc3hdc3', description='zinc finger CCCH type containing 3', other_designations='smad-interacting CPSF-like factor|zinc finger CCCH type domain containing 3', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75754434, end_position_on_the_genomic_accession=75841971, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  223646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223646, CurrentID=0, Status='live', Symbol='Naprt', Aliases='9130210N20Rik1, Naprt', description='nicotinate phosphoribosyltransferase', other_designations='NAPRTase|nicotinate phosphoribosyltransferase domain containing 1|nicotinate phosphoribosyltransferase domain-containing protein 1', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75890964, end_position_on_the_genomic_accession=75894481, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  223648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223648, CurrentID=0, Status='live', Symbol='Ccdc166', Aliases='2410075B13Rik', description='coiled-coil domain containing 166', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75979872, end_position_on_the_genomic_accession=75982285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  223649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223649, CurrentID=0, Status='live', Symbol='Nrbp2', Aliases='BC011468', description='nuclear receptor binding protein 2', other_designations='nrbp2', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76085589, end_position_on_the_genomic_accession=76091594, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  223650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223650, CurrentID=0, Status='live', Symbol='Eppk1', Aliases='6230424I18Rik, EPIPL, EPIPL1, EPPK', description='epiplakin 1', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76101488, end_position_on_the_genomic_accession=76120195, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  223658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223658, CurrentID=0, Status='live', Symbol='Mroh1', Aliases='D330001F17Rik, Heatr7a', description='maestro heat-like repeat family member 1', other_designations='HEAT repeat containing 7A', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76380467, end_position_on_the_genomic_accession=76453039, orientation='plus', exon_count=48, OMIM='', no_hdr0=''),
  223664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223664, CurrentID=0, Status='live', Symbol='Lrrc14', Aliases='E130306I01Rik, mKIAA0014', description='leucine rich repeat containing 14', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76710623, end_position_on_the_genomic_accession=76715097, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  223665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223665, CurrentID=0, Status='live', Symbol='C030006K11Rik', Aliases='1110033I13Rik', description='RIKEN cDNA C030006K11 gene', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76721465, end_position_on_the_genomic_accession=76723845, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  223666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223666, CurrentID=0, Status='live', Symbol='Arhgap39', Aliases='9530053N22, AI843066, D15Wsu169e, Kiaa1688, Porf-2', description='Rho GTPase activating protein 39', other_designations='preoptic regulatory factor-2', map_location='15 36.28 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76723985, end_position_on_the_genomic_accession=76818170, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  223669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223669, CurrentID=0, Status='live', Symbol='Zfp7', Aliases='KRAB20, KRAB7, Krox-2, Zfp-7, Zfp65, Zfp80, Zfp86-rs1, mszf73-2', description='zinc finger protein 7', other_designations='zinc finger protein 86, related sequence 1', map_location='15 36.28 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76879217, end_position_on_the_genomic_accession=76892395, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  223690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223690, CurrentID=0, Status='live', Symbol='Ankrd54', Aliases='C730048E16Rik, EST1068184, EST475269, Liar', description='ankyrin repeat domain 54', other_designations='lyn-interacting ankyrin repeat protein', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79053094, end_position_on_the_genomic_accession=79062958, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  223691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223691, CurrentID=0, Status='live', Symbol='Eif3l', Aliases='0610011H21Rik, D15N1e, Eif3eip, Eif3ip, Eif3s6ip, HSP-66Y, PAF67', description='eukaryotic translation initiation factor 3, subunit L', other_designations='66 kDa tyrosine-rich heat shock protein|67 kDa polymerase-associated factor|eIF3l|eukaryotic translation initiation factor 3 subunit 6-interacting protein|eukaryotic translation initiation factor 3 subunit E-interacting protein', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79075223, end_position_on_the_genomic_accession=79094400, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  223693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223693, CurrentID=0, Status='live', Symbol='Tmem184b', Aliases='2610507A11, 4732495E13Rik', description='transmembrane protein 184b', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79360684, end_position_on_the_genomic_accession=79403303, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  223696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223696, CurrentID=0, Status='live', Symbol='Tomm22', Aliases='2310047D01, Tom22', description='translocase of outer mitochondrial membrane 22 homolog (yeast)', other_designations='translocase of outer membrane 22 kDa subunit homolog', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79670868, end_position_on_the_genomic_accession=79672862, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  223697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223697, CurrentID=0, Status='live', Symbol='Sun2', Aliases='B230369L08Rik, C030011B15, Unc84b', description='Sad1 and UNC84 domain containing 2', other_designations='protein unc-84 homolog B|sad1/unc-84 protein-like 2', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79724068, end_position_on_the_genomic_accession=79742536, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  223701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223701, CurrentID=0, Status='live', Symbol='Mkl1', Aliases='AMKL, Bsac, Mal, Mrtf-A', description='MKL (megakaryoblastic leukemia)/myocardin-like 1', other_designations='basic SAP and coiled-coil domains|basic SAP coiled-coil transcription activator|megakaryoblastic leukemia (translocation) 1 homolog|myocardin-related transcription factor A', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81012281, end_position_on_the_genomic_accession=81190757, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  223722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223722, CurrentID=0, Status='live', Symbol='Mcat', Aliases='AI225907, BC025519', description='malonyl CoA:ACP acyltransferase (mitochondrial)', other_designations='MCT|[Acyl-carrier-protein] malonyltransferase|malonyl CoA-acyl carrier protein transacylase, mitochondrial|mitochondrial malonyltransferase', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83546797, end_position_on_the_genomic_accession=83555757, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  223723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223723, CurrentID=0, Status='live', Symbol='Ttll12', Aliases='BC055368, D430005B17', description='tubulin tyrosine ligase-like family, member 12', other_designations='TTL domain protein|ttll12', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83575094, end_position_on_the_genomic_accession=83595157, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  223726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223726, CurrentID=0, Status='live', Symbol='Mpped1', Aliases='6230403M19, AW456965, D15Bwg0669e', description='metallophosphoesterase domain containing 1', other_designations='', map_location='15 39.44 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83779472, end_position_on_the_genomic_accession=83858487, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  223732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223732, CurrentID=0, Status='live', Symbol='Ldoc1l', Aliases='BC058638, Mar6, Mart6', description='leucine zipper, down-regulated in cancer 1-like', other_designations='gag-like protein|mammalian retrotransposon-derived protein 6', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84553398, end_position_on_the_genomic_accession=84557823, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  223739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223739, CurrentID=0, Status='live', Symbol='5031439G07Rik', Aliases='AW554572, mKIAA0930', description='RIKEN cDNA 5031439G07 gene', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84945720, end_position_on_the_genomic_accession=84987971, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  223752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223752, CurrentID=0, Status='live', Symbol='Gramd4', Aliases='9930016O13', description='GRAM domain containing 4', other_designations='death-inducing protein', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=86057695, end_position_on_the_genomic_accession=86137636, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  223753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223753, CurrentID=0, Status='live', Symbol='Cerk', Aliases='AI848610, D330016D08Rik, mKIAA1646', description='ceramide kinase', other_designations='acylsphingosine kinase|mCERK', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=86139101, end_position_on_the_genomic_accession=86186193, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  223754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223754, CurrentID=0, Status='live', Symbol='Tbc1d22a', Aliases='BC023106, D15Ertd781e', description='TBC1 domain family, member 22a', other_designations='', map_location='15 40.56 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=86214421, end_position_on_the_genomic_accession=86498503, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  223770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223770, CurrentID=0, Status='live', Symbol='Brd1', Aliases='1110059H06Rik, AI316859, mKIAA4191', description='bromodomain containing 1', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88687034, end_position_on_the_genomic_accession=88734226, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  223773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223773, CurrentID=0, Status='live', Symbol='Zbed4', Aliases='1700009F06Rik', description='zinc finger, BED type containing 4', other_designations='zinc finger, BED domain containing 4', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88751666, end_position_on_the_genomic_accession=88784517, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  223774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223774, CurrentID=0, Status='live', Symbol='Alg12', Aliases='ECM39', description='asparagine-linked glycosylation 12 (alpha-1,6-mannosyltransferase)', other_designations='asparagine-linked glycosylation 12 homolog (yeast, alpha-1,6-mannosyltransferase)|asparagine-linked glycosylation protein 12 homolog|dolichyl-P-Man:Man(7)GlcNAc(2)-PP-dolichyl-alpha-1,6-mannosyltransferase|mannosyltransferase ALG12 homolog', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88805243, end_position_on_the_genomic_accession=88819318, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  223775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223775, CurrentID=0, Status='live', Symbol='Pim3', Aliases='BC026639, Kid1', description='proviral integration site 3', other_designations='KID-1/kinase induced by depolarization', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88862194, end_position_on_the_genomic_accession=88865726, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  223776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223776, CurrentID=0, Status='live', Symbol='Selo', Aliases='1300018J18Rik', description='selenoprotein O', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89089107, end_position_on_the_genomic_accession=89100340, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  223780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223780, CurrentID=0, Status='live', Symbol='Adm2', Aliases='Am2, IMD, Imdn', description='adrenomedullin 2', other_designations='intermedin', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89322899, end_position_on_the_genomic_accession=89324731, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  223809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223809, CurrentID=0, Status='live', Symbol='Smgc', Aliases='2310010P21Rik, DXImx49e, Muc19, Muc19/Smgc, Sfc21', description='submandibular gland protein C', other_designations='mucin 19|neonatal submandibular gland protein C', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=91838328, end_position_on_the_genomic_accession=91861435, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  233912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233912, CurrentID=0, Status='live', Symbol='Armc5', Aliases='AW049683, BC032200', description='armadillo repeat containing 5', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128237357, end_position_on_the_genomic_accession=128245100, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  223827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223827, CurrentID=0, Status='live', Symbol='Gxylt1', Aliases='Glt8d3, Gm1228, Gm87', description='glucoside xylosyltransferase 1', other_designations='glycosyltransferase 8 domain containing 3|glycosyltransferase 8 domain-containing protein 3', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=93239735, end_position_on_the_genomic_accession=93275336, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  223828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223828, CurrentID=0, Status='live', Symbol='Pphln1', Aliases='CR, HSPC206, HSPC232', description='periphilin 1', other_designations='gastric cancer antigen Ga50', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=93398350, end_position_on_the_genomic_accession=93491913, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  223838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223838, CurrentID=0, Status='live', Symbol='Adamts20', Aliases='Adamst20, bt', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 20', other_designations='ADAM-TS 20|ADAM-TS20|belted', map_location='15 48.2 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=94268314, end_position_on_the_genomic_accession=94465427, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  223843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223843, CurrentID=0, Status='live', Symbol='Dbx2', Aliases='9430056A22Rik, 9630005K15, AI846217, Gm1229', description='developing brain homeobox 2', other_designations='', map_location='15 50.47 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=95623563, end_position_on_the_genomic_accession=95654771, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  223864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223864, CurrentID=0, Status='live', Symbol='Rapgef3', Aliases='2310016P22Rik, 9330170P05Rik, Epac, Epac1', description='Rap guanine nucleotide exchange factor (GEF) 3', other_designations='Rap1 guanine-nucleotide-exchange factor directly activated by cAMP|cAMP-GEFI|cAMP-regulated guanine nucleotide exchange factor I|exchange protein directly activated by cAMP 1', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97744770, end_position_on_the_genomic_accession=97768375, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  223870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223870, CurrentID=0, Status='live', Symbol='Senp1', Aliases='2310046A20Rik, D15Ertd528e, E330036L07Rik, suPr-2', description='SUMO1/sentrin specific peptidase 1', other_designations='SUMO-1 protease 2|SUMO1/sentrin specific protease 1|sentrin/SUMO-specific protease SENP1', map_location='15 54.04 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98038744, end_position_on_the_genomic_accession=98093569, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  223881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223881, CurrentID=0, Status='live', Symbol='Rnd1', Aliases='A830014L09Rik, Arhs', description='Rho family GTPase 1', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98669205, end_position_on_the_genomic_accession=98677461, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  223915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223915, CurrentID=0, Status='live', Symbol='Krt73', Aliases='BC067067, Kb36', description='keratin 73', other_designations='CK-73|K73|cytokeratin-73|keratin-73|type II inner root sheath-specific keratin-K6irs3|type II keratin Kb36|type-II keratin Kb36', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101793308, end_position_on_the_genomic_accession=101802332, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  223917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223917, CurrentID=0, Status='live', Symbol='Krt79', Aliases='BC031593', description='keratin 79', other_designations='CK-79|K79|cytokeratin-79|keratin-79|type-II keratin Kb38', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101929332, end_position_on_the_genomic_accession=101940324, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  223918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223918, CurrentID=0, Status='live', Symbol='Spryd3', Aliases='BC008150', description='SPRY domain containing 3', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102116526, end_position_on_the_genomic_accession=102136245, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  223920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223920, CurrentID=0, Status='live', Symbol='Soat2', Aliases='ACAT2, D15Wsu97e', description='sterol O-acyltransferase 2', other_designations='ACAT-2|acyl coenzyme A:cholesterol acyltransferase 2|acyl-CoA:cholesterol acyltransferase|acyl-coenzyme A:cholesterol acyltransferase 2|cholesterol acyltransferase 2', map_location='15 57.33 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102150414, end_position_on_the_genomic_accession=102163436, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  223921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223921, CurrentID=0, Status='live', Symbol='Aaas', Aliases='AAA, ADRACALA, D030041N15Rik, GL003', description='achalasia, adrenocortical insufficiency, alacrimia', other_designations='adracalin', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102338247, end_position_on_the_genomic_accession=102350759, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  223922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223922, CurrentID=0, Status='live', Symbol='Atf7', Aliases='1110012F10Rik, 9430065F09Rik, AI549878, C130020M04Rik', description='activating transcription factor 7', other_designations='cAMP-dependent transcription factor ATF-7|transcription factor ATF-A', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102536643, end_position_on_the_genomic_accession=102625421, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  223927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223927, CurrentID=0, Status='live', Symbol='BC048502', Aliases='', description='cDNA sequence BC048502', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103438950, end_position_on_the_genomic_accession=103450079, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  212085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=212085, CurrentID=0, Status='live', Symbol='Trim52', Aliases='4921513B05Rik', description='tripartite motif-containing 52', other_designations='', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=106106198, end_position_on_the_genomic_accession=106135862, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  81018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=81018, CurrentID=0, Status='live', Symbol='Rnf114', Aliases='1110008J21Rik, AI225886, AW549494, Zfp228, Zfp313, Znf228', description='ring finger protein 114', other_designations='zinc finger protein 313', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167492645, end_position_on_the_genomic_accession=167516169, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  223970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223970, CurrentID=0, Status='live', Symbol='Rmi2', Aliases='A630055G03Rik, Gm537, Gm929', description='RMI2, RecQ mediated genome instability 2, homolog (S. cerevisiae)', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10835059, end_position_on_the_genomic_accession=10843235, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  223978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223978, CurrentID=0, Status='live', Symbol='Cpped1', Aliases='C530044N13Rik', description='calcineurin-like phosphoesterase domain containing 1', other_designations='Vcalcineurin-like phosphoesterase domain containing 1|calcineurin-like phosphoesterase domain-containing protein 1', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11803684, end_position_on_the_genomic_accession=11909475, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  223989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=223989, CurrentID=0, Status='live', Symbol='Marf1', Aliases='4921513D23Rik, BC031575, C87306, E030019O05, Lkap, mKIAA0430', description='meiosis arrest female 1', other_designations='limkain b1|limkain-b1', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=14109166, end_position_on_the_genomic_accession=14159274, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  224008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224008, CurrentID=0, Status='live', Symbol='Spidr', Aliases='2310008H04Rik, Kiaa0146', description='scaffolding protein involved in DNA repair', other_designations='scaffolding protein involved i DNA repair', map_location='16 A2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=15889224, end_position_on_the_genomic_accession=16146851, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  224014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224014, CurrentID=0, Status='live', Symbol='Fgd4', Aliases='9030023J02Rik, 9330209B17Rik, Frabp, ZFYVE6', description='FYVE, RhoGEF and PH domain containing 4', other_designations='Fgd1-related F-actin-binding protein|actin filament-binding protein frabin', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=16416915, end_position_on_the_genomic_accession=16601112, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  224019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224019, CurrentID=0, Status='live', Symbol='Tmem191c', Aliases='4933405M22Rik, D16Bwg1494e, MNCb-4137', description='transmembrane protein 191C', other_designations='', map_location='16 10.71 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17276300, end_position_on_the_genomic_accession=17278661, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  224020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224020, CurrentID=0, Status='live', Symbol='Pi4ka', Aliases='Pik4ca', description='phosphatidylinositol 4-kinase, catalytic, alpha polypeptide', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17280351, end_position_on_the_genomic_accession=17406314, orientation='minus', exon_count=57, OMIM='', no_hdr0=''),
  224022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224022, CurrentID=0, Status='live', Symbol='Slc7a4', Aliases='AI853530', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 4', other_designations='CAT-4|CAT4|solute carrier family 7 member 4', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17572018, end_position_on_the_genomic_accession=17576892, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  224023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224023, CurrentID=0, Status='live', Symbol='Klhl22', Aliases='2610318I18Rik, Kelchl', description='kelch-like 22', other_designations='', map_location='16 11.01 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17759090, end_position_on_the_genomic_accession=17793382, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  224024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224024, CurrentID=0, Status='live', Symbol='Scarf2', Aliases='NSR1, SREC-II, SRECRP-1, Scatf1, Srec2, SrecII', description='scavenger receptor class F, member 2', other_designations='scavenger receptor expressed by endothelial cells 2 (acetyl LDL receptor like)|scavenger receptor type F', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17797282, end_position_on_the_genomic_accession=17808294, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  68563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68563, CurrentID=0, Status='live', Symbol='Dpm3', Aliases='1110001H19Rik, AV117483', description='dolichyl-phosphate mannosyltransferase polypeptide 3', other_designations='DPM synthase complex subunit 3|DPM synthase subunit 3|MPD synthase subunit 3|dolichol-phosphate mannose synthase subunit 3|dolichyl-phosphate beta-D-mannosyltransferase subunit 3|mannose-P-dolichol synthase subunit 3', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89259263, end_position_on_the_genomic_accession=89267079, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  224044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224044, CurrentID=0, Status='live', Symbol='Cyp2ab1', Aliases='EG224044, F730023N20', description='cytochrome P450, family 2, subfamily ab, polypeptide 1', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20308387, end_position_on_the_genomic_accession=20325412, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  224045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224045, CurrentID=0, Status='live', Symbol='Eif2b5', Aliases='C81315', description='eukaryotic translation initiation factor 2B, subunit 5 epsilon', other_designations='eIF-2B GDP-GTP exchange factor subunit epsilon', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20498808, end_position_on_the_genomic_accession=20509325, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  224055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224055, CurrentID=0, Status='live', Symbol='Rtp2', Aliases='Gm605', description='receptor transporter protein 2', other_designations='receptor transporting protein 2', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23925548, end_position_on_the_genomic_accession=23930847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  224065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224065, CurrentID=0, Status='live', Symbol='Uts2b', Aliases='Gm538, U2B, Urp, Uts2d', description='urotensin 2B', other_designations='U-IIB|UIIB|urotensin 2 domain containing protein|urotensin 2 domain-containing protein|urotensin II-related peptide|urotensin IIB|urotensin-2 domain-containing protein|urotensin-IIB', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=27353322, end_position_on_the_genomic_accession=27370239, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  224079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224079, CurrentID=0, Status='live', Symbol='Atp13a4', Aliases='4631413J11Rik, 4832416L12, 9330174J19Rik', description='ATPase type 13A4', other_designations='P5-ATPase|type V P-type ATPase', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=29396038, end_position_on_the_genomic_accession=29544864, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  224088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224088, CurrentID=0, Status='live', Symbol='Atp13a3', Aliases='AU022875, Gm1745, Gm541, Gm542', description='ATPase type 13A3', other_designations='type V P-type ATPase', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30312423, end_position_on_the_genomic_accession=30388856, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  68565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68565, CurrentID=0, Status='live', Symbol='Mrps18a', Aliases='1110004O20Rik, C79712', description='mitochondrial ribosomal protein S18A', other_designations='28S ribosomal protein S18-3, mitochondrial|MRP-S18-3|MRP-S18-a|S18mt-a', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46111004, end_position_on_the_genomic_accession=46128908, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  224092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224092, CurrentID=0, Status='live', Symbol='Lsg1', Aliases='5830465I20, AA409273, D16Bwg1547e', description='large 60S subunit nuclear export GTPase 1', other_designations='', map_location='16 21.34 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30561369, end_position_on_the_genomic_accession=30587589, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  224093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224093, CurrentID=0, Status='live', Symbol='Fam43a', Aliases='Tuf1', description='family with sequence similarity 43, member A', other_designations='somitovasculin', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=30599723, end_position_on_the_genomic_accession=30602797, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  224105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224105, CurrentID=0, Status='live', Symbol='Pak2', Aliases='5330420P17Rik, A130002K10Rik, AI836325, D16Ertd269e, PAK-2, mKIAA4182', description='p21 protein (Cdc42/Rac)-activated kinase 2', other_designations='gamma-PAK|p21-activated kinase 2', map_location='16 22.4 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32016290, end_position_on_the_genomic_accession=32079342, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  224109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224109, CurrentID=0, Status='live', Symbol='Nrros', Aliases='E430025L02Rik, Lrcc33, Lrrc33', description='negative regulator of reactive oxygen species', other_designations='leucine rich repeat containing 33|leucine-rich repeat-containing protein 33', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32142785, end_position_on_the_genomic_accession=32165589, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  224111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224111, CurrentID=0, Status='live', Symbol='Ubxn7', Aliases='A630090P18, AI196514, Ubxd7, mKIAA0794', description='UBX domain protein 7', other_designations='UBX domain containing 7', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32332184, end_position_on_the_genomic_accession=32393750, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  224116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224116, CurrentID=0, Status='live', Symbol='Muc20', Aliases='BC026367', description='mucin 20', other_designations='MUC-20', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32777419, end_position_on_the_genomic_accession=32797435, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  224129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224129, CurrentID=0, Status='live', Symbol='Adcy5', Aliases='AW121902, Ac5', description='adenylate cyclase 5', other_designations='ATP pyrophosphate-lyase 5|adcy5|adenylate cyclase type V|adenylyl cyclase 5', map_location='16 B-5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35155636, end_position_on_the_genomic_accession=35304549, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  224132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224132, CurrentID=0, Status='live', Symbol='Dirc2', Aliases='RCC4', description='disrupted in renal carcinoma 2 (human)', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=35694903, end_position_on_the_genomic_accession=35769356, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  224139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224139, CurrentID=0, Status='live', Symbol='Golgb1', Aliases='4930428L02Rik, 628101, 6330407A06Rik, AU042952, C130074L01Rik, F730017E11Rik, Gm6840, mKIAA4151', description='golgi autoantigen, golgin subfamily b, macrogolgin 1', other_designations='giantin', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36875173, end_position_on_the_genomic_accession=36933085, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  224143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224143, CurrentID=0, Status='live', Symbol='Poglut1', Aliases='9630046K23Rik, Clp46, Ktelc1, Rumi', description='protein O-glucosyltransferase 1', other_designations='CAP10-like 46 kDa protein|KTEL (Lys-Tyr-Glu-Leu) containing 1|KTEL motif-containing protein 1|O-glucosyltransferase Rumi homolog|protein O-xylosyltransferase', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=38525058, end_position_on_the_genomic_accession=38550256, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  224170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224170, CurrentID=0, Status='live', Symbol='Dzip3', Aliases='2310047C04Rik, 2A-HUB, 6430549P11Rik, A230104G20', description='DAZ interacting protein 3, zinc finger', other_designations='DAZ-interacting protein 3 homolog', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48924228, end_position_on_the_genomic_accession=48994153, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  224171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224171, CurrentID=0, Status='live', Symbol='C330027C09Rik', Aliases='AA408511, AU018569, Cip2a, Kiaa1524', description='RIKEN cDNA C330027C09 gene', other_designations='cancerous inhibitor of PP2A|p90 autoantigen homolog', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48994188, end_position_on_the_genomic_accession=49019711, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  224224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224224, CurrentID=0, Status='live', Symbol='Impg2', Aliases='IPM200, PG10.2, Rsbp, Spacrcan', description='interphotoreceptor matrix proteoglycan 2', other_designations='sialoprotein associated with cones and rods proteoglycan', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56203842, end_position_on_the_genomic_accession=56273753, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  224247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224247, CurrentID=0, Status='live', Symbol='E330017A01Rik', Aliases='', description='RIKEN cDNA E330017A01 gene', other_designations='', map_location='16 C1.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58635262, end_position_on_the_genomic_accession=58638403, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  224250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224250, CurrentID=0, Status='live', Symbol='Cldnd1', Aliases='1110019C08Rik, AA407103, AI849195, AW489850, Cldn25', description='claudin domain containing 1', other_designations='claudin 25|claudin-25', map_location='16 C1.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58727910, end_position_on_the_genomic_accession=58734247, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  224273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224273, CurrentID=0, Status='live', Symbol='Crybg3', Aliases='5031404N19, AU022790', description='beta-gamma crystallin domain containing 3', other_designations='', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59490775, end_position_on_the_genomic_accession=59601399, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  224291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224291, CurrentID=0, Status='live', Symbol='Csnka2ip', Aliases='Ckt2', description='casein kinase 2, alpha prime interacting protein', other_designations='', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=64477810, end_position_on_the_genomic_accession=64602366, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  224318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224318, CurrentID=0, Status='live', Symbol='Speer2', Aliases='SPEER-2', description='spermatogenesis associated glutamate (E)-rich protein 2', other_designations='SPEER 2|sperm-associated glutamate (E)-rich protein 2|spermatogenic-associated glutamate (E)-rich protein 2', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=69856874, end_position_on_the_genomic_accession=69863794, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  224344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224344, CurrentID=0, Status='live', Symbol='Rbm11', Aliases='A330018F01', description='RNA binding motif protein 11', other_designations='RNA-binding motif protein 11|putative RNA-binding protein 11', map_location='16 C3.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=75592844, end_position_on_the_genomic_accession=75602825, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  224405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224405, CurrentID=0, Status='live', Symbol='Cyyr1', Aliases='', description='cysteine and tyrosine-rich protein 1', other_designations='proline-rich domain-containing protein', map_location='16 47.45 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=85450357, end_position_on_the_genomic_accession=85550905, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  224419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224419, CurrentID=0, Status='live', Symbol='Map3k7cl', Aliases='C21orf7, ORF63, Tak1l', description='Map3k7 C-terminal like', other_designations='TAK1-like protein', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=87553330, end_position_on_the_genomic_accession=87595336, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  224432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224432, CurrentID=0, Status='live', Symbol='Scaf4', Aliases='AA517739, Sfrs15, Sra4, Srsf15', description='SR-related CTD-associated factor 4', other_designations='CTD-binding SR-like protein RA4|serine/arginine-rich splicing factor 15', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90228951, end_position_on_the_genomic_accession=90284498, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  224440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224440, CurrentID=0, Status='live', Symbol='Setd4', Aliases='AY037804, C21orf18, ORF21', description='SET domain containing 4', other_designations='', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93583457, end_position_on_the_genomic_accession=93604833, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  224454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224454, CurrentID=0, Status='live', Symbol='Zdhhc14', Aliases='B530001K09Rik, New1cp', description='zinc finger, DHHC domain containing 14', other_designations='DHHC-14|NEW1 domain containing protein|NEW1 domain-containing protein|zinc finger DHHC domain-containing protein 14', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=5491946, end_position_on_the_genomic_accession=5753891, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  224480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224480, CurrentID=0, Status='live', Symbol='Nox3', Aliases='GP91-3, het, nmf250', description='NADPH oxidase 3', other_designations='head-tilt', map_location='17 2.05 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=3635240, end_position_on_the_genomic_accession=3696261, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  224481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224481, CurrentID=0, Status='live', Symbol='Tfb1m', Aliases='AI429207', description='transcription factor B1, mitochondrial', other_designations="S-adenosylmethionine-6-N', N'-adenosyl(rRNA) dimethyltransferase 1|mitochondrial 12S rRNA dimethylase 1|mitochondrial transcription factor B1|mtTFB1", map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=3519256, end_position_on_the_genomic_accession=3557822, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  224530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224530, CurrentID=0, Status='live', Symbol='Acat3', Aliases='ACTL', description='acetyl-Coenzyme A acetyltransferase 3', other_designations='', map_location='17 8.72 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12923959, end_position_on_the_genomic_accession=12940396, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  224585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224585, CurrentID=0, Status='live', Symbol='Zfp160', Aliases='6720480D16Rik, F11, HKr18, HZF5, KR18, mKIAA1611, mszf47', description='zinc finger protein 160', other_designations='', map_location='17 11.4 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21008941, end_position_on_the_genomic_accession=21028857, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  224598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224598, CurrentID=0, Status='live', Symbol='Zfp758', Aliases='BC021442', description='zinc finger protein 758', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=22361453, end_position_on_the_genomic_accession=22393909, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  224613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224613, CurrentID=0, Status='live', Symbol='Flywch1', Aliases='E030034P13Rik, mKIAA1552', description='FLYWCH-type zinc finger 1', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23752797, end_position_on_the_genomic_accession=23771591, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  224617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224617, CurrentID=0, Status='live', Symbol='Tbc1d24', Aliases='9630033P11, C530046L02Rik, mKIAA1171', description='TBC1 domain family, member 24', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24175431, end_position_on_the_genomic_accession=24205562, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  224619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224619, CurrentID=0, Status='live', Symbol='Traf7', Aliases='RFWD1', description='TNF receptor-associated factor 7', other_designations='ring finger and WD repeat domain 1|traf7', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24508850, end_position_on_the_genomic_accession=24527938, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  100101484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100101484, CurrentID=0, Status='live', Symbol='Trac', Aliases='Gm16914, Tcra, Tcra-C', description='T cell receptor alpha constant', other_designations='T-cell receptor alpha, constant region', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54220521, end_position_on_the_genomic_accession=54224198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  224624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224624, CurrentID=0, Status='live', Symbol='Rab40c', Aliases='RAR3', description='Rab40C, member RAS oncogene family', other_designations='RAR (RAS like GTPASE) like|SOCS box containing protein RAR3|SOCS box-containing protein RAR3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25882114, end_position_on_the_genomic_accession=25919714, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  224630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224630, CurrentID=0, Status='live', Symbol='Bnip1', Aliases='2010005M06Rik, 5930429G21Rik, Sec20', description='BCL2/adenovirus E1B interacting protein 1', other_designations='BCL2/adenovirus E1B 19kDa-interacting protein 1, NIP1|BCL2/adenovirus E1B interacting protein 1, NIP1|SNARE protein SEC20|bnip1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26781079, end_position_on_the_genomic_accession=26792521, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  224640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224640, CurrentID=0, Status='live', Symbol='Lemd2', Aliases='BC026588, Lem2, NET25', description='LEM domain containing 2', other_designations='nuclear envelope constituent|nuclear envelope transmembrane protein 25', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27189600, end_position_on_the_genomic_accession=27204515, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  224647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224647, CurrentID=0, Status='live', Symbol='D17Wsu92e', Aliases='AU020189, C030010A15, C80239', description='DNA segment, Chr 17, Wayne State University 92, expressed', other_designations='', map_location='17 14.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27751232, end_position_on_the_genomic_accession=27820542, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  224648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224648, CurrentID=0, Status='live', Symbol='Uhrf1bp1', Aliases='1110020K19Rik, AA408674, F830021D11Rik, mKIAA4127', description='UHRF1 (ICBP90) binding protein 1', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27856507, end_position_on_the_genomic_accession=27900040, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  224650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224650, CurrentID=0, Status='live', Symbol='Anks1', Aliases='Anks1a, mKIAA0229', description='ankyrin repeat and SAM domain containing 1', other_designations='Odin', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27909306, end_position_on_the_genomic_accession=28062779, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  224656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224656, CurrentID=0, Status='live', Symbol='Zfp523', Aliases='2810027J07, BC025615, Znf76', description='zinc finger protein 523', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28174986, end_position_on_the_genomic_accession=28205886, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  224661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224661, CurrentID=0, Status='live', Symbol='Slc26a8', Aliases='', description='solute carrier family 26, member 8', other_designations='anion exchange transporter|slc26a8', map_location='17 14.82 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28637779, end_position_on_the_genomic_accession=28691237, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  332175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=332175, CurrentID=0, Status='live', Symbol='Zdhhc23', Aliases='Gm1751, Gm779, nidd', description='zinc finger, DHHC domain containing 23', other_designations='DHHC-23|DHHC-containing protein 11|nNOS-intereacting DHHC-containing Dem protein-L|probable palmitoyltransferase ZDHHC23|zinc finger DHHC domain-containing protein 23', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=43965033, end_position_on_the_genomic_accession=43979783, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  224671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224671, CurrentID=0, Status='live', Symbol='Btbd9', Aliases='1700023F20Rik, 4930402L05, AI464421', description='BTB (POZ) domain containing 9', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=30215524, end_position_on_the_genomic_accession=30577530, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  224674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224674, CurrentID=0, Status='live', Symbol='Slc37a1', Aliases='G3PP', description='solute carrier family 37 (glycerol-3-phosphate transporter), member 1', other_designations='', map_location='17 A3.3-B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31294798, end_position_on_the_genomic_accession=31350709, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  224691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224691, CurrentID=0, Status='live', Symbol='Zfp472', Aliases='KRIM-1, Krim-1A, Krim-1B, Krim1', description='zinc finger protein 472', other_designations='KRAB box containing zinc finger protein|Krab box proteins interacting with myc 1|zin finger protein 472', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32965831, end_position_on_the_genomic_accession=32979211, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  224694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224694, CurrentID=0, Status='live', Symbol='Zfp81', Aliases='AA409460, AA409461, C330034P10Rik, D330034E10Rik, Hszfp36, KRAB13, Zfp78', description='zinc finger protein 81', other_designations='ZFP-36 zinc finger protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33329342, end_position_on_the_genomic_accession=33359013, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  224697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224697, CurrentID=0, Status='live', Symbol='Adamts10', Aliases='9430006O18, AW045948, Adam-ts10, Adamts-10, Znmp', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 10', other_designations='a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 10|zinc metalloendopeptidase', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33524176, end_position_on_the_genomic_accession=33553782, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  224703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224703, CurrentID=0, Status='live', Symbol='March2', Aliases='9530046H09Rik, MARCH-II', description='membrane-associated ring finger (C3HC4) 2', other_designations='membrane-associated RING finger protein 2|membrane-associated RING-CH protein II', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33685692, end_position_on_the_genomic_accession=33718677, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  224705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224705, CurrentID=0, Status='live', Symbol='Vps52', Aliases='ARE1, D130068D18, D430041K17Rik, SAC2, Sacm2l, tclw5', description='vacuolar protein sorting 52 (yeast)', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33955785, end_position_on_the_genomic_accession=33966495, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  224727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224727, CurrentID=0, Status='live', Symbol='Bag6', Aliases='2410045D21Rik, AA408914, BAG-6, Bat3, D17H6S52E, G3, Scythe', description='BCL2-associated athanogene 6', other_designations='BAG family molecular chaperone regulator 6|HLA-B-associated transcript 3|large proline-rich protein BAT3|protein Scythe', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35134600, end_position_on_the_genomic_accession=35147322, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  93670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93670, CurrentID=0, Status='live', Symbol='Tac4', Aliases='AW489379, PPT-C', description='tachykinin 4', other_designations='hemokinin I (HK-1)|preprotachykinin-C|tac4', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95261529, end_position_on_the_genomic_accession=95269262, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93671, CurrentID=0, Status='live', Symbol='Cd163', Aliases='CD163v2, CD163v3', description='CD163 antigen', other_designations='', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124304631, end_position_on_the_genomic_accession=124330529, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  93672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93672, CurrentID=0, Status='live', Symbol='Il24', Aliases='FISP, Mda-7, Mda7, St16', description='interleukin 24', other_designations='IL-24|IL-4-induced secreted protein|melanoma differentiation-associated gene 7 protein|th2-specific cytokine FISP', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130882074, end_position_on_the_genomic_accession=130887413, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  93673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93673, CurrentID=0, Status='live', Symbol='Nat8f2', Aliases='', description='N-acetyltransferase 8 (GCN5-related) family member 2', other_designations='camello-like 2|camello-like protein 2|cml2', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85865422, end_position_on_the_genomic_accession=85869137, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93674, CurrentID=0, Status='live', Symbol='Nat8f3', Aliases='4833414A14Rik, 4930583K21Rik', description='N-acetyltransferase 8 (GCN5-related) family member 3', other_designations='camello-like 3|camello-like protein 3|cml3', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85760631, end_position_on_the_genomic_accession=85765785, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93675, CurrentID=0, Status='live', Symbol='Clec2i', Aliases='Clr-g, Clrg, Dcl1, LCL-1, OCILrP2', description='C-type lectin domain family 2, member i', other_designations='C-type lectin-related protein DCL1|C-type lectin-related protein G|lymphoid-derived C-type lectin-1|ocil-related protein 2|osteoclast inhibitory lectin related protein 2', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128887573, end_position_on_the_genomic_accession=128898167, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  93677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93677, CurrentID=0, Status='live', Symbol='Lmod2', Aliases='C-Lmod', description='leiomodin 2 (cardiac)', other_designations='cardiac leiomodin', map_location='6 A3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=24597771, end_position_on_the_genomic_accession=24605414, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93679, CurrentID=0, Status='live', Symbol='Trim8', Aliases='AA408830, BB006755, GERP, Rnf27', description='tripartite motif-containing 8', other_designations='glioblastoma-expressed RING finger protein|ring finger protein 27|tripartite motif protein 8|tripartite motif-containing protein 8', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46501648, end_position_on_the_genomic_accession=46516805, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  224753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224753, CurrentID=0, Status='live', Symbol='H2-M10.4', Aliases='15H', description='histocompatibility 2, M region locus 10.4', other_designations='major histocompatibility complex class Ib M10.5', map_location='17 19.09 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36458917, end_position_on_the_genomic_accession=36462329, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  93683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93683, CurrentID=0, Status='live', Symbol='Glce', Aliases='1110017N23Rik, AI747411, C130034A12Rik, Hsepi', description='glucuronyl C5-epimerase', other_designations='heparan sulfate C5-epimerase|heparan sulfate-glucuronic acid C5-epimerase|heparin sulfate C5-epimerase|heparin/heparan sulfate:glucuronic acid C5-epimerase|heparosan-N-sulfate-glucuronate 5-epimerase', map_location='9 C-D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62057248, end_position_on_the_genomic_accession=62122720, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  224756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224756, CurrentID=0, Status='live', Symbol='H2-M1', Aliases='H-2M1, H2-Qa-Mb1, Mb1', description='histocompatibility 2, M region locus 1', other_designations='', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36669030, end_position_on_the_genomic_accession=36672197, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  93685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93685, CurrentID=0, Status='live', Symbol='Entpd7', Aliases='1810012B13Rik, 1810020C02Rik, 2810003F23Rik, LALP1, Lysal2', description='ectonucleoside triphosphate diphosphohydrolase 7', other_designations='NTPDase 7|lysosomal apyrase-like 2|lysosomal apyrase-like protein 1', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43689689, end_position_on_the_genomic_accession=43733853, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  93686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93686, CurrentID=0, Status='live', Symbol='Rbfox2', Aliases='Fbm2, Fxh, Hrnbp2, Rbm9', description='RNA binding protein, fox-1 homolog (C. elegans) 2', other_designations='Fyn-binding molecule 2|RNA-binding motif protein 9|fox-1 homolog B|fox-1 homolog Fxh|hexaribonucleotide-binding protein 2', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77078990, end_position_on_the_genomic_accession=77307053, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  93687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93687, CurrentID=0, Status='live', Symbol='Csnk1a1', Aliases='2610208K14Rik, 4632404G05Rik, 5430427P18Rik, CK1a, Csnk1a', description='casein kinase 1, alpha 1', other_designations='CK1|CKI-alpha|casein kinase I-alpha', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61555165, end_position_on_the_genomic_accession=61590315, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  93688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93688, CurrentID=0, Status='live', Symbol='Klhl1', Aliases='mKIAA1490', description='kelch-like 1', other_designations='kelch-l like', map_location='14 E2.1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=96105265, end_position_on_the_genomic_accession=96519034, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  93689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93689, CurrentID=0, Status='live', Symbol='Lmod1', Aliases='9530015K06Rik, SM-Lmod', description='leiomodin 1 (smooth muscle)', other_designations='', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135324813, end_position_on_the_genomic_accession=135368065, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  224762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224762, CurrentID=0, Status='live', Symbol='Trim31', Aliases='BC026666, HCG1, HCGI, RNF', description='tripartite motif-containing 31', other_designations='tripartite motif-containing protein 31', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36898130, end_position_on_the_genomic_accession=36910217, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  93691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93691, CurrentID=0, Status='live', Symbol='Klf7', Aliases='9830124P08Rik', description='Kruppel-like factor 7 (ubiquitous)', other_designations='', map_location='1 C1-C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=64033727, end_position_on_the_genomic_accession=64122341, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  93692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93692, CurrentID=0, Status='live', Symbol='Glrx', Aliases='C86710, D13Wsu156e1, Grx1, TTase, Glrx', description='glutaredoxin', other_designations='TTase-1|glutaredoxin (thioltransferase)|glutaredoxin 1 (thioltransferase)|thioltransferase-1', map_location='13 40.95 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=75839873, end_position_on_the_genomic_accession=75850151, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93694, CurrentID=0, Status='live', Symbol='Clec2d', Aliases='Clr-b, Clrb, Ocil', description='C-type lectin domain family 2, member d', other_designations='C-type lectin-domain family 2 member D|C-type lectin-related protein B|lectin-like transmembrane protein|osteoclast inhibitory lectin', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129180615, end_position_on_the_genomic_accession=129186535, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  93695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93695, CurrentID=0, Status='live', Symbol='Gpnmb', Aliases='DC-HIL, Dchil, ipd', description='glycoprotein (transmembrane) nmb', other_designations='dendritic cell-associated heparin sulfate proteoglycan-dependent integrin ligand|dendritic cell-associated transmembrane protein|iris pigment dispersion|osteoactivin', map_location='6 23.82 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49036518, end_position_on_the_genomic_accession=49058182, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  93696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93696, CurrentID=0, Status='live', Symbol='Chrac1', Aliases='2410152E03Rik, 2810406L04Rik, YCL1', description='chromatin accessibility complex 1', other_designations='CHRAC-1|DNA polymerase epsilon subunit p15|NF-YC-like protein|YC-like protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73090412, end_position_on_the_genomic_accession=73094075, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  93697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93697, CurrentID=0, Status='live', Symbol='Ice2', Aliases='B230343B06Rik, Narg2', description='interactor of little elongation complex ELL subunit 2', other_designations='NMDA receptor-regulated gene 2|NMDA receptor-regulated protein 2|interactor of little elongator complex ELL subunit 2', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=69397911, end_position_on_the_genomic_accession=69433076, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  93699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93699, CurrentID=0, Status='live', Symbol='Pcdhgb1', Aliases='', description='protocadherin gamma subfamily B, 1', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37680458, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93700, CurrentID=0, Status='live', Symbol='Pcdhgb2', Aliases='', description='protocadherin gamma subfamily B, 2', other_designations="3' RACE product|protocadherin gamma B2", map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37689859, end_position_on_the_genomic_accession=37841872, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93701, CurrentID=0, Status='live', Symbol='Pcdhgb4', Aliases='', description='protocadherin gamma subfamily B, 4', other_designations='protocadherin gamma C-VI', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37720554, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93702, CurrentID=0, Status='live', Symbol='Pcdhgb5', Aliases='', description='protocadherin gamma subfamily B, 5', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37731154, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93703, CurrentID=0, Status='live', Symbol='Pcdhgb6', Aliases='', description='protocadherin gamma subfamily B, 6', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37742094, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93704, CurrentID=0, Status='live', Symbol='Pcdhgb7', Aliases='', description='protocadherin gamma subfamily B, 7', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37751779, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93705, CurrentID=0, Status='live', Symbol='Pcdhgb8', Aliases='', description='protocadherin gamma subfamily B, 8', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37761801, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93706, CurrentID=0, Status='live', Symbol='Pcdhgc3', Aliases='PC43, Pcdh2', description='protocadherin gamma subfamily C, 3', other_designations='protocadherin 2', map_location='18 19.67 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37806410, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93707, CurrentID=0, Status='live', Symbol='Pcdhgc4', Aliases='', description='protocadherin gamma subfamily C, 4', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37815079, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93708, CurrentID=0, Status='live', Symbol='Pcdhgc5', Aliases='', description='protocadherin gamma subfamily C, 5', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37819546, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93709, CurrentID=0, Status='live', Symbol='Pcdhga1', Aliases='', description='protocadherin gamma subfamily A, 1', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37661945, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93710, CurrentID=0, Status='live', Symbol='Pcdhga2', Aliases='mKIAA0588', description='protocadherin gamma subfamily A, 2', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37669105, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93711, CurrentID=0, Status='live', Symbol='Pcdhga3', Aliases='', description='protocadherin gamma subfamily A, 3', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37674335, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93712, CurrentID=0, Status='live', Symbol='Pcdhga4', Aliases='', description='protocadherin gamma subfamily A, 4', other_designations='PCDH-gamma-A4|pcdhga4', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37685400, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93713, CurrentID=0, Status='live', Symbol='Pcdhga5', Aliases='', description='protocadherin gamma subfamily A, 5', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37694501, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93714, CurrentID=0, Status='live', Symbol='Pcdhga6', Aliases='', description='protocadherin gamma subfamily A, 6', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37707229, end_position_on_the_genomic_accession=37841870, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93715, CurrentID=0, Status='live', Symbol='Pcdhga7', Aliases='', description='protocadherin gamma subfamily A, 7', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37714834, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93716, CurrentID=0, Status='live', Symbol='Pcdhga8', Aliases='mKIAA4054', description='protocadherin gamma subfamily A, 8', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37725706, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93717, CurrentID=0, Status='live', Symbol='Pcdhga9', Aliases='', description='protocadherin gamma subfamily A, 9', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37736936, end_position_on_the_genomic_accession=37841863, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93719, CurrentID=0, Status='live', Symbol='Ear6', Aliases='', description='eosinophil-associated, ribonuclease A family, member 6', other_designations='eosinophil-associated ribonuclease 6', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51853768, end_position_on_the_genomic_accession=51854643, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  224792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224792, CurrentID=0, Status='live', Symbol='Adgrf5', Aliases='8430401C09Rik, 9330185D23, AI528491, Gpr116, mKIAA0758', description='adhesion G protein-coupled receptor F5', other_designations='G protein-coupled receptor 116|G-protein coupled receptor 116|probable G-protein coupled receptor 116', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43360329, end_position_on_the_genomic_accession=43459557, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  93721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93721, CurrentID=0, Status='live', Symbol='Cpn1', Aliases='0610011F20Rik, CPN', description='carboxypeptidase N, polypeptide 1', other_designations='anaphylatoxin inactivator|carboxypeptidase N small subunit|carboxypeptidase N, polypeptide 1, 50kD|kininase I', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43956307, end_position_on_the_genomic_accession=43986551, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  224794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224794, CurrentID=0, Status='live', Symbol='Enpp4', Aliases='4933413N07Rik, AA986363, AI195364, E-NPP 4, Gm90, mKIAA0879', description='ectonucleotide pyrophosphatase/phosphodiesterase 4', other_designations='AP3A hydrolase|AP3Aase|E-NPP4|NPP-4|ectonucleotide pyrophosphatase/phosphodiesterase family member 4', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=44096307, end_position_on_the_genomic_accession=44105924, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  93723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93723, CurrentID=0, Status='live', Symbol='Pcdhga11', Aliases='', description='protocadherin gamma subfamily A, 11', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37755773, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93724, CurrentID=0, Status='live', Symbol='Pcdhga12', Aliases='Pcdh13, pc2c', description='protocadherin gamma subfamily A, 12', other_designations='protocadherin 13|protocadherin gamma A12|protocadherin gamma B6', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37765580, end_position_on_the_genomic_accession=37841873, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93725, CurrentID=0, Status='live', Symbol='Ear10', Aliases='Raf2', description='eosinophil-associated, ribonuclease A family, member 10', other_designations='ribonuclease A f2', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=43922898, end_position_on_the_genomic_accession=43923368, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  93726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93726, CurrentID=0, Status='live', Symbol='Rnase2a', Aliases='Ear11', description='ribonuclease, RNase A family, 2A (liver, eosinophil-derived neurotoxin)', other_designations='eosinophil-associated ribonuclease 11|eosinophil-associated, ribonuclease A family, member 11', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51255261, end_position_on_the_genomic_accession=51256112, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  93728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93728, CurrentID=0, Status='live', Symbol='Pabpc5', Aliases='C820015E17', description='poly(A) binding protein, cytoplasmic 5', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=119927196, end_position_on_the_genomic_accession=119930165, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  93730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93730, CurrentID=0, Status='live', Symbol='Lztfl1', Aliases='5530402H04Rik, 6130400H19Rik, AI414725, AW048545', description='leucine zipper transcription factor-like 1', other_designations='', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123694413, end_position_on_the_genomic_accession=123717668, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  93732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93732, CurrentID=0, Status='live', Symbol='Acox2', Aliases='THCCox', description='acyl-Coenzyme A oxidase 2, branched chain', other_designations='3-alpha,7-alpha,12-alpha-trihydroxy-5-beta-cholestanoyl-CoA 24-hydroxylase|3-alpha,7-alpha,12-alpha-trihydroxy-5-beta-cholestanoyl-CoA oxidase|THCA-CoA oxidase|branched chain acyl-CoA oxidase|trihydroxycoprostanoyl-CoA oxidase', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=8225511, end_position_on_the_genomic_accession=8259019, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  224805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224805, CurrentID=0, Status='live', Symbol='Aars2', Aliases='Aarsl, AlaRS, Gm89', description='alanyl-tRNA synthetase 2, mitochondrial', other_designations='alaRS|alanyl-tRNA synthetase 2, mitochondrial (putative)|alanyl-tRNA synthetase like|alanyl-tRNA synthetase, mitochondrial|probable alanyl-tRNA synthetase, mitochondrial', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45506806, end_position_on_the_genomic_accession=45520843, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  93734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93734, CurrentID=0, Status='live', Symbol='Mpv17l', Aliases='M-LP', description='Mpv17 transgene, kidney disease mutant-like', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13903161, end_position_on_the_genomic_accession=13949619, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  224807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224807, CurrentID=0, Status='live', Symbol='Tmem63b', Aliases='BC026370', description='transmembrane protein 63b', other_designations='', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45660172, end_position_on_the_genomic_accession=45686254, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  93736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93736, CurrentID=0, Status='live', Symbol='Aff4', Aliases='AF5Q31, Alf4, Laf4l, MCEF', description='AF4/FMR2 family, member 4', other_designations='ALL1 fused gene from 5q31', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53350307, end_position_on_the_genomic_accession=53421830, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  93737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93737, CurrentID=0, Status='live', Symbol='Pard6g', Aliases='2410049N21Rik, Par6a', description='par-6 family cell polarity regulator gamma', other_designations='PAR-6 gamma|par-6 partitioning defective 6 homolog gamma|partitioning-defective protein 6, alpha', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80046872, end_position_on_the_genomic_accession=80119640, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  93739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93739, CurrentID=0, Status='live', Symbol='Gabarapl2', Aliases='0610012F20Rik, 2900019O08Rik, AI173605, GATE-16, Gef2', description='gamma-aminobutyric acid (GABA) A receptor-associated protein-like 2', other_designations='GABA(A) receptor-associated protein like 2|GABA(A) receptor-associated protein-like 2|gamma-aminobutyric acid (GABA-A) receptor-associated protein-like 2|ganglioside expression factor 2|golgi-associated ATPase enhancer of 16 kDa', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111940703, end_position_on_the_genomic_accession=111952915, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  224813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224813, CurrentID=0, Status='live', Symbol='Lrrc73', Aliases='Gm88', description='leucine rich repeat containing 73', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46253445, end_position_on_the_genomic_accession=46257316, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  224814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224814, CurrentID=0, Status='live', Symbol='Abcc10', Aliases='Mrp7, mFLJ00002', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 10', other_designations='ATP-binding cassette sub-family C member 10', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46303230, end_position_on_the_genomic_accession=46328032, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  93746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93746, CurrentID=0, Status='live', Symbol='Gprc5d', Aliases='', description='G protein-coupled receptor, family C, group 5, member D', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135105991, end_position_on_the_genomic_accession=135137443, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  93747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93747, CurrentID=0, Status='live', Symbol='Echs1', Aliases='C80529', description='enoyl Coenzyme A hydratase, short chain, 1, mitochondrial', other_designations='SCEH|enoyl-CoA hydratase 1|short chain enoyl-CoA hydratase|short-chain enoyl-CoA hydratase', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140105723, end_position_on_the_genomic_accession=140116423, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  224823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224823, CurrentID=0, Status='live', Symbol='Rrp36', Aliases='', description='ribosomal RNA processing 36 homolog (S. cerevisiae)', other_designations='RNA processing factor', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46667442, end_position_on_the_genomic_accession=46674451, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  224824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224824, CurrentID=0, Status='live', Symbol='Pex6', Aliases='AI132582, D130055I09Rik, mKIAA4177', description='peroxisomal biogenesis factor 6', other_designations='PAF-2|peroxin-6|peroxisomal-type ATPase 1', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46711463, end_position_on_the_genomic_accession=46725541, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  224826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224826, CurrentID=0, Status='live', Symbol='Ubr2', Aliases='9930021A08Rik, AI462103, AW540746, E130209G04Rik, mKIAA0349', description='ubiquitin protein ligase E3 component n-recognin 2', other_designations='N-recognin-2|ubiquitin ligase E3 alpha-II|ubiquitin-protein ligase E3-alpha-2|ubiquitin-protein ligase E3-alpha-II', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46928291, end_position_on_the_genomic_accession=47010775, orientation='minus', exon_count=48, OMIM='', no_hdr0=''),
  224829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224829, CurrentID=0, Status='live', Symbol='Trerf1', Aliases='9430096I18Rik, AI429294, B830015H24, RAPA, Trep-132, Trep132', description='transcriptional regulating factor 1', other_designations='transcriptional regulating protein 132|transcriptional-regulating protein 132|zinc finger transcription factor TReP-132', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47140942, end_position_on_the_genomic_accession=47359458, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  93759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93759, CurrentID=0, Status='live', Symbol='Sirt1', Aliases='AA673258, SIR2L1, Sir2, Sir2a, Sir2alpha', description='sirtuin 1', other_designations='SIR2-like protein 1|regulatory protein SIR2 homolog 1|sirtuin 1 ((silent mating type information regulation 2, homolog) 1', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=63319005, end_position_on_the_genomic_accession=63339061, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  93760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93760, CurrentID=0, Status='live', Symbol='Arid1a', Aliases='1110030E03Rik, BAF250a, Osa1, Smarcf1', description='AT rich interactive domain 1A (SWI-like)', other_designations='ARID domain-containing protein 1A|AT rich interactive domain 1A (Swi1 like)|BAF250|BRG1-associated factor 250|BRG1-associated factor 250a|SWI-SNF complex protein p270|SWI-like protein|SWI/SNF complex protein p270|SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily f, member 1|SWI/SNF-related, matrix-associated, actin-dependent regulator of chromatin subfamily F member 1|Swi1 like|osa homolog 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133679008, end_position_on_the_genomic_accession=133753981, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  224833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224833, CurrentID=0, Status='live', Symbol='AI661453', Aliases='', description='expressed sequence AI661453', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47436639, end_position_on_the_genomic_accession=47470638, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  93762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93762, CurrentID=0, Status='live', Symbol='Smarca5', Aliases='4933427E24Rik, D030040M08Rik, D330027N15Rik, MommeD4, Snf2h', description='SWI/SNF related, matrix associated, actin dependent regulator of chromatin, subfamily a, member 5', other_designations='mSnf2h|sucrose nonfermenting protein 2 homolog', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=80699943, end_position_on_the_genomic_accession=80739459, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  224836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224836, CurrentID=0, Status='live', Symbol='Usp49', Aliases='C330046L10Rik, Gm545', description='ubiquitin specific peptidase 49', other_designations='deubiquitinating enzyme 49|ubiquitin specific protease 49|ubiquitin thioesterase 49|ubiquitin thiolesterase 49|ubiquitin-specific-processing protease 49', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47630690, end_position_on_the_genomic_accession=47684067, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  93765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93765, CurrentID=0, Status='live', Symbol='Ube2n', Aliases='1500026J17Rik, AL022654, BB101821, UBC13', description='ubiquitin-conjugating enzyme E2N', other_designations='E2 ubiquitin-conjugating enzyme N|bendless-like ubiquitin-conjugating enzyme|ubiquitin carrier protein N|ubiquitin-protein ligase N', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=95514706, end_position_on_the_genomic_accession=95545658, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  224840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224840, CurrentID=0, Status='live', Symbol='Treml4', Aliases='5031403H21Rik, BB137214, IDCP1, TLT-4, TLT4, Treml3', description='triggering receptor expressed on myeloid cells-like 4', other_designations='IG-like domain containing protein IDCP1c|triggering receptor expressed on myeloid cells-like 3|triggering receptor expressed on myeloid cells-like 3b|triggering receptor expressed on myeloid cells-like protein 3|triggering receptor expressed on myeloid cells-like protein 4', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48264191, end_position_on_the_genomic_accession=48275607, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100504164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504164, CurrentID=0, Status='live', Symbol='Cmtm1', Aliases='CHLFH1a, CKLFH, CKLFH1, Cklfsf1', description='CKLF-like MARVEL transmembrane domain containing 1', other_designations='chemokine-like factor super family 1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104293542, end_position_on_the_genomic_accession=104310145, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  224860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224860, CurrentID=0, Status='live', Symbol='Plcl2', Aliases='PLC-L2, PRIP-2, Plce2, mKIAA1092', description='phospholipase C-like 2', other_designations='PLC-L(2)|PLC-epsilon-2|phospholipase C, epsilon 2|phospholipase C-L2|phospholipase C-epsilon-2', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=50508633, end_position_on_the_genomic_accession=50688494, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  93790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93790, CurrentID=0, Status='live', Symbol='Nipa2', Aliases='2600017P10Rik, 3830408P04Rik, AB041581', description='non imprinted in Prader-Willi/Angelman syndrome 2 homolog (human)', other_designations='non-imprinted in Prader-Willi/Angelman syndrome region protein 2 homolog', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=55931266, end_position_on_the_genomic_accession=55962655, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  68591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68591, CurrentID=0, Status='live', Symbol='Mocos', Aliases='1110018O12Rik, HMCS, MCS, MOS', description='molybdenum cofactor sulfurase', other_designations='MCS|MOS|moCo sulfurase|molybdenum cofactor sulfurtransferase', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24653176, end_position_on_the_genomic_accession=24701577, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  93806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93806, CurrentID=0, Status='live', Symbol='Serpinb9g', Aliases='1600002F03Rik, AW545779, NK21B', description='serine (or cysteine) peptidase inhibitor, clade B, member 9g', other_designations='ovalbumin', map_location='13 13.9 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33484790, end_position_on_the_genomic_accession=33496002, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  224893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224893, CurrentID=0, Status='live', Symbol='Zfp959', Aliases='', description='zinc finger protein 959', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55891097, end_position_on_the_genomic_accession=55898935, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  224897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224897, CurrentID=0, Status='live', Symbol='Dpp9', Aliases='6430584G11Rik, A330078I11, DPP IX, DPRP2', description='dipeptidylpeptidase 9', other_designations='DP9|DPLP9|dipeptidyl peptidase IX|dipeptidyl peptidase-like protein 9', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56186674, end_position_on_the_genomic_accession=56218889, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  224902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224902, CurrentID=0, Status='live', Symbol='Safb2', Aliases='AA389433, AI255170, mKIAA0138', description='scaffold attachment factor B2', other_designations='SAF-B2', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56560971, end_position_on_the_genomic_accession=56584672, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  224903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224903, CurrentID=0, Status='live', Symbol='Safb', Aliases='3110021E02Rik, 5330423C17Rik, AU018122, D18386, E130307D12, HAP, HET, SAF-B1, SAFB1', description='scaffold attachment factor B', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56584760, end_position_on_the_genomic_accession=56606294, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  224904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224904, CurrentID=0, Status='live', Symbol='2410015M20Rik', Aliases='Mic13, QIL1, sr104', description='RIKEN cDNA 2410015M20 gene', other_designations='protein QIL1', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56607452, end_position_on_the_genomic_accession=56609771, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  93834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93834, CurrentID=0, Status='live', Symbol='Peli2', Aliases='AW047589, BB129927', description='pellino 2', other_designations='pellino-2|protein pellino homolog 2', map_location='14 24.6 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=48120778, end_position_on_the_genomic_accession=48308103, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  224907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224907, CurrentID=0, Status='live', Symbol='Dus3l', Aliases='AI662135, AW557805', description='dihydrouridine synthase 3-like (S. cerevisiae)', other_designations='putative zinc finger protein MA23', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56764717, end_position_on_the_genomic_accession=56770093, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  93836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93836, CurrentID=0, Status='live', Symbol='Rnf111', Aliases='ARK, Arkadia', description='ring finger 111', other_designations='RING finger protein 111', map_location='9 39.53 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70425424, end_position_on_the_genomic_accession=70503725, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  93837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93837, CurrentID=0, Status='live', Symbol='Dach2', Aliases='9430028N04Rik', description='dachshund 2 (Drosophila)', other_designations='', map_location='X 49.13 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=113297510, end_position_on_the_genomic_accession=113836386, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  93838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93838, CurrentID=0, Status='live', Symbol='Dqx1', Aliases='2310066E11Rik', description='DEAQ RNA-dependent ATPase', other_designations='DEAQ box polypeptide 1', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83057731, end_position_on_the_genomic_accession=83067219, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  224912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224912, CurrentID=0, Status='live', Symbol='Crb3', Aliases='5730439B18, AU015179, AU041752', description='crumbs family member 3', other_designations='crumbs homolog 3|crumbs protein homolog 3', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57062096, end_position_on_the_genomic_accession=57065920, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  93842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93842, CurrentID=0, Status='live', Symbol='Igsf9', Aliases='644ETD8, Dasm1, Kiaa1355-hp, NRT1, Ncaml, mKIAA1355', description='immunoglobulin superfamily, member 9', other_designations='NCAM-like protein NRT1|dendrite arborization and synapse maturation protein 1|igSF9A|immunoglobulin superfamily member 9A', map_location='1 79.86 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172481776, end_position_on_the_genomic_accession=172498877, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  93843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93843, CurrentID=0, Status='live', Symbol='Pnck', Aliases='Bstk3, CaMKIbeta2, CaMKlb2, Camk1b, Punc, caMKlb1', description='pregnancy upregulated non-ubiquitously expressed CaM kinase', other_designations='caM kinase I beta|caM kinase IB|caM-KI beta|caMKI-beta|calcium/calmodulin-dependent protein kinase 1, beta|pregnancy up-regulated non-ubiquitously-expressed CaM kinase homolog', map_location='X 37.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73655992, end_position_on_the_genomic_accession=73660127, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  224938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224938, CurrentID=0, Status='live', Symbol='Pja2', Aliases='AI447901, AL022700, Neurodap1, mKIAA0438', description='praja 2, RING-H2 motif containing', other_designations='E3 ubiquitin-protein ligase Praja2|RING finger protein 131|neurodegeneration associated protein 1|praja-2|praja2|protein carrying the RING-H2 sequence motif', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=64281006, end_position_on_the_genomic_accession=64331883, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100101806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100101806, CurrentID=0, Status='live', Symbol='Srp54c', Aliases='E330029E12Rik', description='signal recognition particle 54C', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55230105, end_position_on_the_genomic_accession=55279334, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100101807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100101807, CurrentID=0, Status='live', Symbol='1700047I17Rik2', Aliases='', description='RIKEN cDNA 1700047I17 gene 2', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=55199532, end_position_on_the_genomic_accession=55217168, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  93872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93872, CurrentID=0, Status='live', Symbol='Pcdhb1', Aliases='PcdhbA', description='protocadherin beta 1', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37264998, end_position_on_the_genomic_accession=37267454, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93873, CurrentID=0, Status='live', Symbol='Pcdhb2', Aliases='PcdhbB', description='protocadherin beta 2', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37294840, end_position_on_the_genomic_accession=37297614, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93874, CurrentID=0, Status='live', Symbol='Pcdhb3', Aliases='PcdhbC', description='protocadherin beta 3', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37300799, end_position_on_the_genomic_accession=37304585, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93875, CurrentID=0, Status='live', Symbol='Pcdhb4', Aliases='Pcdhb5A, PcdhbD', description='protocadherin beta 4', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37307455, end_position_on_the_genomic_accession=37311173, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93876, CurrentID=0, Status='live', Symbol='Pcdhb5', Aliases='Pcdhb4A, PcdhbE', description='protocadherin beta 5', other_designations='Protocadherin-betaE|cell-cell adhesion molecule', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37320381, end_position_on_the_genomic_accession=37323913, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93877, CurrentID=0, Status='live', Symbol='Pcdhb6', Aliases='Pcdhb5B, PcdhbF', description='protocadherin beta 6', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37334028, end_position_on_the_genomic_accession=37336346, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93878, CurrentID=0, Status='live', Symbol='Pcdhb7', Aliases='Pcdhb4B, PcdhbG', description='protocadherin beta 7', other_designations='protocadherin beta 5', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37341702, end_position_on_the_genomic_accession=37345207, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93879, CurrentID=0, Status='live', Symbol='Pcdhb8', Aliases='Pcdhb5C, PcdhbH', description='protocadherin beta 8', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37355271, end_position_on_the_genomic_accession=37357610, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93880, CurrentID=0, Status='live', Symbol='Pcdhb9', Aliases='Pcdhb4C, PcdhbI', description='protocadherin beta 9', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37400855, end_position_on_the_genomic_accession=37403909, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93881, CurrentID=0, Status='live', Symbol='Pcdhb10', Aliases='Pcdhb5D, PcdhbJ', description='protocadherin beta 10', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37411674, end_position_on_the_genomic_accession=37414514, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93882, CurrentID=0, Status='live', Symbol='Pcdhb11', Aliases='Pcdhb5E, PcdhbK', description='protocadherin beta 11', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37421418, end_position_on_the_genomic_accession=37425032, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93883, CurrentID=0, Status='live', Symbol='Pcdhb12', Aliases='Pcdh3, Pcdhb5F, PcdhbL', description='protocadherin beta 12', other_designations='protocadherin 3', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37435621, end_position_on_the_genomic_accession=37438654, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93884, CurrentID=0, Status='live', Symbol='Pcdhb13', Aliases='Pcdbh6, PcdhbM', description='protocadherin beta 13', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37442517, end_position_on_the_genomic_accession=37446209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93885, CurrentID=0, Status='live', Symbol='Pcdhb14', Aliases='2210006M07Rik, Pcdhb17, PcdhbN', description='protocadherin beta 14', other_designations='PCDH-beta-14|Protocadherin-betaN|cell-cell adhesion molecule', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37447657, end_position_on_the_genomic_accession=37451094, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93886, CurrentID=0, Status='live', Symbol='Pcdhb15', Aliases='Pcdhb7, PcdhbO', description='protocadherin beta 15', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37473546, end_position_on_the_genomic_accession=37476340, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93887, CurrentID=0, Status='live', Symbol='Pcdhb16', Aliases='Pcdhb8, PcdhbP', description='protocadherin beta 16', other_designations='Protocadherin-betaP|cell-cell adhesion molecule', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37477768, end_position_on_the_genomic_accession=37483038, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93888, CurrentID=0, Status='live', Symbol='Pcdhb17', Aliases='Pcdhb16, PcdhbQ, mKIAA1621', description='protocadherin beta 17', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37485021, end_position_on_the_genomic_accession=37488290, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93889, CurrentID=0, Status='live', Symbol='Pcdhb18', Aliases='Pcdhb9, PcdhbR', description='protocadherin beta 18', other_designations='PCDH-beta-18', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37489465, end_position_on_the_genomic_accession=37494503, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93890, CurrentID=0, Status='live', Symbol='Pcdhb19', Aliases='Pcdhb11, PcdhbS', description='protocadherin beta 19', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37496998, end_position_on_the_genomic_accession=37501711, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93891, CurrentID=0, Status='live', Symbol='Pcdhb20', Aliases='Pcdhb14, PcdhbT', description='protocadherin beta 20', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37504264, end_position_on_the_genomic_accession=37507660, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93892, CurrentID=0, Status='live', Symbol='Pcdhb21', Aliases='Pcdhb18, PcdhbU', description='protocadherin beta 21', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37513652, end_position_on_the_genomic_accession=37516363, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93893, CurrentID=0, Status='live', Symbol='Pcdhb22', Aliases='Pcdhb15, PcdhbV', description='protocadherin beta 22', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37518353, end_position_on_the_genomic_accession=37521419, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93896, CurrentID=0, Status='live', Symbol='Glp2r', Aliases='9530092J08Rik, GLP-2', description='glucagon-like peptide 2 receptor', other_designations='G protein-coupled receptor GLP2R|GLP-2 receptor|GLP-2-R|GLP-2R', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=67663299, end_position_on_the_genomic_accession=67771153, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  93897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93897, CurrentID=0, Status='live', Symbol='Fzd10', Aliases='Fz-10', description='frizzled class receptor 10', other_designations='frizzled homolog 10', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=128601106, end_position_on_the_genomic_accession=128604093, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  93898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93898, CurrentID=0, Status='live', Symbol='Cers1', Aliases='Lass1, Uog-1, to', description='ceramide synthase 1', other_designations='LAG1 homolog, ceramide synthase 1|LAG1 longevity assurance homolog 1|longevity assurance gene 1 protein homolog 1|longevity assurance homolog 1|protein UOG-1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70315775, end_position_on_the_genomic_accession=70331588, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  112240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112240, CurrentID=0, Status='live', Symbol='Bhr3', Aliases='', description='bronchial hyperresponsiveness 3', other_designations='', map_location='17 2.11 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  224997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=224997, CurrentID=0, Status='live', Symbol='Dlgap1', Aliases='4933422O14Rik, 9630002F18, AI845682, AI848168, BB075781, D17Bwg0511e, GKAP/SAPAP, GKPA/SAPAP, Gkap, Sapap1, mKIAA4162', description='discs, large (Drosophila) homolog-associated protein 1', other_designations='DAP-1 beta|PSD-95/SAP90-binding protein 1|SAP90/PSD-95 associated protein 1|SAP90/PSD-95-associated protein 1|guanylate kinase associated protein|guanylate kinase-associated protein', map_location='17 40.85 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=69969078, end_position_on_the_genomic_accession=70821413, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  225004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225004, CurrentID=0, Status='live', Symbol='BC027072', Aliases='', description='cDNA sequence BC027072', other_designations='', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71743555, end_position_on_the_genomic_accession=71752885, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  225010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225010, CurrentID=0, Status='live', Symbol='Lclat1', Aliases='AI181996, Agpat8, Alcat1, Gm91, Lycat', description='lysocardiolipin acyltransferase 1', other_designations='acyl-CoA:lysocardiolipin acyltransferase 1', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=73107985, end_position_on_the_genomic_accession=73243368, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  225020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225020, CurrentID=0, Status='live', Symbol='Fez2', Aliases='9030616F16, D17Ertd315e', description='fasciculation and elongation protein zeta 2 (zygin II)', other_designations='zygin 2|zygin II|zygin-2', map_location='17 48.57 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=78369212, end_position_on_the_genomic_accession=78418152, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  225027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225027, CurrentID=0, Status='live', Symbol='Srsf7', Aliases='35kDa, 9430065L19Rik, 9G8, NX-96, Sfrs7', description='serine/arginine-rich splicing factor 7', other_designations='splicing factor, arginine/serine-rich 7', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80200080, end_position_on_the_genomic_accession=80207318, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  225028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225028, CurrentID=0, Status='live', Symbol='Map4k3', Aliases='4833416M01Rik, 4833416M07Rik, 9530052P13Rik, Glk, MAPKKKK3, MEKKK 3, MEKKK3, RAB8IPL1', description='mitogen-activated protein kinase kinase kinase kinase 3', other_designations='MAPK/ERK kinase kinase kinase 3|MEK kinase kinase 3|germinal center kinase-related protein kinase|mitogen activated protein kinase kinase kinase kinase 3', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80580513, end_position_on_the_genomic_accession=80728205, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  225030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225030, CurrentID=0, Status='live', Symbol='Kcng3', Aliases='KV6.3, Kv10.1a, Kv10.1b', description='potassium voltage-gated channel, subfamily G, member 3', other_designations='voltage-gated potassium channel subunit Kv10.1|voltage-gated potassium channel subunit Kv6.3', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=83585411, end_position_on_the_genomic_accession=83631993, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  93960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93960, CurrentID=0, Status='live', Symbol='Nkd1', Aliases='2810434J10Rik, 9030215G15Rik, Nkd', description='naked cuticle 1 homolog (Drosophila)', other_designations='mNkd|mNkd1|naked-1', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88521344, end_position_on_the_genomic_accession=88594887, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  93961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93961, CurrentID=0, Status='live', Symbol='B3galt5', Aliases='1190002B21Rik, AU045265, b3Galt-V', description='UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 5', other_designations='SSEA-3 synthase|UDP-Gal:beta-GlcNAc beta-1,3-galactosyltransferase 5|UDP-galactose:beta-N-acetylglucosamine beta-1,3-galactosyltransferase 5|b3Gal-T5|beta-1,3-GalTase 5|beta-3-Gx-T5|beta3Gal-T5|beta3GalT5|stage-specific embryonic antigen 3 synthase', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96235801, end_position_on_the_genomic_accession=96319859, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  93966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93966, CurrentID=0, Status='live', Symbol='Hemgn', Aliases='4921524M03Rik, AI317176, EDAG, Hgn, mNDR', description='hemogen', other_designations='hemopoietic gene protein|negative differentiation regulator protein|nuclear protein', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46393989, end_position_on_the_genomic_accession=46413504, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  93968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=93968, CurrentID=0, Status='live', Symbol='Klra21', Aliases='Klra-ps1, Ly49H, Ly49u', description='killer cell lectin-like receptor subfamily A, member 21', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130082598, end_position_on_the_genomic_accession=130096486, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  225049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225049, CurrentID=0, Status='live', Symbol='Ttc7', Aliases='1110035E02Rik, 1700007L07Rika, fsn, hea, Ttc7', description='tetratricopeptide repeat domain 7', other_designations='TPR repeat protein 7A', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87282082, end_position_on_the_genomic_accession=87381770, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  68597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68597, CurrentID=0, Status='live', Symbol='Ccdc167', Aliases='1110021J02Rik', description='coiled-coil domain containing 167', other_designations='transmembrane and coiled-coil domain-containing protein C6orf129 homolog', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29695977, end_position_on_the_genomic_accession=29717017, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  225055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225055, CurrentID=0, Status='live', Symbol='Fbxo11', Aliases='C80048, Jf', description='F-box protein 11', other_designations='GENA 104|jeff', map_location='17 57.87 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=87990859, end_position_on_the_genomic_accession=88065285, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  225058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225058, CurrentID=0, Status='live', Symbol='Gm4832', Aliases='EG225058', description='predicted gene 4832', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=88066125, end_position_on_the_genomic_accession=88131186, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  68598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68598, CurrentID=0, Status='live', Symbol='Dnajc8', Aliases='1110021D09Rik, 2010009J04Rik, AL024084, AU019262, AU044514', description='DnaJ heat shock protein family (Hsp40) member C8', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 8', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132535548, end_position_on_the_genomic_accession=132553742, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  94040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94040, CurrentID=0, Status='live', Symbol='Clmn', Aliases='9330188N17Rik, AI428889, AI788815, mKIAA1188', description='calmin', other_designations='calponin-like transmembrane domain protein', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104763114, end_position_on_the_genomic_accession=104865076, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  94041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94041, CurrentID=0, Status='live', Symbol='Allc', Aliases='1700012B22Rik, Alc', description='allantoicase', other_designations='allantoate amidinohydrolase', map_location='12 10.97 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28553755, end_position_on_the_genomic_accession=28582552, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  225115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225115, CurrentID=0, Status='live', Symbol='Svil', Aliases='AU024053, B430302E16Rik', description='supervillin', other_designations='archvillin|p205/p250|supervillin muscle-specific isoform', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=4920236, end_position_on_the_genomic_accession=5119293, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  94044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94044, CurrentID=0, Status='live', Symbol='Bcl2l13', Aliases='BCL-RAMBO, E430016C20Rik, Mil-1, Mil1', description='BCL2-like 13 (apoptosis facilitator)', other_designations='bcl2-L-13|mitochondria located 1 homolog|protein Mil1', map_location='6 57.01 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120836221, end_position_on_the_genomic_accession=120892842, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  94045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94045, CurrentID=0, Status='live', Symbol='P2rx5', Aliases='P2X5', description='purinergic receptor P2X, ligand-gated ion channel, 5', other_designations='purinergic receptor P2X5-2', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73155277, end_position_on_the_genomic_accession=73175327, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  94047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94047, CurrentID=0, Status='live', Symbol='Cecr6', Aliases='AW048664', description='cat eye syndrome chromosome region, candidate 6', other_designations='cat eye syndrome chromosome region, candidate 6 homolog', map_location='6 56.95 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=120488939, end_position_on_the_genomic_accession=120493807, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  225131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225131, CurrentID=0, Status='live', Symbol='Wac', Aliases='1110067P07Rik, A230035H12Rik, AI256735, AI256776, Wwp4', description='WW domain containing adaptor with coiled-coil', other_designations='WW domain-containing protein 4', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=7868459, end_position_on_the_genomic_accession=7929028, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  94060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94060, CurrentID=0, Status='live', Symbol='Lce3c', Aliases='2300007B01Rik, AV076207, Eig3, Sprrl1', description='late cornified envelope 3C', other_designations='small proline rich-like 1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92944486, end_position_on_the_genomic_accession=92945730, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  94061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94061, CurrentID=0, Status='live', Symbol='Mrpl1', Aliases='2410002L03Rik, 5830418D04Rik, AI462572, L1mt', description='mitochondrial ribosomal protein L1', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=96209498, end_position_on_the_genomic_accession=96266727, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  94062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94062, CurrentID=0, Status='live', Symbol='Mrpl3', Aliases='2010320L16Rik, 5930422H18Rik, AA407889', description='mitochondrial ribosomal protein L3', other_designations='L3mt|MRP-L3', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105053268, end_position_on_the_genomic_accession=105077476, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  94063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94063, CurrentID=0, Status='live', Symbol='Mrpl16', Aliases='2310039D06Rik', description='mitochondrial ribosomal protein L16', other_designations='L16mt|MRP-L16|mitochondiral ribosomal protein L16', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11770415, end_position_on_the_genomic_accession=11774956, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  94064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94064, CurrentID=0, Status='live', Symbol='Mrpl27', Aliases='A630055F16Rik, D11Moh47, D18Ertd643e', description='mitochondrial ribosomal protein L27', other_designations='L27mt|MRP-L27', map_location='11 58.9 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94653770, end_position_on_the_genomic_accession=94660087, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  94065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94065, CurrentID=0, Status='live', Symbol='Mrpl34', Aliases='0610007O17Rik, D8Bwg1484e', description='mitochondrial ribosomal protein L34', other_designations='L34mt|MRP-L34', map_location='8 34.43 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71464926, end_position_on_the_genomic_accession=71465753, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  94066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94066, CurrentID=0, Status='live', Symbol='Mrpl36', Aliases='AI646041', description='mitochondrial ribosomal protein L36', other_designations='L36mt|MRP-L36', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73331009, end_position_on_the_genomic_accession=73332178, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  94067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94067, CurrentID=0, Status='live', Symbol='Mrpl43', Aliases='4930442D21Rik, bMRP36a', description='mitochondrial ribosomal protein L43', other_designations='L43mt|MRP-L43|mitochondrial ribosomal protein bMRP36a', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45005014, end_position_on_the_genomic_accession=45006442, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  94071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94071, CurrentID=0, Status='live', Symbol='Clec2h', Aliases='Clrf, clr-f', description='C-type lectin domain family 2, member h', other_designations='C-type lectin-related protein F', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128662385, end_position_on_the_genomic_accession=128677374, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  225152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225152, CurrentID=0, Status='live', Symbol='Gjd4', Aliases='9430022F06Rik, Cx39', description='gap junction protein, delta 4', other_designations='connexin-39|connexin39', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=9137290, end_position_on_the_genomic_accession=9289808, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  94088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94088, CurrentID=0, Status='live', Symbol='Trim6', Aliases='C430046K18Rik, D7Ertd684e', description='tripartite motif-containing 6', other_designations='tripartite motif protein 6', map_location='7 55.55 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104218783, end_position_on_the_genomic_accession=104235152, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  94089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94089, CurrentID=0, Status='live', Symbol='Trim7', Aliases='AI790312', description='tripartite motif-containing 7', other_designations='glycogenin-interacting protein|tripartite motif protein 7', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48826138, end_position_on_the_genomic_accession=48850644, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  94090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94090, CurrentID=0, Status='live', Symbol='Trim9', Aliases='', description='tripartite motif-containing 9', other_designations='tripartite motif-containing protein 9', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70244533, end_position_on_the_genomic_accession=70347761, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  94091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94091, CurrentID=0, Status='live', Symbol='Trim11', Aliases='', description='tripartite motif-containing 11', other_designations='tripartite motif-containing protein 11', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58978075, end_position_on_the_genomic_accession=58991462, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  94092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94092, CurrentID=0, Status='live', Symbol='Trim16', Aliases='9130006M08Rik, AI482483, EBBP', description='tripartite motif-containing 16', other_designations='estrogen-responsive B box protein|tripartite motif protein 16', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=62820195, end_position_on_the_genomic_accession=62842948, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  94093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94093, CurrentID=0, Status='live', Symbol='Trim33', Aliases='8030451N04Rik, AI413936, Ecto, Tif1g', description='tripartite motif-containing 33', other_designations='TIF1-gamma|ectodermin homolog|transcription intermediary factor 1-gamma|tripartite motif-containing protein 33', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103279293, end_position_on_the_genomic_accession=103358770, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  94094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94094, CurrentID=0, Status='live', Symbol='Trim34a', Aliases='Trim34, Trim34-1', description='tripartite motif-containing 34A', other_designations='tripartite motif protein 34|tripartite motif-containing 34|tripartite motif-containing protein 34', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104244437, end_position_on_the_genomic_accession=104262239, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  94109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94109, CurrentID=0, Status='live', Symbol='Csmd1', Aliases='B930082H09, mKIAA1890', description='CUB and Sushi multiple domains 1', other_designations='CUB and sushi multiple domains protein 1', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=15892542, end_position_on_the_genomic_accession=17535385, orientation='minus', exon_count=72, OMIM='', no_hdr0=''),
  225182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225182, CurrentID=0, Status='live', Symbol='Rbbp8', Aliases='9930104E21Rik, CtIP, RBBP-8, RIM, SAE2', description='retinoblastoma binding protein 8', other_designations='ctBP-interacting protein|retinoblastoma-binding protein 8|sporulation in the absence of SPO11 protein 2 homolog', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=11633276, end_position_on_the_genomic_accession=11743207, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  94111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94111, CurrentID=0, Status='live', Symbol='Mepe', Aliases='Of45', description='matrix extracellular phosphoglycoprotein with ASARM motif (bone)', other_designations='osteoregulin', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=104325329, end_position_on_the_genomic_accession=104338611, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  94112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94112, CurrentID=0, Status='live', Symbol='Med15', Aliases='A230074L19Rik, AW536074, Pcqap, mPcqap', description='mediator complex subunit 15', other_designations='PC2 glutamine/Q-rich-associated protein|positive cofactor 2 glutamine/Q-rich-associated protein|positive cofactor 2, multiprotein complex, glutamine/Q-rich-associated protein', map_location='16 10.94 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17651208, end_position_on_the_genomic_accession=17722947, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  225187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225187, CurrentID=0, Status='live', Symbol='Ankrd29', Aliases='G630054C21, G630054C21Rik, Gm95', description='ankyrin repeat domain 29', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=12252357, end_position_on_the_genomic_accession=12305720, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  225192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225192, CurrentID=0, Status='live', Symbol='Hrh4', Aliases='AXOR35, BG26, GPCR105, GPRv53, H4, H4R, HH4R', description='histamine receptor H4', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=13006982, end_position_on_the_genomic_accession=13023701, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  225207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225207, CurrentID=0, Status='live', Symbol='Zfp521', Aliases='B930086A16Rik, Evi3, Znf521', description='zinc finger protein 521', other_designations='ecotropic viral integration site 3 protein', map_location='18 7.68 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=13687013, end_position_on_the_genomic_accession=13972787, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  225215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225215, CurrentID=0, Status='live', Symbol='Rsl24d1', Aliases='2410159K22Rik', description='ribosomal L24 domain containing 1', other_designations='60S ribosomal protein L30 isolog|ribosomal L24 domain-containing protein 1', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=73113469, end_position_on_the_genomic_accession=73123333, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100233175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100233175, CurrentID=0, Status='live', Symbol='AK010878', Aliases='', description='cDNA sequence AK010878', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102753275, end_position_on_the_genomic_accession=102757810, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  94175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94175, CurrentID=0, Status='live', Symbol='Hrg', Aliases='AI265597, AW413091, D16jh2, D18020, Hprgp, Hrg', description='histidine-rich glycoprotein', other_designations='histidine-proline-rich glycoprotein', map_location='16 13.79 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22951072, end_position_on_the_genomic_accession=22961659, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  94176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94176, CurrentID=0, Status='live', Symbol='Dock2', Aliases='AI662014, AW122239, CED-5, Hch, MBC', description='dedicator of cyto-kinesis 2', other_designations='', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=34682290, end_position_on_the_genomic_accession=34783905, orientation='minus', exon_count=53, OMIM='', no_hdr0=''),
  94178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94178, CurrentID=0, Status='live', Symbol='Mcoln1', Aliases='2210015I05Rik, TRPML1, mucolipidin', description='mucolipin 1', other_designations='', map_location='8 1.92 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3500442, end_position_on_the_genomic_accession=3515232, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  94179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94179, CurrentID=0, Status='live', Symbol='Krt23', Aliases='CK23, Haik1, K23, Krt1-23', description='keratin 23', other_designations='CK-23|cytokeratin-23|keratin complex 1, acidic, gene 23|keratin-23|type I intermediate filament cytokeratin', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99477973, end_position_on_the_genomic_accession=99493110, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  94180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94180, CurrentID=0, Status='live', Symbol='Acsbg1', Aliases='BG1, Bgm, E230019G03Rik, Lpd, R75185, mKIAA0631', description='acyl-CoA synthetase bubblegum family member 1', other_designations='GR-LACS|bubblegum|gonadotropin-regulated long chain acyl CoA synthetase|lipidosin|lipidosis-related protein lipidosin|lipodisin|mBG1', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54604880, end_position_on_the_genomic_accession=54661885, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  94181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94181, CurrentID=0, Status='live', Symbol='Nans', Aliases='4632418E04Rik, BB056474, Sas', description='N-acetylneuraminic acid synthase (sialic acid synthase)', other_designations='N-acetylneuraminic acid 9-phosphate synthetase|N-acetylneuraminic acid phosphate synthase', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46489329, end_position_on_the_genomic_accession=46503438, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  94184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94184, CurrentID=0, Status='live', Symbol='Pdxdc1', Aliases='2210010A19Rik, AA415817, Kiaa0251-hp', description='pyridoxal-dependent decarboxylase domain containing 1', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13833161, end_position_on_the_genomic_accession=13903145, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  94185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94185, CurrentID=0, Status='live', Symbol='Tnfrsf21', Aliases='AA959878, DR6, R74815, TR7', description='tumor necrosis factor receptor superfamily, member 21', other_designations='Death receptor 6|TNFR-related death receptor 6|TNFR-related death receptor-6', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=43016555, end_position_on_the_genomic_accession=43089188, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  94186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94186, CurrentID=0, Status='live', Symbol='Strn3', Aliases='Gs2na, SG2NA', description='striatin, calmodulin binding protein 3', other_designations='SG2NA beta isoform|cell cycle autoantigen SG2NA|nuclear autoantigen|s/G2 antigen|striatin-3 35 kDa', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=51608541, end_position_on_the_genomic_accession=51692565, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  94187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94187, CurrentID=0, Status='live', Symbol='Zfp423', Aliases='Ebfaz, Roaz, Zfp104, Znf423, ataxia1, mKIAA0760, nur12', description='zinc finger protein 423', other_designations='early B-cell factor associated zinc finger protein|early B-cell factor associated zinc finger transcription factor|early B-cell factor-associated zinc finger protein|olf1/EBF-associated zinc finger protein|smad- and Olf-interacting zinc finger protein|zinc finger protein 104', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=87661810, end_position_on_the_genomic_accession=87961859, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  94190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94190, CurrentID=0, Status='live', Symbol='Ophn1', Aliases='C130037N19Rik, Wtgr', description='oligophrenin 1', other_designations='', map_location='X C2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=98554280, end_position_on_the_genomic_accession=98891414, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  94191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94191, CurrentID=0, Status='live', Symbol='Adarb2', Aliases='Adar3, RED2', description='adenosine deaminase, RNA-specific, B2', other_designations='RNA dependent adenosine deaminase 3|RNA-dependent adenosine deaminase 3|RNA-editing deaminase 2|RNA-editing enzyme 2|adenosine deaminase 3, RNA dependent|dsRNA adenosine deaminase B2', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=8202318, end_position_on_the_genomic_accession=8768762, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  94192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94192, CurrentID=0, Status='live', Symbol='C1galt1', Aliases='2210410E06Rik, AV284120, T-synthase', description='core 1 synthase, glycoprotein-N-acetylgalactosamine 3-beta-galactosyltransferase, 1', other_designations='T-syn|core 1 O-glycan T-synthase|core 1 UDP-galactose:N-acetylgalactosamine-alpha-R beta 1,3-galactosyltransferase 1|core 1 beta1,3-galactosyltransferase 1|core 1 beta3-Gal-T|core 1 beta3-Gal-T1|core1 UDP-galactose:N-acetylgalactosamine-alpha-R beta 1,3-galactosyltransferase 1', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=7845224, end_position_on_the_genomic_accession=7872042, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  225266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225266, CurrentID=0, Status='live', Symbol='Klhl14', Aliases='6330403N15Rik, AI785250, printor', description='kelch-like 14', other_designations='protein interactor of Torsin-1A|protein interactor of torsinA', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=21550377, end_position_on_the_genomic_accession=21672932, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100233208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100233208, CurrentID=0, Status='live', Symbol='Gm10778', Aliases='ENSMUSG00000074860', description='predicted gene 10778', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81649069, end_position_on_the_genomic_accession=81667361, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  225280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225280, CurrentID=0, Status='live', Symbol='Ino80c', Aliases='D030070L09Rik', description='INO80 complex subunit C', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24104761, end_position_on_the_genomic_accession=24121819, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  225283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225283, CurrentID=0, Status='live', Symbol='Rprd1a', Aliases='C77387, mKIAA4077', description='regulation of nuclear pre-mRNA domain containing 1A', other_designations='cyclin-dependent kinase inhibitor 2B-related protein|p15INK4B-related protein', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24484958, end_position_on_the_genomic_accession=24530222, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  94212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94212, CurrentID=0, Status='live', Symbol='Pag1', Aliases='Cbp, F730007C19Rik, Pag', description='phosphoprotein associated with glycosphingolipid microdomains 1', other_designations='Csk-binding protein|phosphoprotein transmembrane adaptor 1|phosphoprotein-associated with GEMs|transmembrane phosphoprotein Cbp', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=9687479, end_position_on_the_genomic_accession=9833679, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  94213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94213, CurrentID=0, Status='live', Symbol='Ddx50', Aliases='4933429B04Rik, 8430408E17Rik, GU2, RH-II', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 50', other_designations='DEAD box protein 50|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 50|Gubeta|RH-II/Gubeta|gu-beta|nucleolar protein GU2', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=62616015, end_position_on_the_genomic_accession=62651606, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  94214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94214, CurrentID=0, Status='live', Symbol='Spock2', Aliases='AA407235, Gcap26, mKIAA0275', description='sparc/osteonectin, cwcv and kazal-like domains proteoglycan 2', other_designations='granule cell antiserum positive 26', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60106257, end_position_on_the_genomic_accession=60135198, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  94215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94215, CurrentID=0, Status='live', Symbol='Ugt2a1', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide A1', other_designations='UDP glycosyltransferase 2 family, polypeptide A1|UDPGT 2A1|olfactory UDP glucuronosyltransferase', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87459490, end_position_on_the_genomic_accession=87490871, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  225288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225288, CurrentID=0, Status='live', Symbol='Fhod3', Aliases='A930009H06Rik, FHOS2, mKIAA1695', description='formin homology 2 domain containing 3', other_designations='formin homolog overexpressed in spleen 2|formin-family protein FHOS2', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24708456, end_position_on_the_genomic_accession=25133507, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  225289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225289, CurrentID=0, Status='live', Symbol='AW554918', Aliases='4921538P22, AI461742, Kiaa1328, mKIAA1328', description='expressed sequence AW554918', other_designations='', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=25169004, end_position_on_the_genomic_accession=25467321, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  94218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94218, CurrentID=0, Status='live', Symbol='Cnnm3', Aliases='Acdp3, Clp3, mACDP3', description='cyclin M3', other_designations='ancient conserved domain protein 3|ancient conserved domain-containing protein 3|cyclin-M3', map_location='1 15.3 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36511856, end_position_on_the_genomic_accession=36528237, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  94219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94219, CurrentID=0, Status='live', Symbol='Cnnm2', Aliases='AU015877, AW048635, Acdp2, Clp2', description='cyclin M2', other_designations='ancient conserved domain protein 2|ancient conserved domain-containing protein 2|cyclin-M2|cyclin-like protein 2|mACDP2', map_location='19 38.97 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46760381, end_position_on_the_genomic_accession=46878580, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  94220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94220, CurrentID=0, Status='live', Symbol='Cnnm4', Aliases='5430430O18Rik, Acdp4', description='cyclin M4', other_designations='ancient conserved domain protein 4|ancient conserved domain-containing protein 4|cyclin-M4|mACDP4', map_location='1 15.28 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36471597, end_position_on_the_genomic_accession=36508776, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  94221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94221, CurrentID=0, Status='live', Symbol='Gopc', Aliases='2210402P09Rik, AI844555, CAL, FIG, GOPC1, PIST', description='golgi associated PDZ and coiled-coil motif containing', other_designations='PDZ domain-containing protein PIST|PDZ protein interacting specifically with TC10', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=52337024, end_position_on_the_genomic_accession=52382124, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  94222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94222, CurrentID=0, Status='live', Symbol='Olig3', Aliases='Bhlhb7, bHLHe20', description='oligodendrocyte transcription factor 3', other_designations='basic helix-loop-helix domain-containing class B protein 7|class B basic helix-loop-helix protein 7|oligo3|oligodendrocyte-specific bHLH transcription factor 3', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19356539, end_position_on_the_genomic_accession=19358604, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  94223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94223, CurrentID=0, Status='live', Symbol='Dgcr8', Aliases='D16H22S1742E, D16H22S788E, D16Wis2, Gy1, N41, Vo59c07, mir-1306', description='DiGeorge syndrome critical region gene 8', other_designations='diGeorge syndrome critical region 8 homolog', map_location='16 11.31 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18253948, end_position_on_the_genomic_accession=18289252, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  94224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94224, CurrentID=0, Status='live', Symbol='Srd5a2', Aliases='5ART2, S5AR 2', description='steroid 5 alpha-reductase 2', other_designations='5 alpha-SR2|SR type 2|steroid 5 alpha reductase type2|steroid 5-alpha-reductase 2', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=74017706, end_position_on_the_genomic_accession=74047916, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  94226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94226, CurrentID=0, Status='live', Symbol='S1pr5', Aliases='Edg8, S1P5, lpB4', description='sphingosine-1-phosphate receptor 5', other_designations='S1P receptor 5|S1P receptor Edg-8|endothelial differentiation G-protein-coupled receptor 8|endothelial differentiation, sphingolipid G-protein-coupled receptor, 8|lysophospholipid receptor B4|sphingosine 1-phosphate receptor Edg-8|sphingosine-1-phosphate receptor LPB4', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21242917, end_position_on_the_genomic_accession=21248480, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  94227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94227, CurrentID=0, Status='live', Symbol='Pi15', Aliases='P24TI, P25TI, PI-15, SugarCrisp', description='peptidase inhibitor 15', other_designations='cysteine-rich secreted protein|protease inhibitor 15', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=17601901, end_position_on_the_genomic_accession=17630939, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  94229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94229, CurrentID=0, Status='live', Symbol='Slc4a10', Aliases='NCBE, mKIAA4136', description='solute carrier family 4, sodium bicarbonate cotransporter-like, member 10', other_designations="sodium bicarbonate cotransporter NBCn2-E|sodium bicarbonate cotransporter member 10 variant NBCn2-E'", map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=62046515, end_position_on_the_genomic_accession=62326743, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  94230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94230, CurrentID=0, Status='live', Symbol='Cpsf1', Aliases='', description='cleavage and polyadenylation specific factor 1', other_designations='CPSF 160 kDa subunit|CPSF160|cleavage and polyadenylation specificity factor 1|cleavage and polyadenylation specificity factor 160 kDa subunit', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76595808, end_position_on_the_genomic_accession=76607613, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  94232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94232, CurrentID=0, Status='live', Symbol='Ubqln4', Aliases='A1Up, A1u, AI663987, CIP75, UBIN', description='ubiquilin 4', other_designations='ataxin-1 interacting ubiquitin-like protein|ataxin-1 ubiquitin-like interacting protein|ataxin-1 ubiquitin-like-interacting protein A1U|connexin43-interacting protein of 75 kDa', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88553677, end_position_on_the_genomic_accession=88569725, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  94242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94242, CurrentID=0, Status='live', Symbol='Tinagl1', Aliases='1110021J17Rik, AZ-1, AZ1, Arg1, Lcn7, TARP, Tinagl', description='tubulointerstitial nephritis antigen-like 1', other_designations='adrenocortical zonation factor 1|androgen-regulated gene 1 protein|tubulointerstitial nephritis antigen-related protein', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130165600, end_position_on_the_genomic_accession=130175122, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  94244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94244, CurrentID=0, Status='live', Symbol='Fkbp6', Aliases='1700008G22Rik, 36kDa, AU017274, D5Ertd724e, FKBP-36, FKBP-6', description='FK506 binding protein 6', other_designations='36 kDa FK506-binding protein|36 kDa FKBP|FK506-binding protein 6 (36kD)|PPIase FKBP6|immunophilin FKBP36|inactive PPIase FKBP6|peptidyl-prolyl cis-trans isomerase FKBP6|rotamase', map_location='5 75.11 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135291704, end_position_on_the_genomic_accession=135351807, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  94245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94245, CurrentID=0, Status='live', Symbol='Dtnbp1', Aliases='5430437B18Rik, AW048963, Bloc1s8, dysbindin, sdy', description='dystrobrevin binding protein 1', other_designations='BLOC-1 subunit 8|HPS7 protein homolog|biogenesis of lysosome-related organelles complex 1 subunit 8|dysbindin-1|dystrobrevin-binding protein 1|hermansky-Pudlak syndrome 7 protein homolog|sandy', map_location='13 21.73 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=44922076, end_position_on_the_genomic_accession=45002096, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  94246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94246, CurrentID=0, Status='live', Symbol='Arid4b', Aliases='BCAA, BRCAA1, RBBP1L1, Rbp1l1, SAP180', description='AT rich interactive domain 4B (RBP1-like)', other_designations='180 kDa Sin3-associated polypeptide|ARID domain-containing protein 4B|breast cancer-associated antigen|histone deacetylase complex subunit SAP180|retinoblastoma binding protein 1-like 1|sin3-associated polypeptide p180', map_location='13 5.29 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=14063768, end_position_on_the_genomic_accession=14199603, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  94249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94249, CurrentID=0, Status='live', Symbol='Slc24a3', Aliases='NCKX3', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 3', other_designations='Na(+)/K(+)/Ca(2+)-exchange protein 3|solute carrier family 24 member 3', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=145167712, end_position_on_the_genomic_accession=145642166, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  94253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94253, CurrentID=0, Status='live', Symbol='Hecw1', Aliases='9330116H24Rik, AV273951, E130207I19Rik, Nedl1', description='HECT, C2 and WW domain containing E3 ubiquitin protein ligase 1', other_designations='HECT type E3 ubiquitin ligase|HECT, C2 and WW domain-containing protein 1|NEDD4-like E3 ubiquitin-protein ligase 1|mNEDL1', map_location='13 5.29 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=14226438, end_position_on_the_genomic_accession=14523280, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  94254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94254, CurrentID=0, Status='live', Symbol='Wbscr16', Aliases='5730496C04Rik, AU019812', description='Williams-Beuren syndrome chromosome region 16 homolog (human)', other_designations='RCC1-like G exchanging factor-like protein', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134148058, end_position_on_the_genomic_accession=134177301, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  225339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225339, CurrentID=0, Status='live', Symbol='Ammecr1l', Aliases='5430429D03Rik, AU022236, AU040755, AW111353, E230022H04Rik', description='AMME chromosomal region gene 1-like', other_designations='', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31759475, end_position_on_the_genomic_accession=31784083, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  225341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225341, CurrentID=0, Status='live', Symbol='Lims2', Aliases='PINCH2', description='LIM and senescent cell antigen like domains 2', other_designations='PINCH-2|particularly interesting new Cys-His protein 2', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=31922277, end_position_on_the_genomic_accession=31958620, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  94275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94275, CurrentID=0, Status='live', Symbol='Maged1', Aliases='2810433C11Rik, 5430405L04Rik, DXBwg1492e, Dlixin, Dlxin-1, Dlxin1, MAGE-D1, NRAGE, mFLJ00163', description='melanoma antigen, family D, 1', other_designations='MAGE-D1 antigen|neurotrophin receptor-interacting MAGE homolog', map_location='X 41.56 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94535474, end_position_on_the_genomic_accession=94542074, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  225348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225348, CurrentID=0, Status='live', Symbol='Wdr36', Aliases='5730444A13Rik, Ta-wdrp', description='WD repeat domain 36', other_designations='', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=32837225, end_position_on_the_genomic_accession=32866420, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  94279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94279, CurrentID=0, Status='live', Symbol='Sfxn2', Aliases='C79732, F630107H02Rik', description='sideroflexin 2', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46573322, end_position_on_the_genomic_accession=46596901, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  94280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94280, CurrentID=0, Status='live', Symbol='Sfxn3', Aliases='', description='sideroflexin 3', other_designations='', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45047508, end_position_on_the_genomic_accession=45056383, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  94281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94281, CurrentID=0, Status='live', Symbol='Sfxn4', Aliases='', description='sideroflexin 4', other_designations='', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60837270, end_position_on_the_genomic_accession=60861700, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  94282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94282, CurrentID=0, Status='live', Symbol='Sfxn5', Aliases='C230001H08Rik', description='sideroflexin 5', other_designations='', map_location='6 37.23 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85213049, end_position_on_the_genomic_accession=85333483, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  94284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94284, CurrentID=0, Status='live', Symbol='Ugt1a6a', Aliases='UDPGT, UDPGT 1-6, UGT1-06, UGT1.6, Ugt1a6, Ugt1a7', description='UDP glucuronosyltransferase 1 family, polypeptide A6A', other_designations='UDP glycosyltransferase 1 family, polypeptide A6|UDP-glucuronosyltransferase 1A6|phenol UDP-glucuronosyltransferase', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88134809, end_position_on_the_genomic_accession=88220002, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  225358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225358, CurrentID=0, Status='live', Symbol='Fam13b', Aliases='2610024E20Rik, AW060714, AW546153', description='family with sequence similarity 13, member B', other_designations='', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34440571, end_position_on_the_genomic_accession=34523046, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  225362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225362, CurrentID=0, Status='live', Symbol='Reep2', Aliases='', description='receptor accessory protein 2', other_designations='', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34840589, end_position_on_the_genomic_accession=34847463, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  225363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225363, CurrentID=0, Status='live', Symbol='Etf1', Aliases='AI463371, D6Ertd109e, ERF, ERF1, SUP45L1, TB3-1', description='eukaryotic translation termination factor 1', other_designations='eukaryotic release factor 1', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=34902785, end_position_on_the_genomic_accession=34932003, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  225372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225372, CurrentID=0, Status='live', Symbol='Apbb3', Aliases='Fe65l2, Rirl2, TR2S', description='amyloid beta (A4) precursor protein-binding, family B, member 3', other_designations='', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36671154, end_position_on_the_genomic_accession=36679366, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  94315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94315, CurrentID=0, Status='live', Symbol='Prcc', Aliases='', description='papillary renal cell carcinoma (translocation-associated)', other_designations='', map_location='3 F2.1-2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87858903, end_position_on_the_genomic_accession=87885562, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  225392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225392, CurrentID=0, Status='live', Symbol='Rell2', Aliases='4631403P03Rik, ependolin', description='RELT-like 2', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37955084, end_position_on_the_genomic_accession=37959179, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  94332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94332, CurrentID=0, Status='live', Symbol='Cadm3', Aliases='BIgR, Igsf4b, Necl-1, Necl1, SynCAM3, Tsll1', description='cell adhesion molecule 3', other_designations='TSLC1-like protein 1|immunoglobulin superfamily member 4B|immunoglobulin superfamily, member 4B|nectin-like protein 1|nectin-lke 1|synaptic cell adhesion molecule 3', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173334254, end_position_on_the_genomic_accession=173367695, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  94346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94346, CurrentID=0, Status='live', Symbol='Tmem40', Aliases='9030407H20Rik, AV091463, W64236', description='transmembrane protein 40', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115729131, end_position_on_the_genomic_accession=115762466, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  94352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94352, CurrentID=0, Status='live', Symbol='Loxl2', Aliases='1110004B06Rik, 4930526G11Rik, 9430067E15Rik', description='lysyl oxidase-like 2', other_designations='lysyl oxidase-like protein 2', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69608692, end_position_on_the_genomic_accession=69695834, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  94353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=94353, CurrentID=0, Status='live', Symbol='Hmgn3', Aliases='1110002A15Rik, 6330514M13Rik, BB071015, TRIP7', description='high mobility group nucleosomal binding domain 3', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=83109942, end_position_on_the_genomic_accession=83146734, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  225432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225432, CurrentID=0, Status='live', Symbol='Rbm27', Aliases='A730010I06, AI043120, Psc1, mKIAA1311', description='RNA binding motif protein 27', other_designations='RNA-binding motif protein 27|peri implantation stem cell 1|peri-implantation stem cell protein 1|periimplantation stem cell 1', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42275353, end_position_on_the_genomic_accession=42341542, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  225443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225443, CurrentID=0, Status='live', Symbol='Gm94', Aliases='', description='predicted gene 94', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=43777196, end_position_on_the_genomic_accession=43792878, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  225467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225467, CurrentID=0, Status='live', Symbol='Pggt1b', Aliases='2010207C17Rik, 2610100E13, AI451237, AI551093, BGG1, GGT1', description='protein geranylgeranyltransferase type I, beta subunit', other_designations='GGTase-I-beta|geranylgeranyl transferase type I subunit beta|type I protein geranyl-geranyltransferase subunit beta', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46235474, end_position_on_the_genomic_accession=46281089, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  225471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225471, CurrentID=0, Status='live', Symbol='Ticam2', Aliases='B430113A10, TICAM-2, TRAM, Tirp, Trif', description='toll-like receptor adaptor molecule 2', other_designations='TIR domain-containing adapter protein|TRIF-related adapter molecule TRAM|TRIF-related adaptor molecule|Toll-interleukin I receptor domain (TIR)-containing adaptor molecule (TICAM) 2|toll/interleukin-1 receptor domain-containing protein', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46558231, end_position_on_the_genomic_accession=46574533, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  225497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225497, CurrentID=0, Status='live', Symbol='Fam170a', Aliases='Gm93, Znfd', description='family with sequence similarity 170, member A', other_designations='family with sequence similarity 170, member B|zinc finger domain-containing protein|zinc finger protein ZNFD', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=50278369, end_position_on_the_genomic_accession=50283020, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  225518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225518, CurrentID=0, Status='live', Symbol='Prdm6', Aliases='Gm92, PRISM', description='PR domain containing 6', other_designations='PR domain zinc finger protein 6|PR domain-containing protein 6|PR domain-containing protein in smooth muscle|PR-domain zinc finger protein 6', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=53463583, end_position_on_the_genomic_accession=53575857, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  225523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225523, CurrentID=0, Status='live', Symbol='Cep120', Aliases='A230075C01, AU016693, Ccdc100', description='centrosomal protein 120', other_designations='coiled-coil domain containing 100|coiled-coil domain-containing protein 100', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=53681723, end_position_on_the_genomic_accession=53745603, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  225579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225579, CurrentID=0, Status='live', Symbol='Slc27a6', Aliases='4732438L20Rik, FACVL2, FATP6, VLCS-H1', description='solute carrier family 27 (fatty acid transporter), member 6', other_designations='', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=58556240, end_position_on_the_genomic_accession=58612869, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  225583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225583, CurrentID=0, Status='live', Symbol='A730017C20Rik', Aliases='', description='RIKEN cDNA A730017C20 gene', other_designations='', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=59062189, end_position_on_the_genomic_accession=59076962, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  225600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225600, CurrentID=0, Status='live', Symbol='Pde6a', Aliases='Pdea, nmf282', description='phosphodiesterase 6A, cGMP-specific, rod, alpha', other_designations='GMP-PDE alpha|cGMP phosphodiesterase alpha', map_location='18 34.41 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61220080, end_position_on_the_genomic_accession=61289750, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  225608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225608, CurrentID=0, Status='live', Symbol='Sh3tc2', Aliases='D430044G18Rik', description='SH3 domain and tetratricopeptide repeats 2', other_designations='SH3 domain and tetratricopeptide repeats-containing protein 2', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61952925, end_position_on_the_genomic_accession=62015719, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  243197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243197, CurrentID=0, Status='live', Symbol='Mfsd7a', Aliases='4732482E20Rik, Mfsd7', description='major facilitator superfamily domain containing 7A', other_designations='major facilitator superfamily domain containing 7', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108441052, end_position_on_the_genomic_accession=108450260, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  225631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225631, CurrentID=0, Status='live', Symbol='Onecut2', Aliases='C730009D12, OC-2, Oc2', description='one cut domain, family member 2', other_designations='one cut homeobox 2|transcription factor ONECUT-2|transcription factor ONECUT2', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=64340364, end_position_on_the_genomic_accession=64398488, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  103142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103142, CurrentID=0, Status='live', Symbol='Rdh9', Aliases='AI121090, Crad3', description='retinol dehydrogenase 9', other_designations='cis-retinol/androgen dehydrogenase type 3', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127776405, end_position_on_the_genomic_accession=127792697, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  225638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225638, CurrentID=0, Status='live', Symbol='Alpk2', Aliases='Gm549, Hak', description='alpha-kinase 2', other_designations='alpk2|heart alpha-kinase|heart alpha-protein kinase', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65265529, end_position_on_the_genomic_accession=65393888, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  225642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225642, CurrentID=0, Status='live', Symbol='Grp', Aliases='BLP', description='gastrin releasing peptide', other_designations='', map_location='18 38.98 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65872748, end_position_on_the_genomic_accession=65886599, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  225644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225644, CurrentID=0, Status='live', Symbol='Cplx4', Aliases='A930004D23Rik, CPXIV, Gm957', description='complexin 4', other_designations='CPX IV|complexin IV', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65955722, end_position_on_the_genomic_accession=65970178, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  225651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225651, CurrentID=0, Status='live', Symbol='Mppe1', Aliases='A530095G11', description='metallophosphoesterase 1', other_designations='metallo phosphoesterase|post-GPI attachment to proteins factor 5', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67225048, end_position_on_the_genomic_accession=67245830, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  225655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225655, CurrentID=0, Status='live', Symbol='Slmo1', Aliases='A230074A19Rik, BC019561, Prelid3a', description='slowmo homolog 1 (Drosophila)', other_designations='protein slowmo homolog 1', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67464835, end_position_on_the_genomic_accession=67480581, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  225659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225659, CurrentID=0, Status='live', Symbol='Cep76', Aliases='6230425F05Rik, 9630013E05', description='centrosomal protein 76', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67617397, end_position_on_the_genomic_accession=67641336, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  225724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225724, CurrentID=0, Status='live', Symbol='Mapk4', Aliases='A330097D03Rik, Erk3, Prkm4, p63Mapk', description='mitogen-activated protein kinase 4', other_designations='ERK-4|Erk3-related|MAP kinase 4|MAPK 4|extracellular signal-regulated kinase 4|mitogen activated protein kinase 4', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=73928486, end_position_on_the_genomic_accession=74065559, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  225742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225742, CurrentID=0, Status='live', Symbol='St8sia5', Aliases='ST8SiaV, Siat8e', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 5', other_designations='SIAT8-E|sialyltransferase 8 (alpha-2, 8-sialyltransferase) E|sialyltransferase 8E|sialyltransferase St8Sia V|sialytransferase St8Sia V', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77185847, end_position_on_the_genomic_accession=77255450, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  225743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225743, CurrentID=0, Status='live', Symbol='Rnf165', Aliases='2900024M11Rik, AI427432, Akd2, Ark2c, G630064H08Rik, Gm96', description='ring finger protein 165', other_designations='arkadia 2', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77456110, end_position_on_the_genomic_accession=77566378, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  225745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225745, CurrentID=0, Status='live', Symbol='Haus1', Aliases='BC024400, Ccdc5, HEI-C', description='HAUS augmin-like complex, subunit 1', other_designations='coiled-coil domain containing 5|coiled-coil domain-containing protein 5|spindle associated', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77757567, end_position_on_the_genomic_accession=77773847, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  225791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225791, CurrentID=0, Status='live', Symbol='Zadh2', Aliases='C530046K17Rik, Pthr3', description='zinc binding alcohol dehydrogenase, domain containing 2', other_designations='', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84087649, end_position_on_the_genomic_accession=84097528, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  225825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225825, CurrentID=0, Status='live', Symbol='Cd226', Aliases='BC051526, DNAM-1, DNAM1, Pta1, TLiSA1', description='CD226 antigen', other_designations='platelet and T cell activation antigen 1|platelet and T-cell activation antigen 1', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=89176954, end_position_on_the_genomic_accession=89272232, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  225845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225845, CurrentID=0, Status='live', Symbol='Pla2g16', Aliases='C78643, HRSL3, Hrasls3, Hrev107, MLP-3', description='phospholipase A2, group XVI', other_designations='H-rev 107 protein|HRAS like suppressor 3|LRAT-like protein 3|group XVI phospholipase A1/A2|group XVI phospholipase A2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7557434, end_position_on_the_genomic_accession=7588545, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  225849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225849, CurrentID=0, Status='live', Symbol='Ppp2r5b', Aliases="B'beta, BC026670", description="protein phosphatase 2, regulatory subunit B', beta", other_designations='protein phosphatase 2, regulatory subunit B (B56), beta isoform', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6227767, end_position_on_the_genomic_accession=6235840, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  225861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225861, CurrentID=0, Status='live', Symbol='Snx32', Aliases='B930037P14Rik, Snx6b', description='sorting nexin 32', other_designations='sortin nexin-6B|sorting nexin 6B|sorting nexin-6B', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5495274, end_position_on_the_genomic_accession=5510694, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  225865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225865, CurrentID=0, Status='live', Symbol='Catsper1', Aliases='Catsper, KSper', description='cation channel, sperm associated 1', other_designations='cation channel of sperm 1|sperm calcium channel 1|sperm ion channel|sperm protein CATSPER1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5335602, end_position_on_the_genomic_accession=5344281, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  225870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225870, CurrentID=0, Status='live', Symbol='Rin1', Aliases='', description='Ras and Rab interactor 1', other_designations='ras interaction/interference protein 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5050010, end_position_on_the_genomic_accession=5057072, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  225872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225872, CurrentID=0, Status='live', Symbol='Npas4', Aliases='LE-PAS, Nxf', description='neuronal PAS domain protein 4', other_designations='HLH-PAS transcription factor NXF|bHLH-PAS type transcription factor NXF|limbic-enhanced PAS protein|neuronal PAS4', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4984353, end_position_on_the_genomic_accession=5001925, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  225875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225875, CurrentID=0, Status='live', Symbol='Lrfn4', Aliases='BC023156, Salm3', description='leucine rich repeat and fibronectin type III domain containing 4', other_designations='synaptic adhesion-like molecule 3', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4611784, end_position_on_the_genomic_accession=4615667, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  225876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225876, CurrentID=0, Status='live', Symbol='Kdm2a', Aliases='100043628, 5530401A10Rik, AA589516, AW536790, Cxxc8, Fbl11, Fbl7, Fbxl11, Gm4560, Jhdm1, Jhdm1a, lalina', description='lysine (K)-specific demethylase 2A', other_designations='F-box and leucine-rich repeat protein 11|F-box protein containing leucine repeats|F-box/LRR-repeat protein 11|[Histone-H3]-lysine-36 demethylase 1A|jmjC domain-containing histone demethylation protein 1A', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4316144, end_position_on_the_genomic_accession=4398280, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  225887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225887, CurrentID=0, Status='live', Symbol='Ndufs8', Aliases='BC021616, CI-23kD, TYKY', description='NADH dehydrogenase (ubiquinone) Fe-S protein 8', other_designations='NADH-ubiquinone oxidoreductase 23 kDa subunit|complex I-23kD', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3908863, end_position_on_the_genomic_accession=3912774, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  225888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225888, CurrentID=0, Status='live', Symbol='Suv420h1', Aliases='AA117471, C630029K18Rik, KMT5B, Suv4-20h1', description='suppressor of variegation 4-20 homolog 1 (Drosophila)', other_designations='su(var)4-20 homolog 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3764155, end_position_on_the_genomic_accession=3818303, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  225895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225895, CurrentID=0, Status='live', Symbol='Taf6l', Aliases='2810417N14Rik, BB223262, C530024J06Rik, Paf65a', description='TATA-box binding protein associated factor 6 like', other_designations='PAF65-alpha|PCAF-associated factor 65 alpha|PCAF-associated factor 65-alpha|TAF6-like RNA polymerase II, p300/CBP-associated factor (PCAF)-associated factor', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8774354, end_position_on_the_genomic_accession=8786417, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  225896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225896, CurrentID=0, Status='live', Symbol='Ubxn1', Aliases='2B28, 4930455J02Rik, D19Ertd721e, T25529', description='UBX domain protein 1', other_designations='SAPK substrate protein 1|UBA/UBX 33.3 kDa protein|mY33K|protein 2B28', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8871510, end_position_on_the_genomic_accession=8875664, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  225898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225898, CurrentID=0, Status='live', Symbol='Eml3', Aliases='BC022146', description='echinoderm microtubule associated protein like 3', other_designations='EMAP-3', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8929694, end_position_on_the_genomic_accession=8941582, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  225908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225908, CurrentID=0, Status='live', Symbol='Myrf', Aliases='6030439E18, Gm1804, Gm98, Mrf', description='myelin regulatory factor', other_designations='myelin gene regulatory factor', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10208271, end_position_on_the_genomic_accession=10240781, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  225912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225912, CurrentID=0, Status='live', Symbol='Cyb561a3', Aliases='2310004G04Rik, Cybasc3, Lcytb', description='cytochrome b561 family, member A3', other_designations='lysosomal cytochrome b', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10577157, end_position_on_the_genomic_accession=10595287, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  225913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225913, CurrentID=0, Status='live', Symbol='Tkfc', Aliases='BC021917, Dak', description='triokinase, FMN cyclase', other_designations='bifunctional ATP-dependent dihydroxyacetone kinase/FAD-AMP lyase (cyclizing)|dihydroxyacetone kinase 2 homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10592197, end_position_on_the_genomic_accession=10605482, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  225922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225922, CurrentID=0, Status='live', Symbol='Oosp2', Aliases='Gm99, Plac1l, Tmem122', description='oocyte secreted protein 2', other_designations='placenta-specific 1-like protein|transmembrane protein 122', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11647279, end_position_on_the_genomic_accession=11661605, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  225923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225923, CurrentID=0, Status='live', Symbol='Oosp3', Aliases='Gm97', description='oocyte secreted protein 3', other_designations='oocyte-secreted protein 3', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11697055, end_position_on_the_genomic_accession=11711858, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  225929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225929, CurrentID=0, Status='live', Symbol='Patl1', Aliases='AV312086, Pat1b', description='protein associated with topoisomerase II homolog 1 (yeast)', other_designations='PAT1-like protein 1|protein PAT1 homolog b', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11912399, end_position_on_the_genomic_accession=11945096, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  619209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619209, CurrentID=0, Status='live', Symbol='Xce', Aliases='Xca, Xite', description='X-chromosome controlling element', other_designations='X-inactivation intergenic transcription element', map_location='X', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  225994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225994, CurrentID=0, Status='live', Symbol='Nmrk1', Aliases='AA408380, D630020N23Rik, Nrk1', description='nicotinamide riboside kinase 1', other_designations='NRK 1|RNK 1|nicotinic acid riboside kinase 1|nmR-K 1|ribosylnicotinamide kinase 1|ribosylnicotinic acid kinase 1', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=18632016, end_position_on_the_genomic_accession=18652191, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  225995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225995, CurrentID=0, Status='live', Symbol='D030056L22Rik', Aliases='', description='RIKEN cDNA D030056L22 gene', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=18712853, end_position_on_the_genomic_accession=18718433, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  225997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225997, CurrentID=0, Status='live', Symbol='Trpm6', Aliases='CHAK2, TRPM6/CHAK2', description='transient receptor potential cation channel, subfamily M, member 6', other_designations='channel kinase 2|melastatin-related TRP cation channel 6|transient receptor potential cation channel 6/channel-kinase 2', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=18730178, end_position_on_the_genomic_accession=18892516, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  225998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=225998, CurrentID=0, Status='live', Symbol='Rorb', Aliases='Nr1f2, RZR-beta, RZRBeta, Rorb', description='RAR-related orphan receptor beta', other_designations='nuclear receptor RZR-beta|nuclear receptor subfamily 1 group F member 2|retinoid-related orphan receptor-beta', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=18930609, end_position_on_the_genomic_accession=19111196, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  226016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226016, CurrentID=0, Status='live', Symbol='Abhd17b', Aliases='5730446C15Rik, Cgi67, Fam108b, Fam108b1', description='abhydrolase domain containing 17B', other_designations='Cgi67 serine protease|abhydrolase domain-containing protein FAM108B1|family with sequence similarity 108, member B', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=21653309, end_position_on_the_genomic_accession=21685638, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  226025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226025, CurrentID=0, Status='live', Symbol='Trpm3', Aliases='6330504P12Rik, 9330180E14, AU018608, B930001P07Rik, LTRPC3, MLSN2', description='transient receptor potential cation channel, subfamily M, member 3', other_designations='melastatin 2|transient receptor potential melastatin 3 beta 1|transient receptor potential melastatin 3 beta 10|transient receptor potential melastatin 3 beta 11|transient receptor potential melastatin 3 beta 12|transient receptor potential melastatin 3 beta 13|transient receptor potential melastatin 3 beta 14|transient receptor potential melastatin 3 beta 15|transient receptor potential melastatin 3 beta 16|transient receptor potential melastatin 3 beta 17|transient receptor potential melastatin 3 beta 2|transient receptor potential melastatin 3 beta 3|transient receptor potential melastatin 3 beta 4|transient receptor potential melastatin 3 beta 5|transient receptor potential melastatin 3 beta 6|transient receptor potential melastatin 3 beta 7|transient receptor potential melastatin 3 beta 8|transient receptor potential melastatin 3 beta 9', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=22138181, end_position_on_the_genomic_accession=22989897, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  226026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226026, CurrentID=0, Status='live', Symbol='Smc5', Aliases='MLZ-453, SMC-5l1, mKIAA0594, mSMC5, Smc5', description='structural maintenance of chromosomes 5', other_designations='SMC protein 5|SMC5 structural maintenance of chromosomes 5-like 1|protein expressed in male leptotene and zygotene spermatocytes 453', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=23206441, end_position_on_the_genomic_accession=23273908, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  226040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226040, CurrentID=0, Status='live', Symbol='Tmem252', Aliases='E030010A14Rik', description='transmembrane protein 252', other_designations='transmembrane protein C9orf71 homolog', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24674008, end_position_on_the_genomic_accession=24679661, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  226041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226041, CurrentID=0, Status='live', Symbol='Pgm5', Aliases='4833423B07, 9530034F03Rik, D830025G17, aciculin', description='phosphoglucomutase 5', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24678261, end_position_on_the_genomic_accession=24861842, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  226043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226043, CurrentID=0, Status='live', Symbol='Cbwd1', Aliases='AV349248, BC018472', description='COBW domain containing 1', other_designations='cobalamin synthase W domain-containing protein 1|cobalamin synthetase W domain-containing protein 1|dopamine responsive protein', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=24919916, end_position_on_the_genomic_accession=24961616, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  226049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226049, CurrentID=0, Status='live', Symbol='Dmrt2', Aliases='Terra', description='doublesex and mab-3 related transcription factor 2', other_designations='dmrt2|doublesex-like 2 protein|terra homolog', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=25672391, end_position_on_the_genomic_accession=25679013, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  619287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619287, CurrentID=0, Status='live', Symbol='Zcchc16', Aliases='C230031A03Rik, Mar4', description='zinc finger, CCHC domain containing 16', other_designations='gag-like protein|mammalian retrotransposon-derived protein 4|sushi-XF2b protein', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=144688218, end_position_on_the_genomic_accession=145122414, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  619288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619288, CurrentID=0, Status='live', Symbol='Fam71a', Aliases='4933417M04Rik, GARI-L4', description='family with sequence similarity 71, member A', other_designations='Golgi-associated Rab2B interactor-like 4', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191162584, end_position_on_the_genomic_accession=191164817, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619289, CurrentID=0, Status='live', Symbol='Rfx8', Aliases='4933400N17Rik', description='regulatory factor X 8', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39665299, end_position_on_the_genomic_accession=39720999, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  226075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226075, CurrentID=0, Status='live', Symbol='Glis3', Aliases='4833409N03Rik, E330013K21Rik', description='GLIS family zinc finger 3', other_designations='GLI-similar 3', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=28262179, end_position_on_the_genomic_accession=28680077, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  619294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619294, CurrentID=0, Status='live', Symbol='Cldn34b4', Aliases='4930428D18Rik', description='claudin 34B4', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=76385323, end_position_on_the_genomic_accession=76397979, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  619297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619297, CurrentID=0, Status='live', Symbol='C430049E01Rik', Aliases='', description='RIKEN cDNA C430049E01 gene', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71520118, end_position_on_the_genomic_accession=71528186, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  619301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619301, CurrentID=0, Status='live', Symbol='Tmem253', Aliases='G630016D24Rik', description='transmembrane protein 253', other_designations='tranmembrane protein 253|transmembrane protein C14orf176 homolog', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52015275, end_position_on_the_genomic_accession=52019787, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  619303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619303, CurrentID=0, Status='live', Symbol='4930567H17Rik', Aliases='', description='RIKEN cDNA 4930567H17 gene', other_designations='', map_location='X A7.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=70393901, end_position_on_the_genomic_accession=70394740, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  226089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226089, CurrentID=0, Status='live', Symbol='Ric1', Aliases='C030046E11Rik, C130057E09Rik, Kiaa1432', description='RAB6A GEF complex partner 1', other_designations='protein RIC1 homolog', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29522282, end_position_on_the_genomic_accession=29605921, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  226090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226090, CurrentID=0, Status='live', Symbol='Ermp1', Aliases='D19Ertd410e, D19Wsu12e, Fxna, mKIAA1815', description='endoplasmic reticulum metallopeptidase 1', other_designations='aminopeptidase Fxna|felix-ina', map_location='19 24.17 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29609883, end_position_on_the_genomic_accession=29648484, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  619309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619309, CurrentID=0, Status='live', Symbol='Muc3a', Aliases='A630081J09Rik', description='mucin 3A, cell surface associated', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137208805, end_position_on_the_genomic_accession=137212542, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  226098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226098, CurrentID=0, Status='live', Symbol='Hectd2', Aliases='4921524L07, A630025O09Rik, AW212605', description='HECT domain containing 2', other_designations='HECT domain-containing protein 2', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36553969, end_position_on_the_genomic_accession=36621135, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  226101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226101, CurrentID=0, Status='live', Symbol='Myof', Aliases='2310004N10Rik, 2310051D19Rik, E030042N20Rik, Fer1l3', description='myoferlin', other_designations='fer-1-like protein 3', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37899036, end_position_on_the_genomic_accession=38043577, orientation='minus', exon_count=55, OMIM='', no_hdr0=''),
  226105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226105, CurrentID=0, Status='live', Symbol='Cyp2c70', Aliases='', description='cytochrome P450, family 2, subfamily c, polypeptide 70', other_designations='CYPIIC70|cyp2c70', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40153361, end_position_on_the_genomic_accession=40187286, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  619326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619326, CurrentID=0, Status='live', Symbol='9130409I23Rik', Aliases='', description='RIKEN cDNA 9130409I23 gene', other_designations='', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181006612, end_position_on_the_genomic_accession=181060761, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  619331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619331, CurrentID=0, Status='live', Symbol='Zfp551', Aliases='9630004E07Rik, Znf551', description='zinc fingr protein 551', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12412548, end_position_on_the_genomic_accession=12422739, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  226122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226122, CurrentID=0, Status='live', Symbol='Ubtd1', Aliases='BC016129', description='ubiquitin domain containing 1', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41981763, end_position_on_the_genomic_accession=42034650, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  226123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226123, CurrentID=0, Status='live', Symbol='Morn4', Aliases='', description='MORN repeat containing 4', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42074939, end_position_on_the_genomic_accession=42086370, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  226139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226139, CurrentID=0, Status='live', Symbol='Cox15', Aliases='2900026G05Rik', description='cytochrome c oxidase assembly protein 15', other_designations='COX15 homolog, cytochrome c oxidase assembly protein', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43733254, end_position_on_the_genomic_accession=43753000, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  226143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226143, CurrentID=0, Status='live', Symbol='Cyp2c44', Aliases='AW107714, BC034834, b2b386Clo', description='cytochrome P450, family 2, subfamily c, polypeptide 44', other_designations='Mutant line 386|cytochrome P450, 2c44', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44005022, end_position_on_the_genomic_accession=44029247, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  226144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226144, CurrentID=0, Status='live', Symbol='Erlin1', Aliases='2810439N09Rik, C80197, Keo4, Spfh1', description='ER lipid raft associated 1', other_designations='SPFH domain family, member 1|SPFH domain-containing protein 1|endoplasmic reticulum lipid raft-associated protein 1|protein KE04 homolog|stomatin-prohibitin-flotillin-HflC/K domain-containing protein 1', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44034943, end_position_on_the_genomic_accession=44070493, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  226151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226151, CurrentID=0, Status='live', Symbol='Fam178a', Aliases='3632432H19, 6030443O07Rik, AU014917, Slf2', description='family with sequence similarity 178, member A', other_designations='protein FAM178A', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44918505, end_position_on_the_genomic_accession=44983787, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  226153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226153, CurrentID=0, Status='live', Symbol='Peo1', Aliases='D19Ertd626e, PEO, Twinl', description='progressive external ophthalmoplegia 1 (human)', other_designations='T7 gp4-like protein with intramitochondrial nucleoid localization|T7-like mitochondrial DNA helicase|progressive external ophthalmoplegia 1 protein homolog|putative T7-like mitochodrial DNA helicase twinky', map_location='19 38.19 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45006079, end_position_on_the_genomic_accession=45012762, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  226154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226154, CurrentID=0, Status='live', Symbol='Lzts2', Aliases='BC014695, LAPSER1, mKIAA1813', description='leucine zipper, putative tumor suppressor 2', other_designations='protein LAPSER1|putative tumor suppressor LAPSER1', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45015176, end_position_on_the_genomic_accession=45027115, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  226162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226162, CurrentID=0, Status='live', Symbol='Dpcd', Aliases='5330431N19Rik, Ndac', description='deleted in primary ciliary dyskinesia', other_designations='neighbor of Dac', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45560590, end_position_on_the_genomic_accession=45578287, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  226169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226169, CurrentID=0, Status='live', Symbol='Pprc1', Aliases='A930018M19, BC013720, PRC, mKIAA0595', description='peroxisome proliferative activated receptor, gamma, coactivator-related 1', other_designations='PGC-1-related coactivator', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46056496, end_position_on_the_genomic_accession=46072909, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  226178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226178, CurrentID=0, Status='live', Symbol='Wbp1l', Aliases='D19Wsu162e, Opal1', description='WW domain binding protein 1 like', other_designations='outcome predictor in acute leukemia 1 homolog', map_location='19 38.97 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46599106, end_position_on_the_genomic_accession=46657389, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  226180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226180, CurrentID=0, Status='live', Symbol='Ina', Aliases='AV028420, NF-66, NF66', description='internexin neuronal intermediate filament protein, alpha', other_designations='66 kDa neurofilament protein|alpha-Inx|neurofilament-66', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47014698, end_position_on_the_genomic_accession=47024358, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  226182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226182, CurrentID=0, Status='live', Symbol='Taf5', Aliases='6330528C20Rik, AV117817', description='TATA-box binding protein associated factor 5', other_designations='TAF(II)100|TAF5 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF5 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 100kDa|TAFII-100|TAFII100|Tbp-associated factor 5|transcription initiation factor TFIID 100 kDa subunit', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47067748, end_position_on_the_genomic_accession=47083479, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  619441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619441, CurrentID=0, Status='live', Symbol='Tnfsfm13', Aliases='Tnfsf12-Tnfsf13, Tnfsf12Tnfsf13, Tnfsf12tnfsf13, Twe-pril', description='tumor necrosis factor (ligand) superfamily, membrane-bound member 13', other_designations='Tnfsf12-Tnfsf13 readthrough transcript', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69682577, end_position_on_the_genomic_accession=69696098, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  226243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226243, CurrentID=0, Status='live', Symbol='Habp2', Aliases='AI035669, FSAP, HABP, HGFAL, PHBP', description='hyaluronic acid binding protein 2', other_designations='hyaduronic acid-binding protein 2|plasma hyaluronan-binding protein', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56287113, end_position_on_the_genomic_accession=56320089, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  226245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226245, CurrentID=0, Status='live', Symbol='Plekhs1', Aliases='9930023K05Rik', description='pleckstrin homology domain containing, family S member 1', other_designations='PH domain-containing family S member 1|PH domain-containing protein C10orf81 homolog', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56457276, end_position_on_the_genomic_accession=56486729, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  226250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226250, CurrentID=0, Status='live', Symbol='Afap1l2', Aliases='AU041783, C86904, mKIAA1914', description='actin filament associated protein 1-like 2', other_designations='AFAP1-like protein 2', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56912354, end_position_on_the_genomic_accession=57008590, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  226251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226251, CurrentID=0, Status='live', Symbol='Ablim1', Aliases='2210411C18Rik, 2610209L21Rik, 4833406P10Rik, 9330196J19Rik, AV079770, AW060987, AW215784, Limab1, mKIAA0059', description='actin-binding LIM protein 1', other_designations='actin-binding LIM protein family member 1|actin-binding double zinc finger protein', map_location='19 52.09 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=57033264, end_position_on_the_genomic_accession=57239560, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  226252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226252, CurrentID=0, Status='live', Symbol='Fam160b1', Aliases='AI450540, mKIAA1600', description='family with sequence similarity 160, member B1', other_designations='', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=57360971, end_position_on_the_genomic_accession=57389594, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  226255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226255, CurrentID=0, Status='live', Symbol='Atrnl1', Aliases='AI504415, AW555641, Alp, Atrnl, Atrnlp', description='attractin like 1', other_designations='', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=57610963, end_position_on_the_genomic_accession=58133343, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  226265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226265, CurrentID=0, Status='live', Symbol='Eno4', Aliases='6430537H07Rik', description='enolase 4', other_designations='enolase-like protein C10orf134 homolog', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=58939922, end_position_on_the_genomic_accession=58971421, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  226278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226278, CurrentID=0, Status='live', Symbol='Prlhr', Aliases='Gm339, Gpr10, Gr3, Prrpr', description='prolactin releasing hormone receptor', other_designations='G-protein coupled receptor 10|PrRP receptor', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60465641, end_position_on_the_genomic_accession=60469431, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  619517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619517, CurrentID=0, Status='live', Symbol='Esp1', Aliases='EG619517, Gm6084', description='exocrine gland secreted peptide 1', other_designations='esp1|exocrine gland-secreting peptide 1', map_location='17 B2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40727120, end_position_on_the_genomic_accession=40731782, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  226304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226304, CurrentID=0, Status='live', Symbol='Npbwr1', Aliases='Gpr7, Nbpwr1', description='neuropeptides B/W receptor 1', other_designations='G protein-coupled receptor 7|G-protein coupled receptor 7', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=5913707, end_position_on_the_genomic_accession=5917398, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  619548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619548, CurrentID=0, Status='live', Symbol='Defb42', Aliases='BD-42, Defb44, mBD-42', description='defensin beta 42', other_designations='beta-defensin 136|beta-defensin 44|defensin, beta 42', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63046991, end_position_on_the_genomic_accession=63048606, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  226351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226351, CurrentID=0, Status='live', Symbol='Tmem185b', Aliases='2500001K11Rik', description='transmembrane protein 185B', other_designations='ee3_2|erythropoietin-induced EST 3-2', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=119526154, end_position_on_the_genomic_accession=119528983, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  226352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226352, CurrentID=0, Status='live', Symbol='Epb41l5', Aliases='1700030C16Rik, AL022914, BE37, E230025E14Rik, Epb4.1l5, Lulu1, NBL5', description='erythrocyte membrane protein band 4.1 like 5', other_designations='erythrocyte protein band 4.1 like 5|erythrocyte protein band 4.1-like 5', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=119545032, end_position_on_the_genomic_accession=119649136, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  226356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226356, CurrentID=0, Status='live', Symbol='Cfap221', Aliases='Gm101, Pcdp1', description='cilia and flagella associated protein 221', other_designations='primary ciliary dyskinesia protein 1', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=119923340, end_position_on_the_genomic_accession=119997471, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  226359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226359, CurrentID=0, Status='live', Symbol='C1ql2', Aliases='BC040774, CTRP10', description='complement component 1, q subcomponent-like 2', other_designations='C1q and tumor necrosis factor-related protein 10|C1q/TNF-related protein 10|C1qTNF10', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=120340582, end_position_on_the_genomic_accession=120343174, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  619597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619597, CurrentID=0, Status='live', Symbol='Gm6086', Aliases='EG619597, Gm219', description='predicted gene 6086', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93990509, end_position_on_the_genomic_accession=94011539, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  619605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619605, CurrentID=0, Status='live', Symbol='Zcchc17', Aliases='2810055E05Rik, HSPC251, LDC4, pNO40, pS1D', description='zinc finger, CCHC domain containing 17', other_designations='putative S1 RNA binding domain protein|putative S1 RNA-binding domain protein|zinc finger CCHC domain-containing protein 17', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130315902, end_position_on_the_genomic_accession=130359963, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  226407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226407, CurrentID=0, Status='live', Symbol='Rab3gap1', Aliases='1700003B17Rik, 4732493F09Rik, AL117896, Rab3gap, p130', description='RAB3 GTPase activating protein subunit 1', other_designations='RAB3 GTPase-activating protein 130 kDa subunit|rab3-GAP p130', map_location='1 E3-E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=127868737, end_position_on_the_genomic_accession=127943876, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  226409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226409, CurrentID=0, Status='live', Symbol='Zranb3', Aliases='4933425L19Rik, AH2, AI316834, C730006D09', description='zinc finger, RAN-binding domain containing 3', other_designations='annealing helicase 2|zinc finger Ran-binding domain-containing protein 3', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=127954179, end_position_on_the_genomic_accession=128103049, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  226412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226412, CurrentID=0, Status='live', Symbol='R3hdm1', Aliases='R3hdm', description='R3H domain containing 1', other_designations='R3H domain (binds single-stranded nucleic acids)|R3H domain 1 (binds single-stranded nucleic acids)', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=128103293, end_position_on_the_genomic_accession=128237738, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  226413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226413, CurrentID=0, Status='live', Symbol='Lct', Aliases='Gm100, LAC, LPH, Lphl', description='lactase', other_designations='', map_location='1 55.8 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=128284764, end_position_on_the_genomic_accession=128328318, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  226414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226414, CurrentID=0, Status='live', Symbol='Dars', Aliases='5730439G15Rik', description='aspartyl-tRNA synthetase', other_designations='aspRS|aspartyl-tRNA synthetase, cytoplasmic', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=128363707, end_position_on_the_genomic_accession=128417416, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  226418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226418, CurrentID=0, Status='live', Symbol='Yod1', Aliases='9930028C20Rik, Hshin7', description='YOD1 OTU deubiquitinating enzyme 1 homologue (S. cerevisiae)', other_designations='HIV-induced protein-7-like protease', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=130717327, end_position_on_the_genomic_accession=130722073, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  226419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226419, CurrentID=0, Status='live', Symbol='Dyrk3', Aliases='BC006704', description='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 3', other_designations='dyrk3', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131128441, end_position_on_the_genomic_accession=131138234, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  226421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226421, CurrentID=0, Status='live', Symbol='Rab7b', Aliases='5430435G22Rik', description='RAB7B, member RAS oncogene family', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131688680, end_position_on_the_genomic_accession=131715441, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  226422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226422, CurrentID=0, Status='live', Symbol='Rab29', Aliases='Rab7l1', description='RAB29, member RAS oncogene family', other_designations='RAB7, member RAS oncogene family-like 1|rab-7-like protein 1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131867210, end_position_on_the_genomic_accession=131872887, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  404711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=404711, CurrentID=0, Status='live', Symbol='Ighg2c', Aliases='Igh-1b', description='immunoglobulin heavy constant gamma 2C', other_designations='immunoglobulin heavy chain 1b (serum IgG2c)', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113287389, end_position_on_the_genomic_accession=113288932, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  226432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226432, CurrentID=0, Status='live', Symbol='Ipo9', Aliases='0710008K06Rik, AI845704, C78347, Imp9, Ranbp9, mKIAA1192', description='importin 9', other_designations='Ran binding protein 9|imp9a|imp9b|importin-9a|importin-9b|ran-binding protein 9', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135382315, end_position_on_the_genomic_accession=135430502, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  226438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226438, CurrentID=0, Status='live', Symbol='Igfn1', Aliases='9830123M21Rik', description='immunoglobulin-like and fibronectin type III domain containing 1', other_designations='eEF1A2 binding protein|immunoglobulin-like and fibronectin type 3 domain containing 1 protein', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135953205, end_position_on_the_genomic_accession=136040842, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  226439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226439, CurrentID=0, Status='live', Symbol='Ascl5', Aliases='Ascl5-ps, Gm197, bHLHa47', description='achaete-scute family bHLH transcription factor 5', other_designations='achaete-scute complex homolog 5, pseudogene', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136050804, end_position_on_the_genomic_accession=136051370, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  226442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226442, CurrentID=0, Status='live', Symbol='Zfp281', Aliases='Znf281', description='zinc finger protein 281', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136624901, end_position_on_the_genomic_accession=136630391, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  619665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619665, CurrentID=0, Status='live', Symbol='Klf14', Aliases='5330411L03Rik, AL022736, BTEB5', description='Kruppel-like factor 14', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30956021, end_position_on_the_genomic_accession=30958990, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  226470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226470, CurrentID=0, Status='live', Symbol='Zbtb41', Aliases='8430415N23Rik, 9430031N01, 9830132G07Rik, AI316857', description='zinc finger and BTB domain containing 41', other_designations='zinc finger and BTB domain containing 41 homolog', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=139421719, end_position_on_the_genomic_accession=139453007, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  619697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619697, CurrentID=0, Status='live', Symbol='Vmn2r116', Aliases='EG619697, V2Rp5', description='vomeronasal 2, receptor 116', other_designations='vmn2r116|vomeronasal 2 receptor p5|vomeronasal receptor Vmn2r116|vomeronasal type-2 receptor P5', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23384803, end_position_on_the_genomic_accession=23401864, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  103284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103284, CurrentID=0, Status='live', Symbol='Zc3h10', Aliases='AI790326', description='zinc finger CCCH type containing 10', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128543563, end_position_on_the_genomic_accession=128547744, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  226499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226499, CurrentID=0, Status='live', Symbol='BC003331', Aliases='1810053E15Rik, Odr4, odr-4', description='cDNA sequence BC003331', other_designations='mODR-4|olfactory specific odr-4', map_location='1 G1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=150361311, end_position_on_the_genomic_accession=150393055, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  226517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226517, CurrentID=0, Status='live', Symbol='Smg7', Aliases='9430023P16Rik, mKIAA0250', description='Smg-7 homolog, nonsense mediated mRNA decay factor (C. elegans)', other_designations='EST1-like protein C', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152836995, end_position_on_the_genomic_accession=152902686, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  226518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226518, CurrentID=0, Status='live', Symbol='Nmnat2', Aliases='AI843915, D030041I09Rik, PNAT1, PNAT2', description='nicotinamide nucleotide adenylyltransferase 2', other_designations='NMN adenylyltransferase 2|NMN/NaMN adenylyltransferase 2|naMN adenylyltransferase 1|naMN adenylyltransferase 2|nicotinamide mononucleotide adenylyltransferase 2|nicotinate-nucleotide adenylyltransferase 1|nicotinate-nucleotide adenylyltransferase 2|pyridine nucleotide adenylyltransferase 2', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=152954940, end_position_on_the_genomic_accession=153119266, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  226519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226519, CurrentID=0, Status='live', Symbol='Lamc1', Aliases='Lamb2', description='laminin, gamma 1', other_designations='S-LAM gamma|S-laminin subunit gamma|laminin B2 chain|laminin-1 subunit gamma|laminin-10 subunit gamma|laminin-11 subunit gamma|laminin-2 subunit gamma|laminin-3 subunit gamma|laminin-4 subunit gamma|laminin-6 subunit gamma|laminin-7 subunit gamma|laminin-8 subunit gamma|laminin-9 subunit gamma', map_location='1 65.3 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153218922, end_position_on_the_genomic_accession=153332786, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  226525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226525, CurrentID=0, Status='live', Symbol='Rasal2', Aliases='6530401P13, A330066M24Rik', description='RAS protein activator like 2', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=157135183, end_position_on_the_genomic_accession=157412595, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  226527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226527, CurrentID=0, Status='live', Symbol='BC026585', Aliases='', description='cDNA sequence BC026585', other_designations='', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=157458422, end_position_on_the_genomic_accession=157490000, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  226539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226539, CurrentID=0, Status='live', Symbol='Dars2', Aliases='5830468K18Rik, aspRS', description='aspartyl-tRNA synthetase 2 (mitochondrial)', other_designations='', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161040612, end_position_on_the_genomic_accession=161070658, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  226541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226541, CurrentID=0, Status='live', Symbol='Klhl20', Aliases='AI504637, D930050H05Rik, Kleip, mKIAA4210', description='kelch-like 20', other_designations='Kelch-like ECT2 interacting protein|kelch-like 18|kelch-like ECT2-interacting protein', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161088375, end_position_on_the_genomic_accession=161131508, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  226548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226548, CurrentID=0, Status='live', Symbol='Aph1a', Aliases='6530402N02Rik, APH-1a', description='aph1 homolog A, gamma secretase subunit', other_designations='anterior pharynx defective 1a homolog|aph-1alpha', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95893907, end_position_on_the_genomic_accession=95898592, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  226551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226551, CurrentID=0, Status='live', Symbol='Suco', Aliases='4732491B11, AI848100, AU016931, Opt', description='SUN domain containing ossification factor', other_designations='SUN-like protein 1|membrane protein CH1|osteopotentia|protein C1orf9 homolog|protein osteopotentia', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161816112, end_position_on_the_genomic_accession=161876661, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  226562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226562, CurrentID=0, Status='live', Symbol='Prrc2c', Aliases='1810043M20Rik, 9630039I18Rik, A630006J20, Bat2d, Bat2d1, Bat2l2, E130112L15Rik, Prrc3, mKIAA1096', description='proline-rich coiled-coil 2C', other_designations='BAT2 domain containing 1|BAT2 domain-containing protein 1|HLA-B associated transcript 2-like 2|HLA-B-associated transcript 2-like 2|proline-rich and coiled-coil-containing protein 2C|protein BAT2-like 2', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162670727, end_position_on_the_genomic_accession=162740564, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  226564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226564, CurrentID=0, Status='live', Symbol='Fmo4', Aliases='D1Ertd532e', description='flavin containing monooxygenase 4', other_designations='FMO 4|dimethylaniline oxidase 4|flavin-containing monooxygenase 4|hepatic flavin-containing monooxygenase 4', map_location='1 70.34 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162793188, end_position_on_the_genomic_accession=162815869, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  226565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226565, CurrentID=0, Status='live', Symbol='Fmo6', Aliases='', description='flavin containing monooxygenase 6', other_designations='flavin-containing monooxygenase 6', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162916551, end_position_on_the_genomic_accession=162937572, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  619788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=619788, CurrentID=0, Status='live', Symbol='Vmn2r117', Aliases='EG619788, V2Rp6', description='vomeronasal 2, receptor 117', other_designations='vomeronasal 2 receptor p6', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23459675, end_position_on_the_genomic_accession=23479597, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  226591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226591, CurrentID=0, Status='live', Symbol='Tiprl', Aliases='1810011K17Rik', description='TIP41, TOR signalling pathway regulator-like (S. cerevisiae)', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165212286, end_position_on_the_genomic_accession=165282478, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  226594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226594, CurrentID=0, Status='live', Symbol='Rcsd1', Aliases='A430105K13Rik, BC025872', description='RCSD domain containing 1', other_designations='protein kinase substrate CapZIP', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165648945, end_position_on_the_genomic_accession=165710063, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  226601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226601, CurrentID=0, Status='live', Symbol='Gm4846', Aliases='EG226601, Fmo13', description='predicted gene 4846', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166482599, end_position_on_the_genomic_accession=166506135, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  226604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226604, CurrentID=0, Status='live', Symbol='Gm4847', Aliases='EG226604, Fmo12', description='predicted gene 4847', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166628971, end_position_on_the_genomic_accession=166647693, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  226610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226610, CurrentID=0, Status='live', Symbol='Fam78b', Aliases='C030014K22Rik, C030020L09Rik', description='family with sequence similarity 78, member B', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167001417, end_position_on_the_genomic_accession=167091302, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  226641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226641, CurrentID=0, Status='live', Symbol='Atf6', Aliases='9130025P16Rik, 9630036G24, AA789574alpha, ESTM49, Atf6', description='activating transcription factor 6', other_designations='ATF6-alpha|activating transcription factor 6 alpha|cAMP-dependent transcription factor ATF-6 alpha', map_location='1 76.96 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170704457, end_position_on_the_genomic_accession=170867771, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  226646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226646, CurrentID=0, Status='live', Symbol='Ndufs2', Aliases='CI-49kD', description='NADH dehydrogenase (ubiquinone) Fe-S protein 2', other_designations='NADH-coenzyme Q reductase|NADH-ubiquinone oxidoreductase 49 kDa subunit|complex I-49kD', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171234860, end_position_on_the_genomic_accession=171247112, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  226652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226652, CurrentID=0, Status='live', Symbol='Arhgap30', Aliases='6030405P05Rik, Gm102, mFLJ00267', description='Rho GTPase activating protein 30', other_designations='rho-type GTPase-activating protein 30', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171388960, end_position_on_the_genomic_accession=171410240, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  226654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226654, CurrentID=0, Status='live', Symbol='Tstd1', Aliases='AV065366, EG226654, Gm4848, KAT', description='thiosulfate sulfurtransferase (rhodanese)-like domain containing 1', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171419033, end_position_on_the_genomic_accession=171420352, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  226691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226691, CurrentID=0, Status='live', Symbol='AI607873', Aliases='', description='expressed sequence AI607873', other_designations='interferon-activatable protein', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173723427, end_position_on_the_genomic_accession=173741837, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  226695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226695, CurrentID=0, Status='live', Symbol='Ifi205', Aliases='D3, D3cDNAa, Ifi205', description='interferon activated gene 205', other_designations="ifi-205-A|interferon-inducible protein p205-A|protein D3'", map_location='1 80.83 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174011998, end_position_on_the_genomic_accession=174031755, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  226744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226744, CurrentID=0, Status='live', Symbol='Cnst', Aliases='9630058J23Rik', description='consortin, connexin sorting protein', other_designations='', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=179546349, end_position_on_the_genomic_accession=179627473, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  226747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226747, CurrentID=0, Status='live', Symbol='Ahctf1', Aliases='6230412P20Rik, AV011447, Elys', description='AT hook containing transcription factor 1', other_designations='embryonic large molecule derived from yolk sac|putative AT-hook-containing transcription factor 1', map_location='1 83.62 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=179744894, end_position_on_the_genomic_accession=179804015, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  226751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226751, CurrentID=0, Status='live', Symbol='Cdc42bpa', Aliases='A930014J19Rik, DMPK-like', description='CDC42 binding protein kinase alpha', other_designations='CDC42-binding protein kinase alpha|myotonic dystrophy kinase-related CDC-42-binding kinase-alpha', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=179960093, end_position_on_the_genomic_accession=180165606, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  226757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226757, CurrentID=0, Status='live', Symbol='Wdr26', Aliases='1600024A01Rik, AA693241, AI447817, AU044014, C77982, Gid7', description='WD repeat domain 26', other_designations='', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181173226, end_position_on_the_genomic_accession=181211978, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  226777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226777, CurrentID=0, Status='live', Symbol='C130074G19Rik', Aliases='', description='RIKEN cDNA C130074G19 gene', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=184871926, end_position_on_the_genomic_accession=184883036, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  226778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226778, CurrentID=0, Status='live', Symbol='Mark1', Aliases='AW491150, B930025N23Rik, Emk3, Par-1c, mKIAA1477, mPar-1c', description='MAP/microtubule affinity regulating kinase 1', other_designations='ELKL motif serine-threonine protein kinase 3|ELKL motif serine/threonine-protein kinase 3|MAP/microtubule affinity-regulating kinase 1|PAR1 homolog c|mark1', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=184896424, end_position_on_the_genomic_accession=184999572, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  226781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226781, CurrentID=0, Status='live', Symbol='Slc30a10', Aliases='E130106K10Rik, Gm212', description='solute carrier family 30, member 10', other_designations='znT-10', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=185454848, end_position_on_the_genomic_accession=185468762, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  112300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112300, CurrentID=0, Status='live', Symbol='Lmb4', Aliases='', description='lupus in MRL and B6 F2 cross, QTL 4', other_designations='', map_location='10 34.29 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  226791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226791, CurrentID=0, Status='live', Symbol='Lyplal1', Aliases='BC027340, Q96AVO', description='lysophospholipase-like 1', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=186087732, end_position_on_the_genomic_accession=186117310, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  620017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620017, CurrentID=0, Status='live', Symbol='Igkv5-39', Aliases='23-39, Gm1409', description='immunoglobulin kappa variable 5-39', other_designations='Ig kappa chain V-V region L7|anti-HIV-1 reverse transcriptase single-chain variable fragment|immunoglobulin kappa chain variable 5-39', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69900424, end_position_on_the_genomic_accession=69900977, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  226823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226823, CurrentID=0, Status='live', Symbol='Kctd3', Aliases='4930438A20Rik, 9330185B06, E330032J19Rik, NY-REN-45', description='potassium channel tetramerisation domain containing 3', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=188971095, end_position_on_the_genomic_accession=189007868, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  226830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226830, CurrentID=0, Status='live', Symbol='Smyd2', Aliases='1110020E07Rik, 4930402C15, KMT3C, Zmynd14', description='SET and MYND domain containing 2', other_designations='SET and MYND domain-containing protein 2|histone methyltransferase SMYD2', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=189880492, end_position_on_the_genomic_accession=189922379, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  226841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226841, CurrentID=0, Status='live', Symbol='Vash2', Aliases='B130052G07Rik', description='vasohibin 2', other_designations='vasohibin-like protein', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=190947646, end_position_on_the_genomic_accession=190979303, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  226844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226844, CurrentID=0, Status='live', Symbol='Mfsd7b', Aliases='9630055N22Rik, FLVCR, Flvcr1', description='major facilitator superfamily domain containing 7B', other_designations='feline leukemia virus subgroup C cellular receptor 1', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191005832, end_position_on_the_genomic_accession=191026215, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  226849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226849, CurrentID=0, Status='live', Symbol='Ppp2r5a', Aliases='PR61alpha', description="protein phosphatase 2, regulatory subunit B', alpha", other_designations="PP2A B subunit isoform B'-alpha|PP2A B subunit isoform B56-alpha|PP2A B subunit isoform PR61-alpha|PP2A B subunit isoform R5-alpha|protein phosphatase 2, regulatory subunit B (B56), alpha isoform", map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191351981, end_position_on_the_genomic_accession=191397041, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  226856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226856, CurrentID=0, Status='live', Symbol='Lpgat1', Aliases='AI649174, AW112037, BC013667', description='lysophosphatidylglycerol acyltransferase 1', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191717556, end_position_on_the_genomic_accession=191784255, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  226861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226861, CurrentID=0, Status='live', Symbol='Hhat', Aliases='2810432O22Rik, AI462858, AP-2CRE, Skn, Tg(TFAP2A-cre)1Will', description='hedgehog acyltransferase', other_designations='skinny hedgehog protein', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=192512828, end_position_on_the_genomic_accession=192771412, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  620078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620078, CurrentID=0, Status='live', Symbol='C130026I21Rik', Aliases='4930565N07Rik, OTTMUSG00000029174', description='RIKEN cDNA C130026I21 gene', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85241286, end_position_on_the_genomic_accession=85270566, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  226866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226866, CurrentID=0, Status='live', Symbol='Sbspon', Aliases='Gm106, Rpesp', description='somatomedin B and thrombospondin, type 1 domain containing', other_designations='RPE-spondin', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=15853862, end_position_on_the_genomic_accession=15892722, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  226896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226896, CurrentID=0, Status='live', Symbol='Tfap2d', Aliases='Tcfap2d', description='transcription factor AP-2, delta', other_designations='AP2-delta|activating enhancer-binding protein 2 delta|activating enhancer-binding protein 2-delta', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=19103022, end_position_on_the_genomic_accession=19168592, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  620137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620137, CurrentID=0, Status='live', Symbol='Mrgprb13', Aliases='MrgB13', description='MAS-related GPR, member B13', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48311304, end_position_on_the_genomic_accession=48312323, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  226922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226922, CurrentID=0, Status='live', Symbol='Kcnq5', Aliases='7730402H11, 9230107O05Rik, AA589396, D1Mgi1', description='potassium voltage-gated channel, subfamily Q, member 5', other_designations='KQT-like 5|potassium channel subunit alpha KvLQT5|voltage-gated potassium channel subunit Kv7.5|voltage-gated potassium channel type Kv7.5', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=21398399, end_position_on_the_genomic_accession=21962650, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  226970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226970, CurrentID=0, Status='live', Symbol='Arhgef4', Aliases='9330140K16Rik, Asef, ENSMUSG00000070955', description='Rho guanine nucleotide exchange factor (GEF) 4', other_designations='APC-stimulated guanine nucleotide exchange factor', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34801722, end_position_on_the_genomic_accession=34812754, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  226971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226971, CurrentID=0, Status='live', Symbol='Plekhb2', Aliases='2310009M15Rik, AU018239, Evt2, Phdc, evt-2', description='pleckstrin homology domain containing, family B (evectins) member 2', other_designations='PH domain-containing family B member 2|evectin-2', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=34849959, end_position_on_the_genomic_accession=34879585, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  226976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226976, CurrentID=0, Status='live', Symbol='Kansl3', Aliases='4632411B12Rik, 4932435K23, AI431067, AI647574, C85723, Kiaa1310, mKIAA1310', description='KAT8 regulatory NSL complex subunit 3', other_designations='NSL complex protein NSL3|non-specific lethal 3 homolog', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36329947, end_position_on_the_genomic_accession=36369194, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  226977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226977, CurrentID=0, Status='live', Symbol='Actr1b', Aliases='2310066K23Rik, AA960180, AI851923, Arp1b', description='ARP1 actin-related protein 1B, centractin beta', other_designations='ARP1 actin-related protein 1 homolog B, centractin beta|actin-related protein 1B', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36698114, end_position_on_the_genomic_accession=36710017, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  226982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226982, CurrentID=0, Status='live', Symbol='Eif5b', Aliases='A030003E17Rik, BC018347, IF2', description='eukaryotic translation initiation factor 5B', other_designations='eIF-5B|translation initiation factor IF-2', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37998000, end_position_on_the_genomic_accession=38055579, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  226999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=226999, CurrentID=0, Status='live', Symbol='Slc9a2', Aliases='2210416H12Rik, 4932415O19Rik, AV333884, NHE2', description='solute carrier family 9 (sodium/hydrogen exchanger), member 2', other_designations='', map_location='1 19.66 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40681712, end_position_on_the_genomic_accession=40768885, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  620235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620235, CurrentID=0, Status='live', Symbol='Siglec15', Aliases='Cd33l3, EG620235, SIGLEC-I', description='sialic acid binding Ig-like lectin 15', other_designations='CD33 molecule-like 3', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=78042499, end_position_on_the_genomic_accession=78057435, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  620246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620246, CurrentID=0, Status='live', Symbol='Gpr52', Aliases='EG620246', description='G protein-coupled receptor 52', other_designations='', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160576668, end_position_on_the_genomic_accession=160577753, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  620253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620253, CurrentID=0, Status='live', Symbol='Dcpp3', Aliases='Dcpp-3, EG620253, SPT-3', description='demilune cell and parotid protein 3', other_designations='sublingual parotid tongue-3 salivary protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23916698, end_position_on_the_genomic_accession=23919441, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  227058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227058, CurrentID=0, Status='live', Symbol='Dnah7b', Aliases='Dnahc7b, Gm107', description='dynein, axonemal, heavy chain 7B', other_designations='', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=46065786, end_position_on_the_genomic_accession=46373551, orientation='plus', exon_count=78, OMIM='', no_hdr0=''),
  227059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227059, CurrentID=0, Status='live', Symbol='Slc39a10', Aliases='2900042E17Rik, 5430433I10, mKIAA1265', description='solute carrier family 39 (zinc transporter), member 10', other_designations='ZIP-10|solute carrier family 39 member 10|zrt- and Irt-like protein 10', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=46807544, end_position_on_the_genomic_accession=46925772, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  620292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620292, CurrentID=0, Status='live', Symbol='Cntnap5c', Aliases='Caspr5-3, EG620292, Gm325', description='contactin associated protein-like 5C', other_designations='cell recognition molecule Caspr5-3|cell recognition molecule Caspr5c|contactin associated protein-like 5-3|contactin-associated protein-like 5c', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57769077, end_position_on_the_genomic_accession=58410429, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  227094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227094, CurrentID=0, Status='live', Symbol='Tmem194b', Aliases='5330401P04Rik, 9830163O08, Nemp2', description='transmembrane protein 194B', other_designations='', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52568535, end_position_on_the_genomic_accession=52651919, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  227095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227095, CurrentID=0, Status='live', Symbol='Hibch', Aliases='2610509I15Rik, HIBYL-COA-H', description='3-hydroxyisobutyryl-Coenzyme A hydrolase', other_designations='HIB-CoA hydrolase', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52845046, end_position_on_the_genomic_accession=52920860, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  227099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227099, CurrentID=0, Status='live', Symbol='Pms1', Aliases='', description='postmeiotic segregation increased 1 (S. cerevisiae)', other_designations='', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53189187, end_position_on_the_genomic_accession=53297026, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  227102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227102, CurrentID=0, Status='live', Symbol='Ormdl1', Aliases='C730042F17Rik', description='ORM1-like 1 (S. cerevisiae)', other_designations='', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53297095, end_position_on_the_genomic_accession=53310245, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  227120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227120, CurrentID=0, Status='live', Symbol='Plcl1', Aliases='C230017K02Rik, PLC-L, PLC-L1, PLCE, PLCL, PLDL1, PRIP, PRIP-1', description='phospholipase C-like 1', other_designations='phospholipase C, eta 2 pseudogene|phospholipase C-related but catalytically inactive protein', map_location='1 C1.2-C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=55405946, end_position_on_the_genomic_accession=55754285, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  227154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227154, CurrentID=0, Status='live', Symbol='Stradb', Aliases='AA792893, Als2cr2, B830008M19, D1Ucla2, ILPIP, ILPIPA, PRO1038, Papk, Syradb', description='STE20-related kinase adaptor beta', other_designations='ILP-interacting protein homolog|STRAD beta|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 2 protein homolog|polyploidy associated protein kinase|polyploidy-associated protein kinase|pseudokinase ALS2CR2', map_location='1 29.2 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58973527, end_position_on_the_genomic_accession=58995126, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  227157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227157, CurrentID=0, Status='live', Symbol='Mpp4', Aliases='Dlg6', description='membrane protein, palmitoylated 4 (MAGUK p55 subfamily member 4)', other_designations='discs large homolog 6|mDLG6|membrane palmitoylated protein 4', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59120935, end_position_on_the_genomic_accession=59163389, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  112313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112313, CurrentID=0, Status='live', Symbol='Char1', Aliases='', description='P. chabaudi malaria resistance QTL 1', other_designations='', map_location='9 51.41 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  68668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=68668, CurrentID=0, Status='live', Symbol='Klk5', Aliases='1110030O19Rik', description='kallikrein related-peptidase 5', other_designations='kallikrein 5', map_location='7 28.3 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43834157, end_position_on_the_genomic_accession=43851181, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  227195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227195, CurrentID=0, Status='live', Symbol='Ino80d', Aliases='7330405I11, A430093A21Rik, BC010584, OTTMUSG00000002594', description='INO80 complex subunit D', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63047801, end_position_on_the_genomic_accession=63114267, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  227197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227197, CurrentID=0, Status='live', Symbol='Ndufs1', Aliases='5830412M15Rik, 9930026A05Rik', description='NADH dehydrogenase (ubiquinone) Fe-S protein 1', other_designations='CI-75kD|complex I-75kD', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63143592, end_position_on_the_genomic_accession=63176822, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  620419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620419, CurrentID=0, Status='live', Symbol='Zfp963', Aliases='', description='zinc finger protein 963', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69741639, end_position_on_the_genomic_accession=69749962, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  227210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227210, CurrentID=0, Status='live', Symbol='Ccnyl1', Aliases='9630037P07Rik, AW554339', description='cyclin Y-like 1', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=64691005, end_position_on_the_genomic_accession=64725646, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  227231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227231, CurrentID=0, Status='live', Symbol='Cps1', Aliases='4732433M03Rik, CPS, D1Ucla3', description='carbamoyl-phosphate synthetase 1', other_designations='CPSase I|carbamoyl-phosphate synthetase I|carbamoylphosphate synthetase I', map_location='1 33.75 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=67122426, end_position_on_the_genomic_accession=67231267, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  751522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751522, CurrentID=0, Status='live', Symbol='Mir495', Aliases='Mirn495, mir-495, mmu-mir-495', description='microRNA 495', other_designations='mir 495', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109718754, end_position_on_the_genomic_accession=109718816, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751527, CurrentID=0, Status='live', Symbol='Mir19b-1', Aliases='Mirn19b-1, mir-19b-1', description='microRNA 19b-1', other_designations='mir 19b-1|mmu-mir-19b-1', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115044305, end_position_on_the_genomic_accession=115044391, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751532, CurrentID=0, Status='live', Symbol='Mir450b', Aliases='Mirn450b, mir-450b', description='microRNA 450b', other_designations='mir 450b|mmu-mir-450b', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=53047997, end_position_on_the_genomic_accession=53048078, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  751534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751534, CurrentID=0, Status='live', Symbol='Mir676', Aliases='Mirn676, mir-676, mmu-mir-676', description='microRNA 676', other_designations='mir 676', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100381097, end_position_on_the_genomic_accession=100381185, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751535, CurrentID=0, Status='live', Symbol='Mir672', Aliases='Mirn672, mir-672, mmu-mir-672', description='microRNA 672', other_designations='mir 672', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=104116175, end_position_on_the_genomic_accession=104116274, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  751537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751537, CurrentID=0, Status='live', Symbol='Mir497', Aliases='Mir, Mirn497, mir-497a, mmu-mir-497, mmu-mir-497a', description='microRNA 497', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70234717, end_position_on_the_genomic_accession=70234800, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751546, CurrentID=0, Status='live', Symbol='Mir374b', Aliases='Mir374, Mirn374, mir-374b, mmu-mir-374, mmu-mir-374b', description='microRNA 374b', other_designations='microRNA 374|mir 374', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=103573060, end_position_on_the_genomic_accession=103573154, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  751547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751547, CurrentID=0, Status='live', Symbol='Mir673', Aliases='Mirn673, mir-673, mmu-mir-673', description='microRNA 673', other_designations='mir 673', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109571990, end_position_on_the_genomic_accession=109572080, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751549, CurrentID=0, Status='live', Symbol='Mir92-1', Aliases='Mirn92-1, mir-92a-1', description='microRNA 92-1', other_designations='mir 92-1|mmu-mir-92-1|mmu-mir-92a-1', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115044427, end_position_on_the_genomic_accession=115044506, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751550, CurrentID=0, Status='live', Symbol='Mir146b', Aliases='Mirn146b, mir-146b', description='microRNA 146b', other_designations='mir 146b|mmu-mir-146b', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46342762, end_position_on_the_genomic_accession=46342870, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751556, CurrentID=0, Status='live', Symbol='Mir682', Aliases='Mirn682, mir-682, mmu-mir-682', description='microRNA 682', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=75645046, end_position_on_the_genomic_accession=75645141, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  751557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751557, CurrentID=0, Status='live', Symbol='Mir615', Aliases='Mir, Mirn615, mir-615, mmu-mir-615', description='microRNA 615', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103014910, end_position_on_the_genomic_accession=103015001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  227288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227288, CurrentID=0, Status='live', Symbol='Cxcr1', Aliases='Il8ra', description='chemokine (C-X-C motif) receptor 1', other_designations='CXC chemokine receptor 1|CXC-R1|CXCR-1|IL-8R A|high affinity interleukin-8 receptor A|interleukin 8 receptor, alpha', map_location='1 38.44 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74191785, end_position_on_the_genomic_accession=74198144, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  227289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227289, CurrentID=0, Status='live', Symbol='Gpbar1', Aliases='BG37, GPCR, GPR131, M-BAR, TGR5', description='G protein-coupled bile acid receptor 1', other_designations='G protein-coupled receptor TGR5|membrane bile acid receptor|membrane-type receptor for bile acids', map_location='1 38.53 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74278600, end_position_on_the_genomic_accession=74279589, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  227290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227290, CurrentID=0, Status='live', Symbol='Aamp', Aliases='AU040907-rs, Aamp', description='angio-associated migratory protein', other_designations='', map_location='1 38.53 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74279840, end_position_on_the_genomic_accession=74284738, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  227292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227292, CurrentID=0, Status='live', Symbol='Ctdsp1', Aliases='GIP, NLIIF, Nif3, SCP1', description='CTD (carboxy-terminal domain, RNA polymerase II, polypeptide A) small phosphatase 1', other_designations='NLI-interacting factor 3|golli-interacting protein|nuclear LIM interactor-interacting factor 3|small C-terminal domain phosphatase 1|small CTD phosphatase 1', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74391609, end_position_on_the_genomic_accession=74397285, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  227298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227298, CurrentID=0, Status='live', Symbol='Fam134a', Aliases='', description='family with sequence similarity 134, member A', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75142786, end_position_on_the_genomic_accession=75147909, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  227325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227325, CurrentID=0, Status='live', Symbol='Dner', Aliases='A930026D19Rik, BET, Bret', description='delta/notch-like EGF repeat containing', other_designations='brain EGF repeat-containing transmembrane protein|delta/notch-like EGF-related receptor|transmembrane protein Bet', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=84369839, end_position_on_the_genomic_accession=84696221, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  227326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227326, CurrentID=0, Status='live', Symbol='Gpr55', Aliases='Gm218, Lpir1', description='G protein-coupled receptor 55', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85933206, end_position_on_the_genomic_accession=85991567, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  227327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227327, CurrentID=0, Status='live', Symbol='B3gnt7', Aliases='C330001H22Rik, beta-3GnT7', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 7', other_designations='BGnT-7|beta-1,3-Gn-T7|beta-1,3-N-acetylglucosaminyltransferase 7|beta3Gn-T7', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=86303221, end_position_on_the_genomic_accession=86307305, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  227331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227331, CurrentID=0, Status='live', Symbol='Gigyf2', Aliases='2610016F01Rik, A830080H02Rik, AI852361, AW259676, BC006835, Tnrc15, mKIAA0642', description='GRB10 interacting GYF protein 2', other_designations='GRB10-interacting GYF protein 2|trinucleotide repeat containing 15', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87326981, end_position_on_the_genomic_accession=87450810, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  227333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227333, CurrentID=0, Status='live', Symbol='Dgkd', Aliases='AI841987, D330025K09, DGKdelta, dgkd-2', description='diacylglycerol kinase, delta', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87853273, end_position_on_the_genomic_accession=87944490, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  227334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227334, CurrentID=0, Status='live', Symbol='Usp40', Aliases='B230215L03Rik, C730029K03', description='ubiquitin specific peptidase 40', other_designations='deubiquitinating enzyme 40|ubiquitin specific protease 40|ubiquitin thioesterase 40|ubiquitin thiolesterase 40|ubiquitin-specific-processing protease 40', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87945121, end_position_on_the_genomic_accession=88008577, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  227357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227357, CurrentID=0, Status='live', Symbol='Espnl', Aliases='Gm556', description='espin-like', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91321164, end_position_on_the_genomic_accession=91348303, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  227358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227358, CurrentID=0, Status='live', Symbol='Fam132b', Aliases='4832406C22, myonectin', description='family with sequence similarity 132, member B', other_designations='complement C1q tumor necrosis factor-related protein 15|protein FAM132B', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=91366162, end_position_on_the_genomic_accession=91374217, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  227377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227377, CurrentID=0, Status='live', Symbol='Farp2', Aliases='AI465173, BC009153, D030026M03Rik, Fir, mKIAA0793', description='FERM, RhoGEF and pleckstrin domain protein 2', other_designations='FERM domain including RhoGEF', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93512104, end_position_on_the_genomic_accession=93621976, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  227394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227394, CurrentID=0, Status='live', Symbol='Slco4c1', Aliases='9530051F04, C330017E21Rik, OATP-H, OATP-M1, OATP4C1, OATPX, PRO2176, SLC21A20, oatp-R', description='solute carrier organic anion transporter family, member 4C1', other_designations='slco4c1|solute carrier family 21 member 20', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=96818784, end_position_on_the_genomic_accession=96872171, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  227399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227399, CurrentID=0, Status='live', Symbol='Ppip5k2', Aliases='AW555814, D330021B20, Hisppd1, Vip2, mKIAA0433', description='diphosphoinositol pentakisphosphate kinase 2', other_designations='VIP1 homolog 2|histidine acid phosphatase domain containing 1|histidine acid phosphatase domain-containing protein 1|insP6 and PP-IP5 kinase 2', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=97706043, end_position_on_the_genomic_accession=97770250, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  227446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227446, CurrentID=0, Status='live', Symbol='2310035C23Rik', Aliases='6430401N10, Kiaa1468, mKIAA1468', description='RIKEN cDNA 2310035C23 gene', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=105663843, end_position_on_the_genomic_accession=105755131, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  227449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227449, CurrentID=0, Status='live', Symbol='Zcchc2', Aliases='9930114B20Rik, AW212015', description='zinc finger, CCHC domain containing 2', other_designations='', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=105990406, end_position_on_the_genomic_accession=106034077, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  620695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620695, CurrentID=0, Status='live', Symbol='Gm13889', Aliases='Ag2, OTTMUSG00000014994', description='predicted gene 13889', other_designations='protein Ag2 homolog', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93955810, end_position_on_the_genomic_accession=93957100, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  227485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227485, CurrentID=0, Status='live', Symbol='Cdh19', Aliases='', description='cadherin 19, type 2', other_designations='', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=110888002, end_position_on_the_genomic_accession=110977593, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  620709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620709, CurrentID=0, Status='live', Symbol='Lcn6', Aliases='9230101D24Rik', description='lipocalin 6', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25676632, end_position_on_the_genomic_accession=25681608, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  227522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227522, CurrentID=0, Status='live', Symbol='Rpp38', Aliases='C330006A15Rik', description='ribonuclease P/MRP 38 subunit', other_designations='RNaseP protein p38|ribonuclease P/MRP 38kDa subunit', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3328943, end_position_on_the_genomic_accession=3333024, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  227525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227525, CurrentID=0, Status='live', Symbol='Dclre1c', Aliases='9930121L06Rik, AI661365, Art, Snm1l', description='DNA cross-link repair 1C', other_designations='DNA cross-link repair 1A, PSO2 homolog|DNA cross-link repair 1C protein|DNA cross-link repair 1C, PSO2 homolog|SNM1-like protein', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3424062, end_position_on_the_genomic_accession=3474330, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  227526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227526, CurrentID=0, Status='live', Symbol='Cdnf', Aliases='9330140G23, Armetl1', description='cerebral dopamine neurotrophic factor', other_designations='ARMET-like protein 1|arginine-rich, mutated in early stage tumors-like 1|conserved dopamine neurotrophic factor', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3512966, end_position_on_the_genomic_accession=3526376, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  227541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227541, CurrentID=0, Status='live', Symbol='Camk1d', Aliases='A630059D12Rik, CKLiK, CaMKIdelta, E030025C11Rik, mCKLiK', description='calcium/calmodulin-dependent protein kinase ID', other_designations='caM kinase I delta|caM kinase ID|caM-KI delta|caMKI delta|calcium/calmodulin-dependent protein kinase 1D|camKI-like protein kinase', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=5293457, end_position_on_the_genomic_accession=5715364, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  227545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227545, CurrentID=0, Status='live', Symbol='Proser2', Aliases='5430407P10Rik, AU067770', description='proline and serine rich 2', other_designations='', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=6098493, end_position_on_the_genomic_accession=6130294, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  751865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751865, CurrentID=0, Status='live', Symbol='Sap25', Aliases='', description='sin3 associated polypeptide', other_designations='25 kDa Sin3-associated polypeptide|mSin3A-binding protein|sin3 corepressor complex subunit SAP25|sin3A-binding protein, SAP25', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137641473, end_position_on_the_genomic_accession=137642902, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  751866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=751866, CurrentID=0, Status='live', Symbol='Kis2', Aliases='Kis2-T1-T2, Kis2-T3, Kis2-T4, Kis2-T5, Platr24, Xpcl1, Kis2', description='Kaplan integration site 2', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52742563, end_position_on_the_genomic_accession=52744593, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  227580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227580, CurrentID=0, Status='live', Symbol='C1ql3', Aliases='1110065A22Rik, AI661623, C1ql, C1qtnf13, CTRP13, K100', description='C1q-like 3', other_designations='C1q and tumor necrosis factor-related protein 13|C1q/TNF-related protein 13|gliacolin', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=13001887, end_position_on_the_genomic_accession=13010864, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  620807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=620807, CurrentID=0, Status='live', Symbol='Mup6', Aliases='Gm12544, Mup2, OTTMUSG00000007423', description='major urinary protein 6', other_designations='major urinary protein 2', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=60003549, end_position_on_the_genomic_accession=60006214, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  227606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227606, CurrentID=0, Status='live', Symbol='Tbpl2', Aliases='Gm348, Trf3', description='TATA box binding protein like 2', other_designations='TATA box-binding protein-related factor 3|TBP-like protein 2|TBP-related factor 3 (Trf3)|tbpl2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24071366, end_position_on_the_genomic_accession=24096595, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  227612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227612, CurrentID=0, Status='live', Symbol='Tor4a', Aliases='A830007P12Rik', description='torsin family 4, member A', other_designations='torsin family 4 member A|torsin family protein C9orf167 homolog', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25192720, end_position_on_the_genomic_accession=25196813, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  227613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227613, CurrentID=0, Status='live', Symbol='Tubb4b', Aliases='4930542G03Rik, Tubb2c, Tubb2c1', description='tubulin, beta 4B class IVB', other_designations='novel tubulin protein|tubulin beta-2C chain|tubulin, beta 2C|tubulin, beta 2c1|tubulin, beta, 2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25222158, end_position_on_the_genomic_accession=25224702, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  227615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227615, CurrentID=0, Status='live', Symbol='Tmem203', Aliases='C730025P13Rik, Hbebp1', description='transmembrane protein 203', other_designations='HBeAg-binding protein 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25255439, end_position_on_the_genomic_accession=25256352, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  227618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227618, CurrentID=0, Status='live', Symbol='Lrrc26', Aliases='', description='leucine rich repeat containing 26', other_designations='BK channel auxiliary gamma subunit LRRC26|BK channel auxilliary gamma subunit LRRC26', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25289911, end_position_on_the_genomic_accession=25291193, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  227619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227619, CurrentID=0, Status='live', Symbol='Man1b1', Aliases='E430019H13Rik, ERMan1, Gm108, MANA-ER', description='mannosidase, alpha, class 1B, member 1', other_designations='ER alpha-1,2-mannosidase|ER mannosidase 1|man9GlcNAc2-specific-processing alpha-mannosidase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25332743, end_position_on_the_genomic_accession=25352213, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  227620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227620, CurrentID=0, Status='live', Symbol='Uap1l1', Aliases='5730445F03Rik', description='UDP-N-acteylglucosamine pyrophosphorylase 1-like 1', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25361489, end_position_on_the_genomic_accession=25365686, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  227622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227622, CurrentID=0, Status='live', Symbol='BC029214', Aliases='D930050G13Rik, Paxx', description='cDNA sequence BC029214', other_designations='paralog of XRCC4 and XLF', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25455120, end_position_on_the_genomic_accession=25461399, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  227624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227624, CurrentID=0, Status='live', Symbol='Rabl6', Aliases='B230208H17Rik, Rbel1, Rbel1a, Rbel1b, parf', description='RAB, member RAS oncogene family-like 6', other_designations='GTP-binding protein Parf|RAB, member of RAS oncogene family-like 6|putative GTP-binding protein Parf|rab-like protein 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25583018, end_position_on_the_genomic_accession=25608446, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  227627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227627, CurrentID=0, Status='live', Symbol='Obp2a', Aliases='BC027556, Lcn13', description='odorant binding protein 2A', other_designations='epididymal-specific lipocalin-13|lipocalin 13', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25697543, end_position_on_the_genomic_accession=25703326, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  227631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227631, CurrentID=0, Status='live', Symbol='Sohlh1', Aliases='Gm110, Nohlh', description='spermatogenesis and oogenesis specific basic helix-loop-helix 1', other_designations='helix-loop-helix protein NOHLH', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25842995, end_position_on_the_genomic_accession=25847248, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  227632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227632, CurrentID=0, Status='live', Symbol='Kcnt1', Aliases='C030030G16Rik, Slack, Slo2.2, slo2', description='potassium channel, subfamily T, member 1', other_designations='potassium channel subunit homolog|sodium-activated potassium channel', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25863791, end_position_on_the_genomic_accession=25918273, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  227634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227634, CurrentID=0, Status='live', Symbol='Camsap1', Aliases='9530003A05Rik, C77823, PRO2405', description='calmodulin regulated spectrin-associated protein 1', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25926838, end_position_on_the_genomic_accession=25983452, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  227638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227638, CurrentID=0, Status='live', Symbol='Qsox2', Aliases='BC030934, Qscn6l1, Soxn', description='quiescin Q6 sulfhydryl oxidase 2', other_designations='quiescin Q6-like 1', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26209124, end_position_on_the_genomic_accession=26237399, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  227644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227644, CurrentID=0, Status='live', Symbol='Snapc4', Aliases='5730436L13Rik', description='small nuclear RNA activating complex, polypeptide 4', other_designations='SNAPc subunit 4|snRNA-activating protein complex 190 kDa subunit|snapc4', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26362765, end_position_on_the_genomic_accession=26381072, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  227648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227648, CurrentID=0, Status='live', Symbol='Sec16a', Aliases='AU024582, C230052J16Rik', description='SEC16 homolog A, endoplasmic reticulum export factor', other_designations='SEC16 homolog A', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26409431, end_position_on_the_genomic_accession=26446162, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  227656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227656, CurrentID=0, Status='live', Symbol='Rexo4', Aliases='Gm111, Xpmc2h', description='REX4, RNA exonuclease 4 homolog (S. cerevisiae)', other_designations='XPMC2 prevents mitotic catastrophe 2 homolog|Xenopus prevents mitotic catastrophe 2 homolog|exonuclease XPMC2|prevents mitotic catastrophe 2 protein homolog', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26953563, end_position_on_the_genomic_accession=26964539, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  227659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227659, CurrentID=0, Status='live', Symbol='Slc2a6', Aliases='A330096C23, F630103L12Rik, Glut6, Glut9', description='solute carrier family 2 (facilitated glucose transporter), member 6', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27021363, end_position_on_the_genomic_accession=27028001, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  227671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227671, CurrentID=0, Status='live', Symbol='Gbgt1', Aliases='Fgs, Fs', description='globoside alpha-1,3-N-acetylgalactosaminyltransferase 1', other_designations='forssman glycolipid synthase|forssman glycolipid synthetase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28479229, end_position_on_the_genomic_accession=28506030, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  227674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227674, CurrentID=0, Status='live', Symbol='Ddx31', Aliases='5830444G11Rik, Gm997', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 31', other_designations='DEAD box protein 31', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28840406, end_position_on_the_genomic_accession=28905575, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  227682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227682, CurrentID=0, Status='live', Symbol='Trub2', Aliases='G430055L02Rik', description='TruB pseudouridine (psi) synthase homolog 2 (E. coli)', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29761223, end_position_on_the_genomic_accession=29787810, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  227683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227683, CurrentID=0, Status='live', Symbol='Coq4', Aliases='A330041N06, D2Ertd97e', description='coenzyme Q4', other_designations='EST-MNCb4625|coenzyme Q biosynthesis protein 4 homolog|coenzyme Q4 homolog', map_location='2 20.62 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29788226, end_position_on_the_genomic_accession=29797743, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  227693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227693, CurrentID=0, Status='live', Symbol='Zer1', Aliases='C230075L19Rik, Zyg11bl', description='zyg-11 related, cell cycle regulator', other_designations='zer-1 homolog|zyg-11 homolog B-like protein', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30097283, end_position_on_the_genomic_accession=30124804, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  227695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227695, CurrentID=0, Status='live', Symbol='D2Wsu81e', Aliases='9930028P05, AL033358', description='DNA segment, Chr 2, Wayne State University 81, expressed', other_designations='', map_location='2 21.29 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30173447, end_position_on_the_genomic_accession=30178495, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  227696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227696, CurrentID=0, Status='live', Symbol='Phyhd1', Aliases='', description='phytanoyl-CoA dioxygenase domain containing 1', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30266203, end_position_on_the_genomic_accession=30282149, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  227697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227697, CurrentID=0, Status='live', Symbol='Dolk', Aliases='BC026973, Tmem15, mKIAA1094', description='dolichol kinase', other_designations='dolk|transmembrane protein 15', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30284229, end_position_on_the_genomic_accession=30286354, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  227699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227699, CurrentID=0, Status='live', Symbol='Nup188', Aliases='BC025526, U89435, mKIAA0169', description='nucleoporin 188', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30286422, end_position_on_the_genomic_accession=30344263, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  227700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227700, CurrentID=0, Status='live', Symbol='Sh3glb2', Aliases='SPAS-1', description='SH3-domain GRB2-like endophilin B2', other_designations='SH3 domain-containing GRB2-like protein B2|SH3-containing protein SH3GLB2|stimulator of prostatic adenocarcinoma-specific T-cells-1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30344777, end_position_on_the_genomic_accession=30359316, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  227707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227707, CurrentID=0, Status='live', Symbol='BC005624', Aliases='', description='cDNA sequence BC005624', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30972833, end_position_on_the_genomic_accession=30981941, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  227715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227715, CurrentID=0, Status='live', Symbol='Exosc2', Aliases='Rrp4', description='exosome component 2', other_designations="exosome complex exonuclease RRP4|homolog of Yeast RRP4 (ribosomal RNA processing 4), 3'-5'-exoribonuclease|ribosomal RNA-processing protein 4", map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31670737, end_position_on_the_genomic_accession=31681349, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  227717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227717, CurrentID=0, Status='live', Symbol='Qrfp', Aliases='P518', description='pyroglutamylated RFamide peptide', other_designations='P518 precursor protein (P518)', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31806168, end_position_on_the_genomic_accession=31810518, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  227720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227720, CurrentID=0, Status='live', Symbol='Nup214', Aliases='BC039282, CAN, D2H9S46E', description='nucleoporin 214', other_designations='214 kDa nucleoporin|nucleoporin Nup214', map_location='2 21.97 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31974420, end_position_on_the_genomic_accession=32053975, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  227721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227721, CurrentID=0, Status='live', Symbol='Plpp7', Aliases='D830019K17Rik, NET39, Ppapdc3', description='phospholipid phosphatase 7 (inactive)', other_designations='nuclear envelope transmembrane protein 39|phosphatidic acid phosphatase type 2 domain containing 3|phosphatidic acid phosphatase type 2 domain-containing protein 3|probable lipid phosphate phosphatase PPAPDC3', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32095516, end_position_on_the_genomic_accession=32110820, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  227723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227723, CurrentID=0, Status='live', Symbol='Prrc2b', Aliases='5830434P21Rik, AI173903, Bat2l, Bat2l1, D430039P21, mKIAA0515', description='proline-rich coiled-coil 2B', other_designations='HLA-B-associated transcript 2-like 1|proline-rich coiled-coil protein 2B|protein BAT2-like 1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32151148, end_position_on_the_genomic_accession=32234537, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  227731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227731, CurrentID=0, Status='live', Symbol='Slc25a25', Aliases='1110030N17Rik, MCSC, mKIAA1896', description='solute carrier family 25 (mitochondrial carrier, phosphate carrier), member 25', other_designations='mitochondrial Ca2+-dependent solute carrier|small calcium-binding mitochondrial carrier protein 2|solute carrier family 25 member 25', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32414487, end_position_on_the_genomic_accession=32451470, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  227733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227733, CurrentID=0, Status='live', Symbol='Pip5kl1', Aliases='BC028795', description='phosphatidylinositol-4-phosphate 5-kinase-like 1', other_designations='PI(4)P 5-kinase-like protein 1|PIPKH|phosphatidylinositol phosphate kinase homolog|phosphatidylinositol-4-phosphate 5-kinase-like protein 1|ptdIns(4)P-5-kinase-like protein 1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32574788, end_position_on_the_genomic_accession=32583780, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  227736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227736, CurrentID=0, Status='live', Symbol='Cfap157', Aliases='1700019L03Rik', description='cilia and flagella associated protein 157', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32777413, end_position_on_the_genomic_accession=32784405, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  227737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227737, CurrentID=0, Status='live', Symbol='Fam129b', Aliases='9130404D14Rik', description='family with sequence similarity 129, member B', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32876134, end_position_on_the_genomic_accession=32925253, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  227738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227738, CurrentID=0, Status='live', Symbol='Lrsam1', Aliases='', description='leucine rich repeat and sterile alpha motif containing 1', other_designations='leucine-rich repeat and sterile alpha motif-containing protein 1|tsg101-associated ligase', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32925215, end_position_on_the_genomic_accession=32961654, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  227743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227743, CurrentID=0, Status='live', Symbol='Mapkap1', Aliases='AI591529, D230039K05Rik, Sin1', description='mitogen-activated protein kinase associated protein 1', other_designations='SAPK-interacting protein 1|TORC2 subunit MAPKAP1|mitogen activated protein kinase associated protein 1|mitogen-activated protein kinase 2-associated protein 1|stress-activated map kinase-interacting protein 1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34406771, end_position_on_the_genomic_accession=34624958, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  227746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227746, CurrentID=0, Status='live', Symbol='Rabepk', Aliases='8430412M01Rik, 9530020D24Rik, AV073337, C87311, Rab9p40', description='Rab9 effector protein with kelch motifs', other_designations='Rab9 effector p40', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=34777892, end_position_on_the_genomic_accession=34799912, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  227753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227753, CurrentID=0, Status='live', Symbol='Gsn', Aliases='ADF', description='gelsolin', other_designations='actin-depolymerizing factor|brevin', map_location='2 23.5 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35256359, end_position_on_the_genomic_accession=35307902, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  227800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227800, CurrentID=0, Status='live', Symbol='Rabgap1', Aliases='Gapcena, mKIAA4104', description='RAB GTPase activating protein 1', other_designations='GAP and centrosome-associated protein|rab6 GTPase activating protein (GAP and centrosome-associated)|rab6 GTPase-activating protein GAPCenA|rabgap1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37443261, end_position_on_the_genomic_accession=37566454, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  227801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227801, CurrentID=0, Status='live', Symbol='Dennd1a', Aliases='6030446I19Rik, connecdenn', description='DENN/MADD domain containing 1A', other_designations='connecdenn 1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37798990, end_position_on_the_genomic_accession=38287384, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  227835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227835, CurrentID=0, Status='live', Symbol='Gtdc1', Aliases='E330008O22Rik', description='glycosyltransferase-like domain containing 1', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=44564412, end_position_on_the_genomic_accession=44927918, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  103511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103511, CurrentID=0, Status='live', Symbol='Fam26e', Aliases='BB146404', description='family with sequence similarity 26, member E', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=34091352, end_position_on_the_genomic_accession=34096519, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  621080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621080, CurrentID=0, Status='live', Symbol='AI429214', Aliases='', description='expressed sequence AI429214', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=36993575, end_position_on_the_genomic_accession=36995533, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  227867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227867, CurrentID=0, Status='live', Symbol='Epc2', Aliases='5830499L14, D2Ertd694e', description='enhancer of polycomb homolog 2 (Drosophila)', other_designations='', map_location='2 28.62 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=49451486, end_position_on_the_genomic_accession=49551942, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  227929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227929, CurrentID=0, Status='live', Symbol='Cytip', Aliases='A130053M09Rik, AI462064, C80816, Cbp, Cybr, Pscdbp', description='cytohesin 1 interacting protein', other_designations='cbp HE|cytohesin binding protein|cytohesin-binding protein HE|pleckstrin homology Sec7 and coiled-coil domains-binding protein|pleckstrin homology, Sec7 and coiled-coil domains, binding protein|tamalin', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=58129139, end_position_on_the_genomic_accession=58195532, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  227933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227933, CurrentID=0, Status='live', Symbol='Ccdc148', Aliases='5830402J09', description='coiled-coil domain containing 148', other_designations='', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=58821697, end_position_on_the_genomic_accession=59018606, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  227937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227937, CurrentID=0, Status='live', Symbol='Pkp4', Aliases='5031422I09Rik, 9430019K17Rik, Armrp, p0071', description='plakophilin 4', other_designations='armadillo-related protein', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=59160850, end_position_on_the_genomic_accession=59355205, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  96875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=96875, CurrentID=0, Status='live', Symbol='Prg4', Aliases='CACP, DOL54, JCAP, MSF, SZP, lubricin', description='proteoglycan 4 (megakaryocyte stimulating factor, articular superficial zone protein)', other_designations='megakaryocyte-stimulating factor|proteoglycan 3|superficial zone proteoglycan', map_location='1 G', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=150449412, end_position_on_the_genomic_accession=150466221, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  227960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=227960, CurrentID=0, Status='live', Symbol='Gca', Aliases='5133401E04Rik, AI573844', description='grancalcin', other_designations='', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=62664327, end_position_on_the_genomic_accession=62694109, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  228003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228003, CurrentID=0, Status='live', Symbol='Klhl41', Aliases='Gm112, Kbtbd10, SARCOSIN', description='kelch-like 41', other_designations='kelch repeat and BTB (POZ) domain containing 10|kelch repeat and BTB domain-containing protein 10', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69670120, end_position_on_the_genomic_accession=69684239, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  228005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228005, CurrentID=0, Status='live', Symbol='Ppig', Aliases='AU019516, AU022200, B230312B02Rik, CYP, SRCyp', description='peptidyl-prolyl isomerase G (cyclophilin G)', other_designations='PPIase G|cyclophilin G|peptidylprolyl isomerase G|rotamase G', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69722617, end_position_on_the_genomic_accession=69754059, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  96935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=96935, CurrentID=0, Status='live', Symbol='Susd4', Aliases='AI848994, E430021N18Rik, N28096', description='sushi domain containing 4', other_designations='', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=182763865, end_position_on_the_genomic_accession=182896596, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  228012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228012, CurrentID=0, Status='live', Symbol='Tlk1', Aliases='4930545J15Rik', description='tousled-like kinase 1', other_designations='', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=70712407, end_position_on_the_genomic_accession=70825480, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  228019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228019, CurrentID=0, Status='live', Symbol='Mettl8', Aliases='BC004636, Tip', description='methyltransferase like 8', other_designations='tension induced/inhibited protein|tension-induced/inhibited protein', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=70964558, end_position_on_the_genomic_accession=71055643, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  621239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621239, CurrentID=0, Status='live', Symbol='Nhlrc4', Aliases='F430201B04Rik', description='NHL repeat containing 4', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25942233, end_position_on_the_genomic_accession=25944931, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  228026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228026, CurrentID=0, Status='live', Symbol='Pdk1', Aliases='B830012B01, D530020C15Rik', description='pyruvate dehydrogenase kinase, isoenzyme 1', other_designations='PDH kinase 1|[Pyruvate dehydrogenase (acetyl-transferring)] kinase isozyme 1, mitochondrial|[Pyruvate dehydrogenase [lipoamide]] kinase isozyme 1, mitochondrial', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71873218, end_position_on_the_genomic_accession=71903858, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  96957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=96957, CurrentID=0, Status='live', Symbol='Tmem62', Aliases='B830009D23Rik, C77355', description='transmembrane protein 62', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120976719, end_position_on_the_genomic_accession=121007849, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  16160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16160, CurrentID=0, Status='live', Symbol='Il12b', Aliases='Il-12b, Il-12p40, Il12p40, p40', description='interleukin 12b', other_designations='CLMF p40|IL-12 p40|IL-12 subunit p40|IL-23 subunit p40|cytotoxic lymphocyte maturation factor 40 kDa subunit', map_location='11 A5-B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=44400063, end_position_on_the_genomic_accession=44414677, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  96979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=96979, CurrentID=0, Status='live', Symbol='Ptges2', Aliases='0610038H10Rik, C79137, Gbf1, Mpges2, Pges2', description='prostaglandin E synthase 2', other_designations='GATE-binding factor 1|mPGES-2|microsomal prostaglandin E synthase 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32395890, end_position_on_the_genomic_accession=32402740, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  621269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621269, CurrentID=0, Status='live', Symbol='Actg-ps1', Aliases='EG621269, Gm1841, Gm6208', description='actin, gamma, pseudogene 1', other_designations='actin, beta pseudogene|beta-actin pseudogene', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46228985, end_position_on_the_genomic_accession=46230847, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  228061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228061, CurrentID=0, Status='live', Symbol='Agps', Aliases='5832437L22, 9930035G10Rik, AW123847, Adaps, Adas, Adhaps, Adps, Aldhpsy, bs2', description='alkylglycerone phosphate synthase', other_designations='alkyl-DHAP synthase|alkylglycerone-phosphate synthase', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=75832177, end_position_on_the_genomic_accession=75931350, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  228071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228071, CurrentID=0, Status='live', Symbol='Sestd1', Aliases='1500031J16Rik, A930010O20, Solo/Trio8', description='SEC14 and spectrin domains 1', other_designations='huntingtin-interacting protein-like protein', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=77180340, end_position_on_the_genomic_accession=77280592, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  228094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228094, CurrentID=0, Status='live', Symbol='Cerkl', Aliases='Gm1958, Rp26', description='ceramide kinase-like', other_designations='retinitis pigmentosa 26 (autosomal recessive)', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=79331105, end_position_on_the_genomic_accession=79430520, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  97031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97031, CurrentID=0, Status='live', Symbol='Tprn', Aliases='C430004E15Rik, C87750', description='taperin', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25262598, end_position_on_the_genomic_accession=25269886, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  16175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16175, CurrentID=0, Status='live', Symbol='Il1a', Aliases='Il-1a', description='interleukin 1 alpha', other_designations='IL-1 alpha', map_location='2 F', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129299609, end_position_on_the_genomic_accession=129310901, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  228136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228136, CurrentID=0, Status='live', Symbol='Zdhhc5', Aliases='1110032A17Rik, AI451382, Zisp', description='zinc finger, DHHC domain containing 5', other_designations='DHHC-5|probable palmitoyltransferase ZDHHC5|zinc finger DHHC domain-containing protein 5', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84687920, end_position_on_the_genomic_accession=84715164, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  228139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228139, CurrentID=0, Status='live', Symbol='P2rx3', Aliases='4930513E20Rik, P2X3', description='purinergic receptor P2X, ligand-gated ion channel, 3', other_designations='ATP receptor|purinergic receptor P2X3', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84996396, end_position_on_the_genomic_accession=85037029, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  228140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228140, CurrentID=0, Status='live', Symbol='Tnks1bp1', Aliases='Tab182, mKIAA1741', description='tankyrase 1 binding protein 1', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85048012, end_position_on_the_genomic_accession=85073048, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  228151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228151, CurrentID=0, Status='live', Symbol='4833423E24Rik', Aliases='Fads2p1, Gm113', description='RIKEN cDNA 4833423E24 gene', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85484092, end_position_on_the_genomic_accession=85518935, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  97086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97086, CurrentID=0, Status='live', Symbol='Slc9b2', Aliases='C80638, NHA2, NHE10, Nhedc2, nha-oc, nhaoc', description='solute carrier family 9, subfamily B (NHA2, cation proton antiporter 2), member 2', other_designations='NHE domain-containing protein 2|Na(+)/H(+) exchanger-like domain-containing protein 2|Na+/H+ exchanger domain containing 2|mitochondrial Na(+)/H(+) exchanger NHA2|mitochondrial sodium/hydrogen exchanger NHA2|sodium/hydrogen exchanger-like domain-containing protein 2|solute carrier family 9 subfamily B member 2', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135307618, end_position_on_the_genomic_accession=135342845, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  621377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621377, CurrentID=0, Status='live', Symbol='Aprt-ps', Aliases='', description='adenine phosphoribosyl transferase, pseudogene', other_designations='', map_location='8 34.43 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71223915, end_position_on_the_genomic_accession=71224450, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  97112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97112, CurrentID=0, Status='live', Symbol='Nmd3', Aliases='C87860', description='NMD3 homolog (S. cerevisiae)', other_designations='', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69722055, end_position_on_the_genomic_accession=69749047, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  97114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97114, CurrentID=0, Status='live', Symbol='Hist2h3c2', Aliases='H3-614, H3.2-614-ps, Hist2h3ca1, M32461, Hist2h3c2', description='histone cluster 2, H3c2', other_designations='histone 2, H3c2|histone 2, H3ca2|histone cluster 2, H3c2, pseudogene', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96238479, end_position_on_the_genomic_accession=96239127, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  97122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97122, CurrentID=0, Status='live', Symbol='Hist2h4', Aliases='H4, X04652', description='histone cluster 2, H4', other_designations='histone 2, H4', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96262934, end_position_on_the_genomic_accession=96263317, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  97130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97130, CurrentID=0, Status='live', Symbol='C77080', Aliases='Kiaa1522, mKIAA1522', description='expressed sequence C77080', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129219578, end_position_on_the_genomic_accession=129261404, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  228228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228228, CurrentID=0, Status='live', Symbol='Olfr1102', Aliases='MOR179-4', description='olfactory receptor 1102', other_designations='GA_x6K02T2Q125-48487992-48488966|olfactory receptor 179-4', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87001932, end_position_on_the_genomic_accession=87002994, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  97159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97159, CurrentID=0, Status='live', Symbol='A430005L14Rik', Aliases='C79672', description='RIKEN cDNA A430005L14 gene', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=153957237, end_position_on_the_genomic_accession=153961925, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  97165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97165, CurrentID=0, Status='live', Symbol='Hmgb2', Aliases='C80539, HMG-2, Hmg2', description='high mobility group box 2', other_designations='high mobility group protein 2', map_location='8 29.9 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=57511843, end_position_on_the_genomic_accession=57515999, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  97187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97187, CurrentID=0, Status='live', Symbol='C87977', Aliases='C85819', description='expressed sequence C87977', other_designations='novel oogenesin (Oog) protein', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144206762, end_position_on_the_genomic_accession=144213017, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  97212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97212, CurrentID=0, Status='live', Symbol='Hadha', Aliases='C77020, Mtpa', description='hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), alpha subunit', other_designations='TP-alpha', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30118304, end_position_on_the_genomic_accession=30154980, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  16204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16204, CurrentID=0, Status='live', Symbol='Fabp6', Aliases='GT, I-15P, I-BABP, ILBP, ILBP3, Illbp', description='fatty acid binding protein 6, ileal (gastrotropin)', other_designations='fatty acid-binding protein 6|ileal bile acid-binding protein|ileal lipid-binding protein', map_location='11 25.81 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43596040, end_position_on_the_genomic_accession=43601562, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  97243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97243, CurrentID=0, Status='live', Symbol='Naa11', Aliases='Ard1b, Ard2, C80008', description='N(alpha)-acetyltransferase 11, NatA catalytic subunit', other_designations='ARD1 homolog B|N-alpha-acetyltransferase 11, NatA catalytic subunit|N-terminal acetyltransferase complex ARD1 subunit homolog B|Nalpha acteyltransferase 11|arrest defective 2|natA catalytic subunit Naa11', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=97382209, end_position_on_the_genomic_accession=97392330, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  228355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228355, CurrentID=0, Status='live', Symbol='Madd', Aliases='9630059K23Rik, IG20', description='MAP-kinase activating death domain', other_designations='Rab3 GDP/GTP exchange protein|Rab3 GEP|insulinoma-glucagonoma protein 20|madd|rab3 GDP/GTP exchange factor', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91137360, end_position_on_the_genomic_accession=91183991, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  228356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228356, CurrentID=0, Status='live', Symbol='1110051M20Rik', Aliases='AI586322', description='RIKEN cDNA 1110051M20 gene', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91273240, end_position_on_the_genomic_accession=91444713, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  228357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228357, CurrentID=0, Status='live', Symbol='Lrp4', Aliases='6430526J12Rik, D230026E03, Megf7, mdig', description='low density lipoprotein receptor-related protein 4', other_designations='LDLR dan|LRP-4', map_location='2 50.63 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91457470, end_position_on_the_genomic_accession=91513901, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  97287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97287, CurrentID=0, Status='live', Symbol='Mtmr14', Aliases='1110061O04Rik, AW553738, C76151', description='myotubularin related protein 14', other_designations='mJumpy', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113237635, end_position_on_the_genomic_accession=113281392, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  228361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228361, CurrentID=0, Status='live', Symbol='Ambra1', Aliases='2310079H06Rik, A130023A14, AA474864, AV021921, D030051N19Rik, mKIAA1736', description='autophagy/beclin 1 regulator 1', other_designations='activating molecule in Beclin1-regulated autophagy', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91730111, end_position_on_the_genomic_accession=91918849, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  228366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228366, CurrentID=0, Status='live', Symbol='Gyltl1b', Aliases='5730485C17Rik, AI891893, Large2, Largel, mKIAA4105', description='glycosyltransferase-like 1B', other_designations='glycoslytransferase-like 1B', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92365046, end_position_on_the_genomic_accession=92371940, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  228368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228368, CurrentID=0, Status='live', Symbol='Slc35c1', Aliases='E430007K15Rik, fuct1', description='solute carrier family 35, member C1', other_designations='multispan transmembrane protein fuct1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92452764, end_position_on_the_genomic_accession=92460518, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  621603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621603, CurrentID=0, Status='live', Symbol='Aldh3b2', Aliases='AI848594, C130048D07Rik', description='aldehyde dehydrogenase 3 family, member B2', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3972328, end_position_on_the_genomic_accession=3981665, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  228410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228410, CurrentID=0, Status='live', Symbol='Cstf3', Aliases='4732468G05Rik, C79532, CstF-77', description="cleavage stimulation factor, 3' pre-RNA, subunit 3", other_designations='CF-1 77 kDa subunit|CSTF 77 kDa subunit|cleavage stimulation factor 77 kDa subunit', map_location='2 54.84 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104590484, end_position_on_the_genomic_accession=104665425, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  621628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621628, CurrentID=0, Status='live', Symbol='Cldn20', Aliases='EG621628', description='claudin 20', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=3532554, end_position_on_the_genomic_accession=3533213, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  228413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228413, CurrentID=0, Status='live', Symbol='Prrg4', Aliases='9930111I18Rik, TMG4', description='proline rich Gla (G-carboxyglutamic acid) 4 (transmembrane)', other_designations='proline-rich Gla protein 4|proline-rich gamma-carboxyglutamic acid protein 4', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104830741, end_position_on_the_genomic_accession=104849850, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  228421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228421, CurrentID=0, Status='live', Symbol='Kif18a', Aliases='AU024633, B130001M12Rik, gcd2', description='kinesin family member 18A', other_designations='N-8 kinesin|kinesin superfamily protein 18A', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=109280738, end_position_on_the_genomic_accession=109341749, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  228432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228432, CurrentID=0, Status='live', Symbol='Ano3', Aliases='AI838058, B230324K02Rik, Tmem16c', description='anoctamin 3', other_designations='ano3', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=110655201, end_position_on_the_genomic_accession=110950395, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  228443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228443, CurrentID=0, Status='live', Symbol='Olfr1283', Aliases='GA_x6K02T2Q125-72420217-72421134, MOR248-19', description='olfactory receptor 1283', other_designations='olfactory receptor GA_x6K02T2Q125-72420217-72421134', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111368634, end_position_on_the_genomic_accession=111369551, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  97387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97387, CurrentID=0, Status='live', Symbol='Strn4', Aliases='C80611, ZIN, zinedin', description='striatin, calmodulin binding protein 4', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16815889, end_position_on_the_genomic_accession=16840931, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  228482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228482, CurrentID=0, Status='live', Symbol='Arhgap11a', Aliases='6530401L14Rik, mKIAA0013', description='Rho GTPase activating protein 11A', other_designations='GAP (1-12)|rho-type GTPase-activating protein 11A', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=113831492, end_position_on_the_genomic_accession=113848661, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  228491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228491, CurrentID=0, Status='live', Symbol='Zfp770', Aliases='6430601A21Rik, Znf770', description='zinc finger protein 770', other_designations='', map_location='2 E4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=114193456, end_position_on_the_genomic_accession=114201432, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  97440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97440, CurrentID=0, Status='live', Symbol='B3gnt9', Aliases='3-Gn-T9-ps, BGnT-9, C76565, C76566, beta-1, beta3Gn-T9, B3gnt9', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 9', other_designations='BGnT-9|UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 9, pseudogene|beta-1,3-Gn-T9|beta-1,3-N-acetylglucosaminyltransferase 9|beta-1,3-galactosyltransferase-related protein|beta3Gn-T9', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105252638, end_position_on_the_genomic_accession=105255312, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  228536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228536, CurrentID=0, Status='live', Symbol='Bahd1', Aliases='AL022997, AW541238, Gm117, mKIAA0945', description='bromo adjacent homology domain containing 1', other_designations='BAH domain-containing protein 1', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118900332, end_position_on_the_genomic_accession=118924528, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  228543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228543, CurrentID=0, Status='live', Symbol='Rhov', Aliases='A030005A06Rik, Arhv', description='ras homolog gene family, member V', other_designations='rhov', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119269201, end_position_on_the_genomic_accession=119278008, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  228545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228545, CurrentID=0, Status='live', Symbol='Vps18', Aliases='9930024E13Rik', description='vacuolar protein sorting 18 (yeast)', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119288742, end_position_on_the_genomic_accession=119298453, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  97476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97476, CurrentID=0, Status='live', Symbol='Fam90a1a', Aliases='C86695', description='family with sequence similarity 90, member A1A', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21958714, end_position_on_the_genomic_accession=21964303, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  228550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228550, CurrentID=0, Status='live', Symbol='Itpka', Aliases='', description='inositol 1,4,5-trisphosphate 3-kinase A', other_designations='IP3 3-kinase A|IP3-kinase A|IP3K A|insP 3-kinase A', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119742239, end_position_on_the_genomic_accession=119751253, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  97484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97484, CurrentID=0, Status='live', Symbol='Cog8', Aliases='BB235941, C87832', description='component of oligomeric golgi complex 8', other_designations='COG complex subunit 8|conserved oligomeric Golgi complex protein 8', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107048709, end_position_on_the_genomic_accession=107056737, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  97487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97487, CurrentID=0, Status='live', Symbol='Cmtm4', Aliases='Cklfsf4, D19397, ENSMUSG00000051554, Gm9853', description='CKLF-like MARVEL transmembrane domain containing 4', other_designations='chemokine-like factor super family 4|chemokine-like factor superfamily member 4', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104348191, end_position_on_the_genomic_accession=104395807, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  228564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228564, CurrentID=0, Status='live', Symbol='Frmd5', Aliases='1500032A09Rik, 5830442F04, A930004K21Rik, R75430', description='FERM domain containing 5', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121545529, end_position_on_the_genomic_accession=121807757, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  228576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228576, CurrentID=0, Status='live', Symbol='Mall', Aliases='BC012256, BENE', description='mal, T cell differentiation protein-like', other_designations='mal, T-cell differentiation protein-like', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127704390, end_position_on_the_genomic_accession=127729897, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  228592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228592, CurrentID=0, Status='live', Symbol='F830045P16Rik', Aliases='Sirpb3', description='RIKEN cDNA F830045P16 gene', other_designations='novel immunoglobulin C1-set domain containing protein', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=129457797, end_position_on_the_genomic_accession=129536602, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  228598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228598, CurrentID=0, Status='live', Symbol='Ebf4', Aliases='Ebf3, O/E-4, Olf-1', description='early B cell factor 4', other_designations='EBF-4|OE-4|Olf-1/EBF-like 4|early B-cell factor 3|early B-cell factor 4', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130295249, end_position_on_the_genomic_accession=130370481, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  228602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228602, CurrentID=0, Status='live', Symbol='4930402H24Rik', Aliases='2810487F15Rik, AI413763, C78926', description='RIKEN cDNA 4930402H24 gene', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130706200, end_position_on_the_genomic_accession=130840807, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  621823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621823, CurrentID=0, Status='live', Symbol='Psme2b', Aliases='PA28b-p, PA28b2, Psme2-like-ps, Psme2b', description='protease (prosome, macropain) activator subunit 2B', other_designations='PA28 beta subunit (PA28b) pseudogene|protease (prosome, macropain) 28 subunit beta B, pseudogene|protease (prosome, macropain) 28 subunit, beta, b|proteasome activator complex like', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=48945345, end_position_on_the_genomic_accession=48946410, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  228608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228608, CurrentID=0, Status='live', Symbol='Smox', Aliases='B130066H01Rik, PAO, PAOh1, SMO', description='spermine oxidase', other_designations='PAO-1|polyamine oxidase 1', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131491782, end_position_on_the_genomic_accession=131525183, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  97541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97541, CurrentID=0, Status='live', Symbol='Qars', Aliases='1110018N24Rik, 1200016L19Rik, C80286, GLNRS', description='glutaminyl-tRNA synthetase', other_designations='glutamine-tRNA ligase', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108508005, end_position_on_the_genomic_accession=108515941, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  621852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621852, CurrentID=0, Status='live', Symbol='Rhox3f', Aliases='OTTMUSG00000017155, Rhox3.6, Rhox3e', description='reproductive homeobox 3F', other_designations='PEPP subfamily member|reproductive homeobox 3E', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37581352, end_position_on_the_genomic_accession=37585496, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  77446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=77446, CurrentID=0, Status='live', Symbol='Heg1', Aliases='4632417D23Rik, 5530401I02Rik, 9530025L16Rik, BC038614, Gm629, OTTMUSG00000027315', description='heart development protein with EGF-like domains 1', other_designations='HEG homolog 1', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33684430, end_position_on_the_genomic_accession=33771576, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  228662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228662, CurrentID=0, Status='live', Symbol='Btbd3', Aliases='mKIAA0952', description='BTB (POZ) domain containing 3', other_designations='', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=138256584, end_position_on_the_genomic_accession=138287424, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  621893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621893, CurrentID=0, Status='live', Symbol='Hist2h2ab', Aliases='EG621893, H2a-613a', description='histone cluster 2, H2ab', other_designations='histone 2, H2ab', map_location='3 F1-F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96219916, end_position_on_the_genomic_accession=96220353, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  228684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228684, CurrentID=0, Status='live', Symbol='Sel1l2', Aliases='Gm118', description='sel-1 suppressor of lin-12-like 2 (C. elegans)', other_designations='sel-1L2|suppressor of lin-12-like protein 2', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=140229599, end_position_on_the_genomic_accession=140389716, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  228714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228714, CurrentID=0, Status='live', Symbol='Csrp2bp', Aliases='2510008M08Rik, ATAC2, AU023459, D2Ertd473e, D2Wsu131e, E430020F17', description='cysteine and glycine-rich protein 2 binding protein', other_designations='ADA2A-containing complex subunit 2|CRP2-binding partner|CRP2BP|CSRP2-binding protein|cysteine-rich protein 2 binding protein', map_location='2 71.01 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144369032, end_position_on_the_genomic_accession=144407675, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  228715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228715, CurrentID=0, Status='live', Symbol='Gm561', Aliases='', description='predicted gene 561', other_designations='', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144594065, end_position_on_the_genomic_accession=144595365, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  228730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228730, CurrentID=0, Status='live', Symbol='Kiz', Aliases='Gm114, Ncrna00153, Plk1s1', description='kizuna centrosomal protein', other_designations='polo-like kinase 1 substrate 1', map_location='2 G2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=146855824, end_position_on_the_genomic_accession=146970098, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  228731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228731, CurrentID=0, Status='live', Symbol='Nkx2-4', Aliases='1700001P03Rik, Nkx-2.4, tinman', description='NK2 homeobox 4', other_designations='NK2 transcription factor related, locus 4|homeobox protein NK-2 homolog D', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=147083337, end_position_on_the_genomic_accession=147085439, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  621968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621968, CurrentID=0, Status='live', Symbol='Trbv1', Aliases='Gm6273, TCRBV2S1, Tcrb-V2', description='T cell receptor beta, variable 1', other_designations='T-cell receptor beta, variable 1|T-cell receptor beta, variable 2|TCR V beta chain', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40891296, end_position_on_the_genomic_accession=40891885, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  228756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228756, CurrentID=0, Status='live', Symbol='Cstl1', Aliases='RCET1, RCET2, RCET3', description='cystatin-like 1', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148750361, end_position_on_the_genomic_accession=148755378, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  621976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=621976, CurrentID=0, Status='live', Symbol='Tmem170b', Aliases='EG621976', description='transmembrane protein 170B', other_designations='tmem170b', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41606216, end_position_on_the_genomic_accession=41641357, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  228765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228765, CurrentID=0, Status='live', Symbol='Sdcbp2', Aliases='', description='syndecan binding protein (syntenin) 2', other_designations='syndecan-binding protein 2', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151572609, end_position_on_the_genomic_accession=151590005, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  228767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228767, CurrentID=0, Status='live', Symbol='Tmem74bos', Aliases='', description='transmembrane 74B, opposite strand', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151679361, end_position_on_the_genomic_accession=151687432, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  228769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228769, CurrentID=0, Status='live', Symbol='Psmf1', Aliases='AW048666, BC012260, PI31', description='proteasome (prosome, macropain) inhibitor subunit 1', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151716062, end_position_on_the_genomic_accession=151744188, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  228770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228770, CurrentID=0, Status='live', Symbol='Rspo4', Aliases='A730099F22, A930029K19Rik', description='R-spondin 4', other_designations='R-spondin family, member 4|R-spondin-like protein|cristin-4|cysteine-rich and single thrombospondin domain-containing protein 4|mCristin-4|roof plate-specific spondin-4', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151842579, end_position_on_the_genomic_accession=151874668, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  228775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228775, CurrentID=0, Status='live', Symbol='Trib3', Aliases='Ifld2, Nipk, SINK, SKIP3, TRB-3, Trb3', description='tribbles homolog 3 (Drosophila)', other_designations='induced in fatty liver dystrophy 2|neuronal cell death-inducible putative kinase', map_location='2 74.83 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152337425, end_position_on_the_genomic_accession=152344060, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  228777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228777, CurrentID=0, Status='live', Symbol='Nrsn2', Aliases='Gm123, Neurensin-2', description='neurensin 2', other_designations='nrsn2', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152368755, end_position_on_the_genomic_accession=152377153, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  228778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228778, CurrentID=0, Status='live', Symbol='6820408C15Rik', Aliases='', description='RIKEN cDNA 6820408C15 gene', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152415587, end_position_on_the_genomic_accession=152444336, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  228785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228785, CurrentID=0, Status='live', Symbol='Mylk2', Aliases='9830004H17Rik, BB138278', description='myosin, light polypeptide kinase 2, skeletal muscle', other_designations='MLCK2', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152911276, end_position_on_the_genomic_accession=152923065, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  228788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228788, CurrentID=0, Status='live', Symbol='Ccm2l', Aliases='', description='cerebral cavernous malformation 2-like', other_designations='CCM2-like', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153065933, end_position_on_the_genomic_accession=153081735, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  228790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228790, CurrentID=0, Status='live', Symbol='Asxl1', Aliases='mKIAA0978', description='additional sex combs like 1', other_designations='additional sex combs-like protein 1', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153346139, end_position_on_the_genomic_accession=153404007, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  228796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228796, CurrentID=0, Status='live', Symbol='Bpifb6', Aliases='Bpil3, Gm119, Lplunc6', description='BPI fold containing family B, member 6', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153900388, end_position_on_the_genomic_accession=153912795, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  228801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228801, CurrentID=0, Status='live', Symbol='Bpifb1', Aliases='Lplunc1', description='BPI fold containing family B, member 1', other_designations='long palate, lung and nasal epithelium carcinoma associated 1|long palate, lung and nasal epithelium carcinoma-associated protein 1|von Ebner minor salivary gland protein|von Ebner minor salivary protein', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154190818, end_position_on_the_genomic_accession=154220343, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  228802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228802, CurrentID=0, Status='live', Symbol='Bpifb5', Aliases='Lplunc5', description='BPI fold containing family B, member 5', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154223742, end_position_on_the_genomic_accession=154240902, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  228807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228807, CurrentID=0, Status='live', Symbol='Zfp341', Aliases='Znf341', description='zinc finger protein 341', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=154611797, end_position_on_the_genomic_accession=154646821, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  228812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228812, CurrentID=0, Status='live', Symbol='Pigu', Aliases='5430426F17Rik, BC028278, Cdc91l1', description='phosphatidylinositol glycan anchor biosynthesis, class U', other_designations='CDC91 cell division cycle 91-like 1|GPI transamidase component PIG-U|cell division cycle 91-like 1|cell division cycle protein 91-like 1|phosphatidylinositol-glycan biosynthesis, class U protein|protein CDC91-like 1', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155278243, end_position_on_the_genomic_accession=155357424, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  228829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228829, CurrentID=0, Status='live', Symbol='Phf20', Aliases='6820402O20Rik', description='PHD finger protein 20', other_designations='hepatocellular carcinoma-associated antigen 58 homolog', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156196647, end_position_on_the_genomic_accession=156309953, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  97761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97761, CurrentID=0, Status='live', Symbol='Sgsm2', Aliases='D630003G22Rik, R74628, Rutbc1, mKIAA0397', description='small G protein signaling modulator 2', other_designations='RUN and TBC1 domain containing 1|RUN and TBC1 domain-containing protein 1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74849262, end_position_on_the_genomic_accession=74897093, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  228836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228836, CurrentID=0, Status='live', Symbol='Dlgap4', Aliases='AI225853, BC024558, DAP-4, DAP4, SAPAP-4, Sapap4, WBP16', description='discs, large homolog-associated protein 4 (Drosophila)', other_designations='PSD-95/SAP90-binding protein 4|SAP90/PSD-95-associated protein 4|WW domain-binding protein 16', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156613633, end_position_on_the_genomic_accession=156764363, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  228839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228839, CurrentID=0, Status='live', Symbol='Tgif2', Aliases='4921501K24, 5730599O09Rik, C80753, C81206', description='TGFB-induced factor homeobox 2', other_designations="5'-TG-3'-interacting factor 2|TALE family homeobox|TGF-beta-induced transcription factor 2|TGFB-induced factor 2", map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156839817, end_position_on_the_genomic_accession=156855575, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  228846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228846, CurrentID=0, Status='live', Symbol='D630003M21Rik', Aliases='', description='RIKEN cDNA D630003M21 gene', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158182533, end_position_on_the_genomic_accession=158229289, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  228850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228850, CurrentID=0, Status='live', Symbol='Ralgapb', Aliases='9330195P16, AI507211, B230339M05Rik, mKIAA1219, p170', description='Ral GTPase activating protein, beta subunit (non-catalytic)', other_designations='ral GTPase-activating protein beta subunit', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158409829, end_position_on_the_genomic_accession=158499253, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  228852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228852, CurrentID=0, Status='live', Symbol='Ppp1r16b', Aliases='ANKRD4, C130078N17Rik, TIMAP, Wdt4', description='protein phosphatase 1, regulatory (inhibitor) subunit 16B', other_designations='CAAX box protein TIMAP|TGF-beta-inhibited membrane-associated protein|whn-dependent transcript 4', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158666733, end_position_on_the_genomic_accession=158766334, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  228858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228858, CurrentID=0, Status='live', Symbol='Gdap1l1', Aliases='Gdap1l', description='ganglioside-induced differentiation-associated protein 1-like 1', other_designations='GDAP1-L1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163438419, end_position_on_the_genomic_accession=163455324, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  228859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228859, CurrentID=0, Status='live', Symbol='Fitm2', Aliases='AI414861, D930001I22Rik, Fit2', description='fat storage-inducing transmembrane protein 2', other_designations='fat-inducing protein 2|fat-inducing transcript 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163468703, end_position_on_the_genomic_accession=163472629, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  228866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228866, CurrentID=0, Status='live', Symbol='Pcif1', Aliases='2310022K11Rik, C20orf67, F730014I05Rik', description='PDX1 C-terminal inhibiting factor 1', other_designations='phosphorylated CTD interacting factor PCIF1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164879368, end_position_on_the_genomic_accession=164891437, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  228869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228869, CurrentID=0, Status='live', Symbol='Ncoa5', Aliases='CIA', description='nuclear receptor coactivator 5', other_designations='NCoA-5|coactivator independent of AF-2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165000357, end_position_on_the_genomic_accession=165034868, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  228875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228875, CurrentID=0, Status='live', Symbol='Slc35c2', Aliases='C85957, CGI-15, D2Wsu58e, Ovcov1', description='solute carrier family 35, member C2', other_designations='ovarian cancer overexpressed gene 1 protein', map_location='2 85.53 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165276522, end_position_on_the_genomic_accession=165287888, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  228876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228876, CurrentID=0, Status='live', Symbol='Zfp334', Aliases='D2Ertd535e', description='zinc finger protein 334', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165377434, end_position_on_the_genomic_accession=165388317, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  228880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228880, CurrentID=0, Status='live', Symbol='Zmynd8', Aliases='1110013E22Rik, 2010005I16Rik, 3632413B07Rik, AI316811, AL024039, Prkcbp1, RACK7', description='zinc finger, MYND-type containing 8', other_designations='protein kinase C binding protein 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165784152, end_position_on_the_genomic_accession=165898984, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  228889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228889, CurrentID=0, Status='live', Symbol='Ddx27', Aliases='C86129', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 27', other_designations='DEAD box protein 27|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 27|RNA-helicase like protein', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167015255, end_position_on_the_genomic_accession=167034950, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  97820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97820, CurrentID=0, Status='live', Symbol='4833439L19Rik', Aliases='4930558H15Rik, C81457, Kiaa1191, P33monox', description='RIKEN cDNA 4833439L19 gene', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54551218, end_position_on_the_genomic_accession=54565382, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  97827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97827, CurrentID=0, Status='live', Symbol='Exd2', Aliases='4930539P14Rik, C85658, Exdl2', description="exonuclease 3'-5' domain containing 2", other_designations="exonuclease 3''-5'' domain-like 2|exonuclease 3'-5' domain-like-containing protein 2", map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80463095, end_position_on_the_genomic_accession=80498135, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  622127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622127, CurrentID=0, Status='live', Symbol='Cyp3a57', Aliases='EG622127', description='cytochrome P450, family 3, subfamily a, polypeptide 57', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145345145, end_position_on_the_genomic_accession=145390926, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  228913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228913, CurrentID=0, Status='live', Symbol='Zfp217', Aliases='4933431C08Rik, AW987152, Gm562, ZABC1, Znf217', description='zinc finger protein 217', other_designations='', map_location='2 89.91 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=170108643, end_position_on_the_genomic_accession=170151677, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  97848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97848, CurrentID=0, Status='live', Symbol='Serpinb6c', Aliases='C76171, C76174, SPI3C, SPIC, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 6c', other_designations='serine (or cysteine) proteinase inhibitor, clade B, member 6c', map_location='13 13.99 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33879816, end_position_on_the_genomic_accession=33905708, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  97863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97863, CurrentID=0, Status='live', Symbol='Fam8a1', Aliases='C78339', description='family with sequence similarity 8, member A1', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=46669443, end_position_on_the_genomic_accession=46678064, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  360013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360013, CurrentID=0, Status='live', Symbol='Myo18a', Aliases='MAJN, MyoPDZ, MysPDZ', description='myosin XVIIIA', other_designations='molecule associated with JAK3 N-terminus|myosin XVIIIb|myosin containing a PDZ domain', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77763251, end_position_on_the_genomic_accession=77865988, orientation='plus', exon_count=50, OMIM='', no_hdr0=''),
  228942 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228942, CurrentID=0, Status='live', Symbol='Cbln4', Aliases='AI848962', description='cerebellin 4 precursor protein', other_designations='cerebellin-like glycoprotein 1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172036336, end_position_on_the_genomic_accession=172043466, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  97884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97884, CurrentID=0, Status='live', Symbol='B3galnt2', Aliases='A930105D20Rik, C80633, D230016N13Rik', description='UDP-GalNAc:betaGlcNAc beta 1,3-galactosaminyltransferase, polypeptide 2', other_designations='beta-1,3-GalNAc-T2|beta-1,3-N-acetylgalactosaminyltransferase II', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=13954474, end_position_on_the_genomic_accession=13999098, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  228960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228960, CurrentID=0, Status='live', Symbol='Stx16', Aliases='4930401D03, 5430410K23Rik, 6330500A18Rik, AI648908, AW553605, Syn16', description='syntaxin 16', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174076262, end_position_on_the_genomic_accession=174099771, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  228961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228961, CurrentID=0, Status='live', Symbol='Npepl1', Aliases='BC023239, C85514, mKIAA1974', description='aminopeptidase-like 1', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=174110335, end_position_on_the_genomic_accession=174122702, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  228966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228966, CurrentID=0, Status='live', Symbol='Ppp1r3d', Aliases='2610024M03Rik, Gm359', description='protein phosphatase 1, regulatory subunit 3D', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=178411206, end_position_on_the_genomic_accession=178414463, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  97895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97895, CurrentID=0, Status='live', Symbol='Nlrp4f', Aliases='C330026N02Rik, C86045, Nalp-kappa, Nalp4f', description='NLR family, pyrin domain containing 4F', other_designations='NACHT, leucine rich repeat and PYD containing 4F|NALP, kappa', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=65177111, end_position_on_the_genomic_accession=65205716, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  228980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228980, CurrentID=0, Status='live', Symbol='Taf4', Aliases='135kDa, AI450312, TAFII130, TAFII135, Taf2c1a, Taf4', description='TATA-box binding protein associated factor 4', other_designations='TAF4A RNA polymerase II, TATA box binding protein (TBP)-associated factor|TATA box binding protein (TBP)-associated factor, RNA polymerase II, C1, 130kd', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=179912146, end_position_on_the_genomic_accession=179976646, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  228983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228983, CurrentID=0, Status='live', Symbol='Osbpl2', Aliases='C130070J12Rik, ORP-2, Orp2, mKIAA0772', description='oxysterol binding protein-like 2', other_designations='OSBP-related protein 2|oxysterol-binding protein-like 2', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180119366, end_position_on_the_genomic_accession=180162680, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  228993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228993, CurrentID=0, Status='live', Symbol='Slc17a9', Aliases='1700019H03Rik, BC019537, Vnut', description='solute carrier family 17, member 9', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180725213, end_position_on_the_genomic_accession=180742280, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  228994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228994, CurrentID=0, Status='live', Symbol='Ythdf1', Aliases='2210410K23Rik, 8030473O16', description='YTH domain family 1', other_designations='DACA-1 homolog|YTH domain family protein 1|dermatomyositis associated with cancer putative autoantigen 1 homolog', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180904377, end_position_on_the_genomic_accession=180920936, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  228998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=228998, CurrentID=0, Status='live', Symbol='Arfgap1', Aliases='AI115377, Arf1gap', description='ADP-ribosylation factor GTPase activating protein 1', other_designations='ADP-ribosylation factor 1 GTPase activating protein|ADP-ribosylation factor 1 GTPase-activating protein|ARF GAP 1|ARF1 GAP|ARF1-directed GTPase-activating protein', map_location='2 103.53 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180967225, end_position_on_the_genomic_accession=180982524, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  229003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229003, CurrentID=0, Status='live', Symbol='Helz2', Aliases='PDIP1, Pric285, mKIAA1769, mPDIP1', description='helicase with zinc finger 2, transcriptional coactivator', other_designations='PPAR gamma-DNA-binding domain interacting protein1|PPAR-gamma DBD-interacting protein 1|PPAR-gamma DNA-binding domain-interacting protein 1|peroxisomal proliferator-activated receptor A interacting complex 285', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181227615, end_position_on_the_genomic_accession=181242058, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  229004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229004, CurrentID=0, Status='live', Symbol='Gmeb2', Aliases='AI839884', description='glucocorticoid modulatory element binding protein 2', other_designations='GMEB-2', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181251449, end_position_on_the_genomic_accession=181287974, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  229007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229007, CurrentID=0, Status='live', Symbol='Zgpat', Aliases='1500006I01Rik, BC021513', description='zinc finger, CCCH-type with G patch domain', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181364931, end_position_on_the_genomic_accession=181380793, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  229011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229011, CurrentID=0, Status='live', Symbol='Samd10', Aliases='D930033N23Rik, E130318E08', description='sterile alpha motif domain containing 10', other_designations='SAM domain-containing protein 10|samd10', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181595218, end_position_on_the_genomic_accession=181599491, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  97961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97961, CurrentID=0, Status='live', Symbol='Nol12', Aliases='C78541, Nop25', description='nucleolar protein 12', other_designations='25 kDa nucleolar protein|nucleolar protein of 25 kDa', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78934933, end_position_on_the_genomic_accession=78941910, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  102465212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102465212, CurrentID=0, Status='live', Symbol='Mir21c', Aliases='Gm28013, mmu-mir-21c', description='microRNA 21c', other_designations='microRNA mir-21c', map_location='', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=128278148, end_position_on_the_genomic_accession=128278225, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  229055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229055, CurrentID=0, Status='live', Symbol='Zbtb10', Aliases='4832414A18, AU021107, RINZF', description='zinc finger and BTB domain containing 10', other_designations='', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=9250567, end_position_on_the_genomic_accession=9285333, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  97998 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=97998, CurrentID=0, Status='live', Symbol='Deptor', Aliases='Depdc6', description='DEP domain containing MTOR-interacting protein', other_designations='DEP domain-containing protein 6', map_location='15 21.96 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=55112317, end_position_on_the_genomic_accession=55259273, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  622301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622301, CurrentID=0, Status='live', Symbol='Rhox2h', Aliases='OTTMUSG00000017156, Rhox2.8, Rhox2g', description='reproductive homeobox 2H', other_designations='reproductive homeobox 2G', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37668997, end_position_on_the_genomic_accession=37673277, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  622307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622307, CurrentID=0, Status='live', Symbol='5830473C10Rik', Aliases='Arg, Gm17754', description='RIKEN cDNA 5830473C10 gene', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90560632, end_position_on_the_genomic_accession=90598016, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  229096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229096, CurrentID=0, Status='live', Symbol='Ythdf3', Aliases='9130022A11Rik', description='YTH domain family 3', other_designations='YTH domain family protein 3', map_location='3 A1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=16183005, end_position_on_the_genomic_accession=16217037, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  622320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622320, CurrentID=0, Status='live', Symbol='Kctd21', Aliases='EG622320', description='potassium channel tetramerisation domain containing 21', other_designations='K+ channel tetramerisation/BTB/POZ/BTB domain profile containing protein', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97332311, end_position_on_the_genomic_accession=97350216, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  98053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98053, CurrentID=0, Status='live', Symbol='Gtf2f1', Aliases='2810405L04Rik, C76800, RAP74', description='general transcription factor IIF, polypeptide 1', other_designations='TFIIF-alpha|general transcription factor IIF polypeptide 1 (74kD subunit)|transcription initiation factor IIF subunit alpha', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57003401, end_position_on_the_genomic_accession=57011393, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  360198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360198, CurrentID=0, Status='live', Symbol='Hist1h3a', Aliases='', description='histone cluster 1, H3a', other_designations='histone 1, H3a', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23761885, end_position_on_the_genomic_accession=23762386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  103724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103724, CurrentID=0, Status='live', Symbol='Tbc1d10a', Aliases='AI447804, EPI64, Tbc1d10, mFLJ00288', description='TBC1 domain family, member 10a', other_designations='EBP50-PDX interactor of 64 kDa', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4186833, end_position_on_the_genomic_accession=4215507, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  360213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360213, CurrentID=0, Status='live', Symbol='Trim46', Aliases='TRIFIC', description='tripartite motif-containing 46', other_designations='geneY|tripartite, fibronectin type III and C-terminal SPRY motif-containing', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89234169, end_position_on_the_genomic_accession=89246154, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  360214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360214, CurrentID=0, Status='live', Symbol='Defb39', Aliases='BD-39', description='defensin beta 39', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19052826, end_position_on_the_genomic_accession=19064810, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  360216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360216, CurrentID=0, Status='live', Symbol='Zranb1', Aliases='9330160G10Rik, D7Wsu87e, Gm1956, Trabid', description='zinc finger, RAN-binding domain containing 1', other_designations='TRAF-binding protein|zinc finger Ran-binding domain-containing protein 1', map_location='7 76.32 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132931147, end_position_on_the_genomic_accession=132986391, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  360217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360217, CurrentID=0, Status='live', Symbol='Defb40', Aliases='BD-40', description='defensin beta 40', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=18974940, end_position_on_the_genomic_accession=18978116, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  360220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360220, CurrentID=0, Status='live', Symbol='Speer4d', Aliases='4933431D05Rik, SPEER-4D', description='spermatogenesis associated glutamate (E)-rich protein 4D', other_designations='sperm-associated glutamate (E)-rich protein 4d|spermatogenic-associated glutamate (E)-rich protein 4d', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=15619099, end_position_on_the_genomic_accession=15623864, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  360223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=360223, CurrentID=0, Status='live', Symbol='jams1', Aliases='', description='juvenile audiogenic monogenic seizures 1', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  622402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622402, CurrentID=0, Status='live', Symbol='Akr1c12', Aliases='AI266976, Akra', description='aldo-keto reductase family 1, member C12', other_designations='aldo-keto reductase a', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4268172, end_position_on_the_genomic_accession=4279399, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  622404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622404, CurrentID=0, Status='live', Symbol='Ccdc107', Aliases='', description='coiled-coil domain containing 107', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43493340, end_position_on_the_genomic_accession=43495921, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  622408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622408, CurrentID=0, Status='live', Symbol='Mcidas', Aliases='EG622408, Gm6320, Idas, Mci, Mcin', description='multiciliate differentiation and DNA synthesis associated cell cycle protein', other_designations='multiciliate cell differentiation|multiciliate differentiation and DNA synthesis-associated cell cycle protein|protein Idas', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=112993472, end_position_on_the_genomic_accession=113000394, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  229211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229211, CurrentID=0, Status='live', Symbol='Acad9', Aliases='2600017P15Rik, 4732402K02, C630012L17Rik, NPD002', description='acyl-Coenzyme A dehydrogenase family, member 9', other_designations='ACAD-9|very-long-chain acyl-CoA dehydrogenase VLCAD homolog', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36065984, end_position_on_the_genomic_accession=36092857, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  229214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229214, CurrentID=0, Status='live', Symbol='Qrfpr', Aliases='AQ27, Gpr103, SP9155', description='pyroglutamylated RFamide peptide receptor', other_designations='G protein-coupled receptor 103|G-protein coupled receptor 103|QRFP receptor|orexigenic neuropeptide QRFP receptor', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36179426, end_position_on_the_genomic_accession=36222275, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  622434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622434, CurrentID=0, Status='live', Symbol='Arhgef26', Aliases='4631416L12Rik, 8430436L14Rik, AI303572, AW491953, SGEF', description='Rho guanine nucleotide exchange factor (GEF) 26', other_designations='Src homology 3 domain-containing guanine nucleotide exchange factor', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=62338777, end_position_on_the_genomic_accession=62462221, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  229227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229227, CurrentID=0, Status='live', Symbol='4932438A13Rik', Aliases='4732443H21, B830039D19Rik, D630029K19Rik, FSA, Kiaa1109, Tweek', description='RIKEN cDNA 4932438A13 gene', other_designations='fragile site-associated protein homolog', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36863106, end_position_on_the_genomic_accession=37053033, orientation='plus', exon_count=88, OMIM='', no_hdr0=''),
  229228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229228, CurrentID=0, Status='live', Symbol='Nudt6', Aliases='Asfgf2b', description='nudix (nucleoside diphosphate linked moiety X)-type motif 6', other_designations='antisense basic fibroblast growth factor B|nudix motif 6', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37404982, end_position_on_the_genomic_accession=37420186, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  234814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234814, CurrentID=0, Status='live', Symbol='Mthfsd', Aliases='AW049977, BC052066, C920004D05', description='methenyltetrahydrofolate synthetase domain containing', other_designations='methenyltetrahydrofolate synthetase domain-containing protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121097555, end_position_on_the_genomic_accession=121108409, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  98170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98170, CurrentID=0, Status='live', Symbol='Tmem132a', Aliases='6720481D13Rik, Hspa5bp1, Orai1, R74613', description='transmembrane protein 132A', other_designations='HSPA5-binding protein 1|heat shock 70kDa protein 5 binding protein 1|heat shock protein 5 binding protein 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10857829, end_position_on_the_genomic_accession=10869935, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100417514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100417514, CurrentID=0, Status='live', Symbol='Adh6b', Aliases='Adh5b', description='alcohol dehydrogenase 6B (class V)', other_designations='alcohol dehydrogenase 6A (class V) pseudogene', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138341185, end_position_on_the_genomic_accession=138359308, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  278507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278507, CurrentID=0, Status='live', Symbol='Wfikkn2', Aliases='2610304F08Rik, AY100450, Gasp1, WFIKKNRP', description='WAP, follistatin/kazal, immunoglobulin, kunitz and netrin domain containing 2', other_designations='GASP-1|growth and differentiation factor-associated serum protein 1|mGASP-1', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94235951, end_position_on_the_genomic_accession=94247793, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  622473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622473, CurrentID=0, Status='live', Symbol='Ripply1', Aliases='', description='ripply1 homolog (zebrafish)', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139779681, end_position_on_the_genomic_accession=139782353, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  622474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622474, CurrentID=0, Status='live', Symbol='Smok3b', Aliases='4931427P12Rik, Smok2b2', description='sperm motility kinase 3B', other_designations='sperm motility kinase 2B', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138037224, end_position_on_the_genomic_accession=138050633, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  98193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98193, CurrentID=0, Status='live', Symbol='Dcaf8', Aliases='AA408877, AA433120, AA792214, AU018973, C76055, D1Dau35e, D1Ucla4, H326, Wdr42a', description='DDB1 and CUL4 associated factor 8', other_designations='WD repeat domain 42A|WD repeat-containing protein 42A', map_location='1 79.54 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172148015, end_position_on_the_genomic_accession=172196393, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  229277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229277, CurrentID=0, Status='live', Symbol='Stoml3', Aliases='Epb7.2l, SLP3, sro', description='stomatin (Epb7.2)-like 3', other_designations='SLP-3|stomatin related olfactory protein SRO|stomatin-related olfactory protein', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=53488793, end_position_on_the_genomic_accession=53507652, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  229279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229279, CurrentID=0, Status='live', Symbol='Hnrnpa3', Aliases='2410013L13Rik, 2610209F03Rik, 2610510D13Rik, Hnrpa3', description='heterogeneous nuclear ribonucleoprotein A3', other_designations='hnRNP A3', map_location='2 A7.2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=75659230, end_position_on_the_genomic_accession=75669407, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  229285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229285, CurrentID=0, Status='live', Symbol='Spg20', Aliases='AI840044, C79168, SPARTIN, TAHCCP1, mKIAA0610', description='spastic paraplegia 20, spartin (Troyer syndrome) homolog (human)', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=55112074, end_position_on_the_genomic_accession=55137332, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  98221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98221, CurrentID=0, Status='live', Symbol='Eif3m', Aliases='Ga17, Pcid1, Tango7', description='eukaryotic translation initiation factor 3, subunit M', other_designations='PCI domain containing 1 (herpesvirus entry mediator)|PCI domain-containing protein 1|dendritic cell protein GA17|eIF3m', map_location='2 55.06 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104999656, end_position_on_the_genomic_accession=105017027, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  229302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229302, CurrentID=0, Status='live', Symbol='Tm4sf4', Aliases='Iltmp', description='transmembrane 4 superfamily member 4', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=57425410, end_position_on_the_genomic_accession=57441677, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  98238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98238, CurrentID=0, Status='live', Symbol='Lrrc59', Aliases='AA959742, C78668', description='leucine rich repeat containing 59', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94629824, end_position_on_the_genomic_accession=94645216, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  229317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229317, CurrentID=0, Status='live', Symbol='Eif2a', Aliases='D030048D22, D3Ertd194e', description='eukaryotic translation initiation factor 2A', other_designations='eIF-2A', map_location='3 28.58 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=58525821, end_position_on_the_genomic_accession=58557501, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  229320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229320, CurrentID=0, Status='live', Symbol='Clrn1', Aliases='A130002D11Rik, USH3, Ush3a', description='clarin 1', other_designations='Usher syndrome 3A homolog|usher syndrome type-3 protein homolog', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=58844028, end_position_on_the_genomic_accession=58885212, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  229323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229323, CurrentID=0, Status='live', Symbol='Gpr171', Aliases='A630006E05, F730001G15Rik, H963', description='G protein-coupled receptor 171', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=59096448, end_position_on_the_genomic_accession=59101821, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  98256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98256, CurrentID=0, Status='live', Symbol='Kmo', Aliases='AI046660', description='kynurenine 3-monooxygenase (kynurenine 3-hydroxylase)', other_designations='kmo|kynurenine 3-hydroxylase', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=175632193, end_position_on_the_genomic_accession=175660853, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  98258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98258, CurrentID=0, Status='live', Symbol='Txndc9', Aliases='AI098020, APACD', description='thioredoxin domain containing 9', other_designations='ATP binding protein associated with cell differentiation|ATP-binding protein associated with cell differentiation', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37983867, end_position_on_the_genomic_accession=37997725, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  622554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622554, CurrentID=0, Status='live', Symbol='1700123I01Rik', Aliases='MAJIN', description='RIKEN cDNA 1700123I01 gene', other_designations='uncharacterized protein C11orf85 homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6175080, end_position_on_the_genomic_accession=6225885, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  98267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98267, CurrentID=0, Status='live', Symbol='Stk17b', Aliases='3110009A03Rik, AI120141, Drak2', description='serine/threonine kinase 17b (apoptosis-inducing)', other_designations='DAP kinase-related apoptosis-inducing protein kinase 2', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53755512, end_position_on_the_genomic_accession=53785250, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  229357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229357, CurrentID=0, Status='live', Symbol='Gpr149', Aliases='9630018L10Rik, Ieda, PGR10, R35', description='G protein-coupled receptor 149', other_designations='G protein-coupled receptor PGR10|G-protein coupled receptor PGR10', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=62529963, end_position_on_the_genomic_accession=62605442, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  229363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229363, CurrentID=0, Status='live', Symbol='Gmps', Aliases='AA591640, AI047208', description='guanine monophosphate synthetase', other_designations='GMP synthase|GMP synthetase|glutamine amidotransferase', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=63964819, end_position_on_the_genomic_accession=64021490, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  98303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98303, CurrentID=0, Status='live', Symbol='D630023F18Rik', Aliases='AI314969', description='RIKEN cDNA D630023F18 gene', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=65105284, end_position_on_the_genomic_accession=65123351, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  98314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98314, CurrentID=0, Status='live', Symbol='D2hgdh', Aliases='AA408776, AA408778, AI325464', description='D-2-hydroxyglutarate dehydrogenase', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93824808, end_position_on_the_genomic_accession=93852358, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  229389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229389, CurrentID=0, Status='live', Symbol='Otol1', Aliases='Gm414', description='otolin 1 homolog (zebrafish)', other_designations='', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=70007613, end_position_on_the_genomic_accession=70028708, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100499513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100499513, CurrentID=0, Status='live', Symbol='Mir3099', Aliases='mir-3099, mmu-mir-3099', description='microRNA 3099', other_designations='microRNA mir-3099', map_location='7 3.95 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6803587, end_position_on_the_genomic_accession=6803670, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  98363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98363, CurrentID=0, Status='live', Symbol='Efhd1', Aliases='4931430I01Rik, AI452351, PP3051, mitocalcin', description='EF hand domain containing 1', other_designations='EF-hand domain-containing protein 1|swiprosin-2', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=87264308, end_position_on_the_genomic_accession=87310839, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  98365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98365, CurrentID=0, Status='live', Symbol='Slamf9', Aliases='2310026I04Rik, AI462096, CD2F-10, CD84-H1, Cd2f10, SF2001', description='SLAM family member 9', other_designations='CD2 antigen family, member 10', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172458348, end_position_on_the_genomic_accession=172478421, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  98366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98366, CurrentID=0, Status='live', Symbol='Smap1', Aliases='1700056O10Rik, 4921514B13Rik, 4921525H11Rik, AI462175', description='small ArfGAP 1', other_designations='', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=23844825, end_position_on_the_genomic_accession=23922450, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  229445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229445, CurrentID=0, Status='live', Symbol='Ctso', Aliases='A330105D01Rik, AI118514', description='cathepsin O', other_designations='', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=81932616, end_position_on_the_genomic_accession=81956725, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  98376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98376, CurrentID=0, Status='live', Symbol='Gorab', Aliases='NTKL-BP1, SCYL1-BP1, Scyl1bp1', description='golgin, RAB6-interacting', other_designations='N-terminal kinase-like-binding protein 1|NTKL-binding protein 1|SCY1-like 1 binding protein 1|SCYL1-binding protein 1|dynactin 1, isoform 1; dynactin 1 (p150, Glued homolog); p150, Glued homolog; 150 kDa dynein-associated polypeptide; p150-glued', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=163384906, end_position_on_the_genomic_accession=163403908, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  622665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622665, CurrentID=0, Status='live', Symbol='Ccdc17', Aliases='', description='coiled-coil domain containing 17', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116596730, end_position_on_the_genomic_accession=116600266, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  98386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98386, CurrentID=0, Status='live', Symbol='Lbr', Aliases='AI505894, ic', description='lamin B receptor', other_designations='ichthyosis|integral nuclear envelope inner membrane protein', map_location='1 84.89 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181815315, end_position_on_the_genomic_accession=181842401, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  622675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622675, CurrentID=0, Status='live', Symbol='Zfp827', Aliases='2810449M09Rik, D630040G17Rik, Znf827', description='zinc finger protein 827', other_designations='', map_location='8 C1-C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79027983, end_position_on_the_genomic_accession=79193768, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  98388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98388, CurrentID=0, Status='live', Symbol='Chst10', Aliases='AI507003, AU041319, Hnk-1st, ST', description='carbohydrate sulfotransferase 10', other_designations='HNK-1 sulfotransferase|HNK1ST', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=38863867, end_position_on_the_genomic_accession=38898262, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  98396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98396, CurrentID=0, Status='live', Symbol='Slc41a1', Aliases='AI573938, B230315F01Rik', description='solute carrier family 41, member 1', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131828012, end_position_on_the_genomic_accession=131848863, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  229473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229473, CurrentID=0, Status='live', Symbol='D930015E06Rik', Aliases='Kiaa0922, mKIAA0922', description='RIKEN cDNA D930015E06 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=83898205, end_position_on_the_genomic_accession=84040182, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  98402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98402, CurrentID=0, Status='live', Symbol='Sh3bp4', Aliases='AI594717, AW227605, BOG25', description='SH3-domain binding protein 4', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=89070462, end_position_on_the_genomic_accession=89155084, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  98403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98403, CurrentID=0, Status='live', Symbol='Zfp451', Aliases='4930515K21Rik, 4933435G09Rik, AI596398, COASTER, Kiaa0576-hp, Znf451, mKIAA1702', description='zinc finger protein 451', other_designations='', map_location='1 12.81 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=33761541, end_position_on_the_genomic_accession=33815452, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  98404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98404, CurrentID=0, Status='live', Symbol='AI597479', Aliases='', description='expressed sequence AI597479', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=43098710, end_position_on_the_genomic_accession=43115946, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  98415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98415, CurrentID=0, Status='live', Symbol='Nucks1', Aliases='2700010L10Rik, 8430423A01Rik, AI647518, C78391, Nucks', description='nuclear casein kinase and cyclin-dependent kinase substrate 1', other_designations='JC7|nuclear ubiquitous casein and cyclin-dependent kinases substrate|nuclear, casein kinase and cyclin-dependent kinase substrate', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131910458, end_position_on_the_genomic_accession=131936321, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  229488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229488, CurrentID=0, Status='live', Symbol='Fam160a1', Aliases='9930021J17Rik', description='family with sequence similarity 160, member A1', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=85660063, end_position_on_the_genomic_accession=85845419, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  98417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98417, CurrentID=0, Status='live', Symbol='Cnih4', Aliases='AI647760, AW050376, CNIH-4, D530030D03Rik, E430023H19Rik, HSPC163', description='cornichon family AMPA receptor auxiliary protein 4', other_designations='cornichon homolog 4', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181150336, end_position_on_the_genomic_accession=181168994, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  229499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229499, CurrentID=0, Status='live', Symbol='Fcrl1', Aliases='A230020G22Rik, Bxmas1, Bxmh1, Bxmh1b, Fcrh1, Fcrh1l, Fcrh1s, Ifgp1', description='Fc receptor-like 1', other_designations='BXMAS1-like protein 1|BXMAS1-like protein 1b|IFGP family protein 1|fc receptor homolog 1|fcR-like protein 1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87376333, end_position_on_the_genomic_accession=87402934, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  229503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229503, CurrentID=0, Status='live', Symbol='Rrnad1', Aliases='', description='ribosomal RNA adenine dimethylase domain containing 1', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=87922601, end_position_on_the_genomic_accession=87930588, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  98432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98432, CurrentID=0, Status='live', Symbol='Phlpp1', Aliases='AI836256, Phlpp, Plekhe1, SCOP, mKIAA0606', description='PH domain and leucine rich repeat protein phosphatase 1', other_designations='PH domain-containing family E member 1|pleckstrin homology domain containing, family E (with leucine rich repeats) member 1|pleckstrin homology domain-containing family E member 1|suprachiasmatic nucleus circadian oscillatory protein', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=106171869, end_position_on_the_genomic_accession=106394245, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  229512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229512, CurrentID=0, Status='live', Symbol='Smg5', Aliases='BC024683, mKIAA1089', description='Smg-5 homolog, nonsense mediated mRNA decay factor (C. elegans)', other_designations='EST1-like protein B', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88336260, end_position_on_the_genomic_accession=88362338, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  229517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229517, CurrentID=0, Status='live', Symbol='Slc25a44', Aliases='6720482A19, B430110G05Rik', description='solute carrier family 25, member 44', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88410494, end_position_on_the_genomic_accession=88425141, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  229521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229521, CurrentID=0, Status='live', Symbol='Syt11', Aliases='1500004A13Rik, 3632445O20Rik, 5430404N14Rik, 6530420C11Rik, AI851970', description='synaptotagmin XI', other_designations='synaptotagmin 11|sytXI', map_location='3 38.96 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88744699, end_position_on_the_genomic_accession=88775159, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  229524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229524, CurrentID=0, Status='live', Symbol='Msto1', Aliases='BC008103, mst', description='misato homolog 1 (Drosophila)', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88909616, end_position_on_the_genomic_accession=88913950, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  229534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229534, CurrentID=0, Status='live', Symbol='Pbxip1', Aliases='4732463H20Rik', description='pre B cell leukemia transcription factor interacting protein 1', other_designations='hematopoietic PBX-interacting protein|pre-B-cell leukemia transcription factor interacting protein 1', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89436675, end_position_on_the_genomic_accession=89450952, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  229541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229541, CurrentID=0, Status='live', Symbol='Dennd4b', Aliases='', description='DENN/MADD domain containing 4B', other_designations='brain specific protein 4|brain-specific gene 4 protein', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90264920, end_position_on_the_genomic_accession=90280665, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  229542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229542, CurrentID=0, Status='live', Symbol='Gatad2b', Aliases='AL118180, C430014D17Rik, P66beta, mKIAA1150', description='GATA zinc finger domain containing 2B', other_designations='GATA zinc finger domain-containing protein 2B|p66/p68|transcription repressor p66 beta component of the MeCP1 complex|transcription repressor p66 component of the MeCP1 complex', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90293285, end_position_on_the_genomic_accession=90363406, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  229543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229543, CurrentID=0, Status='live', Symbol='Ints3', Aliases='C77668', description='integrator complex subunit 3', other_designations='SOSS complex subunit A|SOSS-A|int3|sensor of single-strand DNA complex subunit A|sensor of ssDNA subunit A', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90391380, end_position_on_the_genomic_accession=90433636, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  229550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229550, CurrentID=0, Status='live', Symbol='9130204L05Rik', Aliases='Gm127', description='RIKEN cDNA 9130204L05 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=91088137, end_position_on_the_genomic_accession=91107429, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  98488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98488, CurrentID=0, Status='live', Symbol='Gtf3c3', Aliases='TF3C-gamma, TFIIIC102', description='general transcription factor IIIC, polypeptide 3', other_designations='TFIIIC 102 kDa subunit|transcription factor IIIC-gamma subunit', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=54397581, end_position_on_the_genomic_accession=54438971, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  98496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98496, CurrentID=0, Status='live', Symbol='Pid1', Aliases='5033414K04Rik, NYGGF4', description='phosphotyrosine interaction domain containing 1', other_designations='P-CLI1|phosphotyrosine interaction domain-containing protein 1', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=84036293, end_position_on_the_genomic_accession=84284675, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  229571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229571, CurrentID=0, Status='live', Symbol='Gm4858', Aliases='EG229571', description='predicted gene 4858', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93068823, end_position_on_the_genomic_accession=93075505, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  229574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229574, CurrentID=0, Status='live', Symbol='Flg2', Aliases='EG229574', description='filaggrin family member 2', other_designations='FLG-2|filaggrin 2|intermediate filament-associated protein', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93197273, end_position_on_the_genomic_accession=93221377, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  229584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229584, CurrentID=0, Status='live', Symbol='Pogz', Aliases='9530006B08Rik', description='pogo transposable element with ZNF domain', other_designations='', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94823043, end_position_on_the_genomic_accession=94883569, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  229588 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229588, CurrentID=0, Status='live', Symbol='Gm128', Aliases='Ment, pmis1', description='predicted gene 128', other_designations='methylated in normal thymocytes protein|protein missing in infertile spermatozoa 1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95236920, end_position_on_the_genomic_accession=95241109, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  229589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229589, CurrentID=0, Status='live', Symbol='Prune', Aliases='9230112O05Rik, C130058A12, DRES-17, HTCD37, PRUNEM1-M1, Prune', description='prune exopolyphosphatase', other_designations='prune homolog', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95253674, end_position_on_the_genomic_accession=95282076, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  229593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229593, CurrentID=0, Status='live', Symbol='Golph3l', Aliases='2010204I15Rik, AU015439, AW549227, C77637, Gpp34r', description='golgi phosphoprotein 3-like', other_designations='', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95588915, end_position_on_the_genomic_accession=95619247, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  229595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229595, CurrentID=0, Status='live', Symbol='Adamtsl4', Aliases='ADAMTSL-4, Tsrc1', description='ADAMTS-like 4', other_designations='adamtsl4|thrombospondin repeat containing 1|thrombospondin repeat protein 1|thrombospondin repeat-containing protein 1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95676201, end_position_on_the_genomic_accession=95688444, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  229599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229599, CurrentID=0, Status='live', Symbol='Ciart', Aliases='Chrono, Gm129', description='circadian associated repressor of transcription', other_designations='chIP-derived repressor of network oscillator|computationally highlighted repressor of the network oscillator', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=95878500, end_position_on_the_genomic_accession=95882421, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  229603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229603, CurrentID=0, Status='live', Symbol='Otud7b', Aliases='2900060B22Rik, 4930463P07Rik, AI462125, CEZANNE, Trabid2, Za20d1', description='OTU domain containing 7B', other_designations='TRAF binding domain containing 2|cellular zinc finger anti-NF-kappa-B protein|zinc finger A20 domain-containing protein 1|zinc finger protein Cezanne|zinc finger, A20 domain containing 1', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96104451, end_position_on_the_genomic_accession=96161129, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  229615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229615, CurrentID=0, Status='live', Symbol='Pias3', Aliases='Pias3l', description='protein inhibitor of activated STAT 3', other_designations='protein inhibitor of activated STAT protein 3', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96696375, end_position_on_the_genomic_accession=96706070, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  234878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234878, CurrentID=0, Status='live', Symbol='BC021891', Aliases='Mlk4, mKIAA1804', description='cDNA sequence BC021891', other_designations='mixed lineage kinase 4', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125910450, end_position_on_the_genomic_accession=125947440, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  98558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98558, CurrentID=0, Status='live', Symbol='Mael', Aliases='4933405K18Rik, AU019877', description='maelstrom spermatogenic transposon silencer', other_designations='maelstrom homolog', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166201385, end_position_on_the_genomic_accession=166238744, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  229644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229644, CurrentID=0, Status='live', Symbol='Trim45', Aliases='4921530N01Rik', description='tripartite motif-containing 45', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100922493, end_position_on_the_genomic_accession=100936925, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  98582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98582, CurrentID=0, Status='live', Symbol='Khdc1b', Aliases='AU014713, AU015226, AU024076, AY238603, C85331, C86372, C88118, Khdc1, Khdc1c', description='KH domain containing 1B', other_designations='KH homology domain containing 1|KH homology domain containing 1B|oocyte-specific Khdc1|peptidase M20 domain containing 2', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=21383556, end_position_on_the_genomic_accession=21386384, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  229658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229658, CurrentID=0, Status='live', Symbol='Vangl1', Aliases='Kitenin, Lpp2, Stb2, Stbm', description='vang-like 1 (van gogh, Drosophila)', other_designations='KAI1 COOH-terminal interacting tetraspanin|loop-tail protein 2|strabismus|van Gogh-like protein 1|vang, van gogh-like 1', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102155899, end_position_on_the_genomic_accession=102205011, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  229663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229663, CurrentID=0, Status='live', Symbol='Csde1', Aliases='AA960392, BC016898, D3Jfr1, mKIAA0885, unr', description='cold shock domain containing E1, RNA binding', other_designations='NRAS-related|upstream of NRAS', map_location='3 45.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103020431, end_position_on_the_genomic_accession=103058189, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  229665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229665, CurrentID=0, Status='live', Symbol='Ampd1', Aliases='AI553520, Ampd-1', description='adenosine monophosphate deaminase 1', other_designations='adenosine monophosphate deaminase 1 (isoform M)|myoadenylate deaminase', map_location='3 45.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103073568, end_position_on_the_genomic_accession=103099721, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  229672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229672, CurrentID=0, Status='live', Symbol='Bcl2l15', Aliases='Bfk, Gm566', description='BCLl2-like 15', other_designations='Bcl2-like 15|bcl2-L-15', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103831039, end_position_on_the_genomic_accession=103854618, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  229675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229675, CurrentID=0, Status='live', Symbol='Rsbn1', Aliases='C230004D03Rik, E330004N01, Rosbin, Rsbn, Rsbp, mKIAA3002', description='rosbin, round spermatid basic protein 1', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103914067, end_position_on_the_genomic_accession=103966620, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  229681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229681, CurrentID=0, Status='live', Symbol='St7l', Aliases='St7r', description='suppression of tumorigenicity 7-like', other_designations='ST7-related protein|tumor suppressor ST7 like', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104864506, end_position_on_the_genomic_accession=104930064, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  229687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229687, CurrentID=0, Status='live', Symbol='Chil5', Aliases='Bclp, Bclp1, Bclp2, Chi3l7', description='chitinase-like 5', other_designations='chitinase 3-like 7|chitinase like protein 2', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106016919, end_position_on_the_genomic_accession=106033131, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  229688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229688, CurrentID=0, Status='live', Symbol='Chil6', Aliases='BYm, Chil8', description='chitinase-like 6', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106387384, end_position_on_the_genomic_accession=106409632, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  229699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229699, CurrentID=0, Status='live', Symbol='Slc16a4', Aliases='AW146050, MCT 5', description='solute carrier family 16 (monocarboxylic acid transporters), member 4', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107289865, end_position_on_the_genomic_accession=107312115, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  229700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229700, CurrentID=0, Status='live', Symbol='Rbm15', Aliases='C230088J01Rik, mKIAA1438', description='RNA binding motif protein 15', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107326110, end_position_on_the_genomic_accession=107333289, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  229706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229706, CurrentID=0, Status='live', Symbol='Slc6a17', Aliases='D130012J15Rik, Ntt4', description='solute carrier family 6 (neurotransmitter transporter), member 17', other_designations='orphan sodium- and chloride-dependent neurotransmitter transporter NTT4|sodium-dependent neurotransmitter transporter NTT4|sodium-dependent orphan neurotramsmitter transporter NTT4', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107467548, end_position_on_the_genomic_accession=107518018, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  229707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229707, CurrentID=0, Status='live', Symbol='Strip1', Aliases='6330569M22Rik, 6530401O14Rik, Fam40a', description='striatin interacting protein 1', other_designations='family with sequence similarity 40, member A|protein FAM40A', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107612532, end_position_on_the_genomic_accession=107631710, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  229709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229709, CurrentID=0, Status='live', Symbol='Ahcyl1', Aliases='1110034F20Rik, AA409031, AA414901, Ahcy-rs3, DCAL, Irbit', description='S-adenosylhomocysteine hydrolase-like 1', other_designations='IP3R binding protein released with inositol 1,4,5-trisphosphate|IP3R-binding protein released with inositol 1,4,5-trisphosphate|S-adenosyl-L-homocysteine hydrolase 2|S-adenosylhomocysteine hydrolase, related sequence 3|S-adenosylhomocysteine hydrolase-like protein 1|adoHcyase 2', map_location='3 46.83 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107663120, end_position_on_the_genomic_accession=107696548, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  229714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229714, CurrentID=0, Status='live', Symbol='Gpr61', Aliases='', description='G protein-coupled receptor 61', other_designations='', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108148321, end_position_on_the_genomic_accession=108154986, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  229715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229715, CurrentID=0, Status='live', Symbol='Amigo1', Aliases='Amigo, ali2', description='adhesion molecule with Ig like domain 1', other_designations='AMIGO-1|alivin 2|alivin-2|amigo1|amphoterin induced gene and ORF (Amigo)', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108186289, end_position_on_the_genomic_accession=108192286, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  229722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229722, CurrentID=0, Status='live', Symbol='5330417C22Rik', Aliases='BB183350, mKIAA1324', description='RIKEN cDNA 5330417C22 gene', other_designations='', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108455694, end_position_on_the_genomic_accession=108536748, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  229725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229725, CurrentID=0, Status='live', Symbol='Clcc1', Aliases='Mclc', description='chloride channel CLIC-like 1', other_designations='Mid-1-related chloride channel 1', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108653913, end_position_on_the_genomic_accession=108678840, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  229731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229731, CurrentID=0, Status='live', Symbol='Slc25a24', Aliases='2610016M12Rik', description='solute carrier family 25 (mitochondrial carrier, phosphate carrier), member 24', other_designations='small calcium-binding mitochondrial carrier protein 1|solute carrier family 25 member 24', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=109123149, end_position_on_the_genomic_accession=109168409, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  98660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98660, CurrentID=0, Status='live', Symbol='Atp1a2', Aliases='AW060654, Atpa-3, mKIAA0778', description='ATPase, Na+/K+ transporting, alpha 2 polypeptide', other_designations='Na(+)/K(+) ATPase alpha(+) subunit|Na(+)/K(+) ATPase alpha-2 subunit|sodium pump subunit alpha-2', map_location='1 79.6 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172271709, end_position_on_the_genomic_accession=172298064, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  98682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98682, CurrentID=0, Status='live', Symbol='Mfsd6', Aliases='2210010L05Rik, 9630025I22Rik, AI836200, AW212394, MMR2', description='major facilitator superfamily domain containing 6', other_designations='macrophage MHC class I receptor 2|macrophage MHC receptor 2', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=52656299, end_position_on_the_genomic_accession=52727462, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  98685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98685, CurrentID=0, Status='live', Symbol='Trmt1l', Aliases='1190005F20Rik, AW226554, C1orf25, Trm1-like, Trm1l', description='tRNA methyltransferase 1 like', other_designations='', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151428547, end_position_on_the_genomic_accession=151458183, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  229759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229759, CurrentID=0, Status='live', Symbol='Olfm3', Aliases='B230206G02Rik', description='olfactomedin 3', other_designations='olfactomedin-3|optimedin', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=114904078, end_position_on_the_genomic_accession=115125764, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  622976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=622976, CurrentID=0, Status='live', Symbol='Gm6377', Aliases='EG622976', description='predicted gene 6377', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=109191425, end_position_on_the_genomic_accession=109200448, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  229776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229776, CurrentID=0, Status='live', Symbol='Cdc14a', Aliases='A830059A17Rik, CDC14A2, CDC14a1, Cdc14', description='CDC14 cell division cycle 14A', other_designations='CDC14 cell division cycle 14 homolog A', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116272553, end_position_on_the_genomic_accession=116428743, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  229780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229780, CurrentID=0, Status='live', Symbol='Trmt13', Aliases='4631408H19Rik, A330067P21, A930028L21Rik, Ccdc76', description='tRNA methyltransferase 13', other_designations="coiled-coil domain containing 76|coiled-coil domain-containing protein 76|tRNA [Gm4] methyltransferase|tRNA guanosine-2'-O-methyltransferase TRM13 homolog|tRNA methyltransferase 13 homolog", map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116581093, end_position_on_the_genomic_accession=116614587, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  98710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98710, CurrentID=0, Status='live', Symbol='Rabif', Aliases='AI842864, AW549708, E130318E12Rik, MSS4, RASGFR3', description='RAB interacting factor', other_designations='rab-interacting factor', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134494660, end_position_on_the_genomic_accession=134507885, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  98711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98711, CurrentID=0, Status='live', Symbol='Rdh10', Aliases='3110069K09Rik, 4921506A21Rik, AI875664, AW549993, D1Ertd762e, m366Asp', description='retinol dehydrogenase 10 (all-trans)', other_designations='', map_location='1 4.94 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=16105780, end_position_on_the_genomic_accession=16133734, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  229791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229791, CurrentID=0, Status='live', Symbol='Plppr4', Aliases='A330086D10, D3Bwg0562e, Lppr4, PRG-1, mKIAA0455', description='phospholipid phosphatase related 4', other_designations='brain-specific phosphatidic acid phosphatase-like protein 1|plasticity related gene 1|plasticity-related gene 1 protein', map_location='3 50.57 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=117319145, end_position_on_the_genomic_accession=117360876, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  229801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229801, CurrentID=0, Status='live', Symbol='Tram1l1', Aliases='A830091N21Rik', description='translocation associated membrane protein 1-like 1', other_designations='', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=124321037, end_position_on_the_genomic_accession=124323260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  98732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98732, CurrentID=0, Status='live', Symbol='Rab3gap2', Aliases='1110059F07Rik, 2010002H18Rik, 5830469C09, AI851069, AW743433, RAB3-GAP150, mKIAA0839', description='RAB3 GTPase activating protein subunit 2', other_designations='Rab3 GTPase-activating protein 150 kDa subunit|Rab3-GAP regulatory subunit|rab3-GAP p150', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=185204088, end_position_on_the_genomic_accession=185286746, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  98741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98741, CurrentID=0, Status='live', Symbol='Kcnb2', Aliases='9630047L19Rik, BB130875, Kv2.2', description='potassium voltage gated channel, Shab-related subfamily, member 2', other_designations='voltage-gated potassium channel subunit Kv2.2', map_location='1 4.5 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=15287383, end_position_on_the_genomic_accession=15714214, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  628127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628127, CurrentID=0, Status='live', Symbol='Igkv14-126', Aliases='EG628127, Gm6844, IgK, IgM, br9, mga5th128, mga5th137', description='immunoglobulin kappa variable 14-126', other_designations='ASWS1 light chain variable region|Vk ASWS1|anti-glycyrrhetic acid antibody GA128 light chain|anti-glycyrrhetic acid antibody GA137 light chain|anti-polycation monoclonal antibody, Ig light chain variable region|immunoglobulin kappa chain variable 14-126|immunoglobulin kappa-chain VK-1|monoclonal antibody kappa light chain', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67896177, end_position_on_the_genomic_accession=67896642, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  98752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98752, CurrentID=0, Status='live', Symbol='Fcrla', Aliases='BB219290, FCRL, FCRL1, FREB, Fcrlm1, Fcrx, Freb1, mFREB, mFcrX', description='Fc receptor-like A', other_designations='Fc receptor related X|Fc receptor related protein X|Fc receptor-like mucin-like 1|fc receptor homolog expressed in B-cells|fc receptor-like and mucin-like protein 1|fc receptor-like protein|fc receptor-related protein X', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170917594, end_position_on_the_genomic_accession=170927590, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  98758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98758, CurrentID=0, Status='live', Symbol='Hnrnpf', Aliases='4833420I20Rik, AA407306, Hnrpf', description='heterogeneous nuclear ribonucleoprotein F', other_designations='hnRNP F', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=117900340, end_position_on_the_genomic_accession=117925622, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  98766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98766, CurrentID=0, Status='live', Symbol='Ubac1', Aliases='1110033G07Rik, AA407978, GBDR1, Gdbr1, Ubadc1', description='ubiquitin associated domain containing 1', other_designations='E3 ubiquitin-protein ligase subunit KPC2|UBA domain-containing protein 1|kip1 ubiquitination-promoting complex protein 2|putative glialblastoma cell differentiation-related', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25996958, end_position_on_the_genomic_accession=26021781, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  229841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229841, CurrentID=0, Status='live', Symbol='Cenpe', Aliases='312kDa, AU019344, BC049989, C530022J18, CENP-E, Kif10', description='centromere protein E', other_designations='CENP-E|KIF10|N-7 kinesin|cenpe|centromere autoantigen E|centromeric protein E|kinesin 10|kinesin superfamily protein 10|motor domain of KIF10', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135212563, end_position_on_the_genomic_accession=135273540, orientation='plus', exon_count=46, OMIM='', no_hdr0=''),
  282619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=282619, CurrentID=0, Status='live', Symbol='Sbsn', Aliases='1110005D19Rik, AI747508', description='suprabasin', other_designations='suprabasal-specific protein', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30748971, end_position_on_the_genomic_accession=30756134, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  229877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229877, CurrentID=0, Status='live', Symbol='Rap1gds1', Aliases='BC011279, GDS1', description='RAP1, GTP-GDP dissociation stimulator 1', other_designations='', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138925897, end_position_on_the_genomic_accession=139075201, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  623112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623112, CurrentID=0, Status='live', Symbol='Stmn1-rs2', Aliases='EG623112, Gm6393, Lag-rs2, Lap18rs2', description='stathmin 1, related sequence 2', other_designations='leukemia-associated gene, related sequence 2', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=70392597, end_position_on_the_genomic_accession=70393569, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  229898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229898, CurrentID=0, Status='live', Symbol='Gbp5', Aliases='5330409J06Rika, Gbp5', description='guanylate binding protein 5', other_designations='GBP-5|GTP-binding protein 5|guanine nucleotide-binding protein 5|guanylate nucleotide binding protein 5|muGBP-5', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142496934, end_position_on_the_genomic_accession=142522344, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  229900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229900, CurrentID=0, Status='live', Symbol='Gbp7', Aliases='9830147J24Rik', description='guanylate binding protein 7', other_designations='', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142530336, end_position_on_the_genomic_accession=142550151, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  229905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229905, CurrentID=0, Status='live', Symbol='Ccbl2', Aliases='KATIII, Kat3', description='cysteine conjugate-beta lyase 2', other_designations='cysteine-S-conjugate beta-lyase 2|kynurenine aminotransferase III|kynurenine--glyoxylate transaminase|kynurenine--oxoglutarate transaminase III', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142701048, end_position_on_the_genomic_accession=142744911, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  229906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229906, CurrentID=0, Status='live', Symbol='Gtf2b', Aliases='', description='general transcription factor IIB', other_designations='RNA polymerase II alpha initiation factor|general transcription factor TFIIB', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142765247, end_position_on_the_genomic_accession=142783606, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  98845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98845, CurrentID=0, Status='live', Symbol='Eps8l2', Aliases='AI042819, AW545405, Eps8r2', description='EPS8-like 2', other_designations='EPS8 related protein 2|EPS8-like protein 2|EPS8-related protein 2|epidermal growth factor receptor pathway substrate 8-like protein 2|epidermal growth factor receptor pathway substrate 8-related protein 2', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141338911, end_position_on_the_genomic_accession=141363016, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  229927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229927, CurrentID=0, Status='live', Symbol='Clca3b', Aliases='Clca4', description='chloride channel accessory 3B', other_designations='calcium-activated chloride channel CLCA4|chloride channel calcium activated 3B|chloride channel calcium activated 4', map_location='3 H2-H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144822623, end_position_on_the_genomic_accession=144849302, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  229933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229933, CurrentID=0, Status='live', Symbol='Clca2', Aliases='4732440A06, AI586120, Clca5', description='chloride channel accessory 2', other_designations='calcium-activated chloride channel family member 5|chloride channel calcium activated 2|chloride channel calcium activated 5|mCLCA5', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145070260, end_position_on_the_genomic_accession=145099495, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  229937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229937, CurrentID=0, Status='live', Symbol='Znhit6', Aliases='2410019A14Rik', description='zinc finger, HIT type 6', other_designations='zinc finger HIT domain-containing protein 6', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=145576197, end_position_on_the_genomic_accession=145605245, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  98870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98870, CurrentID=0, Status='live', Symbol='AI182371', Aliases='AI046667', description='expressed sequence AI182371', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=35081861, end_position_on_the_genomic_accession=35100685, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  229949 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=229949, CurrentID=0, Status='live', Symbol='Ak5', Aliases='', description='adenylate kinase 5', other_designations='AK 5|ATP-AMP transphosphorylase 5', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152462815, end_position_on_the_genomic_accession=152668376, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  98878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98878, CurrentID=0, Status='live', Symbol='Ehd4', Aliases='2210022F10Rik, AI197390, AI846352, AV006278, Past2', description='EH-domain containing 4', other_designations='EH domain containing protein MPAST2|PAST homolog 2|mPAST2', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120089487, end_position_on_the_genomic_accession=120154575, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  98910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98910, CurrentID=0, Status='live', Symbol='Usp6nl', Aliases='AI316785, RNTRE, TRE2NL, mKIAA0019', description='USP6 N-terminal like', other_designations='novel TBC domain containing protein', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=6322757, end_position_on_the_genomic_accession=6444363, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  98932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98932, CurrentID=0, Status='live', Symbol='Myl9', Aliases='AI327049, MLC20, Mylc2c, RLC-C', description='myosin, light polypeptide 9, regulatory', other_designations='myosin light chain, regulatory C|myosin regulatory light chain 2, smooth muscle isoform|myosin regulatory light chain 9', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156775435, end_position_on_the_genomic_accession=156781658, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  98952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98952, CurrentID=0, Status='live', Symbol='Fam102a', Aliases='AI426465, C230093N12Rik, Eeig1', description='family with sequence similarity 102, member A', other_designations='early estrogen-induced gene 1 protein', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32535359, end_position_on_the_genomic_accession=32569750, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  230025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230025, CurrentID=0, Status='live', Symbol='Prdm13', Aliases='', description='PR domain containing 13', other_designations='PR domain-containing protein 13', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=21675865, end_position_on_the_genomic_accession=21685963, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  230027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230027, CurrentID=0, Status='live', Symbol='Coq3', Aliases='4732433J24, C77934', description='coenzyme Q3 methyltransferase', other_designations='2-polyprenyl-6-hydroxyphenol methylase|2-polyprenyl-6-hydroxyphenyl methylase|3,4-dihydroxy-5-hexaprenylbenzoate methyltransferase|3-demethylubiquinol 3-O-methyltransferase|3-demethylubiquinone-9 3-methyltransferase|Coenzyme Q (ubiquinone)|DHHB methyltransferase|DHHB-MT|DHHB-MTase|coenzyme Q3 homolog methyltransferase|coenzyme Q3 homolog, methyltransferase|dihydroxyhexaprenylbenzoate methyltransferase|hexaprenyldihydroxybenzoate methyltransferase, mitochondrial|polyprenyldihydroxybenzoate methyltransferase', map_location='4 9.11 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=21879675, end_position_on_the_genomic_accession=21912163, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  98956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98956, CurrentID=0, Status='live', Symbol='Nat10', Aliases='AI429152', description='N-acetyltransferase 10', other_designations='18S rRNA cytosine acetyltransferase|novel putative ATPase domain containing protein', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103721258, end_position_on_the_genomic_accession=103762647, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  98970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98970, CurrentID=0, Status='live', Symbol='Fibcd1', Aliases='AI448887', description='fibrinogen C domain containing 1', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31813290, end_position_on_the_genomic_accession=31846647, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  98985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98985, CurrentID=0, Status='live', Symbol='Clp1', Aliases='AI462438, Heab', description='CLP1, cleavage and polyadenylation factor I subunit', other_designations='ATP/GTP-binding protein|CLP1, cleavage and polyadenylation factor I subunit, homolog|polyadenylation factor Clp1|polynucleotide kinase Clp1|pre-mRNA cleavage complex II protein Clp1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84723122, end_position_on_the_genomic_accession=84727268, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  623279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623279, CurrentID=0, Status='live', Symbol='Dok6', Aliases='Dok-6', description='docking protein 6', other_designations='downstream of tyrosine kinase 6|downstream of tyrosine kinases 6', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=89301135, end_position_on_the_genomic_accession=89769136, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  98999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=98999, CurrentID=0, Status='live', Symbol='Znfx1', Aliases='AI481105, Astx3', description='zinc finger, NFX1-type containing 1', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=167035790, end_position_on_the_genomic_accession=167063015, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  230073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230073, CurrentID=0, Status='live', Symbol='Ddx58', Aliases='6430573D20Rik, C330021E21, RIG-I, RLR-1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 58', other_designations='DEAD box protein 58|DEAD-box protein 58|DEAD/H box polypeptide RIG-I|RIG-1|RIG-I-like receptor 1|retinoic acid-inducible gene 1 protein|retinoic acid-inducible gene I protein|retinoic acid-inducible gene-I', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40203777, end_position_on_the_genomic_accession=40239825, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  230075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230075, CurrentID=0, Status='live', Symbol='Ndufb6', Aliases='CI-B17, Gm137', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 6', other_designations='NADH-ubiquinone oxidoreductase B17 subunit|complex I-B17', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40270663, end_position_on_the_genomic_accession=40279368, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  230082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230082, CurrentID=0, Status='live', Symbol='Nol6', Aliases='AA410091, Nrap', description='nucleolar protein family 6 (RNA-associated)', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41114427, end_position_on_the_genomic_accession=41124339, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  99011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99011, CurrentID=0, Status='live', Symbol='Pomt1', Aliases='AI505244', description='protein-O-mannosyltransferase 1', other_designations='dolichyl-phosphate-mannose--protein mannosyltransferase 1|protein O-mannosyltransferase 1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32236595, end_position_on_the_genomic_accession=32255005, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  230085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230085, CurrentID=0, Status='live', Symbol='Phf24', Aliases='A230025O18, Kiaa1045, N28178, mKIAA1045', description='PHD finger protein 24', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=42872619, end_position_on_the_genomic_accession=42944151, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  230088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230088, CurrentID=0, Status='live', Symbol='Fam214b', Aliases='B230312A22Rik', description='family with sequence similarity 214, member B', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43032414, end_position_on_the_genomic_accession=43046220, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  230098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230098, CurrentID=0, Status='live', Symbol='Arhgef39', Aliases='E130306D19Rik', description='Rho guanine nucleotide exchange factor (GEF) 39', other_designations='vav-like protein C9orf100 homolog', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43496139, end_position_on_the_genomic_accession=43508769, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  230099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230099, CurrentID=0, Status='live', Symbol='Car9', Aliases='CAIX, Ca9, MN/CA9', description='carbonic anhydrase 9', other_designations='CA-IX|carbonate dehydratase IX|carbonic anhydrase IX|membrane antigen MN homolog', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43507026, end_position_on_the_genomic_accession=43513810, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  230101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230101, CurrentID=0, Status='live', Symbol='Gba2', Aliases='F630034E04', description='glucosidase beta 2', other_designations='NLGase|beta-glucocerebrosidase 2|beta-glucosidase 2|bile acid|glucosidase, beta (bile acid) 2|glucosylceramidase 2', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43566928, end_position_on_the_genomic_accession=43578864, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  230103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230103, CurrentID=0, Status='live', Symbol='Npr2', Aliases='GC-B, GC-B2, GC-B3, cn, mNPR-B, pwe', description='natriuretic peptide receptor 2', other_designations='ANP-B|ANPR-B|NPR-B|Nppb receptor|atrial natriuretic peptide receptor type B|guanylate cyclase B|guanylyl cyclase-B|natriuretic peptide receptor B type', map_location='4 23.05 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43630332, end_position_on_the_genomic_accession=43651437, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  99035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99035, CurrentID=0, Status='live', Symbol='Olah', Aliases='AI607300, E230009B14Rik, Thedc1', description='oleoyl-ACP hydrolase', other_designations='thioesterase II|thioesterase domain containing 1|thioesterase domain-containing protein 1', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3341982, end_position_on_the_genomic_accession=3397238, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  99045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99045, CurrentID=0, Status='live', Symbol='Mrps26', Aliases='AI648866, GI008, MRP-S13, MRP-S26, Rpms13', description='mitochondrial ribosomal protein S26', other_designations='S26mt', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130563757, end_position_on_the_genomic_accession=130565394, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  230119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230119, CurrentID=0, Status='live', Symbol='Zbtb5', Aliases='5930421I10, 9430083K24Rik, AI646847, mKIAA0354', description='zinc finger and BTB domain containing 5', other_designations='expressed in brain and neural tube|transcription factor ZNF-POZ', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=44991242, end_position_on_the_genomic_accession=45012412, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  230125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230125, CurrentID=0, Status='live', Symbol='Slc25a51', Aliases='9130208E07Rik, D130005A03Rik, Gm138, Mcart1', description='solute carrier family 25, member 51', other_designations='mitochondrial carrier triple repeat 1|mitochondrial carrier triple repeat protein 1', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45395923, end_position_on_the_genomic_accession=45408783, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  230126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230126, CurrentID=0, Status='live', Symbol='Shb', Aliases='BC028832', description='src homology 2 domain-containing transforming protein B', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45423276, end_position_on_the_genomic_accession=45530828, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  230145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230145, CurrentID=0, Status='live', Symbol='Galnt12', Aliases='9130206E10, A630062B03Rik', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 12', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 12|galNAc-T12|mpp-GalNAc-T12|polypeptide GalNAc transferase 12|pp-GaNTase 12|protein-UDP acetylgalactosaminyltransferase 12', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=47091893, end_position_on_the_genomic_accession=47123042, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  230157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230157, CurrentID=0, Status='live', Symbol='Tmeff1', Aliases='A830033E11, M7365, Tr1', description='transmembrane protein with EGF-like and two follistatin-like domains 1', other_designations='TR-1|transmembrane protein with EGF-like and one follistatin-like domain', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=48585114, end_position_on_the_genomic_accession=48663131, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  230161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230161, CurrentID=0, Status='live', Symbol='Acnat1', Aliases='AI132189', description='acyl-coenzyme A amino acid N-acyltransferase 1', other_designations='acnat1', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49443532, end_position_on_the_genomic_accession=49481312, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  230162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230162, CurrentID=0, Status='live', Symbol='Zfp189', Aliases='BB131811, C430015I23Rik', description='zinc finger protein 189', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49521176, end_position_on_the_genomic_accession=49531558, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  230163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230163, CurrentID=0, Status='live', Symbol='Aldob', Aliases='Aldo-2, Aldo2, BC016435', description='aldolase B, fructose-bisphosphate', other_designations='aldolase 2, B isoform|liver-type aldolase', map_location='4 26.57 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49535995, end_position_on_the_genomic_accession=49549483, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  628185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628185, CurrentID=0, Status='live', Symbol='Vmn2r112', Aliases='EG628185, V2rp1', description='vomeronasal 2, receptor 112', other_designations='vomeronasal 2 receptor p1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=22601148, end_position_on_the_genomic_accession=22619133, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  99100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99100, CurrentID=0, Status='live', Symbol='Cep152', Aliases='AI851464, mKIAA0912', description='centrosomal protein 152', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125563088, end_position_on_the_genomic_accession=125625176, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  99138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99138, CurrentID=0, Status='live', Symbol='Stard7', Aliases='AI852671, AL022671, AW544915', description='START domain containing 7', other_designations='START domain-containing protein 7', map_location='2 61.88 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127270203, end_position_on_the_genomic_accession=127298935, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  99151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99151, CurrentID=0, Status='live', Symbol='Cercam', Aliases='2600006K01Rik, AL024097, Ceecam1', description='cerebral endothelial cell adhesion molecule', other_designations='cerebral endothelial cell adhesion molecule 1|glycosyltransferase 25 family member 3', map_location='2 20.76 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29845105, end_position_on_the_genomic_accession=29882845, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  99152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99152, CurrentID=0, Status='live', Symbol='Anapc2', Aliases='9230107K09Rik, AL024279, Apc2, Emi4, Imi4, mKIAA1406', description='anaphase promoting complex subunit 2', other_designations='cyclosome subunit 2|expressed during mesenchymal induction 4|induced during mesenchymal induction 4', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25272466, end_position_on_the_genomic_accession=25285916, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  230233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230233, CurrentID=0, Status='live', Symbol='Ikbkap', Aliases='3110040G09Rik, 6030413P05, C78473, Elp1, IKAP', description='inhibitor of kappa light polypeptide enhancer in B cells, kinase complex-associated protein', other_designations='IKK complex-associated protein|IKK-complex-associated protein IKAP|ikappaB kinase complex-associated protein|inhibitor of kappa light polypeptide enhancer in B-cells, kinase complex-associated protein', map_location='4 31.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56749680, end_position_on_the_genomic_accession=56802331, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  230234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230234, CurrentID=0, Status='live', Symbol='Fam206a', Aliases='simiate', description='family with sequence similarity 206, member A', other_designations='protein FAM206A', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56802329, end_position_on_the_genomic_accession=56809605, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  230235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230235, CurrentID=0, Status='live', Symbol='Frrs1l', Aliases='6430704M03Rik', description='ferric-chelate reductase 1 like', other_designations='ferric-chelate reductase 1-like protein', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56960136, end_position_on_the_genomic_accession=56990391, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  99167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99167, CurrentID=0, Status='live', Symbol='Ssx2ip', Aliases='AU014939, AU042321, Adip', description='synovial sarcoma, X breakpoint 2 interacting protein', other_designations='afadin DIL domain-interacting protein', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146404617, end_position_on_the_genomic_accession=146440144, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  278672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=278672, CurrentID=0, Status='live', Symbol='Duxbl1', Aliases='1110051B16Rik, Duxbl, Duxl', description='double homeobox B-like 1', other_designations='Dux in lymphoid lineage|double homeodomain protein', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25978918, end_position_on_the_genomic_accession=25990145, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  230249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230249, CurrentID=0, Status='live', Symbol='AI314180', Aliases='AW558785, BB181316, mKIAA0368', description='expressed sequence AI314180', other_designations='ecm29', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58797237, end_position_on_the_genomic_accession=58913123, orientation='minus', exon_count=52, OMIM='', no_hdr0=''),
  230257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230257, CurrentID=0, Status='live', Symbol='Ptbp3', Aliases='5830471K22Rik, AA407443, AI462022, AW107884, C86549, Rod1', description='polypyrimidine tract binding protein 3', other_designations='ROD1 regulator of differentiation 1|regulator of differentiation 1', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59471868, end_position_on_the_genomic_accession=59549379, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  623474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623474, CurrentID=0, Status='live', Symbol='Rad54b', Aliases='E130016E03Rik', description='RAD54 homolog B (S. cerevisiae)', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11558920, end_position_on_the_genomic_accession=11615808, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  230259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230259, CurrentID=0, Status='live', Symbol='E130308A19Rik', Aliases='', description='RIKEN cDNA E130308A19 gene', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59626200, end_position_on_the_genomic_accession=59759287, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  230279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230279, CurrentID=0, Status='live', Symbol='6330416G13Rik', Aliases='AW492431, bM230C2.2', description='RIKEN cDNA 6330416G13 gene', other_designations='', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63558416, end_position_on_the_genomic_accession=63586357, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  16535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16535, CurrentID=0, Status='live', Symbol='Kcnq1', Aliases='AW559127, KVLQT1, Kcna9', description='potassium voltage-gated channel, subfamily Q, member 1', other_designations='IKs producing slow voltage-gated potassium channel subunit alpha KvLQT1|KQT-like 1|voltage-gated potassium channel subunit Kv7.1', map_location='7 88.12 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143106742, end_position_on_the_genomic_accession=143427050, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  623503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623503, CurrentID=0, Status='live', Symbol='Prlh', Aliases='EG623503', description='prolactin releasing hormone', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90951256, end_position_on_the_genomic_accession=90954369, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  99237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99237, CurrentID=0, Status='live', Symbol='Tm9sf4', Aliases='AA986553, AU045326, B930079E06, mKIAA0255', description='transmembrane 9 superfamily protein member 4', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153161301, end_position_on_the_genomic_accession=153210466, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  230316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230316, CurrentID=0, Status='live', Symbol='Megf9', Aliases='4933405H16Rik, 9630025C07, Egfl5, mKIAA0818', description='multiple EGF-like-domains 9', other_designations='EGF-like protein 5|EGF-like-domain, multiple 5|epidermal growth factor-like protein 5|multiple EGF-like domains protein 9', map_location='4 C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=70431927, end_position_on_the_genomic_accession=70534928, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  623534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623534, CurrentID=0, Status='live', Symbol='Nme9', Aliases='Txndc6', description='NME/NM23 family member 9', other_designations='thioredoxin domain containing 6', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99456243, end_position_on_the_genomic_accession=99479899, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  99296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99296, CurrentID=0, Status='live', Symbol='Hrh3', Aliases='AW049250, H3R, HH3R', description='histamine receptor H3', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180099465, end_position_on_the_genomic_accession=180104407, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  230376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230376, CurrentID=0, Status='live', Symbol='Haus6', Aliases='6230416J20Rik, 9830144E06, D4Ertd27e, mKIAA1574', description='HAUS augmin-like complex, subunit 6', other_designations='likely ortholog of H. sapiens family with sequence similarity 29, member A (FAM29A)', map_location='4 40.69 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86578855, end_position_on_the_genomic_accession=86612076, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  230379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230379, CurrentID=0, Status='live', Symbol='Acer2', Aliases='2410116I05Rik, Asah3l, CRG-L1, maCER2', description='alkaline ceramidase 2', other_designations='alkCDase 2|alkaline CDase 2|cancer related gene-liver 1 (CRG-L1)|cancer-related gene liver 1 protein|ceramide hydrolase', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=86874342, end_position_on_the_genomic_accession=86932490, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  99311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99311, CurrentID=0, Status='live', Symbol='Commd7', Aliases='2310010I22Rik, AW060956, mU3', description='COMM domain containing 7', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153616930, end_position_on_the_genomic_accession=153632794, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  230393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230393, CurrentID=0, Status='live', Symbol='Focad', Aliases='Kiaa1797, mKIAA1797', description='focadhesin', other_designations='', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88094630, end_position_on_the_genomic_accession=88411011, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  230396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230396, CurrentID=0, Status='live', Symbol='Ifna13', Aliases='', description='interferon alpha 13', other_designations='IFN-alpha-13|interferon alpha family, gene 13', map_location='4 42.02 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88643641, end_position_on_the_genomic_accession=88644459, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  230398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230398, CurrentID=0, Status='live', Symbol='Ifna16', Aliases='BC114392, Gm13280, Ifna6T, Ifnat6, OTTMUSG00000011275', description='interferon alpha 16', other_designations='interferon alpha 6T', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88676287, end_position_on_the_genomic_accession=88676856, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  230405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230405, CurrentID=0, Status='live', Symbol='Ifne', Aliases='Ifn-tau-11, Ifnt1, Infe1, Ifne', description='interferon epsilon', other_designations='IFN-epsilon|interferon epsilon 1|interferon epsilon-1|interferon tau-1', map_location='4 42.04 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88879538, end_position_on_the_genomic_accession=88880201, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  99334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99334, CurrentID=0, Status='live', Symbol='Zscan29', Aliases='AW125441, Zfp690', description='zinc finger SCAN domains 29', other_designations='zinc finger protein 690', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121158273, end_position_on_the_genomic_accession=121171125, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  99349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99349, CurrentID=0, Status='live', Symbol='Dnajc24', Aliases='1700030A21Rik, 2610027M02Rik, AV066965, AW240712, Dph4, MmDjC7, Zcsl3', description='DnaJ heat shock protein family (Hsp40) member C24', other_designations='CSL-type zinc finger-containing protein 3|DPH4 homolog (JJJ3, S. cerevisiae)|DnaJ (Hsp40) homolog, subfamily C, member 24|diphthamide biosynthesis protein 4|j domain protein DjC7|zinc finger, CSL-type containing 3', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=105966708, end_position_on_the_genomic_accession=106003868, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  99371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99371, CurrentID=0, Status='live', Symbol='Arfgef2', Aliases='AI463430, AW493672, BIG2, E230011G24Rik', description='ADP-ribosylation factor guanine nucleotide-exchange factor 2 (brefeldin A-inhibited)', other_designations='brefeldin A-inhibited GEP 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=166805465, end_position_on_the_genomic_accession=166898052, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  623661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623661, CurrentID=0, Status='live', Symbol='Lipt1', Aliases='EG623661', description='lipoyltransferase 1', other_designations='lipoate biosynthesis protein|lipoate-protein ligase|lipoyl ligase', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37868507, end_position_on_the_genomic_accession=37876298, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  99375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99375, CurrentID=0, Status='live', Symbol='Cul4a', Aliases='2810470J21Rik, AW495282', description='cullin 4A', other_designations='CUL-4A', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13104862, end_position_on_the_genomic_accession=13147940, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  99377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99377, CurrentID=0, Status='live', Symbol='Sall4', Aliases='5730441M18Rik, AA407717, AL022809, AW536104, C330011P20Rik, C78083, C78563, Tex20', description='sal-like 4 (Drosophila)', other_designations='zinc finger protein SALL4', map_location='2 88.99 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168748332, end_position_on_the_genomic_accession=168767201, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  99382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99382, CurrentID=0, Status='live', Symbol='Abtb2', Aliases='AW539457, BPOZ-2', description='ankyrin repeat and BTB (POZ) domain containing 2', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=103566310, end_position_on_the_genomic_accession=103718423, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  230459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230459, CurrentID=0, Status='live', Symbol='Cyp2j13', Aliases='', description='cytochrome P450, family 2, subfamily j, polypeptide 13', other_designations='cytochrome P450, 2j13', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=96042660, end_position_on_the_genomic_accession=96077566, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  102466637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102466637, CurrentID=0, Status='live', Symbol='Mir21b', Aliases='Gm27951, mmu-mir-21b', description='microRNA 21b', other_designations='microRNA mir-21b', map_location='', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=29636744, end_position_on_the_genomic_accession=29636851, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  99412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99412, CurrentID=0, Status='live', Symbol='Golga2', Aliases='AW555139, GM130, mKIAA4150', description='golgi autoantigen, golgin subfamily a, 2', other_designations='130 kDa cis-Golgi matrix protein|SY11 protein|golgin-95', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32288221, end_position_on_the_genomic_accession=32307921, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  230500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230500, CurrentID=0, Status='live', Symbol='Efcab7', Aliases='mKIAA1799', description='EF-hand calcium binding domain 7', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=99829500, end_position_on_the_genomic_accession=99912785, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  99439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99439, CurrentID=0, Status='live', Symbol='Duox1', Aliases='9930101G15Rik, AW987690, Duox2, LNOX1, LNOX2, NOXEF1, NOXEF2, P138-TOX, THOX1, THOX2', description='dual oxidase 1', other_designations='dual oxidase 2', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122313811, end_position_on_the_genomic_accession=122349226, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  230514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230514, CurrentID=0, Status='live', Symbol='Leprot', Aliases='LEPR, OB-RGRP, Obrgrp, VPS55', description='leptin receptor overlapping transcript', other_designations='OB-R gene-related protein|endospanin-1', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101647783, end_position_on_the_genomic_accession=101659358, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  99470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99470, CurrentID=0, Status='live', Symbol='Magi3', Aliases='4732496O19Rik, 6530407C02Rik, AA407180, AI120132, mKIAA1634', description='membrane associated guanylate kinase, WW and PDZ domain containing 3', other_designations='MAGI-3|membrane-associated guanylate kinase inverted 3|membrane-associated guanylate kinase-related (MAGI-3)', map_location='3 45.52 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=104013259, end_position_on_the_genomic_accession=104220406, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  99480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99480, CurrentID=0, Status='live', Symbol='Dnttip2', Aliases='4930588M11Rik, AA408582, AU014960, HSU15552', description='deoxynucleotidyltransferase, terminal, interacting protein 2', other_designations='acidic 82 kDa protein mRNA', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122274414, end_position_on_the_genomic_accession=122285271, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  230558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230558, CurrentID=0, Status='live', Symbol='C8a', Aliases='', description='complement component 8, alpha polypeptide', other_designations='complement component 8 alpha subunit', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=104815679, end_position_on_the_genomic_accession=104876486, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  623781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=623781, CurrentID=0, Status='live', Symbol='Gm14137', Aliases='OTTMUSG00000015762', description='predicted gene 14137', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119174509, end_position_on_the_genomic_accession=119177575, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103964, CurrentID=0, Status='live', Symbol='Try5', Aliases='1810049H19Rik, Tc, Tesp4', description='trypsin 5', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41311232, end_position_on_the_genomic_accession=41314710, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  230576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230576, CurrentID=0, Status='live', Symbol='Ttc22', Aliases='4732467L16Rik', description='tetratricopeptide repeat domain 22', other_designations='TPR repeat protein 22', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106622449, end_position_on_the_genomic_accession=106640196, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  230577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230577, CurrentID=0, Status='live', Symbol='Pars2', Aliases='BC027073', description='prolyl-tRNA synthetase (mitochondrial)(putative)', other_designations='class II tRNA synthetase|proRS|probable prolyl-tRNA synthetase, mitochondrial|proline--tRNA ligase', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106651021, end_position_on_the_genomic_accession=106655282, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  230579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230579, CurrentID=0, Status='live', Symbol='Fam151a', Aliases='2010309H15Rik', description='family with sequence simliarity 151, member A', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106733915, end_position_on_the_genomic_accession=106748292, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  230582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230582, CurrentID=0, Status='live', Symbol='Cyb5rl', Aliases='2810410C14Rik, mFLJ00377', description='cytochrome b5 reductase-like', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107066844, end_position_on_the_genomic_accession=107087993, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  99512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99512, CurrentID=0, Status='live', Symbol='Wdr47', Aliases='1810073M12Rik, AA552998, mKIAA0893', description='WD repeat domain 47', other_designations='nemitin|neuronal enriched MAP interacting protein', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108591256, end_position_on_the_genomic_accession=108645719, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  230587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230587, CurrentID=0, Status='live', Symbol='Glis1', Aliases='Gli5, Gli6, GliH1', description='GLIS family zinc finger 1', other_designations='GLI-Kruppel family member GLI6|GLI-similar 1|Glis1 Kruppel-like zinc finger protein|gli homologous protein 1|zinc finger transcription factor Gli5', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107426640, end_position_on_the_genomic_accession=107635061, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  230590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230590, CurrentID=0, Status='live', Symbol='Zyg11a', Aliases='BC022150', description='zyg-11 family member A, cell cycle regulator', other_designations='zyg-11 homolog A', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108180245, end_position_on_the_genomic_accession=108218160, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  103968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103968, CurrentID=0, Status='live', Symbol='Plin1', Aliases='6030432J05Rik, Peri, Plin', description='perilipin 1', other_designations='lipid droplet-associated protein|perilipin A|perilipin B', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=79721164, end_position_on_the_genomic_accession=79732904, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  230594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230594, CurrentID=0, Status='live', Symbol='Zcchc11', Aliases='6030404K05Rik, 9230115F04Rik, PPAPD3, mKIAA0191', description='zinc finger, CCHC domain containing 11', other_designations='TUTase 4|zinc finger CCHC domain-containing protein 11', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108459426, end_position_on_the_genomic_accession=108559415, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  230596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230596, CurrentID=0, Status='live', Symbol='Prpf38a', Aliases='2410002M20Rik', description='PRP38 pre-mRNA processing factor 38 (yeast) domain containing A', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108564867, end_position_on_the_genomic_accession=108579336, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  230597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230597, CurrentID=0, Status='live', Symbol='Zfyve9', Aliases='E030027L17, Madhip, NSP, SARA, SMADIP', description='zinc finger, FYVE domain containing 9', other_designations='MAD, mothers against decapentaplegic homolog interacting protein, receptor activation anchor', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108637834, end_position_on_the_genomic_accession=108781148, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  230598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230598, CurrentID=0, Status='live', Symbol='Nrd1', Aliases='2600011I06Rik, AI875733, NRD-C', description='nardilysin, N-arginine dibasic convertase, NRD convertase 1', other_designations='N-arginine dibasic convertase|NRD convertase', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109000615, end_position_on_the_genomic_accession=109061771, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  230603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230603, CurrentID=0, Status='live', Symbol='Ttc39a', Aliases='4922503N01Rik', description='tetratricopeptide repeat domain 39A', other_designations='TPR repeat protein 39A', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109394148, end_position_on_the_genomic_accession=109444745, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  230612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230612, CurrentID=0, Status='live', Symbol='Slc5a9', Aliases='AI159731, SGLT4', description='solute carrier family 5 (sodium/glucose cotransporter), member 9', other_designations='Na(+)/glucose cotransporter 4|mSGLT4|solute carrier family 5 member 9', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=111875374, end_position_on_the_genomic_accession=111902796, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  230613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230613, CurrentID=0, Status='live', Symbol='Skint10', Aliases='A030001H23Rik, Gm570', description='selection and upkeep of intraepithelial T cells 10', other_designations='skint 10|skint-10', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=112710830, end_position_on_the_genomic_accession=112774861, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  99543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99543, CurrentID=0, Status='live', Symbol='Olfml3', Aliases='2810002E22Rik, AI115209, AW550633, HNOEL-iso, ONT3, mONT3', description='olfactomedin-like 3', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=103735394, end_position_on_the_genomic_accession=103738001, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  230623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230623, CurrentID=0, Status='live', Symbol='Skint11', Aliases='A630098G03Rik, Gm569', description='selection and upkeep of intraepithelial T cells 11', other_designations='skint 11|skint-11', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=114163384, end_position_on_the_genomic_accession=114245028, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  99571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99571, CurrentID=0, Status='live', Symbol='Fgg', Aliases='3010002H13Rik, AI256424', description='fibrinogen gamma chain', other_designations='fibrinogen gamma-chain|fibrinogen, gamma polypeptide', map_location='3 36.94 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=83007896, end_position_on_the_genomic_accession=83015049, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  230648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230648, CurrentID=0, Status='live', Symbol='Efcab14', Aliases='4732418C07Rik', description='EF-hand calcium binding domain 14', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115737716, end_position_on_the_genomic_accession=115777328, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  230649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230649, CurrentID=0, Status='live', Symbol='Atpaf1', Aliases='6330547J17Rik, AI593252, ATP11, ATP11p', description='ATP synthase mitochondrial F1 complex assembly factor 1', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115784814, end_position_on_the_genomic_accession=115812314, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  103978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103978, CurrentID=0, Status='live', Symbol='Gpc5', Aliases='A230034F01Rik', description='glypican 5', other_designations='', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115092212, end_position_on_the_genomic_accession=116525192, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  230654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230654, CurrentID=0, Status='live', Symbol='Lrrc41', Aliases='AA409966, AW555107, D630045E04Rik, D730026A16Rik, MUF1', description='leucine rich repeat containing 41', other_designations='protein Muf1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116075465, end_position_on_the_genomic_accession=116097043, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  230657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230657, CurrentID=0, Status='live', Symbol='Tmem69', Aliases='A630048M13Rik', description='transmembrane protein 69', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116551528, end_position_on_the_genomic_accession=116557042, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  99586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99586, CurrentID=0, Status='live', Symbol='Dpyd', Aliases='AI315208, DPD, E330028L06Rik', description='dihydropyrimidine dehydrogenase', other_designations='DHPDHase|dihydrothymine dehydrogenase|dihydrouracil dehydrogenase', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=118562109, end_position_on_the_genomic_accession=119432918, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  230661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230661, CurrentID=0, Status='live', Symbol='Tesk2', Aliases='', description='testis-specific kinase 2', other_designations='testicular protein kinase 2', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116720909, end_position_on_the_genomic_accession=116804248, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  230673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230673, CurrentID=0, Status='live', Symbol='Ipo13', Aliases='Imp13, Kap13, Ranbp13', description='importin 13', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117894486, end_position_on_the_genomic_accession=117915176, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  230674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230674, CurrentID=0, Status='live', Symbol='Kdm4a', Aliases='D4Ertd222e, JHDM3A, Jmjd2, Jmjd2a, mKIAA0677', description='lysine (K)-specific demethylase 4A', other_designations='jmjC domain-containing histone demethylation protein 3A|jumonji domain containing 2A|jumonji domain-containing protein 2A', map_location='4 54.31 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118136957, end_position_on_the_genomic_accession=118180043, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  230676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230676, CurrentID=0, Status='live', Symbol='Szt2', Aliases='6430407N22, ENSMUSG00000058432, TIGR', description='seizure threshold 2', other_designations='Seizure threshold 2 protein|Transcript increased in glutamate resistance', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118362740, end_position_on_the_genomic_accession=118409276, orientation='minus', exon_count=72, OMIM='', no_hdr0=''),
  230678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230678, CurrentID=0, Status='live', Symbol='Tmem125', Aliases='6330530A05Rik, AV223516', description='transmembrane protein 125', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118540941, end_position_on_the_genomic_accession=118544953, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  492824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492824, CurrentID=0, Status='live', Symbol='Ath11', Aliases='', description='atherosclerosis 11', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  492831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492831, CurrentID=0, Status='live', Symbol='Bmd13', Aliases='', description='bone mineral density 13', other_designations='', map_location='13 21.73 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  492834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492834, CurrentID=0, Status='live', Symbol='Bmd5', Aliases='', description='bone mineral density 5', other_designations='', map_location='1 37.02 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  492836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492836, CurrentID=0, Status='live', Symbol='Bmd7', Aliases='', description='bone mineral density 7', other_designations='', map_location='4 42.13 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230696, CurrentID=0, Status='live', Symbol='AU022252', Aliases='AI450326', description='expressed sequence AU022252', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119221073, end_position_on_the_genomic_accession=119232765, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  492842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492842, CurrentID=0, Status='live', Symbol='Hrec', Aliases='', description='hemotopoietic response to early cytokines', other_designations='', map_location='2 65.46 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230700, CurrentID=0, Status='live', Symbol='Foxj3', Aliases='C330039G02Rik, Fhd6', description='forkhead box J3', other_designations='forkhead domain gene 6', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119539661, end_position_on_the_genomic_accession=119629119, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  99633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99633, CurrentID=0, Status='live', Symbol='Adgrl2', Aliases='AI450192, CIRL-2, Gm619, Lec1, Lphh1, Lphn2, mKIAA0786', description='adhesion G protein-coupled receptor L2', other_designations='calcium-independent alpha-latrotoxin receptor 2|calcium-independent alpha-latrotoxin receptor homolog 2|latrophilin 1|latrophilin 2|latrophilin-2', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=148815586, end_position_on_the_genomic_accession=149445423, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  230709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230709, CurrentID=0, Status='live', Symbol='Zmpste24', Aliases='A530043O15Rik, D030046F19, FACE1, Face-1, MADB, STE24, Ste24p', description='zinc metallopeptidase, STE24', other_designations='farnesylated proteins-converting enzyme 1|farnesylated-proteins converting enzyme 1|prenyl protein-specific endoprotease 1|zinc metallopeptidase STE24|zinc metallopeptidase, STE24 homolog|zinc metalloproteinase Ste24 homolog|zinc metalloproteinase, STE24 homolog', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=121059237, end_position_on_the_genomic_accession=121098249, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  492854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492854, CurrentID=0, Status='live', Symbol='Sluc20', Aliases='', description='susceptibility to lung cancer 20', other_designations='', map_location='8 12.86 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230718, CurrentID=0, Status='live', Symbol='Nt5c1a', Aliases='Cn1a, Gm49', description="5'-nucleotidase, cytosolic IA", other_designations="cN-IA|cytosolic 5' nucleotidase, type 1A|cytosolic 5'-nucleotidase IA", map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123201210, end_position_on_the_genomic_accession=123225333, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  230721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230721, CurrentID=0, Status='live', Symbol='Pabpc4', Aliases='', description='poly(A) binding protein, cytoplasmic 4', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123262387, end_position_on_the_genomic_accession=123298832, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  99650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99650, CurrentID=0, Status='live', Symbol='4933434E20Rik', Aliases='5730552F22Rik, AI462154, NICE-3, NS5ATP4', description='RIKEN cDNA 4933434E20 gene', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90051636, end_position_on_the_genomic_accession=90063341, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  230726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230726, CurrentID=0, Status='live', Symbol='Rhbdl2', Aliases='9130416B15, RRP2', description='rhomboid, veinlet-like 2 (Drosophila)', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=123787875, end_position_on_the_genomic_accession=123830200, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  99662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99662, CurrentID=0, Status='live', Symbol='Eps8l3', Aliases='AI504489, Eps8r3', description='EPS8-like 3', other_designations='EPS8-like protein 3|EPS8-related protein 3|ESP8-like 3|epidermal growth factor receptor pathway substrate 8 related protein 3|epidermal growth factor receptor pathway substrate 8-related protein 3', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107873860, end_position_on_the_genomic_accession=107892900, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  99663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99663, CurrentID=0, Status='live', Symbol='Clca4a', Aliases='9130020L07Rik, AI504701, Clca4, Clca6', description='chloride channel accessory 4A', other_designations='calcium activated chloride channel|calcium-activated chloride channel regulator 4|calcium-activated chloride channel regulator 6|chloride channel calcium activated 4A|chloride channel calcium activated 6|mClca6', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144952486, end_position_on_the_genomic_accession=144975045, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  230737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230737, CurrentID=0, Status='live', Symbol='Gnl2', Aliases='BC003262, HUMAUANTIG, Ngp-1', description='guanine nucleotide binding protein-like 2 (nucleolar)', other_designations='testicular antigen', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=125029988, end_position_on_the_genomic_accession=125055382, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  230738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230738, CurrentID=0, Status='live', Symbol='Zc3h12a', Aliases='BC036563, MCPIP, Mcpip1', description='zinc finger CCCH type containing 12A', other_designations='MCP-1 treatment-induced protein|MCP-induced protein 1|zc3h12a|zinc finger CCCH domain-containing protein 12A', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=125118414, end_position_on_the_genomic_accession=125127881, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  230751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230751, CurrentID=0, Status='live', Symbol='Oscp1', Aliases='1810007P19Rik, 5730415O04, 6030436A01Rik', description='organic solute carrier partner 1', other_designations='organic solute carrier 1|organic solute carrier protein 1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126057287, end_position_on_the_genomic_accession=126089334, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  230752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230752, CurrentID=0, Status='live', Symbol='Eva1b', Aliases='2610027C15Rik, Fam176b', description='eva-1 homolog B (C. elegans)', other_designations='family with sequence similarity 176, member B|protein FAM176B', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126145704, end_position_on_the_genomic_accession=126149879, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  99681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99681, CurrentID=0, Status='live', Symbol='Tchh', Aliases='AHF, AI597080, Thh', description='trichohyalin', other_designations='hair follicle protein AHF', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93442330, end_position_on_the_genomic_accession=93449077, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  99683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99683, CurrentID=0, Status='live', Symbol='Sec24b', Aliases='AI605202, SEC24', description='Sec24 related gene family, member B (S. cerevisiae)', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129982764, end_position_on_the_genomic_accession=130061573, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  230757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230757, CurrentID=0, Status='live', Symbol='5730409E04Rik', Aliases='7530403E16Rik, AI849033', description='RIKEN cDNA 5730409E04Rik gene', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126609854, end_position_on_the_genomic_accession=126614371, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  492903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492903, CurrentID=0, Status='live', Symbol='Lmr9', Aliases='', description='leishmaniasis resistance 9', other_designations='', map_location='4 2.04 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230761, CurrentID=0, Status='live', Symbol='Zfp362', Aliases='', description='zinc finger protein 362', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128773085, end_position_on_the_genomic_accession=128806112, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  230766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230766, CurrentID=0, Status='live', Symbol='Fam167b', Aliases='SEC', description='family with sequence similarity 167, member B', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129576815, end_position_on_the_genomic_accession=129578580, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  492911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492911, CurrentID=0, Status='live', Symbol='Mptp1', Aliases='', description='MPTP sensitivity 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  99696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99696, CurrentID=0, Status='live', Symbol='Ankrd50', Aliases='AI662170, E430012K20Rik, mKIAA1223, mKIAA3019', description='ankyrin repeat domain 50', other_designations='ankrin repeat domain 50', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=38449261, end_position_on_the_genomic_accession=38484861, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  230770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230770, CurrentID=0, Status='live', Symbol='Tmem39b', Aliases='6330509E05Rik, BC037095', description='transmembrane protein 39b', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129676355, end_position_on_the_genomic_accession=129696851, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  230775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230775, CurrentID=0, Status='live', Symbol='Adgrb2', Aliases='Bai2', description='adhesion G protein-coupled receptor B2', other_designations='G protein-coupled receptor brain-specific angiogenesis inhibitor 1|brain-specific angiogenesis inhibitor 2', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129984870, end_position_on_the_genomic_accession=130022633, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  230777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230777, CurrentID=0, Status='live', Symbol='Hcrtr1', Aliases='Ox1r, ox-1-R, ox1-R', description='hypocretin (orexin) receptor 1', other_designations='hypocretin receptor type 1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130130217, end_position_on_the_genomic_accession=130139691, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  230779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230779, CurrentID=0, Status='live', Symbol='Serinc2', Aliases='2310004K20Rik, AW121759, FKSG84, TDE2, Tde2l', description='serine incorporator 2', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130253495, end_position_on_the_genomic_accession=130279200, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  492924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492924, CurrentID=0, Status='live', Symbol='Ssial3', Aliases='', description='susceptibility to sialadenitis 3', other_designations='', map_location='7 57.86 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  99709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99709, CurrentID=0, Status='live', Symbol='Clca4b', Aliases='AI747448, Clca7', description='chloride channel accessory 4B', other_designations='calcium activated chloride channel|chloride channel calcium activated 4B|chloride channel calcium activated 7', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=144910921, end_position_on_the_genomic_accession=144932529, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  230784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230784, CurrentID=0, Status='live', Symbol='Sesn2', Aliases='HI95, SEST2, Ses2', description='sestrin 2', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132492804, end_position_on_the_genomic_accession=132510456, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  230787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230787, CurrentID=0, Status='live', Symbol='Themis2', Aliases='ICB-1', description='thymocyte selection associated family member 2', other_designations='basement membrane-induced protein|induced by contact to basement membrane 1 protein|protein ICB-1|thymocyte-expressed molecule involved in selection protein 2', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132781837, end_position_on_the_genomic_accession=132796364, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  230789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230789, CurrentID=0, Status='live', Symbol='Fam76a', Aliases='', description='family with sequence similarity 76, member A', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132899213, end_position_on_the_genomic_accession=132923226, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  230793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230793, CurrentID=0, Status='live', Symbol='Ahdc1', Aliases='D030015G18Rik', description='AT hook, DNA binding motif, containing 1', other_designations='a.T hook DNA-binding motif-containing protein 1', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133011265, end_position_on_the_genomic_accession=133078110, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  230796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230796, CurrentID=0, Status='live', Symbol='Wdtc1', Aliases='Gm695, adipose, adp', description='WD and tetratricopeptide repeats 1', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133292466, end_position_on_the_genomic_accession=133353476, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  230801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230801, CurrentID=0, Status='live', Symbol='Pigv', Aliases='B330013B03, D430024F16Rik', description='phosphatidylinositol glycan anchor biosynthesis, class V', other_designations='GPI mannosyltransferase II|GPI-MT-II|PIG-V|phosphatidylinositol glycan class V|phosphatidylinositol glycan, class V|phosphatidylinositol-glycan biosynthesis class V protein', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133659914, end_position_on_the_genomic_accession=133673997, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  99730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99730, CurrentID=0, Status='live', Symbol='Taf13', Aliases='1810004N01Rik, 2010309N11Rik, AI847295, TAFII18', description='TATA-box binding protein associated factor 13', other_designations='TAF(II)18|TAF13 RNA polymerase II, TATA box binding protein (TBP)-associated factor|TAF13 RNA polymerase II, TATA box-binding protein-associated factor|TAFII-18|TBP-associated factor, RNA polymerase II, K, 18kD|transcription initiation factor TFIID 18 kDa subunit', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108571699, end_position_on_the_genomic_accession=108582068, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  492947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492947, CurrentID=0, Status='live', Symbol='Alcdp1', Aliases='', description='alcohol dependency 1', other_designations='', map_location='1 80.33 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230806, CurrentID=0, Status='live', Symbol='Aim1l', Aliases='', description='absent in melanoma 1-like', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134060728, end_position_on_the_genomic_accession=134092504, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  230809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230809, CurrentID=0, Status='live', Symbol='Pdik1l', Aliases='BC027088, stk35l2', description='PDLIM1 interacting kinase 1 like', other_designations='PDLIM1-interacting kinase 1-like', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134275002, end_position_on_the_genomic_accession=134287846, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  230810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230810, CurrentID=0, Status='live', Symbol='Slc30a2', Aliases='6030430C21, ZnT-2, Znt2', description='solute carrier family 30 (zinc transporter), member 2', other_designations='solute carrier family 30 member 2', map_location='4 66.61 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134342961, end_position_on_the_genomic_accession=134356528, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  230815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230815, CurrentID=0, Status='live', Symbol='Man1c1', Aliases='AI593348', description='mannosidase, alpha, class 1C, member 1', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134561690, end_position_on_the_genomic_accession=134704290, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  492961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492961, CurrentID=0, Status='live', Symbol='Gcsfis', Aliases='', description='G-CSF induced splenomegaly', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  492966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492966, CurrentID=0, Status='live', Symbol='Hdlq15', Aliases='', description='HDL QTL 15', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230824, CurrentID=0, Status='live', Symbol='Grhl3', Aliases='AI561912, Get1, Som, Tfcp2l4, ct', description='grainyhead-like 3 (Drosophila)', other_designations='grainy head-like 3|sister-of-mammalian grainyhead|transcription factor CP2-like 4|transcription factor hSOM1', map_location='4 67.76 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135541887, end_position_on_the_genomic_accession=135573628, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  230828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230828, CurrentID=0, Status='live', Symbol='Il22ra1', Aliases='9130219A07Rik, ENSMUSG00000073746, IL-22R, Il22r', description='interleukin 22 receptor, alpha 1', other_designations='IL-22 receptor subunit alpha-1|IL-22R-alpha-1|IL-22RA1|interleukin-22 receptor alpha chain', map_location='4 67.99 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135728159, end_position_on_the_genomic_accession=135755383, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  492979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492979, CurrentID=0, Status='live', Symbol='Igf1sl1', Aliases='', description='IGF-1 serum levels 1', other_designations='', map_location='6 53.75 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230837, CurrentID=0, Status='live', Symbol='Asap3', Aliases='9430088F20Rik, Ddefl1, Gm140, UPLC1', description='ArfGAP with SH3 domain, ankyrin repeat and PH domain 3', other_designations='development and differentiation enhancing factor-like 1|development and differentiation-enhancing factor-like 1', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136206349, end_position_on_the_genomic_accession=136246578, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  230848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230848, CurrentID=0, Status='live', Symbol='Zbtb40', Aliases='BC059177, C230087D24, Gm571, mKIAA0478', description='zinc finger and BTB domain containing 40', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136979732, end_position_on_the_genomic_accession=137049442, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  230857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230857, CurrentID=0, Status='live', Symbol='Ece1', Aliases='AW322500, ECE-1, ECE-1a, ECE-1b, ECE-1d', description='endothelin converting enzyme 1', other_designations='endothelin-converting enzyme-1a|endothelin-converting enzyme-1b|endothelin-converting enzyme-1d', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137862216, end_position_on_the_genomic_accession=137965229, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  230861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230861, CurrentID=0, Status='live', Symbol='Eif4g3', Aliases='1500002J22Rik, 4833436O05, 4930523M17Rik, G1-419-52, eIF4GII, repro8', description='eukaryotic translation initiation factor 4 gamma, 3', other_designations='eIF-4-gamma 3|eIF-4-gamma II|eIF-4G 3|eIF4G 3', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137992525, end_position_on_the_genomic_accession=138207081, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  230863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230863, CurrentID=0, Status='live', Symbol='Sh2d5', Aliases='BC036961', description='SH2 domain containing 5', other_designations='vascular SH2 domain-containing protein', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138250411, end_position_on_the_genomic_accession=138260968, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  230866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230866, CurrentID=0, Status='live', Symbol='Emc1', Aliases='2700016F22Rik', description='ER membrane protein complex subunit 1', other_designations='novel DUF1620 domain containing protein', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139352587, end_position_on_the_genomic_accession=139378735, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  230868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230868, CurrentID=0, Status='live', Symbol='Igsf21', Aliases='BC055811, Gm141', description='immunoglobulin superfamily, member 21', other_designations='immunoglobin superfamily, member 21', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140026852, end_position_on_the_genomic_accession=140246811, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  230872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230872, CurrentID=0, Status='live', Symbol='Crocc', Aliases='', description='ciliary rootlet coiled-coil, rootletin', other_designations='ciliary rootlet coiled-coil protein|crocc', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141016637, end_position_on_the_genomic_accession=141060615, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  493020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493020, CurrentID=0, Status='live', Symbol='Tlyr1', Aliases='', description='thymic lymphoma resistance 1', other_designations='', map_location='19 21.34 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  493025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493025, CurrentID=0, Status='live', Symbol='Nktcn1', Aliases='Nkt1', description='natural killer T cell numbers 1', other_designations='', map_location='1 76.25 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  493033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493033, CurrentID=0, Status='live', Symbol='Lith6', Aliases='', description='lithogenic gene 6', other_designations='', map_location='6 77.64 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  493036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493036, CurrentID=0, Status='live', Symbol='Ahl2', Aliases='', description='age related hearing loss 2', other_designations='', map_location='5 42.22 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230895, CurrentID=0, Status='live', Symbol='Vps13d', Aliases='', description='vacuolar protein sorting 13 D (yeast)', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144871045, end_position_on_the_genomic_accession=145195095, orientation='minus', exon_count=76, OMIM='', no_hdr0=''),
  230899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230899, CurrentID=0, Status='live', Symbol='Nppa', Aliases='ANP, Anf, Pnd', description='natriuretic peptide type A', other_designations='atrial natriuretic factor|atrial natriuretic peptide|natriuretic peptide precursor type A|prepronatriodilatin', map_location='4 78.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148000746, end_position_on_the_genomic_accession=148002067, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  230903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230903, CurrentID=0, Status='live', Symbol='Fbxo44', Aliases='5730411K09, AV001623, FBG3, FBX30, Fbx6a, Fbxo6a', description='F-box protein 44', other_designations='F-box only protein 6a', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148152799, end_position_on_the_genomic_accession=148160094, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  230904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230904, CurrentID=0, Status='live', Symbol='Fbxo2', Aliases='FBG1, FBX2, Fbs1, Fbs2, NFB42, Prpl4', description='F-box protein 2', other_designations='prion protein ligand 4', map_location='4 78.68 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148160226, end_position_on_the_genomic_accession=148166417, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  230908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230908, CurrentID=0, Status='live', Symbol='Tardbp', Aliases='1190002A23Rik, C85084, TDP-43, Tdp43', description='TAR DNA binding protein', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148612382, end_position_on_the_genomic_accession=148626996, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  230909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230909, CurrentID=0, Status='live', Symbol='Gm572', Aliases='b2b1167Clo', description='predicted gene 572', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148643317, end_position_on_the_genomic_accession=148676969, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  493058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493058, CurrentID=0, Status='live', Symbol='Heal10', Aliases='', description='wound healing/regeneration 10', other_designations='', map_location='11 58.9 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230917, CurrentID=0, Status='live', Symbol='Tmem201', Aliases='AV028368, D4Ertd429e, Net5, Samp1', description='transmembrane protein 201', other_designations='spindle-associated membrane protein 1', map_location='4 80.15 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149715375, end_position_on_the_genomic_accession=149738071, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  230935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230935, CurrentID=0, Status='live', Symbol='Dnajc11', Aliases='E030019A03Rik', description='DnaJ heat shock protein family (Hsp40) member C11', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 11', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=151933720, end_position_on_the_genomic_accession=151981959, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  230936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230936, CurrentID=0, Status='live', Symbol='Phf13', Aliases='BC029632, Phf5, SPOC1', description='PHD finger protein 13', other_designations='PHD zinc finger protein PHF5', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=151989631, end_position_on_the_genomic_accession=151996179, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  624153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624153, CurrentID=0, Status='live', Symbol='H2afb2', Aliases='EG624153, H2A.Bbd4-ps, H2afb2', description='H2A histone family, member B2', other_designations='H2A histone family, member B2, pseudogene', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=116681178, end_position_on_the_genomic_accession=116681525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  493083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493083, CurrentID=0, Status='live', Symbol='Scon2', Aliases='', description='sucrose consumption 2', other_designations='', map_location='4 82.64 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  493086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493086, CurrentID=0, Status='live', Symbol='Sysbp1', Aliases='', description='systolic blood pressure 1', other_designations='', map_location='10 34.83 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  99887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99887, CurrentID=0, Status='live', Symbol='Tmem56', Aliases='4930577M16Rik, AU018104, AW109744, C730036B01Rik', description='transmembrane protein 56', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=121202008, end_position_on_the_genomic_accession=121283108, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  99889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99889, CurrentID=0, Status='live', Symbol='Arfip1', Aliases='2310030I10Rik, Arfaptin, D3Ertd598e', description='ADP-ribosylation factor interacting protein 1', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84496093, end_position_on_the_genomic_accession=84582835, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  99890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99890, CurrentID=0, Status='live', Symbol='Prmt6', Aliases='AW124876, BB233495, Hrmt1l6', description='protein arginine N-methyltransferase 6', other_designations='HMT1 hnRNP methyltransferase-like 6|histone-arginine N-methyltransferase PRMT6', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=110246104, end_position_on_the_genomic_accession=110250999, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  230967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230967, CurrentID=0, Status='live', Symbol='Cep104', Aliases='A930027E11, AI115523, Kiaa0562, mKIAA0562', description='centrosomal protein 104', other_designations='glycine-, glutamate-, thienylcyclohexylpiperidine-binding protein|likely ortholog of H. sapiens glycine-, glutamate-, thienylcyclohexylpiperidine-binding protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=153975126, end_position_on_the_genomic_accession=154008732, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  230971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230971, CurrentID=0, Status='live', Symbol='Megf6', Aliases='2600001P17Rik, Egfl3', description='multiple EGF-like-domains 6', other_designations='EGF-like protein 3|EGF-like-domain, multiple 3|epidermal growth factor-like protein 3|multiple EGF-like domains protein 6', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154170713, end_position_on_the_genomic_accession=154275721, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  230972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230972, CurrentID=0, Status='live', Symbol='Arhgef16', Aliases='', description='Rho guanine nucleotide exchange factor (GEF) 16', other_designations='Neuroblastoma|ephexin-4', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154278463, end_position_on_the_genomic_accession=154301671, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  230979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230979, CurrentID=0, Status='live', Symbol='Tnfrsf14', Aliases='Atar, HveA, Hvem, Tnfrs14', description='tumor necrosis factor receptor superfamily, member 14 (herpesvirus entry mediator)', other_designations='herpes virus entry mediator', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154921442, end_position_on_the_genomic_accession=154928620, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  493131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493131, CurrentID=0, Status='live', Symbol='Hivan1', Aliases='', description='HIV-associated nephropathy 1', other_designations='', map_location='3 9.61 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  230991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230991, CurrentID=0, Status='live', Symbol='B930041F14Rik', Aliases='MNCb-3966', description='RIKEN cDNA B930041F14 gene', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155694342, end_position_on_the_genomic_accession=155696485, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  230996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230996, CurrentID=0, Status='live', Symbol='9430015G10Rik', Aliases='', description='RIKEN cDNA 9430015G10 gene', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156109998, end_position_on_the_genomic_accession=156127263, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  99929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99929, CurrentID=0, Status='live', Symbol='Tiparp', Aliases='ARTD14, AW558171, PARP7', description='TCDD-inducible poly(ADP-ribose) polymerase', other_designations='ADP-ribosyltransferase diphtheria toxin-like 14', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=65528447, end_position_on_the_genomic_accession=65555518, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  231002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231002, CurrentID=0, Status='live', Symbol='Plekhn1', Aliases='BC025458', description='pleckstrin homology domain containing, family N member 1', other_designations='PH domain-containing family N member 1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156221456, end_position_on_the_genomic_accession=156228542, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  624219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624219, CurrentID=0, Status='live', Symbol='Angptl8', Aliases='EG624219, Gm6484, Rifl', description='angiopoietin-like 8', other_designations='betatrophin|hepatocellular carcinoma-associated protein TD26 homolog|lipasin|refeeding-induced fat and liver protein', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21835510, end_position_on_the_genomic_accession=21837347, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  231004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231004, CurrentID=0, Status='live', Symbol='Samd11', Aliases='BC026991, mr-s', description='sterile alpha motif domain containing 11', other_designations='SAM domain-containing protein 11|major retinal SAM domain-containing protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156246966, end_position_on_the_genomic_accession=156255657, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  231014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231014, CurrentID=0, Status='live', Symbol='9330182L06Rik', Aliases='Eig121l', description='RIKEN cDNA 9330182L06 gene', other_designations='estrogen-induced gene 121-like protein', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=9266123, end_position_on_the_genomic_accession=9481827, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  231042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231042, CurrentID=0, Status='live', Symbol='Nupl2', Aliases='AV116043, CG1, NLP-1, Nlp1', description='nucleoporin like 2', other_designations='nucleoporin-like protein 1', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24164700, end_position_on_the_genomic_accession=24184008, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  231044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231044, CurrentID=0, Status='live', Symbol='Gbx1', Aliases='Gbx-1', description='gastrulation brain homeobox 1', other_designations='gastrulation and brain-specific homeobox protein 1', map_location='5 11.93 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24504426, end_position_on_the_genomic_accession=24526848, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  231045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231045, CurrentID=0, Status='live', Symbol='4931409K22Rik', Aliases='Iqca1l, Iqca1p1', description='RIKEN cDNA 4931409K22 gene', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24543434, end_position_on_the_genomic_accession=24555469, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  231050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231050, CurrentID=0, Status='live', Symbol='Galnt11', Aliases='A430075I06Rik, AI648252, E430002F06Rik', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 11', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 11|galNAc-T11|polypeptide GalNAc transferase 11|pp-GaNTase 11|protein-UDP acetylgalactosaminyltransferase 11', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=25222893, end_position_on_the_genomic_accession=25265918, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  231051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231051, CurrentID=0, Status='live', Symbol='Kmt2c', Aliases='E330008K23Rik, HALR, Mll3, mKIAA1506', description='lysine (K)-specific methyltransferase 2C', other_designations='histone-lysine N-methyltransferase MLL3|lysine N-methyltransferase 2C|myeloid/lymphoid or mixed-lineage leukemia 3|myeloid/lymphoid or mixed-lineage leukemia protein 3 homolog', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=25271794, end_position_on_the_genomic_accession=25498818, orientation='minus', exon_count=62, OMIM='', no_hdr0=''),
  99982 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=99982, CurrentID=0, Status='live', Symbol='Kdm1a', Aliases='1810043O07Rik, AA408884, Aof2, D4Ertd478e, Kdm1, Lsd1, mKIAA0601', description='lysine (K)-specific demethylase 1A', other_designations='BRAF35-HDAC complex protein BHC110|amine oxidase (flavin containing) domain 2|flavin-containing amine oxidase domain-containing protein 2|lysine (K)-specific demethylase 1|lysine-specific histone demethylase 1', map_location='4 68.8 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136550533, end_position_on_the_genomic_accession=136602789, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  231070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231070, CurrentID=0, Status='live', Symbol='Insig1', Aliases='1810013C12Rik, Insig-1', description='insulin induced gene 1', other_designations='INSIG-1 membrane protein', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=28071412, end_position_on_the_genomic_accession=28078662, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100012, CurrentID=0, Status='live', Symbol='Oog3', Aliases='AA617276, C86881', description='oogenesin 3', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144157556, end_position_on_the_genomic_accession=144162651, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  231086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231086, CurrentID=0, Status='live', Symbol='Hadhb', Aliases='4930479F15Rik, Mtpb, TP-beta', description='hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), beta subunit', other_designations='hydroxyacyl-Coenzyme A dehydrogenase/3-ketoacyl-Coenzyme A thiolase/enoyl-Coenzyme A hydratase (trifunctional protein), beta aubunit', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30155253, end_position_on_the_genomic_accession=30184593, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100017, CurrentID=0, Status='live', Symbol='Ldlrap1', Aliases='AA691260, ARH2, Arh, Arh1, FHCB1, FHCB2', description='low density lipoprotein receptor adaptor protein 1', other_designations='LDL receptor adaptor protein|autosomal recessive hypercholesterolemia protein homolog', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134741554, end_position_on_the_genomic_accession=134768004, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100019, CurrentID=0, Status='live', Symbol='Mdn1', Aliases='4833432B22Rik, A130070M06, AA958993, D4Abb1e, Gm135', description='midasin AAA ATPase 1', other_designations='MIDAS-containing protein|midasin homolog', map_location='4 14.3 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=32657112, end_position_on_the_genomic_accession=32775217, orientation='plus', exon_count=102, OMIM='', no_hdr0=''),
  231093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231093, CurrentID=0, Status='live', Symbol='Agbl5', Aliases='4930455N08, 9430057O19Rik, CCP5', description='ATP/GTP binding protein-like 5', other_designations='ATP/GTP-binding protein-like 5|carboxypeptidase 5, cytosolic', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30888476, end_position_on_the_genomic_accession=30906666, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  231098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231098, CurrentID=0, Status='live', Symbol='Dnajc5g', Aliases='4922503E23Rik', description='DnaJ heat shock protein family (Hsp40) member C5 gamma', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 5 gamma', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31108292, end_position_on_the_genomic_accession=31112526, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  231103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231103, CurrentID=0, Status='live', Symbol='Gckr', Aliases='GKRP', description='glucokinase regulatory protein', other_designations='glucokinase regulator', map_location='5 17.27 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31297254, end_position_on_the_genomic_accession=31327302, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  231123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231123, CurrentID=0, Status='live', Symbol='Haus3', Aliases='D4S43h, D5H4S43, D5H4S43E', description='HAUS augmin-like complex, subunit 3', other_designations='', map_location='5 17.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34153896, end_position_on_the_genomic_accession=34169424, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  231125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231125, CurrentID=0, Status='live', Symbol='Zfyve28', Aliases='9630058O20Rik, Gm146, mKIAA1643', description='zinc finger, FYVE domain containing 28', other_designations='zinc finger FYVE domain-containing protein 28', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34194893, end_position_on_the_genomic_accession=34288469, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  231128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231128, CurrentID=0, Status='live', Symbol='Fam193a', Aliases='', description='family with sequence similarity 193, member A', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34369933, end_position_on_the_genomic_accession=34486458, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  231130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231130, CurrentID=0, Status='live', Symbol='Tnip2', Aliases='1810020H16Rik, ABIN-2, AI428870', description='TNFAIP3 interacting protein 2', other_designations='A20 binding inhibitor of NF-kappaB activation-2|A20-binding inhibitor of NF-kappa-B activation 2|tnip2', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=34496096, end_position_on_the_genomic_accession=34513979, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100061, CurrentID=0, Status='live', Symbol='Lrrc19', Aliases='9130022A01Rik, AI314124, AL022954, AW261791', description='leucine rich repeat containing 19', other_designations='', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=94636154, end_position_on_the_genomic_accession=94650144, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231134, CurrentID=0, Status='live', Symbol='Dok7', Aliases='A930013K19Rik, AW049091, Dok-7, EF-12, Oit5', description='docking protein 7', other_designations='downstream of tyrosine kinase 7|oncoprotein induced transcript 5', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35056566, end_position_on_the_genomic_accession=35087839, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100066, CurrentID=0, Status='live', Symbol='Cyp2j11', Aliases='AI314783-ps, Cyp2j11', description='cytochrome P450, family 2, subfamily j, polypeptide 11', other_designations='cytochrome P450, 2j11|cytochrome P450, family 2, subfamily j, polypeptide 11, pseudogene', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=96294508, end_position_on_the_genomic_accession=96348697, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100072, CurrentID=0, Status='live', Symbol='Camta1', Aliases='1810059M14Rik, 2310058O09Rik, AI316882, AW987475, mKIAA0833', description='calmodulin binding transcription activator 1', other_designations='camta1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=151059523, end_position_on_the_genomic_accession=151861768, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  231147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231147, CurrentID=0, Status='live', Symbol='Sh3tc1', Aliases='4631428G15, BC024909', description='SH3 domain and tetratricopeptide repeats 1', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35697173, end_position_on_the_genomic_accession=35740080, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  231148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231148, CurrentID=0, Status='live', Symbol='Ablim2', Aliases='AI606905, C230091L11', description='actin-binding LIM protein 2', other_designations='abLIM-2|actin-binding LIM protein family member 2', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35757855, end_position_on_the_genomic_accession=35884979, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  231151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231151, CurrentID=0, Status='live', Symbol='Tada2b', Aliases='AA474455, Gm145', description='transcriptional adaptor 2B', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36473670, end_position_on_the_genomic_accession=36484285, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100087, CurrentID=0, Status='live', Symbol='Kti12', Aliases='1110001A12Rik, AI326810, AU014758', description='KTI12 homolog, chromatin associated', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108847857, end_position_on_the_genomic_accession=108849412, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100088, CurrentID=0, Status='live', Symbol='Rcc1', Aliases='4931417M11Rik, AI326872, Chc1', description='regulator of chromosome condensation 1', other_designations='chromosome condensation protein 1', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132331919, end_position_on_the_genomic_accession=132345750, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  231162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231162, CurrentID=0, Status='live', Symbol='Cytl1', Aliases='4930443F05Rik, 4Cytl1, C17, Cyt1, Gm147', description='cytokine-like 1', other_designations='cytokine like 1', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37735519, end_position_on_the_genomic_accession=37739820, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100102, CurrentID=0, Status='live', Symbol='Pcsk9', Aliases='AI415265, AI747682, FH3, HCHOLA3, Narc1, PC9', description='proprotein convertase subtilisin/kexin type 9', other_designations='NARC-1|convertase subtilisin|neural apoptosis regulated convertase 1|neural apoptosis-regulated convertase 1|proprotein convertase 9|proprotein convertase PC9|subtilisin/kexin-like protease PC9', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106442334, end_position_on_the_genomic_accession=106464325, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  100121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100121, CurrentID=0, Status='live', Symbol='Tdrd7', Aliases='5730495N10Rik, AI447470, PCTAIRE2BP', description='tudor domain containing 7', other_designations='PCTAIRE2-binding protein|trap|tudor repeat associator with PCTAIRE 2', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45965027, end_position_on_the_genomic_accession=46034766, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  100129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100129, CurrentID=0, Status='live', Symbol='Gpr153', Aliases='1110065N12Rik, AI449320, PGR1', description='G protein-coupled receptor 153', other_designations='G protein-coupled receptor PGR1|G-protein coupled receptor PGR1', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152273938, end_position_on_the_genomic_accession=152285343, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  231207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231207, CurrentID=0, Status='live', Symbol='Cpeb2', Aliases='A630055H10Rik, Cpe-bp2', description='cytoplasmic polyadenylation element binding protein 2', other_designations='CPE-binding protein 2', map_location='5 B', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43232254, end_position_on_the_genomic_accession=43289724, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  231214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231214, CurrentID=0, Status='live', Symbol='Cc2d2a', Aliases='5730509K17Rik, b2b1035Clo', description='coiled-coil and C2 domain containing 2A', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43662373, end_position_on_the_genomic_accession=43740972, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  624439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624439, CurrentID=0, Status='live', Symbol='Scgb2b15', Aliases='Abpb15, EG624439, Gm6504', description='secretoglobin, family 2B, member 15', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=32827683, end_position_on_the_genomic_accession=32829653, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  231225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231225, CurrentID=0, Status='live', Symbol='Tapt1', Aliases='4932414K18Rik', description='transmembrane anterior posterior transformation 1', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=44175156, end_position_on_the_genomic_accession=44229281, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100155, CurrentID=0, Status='live', Symbol='AI481877', Aliases='Gm426', description='expressed sequence AI481877', other_designations='likely orthologue of H. sapiens chromosome 9 open reading frame 84', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59041577, end_position_on_the_genomic_accession=59117809, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  100163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100163, CurrentID=0, Status='live', Symbol='Pafah2', Aliases='2310074E22Rik, AI507170', description='platelet-activating factor acetylhydrolase 2', other_designations='SD-PLA2|serine-dependent phospholipase A2', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=134396320, end_position_on_the_genomic_accession=134427412, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  231238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231238, CurrentID=0, Status='live', Symbol='Sel1l3', Aliases='2310045A20Rik, AI429585, mKIAA0746', description='sel-1 suppressor of lin-12-like 3 (C. elegans)', other_designations='sel-1L3|suppressor of lin-12-like protein 3', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=53107083, end_position_on_the_genomic_accession=53213452, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  100169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100169, CurrentID=0, Status='live', Symbol='Phactr4', Aliases='3110001B12Rik, AI527228, AW495572, C330013F19Rik, N28169, mKIAA4120', description='phosphatase and actin regulator 4', other_designations='humdy|protein Humpty dumpty', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132355922, end_position_on_the_genomic_accession=132422446, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  100177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100177, CurrentID=0, Status='live', Symbol='Zmym6', Aliases='9330177P20Rik, AI593204, AI661340, D4Wsu24e, Zfp258', description='zinc finger, MYM-type 6', other_designations='likely ortholog of H. sapiens zinc finger protein 258 (ZNF258)|probably fused to an N terminal zinc finger domain|transposase-like protein|zinc finger protein 258', map_location='4 61.25 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127077383, end_position_on_the_genomic_accession=127124374, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  231252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231252, CurrentID=0, Status='live', Symbol='Chrna9', Aliases='2410015I05Rik, Acra9, EG666827, Gm8311', description='cholinergic receptor, nicotinic, alpha polypeptide 9', other_designations='', map_location='5 33.84 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=65934764, end_position_on_the_genomic_accession=65977560, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100182, CurrentID=0, Status='live', Symbol='Akna', Aliases='AI597013', description='AT-hook transcription factor', other_designations='', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63367127, end_position_on_the_genomic_accession=63410758, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  100198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100198, CurrentID=0, Status='live', Symbol='H6pd', Aliases='AI785303, G6pd1, Gpd-1, Gpd1', description='hexose-6-phosphate dehydrogenase (glucose 1-dehydrogenase)', other_designations='glucose 1-dehydrogenase|glucose-6-phosphate dehydrogenase 1', map_location='4 80.65 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149979474, end_position_on_the_genomic_accession=150009023, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100201, CurrentID=0, Status='live', Symbol='Tmem64', Aliases='9630015D15Rik, AI790744, AV300874', description='transmembrane protein 64', other_designations='CG11367 gene product', map_location='4 A2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=15265820, end_position_on_the_genomic_accession=15286753, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100206, CurrentID=0, Status='live', Symbol='Adprhl2', Aliases='AI836109, Arh3', description='ADP-ribosylhydrolase like 2', other_designations='ADP-ribosylhydrolase 2|ADP-ribosylhydrolase 3|[Protein ADP-ribosylarginine] hydrolase-like protein 2|ecto ADP-ribosylhydrolase', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126316351, end_position_on_the_genomic_accession=126321703, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231279, CurrentID=0, Status='live', Symbol='Guf1', Aliases='4631409J12, AA407526, EF-4', description='GUF1 homolog, GTPase', other_designations='GTP-binding protein GUF1 homolog|GUF1 GTPase homolog|elongation factor 4 homolog|ribosomal back-translocase', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=69556908, end_position_on_the_genomic_accession=69573630, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100210, CurrentID=0, Status='live', Symbol='Gpn2', Aliases='AI838661, Atpbd1b, R74630', description='GPN-loop GTPase 2', other_designations='ATP binding domain 1 family, member B', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133584373, end_position_on_the_genomic_accession=133591735, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100213, CurrentID=0, Status='live', Symbol='Rusc2', Aliases='AI840675, mKIAA0375', description='RUN and SH3 domain containing 2', other_designations='RUN and SH3 domain-containing protein 2|interacting protein of Rab1', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43381982, end_position_on_the_genomic_accession=43427092, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  231287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231287, CurrentID=0, Status='live', Symbol='Atp10d', Aliases='9830145H18Rik, ATPVD, D5Buc24e', description='ATPase, class V, type 10D', other_designations='P4-ATPase flippase complex alpha subunit ATP10D|putative type IV aminophospholipid transporting ATPase', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72203329, end_position_on_the_genomic_accession=72298771, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  231290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231290, CurrentID=0, Status='live', Symbol='Slc10a4', Aliases='E130304D01', description='solute carrier family 10 (sodium/bile acid cotransporter family), member 4', other_designations='Na(+)/bile acid cotransporter 4|solute carrier family 10 member 4', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73006904, end_position_on_the_genomic_accession=73012955, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  231293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231293, CurrentID=0, Status='live', Symbol='Cwh43', Aliases='C130090K23Rik, imds-60', description='cell wall biogenesis 43 C-terminal homolog', other_designations='cell wall biogenesis protein 43 C-terminal homolog', map_location='5 C3.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73406051, end_position_on_the_genomic_accession=73453435, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  231296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231296, CurrentID=0, Status='live', Symbol='Lrrc66', Aliases='', description='leucine rich repeat containing 66', other_designations='', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73606642, end_position_on_the_genomic_accession=73632421, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100226, CurrentID=0, Status='live', Symbol='Stx12', Aliases='AI850350, AU041521, Stx13', description='syntaxin 12', other_designations='syntaxin 13', map_location='4 66.01 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132854064, end_position_on_the_genomic_accession=132884458, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  330677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330677, CurrentID=0, Status='live', Symbol='Nctc1', Aliases='9530074L01, Rhit1', description='non-coding transcript 1', other_designations='', map_location='7 87.96 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142544609, end_position_on_the_genomic_accession=142558598, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  231326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231326, CurrentID=0, Status='live', Symbol='Aasdh', Aliases='A230062G08Rik, A830035E16, Acsf4, U26', description='aminoadipate-semialdehyde dehydrogenase', other_designations='2-aminoadipic 6-semialdehyde dehydrogenase|protein LYS2 homolog|putative aminoadipate-semialdehyde dehydrogenase', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76873659, end_position_on_the_genomic_accession=76905518, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  231327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231327, CurrentID=0, Status='live', Symbol='Ppat', Aliases='5730454C12Rik, AA408689, AA675351, AI507113, C79945', description='phosphoribosyl pyrophosphate amidotransferase', other_designations='', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76913249, end_position_on_the_genomic_accession=76951578, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  231329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231329, CurrentID=0, Status='live', Symbol='Polr2b', Aliases='Pol2rb, Rpb140, Rpb2', description='polymerase (RNA) II (DNA directed) polypeptide B', other_designations='DNA-directed RNA polymerase II 140 kDa polypeptide|DNA-directed RNA polymerase II subunit B|RNA polymerase II subunit 2|RNA polymerase II subunit B2', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77310484, end_position_on_the_genomic_accession=77349328, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  100273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100273, CurrentID=0, Status='live', Symbol='Osbpl9', Aliases='2600011I06Rik, AU015843, Orp-9', description='oxysterol binding protein-like 9', other_designations='OSBP-related protein 9', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109061145, end_position_on_the_genomic_accession=109202272, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  624584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624584, CurrentID=0, Status='live', Symbol='Scgb2b10', Aliases='Abpbg10, EG624584, Gm6513', description='secretoglobin, family 2B, member 10', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=32091996, end_position_on_the_genomic_accession=32093939, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  231380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231380, CurrentID=0, Status='live', Symbol='Uba6', Aliases='4930542H01, 5730469D23Rik, AU021846, AW124799, E1-L2, Ube1l2', description='ubiquitin-like modifier activating enzyme 6', other_designations='ubiquitin-activating enzyme 6|ubiquitin-activating enzyme E1-like 2|ubiquitin-activating enzyme E1-like protein 2', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86110720, end_position_on_the_genomic_accession=86172743, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  231382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231382, CurrentID=0, Status='live', Symbol='Tmprss11d', Aliases='AST, AsP, BC020151', description='transmembrane protease, serine 11d', other_designations='AT|adrenal secretory serine protease|airway spesific trypsin-like protease|airway trypsin-like protease', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86302854, end_position_on_the_genomic_accession=86373387, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  231386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231386, CurrentID=0, Status='live', Symbol='Ythdc1', Aliases='A730098D12Rik, C80342, mKIAA1966', description='YTH domain containing 1', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86803980, end_position_on_the_genomic_accession=86836659, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  100317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100317, CurrentID=0, Status='live', Symbol='AU040320', Aliases='A730047D20Rik, Kiaa0319l', description='expressed sequence AU040320', other_designations='likely ortholog of H. sapiens polycystic kidney disease 1-like (PKD1-like)', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126753555, end_position_on_the_genomic_accession=126869694, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  231396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231396, CurrentID=0, Status='live', Symbol='Ugt2b36', Aliases='', description='UDP glucuronosyltransferase 2 family, polypeptide B36', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87065927, end_position_on_the_genomic_accession=87092555, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100336, CurrentID=0, Status='live', Symbol='Ppp1r8', Aliases='6330548N22Rik, AU044684, NIPP1', description='protein phosphatase 1, regulatory (inhibitor) subunit 8', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132826924, end_position_on_the_genomic_accession=132843169, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100340, CurrentID=0, Status='live', Symbol='Smpdl3b', Aliases='1110054A24Rik, AU045240, Asml3b', description='sphingomyelin phosphodiesterase, acid-like 3B', other_designations='ASM-like phosphodiesterase 3b', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132732966, end_position_on_the_genomic_accession=132757171, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  231413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231413, CurrentID=0, Status='live', Symbol='Grsf1', Aliases='BB232551, C80280, D5Wsu31e', description='G-rich RNA sequence binding factor 1', other_designations='', map_location='5 43.85 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=88659233, end_position_on_the_genomic_accession=88676171, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100342, CurrentID=0, Status='live', Symbol='Fam46b', Aliases='4732473B16Rik, AU045678, AU046069', description='family with sequence similarity 46, member B', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133480133, end_position_on_the_genomic_accession=133487940, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  231430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231430, CurrentID=0, Status='live', Symbol='Cox18', Aliases='BC038311', description='cytochrome c oxidase assembly protein 18', other_designations='COX18 cytochrome c oxidase assembly homolog', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=90214724, end_position_on_the_genomic_accession=90223996, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  493583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493583, CurrentID=0, Status='live', Symbol='Itlnb', Aliases='Itln1b, Itln2', description='intelectin b', other_designations='intelectin-2', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  231440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231440, CurrentID=0, Status='live', Symbol='Parm1', Aliases='2210012L08Rik, 9130213B05Rik, W91666', description='prostate androgen-regulated mucin-like protein 1', other_designations='PARM-1|prostatic androgen-repressed message-1', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=91517700, end_position_on_the_genomic_accession=91624011, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  231452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231452, CurrentID=0, Status='live', Symbol='Sdad1', Aliases='4931421J16, AA591032, AW538460', description='SDA1 domain containing 1', other_designations='SDA1 domain-containing protein 1', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92284010, end_position_on_the_genomic_accession=92310024, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  100383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100383, CurrentID=0, Status='live', Symbol='Bsdc1', Aliases='1110063F24Rik, AW011758', description='BSD domain containing 1', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129461679, end_position_on_the_genomic_accession=129488498, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  104112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104112, CurrentID=0, Status='live', Symbol='Acly', Aliases='A730098H14Rik, AW538652', description='ATP citrate lyase', other_designations='ATP-citrate (pro-S-)-lyase|citrate cleavage enzyme', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100476352, end_position_on_the_genomic_accession=100528000, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  231464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231464, CurrentID=0, Status='live', Symbol='Cnot6l', Aliases='4932442K20Rik', description='CCR4-NOT transcription complex, subunit 6-like', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=96070333, end_position_on_the_genomic_accession=96164313, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  231470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231470, CurrentID=0, Status='live', Symbol='Fras1', Aliases='E130113P14Rik, bl, mKIAA1500', description='Fraser extracellular matrix complex subunit 1', other_designations='Fraser syndrome 1 homolog|Fraser syndrome 1 protein|blebbed', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=96373825, end_position_on_the_genomic_accession=96784732, orientation='plus', exon_count=74, OMIM='', no_hdr0=''),
  231474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231474, CurrentID=0, Status='live', Symbol='Paqr3', Aliases='6330415A20Rik, AY424292, RKTG', description='progestin and adipoQ receptor family member III', other_designations='Raf kinase trapping to Golgi', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=97082329, end_position_on_the_genomic_accession=97111748, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  231503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231503, CurrentID=0, Status='live', Symbol='Tmem150c', Aliases='', description='transmembrane protein 150C', other_designations='', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100077872, end_position_on_the_genomic_accession=100165604, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100434, CurrentID=0, Status='live', Symbol='Slc44a1', Aliases='2210409B22Rik, 4833416H08Rik, AW547365, CHTL1, CTL1, Cdw92', description='solute carrier family 44, member 1', other_designations='CDW92 antigen', map_location='4 B2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=53440413, end_position_on_the_genomic_accession=53622478, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  231507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231507, CurrentID=0, Status='live', Symbol='Plac8', Aliases='C15, D5Wsu111e', description='placenta-specific 8', other_designations='onzin', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100553726, end_position_on_the_genomic_accession=100572205, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231510, CurrentID=0, Status='live', Symbol='Agpat9', Aliases='1-AGPAT, 1-AGPAT 9, 4933407I02Rik, 4933408F15, A230097K15Rik, GPAT-3, Gpat3, mGPAT3', description='1-acylglycerol-3-phosphate O-acyltransferase 9', other_designations='1-AGP acyltransferase 9|1-acyl-sn-glycerol-3-phosphate O-acyltransferase 9|LPAAT-theta|acyl-CoA:glycerol-3-phosphate acyltransferase 3|lysophosphatidic acid acyltransferase theta', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100845710, end_position_on_the_genomic_accession=100899102, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  231532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231532, CurrentID=0, Status='live', Symbol='Arhgap24', Aliases='0610025G21Rik', description='Rho GTPase activating protein 24', other_designations='rho-type GTPase-activating protein 24', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=102207198, end_position_on_the_genomic_accession=102897937, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100465, CurrentID=0, Status='live', Symbol='Mob3c', Aliases='AW822253, D130076I06Rik, Mobkl2c', description='MOB kinase activator 3C', other_designations='MOB1, Mps One Binder kinase activator-like 2C|mob1 homolog 2C|mps one binder kinase activator-like 2C', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115818756, end_position_on_the_genomic_accession=115836185, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100470, CurrentID=0, Status='live', Symbol='Lao1', Aliases='AW990848', description='L-amino acid oxidase 1', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118961967, end_position_on_the_genomic_accession=118968910, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  231549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231549, CurrentID=0, Status='live', Symbol='Lrrc8d', Aliases='2810473G09Rik, 4930525N13Rik, A930019F03, Lrrc5', description='leucine rich repeat containing 8D', other_designations='leucine rich repeat containing 5|leucine-rich repeat-containing 5|leucine-rich repeat-containing protein 5|leucine-rich repeat-containing protein 8D', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105699969, end_position_on_the_genomic_accession=105815215, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100494, CurrentID=0, Status='live', Symbol='Zfand2a', Aliases='AA407930, AU016206, Airap', description='zinc finger, AN1-type domain 2A', other_designations='arsenite inducible RNA associated protein|arsenite-inducible RNA-associated protein', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139471216, end_position_on_the_genomic_accession=139484491, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  231570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231570, CurrentID=0, Status='live', Symbol='A830010M20Rik', Aliases='AI849226, Kiaa1107, mKIAA1107', description='RIKEN cDNA A830010M20 gene', other_designations='', map_location='5 E5-F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107497745, end_position_on_the_genomic_accession=107512556, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  231571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231571, CurrentID=0, Status='live', Symbol='Rpap2', Aliases='AW060207', description='RNA polymerase II associated protein 2', other_designations='RNA polymerase II-associated protein 2', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107597373, end_position_on_the_genomic_accession=107661853, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  231580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231580, CurrentID=0, Status='live', Symbol='Gak', Aliases='D130045N16Rik', description='cyclin G associated kinase', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108569107, end_position_on_the_genomic_accession=108629777, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  231583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231583, CurrentID=0, Status='live', Symbol='Slc26a1', Aliases='Sat1', description='solute carrier family 26 (sulfate transporter), member 1', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108669970, end_position_on_the_genomic_accession=108678097, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  100515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100515, CurrentID=0, Status='live', Symbol='Zfp518b', Aliases='6820424L24Rik, AA410078, AI661722, Znf518b', description='zinc finger protein 518B', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38668484, end_position_on_the_genomic_accession=38684826, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  231600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231600, CurrentID=0, Status='live', Symbol='Chfr', Aliases='5730484M20Rik, C230082M18, RNF116', description='checkpoint with forkhead and ring finger domains', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110135759, end_position_on_the_genomic_accession=110171973, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  231602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231602, CurrentID=0, Status='live', Symbol='P2rx2', Aliases='P2X2a, P2x2', description='purinergic receptor P2X, ligand-gated ion channel, 2', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110339812, end_position_on_the_genomic_accession=110343446, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100532, CurrentID=0, Status='live', Symbol='Rell1', Aliases='AA536743', description='RELT-like 1', other_designations='', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=63888667, end_position_on_the_genomic_accession=63968900, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  231605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231605, CurrentID=0, Status='live', Symbol='Galnt9', Aliases='GalNAc-T9', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 9', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110544350, end_position_on_the_genomic_accession=110621383, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  100535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100535, CurrentID=0, Status='live', Symbol='Oas1d', Aliases='AA536864, C85127, oasl8', description="2'-5' oligoadenylate synthetase 1D", other_designations="2',5'-oligoadenylate synthetase-like 8|2'-5'olygoadenylate synthetase 1D", map_location='5 60.87 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120914818, end_position_on_the_genomic_accession=120921648, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100554, CurrentID=0, Status='live', Symbol='AA792892', Aliases='', description='expressed sequence AA792892', other_designations='', map_location='5 E3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=94339763, end_position_on_the_genomic_accession=94384334, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  231630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231630, CurrentID=0, Status='live', Symbol='Ficd', Aliases='D5Ertd40e, Hype', description='FIC domain containing', other_designations='AMPylator FICD|FIC domain-containing protein|huntingtin interacting protein E|huntingtin interactor protein E', map_location='5 55.55 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113735677, end_position_on_the_genomic_accession=113740607, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100559, CurrentID=0, Status='live', Symbol='Ugt2b38', Aliases='9430041C03Rik, AA986709', description='UDP glucuronosyltransferase 2 family, polypeptide B38', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87409939, end_position_on_the_genomic_accession=87424203, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100561, CurrentID=0, Status='live', Symbol='Slc15a4', Aliases='AA987064, AW742963, C130069N12Rik, PHT1, PTR4', description='solute carrier family 15, member 4', other_designations='peptide transporter 4|peptide-histidine transporter 4|peptide/histidine transporter 1', map_location='5 G1.2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=127595664, end_position_on_the_genomic_accession=127632907, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  231637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231637, CurrentID=0, Status='live', Symbol='Ssh1', Aliases='AW551225, Gm1394, Gm1395, SSH-1, mSSH-1L', description='slingshot homolog 1 (Drosophila)', other_designations='SSH-1L|SSH-like protein 1|slingshot 1L|slingshot-like 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113937096, end_position_on_the_genomic_accession=113993757, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  624855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624855, CurrentID=0, Status='live', Symbol='Cks1brt', Aliases='EG624855, Gm6531', description='CDC28 protein kinase 1b, retrogene', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85151924, end_position_on_the_genomic_accession=85173622, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  231642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231642, CurrentID=0, Status='live', Symbol='Alkbh2', Aliases='9530023G02, AU016977, Abh2, mABH2', description='alkB homolog 2, alpha-ketoglutarate-dependent dioxygenase', other_designations='alkB, alkylation repair homolog 2|alkylated DNA repair protein alkB homolog 2|alpha-ketoglutarate-dependent dioxygenase alkB homolog 2', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114123926, end_position_on_the_genomic_accession=114130546, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231646, CurrentID=0, Status='live', Symbol='Myo1h', Aliases='4631401O15Rik', description='myosin 1H', other_designations='myosin-Ih', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114288899, end_position_on_the_genomic_accession=114364576, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  624866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624866, CurrentID=0, Status='live', Symbol='Lekr1', Aliases='EG546798, EG624866, Gm6534', description='leucine, glutamate and lysine rich 1', other_designations='', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=65659539, end_position_on_the_genomic_accession=65832480, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  231655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231655, CurrentID=0, Status='live', Symbol='Oasl1', Aliases='7530414C13Rik, oasl9', description="2'-5' oligoadenylate synthetase-like 1", other_designations="2',5'-oligoadenylate synthetase-like 9|2,5-oligoadenylate synthetase-like 9", map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114923190, end_position_on_the_genomic_accession=114937912, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  231659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231659, CurrentID=0, Status='live', Symbol='Gcn1l1', Aliases='4932409G22, AL022764, G431004K08Rik, GCN1L, mKIAA0219', description='GCN1 general control of amino-acid synthesis 1-like 1 (yeast)', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115565219, end_position_on_the_genomic_accession=115622654, orientation='plus', exon_count=58, OMIM='', no_hdr0=''),
  493809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493809, CurrentID=0, Status='live', Symbol='Taar3', Aliases='', description='trace amine-associated receptor 3', other_designations='mTaar3|taR-3|trace amine associated receptor 3|trace amine receptor 3', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23949558, end_position_on_the_genomic_accession=23950589, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  231668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231668, CurrentID=0, Status='live', Symbol='Vsig10', Aliases='', description='V-set and immunoglobulin domain containing 10', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117318596, end_position_on_the_genomic_accession=117355006, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  231670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231670, CurrentID=0, Status='live', Symbol='Fbxo21', Aliases='2810425J22Rik, AU016673, mKIAA0875', description='F-box protein 21', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=117976735, end_position_on_the_genomic_accession=118010201, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  231672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231672, CurrentID=0, Status='live', Symbol='Fbxw8', Aliases='4930438M06Rik, FBW6, FBW8, FBXO29, Fbx29', description='F-box and WD-40 domain protein 8', other_designations='F-box and WD-40 domain-containing protein 8|F-box protein FBX29', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=118064981, end_position_on_the_genomic_accession=118155458, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100604, CurrentID=0, Status='live', Symbol='Lrrc8c', Aliases='AD158, AI326115, E430036I04Rik, fad158', description='leucine rich repeat containing 8 family, member C', other_designations='factor for adipocyte differentiation 158|leucine-rich repeat-containing protein 8C', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105519386, end_position_on_the_genomic_accession=105613018, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100608, CurrentID=0, Status='live', Symbol='Noc4l', Aliases='AI326906', description='nucleolar complex associated 4 homolog (S. cerevisiae)', other_designations='NOC4 protein homolog|NOC4-like protein|nucleolar complex-associated protein 4-like protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110648419, end_position_on_the_genomic_accession=110653382, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100609, CurrentID=0, Status='live', Symbol='Nsun5', Aliases='9830109N13Rik, AI326939, Nol1r, Wbscr20, Wbscr20a', description='NOL1/NOP2/Sun domain family, member 5', other_designations='Williams Beuren syndrome chromosome region 20A homolog|Williams-Beuren syndrome critical region protein 20|putative methyltransferase NSUN5|williams-Beuren syndrome chromosomal region 20A protein homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135369930, end_position_on_the_genomic_accession=135376805, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  231691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231691, CurrentID=0, Status='live', Symbol='Sds', Aliases='4432411H13Rik, SDH', description='serine dehydratase', other_designations='L-serine deaminase|L-threonine dehydratase|TDH', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120476547, end_position_on_the_genomic_accession=120484935, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  624910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624910, CurrentID=0, Status='live', Symbol='Rgs21', Aliases='EG624910', description='regulator of G-protein signalling 21', other_designations='', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=144519690, end_position_on_the_genomic_accession=144567667, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  231699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231699, CurrentID=0, Status='live', Symbol='Oas1e', Aliases='C87484, oasl7', description="2'-5' oligoadenylate synthetase 1E", other_designations="2'-5' oligoadenylate synthetase-like 7|2'-5'olygoadenylate synthetase 1E|2,5-oligoadenylate synthetase-like 7", map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120786312, end_position_on_the_genomic_accession=120795530, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100637, CurrentID=0, Status='live', Symbol='N4bp2l1', Aliases='2410024N18Rik, AI428195, B230342M21Rik', description='NEDD4 binding protein 2-like 1', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=150571643, end_position_on_the_genomic_accession=150594525, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231712, CurrentID=0, Status='live', Symbol='Trafd1', Aliases='1110008K06Rik, Fln29', description='TRAF type zinc finger domain containing 1', other_designations='FLN29 gene product', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121371725, end_position_on_the_genomic_accession=121386480, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  231713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231713, CurrentID=0, Status='live', Symbol='Naa25', Aliases='4833422K13Rik, A630046C11, AI450868, C330023M02Rik, Mdm20', description='N(alpha)-acetyltransferase 25, NatB auxiliary subunit', other_designations='N-terminal acetyltransferase B complex subunit MDM20|N-terminal acetyltransferase B complex subunit NAA25|Nalpha acteyltransferase 25|mitochondrial distribution and morphology protein 20|natB complex subunit MDM20', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121397941, end_position_on_the_genomic_accession=121442549, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  231717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231717, CurrentID=0, Status='live', Symbol='Fam109a', Aliases='A230106M15Rik, AU017694, IPIP27A, Ses1', description='family with sequence similarity 109, member A', other_designations='27 kDa inositol polyphosphate phosphatase interacting protein A', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=121849000, end_position_on_the_genomic_accession=121854606, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100647, CurrentID=0, Status='live', Symbol='Upk3b', Aliases='AI447729, AI957187, P35, PMS2L14, UpIIIb', description='uroplakin 3B', other_designations='UP3b|uroplakin IIIb', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=136038496, end_position_on_the_genomic_accession=136044993, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  231724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231724, CurrentID=0, Status='live', Symbol='Rad9b', Aliases='A630082N15Rik, BC021784', description='RAD9 homolog B', other_designations='DNA repair exonuclease rad9 homolog B|cell cycle checkpoint control protein RAD9B homolog|mRAD9B', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122323223, end_position_on_the_genomic_accession=122354272, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  231727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231727, CurrentID=0, Status='live', Symbol='B3gnt4', Aliases='1010001G17Rik, BGnT-4', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 4', other_designations='beta-1,3-Gn-T4|beta3Gn-T4', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123508289, end_position_on_the_genomic_accession=123512099, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  624957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=624957, CurrentID=0, Status='live', Symbol='H2afb3', Aliases='EG624957, H2A.Bbd3-ps, H2afb3', description='H2A histone family, member B3', other_designations='H2A histone family, member B3, pseudogene', map_location='X E2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=120312748, end_position_on_the_genomic_accession=120313095, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100678, CurrentID=0, Status='live', Symbol='Psph', Aliases='AI480570, PSP, PSPase', description='phosphoserine phosphatase', other_designations='O-phosphoserine phosphohydrolase', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129765558, end_position_on_the_genomic_accession=129787253, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100683, CurrentID=0, Status='live', Symbol='Trrap', Aliases='AI481500', description='transformation/transcription domain-associated protein', other_designations='tra1 homolog|transactivation/transformation-domain associated protein', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144768097, end_position_on_the_genomic_accession=144859778, orientation='plus', exon_count=76, OMIM='', no_hdr0=''),
  231760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231760, CurrentID=0, Status='live', Symbol='Rimbp2', Aliases='A930033C01Rik, mKIAA0318', description='RIMS binding protein 2', other_designations='', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=128757791, end_position_on_the_genomic_accession=128953647, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  100689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100689, CurrentID=0, Status='live', Symbol='Spon2', Aliases='2310045I24Rik, AI504350, M-spondin, Mindin, Mspondin', description='spondin 2, extracellular matrix protein', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33213518, end_position_on_the_genomic_accession=33218253, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  231769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231769, CurrentID=0, Status='live', Symbol='Sfswap', Aliases='1190005N23Rik, 6330437E22Rik, AI197402, AW212079, SWAP, Sfrs8, Srsf8', description='splicing factor, suppressor of white-apricot homolog (Drosophila)', other_designations='serine/arginine-rich splicing factor 8|splicing factor, arginine/serine-rich 8|suppressor of white apricot protein homolog', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129501231, end_position_on_the_genomic_accession=129571384, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  100702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100702, CurrentID=0, Status='live', Symbol='Gbp6', Aliases='AI595338, Mpa2l', description='guanylate binding protein 6', other_designations='macrophage activation 2 like', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105270702, end_position_on_the_genomic_accession=105295870, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  493919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=493919, CurrentID=0, Status='live', Symbol='Snord115', Aliases='MBII-52', description='small nucleolar RNA, C/D Box 115 cluster', other_designations='snoRNA MBII-52', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100705, CurrentID=0, Status='live', Symbol='Acacb', Aliases='AI597064, AW743042, Acc2, Accb', description='acetyl-Coenzyme A carboxylase beta', other_designations='ACC-beta', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114146318, end_position_on_the_genomic_accession=114250761, orientation='plus', exon_count=58, OMIM='', no_hdr0=''),
  100710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100710, CurrentID=0, Status='live', Symbol='Pds5b', Aliases='AI646570, AS3, AW212954, Aprin, mKIAA0979', description='PDS5, regulator of cohesion maintenance, homolog B (S. cerevisiae)', other_designations='androgen-induced proliferation inhibitor|androgen-induced prostate proliferative shutoff associated protein AS3|androgen-induced prostate proliferative shutoff-associated protein AS3', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=150673744, end_position_on_the_genomic_accession=150810669, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  100715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100715, CurrentID=0, Status='live', Symbol='Papd4', Aliases='8030446C20Rik, AI649009, GLD-2, Gld2', description='PAP associated domain containing 4', other_designations='PAP-associated domain-containing protein 4|cytoplasmic poly(A) polymerase|mGLD-2', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=93146275, end_position_on_the_genomic_accession=93192970, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  231798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231798, CurrentID=0, Status='live', Symbol='Lrch4', Aliases='2810008P14Rik, 2900069C24Rik, AI558103, LRN, LRRN4, SAP25, mFLJ00248', description='leucine-rich repeats and calponin homology (CH) domain containing 4', other_designations='leucine rich repeat protein 4, neuronal', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137629123, end_position_on_the_genomic_accession=137641099, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  100727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100727, CurrentID=0, Status='live', Symbol='Ugt2b34', Aliases='AI788959', description='UDP glucuronosyltransferase 2 family, polypeptide B34', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86889770, end_position_on_the_genomic_accession=86906937, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231801, CurrentID=0, Status='live', Symbol='Agfg2', Aliases='A630095P14Rik, Hrbl, RAB-R, RABR', description='ArfGAP with FG repeats 2', other_designations='HIV-1 Rev-binding protein-like protein|arf-GAP domain and FG repeats-containing protein 2|rev/Rex activation domain-binding protein related', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137650463, end_position_on_the_genomic_accession=137684806, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  625018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625018, CurrentID=0, Status='live', Symbol='C4a', Aliases='Slp, SlpA', description='complement component 4A (Rodgers blood group)', other_designations='', map_location='17 18.37 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34809092, end_position_on_the_genomic_accession=34823454, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  231803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231803, CurrentID=0, Status='live', Symbol='Mepce', Aliases='Bcdin3, D5Wsu46e', description='methylphosphate capping enzyme', other_designations='bin3, bicoid-interacting 3, homolog', map_location='5 76.7 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137781906, end_position_on_the_genomic_accession=137786701, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100732, CurrentID=0, Status='live', Symbol='Mapre3', Aliases='AI790651, EB2, EB3, EBF3, RP3', description='microtubule-associated protein, RP/EB family, member 3', other_designations='EB1 protein family member 3|end-binding protein 3', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30814614, end_position_on_the_genomic_accession=30866106, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  231805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231805, CurrentID=0, Status='live', Symbol='Pilra', Aliases='AV021745, FDF03', description='paired immunoglobin-like type 2 receptor alpha', other_designations='cell surface receptor FDF03|inhibitory receptor PILR-alpha', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137821952, end_position_on_the_genomic_accession=137836278, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  231807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231807, CurrentID=0, Status='live', Symbol='BC037034', Aliases='', description='cDNA sequence BC037034', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138259658, end_position_on_the_genomic_accession=138264052, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100737, CurrentID=0, Status='live', Symbol='Dcun1d4', Aliases='AI836376', description='DCN1, defective in cullin neddylation 1, domain containing 4 (S. cerevisiae)', other_designations='DCUN1 domain-containing protein 4|defective in cullin neddylation protein 1-like protein 4', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=73481027, end_position_on_the_genomic_accession=73560794, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  231821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231821, CurrentID=0, Status='live', Symbol='Adap1', Aliases='4930431P11Rik, Centa1, GC1L, p42IP4', description='ArfGAP with dual PH domains 1', other_designations='centaurin, alpha 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139271876, end_position_on_the_genomic_accession=139325464, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100756, CurrentID=0, Status='live', Symbol='Usp30', Aliases='6330590F17Rik, AI851327, D5Ertd483e', description='ubiquitin specific peptidase 30', other_designations='deubiquitinating enzyme 30|ub-specific protease 30|ubiquitin specific protease 30|ubiquitin thioesterase 30|ubiquitin thiolesterase 30|ubiquitin-specific protease 30|ubiquitin-specific-processing protease 30', map_location='5 55.96 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114100328, end_position_on_the_genomic_accession=114123510, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  231830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231830, CurrentID=0, Status='live', Symbol='Micall2', Aliases='A930021H16Rik, JRAB, MICAL-L2, mFLJ00139', description='MICAL-like 2', other_designations='junctional Rab13-binding protein|molecule interacting with CasL-like 2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139706692, end_position_on_the_genomic_accession=139736333, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  231832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231832, CurrentID=0, Status='live', Symbol='Tmem184a', Aliases='Sdmg1', description='transmembrane protein 184a', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139804952, end_position_on_the_genomic_accession=139819970, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  231834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231834, CurrentID=0, Status='live', Symbol='Snx8', Aliases='B130023O14Rik', description='sorting nexin 8', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140340298, end_position_on_the_genomic_accession=140389247, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100763, CurrentID=0, Status='live', Symbol='Ube3c', Aliases='AI853514, mKIAA0010', description='ubiquitin protein ligase E3C', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=29569242, end_position_on_the_genomic_accession=29676077, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  100764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100764, CurrentID=0, Status='live', Symbol='Rita1', Aliases='1110008J03Rik, AI853657, C86297, Rita', description='RBPJ interacting and tubulin associated 1', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120609058, end_position_on_the_genomic_accession=120612589, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  231841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231841, CurrentID=0, Status='live', Symbol='Brat1', Aliases='AA881470, Baat1', description='BRCA1-associated ATM activator 1', other_designations='BRCA1-associated protein required for ATM activation protein 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140704786, end_position_on_the_genomic_accession=140719379, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  231842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231842, CurrentID=0, Status='live', Symbol='Amz1', Aliases='5330426I05, 6530401C20Rik', description='archaelysin family metallopeptidase 1', other_designations='archeobacterial metalloproteinase-like 1|archeobacterial metalloproteinase-like protein 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140724127, end_position_on_the_genomic_accession=140753312, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  231855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231855, CurrentID=0, Status='live', Symbol='Ap5z1', Aliases='C330006K01Rik', description='adaptor-related protein complex 5, zeta 1 subunit', other_designations='adapter-related protein complex 5 zeta subunit|adaptor-related protein complex 5 zeta subunit|zeta5', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142463205, end_position_on_the_genomic_accession=142478716, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  231858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231858, CurrentID=0, Status='live', Symbol='Radil', Aliases='AI536456, D930005D10Rik', description='Ras association and DIL domains', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142484839, end_position_on_the_genomic_accession=142551098, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  231861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231861, CurrentID=0, Status='live', Symbol='Tnrc18', Aliases='BC025631, EG381742, Zfp469', description='trinucleotide repeat containing 18', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142724604, end_position_on_the_genomic_accession=142822703, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  231863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231863, CurrentID=0, Status='live', Symbol='Fbxl18', Aliases='AI427077, B130019G13Rik, C330021B20Rik', description='F-box and leucine-rich repeat protein 18', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142871787, end_position_on_the_genomic_accession=142895643, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  231866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231866, CurrentID=0, Status='live', Symbol='Zfp12', Aliases='C530015C18, Krox-7, Zfp-12, Znf12', description='zinc finger protein 12', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143235163, end_position_on_the_genomic_accession=143248834, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  231868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231868, CurrentID=0, Status='live', Symbol='E130309D02Rik', Aliases='A630028N22', description='RIKEN cDNA E130309D02 gene', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143301072, end_position_on_the_genomic_accession=143315360, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231871, CurrentID=0, Status='live', Symbol='Daglb', Aliases='E330036I19Rik', description='diacylglycerol lipase, beta', other_designations='DGL-beta', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143464487, end_position_on_the_genomic_accession=143504442, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  231872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231872, CurrentID=0, Status='live', Symbol='Aimp2', Aliases='AA407136, AA407225(p38), Jtv1, Aimp2', description='aminoacyl tRNA synthetase complex-interacting multifunctional protein 2', other_designations='ARS interacting multifunctional protein 2|multisynthase complex auxiliary component p38|multisynthetase complex auxiliary component p38|protein JTV-1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143902704, end_position_on_the_genomic_accession=143909839, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  231874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231874, CurrentID=0, Status='live', Symbol='Ccz1', Aliases='AU022870', description='CCZ1 vacuolar protein trafficking and biogenesis associated', other_designations='CCZ1 vacuolar protein trafficking and biogenesis associated homolog', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=143987909, end_position_on_the_genomic_accession=144014853, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  231876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231876, CurrentID=0, Status='live', Symbol='Lmtk2', Aliases='2900041G10Rik, A330101P12Rik, BREK, KPI-2, KPI2, aatyk2, cprk', description='lemur tyrosine kinase 2', other_designations='apoptosis-associated tyrosine kinase 2|brain-enriched kinase', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144100366, end_position_on_the_genomic_accession=144188204, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  625098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625098, CurrentID=0, Status='live', Symbol='Slc38a6', Aliases='AW322671, EG625098', description='solute carrier family 38, member 6', other_designations='Na(+)-coupled neutral amino acid transporter 6|solute carrier family 38 member 6', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73286794, end_position_on_the_genomic_accession=73361998, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  231887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231887, CurrentID=0, Status='live', Symbol='Pdap1', Aliases='HASPP28, PAP, PAP1', description='PDGFA associated protein 1', other_designations='PDGF-associated protein|PDGFA-associated protein 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=145128770, end_position_on_the_genomic_accession=145140089, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104184, CurrentID=0, Status='live', Symbol='Blmh', Aliases='AI035728, Bh, Bmh', description='bleomycin hydrolase', other_designations='BLM hydrolase', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76945656, end_position_on_the_genomic_accession=76987389, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  231903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231903, CurrentID=0, Status='live', Symbol='Urad', Aliases='EG231903, Prhoxnb', description='ureidoimidazoline (2-oxo-4-hydroxy-4-carboxy-5) decarboxylase', other_designations='OHCU decarboxylase|parahox cluster neighbor|parahox neighbor|ureidoimidazoline (2-oxo-4-hydroxy-4-carboxy-5-) decarboxylase', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=147314984, end_position_on_the_genomic_accession=147322440, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  231912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231912, CurrentID=0, Status='live', Symbol='Katnal1', Aliases='', description='katanin p60 subunit A-like 1', other_designations='p60 katanin-like 1', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=148871584, end_position_on_the_genomic_accession=148929134, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  231915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231915, CurrentID=0, Status='live', Symbol='Uspl1', Aliases='E430026A01Rik', description='ubiquitin specific peptidase like 1', other_designations='ubiquitin-specific peptidase-like protein 1', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149184560, end_position_on_the_genomic_accession=149215434, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100855, CurrentID=0, Status='live', Symbol='Tbc1d14', Aliases='2810413P16Rik, AU043625, C86258, D5Ertd110e, mKIAA1322', description='TBC1 domain family, member 14', other_designations='', map_location='5 19.15 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36490604, end_position_on_the_genomic_accession=36586273, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  231931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231931, CurrentID=0, Status='live', Symbol='Gimap6', Aliases='4833419H03Rik, Ian6, mFLJ00102', description='GTPase, IMAP family member 6', other_designations='IAN-6|immune associated nucleotide 6|immune-associated nucleotide 6|immunity-associated nucleotide 6 protein|mIAN6', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48701583, end_position_on_the_genomic_accession=48708244, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  231932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231932, CurrentID=0, Status='live', Symbol='Gimap7', Aliases='IAN7, Ian3', description='GTPase, IMAP family member 7', other_designations='immune associated nucleotide 3', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48718621, end_position_on_the_genomic_accession=48724636, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  231946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231946, CurrentID=0, Status='live', Symbol='Fam221a', Aliases='D330028D13Rik', description='family with sequence similarity 221, member A', other_designations='uncharacterized protein C7orf46 homolog', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49367219, end_position_on_the_genomic_accession=49390540, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  60505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=60505, CurrentID=0, Status='live', Symbol='Il21', Aliases='IL-21', description='interleukin 21', other_designations='', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37222759, end_position_on_the_genomic_accession=37232636, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100900, CurrentID=0, Status='live', Symbol='Hscb', Aliases='AI325508, AW049829, Hsc20', description='HscB iron-sulfur cluster co-chaperone', other_designations='HscB iron-sulfur cluster co-chaperone homolog|J-type co-chaperone HSC20', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110829070, end_position_on_the_genomic_accession=110839802, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100910, CurrentID=0, Status='live', Symbol='Chpf2', Aliases='2010209O12Rik, AW060945, mKIAA1402, mir-671', description='chondroitin polymerizing factor 2', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24586750, end_position_on_the_genomic_accession=24592487, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  231986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231986, CurrentID=0, Status='live', Symbol='Jazf1', Aliases='AI591476, C820002C15, Jaz1, Tip27', description='JAZF zinc finger 1', other_designations='', map_location='6 25.74 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52768065, end_position_on_the_genomic_accession=53068624, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  16819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16819, CurrentID=0, Status='live', Symbol='Lcn2', Aliases='24p3, AW212229, Sip24', description='lipocalin 2', other_designations='SV-40-induced 24P3 protein|lipocalin-2|p25|secreted inducible protein 24|siderocalin LCN2', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32384637, end_position_on_the_genomic_accession=32387739, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  231991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231991, CurrentID=0, Status='live', Symbol='Creb5', Aliases='Crebpa, D430026C09Rik', description='cAMP responsive element binding protein 5', other_designations='CRE-BPa|CREB-5|cAMP-responsive element-binding protein 5', map_location='6 25.9 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=53573374, end_position_on_the_genomic_accession=53697850, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  231997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231997, CurrentID=0, Status='live', Symbol='Fkbp14', Aliases='BC029109, FKBP-14, FKBP22', description='FK506 binding protein 14', other_designations='FK506-binding protein 14|PPIase FKBP14|rotamase', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54577605, end_position_on_the_genomic_accession=54593698, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  231999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231999, CurrentID=0, Status='live', Symbol='Plekha8', Aliases='6330400G01, AA517676, BC052360, Fapp-2, Fapp2', description='pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 8', other_designations='PH domain-containing family A member 8|phosphatidylinositol-four-phosphate adapter protein 2|phosphoinositol 4-phosphate adapter protein 2', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54595111, end_position_on_the_genomic_accession=54645823, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100929, CurrentID=0, Status='live', Symbol='Tyw1', Aliases='AW125391, Rsafd1', description='tRNA-yW synthesizing protein 1 homolog (S. cerevisiae)', other_designations='radical S-adenosyl methionine and flavodoxin domain-containing protein 1|radical S-adenosyl methionine and flavodoxin domains 1|tRNA wybutosine-synthesizing protein 1 homolog|tRNA-yW-synthesizing protein', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130255572, end_position_on_the_genomic_accession=130341567, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  232016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232016, CurrentID=0, Status='live', Symbol='Ccdc129', Aliases='D530004J12Rik, Gm458', description='coiled-coil domain containing 129', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=55804628, end_position_on_the_genomic_accession=55978743, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  232023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232023, CurrentID=0, Status='live', Symbol='Vopp1', Aliases='AW146242, Ecop', description='vesicular, overexpressed in cancer, prosurvival protein 1', other_designations='EGFR-coamplified and overexpressed protein', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57752264, end_position_on_the_genomic_accession=57825125, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100952, CurrentID=0, Status='live', Symbol='Emilin1', Aliases='5830419M17Rik, AW229038, BB105748, EMILIN-1, gp115', description='elastin microfibril interfacer 1', other_designations='elastin microfibril interface located protein 1|elastin microfibril interface-located protein 1', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30913381, end_position_on_the_genomic_accession=30921279, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  625249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625249, CurrentID=0, Status='live', Symbol='Gpx4', Aliases='GPx-4, GSHPx-4, PHGPx, mtPHGPx, snGPx', description='glutathione peroxidase 4', other_designations='phospholipid hydroperoxide glutathione peroxidase, mitochondrial|sperm nuclei glutathione peroxidase', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80047166, end_position_on_the_genomic_accession=80056439, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  232035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232035, CurrentID=0, Status='live', Symbol='Ccser1', Aliases='6230405M12Rik, C130092O11Rik, Fam190a', description='coiled-coil serine rich 1', other_designations='coiled-coil serine-rich protein 1|family with sequence similarity 190, member A|protein FAM190A', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=61180325, end_position_on_the_genomic_accession=62382863, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100972, CurrentID=0, Status='live', Symbol='Rab28', Aliases='2700023P08Rik, AW496496', description='RAB28, member RAS oncogene family', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=41624974, end_position_on_the_genomic_accession=41708192, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100978, CurrentID=0, Status='live', Symbol='Nfxl1', Aliases='1700012H24Rik, AW538212, D430033A06Rik, DNABF, GCF, Gm1058, TCF9', description='nuclear transcription factor, X-box binding-like 1', other_designations='ovarian zinc finger protein', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=72513301, end_position_on_the_genomic_accession=72559707, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  100986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100986, CurrentID=0, Status='live', Symbol='Akap9', Aliases='5730481H23Rik, AKAP-9, AKAP450, AW545847, C79026, G1-448-15, PRKA9, mKIAA0803, mei2-5, repro12', description='A kinase (PRKA) anchor protein (yotiao) 9', other_designations='protein kinase A-anchoring protein 9', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3927417, end_position_on_the_genomic_accession=4080204, orientation='plus', exon_count=50, OMIM='', no_hdr0=''),
  625285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625285, CurrentID=0, Status='live', Symbol='H2-Q3', Aliases='EG625285, Gm6571, H-2Q3, Qa-3, Qa3', description='histocompatibility 2, Q region locus 3', other_designations='', map_location='17 18.64 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35358969, end_position_on_the_genomic_accession=35359982, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  625286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625286, CurrentID=0, Status='live', Symbol='Tmem236', Aliases='2010003H20Rik, Fam23a, Gm613', description='transmembrane protein 236', other_designations='family with sequence similarity 23, member A', map_location='2 A2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14174524, end_position_on_the_genomic_accession=14221993, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232077, CurrentID=0, Status='live', Symbol='Foxi3', Aliases='', description='forkhead box I3', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70956606, end_position_on_the_genomic_accession=70961066, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  232078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232078, CurrentID=0, Status='live', Symbol='Thnsl2', Aliases='BC051244, TSH2', description='threonine synthase-like 2 (bacterial)', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71128166, end_position_on_the_genomic_accession=71144380, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  232086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232086, CurrentID=0, Status='live', Symbol='Tmem150a', Aliases='BC014685, Tmem150', description='transmembrane protein 150A', other_designations='fasting-inducible integral membrane protein TM6P1|transmembrane protein 150', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72355483, end_position_on_the_genomic_accession=72359762, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  232087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232087, CurrentID=0, Status='live', Symbol='Mat2a', Aliases='D630045P18Rik', description='methionine adenosyltransferase II, alpha', other_designations='MAT 2|S-adenosylmethionine synthetase isoform type-2|adoMet synthase 2|adoMet synthetase 2|methionine adenosyltransferase 2', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72432796, end_position_on_the_genomic_accession=72439608, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  232089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232089, CurrentID=0, Status='live', Symbol='Elmod3', Aliases='AI844780, C330008I15Rik, RBM29, Rbed1', description='ELMO/CED-12 domain containing 3', other_designations='RNA binding motif and ELMO domain 1|RNA-binding motif and ELMO domain-containing protein 1', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=72565922, end_position_on_the_genomic_accession=72598413, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  101023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101023, CurrentID=0, Status='live', Symbol='Zfp513', Aliases='AW990386, D430028M17Rik, Znf513', description='zinc finger protein 513', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31198981, end_position_on_the_genomic_accession=31202663, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  625328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625328, CurrentID=0, Status='live', Symbol='H3f3c', Aliases='Gm14384, H3mm11, OTTMUSG00000016509', description='H3 histone, family 3C', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119665587, end_position_on_the_genomic_accession=119666142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  232146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232146, CurrentID=0, Status='live', Symbol='Eva1a', Aliases='BC014699, Fam176a, Tmem166', description='eva-1 homolog A (C. elegans)', other_designations='family with sequence similarity 176, member A|protein FAM176A|transmembrane protein 166', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=82041048, end_position_on_the_genomic_accession=82093099, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  232156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232156, CurrentID=0, Status='live', Symbol='Slc4a5', Aliases='C330006M16, C330016K18Rik', description='solute carrier family 4, sodium bicarbonate cotransporter, member 5', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83207580, end_position_on_the_genomic_accession=83304945, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  232157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232157, CurrentID=0, Status='live', Symbol='Mob1a', Aliases='4022402H07Rik, MOB1, MOB4B, Mobk1b, Mobkl1b', description='MOB kinase activator 1A', other_designations='MOB1, Mps One Binder kinase activator-like 1B|mob1 homolog 1B|mps one binder kinase activator-like 1B', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83326033, end_position_on_the_genomic_accession=83343696, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  232164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232164, CurrentID=0, Status='live', Symbol='Paip2b', Aliases='', description='poly(A) binding protein interacting protein 2B', other_designations='PABP-interacting protein 2B|PAIP-2B|poly(A)-binding protein-interacting protein 2B', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83806058, end_position_on_the_genomic_accession=83831741, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  101095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101095, CurrentID=0, Status='live', Symbol='Zfp282', Aliases='AI449432, C81429, E430019K09Rik, HUB1', description='zinc finger protein 282', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47877209, end_position_on_the_genomic_accession=47908485, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101100, CurrentID=0, Status='live', Symbol='Ttll3', Aliases='4833441J24Rik, AI450050', description='tubulin tyrosine ligase-like family, member 3', other_designations='TTL domain protein|ttll3|tubulin--tyrosine ligase-like protein 3', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113389260, end_position_on_the_genomic_accession=113414583, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  232174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232174, CurrentID=0, Status='live', Symbol='Cyp26b1', Aliases='CP26, P450RAI-2', description='cytochrome P450, family 26, subfamily b, polypeptide 1', other_designations='cytochrome P450 retinoic acid-inactivating 2|cytochrome P450, 26, retinoic acid B1|cytochrome P450RAI-2|retinoic acid B1|retinoic acid hydroxylase', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=84571414, end_position_on_the_genomic_accession=84593908, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101113, CurrentID=0, Status='live', Symbol='Snx21', Aliases='5730407K14Rik, AI481716', description='sorting nexin family member 21', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164785762, end_position_on_the_genomic_accession=164793810, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232187, CurrentID=0, Status='live', Symbol='Smyd5', Aliases='AW536703, NN8-4AG, Rai15, Rrg1', description='SET and MYND domain containing 5', other_designations='retinoic acid induced 15|retinoic acid responsive gene 1|retinoic acid-induced protein 15', map_location='6 37.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85431976, end_position_on_the_genomic_accession=85446429, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  101118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101118, CurrentID=0, Status='live', Symbol='Tmem168', Aliases='5730526F17Rik, 8430437G11Rik, AI462344, AI504145', description='transmembrane protein 168', other_designations='', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=13580689, end_position_on_the_genomic_accession=13608222, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101122, CurrentID=0, Status='live', Symbol='Rpusd3', Aliases='AI527266', description='RNA pseudouridylate synthase domain containing 3', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113415319, end_position_on_the_genomic_accession=113419459, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  232196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232196, CurrentID=0, Status='live', Symbol='C87436', Aliases='', description='expressed sequence C87436', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=86437328, end_position_on_the_genomic_accession=86470388, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  232201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232201, CurrentID=0, Status='live', Symbol='Arhgap25', Aliases='A130039I20Rik', description='Rho GTPase activating protein 25', other_designations='rho-type GTPase-activating protein 25', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87458544, end_position_on_the_genomic_accession=87533259, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  625424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625424, CurrentID=0, Status='live', Symbol='Gm6583', Aliases='EG625424', description='predicted gene 6583', other_designations='', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112353773, end_position_on_the_genomic_accession=112356033, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  232210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232210, CurrentID=0, Status='live', Symbol='Hmces', Aliases='8430410A17Rik, C85376', description='5-hydroxymethylcytosine (hmC) binding, ES cell specific', other_designations='ES cell-specific 5hmC-binding protein|SRAP domain-containing protein 1|UPF0361 protein C3orf37 homolog|putative peptidase SRAPD1', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87913907, end_position_on_the_genomic_accession=87936614, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101142, CurrentID=0, Status='live', Symbol='Itfg2', Aliases='2700050P07Rik, AI646725', description='integrin alpha FG-GAP repeat containing 2', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128409444, end_position_on_the_genomic_accession=128424946, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  232217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232217, CurrentID=0, Status='live', Symbol='4933427D06Rik', Aliases='', description='RIKEN cDNA 4933427D06 gene', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89096115, end_position_on_the_genomic_accession=89110037, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101148, CurrentID=0, Status='live', Symbol='B630005N14Rik', Aliases='AI666701', description='RIKEN cDNA B630005N14 gene', other_designations='UPF0532 protein C7orf60 homolog', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=13625675, end_position_on_the_genomic_accession=13677966, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  232223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232223, CurrentID=0, Status='live', Symbol='Txnrd3', Aliases='AI196535, TR2, Tgr', description='thioredoxin reductase 3', other_designations='thioredoxin and glutathione reductase|thioredoxin reductase TR2|txnrd3', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89643988, end_position_on_the_genomic_accession=89675529, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  232227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232227, CurrentID=0, Status='live', Symbol='Iqsec1', Aliases='AW561907, BC026481, BRAG2, D6Ertd349e, cI-43', description='IQ motif and Sec7 domain 1', other_designations='', map_location='6 40.16 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90659596, end_position_on_the_genomic_accession=90988460, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  232232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232232, CurrentID=0, Status='live', Symbol='Hdac11', Aliases='', description='histone deacetylase 11', other_designations='HD11', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91156815, end_position_on_the_genomic_accession=91174692, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  232236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232236, CurrentID=0, Status='live', Symbol='Ccdc174', Aliases='C130022K22Rik', description='coiled-coil domain containing 174', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91877654, end_position_on_the_genomic_accession=91899849, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  232237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232237, CurrentID=0, Status='live', Symbol='Fgd5', Aliases='C330025N11Rik, ZFYVE23, mFLJ00274', description='FYVE, RhoGEF and PH domain containing 5', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91978821, end_position_on_the_genomic_accession=92076005, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  625464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625464, CurrentID=0, Status='live', Symbol='Gm6588', Aliases='EG625464', description='predicted gene 6588', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112449426, end_position_on_the_genomic_accession=112451738, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  101185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101185, CurrentID=0, Status='live', Symbol='Pot1a', Aliases='1500031H18Rik, AI851169, Pot1', description='protection of telomeres 1A', other_designations='POT1-like telomere end-binding protein|mPot1', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=25743735, end_position_on_the_genomic_accession=25809277, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  101187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101187, CurrentID=0, Status='live', Symbol='Parp11', Aliases='5330431N24Rik, AI851877, ARTD11, HIN1L', description='poly (ADP-ribose) polymerase family, member 11', other_designations='ADP-ribosyltransferase diphtheria toxin-like 11|PARP-11', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127446845, end_position_on_the_genomic_accession=127512908, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  101197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101197, CurrentID=0, Status='live', Symbol='Zfp956', Aliases='AI894139', description='zinc finger protein 956', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47953390, end_position_on_the_genomic_accession=47965300, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  235320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235320, CurrentID=0, Status='live', Symbol='Zbtb16', Aliases='AI467657, PLZF, Zfp145, lu', description='zinc finger and BTB domain containing 16', other_designations="Green's luxoid|luxoid|zinc finger protein 145", map_location='9 26.47 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48651797, end_position_on_the_genomic_accession=48836381, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101202, CurrentID=0, Status='live', Symbol='Hepacam2', Aliases='AI987662', description='HEPACAM family member 2', other_designations='mitotic kinetics regulator', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=3457089, end_position_on_the_genomic_accession=3498189, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101206, CurrentID=0, Status='live', Symbol='Tada3', Aliases='1110004B19Rik, ADA3, AI987856l, Tada3', description='transcriptional adaptor 3', other_designations='ADA3 homolog|ADA3-like protein|mADA3', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113366633, end_position_on_the_genomic_accession=113377939, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  101214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101214, CurrentID=0, Status='live', Symbol='Tra2a', Aliases='1500010G04Rik, AL022798, G430041M01Rik, mAWMS1', description='transformer 2 alpha homolog (Drosophila)', other_designations='TRA-2 alpha|TRA2-alpha|Tra2alpha|nuclear transformer-2 alpha|transformer-2 protein homolog A', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49243921, end_position_on_the_genomic_accession=49264248, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  232288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232288, CurrentID=0, Status='live', Symbol='Frmd4b', Aliases='6030440G05Rik, C87375, GOBLIN, GRSP1, R74720, mKIAA1013', description='FERM domain containing 4B', other_designations='GRP1 binding protein GRSP1|GRP1-binding protein GRSP1|golgi-associated band 4.1-like protein', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=97286867, end_position_on_the_genomic_accession=97617657, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  494448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=494448, CurrentID=0, Status='live', Symbol='Cbx6', Aliases='', description='chromobox 6', other_designations='chromobox homolog 6', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79823898, end_position_on_the_genomic_accession=79834333, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101240, CurrentID=0, Status='live', Symbol='Wdr91', Aliases='9530020G05Rik, AI987683, AU018665', description='WD repeat domain 91', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34879062, end_position_on_the_genomic_accession=34910881, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  232313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232313, CurrentID=0, Status='live', Symbol='Gxylt2', Aliases='Glt8d4, Gm152', description='glucoside xylosyltransferase 2', other_designations='glycosyltransferase 8 domain containing 4|glycosyltransferase 8 domain-containing protein 4', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=100704293, end_position_on_the_genomic_accession=100805686, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  232314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232314, CurrentID=0, Status='live', Symbol='Ppp4r2', Aliases='BE691708, C230060M08Rik', description='protein phosphatase 4, regulatory subunit 2', other_designations='', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=100833623, end_position_on_the_genomic_accession=100869934, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  494468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=494468, CurrentID=0, Status='live', Symbol='Armcx5', Aliases='AW556585', description='armadillo repeat containing, X-linked 5', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135742692, end_position_on_the_genomic_accession=135747328, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  232333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232333, CurrentID=0, Status='live', Symbol='Slc6a1', Aliases='A730043E01, GABATHG, GABATR, GAT-1, Gabt, Gabt1, Gat1, XT-1, Xtrp1', description='solute carrier family 6 (neurotransmitter transporter, GABA), member 1', other_designations='gamma-aminobutyric acid (GABA-A) transporter 1|solute carrier family 6 member 1', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=114282635, end_position_on_the_genomic_accession=114317525, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  232334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232334, CurrentID=0, Status='live', Symbol='Vgll4', Aliases='BC048841, VGL-4', description='vestigial like 4 (Drosophila)', other_designations='', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=114860621, end_position_on_the_genomic_accession=114969989, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  232337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232337, CurrentID=0, Status='live', Symbol='Zfp637', Aliases='AI646709, BC021601, Znf32', description='zinc finger protein 637', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=117841156, end_position_on_the_genomic_accession=117845956, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232339, CurrentID=0, Status='live', Symbol='Ankrd26', Aliases='5730521P14Rik, BC008111, mKIAA1074', description='ankyrin repeat domain 26', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=118501308, end_position_on_the_genomic_accession=118562269, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  232341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232341, CurrentID=0, Status='live', Symbol='Wnk1', Aliases='6430573H23Rik, EG406236, Hsn2, Prkwnk1, mKIAA0344', description='WNK lysine deficient protein kinase 1', other_designations='protein kinase lysine-deficient 1|protein kinase with no lysine 1', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119923969, end_position_on_the_genomic_accession=120038655, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  625558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625558, CurrentID=0, Status='live', Symbol='Gm6600', Aliases='287g15T7, C86350, EG625558, Zfp-like', description='predicted gene 6600', other_designations='', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=130728932, end_position_on_the_genomic_accession=130799970, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  232345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232345, CurrentID=0, Status='live', Symbol='A2m', Aliases='A2mp', description='alpha-2-macroglobulin', other_designations='A2mp', map_location='6 57.49 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121636166, end_position_on_the_genomic_accession=121679238, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  494497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=494497, CurrentID=0, Status='live', Symbol='BC037156', Aliases='FDC-SP', description='cDNA sequence BC037156', other_designations='follicular dendritic cell secreted peptide', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87908779, end_position_on_the_genomic_accession=87914171, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  494504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=494504, CurrentID=0, Status='live', Symbol='Apcdd1', Aliases='AB023957, AU041258, Drapc1, EIG180', description='adenomatosis polyposis coli down-regulated 1', other_designations='adenomatosis polyposis coli down-regulated 1 protein homolog|apcdd1|ethanol induced gene product EIG180', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=62922327, end_position_on_the_genomic_accession=62953195, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232370, CurrentID=0, Status='live', Symbol='Clstn3', Aliases='CSTN3, Cs3, Cst-3, mKIAA0726', description='calsyntenin 3', other_designations='alc-beta|alcadein-beta', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124430756, end_position_on_the_genomic_accession=124464784, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  232371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232371, CurrentID=0, Status='live', Symbol='C1rl', Aliases='C1r-LP1, C1rlp, C1rl', description='complement component 1, r subcomponent-like', other_designations='C1r-like protein|complement component 1, r subcomponent like protein', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124493113, end_position_on_the_genomic_accession=124510643, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  494519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=494519, CurrentID=0, Status='live', Symbol='Scgb2b20', Aliases='Abpbg20, Abpd, C2c, Scgb2b2', description='secretoglobin, family 2B, member 20', other_designations='allergen dI chain C2C|androgen binding protein delta|androgen-binding protein delta|lacrimal androgen-binding protein delta|secretoglobin, family 2B, member 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33364343, end_position_on_the_genomic_accession=33366319, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  625599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625599, CurrentID=0, Status='live', Symbol='Gml', Aliases='EG625599, HemT', description='glycosylphosphatidylinositol anchored molecule like', other_designations='GPI anchored molecule like protein|hematopoietic cell transcript 3', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74813435, end_position_on_the_genomic_accession=74818815, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  101314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101314, CurrentID=0, Status='live', Symbol='Brk1', Aliases='6720456B07Rik, AW011779', description='BRICK1, SCAR/WAVE actin-nucleating complex subunit', other_designations='probable protein BRICK1', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113604772, end_position_on_the_genomic_accession=113616951, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101320, CurrentID=0, Status='live', Symbol='Dyrk4', Aliases='AW049118a, Dyrk4b, Dyrk4', description='dual-specificity tyrosine-(Y)-phosphorylation regulated kinase 4', other_designations='dyrk4', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=126876020, end_position_on_the_genomic_accession=126921864, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  232400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232400, CurrentID=0, Status='live', Symbol='BC048546', Aliases='Ovos', description='cDNA sequence BC048546', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=128539822, end_position_on_the_genomic_accession=128581606, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  494546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=494546, CurrentID=0, Status='live', Symbol='Taar8c', Aliases='taR-8c', description='trace amine-associated receptor 8C', other_designations='mTaar8c', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24100843, end_position_on_the_genomic_accession=24101951, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  232408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232408, CurrentID=0, Status='live', Symbol='Klrb1f', Aliases='A630024B12Rik, Nkr-p1f, Nkrp1-f, Nkrp1f', description='killer cell lectin-like receptor subfamily B member 1F', other_designations='CD161 antigen-like family member F|natural killer cell surface protein NKR-P1F', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129036902, end_position_on_the_genomic_accession=129059724, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  232409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232409, CurrentID=0, Status='live', Symbol='Clec2e', Aliases='Clra, clr-a', description='C-type lectin domain family 2, member e', other_designations='C-type lectin-related protein A', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129091816, end_position_on_the_genomic_accession=129100903, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  232413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232413, CurrentID=0, Status='live', Symbol='Clec12a', Aliases='CLL-1, D230024O04, KLRL1, Micl', description='C-type lectin domain family 12, member a', other_designations='C-type lectin-like molecule 1|myeloid inhibitory C-type lectin-like receptor', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129349779, end_position_on_the_genomic_accession=129365303, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  232414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232414, CurrentID=0, Status='live', Symbol='Clec9a', Aliases='9830005G06Rik, DNGR-1', description='C-type lectin domain family 9, member a', other_designations='dendritic cell natural killer lectin group receptor 1', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129408862, end_position_on_the_genomic_accession=129501701, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  232415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232415, CurrentID=0, Status='live', Symbol='Gm156', Aliases='Klrh1', description='predicted gene 156', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129766568, end_position_on_the_genomic_accession=129784270, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  104272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104272, CurrentID=0, Status='live', Symbol='Caf', Aliases='', description='caffeine susceptibility', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  625638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=625638, CurrentID=0, Status='live', Symbol='Fam43b', Aliases='OTTMUSG00000009974', description='family with sequence similarity 43, member B', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=138394092, end_position_on_the_genomic_accession=138396458, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  101351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101351, CurrentID=0, Status='live', Symbol='Eogt', Aliases='A130022J15Rik, AI447490, AW214473, AW259391, Aer61', description='EGF domain-specific O-linked N-acetylglucosamine (GlcNAc) transferase', other_designations='extracellular O-linked N-acetylglucosamine transferase', map_location='6 D3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=97110016, end_position_on_the_genomic_accession=97149703, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  232426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232426, CurrentID=0, Status='live', Symbol='5530400C23Rik', Aliases='GRP', description='RIKEN cDNA 5530400C23 gene', other_designations='submandibular gland secretory Glx-rich protein', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=133292216, end_position_on_the_genomic_accession=133295790, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  101358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101358, CurrentID=0, Status='live', Symbol='Fbxl14', Aliases='AW322056, Fbx14l', description='F-box and leucine-rich repeat protein 14', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119479668, end_position_on_the_genomic_accession=119483886, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  232431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232431, CurrentID=0, Status='live', Symbol='Gprc5a', Aliases='Rai3, Raig1', description='G protein-coupled receptor, family C, group 5, member A', other_designations='G-protein coupled receptor family C group 5 member A|RAIG-1|retinoic acid induced 3|retinoic acid-induced gene 1 protein', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=135065662, end_position_on_the_genomic_accession=135084709, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232440, CurrentID=0, Status='live', Symbol='H2afj', Aliases='E130307C13', description='H2A histone family, member J', other_designations='H2a/j', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136808248, end_position_on_the_genomic_accession=136810074, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  232441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232441, CurrentID=0, Status='live', Symbol='Rerg', Aliases='', description='RAS-like, estrogen-regulated, growth-inhibitor', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=137054817, end_position_on_the_genomic_accession=137170496, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  232449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232449, CurrentID=0, Status='live', Symbol='Dera', Aliases='2010002D22Rik, 2500002K03Rik, DEOC', description='deoxyribose-phosphate aldolase (putative)', other_designations='2-deoxy-D-ribose 5-phosphate aldolase|2-deoxyribose-5-phosphate aldolase homolog|deoxyriboaldolase|phosphodeoxyriboaldolase|putative deoxyribose-phosphate aldolase', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=137754577, end_position_on_the_genomic_accession=137837872, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  101401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101401, CurrentID=0, Status='live', Symbol='Adamts9', Aliases='1810011L16Rik, 8430403M15Rik, AW743315, E030027K14Rik', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 9', other_designations='a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 9', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=92772699, end_position_on_the_genomic_accession=92944047, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  232491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232491, CurrentID=0, Status='live', Symbol='Pyroxd1', Aliases='', description='pyridine nucleotide-disulphide oxidoreductase domain 1', other_designations='', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142345625, end_position_on_the_genomic_accession=142362630, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  232493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232493, CurrentID=0, Status='live', Symbol='Gys2', Aliases='BC021322, LGS', description='glycogen synthase 2', other_designations='glycogen synthase, liver', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142422612, end_position_on_the_genomic_accession=142473124, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  101434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101434, CurrentID=0, Status='live', Symbol='Ceacam15', Aliases='AA407838, C430002N04Rik', description='carcinoembryonic antigen-related cell adhesion molecule 15', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16671331, end_position_on_the_genomic_accession=16675705, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  101437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101437, CurrentID=0, Status='live', Symbol='Dhx32', Aliases='3110079L04Rik, 4732469F02Rik, AA408140, Ddx32', description='DEAH (Asp-Glu-Ala-His) box polypeptide 32', other_designations='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 32|DEAH box protein 32|helicase DDX32|muDDX32', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=133720935, end_position_on_the_genomic_accession=133782785, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100502590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502590, CurrentID=0, Status='live', Symbol='4930469G21Rik', Aliases='', description='RIKEN cDNA 4930469G21 gene', other_designations='', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161156843, end_position_on_the_genomic_accession=161159314, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  232533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232533, CurrentID=0, Status='live', Symbol='Stk38l', Aliases='4930473A22Rik, B230328I19, Ndr2, Ndr54', description='serine/threonine kinase 38 like', other_designations='nuclear Dbf2-related kinase 2|stk38l', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146724930, end_position_on_the_genomic_accession=146778814, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  232536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232536, CurrentID=0, Status='live', Symbol='Mrps35', Aliases='MDSO23, MRP-S28, MRPS28', description='mitochondrial ribosomal protein S35', other_designations='MRP-S35|S35mt|mitochondrial ribosomal protein S34|mrps35', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=147042741, end_position_on_the_genomic_accession=147070903, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  232539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232539, CurrentID=0, Status='live', Symbol='Klhl42', Aliases='A330079N23Rik, AI450355, C230080I20Rik, Ctb9, Klhdc5, mKIAA1340', description='kelch-like 42', other_designations='cullin-3-binding protein 9|kelch domain containing 5|kelch domain-containing protein 5', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=147091075, end_position_on_the_genomic_accession=147112778, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101471, CurrentID=0, Status='live', Symbol='Phrf1', Aliases='AA673488', description='PHD and ring finger domains 1', other_designations='phrf1', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141228263, end_position_on_the_genomic_accession=141262751, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  101476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101476, CurrentID=0, Status='live', Symbol='Plekha1', Aliases='AA960558, C920009D07Rik, TAPP1', description='pleckstrin homology domain containing, family A (phosphoinositide binding specific) member 1', other_designations='PH domain-containing family A member 1|TAPP-1|tandem PH domain-containing protein 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=130865729, end_position_on_the_genomic_accession=130913308, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  232560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232560, CurrentID=0, Status='live', Symbol='Caprin2', Aliases='C1qdc1, Eeg1, rng140', description='caprin family member 2', other_designations='C1q domain containing 1|C1q domain-containing protein 1|RNA granule protein 140|caprin2|cytoplasmic activation/proliferation-associated protein 2', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148842492, end_position_on_the_genomic_accession=148896257, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  101489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101489, CurrentID=0, Status='live', Symbol='Ric8', Aliases='AI114950, RIC-8a, Ric8', description='resistance to inhibitors of cholinesterase 8 homolog (C. elegans)', other_designations='synembryn', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140857397, end_position_on_the_genomic_accession=140863731, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  101490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101490, CurrentID=0, Status='live', Symbol='Inpp5f', Aliases='5830435P03Rik, AI115354, AW561896, SAC2, cI-27, mKIAA0966', description='inositol polyphosphate-5-phosphatase F', other_designations='inositol polyphosphate 5-phosphatase F|sac domain-containing inositol phosphatase 2|sac domain-containing phosphoinositide 5-phosphatase 2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128610775, end_position_on_the_genomic_accession=128696437, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  232566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232566, CurrentID=0, Status='live', Symbol='Amn1', Aliases='5830467E07Rik, A230083D09, C730024G19Rik', description='antagonist of mitotic exit network 1', other_designations='antagonist of mitotic exit network 1 homolog', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=149157349, end_position_on_the_genomic_accession=149188762, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101497, CurrentID=0, Status='live', Symbol='Plekhg2', Aliases='AI194308, Clg, Cslg', description='pleckstrin homology domain containing, family G (with RhoGef domain) member 2', other_designations='PH domain-containing family G member 2|common site lymphoma/leukemia GEF|common site lymphoma/leukemia guanine nucleotide exchange factor|common-site lymphoma/leukemia GEF', map_location='7 16.71 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28359603, end_position_on_the_genomic_accession=28372687, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  101502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101502, CurrentID=0, Status='live', Symbol='Hsd3b7', Aliases='AI195443, BB098564', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase 7', other_designations='3 beta-hydroxysteroid dehydrogenase type VII|3-beta-HSD VII|3-beta-hydroxy-Delta(5)-C27 steroid oxidoreductase|c(27) 3-beta-HSD|cholest-5-ene-3-beta,7-alpha-diol 3-beta-dehydrogenase', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127800609, end_position_on_the_genomic_accession=127803802, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  101513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101513, CurrentID=0, Status='live', Symbol='Mob2', Aliases='1110017M21Rik, 2700078K21Rik, AI256456, Mmh', description='MOB kinase activator 2', other_designations='Mps one binder kinase activator-like 2|mob2 homolog|ovary-specific MOB-like protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142008535, end_position_on_the_genomic_accession=142061034, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101533, CurrentID=0, Status='live', Symbol='Klk9', Aliases='1200016C12Rik, AI324041', description='kallikrein related-peptidase 9', other_designations='kallikrein 9|kallikrein-like protein 3', map_location='7 28.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43791588, end_position_on_the_genomic_accession=43796758, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101540, CurrentID=0, Status='live', Symbol='Prkd2', Aliases='AI325941, PKD2', description='protein kinase D2', other_designations='nPKC-D2', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16842902, end_position_on_the_genomic_accession=16870461, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  101543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101543, CurrentID=0, Status='live', Symbol='Wtip', Aliases='', description='WT1-interacting protein', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34109541, end_position_on_the_genomic_accession=34133525, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  101544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101544, CurrentID=0, Status='live', Symbol='Zfp575', Aliases='AI326876, Znf575', description='zinc finger protein 575', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24583838, end_position_on_the_genomic_accession=24587641, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  16924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=16924, CurrentID=0, Status='live', Symbol='Lnx1', Aliases='Lnx', description='ligand of numb-protein X 1', other_designations='ligand of Numb protein X 1|ligand of Numb-binding protein 1|numb-binding protein 1', map_location='5 D', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=74592447, end_position_on_the_genomic_accession=74703042, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  100502698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502698, CurrentID=0, Status='live', Symbol='Rubcn', Aliases='1700021K19Rik, 5330403K09, Kiaa0226, Rubicon, mKIAA0226', description='RUN domain and cysteine-rich domain containing, Beclin 1-interacting protein', other_designations='run domain Beclin-1 interacting and cystein-rich containing protein|run domain Beclin-1 interacting and cysteine-rich containing protein', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32821702, end_position_on_the_genomic_accession=32877348, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  101565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101565, CurrentID=0, Status='live', Symbol='Ccp110', Aliases='6330503K22Rik, AA415922, AI427129, AW557948, CP110', description='centriolar coiled coil protein 110', other_designations='centrosomal protein CP110|centrosomal protein of 110 kDa|cep110', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118712516, end_position_on_the_genomic_accession=118737024, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  101568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101568, CurrentID=0, Status='live', Symbol='Vrk3', Aliases='AI428238', description='vaccinia related kinase 3', other_designations='serine/threonine-protein kinase VRK3|serine/threonine-protein pseudokinase VRK3|vaccinia-related kinase 3', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44748629, end_position_on_the_genomic_accession=44777515, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  100502723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502723, CurrentID=0, Status='live', Symbol='Tsga8', Aliases='Halap-X, Halapx', description='testis specific gene A8', other_designations='haploid specific alanine rich acidic protein, X chromosome|testis-specific protein A8', map_location='X 38.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=83486674, end_position_on_the_genomic_accession=83487924, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  101592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101592, CurrentID=0, Status='live', Symbol='Efl1', Aliases='4932434J20Rik, 6030468D11Rik, AI451340, AU019507, AU022896, D7Ertd791e, Eftud1', description='elongation factor like GPTase 1', other_designations='elongation factor Tu GTP binding domain containing 1', map_location='7 47.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=82648614, end_position_on_the_genomic_accession=82777852, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  232670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232670, CurrentID=0, Status='live', Symbol='Tspan33', Aliases='1300010A20Rik, Pen', description='tetraspanin 33', other_designations='penumbra|proerythroblast new membrane', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29694214, end_position_on_the_genomic_accession=29718558, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  101602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101602, CurrentID=0, Status='live', Symbol='AI467606', Aliases='', description='expressed sequence AI467606', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127090767, end_position_on_the_genomic_accession=127094049, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101604, CurrentID=0, Status='live', Symbol='E430018J23Rik', Aliases='AI480612', description='RIKEN cDNA E430018J23 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127390830, end_position_on_the_genomic_accession=127393662, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  232679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232679, CurrentID=0, Status='live', Symbol='Zc3hc1', Aliases='1110054L24Rik, AU018540, AU019789, HSPC216, Nipa', description='zinc finger, C3HC type 1', other_designations='nuclear interacting partner of anaplastic lymphoma kinase (Alk)|zinc finger C3HC-type protein 1', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30366384, end_position_on_the_genomic_accession=30391040, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  232680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232680, CurrentID=0, Status='live', Symbol='Cpa2', Aliases='', description='carboxypeptidase A2, pancreatic', other_designations='', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30541642, end_position_on_the_genomic_accession=30564476, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  101612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101612, CurrentID=0, Status='live', Symbol='Grwd1', Aliases='A301, AI504353, Grwd, Wdr28', description='glutamate-rich WD repeat containing 1', other_designations='glutamate rich WD repeat protein GRWD', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45825223, end_position_on_the_genomic_accession=45830796, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101613, CurrentID=0, Status='live', Symbol='Nlrp6', Aliases='AI504961, Avr, Nalp6, Navr, Navr/Avr, Non-AVR, Pypaf5', description='NLR family, pyrin domain containing 6', other_designations='NACHT, leucine rich repeat and PYD containing 6|NACHT/LRR/pyrin domain-containing protein 6|PYRIN-containing APAF1-like protein 5|angiotensin II/vasopressin receptor|non-angiotensin-vasopressin receptor', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140920902, end_position_on_the_genomic_accession=140929192, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100502766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502766, CurrentID=0, Status='live', Symbol='Kifc1', Aliases='100042970, Gm4137, HSET, KNSL2, Kifc5a, Knsl2a, Tctex-7, Tctex-7A, Tctex7, Tctex7a', description='kinesin family member C1', other_designations='kifc1|kinesin family c-terminal 5A|kinesin motor protein KIFC4|kinesin-like 2a|kinesin-related protein KIFC5A', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33875664, end_position_on_the_genomic_accession=33890633, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  101631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101631, CurrentID=0, Status='live', Symbol='Pwwp2b', Aliases='AI594893, D7Ertd517e, D930023J19Rik, Pwwp2', description='PWWP domain containing 2B', other_designations='PWWP domain containing 2', map_location='7 84.39 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139248482, end_position_on_the_genomic_accession=139267253, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  232714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232714, CurrentID=0, Status='live', Symbol='Mgam', Aliases='6030407P20Rik, MGA', description='maltase-glucoamylase', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40628825, end_position_on_the_genomic_accession=40769123, orientation='plus', exon_count=96, OMIM='', no_hdr0=''),
  232717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232717, CurrentID=0, Status='live', Symbol='Prss58', Aliases='1700020H15, Tryx3', description='protease, serine 58', other_designations='putative trypsin-X3|trypsin-X3', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40895262, end_position_on_the_genomic_accession=40900387, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  232748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232748, CurrentID=0, Status='live', Symbol='Tcaf2', Aliases='Fam115c', description='TRPM8 channel-associated factor 2', other_designations='TRP channel-associated factor 2|family with sequence similarity 115, member C|protein FAM115C', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42623043, end_position_on_the_genomic_accession=42645041, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101685, CurrentID=0, Status='live', Symbol='Spty2d1', Aliases='5830435K17Rik, AI852426, P16H6', description='SPT2, Suppressor of Ty, domain containing 1 (S. cerevisiae)', other_designations='SPT2 domain-containing protein 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46990396, end_position_on_the_genomic_accession=47008414, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100502841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502841, CurrentID=0, Status='live', Symbol='Epg5', Aliases='4732475F16, 5430411K18Rik, AI661957, AW456499, mKIAA1632', description='ectopic P-granules autophagy protein 5 homolog (C. elegans)', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77938408, end_position_on_the_genomic_accession=78037323, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  101700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101700, CurrentID=0, Status='live', Symbol='Trim68', Aliases='AI894158, F730114J12Rik, Rnf137, SS-56', description='tripartite motif-containing 68', other_designations='ring finger protein 137|tripartite motif-containing protein 68', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102677579, end_position_on_the_genomic_accession=102687329, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101706, CurrentID=0, Status='live', Symbol='Numa1', Aliases='6720401E04Rik, AA764025, AL022610, AU014979', description='nuclear mitotic apparatus protein 1', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101934002, end_position_on_the_genomic_accession=102014959, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  100502861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502861, CurrentID=0, Status='live', Symbol='Ccdc13', Aliases='2900041A11Rik, Gm188', description='coiled-coil domain containing 13', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121797624, end_position_on_the_genomic_accession=121839515, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  232784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232784, CurrentID=0, Status='live', Symbol='Zfp212', Aliases='Znf212, mKIAA3011', description='Zinc finger protein 212', other_designations='Zinc finger protein C2H2-150', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=47920568, end_position_on_the_genomic_accession=47932639, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232790, CurrentID=0, Status='live', Symbol='Oscar', Aliases='mOSCAR, mOSCAR-M1, mOSCAR-M2, mOSCAR-M3', description='osteoclast associated receptor', other_designations='ocl178|osteoclast-associated receptor OSCAR', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3609813, end_position_on_the_genomic_accession=3616157, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  232791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232791, CurrentID=0, Status='live', Symbol='Cnot3', Aliases='A930039N10Rik', description='CCR4-NOT transcription complex, subunit 3', other_designations='CCR4-associated factor 3', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3644522, end_position_on_the_genomic_accession=3661109, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  232798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232798, CurrentID=0, Status='live', Symbol='Leng8', Aliases='AW049671', description='leukocyte receptor cluster (LRC) member 8', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4137047, end_position_on_the_genomic_accession=4148177, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  232801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232801, CurrentID=0, Status='live', Symbol='Lilra5', Aliases='E030017K20, EG232801, Gm4878, Lilrc1', description='leukocyte immunoglobulin-like receptor, subfamily A (with TM domain), member 5', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4237643, end_position_on_the_genomic_accession=4243690, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  232807 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232807, CurrentID=0, Status='live', Symbol='Ppp1r12c', Aliases='2410197A17Rik, AI839747, B430101M08, Mbs85', description='protein phosphatase 1, regulatory (inhibitor) subunit 12C', other_designations='myosin-binding subunit 85|ppp1r12c|protein phosphatase 1 myosin-binding subunit of 85 kDa|protein phosphatase 1 myosin-binding subunit p85', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4481519, end_position_on_the_genomic_accession=4502222, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  101739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101739, CurrentID=0, Status='live', Symbol='Psip1', Aliases='AA408851, AU015605, Dfs70, Ledgfa, Ledgfb, Psip2', description='PC4 and SFRS1 interacting protein 1', other_designations='PC4 and SFRS1 interacting protein 2|dense fine speckles 70|lens epithelium-derived growth factor', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=83455680, end_position_on_the_genomic_accession=83486638, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  232813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232813, CurrentID=0, Status='live', Symbol='Shisa7', Aliases='', description='shisa family member 7', other_designations='shisa homolog 7', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4825551, end_position_on_the_genomic_accession=4844789, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  232816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232816, CurrentID=0, Status='live', Symbol='Zfp628', Aliases='Zec, Znf628', description='zinc finger protein 628', other_designations='zinc finger protein Zec|zinc finger protein expressed in embryonal cells and certain adult organs', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4915176, end_position_on_the_genomic_accession=4922003, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  232821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232821, CurrentID=0, Status='live', Symbol='Ccdc106', Aliases='BC018462', description='coiled-coil domain containing 106', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5053821, end_position_on_the_genomic_accession=5060785, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  232827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232827, CurrentID=0, Status='live', Symbol='Nlrp2', Aliases='E330007A02Rik, NBS1, Nalp2, PAN1, PYPAF2', description='NLR family, pyrin domain containing 2', other_designations='NACHT, leucine rich repeat and PYD containing 2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5298547, end_position_on_the_genomic_accession=5351035, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  232836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232836, CurrentID=0, Status='live', Symbol='Galp', Aliases='', description='galanin-like peptide', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6197090, end_position_on_the_genomic_accession=6216138, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  101772 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101772, CurrentID=0, Status='live', Symbol='Ano1', Aliases='Tmem16a', description='anoctamin 1, calcium activated chloride channel', other_designations='transmembrane protein 16A', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144588549, end_position_on_the_genomic_accession=144752039, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  232853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232853, CurrentID=0, Status='live', Symbol='Zfp954', Aliases='5730403M16Rik, 8030437F07, AA123315', description='zinc finger protein 954', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7114683, end_position_on_the_genomic_accession=7121476, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  232854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232854, CurrentID=0, Status='live', Symbol='Zfp418', Aliases='A230102I05Rik', description='zinc finger protein 418', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7171353, end_position_on_the_genomic_accession=7183560, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  232855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232855, CurrentID=0, Status='live', Symbol='Zfp772', Aliases='', description='zinc finger protein 772', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=7202122, end_position_on_the_genomic_accession=7209998, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100502950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502950, CurrentID=0, Status='live', Symbol='Gm3336', Aliases='100041434, 2410018E23Rik', description='predicted gene 3336', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70718543, end_position_on_the_genomic_accession=70722635, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  100502953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502953, CurrentID=0, Status='live', Symbol='Gm10100', Aliases='ENSMUSG00000061452', description='predicted gene 10100', other_designations='', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77726486, end_position_on_the_genomic_accession=77726848, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  232878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232878, CurrentID=0, Status='live', Symbol='Zscan22', Aliases='AW227621, D530006B18Rik, Hkr2', description='zinc finger and SCAN domain containing 22', other_designations='GLI-Kruppel family member HKR2', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12897809, end_position_on_the_genomic_accession=12909083, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  232879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232879, CurrentID=0, Status='live', Symbol='Zbtb45', Aliases='BB161562, Gm157, Zfp499', description='zinc finger and BTB domain containing 45', other_designations='zinc finger protein 499', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13005666, end_position_on_the_genomic_accession=13010168, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  101809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101809, CurrentID=0, Status='live', Symbol='Spred3', Aliases='AW047143, D130060H24Rik, Spred-3', description='sprouty-related, EVH1 domain containing 3', other_designations='Sprouty-related EVH1 domain-containing protein', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29157783, end_position_on_the_genomic_accession=29169854, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100502967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100502967, CurrentID=0, Status='live', Symbol='Speer4c', Aliases='4933431D05Rik, SPEER-4C', description='spermatogenesis associated glutamate (E)-rich protein 4C', other_designations='sperm-associated glutamate (E)-rich protein 4c|spermatogenic-associated glutamate (E)-rich protein 4c', map_location='5 6.52 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=15709505, end_position_on_the_genomic_accession=15714675, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  232889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232889, CurrentID=0, Status='live', Symbol='Pla2g4c', Aliases='D7Ertd445e', description='phospholipase A2, group IVC (cytosolic, calcium-independent)', other_designations='CPLA2-gamma|egg and early embryo abundant protein', map_location='7 7.84 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13311522, end_position_on_the_genomic_accession=13360674, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  101831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101831, CurrentID=0, Status='live', Symbol='Faap24', Aliases='AW124591, AW538696, C230052I12Rik', description='Fanconi anemia core complex associated protein 24', other_designations='Fanconi anemia-associated protein, 24 kDa|fanconi anemia-associated protein of 24 kDa', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35392152, end_position_on_the_genomic_accession=35396767, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  232906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232906, CurrentID=0, Status='live', Symbol='Arhgap35', Aliases='6430596G11Rik, AI841135, Grlf1, mKIAA1722, p190A, p190RhoGAP', description='Rho GTPase activating protein 35', other_designations='P190 RhoGAP|glucocorticoid receptor DNA binding factor 1|glucocorticoid receptor DNA-binding factor 1', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16494473, end_position_on_the_genomic_accession=16614993, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101835, CurrentID=0, Status='live', Symbol='AW146154', Aliases='AI451288', description='expressed sequence AW146154', other_designations='', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=41478874, end_position_on_the_genomic_accession=41499890, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  232910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232910, CurrentID=0, Status='live', Symbol='Ap2s1', Aliases='AI043088', description='adaptor-related protein complex 2, sigma 1 subunit', other_designations='adapter-related protein complex 2 sigma subunit|adapter-related protein complex 2 subunit sigma|adaptor protein complex AP-2 subunit sigma|adaptor-related protein complex 2 subunit sigma|clathrin assembly protein 2 sigma small chain|clathrin assembly protein 2 small chain|clathrin coat assembly protein AP17|clathrin coat-associated protein AP17|plasma membrane adaptor AP-2 17 kDa protein|sigma-adaptin 3b|sigma2-adaptin', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16738444, end_position_on_the_genomic_accession=16749290, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232925, CurrentID=0, Status='live', Symbol='Igfl3', Aliases='Gm580, Igfl, Rprc483', description='IGF-like family member 3', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18176494, end_position_on_the_genomic_accession=18181862, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  232933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232933, CurrentID=0, Status='live', Symbol='Ccdc61', Aliases='C530028I08Rik, Gm159', description='coiled-coil domain containing 61', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18890883, end_position_on_the_genomic_accession=18910412, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  232934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232934, CurrentID=0, Status='live', Symbol='Mypop', Aliases='Myodysa, P42pop', description='Myb-related transcription factor, partner of profilin', other_designations='Myb protein P42POP|myb-related protein p42POP|myotonic dystrophy region|partner of profilin', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18990058, end_position_on_the_genomic_accession=19001766, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  101867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101867, CurrentID=0, Status='live', Symbol='Rrp8', Aliases='1500003O22Rik, 2900001K19Rik, AW538116', description='ribosomal RNA processing 8, methyltransferase, homolog (yeast)', other_designations='cerebral protein 1 homolog', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105731730, end_position_on_the_genomic_accession=105737385, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  101869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101869, CurrentID=0, Status='live', Symbol='Unc45a', Aliases='AW538196', description='unc-45 homolog A (C. elegans)', other_designations='SMAP-1|UNC-45A|smooth muscle cell associated protein-1|stromal membrane-associated protein 1', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80325292, end_position_on_the_genomic_accession=80341000, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  232943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232943, CurrentID=0, Status='live', Symbol='Klc3', Aliases='BC017147, Klct', description='kinesin light chain 3', other_designations='', map_location='7 9.62 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19394440, end_position_on_the_genomic_accession=19404442, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  232944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232944, CurrentID=0, Status='live', Symbol='Mark4', Aliases='2410090P21Rik, C79806, Markl1', description='MAP/microtubule affinity regulating kinase 4', other_designations='MAP/microtubule affinity-regulating kinase 4L|MARK4 serine/threonine protein kinase', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19425102, end_position_on_the_genomic_accession=19458795, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  232946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232946, CurrentID=0, Status='live', Symbol='Bloc1s3', Aliases='BC043666, Blos3, E230011O18, rp', description='biogenesis of lysosomal organelles complex-1, subunit 3', other_designations='BLOC-1 subunit 3|Bloc-1 complex component|Bloc1 subunit 3|biogenesis of lysosome-related organelles complex-1, subunit 3|reduced pigmentation protein', map_location='7 9.92 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19505804, end_position_on_the_genomic_accession=19508371, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  232947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232947, CurrentID=0, Status='live', Symbol='Ppp1r37', Aliases='Gm158, Lrrc68', description='protein phosphatase 1, regulatory subunit 37', other_designations='leucine rich repeat containing 68|leucine-rich repeat-containing protein 68', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19530967, end_position_on_the_genomic_accession=19563071, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101883, CurrentID=0, Status='live', Symbol='Igflr1', Aliases='F630003I17Rik, Tmem149', description='IGF-like family receptor 1', other_designations='U2 small nuclear RNA auxiliary factor 1-like 4|transmembrane protein 149', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30565427, end_position_on_the_genomic_accession=30567962, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100503036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503036, CurrentID=0, Status='live', Symbol='1700015G11Rik', Aliases='Ccdc179', description='RIKEN cDNA 1700015G11 gene', other_designations='', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=52011680, end_position_on_the_genomic_accession=52015716, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  232959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232959, CurrentID=0, Status='live', Symbol='Vmn1r178', Aliases='Gm473, V1rd13', description='vomeronasal 1 receptor 178', other_designations='vomernasal 1 receptor Vmn1r178|vomeronasal 1 receptor, D13|vomeronasal receptor V1RD13|vomeronasal type 1 receptor D13', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23893529, end_position_on_the_genomic_accession=23894443, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100503041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503041, CurrentID=0, Status='live', Symbol='Pdzd7', Aliases='9130207N01, EG435601, Pdzk7', description='PDZ domain containing 7', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45026907, end_position_on_the_genomic_accession=45058308, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  232962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232962, CurrentID=0, Status='live', Symbol='Vmn1r180', Aliases='Gm475, V1rd16', description='vomeronasal 1 receptor 180', other_designations='vomernasal 1 receptor Vmn1r180|vomeronasal receptor V1RD16', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23952387, end_position_on_the_genomic_accession=23953356, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100503043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503043, CurrentID=0, Status='live', Symbol='Armcx4', Aliases='AI448196, Gm195', description='armadillo repeat containing, X-linked 4', other_designations='', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134685651, end_position_on_the_genomic_accession=134697775, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  232969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232969, CurrentID=0, Status='live', Symbol='Zfp428', Aliases='2410005H09Rik, AV271806, AW047854, Znf428', description='zinc finger protein 428', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24507087, end_position_on_the_genomic_accession=24515682, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232973, CurrentID=0, Status='live', Symbol='Lypd4', Aliases='4933400F01Rik', description='Ly6/Plaur domain containing 4', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24864619, end_position_on_the_genomic_accession=24869772, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  232975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232975, CurrentID=0, Status='live', Symbol='Atp1a3', Aliases='Atpa-2', description='ATPase, Na+/K+ transporting, alpha 3 polypeptide', other_designations='Na(+)/K(+) ATPase alpha(III) subunit|Na(+)/K(+) ATPase alpha-3 subunit|sodium pump subunit alpha-3', map_location='7 13.73 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24978167, end_position_on_the_genomic_accession=25005937, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  232976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232976, CurrentID=0, Status='live', Symbol='Zfp574', Aliases='A630056B21Rik, Znf574', description='zinc finger protein 574', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25075684, end_position_on_the_genomic_accession=25083492, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  232983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232983, CurrentID=0, Status='live', Symbol='Cxcl17', Aliases='VCC-1, Vcc1', description='chemokine (C-X-C motif) ligand 17', other_designations='13.6 kDa protein|13.6-kDa protein|C-X-C motif chemokine 17|VEGF co-regulated chemokine 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25400053, end_position_on_the_genomic_accession=25412886, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  232984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232984, CurrentID=0, Status='live', Symbol='B3gnt8', Aliases='B3galt7, B7galt7, BC025206', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 8', other_designations='BGnT-8|UDP-Gal:betaGal beta 1,3-galactosyltransferase polypeptide 7|beta-1,3-Gn-T8|beta-1,3-N-acetylglucosaminyltransferase 8|beta3Gn-T8', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25627587, end_position_on_the_genomic_accession=25629492, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  232987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232987, CurrentID=0, Status='live', Symbol='B9d2', Aliases='stumpy', description='B9 protein domain 2', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25680734, end_position_on_the_genomic_accession=25686562, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  232989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=232989, CurrentID=0, Status='live', Symbol='Hnrnpul1', Aliases='E130317O14Rik, E1B-AP5, E1BAP5, Hnrnpul, Hnrpul1', description='heterogeneous nuclear ribonucleoprotein U-like 1', other_designations='EIB-55kDa associated protein 5', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25721165, end_position_on_the_genomic_accession=25756245, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  233001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233001, CurrentID=0, Status='live', Symbol='Nlrp9a', Aliases='D7Ertd565e, Nalp-theta, Nalp9a', description='NLR family, pyrin domain containing 9A', other_designations='', map_location='7 15.08 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26535023, end_position_on_the_genomic_accession=26574148, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  233005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233005, CurrentID=0, Status='live', Symbol='Cyp2a22', Aliases='EG233005', description='cytochrome P450, family 2, subfamily a, polypeptide 22', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26931631, end_position_on_the_genomic_accession=26939412, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  233011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233011, CurrentID=0, Status='live', Symbol='Itpkc', Aliases='9130023N17Rik', description='inositol 1,4,5-trisphosphate 3-kinase C', other_designations='IP3 3-kinase C|IP3K C|insP 3-kinase C', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27207170, end_position_on_the_genomic_accession=27228597, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  101943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101943, CurrentID=0, Status='live', Symbol='Sf3b3', Aliases='1810061H24Rik, 5730409A01Rik, AA409318, D8Ertd633e, RSE1, SAP130, mKIAA0017', description='splicing factor 3b, subunit 3', other_designations='SAP 130|SF3b130|pre-mRNA-splicing factor SF3b 130 kDa subunit|spliceosome-associated protein 130', map_location='8 57.73 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110810491, end_position_on_the_genomic_accession=110846803, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  233016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233016, CurrentID=0, Status='live', Symbol='Blvrb', Aliases='', description='biliverdin reductase B (flavin reductase (NADPH))', other_designations='BVR-B|FLR|FR|NADPH-dependent diaphorase|NADPH-flavin reductase|biliverdin-IX beta-reductase', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27447978, end_position_on_the_genomic_accession=27465981, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  233020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233020, CurrentID=0, Status='live', Symbol='Hipk4', Aliases='Gm162', description='homeodomain interacting protein kinase 4', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27523261, end_position_on_the_genomic_accession=27531180, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  233033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233033, CurrentID=0, Status='live', Symbol='Samd4b', Aliases='BC042901', description='sterile alpha motif domain containing 4B', other_designations='mSmaug2|smaug 2|sterile alpha motif domain-containing protein 4B', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28399522, end_position_on_the_genomic_accession=28436191, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  101966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101966, CurrentID=0, Status='live', Symbol='D8Ertd738e', Aliases='AA589584', description='DNA segment, Chr 8, ERATO Doi 738, expressed', other_designations='', map_location='8 40.72 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84246235, end_position_on_the_genomic_accession=84249761, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  233040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233040, CurrentID=0, Status='live', Symbol='Fbxo27', Aliases='E130008B10Rik, FBG5, Gm161', description='F-box protein 27', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28692779, end_position_on_the_genomic_accession=28699338, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  233046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233046, CurrentID=0, Status='live', Symbol='Rasgrp4', Aliases='', description='RAS guanyl releasing protein 4', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29134933, end_position_on_the_genomic_accession=29153961, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  233056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233056, CurrentID=0, Status='live', Symbol='Zfp790', Aliases='6330581L23Rik', description='zinc finger protein 790', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29816005, end_position_on_the_genomic_accession=29832042, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  101985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101985, CurrentID=0, Status='live', Symbol='Usb1', Aliases='AA960436', description='U6 snRNA biogenesis 1', other_designations='UPF0406 protein C16orf57 homolog|putative U6 snRNA phosphodiesterase', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95332284, end_position_on_the_genomic_accession=95347513, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  233058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233058, CurrentID=0, Status='live', Symbol='Zfp420', Aliases='B230119D13, B230312I18Rik, mszf59-1, mszf70', description='zinc finger protein 420', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29859960, end_position_on_the_genomic_accession=29877302, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  233060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233060, CurrentID=0, Status='live', Symbol='Zfp382', Aliases='5930415A09Rik, Znf382', description='zinc finger protein 382', other_designations='KRAB/zinc finger suppressor protein 1|KS1|multiple zinc finger and krueppel-associated box protein KS1', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30121569, end_position_on_the_genomic_accession=30136060, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  233064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233064, CurrentID=0, Status='live', Symbol='Wdr62', Aliases='2310038K02Rik', description='WD repeat domain 62', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30240138, end_position_on_the_genomic_accession=30280427, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  233065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233065, CurrentID=0, Status='live', Symbol='Alkbh6', Aliases='AI839550, Abh6, C130099A02', description='alkB homolog 6', other_designations='alkB, alkylation repair homolog 6|alkylated DNA repair protein alkB homolog 6|probable alpha-ketoglutarate-dependent dioxygenase ABH6', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30308753, end_position_on_the_genomic_accession=30314303, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  233066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233066, CurrentID=0, Status='live', Symbol='Syne4', Aliases='0610012K07Rik, AI428936, Nesp4', description='spectrin repeat containing, nuclear envelope family member 4', other_designations='nuclear envelope spectrin repeat protein 4', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30314816, end_position_on_the_genomic_accession=30319045, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  233067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233067, CurrentID=0, Status='live', Symbol='Lrfn3', Aliases='A530045B06Rik, SALM4', description='leucine rich repeat and fibronectin type III domain containing 3', other_designations='synaptic adhesion-like molecule 4', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30355514, end_position_on_the_genomic_accession=30362772, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  233071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233071, CurrentID=0, Status='live', Symbol='Arhgap33', Aliases='AW324378, NOMA-GAP, Snx26, TCGAP', description='Rho GTPase activating protein 33', other_designations='TC10/CDC42 GTPase-activating protein|rho-type GTPase-activating protein 33|sorting nexin 26', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30522226, end_position_on_the_genomic_accession=30535092, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  233073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233073, CurrentID=0, Status='live', Symbol='U2af1l4', Aliases='AA407033, AF419339, AI451269, AW553050, U2af26', description='U2 small nuclear RNA auxiliary factor 1-like 4', other_designations='U2 auxiliary factor 26|U2 small nuclear RNA auxiliary factor 1-like protein 4|U2AF1-like 4|auxiliary factor 26', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30563340, end_position_on_the_genomic_accession=30565364, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  233079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233079, CurrentID=0, Status='live', Symbol='Ffar2', Aliases='GPCR43, Gpr43', description='free fatty acid receptor 2', other_designations='G protein-coupled receptor 43|G-protein coupled receptor 43|leukocyte-specific STAT-induced GPCR', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30818032, end_position_on_the_genomic_accession=30823797, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  233080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233080, CurrentID=0, Status='live', Symbol='Ffar3', Aliases='Gm478, Gpr41', description='free fatty acid receptor 3', other_designations='G protein-coupled receptor 41|G-protein coupled receptor 41|putative G protein-coupled receptor GPR41', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30854330, end_position_on_the_genomic_accession=30856178, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  233081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233081, CurrentID=0, Status='live', Symbol='Ffar1', Aliases='Gpr40', description='free fatty acid receptor 1', other_designations='G protein-coupled receptor 40|G protein-coupled receptor GPR40|G-protein coupled receptor 40', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30860568, end_position_on_the_genomic_accession=30861470, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  626305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626305, CurrentID=0, Status='live', Symbol='Scgb1b7', Aliases='Abpa7, EG626305, Gm6662', description='secretoglobin, family 1B, member 7', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31712665, end_position_on_the_genomic_accession=31713980, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  233090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233090, CurrentID=0, Status='live', Symbol='Scgb2b24', Aliases='Abpbg24, Abpz, C2b, Scgb2b3', description='secretoglobin, family 2B, member 24', other_designations='allergen dI chain C2B|androgen binding protein zeta|androgen-binding protein zeta|lacrimal androgen-binding protein zeta|secretoglobin, family 2B, member 3', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33737193, end_position_on_the_genomic_accession=33739295, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102022, CurrentID=0, Status='live', Symbol='Ces2a', Aliases='9130231C15Rik, AI266984, Ces6', description='carboxylesterase 2A', other_designations='carboxylesterase 6', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104734003, end_position_on_the_genomic_accession=104741634, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  233099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233099, CurrentID=0, Status='live', Symbol='Scgb2b27', Aliases='Abpbg27, Abpbg5, Scgb2b5', description='secretoglobin, family 2B, member 27', other_designations='allergen dI chain C2A|androgen binding protein beta|secretoglobin, family 2B, member 5', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34011919, end_position_on_the_genomic_accession=34013942, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  626316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626316, CurrentID=0, Status='live', Symbol='Gm13051', Aliases='OTTMUSG00000010104, OTTMUSG00000010173', description='predicted gene 13051', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=146097312, end_position_on_the_genomic_accession=146126624, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  233103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233103, CurrentID=0, Status='live', Symbol='4931406P16Rik', Aliases='AA553327, mKIAA0355', description='RIKEN cDNA 4931406P16 gene', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34236707, end_position_on_the_genomic_accession=34313575, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  102032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102032, CurrentID=0, Status='live', Symbol='Smim19', Aliases='AI316807, BG694986, EST-c39', description='small integral membrane protein 19', other_designations='UPF0697 protein C8orf40 homolog', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22462614, end_position_on_the_genomic_accession=22476879, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  233107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233107, CurrentID=0, Status='live', Symbol='Kctd15', Aliases='BC031749', description='potassium channel tetramerisation domain containing 15', other_designations='potassium channel tetramerization domain-containing protein 15', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=34639015, end_position_on_the_genomic_accession=34655507, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  233115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233115, CurrentID=0, Status='live', Symbol='Dpy19l3', Aliases='6030410G08, 9330164H19Rik', description='dpy-19-like 3 (C. elegans)', other_designations='dpy-19-like protein 3|protein dpy-19 homolog 3', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35685500, end_position_on_the_genomic_accession=35754454, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  100503199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503199, CurrentID=0, Status='live', Symbol='5430416N02Rik', Aliases='100041797, Adapt33, Gm3514', description='RIKEN cDNA 5430416N02 gene', other_designations='adaptive response 33', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=100420842, end_position_on_the_genomic_accession=100429535, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102058, CurrentID=0, Status='live', Symbol='Exoc8', Aliases='AI414418, EXO84, Exo84p, R74783, SEC84', description='exocyst complex component 8', other_designations='exocyst complex 84 kDa subunit', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124890299, end_position_on_the_genomic_accession=124897705, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  626347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626347, CurrentID=0, Status='live', Symbol='Igkv3-4', Aliases='Igk-V21-4', description='immunoglobulin kappa variable 3-4', other_designations='V kappa 21-4|immunoglobulin kappa chain variable 21 (V21)-4|immunoglobulin kappa chain variable 3-4|immunoglobulin kappa light chain|monoclonal antibody kappa light chain', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70671789, end_position_on_the_genomic_accession=70672377, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  102060 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102060, CurrentID=0, Status='live', Symbol='Gadd45gip1', Aliases='2310040G17Rik, AI425883, Crif1, MRP-L59, Plinp1', description='growth arrest and DNA-damage-inducible, gamma interacting protein 1', other_designations='39S ribosomal protein L59, mitochondrial|growth arrest and DNA-damage-inducible proteins-interacting protein 1|papillomavirus L2 interacting nuclear protein 1', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84832282, end_position_on_the_genomic_accession=84835482, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  233147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233147, CurrentID=0, Status='live', Symbol='Zfp939', Aliases='9430025M13Rik', description='zinc finger protein 939', other_designations='', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=39449518, end_position_on_the_genomic_accession=39477427, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100503240 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503240, CurrentID=0, Status='live', Symbol='Trpc5os', Aliases='ENSMUSG00000072934, Gm15070, OTTMUSG00000018931', description='transient receptor potential cation channel, subfamily C, member 5, opposite strand', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=144455781, end_position_on_the_genomic_accession=144477063, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100503242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503242, CurrentID=0, Status='live', Symbol='Scgb2b6', Aliases='Abpbg6, Gm19607', description='secretoglobin, family 2B, member 6', other_designations='androgen-binding protein delta-like', map_location='7 19.33 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31617738, end_position_on_the_genomic_accession=31619526, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102093, CurrentID=0, Status='live', Symbol='Phkb', Aliases='AI463271', description='phosphorylase kinase beta', other_designations='phosphorylase kinase subunit beta', map_location='8 41.61 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85840984, end_position_on_the_genomic_accession=86064254, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  102098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102098, CurrentID=0, Status='live', Symbol='Arhgef18', Aliases='AI467246, D030053O22Rik', description='rho/rac guanine nucleotide exchange factor (GEF) 18', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3392943, end_position_on_the_genomic_accession=3456601, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  102103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102103, CurrentID=0, Status='live', Symbol='Mtus1', Aliases='AI481402, ATBP135, Atip1, B430010I23Rik, B430305I03Rik, C85752, Cctsg1-440, MD44, MTSG1, mKIAA1288', description='mitochondrial tumor suppressor 1', other_designations='AT2 receptor-binding protein|AT2 receptor-interacting protein 1|angiotensin II AT2 receptor-interacting protein 1|angiotensin-II type 2 receptor-interacting protein|coiled-coiled tumor suppressor gene 1 protein|coiled-coiled tumor supressor gene 1|mitochondrial tumor suppressor 1 homolog', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40990912, end_position_on_the_genomic_accession=41133983, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  233186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233186, CurrentID=0, Status='live', Symbol='Siglecf', Aliases='Siglec5, mSiglec-F', description='sialic acid binding Ig-like lectin F', other_designations='sialic acid binding Ig-like lectin 5|sialic acid-binding Ig-like lectin F|sialic acid-binding lectin Siglec-F|siglec-5|siglec-F', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43351341, end_position_on_the_genomic_accession=43359475, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  233187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233187, CurrentID=0, Status='live', Symbol='Lim2', Aliases='19kDa, 4833403J20, MP19, To3', description='lens intrinsic membrane protein 2', other_designations='MP17|MP18|MP20|total opacity 3', map_location='7 28.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43430101, end_position_on_the_genomic_accession=43435991, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  233189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233189, CurrentID=0, Status='live', Symbol='Ctu1', Aliases='Atpbd3, BC005752', description='cytosolic thiouridylase subunit 1', other_designations='ATP binding domain 3|ATP-binding domain-containing protein 3|cytoplasmic tRNA adenylyltransferase 1|cytosolic thiouridylase subunit 1 homolog', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43672031, end_position_on_the_genomic_accession=43678298, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  102122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102122, CurrentID=0, Status='live', Symbol='Fam192a', Aliases='1700001O11Rik, 2310065K24Rik, AI596259, Nip30', description='family with sequence similarity 192, member A', other_designations='NEFA-interacting nuclear protein NIP30', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94574940, end_position_on_the_genomic_accession=94601726, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  102124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102124, CurrentID=0, Status='live', Symbol='Enkd1', Aliases='AI606951, E130303B06Rik', description='enkurin domain containing 1', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105703649, end_position_on_the_genomic_accession=105708221, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  233199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233199, CurrentID=0, Status='live', Symbol='Mybpc2', Aliases='', description='myosin binding protein C, fast-type', other_designations='C-protein, skeletal muscle fast isoform|fast MyBP-C|fast-type C-protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44501699, end_position_on_the_genomic_accession=44524715, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  100503280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503280, CurrentID=0, Status='live', Symbol='Gm38425', Aliases='', description='predicted gene, 38425', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74966844, end_position_on_the_genomic_accession=74969125, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  233204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233204, CurrentID=0, Status='live', Symbol='Tbc1d17', Aliases='BC017607', description='TBC1 domain family, member 17', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44840776, end_position_on_the_genomic_accession=44849079, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  233210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233210, CurrentID=0, Status='live', Symbol='Prr12', Aliases='6720469B10, BC058674', description='proline rich 12', other_designations='', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45027707, end_position_on_the_genomic_accession=45052881, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  102141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102141, CurrentID=0, Status='live', Symbol='Snx25', Aliases='AI661919, Gm1699, Gm173, Sbbi31', description='sorting nexin 25', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46033261, end_position_on_the_genomic_accession=46151858, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  233221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233221, CurrentID=0, Status='live', Symbol='Mrgpra1', Aliases='MrgA1', description='MAS-related GPR, member A1', other_designations='RF-amide G protein-coupled receptor|RF-amide G-protein coupled receptor', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=47334875, end_position_on_the_genomic_accession=47354240, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  233222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233222, CurrentID=0, Status='live', Symbol='Mrgpra3', Aliases='Mrga3', description='MAS-related GPR, member A3', other_designations='G protein-coupled receptor MRGA3|Mas-related gene A3', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=47588950, end_position_on_the_genomic_accession=47601372, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  233230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233230, CurrentID=0, Status='live', Symbol='Mrgprb4', Aliases='MrgB4', description='MAS-related GPR, member B4', other_designations='MrgB4 G protein-coupled receptor', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48198213, end_position_on_the_genomic_accession=48199178, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100503311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503311, CurrentID=0, Status='live', Symbol='Pifo', Aliases='1700027A23Rik', description='primary cilia formation', other_designations='pitchfork', map_location='3 F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105996957, end_position_on_the_genomic_accession=106014646, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  102162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102162, CurrentID=0, Status='live', Symbol='Taf5l', Aliases='1110005N04Rik, AI849020', description='TATA-box binding protein associated factor 5 like', other_designations='PAF65-beta|PCAF-associated factor 65 beta|TAF5-like RNA polymerase II, p300/CBP-associated factor (PCAF)-associated factor', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123996310, end_position_on_the_genomic_accession=124021309, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  233246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233246, CurrentID=0, Status='live', Symbol='Ano5', Aliases='9330162L24, Gdd1, Tmem16e', description='anoctamin 5', other_designations='gnathodiaphyseal dysplasia 1 protein homolog|integral membrane protein GDD1|transmembrane protein 16E', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=51510875, end_position_on_the_genomic_accession=51600693, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  102182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102182, CurrentID=0, Status='live', Symbol='Prmt9', Aliases='AI931714, Prmt10', description='protein arginine methyltransferase 9', other_designations='protein arginine methyltransferase 10 (putative)|putative protein arginine N-methyltransferase 10', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=77549387, end_position_on_the_genomic_accession=77581338, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  102193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102193, CurrentID=0, Status='live', Symbol='Zdhhc7', Aliases='AL024087, Gramp2', description='zinc finger, DHHC domain containing 7', other_designations='DHHC-7|GABA-A receptor-associated membrane protein 2|zinc finger DHHC domain-containing protein 7', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120081091, end_position_on_the_genomic_accession=120101472, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  233271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233271, CurrentID=0, Status='live', Symbol='Luzp2', Aliases='9330154K17Rik, D130073E01', description='leucine zipper protein 2', other_designations='leucine zipper motif 2', map_location='7 33.23 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=54835200, end_position_on_the_genomic_accession=55268888, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  233274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233274, CurrentID=0, Status='live', Symbol='Siglech', Aliases='6430529G09Rik, Siglec-H', description='sialic acid binding Ig-like lectin H', other_designations='SIGLEC-like protein', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=55768138, end_position_on_the_genomic_accession=55778833, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  233276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233276, CurrentID=0, Status='live', Symbol='Tubgcp5', Aliases='B130010C12Rik, GCP5, mKIAA1899', description='tubulin, gamma complex associated protein 5', other_designations='GCP-5', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=55794125, end_position_on_the_genomic_accession=55831459, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  233280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233280, CurrentID=0, Status='live', Symbol='Nipa1', Aliases='1110027G09Rik, A830014A18Rik, FSP3, Spg6', description='non imprinted in Prader-Willi/Angelman syndrome 1 homolog (human)', other_designations='non-imprinted in Prader-Willi/Angelman syndrome 1|non-imprinted in Prader-Willi/Angelman syndrome region protein 1 homolog|spastic paraplegia 6 homolog', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=55978484, end_position_on_the_genomic_accession=56019573, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  100503361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503361, CurrentID=0, Status='live', Symbol='Tmem95', Aliases='432576', description='transmembrane protein 95', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69876054, end_position_on_the_genomic_accession=69878173, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100503386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503386, CurrentID=0, Status='live', Symbol='Tpbgl', Aliases='EG245190, Gm4980', description='trophoblast glycoprotein-like', other_designations='', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99624067, end_position_on_the_genomic_accession=99627103, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  235493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235493, CurrentID=0, Status='live', Symbol='Fam214a', Aliases='6330415I01Rik, AI553555, C130047D21Rik, Kiaa1370, mKIAA1370', description='family with sequence similarity 214, member A', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=74952919, end_position_on_the_genomic_accession=75032468, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  233315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233315, CurrentID=0, Status='live', Symbol='Mtmr10', Aliases='BB128963, C79734', description='myotubularin related protein 10', other_designations='', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64287670, end_position_on_the_genomic_accession=64340806, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  102247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102247, CurrentID=0, Status='live', Symbol='Gpat4', Aliases='AU041707, AW545732, Agpat6, Tsarg7', description='glycerol-3-phosphate acyltransferase 4', other_designations='1-AGP acyltransferase 6|1-AGPAT 6|1-acylglycerol-3-phosphate O-acyltransferase 6 (lysophosphatidic acid acyltransferase, zeta)|LPAAT-zeta|acyl-CoA:glycerol-3-phosphate acyltransferase 4|lysophosphatidic acid acyltransferase zeta|putative lysophosphatidic acid acyltransferase|testis spermatogenesis apoptosis-related protein 7', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=23171265, end_position_on_the_genomic_accession=23209171, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  233328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233328, CurrentID=0, Status='live', Symbol='Lrrk1', Aliases='AW319595, C230002E15Rik, D130026O16Rik, mKIAA1790', description='leucine-rich repeat kinase 1', other_designations='lrrk1', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66258745, end_position_on_the_genomic_accession=66388341, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  233332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233332, CurrentID=0, Status='live', Symbol='Adamts17', Aliases='AU023434', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 17', other_designations='ADAM metallopeptidase with thrombospondin type 1 motif, 17|disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 17', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66839723, end_position_on_the_genomic_accession=67153171, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  233335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233335, CurrentID=0, Status='live', Symbol='Synm', Aliases='4930412K21Rik, AI852401, Dmn, E130104F11, Syn, Synemin', description='synemin, intermediate filament protein', other_designations='synm', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=67730161, end_position_on_the_genomic_accession=67759742, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102278, CurrentID=0, Status='live', Symbol='Cpne7', Aliases='AW047065', description='copine VII', other_designations='copine 7 protein|cpne7', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123117315, end_position_on_the_genomic_accession=123135185, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  626578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626578, CurrentID=0, Status='live', Symbol='Gbp10', Aliases='', description='guanylate-binding protein 10', other_designations='', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105215699, end_position_on_the_genomic_accession=105239533, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  102294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102294, CurrentID=0, Status='live', Symbol='Cyp4v3', Aliases='AU043077, AW111961, Cyp4v2', description='cytochrome P450, family 4, subfamily v, polypeptide 3', other_designations='cytochrome P450 4V3|cytochrome P450, 4v3|docosahexaenoic acid omega-hydroxylase CYP4V2', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45305802, end_position_on_the_genomic_accession=45333196, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  626583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626583, CurrentID=0, Status='live', Symbol='Igkv3-2', Aliases='IgK, IgK21-2, Igk-V21-2', description='immunoglobulin kappa variable 3-2', other_designations='V(kappa)21A|immunoglobulin kappa chain variable 21 (V21)-2|noglobulin kappa chain variable 3-2', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70698468, end_position_on_the_genomic_accession=70699067, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  626596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626596, CurrentID=0, Status='live', Symbol='Rgs22', Aliases='', description='regulator of G-protein signalling 22', other_designations='', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=35960511, end_position_on_the_genomic_accession=36140630, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  102323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102323, CurrentID=0, Status='live', Symbol='Dcun1d2', Aliases='AW495713, mFLJ10704', description='DCN1, defective in cullin neddylation 1, domain containing 2 (S. cerevisiae)', other_designations='DCUN1 domain-containing protein 2|defective in cullin neddylation protein 1-like protein 2', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13255963, end_position_on_the_genomic_accession=13288126, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  233405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233405, CurrentID=0, Status='live', Symbol='Vps33b', Aliases='', description='vacuolar protein sorting 33B (yeast)', other_designations='', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80269635, end_position_on_the_genomic_accession=80291579, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  233406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233406, CurrentID=0, Status='live', Symbol='Prc1', Aliases='D7Ertd348e', description='protein regulator of cytokinesis 1', other_designations='prc1', map_location='7 45.62 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80293786, end_position_on_the_genomic_accession=80316259, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  233410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233410, CurrentID=0, Status='live', Symbol='Zfp592', Aliases='8430405N24, A730014M16Rik, Znf592, mKIAA0211', description='zinc finger protein 592', other_designations='zfp-592', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=80993684, end_position_on_the_genomic_accession=81045162, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  102339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102339, CurrentID=0, Status='live', Symbol='Cog4', Aliases='AW554810, D8Ertd515e', description='component of oligomeric golgi complex 4', other_designations='COG complex subunit 4', map_location='8 57.76 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110846675, end_position_on_the_genomic_accession=110882234, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  233424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233424, CurrentID=0, Status='live', Symbol='Tmc3', Aliases='9030203A06', description='transmembrane channel-like gene family 3', other_designations='tmc3', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=83567837, end_position_on_the_genomic_accession=83626852, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  233437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233437, CurrentID=0, Status='live', Symbol='Vmn2r66', Aliases='F830104D24Rik', description='vomeronasal 2, receptor 66', other_designations='vomeronasal receptor Vmn2r66', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=84994645, end_position_on_the_genomic_accession=85012020, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  102371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102371, CurrentID=0, Status='live', Symbol='Myzap', Aliases='AA407270, AV006038, Gcom1, Gm640, Grinl1a7, Myozap', description='myocardial zonula adherens protein', other_designations='GRINL1A complex locus protein 1', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=71504347, end_position_on_the_genomic_accession=71592360, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100503545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503545, CurrentID=0, Status='live', Symbol='Nuggc', Aliases='Gm600, SLIP-GC', description='nuclear GTPase, germinal center associated', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65593499, end_position_on_the_genomic_accession=65648444, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  102414 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102414, CurrentID=0, Status='live', Symbol='Clk3', Aliases='AI256811', description='CDC-like kinase 3', other_designations='cdc2/CDC28-like protein kinase 3', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57750621, end_position_on_the_genomic_accession=57765860, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  233489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233489, CurrentID=0, Status='live', Symbol='Picalm', Aliases='CALM, CLTH, PAP180, fit-1, fit1, mKIAA4114', description='phosphatidylinositol binding clathrin assembly protein', other_designations='clathrin assembly lymphoid myeloid leukemia', map_location='7 50.47 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90130220, end_position_on_the_genomic_accession=90209447, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  233490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233490, CurrentID=0, Status='live', Symbol='Crebzf', Aliases='1110034C16Rik, 6330417B10Rik, LAZip, LAZipII, SMILE, Zf', description='CREB/ATF bZIP transcription factor', other_designations='HCF-binding transcription factor Zhangfei|host cell factor-binding transcription factor Zhangfei|tyrosine kinase-associated leucine zipper protein LAZip|tyrosine kinase-associated leucine zipper protein LAZipII', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=90442735, end_position_on_the_genomic_accession=90448043, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100503572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503572, CurrentID=0, Status='live', Symbol='Bbip1', Aliases='2310002J21Rik, 2610002B11Rik, 4930560M04Rik, Ncrna00081', description='BBSome interacting protein 1', other_designations='BBSome-interacting protein of 10 kDa', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53929656, end_position_on_the_genomic_accession=53944643, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102423, CurrentID=0, Status='live', Symbol='Hinfp', Aliases='AA589481, HiNF-P, Mizf', description='histone H4 transcription factor', other_designations='MBD2 (methyl-CpG-binding protein)-interacting zinc finger protein|MBD2-interacting zinc finger protein|histone nuclear factor P|methyl-CpG-binding protein 2-interacting zinc finger protein', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44292438, end_position_on_the_genomic_accession=44305705, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  100503583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503583, CurrentID=0, Status='live', Symbol='Fsbp', Aliases='', description='fibrinogen silencer binding protein', other_designations='', map_location='4 5.11 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11579662, end_position_on_the_genomic_accession=11587802, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  100503584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503584, CurrentID=0, Status='live', Symbol='Zfp534', Aliases='Gm13159, OTTMUSG00000010681', description='zinc finger protein 534', other_designations='novel KRAB box domain containing protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147674225, end_position_on_the_genomic_accession=147702631, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  102436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102436, CurrentID=0, Status='live', Symbol='Lars2', Aliases='AI035546, Kiaa0028, LEURS', description='leucyl-tRNA synthetase, mitochondrial', other_designations='leucine--tRNA ligase|probable leucyl-tRNA synthetase, mitochondrial', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123366940, end_position_on_the_genomic_accession=123462666, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  102442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102442, CurrentID=0, Status='live', Symbol='Dennd4a', Aliases='AI115600, AI851943, F730015K02Rik', description='DENN/MADD domain containing 4A', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64806123, end_position_on_the_genomic_accession=64919667, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  102448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102448, CurrentID=0, Status='live', Symbol='Xylb', Aliases='AI132223, AI327421', description='xylulokinase homolog (H. influenzae)', other_designations='', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119357359, end_position_on_the_genomic_accession=119393797, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  626740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626740, CurrentID=0, Status='live', Symbol='Scgb2b30', Aliases='Abpbg30, EG626740, Gm6703', description='secretoglobin, family 2B, member 30', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=32146240, end_position_on_the_genomic_accession=32148032, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100503605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503605, CurrentID=0, Status='live', Symbol='Hbb-bs', Aliases='Beta-s, Hbbt1, Hbbt2', description='hemoglobin, beta adult s chain', other_designations='beta s|beta-globin|beta-single|hemoglobin subunit beta-1-like', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103826523, end_position_on_the_genomic_accession=103827928, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  233529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233529, CurrentID=0, Status='live', Symbol='Kctd14', Aliases='AI449310, D7Ertd760e', description='potassium channel tetramerisation domain containing 14', other_designations='potassium channel tetramerization domain containing 14', map_location='7 53.51 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97436380, end_position_on_the_genomic_accession=97459557, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  233532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233532, CurrentID=0, Status='live', Symbol='Rsf1', Aliases='4832420A03Rik, C030033M12Rik, Gm164, Hbxap, XAP8, p325', description='remodeling and spacing factor 1', other_designations='hepatitis B virus x associated protein', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97579596, end_position_on_the_genomic_accession=97692782, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  102462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102462, CurrentID=0, Status='live', Symbol='Imp3', Aliases='1190002L16Rik, AI256594', description='IMP3, U3 small nucleolar ribonucleoprotein', other_designations='IMP3, U3 small nucleolar ribonucleoprotein, homolog|U3 snoRNP protein 3 homolog|U3 snoRNP protein IMP3', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56937500, end_position_on_the_genomic_accession=56938398, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  233537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233537, CurrentID=0, Status='live', Symbol='Gdpd4', Aliases='4921513O04', description='glycerophosphodiester phosphodiesterase domain containing 4', other_designations='glycerophosphodiester phosphodiesterase 6', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97919944, end_position_on_the_genomic_accession=98049663, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  233545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233545, CurrentID=0, Status='live', Symbol='Emsy', Aliases='2210018M11Rik, 2310016L09Rik, AU019521, AU019557', description='EMSY, BRCA2-interacting transcriptional repressor', other_designations='protein EMSY', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98587137, end_position_on_the_genomic_accession=98658289, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  233549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233549, CurrentID=0, Status='live', Symbol='Mogat2', Aliases='DGAT2L5, MGAT2, Mgat1l', description='monoacylglycerol O-acyltransferase 2', other_designations='acyl CoA:monoacylglycerol acyltransferase 2|acyl-CoA:monoacylglycerol acyltransferase 2|diacylglycerol acyltransferase 2-like protein 5', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99219084, end_position_on_the_genomic_accession=99238611, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  233552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233552, CurrentID=0, Status='live', Symbol='Gdpd5', Aliases='BC024955, Gde2', description='glycerophosphodiester phosphodiesterase domain containing 5', other_designations='glycerophosphodiester phosphodiesterase 2', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99381415, end_position_on_the_genomic_accession=99460984, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  233571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233571, CurrentID=0, Status='live', Symbol='P2ry6', Aliases='2010204J23Rik, P2Y6', description='pyrimidinergic receptor P2Y, G-protein coupled, 6', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=100937630, end_position_on_the_genomic_accession=100975492, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102502, CurrentID=0, Status='live', Symbol='Pls1', Aliases='AI427122', description='plastin 1 (I-isoform)', other_designations='I-fimbrin|plastin 1, I', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95752642, end_position_on_the_genomic_accession=95845306, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  233575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233575, CurrentID=0, Status='live', Symbol='Pgap2', Aliases='1810006G21Rik, C76649, Frag1', description='post-GPI attachment to proteins 2', other_designations='FGF receptor activating protein 1|FGF receptor-activating protein 1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102210242, end_position_on_the_genomic_accession=102238564, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100503659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503659, CurrentID=0, Status='live', Symbol='Cbarp', Aliases='9630008F14, Dos, R29144/1', description='calcium channel, voltage-dependent, beta subunit associated regulatory protein', other_designations='downstream of Stk11 protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80130434, end_position_on_the_genomic_accession=80140136, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100503670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503670, CurrentID=0, Status='live', Symbol='Rpl5', Aliases='U21RNA', description='ribosomal protein L5', other_designations='', map_location='5 52.23 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107900528, end_position_on_the_genomic_accession=107909005, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  626832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626832, CurrentID=0, Status='live', Symbol='Gm6710', Aliases='1600017I02Rik, 2310011F05Rik, 2900089K14Rik, 3100002L24Rik, WF-4', description='predicted gene 6710', other_designations='KRAB box and zinc finger, C2H2 type domain containing', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=175869245, end_position_on_the_genomic_accession=175883182, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  102545 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102545, CurrentID=0, Status='live', Symbol='Cmtm7', Aliases='AI481279, Cklfsf7, LNV', description='CKLF-like MARVEL transmembrane domain containing 7', other_designations='chemokine-like factor super family 7|chemokine-like factor superfamily member 7', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114756836, end_position_on_the_genomic_accession=114781993, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  626834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626834, CurrentID=0, Status='live', Symbol='Klk13', Aliases='Egfbp-2, mGk-13', description='kallikrein related-peptidase 13', other_designations='kallikrein 13', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43712567, end_position_on_the_genomic_accession=43727864, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  626848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626848, CurrentID=0, Status='live', Symbol='Etohi1', Aliases='', description='ethanol induced 1', other_designations='KRAB box and zinc finger, C2H2 type domain containing protein', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=178023284, end_position_on_the_genomic_accession=178035859, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102566, CurrentID=0, Status='live', Symbol='Ano10', Aliases='AI604832, Tmem16k', description='anoctamin 10', other_designations='transmembrane protein 16K', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122175874, end_position_on_the_genomic_accession=122294443, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  102570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102570, CurrentID=0, Status='live', Symbol='Slc22a13', Aliases='AI648912, OCTL1, OCTL3, ORCTL3', description='solute carrier family 22 (organic cation transporter), member 13', other_designations='ORCTL-3|organic cation transporter-like 3', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119192968, end_position_on_the_genomic_accession=119209105, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  233649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233649, CurrentID=0, Status='live', Symbol='Cnga4', Aliases='', description='cyclic nucleotide gated channel alpha 4', other_designations='CNG channel alpha-4|CNG-4|CNG4|cyclic nucleotide gated cation channel|cyclic nucleotide-gated cation channel alpha 4|cyclic nucleotide-gated channel alpha-4|olfactory cyclic nucleotide-gated channel', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105400377, end_position_on_the_genomic_accession=105409313, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  233651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233651, CurrentID=0, Status='live', Symbol='Dchs1', Aliases='3110041P15Rik, C130033F22Rik, CDH25, FIB1, Gm165, PCDH16', description='dachsous 1 (Drosophila)', other_designations='protein Dchs1|protein dachsous homolog 1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105752989, end_position_on_the_genomic_accession=105788317, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  102580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102580, CurrentID=0, Status='live', Symbol='Alg9', Aliases='8230402H15Rik, AI747665, B430313H07Rik, Dibd1', description='asparagine-linked glycosylation 9 (alpha 1,2 mannosyltransferase)', other_designations='asparagine-linked glycosylation 9 homolog (yeast, alpha 1,2 mannosyltransferase)|asparagine-linked glycosylation 9 protein|asparagine-linked glycosylation protein 9 homolog|disrupted in bipolar disorder 1 homolog|disrupted in bipolar disorder protein 1 homolog|dol-P-Man:Man(6)GlcNAc(2)-PP-Dol alpha-1,2-mannosyltransferase|dol-P-Man:Man(8)GlcNAc(2)-PP-Dol alpha-1,2-mannosyltransferase', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50774881, end_position_on_the_genomic_accession=50843639, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  626870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626870, CurrentID=0, Status='live', Symbol='Gm11992', Aliases='OTTMUSG00000005148', description='predicted gene 11992', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=9048592, end_position_on_the_genomic_accession=9069354, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  102595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102595, CurrentID=0, Status='live', Symbol='Plekho2', Aliases='AI840980, Plekhq1', description='pleckstrin homology domain containing, family O member 2', other_designations='PH domain-containing family O member 2|PH domain-containing family Q member 1|PH domain-containing protein homolog|pleckstrin homology domain containing, family Q member 1', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=65552577, end_position_on_the_genomic_accession=65580087, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  233670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233670, CurrentID=0, Status='live', Symbol='Olfr6', Aliases='M50, MOR103-16', description='olfactory receptor 6', other_designations='GA_x6K02T2PBJ9-9337331-9336381|m50 olfactory receptor|odorant receptor M50|olfactory receptor 103-16', map_location='7 56.1 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106955984, end_position_on_the_genomic_accession=106956934, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  102607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102607, CurrentID=0, Status='live', Symbol='Snx19', Aliases='3526401K03Rik, AI195321, AI848208, mKIAA0254', description='sorting nexin 19', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=30427329, end_position_on_the_genomic_accession=30466726, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  102614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102614, CurrentID=0, Status='live', Symbol='Rpp25', Aliases='AI851155', description='ribonuclease P/MRP 25 subunit', other_designations='RNase P protein subunit p25|ribonuclease P 25 subunit|ribonuclease P 25kDa subunit', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57504102, end_position_on_the_genomic_accession=57505447, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  102626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102626, CurrentID=0, Status='live', Symbol='Mapkapk3', Aliases='3PK, AI874665, MAPKAP-K3, MAPKAP3, MK-3, MK3, MapkKapk3', description='mitogen-activated protein kinase-activated protein kinase 3', other_designations='MAPK-activated protein kinase 3|MAPKAP kinase 3|MAPKAPK-3|mitogen activated protein kinase-activated protein kinase 3', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107254926, end_position_on_the_genomic_accession=107289884, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  102632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102632, CurrentID=0, Status='live', Symbol='Acad11', Aliases='5730439E10Rik, AI987948', description='acyl-Coenzyme A dehydrogenase family, member 11', other_designations='ACAD-11', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=104063703, end_position_on_the_genomic_accession=104127646, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  102644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102644, CurrentID=0, Status='live', Symbol='Oaf', Aliases='AL024221, D130038B21Rik, D9Ucla1', description='OAF homolog (Drosophila)', other_designations='', map_location='9 24.24 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=43221278, end_position_on_the_genomic_accession=43239816, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  233724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233724, CurrentID=0, Status='live', Symbol='Tmem41b', Aliases='1500015G02Rik, 1500031M19Rik, AU018901, D7Ertd70e, D7Ertd743e', description='transmembrane protein 41B', other_designations='', map_location='7 57.69 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109972187, end_position_on_the_genomic_accession=109986230, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  233726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233726, CurrentID=0, Status='live', Symbol='Ipo7', Aliases='A330055O14Rik, C330016G14, Imp7, Ranbp7', description='importin 7', other_designations='RAN binding protein 7|ran-binding protein 7', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110018227, end_position_on_the_genomic_accession=110056610, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  102657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102657, CurrentID=0, Status='live', Symbol='Cd276', Aliases='6030411F23Rik, AU016588, B7RP-2, B7h3', description='CD276 antigen', other_designations='B7 homolog 3|B7-H3|costimulatory molecule', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58524300, end_position_on_the_genomic_accession=58555033, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  233733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233733, CurrentID=0, Status='live', Symbol='Galnt18', Aliases='2900011G21Rik, BC024988, Galntl4', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 18', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 4|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 4|galNAc-T-like protein 4|galNAc-T18|mpp-GalNAc-T18(O18)|polypeptide GalNAc transferase 18|polypeptide GalNAc transferase-like protein 4|pp-GaNTase-like protein 4|protein-UDP acetylgalactosaminyltransferase-like protein 4|putative polypeptide N-acetylgalactosaminyltransferase-like protein 4', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=111471656, end_position_on_the_genomic_accession=111780067, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  233744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233744, CurrentID=0, Status='live', Symbol='Spon1', Aliases='AI666765, AW455831, BC020531, D330035F22Rik, FSP', description='spondin 1, (f-spondin) extracellular matrix protein', other_designations='F-spondin', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=113765998, end_position_on_the_genomic_accession=114043375, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  102680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102680, CurrentID=0, Status='live', Symbol='Slc6a20a', Aliases='A730081N20Rik, AU022428, Xt3s1, Xtrp3s1', description='solute carrier family 6 (neurotransmitter transporter), member 20A', other_designations='IMINO-B|X transporter protein 3 similar 1|slc6a20a|solute carrier family 6 member 20A', map_location='9 74.26 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123636907, end_position_on_the_genomic_accession=123678832, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  626970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=626970, CurrentID=0, Status='live', Symbol='Scgb2b21', Aliases='Abpbg21, EG626970, Gm6725', description='secretoglobin, family 2B, member 21', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33518644, end_position_on_the_genomic_accession=33520440, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  102693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102693, CurrentID=0, Status='live', Symbol='Phldb1', Aliases='AU041016, D330037A14Rik, Ll5a', description='pleckstrin homology-like domain, family B, member 1', other_designations='Ll5alpha|protein LL5-alpha', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44686308, end_position_on_the_genomic_accession=44736151, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  100503849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503849, CurrentID=0, Status='live', Symbol='Sp3os', Aliases='1700011J10Rik', description='trans-acting transcription factor 3, opposite strand', other_designations='', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=72979432, end_position_on_the_genomic_accession=72989249, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  233781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233781, CurrentID=0, Status='live', Symbol='Xylt1', Aliases='8030490L12', description='xylosyltransferase 1', other_designations='peptide O-xylosyltransferase 1|xylosyltransferase I', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=117380979, end_position_on_the_genomic_accession=117667630, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  233789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233789, CurrentID=0, Status='live', Symbol='Smg1', Aliases='2610207I05Rik, 5430435M13Rik, C130002K18Rik, mKIAA0421', description='SMG1 homolog, phosphatidylinositol 3-kinase-related kinase (C. elegans)', other_designations='PI-3-kinase-related kinase SMG-1|SMG-1|phosphatidylinositol 3-kinase-related protein kinase', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118131308, end_position_on_the_genomic_accession=118243658, orientation='minus', exon_count=64, OMIM='', no_hdr0=''),
  100503872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503872, CurrentID=0, Status='live', Symbol='Hotair', Aliases='Gm16258, OTTMUSG00000032781', description='HOX transcript antisense RNA (non-protein coding)', other_designations='hox transcript antisense intergenic RNA', map_location='15 58.01 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102944062, end_position_on_the_genomic_accession=102947946, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  233799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233799, CurrentID=0, Status='live', Symbol='Acsm2', Aliases='AI315615, BC031140', description='acyl-CoA synthetase medium-chain family member 2', other_designations='butyrate--CoA ligase 2|butyryl coenzyme A synthetase 2|butyryl-coenzyme A synthetase 2|middle-chain acyl-CoA synthetase 2', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119554340, end_position_on_the_genomic_accession=119600694, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  233801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233801, CurrentID=0, Status='live', Symbol='Acsm4', Aliases='O-MACS, OMACS', description='acyl-CoA synthetase medium-chain family member 4', other_designations='olfactory specific medium-chain acyl CoA synthetase', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119690026, end_position_on_the_genomic_accession=119714566, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  233802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233802, CurrentID=0, Status='live', Symbol='Thumpd1', Aliases='6330575P11Rik', description='THUMP domain containing 1', other_designations='', map_location='7 64.13 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119715093, end_position_on_the_genomic_accession=119720823, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  100503884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503884, CurrentID=0, Status='live', Symbol='Ccdc149', Aliases='Gm447', description='coiled-coil domain containing 149', other_designations='', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52374651, end_position_on_the_genomic_accession=52472025, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  233805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233805, CurrentID=0, Status='live', Symbol='Dcun1d3', Aliases='1700020A13Rik, BC030335', description='DCN1, defective in cullin neddylation 1, domain containing 3 (S. cerevisiae)', other_designations='DCUN1 domain-containing protein 3|defective in cullin neddylation protein 1-like protein 3', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119853158, end_position_on_the_genomic_accession=119896169, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  233806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233806, CurrentID=0, Status='live', Symbol='Tmem159', Aliases='8430420C20Rik', description='transmembrane protein 159', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120102426, end_position_on_the_genomic_accession=120120987, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  233810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233810, CurrentID=0, Status='live', Symbol='Abca16', Aliases='AY243472', description='ATP-binding cassette, sub-family A (ABC1), member 16', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120409647, end_position_on_the_genomic_accession=120544892, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  233812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233812, CurrentID=0, Status='live', Symbol='BC030336', Aliases='', description='cDNA sequence BC030336', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120677620, end_position_on_the_genomic_accession=120734854, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  233813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233813, CurrentID=0, Status='live', Symbol='Vwa3a', Aliases='E030013G06Rik', description='von Willebrand factor A domain containing 3A', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=120735693, end_position_on_the_genomic_accession=120805546, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  100503895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503895, CurrentID=0, Status='live', Symbol='Tpsab1', Aliases='MMCP-7, Mcp-7, Mcp7, Mcpt7', description='tryptase alpha/beta 1', other_designations='mast cell protease 7|mast cell-restricted serine protease 7|tryptase alpha/beta-1', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25343245, end_position_on_the_genomic_accession=25345562, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102747, CurrentID=0, Status='live', Symbol='Lrrc49', Aliases='AI430817, AW050057, D430025H09Rik, PGs4, p79', description='leucine rich repeat containing 49', other_designations='tubulin polyglutamylase complex subunit 4', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=60587229, end_position_on_the_genomic_accession=60688134, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  233824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233824, CurrentID=0, Status='live', Symbol='Cog7', Aliases='5630400E24Rik, Gm167', description='component of oligomeric golgi complex 7', other_designations='COG complex subunit 7', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121922836, end_position_on_the_genomic_accession=121981706, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  233826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233826, CurrentID=0, Status='live', Symbol='Palb2', Aliases='BC066140', description='partner and localizer of BRCA2', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=122107262, end_position_on_the_genomic_accession=122132980, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  233833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233833, CurrentID=0, Status='live', Symbol='Tnrc6a', Aliases='2010321I05Rik, 3110054G10Rik, AW557223, CAGH26, D130023A07Rik, GW182, Tnrc6', description='trinucleotide repeat containing 6a', other_designations='GW182 autoantigen', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=123031101, end_position_on_the_genomic_accession=123195296, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  100503915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503915, CurrentID=0, Status='live', Symbol='Smpd5', Aliases='ENSMUSG00000071724, Gm10345, MA-nSMase', description='sphingomyelin phosphodiesterase 5', other_designations='mitochondria-associated nSMase', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76293869, end_position_on_the_genomic_accession=76296896, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  233836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233836, CurrentID=0, Status='live', Symbol='Slc5a11', Aliases='2010013B02Rik, Kst1, RKST2, SMIT2, Slc5a10', description='solute carrier family 5 (sodium/glucose cotransporter), member 11', other_designations='Na(+)/myo-inositol cotransporter 2|na+/myo-inositol cotransporter 2|slc5a11|sodium-coupled cotransporter KST1|sodium-dependent glucose cotransporter|sodium/glucose cotransporter KST1|sodium/myo-inositol transporter 2|solute carrier family 5 (sodium/glucose cotransporter), member 10|solute carrier family 5 member 11', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=123214758, end_position_on_the_genomic_accession=123274390, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  100503924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503924, CurrentID=0, Status='live', Symbol='Fcor', Aliases='2400009B08Rik', description='Foxo1 corepressor', other_designations='Foxo1-corepressor', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3671185, end_position_on_the_genomic_accession=3672358, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  102774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102774, CurrentID=0, Status='live', Symbol='Bbs4', Aliases='AW537059, AW742241, D9Ertd464e', description='Bardet-Biedl syndrome 4 (human)', other_designations='Bardet-Biedl syndrome 4 homolog', map_location='9 32.01 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=59321966, end_position_on_the_genomic_accession=59353508, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  233863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233863, CurrentID=0, Status='live', Symbol='Gtf3c1', Aliases='', description='general transcription factor III C 1', other_designations='TF3C-alpha|TFIIIC 220 kDa subunit|TFIIIC box B-binding subunit|TFIIIC220|transcription factor IIIC 220 kDa subunit|transcription factor IIIC subunit alpha', map_location='7 69.01 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125640954, end_position_on_the_genomic_accession=125707688, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  233865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233865, CurrentID=0, Status='live', Symbol='D430042O09Rik', Aliases='Kiaa0556', description='RIKEN cDNA D430042O09 gene', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125707857, end_position_on_the_genomic_accession=125874797, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  233870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233870, CurrentID=0, Status='live', Symbol='Tufm', Aliases='2300002G02Rik, C76308, C76389, EF-TuMT, EFTU', description='Tu translation elongation factor, mitochondrial', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126487265, end_position_on_the_genomic_accession=126490731, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  233871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233871, CurrentID=0, Status='live', Symbol='Atxn2l', Aliases='A2D, A2LG, A2RP, A2lp, Atxnl2', description='ataxin 2-like', other_designations='ataxin 2 related protein', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126491708, end_position_on_the_genomic_accession=126503644, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  233875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233875, CurrentID=0, Status='live', Symbol='Ino80e', Aliases='AI225782, AI854876, Ccdc85, Ccdc95', description='INO80 complex subunit E', other_designations='NO80 complex subunit E|coiled-coil domain containing 95', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126851584, end_position_on_the_genomic_accession=126862371, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  233876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233876, CurrentID=0, Status='live', Symbol='Hirip3', Aliases='B130036O03, C86302', description='HIRA interacting protein 3', other_designations='hirip3', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126861972, end_position_on_the_genomic_accession=126865377, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  233877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233877, CurrentID=0, Status='live', Symbol='Kctd13', Aliases='1500003N18Rik, AV259508, PDIP1alpha, Pdip1, Poldip1', description='potassium channel tetramerisation domain containing 13', other_designations='BTB/POZ domain-containing protein KCTD13|polymerase delta-interacting protein 1|tumor necrosis factor alpha-inducible protein PDIP1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126928842, end_position_on_the_genomic_accession=126947897, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  233878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233878, CurrentID=0, Status='live', Symbol='Sez6l2', Aliases='AI835913, AW121566, BSRP-A, Psk1', description='seizure related 6 homolog like 2', other_designations='brain-specific receptor-like protein A|seizure related 6 homolog-like 2|type I transmembrane receptor (seizure-related protein)', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126950043, end_position_on_the_genomic_accession=126970610, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  233887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233887, CurrentID=0, Status='live', Symbol='Zfp553', Aliases='2600009K23Rik, AI431057, BC026401, C330013F15Rik, ENSMUSG00000054461, Znf48', description='zinc finger protein 553', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127233255, end_position_on_the_genomic_accession=127238042, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  233890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233890, CurrentID=0, Status='live', Symbol='Zfp768', Aliases='BC026432, Znf768', description='zinc finger protein 768', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127342795, end_position_on_the_genomic_accession=127345314, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  233893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233893, CurrentID=0, Status='live', Symbol='Zfp764', Aliases='8030466O12Rik', description='zinc finger protein 764', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127382263, end_position_on_the_genomic_accession=127406835, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  233895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233895, CurrentID=0, Status='live', Symbol='Prr14', Aliases='BC006909', description='proline rich 14', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127470096, end_position_on_the_genomic_accession=127476767, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  233899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233899, CurrentID=0, Status='live', Symbol='Ccdc189', Aliases='Gm166', description='coiled-coil domain containing 189', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127582383, end_position_on_the_genomic_accession=127588609, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  233900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233900, CurrentID=0, Status='live', Symbol='Rnf40', Aliases='AI848422, mKIAA0661', description='ring finger protein 40', other_designations='BRE1-B', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127588698, end_position_on_the_genomic_accession=127603605, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  233902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233902, CurrentID=0, Status='live', Symbol='Fbxl19', Aliases='6030430O11, BC059812, Fbl19', description='F-box and leucine-rich repeat protein 19', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127744027, end_position_on_the_genomic_accession=127769483, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  233904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233904, CurrentID=0, Status='live', Symbol='Setd1a', Aliases='BC010250, KMT2F, Nsccn1, mKIAA0339, mNSC1', description='SET domain containing 1A', other_designations='non-selective cation channel 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127777389, end_position_on_the_genomic_accession=127800122, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  233905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233905, CurrentID=0, Status='live', Symbol='Zfp646', Aliases='6820429M01', description='zinc finger protein 646', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127877036, end_position_on_the_genomic_accession=127885996, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  233908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233908, CurrentID=0, Status='live', Symbol='Fus', Aliases='D430004D17Rik, D930039C12Rik1, Tls, Fus', description='fused in sarcoma', other_designations='fusion, derived from t(12;16) malignant liposarcoma|hnRNP P2|pigpen protein|protein pigpen|translocated in liposarcoma', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127966788, end_position_on_the_genomic_accession=127982032, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100503992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100503992, CurrentID=0, Status='live', Symbol='Gm6040', Aliases='Defb51, EG574083', description='predicted gene 6040', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=20916490, end_position_on_the_genomic_accession=20922071, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  233913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233913, CurrentID=0, Status='live', Symbol='BC017158', Aliases='', description='cDNA sequence BC017158', other_designations='UPF0420 protein C16orf58 homolog', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128271379, end_position_on_the_genomic_accession=128298165, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  233918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233918, CurrentID=0, Status='live', Symbol='4933402N03Rik', Aliases='4930412O05', description='RIKEN cDNA 4933402N03 gene', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131138347, end_position_on_the_genomic_accession=131146283, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  233919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233919, CurrentID=0, Status='live', Symbol='Gpr26', Aliases='9630036A11Rik, A730021D22', description='G protein-coupled receptor 26', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=131966135, end_position_on_the_genomic_accession=131985633, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102857, CurrentID=0, Status='live', Symbol='Slc6a8', Aliases='AA589632, CRT, CRTR, CT1, CTR5, Creat', description='solute carrier family 6 (neurotransmitter transporter, creatine), member 8', other_designations='creatine transporter 1|creatine transporter SLC6A8 variant D|creatine transporter, solute carrier family 6, member 8|solute carrier family 6 member 8|solute carrier family 6, member 8', map_location='X 37.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73673133, end_position_on_the_genomic_accession=73682502, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  102866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102866, CurrentID=0, Status='live', Symbol='Pls3', Aliases='AI115446, AL024105', description='plastin 3 (T-isoform)', other_designations='T-fimbrin|T-plastin|plastin 3, T', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75785654, end_position_on_the_genomic_accession=75875170, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  102871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102871, CurrentID=0, Status='live', Symbol='D330045A20Rik', Aliases='AI132325', description='RIKEN cDNA D330045A20 gene', other_designations='likely orthologue of H. sapiens Chromosome X open reading frame 57 (CXorf57)', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139480226, end_position_on_the_genomic_accession=139554589, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  627191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=627191, CurrentID=0, Status='live', Symbol='Syndig1l', Aliases='DSPC1, Gm261, Tmem90a', description='synapse differentiation inducing 1 like', other_designations='capucin|caudate and putamen-enriched protein|dispanin subfamily C member 1|synapse differentiation-induced protein 2|transmembrane protein 90a', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84677277, end_position_on_the_genomic_accession=84698907, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  233977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233977, CurrentID=0, Status='live', Symbol='Ppfia1', Aliases='C030014K08Rik, C87158, LIP.1, LIP1', description='protein tyrosine phosphatase, receptor type, f polypeptide (PTPRF), interacting protein (liprin), alpha 1', other_designations='Liprin-alpha1|liprin', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=144476755, end_position_on_the_genomic_accession=144553747, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  233979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233979, CurrentID=0, Status='live', Symbol='Tpcn2', Aliases='D830047E22Rik, Gm35086', description='two pore segment channel 2', other_designations='voltage-dependent calcium channel protein TPC2', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=145181633, end_position_on_the_genomic_accession=145283987, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  233987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=233987, CurrentID=0, Status='live', Symbol='Zfp958', Aliases='', description='zinc finger protein 958', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=4613170, end_position_on_the_genomic_accession=4630231, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  102920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102920, CurrentID=0, Status='live', Symbol='Cenpi', Aliases='CENP-I, Fshprh1', description='centromere protein I', other_designations='FSH primary response protein 1|follicle-stimulating hormone primary response protein', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134308089, end_position_on_the_genomic_accession=134362643, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  627214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=627214, CurrentID=0, Status='live', Symbol='Fam196a', Aliases='B830028B13Rik', description='family with sequence similarity 196, member A', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=134881908, end_position_on_the_genomic_accession=134938430, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  234023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234023, CurrentID=0, Status='live', Symbol='Arglu1', Aliases='9430010O03Rik, C130008N12', description='arginine and glutamate rich 1', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=8666567, end_position_on_the_genomic_accession=8690553, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  102954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102954, CurrentID=0, Status='live', Symbol='Nudt10', Aliases='AI853080, Dipp3a', description='nudix (nucleoside diphosphate linked moiety X)-type motif 10', other_designations="DIPP-3-alpha|DIPP3 alpha|DIPP3-alpha|diadenosine 5',5'''-P1,P6-hexaphosphate hydrolase 3-alpha|diadenosine hexaphosphate hydrolase (AMP-forming)|nucleoside diphosphate-linked moiety X motif 10|nudix motif 10", map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=6168696, end_position_on_the_genomic_accession=6173044, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  627280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=627280, CurrentID=0, Status='live', Symbol='Vmn1r90', Aliases='B430211C08Rik', description='vomeronasal 1 receptor 90', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=14561242, end_position_on_the_genomic_accession=14562171, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  234069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234069, CurrentID=0, Status='live', Symbol='Pcid2', Aliases='A730042J05Rik', description='PCI domain containing 2', other_designations='CSN12-like protein', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13077183, end_position_on_the_genomic_accession=13105343, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  234072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234072, CurrentID=0, Status='live', Symbol='Adprhl1', Aliases='Arh2, D330008N11Rik', description='ADP-ribosylhydrolase like 1', other_designations='ADP-ribosylhydrolase 2', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13235632, end_position_on_the_genomic_accession=13254263, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  234076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234076, CurrentID=0, Status='live', Symbol='Tmco3', Aliases='B230339H12Rik, C87304', description='transmembrane and coiled-coil domains 3', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13288013, end_position_on_the_genomic_accession=13322924, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  100504162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504162, CurrentID=0, Status='live', Symbol='Prss51', Aliases='1700007N14Rik', description='protease, serine 51', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64085525, end_position_on_the_genomic_accession=64097674, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  103012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103012, CurrentID=0, Status='live', Symbol='Firre', Aliases='5830467J12Rik, 6720401G13Rik, AW048145', description='functional intergenic repeating RNA element', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=50555744, end_position_on_the_genomic_accession=50635258, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  234086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234086, CurrentID=0, Status='live', Symbol='Erich1', Aliases='', description='glutamate rich 1', other_designations='glutamate-rich 1', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=14025882, end_position_on_the_genomic_accession=14090338, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  234094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234094, CurrentID=0, Status='live', Symbol='Arhgef10', Aliases='6430549H08Rik, mKIAA0294', description='Rho guanine nucleotide exchange factor (GEF) 10', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=14910944, end_position_on_the_genomic_accession=15001085, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  100504195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504195, CurrentID=0, Status='live', Symbol='Micalcl', Aliases='4921517J23Rik', description='MICAL C-terminal like', other_designations='ERK2-binding testicular protein 1|MICAL C-terminal-like protein-like|ebitein-1|ebitein1', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=112368308, end_position_on_the_genomic_accession=112413104, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  17175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17175, CurrentID=0, Status='live', Symbol='Masp2', Aliases='MASP-2, MAp19', description='mannan-binding lectin serine peptidase 2', other_designations='MBL-associated serine protease 2|mannose-binding protein-associated serine protease 2', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148594877, end_position_on_the_genomic_accession=148615482, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  234129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234129, CurrentID=0, Status='live', Symbol='Tpte', Aliases='Pten2', description='transmembrane phosphatase with tensin homology', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22283441, end_position_on_the_genomic_accession=22371418, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  234130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234130, CurrentID=0, Status='live', Symbol='Dkk4', Aliases='Dkk-4', description='dickkopf WNT signaling pathway inhibitor 4', other_designations='dickkopf homolog 4|dickkopf-4', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22623685, end_position_on_the_genomic_accession=22627547, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  234135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234135, CurrentID=0, Status='live', Symbol='Whsc1l1', Aliases='6720429E03, A530023P05, AI528490, NSD3, WHISTLE', description='Wolf-Hirschhorn syndrome candidate 1-like 1 (human)', other_designations='WHSC1-like protein 1|nuclear SET domain-containing protein 3|wolf-Hirschhorn syndrome candidate 1-like protein 1 homolog', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=25601175, end_position_on_the_genomic_accession=25719667, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  234138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234138, CurrentID=0, Status='live', Symbol='Tti2', Aliases='', description='TELO2 interacting protein 2', other_designations='', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=31150316, end_position_on_the_genomic_accession=31164703, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100504230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504230, CurrentID=0, Status='live', Symbol='AU020206', Aliases='', description='expressed sequence AU020206', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=75769039, end_position_on_the_genomic_accession=75782099, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  103080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103080, CurrentID=0, Status='live', Symbol='Sept10', Aliases='4921515A04Rik, 9430099J10Rik, AA408298, AI874685', description='septin 10', other_designations='', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=59141627, end_position_on_the_genomic_accession=59221847, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  234155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234155, CurrentID=0, Status='live', Symbol='Mboat4', Aliases='GOAT, Gm171', description='membrane bound O-acyltransferase domain containing 4', other_designations='O-acyltransferase domain-containing protein 4|membrane-bound O-acyltransferase domain-containing protein 4', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=34108219, end_position_on_the_genomic_accession=34125185, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  103098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103098, CurrentID=0, Status='live', Symbol='Slc6a15', Aliases='AA536730, AI326450, AI326451, v7-3', description='solute carrier family 6 (neurotransmitter transporter), member 15', other_designations='B0AT2 protein|orphan sodium- and chloride-dependent neurotransmitter transporter NTT73|orphan transporter v7-3|sodium- and chloride-dependent neurotransmitter transporter NTT73|solute carrier family 6 member 15|transporter v7-3', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=103367458, end_position_on_the_genomic_accession=103419379, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  100504267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504267, CurrentID=0, Status='live', Symbol='Trp53cor1', Aliases='Gm16197, OTTMUSG00000031656, linc-p21, lincRNA-p21', description='tumor protein p53 pathway corepressor 1', other_designations='', map_location='17 15.04 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29057474, end_position_on_the_genomic_accession=29079126, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  234199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234199, CurrentID=0, Status='live', Symbol='Fgl1', Aliases='Mfire1', description='fibrinogen-like protein 1', other_designations='MFIRE-1|fibrinogen-related protein 1', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=41191434, end_position_on_the_genomic_accession=41217978, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  100504285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504285, CurrentID=0, Status='live', Symbol='Abhd12b', Aliases='Gm804', description='abhydrolase domain containing 12B', other_designations='', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70154171, end_position_on_the_genomic_accession=70183206, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  103135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103135, CurrentID=0, Status='live', Symbol='Pan2', Aliases='1200014O24Rik, AI047843, AW742773, Usp52, mKIAA0710', description='PAN2 polyA specific ribonuclease subunit homolog (S. cerevisiae)', other_designations='AN2 polyA specific ribonuclease subunit homolog|PAB-dependent poly(A)-specific ribonuclease subunit 2|PAB1P-dependent poly(A)-nuclease|PAN deadenylation complex catalytic subunit 2|inactive ubiquitin carboxyl-terminal hydrolase 52|ubiquitin specific peptidase 52|ubiquitin specific protease 52', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128303331, end_position_on_the_genomic_accession=128321358, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  103136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103136, CurrentID=0, Status='live', Symbol='Pwp1', Aliases='2310058A11Rik, 2610205J09Rik, AI098170, AL022879, AV083409, AW550690', description='PWP1 homolog, endonuclein', other_designations='PWP1 homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85871831, end_position_on_the_genomic_accession=85889103, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  103140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103140, CurrentID=0, Status='live', Symbol='Gstt3', Aliases='AI118089', description='glutathione S-transferase, theta 3', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75774122, end_position_on_the_genomic_accession=75781483, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  234214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234214, CurrentID=0, Status='live', Symbol='Sorbs2', Aliases='2010203O03Rik, 9430041O17Rik, A530071H08, Argbp2, mKIAA0777, nArgBP2', description='sorbin and SH3 domain containing 2', other_designations='Arg/Abl-interacting protein ArgBP2|arg-binding protein 2|arg/Abl-interacting protein 2', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45507629, end_position_on_the_genomic_accession=45827906, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  234219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234219, CurrentID=0, Status='live', Symbol='Helt', Aliases='A830086M02, Hesl, Mgn, megane', description='helt bHLH transcription factor', other_designations='HES/HEY-like transcription factor|Heslike|bHLH transcription factor megane|mesencephalic GABAergic neurons enhancer of split related protein|protein megane', map_location='8 26.32 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=46292043, end_position_on_the_genomic_accession=46294671, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  103149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103149, CurrentID=0, Status='live', Symbol='Upb1', Aliases='AI195023', description='ureidopropionase, beta', other_designations='N-carbamoyl-beta-alanine amidohydrolase|beta-alanine synthase', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75406911, end_position_on_the_genomic_accession=75441679, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  103161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103161, CurrentID=0, Status='live', Symbol='Apof', Aliases='AI255964', description='apolipoprotein F', other_designations='LVIF|apo-F|leukemia virus-inactivating factor', map_location='10 76.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128267997, end_position_on_the_genomic_accession=128270151, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  103172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103172, CurrentID=0, Status='live', Symbol='Chchd10', Aliases='1620401E04Rik, AI267078, Ndg2', description='coiled-coil-helix-coiled-coil-helix domain containing 10', other_designations='Nur77 downstream gene 2|Nur77 downstream protein 2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75932970, end_position_on_the_genomic_accession=75937747, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  234258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234258, CurrentID=0, Status='live', Symbol='Neil3', Aliases='AI449477, BC034753, C85903', description='nei like 3 (E. coli)', other_designations='DNA glycosylase FPG2|DNA glycosylase/AP lyase Neil3|endonuclease VIII-like 3|nei endonuclease VIII-like 3|nei-like protein 3', map_location='8 B1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=53586867, end_position_on_the_genomic_accession=53639065, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  234267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234267, CurrentID=0, Status='live', Symbol='Gpm6a', Aliases='Gpm6, M6A', description='glycoprotein m6a', other_designations='', map_location='8 B3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=54779434, end_position_on_the_genomic_accession=55060878, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  103199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103199, CurrentID=0, Status='live', Symbol='Fig4', Aliases='A530089I17Rik, AI326867, Sac3', description='FIG4 phosphoinositide 5-phosphatase', other_designations='FIG4 homolog|SAC domain-containing protein 3|Sac domain-containing inositol phosphatase 3|phosphatidylinositol 3,5-bisphosphate 5-phosphatase', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41188172, end_position_on_the_genomic_accession=41303241, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  628870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628870, CurrentID=0, Status='live', Symbol='Otogl', Aliases='EG628870, Gm6924', description='otogelin-like', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=107760530, end_position_on_the_genomic_accession=107912174, orientation='minus', exon_count=60, OMIM='', no_hdr0=''),
  103213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103213, CurrentID=0, Status='live', Symbol='Traf3ip2', Aliases='AI429613, Act1, CIKS', description='TRAF3 interacting protein 2', other_designations='', map_location='10 B1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=39612934, end_position_on_the_genomic_accession=39655307, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  103214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103214, CurrentID=0, Status='live', Symbol='Tmevpg1', Aliases='AI447422', description="Theiler's murine encephalomyelitis virus persistence candidate gene 1", other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=118502035, end_position_on_the_genomic_accession=118556525, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  103220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103220, CurrentID=0, Status='live', Symbol='Ttc41', Aliases='AI449705, AU041509, Gnn', description='tetratricopeptide repeat domain 41', other_designations='TPR repeat protein 41|TPR repeat protein GNN|grp94-neighboring nucleotidase|tetratricopeptide repeat protein GNN', map_location='10 43.07 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=86705632, end_position_on_the_genomic_accession=86776844, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  103236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103236, CurrentID=0, Status='live', Symbol='Csnk1g2', Aliases='2810429I12Rik, AI463719', description='casein kinase 1, gamma 2', other_designations='CKI-gamma 2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80622705, end_position_on_the_genomic_accession=80640771, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  234309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234309, CurrentID=0, Status='live', Symbol='Cbr4', Aliases='A730083J17Rik, BC009118', description='carbonyl reductase 4', other_designations='3-oxoacyl-[acyl-carrier-protein] reductase|quinone reductase CBR4', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=61487734, end_position_on_the_genomic_accession=61503500, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  234311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234311, CurrentID=0, Status='live', Symbol='Ddx60', Aliases='9830118M07', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 60', other_designations='', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=61928066, end_position_on_the_genomic_accession=62037704, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  100504399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504399, CurrentID=0, Status='live', Symbol='C130021I20Rik', Aliases='', description='Riken cDNA C130021I20 gene', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33641193, end_position_on_the_genomic_accession=33645663, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100504404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504404, CurrentID=0, Status='live', Symbol='H2-Ea-ps', Aliases='AI323765, E-alpha-f, H-2Ea, H2-Ea, I-Ealpha, Ia-3, Ia3, MHC-H2-Ea', description='histocompatibility 2, class II antigen E alpha, pseudogene', other_designations='H-2 class II histocompatibility antigen, E-D alpha chain|H-2 class II histocompatibility antigen, E-K alpha chain|H-2 class II histocompatibility antigen, E-U alpha chain|H2-IE-alpha|I-E alpha MHC class II|MHC class II H2-E alpha chain|MHC class II alpha chain|MHC class II antigen E alpha|MHC class II protein', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34341959, end_position_on_the_genomic_accession=34347547, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  234329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234329, CurrentID=0, Status='live', Symbol='Trim60', Aliases='2czf45, Rnf33', description='tripartite motif-containing 60', other_designations='ring finger protein 33', map_location='8 B3.1-3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=64998448, end_position_on_the_genomic_accession=65018688, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103266, CurrentID=0, Status='live', Symbol='Tmem263', Aliases='AI597468', description='transmembrane protein 263', other_designations='UPF0444 transmembrane protein C12orf23 homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=85102627, end_position_on_the_genomic_accession=85117747, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103268, CurrentID=0, Status='live', Symbol='Cep57l1', Aliases='2410017P07Rik, 4930484D11Rik, AI607327, cep57R', description='centrosomal protein 57-like 1', other_designations='centrosomal protein 57kDa-like protein 1|centrosomal protein of 57 kDa-related protein|cep57-related protein', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41718840, end_position_on_the_genomic_accession=41809868, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  234344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234344, CurrentID=0, Status='live', Symbol='Naf1', Aliases='Gm174', description='nuclear assembly factor 1 homolog (S. cerevisiae)', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=66859957, end_position_on_the_genomic_accession=66890564, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  100504429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504429, CurrentID=0, Status='live', Symbol='Spin2d', Aliases='4930408F14Rik', description='spindlin family, member 2D', other_designations='', map_location='X 37.29 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73175302, end_position_on_the_genomic_accession=73176989, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  234353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234353, CurrentID=0, Status='live', Symbol='Psd3', Aliases='4931420C21Rik, AI661273, BC003498, D430018P08, EFA6D', description='pleckstrin and Sec7 domain containing 3', other_designations='exchange factor for ADP-ribosylation factor guanine nucleotide factor 6|guanine nucleotide exchange factor for ADP ribosylation factor 6|pleckstrin homology and SEC7 domain-containing protein 3', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=67689082, end_position_on_the_genomic_accession=68206052, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  234356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234356, CurrentID=0, Status='live', Symbol='Csgalnact1', Aliases='4732435N03Rik, CSGalNAcT-1', description='chondroitin sulfate N-acetylgalactosaminyltransferase 1', other_designations='beta1,4 N-acetylgalactosaminyltransferase|beta4GalNAcT-1|chondroitin beta-1,4-N-acetylgalactosaminyltransferase 1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=68356781, end_position_on_the_genomic_accession=68735673, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  234362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234362, CurrentID=0, Status='live', Symbol='Zfp868', Aliases='AI449175, AW108241, C530032E20, mszf62', description='zinc finger protein 868', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69610654, end_position_on_the_genomic_accession=69625562, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  234366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234366, CurrentID=0, Status='live', Symbol='Gatad2a', Aliases='1110066C11Rik, BC031407, C80248', description='GATA zinc finger domain containing 2A', other_designations='GATA zinc finger domain-containing protein 2A', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69907069, end_position_on_the_genomic_accession=69996384, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  234371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234371, CurrentID=0, Status='live', Symbol='Tmem161a', Aliases='AI428876, BB161850, BC021367', description='transmembrane protein 161A', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70172329, end_position_on_the_genomic_accession=70183681, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  234373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234373, CurrentID=0, Status='live', Symbol='Sugp2', Aliases='Sfrs14, Srsf14, mKIAA0365', description='SURP and G patch domain containing 2', other_designations='arginine/serine-rich-splicing factor 14|putative splicing factor, arginine/serine-rich 14|serine/arginine-rich splicing factor 14|splicing factor, arginine/serine-rich 14', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70234209, end_position_on_the_genomic_accession=70263106, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  234374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234374, CurrentID=0, Status='live', Symbol='Ddx49', Aliases='R27090_2', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 49', other_designations='DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 27|DEAD box protein 49|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 49', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70292866, end_position_on_the_genomic_accession=70302511, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  234378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234378, CurrentID=0, Status='live', Symbol='Klhl26', Aliases='C630013N10Rik, Klkl26', description='kelch-like 26', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70450221, end_position_on_the_genomic_accession=70476988, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  234384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234384, CurrentID=0, Status='live', Symbol='Mpv17l2', Aliases='Fksg24', description='MPV17 mitochondrial membrane protein-like 2', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70758649, end_position_on_the_genomic_accession=70760921, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  234388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234388, CurrentID=0, Status='live', Symbol='Ccdc124', Aliases='1810023B24Rik', description='coiled-coil domain containing 124', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70868226, end_position_on_the_genomic_accession=70873490, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  100504474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504474, CurrentID=0, Status='live', Symbol='Smlr1', Aliases='2010003K15Rik', description='small leucine-rich protein 1', other_designations='', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=25527943, end_position_on_the_genomic_accession=25536272, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  234395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234395, CurrentID=0, Status='live', Symbol='Ushbp1', Aliases='2210404N08Rik, 9430064H24, AIEBP, MCC2', description='Usher syndrome 1C binding protein 1', other_designations='AIE-75 binding protein|USH1C-binding protein 1|ushbp1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71384274, end_position_on_the_genomic_accession=71395835, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  234396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234396, CurrentID=0, Status='live', Symbol='Ankle1', Aliases='8430438L13Rik, Ankrd41, C430003I23, Lem3', description='ankyrin repeat and LEM domain containing 1', other_designations='LEM-domain containing 3|ankyrin repeat domain 41', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71406005, end_position_on_the_genomic_accession=71410543, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  234404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234404, CurrentID=0, Status='live', Symbol='Nxnl1', Aliases='A930031O08Rik, RdCVF, Txnl6', description='nucleoredoxin-like 1', other_designations='rod-derived cone viability factor|thioredoxin-like 6|thioredoxin-like protein 6', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71560555, end_position_on_the_genomic_accession=71568666, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  234407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234407, CurrentID=0, Status='live', Symbol='Colgalt1', Aliases='2810024B22Rik, Glt25d1', description='collagen beta(1-O)galactosyltransferase 1', other_designations='glycosyltransferase 25 domain containing 1|glycosyltransferase 25 family member 1|hydroxylysine galactosyltransferase 1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71610991, end_position_on_the_genomic_accession=71624911, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  627626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=627626, CurrentID=0, Status='live', Symbol='Ptchd4', Aliases='3110082D06Rik, Gm321', description='patched domain containing 4', other_designations='patched domain-containing protein C6orf138 homolog', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=42315810, end_position_on_the_genomic_accession=42507742, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100504491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504491, CurrentID=0, Status='live', Symbol='Dcdc2b', Aliases='Gm12964, Gm1366, OTTMUSG00000009523', description='doublecortin domain containing 2b', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129608331, end_position_on_the_genomic_accession=129614477, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  234413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234413, CurrentID=0, Status='live', Symbol='Zfp961', Aliases='A230105L22Rik, BC065788, mszf37, mszf55-2', description='zinc finger protein 961', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71951066, end_position_on_the_genomic_accession=71970333, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  234421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234421, CurrentID=0, Status='live', Symbol='Cib3', Aliases='C730014M21Rik, Gm1107, KIP3', description='calcium and integrin binding family member 3', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72192699, end_position_on_the_genomic_accession=72213092, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  100504518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504518, CurrentID=0, Status='live', Symbol='3425401B19Rik', Aliases='', description='RIKEN cDNA 3425401B19 gene', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32659118, end_position_on_the_genomic_accession=32707821, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  234463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234463, CurrentID=0, Status='live', Symbol='Tmem184c', Aliases='8430433H16Rik, AI645453, AL033298, BC004056, Tmem34', description='transmembrane protein 184C', other_designations='transmembrane protein 34', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=77595978, end_position_on_the_genomic_accession=77610759, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  103406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103406, CurrentID=0, Status='live', Symbol='Zfr2', Aliases='2010013I23Rik, 9130206N08Rik, AW045600, mKIAA1086', description='zinc finger RNA binding protein 2', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81233163, end_position_on_the_genomic_accession=81252123, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  103425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103425, CurrentID=0, Status='live', Symbol='Ncln', Aliases='3100002P13Rik, AW120537', description='nicalin homolog (zebrafish)', other_designations='nicastrin-like protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81486459, end_position_on_the_genomic_accession=81496363, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  100504594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504594, CurrentID=0, Status='live', Symbol='A630012P03Rik', Aliases='', description='RIKEN cDNA A630012P03 gene', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52575843, end_position_on_the_genomic_accession=52619047, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  234515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234515, CurrentID=0, Status='live', Symbol='Inpp4b', Aliases='E130107I17Rik', description='inositol polyphosphate-4-phosphatase, type II', other_designations='D-myo-inositol-3,4-bisphosphate 4-phosphohydrolase|inositol polyphosphate 4-phosphatase type II beta isoform|phosphatidylinositol-3,4-bisphosphate 4-phosphatase|phosphoinositide 4-phosphatase|type II inositol-3,4-bisphosphate 4-phosphatase', map_location='8 39.02 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=81342342, end_position_on_the_genomic_accession=82152019, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  103466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103466, CurrentID=0, Status='live', Symbol='Nt5dc3', Aliases='AU040402, AW540062, C630002B14Rik, TU12B1-TY', description="5'-nucleotidase domain containing 3", other_designations='TU12B1-TY protein homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=86778990, end_position_on_the_genomic_accession=86838389, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  103468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103468, CurrentID=0, Status='live', Symbol='Nup107', Aliases='AW541137, C76801', description='nucleoporin 107', other_designations='107 kDa nucleoporin|nucleoporin Nup107', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117750621, end_position_on_the_genomic_accession=117792736, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  234542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234542, CurrentID=0, Status='live', Symbol='Rtbdn', Aliases='A330096I21Rik, BC022744, Rtbnd', description='retbindin', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84946263, end_position_on_the_genomic_accession=84956606, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  234549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234549, CurrentID=0, Status='live', Symbol='Heatr3', Aliases='C030036P15Rik', description='HEAT repeat containing 3', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88137885, end_position_on_the_genomic_accession=88171952, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  17248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17248, CurrentID=0, Status='live', Symbol='Mdm4', Aliases='4933417N07Rik, AA414968, AL023055, AU018793, AU021806, C85810, Mdmx', description='transformed mouse 3T3 cell double minute 4', other_designations='double minute 4 protein|mdm2-like p53-binding protein|p53-binding protein Mdm4|protein Mdmx', map_location='1 G-G', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132986105, end_position_on_the_genomic_accession=133030587, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  234564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234564, CurrentID=0, Status='live', Symbol='Ces1f', Aliases='AU018778, CesML1, TGH-2', description='carboxylesterase 1F', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93256236, end_position_on_the_genomic_accession=93279766, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  234577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234577, CurrentID=0, Status='live', Symbol='Cpne2', Aliases='3322401K10Rik', description='copine II', other_designations='copine 2', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94533028, end_position_on_the_genomic_accession=94570529, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  234582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234582, CurrentID=0, Status='live', Symbol='Ccdc102a', Aliases='BC027663, Gm586', description='coiled-coil domain containing 102A', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94902869, end_position_on_the_genomic_accession=94918098, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  100504663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504663, CurrentID=0, Status='live', Symbol='Atg14', Aliases='4832427M01L, D14Ertd114e, D14Ertd436e, Atg14', description='autophagy related 14', other_designations='Atg14-like protein|VATG14 autophagy related 14 homolog|autophagy-related protein 14-like protein|barkor', map_location='14 24.6 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=47540893, end_position_on_the_genomic_accession=47568434, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  234593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234593, CurrentID=0, Status='live', Symbol='Ndrg4', Aliases='D8Bwg1337e, Ndr1-rs, Ndr4, R74996, SMAP-8', description='N-myc downstream regulated gene 4', other_designations='N-myc downstream regulated 1, related sequence|N-myc downstream-regulated gene 4 protein|protein Ndr4', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95676985, end_position_on_the_genomic_accession=95715119, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  234594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234594, CurrentID=0, Status='live', Symbol='Cnot1', Aliases='6030411K04Rik, AA815922, D830048B13', description='CCR4-NOT transcription complex, subunit 1', other_designations='CCR4-associated factor 1', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95719451, end_position_on_the_genomic_accession=95807479, orientation='minus', exon_count=49, OMIM='', no_hdr0=''),
  234595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234595, CurrentID=0, Status='live', Symbol='Slc38a7', Aliases='D430050E18', description='solute carrier family 38, member 7', other_designations='amino acid transporter|solute carrier family 38 member 7', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95835920, end_position_on_the_genomic_accession=95853577, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  103534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103534, CurrentID=0, Status='live', Symbol='Mgat4b', Aliases='AA407995, GnTIVb', description='mannoside acetylglucosaminyltransferase 4, isoenzyme B', other_designations='GlcNAc-T IVb|N-acetylglucosaminyltransferase IVb|N-glycosyl-oligosaccharide-glycoprotein N-acetylglucosaminyltransferase IVb|UDP-N-acetylglucosamine: alpha-1,3-D-mannoside beta-1,4-N-acetylglucosaminyltransferase IVb|gnT-IVb|mannosyl (alpha-1,3-)-glycoprotein beta-1,4-N-acetylglucosaminyltransferase, isoenzyme B', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50225335, end_position_on_the_genomic_accession=50235103, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  103537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103537, CurrentID=0, Status='live', Symbol='Mbtd1', Aliases='AA408199, AI194990, hemp', description='mbt domain containing 1', other_designations='', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=93885850, end_position_on_the_genomic_accession=93946985, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  103551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103551, CurrentID=0, Status='live', Symbol='E130012A19Rik', Aliases='AA409164, AI413509, AW539173, E13, esPRC2p48', description='RIKEN cDNA E130012A19 gene', other_designations='ES cell-specific PRC2 subunit p48', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97627387, end_position_on_the_genomic_accession=97629716, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  103554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103554, CurrentID=0, Status='live', Symbol='Psme4', Aliases='AA409398, AU041366, TEMO, mKIAA0077', description='proteasome (prosome, macropain) activator subunit 4', other_designations='proteasome activator PA200', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30771328, end_position_on_the_genomic_accession=30881150, orientation='plus', exon_count=49, OMIM='', no_hdr0=''),
  100504715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100504715, CurrentID=0, Status='live', Symbol='Tmppe', Aliases='', description='transmembrane protein with metallophosphoesterase domain', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114401095, end_position_on_the_genomic_accession=114411201, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  103573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103573, CurrentID=0, Status='live', Symbol='Xpo1', Aliases='AA420417, Crm1, Exp1', description='exportin 1', other_designations='chromosome region maintenance 1 protein homolog|exportin 1, CRM1 homolog', map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=23255914, end_position_on_the_genomic_accession=23297597, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  103583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103583, CurrentID=0, Status='live', Symbol='Fbxw11', Aliases='2310065A07Rik, AA536858, BTRC2, BTRCP2, Fbxw1b, HOS', description='F-box and WD-40 domain protein 11', other_designations='F-box and WD repeats protein beta-TrCP2|F-box and WD-40 domain protein 1B|F-box/WD repeat-containing protein 1B|homologous to Slimb protein', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32642003, end_position_on_the_genomic_accession=32746816, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  627872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=627872, CurrentID=0, Status='live', Symbol='Dnah7a', Aliases='Dnahc7, Dnahc7a, Gm992', description='dynein, axonemal, heavy chain 7A', other_designations='dynein, axonemal, heavy chain 7', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=53396494, end_position_on_the_genomic_accession=53708118, orientation='minus', exon_count=65, OMIM='', no_hdr0=''),
  234663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234663, CurrentID=0, Status='live', Symbol='Dync1li2', Aliases='AA409702, C920003I06, Dncli2, Dnclic2, LIC2', description='dynein, cytoplasmic 1 light intermediate chain 2', other_designations='dynein light intermediate chain 2, cytosolic|dynein, cytoplasmic, light intermediate chain 2|dynein, cytoplasmic, light intermediate polypeptide 2', map_location='8 53.04 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104417674, end_position_on_the_genomic_accession=104443224, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  234664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234664, CurrentID=0, Status='live', Symbol='Nae1', Aliases='59kDa, Appbp1', description='NEDD8 activating enzyme E1 subunit 1', other_designations='APP-BP1|amyloid beta precursor protein binding protein 1, 59kDa|amyloid beta precursor protein-binding protein 1, 59 kDa|amyloid protein-binding protein 1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104511028, end_position_on_the_genomic_accession=104534655, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  234669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234669, CurrentID=0, Status='live', Symbol='Ces2b', Aliases='', description='carboxyesterase 2B', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104828494, end_position_on_the_genomic_accession=104840093, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  234671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234671, CurrentID=0, Status='live', Symbol='Ces2c', Aliases='ACH M1, CES 2, Ces2, ces2A3', description='carboxylesterase 2C', other_designations='acylcarnitine hydrolase M1|carboxylesterase 2|carboxylesterase ML3|peroxisome proliferator-inducible acylcarnitine hydrolase', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104847068, end_position_on_the_genomic_accession=104854483, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  234673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234673, CurrentID=0, Status='live', Symbol='Ces2e', Aliases='9030624L02Rik, Ces5', description='carboxylesterase 2E', other_designations='carboxylesterase 5', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104926260, end_position_on_the_genomic_accession=104934672, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  234677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234677, CurrentID=0, Status='live', Symbol='Ces4a', Aliases='Ces8', description='carboxylesterase 4A', other_designations='carboxylesterase 8 (putative)', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105131800, end_position_on_the_genomic_accession=105150109, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  234678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234678, CurrentID=0, Status='live', Symbol='D230025D16Rik', Aliases='BC006874, Lin10', description='RIKEN cDNA D230025D16 gene', other_designations='lin 10 homolog|lin-10 homolog', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105225150, end_position_on_the_genomic_accession=105253053, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  234683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234683, CurrentID=0, Status='live', Symbol='Elmo3', Aliases='9930107J06, BC058752, CED-12, Ced12', description='engulfment and cell motility 3', other_designations='engulfment and cell motility 3, ced-12 homolog', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105305399, end_position_on_the_genomic_accession=105310623, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  234684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234684, CurrentID=0, Status='live', Symbol='Lrrc29', Aliases='A630024J02Rik, FBL9, FBXL9', description='leucine rich repeat containing 29', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105312340, end_position_on_the_genomic_accession=105326331, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  234686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234686, CurrentID=0, Status='live', Symbol='Fhod1', Aliases='BC017144, FHOS, FHOS1', description='formin homology 2 domain containing 1', other_designations='formin homolog overexpressed in spleen 1|formin homology 2 domain-containing protein 1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105329160, end_position_on_the_genomic_accession=105347970, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  234695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234695, CurrentID=0, Status='live', Symbol='Rltpr', Aliases='CG1399-PB, D130029J02Rik, Gm585', description='RGD motif, leucine rich repeats, tropomodulin domain and proline-rich containing', other_designations='RGD, leucine-rich repeat, tropomodulin and proline-rich containing protein|RGD, leucine-rich repeat, tropomodulin and proline-rich-containing protein', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105686010, end_position_on_the_genomic_accession=105698195, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  234699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234699, CurrentID=0, Status='live', Symbol='Edc4', Aliases='BC022641', description='enhancer of mRNA decapping 4', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105880873, end_position_on_the_genomic_accession=105893225, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  234700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234700, CurrentID=0, Status='live', Symbol='Nrn1l', Aliases='Cpg15-2, G630049C14Rik', description='neuritin 1-like', other_designations='MRCC2446-like protein|candidate plasticity gene 15-2 protein', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105893567, end_position_on_the_genomic_accession=105895023, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  234723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234723, CurrentID=0, Status='live', Symbol='Txnl4b', Aliases='AI595343, D530025J19Rik, DLP, Dim2', description='thioredoxin-like 4B', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109565986, end_position_on_the_genomic_accession=109574051, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  234724 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234724, CurrentID=0, Status='live', Symbol='Tat', Aliases='', description='tyrosine aminotransferase', other_designations='L-tyrosine:2-oxoglutarate aminotransferase|tyrosine transaminase', map_location='8 57.38 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109990436, end_position_on_the_genomic_accession=109999804, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  234725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234725, CurrentID=0, Status='live', Symbol='Zfp612', Aliases='B230354B21Rik', description='zinc finger protein 612', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110079692, end_position_on_the_genomic_accession=110092753, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  103655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103655, CurrentID=0, Status='live', Symbol='Sec14l4', Aliases='AI256582, AW536553, SPF, TAP', description='SEC14-like lipid binding 4', other_designations='SEC14-like 2|SEC14-like 4|alpha-tocopherol associated protein|squalene transfer protein|supernatant protein factor', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4031378, end_position_on_the_genomic_accession=4052571, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  234728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234728, CurrentID=0, Status='live', Symbol='Cmtr2', Aliases='AU022703, C730036L12Rik, Ftsjd1, MTr2', description='cap methyltransferase 2', other_designations="FtsJ methyltransferase domain containing 1|cap2 2'O-ribose methyltransferase 2|ftsJ methyltransferase domain-containing protein 1", map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110215347, end_position_on_the_genomic_accession=110224489, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  234729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234729, CurrentID=0, Status='live', Symbol='Vac14', Aliases='AA959718, BC032215, D8Wsu151e, Tax1bp2, Trx, ingls', description='Vac14 homolog (S. cerevisiae)', other_designations='Tax1 (human T-cell leukemia virus type I) binding protein 1|Tax1 (human T-cell leukemia virus type I) binding protein 2|vacuole morphology 14', map_location='8 57.59 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110618566, end_position_on_the_genomic_accession=110720398, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  234730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234730, CurrentID=0, Status='live', Symbol='Fuk', Aliases='1110046B12Rik, AI303278', description='fucokinase', other_designations='mFLJ00241 protein', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110882456, end_position_on_the_genomic_accession=110902488, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  234733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234733, CurrentID=0, Status='live', Symbol='Ddx19b', Aliases='2810457M08Rik, 4921519L13Rik, AW260119', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 19b', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111003186, end_position_on_the_genomic_accession=111031751, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  234734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234734, CurrentID=0, Status='live', Symbol='Aars', Aliases='AI316495, C76919, sti', description='alanyl-tRNA synthetase', other_designations='alaRS|alanyl-tRNA synthetase, cytoplasmic', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111027778, end_position_on_the_genomic_accession=111055569, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  234736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234736, CurrentID=0, Status='live', Symbol='Rfwd3', Aliases='BC027246', description='ring finger and WD repeat domain 3', other_designations='RING finger and WD repeat domain-containing protein 3|RING finger protein 201', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111270944, end_position_on_the_genomic_accession=111300260, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  234740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234740, CurrentID=0, Status='live', Symbol='Tmem231', Aliases='4932417I16Rik, Gm178', description='transmembrane protein 231', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=111912011, end_position_on_the_genomic_accession=111933811, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  103677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103677, CurrentID=0, Status='live', Symbol='Smg6', Aliases='AI317223, AU041178', description='Smg-6 homolog, nonsense mediated mRNA decay factor (C. elegans)', other_designations='EST1-like protein A|ever shorter telomeres 1A|telomerase subunit EST1A', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74925872, end_position_on_the_genomic_accession=75164448, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  103694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103694, CurrentID=0, Status='live', Symbol='Tmed4', Aliases='1110014L17Rik, AI326346', description='transmembrane emp24 protein transport domain containing 4', other_designations='ERS25|endoplasmic reticulum stress-response protein 25|p24 family protein alpha-3|p24alpha3|p26', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6270714, end_position_on_the_genomic_accession=6274837, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  234776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234776, CurrentID=0, Status='live', Symbol='Atmin', Aliases='Asciz, gpg6, mKIAA0431', description='ATM interactor', other_designations='ASCIZ|ATM/ATR-substrate CHEK2-interacting zinc finger protein|ATM/ATR-substrate CHK2-interacting zinc finger protein|atmin', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=116943386, end_position_on_the_genomic_accession=116960445, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  234779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234779, CurrentID=0, Status='live', Symbol='Plcg2', Aliases='PLC-gamma-2, PLCgamma2, Plcg-2', description='phospholipase C, gamma 2', other_designations='phosphoinositide phospholipase C-gamma-2|phospholipase C-gamma-2', map_location='8 64.26 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=117498291, end_position_on_the_genomic_accession=117635142, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  103710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103710, CurrentID=0, Status='live', Symbol='Slc35e4', Aliases='A330108F03Rik, AI414849, AI536442', description='solute carrier family 35, member E4', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3907022, end_position_on_the_genomic_accession=3914664, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  103711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103711, CurrentID=0, Status='live', Symbol='Pnpo', Aliases='AI415282', description="pyridoxine 5'-phosphate oxidase", other_designations='pyridoxamine-phosphate oxidase', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96937816, end_position_on_the_genomic_accession=96944019, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  103712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103712, CurrentID=0, Status='live', Symbol='6330403K07Rik', Aliases='AI415455, AV161034, UGS148', description='RIKEN cDNA 6330403K07 gene', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=71031941, end_position_on_the_genomic_accession=71033617, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  234796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234796, CurrentID=0, Status='live', Symbol='Klhl36', Aliases='', description='kelch-like 36', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119862231, end_position_on_the_genomic_accession=119876989, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  234797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234797, CurrentID=0, Status='live', Symbol='6430548M08Rik', Aliases='AW049007, Kiaa0513, mKIAA0513', description='RIKEN cDNA 6430548M08 gene', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120114152, end_position_on_the_genomic_accession=120165307, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  103733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103733, CurrentID=0, Status='live', Symbol='Tubg1', Aliases='1500010O08Rik, AI451582, AI503389, Tubg', description='tubulin, gamma 1', other_designations='GCP-1|gamma-1-tubulin|gamma-tubulin complex component 1|tubulin, gamma polypeptide', map_location='11 64.24 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101120131, end_position_on_the_genomic_accession=101126419, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  103737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103737, CurrentID=0, Status='live', Symbol='Pex12', Aliases='AI451906', description='peroxisomal biogenesis factor 12', other_designations='peroxin-12', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83294645, end_position_on_the_genomic_accession=83298996, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  628027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628027, CurrentID=0, Status='live', Symbol='Igkv1-133', Aliases='EG628027, Gm6827, Igk-Vk2(70/1)', description='immunoglobulin kappa variable 1-133', other_designations='immunoglobulin kappa chain variable 1-133', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67724914, end_position_on_the_genomic_accession=67725661, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103742, CurrentID=0, Status='live', Symbol='Mien1', Aliases='1810046J19Rik, AI463380, Rdx12', description='migration and invasion enhancer 1', other_designations='protein C17orf37 homolog', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=98437708, end_position_on_the_genomic_accession=98438988, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  103743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103743, CurrentID=0, Status='live', Symbol='Tmem98', Aliases='6530411B15Rik, AI463522', description='transmembrane protein 98', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80810415, end_position_on_the_genomic_accession=80822033, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  234825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234825, CurrentID=0, Status='live', Symbol='Klhdc4', Aliases='AA408426, AV352552, BC012312, G430025P05Rik', description='kelch domain containing 4', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=121796308, end_position_on_the_genomic_accession=121829577, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  234836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234836, CurrentID=0, Status='live', Symbol='Il17c', Aliases='IL-17C', description='interleukin 17C', other_designations='cytokine CX2', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122422085, end_position_on_the_genomic_accession=122423639, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  103765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103765, CurrentID=0, Status='live', Symbol='Tmem17', Aliases='AI503894, AW049890', description='transmembrane protein 17', other_designations='', map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=22512283, end_position_on_the_genomic_accession=22519231, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  234839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234839, CurrentID=0, Status='live', Symbol='Piezo1', Aliases='9630020g22, Fam38a, mKIAA0233', description='piezo-type mechanosensitive ion channel component 1', other_designations='family with sequence similarity 38, member A|protein PIEZO1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122481698, end_position_on_the_genomic_accession=122551657, orientation='minus', exon_count=52, OMIM='', no_hdr0=''),
  103768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103768, CurrentID=0, Status='live', Symbol='Tubg2', Aliases='AI504772, Tubgl', description='tubulin, gamma 2', other_designations='gamma-2-tubulin', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101155871, end_position_on_the_genomic_accession=101161787, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  103775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103775, CurrentID=0, Status='live', Symbol='Slc25a41', Aliases='4933406J04Rik, AI507137', description='solute carrier family 25, member 41', other_designations='mitochondrial ATP-Mg/Pi carrier protein|slc25a41', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57031454, end_position_on_the_genomic_accession=57042086, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  234852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234852, CurrentID=0, Status='live', Symbol='Chmp1a', Aliases='2900018H07Rik, Pcoln3', description='charged multivesicular body protein 1A', other_designations='chromatin modifying protein 1A|chromatin-modifying protein 1a|procollagen (type III) N-endopeptidase', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123204261, end_position_on_the_genomic_accession=123212788, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  234854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234854, CurrentID=0, Status='live', Symbol='Cdk10', Aliases='BC017131', description='cyclin-dependent kinase 10', other_designations='cell division protein kinase 10|cyclin dependent kinase 10|cyclin-dependent kinase (CDC2-like) 10', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123224840, end_position_on_the_genomic_accession=123232256, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  103784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103784, CurrentID=0, Status='live', Symbol='Wdr92', Aliases='AI553587, HZGJ', description='WD repeat domain 92', other_designations='WD repeat-containing protein Monad', map_location='11 A2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=17211893, end_position_on_the_genomic_accession=17235200, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  234857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234857, CurrentID=0, Status='live', Symbol='Spire2', Aliases='BC026502, Spir-2, Spir2', description='spire homolog 2 (Drosophila)', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123332713, end_position_on_the_genomic_accession=123369518, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  234865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234865, CurrentID=0, Status='live', Symbol='Nup133', Aliases='mermaid', description='nucleoporin 133', other_designations='133 kDa nucleoporin|nucleoporin Nup133', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123897123, end_position_on_the_genomic_accession=123949265, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  234875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234875, CurrentID=0, Status='live', Symbol='Ttc13', Aliases='BC017545', description='tetratricopeptide repeat domain 13', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124671327, end_position_on_the_genomic_accession=124722013, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  103806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103806, CurrentID=0, Status='live', Symbol='Maml1', Aliases='AI644666, AW743257, D930008C07Rik, Mam-1, Mam1, mKIAA0200', description='mastermind like 1 (Drosophila)', other_designations='maml1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50255634, end_position_on_the_genomic_accession=50292336, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  628100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628100, CurrentID=0, Status='live', Symbol='Fbxo39', Aliases='1700010H23Rik', description='F-box protein 39', other_designations='F-box protein 38', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72314370, end_position_on_the_genomic_accession=72319419, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  234889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234889, CurrentID=0, Status='live', Symbol='Gucy1a2', Aliases='6330407I18Rik, A230060L24Rik', description='guanylate cyclase 1, soluble, alpha 2', other_designations='', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=3532349, end_position_on_the_genomic_accession=3905791, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  103836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103836, CurrentID=0, Status='live', Symbol='Zfp692', Aliases='AI746306, AI839920-ps, Znf692, Zfp692', description='zinc finger protein 692', other_designations='zinc finger protein 692, pseudogene', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58306733, end_position_on_the_genomic_accession=58314613, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  234911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234911, CurrentID=0, Status='live', Symbol='Mmp27', Aliases='Gm180', description='matrix metallopeptidase 27', other_designations='matrix metalloproteinase 27', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=7571396, end_position_on_the_genomic_accession=7581885, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  234912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234912, CurrentID=0, Status='live', Symbol='9230110C19Rik', Aliases='', description='RIKEN cDNA 9230110C19 gene', other_designations='', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=8021672, end_position_on_the_genomic_accession=8042894, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  103841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103841, CurrentID=0, Status='live', Symbol='Cuedc1', Aliases='AI841487, C330016O16Rik', description='CUE domain containing 1', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88099146, end_position_on_the_genomic_accession=88194140, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  234915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234915, CurrentID=0, Status='live', Symbol='Cep126', Aliases='Kiaa1377, mKIAA1377', description='centrosomal protein 126', other_designations='', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=8076460, end_position_on_the_genomic_accession=8134294, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  103844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103844, CurrentID=0, Status='live', Symbol='Inca1', Aliases='AI842396', description='inhibitor of CDK, cyclin A1 interacting protein 1', other_designations='inhibitor of CDK interacting with cyclin A1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70688361, end_position_on_the_genomic_accession=70700206, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  103850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103850, CurrentID=0, Status='live', Symbol='Nt5m', Aliases='2010013E09Rik, AI846937, dNT-2', description="5',3'-nucleotidase, mitochondrial", other_designations="5',3'-deoxyribonucleotidase, mitochondrial|deoxy-5'-nucleotidase 2", map_location='11 B', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59847324, end_position_on_the_genomic_accession=59876533, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  497097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=497097, CurrentID=0, Status='live', Symbol='Xkr4', Aliases='AY534250, Gm210, XRG4, mKIAA1889', description='X-linked Kx blood group related 4', other_designations='X Kell blood group precursor related family member 4', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=3199731, end_position_on_the_genomic_accession=3671742, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  234959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234959, CurrentID=0, Status='live', Symbol='Med17', Aliases='77kDa, C330002H14Rik, Crsp6, Trap80', description='mediator complex subunit 17', other_designations='CRSP complex subunit 6|cofactor required for Sp1 transcriptional activation subunit 6|cofactor required for Sp1 transcriptional activation, subunit 6 (77kD)|thyroid hormone receptor-associated protein complex 80 kDa component', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15260351, end_position_on_the_genomic_accession=15279948, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  103889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103889, CurrentID=0, Status='live', Symbol='Hoxb2', Aliases='AI894218, Hox-2.8, Hoxbes2', description='homeobox B2', other_designations='homeo box B2|homeobox protein Hox-2.8', map_location='11 59.85 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96351632, end_position_on_the_genomic_accession=96354020, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  497106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=497106, CurrentID=0, Status='live', Symbol='Rnase12', Aliases='Rai1', description='ribonuclease, RNase A family, 12 (non-active)', other_designations='ribonuclease A i1|ribonuclease-like protein 12', map_location='14 B-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51056698, end_position_on_the_genomic_accession=51057242, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  234964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234964, CurrentID=0, Status='live', Symbol='Ccdc67', Aliases='4933401K09Rik, Deup1', description='coiled-coil domain containing 67', other_designations='coiled-coil domain-containing protein 67|deuterosomal protein 1', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15559864, end_position_on_the_genomic_accession=15628338, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  234967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234967, CurrentID=0, Status='live', Symbol='Slc36a4', Aliases='6330573I15Rik, PAT4', description='solute carrier family 36 (proton/amino acid symporter), member 4', other_designations='proton/amino acid transporter 4|solute carrier family 36 member 4', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15709769, end_position_on_the_genomic_accession=15742388, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  497113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=497113, CurrentID=0, Status='live', Symbol='Rnase11', Aliases='Raj1', description='ribonuclease, RNase A family, 11 (non-active)', other_designations='ribonuclease A j1', map_location='14 B-C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=51048694, end_position_on_the_genomic_accession=51055702, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  497114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=497114, CurrentID=0, Status='live', Symbol='Defa23', Aliases='Defcr23', description='defensin, alpha, 23', other_designations='defensin-related cryptdin-23', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21055040, end_position_on_the_genomic_accession=21056016, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  497118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=497118, CurrentID=0, Status='live', Symbol='Dcr3', Aliases='', description='decoy receptor 3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103912, CurrentID=0, Status='live', Symbol='Wo', Aliases='', description='wocko', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  234988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234988, CurrentID=0, Status='live', Symbol='Mbd3l2', Aliases='', description='methyl-CpG binding domain protein 3-like 2', other_designations='', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18430615, end_position_on_the_genomic_accession=18446316, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  103935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103935, CurrentID=0, Status='live', Symbol='Eh', Aliases='', description='hairy ears', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100505096, CurrentID=0, Status='live', Symbol='Gm21002', Aliases='', description='predicted gene, 21002', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21391790, end_position_on_the_genomic_accession=21418342, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  103946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103946, CurrentID=0, Status='live', Symbol='Br', Aliases='', description='brachyrrhine', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  628236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628236, CurrentID=0, Status='live', Symbol='Lipo4', Aliases='Gm6857', description='lipase, member O4', other_designations='', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=33499049, end_position_on_the_genomic_accession=33517770, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  235028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235028, CurrentID=0, Status='live', Symbol='Zfp426', Aliases='2900057C04Rik, KRAB1, Zfo61, Zfp68-rs1, Znf426, mKIAA4237', description='zinc finger protein 426', other_designations='zinc finger protein 68, related sequence 1', map_location='9 7.55 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20468548, end_position_on_the_genomic_accession=20493922, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  235033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235033, CurrentID=0, Status='live', Symbol='Rdh8', Aliases='Gm182, prRDH', description='retinol dehydrogenase 8', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20818504, end_position_on_the_genomic_accession=20826163, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  103962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103962, CurrentID=0, Status='live', Symbol='Rf', Aliases='', description='rib fusion', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103963, CurrentID=0, Status='live', Symbol='Rpn1', Aliases='AU018702, D6Wsu137e, Rpn-1', description='ribophorin I', other_designations='RPN-I|dolichyl-diphosphooligosaccharide--protein glycosyltransferase 67 kDa subunit|ribophorin-1', map_location='6 39.13 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=88084473, end_position_on_the_genomic_accession=88105304, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  235036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235036, CurrentID=0, Status='live', Symbol='Ppan', Aliases='A230087P06Rik, SSF1', description='peter pan homolog (Drosophila)', other_designations='second-step splicing factor 1|ssf-1|suppressor of sterile four 1', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20888175, end_position_on_the_genomic_accession=20892179, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  103967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103967, CurrentID=0, Status='live', Symbol='Dnm3', Aliases='9630020E24Rik, AW061159, B230343F03Rik, C530045C17, mKIAA0820', description='dynamin 3', other_designations='Dyna III|Dynamin III|t-dynamin|testicular dynamin', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161982450, end_position_on_the_genomic_accession=162478323, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  235040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235040, CurrentID=0, Status='live', Symbol='Atg4d', Aliases='9830134P12Rik, APG4-D, Apg4dl, Autl4, Atg4d', description='autophagy related 4D, cysteine peptidase', other_designations='APG4 (ATG4) autophagy-related homolog D|AUT-like 4 cysteine endopeptidase|AUT-like 4, cysteine endopeptidase|autophagin-4|autophagy-related 4D|autophagy-related cysteine endopeptidase 4|autophagy-related protein 4 homolog D|cysteine protease involved in autophagy APG4-D', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21265285, end_position_on_the_genomic_accession=21274837, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  235041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235041, CurrentID=0, Status='live', Symbol='Kank2', Aliases='AI504612, Ankrd25, BC010245', description='KN motif and ankyrin repeat domains 2', other_designations='ankyrin repeat domain 25|ankyrin repeat domain-containing protein 25', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21766773, end_position_on_the_genomic_accession=21798546, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  235043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235043, CurrentID=0, Status='live', Symbol='Tmem205', Aliases='', description='transmembrane protein 205', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21921009, end_position_on_the_genomic_accession=21927535, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  235044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235044, CurrentID=0, Status='live', Symbol='Plppr2', Aliases='Lppr2, PRG-4', description='phospholipid phosphatase related 2', other_designations='plasticity-related gene 4 protein', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21936422, end_position_on_the_genomic_accession=21948907, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  628261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628261, CurrentID=0, Status='live', Symbol='Bb1', Aliases='', description='Ig kappa chain variable segment', other_designations='', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.3', start_position_on_the_genomic_accession=68246582, end_position_on_the_genomic_accession=68247613, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  235047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235047, CurrentID=0, Status='live', Symbol='Zfp809', Aliases='BB114266, C330026E23', description='zinc finger protein 809', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22225697, end_position_on_the_genomic_accession=22253855, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  235048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235048, CurrentID=0, Status='live', Symbol='Zfp599', Aliases='6720416P20', description='zinc finger protein 599', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22247430, end_position_on_the_genomic_accession=22259895, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  235050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235050, CurrentID=0, Status='live', Symbol='Zfp810', Aliases='BC005471', description='zinc finger protein 810', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=22276737, end_position_on_the_genomic_accession=22307640, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  103979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103979, CurrentID=0, Status='live', Symbol='Alph1', Aliases='Aph-1, Aph1', description='alphaprotein 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  628268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628268, CurrentID=0, Status='live', Symbol='Igkv2-109', Aliases='ENSMUSG00000060986, ENSMUSG00000076521, Gm10878, he24', description='immunoglobulin kappa variable 2-109', other_designations='immunoglobulin kappa chain variable 2-109', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68302439, end_position_on_the_genomic_accession=68303158, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  103985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103985, CurrentID=0, Status='live', Symbol='Skn1', Aliases='Sk-1, Skn-1', description='skin antigen 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103986, CurrentID=0, Status='live', Symbol='Wt', Aliases='', description='waltzer-type', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103988, CurrentID=0, Status='live', Symbol='Gck', Aliases='GLK, Gk, Gls006, HK4, HKIV, HXKP, Hlb62, MODY2', description='glucokinase', other_designations='HK IV|hexokinase 4|hexokinase type IV|hexokinase-4|hexokinase-D', map_location='11 3.88 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5900816, end_position_on_the_genomic_accession=5950081, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  103993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103993, CurrentID=0, Status='live', Symbol='Bsp', Aliases='', description='black spleen', other_designations='splenic lipofuscinosis', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235067, CurrentID=0, Status='live', Symbol='Cypt4', Aliases='1700009B20Rik, Ckt1r2', description='cysteine-rich perinuclear theca 4', other_designations='basic protein CKT1R2|cysteine-rich perinuclear theca protein 4', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=24625185, end_position_on_the_genomic_accession=24625828, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  103996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103996, CurrentID=0, Status='live', Symbol='Lse', Aliases='', description='low set ears', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  103997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103997, CurrentID=0, Status='live', Symbol='Dpa', Aliases='', description='D-phenylalanine', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235072, CurrentID=0, Status='live', Symbol='Sept7', Aliases='Cdc10, E430034N22', description='septin 7', other_designations='CDC10 protein homolog|cell division cycle 10 homolog', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=25251090, end_position_on_the_genomic_accession=25308571, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  104001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104001, CurrentID=0, Status='live', Symbol='Rtn1', Aliases='Nsp', description='reticulon 1', other_designations='neuroendocrine-specific protein', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72211749, end_position_on_the_genomic_accession=72409054, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  104002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104002, CurrentID=0, Status='live', Symbol='Ctsq', Aliases='1600010J02Rik', description='cathepsin Q', other_designations='', map_location='13 32.65 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61035038, end_position_on_the_genomic_accession=61040597, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  104009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104009, CurrentID=0, Status='live', Symbol='Qsox1', Aliases='1300003H02Rik, QSOX, Qscn6, SOx', description='quiescin Q6 sulfhydryl oxidase 1', other_designations='skin sulfhydryl oxidase', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=155778155, end_position_on_the_genomic_accession=155812899, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  104010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104010, CurrentID=0, Status='live', Symbol='Cdh22', Aliases='', description='cadherin 22', other_designations='PB-cadherin|pituitary and brain cadherin', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165111501, end_position_on_the_genomic_accession=165234879, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  104012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104012, CurrentID=0, Status='live', Symbol='Axd', Aliases='', description='axial defects', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235086, CurrentID=0, Status='live', Symbol='Igsf9b', Aliases='Gm508, mKIAA1030', description='immunoglobulin superfamily, member 9B', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=27299104, end_position_on_the_genomic_accession=27357549, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  104015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104015, CurrentID=0, Status='live', Symbol='Synj1', Aliases='A930006D20Rik, AA675315, mKIAA0910', description='synaptojanin 1', other_designations='synaptic inositol 1,4,5-trisphosphate 5-phosphatase 1|synaptic inositol-1,4,5-trisphosphate 5-phosphatase 1', map_location='16 C3-C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=90936092, end_position_on_the_genomic_accession=91012137, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  104017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104017, CurrentID=0, Status='live', Symbol='Em', Aliases='', description='Emory cataract', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104027, CurrentID=0, Status='live', Symbol='Synpo', Aliases='9030217H17Rik, 9130229N11, 9330140I15Rik, AW046661', description='synaptopodin', other_designations='', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60593981, end_position_on_the_genomic_accession=60660174, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  235106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235106, CurrentID=0, Status='live', Symbol='Ntm', Aliases='6230410L23Rik, B230210G24Rik, Hnt, Igdcc2, R75390', description='neurotrimin', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=28994923, end_position_on_the_genomic_accession=29963486, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  628324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628324, CurrentID=0, Status='live', Symbol='S100a2', Aliases='CaN19, EG628324, S100L', description='S100 calcium binding protein A2', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90590247, end_position_on_the_genomic_accession=90591508, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  104054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104054, CurrentID=0, Status='live', Symbol='Ga', Aliases='', description='greying with age', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104055, CurrentID=0, Status='live', Symbol='Hst3', Aliases='Hst-3', description='hybrid sterility 3', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235130, CurrentID=0, Status='live', Symbol='Adamts15', Aliases='', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 15', other_designations='ADAM-TS 15|ADAM-TS15|ADAMTS-15|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 15', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=30899155, end_position_on_the_genomic_accession=30922452, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  235132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235132, CurrentID=0, Status='live', Symbol='Zbtb44', Aliases='6030404E16Rik, Btbd15, E230011F09', description='zinc finger and BTB domain containing 44', other_designations='BTB (POZ) domain containing 15|BTB/POZ domain-containing protein 15', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=31030394, end_position_on_the_genomic_accession=31075885, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  61034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61034, CurrentID=0, Status='live', Symbol='D8Mit155', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 155', other_designations='', map_location='8 2.14 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235134, CurrentID=0, Status='live', Symbol='Nfrkb', Aliases='A530090G11Rik', description='nuclear factor related to kappa B binding protein', other_designations='DNA-binding protein R kappa-B', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=31386161, end_position_on_the_genomic_accession=31421333, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  235135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235135, CurrentID=0, Status='live', Symbol='Tmem45b', Aliases='BC018222', description='transmembrane protein 45b', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=31426196, end_position_on_the_genomic_accession=31464238, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  104067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104067, CurrentID=0, Status='live', Symbol='Pol1', Aliases='Pol-1', description='viral polymerase 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104069, CurrentID=0, Status='live', Symbol='Sncb', Aliases='AI838531, betaSYN', description='synuclein, beta', other_designations='', map_location='13 29.28 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54758860, end_position_on_the_genomic_accession=54766440, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  104070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104070, CurrentID=0, Status='live', Symbol='Gdr1', Aliases='Gdr-1', description='G6PD regulator 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104072, CurrentID=0, Status='live', Symbol='Tol1', Aliases='Tol-1', description='tolerance to BCG', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104079, CurrentID=0, Status='live', Symbol='Nxph3', Aliases='', description='neurexophilin 3', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95509846, end_position_on_the_genomic_accession=95514565, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104080, CurrentID=0, Status='live', Symbol='Nxph4', Aliases='1110036M10Rik, AI851051', description='neurexophilin 4', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127525473, end_position_on_the_genomic_accession=127534559, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104082, CurrentID=0, Status='live', Symbol='Wdr7', Aliases='AI462727, TRAG, mKIAA0541', description='WD repeat domain 7', other_designations='TGF-beta resistance associated|TGF-beta resistance-associated protein TRAG', map_location='18 D1-E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=63708695, end_position_on_the_genomic_accession=63989760, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  104086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104086, CurrentID=0, Status='live', Symbol='Cyp27a1', Aliases='1300013A03Rik, Cyp27', description='cytochrome P450, family 27, subfamily a, polypeptide 1', other_designations='5-beta-cholestane-3-alpha,7-alpha,12-alpha-triol 27-hydroxylase|cholesterol 27 hydroxylase|cytochrome P-450C27/25|cytochrome P450, 27a1|sterol 27-hydroxylase|vitamin D(3) 25-hydroxylase', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74713146, end_position_on_the_genomic_accession=74737896, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  104089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104089, CurrentID=0, Status='live', Symbol='Hct', Aliases='', description='hair constriction', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104090, CurrentID=0, Status='live', Symbol='Hxa', Aliases='H-X, H-hix', description='histocompatibility X', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104092, CurrentID=0, Status='live', Symbol='Ir3', Aliases='Ir-3', description='immune response 3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104093, CurrentID=0, Status='live', Symbol='Ir4', Aliases='Ir-4', description='immune response 4', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235169, CurrentID=0, Status='live', Symbol='Foxred1', Aliases='BC024806, TEG-23, Tex23', description='FAD-dependent oxidoreductase domain containing 1', other_designations='testis expressed gene 23', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35204208, end_position_on_the_genomic_accession=35211055, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  104098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104098, CurrentID=0, Status='live', Symbol='Xpl', Aliases='', description='X-linked polydactyly', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104099, CurrentID=0, Status='live', Symbol='Itga9', Aliases='(alpha)9, 2610002H11Rik, 6720458D17Rik, AI461869, D130073C02, D9Ertd428e', description='integrin alpha 9', other_designations='', map_location='9 70.32 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=118606709, end_position_on_the_genomic_accession=118901003, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  104103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104103, CurrentID=0, Status='live', Symbol='Airn', Aliases='2810051F02Rik, 2810434M15Rik, AI256653, AI597500, AW049873, Air, B930018I07Rik, D17Ertd663e, IGF2RAS', description='antisense Igf2r RNA', other_designations='antisense Igf2r RNA noncoding', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12741311, end_position_on_the_genomic_accession=12859884, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  235180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235180, CurrentID=0, Status='live', Symbol='Fez1', Aliases='UNC-76, UNC76', description='fasciculation and elongation protein zeta 1 (zygin I)', other_designations='zygin I|zygin-1', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=36821944, end_position_on_the_genomic_accession=36878926, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  104110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104110, CurrentID=0, Status='live', Symbol='Adcy4', Aliases='mKIAA4004', description='adenylate cyclase 4', other_designations='ATP pyrophosphate-lyase 4|adenylate cyclase type IV|adenylyl cyclase 4', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55769056, end_position_on_the_genomic_accession=55784095, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  104111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104111, CurrentID=0, Status='live', Symbol='Adcy3', Aliases='AC3, mKIAA0511', description='adenylate cyclase 3', other_designations='AC-III|ATP pyrophosphate-lyase 3|adenylate cyclase type III|adenylate cyclase, olfactive type|adenylyl cyclase 3', map_location='12 A-B', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4133087, end_position_on_the_genomic_accession=4213525, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  235184 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235184, CurrentID=0, Status='live', Symbol='Msantd2', Aliases='2810450G17Rik, 9530092B10Rik', description='Myb/SANT-like DNA-binding domain containing 2', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37487773, end_position_on_the_genomic_accession=37526113, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  104117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104117, CurrentID=0, Status='live', Symbol='H35', Aliases='H-35', description='histocompatibility 35', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104118, CurrentID=0, Status='live', Symbol='H37', Aliases='H-37', description='histocompatibility 37', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104124, CurrentID=0, Status='live', Symbol='Meta', Aliases='', description='meth A tumor antigen', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104126, CurrentID=0, Status='live', Symbol='Lith3', Aliases='', description='lithogenic gene 3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104127 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104127, CurrentID=0, Status='live', Symbol='Lith2', Aliases='', description='lithogenic gene 2', other_designations='', map_location='19 44.88 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104130, CurrentID=0, Status='live', Symbol='Ndufb11', Aliases='D5Bwg0566e, D5Bwg0577e, NP15.6, Np15', description='NADH dehydrogenase (ubiquinone) 1 beta subcomplex, 11', other_designations='CI-ESSS|NADH-ubiquinone oxidoreductase ESSS subunit|complex I-ESSS|neuronal protein 15.6|nuclear protein 15.6|p15.6', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20615326, end_position_on_the_genomic_accession=20617564, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  279499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=279499, CurrentID=0, Status='live', Symbol='Kctd19', Aliases='4922504H04Rik', description='potassium channel tetramerisation domain containing 19', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105382350, end_position_on_the_genomic_accession=105413502, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  104137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104137, CurrentID=0, Status='live', Symbol='H40', Aliases='H(Igh), H-40', description='histocompatibility 40', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104142, CurrentID=0, Status='live', Symbol='Len2', Aliases='Len-2', description='eye lens protein 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104145, CurrentID=0, Status='live', Symbol='Pp', Aliases='', description='passive performance', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104156, CurrentID=0, Status='live', Symbol='Etv5', Aliases='1110005E01Rik, 8430401F14Rik, ERM', description='ets variant 5', other_designations='ets variant gene 5', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22381313, end_position_on_the_genomic_accession=22439570, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  104158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104158, CurrentID=0, Status='live', Symbol='Ces1d', Aliases='Ces3, TGH', description='carboxylesterase 1D', other_designations='FAEE synthase|carboxylesterase 3|fatty acid ethyl ester synthase|triacylglycerol hydrolase', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93166072, end_position_on_the_genomic_accession=93197804, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  104159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104159, CurrentID=0, Status='live', Symbol='Epa1', Aliases='Epa-1', description='epidermal antigen 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104164, CurrentID=0, Status='live', Symbol='Lyx3', Aliases='Lyx-3', description='lymphocyte antigen X 3', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104167, CurrentID=0, Status='live', Symbol='Pd', Aliases='', description='pyrimidine degrading', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104168, CurrentID=0, Status='live', Symbol='Rmp1', Aliases='Rmp-1', description='resistance to mousepox', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104174, CurrentID=0, Status='live', Symbol='Gldc', Aliases='D030049L12Rik, D19Wsu57e', description='glycine decarboxylase', other_designations='glycine cleavage system P protein|glycine dehydrogenase (aminomethyl-transferring)|glycine dehydrogenase [decarboxylating], mitochondrial', map_location='19 C', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=30098441, end_position_on_the_genomic_accession=30175441, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  104175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104175, CurrentID=0, Status='live', Symbol='Sbk1', Aliases='Sbk', description='SH3-binding kinase 1', other_designations='', map_location='7 F1-F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126272619, end_position_on_the_genomic_accession=126294999, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  235248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235248, CurrentID=0, Status='live', Symbol='Olfr952', Aliases='MOR171-21', description='olfactory receptor 952', other_designations='GA_x6K02T2PVTD-33124064-33123120|olfactory receptor MOR171-21', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39426125, end_position_on_the_genomic_accession=39427069, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  104177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104177, CurrentID=0, Status='live', Symbol='Lyb7', Aliases='Lyb-7', description='B-lymphocyte antigen 7', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104183, CurrentID=0, Status='live', Symbol='Chil4', Aliases='Chi3l4, Ym2', description='chitinase-like 4', other_designations='chitinase 3-like 4, Chil4|chitinase-3-like protein 4|secreted protein Ym2', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106201491, end_position_on_the_genomic_accession=106219479, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  235256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235256, CurrentID=0, Status='live', Symbol='Olfr149', Aliases='M31, MOR224-8', description='olfactory receptor 149', other_designations='GA_x6K02T2PVTD-33400306-33399371|odorant receptor M31|olfactory receptor 224-8|olfactory receptor 7G|olfactory receptor GA_x5J8B7W3WPH-39574-39485|olfactory receptor MOR224-8', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39701832, end_position_on_the_genomic_accession=39702767, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  104187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104187, CurrentID=0, Status='live', Symbol='Lm1', Aliases='Lm-1', description='lymphomyeloid antigen 1', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104193, CurrentID=0, Status='live', Symbol='Ce1', Aliases='Ce-1', description='catalase activity, liver', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104194, CurrentID=0, Status='live', Symbol='Ea2', Aliases='Ea-2, H-14, R, R-Z, RHO', description='erythrocyte antigen 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104207, CurrentID=0, Status='live', Symbol='Mcpt3', Aliases='Mcp-3', description='mast cell protease 3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235281, CurrentID=0, Status='live', Symbol='Scn3b', Aliases='1110001K16Rik, 4833414B02Rik, Scnb3', description='sodium channel, voltage-gated, type III, beta', other_designations='brain and heart sodium channel beta 3 subunit', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40269187, end_position_on_the_genomic_accession=40291618, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  235283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235283, CurrentID=0, Status='live', Symbol='Gramd1b', Aliases='3222402H23, A930008A22Rik, AI593249, mKIAA1201', description='GRAM domain containing 1B', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40293233, end_position_on_the_genomic_accession=40531499, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  104215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104215, CurrentID=0, Status='live', Symbol='Rhoq', Aliases='Arhq, TC10A, Tc10', description='ras homolog gene family, member Q', other_designations='ras-like protein TC10|ras-related GTP-binding protein TC10|small GTPase Tc10', map_location='17 56.9 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=86963111, end_position_on_the_genomic_accession=87000069, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  235293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235293, CurrentID=0, Status='live', Symbol='Sc5d', Aliases='A830037K02, A830073K23Rikl, Sc5d', description='sterol-C5-desaturase (fungal ERG3, delta-5-desaturase) homolog (S. cerevisae)', other_designations='C-5 sterol desaturase|delta(7)-sterol 5-desaturase|delta(7)-sterol C5(6)-desaturase|lathosterol 5-desaturase', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=42251592, end_position_on_the_genomic_accession=42264300, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  235300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235300, CurrentID=0, Status='live', Symbol='Tmem136', Aliases='AA396355, Gm184', description='transmembrane protein 136', other_designations='', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=43108650, end_position_on_the_genomic_accession=43116714, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104231, CurrentID=0, Status='live', Symbol='Ly76', Aliases='TER-119, Ter119', description='lymphocyte antigen 76', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104232, CurrentID=0, Status='live', Symbol='Ly79', Aliases='', description='lymphocyte antigen 79', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100505386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100505386, CurrentID=0, Status='live', Symbol='Gm21949', Aliases='Iqcj-schip1, Schip-1-iqcj, Schip1', description='predicted gene, 21949', other_designations='IQ motif containing J-schwannomin interacting protein 1 readthrough|schwannomin interacting protein 1 isoform IQCJ|schwannomin interacting protein 1 isoform IQCJs', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=67892220, end_position_on_the_genomic_accession=68626482, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  235312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235312, CurrentID=0, Status='live', Symbol='C1qtnf5', Aliases='Ctrp5', description='C1q and tumor necrosis factor related protein 5', other_designations='', map_location='9 24.62 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44101770, end_position_on_the_genomic_accession=44109187, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  235315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235315, CurrentID=0, Status='live', Symbol='Rnf214', Aliases='D130054N24Rik', description='ring finger protein 214', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45863532, end_position_on_the_genomic_accession=45906906, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  104245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104245, CurrentID=0, Status='live', Symbol='Slc6a5', Aliases='Glyt2, prestin', description='solute carrier family 6 (neurotransmitter transporter, glycine), member 5', other_designations='glyT-2|glycine transporter 2|slc6a5|solute carrier family 6 member 5', map_location='7 B5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=49895702, end_position_on_the_genomic_accession=49963864, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  104248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104248, CurrentID=0, Status='live', Symbol='Cabin1', Aliases='A330070M20Rik, Cain, Ppp3in', description='calcineurin binding protein 1', other_designations='calcineurin inhibitor|protein phosphatase 3 inhibitor', map_location='10 38.56 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75646110, end_position_on_the_genomic_accession=75764358, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  235323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235323, CurrentID=0, Status='live', Symbol='Usp28', Aliases='9830148O20Rik, AU022237, mKIAA1515', description='ubiquitin specific peptidase 28', other_designations='deubiquitinating enzyme 28|ubiquitin specific protease 28|ubiquitin thioesterase 28|ubiquitin thiolesterase 28|ubiquitin-specific-processing protease 28', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48985331, end_position_on_the_genomic_accession=49042517, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  104252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104252, CurrentID=0, Status='live', Symbol='Cdc42ep2', Aliases='1110008C05Rik, Borg1, Cep2, D19Bwg1013e', description='CDC42 effector protein (Rho GTPase binding) 2', other_designations='binder of Rho GTPases 1', map_location='19 4.34 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5917556, end_position_on_the_genomic_accession=5924816, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104253, CurrentID=0, Status='live', Symbol='Ad', Aliases='', description='adult obesity and diabetes', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104258, CurrentID=0, Status='live', Symbol='Lyb3', Aliases='Lyb-3', description='B-lymphocyte antigen 3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104259, CurrentID=0, Status='live', Symbol='Lyb5', Aliases='Lyb-5', description='B-lymphocyte antigen 5', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104263, CurrentID=0, Status='live', Symbol='Kdm3a', Aliases='1700105C21Rik, C230043E16Rik, JHDM2a, Jmjd1, Jmjd1a, KDM2A, TGSA, Tsga', description='lysine (K)-specific demethylase 3A', other_designations='jmjC domain-containing histone demethylation protein 2A|jumonji domain-containing protein 1A|testis specific gene A', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=71588970, end_position_on_the_genomic_accession=71633297, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  104266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104266, CurrentID=0, Status='live', Symbol='Tsp2', Aliases='MTS', description='tumor suppressor region 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235339, CurrentID=0, Status='live', Symbol='Dlat', Aliases='6332404G05Rik, DLTA, PDC-E2', description='dihydrolipoamide S-acetyltransferase (E2 component of pyruvate dehydrogenase complex)', other_designations='PDCE2|dihydrolipoamide acetyltransferase component of pyruvate dehydrogenase complex|pyruvate dehydrogenase complex component E2', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50634633, end_position_on_the_genomic_accession=50659780, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  104271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104271, CurrentID=0, Status='live', Symbol='Tex15', Aliases='2210014E14Rik, AL022622, AU022940', description='testis expressed gene 15', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33516684, end_position_on_the_genomic_accession=33585585, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  235344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235344, CurrentID=0, Status='live', Symbol='Sik2', Aliases='G630080D20Rik, Snf1lk2', description='salt inducible kinase 2', other_designations='SIK-2|SNF1-like kinase 2|salt induceable kinase 2|salt-inducible kinase 2|salt-inducible protein kinase 2|serine/threonine-protein kinase SNF1-like kinase 2', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50889644, end_position_on_the_genomic_accession=51009073, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  235345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235345, CurrentID=0, Status='live', Symbol='4833427G06Rik', Aliases='', description='RIKEN cDNA 4833427G06 gene', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51081313, end_position_on_the_genomic_accession=51102078, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104277, CurrentID=0, Status='live', Symbol='Glud-ps', Aliases='EG545999, Glud-2, Glud2, Gm5902', description='glutamate dehydrogenase, pseudogene', other_designations='glutamate dehydrogenase 2', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=125075348, end_position_on_the_genomic_accession=125076497, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  104290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104290, CurrentID=0, Status='live', Symbol='Pctr2', Aliases='Pcts3', description='plasmacytoma resistance 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104297, CurrentID=0, Status='live', Symbol='Igis1', Aliases='', description='immunoglobulin isotype 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104299, CurrentID=0, Status='live', Symbol='Prt1', Aliases='Prt-1', description='pancreatic proteinase 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104300, CurrentID=0, Status='live', Symbol='Prt3', Aliases='Prt-3', description='pancreatic proteinase 3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104303, CurrentID=0, Status='live', Symbol='Arl1', Aliases='2310008D22Rik, AU019749', description='ADP-ribosylation factor-like 1', other_designations='ADP-ribosylation-like 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88731414, end_position_on_the_genomic_accession=88743202, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  235379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235379, CurrentID=0, Status='live', Symbol='Gldn', Aliases='CRG-L2, Clom, Colm, Crgl2, Crlg2', description='gliomedin', other_designations='cancer related gene - liver 2|cancer-related gene liver 2 protein|collomin|olfactomedin-related protein', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54286486, end_position_on_the_genomic_accession=54341777, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  235380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235380, CurrentID=0, Status='live', Symbol='Dmxl2', Aliases='6330586A12, 6430411K14Rik, E130119P06Rik, mKIAA0856', description='Dmx-like 2', other_designations='rabconnectin-3', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54365158, end_position_on_the_genomic_accession=54501697, orientation='minus', exon_count=44, OMIM='', no_hdr0=''),
  235386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235386, CurrentID=0, Status='live', Symbol='Hykk', Aliases='Agphd1, C630028N24Rik', description='hydroxylysine kinase 1', other_designations='5-hydroxy-L-lysine kinase|aminoglycoside phosphotransferase domain containing 1|aminoglycoside phosphotransferase domain-containing protein 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54917290, end_position_on_the_genomic_accession=54949924, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  104318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104318, CurrentID=0, Status='live', Symbol='Csnk1d', Aliases='1200006A05Rik, AA409348, D930010H05Rik', description='casein kinase 1, delta', other_designations='CKI-delta|CKId|tau-protein kinase CSNK1D', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120958010, end_position_on_the_genomic_accession=120991333, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  104320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104320, CurrentID=0, Status='live', Symbol='Cfp1', Aliases='Cfp-1', description='cerebellar folial pattern 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104323, CurrentID=0, Status='live', Symbol='Emi1', Aliases='Emi-1', description='proviral insertion Emi 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104324, CurrentID=0, Status='live', Symbol='Hrs2', Aliases='Hrs-2', description='mouse minor satellite DNA 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235402 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235402, CurrentID=0, Status='live', Symbol='Lingo1', Aliases='4930471K13Rik, AV148400, LERN1, LINGO-1, Lrrn6a, UNQ201', description='leucine rich repeat and Ig domain containing 1', other_designations='leucine rich repeat neuronal 6A|leucine-rich repeat neuronal 6A|leucine-rich repeat neuronal protein 1|leucine-rich repeat neuronal protein 6A', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56618460, end_position_on_the_genomic_accession=56796784, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  235406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235406, CurrentID=0, Status='live', Symbol='Snx33', Aliases='E130307J07Rik, Sh3px3', description='sorting nexin 33', other_designations='SH3 and PX domain containing 3|SH3 and PX domain-containing protein 3', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56917191, end_position_on_the_genomic_accession=56928442, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104337, CurrentID=0, Status='live', Symbol='Gct1', Aliases='Gct', description='granulosa cell tumorigenesis 1', other_designations='', map_location='4 78.17 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  629059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629059, CurrentID=0, Status='live', Symbol='Fam124a', Aliases='EG629059', description='family with sequence similarity 124, member A', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62555737, end_position_on_the_genomic_accession=62608486, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  235415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235415, CurrentID=0, Status='live', Symbol='Cplx3', Aliases='BC016632, CpxIII, ERGIC-53L, ERGL, Lamn1l, Lman1l', description='complexin 3', other_designations='CPX III|complexin III|lectin, mannose-binding, 1 like', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57599992, end_position_on_the_genomic_accession=57606281, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  235416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235416, CurrentID=0, Status='live', Symbol='Lman1l', Aliases='BC020188, slamp', description='lectin, mannose-binding 1 like', other_designations='ERGIC53-like protein|LMAN1-like protein|lectin mannose-binding 1-like|sublingual acinar membrane protein', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57606717, end_position_on_the_genomic_accession=57621419, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  104346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104346, CurrentID=0, Status='live', Symbol='Gas8', Aliases='Gas11', description='growth arrest specific 8', other_designations='GAS-11|GAS-8|beta growth arrest specific protein 11|growth arrest specific 11|growth arrest specific protein 11|growth arrest-specific protein 11', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123518835, end_position_on_the_genomic_accession=123536657, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  104348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104348, CurrentID=0, Status='live', Symbol='Zfp120', Aliases='AI315103, E030042N05Rik, F80, MZF31, Zfp844', description='zinc finger protein 120', other_designations='zfp120|zinc finger protein 31|zinc finger protein 844', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150114407, end_position_on_the_genomic_accession=150136678, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104349, CurrentID=0, Status='live', Symbol='Zfp119a', Aliases='AL024077, Mzf13, Zfp119', description='zinc finger protein 119a', other_designations='KRAB zinc finger protein|zinc finger protein 119', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55864886, end_position_on_the_genomic_accession=55879685, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104353, CurrentID=0, Status='live', Symbol='Rec2', Aliases='Rec-2', description='ecotropic MuLV M813 receptor', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235431, CurrentID=0, Status='live', Symbol='Coro2b', Aliases='CLIPINC, E130012P22Rik', description='coronin, actin binding protein, 2B', other_designations='E130012P22', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=62419490, end_position_on_the_genomic_accession=62537044, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  104360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104360, CurrentID=0, Status='live', Symbol='Isl2', Aliases='3110001N10Rik', description='insulin related protein 2 (islet 2)', other_designations='islet 2', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55538677, end_position_on_the_genomic_accession=55547667, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  104362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104362, CurrentID=0, Status='live', Symbol='Meig1', Aliases='MLZ-278, Meg1', description='meiosis expressed gene 1', other_designations='protein expressed in male leptotene and zygotene spermatocytes 278', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=3409043, end_position_on_the_genomic_accession=3422648, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  235435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235435, CurrentID=0, Status='live', Symbol='Lctl', Aliases='E130104I05Rik, KLPH', description='lactase-like', other_designations='Klotho-LPH related|klotho/lactase-phlorizin hydrolase-related protein', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=64117123, end_position_on_the_genomic_accession=64138118, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  235439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235439, CurrentID=0, Status='live', Symbol='Herc1', Aliases='2810449H11Rik, B230218H07, D130015N03Rik, tbl', description='HECT and RLD domain containing E3 ubiquitin protein ligase family member 1', other_designations='hect (homologous to the E6-AP (UBE3A) carboxyl terminus) domain and RCC1 (CHC1)-like domain (RLD) 1', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66350450, end_position_on_the_genomic_accession=66508775, orientation='plus', exon_count=78, OMIM='', no_hdr0=''),
  235441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235441, CurrentID=0, Status='live', Symbol='Usp3', Aliases='AA409661, BC017156', description='ubiquitin specific peptidase 3', other_designations='deubiquitinating enzyme 3|ubiquitin specific protease 3|ubiquitin thioesterase 3|ubiquitin thiolesterase 3|ubiquitin-specific-processing protease 3', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66514639, end_position_on_the_genomic_accession=66593115, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  235442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235442, CurrentID=0, Status='live', Symbol='Rab8b', Aliases='5930437D16, D330025I23Rik', description='RAB8B, member RAS oncogene family', other_designations='RAB44, member RAS oncogene family', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66843664, end_position_on_the_genomic_accession=66919705, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  104375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104375, CurrentID=0, Status='live', Symbol='Ezg', Aliases='', description='extent of zona glomerulosa', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104382, CurrentID=0, Status='live', Symbol='Barhl2', Aliases='E130309B19Rik, MBH1', description='BarH-like 2 (Drosophila)', other_designations='BarH-class homeodomain transcription factor|bar-class homeodomain protein MBH1|homeobox protein B-H1', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=106452523, end_position_on_the_genomic_accession=106458166, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  104383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104383, CurrentID=0, Status='live', Symbol='Rcor2', Aliases='1A13, AW122124, CoREST, Rcor, Rcor1', description='REST corepressor 2', other_designations='M-CoREST|RE1-silencing transcription factor (REST) co-repressor 1|REST co-repressor 2|co-repressor of Rest', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7267330, end_position_on_the_genomic_accession=7275225, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  104384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104384, CurrentID=0, Status='live', Symbol='Rhox9', Aliases='1600026O01Rik, Gpbox, Psx2', description='reproductive homeobox 9', other_designations='placenta specific homeobox 2', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37899097, end_position_on_the_genomic_accession=37901770, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104385, CurrentID=0, Status='live', Symbol='Asc1', Aliases='Asc-1', description='anterior suture cataract 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  235459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235459, CurrentID=0, Status='live', Symbol='Gtf2a2', Aliases='TFIIA-gamma, TfIIg, Tfiia2', description='general transcription factor II A, 2', other_designations='general transcription factor IIA, 2 (12kD subunit)|transcription initiation factor IIA gamma chain', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70011851, end_position_on_the_genomic_accession=70022866, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  235461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235461, CurrentID=0, Status='live', Symbol='Fam63b', Aliases='AI465301, B230380D07Rik, mKIAA1164', description='family with sequence similarity 63, member B', other_designations='', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=70599014, end_position_on_the_genomic_accession=70657296, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  104394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104394, CurrentID=0, Status='live', Symbol='E2f4', Aliases='2010111M04Rik, AI427446', description='E2F transcription factor 4', other_designations='E2F-4', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105297663, end_position_on_the_genomic_accession=105305370, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  235469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235469, CurrentID=0, Status='live', Symbol='Zfp280d', Aliases='A930005F02Rik, BC027163, Suhw4, ZNF634, Znf280d', description='zinc finger protein 280D', other_designations='suppressor of hairy wing homolog 4', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72274893, end_position_on_the_genomic_accession=72363779, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  235472 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235472, CurrentID=0, Status='live', Symbol='Prtg', Aliases='A230098A12Rik, Igdcc5', description='protogenin', other_designations='protein Shen-Dan|protogenin homolog (Gallus gallus)|prtg|shen-dan', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=72807274, end_position_on_the_genomic_accession=72917307, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  104401 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104401, CurrentID=0, Status='live', Symbol='Pcnxl3', Aliases='mKIAA4073', description='pecanex-like 3 (Drosophila)', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5664635, end_position_on_the_genomic_accession=5688908, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  104408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104408, CurrentID=0, Status='live', Symbol='Hsp84-2', Aliases='84kDa, Hsp90, hsp2', description='heat shock protein, 2', other_designations='heat shock protein, 84 kDa 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104416, CurrentID=0, Status='live', Symbol='Bap1', Aliases='2300006C11Rik, AA989761, AW553466, mKIAA0272, uch-x4', description='Brca1 associated protein 1', other_designations='BRCA1-associated protein 1|ubiquitin C-terminal hydrolase X4', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31251455, end_position_on_the_genomic_accession=31259930, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  628705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628705, CurrentID=0, Status='live', Symbol='Phf11c', Aliases='EG628705, Gm6907, Phf11', description='PHD finger protein 11C', other_designations='PHD finger protein 11', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59380833, end_position_on_the_genomic_accession=59440890, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  104418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104418, CurrentID=0, Status='live', Symbol='Dgkz', Aliases='E130307B02Rik, ENSMUSG00000075035, F730209L11Rik, mDGK[z]', description='diacylglycerol kinase zeta', other_designations='DAG kinase zeta|DGK-zeta|diglyceride kinase zeta', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91932822, end_position_on_the_genomic_accession=91976443, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  628709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628709, CurrentID=0, Status='live', Symbol='Gm10324', Aliases='ENSMUSG00000071302', description='predicted gene 10324', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=66113453, end_position_on_the_genomic_accession=66122836, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  235497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235497, CurrentID=0, Status='live', Symbol='Leo1', Aliases='Gm185', description='Leo1, Paf1/RNA polymerase II complex component', other_designations='Leo1, Paf1/RNA polymerase II complex component, homolog', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=75441524, end_position_on_the_genomic_accession=75466432, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  192167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192167, CurrentID=0, Status='live', Symbol='Nlgn1', Aliases='6330415N05Rik, BB179718, NL1, Nlg1, mKIAA1070', description='neuroligin 1', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=25431841, end_position_on_the_genomic_accession=26332056, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  235504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235504, CurrentID=0, Status='live', Symbol='Slc17a5', Aliases='4631416G20Rik, 4732491M05, AST, ISSD, NSD, SD, SIALIN, SIASD, SLD', description='solute carrier family 17 (anion/sugar transporter), member 5', other_designations='H(+)/nitrate cotransporter|sodium/sialic acid cotransporter|vesicular H(+)/Aspartate-glutamate cotransporter', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78536487, end_position_on_the_genomic_accession=78588065, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  235505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235505, CurrentID=0, Status='live', Symbol='Cd109', Aliases='9930012E15Rik, AI480638, GARP', description='CD109 antigen', other_designations='GPI-anchored alpha 2 macroglobulin-related protein|GPI-anchored alpha-2 macroglobulin-related protein|Gov platelet alloantigens', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78615546, end_position_on_the_genomic_accession=78716260, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  104434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104434, CurrentID=0, Status='live', Symbol='Hsp84-3', Aliases='84kDa, Hsp90, hsp3', description='heat shock protein, 3', other_designations='heat shock protein, 84 kDa 3', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  497652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=497652, CurrentID=0, Status='live', Symbol='Acd', Aliases='', description='adrenocortical dysplasia', other_designations='', map_location='8 53.04 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105698157, end_position_on_the_genomic_accession=105701114, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  61097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61097, CurrentID=0, Status='live', Symbol='D1Mit488', Aliases='', description='DNA Segment, Chr 1 Massachusetts Institute of Technology 488', other_designations='', map_location='1 45.12 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  104443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104443, CurrentID=0, Status='live', Symbol='Npffr2', Aliases='Gpr74, HG31, NPFF2', description='neuropeptide FF receptor 2', other_designations='G protein-coupled receptor 74|G-protein coupled receptor 74|neuropeptide NPFF receptor', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=89526577, end_position_on_the_genomic_accession=89583914, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  104444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104444, CurrentID=0, Status='live', Symbol='Rexo2', Aliases='1810038D15Rik, AW107347, Sfn, Smfn', description='REX2, RNA exonuclease 2 homolog (S. cerevisiae)', other_designations='RNA exonuclease 2 homolog|small fragment nuclease', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48468512, end_position_on_the_genomic_accession=48480714, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  104445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104445, CurrentID=0, Status='live', Symbol='Cdc42ep1', Aliases='1810058K22Rik, AA980734, Borg5, CEP1, MSE55', description='CDC42 effector protein (Rho GTPase binding) 1', other_designations='binder of Rho GTPase 5|binder of Rho GTPases 5', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78842647, end_position_on_the_genomic_accession=78850902, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  235527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235527, CurrentID=0, Status='live', Symbol='Plscr4', Aliases='AV245873', description='phospholipid scramblase 4', other_designations='PL scramblase 4|ca(2+)-dependent phospholipid scramblase 4', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=92457372, end_position_on_the_genomic_accession=92492516, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  104457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104457, CurrentID=0, Status='live', Symbol='0610010K14Rik', Aliases='1110020A23Rik, AL033328, AU045833, Bap18', description='RIKEN cDNA 0610010K14 gene', other_designations='BPTF-associated protein of 18 kDa|MLL1/MLL complex subunit C17orf49 homolog', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70235204, end_position_on_the_genomic_accession=70237914, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104458, CurrentID=0, Status='live', Symbol='Rars', Aliases='2610011N19Rik, 2610037E21Rik, AL033339, AW985894', description='arginyl-tRNA synthetase', other_designations='argRS|arginyl-tRNA synthetase, cytoplasmic', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=35808381, end_position_on_the_genomic_accession=35834528, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  235533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235533, CurrentID=0, Status='live', Symbol='Gk5', Aliases='AV095337, C330018K18Rik, G630067D24Rik', description='glycerol kinase 5 (putative)', other_designations='ATP:glycerol 3-phosphotransferase 5|GK 5|glycerokinase 5', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96119344, end_position_on_the_genomic_accession=96184598, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  235534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235534, CurrentID=0, Status='live', Symbol='Pxylp1', Aliases='9430094M07Rik, Acpl2, BB177120, C130099A20Rik', description='2-phosphoxylose phosphatase 1', other_designations='acid phosphatase-like 2|acid phosphatase-like protein 2', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96823336, end_position_on_the_genomic_accession=96892680, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  235542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235542, CurrentID=0, Status='live', Symbol='Ppp2r3a', Aliases='3222402P14Rik, A730042E07', description="protein phosphatase 2, regulatory subunit B'', alpha", other_designations='', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=101104989, end_position_on_the_genomic_accession=101251832, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  104479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104479, CurrentID=0, Status='live', Symbol='Ccdc117', Aliases='1110004K02Rik, 1700026O03Rik, AU018638, AU042822, AV173073, BC018601', description='coiled-coil domain containing 117', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=5528887, end_position_on_the_genomic_accession=5542217, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  235559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235559, CurrentID=0, Status='live', Symbol='Topbp1', Aliases='1110031N14Rik, 2810429C13Rik, AI256758, D430026L04Rik, mKIAA0259', description='topoisomerase (DNA) II binding protein 1', other_designations='DNA topoisomerase II-beta-binding protein 1|DNA topoisomerase II-binding protein 1|topoisomerase (DNA) II beta binding protein', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103305169, end_position_on_the_genomic_accession=103350428, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  628779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628779, CurrentID=0, Status='live', Symbol='Hs3st4', Aliases='EG628779, Gm6915', description='heparan sulfate (glucosamine) 3-O-sulfotransferase 4', other_designations='', map_location='7 67.56 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=123983181, end_position_on_the_genomic_accession=124398989, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  235567 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235567, CurrentID=0, Status='live', Symbol='Dnajc13', Aliases='D030002L11Rik, Gm1124, RME-8, Rme8, mKIAA0678', description='DnaJ heat shock protein family (Hsp40) member C13', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 13|j-domain containing protein, required for endocytosis|receptor mediated endocytosis RME-8', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=104151597, end_position_on_the_genomic_accession=104263184, orientation='minus', exon_count=59, OMIM='', no_hdr0=''),
  235574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235574, CurrentID=0, Status='live', Symbol='Atp2c1', Aliases='1700121J11Rik, ATP2C1A, AW061228, BCPM, D930003G21Rik, HHD, SPCA, pmr1', description='ATPase, Ca++-sequestering', other_designations='ATP-dependent Ca(2+) pump PMR1|ATPase 2C1|ATPase, Ca++ transporting, type 2C, member 1', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105411362, end_position_on_the_genomic_accession=105521257, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  235582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235582, CurrentID=0, Status='live', Symbol='Glyctk', Aliases='6230410P16Rik, C430006L12', description='glycerate kinase', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106152860, end_position_on_the_genomic_accession=106161065, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  235584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235584, CurrentID=0, Status='live', Symbol='Dusp7', Aliases='AU015694, Mkpx, Pyst2', description='dual specificity phosphatase 7', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106368610, end_position_on_the_genomic_accession=106375724, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  235587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235587, CurrentID=0, Status='live', Symbol='Parp3', Aliases='A930002C11Rik, AW990611, Adprt3, Adprtl3, PARP-3, pADPRT-3', description='poly (ADP-ribose) polymerase family, member 3', other_designations='A DP-ribosyltransferase (NAD+; poly (ADP-ribose polymerase)-like 3|ADP-ribosyltransferase (NAD+, poly (ADP-ribose polymerase)-like 3|ADP-ribosyltransferase (NAD+; poly (ADP-ribose polymerase)-like 3|NAD+ ADP-ribosyltransferase 3|Poly[ADP-ribose] synthetase-3', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106470353, end_position_on_the_genomic_accession=106476944, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  235599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235599, CurrentID=0, Status='live', Symbol='6430571L13Rik', Aliases='', description='RIKEN cDNA 6430571L13 gene', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107340640, end_position_on_the_genomic_accession=107349683, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  235604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235604, CurrentID=0, Status='live', Symbol='Camkv', Aliases='1G5, BB074618, BC017634', description='CaM kinase-like vesicle-associated', other_designations='vesicle-associated calmodulin-binding protein', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107935851, end_position_on_the_genomic_accession=107949691, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  235606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235606, CurrentID=0, Status='live', Symbol='Apeh', Aliases='', description='acylpeptide hydrolase', other_designations='AARE|APH|N-acylaminoacyl peptide hydrolase|acyl-peptide hydrolase|acylaminoacyl-peptidase', map_location='9 59.07 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108085413, end_position_on_the_genomic_accession=108094548, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  235610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235610, CurrentID=0, Status='live', Symbol='Atrip', Aliases='6620401K05Rik', description='ATR interacting protein', other_designations='ATM and Rad3 related interacting protein|ATM and Rad3-related-interacting protein', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109059747, end_position_on_the_genomic_accession=109075580, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  235611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235611, CurrentID=0, Status='live', Symbol='Plxnb1', Aliases='2900002G15Rik, AU042020, mKIAA0407', description='plexin B1', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109093849, end_position_on_the_genomic_accession=109119915, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  235612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235612, CurrentID=0, Status='live', Symbol='Fbxw19', Aliases='E330023O18', description='F-box and WD-40 domain protein 19', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109478507, end_position_on_the_genomic_accession=109495863, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  192187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192187, CurrentID=0, Status='live', Symbol='Stab1', Aliases='FEEL-1, FELE-1, MFEEL-1, MS-1, STAB-1, mKIAA0246', description='stabilin 1', other_designations='fasciclin, EGF-like, laminin-type EGF-like and link domain-containing scavenger receptor 1|scavenger receptor FEEL-1', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31139017, end_position_on_the_genomic_accession=31168652, orientation='minus', exon_count=69, OMIM='', no_hdr0=''),
  235623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235623, CurrentID=0, Status='live', Symbol='Scap', Aliases='9530044G19, mKIAA0199', description='SREBF chaperone', other_designations='SREBP cleavage activating protein|SREBP cleavage-activating protein|Sreb cleavage-activating protein|scap', map_location='9 59.91 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110333288, end_position_on_the_genomic_accession=110384949, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  192188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192188, CurrentID=0, Status='live', Symbol='Stab2', Aliases='FEEL-2, FELL, MFEEL-2, STAB-2', description='stabilin 2', other_designations='fasciclin, EGF-like, laminin-type EGF-like and link domain-containing scavenger receptor 2|scavenger receptor FEEL-2|stab2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=86841210, end_position_on_the_genomic_accession=87007942, orientation='minus', exon_count=69, OMIM='', no_hdr0=''),
  235626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235626, CurrentID=0, Status='live', Symbol='Setd2', Aliases='4921524K10Rik, BC031601, KMT3A', description='SET domain containing 2', other_designations='SET domain-containing protein 2|lysine N-methyltransferase 3A', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110529296, end_position_on_the_genomic_accession=110618633, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  235627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235627, CurrentID=0, Status='live', Symbol='Nbeal2', Aliases='1110014F23Rik, BC042396, mKIAA0540', description='neurobeachin-like 2', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110624789, end_position_on_the_genomic_accession=110654161, orientation='minus', exon_count=54, OMIM='', no_hdr0=''),
  235628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235628, CurrentID=0, Status='live', Symbol='Prss42', Aliases='Tessp2', description='protease, serine 42', other_designations='protease, serine, 42|testis serine protease 2', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110797261, end_position_on_the_genomic_accession=110803744, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  235631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235631, CurrentID=0, Status='live', Symbol='Prss50', Aliases='79H19C, Tsp50', description='protease, serine 50', other_designations='protease, serine, 50|serine protease 50|testes-specific protease 50|testis-specific protease-like protein 50', map_location='9 60.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110857899, end_position_on_the_genomic_accession=110864629, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  235633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235633, CurrentID=0, Status='live', Symbol='Als2cl', Aliases='79H19D, 79H19E, BC022645, D930044G19Rik, mRn.49018', description='ALS2 C-terminal like', other_designations='mFLJ00189 protein', map_location='9 60.79 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110880136, end_position_on_the_genomic_accession=110900530, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  235636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235636, CurrentID=0, Status='live', Symbol='Rtp3', Aliases='Tmem7', description='receptor transporter protein 3', other_designations='receptor transporting protein 3|transmembrane protein 7', map_location='9 F', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110984938, end_position_on_the_genomic_accession=110990578, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  17428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17428, CurrentID=0, Status='live', Symbol='Mnt', Aliases='Rox, bHLHd3', description='max binding protein', other_designations='myc antagonist MNT', map_location='11 B-C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74830796, end_position_on_the_genomic_accession=74845725, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  104570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104570, CurrentID=0, Status='live', Symbol='Smek2', Aliases='AW011752, AW557776, Ppp4r3b, mKIAA1387', description='SMEK homolog 2, suppressor of mek1 (Dictyostelium)', other_designations='', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29167939, end_position_on_the_genomic_accession=29220797, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  104582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104582, CurrentID=0, Status='live', Symbol='Rprml', Aliases='AW049332', description='reprimo-like', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103649509, end_position_on_the_genomic_accession=103650580, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  235661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235661, CurrentID=0, Status='live', Symbol='Dync1li1', Aliases='1110053F02Rik, Dnclic1, LIC-1', description='dynein cytoplasmic 1 light intermediate chain 1', other_designations='DLC-A|dynein light chain A|dynein light intermediate chain 1, cytosolic|dynein, cytoplasmic, light intermediate chain 1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114688831, end_position_on_the_genomic_accession=114723788, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  192195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192195, CurrentID=0, Status='live', Symbol='Ash1l', Aliases='8030453L17Rik, Ash1, E430018P19Rik, Kmt2h', description='ash1 (absent, small, or homeotic)-like (Drosophila)', other_designations='ASH1-like protein|absent small and homeotic disks protein 1 homolog|absent, small, or homeotic discs 1|chromatin remodeling factor|probable histone-lysine N-methyltransferase ASH1L', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88950322, end_position_on_the_genomic_accession=89079375, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  104601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104601, CurrentID=0, Status='live', Symbol='Mycbpap', Aliases='4932408B01Rik, AMAP-1, AW125474', description='MYCBP associated protein', other_designations='AMAM-1|AMY-1-binding protein 1|AMY1-associating protein 1|mycbpap', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94501347, end_position_on_the_genomic_accession=94532388, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  235674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235674, CurrentID=0, Status='live', Symbol='Acaa1b', Aliases='', description='acetyl-Coenzyme A acyltransferase 1B', other_designations='acetyl-CoA acyltransferase B|beta-ketothiolase B|peroxisomal 3-oxoacyl-CoA thiolase B|thiolase B', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119148043, end_position_on_the_genomic_accession=119157093, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  192197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192197, CurrentID=0, Status='live', Symbol='Bcas3', Aliases='1500019F07Rik, 2610028P08Rik, AU021018, BC028339, K20D4, rudhira', description='breast carcinoma amplified sequence 3', other_designations='protein rudhira', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=85353116, end_position_on_the_genomic_accession=85826058, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  235682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235682, CurrentID=0, Status='live', Symbol='Zfp445', Aliases='AW610627, ZNF168, Znf445, mszf29, mszf50', description='zinc finger protein 445', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122848910, end_position_on_the_genomic_accession=122886850, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  628900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628900, CurrentID=0, Status='live', Symbol='Serpina3i', Aliases='2B2, EG628900, Gm6930', description='serine (or cysteine) peptidase inhibitor, clade A, member 3I', other_designations='alpha-1 antiproteinase|antitrypsin|serine (or cysteine) proteinase inhibitor, clade A, member 3I', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104263045, end_position_on_the_genomic_accession=104269365, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  104625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104625, CurrentID=0, Status='live', Symbol='Cnot6', Aliases='CCR4', description='CCR4-NOT transcription complex, subunit 6', other_designations='carbon catabolite repressor protein 4 homolog|cytoplasmic deadenylase', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49671497, end_position_on_the_genomic_accession=49714004, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  235712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235712, CurrentID=0, Status='live', Symbol='Mrgpra2b', Aliases='MrgA2, Mrgpra2, OTTMUSG00000028717', description='MAS-related GPR, member A2B', other_designations='G protein-coupled receptor MrgA2|MAS-related GPR, member A2|Mas-related gene A2', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=47463799, end_position_on_the_genomic_accession=47489585, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104662, CurrentID=0, Status='live', Symbol='Tsr1', Aliases='AU040765, AW550801, mKIAA1401', description='TSR1 20S rRNA accumulation', other_designations='TSR1, 20S rRNA accumulation, homolog', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74898080, end_position_on_the_genomic_accession=74909340, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  104681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104681, CurrentID=0, Status='live', Symbol='Slc16a6', Aliases='AW743111, ESTM12, MCT 6, MCT 7, MCT6', description='solute carrier family 16 (monocarboxylic acid transporters), member 6', other_designations='monocarboxylate transporter 6|solute carrier family 16 member 6', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=109450855, end_position_on_the_genomic_accession=109473596, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  628991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=628991, CurrentID=0, Status='live', Symbol='Obp1b', Aliases='EG628991, OBP-1b, OBP7', description='odorant binding protein IB', other_designations='odorant binding protein 7|odorant-binding protein Ib', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=78186827, end_position_on_the_genomic_accession=78191461, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  235779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235779, CurrentID=0, Status='live', Symbol='Nlrp4g', Aliases='nalp4g', description='NLR family, pyrin domain containing 4G', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=124348831, end_position_on_the_genomic_accession=124354491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  104709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104709, CurrentID=0, Status='live', Symbol='Pik3r6', Aliases='BB220380, p84, p84 PIKAP, p87(PIKAP), p87PIKAP', description='phosphoinositide-3-kinase, regulatory subunit 6', other_designations='p84 PI3K adapter protein|p87 phosphoinositide 3-kinase gamma (PI3Kg) adapter protein|phosphoinositide 3-kinase gamma adapter protein of 87 kDa', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68503019, end_position_on_the_genomic_accession=68552698, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  104718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104718, CurrentID=0, Status='live', Symbol='Ttc7b', Aliases='AA408451, Ttc7l1', description='tetratricopeptide repeat domain 7B', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100300770, end_position_on_the_genomic_accession=100520822, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  104721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104721, CurrentID=0, Status='live', Symbol='Ddx1', Aliases='AA409185, DBP-RB', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 1', other_designations='DEAD box protein 1|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 1', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=13219307, end_position_on_the_genomic_accession=13249174, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  104725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104725, CurrentID=0, Status='live', Symbol='Sptssa', Aliases='1110002B05Rik, AA407909, AA409588, AU041967, Ssspta', description='serine palmitoyltransferase, small subunit A', other_designations='small subunit of serine palmitoyltransferase A', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=54645374, end_position_on_the_genomic_accession=54656572, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  104732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104732, CurrentID=0, Status='live', Symbol='4930427A07Rik', Aliases='AA536649, mFLJ00116', description='RIKEN cDNA 4930427A07 gene', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113156421, end_position_on_the_genomic_accession=113165458, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  104759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104759, CurrentID=0, Status='live', Symbol='Pld4', Aliases='AI132321, thss', description='phospholipase D family, member 4', other_designations='PLD 4|choline phosphatase 4|phosphatidylcholine-hydrolyzing phospholipase D4', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112760655, end_position_on_the_genomic_accession=112768988, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  104771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104771, CurrentID=0, Status='live', Symbol='Jkamp', Aliases='1200003C05Rik, 2310047L11Rik, Jamp', description='JNK1/MAPK8-associated membrane protein', other_designations='medulloblastoma antigen MU-MB-50.4 homolog', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72085839, end_position_on_the_genomic_accession=72101527, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  104776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104776, CurrentID=0, Status='live', Symbol='Aldh6a1', Aliases='1110038I05Rik, Mmsdh', description='aldehyde dehydrogenase family 6, subfamily A1', other_designations='aldehyde dehydrogenase family 6 member A1|malonate-semialdehyde dehydrogenase', map_location='12 39.21 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84430717, end_position_on_the_genomic_accession=84450950, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  235854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235854, CurrentID=0, Status='live', Symbol='Mrgpra4', Aliases='Mgrg10, MrgA4', description='MAS-related GPR, member A4', other_designations='Mas-related G protein-coupled receptor g10|MrgA4 RF-amide G protein-coupled receptor|RF-amide G-protein coupled receptor', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=47980794, end_position_on_the_genomic_accession=47982296, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  104799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104799, CurrentID=0, Status='live', Symbol='Vipas39', Aliases='6720456H09Rik, 9330175H22Rik, AI413782, SPE-39, Spe39, Vipar, hSPE-39', description='VPS33B interacting protein, apical-basolateral polarity regulator, spe-39 homolog', other_designations='VPS33B-interacting protein in apical-basolateral polarity regulator|VPS33B-interacting protein in polarity and apical restriction|protein spe-39 homolog', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87238875, end_position_on_the_genomic_accession=87266378, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  104806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104806, CurrentID=0, Status='live', Symbol='Fancm', Aliases='AI427100, C730036B14Rik, D12Ertd364e', description='Fanconi anemia, complementation group M', other_designations='ATP-dependent RNA helicase FANCM', map_location='12 27.21 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=65074091, end_position_on_the_genomic_accession=65132058, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  104816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104816, CurrentID=0, Status='live', Symbol='Aspg', Aliases='A530050D06Rik, AI429460, Gm690', description='asparaginase', other_designations='60 kDa lysophospholipase [Includes: L-asparaginase|asparaginase homolog', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112106683, end_position_on_the_genomic_accession=112127573, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  192232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192232, CurrentID=0, Status='live', Symbol='Hps4', Aliases='2010205O06Rik, AU040608, BLOC-3, C130020P05Rik, le, mKIAA1667', description='Hermansky-Pudlak syndrome 4', other_designations='Hermansky-Pudlak syndrome 4 homolog|light ear|light-ear protein', map_location='5 54.69 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=112343056, end_position_on_the_genomic_accession=112378424, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  104831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104831, CurrentID=0, Status='live', Symbol='Ptpn23', Aliases='AI462446, PTP-TD14', description='protein tyrosine phosphatase, non-receptor type 23', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110385084, end_position_on_the_genomic_accession=110408210, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  235907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235907, CurrentID=0, Status='live', Symbol='Zfp65', Aliases='KRAB5, Zfp71-rs1, pz114', description='zinc finger protein 65', other_designations='zinc finger protein 71, related sequence 1', map_location='13 34.7 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67705306, end_position_on_the_genomic_accession=67729173, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104836, CurrentID=0, Status='live', Symbol='Cbll1', Aliases='AI467391, Hakai, c-Cbl-like', description='Casitas B-lineage lymphoma-like 1', other_designations='E-cadherin binding protein E7|c-Cbl-like protein 1|casitas B-lineage lymphoma-transforming sequence-like protein 1', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31484829, end_position_on_the_genomic_accession=31500741, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  192236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192236, CurrentID=0, Status='live', Symbol='Hps1', Aliases='6030422N11Rik, BB405864, Gm21361, Hps, ep', description='Hermansky-Pudlak syndrome 1', other_designations='Hermansky-Pudlak syndrome 1 homolog|Hermansky-Pudlak syndrome protein homolog|pale ear', map_location='19 36.56 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=42755191, end_position_on_the_genomic_accession=42780437, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  104859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104859, CurrentID=0, Status='live', Symbol='Tecpr2', Aliases='4930573I19Rik, AI604920, mKIAA0297', description='tectonin beta-propeller repeat containing 2', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110889259, end_position_on_the_genomic_accession=110972394, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  629147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629147, CurrentID=0, Status='live', Symbol='Ctxn3', Aliases='ENSMUSG00000069372', description='cortexin 3', other_designations='', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=57467656, end_position_on_the_genomic_accession=57478134, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  104871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104871, CurrentID=0, Status='live', Symbol='Spata7', Aliases='AI661438, B230306G18Rik, HSD3', description='spermatogenesis associated 7', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=98628123, end_position_on_the_genomic_accession=98669819, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  104884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104884, CurrentID=0, Status='live', Symbol='Tdp1', Aliases='2810481F14Rik, 4921509N21Rik, AI838772, AW493413, E430034L06Rik, MLZ-501, SCAN1', description='tyrosyl-DNA phosphodiesterase 1', other_designations='protein expressed in male leptotene and zygotene spermatocytes 501|tyr-DNA phosphodiesterase 1|tyrosyl-DNA phodphodiesterase 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=99884515, end_position_on_the_genomic_accession=99955219, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  104885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104885, CurrentID=0, Status='live', Symbol='Tmem179', Aliases='AI839735', description='transmembrane protein 179', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112500184, end_position_on_the_genomic_accession=112511160, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104886, CurrentID=0, Status='live', Symbol='Rab15', Aliases='2310012G06Rik, AI840042', description='RAB15, member RAS oncogene family', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76797960, end_position_on_the_genomic_accession=76822761, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  235973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235973, CurrentID=0, Status='live', Symbol='A630095E13Rik', Aliases='Gm191, Sslp-1, Sslp1', description='RIKEN cDNA A630095E13 gene', other_designations='sslp1', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=36635754, end_position_on_the_genomic_accession=36638602, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  104910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104910, CurrentID=0, Status='live', Symbol='Slc25a47', Aliases='AI132487, AI876593, HDMCP', description='solute carrier family 25, member 47', other_designations='hepatocellular carcinoma down-regulated mitochondrial carrier homolog|mitochondrial hepatocellular carcinoma-downregulated carrier protein', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=108849823, end_position_on_the_genomic_accession=108856815, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  104923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104923, CurrentID=0, Status='live', Symbol='Adi1', Aliases='AL024210, Fe-ARD, MTCBP-1, Ni-ARD', description='acireductone dioxygenase 1', other_designations='ARD|MTCBP-1|aci-reductone dioxygenase|acireductone dioxygenase (Fe(2+)-requiring)|acireductone dioxygenase (Ni(2+)-requiring)|fe-ARD|membrane-type 1 matrix metalloproteinase cytoplasmic tail-binding protein 1', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28675207, end_position_on_the_genomic_accession=28682175, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  629219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629219, CurrentID=0, Status='live', Symbol='Sult2a6', Aliases='EG629219, Gm6957', description='sulfotransferase family 2A, dehydroepiandrosterone (DHEA)-preferring, member 6', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=14222403, end_position_on_the_genomic_accession=14254870, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  104943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=104943, CurrentID=0, Status='live', Symbol='Fam110c', Aliases='9030611O19Rik, AU019489', description='family with sequence similarity 110, member C', other_designations='', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31073968, end_position_on_the_genomic_accession=31079940, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  629242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629242, CurrentID=0, Status='live', Symbol='Gm6958', Aliases='', description='predicted gene 6958', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=55961315, end_position_on_the_genomic_accession=55965068, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  236047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236047, CurrentID=0, Status='live', Symbol='Igkv4-74', Aliases='ENSMUSG00000076543, Gm10880, Gm192, IgVkai4', description='immunoglobulin kappa variable 4-74', other_designations='V kappa|immunoglobulin kappa chain variable 4-74|immunoglobulin light chain variable region ai4|light chain of the monoclonal antibody MST2', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69184826, end_position_on_the_genomic_accession=69185361, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105000, CurrentID=0, Status='live', Symbol='Dnal1', Aliases='1700010H15Rik, AW121714, Dnalc1, E330027P08Rik', description='dynein, axonemal, light chain 1', other_designations='', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84114280, end_position_on_the_genomic_accession=84143517, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105005, CurrentID=0, Status='live', Symbol='Fam84a', Aliases='2310003N02Rik, 4731402F03Rik, AW125753, Nse1', description='family with sequence similarity 84, member A', other_designations='', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=14147598, end_position_on_the_genomic_accession=14152038, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  236082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236082, CurrentID=0, Status='live', Symbol='Dhrsx', Aliases='C80877, D4Ertd296ey, Pscad, Dhrsx', description='dehydrogenase/reductase (SDR family) X chromosome', other_designations='SCAD family protein', map_location='4', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105014, CurrentID=0, Status='live', Symbol='Rdh14', Aliases='3110030G19Rik, AW215439, PAN2', description='retinol dehydrogenase 14 (all-trans and 9-cis)', other_designations='alcohol dehydrogenase PAN2|alcohol dehydrogenase PAN2; expressed sequence AW215439', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=10390690, end_position_on_the_genomic_accession=10395562, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105083, CurrentID=0, Status='live', Symbol='Pelo', Aliases='AA409897, AI852218', description='pelota homolog (Drosophila)', other_designations='', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=115088355, end_position_on_the_genomic_accession=115090158, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  629378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629378, CurrentID=0, Status='live', Symbol='Dact3', Aliases='EG666671, naa-3', description='dishevelled-binding antagonist of beta-catenin 3', other_designations='dact 3|dapper antagonist of catenin 3|dapper homolog 3, antagonist of beta-catenin', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16875317, end_position_on_the_genomic_accession=16887301, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  236193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236193, CurrentID=0, Status='live', Symbol='Zfp709', Aliases='BC021921, GIOT-4, HIT-40', description='zinc finger protein 709', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71882068, end_position_on_the_genomic_accession=71892565, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  192285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192285, CurrentID=0, Status='live', Symbol='Phf21a', Aliases='80kDa, Bhc, Bhc80, D030065N23Rik, PFTF1', description='PHD finger protein 21A', other_designations='BHC80a|BRAF35-HDAC complex protein BHC80|BRAF35/HDAC2 complex|Braf35/HDAC complex (Bhc)|mBHC80', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=92183746, end_position_on_the_genomic_accession=92364667, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  105148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105148, CurrentID=0, Status='live', Symbol='Iars', Aliases='2510016L12Rik, AI327140, AU044614, E430001P04Rik, ILRS', description='isoleucine-tRNA synthetase', other_designations='IRS|ileRS|isoleucyl-tRNA synthetase, cytoplasmic', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=49682130, end_position_on_the_genomic_accession=49734267, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  61217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61217, CurrentID=0, Status='live', Symbol='D1Mit206', Aliases='', description='DNA segment, Chr 1, Massachusetts Institute of Technology 206', other_designations='', map_location='1 80.33 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105171, CurrentID=0, Status='live', Symbol='Arrdc3', Aliases='AI450344, mKIAA1376', description='arrestin domain containing 3', other_designations='', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=80883338, end_position_on_the_genomic_accession=80896042, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  192292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192292, CurrentID=0, Status='live', Symbol='Nrbp1', Aliases='B230344L17Rik, Nrbp', description='nuclear receptor binding protein 1', other_designations='HLS7-interacting protein kinase|MLF1 adapter molecule', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31228089, end_position_on_the_genomic_accession=31251566, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  105193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105193, CurrentID=0, Status='live', Symbol='Nhlrc1', Aliases='AI505271, B230309E09Rik, EPM2B', description='NHL repeat containing 1', other_designations='NHL repeat-containing protein 1|malin', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=47012557, end_position_on_the_genomic_accession=47014850, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  236266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236266, CurrentID=0, Status='live', Symbol='Alms1', Aliases='bbb', description='Alstrom syndrome 1', other_designations='Alstrom syndrome 1 homolog', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85587499, end_position_on_the_genomic_accession=85726611, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  17533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17533, CurrentID=0, Status='live', Symbol='Mrc1', Aliases='AW259686, CD206, MR', description='mannose receptor, C type 1', other_designations='MMR', map_location='2 10.46 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14229414, end_position_on_the_genomic_accession=14332023, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  105203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105203, CurrentID=0, Status='live', Symbol='Fam208b', Aliases='AI645998, Pap20, mKIAA2006', description='family with sequence similarity 208, member B', other_designations='PBR associated protein|peripheral benzodiazepine receptor-associated protein 20|peripherial benzodiazepine receptor associated protein', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3566035, end_position_on_the_genomic_accession=3611231, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  629499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629499, CurrentID=0, Status='live', Symbol='Mroh8', Aliases='4922505G16Rik', description='maestro heat-like repeat family member 8', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=157208548, end_position_on_the_genomic_accession=157279549, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  236285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236285, CurrentID=0, Status='live', Symbol='Lancl3', Aliases='6030463G20Rik', description='LanC lantibiotic synthetase component C-like 3 (bacterial)', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9199973, end_position_on_the_genomic_accession=9268085, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  236293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236293, CurrentID=0, Status='live', Symbol='Slc22a29', Aliases='D630002G06Rik', description='solute carrier family 22. member 29', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8160168, end_position_on_the_genomic_accession=8253519, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  629524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629524, CurrentID=0, Status='live', Symbol='Olfr286', Aliases='EG629524, MOR286-2', description='olfactory receptor 286', other_designations='GA_x6K02T2NBG7-5528233-5529186|olfactory receptor MOR286-2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98226675, end_position_on_the_genomic_accession=98234450, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105239, CurrentID=0, Status='live', Symbol='Rnf44', Aliases='AI854545, mKIAA1100', description='ring finger protein 44', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54679399, end_position_on_the_genomic_accession=54693995, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  236312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236312, CurrentID=0, Status='live', Symbol='Pyhin1', Aliases='4930422C14, AI447904, Ifi209, Ifix', description='pyrin and HIN domain family, member 1', other_designations='ifi-209|interferon-inducible protein 209|interferon-inducible protein X|interferon-inducible protein p209', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173630859, end_position_on_the_genomic_accession=173647928, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105243, CurrentID=0, Status='live', Symbol='Slc9a3', Aliases='9030624O13Rik, AI930210, NHE-3, NHE3', description='solute carrier family 9 (sodium/hydrogen exchanger), member 3', other_designations='', map_location='13 40.15 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74086254, end_position_on_the_genomic_accession=74169442, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105245, CurrentID=0, Status='live', Symbol='Txndc5', Aliases='AL022641, ERp46, PC-TRP', description='thioredoxin domain containing 5', other_designations='ER protein 46|endoplasmic reticulum protein ERp46|endoplasmic reticulum resident protein 46|plasma cell-specific thioredoxin-related protein|thioredoxin-like protein p46', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38500266, end_position_on_the_genomic_accession=38528824, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  105246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105246, CurrentID=0, Status='live', Symbol='Brd9', Aliases='AL022779', description='bromodomain containing 9', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=73937825, end_position_on_the_genomic_accession=73960895, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  105278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105278, CurrentID=0, Status='live', Symbol='Cdk20', Aliases='4932702G04Rik, AU019450, AW558027, CDCH, Ccrk, PNQLARE, p42', description='cyclin-dependent kinase 20', other_designations='CAK-kinase p42|CDK-activating kinase p42|CDK-related protein kinase PNQLARE|cell cycle related kinase|cell cycle-related kinase|cell division protein kinase 20|cyclin-dependent protein kinase H|cyclin-kinase-activating kinase p42', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64432363, end_position_on_the_genomic_accession=64439722, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105298, CurrentID=0, Status='live', Symbol='Epdr1', Aliases='AU040950, AV006210, Epdm2, Epdr2, MERP-1, MERP-2, MERP2, Ucc1', description='ependymin related protein 1 (zebrafish)', other_designations='ependymin 2|ependymin related protein 2|mammalian ependymin related protein 1|upregulated in colorectal cancer gene 1', map_location='13 A2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19591708, end_position_on_the_genomic_accession=19619952, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105348, CurrentID=0, Status='live', Symbol='Golm1', Aliases='2310001L02Rik, AW125446, D030064E01Rik, GP73, Golph2, PSEC0257', description='golgi membrane protein 1', other_designations='golgi membrane protein GP73|golgi phosphoprotein 2', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59634996, end_position_on_the_genomic_accession=59675784, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  105349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105349, CurrentID=0, Status='live', Symbol='Akr1c18', Aliases='AW146047', description='aldo-keto reductase family 1, member C18', other_designations='20-alpha-HSD|20-alpha-hydroxysteroid dehydrogenase|20alpha-HSD|20alpha-hydroxysteroid dehydrogenase', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4132615, end_position_on_the_genomic_accession=4150651, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105351, CurrentID=0, Status='live', Symbol='AW209491', Aliases='', description='expressed sequence AW209491', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=14630245, end_position_on_the_genomic_accession=14638202, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105352, CurrentID=0, Status='live', Symbol='Dusp22', Aliases='JKAP, JSP1, LMW-DSP2, MKPX', description='dual specificity phosphatase 22', other_designations='low molecular weight dual specificity phosphatase 2', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=30660059, end_position_on_the_genomic_accession=30711232, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  629643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629643, CurrentID=0, Status='live', Symbol='Mif-ps9', Aliases='EG629643, Gm6990', description='macrophage migration inhibitory factor, pseudogene 9', other_designations='predicted pseudogene 6990|rp1 MIF pseudogene', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=56754878, end_position_on_the_genomic_accession=56755428, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105372, CurrentID=0, Status='live', Symbol='Utp15', Aliases='AW544865', description='UTP15, U3 small nucleolar ribonucleoprotein, homolog (yeast)', other_designations='src-associated protein SAW', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98246845, end_position_on_the_genomic_accession=98262992, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  105377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105377, CurrentID=0, Status='live', Symbol='Slf1', Aliases='2700017A04Rik, AW545819, Ankrd32, Brctd1, Brctx, C730024G01Rik', description='SMC5-SMC6 complex localization factor 1', other_designations='BRCT domain containing 1|BRCT domain-containing protein 1|ankyrin repeat domain 32|ankyrin repeat domain-containing protein 32|protein BRCTx', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=77043088, end_position_on_the_genomic_accession=77135468, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  236451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236451, CurrentID=0, Status='live', Symbol='Phf11b', Aliases='EG236451, Gm4902', description='PHD finger protein 11B', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59320964, end_position_on_the_genomic_accession=59341377, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105387, CurrentID=0, Status='live', Symbol='Akr1c14', Aliases='9030611N15Rik, AW557061', description='aldo-keto reductase family 1, member C14', other_designations='3-alpha-hydroxysteroid dehydrogenase type 1', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4048999, end_position_on_the_genomic_accession=4090560, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  432628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432628, CurrentID=0, Status='live', Symbol='Mfsd2b', Aliases='Gm1964', description='major facilitator superfamily domain containing 2B', other_designations='', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4862438, end_position_on_the_genomic_accession=4875044, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105418, CurrentID=0, Status='live', Symbol='E330034G19Rik', Aliases='AA415275, Zpac', description='RIKEN cDNA E330034G19 gene', other_designations='zygote-specific proteasome assembly chaperone', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=24293035, end_position_on_the_genomic_accession=24309960, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105428, CurrentID=0, Status='live', Symbol='Fam149b', Aliases='AA536717, AI4287731, mKIAA0974, Fam149b', description='family with sequence similarity 149, member B', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20348162, end_position_on_the_genomic_accession=20383491, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100113365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100113365, CurrentID=0, Status='live', Symbol='Nlgn4l', Aliases='NL-4, NL4, Nlgn4, Nlgn4x', description='neuroligin 4-like', other_designations='neuroligin 4, X-linked', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  236509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236509, CurrentID=0, Status='live', Symbol='Olfr366', Aliases='MOR138-5P, MOR138-6', description='olfactory receptor 366', other_designations='GA_x6K02T2NLDC-33902472-33903401|olfactory receptor MOR138-5P', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37219491, end_position_on_the_genomic_accession=37220420, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  236511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236511, CurrentID=0, Status='live', Symbol='Ago1', Aliases='Eif2c1', description='argonaute RISC catalytic subunit 1', other_designations='Piwi/Argonaute family protein meIF2C1|argonaute 1|argonaute RISC catalytic component 1|argonaute1|eIF-2C 1|eIF2C 1|eukaryotic translation initiation factor 2C, 1|mAgo1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126435012, end_position_on_the_genomic_accession=126468709, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  105440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105440, CurrentID=0, Status='live', Symbol='Kctd9', Aliases='', description='potassium channel tetramerisation domain containing 9', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=67716098, end_position_on_the_genomic_accession=67742310, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  105445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105445, CurrentID=0, Status='live', Symbol='Dock9', Aliases='AA959601, AW538057, B230309H04Rik, D14Wsu89e, Zizimin1, mKIAA1058', description='dedicator of cytokinesis 9', other_designations='cdc42 guanine nucleotide exchange factor zizimin-1', map_location='14 65.28 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=121542039, end_position_on_the_genomic_accession=121797897, orientation='minus', exon_count=62, OMIM='', no_hdr0=''),
  105446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105446, CurrentID=0, Status='live', Symbol='Gmpr2', Aliases='1810008P16Rik, 5730544D12Rik, AA959850, GMPR 2', description='guanosine monophosphate reductase 2', other_designations="guanosine 5'-monophosphate oxidoreductase 2", map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55672235, end_position_on_the_genomic_accession=55687634, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  105450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105450, CurrentID=0, Status='live', Symbol='Mmrn2', Aliases='AA986839, ENDOGLYX1, Emilin3, EndoGlyx-1', description='multimerin 2', other_designations='elastin microfibril interfacer 3|mmrn2', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34375504, end_position_on_the_genomic_accession=34404287, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  629747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629747, CurrentID=0, Status='live', Symbol='Spint3', Aliases='Gm14315, OTTMUSG00000016288', description='serine peptidase inhibitor, Kunitz type, 3', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164569695, end_position_on_the_genomic_accession=164573456, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  100113398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100113398, CurrentID=0, Status='live', Symbol='Adat3', Aliases='A430024H01Rik', description='adenosine deaminase, tRNA-specific 3', other_designations='adenosine deaminase, tRNA-specific 3, TAD2 homolog|tRNA-specific adenosine deaminase 3|tRNA-specific adenosine deaminase-like protein 3|tRNA-specific adenosine-34 deaminase subunit ADAT3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80602880, end_position_on_the_genomic_accession=80607654, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  236537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236537, CurrentID=0, Status='live', Symbol='Zfp352', Aliases='2czf48', description='zinc finger protein 352', other_designations='', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=90218820, end_position_on_the_genomic_accession=90225687, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  629754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629754, CurrentID=0, Status='live', Symbol='Wfdc9', Aliases='OTTMUSG00000001088', description='WAP four-disulfide core domain 9', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164649625, end_position_on_the_genomic_accession=164657094, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  236539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236539, CurrentID=0, Status='live', Symbol='Phgdh', Aliases='3-PGDH, 3PGDH, 4930479N23, A10, PGAD, PGD, PGDH, SERA', description='3-phosphoglycerate dehydrogenase', other_designations='3-phosphoglycerate dehyrogenase', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98313171, end_position_on_the_genomic_accession=98339969, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  629756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629756, CurrentID=0, Status='live', Symbol='Wfdc10', Aliases='Gm120, OTTMUSG00000016289', description='WAP four-disulfide core domain 10', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164656046, end_position_on_the_genomic_accession=164657368, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  629761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629761, CurrentID=0, Status='live', Symbol='Wfdc11', Aliases='', description='WAP four-disulfide core domain 11', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164662894, end_position_on_the_genomic_accession=164674253, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  105501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105501, CurrentID=0, Status='live', Symbol='Abhd4', Aliases='1110035H23Rik, AI429574, Abh4', description='abhydrolase domain containing 4', other_designations='alpha/beta-hydrolase 4|lyso-N-acylphosphatidylethanolamine lipase', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54254129, end_position_on_the_genomic_accession=54269169, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  236574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236574, CurrentID=0, Status='live', Symbol='Smok2b', Aliases='Gm1847, Smok+129', description='sperm motility kinase 2B', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13230263, end_position_on_the_genomic_accession=13237189, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  236576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236576, CurrentID=0, Status='live', Symbol='Spry3', Aliases='Gm1409, Gm391, sprouty3', description='sprouty homolog 3 (Drosophila)', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105511, CurrentID=0, Status='live', Symbol='Fam170b', Aliases='4922501K12Rik, AI449756', description='family with sequence similarity 170, member B', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32833962, end_position_on_the_genomic_accession=32836789, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105513, CurrentID=0, Status='live', Symbol='Chmp7', Aliases='4930596K11Rik, 6330407G04Rik, AI450338, AW550775', description='charged multivesicular body protein 7', other_designations='CHMP family, member 7|chromatin-modifying protein 7', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69716979, end_position_on_the_genomic_accession=69732570, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  105522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105522, CurrentID=0, Status='live', Symbol='Ankrd28', Aliases='AI465466, AI604979, E430019N21Rik, mKIAA0379', description='ankyrin repeat domain 28', other_designations='PITK|PP6-ARS-A|ankyrin repeat domain-containing protein 28|phosphatase interactor targeting protein hnRNP K|serine/threonine-protein phosphatase 6 regulatory subunit ARS-A', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31698768, end_position_on_the_genomic_accession=31830415, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  236604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236604, CurrentID=0, Status='live', Symbol='Pisd-ps1', Aliases='', description='phosphatidylserine decarboxylase, pseudogene 1', other_designations='phosphatidylserine decarboxylase pseudogene', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3124021, end_position_on_the_genomic_accession=3131945, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  629822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629822, CurrentID=0, Status='live', Symbol='Ighv7-3', Aliases='Gm16842, Igh-VS107, V(11)-djn', description='immunoglobulin heavy variable 7-3', other_designations='immunoglobulin heavy variable V7-3|monoclonal antibody heavy chain', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114153180, end_position_on_the_genomic_accession=114153644, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  629826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629826, CurrentID=0, Status='live', Symbol='Ighv14-4', Aliases='Gm16841, VH-D-J', description='immunoglobulin heavy variable 14-4', other_designations='immunoglobulin heavy variable V14-4', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114176438, end_position_on_the_genomic_accession=114176867, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  105559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105559, CurrentID=0, Status='live', Symbol='Mbnl2', Aliases='1110002M11Rik, AI047808, AI837313, AI849185, AL118326, MBLL, R75232, mKIAA4072', description='muscleblind-like 2', other_designations='', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=120275661, end_position_on_the_genomic_accession=120431698, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  236643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236643, CurrentID=0, Status='live', Symbol='Sytl5', Aliases='ENSMUSG00000054453, slp5', description='synaptotagmin-like 5', other_designations='', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=9747557, end_position_on_the_genomic_accession=9998864, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  629860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629860, CurrentID=0, Status='live', Symbol='Ighv1-12', Aliases='EG629860, Gm7011', description='immunoglobulin heavy variable V1-12', other_designations='anti-human CD20 antibody 1F5 gamma heavy chain variable region', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114615850, end_position_on_the_genomic_accession=114616143, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  629873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=629873, CurrentID=0, Status='live', Symbol='Krtap1-4', Aliases='OTTMUSG00000004966', description='keratin associated protein 1-4', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99582549, end_position_on_the_genomic_accession=99583677, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  105594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105594, CurrentID=0, Status='live', Symbol='Cphx1', Aliases='AU019881, AU023336, C330003B14Rik, C80129, Cphx, Eso-1', description='cytoplasmic polyadenylated homeobox 1', other_designations='', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25939951, end_position_on_the_genomic_accession=25956374, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  236690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236690, CurrentID=0, Status='live', Symbol='Nyx', Aliases='CLNP, CSNB1, CSNB4, nob', description='nyctalopin', other_designations='no b wave', map_location='X 8.37 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=13461095, end_position_on_the_genomic_accession=13489314, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105638, CurrentID=0, Status='live', Symbol='Dph3', Aliases='DELGIP1, DelgipP1, Desr1, KTI11, Zcsl2', description='diphthamine biosynthesis 3', other_designations='CSL-type zinc finger-containing protein 2|DPH3 homolog (KTI11, S. cerevisiae)|delGEF-interacting protein 1|zinc finger, CSL domain containing 2', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32080517, end_position_on_the_genomic_accession=32085692, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  105651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105651, CurrentID=0, Status='live', Symbol='Ppp1r3e', Aliases='A630071A11Rik, AW049239, Gm1211', description='protein phosphatase 1, regulatory (inhibitor) subunit 3E', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54875597, end_position_on_the_genomic_accession=54877538, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105653, CurrentID=0, Status='live', Symbol='Phyhip', Aliases='AW049870, C630010D02Rik, Lnap1ip, PAHX-AP#1, PAHX-AP1', description='phytanoyl-CoA hydroxylase interacting protein', other_designations='PAHXAP1|lupus nephritis-associated peptide 1 interacting protein|phytanoyl-CoA hydroxylase-associated protein 1', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70457517, end_position_on_the_genomic_accession=70468826, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  236727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236727, CurrentID=0, Status='live', Symbol='Slc9a7', Aliases='A530087D17Rik, NHE7', description='solute carrier family 9 (sodium/hydrogen exchanger), member 7', other_designations='NHE-7|Na(+)/H(+) exchanger 7|solute carrier family 9 (sodium/hydrogen exchanger)|solute carrier family 9 member 7', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20098851, end_position_on_the_genomic_accession=20291970, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  236732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236732, CurrentID=0, Status='live', Symbol='Rbm10', Aliases='E430039K10Rik', description='RNA binding motif protein 10', other_designations='RNA-binding motif protein 10', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20617503, end_position_on_the_genomic_accession=20650905, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  236733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236733, CurrentID=0, Status='live', Symbol='Usp11', Aliases='6230415D12Rik, mKIAA4085', description='ubiquitin specific peptidase 11', other_designations='deubiquitinating enzyme 11|ubiquitin specific protease 11|ubiquitin thioesterase 11|ubiquitin thiolesterase 11|ubiquitin-specific-processing protease 11', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20703909, end_position_on_the_genomic_accession=20720539, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  105663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105663, CurrentID=0, Status='live', Symbol='Thtpa', Aliases='AW122966', description='thiamine triphosphatase', other_designations='ThTPase', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55094784, end_position_on_the_genomic_accession=55098995, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  105670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105670, CurrentID=0, Status='live', Symbol='Rcbtb2', Aliases='2610028E02Rik, 2810420M18Rik, AI448345, AW240694, Chc1l, RC/BTB2', description='regulator of chromosome condensation (RCC1) and BTB (POZ) domain containing protein 2', other_designations='regulator of chromosome condensation and BTB domain-containing protein 2', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=73138773, end_position_on_the_genomic_accession=73184055, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105675, CurrentID=0, Status='live', Symbol='Ppif', Aliases='AW457192, CyP-D, CyP-F, CypD, PPIase', description='peptidylprolyl isomerase F (cyclophilin F)', other_designations='PPIase F|cyclophilin D|cyclophilin F|mitochondrial Cyclophilin D|peptidyl-prolyl cis-trans isomerase, mitochondrial|rotamase F', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=25693574, end_position_on_the_genomic_accession=25700468, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  105689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105689, CurrentID=0, Status='live', Symbol='Mycbp2', Aliases='AU023734, AW107953, AW546647, C130061D10Rik, Pam, Phr1, R75243', description='MYC binding protein 2', other_designations='Myc-binding protein 2|mycbp2|pam, highwire, rpm 1|pam/highwire/rpm-1 protein|probable E3 ubiquitin-protein ligase MYCBP2|protein associated with Myc', map_location='14 51.79 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=103113411, end_position_on_the_genomic_accession=103346809, orientation='minus', exon_count=86, OMIM='', no_hdr0=''),
  236781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236781, CurrentID=0, Status='live', Symbol='Gpr119', Aliases='', description='G-protein coupled receptor 119', other_designations='G protein-coupled receptor 119', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=48667979, end_position_on_the_genomic_accession=48674605, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  236785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236785, CurrentID=0, Status='live', Symbol='Olfr1321', Aliases='MOR264-26', description='olfactory receptor 1321', other_designations='GA_x5J8B7W1V10-1157-554|GA_x6K02T2Q64A-11768161-11767223', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=49726973, end_position_on_the_genomic_accession=49727932, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  236790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236790, CurrentID=0, Status='live', Symbol='Ddx26b', Aliases='4930535D10Rik, 6330505F04Rik, D130066O12', description='DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 26B', other_designations='', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56454821, end_position_on_the_genomic_accession=56507843, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  236792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236792, CurrentID=0, Status='live', Symbol='Mmgt1', Aliases='9630048L06Rik, BC032271, Emc5, Mir5116, Tmem32, mir-5116', description='membrane magnesium transporter 1', other_designations='ER membrane protein complex subunit 5|transmembrane protein 32', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56585512, end_position_on_the_genomic_accession=56597919, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105722, CurrentID=0, Status='live', Symbol='Ano6', Aliases='2900059G15Rik, AA407480, AW554778, F730003B03Rik, Tmem16f', description='anoctamin 6', other_designations='SCAN channel|small-conductance calcium-activated nonselective cation channel|transmembrane protein 16F', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=95790843, end_position_on_the_genomic_accession=95975472, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  236798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236798, CurrentID=0, Status='live', Symbol='Adgrg4', Aliases='Gm367, Gpr112, PGR17', description='adhesion G protein-coupled receptor G4', other_designations='G protein-coupled receptor 112|G-protein coupled receptor 112|probable G-protein coupled receptor 112', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56894234, end_position_on_the_genomic_accession=56980270, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  105727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105727, CurrentID=0, Status='live', Symbol='Slc38a1', Aliases='AA408026, AA409865, AL022800, AU015942, NAT2', description='solute carrier family 38, member 1', other_designations='MNat2|N-system amino acid transporter 2|amino acid transporter A1|system A amino acid transporter 1|system N amino acid transporter 1', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=96571418, end_position_on_the_genomic_accession=96642913, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  105732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105732, CurrentID=0, Status='live', Symbol='Fam83h', Aliases='AA409316', description='family with sequence similarity 83, member H', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76001092, end_position_on_the_genomic_accession=76014336, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  105734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105734, CurrentID=0, Status='live', Symbol='Tigd5', Aliases='AA409802, AA409995, AI666797', description='tigger transposable element derived 5', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75909735, end_position_on_the_genomic_accession=75914535, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  236848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236848, CurrentID=0, Status='live', Symbol='BC023829', Aliases='Tmem185a', description='cDNA sequence BC023829', other_designations='ee3|erythropoietin-induced EST 3', map_location='X A7.1-A7.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=70459753, end_position_on_the_genomic_accession=70477193, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  236852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236852, CurrentID=0, Status='live', Symbol='Magea10', Aliases='', description='melanoma antigen family A, 10', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=72381870, end_position_on_the_genomic_accession=72386858, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  105782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105782, CurrentID=0, Status='live', Symbol='Scrib', Aliases='AI118201, CRIB, Crc, SCRIB1, Scrb1, mKIAA0147, vartul', description='scribbled homolog (Drosophila)', other_designations='PDZ-domain protein scribble|circletail|scribble homolog 1', map_location='15 35.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76047162, end_position_on_the_genomic_accession=76069779, orientation='minus', exon_count=38, OMIM='', no_hdr0=''),
  105785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105785, CurrentID=0, Status='live', Symbol='Kdelr3', Aliases='AI173274', description='KDEL (Lys-Asp-Glu-Leu) endoplasmic reticulum protein retention receptor 3', other_designations='ER lumen protein retaining receptor 3|KDEL endoplasmic reticulum protein retention receptor 3|KDEL receptor 3', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79516408, end_position_on_the_genomic_accession=79527739, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  105787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105787, CurrentID=0, Status='live', Symbol='Prkaa1', Aliases='AI194361, AI450832, AL024255, AMPKalpha1, C130083N04Rik', description='protein kinase, AMP-activated, alpha 1 catalytic subunit', other_designations='ACACA kinase|AMP-activated protein kinase, alpha 1 catalytic subunit|AMPK subunit alpha-1|HMGCR kinase|acetyl-CoA carboxylase kinase|hydroxymethylglutaryl-CoA reductase kinase|tau-protein kinase PRKAA1', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=5143861, end_position_on_the_genomic_accession=5181899, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  236899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236899, CurrentID=0, Status='live', Symbol='Pcyt1b', Aliases='AW045697, CTTbeta', description='phosphate cytidylyltransferase 1, choline, beta isoform', other_designations='CCT B|CCT-beta|CT B|CTP:phosphocholine cytidylyltransferase B|phosphorylcholine transferase B', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=93654863, end_position_on_the_genomic_accession=93749951, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  236900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236900, CurrentID=0, Status='live', Symbol='Pdk3', Aliases='2610001M10Rik, AI035637', description='pyruvate dehydrogenase kinase, isoenzyme 3', other_designations='[Pyruvate dehydrogenase (acetyl-transferring)] kinase isozyme 3, mitochondrial|[Pyruvate dehydrogenase [lipoamide]] kinase isozyme 3, mitochondrial|pyruvate dehydrogenase kinase 3', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=93764616, end_position_on_the_genomic_accession=93832095, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  236904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236904, CurrentID=0, Status='live', Symbol='Klhl15', Aliases='6330500C13Rik', description='kelch-like 15', other_designations='', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94234495, end_position_on_the_genomic_accession=94277724, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105833, CurrentID=0, Status='live', Symbol='Ccdc65', Aliases='4933417K04Rik, AI428139, AU020178, NYD-SP28', description='coiled-coil domain containing 65', other_designations='testis development protein NYD-SP28', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98708227, end_position_on_the_genomic_accession=98723336, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  105835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105835, CurrentID=0, Status='live', Symbol='Sgsm3', Aliases='1810012I01Rik, AI428557, BB175482, Cip85, R75178, RUSC3, Rutbc3', description='small G protein signaling modulator 3', other_designations='Cx43-interacting protein of 85-kDa|RUN and TBC1 domain-containing protein 3', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80977765, end_position_on_the_genomic_accession=81012290, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  105837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105837, CurrentID=0, Status='live', Symbol='Mtbp', Aliases='AI429604, MDM2BP', description='Mdm2, transformed 3T3 cell double minute p53 binding protein', other_designations='mMTBP', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=55557408, end_position_on_the_genomic_accession=55626423, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  105841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105841, CurrentID=0, Status='live', Symbol='Dennd3', Aliases='AI447457, E030003N15Rik', description='DENN/MADD domain containing 3', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73512560, end_position_on_the_genomic_accession=73572242, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  236915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236915, CurrentID=0, Status='live', Symbol='Arhgef9', Aliases='9630036L12Rik', description='CDC42 guanine nucleotide exchange factor (GEF) 9', other_designations='collybistin|rac/Cdc42 guanine nucleotide exchange factor 9', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=95048935, end_position_on_the_genomic_accession=95196852, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  105844 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105844, CurrentID=0, Status='live', Symbol='Card10', Aliases='AI449026, Bimp1, Carma3', description='caspase recruitment domain family, member 10', other_designations='bcl10-interacting MAGUK protein 1', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78775136, end_position_on_the_genomic_accession=78803075, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  105847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105847, CurrentID=0, Status='live', Symbol='Lmf2', Aliases='AI451006, Tmem112b, Tmem153', description='lipase maturation factor 2', other_designations='transmembrane protein 112B|transmembrane protein 153', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89351004, end_position_on_the_genomic_accession=89355659, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  236920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236920, CurrentID=0, Status='live', Symbol='Stard8', Aliases='Dlc3, mKIAA0189', description='START domain containing 8', other_designations='START domain-containing protein 8', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=99003228, end_position_on_the_genomic_accession=99074728, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  105853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105853, CurrentID=0, Status='live', Symbol='Mal2', Aliases='AI461653', description='mal, T cell differentiation protein 2', other_designations='mal, T-cell differentiation protein 2', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=54571366, end_position_on_the_genomic_accession=54602846, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105855, CurrentID=0, Status='live', Symbol='Nckap1l', Aliases='4930568P13Rik, AI463083, Hem1, Hemp1', description='NCK associated protein 1 like', other_designations='hematopoietic protein 1', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=103453784, end_position_on_the_genomic_accession=103502860, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  236930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236930, CurrentID=0, Status='live', Symbol='Ercc6l', Aliases='BC004701, D330021P09Rik', description='excision repair cross-complementing rodent repair deficiency complementation group 6 like', other_designations='ATP-dependent helicase ERCC6-like|SNF2/RAD54 family protein|excision repair cross-complementing rodent repair deficiency complementation group 6 - like|excision repair cross-complementing rodent repair deficiency complementation group 6-like', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102142820, end_position_on_the_genomic_accession=102157091, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  105859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105859, CurrentID=0, Status='live', Symbol='Csdc2', Aliases='AI415250, AI481750, Pippin', description='cold shock domain containing C2, RNA binding', other_designations='RNA-binding protein PIPPin', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=81936759, end_position_on_the_genomic_accession=81950941, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  105866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105866, CurrentID=0, Status='live', Symbol='Krt72', Aliases='AI507495, K6irs2, Krt2-35-ps, Krt72', description='keratin 72', other_designations='CK-72|K72|cytokeratin-72|keratin 72, pseudogene|keratin complex 2, basic, gene 35|keratin-72|type II inner root sheath-specific keratin-K6irs2|type-II keratin Kb35', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101776560, end_position_on_the_genomic_accession=101786458, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  105887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105887, CurrentID=0, Status='live', Symbol='Ugt3a1', Aliases='AI746432', description='UDP glycosyltransferases 3 family, polypeptide A1', other_designations='UDP glucuronosyltransferase 3 family, polypeptide A1|UDPGT 3A1', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=9279826, end_position_on_the_genomic_accession=9315038, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  234847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=234847, CurrentID=0, Status='live', Symbol='Spg7', Aliases='AI452278, AU015315, Cmar, PGN', description='spastic paraplegia 7 homolog (human)', other_designations='cell adhesion regulator|cell matrix adhesion regulator', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123065508, end_position_on_the_genomic_accession=123097751, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  237038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237038, CurrentID=0, Status='live', Symbol='Nox1', Aliases='GP91-2, MOX1, NOH-1, NOH1, NOX1a, NOX1alpha, Nox-1', description='NADPH oxidase 1', other_designations='GP91 phox homolog|NADH/NADPH mitogenic oxidase subunit p65-mox|NADPH oxidase 1 alpha', map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134086012, end_position_on_the_genomic_accession=134137711, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  237052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237052, CurrentID=0, Status='live', Symbol='Tceal1', Aliases='0610011M09Rik, BC011290, P21, SIIR, pp21', description='transcription elongation factor A (SII)-like 1', other_designations='TCEA-like protein 1|transcription elongation factor S-II protein-like 1', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136707966, end_position_on_the_genomic_accession=136709873, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  630270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630270, CurrentID=0, Status='live', Symbol='Ighv4s2', Aliases='', description='Ig heavy chain V region', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105988, CurrentID=0, Status='live', Symbol='Espl1', Aliases='AL024103, AU045071, Cerp, ESP1, PRCE, SSE', description='extra spindle pole bodies 1, separase', other_designations='Cut1/ESP1 related protein|caspase-like protein ESPL1|extra spindle pole bodies 1|extra spindle poles-like 1 protein|separase', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102296253, end_position_on_the_genomic_accession=102324357, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  237073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237073, CurrentID=0, Status='live', Symbol='Rbm41', Aliases='BB131279, D330023I21Rik', description='RNA binding motif protein 41', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139889511, end_position_on_the_genomic_accession=139998595, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  630294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630294, CurrentID=0, Status='live', Symbol='Gm7030', Aliases='EG630294, H2-t9', description='predicted gene 7030', other_designations='MHC class Ib T9', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36127609, end_position_on_the_genomic_accession=36129425, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  237082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237082, CurrentID=0, Status='live', Symbol='Nxt2', Aliases='6330587F24Rik, AI787442, P15-2', description='nuclear transport factor 2-like export factor 2', other_designations='', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=142226770, end_position_on_the_genomic_accession=142239700, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  106014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106014, CurrentID=0, Status='live', Symbol='Fam19a5', Aliases='TAFA5, Tafa-5, Tara-5', description='family with sequence similarity 19, member A5', other_designations='chemokine-like protein TAFA-5', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=87544299, end_position_on_the_genomic_accession=87759364, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  237091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237091, CurrentID=0, Status='live', Symbol='Lhfpl1', Aliases='Lhfp', description='lipoma HMGIC fusion partner-like 1', other_designations='', map_location='X F2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=145290359, end_position_on_the_genomic_accession=145372304, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106021, CurrentID=0, Status='live', Symbol='Topors', Aliases='AW105885, LUN, TP53BPL, p53BP3/LUN', description='topoisomerase I binding, arginine/serine-rich', other_designations='SUMO1-protein E3 ligase Topors|p53-binding protein 3|p53BP3|topoisomerase 1-binding RING finger|topoisomerase I-binding RING finger protein|topoisomerase I-binding arginine/serine-rich protein|tumor suppressor p53-binding protein 3', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40259606, end_position_on_the_genomic_accession=40269841, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  106025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106025, CurrentID=0, Status='live', Symbol='Sharpin', Aliases='0610041B22Rik, AW121341, RBCKL1, SIPL1, cpdm', description='SHANK-associated RH domain interacting protein', other_designations='mSIPL1|protein kinase C-interacting protein RBCC like 1|shank-associated RH domain-interacting protein|shank-interacting protein-like 1', map_location='15 35.75 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76347040, end_position_on_the_genomic_accession=76351115, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  237107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237107, CurrentID=0, Status='live', Symbol='Gnl3l', Aliases='BC020354', description='guanine nucleotide binding protein-like 3 (nucleolar)-like', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=150983133, end_position_on_the_genomic_accession=151017322, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  106039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106039, CurrentID=0, Status='live', Symbol='Gga1', Aliases='4930406E12Rik, AU016030, AW209092', description='golgi associated, gamma adaptin ear containing, ARF binding protein 1', other_designations='gamma-adaptin-related protein 1|golgi-localized, gamma ear-containing, ARF-binding protein 1', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78877190, end_position_on_the_genomic_accession=78894585, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  106042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106042, CurrentID=0, Status='live', Symbol='Prickle1', Aliases='1110058P22Rik, AW215793, Prickle, b2b019Clo, mpk1', description='prickle homolog 1', other_designations='prickle like 1', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=93499111, end_position_on_the_genomic_accession=93596266, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  106052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106052, CurrentID=0, Status='live', Symbol='Fbxo4', Aliases='1700096C12Rik, AI851261, AW494535, Fbx4', description='F-box protein 4', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=3963564, end_position_on_the_genomic_accession=3979573, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  368202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=368202, CurrentID=0, Status='live', Symbol='Prss48', Aliases='Esspl, Gm1019', description='protease, serine 48', other_designations='epidermis-specific serine protease-like protein|protease, serine, 48', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=85973843, end_position_on_the_genomic_accession=86006002, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  368204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=368204, CurrentID=0, Status='live', Symbol='Khdc1a', Aliases='AU014713, C88118, Ndg1', description='KH domain containing 1A', other_designations='KH homology domain containing 1A|Nur77 downstream protein 1|nur77 downstream gene 1 protein', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=21329476, end_position_on_the_genomic_accession=21352200, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106064, CurrentID=0, Status='live', Symbol='AW549877', Aliases='AI195826', description='expressed sequence AW549877', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=3982035, end_position_on_the_genomic_accession=3995752, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106068, CurrentID=0, Status='live', Symbol='Slc45a4', Aliases='9330175B01Rik, AW552276', description='solute carrier family 45, member 4', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73577418, end_position_on_the_genomic_accession=73645757, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  106073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106073, CurrentID=0, Status='live', Symbol='Mfsd5', Aliases='AA409375, AW556797, D15Mgi27', description='major facilitator superfamily domain containing 5', other_designations='major facilitator superfamily domain-containing protein 5|molybdate transporter 2 homolog', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102279456, end_position_on_the_genomic_accession=102281744, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  237175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237175, CurrentID=0, Status='live', Symbol='Adgrg2', Aliases='AW212196, B830041D06Rik, Gpr64, Me6', description='adhesion G protein-coupled receptor G2', other_designations='G protein-coupled receptor 64|G-protein coupled receptor 64|LNB-TM7 heptahelical receptor Me6|epididymis-specific protein 6', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=160390646, end_position_on_the_genomic_accession=160498076, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  237178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237178, CurrentID=0, Status='live', Symbol='Ppef1', Aliases='PPEF-1', description='protein phosphatase with EF hand calcium-binding domain 1', other_designations='DRES10|PPEF 1|protein phosphatase with EF calcium-binding domain', map_location='X 73.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=160622419, end_position_on_the_genomic_accession=160766000, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  237211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237211, CurrentID=0, Status='live', Symbol='Fancb', Aliases='BC022692', description='Fanconi anemia, complementation group B', other_designations='', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=164980536, end_position_on_the_genomic_accession=164997272, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  237213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237213, CurrentID=0, Status='live', Symbol='Glra2', Aliases='', description='glycine receptor, alpha 2 subunit', other_designations='', map_location='X 76.75 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=165129017, end_position_on_the_genomic_accession=165327542, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  106143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106143, CurrentID=0, Status='live', Symbol='Cggbp1', Aliases='AA960172, AL023996, AW045841', description='CGG triplet repeat binding protein 1', other_designations='20 kDa CGG-binding protein|CGG-binding protein 1|p20-CGGBP DNA-binding protein', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=64851891, end_position_on_the_genomic_accession=64859491, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  237221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237221, CurrentID=0, Status='live', Symbol='Gemin8', Aliases='B130034M20Rik, B930082C09Rik, BC023488, gemin-8', description='gem (nuclear organelle) associated protein 8', other_designations='', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=166170412, end_position_on_the_genomic_accession=166190512, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  237222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237222, CurrentID=0, Status='live', Symbol='Ofd1', Aliases='Cxorf5, DXGgc7e, ORF2', description='oral-facial-digital syndrome 1 gene homolog (human)', other_designations='', map_location='X 77.28 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=166390033, end_position_on_the_genomic_accession=166440838, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  237253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237253, CurrentID=0, Status='live', Symbol='Lrp11', Aliases='1700034J19Rik, 6330533B21, 9830160H19Rik', description='low density lipoprotein receptor-related protein 11', other_designations='LRP-11', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7589767, end_position_on_the_genomic_accession=7625483, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  237256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237256, CurrentID=0, Status='live', Symbol='Zc3h12d', Aliases='D730019B10Rik, TFL', description='zinc finger CCCH type containing 12D', other_designations='transformed follicular lymphoma homolog|zc3h12d|zinc finger CCCH-type containing 12D', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=7832323, end_position_on_the_genomic_accession=7870397, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106200, CurrentID=0, Status='live', Symbol='Txndc11', Aliases='2810408E11Rik, AI427833, EFP1, Txdc11', description='thioredoxin domain containing 11', other_designations='EF-hand binding protein 1', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11074911, end_position_on_the_genomic_accession=11134681, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  106205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106205, CurrentID=0, Status='live', Symbol='Zc3h7a', Aliases='A430104C18Rik, AI447294, AW556219, HSPC055, Zc3h7, Zc3hdc7', description='zinc finger CCCH type containing 7 A', other_designations='zinc finger CCCH type domain containing 7', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=11136594, end_position_on_the_genomic_accession=11176393, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  630499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630499, CurrentID=0, Status='live', Symbol='H2-K2', Aliases='EG630499, Gm7035, H-2K1, H-2K1K, H-2K2, H2-K1', description='histocompatibility 2, K region locus 2', other_designations='histocompatibility 2, K region locus 1|predicted pseudogene 7035', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33974659, end_position_on_the_genomic_accession=33978791, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  237300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237300, CurrentID=0, Status='live', Symbol='Gm4922', Aliases='4933423E17, EG237300', description='predicted gene 4922', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=18777168, end_position_on_the_genomic_accession=18786793, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  237310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237310, CurrentID=0, Status='live', Symbol='Il22ra2', Aliases='CRF2-10, CRF2-s1, CRF2X, Il-22bp', description='interleukin 22 receptor, alpha 2', other_designations='IL-22 receptor subunit alpha-2|IL-22R-alpha-2|IL-22RA2|IL22BP|cytokine receptor family II soluble 1|cytokine receptor family type 2, soluble 1|interleukin-22-binding protein|zcytoR16', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19622028, end_position_on_the_genomic_accession=19634681, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  237313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237313, CurrentID=0, Status='live', Symbol='Il20ra', Aliases='E230031K19', description='interleukin 20 receptor, alpha', other_designations='IL-20 receptor subunit alpha|IL-20R-alpha|IL-20R1|IL-20RA', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=19711673, end_position_on_the_genomic_accession=19760053, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  432769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432769, CurrentID=0, Status='live', Symbol='Zfp708', Aliases='Rslcan11', description='zinc finger protein 708', other_designations='regulator of sex-limitation candidate 11', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67069398, end_position_on_the_genomic_accession=67097976, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  237320 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237320, CurrentID=0, Status='live', Symbol='Aldh8a1', Aliases='Raldh4', description='aldehyde dehydrogenase 8 family, member A1', other_designations='retinal dehydrogenase 4|retinal dehydrogenase type 4', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=21377300, end_position_on_the_genomic_accession=21396578, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  630537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630537, CurrentID=0, Status='live', Symbol='Dcpp2', Aliases='Dcpp-2, EG630537, SPT-2', description='demilune cell and parotid protein 2', other_designations='sublingual parotid tongue-2 salivary protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23898722, end_position_on_the_genomic_accession=23900787, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  237336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237336, CurrentID=0, Status='live', Symbol='Tbpl1', Aliases='4732475G08, AW011832, AW491032, D18347, STD, TLF, TRF2, TRP, Tlp', description='TATA box binding protein-like 1', other_designations='21 kDa TBP-like protein|TATA box binding protein-like protein|TATA box-binding protein-related factor 2|TBP-like factor|TBP-like protein 1|TBP-related factor 2|TBP-related protein', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=22703877, end_position_on_the_genomic_accession=22732573, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  237339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237339, CurrentID=0, Status='live', Symbol='L3mbtl3', Aliases='AI481284, MBT-1', description='l(3)mbt-like 3 (Drosophila)', other_designations='l(3)mbt-like protein 3', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=26274470, end_position_on_the_genomic_accession=26375989, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  237353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237353, CurrentID=0, Status='live', Symbol='Sh3rf3', Aliases='4831416G18Rik, Igfbp5-ip, Sh3md41, Sh3rf3', description='SH3 domain containing ring finger 3', other_designations='SH3 multiple domains 4|SH3 multiple domains protein 4|insulin-like growth factor binding protein 5 interacting protein|plenty of SH3s 2', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=58812372, end_position_on_the_genomic_accession=59138916, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  237360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237360, CurrentID=0, Status='live', Symbol='Adamts14', Aliases='Adamts-14, TS14', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 14', other_designations='a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 14', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61197112, end_position_on_the_genomic_accession=61273438, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  237362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237362, CurrentID=0, Status='live', Symbol='Npffr1', Aliases='Gm236, Gpr147, NPFF1, OT7T022', description='neuropeptide FF receptor 1', other_designations='G protein-coupled receptor 147|G-protein coupled receptor NPFF1R|RFamide-related peptide receptor', map_location='10 B4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=61595484, end_position_on_the_genomic_accession=61628762, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  630579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630579, CurrentID=0, Status='live', Symbol='Zfp808', Aliases='EG630579, Gm7036', description='zinc finger protein 80', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=62129864, end_position_on_the_genomic_accession=62173936, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106298, CurrentID=0, Status='live', Symbol='Rrn3', Aliases='AL023001, E130302O19Rik, R75565, TIF-1A, Tif1a', description='RRN3 RNA polymerase I transcription factor homolog (yeast)', other_designations='transcription initiation factor TIF-IA', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13780699, end_position_on_the_genomic_accession=13814841, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  237387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237387, CurrentID=0, Status='live', Symbol='Lrrc3', Aliases='1300011L04Rik, C21orf102', description='leucine rich repeat containing 3', other_designations='leucine-rich repeat-containing 3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77897575, end_position_on_the_genomic_accession=77902536, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  237397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237397, CurrentID=0, Status='live', Symbol='C2cd4c', Aliases='4932409I22Rik, Fam148c, Gm238', description='C2 calcium-dependent domain containing 4C', other_designations='C2cd4cC2CD4 family|family with sequence similarity 148, member C|nuclear-localized factor 3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79606854, end_position_on_the_genomic_accession=79614025, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106326, CurrentID=0, Status='live', Symbol='Osbpl11', Aliases='9430097N02Rik, AI132306, AU019795, ORP-11', description='oxysterol binding protein-like 11', other_designations='OSBP-related protein 11', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33169763, end_position_on_the_genomic_accession=33243312, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  237400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237400, CurrentID=0, Status='live', Symbol='Mex3d', Aliases='BC059858, Rkhd1', description='mex3 RNA binding family member D', other_designations='RING finger and KH domain-containing protein 1|mex3 homolog D|ring finger (C3HC4 type) and KH domain containing 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80380355, end_position_on_the_genomic_accession=80387651, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  237403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237403, CurrentID=0, Status='live', Symbol='Lingo3', Aliases='AI851712, LERN2', description='leucine rich repeat and Ig domain containing 3', other_designations='leucine-rich repeat neuronal protein 6B', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80832801, end_position_on_the_genomic_accession=80844039, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  106338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106338, CurrentID=0, Status='live', Symbol='Nsun3', Aliases='6720484A09Rik, AU022521', description='NOL1/NOP2/Sun domain family member 3', other_designations='NOL1/NOP2/Sun domain family 3', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=62732444, end_position_on_the_genomic_accession=62786794, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  237411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237411, CurrentID=0, Status='live', Symbol='Zfp938', Aliases='B230315N10Rik', description='zinc finger protein 938', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82224845, end_position_on_the_genomic_accession=82241318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106344, CurrentID=0, Status='live', Symbol='Rfc4', Aliases='A1, AI894123, AU040575, RFC37', description='replication factor C (activator 1) 4', other_designations='activator 1 subunit 4', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23113948, end_position_on_the_genomic_accession=23127730, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  106347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106347, CurrentID=0, Status='live', Symbol='Ildr1', Aliases='AU041483, AU044638', description='immunoglobulin-like domain containing receptor 1', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36693978, end_position_on_the_genomic_accession=36726804, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  237422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237422, CurrentID=0, Status='live', Symbol='Ric8b', Aliases='BC051080, Ric-8, Ric-8b', description='resistance to inhibitors of cholinesterase 8 homolog B (C. elegans)', other_designations='resistance to inhibitors of cholinesterase 8B', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=84917249, end_position_on_the_genomic_accession=85018337, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  237436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237436, CurrentID=0, Status='live', Symbol='Gas2l3', Aliases='8430435B07Rik, Gm240', description='growth arrest-specific 2 like 3', other_designations='growth arrest-specific protein 2-like 3', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=89408822, end_position_on_the_genomic_accession=89444374, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  106369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106369, CurrentID=0, Status='live', Symbol='Ypel1', Aliases='0610009L05Rik, 1700016N17Rik, 1700019O22Rik, 4921520K19Rik, 4930511F14Rik, AV118478, Dgl1, Ppil2, mdgl-1', description='yippee-like 1 (Drosophila)', other_designations='DGL-1|cardiac Abnormality/abnormal facies (CATCH22), microdeletion syndrome', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17070138, end_position_on_the_genomic_accession=17086736, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  237459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237459, CurrentID=0, Status='live', Symbol='Cdk17', Aliases='6430598J10Rik, Pctk2', description='cyclin-dependent kinase 17', other_designations='PCTAIRE-motif protein kinase 2|Pctaire2|cell division protein kinase 17|serine/threonine-protein kinase PCTAIRE-2', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93160876, end_position_on_the_genomic_accession=93250688, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  106389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106389, CurrentID=0, Status='live', Symbol='Eaf2', Aliases='AW048865, FESTA-L, FESTA-S, Festa, Traits, U19', description='ELL associated factor 2', other_designations='ehrlich S-II transcriptional activator factor|testosterone regulated apoptosis inducer and tumor suppressor|testosterone-regulated apoptosis inducer and tumor suppressor protein', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36792884, end_position_on_the_genomic_accession=36874826, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106393, CurrentID=0, Status='live', Symbol='Srl', Aliases='9830004M20Rik, AW108387, sar', description='sarcalumenin', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4480212, end_position_on_the_genomic_accession=4523627, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  106407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106407, CurrentID=0, Status='live', Symbol='Slc51a', Aliases='AV001382, AW261577, D630035O19Rik, OSTalpha, Osta', description='solute carrier family 51, alpha subunit', other_designations='OST-alpha|organic solute transporter alpha|solute carrier family 51 subunit alpha', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32475578, end_position_on_the_genomic_accession=32487879, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  237500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237500, CurrentID=0, Status='live', Symbol='Tmtc3', Aliases='9130014E20Rik, B130008E12Rik, mSmile', description='transmembrane and tetratricopeptide repeat containing 3', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=100443902, end_position_on_the_genomic_accession=100487347, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  237504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237504, CurrentID=0, Status='live', Symbol='Rassf9', Aliases='AW322107, Pamci', description='Ras association (RalGDS/AF-6) domain family (N-terminal) member 9', other_designations='P-CIP1|PAM COOH-terminal interactor protein 1|peptidylglycine alpha-amidating monooxygenase COOH-terminal interactor', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=102512222, end_position_on_the_genomic_accession=102549737, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  237523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237523, CurrentID=0, Status='live', Symbol='Ptprq', Aliases='', description='protein tyrosine phosphatase, receptor type, Q', other_designations='PTP-RQ|R-PTP-Q|phosphotidylinositol phosphatase PTPRQ|receptor-type tyrosine-protein phosphatase Q', map_location='10 55.97 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=107517360, end_position_on_the_genomic_accession=107720078, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  208677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=208677, CurrentID=0, Status='live', Symbol='Creb3l3', Aliases='BC010786, CREB-H, D10Bur1e', description='cAMP responsive element binding protein 3-like 3', other_designations='cAMP-responsive element-binding protein 3-like protein 3|transcription factor CREB-H', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81084324, end_position_on_the_genomic_accession=81098872, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  237542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237542, CurrentID=0, Status='live', Symbol='Osbpl8', Aliases='AA536976, AA536995, C730029P18Rik, D330025H14Rik, ORP-8', description='oxysterol binding protein-like 8', other_designations='OSBP-related protein 8', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=111164757, end_position_on_the_genomic_accession=111297249, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  237553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237553, CurrentID=0, Status='live', Symbol='Trhde', Aliases='9330155P21Rik', description='TRH-degrading enzyme', other_designations='PAP-II|TRH-DE|TRH-degrading ectoenzyme|TRH-specific aminopeptidase|pyroglutamyl-peptidase II|thyroliberinase|thyrotropin-releasing hormone degrading ectoenzyme', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=114398821, end_position_on_the_genomic_accession=114801370, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  237558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237558, CurrentID=0, Status='live', Symbol='Myrfl', Aliases='Gm239', description='myelin regulatory factor-like', other_designations='', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=116776545, end_position_on_the_genomic_accession=116896879, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  237560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237560, CurrentID=0, Status='live', Symbol='Lrrc10', Aliases='D330003I11Rik, Hrlrrp, Serdin1', description='leucine rich repeat containing 10', other_designations='heart-restricted leucine-rich repeat protein|leucine-rich containing 10|leucine-rich repeat protein|leucine-rich repeat-containing 10', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=117045341, end_position_on_the_genomic_accession=117046768, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  106489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106489, CurrentID=0, Status='live', Symbol='Sft2d1', Aliases='5630401J11Rik, AA409698', description='SFT2 domain containing 1', other_designations='SFT2 domain-containing protein 1|sft2d1', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=8311103, end_position_on_the_genomic_accession=8327442, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  106504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106504, CurrentID=0, Status='live', Symbol='Stk38', Aliases='5830476G13Rik, 9530097A09Rik, AA617404, Ndr1', description='serine/threonine kinase 38', other_designations='NDR1 protein kinase|nuclear Dbf2-related kinase 1|stk38', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28970885, end_position_on_the_genomic_accession=29008006, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  106512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106512, CurrentID=0, Status='live', Symbol='Gpsm3', Aliases='AA960287, AGS4', description='G-protein signalling modulator 3 (AGS3-like, C. elegans)', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34589806, end_position_on_the_genomic_accession=34591754, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106522, CurrentID=0, Status='live', Symbol='Pkdcc', Aliases='AI115348, AW548124, Adtk1, ESTM17, MAd1, Sgk493, Vlk, X83346', description='protein kinase domain containing, cytoplasmic', other_designations='protein kinase domain-containing protein, cytoplasmic|protein kinase-like protein SgK493|sugen kinase 493|vertebrate lonesome kinase', map_location='17 52.25 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=83215283, end_position_on_the_genomic_accession=83225069, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  106529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106529, CurrentID=0, Status='live', Symbol='Tecr', Aliases='2410016D23Rik, A230102P12Rik, AI173355, D17Ertd178e, Gpsn2, SC2', description='trans-2,3-enoyl-CoA reductase', other_designations='TER|glycoprotein, synaptic 2|synaptic glycoprotein SC2', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83571698, end_position_on_the_genomic_accession=83594491, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  237611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237611, CurrentID=0, Status='live', Symbol='Stac3', Aliases='9830125E18', description='SH3 and cysteine rich domain 3', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127501617, end_position_on_the_genomic_accession=127508823, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  237615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237615, CurrentID=0, Status='live', Symbol='Ankrd52', Aliases='6430544C07, G431002C21Rik', description='ankyrin repeat domain 52', other_designations='PP6-ARS-C|serine/threonine-protein phosphatase 6 regulatory subunit ARS-C', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128377124, end_position_on_the_genomic_accession=128394006, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  630836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630836, CurrentID=0, Status='live', Symbol='2010315B03Rik', Aliases='BC029716', description='RIKEN cDNA 2010315B03 gene', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=124291804, end_position_on_the_genomic_accession=124312721, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  237625 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237625, CurrentID=0, Status='live', Symbol='Pla2g3', Aliases='', description='phospholipase A2, group III', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3487596, end_position_on_the_genomic_accession=3499346, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  630845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630845, CurrentID=0, Status='live', Symbol='Tas2r122', Aliases='T2R22, Tas2r22, mGR22', description='taste receptor, type 2, member 122', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132710999, end_position_on_the_genomic_accession=132711928, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  237636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237636, CurrentID=0, Status='live', Symbol='Npc1l1', Aliases='9130221N23Rik, Gm243', description='NPC1-like 1', other_designations='Niemann-Pick disease, type C1|Niemann-Pick disease, type C1-like 1', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6211011, end_position_on_the_genomic_accession=6230245, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  106565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106565, CurrentID=0, Status='live', Symbol='Dlk2', Aliases='AI413481, DLK-2, Egfl9', description='delta-like 2 homolog (Drosophila)', other_designations='EGF-like protein 9|EGF-like-domain, multiple 9|endothelial cell-specific protein S-1|epidermal growth factor-like protein 9', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46297248, end_position_on_the_genomic_accession=46303271, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106572, CurrentID=0, Status='live', Symbol='Rab31', Aliases='1700093E07Rik, AI415285, Rab22B', description='RAB31, member RAS oncogene family', other_designations='', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65651726, end_position_on_the_genomic_accession=65772752, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  106581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106581, CurrentID=0, Status='live', Symbol='Fam234a', Aliases='Itfg3', description='family with sequence similarity 234, member A', other_designations='integrin alpha FG-GAP repeat containing 3', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26212692, end_position_on_the_genomic_accession=26244242, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  106582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106582, CurrentID=0, Status='live', Symbol='Nrm', Aliases='2610307M02Rik, AI429796', description='nurim (nuclear envelope membrane protein)', other_designations='nuclear envelope membrane protein|nuclear rim protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35861318, end_position_on_the_genomic_accession=35865400, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106583, CurrentID=0, Status='live', Symbol='Scaf8', Aliases='A630086M08Rik, A930036P18Rik, AA408306, AI447644, AI448652, AU015035, Rbm16, mKIAA1116', description='SR-related CTD-associated factor 8', other_designations='CTD-binding SR-like protein rA8|RNA binding motif protein 16|RNA-binding motif protein 16|SR-like CTD-associated factor 8|SR-related and CTD-associated factor 8|putative RNA-binding protein 16', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=3114972, end_position_on_the_genomic_accession=3198859, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  106585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106585, CurrentID=0, Status='live', Symbol='Ankrd12', Aliases='2900001A12Rik, AI447928, ANCO-2, AV347965, GAC-1, mKIAA0874', description='ankyrin repeat domain 12', other_designations='', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65965703, end_position_on_the_genomic_accession=66077046, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  106618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106618, CurrentID=0, Status='live', Symbol='Wdr90', Aliases='3230401M21Rik, AI551153, mKIAA1924', description='WD repeat domain 90', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25844734, end_position_on_the_genomic_accession=25861515, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  106628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106628, CurrentID=0, Status='live', Symbol='Trip10', Aliases='AI646975, Cip4', description='thyroid hormone receptor interactor 10', other_designations='Cdc42 interacting protein-4|TR-interacting protein 10|TRIP-10|thyroid receptor-interacting protein 10', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=57233207, end_position_on_the_genomic_accession=57263697, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  106633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106633, CurrentID=0, Status='live', Symbol='Ift140', Aliases='AI661311, Tce5, Wdtc2, mKIAA0590', description='intraflagellar transport 140', other_designations='T-complex expressed gene 5|WD and tetratricopeptide repeats 2|intraflagellar transport 140 homolog', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25016084, end_position_on_the_genomic_accession=25099499, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  106639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106639, CurrentID=0, Status='live', Symbol='Vmac', Aliases='AI662250', description='vimentin-type intermediate filament associated coiled-coil protein', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56713932, end_position_on_the_genomic_accession=56717699, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  237716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237716, CurrentID=0, Status='live', Symbol='Gpr75', Aliases='', description='G protein-coupled receptor 75', other_designations='', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=30884600, end_position_on_the_genomic_accession=30893729, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106648, CurrentID=0, Status='live', Symbol='Cyp4f15', Aliases='AI787289, Cyp4f-15', description='cytochrome P450, family 4, subfamily f, polypeptide 15', other_designations='cytochrome P450 CYP4F15|cytochrome P450, 4f15', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32685601, end_position_on_the_genomic_accession=32703352, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  237730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237730, CurrentID=0, Status='live', Symbol='Fbll1', Aliases='AI595406', description='fibrillarin-like 1', other_designations='protein-glutamine methyltransferase', map_location='11 A4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=35797380, end_position_on_the_genomic_accession=35798884, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  106672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106672, CurrentID=0, Status='live', Symbol='AI413582', Aliases='AI854251', description='expressed sequence AI413582', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=27563766, end_position_on_the_genomic_accession=27565799, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  237754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237754, CurrentID=0, Status='live', Symbol='Btnl9', Aliases='B430208I01, Btn3, D330012D11Rik', description='butyrophilin-like 9', other_designations='butyrophilin 3', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49168325, end_position_on_the_genomic_accession=49187089, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  237758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237758, CurrentID=0, Status='live', Symbol='Zfp454', Aliases='B230342N19', description='zinc finger protein 454', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50872715, end_position_on_the_genomic_accession=50887635, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  237759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237759, CurrentID=0, Status='live', Symbol='Col23a1', Aliases='2810458L13Rik, AI429655', description='collagen, type XXIII, alpha 1', other_designations='procollagen type XXIII alpha 1|procollagen, type XXIII, alpha 1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51289678, end_position_on_the_genomic_accession=51583925, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  237761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237761, CurrentID=0, Status='live', Symbol='Sowaha', Aliases='A830006N08Rik, Ankrd43', description='sosondowah ankyrin repeat domain family member A', other_designations='ankyrin repeat domain 43|ankyrin repeat domain-containing protein 43|protein sosondowah homolog A', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53476578, end_position_on_the_genomic_accession=53480195, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  237775 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237775, CurrentID=0, Status='live', Symbol='Zfp867', Aliases='', description='zinc finger protein 867', other_designations='zinc finger protein LOC237775', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59461197, end_position_on_the_genomic_accession=59472566, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  106707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106707, CurrentID=0, Status='live', Symbol='Rpusd1', Aliases='2310051D06Rik, AU019540, Rlucl', description='RNA pseudouridylate synthase domain containing 1', other_designations='ribosomal large subunit pseudouridine synthase C like|ribosomal large subunit pseudouridine synthase C-like protein', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25727703, end_position_on_the_genomic_accession=25731464, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  237781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237781, CurrentID=0, Status='live', Symbol='Mief2', Aliases='AI482195, Gm11, Smcr7', description='mitochondrial elongation factor 2', other_designations='mitochondrial dynamic protein MID49|mitochondrial dynamic protein of 49 kDa homolog|mitochondrial dynamics protein of 49 kDa homolog|smith-Magenis syndrome chromosomal region candidate gene 7 protein homolog', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60728398, end_position_on_the_genomic_accession=60732951, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  237782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237782, CurrentID=0, Status='live', Symbol='Smcr8', Aliases='2310076G09Rik, AI642055, D030073L15Rik', description='Smith-Magenis syndrome chromosome region, candidate 8 homolog (human)', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60777525, end_position_on_the_genomic_accession=60788287, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106722, CurrentID=0, Status='live', Symbol='G6b', Aliases='AU023871, C6orf25, Ng31', description='immunoreceptor tyrosine-based inhibitory motif (ITIM) containing platelet receptor', other_designations='ITIM-receptor G6b-B', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35062693, end_position_on_the_genomic_accession=35066785, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  237806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237806, CurrentID=0, Status='live', Symbol='Dnah9', Aliases='9030022M04, A230091C01, C230051G16, D11Ertd686e, Dnahc9, mKIAA0357', description='dynein, axonemal, heavy chain 9', other_designations='dynein, axonemal, heavy polypeptide 9', map_location='11 40.53 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=65831324, end_position_on_the_genomic_accession=66169186, orientation='minus', exon_count=69, OMIM='', no_hdr0=''),
  237823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237823, CurrentID=0, Status='live', Symbol='Pfas', Aliases='4432409B16Rik, FGAMS, FGAR-AT, FGARAT, Gm18, PURL, Sofa', description='phosphoribosylformylglycinamidine synthase (FGAR amidotransferase)', other_designations='FGAM synthase|FGAR amidotransferase|formylglycinamide ribonucleotide amidotransferase|formylglycinamide ribotide amidotransferase|formylglycinamide ribotide synthetase', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68985701, end_position_on_the_genomic_accession=69008460, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  106757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106757, CurrentID=0, Status='live', Symbol='Catsperd', Aliases='4921529N20Rik, 4933402B14Rik, 619718, AW045609, Gm6095, Tmem146', description='catsper channel auxiliary subunit delta', other_designations='catSper-delta|catSperdelta|transmembrane protein 146', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56626896, end_position_on_the_genomic_accession=56664460, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  106759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106759, CurrentID=0, Status='live', Symbol='Ticam1', Aliases='AW046014, AW547018, TICAM-1, TRIF', description='toll-like receptor adaptor molecule 1', other_designations='TIR domain-containing adapter protein inducing IFN-beta|TIR domain-containing adaptor inducing interferon-beta|toll-interleukin 1 receptor domain-containing adapter protein inducing interferon-beta|toll-interleukin-1 receptor domain-containing adapter protein inducing interferon beta', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=56269462, end_position_on_the_genomic_accession=56276767, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106763, CurrentID=0, Status='live', Symbol='Ttbk1', Aliases='AU017937, AW048023, C330008L01Rik', description='tau tubulin kinase 1', other_designations='', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46442448, end_position_on_the_genomic_accession=46487675, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  106766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106766, CurrentID=0, Status='live', Symbol='Stap2', Aliases='AW049765, STAP-2', description='signal transducing adaptor family member 2', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55997074, end_position_on_the_genomic_accession=56005643, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  237847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237847, CurrentID=0, Status='live', Symbol='Rtn4rl1', Aliases='Ngr3, Ngrh2, Ngrl2', description='reticulon 4 receptor-like 1', other_designations='nogo receptor-like 2|nogo-66 receptor homolog 2|nogo-66 receptor-related protein 3|rtn4rl1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75193993, end_position_on_the_genomic_accession=75267769, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  237858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237858, CurrentID=0, Status='live', Symbol='Tusc5', Aliases='C130069F04Rik, DSPB1, Lost1', description='tumor suppressor candidate 5', other_designations='dispanin subfamily B member 1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76679873, end_position_on_the_genomic_accession=76698664, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  237859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237859, CurrentID=0, Status='live', Symbol='Nsrp1', Aliases='8030451K07, AI851076, AL022898, Ccdc55, NSpr70, NSrp70', description='nuclear speckle regulatory protein 1', other_designations='coiled-coil domain containing 55|coiled-coil domain-containing protein 55|nuclear speckle-related protein 70', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77044292, end_position_on_the_genomic_accession=77078437, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  237860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237860, CurrentID=0, Status='live', Symbol='Ssh2', Aliases='SSH-2, SSH-2L, mSSH-2L', description='slingshot homolog 2 (Drosophila)', other_designations='SSH-like protein 2|slingshot 2L|slingshot-like 2', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=77216169, end_position_on_the_genomic_accession=77460220, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  432860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=432860, CurrentID=0, Status='live', Symbol='B020004C17Rik', Aliases='', description='RIKEN cDNA B020004C17 gene', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57015134, end_position_on_the_genomic_accession=57018982, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106794, CurrentID=0, Status='live', Symbol='Dhx57', Aliases='AW494914', description='DEAH (Asp-Glu-Ala-Asp/His) box polypeptide 57', other_designations='DEAH (Asp-Glu-Ala-Asp/His) box polypeptide 57 homolog|DEAH box protein 57', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80238304, end_position_on_the_genomic_accession=80290476, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  106795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106795, CurrentID=0, Status='live', Symbol='Tcf19', Aliases='5730403J10Rik, AW495861', description='transcription factor 19', other_designations='SC1 protein homolog|TCF-19', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35512730, end_position_on_the_genomic_accession=35516824, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  237868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237868, CurrentID=0, Status='live', Symbol='Sarm1', Aliases='A830091I15Rik, C78606, MyD885, Sarm', description='sterile alpha and HEAT/Armadillo motif containing 1', other_designations='Tir-1 homolog|sterile alpha and TIR motif containing 1', map_location='11 46.74 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78472330, end_position_on_the_genomic_accession=78497820, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  237877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237877, CurrentID=0, Status='live', Symbol='Atad5', Aliases='C130052G03Rik, FRAG1, Gm17', description='ATPase family, AAA domain containing 5', other_designations='chromosome fragility associated gene 1|chromosome fragility-associated gene 1 protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80089379, end_position_on_the_genomic_accession=80135797, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  237886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237886, CurrentID=0, Status='live', Symbol='Slfn9', Aliases='9830137M10Rik', description='schlafen 9', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=82978382, end_position_on_the_genomic_accession=82992233, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  237887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237887, CurrentID=0, Status='live', Symbol='Slfn10-ps', Aliases='Slfn10, mFLJ00257', description='schlafen 10, pseudogene', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83028123, end_position_on_the_genomic_accession=83040533, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  237890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237890, CurrentID=0, Status='live', Symbol='Slfn14', Aliases='Gm20-ps, Slfn14', description='schlafen 14', other_designations='schlafen family member 14, pseudogene', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83275110, end_position_on_the_genomic_accession=83286726, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  237891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237891, CurrentID=0, Status='live', Symbol='Gas2l2', Aliases='', description='growth arrest-specific 2 like 2', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83421635, end_position_on_the_genomic_accession=83429521, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  106821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106821, CurrentID=0, Status='live', Symbol='Oard1', Aliases='AI314976, AW558560', description='O-acyl-ADP-ribose deacylase 1', other_designations='O-acetyl-ADP-ribose deacetylase C6orf130 homolog', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=48409985, end_position_on_the_genomic_accession=48417270, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  237898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237898, CurrentID=0, Status='live', Symbol='Usp32', Aliases='2900074J03Rik, 6430526O11Rik, AW045245', description='ubiquitin specific peptidase 32', other_designations='ubiquitin specific protease 32', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84984437, end_position_on_the_genomic_accession=85140042, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  237911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237911, CurrentID=0, Status='live', Symbol='Brip1', Aliases='3110009N10Rik, 8030460J03Rik, Bach1, FACJ, Fancj, OF', description='BRCA1 interacting protein C-terminal helicase 1', other_designations='ATP-dependent RNA helicase BRIP1|BRCA1-associated C-terminal helicase 1', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86058131, end_position_on_the_genomic_accession=86201267, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  106840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106840, CurrentID=0, Status='live', Symbol='Unc119b', Aliases='AA407659, AI414892', description='unc-119 homolog B (C. elegans)', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115122566, end_position_on_the_genomic_accession=115134975, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  237926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237926, CurrentID=0, Status='live', Symbol='Rsad1', Aliases='B430319G23, BC056485', description='radical S-adenosyl methionine domain containing 1', other_designations='oxygen-independent coproporphyrinogen-III oxidase-like protein RSAD1', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=94539798, end_position_on_the_genomic_accession=94556861, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  237928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237928, CurrentID=0, Status='live', Symbol='Phospho1', Aliases='D11Moh36, Phospo1', description='phosphatase, orphan 1', other_designations='', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95824500, end_position_on_the_genomic_accession=95832140, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  631145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=631145, CurrentID=0, Status='live', Symbol='Fam90a1b', Aliases='4932442L08Rik', description='family with sequence similarity 90, member A1B', other_designations='', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=94339903, end_position_on_the_genomic_accession=94366937, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  237930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237930, CurrentID=0, Status='live', Symbol='Ttll6', Aliases='4932418K24Rik, D11Moh43e, D11Moh44e, t8130b59', description='tubulin tyrosine ligase-like family, member 6', other_designations='polyglutamylase|ttll6|tubulin--tyrosine ligase-like protein 6', map_location='11 59.48 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96133786, end_position_on_the_genomic_accession=96165452, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  106861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106861, CurrentID=0, Status='live', Symbol='Abhd3', Aliases='AA675331, LABH3', description='abhydrolase domain containing 3', other_designations='abhydrolase domain-containing protein 3|lung alpha/beta hydrolase 3|mmLABH3', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10644411, end_position_on_the_genomic_accession=10706771, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  237934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237934, CurrentID=0, Status='live', Symbol='Krt39', Aliases='4732494G06Rik, Ka35', description='keratin 39', other_designations='CK-39|K39|cytokeratin-39|keratin-39|ortholog of type I hair keratin KA35|type I hair keratin KA35', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=99514293, end_position_on_the_genomic_accession=99521338, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  237940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237940, CurrentID=0, Status='live', Symbol='Aoc2', Aliases='RAO', description='amine oxidase, copper containing 2 (retina-specific)', other_designations='retina-specific amine oxidase', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=101325063, end_position_on_the_genomic_accession=101329699, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  106869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106869, CurrentID=0, Status='live', Symbol='Tnfaip8', Aliases='AA987150, E130304C20Rik, ENSMUSG00000073567, Gg2-1, Gm10539, Nded, Ssc-2', description='tumor necrosis factor, alpha-induced protein 8', other_designations='TNF alpha-induced protein 8|TNF-induced protein', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=49979427, end_position_on_the_genomic_accession=50093229, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  237943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237943, CurrentID=0, Status='live', Symbol='Gpatch8', Aliases='5430405G24Rik, AU018890, ENSMUSG00000075516, Fbm1, Gpatc8, mKIAA0553', description='G patch domain containing 8', other_designations='Fyn-binding molecule 1', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=102475911, end_position_on_the_genomic_accession=102556158, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  106877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106877, CurrentID=0, Status='live', Symbol='Afap1l1', Aliases='AI173486', description='actin filament associated protein 1-like 1', other_designations='AFAP1-like protein 1', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61730262, end_position_on_the_genomic_accession=61786662, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  106878 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106878, CurrentID=0, Status='live', Symbol='Smim3', Aliases='2010002N04Rik, AI195350, AW561905, Nid67, cI-41', description='small integral membrane protein 3', other_designations='NGF-induced differentiation clone 67 protein|putative small membrane protein NID67|small membrane protein NID67', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60474191, end_position_on_the_genomic_accession=60518947, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  237958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237958, CurrentID=0, Status='live', Symbol='Sppl2c', Aliases='4933407P14Rik, IMP5', description='signal peptide peptidase 2C', other_designations='IMP-5|SPP-like 2C|intramembrane protease 5|ortholog of human intramembrane protease 5 IMP5|protein SPP-like 2C|protein SPPL2c', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=104186327, end_position_on_the_genomic_accession=104191166, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106894, CurrentID=0, Status='live', Symbol='Hmgxb3', Aliases='2510002C16Rik, A630042L21Rik, AI413166, mKIAA0194', description='HMG box domain containing 3', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61131277, end_position_on_the_genomic_accession=61177050, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  237979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237979, CurrentID=0, Status='live', Symbol='Sdk2', Aliases='4632412F08Rik, 5330435L01Rik, Sdk-2, mKIAA1514', description='sidekick homolog 2 (chicken)', other_designations='sidekick 2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=113776374, end_position_on_the_genomic_accession=114066181, orientation='minus', exon_count=47, OMIM='', no_hdr0=''),
  237987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237987, CurrentID=0, Status='live', Symbol='Otop2', Aliases='4732464P15', description='otopetrin 2', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115307163, end_position_on_the_genomic_accession=115332306, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  237988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237988, CurrentID=0, Status='live', Symbol='Cdr2l', Aliases='D030068L24Rik, Gm21', description='cerebellar degeneration-related protein 2-like', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115381916, end_position_on_the_genomic_accession=115396132, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  106931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106931, CurrentID=0, Status='live', Symbol='Kctd1', Aliases='4933402K10Rik, AI661543, AW553000', description='potassium channel tetramerisation domain containing 1', other_designations='potassium channel tetramerization domain containing 1', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=14968685, end_position_on_the_genomic_accession=15151446, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  238011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238011, CurrentID=0, Status='live', Symbol='Enpp7', Aliases='Alk-SMase, Gm254', description='ectonucleotide pyrophosphatase/phosphodiesterase 7', other_designations='', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=118988108, end_position_on_the_genomic_accession=118993270, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  106947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106947, CurrentID=0, Status='live', Symbol='Slc39a3', Aliases='AI845814, Zip3', description='solute carrier family 39 (zinc transporter), member 3', other_designations='ZIP-3|solute carrier family 39 member 3|zrt- and Irt-like protein 3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81028540, end_position_on_the_genomic_accession=81037421, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  238021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238021, CurrentID=0, Status='live', Symbol='Fscn2', Aliases='A930022G03, Ahl8, C630046B20Rik', description='fascin actin-bundling protein 2', other_designations='fascin homolog 2, actin-bundling protein, retinal|retinal fascin', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120359851, end_position_on_the_genomic_accession=120368173, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  238023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238023, CurrentID=0, Status='live', Symbol='Hexdc', Aliases='BC069960', description='hexosaminidase (glycosyl hydrolase family 20, catalytic domain) containing', other_designations='N-acetyl-beta-galactosaminidase|beta-N-acetylhexosaminidase|beta-hexosaminidase D|hexosaminidase domain-containing protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121204433, end_position_on_the_genomic_accession=121222656, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  106952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106952, CurrentID=0, Status='live', Symbol='Arap3', Aliases='Centd3, Drag1, E030006K04Rik, cnt-d3', description='ArfGAP with RhoGAP domain, ankyrin repeat and PH domain 3', other_designations='ARF-GAP, RHO-GAP, ankyrin repeat and pleckstrin homology domains-containing protein 3|centaurin, delta 3|dual-specificity Rho- and Arf-GTPase activating protein 1', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=37972623, end_position_on_the_genomic_accession=37999533, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  106957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106957, CurrentID=0, Status='live', Symbol='Slc39a6', Aliases='Ermelin', description='solute carrier family 39 (metal ion transporter), member 6', other_designations='ZIP-6|endoplasmic reticulum membrane protein|endoplasmic reticulum membrane-linked protein|solute carrier family 39 member 6|zrt- and Irt-like protein 6', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24579881, end_position_on_the_genomic_accession=24603817, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  17827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17827, CurrentID=0, Status='live', Symbol='Mtx1', Aliases='Gcap6, Mtx', description='metaxin 1', other_designations='mitochondrial outer membrane import complex protein 1', map_location='3 39.01 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89209081, end_position_on_the_genomic_accession=89214347, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  238037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238037, CurrentID=0, Status='live', Symbol='BC068281', Aliases='A430092C04', description='cDNA sequence BC068281', other_designations='WD repeat-containing protein C2orf44 homolog', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4843202, end_position_on_the_genomic_accession=4858901, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100114902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100114902, CurrentID=0, Status='live', Symbol='Trdd1', Aliases='Gm16623', description='T cell receptor delta diversity 1', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54113468, end_position_on_the_genomic_accession=54113476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  238055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238055, CurrentID=0, Status='live', Symbol='Apob', Aliases='AI315052, Apo B-100, apob-100, apob-48', description='apolipoprotein B', other_designations='', map_location='12 3.53 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=7977633, end_position_on_the_genomic_accession=8016839, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  238057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238057, CurrentID=0, Status='live', Symbol='Gdf7', Aliases='BMP12', description='growth differentiation factor 7', other_designations='GDF-7', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=8291018, end_position_on_the_genomic_accession=8302088, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  238076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238076, CurrentID=0, Status='live', Symbol='Kcns3', Aliases='', description='potassium voltage-gated channel, delayed-rectifier, subfamily S, member 3', other_designations='voltage-gated potassium channel subunit Kv9.3', map_location='12 5.58 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=11090202, end_position_on_the_genomic_accession=11158735, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  107022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107022, CurrentID=0, Status='live', Symbol='Gramd3', Aliases='9030613F08Rik, 9130427A09Rik, AW061306, NS3TP2', description='GRAM domain containing 3', other_designations='HCV NS3-transactivated protein 2', map_location='18 D2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=56387485, end_position_on_the_genomic_accession=56503792, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  107029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107029, CurrentID=0, Status='live', Symbol='Me2', Aliases='AW120568, D030040L20Rik, NAD-ME', description='malic enzyme 2, NAD(+)-dependent, mitochondrial', other_designations='', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=73764956, end_position_on_the_genomic_accession=73815478, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  107035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107035, CurrentID=0, Status='live', Symbol='Fbxo38', Aliases='6030410I24Rik, AU044865, AW214031, Moka, SP329', description='F-box protein 38', other_designations='F-box protein modulating KLF7 activity|modulator of KLF7 activity', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=62504059, end_position_on_the_genomic_accession=62548858, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  107045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107045, CurrentID=0, Status='live', Symbol='Lars', Aliases='2310045K21Rik, 3110009L02Rik, AW536573, mKIAA1352', description='leucyl-tRNA synthetase', other_designations='leuRS|leucyl-tRNA synthetase, cytoplasmic', map_location='18 E', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42202349, end_position_on_the_genomic_accession=42262189, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  107047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107047, CurrentID=0, Status='live', Symbol='Psmg2', Aliases='1700017I17Rik, AW545363, Clast3, Tnfsf5ip1', description='proteasome (prosome, macropain) assembly chaperone 2', other_designations='CD40 ligand-activated specific transcript 3|psmg2|tumor necrosis factor superfamily member 5-induced protein 1|tumor necrosis factor superfamily, member 5-induced protein 1', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67641596, end_position_on_the_genomic_accession=67654181, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  238123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238123, CurrentID=0, Status='live', Symbol='Cog5', Aliases='5430405C01Rik, C87247, D18362, GOLTC1, GTC90', description='component of oligomeric golgi complex 5', other_designations='COG complex subunit 5', map_location='12 A2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=31654868, end_position_on_the_genomic_accession=31937630, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  238130 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238130, CurrentID=0, Status='live', Symbol='Dock4', Aliases='5330406C03, 6330411N01Rik, AF263288, C030023J22, mKIAA0716', description='dedicator of cytokinesis 4', other_designations='', map_location='12 18.06 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=40445957, end_position_on_the_genomic_accession=40846874, orientation='plus', exon_count=53, OMIM='', no_hdr0=''),
  107065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107065, CurrentID=0, Status='live', Symbol='Lrrtm2', Aliases='AI851755, BB129880, C630011A14Rik, mKIAA0416', description='leucine rich repeat transmembrane neuronal 2', other_designations='leucine-rich repeat transmembrane neuronal 2 protein', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35209003, end_position_on_the_genomic_accession=35215024, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  107071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107071, CurrentID=0, Status='live', Symbol='Wdr74', Aliases='5730436H21Rik, AA407588', description='WD repeat domain 74', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8735799, end_position_on_the_genomic_accession=8740624, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  238161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238161, CurrentID=0, Status='live', Symbol='Akap6', Aliases='AI482140, Akapalpha, Akapbeta, PRKA', description='A kinase (PRKA) anchor protein 6', other_designations='A-kinase anchoring protein alpha|A-kinase anchoring protein beta|anchor protein 6', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=52699016, end_position_on_the_genomic_accession=53155568, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  107094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107094, CurrentID=0, Status='live', Symbol='Rrp12', Aliases='AA408556, AA536972, mKIAA0690', description='ribosomal RNA processing 12 homolog (S. cerevisiae)', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41862851, end_position_on_the_genomic_accession=41896153, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  238205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238205, CurrentID=0, Status='live', Symbol='Lrfn5', Aliases='AI427653, AI604817, C130061B21, mKIAA4208', description='leucine rich repeat and fibronectin type III domain containing 5', other_designations='', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=61522817, end_position_on_the_genomic_accession=61857930, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  107141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107141, CurrentID=0, Status='live', Symbol='Cyp2c50', Aliases='AI266900', description='cytochrome P450, family 2, subfamily c, polypeptide 50', other_designations='CYPIIC50|cyp2c50|cytochrome P450, 2c50', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40089679, end_position_on_the_genomic_accession=40113955, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  238217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238217, CurrentID=0, Status='live', Symbol='Rpl10l', Aliases='EG238217', description='ribosomal protein L10-like', other_designations='rpl10l', map_location='12 C2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=66283379, end_position_on_the_genomic_accession=66284401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  107146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107146, CurrentID=0, Status='live', Symbol='Glyat', Aliases='A330009E03Rik, ACGNAT, AI195249, AI315345, CAT, GAT', description='glycine-N-acyltransferase', other_designations='AAc|acyl-CoA:glycine N-acyltransferase|aralkyl acyl-CoA N-acyltransferase|aralkyl acyl-CoA:amino acid N-acyltransferase|benzoyl-coenzyme A:glycine N-acyltransferase|glycine N-benzoyltransferase', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12633308, end_position_on_the_genomic_accession=12651737, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  83397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83397, CurrentID=0, Status='live', Symbol='Akap12', Aliases='AI317366, SSeCKS, Srcs5, Tsga12', description='A kinase (PRKA) anchor protein (gravin) 12', other_designations='AKAP-12|gercelin|germ cell lineage protein gercelin|src-suppressed C kinase substrate|testis specific gene A12|v-src suppressed transcript 5', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=4266329, end_position_on_the_genomic_accession=4359471, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  107173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107173, CurrentID=0, Status='live', Symbol='Gpr137', Aliases='AI428855', description='G protein-coupled receptor 137', other_designations='G protein-coupled receptor TM7SF1L1|transmembrane 7 superfamily member 1-like 1 protein', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6938062, end_position_on_the_genomic_accession=6942450, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  238247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238247, CurrentID=0, Status='live', Symbol='Arid4a', Aliases='A630009N03, A630067N03Rik, MmRBBP1, Rbbp1', description='AT rich interactive domain 4A (RBP1-like)', other_designations='AT rich interactive domain 4A (Rbp1 like)|retinoblastoma binding protein 1|retinoblastoma-binding protein 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=71015967, end_position_on_the_genomic_accession=71099351, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  238252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238252, CurrentID=0, Status='live', Symbol='Gpr135', Aliases='PAFR', description='G protein-coupled receptor 135', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=72069618, end_position_on_the_genomic_accession=72070991, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  107182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107182, CurrentID=0, Status='live', Symbol='Btaf1', Aliases='AI414500, AI447930, E430027O22Rik, TAF170', description='B-TFIID TATA-box binding protein associated factor 1', other_designations='BTAF1 RNA polymerase II, B-TFIID transcription factor-associated, (Mot1 homolog, S. cerevisiae)|TBP-associated factor 170', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36926079, end_position_on_the_genomic_accession=37014057, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  238257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238257, CurrentID=0, Status='live', Symbol='Tmem30b', Aliases='9130011B11Rik', description='transmembrane protein 30B', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73543114, end_position_on_the_genomic_accession=73546395, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  238266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238266, CurrentID=0, Status='live', Symbol='Syt16', Aliases='6430703N11, Strep14, Syt14l, syt14r', description='synaptotagmin XVI', other_designations='synaptotagmin 14-like protein|synaptotagmin XIV-related protein', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=73997608, end_position_on_the_genomic_accession=74267916, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  107197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107197, CurrentID=0, Status='live', Symbol='Uqcc3', Aliases='AI462493, Cbp4', description='ubiquinol-cytochrome c reductase complex assembly factor 3', other_designations='UPF0723 protein C11orf83 homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8880014, end_position_on_the_genomic_accession=8880933, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  238271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238271, CurrentID=0, Status='live', Symbol='Kcnh5', Aliases='Eag2', description='potassium voltage-gated channel, subfamily H (eag-related), member 5', other_designations='ether-a-go-go potassium channel 2|voltage-gated potassium channel subunit Kv10.2', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=74897217, end_position_on_the_genomic_accession=75177332, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  238276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238276, CurrentID=0, Status='live', Symbol='Akap5', Aliases='3526401B18Rik, AKAP 150, AKAP-5, AKAP150, BB098886, Gm258, P150', description='A kinase (PRKA) anchor protein 5', other_designations='A-kinase anchor protein 150 kDa|cAMP-dependent protein kinase regulatory subunit II high affinity-binding protein', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76324402, end_position_on_the_genomic_accession=76334151, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  107221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107221, CurrentID=0, Status='live', Symbol='Ffar4', Aliases='AI552415, GT01, Gpr120, Gpr129, KPG_013, O3far1, Pgr4', description='free fatty acid receptor 4', other_designations='G protein-coupled receptor 120|G-protein coupled receptor 120|G-protein coupled receptor GT01|G-protein-coupled receptor GT01|omega-3 fatty acid receptor 1', map_location='19 C2-C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38097079, end_position_on_the_genomic_accession=38114263, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  107227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107227, CurrentID=0, Status='live', Symbol='Macrod1', Aliases='AI604841, AW743046, D930010J01Rik, Lrp16', description='MACRO domain containing 1', other_designations='MACRO domain-containing protein 1|[Protein ADP-ribosylglutamate] hydrolase', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7056715, end_position_on_the_genomic_accession=7198062, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  107239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107239, CurrentID=0, Status='live', Symbol='Carns1', Aliases='AI790298, Atpgd1, mKIAA1394', description='carnosine synthase 1', other_designations='ATP-grasp domain containing 1|ATP-grasp domain-containing protein 1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4164324, end_position_on_the_genomic_accession=4175479, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  107242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107242, CurrentID=0, Status='live', Symbol='AI837181', Aliases='Bles03, N28173', description='expressed sequence AI837181', other_designations='basophilic leukemia expressed protein|basophilic leukemia-expressed protein Bles03|protein WF-3', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5425144, end_position_on_the_genomic_accession=5427317, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  238317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238317, CurrentID=0, Status='live', Symbol='Elmsan1', Aliases='9430029N19Rik, C130039O16Rik, Gm260', description='ELM2 and Myb/SANT-like domain containing 1', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=84149168, end_position_on_the_genomic_accession=84218881, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  107250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107250, CurrentID=0, Status='live', Symbol='Kazald1', Aliases='AI842353, Bono1, IGFBP-rP10', description='Kazal-type serine peptidase inhibitor domain 1', other_designations='IGFBP-related protein 10|Kazal-type serine protease inhibitor domain 1|bone and odontoblast-expressed protein 1|insulin-like growth factor-binding-related protein 10|insulin-like growth factor-binding-related protein 4', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=45074485, end_position_on_the_genomic_accession=45079289, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  238323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238323, CurrentID=0, Status='live', Symbol='Rps6kl1', Aliases='A830084F09Rik, AW046440', description='ribosomal protein S6 kinase-like 1', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=85135248, end_position_on_the_genomic_accession=85151277, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  238328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238328, CurrentID=0, Status='live', Symbol='Vash1', Aliases='AI834978, D930046M13Rik, G630009D10Rik', description='vasohibin 1', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86678700, end_position_on_the_genomic_accession=86695688, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  238330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238330, CurrentID=0, Status='live', Symbol='Irf2bpl', Aliases='6430527G18Rik, Eap1', description='interferon regulatory factor 2 binding protein-like', other_designations='enhanced at puberty protein 1|polyglutamine-containing protein', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86880703, end_position_on_the_genomic_accession=86884814, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  238331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238331, CurrentID=0, Status='live', Symbol='Zdhhc22', Aliases='Gm262', description='zinc finger, DHHC-type containing 22', other_designations='DHHC-22|putative palmitoyltransferase ZDHHC22|zinc finger DHHC domain-containing protein 22', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=86980763, end_position_on_the_genomic_accession=86991867, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  107260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107260, CurrentID=0, Status='live', Symbol='Otub1', Aliases='AI850305', description='OTU domain, ubiquitin aldehyde binding 1', other_designations='OTU domain-containing ubiquitin aldehyde-binding protein 1|deubiquitinating enzyme OTUB1|otubain 1|otubain-1|ubiquitin-specific-processing protease OTUB1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7198206, end_position_on_the_genomic_accession=7206284, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  238333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238333, CurrentID=0, Status='live', Symbol='Samd15', Aliases='Gm263', description='sterile alpha motif domain containing 15', other_designations='SAM domain-containing protein 15', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87196168, end_position_on_the_genomic_accession=87213541, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  107271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107271, CurrentID=0, Status='live', Symbol='Yars', Aliases='AL024047, AU018965', description='tyrosyl-tRNA synthetase', other_designations='tyrRS|tyrosyl--tRNA ligase|tyrosyl-tRNA synthetase, cytoplasmic', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129189765, end_position_on_the_genomic_accession=129219607, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  107272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107272, CurrentID=0, Status='live', Symbol='Psat1', Aliases='D8Ertd814e, EPIP, PSA, Psat', description='phosphoserine aminotransferase 1', other_designations='endometrial progesterone-induced protein|phosphohydroxythreonine aminotransferase', map_location='19 10.86 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=15905123, end_position_on_the_genomic_accession=15925059, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  238377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238377, CurrentID=0, Status='live', Symbol='Gpr68', Aliases='BB131428, Ogr1', description='G protein-coupled receptor 68', other_designations='G-protein coupled receptor 68|sphingosylphosphorylcholine receptor', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=100876682, end_position_on_the_genomic_accession=100908198, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  238384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238384, CurrentID=0, Status='live', Symbol='Slc24a4', Aliases='A930002M03Rik, AW492432, NCKX4', description='solute carrier family 24 (sodium/potassium/calcium exchanger), member 4', other_designations='K-dependent Na/Ca exchanger NCKX4|Na(+)/K(+)/Ca(2+)-exchange protein 4|solute carrier family 24 member 4', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102128652, end_position_on_the_genomic_accession=102269083, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  238386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238386, CurrentID=0, Status='live', Symbol='Btbd7', Aliases='5730507E09Rik, 8030448M07, E130118E17Rik, FUP1, mKIAA1525', description='BTB (POZ) domain containing 7', other_designations='function-unkown protein 1', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=102780797, end_position_on_the_genomic_accession=102878471, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  107321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107321, CurrentID=0, Status='live', Symbol='Lpxn', Aliases='4933402K05Rik, A530083L21Rik, AV278559, LDPL', description='leupaxin', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12796198, end_position_on_the_genomic_accession=12833808, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  238395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238395, CurrentID=0, Status='live', Symbol='Serpina3j', Aliases='AI506554, EG238395, Gm4931', description='serine (or cysteine) peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 3J', other_designations='alpha-1 antiproteinase|serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 3J', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104300722, end_position_on_the_genomic_accession=104320586, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  107328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107328, CurrentID=0, Status='live', Symbol='Trpt1', Aliases='AU079016, AW045591, Tpt1, Tpt1h', description='tRNA phosphotransferase 1', other_designations="EST-MNCb3719|mTPT1|tRNA splicing 2' phosphotransferase 1 homolog", map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6996131, end_position_on_the_genomic_accession=6999048, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  238405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238405, CurrentID=0, Status='live', Symbol='Adam6b', Aliases='4930523C11Rik', description='a disintegrin and metallopeptidase domain 6B', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113489565, end_position_on_the_genomic_accession=113491835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  238406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238406, CurrentID=0, Status='live', Symbol='Adam6a', Aliases='Adam6, tMDCIV', description='a disintegrin and metallopeptidase domain 6A', other_designations='a disintegrin and metallopeptidase domain 6|disintegrin and metalloproteinase domain 6', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113543908, end_position_on_the_genomic_accession=113546414, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  107338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107338, CurrentID=0, Status='live', Symbol='Gbf1', Aliases='1700083E03Rik, AI035702, AW049900, mKIAA0248', description='golgi-specific brefeldin A-resistance factor 1', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=46152509, end_position_on_the_genomic_accession=46286510, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  238412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238412, CurrentID=0, Status='live', Symbol='Ighv2-3', Aliases='Gm16948', description='immunoglobulin heavy variable 2-3', other_designations='immunoglobulin heavy chain|immunoglobulin heavy variable V2-3', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113611184, end_position_on_the_genomic_accession=113611476, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  192653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192653, CurrentID=0, Status='live', Symbol='Ttc36', Aliases='', description='tetratricopeptide repeat domain 36', other_designations='TPR repeat protein 36', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44799400, end_position_on_the_genomic_accession=44802951, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  238418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238418, CurrentID=0, Status='live', Symbol='Ighv14-3', Aliases='Gm16840', description='immunoglobulin heavy variable V14-3', other_designations='V(H) region of G5 Bb 2.2|VDJ complex', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114059845, end_position_on_the_genomic_accession=114060273, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  107351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107351, CurrentID=0, Status='live', Symbol='Kank1', Aliases='A930031B09Rik, AU015049, AW121052, Ankrd15, D330024H06Rik, mKIAA0172', description='KN motif and ankyrin repeat domains 1', other_designations='ankyrin repeat domain 15', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=25236732, end_position_on_the_genomic_accession=25434498, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  238427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238427, CurrentID=0, Status='live', Symbol='Ighv6-6', Aliases='Gm16695, VH10G1', description='immunoglobulin heavy variable 6-6', other_designations='anti-dsRNA (RDV-RNA) antibody|immunoglobulin heavy variable V6-6', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114434788, end_position_on_the_genomic_accession=114435087, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  107358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107358, CurrentID=0, Status='live', Symbol='Tm9sf3', Aliases='1810073M23Rik, 2810031D16Rik, AI115521, AI413748, AW146116, AW549777, Smbp, mKIAA4036', description='transmembrane 9 superfamily member 3', other_designations='SM-11044 binding protein|transmembrane protein 9 superfamily member 3', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=41210842, end_position_on_the_genomic_accession=41264004, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  238440 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238440, CurrentID=0, Status='live', Symbol='LOC238440', Aliases='', description='Ig heavy chain V region', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107371, CurrentID=0, Status='live', Symbol='Exoc6', Aliases='4833405E05Rik, AW413330, C430002C19, Sec15, Sec15l1, hbd, msec15', description='exocyst complex component 6', other_designations='SEC15 homolog|SEC15-like 1|SEC15-like protein 1|exocyst complex component Sec15A', map_location='19 32.32 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=37520874, end_position_on_the_genomic_accession=37684052, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  107373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107373, CurrentID=0, Status='live', Symbol='Fam111a', Aliases='4632417K18Rik, AW413625', description='family with sequence similarity 111, member A', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12573075, end_position_on_the_genomic_accession=12589769, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  238447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238447, CurrentID=0, Status='live', Symbol='Igha', Aliases='IgA, Igh-2', description='immunoglobulin heavy constant alpha', other_designations='immunoglobulin heavy chain 2 (serum IgA)', map_location='12 62.09 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113256204, end_position_on_the_genomic_accession=113260236, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  107376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107376, CurrentID=0, Status='live', Symbol='E330013P04Rik', Aliases='9930018I18Rik, AW492152', description='RIKEN cDNA E330013P04 gene', other_designations='', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=60144675, end_position_on_the_genomic_accession=60162591, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  214505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=214505, CurrentID=0, Status='live', Symbol='Gnptg', Aliases='6430527N14Rik, A830081F19, AU067667, AU067744, Mdcp1, Tce7', description='N-acetylglucosamine-1-phosphotransferase, gamma subunit', other_designations='M6PR domain containing protein 1|M6PR domain-containing protein 1|T-complex expressed 7|UDP-N-acetylglucosamine-1-phosphotransferase subunit gamma|glcNAc-1-phosphotransferase subunit gamma|mannose-6-phosphate receptor domain', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25234317, end_position_on_the_genomic_accession=25240170, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  238463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238463, CurrentID=0, Status='live', Symbol='Tubal3', Aliases='', description='tubulin, alpha-like 3', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=3924635, end_position_on_the_genomic_accession=3935277, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  107392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107392, CurrentID=0, Status='live', Symbol='Brms1', Aliases='AV003220, AW554636', description='breast cancer metastasis-suppressor 1', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5041236, end_position_on_the_genomic_accession=5049917, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  61591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=61591, CurrentID=0, Status='live', Symbol='D9Mit285', Aliases='', description='DNA segment, Chr 9, Massachusetts Institute of Technology 285', other_designations='', map_location='9 21.42 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107417, CurrentID=0, Status='live', Symbol='Aap', Aliases='', description='active avoidance performance', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107425, CurrentID=0, Status='live', Symbol='Gcna1', Aliases='Gcna', description='germ cell nuclear antigen 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  238505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238505, CurrentID=0, Status='live', Symbol='Mtr', Aliases='AI894170, D830038K18Rik, MS', description='5-methyltetrahydrofolate-homocysteine methyltransferase', other_designations='5-methyltetrahydrofolate--homocysteine methyltransferase|vitamin-B12 dependent methionine synthase', map_location='13 4.52 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=12182720, end_position_on_the_genomic_accession=12258195, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  107435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107435, CurrentID=0, Status='live', Symbol='Hat1', Aliases='2410071B14Rik, AA536933, Hat-1, KAT1', description='histone aminotransferase 1', other_designations='histidine aminotransferase 1|histone acetyltransferase 1', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71389260, end_position_on_the_genomic_accession=71441622, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  107436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107436, CurrentID=0, Status='live', Symbol='Pchr', Aliases='', description='Plasmodium chaubaudi resistance', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107448, CurrentID=0, Status='live', Symbol='Unc5a', Aliases='Unc5h1, mKIAA1976', description='unc-5 homolog A (C. elegans)', other_designations='protein unc-5 homolog 1|protein unc-5 homolog A|unc-5 homolog 1|unc5 homolog 1', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=54949417, end_position_on_the_genomic_accession=55006018, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  107449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107449, CurrentID=0, Status='live', Symbol='Unc5b', Aliases='6330415E02Rik, A630020F16, D10Bwg0792e, Unc5h2', description='unc-5 homolog B (C. elegans)', other_designations='protein unc-5 homolog 2|protein unc-5 homolog B|unc-5 homolog 2|unc5 homolog 2', map_location='10 31.52 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=60762594, end_position_on_the_genomic_accession=60831604, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  107450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107450, CurrentID=0, Status='live', Symbol='Fv3', Aliases='Fv-3', description='Friend virus susceptibility 3', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107451, CurrentID=0, Status='live', Symbol='Fv6', Aliases='Fv-6', description='Friend virus susceptibility 6', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107452, CurrentID=0, Status='live', Symbol='Hld', Aliases='', description='hippocampal lamination defect', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107469, CurrentID=0, Status='live', Symbol='Nba1', Aliases='', description='New Zealand Black autoimmunity 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107476, CurrentID=0, Status='live', Symbol='Acaca', Aliases='A530025K05Rik, Acac, Acc1, Gm738', description='acetyl-Coenzyme A carboxylase alpha', other_designations='ACC-alpha|acaca|acetyl-CoA carboxylase 265', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=84129677, end_position_on_the_genomic_accession=84401651, orientation='plus', exon_count=58, OMIM='', no_hdr0=''),
  107477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107477, CurrentID=0, Status='live', Symbol='Guca1b', Aliases='BC018258, GCAP2', description='guanylate cyclase activator 1B', other_designations='GCAP 2|GCAP2 (retina)', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47384694, end_position_on_the_genomic_accession=47392967, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  107482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107482, CurrentID=0, Status='live', Symbol='Ack', Aliases='ack, acm', description='anti-Corynebacterium kutscheri', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  238555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238555, CurrentID=0, Status='live', Symbol='Btn2a2', Aliases='Btn2, D030063K05', description='butyrophilin, subfamily 2, member A2', other_designations='butyrophilin 2', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23477049, end_position_on_the_genomic_accession=23488858, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  107487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107487, CurrentID=0, Status='live', Symbol='Lus', Aliases='', description='lymphoid cytostasis suppressor', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  238564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238564, CurrentID=0, Status='live', Symbol='Mylk4', Aliases='EG238564', description='myosin light chain kinase family, member 4', other_designations='sgK085|sugen kinase 85', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=32700827, end_position_on_the_genomic_accession=32806771, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  631784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=631784, CurrentID=0, Status='live', Symbol='Gm7073', Aliases='', description='predicted gene 7073', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=60436052, end_position_on_the_genomic_accession=60456195, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  107503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107503, CurrentID=0, Status='live', Symbol='Atf5', Aliases='AFTA, Atf7, Atfx, ODA-10', description='activating transcription factor 5', other_designations='BZIP protein ATF7|NAP1|NRIF3-associated protein|activating transcription factor 5-alpha/beta|activating transcription factor 7|activating transcription factor X|cAMP-dependent transcription factor ATF-5|transcription factor ATFx|transcription factor-like protein ODA-10', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44812256, end_position_on_the_genomic_accession=44816658, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  107508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107508, CurrentID=0, Status='live', Symbol='Eprs', Aliases='2410081F06Rik, 3010002K18Rik, C79379, Qprs', description='glutamyl-prolyl-tRNA synthetase', other_designations='bifunctional aminoacyl-tRNA synthetase|glutamine-proline-tRNA synthetase', map_location='1 H4-H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=185363028, end_position_on_the_genomic_accession=185428355, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  107513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107513, CurrentID=0, Status='live', Symbol='Ssr1', Aliases='2510001K09Rik, 6330400D04, AI159733, AI452176, SSR, TRAPA', description='signal sequence receptor, alpha', other_designations='SSR-alpha|TRAP-alpha|signal sequence receptor subunit alpha|translocon-associated protein alpha', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=37971401, end_position_on_the_genomic_accession=37994203, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  107515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107515, CurrentID=0, Status='live', Symbol='Lgr4', Aliases='9130225G07, A330106J01Rik, A930009A08Rik, Gpr48', description='leucine-rich repeat-containing G protein-coupled receptor 4', other_designations='G protein-coupled receptor 48|G protein-coupled receptor GPR48|G-protein coupled receptor 48', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=109917647, end_position_on_the_genomic_accession=110014257, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  107522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107522, CurrentID=0, Status='live', Symbol='Ece2', Aliases='1810009K13Rik, 6330509A19Rik, 9630025D12Rik, BB127715, Ece-2', description='endothelin converting enzyme 2', other_designations='', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20611601, end_position_on_the_genomic_accession=20646480, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  107526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107526, CurrentID=0, Status='live', Symbol='Gimap4', Aliases='AU019574, E430007K16Rik, Ian-1, Ian1, Imap4', description='GTPase, IMAP family member 4', other_designations='immune associated nucleotide 1|immunity-associated protein 4', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48684487, end_position_on_the_genomic_accession=48692062, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  107527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107527, CurrentID=0, Status='live', Symbol='Il1rl2', Aliases='AI481289, AW551444, IL-1Rrp2', description='interleukin 1 receptor-like 2', other_designations='IL-36 receptor|IL1R-rp2|interleukin-1 receptor related protein 2|interleukin-1 receptor-related protein 2', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40324441, end_position_on_the_genomic_accession=40367585, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  107528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107528, CurrentID=0, Status='live', Symbol='Magee1', Aliases='AI847422, DAMAGE, mMage-e1', description='melanoma antigen, family E, 1', other_designations='MAGE-E1 antigen|alpha-dystrobrevin-associated MAGE Protein', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=105120396, end_position_on_the_genomic_accession=105123912, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  107539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107539, CurrentID=0, Status='live', Symbol='Lap2', Aliases='Lap-2', description='leucine aminopeptidase 2, serum', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107547, CurrentID=0, Status='live', Symbol='Bkm4', Aliases='Bkm-4', description='banded krait minor satellite DNA 4', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107566, CurrentID=0, Status='live', Symbol='Arl2bp', Aliases='1700010P10Rik, 1700027H16Rik, 6330544B05Rik, AI482273, AI849834, BART, Bart1', description='ADP-ribosylation factor-like 2 binding protein', other_designations='ADP-ribosylation-like 2 binding protein|ARF-like 2-binding protein|binder of ARF2 protein 1', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94666600, end_position_on_the_genomic_accession=94674457, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  107568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107568, CurrentID=0, Status='live', Symbol='Wwp1', Aliases='8030445B08Rik, AIP5, SDRP1, Tiul1', description='WW domain containing E3 ubiquitin protein ligase 1', other_designations='WW domain-containing protein 1', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=19608296, end_position_on_the_genomic_accession=19709004, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  107569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107569, CurrentID=0, Status='live', Symbol='Nt5c3', Aliases='1600024P05Rik, 2610206B05Rik, 3110004A18Rik, AI842776a, Nt5y, PN-1, PN-I, PSN1, UMPH, Umph-1, Umph1, cN-III, lupin, p36, Nt5c3', description="5'-nucleotidase, cytosolic III", other_designations="5'-nucleotidase, pyrimidine|7-methylguanosine nucleotidase|7-methylguanosine phosphate-specific 5'-nucleotidase|P5'N-1|P5N-1|cytosolic 5'-nucleotidase 3|cytosolic 5'-nucleotidase III|pyrimidine 5'-nucleotidase 1|uridine monophosphatase 1", map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=56882400, end_position_on_the_genomic_accession=56923932, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  107572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107572, CurrentID=0, Status='live', Symbol='Mchm2', Aliases='Mcm-2, Mcm2', description='modifier of chinchilla-mottled 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107573, CurrentID=0, Status='live', Symbol='Tcrg-C1', Aliases='Cgamma1, Gm17004, TRGC1', description='T cell receptor gamma, constant 1', other_designations='T-cell receptor gamma, constant 1', map_location='', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19214103, end_position_on_the_genomic_accession=19216618, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  107574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107574, CurrentID=0, Status='live', Symbol='Tcrg-C2', Aliases='Gm16603, Trgc2', description='T-cell receptor gamma, constant 2', other_designations='T cell receptor gamma constant 2', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19305060, end_position_on_the_genomic_accession=19307551, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  107577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107577, CurrentID=0, Status='live', Symbol='Inf1', Aliases='', description='acute ozone induced inflammation', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107581, CurrentID=0, Status='live', Symbol='Col16a1', Aliases='2700007F12Rik, A530052M23Rik, AI838662, CA1F', description='collagen, type XVI, alpha 1', other_designations='[a]1 (XVI) collagen|col16a1|procollagen type XVI alpha 1|procollagen, type XVI, alpha 1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=130047425, end_position_on_the_genomic_accession=130099277, orientation='plus', exon_count=74, OMIM='', no_hdr0=''),
  107585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107585, CurrentID=0, Status='live', Symbol='Dio3', Aliases='', description='deiodinase, iodothyronine type III', other_designations="5DIII|DIOIII|type 3 DI|type III iodothyronine deiodinase|type-III 5'-deiodinase", map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=110279230, end_position_on_the_genomic_accession=110281097, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  107586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107586, CurrentID=0, Status='live', Symbol='Ovol2', Aliases='1700108N11Rik, 1810007D21Rik, M-OVO, M-OVO-A, M-OVO-B, MOVO, Ovo2, Zfp339, movo2', description='ovo-like 2 (Drosophila)', other_designations='MOVO-B zinc finger protein|MOVO-C zinc finger protein|OVO homolog-like 2|zinc finger OVO2|zinc finger protein 339|zinc finger protein mOVO', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=144305175, end_position_on_the_genomic_accession=144332080, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  107587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107587, CurrentID=0, Status='live', Symbol='Osr2', Aliases='5430409I15RikA, Osr2B, Osr2', description='odd-skipped related 2', other_designations='', map_location='15 14.46 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=35296112, end_position_on_the_genomic_accession=35303308, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  107589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107589, CurrentID=0, Status='live', Symbol='Mylk', Aliases='9530072E15Rik, A930019C19Rik, AW489456, KRP, MLCK108, MLCK210, Mlck, nmMlck, smMLCK', description='myosin, light polypeptide kinase', other_designations='kinase-related protein|mylk|telokin', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=34745199, end_position_on_the_genomic_accession=35002436, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  238663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238663, CurrentID=0, Status='live', Symbol='Spata31d1d', Aliases='4932411G14Rik, Fam75d1d', description='spermatogenesis associated 31 subfamily D, member 1D', other_designations='family with sequence similarity 75, member D1D', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=59725925, end_position_on_the_genomic_accession=59731752, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  238673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238673, CurrentID=0, Status='live', Symbol='Zfp367', Aliases='8030486J21Rik, ZFF29a, ZFF29b, Znf367', description='zinc finger protein 367', other_designations='C2H2 zinc finger protein ZFF29', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64133057, end_position_on_the_genomic_accession=64153199, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  107605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107605, CurrentID=0, Status='live', Symbol='Rdh1', Aliases='', description='retinol dehydrogenase 1 (all trans)', other_designations='retinol dehydrogenase 1 (11-cis)', map_location='10 75.21 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127759726, end_position_on_the_genomic_accession=127768299, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  107607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107607, CurrentID=0, Status='live', Symbol='Nod1', Aliases='C230079P11, Card4, F830007N14Rik, Nlrc1', description='nucleotide-binding oligomerization domain containing 1', other_designations='caspase recruitment domain 4|caspase recruitment domain-containing protein 4', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54923942, end_position_on_the_genomic_accession=54972672, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  238680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238680, CurrentID=0, Status='live', Symbol='Cntnap3', Aliases='', description='contactin associated protein-like 3', other_designations='cell recognition molecule Caspr3', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=64736806, end_position_on_the_genomic_accession=64904028, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  238690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238690, CurrentID=0, Status='live', Symbol='Zfp458', Aliases='BC062958, Rslcan-7, Znf728, mszf59-2', description='zinc finger protein 458', other_designations='', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67254918, end_position_on_the_genomic_accession=67269068, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  238692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238692, CurrentID=0, Status='live', Symbol='Zfp874a', Aliases='C330011K17Rik, Rslcan15, Zfp874', description='zinc finger protein 874a', other_designations='KRAB zinc finger protein|regulator of sex-limitation candidate 15', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67440431, end_position_on_the_genomic_accession=67451624, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  238693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238693, CurrentID=0, Status='live', Symbol='Zfp58', Aliases='A530094I17Rik, Mfg-1, Mfg1, Rslcan5, Zfp817', description='zinc finger protein 58', other_designations='KRAB zinc finger protein|regulator of sex-limitation candidate 5|zfp-58|zinc finger protein 817|zinc finger protein Mfg-1', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67490167, end_position_on_the_genomic_accession=67500522, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  107626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107626, CurrentID=0, Status='live', Symbol='Asmt', Aliases='Hiomt', description='acetylserotonin O-methyltransferase', other_designations='hydroxyindole O-methyltransfease', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=170672644, end_position_on_the_genomic_accession=170678054, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  107627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107627, CurrentID=0, Status='live', Symbol='If2', Aliases='If-2', description='MTV-induced circulating interferon', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107650, CurrentID=0, Status='live', Symbol='Pi4kb', Aliases='ESTM41, PI4K-beta, PI4Kbeta, Pik4cb', description='phosphatidylinositol 4-kinase, catalytic, beta polypeptide', other_designations='ptdIns 4-kinase beta', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=94974731, end_position_on_the_genomic_accession=95006937, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  107652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107652, CurrentID=0, Status='live', Symbol='Uap1', Aliases='AA420407, AA437972, AGX-1, AGX-2, AGX1, AgX, ESTM38, SPAG2', description='UDP-N-acetylglucosamine pyrophosphorylase 1', other_designations='antigen X|sperm associated antigen 2', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=170142002, end_position_on_the_genomic_accession=170174971, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  238725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238725, CurrentID=0, Status='live', Symbol='Gpr150', Aliases='C030001A19Rik, PGR11', description='G protein-coupled receptor 150', other_designations='G-protein coupled receptor PGR11', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=76054851, end_position_on_the_genomic_accession=76056996, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  238726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238726, CurrentID=0, Status='live', Symbol='Fam81b', Aliases='Gm276', description='family with sequence similarity 81, member B', other_designations='', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=76201708, end_position_on_the_genomic_accession=76296477, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  107656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107656, CurrentID=0, Status='live', Symbol='Krt9', Aliases='EPPK, K9, Krt1-9', description='keratin 9', other_designations='CK-9|cytokeratin-9|keratin complex 1, acidic, gene 9|keratin-9|krt9', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100186781, end_position_on_the_genomic_accession=100193875, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  107684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107684, CurrentID=0, Status='live', Symbol='Coro2a', Aliases='9030208C03Rik, AI563590, IR10', description='coronin, actin binding protein 2A', other_designations='coronin 4', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46536937, end_position_on_the_genomic_accession=46601929, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  107685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107685, CurrentID=0, Status='live', Symbol='Carg2', Aliases='', description='Candida albicans resistance gene 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107686, CurrentID=0, Status='live', Symbol='Snrpd2', Aliases='1810009A06Rik, SMD2', description='small nuclear ribonucleoprotein D2', other_designations='sm-D2|snRNP core protein D2', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19149838, end_position_on_the_genomic_accession=19152726, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  107688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107688, CurrentID=0, Status='live', Symbol='Inf2q', Aliases='Inf2', description='subacute ozone induced inflammation QTL', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107689, CurrentID=0, Status='live', Symbol='Pet1', Aliases='Pex1', description='plasmacytoma expressed transcript 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107693, CurrentID=0, Status='live', Symbol='Hfm', Aliases='', description='hemifacial microsomia-associated locus', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107697, CurrentID=0, Status='live', Symbol='Tcrg-J', Aliases='Jgamma1', description='T cell receptor gamma, joining region', other_designations='T-cell receptor gamma, joining region', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107701, CurrentID=0, Status='live', Symbol='Sf3b4', Aliases='SF3b49, SF3b50, Sap49', description='splicing factor 3b, subunit 4', other_designations='spliceosome associated protein 49|splicing factor 3b, subunit 4, 49kD', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96172550, end_position_on_the_genomic_accession=96177564, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  631990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=631990, CurrentID=0, Status='live', Symbol='Cdr1', Aliases='100039759, Cdr, Cdr34, EG631990, Gm2409, Gm7077', description='cerebellar degeneration related antigen 1', other_designations='', map_location='X 34.33 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=61183246, end_position_on_the_genomic_accession=61185558, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  107711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107711, CurrentID=0, Status='live', Symbol='Tcrbe', Aliases='', description='T cell receptor beta chain enhancer', other_designations='T-cell receptor beta chain enhancer', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107723, CurrentID=0, Status='live', Symbol='Slc12a6', Aliases='9530023I19Rik, KCC3, gaxp', description='solute carrier family 12, member 6', other_designations='K-Cl cotransporter 3|electroneutral potassium-chloride cotransporter 3', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112265538, end_position_on_the_genomic_accession=112363163, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  107727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107727, CurrentID=0, Status='live', Symbol='Tcrg-V', Aliases='', description='T cell receptor gamma, variable region', other_designations='T-cell receptor gamma, variable region', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  238803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238803, CurrentID=0, Status='live', Symbol='Zfp366', Aliases='DC-SCRIPT, Znf366', description='zinc finger protein 366', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=99184823, end_position_on_the_genomic_accession=99247032, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  107732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107732, CurrentID=0, Status='live', Symbol='Mrpl10', Aliases='0610040E02Rik, C78440, MRP-L8, Rpml8', description='mitochondrial ribosomal protein L10', other_designations='L10mt|MRP-L10', map_location='11 D', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97041567, end_position_on_the_genomic_accession=97049213, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  107733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107733, CurrentID=0, Status='live', Symbol='Mrpl41', Aliases='MRP-L27, Rpml27', description='mitochondrial ribosomal protein L41', other_designations='L41mt|MRP-L41', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24972470, end_position_on_the_genomic_accession=24975098, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  107734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107734, CurrentID=0, Status='live', Symbol='Mrpl30', Aliases='2310001L22Rik, MRP-L28, Rpml28', description='mitochondrial ribosomal protein L30', other_designations='L30mt|MRP-L30', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=37890498, end_position_on_the_genomic_accession=37898338, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  107739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107739, CurrentID=0, Status='live', Symbol='Adre', Aliases='', description='age at disappearance of rooting response', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107742, CurrentID=0, Status='live', Symbol='Tse3', Aliases='Tse-3', description='tissue specific extinction 3, of albumin', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107746, CurrentID=0, Status='live', Symbol='Rapgef1', Aliases='4932418O06Rik, C3G, Grf2', description='Rap guanine nucleotide exchange factor (GEF) 1', other_designations='guanine nucleotide releasing factor 2', map_location='2 20.32 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=29619720, end_position_on_the_genomic_accession=29740364, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  107747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107747, CurrentID=0, Status='live', Symbol='Aldh1l1', Aliases='1810048F20Rik, FDH, Fthfd, Neut2', description='aldehyde dehydrogenase 1 family, member L1', other_designations='10-FTHFDH|10-formyltetrahydrofolate dehydrogenase|aldehyde dehydrogenase family 1 member L1|formyltetrahydrofolate dehydrogenase', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90550767, end_position_on_the_genomic_accession=90599171, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  107751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107751, CurrentID=0, Status='live', Symbol='Prrxl1', Aliases='Drg11, Drgx', description='paired related homeobox protein-like 1', other_designations='dorsal root ganglion 11|homeobox protein DRG11|paired homeodomain protein DRG11|paired related homeobox protein-like 1b|paired-related homeobox protein-like 1', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32595838, end_position_on_the_genomic_accession=32649320, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  107753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107753, CurrentID=0, Status='live', Symbol='Lgals2', Aliases='2200008F12Rik, AI324147', description='lectin, galactose-binding, soluble 2', other_designations='gal-2', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78850860, end_position_on_the_genomic_accession=78855529, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  107755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107755, CurrentID=0, Status='live', Symbol='Hv1', Aliases='Hv-1', description='hepatitis virus (MHV-2) susceptibility', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107758, CurrentID=0, Status='live', Symbol='Ril1', Aliases='Ril-1', description='radiation-induced leukemia sensitivity 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  238831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238831, CurrentID=0, Status='live', Symbol='Ppwd1', Aliases='4632422M10Rik, A330090G21Rik', description='peptidylprolyl isomerase domain and WD repeat containing 1', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104204869, end_position_on_the_genomic_accession=104228844, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  107765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107765, CurrentID=0, Status='live', Symbol='Ankrd1', Aliases='Alrp, CARP, Crap, MARP1', description='ankyrin repeat domain 1 (cardiac muscle)', other_designations='ankyrin-like repeat protein|cardiac ankyrin repeat protein|cardiac responsive adriamycin protein', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=36111965, end_position_on_the_genomic_accession=36119844, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  107766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107766, CurrentID=0, Status='live', Symbol='Haao', Aliases='0610007K21Rik, 0610012J07Rik, 3-HAO, 3-HAOxase, 3HAO', description='3-hydroxyanthranilate 3,4-dioxygenase', other_designations='3-HAO|3-hydroxyanthranilate oxygenase|3-hydroxyanthranilic acid dioxygenase|HAD', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=83831328, end_position_on_the_genomic_accession=83847950, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  107767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107767, CurrentID=0, Status='live', Symbol='Scamp1', Aliases='4930505M11Rik, AI415563, AW122395', description='secretory carrier membrane protein 1', other_designations='', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=94201433, end_position_on_the_genomic_accession=94285826, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  107769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107769, CurrentID=0, Status='live', Symbol='Tm6sf1', Aliases='AI428514, BC023123, C630016D09Rik', description='transmembrane 6 superfamily member 1', other_designations='', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81859001, end_position_on_the_genomic_accession=81884079, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  107770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107770, CurrentID=0, Status='live', Symbol='Tm6sf2', Aliases='', description='transmembrane 6 superfamily member 2', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70072924, end_position_on_the_genomic_accession=70080066, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  107771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107771, CurrentID=0, Status='live', Symbol='Bmyc', Aliases='2900002K07Rik, AW060705, Mycb', description='brain expressed myelocytomatosis oncogene', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25706879, end_position_on_the_genomic_accession=25707721, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  107774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107774, CurrentID=0, Status='live', Symbol='Tse2', Aliases='Tse-2', description='tissue specific extinction 2, of L-ADH', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  238871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238871, CurrentID=0, Status='live', Symbol='Pde4d', Aliases='9630011N22Rik, Dpde3', description='phosphodiesterase 4D, cAMP specific', other_designations='cAMP-specific phosphodiesterase 4D|cyclic AMP specific phosphodiesterase PDE4D5A|dunce', map_location='13 59.69 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=108654177, end_position_on_the_genomic_accession=109955969, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  238875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238875, CurrentID=0, Status='live', Symbol='Gapt', Aliases='9830130M13Rik', description='Grb2-binding adaptor, transmembrane', other_designations='growth factor receptor-bound protein 2-binding adapter protein, transmembrane', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=110352614, end_position_on_the_genomic_accession=110357273, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  238880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238880, CurrentID=0, Status='live', Symbol='Actbl2', Aliases='4732495G21Rik', description='actin, beta-like 2', other_designations='kappa-actin', map_location='13 D2.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=111255013, end_position_on_the_genomic_accession=111257749, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  107815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107815, CurrentID=0, Status='live', Symbol='Scml2', Aliases='4932420G07Rik', description='sex comb on midleg-like 2 (Drosophila)', other_designations='', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=161082629, end_position_on_the_genomic_accession=161258218, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  107817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107817, CurrentID=0, Status='live', Symbol='Jmjd6', Aliases='5730436I23Rik, D11Ertd195e, PSR, PtdSerR, Ptdsr, mKIAA0585', description='jumonji domain containing 6', other_designations='apoptotic cell clearance receptor PtdSerR|histone arginine demethylase JMJD6|jmjC domain-containing protein 6|jumonji domain-containing protein 6|lysyl-hydroxylase JMJD6|peptide-lysine 5-dioxygenase JMJD6|phosphatidylserine receptor', map_location='11 81.49 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116837432, end_position_on_the_genomic_accession=116843449, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  107819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107819, CurrentID=0, Status='live', Symbol='Myhn4', Aliases='Myhn-4', description='myosin heavy chain, nonmuscular, type D', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107823, CurrentID=0, Status='live', Symbol='Whsc1', Aliases='5830445G22Rik, 9430010A17Rik, AW555663, C130020C13Rik, D030027O06Rik, D930023B08Rik, MMSET, NSD2l, mKIAA1090, Whsc1', description='Wolf-Hirschhorn syndrome candidate 1 (human)', other_designations='IL5 promoter REII region-binding protein|multiple myeloma SET domain-containing protein|nuclear SET domain-containing protein 2|probable histone-lysine N-methyltransferase NSD2|trithorax/ash1-related protein 5|wolf-Hirschhorn syndrome candidate 1 protein homolog', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33820775, end_position_on_the_genomic_accession=33897966, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  107829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107829, CurrentID=0, Status='live', Symbol='Thoc5', Aliases='1700060C24Rik, A430085L24Rik, Fmip, PK1.3, Sgk2', description='THO complex 5', other_designations='Fms interacting protein|fms-interacting protein|serum/glucocorticoid regulated kinase 2', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4895283, end_position_on_the_genomic_accession=4928867, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  107831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107831, CurrentID=0, Status='live', Symbol='Adgrb1', Aliases='B830018M07Rik, Bai1, R75078, mKIAA4089', description='adhesion G protein-coupled receptor B1', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74516196, end_position_on_the_genomic_accession=74589465, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  107833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107833, CurrentID=0, Status='live', Symbol='Cdc2b', Aliases='Cdc2', description='cell division cycle 2 homolog B (S. pombe)', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107835, CurrentID=0, Status='live', Symbol='Idd6', Aliases='Idd-6', description='insulin dependent diabetes susceptibility 6', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105355, CurrentID=0, Status='live', Symbol='Slc17a3', Aliases='AW261723, Npt4', description='solute carrier family 17 (sodium phosphate), member 3', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23836918, end_position_on_the_genomic_accession=23860718, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  107849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107849, CurrentID=0, Status='live', Symbol='Prl2c5', Aliases='Ghd3, MRP-4, Mrpplf4, PLF-4', description='prolactin family 2, subfamily c, member 5', other_designations='growth hormone d3|mitogen regulated protein, proliferin 4', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=13182703, end_position_on_the_genomic_accession=13191930, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  107851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107851, CurrentID=0, Status='live', Symbol='Ifx', Aliases='Hsv-1, If-X', description='NDV-induced circulating interferon, X-linked', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107853, CurrentID=0, Status='live', Symbol='Scl1', Aliases='Scl-1', description='susceptibility to cutaneous leishmaniasis 1', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107864, CurrentID=0, Status='live', Symbol='Jevr', Aliases='', description='resistance to Japanese encephalitis virus', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107868, CurrentID=0, Status='live', Symbol='Usp9y', Aliases='Dffry, Fafl2', description='ubiquitin specific peptidase 9, Y chromosome', other_designations='ubiquitin specific protease 9, Y chromosome', map_location='Y A1', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=1298961, end_position_on_the_genomic_accession=1459782, orientation='minus', exon_count=47, OMIM='', no_hdr0=''),
  107869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107869, CurrentID=0, Status='live', Symbol='Cth', Aliases='0610010I13Rik, AI314617, BC019483, CSE, Cys3', description='cystathionase (cystathionine gamma-lyase)', other_designations='cysteine-protein sulfhydrase|gamma-cystathionase', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=157894248, end_position_on_the_genomic_accession=157925081, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  107884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107884, CurrentID=0, Status='live', Symbol='Idd14', Aliases='', description='insulin dependent diabetes susceptibility 14', other_designations='', map_location='13 20.16 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107885, CurrentID=0, Status='live', Symbol='Mthfs', Aliases='1110034I12Rik, 2310020H23Rik, AI119695', description='5, 10-methenyltetrahydrofolate synthetase', other_designations='5,10-methenyl-tetrahydrofolate synthetase|methenyl-THF synthetase', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89211190, end_position_on_the_genomic_accession=89240225, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  107888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107888, CurrentID=0, Status='live', Symbol='Pbr', Aliases='', description='resistance to Paracoccidioides brasiliensis', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107889, CurrentID=0, Status='live', Symbol='Gcm2', Aliases='Gcm-rs1, Gcm1, Gcm1-rs2', description='glial cells missing homolog 2 (Drosophila)', other_designations='GCM motif protein 2|mGCMb', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41101427, end_position_on_the_genomic_accession=41111035, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  107895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107895, CurrentID=0, Status='live', Symbol='Mgat5', Aliases='4930471A21Rik, 5330407H02Rik, AI480971, GlcNAc-TV', description='mannoside acetylglucosaminyltransferase 5', other_designations='GNT-V|N-acetylglucosaminyl-transferase V|alpha-mannoside beta-1,6-N-acetylglucosaminyltransferase|beta1,6N-acetylglucosaminyltransferase V|glcNAc-T V', map_location='1 E3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=127204977, end_position_on_the_genomic_accession=127488334, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  238988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238988, CurrentID=0, Status='live', Symbol='Erc2', Aliases='6430531D06, CAST, CAST1/ERC2, D14Ertd171e, ELKS2alpha', description='ELKS/RAB6-interacting/CAST family member 2', other_designations='CAST1|CAZ-associated structural protein 1', map_location='14 16.51 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=27622410, end_position_on_the_genomic_accession=28478537, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  107924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107924, CurrentID=0, Status='live', Symbol='Ema', Aliases='', description='electrophoretic mobility and agglutinibility', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107925, CurrentID=0, Status='live', Symbol='Gdcr1', Aliases='Gdcr-1', description='glycerolphosphate dehydrogenase regulator 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107927, CurrentID=0, Status='live', Symbol='Htr1e', Aliases='Htr1ea', description='5-hydroxytryptamine (serotonin) receptor 1E', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107932, CurrentID=0, Status='live', Symbol='Chd4', Aliases='9530019N15Rik, AA617397, BC005710, D6Ertd380e, Mi-2beta, mKIAA4075', description='chromodomain helicase DNA binding protein 4', other_designations='CHD-4|Mi-2 beta', map_location='6 59.28 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125095986, end_position_on_the_genomic_accession=125130591, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  107934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107934, CurrentID=0, Status='live', Symbol='Celsr3', Aliases='Adgrc3, Fmi1, flamingo, mKIAA0812', description='cadherin, EGF LAG seven-pass G-type receptor 3', other_designations='cadherin, EGF LAG seven-pass G-type receptor 3 (flamingo homolog, Drosophila)', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108826045, end_position_on_the_genomic_accession=108852969, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  107935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107935, CurrentID=0, Status='live', Symbol='Idd9.1', Aliases='', description='insulin dependent diabetes susceptibility 9.1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107936, CurrentID=0, Status='live', Symbol='Idd9.2', Aliases='', description='insulin dependent diabetes susceptibility 9.2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107937, CurrentID=0, Status='live', Symbol='Idd9.3', Aliases='', description='insulin dependent diabetes susceptibility 9.3', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107939, CurrentID=0, Status='live', Symbol='Pom121', Aliases='2610027A18Rik, C80273, mKIAA0618', description='nuclear pore membrane protein 121', other_designations='nucleoporin Nup121|pore membrane protein of 121 kDa', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135376140, end_position_on_the_genomic_accession=135404733, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  107944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107944, CurrentID=0, Status='live', Symbol='Yaa', Aliases='Tp(X;Y)1Ekw', description='accelerated autoimmunity and lymphoproliferation transposition', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239017, CurrentID=0, Status='live', Symbol='Ogdhl', Aliases='', description='oxoglutarate dehydrogenase-like', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32322012, end_position_on_the_genomic_accession=32348150, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  107951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107951, CurrentID=0, Status='live', Symbol='Cdk9', Aliases='PITALRE', description='cyclin-dependent kinase 9 (CDC2-related kinase)', other_designations='cell division protein kinase 9', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32705782, end_position_on_the_genomic_accession=32712784, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  239027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239027, CurrentID=0, Status='live', Symbol='Arhgap22', Aliases='B230341L19Rik, RHOGAP2, p68RacGAP', description='Rho GTPase activating protein 22', other_designations='rho-type GTPase-activating protein 22', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33214003, end_position_on_the_genomic_accession=33369936, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  239029 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239029, CurrentID=0, Status='live', Symbol='Antxrl', Aliases='1700112N15Rik, 4933430J11', description='anthrax toxin receptor-like', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=34052713, end_position_on_the_genomic_accession=34076323, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  239037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239037, CurrentID=0, Status='live', Symbol='Lrit1', Aliases='BC032270, Lrrc21', description='leucine-rich repeat, immunoglobulin-like and transmembrane domains 1', other_designations='leucine rich repeat containing 21|leucine-rich repeat-containing protein 21|photoreceptor-associated LRR superfamily protein|retina-specific protein PAL', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=37054830, end_position_on_the_genomic_accession=37064938, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  239038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239038, CurrentID=0, Status='live', Symbol='Lrit2', Aliases='A930010E21Rik, Lrrc22', description='leucine-rich repeat, immunoglobulin-like and transmembrane domains 2', other_designations='leucine rich repeat containing 22|leucine-rich repeat-containing protein 22', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=37068049, end_position_on_the_genomic_accession=37073735, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  107970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107970, CurrentID=0, Status='live', Symbol='Hist1h1t', Aliases='H1.6, H1ft, H1t', description='histone cluster 1, H1t', other_designations='H1 histone family, member T (testis-specific)|histone 1, H1t|testicular H1 histone', map_location='13 A3.2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23695811, end_position_on_the_genomic_accession=23696545, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  107971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107971, CurrentID=0, Status='live', Symbol='Frs3', Aliases='4930417B13Rik, AI449674, Frs2beta, Snt2', description='fibroblast growth factor receptor substrate 3', other_designations='FGFR substrate 3|FRS2-beta|suc1-associated neurotrophic factor target 2 (FGFR signalling adaptor)', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47687296, end_position_on_the_genomic_accession=47704286, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  107975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107975, CurrentID=0, Status='live', Symbol='Pacs1', Aliases='AI325977', description='phosphofurin acidic cluster sorting protein 1', other_designations='PACS-1|cytosolic sorting protein PACS-1', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5133685, end_position_on_the_genomic_accession=5273302, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  107976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107976, CurrentID=0, Status='live', Symbol='Bre', Aliases='6030405P19Rik, AI429776, B830038C02Rik', description='brain and reproductive organ-expressed protein', other_designations='BRCA1/BRCA2-containing complex subunit 45|bre', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31697945, end_position_on_the_genomic_accession=32084948, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  107980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107980, CurrentID=0, Status='live', Symbol='Ifaprc2', Aliases='Marc2, RC2', description='intermediate filament-associated protein RC2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  107986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107986, CurrentID=0, Status='live', Symbol='Ddb2', Aliases='2610043A19Rik', description='damage specific DNA binding protein 2', other_designations='damage-specific DNA-binding protein 2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91211579, end_position_on_the_genomic_accession=91237132, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  107993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107993, CurrentID=0, Status='live', Symbol='Bfsp2', Aliases='AI448593, CP49', description='beaded filament structural protein 2, phakinin', other_designations='49 kDa cytoskeletal protein|lens fiber cell beaded filament protein CP 49', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=103424924, end_position_on_the_genomic_accession=103480414, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  107995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107995, CurrentID=0, Status='live', Symbol='Cdc20', Aliases='2310042N09Rik, C87100, p55CDC', description='cell division cycle 20', other_designations='cell division cycle 20 homolog|mmCdc20', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118432901, end_position_on_the_genomic_accession=118437343, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  107999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107999, CurrentID=0, Status='live', Symbol='Gtpbp6', Aliases='AV119224, Pgbpll, pgpl', description='GTP binding protein 6 (putative)', other_designations='pseudoautosomal GTP binding protein-like, like', map_location='5 E', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110103976, end_position_on_the_genomic_accession=110108197, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  108000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108000, CurrentID=0, Status='live', Symbol='Cenpf', Aliases='6530404A22Rik, AI325968, CENF, Lek1, PRO1779, mitosin', description='centromere protein F', other_designations='centromere autoantigen F|leucine, glutamic acid, lysine family 1 protein', map_location='1 95.03 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=189640606, end_position_on_the_genomic_accession=189688105, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  108005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108005, CurrentID=0, Status='live', Symbol='Igkv3-7', Aliases='IgK21-7, Igk-V21-7', description='immunoglobulin kappa variable 3-7', other_designations='V kappa 21-7|immunoglobulin kappa chain variable 21 (V21)-7', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70607437, end_position_on_the_genomic_accession=70608036, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  239081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239081, CurrentID=0, Status='live', Symbol='Tlr11', Aliases='Gm287', description='toll-like receptor 11', other_designations='', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50357914, end_position_on_the_genomic_accession=50363663, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  239083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239083, CurrentID=0, Status='live', Symbol='Ccnb1ip1', Aliases='Gm288, Hei10, mei4', description='cyclin B1 interacting protein 1', other_designations='cyclin-B1-interacting protein 1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50789249, end_position_on_the_genomic_accession=50795983, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  108012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108012, CurrentID=0, Status='live', Symbol='Ap1s2', Aliases='1500012A13Rik, AI853648, EST1', description='adaptor-related protein complex 1, sigma 2 subunit', other_designations='adapter-related protein complex 1 sigma-1B subunit|adaptor protein complex AP-1 sigma-1B subunit|adaptor-related protein complex 1 subunit sigma-1B|clathrin assembly protein complex 1 sigma-1B small chain|golgi adaptor HA1/AP1 adaptin sigma-1B subunit|sigma 1B subunit of AP-1 clathrin', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=163907377, end_position_on_the_genomic_accession=163933666, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  108013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108013, CurrentID=0, Status='live', Symbol='Celf4', Aliases='A230070D14Rik, BRUNOL-4, Brul4, Brunol4, C130060B05Rik', description='CUGBP, Elav-like family member 4', other_designations='CELF-4|CUG-BP- and ETR-3-like factor 4|RNA-binding protein BRUNOL-4|bruno-like 4, RNA binding protein|bruno-like protein 4', map_location='18 B1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=25477620, end_position_on_the_genomic_accession=25754162, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  632302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=632302, CurrentID=0, Status='live', Symbol='Cyp21a2-ps', Aliases='21OHB, Cyp21-ps1, Cyp21B1, Cyp21a2ps, EG632302, Oh21-2, Cyp21a2-ps', description='cytochrome P450, family 21, subfamily a, polypeptide 2 pseudogene', other_designations='21-OHase B|steroid 21-hydroxylase B', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34719431, end_position_on_the_genomic_accession=34722279, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  108015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108015, CurrentID=0, Status='live', Symbol='Chrnb4', Aliases='Acrb-4, Acrb4', description='cholinergic receptor, nicotinic, beta polypeptide 4', other_designations='nicotinic acetylcholine receptor beta4 subunit', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55028154, end_position_on_the_genomic_accession=55048947, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  108017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108017, CurrentID=0, Status='live', Symbol='Fxyd4', Aliases='0610008I02Rik, AI267073, Chif', description='FXYD domain-containing ion transport regulator 4', other_designations='channel inducing factor|channel-inducing factor', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=117933559, end_position_on_the_genomic_accession=117938510, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  108024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108024, CurrentID=0, Status='live', Symbol='Igkv6-20', Aliases='Igk-V19-20', description='immunoglobulin kappa variable 6-20', other_designations='V kappa 19-20|immunoglobulin kappa chain variable 19 (V19)-20|immunoglobulin kappa chain variable 6-20', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70335841, end_position_on_the_genomic_accession=70336479, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  239099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239099, CurrentID=0, Status='live', Symbol='Homez', Aliases='mKIAA1443', description='homeodomain leucine zipper-encoding gene', other_designations='homeodomain leucine zipper-containing factor', map_location='14 C2-C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54852736, end_position_on_the_genomic_accession=54871352, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  108030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108030, CurrentID=0, Status='live', Symbol='Lin7a', Aliases='AI848705, LIN-7A, MALS-1, TIP-33, Veli, Veli1', description='lin-7 homolog A (C. elegans)', other_designations='lin 7 homolog a|mLin-7|mammalian LIN-7 1|mammalian lin-seven protein 1|veli-1|vertebrate homolog of C. elegans Lin-7 type 1|vertebrate lin-7 homolog 1', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=107271798, end_position_on_the_genomic_accession=107425146, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  108032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108032, CurrentID=0, Status='live', Symbol='Fhe', Aliases='', description='Friend helper virus erythroblastosis susceptibility', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  108037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108037, CurrentID=0, Status='live', Symbol='Shmt2', Aliases='2700043D08Rik, AA408223, AA986903', description='serine hydroxymethyltransferase 2 (mitochondrial)', other_designations='', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127517123, end_position_on_the_genomic_accession=127522444, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  239114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239114, CurrentID=0, Status='live', Symbol='Il17d', Aliases='AI462269, IL-17D, Il27a', description='interleukin 17D', other_designations='interleukin 27A', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=57524829, end_position_on_the_genomic_accession=57543166, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  108043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108043, CurrentID=0, Status='live', Symbol='Chrnb3', Aliases='5730417K16Rik, Acrb3', description='cholinergic receptor, nicotinic, beta polypeptide 3', other_designations='', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27368711, end_position_on_the_genomic_accession=27399730, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  239122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239122, CurrentID=0, Status='live', Symbol='Setdb2', Aliases='Clld8, Gm293, KMT1F', description='SET domain, bifurcated 2', other_designations='Setdb2-Phf11-1-Phf11-2 fusion', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59402011, end_position_on_the_genomic_accession=59440877, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  108052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108052, CurrentID=0, Status='live', Symbol='Slc14a1', Aliases='2610507K20Rik, 3021401A05Rik, UT-B, Utb1', description='solute carrier family 14 (urea transporter), member 1', other_designations='solute carrier family 14 member 1|urea transporter B|urea transporter, erythrocyte', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=78100091, end_position_on_the_genomic_accession=78142119, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  239126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239126, CurrentID=0, Status='live', Symbol='C1qtnf9', Aliases='9130217G22Rik, CTRP9, Ciqtnf9', description='C1q and tumor necrosis factor related protein 9', other_designations='complement C1q tumor necrosis factor-related protein 9', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=60768134, end_position_on_the_genomic_accession=60780869, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  108058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108058, CurrentID=0, Status='live', Symbol='Camk2d', Aliases='2810011D23Rik, 8030469K03Rik, CaMK II, [d]-CaMKII', description='calcium/calmodulin-dependent protein kinase II, delta', other_designations='caM kinase II subunit delta|caMK-II subunit delta', map_location='3 G1-G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=126596358, end_position_on_the_genomic_accession=126846326, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  239133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239133, CurrentID=0, Status='live', Symbol='Dleu7', Aliases='BC038059', description='deleted in lymphocytic leukemia, 7', other_designations='deleted in lymphocytic leukemia 7 homolog', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62276229, end_position_on_the_genomic_accession=62292979, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  108062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108062, CurrentID=0, Status='live', Symbol='Cstf2', Aliases='C630034J23Rik, Cstf64, mir-3112', description="cleavage stimulation factor, 3' pre-RNA subunit 2", other_designations="CF-1 64 kDa subunit|CSTF 64 kDa subunit|alphaCstF-64 variant 4|betaCstF-64 variant 1|betaCstF-64 variant 3|cleavage stimulation factor, 3' pre-RNA subunit 2, 64 kDa", map_location='X E3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=134059139, end_position_on_the_genomic_accession=134086822, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  108067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108067, CurrentID=0, Status='live', Symbol='Eif2b3', Aliases='1190002P15Rik', description='eukaryotic translation initiation factor 2B, subunit 3', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117019399, end_position_on_the_genomic_accession=117086852, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  108068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108068, CurrentID=0, Status='live', Symbol='Grm2', Aliases='4930441L02Rik, Gprc1b, mGluR2, mGluR7', description='glutamate receptor, metabotropic 2', other_designations='G protein coupled receptor, family C, group 1, member B|G protein-coupled receptor, family C, group 1, member B', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106644534, end_position_on_the_genomic_accession=106656109, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  108069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108069, CurrentID=0, Status='live', Symbol='Grm3', Aliases='0710001G23Rik, Gprc1c, mGlu3, mGluR3', description='glutamate receptor, metabotropic 3', other_designations='G protein-coupled receptor, family C, group 1, member C', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=9485236, end_position_on_the_genomic_accession=9725352, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  108071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108071, CurrentID=0, Status='live', Symbol='Grm5', Aliases='6430542K11Rik, AI850523, Glu5R, Gprc1e, mGluR5, mGluR5b', description='glutamate receptor, metabotropic 5', other_designations='G protein coupled receptor, family C, group 1, member E|G protein-coupled receptor GRM5|G protein-coupled receptor, family C, group 1, member E|metabotropic glutamate receptor subtype 5b|metabotropic glutamate receptor type 5', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=87583981, end_position_on_the_genomic_accession=88135063, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  108072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108072, CurrentID=0, Status='live', Symbol='Grm6', Aliases='BC021919, Gm3, Gprc1f, mGluR6, nerg1, nob2, nob3, nob4', description='glutamate receptor, metabotropic 6', other_designations='G protein-coupled receptor GRM6|G protein-coupled receptor, family C, group 1, member F', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50850008, end_position_on_the_genomic_accession=50866208, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  108073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108073, CurrentID=0, Status='live', Symbol='Grm7', Aliases='6330570A01Rik, BB176677, C030018L03, E130018M02Rik, Gpr1g, Gprc1g, mGluR7', description='glutamate receptor, metabotropic 7', other_designations='G protein-coupled receptor GRM7|G protein-coupled receptor, family C, group 1, member G|mGlu7a receptor', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=110645458, end_position_on_the_genomic_accession=111567230, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  108075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108075, CurrentID=0, Status='live', Symbol='Ltbp4', Aliases='2310046A13Rik', description='latent transforming growth factor beta binding protein 4', other_designations='LTBP-4', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27305140, end_position_on_the_genomic_accession=27339709, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  108077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108077, CurrentID=0, Status='live', Symbol='Skiv2l', Aliases='4930534J06Rik, AW214248, Ddx13, SKI, Ski2w', description='superkiller viralicidic activity 2-like (S. cerevisiae)', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34839226, end_position_on_the_genomic_accession=34850204, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  108078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108078, CurrentID=0, Status='live', Symbol='Olr1', Aliases='LOX-1, SR-EI, Scare1', description='oxidized low density lipoprotein (lectin-like) receptor 1', other_designations='lectin-like oxLDL receptor 1|lectin-like oxidized LDL receptor 1|lectin-type oxidized LDL receptor 1|ox-LDL receptor 1|oxidised low density lipoprotein (lectin-like) receptor 1', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129485247, end_position_on_the_genomic_accession=129507165, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  108079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108079, CurrentID=0, Status='live', Symbol='Prkaa2', Aliases='2310008I11Rik, A830082D05, AMPKalpha2', description='protein kinase, AMP-activated, alpha 2 catalytic subunit', other_designations='ACACA kinase|AMP-activated protein kinase alpha-2 variant B|AMPK subunit alpha-2|HMGCR kinase|acetyl-CoA carboxylase kinase|hydroxymethylglutaryl-CoA reductase kinase|prkaa2', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=105029650, end_position_on_the_genomic_accession=105109898, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  108083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108083, CurrentID=0, Status='live', Symbol='Pip4k2b', Aliases='AI848124, PI5P4Kbeta, Pip5k2b, c11', description='phosphatidylinositol-5-phosphate 4-kinase, type II, beta', other_designations='1-phosphatidylinositol 5-phosphate 4-kinase 2-beta|1-phosphatidylinositol-5-phosphate 4-kinase 2-beta|PI(5)P 4-kinase type II beta|PIP4KII-beta|diphosphoinositide kinase 2-beta|phosphatidylinositol 5-phosphate 4-kinase type II beta|phosphatidylinositol-4-phosphate 5-kinase, type II, beta|phosphatidylinositol-5-phosphate 4-kinase type-2 beta|ptdIns(5)P-4-kinase isoform 2-beta', map_location='11 61.07 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97715157, end_position_on_the_genomic_accession=97744704, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  239157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239157, CurrentID=0, Status='live', Symbol='Pnma2', Aliases='A830049P17Rik, mKIAA0883', description='paraneoplastic antigen MA2', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66900682, end_position_on_the_genomic_accession=66920063, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  108086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108086, CurrentID=0, Status='live', Symbol='Rnf216', Aliases='2810055G22Rik, AI647468, AU019462, C86502, F830018F18Rik, TRIAD3, UIP83, Ubce7ip1', description='ring finger protein 216', other_designations='E3 ubiquitin ligase|triad domain-containing protein 3|ubcM4-interacting protein 83|ubiquitin-conjugating enzyme 7-interacting protein 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=142990893, end_position_on_the_genomic_accession=143113061, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  108089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108089, CurrentID=0, Status='live', Symbol='Rnf144a', Aliases='Rnf144, UIP4, Ubce7ip4', description='ring finger protein 144A', other_designations='probable E3 ubiquitin-protein ligase RNF144A|ubcM4-interacting protein 4|ubiquitin-conjugating enzyme 7-interacting protein 4', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=26306794, end_position_on_the_genomic_accession=26415321, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  239167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239167, CurrentID=0, Status='live', Symbol='Synb', Aliases='D930020E02Rik, syncytin-B', description='syncytin b', other_designations='placenta-specific/fusogenic', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69290378, end_position_on_the_genomic_accession=69313532, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  108096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108096, CurrentID=0, Status='live', Symbol='Slco1a5', Aliases='Oatp-3, Oatp3, Slc21a7', description='solute carrier organic anion transporter family, member 1a5', other_designations='organic anion-transporting polypeptide 3|sodium-independent organic anion transporter 3|solute carrier family 21 (organic anion transporter), member 7|solute carrier family 21 member 7', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=142234227, end_position_on_the_genomic_accession=142322981, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  108097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108097, CurrentID=0, Status='live', Symbol='Prkab2', Aliases='5730553K21Rik, AW049591, BB124140', description='protein kinase, AMP-activated, beta 2 non-catalytic subunit', other_designations='AMPK subunit beta-2', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97658212, end_position_on_the_genomic_accession=97673812, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  108098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108098, CurrentID=0, Status='live', Symbol='Med21', Aliases='0610007L03Rik, AI449604, D19234, D6Ertd782e, Srb7, Surb7', description='mediator complex subunit 21', other_designations='RNA polymerase II holoenzyme component SRB7|RNAPII complex component SRB7|SRB7 (supressor of RNA polymerase B) homolog', map_location='6 77.7 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=146642579, end_position_on_the_genomic_accession=146650600, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  108099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108099, CurrentID=0, Status='live', Symbol='Prkag2', Aliases='2410051C13Rik, AAKG, AAKG2, AI854673, H91620p, WPWS', description='protein kinase, AMP-activated, gamma 2 non-catalytic subunit', other_designations='AMPK gamma-2 chain|AMPK subunit gamma-2', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24862735, end_position_on_the_genomic_accession=25101906, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  108100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108100, CurrentID=0, Status='live', Symbol='Baiap2', Aliases='IRSp53, R75030', description='brain-specific angiogenesis inhibitor 1-associated protein 2', other_designations='BAI-associated protein 2|BAI1-associated protein 2|insulin receptor substrate p53|insulin receptor substrate protein of 53 kDa|insulin receptor tyrosine kinase 53 kDa substrate', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=119942869, end_position_on_the_genomic_accession=120006782, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  108101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108101, CurrentID=0, Status='live', Symbol='Fermt3', Aliases='C79673, Kindlin3', description='fermitin family member 3', other_designations='unc-112-related protein 2', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6998958, end_position_on_the_genomic_accession=7019469, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  108105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108105, CurrentID=0, Status='live', Symbol='B3gnt5', Aliases='beta3GnT5', description='UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 5', other_designations='BGnT-5|UDP-GlcNAc:beta-Gal beta-1,3-N-acetylglucosaminyltransferase 5|beta-1,3-Gn-T5|beta-1,3-N-acetylglucosaminyltransferase 5|beta3Gn-T5|lactotriaosylceramide synthase|lc(3)Cer synthase|lc3 synthase', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19760192, end_position_on_the_genomic_accession=19772758, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  108109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108109, CurrentID=0, Status='live', Symbol='Fv5', Aliases='Fv-5', description='Friend virus P-induced early anemia, polycythemia susceptibility', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  108112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108112, CurrentID=0, Status='live', Symbol='Eif4ebp3', Aliases='4e-bp3', description='eukaryotic translation initiation factor 4E binding protein 3', other_designations='eIF4E-binding protein 3', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36664060, end_position_on_the_genomic_accession=36666324, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  108114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108114, CurrentID=0, Status='live', Symbol='Slc22a7', Aliases='NLT, OAT2', description='solute carrier family 22 (organic anion transporter), member 7', other_designations='mOAT2|organic anion transporter 2', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46432185, end_position_on_the_genomic_accession=46438477, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  18019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18019, CurrentID=0, Status='live', Symbol='Nfatc2', Aliases='AI607462, NF-ATc2, NF-ATp, NFAT1, NFAT1-D, Nfatp', description='nuclear factor of activated T cells, cytoplasmic, calcineurin dependent 2', other_designations='NFAT pre-existing subunit|T-cell transcription factor NFAT1|nuclear factor of activated T-cells, cytoplasmic, calcineurin-dependent 2', map_location='2 88.91 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168476410, end_position_on_the_genomic_accession=168601657, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  108116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108116, CurrentID=0, Status='live', Symbol='Slco3a1', Aliases='5830414C08Rik, Anr1, MJAM, OATP-D, R75096, Slc21a11', description='solute carrier organic anion transporter family, member 3a1', other_designations='androgen regulated protein 1|organic anion-transporting polypeptide D|sodium-independent organic anion transporter D|solute carrier family 21 (organic anion transporter), member 11', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=74275418, end_position_on_the_genomic_accession=74554780, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  108121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108121, CurrentID=0, Status='live', Symbol='U2af1', Aliases='2010107D16Rik, 35kDa', description='U2 small nuclear ribonucleoprotein auxiliary factor (U2AF) 1', other_designations='U2 auxiliary factor 35 kDa subunit|U2 small nuclear ribonucleoprotein auxiliary factor (U2AF),35 kDa|U2 snRNP auxiliary factor small subunit', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=31647081, end_position_on_the_genomic_accession=31658764, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  108123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108123, CurrentID=0, Status='live', Symbol='Napg', Aliases='2400003O04Rik, SNARE', description='N-ethylmaleimide sensitive fusion protein attachment protein gamma', other_designations='N-ethylmaleimide-sensitive factor attachment protein gamma|SNAP-gamma', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=62977841, end_position_on_the_genomic_accession=62999450, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  108124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108124, CurrentID=0, Status='live', Symbol='Napa', Aliases='1500039N14Rik, AW209189, RA81, SNAPA, SNARE, a-SNAP, hyh', description='N-ethylmaleimide sensitive fusion protein attachment protein alpha', other_designations='N-ethylmaleimide-sensitive factor attachment protein alpha|SNAP-alpha|hydrocephaly with hop gait', map_location='7 8.74 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16098456, end_position_on_the_genomic_accession=16117975, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  108138 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108138, CurrentID=0, Status='live', Symbol='Xrcc4', Aliases='2310057B22Rik, AW413319, AW545101', description='X-ray repair complementing defective repair in Chinese hamster cells 4', other_designations='X-ray repair cross-complementing protein 4', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=89848904, end_position_on_the_genomic_accession=90089608, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  192786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192786, CurrentID=0, Status='live', Symbol='Rapgef6', Aliases='A530068K01, AI844632, BB085664, C030018K18Rik, PDZ-GEF2, Pdzgef2, RA-GEF-2, Ragef2, mKIAA4052', description='Rap guanine nucleotide exchange factor (GEF) 6', other_designations='PDZ domain containing guanine nucleotide exchange factor (GEF) 2|PDZ domain containing guanine nucleotide exchange factor(GEF)1|rap guanine nucleotide exchange factor 2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=54521754, end_position_on_the_genomic_accession=54699286, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  108143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108143, CurrentID=0, Status='live', Symbol='Taf9', Aliases='2310012M09Rik, AA673500, AU045423, TAFII31, TAFII32, Taf2g', description='TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor', other_designations='RNA polymerase II TBP-associated factor subunit G|TAF9 RNA polymerase II, TATA box binding protein (TBP)-associated factor, 32 kDa|transcription initiation factor TFIID 31 kDa subunit|transcription initiation factor TFIID 32 kDa subunit', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100651343, end_position_on_the_genomic_accession=100656060, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  239217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239217, CurrentID=0, Status='live', Symbol='Kctd12', Aliases='AU046135, AW538430, Pfet1, Pfetin', description='potassium channel tetramerisation domain containing 12', other_designations='potassium channel tetramerization domain containing 12|predominantly fetal expressed T1 domain', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=102976581, end_position_on_the_genomic_accession=102982637, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  108147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108147, CurrentID=0, Status='live', Symbol='Atic', Aliases='2610509C24Rik, AA536954, AW212393', description='5-aminoimidazole-4-carboxamide ribonucleotide formyltransferase/IMP cyclohydrolase', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=71557156, end_position_on_the_genomic_accession=71579403, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  108148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108148, CurrentID=0, Status='live', Symbol='Galnt2', Aliases='AI480629', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 2', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 2|galNAc-T2|polypeptide GalNAc transferase 2|pp-GaNTase 2|ppGaNTase-T2|protein-UDP acetylgalactosaminyltransferase 2', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124231394, end_position_on_the_genomic_accession=124345722, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  108150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108150, CurrentID=0, Status='live', Symbol='Galnt7', Aliases='AI225872', description='UDP-N-acetyl-alpha-D-galactosamine: polypeptide N-acetylgalactosaminyltransferase 7', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 7|galNAc-T7|polypeptide GalNAc transferase 7|pp-GaNTase 7|ppGaNTase-T7|protein-UDP acetylgalactosaminyltransferase 7', map_location='8 B3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=57523825, end_position_on_the_genomic_accession=57653055, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  108151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108151, CurrentID=0, Status='live', Symbol='Sema3d', Aliases='4631426B19Rik', description='sema domain, immunoglobulin domain (Ig), short basic domain, secreted, (semaphorin) 3D', other_designations='', map_location='5 A2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=12383166, end_position_on_the_genomic_accession=12588943, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  108153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108153, CurrentID=0, Status='live', Symbol='Adamts7', Aliases='ADAM-TS7, ADAMTS7B', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 7', other_designations='ADAM-TS 7|ADAMTS-7|COMPase|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 7|metalloprotease', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=90162978, end_position_on_the_genomic_accession=90200102, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  108154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108154, CurrentID=0, Status='live', Symbol='Adamts6', Aliases='5031426K13, A930019D11Rik, ADAM-TS6, b2b1879.1Clo, b2b2029Clo, b2b2182Clo, b2b2187.1Clo, b2b2228Clo', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 6', other_designations='a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 6', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=104287347, end_position_on_the_genomic_accession=104496694, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  108155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108155, CurrentID=0, Status='live', Symbol='Ogt', Aliases='1110038P24Rik, 4831420N21Rik, AI115525l, Ogt', description='O-linked N-acetylglucosamine (GlcNAc) transferase (UDP-N-acetylglucosamine:polypeptide-N-acetylglucosaminyl transferase)', other_designations='O-GlcNAc transferase subunit p110|O-linked N-acetylglucosamine transferase 110 kDa subunit', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101640012, end_position_on_the_genomic_accession=101684351, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  108156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108156, CurrentID=0, Status='live', Symbol='Mthfd1', Aliases='Dcs, E430024A07Rik, Mthfd', description='methylenetetrahydrofolate dehydrogenase (NADP+ dependent), methenyltetrahydrofolate cyclohydrolase, formyltetrahydrofolate synthase', other_designations='C1-THF synthase|C1-tetrahydrofolate synthase', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76255232, end_position_on_the_genomic_accession=76319820, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  108159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108159, CurrentID=0, Status='live', Symbol='Ubxn8', Aliases='D0H8S2298E, Rep-8, Rep8h, Ubxd6', description='UBX domain protein 8', other_designations='UBX domain containing 6|UBX domain-containing protein 6|protein Rep-8|reproduction 8 protein', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=33619586, end_position_on_the_genomic_accession=33643762, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  108160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108160, CurrentID=0, Status='live', Symbol='Fam50a', Aliases='D0HXS9928E, XAP-5', description='family with sequence similarity 50, member A', other_designations='protein XAP-5', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74313033, end_position_on_the_genomic_accession=74320149, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  108161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108161, CurrentID=0, Status='live', Symbol='Fam50b', Aliases='D0H6S2654E, X5L, XAP-5-like', description='family with sequence similarity 50, member B', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=34734828, end_position_on_the_genomic_accession=34747623, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  239250 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239250, CurrentID=0, Status='live', Symbol='Slitrk6', Aliases='4832410J21Rik, Sltk6', description='SLIT and NTRK-like family, member 6', other_designations='neuronal transmembrane protein Slitrk6|slit and trk like gene 6', map_location='14 E3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=110748041, end_position_on_the_genomic_accession=110755205, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  433091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=433091, CurrentID=0, Status='live', Symbol='Pnpla1', Aliases='', description='patatin-like phospholipase domain containing 1', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28857965, end_position_on_the_genomic_accession=28890308, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  239273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239273, CurrentID=0, Status='live', Symbol='Abcc4', Aliases='D630049P08Rik, MOATB, MRP4', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 4', other_designations='', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=118482692, end_position_on_the_genomic_accession=118707620, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  239283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239283, CurrentID=0, Status='live', Symbol='Oxgr1', Aliases='Cysltr3, Gm299, Gpr80, Gpr99, P2Y15', description='oxoglutarate (alpha-ketoglutarate) receptor 1', other_designations='G protein-coupled receptor 80|G protein-coupled receptor 99|G-protein coupled receptor 99|alpha-ketoglutarate receptor 1', map_location='14 E4', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=120019585, end_position_on_the_genomic_accession=120042435, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  108217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108217, CurrentID=0, Status='live', Symbol='Rps19X-ps1', Aliases='pS19X', description='ribosomal protein S19, X-linked pseudogene 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239318, CurrentID=0, Status='live', Symbol='Plcxd3', Aliases='B130016O10Rik', description='phosphatidylinositol-specific phospholipase C, X domain containing 3', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=4375491, end_position_on_the_genomic_accession=4575585, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  239319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239319, CurrentID=0, Status='live', Symbol='Card6', Aliases='D730008L15', description='caspase recruitment domain family, member 6', other_designations='', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=5096920, end_position_on_the_genomic_accession=5116483, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  108263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108263, CurrentID=0, Status='live', Symbol='Eef2-ps1', Aliases='Ef2-ps', description='eukaryotic translation elongation factor 2 pseudogene 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239336, CurrentID=0, Status='live', Symbol='Rxfp3', Aliases='BC053073, GPCR135, Rln3r1, Salpr', description='relaxin family peptide receptor 3', other_designations='G-protein coupled receptor SALPR|RLN3 receptor 1|g protein-coupled receptor SALPR homolog|relaxin 3 receptor 1|relaxin-3/INSL7 receptor 1|somatostatin and angiotensin-like peptide receptor', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=11033717, end_position_on_the_genomic_accession=11037968, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  239337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239337, CurrentID=0, Status='live', Symbol='Adamts12', Aliases='ADAMTS-12, AI605170', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 12', other_designations='ADAM-TS 12|ADAM-TS12|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 12|a disintegrin-like and metalloprotease with thrombospondin type 1 motif, 12|metalloprotease disintegrin 12', map_location='15 A1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=11064790, end_position_on_the_genomic_accession=11346867, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  239364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239364, CurrentID=0, Status='live', Symbol='Tspyl5', Aliases='6330530B20, E130308C19Rik', description='testis-specific protein, Y-encoded-like 5', other_designations='TSPY-like protein 5', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=33683875, end_position_on_the_genomic_accession=33687883, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  108293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108293, CurrentID=0, Status='live', Symbol='Nzv1', Aliases='Nxv-1, Nxv1, Nzv-1', description='NZB virus 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  108294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108294, CurrentID=0, Status='live', Symbol='Nzv2', Aliases='Nxv-2, Nxv2, Nzv-2', description='NZB virus 2', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  18049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18049, CurrentID=0, Status='live', Symbol='Ngf', Aliases='Ngfb', description='nerve growth factor', other_designations='beta-NGF|nerve growth factor, beta', map_location='3 45.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102469919, end_position_on_the_genomic_accession=102521013, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  239368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239368, CurrentID=0, Status='live', Symbol='Erich5', Aliases='', description='glutamate rich 5', other_designations='', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34453312, end_position_on_the_genomic_accession=34473892, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  108314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108314, CurrentID=0, Status='live', Symbol='Tsp1', Aliases='MTS', description='tumor suppressor region 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239393, CurrentID=0, Status='live', Symbol='Lrp12', Aliases='AI848829, C820005L12Rik, LRP-12', description='low density lipoprotein-related protein 12', other_designations='', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39870602, end_position_on_the_genomic_accession=39944009, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  108328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108328, CurrentID=0, Status='live', Symbol='Mtv4', Aliases='Mtv-4', description='mammary tumor virus locus 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239405, CurrentID=0, Status='live', Symbol='Rspo2', Aliases='2610028F08Rik, AA673245, D430027K22, ftls', description='R-spondin 2', other_designations='R-spondin 2 homolog|cristin-2|cysteine-rich and single thrombospondin domain-containing protein 2|mCristin-2|roof plate-specific spondin-2', map_location='15 16.73 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=43020788, end_position_on_the_genomic_accession=43171595, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  239408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239408, CurrentID=0, Status='live', Symbol='Tmem74', Aliases='AA549547, B230382K22Rik', description='transmembrane protein 74', other_designations='', map_location='15 B3.2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=43866695, end_position_on_the_genomic_accession=43870612, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  108341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108341, CurrentID=0, Status='live', Symbol='Mtv42', Aliases='Mtv-42', description='mammary tumor virus locus 42', other_designations='', map_location='', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  108342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108342, CurrentID=0, Status='live', Symbol='Mtv50', Aliases='Mtv-50', description='mammary tumor virus locus 50', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  105439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105439, CurrentID=0, Status='live', Symbol='Slain1', Aliases='9630044O09Rik, AA675320, AW742596', description='SLAIN motif family, member 1', other_designations='', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=103650071, end_position_on_the_genomic_accession=103704912, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  239420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239420, CurrentID=0, Status='live', Symbol='Csmd3', Aliases='4930500N14Rik, mKIAA1894', description='CUB and Sushi multiple domains 3', other_designations='CUB and sushi multiple domains protein 3', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=47580633, end_position_on_the_genomic_accession=48792231, orientation='minus', exon_count=72, OMIM='', no_hdr0=''),
  108357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108357, CurrentID=0, Status='live', Symbol='Olfr37', Aliases='mOR37A', description='olfactory receptor 37 cluster', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239435, CurrentID=0, Status='live', Symbol='Aard', Aliases='A5D3, AV328152', description='alanine and arginine rich domain containing protein', other_designations='', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=52040107, end_position_on_the_genomic_accession=52045722, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  239436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239436, CurrentID=0, Status='live', Symbol='Slc30a8', Aliases='C820002P14Rik, ZnT-8, ZnT8', description='solute carrier family 30 (zinc transporter), member 8', other_designations='solute carrier family 30 member 8', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=52295553, end_position_on_the_genomic_accession=52335733, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  108366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108366, CurrentID=0, Status='live', Symbol='Mtv48', Aliases='Mtv-44, Mtv-48', description='mammary tumor virus locus 48', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  108368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108368, CurrentID=0, Status='live', Symbol='Mtv51', Aliases='Mtv-51', description='mammary tumor virus locus 51', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239447, CurrentID=0, Status='live', Symbol='Colec10', Aliases='CL-L1', description='collectin sub-family member 10', other_designations='collectin liver 1|collectin liver protein 1|collectin-L1', map_location='15 21.15 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=54410774, end_position_on_the_genomic_accession=54466360, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  632687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=632687, CurrentID=0, Status='live', Symbol='March10', Aliases='4933417C16Rik, AV039223, OTTMUSG00000002847, Rnf190', description='membrane-associated ring finger (C3HC4) 10', other_designations='ring finger protein 190', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=105360798, end_position_on_the_genomic_accession=105456863, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  632708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=632708, CurrentID=0, Status='live', Symbol='Dmrtc1b', Aliases='Dmrt8.2, Gm714', description='DMRT-like family C1b', other_designations='doublesex and mab-3 related transcription factor 8.2', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=102707881, end_position_on_the_genomic_accession=102715354, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  239510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239510, CurrentID=0, Status='live', Symbol='Phf20l1', Aliases='CGI-72, E130113K22Rik', description='PHD finger protein 20-like 1', other_designations='tudor domain-containing protein PHF20L1', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=66577517, end_position_on_the_genomic_accession=66647043, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  103995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=103995, CurrentID=0, Status='live', Symbol='Hist1-1s', Aliases='', description='H1.S histone', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239528, CurrentID=0, Status='live', Symbol='Ago2', Aliases='1110029L17Rik, 2310051F07Rik, AI225898, AL022874, AW546247, ENSMUSG00000072493, Eif2c2, Gerp95, Gm10365, mKIAA4215', description='argonaute RISC catalytic subunit 2', other_designations='Piwi/Argonaute family protein meIF2C2|argonaute 2|argonaute RISC catalytic component 2|eukaryotic translation initiation factor 2C, 2|mAgo2', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73101625, end_position_on_the_genomic_accession=73184947, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  239530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239530, CurrentID=0, Status='live', Symbol='Gpr20', Aliases='A430106B11Rik', description='G protein-coupled receptor 20', other_designations='G protein-coupled receptor GPR20', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73694607, end_position_on_the_genomic_accession=73707505, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  239546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239546, CurrentID=0, Status='live', Symbol='Zfp647', Aliases='6030449J21, Es492, Znf250', description='zinc finger protein 647', other_designations='KRAB zinc-finger protein', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76910370, end_position_on_the_genomic_accession=76925499, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  239552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239552, CurrentID=0, Status='live', Symbol='Apol8', Aliases='9830006J20Rik, APOL-II, Apol2', description='apolipoprotein L 8', other_designations='apolipoprotein L, 2', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77745109, end_position_on_the_genomic_accession=77757044, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  239554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239554, CurrentID=0, Status='live', Symbol='Foxred2', Aliases='A430097D04Rik, D15Bwg0759e, Gm304', description='FAD-dependent oxidoreductase domain containing 2', other_designations='', map_location='15 36.92 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77940522, end_position_on_the_genomic_accession=77956771, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  239555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239555, CurrentID=0, Status='live', Symbol='Mief1', Aliases='A230016E22, AI452372, Smcr7l', description='mitochondrial elongation factor 1', other_designations='Smith-Magenis syndrome chromosome region, candidate 7-like|mitochondrial dynamic protein MID51|mitochondrial dynamic protein of 51 kDa homolog|mitochondrial dynamics protein of 51 kDa homolog', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80233940, end_position_on_the_genomic_accession=80253373, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  239556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239556, CurrentID=0, Status='live', Symbol='Cacna1i', Aliases='Ca(V)3.3', description='calcium channel, voltage-dependent, alpha 1I subunit', other_designations='low voltage-activated T-type calcium channel alpha-1 subunit (CACNA1I)|low-voltage-activated calcium channel alpha13.3', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=80287238, end_position_on_the_genomic_accession=80398292, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  239559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239559, CurrentID=0, Status='live', Symbol='A4galt', Aliases='', description='alpha 1,4-galactosyltransferase', other_designations='Gb3 synthase|Gb3/CD77 synthase|UDP-galactose:beta-D-galactosyl-beta1-R 4-alpha-D-galactosyltransferase|alpha-1,4-N-acetylglucosaminyltransferase|alpha-1,4-galactosyltransferase|alpha4Gal-T1|globotriaosylceramide synthase', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83226722, end_position_on_the_genomic_accession=83251774, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  632778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=632778, CurrentID=0, Status='live', Symbol='Erich4', Aliases='EG632778, Gm7092, Kseg1', description='glutamate rich 4', other_designations='kidney specific expressed protein 1', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=25614621, end_position_on_the_genomic_accession=25615892, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  239570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239570, CurrentID=0, Status='live', Symbol='Ttc38', Aliases='9030019H09', description='tetratricopeptide repeat domain 38', other_designations='TPR repeat protein 38', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85832304, end_position_on_the_genomic_accession=85858822, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  239591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239591, CurrentID=0, Status='live', Symbol='Ttll8', Aliases='1700019P01Rik, 4933401B17', description='tubulin tyrosine ligase-like family, member 8', other_designations='TTL domain protein|ttll8|tubulin--tyrosine ligase-like protein 8', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=88889778, end_position_on_the_genomic_accession=88955641, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  239606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239606, CurrentID=0, Status='live', Symbol='Slc2a13', Aliases='6530403A04, A630029G22Rik, AI505012, Gm308', description='solute carrier family 2 (facilitated glucose transporter), member 13', other_designations='h(+)-myo-inositol cotransporter|h(+)-myo-inositol symporter|hmit|solute carrier family 2 member 13', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=91267691, end_position_on_the_genomic_accession=91573261, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  239611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239611, CurrentID=0, Status='live', Symbol='Muc19', Aliases='MUC-19, sld', description='mucin 19', other_designations='gel-forming secreted mucin-19|sublingual apomucin', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=91838326, end_position_on_the_genomic_accession=91934555, orientation='plus', exon_count=44, OMIM='', no_hdr0=''),
  239618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239618, CurrentID=0, Status='live', Symbol='Pdzrn4', Aliases='1110017D07Rik, LNX4, SAMCAP3L, mKIAA3027', description='PDZ domain containing RING finger 4', other_designations='', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=92396810, end_position_on_the_genomic_accession=92771827, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  18091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18091, CurrentID=0, Status='live', Symbol='Nkx2-5', Aliases='Csx, Nkx-2.5, Nkx2.5, tinman', description='NK2 homeobox 5', other_designations='Drosophila NK2 transcription factor related, locus 5|NK2 transcription factor related, locus 5|cardiac-specific homeobox|homeobox protein CSX|homeobox protein NK-2 homolog E', map_location='17 13.6 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26838664, end_position_on_the_genomic_accession=26844994, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  239647 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239647, CurrentID=0, Status='live', Symbol='Pced1b', Aliases='Fam113b', description='PC-esterase domain containing 1B', other_designations='family with sequence similarity 113, member B|protein FAM113B', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97247080, end_position_on_the_genomic_accession=97385696, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  239650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239650, CurrentID=0, Status='live', Symbol='Ccdc184', Aliases='5630400A08, AI836003', description='coiled-coil domain containing 184', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98167806, end_position_on_the_genomic_accession=98170134, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  239652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239652, CurrentID=0, Status='live', Symbol='Zfp641', Aliases='9930016F01Rik, Znf641', description='zinc finger protein 641', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98285590, end_position_on_the_genomic_accession=98296508, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  239659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239659, CurrentID=0, Status='live', Symbol='C1ql4', Aliases='C1qtnf11, CTRP11', description='complement component 1, q subcomponent-like 4', other_designations='C1q and tumor necrosis factor-related protein 11|C1q/TNF-related protein 11|C1qTNF11', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99084086, end_position_on_the_genomic_accession=99090694, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  239667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239667, CurrentID=0, Status='live', Symbol='Dip2b', Aliases='4932422C22, AI317237, AI854602, mKIAA1463', description='disco interacting protein 2 homolog B', other_designations='DIP2 disco-interacting protein 2 homolog B|DIP2 homolog B', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100038318, end_position_on_the_genomic_accession=100219474, orientation='plus', exon_count=40, OMIM='', no_hdr0=''),
  239673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239673, CurrentID=0, Status='live', Symbol='4732456N10Rik', Aliases='Krt90', description='RIKEN cDNA 4732456N10 gene', other_designations='', map_location='15 F2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101552356, end_position_on_the_genomic_accession=101562950, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  18103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18103, CurrentID=0, Status='live', Symbol='Nme2', Aliases='NM23-H2, NM23B, nm23-M2', description='NME/NM23 nucleoside diphosphate kinase 2', other_designations='NDK B|NDP kinase B|P18|expressed in non-metastatic cells 2 protein|histidine protein kinase NDKB|non-metastatic cells 2, protein (NM23B) expressed in|nucleoside-diphosphate kinase 2', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=93949814, end_position_on_the_genomic_accession=93956256, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  239706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239706, CurrentID=0, Status='live', Symbol='Mettl22', Aliases='', description='methyltransferase like 22', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8470753, end_position_on_the_genomic_accession=8490205, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  108637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108637, CurrentID=0, Status='live', Symbol='Snord14c', Aliases='Gm23107, Rnu14, U14', description='small nucleolar RNA, C/D box 14C', other_designations='U14 small nuclear RNA|U14 snRNA', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40803583, end_position_on_the_genomic_accession=40803669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  108645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108645, CurrentID=0, Status='live', Symbol='Mat2b', Aliases='1110064C04Rik, 2410018D16Rik, AI182287, AU022853, MAT-II, MATIIbeta, TGR', description='methionine adenosyltransferase II, beta', other_designations='MAT II beta|methionine adenosyltransferase 2 beta subunit', map_location='11 A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=40679314, end_position_on_the_genomic_accession=40695203, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  239719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239719, CurrentID=0, Status='live', Symbol='Mkl2', Aliases='ENSMUSG00000075401, Gt4-1, MRTF-B, Mrtfb, mKIAA1243', description='MKL/myocardin-like 2', other_designations='gene trap insertion site 4-1|myocardin-related transcription factor B', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13256017, end_position_on_the_genomic_accession=13417529, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  108652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108652, CurrentID=0, Status='live', Symbol='Slc35b3', Aliases='4921526O06Rik, AI428480, CGI-19, PABST2, PAPST2', description='solute carrier family 35, member B3', other_designations='PAPS transporter 2', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38932140, end_position_on_the_genomic_accession=38960875, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  108653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108653, CurrentID=0, Status='live', Symbol='Rimklb', Aliases='4931417E21Rik, 4933426K21Rik, AA097693, AI325948, AU015629, Naags', description='ribosomal modification protein rimK-like family member B', other_designations='N-acetyl-aspartyl-glutamate synthetase B|N-acetyl-aspartylglutamate synthetase B|N-acetylaspartyl-glutamate synthetase B|NAAG synthetase B|beta-citryl-glutamate synthase B|ribosomal protein S6 modification-like protein B', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=122447296, end_position_on_the_genomic_accession=122486532, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  108654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108654, CurrentID=0, Status='live', Symbol='Fam210a', Aliases='4933403F05Rik, AA987140, AV134576, AW491926', description='family with sequence similarity 210, member A', other_designations='uncharacterized protein C18orf19 homolog', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=68260184, end_position_on_the_genomic_accession=68300333, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  108655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108655, CurrentID=0, Status='live', Symbol='Foxp1', Aliases='3110052D19Rik, 4932443N09Rik, AI461938, AW494214', description='forkhead box P1', other_designations='forkhead-related transcription factor 1', map_location='6 E1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=98925342, end_position_on_the_genomic_accession=99523024, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  108657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108657, CurrentID=0, Status='live', Symbol='Rnpepl1', Aliases='1110014H17Rik, AI894167', description='arginyl aminopeptidase (aminopeptidase B)-like 1', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92910825, end_position_on_the_genomic_accession=92920585, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  239731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239731, CurrentID=0, Status='live', Symbol='Rimbp3', Aliases='Gm1759, Gm603, RIM-BP3, mKIAA1666', description='RIMS binding protein 3', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=17208135, end_position_on_the_genomic_accession=17213982, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  108660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108660, CurrentID=0, Status='live', Symbol='Rnf187', Aliases='2410016F01Rik, AB030190, RACO-1', description='ring finger protein 187', other_designations='RING domain AP1 coactivator 1|protein RNF187', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58932288, end_position_on_the_genomic_accession=58938906, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  108664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108664, CurrentID=0, Status='live', Symbol='Atp6v1h', Aliases='0710001F19Rik, AU022349, CGI-11, SFD, SFDalpha, SFDbeta, VMA13', description='ATPase, H+ transporting, lysosomal V1 subunit H', other_designations='ATPase, H+ transporting, lysosomal 50/57kDa, V1 subunit H|H+ transporting ATPase lysosomal V1 subunit H|V-ATPase H subunit|V-ATPase subunit H|vacuolar ATP synthase subunit H|vacuolar ATPase subunit H|vacuolar proton pump H subunit|vacuolar proton pump subunit H', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=5082978, end_position_on_the_genomic_accession=5162549, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  239739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239739, CurrentID=0, Status='live', Symbol='Lamp3', Aliases='1200002D17Rik, Cd208, DC-LAMP, DCLAMP, LAMP, TSC403', description='lysosomal-associated membrane protein 3', other_designations='DC LAMP|DC-lysosome-associated membrane glycoprotein|LAMP-3|dendritic cells-lysosome associated membrane glycoprotein', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19653381, end_position_on_the_genomic_accession=19706365, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  108670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108670, CurrentID=0, Status='live', Symbol='Epsti1', Aliases='2310046K10Rik, 5033415K03Rik, BRESI1', description='epithelial stromal interaction 1 (breast)', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=77904239, end_position_on_the_genomic_accession=78002656, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  239743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239743, CurrentID=0, Status='live', Symbol='Klhl6', Aliases='', description='kelch-like 6', other_designations='', map_location='16 A3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19946492, end_position_on_the_genomic_accession=19983049, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  108672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108672, CurrentID=0, Status='live', Symbol='Zdhhc15', Aliases='6030457O13Rik', description='zinc finger, DHHC domain containing 15', other_designations='DHHC-15|zinc finger DHHC domain-containing protein 15', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=104536969, end_position_on_the_genomic_accession=104671090, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  108673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108673, CurrentID=0, Status='live', Symbol='Ccdc86', Aliases='4933411H20Rik, 6720480F16Rik, D19Ertd678e', description='coiled-coil domain containing 86', other_designations='cyclon|cytokine-induced protein with coiled-coil domain|mCyclon', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10941481, end_position_on_the_genomic_accession=10949266, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  108679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108679, CurrentID=0, Status='live', Symbol='Cops8', Aliases='9430009J09Rik, AA408242, Csn8, Sgn8', description='COP9 signalosome subunit 8', other_designations='COP9 (constitutive photomorphogenic) homolog, subunit 8|COP9 homolog|JAB1-containing signalosome subunit 8|signalosome subunit 8', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=90603425, end_position_on_the_genomic_accession=90613341, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  108682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108682, CurrentID=0, Status='live', Symbol='Gpt2', Aliases='4631422C05Rik, ALT2, AU014768, AU041193, C87201', description='glutamic pyruvate transaminase (alanine aminotransferase) 2', other_designations='GPT 2|glutamate pyruvate transaminase 2|glutamic--alanine transaminase 2|glutamic--pyruvic transaminase 2', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85492474, end_position_on_the_genomic_accession=85527560, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  108686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108686, CurrentID=0, Status='live', Symbol='Ccdc88a', Aliases='A430106J12Rik, AI848406, Ape, C130096N06Rik, C330012F17Rik, D130005J21Rik, Girdin, Giv, Grdn, Hkrp1', description='coiled coil domain containing 88A', other_designations='Akt-phosphorylation enhancer|Hook-related protein 1|coiled-coil domain-containing protein 88A|g alpha-interacting vesicle-associated protein|girders of actin filament', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29373595, end_position_on_the_genomic_accession=29510808, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  108687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108687, CurrentID=0, Status='live', Symbol='Edem2', Aliases='9530090G24Rik, AI327354', description='ER degradation enhancer, mannosidase alpha-like 2', other_designations='', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=155701673, end_position_on_the_genomic_accession=155729475, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  108689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108689, CurrentID=0, Status='live', Symbol='Obfc1', Aliases='0610009H20Rik, 2310057J23Rik, AAF-44, AAF44, AI413458, RPA-32', description='oligonucleotide/oligosaccharide-binding fold containing 1', other_designations='alpha accessory factor 44|alpha-accessory factor of 44 kDa|oligonucleotide/oligosaccharide-binding fold-containing protein 1|replication protein A 32 kDa subunit|suppressor of cdc thirteen homolog', map_location='19 D1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=47501031, end_position_on_the_genomic_accession=47545776, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  239766 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239766, CurrentID=0, Status='live', Symbol='Rtp1', Aliases='Gm1766, Gm604', description='receptor transporter protein 1', other_designations='receptor transporting protein 1', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=23428601, end_position_on_the_genomic_accession=23433960, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  108699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108699, CurrentID=0, Status='live', Symbol='Chn1', Aliases='0610007I19Rik, 0710001E19Rik, 1700112L09Rik, 2900046J01Rik, AI413815, ARHGAP2', description='chimerin 1', other_designations='A-chimaerin|N-chimerin|NC|alpha-chimerin|alpha1 chimaerin|alpha2 chimaerin|alpha3-chimerin|chimaerin|chimerin (chimaerin) 1|rho GTPase-activating protein 2', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73610660, end_position_on_the_genomic_accession=73775367, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  108705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108705, CurrentID=0, Status='live', Symbol='Pttg1ip', Aliases='1810010L20Rik, AI314311, AU018448, C79540', description='pituitary tumor-transforming 1 interacting protein', other_designations='PBF|PTTG-binding factor|pituitary tumor-transforming gene protein-binding factor', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77581767, end_position_on_the_genomic_accession=77598732, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  108707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108707, CurrentID=0, Status='live', Symbol='Fam207a', Aliases='1810008A18Rik, AI303072', description='family with sequence similarity 207, member A', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77486655, end_position_on_the_genomic_accession=77537969, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  239789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239789, CurrentID=0, Status='live', Symbol='Gmnc', Aliases='Gemc1, Gm1778, Gm606', description='geminin coiled-coil domain containing', other_designations='', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=26957235, end_position_on_the_genomic_accession=26991667, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  239790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239790, CurrentID=0, Status='live', Symbol='Ostn', Aliases='Ostc', description='osteocrin', other_designations='musclin', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=27306983, end_position_on_the_genomic_accession=27351209, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  236573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236573, CurrentID=0, Status='live', Symbol='Gbp9', Aliases='', description='guanylate-binding protein 9', other_designations='', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=105076464, end_position_on_the_genomic_accession=105110351, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  108723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108723, CurrentID=0, Status='live', Symbol='Card11', Aliases='0610008L17Rik, 2410011D02Rik, BIMP3, CARMA1', description='caspase recruitment domain family, member 11', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=140872999, end_position_on_the_genomic_accession=141000596, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  239796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239796, CurrentID=0, Status='live', Symbol='Mb21d2', Aliases='1600021P15Rik, A430031N04, C87006', description='Mab-21 domain containing 2', other_designations='mab-21 domain-containing protein 2', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=28826176, end_position_on_the_genomic_accession=28929698, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  18121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18121, CurrentID=0, Status='live', Symbol='Nog', Aliases='', description='noggin', other_designations='', map_location='11 54.34 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=89300638, end_position_on_the_genomic_accession=89302559, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  108735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108735, CurrentID=0, Status='live', Symbol='Sft2d2', Aliases='2010005O13Rik, AI506168, AW561909, cI-45', description='SFT2 domain containing 2', other_designations='SFT2 domain-containing protein 2|sft2d2', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165174341, end_position_on_the_genomic_accession=165194433, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  108737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108737, CurrentID=0, Status='live', Symbol='Oxsr1', Aliases='2210022N24Rik, 2810422B09Rik, AI462649, AW209236, Osr1, mKIAA1101', description='oxidative-stress responsive 1', other_designations='oxidative stress-responsive 1 protein|oxsr1', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119238432, end_position_on_the_genomic_accession=119322710, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  108755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108755, CurrentID=0, Status='live', Symbol='Lyrm2', Aliases='2610208E05Rik', description='LYR motif containing 2', other_designations='novel complex 1 (LYR family) protein', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=32800259, end_position_on_the_genomic_accession=32802268, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  108760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108760, CurrentID=0, Status='live', Symbol='Galnt16', Aliases='5730405L21Rik, AI415388, Galntl1, mpp-GalNAc-T16', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 16', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase-like protein 1|UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase-like 1|galNAc-T-like protein 1|galNAc-T16|polypeptide GalNAc transferase 16|polypeptide GalNAc transferase-like protein 1|polypeptide N-acetylgalactosaminyltransferase-like 1|polypeptide N-acetylgalactosaminyltransferase-like protein 1|pp-GaNTase-like protein 1|protein-UDP acetylgalactosaminyltransferase-like protein 1|putative polypeptide N-acetylgalactosaminyltransferase-like protein 1', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80518313, end_position_on_the_genomic_accession=80603896, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  239833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239833, CurrentID=0, Status='live', Symbol='Lmln', Aliases='5330415H22Rik, AI465495', description='leishmanolysin-like (metallopeptidase M8 family)', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33061871, end_position_on_the_genomic_accession=33127695, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  108767 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108767, CurrentID=0, Status='live', Symbol='Pnrc1', Aliases='5730463C12Rik, B4-2, PRR2, Prol2', description='proline-rich nuclear receptor coactivator 1', other_designations='proline rich 2', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=33245423, end_position_on_the_genomic_accession=33248787, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  239845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239845, CurrentID=0, Status='live', Symbol='Gpr156', Aliases='Gababl', description='G protein-coupled receptor 156', other_designations='GABAB-related G-protein coupled receptor', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=37915999, end_position_on_the_genomic_accession=38007530, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  239849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239849, CurrentID=0, Status='live', Symbol='Cd200r4', Aliases='F630107N04Rik, MCD200RLa', description='CD200 receptor 4', other_designations='CD200 cell surface glycoprotein receptor|CD200 cell surface glycoprotein receptor-like 4|CD200 cell surface glycoprotein receptor-like a|CD200 receptor-like 4|CD200RLa|cell surface glycoprotein OX2 receptor 4', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44801339, end_position_on_the_genomic_accession=44839150, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  239852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239852, CurrentID=0, Status='live', Symbol='Zpld1', Aliases='9430016A21Rik', description='zona pellucida like domain containing 1', other_designations='ZP domain-containing protein 1', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=55224125, end_position_on_the_genomic_accession=55582475, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  239853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239853, CurrentID=0, Status='live', Symbol='Adgrg7', Aliases='9130020O16Rik, Gpr128', description='adhesion G protein-coupled receptor G7', other_designations='G protein-coupled receptor 128|G-protein coupled receptor 128|probable G-protein coupled receptor 128', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=56724609, end_position_on_the_genomic_accession=56796021, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  239857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239857, CurrentID=0, Status='live', Symbol='Cadm2', Aliases='2900078E11Rik, 9330131D06, A830029E02Rik, Igdf4d, Igsf4d, NECL3, SynCAM 2, SynCAM2', description='cell adhesion molecule 2', other_designations='NECL-3|immunoglobulin superfamily member 4D|immunoglobulin superfamily, member 4|nectin-like protein 3|synaptic cell adhesion molecule 2', map_location='16 C1.3-C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=66655421, end_position_on_the_genomic_accession=67621200, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  108797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108797, CurrentID=0, Status='live', Symbol='Mex3b', Aliases='4931439A04Rik, M-04814, Rkhd3, mKIAA2009', description='mex3 RNA binding family member B', other_designations='RING finger and KH domain-containing protein 3|mex3 homolog B|ring finger and KH domain containing 3', map_location='7 47.35 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=82867333, end_position_on_the_genomic_accession=82871576, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  108800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108800, CurrentID=0, Status='live', Symbol='Ston2', Aliases='4933401N24Rik, AI425999, AW544599', description='stonin 2', other_designations='stoned B', map_location='12 D3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=91633009, end_position_on_the_genomic_accession=91786436, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  108802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108802, CurrentID=0, Status='live', Symbol='Calr4', Aliases='4933403L16Rik, AU017583', description='calreticulin 4', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109204779, end_position_on_the_genomic_accession=109255138, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  108803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108803, CurrentID=0, Status='live', Symbol='4933402P03Rik', Aliases='', description='RIKEN cDNA 4933402P03 gene', other_designations='', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69816560, end_position_on_the_genomic_accession=69818487, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  108811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108811, CurrentID=0, Status='live', Symbol='Ccdc122', Aliases='4933415L06Rik', description='coiled-coil domain containing 122', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=77036772, end_position_on_the_genomic_accession=77112265, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  108812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108812, CurrentID=0, Status='live', Symbol='Als2cr12', Aliases='4933405P16Rik, 4933425F06Rik', description='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 12 (human)', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=58658117, end_position_on_the_genomic_accession=58696010, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  108816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108816, CurrentID=0, Status='live', Symbol='4933409K07Rik', Aliases='AV071682, b1402', description='RIKEN cDNA 4933409K07 gene', other_designations='', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  108829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108829, CurrentID=0, Status='live', Symbol='Jmjd1c', Aliases='5430433L24Rik, D630035I23Rik, Jmjdic, TRIP8', description='jumonji domain containing 1C', other_designations='jumonji domain-containing protein 1C', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=67096314, end_position_on_the_genomic_accession=67256326, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  108832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108832, CurrentID=0, Status='live', Symbol='Tmem74b', Aliases='5430405G05Rik', description='transmembrane protein 74B', other_designations='', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151700947, end_position_on_the_genomic_accession=151718253, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  108837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108837, CurrentID=0, Status='live', Symbol='Ibtk', Aliases='5430411K16Rik, AA409502, BTKI, mKIAA1417', description='inhibitor of Bruton agammaglobulinemia tyrosine kinase', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=85687359, end_position_on_the_genomic_accession=85750435, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  108841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108841, CurrentID=0, Status='live', Symbol='Rdh13', Aliases='8430425D21Rik', description='retinol dehydrogenase 13 (all-trans and 9-cis)', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4425665, end_position_on_the_genomic_accession=4445844, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  108853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108853, CurrentID=0, Status='live', Symbol='Mtrf1l', Aliases='9130004K12Rik', description='mitochondrial translational release factor 1-like', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=5811119, end_position_on_the_genomic_accession=5824390, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  108857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108857, CurrentID=0, Status='live', Symbol='Ankhd1', Aliases='1110004O12Rik, 4933432B13Rik, 9130019P20Rik, A530027J04Rik, A630021B20Rik, AA571404, Mask, mFLJ00246, mKIAA1085', description='ankyrin repeat and KH domain containing 1', other_designations='multiple ankyrin repeats, single KH-domain homolog', map_location='18 B2-B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36560603, end_position_on_the_genomic_accession=36658908, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  280287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=280287, CurrentID=0, Status='live', Symbol='Kiss1', Aliases='kisspeptin, metastatin', description='KiSS-1 metastasis-suppressor', other_designations='0610012C01Rik|Kiss1 metastasis-suppressor|kisspeptin-1|metastin', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133321887, end_position_on_the_genomic_accession=133329722, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  239932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239932, CurrentID=0, Status='live', Symbol='Krtap24-1', Aliases='Gm312', description='keratin associated protein 24-1', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88610709, end_position_on_the_genomic_accession=88612279, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  108888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108888, CurrentID=0, Status='live', Symbol='Atad3a', Aliases='2400004H09Rik, Atad3, Tob3, mKIAA1273', description='ATPase family, AAA domain containing 3A', other_designations='AAA-ATPase TOB3', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155740640, end_position_on_the_genomic_accession=155761108, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  108897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108897, CurrentID=0, Status='live', Symbol='Aif1l', Aliases='2810003C17Rik, AI043124, C87647, Iba2', description='allograft inflammatory factor 1-like', other_designations='ionized calcium binding adapter molecule 2|ionized calcium-binding adapter molecule 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31950303, end_position_on_the_genomic_accession=31973442, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  108899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108899, CurrentID=0, Status='live', Symbol='2700081O15Rik', Aliases='AI255177, AI848741, AW742297', description='RIKEN cDNA 2700081O15 gene', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7417625, end_position_on_the_genomic_accession=7425904, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  108900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108900, CurrentID=0, Status='live', Symbol='Fam72a', Aliases='2700049P18Rik, P17', description='family with sequence similarity 72, member A', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=131527989, end_position_on_the_genomic_accession=131539872, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  108902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108902, CurrentID=0, Status='live', Symbol='B4gat1', Aliases='1500032M01Rik, B3gnt1, B3gnt6, BETA3GNT1, iGAT, iGNT', description='beta-1,4-glucuronyltransferase 1', other_designations='I-beta-1,3-N-acetylglucosaminyltransferase|N-acetyllactosaminide beta-1,3-N-acetylglucosaminyltransferase|UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 1|UDP-GlcNAc:betaGal beta-1,3-N-acetylglucosaminyltransferase 6|poly-N-acetyllactosamine extension enzyme', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5038826, end_position_on_the_genomic_accession=5041134, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  108903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108903, CurrentID=0, Status='live', Symbol='Tbcd', Aliases='2310057L06Rik, A030005L14Rik, mKIAA0988', description='tubulin-specific chaperone d', other_designations='beta-tubulin cofactor D|tubulin-folding cofactor D', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121450584, end_position_on_the_genomic_accession=121617170, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  108907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108907, CurrentID=0, Status='live', Symbol='Nusap1', Aliases='2610201A12Rik, AI481307, ANKT, AW547774, BB165529, BM037, LNP, NuSAP, Q0310, SAPL', description='nucleolar and spindle associated protein 1', other_designations='nucleolar protein ANKT', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119618298, end_position_on_the_genomic_accession=119650160, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  108909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108909, CurrentID=0, Status='live', Symbol='Aida', Aliases='2610208M17Rik, BC004835', description='axin interactor, dorsalization associated', other_designations='axin interaction partner and dorsalization antagonist', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=183297060, end_position_on_the_genomic_accession=183324501, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  108911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108911, CurrentID=0, Status='live', Symbol='Rcc2', Aliases='2610510H01Rik, 2610529N02Rik, AA536646, AA675016, Td60, mKIAA1470', description='regulator of chromosome condensation 2', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=140701473, end_position_on_the_genomic_accession=140723220, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  108912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108912, CurrentID=0, Status='live', Symbol='Cdca2', Aliases='2610311M19Rik, AI586158', description='cell division cycle associated 2', other_designations='', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=67673180, end_position_on_the_genomic_accession=67715841, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  239985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239985, CurrentID=0, Status='live', Symbol='Arid1b', Aliases='8030481M12, 9330189K18Rik, AI836955, Ardi1b, B230217J03Rik, mKIAA1235', description='AT rich interactive domain 1B (SWI-like)', other_designations='AT rich interactive domain 1B (Swi1 like)', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=4992755, end_position_on_the_genomic_accession=5347656, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  108927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108927, CurrentID=0, Status='live', Symbol='Lhfp', Aliases='2810489O06Rik, AI194968', description='lipoma HMGIC fusion partner', other_designations='', map_location='3 C', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=53041547, end_position_on_the_genomic_accession=53261679, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  83691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=83691, CurrentID=0, Status='live', Symbol='Crispld1', Aliases='Cocoacrisp, Lcrisp1', description='cysteine-rich secretory protein LCCL domain containing 1', other_designations='LCCL domain containing cysteine-rich secretory protein 1', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=17727030, end_position_on_the_genomic_accession=17766352, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  108934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108934, CurrentID=0, Status='live', Symbol='Smim13', Aliases='2900036K02Rik', description='small integral membrane protein 13', other_designations='UPF0766 protein C6orf228 homolog', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41249844, end_position_on_the_genomic_accession=41276577, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  108937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108937, CurrentID=0, Status='live', Symbol='Rnf169', Aliases='2900057K09Rik', description='ring finger protein 169', other_designations='', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99920254, end_position_on_the_genomic_accession=99980693, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  108943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108943, CurrentID=0, Status='live', Symbol='Trmt10a', Aliases='3110023L08Rik, AA794508, Rg9mtd2, Rnmtd2', description='tRNA methyltransferase 10A', other_designations='RNA (guanine-9-) methyltransferase domain containing 2|RNA (guanine-9-)-methyltransferase domain-containing protein 2', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138143517, end_position_on_the_genomic_accession=138159821, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  108946 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108946, CurrentID=0, Status='live', Symbol='Zzz3', Aliases='3110065C23Rik, 6430567E01Rik, AA408566, AI481057, AW556059, R75514', description='zinc finger, ZZ domain containing 3', other_designations='', map_location='3 H3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=152394486, end_position_on_the_genomic_accession=152462826, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  240023 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240023, CurrentID=0, Status='live', Symbol='Pnldc1', Aliases='Gm313', description='poly(A)-specific ribonuclease (PARN)-like domain containing 1', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12888723, end_position_on_the_genomic_accession=12910224, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  240025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240025, CurrentID=0, Status='live', Symbol='Dact2', Aliases='2900084M21Rik, A630024E20, Dpr2, Frd2, dapper2', description='dishevelled-binding antagonist of beta-catenin 2', other_designations='dapper antagonist of catenin 2|dapper homolog 2, antagonist of beta-catenin|dapper, antagonist of beta-catenin, homolog 2|mDpr2', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=14195230, end_position_on_the_genomic_accession=14204338, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  108954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108954, CurrentID=0, Status='live', Symbol='Ppp1r15b', Aliases='1810033K10Rik, AI606441, C530022L24Rik, CReP', description='protein phosphatase 1, regulatory (inhibitor) subunit 15b', other_designations='constitutive repressor of eIF2alpha phosphorylation', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133131166, end_position_on_the_genomic_accession=133139800, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  240028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240028, CurrentID=0, Status='live', Symbol='Lnpep', Aliases='2010309L07Rik, 4732490P18Rik, CAP, IRAP, PLAP, gp160, vp165', description='leucyl/cystinyl aminopeptidase', other_designations='OTase|cystinyl aminopeptidase|oxytocinase', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17527723, end_position_on_the_genomic_accession=17624489, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  108958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108958, CurrentID=0, Status='live', Symbol='Fam73b', Aliases='5730472N09Rik, AA408683, AI790341, C9orf54, R74766', description='family with sequence similarity 73, member B', other_designations='mFLJ00199 protein|ortholog of family with sequence similarity 73, member B FAM73B', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30364198, end_position_on_the_genomic_accession=30385520, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  108960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108960, CurrentID=0, Status='live', Symbol='Irak2', Aliases='6330415L08Rik, AI649099, IRAK-2', description='interleukin-1 receptor-associated kinase 2', other_designations='interleukin-1 receptor-associated kinase 2c|mu-IRAK-2', map_location='6 52.82 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113638467, end_position_on_the_genomic_accession=113695026, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  108961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108961, CurrentID=0, Status='live', Symbol='E2f8', Aliases='4432406C08Rik, AA410048', description='E2F transcription factor 8', other_designations='E2F-8|E2f family member 8', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48866429, end_position_on_the_genomic_accession=48882142, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  240038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240038, CurrentID=0, Status='live', Symbol='Gm4944', Aliases='EG240038', description='predicted gene 4944', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=22197265, end_position_on_the_genomic_accession=22225659, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  240041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240041, CurrentID=0, Status='live', Symbol='Zfp945', Aliases='A630033E08Rik, C730040L01Rik', description='zinc finger protein 945', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=22846697, end_position_on_the_genomic_accession=22867287, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  240047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240047, CurrentID=0, Status='live', Symbol='Mmp25', Aliases='F730048C11Rik, Leukolysin, MMP20, MT6-MMP', description='matrix metallopeptidase 25', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23629458, end_position_on_the_genomic_accession=23645269, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  108978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108978, CurrentID=0, Status='live', Symbol='4930555G01Rik', Aliases='', description='RIKEN cDNA 4930555G01 gene', other_designations='', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=5050629, end_position_on_the_genomic_accession=5059243, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  240055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240055, CurrentID=0, Status='live', Symbol='Neurl1b', Aliases='C230078M08Rik, EG240055', description='neuralized E3 ubiquitin protein ligase 1B', other_designations='neuralized homolog 1b|neuralized-like protein 1B', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26414965, end_position_on_the_genomic_accession=26446349, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  240057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240057, CurrentID=0, Status='live', Symbol='Syngap1', Aliases='Gm1963, Syngap', description='synaptic Ras GTPase activating protein 1 homolog (rat)', other_designations='neuronal RasGAP|ras GTPase-activating protein SynGAP|synaptic Ras GTPase-activating protein 1|synaptic Ras-GAP 1', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26941408, end_position_on_the_genomic_accession=26972434, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  240058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240058, CurrentID=0, Status='live', Symbol='Cpne5', Aliases='A830083G22Rik', description='copine V', other_designations='copine 5', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29156521, end_position_on_the_genomic_accession=29238118, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  108989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108989, CurrentID=0, Status='live', Symbol='Tpr', Aliases='2610029M07Rik, C77892', description='translocated promoter region, nuclear basket protein', other_designations='NPC-associated intranuclear protein|nuclear pore complex-associated protein Tpr|translocated promoter region and nuclear basket protein', map_location='1 G1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=150392838, end_position_on_the_genomic_accession=150449935, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  240064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240064, CurrentID=0, Status='live', Symbol='Zfp799', Aliases='6030490I01Rik', description='zinc finger protein 799', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32815449, end_position_on_the_genomic_accession=32830261, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  240066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240066, CurrentID=0, Status='live', Symbol='Zfp870', Aliases='AI173952', description='zinc finger protein 870', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32879221, end_position_on_the_genomic_accession=32891603, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  108995 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108995, CurrentID=0, Status='live', Symbol='Tbc1d10c', Aliases='1810062O14Rik, AI428527', description='TBC1 domain family, member 10c', other_designations='EBP50-PDZ interactor of 64 kD', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4184357, end_position_on_the_genomic_accession=4191195, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  240068 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240068, CurrentID=0, Status='live', Symbol='Zfp563', Aliases='D10628, D130054H01, NTfin7, Zfp413', description='zinc finger protein 563', other_designations='Novel zinc finger protein|zinc finger protein NTfin7|zinc finger protein ZFP413', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33089367, end_position_on_the_genomic_accession=33110705, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  633285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=633285, CurrentID=0, Status='live', Symbol='Rbm46', Aliases='EG633285, ENSMUSG00000033882, Gm761', description='RNA binding motif protein 46', other_designations='rbm46', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=82836521, end_position_on_the_genomic_accession=82876533, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  109006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109006, CurrentID=0, Status='live', Symbol='Ciapin1', Aliases='2810413N20Rik, AA617265, AU021794, anamorsin', description='cytokine induced apoptosis inhibitor 1', other_designations='cytokine-induced apoptosis inhibitor 1|fe-S cluster assembly protein DRE2 homolog', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94819818, end_position_on_the_genomic_accession=94838340, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  240084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240084, CurrentID=0, Status='live', Symbol='Cchcr1', Aliases='Hcr', description='coiled-coil alpha-helical rod protein 1', other_designations='HCR (a-helix coiled-coil rod homolog)|alpha-helical coiled-coil rod protein', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35517028, end_position_on_the_genomic_accession=35531015, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  240087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240087, CurrentID=0, Status='live', Symbol='Mdc1', Aliases='6820401C03, AA413496, Nfbd1, mKIAA0170', description='mediator of DNA damage checkpoint 1', other_designations='mediation of DNA damage checkpoint 1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35841490, end_position_on_the_genomic_accession=35859670, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  109019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109019, CurrentID=0, Status='live', Symbol='Nabp1', Aliases='4930434H03Rik, 4930442A21Rik, 4930488J04Rik, 4933440J18Rik, 5830411E10Rik, AI852561, Nbp1, Obfc2a, Ssb2', description='nucleic acid binding protein 1', other_designations='SOSS-B2|nucleic acid-binding protein 1|oligonucleotide/oligosaccharide-binding fold containing 2A|oligonucleotide/oligosaccharide-binding fold-containing protein 2A|sensor of single-strand DNA complex subunit B2|sensor of ssDNA subunit B2|single-stranded DNA-binding protein 2', map_location='1 C1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=51465297, end_position_on_the_genomic_accession=51478458, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  240095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240095, CurrentID=0, Status='live', Symbol='H2-M5', Aliases='CRW2, D130003B22Rik, H-2M5', description='histocompatibility 2, M region locus 5', other_designations='MHC class I like protein M5|MHC classIb protein M5', map_location='17 19.16 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36986855, end_position_on_the_genomic_accession=36989504, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  109032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109032, CurrentID=0, Status='live', Symbol='Sp110', Aliases='5031415C07Rik, 52kDa, 5830484A20Rik, ENSMUSG00000075603, Ifi75, Ipr1', description='Sp110 nuclear body protein', other_designations='interferon-induced protein 75, 52kD|intracellular pathogen resistance 1|intracellular pathogen resistance protein 1', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85570741, end_position_on_the_genomic_accession=85603891, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  109042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109042, CurrentID=0, Status='live', Symbol='Prkcdbp', Aliases='3110015B12Rik, 6330514M23Rik, SRBC', description='protein kinase C, delta binding protein', other_designations='cavin 3|cavin-3|serum deprivation response factor-related gene product that binds to C-kinase', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105480616, end_position_on_the_genomic_accession=105482197, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  240119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240119, CurrentID=0, Status='live', Symbol='St6gal2', Aliases='C230064G14Rik, St6galII, mKIAA1877', description='beta galactoside alpha 2,6 sialyltransferase 2', other_designations='CMP-N-acetylneuraminate-beta-galactosamide-alpha-2,6-sialyltransferase 2|ST6Gal II|alpha 2,6-ST 2|beta-galactoside alpha-2,6-sialyltransferase II|sialyltransferase 2', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55444358, end_position_on_the_genomic_accession=55514581, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  240121 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240121, CurrentID=0, Status='live', Symbol='Fsd1', Aliases='', description='fibronectin type 3 and SPRY domain-containing protein', other_designations='', map_location='17 D', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=55986462, end_position_on_the_genomic_accession=55996883, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  109050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109050, CurrentID=0, Status='live', Symbol='Fam212b', Aliases='6530418L21Rik', description='family with sequence similarity 212, member B', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105704599, end_position_on_the_genomic_accession=105721899, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109052, CurrentID=0, Status='live', Symbol='Krt75', Aliases='4732468K03Rik, AA589387, K6hf, Krt2-6hf, Krtcap1', description='keratin 75', other_designations='CK-75|K75|cytokeratin KRT2-6HF|cytokeratin-75|keratin-6 hair follicle|keratin-75|keratinocyte associated protein 1|mK6hf|type II keratin-K6hf|type-II keratin Kb18', map_location='15 56.91 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101563343, end_position_on_the_genomic_accession=101573904, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  109054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109054, CurrentID=0, Status='live', Symbol='Pfdn4', Aliases='0610025M10Rik, 1810063B06Rik, C-1, C1, Pfd4', description='prefoldin 4', other_designations='', map_location='2 91.91 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=170496428, end_position_on_the_genomic_accession=170519070, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  240131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240131, CurrentID=0, Status='live', Symbol='Lrrc30', Aliases='Gm326', description='leucine rich repeat containing 30', other_designations='', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=67630965, end_position_on_the_genomic_accession=67632723, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  109065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109065, CurrentID=0, Status='live', Symbol='Dnaaf2', Aliases='Ktu, TU54, kintoun, mknt', description='dynein, axonemal assembly factor 2', other_designations='dynein assembly factor 2, axonemal|kintoun', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=69189838, end_position_on_the_genomic_accession=69198429, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  109075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109075, CurrentID=0, Status='live', Symbol='Exosc4', Aliases='1110039I09Rik, 1500001N04Rik, Rrp41', description='exosome component 4', other_designations='exosome complex exonuclease RRP41|ribosomal RNA-processing protein 41', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76327397, end_position_on_the_genomic_accession=76330670, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109077, CurrentID=0, Status='live', Symbol='Ints5', Aliases='1110055N21Rik, mKIAA1698', description='integrator complex subunit 5', other_designations='int5', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=8892987, end_position_on_the_genomic_accession=8897890, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  109079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109079, CurrentID=0, Status='live', Symbol='Sephs1', Aliases='1110046B24Rik, AA589574, AI505014, AW111620, SELD, SPS, SPS1', description='selenophosphate synthetase 1', other_designations='selenium donor protein 1|selenophosphate synthase 1', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=4881564, end_position_on_the_genomic_accession=4910557, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  109082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109082, CurrentID=0, Status='live', Symbol='Fbxw17', Aliases='1110064L07Rik, C85886', description='F-box and WD-40 domain protein 17', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=50417834, end_position_on_the_genomic_accession=50433781, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  18181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18181, CurrentID=0, Status='live', Symbol='Nrf1', Aliases='C87038, D6Ertd415e', description='nuclear respiratory factor 1', other_designations='NRF-1|alpha palindromic-binding protein|alpha-pal', map_location='6 12.47 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30047988, end_position_on_the_genomic_accession=30153458, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  109093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109093, CurrentID=0, Status='live', Symbol='Rars2', Aliases='1500002I10Rik, PRO1992, Rarsl', description='arginyl-tRNA synthetase 2, mitochondrial', other_designations='argRS|arginine--tRNA ligase|arginyl-tRNA synthetase 2, mitochondrial (putative)|probable arginyl-tRNA synthetase, mitochondrial', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=34614941, end_position_on_the_genomic_accession=34660167, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  109095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109095, CurrentID=0, Status='live', Symbol='Rbm15b', Aliases='1810017N16Rik', description='RNA binding motif protein 15B', other_designations='RNA binding motif protein 15 pseudogene|RNA-binding motif protein 15B', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106883985, end_position_on_the_genomic_accession=106887000, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  240168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240168, CurrentID=0, Status='live', Symbol='Rasgrp3', Aliases='BC066069, Gm327', description='RAS, guanyl releasing protein 3', other_designations='', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=75435871, end_position_on_the_genomic_accession=75529054, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  240174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240174, CurrentID=0, Status='live', Symbol='Thada', Aliases='6530405K19, A930021O22, AU019899, BC052885, GITA', description='thyroid adenoma associated', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=84190055, end_position_on_the_genomic_accession=84466261, orientation='minus', exon_count=43, OMIM='', no_hdr0=''),
  109108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109108, CurrentID=0, Status='live', Symbol='Slc30a9', Aliases='2310024J23Rik, AL024256, GAC63, HUEL, znT-9', description='solute carrier family 30 (zinc transporter), member 9', other_designations='GRIP1-associated coactivator 1|GRIP1-associated coactivator 63', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=67306937, end_position_on_the_genomic_accession=67356834, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  109113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109113, CurrentID=0, Status='live', Symbol='Uhrf2', Aliases='2310065A22Rik, AI426270, AW214556, D130071B19Rik, Nirf', description='ubiquitin-like, containing PHD and RING finger domains 2', other_designations='Np95-like ring finger protein|nuclear protein 97|nuclear zinc finger protein Np97|ubiquitin-like PHD and RING finger domain-containing protein 2|ubiquitin-like-containing PHD and RING finger domains protein 2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=30030433, end_position_on_the_genomic_accession=30093729, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  240186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240186, CurrentID=0, Status='live', Symbol='Zfp438', Aliases='9430031G09, 9430091M14Rik, B830013J05Rik', description='zinc finger protein 438', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=5210027, end_position_on_the_genomic_accession=5334962, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  109115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109115, CurrentID=0, Status='live', Symbol='Supt3', Aliases='2310066G22Rik, AI315192, SPT3, SPT3Lh, Supt3', description='suppressor of Ty 3', other_designations='suppressor of Ty 3 homolog', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=44777171, end_position_on_the_genomic_accession=45119294, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  109129 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109129, CurrentID=0, Status='live', Symbol='Mmadhc', Aliases='2010311D03Rik, AI314967', description='methylmalonic aciduria (cobalamin deficiency) cblD type, with homocystinuria', other_designations='likely ortholog of H. sapiens chromosome 2 open reading frame 25 (C2orf25)', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=50279881, end_position_on_the_genomic_accession=50296799, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  109135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109135, CurrentID=0, Status='live', Symbol='Plekha5', Aliases='2810431N21Rik, AI428202, AK129423, AMH-Cre, Ayu21-9, Gt(pU21)9Imeg, Pepp2, Tg(AMH-cre)1Flor', description='pleckstrin homology domain containing, family A member 5', other_designations='Gt(Ayu21)9Imeg|phosphoinositol 3-phosphate-binding protein-2', map_location='6 G2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=140423639, end_position_on_the_genomic_accession=140597110, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  109136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109136, CurrentID=0, Status='live', Symbol='Mmaa', Aliases='2810018E08Rik, AI840684', description='methylmalonic aciduria (cobalamin deficiency) type A', other_designations='', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=79263597, end_position_on_the_genomic_accession=79294973, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  240215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240215, CurrentID=0, Status='live', Symbol='Slc4a9', Aliases='AE4, D630003B07, D630024F24Rik', description='solute carrier family 4, sodium bicarbonate cotransporter, member 9', other_designations='', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36528063, end_position_on_the_genomic_accession=36556272, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  240216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240216, CurrentID=0, Status='live', Symbol='E230025N22Rik', Aliases='EG240216', description='Riken cDNA E230025N22 gene', other_designations='', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36683078, end_position_on_the_genomic_accession=36696091, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  109145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109145, CurrentID=0, Status='live', Symbol='Gins4', Aliases='2810037C03Rik, 4933405K01Rik, Sld5', description='GINS complex subunit 4 (Sld5 homolog)', other_designations='', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=23226610, end_position_on_the_genomic_accession=23237668, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  109151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109151, CurrentID=0, Status='live', Symbol='Chd9', Aliases='1810014J18Rik, 9030205D12Rik, A330063D19Rik, AD013, PRIC320, mKIAA0308', description='chromodomain helicase DNA binding protein 9', other_designations='ATP-dependent helicase CHD9|CHD-9|PPAR-alpha-interacting complex protein 320 kDa|ciprofibrate-bound protein PRIC320|peroxisomal proliferator-activated receptor A-interacting complex 320 kDa protein', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=90828357, end_position_on_the_genomic_accession=91054511, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  109154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109154, CurrentID=0, Status='live', Symbol='Mlec', Aliases='2410014A08Rik, AI850530, ESTM19, X83352, mKIAA0152', description='malectin', other_designations='', map_location='5 55.99 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115142981, end_position_on_the_genomic_accession=115158176, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  109161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109161, CurrentID=0, Status='live', Symbol='Ube2q2', Aliases='3010021M21Rik', description='ubiquitin-conjugating enzyme E2Q family member 2', other_designations='E2 ubiquitin-conjugating enzyme Q2|ubiquitin carrier protein Q2|ubiquitin-conjugating enzyme E2Q (putative) 2|ubiquitin-protein ligase Q2', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55148459, end_position_on_the_genomic_accession=55207529, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  240239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240239, CurrentID=0, Status='live', Symbol='Gpr151', Aliases='C130082O03Rik, GalRL, PGR7, nGPCR-2037', description='G protein-coupled receptor 151', other_designations='G-protein coupled receptor PGR7|GPCR-2037|putative G protein-coupled receptor nGPCR-2037', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=42578020, end_position_on_the_genomic_accession=42579652, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  109168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109168, CurrentID=0, Status='live', Symbol='Atl3', Aliases='4633402C03Rik, 5730596K20Rik, AI465397, AW228836', description='atlastin GTPase 3', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=7493816, end_position_on_the_genomic_accession=7538609, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  109169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109169, CurrentID=0, Status='live', Symbol='Igip', Aliases='6330403M23Rik', description='IgA inducing protein', other_designations='IgA-inducing protein', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36300101, end_position_on_the_genomic_accession=36301478, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  109181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109181, CurrentID=0, Status='live', Symbol='Trip11', Aliases='2610511G22Rik, 3110031G15Rik, 6030460N08Rik, AI450776, GMAP-210, TRIP230', description='thyroid hormone receptor interactor 11', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=101834043, end_position_on_the_genomic_accession=101913649, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  240255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240255, CurrentID=0, Status='live', Symbol='Ythdc2', Aliases='3010002F02Rik, BC037178', description='YTH domain containing 2', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=44827744, end_position_on_the_genomic_accession=44889720, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  240261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240261, CurrentID=0, Status='live', Symbol='Ccdc112', Aliases='8430438M01Rik, AW108467', description='coiled-coil domain containing 112', other_designations='', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46282151, end_position_on_the_genomic_accession=46311955, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  240263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240263, CurrentID=0, Status='live', Symbol='Fem1c', Aliases='2610312A07Rik, 3632443A22Rik, AI646567', description='fem-1 homolog c (C.elegans)', other_designations='FEM1-gamma', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46504606, end_position_on_the_genomic_accession=46525971, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  109205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109205, CurrentID=0, Status='live', Symbol='Sobp', Aliases='2900009C16Rik, 5330439J01Rik, JXC1, jc', description='sine oculis-binding protein homolog (Drosophila)', other_designations='jackson circler protein 1|sobp', map_location='10 22.89 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=43002500, end_position_on_the_genomic_accession=43174588, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  240283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240283, CurrentID=0, Status='live', Symbol='Dmxl1', Aliases='C630007L23Rik, D030030C22', description='Dmx-like 1', other_designations='X-like 1 protein', map_location='18 D1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=49832993, end_position_on_the_genomic_accession=49965473, orientation='plus', exon_count=45, OMIM='', no_hdr0=''),
  109212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109212, CurrentID=0, Status='live', Symbol='Fam64a', Aliases='2610008F03Rik, 6720460F02Rik, AI115087, CATS', description='family with sequence similarity 64, member A', other_designations='CALM-interactor expressed in thymus and spleen homolog|regulator of chromosome segregation protein 1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72042439, end_position_on_the_genomic_accession=72047371, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  109215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109215, CurrentID=0, Status='live', Symbol='Lncbate1', Aliases='3010033K07Rik, Lnc-BATE1', description='brown adipose tissue enriched long non-coding RNA 1', other_designations='', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=108553252, end_position_on_the_genomic_accession=108584677, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  109218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109218, CurrentID=0, Status='live', Symbol='Tmem139', Aliases='A930027H06Rik', description='transmembrane protein 139', other_designations='', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42261940, end_position_on_the_genomic_accession=42264555, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  109222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109222, CurrentID=0, Status='live', Symbol='Rarres1', Aliases='5430417P09Rik, AI662122', description='retinoic acid receptor responder (tazarotene induced) 1', other_designations='', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=67478886, end_position_on_the_genomic_accession=67515523, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  109225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109225, CurrentID=0, Status='live', Symbol='Ms4a7', Aliases='9130422I10Rik, A430103C15Rik, CD20l4, CFFMA', description='membrane-spanning 4-domains, subfamily A, member 7', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11321039, end_position_on_the_genomic_accession=11336146, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  109229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109229, CurrentID=0, Status='live', Symbol='Fam118b', Aliases='2310022O21Rik, 2700018L24Rik, C030004A17Rik, Ppp4r1l', description='family with sequence similarity 118, member B', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35216965, end_position_on_the_genomic_accession=35267812, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  109232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109232, CurrentID=0, Status='live', Symbol='Sccpdh', Aliases='AW214504, C330023F11Rik', description='saccharopine dehydrogenase (putative)', other_designations='probable saccharopine dehydrogenase', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=179668231, end_position_on_the_genomic_accession=179687184, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  109241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109241, CurrentID=0, Status='live', Symbol='Mbd5', Aliases='9430004D19Rik, AA536666, AI426407, C030040A15Rik, Gm1630, OTTMUSG00000012483', description='methyl-CpG binding domain protein 5', other_designations='methyl-CpG-binding protein MBD5', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=48949508, end_position_on_the_genomic_accession=49317070, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  109242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109242, CurrentID=0, Status='live', Symbol='Kif24', Aliases='4933425J19Rik, 9430029L23Rik', description='kinesin family member 24', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=41390748, end_position_on_the_genomic_accession=41464933, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  109245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109245, CurrentID=0, Status='live', Symbol='Lrrc39', Aliases='2010005E21Rik, 9430028I06Rik, Myomasp', description='leucine rich repeat containing 39', other_designations='', map_location='3 G2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116562953, end_position_on_the_genomic_accession=116583134, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  109246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109246, CurrentID=0, Status='live', Symbol='Tspan9', Aliases='6720474K14Rik, 9430079M16Rik, AU018597', description='tetraspanin 9', other_designations='tetraspan NET-5|tspan-9', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127961396, end_position_on_the_genomic_accession=128143598, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  240322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240322, CurrentID=0, Status='live', Symbol='Adamts19', Aliases='4831407I23Rik, AU015154, D230034E10Rik', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 19', other_designations='ADAM-TS 19|ADAM-TS19|ADAMTS-19|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 19', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=58836143, end_position_on_the_genomic_accession=59053682, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  109254 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109254, CurrentID=0, Status='live', Symbol='Adtrp', Aliases='9530008L14Rik', description='androgen dependent TFPI regulating protein', other_designations='androgen-dependent TPF1-regulating protein', map_location='13 A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41763148, end_position_on_the_genomic_accession=41847616, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  240328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240328, CurrentID=0, Status='live', Symbol='F830016B08Rik', Aliases='Ifgga4', description='RIKEN cDNA F830016B08 gene', other_designations='interferon-gamma-inducible GTPase Ifgga4 protein', map_location='18 D3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60257185, end_position_on_the_genomic_accession=60303016, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  240332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240332, CurrentID=0, Status='live', Symbol='Slc6a7', Aliases='AW455934, PROT', description='solute carrier family 6 (neurotransmitter transporter, L-proline), member 7', other_designations='slc6a7|solute carrier family 6 member 7', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60995376, end_position_on_the_genomic_accession=61014199, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  240334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240334, CurrentID=0, Status='live', Symbol='Pcyox1l', Aliases='BC060677, C630049M13', description='prenylcysteine oxidase 1 like', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61696164, end_position_on_the_genomic_accession=61707635, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  109263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109263, CurrentID=0, Status='live', Symbol='Rlf', Aliases='9230110M18Rik, AI195322, MommeD8', description='rearranged L-myc fusion sequence', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=121145373, end_position_on_the_genomic_accession=121215120, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  109264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109264, CurrentID=0, Status='live', Symbol='Me3', Aliases='1700020C08Rik, B230207H15Rik', description='malic enzyme 3, NADP(+)-dependent, mitochondrial', other_designations='NADP-ME', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=89632197, end_position_on_the_genomic_accession=89854359, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  109267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109267, CurrentID=0, Status='live', Symbol='Ssc4d', Aliases='C330016E03Rik, S4D-SRCRB, SRCRB-S4D, Srcrb4d', description='scavenger receptor cysteine rich family, 4 domains', other_designations='four scavenger receptor cysteine-rich domains-containing protein|scavenger receptor cysteine rich domain containing, group B (4 domains)|srcrb4d', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=135960220, end_position_on_the_genomic_accession=135974535, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  109270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109270, CurrentID=0, Status='live', Symbol='Prr5', Aliases='AU043908, Arhgap8, C030017C09Rik, C78947, Protor-1, Protor1', description='proline rich 5 (renal)', other_designations='protein observed with Rictor-1|prr5', map_location='15 E', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84663171, end_position_on_the_genomic_accession=84706409, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  109272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109272, CurrentID=0, Status='live', Symbol='Mybpc1', Aliases='8030451F13Rik', description='myosin binding protein C, slow-type', other_designations='MyBP-C slow|Slow-type C-protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88518279, end_position_on_the_genomic_accession=88605234, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  109275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109275, CurrentID=0, Status='live', Symbol='Actr5', Aliases='AA545173, B430109J19Rik', description='ARP5 actin-related protein 5', other_designations='ARP5 actin-related protein 5 homolog', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158624638, end_position_on_the_genomic_accession=158639211, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  240354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240354, CurrentID=0, Status='live', Symbol='Malt1', Aliases='A630046N12, D430033E09Rik, Pcasp1', description='MALT1 paracaspase', other_designations='malt1|mucosa associated lymphoid tissue lymphoma translocation gene 1|paracaspase|paracaspase-1', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65430969, end_position_on_the_genomic_accession=65478888, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  109284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109284, CurrentID=0, Status='live', Symbol='R3hdm4', Aliases='AI503502, C030046I01Rik', description='R3H domain containing 4', other_designations='R3H domain-containing protein C19orf22 homolog', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79910052, end_position_on_the_genomic_accession=79916941, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  109294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109294, CurrentID=0, Status='live', Symbol='Prex2', Aliases='6230420N16Rik, AI316880, AI553603, C030045D06Rik, D430013K02, Depdc2, P-Rex2', description='phosphatidylinositol-3,4,5-trisphosphate-dependent Rac exchange factor 2', other_designations='DEP domain containing 2|DEP domain-containing protein 2|ptdIns(3,4,5)-dependent Rac exchanger 2', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=10993273, end_position_on_the_genomic_accession=11303683, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  109299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109299, CurrentID=0, Status='live', Symbol='C330006A16Rik', Aliases='AI551216', description='RIKEN cDNA C330006A16 gene', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26136807, end_position_on_the_genomic_accession=26140506, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  18217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18217, CurrentID=0, Status='live', Symbol='Ntsr2', Aliases='NT2R, NTRL', description='neurotensin receptor 2', other_designations='NT-R-2|NTR2|low affinity neurotensin receptor|low-affinity levocabastine-sensitive neurotensin receptor', map_location='12 7.98 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=16653452, end_position_on_the_genomic_accession=16660236, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  109305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109305, CurrentID=0, Status='live', Symbol='Orai1', Aliases='D730049H07Rik, Tmem142a, orai-1', description='ORAI calcium release-activated calcium modulator 1', other_designations='calcium release-activated calcium channel 1|protein orai-1|transmembrane protein 142A', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123015074, end_position_on_the_genomic_accession=123030452, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109314, CurrentID=0, Status='live', Symbol='Prr9', Aliases='A030004J04Rik', description='proline rich 9', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92122204, end_position_on_the_genomic_accession=92123947, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  109323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109323, CurrentID=0, Status='live', Symbol='C1qtnf7', Aliases='5530401N20Rik, 8430425G24Rik, Ctrp7', description='C1q and tumor necrosis factor related protein 7', other_designations='C1q and tumor necrosis factor related protein 5', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43515402, end_position_on_the_genomic_accession=43619126, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  240396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240396, CurrentID=0, Status='live', Symbol='Mex3c', Aliases='A130001D14Rik, BC035207, BM-013, Rkhd2', description='mex3 RNA binding family member C', other_designations='MEX3C variant 1|MEX3C variant 10|MEX3C variant 11|MEX3C variant 2|MEX3C variant 3|MEX3C variant 4|MEX3C variant 5|MEX3C variant 6|MEX3C variant 9|RING finger and KH domain-containing protein 2|RNA-binding protein MEX3C|mex3 homolog C|ring finger and KH domain containing 2', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=73572585, end_position_on_the_genomic_accession=73592578, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109331, CurrentID=0, Status='live', Symbol='Rnf20', Aliases='4833430L21Rik, AW540162, C79397, mKIAA4116', description='ring finger protein 20', other_designations='BRE1-A', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49632011, end_position_on_the_genomic_accession=49656886, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  109332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109332, CurrentID=0, Status='live', Symbol='Cdcp1', Aliases='9030022E12Rik, AA409659, E030027H19Rik', description='CUB domain containing protein 1', other_designations='membrane glycoprotein gp140|transmembrane and associated with src kinases', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=123170805, end_position_on_the_genomic_accession=123216124, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  109333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109333, CurrentID=0, Status='live', Symbol='Pkn2', Aliases='6030436C20Rik, AI507382, PRK2, Prkcl2, Stk7', description='protein kinase N2', other_designations='PKN gamma|protein kinase C-like 2|protein-kinase C-related kinase 2|serine/threonine kinase 7', map_location='3 66.69 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=142790902, end_position_on_the_genomic_accession=142882008, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  240411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240411, CurrentID=0, Status='live', Symbol='Loxhd1', Aliases='1700096C21Rik, 4932417K07, sba', description='lipoxygenase homology domains 1', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=77281949, end_position_on_the_genomic_accession=77442932, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  192986 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=192986, CurrentID=0, Status='live', Symbol='Cyb5d2', Aliases='9330151E16Rik, Gm2', description='cytochrome b5 domain containing 2', other_designations='cytochrome b5 domain-containing protein 2', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72777232, end_position_on_the_genomic_accession=72795839, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  109342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109342, CurrentID=0, Status='live', Symbol='Slc5a10', Aliases='C330021F16Rik, SGLT5', description='solute carrier family 5 (sodium/glucose cotransporter), member 10', other_designations='Na(+)/glucose cotransporter 5|solute carrier family 5 member 10', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61672782, end_position_on_the_genomic_accession=61720811, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  109346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109346, CurrentID=0, Status='live', Symbol='Ankrd39', Aliases='9130416N05Rik, C030004B10Rik', description='ankyrin repeat domain 39', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36538173, end_position_on_the_genomic_accession=36547440, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  109349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109349, CurrentID=0, Status='live', Symbol='Fam163b', Aliases='C630035N08Rik', description='family with sequence similarity 163, member B', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27110379, end_position_on_the_genomic_accession=27142477, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  633640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=633640, CurrentID=0, Status='live', Symbol='Gm7120', Aliases='EG633640', description='predicted gene 7120', other_designations='', map_location='13 D2.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119488039, end_position_on_the_genomic_accession=119610459, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  240427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240427, CurrentID=0, Status='live', Symbol='Setbp1', Aliases='C130092E12, Seb, mKIAA0437', description='SET binding protein 1', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=78750378, end_position_on_the_genomic_accession=79109391, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  109359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109359, CurrentID=0, Status='live', Symbol='Fam175b', Aliases='AA589499, AI853413, Abro1, C430003P19Rik', description='family with sequence similarity 175, member B', other_designations='abraxas brother protein 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=132859225, end_position_on_the_genomic_accession=132885114, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  240442 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240442, CurrentID=0, Status='live', Symbol='Adnp2', Aliases='8430420L05Rik, BC024969, Kiaa0863, Zfp508, mKIAA0863', description='ADNP homeobox 2', other_designations='zinc finger protein 508', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80126315, end_position_on_the_genomic_accession=80152063, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  240444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240444, CurrentID=0, Status='live', Symbol='Kcng2', Aliases='AA591524', description='potassium voltage-gated channel, subfamily G, member 2', other_designations='', map_location='18 E3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=80294544, end_position_on_the_genomic_accession=80365105, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  109375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109375, CurrentID=0, Status='live', Symbol='Dob1', Aliases='Do1', description='dietary obesity 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109376, CurrentID=0, Status='live', Symbol='Lbw5', Aliases='', description='lupus NZB x NZW 5', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109379, CurrentID=0, Status='live', Symbol='Lbw2', Aliases='', description='lupus NZB x NZW 2', other_designations='', map_location='4 57.9 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109386, CurrentID=0, Status='live', Symbol='Tms', Aliases='', description='thermosensitivity', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109403, CurrentID=0, Status='live', Symbol='Etohc3', Aliases='', description='ethanol consumption 3', other_designations='', map_location='9 28.62 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240476, CurrentID=0, Status='live', Symbol='Zfp407', Aliases='6430585N13Rik, Gm333, Gm334, Gm948, ZNF407', description='zinc finger protein 407', other_designations='', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84207701, end_position_on_the_genomic_accession=84591854, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  240479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240479, CurrentID=0, Status='live', Symbol='Fam69c', Aliases='B230399E16Rik', description='family with sequence similarity 69, member C', other_designations='', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84718987, end_position_on_the_genomic_accession=84741731, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  109409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109409, CurrentID=0, Status='live', Symbol='Tbs1', Aliases='', description='tuberculosis severity 1', other_designations='', map_location='3 50.18 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109410, CurrentID=0, Status='live', Symbol='Aliq2', Aliases='Ali2', description='acute lung injury QTL 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109411, CurrentID=0, Status='live', Symbol='Aliq3', Aliases='Ali3', description='acute lung injury QTL 3', other_designations='', map_location='17 50.97 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240505, CurrentID=0, Status='live', Symbol='Cdc42bpg', Aliases='BC046418, MRCKgamma', description='CDC42 binding protein kinase gamma (DMPK-like)', other_designations='CDC42-binding protein kinase gamma|DMPK-like gamma|MRCK gamma|cdc42bpg|myotonic dystrophy kinase-related CDC42-binding kinase gamma|myotonic dystrophy protein kinase-like alpha|myotonic dystrophy protein kinase-like gamma', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6306457, end_position_on_the_genomic_accession=6325655, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  240514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240514, CurrentID=0, Status='live', Symbol='Ccdc85b', Aliases='AI842788, BB115872', description='coiled-coil domain containing 85B', other_designations='delta-interacting protein A homolog|hepatitis delta antigen-interacting protein A homolog', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5453163, end_position_on_the_genomic_accession=5457563, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  109443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109443, CurrentID=0, Status='live', Symbol='Rapop1', Aliases='', description='radiation-induced apoptosis 1', other_designations='', map_location='16 22.4 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109445, CurrentID=0, Status='live', Symbol='Pbd2', Aliases='Cti1', description='peak bone density 2', other_designations='', map_location='13 12.94 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240518, CurrentID=0, Status='live', Symbol='Peli3', Aliases='6030441F14Rik, A930011L17, BC028931', description='pellino 3', other_designations='pellino-3|protein pellino homolog 3', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4929717, end_position_on_the_genomic_accession=4943127, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  109447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109447, CurrentID=0, Status='live', Symbol='Par2', Aliases='', description='pulmonary adenoma resistance 2', other_designations='', map_location='18 44.19 cM', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109449, CurrentID=0, Status='live', Symbol='Bhr5', Aliases='', description='bronchial hyperresponsiveness 5', other_designations='', map_location='6 53.29 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109465, CurrentID=0, Status='live', Symbol='Cms1', Aliases='Cms', description='resistance to Coccidioides immitis 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109467 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109467, CurrentID=0, Status='live', Symbol='Pas2', Aliases='Pas-2', description='pulmonary adenoma susceptibility 2', other_designations='', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109471 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109471, CurrentID=0, Status='live', Symbol='Lore1', Aliases='', description='loss of righting induced by ethanol 1', other_designations='', map_location='1 44.69 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240549, CurrentID=0, Status='live', Symbol='Gm4952', Aliases='EG240549', description='predicted gene 4952', other_designations='', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12599984, end_position_on_the_genomic_accession=12627616, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  109491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109491, CurrentID=0, Status='live', Symbol='Sles1', Aliases='', description='systemic lupus erythmatosus suppressor 1', other_designations='', map_location='17 18.31 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109514, CurrentID=0, Status='live', Symbol='Sle5', Aliases='', description='systemic lupus erythematosus susceptibility 5', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240590, CurrentID=0, Status='live', Symbol='Dmrt3', Aliases='', description='doublesex and mab-3 related transcription factor 3', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=25610247, end_position_on_the_genomic_accession=25623921, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109520, CurrentID=0, Status='live', Symbol='Abhr1', Aliases='', description='antigen-induced bronchial hyperresponsiveness 1', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240595, CurrentID=0, Status='live', Symbol='Kcnv2', Aliases='KV11.1', description='potassium channel, subfamily V, member 2', other_designations='voltage-gated potassium channel subunit Kv8.2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=27322060, end_position_on_the_genomic_accession=27337179, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109540, CurrentID=0, Status='live', Symbol='Re', Aliases='Den, Ri, denuded, riccioli', description='rex', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240613, CurrentID=0, Status='live', Symbol='9930021J03Rik', Aliases='ENSMUSG00000048936, Gm9832, mKIAA2026', description='RIKEN cDNA 9930021J03 gene', other_designations='bromo domain', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29714402, end_position_on_the_genomic_accession=29806106, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  240614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240614, CurrentID=0, Status='live', Symbol='Ranbp6', Aliases='C630001B19', description='RAN binding protein 6', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=29808108, end_position_on_the_genomic_accession=29812974, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  109544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109544, CurrentID=0, Status='live', Symbol='N', Aliases='', description='naked', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109548, CurrentID=0, Status='live', Symbol='Bld', Aliases='', description='blind', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109551, CurrentID=0, Status='live', Symbol='In(15)4H', Aliases='Koa', description='inversion, Chr 15, Harwell 4', other_designations='koala', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109552, CurrentID=0, Status='live', Symbol='Sri', Aliases='2210417O06Rik, 2900070H08Rik, Sor', description='sorcin', other_designations='', map_location='5 3.38 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8046078, end_position_on_the_genomic_accession=8069314, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  109553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109553, CurrentID=0, Status='live', Symbol='Whl', Aliases='', description='wheels', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109555, CurrentID=0, Status='live', Symbol='Pt', Aliases='', description='pintail', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240633, CurrentID=0, Status='live', Symbol='Lipk', Aliases='9930115F20, BC055815, Lipl2', description='lipase, family member K', other_designations='lipase-like abhydrolase domain-containing protein 2|lipase-like, ab-hydrolase domain containing 2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34008254, end_position_on_the_genomic_accession=34047903, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  109565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109565, CurrentID=0, Status='live', Symbol='Sig', Aliases='', description='sightless', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240638, CurrentID=0, Status='live', Symbol='Slc16a12', Aliases='3230401A21, AW210596', description='solute carrier family 16 (monocarboxylic acid transporters), member 12', other_designations='MCT 12|solute carrier family 16 member 12', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34668401, end_position_on_the_genomic_accession=34748269, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  109568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109568, CurrentID=0, Status='live', Symbol='Rn7s6', Aliases='Gm22179, Rn7s-6', description='7S RNA 6', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240641, CurrentID=0, Status='live', Symbol='Kif20b', Aliases='33cex, B130024C23, C330014J10Rik, Mphosph1, magoo', description='kinesin family member 20B', other_designations='M-phase phosphoprotein 1|MPP1|N-6 kinesin|kinesin-related motor interacting with PIN1', map_location='19 C2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34922338, end_position_on_the_genomic_accession=34975750, orientation='plus', exon_count=35, OMIM='', no_hdr0=''),
  109572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109572, CurrentID=0, Status='live', Symbol='Hm', Aliases='', description='hammer toe', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109573, CurrentID=0, Status='live', Symbol='Rw', Aliases='In(5)6H-d, In(5)6H-p, rump-white', description='rump white, inversion', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109575, CurrentID=0, Status='live', Symbol='Tbx10', Aliases='Dc, Tbx13, Tbx7', description='T-box 10', other_designations='T-box 13|T-box protein 10|T-box protein 13|dancer|mmTBX7', map_location='19 3.71 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=3992752, end_position_on_the_genomic_accession=3999512, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  109577 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109577, CurrentID=0, Status='live', Symbol='Ssq', Aliases='', description='sasquatch', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109578, CurrentID=0, Status='live', Symbol='Dbf', Aliases='', description='doublefoot', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109579, CurrentID=0, Status='live', Symbol='Es2', Aliases='Es-2', description='esterase 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109580, CurrentID=0, Status='live', Symbol='Es5', Aliases='Es-5', description='esterase 5', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109581, CurrentID=0, Status='live', Symbol='Es6', Aliases='Es-6', description='esterase 6', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109582, CurrentID=0, Status='live', Symbol='Es7', Aliases='Es-7', description='esterase 7', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109583, CurrentID=0, Status='live', Symbol='Es9', Aliases='Es-9', description='esterase 9', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109584, CurrentID=0, Status='live', Symbol='Far', Aliases='', description='first arch', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109585, CurrentID=0, Status='live', Symbol='Paf', Aliases='', description='patchy fur', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240660, CurrentID=0, Status='live', Symbol='Slc35g1', Aliases='AA590464, D330039I19Rik, Tmem20', description='solute carrier family 35, member G1', other_designations='transmembrane protein 20', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38395980, end_position_on_the_genomic_accession=38405607, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109590, CurrentID=0, Status='live', Symbol='Es3', Aliases='Ee-2, Es-3', description='esterase 3', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240665, CurrentID=0, Status='live', Symbol='Ccnj', Aliases='D430039C20Rik', description='cyclin J', other_designations='', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=40831279, end_position_on_the_genomic_accession=40848572, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  109594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109594, CurrentID=0, Status='live', Symbol='Lmo1', Aliases='Rbtn-1, Rbtn1, Ttg1', description='LIM domain only 1', other_designations='LIM domain only protein 1|LIM only 1|LMO-1|T-cell translocation protein 1|cysteine-rich protein TTG-1', map_location='7 57.21 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109138565, end_position_on_the_genomic_accession=109175343, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  240667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240667, CurrentID=0, Status='live', Symbol='Sec31b', Aliases='Gm341, Sec31l2', description='Sec31 homolog B (S. cerevisiae)', other_designations='SEC31-like 2|SEC31-like protein 2|SEC31-related protein B', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=44516957, end_position_on_the_genomic_accession=44545848, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  109599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109599, CurrentID=0, Status='live', Symbol='Om', Aliases='', description='ovum mutant', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109600, CurrentID=0, Status='live', Symbol='Ve', Aliases='', description='velvet coat', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109603, CurrentID=0, Status='live', Symbol='Es11', Aliases='Es-11', description='esterase 11', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109605, CurrentID=0, Status='live', Symbol='Es13', Aliases='Es-13, Es-21, Es21', description='esterase 13', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109606, CurrentID=0, Status='live', Symbol='Es14', Aliases='Es-14', description='esterase 14', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109608, CurrentID=0, Status='live', Symbol='Esa4', Aliases='Es-17, Es17', description='esterase A4', other_designations='esterase 17', map_location='9', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109609, CurrentID=0, Status='live', Symbol='Es23', Aliases='Es-23, Es-23e, Es-23t-e, Es23-t, Es23', description='esterase 23', other_designations='esterase 23, electrophoretic|esterase 23, temporal', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109610, CurrentID=0, Status='live', Symbol='Es25', Aliases='Es-25', description='esterase 25', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109611, CurrentID=0, Status='live', Symbol='Es26', Aliases='Es-26', description='esterase 26', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109615, CurrentID=0, Status='live', Symbol='Fcr', Aliases='', description='Fc receptor', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109616, CurrentID=0, Status='live', Symbol='Pepc', Aliases='Dip-1, Pep-3, Pep3', description='peptidase C', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240690, CurrentID=0, Status='live', Symbol='St18', Aliases='AV348974, Myt3, NZF-3, Nzf3, mKIAA0535', description='suppression of tumorigenicity 18', other_designations='neural zinc finger protein NZF-3', map_location='1 A1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=6487231, end_position_on_the_genomic_accession=6860940, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  109620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109620, CurrentID=0, Status='live', Symbol='Dsp', Aliases='2300002E22Rik, 5730453H04Rik, AA407887, AA407888, AW109828, DP, rul', description='desmoplakin', other_designations='desmoplakin I/II|desmosomal cytoskeletal connector molecule', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=38151084, end_position_on_the_genomic_accession=38198577, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  109624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109624, CurrentID=0, Status='live', Symbol='Cald1', Aliases='4833423D12Rik, AI195384, AV071549, AW536160, C920027I18Rik', description='caldesmon 1', other_designations='', map_location='6 15.07 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34598500, end_position_on_the_genomic_accession=34775469, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  240697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240697, CurrentID=0, Status='live', Symbol='Mcmdc2', Aliases='6030422M02Rik', description='minichromosome maintenance domain containing 2', other_designations='MCM domain-containing protein C8orf45 homolog', map_location='1 A2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=9906159, end_position_on_the_genomic_accession=9940954, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  109628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109628, CurrentID=0, Status='live', Symbol='Pepb', Aliases='Pep-2, Pep2, Trip-1', description='peptidase B', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109636, CurrentID=0, Status='live', Symbol='Pre2', Aliases='Pre-2, Pre-4', description='prealbumin 2', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109637, CurrentID=0, Status='live', Symbol='Upk1a', Aliases='1110031P12Rik, AU019298, C87192, Tspan21, Upk1', description='uroplakin 1A', other_designations='UP1a|UPIa|UPKa|uroplakin Ia', map_location='7 18.78 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30603092, end_position_on_the_genomic_accession=30612734, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  109641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109641, CurrentID=0, Status='live', Symbol='Ps', Aliases='', description='polysyndactyly', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109648, CurrentID=0, Status='live', Symbol='Npy', Aliases='0710005A05Rik', description='neuropeptide Y', other_designations='', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=49822729, end_position_on_the_genomic_accession=49829507, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  109652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109652, CurrentID=0, Status='live', Symbol='Acy1', Aliases='1110014J22Rik, Acy-1', description='aminoacylase 1', other_designations='N-acyl-L-amino-acid amidohydrolase', map_location='9 57.49 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106432981, end_position_on_the_genomic_accession=106438236, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  240725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240725, CurrentID=0, Status='live', Symbol='Sulf1', Aliases='AI467640, AW121680, mKIAA1077', description='sulfatase 1', other_designations='MSulf-1', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=12691695, end_position_on_the_genomic_accession=12861192, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  240726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240726, CurrentID=0, Status='live', Symbol='Slco5a1', Aliases='4732454B15, A630033C23Rik, AA407151', description='solute carrier organic anion transporter family, member 5A1', other_designations='', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=12866549, end_position_on_the_genomic_accession=12992932, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  109658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109658, CurrentID=0, Status='live', Symbol='Txlna', Aliases='2600010N21Rik, AI425952, AW209241, Il14, Txln', description='taxilin alpha', other_designations='interleukin 14|taxilin Txln', map_location='4 63.26 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129626077, end_position_on_the_genomic_accession=129641084, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  109660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109660, CurrentID=0, Status='live', Symbol='Ctrl', Aliases='0910001G08Rik, 1810004D15Rik, AV005227, Ctra-1, Ctra1, chymopasin, mFLJ00366', description='chymotrypsin-like', other_designations='chymotrypsin A CTRA-1', map_location='8 53.06 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105931994, end_position_on_the_genomic_accession=105933862, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  109661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109661, CurrentID=0, Status='live', Symbol='Ds', Aliases='', description='disorganization', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109663, CurrentID=0, Status='live', Symbol='Hoxc11', Aliases='Hox-3.7', description='homeobox C11', other_designations='homeo box C11|homeobox protein Hox-3.7', map_location='15 58.02 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102954526, end_position_on_the_genomic_accession=102956701, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  109667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109667, CurrentID=0, Status='live', Symbol='Brp1', Aliases='A1, Brp-1', description='brain protein 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109672, CurrentID=0, Status='live', Symbol='Cyb5a', Aliases='0610009N12Rik, Cyb5', description='cytochrome b5 type A (microsomal)', other_designations='cytochrome b-5', map_location='18 57.53 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84851414, end_position_on_the_genomic_accession=84879863, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  109674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109674, CurrentID=0, Status='live', Symbol='Ampd2', Aliases='1200014F01Rik, AI552571, Ampd-2', description='adenosine monophosphate deaminase 2', other_designations='adenosine monophosphate deaminase 2 (isoform L)', map_location='3 46.83 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=108074061, end_position_on_the_genomic_accession=108086666, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  109675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109675, CurrentID=0, Status='live', Symbol='Brp8', Aliases='A8, Brp-8', description='brain protein 8', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109676 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109676, CurrentID=0, Status='live', Symbol='Ank2', Aliases='AI835472, AW491075, Ank-2', description='ankyrin 2, brain', other_designations='ank2|ankyrin-B', map_location='3 56.07 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=126921607, end_position_on_the_genomic_accession=127499870, orientation='minus', exon_count=60, OMIM='', no_hdr0=''),
  109678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109678, CurrentID=0, Status='live', Symbol='Os', Aliases='94-A, 94-K, PlmTgN(Pgk1)1Ddp, PlmTgN(Pgk1)2Ddp', description='oligosyndactylism', other_designations='postimplantation lethal mutation induced by Pgk1 transgene insertion-Dimitrina D. Pravtcheva 1|postimplantation lethal mutation induced by Pgk1 transgene insertion-Dimitrina D. Pravtcheva 2', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240752, CurrentID=0, Status='live', Symbol='Pik3c2b', Aliases='C330011J12Rik, Gm199, PI3K-C2beta, PI3KC2beta', description='phosphoinositide-3-kinase, class 2, beta polypeptide', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133045672, end_position_on_the_genomic_accession=133108688, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  240753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240753, CurrentID=0, Status='live', Symbol='Plekha6', Aliases='9530028L01Rik, AU042621, Pepp3', description='pleckstrin homology domain containing, family A member 6', other_designations='PEPP-3|PH domain-containing family A member 6|phosphoinositol 3-phosphate-binding protein 3|phosphoinositol 3-phosphate-binding protein-3', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133163771, end_position_on_the_genomic_accession=133303435, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  240754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240754, CurrentID=0, Status='live', Symbol='Lax1', Aliases='A530029E09, E430019B13Rik', description='lymphocyte transmembrane adaptor 1', other_designations='linker for activation of X cells|membrane-associated adapter protein LAX', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133679029, end_position_on_the_genomic_accession=133693033, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  240755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240755, CurrentID=0, Status='live', Symbol='4933406M09Rik', Aliases='', description='RIKEN cDNA 4933406M09 gene', other_designations='', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134385940, end_position_on_the_genomic_accession=134391516, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  240756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240756, CurrentID=0, Status='live', Symbol='Klhl12', Aliases='C3ip1', description='kelch-like 12', other_designations='CUL3-interacting protein 1|kelch-like protein C3IP1', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=134455495, end_position_on_the_genomic_accession=134490875, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  109685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109685, CurrentID=0, Status='live', Symbol='Hyal3', Aliases='Hyl3', description='hyaluronoglucosaminidase 3', other_designations='hyal-3|hyaluronidase 3|hyaluronoglucosaminidase-3', map_location='9 58.19 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107581296, end_position_on_the_genomic_accession=107587359, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  109687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109687, CurrentID=0, Status='live', Symbol='Ir2', Aliases='Ir-2', description='immune response 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109689, CurrentID=0, Status='live', Symbol='Arrb1', Aliases='1200006I17Rik, AW208571, G430100A01Rik', description='arrestin, beta 1', other_designations='beta-arrestin1', map_location='7 54.09 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99535486, end_position_on_the_genomic_accession=99606772, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  633979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=633979, CurrentID=0, Status='live', Symbol='Ak9', Aliases='Akd1, Akd2, Gm234, Gm7127', description='adenylate kinase 9', other_designations='adenylate kinase domain containing 1|adenylate kinase domain containing 2', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=41303420, end_position_on_the_genomic_accession=41433568, orientation='plus', exon_count=42, OMIM='', no_hdr0=''),
  109697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109697, CurrentID=0, Status='live', Symbol='Cpa1', Aliases='0910001L12Rik, Cpa', description='carboxypeptidase A1, pancreatic', other_designations='', map_location='6 12.52 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30639221, end_position_on_the_genomic_accession=30645361, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  109698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109698, CurrentID=0, Status='live', Symbol='Dh', Aliases='', description='dominant hemimelia', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109699, CurrentID=0, Status='live', Symbol='Ir5', Aliases='Ir-5', description='immune response 5', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109700, CurrentID=0, Status='live', Symbol='Itga1', Aliases='CD49A, E130012M19Rik, Vla1', description='integrin alpha 1', other_designations='CD49 antigen-like family member A|VLA-1|laminin and collagen receptor', map_location='13 64.61 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=114958079, end_position_on_the_genomic_accession=115101964, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  240776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240776, CurrentID=0, Status='live', Symbol='Kcnt2', Aliases='E330038N15Rik, Slick', description='potassium channel, subfamily T, member 2', other_designations='sodium- and chloride-activated ATP-sensitive potassium channel subunit', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=140245659, end_position_on_the_genomic_accession=140613962, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  109709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109709, CurrentID=0, Status='live', Symbol='Lith1', Aliases='', description='lithogenic gene 1', other_designations='', map_location='2 47.93 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109711, CurrentID=0, Status='live', Symbol='Actn1', Aliases='3110023F10Rika, Actn1', description='actinin, alpha 1', other_designations='F-actin cross-linking protein|alpha actinin 1a|alpha-actinin cytoskeletal isoform|non-muscle alpha-actinin-1', map_location='12 36.49 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80167542, end_position_on_the_genomic_accession=80260483, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  109715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109715, CurrentID=0, Status='live', Symbol='Len1', Aliases='Len-1', description='eye lens protein 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109727, CurrentID=0, Status='live', Symbol='Hst1', Aliases='Hst-1', description='hybrid sterility 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109729, CurrentID=0, Status='live', Symbol='Hst6', Aliases='', description='hybrid sterility 6', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109731, CurrentID=0, Status='live', Symbol='Maob', Aliases='6330414K01Rik, MAO-B', description='monoamine oxidase B', other_designations='monoamine oxidase type B', map_location='X 11.88 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=16709281, end_position_on_the_genomic_accession=16817366, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  109735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109735, CurrentID=0, Status='live', Symbol='War', Aliases='', description='warfarin resistance', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109738, CurrentID=0, Status='live', Symbol='Cat4', Aliases='Anc, Apc-1, Apcat-1, Apcat1, Apyc, Pcs-2, anisocoria', description='dominant cataract 4', other_designations='anterior pyramidal cataract|polar cataract and small eye 2', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109740 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109740, CurrentID=0, Status='live', Symbol='H1', Aliases='H-1', description='histocompatibility 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109741, CurrentID=0, Status='live', Symbol='H3', Aliases='H-3', description='histocompatibility 3', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109742, CurrentID=0, Status='live', Symbol='H7', Aliases='H-7', description='histocompatibility 7', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109743, CurrentID=0, Status='live', Symbol='Hst5', Aliases='', description='hybrid sterility 5', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240816, CurrentID=0, Status='live', Symbol='Rgsl1', Aliases='4930415K13Rik, Gm202, Rgsl2', description='regulator of G-protein signaling like 1', other_designations='regulator of G-protein signaling 1', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153777929, end_position_on_the_genomic_accession=153847342, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  109745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109745, CurrentID=0, Status='live', Symbol='Lsr1', Aliases='Lr-1, Lsr-1', description='listeria resistance', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240819, CurrentID=0, Status='live', Symbol='Teddm1a', Aliases='Epdd1, Teddm1, e9-1', description='transmembrane epididymal protein 1A', other_designations='epididymal protein e9-1|transmembrane epididymal protein 1|transmembrane epididymal protein 1-1', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=153891646, end_position_on_the_genomic_accession=153893060, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  109749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109749, CurrentID=0, Status='live', Symbol='Szf1', Aliases='', description='seizure frequency 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109752, CurrentID=0, Status='live', Symbol='Cat3', Aliases='Cat-3, Vlm', description='dominant cataract 3', other_designations='vacuolated lens with microphthalmia', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109754, CurrentID=0, Status='live', Symbol='Cyb5r3', Aliases='0610016L08Rik, 2500002N19Rik, C85115, Dia-1, Dia1, WU:AL591952.1-001, WU:AL591952.1-002, WU:AL591952.1-003, WU:Cyb5r3', description='cytochrome b5 reductase 3', other_designations='B5R|diaphorase 1 (NADH)|diaphorase-1', map_location='15 39.4 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=83153495, end_position_on_the_genomic_accession=83172208, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  109756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109756, CurrentID=0, Status='live', Symbol='H8', Aliases='H-8', description='histocompatibility 8', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109757, CurrentID=0, Status='live', Symbol='Hya', Aliases='H-Y, Sdma', description='histocompatibility Y', other_designations='Y linked serologically defined male antigen', map_location='Y', chromosome='Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240832, CurrentID=0, Status='live', Symbol='Tor1aip2', Aliases='1110020D10Rik, 15kDa, A130072J07, AA103493, AW060462, AW610675, C77739, Ifrg15, Lull1', description='torsin A interacting protein 2', other_designations='15 kDa interferon-responsive protein|interferon alpha responsive gene, 15 kDa|interferon alpha responsive protein (15 kDa)|lumenal domain like LAP1', map_location='1 G3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=156035664, end_position_on_the_genomic_accession=156068861, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  109765 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109765, CurrentID=0, Status='live', Symbol='Etn2', Aliases='Sd', description='early transposon element insertion site 2', other_designations='early transposon element 2', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240843 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240843, CurrentID=0, Status='live', Symbol='Brinp2', Aliases='6430517E21Rik, BB361039, Fam5b, mKIAA1747', description='bone morphogenic protein/retinoic acid inducible neural-specific 2', other_designations='family with sequence similarity 5, member B|protein FAM5B', map_location='1 H1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=158245265, end_position_on_the_genomic_accession=158356528, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  109776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109776, CurrentID=0, Status='live', Symbol='Act1', Aliases='Act-1', description='actin related gene 1', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109778, CurrentID=0, Status='live', Symbol='Blvra', Aliases='0610006A11Rik, 2500001N03Rik, Blvr', description='biliverdin reductase A', other_designations='BVR A|biliverdin-IX alpha-reductase', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127070629, end_position_on_the_genomic_accession=127097084, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  109781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109781, CurrentID=0, Status='live', Symbol='Ly7', Aliases='Ly-7', description='lymphocyte antigen 7', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109782, CurrentID=0, Status='live', Symbol='Mph1', Aliases='Mph-1', description='macrophage antigen 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109785, CurrentID=0, Status='live', Symbol='Pgm3', Aliases='2810473H05Rik, Agm1, BB187688, C77933, PAGM, Pgm-3', description='phosphoglucomutase 3', other_designations='GlcNAc-P mutase|N-acetylglucosamine-phosphate mutase|PGM 3|acetylglucosamine phosphomutase|phosphoglucomutase-3', map_location='9 E3.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=86542505, end_position_on_the_genomic_accession=86571842, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  109791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109791, CurrentID=0, Status='live', Symbol='Clps', Aliases='2200003J09Rik', description='colipase, pancreatic', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=28558214, end_position_on_the_genomic_accession=28560714, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  109794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109794, CurrentID=0, Status='live', Symbol='Cyx', Aliases='', description='cycloheximide tasting', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109797, CurrentID=0, Status='live', Symbol='Ea6', Aliases='Ea-6, H-6', description='erythrocyte antigen 6', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240873, CurrentID=0, Status='live', Symbol='Tnfsf18', Aliases='Gitrl, Tnlg2a', description='tumor necrosis factor (ligand) superfamily, member 18', other_designations='GITR ligand|glucocorticoid-induced TNF-related ligand|glucocorticoid-induced-tumor necrosis factor receptor ligand|tumor necrosis factor ligand 2a', map_location='1 69.75 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=161494657, end_position_on_the_genomic_accession=161505290, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109803, CurrentID=0, Status='live', Symbol='Gv2', Aliases='Gv-2', description='Gross virus antigen 2', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109805, CurrentID=0, Status='live', Symbol='H16', Aliases='H-16', description='histocompatibility 16', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109806, CurrentID=0, Status='live', Symbol='H18', Aliases='H-18', description='histocompatibility 18', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109808, CurrentID=0, Status='live', Symbol='H21', Aliases='H-21', description='histocompatibility 21', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109809, CurrentID=0, Status='live', Symbol='H23', Aliases='H-23', description='histocompatibility 23', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109811, CurrentID=0, Status='live', Symbol='H42', Aliases='H-42', description='histocompatibility 42', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109813, CurrentID=0, Status='live', Symbol='H45', Aliases='H-45', description='histocompatibility 45', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109814, CurrentID=0, Status='live', Symbol='H46', Aliases='H-4, H-46, H4', description='histocompatibility 46', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109815 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109815, CurrentID=0, Status='live', Symbol='Vimp', Aliases='1500011E07Rik, C78786, H-4, H-47, H4, H47, Sels', description='VCP-interacting membrane protein', other_designations='histocompatibility 47|minor histocompatibility antigen H47', map_location='7 35.49 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66079649, end_position_on_the_genomic_accession=66089405, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  240888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240888, CurrentID=0, Status='live', Symbol='Gpr161', Aliases='Gm208, Gm208Gpr, vl', description='G protein-coupled receptor 161', other_designations='', map_location='1 72.64 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=165281174, end_position_on_the_genomic_accession=165326746, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  109820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109820, CurrentID=0, Status='live', Symbol='Pgc', Aliases='2210410L06Rik, Upg-1, Upg1', description='progastricsin (pepsinogen C)', other_designations='pepsinogen C|urinary pepsinogen 1', map_location='17 23.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=47726842, end_position_on_the_genomic_accession=47734478, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  109821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109821, CurrentID=0, Status='live', Symbol='F11', Aliases='1600027G01Rik, AI503996, Cf11, FXI, PTA', description='coagulation factor XI', other_designations='plasma thromboplastin antecedent', map_location='8 25.14 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45240658, end_position_on_the_genomic_accession=45262048, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  240894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240894, CurrentID=0, Status='live', Symbol='Fmo9', Aliases='4831428F09Rik', description='flavin containing monooxygenase 9', other_designations='', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=166662055, end_position_on_the_genomic_accession=166681845, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  240899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240899, CurrentID=0, Status='live', Symbol='Lrrc52', Aliases='4930413P14Rik', description='leucine rich repeat containing 52', other_designations='BK channel auxiliary gamma subunit LRRC52|BK channel auxilliary gamma subunit LRRC52', map_location='1 H2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167445675, end_position_on_the_genomic_accession=167466780, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  109828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109828, CurrentID=0, Status='live', Symbol='C7', Aliases='A530026G17, Gm1238', description='complement component 7', other_designations='', map_location='15 1.98 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=4988762, end_position_on_the_genomic_accession=5063773, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  109829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109829, CurrentID=0, Status='live', Symbol='Ea4', Aliases='Ea-4', description='erythrocyte antigen 4', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109831, CurrentID=0, Status='live', Symbol='Gv1', Aliases='Gv-1', description='Gross virus antigen 1', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109832, CurrentID=0, Status='live', Symbol='H25', Aliases='H-25', description='histocompatibility 25', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109833, CurrentID=0, Status='live', Symbol='H30', Aliases='H-30', description='histocompatibility 30', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109837, CurrentID=0, Status='live', Symbol='H43', Aliases='H-43', description='histocompatibility 43', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109839, CurrentID=0, Status='live', Symbol='Lvp1', Aliases='Lvp-1', description='major liver protein 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109840, CurrentID=0, Status='live', Symbol='Ly11', Aliases='Ly-11', description='lymphocyte antigen 11', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240913, CurrentID=0, Status='live', Symbol='Adamts4', Aliases='3830423K05, ADAM-TS4, ADAMTS-2, ADMP-1, mKIAA0688', description='a disintegrin-like and metallopeptidase (reprolysin type) with thrombospondin type 1 motif, 4', other_designations='ADAM-TS 4|ADAMTS-4|a disintegrin-like and metalloprotease (reprolysin type) with thrombospondin type 1 motif, 4|aggrecanase-1', map_location='1 79.28 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171250422, end_position_on_the_genomic_accession=171259922, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  240916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240916, CurrentID=0, Status='live', Symbol='Vsig8', Aliases='A030011M19, EG240916', description='V-set and immunoglobulin domain containing 8', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=172555938, end_position_on_the_genomic_accession=172563717, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  109846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109846, CurrentID=0, Status='live', Symbol='Ly31', Aliases='Ly-31, Ly-m31', description='lymphocyte antigen 31', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109849, CurrentID=0, Status='live', Symbol='Zfp4', Aliases='Fnp-2, Zfp-4', description='zinc finger protein 4', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109857, CurrentID=0, Status='live', Symbol='Cbr3', Aliases='1110001J05Rik, C81353', description='carbonyl reductase 3', other_designations='NADPH-dependent carbonyl reductase 3', map_location='16 54.58 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93683219, end_position_on_the_genomic_accession=93690991, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  109858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109858, CurrentID=0, Status='live', Symbol='H61', Aliases='', description='histocompatibility 61', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109863, CurrentID=0, Status='live', Symbol='Ly18', Aliases='Ly-18, Ly-m18', description='lymphocyte antigen 18', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109864, CurrentID=0, Status='live', Symbol='Ly28', Aliases='Ly-28', description='lymphocyte antigen 28', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109869, CurrentID=0, Status='live', Symbol='Tcd1a', Aliases='TRD, Tcd-1, Tcd1, tcs-1, tcs1', description='t-complex distorter 1A', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109870, CurrentID=0, Status='live', Symbol='Tcd3', Aliases='Tcd-3, tcs-3, tcs3', description='t-complex distorter 3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109873, CurrentID=0, Status='live', Symbol='Tthy', Aliases='', description='thymocyte alloantigen', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109875, CurrentID=0, Status='live', Symbol='Clcn4-1', Aliases='Clc4-1', description='chloride channel 4-1', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109879, CurrentID=0, Status='live', Symbol='Akp4', Aliases='Akp-4', description='alkaline phosphatase 4', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109880, CurrentID=0, Status='live', Symbol='Braf', Aliases='9930012E13Rik, AA120551, AA387315, AA473386, B-raf-2, Braf2, C230098H17, C87398, D6Ertd631e, Braf', description='Braf transforming gene', other_designations='B-Raf proto-oncogene serine/threonine-protein kinase (p94)|proto-oncogene B-Raf', map_location='6 18.43 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39603231, end_position_on_the_genomic_accession=39725656, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  171231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171231, CurrentID=0, Status='live', Symbol='Vmn1r230', Aliases='V1re8', description='vomeronasal 1 receptor 230', other_designations='vomernasal 1 receptor Vmn1r230|vomeronasal 1 receptor, E8|vomeronasal type 1 receptor E8', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=20846551, end_position_on_the_genomic_accession=20847501, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  109884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109884, CurrentID=0, Status='live', Symbol='Lyb4', Aliases='Lyb-4', description='B-lymphocyte antigen 4', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  240960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240960, CurrentID=0, Status='live', Symbol='Dnah14', Aliases='A230079K17Rik, Dnahc14, Gm18554, Gm980', description='dynein, axonemal, heavy chain 14', other_designations='dynein, axonemal, heavy chain 14 pseudogene', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=181576562, end_position_on_the_genomic_accession=181815533, orientation='plus', exon_count=87, OMIM='', no_hdr0=''),
  109889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109889, CurrentID=0, Status='live', Symbol='Mzf1', Aliases='Mzf-2, Mzf2, Zfp121, Zfp98, Znf42', description='myeloid zinc finger 1', other_designations='Myeloid zinc finger protein 1|Myeloid zinc finger protein 2|myeloid-specific retinoic acid-responsive zinc finger protein|zinc finger protein 121|zinc finger protein 98', map_location='7 7.73 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13042303, end_position_on_the_genomic_accession=13054764, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  109899 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109899, CurrentID=0, Status='live', Symbol='Akp1', Aliases='Akp-1', description='alkaline phosphatase 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109900, CurrentID=0, Status='live', Symbol='Asl', Aliases='2510006M18Rik', description='argininosuccinate lyase', other_designations='ASAL|arginosuccinase', map_location='5 68.37 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=130011495, end_position_on_the_genomic_accession=130024331, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  109901 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109901, CurrentID=0, Status='live', Symbol='Cela1', Aliases='1810009A17Rik, 1810062B19Rik, Ela-1, Ela1, PC-TsF', description='chymotrypsin-like elastase family, member 1', other_designations='elastase 1, pancreatic|elastase-1', map_location='15 56.38 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100674422, end_position_on_the_genomic_accession=100687920, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  109902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109902, CurrentID=0, Status='live', Symbol='Esl1', Aliases='', description='early lymphoma of SL 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109904, CurrentID=0, Status='live', Symbol='Mcf2', Aliases='B230117G22Rik, Dbl, Mcf-2', description='mcf.2 transforming sequence', other_designations='', map_location='X 33.5 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=60055956, end_position_on_the_genomic_accession=60179187, orientation='minus', exon_count=31, OMIM='', no_hdr0=''),
  109905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109905, CurrentID=0, Status='live', Symbol='Rap1a', Aliases='AI848598, G-22K, Krev-1, Rap1', description='RAS-related protein-1a', other_designations='GTP-binding protein SMG-P21A|ras-related protein Krev-1', map_location='3 46.45 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=105727260, end_position_on_the_genomic_accession=105801381, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  109906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109906, CurrentID=0, Status='live', Symbol='Rig1', Aliases='Rig-1', description='regulation of Igh-1b 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109910, CurrentID=0, Status='live', Symbol='Zfp91', Aliases='9130014I08Rik, A530054C17Rik, AL024263, AW545902, Pzf, Zfp-91', description='zinc finger protein 91', other_designations='penta Zf protein|zinc finger protein 91 homolog', map_location='19 8.73 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12766939, end_position_on_the_genomic_accession=12796123, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  109920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109920, CurrentID=0, Status='live', Symbol='Ccd', Aliases='', description='cleidocranial dysplasia', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109928, CurrentID=0, Status='live', Symbol='Rmcf', Aliases='', description='resistance to MCF virus', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109929 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109929, CurrentID=0, Status='live', Symbol='Zbtb25', Aliases='2810462M08Rik, 2900064P18Rik, AI842128, Kup, Zfp-50, Zfp50', description='zinc finger and BTB domain containing 25', other_designations='zinc finger protein 50', map_location='12 33.73 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76348900, end_position_on_the_genomic_accession=76370218, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  111590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111590, CurrentID=0, Status='live', Symbol='Pmv41', Aliases='Pmv-41', description='polytropic murine leukemia virus 41', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109934 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109934, CurrentID=0, Status='live', Symbol='Abr', Aliases='', description='active BCR-related gene', other_designations='', map_location='11 45.92 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76416732, end_position_on_the_genomic_accession=76577739, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  171242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171242, CurrentID=0, Status='live', Symbol='Vmn1r78', Aliases='V1rg7', description='vomeronasal 1 receptor 78', other_designations='vomernasal 1 receptor Vmn1r78', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12152464, end_position_on_the_genomic_accession=12153405, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  109959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109959, CurrentID=0, Status='live', Symbol='Amy2a5', Aliases='1810008N23Rik, Amy-2, Amy2, mAmy2-2', description='amylase 2a5', other_designations='1,4-alpha-D-glucan glucanohydrolase|PA|amylase 2, pancreatic|amylase 2a5, pancreatic', map_location='3 49.35 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=113249178, end_position_on_the_genomic_accession=113258699, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  241035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241035, CurrentID=0, Status='live', Symbol='Pkhd1', Aliases='AI118496, AI182499, FPC, Tigm1', description='polycystic kidney and hepatic disease 1', other_designations='TIG multiple domains 1|polyductin|tigmin', map_location='1 A2-A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20057779, end_position_on_the_genomic_accession=20618082, orientation='minus', exon_count=70, OMIM='', no_hdr0=''),
  109965 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109965, CurrentID=0, Status='live', Symbol='Ggm1', Aliases='Ggm-1', description='ganglioside expression 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109971, CurrentID=0, Status='live', Symbol='Rua', Aliases='', description='raffinose acetate tasting', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109978, CurrentID=0, Status='live', Symbol='Art4', Aliases='4432404K01Rik, ARTC4, DO, DOK1', description='ADP-ribosyltransferase 4', other_designations='ADP-ribosyltransferase C2 and C3 toxin-like 4|NAD(P)(+)--arginine ADP-ribosyltransferase 4|mono(ADP-ribosyl)transferase 4', map_location='6 66.72 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136848451, end_position_on_the_genomic_accession=136857600, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  109979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109979, CurrentID=0, Status='live', Symbol='Art3', Aliases='4930569O04Rik, ARTC3', description='ADP-ribosyltransferase 3', other_designations='ADP-ribosyltransferase C2 and C3 toxin-like 3|ART3 mono(ADP-ribosyl)transferase|NAD(P)(+)--arginine ADP-ribosyltransferase 3|mono(ADP-ribosyl)transferase 3', map_location='5 46.54 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92331799, end_position_on_the_genomic_accession=92414628, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  109983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109983, CurrentID=0, Status='live', Symbol='Ce2', Aliases='Ce-2', description='catalase activity, kidney', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  109988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=109988, CurrentID=0, Status='live', Symbol='Ri1', Aliases='Ri-1', description='recognition of identity 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241062, CurrentID=0, Status='live', Symbol='Pgap1', Aliases='5033403E17Rik, 9030223K07Rik, A530084K22, D230012E17Rik, oto', description='post-GPI attachment to proteins 1', other_designations='post-GPI attachment to proteins factor 1', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=54473000, end_position_on_the_genomic_accession=54557684, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  241066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241066, CurrentID=0, Status='live', Symbol='Carf', Aliases='Als2cr8, ECBRC-FC1, ECBRC-FC2', description='calcium response factor', other_designations='amyotrophic lateral sclerosis 2 chromosomal region candidate gene 8 protein homolog|calcium-response factor CaRF', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=60098221, end_position_on_the_genomic_accession=60153953, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  241070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241070, CurrentID=0, Status='live', Symbol='Gpr1', Aliases='', description='G protein-coupled receptor 1', other_designations='probable G-protein coupled receptor 1', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63182571, end_position_on_the_genomic_accession=63223278, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  241073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241073, CurrentID=0, Status='live', Symbol='Dytn', Aliases='Gm215', description='dystrotelin', other_designations='', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=63622851, end_position_on_the_genomic_accession=63686927, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  241075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241075, CurrentID=0, Status='live', Symbol='Plekhm3', Aliases='9430067K14Rik, A230102O09Rik, AI449570, DAPR, Plekhm1l', description='pleckstrin homology domain containing, family M, member 3', other_designations='PH domain-containing family M member 3|differentiation associated protein', map_location='1 C2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=64789120, end_position_on_the_genomic_accession=64956824, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110006, CurrentID=0, Status='live', Symbol='Gusb', Aliases='AI747421, Gur, Gus, Gus-r, Gus-s, Gus-t, Gus-u, Gut, asd, g', description='glucuronidase, beta', other_designations='adipose storage deficiency', map_location='5 68.32 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129989011, end_position_on_the_genomic_accession=130003044, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  110007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110007, CurrentID=0, Status='live', Symbol='Qui', Aliases='', description='quinine sensitivity, taste', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110009, CurrentID=0, Status='live', Symbol='Sxa', Aliases='', description='sex chromosome association', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110010, CurrentID=0, Status='live', Symbol='Tind', Aliases='', description='T-peripheral cell antigen', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110012, CurrentID=0, Status='live', Symbol='Tpgs1', Aliases='AV005629, Gm16517, Gtrgeo22, PGs1', description='tubulin polyglutamylase complex subunit 1', other_designations='PGs1 protein p32|gene trap ROSA b-geo 22|p32', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79669410, end_position_on_the_genomic_accession=79676126, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  110026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110026, CurrentID=0, Status='live', Symbol='Krd', Aliases='', description='kidney and retinal defects', other_designations='Del(19)TgN8052Mm', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110027, CurrentID=0, Status='live', Symbol='Cpu3', Aliases='', description='caudate putamen enriched 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110032, CurrentID=0, Status='live', Symbol='Kif22-ps', Aliases='AL033313, Kif22b', description='kinesin family member 22, pseudogene', other_designations='kinesin family member 22B', map_location='1 33.75 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=67585049, end_position_on_the_genomic_accession=67587101, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110033, CurrentID=0, Status='live', Symbol='Kif22', Aliases='AU021460, C81217, Kida, Kif22', description='kinesin family member 22', other_designations='kinesin family member 22A', map_location='7 69.29 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127027729, end_position_on_the_genomic_accession=127042466, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  634322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=634322, CurrentID=0, Status='live', Symbol='Trav14d-3-dv8', Aliases='TCRAV2.5, TRAV14D-3, Trav14d-3/dv8', description='T cell receptor alpha variable 14D-3-DV8', other_designations='T cell receptor alpha variable 14D-3/DV8|T-cell receptor alpha variable region 14D-3', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53078547, end_position_on_the_genomic_accession=53079045, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110036, CurrentID=0, Status='live', Symbol='Agp1', Aliases='Agp-1', description='anti gp70 immune complex 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241113, CurrentID=0, Status='live', Symbol='Prkag3', Aliases='AMPKg3L, AMPKg3S, Amkg3, Ampkg3', description='protein kinase, AMP-activated, gamma 3 non-catatlytic subunit', other_designations='AMP-activated protein kinase gamma subunit|AMPK gamma3|AMPK subunit gamma-3', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74738922, end_position_on_the_genomic_accession=74749590, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  241116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241116, CurrentID=0, Status='live', Symbol='Ccdc108', Aliases='A230051F10, B230363K08Rik', description='coiled-coil domain containing 108', other_designations='', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74902071, end_position_on_the_genomic_accession=74935628, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  236794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=236794, CurrentID=0, Status='live', Symbol='Slc9a6', Aliases='3732426M05, 6430520C02Rik, NHE6, mKIAA0267', description='solute carrier family 9 (sodium/hydrogen exchanger), member 6', other_designations='solute carrier family 9 (sodium/hydrogen exchanger)', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=56609557, end_position_on_the_genomic_accession=56664230, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  241118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241118, CurrentID=0, Status='live', Symbol='Asic4', Aliases='Accn4, BNAC4', description='acid-sensing (proton-gated) ion channel family member 4', other_designations='amiloride-sensitive cation channel 4, pituitary', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75450510, end_position_on_the_genomic_accession=75474340, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  110049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110049, CurrentID=0, Status='live', Symbol='Lmr2', Aliases='', description='leishmaniasis resistance 2', other_designations='', map_location='9 23.41 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110052, CurrentID=0, Status='live', Symbol='Dek', Aliases='1810019E15Rik, D13H6S231E', description='DEK oncogene (DNA binding)', other_designations='', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=47084767, end_position_on_the_genomic_accession=47106221, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  110053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110053, CurrentID=0, Status='live', Symbol='Dsi1', Aliases='Dsi-1', description='David Steffen integration 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241128, CurrentID=0, Status='live', Symbol='Fam124b', Aliases='A830043J08Rik', description='family with sequence similarity 124, member B', other_designations='', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=80186969, end_position_on_the_genomic_accession=80218504, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  110058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110058, CurrentID=0, Status='live', Symbol='Syt17', Aliases='Bk', description='synaptotagmin XVII', other_designations='B/K protein|brain and kidney protein|protein B/K|sytXVII', map_location='7 63.44 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=118380713, end_position_on_the_genomic_accession=118448913, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  110059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110059, CurrentID=0, Status='live', Symbol='Lmr1', Aliases='', description='leishmaniasis resistance 1', other_designations='', map_location='17 20.67 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241134, CurrentID=0, Status='live', Symbol='Nyap2', Aliases='9430031J16Rik, Jr6, Kiaa1486', description='neuronal tyrosine-phophorylated phosphoinositide 3-kinase adaptor 2', other_designations='neuronal tyrosine phosphorylated adaptor for phosphoinositide 3-kinase adaptor 2', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=81076553, end_position_on_the_genomic_accession=81365845, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  110065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110065, CurrentID=0, Status='live', Symbol='Soa', Aliases='', description='sucrose octaacetate aversion', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110066, CurrentID=0, Status='live', Symbol='Tcrd', Aliases='Tcrdelta', description='T cell receptor delta chain', other_designations='T-cell receptor delta chain|Tcr delta', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=53946073, end_position_on_the_genomic_accession=54159198, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110067, CurrentID=0, Status='live', Symbol='Tcrg', Aliases='TCRGV1S1, TCRGV2S1, TCRGV3S1, TCRGV5S3', description='T cell receptor gamma chain', other_designations='T-cell receptor gamma chain|TCR V gamma 1.3|TCR V gamma 2|TCR V gamma 3|TCR V gamma 4', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=19178042, end_position_on_the_genomic_accession=19356476, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110072, CurrentID=0, Status='live', Symbol='Pad4', Aliases='', description='MMTV LTR integration site 4', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110074, CurrentID=0, Status='live', Symbol='Dut', Aliases='5031412I06Rik, 5133400F09Rik, D2Bwg0749ep, dUTPase, Dut', description='deoxyuridine triphosphatase', other_designations='', map_location='2 61.33 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125247248, end_position_on_the_genomic_accession=125259049, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  110075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110075, CurrentID=0, Status='live', Symbol='Bmp3', Aliases='9530029I04Rik', description='bone morphogenetic protein 3', other_designations='', map_location='5 48.24 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=98854432, end_position_on_the_genomic_accession=98884052, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  110078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110078, CurrentID=0, Status='live', Symbol='Pygb', Aliases='', description='brain glycogen phosphorylase', other_designations='', map_location='2 74.74 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150786796, end_position_on_the_genomic_accession=150831748, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  110080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110080, CurrentID=0, Status='live', Symbol='Tme', Aliases='', description='T-associated maternal effect', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110082, CurrentID=0, Status='live', Symbol='Dnah5', Aliases='AU022615, Dnahc5, Mdnah5, b2b1134Clo, b2b1154Clo, b2b1537Clo, b2b1565Clo, b2b601Clo, mKIAA1603', description='dynein, axonemal, heavy chain 5', other_designations='axonemal beta dynein heavy chain 5|axonemal dynein heavy chain 5|ciliary dynein heavy chain 5|dynein, axon, heavy chain 5', map_location='15 10.9 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=28154854, end_position_on_the_genomic_accession=28472045, orientation='plus', exon_count=80, OMIM='', no_hdr0=''),
  110083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110083, CurrentID=0, Status='live', Symbol='Dnah12', Aliases='4921531P07Rik, DHC3, DLP12l, Dnahc12, Dnahc3, Dnahc7c, Dnahc7l, Gm284, Gm74, Gm907, HL-19, HL19, Hdhc3, Dnah12', description='dynein, axonemal, heavy chain 12', other_designations='axonemal beta dynein heavy chain 12|axonemal dynein heavy chain 12-like protein|axonemal dynein heavy chain 7-like protein|ciliary dynein heavy chain 12|dynein, axonemal, heavy chain 3', map_location='14 16.09 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=26691423, end_position_on_the_genomic_accession=26892284, orientation='plus', exon_count=78, OMIM='', no_hdr0=''),
  110084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110084, CurrentID=0, Status='live', Symbol='Dnah1', Aliases='B230373P09Rik, Dnahc1, E030034C22Rik, MDHC7', description='dynein, axonemal, heavy chain 1', other_designations='dynein, axon, heavy chain 1|dynein, axonemal, heavy polypeptide 1', map_location='14 19.1 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31260375, end_position_on_the_genomic_accession=31324063, orientation='minus', exon_count=78, OMIM='', no_hdr0=''),
  241158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241158, CurrentID=0, Status='live', Symbol='Ankmy1', Aliases='4930483I10Rik, ZMYND13', description='ankyrin repeat and MYND domain containing 1', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92859844, end_position_on_the_genomic_accession=92902986, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  241159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241159, CurrentID=0, Status='live', Symbol='Neu4', Aliases='9330166I04', description='sialidase 4', other_designations='N-acetyl-alpha-neuraminidase 4|neuraminidase 4', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=94020446, end_position_on_the_genomic_accession=94028334, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  110094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110094, CurrentID=0, Status='live', Symbol='Phka2', Aliases='6330505C01Rik, D330034O08, Phk', description='phosphorylase kinase alpha 2', other_designations='phosphorylase kinase alpha L subunit', map_location='X 73.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=160502156, end_position_on_the_genomic_accession=160598879, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  110095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110095, CurrentID=0, Status='live', Symbol='Pygl', Aliases='', description='liver glycogen phosphorylase', other_designations='', map_location='12 29.01 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=70190815, end_position_on_the_genomic_accession=70227683, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  110101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110101, CurrentID=0, Status='live', Symbol='Rfv1', Aliases='Rfv-1', description='recovery from Friend virus 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241175, CurrentID=0, Status='live', Symbol='Cntnap5b', Aliases='C230078M14Rik, Caspr5-2', description='contactin associated protein-like 5B', other_designations='cell recognition molecule Caspr5-2|cell recognition molecule Caspr5b|contactin associated protein-like 5-2|contactin-associated protein-like 5b', map_location='1 E1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=99772247, end_position_on_the_genomic_accession=100486066, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  110109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110109, CurrentID=0, Status='live', Symbol='Nop2', Aliases='120kDa, A530002O17, Nol1', description='NOP2 nucleolar protein', other_designations='NOP2 nucleolar protein homolog|nucleolar protein 1, 120 kDa|nucleolar protein 2 homolog|proliferating-cell nucleolar antigen p120|proliferation-associated nucleolar protein p120|putative ribosomal RNA methyltransferase NOP2', map_location='6 59.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125131883, end_position_on_the_genomic_accession=125144753, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  110112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110112, CurrentID=0, Status='live', Symbol='Acf1', Aliases='Acf-1', description='albumin conformation factor 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110114, CurrentID=0, Status='live', Symbol='Cxv2', Aliases='Cxv-2, XenCSA', description='control of xenotropic virus 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110115, CurrentID=0, Status='live', Symbol='Cyp11b1', Aliases='AA389061, CPN1, Cyp11b, Cyp11b-1, FHI', description='cytochrome P450, family 11, subfamily b, polypeptide 1', other_designations='steroid-11-beta-hydroxylase', map_location='15 34.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74834892, end_position_on_the_genomic_accession=74841643, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110119, CurrentID=0, Status='live', Symbol='Mpi', Aliases='1110002E17Rik, AI315153-1, Mpi1, Mpi', description='mannose phosphate isomerase', other_designations='PMI|mannose phosphate isomerase 1|phosphohexomutase|phosphomannose isomerase', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=57544260, end_position_on_the_genomic_accession=57552812, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110124, CurrentID=0, Status='live', Symbol='Tsu', Aliases='', description='T-suppressor cell alloantigen', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241197, CurrentID=0, Status='live', Symbol='Serpinb10', Aliases='9830131G07, BB233602-ps, Serpinb10', description='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 10', other_designations='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 10, pseudogene|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 10', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107528994, end_position_on_the_genomic_accession=107560735, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  171272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=171272, CurrentID=0, Status='live', Symbol='Vmn1r219', Aliases='V1rh13', description='vomeronasal 1 receptor 219', other_designations='vomernasal 1 receptor Vmn1r219|vomeronasal 1 receptor 213 pseudogene|vomeronasal 1 receptor, H13', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23162643, end_position_on_the_genomic_accession=23163581, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  110131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110131, CurrentID=0, Status='live', Symbol='H2-Ke1', Aliases='H-2Ke1, Ke-1', description='H2-K region expressed gene 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110135, CurrentID=0, Status='live', Symbol='Fgb', Aliases='2510049G14Rik', description='fibrinogen beta chain', other_designations='fibrinogen, B beta polypeptide', map_location='3 36.98 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=83042303, end_position_on_the_genomic_accession=83049790, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  110137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110137, CurrentID=0, Status='live', Symbol='Tlag1', Aliases='Tls-1, Tls1', description='T lymphoma attenuation gene 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110147, CurrentID=0, Status='live', Symbol='Ehmt2', Aliases='Bat8, D17Ertd710e, G9a, KMT1C, NG36', description='euchromatic histone lysine N-methyltransferase 2', other_designations='H3-K9-HMTase 3|HLA-B associated transcript 8|HLA-B-associated transcript 8|euchromatic histone-lysine N-methyltransferase 2|histone H3-K9 methyltransferase 3|histone-lysine N-methyltransferase, H3 lysine-9 specific 3|protein G9a', map_location='17 18.45 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34898464, end_position_on_the_genomic_accession=34914051, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  110150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110150, CurrentID=0, Status='live', Symbol='Fim3', Aliases='Fim-3', description='Friend MuLV integration site 3', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110151, CurrentID=0, Status='live', Symbol='H2-Ke5', Aliases='H-2Ke5, Ke-5', description='H2-K region expressed gene 5', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110153, CurrentID=0, Status='live', Symbol='Vis1', Aliases='Gm27276, His-1, His1', description='viral integration site 1', other_designations='hematopoietic insertion site 1', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=45238980, end_position_on_the_genomic_accession=45282900, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  241226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241226, CurrentID=0, Status='live', Symbol='Itga8', Aliases='AI447669', description='integrin alpha 8', other_designations='', map_location='2 9.12 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=12106632, end_position_on_the_genomic_accession=12301996, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  110157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110157, CurrentID=0, Status='live', Symbol='Raf1', Aliases='6430402F14Rik, AA990557, BB129353, Craf1, D830050J10Rik, Raf-1, c-Raf, v-Raf', description='v-raf-leukemia viral oncogene 1', other_designations='Craf1 transforming|cRaf|murine sarcoma 3611 oncogene 1|proto-oncogene c-RAF|sarcoma 3611 oncogene|v-raf-1 leukemia viral oncogene 1', map_location='6 53.62 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115618573, end_position_on_the_genomic_accession=115677129, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  241230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241230, CurrentID=0, Status='live', Symbol='St8sia6', Aliases='1700007J08Rik, AI314453, AI875066, ST8SiaVI, Siat8f', description='ST8 alpha-N-acetyl-neuraminide alpha-2,8-sialyltransferase 6', other_designations='SIAT8-F|ST8Sia VI|alpha-2, 8-sialyltransferase|alpha2,8-sialyltransferase VI|sialyltransferase 8 (alpha-2, 8-sialyltransferase) F|sialyltransferase 8F|sialyltransferase St8Sia VI|sialytransferase St8Sia VI', map_location='2 A1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=13651018, end_position_on_the_genomic_accession=13794734, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  110159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110159, CurrentID=0, Status='live', Symbol='Gct4', Aliases='Sot', description='granulosa cell tumorigenesis 4', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110168, CurrentID=0, Status='live', Symbol='Gpr18', Aliases='', description='G protein-coupled receptor 18', other_designations='G-protein coupled receptor 18|NAGly receptor|gpr18', map_location='14 65.86 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=121911434, end_position_on_the_genomic_accession=121915774, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  110172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110172, CurrentID=0, Status='live', Symbol='Slc35b1', Aliases='Ugalt2', description='solute carrier family 35, member B1', other_designations='UDP-galactose translocator 2|UDP-galactose transporter-related protein 1|UGTrel1', map_location='11 59.01 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=95384891, end_position_on_the_genomic_accession=95391776, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  110173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110173, CurrentID=0, Status='live', Symbol='Manba', Aliases='2410030O07Rik, B930014J03Rik, Bmn', description='mannosidase, beta A, lysosomal', other_designations='lysosomal beta A mannosidase|manba|mannanase|mannase', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=135485575, end_position_on_the_genomic_accession=135571404, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  110174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110174, CurrentID=0, Status='live', Symbol='Fim1', Aliases='Fim-1', description='Friend MuLV integration site 1', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110175, CurrentID=0, Status='live', Symbol='Ggct', Aliases='A030007L17Rik, AA673177, Gctg, Ggc', description='gamma-glutamyl cyclotransferase', other_designations='', map_location='6 27.26 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54985095, end_position_on_the_genomic_accession=54992965, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  110176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110176, CurrentID=0, Status='live', Symbol='Hh1', Aliases='Hh-1', description='hemopoietic histocompatibility', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110177, CurrentID=0, Status='live', Symbol='His2', Aliases='His-2', description='hematopoietic insertion site 2', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110178, CurrentID=0, Status='live', Symbol='Igh-Np', Aliases='', description='immunoglobulin heavy chain Np', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110179, CurrentID=0, Status='live', Symbol='Igh-Pc', Aliases='', description='immunoglobulin heavy chain Pc', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110181, CurrentID=0, Status='live', Symbol='Xlr1', Aliases='Xlr-1', description='X-linked lymphocyte-regulated 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110182, CurrentID=0, Status='live', Symbol='Penk-rs', Aliases='Penk, Penk2', description='preproenkephalin, related sequence', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110183, CurrentID=0, Status='live', Symbol='Rn18s-rs5', Aliases='Rnr17', description='18s RNA, related sequence 5', other_designations='18S ribosomal RNA pseudogene', map_location='', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=39846354, end_position_on_the_genomic_accession=39848202, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110187, CurrentID=0, Status='live', Symbol='Scgb2b26', Aliases='Abpbg26, Abpbg4, Abpg, C2d, Scgb2b4', description='secretoglobin, family 2B, member 26', other_designations='allergen dI chain C2D|secretoglobin, family 2B, member 4', map_location='7 19.38 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=33942997, end_position_on_the_genomic_accession=33944985, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  110188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110188, CurrentID=0, Status='live', Symbol='Fvis1', Aliases='Fis-1, Fis1', description='Friend virus integration site 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241263, CurrentID=0, Status='live', Symbol='Gpr158', Aliases='5330427M13Rik, mKIAA1136', description='G protein-coupled receptor 158', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=21367567, end_position_on_the_genomic_accession=21830544, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  110192 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110192, CurrentID=0, Status='live', Symbol='Tam1', Aliases='Tam-1', description='tosyl arginine methylesterase 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110193, CurrentID=0, Status='live', Symbol='Tcrb-D', Aliases='', description='T cell receptor beta, D region', other_designations='T-cell receptor beta, D region', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110196, CurrentID=0, Status='live', Symbol='Fdps', Aliases='6030492I17Rik, AI256750l1, mKIAA1293, Fdps', description='farnesyl diphosphate synthetase', other_designations='(2E,6E)-farnesyl diphosphate synthase|CR 39|FPP synthase|FPP synthetase|FPS|cholesterol-regulated 39 kDa protein|dimethylallyltranstransferase|farnesyl diphosphate synthase|farnesyl pyrophosphate synthetase|geranyltranstransferase', map_location='3 39.01 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89093588, end_position_on_the_genomic_accession=89101967, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  110197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110197, CurrentID=0, Status='live', Symbol='Dgkg', Aliases='2900055E17Rik, 90kDa, AI854428, Dagk3, E430001K23Rik, mKIAA4131', description='diacylglycerol kinase, gamma', other_designations='88 kDa diacylglycerol kinase|DAG kinase gamma|DGK-gamma|diacylglycerol kinase, gamma 3|diacylglycerol kinase, gamma 90kDa|diglyceride kinase gamma', map_location='16 13.37 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22466568, end_position_on_the_genomic_accession=22657299, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  110198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110198, CurrentID=0, Status='live', Symbol='Akr7a5', Aliases='0610025K21Rik, Afar, Afar1, Akr7a2', description='aldo-keto reductase family 7, member A5 (aflatoxin aldehyde reductase)', other_designations='SSA reductase|akr7a2|succinic semialdehyde reductase', map_location='4 70.64 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139310738, end_position_on_the_genomic_accession=139318786, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  241274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241274, CurrentID=0, Status='live', Symbol='Pnpla7', Aliases='BC027342, E430013P11Rik, Nre', description='patatin-like phospholipase domain containing 7', other_designations='NTE-related esterase|neuropathy target esterase-related esterase', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=24975171, end_position_on_the_genomic_accession=25054072, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  241275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241275, CurrentID=0, Status='live', Symbol='Noxa1', Aliases='NY-CO-31, SDCCAG31', description='NADPH oxidase activator 1', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25085667, end_position_on_the_genomic_accession=25095205, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  110205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110205, CurrentID=0, Status='live', Symbol='Igk-Ef2', Aliases='', description='immunoglobulin kappa chain Ef2', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110206, CurrentID=0, Status='live', Symbol='Mlvi2', Aliases='Mlvi-2', description='Moloney-MuLV integration site 2', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110208, CurrentID=0, Status='live', Symbol='Pgd', Aliases='0610042A05Rik, AU019875, C78335', description='phosphogluconate dehydrogenase', other_designations='', map_location='4 78.96 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=149149985, end_position_on_the_genomic_accession=149166707, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  110212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110212, CurrentID=0, Status='live', Symbol='Xic', Aliases='Cg, Xce', description='X chromosome inactivation center', other_designations='X chromosome controlling element', map_location='X', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110213, CurrentID=0, Status='live', Symbol='Tmbim6', Aliases='5031406P05Rik, Bi1, Tegt', description='transmembrane BAX inhibitor motif containing 6', other_designations='BI-1|testis enhanced gene transcript|testis-enhanced gene transcript protein|transmembrane BAX inhibitor motif-containing protein 6', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99392947, end_position_on_the_genomic_accession=99410049, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  241289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241289, CurrentID=0, Status='live', Symbol='Ppp1r26', Aliases='Gm347, Kiaa0649, mKIAA0649', description='protein phosphatase 1, regulatory subunit 26', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28446155, end_position_on_the_genomic_accession=28455508, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  241296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241296, CurrentID=0, Status='live', Symbol='Lrrc8a', Aliases='Lrrc8, mKIAA1437', description='leucine rich repeat containing 8A', other_designations='leucine-rich repeat-containing 8|leucine-rich repeat-containing protein 8A', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30237769, end_position_on_the_genomic_accession=30263790, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  110225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110225, CurrentID=0, Status='live', Symbol='Igh-Dex', Aliases='', description='immunoglobulin heavy chain Dex', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110229, CurrentID=0, Status='live', Symbol='Igh-Nbp', Aliases='', description='immunoglobulin heavy chain Nbp', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241303, CurrentID=0, Status='live', Symbol='Fam78a', Aliases='A130092J06Rik', description='family with sequence similarity 78, member A', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32052166, end_position_on_the_genomic_accession=32087436, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  241308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241308, CurrentID=0, Status='live', Symbol='Ralgps1', Aliases='5830418G11Rik, AI853783, AI854138, RALGEF2, RALGPS1A, RalGEF 2, mKIAA0351', description='Ral GEF with PH domain and SH3 binding motif 1', other_designations='Ral guanine nucleotide exchange factor RalGPS1A|ral GEF with PH domain and SH3-binding motif 1|ral guanine nucleotide exchange factor 2|ralA exchange factor RalGPS1', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33133419, end_position_on_the_genomic_accession=33371511, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  241311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241311, CurrentID=0, Status='live', Symbol='Zbtb34', Aliases='Gm349, mKIAA1993', description='zinc finger and BTB domain containing 34', other_designations='', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=33406108, end_position_on_the_genomic_accession=33431324, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  241322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241322, CurrentID=0, Status='live', Symbol='Zbtb6', Aliases='A830092L04Rik, ZID, Zfp482', description='zinc finger and BTB domain containing 6', other_designations='zinc finger protein 482', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37425500, end_position_on_the_genomic_accession=37430919, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  110251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110251, CurrentID=0, Status='live', Symbol='Papg1', Aliases='', description='pulmonary adenoma progression 1', other_designations='', map_location='4 47.21 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241324, CurrentID=0, Status='live', Symbol='Crb2', Aliases='5930402A21, BC043114', description='crumbs family member 2', other_designations='crumbs homolog 2|crumbs-like protein 2', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37776225, end_position_on_the_genomic_accession=37799103, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  110253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110253, CurrentID=0, Status='live', Symbol='Triobp', Aliases='AI428493, EST478828, Tara, mKIAA1662', description='TRIO and F-actin binding protein', other_designations='protein Tara|trio-associated repeat on actin', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78947724, end_position_on_the_genomic_accession=79005869, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  110255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110255, CurrentID=0, Status='live', Symbol='Tir3', Aliases='', description='trypanosome infection response 3', other_designations='', map_location='1 67.71 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110256, CurrentID=0, Status='live', Symbol='Tir2', Aliases='', description='trypanosome infection response 2', other_designations='', map_location='5 42.62 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110257, CurrentID=0, Status='live', Symbol='Hba-a2', Aliases='', description='hemoglobin alpha, adult chain 2', other_designations='alpha 2 globin', map_location='11 18.87 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=32296492, end_position_on_the_genomic_accession=32297310, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  110258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110258, CurrentID=0, Status='live', Symbol='Hh3', Aliases='Hh-3', description='hemopoietic histocompatibility 3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110261, CurrentID=0, Status='live', Symbol='Tcrd-D', Aliases='', description='T cell receptor delta, D region', other_designations='T-cell receptor delta, D region', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110262, CurrentID=0, Status='live', Symbol='Tel4q', Aliases='', description='telomeric sequence, Chr 4, q arm', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110265, CurrentID=0, Status='live', Symbol='Msra', Aliases='2310045J23Rik, 6530413P12Rik, MSR-A', description='methionine sulfoxide reductase A', other_designations='PMSR|peptide Met(O) reductase|peptide methionine sulfoxide reductase|peptide-methionine (S)-S-oxide reductase|protein-methionine-S-oxide reductase', map_location='14 33.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64122621, end_position_on_the_genomic_accession=64455903, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  110273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110273, CurrentID=0, Status='live', Symbol='Nba2', Aliases='', description='New Zealand Black autoimmunity 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  503491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=503491, CurrentID=0, Status='live', Symbol='Defa24', Aliases='Defcr24', description='defensin, alpha, 24', other_designations='defensin-related cryptdin-24', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21734494, end_position_on_the_genomic_accession=21735471, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  110279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110279, CurrentID=0, Status='live', Symbol='Bcr', Aliases='5133400C09Rik, AI561783, AI853148, mKIAA3017', description='breakpoint cluster region', other_designations='breakpoint cluster region homolog', map_location='10 38.49 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=75060557, end_position_on_the_genomic_accession=75184923, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  110280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110280, CurrentID=0, Status='live', Symbol='Ctl1', Aliases='Ctl-1', description='cytotoxic T lymphocyte response 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110283, CurrentID=0, Status='live', Symbol='Tir1', Aliases='', description='trypanosome infection response 1', other_designations='', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110284, CurrentID=0, Status='live', Symbol='Es27', Aliases='Es-27', description='esterase 27, serum cholinesterase', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110287, CurrentID=0, Status='live', Symbol='Ldr1', Aliases='Ldr-1', description='lactate dehydrogenase regulator 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110295, CurrentID=0, Status='live', Symbol='Bis1', Aliases='', description='beta-carboline-induced seizures 1', other_designations='', map_location='4 69.44 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110301, CurrentID=0, Status='live', Symbol='Arp', Aliases='', description='Arp lymphoid/erythroid hyperplasia', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110304, CurrentID=0, Status='live', Symbol='Glra3', Aliases='', description='glycine receptor, alpha 3 subunit', other_designations='', map_location='8 29.69 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=55940411, end_position_on_the_genomic_accession=56130260, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  110305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110305, CurrentID=0, Status='live', Symbol='If1', Aliases='If-1', description='NDV-induced circulating interferon', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110307, CurrentID=0, Status='live', Symbol='Int41', Aliases='Int-41', description='mammary tumor integration site 41', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110308, CurrentID=0, Status='live', Symbol='Krt5', Aliases='3300001P10Rik, AW146334, CK5, K5, Krt2-5, Tfip8', description='keratin 5', other_designations='58 kDa cytokeratin|CK-5|cytokeratin 5|cytokeratin-5|keratin complex 2, basic, gene 5|keratin-5|krt5|tuftelin interacting protein 8|type-II keratin Kb5', map_location='15 56.98 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101707070, end_position_on_the_genomic_accession=101712891, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110309, CurrentID=0, Status='live', Symbol='Krt6', Aliases='K6, Krt-2.6, Krt2-6', description='keratin 6', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110310, CurrentID=0, Status='live', Symbol='Krt7', Aliases='D15Wsu77e, K7, Krt2-7', description='keratin 7', other_designations='CK-7|Cytokeratin 7|cytokeratin-7|keratin complex 2, basic, gene 7|keratin-7|krt7|sarcolectin|type-II keratin Kb7', map_location='15 56.88 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101412403, end_position_on_the_genomic_accession=101427806, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  110312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110312, CurrentID=0, Status='live', Symbol='Pmch', Aliases='A230109K23Rik', description='pro-melanin-concentrating hormone', other_designations='neuropeptide-glutamic acid-isoleucine (NEI)|neuropeptide-glycine-glutamic acid (NGE)', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=88091072, end_position_on_the_genomic_accession=88092374, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  110315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110315, CurrentID=0, Status='live', Symbol='Trbv26', Aliases='Gm16909, Tcrb-V3, Tcrb-V3.1', description='T cell receptor beta, variable 26', other_designations='T cell receptor beta variable 26|T-cell receptor beta, variable 26|T-cell receptor beta, variable 3|T-cell receptor beta, variable 3.1', map_location='6 19.33 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41227525, end_position_on_the_genomic_accession=41227913, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  241391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241391, CurrentID=0, Status='live', Symbol='Galnt5', Aliases='4832424J23', description='UDP-N-acetyl-alpha-D-galactosamine:polypeptide N-acetylgalactosaminyltransferase 5', other_designations='UDP-GalNAc:polypeptide N-acetylgalactosaminyltransferase 5|galNAc-T5|polypeptide GalNAc transferase 5|pp-GaNTase 5|ppGaNTase-T5|protein-UDP acetylgalactosaminyltransferase 5', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=57997264, end_position_on_the_genomic_accession=58041027, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  110321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110321, CurrentID=0, Status='live', Symbol='Jckm2', Aliases='', description='juvenile cystic kidney modifier 2', other_designations='', map_location='1 57.91 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110323, CurrentID=0, Status='live', Symbol='Cox6b1', Aliases='2010000G05Rik, Cox6b', description='cytochrome c oxidase, subunit VIb polypeptide 1', other_designations='COX VIb-1|cytochrome c oxidase subunit VIb', map_location='7 18.84 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30616974, end_position_on_the_genomic_accession=30626151, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  110326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110326, CurrentID=0, Status='live', Symbol='Tas1r1', Aliases='Gpr70, T1r1, TR1', description='taste receptor, type 1, member 1', other_designations='G protein-coupled receptor 70|G-protein coupled receptor 70|candidate taste receptor T1R1', map_location='4 82.83 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152027914, end_position_on_the_genomic_accession=152038490, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  110328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110328, CurrentID=0, Status='live', Symbol='Evi4', Aliases='Evi-4', description='ecotropic viral integration site 4', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110331, CurrentID=0, Status='live', Symbol='Jckm1', Aliases='', description='juvenile cystic kidney modifier 1', other_designations='', map_location='10 34.83 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110332, CurrentID=0, Status='live', Symbol='Pp2d1', Aliases='4921523A10Rik', description='protein phosphatase 2C-like domain containing 1', other_designations='PP2C-like domain-containing protein C3orf48 homolog', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=53507460, end_position_on_the_genomic_accession=53539451, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  110333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110333, CurrentID=0, Status='live', Symbol='Rmst', Aliases='AI853140, C230053E11Rik, D930049J19Rik, Dmt2, M2, Ncrms', description='rhabdomyosarcoma 2 associated transcript (non-coding RNA)', other_designations='dorso-medial telencephalon gene 2', map_location='10 46.68 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=92081746, end_position_on_the_genomic_accession=92165178, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  503550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=503550, CurrentID=0, Status='live', Symbol='Klri1', Aliases='', description='killer cell lectin-like receptor family I member 1', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129696118, end_position_on_the_genomic_accession=129720610, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  110338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110338, CurrentID=0, Status='live', Symbol='Anth2', Aliases='Anth-2', description='resistance to Bacillus anthracis 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110340, CurrentID=0, Status='live', Symbol='Ped', Aliases='', description='preimplantation embryo development', other_designations='Qa-2 antigen', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110341, CurrentID=0, Status='live', Symbol='Tel13q', Aliases='TD-2, TD-3', description='telomeric sequence, Chr 13, q arm', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  503558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=503558, CurrentID=0, Status='live', Symbol='Taar9', Aliases='Gm231, taR-9', description='trace amine-associated receptor 9', other_designations='mTaar9', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24108488, end_position_on_the_genomic_accession=24109534, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  110350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110350, CurrentID=0, Status='live', Symbol='Dync2h1', Aliases='4432416O06Rik, AI448217, D030010H02Rik, D330044F14Rik, DHC1b, DHC2, Dnchc2, b2b414Clo, mKIAA1997', description='dynein cytoplasmic 2 heavy chain 1', other_designations='cytoplasmic dynein heavy chain 2|dynein cytoplasmic heavy chain 2|dynein heavy chain 11|dynein heavy chain 2|dynein heavy chain isotype 1B|dynein, cytoplasmic, heavy chain 2|mDHC11', map_location='9 2.46 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=6928503, end_position_on_the_genomic_accession=7177231, orientation='minus', exon_count=90, OMIM='', no_hdr0=''),
  110351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110351, CurrentID=0, Status='live', Symbol='Rap1gap', Aliases='1300019I11Rik, 2310004O14Rik, AI427470, ARPP-90, Rap1GAP1, Rap1ga1', description='Rap1 GTPase-activating protein', other_designations='Rap1, GTPase-activating protein 1', map_location='4 70.02 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=137664711, end_position_on_the_genomic_accession=137729861, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  110355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110355, CurrentID=0, Status='live', Symbol='Adrbk1', Aliases='Adrbk-1, Bark-1, GRK2, beta ARK, betaARK1', description='adrenergic receptor kinase, beta 1', other_designations='G-protein-coupled receptor kinase 2|beta-AR kinase-1', map_location='19 3.98 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4285999, end_position_on_the_genomic_accession=4306222, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  110357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110357, CurrentID=0, Status='live', Symbol='Mod2', Aliases='Cd243, Mdr, Mdr-1, Mod-2, Mod-2r, Mod-2s-r, Mod2-s, Mod2', description='malic enzyme complex, mitochondrial', other_designations='malic enzyme, mitochondrial regulatory|malic enzyme, mitochondrial structural', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241431, CurrentID=0, Status='live', Symbol='Xirp2', Aliases='2310003D02Rik, 2310008C07Rik, A530024P18Rik, AI452089, Cmya3, Gm352, Xin2', description='xin actin-binding repeat containing 2', other_designations='Xin actin-binding repeat containing 2 short splice form|beta-xin|cardiomyopathy associated 3|cardiomyopathy-associated protein 3|mXinbeta|myogenic MEF2-activated Xin-related protein|myomaxin|xeplin|xin-beta', map_location='2 C1.3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=67446000, end_position_on_the_genomic_accession=67526620, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  110367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110367, CurrentID=0, Status='live', Symbol='Tcrd-V2', Aliases='', description='T cell receptor delta, variable 2', other_designations='T-cell receptor delta, variable 2', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110370, CurrentID=0, Status='live', Symbol='Tcrd-V5', Aliases='', description='T cell receptor delta, variable 5', other_designations='T-cell receptor delta, variable 5', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110372, CurrentID=0, Status='live', Symbol='Tcrd-V7', Aliases='', description='T cell receptor delta, variable 7', other_designations='T-cell receptor delta, variable 7', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241447, CurrentID=0, Status='live', Symbol='Cers6', Aliases='4732462C07Rik, AW544719, Lass6, T1L', description='ceramide synthase 6', other_designations='LAG1 homolog, ceramide synthase 6|LAG1 longevity assurance homolog 6|longevity assurance homolog 6', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=68861137, end_position_on_the_genomic_accession=69114282, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  110379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110379, CurrentID=0, Status='live', Symbol='Sec13', Aliases='1110003H02Rikl1, Sec13r, Sec13', description='SEC13 homolog, nuclear pore and COPII coat complex component', other_designations='SEC13 homolog|SEC13 related|SEC13-like 1|SEC13-like protein 1|SEC13-related protein', map_location='6 52.84 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113728052, end_position_on_the_genomic_accession=113740681, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  241452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241452, CurrentID=0, Status='live', Symbol='Dhrs9', Aliases='C730025I08Rik, Rdh15', description='dehydrogenase/reductase (SDR family) member 9', other_designations='3-alpha hydroxysteroid dehydrogenase|3-alpha-HSD|3alpha-HSD|epithelial retinol dehydrogenase|short-chain dehydrogenase/reductase retSDR8', map_location='2 C2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=69380462, end_position_on_the_genomic_accession=69403086, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  110382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110382, CurrentID=0, Status='live', Symbol='C8b', Aliases='4930439B20Rik, AI595927', description='complement component 8, beta polypeptide', other_designations='complement component 8, beta subunit', map_location='4 48.61 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=104764576, end_position_on_the_genomic_accession=104804549, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  110383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110383, CurrentID=0, Status='live', Symbol='Carg1', Aliases='', description='Candida albicans resistance gene 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110384, CurrentID=0, Status='live', Symbol='Myhs', Aliases='', description='myosin heavy chain, skeletal muscle', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110385, CurrentID=0, Status='live', Symbol='Pde4c', Aliases='Dpde1, E130301F19Rik, dunce', description='phosphodiesterase 4C, cAMP specific', other_designations='', map_location='8 34.15 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70723689, end_position_on_the_genomic_accession=70751186, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  110391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110391, CurrentID=0, Status='live', Symbol='Qdpr', Aliases='2610008L04Rik, D5Ertd371e, Dhpr, PKU2', description='quinoid dihydropteridine reductase', other_designations='HDHPR', map_location='5 24.9 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=45434032, end_position_on_the_genomic_accession=45450229, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  110392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110392, CurrentID=0, Status='live', Symbol='rsl', Aliases='Rslcan-9', description='regulator of sex limited protein-Slp', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  503610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=503610, CurrentID=0, Status='live', Symbol='Zdhhc18', Aliases='', description='zinc finger, DHHC domain containing 18', other_designations='DHHC-18|zinc finger DHHC domain-containing protein 18', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133605301, end_position_on_the_genomic_accession=133633480, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  110417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110417, CurrentID=0, Status='live', Symbol='Pigh', Aliases='2210416H01Rik, A930028P05Rik', description='phosphatidylinositol glycan anchor biosynthesis, class H', other_designations='PIG-H|phosphatidylinositol glycan, class H|phosphatidylinositol-glycan biosynthesis class H protein', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=79080664, end_position_on_the_genomic_accession=79089754, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  241490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241490, CurrentID=0, Status='live', Symbol='Rbm45', Aliases='Drb1, Drbp1, G430095G15Rik', description='RNA binding motif protein 45', other_designations='RB-1|RNA-binding motif protein 45|developmentally regulated RNA binding protein 1|developmentally-regulated RNA-binding protein 1|dopamine receptor binding 1', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76369984, end_position_on_the_genomic_accession=76383768, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  110421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110421, CurrentID=0, Status='live', Symbol='H3b', Aliases='Hd-1, Hd1, Hd2', description='histocompatibility 3b, Th stimulating', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241494, CurrentID=0, Status='live', Symbol='Zfp385b', Aliases='B830010L13Rik, C130013B13Rik, Zfp533, Znf385b', description='zinc finger protein 385B', other_designations='zinc finger protein 533', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=77410622, end_position_on_the_genomic_accession=77817706, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  110438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110438, CurrentID=0, Status='live', Symbol='Dxcp2', Aliases='Cps-1, Dcp-2, Dcp2', description='dexamethasone induced cleft palate 2', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241514, CurrentID=0, Status='live', Symbol='Zfp804a', Aliases='C630007C17Rik, Znf804a', description='zinc finger protein 804A', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=82052874, end_position_on_the_genomic_accession=82259879, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  634731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=634731, CurrentID=0, Status='live', Symbol='Susd1', Aliases='A530080P10, Gm12528, OTTMUSG00000007332', description='sushi domain containing 1', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=59314202, end_position_on_the_genomic_accession=59439289, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  241516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241516, CurrentID=0, Status='live', Symbol='Fsip2', Aliases='OTTMUSG00000013335', description='fibrous sheath-interacting protein 2', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=82943609, end_position_on_the_genomic_accession=83008934, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  110446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110446, CurrentID=0, Status='live', Symbol='Acat1', Aliases='6330585C21Rik, Acat', description='acetyl-Coenzyme A acetyltransferase 1', other_designations='acetoacetyl-CoA thiolase', map_location='9 29.12 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53580522, end_position_on_the_genomic_accession=53610350, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  241520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241520, CurrentID=0, Status='live', Symbol='Fam171b', Aliases='D430039N05Rik', description='family with sequence similarity 171, member B', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=83812728, end_position_on_the_genomic_accession=83883486, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  110450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110450, CurrentID=0, Status='live', Symbol='Lmyc2', Aliases='Lmyc-2', description='lung carcinoma myc related oncogene 2', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241525, CurrentID=0, Status='live', Symbol='Ypel4', Aliases='', description='yippee-like 4 (Drosophila)', other_designations='', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84733742, end_position_on_the_genomic_accession=84738655, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  110454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110454, CurrentID=0, Status='live', Symbol='Ly6a', Aliases='Ly-6A.2, Ly-6A/E, Ly-6E.1, Sca-1, Sca1, TAP', description='lymphocyte antigen 6 complex, locus A', other_designations='T-cell-activating protein|stem cell antigen 1', map_location='15 34.29 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=74994877, end_position_on_the_genomic_accession=74998031, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  241528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241528, CurrentID=0, Status='live', Symbol='Lrrc55', Aliases='A330097E02Rik, Gm351', description='leucine rich repeat containing 55', other_designations='BK channel auxiliary gamma subunit LRRC55|BK channel auxilliary gamma subunit LRRC55', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85188067, end_position_on_the_genomic_accession=85198932, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  110459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110459, CurrentID=0, Status='live', Symbol='Ods', Aliases='', description='ocular degeneration with sex reversal', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110460, CurrentID=0, Status='live', Symbol='Acat2', Aliases='AW742799, Tcp-1x, Tcp1-rs1', description='acetyl-Coenzyme A acetyltransferase 2', other_designations='cytosolic acetoacetyl-CoA thiolase|t-complex protein 1, related sequence 1', map_location='17 A2-A3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12943042, end_position_on_the_genomic_accession=12960725, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110465, CurrentID=0, Status='live', Symbol='Tcrb-C', Aliases='', description='T cell receptor beta, constant region', other_designations='T-cell receptor beta, constant region', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110466, CurrentID=0, Status='live', Symbol='Tcrb-V', Aliases='', description='T cell receptor beta, variable region', other_designations='T-cell receptor beta, variable region', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241547, CurrentID=0, Status='live', Symbol='Harbi1', Aliases='D230010M03Rik', description='harbinger transposase derived 1', other_designations='harbinger transposase-derived nuclease', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91710855, end_position_on_the_genomic_accession=91721567, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  241556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241556, CurrentID=0, Status='live', Symbol='Tspan18', Aliases='2610042G18Rik, 6720430O15, BB226562', description='tetraspanin 18', other_designations='tspan-18', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93201760, end_position_on_the_genomic_accession=93334555, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  241568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241568, CurrentID=0, Status='live', Symbol='Lrrc4c', Aliases='6430556C10Rik, NGL-1', description='leucine rich repeat containing 4C', other_designations='netrin-G1 ligand', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=96318169, end_position_on_the_genomic_accession=97631672, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  241576 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241576, CurrentID=0, Status='live', Symbol='Ldlrad3', Aliases='6430500P08, AI194318, Lrad3', description='low density lipoprotein receptor class A domain containing 3', other_designations='', map_location='2 E2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=101950201, end_position_on_the_genomic_accession=102186460, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  110511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110511, CurrentID=0, Status='live', Symbol='Olfr153', Aliases='GA_x6K02T2Q125-49033418-49034341, MOR177-5, Olfr4-2, V5', description='olfactory receptor 153', other_designations='olfactory receptor 4 cluster, gene 2', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87532035, end_position_on_the_genomic_accession=87532958, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  241589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241589, CurrentID=0, Status='live', Symbol='D430041D05Rik', Aliases='9630056G07Rik, G2', description='RIKEN cDNA D430041D05 gene', other_designations='', map_location='2 54.59 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104143073, end_position_on_the_genomic_accession=104411779, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  241593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241593, CurrentID=0, Status='live', Symbol='Pin1rt1', Aliases='OTTMUSG00000014964, Pin1-ps1, Pin1l', description='protein (peptidyl-prolyl cis/trans isomerase) NIMA-interacting 1, retrogene 1', other_designations='peptidylprolyl cis/trans isomerase, NIMA-interacting 1, pseudogene 1', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=104713887, end_position_on_the_genomic_accession=104716362, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  110522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110522, CurrentID=0, Status='live', Symbol='Igl1-r', Aliases='Igl-1r, Igl-Lo', description='immunoglobulin lambda chain regulator', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110523, CurrentID=0, Status='live', Symbol='Tcrd-J', Aliases='', description='T cell receptor delta, joining region', other_designations='T-cell receptor delta, joining region', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110524, CurrentID=0, Status='live', Symbol='Dgkq', Aliases='110kDa, DAGK, DAGK7, Dagk4, Dgkd', description='diacylglycerol kinase, theta', other_designations='DAG kinase theta|DGK-theta|Dgk theta|diacylglycerol kinase, delta (110 kDa)|diglyceride kinase theta', map_location='5 53.24 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108647044, end_position_on_the_genomic_accession=108660828, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  110532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110532, CurrentID=0, Status='live', Symbol='Adarb1', Aliases='1700057H01Rik, AW124433, AW558573, Adar2, BB220382, D10Bwg0447e, Red1', description='adenosine deaminase, RNA-specific, B1', other_designations='RNA-editing deaminase 1|RNA-editing enzyme 1|dsRNA adenosine deaminase', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77290727, end_position_on_the_genomic_accession=77418273, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  110536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110536, CurrentID=0, Status='live', Symbol='H2-D2', Aliases='H-2D2', description='histocompatibility 2, D region locus 2', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110539, CurrentID=0, Status='live', Symbol='Ltw3', Aliases='Ltw-3, Lvtw-3', description='liver 20-30 thousand Mol. Wt. protein 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241612, CurrentID=0, Status='live', Symbol='Slc5a12', Aliases='AI315119, D630015G19, SMCT2', description='solute carrier family 5 (sodium/glucose cotransporter), member 12', other_designations='electroneutral sodium monocarboxylate cotransporter|low-affinity sodium-lactate cotransporter|sodium solute cotransporter|solute carrier family 5 member 12', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=110597299, end_position_on_the_genomic_accession=110649346, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  110542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110542, CurrentID=0, Status='live', Symbol='Amhr2', Aliases='Misiir, Misrii, Mrii', description='anti-Mullerian hormone type 2 receptor', other_designations='AMH type II receptor|MIS type II receptor|anti-Muellerian hormone type II receptor|anti-Mullerian hormone type 2 receptor delta 2|anti-Mullerian hormone type 2 receptor delta 9/10', map_location='15 57.58 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102445028, end_position_on_the_genomic_accession=102454640, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  241624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241624, CurrentID=0, Status='live', Symbol='Exd1', Aliases='4932702D22Rik, Exdl1', description="exonuclease 3'-5' domain containing 1", other_designations="Vexonuclease 3'-5' domain-like 1|exonuclease 3'-5' domain-like 1|exonuclease 3'-5' domain-like-containing protein 1", map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=119516505, end_position_on_the_genomic_accession=119547627, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  110553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110553, CurrentID=0, Status='live', Symbol='H2-D3', Aliases='H-2D3', description='histocompatibility 2, D region locus 3', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241627, CurrentID=0, Status='live', Symbol='Wdr76', Aliases='5830411K18Rik, 6030457D09, AA517836', description='WD repeat domain 76', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121506723, end_position_on_the_genomic_accession=121544860, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  110557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110557, CurrentID=0, Status='live', Symbol='H2-Q6', Aliases='0610037M15Rik, H-2Q6, Qa-6, Qa6', description='histocompatibility 2, Q region locus 6', other_designations='put. Q8 antigen', map_location='17 18.66 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35424799, end_position_on_the_genomic_accession=35430053, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  110558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110558, CurrentID=0, Status='live', Symbol='H2-Q9', Aliases='H-2Q9, Ped, Qa-2, Qa-9, Qa9', description='histocompatibility 2, Q region locus 9', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241633, CurrentID=0, Status='live', Symbol='Atp8b4', Aliases='A530043E15, Im', description='ATPase, class I, type 8B, member 4', other_designations='ATPase 8B4, p type|ATPase B4, class I', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=126320973, end_position_on_the_genomic_accession=126500669, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  241634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241634, CurrentID=0, Status='live', Symbol='Spdye4c', Aliases='Gm355, SpeedyB3, SpyB3', description='speedy/RINGO cell cycle regulator family, member E4C', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=128591122, end_position_on_the_genomic_accession=128607940, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  241636 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241636, CurrentID=0, Status='live', Symbol='Tgm6', Aliases='TGM3L, TGY', description='transglutaminase 6', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130112416, end_position_on_the_genomic_accession=130154232, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  241638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241638, CurrentID=0, Status='live', Symbol='Lzts3', Aliases='Prosapip1', description='leucine zipper, putative tumor suppressor family member 3', other_designations='proSAP-interacting protein 1|proline rich synapse associated protein interacting protein 1 (Prosapip1)', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130632765, end_position_on_the_genomic_accession=130642844, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  241639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241639, CurrentID=0, Status='live', Symbol='Fermt1', Aliases='5830467P10Rik, Kindlin-1', description='fermitin family member 1', other_designations='unc-112-related protein 1', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132904177, end_position_on_the_genomic_accession=132946036, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  241656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241656, CurrentID=0, Status='live', Symbol='Pak7', Aliases='2900083L08Rik, 6430627N20, Pak5', description='p21 protein (Cdc42/Rac)-activated kinase 7', other_designations='PAK-5|PAK-7|p21-activated kinase 5|p21-activated kinase 7', map_location='2 F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=136081088, end_position_on_the_genomic_accession=136387967, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  110593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110593, CurrentID=0, Status='live', Symbol='Prdm2', Aliases='4833427P12Rik, E330024L24Rik, Gm1033, Gm19732, KMT8, Riz, Riz1, Znfpr1c1', description='PR domain containing 2, with ZNF domain', other_designations='', map_location='4 76.84 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143107391, end_position_on_the_genomic_accession=143212737, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  110595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110595, CurrentID=0, Status='live', Symbol='Timp4', Aliases='TIMP-4', description='tissue inhibitor of metalloproteinase 4', other_designations='tissue inhibitor of metalloproteinases 4', map_location='6 53.29 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115245616, end_position_on_the_genomic_accession=115251867, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  110596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110596, CurrentID=0, Status='live', Symbol='Arhgef28', Aliases='9230110L08Rik, AI323540, D13Bwg1089e, RIP2, Rgnef, RhoGEF, Rhoip2, p190RhoGEF', description='Rho guanine nucleotide exchange factor (GEF) 28', other_designations='190 kDa guanine nucleotide exchange factor|Rho interacting protein 2|Rho specific exchange factor|Rho-guanine nucleotide exchange factor|p190-RhoGEF|rho-interacting protein 2', map_location='13 51.6 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=97898594, end_position_on_the_genomic_accession=98206624, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  110598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110598, CurrentID=0, Status='live', Symbol='Cnpl', Aliases='Cnp-2, Cnp2', description='cyclic nucleotide phosphodiesterase-like', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110606, CurrentID=0, Status='live', Symbol='Fntb', Aliases='2010013E13Rik, AA409500', description='farnesyltransferase, CAAX box, beta', other_designations='CAAX farnesyltransferase subunit beta|FTase-beta|ras proteins prenyltransferase subunit beta', map_location='12 33.73 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=76837398, end_position_on_the_genomic_accession=76921412, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  110611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110611, CurrentID=0, Status='live', Symbol='Hdlbp', Aliases='1110005P14Rik, AA960365, AI118566, D1Ertd101e', description='high density lipoprotein (HDL) binding protein', other_designations='HDL-binding protein|high density lipoprotein-binding protein', map_location='1 47.24 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93405940, end_position_on_the_genomic_accession=93478917, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  110612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110612, CurrentID=0, Status='live', Symbol='Iglv2', Aliases='Igl-V2', description='immunoglobulin lambda variable 2', other_designations='immunoglobulin lambda chain, variable 2|immunoglobulin lambda-2|light chain variable region', map_location='16 11.93 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19260403, end_position_on_the_genomic_accession=19260844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  110615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110615, CurrentID=0, Status='live', Symbol='Ihpd', Aliases='', description='interspecific hybrid placental dysplasia', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110616, CurrentID=0, Status='live', Symbol='Atxn3', Aliases='2210008M02Rik, AI463012, AI647473, ATX3, MJD1, Mjd, Sca3, ataxin-3', description='ataxin 3', other_designations='Machado-Joseph disease (spinocerebellar ataxia 3, olivopontocerebellar ataxia 3, autosomal dominant, ataxin 3) homolog|machado-Joseph disease protein 1 homolog', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=101918901, end_position_on_the_genomic_accession=101958800, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  110621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110621, CurrentID=0, Status='live', Symbol='H2-T11-ps', Aliases='H-2T11, H2-T11, T10c, T11', description='histocompatibility 2, T region locus 11, pseudogene', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241694, CurrentID=0, Status='live', Symbol='Ralgapa2', Aliases='A230067G21Rik, AS250, BC053994, RGC2, pp250', description='Ral GTPase activating protein, alpha subunit 2 (catalytic)', other_designations='250 kDa substrate of Akt|P220|ral GTPase-activating protein alpha subunit 2', map_location='2 G1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=146239124, end_position_on_the_genomic_accession=146512222, orientation='minus', exon_count=46, OMIM='', no_hdr0=''),
  110637 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110637, CurrentID=0, Status='live', Symbol='Grik4', Aliases='6330551K01Rik, GluK4, GluRgamma1, KA-1, KA1', description='glutamate receptor, ionotropic, kainate 4', other_designations='glutamate receptor KA1', map_location='9 23.8 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=42520411, end_position_on_the_genomic_accession=42945135, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  110639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110639, CurrentID=0, Status='live', Symbol='Prps2', Aliases='2610101M19Rik, AA589463, AI464149, Prps-2', description='phosphoribosyl pyrophosphate synthetase 2', other_designations='PRS-II|phosphoribosyl pyrophosphate synthase II|phosphoribosyl pyrophosphate synthetase II', map_location='X 78.32 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=167346322, end_position_on_the_genomic_accession=167383100, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110641, CurrentID=0, Status='live', Symbol='Hoxb3os', Aliases='Gm11530, Hoxb3s, OTTMUSG00000001778', description='homeobox B3 and homeobox B2, opposite strand', other_designations='homeo box B3, opposite strand transcript|homeobox B3, opposite strand transcript', map_location='11 59.85 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110648, CurrentID=0, Status='live', Symbol='Lmx1a', Aliases='Lmx1.1, dr, dreher, sst', description='LIM homeobox transcription factor 1 alpha', other_designations='LIM/homeobox protein 1.1|LIM/homeobox protein LMX1A|LMX-1.1|shaker short-tail', map_location='1 75.08 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=167688300, end_position_on_the_genomic_accession=167848741, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  110649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110649, CurrentID=0, Status='live', Symbol='Idd7', Aliases='Idd-7', description='insulin dependent diabetes susceptibility 7', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110650, CurrentID=0, Status='live', Symbol='Mdm3', Aliases='Mdm-3', description='transformed mouse 3T3 cell double minute 3', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110651, CurrentID=0, Status='live', Symbol='Rps6ka3', Aliases='MAPKAPK-1b, MPK-9, Rsk2, S6K-alpha3, p90RSK3, pp90RSK2', description='ribosomal protein S6 kinase polypeptide 3', other_designations='90 kDa ribosomal protein S6 kinase 3|MAP kinase-activated protein kinase 1b|MAPK-activated protein kinase 1b|MAPKAP kinase 1b|ribosomal protein S6 kinase 2', map_location='X 73.27 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=159255782, end_position_on_the_genomic_accession=159368244, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  110654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110654, CurrentID=0, Status='live', Symbol='Idd2', Aliases='Idd-2', description='insulin dependent diabetes susceptibility 2', other_designations='', map_location='9 52.24 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110655, CurrentID=0, Status='live', Symbol='Idd3', Aliases='Idd-3', description='insulin dependent diabetes susceptibility 3', other_designations='', map_location='3 29.17 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110656, CurrentID=0, Status='live', Symbol='Idd5', Aliases='Idd-5', description='insulin dependent diabetes susceptibility 5', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110658, CurrentID=0, Status='live', Symbol='Idd9', Aliases='Idd-9', description='insulin dependent diabetes susceptibility 9', other_designations='', map_location='4 57.9 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110660, CurrentID=0, Status='live', Symbol='Scl2', Aliases='Scl-2', description='susceptibility to cutaneous leishmaniasis 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110673, CurrentID=0, Status='live', Symbol='Hbvi', Aliases='D13Pas2', description='hepatitis B virus transgenic insertion site', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110674, CurrentID=0, Status='live', Symbol='Idd1', Aliases='Idd-1', description='insulin dependent diabetes susceptibility 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110675, CurrentID=0, Status='live', Symbol='Idd4', Aliases='Idd-4', description='insulin dependent diabetes susceptibility 4', other_designations='', map_location='11 45.76 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110677, CurrentID=0, Status='live', Symbol='TelXqYq1', Aliases='TelXqYq-1', description='telomeric sequence, Chr X, Chr Y, q arm 1', other_designations='', map_location='XY 76.0 cM', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110685, CurrentID=0, Status='live', Symbol='Afr2', Aliases='Afr-2, Rif', description='alpha fetoprotein regulation 2, inducibility', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110690 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110690, CurrentID=0, Status='live', Symbol='Cyp2a', Aliases='', description='cytochrome P450, family 2, subfamily a', other_designations='cytochrome P450, 2a, phenobarbitol inducible|phenobarbital inducible', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241764, CurrentID=0, Status='live', Symbol='L3mbtl1', Aliases='C630004G01, L3mbtl, mKIAA0681', description='l(3)mbt-like (Drosophila)', other_designations='H-l(3)mbt protein|L(3)mbt protein homolog', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=162942350, end_position_on_the_genomic_accession=162977806, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  110695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110695, CurrentID=0, Status='live', Symbol='Aldh7a1', Aliases='Atq1, D18Wsu181e', description='aldehyde dehydrogenase family 7, member A1', other_designations='P6c dehydrogenase|aldehyde dehydrogenase 7 family, member A1|alpha-AASA dehydrogenase|antiquitin-1|betaine aldehyde dehydrogenase|delta1-piperideine-6-carboxylate dehydrogenase|delta1-piperideine-6-carboxylate dehydrogenease', map_location='18 30.55 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=56524722, end_position_on_the_genomic_accession=56572939, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  110696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110696, CurrentID=0, Status='live', Symbol='H2-M10.3', Aliases='5.3H', description='histocompatibility 2, M region locus 10.3', other_designations='', map_location='17 19.01 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36365004, end_position_on_the_genomic_accession=36368417, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  241770 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241770, CurrentID=0, Status='live', Symbol='Rims4', Aliases='6530406L02, Rim4', description='regulating synaptic membrane exocytosis 4', other_designations='RIM 4|RIM4 gamma|rab3-interacting molecule 4', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163859756, end_position_on_the_genomic_accession=163918990, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  110709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110709, CurrentID=0, Status='live', Symbol='Braf-rs1', Aliases='Braf', description='Braf transforming gene, related sequence 1', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110712, CurrentID=0, Status='live', Symbol='Idd10', Aliases='Idd-10', description='insulin dependent diabetes susceptibility 10', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110713, CurrentID=0, Status='live', Symbol='Idd11', Aliases='', description='insulin dependent diabetes susceptibility 11', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110716, CurrentID=0, Status='live', Symbol='Idd13', Aliases='', description='insulin dependent diabetes susceptibility 13', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241794, CurrentID=0, Status='live', Symbol='Kcng1', Aliases='AW536275, OTTMUSG00000016048', description='potassium voltage-gated channel, subfamily G, member 1', other_designations='orthologue of H. sapiens potassium voltage-gated channel, subfamily G, member 1 (KCNG1)|voltage-gated potassium channel subunit Kv6.1', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=168260110, end_position_on_the_genomic_accession=168283440, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  110733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110733, CurrentID=0, Status='live', Symbol='Idd16', Aliases='', description='insulin dependent diabetes susceptibility 16', other_designations='', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110735, CurrentID=0, Status='live', Symbol='Idd17', Aliases='', description='insulin dependent diabetes susceptibility 17', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110736, CurrentID=0, Status='live', Symbol='Idd18', Aliases='', description='insulin dependent diabetes susceptibility 18', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110746, CurrentID=0, Status='live', Symbol='Psp-rs1', Aliases='Psp-2, Psp2', description='parotid secretory protein related sequence 1', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110749, CurrentID=0, Status='live', Symbol='Chaf1b', Aliases='2600017H24Rik, C76145, CAF-I p60, CAF-Ip60, CAF1, CAF1A, CAF1P60, MPHOSPH7', description='chromatin assembly factor 1, subunit B (p60)', other_designations='CAF-1 subunit B|CAF-I 60 kDa subunit|chromatin assembly factor I p60 subunit', map_location='16 54.96 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=93883895, end_position_on_the_genomic_accession=93906115, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  110750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110750, CurrentID=0, Status='live', Symbol='Cse1l', Aliases='2610100P18Rik, AA407533, Capts, Cas, Xpo2', description='chromosome segregation 1-like (S. cerevisiae)', other_designations='chromosome segregation 1-like protein|exp2|importin-alpha re-exporter', map_location='2 87.22 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=166906051, end_position_on_the_genomic_accession=166946389, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  110751 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110751, CurrentID=0, Status='live', Symbol='Adam33', Aliases='Adaml', description='a disintegrin and metallopeptidase domain 33', other_designations='ADAM 33|a disintegrin and metalloprotease domain 33', map_location='2 63.26 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131050817, end_position_on_the_genomic_accession=131063814, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  110754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110754, CurrentID=0, Status='live', Symbol='Igkrs', Aliases='Igk-rs', description='immunoglobulin kappa chain recombining segment', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110755, CurrentID=0, Status='live', Symbol='Spe1r', Aliases='Msp-1, Msp1, Spe1-r, Ssp', description='salivary protein electrophoretic 1, regulator', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110759, CurrentID=0, Status='live', Symbol='Igkj1', Aliases='Igk-J1', description='immunoglobulin kappa joining 1', other_designations='immunoglobulin kappa chain, joining region, 1', map_location='6 32.04 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70722562, end_position_on_the_genomic_accession=70722599, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110760, CurrentID=0, Status='live', Symbol='Igkj2', Aliases='Igk-J2', description='immunoglobulin kappa joining 2', other_designations='immunoglobulin kappa chain, joining region, 2', map_location='6 32.04 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70722916, end_position_on_the_genomic_accession=70722954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110761 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110761, CurrentID=0, Status='live', Symbol='Igkj3', Aliases='Igk-J3', description='immunoglobulin kappa joining 3', other_designations='immunoglobulin kappa chain, joining region, 2|immunoglobulin kappa chain, joining region, 3', map_location='6 32.04 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70723223, end_position_on_the_genomic_accession=70723260, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110762 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110762, CurrentID=0, Status='live', Symbol='Igkj4', Aliases='Igk-J4', description='immunoglobulin kappa joining 4', other_designations='immunoglobulin kappa chain, joining region, 2|immunoglobulin kappa chain, joining region, 4', map_location='6 32.04 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70723548, end_position_on_the_genomic_accession=70723585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110763, CurrentID=0, Status='live', Symbol='Igkj5', Aliases='Igk-J5', description='immunoglobulin kappa joining 5', other_designations='immunoglobulin kappa chain, joining region, 2|immunoglobulin kappa chain, joining region, 5', map_location='6 32.04 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70723886, end_position_on_the_genomic_accession=70723923, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  110773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110773, CurrentID=0, Status='live', Symbol='Hsdr1', Aliases='Alf, hsdr-1', description='hepatocyte specific developmental regulation 1', other_designations='albino-lethal factor', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110774, CurrentID=0, Status='live', Symbol='Igl-C5', Aliases='', description='immunoglobulin lambda chain, constant region 5', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241850, CurrentID=0, Status='live', Symbol='Abhd16b', Aliases='1700095H12', description='abhydrolase domain containing 16B', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181493206, end_position_on_the_genomic_accession=181494980, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  110784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110784, CurrentID=0, Status='live', Symbol='Nr3c2', Aliases='MR, Mlr', description='nuclear receptor subfamily 3, group C, member 2', other_designations='aldosterone receptor', map_location='8 36.34 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=76899261, end_position_on_the_genomic_accession=77245232, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  110785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110785, CurrentID=0, Status='live', Symbol='Iglc1', Aliases='Clambda1, Igl-C1', description='immunoglobulin lambda constant 1', other_designations='immunoglobulin lambda chain, constant region 1', map_location='16 11.8 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19061752, end_position_on_the_genomic_accession=19062071, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  110786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110786, CurrentID=0, Status='live', Symbol='Iglc2', Aliases='Igl-C2', description='immunoglobulin lambda constant 2', other_designations='immunoglobulin lambda chain, constant region 2|immunoglobulin lambda-chain', map_location='16 11.89 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19198536, end_position_on_the_genomic_accession=19198852, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  110787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110787, CurrentID=0, Status='live', Symbol='Iglc3', Aliases='Igl-C3', description='immunoglobulin lambda constant 3', other_designations='immunoglobulin lambda chain, constant region 3', map_location='16 11.81 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19065365, end_position_on_the_genomic_accession=19065681, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  110789 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110789, CurrentID=0, Status='live', Symbol='Adgrv1', Aliases='Frings, Gpr98, Mass1, Mgr1, VLGR1', description='adhesion G protein-coupled receptor V1', other_designations='G protein-coupled receptor 98|monogenic audiogenic seizure susceptibility protein 1|monogenic, audiogenic seizure susceptibility 1|neurepin|very large G protein-coupled receptor 1|very large G-protein coupled receptor 1', map_location='13 42.18 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=81095068, end_position_on_the_genomic_accession=81633197, orientation='minus', exon_count=93, OMIM='', no_hdr0=''),
  110794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110794, CurrentID=0, Status='live', Symbol='Cebpe', Aliases='C/EBPe, CRP1, Gm294', description='CCAAT/enhancer binding protein (C/EBP), epsilon', other_designations='C/EBP-related protein 1|C/EBPepsilon|CCAAT/enhancer binding protein epsilon|c/EBP epsilon', map_location='14 27.9 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54710363, end_position_on_the_genomic_accession=54712174, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  110796 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110796, CurrentID=0, Status='live', Symbol='Tshz1', Aliases='5730407I04Rik, D18Bwg1409e, Mtsh1, NY-CO-33, Sdccag33, Tsh1, mKIAA4206, teashirt1', description='teashirt zinc finger family member 1', other_designations='serologically defined colon cancer antigen 3 homolog|serologically defined colon cancer antigen 33', map_location='18 57.53 cM', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=84011627, end_position_on_the_genomic_accession=84086562, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  110798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110798, CurrentID=0, Status='live', Symbol='H2-S', Aliases='H-2S', description='histocompatibility 2, S region (C4, Slp, Bf, C2)', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110800, CurrentID=0, Status='live', Symbol='Camkl', Aliases='', description='calcium/calmodulin dependent protein kinase-like', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110805, CurrentID=0, Status='live', Symbol='Foxe1', Aliases='Titf2', description='forkhead box E1', other_designations='TTF-2|thyroid transcription factor 2', map_location='4 24.68 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46344194, end_position_on_the_genomic_accession=46345309, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  110809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110809, CurrentID=0, Status='live', Symbol='Srsf1', Aliases='1110054N12Rik, 5730507C05Rik, 6330415C05Rik, AI482334, AW491331, Asf, Sf2, Sfrs1', description='serine/arginine-rich splicing factor 1', other_designations='ASF/SF2|pre-mRNA-splicing factor SRp30a|splicing factor, arginine/serine-rich 1 (ASF/SF2)', map_location='11 52.4 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=88046902, end_position_on_the_genomic_accession=88053757, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  110810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110810, CurrentID=0, Status='live', Symbol='Snrp2a', Aliases='Rnu2pa', description='U2 small nuclear ribonucleoprotein polypeptide 2A', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110811, CurrentID=0, Status='live', Symbol='Tblvi1', Aliases='Tblvi-1', description='thymotropic type B leukemogenic virus integration', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110816, CurrentID=0, Status='live', Symbol='Pwp2', Aliases='6530411D08Rikh, wdp103, Pwp2', description='PWP2 periodic tryptophan protein homolog (yeast)', other_designations='WD-repeat protein p103', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78170910, end_position_on_the_genomic_accession=78185171, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  110821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110821, CurrentID=0, Status='live', Symbol='Pcca', Aliases='C79630', description='propionyl-Coenzyme A carboxylase, alpha polypeptide', other_designations='PCCase subunit alpha|propanoyl-CoA:carbon dioxide ligase subunit alpha|propionyl CoA-carboxylase alpha', map_location='14 65.99 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=122484414, end_position_on_the_genomic_accession=122889945, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  110826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110826, CurrentID=0, Status='live', Symbol='Etfb', Aliases='0610009I16Rik, 2810441H06Rik', description='electron transferring flavoprotein, beta polypeptide', other_designations='beta-ETF', map_location='7 28.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43444072, end_position_on_the_genomic_accession=43457800, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  110829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110829, CurrentID=0, Status='live', Symbol='Lims1', Aliases='2310016J22Rik, 4921524A02Rik, AI507642, AU021743, AW551584, C430041B13Rikl, PINCH, PINCH1, Lims1', description='LIM and senescent cell antigen-like domains 1', other_designations='particularly interesting new Cys-His protein 1', map_location='10 29.31 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=58323435, end_position_on_the_genomic_accession=58424691, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  110831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110831, CurrentID=0, Status='live', Symbol='Cbx-rs1', Aliases='M31', description='chromobox related sequence 1 (Drosophila HP1 class)', other_designations='', map_location='X 70.0 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110834, CurrentID=0, Status='live', Symbol='Chrna3', Aliases='(a)3, A730007P14Rik, Acra-3, Acra3', description='cholinergic receptor, nicotinic, alpha polypeptide 3', other_designations='alpha 3|neuronal nicotinic acetylcholine receptor, alpha 3 subunit', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55011343, end_position_on_the_genomic_accession=55026559, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  110835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110835, CurrentID=0, Status='live', Symbol='Chrna5', Aliases='Acra-5, Acra5', description='cholinergic receptor, nicotinic, alpha polypeptide 5', other_designations='nicotinic acetylcholine receptor alpha5 subunit', map_location='9 29.84 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54979828, end_position_on_the_genomic_accession=55009231, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  110842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110842, CurrentID=0, Status='live', Symbol='Etfa', Aliases='2010200I21Rik, D9Ertd394e', description='electron transferring flavoprotein, alpha polypeptide', other_designations='Alpha-ETF', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55454436, end_position_on_the_genomic_accession=55512243, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  241915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241915, CurrentID=0, Status='live', Symbol='Phc3', Aliases='E030046K01Rik, Edr3, Hph3', description='polyhomeotic-like 3 (Drosophila)', other_designations='polyhomeotic 3', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=30899295, end_position_on_the_genomic_accession=30969479, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  241919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241919, CurrentID=0, Status='live', Symbol='Slc7a14', Aliases='A930013N06, BC061928', description='solute carrier family 7 (cationic amino acid transporter, y+ system), member 14', other_designations='solute carrier family 7 member 14|solute carrier family 7, member 14', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=31202854, end_position_on_the_genomic_accession=31310574, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  110854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110854, CurrentID=0, Status='live', Symbol='Ppp2r4', Aliases='2610042B21Rik, C77440, N28142, PR53', description='protein phosphatase 2A activator, regulatory subunit B', other_designations="PP2A, subunit B', PR53 isoform|PTPA|phosphotyrosyl phosphatase activator|protein phosphatase 2A, regulatory subunit B (PR 53)|serine/threonine-protein phosphatase 2A regulatory subunit 4|serine/threonine-protein phosphatase 2A regulatory subunit B'", map_location='2 21.71 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=30416050, end_position_on_the_genomic_accession=30447807, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  110855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110855, CurrentID=0, Status='live', Symbol='Pde6c', Aliases='cpfl1', description='phosphodiesterase 6C, cGMP specific, cone, alpha prime', other_designations='cGMP phosphodiesterase 6C', map_location='19 C3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=38132781, end_position_on_the_genomic_accession=38183955, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  630146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=630146, CurrentID=0, Status='live', Symbol='Cd101', Aliases='EWI-101, Gm1016, Gm734, Igsf2', description='CD101 antigen', other_designations='EWI immunoglobulin subfamily member EWI-101|glu-Trp-Ile EWI motif-containing protein 101|immunoglobulin superfamily, member 2|leukocyte surface protein', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100993529, end_position_on_the_genomic_accession=101029495, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  110862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110862, CurrentID=0, Status='live', Symbol='Kcnq3', Aliases='', description='potassium voltage-gated channel, subfamily Q, member 3', other_designations='KQT-like 3|potassium channel subunit alpha KvLQT3|voltage-gated potassium channel subunit Kv7.3', map_location='15 29.16 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=65992662, end_position_on_the_genomic_accession=66286813, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  18477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18477, CurrentID=0, Status='live', Symbol='Prdx1', Aliases='MSP23, NkefA, OSF-3, OSF3, PAG, Paga, PrdxI, PrxI, TDX2, TPxA, Tdpx2, prx1', description='peroxiredoxin 1', other_designations='Trx dependent peroxide reductase 2|macrophage 23 Kd stress protein|macrophage 23 kDa stress protein|macrophage 23kDa stress protein|macrophase stress protein 22kDa|macrophase stress protein 23 kd|osteoblast specific factor 3|osteoblast-specific factor 3|proliferation-associated gene A|thioredoxin dependent peroxide reductase 2|thioredoxin peroxidase 2|thioredoxin-dependent peroxide reductase 2', map_location='4 53.28 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116685599, end_position_on_the_genomic_accession=116700000, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  110864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110864, CurrentID=0, Status='live', Symbol='Fert1', Aliases='', description='fer (fms/fps related) protein kinase, testis specific 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  241943 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241943, CurrentID=0, Status='live', Symbol='Ccdc144b', Aliases='', description='coiled-coil domain containing 144B', other_designations='', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36007247, end_position_on_the_genomic_accession=36053547, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  241944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241944, CurrentID=0, Status='live', Symbol='D3Ertd254e', Aliases='', description='DNA segment, Chr 3, ERATO Doi 254, expressed', other_designations='', map_location='3 17.41 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=36151080, end_position_on_the_genomic_accession=36170341, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  110876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110876, CurrentID=0, Status='live', Symbol='Scn2a1', Aliases='6430408L10, A230052E19Rik, Nav1.2, Scn2a', description='sodium channel, voltage-gated, type II, alpha 1', other_designations='sodium channel, voltage-gated, type II, alpha polypeptide 1|voltage-gated sodium channel Nav1.2 variant a|voltage-gated sodium channel Nav1.2 variant b', map_location='2 38.61 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=65620764, end_position_on_the_genomic_accession=65767449, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  110877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110877, CurrentID=0, Status='live', Symbol='Slc18a1', Aliases='4832416I10Rik, Vat1, Vmat1', description='solute carrier family 18 (vesicular monoamine), member 1', other_designations='solute carrier family 18 member 1|vesicular amine transporter 1', map_location='8 33.88 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69037708, end_position_on_the_genomic_accession=69090629, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  241950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241950, CurrentID=0, Status='live', Symbol='Bbs12', Aliases='Gm1805, Gm407, Gm721', description='Bardet-Biedl syndrome 12 (human)', other_designations='DN-452A22.3', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37312547, end_position_on_the_genomic_accession=37321450, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  110880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110880, CurrentID=0, Status='live', Symbol='Scn4a', Aliases='Nav1.4, SkM1, mH2', description='sodium channel, voltage-gated, type IV, alpha', other_designations='sodium channel alpha-subunit|sodium channel protein skeletal muscle subunit alpha|sodium channel protein type IV subunit alpha|sodium channel, voltage-gated, type IV, alpha polypeptide|voltage-gated sodium channel subunit alpha Nav1.4', map_location='11 68.91 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106317545, end_position_on_the_genomic_accession=106365552, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  110886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110886, CurrentID=0, Status='live', Symbol='Gabra5', Aliases='A230018I05Rik', description='gamma-aminobutyric acid (GABA) A receptor, subunit alpha 5', other_designations='GABA(A) receptor subunit alpha-5|GABAA receptor alpha5 subunit|gamma-aminobutyric acid (GABA-A) receptor, subunit alpha 5|ligand-gated ion channel subunit', map_location='7 33.52 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=57407669, end_position_on_the_genomic_accession=57510100, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  110891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110891, CurrentID=0, Status='live', Symbol='Slc8a2', Aliases='Ncx2', description='solute carrier family 8 (sodium/calcium exchanger), member 2', other_designations='Na(+)/Ca(2+)-exchange protein 2|solute carrier family 8 member 2', map_location='7 8.75 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16128313, end_position_on_the_genomic_accession=16161063, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  110893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110893, CurrentID=0, Status='live', Symbol='Slc8a3', Aliases='AW742262, Ncx3', description='solute carrier family 8 (sodium/calcium exchanger), member 3', other_designations='Na(+)/Ca(2+)-exchange protein 3|solute carrier family 8 member 3', map_location='12 37.44 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=81197915, end_position_on_the_genomic_accession=81333728, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  110895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110895, CurrentID=0, Status='live', Symbol='Slc9a4', Aliases='AW990558, D730009J23Rik, NHE4', description='solute carrier family 9 (sodium/hydrogen exchanger), member 4', other_designations='NHE-4|Na(+)/H(+) exchanger 4|solute carrier family 9 member 4', map_location='1 19.51 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40579715, end_position_on_the_genomic_accession=40630731, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  110902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110902, CurrentID=0, Status='live', Symbol='Chrna2', Aliases='Acra-2, Acra2, BC011490', description='cholinergic receptor, nicotinic, alpha polypeptide 2 (neuronal)', other_designations='', map_location='14 34.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=66133238, end_position_on_the_genomic_accession=66153396, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  110907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110907, CurrentID=0, Status='live', Symbol='Hsp70-2os', Aliases='70kDa, Hsp70-2s', description='heat shock protein, 2 opposite strand transcription unit', other_designations='heat shock protein, 70 kDa-2 opposite strand transcription unit', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110911, CurrentID=0, Status='live', Symbol='Cds2', Aliases='5730450N06Rik, 5730460C18Rik, AI854580, D2Wsu127e', description='CDP-diacylglycerol synthase (phosphatidate cytidylyltransferase) 2', other_designations='CDP-DAG synthase 2|CDP-DG synthase 2|CDP-diglyceride pyrophosphorylase 2|CDP-diglyceride synthase 2|CTP:phosphatidate cytidylyltransferase 2', map_location='2 64.15 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132262609, end_position_on_the_genomic_accession=132312050, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  241989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=241989, CurrentID=0, Status='live', Symbol='Pabpc4l', Aliases='C330050A14Rik, EG241989', description='poly(A) binding protein, cytoplasmic 4-like', other_designations='', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=46442197, end_position_on_the_genomic_accession=46447996, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  110920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110920, CurrentID=0, Status='live', Symbol='Hspa13', Aliases='1600002I10Rik, AV006182, B230217N24Rik, Stch', description='heat shock protein 70 family, member 13', other_designations='heat shock protein 70kDa family, member 13|microsomal stress 70 protein ATPase core|microsomal stress-70 protein ATPase core|stress 70 protein chaperone microsome-associated 60 kDa protein|stress 70 protein chaperone, microsome-associated, 60kD human homolog|stress 70 protein chaperone, microsome-associated, human homolog|stress-70 protein chaperone microsome-associated 60 kDa protein', map_location='16 43.36 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=75755190, end_position_on_the_genomic_accession=75767277, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  110924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110924, CurrentID=0, Status='live', Symbol='Tmevp3', Aliases='NeST', description="Theiler's murine encephalomyelitis virus persistence 3", other_designations="Theiler's murine encephalomyelitis virus persistence susceptibility 3", map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  110931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110931, CurrentID=0, Status='live', Symbol='Vmn2r-ps57', Aliases='Casr-rs1, Gprc2a-rs1', description='vomeronasal 2, receptor, pseudogene 57', other_designations='G protein-coupled receptor, family C, group 2, member A, related sequence 1|calcium sensing receptor, related sequence 1', map_location='7 28.25 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=41937560, end_position_on_the_genomic_accession=41980928, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  110935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110935, CurrentID=0, Status='live', Symbol='Atp6v1b1', Aliases='AW208839, Atp6b1, D630003L15, D630030L16Rik, D630039P21Rik, Vpp-3, Vpp3', description='ATPase, H+ transporting, lysosomal V1 subunit B1', other_designations='ATPase, H+ transporting, V1 subunit B|ATPase, H+ transporting, lysosomal (vacuolar proton pump), beta 56/58 kDa|ATPase, H+ transporting, lysosomal 56/58kDa, V1 subunit B|lysosomal 56/58kDa', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83743017, end_position_on_the_genomic_accession=83758810, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  18491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18491, CurrentID=0, Status='live', Symbol='Pappa', Aliases='8430414N03Rik, IGFBP-4ase, PAG1, PAPP-A', description='pregnancy-associated plasma protein A', other_designations='IGF-dependent IGFBP-4 protease|insulin-like growth factor-dependent IGF-binding protein 4 protease', map_location='4 34.14 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=65124174, end_position_on_the_genomic_accession=65357509, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  110948 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110948, CurrentID=0, Status='live', Symbol='Hlcs', Aliases='410I21.SP6, D16Jhu34', description='holocarboxylase synthetase (biotin- [propriony-Coenzyme A-carboxylase (ATP-hydrolysing)] ligase)', other_designations='biotin apo-protein ligase|biotin- [propriony-Coenzyme A-carboxylase (ATP-hydrolysing)] ligase', map_location='16 55.12 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=94129195, end_position_on_the_genomic_accession=94313905, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  100135654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100135654, CurrentID=0, Status='live', Symbol='Rhox3c', Aliases='EG665203, Rhox3.3, Rhox3b', description='reproductive homeobox 3C', other_designations='reproductive homeobox 3B', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37469868, end_position_on_the_genomic_accession=37473961, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  242022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242022, CurrentID=0, Status='live', Symbol='Frem2', Aliases='6030440P17Rik, 8430406N05Rik, Gm409, b2b1562Clo, my, ne, nv1', description='Fras1 related extracellular matrix protein 2', other_designations='ECM3 homolog|NV domain containing protein NV1|NV domain-containing protein 1', map_location='3 25.24 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=53513938, end_position_on_the_genomic_accession=53657912, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  110954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110954, CurrentID=0, Status='live', Symbol='Rpl10', Aliases='D0HXS648, DXHXS648, DXHXS648E, QM, QM16', description='ribosomal protein L10', other_designations='24.6 kDa protein|24.6kDa protein|protein QM homolog|ribosomal protein 10', map_location='X 37.94 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74270816, end_position_on_the_genomic_accession=74273135, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  110956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110956, CurrentID=0, Status='live', Symbol='D17H6S56E-5', Aliases='G7e', description='DNA segment, Chr 17, human D6S56E 5', other_designations='', map_location='17 18.54 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34996678, end_position_on_the_genomic_accession=35000741, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  110957 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110957, CurrentID=0, Status='live', Symbol='D1Pas1', Aliases='Pl10', description='DNA segment, Chr 1, Pasteur Institute 1', other_designations='', map_location='1 90.09 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=186967416, end_position_on_the_genomic_accession=186970627, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  110958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110958, CurrentID=0, Status='live', Symbol='M1ap', Aliases='D6Mm5e', description='meiosis 1 associated protein', other_designations='meiosis 1-arresting protein|meiosis 1-associated protein|spermatogenesis-associated protein 37', map_location='6 35.94 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=82946908, end_position_on_the_genomic_accession=83030309, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  110959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110959, CurrentID=0, Status='live', Symbol='Nudt19', Aliases='D7Rp2, D7Rp2-r, D7Rp2-s, D7Rp2e, RP2-r, RP2-s', description='nudix (nucleoside diphosphate linked moiety X)-type motif 19', other_designations='RP2p|androgen regulated gene RP2|androgen-regulated protein RP2|nucleoside diphosphate-linked moiety X motif 19, mitochondrial|nudix motif 19|testosterone-regulated RP2 protein', map_location='7 21.37 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35547185, end_position_on_the_genomic_accession=35556417, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  110960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110960, CurrentID=0, Status='live', Symbol='Tars', Aliases='D15Wsu59e, ThrRS', description='threonyl-tRNA synthetase', other_designations='threonyl-tRNA synthetase, cytoplasmic', map_location='15 5.6 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=11383663, end_position_on_the_genomic_accession=11399658, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  110962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110962, CurrentID=0, Status='live', Symbol='Mbd6', Aliases='D10Wsu93e', description='methyl-CpG binding domain protein 6', other_designations='methyl-CpG-binding protein MBD6', map_location='10 74.5 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127281956, end_position_on_the_genomic_accession=127290719, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  110964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110964, CurrentID=0, Status='live', Symbol='Lamb1-2', Aliases='Lamb-1.13, Lamb-1ps, Lamb1-ps', description='laminin B1 subunit 2', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242037, CurrentID=0, Status='live', Symbol='Ankub1', Aliases='Gm410', description='ankrin repeat and ubiquitin domain containing 1', other_designations='ankyrin repeat and ubiquitin domain-containing 1', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=57656493, end_position_on_the_genomic_accession=57692780, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  110966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110966, CurrentID=0, Status='live', Symbol='Rnu1-2', Aliases='', description='U1 small nuclear RNA 2', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100135657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100135657, CurrentID=0, Status='live', Symbol='Rhox3e', Aliases='ENSMUSG00000071772, OTTMUSG00000017148, Rhox3.5, Rhox3d', description='reproductive homeobox 3E', other_designations='reproductive homeobox 3D', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37546975, end_position_on_the_genomic_accession=37550897, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  504186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=504186, CurrentID=0, Status='live', Symbol='Chrna10', Aliases='', description='cholinergic receptor, nicotinic, alpha polypeptide 10', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102111262, end_position_on_the_genomic_accession=102116828, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  504193 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=504193, CurrentID=0, Status='live', Symbol='Npcd', Aliases='Cbx6-Nptxr', description='neuronal pentraxin chromo domain', other_designations='Cbx6-Nptxr readthrough transcripts', map_location='15E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=79786351, end_position_on_the_genomic_accession=79834333, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  242050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242050, CurrentID=0, Status='live', Symbol='Igsf10', Aliases='6530405F15Rik, 9030224D03, AA409708, AA536958, AI414626, Adlican2, CMF608', description='immunoglobulin superfamily, member 10', other_designations='', map_location='3 D', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=59316735, end_position_on_the_genomic_accession=59344764, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  110989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=110989, CurrentID=0, Status='live', Symbol='Rnu1-ps1', Aliases='Rnu1-1', description='U1 small nuclear RNA pseudogene 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111008, CurrentID=0, Status='live', Symbol='Cbl-ps1', Aliases='Cbl-1', description='Casitas B-lineage lymphoma, pseudogene 1', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242083, CurrentID=0, Status='live', Symbol='Ppm1l', Aliases='3222401G13, 5930404J21Rik, AI481720, AW045850, PP2C, PP2Cepsilon, Pp2ce, Ppp2ce, mKIAA4175', description='protein phosphatase 1 (formerly 2C)-like', other_designations='PP2C-epsilon|protein phosphatase 1-like|protein phosphatase 2C epsilon|protein phosphatase 2C isoform epsilon|protein phosphatase 2a, catalytic subunit, epsilon isoform', map_location='3 E1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69316918, end_position_on_the_genomic_accession=69560800, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  242093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242093, CurrentID=0, Status='live', Symbol='Rxfp4', Aliases='GPCR142, Gpr100, Rln3r2, SALPR', description='relaxin family peptide receptor 4', other_designations='G protein-coupled receptor 100|G-protein coupled receptor 100|RLN3 receptor 2|insulin-like peptide INSL5 receptor|relaxin 3 receptor 2|relaxin-3/INSL7 receptor 2', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88651898, end_position_on_the_genomic_accession=88653142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  242100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242100, CurrentID=0, Status='live', Symbol='Pglyrp3', Aliases='Gm420', description='peptidoglycan recognition protein 3', other_designations='PGLYRPIalpha|PGRP-I-alpha|peptidoglycan recognition protein I-alpha|peptidoglycan recognition protein intermediate alpha', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=92012477, end_position_on_the_genomic_accession=92032672, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  242109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242109, CurrentID=0, Status='live', Symbol='Zfp697', Aliases='9430053K19, AI467503, Znf697', description='zinc finger protein 697', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=98381341, end_position_on_the_genomic_accession=98431951, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  111040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111040, CurrentID=0, Status='live', Symbol='Rrm2-ps1', Aliases='', description='ribonucleotide reductase M2 pseudogene 1', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111042, CurrentID=0, Status='live', Symbol='Hsp86-ps2', Aliases='86kDa, Hsp86-3, Hsp90', description='heat shock protein 86, pseudogene 2', other_designations='heat shock protein, 86 kDa, pseudogene 2|heat shock protein, pseudogene 2', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242122, CurrentID=0, Status='live', Symbol='Vtcn1', Aliases='B7h4, B7s1, B7x, BC032925', description='V-set domain containing T cell activation inhibitor 1', other_designations='B7 homolog 4|immune costimulatory protein B7-H4|t cell costimulatory molecule B7x|vtcn1', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=100825459, end_position_on_the_genomic_accession=100896922, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  242125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242125, CurrentID=0, Status='live', Symbol='Mab21l3', Aliases='', description='mab-21-like 3 (C. elegans)', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=101813076, end_position_on_the_genomic_accession=101836286, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  242126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242126, CurrentID=0, Status='live', Symbol='Slc22a15', Aliases='2610034P21Rik, 4930477M19, A530052I06Rik, BB219042', description='solute carrier family 22 (organic anion/cation transporter), member 15', other_designations='solute carrier family 22 (organic cation transporter), member 15', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=101855768, end_position_on_the_genomic_accession=101924453, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  111055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111055, CurrentID=0, Status='live', Symbol='Rrm2-ps3', Aliases='', description='ribonucleotide reductase M2 pseudogene 3', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111058, CurrentID=0, Status='live', Symbol='Hsp86-ps1', Aliases='86kDa, Hsp86-2, Hsp90', description='heat shock protein 86, pseudogene 1', other_designations='heat shock protein, 86 kDa, pseudogene 1|heat shock protein, pseudogene 1', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111065, CurrentID=0, Status='live', Symbol='Odc-rs1', Aliases='BRS-1, Odc-1', description='ornithine decarboxylase, related sequence 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111069, CurrentID=0, Status='live', Symbol='Odc-rs5', Aliases='BRS-5, Odc-5', description='ornithine decarboxylase, related sequence 5', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111070, CurrentID=0, Status='live', Symbol='Odc-rs6', Aliases='BRS-6, Odc-6', description='ornithine decarboxylase, related sequence 6', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111071, CurrentID=0, Status='live', Symbol='Odc-rs7', Aliases='BRS-7, Odc-7', description='ornithine decarboxylase, related sequence 7', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111072, CurrentID=0, Status='live', Symbol='Odc-rs9', Aliases='BRS-9, Odc-9', description='ornithine decarboxylase, related sequence 9', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  258811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258811, CurrentID=0, Status='live', Symbol='Olfr926', Aliases='MOR171-8', description='olfactory receptor 926', other_designations='GA_x6K02T2PVTD-32573036-32573962|olfactory receptor MOR171-8', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38877178, end_position_on_the_genomic_accession=38878104, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  242151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242151, CurrentID=0, Status='live', Symbol='Kcna10', Aliases='Gm1962, Kcna8, Kv1.8', description='potassium voltage-gated channel, shaker-related subfamily, member 10', other_designations='cyclic GMP gated potassium channel|potassium voltage gated channel, shaker related subfamily, member 8|potassium voltage-gated channel, shaker-related subfamily, member 8', map_location='3 46.64 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107183143, end_position_on_the_genomic_accession=107195721, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  18518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=18518, CurrentID=0, Status='live', Symbol='Igbp1', Aliases='AI115586, AI662457, AW208785, C81413, Pc52, Tap42, p52', description='immunoglobulin (CD79A) binding protein 1', other_designations='CD79a-binding protein 1|alpha 4|alpha4 phosphoprotein|lymphocyte signal transduction molecule alpha 4|phosphoprotein, component, 52 kDa|protein phosphatase 2/4/6 regulatory subunit', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100494291, end_position_on_the_genomic_accession=100516125, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  111118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111118, CurrentID=0, Status='live', Symbol='Odc-rs8', Aliases='BRS-8, Odc-8', description='ornithine decarboxylase, related sequence 8', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242202, CurrentID=0, Status='live', Symbol='Pde5a', Aliases='Cgbpde, Cn5n, PDE5a2, Pde51, Pde5a', description='phosphodiesterase 5A, cGMP-specific', other_designations='cGMP-binding cGMP-specific phosphodiesterase|phosphodiesterase 5A', map_location='3 G1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=122729158, end_position_on_the_genomic_accession=122859374, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  111150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111150, CurrentID=0, Status='live', Symbol='Odc-rs13', Aliases='Odc-13', description='ornithine decarboxylase, related sequence 13', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242235, CurrentID=0, Status='live', Symbol='Lrit3', Aliases='Gm421', description='leucine-rich repeat, immunoglobulin-like and transmembrane domains 3', other_designations='', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=129786929, end_position_on_the_genomic_accession=129808884, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  111172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111172, CurrentID=0, Status='live', Symbol='Tpi-rs3', Aliases='Tpi-3', description='triosephosphate isomerase related sequence 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111173, CurrentID=0, Status='live', Symbol='Erc1', Aliases='4930404L01Rik, 5033405M01Rik, 9630025C19Rik, B430107L16Rik, CAST2, ELKS, RAB6IP2A, RAB6IP2B, Rab6ip2, mKIAA1081', description='ELKS/RAB6-interacting/CAST family member 1', other_designations='CAZ-associated structural protein 2|RAB6-interacting protein 2|Rab6 interacting protein 2', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=119570796, end_position_on_the_genomic_accession=119849739, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  111174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111174, CurrentID=0, Status='live', Symbol='Taar1', Aliases='Tar1, Trar1, taR-1', description='trace amine-associated receptor 1', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=23920406, end_position_on_the_genomic_accession=23921404, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  111175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111175, CurrentID=0, Status='live', Symbol='Pecr', Aliases='2400003B18Rik, TERP', description='peroxisomal trans-2-enoyl-CoA reductase', other_designations='perosisomal 2-enoyl-CoA reductase', map_location='1 36.46 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72259172, end_position_on_the_genomic_accession=72284314, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  242248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242248, CurrentID=0, Status='live', Symbol='Bank1', Aliases='A530094C12Rik, AI451642, AVIEF, BANK', description='B cell scaffold protein with ankyrin repeats 1', other_designations='B-cell scaffold protein with ankyrin repeats 1 delta 2 transcript', map_location='3 63.04 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=136053363, end_position_on_the_genomic_accession=136326141, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  242253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242253, CurrentID=0, Status='live', Symbol='Wdr63', Aliases='4931433A13Rik', description='WD repeat domain 63', other_designations='', map_location='3 H2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=146040526, end_position_on_the_genomic_accession=146108107, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  102216272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=102216272, CurrentID=0, Status='live', Symbol='Ak6', Aliases='2810046E22Rik, 4921516M08Rik, AK/ATPase', description='adenylate kinase 6', other_designations='AK/ATPase|ATP-AMP transphosphorylase 6|coilin-interacting nuclear ATPase protein|dual activity adenylate kinase/ATPase', map_location='13 D1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=100651343, end_position_on_the_genomic_accession=100666415, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  242259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242259, CurrentID=0, Status='live', Symbol='Slc44a5', Aliases='AA117069, Gm422', description='solute carrier family 44, member 5', other_designations='', map_location='3 H3-H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=153973436, end_position_on_the_genomic_accession=154271747, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  105886298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105886298, CurrentID=0, Status='live', Symbol='Cmc4', Aliases='Mtcp1nb, p8MTCP1', description='C-x(9)-C motif containing 4', other_designations='MTCP-1 type A|mature T-cell proliferation 1 neighbor protein|mature T-cell proliferation-1 type A|protein p8 MTCP-1', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242274, CurrentID=0, Status='live', Symbol='Lrrc7', Aliases='AK122498, densin, mKIAA1365', description='leucine rich repeat containing 7', other_designations='densin 180', map_location='3 H4', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=158082888, end_position_on_the_genomic_accession=158562221, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  111205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111205, CurrentID=0, Status='live', Symbol='Tpi-rs2', Aliases='Tpi-2', description='triosephosphate isomerase related sequence 2', other_designations='triosephosphate isomerase 1 pseudogene', map_location='3 30.25 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=66128398, end_position_on_the_genomic_accession=66129752, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  111208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111208, CurrentID=0, Status='live', Symbol='Dhfr-rs1', Aliases='Dhfr-2', description='dihydrofolate reductase, related sequence 1', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242285, CurrentID=0, Status='live', Symbol='Sdr16c5', Aliases='Rdhe2, Scdr9', description='short chain dehydrogenase/reductase family 16C, member 5', other_designations='EPHD-2|RDH-E2|epidermal retinal dehydrogenase 2|retSDR2|retinal short chain dehydrogenase reductase (Rdhe2)|retinal short chain dehydrogenase reductase 2|retinal short-chain dehydrogenase reductase 2|short chain dehydrogenase reductase 9|short-chain dehydrogenase reductase 9|short-chain dehydrogenase/reductase family 16C member 5', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3995942, end_position_on_the_genomic_accession=4019663, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  242286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242286, CurrentID=0, Status='live', Symbol='Sdr16c6', Aliases='4833413O15Rik, Gm423', description='short chain dehydrogenase/reductase family 16C, member 6', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=4055926, end_position_on_the_genomic_accession=4077594, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  242291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242291, CurrentID=0, Status='live', Symbol='Impad1', Aliases='1110001C20Rik, AA408880, AI451589, AL022796, B230207P20, Jaws, gPAPP', description='inositol monophosphatase domain containing 1', other_designations='Golgi 3-prime phosphoadenosine 5-prime phosphate 3-prime phosphatase|IMP 3|IMPase 3|golgi-resident PAP phosphatase|inositol monophosphatase domain-containing protein 1|inositol-1(or 4)-monophosphatase 3|myo-inositol monophosphatase A3', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=4764342, end_position_on_the_genomic_accession=4793306, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  242297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242297, CurrentID=0, Status='live', Symbol='Fam110b', Aliases='1700012H17Rik, 5031405P22, AW049855', description='family with sequence similarity 110, member B', other_designations='', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=5644078, end_position_on_the_genomic_accession=5801101, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  111234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111234, CurrentID=0, Status='live', Symbol='Ppib-rs2', Aliases='Cphn2-rs2', description='peptidylprolyl isomerase B, related sequence 2', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111236, CurrentID=0, Status='live', Symbol='Hras1-ps1', Aliases='Hras2', description='Harvey rat sarcoma virus oncogene, pseudogene 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111241, CurrentID=0, Status='live', Symbol='Hmga1-rs1', Aliases='ENSMUSG00000041059, Hmgi-rs1', description='high mobility group AT-hook I, related sequence 1', other_designations='HMG-I(Y)|high mobility group AT-hook protein 1|high mobility group protein A1', map_location='11 84.51 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=120762794, end_position_on_the_genomic_accession=120764392, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  242316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242316, CurrentID=0, Status='live', Symbol='Gdf6', Aliases='BMP-13, BMP13, GDF16', description='growth differentiation factor 6', other_designations='GDF-6|bone morphogenetic protein 13|growth differentiation factor 16|growth/differentiation factor 16', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=9844372, end_position_on_the_genomic_accession=9862345, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  242341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242341, CurrentID=0, Status='live', Symbol='Atp6v0d2', Aliases='1620401A02Rik, AI324824, V-ATPase', description='ATPase, H+ transporting, lysosomal V0 subunit D2', other_designations='ATPase, H+ transporting, V0 subunit D|V-ATPase subunit d 2|osteoclast-specific vacuolar ATP synthase|vacuolar proton pump subunit d 2|vacuolar proton-translocating ATPase d subunit d2', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=19876838, end_position_on_the_genomic_accession=19922566, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  111288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111288, CurrentID=0, Status='live', Symbol='Hmgn1-rs7', Aliases='Hmg14-rs7', description='high mobility group nucleosomal binding domain 1, related sequence 7', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242362, CurrentID=0, Status='live', Symbol='Manea', Aliases='4932703L02Rik', description='mannosidase, endo-alpha', other_designations='endo-alpha mannosidase|endomannosidase|mEndo', map_location='4 A3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=26324506, end_position_on_the_genomic_accession=26346652, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  111299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111299, CurrentID=0, Status='live', Symbol='Hmgn1-rs4', Aliases='Hmg14-rs4', description='high mobility group nucleosomal binding domain 1, related sequence 4', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242377, CurrentID=0, Status='live', Symbol='Pm20d2', Aliases='Acy1l2, Gm424', description='peptidase M20 domain containing 2', other_designations='aminoacylase 1-like 2|aminoacylase 1-like protein 2|aminoacylase-1-like protein 2', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=33170401, end_position_on_the_genomic_accession=33189737, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  242384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242384, CurrentID=0, Status='live', Symbol='Lingo2', Aliases='B230217C06Rik, Lern3, Lrrn6c', description='leucine rich repeat and Ig domain containing 2', other_designations='leucine rich repeat neuronal 6C|leucine-rich repeat neuronal protein 6C', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=35706647, end_position_on_the_genomic_accession=36952254, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  111326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111326, CurrentID=0, Status='live', Symbol='Ddx5-ps1', Aliases='Ddx5-rs1, Hlr1-ps1, Hlr1-rs1', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 5, pseudogene 1', other_designations='DEAD (aspartate-glutamate-alanine-aspartate) box polypeptide 5, pseudogene 1|DEAD/H (Asp-Glu-Ala-Asp/His) box polypeptide 5, pseudogene 1', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111333, CurrentID=0, Status='live', Symbol='Amy', Aliases='', description='amylase complex', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242406, CurrentID=0, Status='live', Symbol='Rgp1', Aliases='1110029E03Rik, AI527196, F730001J03, mKIAA0258', description='RGP1 retrograde golgi transport homolog (S. cerevisiae)', other_designations='retrograde Golgi transport protein RGP1 homolog', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43578735, end_position_on_the_genomic_accession=43587487, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  111335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111335, CurrentID=0, Status='live', Symbol='Cat2', Aliases='Cat-2, Nzc, Rop, Scat', description='dominant cataract 2', other_designations='nuclear and zonular cataract', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111336, CurrentID=0, Status='live', Symbol='Hoxa', Aliases='Hox-1', description='homeobox A cluster', other_designations='homeo box A cluster', map_location='6', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111337, CurrentID=0, Status='live', Symbol='Hoxd', Aliases='Hox-4, Hox-5, Ul, ulnaless', description='homeobox D cluster', other_designations='homeo box D cluster', map_location='2', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111339, CurrentID=0, Status='live', Symbol='Hoxc', Aliases='Hox-3, Hox-6', description='homeobox C cluster', other_designations='homeo box C cluster', map_location='15', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  237010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237010, CurrentID=0, Status='live', Symbol='Klhl4', Aliases='B130014A11, C130018J01Rik', description='kelch-like 4', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=114474333, end_position_on_the_genomic_accession=114561129, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  111342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111342, CurrentID=0, Status='live', Symbol='Surf', Aliases='', description='surfeit gene cluster', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111343, CurrentID=0, Status='live', Symbol='Il1', Aliases='Il-1', description='interleukin 1 complex', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111345, CurrentID=0, Status='live', Symbol='Saa', Aliases='', description='serum amyloid A cluster', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242418, CurrentID=0, Status='live', Symbol='Dcaf10', Aliases='AA959934, Wdr32', description='DDB1 and CUL4 associated factor 10', other_designations='WD repeat domain 32|WD repeat-containing protein 32', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45341735, end_position_on_the_genomic_accession=45379732, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  242425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242425, CurrentID=0, Status='live', Symbol='Gabbr2', Aliases='GABABR2, Gb2, Gm425, Gpr51', description='gamma-aminobutyric acid (GABA) B receptor, 2', other_designations='G protein-coupled receptor 51|G-protein coupled receptor 51|GABA-B receptor 2|GABA-B-R2|GABA-BR2|gamma-aminobutyric acid type B receptor, subunit 2|ortholog of human G protein-coupled receptor 51 GPR51', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46663898, end_position_on_the_genomic_accession=46991714, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  193322 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193322, CurrentID=0, Status='live', Symbol='Oog1', Aliases='Oog, c-1', description='oogenesin 1', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87552648, end_position_on_the_genomic_accession=87608845, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  111359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111359, CurrentID=0, Status='live', Symbol='Cryg', Aliases='Len-1', description='crystallin gamma complex', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111364, CurrentID=0, Status='live', Symbol='H2', Aliases='H-2, MHC-II', description='histocompatibility-2, MHC', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111368, CurrentID=0, Status='live', Symbol='Prn', Aliases='Gm28045p-Prnd, Prn', description='prion protein gene complex', other_designations='Prnp-Prnd readthrough', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=131909928, end_position_on_the_genomic_accession=131956131, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  111369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111369, CurrentID=0, Status='live', Symbol='Mov15', Aliases='Mov-15', description='Moloney leukemia virus 15', other_designations='', map_location='XY 76.0 cM', chromosome='X, Y', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111370, CurrentID=0, Status='live', Symbol='Cyp2d', Aliases='', description='cytochrome P450, 2d region', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242443, CurrentID=0, Status='live', Symbol='Grin3a', Aliases='6430537F04, A830097C19Rik, NMDAR-L, NR3A, mKIAA1973', description='glutamate receptor ionotropic, NMDA3A', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=49661611, end_position_on_the_genomic_accession=49846003, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  111372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111372, CurrentID=0, Status='live', Symbol='Emv2', Aliases='Blv-1, Bv-1, Emv-2, Inb-1, N-tropic', description='endogenous ecotropic MuLV 2', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111373, CurrentID=0, Status='live', Symbol='Emv3', Aliases='Emv-3, Sev-1', description='endogenous ecotropic MuLV 3', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111374, CurrentID=0, Status='live', Symbol='Mtv1', Aliases='Mtv-1', description='mammary tumor virus locus 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111375, CurrentID=0, Status='live', Symbol='Mtv7', Aliases='Mls, Mls-1, Mls1, Mtv-10, Mtv-19, Mtv-22, Mtv-7, Pyv', description='mammary tumor virus locus 7', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111376, CurrentID=0, Status='live', Symbol='Mtv8', Aliases='Mtv-16, Mtv-8', description='mammary tumor virus locus 8', other_designations='Unit 11', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111378 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111378, CurrentID=0, Status='live', Symbol='Mtv2', Aliases='Mtv-2', description='mammary tumor virus locus 2', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111379, CurrentID=0, Status='live', Symbol='Mtv3', Aliases='Mtv-18, Mtv-3', description='mammary tumor virus locus 3', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111380, CurrentID=0, Status='live', Symbol='Mtv6', Aliases='Mls-3, Mls3, Mtv-6', description='mammary tumor virus locus 6', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111381, CurrentID=0, Status='live', Symbol='Mtv9', Aliases='Mtv-9', description='mammary tumor virus locus 9', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111382, CurrentID=0, Status='live', Symbol='Pad1', Aliases='Pad-1', description='MMTV LTR integration site 1', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111383, CurrentID=0, Status='live', Symbol='Mar1', Aliases='L2a', description='matrix associated region 1', other_designations='matrix-associated region  1', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111384, CurrentID=0, Status='live', Symbol='Emv11', Aliases='Akv-1, Emv-11', description='endogenous ecotropic MuLV 11', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111386, CurrentID=0, Status='live', Symbol='Rnr4', Aliases='', description='ribosomal RNA, Chr 4 cluster', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111388, CurrentID=0, Status='live', Symbol='Emv13', Aliases='Akv-3, Akvp, Emv-13', description='endogenous ecotropic MuLV 13', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111389, CurrentID=0, Status='live', Symbol='Emv16', Aliases='Emv-16', description='endogenous ecotropic MuLV 16', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111390, CurrentID=0, Status='live', Symbol='Emv17', Aliases='Emv-17', description='endogenous ecotropic MuLV 17', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111391, CurrentID=0, Status='live', Symbol='Emv18', Aliases='Emv-18', description='endogenous ecotropic MuLV 18', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242466, CurrentID=0, Status='live', Symbol='Zfp462', Aliases='6030417H05, 9430078C22Rik, Gt4-2, Zfpip', description='zinc finger protein 462', other_designations='gene trap insertion site 4-2', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=54944583, end_position_on_the_genomic_accession=55083563, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  111395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111395, CurrentID=0, Status='live', Symbol='Emv23', Aliases='C58v-4, Emv-23', description='endogenous ecotropic MuLV 23', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111397, CurrentID=0, Status='live', Symbol='Emv26', Aliases='C58v-1, Emv-26', description='endogenous ecotropic MuLV 26', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111398, CurrentID=0, Status='live', Symbol='Emv27', Aliases='Emv-19, Emv-27', description='endogenous ecotropic MuLV 27', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242474, CurrentID=0, Status='live', Symbol='Tmem245', Aliases='A630051L19Rik, AI957324, D730040F13Rik', description='transmembrane protein 245', other_designations='likely ortholog of H. sapiens chromosome 9 open reading frame 5 (C9orf5)|transmembrane 245', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=56876003, end_position_on_the_genomic_accession=56947527, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  242481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242481, CurrentID=0, Status='live', Symbol='Palm2', Aliases='53-H12, B130066I24, R75368, pEN70', description='paralemmin 2', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=57568248, end_position_on_the_genomic_accession=57717146, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  242484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242484, CurrentID=0, Status='live', Symbol='D630039A03Rik', Aliases='', description='RIKEN cDNA D630039A03 gene', other_designations='', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=57908384, end_position_on_the_genomic_accession=57916264, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  635702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=635702, CurrentID=0, Status='live', Symbol='Naaladl2', Aliases='2810043G22Rik, EG635702, Gm1021', description='N-acetylated alpha-linked acidic dipeptidase-like 2', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=23798103, end_position_on_the_genomic_accession=24784006, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  111415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111415, CurrentID=0, Status='live', Symbol='Mtv13', Aliases='Mls, Mls-2, Mls2, Mtv-13', description='mammary tumor virus locus 13', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111416, CurrentID=0, Status='live', Symbol='Mtv14', Aliases='Mtv-14', description='mammary tumor virus locus 14', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111417, CurrentID=0, Status='live', Symbol='Mtv17', Aliases='Mtv-15, Mtv-17, Mtv-20', description='mammary tumor virus locus 17', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111418, CurrentID=0, Status='live', Symbol='Mtv23', Aliases='Mtv-23', description='mammary tumor virus locus 23', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111424, CurrentID=0, Status='live', Symbol='Mtv35', Aliases='Mtv-35', description='mammary tumor virus locus 35', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111428, CurrentID=0, Status='live', Symbol='Mtv44', Aliases='Mtv-44', description='mammary tumor virus locus 44', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242502, CurrentID=0, Status='live', Symbol='Gm428', Aliases='', description='predicted gene 428', other_designations='', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=73681151, end_position_on_the_genomic_accession=73687559, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  111431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111431, CurrentID=0, Status='live', Symbol='Emv12', Aliases='Akv-2, Emv-12', description='endogenous ecotropic MuLV 12', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111432, CurrentID=0, Status='live', Symbol='Emv14', Aliases='Akv-2J, Akv-4, Emv-14', description='endogenous ecotropic MuLV 14', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242505, CurrentID=0, Status='live', Symbol='Rasef', Aliases='BC030882, RAB45', description='RAS and EF hand domain containing', other_designations='', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=73714577, end_position_on_the_genomic_accession=73791401, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  242506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242506, CurrentID=0, Status='live', Symbol='Frmd3', Aliases='4.1O, 9430066I12Rik, EPB41L4O, P410', description='FERM domain containing 3', other_designations='', map_location='4 C3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=74013603, end_position_on_the_genomic_accession=74202214, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  242509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242509, CurrentID=0, Status='live', Symbol='Bnc2', Aliases='5031434M05Rik, 8430420F16Rik', description='basonuclin 2', other_designations='bnc2', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=84266804, end_position_on_the_genomic_accession=84676342, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  242517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242517, CurrentID=0, Status='live', Symbol='Ifna15', Aliases='Gm12597, Ifnaa, MuIFN-alpha-A, OTTMUSG00000007655', description='interferon alpha 15', other_designations='novel interferon alpha (Ifna) family member', map_location='4 C4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88557673, end_position_on_the_genomic_accession=88558245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  111447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111447, CurrentID=0, Status='live', Symbol='Mtv11', Aliases='Mtv-11, Mtv-12, Mtv-7a', description='mammary tumor virus locus 11', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  667118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667118, CurrentID=0, Status='live', Symbol='Zbed6', Aliases='Gm8466, MGR', description='zinc finger, BED type containing 6', other_designations='muscle growth regulator|zbed6|zinc finger, BED domain containing 6', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133655879, end_position_on_the_genomic_accession=133660885, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  242523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242523, CurrentID=0, Status='live', Symbol='Dmrta1', Aliases='Dmrt4', description='doublesex and mab-3 related transcription factor like family A1', other_designations='doublesex and mab-3 related transcription factor 4|doublesex- and mab-3-related transcription factor 4', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=89688198, end_position_on_the_genomic_accession=89694766, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  111452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111452, CurrentID=0, Status='live', Symbol='Krt1c', Aliases='Krt-2, Krt2', description='keratin 1 complex', other_designations='keratin gene complex 2, basic', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111453, CurrentID=0, Status='live', Symbol='Rnr11', Aliases='', description='ribosomal RNA, Chr 11 cluster', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111454, CurrentID=0, Status='live', Symbol='Rnr12', Aliases='', description='ribosomal RNA, Chr 12 cluster', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111455, CurrentID=0, Status='live', Symbol='Rnr15', Aliases='', description='ribosomal RNA, Chr 15 cluster', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111456 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111456, CurrentID=0, Status='live', Symbol='Rnr16', Aliases='', description='ribosomal RNA, Chr 16 cluster', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111457, CurrentID=0, Status='live', Symbol='Rnr18', Aliases='', description='ribosomal RNA, Chr 18 cluster', other_designations='', map_location='', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111458, CurrentID=0, Status='live', Symbol='Qa2', Aliases='H2-Qa2, Ped, Qa-2', description='Qa lymphocyte antigen 2 region', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111459, CurrentID=0, Status='live', Symbol='Rnr13', Aliases='', description='rRNA, Chr 13 cluster', other_designations='ribosomal RNA, Chr 13 cluster', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111462, CurrentID=0, Status='live', Symbol='H2-Q', Aliases='H-2Q', description='histocompatibility 2, Q region', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111463, CurrentID=0, Status='live', Symbol='H2-T', Aliases='H-2T', description='histocompatibility 2, T region', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111465, CurrentID=0, Status='live', Symbol='L1Md-A1', Aliases='L1Md-9(Hbb), L1Md9-1', description='L1 repeat, A subfamily, member 1', other_designations='L1 repeat in Hbb<d> haplotype|LINE-1 element ORF1', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111470, CurrentID=0, Status='live', Symbol='Xmmv2', Aliases='Env-2, Xmmv-2', description='xenotropic-MCF leukemia virus 2', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111473, CurrentID=0, Status='live', Symbol='Xmmv9', Aliases='Env-9, Xmmv-61, Xmmv-9, Xp-20', description='xenotropic-MCF leukemia virus 9', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242546, CurrentID=0, Status='live', Symbol='Cyp2j12', Aliases='BB197342-ps, OTTMUSG00000007939, Cyp2j12', description='cytochrome P450, family 2, subfamily j, polypeptide 12', other_designations='', map_location='4 C5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=96099314, end_position_on_the_genomic_accession=96141272, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  242553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242553, CurrentID=0, Status='live', Symbol='Kank4', Aliases='AU021071, Ankrd38, BC060737, C130031J23', description='KN motif and ankyrin repeat domains 4', other_designations='ankyrin repeat domain 38|ankyrin repeat domain-containing protein 38', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=98754892, end_position_on_the_genomic_accession=98817800, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  111484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111484, CurrentID=0, Status='live', Symbol='Gsta', Aliases='GSTs, Gst-2', description='glutathione S-transferase cluster', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242557, CurrentID=0, Status='live', Symbol='Atg4c', Aliases='Apg4-C, Apg4cl, Autl1, Atg4c', description='autophagy related 4C, cysteine peptidase', other_designations='APG4 (ATG4) autophagy-related homolog C|AUT-like 1, cysteine endopeptidase|AUT-like 3 cysteine endopeptidase|autophagin 3|autophagin-3|autophagy-related 4C|autophagy-related cysteine endopeptidase 3|autophagy-related protein 4 homolog C|cysteine protease involved in autophagy', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=99190891, end_position_on_the_genomic_accession=99297258, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  111491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111491, CurrentID=0, Status='live', Symbol='Xmmv42', Aliases='Xmmv-42, Xp-1', description='xenotropic-MCF leukemia virus 42', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111495, CurrentID=0, Status='live', Symbol='Xmmv65', Aliases='Xmmv-65, Xp-24', description='xenotropic-MCF leukemia virus 65', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111496 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111496, CurrentID=0, Status='live', Symbol='Xmmv71', Aliases='Mmv-71, Xmmv-71', description='xenotropic-MCF leukemia virus 71', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242570, CurrentID=0, Status='live', Symbol='Raver2', Aliases='A430091O22Rik', description='ribonucleoprotein, PTB-binding 2', other_designations='protein raver-2', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101069038, end_position_on_the_genomic_accession=101152370, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  111501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111501, CurrentID=0, Status='live', Symbol='Rnr19-1', Aliases='Rnr19', description='ribosomal RNA, Chr 19-1 cluster', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111503, CurrentID=0, Status='live', Symbol='Pmv1', Aliases='Pmv-1', description='polytropic murine leukemia virus 1', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111504, CurrentID=0, Status='live', Symbol='Pmv4', Aliases='Pmv-4', description='polytropic murine leukemia virus 4', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111507, CurrentID=0, Status='live', Symbol='Igh', Aliases='', description='immunoglobulin heavy chain complex', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113258768, end_position_on_the_genomic_accession=116009954, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  111508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111508, CurrentID=0, Status='live', Symbol='Pmv2', Aliases='Pmv-2', description='polytropic murine leukemia virus 2', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111509, CurrentID=0, Status='live', Symbol='Pmv5', Aliases='Pmv-5', description='polytropic murine leukemia virus 5', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111511, CurrentID=0, Status='live', Symbol='Pmv7', Aliases='Pmv-7', description='polytropic murine leukemia virus 7', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242584, CurrentID=0, Status='live', Symbol='Wdr78', Aliases='BC028975', description='WD repeat domain 78', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=103034757, end_position_on_the_genomic_accession=103114299, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  242585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242585, CurrentID=0, Status='live', Symbol='Slc35d1', Aliases='AI834976, C330011J09, UGTREL7, mKIAA0260', description='solute carrier family 35 (UDP-glucuronic acid/UDP-N-acetylgalactosamine dual transporter), member D1', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=103170649, end_position_on_the_genomic_accession=103217826, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  111515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111515, CurrentID=0, Status='live', Symbol='Pmv3', Aliases='Pmv-3', description='polytropic murine leukemia virus 3', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111517, CurrentID=0, Status='live', Symbol='Pmv9', Aliases='Pmv-9', description='polytropic murine leukemia virus 9', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111518, CurrentID=0, Status='live', Symbol='Cea', Aliases='', description='carcinoembryonic antigen gene family', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111519, CurrentID=0, Status='live', Symbol='Igl', Aliases='Igl-x', description='immunoglobulin lambda chain complex', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19026858, end_position_on_the_genomic_accession=19260844, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  111520 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111520, CurrentID=0, Status='live', Symbol='Pmv18', Aliases='Env-35, Pmv-18, Xmmv-35', description='polytropic murine leukemia virus 18', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242594, CurrentID=0, Status='live', Symbol='1700024P16Rik', Aliases='C1orf168, Gm620', description='RIKEN cDNA 1700024P16 gene', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=104833834, end_position_on_the_genomic_accession=105016864, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  111524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111524, CurrentID=0, Status='live', Symbol='Tcra-V1', Aliases='', description='T cell receptor alpha, variable 1', other_designations='T-cell receptor alpha, variable 1', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111525, CurrentID=0, Status='live', Symbol='Tcra-V3', Aliases='', description='T cell receptor alpha, variable 3', other_designations='T-cell receptor alpha, variable 3', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111526, CurrentID=0, Status='live', Symbol='Tcra-V4', Aliases='', description='T cell receptor alpha, variable 4', other_designations='T-cell receptor alpha, variable 4', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111527, CurrentID=0, Status='live', Symbol='Tcra-V5', Aliases='', description='T cell receptor alpha, variable 5', other_designations='T-cell receptor alpha, variable 5', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111528, CurrentID=0, Status='live', Symbol='Tcra-V6', Aliases='', description='T cell receptor alpha, variable 6', other_designations='T-cell receptor alpha, variable 6', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111529, CurrentID=0, Status='live', Symbol='Tcra-V7', Aliases='', description='T cell receptor alpha, variable 7', other_designations='T-cell receptor alpha, variable 7', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242602, CurrentID=0, Status='live', Symbol='Lexm', Aliases='', description='lymphocyte expansion molecule', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106588958, end_position_on_the_genomic_accession=106617500, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  242603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242603, CurrentID=0, Status='live', Symbol='Cdcp2', Aliases='D030010E02Rik', description='CUB domain containing protein 2', other_designations='', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107096891, end_position_on_the_genomic_accession=107114304, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  111532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111532, CurrentID=0, Status='live', Symbol='Xmv26', Aliases='Xmv-26', description='xenotropic murine leukemia virus 26', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111533, CurrentID=0, Status='live', Symbol='Xmv45', Aliases='Xmv-45', description='xenotropic murine leukemia virus 45', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242607, CurrentID=0, Status='live', Symbol='Slc1a7', Aliases='A930031E15Rik, Eaat5', description='solute carrier family 1 (glutamate transporter), member 7', other_designations='solute carrier family 1 member 7', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=107968334, end_position_on_the_genomic_accession=108013532, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  111536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111536, CurrentID=0, Status='live', Symbol='Pmv11', Aliases='Pmv-11', description='polytropic murine leukemia virus 11', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111537, CurrentID=0, Status='live', Symbol='Pmv12', Aliases='Pmv-12', description='polytropic murine leukemia virus 12', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111538, CurrentID=0, Status='live', Symbol='Pmv15', Aliases='Pmv-15', description='polytropic murine leukemia virus 15', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111539 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111539, CurrentID=0, Status='live', Symbol='Pmv19', Aliases='Pmv-19', description='polytropic murine leukemia virus 19', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111540 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111540, CurrentID=0, Status='live', Symbol='Pmv22', Aliases='Pmv-22', description='polytropic murine leukemia virus 22', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111541, CurrentID=0, Status='live', Symbol='Pmv23', Aliases='Pmv-23', description='polytropic murine leukemia virus 23', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111542, CurrentID=0, Status='live', Symbol='Pmv24', Aliases='Pmv-24', description='polytropic murine leukemia virus 24', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111547, CurrentID=0, Status='live', Symbol='Pmv31', Aliases='Pmv-31', description='polytropic murine leukemia virus 31', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242620, CurrentID=0, Status='live', Symbol='Dmrta2', Aliases='Dmrt5', description='doublesex and mab-3 related transcription factor like family A2', other_designations='doublesex- and mab-3-related transcription factor 5', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=109978006, end_position_on_the_genomic_accession=109983684, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  111560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111560, CurrentID=0, Status='live', Symbol='Xmv10', Aliases='Xmv-10', description='xenotropic murine leukemia virus 10', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111561, CurrentID=0, Status='live', Symbol='Xmv12', Aliases='Xmv-12', description='xenotropic murine leukemia virus 12', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111562, CurrentID=0, Status='live', Symbol='Xmv15', Aliases='Xmv-15', description='xenotropic murine leukemia virus 15', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111563, CurrentID=0, Status='live', Symbol='Xmv16', Aliases='Xmv-16', description='xenotropic murine leukemia virus 16', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111565, CurrentID=0, Status='live', Symbol='Xmv24', Aliases='Xmv-24', description='xenotropic murine leukemia virus 24', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242642, CurrentID=0, Status='live', Symbol='Hpdl', Aliases='A830048M07Rik, BC034099, Gloxd1', description='4-hydroxyphenylpyruvate dioxygenase-like', other_designations='glyoxalase domain containing 1|glyoxalase domain-containing protein 1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=116819907, end_position_on_the_genomic_accession=116821508, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  111573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111573, CurrentID=0, Status='live', Symbol='Xmv41', Aliases='Xmv-41', description='xenotropic murine leukemia virus 41', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242646, CurrentID=0, Status='live', Symbol='Tctex1d4', Aliases='4833401D15Rik', description='Tctex1 domain containing 4', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117124524, end_position_on_the_genomic_accession=117128946, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  242653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242653, CurrentID=0, Status='live', Symbol='Cldn19', Aliases='', description='claudin 19', other_designations='', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119255441, end_position_on_the_genomic_accession=119262438, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  111582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111582, CurrentID=0, Status='live', Symbol='Pmv14', Aliases='Pmv-14', description='polytropic murine leukemia virus 14', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111583, CurrentID=0, Status='live', Symbol='Pmv16', Aliases='Pmv-16', description='polytropic murine leukemia virus 16', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111584, CurrentID=0, Status='live', Symbol='Pmv17', Aliases='Pmv-17, Pol-5, Xmmv-55, Xmmv-72, Xmmv-75, Xp-14', description='polytropic murine leukemia virus 17', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111586, CurrentID=0, Status='live', Symbol='Pmv21', Aliases='Pmv-21', description='polytropic murine leukemia virus 21', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242662, CurrentID=0, Status='live', Symbol='Rims3', Aliases='A730060M23Rik, Nim3, Rim3, mKIAA0237', description='regulating synaptic membrane exocytosis 3', other_designations='RIM 3|RIM3 gamma|rab-3-interacting molecule 3', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120854779, end_position_on_the_genomic_accession=120896580, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  111593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111593, CurrentID=0, Status='live', Symbol='Pmv46', Aliases='Pmv-46', description='polytropic murine leukemia virus 46', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242667, CurrentID=0, Status='live', Symbol='Dlgap3', Aliases='BC058433, DAP-3, DAP3, Prpl8, Sapap3', description='discs, large (Drosophila) homolog-associated protein 3', other_designations='PSD-95/SAP90-binding protein 3|SAP90/PSD 95 associated protein 3|prion protein ligand 8', map_location='4 61.33 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=127169204, end_position_on_the_genomic_accession=127237022, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  242669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242669, CurrentID=0, Status='live', Symbol='Azin2', Aliases='4933429I20Rik, Adc, Azi2, B930082O19, ODC-p, Odcp', description='antizyme inhibitor 2', other_designations='ODC antizyme inhibitor-2|ODC-like protein|arginine decarboxylase|ornithine decarboxylase paralog|ornithine decarboxylase-like protein', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128930233, end_position_on_the_genomic_accession=128962455, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  111602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111602, CurrentID=0, Status='live', Symbol='Xmv13', Aliases='Xmv-13', description='xenotropic murine leukemia virus 13', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111603, CurrentID=0, Status='live', Symbol='Xmv18', Aliases='Xmv-18', description='xenotropic murine leukemia virus 18', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111604, CurrentID=0, Status='live', Symbol='Xmv19', Aliases='Xmv-19', description='xenotropic murine leukemia virus 19', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242681, CurrentID=0, Status='live', Symbol='Rab42', Aliases='9530096D07Rik-ps, Rab42', description='RAB42, member RAS oncogene family', other_designations='RAB42, member RAS oncogene family, pseudogene', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132300438, end_position_on_the_genomic_accession=132303403, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  111610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111610, CurrentID=0, Status='live', Symbol='Xmv31', Aliases='Xmv-31', description='xenotropic murine leukemia virus 31', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111613 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111613, CurrentID=0, Status='live', Symbol='Xmv42', Aliases='Xmv-42', description='xenotropic murine leukemia virus 42', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242687, CurrentID=0, Status='live', Symbol='Wasf2', Aliases='AW742646, D4Ertd13e, WAVE2', description='WAS protein family, member 2', other_designations='WASP family protein member 2|protein WAVE-2', map_location='4 66.22 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133130527, end_position_on_the_genomic_accession=133199756, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  242691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242691, CurrentID=0, Status='live', Symbol='Gpatch3', Aliases='D930035B09Rik, Gpatc3', description='G patch domain containing 3', other_designations='', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=133574745, end_position_on_the_genomic_accession=133584242, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  111620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111620, CurrentID=0, Status='live', Symbol='Tcra-V11', Aliases='', description='T cell receptor alpha, variable 11', other_designations='T-cell receptor alpha, variable 11', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111622, CurrentID=0, Status='live', Symbol='Tcra-V14', Aliases='Valpha14', description='T cell receptor alpha, variable 14', other_designations='T-cell receptor alpha, variable 14', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242700, CurrentID=0, Status='live', Symbol='Ifnlr1', Aliases='CRF2-12, Il28ra', description='interferon lambda receptor 1', other_designations='IFN-lambda R1|IL-28 receptor subunit alpha|IL-28R-alpha|IL-28RA|class II cytokine receptor 12|cytokine receptor class-II member 12|cytokine receptor family 2 member 12|interferon-lambda receptor 1|interleukin 28 receptor alpha|interleukin-28 receptor subunit alpha', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135686457, end_position_on_the_genomic_accession=135708181, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  242702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242702, CurrentID=0, Status='live', Symbol='Myom3', Aliases='8430427K15Rik, Gm434', description='myomesin family, member 3', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135759480, end_position_on_the_genomic_accession=135815567, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  242705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242705, CurrentID=0, Status='live', Symbol='E2f2', Aliases='E2F-2', description='E2F transcription factor 2', other_designations='', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=136172394, end_position_on_the_genomic_accession=136196057, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  242721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242721, CurrentID=0, Status='live', Symbol='Klhdc7a', Aliases='B230308G19Rik', description='kelch domain containing 7A', other_designations='novel kelch domain containing protein', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=139962173, end_position_on_the_genomic_accession=139968026, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  111654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111654, CurrentID=0, Status='live', Symbol='Ifna', Aliases='Ifa, Ifa8', description='interferon alpha', other_designations='interferon alpha gene family, leukocyte', map_location='4', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111660, CurrentID=0, Status='live', Symbol='Tcra-V', Aliases='', description='T cell receptor alpha, variable region', other_designations='T-cell receptor alpha, variable region', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111662, CurrentID=0, Status='live', Symbol='Iap11-1', Aliases='Iap(Il-3)', description='intracisternal A particles, Il3 linked', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242735, CurrentID=0, Status='live', Symbol='Lrrc38', Aliases='A230053A07Rik', description='leucine rich repeat containing 38', other_designations='BK channel auxiliary gamma subunit LRRC38|BK channel auxilliary gamma subunit LRRC38', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143346981, end_position_on_the_genomic_accession=143371028, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  242736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242736, CurrentID=0, Status='live', Symbol='Pramef8', Aliases='4732496O08Rik, AU023234', description='PRAME family member 8', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143412309, end_position_on_the_genomic_accession=143421091, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  242737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242737, CurrentID=0, Status='live', Symbol='Oog4', Aliases='E330017M15', description='oogenesin 4', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143437164, end_position_on_the_genomic_accession=143450308, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  111667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111667, CurrentID=0, Status='live', Symbol='Igh-VQ52', Aliases='', description='immunoglobulin heavy chain (Q52 family)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111671, CurrentID=0, Status='live', Symbol='Myhc', Aliases='', description='myosin heavy chain, cardiac muscle complex', other_designations='', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111672, CurrentID=0, Status='live', Symbol='Igh-C', Aliases='', description='immunoglobulin heavy chain constant region', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111674, CurrentID=0, Status='live', Symbol='Minta', Aliases='', description='Moloney MuLV integration site, EC cells, a', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242747, CurrentID=0, Status='live', Symbol='Zfp933', Aliases='2810408P10Rik', description='zinc finger protein 933', other_designations='zinc finger-like protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147822986, end_position_on_the_genomic_accession=147848375, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  242748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242748, CurrentID=0, Status='live', Symbol='Ptchd2', Aliases='G630052C06Rik, RNDEu-2', description='patched domain containing 2', other_designations='RND-type protein RNDEu-2', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148236433, end_position_on_the_genomic_accession=148288161, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  111677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111677, CurrentID=0, Status='live', Symbol='Ms15-1', Aliases='', description='minisatellite 15-1 detected by probe 33.15', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111683, CurrentID=0, Status='live', Symbol='Mpmv1', Aliases='Mpmv-1', description='modified polytropic murine leukemia virus 1', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111684, CurrentID=0, Status='live', Symbol='Mpmv5', Aliases='Mpmv-5', description='modified polytropic murine leukemia virus 5', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111685, CurrentID=0, Status='live', Symbol='Mpmv6', Aliases='Mpmv-6', description='modified polytropic murine leukemia virus 6', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111686, CurrentID=0, Status='live', Symbol='Mpmv7', Aliases='Env-5, Mpmv-7, Xmmv-5', description='modified polytropic murine leukemia virus 7', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111687, CurrentID=0, Status='live', Symbol='Mpmv9', Aliases='Mpmv-9', description='modified polytropic murine leukemia virus 9', other_designations='', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111695, CurrentID=0, Status='live', Symbol='Igh-D', Aliases='', description='immunoglobulin heavy chain diversity region', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111697, CurrentID=0, Status='live', Symbol='Mpmv2', Aliases='Mpmv-2', description='modified polytropic murine leukemia virus 2', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111698, CurrentID=0, Status='live', Symbol='Mpmv4', Aliases='Env-3, Mpmv-4, Xmmv-3', description='modified polytropic murine leukemia virus 4', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111699, CurrentID=0, Status='live', Symbol='Mpmv8', Aliases='Mpmv-8', description='modified polytropic murine leukemia virus 8', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111700, CurrentID=0, Status='live', Symbol='Igl-J', Aliases='', description='immunoglobulin lambda chain, joining region', other_designations='', map_location='', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242773, CurrentID=0, Status='live', Symbol='Slc45a1', Aliases='C230078B22, Dnb5', description='solute carrier family 45, member 1', other_designations='DNb-5 homolog|PAST-A|deleted in neuroblastoma 5 protein homolog', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=150629396, end_position_on_the_genomic_accession=150652174, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  111705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111705, CurrentID=0, Status='live', Symbol='Igh-V3609', Aliases='', description='immunoglobulin heavy chain (V3609 family)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111707, CurrentID=0, Status='live', Symbol='Mpmv14', Aliases='Mpmv-14', description='modified polytropic murine leukemia virus 14', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111709, CurrentID=0, Status='live', Symbol='Mpmv19', Aliases='Mpmv-19', description='modified polytropic murine leukemia virus 19', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111712, CurrentID=0, Status='live', Symbol='Mpmv23', Aliases='Mpmv-23, Xmmv-52, Xp-11', description='modified polytropic murine leukemia virus 23', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242785, CurrentID=0, Status='live', Symbol='Klhl21', Aliases='1810045K06Rik, D330008A20, mKIAA0469', description='kelch-like 21', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152008891, end_position_on_the_genomic_accession=152017677, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  111714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111714, CurrentID=0, Status='live', Symbol='Mpmv27', Aliases='Mpmv-27', description='modified polytropic murine leukemia virus 27', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111715, CurrentID=0, Status='live', Symbol='Mpmv28', Aliases='Mpmv-28', description='modified polytropic murine leukemia virus 28', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  373864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=373864, CurrentID=0, Status='live', Symbol='Col27a1', Aliases='5730512J02Rik, AI449266', description='collagen, type XXVII, alpha 1', other_designations='procollagen, type XXVII, alpha 1', map_location='4 C1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=63214009, end_position_on_the_genomic_accession=63334991, orientation='plus', exon_count=64, OMIM='', no_hdr0=''),
  111727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111727, CurrentID=0, Status='live', Symbol='Iapt15', Aliases='', description='intracisternal A particle, tumor specific 15', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242800, CurrentID=0, Status='live', Symbol='Ttc34', Aliases='4932418L02, B230396O12Rik', description='tetratricopeptide repeat domain 34', other_designations='TPR repeat protein 34', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154833443, end_position_on_the_genomic_accession=154867127, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  111729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111729, CurrentID=0, Status='live', Symbol='Emv15', Aliases='Emv-15', description='endogenous ecotropic MuLV 15, A<y>-associated', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111731, CurrentID=0, Status='live', Symbol='Igk-V10', Aliases='', description='immunoglobulin kappa chain variable 10 (V10)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111732, CurrentID=0, Status='live', Symbol='Igk-V11', Aliases='', description='immunoglobulin kappa chain variable 11 (V11)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111733, CurrentID=0, Status='live', Symbol='Igk-V19', Aliases='', description='immunoglobulin kappa chain variable 19 (V19)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111734, CurrentID=0, Status='live', Symbol='Igk-V22', Aliases='', description='immunoglobulin kappa chain variable 22 (V22)', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111737, CurrentID=0, Status='live', Symbol='Mpmv11', Aliases='Mpmv-11', description='modified polytropic murine leukemia virus 11', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111738, CurrentID=0, Status='live', Symbol='Mpmv12', Aliases='Mpmv-12', description='modified polytropic murine leukemia virus 12', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242819, CurrentID=0, Status='live', Symbol='Rundc3b', Aliases='Gm440', description='RUN domain containing 3B', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=8490336, end_position_on_the_genomic_accession=8623027, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  111748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111748, CurrentID=0, Status='live', Symbol='Tic1', Aliases='Pim2, pim-2', description='transplantation-specific integration cluster 1', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111749, CurrentID=0, Status='live', Symbol='Igh-VGAM3.8', Aliases='', description='immunoglobulin heavy chain (VGAM3.8 family)', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111756, CurrentID=0, Status='live', Symbol='Iapls3-4', Aliases='', description='intracisternal A particle, lymphocyte specific 3-4', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242838, CurrentID=0, Status='live', Symbol='Lrrd1', Aliases='4932412H11Rik', description='leucine rich repeats and death domain containing 1', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=3845173, end_position_on_the_genomic_accession=3866596, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  111768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111768, CurrentID=0, Status='live', Symbol='Iapls3-9', Aliases='', description='intracisternal A particle, lymphocyte specific 3-9', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242851, CurrentID=0, Status='live', Symbol='Gnat3', Aliases='Ggust, Gtn', description='guanine nucleotide binding protein, alpha transducing 3', other_designations='alpha-gustducin|gustducin alpha-3 chain', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=17962570, end_position_on_the_genomic_accession=18019668, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  111785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111785, CurrentID=0, Status='live', Symbol='Hsd3b', Aliases='', description='hydroxy-delta-5-steroid dehydrogenase, 3 beta- and steroid delta-isomerase cluster', other_designations='hydroxysteroid dehydrogenase cluster, delta<5>-3-beta', map_location='', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242860, CurrentID=0, Status='live', Symbol='Rsbn1l', Aliases='8430412F05Rik, AI447441, C330002G24Rik', description='round spermatid basic protein 1-like', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=20893024, end_position_on_the_genomic_accession=20953441, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  111792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111792, CurrentID=0, Status='live', Symbol='Iapls3-10', Aliases='', description='intracisternal A particle, lymphocyte specific 3-10', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111801, CurrentID=0, Status='live', Symbol='Iapls3-41', Aliases='', description='intracisternal A particle, lymphocyte specific 3-41', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111813, CurrentID=0, Status='live', Symbol='Iapls1-19', Aliases='', description='intracisternal A particle, lymphocyte specific 1-19', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242891, CurrentID=0, Status='live', Symbol='Cct8l1', Aliases='Gm443', description='chaperonin containing TCP1, subunit 8 (theta)-like 1', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=25516067, end_position_on_the_genomic_accession=25518027, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  242894 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242894, CurrentID=0, Status='live', Symbol='Actr3b', Aliases='9630005C02, ARP11, AW047569, Arp3b, Arp3beta', description='ARP3 actin-related protein 3B', other_designations='ARP3 actin-related protein 3 homolog B|actin-related protein 3-beta', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=25759972, end_position_on_the_genomic_accession=25850688, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  111861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111861, CurrentID=0, Status='live', Symbol='H2-I', Aliases='H-2I', description='H2 I region, includes H2-Aa, Ab, Bl, Ea, Eb, Eb2, Ob, Pb', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111863, CurrentID=0, Status='live', Symbol='Pltr3', Aliases='', description='polytropic long terminal repeat containing proviral locus 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111864, CurrentID=0, Status='live', Symbol='Pltr6', Aliases='', description='polytropic long terminal repeat containing proviral locus 6', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  242939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242939, CurrentID=0, Status='live', Symbol='Cpz', Aliases='AV136066', description='carboxypeptidase Z', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=35502218, end_position_on_the_genomic_accession=35525626, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  242960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=242960, CurrentID=0, Status='live', Symbol='Fbxl5', Aliases='Fbl4, Fir4', description='F-box and leucine-rich repeat protein 5', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=43744618, end_position_on_the_genomic_accession=43782149, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  636177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=636177, CurrentID=0, Status='live', Symbol='Rhox4f', Aliases='Rhox4.6, Rhox4e, Rhox4h', description='reproductive homeobox 4F', other_designations='reproductive homeobox 4H', map_location='X A3.2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37602895, end_position_on_the_genomic_accession=37607686, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  111900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111900, CurrentID=0, Status='live', Symbol='Pltr4', Aliases='', description='polytropic long terminal repeat containing proviral locus 4', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111961, CurrentID=0, Status='live', Symbol='Csn', Aliases='', description='casein gene family, alpha, beta, gamma, delta/epsilon, kappa', other_designations='', map_location='', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111966, CurrentID=0, Status='live', Symbol='Mltr4', Aliases='', description='modified polytropic long terminal repeat containing proviral locus 4', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111970, CurrentID=0, Status='live', Symbol='Dlx1as', Aliases='Dix1as, Dlx1os', description='distal-less homeobox 1, antisense', other_designations='Dlx1 antisense|distal-less homeobox 1, opposite strand', map_location='2 42.58 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=71530638, end_position_on_the_genomic_accession=71537891, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  243043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243043, CurrentID=0, Status='live', Symbol='Kctd8', Aliases='A730087N02Rik', description='potassium channel tetramerisation domain containing 8', other_designations='', map_location='5 C3.1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=69109285, end_position_on_the_genomic_accession=69341709, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  111973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111973, CurrentID=0, Status='live', Symbol='Rnu3b', Aliases='', description='U3B small nuclear RNA complex', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111975, CurrentID=0, Status='live', Symbol='Igf2os', Aliases='Igf2as, Peg8', description='insulin-like growth factor 2, opposite strand', other_designations='insulin-like growth factor 2, antisense', map_location='7 87.99 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=142659693, end_position_on_the_genomic_accession=142670356, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  111977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111977, CurrentID=0, Status='live', Symbol='El1', Aliases='El, El-1', description='epilepsy 1', other_designations='', map_location='', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111978, CurrentID=0, Status='live', Symbol='El2', Aliases='El-2', description='epilepsy 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111984, CurrentID=0, Status='live', Symbol='Derm1', Aliases='', description='dermititis 1', other_designations='', map_location='9 33.34 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111987, CurrentID=0, Status='live', Symbol='Obq5', Aliases='', description='obesity QTL 5', other_designations='', map_location='9 23.53 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111992, CurrentID=0, Status='live', Symbol='Cq3', Aliases='Chol5', description='cholesterol QTL 3', other_designations='', map_location='3 44.3 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  111994 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=111994, CurrentID=0, Status='live', Symbol='Emo3', Aliases='', description='emotionality 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243078, CurrentID=0, Status='live', Symbol='Tecrl', Aliases='D330017N19Rik, Srd5a2l2', description='trans-2,3-enoyl-CoA reductase-like', other_designations='steroid 5 alpha-reductase 2-like 2|steroid 5-alpha-reductase 2-like 2 protein|synaptic glycoprotein SC2-like', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=83278122, end_position_on_the_genomic_accession=83355195, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  112009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112009, CurrentID=0, Status='live', Symbol='Lbw7', Aliases='', description='lupus NZB x NZW 7', other_designations='', map_location='1 76.73 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243083, CurrentID=0, Status='live', Symbol='Tmprss11f', Aliases='4631408C20, 4732406D01Rik', description='transmembrane protease, serine 11f', other_designations='airway trypsin-like protease 4', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86521226, end_position_on_the_genomic_accession=86632424, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  243084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243084, CurrentID=0, Status='live', Symbol='Tmprss11e', Aliases='A030012E10, AY787860, DESC1, TTSP', description='transmembrane protease, serine 11e', other_designations='serine protease DESC1|type II transmembrane serine protease', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=86705186, end_position_on_the_genomic_accession=86745807, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  243085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243085, CurrentID=0, Status='live', Symbol='Ugt2b35', Aliases='C730031G17', description='UDP glucuronosyltransferase 2 family, polypeptide B35', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87000860, end_position_on_the_genomic_accession=87013274, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  112052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112052, CurrentID=0, Status='live', Symbol='Mobq1', Aliases='Mob1', description='multigenic obesity QTL 1', other_designations='', map_location='7 66.7 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112056, CurrentID=0, Status='live', Symbol='Bomb1', Aliases='', description='bone marrow pre-B 1', other_designations='', map_location='3 56.89 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112077, CurrentID=0, Status='live', Symbol='Mop3', Aliases='', description='morphine preference 3', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112080, CurrentID=0, Status='live', Symbol='Alcw1', Aliases='', description='alcohol withdrawal 1', other_designations='', map_location='1 80.33 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112081, CurrentID=0, Status='live', Symbol='Alcw2', Aliases='', description='alcohol withdrawal 2', other_designations='', map_location='', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112087, CurrentID=0, Status='live', Symbol='Mop2', Aliases='', description='morphine preference 2', other_designations='', map_location='10 3.04 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112088, CurrentID=0, Status='live', Symbol='Alcw3', Aliases='', description='alcohol withdrawal 3', other_designations='', map_location='11 25.18 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243168, CurrentID=0, Status='live', Symbol='Hsd17b13', Aliases='AI047820, PAN1B-like, Pan1b', description='hydroxysteroid (17-beta) dehydrogenase 13', other_designations='17-beta hydroxysteroid dehydrogenase 13|17-beta-HSD 13|short-chain dehydrogenase/reductase 9', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=103955442, end_position_on_the_genomic_accession=103977388, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  112100 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112100, CurrentID=0, Status='live', Symbol='Bmd1', Aliases='', description='bone mineral density 1', other_designations='', map_location='1 67.71 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112101, CurrentID=0, Status='live', Symbol='Etohc1', Aliases='', description='ethanol consumption 1', other_designations='', map_location='2 24.26 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112102, CurrentID=0, Status='live', Symbol='Etohc2', Aliases='', description='ethanol consumption 2', other_designations='', map_location='2 49.68 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112114, CurrentID=0, Status='live', Symbol='Tbs2', Aliases='', description='tuberculosis severity 2', other_designations='', map_location='9 14.79 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112123, CurrentID=0, Status='live', Symbol='Szs1', Aliases='Kasz1', description='seizure susceptibility 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112125, CurrentID=0, Status='live', Symbol='Aliq1', Aliases='Ali1', description='acute lung injury QTL 1', other_designations='', map_location='11 54.6 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112133, CurrentID=0, Status='live', Symbol='Aaq1', Aliases='', description='alcohol acceptance QTL 1', other_designations='', map_location='15 16.82 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  193452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193452, CurrentID=0, Status='live', Symbol='Zfp184', Aliases='4930500C15Rik', description='zinc finger protein 184 (Kruppel-like)', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21944940, end_position_on_the_genomic_accession=21960495, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  243219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243219, CurrentID=0, Status='live', Symbol='2900026A02Rik', Aliases='8430408O14Rik, A530094D01, AW539426, Gm449, Kiaa1671, mKIAA1671', description='RIKEN cDNA 2900026A02 gene', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=113086323, end_position_on_the_genomic_accession=113221287, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  112154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112154, CurrentID=0, Status='live', Symbol='Skts1', Aliases='', description='skin tumor susceptibility 1', other_designations='', map_location='7 33.37 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112163, CurrentID=0, Status='live', Symbol='Scc2', Aliases='', description='colon tumor susceptibility 2', other_designations='', map_location='2 23.71 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243262 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243262, CurrentID=0, Status='live', Symbol='Oas1f', Aliases='', description="2'-5' oligoadenylate synthetase 1F", other_designations='', map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120847367, end_position_on_the_genomic_accession=120857986, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  243270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243270, CurrentID=0, Status='live', Symbol='Hcar1', Aliases='Gpr81', description='hydrocarboxylic acid receptor 1', other_designations='G protein-coupled receptor 81|G-protein coupled receptor 81', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123876736, end_position_on_the_genomic_accession=123880020, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  243272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243272, CurrentID=0, Status='live', Symbol='Sbno1', Aliases='9330180L10Rik, AI849128, AW557836, BC021875, Sbno, mSno1, sno', description='sno, strawberry notch homolog 1 (Drosophila)', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=124368702, end_position_on_the_genomic_accession=124426011, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  112201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112201, CurrentID=0, Status='live', Symbol='Skts5', Aliases='', description='skin tumor susceptibility 5', other_designations='', map_location='12 18.06 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243274, CurrentID=0, Status='live', Symbol='Tmem132d', Aliases='C630028F04Rik', description='transmembrane protein 132D', other_designations='mature OL transmembrane protein|mature oligodendrocytes transmembrane protein', map_location='5 G1.2-G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=127781629, end_position_on_the_genomic_accession=128433077, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  243277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243277, CurrentID=0, Status='live', Symbol='Adgrd1', Aliases='E230012M21Rik, Gpr133', description='adhesion G protein-coupled receptor D1', other_designations='G protein-coupled receptor 133|G-protein coupled receptor 133|G-protein coupled receptor PGR25|probable G-protein coupled receptor 133', map_location='5 G1.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=129096666, end_position_on_the_genomic_accession=129204599, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  112208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112208, CurrentID=0, Status='live', Symbol='Ity3', Aliases='', description='immunity to S. typhimurium 3', other_designations='', map_location='1 56.92 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112210, CurrentID=0, Status='live', Symbol='Tlsm1', Aliases='', description='thymic lymphoma susceptible 1', other_designations='', map_location='7 81.27 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243300, CurrentID=0, Status='live', Symbol='Nyap1', Aliases='6430598A04Rik', description='neuronal tyrosine-phosphorylated phosphoinositide 3-kinase adaptor 1', other_designations='neuronal tyrosine phosphorylated adaptor for phosphoinositide 3-kinase adaptor 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137730882, end_position_on_the_genomic_accession=137741614, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  112229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112229, CurrentID=0, Status='live', Symbol='Psl1', Aliases='Tpa', description='promotion susceptibility QTL 1', other_designations='', map_location='9 57.46 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112235, CurrentID=0, Status='live', Symbol='Par3', Aliases='', description='pulmonary adenoma resistance 3', other_designations='', map_location='12 37.16 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243308, CurrentID=0, Status='live', Symbol='A430033K04Rik', Aliases='C86988', description='RIKEN cDNA A430033K04 gene', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138622859, end_position_on_the_genomic_accession=138648917, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  112239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112239, CurrentID=0, Status='live', Symbol='Rapop5', Aliases='', description='radiation-induced apoptosis 5', other_designations='', map_location='15 19.26 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243312 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243312, CurrentID=0, Status='live', Symbol='Elfn1', Aliases='A930017N06Rik, AI197272, Ppp1r28', description='leucine rich repeat and fibronectin type III, extracellular 1', other_designations='extracellular leucine-rich repeat and fibronectin type-III domain-containing protein 1|protein phosphatase 1 regulatory subunit 28', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=139907640, end_position_on_the_genomic_accession=139974725, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  112242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112242, CurrentID=0, Status='live', Symbol='Bhr4', Aliases='', description='bronchial hyperresponsiveness 4', other_designations='', map_location='', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112244, CurrentID=0, Status='live', Symbol='Aod2', Aliases='', description='autoimmune ovarian dysgenesis 2', other_designations='', map_location='3 18.3 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112252, CurrentID=0, Status='live', Symbol='Cia2', Aliases='', description='collagen induced arthritis QTL 2', other_designations='', map_location='2 34.9 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112256, CurrentID=0, Status='live', Symbol='Sst1', Aliases='', description='susceptibility to tuberculosis 1', other_designations='', map_location='1 44.07 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112257, CurrentID=0, Status='live', Symbol='Pbwg1', Aliases='C10bw1', description='postnatal body weight growth 1', other_designations='', map_location='2 33.33 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112265, CurrentID=0, Status='live', Symbol='Cia8', Aliases='', description='collagen induced arthritis QTL 8', other_designations='', map_location='10 38.48 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243339, CurrentID=0, Status='live', Symbol='Tmem130', Aliases='BC067004, C130036G08', description='transmembrane protein 130', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144735915, end_position_on_the_genomic_accession=144761578, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  112272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112272, CurrentID=0, Status='live', Symbol='Pbwg12', Aliases='C10bw7', description='postnatal body weight growth 12', other_designations='', map_location='12 35.51 cM', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112274, CurrentID=0, Status='live', Symbol='Mvwf', Aliases='', description='modifier of von Willebrand factor', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112278, CurrentID=0, Status='live', Symbol='Pas1', Aliases='Pas-1', description='pulmonary adenoma susceptibility 1', other_designations='', map_location='6 78.22 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112284, CurrentID=0, Status='live', Symbol='Pas3', Aliases='Pas-3', description='pulmonary adenoma susceptibility 3', other_designations='', map_location='19 13.82 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112287, CurrentID=0, Status='live', Symbol='Exq2', Aliases='', description='exploratory and excitability QTL 2', other_designations='', map_location='1 58.27 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112289, CurrentID=0, Status='live', Symbol='Exq1', Aliases='', description='exploratory and excitability QTL 1', other_designations='', map_location='10 67.89 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243362, CurrentID=0, Status='live', Symbol='Stard13', Aliases='DLC2, GT650', description='StAR-related lipid transfer (START) domain containing 13', other_designations='START domain-containing protein 13|serologically defined colon cancer antigen 13|serologically defined colon cancer antigen 28', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=151037515, end_position_on_the_genomic_accession=151428256, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  112295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112295, CurrentID=0, Status='live', Symbol='Lmb1', Aliases='', description='lupus in MRL and B6 F2 cross, QTL 1', other_designations='', map_location='4 57.76 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243369, CurrentID=0, Status='live', Symbol='Sspo', Aliases='C79529, Scospondin', description='SCO-spondin', other_designations='subcommissural organ spondin', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48446243, end_position_on_the_genomic_accession=48501262, orientation='plus', exon_count=104, OMIM='', no_hdr0=''),
  243371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243371, CurrentID=0, Status='live', Symbol='Lrrc61', Aliases='BC027309', description='leucine rich repeat containing 61', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48554779, end_position_on_the_genomic_accession=48570722, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  243372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243372, CurrentID=0, Status='live', Symbol='Zfp775', Aliases='C130032F08Rik, Znf775', description='zinc finger protein 775', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48602001, end_position_on_the_genomic_accession=48623227, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  243374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243374, CurrentID=0, Status='live', Symbol='Gimap8', Aliases='Gm457, IAN9, IMAP8', description='GTPase, IMAP family member 8', other_designations='immune-associated nucleotide-binding protein 9|immunity-associated protein 8', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48647186, end_position_on_the_genomic_accession=48660875, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  243376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243376, CurrentID=0, Status='live', Symbol='Doxl2', Aliases='', description='diamine oxidase-like protein 2', other_designations='', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48973816, end_position_on_the_genomic_accession=48978745, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  243377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243377, CurrentID=0, Status='live', Symbol='Svs1', Aliases='9530003A11', description='seminal vesicle secretory protein 1', other_designations='seminal vesicle secretion 1', map_location='6 B2.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=48986861, end_position_on_the_genomic_accession=48991725, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  112308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112308, CurrentID=0, Status='live', Symbol='Hcs1', Aliases='Hcs', description='hepatocarcinogenesis susceptibility 1', other_designations='', map_location='7 43.46 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243382 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243382, CurrentID=0, Status='live', Symbol='Ppm1k', Aliases='2900063A19Rik, A930026L03Rik, PP2Cm', description='protein phosphatase 1K (PP2C domain containing)', other_designations='PP2C-kappa|protein phosphatase 2C isoform kappa', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57506502, end_position_on_the_genomic_accession=57535516, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  243385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243385, CurrentID=0, Status='live', Symbol='Gprin3', Aliases='C030038J10Rik, C730021L23', description='GPRIN family member 3', other_designations='GRIN3', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=59344503, end_position_on_the_genomic_accession=59426290, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  112314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112314, CurrentID=0, Status='live', Symbol='Char2', Aliases='', description='P. chabaudi malaria resistance QTL 2', other_designations='', map_location='8 48.27 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112326, CurrentID=0, Status='live', Symbol='Wg2', Aliases='', description='weight gain in high growth mice 2', other_designations='Q2Ucd2-wg29', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112332, CurrentID=0, Status='live', Symbol='Athsq1', Aliases='', description='atherosclerosis susceptibility QTL 1', other_designations='', map_location='4 70.02 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112335, CurrentID=0, Status='live', Symbol='Lore2', Aliases='', description='loss of righting induced by ethanol 2', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112340, CurrentID=0, Status='live', Symbol='Lore4', Aliases='', description='loss of righting induced by ethanol 4', other_designations='', map_location='', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112341, CurrentID=0, Status='live', Symbol='Lore5', Aliases='', description='loss of righting induced by ethanol 5', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243420, CurrentID=0, Status='live', Symbol='Igkv1-135', Aliases='Gm4964, bd2', description='immunoglobulin kappa variable 1-135', other_designations='Ig kappa chain|IgG1/kappa antibody|immunoglobulin kappa chain variable 1-135|immunoglobulin light chain variable region|predictd gene, 243420', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67609745, end_position_on_the_genomic_accession=67610508, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  112349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112349, CurrentID=0, Status='live', Symbol='Gasa2', Aliases='', description='gastritis type A susceptibility locus 2', other_designations='', map_location='4 80.52 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112350, CurrentID=0, Status='live', Symbol='Gasa1', Aliases='', description='gastritis type A susceptibility locus 1', other_designations='', map_location='4 69.48 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112354, CurrentID=0, Status='live', Symbol='Alcp1', Aliases='', description='alcohol preference locus 1, male specific', other_designations='', map_location='2 27.37 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243428, CurrentID=0, Status='live', Symbol='Igkv11-125', Aliases='ENSMUSG00000050218, ENSMUSG00000076510, Gm10877, if11', description='immunoglobulin kappa variable 11-125', other_designations='immunoglobulin kappa chain variable 11-125|immunoglobulin light chain variable region', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67913573, end_position_on_the_genomic_accession=67914052, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  112358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112358, CurrentID=0, Status='live', Symbol='Nidd5', Aliases='', description='non-insulin-dependent diabetes mellitus 5', other_designations='', map_location='2 31.66 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243431, CurrentID=0, Status='live', Symbol='Igkv9-124', Aliases='EG243431, Gm4966', description='immunoglobulin kappa chain variable 9-124', other_designations='Immnogloblin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67942076, end_position_on_the_genomic_accession=67942540, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  112365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112365, CurrentID=0, Status='live', Symbol='Sle3', Aliases='', description='systemic lupus erythmatosus susceptibility 3', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112366, CurrentID=0, Status='live', Symbol='Sle2', Aliases='', description='systemic lupus erythmatosus susceptibility 2', other_designations='', map_location='4 39.4 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112367, CurrentID=0, Status='live', Symbol='Sle1', Aliases='', description='systemic lupus erythmatosus susceptibility 1', other_designations='', map_location='1 76.73 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112375, CurrentID=0, Status='live', Symbol='Sle1a', Aliases='', description='systematic lupus erythematosus susceptibility 1a', other_designations='', map_location='1 74.86 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112376, CurrentID=0, Status='live', Symbol='Sle1b', Aliases='', description='systematic lupus erythematosus susceptibility 1b', other_designations='', map_location='1 75.1 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112377, CurrentID=0, Status='live', Symbol='Sle1c', Aliases='', description='systematic lupus erythematosus susceptibility 1c', other_designations='', map_location='1 94.73 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243451, CurrentID=0, Status='live', Symbol='Igkv4-86', Aliases='Gm459, kb4', description='immunoglobulin kappa variable 4-86', other_designations='Ig kappa chain V-VI region XRPC|immunoglobulin kappa chain variable 4-86|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68910411, end_position_on_the_genomic_accession=68910938, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  112380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112380, CurrentID=0, Status='live', Symbol='Igfbp3q2', Aliases='', description='insulin-like growth factor binding protein 3 QTL 2', other_designations='', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112381, CurrentID=0, Status='live', Symbol='Eae3', Aliases='', description='experimental allergic encephalomyelitis susceptibility 3', other_designations='', map_location='3 41.93 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112383, CurrentID=0, Status='live', Symbol='Eae2', Aliases='', description='experimental allergic encephalomyelitis susceptibility 2', other_designations='', map_location='', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  112388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112388, CurrentID=0, Status='live', Symbol='Eae7', Aliases='', description='experimental allergic encephalomyelitis susceptibility 7', other_designations='', map_location='11 51.88 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243461, CurrentID=0, Status='live', Symbol='Igkv7-33', Aliases='Igk-V22, U29423, vk22/Jk5', description='immunoglobulin kappa chain variable 7-33', other_designations='anti-PC Ig kappa chain|anti-PC rearranged Ig kappa chain V-J region mRNA', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70058632, end_position_on_the_genomic_accession=70059199, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  112394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112394, CurrentID=0, Status='live', Symbol='Eae11', Aliases='', description='experimental allergic encephalomyelitis susceptibility 11', other_designations='experimental allergic encephalomyelitis susceptibility11', map_location='16 40.3 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  243469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243469, CurrentID=0, Status='live', Symbol='Igk', Aliases='kappa', description='immunoglobulin kappa chain complex', other_designations='kappa light chain variable region', map_location='6 30.89 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=67555636, end_position_on_the_genomic_accession=70726754, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  112403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112403, CurrentID=0, Status='live', Symbol='Dxo', Aliases='Dom3z, NG6', description='decapping exoribonuclease', other_designations='protein Dom3Z', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34837019, end_position_on_the_genomic_accession=34839186, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  112405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112405, CurrentID=0, Status='live', Symbol='Egln1', Aliases='AI503754, C1orf12, HIF-PH2, HPH-2, Hif-p4h-2, ORF13, Phd2, SM-20', description='egl-9 family hypoxia-inducible factor 1', other_designations='HIF-prolyl hydroxylase 2|hypoxia-inducible factor prolyl hydroxylase 2|prolyl hydroxylase domain-containing protein 2', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124908586, end_position_on_the_genomic_accession=124949319, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  112406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112406, CurrentID=0, Status='live', Symbol='Egln2', Aliases='0610011A13Rik, C85656, Hif-p4h-1, Ier4, Phd1, SM-20', description='egl-9 family hypoxia-inducible factor 2', other_designations='HIF-PH1|HIF-prolyl hydroxylase 1|HPH-1|falkor|hypoxia-inducible factor prolyl hydroxylase 1|immediate early response 4|prolyl hydroxylase domain-containing protein 1', map_location='7 15.83 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27158658, end_position_on_the_genomic_accession=27166802, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  112407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112407, CurrentID=0, Status='live', Symbol='Egln3', Aliases='2610021G09Rik, AI505553, AI648162, Hif-p4h-3, Phd3, SM-20', description='egl-9 family hypoxia-inducible factor 3', other_designations='HIF-PH3|HIF-prolyl hydroxylase 3|HPH-3|hypoxia-inducible factor prolyl hydroxylase 3|prolyl hydroxylase domain-containing protein 3', map_location='12 C1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=54178981, end_position_on_the_genomic_accession=54203874, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  112408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112408, CurrentID=0, Status='live', Symbol='Tas2r116', Aliases='T2R16, TRB1, TRB4, Tas2r14, Tas2r16, Tas2r7, mGR16, mt2r56', description='taste receptor, type 2, member 116', other_designations='T2R116|candidate taste receptor mt2r56|tas2r116|taste receptor, type 2, member 14|taste receptor, type 2, member 7', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132855438, end_position_on_the_genomic_accession=132856355, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  112415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112415, CurrentID=0, Status='live', Symbol='C030039L03Rik', Aliases='', description='RIKEN cDNA C030039L03 gene', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=27674661, end_position_on_the_genomic_accession=27706484, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  112417 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112417, CurrentID=0, Status='live', Symbol='Ugt2b37', Aliases='0610033E06Rik', description='UDP glucuronosyltransferase 2 family, polypeptide B37', other_designations='', map_location='5 E1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=87240492, end_position_on_the_genomic_accession=87254788, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  112418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112418, CurrentID=0, Status='live', Symbol='1700102P08Rik', Aliases='AI593206', description='RIKEN cDNA 1700102P08 gene', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108392834, end_position_on_the_genomic_accession=108410966, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  112419 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112419, CurrentID=0, Status='live', Symbol='Ifit1bl2', Aliases='2010002M12Rik, AW412491, Ifit1b', description='interferon induced protein with tetratricopeptide repeats 1B like 2', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34617051, end_position_on_the_genomic_accession=34640743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  112422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=112422, CurrentID=0, Status='live', Symbol='2610305D13Rik', Aliases='Ssm1b', description='RIKEN cDNA 2610305D13 gene', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=147611936, end_position_on_the_genomic_accession=147642508, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  243499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243499, CurrentID=0, Status='live', Symbol='Lrrtm4', Aliases='7530419J18Rik, A230052N11', description='leucine rich repeat transmembrane neuronal 4', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=80018859, end_position_on_the_genomic_accession=80810162, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  243510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243510, CurrentID=0, Status='live', Symbol='Ccdc142', Aliases='A230058J24Rik', description='coiled-coil domain containing 142', other_designations='', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83101516, end_position_on_the_genomic_accession=83109121, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  636741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=636741, CurrentID=0, Status='live', Symbol='Zfp964', Aliases='4732490P12, EG636741, Gm7187', description='zinc finger protein 964', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=69654463, end_position_on_the_genomic_accession=69664656, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  243529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243529, CurrentID=0, Status='live', Symbol='H1fx', Aliases='Gm461, H1.10, H1X', description='H1 histone family, member X', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=87980421, end_position_on_the_genomic_accession=87981482, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  243535 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243535, CurrentID=0, Status='live', Symbol='BC048671', Aliases='', description='cDNA sequence BC048671', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90301270, end_position_on_the_genomic_accession=90305449, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  243537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243537, CurrentID=0, Status='live', Symbol='Uroc1', Aliases='AI265605, BC022133', description='urocanase domain containing 1', other_designations='imidazolonepropionate hydrolase', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90333239, end_position_on_the_genomic_accession=90364551, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  243538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243538, CurrentID=0, Status='live', Symbol='Cfap100', Aliases='AW048498, C030041G11Rik, C230069K22Rik, Ccdc37', description='cilia and flagella associated protein 100', other_designations='coiled-coil domain containing 37', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90403731, end_position_on_the_genomic_accession=90430246, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  243547 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243547, CurrentID=0, Status='live', Symbol='Grip2', Aliases='', description='glutamate receptor interacting protein 2', other_designations='', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91761508, end_position_on_the_genomic_accession=91843954, orientation='minus', exon_count=28, OMIM='', no_hdr0=''),
  243548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243548, CurrentID=0, Status='live', Symbol='Prickle2', Aliases='6230400G14Rik, 6720451F06Rik, mpk2', description='prickle homolog 2', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=92370891, end_position_on_the_genomic_accession=92706335, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  243574 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243574, CurrentID=0, Status='live', Symbol='Kbtbd8', Aliases='AU040152, SSEC-51, SSEC51, Takrp, mKIAA1842', description='kelch repeat and BTB (POZ) domain containing 8', other_designations='SSEC 51|T-cell activation kelch repeat protein', map_location='6 D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=95117880, end_position_on_the_genomic_accession=95129793, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  636808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=636808, CurrentID=0, Status='live', Symbol='Cntnap5a', Aliases='Caspr5-1, EG636808', description='contactin associated protein-like 5A', other_designations='cell recognition molecule Caspr5-1|cell recognition molecule Caspr5a|contactin associated protein-like 5-1|contactin-associated protein-like 5a', map_location='1 E2.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=115684497, end_position_on_the_genomic_accession=116587323, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  243612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243612, CurrentID=0, Status='live', Symbol='Ssu2', Aliases='D630042P16Rik, Fls485, Ssuh2', description='ssu-2 homolog (C. elegans)', other_designations='protein ssu-2 homolog', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=112359322, end_position_on_the_genomic_accession=112388199, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  243616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243616, CurrentID=0, Status='live', Symbol='Slc6a11', Aliases='D930045G19Rik, E130202I16Rik, GAT4, Gabt4, Gat3', description='solute carrier family 6 (neurotransmitter transporter, GABA), member 11', other_designations='GAT-3|GAT-4|gamma-aminobutyric acid (GABA-A) transporter 4|sodium- and chloride-dependent GABA transporter 4|solute carrier family 6 member 11', map_location='6 52.98 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=114131241, end_position_on_the_genomic_accession=114249886, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  243621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243621, CurrentID=0, Status='live', Symbol='Iqsec3', Aliases='AK122446, BRAG3, mKIAA1110, synarfGEF', description='IQ motif and Sec7 domain 3', other_designations='', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=121372927, end_position_on_the_genomic_accession=121473703, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  243634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243634, CurrentID=0, Status='live', Symbol='Ano2', Aliases='BC033409, Tmem16b', description='anoctamin 2', other_designations='ano2|anoctamin 2 isoform Adelta4|anoctamin 2 isoform Bdelta4', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125690387, end_position_on_the_genomic_accession=126040133, orientation='plus', exon_count=30, OMIM='', no_hdr0=''),
  243653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243653, CurrentID=0, Status='live', Symbol='Clec1a', Aliases='5930406N14Rik', description='C-type lectin domain family 1, member a', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129423358, end_position_on_the_genomic_accession=129452000, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  243655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243655, CurrentID=0, Status='live', Symbol='Klre1', Aliases='Klre-1, NKG2I', description='killer cell lectin-like receptor family E member 1', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129559032, end_position_on_the_genomic_accession=129585827, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  243659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243659, CurrentID=0, Status='live', Symbol='Styk1', Aliases='9130025L13, AI326477, Nok', description='serine/threonine/tyrosine kinase 1', other_designations='NOK kinase|mNOK|novel oncogene with kinase domain', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131299133, end_position_on_the_genomic_accession=131317009, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  636931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=636931, CurrentID=0, Status='live', Symbol='Trim71', Aliases='2610206G21Rik, AL022943, Gm1127, Lin41, lin-41, mLin41, mlin-41', description='tripartite motif-containing 71', other_designations='lin-41 homolog|protein lin-41 homolog|tripartite motif-containing protein 71', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114511268, end_position_on_the_genomic_accession=114564369, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  243725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243725, CurrentID=0, Status='live', Symbol='Ppp1r9a', Aliases='2810430P21Rik, 4930518N04Rik, 5330407E15, A230094E16Rik, BB181831, NRB, mKIAA1222, neurabin-I', description='protein phosphatase 1, regulatory (inhibitor) subunit 9A', other_designations='Neurabin I|neural tissue-specific F-actin binding protein', map_location='6 1.86 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=4902747, end_position_on_the_genomic_accession=5165661, orientation='plus', exon_count=25, OMIM='', no_hdr0=''),
  243743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243743, CurrentID=0, Status='live', Symbol='Plxna4', Aliases='9330117B14, Plxa4, mKIAA1550', description='plexin A4', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=32144543, end_position_on_the_genomic_accession=32588192, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  243753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243753, CurrentID=0, Status='live', Symbol='2010107G12Rik', Aliases='Gm468', description='RIKEN cDNA 2010107G12 gene', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=34945294, end_position_on_the_genomic_accession=34988493, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  243755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243755, CurrentID=0, Status='live', Symbol='Slc13a4', Aliases='9630060C05Rik, SUT-1, SUT1', description='solute carrier family 13 (sodium/sulfate symporters), member 4', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=35267953, end_position_on_the_genomic_accession=35308126, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  243764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243764, CurrentID=0, Status='live', Symbol='Chrm2', Aliases='AChR-M2, Chrm-2, M2', description='cholinergic receptor, muscarinic 2, cardiac', other_designations='AChR M2|muscarinic acetylcholine receptor 2', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=36387950, end_position_on_the_genomic_accession=36528638, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  243771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243771, CurrentID=0, Status='live', Symbol='Parp12', Aliases='9930021O16, AA409132, AA536654, ARTD12, PARP-12, Zc3hdc1', description='poly (ADP-ribose) polymerase family, member 12', other_designations='ADP-ribosyltransferase diphtheria toxin-like 12|poly ADP-ribose polymerase 12|zinc finger CCCH domain-containing protein 1|zinc finger CCCH type domain containing 1', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39086412, end_position_on_the_genomic_accession=39118349, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  243780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243780, CurrentID=0, Status='live', Symbol='E330009J07Rik', Aliases='AI841796, Lchn, mKIAA1147', description='RIKEN cDNA E330009J07 gene', other_designations='', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40401375, end_position_on_the_genomic_accession=40436491, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  637021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=637021, CurrentID=0, Status='live', Symbol='Vmn2r124', Aliases='EG637021, Gm7196, Vmn2r-ps113', description='vomeronasal 2, receptor 124', other_designations='vomeronasal 2, receptor 93 pseudogene|vomeronasal 2, receptor, pseudogene 113', map_location='17 A3.2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=18049484, end_position_on_the_genomic_accession=18074220, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  243813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243813, CurrentID=0, Status='live', Symbol='Leng9', Aliases='9530024C23Rik, F630035L11Rik', description='leukocyte receptor cluster (LRC) member 9', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4148183, end_position_on_the_genomic_accession=4149872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  243816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243816, CurrentID=0, Status='live', Symbol='Gp6', Aliases='9830166G18Rik, Gm469, Gpvi', description='glycoprotein 6 (platelet)', other_designations='glycoprotein 5', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4362853, end_position_on_the_genomic_accession=4397905, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  243819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243819, CurrentID=0, Status='live', Symbol='Ppp6r1', Aliases='2010309P17Rik, AI836219, B430201G11Rik, D030074N20Rik, Pp6r1, Saps1, mKIAA1115', description='protein phosphatase 6, regulatory subunit 1', other_designations='SAPS domain family, member 1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4631495, end_position_on_the_genomic_accession=4658950, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  243822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243822, CurrentID=0, Status='live', Symbol='Fam71e2', Aliases='4930401F20Rik', description='family with sequence similarity 71, member E2', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4753226, end_position_on_the_genomic_accession=4771361, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  243833 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243833, CurrentID=0, Status='live', Symbol='Zfp128', Aliases='9630016P15Rik, Znf8, mZnf8', description='zinc finger protein 128', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12881178, end_position_on_the_genomic_accession=12893422, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  243834 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243834, CurrentID=0, Status='live', Symbol='Zfp324', Aliases='A930002M05, D430030K24Rik, ZF5128', description='zinc finger protein 324', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12965811, end_position_on_the_genomic_accession=12974244, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  243842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243842, CurrentID=0, Status='live', Symbol='Gltscr1', Aliases='', description='glioma tumor suppressor candidate region gene 1', other_designations='', map_location='7 8.67 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15970671, end_position_on_the_genomic_accession=16048082, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  243846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243846, CurrentID=0, Status='live', Symbol='Ccdc9', Aliases='2600011L02Rik', description='coiled-coil domain containing 9', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16273534, end_position_on_the_genomic_accession=16286815, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  243853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243853, CurrentID=0, Status='live', Symbol='Fkrp', Aliases='A830029B19Rik, AI842067, AI847300, LGMD1I, MDC1C', description='fukutin related protein', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=16809267, end_position_on_the_genomic_accession=16820668, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  243862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243862, CurrentID=0, Status='live', Symbol='Psg22', Aliases='AA408604, cea9', description='pregnancy-specific glycoprotein 22', other_designations='carcinoembryonic antigen-related cell adhesion molecule 1 pseudogene', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18718090, end_position_on_the_genomic_accession=18727248, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  637079 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=637079, CurrentID=0, Status='live', Symbol='Gm16486', Aliases='ENSMUSG00000035586', description='predicted gene 16486', other_designations='', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70702025, end_position_on_the_genomic_accession=70717650, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  243864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243864, CurrentID=0, Status='live', Symbol='Mill2', Aliases='Micb', description='MHC I like leukocyte 2', other_designations='MHC I-like leukocyte 2|MHC class I-like located near the LRC, 2|VMHC class I polypeptide-related sequence B', map_location='7 9.35 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18839955, end_position_on_the_genomic_accession=18863609, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  243866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243866, CurrentID=0, Status='live', Symbol='Gm4969', Aliases='9330180I18, EG243866', description='predicted gene 4969', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19098299, end_position_on_the_genomic_accession=19118531, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  243867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243867, CurrentID=0, Status='live', Symbol='Fbxo46', Aliases='20D7-FC4, 4932704E22Rik, Fbxo34l', description='F-box protein 46', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19119859, end_position_on_the_genomic_accession=19138261, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  243874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243874, CurrentID=0, Status='live', Symbol='Nlrp9b', Aliases='Nalp-delta, Nalp9b, Nlrp9', description='NLR family, pyrin domain containing 9B', other_designations='NACHT, LRR and PYD containing protein 9b|NLR family, pyrin domain containing 9', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19991393, end_position_on_the_genomic_accession=20062942, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  243880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243880, CurrentID=0, Status='live', Symbol='Nlrp4a', Aliases='E330021B02, E330028A19Rik, NALP4, Nalp-eta, Nalp-ita, Nalp4a', description='NLR family, pyrin domain containing 4A', other_designations='NACHT, leucine rich repeat and PYD containing 4A|NACHT/LRR/pyrin domain-containing protein 4', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26435113, end_position_on_the_genomic_accession=26476142, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  243881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243881, CurrentID=0, Status='live', Symbol='Cyp2b23', Aliases='EG243881', description='cytochrome P450, family 2, subfamily b, polypeptide 23', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=26665227, end_position_on_the_genomic_accession=26690372, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  243897 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243897, CurrentID=0, Status='live', Symbol='Ggn', Aliases='AI593290', description='gametogenetin', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29170210, end_position_on_the_genomic_accession=29173939, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  243905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243905, CurrentID=0, Status='live', Symbol='Zfp568', Aliases='C80731, Gm1691, chato', description='zinc finger protein 568', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=29983955, end_position_on_the_genomic_accession=30028282, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  243906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243906, CurrentID=0, Status='live', Symbol='Zfp14', Aliases='4732429I09Rik, Krox-9, Zfp-14, mKIAA1559', description='zinc finger protein 14', other_designations='zinc finger protein Krox-9', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30036357, end_position_on_the_genomic_accession=30051935, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  243910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243910, CurrentID=0, Status='live', Symbol='Nfkbid', Aliases='IkappaBNS', description='nuclear factor of kappa light polypeptide gene enhancer in B cells inhibitor, delta', other_designations='I-kappa-B-delta|NF-kappa B inhibitor|ikB-delta|ikappaBdelta|nuclear factor of kappa light polypeptide gene enhancer in B-cells inhibitor, delta', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30421729, end_position_on_the_genomic_accession=30428746, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  243911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243911, CurrentID=0, Status='live', Symbol='Kirrel2', Aliases='C330019F22Rik, NEPH3, NLG1', description='kin of IRRE like 2 (Drosophila)', other_designations='X kin of IRRE like 2|kin of irregular chiasm-like 2|kin of irregular chiasm-like protein 2', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30447530, end_position_on_the_genomic_accession=30460680, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  243912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243912, CurrentID=0, Status='live', Symbol='Hspb6', Aliases='AA387366, Gm479, Hsp20', description='heat shock protein, alpha-crystallin-related, B6', other_designations='cardiac heat shock protein p20|cardiac p20', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30552634, end_position_on_the_genomic_accession=30555439, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  243914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243914, CurrentID=0, Status='live', Symbol='Lgi4', Aliases='Lgil3, clp', description='leucine-rich repeat LGI family, member 4', other_designations='LGI1-like protein 3|leucine-rich glioma inactivated protein 4|leucine-rich glioma-inactivated protein 4', map_location='7 19.29 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31058953, end_position_on_the_genomic_accession=31070935, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  243923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243923, CurrentID=0, Status='live', Symbol='Rgs9bp', Aliases='R9AP, Rgs9-1bp', description='regulator of G-protein signalling 9 binding protein', other_designations='RGS9 anchor protein|RGS9-1 anchor|RGS9-anchoring protein', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35578993, end_position_on_the_genomic_accession=35585582, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  243931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243931, CurrentID=0, Status='live', Symbol='Tshz3', Aliases='A630038G13Rik, Tsh3, Zfp537, mKIAA1474, teashirt3', description='teashirt zinc finger family member 3', other_designations='teashirt 3|zinc finger protein 537', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=36697742, end_position_on_the_genomic_accession=36773553, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  243937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243937, CurrentID=0, Status='live', Symbol='Zfp536', Aliases='9630010P11Rik, Znf536, mKIAA0390', description='zinc finger protein 536', other_designations='', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=37318023, end_position_on_the_genomic_accession=37774177, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  243944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243944, CurrentID=0, Status='live', Symbol='4930433I11Rik', Aliases='Gm481', description='RIKEN cDNA 4930433I11 gene', other_designations='', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=40987609, end_position_on_the_genomic_accession=40994833, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  243958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243958, CurrentID=0, Status='live', Symbol='Siglecg', Aliases='9830164H23, A630096C01Rik, Siglec-G, Siglec10, mSiglec-G', description='sialic acid binding Ig-like lectin G', other_designations='sialic acid binding Ig-like lectin 10', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43408193, end_position_on_the_genomic_accession=43418350, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  243961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243961, CurrentID=0, Status='live', Symbol='Shank1', Aliases='', description='SH3/ankyrin domain gene 1', other_designations='SH3 and multiple ankyrin repeat domains 1', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44307848, end_position_on_the_genomic_accession=44358353, orientation='plus', exon_count=32, OMIM='', no_hdr0=''),
  243963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243963, CurrentID=0, Status='live', Symbol='Zfp473', Aliases='D030014N22Rik, Znf473, zfp-100', description='zinc finger protein 473', other_designations='zinc finger protein 100', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=44731480, end_position_on_the_genomic_accession=44748617, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  768252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=768252, CurrentID=0, Status='live', Symbol='Foxl2os', Aliases='ENSMUSG00000074122', description='forkhead box L2, opposite strand', other_designations='forkhead box L2 opposite strand transcript', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=98949155, end_position_on_the_genomic_accession=98955310, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  243967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243967, CurrentID=0, Status='live', Symbol='Ntn5', Aliases='Gm484', description='netrin 5', other_designations='netrin-1-like protein', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45684022, end_position_on_the_genomic_accession=45694556, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  243978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243978, CurrentID=0, Status='live', Symbol='Mrgprx2', Aliases='G370024M05Rik, MrgB10, Mrgprb10', description='MAS-related GPR, member X2', other_designations='MAS-related GPR, member B10|mas-related G-protein coupled receptor member B10', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48478619, end_position_on_the_genomic_accession=48499270, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  243979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243979, CurrentID=0, Status='live', Symbol='Mrgprb2', Aliases='4833406I20Rik, Mgrg14', description='MAS-related GPR, member B2', other_designations='Mas-related G protein-coupled receptor g14', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48550965, end_position_on_the_genomic_accession=48658094, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  243983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=243983, CurrentID=0, Status='live', Symbol='Zdhhc13', Aliases='2410004E01Rik, C530010M18, HIP3RP, Hip14l, kojak, skc4', description='zinc finger, DHHC domain containing 13', other_designations='DHHC-13|HIP14-related protein|huntingtin-interacting protein 14-related protein|probable palmitoyltransferase ZDHHC13|zinc finger DHHC domain-containing protein 13', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=48789003, end_position_on_the_genomic_accession=48827440, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  244049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244049, CurrentID=0, Status='live', Symbol='Mctp2', Aliases='Gm489', description='multiple C2 domains, transmembrane 2', other_designations='', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=72077827, end_position_on_the_genomic_accession=72306595, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  244058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244058, CurrentID=0, Status='live', Symbol='Rgma', Aliases='BC059072, C230063O06', description='repulsive guidance molecule family member A', other_designations='RGM domain family, member A', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=73375520, end_position_on_the_genomic_accession=73419899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  244059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244059, CurrentID=0, Status='live', Symbol='Chd2', Aliases='2810013C04Rik, 2810040A01Rik, 5630401D06Rik, AI851092, BC029703, CHD-2', description='chromodomain helicase DNA binding protein 2', other_designations='ATP-dependent helicase CHD2', map_location='7 D1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=73426652, end_position_on_the_genomic_accession=73541746, orientation='minus', exon_count=39, OMIM='', no_hdr0=''),
  637277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=637277, CurrentID=0, Status='live', Symbol='Sycp2l', Aliases='EG621792, Gm268', description='synaptonemal complex protein 2-like', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=41141914, end_position_on_the_genomic_accession=41178920, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  244071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244071, CurrentID=0, Status='live', Symbol='Agbl1', Aliases='Ccp4, D430020F16, EG244071, Nna1-l1', description='ATP/GTP binding protein-like 1', other_designations='ATP/GTP-binding protein-like 1|carboxypeptidase 4, cytosolic', map_location='7 D2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=76229887, end_position_on_the_genomic_accession=77124698, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  244091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244091, CurrentID=0, Status='live', Symbol='Fsd2', Aliases='9830160G03Rik, Spryd1', description='fibronectin type III and SPRY domain containing 2', other_designations='SPRY domain containing 1|SPRY domain-containing protein 1', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81534354, end_position_on_the_genomic_accession=81566981, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  244141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244141, CurrentID=0, Status='live', Symbol='Nars2', Aliases='AI875199', description='asparaginyl-tRNA synthetase 2 (mitochondrial)(putative)', other_designations='asnRS|asparagine--tRNA ligase|probable asparaginyl-tRNA synthetase, mitochondrial', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=96951510, end_position_on_the_genomic_accession=97064758, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  244152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244152, CurrentID=0, Status='live', Symbol='Tsku', Aliases='9530051K01Rik, E2ig4, Lrrc54, Tsk', description='tsukushi, small leucine rich proteoglycan', other_designations='leucine rich repeat containing 54|leucine-rich repeat-containing protein 54', map_location='7 E2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=98350668, end_position_on_the_genomic_accession=98361393, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  244178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244178, CurrentID=0, Status='live', Symbol='Ubqln3', Aliases='4933400K24Rik', description='ubiquilin 3', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104140621, end_position_on_the_genomic_accession=104143321, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  244179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244179, CurrentID=0, Status='live', Symbol='Ubqlnl', Aliases='4922504M18Rik, Gm490', description='ubiquilin-like', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104148259, end_position_on_the_genomic_accession=104150556, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  244180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244180, CurrentID=0, Status='live', Symbol='Olfm5', Aliases='', description='olfactomedin 5', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104153013, end_position_on_the_genomic_accession=104164816, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  244187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244187, CurrentID=0, Status='live', Symbol='Olfr684', Aliases='GA_x6K02T2PBJ9-7786441-7785503, MOR40-8P', description='olfactory receptor 684', other_designations='olfactory receptor GA_x6K02T2PBJ9-7786441-7785503', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105156742, end_position_on_the_genomic_accession=105157680, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  244198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244198, CurrentID=0, Status='live', Symbol='Olfml1', Aliases='6720478C22, AV321997, BC047207, MVAL564, ONT2, mONT2', description='olfactomedin-like 1', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107567433, end_position_on_the_genomic_accession=107591365, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  244199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244199, CurrentID=0, Status='live', Symbol='Ovch2', Aliases='9230106D23Rik, Ovtn', description='ovochymase 2', other_designations='oviductin', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107781544, end_position_on_the_genomic_accession=107801179, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  244202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244202, CurrentID=0, Status='live', Symbol='Nlrp10', Aliases='6430548I20Rik, Nalp10, Napl10, PAN5, Pynod', description='NLR family, pyrin domain containing 10', other_designations='NACHT, LRR and PYD containing protein 10|NACHT, leucine rich repeat and PYD containing 10|NACHT/LRR/pyrin domain-containing protein 10|neuronal apoptosis inhibitor protein 10', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108921853, end_position_on_the_genomic_accession=108930158, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  244209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244209, CurrentID=0, Status='live', Symbol='Cyp2r1', Aliases='', description='cytochrome P450, family 2, subfamily r, polypeptide 1', other_designations='cytochrome P450, 2r1', map_location='7 F1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114550123, end_position_on_the_genomic_accession=114565618, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  244216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244216, CurrentID=0, Status='live', Symbol='Zfp771', Aliases='G630024C07Rik, Znf771', description='zinc finger protein 771', other_designations='mesenchymal stem cell protein DSC43', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127244480, end_position_on_the_genomic_accession=127254807, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  244218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244218, CurrentID=0, Status='live', Symbol='Ctf2', Aliases='Gm494, NP', description='cardiotrophin 2', other_designations='CT-2|neuropoietin', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127717816, end_position_on_the_genomic_accession=127725616, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  244219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244219, CurrentID=0, Status='live', Symbol='Zfp668', Aliases='BC030314, E130018B19Rik, Znf668', description='zinc finger protein 668', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=127864543, end_position_on_the_genomic_accession=127876823, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  244233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244233, CurrentID=0, Status='live', Symbol='Cd163l1', Aliases='B430307C05, E430002D04Rik, Scart1', description='CD163 molecule-like 1', other_designations='scavenger receptor cysteine-rich type 1 protein CD163c-alpha', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140197218, end_position_on_the_genomic_accession=140231145, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  244234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244234, CurrentID=0, Status='live', Symbol='5830411N06Rik', Aliases='Scart2', description='RIKEN cDNA 5830411N06 gene', other_designations='mFLJ00268 protein', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140247279, end_position_on_the_genomic_accession=140300734, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  244237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244237, CurrentID=0, Status='live', Symbol='Tnfrsf26', Aliases='Tnfrh3, Tnfrsf24', description='tumor necrosis factor receptor superfamily, member 26', other_designations='TNF receptor homolog 3|tumor necrosis factor receptor superfamily, member 24', map_location='7 88.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143607685, end_position_on_the_genomic_accession=143627845, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  244238 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244238, CurrentID=0, Status='live', Symbol='Mrgpre', Aliases='C130069N09Rik, Ebrt3, Mgrf, MrgE', description='MAS-related GPR, member E', other_designations='G protein-coupled receptor MrgE|Mas-related G protein-coupled receptor f|evolutionary breakpoint transcript 3 protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143778363, end_position_on_the_genomic_accession=143784500, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  106248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106248, CurrentID=0, Status='live', Symbol='Qtrtd1', Aliases='3110012M05Rik, 4930470H18Rik, AI648807', description='queuine tRNA-ribosyltransferase domain containing 1', other_designations='queuine tRNA-ribosyltransferase domain-containing protein 1', map_location='16 B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=43861407, end_position_on_the_genomic_accession=43889680, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  244281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244281, CurrentID=0, Status='live', Symbol='Myo16', Aliases='BM140241, C230040D10Rik, Gm1105, NYAP3, mKIAA0865', description='myosin XVI', other_designations='myosin-XVI|neuronal tyrosine phosphorylated adaptor for phosphoinositide 3-kinase adaptor 3|neuronal tyrosine-phosphorylated phosphoinositide-3-kinase adapter 3|unconventional myosin-16', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=10153916, end_position_on_the_genomic_accession=10634742, orientation='plus', exon_count=39, OMIM='', no_hdr0=''),
  637515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=637515, CurrentID=0, Status='live', Symbol='Nlrp1b', Aliases='ENSMUSG00000070390, Nalp1b, OTTMUSG00000006087, OTTMUSG00000006089', description='NLR family, pyrin domain containing 1B', other_designations='NACHT-, LRR-, and PYD-containing protein 1 paralog b', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=71153014, end_position_on_the_genomic_accession=71230774, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  244310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244310, CurrentID=0, Status='live', Symbol='Dlgap2', Aliases='6430596N04Rik, DAP2, Sapap2', description='discs, large (Drosophila) homolog-associated protein 2', other_designations='DAP-2|PSD-95/SAP90-binding protein 2|SAP90/PSD-95 associated protein 2|SAP90/PSD-95-associated protein 2', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=14094615, end_position_on_the_genomic_accession=14848626, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  244329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244329, CurrentID=0, Status='live', Symbol='Mcph1', Aliases='5430437K10Rik, BRIT1, D030046N04Rik, MCT', description='microcephaly, primary autosomal recessive 1', other_designations='', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=18595136, end_position_on_the_genomic_accession=18803189, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  244332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244332, CurrentID=0, Status='live', Symbol='Defb14', Aliases='BD-14', description='defensin beta 14', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19163156, end_position_on_the_genomic_accession=19195309, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  100048658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100048658, CurrentID=0, Status='live', Symbol='Ddx43', Aliases='ENSMUSG00000070291, OTTMUSG00000019690', description='DEAD (Asp-Glu-Ala-Asp) box polypeptide 43', other_designations='', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78395777, end_position_on_the_genomic_accession=78423589, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  244334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244334, CurrentID=0, Status='live', Symbol='Defb8', Aliases='BD-8, Defr1', description='defensin beta 8', other_designations='defensin related 1|defensin-related peptide', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19445769, end_position_on_the_genomic_accession=19447606, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  244349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244349, CurrentID=0, Status='live', Symbol='Kat6a', Aliases='1500036M03, 9930021N24Rik, MOZ, Myst3, Zfp220', description='K(lysine) acetyltransferase 6A', other_designations='MOZ, YBF2/SAS3, SAS2 and TIP60 protein 3|MYST histone acetyltransferase (monocytic leukemia) 3|MYST-3|histone acetyltransferase MYST3|monocytic leukemia zinc finger homolog|monocytic leukemia zinc finger protein|zinc finger protein 220', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22859516, end_position_on_the_genomic_accession=22943262, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  244373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244373, CurrentID=0, Status='live', Symbol='Erlin2', Aliases='BC036333, C87251, Spfh2', description='ER lipid raft associated 2', other_designations='SPFH domain family, member 2|SPFH domain-containing protein 2|endoplasmic reticulum lipid raft-associated protein 2|stomatin-prohibitin-flotillin-HflC/K domain-containing protein 2', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27023266, end_position_on_the_genomic_accession=27039437, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  244416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244416, CurrentID=0, Status='live', Symbol='Ppp1r3b', Aliases='6430576E21, AW821953, GL, PTG', description='protein phosphatase 1, regulatory (inhibitor) subunit 3B', other_designations='PP1 subunit R4|hepatic glycogen-targeting protein phosphatase 1 regulatory subunit GL|protein phosphatase 1 regulatory subunit 4|protein phosphatase 1 subunit GL', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=35375668, end_position_on_the_genomic_accession=35388139, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  244418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244418, CurrentID=0, Status='live', Symbol='D8Ertd82e', Aliases='9830148H23, NACK, Sgk223, mFLJ00269', description='DNA segment, Chr 8, ERATO Doi 82, expressed', other_designations='pragmin|sugen kinase 223', map_location='8 22.78 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=36094625, end_position_on_the_genomic_accession=36147787, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  244421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244421, CurrentID=0, Status='live', Symbol='Lonrf1', Aliases='', description='LON peptidase N-terminal domain and ring finger 1', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=36216064, end_position_on_the_genomic_accession=36249516, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  244431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244431, CurrentID=0, Status='live', Symbol='Sgcz', Aliases='C230085N17Rik', description='sarcoglycan zeta', other_designations='ZSG1|zeta-SG', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=37517840, end_position_on_the_genomic_accession=38661508, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  244448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244448, CurrentID=0, Status='live', Symbol='Triml1', Aliases='4933403D14, BC050188', description='tripartite motif family-like 1', other_designations='tripartite motif family-like protein 1', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=43129806, end_position_on_the_genomic_accession=43180391, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  244484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244484, CurrentID=0, Status='live', Symbol='Wdr17', Aliases='3010002I12Rik, B230207L18Rik', description='WD repeat domain 17', other_designations='', map_location='8 B3.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=54629611, end_position_on_the_genomic_accession=54724517, orientation='minus', exon_count=33, OMIM='', no_hdr0=''),
  244486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244486, CurrentID=0, Status='live', Symbol='Adam29', Aliases='', description='a disintegrin and metallopeptidase domain 29', other_designations='a disintegrin and metalloprotease domain 29|a disintegrin and metalloproteinase domain 29', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=55870912, end_position_on_the_genomic_accession=55906964, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  723924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723924, CurrentID=0, Status='live', Symbol='Mir218-2', Aliases='Mirn218-2, mir-218-2', description='microRNA 218-2', other_designations='mir 218-2|mmu-mir-218-2', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=35616816, end_position_on_the_genomic_accession=35616925, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  244548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244548, CurrentID=0, Status='live', Symbol='Elmod2', Aliases='9830169G11Rik', description='ELMO/CED-12 domain containing 2', other_designations='ELMO domain containing 2', map_location='8 C2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83312632, end_position_on_the_genomic_accession=83332486, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  244550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244550, CurrentID=0, Status='live', Symbol='Podnl1', Aliases='5832418A03Rik, Gm506', description='podocan-like 1', other_designations='', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84125927, end_position_on_the_genomic_accession=84143016, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  244551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244551, CurrentID=0, Status='live', Symbol='Nanos3', Aliases='nos3', description='nanos homolog 3 (Drosophila)', other_designations='NOS-3', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84173730, end_position_on_the_genomic_accession=84176945, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  244556 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244556, CurrentID=0, Status='live', Symbol='Zfp791', Aliases='EG244556', description='zinc finger protein 791', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85109167, end_position_on_the_genomic_accession=85123095, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  244562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244562, CurrentID=0, Status='live', Symbol='Abcc12', Aliases='4930467B22Rik, MRP9', description='ATP-binding cassette, sub-family C (CFTR/MRP), member 12', other_designations='ATP-binding cassette sub-family C member 12|ATP-binding cassette transporter sub-family C member 12', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=86504576, end_position_on_the_genomic_accession=86580690, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  244579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244579, CurrentID=0, Status='live', Symbol='Tox3', Aliases='500-9, BC052044, C230068E13, CAGF9, Tnrc9', description='TOX high mobility group box family member 3', other_designations='trinucleotide repeat containing 9|trinucleotide repeat-containing gene 9 protein', map_location='8 43.96 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=90247037, end_position_on_the_genomic_accession=90349066, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  244585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244585, CurrentID=0, Status='live', Symbol='Rpgrip1l', Aliases='1700047E16Rik, 4931437C01, Ftm, Nphp8', description='Rpgrip1-like', other_designations='RPGR-interacting protein 1-like protein|RPGRIP1-like protein|fantom|nephrocystin-8', map_location='8 44.29 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=91217030, end_position_on_the_genomic_accession=91313291, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  244595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244595, CurrentID=0, Status='live', Symbol='Ces1a', Aliases='Ces1l, EG244595, Gm4976', description='carboxylesterase 1A', other_designations='', map_location='8 C5', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=93020214, end_position_on_the_genomic_accession=93048192, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  244608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244608, CurrentID=0, Status='live', Symbol='Ccdc113', Aliases='4933409I22, BC060957', description='coiled-coil domain containing 113', other_designations='', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=95534060, end_position_on_the_genomic_accession=95558888, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  723927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723927, CurrentID=0, Status='live', Symbol='Mir299a', Aliases='Mir299, Mirn299, mir-299a, mmu-mir-298, mmu-mir-299, mmu-mir-299a', description='microRNA 299a', other_designations='microRNA 299|mir 299', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109710638, end_position_on_the_genomic_accession=109710700, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  244631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244631, CurrentID=0, Status='live', Symbol='Pskh1', Aliases='AW539964, E130013P03Rik, b2b1230Clo', description='protein serine kinase H1', other_designations='Mutant line 1230|PSK-H1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105900474, end_position_on_the_genomic_accession=105931802, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  244646 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244646, CurrentID=0, Status='live', Symbol='Pkd1l3', Aliases='', description='polycystic kidney disease 1 like 3', other_designations='PC1-like 3 protein|polycystin-1L3', map_location='8 A', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109613994, end_position_on_the_genomic_accession=109673526, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  244650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244650, CurrentID=0, Status='live', Symbol='Phlpp2', Aliases='AI481772, C130044A18Rik, Phlppl', description='PH domain and leucine rich repeat protein phosphatase 2', other_designations='GTPase activating protein and VPS9 domains 1', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=109868519, end_position_on_the_genomic_accession=109944671, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  244653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244653, CurrentID=0, Status='live', Symbol='Hydin', Aliases='1700034M11Rik, 4930545D19Rik, 4932703P14, A830061H17, AC069308.21gm4, hy-3, hy3, hyrh', description='HYDIN, axonemal central pair apparatus protein', other_designations='hydrocephalus 3|hydrocephalus inducing|protein Hy-3', map_location='8 57.5 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110266977, end_position_on_the_genomic_accession=110610253, orientation='plus', exon_count=87, OMIM='', no_hdr0=''),
  244654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244654, CurrentID=0, Status='live', Symbol='Mtss1l', Aliases='ABBA', description='metastasis suppressor 1-like', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=110721282, end_position_on_the_genomic_accession=110741400, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  637873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=637873, CurrentID=0, Status='live', Symbol='Vmn2r61', Aliases='Casr-rs2, EG637873, Gprc2a-rs2', description='vomeronasal 2, receptor 61', other_designations='G protein-coupled receptor, family C, group 2, member A, related sequence 2|calcium sensing receptor, related sequence 2', map_location='7 B3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=42260053, end_position_on_the_genomic_accession=42300755, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  244666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244666, CurrentID=0, Status='live', Symbol='Sprtn', Aliases='Gm505', description='SprT-like N-terminal domain', other_designations='protein with SprT-like domain at the N terminus|spartan', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124897771, end_position_on_the_genomic_accession=124906164, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  244667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244667, CurrentID=0, Status='live', Symbol='Disc1', Aliases='', description='disrupted in schizophrenia 1', other_designations='', map_location='8 73.26 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125050148, end_position_on_the_genomic_accession=125263822, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  244668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244668, CurrentID=0, Status='live', Symbol='Sipa1l2', Aliases='BC058408, mKIAA1389', description='signal-induced proliferation-associated 1 like 2', other_designations='SIPA1-like protein 2', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=125418063, end_position_on_the_genomic_accession=125569863, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  244672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244672, CurrentID=0, Status='live', Symbol='Cwf19l2', Aliases='3230401L03Rik, 9030209E10', description='CWF19-like 2, cell cycle control (S. pombe)', other_designations='', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=3403592, end_position_on_the_genomic_accession=3479236, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  244682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244682, CurrentID=0, Status='live', Symbol='Cntn5', Aliases='6720426O10Rik, A830025P08Rik, Gm507, NB-2', description='contactin 5', other_designations='neural recognition molecule NB-2', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=9582665, end_position_on_the_genomic_accession=10904775, orientation='minus', exon_count=36, OMIM='', no_hdr0=''),
  637908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=637908, CurrentID=0, Status='live', Symbol='Vmn2r53', Aliases='EG637908, Gm471', description='vomeronasal 2, receptor 53', other_designations='vomeronasal receptor Vmn2r53', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12581446, end_position_on_the_genomic_accession=12606544, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  244694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244694, CurrentID=0, Status='live', Symbol='Kdm4d', Aliases='4932416A15, Jmjd2d', description='lysine (K)-specific demethylase 4D', other_designations='jmjC domain-containing histone demethylation protein 3D|jumonji domain containing 2D|jumonji domain-containing protein 2D', map_location='9 A1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=14462548, end_position_on_the_genomic_accession=14500522, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  244698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244698, CurrentID=0, Status='live', Symbol='Hephl1', Aliases='Gm509, Zp, zyklopen', description='hephaestin-like 1', other_designations='', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15051841, end_position_on_the_genomic_accession=15112108, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  244701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244701, CurrentID=0, Status='live', Symbol='Mtnr1b', Aliases='Mel-1B-R, Mel1b, Mt2', description='melatonin receptor 1B', other_designations='MT2 melatonin receptor|Mel-1b melatonin receptor|mel1b receptor', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=15862614, end_position_on_the_genomic_accession=15874556, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  244713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244713, CurrentID=0, Status='live', Symbol='Zfp317', Aliases='4932416G03, D230022C05Rik, KRAB9, Zfp67, Zfp75, mszf40', description='zinc finger protein 317', other_designations='zinc finger protein 75', map_location='9 7.35 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19622039, end_position_on_the_genomic_accession=19649734, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  244721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244721, CurrentID=0, Status='live', Symbol='Zfp846', Aliases='2210010B09Rik, 9130221F13, Znf846', description='zinc finger protein 846', other_designations='', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20581282, end_position_on_the_genomic_accession=20605404, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  244723 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244723, CurrentID=0, Status='live', Symbol='Olfm2', Aliases='A030009A06Rik', description='olfactomedin 2', other_designations='olfactomedin-2', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20667691, end_position_on_the_genomic_accession=20746311, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  244745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244745, CurrentID=0, Status='live', Symbol='Dpy19l1', Aliases='1100001I19Rik, 6030402K12, AI225861, AW558481', description='dpy-19-like 1 (C. elegans)', other_designations='dpy-19-like protein 1|protein dpy-19 homolog 1', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=24411776, end_position_on_the_genomic_accession=24503341, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  434008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=434008, CurrentID=0, Status='live', Symbol='Tmem178b', Aliases='EG434008, Gm5567', description='transmembrane protein 178B', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=39872540, end_position_on_the_genomic_accession=40248353, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  244757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244757, CurrentID=0, Status='live', Symbol='Glb1l2', Aliases='', description='galactosidase, beta 1-like 2', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=26763042, end_position_on_the_genomic_accession=26806472, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  244810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244810, CurrentID=0, Status='live', Symbol='AW551984', Aliases='3222402N16', description='expressed sequence AW551984', other_designations='', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39587396, end_position_on_the_genomic_accession=39604774, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  244813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244813, CurrentID=0, Status='live', Symbol='Bsx', Aliases='Bsx1a1b, Bsx', description='brain specific homeobox', other_designations='bsx', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40871854, end_position_on_the_genomic_accession=40878962, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  244853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244853, CurrentID=0, Status='live', Symbol='Nxpe4', Aliases='C130036J11, D930028F11Rik, Fam55d', description='neurexophilin and PC-esterase domain family, member 4', other_designations='NXPE family member 4|family with sequence similarity 55, member D|protein FAM55D', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=48162001, end_position_on_the_genomic_accession=48400025, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  244859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244859, CurrentID=0, Status='live', Symbol='Ankk1', Aliases='9930020N01Rik', description='ankyrin repeat and kinase domain containing 1', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=49412955, end_position_on_the_genomic_accession=49427791, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  244864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244864, CurrentID=0, Status='live', Symbol='Layn', Aliases='E030012M19Rik, Gm511', description='layilin', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51056780, end_position_on_the_genomic_accession=51077209, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  244867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244867, CurrentID=0, Status='live', Symbol='Arhgap20', Aliases='6530403F17Rik, A530023E23Rik, RARhoGAP, mKIAA1391', description='Rho GTPase activating protein 20', other_designations='RA and RhoGAP domain-containing protein|rho-type GTPase-activating protein 20', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=51765087, end_position_on_the_genomic_accession=51853856, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  244871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244871, CurrentID=0, Status='live', Symbol='Zc3h12c', Aliases='A230108E06, C230027N18Rik, mKIAA1726', description='zinc finger CCCH type containing 12C', other_designations='zinc finger CCCH domain-containing protein 12C|zinc finger CCCH-type containing 12C', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=52109583, end_position_on_the_genomic_accession=52169154, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  244879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244879, CurrentID=0, Status='live', Symbol='Npat', Aliases='6820401K01, AI427561, BB112559', description='nuclear protein in the AT region', other_designations='', map_location='9 29.12 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=53537041, end_position_on_the_genomic_accession=53575627, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  244882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244882, CurrentID=0, Status='live', Symbol='Tnfaip8l3', Aliases='9930029P06Rik, Gm517', description='tumor necrosis factor, alpha-induced protein 8-like 3', other_designations='TNF alpha-induced protein 8-like protein 3|TNFAIP8-like protein 3|tumor necrosis factor, alpha-induced protein 8-like protein 3', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54025606, end_position_on_the_genomic_accession=54068411, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  244885 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244885, CurrentID=0, Status='live', Symbol='Sh2d7', Aliases='4933412E14Rik', description='SH2 domain containing 7', other_designations='', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54534140, end_position_on_the_genomic_accession=54545030, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  638102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=638102, CurrentID=0, Status='live', Symbol='Vmn2r115', Aliases='EG638102, V2Rp4', description='vomeronasal 2, receptor 115', other_designations='vomeronasal 2 receptor p4', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23343977, end_position_on_the_genomic_accession=23360128, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  244891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244891, CurrentID=0, Status='live', Symbol='Scaper', Aliases='C430017I08, D530014O03Rik, Zfp291', description='S phase cyclin A-associated protein in the ER', other_designations='zinc finger protein 291', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=55549883, end_position_on_the_genomic_accession=55938158, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  244895 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244895, CurrentID=0, Status='live', Symbol='Peak1', Aliases='1110049L02Rik, 9530046P14, AA690169, BE132841, C230081A13Rik, SGK269, mKIAA2002', description='pseudopodium-enriched atypical kinase 1', other_designations='NKF3 kinase family member|sugen kinase 269|tyrosine-protein kinase SgK269', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56201126, end_position_on_the_genomic_accession=56418050, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  723936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=723936, CurrentID=0, Status='live', Symbol='Mir411', Aliases='Mirn411, mir-411, mmu-mir-411', description='microRNA 411', other_designations='mir 411', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109710175, end_position_on_the_genomic_accession=109710256, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  244911 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244911, CurrentID=0, Status='live', Symbol='C2cd4a', Aliases='EG244911, Fam148a', description='C2 calcium-dependent domain containing 4A', other_designations='family with sequence similarity 148, member A', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=67830532, end_position_on_the_genomic_accession=67832330, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  62664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=62664, CurrentID=0, Status='live', Symbol='D15Mit119', Aliases='', description='DNA segment, Chr 15, Massachusetts Institute of Technology 119', other_designations='', map_location='15 37.7 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  113845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113845, CurrentID=0, Status='live', Symbol='Vmn1r48', Aliases='V1ra3', description='vomeronasal 1 receptor 48', other_designations='vomernasal 1 receptor Vmn1r48|vomeronasal 1 receptor, A3|vomeronasal type-1 receptor A3|vomeronasal type-1 receptor A6', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90035933, end_position_on_the_genomic_accession=90036841, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  113846 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113846, CurrentID=0, Status='live', Symbol='Vmn1r47', Aliases='V1RA7, V1ra4', description='vomeronasal 1 receptor 47', other_designations='vomernasal 1 receptor Vmn1r47|vomeronasal 1 receptor, A4|vomeronasal receptor 1 A7|vomeronasal type-1 receptor A4|vomeronasal type-1 receptor A7', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90021888, end_position_on_the_genomic_accession=90022820, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  113847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113847, CurrentID=0, Status='live', Symbol='Vmn1r43', Aliases='V1ra2, V1ra5', description='vomeronasal 1 receptor 43', other_designations='vomernasal 1 receptor Vmn1r43|vomeronasal receptor 1 A2|vomeronasal type-1 receptor A5', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89869461, end_position_on_the_genomic_accession=89870529, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  113848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113848, CurrentID=0, Status='live', Symbol='Vmn1r42', Aliases='V1ra1, V1ra3, V1ra6', description='vomeronasal 1 receptor 42', other_designations='vomernasal 1 receptor Vmn1r42|vomeronasal receptor 1 A1|vomeronasal receptor 1 A3|vomeronasal receptor V1RA10|vomeronasal type-1 receptor A10|vomeronasal type-1 receptor A6', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89844518, end_position_on_the_genomic_accession=89845615, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  244923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244923, CurrentID=0, Status='live', Symbol='Klhl31', Aliases='9830147P19Rik, D930047P17Rik, Kbtbd1', description='kelch-like 31', other_designations='kelch repeat and BTB (POZ) domain containing 1|kelch repeat and BTB domain-containing protein 1|kelch-like protein KLHL', map_location='9 E1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=77636732, end_position_on_the_genomic_accession=77660125, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113852, CurrentID=0, Status='live', Symbol='Vmn1r50', Aliases='V1RA5, V1rb1, VN2', description='vomeronasal 1 receptor 50', other_designations='pheromone receptor VN2|vomernasal 1 receptor Vmn1r50|vomeronasal 1 receptor, B1|vomeronasal receptor 1 A5|vomeronasal receptor 2|vomeronasal type 1 receptor B1|vomeronasal type-1 receptor A5|vomeronasal type-1 receptor B1', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90107275, end_position_on_the_genomic_accession=90108207, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  113853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113853, CurrentID=0, Status='live', Symbol='Vmn1r53', Aliases='V1rb3, VN5', description='vomeronasal 1 receptor 53', other_designations='pheromone receptor VN5|vomernasal 1 receptor Vmn1r53|vomeronasal 1 receptor, B3|vomeronasal receptor 5|vomeronasal type 1 receptor B3|vomeronasal type-1 receptor B3', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=90223317, end_position_on_the_genomic_accession=90224438, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  113854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113854, CurrentID=0, Status='live', Symbol='Vmn1r44', Aliases='V1RA10, V1rb11, V1rb4', description='vomeronasal 1 receptor 44', other_designations='vomernasal 1 receptor Vmn1r44|vomeronasal 1 receptor, B4|vomeronasal receptor 1 A10|vomeronasal receptor V1RB11|vomeronasal type-1 receptor A10|vomeronasal type-1 receptor B11|vomeronasal type-1 receptor B4', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89892161, end_position_on_the_genomic_accession=89894206, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  113855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113855, CurrentID=0, Status='live', Symbol='Vmn1r40', Aliases='V1rb7', description='vomeronasal 1 receptor 40', other_designations='vomernasal 1 receptor Vmn1r40|vomeronasal 1 receptor, B7|vomeronasal type 1 receptor B7|vomeronasal type-1 receptor A11|vomeronasal type-1 receptor B7', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89711264, end_position_on_the_genomic_accession=89715135, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113856 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113856, CurrentID=0, Status='live', Symbol='Vmn1r46', Aliases='V1rb8', description='vomeronasal 1 receptor 46', other_designations='vomernasal 1 receptor Vmn1r46|vomeronasal 1 receptor, B8|vomeronasal type 1 receptor B8|vomeronasal type-1 receptor A13|vomeronasal type-1 receptor B8', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89969304, end_position_on_the_genomic_accession=89977100, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113857, CurrentID=0, Status='live', Symbol='Vmn1r41', Aliases='V1ra12, V1rb9', description='vomeronasal 1 receptor 41', other_designations='vomernasal 1 receptor Vmn1r41|vomeronasal receptor 1 A12|vomeronasal type-1 receptor B9', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=89740888, end_position_on_the_genomic_accession=89747414, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  113858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113858, CurrentID=0, Status='live', Symbol='Vmn1r10', Aliases='V1rc1', description='vomeronasal 1 receptor 10', other_designations='vomernasal 1 receptor Vmn1r10', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57113425, end_position_on_the_genomic_accession=57114360, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  113859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113859, CurrentID=0, Status='live', Symbol='Vmn1r29', Aliases='V1rc2', description='vomeronasal 1 receptor 29', other_designations='vomernasal 1 receptor Vmn1r29|vomeronasal 1 receptor, C2', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=58301384, end_position_on_the_genomic_accession=58308284, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  113860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113860, CurrentID=0, Status='live', Symbol='Vmn1r11', Aliases='V1rc3', description='vomeronasal 1 receptor 11', other_designations='vomernasal 1 receptor Vmn1r11', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57137353, end_position_on_the_genomic_accession=57138252, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  113862 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113862, CurrentID=0, Status='live', Symbol='Vmn1r13', Aliases='V1rc5', description='vomeronasal 1 receptor 13', other_designations='vomernasal 1 receptor Vmn1r13|vomeronasal type 1 receptor C5', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57203298, end_position_on_the_genomic_accession=57210760, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  113863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113863, CurrentID=0, Status='live', Symbol='Vmn1r15', Aliases='V1rc6', description='vomeronasal 1 receptor 15', other_designations='vomernasal 1 receptor Vmn1r15|vomeronasal type 1 receptor C6', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57258149, end_position_on_the_genomic_accession=57259048, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  113864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113864, CurrentID=0, Status='live', Symbol='Vmn1r14', Aliases='V1rc7', description='vomeronasal 1 receptor 14', other_designations='vomernasal 1 receptor Vmn1r14', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57233439, end_position_on_the_genomic_accession=57234350, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  113865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113865, CurrentID=0, Status='live', Symbol='Vmn1r25', Aliases='V1rc8', description='vomeronasal 1 receptor 25', other_designations='vomernasal 1 receptor Vmn1r25', map_location='6 B3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=57978394, end_position_on_the_genomic_accession=57979302, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  113868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=113868, CurrentID=0, Status='live', Symbol='Acaa1a', Aliases='Acaa, Acaa1, D9Ertd25e, PTL', description='acetyl-Coenzyme A acyltransferase 1A', other_designations='acetyl-CoA acyltransferase A|beta-ketothiolase A|peroxisomal 3-ketoacyl-CoA thiolase|peroxisomal 3-oxoacyl-CoA thiolase A|peroxisomal 3-oxoacyl-Coenzyme A thiolase|thiolase A', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119341052, end_position_on_the_genomic_accession=119350296, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  193742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=193742, CurrentID=0, Status='live', Symbol='Abhd16a', Aliases='AI326074, Bat-5, Bat5, D17H6S82E, NG26', description='abhydrolase domain containing 16A', other_designations='HLA-B associated transcript 5|HLA-B-associated transcript 5|protein BAT5', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35089233, end_position_on_the_genomic_accession=35102987, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  244954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244954, CurrentID=0, Status='live', Symbol='Prss35', Aliases='6030424L22Rik, P3D9', description='protease, serine 35', other_designations='protease, serine, 35', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=86743240, end_position_on_the_genomic_accession=86758443, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  244958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244958, CurrentID=0, Status='live', Symbol='Mrap2', Aliases='BB633055', description='melanocortin 2 receptor accessory protein 2', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=87144224, end_position_on_the_genomic_accession=87184051, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  244962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=244962, CurrentID=0, Status='live', Symbol='Snx14', Aliases='B830022K16, C330035N22Rik, YR-14', description='sorting nexin 14', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=88376747, end_position_on_the_genomic_accession=88439565, orientation='minus', exon_count=29, OMIM='', no_hdr0=''),
  245000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245000, CurrentID=0, Status='live', Symbol='Atr', Aliases='', description='ataxia telangiectasia and Rad3 related', other_designations='ataxia telangiectasia and Rad3-related protein', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=95857597, end_position_on_the_genomic_accession=95952760, orientation='plus', exon_count=47, OMIM='', no_hdr0=''),
  245007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245007, CurrentID=0, Status='live', Symbol='Zbtb38', Aliases='A930014K01Rik, CIBZ', description='zinc finger and BTB domain containing 38', other_designations='CtBP-interacting BTB zinc finger protein|Zenon homolog', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96685393, end_position_on_the_genomic_accession=96731675, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  245026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245026, CurrentID=0, Status='live', Symbol='Col6a6', Aliases='E330019B14, E330026B02Rik', description='collagen, type VI, alpha 6', other_designations='', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=105689417, end_position_on_the_genomic_accession=105828245, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  245038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245038, CurrentID=0, Status='live', Symbol='Dclk3', Aliases='BC056929, C730036H08, Dcamkl3', description='doublecortin-like kinase 3', other_designations='CLICK-I and II-related|CLr|Click-I, -II related|doublecortin and CaM kinase-like 3|doublecortin-like and CAM kinase-like 3', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=111439081, end_position_on_the_genomic_accession=111489611, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  245049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245049, CurrentID=0, Status='live', Symbol='Myrip', Aliases='A230081N12Rik, Slac2-c, Slac2c', description='myosin VIIA and Rab interacting protein', other_designations='exophilin 8|exophilin-8|myosin-VIIa- and Rab-interacting protein|slp homolog lacking C2 domains c|synaptotagmin-like protein lacking C2 domains C', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=120303919, end_position_on_the_genomic_accession=120476101, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  245050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245050, CurrentID=0, Status='live', Symbol='Fam198a', Aliases='C730027P07Rik', description='family with sequence similarity 198, member A', other_designations='', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=121950988, end_position_on_the_genomic_accession=121988210, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  245109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245109, CurrentID=0, Status='live', Symbol='Zscan4c', Aliases='Gm397, XM_142517, Zscan4d', description='zinc finger and SCAN domain containing 4C', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11005745, end_position_on_the_genomic_accession=11010547, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  215615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=215615, CurrentID=0, Status='live', Symbol='Rnpep', Aliases='', description='arginyl aminopeptidase (aminopeptidase B)', other_designations='AP-B|arginine aminopeptidase|cytosol aminopeptidase IV', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=135262699, end_position_on_the_genomic_accession=135284423, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  245126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245126, CurrentID=0, Status='live', Symbol='Tarm1', Aliases='9930022N03Rik, ENSMUSG00000053338, Gm9904, OLT-2', description='T cell-interacting, activating receptor on myeloid cells 1', other_designations='OSCAR-like transcript-2 protein', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3486499, end_position_on_the_genomic_accession=3502643, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  245128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245128, CurrentID=0, Status='live', Symbol='AU018091', Aliases='', description='expressed sequence AU018091', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3154658, end_position_on_the_genomic_accession=3169252, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  114066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114066, CurrentID=0, Status='live', Symbol='Heal2', Aliases='', description='wound healing/regeneration 2', other_designations='', map_location='13 63.03 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  237465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237465, CurrentID=0, Status='live', Symbol='Ccdc38', Aliases='4933417K05Rik', description='coiled-coil domain containing 38', other_designations='', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=93540567, end_position_on_the_genomic_accession=93584326, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  114076 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114076, CurrentID=0, Status='live', Symbol='Axtofa1', Aliases='', description='anxiety-open field activity 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245195, CurrentID=0, Status='live', Symbol='Retnlg', Aliases='Fizz3, Relmg, Xcp1', description='resistin like gamma', other_designations='RELMgamma|resistin-like molecule gamma', map_location='16 30.75 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=48872608, end_position_on_the_genomic_accession=48874498, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114128 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114128, CurrentID=0, Status='live', Symbol='Laptm4b', Aliases='C330023P13Rik, LAPTM4beta', description='lysosomal-associated protein transmembrane 4B', other_designations='lysosomal-associated transmembrane protein 4', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=34238026, end_position_on_the_genomic_accession=34284295, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  114141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114141, CurrentID=0, Status='live', Symbol='Cldn16', Aliases='PCLN1, claudin-16', description='claudin 16', other_designations='H59D2a protein|paracellin-1', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=26463135, end_position_on_the_genomic_accession=26482765, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  114142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114142, CurrentID=0, Status='live', Symbol='Foxp2', Aliases='2810043D05Rik, AI449000, CAG-16, D0Kist7', description='forkhead box P2', other_designations='', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=14901226, end_position_on_the_genomic_accession=15441977, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  114143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114143, CurrentID=0, Status='live', Symbol='Atp6v0b', Aliases='2310024H13Rik, Atp6f, VMA16', description='ATPase, H+ transporting, lysosomal V0 subunit B', other_designations='23 kDa subunit of V-ATPase|ATPase, H+ transporting, V0 subunit B|ATPase, H+ transporting, lysosomal (vacuolar proton pump)|ATPase, H+ transporting, lysosomal 21kDa, V0 subunit B|V-ATPase 21 kDa proteolipid subunit|lysosomal 21kDa|vacuolar proton pump 21 kDa proteolipid subunit', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117884326, end_position_on_the_genomic_accession=117887340, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  245263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245263, CurrentID=0, Status='live', Symbol='Gm4981', Aliases='EG245263', description='predicted gene 4981', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=58234848, end_position_on_the_genomic_accession=58236660, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  245269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245269, CurrentID=0, Status='live', Symbol='Nim1k', Aliases='E130304F04Rik, Nim1', description='NIM1 serine/threonine protein kinase', other_designations='NIM1 serine/threonine-protein kinase', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=119710094, end_position_on_the_genomic_accession=119755882, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  245282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245282, CurrentID=0, Status='live', Symbol='Apol10a', Aliases='9030421J09Rik', description='apolipoprotein L 10A', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77477047, end_position_on_the_genomic_accession=77491069, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  114228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114228, CurrentID=0, Status='live', Symbol='Prss1', Aliases='Try-1, Try1, Trygn16', description='protease, serine 1 (trypsin 1)', other_designations='protease, serine, 1 (trypsin 1)|trypsin 1|trypsinogen 16', map_location='6 19.6 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41458930, end_position_on_the_genomic_accession=41463786, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  114229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114229, CurrentID=0, Status='live', Symbol='Kiss1r', Aliases='Gpr54, KiSS-1', description='KISS1 receptor', other_designations='G protein-coupled receptor 54|G-protein coupled receptor 54|G-protein coupled receptor OT7T175|kiSS-1R|kisspeptins receptor|mOT7T175|metastin receptor|orphan G protein-coupled receptor 54', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79916971, end_position_on_the_genomic_accession=79922282, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114230, CurrentID=0, Status='live', Symbol='Aipl1', Aliases='A930007I01Rik', description='aryl hydrocarbon receptor-interacting protein-like 1', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72028717, end_position_on_the_genomic_accession=72042375, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  245308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245308, CurrentID=0, Status='live', Symbol='Zdhhc19', Aliases='Gm1744, Gm616', description='zinc finger, DHHC domain containing 19', other_designations='DHHC-19|zinc finger DHHC domain-containing protein 19', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=32496241, end_position_on_the_genomic_accession=32507214, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  114249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114249, CurrentID=0, Status='live', Symbol='Npnt', Aliases='1110009H02Rik, AA682063, AI314031, Nctn, POEM', description='nephronectin', other_designations='preosteoblast EGF-like repeat protein with MAM domain', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=132881745, end_position_on_the_genomic_accession=132950291, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  114255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114255, CurrentID=0, Status='live', Symbol='Dok4', Aliases='', description='docking protein 4', other_designations='downstream of tyrosine kinase 4', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94863828, end_position_on_the_genomic_accession=94876312, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  245350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245350, CurrentID=0, Status='live', Symbol='AA414768', Aliases='OTTMUSG00000017000', description='expressed sequence AA414768', other_designations='ubiquitin-conjugating enzyme E2Q (putative) 2 pseudogene', map_location='X A1.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=12936873, end_position_on_the_genomic_accession=12938541, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  245368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245368, CurrentID=0, Status='live', Symbol='Zfp300', Aliases='D930016N04Rik', description='zinc finger protein 300', other_designations='Novel KRAB box containing protein', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=21079150, end_position_on_the_genomic_accession=21089700, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  114301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114301, CurrentID=0, Status='live', Symbol='Palmd', Aliases='4631423C22Rik, PALML', description='palmdelphin', other_designations='paralemmin-related', map_location='3 50.49 cM', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=116918262, end_position_on_the_genomic_accession=116968952, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  114304 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114304, CurrentID=0, Status='live', Symbol='Slc28a3', Aliases='Cnt3', description='solute carrier family 28 (sodium-coupled nucleoside transporter), member 3', other_designations='CNT 3|concentrative Na(+)-nucleoside cotransporter 3|concentrative Na+-nucleoside cotransporter|mCNT3', map_location='13 B1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58553307, end_position_on_the_genomic_accession=58610877, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  245381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245381, CurrentID=0, Status='live', Symbol='Sowahd', Aliases='A630014H24Rik, Ankrd58', description='sosondowah ankyrin repeat domain family member D', other_designations='ankyrin repeat domain 58|ankyrin repeat domain-containing protein 58|protein sosondowah homolog D', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37048845, end_position_on_the_genomic_accession=37050418, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  245386 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245386, CurrentID=0, Status='live', Symbol='Tmem255a', Aliases='4933417N17, 6430550H21Rik, Fam70a', description='transmembrane protein 255A', other_designations='family with sequence similarity 70, member A|protein FAM70A', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38196573, end_position_on_the_genomic_accession=38252489, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  114323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114323, CurrentID=0, Status='live', Symbol='Yaa4', Aliases='Bxs3', description='Y-linked autoimmune acceleration QTL 4', other_designations='', map_location='1 81.63 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114332 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114332, CurrentID=0, Status='live', Symbol='Lyve1', Aliases='1200012G08Rik, Crsbp-1, Lyve-1, Xlkd1', description='lymphatic vessel endothelial hyaluronan receptor 1', other_designations='cell surface retention sequence binding protein-1|cell surface retention sequence-binding protein 1|extra cellular link domain-containing 1|extracellular link domain-containing protein 1|lymphatic vessel endothelial HA receptor-1|lymphatic vessel endothelial HA recptor-1', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=110850607, end_position_on_the_genomic_accession=110862953, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  245424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245424, CurrentID=0, Status='live', Symbol='Gpr101', Aliases='GPCR6, Gm365', description='G protein-coupled receptor 101', other_designations='G protein-coupled receptor GPR101', map_location='X A6', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=57496668, end_position_on_the_genomic_accession=57503757, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245446, CurrentID=0, Status='live', Symbol='Slitrk4', Aliases='6430513G13, D930039I09Rik', description='SLIT and NTRK-like family, member 4', other_designations='slit and trk like gene 4', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=64264883, end_position_on_the_genomic_accession=64277367, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  245450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245450, CurrentID=0, Status='live', Symbol='Slitrk2', Aliases='A930040J07', description='SLIT and NTRK-like family, member 2', other_designations='slit and trk like gene 2', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=66649318, end_position_on_the_genomic_accession=66661402, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  245468 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245468, CurrentID=0, Status='live', Symbol='Pnma3', Aliases='', description='paraneoplastic antigen MA3', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73064787, end_position_on_the_genomic_accession=73068191, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  245469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245469, CurrentID=0, Status='live', Symbol='Pdzd4', Aliases='B230341P03, D130075J17Rik, Kiaa1444-hp, LU1, PDZRN4L, Pdzk4, Pdzrnk4l, Xlu', description='PDZ domain containing 4', other_designations='PDZ domain-containing RING finger protein 4-like protein', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73793357, end_position_on_the_genomic_accession=73824969, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  245474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245474, CurrentID=0, Status='live', Symbol='Dkc1', Aliases='BC068171', description='dyskeratosis congenita 1, dyskerin', other_designations='dyskeratosis congenita 1, dyskerin homolog|dyskerin|nopp140-associated protein of 57 kDa|nucleolar protein NAP57|nucleolar protein family A member 4|snoRNP protein DKC1', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=75095854, end_position_on_the_genomic_accession=75109777, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  245492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245492, CurrentID=0, Status='live', Symbol='4930595M18Rik', Aliases='', description='RIKEN cDNA 4930595M18 gene', other_designations='', map_location='X B', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=81419575, end_position_on_the_genomic_accession=81469912, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  245509 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245509, CurrentID=0, Status='live', Symbol='4932429P05Rik', Aliases='', description='RIKEN cDNA 4932429P05 gene', other_designations='DUF625 domain containing protein', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=89752268, end_position_on_the_genomic_accession=89755491, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  245511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245511, CurrentID=0, Status='live', Symbol='4930415L06Rik', Aliases='Ppp4r3cp, Smek3p', description='RIKEN cDNA 4930415L06 gene', other_designations='SMEK homolog 2, suppressor of mek1|Smek homolog 3', map_location='X C1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=89930097, end_position_on_the_genomic_accession=89932852, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  245522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245522, CurrentID=0, Status='live', Symbol='Zc4h2', Aliases='Gm372, mKIAA1166', description='zinc finger, C4H2 domain containing', other_designations='hepatocellular carcinoma-associated antigen 127 homolog', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=95639190, end_position_on_the_genomic_accession=95658509, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  245525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245525, CurrentID=0, Status='live', Symbol='Hsf3', Aliases='B230358A15Rik, HSF 3, HSTF 3', description='heat shock transcription factor 3', other_designations='', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=96306184, end_position_on_the_genomic_accession=96367919, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  245526 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245526, CurrentID=0, Status='live', Symbol='Pgr15l', Aliases='', description='G protein-coupled receptor 15-like', other_designations='G protein-coupled receptor PGR15L', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=97072596, end_position_on_the_genomic_accession=97082104, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  245527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245527, CurrentID=0, Status='live', Symbol='Eda2r', Aliases='9430060M22Rik, TNFRSF27, Xedar', description='ectodysplasin A2 receptor', other_designations='EDA-A2 receptor|X linked ectodysplasin receptor|X-linked ectodysplasin-A2 receptor|ectodysplasin A2 isoform receptor', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=97333837, end_position_on_the_genomic_accession=97377218, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  245532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245532, CurrentID=0, Status='live', Symbol='Awat2', Aliases='9430062J17Rik, ARAT, Dgat2l4, WS', description='acyl-CoA wax alcohol acyltransferase 2', other_designations='acyl-CoA retinol O-fatty-acyltransferase|diacylglycerol O-acyltransferase 2-like protein 4|long-chain-alcohol O-fatty-acyltransferase 2|wax synthase 2', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100399336, end_position_on_the_genomic_accession=100494221, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  245533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245533, CurrentID=0, Status='live', Symbol='Awat1', Aliases='Dgat2l3', description='acyl-CoA wax alcohol acyltransferase 1', other_designations='diacylglycerol O-acyltransferase 2-like 3|diacylglycerol O-acyltransferase 2-like protein 3|long-chain-alcohol O-fatty-acyltransferase 1|wax synthase 1', map_location='X C3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=100572247, end_position_on_the_genomic_accession=100578206, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  245536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245536, CurrentID=0, Status='live', Symbol='Gm614', Aliases='', description='predicted gene 614', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101261377, end_position_on_the_genomic_accession=101263992, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  245537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245537, CurrentID=0, Status='live', Symbol='Nlgn3', Aliases='A230085M13Rik, HNL3, NL3', description='neuroligin 3', other_designations='gliotactin homolog|neuroligin 3 protein', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101299179, end_position_on_the_genomic_accession=101321350, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  114479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114479, CurrentID=0, Status='live', Symbol='Slc5a5', Aliases='NIS', description='solute carrier family 5 (sodium iodide symporter), member 5', other_designations='Na(+)/I(-) cotransporter|Na(+)/I(-) symporter|Na(+)/I(-)-symporter|sodium-iodide symporter|solute carrier family 5 member 5', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70882885, end_position_on_the_genomic_accession=70893053, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  245555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245555, CurrentID=0, Status='live', Symbol='C77370', Aliases='A230051P11, C77386, Kiaa2022, Xpn, mKIAA2022', description='expressed sequence C77370', other_designations='XLMR protein related to neurite extension', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=104077435, end_position_on_the_genomic_accession=104201757, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  245572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245572, CurrentID=0, Status='live', Symbol='Tbx22', Aliases='D230020M15Rik', description='T-box 22', other_designations='T-box protein 22', map_location='X 47.59 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=107667964, end_position_on_the_genomic_accession=107688980, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  245578 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245578, CurrentID=0, Status='live', Symbol='Pcdh11x', Aliases='PCDHX, PCDHX11, Pcdh11', description='protocadherin 11 X-linked', other_designations='protocadherin X-linked 11', map_location='X E2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=120290264, end_position_on_the_genomic_accession=120910622, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  245583 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245583, CurrentID=0, Status='live', Symbol='Tgif2lx1', Aliases='OTTMUSG00000018465, Tex1, Tgif2lx, Tgifx1', description='TGFB-induced factor homeobox 2-like, X-linked 1', other_designations='TGF beta-induced transcription factor 2-like protein|TGIF homeobox 1|testis expressed homeobox Tex1', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=118479708, end_position_on_the_genomic_accession=118480729, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  245595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245595, CurrentID=0, Status='live', Symbol='Zfp711', Aliases='2810409C01Rik, A230078I01Rik, Znf711', description='zinc finger protein 711', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=112600526, end_position_on_the_genomic_accession=112635070, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  245596 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245596, CurrentID=0, Status='live', Symbol='Hdx', Aliases='D030011N01Rik, Gm380', description='highly divergent homeobox', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=111575204, end_position_on_the_genomic_accession=111697079, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  114534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114534, CurrentID=0, Status='live', Symbol='Pgia17', Aliases='', description='proteoglycan induced arthritis 17', other_designations='', map_location='', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245607, CurrentID=0, Status='live', Symbol='Gprasp2', Aliases='5330440H13Rik, GASP-2, GASP2, Prpl5', description='G protein-coupled receptor associated sorting protein 2', other_designations='prion protein ligand 5', map_location='X 57.3 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=135839034, end_position_on_the_genomic_accession=135844730, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  114536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114536, CurrentID=0, Status='live', Symbol='Bbaa2', Aliases='Bb2', description='B.burgdorferi-associated arthritis 2', other_designations='', map_location='5 60.4 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114537, CurrentID=0, Status='live', Symbol='Bbaa3', Aliases='Bb3', description='B.burgdorferi-associated arthritis 3', other_designations='', map_location='5 47.29 cM', chromosome='5', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245610 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245610, CurrentID=0, Status='live', Symbol='Nxf3', Aliases='Gm384', description='nuclear RNA export factor 3', other_designations='', map_location='X 57.38 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136072099, end_position_on_the_genomic_accession=136085553, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  245615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245615, CurrentID=0, Status='live', Symbol='Kir3dl2', Aliases='Kirl2', description='killer cell immunoglobulin-like receptor, three domains, long cytoplasmic tail, 2', other_designations='KIR-like receptor 2|killer immunoglobulin-like receptor-like 2', map_location='X 58.45 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136447841, end_position_on_the_genomic_accession=136544052, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  245616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245616, CurrentID=0, Status='live', Symbol='Kir3dl1', Aliases='Kirl1, Kirl2, Krl1', description='killer cell immunoglobulin-like receptor, three domains, long cytoplasmic tail, 1', other_designations='KIR-like receptor 1|killer cell immunoglobulin-like receptor Kir3dl1|killer immunoglobulin receptor-like protein 1', map_location='X 58.77 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=136517999, end_position_on_the_genomic_accession=136534308, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  245622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245622, CurrentID=0, Status='live', Symbol='Fam199x', Aliases='', description='family with sequence similarity 199, X-linked', other_designations='', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=137049594, end_position_on_the_genomic_accession=137082503, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  245631 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245631, CurrentID=0, Status='live', Symbol='Mum1l1', Aliases='9430086K21Rik', description='melanoma associated antigen (mutated) 1-like 1', other_designations='MUM1-like protein 1|mutated melanoma-associated antigen 1-like protein 1', map_location='X F1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=139210030, end_position_on_the_genomic_accession=139238335, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  114564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114564, CurrentID=0, Status='live', Symbol='Csprs', Aliases='D1Lub1, HSR', description='component of Sp100-rs', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114565, CurrentID=0, Status='live', Symbol='Zbtb21', Aliases='5430437K12Rik, B430213I24Rik, Zfp295, Znf295, mKIAA1227', description='zinc finger and BTB domain containing 21', other_designations='zinc finger protein 295', map_location='16 C3-4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97947435, end_position_on_the_genomic_accession=97962629, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114566 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114566, CurrentID=0, Status='live', Symbol='Krt82', Aliases='Krt2-20', description='keratin 82', other_designations='K82|keratin complex 2, basic, gene 20|keratin-82|type II hair keratin Hb2|type-II keratin Kb22', map_location='15 56.9 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101541216, end_position_on_the_genomic_accession=101550659, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  114570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114570, CurrentID=0, Status='live', Symbol='Crip3', Aliases='TLP, TLP-A, TLP-B, Tlmp', description='cysteine-rich protein 3', other_designations='CRP-3|m17TLP|thymus LIM protein', map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46428195, end_position_on_the_genomic_accession=46431780, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  245650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245650, CurrentID=0, Status='live', Symbol='Gucy2f', Aliases='A930002I06, GC-F, GC2', description='guanylate cyclase 2f', other_designations='', map_location='X 62.85 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=142079286, end_position_on_the_genomic_accession=142197276, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  114584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114584, CurrentID=0, Status='live', Symbol='Clic1', Aliases='Clcp, G6', description='chloride intracellular channel 1', other_designations='NCC27|nuclear chloride ion channel 27', map_location='17 18.58 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35050243, end_position_on_the_genomic_accession=35058719, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114585, CurrentID=0, Status='live', Symbol='D17H6S53E', Aliases='G4, NG34', description='DNA segment, Chr 17, human D6S53E', other_designations='', map_location='17 18.59 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35126402, end_position_on_the_genomic_accession=35128855, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  245666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245666, CurrentID=0, Status='live', Symbol='Iqsec2', Aliases='Brag1', description='IQ motif and Sec7 domain 2', other_designations='', map_location='X 68.46 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=152144183, end_position_on_the_genomic_accession=152225237, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  245670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245670, CurrentID=0, Status='live', Symbol='Rragb', Aliases='', description='Ras-related GTP binding B', other_designations='rag B|ragB|rragb', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153139796, end_position_on_the_genomic_accession=153190192, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  245671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245671, CurrentID=0, Status='live', Symbol='Klf8', Aliases='A830097P10Rik, BKLF3, ZNF74', description='Kruppel-like factor 8', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=153225339, end_position_on_the_genomic_accession=153398453, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  114600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114600, CurrentID=0, Status='live', Symbol='Gm4736', Aliases='EG114600, Mp4, Mp6', description='predicted gene 4736', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.5', start_position_on_the_genomic_accession=132311063, end_position_on_the_genomic_accession=132314153, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114601, CurrentID=0, Status='live', Symbol='Ehbp1l1', Aliases='G430002G23Rik', description='EH domain binding protein 1-like 1', other_designations='tangerin', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5707369, end_position_on_the_genomic_accession=5726431, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  114602 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114602, CurrentID=0, Status='live', Symbol='Zmynd10', Aliases='Blu', description='zinc finger, MYND domain containing 10', other_designations='protein BLu', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=107545991, end_position_on_the_genomic_accession=107551321, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  114604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114604, CurrentID=0, Status='live', Symbol='Prdm15', Aliases='C21orf83, E130018M06Rik, ORF62, Zfp298', description='PR domain containing 15', other_designations='zinc finger protein 298', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=97791467, end_position_on_the_genomic_accession=97852427, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  114606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114606, CurrentID=0, Status='live', Symbol='Tle6', Aliases='1810057E06Rik, Grg6', description='transducin-like enhancer of split 6, homolog of Drosophila E(spl)', other_designations='Groucho-related protein 6', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=81590904, end_position_on_the_genomic_accession=81601190, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  114608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114608, CurrentID=0, Status='live', Symbol='Cia5', Aliases='mcia2', description='collagen induced arthritis QTL 5', other_designations='', map_location='3 37.01 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245684, CurrentID=0, Status='live', Symbol='Cnksr2', Aliases='Cnk2, Ksr2', description='connector enhancer of kinase suppressor of Ras 2', other_designations='CNK homolog protein 2|connector enhancer of KSR 2|connector enhancer of KSR2', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=157821573, end_position_on_the_genomic_accession=158043313, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  114615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114615, CurrentID=0, Status='live', Symbol='Elac1', Aliases='2610018O07Rik, 8430417G19Rik', description='elaC ribonuclease Z 1', other_designations='RNase Z 1|elaC homolog 1|elaC homolog protein 1|ribonuclease Z 1|tRNA 3 endonuclease 1|tRNase Z 1', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=73735038, end_position_on_the_genomic_accession=73754758, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  245688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245688, CurrentID=0, Status='live', Symbol='Rbbp7', Aliases='AA409861, AI173248, AU019541, BB114024, mRbAp46', description='retinoblastoma binding protein 7', other_designations='G1/S transition control protein-binding protein|RBBP-7|histone acetyltransferase type B subunit 2|nucleosome-remodeling factor subunit RBAP46|retinoblastoma-binding protein 7|retinoblastoma-binding protein p46', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=162760372, end_position_on_the_genomic_accession=162779090, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  114621 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114621, CurrentID=0, Status='live', Symbol='Mobq5', Aliases='Mob5', description='multigenic obesity QTL 5', other_designations='', map_location='2 38.54 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114622, CurrentID=0, Status='live', Symbol='Mobq6', Aliases='Mob6', description='multigenic obesity QTL 6', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245695, CurrentID=0, Status='live', Symbol='Tceanc', Aliases='A630018P17Rik, Gm390', description='transcription elongation factor A (SII) N-terminal and central domain containing', other_designations='TFIIS central domain-containing protein 1|TFS2-M domain-containing protein 1|novel transcription factor S-II (TFIIS), central domain containing protein', map_location='X F5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=166499812, end_position_on_the_genomic_accession=166518598, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  245704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245704, CurrentID=0, Status='live', Symbol='Aliq4', Aliases='', description='acute lung injury QTL 4', other_designations='', map_location='6 25.82 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114640, CurrentID=0, Status='live', Symbol='Pth2', Aliases='Tifp39, Tip39', description='parathyroid hormone 2', other_designations='tuberoinfundibular 39 residue protein|tuberoinfundibular peptide of 39 residues (TIP39) preprohormone', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45176565, end_position_on_the_genomic_accession=45181838, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  114641 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114641, CurrentID=0, Status='live', Symbol='Rpl31', Aliases='', description='ribosomal protein L31', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=39367851, end_position_on_the_genomic_accession=39371911, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  114642 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114642, CurrentID=0, Status='live', Symbol='Brdt', Aliases='7420412D09Rik, Brd6, Fsrg3', description='bromodomain, testis-specific', other_designations='RING3-like protein|bromodomain-containing female sterile homeotic-like protein|female sterile homeotic-related gene 3', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107331091, end_position_on_the_genomic_accession=107387061, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  114643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114643, CurrentID=0, Status='live', Symbol='Oas1c', Aliases='Oasl2, Oasl5', description="2'-5' oligoadenylate synthetase 1C", other_designations="2'-5' oligoadenylate synthetase-like 2", map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120800194, end_position_on_the_genomic_accession=120812540, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  114644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114644, CurrentID=0, Status='live', Symbol='Slc13a3', Aliases='NaDC-3, NaDC3, SDCT2', description='solute carrier family 13 (sodium-dependent dicarboxylate transporter), member 3', other_designations='Na(+)/dicarboxylate cotransporter 3|mNaDC3|sodium-dependent high-affinity dicarboxylate transporter 2', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165405295, end_position_on_the_genomic_accession=165473225, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  114652 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114652, CurrentID=0, Status='live', Symbol='Ly6g5c', Aliases='G5c, NG33', description='lymphocyte antigen 6 complex, locus G5C', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35108300, end_position_on_the_genomic_accession=35111953, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  114654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114654, CurrentID=0, Status='live', Symbol='Ly6g6d', Aliases='A930024F17Rik, G6d, G6f, MEGT1, NG25, NG32', description='lymphocyte antigen 6 complex, locus G6D', other_designations='megakaryocyte-enhanced gene transcript 1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35071342, end_position_on_the_genomic_accession=35080567, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114661, CurrentID=0, Status='live', Symbol='Prss28', Aliases='Isp1, mIsp-1', description='protease, serine 28', other_designations='ISP-1|implantation serine protease 1|implantation serine proteinase 1|protease, serine, 28|strypsin|tryptase-like proteinase', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25308646, end_position_on_the_genomic_accession=25311876, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114662, CurrentID=0, Status='live', Symbol='Prss29', Aliases='D19363, Isp2, mISP-2', description='protease, serine 29', other_designations='ISP-2|implantation serine protease 2|implantation serine proteinase 2|protease, serine, 29|strypsin-2|strypsin-related protein|tryptase-like proteinase', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25318654, end_position_on_the_genomic_accession=25322684, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114663 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114663, CurrentID=0, Status='live', Symbol='Impa2', Aliases='2210415D20Rik, AI326924, AW259601', description='inositol (myo)-1(or 4)-monophosphatase 2', other_designations='IMP 2|IMPase 2|inositol-1(or 4)-monophosphatase 2|myo-inositol monophosphatase A2', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=67289223, end_position_on_the_genomic_accession=67318841, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  114664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114664, CurrentID=0, Status='live', Symbol='Hsd17b11', Aliases='Dhrs8, Pan1b, SDR2, retSDR2', description='hydroxysteroid (17-beta) dehydrogenase 11', other_designations='17-beta-HSD 11|17-beta-HSD XI|17-beta-hydroxysteroid dehydrogenase 11|17-beta-hydroxysteroid dehydrogenase XI|17bHSD11|17beta-HSD11|17betaHSD11|17betaHSDXI|dehydrogenase/reductase (SDR family) member 8|dehydrogenase/reductase SDR family member 8|retinal short-chain dehydrogenase/reductase 2|retinal short-chain dehydrogenase/reductase SDR2', map_location='5 E4', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=103989765, end_position_on_the_genomic_accession=104021796, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  114671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114671, CurrentID=0, Status='live', Symbol='4930444G20Rik', Aliases='', description='RIKEN cDNA 4930444G20 gene', other_designations='', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=22066309, end_position_on_the_genomic_accession=22068079, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  114674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114674, CurrentID=0, Status='live', Symbol='Gtf2ird2', Aliases='1700012P16Rik', description='GTF2I repeat domain containing 2', other_designations='GTF2I repeat domain-containing protein 2|transcription factor GTF2IRD2', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=134183215, end_position_on_the_genomic_accession=134218146, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  114679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114679, CurrentID=0, Status='live', Symbol='Selm', Aliases='1500040L08Rik, A230103K18, Sepm', description='selenoprotein M', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3514702, end_position_on_the_genomic_accession=3517351, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  414084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=414084, CurrentID=0, Status='live', Symbol='Tnip3', Aliases='9030611K07Rik, ABIN-3, LIND', description='TNFAIP3 interacting protein 3', other_designations='ABIN-3 beta|TNFAIP3-interacting protein 3 beta|TNIP3 beta', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=65525335, end_position_on_the_genomic_accession=65634040, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  114712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114712, CurrentID=0, Status='live', Symbol='Ferd3l', Aliases='Mnato3, N-twist, Nato, Nato3, Ntwist, bHLHa31, fer3', description='Fer3-like (Drosophila)', other_designations='basic helix-loop-helix protein N-twist|nephew of atonal 3|neuronal twist', map_location='12 B2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=33928425, end_position_on_the_genomic_accession=33929310, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  114713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114713, CurrentID=0, Status='live', Symbol='Rasa2', Aliases='5430433H21Rik, AA517451, AU023900, GAP1m', description='RAS p21 protein activator 2', other_designations='', map_location='9 E4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=96539300, end_position_on_the_genomic_accession=96631615, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  114714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114714, CurrentID=0, Status='live', Symbol='Rad51c', Aliases='R51H3, Rad51l2', description='RAD51 homolog C', other_designations='RAD51-like protein 2', map_location='11 52.08 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87377607, end_position_on_the_genomic_accession=87404954, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  114715 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114715, CurrentID=0, Status='live', Symbol='Spred1', Aliases='5730461F13Rik, AL024345, AW047647', description='sprouty protein with EVH-1 domain 1, related sequence', other_designations='sprouty-related protein 1 with EVH-1 domain', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=117121071, end_position_on_the_genomic_accession=117182279, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  114716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114716, CurrentID=0, Status='live', Symbol='Spred2', Aliases='C79158', description='sprouty-related, EVH1 domain containing 2', other_designations='sprouty protein with EVH-1 domain 2, related sequence', map_location='11 A3.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=19924442, end_position_on_the_genomic_accession=20024026, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  114737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114737, CurrentID=0, Status='live', Symbol='Iba1', Aliases='', description='induction of brown adipocytes 1', other_designations='', map_location='2 49.45 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  114741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114741, CurrentID=0, Status='live', Symbol='Supt16', Aliases='Cdc68, Fact140, Spt16h, Supt16', description='suppressor of Ty 16', other_designations='FACT 140 kDa subunit|FACTp140|chromatin-specific transcription elongation factor 140 kDa subunit|facilitates chromatin transcription complex subunit SPT16|suppressor of Ty 16 homolog', map_location='14 26.83 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52160419, end_position_on_the_genomic_accession=52197239, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  245827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245827, CurrentID=0, Status='live', Symbol='Fat2', Aliases='EMI2, Fath2, Gm523, mKIAA0811', description='FAT atypical cadherin 2', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55250609, end_position_on_the_genomic_accession=55336621, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  245828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245828, CurrentID=0, Status='live', Symbol='Trappc1', Aliases='BET5, D11Ertd172e, MUM2', description='trafficking protein particle complex 1', other_designations='novel Sybindin-like family domain-containing protein', map_location='11 42.51 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69323986, end_position_on_the_genomic_accession=69325793, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  245839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245839, CurrentID=0, Status='live', Symbol='Gzmn', Aliases='GrN', description='granzyme N', other_designations='', map_location='14 28.19 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=56165786, end_position_on_the_genomic_accession=56174612, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  245841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245841, CurrentID=0, Status='live', Symbol='Polr2h', Aliases='', description='polymerase (RNA) II (DNA directed) polypeptide H', other_designations='DNA-directed RNA polymerase II subunit H|RNA polymerases I, II, and III subunit ABC3|RPB17|RPB8 homolog', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20717826, end_position_on_the_genomic_accession=20722272, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  114774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114774, CurrentID=0, Status='live', Symbol='Pawr', Aliases='2310001G03Rik, PAR4, Par-4', description='PRKC, apoptosis, WT1, regulator', other_designations='prostate apoptosis response 4 protein', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=108332189, end_position_on_the_genomic_accession=108414391, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  245847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245847, CurrentID=0, Status='live', Symbol='Amdhd2', Aliases='5730457F11Rik', description='amidohydrolase domain containing 2', other_designations='amidohydrolase domain-containing protein 2|glcNAc 6-P deacetylase', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24155833, end_position_on_the_genomic_accession=24163733, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  245857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245857, CurrentID=0, Status='live', Symbol='Ssh3', Aliases='BC028922, SSH-3, mSSH-3L', description='slingshot homolog 3 (Drosophila)', other_designations='SSH-3L|SSH-like protein 3|slingshot-3L|slingshot-like 3', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4261668, end_position_on_the_genomic_accession=4270276, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  245860 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245860, CurrentID=0, Status='live', Symbol='Atg9a', Aliases='AU019532, Apg9l1, Atg9, Atg9l1', description='autophagy related 9A', other_designations='APG9 autophagy 9-like 1|APG9-like 1|autophagy protein 9|autophagy-related 9-like 1|autophagy-related 9A', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=75180861, end_position_on_the_genomic_accession=75192010, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  245865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245865, CurrentID=0, Status='live', Symbol='Spag4', Aliases='1700041K21Rik, MNCb-0953, Sun4, mKIAA4118', description='sperm associated antigen 4', other_designations='SUN domain-containing protein 4|outer dense fiber-associated protein SPAG4|sperm antigen 4', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=156065015, end_position_on_the_genomic_accession=156069501, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  245866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245866, CurrentID=0, Status='live', Symbol='Ift52', Aliases='BC037708, NGD5', description='intraflagellar transport 52', other_designations='intraflagellar transport 52 homolog', map_location='2 H2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=163017298, end_position_on_the_genomic_accession=163046155, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  245867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245867, CurrentID=0, Status='live', Symbol='Pcmtd2', Aliases='5330414D10Rik, AI314201', description='protein-L-isoaspartate (D-aspartate) O-methyltransferase domain containing 2', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=181835932, end_position_on_the_genomic_accession=181857460, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  245877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245877, CurrentID=0, Status='live', Symbol='Map7d1', Aliases='AV028413, BC019977, Mtap7d1, Parcc1, Rprc1, mKIAA1187', description='MAP7 domain containing 1', other_designations='arginine/proline rich coiled-coil 1|microtubule-associated protein 7 domain containing 1|proline arginine rich coiled coil 1', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=126232167, end_position_on_the_genomic_accession=126256417, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  114806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114806, CurrentID=0, Status='live', Symbol='Tlln', Aliases='', description='tail length', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245880, CurrentID=0, Status='live', Symbol='Wasf3', Aliases='Scar3, Wave3', description='WAS protein family, member 3', other_designations='WASP family protein member 3|protein WAVE-3', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=146384990, end_position_on_the_genomic_accession=146471782, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  245884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245884, CurrentID=0, Status='live', Symbol='Fam71f2', Aliases='AI507611, GARI', description='family with sequence similarity 71, member F2', other_designations='Golgi-associated Rab2B interactor', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29277539, end_position_on_the_genomic_accession=29290684, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  114813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114813, CurrentID=0, Status='live', Symbol='Bw19', Aliases='QbwX', description='body weight QTL 19', other_designations='', map_location='X 24.88 cM', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  245886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245886, CurrentID=0, Status='live', Symbol='Ankrd27', Aliases='AA408090, BC016493, D330003H11Rik, Varp', description='ankyrin repeat domain 27 (VPS9 domain)', other_designations='VPS9 domain-containing protein|VPS9-ankyrin repeat-containing protein|VPS9-ankyrin-repeat protein|mFLJ00040 protein', map_location='7 B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=35586226, end_position_on_the_genomic_accession=35639237, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  245902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=245902, CurrentID=0, Status='live', Symbol='Ccdc15', Aliases='A630039F14Rik, D030074L19', description='coiled-coil domain containing 15', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37271292, end_position_on_the_genomic_accession=37353528, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  114863 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114863, CurrentID=0, Status='live', Symbol='Prosc', Aliases='1700024N20Rik, 2200002F22Rik', description='proline synthetase co-transcribed', other_designations='proline synthetase co-transcribed bacterial homolog protein', map_location='8 A3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=27042555, end_position_on_the_genomic_accession=27056132, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  114871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114871, CurrentID=0, Status='live', Symbol='Psg28', Aliases='cea16', description='pregnancy-specific glycoprotein 28', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18422536, end_position_on_the_genomic_accession=18432055, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  114872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114872, CurrentID=0, Status='live', Symbol='Psg29', Aliases='AA102956, AA409969, cea17', description='pregnancy-specific glycoprotein 29', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17203477, end_position_on_the_genomic_accession=17215756, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  114873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114873, CurrentID=0, Status='live', Symbol='Dscaml1', Aliases='4921507G06Rik, 4930435C18Rik, BB239540, mKIAA1132', description='Down syndrome cell adhesion molecule like 1', other_designations='Down syndrome cell adhesion molecule-like 1', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=45429495, end_position_on_the_genomic_accession=45753713, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  114874 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114874, CurrentID=0, Status='live', Symbol='Ddhd1', Aliases='Mir5131, PA-PLA1', description='DDHD domain containing 1', other_designations='DDHD domain-containing protein 1|phosphatidic acid-preferring phospholipase A1 homolog', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=45593171, end_position_on_the_genomic_accession=45658143, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  114875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114875, CurrentID=0, Status='live', Symbol='Plcz1', Aliases='1700041H07Rik, PLCzeta', description='phospholipase C, zeta 1', other_designations='1-phosphatidylinositol-4,5-bisphosphate phosphodiesterase zeta-1|PLC-zeta-1|phosphoinositide phospholipase C-zeta-1|phospholipase C zeta|phospholipase C-zeta-1|plcz1', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=139988274, end_position_on_the_genomic_accession=140041419, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  114886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114886, CurrentID=0, Status='live', Symbol='Cygb', Aliases='3110001K20Rik, HGb, Staap', description='cytoglobin', other_designations='histoglobin|stellate cell activation associated protein', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=116645595, end_position_on_the_genomic_accession=116654313, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  114889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114889, CurrentID=0, Status='live', Symbol='Vsx1', Aliases='CHX10-like', description='visual system homeobox 1 homolog (zebrafish)', other_designations='homeodomain protein RINX|retinal inner nuclear layer homeobox protein|transcription factor VSX1', map_location='2 74.74 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=150680702, end_position_on_the_genomic_accession=150689137, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  114893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114893, CurrentID=0, Status='live', Symbol='Dcun1d1', Aliases='Rp42, SCCRO, Tes3, pTes3', description='DCN1, defective in cullin neddylation 1, domain containing 1 (S. cerevisiae)', other_designations='DCUN1 domain-containing protein 1|defective in cullin neddylation protein 1-like protein 1|testis derived transcript 3|testis-specific protein 3', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=35892103, end_position_on_the_genomic_accession=35937605, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  114896 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=114896, CurrentID=0, Status='live', Symbol='Afg3l1', Aliases='1700047G05Rik, 3110061K15Rik', description='AFG3-like AAA ATPase 1', other_designations='AFG3(ATPase family gene 3)-like 1', map_location='8 E1-E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=123477862, end_position_on_the_genomic_accession=123503916, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  330721 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=330721, CurrentID=0, Status='live', Symbol='Nek5', Aliases='6530411J22, BB049667', description='NIMA (never in mitosis gene a)-related expressed kinase 5', other_designations='NIMA-related kinase 5|never in mitosis A-related kinase 5|nimA-related protein kinase 5', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=22073615, end_position_on_the_genomic_accession=22125415, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  246048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246048, CurrentID=0, Status='live', Symbol='Chodl', Aliases='3110074E07Rik, MT75, PRED12', description='chondrolectin', other_designations='transmembrane protein MT75', map_location='16 C3.2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=78930864, end_position_on_the_genomic_accession=78951733, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  246049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246049, CurrentID=0, Status='live', Symbol='Slc36a2', Aliases='A530067G19Rik, PAT2, Tramd1', description='solute carrier family 36 (proton/amino acid symporter), member 2', other_designations='proton/amino acid transporter 2|solute carrier family 36 member 2|tramdorin 1 (transmembrane domain rich protein)|tramdorin-1', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=55158468, end_position_on_the_genomic_accession=55185095, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  246080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246080, CurrentID=0, Status='live', Symbol='Defb7', Aliases='BD-7, Defb15', description='defensin beta 7', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19495097, end_position_on_the_genomic_accession=19497775, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246082, CurrentID=0, Status='live', Symbol='Defb15', Aliases='BD-15', description='defensin beta 15', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21929813, end_position_on_the_genomic_accession=21932710, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246083, CurrentID=0, Status='live', Symbol='Defb13', Aliases='BD-13', description='defensin beta 13', other_designations='', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21946762, end_position_on_the_genomic_accession=21948622, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246086, CurrentID=0, Status='live', Symbol='Onecut3', Aliases='OC-3, Oc3', description='one cut domain, family member 3', other_designations='one cut homeobox 3|transcription factor ONECUT-3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80494906, end_position_on_the_genomic_accession=80517266, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  246088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246088, CurrentID=0, Status='live', Symbol='Mtes1', Aliases='', description='Metastasis efficiency suppressor gene 1', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  246102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246102, CurrentID=0, Status='live', Symbol='Rttn', Aliases='4921538A15Rik, AI666264, C530033I08Rik, D230040K24', description='rotatin', other_designations='', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=88971738, end_position_on_the_genomic_accession=89131016, orientation='plus', exon_count=52, OMIM='', no_hdr0=''),
  246103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246103, CurrentID=0, Status='live', Symbol='Atxn7', Aliases='A430107N12Rik, AI627028, Sca7, ataxin-7', description='ataxin 7', other_designations='spinocerebellar ataxia 7 homolog|spinocerebellar ataxia type 7 protein homolog', map_location='14 A1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=13961439, end_position_on_the_genomic_accession=14107302, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  246104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246104, CurrentID=0, Status='live', Symbol='Rhbdl3', Aliases='AI847581, Rhbdl4, Rhomr1, Vrho', description='rhomboid, veinlet-like 3 (Drosophila)', other_designations='rhomboid protease 3|rhomboid related gene 1|rhomboid, veinlet-like 4|ventrhoid transmembrane protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=80300432, end_position_on_the_genomic_accession=80355986, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  246108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246108, CurrentID=0, Status='live', Symbol='Lifespan2', Aliases='', description='life span 2', other_designations='', map_location='10 34.83 cM', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  106557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106557, CurrentID=0, Status='live', Symbol='Ldhal6b', Aliases='4933402O15Rik, AI326310, LDHL, Ldhal', description='lactate dehydrogenase A-like 6B', other_designations='', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=5417323, end_position_on_the_genomic_accession=5418767, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246133, CurrentID=0, Status='live', Symbol='Kcne2', Aliases='2200002I16Rik, AW048273, MiRP1', description='potassium voltage-gated channel, Isk-related subfamily, gene 2', other_designations='minK-related peptide 1|minimum potassium ion channel related peptide 1|minimum potassium ion channel-related peptide 1|potassium channel subunit beta MiRP1', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92292389, end_position_on_the_genomic_accession=92298137, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246154, CurrentID=0, Status='live', Symbol='Vasn', Aliases='2610528G05Rik, Atia, Slitl2', description='vasorin', other_designations='Slit-like 2|protein slit-like 2', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=4639945, end_position_on_the_genomic_accession=4651166, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  106564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=106564, CurrentID=0, Status='live', Symbol='Ppcs', Aliases='6330579B17Rik, AI413471, AI790434', description='phosphopantothenoylcysteine synthetase', other_designations='PPC synthetase', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=119418527, end_position_on_the_genomic_accession=119422420, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  639390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=639390, CurrentID=0, Status='live', Symbol='BC094435', Aliases='ENSMUSG00000073624', description='cDNA sequence BC094435', other_designations='gag protein', map_location='18', chromosome='18', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  246177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246177, CurrentID=0, Status='live', Symbol='Myo1g', Aliases='E430002D17Rik', description='myosin IG', other_designations='myosin 1G|myosin-Ig', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=6506548, end_position_on_the_genomic_accession=6520960, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  246179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246179, CurrentID=0, Status='live', Symbol='Fktn', Aliases='D830030O17Rik, Fcmd', description='fukutin', other_designations='Fukuyama type congenital muscular dystrophy (fukutin)|Fukuyama type congenital muscular dystrophy homolog', map_location='4 28.74 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=53714107, end_position_on_the_genomic_accession=53765785, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  246190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246190, CurrentID=0, Status='live', Symbol='Otoa', Aliases='', description='otoancorin', other_designations='', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=121083438, end_position_on_the_genomic_accession=121163470, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  246196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246196, CurrentID=0, Status='live', Symbol='Zfp277', Aliases='2410017E24Rik, NIRF4', description='zinc finger protein 277', other_designations='', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=40315046, end_position_on_the_genomic_accession=40445790, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  246198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246198, CurrentID=0, Status='live', Symbol='Mllt6', Aliases='AI315037, Af17', description='myeloid/lymphoid or mixed-lineage leukemia (trithorax homolog, Drosophila); translocated to, 6', other_designations='myeloid/lymphoid or mixed lineage-leukemia translocation to 6 homolog', map_location='11 61.05 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97663412, end_position_on_the_genomic_accession=97685461, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  246221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246221, CurrentID=0, Status='live', Symbol='Mpst', Aliases='Mst', description='mercaptopyruvate sulfurtransferase', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=78406712, end_position_on_the_genomic_accession=78414015, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  246228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246228, CurrentID=0, Status='live', Symbol='Vwa1', Aliases='4932416A11Rik, WARP', description='von Willebrand factor A domain containing 1', other_designations='von Willebrand factor A domain-related protein', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155768495, end_position_on_the_genomic_accession=155774561, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246229, CurrentID=0, Status='live', Symbol='Bivm', Aliases='AI854267', description='basic, immunoglobulin-like variable motif containing', other_designations='', map_location='1 C1.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=44118888, end_position_on_the_genomic_accession=44144771, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  246256 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246256, CurrentID=0, Status='live', Symbol='Fcgr4', Aliases='4833442P21Rik, CD16-2, FcgRIV, FcgammaRIV, Fcgr3a, Fcrl3', description='Fc receptor, IgG, low affinity IV', other_designations='Fc fragment of IgG, low affinity IIIa, receptor for (CD16); Fc fragment of IgG, low affinity III, receptor for (CD16)|Fc gammaRIV|Fc receptor-like 3', map_location='1 78.53 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171018926, end_position_on_the_genomic_accession=171029761, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  246257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246257, CurrentID=0, Status='live', Symbol='Ovca2', Aliases='9030414M07Rik', description='candidate tumor suppressor in ovarian cancer 2', other_designations='candidate tumor suppressor OVCA2|ortholog of human candidate tumor suppressor in ovarian cancer 2 OVCA2', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75175943, end_position_on_the_genomic_accession=75178808, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246277, CurrentID=0, Status='live', Symbol='Csad', Aliases='Csd', description='cysteine sulfinic acid decarboxylase', other_designations='cysteine-sulfinate decarboxylase|sulfinoalanine decarboxylase', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102176997, end_position_on_the_genomic_accession=102204231, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  246278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246278, CurrentID=0, Status='live', Symbol='Cd207', Aliases='', description='CD207 antigen', other_designations='Langerhans cell specific c-type lectin|langerin', map_location='6 D1-D2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83671207, end_position_on_the_genomic_accession=83677857, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246293, CurrentID=0, Status='live', Symbol='Klhl8', Aliases='2310001P09Rik, D5Ertd431e', description='kelch-like 8', other_designations='', map_location='5 50.45 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=103862050, end_position_on_the_genomic_accession=103911273, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  246313 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246313, CurrentID=0, Status='live', Symbol='Prokr2', Aliases='B830005M06Rik, EG-VEGRF2, Gpcr73l1, Gpr73l1, PKR2', description='prokineticin receptor 2', other_designations='G protein-coupled receptor 73-like 1|G-protein coupled receptor 73-like 1|PK-R2', map_location='2 F2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=132370329, end_position_on_the_genomic_accession=132385498, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246316, CurrentID=0, Status='live', Symbol='Lgi2', Aliases='mKIAA1916', description='leucine-rich repeat LGI family, member 2', other_designations='leucine-rich glioma-inactivated protein 2', map_location='5 C1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=52533517, end_position_on_the_genomic_accession=52566462, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  246317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246317, CurrentID=0, Status='live', Symbol='Neto1', Aliases='AI851453, Btcl1, C130005O10Rik', description='neuropilin (NRP) and tolloid (TLL)-like 1', other_designations='brain-specific transmembrane protein containing 2 CUB and 1 LDL-receptor class A domains protein 1', map_location='18 E4', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=86394952, end_position_on_the_genomic_accession=86506721, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  639769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=639769, CurrentID=0, Status='live', Symbol='Adh6-ps1', Aliases='1300002P07Rik, Adh5ps, Gm7277', description='alcohol dehydrogenase 6 (class V), pseudogene 1', other_designations='', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=138374121, end_position_on_the_genomic_accession=138388291, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  639774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=639774, CurrentID=0, Status='live', Symbol='Skint8', Aliases='OTTMUSG00000009475', description='selection and upkeep of intraepithelial T cells 8', other_designations='skint-8', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=111919446, end_position_on_the_genomic_accession=111950361, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  639781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=639781, CurrentID=0, Status='live', Symbol='Skint1', Aliases='OTTMUSG00000008542, Skint-1', description='selection and upkeep of intraepithelial T cells 1', other_designations='Immunoglobulin-like and transmembrane domain-containing protein expressed in skin and thymus protein 1|skint 1', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=112006269, end_position_on_the_genomic_accession=112029538, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  100123473 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100123473, CurrentID=0, Status='live', Symbol='Trdc', Aliases='Tcrd-C', description='T cell receptor delta, constant region', other_designations='T-cell receptor delta, constant region', map_location='14 27.66 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54142851, end_position_on_the_genomic_accession=54146108, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  237711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237711, CurrentID=0, Status='live', Symbol='Eml6', Aliases='2900083P10Rik, C230094A16Rik, EMAP-6', description='echinoderm microtubule associated protein like 6', other_designations='', map_location='11 A3.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=29742586, end_position_on_the_genomic_accession=30026034, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  246691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246691, CurrentID=0, Status='live', Symbol='Prok1', Aliases='EG-VEGF', description='prokineticin 1', other_designations='EG-VEGF|endocrine-gland-derived vascular endothelial growth factor|prok1', map_location='3 F2.3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=107231916, end_position_on_the_genomic_accession=107239707, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246694, CurrentID=0, Status='live', Symbol='Hps5', Aliases='AI646796, AL022647, C85120, haze, hz, maroon, mr, ru-2, ru2', description='Hermansky-Pudlak syndrome 5', other_designations='Hermansky-Pudlak syndrome 5 homolog|ruby eye 2|ruby-eye protein 2', map_location='7 30.56 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=46760466, end_position_on_the_genomic_accession=46796064, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  246696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246696, CurrentID=0, Status='live', Symbol='Slc25a28', Aliases='2210403D18Rik, Mfrn2, Mrs3/4', description='solute carrier family 25, member 28', other_designations='mitochondrial RNA-splicing protein 3/4 homolog|mitochondrial iron transporter 2|mitoferrin 2', map_location='19 36.67 cM', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=43663801, end_position_on_the_genomic_accession=43677099, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246700, CurrentID=0, Status='live', Symbol='Defb19', Aliases='BD-19, Defb24, Tdl, mBD-19', description='defensin beta 19', other_designations='beta-defensin 24|testis-specific beta-defensin-like protein', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152576086, end_position_on_the_genomic_accession=152580312, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246703, CurrentID=0, Status='live', Symbol='Apoa1bp', Aliases='AA087124, AI-BP, AIBP, Apoa1ip, ESTM37', description='apolipoprotein A-I binding protein', other_designations='AI-BP|NAD(P)HX epimerase|apoA-I binding protein|apolipoprotein A-I interacting protein|apolipoprotein A-I-binding protein', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88056523, end_position_on_the_genomic_accession=88058541, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246707 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246707, CurrentID=0, Status='live', Symbol='Emilin2', Aliases='FOAP-10', description='elastin microfibril interfacer 2', other_designations='basilin|elastin microfibril interface located protein 2|elastin microfibril interface-located protein 2', map_location='17 E1.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=71252172, end_position_on_the_genomic_accession=71311639, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  246709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246709, CurrentID=0, Status='live', Symbol='Rgs13', Aliases='', description='regulator of G-protein signaling 13', other_designations='', map_location='1 62.56 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=144137386, end_position_on_the_genomic_accession=144177372, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  246710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246710, CurrentID=0, Status='live', Symbol='Rhobtb2', Aliases='Dbc2, E130206H14Rik, mKIAA0717', description='Rho-related BTB domain containing 2', other_designations='deleted in breast cancer 2 gene protein homolog', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=69784990, end_position_on_the_genomic_accession=69824847, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  246722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246722, CurrentID=0, Status='live', Symbol='Gy', Aliases='', description='gyro deletion region', other_designations='', map_location='', chromosome='X', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  246727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246727, CurrentID=0, Status='live', Symbol='Oas3', Aliases='Oasl10', description="2'-5' oligoadenylate synthetase 3", other_designations="(2-5')oligo(A) synthase 3|2',5'-oligoadenylate synthetase-like 10|2'-5' oligoadenylate synthetase-like 10|2-5A synthase 3", map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120753098, end_position_on_the_genomic_accession=120777659, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  246728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246728, CurrentID=0, Status='live', Symbol='Oas2', Aliases='Oasl11', description="2'-5' oligoadenylate synthetase 2", other_designations="(2-5')oligo(A) synthase 2|2',5'-oligoadenylate synthetase-like 11|2'-5' oligoadenylate synthetase-like 11|2,5-oligoadenylate synthetase-like 11|2-5A synthase 2", map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120730333, end_position_on_the_genomic_accession=120749884, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  246729 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246729, CurrentID=0, Status='live', Symbol='Oas1h', Aliases='', description="2'-5' oligoadenylate synthetase 1H", other_designations='', map_location='5 60.64 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120861422, end_position_on_the_genomic_accession=120873506, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  246730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246730, CurrentID=0, Status='live', Symbol='Oas1a', Aliases='L3', description="2'-5' oligoadenylate synthetase 1A", other_designations="(2-5')oligo(A) synthase 1A|(2-5')oligo(A) synthetase 1A|2-5A synthase 1A|2-5A synthetase 1A|p42 OAS", map_location='5 60.65 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120896256, end_position_on_the_genomic_accession=120907543, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  246735 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246735, CurrentID=0, Status='live', Symbol='AY074887', Aliases='Mcpr1', description='cDNA sequence AY074887', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=54950244, end_position_on_the_genomic_accession=54950954, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  246738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246738, CurrentID=0, Status='live', Symbol='Dnajc28', Aliases='BC020175, ORF28', description='DnaJ heat shock protein family (Hsp40) member C28', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 28', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=91614257, end_position_on_the_genomic_accession=91618999, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  246746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246746, CurrentID=0, Status='live', Symbol='Cd300lf', Aliases='CLIM1, CLM-1, CLM1, Digr2, F730004D16Rik, IREM1, IgSF13, LMIR3, Pigr3', description='CD300 antigen like family member F', other_designations='CD300 antigen-like family member F|CMRF-35-like molecule 1|DC-derived Ig-like receptor 2|MAIR-5|MAIR-V|leukocyte mono-Ig-like receptor 3|myeloid-associated immunoglobulin-like receptor 5', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115116214, end_position_on_the_genomic_accession=115134150, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  246747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246747, CurrentID=0, Status='live', Symbol='Adig', Aliases='SMAF-1, Smaf1', description='adipogenin', other_designations='small adipocyte factor 1', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=158502612, end_position_on_the_genomic_accession=158508198, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  246779 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246779, CurrentID=0, Status='live', Symbol='Il27', Aliases='IL-27, IL-27p28, Il30, p28', description='interleukin 27', other_designations='IL-27 p28 subunit|IL-27 subunit alpha|IL-27-A|IL27-A|interleukin 30', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126589009, end_position_on_the_genomic_accession=126595021, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  246782 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246782, CurrentID=0, Status='live', Symbol='Atpaf2', Aliases='ATP12, ATP12p', description='ATP synthase mitochondrial F1 complex assembly factor 2', other_designations='', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=60400624, end_position_on_the_genomic_accession=60418791, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  246787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246787, CurrentID=0, Status='live', Symbol='Slc5a2', Aliases='Sglt2', description='solute carrier family 5 (sodium/glucose cotransporter), member 2', other_designations='Na(+)/glucose cotransporter 2|low affinity sodium-dependent glucose cotransporter|low affinity sodium-glucose cotransporter|solute carrier family 5 member 2', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128265697, end_position_on_the_genomic_accession=128272433, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  246788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246788, CurrentID=0, Status='live', Symbol='Trpv3', Aliases='1110036I10Rik, AI644701, Nh, VRL3', description='transient receptor potential cation channel, subfamily V, member 3', other_designations='TRP ion channel TRPV3|non-hair', map_location='11 45.25 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73267489, end_position_on_the_genomic_accession=73300368, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  246791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246791, CurrentID=0, Status='live', Symbol='Obox3', Aliases='Ohx', description='oocyte specific homeobox 3', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15625306, end_position_on_the_genomic_accession=15639777, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  246792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=246792, CurrentID=0, Status='live', Symbol='Obox2', Aliases='', description='oocyte specific homeobox 2', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15388851, end_position_on_the_genomic_accession=15398545, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  666794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=666794, CurrentID=0, Status='live', Symbol='Rbm24', Aliases='6330546B05Rik, AI606861', description='RNA binding motif protein 24', other_designations='RNA-binding motif protein 24', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=46418180, end_position_on_the_genomic_accession=46431099, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  230822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=230822, CurrentID=0, Status='live', Symbol='Ncmap', Aliases='A330049M08Rik, AA545175, Mp11', description='noncompact myelin associated protein', other_designations='myelin protein of 11 kDa', map_location='4 D3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=135369577, end_position_on_the_genomic_accession=135398227, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  19369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19369, CurrentID=0, Status='live', Symbol='Raet1b', Aliases='RAE-1beta', description='retinoic acid early transcript beta', other_designations='RAE-1-beta|Rae-1 beta', map_location='', chromosome='10', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  640524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=640524, CurrentID=0, Status='live', Symbol='Sptbn5', Aliases='EG640524, Gm354, Spnb5', description='spectrin beta, non-erythrocytic 5', other_designations='spectrin beta 4|spectrin beta 5|spectrin beta non-erythrocytic 5 (SPTBN5) orthologue', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=120044322, end_position_on_the_genomic_accession=120089175, orientation='minus', exon_count=66, OMIM='', no_hdr0=''),
  640543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=640543, CurrentID=0, Status='live', Symbol='Tgm7', Aliases='TGMZ, TGz', description='transglutaminase 7', other_designations='', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121093498, end_position_on_the_genomic_accession=121127516, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  237831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=237831, CurrentID=0, Status='live', Symbol='Slc13a5', Aliases='Indy, NaC2/NaCT, Nact, mINDY', description='solute carrier family 13 (sodium-dependent citrate transporter), member 5', other_designations='Na(+)/citrate cotransporter|novel solute carrier family 13 (sodium-dependent dicarboxylate transporter) (Slc13a2 or 3) member|sodium-coupled citrate transporter|sodium-dependent citrate transporter', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=72241990, end_position_on_the_genomic_accession=72267032, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  378425 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378425, CurrentID=0, Status='live', Symbol='Nlrp12', Aliases='Nalp12, PYPAF7', description='NLR family, pyrin domain containing 12', other_designations='NACHT, LRR and PYD containing protein 12|PYRIN-containing APAF1-like protein 7|monarch-1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=3218643, end_position_on_the_genomic_accession=3250159, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  378430 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378430, CurrentID=0, Status='live', Symbol='Nanos2', Aliases='nos2', description='nanos homolog 2 (Drosophila)', other_designations='NOS-2', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18987524, end_position_on_the_genomic_accession=18988962, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  378431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378431, CurrentID=0, Status='live', Symbol='Txlnb', Aliases='2310001N14Rik, Mdp77', description='taxilin beta', other_designations='muscle specific protein|muscle-derived protein 77', map_location='10 A2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=17796091, end_position_on_the_genomic_accession=17845665, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  378435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378435, CurrentID=0, Status='live', Symbol='Mafa', Aliases='RIPE3b1', description='v-maf musculoaponeurotic fibrosarcoma oncogene family, protein A (avian)', other_designations='Mafa homolog|V-maf musculoaponeurotic fibrosarcoma oncogene homolog A|pancreatic beta-cell-specific transcriptional activator', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=75746843, end_position_on_the_genomic_accession=75747922, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  378460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378460, CurrentID=0, Status='live', Symbol='Pram1', Aliases='AY665714', description='PML-RAR alpha-regulated adaptor molecule 1', other_designations='PRAM-1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33638056, end_position_on_the_genomic_accession=33645706, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  378462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378462, CurrentID=0, Status='live', Symbol='Morn2', Aliases='Mopt', description='MORN repeat containing 2', other_designations='MORN motif protein in testis|protein containing single MORN motif in testis', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80290212, end_position_on_the_genomic_accession=80305691, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  100124287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124287, CurrentID=0, Status='live', Symbol='Traj61', Aliases='Gm17363', description='T cell receptor alpha joining 61', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54155005, end_position_on_the_genomic_accession=54155076, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124288 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124288, CurrentID=0, Status='live', Symbol='Traj60', Aliases='Gm17534', description='T cell receptor alpha joining 60', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54155933, end_position_on_the_genomic_accession=54155985, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124293, CurrentID=0, Status='live', Symbol='Traj59', Aliases='Gm16770', description='T cell receptor alpha joining 59', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54156133, end_position_on_the_genomic_accession=54156194, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124294, CurrentID=0, Status='live', Symbol='Traj58', Aliases='Gm16771', description='T cell receptor alpha joining 58', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54157280, end_position_on_the_genomic_accession=54157342, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124295, CurrentID=0, Status='live', Symbol='Traj57', Aliases='Gm16768', description='T cell receptor alpha joining 57', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54158507, end_position_on_the_genomic_accession=54158569, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124296, CurrentID=0, Status='live', Symbol='Traj56', Aliases='Gm16769', description='T cell receptor alpha joining 56', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54159263, end_position_on_the_genomic_accession=54159325, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124297, CurrentID=0, Status='live', Symbol='Traj55', Aliases='Gm17471', description='T cell receptor alpha joining 55', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54161439, end_position_on_the_genomic_accession=54161501, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124333, CurrentID=0, Status='live', Symbol='Traj54', Aliases='Gm17479', description='T cell receptor alpha joining 54', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54161990, end_position_on_the_genomic_accession=54162043, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124335, CurrentID=0, Status='live', Symbol='Traj53', Aliases='Gm16772', description='T cell receptor alpha joining 53', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54162644, end_position_on_the_genomic_accession=54162709, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124336, CurrentID=0, Status='live', Symbol='Traj52', Aliases='Gm16773', description='T cell receptor alpha joining 52', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54165316, end_position_on_the_genomic_accession=54165381, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124337, CurrentID=0, Status='live', Symbol='Traj51', Aliases='Gm16924', description='T cell receptor alpha joining 51', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54166239, end_position_on_the_genomic_accession=54166281, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124338, CurrentID=0, Status='live', Symbol='Traj50', Aliases='Gm17007', description='T cell receptor alpha joining 50', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54167590, end_position_on_the_genomic_accession=54167652, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124339, CurrentID=0, Status='live', Symbol='Traj49', Aliases='Gm17012', description='T cell receptor alpha joining 49', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54168686, end_position_on_the_genomic_accession=54168744, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124340, CurrentID=0, Status='live', Symbol='Traj48', Aliases='Gm16923', description='T cell receptor alpha joining 48', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54169808, end_position_on_the_genomic_accession=54169868, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124341, CurrentID=0, Status='live', Symbol='Traj47', Aliases='', description='T cell receptor alpha joining 47', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54171802, end_position_on_the_genomic_accession=54171856, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124342, CurrentID=0, Status='live', Symbol='Traj46', Aliases='Gm16928', description='T cell receptor alpha joining 46', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54172338, end_position_on_the_genomic_accession=54172398, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124343, CurrentID=0, Status='live', Symbol='Traj45', Aliases='Gm16927', description='T cell receptor alpha joining 45', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54172831, end_position_on_the_genomic_accession=54172890, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124344, CurrentID=0, Status='live', Symbol='Traj44', Aliases='Gm16926', description='T cell receptor alpha joining 44', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54173690, end_position_on_the_genomic_accession=54173750, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124345, CurrentID=0, Status='live', Symbol='Traj43', Aliases='Gm16925', description='T cell receptor alpha joining 43', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54174742, end_position_on_the_genomic_accession=54174798, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124346, CurrentID=0, Status='live', Symbol='Traj42', Aliases='Gm16922', description='T cell receptor alpha joining 42', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54175773, end_position_on_the_genomic_accession=54175836, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124347, CurrentID=0, Status='live', Symbol='Traj41', Aliases='Gm16921', description='T cell receptor alpha joining 41', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54176271, end_position_on_the_genomic_accession=54176325, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124348, CurrentID=0, Status='live', Symbol='Traj40', Aliases='Gm16668', description='T cell receptor alpha joining 40', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54177921, end_position_on_the_genomic_accession=54177981, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124349, CurrentID=0, Status='live', Symbol='Traj39', Aliases='Gm16669', description='T cell receptor alpha joining 39', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54179962, end_position_on_the_genomic_accession=54180024, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124351, CurrentID=0, Status='live', Symbol='Traj38', Aliases='Gm16670', description='T cell receptor alpha joining 38', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54180574, end_position_on_the_genomic_accession=54180635, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124352, CurrentID=0, Status='live', Symbol='Traj37', Aliases='Gm16671', description='T cell receptor alpha joining 37', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54181518, end_position_on_the_genomic_accession=54181577, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124353, CurrentID=0, Status='live', Symbol='Traj36', Aliases='Gm17610', description='T cell receptor alpha joining 36', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54182402, end_position_on_the_genomic_accession=54182464, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124354, CurrentID=0, Status='live', Symbol='Traj35', Aliases='Gm16664', description='T cell receptor alpha joining 35', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54183774, end_position_on_the_genomic_accession=54183838, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124355, CurrentID=0, Status='live', Symbol='Traj34', Aliases='Gm16665', description='T cell receptor alpha joining 34', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54184699, end_position_on_the_genomic_accession=54184756, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124356, CurrentID=0, Status='live', Symbol='Traj33', Aliases='Gm16666', description='T cell receptor alpha joining 33', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54185358, end_position_on_the_genomic_accession=54185414, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124357, CurrentID=0, Status='live', Symbol='Traj32', Aliases='Gm16667', description='T cell receptor alpha joining 32', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54186101, end_position_on_the_genomic_accession=54186166, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124358, CurrentID=0, Status='live', Symbol='Traj31', Aliases='Gm17008', description='T cell receptor alpha joining 31', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54187895, end_position_on_the_genomic_accession=54187951, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124359, CurrentID=0, Status='live', Symbol='Traj30', Aliases='Gm16672', description='T cell receptor alpha joining 30', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54189866, end_position_on_the_genomic_accession=54189924, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124360, CurrentID=0, Status='live', Symbol='Traj29', Aliases='Gm16819', description='T cell receptor alpha joining 29', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54190946, end_position_on_the_genomic_accession=54191005, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124361, CurrentID=0, Status='live', Symbol='Traj28', Aliases='Gm16820', description='T cell receptor alpha joining 28', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54191661, end_position_on_the_genomic_accession=54191725, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124362, CurrentID=0, Status='live', Symbol='Traj27', Aliases='Gm16817', description='T cell receptor alpha joining 27', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54192303, end_position_on_the_genomic_accession=54192361, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124363, CurrentID=0, Status='live', Symbol='Traj26', Aliases='Gm16818', description='T cell receptor alpha joining 26', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54194490, end_position_on_the_genomic_accession=54194549, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124364, CurrentID=0, Status='live', Symbol='Traj25', Aliases='Gm16823', description='T cell receptor alpha joining 25', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54194815, end_position_on_the_genomic_accession=54194871, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124365, CurrentID=0, Status='live', Symbol='Traj24', Aliases='Gm16824', description='T cell receptor alpha joining 24', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54195645, end_position_on_the_genomic_accession=54195700, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124366, CurrentID=0, Status='live', Symbol='Traj23', Aliases='Gm16821', description='T cell receptor alpha joining 23', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54196081, end_position_on_the_genomic_accession=54196140, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124367, CurrentID=0, Status='live', Symbol='Traj22', Aliases='Gm16822', description='T cell receptor alpha joining 22', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54197248, end_position_on_the_genomic_accession=54197307, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124368, CurrentID=0, Status='live', Symbol='Traj21', Aliases='Gm16815', description='T cell receptor alpha joining 21', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54198790, end_position_on_the_genomic_accession=54198846, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124369, CurrentID=0, Status='live', Symbol='Traj20', Aliases='Gm16816', description='T cell receptor alpha joining 20', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54199441, end_position_on_the_genomic_accession=54199498, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124370, CurrentID=0, Status='live', Symbol='Traj19', Aliases='Gm16655', description='T cell receptor alpha joining 19', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54200386, end_position_on_the_genomic_accession=54200446, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124371, CurrentID=0, Status='live', Symbol='Traj18', Aliases='Gm17009', description='T cell receptor alpha joining 18', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54200777, end_position_on_the_genomic_accession=54200842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124372 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124372, CurrentID=0, Status='live', Symbol='Traj17', Aliases='Gm17010', description='T cell receptor alpha joining 17', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54201775, end_position_on_the_genomic_accession=54201837, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116450, CurrentID=0, Status='live', Symbol='Etia', Aliases='', description='ethanol induced activation', other_designations='', map_location='16 45.89 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100124387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124387, CurrentID=0, Status='live', Symbol='Traj16', Aliases='Gm16656', description='T cell receptor alpha joining 16', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54203134, end_position_on_the_genomic_accession=54203194, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124388, CurrentID=0, Status='live', Symbol='Traj15', Aliases='Gm16658', description='T cell receptor alpha joining 15', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54204422, end_position_on_the_genomic_accession=54204481, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124389, CurrentID=0, Status='live', Symbol='Traj14', Aliases='Gm16657', description='T cell receptor alpha joining 14', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54205111, end_position_on_the_genomic_accession=54205144, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124390, CurrentID=0, Status='live', Symbol='Traj13', Aliases='Gm16660', description='T cell receptor alpha joining 13', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54205741, end_position_on_the_genomic_accession=54205797, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124391, CurrentID=0, Status='live', Symbol='Traj12', Aliases='Gm16659', description='T cell receptor alpha joining 12', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54206552, end_position_on_the_genomic_accession=54206610, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124392, CurrentID=0, Status='live', Symbol='Traj11', Aliases='Gm16654', description='T cell receptor alpha joining 11', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54207139, end_position_on_the_genomic_accession=54207197, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124393, CurrentID=0, Status='live', Symbol='Traj9', Aliases='Gm16653', description='T cell receptor alpha joining 9', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54209393, end_position_on_the_genomic_accession=54209450, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124395, CurrentID=0, Status='live', Symbol='Traj8', Aliases='Gm16918', description='T cell receptor alpha joining 8', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54209918, end_position_on_the_genomic_accession=54209954, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124396, CurrentID=0, Status='live', Symbol='Traj7', Aliases='Gm16919', description='T cell receptor alpha joining 7', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54211470, end_position_on_the_genomic_accession=54211528, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124397, CurrentID=0, Status='live', Symbol='Traj6', Aliases='Gm16920', description='T cell receptor alpha joining 6', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54212688, end_position_on_the_genomic_accession=54212749, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124398, CurrentID=0, Status='live', Symbol='Traj4', Aliases='Gm16915', description='T cell receptor alpha joining 4', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54216291, end_position_on_the_genomic_accession=54216353, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124399, CurrentID=0, Status='live', Symbol='Traj3', Aliases='Gm16916', description='T cell receptor alpha joining 3', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54217292, end_position_on_the_genomic_accession=54217357, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124400 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124400, CurrentID=0, Status='live', Symbol='Traj1', Aliases='Gm16917', description='T cell receptor alpha joining 1', other_designations='', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54218814, end_position_on_the_genomic_accession=54218842, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124432 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124432, CurrentID=0, Status='live', Symbol='Mir193b', Aliases='Mir193, Mirn193b, mir-193b', description='microRNA 193b', other_designations='microRNA mir-193b|mmu-mir-193b', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=13449523, end_position_on_the_genomic_accession=13449601, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124433, CurrentID=0, Status='live', Symbol='Mir208b', Aliases='Mirn208b, mir-208b', description='microRNA 208b', other_designations='microRNA mir-208b|mmu-mir-208b', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54975700, end_position_on_the_genomic_accession=54975776, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100124446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124446, CurrentID=0, Status='live', Symbol='Mir466h', Aliases='Mirn466h', description='microRNA 466h', other_designations='microRNA Mirn466h|microRNA mir-466h|mmu-mir-466h', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10514891, end_position_on_the_genomic_accession=10514971, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124451, CurrentID=0, Status='live', Symbol='Mir574', Aliases='Mirn574, mir-574, mmu-mir-574', description='microRNA 574', other_designations='microRNA Mirn574|microRNA mir-574', map_location='5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=64970318, end_position_on_the_genomic_accession=64970395, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124457, CurrentID=0, Status='live', Symbol='Mir873a', Aliases='Mir873, Mirn873, mir-873a, mmu-mir-873, mmu-mir-873a', description='microRNA 873a', other_designations='microRNA 873|microRNA Mirn873|microRNA mir-873', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=36668510, end_position_on_the_genomic_accession=36668586, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  116527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116527, CurrentID=0, Status='live', Symbol='Obq14', Aliases='', description='obesity QTL 14', other_designations='', map_location='6 45.74 cM', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100124465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124465, CurrentID=0, Status='live', Symbol='Mir466d', Aliases='Mirn466d', description='microRNA 466d', other_designations='microRNA Mirn466d|microRNA mir-466d|mmu-mir-466d', map_location='', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=10511967, end_position_on_the_genomic_accession=10512062, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124466, CurrentID=0, Status='live', Symbol='Mir493', Aliases='Mirn493, mmu-mir-493', description='microRNA 493', other_designations='microRNA Mirn493|microRNA mir-493', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109580233, end_position_on_the_genomic_accession=109580315, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124470, CurrentID=0, Status='live', Symbol='Mir92b', Aliases='Mirn92b, mir-92b, mmu-mir-92b', description='microRNA 92b', other_designations='microRNA Mirn92b|microRNA mir-92b', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=89227116, end_position_on_the_genomic_accession=89227198, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  100124477 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124477, CurrentID=0, Status='live', Symbol='Mir590', Aliases='Mirn590, mmu-mir-590', description='microRNA 590', other_designations='microRNA Mirn590|microRNA mir-590', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100124488 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124488, CurrentID=0, Status='live', Symbol='Mir511', Aliases='Mirn511, mir-511, mmu-mir-511', description='microRNA 511', other_designations='microRNA Mirn511|microRNA mir-511', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=14261003, end_position_on_the_genomic_accession=14261081, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  100124490 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124490, CurrentID=0, Status='live', Symbol='Mir712', Aliases='Mirn712, mmu-mir-712', description='microRNA 712', other_designations='microRNA Mirn712|microRNA mir-712', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100124491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124491, CurrentID=0, Status='live', Symbol='Mir874', Aliases='Mirn874, mir-874, mmu-mir-874', description='microRNA 874', other_designations='microRNA Mirn874|microRNA mir-874', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=58023125, end_position_on_the_genomic_accession=58023200, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  378700 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378700, CurrentID=0, Status='live', Symbol='Bpifb3', Aliases='Lplunc3, Rya3', description='BPI fold containing family B, member 3', other_designations='antimicrobial peptide RYA3|ligand-binding protein RYA3|long palate, lung and nasal epithelium carcinoma-associated protein 3', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153918230, end_position_on_the_genomic_accession=153932996, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  100124493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124493, CurrentID=0, Status='live', Symbol='Mir883b', Aliases='Mirn883b, mir-883b', description='microRNA 883b', other_designations='microRNA Mirn883b|microRNA mir-883b|mmu-mir-883b', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=66789890, end_position_on_the_genomic_accession=66789967, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  378702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378702, CurrentID=0, Status='live', Symbol='Serf2', Aliases='C80083, Msmac1l, m4F5rel', description='small EDRK-rich factor 2', other_designations='4F5rel (4F5 related)|protein 4F5-related', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121449198, end_position_on_the_genomic_accession=121453420, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  19447 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19447, CurrentID=0, Status='live', Symbol='Rb(1.3)1Bnr', Aliases='Rb1Bnr', description='Robertsonian translocation, Chr 1 and 3, Universitat Bonn/Rhein 1', other_designations='', map_location='', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116701, CurrentID=0, Status='live', Symbol='Fgfrl1', Aliases='FGFR5, FGFR5beta, FGFR5gamma', description='fibroblast growth factor receptor-like 1', other_designations='FGF receptor-like protein 1|fibroblast growth factor receptor 5', map_location='5 E3-F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=108694229, end_position_on_the_genomic_accession=108706950, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  116703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116703, CurrentID=0, Status='live', Symbol='Listr2', Aliases='', description='listeriosis resistance 2', other_designations='', map_location='13 43.68 cM', chromosome='13', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100124659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124659, CurrentID=0, Status='live', Symbol='Trbv2', Aliases='Gm16752, Tcrb-V4', description='T cell receptor beta, variable 2', other_designations='T-cell receptor beta, variable 2|T-cell receptor beta, variable 4', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41047556, end_position_on_the_genomic_accession=41047995, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124660, CurrentID=0, Status='live', Symbol='Trbv4', Aliases='Tcrb-V10', description='T cell receptor beta, variable 10', other_designations='T cell receptor beta variable 4|T-cell receptor beta, variable 10', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41059443, end_position_on_the_genomic_accession=41059886, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124661 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124661, CurrentID=0, Status='live', Symbol='Trbv5', Aliases='Tcrb-V1', description='T cell receptor beta, variable 5', other_designations='T-cell receptor beta, variable 1|T-cell receptor beta, variable 5', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41062359, end_position_on_the_genomic_accession=41062803, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116731 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116731, CurrentID=0, Status='live', Symbol='Pcdha1', Aliases='CNR', description='protocadherin alpha 1', other_designations='protocadherin alpha 2', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=36930285, end_position_on_the_genomic_accession=37187657, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  116732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116732, CurrentID=0, Status='live', Symbol='Tsga13', Aliases='1700023D02Rik, Vme1', description='testis specific gene A13', other_designations='variegated monoallelic expression 1', map_location='6 A3.3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30896980, end_position_on_the_genomic_accession=30925154, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  116733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116733, CurrentID=0, Status='live', Symbol='Vps4a', Aliases='4930589C15Rik, AI325971, AW553189', description='vacuolar protein sorting 4a (yeast)', other_designations='vacuolar protein sorting protein 4a|vps4a', map_location='8 D', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107031326, end_position_on_the_genomic_accession=107045756, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  100124673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124673, CurrentID=0, Status='live', Symbol='Trbv13-1', Aliases='Tcrb, Tcrb-V8.3', description='T cell receptor beta, variable 13-1', other_designations='T-cell receptor beta, variable 13-1|T-cell receptor beta, variable 8.3', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41116036, end_position_on_the_genomic_accession=41116468, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124674, CurrentID=0, Status='live', Symbol='Trbv12-2', Aliases='Tcrb, Tcrb-V5.1', description='T cell receptor beta, variable 12-2', other_designations='T-cell receptor beta, variable 12-2|T-cell receptor beta, variable 5.1', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41118864, end_position_on_the_genomic_accession=41119364, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124675 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124675, CurrentID=0, Status='live', Symbol='Trbv13-2', Aliases='TCRBV8S2, Tcrb, Tcrb-V8.2', description='T cell receptor beta, variable 13-2', other_designations='T-cell receptor beta, variable 13-2|T-cell receptor beta, variable 8.2', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41121396, end_position_on_the_genomic_accession=41121832, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124677, CurrentID=0, Status='live', Symbol='Trbv13-3', Aliases='Gm16777, Tcrb-V8, Tcrb-V8.1', description='T cell receptor beta, variable 13-3', other_designations='T cell receptor beta variable 13-3|T-cell receptor beta, variable 13-3|T-cell receptor beta, variable 8.1', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41130147, end_position_on_the_genomic_accession=41130585, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124678, CurrentID=0, Status='live', Symbol='Trbv14', Aliases='BV13S1, Gm16910, Tcrb-V13', description='T cell receptor beta, variable 14', other_designations='T-cell receptor beta, variable 14', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41135167, end_position_on_the_genomic_accession=41135616, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124679, CurrentID=0, Status='live', Symbol='Trbv15', Aliases='Gm16775, Tcrb-V12', description='T cell receptor beta, variable 15', other_designations='T-cell receptor beta, variable 12|T-cell receptor beta, variable 15', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41141211, end_position_on_the_genomic_accession=41141658, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124680, CurrentID=0, Status='live', Symbol='Trbv16', Aliases='Gm16776, Tcrb-V11', description='T cell receptor beta, variable 16', other_designations='T cell receptor beta variable 16|T-cell receptor beta chain V region|T-cell receptor beta, variable 11|T-cell receptor beta, variable 16', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41151791, end_position_on_the_genomic_accession=41152230, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124681 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124681, CurrentID=0, Status='live', Symbol='Trbv17', Aliases='Gm16779, Tcrb-V9', description='T cell receptor beta, variable 17', other_designations='T-cell receptor beta, variable 17|T-cell receptor beta, variable 9', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41163092, end_position_on_the_genomic_accession=41163556, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  116746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116746, CurrentID=0, Status='live', Symbol='Defb6', Aliases='BD-6', description='defensin beta 6', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19225478, end_position_on_the_genomic_accession=19228209, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  116747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116747, CurrentID=0, Status='live', Symbol='Rmcs2', Aliases='', description='response to metastatic cancers 2', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  100124684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124684, CurrentID=0, Status='live', Symbol='Trbv19', Aliases='Gm16780, Tcrb-V6', description='T cell receptor beta, variable 19', other_designations='T cell receptor beta variable 19|T-cell receptor beta, variable 19|T-cell receptor beta, variable 6', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41178589, end_position_on_the_genomic_accession=41179048, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124685 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124685, CurrentID=0, Status='live', Symbol='Trbv20', Aliases='Gm16911, Tcrb-V15', description='T cell receptor beta, variable 20', other_designations='T cell receptor beta variable 20|T-cell receptor beta, variable 15|T-cell receptor beta, variable 20', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41188273, end_position_on_the_genomic_accession=41188977, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124688, CurrentID=0, Status='live', Symbol='Trbv23', Aliases='Tcrb-V20, VB20', description='T cell receptor beta, variable V23', other_designations='T cell receptor beta variable 23|T-cell receptor beta, variable V20|T-cell receptor beta, variable V23', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41216073, end_position_on_the_genomic_accession=41216526, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124689 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124689, CurrentID=0, Status='live', Symbol='Trbv24', Aliases='Tcrb, Tcrb-V17, Tcrb-V3.2, VB17a(cz)', description='T cell receptor beta, variable 24', other_designations='T cell receptor beta variable 24|T-cell receptor beta, variable 17|T-cell receptor beta, variable 24|T-cell receptor beta, variable 3.2', map_location='6 19.32 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41218091, end_position_on_the_genomic_accession=41218557, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  100124698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100124698, CurrentID=0, Status='live', Symbol='Trbv29', Aliases='Gm16649, Tcrb, Tcrb-V7, Vbeta7', description='T cell receptor beta, variable 29', other_designations='T cell receptor beta chain|T cell receptor beta variable 29|T-cell receptor beta, variable 29|T-cell receptor beta, variable 7', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41271403, end_position_on_the_genomic_accession=41271881, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  378937 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=378937, CurrentID=0, Status='live', Symbol='Lrrc24', Aliases='6430402H13Rik', description='leucine rich repeat containing 24', other_designations='', map_location='15 D3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76715276, end_position_on_the_genomic_accession=76722173, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  116810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116810, CurrentID=0, Status='live', Symbol='Foxn4', Aliases='', description='forkhead box N4', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114254164, end_position_on_the_genomic_accession=114274598, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  116811 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116811, CurrentID=0, Status='live', Symbol='Zim3', Aliases='1700128I23Rik', description='zinc finger, imprinted 3', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6955685, end_position_on_the_genomic_accession=6976662, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  116837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116837, CurrentID=0, Status='live', Symbol='Rims1', Aliases='C030033M19Rik, RIM1, RIM1a, RIM1alpha, Rab3ip1, Rim, Serg1, mKIAA0340', description='regulating synaptic membrane exocytosis 1', other_designations='RIM 1|Rab3 effector|Rab3 interacting protein 1|rab-3-interacting molecule 1|rab-3-interacting protein 1|rab3-interacting molecule 1|rab3-interacting protein 1|regulating synaptic membrane exocytosis 1 beta|synaptic exocytosis regulator 1', map_location='1 A5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=22468191, end_position_on_the_genomic_accession=22805724, orientation='minus', exon_count=32, OMIM='', no_hdr0=''),
  116838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116838, CurrentID=0, Status='live', Symbol='Rims2', Aliases='2810036I15Rik, AW048769, RIM2, Rab3ip2, Rim2(+40A), Rim2(+44A), Rim2(+4A), Serg2, Syt3-rs, mKIAA0751', description='regulating synaptic membrane exocytosis 2', other_designations='RIM 2|Rab3 interacting protein 1|Rab3 interacting protein 2|rab-3-interacting protein 2|rab3-interacting protein 2|synaptic exocytosis regulator 2|synaptotagmin 3, related sequence', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=39198064, end_position_on_the_genomic_accession=39684375, orientation='plus', exon_count=43, OMIM='', no_hdr0=''),
  116847 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116847, CurrentID=0, Status='live', Symbol='Prelp', Aliases='7330409J17Rik, SLRR2A', description='proline arginine-rich end leucine-rich repeat', other_designations='proline-arginine-rich end leucine-rich repeat protein', map_location='1 58.02 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133910304, end_position_on_the_genomic_accession=133929976, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  116848 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116848, CurrentID=0, Status='live', Symbol='Baz2a', Aliases='AA415431, C030005G16Rik, C78388, Tip5, Walp3, mKIAA0314', description='bromodomain adjacent to zinc finger domain, 2A', other_designations='TTF-I-interacting protein 5|transcription termination factor I-interacting protein 5', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=128091056, end_position_on_the_genomic_accession=128129304, orientation='plus', exon_count=33, OMIM='', no_hdr0=''),
  116849 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116849, CurrentID=0, Status='live', Symbol='Iltifb', Aliases='IL-22b, IL-TIFb, Il22b', description='interleukin 10-related T cell-derived inducible factor beta', other_designations='IL-10-related T-cell-derived-inducible factor beta|IL-TIF beta', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=118289629, end_position_on_the_genomic_accession=118295038, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  116852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116852, CurrentID=0, Status='live', Symbol='Akr1c20', Aliases='2610528B18Rik', description='aldo-keto reductase family 1, member C20', other_designations='', map_location='13 A1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=4486972, end_position_on_the_genomic_accession=4523354, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  116870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116870, CurrentID=0, Status='live', Symbol='Mta1', Aliases='', description='metastasis associated 1', other_designations='', map_location='12 F', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113098278, end_position_on_the_genomic_accession=113137206, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  116871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116871, CurrentID=0, Status='live', Symbol='Mta3', Aliases='1110002J22Rik, mKIAA1266', description='metastasis associated 3', other_designations='', map_location='17 E4', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=83706163, end_position_on_the_genomic_accession=83823715, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  116872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116872, CurrentID=0, Status='live', Symbol='Serpinb7', Aliases='4631416M05Rik, megsin, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 7', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 7|serine (or cysteine) proteinase inhibitor, clade B, member 7', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107399486, end_position_on_the_genomic_accession=107452689, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  116873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116873, CurrentID=0, Status='live', Symbol='Stim2', Aliases='', description='stromal interaction molecule 2', other_designations='', map_location='5 29.69 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=53998445, end_position_on_the_genomic_accession=54121057, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  116891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116891, CurrentID=0, Status='live', Symbol='Derl2', Aliases='CGI-101, Derlin-2, F-lana, Flana', description='Der1-like domain family, member 2', other_designations='carcinoma related|degradation in endoplasmic reticulum protein 2|der1-like protein 2|derlin 2', map_location='11 43.21 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=71007440, end_position_on_the_genomic_accession=71019841, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  379043 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=379043, CurrentID=0, Status='live', Symbol='Raet1e', Aliases='', description='retinoic acid early transcript 1E', other_designations='RAE-1-epsilon|RAE1epsilon|Rae-1 epsilon', map_location='10 A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=22158532, end_position_on_the_genomic_accession=22191791, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  116903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116903, CurrentID=0, Status='live', Symbol='Calcb', Aliases='CGRP2, Calc2', description='calcitonin-related polypeptide, beta', other_designations='CGRP-II|beta-type CGRP|calcitonin gene-related peptide II', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=114716556, end_position_on_the_genomic_accession=114723368, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  116904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116904, CurrentID=0, Status='live', Symbol='Alpk3', Aliases='AW319487, D330016D04, MAK, Midori, mKIAA1330', description='alpha-kinase 3', other_designations='alpk3|myocyte induction differentiation originator|myocytic induction/differentiation originator', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=81057600, end_position_on_the_genomic_accession=81105612, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  116905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116905, CurrentID=0, Status='live', Symbol='Dph1', Aliases='2310011M22Rik, 4930488F09Rik, AW551873, Dph2l1, Ovca1', description='diphthamide biosynthesis 1', other_designations='DPH1 homolog|DPH2-like 1|diphthamide biosynthesis protein 2 homolog-like 1|diptheria toxin resistance protein required for diphthamide biosynthesis (Saccharomyces)-like 1|mmDph1|ovarian cancer-associated gene 1 protein homolog', map_location='11 45.76 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75177643, end_position_on_the_genomic_accession=75190483, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  116906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116906, CurrentID=0, Status='live', Symbol='Char4', Aliases='', description='P. chabaudi malaria resistance QTL 4', other_designations='', map_location='3 56.54 cM', chromosome='3', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116913, CurrentID=0, Status='live', Symbol='Tpbpb', Aliases='1600012N09Rik', description='trophoblast specific protein beta', other_designations='', map_location='13 B2', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=60901295, end_position_on_the_genomic_accession=60904847, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  116914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116914, CurrentID=0, Status='live', Symbol='Slc19a2', Aliases='AV276020, AW322295, DDA1, THTR1, TRMA', description='solute carrier family 19 (thiamine transporter), member 2', other_designations='', map_location='1 71.56 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=164249046, end_position_on_the_genomic_accession=164265385, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  116927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116927, CurrentID=0, Status='live', Symbol='Artles', Aliases='', description='arterial lesions', other_designations='', map_location='', chromosome='6', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  116939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116939, CurrentID=0, Status='live', Symbol='Pnpla3', Aliases='Adpn', description='patatin-like phospholipase domain containing 3', other_designations='acylglycerol O-acyltransferase|adiponutrin|calcium-independent phospholipase A2-epsilon|iPLA2-epsilon', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84167783, end_position_on_the_genomic_accession=84189521, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  116940 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116940, CurrentID=0, Status='live', Symbol='Tgs1', Aliases='D4Ertd800e, Ncoa6ip, Pimt', description='trimethylguanosine synthase homolog (S. cerevisiae)', other_designations='PIPMT|PRIP-interacting protein with methyltransferase motif|nuclear receptor coactivator 6 interacting protein|nuclear receptor coactivator 6-interacting protein', map_location='4 1.99 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=3574879, end_position_on_the_genomic_accession=3616624, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  641247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=641247, CurrentID=0, Status='live', Symbol='Ighv1-33', Aliases='Gm16693', description='immunoglobulin heavy variable 1-33', other_designations='immunoglobulin heavy variable V1-33', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114843693, end_position_on_the_genomic_accession=114844083, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  116972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=116972, CurrentID=0, Status='live', Symbol='Fam57a', Aliases='2310047D13Rik, 4932415L08Rik, 5430402E13Rik, 5430420K21Rik, Wdt3', description='family with sequence similarity 57, member A', other_designations='whn-dependent transcript 3', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=76179676, end_position_on_the_genomic_accession=76209417, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  117004 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117004, CurrentID=0, Status='live', Symbol='Olfr73', Aliases='MOR174-9, fascin-1, mOR-EG', description='olfactory receptor 73', other_designations='GA_x6K02T2Q125-49527073-49526132|olfactory receptor MOR174-9', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88034196, end_position_on_the_genomic_accession=88037686, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  117005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117005, CurrentID=0, Status='live', Symbol='Olfr74', Aliases='MOR174-4, mOR-EV', description='olfactory receptor 74', other_designations='GA_x6K02T2Q125-49458388-49457432|olfactory receptor MOR174-4', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87973707, end_position_on_the_genomic_accession=87974663, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  117010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117010, CurrentID=0, Status='live', Symbol='Datd', Aliases='', description='dopamine transporter density', other_designations='', map_location='', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117012, CurrentID=0, Status='live', Symbol='Cgnz1', Aliases='', description='chronic glomerulonephritis in NZM 1', other_designations='', map_location='1 76.73 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=641340, CurrentID=0, Status='live', Symbol='Nrbf2', Aliases='NRBF-2', description='nuclear receptor binding factor 2', other_designations='', map_location='10 B5.1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=67266689, end_position_on_the_genomic_accession=67285347, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  117066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117066, CurrentID=0, Status='live', Symbol='Cts3', Aliases='1600000I23Rik', description='cathepsin 3', other_designations='', map_location='13 B3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=61564630, end_position_on_the_genomic_accession=61570162, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  117072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117072, CurrentID=0, Status='live', Symbol='T2dm2', Aliases='', description='type 2 diabetes mellitus 2', other_designations='insulin level locus 2', map_location='19 50.72 cM', chromosome='19', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641361 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=641361, CurrentID=0, Status='live', Symbol='Pinlyp', Aliases='2310033E01Rik', description='phospholipase A2 inhibitor and LY6/PLAUR domain containing', other_designations='X-ray repair complementing defective repair in Chinese hamster cells 1', map_location='7 A2-A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24541654, end_position_on_the_genomic_accession=24546313, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  117074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117074, CurrentID=0, Status='live', Symbol='Actre1', Aliases='actre1q', description='activity response to ethanol 1', other_designations='', map_location='1 81.08 cM', chromosome='1', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117075, CurrentID=0, Status='live', Symbol='Actre2', Aliases='Actre2q', description='activity response to ethanol 2', other_designations='', map_location='2 57.65 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  641368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=641368, CurrentID=0, Status='live', Symbol='Spint5', Aliases='Gm14316, OTTMUSG00000016293', description='serine protease inhibitor, Kunitz type 5', other_designations='KUNITZ5|Spint5 (KUNITZ5)', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164715305, end_position_on_the_genomic_accession=164718068, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  641376 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=641376, CurrentID=0, Status='live', Symbol='Tomm40l', Aliases='', description='translocase of outer mitochondrial membrane 40 homolog-like (yeast)', other_designations='', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=171217802, end_position_on_the_genomic_accession=171222514, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  117109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117109, CurrentID=0, Status='live', Symbol='Pop5', Aliases='1500019J17Rik, 2700077E03Rik, AW049989, Rnasep3', description='processing of precursor 5, ribonuclease P/MRP family (S. cerevisiae)', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115235851, end_position_on_the_genomic_accession=115240972, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  216131 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216131, CurrentID=0, Status='live', Symbol='Trappc10', Aliases='B230307C21Rik, Gm870, Tmem1', description='trafficking protein particle complex 10', other_designations='TRAPP subunit TMEM1|trafficking protein particle complex subunit TMEM1|transmembrane protein 1|transport protein particle subunit TMEM1', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78186725, end_position_on_the_genomic_accession=78244642, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  117146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117146, CurrentID=0, Status='live', Symbol='Ube3b', Aliases='AI449831, AU020130', description='ubiquitin protein ligase E3B', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=114380607, end_position_on_the_genomic_accession=114421166, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  117147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117147, CurrentID=0, Status='live', Symbol='Acsm1', Aliases='Acas3, Bucs1, Macs', description='acyl-CoA synthetase medium-chain family member 1', other_designations='acetyl-Coenzyme A synthetase 3|butyrate--CoA ligase 1|butyryl Coenzyme A synthetase 1|butyryl-coenzyme A synthetase 1|lipoate-activating enzyme|middle-chain acyl-CoA synthetase 1', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119617756, end_position_on_the_genomic_accession=119662515, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  117148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117148, CurrentID=0, Status='live', Symbol='Necab2', Aliases='Efcbp2', description='N-terminal EF-hand calcium binding protein 2', other_designations='EF hand calcium binding protein 2|EF-hand calcium-binding protein 2|neuronal calcium binding 2|neuronal calcium-binding protein 2', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=119446719, end_position_on_the_genomic_accession=119472640, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  117149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117149, CurrentID=0, Status='live', Symbol='Tirap', Aliases='AA407980, C130027E04Rik, Mal, Tlr4ap, Wyatt', description='toll-interleukin 1 receptor (TIR) domain-containing adaptor protein', other_designations='TIR domain-containing adapter protein|Toll-like receptor 4 adaptor protein|adaptor protein Wyatt|myD88 adapter-like protein', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=35184391, end_position_on_the_genomic_accession=35200291, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  117150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117150, CurrentID=0, Status='live', Symbol='Pip4k2c', Aliases='Pip5k2c', description='phosphatidylinositol-5-phosphate 4-kinase, type II, gamma', other_designations='PI(5)P 4-kinase type II gamma|PIP4KII-gamma|phosphatidylinositol 5-phosphate 4-kinase type II gamma|phosphatidylinositol-4-phosphate 5-kinase, type II, gamma|phosphatidylinositol-5-phosphate 4-kinase type-2 gamma', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127197067, end_position_on_the_genomic_accession=127211644, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  216134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216134, CurrentID=0, Status='live', Symbol='Pdxk', Aliases='2310036D04Rik, AA119688, C77999', description='pyridoxal (pyridoxine, vitamin B6) kinase', other_designations='pyridoxine kinase', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78436747, end_position_on_the_genomic_accession=78464948, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  117158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117158, CurrentID=0, Status='live', Symbol='Scgb3a2', Aliases='LuLeu1, Pnsp1, UGRP1, Utgrp1', description='secretoglobin, family 3A, member 2', other_designations='pneumo secretory protein 1|uteroglobin related protein 1|uteroglobin-related protein 1', map_location='18 C-D', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=43764281, end_position_on_the_genomic_accession=43767399, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  117160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117160, CurrentID=0, Status='live', Symbol='Ttyh2', Aliases='1110001A03Rik', description='tweety homolog 2 (Drosophila)', other_designations='mTTY2', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=114675468, end_position_on_the_genomic_accession=114720984, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  117167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117167, CurrentID=0, Status='live', Symbol='Steap4', Aliases='1110021O17Rik, AI481214, Tiarp, Tnfaip9', description='STEAP family member 4', other_designations='Tnfa-induced adipose-related protein|dudulin 4|dudulin-4|six-transmembrane epithelial antigen of prostate 4|tumor necrosis factor, alpha-induced protein 9|tumor necrosis factor-alpha-induced adipose-related protein', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=7960472, end_position_on_the_genomic_accession=7982213, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  216136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216136, CurrentID=0, Status='live', Symbol='Ilvbl', Aliases='5830463I21, AI415009', description='ilvB (bacterial acetolactate synthase)-like', other_designations='ilvB-like protein', map_location='10 39.72 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78574345, end_position_on_the_genomic_accession=78584502, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  117171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117171, CurrentID=0, Status='live', Symbol='1110038F14Rik', Aliases='', description='RIKEN cDNA 1110038F14 gene', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=76948266, end_position_on_the_genomic_accession=76950734, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  117172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117172, CurrentID=0, Status='live', Symbol='2310034C09Rik', Aliases='', description='RIKEN cDNA 2310034C09 gene', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88758847, end_position_on_the_genomic_accession=88759789, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  117197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117197, CurrentID=0, Status='live', Symbol='Bloc1s4', Aliases='2610101N07Rik, Blos4, Cno', description='biogenesis of lysosomal organelles complex-1, subunit 4, cappuccino', other_designations='BLOC-1 subunit 4|biogenesis of organelles complex-1, subunit 4, cappuccino|cappuccino|protein cappuccino homolog', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=36747374, end_position_on_the_genomic_accession=36748679, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  117198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117198, CurrentID=0, Status='live', Symbol='Ivns1abp', Aliases='1190004M08Rik, 1700126I16Rik, AA960440, HSPC068, ND1, NS-1, NS1-BP, Nd1-L, Nd1-S, mKIAA0850', description='influenza virus NS1A binding protein', other_designations='ND1-L2|NS1-binding protein homolog|kelch family protein Nd1|kelch family protein Nd1-L', map_location='1 G2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=151344308, end_position_on_the_genomic_accession=151364445, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  117216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117216, CurrentID=0, Status='live', Symbol='Thim', Aliases='Tim-1', description='thymoma insertional mutation', other_designations='', map_location='', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117218, CurrentID=0, Status='live', Symbol='Ap2q', Aliases='', description='alcohol preference 2 QTL', other_designations='', map_location='2 34.9 cM', chromosome='2', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117219, CurrentID=0, Status='live', Symbol='Ap3q', Aliases='', description='alcohol preference 3 QTL', other_designations='', map_location='4 82.64 cM', chromosome='4', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117220, CurrentID=0, Status='live', Symbol='Ap5q', Aliases='', description='alcohol preference 5 QTL', other_designations='', map_location='9 25.7 cM', chromosome='9', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117223, CurrentID=0, Status='live', Symbol='pogo', Aliases='', description='pogostick', other_designations='', map_location='', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  117229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117229, CurrentID=0, Status='live', Symbol='Stk33', Aliases='4921505G21Rik, AW061264', description='serine/threonine kinase 33', other_designations='', map_location='7 57.21 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=109279213, end_position_on_the_genomic_accession=109439053, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  85078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85078, CurrentID=0, Status='live', Symbol='D8Mit77', Aliases='', description='DNA segment, Chr 8, Massachusetts Institute of Technology 77', other_designations='', map_location='8 39.33 cM', chromosome='8', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  216151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216151, CurrentID=0, Status='live', Symbol='Polrmt', Aliases='1110018N15Rik, 4932416K13', description='polymerase (RNA) mitochondrial (DNA directed)', other_designations='mitochondrial DNA-directed RNA polymerase|mtRPOL', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79736125, end_position_on_the_genomic_accession=79746589, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  216152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216152, CurrentID=0, Status='live', Symbol='Plppr3', Aliases='Lppr3, Prg2', description='phospholipid phosphatase related 3', other_designations='PRG-2|plasticity related gene 2|plasticity-related gene 2 protein', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79860475, end_position_on_the_genomic_accession=79874634, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  85089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85089, CurrentID=0, Status='live', Symbol='D16Mit2', Aliases='', description='DNA segment, Chr 16, Massachusetts Institute of Technology 2', other_designations='', map_location='16 13.8 cM', chromosome='16', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  85094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85094, CurrentID=0, Status='live', Symbol='D7Mit63', Aliases='', description='DNA segment, Chr 7, Massachusetts Institute of Technology 63', other_designations='', map_location='7 57.21 cM', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  238024 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238024, CurrentID=0, Status='live', Symbol='Fn3krp', Aliases='FN3K-RP', description='fructosamine 3 kinase related protein', other_designations='fructosamine 3-kinase-related protein|fructosamine-3-kinase-related protein|ortholog of human fructosamine-3-kinase-related protein FN3KRP', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=121421373, end_position_on_the_genomic_accession=121431288, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  216188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216188, CurrentID=0, Status='live', Symbol='Aldh1l2', Aliases='D330038I09Rik, mtFDH', description='aldehyde dehydrogenase 1 family, member L2', other_designations='aldehyde dehydrogenase 1L2|aldehyde dehydrogenase family 1 member L2, mitochondrial|mitochondrial 10-FTHFDH|probable 10-formyltetrahydrofolate dehydrogenase ALDH1L2', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=83487283, end_position_on_the_genomic_accession=83534146, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  117586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117586, CurrentID=0, Status='live', Symbol='A1bg', Aliases='C44', description='alpha-1-B glycoprotein', other_designations='', map_location='15 D1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=60917589, end_position_on_the_genomic_accession=60921270, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  117589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117589, CurrentID=0, Status='live', Symbol='Asb7', Aliases='AI449039, Asb-7, D030055C23Rik', description='ankyrin repeat and SOCS box-containing 7', other_designations='ankyrin repeat and SOCS box-containing protein 7', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66644565, end_position_on_the_genomic_accession=66689650, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  117590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117590, CurrentID=0, Status='live', Symbol='Asb10', Aliases='Asb-10, BB137407', description='ankyrin repeat and SOCS box-containing 10', other_designations='', map_location='5 A3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=24532685, end_position_on_the_genomic_accession=24541047, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  117591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117591, CurrentID=0, Status='live', Symbol='Slc2a9', Aliases='Glut9, SLC2A9B, SLC2a9A', description='solute carrier family 2 (facilitated glucose transporter), member 9', other_designations='facilitated glucose transporter, type 9', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=38349273, end_position_on_the_genomic_accession=38503149, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  117592 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117592, CurrentID=0, Status='live', Symbol='B3galt6', Aliases='BB129894, GalTII', description='UDP-Gal:betaGal beta 1,3-galactosyltransferase, polypeptide 6', other_designations='GAG GalTII|UDP-Gal:betaGlcNAc beta 1,3-galactosyltransferase, polypeptide 6|beta-1,3-GalTase 6|beta3Gal-T6|beta3GalT6|galactosyltransferase II|galactosylxylosylprotein 3-beta-galactosyltransferase', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155989466, end_position_on_the_genomic_accession=155992678, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  117599 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117599, CurrentID=0, Status='live', Symbol='Helb', Aliases='AI447783, D10Ertd664e', description='helicase (DNA) B', other_designations='', map_location='10 67.94 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=120083608, end_position_on_the_genomic_accession=120112965, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  117600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117600, CurrentID=0, Status='live', Symbol='Srgap1', Aliases='4930572H05Rik, Arhgap13', description='SLIT-ROBO Rho GTPase activating protein 1', other_designations='Rho GTPase activating protein 13|rho GTPase-activating protein 13', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=121780991, end_position_on_the_genomic_accession=122047315, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  117606 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=117606, CurrentID=0, Status='live', Symbol='Boc', Aliases='4732455C11, mFLJ00376', description='biregional cell adhesion molecule-related/down-regulated by oncogenes (Cdon) binding protein', other_designations='biregional Cdon binding protein', map_location='16 B5', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=44485045, end_position_on_the_genomic_accession=44558870, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  216227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216227, CurrentID=0, Status='live', Symbol='Slc17a8', Aliases='BC042593, Vglut3', description='solute carrier family 17 (sodium-dependent inorganic phosphate cotransporter), member 8', other_designations='solute carrier family 17 member 8|vesicular glutamate transporter-3', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=89574020, end_position_on_the_genomic_accession=89621249, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  216233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=216233, CurrentID=0, Status='live', Symbol='Socs2', Aliases='8030460M17, AI527257, AW108012, CIS2, Cish2, D130043N08Rik, JAB, SOCS-2, SSI-2, hg', description='suppressor of cytokine signaling 2', other_designations='CIS-2|STAT-induced STAT inhibitor 2|cytokine inducible SH2-containing protein 2|cytokine-inducible SH2 protein 2', map_location='10 49.35 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=95383903, end_position_on_the_genomic_accession=95417253, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  631323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=631323, CurrentID=0, Status='live', Symbol='Gm12250', Aliases='Irgb10, OTTMUSG00000005723', description='predicted gene 12250', other_designations='interferon-gamma-inducible p47 GTPase', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58183843, end_position_on_the_genomic_accession=58190198, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  85213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=85213, CurrentID=0, Status='live', Symbol='D17Mit133', Aliases='', description='DNA segment, Chr 17, Massachusetts Institute of Technology 133', other_designations='', map_location='17 12.53 cM', chromosome='17', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  19693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=19693, CurrentID=0, Status='live', Symbol='Reg2', Aliases='RegII', description='regenerating islet-derived 2', other_designations='PSP|PTP|REG-2|islet of Langerhans regenerating protein 2|pancreatic stone protein 2|pancreatic thread protein 2|regenerating islet-derived, mouse homolog 2', map_location='6 C3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=78405155, end_position_on_the_genomic_accession=78408097, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  100126226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100126226, CurrentID=0, Status='live', Symbol='5430421N21Rik', Aliases='Krt83', description='RIKEN cDNA 5430421N21 gene', other_designations='keratin type II', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=101485130, end_position_on_the_genomic_accession=101491512, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  118445 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=118445, CurrentID=0, Status='live', Symbol='Klf16', Aliases='AI843742, BTEB4, DRRF', description='Kruppel-like factor 16', other_designations='BTE-binding protein 4|basic transcription element-binding protein 4|dopamine receptor regulating factor|dopamine receptor-regulating factor|transcription factor BTEB4', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=80567121, end_position_on_the_genomic_accession=80577296, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  118446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=118446, CurrentID=0, Status='live', Symbol='Gjc3', Aliases='Cx29, Gje1', description='gap junction protein, gamma 3', other_designations='connexin 29|connexin-29|gap junction epsilon-1 protein|gap junction membrane channel protein epsilon 1', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137953809, end_position_on_the_genomic_accession=137962959, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  118449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=118449, CurrentID=0, Status='live', Symbol='Synpo2', Aliases='1110069I04Rik, 2310068J10Rik, 9530006G20Rik, AI848603, Myo', description='synaptopodin 2', other_designations='myopodin|mypodin', map_location='3 G3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=123076519, end_position_on_the_genomic_accession=123236169, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  118451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=118451, CurrentID=0, Status='live', Symbol='Mrps2', Aliases='1500019M10Rik', description='mitochondrial ribosomal protein S2', other_designations='MRP-S2|S2mt', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=28465956, end_position_on_the_genomic_accession=28471177, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  118452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=118452, CurrentID=0, Status='live', Symbol='Baalc', Aliases='2810457D07Rik', description='brain and acute leukemia, cytoplasmic', other_designations='', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=38933909, end_position_on_the_genomic_accession=38952917, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  118453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=118453, CurrentID=0, Status='live', Symbol='Mmp28', Aliases='D130023P12', description='matrix metallopeptidase 28 (epilysin)', other_designations='matrix metalloproteinase 28 (epilysin)', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=83441241, end_position_on_the_genomic_accession=83463064, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  118454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=118454, CurrentID=0, Status='live', Symbol='Gjc2', Aliases='B230382L12Rik, Cx47, Gja12', description='gap junction protein, gamma 2', other_designations='connexin 47 type A|connexin 47 type C|connexin 47 type D|connexin-47|gap junction alpha-12 protein|gap junction membrane channel protein alpha 12|gap junction protein, chi 2', map_location='11 37.05 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59175564, end_position_on_the_genomic_accession=59183213, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  380601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380601, CurrentID=0, Status='live', Symbol='Fastkd5', Aliases='C78212, mKIAA1792', description='FAST kinase domains 5', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=130613838, end_position_on_the_genomic_accession=130630027, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  380608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380608, CurrentID=0, Status='live', Symbol='Tagap1', Aliases='2610315E15Rik, AA415005, FKSG15', description='T cell activation GTPase activating protein 1', other_designations='T-cell activation GTPase activating protein 1', map_location='17 A1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=6954965, end_position_on_the_genomic_accession=6961334, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  380614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380614, CurrentID=0, Status='live', Symbol='Intu', Aliases='9230116I04Rik, 9430087H23Rik, Pdzd6, Pdzk6, mKIAA1284', description='inturned planar cell polarity protein', other_designations='PDZ domain-containing protein 6|inturned planar cell polarity effector homolog', map_location='3 B', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=40653702, end_position_on_the_genomic_accession=40704774, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  380629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380629, CurrentID=0, Status='live', Symbol='Heca', Aliases='Gm869, HDC, HDCL', description='hdc homolog, cell cycle regulator', other_designations='headcase homolog', map_location='10 A2-A3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=17899042, end_position_on_the_genomic_accession=17948067, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  380654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380654, CurrentID=0, Status='live', Symbol='Cfap54', Aliases='4930485B16Rik, Gm872', description='cilia and flagella associated protein 54', other_designations='ciliary and flagellar protein 54', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=92775612, end_position_on_the_genomic_accession=93083592, orientation='minus', exon_count=71, OMIM='', no_hdr0=''),
  380660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380660, CurrentID=0, Status='live', Symbol='Acss3', Aliases='8430416H19Rik, Gm874', description='acyl-CoA synthetase short-chain family member 3', other_designations='', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=106936164, end_position_on_the_genomic_accession=107123664, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  380664 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380664, CurrentID=0, Status='live', Symbol='Lemd3', Aliases='AI316861, Man1', description='LEM domain containing 3', other_designations='LEM domain-containing protein 3|integral inner nuclear membrane protein MAN1', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=120923411, end_position_on_the_genomic_accession=120979949, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  380669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380669, CurrentID=0, Status='live', Symbol='Lin28b', Aliases='2810403D23Rik, D030047M17Rik, Lin-28.2', description='lin-28 homolog B (C. elegans)', other_designations='lin-28B', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=45376619, end_position_on_the_genomic_accession=45487303, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  380683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380683, CurrentID=0, Status='live', Symbol='Sec14l3', Aliases='1110069O07Rik', description='SEC14-like lipid binding 3', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4064853, end_position_on_the_genomic_accession=4078990, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  380684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380684, CurrentID=0, Status='live', Symbol='Nefh', Aliases='NF-H, NF200, Nfh, mKIAA0845', description='neurofilament, heavy polypeptide', other_designations='200 kDa neurofilament protein|neurofilament 200kDa|neurofilament triplet H protein', map_location='11 A1-A5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=4938754, end_position_on_the_genomic_accession=4948064, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  380686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380686, CurrentID=0, Status='live', Symbol='Cnrip1', Aliases='1500041B16Rik, 3110054C06Rik, 5330437A18Rik, AI854501, C2orf32', description='cannabinoid receptor interacting protein 1', other_designations='CRIP-1', map_location='11 A3.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=17051934, end_position_on_the_genomic_accession=17079372, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  380687 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380687, CurrentID=0, Status='live', Symbol='Gm5138', Aliases='ENSMUSG00000079501', description='predicted gene 5138', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=83874199, end_position_on_the_genomic_accession=83875398, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  380694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380694, CurrentID=0, Status='live', Symbol='Ccnjl', Aliases='Gm877', description='cyclin J-like', other_designations='', map_location='11 B1.1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43528749, end_position_on_the_genomic_accession=43586999, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  380698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380698, CurrentID=0, Status='live', Symbol='Obscn', Aliases='BC046431, Gm878, OTTMUSG00000005786, UNC89', description='obscurin, cytoskeletal calmodulin and titin-interacting RhoGEF', other_designations='obscn|obscurin-MLCK|obscurin-RhoGEF|obscurin-myosin light chain kinase', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58994256, end_position_on_the_genomic_accession=59136375, orientation='minus', exon_count=114, OMIM='', no_hdr0=''),
  380701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380701, CurrentID=0, Status='live', Symbol='Slc47a2', Aliases='4933429E10Rik', description='solute carrier family 47, member 2', other_designations='MATE-2|h(+)/organic cation antiporter kidney-specific|mMATE-2', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=61301627, end_position_on_the_genomic_accession=61342860, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  380702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380702, CurrentID=0, Status='live', Symbol='Shisa6', Aliases='Gm879', description='shisa family member 6', other_designations='shisa homolog 6', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=66211725, end_position_on_the_genomic_accession=66526126, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  380705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380705, CurrentID=0, Status='live', Symbol='Tmem102', Aliases='4632403M07Rik, Cbap-ps, Tmem102', description='transmembrane protein 102', other_designations='transmembrane protein 102, pseudogene', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=69803595, end_position_on_the_genomic_accession=69805624, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  380709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380709, CurrentID=0, Status='live', Symbol='Spata22', Aliases='Gm882', description='spermatogenesis associated 22', other_designations='', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73326537, end_position_on_the_genomic_accession=73346434, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  380711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380711, CurrentID=0, Status='live', Symbol='Rap1gap2', Aliases='AU067654, Garnl4, Gm1561, mKIAA1039', description='RAP1 GTPase activating protein 2', other_designations='GTPase activating RANGAP domain-like 4|GTPase-activating Rap/Ran-GAP domain-like protein 4', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74383483, end_position_on_the_genomic_accession=74610910, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  380712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380712, CurrentID=0, Status='live', Symbol='Tlcd2', Aliases='', description='TLC domain containing 2', other_designations='', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75468024, end_position_on_the_genomic_accession=75470899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  380713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380713, CurrentID=0, Status='live', Symbol='Scarf1', Aliases='AA986099, SREC, SREC-I, mKIAA0149', description='scavenger receptor class F, member 1', other_designations='acetyl LDL receptor|scavenger receptor expressed by endothelial cells 1 protein', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75513495, end_position_on_the_genomic_accession=75526580, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  380714 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380714, CurrentID=0, Status='live', Symbol='Rph3al', Aliases='Gm1753, Noc2', description='rabphilin 3A-like (without C2 domains)', other_designations='no C2 domains protein|rabphilin-3A-like protein', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75830992, end_position_on_the_genomic_accession=75967250, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  380718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380718, CurrentID=0, Status='live', Symbol='Mks1', Aliases='AK190930, B8d3', description='Meckel syndrome, type 1', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87852928, end_position_on_the_genomic_accession=87863679, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  380728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380728, CurrentID=0, Status='live', Symbol='Kcnh4', Aliases='BEC2', description='potassium voltage-gated channel, subfamily H (eag-related), member 4', other_designations='', map_location='11 63.57 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=100740378, end_position_on_the_genomic_accession=100760199, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  380730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380730, CurrentID=0, Status='live', Symbol='Gm884', Aliases='LRRC37', description='predicted gene 884', other_designations='', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103534577, end_position_on_the_genomic_accession=103627878, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  380732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380732, CurrentID=0, Status='live', Symbol='Milr1', Aliases='Allergin-1, Gm885, Mca32', description='mast cell immunoglobulin like receptor 1', other_designations='allergy inhibitory receptor 1|mast cell Ag-32|mast cell antigen 32|mast cell immunoglobulin-like receptor 1|surface protein MCA-32', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=106750652, end_position_on_the_genomic_accession=106768794, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  380752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380752, CurrentID=0, Status='live', Symbol='Tssc1', Aliases='D12Ertd604e', description='tumor suppressing subtransferable candidate 1', other_designations='', map_location='12 11.26 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=28751804, end_position_on_the_genomic_accession=28867494, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  380753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380753, CurrentID=0, Status='live', Symbol='Atxn7l1', Aliases='2810423G08Rik, Atxn7l4', description='ataxin 7-like 1', other_designations='ataxin 7-like 4', map_location='12 A3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=33147639, end_position_on_the_genomic_accession=33376408, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  380755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380755, CurrentID=0, Status='live', Symbol='Lsmem1', Aliases='Gm889', description='leucine-rich single-pass membrane protein 1', other_designations='', map_location='12 B1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=40176386, end_position_on_the_genomic_accession=40199315, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  380768 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380768, CurrentID=0, Status='live', Symbol='Ccdc177', Aliases='Gm1568', description='coiled-coil domain containing 177', other_designations='', map_location='12 D1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=80755447, end_position_on_the_genomic_accession=80760715, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  380773 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380773, CurrentID=0, Status='live', Symbol='Slirp', Aliases='1810035L17Rik', description='SRA stem-loop interacting RNA binding protein', other_designations='', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87443896, end_position_on_the_genomic_accession=87449924, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  380780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380780, CurrentID=0, Status='live', Symbol='Serpina11', Aliases='Gm895', description='serine (or cysteine) peptidase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 11', other_designations='serine (or cysteine) proteinase inhibitor, clade A (alpha-1 antiproteinase, antitrypsin), member 11|serine (or cysteine) proteinase inhibitor, clade A, member 11', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=103980243, end_position_on_the_genomic_accession=103989970, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  380785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380785, CurrentID=0, Status='live', Symbol='Begain', Aliases='BM948371, Gm897', description='brain-enriched guanylate kinase-associated', other_designations='brain-enriched guanylate kinase-associated; brain-enriched guanylate kinase-associated protein', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109032181, end_position_on_the_genomic_accession=109068251, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  380787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380787, CurrentID=0, Status='live', Symbol='A230065H16Rik', Aliases='Gm898, Lbh2', description='RIKEN cDNA A230065H16 gene', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=111402047, end_position_on_the_genomic_accession=111412077, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  380792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380792, CurrentID=0, Status='live', Symbol='Ighe', Aliases='Gm900', description='Immunoglobulin heavy constant epsilon', other_designations='', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113269263, end_position_on_the_genomic_accession=113273248, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  380793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380793, CurrentID=0, Status='live', Symbol='Ighg2a', Aliases='1810060O09Rik, Igh-1, Igh-1a', description='immunoglobulin heavy constant gamma 2A', other_designations='Ig heavy chain Mem5|MLT2-23 heavy chain|immunoglobulin gamma-2a heavy chain|immunoglobulin heavy chain 1 (serum IgG2a)|immunoglobulin heavy chain 1a (serum IgG2a)', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.3', start_position_on_the_genomic_accession=110806667, end_position_on_the_genomic_accession=110813558, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  380794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380794, CurrentID=0, Status='live', Symbol='Ighg', Aliases='AU044919, IGHG2A', description='Immunoglobulin heavy chain (gamma polypeptide)', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  380795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380795, CurrentID=0, Status='live', Symbol='Ighg3', Aliases='IgG3', description='Immunoglobulin heavy constant gamma 3', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113357442, end_position_on_the_genomic_accession=113361232, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  380797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380797, CurrentID=0, Status='live', Symbol='Ighd', Aliases='IgD, Igh-5', description='immunoglobulin heavy constant delta', other_designations='immunoglobulin delta chain|immunoglobulin delta-chain|immunoglobulin heavy chain 5 (delta-like heavy chain)', map_location='12 62.1 cM', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113407535, end_position_on_the_genomic_accession=113416324, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  380809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380809, CurrentID=0, Status='live', Symbol='Ighv10-3', Aliases='EG380809, Gm5140', description='immunoglobulin heavy variable V10-3', other_designations='anti-poly(dC) monoclonal antibody heavy chain', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114523441, end_position_on_the_genomic_accession=114523887, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  380836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380836, CurrentID=0, Status='live', Symbol='Mrs2', Aliases='Gm902, HPTl, RPT, Mrs2', description='MRS2 magnesium homeostasis factor homolog (S. cerevisiae)', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=24987483, end_position_on_the_genomic_accession=25020358, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  380839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380839, CurrentID=0, Status='live', Symbol='Serpinb1c', Aliases='EIC, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 1c', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 1c|serine (or cysteine) proteinase inhibitor, clade B, member 1c|serine protease inhibitor EIC|serpin B1c', map_location='13 13.76 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=32881397, end_position_on_the_genomic_accession=32898140, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  380840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380840, CurrentID=0, Status='live', Symbol='Lyrm4', Aliases='BC034664, Gm903', description='LYR motif containing 4', other_designations='', map_location='13 A3.3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=35978797, end_position_on_the_genomic_accession=36117357, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  380845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380845, CurrentID=0, Status='live', Symbol='Gm904', Aliases='Gm1135', description='predicted gene 904', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=50643228, end_position_on_the_genomic_accession=50645838, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  380850 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380850, CurrentID=0, Status='live', Symbol='Gm5141', Aliases='', description='predicted gene 5141', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=62772200, end_position_on_the_genomic_accession=62785808, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  380855 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380855, CurrentID=0, Status='live', Symbol='Rsl1', Aliases='rslcan-9', description='regulator of sex limited protein 1', other_designations='regulator of sex limited protein-Slp 1|regulator of sex-limitation 1', map_location='13 34.54 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=67173182, end_position_on_the_genomic_accession=67183499, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  380882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380882, CurrentID=0, Status='live', Symbol='Gm906', Aliases='', description='predicted gene 906', other_designations='', map_location='13 A5', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=50238757, end_position_on_the_genomic_accession=50253444, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  380907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380907, CurrentID=0, Status='live', Symbol='Gm5142', Aliases='EG380907', description='predicted gene 5142', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=59158503, end_position_on_the_genomic_accession=59178749, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  380912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380912, CurrentID=0, Status='live', Symbol='Zfp395', Aliases='BC053701, Gm910', description='zinc finger protein 395', other_designations='papillomavirus regulatory factor', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=65358450, end_position_on_the_genomic_accession=65398930, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  380916 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380916, CurrentID=0, Status='live', Symbol='Lrch1', Aliases='4832412D13Rik, AI552548, AW215503, Chdc1, mKIAA1016', description='leucine-rich repeats and calponin homology (CH) domain containing 1', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=74754673, end_position_on_the_genomic_accession=74948039, orientation='minus', exon_count=23, OMIM='', no_hdr0=''),
  380921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380921, CurrentID=0, Status='live', Symbol='Dgkh', Aliases='5930402B05Rik, D130015C16, DGK', description='diacylglycerol kinase, eta', other_designations='', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=78559349, end_position_on_the_genomic_accession=78732614, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  380924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380924, CurrentID=0, Status='live', Symbol='Olfm4', Aliases='GC1, GW112, Gm296, Gm913, OlfD, pPD4', description='olfactomedin 4', other_designations='OLM4|PU.1 difference product 4', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=80000154, end_position_on_the_genomic_accession=80022862, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  380928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380928, CurrentID=0, Status='live', Symbol='Lmo7', Aliases='C78582, FBXO20, Gm914, mKIAA0858', description='LIM domain only 7', other_designations='', map_location='14 50.9 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=101729917, end_position_on_the_genomic_accession=101934708, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  380930 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380930, CurrentID=0, Status='live', Symbol='9330188P03Rik', Aliases='Gm1589, mgt-6', description='RIKEN cDNA 9330188P03 gene', other_designations='', map_location='14 E2.3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=105589549, end_position_on_the_genomic_accession=105593929, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  545156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545156, CurrentID=0, Status='live', Symbol='Kalrn', Aliases='2210407G14Rik, AV235988, DUET, E530005C20Rik, Gm539, Hapip, TRAD', description='kalirin, RhoGEF kinase', other_designations='huntingtin-associated protein interacting protein (duo)|kalrn|protein Duo|serine/threonine kinase with Dbl- and pleckstrin homology domain|serine/threonine kinase with Dbl- and pleckstrin homology domains|serine/threonine-protein kinase with Dbl- and pleckstrin homology domain', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=33969073, end_position_on_the_genomic_accession=34573291, orientation='minus', exon_count=69, OMIM='', no_hdr0=''),
  380959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380959, CurrentID=0, Status='live', Symbol='Alg10b', Aliases='AA469671, AU045404, AW123895, C85569, Deaf1, Gm917, nse5', description='asparagine-linked glycosylation 10B (alpha-1,2-glucosyltransferase)', other_designations='alpha-1,2-glucosyltransferase ALG10-A|alpha-2-glucosyltransferase ALG10-B|asparagine-linked glycosylation 10 homolog B (yeast, alpha-1,2-glucosyltransferase)|asparagine-linked glycosylation protein 10 homolog B|potassium channel regulator 1|putative alpha-1,2-glucosyltransferase ALG10-B', map_location='15 E3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=90224311, end_position_on_the_genomic_accession=90230554, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  380967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380967, CurrentID=0, Status='live', Symbol='Tmem106c', Aliases='AI046681, BC046621, D15Ertd405e', description='transmembrane protein 106C', other_designations='', map_location='15 53.96 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=97964197, end_position_on_the_genomic_accession=97970286, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  380969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380969, CurrentID=0, Status='live', Symbol='Nckap5l', Aliases='C230021P08Rik', description='NCK-associated protein 5-like', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=99422034, end_position_on_the_genomic_accession=99457774, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  100126777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100126777, CurrentID=0, Status='live', Symbol='Esp4', Aliases='Gm20580', description='exocrine gland secreted peptide 4', other_designations='', map_location='17 19.54 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40598594, end_position_on_the_genomic_accession=40602618, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  100126778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100126778, CurrentID=0, Status='live', Symbol='Esp8', Aliases='Gm17495', description='exocrine gland secreted peptide 8', other_designations='', map_location='17 19.53 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=40520022, end_position_on_the_genomic_accession=40530702, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  380993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380993, CurrentID=0, Status='live', Symbol='Zfat', Aliases='Gm9221, Zfp406, Zfat', description='zinc finger and AT hook domain containing', other_designations='ZFAT zinc finger 1|zinc finger protein 403|zinc finger protein 406', map_location='15 D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=68083714, end_position_on_the_genomic_accession=68259265, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  380997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=380997, CurrentID=0, Status='live', Symbol='Cyp2d12', Aliases='9030605E09Rik', description='cytochrome P450, family 2, subfamily d, polypeptide 12', other_designations='cytochrome P450, 2d12', map_location='15 38.64 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=82554960, end_position_on_the_genomic_accession=82559413, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  381022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381022, CurrentID=0, Status='live', Symbol='Kmt2d', Aliases='ALR, BC032281, BC058659, C430014K11Rik, KMT2B, Mll2, Mll4', description='lysine (K)-specific methyltransferase 2D', other_designations='ALL1-related protein|histone-lysine N-methyltransferase MLL2|lysine N-methyltransferase 2B|lysine N-methyltransferase 2D|myeloid/lymphoid or mixed-lineage leukemia 2|myeloid/lymphoid or mixed-lineage leukemia protein 2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98831669, end_position_on_the_genomic_accession=98871205, orientation='minus', exon_count=55, OMIM='', no_hdr0=''),
  100126824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100126824, CurrentID=0, Status='live', Symbol='Sco2', Aliases='', description='SCO cytochrome oxidase deficient homolog 2 (yeast)', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=89371637, end_position_on_the_genomic_accession=89373818, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  100126829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100126829, CurrentID=0, Status='live', Symbol='Pif', Aliases='', description='proteolysis-inducing factor', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  381038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381038, CurrentID=0, Status='live', Symbol='Parl', Aliases='D16Ertd607e, PRO2207, PSARL1, PSENIP2, Psarl', description='presenilin associated, rhomboid-like', other_designations='mitochondrial intramembrane-cleaving protease PARL', map_location='16 12.4 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=20279820, end_position_on_the_genomic_accession=20302395, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  381045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381045, CurrentID=0, Status='live', Symbol='Ccdc58', Aliases='A930007B11Rik, AI413631', description='coiled-coil domain containing 58', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=36071660, end_position_on_the_genomic_accession=36092118, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  381058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381058, CurrentID=0, Status='live', Symbol='Unc93a', Aliases='Unc93l, unc-93A', description='unc-93 homolog A (C. elegans)', other_designations='unc93 homolog A', map_location='17 8.78 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=13108617, end_position_on_the_genomic_accession=13143625, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  381062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381062, CurrentID=0, Status='live', Symbol='Ermard', Aliases='2210404J11Rik, 2410011O22Rik, AI448938', description='ER membrane associated RNA degradation', other_designations='', map_location='17 A2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=15041563, end_position_on_the_genomic_accession=15090045, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  381067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381067, CurrentID=0, Status='live', Symbol='Zfp229', Aliases='', description='zinc finger protein 229', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=21733724, end_position_on_the_genomic_accession=21748969, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  381072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381072, CurrentID=0, Status='live', Symbol='Abca17', Aliases='', description='ATP-binding cassette, sub-family A (ABC1), member 17', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24264258, end_position_on_the_genomic_accession=24351897, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  381073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381073, CurrentID=0, Status='live', Symbol='Npw', Aliases='Gm935', description='neuropeptide W', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=24657330, end_position_on_the_genomic_accession=24658425, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  381077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381077, CurrentID=0, Status='live', Symbol='Ccdc78', Aliases='Gm938, Gm950', description='coiled-coil domain containing 78', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=25785619, end_position_on_the_genomic_accession=25790513, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  381085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381085, CurrentID=0, Status='live', Symbol='Tbc1d22b', Aliases='BC045600', description='TBC1 domain family, member 22B', other_designations='', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=29549788, end_position_on_the_genomic_accession=29606894, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  381091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381091, CurrentID=0, Status='live', Symbol='H2-Eb2', Aliases='A130038H09Rik, H-2Eb2, Ia-5, Ia5, Iebeta2', description='histocompatibility 2, class II antigen E beta2', other_designations='H-2 class II histocompatibility antigen E-a/k beta-2|MHC class II E-beta-2|MHC class II antigen E beta2|MHC class II antigen beta chain', map_location='17 17.98 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34325500, end_position_on_the_genomic_accession=34341410, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  381101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381101, CurrentID=0, Status='live', Symbol='Dnph1', Aliases='C76683, Rcl', description="2'-deoxynucleoside 5'-phosphate N-hydrolase 1", other_designations="c-Myc-responsive protein Rcl|deoxyribonucleoside 5'-monophosphate N-glycosidase|putative c-Myc-responsive", map_location='17 C', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=46496789, end_position_on_the_genomic_accession=46499618, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  381107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381107, CurrentID=0, Status='live', Symbol='Tmem232', Aliases='E130009J12Rik, Gm940, Tes13-L, Tes13-S', description='transmembrane protein 232', other_designations='testis-specific protein 13', map_location='17 E1.1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=65255999, end_position_on_the_genomic_accession=65541102, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  381110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381110, CurrentID=0, Status='live', Symbol='Rmdn2', Aliases='AW061290, Fam82a1, RMD-2, mRMD-2', description='regulator of microtubule dynamics 2', other_designations='family with sequence similarity 82, member A1|microtubule-associated protein', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=79578598, end_position_on_the_genomic_accession=79685067, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  381112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381112, CurrentID=0, Status='live', Symbol='Arhgef33', Aliases='Gm941', description='Rho guanine nucleotide exchange factor (GEF) 33', other_designations='DH and coiled-coil domain-containing protein ENSP00000381780 homolog', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80306930, end_position_on_the_genomic_accession=80427079, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  381113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381113, CurrentID=0, Status='live', Symbol='Cdkl4', Aliases='AU067824, Gm942', description='cyclin-dependent kinase-like 4', other_designations='', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=80523550, end_position_on_the_genomic_accession=80577839, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  381122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381122, CurrentID=0, Status='live', Symbol='Capn13', Aliases='CANP 13, Gm943', description='calpain 13', other_designations='calcium-activated neutral proteinase 13', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=73306456, end_position_on_the_genomic_accession=73399306, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  381126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381126, CurrentID=0, Status='live', Symbol='Garem', Aliases='C86169, Fam59a, Gm944, mKIAA4238', description='GRB2 associated, regulator of MAPK1', other_designations='GRB2-associated and regulator of MAPK1|family with sequence similarity 59, member A|protein FAM59A', map_location='18 A2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=21127342, end_position_on_the_genomic_accession=21300139, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381142, CurrentID=0, Status='live', Symbol='Arl14epl', Aliases='Gm949', description='ADP-ribosylation factor-like 14 effector protein-like', other_designations='', map_location='18 C', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=46921812, end_position_on_the_genomic_accession=46934222, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381148, CurrentID=0, Status='live', Symbol='Prob1', Aliases='Gm1614', description='proline rich basic protein 1', other_designations='', map_location='18 B2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=35650351, end_position_on_the_genomic_accession=35655199, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  381157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381157, CurrentID=0, Status='live', Symbol='Greb1l', Aliases='mKIAA4095', description='growth regulation by estrogen in breast cancer-like', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10324963, end_position_on_the_genomic_accession=10562941, orientation='plus', exon_count=36, OMIM='', no_hdr0=''),
  667370 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=667370, CurrentID=0, Status='live', Symbol='Ifit3b', Aliases='I830012O16Rik', description='interferon-induced protein with tetratricopeptide repeats 3B', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=34607539, end_position_on_the_genomic_accession=34613409, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  381175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381175, CurrentID=0, Status='live', Symbol='Ccdc68', Aliases='BC046343', description='coiled-coil domain containing 68', other_designations='', map_location='18 E2', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=69924858, end_position_on_the_genomic_accession=69969484, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  381196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381196, CurrentID=0, Status='live', Symbol='Gm960', Aliases='', description='predicted gene 960', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=4625734, end_position_on_the_genomic_accession=4698769, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  381199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381199, CurrentID=0, Status='live', Symbol='Tmem151a', Aliases='AW125688, Gm961, Tmem151', description='transmembrane protein 151A', other_designations='transmembrane protein 151', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5079337, end_position_on_the_genomic_accession=5085477, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  381201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381201, CurrentID=0, Status='live', Symbol='Ap5b1', Aliases='Gm962', description='adaptor-related protein complex 5, beta 1 subunit', other_designations='AP-5 complex subunit beta|adapter-related protein complex 5 beta subunit|adaptor-related protein complex 5 beta subunit|beta-5|beta5', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5560250, end_position_on_the_genomic_accession=5571262, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  381203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381203, CurrentID=0, Status='live', Symbol='Slc22a20', Aliases='Gm963, OAT1-L, mOAT6', description='solute carrier family 22 (organic anion transporter), member 20', other_designations='OAT1-like organic anion transporter|SLC22 organic anion transporter|organic anion transporter 6', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5970232, end_position_on_the_genomic_accession=5986551, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  381217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381217, CurrentID=0, Status='live', Symbol='Fam189a2', Aliases='AI414343, AI666772, Gm967', description='family with sequence similarity 189, member A2', other_designations='', map_location='19 B', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=23972750, end_position_on_the_genomic_accession=24031273, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  381218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381218, CurrentID=0, Status='live', Symbol='4430402I18Rik', Aliases='AW122618, Spata6l', description='RIKEN cDNA 4430402I18 gene', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=28845528, end_position_on_the_genomic_accession=28970888, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  381232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381232, CurrentID=0, Status='live', Symbol='Mirt1', Aliases='4833407H14Rik, 5430437C04Rik, 5830416P10Rik, ENSMUSG00000074788, Gm10758', description='myocardial infarction associated transcript 1', other_designations='myocardial infarction-associated transcript 1', map_location='19 D2', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=53441212, end_position_on_the_genomic_accession=53464796, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  381236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381236, CurrentID=0, Status='live', Symbol='Lipo1', Aliases='AI747699', description='lipase, member O1', other_designations='', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=33555160, end_position_on_the_genomic_accession=33591172, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  381246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381246, CurrentID=0, Status='live', Symbol='Xkr9', Aliases='Gm1620, XRG9', description='X-linked Kx blood group related 9', other_designations='X Kell blood group precursor related family member 9 homolog', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=13668738, end_position_on_the_genomic_accession=13701723, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381259, CurrentID=0, Status='live', Symbol='Tmem237', Aliases='AI853305, Als2cr4, Gm972', description='transmembrane protein 237', other_designations='amyotrophic lateral sclerosis 2 (juvenile) chromosome region, candidate 4|amyotrophic lateral sclerosis 2 chromosomal region candidate gene 4 protein homolog', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59100584, end_position_on_the_genomic_accession=59120480, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  381260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381260, CurrentID=0, Status='live', Symbol='Gm973', Aliases='AW554518', description='predicted gene 973', other_designations='', map_location='1 C1.3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=59514723, end_position_on_the_genomic_accession=59636417, orientation='plus', exon_count=26, OMIM='', no_hdr0=''),
  381269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381269, CurrentID=0, Status='live', Symbol='Mreg', Aliases='Gm974, Wdt2, dsu', description='melanoregulin', other_designations='dilute suppressor protein|whn-dependent transcript 2', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72159233, end_position_on_the_genomic_accession=72212307, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  381270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381270, CurrentID=0, Status='live', Symbol='March4', Aliases='BC056494, mKIAA1399', description='membrane-associated ring finger (C3HC4) 4', other_designations='MARCH-IV|membrane-associated RING finger protein 4|membrane-associated RING-CH protein IV', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=72427112, end_position_on_the_genomic_accession=72536564, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381280, CurrentID=0, Status='live', Symbol='Hjurp', Aliases='6430706D22Rik, A730008H23Rik, C330011F01Rik, Gm975, Mfleg1', description='Holliday junction recognition protein', other_designations='fetal liver expressing gene 1 protein homolog', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=88258856, end_position_on_the_genomic_accession=88277633, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  381284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381284, CurrentID=0, Status='live', Symbol='Crocc2', Aliases='E030010N08Rik, Gm976', description='ciliary rootlet coiled-coil, rootletin family member 2', other_designations='', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93168423, end_position_on_the_genomic_accession=93231067, orientation='plus', exon_count=37, OMIM='', no_hdr0=''),
  381286 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381286, CurrentID=0, Status='live', Symbol='Serpinb3c', Aliases='1110001H02Rik, 1110013A16Rik, Scca2, Serpinb4', description='serine (or cysteine) peptidase inhibitor, clade B, member 3C', other_designations='ovalbumin', map_location='1 E1-E2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107271201, end_position_on_the_genomic_accession=107278371, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  381287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381287, CurrentID=0, Status='live', Symbol='A530032D15Rik', Aliases='Gm1888', description='RIKEN cDNA A530032D15Rik gene', other_designations='', map_location='1 C5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=85088139, end_position_on_the_genomic_accession=85110501, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  381290 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381290, CurrentID=0, Status='live', Symbol='Atp2b4', Aliases='Pmca4', description='ATPase, Ca++ transporting, plasma membrane 4', other_designations='', map_location='1 57.95 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=133702674, end_position_on_the_genomic_accession=133801037, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  381293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381293, CurrentID=0, Status='live', Symbol='Kif14', Aliases='D1Ertd367e, E130203M01', description='kinesin family member 14', other_designations='N-3 kinesin|truncated kinesin family member 14', map_location='1 F', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=136466290, end_position_on_the_genomic_accession=136555938, orientation='plus', exon_count=41, OMIM='', no_hdr0=''),
  381305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381305, CurrentID=0, Status='live', Symbol='Rc3h1', Aliases='5730557L09Rik, Gm551, N28103, mKIAA2025', description='RING CCCH (C3H) domains 1', other_designations='RING finger and C3H zinc finger protein 1|RING finger and CCCH-type zinc finger domain-containing protein 1|probable E3 ubiquitin-protein ligase Roquin|protein Sanroque|roquin', map_location='1 H2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=160906411, end_position_on_the_genomic_accession=160974976, orientation='plus', exon_count=21, OMIM='', no_hdr0=''),
  381306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381306, CurrentID=0, Status='live', Symbol='BC055324', Aliases='', description='cDNA sequence BC055324', other_designations='', map_location='1 H2.2', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=163945840, end_position_on_the_genomic_accession=163994900, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  381308 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381308, CurrentID=0, Status='live', Symbol='Mnda', Aliases='A730048F03, Ifi205b, ifi-205-B', description='myeloid cell nuclear differentiation antigen', other_designations='interferon activated|interferon-inducible protein p205-B|lipopolysaccharide-inducible|protein D3', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173896341, end_position_on_the_genomic_accession=173913046, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  381310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381310, CurrentID=0, Status='live', Symbol='6330403A02Rik', Aliases='ENSMUSG00000056615, Gm10001, R75555', description='RIKEN cDNA 6330403A02 gene', other_designations='', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=180432373, end_position_on_the_genomic_accession=180483519, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  381314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381314, CurrentID=0, Status='live', Symbol='Iars2', Aliases='2010002H18Rik, C79125', description='isoleucine-tRNA synthetase 2, mitochondrial', other_designations='ileRS|isoleucyl-tRNA synthetase, mitochondrial', map_location='1 H5', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=185286662, end_position_on_the_genomic_accession=185329419, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  381318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381318, CurrentID=0, Status='live', Symbol='Nsl1', Aliases='4833432M17Rik, AI451597, Gm105, Gm984', description='NSL1, MIND kinetochore complex component, homolog (S. cerevisiae)', other_designations='', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191063012, end_position_on_the_genomic_accession=191084558, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  381319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381319, CurrentID=0, Status='live', Symbol='Batf3', Aliases='9130211I03Rik, Snft', description='basic leucine zipper transcription factor, ATF-like 3', other_designations='B-ATF-3|Jun dimerization protein p21SNFT', map_location='1 H6', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=191098414, end_position_on_the_genomic_accession=191108943, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  381334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381334, CurrentID=0, Status='live', Symbol='Gal3st2', Aliases='GP3ST, Gm988', description='galactose-3-O-sulfotransferase 2', other_designations="beta-galactose-3-O-sulfotransferase 2|gal-beta-1, 3-GalNAc 3'-sulfotransferase 2|gal3ST-2|galactose 3-O-sulfotransferase 2", map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=93861344, end_position_on_the_genomic_accession=93876494, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  381337 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381337, CurrentID=0, Status='live', Symbol='Fam178b', Aliases='1700024G10Rik, Gm990', description='family with sequence similarity 178, member B', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=36562696, end_position_on_the_genomic_accession=36683190, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  381338 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381338, CurrentID=0, Status='live', Symbol='Lonrf2', Aliases='2900060P06Rik, AI851349', description='LON peptidase N-terminal domain and ring finger 2', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=38793669, end_position_on_the_genomic_accession=38836719, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  381339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381339, CurrentID=0, Status='live', Symbol='Tmem182', Aliases='2310079P10Rik, AI414494', description='transmembrane protein 182', other_designations='', map_location='1 B', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=40805601, end_position_on_the_genomic_accession=40856887, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381350, CurrentID=0, Status='live', Symbol='Spag6l', Aliases='', description='sperm associated antigen 6 like', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=18694011, end_position_on_the_genomic_accession=18749746, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  381353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381353, CurrentID=0, Status='live', Symbol='Gm996', Aliases='', description='predicted gene 996', other_designations='', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=25575416, end_position_on_the_genomic_accession=25580099, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  381356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381356, CurrentID=0, Status='live', Symbol='Cacfd1', Aliases='5930434B04Rik, mFwe', description='calcium channel flower domain containing 1', other_designations='calcium channel flower domain-containing protein 1|likely ortholog of H. sapiens chromosome 9 open reading frame 7 (C9orf7)', map_location='2 A3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=27009841, end_position_on_the_genomic_accession=27021089, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  381359 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381359, CurrentID=0, Status='live', Symbol='Prdm12', Aliases='Gm998', description='PR domain containing 12', other_designations='PR domain-containing protein 12', map_location='2 B', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=31633517, end_position_on_the_genomic_accession=31655795, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  381373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381373, CurrentID=0, Status='live', Symbol='Sp9', Aliases='', description='trans-acting transcription factor 9', other_designations='buttonhead-like zinc finger transcription factor Sp9', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=73271926, end_position_on_the_genomic_accession=73275771, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  381375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381375, CurrentID=0, Status='live', Symbol='Dfnb59', Aliases='Gm1001, Pjvk', description='deafness, autosomal recessive 59 (human)', other_designations='autosomal recessive deafness type 59 protein homolog', map_location='2 C3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=76648327, end_position_on_the_genomic_accession=76658554, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  381379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381379, CurrentID=0, Status='live', Symbol='Med19', Aliases='2410018M14Rik, 3110040A13Rik, LCMR1', description='mediator complex subunit 19', other_designations='mediator of RNA polymerase II transcription, subunit 19 homolog', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84678402, end_position_on_the_genomic_accession=84688215, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381393, CurrentID=0, Status='live', Symbol='4921509C19Rik', Aliases='Gm1005', description='RIKEN cDNA 4921509C19 gene', other_designations='MarkmE1', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151460380, end_position_on_the_genomic_accession=151476268, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  381399 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381399, CurrentID=0, Status='live', Symbol='Bpifb4', Aliases='Gm1006, Lplunc4', description='BPI fold containing family B, member 4', other_designations='LBP / BPI / CETP family, N-terminal domain containing protein|Novel LBP/BPI/CETP family, N-terminal domain containing protein|ligand-binding protein RY2G5|long palate, lung and nasal epithelium carcinoma-associated protein 4', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=153938217, end_position_on_the_genomic_accession=153963852, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  381404 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381404, CurrentID=0, Status='live', Symbol='Pabpc1l', Aliases='1810053B01Rik, Epab', description='poly(A) binding protein, cytoplasmic 1-like', other_designations='embryonic poly(A)-binding protein', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=164024770, end_position_on_the_genomic_accession=164050538, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  381405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381405, CurrentID=0, Status='live', Symbol='Zfp663', Aliases='Gm1008', description='zinc finger protein 663', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165351297, end_position_on_the_genomic_accession=165369977, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381406, CurrentID=0, Status='live', Symbol='Trp53rka', Aliases='2810408M09Rik', description='transformation related protein 53 regulating kinase A', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=165490112, end_position_on_the_genomic_accession=165493314, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  381409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381409, CurrentID=0, Status='live', Symbol='Cdh26', Aliases='Gm1010', description='cadherin-like 26', other_designations='', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=178430515, end_position_on_the_genomic_accession=178487366, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  381410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381410, CurrentID=0, Status='live', Symbol='Zfp408', Aliases='Gm1011', description='zinc finger protein 408', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=91643168, end_position_on_the_genomic_accession=91649897, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  381411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381411, CurrentID=0, Status='live', Symbol='Accsl', Aliases='Gm1967', description='1-aminocyclopropane-1-carboxylate synthase (non-functional)-like', other_designations='1-aminocyclopropane-1-carboxylate synthase homolog(non-functional)-like|1-aminocyclopropane-1-carboxylate synthase-like protein 2|ACC synthase-like protein 2|novel aminotransferase class I and II domain containing protein', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=93855360, end_position_on_the_genomic_accession=93931236, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  381413 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381413, CurrentID=0, Status='live', Symbol='Gpr176', Aliases='Gm1012', description='G protein-coupled receptor 176', other_designations='G-protein coupled receptor AGR9', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118277096, end_position_on_the_genomic_accession=118373419, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  381418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381418, CurrentID=0, Status='live', Symbol='Ctxn2', Aliases='BB181834, Gm1638', description='cortexin 2', other_designations='', map_location='2 F1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=125134592, end_position_on_the_genomic_accession=125147841, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381438, CurrentID=0, Status='live', Symbol='Gm5148', Aliases='EG381438', description='predicted gene 5148', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=37714190, end_position_on_the_genomic_accession=37724360, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  381457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381457, CurrentID=0, Status='live', Symbol='Crnn', Aliases='Gm1015', description='cornulin', other_designations='', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=93144787, end_position_on_the_genomic_accession=93149720, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381463, CurrentID=0, Status='live', Symbol='Nr1h5', Aliases='Fxrb', description='nuclear receptor subfamily 1, group H, member 5', other_designations='', map_location='3 F2.2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=102939658, end_position_on_the_genomic_accession=102964290, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  381476 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381476, CurrentID=0, Status='live', Symbol='Stpg2', Aliases='B930007M17Rik, Gm1017', description='sperm tail PG rich repeat containing 2', other_designations='', map_location='3 H1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=139205194, end_position_on_the_genomic_accession=139835622, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  381485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381485, CurrentID=0, Status='live', Symbol='Trim55', Aliases='D830041C10Rik, Murf2, Rnf29', description='tripartite motif-containing 55', other_designations='ring finger protein 29', map_location='3 A2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=19644328, end_position_on_the_genomic_accession=19692348, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  381489 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381489, CurrentID=0, Status='live', Symbol='Rxfp1', Aliases='Gm1018, Lgr7', description='relaxin/insulin-like family peptide receptor 1', other_designations='leucine-rich repeat-containing 7|leucine-rich repeat-containing G protein-coupled receptor 7|leucine-rich repeat-containing G-protein coupled receptor 7|relaxin family peptide receptor 1', map_location='3 E3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=79641600, end_position_on_the_genomic_accession=79737951, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  381493 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381493, CurrentID=0, Status='live', Symbol='S100a7a', Aliases='AY465109, Gm1020, S100A7L1, S100A7f, S100a15, S100a15a, S100a17l1', description='S100 calcium binding protein A7A', other_designations='S100 calcium binding protein A15|S100 calcium binding protein A7-like 1|S100 calcium-binding protein A15|S100 calcium-binding protein A15A|S100 calcium-binding protein A7A', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90651934, end_position_on_the_genomic_accession=90658130, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381510, CurrentID=0, Status='live', Symbol='Dpy19l4', Aliases='Gm1023, Narg3', description='dpy-19-like 4 (C. elegans)', other_designations='NMDA receptor-regulated gene 3|dpy-19-like protein 4|protein dpy-19 homolog 4', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11261315, end_position_on_the_genomic_accession=11322135, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  381511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381511, CurrentID=0, Status='live', Symbol='Pdp1', Aliases='Gm1024, Ppm2c', description='pyruvate dehyrogenase phosphatase catalytic subunit 1', other_designations='PDP 1|PDPC 1|protein phosphatase 2C, magnesium dependent, catalytic subunit|pyruvate dehydrogenase phosphatase catalytic subunit 1|pyruvate dehydrogenase phosphatase isoenzyme 2', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=11958183, end_position_on_the_genomic_accession=11966450, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  381522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381522, CurrentID=0, Status='live', Symbol='Ccdc180', Aliases='E230008N13Rik, Gm1025', description='coiled-coil domain containing 180', other_designations='', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=45890303, end_position_on_the_genomic_accession=45950774, orientation='plus', exon_count=38, OMIM='', no_hdr0=''),
  381530 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381530, CurrentID=0, Status='live', Symbol='Mup20', Aliases='Gm12560, Mup24, OTTMUSG00000007485, darcin', description='major urinary protein 20', other_designations='darcin|major urinary protein 24|mup20', map_location='4 B3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62011406, end_position_on_the_genomic_accession=62054143, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  381531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381531, CurrentID=0, Status='live', Symbol='Mup21', Aliases='Gm11208, Mup, Mup26, OTTMUSG00000000231, bM64F17.1, bM64F17.4', description='major urinary protein 21', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=62147832, end_position_on_the_genomic_accession=62150841, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381534, CurrentID=0, Status='live', Symbol='Ube2u', Aliases='', description='ubiquitin-conjugating enzyme E2U (putative)', other_designations='', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=100478867, end_position_on_the_genomic_accession=100550145, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  381538 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381538, CurrentID=0, Status='live', Symbol='Mroh7', Aliases='Gm1027, Heatr8', description='maestro heat-like repeat family member 7', other_designations='HEAT repeat-containing protein 8', map_location='4 C7', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=106680417, end_position_on_the_genomic_accession=106736620, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  381544 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381544, CurrentID=0, Status='live', Symbol='Gm1661', Aliases='Gm1662, Ncrna00082', description='predicted gene 1661', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=117213326, end_position_on_the_genomic_accession=117251917, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  381549 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381549, CurrentID=0, Status='live', Symbol='Zfp69', Aliases='Gm1029, Krab2, Zfp63', description='zinc finger protein 69', other_designations='Zfp63p', map_location='4 56.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120930133, end_position_on_the_genomic_accession=120951701, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  381560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381560, CurrentID=0, Status='live', Symbol='Xkr8', Aliases='4931440N07Rik, AI597014, Gm1031, XRG8, mXkr8', description='X-linked Kx blood group related 8', other_designations='X Kell blood group precursor related family member 8 homolog', map_location='4 D2.3', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=132724904, end_position_on_the_genomic_accession=132732614, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  381569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381569, CurrentID=0, Status='live', Symbol='Gm13084', Aliases='OTTMUSG00000010328', description='predicted gene 13084', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=143809250, end_position_on_the_genomic_accession=143816093, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  381570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381570, CurrentID=0, Status='live', Symbol='Oog2', Aliases='C85069', description='oogenesin 2', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144190717, end_position_on_the_genomic_accession=144196934, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381572, CurrentID=0, Status='live', Symbol='9430007A20Rik', Aliases='Aadacl4, Gm1034', description='RIKEN cDNA 9430007A20 gene', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144519822, end_position_on_the_genomic_accession=144529353, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  381580 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381580, CurrentID=0, Status='live', Symbol='Ccdc27', Aliases='Gm1035', description='coiled-coil domain containing 27', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=154026644, end_position_on_the_genomic_accession=154042982, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  381582 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381582, CurrentID=0, Status='live', Symbol='Tmem240', Aliases='EG381582, Gm5151', description='transmembrane protein 240', other_designations='', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=155734622, end_position_on_the_genomic_accession=155740568, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  381591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381591, CurrentID=0, Status='live', Symbol='L1td1', Aliases='AA546746, AB211064, D76865, ECAT11', description='LINE-1 type transposase domain containing 1', other_designations='ES cell-associated protein 11', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=98726754, end_position_on_the_genomic_accession=98738486, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381598, CurrentID=0, Status='live', Symbol='2610005L07Rik', Aliases='2810038F24Rik, 6720476A01Rik', description='RIKEN cDNA 2610005L07 gene', other_designations='', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=20385782, end_position_on_the_genomic_accession=20424814, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381605 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381605, CurrentID=0, Status='live', Symbol='Tbc1d2', Aliases='A630005A06Rik, BC022630, Gm1037, PARIS-1, PARIS1', description='TBC1 domain family, member 2', other_designations='likely ortholog of H. sapiens TBC1 domain family, member 2 (TBC1D2)', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=46604390, end_position_on_the_genomic_accession=46650877, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  381622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381622, CurrentID=0, Status='live', Symbol='5031410I06Rik', Aliases='', description='RIKEN cDNA 5031410I06 gene', other_designations='', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=26098668, end_position_on_the_genomic_accession=26107407, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  381626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381626, CurrentID=0, Status='live', Symbol='Rbm33', Aliases='3200001K10Rik, 6430512A10Rik, AA386890, Prr8', description='RNA binding motif protein 33', other_designations='RNA-binding motif protein 33|proline rich 8|proline-rich protein 8', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=28316891, end_position_on_the_genomic_accession=28419244, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  381628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381628, CurrentID=0, Status='live', Symbol='Adgrf3', Aliases='Gm1041, Gpr113, Pgr23', description='adhesion G protein-coupled receptor F3', other_designations='G protein-coupled receptor 113|G-protein coupled receptor 113|G-protein coupled receptor PGR23|probable G-protein coupled receptor 113', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30193431, end_position_on_the_genomic_accession=30205754, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  381629 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381629, CurrentID=0, Status='live', Symbol='Atraid', Aliases='0610007C21Rik, AI316792, Apr3, HSPC013, p18', description='all-trans retinoic acid induced differentiation factor', other_designations='H1E6 protein|apoptosis related protein APR-3', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=31048640, end_position_on_the_genomic_accession=31054633, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  381633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381633, CurrentID=0, Status='live', Symbol='Gm1673', Aliases='', description='predicted gene 1673', other_designations='', map_location='5 B2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=33983472, end_position_on_the_genomic_accession=33985009, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  381634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381634, CurrentID=0, Status='live', Symbol='Gm1043', Aliases='', description='predicted gene 1043', other_designations='', map_location='5 B3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=37153129, end_position_on_the_genomic_accession=37234484, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  100127433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=100127433, CurrentID=0, Status='live', Symbol='Gm11532', Aliases='OTTMUSG00000001860, Sp6os', description='predicted gene 11532', other_designations='Sp6 opposite strand anti-sense RNA', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=97024639, end_position_on_the_genomic_accession=97026491, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  381644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381644, CurrentID=0, Status='live', Symbol='Cep135', Aliases='BC062951, Cep4, Gm1044', description='centrosomal protein 135', other_designations='centrosomal protein 4', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=76587249, end_position_on_the_genomic_accession=76646466, orientation='plus', exon_count=29, OMIM='', no_hdr0=''),
  381651 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381651, CurrentID=0, Status='live', Symbol='Gm1045', Aliases='Gm1045-a-b, Gm1045', description='predicted gene 1045', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=91987473, end_position_on_the_genomic_accession=91995317, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  381654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381654, CurrentID=0, Status='live', Symbol='C87414', Aliases='', description='expressed sequence C87414', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=93635185, end_position_on_the_genomic_accession=93671463, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381667, CurrentID=0, Status='live', Symbol='Lrcol1', Aliases='Clpsl3, Gm1679', description='leucine rich colipase-like 1', other_designations='colipase-like protein 3 variant 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110344507, end_position_on_the_genomic_accession=110356094, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  381668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381668, CurrentID=0, Status='live', Symbol='Fbrsl1', Aliases='2410025L10Rik, AI467568, Gm1049, Gm29766, mKIAA1545', description='fibrosin-like 1', other_designations='AUTS2-like protein', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110429875, end_position_on_the_genomic_accession=110431955, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  381677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381677, CurrentID=0, Status='live', Symbol='Vgf', Aliases='Gm1052', description='VGF nerve growth factor inducible', other_designations='', map_location='5 76.13 cM', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137025302, end_position_on_the_genomic_accession=137033351, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  381678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381678, CurrentID=0, Status='live', Symbol='Zcwpw1', Aliases='Gm1053', description='zinc finger, CW type with PWWP domain 1', other_designations='', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137787395, end_position_on_the_genomic_accession=137822621, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  381680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381680, CurrentID=0, Status='live', Symbol='Nxpe5', Aliases='', description='neurexophilin and PC-esterase domain family, member 5', other_designations='Fam55 related', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=138207499, end_position_on_the_genomic_accession=138251876, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  381686 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381686, CurrentID=0, Status='live', Symbol='Kpna7', Aliases='AW146299, Gm1055', description='karyopherin alpha 7 (importin alpha 8)', other_designations='importin alpha 2|karyopherin subunit alpha-7|nuclear transport receptor', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=144983727, end_position_on_the_genomic_accession=145009636, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  107305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107305, CurrentID=0, Status='live', Symbol='Vps37c', Aliases='5730409F24Rik, AU042646', description='vacuolar protein sorting 37C (yeast)', other_designations='ESCRT-I complex subunit VPS37C', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=10688729, end_position_on_the_genomic_accession=10714628, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  381693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381693, CurrentID=0, Status='live', Symbol='Wdr95', Aliases='4930434E21Rik', description='WD40 repeat domain 95', other_designations='', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149528679, end_position_on_the_genomic_accession=149611894, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  381694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381694, CurrentID=0, Status='live', Symbol='B3glct', Aliases='B3galtl, Gm1057', description='beta-3-glucosyltransferase', other_designations='beta 1,3-galactosyltransferase-like|beta 3-glucosyltransferase|beta-3-glycosyltransferase-like|beta3Glc-T', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=149678105, end_position_on_the_genomic_accession=149762599, orientation='plus', exon_count=17, OMIM='', no_hdr0=''),
  381695 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381695, CurrentID=0, Status='live', Symbol='N4bp2l2', Aliases='2700092H06Rik, zag1', description='NEDD4 binding protein 2-like 2', other_designations='phosphonoformate immuno-associated protein 5 homolog', map_location='5 G3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=150635973, end_position_on_the_genomic_accession=150665612, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381716, CurrentID=0, Status='live', Symbol='1700015F17Rik', Aliases='Gm1686', description='RIKEN cDNA 1700015F17 gene', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=5437827, end_position_on_the_genomic_accession=5479143, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  381738 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381738, CurrentID=0, Status='live', Symbol='Drc1', Aliases='Ccdc164, Gm1060, b2b1654Clo', description='dynein regulatory complex subunit 1', other_designations='coiled-coil domain containing 164|coiled-coil domain-containing protein 164', map_location='5 B1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=30341663, end_position_on_the_genomic_accession=30366708, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  381741 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381741, CurrentID=0, Status='live', Symbol='Lrrc43', Aliases='Gm1061', description='leucine rich repeat containing 43', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=123489264, end_position_on_the_genomic_accession=123508205, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  381759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381759, CurrentID=0, Status='live', Symbol='Wee2', Aliases='Gm1065, Wee1B', description='WEE1 homolog 2 (S. pombe)', other_designations='mWee1B|wee1-like protein kinase 1B|wee1B kinase', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40426699, end_position_on_the_genomic_accession=40466815, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  381760 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381760, CurrentID=0, Status='live', Symbol='Ssbp1', Aliases='2810480P10Rik, G630031O20Rik, MtSSB, mtDBP', description='single-stranded DNA binding protein 1', other_designations='', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40471415, end_position_on_the_genomic_accession=40481825, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  381776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381776, CurrentID=0, Status='live', Symbol='Igkv2-112', Aliases='EG381776, Gm5153, hg24', description='immunoglobulin kappa variable 2-112', other_designations='Ig kappa V-region 24|immunoglobulin kappa chain variable 2-112|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68219981, end_position_on_the_genomic_accession=68220705, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  381777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381777, CurrentID=0, Status='live', Symbol='Igkv1-110', Aliases='Gm16634, Igk, K5.1', description='immunoglobulin kappa variable 1-110', other_designations='anti-fluorescein antibody|immunoglobulin kappa chain variable 1-110', map_location='6 31.12 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68270527, end_position_on_the_genomic_accession=68271265, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  381778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381778, CurrentID=0, Status='live', Symbol='Igkv16-104', Aliases='Gm1067, rf', description='immunoglobulin kappa variable 16-104', other_designations='IgM (kappa)light-chain|immunoglobulin kappa chain variable 16-104|immunoglobulin kappa-chain VK-1|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=68425605, end_position_on_the_genomic_accession=68426072, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  238393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238393, CurrentID=0, Status='live', Symbol='Serpina3f', Aliases='2A1, BC049975', description='serine (or cysteine) peptidase inhibitor, clade A, member 3F', other_designations='alpha-1 antiproteinasin|antitrypsin|serpin A3F', map_location='12 E', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=104214544, end_position_on_the_genomic_accession=104221129, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  381784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381784, CurrentID=0, Status='live', Symbol='Igkv6-25', Aliases='19-25, Gm16943', description='immunoglobulin kappa chain variable 6-25', other_designations='immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=70215433, end_position_on_the_genomic_accession=70215957, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  381798 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381798, CurrentID=0, Status='live', Symbol='4930590J08Rik', Aliases='Gm1071', description='RIKEN cDNA 4930590J08 gene', other_designations='', map_location='6 D1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=91902692, end_position_on_the_genomic_accession=91951736, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  381801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381801, CurrentID=0, Status='live', Symbol='Tatdn2', Aliases='AI646012, mKIAA0218', description='TatD DNase domain containing 2', other_designations='', map_location='6 52.83 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=113697499, end_position_on_the_genomic_accession=113711070, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  381802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381802, CurrentID=0, Status='live', Symbol='Tsen2', Aliases='AU067695', description='tRNA splicing endonuclease 2 homolog (S. cerevisiae)', other_designations='tRNA-intron endonuclease Sen2', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=115544704, end_position_on_the_genomic_accession=115578340, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  381809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381809, CurrentID=0, Status='live', Symbol='Clec4b2', Aliases='Aplra1, F830043G12Rik, mDCAR1', description='C-type lectin domain family 4, member b2', other_designations='antigen presenting cell lectin-like receptor A1', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=123173023, end_position_on_the_genomic_accession=123204749, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  381810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381810, CurrentID=0, Status='live', Symbol='Lpar5', Aliases='GPR93, Gm1072, Gpr92, LPA5', description='lysophosphatidic acid receptor 5', other_designations='G protein-coupled receptor 92|G-protein coupled receptor 92|LPA receptor 5|LPA-5', map_location='6 F2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=125067842, end_position_on_the_genomic_accession=125086129, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  381812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381812, CurrentID=0, Status='live', Symbol='Cracr2a', Aliases='Efcab4b, Gm1073, Gm462', description='calcium release activated channel regulator 2A', other_designations='EF-hand calcium binding domain 4B|calcium release-activated channel regulator 2A', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127561064, end_position_on_the_genomic_accession=127670161, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  381813 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381813, CurrentID=0, Status='live', Symbol='Prmt8', Aliases='Hrmt1l3, Hrmt1l4', description='protein arginine N-methyltransferase 8', other_designations='HMT1 hnRNP methyltransferase-like 3|heterogeneous nuclear ribonucleoprotein methyltransferase-like 4|heterogeneous nuclear ribonucleoprotein methyltransferase-like protein 4|protein arginine N-methyltransferase 4', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=127689009, end_position_on_the_genomic_accession=127769159, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  381816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381816, CurrentID=0, Status='live', Symbol='4922502D21Rik', Aliases='Clec2m, Gm1074, SPATA78', description='RIKEN cDNA 4922502D21 gene', other_designations='', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=129322165, end_position_on_the_genomic_accession=129331808, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381820, CurrentID=0, Status='live', Symbol='Smim10l1', Aliases='2700089E24Rik, AI662692', description='small integral membrane protein 10 like 1', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=133104914, end_position_on_the_genomic_accession=133110899, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  381823 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381823, CurrentID=0, Status='live', Symbol='Apold1', Aliases='Gm1075', description='apolipoprotein L domain containing 1', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=134982001, end_position_on_the_genomic_accession=134986836, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  381832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381832, CurrentID=0, Status='live', Symbol='Prpmp5', Aliases='MP5', description='proline-rich protein MP5', other_designations='proline-rich protein PRP MP5', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132311590, end_position_on_the_genomic_accession=132314743, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  381835 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381835, CurrentID=0, Status='live', Symbol='Sbk3', Aliases='Gm1078, Sgk110', description='SH3 domain binding kinase family, member 3', other_designations='SH3-binding domain kinase family member 3|sugen kinase 110', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4965260, end_position_on_the_genomic_accession=4983395, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  381838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381838, CurrentID=0, Status='live', Symbol='Vmn2r43', Aliases='EC2-V2R, Gm1961', description='vomeronasal 2, receptor 43', other_designations='EC2-V2R pheromone receptor|vomeronasal receptor Vmn2r43', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=8244348, end_position_on_the_genomic_accession=8260599, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381845, CurrentID=0, Status='live', Symbol='Rnf225', Aliases='2310014L17Rik, AV237412, C77118', description='ring finger protein 225', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=12927416, end_position_on_the_genomic_accession=12931072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  381852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381852, CurrentID=0, Status='live', Symbol='Gm5155', Aliases='EG381852', description='predicted gene 5155', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17871744, end_position_on_the_genomic_accession=17919043, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  381853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381853, CurrentID=0, Status='live', Symbol='Gipr', Aliases='GIP-R, Gm1081, Gm160', description='gastric inhibitory polypeptide receptor', other_designations='glucose-dependent insulinotropic polypeptide receptor', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19156063, end_position_on_the_genomic_accession=19170248, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  381867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381867, CurrentID=0, Status='live', Symbol='Ovol3', Aliases='Gm1903, movo3', description='OVO homolog-like 3 (Drosophila)', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=30233297, end_position_on_the_genomic_accession=30237575, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  381903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381903, CurrentID=0, Status='live', Symbol='Alg8', Aliases='AI447372, Gm1089', description='asparagine-linked glycosylation 8 (alpha-1,3-glucosyltransferase)', other_designations='asparagine-linked glycosylation 8 homolog (yeast, alpha-1,3-glucosyltransferase)|asparagine-linked glycosylation protein 8 homolog|dol-P-Glc:Glc(1)Man(9)GlcNAc(2)-PP-dolichyl alpha-1,3-glucosyltransferase|dolichyl pyrophosphate Glc1Man9GlcNAc2 alpha-1,3-glucosyltransferase|dolichyl-P-Glc:Glc1Man9GlcNAc2-PP-dolichyl glucosyltransferase', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=97371594, end_position_on_the_genomic_accession=97392159, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  381917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381917, CurrentID=0, Status='live', Symbol='Dnah3', Aliases='BC051401, Dnahc3', description='dynein, axonemal, heavy chain 3', other_designations='axonemal beta dynein heavy chain 3|ciliary dynein heavy chain 3|dynein, axonemal, heavy polypeptide 3', map_location='7 F2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=119922671, end_position_on_the_genomic_accession=120095675, orientation='minus', exon_count=61, OMIM='', no_hdr0=''),
  381921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381921, CurrentID=0, Status='live', Symbol='Taok2', Aliases='1110033K02Rik, B230344N16, MAP3K17, PSK, PSK1, TAO1, TAO2, mKIAA0881', description='TAO kinase 2', other_designations='TAO kinase 2TAO kinase 2|thousand and one amino acid protein 2', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=126865677, end_position_on_the_genomic_accession=126886296, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  381924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381924, CurrentID=0, Status='live', Symbol='Itgad', Aliases='Cd11d', description='integrin, alpha D', other_designations='', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=128173933, end_position_on_the_genomic_accession=128205521, orientation='plus', exon_count=31, OMIM='', no_hdr0=''),
  381925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381925, CurrentID=0, Status='live', Symbol='Plpp4', Aliases='C030048B12Rik, Gm1090, Ppapdc1a', description='phospholipid phosphatase 4', other_designations='phosphatidate phosphatase PPAPDC1A|phosphatidic acid phosphatase type 2 domain containing 1A|phosphatidic acid phosphatase type 2 domain-containing protein 1A', map_location='7 F3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=129257094, end_position_on_the_genomic_accession=129391307, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  381933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381933, CurrentID=0, Status='live', Symbol='6430531B16Rik', Aliases='Gm1093', description='RIKEN cDNA 6430531B16 gene', other_designations='', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139972303, end_position_on_the_genomic_accession=139978755, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  381944 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381944, CurrentID=0, Status='live', Symbol='Usp17lb', Aliases='DUB-1A, Dub1a', description='ubiquitin specific peptidase 17-like B', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104840257, end_position_on_the_genomic_accession=104842597, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  381970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381970, CurrentID=0, Status='live', Symbol='Scgb2b2', Aliases='Abpbg2, Abpe, lacrimal', description='secretoglobin, family 2B, member 2', other_designations='androgen binding protein epsilon|androgen-binding protein epsilon|lacrimal androgen-binding protein epsilon', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31302769, end_position_on_the_genomic_accession=31304977, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  381974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381974, CurrentID=0, Status='live', Symbol='Mrgprg', Aliases='Gm1098, Mrgg', description='MAS-related GPR, member G', other_designations='G protein-coupled receptor MrgG|evolutionary breakpoint transcript 2 protein', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=143763710, end_position_on_the_genomic_accession=143766993, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  381979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381979, CurrentID=0, Status='live', Symbol='Brsk1', Aliases='Gm1100, SAD-B, SADB', description='BR serine/threonine kinase 1', other_designations='BR serine/threonine-protein kinase 1|brain-specific serine/threonine-protein kinase 1|serine/threonine kinase SAD-B|serine/threonine kinase SADB|serine/threonine-protein kinase SAD-B', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=4690928, end_position_on_the_genomic_accession=4715997, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  381983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381983, CurrentID=0, Status='live', Symbol='Lmtk3', Aliases='BC059845, aatyk3', description='lemur tyrosine kinase 3', other_designations='apoptosis-associated tyrosine kinase 3', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45783743, end_position_on_the_genomic_accession=45806891, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  381990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=381990, CurrentID=0, Status='live', Symbol='Zbtb2', Aliases='Gm1103', description='zinc finger and BTB domain containing 2', other_designations='', map_location='10 A1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=4367073, end_position_on_the_genomic_accession=4392701, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  382000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382000, CurrentID=0, Status='live', Symbol='AY761184', Aliases='CRS1C-3, CRS1C-6', description='cDNA sequence AY761184', other_designations='novel defensin-related cryptidin', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21702524, end_position_on_the_genomic_accession=21703647, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  382007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382007, CurrentID=0, Status='live', Symbol='Adam26b', Aliases='', description='a disintegrin and metallopeptidase domain 26B', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=43519864, end_position_on_the_genomic_accession=43528137, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  382010 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382010, CurrentID=0, Status='live', Symbol='Cep44', Aliases='4933440G23Rik, mKIAA1712', description='centrosomal protein 44', other_designations='', map_location='8 B2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=56531516, end_position_on_the_genomic_accession=56550777, orientation='minus', exon_count=16, OMIM='', no_hdr0=''),
  382014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382014, CurrentID=0, Status='live', Symbol='Ano8', Aliases='BC053460, Tmem16h', description='anoctamin 8', other_designations='transmembrane protein 16H', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71476017, end_position_on_the_genomic_accession=71486067, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  382018 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382018, CurrentID=0, Status='live', Symbol='Unc13a', Aliases='2410078G03Rik, Munc13-1', description='unc-13 homolog A (C. elegans)', other_designations='unc13a', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71626712, end_position_on_the_genomic_accession=71671775, orientation='minus', exon_count=45, OMIM='', no_hdr0=''),
  382019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382019, CurrentID=0, Status='live', Symbol='Zfp882', Aliases='ENSMUSG00000052439, Gm15468, OTTMUSG00000022410', description='zinc finger protein 882', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=71908606, end_position_on_the_genomic_accession=71918851, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  382030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382030, CurrentID=0, Status='live', Symbol='Cnep1r1', Aliases='5033428A16Rik, NEP1-R1, Tmem188', description='CTD nuclear envelope phosphatase 1 regulatory subunit 1', other_designations='transmembrane protein 188', map_location='8 C4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88118735, end_position_on_the_genomic_accession=88135197, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  382034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382034, CurrentID=0, Status='live', Symbol='Gse1', Aliases='2210013I18Rik, mKIAA0182', description='genetic suppressor element 1', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=120228826, end_position_on_the_genomic_accession=120581383, orientation='plus', exon_count=20, OMIM='', no_hdr0=''),
  382035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382035, CurrentID=0, Status='live', Symbol='Pabpn1l', Aliases='Gm1108, Pabpnl1, ePABP2', description='poly(A)binding protein nuclear 1-like', other_designations='T-cell receptor alpha chain V region (8/10-2)|ePABP-2|embryonic poly(A) binding protein 2|embryonic poly(A)-binding protein 2|embryonic poly(A)-binding protein type II|poly(A)-binding protein nuclear-like 1|poly(A)binding protein nuclear-like 1', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122619471, end_position_on_the_genomic_accession=122622735, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  382038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382038, CurrentID=0, Status='live', Symbol='Urb2', Aliases='C77575, mKIAA0133', description='URB2 ribosome biogenesis 2 homolog (S. cerevisiae)', other_designations='', map_location='8 E2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=124021442, end_position_on_the_genomic_accession=124048505, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  382045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382045, CurrentID=0, Status='live', Symbol='Adgrg5', Aliases='Gm1109, Gpr114, PGR27', description='adhesion G protein-coupled receptor G5', other_designations='G protein-coupled receptor 114|G-protein coupled receptor 114|G-protein coupled receptor PGR27|probable G-protein coupled receptor 114', map_location='8 D1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=94921811, end_position_on_the_genomic_accession=94943293, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  382051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382051, CurrentID=0, Status='live', Symbol='Pdp2', Aliases='4833426J09Rik, Gm1705, mKIAA1348', description='pyruvate dehyrogenase phosphatase catalytic subunit 2', other_designations='pyruvate dehydrogenase phosphatase isoenzyme 2', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=104591468, end_position_on_the_genomic_accession=104599026, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  382053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382053, CurrentID=0, Status='live', Symbol='Ces3a', Aliases='Ces3, Ces31, Es31', description='carboxylesterase 3A', other_designations='Es-male carboxylesterase|esterase 31|esterase-31|liver carboxylesterase 31', map_location='8 D3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=105048577, end_position_on_the_genomic_accession=105061928, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  382056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382056, CurrentID=0, Status='live', Symbol='Crtc1', Aliases='AI413414, Mect1, R74955, TORC1, mKIAA0616', description='CREB regulated transcription coactivator 1', other_designations='TORC-1|crtc1|mucoepidermoid carcinoma translocated 1|mucoepidermoid carcinoma translocated protein 1 homolog|transducer of CREB protein 1|transducer of regulated cAMP response element-binding protein 1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=70382352, end_position_on_the_genomic_accession=70439643, orientation='minus', exon_count=15, OMIM='', no_hdr0=''),
  382059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382059, CurrentID=0, Status='live', Symbol='Defa22', Aliases='Defcr22', description='defensin, alpha, 22', other_designations='defensin-related cryptdin-22', map_location='8 A2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=21162277, end_position_on_the_genomic_accession=21163249, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  382062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382062, CurrentID=0, Status='live', Symbol='AB124611', Aliases='Gm1706', description='cDNA sequence AB124611', other_designations='HIDE1|highly expressed in immature dendritic cell transcript 1 protein', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21526166, end_position_on_the_genomic_accession=21545331, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  382064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382064, CurrentID=0, Status='live', Symbol='Gm1110', Aliases='', description='predicted gene 1110', other_designations='', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=26879567, end_position_on_the_genomic_accession=26923184, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  107368 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107368, CurrentID=0, Status='live', Symbol='Pdzd8', Aliases='A630041P07Rik, AV329468, AW228944, Pdzk8', description='PDZ domain containing 8', other_designations='', map_location='19 D3', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=59296084, end_position_on_the_genomic_accession=59345780, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  382066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382066, CurrentID=0, Status='live', Symbol='Prdm10', Aliases='Gm1112, tristanin', description='PR domain containing 10', other_designations='PR domain-containing protein 10', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=31278780, end_position_on_the_genomic_accession=31381731, orientation='plus', exon_count=28, OMIM='', no_hdr0=''),
  382073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382073, CurrentID=0, Status='live', Symbol='Ccdc84', Aliases='D630044F24Rik, Gm1114', description='coiled-coil domain containing 84', other_designations='', map_location='9 A5.2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44410159, end_position_on_the_genomic_accession=44418084, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  382074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382074, CurrentID=0, Status='live', Symbol='Foxr1', Aliases='Foxn5, Gm1115', description='forkhead box R1', other_designations='forkhead box protein N5', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44434234, end_position_on_the_genomic_accession=44440868, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  382075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382075, CurrentID=0, Status='live', Symbol='Odf3l1', Aliases='BC049697, Gm1116', description='outer dense fiber of sperm tails 3-like 1', other_designations='', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56848657, end_position_on_the_genomic_accession=56864674, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  382077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382077, CurrentID=0, Status='live', Symbol='Ccdc33', Aliases='4930535E21Rik, Gm1117, Tzip1', description='coiled-coil domain containing 33', other_designations='speriolin-binding factor', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=58028673, end_position_on_the_genomic_accession=58118823, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  382083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382083, CurrentID=0, Status='live', Symbol='Snx22', Aliases='Gm1120', description='sorting nexin 22', other_designations='', map_location='9 C', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=66065176, end_position_on_the_genomic_accession=66069780, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  382088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382088, CurrentID=0, Status='live', Symbol='Omt2b', Aliases='AU018149, C86230, Mosg, OM2b', description='oocyte maturation, beta', other_designations='', map_location='9 43.65 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=78328030, end_position_on_the_genomic_accession=78329620, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  382089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382089, CurrentID=0, Status='live', Symbol='Ripply2', Aliases='C030002E08Rik, Gm1122', description='ripply2 homolog (zebrafish)', other_designations='', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=87015128, end_position_on_the_genomic_accession=87020441, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  382090 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382090, CurrentID=0, Status='live', Symbol='Cep162', Aliases='4922501C03Rik, mKIAA1009', description='centrosomal protein 162', other_designations='protein QN1 homolog', map_location='9 E3.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=87189570, end_position_on_the_genomic_accession=87255536, orientation='minus', exon_count=27, OMIM='', no_hdr0=''),
  382097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382097, CurrentID=0, Status='live', Symbol='Gm1123', Aliases='AV052824, CLSP', description='predicted gene 1123', other_designations='CAR-like soluble protein', map_location='9 E3.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=99006965, end_position_on_the_genomic_accession=99035690, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  382105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382105, CurrentID=0, Status='live', Symbol='Fbxw15', Aliases='Fbxo12J', description='F-box and WD-40 domain protein 15', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109552602, end_position_on_the_genomic_accession=109569438, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  382106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382106, CurrentID=0, Status='live', Symbol='Fbxw24', Aliases='EG382106, Gm5162', description='F-box and WD-40 domain protein 24', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109601116, end_position_on_the_genomic_accession=109626059, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  107375 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107375, CurrentID=0, Status='live', Symbol='Slc25a45', Aliases='AW491445', description='solute carrier family 25, member 45', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=5877805, end_position_on_the_genomic_accession=5885768, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  382109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382109, CurrentID=0, Status='live', Symbol='Fbxw26', Aliases='EG382109, Gm5163', description='F-box and WD-40 domain protein 26', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109717566, end_position_on_the_genomic_accession=109746089, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  382111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382111, CurrentID=0, Status='live', Symbol='Susd5', Aliases='Gm1126', description='sushi domain containing 5', other_designations='', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=114057354, end_position_on_the_genomic_accession=114098733, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  382113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382113, CurrentID=0, Status='live', Symbol='Slc22a14', Aliases='Gm1128', description='solute carrier family 22 (organic cation transporter), member 14', other_designations='organic cation transporter like 4', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119169453, end_position_on_the_genomic_accession=119190430, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  382117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382117, CurrentID=0, Status='live', Symbol='Tcaim', Aliases='D9Ertd402e, Gm1129, TOAG-1', description='T cell activation inhibitor, mitochondrial', other_designations='tolerance associated gene-1 protein|tolerance-associated gene 1', map_location='9 73.14 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122805361, end_position_on_the_genomic_accession=122836332, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  382118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382118, CurrentID=0, Status='live', Symbol='Zkscan7', Aliases='BC042569, Zfp167', description='zinc finger with KRAB and SCAN domains 7', other_designations='zinc finger protein 167', map_location='9 F4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122886979, end_position_on_the_genomic_accession=122898619, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  382137 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382137, CurrentID=0, Status='live', Symbol='Fdxacb1', Aliases='D630004A14Rik', description='ferredoxin-fold anticodon binding domain containing 1', other_designations='FDX-ACDB domain-containing protein 1', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=50768238, end_position_on_the_genomic_accession=50772670, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  238455 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238455, CurrentID=0, Status='live', Symbol='Macc1', Aliases='4732474O15Rik, Gm267', description='metastasis associated in colon cancer 1', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=119326285, end_position_on_the_genomic_accession=119466934, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  382156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382156, CurrentID=0, Status='live', Symbol='Fbxw22', Aliases='EG382156, Gm5164', description='F-box and WD-40 domain protein 22', other_designations='', map_location='9 F2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=109378400, end_position_on_the_genomic_accession=109404294, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  382207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382207, CurrentID=0, Status='live', Symbol='Jade3', Aliases='5730598B06, AI851988, AW544806, Phf16, mKIAA0215', description='jade family PHD finger 3', other_designations='PHD finger protein 16|jade family PHD finger protein 3', map_location='X A1.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=20425327, end_position_on_the_genomic_accession=20519939, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  382209 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382209, CurrentID=0, Status='live', Symbol='Rhox3a', Aliases='AY147207, Gm1138, Prx, Rhox3, Rhox3.1', description='reproductive homeobox 3A', other_designations='reproductive homeobox 3|reproductive homeobox on X chromosome, 3|testis expressed homeobox 2', map_location='X A2', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37249919, end_position_on_the_genomic_accession=37258978, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  382217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382217, CurrentID=0, Status='live', Symbol='Gm1140', Aliases='', description='predicted gene 1140', other_designations='', map_location='X A7.1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=67682265, end_position_on_the_genomic_accession=67694458, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  382231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382231, CurrentID=0, Status='live', Symbol='8030474K03Rik', Aliases='Gm1142', description='RIKEN cDNA 8030474K03 gene', other_designations='', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=101794538, end_position_on_the_genomic_accession=101798658, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  382236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382236, CurrentID=0, Status='live', Symbol='Brwd3', Aliases='Brodl, D030064D06Rik, Gm596', description='bromodomain and WD repeat domain containing 3', other_designations='bromo domain-containing protein disrupted in leukemia', map_location='X D', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=108737225, end_position_on_the_genomic_accession=108834360, orientation='minus', exon_count=41, OMIM='', no_hdr0=''),
  382239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382239, CurrentID=0, Status='live', Symbol='Cpxcr1', Aliases='Gm1143', description='CPX chromosome region, candidate 1', other_designations='', map_location='X E1', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=116448944, end_position_on_the_genomic_accession=116478783, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  382245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382245, CurrentID=0, Status='live', Symbol='Tmem29', Aliases='2700081K05Rik, 6330540D07Rik, CXorf44-like', description='transmembrane protein 29', other_designations='', map_location='X F3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=150397773, end_position_on_the_genomic_accession=150467331, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  382252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382252, CurrentID=0, Status='live', Symbol='A830080D01Rik', Aliases='2310067E24Rik, Gm1146', description='RIKEN cDNA A830080D01 gene', other_designations='', map_location='X F4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=159532641, end_position_on_the_genomic_accession=159593081, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  382253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382253, CurrentID=0, Status='live', Symbol='Cdkl5', Aliases='BC038161, Stk9', description='cyclin-dependent kinase-like 5', other_designations='serine/threonine kinase 9', map_location='X 73.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=160784308, end_position_on_the_genomic_accession=160994681, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  382282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382282, CurrentID=0, Status='live', Symbol='Rhox12', Aliases='Gm1148', description='reproductive homeobox 12', other_designations='', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=38104062, end_position_on_the_genomic_accession=38110907, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  382301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382301, CurrentID=0, Status='live', Symbol='Sly', Aliases='', description='Sycp3 like Y-linked', other_designations='', map_location='Y C2', chromosome='Y', genomic_nucleotide_accession_version='NC_000087.7', start_position_on_the_genomic_accession=57187206, end_position_on_the_genomic_accession=57213358, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  382348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382348, CurrentID=0, Status='live', Symbol='Taar8b', Aliases='Gm1149, taR-8b', description='trace amine-associated receptor 8B', other_designations='mTaar8b', map_location='10 A4', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=24091260, end_position_on_the_genomic_accession=24092294, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  382384 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382384, CurrentID=0, Status='live', Symbol='Odf3l2', Aliases='Gm1157', description='outer dense fiber of sperm tails 3-like 2', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79639526, end_position_on_the_genomic_accession=79655710, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  382406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382406, CurrentID=0, Status='live', Symbol='Poc1b', Aliases='4933430F16Rik, Wdr51b', description='POC1 centriolar protein B', other_designations='WD repeat domain 51B|WD repeat-containing protein 51B', map_location='10 C3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=99107171, end_position_on_the_genomic_accession=99198074, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  382407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382407, CurrentID=0, Status='live', Symbol='Gad1-ps', Aliases='Gad-1ps', description='glutamate decarboxylase 1, pseudogene', other_designations='glutamic acid decarboxylase 1, pseudogene', map_location='10 51.14 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=99443838, end_position_on_the_genomic_accession=99447029, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  382423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382423, CurrentID=0, Status='live', Symbol='Atxn7l3b', Aliases='4921506J03Rik, 6230409E21Rik, AI315132, ENSMUSG00000074747', description='ataxin 7-like 3B', other_designations='putative ataxin-7-like protein 3-like protein', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=112925428, end_position_on_the_genomic_accession=112929026, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  382427 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382427, CurrentID=0, Status='live', Symbol='Best3', Aliases='Vmd2l3, mBest4', description='bestrophin 3', other_designations='vitelliform macular dystrophy 2-like 3|vitelliform macular dystrophy 2-like protein 3', map_location='10 D2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=116986314, end_position_on_the_genomic_accession=117025687, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  239691 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239691, CurrentID=0, Status='live', Symbol='AU021092', Aliases='', description='expressed sequence AU021092', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=5211814, end_position_on_the_genomic_accession=5222299, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  382522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382522, CurrentID=0, Status='live', Symbol='Hist3h2bb-ps', Aliases='4930534G10Rik, 4933432H21, B2Hist3h2bb, Hist3h2bb, R75370', description='histone cluster 3, H2bb, pseudogene', other_designations='histone 3, H2bb|histone H2B type 3-B', map_location='11 B2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58954048, end_position_on_the_genomic_accession=58954512, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  382543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382543, CurrentID=0, Status='live', Symbol='Ankfn1', Aliases='Gm1175', description='ankyrin-repeat and fibronectin type III domain containing 1', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=89421012, end_position_on_the_genomic_accession=89696936, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  382551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382551, CurrentID=0, Status='live', Symbol='Cd300lh', Aliases='AY457049, Clm3, Lmir7, clm-3', description='CD300 antigen like family member H', other_designations='CMRF-35-like molecule 3', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115040520, end_position_on_the_genomic_accession=115048419, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  382562 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382562, CurrentID=0, Status='live', Symbol='Pfn4', Aliases='2900024P18Rik', description='profilin family, member 4', other_designations='profilin IV', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=4768913, end_position_on_the_genomic_accession=4778808, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  382571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382571, CurrentID=0, Status='live', Symbol='Kcnf1', Aliases='Gm1182, Kv5.1', description='potassium voltage-gated channel, subfamily F, member 1', other_designations='voltage-gated potassium channel subunit Kv5.1', map_location='12 A1.1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=17172100, end_position_on_the_genomic_accession=17176888, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  235330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=235330, CurrentID=0, Status='live', Symbol='Ttc12', Aliases='E330017O07Rik', description='tetratricopeptide repeat domain 12', other_designations='TPR repeat protein 12', map_location='9 A5.3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=49436961, end_position_on_the_genomic_accession=49486226, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  382620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382620, CurrentID=0, Status='live', Symbol='Tmed8', Aliases='6430595O10Rik, AI447224, Gm1184, Mem1', description='transmembrane emp24 domain containing 8', other_designations='maternal embryonic message 1', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87166242, end_position_on_the_genomic_accession=87200229, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  382639 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382639, CurrentID=0, Status='live', Symbol='Zbtb42', Aliases='EG382639, Gm5188, simiRP58', description='zinc finger and BTB domain containing 42', other_designations='protein simiRP58', map_location='12 F1', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112678604, end_position_on_the_genomic_accession=112682747, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  107483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107483, CurrentID=0, Status='live', Symbol='Av1', Aliases='Av-1', description='Abelson virus susceptibility 1', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  382793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382793, CurrentID=0, Status='live', Symbol='Mtx3', Aliases='4930470O13Rik, AA409304, AI853833, AU067765, EG624619, Gm1194, Gm6514', description='metaxin 3', other_designations='pleckstrin homology domain containing, family M (with RUN domain) member 2', map_location='13 C3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=92844743, end_position_on_the_genomic_accession=92858230, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  238568 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238568, CurrentID=0, Status='live', Symbol='Serpinb6d', Aliases='Gm11390, OTTMUSG00000000712, SPI3D', description='serine (or cysteine) peptidase inhibitor, clade B, member 6d', other_designations='serine (or cysteine) proteinase inhibitor, clade B, member 6d', map_location='13 13.94 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=33661405, end_position_on_the_genomic_accession=33671585, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  382864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382864, CurrentID=0, Status='live', Symbol='Colq', Aliases='A130034K24Rik', description='collagen-like tail subunit (single strand of homotrimer) of asymmetric acetylcholinesterase', other_designations='AChE Q subunit|acetylcholinesterase-associated collagen', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=31523082, end_position_on_the_genomic_accession=31577863, orientation='minus', exon_count=19, OMIM='', no_hdr0=''),
  382867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382867, CurrentID=0, Status='live', Symbol='Zfp488', Aliases='Gm1206, Znf488', description='zinc finger protein 488', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=33967070, end_position_on_the_genomic_accession=33978972, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  12750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=12750, CurrentID=0, Status='live', Symbol='Clk4', Aliases='AI987988, C85119', description='CDC like kinase 4', other_designations='CDC-like kinase 4', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=51263167, end_position_on_the_genomic_accession=51281770, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  382913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382913, CurrentID=0, Status='live', Symbol='Neil2', Aliases='Gm1212, NEH2', description='nei like 2 (E. coli)', other_designations='DNA glycosylase/AP lyase Neil2|DNA-(apurinic or apyrimidinic site) lyase Neil2|endonuclease VIII-like 2|nei homolog 2|nei-like protein 2', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63182445, end_position_on_the_genomic_accession=63194244, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  382985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=382985, CurrentID=0, Status='live', Symbol='Rrm2b', Aliases='p53R2', description='ribonucleotide reductase M2 B (TP53 inducible)', other_designations='TP53-inducible ribonucleotide reductase M2 B|p53-inducible ribonucleotide reductase small subunit 2-like protein', map_location='15 B3.1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=37923952, end_position_on_the_genomic_accession=37961108, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  383103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383103, CurrentID=0, Status='live', Symbol='Tvp23a', Aliases='Fam18a', description='trans-golgi network vesicle protein 23A', other_designations='family with sequence similarity 18, member A', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=10420557, end_position_on_the_genomic_accession=10447388, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  121021 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=121021, CurrentID=0, Status='live', Symbol='Cspg4', Aliases='4732461B14Rik, AN2, NG2', description='chondroitin sulfate proteoglycan 4', other_designations='AN2 proteoglycan|AN2/NG2 proteoglycan|chondroitin sulfate proteoglycan NG2|nerve/glial antigen 2|proteoglycan AN2', map_location='9 B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=56865104, end_position_on_the_genomic_accession=56899870, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  121022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=121022, CurrentID=0, Status='live', Symbol='Mrps6', Aliases='AW046321', description='mitochondrial ribosomal protein S6', other_designations='MRP-S6|S6mt', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=92058336, end_position_on_the_genomic_accession=92112227, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  383229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383229, CurrentID=0, Status='live', Symbol='Hba-ps4', Aliases='EG383229, Gm5226, Hba-4ps, Hba-a4', description='hemoglobin alpha, pseudogene 4', other_designations='predicted pseudogene 5226', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=26286279, end_position_on_the_genomic_accession=26287061, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  383243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383243, CurrentID=0, Status='live', Symbol='Olfr128', Aliases='GA_x5J8B7TWBT8-615-998, GA_x5J8B7W5KGR-591839-591418, GA_x6K02T2PSCP-2374126-2375048, MOR218-13', description='olfactory receptor 128', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37921004, end_position_on_the_genomic_accession=37924549, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  383295 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383295, CurrentID=0, Status='live', Symbol='Ypel5', Aliases='2310076K21Rik, CGI-127', description='yippee-like 5 (Drosophila)', other_designations='yippee protein homolog', map_location='17 E2', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=72836625, end_position_on_the_genomic_accession=72851195, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  383348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383348, CurrentID=0, Status='live', Symbol='Kctd16', Aliases='4930434H12Rik, Gm1267', description='potassium channel tetramerisation domain containing 16', other_designations='', map_location='18 B3', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=40258361, end_position_on_the_genomic_accession=40531184, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  383435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383435, CurrentID=0, Status='live', Symbol='Ms4a14', Aliases='Gm1276, SP3111', description='membrane-spanning 4-domains, subfamily A, member 14', other_designations='', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11301249, end_position_on_the_genomic_accession=11315858, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  383491 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383491, CurrentID=0, Status='live', Symbol='Prdm14', Aliases='', description='PR domain containing 14', other_designations='PR domain-containing 14|PR domain-containing protein 14', map_location='1 A3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=13113428, end_position_on_the_genomic_accession=13127163, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  383548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383548, CurrentID=0, Status='live', Symbol='Serpinb3b', Aliases='Scca2-rs, Scca2-rs1', description='serine (or cysteine) peptidase inhibitor, clade B (ovalbumin), member 3B', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 3B|squamous cell carcinoma antigen 2 related protein 1', map_location='1 E2.1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=107153961, end_position_on_the_genomic_accession=107161114, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  383619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383619, CurrentID=0, Status='live', Symbol='Aim2', Aliases='Gm1313, Ifi210', description='absent in melanoma 2', other_designations='ifi-210|interferon-inducible protein 210|interferon-inducible protein p210', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173349246, end_position_on_the_genomic_accession=173466040, orientation='plus', exon_count=19, OMIM='', no_hdr0=''),
  238722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238722, CurrentID=0, Status='live', Symbol='Zfp72', Aliases='Zfp74', description='zinc finger protein 72', other_designations='zinc finger protein 74', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=74368682, end_position_on_the_genomic_accession=74390861, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  383787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=383787, CurrentID=0, Status='live', Symbol='Ankrd63', Aliases='Gm1337', description='ankyrin repeat domain 63', other_designations='ankrd63|novel ankyrin repeat domain containing protein', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=118699103, end_position_on_the_genomic_accession=118703963, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  252828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252828, CurrentID=0, Status='live', Symbol='Obox4', Aliases='', description='oocyte specific homeobox 4', other_designations='', map_location='', chromosome='7', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  252829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252829, CurrentID=0, Status='live', Symbol='Obox5', Aliases='C87544', description='oocyte specific homeobox 5', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15750370, end_position_on_the_genomic_accession=15759274, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  252830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252830, CurrentID=0, Status='live', Symbol='Obox6', Aliases='D17Ertd599e', description='oocyte specific homeobox 6', other_designations='', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15833250, end_position_on_the_genomic_accession=15839679, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  252837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252837, CurrentID=0, Status='live', Symbol='Ackr4', Aliases='A630091E18Rik, CCBP2, CCR11, CCX-CKR, CCX-CKR1, Ccrl1, Cmkbrl1, PPR1, VSHK1', description='atypical chemokine receptor 4', other_designations='C-C CKR-11|C-C chemokine receptor type 11|CC chemokine receptor-like 1|CC-CKR-11|CCR-11|CCX CKR|chemokine (C-C motif) receptor-like 1|chemokine (C-C) receptor-like 1|chemokine receptor CCR11', map_location='9 F1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=104097682, end_position_on_the_genomic_accession=104126643, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  252838 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252838, CurrentID=0, Status='live', Symbol='Tox', Aliases='1700007F02Rik', description='thymocyte selection-associated high mobility group box', other_designations='thymocyte selection-associated HMG box|thymus high mobility group box protein TOX', map_location='4 A1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=6686353, end_position_on_the_genomic_accession=6991552, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  492881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=492881, CurrentID=0, Status='live', Symbol='Blmpf2', Aliases='', description='bleomycin-induced pulmonary fibrosis 2', other_designations='', map_location='11 32.13 cM', chromosome='11', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  252864 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252864, CurrentID=0, Status='live', Symbol='Dusp15', Aliases='AI851682, LMW-DSP10, T-DSP10', description='dual specificity phosphatase-like 15', other_designations='dual specificity protein phosphatase T-DSP10', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152940995, end_position_on_the_genomic_accession=152952277, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  252866 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252866, CurrentID=0, Status='live', Symbol='Adam34', Aliases='', description='a disintegrin and metallopeptidase domain 34', other_designations='testase 4', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=43650309, end_position_on_the_genomic_accession=43853925, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  252868 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252868, CurrentID=0, Status='live', Symbol='Odf4', Aliases='Oppo1', description='outer dense fiber of sperm tails 4', other_designations='outer dense fiber of sperm tails protein 4|testis-specific protein oppo 1', map_location='11 B3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=68921828, end_position_on_the_genomic_accession=68927490, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  252870 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252870, CurrentID=0, Status='live', Symbol='Usp7', Aliases='2210010O09Rik, AA409944, AA617399, AU019296, AW548146, C80752, Hausp', description='ubiquitin specific peptidase 7', other_designations='deubiquitinating enzyme 7|herpesvirus-associated ubiquitin-specific protease|mHAUSP|ubiquitin specific protease 7|ubiquitin thioesterase 7|ubiquitin thiolesterase 7|ubiquitin-specific-processing protease 7', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=8688722, end_position_on_the_genomic_accession=8757067, orientation='minus', exon_count=34, OMIM='', no_hdr0=''),
  252875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252875, CurrentID=0, Status='live', Symbol='Mios', Aliases='C81488', description='missing oocyte, meiosis regulator, homolog (Drosophila)', other_designations='', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=8209004, end_position_on_the_genomic_accession=8236274, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  252876 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252876, CurrentID=0, Status='live', Symbol='Gin1', Aliases='4930429M06Rik, 8430411H10Rik, Gin-1, Zh2c2', description='gypsy retrotransposon integrase 1', other_designations='zinc finger H2C2 domain-containing protein|zinc finger, H2C2 domain containing', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=97770168, end_position_on_the_genomic_accession=97793710, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  252903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252903, CurrentID=0, Status='live', Symbol='Ap1s3', Aliases='1190009B22, Jr2, [s]3A', description='adaptor-related protein complex AP-1, sigma 3', other_designations='adapter-related protein complex 1 sigma-1C subunit|adapter-related protein complex 1 subunit sigma-1C|adaptor protein complex AP-1 sigma-1C subunit|adaptor protein complex AP-1 subunit sigma-1C|adaptor-related protein complex 1 subunit sigma-1C|clathrin assembly protein complex 1 sigma-1C small chain|golgi adaptor HA1/AP1 adaptin sigma-1C subunit|sigma 1C subunit of AP-1 clathrin|sigma-adaptin 1C|sigma1C-adaptin', map_location='1 C4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=79606876, end_position_on_the_genomic_accession=79672020, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  252904 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252904, CurrentID=0, Status='live', Symbol='Vmn1r69', Aliases='V1re9', description='vomeronasal 1 receptor 69', other_designations='vomernasal 1 receptor Vmn1r69', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=10579756, end_position_on_the_genomic_accession=10581487, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  252905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252905, CurrentID=0, Status='live', Symbol='Vmn1r72', Aliases='V1rg1', description='vomeronasal 1 receptor 72', other_designations='vomernasal 1 receptor Vmn1r72|vomeronasal 1 receptor, G1|vomeronasal type 1 receptor G1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=11669599, end_position_on_the_genomic_accession=11679317, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  252906 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252906, CurrentID=0, Status='live', Symbol='Vmn1r189', Aliases='V1rh2', description='vomeronasal 1 receptor 189', other_designations='vomernasal 1 receptor Vmn1r189|vomeronasal 1 receptor, H2', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22101727, end_position_on_the_genomic_accession=22102665, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  252907 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252907, CurrentID=0, Status='live', Symbol='Vmn1r192', Aliases='V1ri1', description='vomeronasal 1 receptor 192', other_designations='vomernasal 1 receptor Vmn1r192|vomeronasal 1 receptor, I1|vomeronasal type 1 receptor I1', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22187146, end_position_on_the_genomic_accession=22188048, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  252909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252909, CurrentID=0, Status='live', Symbol='Vmn1r85', Aliases='V1rj3', description='vomeronasal 1 receptor 85', other_designations='vomernasal 1 receptor Vmn1r85|vomeronasal 1 receptor, J3|vomeronasal type 1 receptor J3', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=13084289, end_position_on_the_genomic_accession=13085215, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  252910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252910, CurrentID=0, Status='live', Symbol='Vmn1r71', Aliases='V1re13', description='vomeronasal 1 receptor 71', other_designations='vomernasal 1 receptor Vmn1r71', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=10747502, end_position_on_the_genomic_accession=10749538, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  252912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252912, CurrentID=0, Status='live', Symbol='Vmn1r188', Aliases='V1rh17', description='vomeronasal 1 receptor 188', other_designations='vomernasal 1 receptor Vmn1r188|vomeronasal 1 receptor, H17', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22087878, end_position_on_the_genomic_accession=22088804, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  17151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=17151, CurrentID=0, Status='live', Symbol='Ccndbp1', Aliases='AU022347, DIP1, GCIP, Maid, SECC-8, SSEC-8', description='cyclin D-type binding-protein 1', other_designations='grap2 and cyclin-D-interacting protein|maternal Id-like protein|maternal inhibition of differentiation|stage specific embryonic cDNA-8 protein', map_location='2 60.37 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=121008383, end_position_on_the_genomic_accession=121016912, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  384001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384001, CurrentID=0, Status='live', Symbol='Prdx6-ps2', Aliases='Aop2-rs2, GPx*, Prdx6-rs2', description='peroxiredoxin 6 pseudogene 2', other_designations='anti-oxidant protein 2, related sequence 2|peroxiredoxin 5, related sequence 2|peroxiredoxin 6, related sequence 2', map_location='4 10.59 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=25307777, end_position_on_the_genomic_accession=25309139, orientation='plus', exon_count=0, OMIM='', no_hdr0=''),
  384009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384009, CurrentID=0, Status='live', Symbol='Glipr2', Aliases='5730414A08Rik, C77180, GAPR-1', description='GLI pathogenesis-related 2', other_designations='gliPR 2|glioma pathogenesis-related protein 2|golgi-associated PR-1 protein', map_location='4 B1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=43957702, end_position_on_the_genomic_accession=43979118, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  252966 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252966, CurrentID=0, Status='live', Symbol='Cables2', Aliases='ik3-2', description='CDK5 and Abl enzyme substrate 2', other_designations='interactor with CDK3 2', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180258539, end_position_on_the_genomic_accession=180273490, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  252967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252967, CurrentID=0, Status='live', Symbol='Ropn1l', Aliases='ASP, AV047578, Akapasp', description='ropporin 1-like', other_designations='A-kinase anchoring protein-associated sperm protein|AKAP-associated sperm protein', map_location='15 B2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=31441210, end_position_on_the_genomic_accession=31453689, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  252972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252972, CurrentID=0, Status='live', Symbol='Tpcn1', Aliases='5730403B01Rik, Tpc1, mKIAA1169', description='two pore channel 1', other_designations='voltage-dependent calcium channel protein TPC1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120534157, end_position_on_the_genomic_accession=120588699, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  252973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252973, CurrentID=0, Status='live', Symbol='Grhl2', Aliases='0610015A08Rik, BOM, Tcfcp2l3, clft3', description='grainyhead-like 2 (Drosophila)', other_designations='brother of mammalian grainyhead|grainyhead like 2|grainyheadlike|transcription factor CP2-like 3', map_location='15 C', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=37233036, end_position_on_the_genomic_accession=37363569, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  252974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=252974, CurrentID=0, Status='live', Symbol='Tspear', Aliases='C330046G03Rik, ORF65, TSP-EAR, Tnep1', description='thrombospondin type laminin G domain and EAR repeats', other_designations='TSP-EAR|thrombospondin N-terminal domain/EPTP protein 1|thrombospondin-type laminin G domain and EAR repeats protein|tspear', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=77686030, end_position_on_the_genomic_accession=77886994, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  384059 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384059, CurrentID=0, Status='live', Symbol='Tlr12', Aliases='Gm1365, Tlr11', description='toll-like receptor 12', other_designations='', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=128615446, end_position_on_the_genomic_accession=128618619, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  384061 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384061, CurrentID=0, Status='live', Symbol='Fndc5', Aliases='1500001L03Rik, AI836596, C87088, PeP, Pxp', description='fibronectin type III domain containing 5', other_designations='fibronectin type III repeat-containing protein 2|peroxisomal protein', map_location='4 D2.2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=129136922, end_position_on_the_genomic_accession=129144593, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  107702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=107702, CurrentID=0, Status='live', Symbol='Rnh1', Aliases='AW546468, C80305, RNH', description='ribonuclease/angiogenin inhibitor 1', other_designations='', map_location='7 F5', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141160326, end_position_on_the_genomic_accession=141172851, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  384071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384071, CurrentID=0, Status='live', Symbol='Slc25a34', Aliases='Gm1369', description='solute carrier family 25, member 34', other_designations='', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=141618824, end_position_on_the_genomic_accession=141623993, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  384077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384077, CurrentID=0, Status='live', Symbol='Pramel5', Aliases='OTTMUSG00000010540', description='preferentially expressed antigen in melanoma like 5', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=144270633, end_position_on_the_genomic_accession=144280466, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  384185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384185, CurrentID=0, Status='live', Symbol='Arl9', Aliases='1700110A09Rik', description='ADP-ribosylation factor-like 9', other_designations='', map_location='5 C3.3', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=77002805, end_position_on_the_genomic_accession=77010959, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  384198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384198, CurrentID=0, Status='live', Symbol='Fam47e', Aliases='C730043O17, Gm1381', description='family with sequence similarity 47, member E', other_designations='', map_location='5 E2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=92554951, end_position_on_the_genomic_accession=92591285, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  384214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384214, CurrentID=0, Status='live', Symbol='Ephx4', Aliases='Abhd7, Ephxrp, Gm1382', description='epoxide hydrolase 4', other_designations='abhydrolase domain containing 7|abhydrolase domain-containing protein 7|epoxide hydrolase-related protein', map_location='5 E5', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=107403513, end_position_on_the_genomic_accession=107430035, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  238799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=238799, CurrentID=0, Status='live', Symbol='Tnpo1', Aliases='AU021749, D13Ertd688e, IPO2, Kpnb2, MIP, MIP1, TRN', description='transportin 1', other_designations='importin beta-2|karyopherin (importin) beta 2|karyopherin beta-2', map_location='13 52.24 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=98839019, end_position_on_the_genomic_accession=98926384, orientation='minus', exon_count=25, OMIM='', no_hdr0=''),
  384281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384281, CurrentID=0, Status='live', Symbol='Gatc', Aliases='2010003O18Rik', description='glutamyl-tRNA(Gln) amidotransferase, subunit C', other_designations='gatC-like protein|glu-AdT subunit C|glutamyl-tRNA(Gln) amidotransferase, subunit C homolog', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115333242, end_position_on_the_genomic_accession=115341161, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  384309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384309, CurrentID=0, Status='live', Symbol='Trim56', Aliases='A130009K11Rik, Gm452, RNF109', description='tripartite motif-containing 56', other_designations='tripartite motif-containing protein 56', map_location='5 G2', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=137111286, end_position_on_the_genomic_accession=137116466, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  384422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384422, CurrentID=0, Status='live', Symbol='Igk-V8', Aliases='B128L, B55L, IgK', description='immunoglobulin kappa chain variable 8 (V8)', other_designations='Ig kappa light chain|immunoglobulin light chain variable region|mAbB128 specific against human apolipoprotein B-100|mAbB55 specific against human apolipoprotein B-100|monoclonal antibody kappa light chain', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.3', start_position_on_the_genomic_accession=70234759, end_position_on_the_genomic_accession=70235422, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  777654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=777654, CurrentID=0, Status='live', Symbol='Ighj2', Aliases='Gm16879', description='immunoglobulin heavy joining 2', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113429468, end_position_on_the_genomic_accession=113429515, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  384452 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384452, CurrentID=0, Status='live', Symbol='Noto', Aliases='Flh, MmNot, Not, tc', description='notochord homolog (Xenopus laevis)', other_designations='Not homeodomain protein|floating head homolog|truncate', map_location='6 37.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=85423886, end_position_on_the_genomic_accession=85428877, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  777688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=777688, CurrentID=0, Status='live', Symbol='Ighv5-4', Aliases='Gm16971', description='immunoglobulin heavy variable 5-4', other_designations='immunoglobulin heavy variable V5-4', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113597448, end_position_on_the_genomic_accession=113597741, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  20390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20390, CurrentID=0, Status='live', Symbol='Sftpd', Aliases='AI573415, SP-D, Sftp4', description='surfactant associated protein D', other_designations='PSP-D|lung surfactant protein D', map_location='14 22.36 cM', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=41172212, end_position_on_the_genomic_accession=41185198, orientation='minus', exon_count=8, OMIM='', no_hdr0=''),
  384515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384515, CurrentID=0, Status='live', Symbol='Igkv4-68', Aliases='Gm1419, Gm151, Gm636, IgVk, aa4', description='immunoglobulin kappa variable 4-68', other_designations='immunoglobulin kappa chain variable 4-68|immunoglobulin kappa light chain|immunoglobulin light chain variable region|variable region of immunoglobulin kappa light chain', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69304834, end_position_on_the_genomic_accession=69305363, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  384522 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384522, CurrentID=0, Status='live', Symbol='Vmn1r55', Aliases='Gm1422, Gm403, V1rd5', description='vomeronasal 1 receptor 55', other_designations='vomernasal 1 receptor Vmn1r55|vomeronasal 1 receptor, D5', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5146490, end_position_on_the_genomic_accession=5147422, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  384557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384557, CurrentID=0, Status='live', Symbol='Ceacam3', Aliases='EG384557, Psg24, cea12', description='carcinoembryonic antigen-related cell adhesion molecule 3', other_designations='pregnancy-specific glycoprotein 24', map_location='7 A2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=17150195, end_position_on_the_genomic_accession=17163754, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  384569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384569, CurrentID=0, Status='live', Symbol='Nova2', Aliases='Gm1424', description='neuro-oncological ventral antigen 2', other_designations='', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=18925888, end_position_on_the_genomic_accession=18965319, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  384572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384572, CurrentID=0, Status='live', Symbol='Vmn1r177', Aliases='V1rd12', description='vomeronasal 1 receptor 177', other_designations='vomernasal 1 receptor Vmn1r177|vomeronasal receptor V1RD12', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=23865520, end_position_on_the_genomic_accession=23866449, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  384585 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384585, CurrentID=0, Status='live', Symbol='Scgb1b3', Aliases='Abpa21, Abpa3, EG384585, Gm5325', description='secretoglobin, family 1B, member 3', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=31375592, end_position_on_the_genomic_accession=31376916, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  384589 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384589, CurrentID=0, Status='live', Symbol='Scgb1b29', Aliases='Abpa29, Abpa29_a7, EG384589, Gm5326', description='secretoglobin, family 1B, member 29', other_designations='', map_location='7 B1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=32441541, end_position_on_the_genomic_accession=32442844, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  544696 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544696, CurrentID=0, Status='live', Symbol='Tbc1d32', Aliases='Bromi, D630037F22Rik, b2b2284Clo', description='TBC1 domain family, member 32', other_designations='TBC1 domain family member 32', map_location='10 B3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=56014293, end_position_on_the_genomic_accession=56229456, orientation='minus', exon_count=35, OMIM='', no_hdr0=''),
  64099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64099, CurrentID=0, Status='live', Symbol='Parvg', Aliases='AI413459', description='parvin, gamma', other_designations='', map_location='15 E2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=84323923, end_position_on_the_genomic_accession=84342978, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  384619 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384619, CurrentID=0, Status='live', Symbol='Ccdc155', Aliases='Gm1434, KASH5', description='coiled-coil domain containing 155', other_designations='KASH domain-containing protein 5|coiled-coil domain-containing protein 155', map_location='7 B4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45183619, end_position_on_the_genomic_accession=45204892, orientation='minus', exon_count=21, OMIM='', no_hdr0=''),
  20422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20422, CurrentID=0, Status='live', Symbol='Shfm1', Aliases='DSS1, Shfdg1, Shfg', description='split hand/foot malformation (ectrodactyly) type 1', other_designations='deleted in split hand/split foot protein 1 homolog|split hand/foot deleted gene 1|split hand/foot deleted protein 1 homolog|split hand/foot malformation type 1 protein homolog', map_location='6 A2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=6558275, end_position_on_the_genomic_accession=6578658, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  384701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384701, CurrentID=0, Status='live', Symbol='Usp17ld', Aliases='Dub2a, Dub4', description='ubiquitin specific peptidase 17-like D', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103250086, end_position_on_the_genomic_accession=103252505, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  544717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544717, CurrentID=0, Status='live', Symbol='1190007I07Rik', Aliases='', description='RIKEN cDNA 1190007I07 gene', other_designations='', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=82619851, end_position_on_the_genomic_accession=82623228, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  384763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384763, CurrentID=0, Status='live', Symbol='Zfp667', Aliases='A830025F02Rik, Mip1, ZNF667', description='zinc finger protein 667', other_designations='myocardial ischemic preconditioning up-regulated protein 1', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6286183, end_position_on_the_genomic_accession=6307883, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  384783 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384783, CurrentID=0, Status='live', Symbol='Irs2', Aliases='Irs-2', description='insulin receptor substrate 2', other_designations='4PS', map_location='8 5.35 cM', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=10986961, end_position_on_the_genomic_accession=11008430, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  384806 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384806, CurrentID=0, Status='live', Symbol='Adam20', Aliases='4930529F22Rik, Adam38', description='a disintegrin and metallopeptidase domain 20', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40793273, end_position_on_the_genomic_accession=40797342, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  195208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195208, CurrentID=0, Status='live', Symbol='Dcdc2a', Aliases='AW492955, Dcdc2, RU2', description='doublecortin domain containing 2a', other_designations='', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=25020435, end_position_on_the_genomic_accession=25210706, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  675812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=675812, CurrentID=0, Status='live', Symbol='Zfp605', Aliases='A830023I12Rik', description='zinc finger protein 605', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=110110022, end_position_on_the_genomic_accession=110130509, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  544752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544752, CurrentID=0, Status='live', Symbol='Tug1', Aliases='AI316828', description='taurine upregulated gene 1', other_designations='', map_location='11 A1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=3639785, end_position_on_the_genomic_accession=3648814, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  20475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20475, CurrentID=0, Status='live', Symbol='Six5', Aliases='Dmahp, MDMAHP, TrexBF', description='sine oculis-related homeobox 5', other_designations='DM locus-associated homeodomain protein homolog|sine oculis homeobox homolog 5|sine oculis-related homeobox 5 homolog', map_location='7 9.46 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=19094544, end_position_on_the_genomic_accession=19098549, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  384997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=384997, CurrentID=0, Status='live', Symbol='Pglyrp4', Aliases='Gm1478, Gm413', description='peptidoglycan recognition protein 4', other_designations='PGLYRPIbeta|PGRP-I-beta|peptidoglycan recognition protein I-beta|peptidoglycan recognition protein intermediate beta', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=90726064, end_position_on_the_genomic_accession=90741517, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  105242658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105242658, CurrentID=0, Status='live', Symbol='Hotairm1', Aliases='Gm15051, Hoxaas1, OTTMUSG00000018870', description='Hoxa transcript antisense RNA, myeloid-specific 1', other_designations='', map_location='6 25.4 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=52158372, end_position_on_the_genomic_accession=52162307, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  385109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385109, CurrentID=0, Status='live', Symbol='Igkv4-72', Aliases='Gm1499, am4', description='immunoglobulin kappa chain variable 4-72', other_designations='anti-human CD20 antibody 1F5 kappa light chain variable region|anti-lipoteichoic acid light chain variable region|anti-myosin immunoglobulin light chain variable region|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69226854, end_position_on_the_genomic_accession=69227383, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  385120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385120, CurrentID=0, Status='live', Symbol='Igkv4-70', Aliases='BH151, Gm1502, kn4', description='immunoglobulin kappa chain variable 4-70', other_designations='anti-MOG Z12 variable light chain|anti-MOG kappa|immunoglobulin light chain variable region', map_location='6 C1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=69267888, end_position_on_the_genomic_accession=69268412, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  20503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20503, CurrentID=0, Status='live', Symbol='Slc16a7', Aliases='4921534N07Rik, Mct2', description='solute carrier family 16 (monocarboxylic acid transporters), member 7', other_designations='MCT 2|solute carrier family 16 member 7', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=125219284, end_position_on_the_genomic_accession=125389562, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  20518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20518, CurrentID=0, Status='live', Symbol='Slc22a2', Aliases='Oct2, Orct2', description='solute carrier family 22 (organic cation transporter), member 2', other_designations='organic cation transporter 2|solute carrier family 22, member 2', map_location='17 8.61 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12583958, end_position_on_the_genomic_accession=12628488, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  20519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20519, CurrentID=0, Status='live', Symbol='Slc22a3', Aliases='EMT, Oct3, Orct3, Slca22a3', description='solute carrier family 22 (organic cation transporter), member 3', other_designations='organic cation transporter 3|solute carrier family 22, member 3', map_location='17 8.52 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=12419972, end_position_on_the_genomic_accession=12507704, orientation='minus', exon_count=11, OMIM='', no_hdr0=''),
  385263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385263, CurrentID=0, Status='live', Symbol='Gm1527', Aliases='Gm1136', description='predicted gene 1527', other_designations='', map_location='3 A3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=28892617, end_position_on_the_genomic_accession=28926724, orientation='plus', exon_count=16, OMIM='', no_hdr0=''),
  282665 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=282665, CurrentID=0, Status='live', Symbol='Serpinb1-ps1', Aliases='EID, ovalbumin', description='serine (or cysteine) peptidase inhibitor, clade B, member 1, pseudogene', other_designations='serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 1, pseudogene|serine (or cysteine) proteinase inhibitor, clade B (ovalbumin), member 1a pseudogene|serine (or cysteine) proteinase inhibitor, clade B, member 1, pseudogene', map_location='13 13.76 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=32859948, end_position_on_the_genomic_accession=32863178, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  544817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544817, CurrentID=0, Status='live', Symbol='Arhgap27', Aliases='2310069I04Rik, 5730442P18Rik, Gm21552, Sh3d20', description='Rho GTPase activating protein 27', other_designations='CIN85-associated multi-domain-containing Rho GTPase-activating protein 1|SH3 domain containing 20|rho-type GTPase-activating protein 27', map_location='11 E1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=103331484, end_position_on_the_genomic_accession=103363692, orientation='minus', exon_count=20, OMIM='', no_hdr0=''),
  385343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385343, CurrentID=0, Status='live', Symbol='Rhox1', Aliases='', description='reproductive homeobox 1', other_designations='reproductive homeobox on X chromosome 1|reproductive homeobox on chromosome X, 1', map_location='X A3.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=37213804, end_position_on_the_genomic_accession=37222258, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  385354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385354, CurrentID=0, Status='live', Symbol='Frmd7', Aliases='EG665849, Gm1533', description='FERM domain containing 7', other_designations='frmd7', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=50892645, end_position_on_the_genomic_accession=50943619, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  385377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385377, CurrentID=0, Status='live', Symbol='Pnma5', Aliases='mKIAA1934', description='paraneoplastic antigen family 5', other_designations='', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=73033981, end_position_on_the_genomic_accession=73040203, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  385643 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385643, CurrentID=0, Status='live', Symbol='Kng2', Aliases='', description='kininogen 2', other_designations='Kininogen-II|kininogen II', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=22985852, end_position_on_the_genomic_accession=23029101, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  385658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385658, CurrentID=0, Status='live', Symbol='Nxpe3', Aliases='Fam55c, Gm1752, Gm611', description='neurexophilin and PC-esterase domain family, member 3', other_designations='family with sequence similarity 55, member C', map_location='16 C1.1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=55839953, end_position_on_the_genomic_accession=55896773, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  385668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385668, CurrentID=0, Status='live', Symbol='Lca5l', Aliases='4921526F01Rik, Lcca5l', description='Leber congenital amaurosis 5-like', other_designations='lca5l|leber congenital amaurosis 5-like protein', map_location='16 C4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=96158405, end_position_on_the_genomic_accession=96193993, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  385674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=385674, CurrentID=0, Status='live', Symbol='Zfp174', Aliases='Gm1754', description='zinc finger protein 174', other_designations='', map_location='16 A1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3847190, end_position_on_the_genomic_accession=3872051, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  20608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20608, CurrentID=0, Status='live', Symbol='Sstr4', Aliases='Smstr4, sst4', description='somatostatin receptor 4', other_designations='SS-4-R|SS4-R|SS4R', map_location='2 73.44 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=148395377, end_position_on_the_genomic_accession=148396764, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  544905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544905, CurrentID=0, Status='live', Symbol='LOC544905', Aliases='', description='Ig heavy chain V region', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.3', start_position_on_the_genomic_accession=111662055, end_position_on_the_genomic_accession=111662556, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  108011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108011, CurrentID=0, Status='live', Symbol='Ap4e1', Aliases='2310033A20Rik, 9930028M04Rik, AV087807', description='adaptor-related protein complex AP-4, epsilon 1', other_designations='AP-4 adapter complex subunit epsilon|AP-4 adaptor complex subunit epsilon|adapter-related protein complex 4 subunit epsilon-1|adaptor-related protein complex 4 subunit epsilon-1|ap4e1|epsilon subunit of AP-4|epsilon-adaptin', map_location='2 61.76 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=127008679, end_position_on_the_genomic_accession=127069814, orientation='plus', exon_count=24, OMIM='', no_hdr0=''),
  108014 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108014, CurrentID=0, Status='live', Symbol='Srsf9', Aliases='25kDa, 2610029M16Rik, SRp30c, Sfrs9', description='serine/arginine-rich splicing factor 9', other_designations='splicing factor, arginine/serine-rich 9', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=115327161, end_position_on_the_genomic_accession=115333084, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  105243590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=105243590, CurrentID=0, Status='live', Symbol='LOC105243590', Aliases='IgH, Igg1', description='Ig heavy chain Mem5-like', other_designations='', map_location='', chromosome='12', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  239096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239096, CurrentID=0, Status='live', Symbol='Cdh24', Aliases='1700040A22Rik, ENSMUSG00000022188', description='cadherin-like 24', other_designations='EY-cadherin', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54631992, end_position_on_the_genomic_accession=54641364, orientation='minus', exon_count=13, OMIM='', no_hdr0=''),
  239102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239102, CurrentID=0, Status='live', Symbol='Zfhx2', Aliases='ZFH-5', description='zinc finger homeobox 2', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55060265, end_position_on_the_genomic_accession=55092433, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  195434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195434, CurrentID=0, Status='live', Symbol='Utp14b', Aliases='4932411L21Rik, jsd', description='UTP14, U3 small nucleolar ribonucleoprotein, homolog B (yeast)', other_designations='juvenile spermatogonial depletion protein', map_location='1 40.84 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=78657825, end_position_on_the_genomic_accession=78667601, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  544963 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=544963, CurrentID=0, Status='live', Symbol='Iqgap2', Aliases='4933417J23Rik, A630053O10, AI788777', description='IQ motif containing GTPase activating protein 2', other_designations='', map_location='13 50.26 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=95627177, end_position_on_the_genomic_accession=95891969, orientation='minus', exon_count=37, OMIM='', no_hdr0=''),
  239134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239134, CurrentID=0, Status='live', Symbol='Gucy1b2', Aliases='B930025B03, GC-SB2', description='guanylate cyclase 1, soluble, beta 2', other_designations='soluble guanylyl cyclase beta2 subunit', map_location='14 D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62373023, end_position_on_the_genomic_accession=62457562, orientation='minus', exon_count=22, OMIM='', no_hdr0=''),
  20683 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20683, CurrentID=0, Status='live', Symbol='Sp1', Aliases='1110003E12Rik, AA450830, AI845540-1, Sp1', description='trans-acting transcription factor 1', other_designations='', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102406316, end_position_on_the_genomic_accession=102436404, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  239170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239170, CurrentID=0, Status='live', Symbol='Fam160b2', Aliases='G430067P06Rik, Rai16', description='family with sequence similarity 160, member B2', other_designations='retinoic acid induced 16|retinoic acid-induced protein 16', map_location='14 D2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=70583293, end_position_on_the_genomic_accession=70599835, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  386454 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386454, CurrentID=0, Status='live', Symbol='Rnf39', Aliases='BC038661, Gm1786, Gm1787, Gm320, LIRF', description='ring finger protein 39', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=36943051, end_position_on_the_genomic_accession=36947986, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  545007 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545007, CurrentID=0, Status='live', Symbol='Gm5796', Aliases='', description='predicted gene 5796', other_designations='alpha41-takusan', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=4018034, end_position_on_the_genomic_accession=4027619, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  386463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386463, CurrentID=0, Status='live', Symbol='Cdsn', Aliases='AI747712', description='corneodesmosin', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=35543082, end_position_on_the_genomic_accession=35557180, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  108115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108115, CurrentID=0, Status='live', Symbol='Slco4a1', Aliases='OATP-E, Slc21a12', description='solute carrier organic anion transporter family, member 4a1', other_designations='organic anion-transporting polypeptide E|sodium-independent organic anion transporter E|solute carrier family 21 (organic anion transporter), member 12|solute carrier family 21 member 12', map_location='2 H4', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180456250, end_position_on_the_genomic_accession=180474853, orientation='plus', exon_count=15, OMIM='', no_hdr0=''),
  239188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239188, CurrentID=0, Status='live', Symbol='Enox1', Aliases='B230207J08, D230005D02Rik', description='ecto-NOX disulfide-thiol exchanger 1', other_designations='cNOX|constitutive Ecto-NOX', map_location='14 D3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=77154001, end_position_on_the_genomic_accession=77721763, orientation='plus', exon_count=23, OMIM='', no_hdr0=''),
  545030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545030, CurrentID=0, Status='live', Symbol='Wdfy4', Aliases='Gm18810', description='WD repeat and FYVE domain containing 4', other_designations='', map_location='14 B', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=32959547, end_position_on_the_genomic_accession=33193855, orientation='minus', exon_count=67, OMIM='', no_hdr0=''),
  386611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386611, CurrentID=0, Status='live', Symbol='Rnf133', Aliases='Greul2', description='ring finger protein 133', other_designations='goliath-related E3 ubiquitin-protein ligase 2', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=23648869, end_position_on_the_genomic_accession=23650305, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  386612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386612, CurrentID=0, Status='live', Symbol='Thoc6', Aliases='F830014G06Rik, Wdr58', description='THO complex 6 homolog (Drosophila)', other_designations='WD repeat domain 58|WD repeat-containing protein 58', map_location='17 A3.3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23668618, end_position_on_the_genomic_accession=23673903, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  386626 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386626, CurrentID=0, Status='live', Symbol='Zfp640', Aliases='Mzf6d', description='zinc finger protein 640', other_designations='Kruppel associated box containing', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.5', start_position_on_the_genomic_accession=66773876, end_position_on_the_genomic_accession=66778625, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  386649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386649, CurrentID=0, Status='live', Symbol='Nsfl1c', Aliases='Munc-18c, Stxbp3a, p47', description='NSFL1 (p97) cofactor (p47)', other_designations='p97 cofactor p47', map_location='2 G3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=151494182, end_position_on_the_genomic_accession=151511310, orientation='plus', exon_count=10, OMIM='', no_hdr0=''),
  386655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386655, CurrentID=0, Status='live', Symbol='Eid2', Aliases='Cri2, EID-2', description='EP300 interacting inhibitor of differentiation 2', other_designations='CREBBP/EP300 inhibitor 2|CREBBP/EP300 inhibitory protein 2|E1A like inhibitor of differentiation 2|EID-1-like inhibitor of differentiation 2|eid2', map_location='7 A3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=28267881, end_position_on_the_genomic_accession=28269168, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  64450 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=64450, CurrentID=0, Status='live', Symbol='Gpr85', Aliases='2900026B03Rik', description='G protein-coupled receptor 85', other_designations='super conserved receptor expressed in brain 2', map_location='6 A1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=13834454, end_position_on_the_genomic_accession=13839937, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  545055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545055, CurrentID=0, Status='live', Symbol='Cma2', Aliases='Mcp10', description='chymase 2, mast cell', other_designations='', map_location='14 C3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55950806, end_position_on_the_genomic_accession=55974031, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  386750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386750, CurrentID=0, Status='live', Symbol='Slitrk3', Aliases='', description='SLIT and NTRK-like family, member 3', other_designations='slit and trk like gene 3', map_location='3 E2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=73047247, end_position_on_the_genomic_accession=73057881, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  386753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=386753, CurrentID=0, Status='live', Symbol='Dbpht2', Aliases='Spot2', description='DNA binding protein with his-thr domain', other_designations='', map_location='12 C3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=74297474, end_position_on_the_genomic_accession=74300468, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387132 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387132, CurrentID=0, Status='live', Symbol='Ssxb2', Aliases='', description='synovial sarcoma, X member B, breakpoint 2', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8454343, end_position_on_the_genomic_accession=8461726, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  387133 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387133, CurrentID=0, Status='live', Symbol='Mir9-1', Aliases='Mirn9, Mirn9-1, mir-9-1, mmu-mir-9-1', description='microRNA 9-1', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=88215598, end_position_on_the_genomic_accession=88215686, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387134 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387134, CurrentID=0, Status='live', Symbol='Mir16-1', Aliases='Mirn16, Mirn16-1, miR-16, mir-16-1', description='microRNA 16-1', other_designations='mmu-mir-16-1', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61631880, end_position_on_the_genomic_accession=61631972, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387135, CurrentID=0, Status='live', Symbol='Mir18', Aliases='Mirn18, mir-18a, mmu-mir-18, mmu-mir-18a', description='microRNA 18', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115043851, end_position_on_the_genomic_accession=115043946, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387136, CurrentID=0, Status='live', Symbol='Mir1a-1', Aliases='Mir1-1, Mirn1-1, Mirn1b, Mirn1c, Mirn1d, mir-1a-1, mmu-mir-1-1, mmu-mir-1a-1', description='microRNA 1a-1', other_designations='microRNA 1-1|microRNA 1b|microRNA 1c|microRNA 1d', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=180389048, end_position_on_the_genomic_accession=180389124, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387139 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387139, CurrentID=0, Status='live', Symbol='Mir20a', Aliases='Mir-20, Mirn20, Mirn20a, mir-20a, mmu-mir-20a', description='microRNA 20a', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=115044157, end_position_on_the_genomic_accession=115044263, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387140 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387140, CurrentID=0, Status='live', Symbol='Mir21a', Aliases='Mir21, Mirn21, mmu-mir-21, mmu-mir-21a', description='microRNA 21a', other_designations='microRNA 21', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=86584067, end_position_on_the_genomic_accession=86584158, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387141 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387141, CurrentID=0, Status='live', Symbol='Mir22', Aliases='Mirn22, mmu-mir-22', description='microRNA 22', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75463716, end_position_on_the_genomic_accession=75463810, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387142 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387142, CurrentID=0, Status='live', Symbol='Mir24-1', Aliases='Mirn189, Mirn24, Mirn24-1, miR-189, miR-24, miR-24-1', description='microRNA 24-1', other_designations='microRNA 189|mmu-mir-24-1', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=63301208, end_position_on_the_genomic_accession=63301275, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387143, CurrentID=0, Status='live', Symbol='Mir101a', Aliases='Mirn101, Mirn101a, miR-101, mir-101a', description='microRNA 101a', other_designations='mmu-mir-101a', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101346945, end_position_on_the_genomic_accession=101347027, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387144, CurrentID=0, Status='live', Symbol='Mir10b', Aliases='Mirn10b, mir-10b, mmu-mir-10b', description='microRNA 10b', other_designations='', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=74726070, end_position_on_the_genomic_accession=74726137, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387145, CurrentID=0, Status='live', Symbol='Mir126a', Aliases='Mir126, Mirn126, mir-126a, mmu-mir-126, mmu-mir-126a', description='microRNA 126a', other_designations='microRNA 126', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=26591357, end_position_on_the_genomic_accession=26591429, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387146 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387146, CurrentID=0, Status='live', Symbol='Mir127', Aliases='Mirn127, mir-127, mmu-mir-127', description='microRNA 127', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109592846, end_position_on_the_genomic_accession=109592915, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387147, CurrentID=0, Status='live', Symbol='Mir128-1', Aliases='Mirn128, Mirn128-1, Mirn128a, mir-128-1', description='microRNA 128-1', other_designations='microRNA 128a|mmu-mir-128-1|mmu-mir-128a', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=128202361, end_position_on_the_genomic_accession=128202430, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387148 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387148, CurrentID=0, Status='live', Symbol='Mir129', Aliases='Mirn129', description='microRNA 129', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387149 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387149, CurrentID=0, Status='live', Symbol='Mir130a', Aliases='Mirn130, Mirn130a, mir-130a', description='microRNA 130a', other_designations='mmu-mir-130a', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=84741115, end_position_on_the_genomic_accession=84741178, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387150, CurrentID=0, Status='live', Symbol='Mir132', Aliases='Mirn132, mir-132, mmu-mir-132', description='microRNA 132', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75173682, end_position_on_the_genomic_accession=75173747, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387151, CurrentID=0, Status='live', Symbol='Mir133a-1', Aliases='Mirn133, Mirn133a-1, mir-133a-1', description='microRNA 133a-1', other_designations='mmu-mir-133a-1', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=10782909, end_position_on_the_genomic_accession=10782976, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387152 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387152, CurrentID=0, Status='live', Symbol='Mir134', Aliases='Mirn134, mir-134, mmu-mir-134', description='microRNA 134', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109734139, end_position_on_the_genomic_accession=109734209, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387153, CurrentID=0, Status='live', Symbol='Mir135a-1', Aliases='Mir135a01, Mirn135, Mirn135a-1, mir-135a-1', description='microRNA 135a-1', other_designations='mmu-mir-135a-1', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106154124, end_position_on_the_genomic_accession=106154213, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387154, CurrentID=0, Status='live', Symbol='Mir136', Aliases='Mirn136, mir-136, mmu-mir-136', description='microRNA 136', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109595327, end_position_on_the_genomic_accession=109595388, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387155 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387155, CurrentID=0, Status='live', Symbol='Mir137', Aliases='Mirn137, mir-137, mmu-mir-137', description='microRNA 137', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=118433857, end_position_on_the_genomic_accession=118433929, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387156, CurrentID=0, Status='live', Symbol='Mir138-1', Aliases='Mirn138, Mirn138-1, mir-138-1', description='microRNA 138-1', other_designations='mmu-mir-138-1', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=122682876, end_position_on_the_genomic_accession=122682974, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387157 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387157, CurrentID=0, Status='live', Symbol='Mir139', Aliases='Mirn139, mir-139, mmu-mir-139', description='microRNA 139', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=101475376, end_position_on_the_genomic_accession=101475443, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387158 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387158, CurrentID=0, Status='live', Symbol='Mir140', Aliases='Mirn140, mir-140, mmu-mir-140', description='microRNA 140', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=107551244, end_position_on_the_genomic_accession=107551313, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387159, CurrentID=0, Status='live', Symbol='Mir141', Aliases='Mirn141, mir-141, mmu-mir-141', description='microRNA 141', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=124717914, end_position_on_the_genomic_accession=124717985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387160, CurrentID=0, Status='live', Symbol='Mir142', Aliases='Mirn142, miR-142, mir-142a, mmu-mir-142, mmu-mir-142a', description='microRNA 142', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87756864, end_position_on_the_genomic_accession=87756927, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387161, CurrentID=0, Status='live', Symbol='Mir143', Aliases='Mirn143, mir-143, mmu-mir-143', description='microRNA 143', other_designations='', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61649196, end_position_on_the_genomic_accession=61649258, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387162, CurrentID=0, Status='live', Symbol='Mir144', Aliases='Mirn144, mir-144, mmu-mir-144', description='microRNA 144', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=78073005, end_position_on_the_genomic_accession=78073070, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387163, CurrentID=0, Status='live', Symbol='Mir145a', Aliases='Mir145, Mirn145, mir-145a, mmu-mir-145, mmu-mir-145a', description='microRNA 145a', other_designations='microRNA 145', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61647825, end_position_on_the_genomic_accession=61647894, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387164 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387164, CurrentID=0, Status='live', Symbol='Mir146', Aliases='Mirn146, miR-146a, mmu-mir-146', description='microRNA 146', other_designations='mmu-mir-146a', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=43374397, end_position_on_the_genomic_accession=43374461, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387165, CurrentID=0, Status='live', Symbol='Mir147', Aliases='Mirn147, mmu-mir-147', description='microRNA 147', other_designations='microRNA mir-147', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=122640803, end_position_on_the_genomic_accession=122640881, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387166, CurrentID=0, Status='live', Symbol='Mir148a', Aliases='Mirn148, Mirn148a, mir-148a', description='microRNA 148a', other_designations='mmu-mir-148a', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=51269812, end_position_on_the_genomic_accession=51269910, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387167, CurrentID=0, Status='live', Symbol='Mir149', Aliases='Mirn149, mir-149, mmu-mir-149', description='microRNA 149', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92850378, end_position_on_the_genomic_accession=92850443, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387168, CurrentID=0, Status='live', Symbol='Mir150', Aliases='Mirn150, mir-150, mmu-mir-150', description='microRNA 150', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=45121757, end_position_on_the_genomic_accession=45121821, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387169 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387169, CurrentID=0, Status='live', Symbol='Mir151', Aliases='Mirn151, mir-151, mmu-mir-151', description='microRNA 151', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=73254815, end_position_on_the_genomic_accession=73254882, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387170 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387170, CurrentID=0, Status='live', Symbol='Mir152', Aliases='Mirn152, mir-152, mmu-mir-152', description='microRNA 152', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96850393, end_position_on_the_genomic_accession=96850465, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387171 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387171, CurrentID=0, Status='live', Symbol='Mir153', Aliases='Mirn153, mir-153, mmu-mir-153', description='microRNA 153', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=117250817, end_position_on_the_genomic_accession=117250885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387172, CurrentID=0, Status='live', Symbol='Mir154', Aliases='Mirn154, mir-154, mmu-mir-154', description='microRNA 154', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=109738433, end_position_on_the_genomic_accession=109738498, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387173, CurrentID=0, Status='live', Symbol='Mir155', Aliases='Mirn155, mir-155, mmu-mir-155', description='microRNA 155', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=84714140, end_position_on_the_genomic_accession=84714204, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387174 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387174, CurrentID=0, Status='live', Symbol='Mir15a', Aliases='Mirn15a, mir-15a, mmu-mir-15a', description='microRNA 15a', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=61632027, end_position_on_the_genomic_accession=61632110, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387175 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387175, CurrentID=0, Status='live', Symbol='Mir15b', Aliases='Mirn15b, mir-15b, mmu-mir-15b', description='microRNA 15b', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=69009772, end_position_on_the_genomic_accession=69009835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387176 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387176, CurrentID=0, Status='live', Symbol='Mir181a-2', Aliases='Mirn181, Mirn181a, Mirn181a-2, miR-181, mir-181a, mir-181a-2', description='microRNA 181a-2', other_designations='microRNA mir-181a-2|mmu-mir-181a|mmu-mir-181a-2', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=38852735, end_position_on_the_genomic_accession=38852810, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387177 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387177, CurrentID=0, Status='live', Symbol='Mir182', Aliases='Mirn182, mir-182, mmu-mir-182', description='microRNA 182', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30165918, end_position_on_the_genomic_accession=30165992, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387178 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387178, CurrentID=0, Status='live', Symbol='Mir183', Aliases='Mirn183, mir-183, mmu-mir-183', description='microRNA 183', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=30169668, end_position_on_the_genomic_accession=30169737, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387179 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387179, CurrentID=0, Status='live', Symbol='Mir184', Aliases='Mirn184, mir-184, mmu-mir-184', description='microRNA 184', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=89802260, end_position_on_the_genomic_accession=89802328, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387180 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387180, CurrentID=0, Status='live', Symbol='Mir185', Aliases='Mirn185, mir-185, mmu-mir-185', description='microRNA 185', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=18327401, end_position_on_the_genomic_accession=18327465, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387181 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387181, CurrentID=0, Status='live', Symbol='Mir186', Aliases='Mirn186, mir-186, mmu-mir-186', description='microRNA 186', other_designations='', map_location='3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=157544279, end_position_on_the_genomic_accession=157544349, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387182 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387182, CurrentID=0, Status='live', Symbol='Mir187', Aliases='Mirn187, mir-187, mmu-mir-187', description='microRNA 187', other_designations='', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=24429110, end_position_on_the_genomic_accession=24429170, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387183 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387183, CurrentID=0, Status='live', Symbol='Mir188', Aliases='Mirn188, mir-188, mmu-mir-188', description='microRNA 188', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=7247989, end_position_on_the_genomic_accession=7248056, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387185, CurrentID=0, Status='live', Symbol='Mir190a', Aliases='Mir190, Mirn190, mir-190, mir-190a, mmu-mir-190, mmu-mir-190a', description='microRNA 190a', other_designations='microRNA 190', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=67236660, end_position_on_the_genomic_accession=67236726, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387186 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387186, CurrentID=0, Status='live', Symbol='Mir191', Aliases='Mirn191, mir-191, mmu-mir-191', description='microRNA 191', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=108568319, end_position_on_the_genomic_accession=108568392, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387187 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387187, CurrentID=0, Status='live', Symbol='Mir192', Aliases='Mirn192, mir-192, mmu-mir-192', description='microRNA 192', other_designations='', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=6264844, end_position_on_the_genomic_accession=6264932, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387188 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387188, CurrentID=0, Status='live', Symbol='Mir193a', Aliases='Mir193, Mirn193, mir-193a, mmu-mir-193, mmu-mir-193a', description='microRNA 193a', other_designations='microRNA 193', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=79711969, end_position_on_the_genomic_accession=79712034, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387189, CurrentID=0, Status='live', Symbol='Mir194-1', Aliases='Mirn194, Mirn194-1, miR-194, mir-194-1', description='microRNA 194-1', other_designations='mmu-mir-194-1', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=185313319, end_position_on_the_genomic_accession=185313385, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387190 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387190, CurrentID=0, Status='live', Symbol='Mir195a', Aliases='Mir195, Mirn195, mir-195a, mmu-mir-195, mmu-mir-195a', description='microRNA 195a', other_designations='microRNA 195', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=70235042, end_position_on_the_genomic_accession=70235135, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387191 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387191, CurrentID=0, Status='live', Symbol='Mir196a-1', Aliases='Mirn196, Mirn196a-1, miR-196, mir-196a-1', description='microRNA 196a-1', other_designations='mmu-mir-196a-1', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=96265164, end_position_on_the_genomic_accession=96265265, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387194 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387194, CurrentID=0, Status='live', Symbol='Mir199a-1', Aliases='Mirn199, Mirn199a-1, mir-199a-1', description='microRNA 199a-1', other_designations='mmu-mir-199a-1', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=21496495, end_position_on_the_genomic_accession=21496564, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387195 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387195, CurrentID=0, Status='live', Symbol='Mir19b-2', Aliases='Mirn19b, Mirn19b-2, miR-19b, mir-19b-2', description='microRNA 19b-2', other_designations='mmu-mir-19b-2', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=52741983, end_position_on_the_genomic_accession=52742066, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387197 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387197, CurrentID=0, Status='live', Symbol='Mir201', Aliases='Mirn201, mir-201, mmu-mir-201', description='microRNA 201', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=67988096, end_position_on_the_genomic_accession=67988161, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387198, CurrentID=0, Status='live', Symbol='Mir202', Aliases='Mirn202, mir-202, mmu-mir-202', description='microRNA 202', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=139957689, end_position_on_the_genomic_accession=139957760, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387199 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387199, CurrentID=0, Status='live', Symbol='Mir203', Aliases='Mirn203, mir-203, mmu-mir-203', description='microRNA 203', other_designations='', map_location='12', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=112130880, end_position_on_the_genomic_accession=112130955, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387200 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387200, CurrentID=0, Status='live', Symbol='Mir204', Aliases='Mirn204, mir-204, mmu-mir-204', description='microRNA 204', other_designations='', map_location='19', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=22750605, end_position_on_the_genomic_accession=22750672, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387201, CurrentID=0, Status='live', Symbol='Mir205', Aliases='Mirn205, mir-205, mmu-mir-205', description='microRNA 205', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=193507463, end_position_on_the_genomic_accession=193507530, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387202 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387202, CurrentID=0, Status='live', Symbol='Mir206', Aliases='Mirn206, mmu-mir-206', description='microRNA 206', other_designations='miR-206', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20679010, end_position_on_the_genomic_accession=20679082, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387203, CurrentID=0, Status='live', Symbol='Mir207', Aliases='Mirn207, mmu-mir-207', description='microRNA 207', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=40722917, end_position_on_the_genomic_accession=40722995, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387204 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387204, CurrentID=0, Status='live', Symbol='Mir208a', Aliases='Mirn208, Mirn208a, mir-208, mir-208a, mmu-mir-208', description='microRNA 208a', other_designations='microRNA 208|mmu-mir-208a', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=54949060, end_position_on_the_genomic_accession=54949142, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387206, CurrentID=0, Status='live', Symbol='Mir210', Aliases='Mirn210, mir-210, mmu-mir-210', description='microRNA 210', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=141221384, end_position_on_the_genomic_accession=141221493, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387207 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387207, CurrentID=0, Status='live', Symbol='Mir211', Aliases='Mirn211, mir-211, mmu-mir-211', description='microRNA 211', other_designations='', map_location='7', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=64205806, end_position_on_the_genomic_accession=64205911, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387208 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387208, CurrentID=0, Status='live', Symbol='Mir212', Aliases='Mirn212, mir-212, mmu-mir-212', description='microRNA 212', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=75173388, end_position_on_the_genomic_accession=75173478, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387210 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387210, CurrentID=0, Status='live', Symbol='Mir214', Aliases='Mirn214, mir-214, mmu-mir-214', description='microRNA 214', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=162223368, end_position_on_the_genomic_accession=162223477, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387211 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387211, CurrentID=0, Status='live', Symbol='Mir215', Aliases='Mirn215, mir-215, mmu-mir-215', description='microRNA 215', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=185313581, end_position_on_the_genomic_accession=185313692, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387212 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387212, CurrentID=0, Status='live', Symbol='Mir216a', Aliases='Mirn216, Mirn216a, mir-216a, mmu-mir-216', description='microRNA 216a', other_designations='mmu-mir-216a', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=28757012, end_position_on_the_genomic_accession=28757083, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387213 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387213, CurrentID=0, Status='live', Symbol='Mir217', Aliases='Mirn217, mir-217, mmu-mir-217', description='microRNA 217', other_designations='', map_location='11', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=28763728, end_position_on_the_genomic_accession=28763835, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387214 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387214, CurrentID=0, Status='live', Symbol='Mir218', Aliases='Mirn218', description='microRNA 218', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387216, CurrentID=0, Status='live', Symbol='Mir23a', Aliases='Mirn23a, mir-23a, mmu-mir-23a', description='microRNA 23a', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84208518, end_position_on_the_genomic_accession=84208592, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387217 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387217, CurrentID=0, Status='live', Symbol='Mir23b', Aliases='Mirn23b, mmu-mir-23b', description='microRNA 23b', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=63300484, end_position_on_the_genomic_accession=63300557, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387218, CurrentID=0, Status='live', Symbol='Mir26a-1', Aliases='Mirn26a, Mirn26a-1, miR-26a, mir-26a-1', description='microRNA 26a-1', other_designations='mmu-mir-26a-1', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=119031796, end_position_on_the_genomic_accession=119031885, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387219, CurrentID=0, Status='live', Symbol='Mir26b', Aliases='Mirn26b, mir-26b, mmu-mir-26b', description='microRNA 26b', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74394310, end_position_on_the_genomic_accession=74394394, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387220 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387220, CurrentID=0, Status='live', Symbol='Mir27a', Aliases='Mirn27a, mir-27a, mmu-mir-27a', description='microRNA 27a', other_designations='', map_location='8', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=84208672, end_position_on_the_genomic_accession=84208758, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387221 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387221, CurrentID=0, Status='live', Symbol='Mir27b', Aliases='Mirn27b, mmu-mir-27b', description='microRNA 27b', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=63300712, end_position_on_the_genomic_accession=63300784, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387222, CurrentID=0, Status='live', Symbol='Mir29a', Aliases='Mirn29a, mir-29a, mmu-mir-29a', description='microRNA 29a', other_designations='', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=31062660, end_position_on_the_genomic_accession=31062747, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387223 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387223, CurrentID=0, Status='live', Symbol='Mir29b-1', Aliases='Mirn29b, Mirn29b-1, mir-29b-1', description='microRNA 29b-1', other_designations='mmu-mir-29b-1', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=31063023, end_position_on_the_genomic_accession=31063093, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387224, CurrentID=0, Status='live', Symbol='Mir29c', Aliases='Mirn29c, mir-29c, mmu-mir-29c', description='microRNA 29c', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=195037547, end_position_on_the_genomic_accession=195037634, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387225, CurrentID=0, Status='live', Symbol='Mir30a', Aliases='Mirn30a, mir-30a, mmu-mir-30a', description='microRNA 30a', other_designations='', map_location='1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=23272269, end_position_on_the_genomic_accession=23272339, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387226 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387226, CurrentID=0, Status='live', Symbol='Mir30b', Aliases='Mirn30b, mir-30b, mmu-mir-30b', description='microRNA 30b', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=68337415, end_position_on_the_genomic_accession=68337510, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387227, CurrentID=0, Status='live', Symbol='Mir30c-1', Aliases='Mirn30c, Mirn30c-1, mir-30c-1', description='microRNA 30c-1', other_designations='mmu-mir-30c-1', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=120769534, end_position_on_the_genomic_accession=120769622, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387228 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387228, CurrentID=0, Status='live', Symbol='Mir30d', Aliases='Mirn30d, mir-30d, mmu-mir-30d', description='microRNA 30d', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=68341208, end_position_on_the_genomic_accession=68341289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387229 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387229, CurrentID=0, Status='live', Symbol='Mir99a', Aliases='Mirn99a, mir-99a, mmu-mir-99a', description='microRNA 99a', other_designations='', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=77598936, end_position_on_the_genomic_accession=77599000, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387230 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387230, CurrentID=0, Status='live', Symbol='Mir99b', Aliases='Mirn99b, mmu-mir-99b', description='microRNA 99b', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17830188, end_position_on_the_genomic_accession=17830257, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387231, CurrentID=0, Status='live', Symbol='Mir122', Aliases='Mir122a, Mirn122a, mir-122, mmu-mir-122', description='microRNA 122', other_designations='microRNA 122a|mmu-mir-122a', map_location='18', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=65248861, end_position_on_the_genomic_accession=65248926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387232, CurrentID=0, Status='live', Symbol='Mir122b', Aliases='Mirn122b', description='microRNA 122b', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387233 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387233, CurrentID=0, Status='live', Symbol='Mir124a-1', Aliases='Mirn124a, Mirn124a-1', description='microRNA 124a-1', other_designations='mmu-mir-124-1|mmu-mir-124a-1', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=64590657, end_position_on_the_genomic_accession=64590741, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387234 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387234, CurrentID=0, Status='live', Symbol='Mir124b', Aliases='Mirn124b', description='microRNA 124b', other_designations='', map_location='', chromosome='', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  387235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387235, CurrentID=0, Status='live', Symbol='Mir125a', Aliases='Mirn125a', description='microRNA 125a', other_designations='mmu-mir-125a', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17830812, end_position_on_the_genomic_accession=17830879, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387236 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387236, CurrentID=0, Status='live', Symbol='Mir125b-1', Aliases='Mirn125b, Mirn125b-1, mir-125b-1', description='microRNA 125b-1', other_designations='mmu-mir-125b-1', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=41581926, end_position_on_the_genomic_accession=41582002, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387237 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387237, CurrentID=0, Status='live', Symbol='Mir129-1', Aliases='Mirn129-1, Mirn129b, miR-129b, mir-129-1', description='microRNA 129-1', other_designations='microRNA 129b|mmu-mir-129-1', map_location='6', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=29022619, end_position_on_the_genomic_accession=29022691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387239 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387239, CurrentID=0, Status='live', Symbol='Mir199b', Aliases='Mirn199b, mir-199b', description='microRNA 199b', other_designations='mmu-mir-199b', map_location='2', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=32318460, end_position_on_the_genomic_accession=32318569, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387242, CurrentID=0, Status='live', Symbol='Mir200a', Aliases='Mirn200a, mir-200a', description='microRNA 200a', other_designations='mmu-mir-200a', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156054896, end_position_on_the_genomic_accession=156054985, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387243 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387243, CurrentID=0, Status='live', Symbol='Mir200b', Aliases='Mirn200b, mir-200b', description='microRNA 200b', other_designations='mmu-mir-200b', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=156055681, end_position_on_the_genomic_accession=156055750, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387244 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387244, CurrentID=0, Status='live', Symbol='Mirlet7a-1', Aliases='Let-7a, Mirnlet7a, Mirnlet7a-1, let-7a-1', description='microRNA let7a-1', other_designations='mmu-let-7a-1', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48538179, end_position_on_the_genomic_accession=48538272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387245 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387245, CurrentID=0, Status='live', Symbol='Mirlet7b', Aliases='Mirnlet7b, let-7b, mmu-let-7b', description='microRNA let7b', other_designations='', map_location='15', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=85707319, end_position_on_the_genomic_accession=85707403, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387246, CurrentID=0, Status='live', Symbol='Mirlet7c-1', Aliases='Mirnlet7c, Mirnlet7c-1, let-7c-1', description='microRNA let7c-1', other_designations='mmu-let-7c-1', map_location='16', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=77599657, end_position_on_the_genomic_accession=77599750, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387247, CurrentID=0, Status='live', Symbol='Mirlet7d', Aliases='Mirnlet7d, let-7d, mmu-let-7d', description='microRNA let7d', other_designations='', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48536012, end_position_on_the_genomic_accession=48536114, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387248, CurrentID=0, Status='live', Symbol='Mirlet7e', Aliases='Mirnlet7e, mmu-let-7e', description='microRNA let7e', other_designations='', map_location='17', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=17830352, end_position_on_the_genomic_accession=17830444, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387249, CurrentID=0, Status='live', Symbol='Mirlet7g', Aliases='Mirnlet7g, let-7g, mmu-let-7g', description='microRNA let7g', other_designations='', map_location='9', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=106178840, end_position_on_the_genomic_accession=106178927, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387251 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387251, CurrentID=0, Status='live', Symbol='Mirlet7i', Aliases='Mirnlet7i, let-7i, mmu-let-7i', description='microRNA let7i', other_designations='', map_location='10', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=122985640, end_position_on_the_genomic_accession=122985724, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387252, CurrentID=0, Status='live', Symbol='Mirlet7f-1', Aliases='Mirlet7f1, Mirnlet7f1, let-7f-1', description='microRNA let7f-1', other_designations='microRNA let7f1|mmu-let-7f-1', map_location='13', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=48537829, end_position_on_the_genomic_accession=48537917, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387253 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387253, CurrentID=0, Status='live', Symbol='Mirlet7f-2', Aliases='Mirlet7f2, Mirnlet7f2, let-7f-2', description='microRNA let7f-2', other_designations='microRNA let7f2|mmu-let-7f-2', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=151912346, end_position_on_the_genomic_accession=151912428, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  20852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20852, CurrentID=0, Status='live', Symbol='Stat6', Aliases='', description='signal transducer and activator of transcription 6', other_designations='', map_location='10 74.59 cM', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=127642986, end_position_on_the_genomic_accession=127660987, orientation='plus', exon_count=22, OMIM='', no_hdr0=''),
  20853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=20853, CurrentID=0, Status='live', Symbol='Stau1', Aliases='5830401L18Rik, AW549911, C85792, Stau', description='staufen (RNA binding protein) homolog 1 (Drosophila)', other_designations='', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=166947549, end_position_on_the_genomic_accession=166996299, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  387285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387285, CurrentID=0, Status='live', Symbol='Hcrtr2', Aliases='OX2R, mOX2aR, mOX2bR, mOXR2', description='hypocretin (orexin) receptor 2', other_designations='hypocretin receptor type 2|ox-2-R|ox2-R', map_location='9 D', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=76225880, end_position_on_the_genomic_accession=76323868, orientation='minus', exon_count=7, OMIM='', no_hdr0=''),
  414077 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=414077, CurrentID=0, Status='live', Symbol='Wdr83os', Aliases='', description='WD repeat domain 83 opposite strand', other_designations='UPF0139 membrane protein C19orf56 homolog|protein WDR83OS homolog', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85080963, end_position_on_the_genomic_accession=85082339, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  387314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387314, CurrentID=0, Status='live', Symbol='Tmtc1', Aliases='Arg99, BC023818', description='transmembrane and tetratricopeptide repeat containing 1', other_designations='ARG99 homolog', map_location='6 G3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=148232430, end_position_on_the_genomic_accession=148444845, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  632534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=632534, CurrentID=0, Status='live', Symbol='Vmn1r191', Aliases='V1rh15', description='vomeronasal 1 receptor 191', other_designations='vomernasal 1 receptor Vmn1r191|vomeronasal 1 receptor, H15|vomeronasal type 1 receptor H15', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=22178686, end_position_on_the_genomic_accession=22179582, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387339, CurrentID=0, Status='live', Symbol='Tas2r102', Aliases='Tas2r2, mGR02, mt2r51', description='taste receptor, type 2, member 102', other_designations='STC9-7|T2R102|tas2r102', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132762131, end_position_on_the_genomic_accession=132763174, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387340 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387340, CurrentID=0, Status='live', Symbol='Tas2r104', Aliases='T2R04, Tas2r4, mGR04, mt2r45', description='taste receptor, type 2, member 104', other_designations='T2R104|tas2r104', map_location='6 63.53 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131684836, end_position_on_the_genomic_accession=131685744, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387341, CurrentID=0, Status='live', Symbol='Tas2r106', Aliases='T2R06, Tas2r6, mGR06, mt2r44', description='taste receptor, type 2, member 106', other_designations='T2R106|tas2r106', map_location='6 63.52 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131677960, end_position_on_the_genomic_accession=131678886, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387342 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387342, CurrentID=0, Status='live', Symbol='Tas2r107', Aliases='STC5-1, T2R07, T2R107, T2R4, Tas2r7, mGR06, mt2r43', description='taste receptor, type 2, member 107', other_designations='', map_location='6 63.49 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131659158, end_position_on_the_genomic_accession=131660084, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387343, CurrentID=0, Status='live', Symbol='Tas2r109', Aliases='T2R09, Tas2r9, mGR09, mt2r62', description='taste receptor, type 2, member 109', other_designations='T2R109|tas2r109', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132980015, end_position_on_the_genomic_accession=132980965, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387344 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387344, CurrentID=0, Status='live', Symbol='Tas2r110', Aliases='T2R10, Tas2r10, mt2r57', description='taste receptor, type 2, member 110', other_designations='T2R110|tas2r110', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132868008, end_position_on_the_genomic_accession=132869009, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387345 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387345, CurrentID=0, Status='live', Symbol='Tas2r113', Aliases='T2R13, Tas2r13, mGR13, mt2r58', description='taste receptor, type 2, member 113', other_designations='T2R113|tas2r113', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132893011, end_position_on_the_genomic_accession=132893940, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387346 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387346, CurrentID=0, Status='live', Symbol='Tas2r114', Aliases='T2R14, Tas2r14, mGR14, mt2r46', description='taste receptor, type 2, member 114', other_designations='T2R114|tas2r114', map_location='6 F3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131689134, end_position_on_the_genomic_accession=131690063, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387347 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387347, CurrentID=0, Status='live', Symbol='Tas2r118', Aliases='Gm465, T2R18, Tas2r16, Tas2r18, mGR18, mt2r40', description='taste receptor, type 2, member 118', other_designations='T2R16|candidate taste receptor mt2r40', map_location='6 A3.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=23969161, end_position_on_the_genomic_accession=23970060, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387348 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387348, CurrentID=0, Status='live', Symbol='Tas2r120', Aliases='T2R20, Tas2r20, mGR20, mt2r47', description='taste receptor, type 2, member 120', other_designations='T2R120|tas2r120', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132656957, end_position_on_the_genomic_accession=132657844, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387349 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387349, CurrentID=0, Status='live', Symbol='Tas2r121', Aliases='T2R21, Tas2r13, Tas2r21, mGR21, mT2r48', description='taste receptor, type 2, member 121', other_designations='T2R121|T2R13|tas2r13', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132700090, end_position_on_the_genomic_accession=132701007, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387351, CurrentID=0, Status='live', Symbol='Tas2r124', Aliases='T2R24, Tas2r24, mGR24, mt2r50', description='taste receptor, type 2, member 124', other_designations='T2R124|tas2r124', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132754730, end_position_on_the_genomic_accession=132755659, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387352, CurrentID=0, Status='live', Symbol='Tas2r125', Aliases='T2R26, Tas2r25, mGR25, mt2r59', description='taste receptor, type 2, member 125', other_designations='T2R125|tas2r125', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132909651, end_position_on_the_genomic_accession=132910586, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387353, CurrentID=0, Status='live', Symbol='Tas2r126', Aliases='T2R12, T2R26, Tas2r26, Tas2r41, mGR26, mt2r35', description='taste receptor, type 2, member 126', other_designations='T2R41', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42434535, end_position_on_the_genomic_accession=42435461, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387354, CurrentID=0, Status='live', Symbol='Tas2r129', Aliases='T2R29, Tas2r29, mGR29, mt2r60', description='taste receptor, type 2, member 129', other_designations='T2R129|tas2r129', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132951102, end_position_on_the_genomic_accession=132952064, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387355, CurrentID=0, Status='live', Symbol='Tas2r130', Aliases='STC74, T2R30, Tas2r30, Tas2r7, mt2r42', description='taste receptor, type 2, member 130', other_designations='STC 7-4|STC7-4|T2R6|T2R7', map_location='6 63.44 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=131629892, end_position_on_the_genomic_accession=131630830, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387356, CurrentID=0, Status='live', Symbol='Tas2r131', Aliases='T2R31, Tas2r31, mGR31, mt2r61', description='taste receptor, type 2, member 131', other_designations='', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=132956912, end_position_on_the_genomic_accession=132957844, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  414089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=414089, CurrentID=0, Status='live', Symbol='Gja6', Aliases='A530015O12Rik, Cnx33, Cx33, Gja-6', description='gap junction protein, alpha 6', other_designations='connexin 33|connexin-33|gap junction membrane channel protein alpha 6', map_location='X 73.95 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=160902116, end_position_on_the_genomic_accession=160907052, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  387510 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387510, CurrentID=0, Status='live', Symbol='Ifnk', Aliases='', description='interferon kappa', other_designations='IFN-kappa', map_location='4 A5', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=35152055, end_position_on_the_genomic_accession=35154019, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  387511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387511, CurrentID=0, Status='live', Symbol='Tas2r134', Aliases='T2R134, Tas2r34', description='taste receptor, type 2, member 134', other_designations='T2R34|taste receptor type 2 member 34', map_location='2 C1.1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=51627511, end_position_on_the_genomic_accession=51628407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387512, CurrentID=0, Status='live', Symbol='Tas2r135', Aliases='Tas2r35, mt2r38', description='taste receptor, type 2, member 135', other_designations='T2R135|T2R35|taste receptor type 2 member 35', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42405529, end_position_on_the_genomic_accession=42406494, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387513, CurrentID=0, Status='live', Symbol='Tas2r138', Aliases='T2R138, Tas2r38, mt2r31', description='taste receptor, type 2, member 138', other_designations='T2R38', map_location='6 B1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40612315, end_position_on_the_genomic_accession=40613310, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  387514 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387514, CurrentID=0, Status='live', Symbol='Tas2r143', Aliases='Gm1404, Tas2r43, mt2r36', description='taste receptor, type 2, member 143', other_designations='T2R143|T2R43|taste receptor type 2 member 43', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42400238, end_position_on_the_genomic_accession=42401119, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387515 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387515, CurrentID=0, Status='live', Symbol='Tas2r144', Aliases='Tas2r40, Tas2r44, mt2r33', description='taste receptor, type 2, member 144', other_designations='T2R40', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42215328, end_position_on_the_genomic_accession=42216287, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  387524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387524, CurrentID=0, Status='live', Symbol='Znrf2', Aliases='1190002C14Rik, AU044205, D6Ertd365e', description='zinc and ring finger 2', other_designations='zinc finger/RING finger 2|zinc/RING finger protein 2', map_location='6 27.15 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=54816916, end_position_on_the_genomic_accession=54890224, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  387564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387564, CurrentID=0, Status='live', Symbol='Sval3', Aliases='', description='seminal vesicle antigen-like 3', other_designations='seminal vesicle autoantigen like protein 3', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41968140, end_position_on_the_genomic_accession=41973090, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  780794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=780794, CurrentID=0, Status='live', Symbol='Ighv5-17', Aliases='Gm16609', description='immunoglobulin heavy variable 5-17', other_designations='immunoglobulin heavy variable V5-17', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113859149, end_position_on_the_genomic_accession=113859442, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  780799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=780799, CurrentID=0, Status='live', Symbol='Ighv7-1', Aliases='Gm16698', description='immunoglobulin heavy variable 7-1', other_designations='Ig heavy chain V region M167|immunoglobulin heavy variable V7-1', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=113896408, end_position_on_the_genomic_accession=113896946, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387586, CurrentID=0, Status='live', Symbol='Ssxb5', Aliases='', description='synovial sarcoma, X member B, breakpoint 5', other_designations='', map_location='X', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=8803691, end_position_on_the_genomic_accession=8809445, orientation='plus', exon_count=7, OMIM='', no_hdr0=''),
  387609 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387609, CurrentID=0, Status='live', Symbol='Zhx2', Aliases='Afr-1, Afr1, Raf, mKIAA0854', description='zinc fingers and homeoboxes 2', other_designations='AFP regulator 1|alpha fetoprotein regulation 1, adult|alpha-fetoprotein regulator 1|regulator of AFP|zinc finger and homeodomain protein 2', map_location='15 24.01 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=57694667, end_position_on_the_genomic_accession=57839832, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  780829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=780829, CurrentID=0, Status='live', Symbol='Ighv3-6', Aliases='Gm16932', description='immunoglobulin heavy variable 3-6', other_designations='Ig heavy chain V region MOPC 315|VH gene product|immunoglobulin heavy variable V3-6', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114288152, end_position_on_the_genomic_accession=114288447, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  387616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=387616, CurrentID=0, Status='live', Symbol='Tas2r140', Aliases='T2R40, T2r34, TRB3, TRB5, Tas2r13, Tas2r137, Tas2r40, Tas2r8, mTRB3, mt2r64', description='taste receptor, type 2, member 140', other_designations='T2R13|T2R140|T2R8|mTRB5|taste receptor family B member 3|taste receptor type 2 member 40', map_location='6 64.03 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=133054855, end_position_on_the_genomic_accession=133055793, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  545205 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545205, CurrentID=0, Status='live', Symbol='Olfr111', Aliases='', description='olfactory receptor 111', other_designations='olfactory receptor MOR249-1P', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37504960, end_position_on_the_genomic_accession=37536128, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  195727 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195727, CurrentID=0, Status='live', Symbol='Nhs', Aliases='Gm389, Gm481, Xcat, Nhs', description='Nance-Horan syndrome (human)', other_designations='', map_location='X 74.17 cM', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=161833290, end_position_on_the_genomic_accession=162159441, orientation='minus', exon_count=12, OMIM='', no_hdr0=''),
  195733 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=195733, CurrentID=0, Status='live', Symbol='Grhl1', Aliases='LBP-32, MGR, Tcfcp2l2', description='grainyhead-like 1 (Drosophila)', other_designations='grainyhead like 1|mammalian grainyhead|p61 MGR|p70 MGR|transcription factor CP2-like 2|transcription factor LBP-32', map_location='12 A1.3', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=24572230, end_position_on_the_genomic_accession=24617391, orientation='plus', exon_count=18, OMIM='', no_hdr0=''),
  545260 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545260, CurrentID=0, Status='live', Symbol='Arsi', Aliases='BC020108', description='arylsulfatase i', other_designations='ASI', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=60912240, end_position_on_the_genomic_accession=60919504, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  545261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545261, CurrentID=0, Status='live', Symbol='Bvht', Aliases='Braveheart, Gm20748', description='braveheart long non-coding RNA', other_designations='', map_location='18 E1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=61639653, end_position_on_the_genomic_accession=61647485, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  257630 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257630, CurrentID=0, Status='live', Symbol='Il17f', Aliases='C87042, IL-17F', description='interleukin 17F', other_designations='', map_location='1 A4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=20777146, end_position_on_the_genomic_accession=20785274, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  257632 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257632, CurrentID=0, Status='live', Symbol='Nod2', Aliases='ACUG, BLAU, CD, Card15, F830032C23Rik, IBD1, Nlrc2', description='nucleotide-binding oligomerization domain containing 2', other_designations='caspase recruitment domain family, member 15|caspase recruitment domain-containing protein 15', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=88647280, end_position_on_the_genomic_accession=88688474, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  257633 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257633, CurrentID=0, Status='live', Symbol='Acsf3', Aliases='BB101783', description='acyl-CoA synthetase family member 3', other_designations='', map_location='8 E1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=122775493, end_position_on_the_genomic_accession=122817882, orientation='plus', exon_count=12, OMIM='', no_hdr0=''),
  257635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257635, CurrentID=0, Status='live', Symbol='Sdsl', Aliases='4432411H13Rik, AI504310, SDH1, SDS-RS1, TDH', description='serine dehydratase-like', other_designations='L-serine deaminase|L-serine dehydratase/L-threonine deaminase|L-threonine dehydratase|SDH|serine dehydratase related sequence 1', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=120458193, end_position_on_the_genomic_accession=120472891, orientation='minus', exon_count=9, OMIM='', no_hdr0=''),
  257662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257662, CurrentID=0, Status='live', Symbol='Olfr1290', Aliases='MOR248-9', description='olfactory receptor 1290', other_designations='', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111489218, end_position_on_the_genomic_accession=111493815, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  257666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257666, CurrentID=0, Status='live', Symbol='Olfr772', Aliases='GA_x5J8B7W1NLL-1408-2343, MOR114-3', description='olfactory receptor 772', other_designations='GA_x6K02T2PULF-10860457-10859522|olfactory receptor GA_x5J8B7W1NLL-1408-2343|olfactory receptor MOR114-3', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129174051, end_position_on_the_genomic_accession=129175054, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257734, CurrentID=0, Status='live', Symbol='Olfr1355', Aliases='GA_x5J8B7W4DUT-12263-11330, MOR139-6', description='olfactory receptor 1355', other_designations='GA_x6K02T2QGN0-2932609-2931677|M65 odorant receptor', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78875549, end_position_on_the_genomic_accession=78880106, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  257871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257871, CurrentID=0, Status='live', Symbol='Olfr1372-ps1', Aliases='GA_x5J8B7TR6JS-3-632, MOR256-54P, NP_TR6JSE50FPA', description='olfactory receptor 1372, pseudogene 1', other_designations='GA_x5J8B7TQSKC-412-5|GA_x6K02T2QP88-3272399-3273332', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=52154876, end_position_on_the_genomic_accession=52157907, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  257875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257875, CurrentID=0, Status='live', Symbol='Olfr1116', Aliases='MOR264-21P, MOR264-8P-ps, Olfr1521-ps1, Olfr1116', description='olfactory receptor 1116', other_designations='GA_x6K02T2Q125-48753766-48754734|olfactory receptor 1116, pseudogene|olfactory receptor 1521, pseudogene 1|olfactory receptor MOR264-21P|olfactory receptor MOR264-8P', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87268846, end_position_on_the_genomic_accession=87269769, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257880, CurrentID=0, Status='live', Symbol='Olfr1066', Aliases='MOR188-8, MOR256-52P', description='olfactory receptor 1066', other_designations='GA_x6K02T2Q125-47925557-47924616|olfactory receptor MOR256-52P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86455328, end_position_on_the_genomic_accession=86456269, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257881, CurrentID=0, Status='live', Symbol='Olfr205', Aliases='GA_x54KRFPKG5P-55543875-55542958, GA_x5J8B7U0478-6-340, MOR182-11P', description='olfactory receptor 205', other_designations='', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59328590, end_position_on_the_genomic_accession=59329507, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257883, CurrentID=0, Status='live', Symbol='Olfr1357', Aliases='MOR128-3, MOR128-4', description='olfactory receptor 1357', other_designations='GA_x6K02T2QGN0-3196801-3197742|olfactory receptor MOR128-3', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78611669, end_position_on_the_genomic_accession=78618074, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  257884 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257884, CurrentID=0, Status='live', Symbol='Olfr744', Aliases='MOR106-13P', description='olfactory receptor 744', other_designations='G-protein-coupled receptor|GA_x6K02T2PMLR-6326342-6327313|odorant receptor S3|olfactory receptor MOR106-13P', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50618224, end_position_on_the_genomic_accession=50619195, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257886, CurrentID=0, Status='live', Symbol='Olfr225', Aliases='MOR256-51', description='olfactory receptor 225', other_designations='GA_x6K02T0073K-490-1545|olfactory receptor MOR256-51', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59610786, end_position_on_the_genomic_accession=59614225, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  257888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257888, CurrentID=0, Status='live', Symbol='Olfr1386', Aliases='MOR256-50', description='olfactory receptor 1386', other_designations='GA_x6K02T2QP88-5964781-5963852|olfactory receptor MOR256-50', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49470054, end_position_on_the_genomic_accession=49471181, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257889 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257889, CurrentID=0, Status='live', Symbol='Olfr132', Aliases='MOR256-49', description='olfactory receptor 132', other_designations='GA_x6K02T2PSCP-2579687-2578746|olfactory receptor MOR256-49', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38130249, end_position_on_the_genomic_accession=38131190, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257890, CurrentID=0, Status='live', Symbol='Olfr12', Aliases='MOR208-5', description='olfactory receptor 12', other_designations='GA_x6K02T2R7CC-81134096-81133095|odorant receptor M76', map_location='1 46.53 cM', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92619881, end_position_on_the_genomic_accession=92621001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257891 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257891, CurrentID=0, Status='live', Symbol='Olfr479', Aliases='MOR267-15', description='olfactory receptor 479', other_designations='GA_x6K02T2PBJ9-10384085-10385068|olfactory receptor MOR267-15', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108054424, end_position_on_the_genomic_accession=108056061, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  257892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257892, CurrentID=0, Status='live', Symbol='Olfr324', Aliases='MOR102-2', description='olfactory receptor 324', other_designations='GA_x6K02T2NKPP-822947-822000|olfactory receptor MOR102-2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58597345, end_position_on_the_genomic_accession=58598466, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257898, CurrentID=0, Status='live', Symbol='Olfr867', Aliases='MOR143-2', description='olfactory receptor 867', other_designations='GA_x6K02T2PVTD-13795933-13794938|olfactory receptor MOR143-2|olfr867', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20054466, end_position_on_the_genomic_accession=20055461, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257900 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257900, CurrentID=0, Status='live', Symbol='Olfr1024', Aliases='MOR197-1', description='olfactory receptor 1024', other_designations='GA_x6K02T2Q125-47384320-47383337|olfactory receptor MOR197-1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85904069, end_position_on_the_genomic_accession=85905052, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257902, CurrentID=0, Status='live', Symbol='Olfr704', Aliases='MOR283-12P', description='olfactory receptor 704', other_designations='GA_x6K02T2PBJ9-9247095-9248042|T3 olfactory receptor|olfactory receptor MOR283-12P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106864982, end_position_on_the_genomic_accession=106865929, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257908 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257908, CurrentID=0, Status='live', Symbol='Olfr115', Aliases='MOR218-11P', description='olfactory receptor 115', other_designations='GA_x6K02T2PSCP-2070203-2069271|olfactory receptor MOR218-11P', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37609785, end_position_on_the_genomic_accession=37610779, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257912, CurrentID=0, Status='live', Symbol='Olfr948', Aliases='MOR171-45, MOR171-51', description='olfactory receptor 948', other_designations='GA_x6K02T2PVTD-33016899-33015934|olfactory receptor MOR171-45', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39318647, end_position_on_the_genomic_accession=39319612, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257913, CurrentID=0, Status='live', Symbol='Olfr141', Aliases='K17, MOR179-5', description='olfactory receptor 141', other_designations='GA_x6K02T2Q125-48299679-48298702|olfactory receptor 4B', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86806021, end_position_on_the_genomic_accession=86806997, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257917, CurrentID=0, Status='live', Symbol='Olfr314', Aliases='MOR278-2', description='olfactory receptor 314', other_designations='GA_x6K02T2NKPP-622179-621244|olfactory receptor MOR278-2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58786139, end_position_on_the_genomic_accession=58787269, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257921 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257921, CurrentID=0, Status='live', Symbol='Olfr1229', Aliases='MOR233-19, MOR233-22', description='olfactory receptor 1229', other_designations='GA_x6K02T2Q125-50727797-50726862|olfactory receptor MOR233-19', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89282196, end_position_on_the_genomic_accession=89283131, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257926 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257926, CurrentID=0, Status='live', Symbol='Olfr544', Aliases='MOR42-3, Ors6', description='olfactory receptor 544', other_designations='G-protein-coupled receptor|GA_x6K02T2PBJ9-5206624-5205620|odorant receptor S6|odorant receptor S79|olfactory receptor MOR42-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102484114, end_position_on_the_genomic_accession=102488307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  257931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257931, CurrentID=0, Status='live', Symbol='Olfr317', Aliases='MOR256-47', description='olfactory receptor 317', other_designations='GA_x5J8B7W3KVV-2795544-2795410|GA_x6K02T2NKPP-680866-681849|olfactory receptor MOR256-47', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58732096, end_position_on_the_genomic_accession=58733223, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257932 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257932, CurrentID=0, Status='live', Symbol='Olfr332', Aliases='MOR284-2', description='olfactory receptor 332', other_designations='GA_x6K02T2NKPP-926908-927915|olfactory receptor MOR284-2', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58489719, end_position_on_the_genomic_accession=58492499, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  257936 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257936, CurrentID=0, Status='live', Symbol='Olfr1028', Aliases='MOR198-3P, MOR198-4, MOR198-6_p, Olfr1534-ps1', description='olfactory receptor 1028', other_designations='GA_x6K02T2Q125-47430129-47431103|olfactory receptor 1534, pseudogene 1|olfactory receptor MOR198-3P|olfactory receptor MOR198-4', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85951065, end_position_on_the_genomic_accession=85952039, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257938, CurrentID=0, Status='live', Symbol='Olfr1419', Aliases='MOR266-10', description='olfactory receptor 1419', other_designations='GA_x6K02T2RE5P-2222521-2221490|olfactory receptor MOR266-10', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11870183, end_position_on_the_genomic_accession=11871214, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257939, CurrentID=0, Status='live', Symbol='Olfr527', Aliases='MOR251-5', description='olfactory receptor 527', other_designations='GA_x6K02T2PBJ9-42486061-42486978|olfactory receptor MOR251-5', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140335864, end_position_on_the_genomic_accession=140336781, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257941, CurrentID=0, Status='live', Symbol='Olfr1042', Aliases='MOR185-10', description='olfactory receptor 1042', other_designations='GA_x6K02T2Q125-47629317-47628376|olfactory receptor MOR185-10', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86159345, end_position_on_the_genomic_accession=86160441, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257947, CurrentID=0, Status='live', Symbol='Olfr543', Aliases='MOR42-2', description='olfactory receptor 543', other_designations='GA_x6K02T2PBJ9-5199377-5198367|olfactory receptor MOR42-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102476773, end_position_on_the_genomic_accession=102477902, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257950, CurrentID=0, Status='live', Symbol='Olfr1039', Aliases='MOR185-5, MOR185-9P, Olfr1517', description='olfactory receptor 1039', other_designations='GA_x6K02T2Q125-47600809-47599850|olfactory receptor 1517|olfactory receptor MOR185-5|olfactory receptor MOR185-9P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86130702, end_position_on_the_genomic_accession=86131661, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257951, CurrentID=0, Status='live', Symbol='Olfr987', Aliases='GA_x5J8B7TRSBM-251-3, GA_x5J8B7W531T-1241080-1240524, MOR203-7P', description='olfactory receptor 987', other_designations='GA_x6K02T2Q125-46808500-46807571|olfactory receptor GA_x5J8B7TRSBM-251-3|olfactory receptor GA_x5J8B7W531T-1241080-1240524|olfactory receptor MOR203-7P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85330967, end_position_on_the_genomic_accession=85331896, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257958, CurrentID=0, Status='live', Symbol='Olfr301', Aliases='MOR211-8P, MOR221-1P, MOR221-4, Olfr1531-ps1', description='olfactory receptor 301', other_designations='GA_x6K02T2NHDJ-9693313-9692378|olfactory receptor 1531, pseudogene 1|olfactory receptor MOR211-8P|olfactory receptor MOR221-1P', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=86403849, end_position_on_the_genomic_accession=86413299, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  257959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257959, CurrentID=0, Status='live', Symbol='Olfr1537', Aliases='K4, MOR171-32P, MOR171-41P, Olfr144-ps1, Olfr1537', description='olfactory receptor 1537', other_designations='GA_x6K02T2PVTD-32935684-32934749|odorant receptor K4|olfactory receptor 144|olfactory receptor 1537, pseudogene 1|olfactory receptor 7B|olfactory receptor MOR171-32P|olfactory receptor MOR171-41P', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39237487, end_position_on_the_genomic_accession=39238431, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257962, CurrentID=0, Status='live', Symbol='Olfr222', Aliases='MOR256-45P', description='olfactory receptor 222', other_designations='GA_x6K02SYYHDF-1415-2371|olfactory receptor MOR256-45P', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59570782, end_position_on_the_genomic_accession=59571738, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257972 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257972, CurrentID=0, Status='live', Symbol='Olfr193', Aliases='MOR113-7P, MOR183-7P, Olfr1540-ps1', description='olfactory receptor 193', other_designations='GA_x54KRFPKG5P-55338697-55337768|olfactory receptor 1540, pseudogene 1|olfactory receptor MOR113-7P|olfactory receptor MOR183-7P', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59109679, end_position_on_the_genomic_accession=59110608, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257974, CurrentID=0, Status='live', Symbol='Olfr1053', Aliases='MOR187-4-ps1, Olfr1053', description='olfactory receptor 1053', other_designations='GA_x6K02T2Q125-47793414-47792471|olfactory receptor 1053, pseudogene 1|olfactory receptor MOR187-4', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86314343, end_position_on_the_genomic_accession=86315285, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257975, CurrentID=0, Status='live', Symbol='Olfr598', Aliases='MOR23-4P, MOR23-5', description='olfactory receptor 598', other_designations='GA_x6K02T2PBJ9-6037823-6038782|olfactory receptor MOR23-4P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103328488, end_position_on_the_genomic_accession=103329447, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257978 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257978, CurrentID=0, Status='live', Symbol='Olfr1322', Aliases='MOR264-16P', description='olfactory receptor 1322', other_designations='GA_x6K02T2Q64A-11927171-11926215|olfactory receptor MOR264-16P', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=49885469, end_position_on_the_genomic_accession=49886401, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257979 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257979, CurrentID=0, Status='live', Symbol='Olfr1281', Aliases='MOR248-14P, MOR248-18', description='olfactory receptor 1281', other_designations='GA_x6K02T2Q125-72379864-72380781|olfactory receptor MOR248-14P', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111328421, end_position_on_the_genomic_accession=111329338, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  257980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257980, CurrentID=0, Status='live', Symbol='Olfr1275', Aliases='MOR245-18', description='olfactory receptor 1275', other_designations='GA_x6K02T2Q125-72283260-72282322|olfactory receptor MOR245-18', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111230853, end_position_on_the_genomic_accession=111231791, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  257996 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=257996, CurrentID=0, Status='live', Symbol='Olfr884', Aliases='MOR162-11P, MOR162-13, MOR162-9P, Olfr1550-ps1', description='olfactory receptor 884', other_designations='GA_x6K02T2PVTD-31726544-31727473|olfactory receptor 1550, pseudogene 1|olfactory receptor MOR162-11P|olfactory receptor MOR162-9P', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38046576, end_position_on_the_genomic_accession=38048153, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258008, CurrentID=0, Status='live', Symbol='Olfr1513', Aliases='MOR223-10, MOR223-7P', description='olfactory receptor 1513', other_designations='GA_x5J8B7TYTVL-817-1341|GA_x6K02T2RJGY-644134-645075|olfactory receptor MOR223-7P', map_location='14 C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52349103, end_position_on_the_genomic_accession=52350044, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258019 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258019, CurrentID=0, Status='live', Symbol='Olfr212', Aliases='4931403F16Rik, MOR119-4', description='olfactory receptor 212', other_designations='GA_x54KRFPKN04-58149882-58150865|olfactory receptor MOR119-4', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116506199, end_position_on_the_genomic_accession=116517965, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  258022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258022, CurrentID=0, Status='live', Symbol='Olfr1318', Aliases='MOR245-16', description='olfactory receptor 1318', other_designations='GA_x6K02T2Q125-73202172-73203134|olfactory receptor MOR245-16', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112155898, end_position_on_the_genomic_accession=112157007, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258025 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258025, CurrentID=0, Status='live', Symbol='Olfr1211', Aliases='MOR233-16P, MOR233-21', description='olfactory receptor 1211', other_designations='GA_x6K02T2Q125-50409367-50408432|olfactory receptor MOR233-16P', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88929378, end_position_on_the_genomic_accession=88934526, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258036, CurrentID=0, Status='live', Symbol='Olfr198', Aliases='MOR182-8', description='olfactory receptor 198', other_designations='GA_x54KRFPKG5P-55416332-55415412|olfactory receptor MOR182-8', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59201504, end_position_on_the_genomic_accession=59202424, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258039, CurrentID=0, Status='live', Symbol='Olfr728', Aliases='MOR246-1P', description='olfactory receptor 728', other_designations='GA_x6K02T2PMLR-5831021-5830086|olfactory receptor MOR246-1P', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50139702, end_position_on_the_genomic_accession=50140637, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258041, CurrentID=0, Status='live', Symbol='Olfr485', Aliases='MOR204-30P', description='olfactory receptor 485', other_designations='GA_x6K02T2PBJ9-10489044-10488091|MOR204-40_p|olfactory receptor MOR204-30P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108158918, end_position_on_the_genomic_accession=108159871, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258046, CurrentID=0, Status='live', Symbol='Olfr951', Aliases='MOR171-33P, MOR171-49', description='olfactory receptor 951', other_designations='GA_x6K02T2PVTD-33090395-33091330|olfactory receptor MOR171-33P', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39393793, end_position_on_the_genomic_accession=39394737, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258051, CurrentID=0, Status='live', Symbol='Olfr93', Aliases='MOR256-39P', description='olfactory receptor 93', other_designations='GA_x6K02T2PSCP-1592036-1591098|olfactory receptor MOR256-39P', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37151032, end_position_on_the_genomic_accession=37151970, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258055, CurrentID=0, Status='live', Symbol='Olfr524', Aliases='MOR103-14P', description='olfactory receptor 524', other_designations='GA_x6K02T2PBJ9-42354580-42353624|olfactory receptor MOR103-14P', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140201336, end_position_on_the_genomic_accession=140205279, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258056 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258056, CurrentID=0, Status='live', Symbol='Olfr533', Aliases='MOR252-3P', description='olfactory receptor 533', other_designations='GA_x6K02T2PBJ9-42615403-42616365|olfactory receptor MOR252-3P', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140466203, end_position_on_the_genomic_accession=140467165, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258064, CurrentID=0, Status='live', Symbol='Olfr316', Aliases='MOR285-3P, MOR285-6_p', description='olfactory receptor 316', other_designations='GA_x6K02T2NKPP-648708-647788|olfactory receptor MOR285-3P', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58751530, end_position_on_the_genomic_accession=58759466, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  258065 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258065, CurrentID=0, Status='live', Symbol='Olfr312', Aliases='GA_x5J8B7W3UM0-1747565-1747342, MOR222-4P', description='olfactory receptor 312', other_designations='GA_x6K02T2NKPP-575829-574903|olfactory receptor GA_x5J8B7W3UM0-1747565-1747342|olfactory receptor MOR222-4P', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58831156, end_position_on_the_genomic_accession=58832082, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258067, CurrentID=0, Status='live', Symbol='Olfr1359', Aliases='MOR256-35, MOR256-60', description='olfactory receptor 1359', other_designations='GA_x6K02T2QHY8-11534186-11533245|olfactory receptor MOR256-35', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21703003, end_position_on_the_genomic_accession=21703944, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258074, CurrentID=0, Status='live', Symbol='Olfr834', Aliases='MOR153-2, MOR153-4_p', description='olfactory receptor 834', other_designations='GA_x6K02T2PVTD-12724921-12725859|olfactory receptor MOR153-2', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18987990, end_position_on_the_genomic_accession=18988928, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258085, CurrentID=0, Status='live', Symbol='Olfr1105', Aliases='MOR0-6P, MOR172-7', description='olfactory receptor 1105', other_designations='GA_x6K02T2Q125-48521031-48520093|olfactory receptor MOR0-6P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87033281, end_position_on_the_genomic_accession=87034219, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258087, CurrentID=0, Status='live', Symbol='Olfr963', Aliases='MOR224-13, MOR224-7P', description='olfactory receptor 963', other_designations='GA_x6K02T2PVTD-33365879-33366814|olfactory receptor MOR224-7P', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39669059, end_position_on_the_genomic_accession=39669994, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258094 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258094, CurrentID=0, Status='live', Symbol='Olfr761', Aliases='MOR218-12, MOR218-5P, MOR218-6P, Olfr1552-ps1', description='olfactory receptor 761', other_designations='GA_x6K02T2PSCP-2403971-2403000|olfactory receptor 1552, pseudogene 1|olfactory receptor MOR218-5P|olfactory receptor MOR218-6P', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37952051, end_position_on_the_genomic_accession=37953022, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258095, CurrentID=0, Status='live', Symbol='Olfr119', Aliases='MOR263-7', description='olfactory receptor 119', other_designations='GA_x6K02T2PSCP-2159633-2160598|olfactory receptor MOR263-7', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37696685, end_position_on_the_genomic_accession=37701720, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258097, CurrentID=0, Status='live', Symbol='Olfr1500', Aliases='MOR212-4P', description='olfactory receptor 1500', other_designations='GA_x6K02T2RE5P-4160554-4159619|olfactory receptor MOR212-4P', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13827459, end_position_on_the_genomic_accession=13828394, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258098 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258098, CurrentID=0, Status='live', Symbol='Olfr1490', Aliases='MOR266-6P', description='olfactory receptor 1490', other_designations='GA_x6K02T2RE5P-3987000-3987950|olfactory receptor MOR266-6P', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13654446, end_position_on_the_genomic_accession=13655396, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258101 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258101, CurrentID=0, Status='live', Symbol='Olfr1137', Aliases='MOR177-20, MOR40-9P', description='olfactory receptor 1137', other_designations='GA_x6K02T2Q125-49215724-49214792|olfactory receptor MOR40-9P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87710972, end_position_on_the_genomic_accession=87711904, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258111, CurrentID=0, Status='live', Symbol='Olfr1129', Aliases='MOR264-23, MOR264-9P', description='olfactory receptor 1129', other_designations='GA_x6K02T2Q125-49078087-49079031|olfactory receptor MOR264-9P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87575086, end_position_on_the_genomic_accession=87576030, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258120 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258120, CurrentID=0, Status='live', Symbol='Olfr1463', Aliases='MOR202-29P', description='olfactory receptor 1463', other_designations='GA_x6K02T2RE5P-3560863-3561795|olfactory receptor MOR202-29P', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13234252, end_position_on_the_genomic_accession=13235184, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258123, CurrentID=0, Status='live', Symbol='Olfr1474', Aliases='MOR202-26P, MOR202-42', description='olfactory receptor 1474', other_designations='GA_x6K02T2RE5P-3803583-3804527|olfactory receptor MOR202-26P', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13470972, end_position_on_the_genomic_accession=13471916, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258135 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258135, CurrentID=0, Status='live', Symbol='Olfr597', Aliases='GA_x5J8B7TU1BB-380-666, MOR23-2', description='olfactory receptor 597', other_designations='GA_x6K02T2PBJ9-6029614-6030561|olfactory receptor GA_x5J8B7TU1BB-380-666|olfactory receptor MOR23-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103319139, end_position_on_the_genomic_accession=103321431, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258136 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258136, CurrentID=0, Status='live', Symbol='Olfr517', Aliases='MOR268-4', description='olfactory receptor 517', other_designations='GA_x6K02T2PBJ9-11199311-11198367|olfactory receptor MOR268-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108868208, end_position_on_the_genomic_accession=108869152, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258150 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258150, CurrentID=0, Status='live', Symbol='Olfr792', Aliases='MOR108-2', description='olfactory receptor 792', other_designations='GA_x6K02T2PULF-11219415-11220350|olfactory receptor MOR108-2', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129540539, end_position_on_the_genomic_accession=129541474, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258151, CurrentID=0, Status='live', Symbol='Olfr1505', Aliases='MOR211-4P', description='olfactory receptor 1505', other_designations='GA_x6K02T2RE5P-4250267-4251217|MOR211-10_i|olfactory receptor MOR211-4P', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13919022, end_position_on_the_genomic_accession=13919972, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258153 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258153, CurrentID=0, Status='live', Symbol='Olfr412', Aliases='MOR127-5P', description='olfactory receptor 412', other_designations='GA_x6K02T2P1NL-4500587-4501525|olfactory receptor MOR127-5P', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74364671, end_position_on_the_genomic_accession=74365609, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258154 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258154, CurrentID=0, Status='live', Symbol='Olfr1029', Aliases='MOR198-1P', description='olfactory receptor 1029', other_designations='GA_x6K02T2Q125-47454152-47455126|olfactory receptor MOR198-1P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85975213, end_position_on_the_genomic_accession=85976273, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258156 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258156, CurrentID=0, Status='live', Symbol='Olfr605', Aliases='GA_x5J8B7TSRWK-430-5, MOR202-22P, MOR24-5', description='olfactory receptor 605', other_designations='GA_x6K02T2PBJ9-6164792-6163848|olfactory receptor GA_x5J8B7TSRWK-430-5|olfactory receptor MOR202-22P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103442140, end_position_on_the_genomic_accession=103443150, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258160 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258160, CurrentID=0, Status='live', Symbol='Olfr685', Aliases='MOR37-1', description='olfactory receptor 685', other_designations='GA_x6K02T2PBJ9-7810071-7809121|olfactory receptor MOR37-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105180361, end_position_on_the_genomic_accession=105181311, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258163, CurrentID=0, Status='live', Symbol='Olfr504', Aliases='MOR40-15, MOR40-7P', description='olfactory receptor 504', other_designations='GA_x6K02T2PBJ9-10895499-10894543|olfactory receptor MOR40-7P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108564837, end_position_on_the_genomic_accession=108565793, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258166 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258166, CurrentID=0, Status='live', Symbol='Olfr988', Aliases='MOR203-5P', description='olfactory receptor 988', other_designations='GA_x6K02T2Q125-46830591-46829662|olfactory receptor MOR203-5P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85352995, end_position_on_the_genomic_accession=85353924, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258167 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258167, CurrentID=0, Status='live', Symbol='Olfr1182', Aliases='MOR225-7P, MOR225-8P, Olfr1536-ps1', description='olfactory receptor 1182', other_designations='GA_x6K02T2Q125-49934557-49933640|MOR225-16_p|olfactory receptor 1536, pseudogene 1|olfactory receptor MOR225-7P|olfactory receptor MOR225-8P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88446019, end_position_on_the_genomic_accession=88449121, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258168 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258168, CurrentID=0, Status='live', Symbol='Olfr566', Aliases='GA_x5J8B7W5P47-379997-380552, MOR14-7P', description='olfactory receptor 566', other_designations='GA_x6K02T2PBJ9-5560696-5559746|olfactory receptor GA_x5J8B7W5P47-379997-380552|olfactory receptor MOR14-7P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102856321, end_position_on_the_genomic_accession=102858746, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258173 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258173, CurrentID=0, Status='live', Symbol='Olfr1532-ps1', Aliases='MOR260-6P, MOR260-9P, Olfr708', description='olfactory receptor 1532, pseudogene 1', other_designations='B5 olfactory receptor|GA_x6K02T2PBJ9-9297671-9298594|olfactory receptor MOR260-6P|olfactory receptor MOR260-9P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106907328, end_position_on_the_genomic_accession=106915123, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  258189 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258189, CurrentID=0, Status='live', Symbol='Olfr624', Aliases='MOR4-2P', description='olfactory receptor 624', other_designations='GA_x6K02T2PBJ9-6394126-6393197|olfactory receptor MOR4-2P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103670103, end_position_on_the_genomic_accession=103671029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258196 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258196, CurrentID=0, Status='live', Symbol='Olfr309', Aliases='MOR251-4P', description='olfactory receptor 309', other_designations='GA_x6K02T2NHDJ-9801340-9802266|olfactory receptor MOR251-4P', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=86306185, end_position_on_the_genomic_accession=86307111, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258198 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258198, CurrentID=0, Status='live', Symbol='Olfr224', Aliases='MOR275-3-ps1, Olfr327-ps1, Olfr224', description='olfactory receptor 224', other_designations='GA_x6K02T00261-652-347|GA_x6K02T2NKPP-858022-858862|MOR275-10_p|olfactory receptor 224, pseudogene 1|olfactory receptor 327, pseudogene 1|olfactory receptor MOR275-3', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58566399, end_position_on_the_genomic_accession=58567343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258203 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258203, CurrentID=0, Status='live', Symbol='Olfr1178', Aliases='MOR225-6P', description='olfactory receptor 1178', other_designations='GA_x6K02T2Q125-49870417-49871388|olfactory receptor MOR225-6P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88391249, end_position_on_the_genomic_accession=88392220, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258206 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258206, CurrentID=0, Status='live', Symbol='Olfr1264', Aliases='MOR236-1, Mor16, Ors16', description='olfactory receptor 1264', other_designations='GA_x6K02T2Q125-51454183-51453257|odorant receptor 16|olfactory receptor MOR236-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90021138, end_position_on_the_genomic_accession=90022064, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258215 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258215, CurrentID=0, Status='live', Symbol='Olfr506', Aliases='MOR204-23', description='olfactory receptor 506', other_designations='GA_x6K02T2PBJ9-10942040-10942984|olfactory receptor MOR204-23', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108612309, end_position_on_the_genomic_accession=108613253, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258216 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258216, CurrentID=0, Status='live', Symbol='Olfr1034', Aliases='MOR227-8P, MOR245-14P, Olfr1533-ps1', description='olfactory receptor 1034', other_designations='GA_x6K02T2Q125-47521463-47522395|olfactory receptor 1533, pseudogene 1|olfactory receptor MOR227-8P|olfactory receptor MOR245-14P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86046444, end_position_on_the_genomic_accession=86047466, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258218 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258218, CurrentID=0, Status='live', Symbol='Olfr102', Aliases='MOR250-4', description='olfactory receptor 102', other_designations='GA_x6K02T2PSCP-1775063-1774137|olfactory receptor MOR250-4', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37313411, end_position_on_the_genomic_accession=37314437, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258219 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258219, CurrentID=0, Status='live', Symbol='Olfr94', Aliases='MOR256-32P', description='olfactory receptor 94', other_designations='GA_x6K02T2PSCP-1637650-1636712|olfactory receptor MOR256-32P', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37196806, end_position_on_the_genomic_accession=37197966, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258222 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258222, CurrentID=0, Status='live', Symbol='Olfr310', Aliases='MOR227-6P', description='olfactory receptor 310', other_designations='GA_x6K02T2NHDJ-9838699-9839697|olfactory receptor MOR227-6P', map_location='7 D3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=86268730, end_position_on_the_genomic_accession=86269811, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258224 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258224, CurrentID=0, Status='live', Symbol='Olfr1358', Aliases='MOR267-9', description='olfactory receptor 1358', other_designations='GA_x6K02T2QGN0-3289955-3289014|olfactory receptor MOR267-9', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78498172, end_position_on_the_genomic_accession=78520566, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258225 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258225, CurrentID=0, Status='live', Symbol='Olfr913', Aliases='MOR165-10, MOR165-9P', description='olfactory receptor 913', other_designations='olfactory receptor MOR165-9P', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38592803, end_position_on_the_genomic_accession=38595161, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258227 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258227, CurrentID=0, Status='live', Symbol='Olfr1418', Aliases='MOR266-5', description='olfactory receptor 1418', other_designations='GA_x6K02T2RE5P-2207258-2206302|olfactory receptor MOR266-5', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11854995, end_position_on_the_genomic_accession=11855951, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258231 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258231, CurrentID=0, Status='live', Symbol='Olfr1471', Aliases='GA_x5J8B7W6KF8-5605544-5605301, MOR202-38, MOR202-47_p', description='olfactory receptor 1471', other_designations='GA_x6K02T2RE5P-3777626-3778570', map_location='19 A', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13445014, end_position_on_the_genomic_accession=13445958, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258232 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258232, CurrentID=0, Status='live', Symbol='Olfr774', Aliases='MOR111-11, MOR111-2', description='olfactory receptor 774', other_designations='GA_x6K02T2PULF-10924432-10925370|olfactory receptor MOR111-11', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129238151, end_position_on_the_genomic_accession=129239089, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258235 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258235, CurrentID=0, Status='live', Symbol='Olfr1084', Aliases='MOR192-2, MOR192-3', description='olfactory receptor 1084', other_designations='GA_x6K02T2Q125-48121788-48120847|olfactory receptor MOR192-2', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86638765, end_position_on_the_genomic_accession=86639718, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258241 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258241, CurrentID=0, Status='live', Symbol='Olfr1212', Aliases='MOR233-17, MOR233-20', description='olfactory receptor 1212', other_designations='GA_x6K02T2Q125-50437014-50437949|olfactory receptor MOR233-17', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88958468, end_position_on_the_genomic_accession=88959403, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258242 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258242, CurrentID=0, Status='live', Symbol='Olfr955', Aliases='MOR171-34, MOR171-50', description='olfactory receptor 955', other_designations='GA_x6K02T2PVTD-33167297-33166353|olfactory receptor MOR171-34', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39469780, end_position_on_the_genomic_accession=39470724, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258246 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258246, CurrentID=0, Status='live', Symbol='Olfr594', Aliases='MOR32-10', description='olfactory receptor 594', other_designations='GA_x6K02T2PBJ9-5935234-5936169|olfactory receptor MOR32-10', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103219720, end_position_on_the_genomic_accession=103220655, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258247 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258247, CurrentID=0, Status='live', Symbol='Olfr645', Aliases='MOR13-5', description='olfactory receptor 645', other_designations='GA_x6K02T2PBJ9-6819097-6818150|olfactory receptor MOR13-5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104084131, end_position_on_the_genomic_accession=104085078, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258248 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258248, CurrentID=0, Status='live', Symbol='Olfr576', Aliases='MOR8-5-ps1, Olfr576', description='olfactory receptor 576', other_designations='GA_x6K02T2PBJ9-5676385-5677371|olfactory receptor 576, pseudogene 1|olfactory receptor MOR8-5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102965102, end_position_on_the_genomic_accession=102966040, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258249 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258249, CurrentID=0, Status='live', Symbol='Olfr845', Aliases='GA_x5J8B7W6RL5-13522154-13521822, MOR150-1, MOR150-1P, MOR150-2, Olfr1522-ps1', description='olfactory receptor 845', other_designations='GA_x6K02T2PVTD-13076685-13077623|olfactory receptor 1522, pseudogene 1|olfactory receptor GA_x5J8B7W6RL5-13522154-13521822|olfactory receptor MOR150-1P|olfactory receptor MOR150-2', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19338459, end_position_on_the_genomic_accession=19339400, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258252 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258252, CurrentID=0, Status='live', Symbol='Olfr813', Aliases='MOR108-3', description='olfactory receptor 813', other_designations='GA_x6K02T2PULF-11535078-11536010|olfactory receptor MOR108-3', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129856520, end_position_on_the_genomic_accession=129857452, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258255 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258255, CurrentID=0, Status='live', Symbol='Olfr1010', Aliases='MOR213-1, MOR213-1_p, MOR213-7P, Olfr1011, Olfr1011-ps1', description='olfactory receptor 1010', other_designations='GA_x6K02T2Q125-47228321-47229256|GA_x6K02T2Q125-47232672-47233052|olfactory receptor 1011, pseudogene 1|olfactory receptor MOR213-1|olfactory receptor MOR213-7P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85753363, end_position_on_the_genomic_accession=85754306, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258257 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258257, CurrentID=0, Status='live', Symbol='Olfr1313', Aliases='MOR245-23', description='olfactory receptor 1313', other_designations='GA_x6K02T2Q125-73119859-73118924|olfactory receptor MOR245-23', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112071646, end_position_on_the_genomic_accession=112072581, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258258 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258258, CurrentID=0, Status='live', Symbol='Olfr1308', Aliases='MOR245-22', description='olfactory receptor 1308', other_designations='GA_x6K02T2Q125-73008844-73007882|olfactory receptor MOR245-22', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111960109, end_position_on_the_genomic_accession=111961071, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258259 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258259, CurrentID=0, Status='live', Symbol='Olfr1338', Aliases='MOR259-4P, MOR259-9, Olfr1524-ps1', description='olfactory receptor 1338', other_designations='GA_x6K02T2QD9B-18795136-18796077|olfactory receptor 1524, pseudogene 1|olfactory receptor MOR259-4P|olfactory receptor MOR259-9', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118753189, end_position_on_the_genomic_accession=118756958, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258261 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258261, CurrentID=0, Status='live', Symbol='Olfr325', Aliases='MOR275-5', description='olfactory receptor 325', other_designations='GA_x6K02T2NKPP-844642-843680|MOR275-11_p|olfactory receptor MOR275-5', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58580837, end_position_on_the_genomic_accession=58581877, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258263 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258263, CurrentID=0, Status='live', Symbol='Olfr117', Aliases='MOR256-33', description='olfactory receptor 117', other_designations='GA_x6K02T2PSCP-2119438-2118485|olfactory receptor MOR256-33', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37659378, end_position_on_the_genomic_accession=37660331, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258264 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258264, CurrentID=0, Status='live', Symbol='Olfr747', Aliases='MOR106-16, MOR106-7', description='olfactory receptor 747', other_designations='GA_x6K02T2PMLR-6420220-6419279|olfactory receptor MOR106-7', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50680691, end_position_on_the_genomic_accession=50681645, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258265 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258265, CurrentID=0, Status='live', Symbol='Olfr1333', Aliases='GA_x5J8B7TQTH0-2-774, MOR259-11, MOR259-6', description='olfactory receptor 1333', other_designations='GA_x6K02T2QD9B-18703033-18703986', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118829485, end_position_on_the_genomic_accession=118830438, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258266 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258266, CurrentID=0, Status='live', Symbol='Olfr247', Aliases='GA_x5J8B7TSHCU-6-512, MOR265-1', description='olfactory receptor 247', other_designations='GA_x6K02T2LVGG-5968-6912|olfactory receptor GA_x5J8B7TSHCU-6-512|olfactory receptor MOR265-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129974605, end_position_on_the_genomic_accession=129975549, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258267 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258267, CurrentID=0, Status='live', Symbol='Olfr370', Aliases='MOR267-16, MOR26716', description='olfactory receptor 370', other_designations='', map_location='8 C1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=83541093, end_position_on_the_genomic_accession=83542188, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258268 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258268, CurrentID=0, Status='live', Symbol='Olfr1511', Aliases='MOR223-9', description='olfactory receptor 1511', other_designations='GA_x6K02T2RJGY-608749-609705|olfactory receptor MOR223-9', map_location='14 C1-C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52389587, end_position_on_the_genomic_accession=52397807, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258269 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258269, CurrentID=0, Status='live', Symbol='Olfr930', Aliases='MOR171-46', description='olfactory receptor 930', other_designations='GA_x6K02T2PVTD-32626123-32627049|olfactory receptor MOR171-46', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38930173, end_position_on_the_genomic_accession=38931099, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258270 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258270, CurrentID=0, Status='live', Symbol='Olfr448', Aliases='MOR261-13', description='olfactory receptor 448', other_designations='GA_x6K02T2P3E9-4663051-4662119|olfactory receptor MOR261-13', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42896453, end_position_on_the_genomic_accession=42897385, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258271 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258271, CurrentID=0, Status='live', Symbol='Olfr1311', Aliases='MOR245-21', description='olfactory receptor 1311', other_designations='GA_x6K02T2Q125-73069292-73068354|olfactory receptor MOR245-21', map_location='2 F2-F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112020914, end_position_on_the_genomic_accession=112021852, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258272 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258272, CurrentID=0, Status='live', Symbol='Olfr1402', Aliases='MOR217-1', description='olfactory receptor 1402', other_designations='GA_x6K02T2R76Y-3201307-3200360|olfactory receptor MOR217-1', map_location='3 F2-F3', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=97410217, end_position_on_the_genomic_accession=97411179, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258273 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258273, CurrentID=0, Status='live', Symbol='Olfr1394', Aliases='MOR280-1', description='olfactory receptor 1394', other_designations='GA_x6K02T2QP88-6274566-6273628|olfactory receptor MOR280-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49160016, end_position_on_the_genomic_accession=49160954, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258274 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258274, CurrentID=0, Status='live', Symbol='Olfr1412', Aliases='MOR208-4', description='olfactory receptor 1412', other_designations='GA_x6K02T2R7CC-81165686-81164721|olfactory receptor MOR208-4', map_location='1 D-E1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92588332, end_position_on_the_genomic_accession=92589297, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258275 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258275, CurrentID=0, Status='live', Symbol='Olfr729', Aliases='MOR246-6', description='olfactory receptor 729', other_designations='GA_x6K02T2PMLR-5839874-5838903|olfactory receptor MOR246-6', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50147901, end_position_on_the_genomic_accession=50148872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258276 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258276, CurrentID=0, Status='live', Symbol='Olfr960', Aliases='MOR224-12', description='olfactory receptor 960', other_designations='GA_x6K02T2PVTD-33320043-33320987|olfactory receptor MOR224-12', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39623125, end_position_on_the_genomic_accession=39624075, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258277, CurrentID=0, Status='live', Symbol='Olfr281', Aliases='MOR160-5', description='olfactory receptor 281', other_designations='GA_x6K02T2NBG7-5275017-5274082|olfactory receptor MOR160-5', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98456312, end_position_on_the_genomic_accession=98457247, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258278 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258278, CurrentID=0, Status='live', Symbol='Olfr284', Aliases='MOR160-4', description='olfactory receptor 284', other_designations='GA_x6K02T2NBG7-5395976-5396893|olfactory receptor MOR160-4', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98340022, end_position_on_the_genomic_accession=98340939, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258280 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258280, CurrentID=0, Status='live', Symbol='Olfr1366', Aliases='MOR130-2', description='olfactory receptor 1366', other_designations='GA_x6K02T2QHY8-11707053-11707994|olfactory receptor MOR130-2', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21536974, end_position_on_the_genomic_accession=21538027, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258281 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258281, CurrentID=0, Status='live', Symbol='Olfr780', Aliases='MOR114-12', description='olfactory receptor 780', other_designations='GA_x6K02T2PULF-11002360-11003298|olfactory receptor MOR114-12', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129321625, end_position_on_the_genomic_accession=129322563, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258282 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258282, CurrentID=0, Status='live', Symbol='Olfr801', Aliases='MOR110-10', description='olfactory receptor 801', other_designations='GA_x6K02T2PULF-11349138-11348182|olfactory receptor MOR110-10', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129669561, end_position_on_the_genomic_accession=129670517, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258283 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258283, CurrentID=0, Status='live', Symbol='Olfr981', Aliases='MOR223-8', description='olfactory receptor 981', other_designations='GA_x6K02T2PVTD-33720892-33721824|olfactory receptor MOR223-8', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40022395, end_position_on_the_genomic_accession=40023327, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258284 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258284, CurrentID=0, Status='live', Symbol='Olfr114', Aliases='MOR218-10', description='olfactory receptor 114', other_designations='GA_x6K02T2PSCP-2049802-2048870|olfactory receptor MOR218-10', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37589413, end_position_on_the_genomic_accession=37590351, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258285 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258285, CurrentID=0, Status='live', Symbol='Olfr122', Aliases='MOR263-10', description='olfactory receptor 122', other_designations='GA_x6K02T2PSCP-2230932-2231897|olfactory receptor MOR263-10', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37768642, end_position_on_the_genomic_accession=37772620, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258287 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258287, CurrentID=0, Status='live', Symbol='Olfr125', Aliases='MOR218-8', description='olfactory receptor 125', other_designations='GA_x6K02T2PSCP-2291580-2292542|olfactory receptor MOR218-8', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37834911, end_position_on_the_genomic_accession=37836028, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258289 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258289, CurrentID=0, Status='live', Symbol='Olfr1324', Aliases='MOR128-2', description='olfactory receptor 1324', other_designations='GA_x6K02T2Q64A-12454598-12453654|olfactory receptor MOR128-2', map_location='X A5', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=50425496, end_position_on_the_genomic_accession=50426494, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258291 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258291, CurrentID=0, Status='live', Symbol='Olfr1167', Aliases='MOR174-16', description='olfactory receptor 1167', other_designations='GA_x6K02T2Q125-49641892-49640942|olfactory receptor MOR174-16', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88148984, end_position_on_the_genomic_accession=88150083, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258292 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258292, CurrentID=0, Status='live', Symbol='Olfr446', Aliases='MOR261-12', description='olfactory receptor 446', other_designations='GA_x6K02T2P3E9-4632269-4631343|olfactory receptor MOR261-12', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42927233, end_position_on_the_genomic_accession=42928159, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258293 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258293, CurrentID=0, Status='live', Symbol='Olfr437', Aliases='MOR261-11', description='olfactory receptor 437', other_designations='GA_x6K02T2P3E9-4391088-4390156|olfactory receptor MOR261-11', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43167060, end_position_on_the_genomic_accession=43167992, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258294 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258294, CurrentID=0, Status='live', Symbol='Olfr1115', Aliases='MOR264-20, MOR273-4P, Olfr1530-ps1', description='olfactory receptor 1115', other_designations='GA_x6K02T2Q125-48736906-48737886|olfactory receptor 1530, pseudogene 1|olfactory receptor MOR264-20|olfactory receptor MOR273-4P', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87251840, end_position_on_the_genomic_accession=87252941, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258296, CurrentID=0, Status='live', Symbol='Olfr745', Aliases='MOR106-11', description='olfactory receptor 745', other_designations='GA_x6K02T2PMLR-6361495-6362481|olfactory receptor MOR106-11', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50642195, end_position_on_the_genomic_accession=50643369, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258297, CurrentID=0, Status='live', Symbol='Olfr827', Aliases='MOR210-5', description='olfactory receptor 827', other_designations='GA_x6K02T2PULF-11878777-11877809|olfactory receptor MOR210-5', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130210160, end_position_on_the_genomic_accession=130211128, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258300, CurrentID=0, Status='live', Symbol='Olfr1449', Aliases='MOR202-34', description='olfactory receptor 1449', other_designations='GA_x6K02T2RE5P-3264213-3265157|olfactory receptor MOR202-34', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12934740, end_position_on_the_genomic_accession=12935684, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258301, CurrentID=0, Status='live', Symbol='Olfr1340', Aliases='MOR258-6', description='olfactory receptor 1340', other_designations='GA_x6K02T2QD9B-18823451-18822504|olfactory receptor MOR258-6', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118726249, end_position_on_the_genomic_accession=118727196, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258302, CurrentID=0, Status='live', Symbol='Olfr420', Aliases='MOR105-10', description='olfactory receptor 420', other_designations='GA_x6K02T2P20D-20995211-20994246|olfactory receptor MOR105-10', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174158722, end_position_on_the_genomic_accession=174159767, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258303 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258303, CurrentID=0, Status='live', Symbol='Olfr518', Aliases='MOR268-5', description='olfactory receptor 518', other_designations='GA_x6K02T2PBJ9-11211854-11210853|olfactory receptor MOR268-5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108880603, end_position_on_the_genomic_accession=108881604, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258305, CurrentID=0, Status='live', Symbol='Olfr1356', Aliases='MOR142-1, MOR142-2_p', description='olfactory receptor 1356', other_designations='GA_x6K02T2QGN0-2962025-2962987|olfactory receptor MOR142-1', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78846951, end_position_on_the_genomic_accession=78847913, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258306 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258306, CurrentID=0, Status='live', Symbol='Olfr1337', Aliases='MOR259-10, Olfr219-ps1', description='olfactory receptor 1337', other_designations='olfactory receptor 219, pseudogene 1', map_location='4 D2.1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118781636, end_position_on_the_genomic_accession=118782586, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258307 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258307, CurrentID=0, Status='live', Symbol='Olfr493', Aliases='MOR204-35', description='olfactory receptor 493', other_designations='GA_x6K02T2PBJ9-10676998-10676054|olfactory receptor 204-35|olfactory receptor MOR204-35', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108346035, end_position_on_the_genomic_accession=108346979, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258309, CurrentID=0, Status='live', Symbol='Olfr657', Aliases='MOR40-13', description='olfactory receptor 657', other_designations='GA_x6K02T2PBJ9-7263864-7264823|olfactory receptor MOR40-13', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104635676, end_position_on_the_genomic_accession=104636635, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258310 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258310, CurrentID=0, Status='live', Symbol='Olfr145', Aliases='K21, MOR161-6', description='olfactory receptor 145', other_designations='GA_x6K02T2PVTD-31578734-31579666|odorant receptor K21|olfactory receptor 161-6|olfactory receptor 7C|olfactory receptor MOR161-6', map_location='9 A4', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37897406, end_position_on_the_genomic_accession=37898338, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258311 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258311, CurrentID=0, Status='live', Symbol='Olfr601', Aliases='MOR24-3', description='olfactory receptor 601', other_designations='GA_x6K02T2PBJ9-6068534-6067590|olfactory receptor MOR24-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103358224, end_position_on_the_genomic_accession=103359213, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258314 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258314, CurrentID=0, Status='live', Symbol='Olfr725', Aliases='MOR246-3, MOR246-7_p', description='olfactory receptor 725', other_designations='GA_x6K02T2PMLR-5725741-5724776|olfactory receptor MOR246-3', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50034391, end_position_on_the_genomic_accession=50035500, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258315, CurrentID=0, Status='live', Symbol='Olfr767', Aliases='MOR115-1, MOR115-4', description='olfactory receptor 767', other_designations='GA_x6K02T2PULF-10765431-10764502|olfactory receptor MOR115-1|olfactory receptor MOR115-4', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129079032, end_position_on_the_genomic_accession=129079961, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258316 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258316, CurrentID=0, Status='live', Symbol='Olfr727', Aliases='MOR246-2', description='olfactory receptor 727', other_designations='GA_x6K02T2PMLR-5817082-5818056|olfactory receptor MOR246-2', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50126553, end_position_on_the_genomic_accession=50127587, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258317 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258317, CurrentID=0, Status='live', Symbol='Olfr1145', Aliases='MOR264-19', description='olfactory receptor 1145', other_designations='GA_x6K02T2Q125-49311440-49312384|olfactory receptor MOR264-19', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87809822, end_position_on_the_genomic_accession=87810799, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258318 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258318, CurrentID=0, Status='live', Symbol='Olfr186', Aliases='MOR183-9', description='olfactory receptor 186', other_designations='GA_x54KRFPKG5P-55257214-55256285|olfactory receptor 183-9|olfactory receptor MOR183-9', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59026976, end_position_on_the_genomic_accession=59027905, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258319 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258319, CurrentID=0, Status='live', Symbol='Olfr187', Aliases='MOR183-8', description='olfactory receptor 187', other_designations='GA_x54KRFPKG5P-55265713-55264787|olfactory receptor 183-8|olfactory receptor MOR183-8', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59035780, end_position_on_the_genomic_accession=59039749, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258321 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258321, CurrentID=0, Status='live', Symbol='Olfr809', Aliases='MOR108-4', description='olfactory receptor 809', other_designations='GA_x6K02T2PULF-11454600-11455541|olfactory receptor MOR108-4', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129775871, end_position_on_the_genomic_accession=129776857, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258323 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258323, CurrentID=0, Status='live', Symbol='Olfr943', Aliases='MOR171-44', description='olfactory receptor 943', other_designations='GA_x6K02T2PVTD-32881408-32882343|olfactory receptor MOR171-44', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39166310, end_position_on_the_genomic_accession=39185124, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258324 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258324, CurrentID=0, Status='live', Symbol='Olfr129', Aliases='MOR263-9', description='olfactory receptor 129', other_designations='GA_x6K02T2PSCP-2503741-2502776|olfactory receptor MOR263-9', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38054599, end_position_on_the_genomic_accession=38059873, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  258325 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258325, CurrentID=0, Status='live', Symbol='Olfr110', Aliases='MOR249-2', description='olfactory receptor 110', other_designations='olfactory receptor MOR249-2', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37492468, end_position_on_the_genomic_accession=37499674, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258326 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258326, CurrentID=0, Status='live', Symbol='Olfr642', Aliases='MOR13-6', description='olfactory receptor 642', other_designations='G-protein-coupled receptor|GA_x6K02T2PBJ9-6784380-6783436|odorant receptor S85|olfactory receptor MOR13-6', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104049408, end_position_on_the_genomic_accession=104050352, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258327 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258327, CurrentID=0, Status='live', Symbol='Olfr958', Aliases='MOR224-9', description='olfactory receptor 958', other_designations='GA_x6K02T2PVTD-33247839-33246901|olfactory receptor 224-9|olfactory receptor MOR224-9', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39549931, end_position_on_the_genomic_accession=39550869, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258328 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258328, CurrentID=0, Status='live', Symbol='Olfr954', Aliases='MOR171-42, MOR171-53', description='olfactory receptor 954', other_designations='GA_x6K02T2PVTD-33158015-33158950|olfactory receptor MOR171-42', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39461433, end_position_on_the_genomic_accession=39462377, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258329 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258329, CurrentID=0, Status='live', Symbol='Olfr135', Aliases='MOR256-48', description='olfactory receptor 135', other_designations='GA_x6K02T2PSCP-2656648-2657586|olfactory receptor MOR256-48', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38208247, end_position_on_the_genomic_accession=38209185, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258330 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258330, CurrentID=0, Status='live', Symbol='Olfr1274-ps', Aliases='GA_x5J8B7W2BV0-3116-4045, MOR228-4, MOR228-5, Olfr1274', description='olfactory receptor 1274, pseudogene', other_designations='GA_x6K02T2Q125-51676372-51677301|olfactory receptor GA_x5J8B7W2BV0-3116-4045|olfactory receptor MOR228-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90400640, end_position_on_the_genomic_accession=90401674, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258331 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258331, CurrentID=0, Status='live', Symbol='Olfr1330', Aliases='MOR259-8', description='olfactory receptor 1330', other_designations='GA_x6K02T2QD9B-18644371-18643424|olfactory receptor MOR259-8', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118893085, end_position_on_the_genomic_accession=118894032, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258333 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258333, CurrentID=0, Status='live', Symbol='Olfr893', Aliases='MOR170-11', description='olfactory receptor 893', other_designations='GA_x6K02T2PVTD-31889215-31890153|olfactory receptor MOR170-11', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38209052, end_position_on_the_genomic_accession=38209993, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258334 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258334, CurrentID=0, Status='live', Symbol='Olfr1396', Aliases='MOR276-2', description='olfactory receptor 1396', other_designations='GA_x6K02T2QP88-6321048-6321995|olfactory receptor MOR276-2', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49111505, end_position_on_the_genomic_accession=49119817, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  258335 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258335, CurrentID=0, Status='live', Symbol='Olfr374', Aliases='MOR130-1', description='olfactory receptor 374', other_designations='olfactory receptor MOR130-1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72107040, end_position_on_the_genomic_accession=72110509, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258336 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258336, CurrentID=0, Status='live', Symbol='Olfr77', Aliases='18A, MOR143-1', description='olfactory receptor 77', other_designations='GA_x6K02T2PVTD-13660026-13660964|olfactory receptor MOR143-1', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19917229, end_position_on_the_genomic_accession=19922510, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258339 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258339, CurrentID=0, Status='live', Symbol='Olfr1269', Aliases='MOR228-3', description='olfactory receptor 1269', other_designations='GA_x6K02T2Q125-51551618-51550689|olfactory receptor MOR228-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90118667, end_position_on_the_genomic_accession=90119596, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258341 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258341, CurrentID=0, Status='live', Symbol='Olfr1495', Aliases='MOR266-9', description='olfactory receptor 1495', other_designations='GA_x6K02T2RE5P-4101369-4102328|olfactory receptor MOR266-9', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13768344, end_position_on_the_genomic_accession=13769303, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258343 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258343, CurrentID=0, Status='live', Symbol='Olfr397', Aliases='MOR135-28', description='olfactory receptor 397', other_designations='GA_x6K02T2P1NL-4121434-4122381|olfactory receptor MOR135-28', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73964610, end_position_on_the_genomic_accession=73965557, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258350 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258350, CurrentID=0, Status='live', Symbol='Olfr706', Aliases='MOR283-11', description='olfactory receptor 706', other_designations='GA_x6K02T2PBJ9-9271198-9270248|T1 olfactory receptor|olfactory receptor MOR283-11', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106885865, end_position_on_the_genomic_accession=106886815, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258351 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258351, CurrentID=0, Status='live', Symbol='Olfr633', Aliases='MOR12-5', description='olfactory receptor 633', other_designations='GA_x6K02T2PBJ9-6681230-6682168|olfactory receptor MOR12-5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103946568, end_position_on_the_genomic_accession=103947506, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258352 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258352, CurrentID=0, Status='live', Symbol='Olfr692', Aliases='MOR36-1', description='olfactory receptor 692', other_designations='GA_x6K02T2PBJ9-7994144-7995106|olfactory receptor MOR36-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105368328, end_position_on_the_genomic_accession=105369317, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258353 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258353, CurrentID=0, Status='live', Symbol='Olfr521', Aliases='MOR101-2', description='olfactory receptor 521', other_designations='GA_x6K02T2PBJ9-2443810-2444775|olfactory receptor MOR101-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99767164, end_position_on_the_genomic_accession=99768130, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258354 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258354, CurrentID=0, Status='live', Symbol='Olfr168', Aliases='MOR271-1', description='olfactory receptor 168', other_designations='GA_x54KRFPKG5P-15979009-15978071|olfactory receptor MOR271-1', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19529980, end_position_on_the_genomic_accession=19530918, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258355 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258355, CurrentID=0, Status='live', Symbol='Olfr677', Aliases='MOR32-11', description='olfactory receptor 677', other_designations='GA_x6K02T2PBJ9-7685262-7686200|olfactory receptor MOR32-11', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105056248, end_position_on_the_genomic_accession=105057186, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258356 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258356, CurrentID=0, Status='live', Symbol='Olfr564', Aliases='MOR14-10', description='olfactory receptor 564', other_designations='GA_x6K02T2PBJ9-5513635-5514627|olfactory receptor MOR14-10', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102803480, end_position_on_the_genomic_accession=102804430, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258357 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258357, CurrentID=0, Status='live', Symbol='Olfr574', Aliases='MOR14-9', description='olfactory receptor 574', other_designations='GA_x6K02T2PBJ9-5659738-5660748|olfactory receptor MOR14-9', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102946411, end_position_on_the_genomic_accession=102949507, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258358 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258358, CurrentID=0, Status='live', Symbol='Olfr557', Aliases='MOR18-3', description='olfactory receptor 557', other_designations='GA_x6K02T2PBJ9-5414934-5415902|olfactory receptor MOR18-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102698216, end_position_on_the_genomic_accession=102699272, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258360 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258360, CurrentID=0, Status='live', Symbol='Olfr731', Aliases='MOR246-5', description='olfactory receptor 731', other_designations='GA_x6K02T2PMLR-5936117-5935137|olfactory receptor MOR246-5', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50237903, end_position_on_the_genomic_accession=50238883, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258362, CurrentID=0, Status='live', Symbol='Olfr1094', Aliases='MOR179-7', description='olfactory receptor 1094', other_designations='GA_x6K02T2Q125-48321457-48322449|olfactory receptor 179-7|olfactory receptor MOR179-7', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86828676, end_position_on_the_genomic_accession=86829822, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258363 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258363, CurrentID=0, Status='live', Symbol='Olfr1093', Aliases='MOR179-6', description='olfactory receptor 1093', other_designations='GA_x6K02T2Q125-48278396-48279364|olfactory receptor MOR179-6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86782546, end_position_on_the_genomic_accession=86786785, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258364 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258364, CurrentID=0, Status='live', Symbol='Olfr976', Aliases='MOR224-10', description='olfactory receptor 976', other_designations='GA_x6K02T2PVTD-33657378-33656440|olfactory receptor MOR224-10', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39954257, end_position_on_the_genomic_accession=39961044, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258365 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258365, CurrentID=0, Status='live', Symbol='Olfr361', Aliases='MOR159-3', description='olfactory receptor 361', other_designations='GA_x6K02T2NLDC-33777519-33776551|olfactory receptor MOR159-3', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37084778, end_position_on_the_genomic_accession=37085746, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258366 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258366, CurrentID=0, Status='live', Symbol='Olfr434', Aliases='MOR261-10', description='olfactory receptor 434', other_designations='GA_x6K02T2P3E9-4341246-4340281|olfactory receptor MOR261-10', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43216915, end_position_on_the_genomic_accession=43217880, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258369 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258369, CurrentID=0, Status='live', Symbol='Olfr509', Aliases='MOR267-14', description='olfactory receptor 509', other_designations='GA_x6K02T2PBJ9-10976304-10975339|olfactory receptor MOR267-14', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108645609, end_position_on_the_genomic_accession=108646574, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258371 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258371, CurrentID=0, Status='live', Symbol='Olfr368', Aliases='MOR178-1', description='olfactory receptor 368', other_designations='GA_x6K02T2NLDC-34015743-34016726|olfactory receptor MOR178-1', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37331749, end_position_on_the_genomic_accession=37332732, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258373 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258373, CurrentID=0, Status='live', Symbol='Olfr323', Aliases='MOR107-1', description='olfactory receptor 323', other_designations='GA_x6K02T2NKPP-794386-795357|olfactory receptor MOR107-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58625073, end_position_on_the_genomic_accession=58626044, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258374 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258374, CurrentID=0, Status='live', Symbol='Olfr127', Aliases='MOR218-7', description='olfactory receptor 127', other_designations='GA_x6K02T2PSCP-2354126-2355093|olfactory receptor MOR218-7', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37903548, end_position_on_the_genomic_accession=37904519, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258377 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258377, CurrentID=0, Status='live', Symbol='Olfr654', Aliases='MOR38-2', description='olfactory receptor 654', other_designations='GA_x6K02T2PBJ9-7215221-7216195|olfactory receptor MOR38-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104579492, end_position_on_the_genomic_accession=104588880, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258379 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258379, CurrentID=0, Status='live', Symbol='Olfr1284', Aliases='MOR245-13', description='olfactory receptor 1284', other_designations='GA_x6K02T2Q125-72430580-72431515|olfactory receptor MOR245-13', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111379002, end_position_on_the_genomic_accession=111379937, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258380 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258380, CurrentID=0, Status='live', Symbol='Olfr461', Aliases='MOR120-3', description='olfactory receptor 461', other_designations='GA_x6K02T2P3E9-6973252-6974193|olfactory receptor MOR120-3', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40544036, end_position_on_the_genomic_accession=40544977, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258381 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258381, CurrentID=0, Status='live', Symbol='Olfr460', Aliases='MOR120-2', description='olfactory receptor 460', other_designations='GA_x6K02T2P3E9-6947292-6946348|olfactory receptor MOR120-2', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=40569152, end_position_on_the_genomic_accession=40572504, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258383 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258383, CurrentID=0, Status='live', Symbol='Olfr1347', Aliases='MOR103-11', description='olfactory receptor 1347', other_designations='GA_x6K02T2QGBW-3218686-3217748|MOR103-19_i|olfactory receptor MOR103-11', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6487913, end_position_on_the_genomic_accession=6488872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258385 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258385, CurrentID=0, Status='live', Symbol='Olfr1323', Aliases='MOR216-1', description='olfactory receptor 1323', other_designations='GA_x6K02T2Q64A-12051754-12050825|olfactory receptor MOR216-1', map_location='X A4', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=50009306, end_position_on_the_genomic_accession=50010235, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258387 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258387, CurrentID=0, Status='live', Symbol='Olfr720', Aliases='MOR274-2', description='olfactory receptor 720', other_designations='GA_x6K02T2PLTE-6544896-6543946|olfactory receptor MOR274-2', map_location='14 A2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=14175130, end_position_on_the_genomic_accession=14176080, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258388 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258388, CurrentID=0, Status='live', Symbol='Olfr1279', Aliases='MOR245-12', description='olfactory receptor 1279', other_designations='GA_x6K02T2Q125-72357646-72358581|olfactory receptor MOR245-12', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111306201, end_position_on_the_genomic_accession=111307142, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258389 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258389, CurrentID=0, Status='live', Symbol='Olfr1278', Aliases='MOR245-11', description='olfactory receptor 1278', other_designations='GA_x6K02T2Q125-72343713-72344654|olfactory receptor MOR245-11', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111292270, end_position_on_the_genomic_accession=111293211, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258390 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258390, CurrentID=0, Status='live', Symbol='Olfr1276', Aliases='MOR245-10', description='olfactory receptor 1276', other_designations='GA_x6K02T2Q125-72308574-72309512|olfactory receptor MOR245-10', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111257117, end_position_on_the_genomic_accession=111258055, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258391 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258391, CurrentID=0, Status='live', Symbol='Olfr1277', Aliases='MOR248-11', description='olfactory receptor 1277', other_designations='GA_x6K02T2Q125-72321818-72320907|olfactory receptor MOR248-11', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111269454, end_position_on_the_genomic_accession=111270365, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258392 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258392, CurrentID=0, Status='live', Symbol='Olfr190', Aliases='MOR183-4', description='olfactory receptor 190', other_designations='GA_x54KRFPKG5P-55303207-55302284|olfactory receptor MOR183-4', map_location='16 C1.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59074155, end_position_on_the_genomic_accession=59075078, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258393 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258393, CurrentID=0, Status='live', Symbol='Olfr1325', Aliases='MOR102-1', description='olfactory receptor 1325', other_designations='GA_x6K02T2QD78-203513-202575|olfactory receptor MOR102-1', map_location='X A7.3', chromosome='X', genomic_nucleotide_accession_version='NC_000086.7', start_position_on_the_genomic_accession=74594328, end_position_on_the_genomic_accession=74595275, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258394 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258394, CurrentID=0, Status='live', Symbol='Olfr1328', Aliases='MOR259-1, MOR259-13, Olfr1519', description='olfactory receptor 1328', other_designations='GA_x6K02T2QD9B-18602750-18603691|olfactory receptor MOR259-1', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118933899, end_position_on_the_genomic_accession=118934840, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258395 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258395, CurrentID=0, Status='live', Symbol='Olfr1288', Aliases='MOR245-9', description='olfactory receptor 1288', other_designations='GA_x6K02T2Q125-72530279-72531217|olfactory receptor MOR245-9', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111478786, end_position_on_the_genomic_accession=111479724, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258396, CurrentID=0, Status='live', Symbol='Olfr1305', Aliases='MOR245-8', description='olfactory receptor 1305', other_designations='olfactory receptor MOR245-8', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111872915, end_position_on_the_genomic_accession=111873853, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258397 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258397, CurrentID=0, Status='live', Symbol='Olfr1303', Aliases='MOR245-7, Olfr276, Olfr276-ps1', description='olfactory receptor 1303', other_designations='GA_x6K02T2N82Q-3465-3764|GA_x6K02T2Q125-72882187-72881249|MOR245-28_p|olfactory receptor 276, pseudogene 1|olfactory receptor MOR245-7', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111813786, end_position_on_the_genomic_accession=111814724, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258398 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258398, CurrentID=0, Status='live', Symbol='Olfr1295', Aliases='MOR248-10', description='olfactory receptor 1295', other_designations='GA_x6K02T2Q125-72616944-72616006|olfactory receptor MOR248-10', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111564504, end_position_on_the_genomic_accession=111569591, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258403 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258403, CurrentID=0, Status='live', Symbol='Olfr1065', Aliases='MOR190-1', description='olfactory receptor 1065', other_designations='GA_x6K02T2Q125-47915274-47914333|odorant receptor B1|olfactory receptor MOR190-1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86445013, end_position_on_the_genomic_accession=86446004, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258405 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258405, CurrentID=0, Status='live', Symbol='Olfr1420', Aliases='MOR266-4', description='olfactory receptor 1420', other_designations='GA_x6K02T2RE5P-2247227-2248156|olfactory receptor MOR266-4', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11896023, end_position_on_the_genomic_accession=11896952, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258406 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258406, CurrentID=0, Status='live', Symbol='Olfr462', Aliases='MOR240-3', description='olfactory receptor 462', other_designations='GA_x6K02T2PAEV-9536824-9535889|olfactory receptor MOR240-3', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87888959, end_position_on_the_genomic_accession=87889894, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258407 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258407, CurrentID=0, Status='live', Symbol='Olfr464', Aliases='MOR240-2', description='olfactory receptor 464', other_designations='GA_x6K02T2PAEV-9555122-9554181|olfactory receptor MOR240-2', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87913920, end_position_on_the_genomic_accession=87915003, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258408, CurrentID=0, Status='live', Symbol='Olfr463', Aliases='MOR240-1', description='olfactory receptor 463', other_designations='GA_x6K02T2PAEV-9540823-9539888|olfactory receptor MOR240-1', map_location='11 B5', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87892987, end_position_on_the_genomic_accession=87893922, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258409, CurrentID=0, Status='live', Symbol='Olfr1431', Aliases='MOR214-5', description='olfactory receptor 1431', other_designations='GA_x6K02T2RE5P-2573738-2574676|olfactory receptor MOR214-5', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12209568, end_position_on_the_genomic_accession=12210506, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258410 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258410, CurrentID=0, Status='live', Symbol='Olfr291', Aliases='MOR254-2', description='olfactory receptor 291', other_designations='GA_x6K02T2NHDJ-11231385-11230438|olfactory receptor MOR254-2', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=84853571, end_position_on_the_genomic_accession=84857318, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258411 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258411, CurrentID=0, Status='live', Symbol='Olfr290', Aliases='MOR254-1', description='olfactory receptor 290', other_designations='GA_x6K02T2NHDJ-11170115-11169168|olfactory receptor MOR254-1', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=84915781, end_position_on_the_genomic_accession=84916728, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258412 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258412, CurrentID=0, Status='live', Symbol='Olfr877', Aliases='MOR161-5', description='olfactory receptor 877', other_designations='GA_x6K02T2PVTD-31540342-31541277|olfactory receptor MOR161-5', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37854820, end_position_on_the_genomic_accession=37855755, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258415 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258415, CurrentID=0, Status='live', Symbol='Olfr887', Aliases='C77344, MOR162-5', description='olfactory receptor 887', other_designations='GA_x6K02T2PVTD-31764095-31765024|olfactory receptor MOR162-5', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38084838, end_position_on_the_genomic_accession=38085767, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258416 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258416, CurrentID=0, Status='live', Symbol='Olfr888', Aliases='MOR162-4', description='olfactory receptor 888', other_designations='GA_x6K02T2PVTD-31787920-31788864|olfactory receptor MOR162-4', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38108128, end_position_on_the_genomic_accession=38109632, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258418 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258418, CurrentID=0, Status='live', Symbol='Olfr469', Aliases='MOR204-21', description='olfactory receptor 469', other_designations='GA_x6K02T2PBJ9-10152980-10152036|olfactory receptor 204-21|olfactory receptor MOR204-21', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107822523, end_position_on_the_genomic_accession=107823467, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258420 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258420, CurrentID=0, Status='live', Symbol='Olfr221', Aliases='MOR205-1, MOR205-2, Olfr1514-ps1, Olfr221', description='olfactory receptor 221', other_designations='GA_x6K02SYYB8M-929-258|GA_x6K02T2RJGY-959918-960853|olfactory receptor 1514|olfactory receptor 221, pseudogene 1|olfactory receptor MOR205-1', map_location='14 C1-C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52035174, end_position_on_the_genomic_accession=52036109, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258421 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258421, CurrentID=0, Status='live', Symbol='Olfr223', Aliases='MOR256-30', description='olfactory receptor 223', other_designations='GA_x6K02T000SA-78-1037|novel 7 transmembrane receptor (rhodopsin family) domain-containing protein|olfactory receptor MOR256-30', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=59589128, end_position_on_the_genomic_accession=59590087, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258422 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258422, CurrentID=0, Status='live', Symbol='Olfr742', Aliases='MOR106-6', description='olfactory receptor 742', other_designations='GA_x6K02T2PMLR-6224293-6225228|olfactory receptor MOR106-6', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50515121, end_position_on_the_genomic_accession=50516240, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258423, CurrentID=0, Status='live', Symbol='Olfr1510', Aliases='MOR223-6', description='olfactory receptor 1510', other_designations='GA_x6K02T2RJGY-583652-584608|olfactory receptor MOR223-6', map_location='14 C1-C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52409889, end_position_on_the_genomic_accession=52410927, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258424 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258424, CurrentID=0, Status='live', Symbol='Olfr1512', Aliases='MOR223-5', description='olfactory receptor 1512', other_designations='GA_x6K02T2RJGY-622120-623061|olfactory receptor MOR223-5', map_location='14 C1-C2', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=52372110, end_position_on_the_genomic_accession=52373051, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258426 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258426, CurrentID=0, Status='live', Symbol='Olfr995', Aliases='MOR203-3', description='olfactory receptor 995', other_designations='GA_x6K02T2Q125-46915844-46914897|olfactory receptor MOR203-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85438209, end_position_on_the_genomic_accession=85439156, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258428 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258428, CurrentID=0, Status='live', Symbol='Olfr998', Aliases='MOR175-5', description='olfactory receptor 998', other_designations='GA_x6K02T2Q125-47069356-47070300|olfactory receptor 175-5|olfactory receptor MOR175-5', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85590510, end_position_on_the_genomic_accession=85591521, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258429 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258429, CurrentID=0, Status='live', Symbol='Olfr996', Aliases='MOR175-4', description='olfactory receptor 996', other_designations='GA_x6K02T2Q125-47058060-47059004|olfactory receptor MOR175-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85579215, end_position_on_the_genomic_accession=85580282, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258431 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258431, CurrentID=0, Status='live', Symbol='Olfr937', Aliases='MOR171-24', description='olfactory receptor 937', other_designations='GA_x6K02T2PVTD-32756567-32755632|olfactory receptor MOR171-24', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39059729, end_position_on_the_genomic_accession=39061229, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258433 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258433, CurrentID=0, Status='live', Symbol='Olfr933', Aliases='MOR171-22', description='olfactory receptor 933', other_designations='GA_x6K02T2PVTD-32671531-32672457|olfactory receptor MOR171-22', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38972313, end_position_on_the_genomic_accession=38976604, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258434 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258434, CurrentID=0, Status='live', Symbol='Olfr934', Aliases='MOR224-6', description='olfactory receptor 934', other_designations='GA_x6K02T2PVTD-32678895-32677963|olfactory receptor MOR224-6', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38982110, end_position_on_the_genomic_accession=38983042, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258435 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258435, CurrentID=0, Status='live', Symbol='Olfr382', Aliases='MOR135-14', description='olfactory receptor 382', other_designations='GA_x6K02T2P1NL-3676608-3675670|MOR135-31_p|olfactory receptor MOR135-14', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73516259, end_position_on_the_genomic_accession=73517197, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258436 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258436, CurrentID=0, Status='live', Symbol='Olfr458', Aliases='MOR211-8P, MOR221-4, MOR257-4, Olfr301', description='olfactory receptor 458', other_designations='GA_x6K02T2P3E9-5100053-5100994|olfactory receptor MOR257-4', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42460076, end_position_on_the_genomic_accession=42461017, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258437 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258437, CurrentID=0, Status='live', Symbol='Olfr450', Aliases='MOR257-3', description='olfactory receptor 450', other_designations='GA_x6K02T2P3E9-4742413-4741481|olfactory receptor MOR257-3', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42817473, end_position_on_the_genomic_accession=42818405, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258438 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258438, CurrentID=0, Status='live', Symbol='Olfr215', Aliases='MOR119-2', description='olfactory receptor 215', other_designations='GA_x54KRFPKN04-58217732-58216800|olfactory receptor MOR119-2', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116582012, end_position_on_the_genomic_accession=116582944, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258439 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258439, CurrentID=0, Status='live', Symbol='Olfr1309', Aliases='MOR245-5', description='olfactory receptor 1309', other_designations='GA_x6K02T2Q125-73031456-73030518|olfactory receptor MOR245-5', map_location='2 F2-F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111983134, end_position_on_the_genomic_accession=111984096, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258441 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258441, CurrentID=0, Status='live', Symbol='Olfr1310', Aliases='MOR245-3', description='olfactory receptor 1310', other_designations='GA_x6K02T2Q125-73056609-73055671|olfactory receptor MOR245-3', map_location='2 F2-F3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112008246, end_position_on_the_genomic_accession=112009184, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258443 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258443, CurrentID=0, Status='live', Symbol='Olfr164', Aliases='MOR279-2', description='olfactory receptor 164', other_designations='GA_x54KRFPKG5P-15738260-15737319|olfactory receptor MOR279-2', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19285794, end_position_on_the_genomic_accession=19286741, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258444 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258444, CurrentID=0, Status='live', Symbol='Olfr694', Aliases='MOR283-9', description='olfactory receptor 694', other_designations='GA_x6K02T2PBJ9-9067220-9066273|olfactory receptor MOR283-9', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106688782, end_position_on_the_genomic_accession=106691307, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258446 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258446, CurrentID=0, Status='live', Symbol='Olfr1231', Aliases='MOR235-2', description='olfactory receptor 1231', other_designations='GA_x6K02T2Q125-50748233-50747292|olfactory receptor MOR235-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89302649, end_position_on_the_genomic_accession=89303590, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258448, CurrentID=0, Status='live', Symbol='Olfr92', Aliases='MOR256-29, bM573K1.4', description='olfactory receptor 92', other_designations='GA_x6K02T2PSCP-1552066-1551128|olfactory receptor MOR256-29', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37111042, end_position_on_the_genomic_accession=37111980, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258449 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258449, CurrentID=0, Status='live', Symbol='Olfr282', Aliases='MOR160-2', description='olfactory receptor 282', other_designations='GA_x6K02T2NBG7-5293798-5292872|olfactory receptor MOR160-2', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98437470, end_position_on_the_genomic_accession=98438397, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258451 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258451, CurrentID=0, Status='live', Symbol='Olfr1215', Aliases='MOR233-13', description='olfactory receptor 1215', other_designations='GA_x6K02T2Q125-50482823-50481885|olfactory receptor MOR233-13', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89001348, end_position_on_the_genomic_accession=89002290, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258453 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258453, CurrentID=0, Status='live', Symbol='Olfr1209', Aliases='MOR230-7', description='olfactory receptor 1209', other_designations='GA_x6K02T2Q125-50386882-50385950|olfactory receptor MOR230-7', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88909459, end_position_on_the_genomic_accession=88910391, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258457, CurrentID=0, Status='live', Symbol='Olfr108', Aliases='MOR156-5', description='olfactory receptor 108', other_designations='GA_x6K02T2PSCP-1893605-1894534|olfactory receptor MOR156-5', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37445480, end_position_on_the_genomic_accession=37446517, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258458, CurrentID=0, Status='live', Symbol='Olfr165', Aliases='MOR279-1', description='olfactory receptor 165', other_designations='GA_x54KRFPKG5P-15855164-15854223|olfactory receptor MOR279-1', map_location='16 B1-B4', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19406991, end_position_on_the_genomic_accession=19445258, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  258459 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258459, CurrentID=0, Status='live', Symbol='Olfr1388', Aliases='MOR256-28', description='olfactory receptor 1388', other_designations='GA_x6K02T2QP88-5991012-5990077|olfactory receptor MOR256-28', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49443853, end_position_on_the_genomic_accession=49444788, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258460 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258460, CurrentID=0, Status='live', Symbol='Olfr1391', Aliases='MOR256-27', description='olfactory receptor 1391', other_designations='GA_x6K02T2QP88-6107233-6106298|olfactory receptor MOR256-27', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49327413, end_position_on_the_genomic_accession=49328348, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258461 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258461, CurrentID=0, Status='live', Symbol='Olfr1381', Aliases='MOR256-26', description='olfactory receptor 1381', other_designations='GA_x6K02T2QP88-5884501-5883566|olfactory receptor MOR256-26', map_location='11 B1.2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49551749, end_position_on_the_genomic_accession=49552684, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258462, CurrentID=0, Status='live', Symbol='Olfr1392', Aliases='MOR256-25', description='olfactory receptor 1392', other_designations='GA_x6K02T2QP88-6141322-6140387|olfactory receptor MOR256-25', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49293297, end_position_on_the_genomic_accession=49294322, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258463 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258463, CurrentID=0, Status='live', Symbol='Olfr1393', Aliases='MOR256-24', description='olfactory receptor 1393', other_designations='GA_x6K02T2QP88-6154577-6153642|MOR256-67_i|olfactory receptor MOR256-24', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49264858, end_position_on_the_genomic_accession=49281429, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  258464 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258464, CurrentID=0, Status='live', Symbol='Olfr1384', Aliases='MOR256-23', description='olfactory receptor 1384', other_designations='GA_x6K02T2QP88-5922712-5921756|olfactory receptor MOR256-23', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49513618, end_position_on_the_genomic_accession=49514651, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258466 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258466, CurrentID=0, Status='live', Symbol='Olfr1261', Aliases='MOR234-3', description='olfactory receptor 1261', other_designations='GA_x6K02T2Q125-51425355-51426275|olfactory receptor MOR234-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89993395, end_position_on_the_genomic_accession=89994315, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258469 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258469, CurrentID=0, Status='live', Symbol='Olfr90', Aliases='MOR256-21, bM573K1.2', description='olfactory receptor 90', other_designations='GA_x6K02T2PSCP-1526227-1525295|olfactory receptor MOR256-21', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37085191, end_position_on_the_genomic_accession=37086236, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258470, CurrentID=0, Status='live', Symbol='Olfr91', Aliases='MOR256-20, bM573K1.3', description='olfactory receptor 91', other_designations='G-protein coupled receptor|GA_x6K02T2PSCP-1533927-1532989|bM573K1.3 (novel 7 transmembrane receptor (rhodopsin family) (olfactory receptor LIKE) protein)|olfactory receptor MOR256-20', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37092934, end_position_on_the_genomic_accession=37093872, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258474 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258474, CurrentID=0, Status='live', Symbol='Olfr890', Aliases='MOR162-3', description='olfactory receptor 890', other_designations='GA_x6K02T2PVTD-31822365-31823309|MOR162-15_p|olfactory receptor MOR162-3', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38143137, end_position_on_the_genomic_accession=38144081, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258475 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258475, CurrentID=0, Status='live', Symbol='Olfr889', Aliases='MOR162-2', description='olfactory receptor 889', other_designations='GA_x6K02T2PVTD-31795028-31795957|olfactory receptor MOR162-2', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38115798, end_position_on_the_genomic_accession=38116727, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258478 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258478, CurrentID=0, Status='live', Symbol='Olfr183', Aliases='MOR183-2', description='olfactory receptor 183', other_designations='GA_x54KRFPKG5P-55229158-55230087|olfactory receptor 183-2|olfactory receptor MOR183-2', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58995404, end_position_on_the_genomic_accession=59000616, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258479 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258479, CurrentID=0, Status='live', Symbol='Olfr203', Aliases='MOR182-5', description='olfactory receptor 203', other_designations='GA_x54KRFPKG5P-55517445-55518365|olfactory receptor MOR182-5', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59303155, end_position_on_the_genomic_accession=59304075, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258480 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258480, CurrentID=0, Status='live', Symbol='Olfr130', Aliases='MOR256-19', description='olfactory receptor 130', other_designations='GA_x6K02T2PSCP-2515350-2516303|olfactory receptor MOR256-19', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38067173, end_position_on_the_genomic_accession=38068126, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258481 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258481, CurrentID=0, Status='live', Symbol='Olfr137', Aliases='MOR256-18', description='olfactory receptor 137', other_designations='GA_x6K02T2PSCP-2749525-2748587|olfactory receptor MOR256-18', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38304521, end_position_on_the_genomic_accession=38305459, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258482 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258482, CurrentID=0, Status='live', Symbol='Olfr266', Aliases='MOR122-2', description='olfactory receptor 266', other_designations='GA_x6K02T2N6GK-529983-529033|olfactory receptor MOR122-2', map_location='3 F2', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=106821607, end_position_on_the_genomic_accession=106822557, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258483 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258483, CurrentID=0, Status='live', Symbol='Olfr1411', Aliases='MOR208-3', description='olfactory receptor 1411', other_designations='GA_x6K02T2R7CC-81157497-81156526|olfactory receptor MOR208-3', map_location='1 D-E1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92596521, end_position_on_the_genomic_accession=92597492, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258484 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258484, CurrentID=0, Status='live', Symbol='Olfr1410', Aliases='MOR208-2', description='olfactory receptor 1410', other_designations='GA_x6K02T2R7CC-81146179-81145211|olfactory receptor MOR208-2', map_location='1 D-E1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92607839, end_position_on_the_genomic_accession=92608807, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258485 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258485, CurrentID=0, Status='live', Symbol='Olfr724', Aliases='MOR247-2', description='olfactory receptor 724', other_designations='GA_x6K02T2PMLR-5645801-5644872|olfactory receptor MOR247-2', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49960091, end_position_on_the_genomic_accession=49961096, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258486 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258486, CurrentID=0, Status='live', Symbol='Olfr730', Aliases='MOR247-1', description='olfactory receptor 730', other_designations='GA_x6K02T2PMLR-5881670-5880717|olfactory receptor MOR247-1', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50186262, end_position_on_the_genomic_accession=50187218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258487 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258487, CurrentID=0, Status='live', Symbol='Olfr722', Aliases='MOR241-3', description='olfactory receptor 722', other_designations='GA_x6K02T2PMLR-5566715-5565789|olfactory receptor MOR241-3', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49894258, end_position_on_the_genomic_accession=49901941, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258492 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258492, CurrentID=0, Status='live', Symbol='Olfr484', Aliases='MOR204-16', description='olfactory receptor 484', other_designations='GA_x6K02T2PBJ9-10454128-10453163|olfactory receptor 204-16|olfactory receptor MOR204-16', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108124296, end_position_on_the_genomic_accession=108125261, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258494 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258494, CurrentID=0, Status='live', Symbol='Olfr318', Aliases='MOR285-1', description='olfactory receptor 318', other_designations='GA_x6K02T2NKPP-692816-693736|olfactory receptor MOR285-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58720069, end_position_on_the_genomic_accession=58721094, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258495 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258495, CurrentID=0, Status='live', Symbol='Olfr328', Aliases='MOR275-2', description='olfactory receptor 328', other_designations='GA_x6K02T2NKPP-873285-874217|olfactory receptor MOR275-2', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58551305, end_position_on_the_genomic_accession=58552237, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258497 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258497, CurrentID=0, Status='live', Symbol='Olfr961', Aliases='MOR224-5', description='olfactory receptor 961', other_designations='GA_x6K02T2PVTD-33343617-33344561|olfactory receptor MOR224-5', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39646728, end_position_on_the_genomic_accession=39647672, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258498, CurrentID=0, Status='live', Symbol='Olfr148', Aliases='M30, MOR224-4', description='olfactory receptor 148', other_designations='GA_x6K02T2PVTD-33310486-33311418|odorant receptor M30|olfactory receptor 224-4|olfactory receptor 7F|olfactory receptor MOR224-4', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39613569, end_position_on_the_genomic_accession=39614501, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258499 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258499, CurrentID=0, Status='live', Symbol='Olfr945', Aliases='MOR171-20', description='olfactory receptor 945', other_designations='GA_x6K02T2PVTD-32955932-32954982|olfactory receptor MOR171-20', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39257720, end_position_on_the_genomic_accession=39258679, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258500 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258500, CurrentID=0, Status='live', Symbol='Olfr944', Aliases='MOR171-19', description='olfactory receptor 944', other_designations='GA_x6K02T2PVTD-32914568-32915503|olfactory receptor MOR171-19', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39217359, end_position_on_the_genomic_accession=39218294, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258501 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258501, CurrentID=0, Status='live', Symbol='Olfr959', Aliases='M34, MOR224-3, ORL583', description='olfactory receptor 959', other_designations='GA_x6K02T2PVTD-33270211-33269276|odorant receptor|olfactory receptor MOR224-3', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39572322, end_position_on_the_genomic_accession=39573257, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258502 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258502, CurrentID=0, Status='live', Symbol='Olfr279', Aliases='MOR122-1, Olfr234, Olfr234-ps1-ps1, Olfr279', description='olfactory receptor 279', other_designations='GA_x6K02T04SN1-554-3|GA_x6K02T2NBG7-5241885-5241019|olfactory receptor 234, pseudogene 1|olfactory receptor 279, pseudogene 1|olfactory receptor MOR122-1', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98492402, end_position_on_the_genomic_accession=98498406, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258503 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258503, CurrentID=0, Status='live', Symbol='Olfr98', Aliases='MOR156-4', description='olfactory receptor 98', other_designations='GA_x6K02T2PSCP-1703582-1702653|olfactory receptor MOR156-4', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37262733, end_position_on_the_genomic_accession=37263662, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258504 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258504, CurrentID=0, Status='live', Symbol='Olfr107', Aliases='MOR156-3', description='olfactory receptor 107', other_designations='GA_x6K02T2PSCP-1867165-1868094|olfactory receptor MOR156-3', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37405477, end_position_on_the_genomic_accession=37406509, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258505 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258505, CurrentID=0, Status='live', Symbol='Olfr97', Aliases='MOR156-2', description='olfactory receptor 97', other_designations='GA_x6K02T2PSCP-1672287-1671355|olfactory receptor MOR156-2', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37231436, end_position_on_the_genomic_accession=37232368, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258506 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258506, CurrentID=0, Status='live', Symbol='Olfr95', Aliases='MOR263-6', description='olfactory receptor 95', other_designations='GA_x6K02T2PSCP-1651760-1650822|olfactory receptor MOR263-6', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37210913, end_position_on_the_genomic_accession=37211851, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258507 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258507, CurrentID=0, Status='live', Symbol='Olfr96', Aliases='MOR121-1', description='olfactory receptor 96', other_designations='GA_x6K02T2PSCP-1665046-1665987|olfactory receptor MOR121-1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37225127, end_position_on_the_genomic_accession=37226068, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258508 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258508, CurrentID=0, Status='live', Symbol='Olfr99', Aliases='MOR156-1', description='olfactory receptor 99', other_designations='GA_x6K02T2PSCP-1720261-1719344|olfactory receptor MOR156-1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37279501, end_position_on_the_genomic_accession=37280418, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258511 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258511, CurrentID=0, Status='live', Symbol='Olfr523', Aliases='MOR104-4', description='olfactory receptor 523', other_designations='GA_x6K02T2PBJ9-42327937-42328872|odorant receptor|olfactory receptor MOR104-4', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140176104, end_position_on_the_genomic_accession=140177057, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258512 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258512, CurrentID=0, Status='live', Symbol='Olfr530', Aliases='MOR252-2', description='olfactory receptor 530', other_designations='GA_x6K02T2PBJ9-42523824-42522901|olfactory receptor MOR252-2', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140372685, end_position_on_the_genomic_accession=140373608, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258513 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258513, CurrentID=0, Status='live', Symbol='Olfr536', Aliases='MOR252-1, MOR252-4', description='olfactory receptor 536', other_designations='GA_x6K02T2PBJ9-42653501-42652578|olfactory receptor MOR252-1', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140500853, end_position_on_the_genomic_accession=140507318, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  258516 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258516, CurrentID=0, Status='live', Symbol='Olfr850', Aliases='MOR147-2', description='olfactory receptor 850', other_designations='GA_x6K02T2PVTD-13214162-13213224|olfactory receptor MOR147-2', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19477301, end_position_on_the_genomic_accession=19478248, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258517 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258517, CurrentID=0, Status='live', Symbol='Olfr855', Aliases='MOR148-1', description='olfactory receptor 855', other_designations='GA_x6K02T2PVTD-13330461-13331399|olfactory receptor MOR148-1', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19584511, end_position_on_the_genomic_accession=19585502, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258518 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258518, CurrentID=0, Status='live', Symbol='Olfr847', Aliases='MOR149-2', description='olfactory receptor 847', other_designations='GA_x6K02T2PVTD-13113073-13112135|olfactory receptor MOR149-2', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19374941, end_position_on_the_genomic_accession=19378669, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258519 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258519, CurrentID=0, Status='live', Symbol='Olfr859', Aliases='MOR146-3', description='olfactory receptor 859', other_designations='GA_x6K02T2PVTD-13548326-13549255|MOR146-10_p|olfactory receptor MOR146-3', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19808320, end_position_on_the_genomic_accession=19809249, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258521, CurrentID=0, Status='live', Symbol='Olfr860', Aliases='MOR146-2', description='olfactory receptor 860', other_designations='GA_x6K02T2PVTD-13586614-13585661|olfactory receptor MOR146-2', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19845623, end_position_on_the_genomic_accession=19849747, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258523 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258523, CurrentID=0, Status='live', Symbol='Olfr1186', Aliases='MOR230-5', description='olfactory receptor 1186', other_designations='GA_x6K02T2Q125-50012960-50013871|olfactory receptor MOR230-5', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88525501, end_position_on_the_genomic_accession=88526590, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258524, CurrentID=0, Status='live', Symbol='Olfr1168', Aliases='MOR174-13', description='olfactory receptor 1168', other_designations='GA_x6K02T2Q125-49669483-49670421|olfactory receptor MOR174-13', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88184879, end_position_on_the_genomic_accession=88185817, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258525 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258525, CurrentID=0, Status='live', Symbol='Olfr1170', Aliases='MOR174-12', description='olfactory receptor 1170', other_designations='GA_x6K02T2Q125-49709783-49708833|olfactory receptor MOR174-12', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88224080, end_position_on_the_genomic_accession=88225030, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258527 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258527, CurrentID=0, Status='live', Symbol='Olfr1368', Aliases='MOR256-15', description='olfactory receptor 1368', other_designations='GA_x6K02T2QHY8-12104556-12105500|olfactory receptor MOR256-15', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21142111, end_position_on_the_genomic_accession=21143055, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258528 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258528, CurrentID=0, Status='live', Symbol='Olfr1370', Aliases='MOR256-14', description='olfactory receptor 1370', other_designations='GA_x6K02T2QHY8-12181473-12182423|olfactory receptor MOR256-14', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21072349, end_position_on_the_genomic_accession=21073299, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258529 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258529, CurrentID=0, Status='live', Symbol='Olfr313', Aliases='MOR222-2', description='olfactory receptor 313', other_designations='GA_x6K02T2NKPP-590035-589109|olfactory receptor MOR222-2', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58816928, end_position_on_the_genomic_accession=58818001, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258531 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258531, CurrentID=0, Status='live', Symbol='Olfr315', Aliases='MOR278-1', description='olfactory receptor 315', other_designations='GA_x6K02T2NKPP-630272-629343|olfactory receptor MOR278-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58778089, end_position_on_the_genomic_accession=58779086, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258532 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258532, CurrentID=0, Status='live', Symbol='Olfr373', Aliases='MOR282-2', description='olfactory receptor 373', other_designations='olfactory receptor MOR282-1', map_location='8 B3.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=72099762, end_position_on_the_genomic_accession=72100706, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258533 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258533, CurrentID=0, Status='live', Symbol='Olfr1364', Aliases='MOR256-13', description='olfactory receptor 1364', other_designations='GA_x6K02T2QHY8-11663090-11664010|olfactory receptor MOR256-13', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21573510, end_position_on_the_genomic_accession=21581441, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258534 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258534, CurrentID=0, Status='live', Symbol='Olfr1361', Aliases='MOR256-12', description='olfactory receptor 1361', other_designations='GA_x6K02T2QHY8-11577590-11578528|olfactory receptor MOR256-12', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21658368, end_position_on_the_genomic_accession=21664523, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258536 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258536, CurrentID=0, Status='live', Symbol='Olfr1360', Aliases='MOR256-10, MOR256-35', description='olfactory receptor 1360', other_designations='GA_x6K02T2QHY8-11561962-11562903|olfactory receptor MOR256-10', map_location='13 A3-A4', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21674001, end_position_on_the_genomic_accession=21674980, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258537 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258537, CurrentID=0, Status='live', Symbol='Olfr777', Aliases='MOR114-9', description='olfactory receptor 777', other_designations='GA_x6K02T2PULF-10955551-10954616|olfactory receptor MOR114-9', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129268386, end_position_on_the_genomic_accession=129270550, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258541 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258541, CurrentID=0, Status='live', Symbol='Olfr800', Aliases='MOR114-7', description='olfactory receptor 800', other_designations='GA_x6K02T2PULF-11338429-11339364|olfactory receptor MOR114-7', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129659808, end_position_on_the_genomic_accession=129660743, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258542 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258542, CurrentID=0, Status='live', Symbol='Olfr786', Aliases='MOR111-5', description='olfactory receptor 786', other_designations='GA_x6K02T2PULF-11116958-11117896|olfactory receptor MOR111-5', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129436814, end_position_on_the_genomic_accession=129437752, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258543 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258543, CurrentID=0, Status='live', Symbol='Olfr810', Aliases='MOR113-4', description='olfactory receptor 810', other_designations='GA_x6K02T2PULF-11470271-11469333|olfactory receptor MOR113-4', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129790649, end_position_on_the_genomic_accession=129791587, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258546 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258546, CurrentID=0, Status='live', Symbol='Olfr806', Aliases='MOR110-12, MOR110-5', description='olfactory receptor 806', other_designations='GA_x6K02T2PULF-11417610-11416669|olfactory receptor MOR110-5', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129737974, end_position_on_the_genomic_accession=129738915, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258548 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258548, CurrentID=0, Status='live', Symbol='Olfr805', Aliases='MOR110-4', description='olfactory receptor 805', other_designations='GA_x6K02T2PULF-11402237-11401278|olfactory receptor MOR110-4', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129722583, end_position_on_the_genomic_accession=129723542, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258550 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258550, CurrentID=0, Status='live', Symbol='Olfr869', Aliases='MOR145-6', description='olfactory receptor 869', other_designations='GA_x6K02T2PVTD-13878275-13879204|olfactory receptor MOR145-6', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20129120, end_position_on_the_genomic_accession=20138089, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258551 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258551, CurrentID=0, Status='live', Symbol='Olfr866', Aliases='MOR145-5', description='olfactory receptor 866', other_designations='GA_x6K02T2PVTD-13768406-13767468|olfactory receptor MOR145-5', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20026959, end_position_on_the_genomic_accession=20028035, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258552 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258552, CurrentID=0, Status='live', Symbol='Olfr868', Aliases='MOR145-4', description='olfactory receptor 868', other_designations='GA_x6K02T2PVTD-13841888-13842802|olfactory receptor MOR145-4', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20098626, end_position_on_the_genomic_accession=20101690, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258553 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258553, CurrentID=0, Status='live', Symbol='Olfr872', Aliases='MOR145-3', description='olfactory receptor 872', other_designations='GA_x6K02T2PVTD-13999915-14000844|olfactory receptor MOR145-3', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20237162, end_position_on_the_genomic_accession=20260913, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258554, CurrentID=0, Status='live', Symbol='Olfr873', Aliases='MOR145-2', description='olfactory receptor 873', other_designations='GA_x6K02T2PVTD-14040245-14041204|olfactory receptor MOR145-2', map_location='9 A3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20300211, end_position_on_the_genomic_accession=20301170, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258555 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258555, CurrentID=0, Status='live', Symbol='Olfr862', Aliases='MOR146-1', description='olfactory receptor 862', other_designations='GA_x6K02T2PVTD-13624132-13623212|olfactory receptor MOR146-1', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19883383, end_position_on_the_genomic_accession=19884303, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258557 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258557, CurrentID=0, Status='live', Symbol='Olfr836', Aliases='MOR152-3', description='olfactory receptor 836', other_designations='GA_x6K02T2PVTD-12857805-12858749|olfactory receptor MOR152-3', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19120957, end_position_on_the_genomic_accession=19121911, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258558 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258558, CurrentID=0, Status='live', Symbol='Olfr837', Aliases='MOR152-2', description='olfactory receptor 837', other_designations='GA_x6K02T2PVTD-12873838-12874776|olfactory receptor MOR152-2', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19136928, end_position_on_the_genomic_accession=19137967, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258559 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258559, CurrentID=0, Status='live', Symbol='Olfr830', Aliases='MOR152-1', description='olfactory receptor 830', other_designations='GA_x6K02T2PVTD-12618399-12619337|olfactory receptor MOR152-1', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18874960, end_position_on_the_genomic_accession=18876276, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258560 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258560, CurrentID=0, Status='live', Symbol='Olfr843', Aliases='MOR155-2', description='olfactory receptor 843', other_designations='GA_x6K02T2PVTD-12986331-12985390|olfactory receptor MOR155-2', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=19248357, end_position_on_the_genomic_accession=19249463, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258561 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258561, CurrentID=0, Status='live', Symbol='Olfr1012', Aliases='MOR213-6', description='olfactory receptor 1012', other_designations='GA_x6K02T2Q125-47239120-47238185|olfactory receptor MOR213-6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85759439, end_position_on_the_genomic_accession=85760374, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258563 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258563, CurrentID=0, Status='live', Symbol='Olfr1006', Aliases='MOR213-4', description='olfactory receptor 1006', other_designations='GA_x6K02T2Q125-47154544-47153606|olfactory receptor MOR213-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85674211, end_position_on_the_genomic_accession=85678741, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258564 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258564, CurrentID=0, Status='live', Symbol='Olfr1015', Aliases='MOR213-3', description='olfactory receptor 1015', other_designations='GA_x6K02T2Q125-47264151-47265089|olfactory receptor MOR213-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85785474, end_position_on_the_genomic_accession=85786489, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258565 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258565, CurrentID=0, Status='live', Symbol='Olfr1009', Aliases='MOR175-3', description='olfactory receptor 1009', other_designations='GA_x6K02T2Q125-47195323-47196267|olfactory receptor 175-3|olfactory receptor MOR175-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85721407, end_position_on_the_genomic_accession=85722351, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258569 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258569, CurrentID=0, Status='live', Symbol='Olfr459', Aliases='MOR120-1', description='olfactory receptor 459', other_designations='GA_x6K02T2P3E9-5780974-5781915|olfactory receptor MOR120-1', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=41771353, end_position_on_the_genomic_accession=41772297, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258570 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258570, CurrentID=0, Status='live', Symbol='Olfr1043', Aliases='MOR185-7', description='olfactory receptor 1043', other_designations='GA_x6K02T2Q125-47631900-47630956|olfactory receptor MOR185-7', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86162003, end_position_on_the_genomic_accession=86162947, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258571 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258571, CurrentID=0, Status='live', Symbol='Olfr1033', Aliases='MOR199-2', description='olfactory receptor 1033', other_designations='GA_x6K02T2Q125-47516301-47517233|olfactory receptor MOR199-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86020640, end_position_on_the_genomic_accession=86044811, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  258572 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258572, CurrentID=0, Status='live', Symbol='Olfr1032', Aliases='MOR199-1', description='olfactory receptor 1032', other_designations='GA_x6K02T2Q125-47485813-47486745|olfactory receptor MOR199-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86007778, end_position_on_the_genomic_accession=86008710, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258573 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258573, CurrentID=0, Status='live', Symbol='Olfr1020', Aliases='MOR201-2', description='olfactory receptor 1020', other_designations='GA_x6K02T2Q125-47327964-47328917|olfactory receptor 201-2|olfactory receptor MOR201-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85849420, end_position_on_the_genomic_accession=85850498, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258575 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258575, CurrentID=0, Status='live', Symbol='Olfr1046', Aliases='MOR194-1', description='olfactory receptor 1046', other_designations='GA_x6K02T2Q125-47692494-47691544|olfactory receptor MOR194-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86216758, end_position_on_the_genomic_accession=86217708, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258579 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258579, CurrentID=0, Status='live', Symbol='Olfr1018', Aliases='MOR260-5', description='olfactory receptor 1018', other_designations='GA_x6K02T2Q125-47301584-47302519|olfactory receptor MOR260-5', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85822973, end_position_on_the_genomic_accession=85823908, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258581 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258581, CurrentID=0, Status='live', Symbol='Olfr1030', Aliases='MOR196-2', description='olfactory receptor 1030', other_designations='GA_x6K02T2Q125-47462755-47463693|olfactory receptor 196-2|olfactory receptor MOR196-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85979312, end_position_on_the_genomic_accession=85984798, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258584 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258584, CurrentID=0, Status='live', Symbol='Olfr1101', Aliases='MOR179-3', description='olfactory receptor 1101', other_designations='GA_x6K02T2Q125-48475870-48474938|olfactory receptor MOR179-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86988242, end_position_on_the_genomic_accession=86989174, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258586 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258586, CurrentID=0, Status='live', Symbol='Olfr1111', Aliases='MOR181-2', description='olfactory receptor 1111', other_designations='GA_x6K02T2Q125-48635468-48634530|olfactory receptor MOR181-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87149720, end_position_on_the_genomic_accession=87150659, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258587 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258587, CurrentID=0, Status='live', Symbol='Olfr1100', Aliases='MOR206-4', description='olfactory receptor 1100', other_designations='GA_x6K02T2Q125-48465387-48464422|olfactory receptor MOR206-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86977829, end_position_on_the_genomic_accession=86978794, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  231201 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=231201, CurrentID=0, Status='live', Symbol='AF366264', Aliases='Susp4', description='cDNA sequence AF366264', other_designations='SUMO-1-specific protease 4', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=13835229, end_position_on_the_genomic_accession=13838089, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258590 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258590, CurrentID=0, Status='live', Symbol='Olfr702', Aliases='MOR260-4', description='olfactory receptor 702', other_designations='B7 olfactory receptor|GA_x6K02T2PBJ9-9202245-9201289|olfactory receptor MOR260-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106823489, end_position_on_the_genomic_accession=106826743, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258591 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258591, CurrentID=0, Status='live', Symbol='Olfr695', Aliases='MOR283-6', description='olfactory receptor 695', other_designations='GA_x6K02T2PBJ9-9092181-9091234|olfactory receptor MOR283-6', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106713732, end_position_on_the_genomic_accession=106716345, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258593 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258593, CurrentID=0, Status='live', Symbol='Olfr700', Aliases='MOR283-4', description='olfactory receptor 700', other_designations='GA_x6K02T2PBJ9-9184187-9183237|olfactory receptor MOR283-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106805510, end_position_on_the_genomic_accession=106806460, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258594 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258594, CurrentID=0, Status='live', Symbol='Olfr710', Aliases='MOR260-3', description='olfactory receptor 710', other_designations='B3 olfactory receptor|GA_x6K02T2PBJ9-9325348-9324416|olfactory receptor MOR260-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106944067, end_position_on_the_genomic_accession=106948312, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258595 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258595, CurrentID=0, Status='live', Symbol='Olfr698', Aliases='MOR283-3', description='olfactory receptor 698', other_designations='GA_x6K02T2PBJ9-9130754-9129519|olfactory receptor MOR283-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106752451, end_position_on_the_genomic_accession=106753386, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258597 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258597, CurrentID=0, Status='live', Symbol='Olfr716', Aliases='MOR260-2', description='olfactory receptor 716', other_designations='B1 olfactory receptor|GA_x6K02T2PBJ9-9497411-9498355|olfactory receptor MOR260-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107147318, end_position_on_the_genomic_accession=107148262, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258598 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258598, CurrentID=0, Status='live', Symbol='Olfr828', Aliases='MOR149-1', description='olfactory receptor 828', other_designations='GA_x6K02T2PVTD-12559294-12558356|olfactory receptor MOR149-1', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18815354, end_position_on_the_genomic_accession=18816292, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258600 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258600, CurrentID=0, Status='live', Symbol='Olfr270', Aliases='MOR262-9', description='olfactory receptor 270', other_designations='GA_x6K02T2N78B-7025964-7025020|olfactory receptor MOR262-9', map_location='4 B2-C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=52970629, end_position_on_the_genomic_accession=52971567, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258601 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258601, CurrentID=0, Status='live', Symbol='Olfr984', Aliases='MOR239-6', description='olfactory receptor 984', other_designations='GA_x6K02T2PVTD-33799484-33798540|olfactory receptor MOR239-6', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40100544, end_position_on_the_genomic_accession=40156029, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  258603 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258603, CurrentID=0, Status='live', Symbol='Olfr972', Aliases='MOR171-17', description='olfactory receptor 972', other_designations='GA_x6K02T2PVTD-33572803-33573747|olfactory receptor MOR171-17', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39873277, end_position_on_the_genomic_accession=39874221, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258604 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258604, CurrentID=0, Status='live', Symbol='Olfr970', Aliases='MOR171-16', description='olfactory receptor 970', other_designations='GA_x6K02T2PVTD-33517322-33518257|olfactory receptor MOR171-16', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39819641, end_position_on_the_genomic_accession=39820576, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258607 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258607, CurrentID=0, Status='live', Symbol='Olfr971', Aliases='MOR171-13', description='olfactory receptor 971', other_designations='GA_x6K02T2PVTD-33539896-33540819|olfactory receptor MOR171-13', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39839436, end_position_on_the_genomic_accession=39840359, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258608 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258608, CurrentID=0, Status='live', Symbol='Olfr986', Aliases='MOR104-3', description='olfactory receptor 986', other_designations='GA_x6K02T2PVTD-33886895-33887833|olfactory receptor MOR104-3', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40187079, end_position_on_the_genomic_accession=40188094, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258611 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258611, CurrentID=0, Status='live', Symbol='Olfr297', Aliases='MOR220-3', description='olfactory receptor 297', other_designations='GA_x6K02T2NHDJ-9587747-9586815|olfactory receptor MOR220-3', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=86526759, end_position_on_the_genomic_accession=86527691, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258612 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258612, CurrentID=0, Status='live', Symbol='Olfr303', Aliases='MOR104-2', description='olfactory receptor 303', other_designations='GA_x6K02T2NHDJ-9712819-9713778|olfactory receptor MOR104-2', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=86393766, end_position_on_the_genomic_accession=86395529, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258614 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258614, CurrentID=0, Status='live', Symbol='Olfr308', Aliases='MOR104-1', description='olfactory receptor 308', other_designations='GA_x6K02T2NHDJ-9786435-9787361|olfactory receptor MOR104-1', map_location='7 E1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=86321024, end_position_on_the_genomic_accession=86321950, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258615 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258615, CurrentID=0, Status='live', Symbol='Olfr360', Aliases='MOR159-1', description='olfactory receptor 360', other_designations='GA_x6K02T2NLDC-33760081-33761034|olfactory receptor MOR159-1', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37068307, end_position_on_the_genomic_accession=37069260, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258616 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258616, CurrentID=0, Status='live', Symbol='Olfr357', Aliases='MOR138-3', description='olfactory receptor 357', other_designations='GA_x6K02T2NLDC-33688556-33689482|olfactory receptor MOR138-3', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36996812, end_position_on_the_genomic_accession=36997738, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258617 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258617, CurrentID=0, Status='live', Symbol='Olfr356', Aliases='MOR134-1', description='olfactory receptor 356', other_designations='GA_x6K02T2NLDC-33631647-33632594|olfactory receptor MOR134-1', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36937121, end_position_on_the_genomic_accession=36941993, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258618 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258618, CurrentID=0, Status='live', Symbol='Olfr355', Aliases='MOR138-2', description='olfactory receptor 355', other_designations='GA_x6K02T2NLDC-33622642-33621710|olfactory receptor MOR138-2', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36927180, end_position_on_the_genomic_accession=36928112, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258620 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258620, CurrentID=0, Status='live', Symbol='Olfr350', Aliases='MOR136-13', description='olfactory receptor 350', other_designations='GA_x6K02T2NLDC-33544602-33545540|olfactory receptor MOR136-13', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36850048, end_position_on_the_genomic_accession=36850986, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258622, CurrentID=0, Status='live', Symbol='Olfr121', Aliases='MOR263-4', description='olfactory receptor 121', other_designations='GA_x6K02T2PSCP-2211113-2212078|olfactory receptor MOR263-4', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37748866, end_position_on_the_genomic_accession=37753004, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258623 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258623, CurrentID=0, Status='live', Symbol='Olfr123', Aliases='MOR256-9', description='olfactory receptor 123', other_designations='GA_x6K02T2PSCP-2255106-2256035|olfactory receptor MOR256-9', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37795446, end_position_on_the_genomic_accession=37796375, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258624 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258624, CurrentID=0, Status='live', Symbol='Olfr120', Aliases='MOR263-3', description='olfactory receptor 120', other_designations='GA_x6K02T2PSCP-2184981-2185946|olfactory receptor MOR263-3', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37725999, end_position_on_the_genomic_accession=37726991, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258627 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258627, CurrentID=0, Status='live', Symbol='Olfr1504', Aliases='MOR212-2', description='olfactory receptor 1504', other_designations='GA_x6K02T2RE5P-4223635-4222688|olfactory receptor MOR212-2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13887261, end_position_on_the_genomic_accession=13897928, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258628 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258628, CurrentID=0, Status='live', Symbol='Olfr1489', Aliases='MOR202-19', description='olfactory receptor 1489', other_designations='GA_x6K02T2RE5P-3965668-3966609|olfactory receptor MOR202-19', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13633113, end_position_on_the_genomic_accession=13634054, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258634 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258634, CurrentID=0, Status='live', Symbol='Olfr1164', Aliases='MOR174-11', description='olfactory receptor 1164', other_designations='GA_x6K02T2Q125-49585842-49584862|olfactory receptor MOR174-11', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88092855, end_position_on_the_genomic_accession=88093966, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258635 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258635, CurrentID=0, Status='live', Symbol='Olfr1140', Aliases='MOR177-6', description='olfactory receptor 1140', other_designations='GA_x6K02T2Q125-49250025-49250960|olfactory receptor MOR177-6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87746161, end_position_on_the_genomic_accession=87747180, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258638 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258638, CurrentID=0, Status='live', Symbol='Olfr1163', Aliases='MOR174-8', description='olfactory receptor 1163', other_designations='GA_x6K02T2Q125-49563265-49562315|olfactory receptor MOR174-8', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88070365, end_position_on_the_genomic_accession=88071407, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258640 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258640, CurrentID=0, Status='live', Symbol='Olfr152', Aliases='MOR181-1, Olfr4-1, V1', description='olfactory receptor 152', other_designations='GA_x6K02T2Q125-49283184-49284128|olfactory receptor 4 cluster, gene 1', map_location='2 49.95 cM', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87782536, end_position_on_the_genomic_accession=87783486, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258644 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258644, CurrentID=0, Status='live', Symbol='Olfr1166', Aliases='MOR174-6', description='olfactory receptor 1166', other_designations='GA_x6K02T2Q125-49616865-49615915|olfactory receptor MOR174-6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88123360, end_position_on_the_genomic_accession=88124993, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258645 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258645, CurrentID=0, Status='live', Symbol='Olfr418', Aliases='MOR267-12P, MOR267-8, Olfr1403-ps1, Olfr418', description='olfactory receptor 418', other_designations='GA_x6K02T2P20D-20826777-20827719|GA_x6K02T2R7CC-581296-580364|olfactory receptor 418, pseudogene 1|olfactory receptor MOR267-12P|olfactory receptor MOR267-8', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173270147, end_position_on_the_genomic_accession=173271137, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258648 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258648, CurrentID=0, Status='live', Symbol='Olfr237-ps1', Aliases='MOR261-4, Olfr438', description='olfactory receptor 237, pseudogene 1', other_designations='GA_x6K02T08UK8-1-481|GA_x6K02T2P3E9-4404793-4403861|olfactory receptor MOR261-4', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43153307, end_position_on_the_genomic_accession=43154239, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258649 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258649, CurrentID=0, Status='live', Symbol='Olfr441', Aliases='MOR261-3', description='olfactory receptor 441', other_designations='GA_x6K02T2P3E9-4442577-4441645|olfactory receptor MOR261-3', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=43115744, end_position_on_the_genomic_accession=43116676, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258650 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258650, CurrentID=0, Status='live', Symbol='Olfr444', Aliases='MOR261-2', description='olfactory receptor 444', other_designations='GA_x6K02T2P3E9-4602571-4601639|olfactory receptor MOR261-2', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42955500, end_position_on_the_genomic_accession=42956432, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258653 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258653, CurrentID=0, Status='live', Symbol='Olfr1136', Aliases='MOR177-3', description='olfactory receptor 1136', other_designations='GA_x6K02T2Q125-49193051-49192119|olfactory receptor MOR177-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87692881, end_position_on_the_genomic_accession=87693947, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258655 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258655, CurrentID=0, Status='live', Symbol='Olfr1112', Aliases='MOR264-6', description='olfactory receptor 1112', other_designations='GA_x6K02T2Q125-48676316-48677272|olfactory receptor MOR264-6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87191689, end_position_on_the_genomic_accession=87192645, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258656 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258656, CurrentID=0, Status='live', Symbol='Olfr365', Aliases='MOR138-1', description='olfactory receptor 365', other_designations='GA_x6K02T2NLDC-33885305-33886243|olfactory receptor MOR138-1', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37201243, end_position_on_the_genomic_accession=37202181, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258657 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258657, CurrentID=0, Status='live', Symbol='Olfr733', Aliases='MOR241-2', description='olfactory receptor 733', other_designations='GA_x6K02T2PMLR-5992342-5991416|olfactory receptor MOR241-2', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50298337, end_position_on_the_genomic_accession=50299330, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258658 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258658, CurrentID=0, Status='live', Symbol='Olfr734', Aliases='MOR242-1', description='olfactory receptor 734', other_designations='GA_x6K02T2PMLR-6013665-6012724|olfactory receptor MOR242-1', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50304110, end_position_on_the_genomic_accession=50320862, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258659 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258659, CurrentID=0, Status='live', Symbol='Olfr732', Aliases='MOR241-1', description='olfactory receptor 732', other_designations='GA_x6K02T2PMLR-5975274-5974348|olfactory receptor MOR241-1', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50281228, end_position_on_the_genomic_accession=50284165, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258660 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258660, CurrentID=0, Status='live', Symbol='Olfr736', Aliases='MOR106-5', description='olfactory receptor 736', other_designations='GA_x6K02T2PMLR-6089963-6090901|olfactory receptor MOR106-5', map_location='14 C3-D1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=50392758, end_position_on_the_genomic_accession=50393696, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258666 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258666, CurrentID=0, Status='live', Symbol='Olfr822', Aliases='MOR113-2', description='olfactory receptor 822', other_designations='GA_x6K02T2PULF-11745102-11746040|olfactory receptor MOR113-2', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130074412, end_position_on_the_genomic_accession=130075350, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258667, CurrentID=0, Status='live', Symbol='Olfr816', Aliases='MOR113-1', description='olfactory receptor 816', other_designations='GA_x6K02T2PULF-11590830-11589892|olfactory receptor MOR113-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129911338, end_position_on_the_genomic_accession=129912276, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258668 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258668, CurrentID=0, Status='live', Symbol='Olfr823', Aliases='MOR210-4', description='olfactory receptor 823', other_designations='GA_x6K02T2PULF-11783479-11782532|olfactory receptor MOR210-4', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130111841, end_position_on_the_genomic_accession=130112788, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258669 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258669, CurrentID=0, Status='live', Symbol='Olfr824', Aliases='MOR210-3', description='olfactory receptor 824', other_designations='GA_x6K02T2PULF-11797746-11796799|olfactory receptor MOR210-3', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130126108, end_position_on_the_genomic_accession=130127055, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258670 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258670, CurrentID=0, Status='live', Symbol='Olfr820', Aliases='MOR116-1', description='olfactory receptor 820', other_designations='GA_x6K02T2PULF-11688441-11689379|olfactory receptor MOR116-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130017363, end_position_on_the_genomic_accession=130018301, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258672 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258672, CurrentID=0, Status='live', Symbol='Olfr825', Aliases='MOR210-1', description='olfactory receptor 825', other_designations='GA_x6K02T2PULF-11834065-11833103|olfactory receptor MOR210-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=130162362, end_position_on_the_genomic_accession=130163324, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258673 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258673, CurrentID=0, Status='live', Symbol='Olfr1428', Aliases='MOR239-5', description='olfactory receptor 1428', other_designations='GA_x6K02T2RE5P-2468394-2467450|olfactory receptor MOR239-5', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12108600, end_position_on_the_genomic_accession=12122095, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  258674 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258674, CurrentID=0, Status='live', Symbol='Olfr1427', Aliases='MOR239-4', description='olfactory receptor 1427', other_designations='GA_x6K02T2RE5P-2458473-2457538|olfactory receptor MOR239-4', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12098702, end_position_on_the_genomic_accession=12099637, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258677, CurrentID=0, Status='live', Symbol='Olfr76', Aliases='MOR215-3, V5', description='olfactory receptor 76', other_designations='GA_x6K02T2RE5P-2479601-2478648', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12119721, end_position_on_the_genomic_accession=12120674, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258678 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258678, CurrentID=0, Status='live', Symbol='Olfr1441', Aliases='MOR215-2', description='olfactory receptor 1441', other_designations='GA_x6K02T2RE5P-2753221-2754177|olfactory receptor MOR215-2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12422311, end_position_on_the_genomic_accession=12423267, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258679 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258679, CurrentID=0, Status='live', Symbol='Olfr1440', Aliases='MOR215-1', description='olfactory receptor 1440', other_designations='GA_x6K02T2RE5P-2725206-2726153|olfactory receptor 202-4|olfactory receptor MOR215-1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12387304, end_position_on_the_genomic_accession=12395212, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258680 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258680, CurrentID=0, Status='live', Symbol='Olfr1434', Aliases='MOR214-4, Olfr1433, Olfr1433-ps1', description='olfactory receptor 1434', other_designations='GA_x6K02T2RE5P-2610001-2610414|GA_x6K02T2RE5P-2618516-2619454', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12279854, end_position_on_the_genomic_accession=12283988, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258682 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258682, CurrentID=0, Status='live', Symbol='Olfr1436', Aliases='MOR214-2', description='olfactory receptor 1436', other_designations='GA_x6K02T2RE5P-2634596-2633658|olfactory receptor MOR214-2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12298183, end_position_on_the_genomic_accession=12299130, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258684 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258684, CurrentID=0, Status='live', Symbol='Olfr1459', Aliases='MOR202-17', description='olfactory receptor 1459', other_designations='GA_x6K02T2RE5P-3473421-3472498|olfactory receptor MOR202-17', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13145734, end_position_on_the_genomic_accession=13146657, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258688 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258688, CurrentID=0, Status='live', Symbol='Olfr1462', Aliases='MOR202-13', description='olfactory receptor 1462', other_designations='GA_x6K02T2RE5P-3517488-3518411|olfactory receptor MOR202-13', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13185734, end_position_on_the_genomic_accession=13191592, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258692 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258692, CurrentID=0, Status='live', Symbol='Olfr1442', Aliases='MOR202-9', description='olfactory receptor 1442', other_designations='GA_x6K02T2RE5P-3000589-3001527|olfactory receptor MOR202-9', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12671020, end_position_on_the_genomic_accession=12675243, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258693 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258693, CurrentID=0, Status='live', Symbol='Olfr1443', Aliases='MOR202-8', description='olfactory receptor 1443', other_designations='GA_x6K02T2RE5P-3006492-3007430|olfactory receptor MOR202-8', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12678166, end_position_on_the_genomic_accession=12683851, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258694 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258694, CurrentID=0, Status='live', Symbol='Olfr1445', Aliases='MOR202-7', description='olfactory receptor 1445', other_designations='GA_x6K02T2RE5P-3213352-3214296|olfactory receptor MOR202-7', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12883883, end_position_on_the_genomic_accession=12884827, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258697 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258697, CurrentID=0, Status='live', Symbol='Olfr1444', Aliases='MOR202-4', description='olfactory receptor 1444', other_designations='GA_x6K02T2RE5P-3191201-3192160|olfactory receptor 202-4|olfactory receptor MOR202-4', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12857421, end_position_on_the_genomic_accession=12862765, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258698 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258698, CurrentID=0, Status='live', Symbol='Olfr1447', Aliases='MOR202-3', description='olfactory receptor 1447', other_designations='GA_x6K02T2RE5P-3231251-3230331|olfactory receptor MOR202-3', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12900808, end_position_on_the_genomic_accession=12903730, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258699 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258699, CurrentID=0, Status='live', Symbol='Olfr1446', Aliases='MOR202-2', description='olfactory receptor 1446', other_designations='GA_x6K02T2RE5P-3220047-3219130|olfactory receptor MOR202-2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12889649, end_position_on_the_genomic_accession=12890575, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258701 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258701, CurrentID=0, Status='live', Symbol='Olfr401', Aliases='MOR255-6', description='olfactory receptor 401', other_designations='GA_x6K02T2P1NL-4278037-4278984|olfactory receptor MOR255-6', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74121291, end_position_on_the_genomic_accession=74122238, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258702 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258702, CurrentID=0, Status='live', Symbol='Olfr410', Aliases='MOR255-5', description='olfactory receptor 410', other_designations='GA_x6K02T2P1NL-4467421-4466474|olfactory receptor MOR255-5', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74334282, end_position_on_the_genomic_accession=74335229, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258703 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258703, CurrentID=0, Status='live', Symbol='Olfr402', Aliases='MOR255-4', description='olfactory receptor 402', other_designations='GA_x6K02T2P1NL-4307199-4308146|olfactory receptor MOR255-4', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74155156, end_position_on_the_genomic_accession=74156103, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258704 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258704, CurrentID=0, Status='live', Symbol='Olfr411', Aliases='MOR255-3', description='olfactory receptor 411', other_designations='GA_x6K02T2P1NL-4481525-4480578|olfactory receptor MOR255-3|olfactory receptor MOR255-6', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74346624, end_position_on_the_genomic_accession=74347672, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258705 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258705, CurrentID=0, Status='live', Symbol='Olfr398', Aliases='MOR157-1', description='olfactory receptor 398', other_designations='GA_x6K02T2P1NL-4141430-4140486|olfactory receptor MOR157-1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73983662, end_position_on_the_genomic_accession=73984606, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258706 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258706, CurrentID=0, Status='live', Symbol='Olfr43', Aliases='IA7, MOR125-1, MOR125-5_p, Olfr403', description='olfactory receptor 43', other_designations='GA_x6K02T2P1NL-4359899-4358958|olfactory receptor 403|olfactory receptor MOR125-1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74206198, end_position_on_the_genomic_accession=74207289, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258708 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258708, CurrentID=0, Status='live', Symbol='Olfr1342', Aliases='MOR258-3', description='olfactory receptor 1342', other_designations='GA_x6K02T2QD9B-18856980-18857927|olfactory receptor MOR258-3', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118688980, end_position_on_the_genomic_accession=118693034, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258710 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258710, CurrentID=0, Status='live', Symbol='Olfr419', Aliases='MOR267-6', description='olfactory receptor 419', other_designations='GA_x6K02T2P20D-20891507-20892448|olfactory receptor MOR267-6', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174249889, end_position_on_the_genomic_accession=174250976, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258711 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258711, CurrentID=0, Status='live', Symbol='Olfr432', Aliases='MOR123-2', description='olfactory receptor 432', other_designations='GA_x6K02T2P20D-21124681-21123743|olfactory receptor MOR123-2', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174050375, end_position_on_the_genomic_accession=174051313, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258712 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258712, CurrentID=0, Status='live', Symbol='Olfr433', Aliases='MOR123-1', description='olfactory receptor 433', other_designations='GA_x6K02T2P20D-21133014-21132070|olfactory receptor MOR123-1', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174041933, end_position_on_the_genomic_accession=174042936, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258713 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258713, CurrentID=0, Status='live', Symbol='Olfr430', Aliases='MOR105-5P', description='olfactory receptor 430', other_designations='GA_x6K02T2P20D-21108443-21107490|olfactory receptor MOR105-5P', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174069300, end_position_on_the_genomic_accession=174070253, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258716 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258716, CurrentID=0, Status='live', Symbol='Olfr424', Aliases='MOR105-2', description='olfactory receptor 424', other_designations='GA_x6K02T2P20D-21025190-21024243|olfactory receptor MOR105-2', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174136746, end_position_on_the_genomic_accession=174137693, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258717 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258717, CurrentID=0, Status='live', Symbol='Olfr429', Aliases='MOR105-1', description='olfactory receptor 429', other_designations='GA_x6K02T2P20D-21090094-21089156|olfactory receptor MOR105-1', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174089038, end_position_on_the_genomic_accession=174089980, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258718 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258718, CurrentID=0, Status='live', Symbol='Olfr513', Aliases='MOR195-1', description='olfactory receptor 513', other_designations='GA_x6K02T2PBJ9-11084889-11085818|olfactory receptor MOR195-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108754858, end_position_on_the_genomic_accession=108755787, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258719 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258719, CurrentID=0, Status='live', Symbol='Olfr512', Aliases='MOR268-3', description='olfactory receptor 512', other_designations='GA_x6K02T2PBJ9-11043421-11044365|olfactory receptor MOR268-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108713355, end_position_on_the_genomic_accession=108714335, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258720 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258720, CurrentID=0, Status='live', Symbol='Olfr516', Aliases='MOR268-2', description='olfactory receptor 516', other_designations='GA_x6K02T2PBJ9-11176324-11175380|olfactory receptor MOR268-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108845064, end_position_on_the_genomic_accession=108846008, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258722 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258722, CurrentID=0, Status='live', Symbol='Olfr611', Aliases='MOR15-1, MOR15-4', description='olfactory receptor 611', other_designations='GA_x6K02T2PBJ9-6236637-6235666|olfactory receptor MOR15-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103517411, end_position_on_the_genomic_accession=103518382, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258725 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258725, CurrentID=0, Status='live', Symbol='Olfr1095', Aliases='MOR179-1', description='olfactory receptor 1095', other_designations='GA_x6K02T2Q125-48336843-48335917|olfactory receptor MOR179-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86850770, end_position_on_the_genomic_accession=86854320, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258726 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258726, CurrentID=0, Status='live', Symbol='Olfr599', Aliases='MOR23-1', description='olfactory receptor 599', other_designations='GA_x6K02T2PBJ9-6047402-6048349|olfactory receptor MOR23-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103338056, end_position_on_the_genomic_accession=103339003, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258728 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258728, CurrentID=0, Status='live', Symbol='Olfr482', Aliases='MOR204-14', description='olfactory receptor 482', other_designations='GA_x6K02T2PBJ9-10424354-10423383|olfactory receptor 204-14|olfactory receptor MOR204-14', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108094597, end_position_on_the_genomic_accession=108095568, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258730 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258730, CurrentID=0, Status='live', Symbol='Olfr483', Aliases='MOR204-12', description='olfactory receptor 483', other_designations='GA_x6K02T2PBJ9-10432095-10433042|olfactory receptor 204-12|olfactory receptor MOR204-12', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108103311, end_position_on_the_genomic_accession=108104258, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258732 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258732, CurrentID=0, Status='live', Symbol='Olfr494', Aliases='MOR204-10', description='olfactory receptor 494', other_designations='GA_x6K02T2PBJ9-10697517-10698461|olfactory receptor 204-10|olfactory receptor MOR204-10', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108367492, end_position_on_the_genomic_accession=108368436, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258734 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258734, CurrentID=0, Status='live', Symbol='Olfr502', Aliases='MOR204-8', description='olfactory receptor 502', other_designations='GA_x6K02T2PBJ9-10853935-10852991|olfactory receptor 204-8|olfactory receptor MOR204-8', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108523004, end_position_on_the_genomic_accession=108523948, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258736 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258736, CurrentID=0, Status='live', Symbol='Olfr1497', Aliases='MOR212-1', description='olfactory receptor 1497', other_designations='GA_x6K02T2RE5P-4127765-4126821|olfactory receptor MOR212-1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13794665, end_position_on_the_genomic_accession=13795609, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258737 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258737, CurrentID=0, Status='live', Symbol='Olfr1316', Aliases='MOR245-1, MOR245-26', description='olfactory receptor 1316', other_designations='GA_x6K02T2Q125-73177039-73176101|olfactory receptor MOR245-1', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=112129871, end_position_on_the_genomic_accession=112130815, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258739 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258739, CurrentID=0, Status='live', Symbol='Olfr1362', Aliases='MOR256-8, Olfr1363-ps1', description='olfactory receptor 1362', other_designations='GA_x6K02T2QHY8-11617756-11618682|GA_x6K02T2QHY8-11628399-11628868|olfactory receptor 1363, pseudogene 1|olfactory receptor MOR256-8', map_location='13 A3.1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=21611020, end_position_on_the_genomic_accession=21612004, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258743 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258743, CurrentID=0, Status='live', Symbol='Olfr661', Aliases='MOR40-4', description='olfactory receptor 661', other_designations='G-protein-coupled receptor|GA_x6K02T2PBJ9-7316375-7317334|odorant receptor S83|olfactory receptor MOR40-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104688017, end_position_on_the_genomic_accession=104688976, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258744 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258744, CurrentID=0, Status='live', Symbol='Olfr875', Aliases='MOR161-4', description='olfactory receptor 875', other_designations='GA_x6K02T2PVTD-31458511-31459443|olfactory receptor MOR161-4', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37772637, end_position_on_the_genomic_accession=37773640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258745 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258745, CurrentID=0, Status='live', Symbol='Olfr689', Aliases='MOR40-3', description='olfactory receptor 689', other_designations='GA_x6K02T2PBJ9-7942985-7943947|olfactory receptor MOR40-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105314006, end_position_on_the_genomic_accession=105314968, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258746 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258746, CurrentID=0, Status='live', Symbol='Olfr648', Aliases='MOR31-12', description='olfactory receptor 648', other_designations='GA_x6K02T2PBJ9-6914780-6913830|olfactory receptor MOR31-12', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104179456, end_position_on_the_genomic_accession=104180406, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258747 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258747, CurrentID=0, Status='live', Symbol='Olfr1106', Aliases='MOR172-6, MOR172-8_p', description='olfactory receptor 1106', other_designations='GA_x6K02T2Q125-48541463-48540525|olfactory receptor MOR172-6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87048296, end_position_on_the_genomic_accession=87049234, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258748 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258748, CurrentID=0, Status='live', Symbol='Olfr1195', Aliases='MOR230-4', description='olfactory receptor 1195', other_designations='GA_x6K02T2Q125-50163514-50162588|MOR230-12_p|olfactory receptor MOR230-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88682804, end_position_on_the_genomic_accession=88683730, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258749 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258749, CurrentID=0, Status='live', Symbol='Olfr556', Aliases='MOR41-1', description='olfactory receptor 556', other_designations='GA_x6K02T2PBJ9-5386601-5387575|olfactory receptor MOR41-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102669901, end_position_on_the_genomic_accession=102670923, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258750 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258750, CurrentID=0, Status='live', Symbol='Olfr551', Aliases='MOR29-1', description='olfactory receptor 551', other_designations='GA_x6K02T2PBJ9-5307445-5306498|olfactory receptor MOR29-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102587741, end_position_on_the_genomic_accession=102588819, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258752 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258752, CurrentID=0, Status='live', Symbol='Olfr583', Aliases='MOR14-6', description='olfactory receptor 583', other_designations='GA_x6K02T2PBJ9-5762668-5763618|olfactory receptor MOR14-6', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103051300, end_position_on_the_genomic_accession=103052259, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258753 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258753, CurrentID=0, Status='live', Symbol='Olfr678', Aliases='MOR32-5', description='olfactory receptor 678', other_designations='GA_x6K02T2PBJ9-7698491-7699432|olfactory receptor MOR32-5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105069469, end_position_on_the_genomic_accession=105070410, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258754 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258754, CurrentID=0, Status='live', Symbol='Olfr214', Aliases='MOR119-1', description='olfactory receptor 214', other_designations='GA_x54KRFPKN04-58190962-58191900|olfactory receptor MOR119-1', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116556427, end_position_on_the_genomic_accession=116557389, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258755 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258755, CurrentID=0, Status='live', Symbol='Olfr672', Aliases='MOR32-4, Ors46', description='olfactory receptor 672', other_designations='GA_x6K02T2PBJ9-7625746-7624808|odorant receptor S46|olfactory receptor MOR32-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104995964, end_position_on_the_genomic_accession=104996902, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258756 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258756, CurrentID=0, Status='live', Symbol='Olfr414', Aliases='MOR103-10', description='olfactory receptor 414', other_designations='GA_x6K02T2P20D-20749615-20748662|olfactory receptor MOR103-10', map_location='1 H4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174430430, end_position_on_the_genomic_accession=174431384, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258757 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258757, CurrentID=0, Status='live', Symbol='Olfr1013', Aliases='MOR213-2', description='olfactory receptor 1013', other_designations='GA_x6K02T2Q125-47248551-47249468|olfactory receptor 213-2|olfactory receptor MOR213-2|olfr1013', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85769803, end_position_on_the_genomic_accession=85770720, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258758 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258758, CurrentID=0, Status='live', Symbol='Olfr1406', Aliases='MOR267-5', description='olfactory receptor 1406', other_designations='GA_x6K02T2R7CC-664297-665229|olfactory receptor MOR267-5', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173183442, end_position_on_the_genomic_accession=173184504, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258759, CurrentID=0, Status='live', Symbol='Olfr1408', Aliases='MOR267-4', description='olfactory receptor 1408', other_designations='GA_x6K02T2R7CC-715776-716708|olfactory receptor MOR267-4', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173130283, end_position_on_the_genomic_accession=173131215, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258763 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258763, CurrentID=0, Status='live', Symbol='Olfr1104', Aliases='MOR207-1', description='olfactory receptor 1104', other_designations='GA_x6K02T2Q125-48508763-48507833|olfactory receptor MOR207-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87021610, end_position_on_the_genomic_accession=87022542, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258764 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258764, CurrentID=0, Status='live', Symbol='Olfr1099', Aliases='MOR206-3', description='olfactory receptor 1099', other_designations='GA_x6K02T2Q125-48446067-48445129|olfactory receptor MOR206-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86958518, end_position_on_the_genomic_accession=86959456, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258769 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258769, CurrentID=0, Status='live', Symbol='Olfr508', Aliases='MOR204-6', description='olfactory receptor 508', other_designations='GA_x6K02T2PBJ9-10959726-10960658|olfactory receptor 204-6|olfactory receptor MOR204-6', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108629994, end_position_on_the_genomic_accession=108630926, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258771 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258771, CurrentID=0, Status='live', Symbol='Olfr473', Aliases='MOR204-4', description='olfactory receptor 473', other_designations='GA_x6K02T2PBJ9-10262759-10263691|olfactory receptor 204-4|olfactory receptor MOR204-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107933522, end_position_on_the_genomic_accession=107934454, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258774 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258774, CurrentID=0, Status='live', Symbol='Olfr1208', Aliases='MOR225-4', description='olfactory receptor 1208', other_designations='GA_x6K02T2Q125-50372411-50371485|olfactory receptor MOR225-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88896669, end_position_on_the_genomic_accession=88897595, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258776 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258776, CurrentID=0, Status='live', Symbol='Olfr715', Aliases='MOR260-1', description='olfactory receptor 715', other_designations='B2 olfactory receptor|GA_x6K02T2PBJ9-9479517-9478573|olfactory receptor MOR260-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107128352, end_position_on_the_genomic_accession=107129482, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258777 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258777, CurrentID=0, Status='live', Symbol='Olfr922', Aliases='MOR161-3', description='olfactory receptor 922', other_designations='GA_x6K02T2PVTD-32518237-32519172|olfactory receptor MOR161-3', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38815408, end_position_on_the_genomic_accession=38816539, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258778 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258778, CurrentID=0, Status='live', Symbol='Olfr921', Aliases='MOR165-8', description='olfactory receptor 921', other_designations='GA_x6K02T2PVTD-32478047-32478988|olfactory receptor MOR165-8', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38773088, end_position_on_the_genomic_accession=38776354, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258780 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258780, CurrentID=0, Status='live', Symbol='Olfr916', Aliases='MOR168-1', description='olfactory receptor 916', other_designations='GA_x6K02T2PVTD-32360710-32359778|olfactory receptor MOR168-1', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38657458, end_position_on_the_genomic_accession=38659875, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258781 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258781, CurrentID=0, Status='live', Symbol='Olfr915', Aliases='MOR167-4', description='olfactory receptor 915', other_designations='GA_x6K02T2PVTD-32349884-32348952|olfactory receptor MOR167-4', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38646590, end_position_on_the_genomic_accession=38647522, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258784 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258784, CurrentID=0, Status='live', Symbol='Olfr1245', Aliases='MOR23112', description='olfactory receptor 1245', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89574712, end_position_on_the_genomic_accession=89575773, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258785 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258785, CurrentID=0, Status='live', Symbol='Olfr1230', Aliases='MOR237-1', description='olfactory receptor 1230', other_designations='GA_x6K02T2Q125-50741934-50741017|olfactory receptor MOR237-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89296351, end_position_on_the_genomic_accession=89297268, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258786 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258786, CurrentID=0, Status='live', Symbol='Olfr1238', Aliases='MOR231-11', description='olfactory receptor 1238', other_designations='GA_x6K02T2Q125-50849945-50848998|olfactory receptor MOR231-11', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89406130, end_position_on_the_genomic_accession=89407077, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258787 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258787, CurrentID=0, Status='live', Symbol='Olfr1248', Aliases='MOR23110', description='olfactory receptor 1248', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89617168, end_position_on_the_genomic_accession=89618245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258788 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258788, CurrentID=0, Status='live', Symbol='Olfr1246', Aliases='MOR2319', description='olfactory receptor 1246', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89590069, end_position_on_the_genomic_accession=89591146, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258790 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258790, CurrentID=0, Status='live', Symbol='Olfr1263', Aliases='MOR234-2', description='olfactory receptor 1263', other_designations='GA_x6K02T2Q125-51447049-51447969|olfactory receptor MOR234-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90014932, end_position_on_the_genomic_accession=90015852, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258791 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258791, CurrentID=0, Status='live', Symbol='Olfr812', Aliases='MOR110-1', description='olfactory receptor 812', other_designations='GA_x6K02T2PULF-11521598-11520666|olfactory receptor MOR110-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129842026, end_position_on_the_genomic_accession=129845997, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258792 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258792, CurrentID=0, Status='live', Symbol='Olfr1499', Aliases='MOR211-2', description='olfactory receptor 1499', other_designations='GA_x6K02T2RE5P-4147744-4146800|olfactory receptor MOR211-2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13814644, end_position_on_the_genomic_accession=13815588, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258793 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258793, CurrentID=0, Status='live', Symbol='Olfr1502', Aliases='MOR211-1', description='olfactory receptor 1502', other_designations='GA_x6K02T2RE5P-4193992-4194942|olfactory receptor MOR211-1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13861795, end_position_on_the_genomic_accession=13862745, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258794 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258794, CurrentID=0, Status='live', Symbol='Olfr878', Aliases='MOR163-1', description='olfactory receptor 878', other_designations='GA_x6K02T2PVTD-31600511-31601440|olfactory receptor MOR163-1', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37918548, end_position_on_the_genomic_accession=37919612, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258797 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258797, CurrentID=0, Status='live', Symbol='Olfr904', Aliases='MOR167-3', description='olfactory receptor 904', other_designations='GA_x6K02T2PVTD-32156773-32157705|olfactory receptor MOR167-3', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38464043, end_position_on_the_genomic_accession=38464975, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258799 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258799, CurrentID=0, Status='live', Symbol='Olfr906', Aliases='MOR167-2', description='olfactory receptor 906', other_designations='GA_x6K02T2PVTD-32194085-32195020|olfactory receptor MOR167-2', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38488031, end_position_on_the_genomic_accession=38488966, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258800 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258800, CurrentID=0, Status='live', Symbol='Olfr905', Aliases='MOR167-1', description='olfactory receptor 905', other_designations='GA_x6K02T2PVTD-32165709-32166641|olfactory receptor MOR167-1', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38472749, end_position_on_the_genomic_accession=38473681, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258802 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258802, CurrentID=0, Status='live', Symbol='Olfr143', Aliases='K18, MOR170-6', description='olfactory receptor 143', other_designations='GA_x6K02T2PVTD-31941084-31942025|odorant receptor K18|olfactory receptor 170-6|olfactory receptor 7A|olfactory receptor MOR170-6', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38253419, end_position_on_the_genomic_accession=38254360, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258803 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258803, CurrentID=0, Status='live', Symbol='Olfr136', Aliases='MOR256-7', description='olfactory receptor 136', other_designations='GA_x6K02T2PSCP-2779375-2780313|olfactory receptor MOR256-7', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38334385, end_position_on_the_genomic_accession=38336097, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258804 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258804, CurrentID=0, Status='live', Symbol='Olfr1240', Aliases='MOR231-8', description='olfactory receptor 1240', other_designations='GA_x6K02T2Q125-50883183-50882239|olfactory receptor MOR231-8', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89439270, end_position_on_the_genomic_accession=89440343, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258805 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258805, CurrentID=0, Status='live', Symbol='Olfr1426', Aliases='MOR239-1', description='olfactory receptor 1426', other_designations='GA_x6K02T2RE5P-2447610-2446675|olfactory receptor MOR239-1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=12087610, end_position_on_the_genomic_accession=12091847, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258808 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258808, CurrentID=0, Status='live', Symbol='Olfr620', Aliases='MOR4-1', description='olfactory receptor 620', other_designations='GA_x6K02T2PBJ9-6335095-6334154|olfactory receptor MOR4-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103611410, end_position_on_the_genomic_accession=103612351, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258809 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258809, CurrentID=0, Status='live', Symbol='Olfr651', Aliases='MOR31-11', description='olfactory receptor 651', other_designations='GA_x6K02T2PBJ9-7179540-7180481|olfactory receptor MOR31-11', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104552884, end_position_on_the_genomic_accession=104553952, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258810 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258810, CurrentID=0, Status='live', Symbol='Olfr665', Aliases='MOR34-7', description='olfactory receptor 665', other_designations='GA_x6K02T2PBJ9-7509539-7510489|olfactory receptor MOR34-7', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104880709, end_position_on_the_genomic_accession=104881659, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  108671 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108671, CurrentID=0, Status='live', Symbol='Dnajc9', Aliases='5330419I01Rik, AU020082', description='DnaJ heat shock protein family (Hsp40) member C9', other_designations='DnaJ (Hsp40) homolog, subfamily C, member 9', map_location='14 A3', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=20384638, end_position_on_the_genomic_accession=20388910, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  258814 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258814, CurrentID=0, Status='live', Symbol='Olfr1156', Aliases='MOR174-3', description='olfactory receptor 1156', other_designations='GA_x6K02T2Q125-49433499-49432537|olfactory receptor MOR174-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87949190, end_position_on_the_genomic_accession=87950265, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258816 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258816, CurrentID=0, Status='live', Symbol='Olfr466', Aliases='MOR209-1', description='olfactory receptor 466', other_designations='GA_x6K02T2PB7A-3051266-3052192|olfactory receptor MOR209-1', map_location='13 C1', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=65152226, end_position_on_the_genomic_accession=65153152, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258817 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258817, CurrentID=0, Status='live', Symbol='Olfr655', Aliases='MOR38-1', description='olfactory receptor 655', other_designations='GA_x6K02T2PBJ9-7224628-7223702|olfactory receptor MOR38-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104596253, end_position_on_the_genomic_accession=104597179, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258818, CurrentID=0, Status='live', Symbol='Olfr629', Aliases='MOR26-2', description='olfactory receptor 629', other_designations='GA_x6K02T2PBJ9-6466772-6465828|olfactory receptor MOR26-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103740201, end_position_on_the_genomic_accession=103741322, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258819 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258819, CurrentID=0, Status='live', Symbol='Olfr640', Aliases='MOR13-4', description='olfactory receptor 640', other_designations='GA_x6K02T2PBJ9-6756759-6755815|olfactory receptor MOR13-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104021372, end_position_on_the_genomic_accession=104022316, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258820 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258820, CurrentID=0, Status='live', Symbol='Olfr1184', Aliases='MOR225-3', description='olfactory receptor 1184', other_designations='GA_x6K02T2Q125-49974190-49975125|olfactory receptor MOR225-3', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88486734, end_position_on_the_genomic_accession=88487669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258821, CurrentID=0, Status='live', Symbol='Olfr273', Aliases='MOR262-8', description='olfactory receptor 273', other_designations='GA_x6K02T2N78B-7137430-7138383|olfactory receptor MOR262-8', map_location='4 B2-C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=52855558, end_position_on_the_genomic_accession=52856511, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258822 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258822, CurrentID=0, Status='live', Symbol='Olfr39', Aliases='MOR144-1', description='olfactory receptor 39', other_designations='GA_x6K02T2PVTD-14025733-14026668', map_location='9 7.51 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20282351, end_position_on_the_genomic_accession=20286648, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258825 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258825, CurrentID=0, Status='live', Symbol='Olfr975', Aliases='MOR224-2', description='olfactory receptor 975', other_designations='GA_x6K02T2PVTD-33651220-33650288|olfactory receptor MOR224-2', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39949837, end_position_on_the_genomic_accession=39950769, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258826 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258826, CurrentID=0, Status='live', Symbol='Olfr27', Aliases='MOR171-6, MTPCR56', description='olfactory receptor 27', other_designations='GA_x6K02T2PVTD-32841223-32842158', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39128167, end_position_on_the_genomic_accession=39145072, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258828 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258828, CurrentID=0, Status='live', Symbol='Olfr133', Aliases='MOR256-6', description='olfactory receptor 133', other_designations='GA_x6K02T2PSCP-2597192-2598130|olfactory receptor MOR256-6', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38145393, end_position_on_the_genomic_accession=38149528, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258829 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258829, CurrentID=0, Status='live', Symbol='Olfr134', Aliases='MOR256-5', description='olfactory receptor 134', other_designations='GA_x6K02T2PSCP-2623613-2624551|olfactory receptor MOR256-5', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38175086, end_position_on_the_genomic_accession=38176024, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258830 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258830, CurrentID=0, Status='live', Symbol='Olfr103', Aliases='MOR250-3, MOR250-8_p', description='olfactory receptor 103', other_designations='GA_x6K02T2PSCP-1798423-1797482|olfactory receptor MOR250-3', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37336289, end_position_on_the_genomic_accession=37339678, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258831 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258831, CurrentID=0, Status='live', Symbol='Olfr101', Aliases='MOR250-2', description='olfactory receptor 101', other_designations='GA_x6K02T2PSCP-1761617-1760691|olfactory receptor MOR250-2', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37299494, end_position_on_the_genomic_accession=37300420, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258832 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258832, CurrentID=0, Status='live', Symbol='Olfr109', Aliases='MOR250-1', description='olfactory receptor 109', other_designations='GA_x6K02T2PSCP-1914078-1915022|olfactory receptor MOR250-1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37466208, end_position_on_the_genomic_accession=37467152, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258836 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258836, CurrentID=0, Status='live', Symbol='Olfr272', Aliases='MOR262-7', description='olfactory receptor 272', other_designations='GA_x6K02T2N78B-7084885-7085844|olfactory receptor MOR262-7', map_location='4 B2-C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=52910833, end_position_on_the_genomic_accession=52911792, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258837 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258837, CurrentID=0, Status='live', Symbol='Olfr545', Aliases='MOR42-1, S50', description='olfactory receptor 545', other_designations='GA_x6K02T2PBJ9-5216160-5215210|olfactory receptor MOR42-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102493823, end_position_on_the_genomic_accession=102494773, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258840 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258840, CurrentID=0, Status='live', Symbol='Olfr1097', Aliases='MOR206-2', description='olfactory receptor 1097', other_designations='GA_x6K02T2Q125-48376288-48375341|olfactory receptor MOR206-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86890226, end_position_on_the_genomic_accession=86892163, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258841 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258841, CurrentID=0, Status='live', Symbol='Olfr1107', Aliases='MOR172-2', description='olfactory receptor 1107', other_designations='GA_x6K02T2Q125-48565383-48564445|olfactory receptor MOR172-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87071115, end_position_on_the_genomic_accession=87072223, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258842 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258842, CurrentID=0, Status='live', Symbol='Olfr1098', Aliases='MOR206-1', description='olfactory receptor 1098', other_designations='GA_x6K02T2Q125-48410458-48409511|olfactory receptor MOR206-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86922583, end_position_on_the_genomic_accession=86924498, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258845 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258845, CurrentID=0, Status='live', Symbol='Olfr1161', Aliases='MOR174-2', description='olfactory receptor 1161', other_designations='GA_x6K02T2Q125-49516664-49517629|olfactory receptor MOR174-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88024702, end_position_on_the_genomic_accession=88025713, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258851 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258851, CurrentID=0, Status='live', Symbol='Olfr1339', Aliases='MOR258-2', description='olfactory receptor 1339', other_designations='GA_x6K02T2QD9B-18815145-18814198|olfactory receptor MOR258-2', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118734531, end_position_on_the_genomic_accession=118735478, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258852 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258852, CurrentID=0, Status='live', Symbol='Olfr1341', Aliases='MOR258-1', description='olfactory receptor 1341', other_designations='GA_x6K02T2QD9B-18838170-18837232|olfactory receptor MOR258-1', map_location='4 D2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=118709409, end_position_on_the_genomic_accession=118710347, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258853 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258853, CurrentID=0, Status='live', Symbol='Olfr982', Aliases='MOR223-4', description='olfactory receptor 982', other_designations='GA_x6K02T2PVTD-33772307-33773260|olfactory receptor MOR223-4', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40074297, end_position_on_the_genomic_accession=40075262, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258854 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258854, CurrentID=0, Status='live', Symbol='Olfr985', Aliases='MOR171-4', description='olfactory receptor 985', other_designations='GA_x6K02T2PVTD-33825763-33824828|olfactory receptor MOR171-4', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40126924, end_position_on_the_genomic_accession=40128039, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258857 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258857, CurrentID=0, Status='live', Symbol='Olfr275', Aliases='MOR262-2', description='olfactory receptor 275', other_designations='GA_x6K02T2N78B-7168533-7167574|olfactory receptor MOR262-2', map_location='4 B2-C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=52825399, end_position_on_the_genomic_accession=52826358, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258858 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258858, CurrentID=0, Status='live', Symbol='Olfr371', Aliases='MOR141-3', description='olfactory receptor 371', other_designations='GA_x6K02T2NUPS-13298842-13299780|olfactory receptor MOR141-3', map_location='8 C3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=85230497, end_position_on_the_genomic_accession=85231435, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258859 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258859, CurrentID=0, Status='live', Symbol='Olfr161', Aliases='MOR131-1, OR912-47M6', description='olfactory receptor 161', other_designations='GA_x54KRFPKG5P-112942-113883|olfactory receptor MOR131-1', map_location='16 B1', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=3592398, end_position_on_the_genomic_accession=3593339, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258861 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258861, CurrentID=0, Status='live', Symbol='Olfr763', Aliases='MOR269-2', description='olfactory receptor 763', other_designations='GA_x6K02T2PULF-10696986-10697915|olfactory receptor MOR269-2', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129011287, end_position_on_the_genomic_accession=129012216, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258865 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258865, CurrentID=0, Status='live', Symbol='Olfr992', Aliases='MOR203-1', description='olfactory receptor 992', other_designations='GA_x6K02T2Q125-46877170-46876241|olfactory receptor MOR203-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85399602, end_position_on_the_genomic_accession=85400531, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258867 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258867, CurrentID=0, Status='live', Symbol='Olfr131', Aliases='MOR256-4', description='olfactory receptor 131', other_designations='GA_x6K02T2PSCP-2531299-2530355|olfactory receptor MOR256-4', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=38082032, end_position_on_the_genomic_accession=38082976, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258869 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258869, CurrentID=0, Status='live', Symbol='Olfr147', Aliases='M3, MOR164-1', description='olfactory receptor 147', other_designations='GA_x6K02T2PVTD-32098059-32099003|odorant receptor M3|olfactory receptor 164-1|olfactory receptor 7E|olfactory receptor MOR164-1', map_location='9 A5.1', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38401711, end_position_on_the_genomic_accession=38403829, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258871 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258871, CurrentID=0, Status='live', Symbol='Olfr898', Aliases='MOR170-3', description='olfactory receptor 898', other_designations='GA_x6K02T2PVTD-32037624-32038565|olfactory receptor MOR170-3', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38349085, end_position_on_the_genomic_accession=38350035, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258872 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258872, CurrentID=0, Status='live', Symbol='Olfr908', Aliases='MOR165-2', description='olfactory receptor 908', other_designations='GA_x6K02T2PVTD-32216179-32217111|olfactory receptor MOR165-2', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38516034, end_position_on_the_genomic_accession=38516965, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258873 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258873, CurrentID=0, Status='live', Symbol='Olfr911-ps1', Aliases='MOR165-1, MOR166-1, Olfr909', description='olfactory receptor 911, pseudogene 1', other_designations='GA_x6K02T2PVTD-32223906-32224841|GA_x6K02T2PVTD-32247224-32248163|olfactory receptor MOR165-1|olfactory receptor MOR166-1', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38523730, end_position_on_the_genomic_accession=38524669, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258875 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258875, CurrentID=0, Status='live', Symbol='Olfr895', Aliases='MOR170-1', description='olfactory receptor 895', other_designations='GA_x6K02T2PVTD-31962461-31963411|olfactory receptor MOR170-1', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=38268001, end_position_on_the_genomic_accession=38269753, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258877 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258877, CurrentID=0, Status='live', Symbol='Olfr1395', Aliases='MOR277-1', description='olfactory receptor 1395', other_designations='GA_x6K02T2QP88-6286247-6285294|olfactory receptor MOR277-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49148259, end_position_on_the_genomic_accession=49149212, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258879 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258879, CurrentID=0, Status='live', Symbol='Olfr330', Aliases='MOR275-1', description='olfactory receptor 330', other_designations='GA_x6K02T2NKPP-895420-896349|olfactory receptor MOR275-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=58528929, end_position_on_the_genomic_accession=58534825, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258880 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258880, CurrentID=0, Status='live', Symbol='Olfr218', Aliases='MOR267-3, Olfr1405-ps1-ps1, Olfr218', description='olfactory receptor 218', other_designations='GA_x6K02SYWG4P-534-1100|GA_x6K02T2R7CC-643715-642847|olfactory receptor 1405, pseudogene 1|olfactory receptor 218, pseudogene 1|olfactory receptor MOR267-3', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173203358, end_position_on_the_genomic_accession=173204299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258881 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258881, CurrentID=0, Status='live', Symbol='Olfr1404', Aliases='MOR267-2', description='olfactory receptor 1404', other_designations='GA_x6K02T2R7CC-630397-629456|olfactory receptor MOR267-2', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=173215607, end_position_on_the_genomic_accession=173216640, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258882 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258882, CurrentID=0, Status='live', Symbol='Olfr874', Aliases='MOR161-2', description='olfactory receptor 874', other_designations='GA_x6K02T2PVTD-31428850-31429782|odorant receptor K22|olfactory receptor MOR161-2', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37743134, end_position_on_the_genomic_accession=37747106, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258883 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258883, CurrentID=0, Status='live', Symbol='Olfr876', Aliases='MOR161-1', description='olfactory receptor 876', other_designations='GA_x6K02T2PVTD-31489645-31490577|olfactory receptor MOR161-1', map_location='9 A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=37803889, end_position_on_the_genomic_accession=37804936, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258886 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258886, CurrentID=0, Status='live', Symbol='Olfr1299', Aliases='MOR248-8', description='olfactory receptor 1299', other_designations='GA_x6K02T2Q125-72715642-72716580|olfactory receptor MOR248-8', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111661453, end_position_on_the_genomic_accession=111665166, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258887 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258887, CurrentID=0, Status='live', Symbol='Olfr1294', Aliases='MOR248-7', description='olfactory receptor 1294', other_designations='GA_x6K02T2Q125-72589785-72588847|olfactory receptor MOR248-7', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111537349, end_position_on_the_genomic_accession=111540150, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258888 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258888, CurrentID=0, Status='live', Symbol='Olfr1298', Aliases='MOR248-6', description='olfactory receptor 1298', other_designations='GA_x6K02T2Q125-72697413-72696475|olfactory receptor MOR248-6', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111645057, end_position_on_the_genomic_accession=111645995, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258890 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258890, CurrentID=0, Status='live', Symbol='Olfr1297', Aliases='MOR248-4', description='olfactory receptor 1297', other_designations='GA_x6K02T2Q125-72673494-72672556|olfactory receptor MOR248-4', map_location='2 E5', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111621134, end_position_on_the_genomic_accession=111626391, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258892 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258892, CurrentID=0, Status='live', Symbol='Olfr126', Aliases='MOR218-1', description='olfactory receptor 126', other_designations='GA_x6K02T2PSCP-2307164-2308123|olfactory receptor MOR218-1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37850594, end_position_on_the_genomic_accession=37851553, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258893 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258893, CurrentID=0, Status='live', Symbol='Olfr1225', Aliases='MOR233-11', description='olfactory receptor 1225', other_designations='GA_x6K02T2Q125-50650037-50649102|olfactory receptor MOR233-11', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89170219, end_position_on_the_genomic_accession=89171245, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258898 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258898, CurrentID=0, Status='live', Symbol='Olfr1205', Aliases='MOR230-1, MOR230-3, Olfr1203-ps1, Olfr1205', description='olfactory receptor 1205', other_designations='olfactory receptor 1203|olfactory receptor 1205, pseudogene 1|olfactory receptor MOR230-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88829757, end_position_on_the_genomic_accession=88832042, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258902 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258902, CurrentID=0, Status='live', Symbol='Olfr1220', Aliases='MOR2335', description='olfactory receptor 1220', other_designations='', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89096938, end_position_on_the_genomic_accession=89097963, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258903 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258903, CurrentID=0, Status='live', Symbol='Olfr1217', Aliases='MOR233-4', description='olfactory receptor 1217', other_designations='GA_x6K02T2Q125-50504545-50503631|olfactory receptor MOR233-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89022995, end_position_on_the_genomic_accession=89024099, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258905 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258905, CurrentID=0, Status='live', Symbol='Olfr871', Aliases='MOR141-2', description='olfactory receptor 871', other_designations='GA_x6K02T2PVTD-13952555-13953490|olfactory receptor MOR141-2', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=20212207, end_position_on_the_genomic_accession=20213353, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  239759 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239759, CurrentID=0, Status='live', Symbol='Liph', Aliases='C130037N08Rik, D16Wsu119e, H06, Lpdlr, P3, PLA1B, lpd1, lpd2, mPA-PLA1', description='lipase, member H', other_designations='lipase H', map_location='16 12.95 cM', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=21953817, end_position_on_the_genomic_accession=21995542, orientation='minus', exon_count=10, OMIM='', no_hdr0=''),
  258909 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258909, CurrentID=0, Status='live', Symbol='Olfr1282', Aliases='MOR248-16, MOR248-2, Olfr1557', description='olfactory receptor 1282', other_designations='GA_x6K02T2Q125-72387537-72386620|olfactory receptor 1557|olfactory receptor MOR248-16|olfactory receptor MOR248-2', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111335159, end_position_on_the_genomic_accession=111336076, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258910 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258910, CurrentID=0, Status='live', Symbol='Olfr1280', Aliases='MOR248-1', description='olfactory receptor 1280', other_designations='GA_x6K02T2Q125-72366920-72367837|olfactory receptor MOR248-1', map_location='2 E3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=111308466, end_position_on_the_genomic_accession=111374131, orientation='plus', exon_count=8, OMIM='', no_hdr0=''),
  258912 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258912, CurrentID=0, Status='live', Symbol='Olfr1378', Aliases='MOR129-2', description='olfactory receptor 1378', other_designations='GA_x6K02T2QP88-4469162-4468215|olfactory receptor MOR129-2', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50968500, end_position_on_the_genomic_accession=50969967, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258913 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258913, CurrentID=0, Status='live', Symbol='Olfr1377', Aliases='MOR129-1', description='olfactory receptor 1377', other_designations='GA_x6K02T2QP88-4453480-4452557|olfactory receptor MOR129-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=50984703, end_position_on_the_genomic_accession=50985626, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258914 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258914, CurrentID=0, Status='live', Symbol='Olfr211', Aliases='MOR253-9', description='olfactory receptor 211', other_designations='GA_x54KRFPKN04-58127726-58128655|olfactory receptor MOR253-9', map_location='6 F1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=116489117, end_position_on_the_genomic_accession=116494540, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258915 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258915, CurrentID=0, Status='live', Symbol='Olfr1348', Aliases='MOR103-9', description='olfactory receptor 1348', other_designations='GA_x6K02T2QGBW-3232059-3231121|olfactory receptor MOR103-9', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6501248, end_position_on_the_genomic_accession=6508435, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258917 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258917, CurrentID=0, Status='live', Symbol='Olfr1336', Aliases='MOR103-7', description='olfactory receptor 1336', other_designations='GA_x6K02T2QGBW-3190370-3191314|olfactory receptor MOR103-7', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6460511, end_position_on_the_genomic_accession=6461455, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258918 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258918, CurrentID=0, Status='live', Symbol='Olfr1346', Aliases='MOR103-6', description='olfactory receptor 1346', other_designations='GA_x6K02T2QGBW-3203957-3204898|olfactory receptor MOR103-6', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6474112, end_position_on_the_genomic_accession=6475053, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258919 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258919, CurrentID=0, Status='live', Symbol='Olfr1179', Aliases='MOR225-2', description='olfactory receptor 1179', other_designations='GA_x6K02T2Q125-49890854-49889931|olfactory receptor MOR225-2', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88402009, end_position_on_the_genomic_accession=88402932, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258920 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258920, CurrentID=0, Status='live', Symbol='Olfr1180', Aliases='MOR225-1', description='olfactory receptor 1180', other_designations='GA_x6K02T2Q125-49900552-49899623|olfactory receptor MOR225-1', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=88411627, end_position_on_the_genomic_accession=88412689, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258922 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258922, CurrentID=0, Status='live', Symbol='Olfr267', Aliases='MOR262-1', description='olfactory receptor 267', other_designations='GA_x6K02T2N78B-1272842-1273783|olfactory receptor MOR262-1', map_location='4 B2-C2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=58783130, end_position_on_the_genomic_accession=58788108, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258923 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258923, CurrentID=0, Status='live', Symbol='Olfr1', Aliases='I54, MOR135-13', description='olfactory receptor 1', other_designations='GA_x6K02T2P1NL-3556334-3555390|odorant receptor I54|olfactory receptor 135-13|olfactory receptor MOR135-13|olfr1', map_location='11 45.33 cM', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73395076, end_position_on_the_genomic_accession=73399495, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258924 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258924, CurrentID=0, Status='live', Symbol='Olfr376', Aliases='MOR135-12', description='olfactory receptor 376', other_designations='GA_x6K02T2P1NL-3535075-3536028|olfactory receptor MOR135-12', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73371246, end_position_on_the_genomic_accession=73375704, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  258925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258925, CurrentID=0, Status='live', Symbol='Olfr20', Aliases='MOR135-11, MTPCR06, MTPCR55, Olfr21', description='olfactory receptor 20', other_designations='GA_x6K02T2P1NL-3514066-3515010|olfactory receptor 21', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73350859, end_position_on_the_genomic_accession=73354699, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258927 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258927, CurrentID=0, Status='live', Symbol='Olfr481', Aliases='MOR204-2, MOR204-39', description='olfactory receptor 481', other_designations='GA_x6K02T2PBJ9-10409785-10410723|olfactory receptor 204-2|olfactory receptor MOR204-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108080796, end_position_on_the_genomic_accession=108081734, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258928 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258928, CurrentID=0, Status='live', Symbol='Olfr477', Aliases='MOR204-1', description='olfactory receptor 477', other_designations='GA_x6K02T2PBJ9-10319672-10320604|olfactory receptor 204-1|olfactory receptor MOR204-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107990367, end_position_on_the_genomic_accession=107991299, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258931, CurrentID=0, Status='live', Symbol='Olfr807', Aliases='MOR117-1', description='olfactory receptor 807', other_designations='GA_x6K02T2PULF-11434134-11433199|olfactory receptor MOR117-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129754495, end_position_on_the_genomic_accession=129757800, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258933 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258933, CurrentID=0, Status='live', Symbol='Olfr796', Aliases='MOR269-1', description='olfactory receptor 796', other_designations='GA_x6K02T2PULF-11287099-11286167|olfactory receptor MOR269-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129607547, end_position_on_the_genomic_accession=129608479, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258935 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258935, CurrentID=0, Status='live', Symbol='Olfr790', Aliases='MOR112-1', description='olfactory receptor 790', other_designations='GA_x6K02T2PULF-11179777-11180721|olfactory receptor MOR112-1', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129500886, end_position_on_the_genomic_accession=129501854, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258938 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258938, CurrentID=0, Status='live', Symbol='Olfr1417', Aliases='MOR266-2', description='olfactory receptor 1417', other_designations='GA_x6K02T2RE5P-2172809-2171862|olfactory receptor MOR266-2', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=11828077, end_position_on_the_genomic_accession=11829024, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258939 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258939, CurrentID=0, Status='live', Symbol='Olfr63', Aliases='M4, MOR267-1', description='olfactory receptor 63', other_designations='GA_x6K02T2NTC5-9778-8828|odorant receptor M4|olfactory receptor 267-1', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33268726, end_position_on_the_genomic_accession=33269676, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258941 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258941, CurrentID=0, Status='live', Symbol='Olfr354', Aliases='MOR127-4', description='olfactory receptor 354', other_designations='GA_x6K02T2NLDC-33601476-33602429|olfactory receptor MOR127-4', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36906948, end_position_on_the_genomic_accession=36907901, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258945 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258945, CurrentID=0, Status='live', Symbol='Olfr347', Aliases='MOR136-9', description='olfactory receptor 347', other_designations='GA_x6K02T2NLDC-33428755-33429693|olfactory receptor MOR136-9', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36734323, end_position_on_the_genomic_accession=36735261, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258947 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258947, CurrentID=0, Status='live', Symbol='Olfr345', Aliases='MOR136-7', description='olfactory receptor 345', other_designations='GA_x6K02T2NLDC-33334179-33335114|olfactory receptor MOR136-7', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36640041, end_position_on_the_genomic_accession=36640976, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258950 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258950, CurrentID=0, Status='live', Symbol='Olfr342', Aliases='MOR136-4', description='olfactory receptor 342', other_designations='GA_x6K02T2NLDC-33222024-33222962|olfactory receptor MOR136-4', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36527414, end_position_on_the_genomic_accession=36528352, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258952 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258952, CurrentID=0, Status='live', Symbol='Olfr341', Aliases='MOR136-2', description='olfactory receptor 341', other_designations='GA_x6K02T2NLDC-33174915-33173974|olfactory receptor MOR136-2', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36479187, end_position_on_the_genomic_accession=36480128, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258953 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258953, CurrentID=0, Status='live', Symbol='Olfr340', Aliases='MOR136-1', description='olfactory receptor 340', other_designations='GA_x6K02T2NLDC-33147742-33148680|olfactory receptor MOR136-1', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=36452587, end_position_on_the_genomic_accession=36453525, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258954 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258954, CurrentID=0, Status='live', Symbol='Olfr522', Aliases='MOR103-5', description='olfactory receptor 522', other_designations='GA_x6K02T2PBJ9-42315125-42314187|olfactory receptor MOR103-5', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140162010, end_position_on_the_genomic_accession=140162948, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258956, CurrentID=0, Status='live', Symbol='Olfr535', Aliases='MOR253-7', description='olfactory receptor 535', other_designations='GA_x6K02T2PBJ9-42641642-42642574|olfactory receptor MOR253-7', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140492640, end_position_on_the_genomic_accession=140493578, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258958 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258958, CurrentID=0, Status='live', Symbol='Olfr525', Aliases='MOR251-2', description='olfactory receptor 525', other_designations='GA_x6K02T2PBJ9-42472898-42473827|olfactory receptor MOR251-2', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140322701, end_position_on_the_genomic_accession=140323630, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258959 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258959, CurrentID=0, Status='live', Symbol='Olfr170', Aliases='MOR273-2', description='olfactory receptor 170', other_designations='GA_x54KRFPKG5P-16052703-16051765|olfactory receptor MOR273-2', map_location='16 B1-B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19605725, end_position_on_the_genomic_accession=19606666, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258960 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258960, CurrentID=0, Status='live', Symbol='Olfr171', Aliases='MOR273-1, MOR273-5', description='olfactory receptor 171', other_designations='GA_x54KRFPKG5P-16071018-16070077|olfactory receptor MOR273-1', map_location='16 B1-B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19624157, end_position_on_the_genomic_accession=19625101, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258961 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258961, CurrentID=0, Status='live', Symbol='Olfr631', Aliases='MOR3-1', description='olfactory receptor 631', other_designations='GA_x6K02T2PBJ9-6662699-6663658|olfactory receptor MOR3-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103904306, end_position_on_the_genomic_accession=103930666, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  258962 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258962, CurrentID=0, Status='live', Symbol='Olfr53', Aliases='IE12, MOR253-5', description='olfactory receptor 53', other_designations='GA_x6K02T2PBJ9-42798103-42799038|olfactory receptor MOR253-5', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140646452, end_position_on_the_genomic_accession=140652919, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  258964 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258964, CurrentID=0, Status='live', Symbol='Olfr541', Aliases='MOR253-3', description='olfactory receptor 541', other_designations='GA_x6K02T2PBJ9-42850324-42851256|olfactory receptor MOR253-3', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140704253, end_position_on_the_genomic_accession=140705191, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258967 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258967, CurrentID=0, Status='live', Symbol='Olfr1250', Aliases='MOR231-7', description='olfactory receptor 1250', other_designations='GA_x6K02T2Q125-51098877-51097933|olfactory receptor MOR231-7', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89656495, end_position_on_the_genomic_accession=89657439, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258968 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258968, CurrentID=0, Status='live', Symbol='Olfr1247', Aliases='MOR231-6', description='olfactory receptor 1247', other_designations='GA_x6K02T2Q125-51051555-51050611|olfactory receptor MOR231-6', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89608750, end_position_on_the_genomic_accession=89614124, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258969 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258969, CurrentID=0, Status='live', Symbol='Olfr1226', Aliases='MOR233-2', description='olfactory receptor 1226', other_designations='GA_x6K02T2Q125-50672630-50671698|olfactory receptor MOR233-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89193100, end_position_on_the_genomic_accession=89194032, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258970 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258970, CurrentID=0, Status='live', Symbol='Olfr1242', Aliases='MOR231-5', description='olfactory receptor 1242', other_designations='GA_x6K02T2Q125-50937307-50936387|olfactory receptor MOR231-5', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89493294, end_position_on_the_genomic_accession=89498213, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258971 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258971, CurrentID=0, Status='live', Symbol='Olfr1243', Aliases='MOR231-4', description='olfactory receptor 1243', other_designations='GA_x6K02T2Q125-50972538-50971621|olfactory receptor MOR231-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89527491, end_position_on_the_genomic_accession=89528408, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258973 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258973, CurrentID=0, Status='live', Symbol='Olfr1228', Aliases='MOR233-1', description='olfactory receptor 1228', other_designations='GA_x6K02T2Q125-50696209-50695274|olfactory receptor MOR233-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89248721, end_position_on_the_genomic_accession=89253707, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  258974 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258974, CurrentID=0, Status='live', Symbol='Olfr1233', Aliases='MOR238-1', description='olfactory receptor 1233', other_designations='GA_x6K02T2Q125-50784973-50784056|olfactory receptor MOR238-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89339312, end_position_on_the_genomic_accession=89342391, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258975, CurrentID=0, Status='live', Symbol='Olfr1234', Aliases='MOR231-2', description='olfactory receptor 1234', other_designations='GA_x6K02T2Q125-50805620-50804676|olfactory receptor MOR231-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89362483, end_position_on_the_genomic_accession=89364861, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  258976 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258976, CurrentID=0, Status='live', Symbol='Olfr1262', Aliases='MOR234-1, or-leaf2', description='olfactory receptor 1262', other_designations='GA_x6K02T2Q125-51434523-51435437|olfactory receptor MOR234-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90002408, end_position_on_the_genomic_accession=90003322, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258977 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258977, CurrentID=0, Status='live', Symbol='Olfr1273-ps', Aliases='MOR227-4, Olfr1273', description='olfactory receptor 1273, pseudogene', other_designations='GA_x6K02T2Q125-51650793-51649870|olfactory receptor MOR227-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90295936, end_position_on_the_genomic_accession=90296859, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258980 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258980, CurrentID=0, Status='live', Symbol='Olfr1258', Aliases='MOR232-3', description='olfactory receptor 1258', other_designations='GA_x6K02T2Q125-51361752-51362687|olfactory receptor MOR232-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89929811, end_position_on_the_genomic_accession=89930746, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258983 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258983, CurrentID=0, Status='live', Symbol='Olfr1260', Aliases='MOR232-2', description='olfactory receptor 1260', other_designations='GA_x6K02T2Q125-51409740-51410672|olfactory receptor MOR232-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89977780, end_position_on_the_genomic_accession=89978712, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258984 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258984, CurrentID=0, Status='live', Symbol='Olfr1257', Aliases='MOR232-1', description='olfactory receptor 1257', other_designations='GA_x6K02T2Q125-51319458-51320387|olfactory receptor MOR232-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89880828, end_position_on_the_genomic_accession=89881757, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258985 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258985, CurrentID=0, Status='live', Symbol='Olfr1256', Aliases='MOR231-1', description='olfactory receptor 1256', other_designations='GA_x6K02T2Q125-51276848-51275928|olfactory receptor MOR231-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89835023, end_position_on_the_genomic_accession=89835943, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258987 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258987, CurrentID=0, Status='live', Symbol='Olfr1270', Aliases='MOR227-1', description='olfactory receptor 1270', other_designations='GA_x6K02T2Q125-51584440-51583526|olfactory receptor MOR227-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=90149019, end_position_on_the_genomic_accession=90150036, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258988 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258988, CurrentID=0, Status='live', Symbol='Olfr38', Aliases='18A, MOR257-2', description='olfactory receptor 38', other_designations='GA_x6K02T2P3E9-4797841-4796888', map_location='6 20.89 cM', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42762054, end_position_on_the_genomic_accession=42763007, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258989 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258989, CurrentID=0, Status='live', Symbol='Olfr457', Aliases='MOR257-1', description='olfactory receptor 457', other_designations='GA_x6K02T2P3E9-5088893-5089834|olfactory receptor MOR257-1', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42471235, end_position_on_the_genomic_accession=42472176, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258990 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258990, CurrentID=0, Status='live', Symbol='Olfr447', Aliases='MOR261-1', description='olfactory receptor 447', other_designations='GA_x6K02T2P3E9-4647978-4647046|olfactory receptor MOR261-1', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42911525, end_position_on_the_genomic_accession=42912457, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258991 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258991, CurrentID=0, Status='live', Symbol='Olfr1496', Aliases='MOR127-1', description='olfactory receptor 1496', other_designations='GA_x6K02T2RE5P-4112771-4113718|olfactory receptor MOR127-1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13780548, end_position_on_the_genomic_accession=13781614, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258992 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258992, CurrentID=0, Status='live', Symbol='Olfr1494', Aliases='MOR266-1', description='olfactory receptor 1494', other_designations='GA_x6K02T2RE5P-4082427-4083374|olfactory receptor MOR266-1', map_location='19 C1', chromosome='19', genomic_nucleotide_accession_version='NC_000085.6', start_position_on_the_genomic_accession=13749108, end_position_on_the_genomic_accession=13750055, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  258993 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258993, CurrentID=0, Status='live', Symbol='Olfr206', Aliases='MOR182-4', description='olfactory receptor 206', other_designations='GA_x54KRFPKG5P-55560552-55559632|olfactory receptor MOR182-4', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59344779, end_position_on_the_genomic_accession=59345699, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258997 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258997, CurrentID=0, Status='live', Symbol='Olfr202', Aliases='MOR182-1', description='olfactory receptor 202', other_designations='GA_x54KRFPKG5P-55498766-55497843|olfactory receptor MOR182-1', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59283572, end_position_on_the_genomic_accession=59284495, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  258999 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=258999, CurrentID=0, Status='live', Symbol='Olfr178', Aliases='MOR184-6', description='olfactory receptor 178', other_designations='GA_x54KRFPKG5P-55108059-55107100|olfactory receptor MOR184-6', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58889259, end_position_on_the_genomic_accession=58890218, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259000 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259000, CurrentID=0, Status='live', Symbol='Olfr195', Aliases='MOR184-5', description='olfactory receptor 195', other_designations='GA_x54KRFPKG5P-55369823-55370749|olfactory receptor MOR184-5', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=59131709, end_position_on_the_genomic_accession=59149817, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  259001 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259001, CurrentID=0, Status='live', Symbol='Olfr181', Aliases='MOR184-4', description='olfactory receptor 181', other_designations='GA_x54KRFPKG5P-55145984-55145034|olfactory receptor 184-4|olfactory receptor MOR184-4', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58925557, end_position_on_the_genomic_accession=58928644, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259002 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259002, CurrentID=0, Status='live', Symbol='Olfr173', Aliases='MOR184-3', description='olfactory receptor 173', other_designations='GA_x54KRFPKG5P-54960233-54959268|olfactory receptor MOR184-3', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58796779, end_position_on_the_genomic_accession=58797942, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259003, CurrentID=0, Status='live', Symbol='Olfr172', Aliases='MOR184-2', description='olfactory receptor 172', other_designations='GA_x54KRFPKG5P-54930346-54929417|olfactory receptor MOR184-2', map_location='16 C2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=58760152, end_position_on_the_genomic_accession=58761240, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259005 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259005, CurrentID=0, Status='live', Symbol='Olfr139', Aliases='M5, MOR255-2', description='olfactory receptor 139', other_designations='GA_x6K02T2P1NL-4202012-4201065|odorant receptor M5|olfactory receptor 1|olfactory receptor 255-2|olfactory receptor MOR255-2', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74044325, end_position_on_the_genomic_accession=74045272, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259006 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259006, CurrentID=0, Status='live', Symbol='Olfr399', Aliases='MOR255-1', description='olfactory receptor 399', other_designations='GA_x6K02T2P1NL-4211516-4210581|olfactory receptor MOR255-1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=74053776, end_position_on_the_genomic_accession=74054778, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259008 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259008, CurrentID=0, Status='live', Symbol='Olfr392', Aliases='MOR135-9', description='olfactory receptor 392', other_designations='GA_x6K02T2P1NL-3966976-3966038|olfactory receptor MOR135-9', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73814124, end_position_on_the_genomic_accession=73816877, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259009 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259009, CurrentID=0, Status='live', Symbol='Olfr394', Aliases='MOR135-8', description='olfactory receptor 394', other_designations='GA_x6K02T2P1NL-4043306-4042374|olfactory receptor MOR135-8', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73887438, end_position_on_the_genomic_accession=73888370, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259011 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259011, CurrentID=0, Status='live', Symbol='Olfr389', Aliases='MOR135-6', description='olfactory receptor 389', other_designations='GA_x6K02T2P1NL-3932085-3931147|olfactory receptor MOR135-6', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73776387, end_position_on_the_genomic_accession=73780589, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  259012 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259012, CurrentID=0, Status='live', Symbol='Olfr1052', Aliases='MOR172-1', description='olfactory receptor 1052', other_designations='GA_x6K02T2Q125-47777498-47778436|olfactory receptor 172-1|olfactory receptor MOR172-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86297818, end_position_on_the_genomic_accession=86298756, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259013 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259013, CurrentID=0, Status='live', Symbol='Olfr1044', Aliases='MOR185-4', description='olfactory receptor 1044', other_designations='GA_x6K02T2Q125-47640742-47639798|olfactory receptor 185-4|olfactory receptor MOR185-4', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86170871, end_position_on_the_genomic_accession=86171815, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259015 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259015, CurrentID=0, Status='live', Symbol='Olfr1038-ps', Aliases='MOR185-3, Olfr1038', description='olfactory receptor 1038, pseudogene', other_designations='GA_x6K02T2Q125-47591072-47592031|olfactory receptor 185-3|olfactory receptor MOR185-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86120224, end_position_on_the_genomic_accession=86123037, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  259017 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259017, CurrentID=0, Status='live', Symbol='Olfr1019', Aliases='MOR180-1', description='olfactory receptor 1019', other_designations='GA_x6K02T2Q125-47320309-47319377|olfactory receptor 180-1|olfactory receptor MOR180-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=85840857, end_position_on_the_genomic_accession=85841789, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259020 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259020, CurrentID=0, Status='live', Symbol='Olfr1056', Aliases='MOR186-2', description='olfactory receptor 1056', other_designations='GA_x6K02T2Q125-47827833-47826892|olfactory receptor MOR186-2', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86355359, end_position_on_the_genomic_accession=86356437, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259022 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259022, CurrentID=0, Status='live', Symbol='Olfr1061', Aliases='MOR188-1, MOR188-7, MOR188-9, Olfr1515', description='olfactory receptor 1061', other_designations='GA_x6K02T2Q125-47883395-47882454|olfactory receptor 1515|olfactory receptor MOR188-1|olfactory receptor MOR188-7', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86413109, end_position_on_the_genomic_accession=86414050, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259026 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259026, CurrentID=0, Status='live', Symbol='Olfr378', Aliases='MOR135-2', description='olfactory receptor 378', other_designations='GA_x6K02T2P1NL-3586282-3585338|odorant receptor|olfactory receptor MOR135-2', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73425037, end_position_on_the_genomic_accession=73428477, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259027 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259027, CurrentID=0, Status='live', Symbol='Olfr380', Aliases='MOR135-1', description='olfactory receptor 380', other_designations='GA_x6K02T2P1NL-3613021-3612086|olfactory receptor MOR135-1', map_location='11 B4', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=73453275, end_position_on_the_genomic_accession=73454210, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259028 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259028, CurrentID=0, Status='live', Symbol='Olfr532', Aliases='MOR251-1', description='olfactory receptor 532', other_designations='GA_x6K02T2PBJ9-42570051-42569122|olfactory receptor MOR251-1', map_location='7 F4', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=140418842, end_position_on_the_genomic_accession=140419771, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259030 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259030, CurrentID=0, Status='live', Symbol='Olfr1124', Aliases='MOR264-24, MOR264-3-ps1, Olfr1125, Olfr1124', description='olfactory receptor 1124', other_designations='GA_x6K02T2Q125-48935224-48935664|GA_x6K02T2Q125-48936945-48937901|olfactory receptor 1124, pseudogene 1|olfactory receptor 1125|olfactory receptor MOR264-3', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87434489, end_position_on_the_genomic_accession=87435445, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259031 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259031, CurrentID=0, Status='live', Symbol='Olfr1120', Aliases='MOR264-2', description='olfactory receptor 1120', other_designations='GA_x6K02T2Q125-48849180-48850100|olfactory receptor MOR264-2', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87354819, end_position_on_the_genomic_accession=87358390, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  259032 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259032, CurrentID=0, Status='live', Symbol='Olfr1134', Aliases='MOR176-1', description='olfactory receptor 1134', other_designations='GA_x6K02T2Q125-49162076-49161138|olfactory receptor MOR176-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87655925, end_position_on_the_genomic_accession=87658457, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259033 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259033, CurrentID=0, Status='live', Symbol='Olfr1122', Aliases='MOR264-1', description='olfactory receptor 1122', other_designations='GA_x6K02T2Q125-48880078-48881058|olfactory receptor MOR264-1', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=87387707, end_position_on_the_genomic_accession=87388687, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259034 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259034, CurrentID=0, Status='live', Symbol='Olfr705', Aliases='MOR283-2', description='olfactory receptor 705', other_designations='GA_x6K02T2PBJ9-9256348-9255398|T2 olfactory receptor|olfactory receptor MOR283-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106873293, end_position_on_the_genomic_accession=106874243, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259035 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259035, CurrentID=0, Status='live', Symbol='Olfr714', Aliases='MOR263-2, P4', description='olfactory receptor 714', other_designations='GA_x6K02T2PBJ9-9453401-9454354|P4 olfactory receptor|olfactory receptor MOR263-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107073830, end_position_on_the_genomic_accession=107074783, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259036 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259036, CurrentID=0, Status='live', Symbol='Olfr713', Aliases='MOR263-1, P3', description='olfactory receptor 713', other_designations='GA_x6K02T2PBJ9-9415724-9416677|P3 olfactory receptor|olfactory receptor MOR263-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=107036136, end_position_on_the_genomic_accession=107037110, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259037 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259037, CurrentID=0, Status='live', Symbol='Olfr711', Aliases='MOR103-4', description='olfactory receptor 711', other_designations='GA_x6K02T2PBJ9-9352783-9351839|m51 olfactory receptor|olfactory receptor MOR103-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=106971368, end_position_on_the_genomic_accession=106975451, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  259038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259038, CurrentID=0, Status='live', Symbol='Olfr283', Aliases='MOR160-1', description='olfactory receptor 283', other_designations='GA_x6K02T2NBG7-5351896-5352825|olfactory receptor MOR160-1', map_location='15 F3', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=98378179, end_position_on_the_genomic_accession=98379108, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259039 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259039, CurrentID=0, Status='live', Symbol='Olfr1413', Aliases='MOR208-1', description='olfactory receptor 1413', other_designations='GA_x6K02T2R7CC-81180849-81179878|olfactory receptor MOR208-1', map_location='1 D-E1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92573125, end_position_on_the_genomic_accession=92574206, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259040 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259040, CurrentID=0, Status='live', Symbol='Olfr1416', Aliases='MOR103-3', description='olfactory receptor 1416', other_designations='GA_x6K02T2R7CC-81277975-81278913|olfactory receptor MOR103-3', map_location='1 D-E1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92479681, end_position_on_the_genomic_accession=92480619, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259041 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259041, CurrentID=0, Status='live', Symbol='Olfr1414', Aliases='MOR103-2', description='olfactory receptor 1414', other_designations='GA_x6K02T2R7CC-81245243-81246181|olfactory receptor MOR103-2', map_location='1 D-E1', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92511048, end_position_on_the_genomic_accession=92518515, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259042 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259042, CurrentID=0, Status='live', Symbol='Olfr1351', Aliases='MOR139-4', description='olfactory receptor 1351', other_designations='GA_x6K02T2QGN0-2794907-2793948|olfactory receptor MOR139-4', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=79017324, end_position_on_the_genomic_accession=79018283, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259044 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259044, CurrentID=0, Status='live', Symbol='Olfr1353', Aliases='MOR139-2', description='olfactory receptor 1353', other_designations='GA_x6K02T2QGN0-2842591-2841662|olfactory receptor MOR139-2', map_location='10 C2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78963309, end_position_on_the_genomic_accession=78970612, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259045 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259045, CurrentID=0, Status='live', Symbol='Olfr669', Aliases='MOR34-6', description='olfactory receptor 669', other_designations='GA_x6K02T2PBJ9-7567376-7568329|olfactory receptor MOR34-6', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104938528, end_position_on_the_genomic_accession=104939481, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259046 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259046, CurrentID=0, Status='live', Symbol='Olfr679', Aliases='MOR40-2', description='olfactory receptor 679', other_designations='GA_x6K02T2PBJ9-7714499-7715446|olfactory receptor MOR40-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105085718, end_position_on_the_genomic_accession=105086665, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259047 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259047, CurrentID=0, Status='live', Symbol='Olfr683', Aliases='MOR40-1', description='olfactory receptor 683', other_designations='GA_x6K02T2PBJ9-7773007-7772066|odorant receptor S51|olfactory receptor MOR40-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105143350, end_position_on_the_genomic_accession=105146459, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  259048 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259048, CurrentID=0, Status='live', Symbol='Olfr600', Aliases='MOR39-1', description='olfactory receptor 600', other_designations='GA_x6K02T2PBJ9-6056235-6055291|olfactory receptor MOR39-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103345982, end_position_on_the_genomic_accession=103346926, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259050 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259050, CurrentID=0, Status='live', Symbol='Olfr652', Aliases='MOR318', description='olfactory receptor 652', other_designations='', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104564152, end_position_on_the_genomic_accession=104565212, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259051 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259051, CurrentID=0, Status='live', Symbol='Olfr658', Aliases='MOR34-5', description='olfactory receptor 658', other_designations='GA_x6K02T2PBJ9-7273558-7272587|olfactory receptor MOR34-5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104642880, end_position_on_the_genomic_accession=104647305, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259052 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259052, CurrentID=0, Status='live', Symbol='Olfr659', Aliases='MOR34-4', description='olfactory receptor 659', other_designations='GA_x6K02T2PBJ9-7298889-7299857|olfactory receptor MOR34-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104670704, end_position_on_the_genomic_accession=104671672, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259053 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259053, CurrentID=0, Status='live', Symbol='Olfr362', Aliases='MOR158-1', description='olfactory receptor 362', other_designations='GA_x6K02T2NLDC-33797415-33796462|olfactory receptor MOR158-1', map_location='2 C1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=37104695, end_position_on_the_genomic_accession=37105648, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259054 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259054, CurrentID=0, Status='live', Symbol='Olfr589', Aliases='MOR32-3', description='olfactory receptor 589', other_designations='GA_x6K02T2PBJ9-5871256-5870303|olfactory receptor MOR32-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103154792, end_position_on_the_genomic_accession=103155745, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259055, CurrentID=0, Status='live', Symbol='Olfr582', Aliases='MOR30-3', description='olfactory receptor 582', other_designations='GA_x6K02T2PBJ9-5752857-5753801|olfactory receptor MOR30-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103041481, end_position_on_the_genomic_accession=103042440, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259057 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259057, CurrentID=0, Status='live', Symbol='Olfr649', Aliases='MOR31-7', description='olfactory receptor 649', other_designations='GA_x6K02T2PBJ9-6924585-6923647|olfactory receptor MOR31-7', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104189267, end_position_on_the_genomic_accession=104190205, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259058 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259058, CurrentID=0, Status='live', Symbol='Olfr646', Aliases='MOR33-2', description='olfactory receptor 646', other_designations='GA_x6K02T2PBJ9-6841330-6842268|olfactory receptor MOR33-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104106281, end_position_on_the_genomic_accession=104107219, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259062 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259062, CurrentID=0, Status='live', Symbol='Olfr667', Aliases='MOR34-11, MOR34-2', description='olfactory receptor 667', other_designations='GA_x6K02T2PBJ9-7546146-7545166|olfactory receptor MOR34-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104916314, end_position_on_the_genomic_accession=104917294, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259063 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259063, CurrentID=0, Status='live', Symbol='Olfr691', Aliases='MOR31-6', description='olfactory receptor 691', other_designations='GA_x6K02T2PBJ9-7966695-7965727|olfactory receptor MOR31-6', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105336746, end_position_on_the_genomic_accession=105337714, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259064 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259064, CurrentID=0, Status='live', Symbol='Olfr124', Aliases='A3, MOR256-3', description='olfactory receptor 124', other_designations='GA_x6K02T2PSCP-2264806-2265753|olfactory receptor MOR256-3', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=37805069, end_position_on_the_genomic_accession=37806190, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259066 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259066, CurrentID=0, Status='live', Symbol='Olfr520', Aliases='MOR101-1', description='olfactory receptor 520', other_designations='GA_x6K02T2PBJ9-2411789-2412739|olfactory receptor MOR101-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=99735145, end_position_on_the_genomic_accession=99736095, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259067, CurrentID=0, Status='live', Symbol='Olfr449', Aliases='MOR103-1', description='olfactory receptor 449', other_designations='GA_x6K02T2P3E9-4722003-4721068|olfactory receptor MOR103-1', map_location='6 B2', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42834384, end_position_on_the_genomic_accession=42838882, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259069, CurrentID=0, Status='live', Symbol='Olfr1389', Aliases='MOR256-1', description='olfactory receptor 1389', other_designations='GA_x6K02T2QP88-6004387-6003452|olfactory receptor MOR256-1', map_location='11 B1', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=49430427, end_position_on_the_genomic_accession=49431499, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259070 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259070, CurrentID=0, Status='live', Symbol='Olfr829', Aliases='MOR147-1', description='olfactory receptor 829', other_designations='GA_x6K02T2PVTD-12599710-12600648|olfactory receptor MOR147-1', map_location='9 A2-A5', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18855062, end_position_on_the_genomic_accession=18857592, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  259071 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259071, CurrentID=0, Status='live', Symbol='Olfr166', Aliases='MOR270-1', description='olfactory receptor 166', other_designations='GA_x54KRFPKG5P-15934912-15935850|olfactory receptor MOR270-1', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=19486840, end_position_on_the_genomic_accession=19487778, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259072 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259072, CurrentID=0, Status='live', Symbol='Olfr686', Aliases='MOR35-1', description='olfactory receptor 686', other_designations='GA_x6K02T2PBJ9-7832633-7831680|olfactory receptor MOR35-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105203388, end_position_on_the_genomic_accession=105204341, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259073 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259073, CurrentID=0, Status='live', Symbol='Olfr603', Aliases='MOR32-14_i, MOR32-2-ps1, Olfr604, Olfr603', description='olfactory receptor 603', other_designations='GA_x6K02T2PBJ9-6092550-6092362|GA_x6K02T2PBJ9-6096387-6095449|olfactory receptor 603, pseudogene 1|olfactory receptor MOR32-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103383062, end_position_on_the_genomic_accession=103384000, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259074 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259074, CurrentID=0, Status='live', Symbol='Olfr1352', Aliases='MOR139-1', description='olfactory receptor 1352', other_designations='GA_x6K02T2QGN0-2828447-2827518|olfactory receptor MOR139-1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78981050, end_position_on_the_genomic_accession=78984721, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  259075 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259075, CurrentID=0, Status='live', Symbol='Olfr641', Aliases='MOR13-3', description='olfactory receptor 641', other_designations='GA_x6K02T2PBJ9-6773690-6774628|olfactory receptor MOR13-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104039798, end_position_on_the_genomic_accession=104040736, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  545622 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545622, CurrentID=0, Status='live', Symbol='Ptpn3', Aliases='9530011I20Rik, PTP-H1, PTPCL', description='protein tyrosine phosphatase, non-receptor type 3', other_designations='protein-tyrosine phosphatase H1', map_location='4 31.66 cM', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=57190841, end_position_on_the_genomic_accession=57339046, orientation='minus', exon_count=30, OMIM='', no_hdr0=''),
  259078 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259078, CurrentID=0, Status='live', Symbol='Olfr656', Aliases='MOR27-1', description='olfactory receptor 656', other_designations='GA_x6K02T2PBJ9-7245486-7246451|olfactory receptor MOR27-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104617657, end_position_on_the_genomic_accession=104618646, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259080 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259080, CurrentID=0, Status='live', Symbol='Olfr619', Aliases='MOR31-5', description='olfactory receptor 619', other_designations='GA_x6K02T2PBJ9-6326488-6327450|olfactory receptor MOR31-5', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103603589, end_position_on_the_genomic_accession=103604712, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259081 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259081, CurrentID=0, Status='live', Symbol='Olfr643', Aliases='MOR13-2', description='olfactory receptor 643', other_designations='GA_x6K02T2PBJ9-6793628-6792684|olfactory receptor MOR13-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104058619, end_position_on_the_genomic_accession=104137416, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259083 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259083, CurrentID=0, Status='live', Symbol='Olfr547', Aliases='MOR31-4', description='olfactory receptor 547', other_designations='GA_x6K02T2PBJ9-5256044-5256988|olfactory receptor MOR31-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102534749, end_position_on_the_genomic_accession=102535688, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259084 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259084, CurrentID=0, Status='live', Symbol='Olfr615', Aliases='MOR19-2', description='olfactory receptor 615', other_designations='GA_x6K02T2PBJ9-6284902-6285843|olfactory receptor MOR19-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103560479, end_position_on_the_genomic_accession=103561420, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259085 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259085, CurrentID=0, Status='live', Symbol='Olfr610', Aliases='MOR9-2', description='olfactory receptor 610', other_designations='GA_x6K02T2PBJ9-6221839-6220892|olfactory receptor MOR9-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103505997, end_position_on_the_genomic_accession=103506944, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259086 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259086, CurrentID=0, Status='live', Symbol='Olfr609', Aliases='MOR9-1', description='olfactory receptor 609', other_designations='GA_x6K02T2PBJ9-6207735-6206776|olfactory receptor MOR9-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103491917, end_position_on_the_genomic_accession=103492876, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259087 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259087, CurrentID=0, Status='live', Symbol='Olfr622', Aliases='MOR26-1', description='olfactory receptor 622', other_designations='GA_x6K02T2PBJ9-6362863-6361910|olfactory receptor MOR26-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103639185, end_position_on_the_genomic_accession=103640138, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259088 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259088, CurrentID=0, Status='live', Symbol='Olfr639', Aliases='MOR12-1', description='olfactory receptor 639', other_designations='GA_x6K02T2PBJ9-6747143-6746193|olfactory receptor MOR12-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104011750, end_position_on_the_genomic_accession=104012700, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259089 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259089, CurrentID=0, Status='live', Symbol='Olfr571', Aliases='MOR21-1', description='olfactory receptor 571', other_designations='GA_x6K02T2PBJ9-5620890-5619922|olfactory receptor MOR21-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102908869, end_position_on_the_genomic_accession=102909837, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259091 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259091, CurrentID=0, Status='live', Symbol='Olfr585', Aliases='MOR14-4', description='olfactory receptor 585', other_designations='GA_x6K02T2PBJ9-5809085-5810035|olfactory receptor MOR14-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103097743, end_position_on_the_genomic_accession=103098699, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259092 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259092, CurrentID=0, Status='live', Symbol='Olfr569', Aliases='MOR30-1', description='olfactory receptor 569', other_designations='GA_x6K02T2PBJ9-5599295-5598351|olfactory receptor MOR30-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102887207, end_position_on_the_genomic_accession=102888151, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259093 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259093, CurrentID=0, Status='live', Symbol='Olfr572', Aliases='MOR10-2', description='olfactory receptor 572', other_designations='GA_x6K02T2PBJ9-5638785-5639741|olfactory receptor MOR10-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102927630, end_position_on_the_genomic_accession=102928586, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259095 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259095, CurrentID=0, Status='live', Symbol='Olfr568', Aliases='MOR14-11, MOR14-3', description='olfactory receptor 568', other_designations='GA_x6K02T2PBJ9-5588278-5589228|olfactory receptor MOR14-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102877122, end_position_on_the_genomic_accession=102878063, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259096 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259096, CurrentID=0, Status='live', Symbol='Olfr561', Aliases='MOR14-2', description='olfactory receptor 561', other_designations='GA_x6K02T2PBJ9-5491151-5492095|olfactory receptor MOR14-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102774526, end_position_on_the_genomic_accession=102775470, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259097 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259097, CurrentID=0, Status='live', Symbol='Olfr558', Aliases='MOR18-1, POGR', description='olfactory receptor 558', other_designations='GA_x6K02T2PBJ9-5425951-5426904|olfactory receptor MOR18-1|prostate overexpressed G protein coupled receptor', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102702323, end_position_on_the_genomic_accession=102712058, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259099 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259099, CurrentID=0, Status='live', Symbol='Olfr676', Aliases='MOR32-1', description='olfactory receptor 676', other_designations='GA_x6K02T2PBJ9-7664016-7664969|olfactory receptor MOR32-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105035200, end_position_on_the_genomic_accession=105036153, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259102 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259102, CurrentID=0, Status='live', Symbol='Olfr630', Aliases='MOR17-1', description='olfactory receptor 630', other_designations='GA_x6K02T2PBJ9-6483085-6482126|olfactory receptor MOR17-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103754150, end_position_on_the_genomic_accession=103757350, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259103 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259103, CurrentID=0, Status='live', Symbol='Olfr616', Aliases='MOR19-1', description='olfactory receptor 616', other_designations='GA_x6K02T2PBJ9-6289676-6288723|olfactory receptor MOR19-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103564324, end_position_on_the_genomic_accession=103565277, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259104 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259104, CurrentID=0, Status='live', Symbol='Olfr613', Aliases='MOR20-1-ps1, Olfr614, Olfr613', description='olfactory receptor 613', other_designations='GA_x6K02T2PBJ9-6271959-6272393|GA_x6K02T2PBJ9-6275524-6276477|olfactory receptor 613, pseudogene 1|olfactory receptor MOR20-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103550368, end_position_on_the_genomic_accession=103555504, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259105 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259105, CurrentID=0, Status='live', Symbol='Olfr549', Aliases='MOR31-3', description='olfactory receptor 549', other_designations='GA_x6K02T2PBJ9-5274337-5275287|olfactory receptor MOR31-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102554286, end_position_on_the_genomic_accession=102555236, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259106 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259106, CurrentID=0, Status='live', Symbol='Olfr552', Aliases='MOR28-1', description='olfactory receptor 552', other_designations='GA_x6K02T2PBJ9-5323062-5324015|olfactory receptor MOR28-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102604356, end_position_on_the_genomic_accession=102605309, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259107 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259107, CurrentID=0, Status='live', Symbol='Olfr555', Aliases='MOR10-1', description='olfactory receptor 555', other_designations='GA_x6K02T2PBJ9-5375442-5376389|olfactory receptor MOR10-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102658823, end_position_on_the_genomic_accession=102659770, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259108 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259108, CurrentID=0, Status='live', Symbol='Olfr550', Aliases='MOR16-1', description='olfactory receptor 550', other_designations='GA_x6K02T2PBJ9-5297243-5298193|olfactory receptor MOR16-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102571431, end_position_on_the_genomic_accession=102579662, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  259109 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259109, CurrentID=0, Status='live', Symbol='Olfr978', Aliases='MOR223-3', description='olfactory receptor 978', other_designations='GA_x6K02T2PVTD-33692287-33693222|olfactory receptor MOR223-3', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39993812, end_position_on_the_genomic_accession=39994747, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259110 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259110, CurrentID=0, Status='live', Symbol='Olfr980', Aliases='MOR223-2', description='olfactory receptor 980', other_designations='GA_x6K02T2PVTD-33705428-33704496|olfactory receptor MOR223-2', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40005994, end_position_on_the_genomic_accession=40007029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259111 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259111, CurrentID=0, Status='live', Symbol='Olfr974', Aliases='MOR171-1', description='olfactory receptor 974', other_designations='GA_x6K02T2PVTD-33640290-33641222|olfactory receptor MOR171-1', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=39942262, end_position_on_the_genomic_accession=39943194, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259112 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259112, CurrentID=0, Status='live', Symbol='Olfr979', Aliases='MOR223-1', description='olfactory receptor 979', other_designations='GA_x6K02T2PVTD-33699706-33698771|olfactory receptor MOR223-1', map_location='9 A5-B', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=40000206, end_position_on_the_genomic_accession=40001250, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259113 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259113, CurrentID=0, Status='live', Symbol='Olfr577', Aliases='MOR7-2', description='olfactory receptor 577', other_designations='GA_x6K02T2PBJ9-5685322-5684384|olfactory receptor MOR7-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102973052, end_position_on_the_genomic_accession=102973990, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259114 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259114, CurrentID=0, Status='live', Symbol='Olfr570', Aliases='MOR8-3', description='olfactory receptor 570', other_designations='GA_x6K02T2PBJ9-5611412-5612350|olfactory receptor MOR8-3', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102813310, end_position_on_the_genomic_accession=102901380, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  259115 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259115, CurrentID=0, Status='live', Symbol='Olfr586', Aliases='MOR8-2, MOR8-7', description='olfactory receptor 586', other_designations='GA_x6K02T2PBJ9-5836380-5835439|olfactory receptor MOR8-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103121829, end_position_on_the_genomic_accession=103122782, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259116 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259116, CurrentID=0, Status='live', Symbol='Olfr559', Aliases='MOR11-1', description='olfactory receptor 559', other_designations='GA_x6K02T2PBJ9-5441154-5440198|olfactory receptor MOR11-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102723532, end_position_on_the_genomic_accession=102724488, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259117 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259117, CurrentID=0, Status='live', Symbol='Olfr560', Aliases='MOR14-1', description='olfactory receptor 560', other_designations='GA_x6K02T2PBJ9-5470572-5469628|olfactory receptor MOR14-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102752983, end_position_on_the_genomic_accession=102753927, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259119 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259119, CurrentID=0, Status='live', Symbol='Olfr578', Aliases='MOR7-1', description='olfactory receptor 578', other_designations='GA_x6K02T2PBJ9-5696486-5695545|olfactory receptor MOR7-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=102984221, end_position_on_the_genomic_accession=102989316, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259122 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259122, CurrentID=0, Status='live', Symbol='Olfr635', Aliases='MOR5-2', description='olfactory receptor 635', other_designations='GA_x6K02T2PBJ9-6713641-6714588|olfactory receptor MOR5-2', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103979176, end_position_on_the_genomic_accession=103980141, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259123 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259123, CurrentID=0, Status='live', Symbol='Olfr632', Aliases='MOR2-1', description='olfactory receptor 632', other_designations='GA_x6K02T2PBJ9-6671256-6672209|olfactory receptor MOR2-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103937382, end_position_on_the_genomic_accession=103938335, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259124 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259124, CurrentID=0, Status='live', Symbol='Olfr638', Aliases='MOR5-1', description='olfactory receptor 638', other_designations='GA_x6K02T2PBJ9-6737723-6738670|olfactory receptor MOR5-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104003259, end_position_on_the_genomic_accession=104023273, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  259125 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259125, CurrentID=0, Status='live', Symbol='Olfr644', Aliases='MOR13-1', description='olfactory receptor 644', other_designations='GA_x6K02T2PBJ9-6803062-6802118|olfactory receptor MOR13-1', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=104068085, end_position_on_the_genomic_accession=104069029, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259126 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259126, CurrentID=0, Status='live', Symbol='Olfr623', Aliases='MOR1-4', description='olfactory receptor 623', other_designations='GA_x6K02T2PBJ9-6384836-6383883|olfactory receptor MOR1-4', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103660244, end_position_on_the_genomic_accession=103661318, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  217951 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=217951, CurrentID=0, Status='live', Symbol='Tmem196', Aliases='Gm528', description='transmembrane protein 196', other_designations='', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=119945686, end_position_on_the_genomic_accession=120021245, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  259143 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259143, CurrentID=0, Status='live', Symbol='Olfr503', Aliases='MOR34-12, MOR34-8P, MOR34-9, Olfr1548', description='olfactory receptor 503', other_designations='GA_x6K02T2PBJ9-10874315-10875286|olfactory receptor 1548|olfactory receptor MOR34-8P|olfactory receptor MOR34-9', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=108544527, end_position_on_the_genomic_accession=108545498, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259144 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259144, CurrentID=0, Status='live', Symbol='Olfr456', Aliases='MOR257-7P', description='olfactory receptor 456', other_designations='GA_x6K02T2P3E9-5073878-5074816|olfactory receptor MOR257-7P', map_location='6 B2.1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=42486163, end_position_on_the_genomic_accession=42487214, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259145 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259145, CurrentID=0, Status='live', Symbol='Olfr1251', Aliases='GA_x5J8B7TT7VM-20-924, MOR231-15P', description='olfactory receptor 1251', other_designations='GA_x6K02T2Q125-51109312-51108356|MOR231-24_p', map_location='2 E1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=89666928, end_position_on_the_genomic_accession=89667884, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259147 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259147, CurrentID=0, Status='live', Symbol='Olfr723', Aliases='MOR247-3P, MOR247-4', description='olfactory receptor 723', other_designations='GA_x6K02T2PMLR-5600424-5599495|olfactory receptor MOR247-3P', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=49928563, end_position_on_the_genomic_accession=49931787, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  259151 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259151, CurrentID=0, Status='live', Symbol='Olfr1037', Aliases='MOR171-52, MOR256-34P', description='olfactory receptor 1037', other_designations='GA_x6K02T2Q125-47560740-47559775|olfactory receptor MOR256-34P', map_location='2 D', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=86084730, end_position_on_the_genomic_accession=86085854, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259159 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259159, CurrentID=0, Status='live', Symbol='Olfr628', Aliases='MOR22-1, MOR22-4P, Olfr1526-ps1', description='olfactory receptor 628', other_designations='GA_x6K02T2PBJ9-6457667-6458617|olfactory receptor 1526, pseudogene 1|olfactory receptor MOR22-1|olfactory receptor MOR22-4P', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=103731928, end_position_on_the_genomic_accession=103732878, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259161 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259161, CurrentID=0, Status='live', Symbol='Olfr688', Aliases='MOR40-16, MOR40-5-ps1, Olfr688', description='olfactory receptor 688', other_designations='GA_x6K02T2PBJ9-7917168-7918148|olfactory receptor 688, pseudogene 1|olfactory receptor MOR40-5|olfr688', map_location='7 E3', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=105288095, end_position_on_the_genomic_accession=105289060, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259162 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259162, CurrentID=0, Status='live', Symbol='Olfr427', Aliases='MOR105-7, MOR105-9', description='olfactory receptor 427', other_designations='GA_x6K02T2P20D-21075927-21074980|olfactory receptor MOR105-9', map_location='1 H3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=174099460, end_position_on_the_genomic_accession=174100407, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259163 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259163, CurrentID=0, Status='live', Symbol='Olfr1354', Aliases='EG257869, GA_x5J8B7TPDN2-1058-1265, GA_x6K02T03FR9-4826-3919, MOR139-5, MOR139-7, MOR185-8-ps1, Olfr233-ps1, Olfr1354', description='olfactory receptor 1354', other_designations='GA_x6K02T2QGN0-2895081-2894349|olfactory receptor 1354, pseudogene 1|olfactory receptor 233, pseudogene 1', map_location='10 C1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=78916842, end_position_on_the_genomic_accession=78917936, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  259165 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259165, CurrentID=0, Status='live', Symbol='Olfr814', Aliases='MOR113-5, MOR113-8', description='olfactory receptor 814', other_designations='GA_x6K02T2PULF-11553313-11552381|olfactory receptor MOR113-5', map_location='10 D3', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=129873823, end_position_on_the_genomic_accession=129874755, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  259172 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259172, CurrentID=0, Status='live', Symbol='Mfrp', Aliases='rd6', description='membrane frizzled-related protein', other_designations='membrane-type frizzled-related protein', map_location='9 24.6 cM', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=44101770, end_position_on_the_genomic_accession=44109187, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  545654 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545654, CurrentID=0, Status='live', Symbol='Gm13285', Aliases='OTTMUSG00000011282', description='predicted gene 13285', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=88800320, end_position_on_the_genomic_accession=88800868, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  21367 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21367, CurrentID=0, Status='live', Symbol='Cntn2', Aliases='D130012K04Rik, TAG-1, Tax', description='contactin 2', other_designations='TAX-1|axonal glycoprotein TAG-1|axonin|axonin-1|transient axonal glycoprotein 1|transiently-expressed axonal glycoprotein', map_location='1 E4', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=132509425, end_position_on_the_genomic_accession=132543251, orientation='minus', exon_count=26, OMIM='', no_hdr0=''),
  259277 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259277, CurrentID=0, Status='live', Symbol='Klk8', Aliases='BSP1, Nrpn, Prss19', description='kallikrein related-peptidase 8', other_designations='Brain Serine protease 1|NP|kallikrein 8|mK8|neuropsin|protease, serine, 19 (neuropsin)|serine protease 19', map_location='7 28.26 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=43797552, end_position_on_the_genomic_accession=43803822, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  259279 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259279, CurrentID=0, Status='live', Symbol='Tubgcp3', Aliases='GCP3, Spc98p', description='tubulin, gamma complex associated protein 3', other_designations='GCP-3|spindle pole body protein', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=12614277, end_position_on_the_genomic_accession=12672243, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  259300 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259300, CurrentID=0, Status='live', Symbol='Ehd2', Aliases='BC027084, C130052H20Rik', description='EH-domain containing 2', other_designations='', map_location='7 8.65 cM', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=15946966, end_position_on_the_genomic_accession=15967535, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  259301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259301, CurrentID=0, Status='live', Symbol='Leap2', Aliases='leap-2', description='liver-expressed antimicrobial peptide 2', other_designations='', map_location='11 B1.3', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=53422181, end_position_on_the_genomic_accession=53423136, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  259302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=259302, CurrentID=0, Status='live', Symbol='Srgap3', Aliases='AI452337, Arhgap14, D130026O08Rik, Gbi, MEGAP, WRP, mKIAA0411', description='SLIT-ROBO Rho GTPase activating protein 3', other_designations='SLIT-ROBO Rho GTPase activating protein 2|WAVE-associated Rac GTPase-activating protein|brain stress early protein Gbi|rho GTPase activating protein 14|rho GTPase-activating protein 14|srGAP2', map_location='6 E3', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=112717971, end_position_on_the_genomic_accession=112947444, orientation='minus', exon_count=24, OMIM='', no_hdr0=''),
  239827 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239827, CurrentID=0, Status='live', Symbol='Pigz', Aliases='F630022B06Rik', description='phosphatidylinositol glycan anchor biosynthesis, class Z', other_designations='GPI mannosyltransferase IV|GPI-MT-IV|PIG-Z|phosphatidylinositol glycan, class Z|phosphatidylinositol-glycan biosynthesis class Z protein', map_location='16 B2', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=31926073, end_position_on_the_genomic_accession=31946048, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  545662 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545662, CurrentID=0, Status='live', Symbol='B020004J07Rik', Aliases='', description='RIKEN cDNA B020004J07 gene', other_designations='novel protein similar to Prame proteins', map_location='4 C6', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=101834969, end_position_on_the_genomic_accession=101844022, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  545667 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545667, CurrentID=0, Status='live', Symbol='Fam159a', Aliases='OTTMUSG00000008243', description='family with sequence similarity 159, member A', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=108367777, end_position_on_the_genomic_accession=108383655, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  239839 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239839, CurrentID=0, Status='live', Symbol='Ccdc14', Aliases='G630039H03Rik', description='coiled-coil domain containing 14', other_designations='', map_location='16 B3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=34690567, end_position_on_the_genomic_accession=34725204, orientation='plus', exon_count=14, OMIM='', no_hdr0=''),
  545677 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545677, CurrentID=0, Status='live', Symbol='Gm12888', Aliases='OTTMUSG00000008911', description='predicted gene 12888', other_designations='', map_location='4', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=121316316, end_position_on_the_genomic_accession=121324917, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  13003 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=13003, CurrentID=0, Status='live', Symbol='Vcan', Aliases='5430420N07Rik, 9430051N09, Cspg2, DPEAAE, NG2, PG-M, PG-M(V0), PG-M(V1), hdf', description='versican', other_designations='PG-M core protein|Versican core protein precursor (Large fibroblast proteoglycan) (Chondroitin sulfate proteoglycan core protein 2) (PG-M)|chondroitin sulfate proteoglycan 2|chondroitin sulfate proteoglycan core protein 2|heart defect|large fibroblast proteoglycan', map_location='13 45.5 cM', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=89655309, end_position_on_the_genomic_accession=89742694, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  652925 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=652925, CurrentID=0, Status='live', Symbol='Tmem243', Aliases='4930420K17Rik, E030031F02Rik', description='transmembrane protein 243, mitochondrial', other_designations='', map_location='5 A1', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=9100737, end_position_on_the_genomic_accession=9118983, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  653016 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=653016, CurrentID=0, Status='live', Symbol='Gm7325', Aliases='EG653016, Esgp', description='predicted gene 7325', other_designations='', map_location='17 B3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=45600961, end_position_on_the_genomic_accession=45602204, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  239931 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=239931, CurrentID=0, Status='live', Symbol='Cldn17', Aliases='', description='claudin 17', other_designations='', map_location='16 C3.3', chromosome='16', genomic_nucleotide_accession_version='NC_000082.6', start_position_on_the_genomic_accession=88505807, end_position_on_the_genomic_accession=88506978, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  21524 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=21524, CurrentID=0, Status='live', Symbol='Tcra-V2', Aliases='', description='T cell receptor alpha, variable 2', other_designations='T-cell receptor alpha, variable 2', map_location='', chromosome='14', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  260296 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260296, CurrentID=0, Status='live', Symbol='Trim61', Aliases='2czf61, AU018704, C85201, E330039K03Rik, Rnf35', description='tripartite motif-containing 61', other_designations='ring finger protein 35', map_location='8 B3.1-3.2', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=65012976, end_position_on_the_genomic_accession=65018688, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  260297 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260297, CurrentID=0, Status='live', Symbol='Prrt1', Aliases='Bat-4, Bat4, D17H6S54E, DSPD1, G5b, NG5, ORF31', description='proline-rich transmembrane protein 1', other_designations='dispanin subfamily D member 1|synapse differentiation-induced protein 4', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=34629569, end_position_on_the_genomic_accession=34633134, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  260298 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260298, CurrentID=0, Status='live', Symbol='Fev', Aliases='Pet-1, Pet1, mPet-1', description='FEV (ETS oncogene family)', other_designations='ETS-domain transcription factor Pet-1|FEV (fifth Ewing variant)|PC12 ETS domain-containing transcription factor 1|PC12 ETS factor 1|pheochromocytoma 12 ETS (E26 transformation-specific)', map_location='1 C3', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=74881509, end_position_on_the_genomic_accession=74885408, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  260299 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260299, CurrentID=0, Status='live', Symbol='Cadm4', Aliases='Igdf4c, Igsf4c, Tsll2', description='cell adhesion molecule 4', other_designations='NECL-4|SynCAM 4|TSLC1-like 2|TSLC1-like protein 2|immunoglobulin superfamily member 4C|immunoglobulin superfamily, member 4C|membrane glycoprotein Tsll2|nectin-like protein 4', map_location='7 A3-B2', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=24482023, end_position_on_the_genomic_accession=24504533, orientation='plus', exon_count=9, OMIM='', no_hdr0=''),
  260301 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260301, CurrentID=0, Status='live', Symbol='Otos', Aliases='OTOSP', description='otospiralin', other_designations='organ of Corti 10 kDa protein', map_location='1 D', chromosome='1', genomic_nucleotide_accession_version='NC_000067.6', start_position_on_the_genomic_accession=92644218, end_position_on_the_genomic_accession=92648841, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  260302 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260302, CurrentID=0, Status='live', Symbol='Gga3', Aliases='C230037M19Rik, mKIAA0154', description='golgi associated, gamma adaptin ear containing, ARF binding protein 3', other_designations='golgi-localized, gamma ear-containing, ARF-binding protein 3', map_location='11 E2', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=115584255, end_position_on_the_genomic_accession=115603917, orientation='minus', exon_count=18, OMIM='', no_hdr0=''),
  260305 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260305, CurrentID=0, Status='live', Symbol='Nphp4', Aliases='4930564O18Rik, nmf192', description='nephronophthisis 4 (juvenile) homolog (human)', other_designations='nephroretinin', map_location='4 E2', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=152473881, end_position_on_the_genomic_accession=152563184, orientation='plus', exon_count=34, OMIM='', no_hdr0=''),
  260315 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260315, CurrentID=0, Status='live', Symbol='Nav3', Aliases='4732483H20Rik, 9630020C08Rik, POMFIL1, Pomfil1p, mKIAA0938, unc53H3', description='neuron navigator 3', other_designations='pore membrane and/or filament-interacting-like protein 1|steerin 3', map_location='10 D1', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=109681573, end_position_on_the_genomic_accession=110000391, orientation='minus', exon_count=40, OMIM='', no_hdr0=''),
  260396 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260396, CurrentID=0, Status='live', Symbol='Pain1', Aliases='Auto', description='pain 1', other_designations='', map_location='15 30.11 cM', chromosome='15', genomic_nucleotide_accession_version='', start_position_on_the_genomic_accession='', end_position_on_the_genomic_accession='', orientation='', exon_count='', OMIM='', no_hdr0=''),
  260408 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260408, CurrentID=0, Status='live', Symbol='Prss45', Aliases='Tespl, Tessp5', description='protease, serine 45', other_designations='inactive testis serine protease 5|protease, serine, 45|testis serine protease 5|trypsin-like protease p98', map_location='9 F3', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=110834588, end_position_on_the_genomic_accession=110841310, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  260409 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260409, CurrentID=0, Status='live', Symbol='Cdc42ep3', Aliases='3200001F04Rik, AA986861, Borg2, Cep3, UB1', description='CDC42 effector protein (Rho GTPase binding) 3', other_designations='binder of Rho GTPases 2', map_location='17 E3', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=79333964, end_position_on_the_genomic_accession=79355091, orientation='minus', exon_count=3, OMIM='', no_hdr0=''),
  260423 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=260423, CurrentID=0, Status='live', Symbol='Hist1h3f', Aliases='H3.2-221, Hist1', description='histone cluster 1, H3f', other_designations='histone 1, H3f|histone H3|histone gene complex 1', map_location='13 A2-A3', chromosome='13', genomic_nucleotide_accession_version='NC_000079.6', start_position_on_the_genomic_accession=23544052, end_position_on_the_genomic_accession=23544954, orientation='plus', exon_count=1, OMIM='', no_hdr0=''),
  108956 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=108956, CurrentID=0, Status='live', Symbol='Apol7c', Aliases='2210421G13Rik', description='apolipoprotein L 7c', other_designations='', map_location='15 E1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=77524852, end_position_on_the_genomic_accession=77533335, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  209448 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209448, CurrentID=0, Status='live', Symbol='Hoxc10', Aliases='AI506621, Hox-3.6', description='homeobox C10', other_designations='homeo box C10|homeobox protein Hox-3.6', map_location='15 58.03 cM', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=102966796, end_position_on_the_genomic_accession=102971898, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  240067 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240067, CurrentID=0, Status='live', Symbol='Zfp952', Aliases='BC057593, C920016K16Rik', description='zinc finger protein 952', other_designations='', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=32993108, end_position_on_the_genomic_accession=33005457, orientation='plus', exon_count=5, OMIM='', no_hdr0=''),
  240069 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240069, CurrentID=0, Status='live', Symbol='Morc2b', Aliases='4932411A10Rik, Tce6', description='microrchidia 2B', other_designations='Morc-like protein|morc2b', map_location='17 B1', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=33135588, end_position_on_the_genomic_accession=33150499, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  207742 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=207742, CurrentID=0, Status='live', Symbol='Rnf43', Aliases='4732452J19Rik', description='ring finger protein 43', other_designations='', map_location='11 C', chromosome='11', genomic_nucleotide_accession_version='NC_000077.6', start_position_on_the_genomic_accession=87662464, end_position_on_the_genomic_accession=87735539, orientation='plus', exon_count=11, OMIM='', no_hdr0=''),
  654309 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654309, CurrentID=0, Status='live', Symbol='Nrp', Aliases='', description='neural regeneration protein', other_designations='', map_location='12 D2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=87442848, end_position_on_the_genomic_accession=87444017, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  209462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=209462, CurrentID=0, Status='live', Symbol='Hace1', Aliases='1700042J16Rik, A730034A22Rik, BC025474', description='HECT domain and ankyrin repeat containing, E3 ubiquitin protein ligase 1', other_designations='HECT domain and ankyrin repeat-containing E3 ubiquitin-protein ligase 1', map_location='10 B2', chromosome='10', genomic_nucleotide_accession_version='NC_000076.6', start_position_on_the_genomic_accession=45577571, end_position_on_the_genomic_accession=45712345, orientation='plus', exon_count=27, OMIM='', no_hdr0=''),
  654362 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654362, CurrentID=0, Status='live', Symbol='Dear1', Aliases='Dear, Despr', description='dual endothelin 1/angiotensin II receptor 1', other_designations='', map_location='3 F1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=84965056, end_position_on_the_genomic_accession=84965580, orientation='minus', exon_count=1, OMIM='', no_hdr0=''),
  545975 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545975, CurrentID=0, Status='live', Symbol='Cers3', Aliases='4930550L11Rik, Gm488, Lass3, T3L', description='ceramide synthase 3', other_designations='LAG1 homolog, ceramide synthase 3|LAG1 longevity assurance homolog 3|dihydroceramide synthase 3|longevity assurance homolog 3|related to TRH3', map_location='7 C', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=66733178, end_position_on_the_genomic_accession=66823697, orientation='plus', exon_count=13, OMIM='', no_hdr0=''),
  654457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654457, CurrentID=0, Status='live', Symbol='Defb26', Aliases='EG654457', description='defensin beta 26', other_designations='beta-defensin 26', map_location='2 H1', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=152507756, end_position_on_the_genomic_accession=152511736, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  654458 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654458, CurrentID=0, Status='live', Symbol='Defb43', Aliases='BD-43, EG654458, mBD-43', description='defensin beta 43', other_designations='', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=63011771, end_position_on_the_genomic_accession=63018088, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  654462 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654462, CurrentID=0, Status='live', Symbol='Kncn', Aliases='Kino, L5', description='kinocilin', other_designations='', map_location='4 D1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=115884400, end_position_on_the_genomic_accession=115887964, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  654465 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654465, CurrentID=0, Status='live', Symbol='Defb47', Aliases='Defb41, EG654465', description='defensin beta 47', other_designations='defensin beta 41', map_location='14', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=62992375, end_position_on_the_genomic_accession=63001180, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  654470 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654470, CurrentID=0, Status='live', Symbol='Tctn1', Aliases='G730031O11Rik, Tect1', description='tectonic family member 1', other_designations='', map_location='5 F', chromosome='5', genomic_nucleotide_accession_version='NC_000071.6', start_position_on_the_genomic_accession=122239495, end_position_on_the_genomic_accession=122264489, orientation='minus', exon_count=14, OMIM='', no_hdr0=''),
  654498 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654498, CurrentID=0, Status='live', Symbol='Hhla1', Aliases='F930104E18Rik', description='HERV-H LTR-associating 1', other_designations='', map_location='15 D1-D2', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=65922443, end_position_on_the_genomic_accession=65976804, orientation='minus', exon_count=17, OMIM='', no_hdr0=''),
  101055709 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=101055709, CurrentID=0, Status='live', Symbol='Zfp264', Aliases='', description='zinc finger protein 264', other_designations='zinc finger protein 264 pseudogene', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=6977413, end_position_on_the_genomic_accession=6985640, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  449521 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=449521, CurrentID=0, Status='live', Symbol='Zfp213', Aliases='BC038178, D17Ertd197e', description='zinc finger protein 213', other_designations='', map_location='17 11.67 cM', chromosome='17', genomic_nucleotide_accession_version='NC_000083.6', start_position_on_the_genomic_accession=23556767, end_position_on_the_genomic_accession=23564226, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  240185 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=240185, CurrentID=0, Status='live', Symbol='9430020K01Rik', Aliases='1110056P05Rik, Gm328, Jcad, mKIAA1462', description='RIKEN cDNA 9430020K01 gene', other_designations='', map_location='18 A1', chromosome='18', genomic_nucleotide_accession_version='NC_000084.6', start_position_on_the_genomic_accession=4590790, end_position_on_the_genomic_accession=4682869, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  546038 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546038, CurrentID=0, Status='live', Symbol='Spag11b', Aliases='EG546038, Ep2c/h, Spag11c/h', description='sperm associated antigen 11B', other_designations='beta-defensin-like peptide|sperm-associated antigen 11c/h', map_location='8 A1.3', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=19140759, end_position_on_the_genomic_accession=19143010, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
  654795 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654795, CurrentID=0, Status='live', Symbol='Sdr39u1', Aliases='2310014G06Rik', description='short chain dehydrogenase/reductase family 39U, member 1', other_designations='short-chain dehydrogenase/reductase family 39U member 1', map_location='14 C1', chromosome='14', genomic_nucleotide_accession_version='NC_000080.6', start_position_on_the_genomic_accession=55897285, end_position_on_the_genomic_accession=55900281, orientation='minus', exon_count=6, OMIM='', no_hdr0=''),
  654801 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654801, CurrentID=0, Status='live', Symbol='Zfp784', Aliases='C730049F21Rik, Znf784', description='zinc finger protein 784', other_designations='', map_location='7 A1', chromosome='7', genomic_nucleotide_accession_version='NC_000073.6', start_position_on_the_genomic_accession=5034446, end_position_on_the_genomic_accession=5038446, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  654812 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654812, CurrentID=0, Status='live', Symbol='Angptl7', Aliases='', description='angiopoietin-like 7', other_designations='angiopoietin-like protein 7', map_location='4 E1', chromosome='4', genomic_nucleotide_accession_version='NC_000070.6', start_position_on_the_genomic_accession=148492979, end_position_on_the_genomic_accession=148500462, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  654818 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654818, CurrentID=0, Status='live', Symbol='Smco3', Aliases='C030030A07Rik', description='single-pass membrane protein with coiled-coil domains 3', other_designations='', map_location='6 G1', chromosome='6', genomic_nucleotide_accession_version='NC_000072.6', start_position_on_the_genomic_accession=136829931, end_position_on_the_genomic_accession=136835450, orientation='minus', exon_count=2, OMIM='', no_hdr0=''),
  654821 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654821, CurrentID=0, Status='live', Symbol='Gcnt7', Aliases='A330041C17Rik, gnt', description='glucosaminyl (N-acetyl) transferase family member 7', other_designations='beta 1,6-N-acetylglucosaminyltransferase|likely orthologue of H. sapiens chromosome 20 open reading frame 105 (C20orf105)', map_location='2 H3', chromosome='2', genomic_nucleotide_accession_version='NC_000068.7', start_position_on_the_genomic_accession=172447120, end_position_on_the_genomic_accession=172458596, orientation='minus', exon_count=4, OMIM='', no_hdr0=''),
  654824 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=654824, CurrentID=0, Status='live', Symbol='Ankrd37', Aliases='', description='ankyrin repeat domain 37', other_designations='', map_location='8 B1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=45996908, end_position_on_the_genomic_accession=46000237, orientation='minus', exon_count=5, OMIM='', no_hdr0=''),
  546049 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546049, CurrentID=0, Status='live', Symbol='C330021F23Rik', Aliases='2610027C06Rik, Rps23r1, Rps23rg1', description='RIKEN cDNA C330021F23 gene', other_designations='', map_location='8 A1.1', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=3567998, end_position_on_the_genomic_accession=3584776, orientation='plus', exon_count=6, OMIM='', no_hdr0=''),
  546055 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546055, CurrentID=0, Status='live', Symbol='Adam39', Aliases='1700056P18Rik', description='a disintegrin and metallopeptidase domain 39', other_designations='', map_location='8 A4', chromosome='8', genomic_nucleotide_accession_version='NC_000074.6', start_position_on_the_genomic_accession=40823009, end_position_on_the_genomic_accession=40826861, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  633457 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=633457, CurrentID=0, Status='live', Symbol='Ighv3-5', Aliases='C72-3A1, EG633457, Gm7112', description='immunoglobulin heavy variable 3-5', other_designations='immunoglobulin heavy variable V3-5', map_location='12 F2', chromosome='12', genomic_nucleotide_accession_version='NC_000078.6', start_position_on_the_genomic_accession=114262652, end_position_on_the_genomic_accession=114262950, orientation='minus', exon_count=0, OMIM='', no_hdr0=''),
  545554 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=545554, CurrentID=0, Status='live', Symbol='Ankrd34a', Aliases='AW060394, Ankrd34', description='ankyrin repeat domain 34A', other_designations='ankyrin repeat domain 34', map_location='3 F2.1', chromosome='3', genomic_nucleotide_accession_version='NC_000069.6', start_position_on_the_genomic_accession=96594826, end_position_on_the_genomic_accession=96600126, orientation='plus', exon_count=4, OMIM='', no_hdr0=''),
  393082 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=393082, CurrentID=0, Status='live', Symbol='Mettl7a2', Aliases='Gm921, UbiE2, Ubie', description='methyltransferase like 7A2', other_designations='', map_location='15 F1', chromosome='15', genomic_nucleotide_accession_version='NC_000081.6', start_position_on_the_genomic_accession=100353200, end_position_on_the_genomic_accession=100361819, orientation='plus', exon_count=2, OMIM='', no_hdr0=''),
  546118 : NtData(tax_id=10090, Org_name='Mus musculus', GeneID=546118, CurrentID=0, Status='live', Symbol='Ubtfl1', Aliases='B020006M18Rik, Hmgpi', description='upstream binding transcription factor, RNA polymerase I-like 1', other_designations='HMG-box preimplantation embryo-specific protein|HMG-box protein, preimplantation-embryo-specific', map_location='9 A2', chromosome='9', genomic_nucleotide_accession_version='NC_000075.6', start_position_on_the_genomic_accession=18404418, end_position_on_the_genomic_accession=18411502, orientation='plus', exon_count=3, OMIM='', no_hdr0=''),
}

